/*
 * Common code for uncompress/gunzip/bunzip2/unlzma
 *
 * Copyright Vladimir Oleynik <dzo@simtreas.ru> (c) 2016
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */

/* convert zero arg usage to "-" */
#define COMPLEMENTARY "\0-0"
#include "opt_gen.h"
OPT_GEN(
opt(c,OPT_STDOUT),
opt(f,OPT_FORCE),
opt(d,OPT_UN),
opt(t,OPT_TEST)
)
#include "opt_gen.h"

#include "busybox.h"
#include "bb_xopen.h"
#include <string.h>
#include <unistd.h>

static int tst_ext(const char *ext_file, const char *ext, const char *filename)
{
	if (ext_file == NULL || (ext != NULL && strcmp(ext_file, ext) != 0)) {
		bb_error_msg("%s: unknow suffix -- ignored", filename);
		return -1;
	}
	return 0;
}

static int drop_ext_open(const char *filename, const char *extension, char **pout)
{
	struct stat stat_buf;
	char *e, *e0 = xstrdup(filename);

	stat(filename, &stat_buf);
	*pout = e0;

	e = strrchr(e0, '.');
	if(tst_ext(e, NULL, filename))
		return -1;

#ifdef CONFIG_GUNZIP
	if(extension == NULL) {
#ifdef CONFIG_FEATURE_GUNZIP_UNCOMPRESS
		if (strcmp(e, ".Z") == 0) {
			*e = '\0';
		} else
#endif
		if (strcmp(e, ".tgz") == 0) {
			e[2] = 'a';
			e[3] = 'r';
		} else {
			extension = ".gz";
			goto tst_gz;
		}
	} else {
		tst_gz:
#endif
		if (tst_ext(e, extension, filename))
			return -1;
		*e = 0;
#ifdef CONFIG_GUNZIP
	}
#endif
	return bb_xopen(e0, O_WRONLY | O_CREAT | BB_OPEN_WARN, stat_buf.st_mode);
}

int unx_main(const char *ext, int flg_stdout_name,
		int (*transformer)(int src_fd, int dst_fd), const char *magic_str)
{
	char **argv;
	int status = EXIT_SUCCESS;
	int flags;
	const char *emsg;

	if(bss_flags == 0)
		flags = bb_xgetopt_ulflags(opt_str);
	else
		flags = bss_flags;      /* from gzip -d */
	if(flg_stdout_name)
		flags |= OPT_STDOUT;

	for (argv = Gargv; *argv != NULL; argv++) {
		const char *compressed_file = *argv;
		const char *delete_path = NULL;
		char *uncompressed_file = NULL;
		int dst_fd;
		int flg_stdin = compressed_file[0] == '-' && compressed_file[1] == '\0';
		int src_fd;

		if(flg_stdin) {
			src_fd = STDIN_FILENO;
			compressed_file = "<stdin>";
		} else {
			src_fd = bb_xopen(compressed_file, O_RDONLY | BB_OPEN_WARN);
			if(src_fd < 0)
				continue;
		}
		/* Check that the input is sane.  */
		if (isatty(src_fd) && (flags & OPT_FORCE) == 0)
			bb_error_msg_and_die("compressed data not read from terminal.  Use -f to force it.");

		if((flags & OPT_TEST) != 0) {
			/* inflate_gunzip use two filenum always :-( */
			dst_fd = bb_xopen(bb_dev_null, O_WRONLY | BB_OPEN_ERR | BB_OPEN_CREAT_0777);
		} else if (flg_stdin || (flags & OPT_STDOUT)) {
			dst_fd = STDOUT_FILENO;
		} else {
			dst_fd = drop_ext_open(compressed_file, ext, &uncompressed_file);
			if(dst_fd < 0)
				goto cont1;
			/* If succeeds remove the old file */
			delete_path = compressed_file;
		}

		if(magic_str != NULL) {
			const unsigned char *m;
			unsigned char mc;
			for(m = (const unsigned char *)magic_str; *m != '\0'; m++) {
				mc = bb_xread_char(src_fd);
				if (mc != *m) {
#if defined(CONFIG_GUNZIP) && defined(CONFIG_FEATURE_GUNZIP_UNCOMPRESS)
					if(m == (const unsigned char *)magic_str || (ext == NULL && mc != 0x9d)) {
#endif
						emsg = "%s: invalid magic";
						goto cont;
#if defined(CONFIG_GUNZIP) && defined(CONFIG_FEATURE_GUNZIP_UNCOMPRESS)
					}
					transformer = bb_uncompress;
#endif
				}
			}
		}

		/* do the decompression */
		status = (*transformer)(src_fd, dst_fd);

		if (status != EXIT_SUCCESS) {
			emsg = "%s: error inflating";
cont:
			bb_error_msg(emsg, compressed_file);
			/* failed, remove the uncomressed file instead of compressed file */
			delete_path = uncompressed_file;
		}

		if (dst_fd != STDOUT_FILENO)
			close(dst_fd);
cont1:
		if (src_fd != STDIN_FILENO)
			close(src_fd);

		/* delete_path will be NULL if in test mode or from stdin */
		if (delete_path && unlink(delete_path) == -1)
			bb_perror_msg("Couldn't remove %s", delete_path);

		free(uncompressed_file);
	}

	return status;
}
