/*
 * Mini date implementation for busybox
 *
 * by Matthew Grant <grantma@anathoth.gen.nz>
 *
 * iso-format handling added by Robert Griebl <griebl@gmx.de>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
*/
#ifdef APPLET_USAGE
#define date_trivial_usage \
	"[OPTION]... [MMDDhhmm[[CC]YY][.ss]] [+FORMAT]"
#define date_full_usage \
	"Displays the current time in the given FORMAT, or sets the system date.\n" \
	"\nOptions:\n" \
	"\t-R\t\tOutputs RFC-822 compliant date string\n" \
	"\t-d STRING\tDisplays time described by STRING, not `now'\n" \
	USE_FEATURE_DATE_ISOFMT( \
		"\t-I[TIMESPEC]\tOutputs an ISO-8601 compliant date/time string\n" \
		"\t\t\tTIMESPEC=`date' (or missing) for date only,\n" \
		"\t\t\t`hours', `minutes', or `seconds' for date and,\n" \
		"\t\t\ttime to the indicated precision\n" \
		"\t-D hint\t\tUse 'hint' as date format, via strptime()\n" \
	) \
	"\t-s\t\tSets time described by STRING\n" \
	"\t-r FILE\t\tDisplays the last modification time of FILE\n" \
	"\t-u\t\tPrints or sets Coordinated Universal Time"
#else

#include "busybox.h"
#include <sys/time.h>
#include <unistd.h>
#include <time.h>
#include <string.h>

#define COMPLEMENTARY "\0?d--s:s--d"
#include "opt_gen.h"
OPT_GEN(
opt(R, DATE_OPT_RFC2822),
opt(s, DATE_OPT_SET, arg),
opt(u, DATE_OPT_UTC),
opt(d, DATE_OPT_DATE, arg),
opt(r, DATE_OPT_REFERENCE, arg),
opt_opt(ENABLE_FEATURE_DATE_ISOFMT, D, DATE_OPT_HINT, arg),
opt_opt(ENABLE_FEATURE_DATE_ISOFMT, I, DATE_OPT_TIMESPEC, arg, optional)
)
#include "opt_gen.h"

/* This 'date' command supports only 2 time setting formats,
   all the GNU strftime stuff (its in libc, lets use it),
   setting time using UTC and displaying int, as well as
   an RFC 822 complient date output for shell scripting
   mail commands */

/* Input parsing code is always bulky - used heavy duty libc stuff as
   much as possible, missed out a lot of bounds checking */

/* Default input handling to save surprising some people */

static struct tm *date_conv_time(struct tm *tm_time, const char *t_string)
{
	int nr;
	char *cp;

	nr = sscanf(t_string, "%2d%2d%2d%2d%d", &(tm_time->tm_mon),
				&(tm_time->tm_mday), &(tm_time->tm_hour), &(tm_time->tm_min),
				&(tm_time->tm_year));

	if (nr < 4 || nr > 5) {
		bb_error_msg_and_die(bb_msg_invalid_date, t_string);
	}

	cp = strchr(t_string, '.');
	if (cp) {
		nr = sscanf(cp + 1, "%2d", &(tm_time->tm_sec));
		if (nr != 1) {
			bb_error_msg_and_die(bb_msg_invalid_date, t_string);
		}
	}

	/* correct for century  - minor Y2K problem here? */
	if (tm_time->tm_year >= 1900) {
		tm_time->tm_year -= 1900;
	}
	/* adjust date */
	tm_time->tm_mon -= 1;

	return (tm_time);

}


/* The new stuff for LRP */

static struct tm *date_conv_ftime(struct tm *tm_time, const char *t_string)
{
	struct tm t;

	/* Parse input and assign appropriately to tm_time */

	if (t = *tm_time, sscanf(t_string, "%d:%d:%d", &t.tm_hour, &t.tm_min,
						 &t.tm_sec) == 3) {
		/* no adjustments needed */
	} else if (t = *tm_time, sscanf(t_string, "%d:%d", &t.tm_hour,
								&t.tm_min) == 2) {
		/* no adjustments needed */
	} else if (t = *tm_time, sscanf(t_string, "%d.%d-%d:%d:%d", &t.tm_mon,
						&t.tm_mday, &t.tm_hour,
						&t.tm_min, &t.tm_sec) == 5) {
		/* Adjust dates from 1-12 to 0-11 */
		t.tm_mon -= 1;
	} else if (t = *tm_time, sscanf(t_string, "%d.%d-%d:%d", &t.tm_mon,
						&t.tm_mday,
						&t.tm_hour, &t.tm_min) == 4) {
		/* Adjust dates from 1-12 to 0-11 */
		t.tm_mon -= 1;
	} else if (t = *tm_time, sscanf(t_string, "%d.%d.%d-%d:%d:%d", &t.tm_year,
						&t.tm_mon, &t.tm_mday,
						&t.tm_hour, &t.tm_min,
							&t.tm_sec) == 6) {
		t.tm_year -= 1900;      /* Adjust years */
		t.tm_mon -= 1;  /* Adjust dates from 1-12 to 0-11 */
	} else if (t = *tm_time, sscanf(t_string, "%d.%d.%d-%d:%d", &t.tm_year,
						&t.tm_mon, &t.tm_mday,
						&t.tm_hour, &t.tm_min) == 5) {
		t.tm_year -= 1900;      /* Adjust years */
		t.tm_mon -= 1;  /* Adjust dates from 1-12 to 0-11 */
	} else {
		bb_error_msg_and_die(bb_msg_invalid_date, t_string);
	}
	*tm_time = t;
	return (tm_time);
}


CALC_BSS struct BB_COMMON_BSS
{
	char bb_common_bufsiz[204];
} bb_common_bss;
#define bb_common_bufsiz bb_common_bss.bb_common_bufsiz

int date_main(void)
{
	char *a;
	char *date_str = NULL;
	char *date_fmt = NULL;
	int set_time;
	int utc;
	time_t tm;
	unsigned long opt;
	struct tm tm_time;
	char *filename = NULL;

	int ifmt = 0;
	char *isofmt_arg;
	char *hintfmt_arg;

	opt = bb_xgetopt_ulflags(opt_str, &date_str, &date_str, &filename
					USE_FEATURE_DATE_ISOFMT(, &hintfmt_arg, &isofmt_arg));
	a = *Gargv;
	set_time = opt & DATE_OPT_SET;
	utc = opt & DATE_OPT_UTC;
	if (utc && putenv("TZ=UTC0") != 0)
		bb_perror_msg_and_die("putenv");

	if(ENABLE_FEATURE_DATE_ISOFMT && (opt & DATE_OPT_TIMESPEC)) {
		if (!isofmt_arg) {
			ifmt = 1;
		} else {
			static const char * const isoformats[] = {
				"date","hours","minutes","seconds"};
			for(ifmt = 4; ifmt; ifmt--)
				if(!strcmp(isofmt_arg,isoformats[ifmt-1]))
					break;
			if (!ifmt)
				bb_show_usage();
		}
	}

	if (a != NULL && a[0] == '+') {
		date_fmt = a+1;         /* Skip over the '+' */
	} else if (date_str == NULL) {
		set_time = 1;
		date_str = a;
	}

	/* Now we have parsed all the information except the date format
	   which depends on whether the clock is being set or read */

	if(filename) {
		struct stat statbuf;
		if(stat(filename,&statbuf))
			bb_perror_msg_and_die("File '%s' not found.", filename);
		tm=statbuf.st_mtime;
	} else time(&tm);
	memcpy(&tm_time, localtime(&tm), sizeof(tm_time));
	/* Zero out fields - take her back to midnight! */
	if (date_str != NULL) {
		tm_time.tm_sec = 0;
		tm_time.tm_min = 0;
		tm_time.tm_hour = 0;

		/* Process any date input to UNIX time since 1 Jan 1970 */
		if (ENABLE_FEATURE_DATE_ISOFMT && (opt & DATE_OPT_HINT)) {
			strptime(date_str, hintfmt_arg, &tm_time);
		} else if (strchr(date_str, ':') != NULL) {
			date_conv_ftime(&tm_time, date_str);
		} else {
			date_conv_time(&tm_time, date_str);
		}

		/* Correct any day of week and day of year etc. fields */
		tm_time.tm_isdst = -1;  /* Be sure to recheck dst. */
		tm = mktime(&tm_time);
		if (tm < 0) {
			bb_error_msg_and_die(bb_msg_invalid_date, date_str);
		}
		if (utc && putenv("TZ=UTC0") != 0)
			bb_perror_msg_and_die("putenv");

		/* if setting time, set it */
		if (set_time) {
			struct timespec ts;
			memset(&ts, 0, sizeof(ts));
			ts.tv_sec = tm;
			if(clock_settime (CLOCK_REALTIME, &ts) < 0)
				bb_perror_msg("cannot set date");
		}
	}

	/* Display output */

	/* Deal with format string */
	if (date_fmt == NULL) {
		/* Start with the default case */

		date_fmt = (opt & DATE_OPT_RFC2822 ?
					(utc ? "%a, %d %b %Y %H:%M:%S GMT" :
					"%a, %d %b %Y %H:%M:%S %z") :
					"%a %b %e %H:%M:%S %Z %Y");

		if (ENABLE_FEATURE_DATE_ISOFMT) {
			if (ifmt == 4)
				date_fmt = utc ? "%Y-%m-%dT%H:%M:%SZ" : "%Y-%m-%dT%H:%M:%S%z";
			else if (ifmt == 3)
				date_fmt = utc ? "%Y-%m-%dT%H:%MZ" : "%Y-%m-%dT%H:%M%z";
			else if (ifmt == 2)
				date_fmt = utc ? "%Y-%m-%dT%HZ" : "%Y-%m-%dT%H%z";
			else if (ifmt == 1)
				date_fmt = "%Y-%m-%d";
		}
	}

	if (*date_fmt == '\0') {

		/* With no format string, just print a blank line */

		*bb_common_bufsiz = 0;
	} else {

		/* Handle special conversions */

		if (strncmp(date_fmt, "%f", 2) == 0) {
			date_fmt = "%Y.%m.%d-%H:%M:%S";
		}

		/* Generate output string */
		strftime(bb_common_bufsiz, 200, date_fmt, &tm_time);
	}
	puts(bb_common_bufsiz);

	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
