/*
 * Mini dd implementation for busybox
 *
 *
 * Copyright (C) 2000,2001  Matt Kraai
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */
#ifdef APPLET_USAGE
#define dd_trivial_usage \
	"[if=FILE] [of=FILE] [bs=N] [count=N] [skip=N]\n" \
	"\t  [seek=N] [conv=notrunc|noerror|sync]"
#define dd_full_usage \
	"Copy a file, converting and formatting according to options\n\n" \
	"\tif=FILE\t\tread from FILE instead of stdin\n" \
	"\tof=FILE\t\twrite to FILE instead of stdout\n" \
	"\tbs=N\t\tread and write N bytes at a time\n" \
	"\tcount=N\t\tcopy only N input blocks\n" \
	"\tskip=N\t\tskip N input blocks\n" \
	"\tseek=N\t\tskip N output blocks\n" \
	"\tconv=notrunc\tdon't truncate output file\n" \
	"\tconv=noerror\tcontinue after read errors\n" \
	"\tconv=sync\tpad blocks with zeros\n" \
	"\n" \
	"Numbers may be suffixed by c (x1), w (x2), b (x512), kD (x1000), k (x1024),\n" \
	"MD (x1000000), M (x1048576), GD (x1000000000) or G (x1073741824)"
#else

#include "busybox.h"
#include <unistd.h>
#include <string.h>
#include "bb_xopen.h"
#include "bb_sig.h"


CALC_BSS struct BB_COMMON_BSS
{
	off_t out_full;
	off_t out_part;
	off_t in_full;
	off_t in_part;
} bb_common_bss;
#define out_full bb_common_bss.out_full
#define out_part bb_common_bss.out_part
#define in_full bb_common_bss.in_full
#define in_part bb_common_bss.in_part

#define cont_l_fmt "%" P_OFFT "d"

static void dd_output_status(int ATTRIBUTE_UNUSED cur_signal)
{

	fprintf(stderr, cont_l_fmt"+"cont_l_fmt" records in\n"
			cont_l_fmt"+"cont_l_fmt" records out\n",
			in_full, in_part, out_full, out_part);
}

#define DD_count_flag   1
#define DD_sync_flag    2
#define DD_noerror_flag 4
#define DD_notrunc_flag 8

static unsigned long parse_number(const char *a)
{
	return bb_xparse_ul_number(10 | BB_PN_BKM_SUFFIXES, a);
}

int dd_main(void)
{
	char **argv;
	off_t count = 0;
	size_t bs = 512;
	ssize_t n;
	off_t seek = 0;
	off_t skip = 0;
	int flag = 0;
	int ifd = STDIN_FILENO;
	int ofd = STDOUT_FILENO;
	const char *infile = bb_msg_standard_input;
	const char *outfile = bb_msg_standard_output;
	char *buf;

	bb_signal((1 << SIGUSR1), dd_output_status);

	for (argv = Gargv + 1; *argv; argv++) {
		if (strncmp("bs=", *argv, 3) == 0)
			bs = parse_number(argv[0]+3);
		else if (strncmp("count=", *argv, 6) == 0) {
			count = parse_number(argv[0]+6);
			flag |= DD_count_flag;
		} else if (strncmp("seek=", *argv, 5) == 0)
			seek = parse_number(argv[0]+5);
		else if (strncmp("skip=", *argv, 5) == 0)
			skip = parse_number(argv[0]+5);
		else if (strncmp("if=", *argv, 3) == 0)
			infile = argv[0]+3;
		else if (strncmp("of=", *argv, 3) == 0)
			outfile = argv[0]+3;
		else if (strncmp("conv=", *argv, 5) == 0) {
			buf = argv[0]+5;
			while (1) {
				if (strncmp("notrunc", buf, 7) == 0) {
					flag |= DD_notrunc_flag;
					buf += 7;
				} else if (strncmp("sync", buf, 4) == 0) {
					flag |= DD_sync_flag;
					buf += 4;
				} else if (strncmp("noerror", buf, 7) == 0) {
					flag |= DD_noerror_flag;
					buf += 7;
				} else {
					bb_error_msg_and_die("invalid conversion `%s'", argv[0]+5);
				}
				if (buf[0] == '\0')
					break;
				if (buf[0] == ',')
					buf++;
			}
		} else
			bb_show_usage();
	}

	buf = xmalloc(bs);

	if (infile != bb_msg_standard_input)
		ifd = bb_xopen(infile, O_RDONLY | BB_OPEN_ERR);

	if (outfile != bb_msg_standard_output) {
		int oflag = O_WRONLY | O_CREAT;

		if (!seek && (flag & DD_notrunc_flag) == 0) {
			oflag |= O_TRUNC;
		}

		ofd = bb_xopen(outfile, oflag | BB_OPEN_ERR, 0666);

		if (seek && (flag & DD_notrunc_flag) == 0) {
			if (ftruncate(ofd, seek * bs) < 0) {
				struct stat st;

				if (fstat (ofd, &st) < 0 || S_ISREG (st.st_mode) ||
						S_ISDIR (st.st_mode)) {
					bb_perror_msg_and_die("%s", outfile);
				}
			}
		}
	}

	if (skip) {
		if (lseek(ifd, skip * bs, SEEK_CUR) < 0)
			bb_perror_msg_and_die("%s", infile);
	}

	if (seek) {
		if (lseek(ofd, seek * bs, SEEK_CUR) < 0)
			bb_perror_msg_and_die("%s", outfile);
	}

	while ((flag & DD_count_flag) == 0 || (in_full + in_part) != count) {
		if ((flag & DD_noerror_flag) != 0) {
			/* Pre-zero the buffer when doing the noerror thing */
			memset(buf, '\0', bs);
		}
		n = safe_read(ifd, buf, bs);
		if (n < 0) {
			if ((flag & DD_noerror_flag) != 0) {
				n = bs;
				bb_perror_msg("%s", infile);
			} else {
				bb_perror_msg_and_die("%s", infile);
			}
		}
		if (n == 0)
			break;
		if ((size_t)n == bs)
			in_full++;
		 else
			in_part++;
		if ((flag & DD_sync_flag) != 0) {
			memset(buf + n, '\0', bs - n);
			n = bs;
		}
		n = bb_full_write(ofd, buf, n);
		if (n < 0)
			bb_perror_msg_and_die("%s", outfile);
		if ((size_t)n == bs)
			out_full++;
		 else
			out_part++;
	}

	if (close (ifd) < 0)
		bb_perror_msg_and_die("%s", infile);

	if (close (ofd) < 0)
		bb_perror_msg_and_die("%s", outfile);

	dd_output_status(0);

	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
