/*
 * Mini df implementation for busybox
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 * based on original code by (I think) Bruce Perens <bruce@pixar.com>.
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */

/* BB_AUDIT SUSv3 _NOT_ compliant -- options -P and -t missing.  Also blocksize. */
/* http://www.opengroup.org/onlinepubs/007904975/utilities/df.html */

/* Mar 16, 2003      Manuel Novoa III   (mjn3@codepoet.org)
 *
 * Size reduction.  Removed floating point dependency.  Added error checking
 * on output.  Output stats on 0-sized filesystems if specifically listed on
 * the command line.  Properly round *-blocks, Used, and Available quantities.
 */
#ifdef APPLET_USAGE
#define df_trivial_usage \
	"[-ki" USE_FEATURE_HUMAN_READABLE("hm") "] [FILESYSTEM ...]"
#define df_full_usage \
	"Print the filesystem space used and space available.\n\n" \
	"Options:\n" \
	"\t-k\tprint sizes in kilobytes(default)\n" \
	"\t-i\tprint inodes" \
	USE_FEATURE_HUMAN_READABLE( \
	"\n\t-h\tprint sizes in human readable format (e.g., 1K 243M 2G )\n" \
	"\t-m\tprint sizes in megabytes")
#else

#define NEED_MNTENT
#include "busybox.h"
#include <string.h>
#include <unistd.h>
#include <mntent.h>
#include <sys/vfs.h>

#ifdef CONFIG_FEATURE_HUMAN_READABLE
# define COMPLEMENTARY "\0h-km:k-hm:m-hk"
#endif
#include "opt_gen.h"
OPT_GEN(
opt_simple(k),
opt(i, INODES_OPT),
opt_opt(ENABLE_FEATURE_HUMAN_READABLE, h, HUMAN_OPT),
opt_opt(ENABLE_FEATURE_HUMAN_READABLE, m, MEGABYTE_OPT)
)
#include "opt_gen.h"


#ifndef CONFIG_FEATURE_HUMAN_READABLE
static long kscale(long b, long bs)
{
	return ( b * (long long) bs + KILOBYTE/2 ) / KILOBYTE;
}
#endif

extern int df_main(void)
{
	char **argv;
	long blocks_used;
#ifdef CONFIG_FEATURE_HUMAN_READABLE
	unsigned long df_disp_hr = KILOBYTE;
	char hr_buf[32];
#endif
	int status = EXIT_SUCCESS;
	unsigned long opt;
	FILE *mount_table;
	struct mntent *mount_entry;
	struct statfs s;
	static const char hdr_1k[] = "1k-blocks"; /* default display is kilobytes */
	const char *disp_units_hdr = hdr_1k;
	int show_inode = 0;
	char percent[6];

	opt =
#ifdef CONFIG_FEATURE_HUMAN_READABLE
	bb_xgetopt_ulflags(opt_str);
#else
	bb_getopt_ulflags(opt_str);
#endif
	if(opt & INODES_OPT) {
		disp_units_hdr = "   Inodes";
		show_inode = 1;
	}
#ifdef CONFIG_FEATURE_HUMAN_READABLE
	if(opt & HUMAN_OPT) {
		df_disp_hr = 0;
		disp_units_hdr = "     Size";
	}
	if(opt & MEGABYTE_OPT) {
		df_disp_hr = MEGABYTE;
		disp_units_hdr = "1M-blocks";
	}
#endif
	argv = Gargv;
	bb_printf("Filesystem%11s%-15sUsed Available Use%% Mounted on\n",
			  "", disp_units_hdr);

	mount_table = NULL;
	if (*argv == NULL) {
		if (!(mount_table = setmntent(bb_path_mtab_file, "r")))
			bb_perror_msg_and_die("setmntent(%s)", bb_path_mtab_file);
	}

	do {
		const char *device;
		const char *mount_point;

		if (mount_table) {
			if (!(mount_entry = getmntent(mount_table))) {
				endmntent(mount_table);
				break;
			}
		} else {
			if (!(mount_point = *argv++))
				break;
			if (!(mount_entry = find_mount_point(mount_point, bb_path_mtab_file))) {
				bb_error_msg("%s: can't find mount point.", mount_point);
			SET_ERROR:
				status = EXIT_FAILURE;
				continue;
			}
		}

		device = mount_entry->mnt_fsname;
		mount_point = mount_entry->mnt_dir;

		if (statfs(mount_point, &s) != 0) {
			bb_perror_msg("%s", mount_point);
			goto SET_ERROR;
		}

#ifdef DEBUG
		fflush(stdout);
		bb_error_msg("mount_point=%s, s.f_blocks=%lld s.f_bfree=%lld",
				mount_point, (long long)s.f_blocks, (long long)s.f_bfree);
#endif
		if ((s.f_blocks > 0) || !mount_table){
			if (show_inode) {
				s.f_blocks = s.f_files;
				s.f_bavail = s.f_bfree = s.f_ffree;
				s.f_bsize = 1;
#ifdef CONFIG_FEATURE_HUMAN_READABLE
				if (df_disp_hr)
					df_disp_hr = 1;
#endif
			}

			blocks_used = s.f_blocks - s.f_bfree;
			if (blocks_used + s.f_bavail) {
				int blocks_percent_used = (((long long) blocks_used) * 100
									   + (blocks_used + s.f_bavail)/2
									   ) / (blocks_used + s.f_bavail);
				sprintf(percent, "%3d%%", blocks_percent_used);
			} else {
				sprintf(percent, "   -");
			}

			if (strcmp(device, "/dev/root") == 0) {
				/* Adjusts device to be the real root device,
				* or leaves device alone if it can't find it */
				if ((device = find_block_device("/")) == NULL) {
					goto SET_ERROR;
				}
			}

#ifdef CONFIG_FEATURE_HUMAN_READABLE
			bb_printf("%-20s %9s ", device,
					  make_human_readable_str(s.f_blocks, s.f_bsize, df_disp_hr, hr_buf));

			bb_printf("%9s ",
					  make_human_readable_str( (s.f_blocks - s.f_bfree),
							s.f_bsize, df_disp_hr, hr_buf));

			bb_printf("%9s %s %s\n",
					  make_human_readable_str(s.f_bavail, s.f_bsize, df_disp_hr, hr_buf),
					  percent, mount_point);
#else
			bb_printf("%-20s %9ld %9ld %9ld %s %s\n",
					  device,
					  kscale(s.f_blocks, s.f_bsize),
					  kscale(s.f_blocks-s.f_bfree, s.f_bsize),
					  kscale(s.f_bavail, s.f_bsize),
					  percent, mount_point);
#endif
		}

	} while (1);

	bb_fflush_stdout_and_exit(status);
}
#endif /* APPLET_USAGE */
