/*
 * Mini du implementation for busybox
 *
 * Copyright (C) 1999,2000,2001 by Lineo, inc. and John Beppu
 * Copyright (C) 1999,2000,2001 by John Beppu <beppu@codepoet.org>
 * Copyright (C) 2002  Edward Betts <edward@debian.org>
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */

/* BB_AUDIT SUSv3 compliant (unless default blocksize set to 1k) */
/* http://www.opengroup.org/onlinepubs/007904975/utilities/du.html */

/* Mar 16, 2003      Manuel Novoa III   (mjn3@codepoet.org)
 *
 * Mostly rewritten for SUSv3 compliance and to fix bugs/defects.
 * 1) Added support for SUSv3 -a, -H, -L, gnu -c, and (busybox) -d options.
 *    The -d option allows setting of max depth (similar to gnu --max-depth).
 * 2) Fixed incorrect size calculations for links and directories, especially
 *    when errors occurred.  Calculates sizes should now match gnu du output.
 * 3) Added error checking of output.
 * 4) Fixed busybox bug #1284 involving long overflow with human_readable.
 */
#ifdef APPLET_USAGE
#define du_trivial_usage \
	"[-aHLdclsx" USE_FEATURE_HUMAN_READABLE("hm") "k] [FILE]..."
#define du_full_usage \
	"Summarizes disk space used for each FILE and/or directory.\n" \
	"Disk space is printed in units of " \
	USE_FEATURE_DU_DEFAULT_BLOCKSIZE_1K("1024") SKIP_FEATURE_DU_DEFAULT_BLOCKSIZE_1K("512") \
	" bytes.\n\n" \
	"Options:\n" \
	"\t-a\tshow sizes of files in addition to directories\n" \
	"\t-H\tfollow symbolic links that are FILE command line args\n" \
	"\t-L\tfollow all symbolic links encountered\n" \
	"\t-d N\tlimit output to directories (and files with -a) of depth < N\n" \
	"\t-c\toutput a grand total\n" \
	"\t-l\tcount sizes many times if hard linked\n" \
	"\t-s\tdisplay only a total for each argument\n" \
	"\t-x\tskip directories on different filesystems\n" \
	USE_FEATURE_HUMAN_READABLE( \
	"\t-h\tprint sizes in human readable format (e.g., 1K 243M 2G )\n" \
	"\t-m\tprint sizes in megabytes\n" ) \
	"\t-k\tprint sizes in kilobytes" USE_FEATURE_DU_DEFAULT_BLOCKSIZE_1K("(default)")
#else

#include "busybox.h"
#include <unistd.h>
#include <dirent.h>

#define COMPLEMENTARY "\0:-d:H-L:L-H:s-d:d-s" USE_FEATURE_HUMAN_READABLE(":h-km:k-hm:m-hk")
#include "opt_gen.h"
OPT_GEN(
opt_simple(a),
opt_simple(H),
opt(k, KILOBYTE_OPT),
opt_simple(L),
opt_simple(s),
opt_simple(x),
opt_simple(l),
opt_simple(c),
opt_simple(d, arg),
opt_opt(ENABLE_FEATURE_HUMAN_READABLE, h, HUMAN_OPT),
opt_opt(ENABLE_FEATURE_HUMAN_READABLE, m, MEGABYTE_OPT)
)
#include "opt_gen.h"

CALC_BSS struct BB_COMMON_BSS
{
	int status;
	int print_files;
	int slink_depth;
	int du_depth;
	int one_file_system;
	dev_t dir_dev;
	nlink_t count_hardlinks;
#ifdef CONFIG_FEATURE_HUMAN_READABLE
	unsigned long disp_hr;
#else
	unsigned int disp_k;
#endif
	int max_print_depth;
	void *ht;
} bb_common_bss;
#define status bb_common_bss.status
#define print_files bb_common_bss.print_files
#define slink_depth bb_common_bss.slink_depth
#define du_depth bb_common_bss.du_depth
#define one_file_system bb_common_bss.one_file_system
#define dir_dev bb_common_bss.dir_dev
#define count_hardlinks bb_common_bss.count_hardlinks
#define disp_hr bb_common_bss.disp_hr
#define disp_k bb_common_bss.disp_k
#define max_print_depth bb_common_bss.max_print_depth
#define ht bb_common_bss.ht

static void print(long size, char *filename)
{
	/* TODO - May not want to defer error checking here. */
#ifdef CONFIG_FEATURE_HUMAN_READABLE
	char hr_buf[32];

	bb_printf("%s\t%s\n",
			make_human_readable_str(size, 512, disp_hr, hr_buf),
			filename);
#else
	if (disp_k) {
		size++;
		size >>= 1;
	}
	bb_printf("%ld\t%s\n", size, filename);
#endif
}

/* tiny recursive du */
static long du(char *filename)
{
	struct stat statbuf;
	long sum;

	if ((lstat(filename, &statbuf)) != 0) {
		bb_perror_msg("%s", filename);
		status = EXIT_FAILURE;
		return 0;
	}

	if (one_file_system) {
		if (du_depth == 0) {
			dir_dev = statbuf.st_dev;
		} else if (dir_dev != statbuf.st_dev) {
			return 0;
		}
	}

	sum = statbuf.st_blocks;

	if (S_ISLNK(statbuf.st_mode)) {
		if (slink_depth > du_depth) {   /* -H or -L */
			if ((stat(filename, &statbuf)) != 0) {
				bb_perror_msg("%s", filename);
				status = EXIT_FAILURE;
				return 0;
			}
			sum = statbuf.st_blocks;
			if (slink_depth == 1) {
				slink_depth = INT_MAX;  /* Convert -H to -L. */
			}
		}
	}

	if (statbuf.st_nlink > count_hardlinks) {
		/* Add files/directories with links only once */
		if (is_in_ino_dev_hashtable(&statbuf, NULL, &ht))
			return 0;
		add_to_ino_dev_hashtable(&statbuf, NULL, &ht);
	}

	if (S_ISDIR(statbuf.st_mode)) {
		DIR *dir;
		struct dirent *entry;
		char *newfile;

		dir = opendir(filename);
		if (!dir) {
			bb_perror_msg("%s", filename);
			status = EXIT_FAILURE;
			return sum;
		}

		newfile = last_char_is(filename, '/');
		if (newfile)
			*newfile = '\0';

		while ((entry = readdir(dir))) {
			char *name = entry->d_name;

			newfile = concat_subpath_file(filename, name);
			if(newfile == NULL)
				continue;
			++du_depth;
			sum += du(newfile);
			--du_depth;
			free(newfile);
		}
		closedir(dir);
	} else if (du_depth > print_files) {
		return sum;
	}
	if (du_depth <= max_print_depth)
		print(sum, filename);
	return sum;
}

int du_main(void)
{
	char **argv;
	long total;
	int slink_depth_save;
	int print_final_total;
	unsigned long opt;

#if EXIT_SUCCESS != 0
	status = EXIT_SUCCESS;
#endif
#ifdef CONFIG_FEATURE_HUMAN_READABLE
# ifdef CONFIG_FEATURE_DU_DEFAULT_BLOCKSIZE_1K
	disp_hr = KILOBYTE;
# else
	disp_hr = 512;
# endif
#else
#if defined CONFIG_FEATURE_DU_DEFAULT_BLOCKSIZE_1K
	disp_k = 1;
#endif
#endif

#ifdef CONFIG_FEATURE_DU_DEFAULT_BLOCKSIZE_1K
	if (getenv("POSIXLY_CORRECT")) {        /* TODO - a new libbb function? */
#ifdef CONFIG_FEATURE_HUMAN_READABLE
		disp_hr = 512;
#else
		disp_k = 0;
#endif
	}
#endif

	max_print_depth = INT_MAX;
	/* Note: SUSv3 specifies that -a and -s options can not be used together
	 * in strictly conforming applications.  However, it also says that some
	 * du implementations may produce output when -a and -s are used together.
	 * gnu du exits with an error code in this case.  We choose to simply
	 * ignore -a.  This is consistent with -s being equivalent to -d 0.
	 */
	opt = bb_xgetopt_ulflags(opt_str, &max_print_depth);
#ifdef CONFIG_FEATURE_HUMAN_READABLE
	if((opt & HUMAN_OPT))
		disp_hr = 0;
	if((opt & MEGABYTE_OPT))
		disp_hr = MEGABYTE;
	if((opt & KILOBYTE_OPT))
		disp_hr = KILOBYTE;
#else
#if !defined CONFIG_FEATURE_DU_DEFAULT_BLOCKSIZE_1K
	if((opt & KILOBYTE_OPT))
		disp_k = 1;
#endif
#endif  /* CONFIG_FEATURE_HUMAN_READABLE */
	if((opt & OPT_a))
		print_files = INT_MAX;
	if((opt & OPT_H))
		slink_depth = 1;
	if((opt & OPT_L))
		slink_depth = INT_MAX;
	if((opt & OPT_s))
		max_print_depth = 0;
	one_file_system = opt & OPT_x;
	if((opt & OPT_l)) {
		count_hardlinks = NLINK_MAX;
	} else {
		count_hardlinks = 1;
	}
	print_final_total = opt & OPT_c;

	argv = Gargv;
	if (*argv == NULL) {
		*--argv = ".";
		if (slink_depth == 1)
			slink_depth = 0;
	}

	slink_depth_save = slink_depth;
	total = 0;
	do {
		total += du(*argv);
		slink_depth = slink_depth_save;
	} while (*++argv);

	if (print_final_total)
		print(total, "total");

	bb_fflush_stdout_and_exit(status);
}
#endif /* APPLET_USAGE */
