/* env - run a program in a modified environment
   Copyright (C) 1986-2022 Free Software Foundation, Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */

/* Richard Mlynarik and David MacKenzie */

#ifdef APPLET_USAGE
#define env_trivial_usage \
	"[-iu] [-] [-C dir] [-S str] [-v] [name=value]... [command [args]...]"
#define env_full_usage \
	"Prints the current environment or runs a program after setting\n" \
	"up the specified environment.\n\n" \
	"Options:\n" \
	USE_FEATURE_LONG_OPTIONS( \
	"\t-i, --ignore-environment  start with an empty environment\n" \
	"\t-0, --null           end each output line with NUL, not newline\n" \
	"\t-u, --unset=NAME     remove variable from the environment\n" \
	"\t-C, --chdir=DIR      change working directory to DIR\n" \
	"\t-S, --split-string=S  process and split S into separate arguments;\n" \
		"\t\tused to pass multiple arguments on shebang lines\n" \
	"\t--block-signal[=SIG]    block delivery of SIG signal(s) to COMMAND\n" \
	"\t--default-signal[=SIG]  reset handling of SIG signal(s) to the default\n" \
	"\t--ignore-signal[=SIG]   set handling of SIG signal(s) to do nothing\n" \
	"\t--list-signal-handling  list non default signal handling to stderr\n" \
	"\t-v, --debug          print verbose information for each processing step" ) \
	SKIP_FEATURE_LONG_OPTIONS( \
	"\t-i, -\t\tstart with an empty environment\n" \
	"\t-0\t\tend each output line with NUL, not newline\n" \
	"\t-u NAME\t\tremove variable from the environment\n" \
	"\t-C DIR\t\tchange working directory to DIR\n" \
	"\t-S SPLIT-STRING\t\tprocess and split S into separate arguments;\n" \
		"\t\tused to pass multiple arguments on shebang lines\n" \
	"\t-B SIGS\t\tblock delivery of SIG signal(s coma seperated) to COMMAND\n" \
	"\t-D SIGS\t\treset handling of SIG signal(s coma seperated) to the default\n" \
	"\t-I SIGS\t\tset handling of SIG signal(s coma seperated) to do nothing\n" \
	"\t-L\t\tlist non default signal handling to stderr\n" \
	"\t-v\t\tprint verbose information for each processing step" )
#else

#include "busybox.h"
#include "bb_sig.h"
#if !ENABLE_FEATURE_LONG_OPTIONS
#include <getopt.h> /* optind */
#endif
#include <string.h>
#include <unistd.h>

/* The isspace characters in the C locale.  */
#define C_ISSPACE_CHARS " \t\n\v\f\r"


/* For long options that have no equivalent short option, use a
   non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
enum
{
  DEFAULT_SIGNAL_OPTION = CHAR_MAX + 1,
  IGNORE_SIGNAL_OPTION,
  BLOCK_SIGNAL_OPTION,
  LIST_SIGNAL_HANDLING_OPTION,
};

/* Possible actions on each signal.  */
enum SIGNAL_MODE {
  UNCHANGED = 0,
  DEFAULT,       /* Set to default handler (SIG_DFL).  */
  DEFAULT_NOERR, /* Ditto, but ignore sigaction(2) errors.  */
  IGNORE,        /* Set to ignore (SIG_IGN).  */
  IGNORE_NOERR   /* Ditto, but ignore sigaction(2) errors.  */
};

#define OPT_PLUS "+"
#define COMPLEMENTARY "\0u::B::D::I::"
#include "opt_gen.h"
OPT_GEN(opt(i, OPT_IGNORE_ENVIRONMENT),
opt(v, OPT_DEBUG),
opt(0, OPT_NULL),
opt(C, OPT_CHDIR, dir),
opt(S, OPT_SPLIT_STRING, str),
opt(u, OPT_UNSET, list),
opt(B, OPT_BLOCK_SIGNAL, list),
opt(D, OPT_DEFAULT_SIGNAL, list),
opt(I, OPT_IGNORE_SIGNAL, list),
opt(L, OPT_LIST_HANDLR))
#include "opt_gen.h"

#if ENABLE_FEATURE_LONG_OPTIONS
static const struct option env_long_options[] =
{
  {"ignore-environment", no_argument, NULL, 'i'},
  {"null", no_argument, NULL, '0'},
  {"unset", required_argument, NULL, 'u'},
  {"chdir", required_argument, NULL, 'C'},
  {"default-signal", optional_argument, NULL, 'D'},
  {"ignore-signal",  optional_argument, NULL, 'I'},
  {"block-signal",   optional_argument, NULL, 'B'},
  {"list-signal-handling", no_argument, NULL,  'L'},
  {"debug", no_argument, NULL, 'v'},
  {"split-string", required_argument, NULL, 'S'},
  {NULL, 0, NULL, 0}
};
#endif

CALC_BSS struct BB_COMMON_BSS
{
	int signum_bound;
	/* Annotate the output with extra info to aid the user.  */
	int dev_debug;
	enum SIGNAL_MODE *signals;
	/* Set of signals to block.  */
	sigset_t block_signals;
	/* Set of signals to unblock.  */
	sigset_t unblock_signals;
	/* Whether signal mask adjustment requested.  */
	int sig_mask_changed;
} bb_common_bss;
#define signum_bound bb_common_bss.signum_bound
#define dev_debug bb_common_bss.dev_debug
#define signals bb_common_bss.signals
#define block_signals bb_common_bss.block_signals
#define unblock_signals bb_common_bss.unblock_signals
#define sig_mask_changed bb_common_bss.sig_mask_changed


static void devmsg(const char *s, ...) __attribute__ ((format (printf, 1, 2)));
static void devmsg(const char *s, ...)
{
	if (dev_debug) {
		va_list p;

		va_start(p, s);
		bb_verror_msg(s, p);
		va_end(p);
	}
}

static void
unset_envvars (const char *u)
{
	devmsg ("unset:    %s", u);

	if (unsetenv (u))
		bb_perror_msg_and_die ("cannot unset %s", u);
}


/* Return a pointer to a buffer containing the VARNAME as
   extracted from a '${VARNAME}' string.
   The returned string will be NUL terminated.
   The returned pointer should not be freed.
   */
static const char *
extract_varname (char const *str, char **pn)
{
  char const *p = NULL;

  if (str[1] == '{' && ((str[2] >= 'A' && str[2] <= 'Z') ||
			(str[2] >= 'a' && str[2] <= 'z') || str[2] == '_')) {
      p = str + 3;
      while ((*p >= 'A' && *p <= 'Z') ||
		(*p >= 'a' && *p <= 'z') ||
		(*p >= '0' && *p <= '9') || *p == '_')
	++p;
      if (*p != '}')
	p = NULL;
  }
  if (!p)
    bb_error_msg_and_die ("only ${VARNAME} expansion is supported, error at: %s", str);
  /* -2 and +2 account for the '${' prefix.  */
  *pn = xstrndup (str + 2, (p - str - 2));
  return p + 1;
}

#define STATE_SQ      1
#define STATE_DQ      2
static char **
split_str(char *arg0, const char *s0, int *pargc)
{
 int newc, state = 0;
 size_t ac;
 const char *s;
 char **argv = NULL;
 char *arg = arg0;

 loop:
	ac = 1;
	s = s0;
 while(1) {
	switch (*s) {
		case '\0': case '#':
			if (state)
				bb_error_msg_and_die ("no terminating quote in -S string");
			if (argv) {
				argv[ac] = NULL;
				*pargc = (int)ac;
				return argv;
			} else {
				/* loop begin again, second scan - store to allocated memory */
				argv = xmalloc((ac + 1) * sizeof(char *));
				argv[0] = arg0;
				if ((arg - arg0))
					arg = xmalloc((arg - arg0));
				goto loop;
			}
		case ' ': case '\t': case '\n': case '\v': case '\f': case '\r':
			s++;
			continue;
		default:
			break;
	}

	if (argv)
		argv[ac] = arg;
	ac++;
	while (1) {
		newc = *s;
		switch (newc) {
			case '\0':
				goto eol;
			case '#':
				if (!state) {
					newc = '\0';
					goto eol;
				}
				break;
			case '\'':
				if (state == STATE_DQ)
					break;
				state = state == STATE_SQ ? 0 : STATE_SQ;
				++s;
				continue;
			case '"':
				if (state == STATE_SQ)
					break;
				state = state == STATE_DQ ? 0 : STATE_DQ;
				++s;
				continue;
			case ' ': case '\t': case '\n': case '\v': case '\f': case '\r':
				if (!state)
					newc = '\0';
				break;
			case '\\':
				/* Backslash inside single-quotes is not special, except \\ and \'.  */
				if (state == STATE_SQ && s[1] != '\\' && s[1] != '\'')
					break;
				/* Skip the backslash and examine the next character.  */
				newc = *++s;
				switch (newc) {
					case '"': case '#': case '$': case '\'': case '\\':
						/* Pass escaped character as-is.  */
						break;
					case 'f': newc = '\f'; break;
					case 'n': newc = '\n'; break;
					case 'r': newc = '\r'; break;
					case 't': newc = '\t'; break;
					case 'v': newc = '\v'; break;
					case '\0':
						bb_error_msg_and_die ("invalid backslash at end of string in -S");
					default:
						bb_error_msg_and_die ("invalid sequence '\\%c' in -S", newc);
				}
				break;
			case '$':
				/* ${VARNAME} are not expanded inside single-quotes.  */
				if (state != STATE_SQ) {
					/* Store the ${VARNAME} value.  */
					char *v, *n;
					s = extract_varname (s, &n);
					v = getenv (n);
					if (v) {
						size_t lv = strlen (v);
						if (argv)
							memcpy (arg, v, lv);
						arg += lv;
						devmsg ("expanding ${%s} into %s", n, v);
					} else {
						devmsg ("replacing ${%s} with null string", n);
					}
					free (n);
					continue;
				}
				break;
		}

		s++;
	eol:
		if (argv)
			*arg = newc;
		arg++;
		if (newc == '\0')
			break;
	}
 }
}

static void
parse_signal_action_params (char const *optarg, int set_default)
{
  const char *signame;
  char *opt_sig;
  char *optarg_writable;

  if (! optarg) {
      int i;
      /* Without an argument, reset all signals.
	 Some signals cannot be set to ignore or default (e.g., SIGKILL,
	 SIGSTOP on most OSes, and SIGCONT on AIX.) - so ignore errors.  */
      for (i = 1 ; i <= signum_bound; i++)
	if (u_signal_num2name (i, 0, NULL) != NULL)
	  signals[i] = set_default ? DEFAULT_NOERR : IGNORE_NOERR;
      return;
  }

  optarg_writable = xstrdup (optarg);

  opt_sig = strtok (optarg_writable, ",");
  while (opt_sig) {
      int signum;
      /* 0 (EXIT) - we reject it.  */
      signame = u_signal_name2num(opt_sig, &signum, 1, NULL);
      if (signame == NULL)
	bb_error_msg ("%s: invalid signal", opt_sig);

      signals[signum] = set_default ? DEFAULT : IGNORE;

      opt_sig = strtok (NULL, ",");
  }

  free (optarg_writable);
}

static void
reset_signal_handlers (void)
{
  int i;
  for (i = 1; i <= signum_bound; i++) {
      struct sigaction act;

      if (signals[i] == UNCHANGED)
	continue;

      int ignore_errors = (signals[i] == DEFAULT_NOERR
			    || signals[i] == IGNORE_NOERR);

      int set_to_default = (signals[i] == DEFAULT
			     || signals[i] == DEFAULT_NOERR);

      int sig_err = sigaction (i, NULL, &act);

      if (sig_err && !ignore_errors)
	bb_perror_msg_and_die ("failed to get signal action for signal %d", i);

      if (! sig_err) {
	  act.sa_handler = set_to_default ? SIG_DFL : SIG_IGN;
	  sig_err = bb_sigaction2 (i, &act);
	  if (sig_err && !ignore_errors)
	    bb_perror_msg_and_die ("failed to set signal action for signal %d", i);
      }

      if (dev_debug) {
	  const char *signame = u_signal_num2name(i, 1, NULL);
	  devmsg ("Reset signal %s (%d) to %s%s", signame, i,
		  set_to_default ? "DEFAULT" : "IGNORE",
		  sig_err ? " (failure ignored)" : "");
	}
    }
}

static void
parse_block_signal_params (char const *optarg, int block)
{
  const char *signame;
  char *opt_sig;
  char *optarg_writable;

  if (! optarg) {
      /* Without an argument, reset all signals.  */
      sigfillset (block ? &block_signals : &unblock_signals);
      sigemptyset (block ? &unblock_signals : &block_signals);
  } else if (! sig_mask_changed) {
      /* Initialize the sets.  */
      sigemptyset (&block_signals);
      sigemptyset (&unblock_signals);
  }

  sig_mask_changed = TRUE;

  if (! optarg)
    return;

  optarg_writable = xstrdup (optarg);

  opt_sig = strtok (optarg_writable, ",");
  while (opt_sig) {
      int signum;

      /* 0 (EXIT) - we reject it.  */
      signame = u_signal_name2num(opt_sig, &signum, 1, NULL);
      if (signame == NULL)
	bb_error_msg ("%s: invalid signal", opt_sig);

      sigaddset (block ? &block_signals : &unblock_signals, signum);
      sigdelset (block ? &unblock_signals : &block_signals, signum);

      opt_sig = strtok (NULL, ",");
  }

  free (optarg_writable);
}

static void
set_signal_proc_mask (void)
{
  /* Get the existing signal mask */
  sigset_t set;
  char const *debug_act;
  int i;

  sigemptyset (&set);

  if (sigprocmask (0, NULL, &set))
    bb_perror_msg_and_die ("failed to get signal process mask");

  for (i = 1; i <= signum_bound; i++) {
      if (sigismember (&block_signals, i)) {
	  sigaddset (&set, i);
	  debug_act = "BLOCK";
      } else if (sigismember (&unblock_signals, i)) {
	  sigdelset (&set, i);
	  debug_act = "UNBLOCK";
      } else {
	  debug_act = NULL;
      }

      if (dev_debug && debug_act) {
	  const char *signame = u_signal_num2name(i, 1, NULL);
	  devmsg ("signal %s (%d) mask set to %s", signame, i, debug_act);
      }
  }

  if (sigprocmask (SIG_SETMASK, &set, NULL))
    bb_perror_msg_and_die ("failed to set signal process mask");
}

static void
list_signal_handling (void)
{
  sigset_t set;
  const char *signame;
  int i;

  sigemptyset (&set);
  if (sigprocmask (0, NULL, &set))
    bb_perror_msg_and_die ("failed to get signal process mask");

  for (i = 1; i <= signum_bound; i++) {
      struct sigaction act;
      char const *ignored = act.sa_handler == SIG_IGN ? "IGNORE" : "";
      char const *blocked = sigismember (&set, i) ? "BLOCK" : "";
      char const *connect = *ignored && *blocked ? "," : "";

      signame = u_signal_num2name (i, 1, NULL);
      if (signame == NULL)
	continue;
      if (sigaction (i, NULL, &act))
	continue;

      if (! *ignored && ! *blocked)
	continue;

      fprintf (stderr, "%-10s (%2d): %s%s%s\n", signame, i, blocked, connect, ignored);
    }
}

static void
initialize_signals (void)
{
	int i;

	for (i = 0; i < 64; i++)
		if (u_signal_num2name(i, 0, NULL) != NULL)
			signum_bound = i;
	signals = zmalloc ((sizeof *signals) * (signum_bound + 1));
}

extern int env_main(void)
{
  int optc;
  int opt_nul_terminate_output;
  char *newdir = NULL, *ss = NULL, *eq;
  llist_t *tl, *unset_env = NULL, *bs = NULL, *ds = NULL, *is = NULL;
  int report_signal_handling;
  char **argv = Gargv;
  int program_specified, exit_status;

  initialize_signals ();

#if ENABLE_FEATURE_LONG_OPTIONS
  bb_applet_long_options = env_long_options;
#endif

  optc = bb_xgetopt_ulflags(opt_str, &newdir, &ss, &unset_env, &bs, &ds, &is);
  if ((optc & OPT_DEBUG)) {
	dev_debug = 1;
	optc &= ~OPT_DEBUG;
  }
  if ((optc & OPT_SPLIT_STRING)) {
	if (optc != OPT_SPLIT_STRING || Gargc > 0) {
		const char *ap = bb_applet_name;
		bb_applet_name = NULL;
		bb_error_msg_and_die("Excuse me, usage with -S support only as: %s -[v]S'opts [cmd [args...]]'", ap);
	}
	devmsg ("split -S:  %s", ss);
	Gargv = split_str((char *)bb_applet_name, ss, &Gargc);
	devmsg (" into:    %s", Gargv[1]);
	for (optc = 2; optc < Gargc; optc++)
		devmsg ("     &    %s", Gargv[optc]);
	optind = 0;     /* reset for rescan */
	optc = bb_xgetopt_ulflags(opt_str, &newdir, &ss, &unset_env, &bs, &ds, &is);
	dev_debug |= optc & OPT_DEBUG;
  }

  argv = Gargv;
  if (*argv && (argv[0][0] == '-') && !argv[0][1]) {
	optc |= OPT_IGNORE_ENVIRONMENT;
	++argv;
  }
  opt_nul_terminate_output = optc & OPT_NULL;
  for (tl = ds; tl != NULL; tl = tl->link) {
	parse_signal_action_params (tl->data, TRUE);
	parse_block_signal_params (tl->data, FALSE);
  }
  for (tl = bs; tl != NULL; tl = tl->link)
	parse_block_signal_params (tl->data, TRUE);
  for (tl = is; tl != NULL; tl = tl->link)
	parse_signal_action_params (tl->data, FALSE);
  report_signal_handling = optc & OPT_LIST_HANDLR;

  if ((optc & OPT_IGNORE_ENVIRONMENT)) {
      devmsg ("cleaning environ");
      clearenv ();
  } else {
    for (tl = unset_env; tl != NULL; tl = tl->link)
	unset_envvars (tl->data);
  }

  while (*argv && (eq = strchr (*argv, '='))) {
      devmsg ("setenv:   %s", *argv);

      if (putenv (*argv)) {
	  *eq = '\0';
	  bb_perror_msg_and_die ("cannot set %s", *argv);
      }
      argv++;
  }

  program_specified = (*argv != NULL);

  if (opt_nul_terminate_output && program_specified)
      bb_error_msg_and_die ("cannot specify --null (-0) with command");

  if (newdir && ! program_specified)
      bb_error_msg_and_die ("must specify command with --chdir (-C)");

  if (! program_specified) {
      /* Print the environment and exit.  */
      char *const *e = environ;
      if (e) {
	while (*e)
		printf ("%s%c", *e++, opt_nul_terminate_output ? '\0' : '\n');
	bb_fflush_stdout_and_exit(0);
      } else
	return 0;
  }

  reset_signal_handlers ();
  if (sig_mask_changed)
    set_signal_proc_mask ();

  if (report_signal_handling)
    list_signal_handling ();

  if (newdir) {
      devmsg ("chdir:    %s", newdir);

      if (chdir (newdir) != 0)
	bb_perror_msg_and_die ("cannot change directory to %s", newdir);
  }

  if (dev_debug) {
      devmsg ("executing: %s", *argv);
      for (optc = 0; argv[optc] != NULL; optc++)
	devmsg ("   arg[%d]= %s", optc, argv[optc]);
  }

  execvp (*argv, argv);

  exit_status = errno == ENOENT ? 127 : 126;
  bb_perror_msg ("%s", *argv);

  if (exit_status == 127 && strpbrk (*argv, C_ISSPACE_CHARS))
    bb_error_msg ("use -[v]S to pass options in shebang lines");

  return exit_status;
}
#endif /* APPLET_USAGE */
