/* fold -- wrap each input line to fit in specified width.

   Written by David MacKenzie, djm@gnu.ai.mit.edu.
   Copyright (C) 91, 1995-2002 Free Software Foundation, Inc.

   Modified for busybox based on coreutils v 5.0
   Copyright (C) 2003 Glenn McGrath <bug1@iinet.net.au>

   Licensed under the GPL v2 or later, see the file LICENSE in this tarball.
*/
#ifdef APPLET_USAGE
#define fold_trivial_usage \
	"[-bs] [-w WIDTH] [FILE]"
#define fold_full_usage \
	"Wrap input lines in each FILE (standard input by default), writing to\n" \
	"standard output.\n\n" \
	"Options:\n" \
	"\t-b\tcount bytes rather than columns\n" \
	"\t-s\tbreak at spaces\n" \
	"\t-w\tuse WIDTH columns instead of 80"
#else

/* zero args to one "-" arg, turn any numeric options into -w options */
#define COMPLEMENTARY "\0-0:==w"
#include "opt_gen.h"
OPT_GEN(
opt(b, FLAG_COUNT_BYTES),
opt(s, FLAG_BREAK_SPACES),
opt(w, FLAG_WIDTH, arg),
)
#include "opt_gen.h"

#include "busybox.h"
#include <ctype.h>
#include <string.h>
#include <unistd.h>

CALC_BSS struct BB_COMMON_BSS
{
	char *line_out;
	int allocated_out;
} bb_common_bss;
#define line_out bb_common_bss.line_out
#define allocated_out bb_common_bss.allocated_out

/* Assuming the current column is COLUMN, return the column that
   printing C will move the cursor to.
   The first column is 0. */

static int adjust_column(int column, char c)
{
	if (c == '\b') {
		if (column > 0)
			column--;
	} else if (c == '\r')
		column = 0;
	else if (c == '\t')
		column = column + 8 - column % 8;
	else                    /* if (isprint (c)) */
		column++;
	return column;
}

extern int fold_main(void)
{
	char **argv;
	char *w_opt;
	int width = 80;
	int errs = 0;
	int flag_count_bytes;
	int flags;

	flags = bb_xgetopt_ulflags(opt_str, &w_opt);

	if (flags & FLAG_WIDTH)
		width = bb_xparse_i_number(10 | BB_PN_SET_LOWER1 | BB_PN_SET_UPPER, w_opt, 10000);
	flag_count_bytes = flags & FLAG_COUNT_BYTES;

	for (argv = Gargv; *argv; argv++) {
		int c;
		int column = 0;         /* Screen column where next char will go. */
		int offset_out = 0;     /* Index in `line_out' for next char. */
		FILE *istream = bb_xfopen(*argv, "W-r");

		if (istream == NULL) {
			errs |= EXIT_FAILURE;
			continue;
		}

		while ((c = getc(istream)) != EOF) {
			if (offset_out + 1 >= allocated_out) {
				allocated_out += 1024;
				line_out = xrealloc(line_out, allocated_out);
			}

			if (c == '\n') {
				line_out[offset_out++] = c;
				fwrite(line_out, sizeof(char), (size_t) offset_out, stdout);
				column = offset_out = 0;
				continue;
			}

rescan:
			if(flag_count_bytes)
				column++;
			else
				column = adjust_column(column, c);

			if (column > width) {
				/* This character would make the line too long.
				   Print the line plus a newline, and make this character
				   start the next line. */
				if (flags & FLAG_BREAK_SPACES) {
					/* Look for the last blank. */
					int logical_end;

					for (logical_end = offset_out - 1; logical_end >= 0; logical_end--) {
						if (isblank(line_out[logical_end])) {
							break;
						}
					}
					if (logical_end >= 0) {
						/* Found a blank.  Don't output the part after it. */
						logical_end++;
						fwrite(line_out, sizeof(char), (size_t) logical_end, stdout);
						putchar('\n');
						/* Move the remainder to the beginning of the next line.
						   The areas being copied here might overlap. */
						memmove(line_out, line_out + logical_end, offset_out - logical_end);
						offset_out -= logical_end;
						if(flag_count_bytes)
							column = offset_out;
						else {
						    int i;
						    for (column = i = 0; i < offset_out; i++)
							column = adjust_column(column, line_out[i]);
						}
						goto rescan;
					}
				} else {
					if (offset_out == 0) {
						line_out[offset_out++] = c;
						continue;
					}
				}
				line_out[offset_out++] = '\n';
				fwrite(line_out, sizeof(char), (size_t) offset_out, stdout);
				column = offset_out = 0;
				goto rescan;
			}

			line_out[offset_out++] = c;
		}

		if (offset_out) {
			fwrite(line_out, sizeof(char), (size_t) offset_out, stdout);
		}

		if (ferror(istream) || bb_fclose_nonstdin(istream)) {
			bb_perror_msg("%s", *argv);     /* Avoid multibyte problems. */
			errs |= EXIT_FAILURE;
		}
	}

	bb_fflush_stdout_and_exit(errs);
}
#endif /* APPLET_USAGE */
