/*
 * tiny-ls.c version 0.1.0: A minimalist 'ls'
 * Copyright (C) 1996 Brian Candler <B.Candler@pobox.com>
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */

/*
 * To achieve a small memory footprint, this version of 'ls' doesn't do any
 * file sorting, and only has the most essential command line switches
 * (i.e., the ones I couldn't live without :-) All features which involve
 * linking in substantial chunks of libc can be disabled.
 *
 * Although I don't really want to add new features to this program to
 * keep it small, I *am* interested to receive bug fixes and ways to make
 * it more portable.
 *
 * KNOWN BUGS:
 * 1. ls -l of a directory doesn't give "total <blocks>" header
 * 2. ls of a symlink to a directory doesn't list directory contents
 * 3. hidden files can make column width too large
 *
 * NON-OPTIMAL BEHAVIOUR:
 * 1. autowidth reads directories twice
 * 2. if you do a short directory listing without filetype characters
 *    appended, there's no need to stat each one
 * PORTABILITY:
 * 1. requires lstat (BSD) - how do you do it without?
 */
#ifdef APPLET_USAGE
#define ls_trivial_usage \
	"[-1Aa" USE_FEATURE_LS_TIMESTAMPS("c") "Cd" \
	USE_FEATURE_LS_FILETYPES("F") "iln" USE_FEATURE_LS_FILETYPES("p") \
	USE_FEATURE_LS_FOLLOWLINKS("L") USE_FEATURE_LS_RECURSIVE("R") \
	USE_FEATURE_LS_SORTFILES("rS") "s" USE_FEATURE_AUTOWIDTH("T") \
	USE_FEATURE_LS_TIMESTAMPS("tu") USE_FEATURE_LS_SORTFILES("v") \
	USE_FEATURE_AUTOWIDTH("w") "x" USE_FEATURE_LS_SORTFILES("X") \
	USE_FEATURE_HUMAN_READABLE("h") "k] filenames...]"
#define ls_full_usage \
	"List directory contents\n\n" \
	"Options:\n" \
	"\t-1\tlist files in a single column\n" \
	"\t-A\tdo not list implied . and ..\n" \
	"\t-a\tdo not hide entries starting with .\n" \
	"\t-C\tlist entries by columns\n" \
	USE_FEATURE_LS_TIMESTAMPS("\t-c\twith -l: show ctime\n") \
	USE_FEATURE_LS_COLOR(USE_FEATURE_LONG_OPTIONS("\t--color[={always,never,auto}]\tto control coloring\n")) \
	"\t-d\tlist directory entries instead of contents\n" \
	USE_FEATURE_LS_TIMESTAMPS(USE_FEATURE_LONG_OPTIONS("\t--full-time\tlist both full date and full time\n")) \
	USE_FEATURE_LS_FILETYPES("\t-F\tappend indicator (one of */=@|) to entries\n") \
	"\t-i\tlist the i-node for each file\n" \
	"\t-l\tuse a long listing format\n" \
	"\t-n\tlist numeric UIDs and GIDs instead of names\n" \
	USE_FEATURE_LS_FILETYPES("\t-p\tappend indicator (one of /=@|) to entries\n") \
	USE_FEATURE_LS_FOLLOWLINKS("\t-L\tlist entries pointed to by symbolic links\n") \
	USE_FEATURE_LS_RECURSIVE("\t-R\tlist subdirectories recursively\n") \
	USE_FEATURE_LS_SORTFILES("\t-r\tsort the listing in reverse order\n") \
	USE_FEATURE_LS_SORTFILES("\t-S\tsort the listing by file size\n") \
	"\t-s\tlist the size of each file, in blocks\n" \
	USE_FEATURE_AUTOWIDTH("\t-T NUM\tassume Tabstop every NUM columns\n") \
	USE_FEATURE_LS_TIMESTAMPS("\t-t\twith -l: show modification time\n") \
	USE_FEATURE_LS_TIMESTAMPS("\t-u\twith -l: show access time\n") \
	USE_FEATURE_LS_SORTFILES("\t-v\tsort the listing by version\n") \
	USE_FEATURE_AUTOWIDTH("\t-w NUM\tassume the terminal is NUM columns wide\n") \
	USE_FEATURE_LS_SORTFILES("\t-X\tsort the listing by extension\n") \
	USE_FEATURE_HUMAN_READABLE( \
	"\t-h\tprint sizes in human readable format (e.g., 1K 243M 2G )\n") \
	"\t-x\tlist entries by lines instead of by columns"
#else

enum {
	TERMINAL_WIDTH = 80,    /* use 79 if terminal has linefold bug */
	COLUMN_GAP = 2,         /* includes the file type char */
};

/************************************************************************/

#include "busybox.h"
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <sys/sysmacros.h>     /* major() and minor() */

#ifdef CONFIG_FEATURE_LS_TIMESTAMPS
#include <time.h>
#endif


#define COMPLEMENTARY "\0l-C1x:C-l1x:1-Cx:x-lC1:el:nl" \
			USE_FEATURE_LS_TIMESTAMPS(":c-u:u-c") \
			USE_FEATURE_LS_SORTFILES(":S-Xv:X-Sv:v-SX") \
  USE_FEATURE_LS_TIMESTAMPS(USE_FEATURE_LS_SORTFILES(":c-tSXv:t-cSXv:S-ct:X-ct:v-ct")) \
  USE_FEATURE_LS_RECURSIVE(":d-R:R-d")
#include "opt_gen.h"
/* 51306 lrwxrwxrwx  1 root     root         2 May 11 01:43 /bin/view -> vi* */
/* what file information will be listed */
BITS_GEN0(LIST, LIST_INO, LIST_BLOCKS, LIST_MODEBITS, LIST_NLINKS, LIST_ID_NAME,
LIST_ID_NUMERIC, LIST_CONTEXT, LIST_SIZE, LIST_DEV, LIST_DATE_TIME,
LIST_FULLTIME, LIST_FILENAME, LIST_SYMLINK, LIST_FILETYPE, LIST_EXEC,

/* what files will be displayed */
DISP_DIRNAME, DISP_HIDDEN, DISP_DOT, DISP_NOLIST, DISP_RECURSIVE, DISP_ROWS,

/* what is the overall style of the listing */
STYLE_COLUMNS, STYLE_SINGLE,

/* which of the three times will be used */
TIME_CHANGE, TIME_ACCESS,

FOLLOW_LINKS, LS_DISP_HR,

/* how will the files be sorted */
SORT_ORDER_REVERSE, SORT_SIZE, SORT_ATIME,
#define SORT_CTIME      (SORT_SIZE|SORT_ATIME)          /* sort by last change time */
SORT_MTIME
#define SORT_VERSION    (SORT_SIZE|SORT_MTIME)          /* sort by version */
#define SORT_EXT        (SORT_ATIME|SORT_MTIME)         /* sort by file name extension */
#define SORT_DIR        (SORT_CTIME|SORT_MTIME)         /* sort by file or directory */
#define SORT_MASK       SORT_DIR
)


/* "[-]Cadil1", POSIX mandated options, busybox always supports */
/* "[-]gnsx", POSIX non-mandated options, busybox always supports */
/* "[-]Ak" GNU options, busybox always supports */
/* "[-]FLRctur", POSIX mandated options, busybox optionally supports */
/* "[-]p", POSIX non-mandated options, busybox optionally supports */
/* "[-]SXvThw", GNU options, busybox optionally supports */
/* "[-]K", SELinux mandated options, busybox optionally supports */
/* "[-]e", I think we made this one up */

OPT_GEN(
opt_simple(C), opt_simple(a), opt_simple(d), opt_simple(i), opt_simple(l),
opt_simple(1), opt_simple(n), opt_simple(s), opt_simple(x),
opt_simple(A),
#ifdef CONFIG_FEATURE_LS_TIMESTAMPS
opt_simple(c), opt_simple(t), opt_simple(u),
opt_opt(ENABLE_FEATURE_LONG_OPTIONS, e, e_OPT),
#endif
#ifdef CONFIG_FEATURE_LS_SORTFILES
opt_simple(S), opt_simple(X), opt_simple(r), opt_simple(v),
#endif
#ifdef CONFIG_FEATURE_LS_FILETYPES
opt_simple(F), opt_simple(p),
#endif
opt_opt(ENABLE_FEATURE_LS_FOLLOWLINKS, L, L_OPT),
opt_opt(ENABLE_FEATURE_LS_RECURSIVE, R, R_OPT),
opt_opt(ENABLE_FEATURE_HUMAN_READABLE, h, h_OPT),
opt_opt(ENABLE_FEATURE_AUTOWIDTH, T, T_OPT, arg),
opt_opt(ENABLE_FEATURE_AUTOWIDTH, w, w_OPT, arg),
opt_simple(g),
opt_opt(ENABLE_FEATURE_LS_COLOR, O, COLOR_OPT, arg, optional)
)
#include "opt_gen.h"

#define LIST_SHORT      LIST_FILENAME
#define LIST_LONG       (LIST_MODEBITS | LIST_NLINKS | LIST_ID_NAME | LIST_SIZE | \
						LIST_DATE_TIME | LIST_FILENAME | LIST_SYMLINK)

static const unsigned opt_flags[] = {
	LIST_SHORT | STYLE_COLUMNS,     /* C */
	DISP_HIDDEN | DISP_DOT,         /* a */
	DISP_NOLIST,                    /* d */
	LIST_INO,                       /* i */
	LIST_LONG | STYLE_SINGLE,       /* l */
	LIST_SHORT | STYLE_SINGLE,      /* 1 */
	LIST_ID_NUMERIC,                /* n */
	LIST_BLOCKS,                    /* s */
	DISP_ROWS,                      /* x */
	DISP_HIDDEN,                    /* A */
#ifdef CONFIG_FEATURE_LS_TIMESTAMPS
# ifdef CONFIG_FEATURE_LS_SORTFILES
	TIME_CHANGE | SORT_CTIME,       /* c */
# else
	TIME_CHANGE,                    /* c */
# endif
# if ENABLE_FEATURE_LONG_OPTIONS
	LIST_FULLTIME,                  /* e */
# endif
# ifdef CONFIG_FEATURE_LS_SORTFILES
	SORT_MTIME,                     /* t */
# else
	0,                              /* t - ignored -- is this correct? */
# endif
# ifdef CONFIG_FEATURE_LS_SORTFILES
	TIME_ACCESS | SORT_ATIME,       /* u */
# else
	TIME_ACCESS,                    /* u */
# endif
#endif
#ifdef CONFIG_FEATURE_LS_SORTFILES
	SORT_SIZE,                      /* S */
	SORT_EXT,                       /* X */
	SORT_ORDER_REVERSE,             /* r */
	SORT_VERSION,                   /* v */
#endif
#ifdef CONFIG_FEATURE_LS_FILETYPES
	LIST_FILETYPE | LIST_EXEC,      /* F */
	LIST_FILETYPE,                  /* p */
#endif
#ifdef CONFIG_FEATURE_LS_FOLLOWLINKS
	FOLLOW_LINKS,                   /* L */
#endif
#ifdef CONFIG_FEATURE_LS_RECURSIVE
	DISP_RECURSIVE,                 /* R */
#endif
#ifdef CONFIG_FEATURE_HUMAN_READABLE
	LS_DISP_HR,                     /* h */
#endif
#if 0
# ifdef CONFIG_FEATURE_AUTOWIDTH
       0, 0,                            /* T, w - ignored */
# endif
	0,                              /* g - ingored */
	COLOR_OPT,                      /* O - internal logic */
#endif
	(1U<<31)
};


#define SPLIT_DIR      1
#define SPLIT_FILE     0
#define SPLIT_SUBDIR   2

#define TYPEINDEX(mode) (((mode) >> 12) & 0x0f)
#define TYPECHAR(mode)  ("0pcCd?bB-?l?s???" [TYPEINDEX(mode)])

#if defined(CONFIG_FEATURE_LS_FILETYPES) || defined(CONFIG_FEATURE_LS_COLOR)
# define APPCHAR(mode)   ("\0|\0\0/\0\0\0\0\0@\0=\0\0\0" [TYPEINDEX(mode)])
#endif

#if ENABLE_FEATURE_LONG_OPTIONS
static const struct option ls_opt[] =
{
#ifdef CONFIG_FEATURE_LS_COLOR
	{"color", optional_argument, NULL, 'O'},
#endif
#ifdef CONFIG_FEATURE_LS_TIMESTAMPS
	{"full-time", no_argument, NULL, 'e'},
#endif
	{NULL, 0, NULL, 0}
};
#endif

/* colored LS support by JaWi, janwillem.janssen@lxtreme.nl */
#ifdef CONFIG_FEATURE_LS_COLOR
#define COLOR(mode)     ("\000\043\043\043\042\000\043\043"\
			 "\000\000\044\000\043\000\000\040" [TYPEINDEX(mode)])
#define ATTR(mode)      ("\00\00\01\00\01\00\01\00"\
			 "\00\00\01\00\01\00\00\01" [TYPEINDEX(mode)])
#endif

/*
 * a directory entry and its stat info are stored here
 */
struct dnode {                  /* the basic node */
	const char *name;       /* the dir entry name */
	char *fullname;         /* the dir entry name */
	struct stat dstat;      /* the file stat info */
	/* point at the next dnode used in list_dir only
	   and not null if fullname have memory allocated */
	void *allocated;
};
typedef struct dnode dnode_t;

static struct dnode **list_dir(const char *, int *pnfiles);
static struct dnode **dnalloc(int);
static int list_single(struct dnode *);

CALC_BSS struct BB_COMMON_BSS
{
	USE_FEATURE_LS_COLOR(int show_color;)
	unsigned int all_fmt;
	int status;
} bb_common_bss;
#define show_color bb_common_bss.show_color
#define all_fmt bb_common_bss.all_fmt
#define selinux_enabled bb_common_bss.selinux_enabled
#define status bb_common_bss.status


#ifdef CONFIG_FEATURE_AUTOWIDTH
static int terminal_width = TERMINAL_WIDTH;
static int tabstops = COLUMN_GAP;
#else
#define tabstops COLUMN_GAP
#define terminal_width TERMINAL_WIDTH
#endif


static struct dnode *my_stat(char *fullname, const char *name)
{
	struct stat dstat;
	struct dnode *cur;
	int rc;

#ifdef CONFIG_FEATURE_LS_FOLLOWLINKS
	if (all_fmt & FOLLOW_LINKS) {
		rc = stat(fullname, &dstat);
		if(rc) {
			bb_perror_msg("%s", fullname);
			status = EXIT_FAILURE;
			return 0;
		}
	} else
#endif
	{
		rc = lstat(fullname, &dstat);
		if(rc) {
			bb_perror_msg("%s", fullname);
			status = EXIT_FAILURE;
			return 0;
		}
	}

	cur = (struct dnode *) xmalloc(sizeof(struct dnode));
	cur->fullname = fullname;
	cur->name = name;
	cur->dstat = dstat;
	return cur;
}

/*----------------------------------------------------------------------*/
#ifdef CONFIG_FEATURE_LS_COLOR
static char fgcolor(mode_t mode)
{
	/* Check wheter the file is existing (if so, color it red!) */
	if (errno == ENOENT) {
		return '\037';
	}
	if (LIST_EXEC && S_ISREG(mode)
		&& (mode & (S_IXUSR | S_IXGRP | S_IXOTH)))
		return COLOR(0xF000);   /* File is executable ... */
	return COLOR(mode);
}

/*----------------------------------------------------------------------*/
static char bgcolor(mode_t mode)
{
	if (LIST_EXEC && S_ISREG(mode)
		&& (mode & (S_IXUSR | S_IXGRP | S_IXOTH)))
		return ATTR(0xF000);    /* File is executable ... */
	return ATTR(mode);
}
#endif

/*----------------------------------------------------------------------*/
#if defined(CONFIG_FEATURE_LS_FILETYPES)
static char append_char(mode_t mode)
{
	if (!(all_fmt & LIST_FILETYPE))
		return '\0';
	if ((all_fmt & LIST_EXEC) && S_ISREG(mode)
		&& (mode & (S_IXUSR | S_IXGRP | S_IXOTH)))
		return '*';
	return APPCHAR(mode);
}
#endif

/*----------------------------------------------------------------------*/

#define countdirs(A,B) count_dirs((A), (B), 1)
#define countsubdirs(A,B) count_dirs((A), (B), 0)

static int count_dirs(struct dnode **dn, int nfiles, int notsubdirs)
{
	int i, dirs;

	if (dn == NULL || nfiles < 1)
		return 0;
	dirs = 0;
	for (i = 0; i < nfiles; i++) {
		if (S_ISDIR(dn[i]->dstat.st_mode)
			&& (notsubdirs ||
			((dn[i]->name[0] != '.') || (dn[i]->name[1]
						&& ((dn[i]->name[1] != '.')
						|| dn[i]->name[2])))))
			dirs++;
	}
	return (dirs);
}

/* get memory to hold an array of pointers */
static struct dnode **dnalloc(int num)
{
	struct dnode **p;

	if (num < 1)
		return (NULL);

	p = (struct dnode **) zmalloc(((size_t) num) * (sizeof(struct dnode *)));
	return (p);
}

static void dfree(struct dnode **dnp, int nfiles)
{
	struct dnode *cur;
	int n;

	for(n = 0; n < nfiles; n++) {
		cur = dnp[n];
		if(cur->allocated)
			free(cur->fullname);    /* free the filename */
		free(cur);              /* free the dnode */
	}
	free(dnp);                      /* free the array holding the dnode pointers */
}

static struct dnode **splitdnarray(struct dnode **dn, int nfiles, int which)
{
	int dncnt, i, d;
	struct dnode **dnp;

	if (dn == NULL || nfiles < 1)
		return (NULL);

	/* count how many dirs and regular files there are */
	if (which == SPLIT_SUBDIR)
		dncnt = countsubdirs(dn, nfiles);
	else {
		dncnt = countdirs(dn, nfiles);  /* assume we are looking for dirs */
		if (which == SPLIT_FILE)
			dncnt = nfiles - dncnt; /* looking for files */
	}

	/* allocate a file array and a dir array */
	dnp = dnalloc(dncnt);

	/* copy the entrys into the file or dir array */
	for (d = i = 0; i < nfiles; i++) {
		if (S_ISDIR(dn[i]->dstat.st_mode)) {
			if (which & (SPLIT_DIR|SPLIT_SUBDIR)) {
				if ((which & SPLIT_DIR)
					|| ((dn[i]->name[0] != '.')
						|| (dn[i]->name[1]
							&& ((dn[i]->name[1] != '.')
								|| dn[i]->name[2])))) {
									dnp[d++] = dn[i];
								}
			}
		} else if (!(which & (SPLIT_DIR|SPLIT_SUBDIR))) {
			dnp[d++] = dn[i];
		}
	}
	return (dnp);
}

/*----------------------------------------------------------------------*/
#ifdef CONFIG_FEATURE_LS_SORTFILES
#define cmp_all_types(a2, a1, d) do { if((a2)==(a1)) d=0; \
					else if((a2)>(a1)) d=1; \
					else d = -1; \
				 } while(0);

static int sortcmp(const void *a, const void *b)
{
	struct dnode *d1 = *(struct dnode **)a;
	struct dnode *d2 = *(struct dnode **)b;
	unsigned int sort_opts = all_fmt & SORT_MASK;
	int dif;

	dif = 0;                        /* assume SORT_NAME */
	if (sort_opts == SORT_SIZE) {
		cmp_all_types(d2->dstat.st_size, d1->dstat.st_size, dif);
	} else if (sort_opts == SORT_ATIME) {
		cmp_all_types(d2->dstat.st_atime, d1->dstat.st_atime, dif);
	} else if (sort_opts == SORT_CTIME) {
		cmp_all_types(d2->dstat.st_ctime, d1->dstat.st_ctime, dif);
	} else if (sort_opts == SORT_MTIME) {
		cmp_all_types(d2->dstat.st_mtime, d1->dstat.st_mtime, dif);
	} else if (sort_opts == SORT_DIR) {
		cmp_all_types(S_ISDIR(d2->dstat.st_mode), S_ISDIR(d1->dstat.st_mode), dif);
	/* } else if (sort_opts == SORT_VERSION) { */
	/* } else if (sort_opts == SORT_EXT) { */
	}

	if (dif == 0) {
		/* sort by name- may be a tie_breaker for time or size cmp */
#ifdef CONFIG_LOCALE_SUPPORT
		dif = strcoll(d1->name, d2->name);
#else
		dif = strcmp(d1->name, d2->name);
#endif
	}

	if (all_fmt & SORT_ORDER_REVERSE)
		dif = -dif;
	return (dif);
}

/*----------------------------------------------------------------------*/
static void dnsort(struct dnode **dn, int size)
{
	qsort(dn, size, sizeof *dn, sortcmp);
}
#endif

/*----------------------------------------------------------------------*/
static void showfiles(struct dnode **dn, int nfiles)
{
	int i, ncols, nrows, row, nc;
	int column = 0;
	int nexttab = 0;
	int column_width = 0; /* for STYLE_SINGLE not used */

	if (dn == NULL || nfiles < 1)
		return;

	if (all_fmt & STYLE_SINGLE) {
		ncols = 1;
	} else {
		/* find the longest file name-  use that as the column width */
		for (i = 0; i < nfiles; i++) {
			int len = mbchar_len(dn[i]->name) +
			((all_fmt & LIST_INO) ? 8 : 0) +
			((all_fmt & LIST_BLOCKS) ? 5 : 0);
			if (column_width < len)
				column_width = len;
		}
		column_width += tabstops;
		ncols = (int) (terminal_width / column_width);
	}

	if (ncols > 1) {
		nrows = nfiles / ncols;
		if ((nrows * ncols) < nfiles)
			nrows++;                /* round up fractionals */
	} else {
		nrows = nfiles;
		ncols = 1;
	}

	for (row = 0; row < nrows; row++) {
		for (nc = 0; nc < ncols; nc++) {
			/* reach into the array based on the column and row */
			i = (nc * nrows) + row; /* assume display by column */
			if (all_fmt & DISP_ROWS)
				i = (row * ncols) + nc; /* display across row */
			if (i < nfiles) {
				if (column > 0) {
					nexttab -= column;
					while (nexttab--) {
						putchar(' ');
						column++;
					}
			}
				nexttab = column + column_width;
				column += list_single(dn[i]);
		}
		}
		putchar('\n');
		column = 0;
	}
}

/*----------------------------------------------------------------------*/
static void showdirs(struct dnode **dn, int ndirs, int first)
{
	int i, nfiles;
	struct dnode **subdnp;

#ifdef CONFIG_FEATURE_LS_RECURSIVE
	int dndirs;
	struct dnode **dnd;
#endif

	if (dn == NULL || ndirs < 1)
		return;

	for (i = 0; i < ndirs; i++) {
		if (all_fmt & (DISP_DIRNAME | DISP_RECURSIVE)) {
			if (!first)
				printf("\n");
			first = 0;
			printf("%s:\n", dn[i]->fullname);
		}
		subdnp = list_dir(dn[i]->fullname, &nfiles);
		if (nfiles > 0) {
			/* list all files at this level */
			USE_FEATURE_LS_SORTFILES(dnsort(subdnp, nfiles));
			showfiles(subdnp, nfiles);
#ifdef CONFIG_FEATURE_LS_RECURSIVE
			if (all_fmt & DISP_RECURSIVE) {
				/* recursive- list the sub-dirs */
				dnd = splitdnarray(subdnp, nfiles, SPLIT_SUBDIR);
				dndirs = countsubdirs(subdnp, nfiles);
				if (dndirs > 0) {
					USE_FEATURE_LS_SORTFILES(dnsort(dnd, dndirs));
					showdirs(dnd, dndirs, 0);
					free(dnd);      /* free the array of dnode pointers to the dirs */
				}
			}
#endif
			dfree(subdnp, nfiles);  /* free the dnodes and the fullname mem */
		}
	}
}

/*----------------------------------------------------------------------*/
static struct dnode **list_dir(const char *path, int *pnfiles)
{
	struct dnode *dn, *cur, **dnp = NULL;
	struct dirent *entry;
	DIR *dir;
	int i, nfiles;

	if (path == NULL)
		return (NULL);

	dn = NULL;
	nfiles = 0;
	dir = opendir(path);
	if (dir == NULL) {
		bb_perror_msg("%s", path);
		status = EXIT_FAILURE;
		return (NULL);  /* could not open the dir */
	}
	while ((entry = readdir(dir)) != NULL) {
		char *fullname;

		/* are we going to list the file- it may be . or .. or a hidden file */
		if (entry->d_name[0] == '.') {
			if ((entry->d_name[1] == 0 || (
					entry->d_name[1] == '.'
					&& entry->d_name[2] == 0))
			     && !(all_fmt & DISP_DOT))
				continue;
			if (!(all_fmt & DISP_HIDDEN))
				continue;
		}
		fullname = concat_path_file(path, entry->d_name);
		cur = my_stat(fullname, bb_basename(fullname));
		if (!cur)
			continue;
		cur->allocated = dn;
		dn = cur;
		nfiles++;
	}
	closedir(dir);

	/* now that we know how many files there are
	   ** allocate memory for an array to hold dnode pointers
	 */
	if (dn != NULL) {
		dnp = dnalloc(nfiles);
		for (i = 0, cur = dn; i < nfiles; i++) {
			dnp[i] = cur;   /* save pointer to node in array */
			cur = cur->allocated;
		}
		/* last fullname need free too */
		dnp[i-1]->allocated = dn;
	}
	*pnfiles = nfiles;
	return (dnp);
}

/*----------------------------------------------------------------------*/
static int list_single(struct dnode *dn)
{
	int i, column = 0;
	char buf[12];
#ifdef CONFIG_FEATURE_LS_USERNAME
	char scratch[16];
#endif
#ifdef CONFIG_FEATURE_LS_TIMESTAMPS
	char *filetime;
	time_t ttime, age;
#endif
#if defined(CONFIG_FEATURE_LS_FILETYPES) || defined (CONFIG_FEATURE_LS_COLOR)
	struct stat info;
#endif
#if defined(CONFIG_FEATURE_LS_FILETYPES)
	char append;
#endif

	if (dn->fullname == NULL)
		return 0;

#ifdef CONFIG_FEATURE_LS_TIMESTAMPS
	ttime = dn->dstat.st_mtime;     /* the default time */
	if (all_fmt & TIME_ACCESS)
		ttime = dn->dstat.st_atime;
	if (all_fmt & TIME_CHANGE)
		ttime = dn->dstat.st_ctime;
	filetime = ctime(&ttime);
#endif
#ifdef CONFIG_FEATURE_LS_FILETYPES
	append = append_char(dn->dstat.st_mode);
#endif

	for (i = 0; i <= 31; i++) {
		switch (all_fmt & (1 << i)) {
		case LIST_INO:
			column += printf("%7ld ", (long int) dn->dstat.st_ino);
			break;
		case LIST_BLOCKS:
#if _FILE_OFFSET_BITS == 64
			column += printf("%4lld ", (long long)dn->dstat.st_blocks >> 1);
#else
			column += printf("%4ld ", (long)(dn->dstat.st_blocks >> 1));
#endif
			break;
		case LIST_MODEBITS:
			column += printf("%-10s ", bb_mode_string(dn->dstat.st_mode, buf));
			break;
		case LIST_NLINKS:
			column += printf("%4ld ", (long) dn->dstat.st_nlink);
			break;
		case LIST_ID_NAME:
#ifdef CONFIG_FEATURE_LS_USERNAME
			bb_getpwuid(scratch, dn->dstat.st_uid, sizeof(scratch));
			printf("%-8.8s ", scratch);
			bb_getgrgid(scratch, dn->dstat.st_gid, sizeof(scratch));
			printf("%-8.8s", scratch);
			column += 17;
			break;
#endif
		case LIST_ID_NUMERIC:
			column += printf("%-8d %-8d", (int)dn->dstat.st_uid, (int)dn->dstat.st_gid);
			break;
		case LIST_SIZE:
		case LIST_DEV:
			if (S_ISBLK(dn->dstat.st_mode) || S_ISCHR(dn->dstat.st_mode)) {
				column += printf("%4d, %3d ", (int) major(dn->dstat.st_rdev),
					   (int) minor(dn->dstat.st_rdev));
			} else {
#ifdef CONFIG_FEATURE_HUMAN_READABLE
				if (all_fmt & LS_DISP_HR) {
					char hr_buf[32];

					column += printf("%9s ",
							make_human_readable_str(dn->dstat.st_size, 1, 0, hr_buf));
				} else
#endif
				{
					column += printf("%9" P_ST_SIZE "d ", dn->dstat.st_size);
				}
			}
			break;
#ifdef CONFIG_FEATURE_LS_TIMESTAMPS
		case LIST_FULLTIME:
			printf("%24.24s ", filetime);
			column += 25;
			break;
		case LIST_DATE_TIME:
			if ((all_fmt & LIST_FULLTIME) == 0) {
				age = time(NULL) - ttime;
				printf("%6.6s ", filetime + 4);
				if (age < 3600L * 24 * 365 / 2 && age > -15 * 60) {
					/* hh:mm if less than 6 months old */
					printf("%5.5s ", filetime + 11);
				} else {
					printf(" %4.4s ", filetime + 20);
				}
				column += 13;
			}
			break;
#endif
		case LIST_FILENAME:
#ifdef CONFIG_FEATURE_LS_COLOR
			errno = 0;
			if (show_color && !lstat(dn->fullname, &info)) {
				printf("\033[%d;%dm", bgcolor(info.st_mode),
					   fgcolor(info.st_mode));
			}
#endif
			column += printf("%s", dn->name);
#ifdef CONFIG_FEATURE_LS_COLOR
			if (show_color) {
				printf("\033[0m");
			}
#endif
			break;
		case LIST_SYMLINK:
			if (S_ISLNK(dn->dstat.st_mode)) {
				char *lpath = xreadlink(dn->fullname);

				if (lpath) {
					printf(" -> ");
#if defined(CONFIG_FEATURE_LS_FILETYPES)
					if (!stat(dn->fullname, &info)) {
						append = append_char(info.st_mode);
					}
#endif
#ifdef CONFIG_FEATURE_LS_COLOR
					if (show_color) {
#if !defined(CONFIG_FEATURE_LS_FILETYPES)
						stat(dn->fullname, &info);
#endif
						errno = 0;
						printf("\033[%d;%dm", bgcolor(info.st_mode),
							   fgcolor(info.st_mode));
					}
#endif
					column += printf("%s", lpath) + 4;
#ifdef CONFIG_FEATURE_LS_COLOR
					if (show_color) {
						printf("\033[0m");
					}
#endif
					free(lpath);
				}
			}
			break;
#ifdef CONFIG_FEATURE_LS_FILETYPES
		case LIST_FILETYPE:
			if (append != '\0') {
				printf("%1c", append);
				column++;
			}
			break;
#endif
		}
	}

	return column;
}


/*----------------------------------------------------------------------*/
extern int ls_main(void)
{
	struct dnode **dnd;
	struct dnode **dnf;
	struct dnode **dnp;
	struct dnode *cur;
	long opt;
	int nfiles = 0;
	int dnfiles;
	int dndirs;
	int oi;
	int ac;
	int i;
	char **av;
#if (ENABLE_FEATURE_LONG_OPTIONS+ENABLE_FEATURE_LS_COLOR) > 1
	char *color_opt = NULL;
#endif

	all_fmt = LIST_SHORT;

#ifdef CONFIG_FEATURE_AUTOWIDTH
	/* Obtain the terminal width.  */
	get_terminal_width_height(&terminal_width, NULL);
	/* Go one less... */
	terminal_width--;
#endif

#if ENABLE_FEATURE_LONG_OPTIONS
	bb_applet_long_options = ls_opt;
#endif

	/* process options */
	opt = bb_xgetopt_ulflags(opt_str
			USE_FEATURE_AUTOWIDTH(, &tabstops, &terminal_width)
#if (ENABLE_FEATURE_LONG_OPTIONS+ENABLE_FEATURE_LS_COLOR) > 1
			USE_FEATURE_LS_COLOR(, &color_opt)
#endif
			);
	for (i = 0; opt_flags[i] != (1U<<31); i++) {
		if (opt & (1 << i))
			all_fmt |= opt_flags[i];
	}

#ifdef CONFIG_FEATURE_LS_COLOR
	{
		/* find color bit value - last position for short getopt */

		char *p = getenv ("LS_COLORS");

		if (p == NULL || (p != NULL &&
			(*p == '\0' || (strcmp(p, "none") == 0)))) {
#if ENABLE_FEATURE_LONG_OPTIONS
			if (color_opt == NULL)
				color_opt = "never";
#else
			;
#endif
		} else if (isatty(STDOUT_FILENO)) {
			show_color = 1;
		}

#if ENABLE_FEATURE_LONG_OPTIONS
		if((opt & COLOR_OPT)) {
			if (color_opt == NULL || strcmp("always", color_opt) == 0)
				show_color = 1;
			else if (strcmp("never", color_opt) == 0)
				show_color = 0;
			else if (strcmp("auto", color_opt) == 0 && isatty(STDOUT_FILENO))
				show_color = 1;
		}
#endif
	}
#endif  /* CONFIG_FEATURE_LS_COLOR */

#ifdef CONFIG_FEATURE_LS_USERNAME
	if ((all_fmt & LIST_ID_NUMERIC))
		all_fmt &= ~LIST_ID_NAME;       /* don't list names if numeric uid */
#endif

	/* choose a display format */
	if ((all_fmt & (STYLE_COLUMNS | STYLE_SINGLE)) == 0)
		all_fmt |= (isatty(STDOUT_FILENO) ? STYLE_COLUMNS : STYLE_SINGLE);

	/*
	 * when there are no cmd line args we have to supply a default "." arg.
	 * we will create a second argv array, "av" that will hold either
	 * our created "." arg, or the real cmd line args.  The av array
	 * just holds the pointers- we don't move the date the pointers
	 * point to.
	 */
	ac = Gargc;             /* how many cmd line args are left */
	if (ac < 1) {
		static const char * const dotdir[] = { "." };

		av = (char **) dotdir;
		ac = 1;
	} else {
		av = Gargv;
	}

	/* now, everything is in the av array */
	if (ac > 1)
		all_fmt |= DISP_DIRNAME;        /* 2 or more items? label directories */

	/* stuff the command line file names into an dnode array */
	dnp = dnalloc(ac);
	for (oi = 0; oi < ac; oi++) {
		cur = my_stat(av[oi], av[oi]);
		if (!cur)
			continue;
		cur->allocated = NULL;
		dnp[nfiles++] = cur;
	}

	if (all_fmt & DISP_NOLIST) {
		USE_FEATURE_LS_SORTFILES(dnsort(dnp, nfiles));
		if (nfiles > 0)
			showfiles(dnp, nfiles);
	} else {
		dnd = splitdnarray(dnp, nfiles, SPLIT_DIR);
		dnf = splitdnarray(dnp, nfiles, SPLIT_FILE);
		dndirs = countdirs(dnp, nfiles);
		dnfiles = nfiles - dndirs;
		if (dnfiles > 0) {
			USE_FEATURE_LS_SORTFILES(dnsort(dnf, dnfiles));
			showfiles(dnf, dnfiles);
		}
		if (dndirs > 0) {
			USE_FEATURE_LS_SORTFILES(dnsort(dnd, dndirs));
			showdirs(dnd, dndirs, dnfiles == 0);
		}
	}
	return (status);
}
#endif /* APPLET_USAGE */
