/*
 *  Copyright (C) 2003 Glenn L. McGrath
 *  Copyright (C) 2003-2004 Erik Andersen
 *
 * Licensed under the GPL v2 or later, see the file LICENSE in this tarball.
 */
#ifdef APPLET_USAGE
#define md5sum_trivial_usage \
	"[OPTION] [FILEs...]" \
	USE_FEATURE_MD5_SHA1_SUM_CHECK("\n or: md5sum [OPTION] -c [FILE]")
#define md5sum_full_usage \
	"Print" USE_FEATURE_MD5_SHA1_SUM_CHECK(" or check") " MD5 checksums.\n\n" \
	"Options:\n" \
	"With no FILE, or when FILE is -, read standard input." \
	USE_FEATURE_MD5_SHA1_SUM_CHECK("\n\n" \
	"\t-c\tcheck MD5 sums against given list\n" \
	"\nThe following two options are useful only when verifying checksums:\n" \
	"\t-s\tdon't output anything, status code shows success\n" \
	"\t-w\twarn about improperly formated MD5 checksum lines")

#define sha1sum_trivial_usage \
	"[OPTION] [FILEs...]" \
	USE_FEATURE_MD5_SHA1_SUM_CHECK("\n or: sha1sum [OPTION] -c [FILE]")
#define sha1sum_full_usage \
	"Print" USE_FEATURE_MD5_SHA1_SUM_CHECK(" or check") " SHA1 checksums.\n\n" \
	"Options:\n" \
	"With no FILE, or when FILE is -, read standard input." \
	USE_FEATURE_MD5_SHA1_SUM_CHECK("\n\n" \
	"\t-c\tcheck SHA1 sums against given list\n" \
	"\nThe following two options are useful only when verifying checksums:\n" \
	"\t-s\tdon't output anything, status code shows success\n" \
	"\t-w\twarn about improperly formated SHA1 checksum lines")
#else

#include "busybox.h"
#include <string.h>
#include <unistd.h>

#include "bb_xopen.h"
#ifdef CONFIG_SHA1SUM
#include "sha1.h"
#endif
#ifdef CONFIG_MD5SUM
#include "md5.h"
#endif

typedef enum { HASH_SHA1, HASH_MD5 } hash_algo_t;

#define FLAG_SILENT     1
#define FLAG_CHECK      2
#define FLAG_WARN       4

/* This might be useful elsewhere */
static unsigned char *hash_bin_to_hex(unsigned char *hash_value, int hash_length)
{
	int x;
	unsigned char *hex_value;

	hex_value = xmalloc((hash_length * 2) + 2);
	for (x = 0; x < hash_length; x++)
		sprintf((char*)(hex_value + x*2), "%02x", hash_value[x]);
	return hex_value;
}

CALC_BSS struct BB_COMMON_BSS
{
	unsigned char in_buf[4096];
} bb_common_bss;
#define in_buf bb_common_bss.in_buf

static uint8_t *hash_file(const char *filename, hash_algo_t hash_algo)
{
	int src_fd, hash_len, count;
	union _ctx_ {
		sha1_ctx_t sha1;
		md5_ctx_t md5;
	} context;
	uint8_t *hash_value = NULL;
	void (*update)(const void*, size_t, void*);
	void (*final)(void*, void*);

	if((src_fd = bb_xopen(filename, O_RDONLY | BB_OPEN_WARN | BB_OPEN_MINUS_IS_STDIN)) >= 0)
		return NULL;

	// figure specific hash algorithims
	if(ENABLE_MD5SUM && hash_algo == HASH_MD5) {
		md5_begin(&context.md5);
		update = (void (*)(const void*, size_t, void*))md5_hash;
		final = (void (*)(void*, void*))md5_end;
		hash_len = 16;
	} else if(ENABLE_SHA1SUM && hash_algo == HASH_SHA1) {
		sha1_begin(&context.sha1);
		update = (void (*)(const void*, size_t, void*))sha1_hash;
		final = (void (*)(void*, void*))sha1_end;
		hash_len = 20;
	} else {
		bb_error_msg_and_die("algorithm not supported");
	}


	while(0 < (count = read(src_fd, in_buf, sizeof in_buf))) {
		update(in_buf, count, &context);
	}

	if(count == 0) {
		final(in_buf, &context);
		hash_value = hash_bin_to_hex(in_buf, hash_len);
	}

	if(src_fd != STDIN_FILENO)
		close(src_fd);

	return hash_value;
}

/* This could become a common function for md5 as well, by using md5_stream */
static int hash_files(hash_algo_t hash_algo)
{
	char **argv;
	int return_value = EXIT_SUCCESS;
	uint8_t *hash_value;
#if ENABLE_FEATURE_MD5_SHA1_SUM_CHECK
	int flags;
#else
# define flags 0
#endif

	if (ENABLE_FEATURE_MD5_SHA1_SUM_CHECK) {
		/* if no arg then set "-" */
		flags = (int)bb_xgetopt_ulflags("scw\0-0");
	} else {
		bb_xgetopt_ulflags("\0-0");
	}
	argv = Gargv;

	if (ENABLE_FEATURE_MD5_SHA1_SUM_CHECK && !(flags & FLAG_CHECK)) {
		if (flags & FLAG_SILENT) {
			bb_error_msg_and_die
				("the -%c option is meaningful only when verifying checksums", 's');
		} else if (flags & FLAG_WARN) {
			bb_error_msg_and_die
				("the -%c option is meaningful only when verifying checksums", 'w');
		}
	}

	if (ENABLE_FEATURE_MD5_SHA1_SUM_CHECK && flags & FLAG_CHECK) {
		FILE *pre_computed_stream;
		int count_total = 0;
		int count_failed = 0;
		char *file_ptr = *argv;
		char *line;

		if (argv[1] != NULL)
			bb_error_msg_and_die("only one argument may be specified when using -c");

		pre_computed_stream = bb_xfopen(file_ptr, "A-r");

		while ((line = bb_get_chomped_line_from_file(pre_computed_stream)) != NULL) {
			char *filename_ptr;

			count_total++;
			filename_ptr = strstr(line, "  ");
			if (filename_ptr == NULL) {
				if (flags & FLAG_WARN) {
					bb_error_msg("Invalid format");
				}
				count_failed++;
				return_value = EXIT_FAILURE;
				free(line);
				continue;
			}
			*filename_ptr = '\0';
			filename_ptr += 2;

			hash_value = hash_file(filename_ptr, hash_algo);

			if (hash_value && (strcmp((char*)hash_value, line) == 0)) {
				if (!(flags & FLAG_SILENT))
					printf("%s: OK\n", filename_ptr);
			} else {
				if (!(flags & FLAG_SILENT))
					printf("%s: FAILED\n", filename_ptr);
				count_failed++;
				return_value = EXIT_FAILURE;
			}
			/* possible free(NULL) */
			free(hash_value);
			free(line);
		}
		if (count_failed && !(flags & FLAG_SILENT)) {
			bb_error_msg("WARNING: %d of %d computed checksums did NOT match",
						 count_failed, count_total);
		}
		if (bb_fclose_nonstdin(pre_computed_stream) == EOF) {
			bb_perror_msg_and_die("Couldnt close file %s", file_ptr);
		}
	} else {
		for(; *argv; argv++) {
			char *file_ptr = *argv;

			hash_value = hash_file(file_ptr, hash_algo);
			if (hash_value == NULL) {
				return_value = EXIT_FAILURE;
			} else {
				printf("%s  %s\n", hash_value, file_ptr);
				free(hash_value);
			}
		}
	}
	return return_value;
}

#ifdef CONFIG_MD5SUM
int md5sum_main(void)
{
	return hash_files(HASH_MD5);
}
#endif

#ifdef CONFIG_SHA1SUM
int sha1sum_main(void)
{
	return hash_files(HASH_SHA1);
}
#endif
#endif /* APPLET_USAGE */
