/* printf - format and print data

   Copyright 1999 Dave Cinege
   Portions copyright (C) 1990-1996 Free Software Foundation, Inc.

   Licensed under GPL v2 or later, see file LICENSE in this tarball for details.
*/

/*
   A front end to the printf function that lets it be used from the shell.

   Backslash escapes:

   \" = double quote
   \\ = backslash
   \a = alert (bell)
   \b = backspace
   \c = produce no further output
   \f = form feed
   \n = new line
   \r = carriage return
   \t = horizontal tab
   \v = vertical tab
   \0ooo = octal number (ooo is 0 to 3 digits)
   \xhhh = hexadecimal number (hhh is 1 to 3 digits)

   Additional directive:

   %b = print an argument string, interpreting backslash escapes

   The `format' argument is re-used as many times as necessary
   to convert all of the given arguments.

   David MacKenzie <djm@gnu.ai.mit.edu>

   19990508 Busy Boxed! Dave Cinege
   */
#ifdef APPLET_USAGE
#define printf_trivial_usage \
	"FORMAT [ARGUMENT...]"
#define printf_full_usage \
	"Formats and prints ARGUMENT(s) according to FORMAT,\n" \
	"Where FORMAT controls the output exactly as in C printf."
#else

#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include "busybox.h"

typedef int (*converter)(const char *arg, void *result);

static void multiconvert(const char *arg, void *result, converter convert)
{
	char s[16];
	if (*arg == '"' || *arg == '\'') {
		sprintf(s, "%d", (unsigned)*(++arg));
		arg = s;
	}
	if(convert(arg, result))
		bb_error_msg("%s: invalid number", arg);
}

static int printf_safe_strtoul(char *arg, long* value)
{
	*value = bb_xparse_ul_number(0 | BB_PN_SILENT_RE, arg);
	if (errno != 0)
		return 1;
	return 0;
}

static unsigned long xstrtoul(char *arg)
{
	unsigned long result;

	multiconvert(arg, &result, (converter)printf_safe_strtoul);
	return result;
}

static int printf_safe_strtol(char *arg, long* value)
{
	*value = bb_xparse_l_number(BB_PN_LOWER_MIN | BB_PN_SILENT_RE, arg);
	if (errno != 0)
		return 1;
	return 0;
}

static long xstrtol(char *arg)
{
	long result;
	multiconvert(arg, &result, (converter)printf_safe_strtol);
	return result;
}

static double xstrtod(char *arg)
{
	double result;
	multiconvert(arg, &result, (converter)safe_strtod);
	return result;
}

static void print_esc_string(char *str)
{
	for (; *str; str++) {
		if (*str == '\\') {
			str++;
			putchar(bb_process_escape_sequence((const char **)&str));
		} else {
			putchar(*str);
		}

	}
}

static void
print_direc(char *start, size_t length, int field_width, int precision,
			char *argument)
{
	char *p;        /* Null-terminated copy of % directive. */

	p = xstrndup(start, length);

	switch (p[length - 1]) {
	case 'd':
	case 'i':
		if (field_width < 0) {
			if (precision < 0)
				printf(p, xstrtol(argument));
			else
				printf(p, precision, xstrtol(argument));
		} else {
			if (precision < 0)
				printf(p, field_width, xstrtol(argument));
			else
				printf(p, field_width, precision, xstrtol(argument));
		}
		break;

	case 'o':
	case 'u':
	case 'x':
	case 'X':
		if (field_width < 0) {
			if (precision < 0)
				printf(p, xstrtoul(argument));
			else
				printf(p, precision, xstrtoul(argument));
		} else {
			if (precision < 0)
				printf(p, field_width, xstrtoul(argument));
			else
				printf(p, field_width, precision, xstrtoul(argument));
		}
		break;

	case 'f':
	case 'e':
	case 'E':
	case 'g':
	case 'G':
		if (field_width < 0) {
			if (precision < 0)
				printf(p, xstrtod(argument));
			else
				printf(p, precision, xstrtod(argument));
		} else {
			if (precision < 0)
				printf(p, field_width, xstrtod(argument));
			else
				printf(p, field_width, precision, xstrtod(argument));
		}
		break;

	case 'c':
		printf(p, *argument);
		break;

	case 's':
		if (field_width < 0) {
			if (precision < 0)
				printf(p, argument);
			else
				printf(p, precision, argument);
		} else {
			if (precision < 0)
				printf(p, field_width, argument);
			else
				printf(p, field_width, precision, argument);
		}
		break;
	case '%':
		bb_error_msg_and_die("`%%': missing format character");
	default:
		bb_error_msg_and_die("`%c': invalid format character", p[length - 1]);
	}


	free(p);
}

/* Print the text in FORMAT, using ARGV (with ARGC elements) for
   arguments to any `%' directives.
   Return the number of elements of ARGV used.  */

static int print_formatted(char *format, int argc, char **argv)
{
	int save_argc = argc;           /* Preserve original value.  */
	char *f;                        /* Pointer into `format'.  */
	char *direc_start;              /* Start of % directive.  */
	size_t direc_length;            /* Length of % directive.  */
	int field_width;                /* Arg to first '*', or -1 if none.  */
	int precision;                  /* Arg to second '*', or -1 if none.  */

	for (f = format; *f; ++f) {
		switch (*f) {
		case '%':
			direc_start = f++;
			direc_length = 1;
			field_width = precision = -1;
			if (*f == '%') {
				putchar('%');
				break;
			}
			if (*f == 'b') {
				if (argc > 0) {
					print_esc_string(*argv);
					++argv;
					--argc;
				}
				break;
			}
			if (*f && strchr("-+ #", *f)) {
				++f;
				++direc_length;
			}
			if (*f == '*') {
				++f;
				++direc_length;
				if (argc > 0) {
					field_width = xstrtoul(*argv);
					++argv;
					--argc;
				} else
					field_width = 0;
			} else
				while (isdigit(*f)) {
					++f;
					++direc_length;
				}
			if (*f == '.') {
				++f;
				++direc_length;
				if (*f == '*') {
					++f;
					++direc_length;
					if (argc > 0) {
						precision = xstrtoul(*argv);
						++argv;
						--argc;
					} else
						precision = 0;
				} else
					while (isdigit(*f)) {
						++f;
						++direc_length;
					}
			}
			if (*f == 'l' || *f == 'L' || *f == 'h') {
				++f;
				++direc_length;
			}
			if(*f != '\0')
				++direc_length;
			/*
			   if (!strchr ("diouxXfeEgGcs", *f))
			   fprintf(stderr, "%%%c: invalid directive", *f);
			 */

			if (argc > 0) {
				print_direc(direc_start, direc_length, field_width,
							precision, *argv);
				++argv;
				--argc;
			} else
				print_direc(direc_start, direc_length, field_width,
							precision, "");
			break;

		case '\\':
			if (*++f == 'c')
				exit(0);
			putchar(bb_process_escape_sequence((const char **)&f));
			f--;
			break;

		default:
			putchar(*f);
		}
	}

	return save_argc - argc;
}

int printf_main(void)
{
	int argc = Gargc;
	char **argv = Gargv;
	char *format;
	int args_used;

	if (argc <= 1 || **(argv + 1) == '-') {
		bb_show_usage();
	}

	format = argv[1];
	argc -= 2;
	argv += 2;

	do {
		args_used = print_formatted(format, argc, argv);
		argc -= args_used;
		argv += args_used;
	}
	while (args_used > 0 && argc > 0);

/*
  if (argc > 0)
    fprintf(stderr, "excess args ignored");
*/

	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
