/* Copyright (C) 2009 Gabor Kovesdan <gabor@FreeBSD.org>
 * Copyright (C) 2012 Oleg Moskalenko <mom040267@gmail.com>
 * All rights reserved.
 *
 * Copyright (c) 2016 Vladimir Oleynik <dzo@simtreas.ru>
 * may safely be consumed by a BSD or GPL license.
 *  Busybox port
 *  restore configurable for non multibyte
 *  support -t C-ESCape char (multibyte too)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#ifdef APPLET_USAGE
#define sort_trivial_usage \
	"[-bcCdfgimnrsuzR] [-k field1[,field2]] [-o output] [-S size] [-T dir] [-t char] [file ...]"
#define sort_full_usage \
	"Sorts lines of text in the specified files\n\n" \
	"Options:\n" \
	"\t-b\tignore leading blanks\n" \
	"\t-c\tcheck whether input is sorted\n" \
	"\t-C\tlike -c, but do not report first bad line\n" \
	"\t-d\tdictionary order (blank or alphanumeric only)\n" \
	"\t-f\tignore case\n" \
	"\t-g\tcompare according to general numerical value\n" \
	"\t-i\tignore unprintable characters\n" \
	"\t-m\tmerge only; the input files are assumed to be pre-sorted\n" \
	"\t-n\tsort numbers\n" \
	"\t-o\toutput to file\n" \
	"\t-k\tsort by key\n" \
	"\t-t\tuse key separator other than whitespace\n" \
	"\t-r\treverse sort order\n" \
	"\t-s\tstabilize sort by disabling last-resort comparison\n" \
	"\t-u\tsuppress duplicate lines\n" \
	"\t-z\tend lines with 0 byte, not newline\n" \
	"\t-R\tsort by random hash of keys"
#else

#define VERSION "2.3-OpenBSD" /* with GNUSORT_COMPATIBILITY */

#include "busybox.h"
#include <sys/mman.h>
#include <sys/resource.h>
#include <stdbool.h>
#include <ctype.h>
#ifdef HAVE_LANGINFO
# include <langinfo.h>
#endif
#include <math.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <getopt.h>
#include <locale.h>
#include <regex.h>
#include "bb_sig.h"
#include "bb_xopen.h"
#include "md5.h"
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
# include <wchar.h>
# include <wctype.h>
#endif

/* #include "md5.h" */
#define MD5_BLOCK_LENGTH                64
#define MD5_DIGEST_LENGTH               16
#define MD5_DIGEST_STRING_LENGTH        (MD5_DIGEST_LENGTH * 2 + 1)

#define OPTIONS "+bCcdfgHhik:Mmno:RrS:st:T:uVz"

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
#define WCHAR_T_UTF8 wchar_t
#define Lc(c) L##c
#define WCSCMP(s1, s2) wcscmp(s1, s2)
#define ISWBLANK(c) iswblank(c)
#define ISWDIGIT(c) iswdigit(c)
#define WCSTOULL(s, e, n) wcstoull(s, e, n)
#else
typedef unsigned char WCHAR_T_UTF8;
#define Lc(c) c
#define WCSCMP(s1, s2) strcmp((char *)s1, (char *)s2)
#define ISWBLANK(c) isblank(c)
#define ISWDIGIT(c) isdigit(c)
#define WCSTOULL(s, e, n) strtoull((char *)s, e, n)
#endif

#define SIZEOF_WCHAR_STRING(LEN) ((LEN)*sizeof(WCHAR_T_UTF8))



/* Binary "wide" string */
struct bwstring {
	size_t len;
	union {
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
		wchar_t         wstr[0];
#endif
		unsigned char   cstr[0];
	} data;
};

struct reader_buffer {
	WCHAR_T_UTF8 *fgetwln_z_buffer;
	size_t   fgetwln_z_buffer_size;
};

typedef void *bwstring_iterator;

#define BWSLEN(s) ((s)->len)

/* Placeholder symbols for options which have no single-character equivalent */
enum {
	SORT_OPT = CHAR_MAX + 1,
	FF_OPT,
	BS_OPT,
	VERSION_OPT,
	DEBUG_OPT,
	RANDOMSOURCE_OPT,
	COMPRESSPROGRAM_OPT,
	QSORT_OPT,
	HEAPSORT_OPT,
	RADIXSORT_OPT,
	MMAP_OPT
};

/* Status of a sort hint object */
typedef enum {
	HS_ERROR = -1, HS_UNINITIALIZED = 0, HS_INITIALIZED = 1
} hint_status;

/* Sort hint data for -n */
struct n_hint {
	unsigned long long       n1;
	unsigned char            si;
	bool                     empty;
	bool                     neg;
};

/* Sort hint data for -g */
struct g_hint {
	double d;
#ifdef HAVE_ISNAN
	bool   nan;
#endif
	bool   notnum;
};

/* Sort hint data for -M */
struct M_hint {
	int m;
};

/* Sort hint object */
struct key_hint {
	hint_status status;
	union {
		struct n_hint           nh;
		struct g_hint           gh;
		struct M_hint           Mh;
	} v;
};

/* Key value */
struct key_value {
	struct bwstring         *k; /* key string */
	struct key_hint          hint[0]; /* key sort hint */
};

/* Set of keys container object. */
struct keys_array {
	struct key_value        key[0];
};

/* Cmp function */
typedef int (*cmpcoll_t)(struct key_value *kv1, struct key_value *kv2, size_t offset);

/* This structure holds "sort modifiers" - options which are affecting the sort ordering. */
struct sort_mods {
	cmpcoll_t       func;
	bool            bflag;
	bool            dflag;
	bool            fflag;
	bool            gflag;
	bool            iflag;
	bool            Mflag;
	bool            nflag;
	bool            rflag;
	bool            Rflag;
	bool            Vflag;
	bool            hflag;
};

/* Parsed -k option data */
struct key_specs {
	struct sort_mods        sm;
	size_t                  c1;
	size_t                  c2;
	size_t                  f1;
	size_t                  f2;
	bool                    pos1b;
	bool                    pos2b;
};

/* Single entry in sort list. */
struct sort_list_item {
	struct bwstring         *str;
	struct keys_array        ka;
};

#define DEFAULT_SORT_FUNC_RADIXSORT mergesort

#define TINY_NODE(sl) ((sl)->tosort_num < 65)
#define SMALL_NODE(sl) ((sl)->tosort_num < 5)

/* sort sub-levels array size */
static const size_t slsz = 256 * sizeof(struct sort_level *);

/* one sort level structure */
struct sort_level {
	struct sort_level       **sublevels;
	struct sort_list_item   **leaves;
	struct sort_list_item   **sorted;
	struct sort_list_item   **tosort;
	size_t                    leaves_num;
	size_t                    leaves_sz;
	size_t                    level;
	size_t                    real_sln;
	size_t                    start_position;
	size_t                    sln;
	size_t                    tosort_num;
	size_t                    tosort_sz;
};

/* stack of sort levels ready to be sorted */
struct level_stack {
	struct level_stack       *next;
	struct sort_level        *sl;
};



/* Function type, used to compare two list objects */
typedef int (*listcoll_t)(struct sort_list_item **ss1, struct sort_list_item **ss2);

#define SORT_DEFAULT    0
#define SORT_QSORT      1
#define SORT_MERGESORT  2
#define SORT_HEAPSORT   3
#define SORT_RADIXSORT  4

#define DEFAULT_SORT_ALGORITHM SORT_HEAPSORT
#define DEFAULT_SORT_FUNC heapsort

/* List of data to be sorted. */
struct sort_list {
	struct sort_list_item   **list;
	unsigned long long       memsize;
	size_t                   count;
	size_t                   size;
	size_t                   sub_list_pos;
};

/* File reader structure */
struct file_reader {
	struct reader_buffer     rb;
	FILE                    *file;
	char                    *fname;
	unsigned char           *buffer;
	unsigned char           *mmapaddr;
	unsigned char           *mmapptr;
	size_t                   bsz;
	size_t                   cbsz;
	size_t                   mmapsize;
	size_t                   strbeg;
	char                     elsymb;
};

/* List of files to be sorted */
struct file_list {
	char                    **fns;
	size_t                   count;
	size_t                   sz;
	bool                     tmp;
};

CALC_BSS struct BB_COMMON_BSS
{
	bool need_random;
	const char *random_source;
#ifdef CONFIG_LOCALE_SUPPORT
	bool byte_sort;
#endif
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	size_t sort_mb_cur_max;
	wchar_t **wmonths;
#endif
	char **cmonths;
	md5_ctx_t md5_ctx;
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	wint_t          field_sep;
#else
	int             field_sep;
#endif
	int             sort_method;
	bool            cflag;
	bool            csilentflag;
	bool            kflag;
	bool            mflag;
	bool            sflag;
	bool            uflag;
	bool            zflag;
	bool            tflag;
	bool            complex_sort;
	/* If true, we output some debug information. */
#ifdef CONFIG_FEATURE_SORT_BIG
	bool debug_sort;
#endif
	bool need_hint;

	struct sort_mods default_sort_mods_object;

	bool print_symbols_on_debug;
	/* Arguments from file (when file0-from option is used: */
	size_t argc_from_file0;
	char **argv_from_file0;

	struct key_specs *keys;
	size_t keys_num;
	/* are we sorting in reverse order ? */
	bool reverse_sort;

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	/* there is no default thousands separator in collate rules: */
	wint_t symbol_thousands_sep;
	/* Main localised symbols. These must be wint_t as they may hold WEOF. */
	wint_t symbol_decimal_point;
	wint_t symbol_negative_sign;
	wint_t symbol_positive_sign;
#else
	int symbol_decimal_point;
	int symbol_negative_sign;
	int symbol_positive_sign;
#endif
	unsigned long long available_free_memory;
	/* Are we using mmap ? */
	bool use_mmap;
	/* temporary file dir */
	const char *tmpdir;
	/* Compress program */
	const char *compress_program;
	/* Max number of simultaneously open files (including the output file). */
	size_t max_open_files;
	struct level_stack *g_ls;
	/* List header of "cleanable" files list. */
	llist_t *tmp_files;
} bb_common_bss;
#ifdef CONFIG_LOCALE_SUPPORT
# define byte_sort bb_common_bss.byte_sort
# define wmonths bb_common_bss.wmonths
#else
# define byte_sort 1
#endif
#define cmonths bb_common_bss.cmonths
#define sort_mb_cur_max bb_common_bss.sort_mb_cur_max
#ifdef CONFIG_FEATURE_SORT_BIG
# define debug_sort bb_common_bss.debug_sort
#else
# define debug_sort 0
#endif
#define md5_ctx bb_common_bss.md5_ctx
#define keys_num bb_common_bss.keys_num
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
# define symbol_thousands_sep bb_common_bss.symbol_thousands_sep
#else
# define symbol_thousands_sep 0
#endif
#define symbol_decimal_point bb_common_bss.symbol_decimal_point
#define symbol_negative_sign bb_common_bss.symbol_negative_sign
#define symbol_positive_sign bb_common_bss.symbol_positive_sign
#define keys bb_common_bss.keys
#define default_sort_mods_object bb_common_bss.default_sort_mods_object
#define g_ls bb_common_bss.g_ls
#define tmp_files bb_common_bss.tmp_files

struct sort_mods * const default_sort_mods = &default_sort_mods_object;

/* This is sqrt(SIZE_MAX+1), as s1*s2 <= SIZE_MAX
 * if both s1 < MUL_NO_OVERFLOW and s2 < MUL_NO_OVERFLOW */
#define MUL_NO_OVERFLOW        ((size_t)1 << (sizeof(size_t) * 4))

static void *sort_reallocarray(void *optr, size_t nmemb, size_t size)
{
	if ((nmemb >= MUL_NO_OVERFLOW || size >= MUL_NO_OVERFLOW) &&
	    nmemb > 0 && SIZE_MAX / nmemb < size) {
		errno = ENOMEM;
		bb_perror_msg_and_die("reallocarray");
	}
	return xrealloc(optr, size * nmemb);
}

static inline bwstring_iterator bws_begin(struct bwstring *bws)
{

	return (bwstring_iterator) (&(bws->data));
}

static inline bwstring_iterator bws_end(struct bwstring *bws)
{
	return
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
		((sort_mb_cur_max == 1) ?
			(bwstring_iterator) (bws->data.cstr + bws->len) :
			(bwstring_iterator) (bws->data.wstr + bws->len));
#else
	    (bwstring_iterator) (bws->data.cstr + bws->len);
#endif
}

static inline bwstring_iterator bws_iterator_inc(bwstring_iterator iter, size_t pos)
{
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	if (sort_mb_cur_max == 1)
#endif
		return ((unsigned char *) iter) + pos;
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	else
		return ((wchar_t*) iter) + pos;
#endif
}

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
static inline wchar_t bws_get_iter_value(bwstring_iterator iter)
{
	if (sort_mb_cur_max == 1)
		return *((unsigned char *) iter);
	else
		return *((wchar_t*) iter);
}
#else
static inline unsigned char bws_get_iter_value(bwstring_iterator iter)
{
	return *((unsigned char *) iter);
}
#endif

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
#define BWS_GET(bws, pos) ((sort_mb_cur_max == 1) ? ((bws)->data.cstr[(pos)]) : (bws)->data.wstr[(pos)])
#else
#define BWS_GET(bws, pos) ((bws)->data.cstr[(pos)])
#endif


static void initialise_months(void)
{
#ifdef HAVE_LANGINFO
# define NL_LANGINFO(i) nl_langinfo(i)
	const nl_item item[12] = { ABMON_1, ABMON_2, ABMON_3, ABMON_4,
	    ABMON_5, ABMON_6, ABMON_7, ABMON_8, ABMON_9, ABMON_10,
	    ABMON_11, ABMON_12 };
#else
# define NL_LANGINFO(i) i
	const char * const item[12] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
#endif
	const char *tmp;
	size_t len;

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	if (sort_mb_cur_max == 1) {
#endif
		if (cmonths == NULL) {
			char *m;
			unsigned int j;
			int i;

			cmonths = xmalloc(sizeof(char *) * 12);
			for (i = 0; i < 12; i++) {
				cmonths[i] = NULL;
				tmp = NL_LANGINFO(item[i]);
				if (debug_sort)
					printf("month[%d]=%s\n", i, tmp);
				if (*tmp == '\0')
					continue;
				m = xstrdup(tmp);
				len = strlen(tmp);
				for (j = 0; j < len; j++)
					m[j] = toupper(m[j]);
				cmonths[i] = m;
			}
		}
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	} else {
		if (wmonths == NULL) {
			unsigned int j;
			wchar_t *m;
			int i;

			wmonths = xmalloc(sizeof(wchar_t *) * 12);
			for (i = 0; i < 12; i++) {
				wmonths[i] = NULL;
				tmp = nl_langinfo(item[i]);
				if (debug_sort)
					printf("month[%d]=%s\n", i, tmp);
				if (*tmp == '\0')
					continue;
				len = strlen(tmp);
				m = sort_reallocarray(NULL, len + 1, sizeof(wchar_t));
				if (mbstowcs(m, tmp, len) == (size_t)-1) {
					free(m);
					continue;
				}
				m[len] = L'\0';
				for (j = 0; j < len; j++)
					m[j] = towupper(m[j]);
				wmonths[i] = m;
			}
		}
	}
#endif
}

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
/* Compare two wide-character strings */
static int wide_str_coll(const wchar_t *s1, const wchar_t *s2)
{
	int ret = 0;

	errno = 0;
	ret = wcscoll(s1, s2);
	if (errno == EILSEQ) {
		errno = 0;
		ret = wcscmp(s1, s2);
		if (errno != 0) {
			size_t i;
			for (i = 0; ; ++i) {
				wchar_t c1 = s1[i];
				wchar_t c2 = s2[i];
				if (c1 == L'\0')
					return (c2 == L'\0') ? 0 : -1;
				if (c2 == L'\0')
					return 1;
				if (c1 == c2)
					continue;
				return (int)c1 - (int)c2;
			}
		}
	}
	return ret;
}
#endif

/* counterparts of wcs functions */

static void bwsprintf(FILE *f, struct bwstring *bws, const char *prefix, const char *suffix)
{
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	if (sort_mb_cur_max == 1)
#endif
		fprintf(f, "%s%s%s", prefix, bws->data.cstr, suffix);
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	else
		fprintf(f, "%s%S%s", prefix, bws->data.wstr, suffix);
#endif
}

static inline const void *bwsrawdata(const struct bwstring *bws)
{
	return &(bws->data);
}

static inline size_t bwsrawlen(const struct bwstring *bws)
{
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	return (sort_mb_cur_max == 1) ? bws->len : SIZEOF_WCHAR_STRING(bws->len);
#else
	return bws->len;
#endif
}

static size_t bws_memsize(const struct bwstring *bws)
{
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	return (sort_mb_cur_max == 1) ? (bws->len + 2 + sizeof(struct bwstring)) :
	    (SIZEOF_WCHAR_STRING(bws->len + 1) + sizeof(struct bwstring));
#else
	return bws->len + 2 + sizeof(struct bwstring);
#endif
}

static void bws_setlen(struct bwstring *bws, size_t newlen)
{
	if (bws && newlen != bws->len && newlen <= bws->len) {
		bws->len = newlen;
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
		if (sort_mb_cur_max == 1)
#endif
			bws->data.cstr[newlen] = '\0';
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
		else
			bws->data.wstr[newlen] = L'\0';
#endif
	}
}

/* Allocate a new binary string of specified size */
static struct bwstring *bwsalloc(size_t sz)
{
	struct bwstring *ret;

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	if (sort_mb_cur_max == 1) {
#endif
		ret = zmalloc(sizeof(struct bwstring) + 1 + sz);
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	} else {
		ret = xmalloc(sizeof(struct bwstring) + SIZEOF_WCHAR_STRING(sz + 1));
		ret->data.wstr[sz] = L'\0';
	}
#endif
	ret->len = sz;

	return ret;
}

/* Create a copy of binary string.
 * New string size equals the length of the old string. */
static struct bwstring *bwsdup(const struct bwstring *s)
{
	struct bwstring *ret;

	if (s == NULL)
		return NULL;

	ret = bwsalloc(s->len);

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	if (sort_mb_cur_max == 1)
#endif
		memcpy(ret->data.cstr, s->data.cstr, s->len);
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	else
		memcpy(ret->data.wstr, s->data.wstr, SIZEOF_WCHAR_STRING(s->len));
#endif
	return ret;
}

/* Create a new binary string from a wide character buffer. */
static struct bwstring *bwssbdup(const WCHAR_T_UTF8 *str, size_t len)
{
	if (str == NULL)
		return (len == 0) ? bwsalloc(0) : NULL;
	else {
		struct bwstring *ret;
		size_t i;

		ret = bwsalloc(len);

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
		if (sort_mb_cur_max == 1)
#endif
			for (i = 0; i < len; ++i)
				ret->data.cstr[i] = (unsigned char) str[i];
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
		else
			memcpy(ret->data.wstr, str, SIZEOF_WCHAR_STRING(len));
#endif
		return ret;
	}
}

/* Create a new binary string from a raw binary buffer. */
static struct bwstring *bwscsbdup(const unsigned char *str, size_t len)
{
	struct bwstring *ret;

	ret = bwsalloc(len);

	if (str) {
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
		if (sort_mb_cur_max == 1)
#endif
			memcpy(ret->data.cstr, str, len);
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
		else {
			mbstate_t mbs;
			const char *s;
			size_t charlen, chars, cptr;

			chars = 0;
			cptr = 0;
			s = (const char *) str;

			memset(&mbs, 0, sizeof(mbs));

			while (cptr < len) {
				size_t n = sort_mb_cur_max;

				if (n > len - cptr)
					n = len - cptr;
				charlen = mbrlen(s + cptr, n, &mbs);
				switch (charlen) {
				case 0:
					/* FALLTHROUGH */
				case (size_t) -1:
					/* FALLTHROUGH */
				case (size_t) -2:
					ret->data.wstr[chars++] =
					    (unsigned char) s[cptr];
					++cptr;
					break;
				default:
					n = mbrtowc(ret->data.wstr + (chars++),
					    s + cptr, charlen, &mbs);
					if ((n == (size_t)-1) || (n == (size_t)-2))
						/* NOTREACHED */
						bb_perror_msg_and_die("mbrtowc error");
					cptr += charlen;
				};
			}

			ret->len = chars;
			ret->data.wstr[ret->len] = L'\0';
		}
#endif
	}
	return ret;
}

/* Copy content of src binary string to dst,
 * with specified number of symbols to be copied.
 * An offset value can be specified, from the start of src string.
 * If the capacity of the dst string is not sufficient,
 * then the data is truncated. */
static struct bwstring *bwsnocpy(struct bwstring *dst, const struct bwstring *src, size_t offset, size_t size)
{
	if (offset >= src->len) {
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
		dst->data.wstr[0] = 0;
#else
		dst->data.cstr[0] = '\0';
#endif
		dst->len = 0;
	} else {
		size_t nums = src->len - offset;

		if (nums > dst->len)
			nums = dst->len;
		if (nums > size)
			nums = size;
		dst->len = nums;
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
		if (sort_mb_cur_max == 1) {
#endif
			memcpy(dst->data.cstr, src->data.cstr + offset, (nums));
			dst->data.cstr[dst->len] = '\0';
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
		} else {
			memcpy(dst->data.wstr, src->data.wstr + offset, SIZEOF_WCHAR_STRING(nums));
			dst->data.wstr[dst->len] = L'\0';
		}
#endif
	}
	return dst;
}

/* Write binary string to the file.
 * The output is ended either with '\n' (nl == true)
 * or '\0' (nl == false). */
static size_t bwsfwrite(struct bwstring *bws, FILE *f, bool zero_ended)
{
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	if (sort_mb_cur_max == 1)
#endif
	{

		size_t len = bws->len;

		if (!zero_ended) {
			bws->data.cstr[len] = '\n';

			if (fwrite(bws->data.cstr, len + 1, 1, f) < 1)
				bb_perror_msg_and_die("%s", bb_msg_write_error);

			bws->data.cstr[len] = '\0';
		} else if (fwrite(bws->data.cstr, len + 1, 1, f) < 1)
			bb_perror_msg_and_die("%s", bb_msg_write_error);

		return len + 1;

	}
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	  else {
		wchar_t eols;
		size_t printed = 0;

		eols = zero_ended ? btowc('\0') : btowc('\n');

		while (printed < BWSLEN(bws)) {
			const wchar_t *s = bws->data.wstr + printed;

			if (*s == L'\0') {
				int nums;

				nums = fwprintf(f, L"%lc", *s);

				if (nums != 1)
					bb_perror_msg_and_die(bb_msg_write_error);
				++printed;
			} else {
				int nums;

				nums = fwprintf(f, L"%ls", s);

				if (nums < 1)
					bb_perror_msg_and_die(bb_msg_write_error);
				printed += nums;
			}
		}
		fwprintf(f, L"%lc", eols);
		return printed + 1;
	}
#endif
}

/* Allocate and read a binary string from file.
 * The strings are nl-ended or zero-ended, depending on the sort setting. */
static struct bwstring *bwsfgetln(FILE *f, size_t *len, bool zero_ended, struct reader_buffer *rb)
{
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
#define EOLS_TO_WINT(c) (wint_t)c
	wint_t eols = zero_ended ? btowc('\0') : btowc('\n');
#else
#define EOLS_TO_WINT(c) c
	int eols = zero_ended ? '\0' : '\n';
#endif

	*len = 0;

	if (feof(f))
		return NULL;

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	if (2 >= rb->fgetwln_z_buffer_size) {
		rb->fgetwln_z_buffer_size += 256;
		rb->fgetwln_z_buffer = sort_reallocarray(rb->fgetwln_z_buffer,
		    rb->fgetwln_z_buffer_size, sizeof(wchar_t));
	}
	rb->fgetwln_z_buffer[0] = 0;

	if (sort_mb_cur_max == 1)
#endif
	{
		while (!feof(f)) {
			int c;

			c = fgetc(f);

			if (c == EOF) {
				if (*len == 0)
					return NULL;
				goto line_read_done;
			}
			if (EOLS_TO_WINT(c) == eols)
				goto line_read_done;

			if (*len + 1 >= rb->fgetwln_z_buffer_size) {
				rb->fgetwln_z_buffer_size += 256;
				rb->fgetwln_z_buffer =
				    sort_reallocarray(rb->fgetwln_z_buffer,
				    rb->fgetwln_z_buffer_size, sizeof(WCHAR_T_UTF8));
			}

			rb->fgetwln_z_buffer[*len] = c;
			rb->fgetwln_z_buffer[++(*len)] = 0;
		}
	}
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	  else {
		while (!feof(f)) {
			wint_t c = 0;

			c = fgetwc(f);

			if (c == WEOF) {
				if (*len == 0)
					return NULL;
				goto line_read_done;
			}
			if (c == eols)
				goto line_read_done;

			if (*len + 1 >= rb->fgetwln_z_buffer_size) {
				rb->fgetwln_z_buffer_size += 256;
				rb->fgetwln_z_buffer =
				    sort_reallocarray(rb->fgetwln_z_buffer,
				    rb->fgetwln_z_buffer_size, sizeof(wchar_t));
			}

			rb->fgetwln_z_buffer[*len] = c;
			rb->fgetwln_z_buffer[++(*len)] = 0;
		}
	}
#endif
line_read_done:
	/* we do not count the last 0 */
	return bwssbdup(rb->fgetwln_z_buffer, *len);
}

static int bwsncmp(const struct bwstring *bws1, const struct bwstring *bws2, size_t offset, size_t len)
{
	size_t cmp_len, len1, len2;
	int res = 0;

	len1 = bws1->len;
	len2 = bws2->len;

	if (len1 <= offset) {
		return (len2 <= offset) ? 0 : -1;
	} else {
		if (len2 <= offset)
			return 1;
		else {
			len1 -= offset;
			len2 -= offset;

			cmp_len = len1;

			if (len2 < cmp_len)
				cmp_len = len2;

			if (len < cmp_len)
				cmp_len = len;

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
			if (sort_mb_cur_max == 1)
#endif
			{
				const unsigned char *s1, *s2;

				s1 = bws1->data.cstr + offset;
				s2 = bws2->data.cstr + offset;

				res = memcmp(s1, s2, cmp_len);

			}
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
			 else {
				const wchar_t *s1, *s2;

				s1 = bws1->data.wstr + offset;
				s2 = bws2->data.wstr + offset;

				res = memcmp(s1, s2, SIZEOF_WCHAR_STRING(cmp_len));
			}
#endif
		}
	}

	if (res == 0) {
		if (len1 < cmp_len && len1 < len2)
			res = -1;
		else if (len2 < cmp_len && len2 < len1)
			res = +1;
	}

	return res;
}

static int bwscmp(const struct bwstring *bws1, const struct bwstring *bws2, size_t offset)
{
	size_t len1, len2, cmp_len;
	int res;

	len1 = bws1->len;
	len2 = bws2->len;

	len1 -= offset;
	len2 -= offset;

	cmp_len = len1;

	if (len2 < cmp_len)
		cmp_len = len2;

	res = bwsncmp(bws1, bws2, offset, cmp_len);

	if (res == 0) {
		if (len1 < len2)
			res = -1;
		else if (len2 < len1)
			res = +1;
	}

	return res;
}

static int bws_iterator_cmp(bwstring_iterator iter1, bwstring_iterator iter2, size_t len)
{
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	wchar_t c1, c2;
#else
	int c1, c2;
#endif
	size_t i = 0;

	for (i = 0; i < len; ++i) {
		c1 = bws_get_iter_value(iter1);
		c2 = bws_get_iter_value(iter2);
		if (c1 != c2)
			return c1 - c2;
		iter1 = bws_iterator_inc(iter1, 1);
		iter2 = bws_iterator_inc(iter2, 1);
	}

	return 0;
}

static int bwscoll(const struct bwstring *bws1, const struct bwstring *bws2, size_t offset)
{
	size_t len1, len2;

	len1 = bws1->len;
	len2 = bws2->len;

	if (len1 <= offset)
		return (len2 <= offset) ? 0 : -1;
	else {
		if (len2 <= offset)
			return 1;
		else {
			len1 -= offset;
			len2 -= offset;

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
			if (sort_mb_cur_max == 1)
#endif
			{
				const unsigned char *s1, *s2;

				s1 = bws1->data.cstr + offset;
				s2 = bws2->data.cstr + offset;

				if (byte_sort) {
					int res = 0;

					if (len1 > len2) {
						res = memcmp(s1, s2, len2);
						if (!res)
							res = +1;
					} else if (len1 < len2) {
						res = memcmp(s1, s2, len1);
						if (!res)
							res = -1;
					} else
						res = memcmp(s1, s2, len1);

					return res;

				} else {
					int res = 0;
					size_t i, maxlen;

					i = 0;
					maxlen = len1;

					if (maxlen > len2)
						maxlen = len2;

					while (i < maxlen) {
						/* goto next non-zero part: */
						while ((i < maxlen) &&
						    !s1[i] && !s2[i])
							++i;

						if (i >= maxlen)
							break;

						if (s1[i] == 0) {
							if (s2[i] == 0)
								/* NOTREACHED */
								bb_perror_msg_and_die("bwscoll error 01");
							else
								return -1;
						} else if (s2[i] == 0)
							return 1;

						res = strcoll((const char *)(s1 + i), (const char *)(s2 + i));
						if (res)
							return res;

						while ((i < maxlen) &&
						    s1[i] && s2[i])
							++i;

						if (i >= maxlen)
							break;

						if (s1[i] == 0) {
							if (s2[i] == 0) {
								++i;
								continue;
							} else
								return -1;
						} else if (s2[i] == 0)
							return 1;
						else
							/* NOTREACHED */
							bb_perror_msg_and_die("bwscoll error 02");
					}

					if (len1 < len2)
						return -1;
					else if (len1 > len2)
						return 1;

					return 0;
				}
			}
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
			  else {
				const wchar_t *s1, *s2;
				size_t i, maxlen;
				int res = 0;

				s1 = bws1->data.wstr + offset;
				s2 = bws2->data.wstr + offset;

				i = 0;
				maxlen = len1;

				if (maxlen > len2)
					maxlen = len2;

				while (i < maxlen) {

					/* goto next non-zero part: */
					while ((i < maxlen) &&
					    !s1[i] && !s2[i])
						++i;

					if (i >= maxlen)
						break;

					if (s1[i] == 0) {
						if (s2[i] == 0)
							/* NOTREACHED */
							bb_perror_msg_and_die("bwscoll error 1");
						else
							return -1;
					} else if (s2[i] == 0)
						return 1;

					res = wide_str_coll(s1 + i, s2 + i);
					if (res)
						return res;

					while ((i < maxlen) && s1[i] && s2[i])
						++i;

					if (i >= maxlen)
						break;

					if (s1[i] == 0) {
						if (s2[i] == 0) {
							++i;
							continue;
						} else
							return -1;
					} else if (s2[i] == 0)
						return 1;
					else
						/* NOTREACHED */
						bb_perror_msg_and_die("bwscoll error 2");
				}

				if (len1 == len2)
					return 0;
				return len1 < len2 ? -1 : 1;
			}
#endif
		}
	}
}

/* Correction of the system API */
static double bwstod(struct bwstring *s0, bool *empty)
{
	double ret = 0;

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	if (sort_mb_cur_max == 1)
#endif
	{
		char *ep, *end, *s;

		s = (char *)s0->data.cstr;
		end = s + s0->len;
		ep = NULL;

		while (isblank((unsigned char)*s) && s < end)
			++s;

		if (!isprint((unsigned char)*s)) {
			*empty = true;
			return 0;
		}

		ret = strtod(s, &ep);
		if (ep == s) {
			*empty = true;
			return 0;
		}
	}
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	  else {
		wchar_t *end, *ep, *s;

		s = s0->data.wstr;
		end = s + s0->len;
		ep = NULL;

		while (iswblank(*s) && s < end)
			++s;

		if (!iswprint(*s)) {
			*empty = true;
			return 0;
		}

		ret = wcstod(s, &ep);
		if (ep == s) {
			*empty = true;
			return 0;
		}
	}
#endif
	*empty = false;
	return ret;
}

/* A helper function for monthcoll.  If a line matches
 * a month name, it returns (number of the month - 1),
 * while if there is no match, it just return -1. */
static int bws_month_score(const struct bwstring *s0)
{
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	if (sort_mb_cur_max == 1) {
#endif
		const char *end, *s;
		int i;

		s = (char *)s0->data.cstr;
		end = s + s0->len;

		while (isblank((unsigned char)*s) && s < end)
			++s;

		for (i = 11; i >= 0; --i) {
			if (cmonths[i] &&
			    (s == strstr(s, cmonths[i])))
				return i;
		}
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	} else {
		const wchar_t *end, *s;
		int i;

		s = s0->data.wstr;
		end = s + s0->len;

		while (iswblank(*s) && s < end)
			++s;

		for (i = 11; i >= 0; --i) {
			if (wmonths[i] && (s == wcsstr(s, wmonths[i])))
				return i;
		}
	}
#endif
	return -1;
}

/* Rips out leading blanks (-b). */
static struct bwstring *ignore_leading_blanks(struct bwstring *str)
{
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	if (sort_mb_cur_max == 1) {
#endif
		unsigned char *dst, *end, *src;

		src = str->data.cstr;
		dst = src;
		end = src + str->len;

		while (src < end && isblank(*src))
			++src;

		if (src != dst) {
			size_t newlen;

			newlen = BWSLEN(str) - (src - dst);

			while (src < end) {
				*dst = *src;
				++dst;
				++src;
			}
			bws_setlen(str, newlen);
		}
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	} else {
		wchar_t *dst, *end, *src;

		src = str->data.wstr;
		dst = src;
		end = src + str->len;

		while (src < end && iswblank(*src))
			++src;

		if (src != dst) {

			size_t newlen = BWSLEN(str) - (src - dst);

			while (src < end) {
				*dst = *src;
				++dst;
				++src;
			}
			bws_setlen(str, newlen);

		}
	}
#endif
	return str;
}

/* Rips out nonprinting characters (-i). */
static struct bwstring *ignore_nonprinting(struct bwstring *str)
{
	size_t newlen = str->len;

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	if (sort_mb_cur_max == 1) {
#endif
		unsigned char *dst, *end, *src;
		unsigned char c;

		src = str->data.cstr;
		dst = src;
		end = src + str->len;

		while (src < end) {
			c = *src;
			if (isprint(c)) {
				*dst = c;
				++dst;
				++src;
			} else {
				++src;
				--newlen;
			}
		}
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	} else {
		wchar_t *dst, *end, *src;
		wchar_t c;

		src = str->data.wstr;
		dst = src;
		end = src + str->len;

		while (src < end) {
			c = *src;
			if (iswprint(c)) {
				*dst = c;
				++dst;
				++src;
			} else {
				++src;
				--newlen;
			}
		}
	}
#endif
	bws_setlen(str, newlen);

	return str;
}

/* Rips out any characters that are not alphanumeric characters
 * nor blanks (-d). */
static struct bwstring *dictionary_order(struct bwstring *str)
{
	size_t newlen = str->len;

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	if (sort_mb_cur_max == 1) {
#endif
		unsigned char *dst, *end, *src;
		unsigned char c;

		src = str->data.cstr;
		dst = src;
		end = src + str->len;

		while (src < end) {
			c = *src;
			if (isalnum(c) || isblank(c)) {
				*dst = c;
				++dst;
				++src;
			} else {
				++src;
				--newlen;
			}
		}
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	} else {
		wchar_t *dst, *end, *src;
		wchar_t c;

		src = str->data.wstr;
		dst = src;
		end = src + str->len;

		while (src < end) {
			c = *src;
			if (iswalnum(c) || iswblank(c)) {
				*dst = c;
				++dst;
				++src;
			} else {
				++src;
				--newlen;
			}
		}
	}
#endif
	bws_setlen(str, newlen);

	return str;
}

/* Converts string to lower case(-f). */
static struct bwstring *ignore_case(struct bwstring *str)
{
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	if (sort_mb_cur_max == 1) {
#endif
		unsigned char *end, *s;

		s = str->data.cstr;
		end = s + str->len;

		while (s < end) {
			*s = toupper(*s);
			++s;
		}
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	} else {
		wchar_t *end, *s;

		s = str->data.wstr;
		end = s + str->len;

		while (s < end) {
			*s = towupper(*s);
			++s;
		}
	}
#endif
	return str;
}

static void bws_disorder_warnx(struct bwstring *s, const char *fn, size_t pos)
{
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	if (sort_mb_cur_max == 1)
#endif
		bb_error_msg("%s:%zu: disorder: %s", fn, pos + 1, s->data.cstr);
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	else
		bb_error_msg("%s:%zu: disorder: %ls", fn, pos + 1, s->data.wstr);
#endif
}

/* How much space we read from file at once */
#define READ_CHUNK 4096


/* Structure to be used in file merge process. */
struct file_header {
	struct file_reader              *fr;
	struct sort_list_item           *si; /* current top line */
	size_t                           file_pos;
};

/* Save name of a tmp file for signal cleanup */
static void tmp_file_atexit(const char *tmp_file)
{
	sigset_t mask, oldmask;

	sigfillset(&mask);
	sigprocmask(SIG_BLOCK, &mask, &oldmask);
	tmp_files = llist_add_to(tmp_files, xstrdup(tmp_file));
	sigprocmask(SIG_SETMASK, &oldmask, NULL);
}

/* Clear tmp files */
static void clear_tmp_files(void)
{
	llist_t *item;

	for(item = tmp_files; item != NULL; item = item->link)
		unlink(item->data);
}

/* Check whether a file is a temporary file */
static bool file_is_tmp(const char *fn)
{
	llist_t *item;

	for(item = tmp_files; item != NULL; item = item->link) {
		if (strcmp(item->data, fn) == 0)
			return true;
	}

	return false;
}

/* Generate new temporary file name */
static char *new_tmp_file_name(void)
{
	char *ret;
	int fd;

	ret = xasprintf("%s/.bsdsort.XXXXXXXXXX", bb_common_bss.tmpdir);
	if ((fd = mkstemp(ret)) == -1)
		bb_perror_msg_and_die("%s", ret);
	close(fd);
	tmp_file_atexit(ret);
	return ret;
}

/* Initialize file list */
static void file_list_init(struct file_list *fl, bool tmp)
{
	fl->count = 0;
	fl->sz = 0;
	fl->fns = NULL;
	fl->tmp = tmp;
}

/* Add a file name to the list */
static void file_list_add(struct file_list *fl, char *fn, bool allocate)
{
	if (fl->count >= fl->sz) {
		fl->fns = sort_reallocarray(fl->fns,
		    fl->sz ? fl->sz : (fl->sz = 1), 2 * sizeof(char *));
		fl->sz *= 2;
	}
	fl->fns[fl->count] = allocate ? xstrdup(fn) : fn;
	fl->count += 1;
}

/* Populate file list from array of file names */
static void file_list_populate(struct file_list *fl, int argc, char **argv, bool allocate)
{
	int i;

	for (i = 0; i < argc; i++)
		file_list_add(fl, argv[i], allocate);
}

/* Clean file list data and delete the files,
 * if this is a list of temporary files */
static void file_list_clean(struct file_list *fl)
{
	if (fl->fns) {
		size_t i;

		for (i = 0; i < fl->count; i++) {
			if (fl->fns[i]) {
				if (fl->tmp)
					unlink(fl->fns[i]);
				free(fl->fns[i]);
				fl->fns[i] = NULL;
			}
		}
		free(fl->fns);
		fl->fns = NULL;
	}
	fl->sz = 0;
	fl->count = 0;
	fl->tmp = false;
}

/* Init sort list */
static void sort_list_init(struct sort_list *l)
{
	l->count = 0;
	l->size = 0;
	l->memsize = sizeof(struct sort_list);
	l->list = NULL;
}

/* Calculate keys array size */
static size_t keys_array_size(void)
{
	return keys_num * (sizeof(struct key_value) +
		(bb_common_bss.need_hint ? sizeof(struct key_hint) : 0));
}

/* Initialize a sort list item */
static struct sort_list_item *sort_list_item_alloc(void)
{
	struct sort_list_item *si;
	size_t sz;

	sz = sizeof(struct sort_list_item) + keys_array_size();
	si = zmalloc(sz);

	return si;
}

static size_t sort_list_item_size(struct sort_list_item *si)
{
	size_t i, ret = 0;

	if (si) {
		ret = sizeof(struct sort_list_item) + keys_array_size();
		if (si->str)
			ret += bws_memsize(si->str);
		for (i = 0; i < keys_num; ++i) {
			struct key_value *kv;

			kv = &(si->ka.key[i]);

			if (kv->k != si->str)
				ret += bws_memsize(kv->k);
		}
	}
	return ret;
}

/* Clean data of keys array */
static void clean_keys_array(const struct bwstring *s, struct keys_array *ka)
{
	if (ka) {
		size_t i;

		for (i = 0; i < keys_num; ++i)
			if (ka->key[i].k && ka->key[i].k != s)
				free(ka->key[i].k);
		memset(ka, 0, keys_array_size());
	}
}

/* Opens a file.  If the given filename is "-", stdout will be opened. */
static FILE *openfile(const char *fn, const char *mode)
{
	FILE *file;

	if (strcmp(fn, "-") == 0)
		return (mode[0] == 'r') ? stdin : stdout;

	if (file_is_tmp(fn) && (bb_common_bss.compress_program != NULL)) {
		char *cmd;

		fflush(stdout);

		if (mode[0] == 'r')
			cmd = xasprintf("%s -d < %s", bb_common_bss.compress_program, fn);
		else if (mode[0] == 'w')
			cmd = xasprintf("%s > %s", bb_common_bss.compress_program, fn);
		else
			bb_error_msg_and_die("invalid file mode");

		if ((file = popen(cmd, mode)) == NULL)
			bb_perror_msg_and_die("%s", bb_common_bss.compress_program);

		free(cmd);
	} else {
		file = bb_xfopen(fn, (mode[0] == 'r' ? "Ar" : "Aw"));
	}

	return file;
}

/* Close file */
static void closefile(FILE *f, const char *fn)
{
	if (f == NULL) {
		;
	} else if (f == stdin) {
		;
	} else if (f == stdout) {
		fflush(f);
	} else {
		if (file_is_tmp(fn) && bb_common_bss.compress_program != NULL) {
			if (pclose(f) < 0)
				bb_perror_nomsg_and_die();
		} else
			fclose(f);
	}
}

/* Compares the given strings.  Returns a positive number if
 * the first precedes the second, a negative number if the second is
 * the preceding one, and zero if they are equal.  This function calls
 * the underlying collate functions, which done the actual comparison. */
static int key_coll(struct keys_array *ps1, struct keys_array *ps2, size_t offset)
{
	struct sort_mods *sm;
	int res = 0;
	size_t i;

	for (i = 0; i < keys_num; ++i) {
		sm = &(keys[i].sm);

		if (sm->rflag)
			res = sm->func(&(ps2->key[i]), &(ps1->key[i]), offset);
		else
			res = sm->func(&(ps1->key[i]), &(ps2->key[i]), offset);

		if (res)
			break;

		/* offset applies to only the first key */
		offset = 0;
	}
	return res;
}

/* Compare two strings. Plain symbol-by-symbol comparison. */
static int top_level_str_coll(const struct bwstring *s1, const struct bwstring *s2)
{
	if (default_sort_mods->rflag) {
		const struct bwstring *tmp;

		tmp = s1;
		s1 = s2;
		s2 = tmp;
	}

	return bwscoll(s1, s2, 0);
}

/* Compare two sort list items, according to the sort specs. */
static int list_coll_offset(struct sort_list_item **ss1, struct sort_list_item **ss2, size_t offset)
{
	int ret;

	ret = key_coll(&((*ss1)->ka), &((*ss2)->ka), offset);

	if (debug_sort) {
		if (offset)
			printf("; offset=%zu", offset);
		bwsprintf(stdout, ((*ss1)->str), "; s1=<", ">");
		bwsprintf(stdout, ((*ss2)->str), ", s2=<", ">");
		printf("; cmp1=%d\n", ret);
	}

	if (ret)
		return ret;

	if (!(bb_common_bss.sflag) && bb_common_bss.complex_sort) {
		ret = top_level_str_coll(((*ss1)->str), ((*ss2)->str));
		if (debug_sort)
			printf("; cmp2=%d\n", ret);
	}

	return ret;
}

/* Compare two sort list items, according to the sort specs. */
static int list_coll(const void *ss1, const void *ss2)
{
	return list_coll_offset((struct sort_list_item **)ss1,
	    (struct sort_list_item **)ss2, 0);
}

/* Write sort list to file */
static void sort_list_dump(struct sort_list *l, const char *fn)
{
	FILE *f;

	f = openfile(fn, "w");

	if (l->list) {
		size_t i;

		if (!bb_common_bss.uflag) {
			for (i = 0; i < l->count; ++i)
				bwsfwrite(l->list[i]->str, f, bb_common_bss.zflag);
		} else {
			struct sort_list_item *last_printed_item = NULL;
			struct sort_list_item *item;
			for (i = 0; i < l->count; ++i) {
				item = l->list[i];
				if ((last_printed_item == NULL) ||
				    list_coll(&last_printed_item, &item)) {
					bwsfwrite(item->str, f, bb_common_bss.zflag);
					last_printed_item = item;
				}
			}
		}
	}

	closefile(f, fn);
}

/* Reads a file into the internal buffer. */
static struct file_reader *file_reader_init(const char *fsrc)
{
	struct file_reader *ret;

	if (fsrc == NULL)
		fsrc = "-";

	ret = zmalloc(sizeof(struct file_reader));

	if (!bb_common_bss.zflag)
		ret->elsymb = '\n';

	ret->fname = xstrdup(fsrc);

	if (strcmp(fsrc, "-") && (bb_common_bss.compress_program == NULL) && bb_common_bss.use_mmap) {
		struct stat stat_buf;
		void *addr;
		size_t sz = 0;
		int fd;

		fd = bb_xopen(fsrc, O_RDONLY | BB_OPEN_ERR);

		if (fstat(fd, &stat_buf) < 0)
			bb_perror_msg_and_die("%s", fsrc);
		sz = stat_buf.st_size;

		addr = mmap(NULL, sz, PROT_READ, 0, fd, 0);
		close(fd);
		if (addr != MAP_FAILED) {
			ret->mmapaddr = addr;
			ret->mmapsize = sz;
			ret->mmapptr = ret->mmapaddr;
#if 0
			posix_madvise(addr, sz, POSIX_MADV_SEQUENTIAL);
#endif
		}
	}

	if (ret->mmapaddr == NULL) {
		ret->file = openfile(fsrc, "r");

		if (strcmp(fsrc, "-")) {
			ret->cbsz = READ_CHUNK;
			ret->buffer = xmalloc(ret->cbsz);
			ret->bsz = 0;
			ret->strbeg = 0;

			ret->bsz = fread(ret->buffer, 1, ret->cbsz, ret->file);
			if (ret->bsz == 0) {
				if (ferror(ret->file))
					bb_perror_nomsg_and_die();
			}
		}
	}

	return ret;
}

static struct bwstring *file_reader_readline(struct file_reader *fr)
{
	struct bwstring *ret = NULL;

	if (fr->mmapaddr) {
		unsigned char *mmapend;

		mmapend = fr->mmapaddr + fr->mmapsize;
		if (fr->mmapptr >= mmapend)
			return NULL;
		else {
			unsigned char *strend;
			size_t sz;

			sz = mmapend - fr->mmapptr;
			strend = memchr(fr->mmapptr, fr->elsymb, sz);

			if (strend == NULL) {
				ret = bwscsbdup(fr->mmapptr, sz);
				fr->mmapptr = mmapend;
			} else {
				ret = bwscsbdup(fr->mmapptr, strend -
				    fr->mmapptr);
				fr->mmapptr = strend + 1;
			}
		}

	} else if (fr->file != stdin) {
		unsigned char *strend;
		size_t bsz1, remsz, search_start;

		search_start = 0;
		remsz = 0;
		strend = NULL;

		if (fr->bsz > fr->strbeg)
			remsz = fr->bsz - fr->strbeg;

		/* line read cycle */
		for (;;) {
			if (remsz > search_start)
				strend = memchr(fr->buffer + fr->strbeg +
				    search_start, fr->elsymb, remsz - search_start);
			else
				strend = NULL;

			if (strend)
				break;
			if (feof(fr->file))
				break;

			if (fr->bsz != fr->cbsz)
				/* NOTREACHED */
				bb_error_msg_and_die("File read software error 1");

			if (remsz > (READ_CHUNK >> 1)) {
				search_start = fr->cbsz - fr->strbeg;
				fr->cbsz += READ_CHUNK;
				fr->buffer = sort_reallocarray(fr->buffer,
				    1, fr->cbsz);
				bsz1 = fread(fr->buffer + fr->bsz, 1, READ_CHUNK, fr->file);
				if (bsz1 == 0) {
					if (ferror(fr->file))
						bb_perror_nomsg_and_die();
					break;
				}
				fr->bsz += bsz1;
				remsz += bsz1;
			} else {
				if (remsz > 0 && fr->strbeg > 0) {
					memmove(fr->buffer,
					    fr->buffer + fr->strbeg, remsz);
				}
				fr->strbeg = 0;
				search_start = remsz;
				bsz1 = fread(fr->buffer + remsz, 1,
				    fr->cbsz - remsz, fr->file);
				if (bsz1 == 0) {
					if (ferror(fr->file))
						bb_perror_nomsg_and_die();
					break;
				}
				fr->bsz = remsz + bsz1;
				remsz = fr->bsz;
			}
		}

		if (strend == NULL)
			strend = fr->buffer + fr->bsz;

		if ((fr->buffer + fr->strbeg <= strend) &&
		    (fr->strbeg < fr->bsz) && (remsz>0))
			ret = bwscsbdup(fr->buffer + fr->strbeg, strend -
			    fr->buffer - fr->strbeg);

		fr->strbeg = (strend - fr->buffer) + 1;
	} else {
		size_t len = 0;

		ret = bwsfgetln(fr->file, &len, bb_common_bss.zflag,
		    &(fr->rb));
	}

	return ret;
}

static void file_reader_clean(struct file_reader *fr)
{
	if (fr->mmapaddr)
		munmap(fr->mmapaddr, fr->mmapsize);

	free(fr->buffer);

	if (fr->file)
		closefile(fr->file, fr->fname);

	free(fr->fname);

	memset(fr, 0, sizeof(struct file_reader));
}

static void file_reader_free(struct file_reader *fr)
{
	file_reader_clean(fr);
	free(fr);
}

/* Compare file headers. Files with EOF always go to the end of the list. */
static int file_header_cmp(struct file_header *f1, struct file_header *f2)
{
	int ret;

	if (f1 == f2)
		return 0;
	if (f1->fr == NULL)
		return (f2->fr == NULL) ? 0 : 1;
	if (f2->fr == NULL)
		return -1;

	ret = list_coll(&(f1->si), &(f2->si));
	if (!ret)
		return (f1->file_pos < f2->file_pos) ? -1 : 1;
	return ret;
}

/* De-allocate a sort list item object memory */
static void sort_list_item_clean(struct sort_list_item *si)
{
	if (si) {
		clean_keys_array(si->str, &(si->ka));
		if (si->str) {
			free(si->str);
			si->str = NULL;
		}
	}
}

/* Close file */
static void file_header_close(struct file_header **fh)
{
	if ((*fh)->fr) {
		file_reader_free((*fh)->fr);
		(*fh)->fr = NULL;
	}
	if ((*fh)->si) {
		sort_list_item_clean((*fh)->si);
		free((*fh)->si);
		(*fh)->si = NULL;
	}
	free(*fh);
	*fh = NULL;
}

/* Swap two array elements */
static void file_header_swap(struct file_header **fh, size_t i1, size_t i2)
{
	struct file_header *tmp;

	tmp = fh[i1];
	fh[i1] = fh[i2];
	fh[i2] = tmp;
}

/* heap algorithm ==>> */

/* Swap two areas of size number of bytes.  Although qsort(3) permits random
 * blocks of memory to be sorted, sorting pointers is almost certainly the
 * common case (and, were it not, could easily be made so).  Regardless, it
 * isn't worth optimizing; the SWAP's get sped up by the cache, and pointer
 * arithmetic gets lost in the time required for comparison function calls. */
#define SWAP_for_heap(a, b, count, size, tmp) { \
	count = size; \
	do { \
		tmp = *a; \
		*a++ = *b; \
		*b++ = tmp; \
	} while (--count); \
}

/* Copy one block of size size to another. */
#define COPY_for_heap(a, b, count, size, tmp1, tmp2) { \
	count = size; \
	tmp1 = a; \
	tmp2 = b; \
	do { \
		*tmp1++ = *tmp2++; \
	} while (--count); \
}

/* Build the list into a heap, where a heap is defined such that for
 * the records K1 ... KN, Kj/2 >= Kj for 1 <= j/2 <= j <= N.
 *
 * There are two cases.  If j == nmemb, select largest of Ki and Kj.  If
 * j < nmemb, select largest of Ki, Kj and Kj+1. */
#define CREATE_for_heap(initval, nmemb, par_i, child_i, par, child, size, count, tmp) { \
	for (par_i = initval; (child_i = par_i * 2) <= nmemb; \
	    par_i = child_i) { \
		child = base + child_i * size; \
		if (child_i < nmemb && compar(child, child + size) < 0) { \
			child += size; \
			++child_i; \
		} \
		par = base + par_i * size; \
		if (compar(child, par) <= 0) \
			break; \
		SWAP_for_heap(par, child, count, size, tmp); \
	} \
}

/* Select the top of the heap and 'heapify'.  Since by far the most expensive
 * action is the call to the compar function, a considerable optimization
 * in the average case can be achieved due to the fact that k, the displaced
 * element, is usually quite small, so it would be preferable to first
 * heapify, always maintaining the invariant that the larger child is copied
 * over its parent's record.
 *
 * Then, starting from the *bottom* of the heap, finding k's correct place,
 * again maintaining the invariant.  As a result of the invariant no element
 * is 'lost' when k is assigned its correct place in the heap.
 *
 * The time savings from this optimization are on the order of 15-20% for the
 * average case. See Knuth, Vol. 3, page 158, problem 18.
 *
 * XXX Don't break the #define SELECT line, below.  Reiser cpp gets upset. */
#define SELECT_for_heap(par_i, child_i, nmemb, par, child, size, k, count, tmp1, tmp2) { \
	for (par_i = 1; (child_i = par_i * 2) <= nmemb; par_i = child_i) { \
		child = base + child_i * size; \
		if (child_i < nmemb && compar(child, child + size) < 0) { \
			child += size; \
			++child_i; \
		} \
		par = base + par_i * size; \
		COPY_for_heap(par, child, count, size, tmp1, tmp2); \
	} \
	for (;;) { \
		child_i = par_i; \
		par_i = child_i / 2; \
		child = base + child_i * size; \
		par = base + par_i * size; \
		if (child_i == 1 || compar(k, par) < 0) { \
			COPY_for_heap(child, k, count, size, tmp1, tmp2); \
			break; \
		} \
		COPY_for_heap(child, par, count, size, tmp1, tmp2); \
	} \
}

/* Heapsort -- Knuth, Vol. 3, page 145.  Runs in O (N lg N), both average
 * and worst.  While heapsort is faster than the worst case of quicksort,
 * the BSD quicksort does median selection so that the chance of finding
 * a data set that will trigger the worst case is nonexistent.  Heapsort's
 * only advantage over quicksort is that it requires little additional memory. */
static int heapsort(void *vbase, size_t nmemb, size_t size,
    int (*compar)(const void *, const void *))
{
	size_t cnt, i, j, l;
	char tmp, *tmp1, *tmp2;
	char *base, *k, *p, *t;

	if (nmemb <= 1)
		return 0;

	if (!size) {
		errno = EINVAL;
		return -1;
	}

	if ((k = malloc(size)) == NULL)
		return -1;

	/*
	 * Items are numbered from 1 to nmemb, so offset from size bytes
	 * below the starting address.
	 */
	base = (char *)vbase - size;

	for (l = nmemb / 2 + 1; --l;)
		CREATE_for_heap(l, nmemb, i, j, t, p, size, cnt, tmp);

	/*
	 * For each element of the heap, save the largest element into its
	 * final slot, save the displaced element (k), then recreate the
	 * heap.
	 */
	while (nmemb > 1) {
		COPY_for_heap(k, base + nmemb * size, cnt, size, tmp1, tmp2);
		COPY_for_heap(base + nmemb * size, base + size, cnt, size, tmp1, tmp2);
		--nmemb;
		SELECT_for_heap(i, j, nmemb, t, p, size, k, cnt, tmp1, tmp2);
	}
	free(k);
	return 0;
}

/* See heap sort algorithm
 * "Raises" last element to its right place */
static void file_header_heap_swim(struct file_header **fh, size_t indx)
{
	if (indx > 0) {
		size_t parent_index;

		parent_index = (indx - 1) >> 1;

		if (file_header_cmp(fh[indx], fh[parent_index]) < 0) {
			/* swap child and parent and continue */
			file_header_swap(fh, indx, parent_index);
			file_header_heap_swim(fh, parent_index);
		}
	}
}

/* Sink the top element to its correct position */
static void file_header_heap_sink(struct file_header **fh, size_t indx, size_t size)
{
	size_t left_child_index;
	size_t right_child_index;

	left_child_index = indx + indx + 1;
	right_child_index = left_child_index + 1;

	if (left_child_index < size) {
		size_t min_child_index;

		min_child_index = left_child_index;

		if ((right_child_index < size) &&
		    (file_header_cmp(fh[left_child_index],
		    fh[right_child_index]) > 0))
			min_child_index = right_child_index;
		if (file_header_cmp(fh[indx], fh[min_child_index]) > 0) {
			file_header_swap(fh, indx, min_child_index);
			file_header_heap_sink(fh, min_child_index, size);
		}
	}
}

/* <<== heap algorithm */

/* Adds element to the "left" end */
static void file_header_list_rearrange_from_header(struct file_header **fh, size_t size)
{
	file_header_heap_sink(fh, 0, size);
}

/* Adds element to the "right" end */
static void file_header_list_push(struct file_header *f, struct file_header **fh, size_t size)
{
	fh[size++] = f;
	file_header_heap_swim(fh, size - 1);
}


static const char *get_sort_method_name(int sm)
{
	if (sm == SORT_MERGESORT)
		return "mergesort";
	else if (bb_common_bss.sort_method == SORT_RADIXSORT)
		return "radixsort";
	else if (bb_common_bss.sort_method == SORT_HEAPSORT)
		return "heapsort";
	else
		return "quicksort";
}


/* #define NATURAL to get hybrid natural merge.
 * (The default is pairwise merging.)
 */
#define NATURAL
#define THRESHOLD 16    /* Best choice for natural merge cut-off. */

#define ISIZE sizeof(int)
#define PSIZE sizeof(u_char *)
#define ICOPY_LIST(src, dst, last)                              \
	do                                                      \
	*(int*)dst = *(int*)src, src += ISIZE, dst += ISIZE;    \
	while(src < last)
#define ICOPY_ELT(src, dst, i)                                  \
	do                                                      \
	*(int*) dst = *(int*) src, src += ISIZE, dst += ISIZE;  \
	while (i -= ISIZE)

#define CCOPY_LIST(src, dst, last)              \
	do                                      \
		*dst++ = *src++;                \
	while (src < last)
#define CCOPY_ELT(src, dst, i)                  \
	do                                      \
		*dst++ = *src++;                \
	while (i -= 1)


#define swap_for_merge(a, b) {                                    \
		s = b;                                  \
		i = size;                               \
		do {                                    \
			tmp = *a; *a++ = *s; *s++ = tmp; \
		} while (--i);                          \
		a -= size;                              \
	}
#define reverse_for_merge(bot, top) {                             \
	s = top;                                        \
	do {                                            \
		i = size;                               \
		do {                                    \
			tmp = *bot; *bot++ = *s; *s++ = tmp; \
		} while (--i);                          \
		s -= size2;                             \
	} while(bot < s);                               \
}

/*
 * This is to avoid out-of-bounds addresses in sorting the
 * last 4 elements.
 */
static void
insertionsort(u_char *a, size_t n, size_t size,
    int (*cmp)(const void *, const void *))
{
	u_char *ai, *s, *t, *u, tmp;
	int i;

	for (ai = a+size; --n >= 1; ai += size)
		for (t = ai; t > a; t -= size) {
			u = t - size;
			if (cmp(u, t) <= 0)
				break;
			swap_for_merge(u, t);
		}
}

/* Find the next possible pointer head.  (Trickery for forcing an array
 * to do double duty as a linked list when objects do not align with word
 * boundaries.
 */
/* Assumption: PSIZE is a power of 2. */
#define EVAL(p) (u_char **)                                             \
	((u_char *)0 +                                                  \
	    (((u_char *)p + PSIZE - 1 - (u_char *) 0) & ~(PSIZE - 1)))

/* Optional hybrid natural/pairwise first pass.  Eats up list1 in runs of
 * increasing order, list2 in a corresponding linked list.  Checks for runs
 * when THRESHOLD/2 pairs compare with same sense.  (Only used when NATURAL
 * is defined.  Otherwise simple pairwise merging is used.)
 */
static void
setup_merge(u_char *list1, u_char *list2, size_t n, size_t size,
    int (*cmp)(const void *, const void *))
{
	int i, length, size2, sense;
	u_char tmp, *f1, *f2, *s, *l2, *last, *p2;

	size2 = size*2;
	if (n <= 5) {
		insertionsort(list1, n, size, cmp);
		*EVAL(list2) = (u_char*) list2 + n*size;
		return;
	}
	/* Avoid running pointers out of bounds; limit n to evens
	 * for simplicity.
	 */
	i = 4 + (n & 1);
	insertionsort(list1 + (n - i) * size, i, size, cmp);
	last = list1 + size * (n - i);
	*EVAL(list2 + (last - list1)) = list2 + n * size;

#ifdef NATURAL
	p2 = list2;
	f1 = list1;
	sense = (cmp(f1, f1 + size) > 0);
	for (; f1 < last; sense = !sense) {
		length = 2;
					/* Find pairs with same sense. */
		for (f2 = f1 + size2; f2 < last; f2 += size2) {
			if ((cmp(f2, f2+ size) > 0) != sense)
				break;
			length += 2;
		}
		if (length < THRESHOLD) {               /* Pairwise merge */
			do {
				p2 = *EVAL(p2) = f1 + size2 - list1 + list2;
				if (sense > 0)
					swap_for_merge (f1, f1 + size);
			} while ((f1 += size2) < f2);
		} else {                                /* Natural merge */
			l2 = f2;
			for (f2 = f1 + size2; f2 < l2; f2 += size2) {
				if ((cmp(f2-size, f2) > 0) != sense) {
					p2 = *EVAL(p2) = f2 - list1 + list2;
					if (sense > 0)
						reverse_for_merge(f1, f2-size);
					f1 = f2;
				}
			}
			if (sense > 0)
				reverse_for_merge (f1, f2-size);
			f1 = f2;
			if (f2 < last || cmp(f2 - size, f2) > 0)
				p2 = *EVAL(p2) = f2 - list1 + list2;
			else
				p2 = *EVAL(p2) = list2 + n*size;
		}
	}
#else           /* pairwise merge only. */
	for (f1 = list1, p2 = list2; f1 < last; f1 += size2) {
		p2 = *EVAL(p2) = p2 + size2;
		if (cmp (f1, f1 + size) > 0)
			swap_for_merge(f1, f1 + size);
	}
#endif /* NATURAL */
}

/* Arguments are as for qsort. */
static int
mergesort(void *base, size_t nmemb, size_t size,
    int (*cmp)(const void *, const void *))
{
	size_t i;
	int sense;
	int big, iflag;
	u_char *f1, *f2, *t, *b, *tp2, *q, *l1, *l2;
	u_char *list2, *list1, *p2, *p, *last, **p1;

	if (size < PSIZE / 2) {         /* Pointers must fit into 2 * size. */
		errno = EINVAL;
		return -1;
	}

	if (nmemb == 0)
		return 0;

	/* XXX Stupid subtraction for the Cray. */
	iflag = 0;
	if (!(size % ISIZE) && !(((char *)base - (char *)0) % ISIZE))
		iflag = 1;

	if ((list2 = malloc(nmemb * size + PSIZE)) == NULL)
		return -1;

	list1 = base;
	setup_merge(list1, list2, nmemb, size, cmp);
	last = list2 + nmemb * size;
	i = big = 0;
	while (*EVAL(list2) != last) {
	    l2 = list1;
	    p1 = EVAL(list1);
	    for (tp2 = p2 = list2; p2 != last; p1 = EVAL(l2)) {
		p2 = *EVAL(p2);
		f1 = l2;
		f2 = l1 = list1 + (p2 - list2);
		if (p2 != last)
			p2 = *EVAL(p2);
		l2 = list1 + (p2 - list2);
		while (f1 < l1 && f2 < l2) {
			if ((*cmp)(f1, f2) <= 0) {
				q = f2;
				b = f1, t = l1;
				sense = -1;
			} else {
				q = f1;
				b = f2, t = l2;
				sense = 0;
			}
			if (!big) {     /* here i = 0 */
				while ((b += size) < t && cmp(q, b) >sense)
					if (++i == 6) {
						big = 1;
						goto EXPONENTIAL;
					}
			} else {
EXPONENTIAL:                    for (i = size; ; i <<= 1)
					if ((p = (b + i)) >= t) {
						if ((p = t - size) > b &&
						    (*cmp)(q, p) <= sense)
							t = p;
						else
							b = p;
						break;
					} else if ((*cmp)(q, p) <= sense) {
						t = p;
						if (i == size)
							big = 0;
						goto FASTCASE;
					} else
						b = p;
				while (t > b+size) {
					i = (((t - b) / size) >> 1) * size;
					if ((*cmp)(q, p = b + i) <= sense)
						t = p;
					else
						b = p;
				}
				goto COPY;
FASTCASE:                       while (i > size)
					if ((*cmp)(q,
						p = b + (i >>= 1)) <= sense)
						t = p;
					else
						b = p;
COPY:                           b = t;
			}
			i = size;
			if (q == f1) {
				if (iflag) {
					ICOPY_LIST(f2, tp2, b);
					ICOPY_ELT(f1, tp2, i);
				} else {
					CCOPY_LIST(f2, tp2, b);
					CCOPY_ELT(f1, tp2, i);
				}
			} else {
				if (iflag) {
					ICOPY_LIST(f1, tp2, b);
					ICOPY_ELT(f2, tp2, i);
				} else {
					CCOPY_LIST(f1, tp2, b);
					CCOPY_ELT(f2, tp2, i);
				}
			}
		}
		if (f2 < l2) {
			if (iflag)
				ICOPY_LIST(f2, tp2, l2);
			else
				CCOPY_LIST(f2, tp2, l2);
		} else if (f1 < l1) {
			if (iflag)
				ICOPY_LIST(f1, tp2, l1);
			else
				CCOPY_LIST(f1, tp2, l1);
		}
		*p1 = l2;
	    }
	    tp2 = list1;        /* swap list1, list2 */
	    list1 = list2;
	    list2 = tp2;
	    last = list2 + nmemb*size;
	}
	if (base == list2) {
		memmove(list2, list1, nmemb*size);
		list2 = list1;
	}
	free(list2);
	return 0;
}

/* Allocate keys array */
static struct keys_array *keys_array_alloc(void)
{
	struct keys_array *ka;
	size_t sz;

	sz = keys_array_size();
	ka = zmalloc(sz);

	return ka;
}

/* Set value of a key in the keys set */
static void set_key_on_keys_array(struct keys_array *ka, struct bwstring *s, size_t ind)
{
	if (ka && keys_num > ind) {
		struct key_value *kv;

		kv = &(ka->key[ind]);

		if (kv->k != s)
			free(kv->k);
		kv->k = s;
	}
}

/* Skip columns according to specs */
static size_t skip_cols_to_start(const struct bwstring *s, size_t cols, size_t start, bool skip_blanks, bool *empty_key)
{
	if (cols < 1)
		return BWSLEN(s) + 1;

	if (skip_blanks)
		while (start < BWSLEN(s) && ISWBLANK(BWS_GET(s, start)))
			++start;

	while (start < BWSLEN(s) && cols > 1) {
		--cols;
		++start;
	}

	if (start >= BWSLEN(s))
		*empty_key = true;

	return start;
}

/* Skip fields according to specs */
static size_t skip_fields_to_start(const struct bwstring *s, size_t fields, bool *empty_field)
{
	if (fields < 2) {
		if (BWSLEN(s) == 0)
			*empty_field = true;
		return 0;
	} else if (!(bb_common_bss.tflag)) {
		size_t cpos = 0;
		bool pb = true;

		while (cpos < BWSLEN(s)) {
			bool f_isblank;

			f_isblank = ISWBLANK(BWS_GET(s, cpos));

			if (f_isblank && !pb) {
				--fields;
				if (fields <= 1)
					return cpos;
			}
			pb = f_isblank;
			++cpos;
		}
		if (fields > 1)
			*empty_field = true;
		return cpos;
	} else {
		size_t cpos = 0;

		while (cpos < BWSLEN(s)) {
			if (BWS_GET(s, cpos) == (WCHAR_T_UTF8)bb_common_bss.field_sep) {
				--fields;
				if (fields <= 1)
					return cpos + 1;
			}
			++cpos;
		}
		if (fields > 1)
			*empty_field = true;
		return cpos;
	}
}

/* Find fields start */
static void find_field_start(const struct bwstring *s, struct key_specs *ks,
    size_t *field_start, size_t *key_start, bool *empty_field, bool *empty_key)
{
	*field_start = skip_fields_to_start(s, ks->f1, empty_field);
	if (!*empty_field)
		*key_start = skip_cols_to_start(s, ks->c1, *field_start,
		    ks->pos1b, empty_key);
	else
		*empty_key = true;
}

/* Find end key position */
static size_t find_field_end(const struct bwstring *s, struct key_specs *ks)
{
	size_t f2, next_field_start, pos_end;
	bool empty_field, empty_key;

	empty_field = false;
	empty_key = false;
	f2 = ks->f2;

	if (f2 == 0)
		return BWSLEN(s) + 1;
	else {
		if (ks->c2 == 0) {
			next_field_start = skip_fields_to_start(s, f2 + 1,
			    &empty_field);
			if ((next_field_start > 0) && bb_common_bss.tflag &&
			    ((WCHAR_T_UTF8)bb_common_bss.field_sep == BWS_GET(s,
			    next_field_start - 1)))
				--next_field_start;
		} else
			next_field_start = skip_fields_to_start(s, f2,
			    &empty_field);
	}

	if (empty_field || (next_field_start >= BWSLEN(s)))
		return BWSLEN(s) + 1;

	if (ks->c2) {
		pos_end = skip_cols_to_start(s, ks->c2, next_field_start,
		    ks->pos2b, &empty_key);
		if (pos_end < BWSLEN(s))
			++pos_end;
	} else
		pos_end = next_field_start;

	return pos_end;
}

/* Cut a field according to the key specs */
static struct bwstring *cut_field(const struct bwstring *s, struct key_specs *ks)
{
	struct bwstring *ret = NULL;

	if (s && ks) {
		size_t field_start, key_end, key_start, sz;
		bool empty_field, empty_key;

		field_start = 0;
		key_start = 0;
		empty_field = false;
		empty_key = false;

		find_field_start(s, ks, &field_start, &key_start,
		    &empty_field, &empty_key);

		if (empty_key)
			sz = 0;
		else {
			key_end = find_field_end(s, ks);
			sz = (key_end < key_start) ? 0 : (key_end - key_start);
		}

		ret = bwsalloc(sz);
		if (sz)
			bwsnocpy(ret, s, key_start, sz);
	} else
		ret = bwsalloc(0);

	return ret;
}

/* Preprocesses a line applying the necessary transformations
 * specified by command line options and returns the preprocessed
 * string, which can be used to compare. */
static int preproc(struct bwstring *s, struct keys_array *ka)
{
	if (bb_common_bss.kflag) {
		size_t i;
		for (i = 0; i < keys_num; i++) {
			struct bwstring *key;
			struct key_specs *kspecs;
			struct sort_mods *sm;

			kspecs = &(keys[i]);
			key = cut_field(s, kspecs);

			sm = &(kspecs->sm);
			if (sm->dflag)
				key = dictionary_order(key);
			else if (sm->iflag)
				key = ignore_nonprinting(key);
			if (sm->fflag || sm->Mflag)
				key = ignore_case(key);

			set_key_on_keys_array(ka, key, i);
		}
	} else {
		struct bwstring *ret = NULL;
		struct sort_mods *sm = default_sort_mods;

		if (sm->bflag) {
			if (ret == NULL)
				ret = bwsdup(s);
			ret = ignore_leading_blanks(ret);
		}
		if (sm->dflag) {
			if (ret == NULL)
				ret = bwsdup(s);
			ret = dictionary_order(ret);
		} else if (sm->iflag) {
			if (ret == NULL)
				ret = bwsdup(s);
			ret = ignore_nonprinting(ret);
		}
		if (sm->fflag || sm->Mflag) {
			if (ret == NULL)
				ret = bwsdup(s);
			ret = ignore_case(ret);
		}
		if (ret == NULL)
			set_key_on_keys_array(ka, s, 0);
		else
			set_key_on_keys_array(ka, ret, 0);
	}

	return 0;
}


#define LSCDEF(N)                                                                                       \
static int                                                                                              \
list_coll_##N(struct sort_list_item **ss1, struct sort_list_item **ss2)                                 \
{                                                                                                       \
													\
	return list_coll_offset(ss1, ss2, N);                                                           \
}

LSCDEF(1)
LSCDEF(2)
LSCDEF(3)
LSCDEF(4)
LSCDEF(5)
LSCDEF(6)
LSCDEF(7)
LSCDEF(8)
LSCDEF(9)
LSCDEF(10)
LSCDEF(11)
LSCDEF(12)
LSCDEF(13)
LSCDEF(14)
LSCDEF(15)
LSCDEF(16)
LSCDEF(17)
LSCDEF(18)
LSCDEF(19)
LSCDEF(20)

static listcoll_t get_list_call_func(size_t offset)
{
	static const listcoll_t lsarray[] = { (listcoll_t)list_coll, list_coll_1,
	    list_coll_2, list_coll_3, list_coll_4, list_coll_5,
	    list_coll_6, list_coll_7, list_coll_8, list_coll_9,
	    list_coll_10, list_coll_11, list_coll_12, list_coll_13,
	    list_coll_14, list_coll_15, list_coll_16, list_coll_17,
	    list_coll_18, list_coll_19, list_coll_20 };

	if (offset <= 20)
		return lsarray[offset];

	return (listcoll_t)list_coll;
}

/*
 * Maximum size of a number in the string (before or after decimal point)
 */
#define MAX_NUM_SIZE (128)

/* Set suffix value */
static void setsuffix(WCHAR_T_UTF8 c, unsigned char *si)
{
	switch (c){
	case Lc('k'):
	case Lc('K'):
		*si = 1;
		break;
	case Lc('M'):
		*si = 2;
		break;
	case Lc('G'):
		*si = 3;
		break;
	case Lc('T'):
		*si = 4;
		break;
	case Lc('P'):
		*si = 5;
		break;
	case Lc('E'):
		*si = 6;
		break;
	case Lc('Z'):
		*si = 7;
		break;
	case Lc('Y'):
		*si = 8;
		break;
	default:
		*si = 0;
	};
}

/* Read string s and parse the string into a fixed-decimal-point number.
 * sign equals -1 if the number is negative (explicit plus is not allowed,
 * according to GNU sort's "info sort".
 * The number part before decimal point is in the smain, after the decimal
 * point is in sfrac, tail is the pointer to the remainder of the string. */
static int read_number(struct bwstring *s0, int *sign, WCHAR_T_UTF8 *smain,
		size_t *main_len, WCHAR_T_UTF8 *sfrac, size_t *frac_len,
		unsigned char *si)
{
	bwstring_iterator s;

	s = bws_begin(s0);

	/* always end the fraction with zero, even if we have no fraction */
	sfrac[0] = 0;

	while (ISWBLANK(bws_get_iter_value(s)))
		s = bws_iterator_inc(s, 1);

	if (bws_get_iter_value(s) == (WCHAR_T_UTF8)symbol_negative_sign) {
		*sign = -1;
		s = bws_iterator_inc(s, 1);
	}

	// This is '0', not '\0', do not change this
	while (ISWDIGIT(bws_get_iter_value(s)) &&
	    (bws_get_iter_value(s) == Lc('0')))
		s = bws_iterator_inc(s, 1);

	while (bws_get_iter_value(s) && *main_len < MAX_NUM_SIZE) {
		if (ISWDIGIT(bws_get_iter_value(s))) {
			smain[*main_len] = bws_get_iter_value(s);
			s = bws_iterator_inc(s, 1);
			*main_len += 1;
		} else if (symbol_thousands_sep &&
		    (bws_get_iter_value(s) == (WCHAR_T_UTF8)symbol_thousands_sep))
			s = bws_iterator_inc(s, 1);
		else
			break;
	}

	smain[*main_len] = 0;

	if (bws_get_iter_value(s) == (WCHAR_T_UTF8)symbol_decimal_point) {
		s = bws_iterator_inc(s, 1);
		while (ISWDIGIT(bws_get_iter_value(s)) &&
		    *frac_len < MAX_NUM_SIZE) {
			sfrac[*frac_len] = bws_get_iter_value(s);
			s = bws_iterator_inc(s, 1);
			*frac_len += 1;
		}
		sfrac[*frac_len] = 0;

		while (*frac_len > 0 && sfrac[*frac_len - 1] == Lc('0')) {
			--(*frac_len);
			sfrac[*frac_len] = Lc('\0');
		}
	}

	setsuffix(bws_get_iter_value(s), si);

	if ((*main_len + *frac_len) == 0)
		*sign = 0;

	return 0;
}


/* Compare two suffixes */
static inline int cmpsuffix(unsigned char si1, unsigned char si2)
{
	return (char)si1 - (char)si2;
}

/* Implements numeric sort for -n and -h. */
static int numcoll_impl(struct key_value *kv1, struct key_value *kv2, bool use_suffix)
{
	struct bwstring *s1, *s2;
	WCHAR_T_UTF8 sfrac1[MAX_NUM_SIZE + 1], sfrac2[MAX_NUM_SIZE + 1];
	WCHAR_T_UTF8 smain1[MAX_NUM_SIZE + 1], smain2[MAX_NUM_SIZE + 1];
	int cmp_res, sign1, sign2;
	size_t frac1, frac2, main1, main2;
	unsigned char SI1, SI2;
	bool e1, e2, key1_read, key2_read;

	s1 = kv1->k;
	s2 = kv2->k;
	sign1 = sign2 = 0;
	main1 = main2 = 0;
	frac1 = frac2 = 0;

	key1_read = key2_read = false;

	if (debug_sort) {
		bwsprintf(stdout, s1, "; k1=<", ">");
		bwsprintf(stdout, s2, ", k2=<", ">");
	}

	if (s1 == s2)
		return 0;

	if (kv1->hint->status == HS_UNINITIALIZED) {
		/* read the number from the string */
		read_number(s1, &sign1, smain1, &main1, sfrac1, &frac1, &SI1);
		key1_read = true;
		kv1->hint->v.nh.n1 = WCSTOULL(smain1, NULL, 10);
		if (main1 < 1 && frac1 < 1)
			kv1->hint->v.nh.empty=true;
		kv1->hint->v.nh.si = SI1;
		kv1->hint->status = (kv1->hint->v.nh.n1 != ULLONG_MAX) ?
		    HS_INITIALIZED : HS_ERROR;
		kv1->hint->v.nh.neg = (sign1 < 0) ? true : false;
	}

	if (kv2->hint->status == HS_UNINITIALIZED) {
		/* read the number from the string */
		read_number(s2, &sign2, smain2, &main2, sfrac2, &frac2, &SI2);
		key2_read = true;
		kv2->hint->v.nh.n1 = WCSTOULL(smain2, NULL, 10);
		if (main2 < 1 && frac2 < 1)
			kv2->hint->v.nh.empty=true;
		kv2->hint->v.nh.si = SI2;
		kv2->hint->status = (kv2->hint->v.nh.n1 != ULLONG_MAX) ?
		    HS_INITIALIZED : HS_ERROR;
		kv2->hint->v.nh.neg = (sign2 < 0) ? true : false;
	}

	if (kv1->hint->status == HS_INITIALIZED && kv2->hint->status ==
	    HS_INITIALIZED) {
		unsigned long long n1, n2;
		bool neg1, neg2;

		e1 = kv1->hint->v.nh.empty;
		e2 = kv2->hint->v.nh.empty;

		if (e1 && e2)
			return 0;

		neg1 = kv1->hint->v.nh.neg;
		neg2 = kv2->hint->v.nh.neg;

		if (neg1 && !neg2)
			return -1;
		if (neg2 && !neg1)
			return 1;

		if (e1)
			return neg2 ? 1 : -1;
		else if (e2)
			return neg1 ? -1 : 1;


		if (use_suffix) {
			cmp_res = cmpsuffix(kv1->hint->v.nh.si, kv2->hint->v.nh.si);
			if (cmp_res)
				return neg1 ? -cmp_res : cmp_res;
		}

		n1 = kv1->hint->v.nh.n1;
		n2 = kv2->hint->v.nh.n1;
		if (n1 < n2)
			return neg1 ? 1 : -1;
		else if (n1 > n2)
			return neg1 ? -1 : 1;
	}

	/* read the numbers from the strings */
	if (!key1_read)
		read_number(s1, &sign1, smain1, &main1, sfrac1, &frac1, &SI1);
	if (!key2_read)
		read_number(s2, &sign2, smain2, &main2, sfrac2, &frac2, &SI2);

	e1 = ((main1 + frac1) == 0);
	e2 = ((main2 + frac2) == 0);

	if (e1 && e2)
		return 0;

	/* we know the result if the signs are different */
	if (sign1 < 0 && sign2 >= 0)
		return -1;
	if (sign1 >= 0 && sign2 < 0)
		return 1;

	if (e1)
		return (sign2 < 0) ? +1 : -1;
	else if (e2)
		return (sign1 < 0) ? -1 : 1;

	if (use_suffix) {
		cmp_res = cmpsuffix(SI1, SI2);
		if (cmp_res)
			return (sign1 < 0) ? -cmp_res : cmp_res;
	}

	/* if both numbers are empty assume that the strings are equal */
	if (main1 < 1 && main2 < 1 && frac1 < 1 && frac2 < 1)
		return 0;

	/*
	 * if the main part is of different size, we know the result
	 * (because the leading zeros are removed)
	 */
	if (main1 < main2)
		cmp_res = -1;
	else if (main1 > main2)
		cmp_res = +1;
	/* if the sizes are equal then simple non-collate string compare gives the correct result */
	else
		cmp_res = WCSCMP(smain1, smain2);

	/* check fraction */
	if (!cmp_res)
		cmp_res = WCSCMP(sfrac1, sfrac2);

	if (!cmp_res)
		return 0;

	/* reverse result if the signs are negative */
	if (sign1 < 0 && sign2 < 0)
		cmp_res = -cmp_res;

	return cmp_res;
}

/* Check for plus infinity */
static inline bool huge_plus(double d, int err1)
{
	if (err1 == ERANGE)
		if (d == HUGE_VAL || d == HUGE_VALF || d == HUGE_VALL)
			return 1;

	return 0;
}

#ifdef HAVE_ISNAN
/* Check whether a function is a NAN */
static bool
is_nan(double d)
{
#if defined(NAN)
	return (d == NAN || isnan(d));
#else
	return (isnan(d));
#endif
}
#endif

#ifdef HAVE_ISNAN
/* Compare two NANs */
static int cmp_nans(double d1, double d2)
{
	if (d1 == d2)
		return 0;
	return d1 < d2 ? -1 : 1;
}
#endif

/* Push sort level to the stack */
static inline void push_ls(struct sort_level *sl)
{
	struct level_stack *new_ls;

	new_ls = xmalloc(sizeof(struct level_stack));
	new_ls->sl = sl;

	new_ls->next = g_ls;
	g_ls = new_ls;
}

/* Pop sort level from the stack (single-threaded style) */
static inline struct sort_level *pop_ls_st(void)
{
	struct sort_level *sl;

	if (g_ls) {
		struct level_stack *saved_ls;

		sl = g_ls->sl;
		saved_ls = g_ls;
		g_ls = g_ls->next;
		free(saved_ls);
	} else
		sl = NULL;

	return sl;
}

static void add_to_sublevel(struct sort_level *sl, struct sort_list_item *item, size_t indx)
{
	struct sort_level *ssl;

	ssl = sl->sublevels[indx];

	if (ssl == NULL) {
		ssl = zmalloc(sizeof(struct sort_level));
		ssl->level = sl->level + 1;
		sl->sublevels[indx] = ssl;

		++(sl->real_sln);
	}

	if (++(ssl->tosort_num) > ssl->tosort_sz) {
		ssl->tosort_sz = ssl->tosort_num + 128;
		ssl->tosort = sort_reallocarray(ssl->tosort, ssl->tosort_sz,
		    sizeof(struct sort_list_item *));
	}

	ssl->tosort[ssl->tosort_num - 1] = item;
}

static inline void add_leaf(struct sort_level *sl, struct sort_list_item *item)
{
	if (++(sl->leaves_num) > sl->leaves_sz) {
		sl->leaves_sz = sl->leaves_num + 128;
		sl->leaves = sort_reallocarray(sl->leaves, sl->leaves_sz,
		    sizeof(struct sort_list_item *));
	}
	sl->leaves[sl->leaves_num - 1] = item;
}

static inline int get_wc_index(struct sort_list_item *sli, size_t level)
{
	const struct bwstring *bws;

	bws = sli->ka.key[0].k;

	if ((BWSLEN(bws) > level))
		return (unsigned char)BWS_GET(bws, level);
	return -1;
}

static void place_item(struct sort_level *sl, size_t item)
{
	struct sort_list_item *sli;
	int c;

	sli = sl->tosort[item];
	c = get_wc_index(sli, sl->level);

	if (c == -1)
		add_leaf(sl, sli);
	else
		add_to_sublevel(sl, sli, c);
}

static void free_sort_level(struct sort_level *sl)
{
	if (sl) {
		free(sl->leaves);

		if (sl->level > 0)
			free(sl->tosort);

		if (sl->sublevels) {
			struct sort_level *slc;
			size_t i, sln;

			sln = sl->sln;

			for (i = 0; i < sln; ++i) {
				slc = sl->sublevels[i];
				free_sort_level(slc);
			}

			free(sl->sublevels);
		}

		free(sl);
	}
}

static void run_sort_level_next(struct sort_level *sl)
{
	struct sort_level *slc;
	size_t i, sln, tosort_num;

	free(sl->sublevels);
	sl->sublevels = NULL;

	switch (sl->tosort_num) {
	case 0:
		goto end;
	case 1:
		sl->sorted[sl->start_position] = sl->tosort[0];
		goto end;
	case 2:
		if (list_coll_offset(&(sl->tosort[0]), &(sl->tosort[1]),
		    sl->level) > 0) {
			sl->sorted[sl->start_position++] = sl->tosort[1];
			sl->sorted[sl->start_position] = sl->tosort[0];
		} else {
			sl->sorted[sl->start_position++] = sl->tosort[0];
			sl->sorted[sl->start_position] = sl->tosort[1];
		}

		goto end;
	default:
		if (TINY_NODE(sl) || (sl->level > 15)) {
			listcoll_t func;

			func = get_list_call_func(sl->level);

			sl->leaves = sl->tosort;
			sl->leaves_num = sl->tosort_num;
			sl->leaves_sz = sl->leaves_num;
			sl->leaves = sort_reallocarray(sl->leaves,
			    sl->leaves_sz, sizeof(struct sort_list_item *));
			sl->tosort = NULL;
			sl->tosort_num = 0;
			sl->tosort_sz = 0;
			sl->sln = 0;
			sl->real_sln = 0;
			if (bb_common_bss.sflag) {
				if (mergesort(sl->leaves, sl->leaves_num,
				    sizeof(struct sort_list_item *),
				    (int(*)(const void *, const void *)) func) == -1)
					/* NOTREACHED */
					bb_error_msg_and_die("Radix sort error 3");
			} else
				DEFAULT_SORT_FUNC_RADIXSORT(sl->leaves, sl->leaves_num,
				    sizeof(struct sort_list_item *),
				    (int(*)(const void *, const void *)) func);

			memcpy(sl->sorted + sl->start_position,
			    sl->leaves, sl->leaves_num *
			    sizeof(struct sort_list_item *));

			goto end;
		} else {
			sl->tosort_sz = sl->tosort_num;
			sl->tosort = sort_reallocarray(sl->tosort,
			    sl->tosort_sz, sizeof(struct sort_list_item *));
		}
	}

	sl->sln = 256;
	sl->sublevels = zmalloc(slsz);
	sl->real_sln = 0;

	tosort_num = sl->tosort_num;
	for (i = 0; i < tosort_num; ++i)
		place_item(sl, i);

	free(sl->tosort);
	sl->tosort = NULL;
	sl->tosort_num = 0;
	sl->tosort_sz = 0;

	if (sl->leaves_num > 1) {
		if (keys_num > 1) {
			if (bb_common_bss.sflag) {
				mergesort(sl->leaves, sl->leaves_num,
				    sizeof(struct sort_list_item *), list_coll);
			} else {
				DEFAULT_SORT_FUNC_RADIXSORT(sl->leaves, sl->leaves_num,
				    sizeof(struct sort_list_item *), list_coll);
			}
		} else if (!bb_common_bss.sflag && bb_common_bss.complex_sort) {
			DEFAULT_SORT_FUNC_RADIXSORT(sl->leaves, sl->leaves_num,
			    sizeof(struct sort_list_item *), list_coll);
		}
	}

	sl->leaves_sz = sl->leaves_num;
	sl->leaves = sort_reallocarray(sl->leaves, sl->leaves_sz,
	    sizeof(struct sort_list_item *));

	if (!bb_common_bss.reverse_sort) {
		memcpy(sl->sorted + sl->start_position, sl->leaves,
		    sl->leaves_num * sizeof(struct sort_list_item *));
		sl->start_position += sl->leaves_num;

		free(sl->leaves);
		sl->leaves = NULL;
		sl->leaves_num = 0;
		sl->leaves_sz = 0;

		sln = sl->sln;

		for (i = 0; i < sln; ++i) {
			slc = sl->sublevels[i];

			if (slc) {
				slc->sorted = sl->sorted;
				slc->start_position = sl->start_position;
				sl->start_position += slc->tosort_num;
				if (SMALL_NODE(slc))
					run_sort_level_next(slc);
				else
					push_ls(slc);
				sl->sublevels[i] = NULL;
			}
		}

	} else {
		size_t n;

		sln = sl->sln;

		for (i = 0; i < sln; ++i) {
			n = sln - i - 1;
			slc = sl->sublevels[n];

			if (slc) {
				slc->sorted = sl->sorted;
				slc->start_position = sl->start_position;
				sl->start_position += slc->tosort_num;
				if (SMALL_NODE(slc))
					run_sort_level_next(slc);
				else
					push_ls(slc);
				sl->sublevels[n] = NULL;
			}
		}

		memcpy(sl->sorted + sl->start_position, sl->leaves,
		    sl->leaves_num * sizeof(struct sort_list_item *));
	}

end:
	free_sort_level(sl);
}

/* Single-threaded sort cycle */
static void run_sort_cycle_st(void)
{
	struct sort_level *slc;

	for (;;) {
		slc = pop_ls_st();
		if (slc == NULL) {
			break;
		}
		run_sort_level_next(slc);
	}
}

static void run_top_sort_level(struct sort_level *sl)
{
	struct sort_level *slc;
	size_t i;

	bb_common_bss.reverse_sort = bb_common_bss.kflag ? keys[0].sm.rflag :
	    default_sort_mods->rflag;

	sl->start_position = 0;
	sl->sln = 256;
	sl->sublevels = zmalloc(slsz);

	for (i = 0; i < sl->tosort_num; ++i)
		place_item(sl, i);

	if (sl->leaves_num > 1) {
		if (keys_num > 1) {
			if (bb_common_bss.sflag) {
				mergesort(sl->leaves, sl->leaves_num,
				    sizeof(struct sort_list_item *), list_coll);
			} else {
				DEFAULT_SORT_FUNC_RADIXSORT(sl->leaves, sl->leaves_num,
				    sizeof(struct sort_list_item *), list_coll);
			}
		} else if (!bb_common_bss.sflag && bb_common_bss.complex_sort) {
			DEFAULT_SORT_FUNC_RADIXSORT(sl->leaves, sl->leaves_num,
			    sizeof(struct sort_list_item *), list_coll);
		}
	}

	if (!bb_common_bss.reverse_sort) {

		memcpy(sl->tosort + sl->start_position, sl->leaves,
		    sl->leaves_num * sizeof(struct sort_list_item *));
		sl->start_position += sl->leaves_num;

		for (i = 0; i < sl->sln; ++i) {
			slc = sl->sublevels[i];

			if (slc) {
				slc->sorted = sl->tosort;
				slc->start_position = sl->start_position;
				sl->start_position += slc->tosort_num;
				push_ls(slc);
				sl->sublevels[i] = NULL;
			}
		}

	} else {
		size_t n;

		for (i = 0; i < sl->sln; ++i) {

			n = sl->sln - i - 1;
			slc = sl->sublevels[n];

			if (slc) {
				slc->sorted = sl->tosort;
				slc->start_position = sl->start_position;
				sl->start_position += slc->tosort_num;
				push_ls(slc);
				sl->sublevels[n] = NULL;
			}
		}

		memcpy(sl->tosort + sl->start_position, sl->leaves,
		    sl->leaves_num * sizeof(struct sort_list_item *));
	}
	run_sort_cycle_st();
}

static void rxsort(struct sort_list_item **base, size_t nmemb)
{
	struct sort_level *sl;

	sl = zmalloc(sizeof(struct sort_level));
	sl->tosort = base;
	sl->tosort_num = nmemb;
	sl->tosort_sz = nmemb;

	run_top_sort_level(sl);

	free_sort_level(sl);
}

static inline bool isdigit_clocale(WCHAR_T_UTF8 c)
{
	return (c >= Lc('0') && c <= Lc('9'));
}

static inline bool isalpha_clocale(WCHAR_T_UTF8 c)
{
	return ((c >= Lc('a') && c <= Lc('z')) || (c >= Lc('A') && c <= Lc('Z')));
}

static inline bool isalnum_clocale(WCHAR_T_UTF8 c)
{
	return ((c >= Lc('a') && c <= Lc('z')) || (c >= Lc('A') && c <= Lc('Z')) ||
	    (c >= Lc('0') && c <= Lc('9')));
}

/* Find string suffix of format: (\.[A-Za-z~][A-Za-z0-9~]*)*$
 * Set length of string before suffix. */
static void find_suffix(bwstring_iterator si, bwstring_iterator se, size_t *len)
{
	WCHAR_T_UTF8 c;
	size_t clen;
	bool expect_alpha, sfx;

	sfx = false;
	expect_alpha = false;
	*len = 0;
	clen = 0;

	while ((si < se) && (c = bws_get_iter_value(si))) {
		if (expect_alpha) {
			expect_alpha = false;
			if (!isalpha_clocale(c) && (c != Lc('~')))
				sfx = false;
		} else if (c == Lc('.')) {
			expect_alpha = true;
			if (!sfx) {
				sfx = true;
				*len = clen;
			}
		} else if (!isalnum_clocale(c) && (c != Lc('~')))
			sfx = false;

		si = bws_iterator_inc(si, 1);
		++clen;
	}

	/* This code must be here to make the implementation compatible
	 * with WORDING of GNU sort documentation.
	 * But the GNU sort implementation is not following its own
	 * documentation.  GNU sort allows empty file extensions
	 * (just dot with nothing after); but the regular expression in
	 * their documentation does not allow empty file extensions.
	 * We chose to make our implementation compatible with GNU sort
	 * implementation.  If they will ever fix their bug, this code
	 * must be uncommented. Or they may choose to fix the info page,
	 * then the code stays commented.
	 *
	 if (expect_alpha)
		sfx = false;
	 */

	if (!sfx)
		*len = clen;
}

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
static inline int cmp_chars(wchar_t c1, wchar_t c2)
#else
static inline int cmp_chars(int c1, int c2)
#endif
{
	if (c1 == c2)
		return 0;

	if (c1 == Lc('~'))
		return -1;
	if (c2 == Lc('~'))
		return 1;

	if (isdigit_clocale(c1) || !c1)
		return (isdigit_clocale(c2) || !c2) ? 0 : -1;

	if (isdigit_clocale(c2) || !c2)
		return 1;

	if (isalpha_clocale(c1))
		return isalpha_clocale(c2) ? (int)c1 - (int)c2 : -1;

	if (isalpha_clocale(c2))
		return 1;

	return (int)c1 - (int)c2;
}

static int cmpversions(bwstring_iterator si1, bwstring_iterator se1, bwstring_iterator si2, bwstring_iterator se2)
{
	int cmp, diff;

	while ((si1 < se1) || (si2 < se2)) {
		diff = 0;

		while (((si1 < se1) &&
		    !isdigit_clocale(bws_get_iter_value(si1))) ||
		    ((si2 < se2) && !isdigit_clocale(bws_get_iter_value(si2)))) {
			WCHAR_T_UTF8 c1, c2;

			c1 = (si1 < se1) ? bws_get_iter_value(si1) : 0;
			c2 = (si2 < se2) ? bws_get_iter_value(si2) : 0;

			cmp = cmp_chars(c1, c2);
			if (cmp)
				return cmp;

			if (si1 < se1)
				si1 = bws_iterator_inc(si1, 1);
			if (si2 < se2)
				si2 = bws_iterator_inc(si2, 1);
		}

		while (bws_get_iter_value(si1) == Lc('0'))
			si1 = bws_iterator_inc(si1, 1);

		while (bws_get_iter_value(si2) == Lc('0'))
			si2 = bws_iterator_inc(si2, 1);

		while (isdigit_clocale(bws_get_iter_value(si1)) &&
		    isdigit_clocale(bws_get_iter_value(si2))) {
			if (!diff)
				diff = ((int)bws_get_iter_value(si1) -
				    (int)bws_get_iter_value(si2));
			si1 = bws_iterator_inc(si1, 1);
			si2 = bws_iterator_inc(si2, 1);
		}

		if (isdigit_clocale(bws_get_iter_value(si1)))
			return 1;

		if (isdigit_clocale(bws_get_iter_value(si2)))
			return -1;

		if (diff)
			return diff;
	}

	return 0;
}

/* Compare two version strings */
static int vcmp(struct bwstring *s1, struct bwstring *s2)
{
	bwstring_iterator si1, si2;
	WCHAR_T_UTF8 c1, c2;
	size_t len1, len2, slen1, slen2;
	int cmp_bytes, cmp_res;

	if (s1 == s2)
		return 0;

	cmp_bytes = bwscmp(s1, s2, 0);
	if (cmp_bytes == 0)
		return 0;

	len1 = slen1 = BWSLEN(s1);
	len2 = slen2 = BWSLEN(s2);

	if (slen1 < 1)
		return -1;
	if (slen2 < 1)
		return 1;

	si1 = bws_begin(s1);
	si2 = bws_begin(s2);

	c1 = bws_get_iter_value(si1);
	c2 = bws_get_iter_value(si2);

	if (c1 == Lc('.') && (slen1 == 1))
		return -1;

	if (c2 == Lc('.') && (slen2 == 1))
		return 1;

	if (slen1 == 2 && c1 == Lc('.') &&
	    bws_get_iter_value(bws_iterator_inc(si1, 1)) == Lc('.'))
		return -1;
	if (slen2 == 2 && c2 == Lc('.') &&
	    bws_get_iter_value(bws_iterator_inc(si2, 1)) == Lc('.'))
		return 1;

	if (c1 == Lc('.') && c2 != Lc('.'))
		return -1;
	if (c1 != Lc('.') && c2 == Lc('.'))
		return 1;

	if (c1 == Lc('.') && c2 == Lc('.')) {
		si1 = bws_iterator_inc(si1, 1);
		si2 = bws_iterator_inc(si2, 1);
	}

	find_suffix(si1, bws_end(s1), &len1);
	find_suffix(si2, bws_end(s2), &len2);

	if ((len1 == len2) && (bws_iterator_cmp(si1, si2, len1) == 0))
		return cmp_bytes;

	cmp_res = cmpversions(si1, bws_iterator_inc(si1, len1), si2,
	    bws_iterator_inc(si2, len2));

	if (cmp_res == 0)
	  cmp_res = cmp_bytes;

	return cmp_res;
}

#define NUMBER_OF_MUTUALLY_EXCLUSIVE_FLAGS 6
static const char mutually_exclusive_flags[NUMBER_OF_MUTUALLY_EXCLUSIVE_FLAGS] = { 'M', 'n', 'g', 'R', 'h', 'V' };

static const struct option long_options[] = {
	{ "batch-size", required_argument, NULL, BS_OPT },
	{ "buffer-size", required_argument, NULL, 'S' },
	{ "check", optional_argument, NULL, 'c' },
	{ "check=silent|quiet", optional_argument, NULL, 'C' },
	{ "compress-program", required_argument, NULL, COMPRESSPROGRAM_OPT },
#ifdef CONFIG_FEATURE_SORT_BIG
	{ "debug", no_argument, NULL, DEBUG_OPT },
#endif
	{ "dictionary-order", no_argument, NULL, 'd' },
	{ "field-separator", required_argument, NULL, 't' },
	{ "files0-from", required_argument, NULL, FF_OPT },
	{ "general-numeric-sort", no_argument, NULL, 'g' },
	{ "heapsort", no_argument, NULL, HEAPSORT_OPT },
	{ "human-numeric-sort", no_argument, NULL, 'h' },
	{ "ignore-leading-blanks", no_argument, NULL, 'b' },
	{ "ignore-case", no_argument, NULL, 'f' },
	{ "ignore-nonprinting", no_argument, NULL, 'i' },
	{ "key", required_argument, NULL, 'k' },
	{ "merge", no_argument, NULL, 'm' },
	{ "mergesort", no_argument, NULL, 'H' },
	{ "mmap", no_argument, NULL, MMAP_OPT },
	{ "month-sort", no_argument, NULL, 'M' },
	{ "numeric-sort", no_argument, NULL, 'n' },
	{ "output", required_argument, NULL, 'o' },
	{ "qsort", no_argument, NULL, QSORT_OPT },
	{ "radixsort", no_argument, NULL, RADIXSORT_OPT },
	{ "random-sort", no_argument, NULL, 'R' },
	{ "random-source", required_argument, NULL, RANDOMSOURCE_OPT },
	{ "reverse", no_argument, NULL, 'r' },
	{ "sort", required_argument, NULL, SORT_OPT },
	{ "stable", no_argument, NULL, 's' },
	{ "temporary-directory", required_argument, NULL, 'T' },
	{ "unique", no_argument, NULL, 'u' },
	{ "version", no_argument, NULL, VERSION_OPT },
	{ "version-sort", no_argument, NULL, 'V' },
	{ "zero-terminated", no_argument, NULL, 'z' },
	{ NULL, no_argument, NULL, 0 }
};

/* Check where sort modifier is present */
static bool sort_modifier_empty(struct sort_mods *sm)
{
	return !(sm->Mflag || sm->Vflag || sm->nflag || sm->gflag ||
	    sm->rflag || sm->Rflag || sm->hflag || sm->dflag || sm->fflag);
}

/* Read input file names from a file (file0-from option). */
static void read_fns_from_file0(const char *fn)
{
	FILE *f;
	char *line = NULL;
	size_t linesize = 0;
	ssize_t linelen;

	f = bb_xfopen(fn, "Ar");

	while ((linelen = getdelim(&line, &linesize, '\0', f)) != -1) {
		if (*line != '\0') {
			if (bb_common_bss.argc_from_file0 == (size_t)-1)
				bb_common_bss.argc_from_file0 = 0;
			++bb_common_bss.argc_from_file0;
			bb_common_bss.argv_from_file0 = sort_reallocarray(bb_common_bss.argv_from_file0,
			    bb_common_bss.argc_from_file0, sizeof(char *));
			bb_common_bss.argv_from_file0[bb_common_bss.argc_from_file0 - 1] = line;
		} else {
			free(line);
		}
		line = NULL;
		linesize = 0;
	}
	if (ferror(f))
		bb_perror_msg_and_die("%s: getdelim", fn);

	closefile(f, fn);
}

/* Check how much RAM is available for the sort. */
static void set_hw_params(void)
{
	unsigned long long free_memory;
	struct rlimit rl;

	/* Increase our data size to the max */
	if (getrlimit(RLIMIT_DATA, &rl) == 0) {
		free_memory = (unsigned long long)rl.rlim_cur;
		rl.rlim_cur = rl.rlim_max;
		if (setrlimit(RLIMIT_DATA, &rl) == 0) {
			free_memory = (unsigned long long)rl.rlim_max;
		} else {
			bb_perror_msg("Can't set resource limit to max data size");
		}
	} else {
		free_memory = 1000000;
		bb_perror_msg("Can't get resource limit for data size");
	}

	bb_common_bss.available_free_memory = free_memory / 2;
}

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
/* Convert "plain" symbol to wide symbol, with default value. */
static void conv_mbtowc(wchar_t *wc, const char *c, const wchar_t def)
{
	int res;

	res = mbtowc(wc, c, MB_CUR_MAX);
	if (res < 1)
		*wc = def;
}
#endif

#ifdef CONFIG_LOCALE_SUPPORT
/* Set current locale symbols. */
static void set_locale(void)
{
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	struct lconv *lc;
#endif
	const char *locale;

	setlocale(LC_ALL, "");

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	/* Obtain LC_NUMERIC info */
	lc = localeconv();

	/* Convert to wide char form */
	conv_mbtowc((wchar_t *)&symbol_decimal_point, lc->decimal_point, (wchar_t)symbol_decimal_point);
	conv_mbtowc((wchar_t *)&symbol_thousands_sep, lc->thousands_sep, (wchar_t)symbol_thousands_sep);
	conv_mbtowc((wchar_t *)&symbol_positive_sign, lc->positive_sign, (wchar_t)symbol_positive_sign);
	conv_mbtowc((wchar_t *)&symbol_negative_sign, lc->negative_sign, (wchar_t)symbol_negative_sign);
#endif
	locale = setlocale(LC_COLLATE, NULL);
	if (locale != NULL) {
		char *tmpl;
		const char *byteclocale;

		tmpl = xstrdup(locale);
		byteclocale = setlocale(LC_COLLATE, "C");
		if (byteclocale && strcmp(byteclocale, tmpl) == 0) {
			byte_sort = true;
		} else {
			byteclocale = setlocale(LC_COLLATE, "POSIX");
			if (byteclocale && strcmp(byteclocale, tmpl) == 0)
				byte_sort = true;
			else
				setlocale(LC_COLLATE, tmpl);
		}
		free(tmpl);
	}
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	if (!byte_sort)
		sort_mb_cur_max = MB_CUR_MAX;
#endif
}
#endif /* CONFIG_LOCALE_SUPPORT */

/* Parse -S option. */
static unsigned long long parse_memory_buffer_value(const char *value)
{
	char *endptr;
	unsigned long long membuf;

	membuf = bb_xparse_uim_number(10 | BB_PN_SILENT_RE |
			BB_PN_SET_UPPER | BB_PN_END_PTR,
			value, (uintmax_t)LLONG_MAX, &endptr);
	if (errno)
		goto invalid;

	switch (*endptr) {
	case 'Y':
		if (membuf > ULLONG_MAX / 1024)
			goto invalid;
		membuf *= 1024;
		/* FALLTHROUGH */
	case 'Z':
		if (membuf > ULLONG_MAX / 1024)
			goto invalid;
		membuf *= 1024;
		/* FALLTHROUGH */
	case 'E':
		if (membuf > ULLONG_MAX / 1024)
			goto invalid;
		membuf *= 1024;
		/* FALLTHROUGH */
	case 'P':
		if (membuf > ULLONG_MAX / 1024)
			goto invalid;
		membuf *= 1024;
		/* FALLTHROUGH */
	case 'T':
		if (membuf > ULLONG_MAX / 1024)
			goto invalid;
		membuf *= 1024;
		/* FALLTHROUGH */
	case 'G':
		if (membuf > ULLONG_MAX / 1024)
			goto invalid;
		membuf *= 1024;
		/* FALLTHROUGH */
	case 'M':
		if (membuf > ULLONG_MAX / 1024)
			goto invalid;
		membuf *= 1024;
		/* FALLTHROUGH */
	case '\0':
	case 'K':
		if (membuf > ULLONG_MAX / 1024)
			goto invalid;
		membuf *= 1024;
		/* FALLTHROUGH */
	case 'b':
		break;
	case '%':
		if (bb_common_bss.available_free_memory != 0 &&
		    membuf > ULLONG_MAX / bb_common_bss.available_free_memory)
			goto invalid;
		membuf = (bb_common_bss.available_free_memory * membuf) / 100;
		break;
	default:
		errno = EINVAL;
		bb_perror_msg("%s", optarg);
		membuf = bb_common_bss.available_free_memory;
	}
	if (membuf > SIZE_MAX)
		goto invalid;
	return membuf;
invalid:
	bb_error_msg_and_die("invalid memory buffer size: %s", value);
}

/* Signal handler that clears the temporary files. */
static void sig_handler(int sig ATTRIBUTE_UNUSED)
{
	clear_tmp_files();
	_exit(2);
}

/* Print "unknown" message and exit with status 2. */
static void unknown(const char *what)
{
	bb_error_msg_and_die("Unknown feature: %s", what);
}

/* Check whether contradictory input options are used. */
static void check_mutually_exclusive_flags(char c, bool *mef_flags)
{
	int i, fo_index, mec;
	bool found_others, found_this;

	found_others = found_this = false;
	fo_index = 0;

	for (i = 0; i < NUMBER_OF_MUTUALLY_EXCLUSIVE_FLAGS; i++) {
		mec = mutually_exclusive_flags[i];

		if (mec != c) {
			if (mef_flags[i]) {
				if (found_this) {
					bb_error_msg_and_die("%c:%c: mutually exclusive flags", c, mec);
				}
				found_others = true;
				fo_index = i;
			}
		} else {
			if (found_others) {
				bb_error_msg_and_die("%c:%c: mutually exclusive flags", c, mutually_exclusive_flags[fo_index]);
			}
			mef_flags[i] = true;
			found_this = true;
		}
	}
}

/* Implements numeric sort (-n). */
static int numcoll(struct key_value *kv1, struct key_value *kv2, size_t offset ATTRIBUTE_UNUSED)
{
	return numcoll_impl(kv1, kv2, false);
}

/* Implements 'human' numeric sort (-h). */
static int hnumcoll(struct key_value *kv1, struct key_value *kv2, size_t offset ATTRIBUTE_UNUSED)
{
	return numcoll_impl(kv1, kv2, true);
}

/* Check for minus infinity */
static inline bool huge_minus(double d, int err1)
{
	if (err1 == ERANGE)
		if (d == -HUGE_VAL || d == -HUGE_VALF || d == -HUGE_VALL)
			return 1;

	return 0;
}

/* Implements general numeric sort (-g). */
static int gnumcoll(struct key_value *kv1, struct key_value *kv2, size_t offset ATTRIBUTE_UNUSED)
{
	double d1, d2;
	int err1, err2;
	bool empty1, empty2, key1_read, key2_read;

	d1 = d2 = 0;
	err1 = err2 = 0;
	key1_read = key2_read = false;

	if (debug_sort) {
		bwsprintf(stdout, kv1->k, "; k1=<", ">");
		bwsprintf(stdout, kv2->k, "; k2=<", ">");
	}

	if (kv1->hint->status == HS_UNINITIALIZED) {
		errno = 0;
		d1 = bwstod(kv1->k, &empty1);
		err1 = errno;

		if (empty1) {
			kv1->hint->v.gh.notnum = true;
			kv1->hint->status = HS_INITIALIZED;
		} else if (err1 == 0) {
			kv1->hint->v.gh.d = d1;
#ifdef HAVE_ISNAN
			kv1->hint->v.gh.nan = is_nan(d1);
#endif
			kv1->hint->status = HS_INITIALIZED;
		} else
			kv1->hint->status = HS_ERROR;

		key1_read = true;
	}

	if (kv2->hint->status == HS_UNINITIALIZED) {
		errno = 0;
		d2 = bwstod(kv2->k, &empty2);
		err2 = errno;

		if (empty2) {
			kv2->hint->v.gh.notnum = true;
			kv2->hint->status = HS_INITIALIZED;
		} else if (err2 == 0) {
			kv2->hint->v.gh.d = d2;
#ifdef HAVE_ISNAN
			kv2->hint->v.gh.nan = is_nan(d2);
#endif
			kv2->hint->status = HS_INITIALIZED;
		} else
			kv2->hint->status = HS_ERROR;

		key2_read = true;
	}

	if (kv1->hint->status == HS_INITIALIZED &&
	    kv2->hint->status == HS_INITIALIZED) {
		if (kv1->hint->v.gh.notnum)
			return kv2->hint->v.gh.notnum ? 0 : -1;
		else if (kv2->hint->v.gh.notnum)
			return 1;

#ifdef HAVE_ISNAN
		if (kv1->hint->v.gh.nan)
			return kv2->hint->v.gh.nan ?
			    cmp_nans(kv1->hint->v.gh.d, kv2->hint->v.gh.d) : -1;
		else if (kv2->hint->v.gh.nan)
			return 1;
#endif
		d1 = kv1->hint->v.gh.d;
		d2 = kv2->hint->v.gh.d;

		if (d1 < d2)
			return -1;
		else if (d1 > d2)
			return 1;
		else
			return 0;
	}

	if (!key1_read) {
		errno = 0;
		d1 = bwstod(kv1->k, &empty1);
		err1 = errno;
	}

	if (!key2_read) {
		errno = 0;
		d2 = bwstod(kv2->k, &empty2);
		err2 = errno;
	}

	/* Non-value case */
	if (empty1)
		return empty2 ? 0 : -1;
	else if (empty2)
		return 1;

#ifdef HAVE_ISNAN
	/* NAN case */
	if (is_nan(d1))
		return is_nan(d2) ? cmp_nans(d1, d2) : -1;
	else if (is_nan(d2))
		return 1;
#endif

	/* Infinities */
	if (err1 == ERANGE || err2 == ERANGE) {
		/* Minus infinity case */
		if (huge_minus(d1, err1)) {
			if (huge_minus(d2, err2)) {
				if (d1 == d2)
					return 0;
				return d1 < d2 ? -1 : 1;
			} else
				return -1;

		} else if (huge_minus(d2, err2)) {
			if (huge_minus(d1, err1)) {
				if (d1 == d2)
					return 0;
				return d1 < d2 ? -1 : 1;
			} else
				return 1;
		}

		/* Plus infinity case */
		if (huge_plus(d1, err1)) {
			if (huge_plus(d2, err2)) {
				if (d1 == d2)
					return 0;
				return d1 < d2 ? -1 : 1;
			} else
				return 1;
		} else if (huge_plus(d2, err2)) {
			if (huge_plus(d1, err1)) {
				if (d1 == d2)
					return 0;
				return d1 < d2 ? -1 : 1;
			} else
				return -1;
		}
	}

	if (d1 == d2)
		return 0;
	return d1 < d2 ? -1 : 1;
}

/* Implements month sort (-M). */
static int monthcoll(struct key_value *kv1, struct key_value *kv2, size_t offset ATTRIBUTE_UNUSED)
{
	int val1, val2;
	bool key1_read, key2_read;

	val1 = val2 = 0;
	key1_read = key2_read = false;

	if (debug_sort) {
		bwsprintf(stdout, kv1->k, "; k1=<", ">");
		bwsprintf(stdout, kv2->k, "; k2=<", ">");
	}

	if (kv1->hint->status == HS_UNINITIALIZED) {
		kv1->hint->v.Mh.m = bws_month_score(kv1->k);
		key1_read = true;
		kv1->hint->status = HS_INITIALIZED;
	}

	if (kv2->hint->status == HS_UNINITIALIZED) {
		kv2->hint->v.Mh.m = bws_month_score(kv2->k);
		key2_read = true;
		kv2->hint->status = HS_INITIALIZED;
	}

	if (kv1->hint->status == HS_INITIALIZED) {
		val1 = kv1->hint->v.Mh.m;
		key1_read = true;
	}

	if (kv2->hint->status == HS_INITIALIZED) {
		val2 = kv2->hint->v.Mh.m;
		key2_read = true;
	}

	if (!key1_read)
		val1 = bws_month_score(kv1->k);
	if (!key2_read)
		val2 = bws_month_score(kv2->k);

	if (val1 == val2)
		return 0;
	return val1 < val2 ? -1 : 1;
}

/* Implements random sort (-R). */
static int randomcoll(struct key_value *kv1, struct key_value *kv2, size_t offset ATTRIBUTE_UNUSED)
{
	struct bwstring *s1, *s2;
	md5_ctx_t ctx1, ctx2;
	unsigned char digest1[MD5_DIGEST_LENGTH];
	unsigned char digest2[MD5_DIGEST_LENGTH];
	int cmp_res;

	s1 = kv1->k;
	s2 = kv2->k;

	if (debug_sort) {
		bwsprintf(stdout, s1, "; k1=<", ">");
		bwsprintf(stdout, s2, ", k2=<", ">");
	}

	if (s1 == s2)
		return 0;

	memcpy(&ctx1, &md5_ctx, sizeof(ctx1));
	memcpy(&ctx2, &md5_ctx, sizeof(ctx2));

	md5_hash(bwsrawdata(s1), bwsrawlen(s1), &ctx1);
	md5_hash(bwsrawdata(s2), bwsrawlen(s2), &ctx2);
	md5_end(digest1, &ctx1);
	md5_end(digest2, &ctx2);

	cmp_res = memcmp(digest1, digest2, sizeof(MD5_DIGEST_LENGTH));
	if (!cmp_res)
		cmp_res = bwscoll(s1, s2, 0);

	return cmp_res;
}

/* Implements version sort (-V). */
static int versioncoll(struct key_value *kv1, struct key_value *kv2, size_t offset ATTRIBUTE_UNUSED)
{
	struct bwstring *s1, *s2;

	s1 = kv1->k;
	s2 = kv2->k;

	if (debug_sort) {
		bwsprintf(stdout, s1, "; k1=<", ">");
		bwsprintf(stdout, s2, ", k2=<", ">");
	}

	if (s1 == s2)
		return 0;

	return vcmp(s1, s2);
}

/* Implements string sort. */
static int wstrcoll(struct key_value *kv1, struct key_value *kv2, size_t offset)
{

	if (debug_sort) {
		if (offset)
			printf("; offset=%zu\n", offset);
		bwsprintf(stdout, kv1->k, "; k1=<", ">");
		printf("(%zu)", BWSLEN(kv1->k));
		bwsprintf(stdout, kv2->k, ", k2=<", ">");
		printf("(%zu)", BWSLEN(kv2->k));
	}

	return bwscoll(kv1->k, kv2->k, offset);
}

static cmpcoll_t get_sort_func(struct sort_mods *sm)
{
	if (sm->nflag)
		return numcoll;
	else if (sm->hflag)
		return hnumcoll;
	else if (sm->gflag)
		return gnumcoll;
	else if (sm->Mflag)
		return monthcoll;
	else if (sm->Rflag)
		return randomcoll;
	else if (sm->Vflag)
		return versioncoll;
	else
		return wstrcoll;
}

/* Set a sort modifier on a sort modifiers object. */
static bool set_sort_modifier(struct sort_mods *sm, int c)
{
	switch (c) {
	case 'b':
		sm->bflag = true;
		break;
	case 'd':
		sm->dflag = true;
		break;
	case 'f':
		sm->fflag = true;
		break;
	case 'g':
		sm->gflag = true;
		bb_common_bss.need_hint = true;
		break;
	case 'i':
		sm->iflag = true;
		break;
	case 'R':
		sm->Rflag = true;
		bb_common_bss.need_random = true;
		break;
	case 'M':
		initialise_months();
		sm->Mflag = true;
		bb_common_bss.need_hint = true;
		break;
	case 'n':
		sm->nflag = true;
		bb_common_bss.need_hint = true;
		bb_common_bss.print_symbols_on_debug = true;
		break;
	case 'r':
		sm->rflag = true;
		break;
	case 'V':
		sm->Vflag = true;
		break;
	case 'h':
		sm->hflag = true;
		bb_common_bss.need_hint = true;
		bb_common_bss.print_symbols_on_debug = true;
		break;
	default:
		return false;
	}
	bb_common_bss.complex_sort = true;
	sm->func = get_sort_func(sm);

	return true;
}

/* Parse POS in -k option. */
static int parse_pos(const char *s, struct key_specs *ks, bool *mef_flags, bool second)
{
	regmatch_t pmatch[4];
	regex_t re;
	char *c, *f;
	size_t len, nmatch;
	int ret;

	ret = -1;
	nmatch = 4;
	c = f = NULL;

	if (regcomp(&re, "^([0-9]+)(\\.[0-9]+)?([bdfirMngRhV]+)?$", REG_EXTENDED) != 0)
		return -1;

	if (regexec(&re, s, nmatch, pmatch, 0) != 0)
		goto end;

	if (pmatch[0].rm_eo <= pmatch[0].rm_so)
		goto end;

	if (pmatch[1].rm_eo <= pmatch[1].rm_so)
		goto end;

	len = pmatch[1].rm_eo - pmatch[1].rm_so;

	f = xmalloc(len + 1);
	memcpy(f, s + pmatch[1].rm_so, len);
	f[len] = '\0';

	if (second) {
		errno = 0;
		ks->f2 = (size_t)strtoul(f, NULL, 10);
		if (errno != 0)
			goto end;
		if (ks->f2 == 0) {
			bb_error_msg("0 field in key specs");
			goto end;
		}
	} else {
		errno = 0;
		ks->f1 = (size_t)strtoul(f, NULL, 10);
		if (errno != 0)
			goto end;
		if (ks->f1 == 0) {
			bb_error_msg("0 field in key specs");
			goto end;
		}
	}

	if (pmatch[2].rm_eo > pmatch[2].rm_so) {
		len = pmatch[2].rm_eo - pmatch[2].rm_so - 1;

		c = xmalloc(len + 1);
		memcpy(c, s + pmatch[2].rm_so + 1, len);
		c[len] = '\0';

		if (second) {
			errno = 0;
			ks->c2 = (size_t)strtoul(c, NULL, 10);
			if (errno != 0)
				goto end;
		} else {
			errno = 0;
			ks->c1 = (size_t)strtoul(c, NULL, 10);
			if (errno != 0)
				goto end;
			if (ks->c1 == 0) {
				bb_error_msg("0 column in key specs");
				goto end;
			}
		}
	} else {
		if (second)
			ks->c2 = 0;
		else
			ks->c1 = 1;
	}

	if (pmatch[3].rm_eo > pmatch[3].rm_so) {
		regoff_t i = 0;

		for (i = pmatch[3].rm_so; i < pmatch[3].rm_eo; i++) {
			check_mutually_exclusive_flags(s[i], mef_flags);
			if (s[i] == 'b') {
				if (second)
					ks->pos2b = true;
				else
					ks->pos1b = true;
			} else if (!set_sort_modifier(&(ks->sm), s[i]))
				goto end;
		}
	}

	ret = 0;

end:
	free(c);
	free(f);
	regfree(&re);

	return ret;
}

/* Parse -k option value. */
static int parse_k(const char *s, struct key_specs *ks)
{
	int ret = -1;
	bool mef_flags[NUMBER_OF_MUTUALLY_EXCLUSIVE_FLAGS] =
	    { false, false, false, false, false, false };

	if (*s != '\0') {
		char *sptr;

		sptr = strchr(s, ',');
		if (sptr) {
			size_t size1;
			char *pos1, *pos2;

			size1 = sptr - s;

			if (size1 < 1)
				return -1;

			pos1 = xstrndup(s, size1);

			ret = parse_pos(pos1, ks, mef_flags, false);

			free(pos1);
			if (ret < 0)
				return ret;

			pos2 = xstrdup(sptr + 1);
			ret = parse_pos(pos2, ks, mef_flags, true);
			free(pos2);
		} else
			ret = parse_pos(s, ks, mef_flags, false);
	}

	return ret;
}

/* Parse POS in +POS -POS option. */
static int parse_pos_obs(const char *s, size_t *nf, size_t *nc, char *sopts, size_t sopts_size)
{
	regex_t re;
	regmatch_t pmatch[4];
	char *c, *f;
	int ret;
	size_t len, nmatch;

	ret = -1;
	nmatch = 4;
	c = f = NULL;
	*nc = *nf = 0;

	if (regcomp(&re, "^([0-9]+)(\\.[0-9]+)?([A-Za-z]+)?$", REG_EXTENDED) != 0)
		return -1;

	if (regexec(&re, s, nmatch, pmatch, 0) != 0)
		goto end;

	if (pmatch[0].rm_eo <= pmatch[0].rm_so)
		goto end;

	if (pmatch[1].rm_eo <= pmatch[1].rm_so)
		goto end;

	len = pmatch[1].rm_eo - pmatch[1].rm_so;

	f = xmalloc(len + 1);
	memcpy(f, s + pmatch[1].rm_so, len);
	f[len] = '\0';

	*nf = (size_t)bb_xparse_ui_number(10 | BB_PN_ERR_STR1, f, "Invalid key position");

	if (pmatch[2].rm_eo > pmatch[2].rm_so) {
		len = pmatch[2].rm_eo - pmatch[2].rm_so - 1;

		c = xmalloc(len + 1);
		memcpy(c, s + pmatch[2].rm_so + 1, len);
		c[len] = '\0';

		*nc = (size_t)bb_xparse_ui_number(10 | BB_PN_ERR_STR1, c, "Invalid key position");
	}

	if (pmatch[3].rm_eo > pmatch[3].rm_so) {

		len = pmatch[3].rm_eo - pmatch[3].rm_so;

		if (len >= sopts_size)
			bb_error_msg_and_die("Invalid key position");
		memcpy(sopts, s + pmatch[3].rm_so, len);
		sopts[len] = '\0';
	}

	ret = 0;

end:
	free(c);
	free(f);
	regfree(&re);

	return ret;
}

/* "Translate" obsolete +POS1 -POS2 syntax into new -kPOS1,POS2 syntax */
static void fix_obsolete_keys(int *argc, char **argv)
{
	int i;

	for (i = 1; i < *argc; i++) {
		const char *arg1 = argv[i];

		if (arg1[0] == '+') {
			size_t c1, f1;
			char sopts1[128];

			sopts1[0] = 0;
			c1 = f1 = 0;

			if (parse_pos_obs(arg1 + 1, &f1, &c1, sopts1, sizeof(sopts1)) < 0)
				continue;

			f1 += 1;
			c1 += 1;
			if (i + 1 < *argc) {
				const char *arg2 = argv[i + 1];

				if (arg2[0] == '-') {
					size_t c2 =0, f2 = 0;
					char sopts2[128];

					sopts2[0] = 0;

					if (parse_pos_obs(arg2 + 1, &f2, &c2, sopts2, sizeof(sopts2)) >= 0) {
						int j;
						if (c2 > 0)
							f2 += 1;
						argv[i] = xasprintf("-k%zu.%zu%s,%zu.%zu%s",
						    f1, c1, sopts1, f2, c2, sopts2);
						for (j = i + 1; j + 1 < *argc; j++)
							argv[j] = argv[j + 1];
						*argc -= 1;
						continue;
					}
				}
			}
			argv[i] = xasprintf("-k%zu.%zu%s", f1, c1, sopts1);
		}
	}
}

/* Set random seed */
static void set_random_seed(void)
{
	unsigned char buf[MD5_BLOCK_LENGTH];
	FILE *fp;

	if (!bb_common_bss.need_random)
		return;

	md5_begin(&md5_ctx);
	if (bb_common_bss.random_source == NULL)
		bb_common_bss.random_source = "/dev/urandom";

	if ((fp = fopen(bb_common_bss.random_source, "r")) == NULL || fread(buf, 1, sizeof(buf), fp) != sizeof(buf))
		bb_perror_msg_and_die("%s", bb_common_bss.random_source);
	md5_hash(buf, sizeof(buf), &md5_ctx);
	if (ferror(fp))
		bb_perror_msg_and_die("%s", bb_common_bss.random_source);
	fclose(fp);
}

/* Calculate key for a sort list item */
static void sort_list_item_make_key(struct sort_list_item *si)
{
	preproc(si->str, &(si->ka));
}

/* Set value of a sort list item. Return combined string and keys memory size. */
static void sort_list_item_set(struct sort_list_item *si, struct bwstring *str)
{
	if (si) {
		clean_keys_array(si->str, &(si->ka));
		if (si->str) {
			if (si->str == str) {
				/* we are trying to reset the same string */
				return;
			} else {
				free(si->str);
				si->str = NULL;
			}
		}
		si->str = str;
		sort_list_item_make_key(si);
	}
}

/* Add string to sort list */
static void sort_list_add(struct sort_list *l, struct bwstring *str)
{
	size_t indx = l->count;

	if ((l->list == NULL) || (indx >= l->size)) {
		size_t newsize = (l->size + 1) + 1024;

		l->list = sort_reallocarray(l->list, newsize,
		    sizeof(struct sort_list_item *));
		l->memsize += (newsize - l->size) *
		    sizeof(struct sort_list_item *);
		l->size = newsize;
	}
	l->list[indx] = sort_list_item_alloc();
	sort_list_item_set(l->list[indx], str);
	l->memsize += sort_list_item_size(l->list[indx]);
	l->count += 1;
}

/* Clean sort list data */
static void sort_list_clean(struct sort_list *l)
{
	if (l->list) {
		size_t i;

		for (i = 0; i < l->count; i++) {
			struct sort_list_item *item;

			item = l->list[i];

			if (item) {
				sort_list_item_clean(item);
				free(item);
				l->list[i] = NULL;
			}
		}
		free(l->list);
		l->list = NULL;
	}
	l->count = 0;
	l->size = 0;
	l->memsize = sizeof(struct sort_list);
}

/* Sort list of lines and writes it to the file */
static void sort_list_to_file(struct sort_list *list, const char *outfile)
{
	struct sort_mods *sm = &(keys[0].sm);

	if (!sm->Mflag && !sm->Rflag && !sm->Vflag &&
	    !sm->gflag && !sm->hflag && !sm->nflag) {
		if ((bb_common_bss.sort_method == SORT_DEFAULT) && byte_sort)
			bb_common_bss.sort_method = SORT_RADIXSORT;

	} else if (bb_common_bss.sort_method == SORT_RADIXSORT)
		bb_error_msg_and_die("Radix sort cannot be used with these sort options");

	/* To handle stable sort and the unique cases in the
	 * right order, we need to use a stable algorithm. */
	if (bb_common_bss.sflag) {
		switch (bb_common_bss.sort_method){
		case SORT_MERGESORT:
			break;
		case SORT_RADIXSORT:
			break;
		case SORT_DEFAULT:
			bb_common_bss.sort_method = SORT_MERGESORT;
			break;
		default:
			bb_error_msg_and_die("The chosen sort method cannot be used with "
			    "stable and/or unique sort");
		};
	}

	if (bb_common_bss.sort_method == SORT_DEFAULT)
		bb_common_bss.sort_method = DEFAULT_SORT_ALGORITHM;

	if (debug_sort)
		printf("sort_method=%s\n",
		    get_sort_method_name(bb_common_bss.sort_method));

	switch (bb_common_bss.sort_method){
	case SORT_RADIXSORT:
		rxsort(list->list, list->count);
		break;
	case SORT_MERGESORT:
		mergesort(list->list, list->count,
		    sizeof(struct sort_list_item *), list_coll);
		break;
	case SORT_HEAPSORT:
		heapsort(list->list, list->count,
		    sizeof(struct sort_list_item *), list_coll);
		break;
	case SORT_QSORT:
		qsort(list->list, list->count,
		    sizeof(struct sort_list_item *), list_coll);
		break;
	default:
		DEFAULT_SORT_FUNC(list->list, list->count,
		    sizeof(struct sort_list_item *), list_coll);
		break;
	}
	sort_list_dump(list, outfile);
}

static int procfile(const char *fsrc, struct sort_list *list, struct file_list *fl)
{
	struct file_reader *fr;

	fr = file_reader_init(fsrc);

	/* file browse cycle */
	for (;;) {
		struct bwstring *bws;

		bws = file_reader_readline(fr);

		if (bws == NULL)
			break;

		sort_list_add(list, bws);

		if (list->memsize >= bb_common_bss.available_free_memory) {
			char *fn;

			fn = new_tmp_file_name();
			sort_list_to_file(list, fn);
			file_list_add(fl, fn, false);
			sort_list_clean(list);
		}
	}

	file_reader_free(fr);

	return 0;
}

/* Checks if the given file is sorted.  Stops at the first disorder,
 * prints the disordered line and returns 1. */
static int check(const char *fn)
{
	struct bwstring *s1, *s2;
	struct file_reader *fr;
	struct keys_array *ka1, *ka2;
	int res;
	size_t pos, posdisorder;

	s1 = s2 = NULL;
	ka1 = ka2 = NULL;

	fr = file_reader_init(fn);

	res = 0;
	pos = 1;
	posdisorder = 1;

	s1 = file_reader_readline(fr);
	if (s1 == NULL)
		goto end;

	ka1 = keys_array_alloc();
	preproc(s1, ka1);

	s2 = file_reader_readline(fr);
	if (s2 == NULL)
		goto end;

	ka2 = keys_array_alloc();
	preproc(s2, ka2);

	for (;;) {

		if (debug_sort) {
			bwsprintf(stdout, s2, "s1=<", ">");
			bwsprintf(stdout, s1, "s2=<", ">");
		}
		int cmp = key_coll(ka2, ka1, 0);
		if (debug_sort)
			printf("; cmp1=%d", cmp);

		if (!cmp && bb_common_bss.complex_sort &&
		    !(bb_common_bss.uflag) && !(bb_common_bss.sflag)) {
			cmp = top_level_str_coll(s2, s1);
			if (debug_sort)
				printf("; cmp2=%d", cmp);
		}
		if (debug_sort)
			printf("\n");

		if ((bb_common_bss.uflag && (cmp <= 0)) || (cmp < 0)) {
			if (!(bb_common_bss.csilentflag)) {
				bws_disorder_warnx(s2, fn, posdisorder);
				posdisorder = pos;
				if (debug_sort)
					bws_disorder_warnx(s1, fn, posdisorder);
			}
			res = 1;
			goto end;
		}

		pos++;

		clean_keys_array(s1, ka1);
		free(ka1);
		ka1 = ka2;
		ka2 = NULL;

		free(s1);
		s1 = s2;

		s2 = file_reader_readline(fr);
		if (s2 == NULL)
			goto end;

		ka2 = keys_array_alloc();
		preproc(s2, ka2);
	}

end:
	if (ka1) {
		clean_keys_array(s1, ka1);
		free(ka1);
	}

	free(s1);

	if (ka2) {
		clean_keys_array(s2, ka2);
		free(ka2);
	}

	free(s2);

	if (fn == NULL || *fn == 0 || strcmp(fn, "-") == 0) {
		for (;;) {
			s2 = file_reader_readline(fr);
			if (s2 == NULL)
				break;
			free(s2);
		}
	}

	file_reader_free(fr);

	return res;
}

/* Allocate and init file header structure */
static void file_header_init(struct file_header **fh, const char *fn, size_t file_pos)
{
	struct bwstring *line;

	*fh = xmalloc(sizeof(struct file_header));
	(*fh)->file_pos = file_pos;
	(*fh)->fr = file_reader_init(fn);
	line = file_reader_readline((*fh)->fr);
	if (line == NULL) {
		file_reader_free((*fh)->fr);
		(*fh)->fr = NULL;
		(*fh)->si = NULL;
	} else {
		(*fh)->si = sort_list_item_alloc();
		sort_list_item_set((*fh)->si, line);
	}
}

/* Read next line */
static void file_header_read_next(struct file_header *fh)
{
	struct bwstring *tmp;

	tmp = file_reader_readline(fh->fr);
	if (tmp == NULL) {
		file_reader_free(fh->fr);
		fh->fr = NULL;
		if (fh->si) {
			sort_list_item_clean(fh->si);
			free(fh->si);
			fh->si = NULL;
		}
	} else {
		if (fh->si == NULL)
			fh->si = sort_list_item_alloc();
		sort_list_item_set(fh->si, tmp);
	}
}

/* Compare a string and a sort list item, according to the sort specs. */
static int str_list_coll(struct bwstring *str1, struct sort_list_item **ss2)
{
	struct keys_array *ka1;
	int ret = 0;

	ka1 = keys_array_alloc();

	preproc(str1, ka1);

	sort_list_item_make_key(*ss2);

	if (debug_sort) {
		bwsprintf(stdout, str1, "; s1=<", ">");
		bwsprintf(stdout, (*ss2)->str, ", s2=<", ">");
	}

	ret = key_coll(ka1, &((*ss2)->ka), 0);

	if (debug_sort)
		printf("; cmp1=%d", ret);

	clean_keys_array(str1, ka1);
	free(ka1);

	if ((ret == 0) && !(bb_common_bss.sflag) && bb_common_bss.complex_sort) {
		ret = top_level_str_coll(str1, ((*ss2)->str));
		if (debug_sort)
			printf("; cmp2=%d", ret);
	}

	if (debug_sort)
		putchar('\n');

	return ret;
}

struct last_printed
{
	struct bwstring *str;
};

/* Prints the current line of the file */
static void file_header_print(struct file_header *fh, FILE *f_out, struct last_printed *lp)
{
	if (bb_common_bss.uflag) {
		if ((lp->str == NULL) || (str_list_coll(lp->str, &(fh->si)))) {
			bwsfwrite(fh->si->str, f_out, bb_common_bss.zflag);
			if (lp->str)
				free(lp->str);
			lp->str = bwsdup(fh->si->str);
		}
	} else
		bwsfwrite(fh->si->str, f_out, bb_common_bss.zflag);
}

/* Merge array of "files headers" */
static void file_headers_merge(size_t fnum, struct file_header **fh, FILE *f_out)
{
	struct last_printed lp;
	size_t i;

	memset(&lp, 0, sizeof(lp));

	/*
	 * construct the initial sort structure
	 */
	for (i = 0; i < fnum; i++)
		file_header_list_push(fh[i], fh, i);

	while (fh[0]->fr) { /* unfinished files are always in front */
		/* output the smallest line: */
		file_header_print(fh[0], f_out, &lp);
		/* read a new line, if possible: */
		file_header_read_next(fh[0]);
		/* re-arrange the list: */
		file_header_list_rearrange_from_header(fh, fnum);
	}

	if (lp.str)
		free(lp.str);
}

/* Merges the given files into the output file, which can be stdout. */
static void merge_files_array(size_t argc, char **argv, const char *fn_out)
{
	struct file_header **fh;
	FILE *f_out;
	size_t i;

	f_out = openfile(fn_out, "w");

	fh = sort_reallocarray(NULL, argc + 1, sizeof(struct file_header *));

	for (i = 0; i < argc; i++)
		file_header_init(fh + i, argv[i], i);

	file_headers_merge(argc, fh, f_out);

	for (i = 0; i < argc; i++)
		file_header_close(fh + i);

	free(fh);

	closefile(f_out, fn_out);
}

/* Shrinks the file list until its size smaller than max number of opened files */
static int shrink_file_list(struct file_list *fl)
{
	struct file_list new_fl;
	size_t indx = 0;

	if (fl->count < bb_common_bss.max_open_files)
		return 0;

	file_list_init(&new_fl, true);
	while (indx < fl->count) {
		char *fnew;
		size_t num;

		num = fl->count - indx;
		fnew = new_tmp_file_name();

		if (num >= bb_common_bss.max_open_files)
			num = bb_common_bss.max_open_files - 1;
		merge_files_array(num, fl->fns + indx, fnew);
		if (fl->tmp) {
			size_t i;

			for (i = 0; i < num; i++)
				unlink(fl->fns[indx + i]);
		}
		file_list_add(&new_fl, fnew, false);
		indx += num;
	}
	fl->tmp = false; /* already taken care of */
	file_list_clean(fl);

	fl->count = new_fl.count;
	fl->fns = new_fl.fns;
	fl->sz = new_fl.sz;
	fl->tmp = new_fl.tmp;

	return 1;
}

/* Merge list of files */
static void merge_files(struct file_list *fl, const char *fn_out)
{
	while (shrink_file_list(fl))
		;

	merge_files_array(fl->count, fl->fns, fn_out);
}

/* Main function. */
int sort_main(void)
{
	int argc = Gargc;
	char **argv = Gargv;
	char *outfile, *real_outfile, *sflag;
	int c;
	size_t i;
	struct sort_mods *sm = &default_sort_mods_object;
	bool mef_flags[NUMBER_OF_MUTUALLY_EXCLUSIVE_FLAGS] =
	    { false, false, false, false, false, false };

	bb_common_bss.argc_from_file0 = (size_t)-1;
	bb_common_bss.available_free_memory = 1000000;
	bb_common_bss.max_open_files = 16;
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
	sort_mb_cur_max = 1;
#endif
	symbol_decimal_point = Lc('.');
	symbol_negative_sign = Lc('-');
	symbol_positive_sign = Lc('+');

	bb_common_bss.tmpdir = "/var/tmp";
	bb_default_error_retval = 2;
	default_sort_mods_object.func = get_sort_func(&default_sort_mods_object);

	set_hw_params();

	outfile = "-";
	real_outfile = NULL;
	sflag = NULL;

	/* SA_RESTART */
	bb_signal(1 | (1<<SIGTERM) | (1<<SIGHUP) | (1<<SIGINT) | (1<<SIGUSR1) |
			(1<<SIGUSR2) | (1<<SIGPIPE) | (1<<SIGXCPU) |
			(1<<SIGXFSZ) , sig_handler);

	atexit(clear_tmp_files);

#ifdef CONFIG_LOCALE_SUPPORT
	set_locale();
#endif
	fix_obsolete_keys(&argc, argv);

	while (((c = getopt_long(argc, argv, OPTIONS, long_options, NULL)) != -1)) {

		check_mutually_exclusive_flags(c, mef_flags);

		if (!set_sort_modifier(sm, c)) {
			switch (c) {
			case 'c':
				bb_common_bss.cflag = true;
				if (optarg) {
					if (!strcmp(optarg, "diagnose-first"))
						;
					else if (!strcmp(optarg, "silent") ||
					    !strcmp(optarg, "quiet"))
						bb_common_bss.csilentflag = true;
					else if (*optarg)
						unknown(optarg);
				}
				break;
			case 'C':
				bb_common_bss.cflag = true;
				bb_common_bss.csilentflag = true;
				break;
			case 'k':
			{
				bb_common_bss.complex_sort = true;
				bb_common_bss.kflag = true;

				keys = sort_reallocarray(keys, keys_num + 1, sizeof(struct key_specs));
				memset(&(keys[keys_num]), 0, sizeof(struct key_specs));

				if (parse_k(optarg, &(keys[keys_num++])) < 0) {
					errno = EINVAL;
					bb_perror_msg_and_die("-k %s", optarg);
				}

				break;
			}
			case 'm':
				bb_common_bss.mflag = true;
				break;
			case 'o':
				outfile = optarg;
				break;
			case 's':
				bb_common_bss.sflag = true;
				break;
			case 'S':
				sflag = optarg;
				break;
			case 'T':
				bb_common_bss.tmpdir = optarg;
				break;
			case 't':
				if(optarg[0] == '\\') {
					const char *pe0 = optarg + 1;
					int e = bb_process_escape_sequence(&pe0);
					if(*pe0 != '\0') {
						errno = EINVAL;
						bb_perror_msg_and_die("-t %s", optarg);
					}
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
					bb_common_bss.field_sep = btowc(e);
					if (bb_common_bss.field_sep == WEOF) {
						errno = EINVAL;
						bb_perror_msg_and_die("-t %s", optarg);
					}
#else
					bb_common_bss.field_sep = e;
#endif
				} else {
					size_t n = mbchar_len(optarg);
					if(n > 1) {
						errno = EINVAL;
						bb_perror_msg_and_die("-t %s", optarg);
					}
#ifdef CONFIG_FEATURE_UTF8_SUPPORT
					 else {
						mbstate_t mbs;

						memset(&mbs, 0, sizeof(mbs));
						errno = 0;
						n = mbrtowc((WCHAR_T_UTF8 *)&bb_common_bss.field_sep, optarg, n, &mbs);
						if( n == (size_t)(-1) || n == (size_t)(-2)) {
							if(errno == 0)
								errno = EINVAL;
							bb_perror_msg_and_die("-t %s", optarg);
						}
					}
#else
					bb_common_bss.field_sep = optarg[0];
#endif
				}
				bb_common_bss.tflag = true;
				break;
			case 'u':
				bb_common_bss.uflag = true;
				/* stable sort for the correct unique val */
				bb_common_bss.sflag = true;
				break;
			case 'z':
				bb_common_bss.zflag = true;
				break;
			case SORT_OPT:
				if (!strcmp(optarg, "general-numeric"))
					set_sort_modifier(sm, 'g');
				else if (!strcmp(optarg, "human-numeric"))
					set_sort_modifier(sm, 'h');
				else if (!strcmp(optarg, "numeric"))
					set_sort_modifier(sm, 'n');
				else if (!strcmp(optarg, "month"))
					set_sort_modifier(sm, 'M');
				else if (!strcmp(optarg, "random"))
					set_sort_modifier(sm, 'R');
				else
					unknown(optarg);
				break;
			case QSORT_OPT:
				bb_common_bss.sort_method = SORT_QSORT;
				break;
			case 'H':
				bb_common_bss.sort_method = SORT_MERGESORT;
				break;
			case MMAP_OPT:
				bb_common_bss.use_mmap = true;
				break;
			case HEAPSORT_OPT:
				bb_common_bss.sort_method = SORT_HEAPSORT;
				break;
			case RADIXSORT_OPT:
				bb_common_bss.sort_method = SORT_RADIXSORT;
				break;
			case RANDOMSOURCE_OPT:
				bb_common_bss.random_source = optarg;
				break;
			case COMPRESSPROGRAM_OPT:
				bb_common_bss.compress_program = optarg;
				break;
			case FF_OPT:
				read_fns_from_file0(optarg);
				break;
			case BS_OPT:
				bb_common_bss.max_open_files = bb_xparse_ul_number(
					10 | BB_PN_SET_LOWER | BB_PN_SET_UPPER | BB_PN_ERR_STRV,
					  optarg, (unsigned long)2, (unsigned long)(UINT_MAX - 1),
					  "--batch-size argument is %s: %m", optarg) + 1;
				break;
			case VERSION_OPT:
				printf("%s\n", VERSION);
				return EXIT_SUCCESS;
#ifdef CONFIG_FEATURE_SORT_BIG
			case DEBUG_OPT:
				debug_sort = true;
				break;
#endif
			default:
				bb_show_usage();
				/* NOTREACHED */
			}
		}
	}
	argc -= optind;
	argv += optind;

	if (bb_common_bss.argv_from_file0) {
		argc = bb_common_bss.argc_from_file0;
		argv = bb_common_bss.argv_from_file0;
	}

	if (bb_common_bss.cflag) {
		if (argc > 1)
			bb_error_msg_and_die("only one input file is allowed with the -%c flag",
			    bb_common_bss.csilentflag ? 'C' : 'c');

	} else {
		/* Case when the outfile equals one of the input files: */
		if (strcmp(outfile, "-") != 0) {
			struct stat sb;
			int fd;

			for (i = 0; i < argc; ++i) {
				if (strcmp(argv[i], outfile) == 0) {
					if (stat(outfile, &sb) == -1)
						bb_perror_msg_and_die("%s", outfile);
					if (access(outfile, W_OK) == -1)
						bb_perror_msg_and_die("%s", outfile);
					real_outfile = outfile;
					outfile = xasprintf("%s.XXXXXXXXXX", real_outfile);
					if ((fd = mkstemp(outfile)) == -1)
						bb_perror_msg_and_die("%s", outfile);
					(void)fchown(fd, sb.st_uid, sb.st_gid);
#ifndef ACCESSPERMS
# define ACCESSPERMS (S_IRWXU|S_IRWXG|S_IRWXO) /* 0777 */
#endif
					if (fchmod(fd, sb.st_mode & ACCESSPERMS) == -1)
						bb_perror_msg_and_die("%s", outfile);
					close(fd);
					tmp_file_atexit(outfile);
					break;
				}
			}
		}

	}

	if (sflag != NULL)
		bb_common_bss.available_free_memory = parse_memory_buffer_value(sflag);

	if (keys_num == 0) {
		keys_num = 1;
		keys = zmalloc(sizeof(struct key_specs));
		keys[0].c1 = 1;
		keys[0].pos1b = sm->bflag;
		keys[0].pos2b = sm->bflag;
		memcpy(&(keys[0].sm), sm, sizeof(struct sort_mods));
	}

	for (i = 0; i < keys_num; i++) {
		struct key_specs *ks;

		ks = &(keys[i]);

		if (sort_modifier_empty(&(ks->sm))) {
			if (!(ks->pos1b) && !(ks->pos2b)) {
				ks->pos1b = sm->bflag;
				ks->pos2b = sm->bflag;
			}
			memcpy(&(ks->sm), sm, sizeof(struct sort_mods));
		}

		ks->sm.func = get_sort_func(&(ks->sm));
	}

	if (debug_sort) {
		printf("Memory to be used for sorting: %llu\n", bb_common_bss.available_free_memory);
		printf("Using collate rules of %s locale\n", setlocale(LC_COLLATE, NULL));
		if (byte_sort)
			printf("Byte sort is used\n");
		if (bb_common_bss.print_symbols_on_debug) {
			printf("Decimal Point: <%lc>\n", symbol_decimal_point);
			if (symbol_thousands_sep)
				printf("Thousands separator: <%lc>\n", symbol_thousands_sep);
			printf("Positive sign: <%lc>\n", symbol_positive_sign);
			printf("Negative sign: <%lc>\n", symbol_negative_sign);
		}
	}

	if (bb_common_bss.cflag)
		return check(argc ? *argv : "-");

	set_random_seed();

	if (!bb_common_bss.mflag) {
		struct file_list fl;
		struct sort_list list;

		sort_list_init(&list);
		file_list_init(&fl, true);

		if (argc < 1)
			procfile("-", &list, &fl);
		else {
			while (argc > 0) {
				procfile(*argv, &list, &fl);
				--argc;
				++argv;
			}
		}

		if (fl.count < 1)
			sort_list_to_file(&list, outfile);
		else {
			if (list.count > 0) {
				char *flast = new_tmp_file_name();

				sort_list_to_file(&list, flast);
				file_list_add(&fl, flast, false);
			}
			merge_files(&fl, outfile);
		}

		file_list_clean(&fl);

		/* We are about to exit the program, so we can ignore
		 * the clean-up for speed
		 *
		 * sort_list_clean(&list); */

	} else {
		struct file_list fl;

		file_list_init(&fl, false);
		file_list_populate(&fl, argc, argv, true);
		merge_files(&fl, outfile);
		file_list_clean(&fl);
	}

	if (real_outfile) {
		if (rename(outfile, real_outfile) < 0)
			bb_perror_msg_and_die("%s", real_outfile);
		free(outfile);
	}

	return 0;
}
#endif
