/*
 * sum -- checksum and count the blocks in a file
 *     Like BSD sum or SysV sum -r, except like SysV sum if -s option is given.
 *
 * Copyright (C) 86, 89, 91, 1995-2002, 2004 Free Software Foundation, Inc.
 * Copyright (C) 2005 by Erik Andersen <andersen@codepoet.org>
 * Copyright (C) 2005 by Mike Frysinger <vapier@gentoo.org>
 *
 * Written by Kayvan Aghaiepour and David MacKenzie
 * Taken from coreutils and turned into a busybox applet by Mike Frysinger
 *
 * Licensed under the GPL v2, see the file LICENSE in this tarball.
 */
#ifdef APPLET_USAGE
#define sum_trivial_usage \
	"[rs] [files...]"
#define sum_full_usage \
	"checksum and count the blocks in a file\n\n" \
	"Options:\n" \
	"\t-r\tuse BSD sum algorithm (1K blocks)\n" \
	"\t-s\tuse System V sum algorithm (512byte blocks)"
#else

/* zero arg to "-" arg, give the bsd func priority over sysv func */
#define COMPLEMENTARY "\0-0:s-r:r-s"
#include "opt_gen.h"
OPT_GEN(opt(s, SYSV_OPT), opt(r, BSD_OPT))
#include "opt_gen.h"

#include "busybox.h"
#include <unistd.h>
#include <fcntl.h>


#define MY_BUF_SIZE 8192

CALC_BSS struct BB_COMMON_BSS {
/* 1 if any of the files read were the standard input */
	int have_read_stdin;
	unsigned char buf[MY_BUF_SIZE];
} bb_common_bss;
#define have_read_stdin bb_common_bss.have_read_stdin
#define buf bb_common_bss.buf


/* make a little more readable and avoid using strcmp for just 2 bytes */
#define IS_STDIN(s) (s[0] == '-' && s[1] == '\0')

/* Calculate and print the rotated checksum and the size in 1K blocks
   of file FILE, or of the standard input if FILE is "-".
   If PRINT_NAME is >1, print FILE next to the checksum and size.
   The checksum varies depending on sizeof (int).
   Return 1 if successful.  */
static int bsd_sum_file(const char *file)
{
	int checksum = 0;          /* The checksum mod 2^16. */
	uintmax_t total_bytes = 0; /* The number of bytes. */
	int ch;                    /* Each character read. */
	FILE *fp = bb_xfopen(file, "W-r");

	if (fp == NULL)
		return 0;
	if (fp == stdin)
		have_read_stdin = 1;

	while ((ch = getc(fp)) != EOF) {
		++total_bytes;
		checksum = (checksum >> 1) + ((checksum & 1) << 15);
		checksum += ch;
		checksum &= 0xffff;             /* Keep it within bounds. */
	}

	if (ferror(fp)) {
		bb_perror_msg(file);
		bb_fclose_nonstdin(fp);
		return 0;
	}

	if (bb_fclose_nonstdin(fp) == EOF) {
		bb_perror_msg(file);
		return 0;
	}

	printf("%05d %5ju", checksum, (total_bytes+1023)/1024);
	return 1;
}

/* Calculate and print the checksum and the size in 512-byte blocks
   of file FILE, or of the standard input if FILE is "-".
   If PRINT_NAME is >0, print FILE next to the checksum and size.
   Return 1 if successful.  */
static int sysv_sum_file(const char *file)
{
	int fd;
	uintmax_t total_bytes = 0;

	/* The sum of all the input bytes, modulo (UINT_MAX + 1).  */
	unsigned int s = 0;

	if (IS_STDIN(file)) {
		fd = 0;
		have_read_stdin = 1;
	} else {
		fd = open(file, O_RDONLY);
		if (fd == -1)
			goto release_and_ret;
	}

	while (1) {
		size_t bytes_read = safe_read(fd, buf, MY_BUF_SIZE);

		if (bytes_read == 0)
			break;

		if (bytes_read == -1) {
release_and_ret:
			bb_perror_msg(file);
			if (!IS_STDIN(file))
				close(fd);
			return 0;
		}

		total_bytes += bytes_read;
		while (bytes_read--)
			s += buf[bytes_read];
	}

	if (!IS_STDIN(file) && close(fd) == -1)
		goto release_and_ret;
	else
		;

	{
		int r = (s & 0xffff) + ((s & 0xffffffff) >> 16);
		s = (r & 0xffff) + (r >> 16);
		printf("%d %ju", s, (total_bytes+511)/512);
	}
	return 1;
}

int sum_main(void)
{
	char **argv;
	int flags;
	int ok;
	int print_name;
	int (*sum_func)(const char *);

	flags = bb_xgetopt_ulflags(opt_str);

	if (flags == SYSV_OPT)
		sum_func = sysv_sum_file;
	else
		sum_func = bsd_sum_file;

	argv = Gargv;
	print_name = argv[1] != NULL;
	ok = EXIT_SUCCESS;
	for (; *argv; argv++) {
		flags = sum_func(*argv);
		if(flags) {
			if(print_name)
				printf(" %s", *argv);
			putchar('\n');
		} else {
			ok = EXIT_FAILURE;
		}
	}

	if (have_read_stdin && fclose(stdin) == EOF)
		bb_perror_msg_and_die("-");

	return ok;
}
#endif /* APPLET_USAGE */
