/*
 * tee implementation for busybox
 *
 * Copyright (C) 2003  Manuel Novoa III  <mjn3@codepoet.org>
 * modified by Vladimir Oleynik for can usage "-" in any position in files now
 *
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */

/* BB_AUDIT SUSv3 compliant */
/* http://www.opengroup.org/onlinepubs/007904975/utilities/tee.html */
#ifdef APPLET_USAGE
#define tee_trivial_usage \
	"[OPTION]... [FILE]..."
#define tee_full_usage \
	"Copy standard input to each FILE, and also to standard output.\n\n" \
	"Options:\n" \
	"\t-a\tappend to the given FILEs, do not overwrite\n" \
	"\t-i\tignore interrupt signals (SIGINT)"
#else

#include "busybox.h"
#include <unistd.h>
#include <string.h>
#include "bb_sig.h"

#ifdef CONFIG_FEATURE_TEE_USE_BLOCK_IO
CALC_BSS struct BB_COMMON_BSS {
	char buf[BUFSIZ];
} bb_common_bss;
#define buf bb_common_bss.buf
#endif


int tee_main(void)
{
	const char *mode = "w\0a";
	FILE **files;
	FILE **p;
	char **filenames;
	int i;
	int retval = EXIT_SUCCESS;
#ifdef CONFIG_FEATURE_TEE_USE_BLOCK_IO
	ssize_t c;
#else
	int c;
#endif
	/* zero args is "-" arg */
	i = bb_xgetopt_ulflags("ia\0-0");    /* 'a' must be 2nd */

	mode += (i & 2);        /* Since 'a' is the 2nd option... */

	if (i & 1)
		bb_signal((1 << SIGINT), SIG_IGN);

	/* gnu tee ignores SIGPIPE in case one of the output files is a pipe
	 * that doesn't consume all its input.  Good idea... */
	bb_signal((1 << SIGPIPE), SIG_IGN);

	/* Allocate an array of FILE *'s, with one extra for a sentinal. */
	files = (FILE **)zmalloc(sizeof(FILE *) * (Gargc + 1));
	for(i = 0, filenames = Gargv; filenames[i]; i++) {
		const char *filename = filenames[i];
		FILE *f;
		int n;

		/* skip duplicate names, find the really count */
		for(n = 0; files[n] != NULL; n++) {
			if(strcmp(filenames[n], filename) == 0)
				break;
		}
		if(files[n] != NULL)
			continue;

		if (filename[0] == '-' && filename[1] == '\0') {
			f = stdout;
			/* but check ferror stdin too, see later */
			filename = bb_msg_standard_input;
		} else if ((f = fopen(filename, mode)) == NULL) {
			bb_perror_msg("%s", filename);
			retval = EXIT_FAILURE;
			continue;
		}
		filenames[n] = (char *)filename;
		files[n] = f;
		setbuf(f, NULL);       /* tee must not buffer output. */
	}

#ifdef CONFIG_FEATURE_TEE_USE_BLOCK_IO
	while ((c = safe_read(STDIN_FILENO, buf, BUFSIZ)) > 0) {
		for (p = files; *p; p++) {
			fwrite(buf, 1, c, *p);
		}
	}

	if (c < 0) {                    /* Make sure read errors are signaled. */
		retval = EXIT_FAILURE;
	}

#else
	setvbuf(stdout, NULL, _IONBF, 0);
	while ((c = getchar()) != EOF) {
		for (p = files; *p; p++) {
			putc(c, *p);
		}
	}
#endif

	/* Now we need to check for i/o errors on stdin and the various
	 * output files.  Since we know that the stdout entry in the output
	 * file table we can save one "if ferror" test by
	 * setting the stdout entry to stdin and checking stdout error
	 * status with bb_fflush_stdout_and_exit()... although fflush()ing
	 * is unnecessary here. */

	p = files;
	do {            /* Now check for (input and) output errors. */
		/* Checking ferror should be sufficient, but we may want to fclose.
		 * If we do, remember not to close stdin! */
		if(*p == stdout)
			*p = stdin;
		bb_xferror(*p, filenames[(int)(p - files)]);
	} while (*++p);

	bb_fflush_stdout_and_exit(retval);
}
#endif /* APPLET_USAGE */
