/*
 * test implementation for busybox
 *
 * Copyright (c) by a whole pile of folks:
 *
 *      test(1); version 7-like  --  author Erik Baalbergen
 *      modified by Eric Gisin to be used as built-in.
 *      modified by Arnold Robbins to add SVR3 compatibility
 *      (-x -c -b -p -u -g -k) plus Korn's -L -nt -ot -ef and new -S (socket).
 *      modified by J.T. Conklin for NetBSD.
 *      modified by Herbert Xu to be used as built-in in ash.
 *      modified by Erik Andersen <andersen@codepoet.org> to be used
 *      in busybox.
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 *
 * Original copyright notice states:
 *      "This program is in the Public Domain."
 */
#ifdef APPLET_USAGE
#define test_trivial_usage \
	"EXPRESSION\n  or  [ EXPRESSION ]"
#define test_full_usage \
	"Checks file types and compares values returning an exit\n" \
	"code determined by the value of EXPRESSION."
#else

#include "busybox.h"
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <setjmp.h>

/* test(1) accepts the following grammar:
	oexpr   ::= aexpr | aexpr "-o" oexpr ;
	aexpr   ::= nexpr | nexpr "-a" aexpr ;
	nexpr   ::= primary | "!" primary
	primary ::= unary-operator operand
		| operand binary-operator operand
		| operand
		| "(" oexpr ")"
		;
	unary-operator ::= "-r"|"-w"|"-x"|"-f"|"-d"|"-c"|"-b"|"-p"|
		"-u"|"-g"|"-k"|"-s"|"-t"|"-z"|"-n"|"-o"|"-O"|"-G"|"-L"|"-S";

	binary-operator ::= "="|"=="|"!="|"-eq"|"-ne"|"-ge"|"-gt"|"-le"|"-lt"|
			"-nt"|"-ot"|"-ef";
	operand ::= <any legal UNIX file name>
*/

enum token {
	EOI,
	FILRD,
	FILWR,
	FILEX,
	FILEXIST,
	FILREG,
	FILDIR,
	FILCDEV,
	FILBDEV,
	FILFIFO,
	FILSOCK,
	FILSYM,
	FILGZ,
	FILTT,
	FILSUID,
	FILSGID,
	FILSTCK,
	FILNT,
	FILOT,
	FILEQ,
	FILUID,
	FILGID,
	STREZ,
	STRNZ,
	STREQ,
	STRNE,
	STRLT,
	STRGT,
	INTEQ,
	INTNE,
	INTGE,
	INTGT,
	INTLE,
	INTLT,
	UNOT,
	BAND,
	BOR,
	LPAREN,
	RPAREN,
	OPERAND
};

enum token_types {
	UNOP,
	BINOP,
	BUNOP,
	BBINOP,
	PAREN
};

static const struct t_op {
	const char *op_text;
	short op_num, op_type;
} ops[] = {
	{
	"-r", FILRD, UNOP}, {
	"-w", FILWR, UNOP}, {
	"-x", FILEX, UNOP}, {
	"-e", FILEXIST, UNOP}, {
	"-f", FILREG, UNOP}, {
	"-d", FILDIR, UNOP}, {
	"-c", FILCDEV, UNOP}, {
	"-b", FILBDEV, UNOP}, {
	"-p", FILFIFO, UNOP}, {
	"-u", FILSUID, UNOP}, {
	"-g", FILSGID, UNOP}, {
	"-k", FILSTCK, UNOP}, {
	"-s", FILGZ, UNOP}, {
	"-t", FILTT, UNOP}, {
	"-z", STREZ, UNOP}, {
	"-n", STRNZ, UNOP}, {
	"-h", FILSYM, UNOP},    /* for backwards compat */
	{
	"-O", FILUID, UNOP}, {
	"-G", FILGID, UNOP}, {
	"-L", FILSYM, UNOP}, {
	"-S", FILSOCK, UNOP}, {
	"=", STREQ, BINOP}, {
	"==", STREQ, BINOP}, {
	"!=", STRNE, BINOP}, {
	"<", STRLT, BINOP}, {
	">", STRGT, BINOP}, {
	"-eq", INTEQ, BINOP}, {
	"-ne", INTNE, BINOP}, {
	"-ge", INTGE, BINOP}, {
	"-gt", INTGT, BINOP}, {
	"-le", INTLE, BINOP}, {
	"-lt", INTLT, BINOP}, {
	"-nt", FILNT, BINOP}, {
	"-ot", FILOT, BINOP}, {
	"-ef", FILEQ, BINOP}, {
	"!", UNOT, BUNOP}, {
	"-a", BAND, BBINOP}, {
	"-o", BOR, BBINOP}, {
	"(", LPAREN, PAREN}, {
	")", RPAREN, PAREN}, {
	0, OPERAND, 0}
};

#ifdef CONFIG_FEATURE_TEST_64
typedef int64_t arith_t;
#else
typedef int arith_t;
#endif

#ifndef CONFIG_ASH_BUILTIN_TEST
CALC_BSS struct BB_COMMON_BSS
{
	char **t_wp;
	struct t_op const *t_wp_op;
	gid_t *group_array;
	int ngroups;
	jmp_buf jmp_env;
} bb_common_bss;
#define BSS_VAR
#define TA0 void
#define TA
#define T_A
#define t_wp bb_common_bss.t_wp
#define t_wp_op bb_common_bss.t_wp_op
#define group_array bb_common_bss.group_array
#define ngroups bb_common_bss.ngroups
#define jmp_env bb_common_bss.jmp_env
#else
struct BSS
{
	char **t_wp;
	struct t_op const *t_wp_op;
	gid_t *group_array;
	int ngroups;
	jmp_buf jmp_env;
};
#define BSS_VAR test_bss
#define TA0 struct BSS *BSS_VAR
#define TA , TA0
#define T_A , BSS_VAR
#define t_wp (BSS_VAR->t_wp)
#define t_wp_op (BSS_VAR->t_wp_op)
#define group_array (BSS_VAR->group_array)
#define ngroups (BSS_VAR->ngroups)
#define jmp_env (BSS_VAR->jmp_env)
#endif


static void syntax(const char *op, const char *msg TA)
{
	if (op && *op) {
		bb_error_msg("%s: %s", op, msg);
	} else {
		bb_error_msg("%s", msg);
	}
	longjmp(jmp_env, 1);
}

static enum token t_lex(char *s TA)
{
	struct t_op const *op = ops;

	if (s == 0) {
		t_wp_op = (struct t_op *) 0;
		return EOI;
	}
	op = find_string_from_struct(ops, sizeof(struct t_op), s);
	t_wp_op = op->op_text != NULL ? op : NULL;
	return op->op_num; /* not found - OPERAND */
}

static arith_t nexpr(enum token n TA);

static arith_t aexpr(enum token n TA)
{
	arith_t res;

	res = nexpr(n T_A);
	if (t_lex(*++t_wp T_A) == BAND)
		return aexpr(t_lex(*++t_wp T_A) T_A) && res;
	t_wp--;
	return res;
}

static arith_t oexpr(enum token n TA)
{
	arith_t res;

	res = aexpr(n T_A);
	if (t_lex(*++t_wp T_A) == BOR) {
		return oexpr(t_lex(*++t_wp T_A) T_A) || res;
	}
	t_wp--;
	return res;
}

/* atoi with error detection */
static arith_t getn(const char *s TA)
{
	char *p;
#ifdef CONFIG_FEATURE_TEST_64
	bb64_t r;
#else
	long r;
#endif

	errno = 0;
#ifdef CONFIG_FEATURE_TEST_64
	r = BB_strtoll(s, &p, 10);
#else
	r = strtol(s, &p, 10);
#endif

	if (errno != 0) {
		bb_error_msg("%s: out of range", s);
		longjmp(jmp_env, 1);
	}

	if (*(bb_skip_whitespace(p))) {
		bb_error_msg("%s: bad number", s);
		longjmp(jmp_env, 1);
	}

	return r;
}

static int newerf(const char *f1, const char *f2)
{
	struct stat b1, b2;

	return (stat(f1, &b1) == 0 &&
			stat(f2, &b2) == 0 && b1.st_mtime > b2.st_mtime);
}

static int olderf(const char *f1, const char *f2)
{
	struct stat b1, b2;

	return (stat(f1, &b1) == 0 &&
			stat(f2, &b2) == 0 && b1.st_mtime < b2.st_mtime);
}

static int equalf(const char *f1, const char *f2)
{
	struct stat b1, b2;

	return (stat(f1, &b1) == 0 &&
			stat(f2, &b2) == 0 &&
			b1.st_dev == b2.st_dev && b1.st_ino == b2.st_ino);
}

static int binop(TA0)
{
	const char *opnd1, *opnd2;
	struct t_op const *op;

	opnd1 = *t_wp;
	(void) t_lex(*++t_wp T_A);
	op = t_wp_op;

	if ((opnd2 = *++t_wp) == (char *) 0)
		syntax(op->op_text, "argument expected" T_A);

	switch (op->op_num) {
	case STREQ:
		return strcmp(opnd1, opnd2) == 0;
	case STRNE:
		return strcmp(opnd1, opnd2) != 0;
	case STRLT:
		return strcmp(opnd1, opnd2) < 0;
	case STRGT:
		return strcmp(opnd1, opnd2) > 0;
	case INTEQ:
		return getn(opnd1 T_A) == getn(opnd2 T_A);
	case INTNE:
		return getn(opnd1 T_A) != getn(opnd2 T_A);
	case INTGE:
		return getn(opnd1 T_A) >= getn(opnd2 T_A);
	case INTGT:
		return getn(opnd1 T_A) > getn(opnd2 T_A);
	case INTLE:
		return getn(opnd1 T_A) <= getn(opnd2 T_A);
	case INTLT:
		return getn(opnd1 T_A) < getn(opnd2 T_A);
	case FILNT:
		return newerf(opnd1, opnd2);
	case FILOT:
		return olderf(opnd1, opnd2);
	case FILEQ:
		return equalf(opnd1, opnd2);
	}
	/* NOTREACHED */
	return 1;
}

/* Return non-zero if GID is one that we have in our groups list. */
static int is_a_group_member(gid_t gid TA)
{
	int i;

	/* Short-circuit if possible, maybe saving a call to getgroups(). */
	if (gid == getgid() || gid == getegid())
		return 1;

	if (ngroups == 0) {
		ngroups = getgroups(0, NULL);
		if(ngroups > 0) {
			group_array = xmalloc(ngroups * sizeof(gid_t));
			getgroups(ngroups, group_array);
		}
	}

	/* Search through the list looking for GID. */
	for (i = 0; i < ngroups; i++)
		if (gid == group_array[i])
			return 1;

	return 0;
}

/* Do the same thing access(2) does, but use the effective uid and gid,
   and don't make the mistake of telling root that any file is
   executable. */
static int test_eaccess(char *path, int mode TA)
{
	struct stat st;
	unsigned int euid = geteuid();

	if (stat(path, &st) < 0)
		return -1;

	if (euid == 0) {
		/* Root can read or write any file. */
		if (mode != X_OK)
			return 0;

		/* Root can execute any file that has any one of the execute
		   bits set. */
		if (st.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH))
			return 0;
	}

	if (st.st_uid == euid)  /* owner */
		mode <<= 6;
	else if (is_a_group_member(st.st_gid T_A))
		mode <<= 3;

	if (st.st_mode & mode)
		return 0;

	return -1;
}

static int filstat(char *nm, enum token mode TA)
{
	struct stat s;
	unsigned int i;

	if (mode == FILSYM) {
#ifdef S_IFLNK
		if (lstat(nm, &s) == 0) {
			i = S_IFLNK;
			goto filetype;
		}
#endif
		return 0;
	}

	if (stat(nm, &s) != 0)
		return 0;

	switch (mode) {
	case FILRD:
		return test_eaccess(nm, R_OK T_A) == 0;
	case FILWR:
		return test_eaccess(nm, W_OK T_A) == 0;
	case FILEX:
		return test_eaccess(nm, X_OK T_A) == 0;
	case FILEXIST:
		return 1;
	case FILREG:
		i = S_IFREG;
		goto filetype;
	case FILDIR:
		i = S_IFDIR;
		goto filetype;
	case FILCDEV:
		i = S_IFCHR;
		goto filetype;
	case FILBDEV:
		i = S_IFBLK;
		goto filetype;
	case FILFIFO:
#ifdef S_IFIFO
		i = S_IFIFO;
		goto filetype;
#else
		return 0;
#endif
	case FILSOCK:
#ifdef S_IFSOCK
		i = S_IFSOCK;
		goto filetype;
#else
		return 0;
#endif
	case FILSUID:
		i = S_ISUID;
		goto filebit;
	case FILSGID:
		i = S_ISGID;
		goto filebit;
	case FILSTCK:
		i = S_ISVTX;
		goto filebit;
	case FILGZ:
		return s.st_size > 0L;
	case FILUID:
		return s.st_uid == geteuid();
	case FILGID:
		return s.st_gid == getegid();
	default:
		return 1;
	}

  filetype:
	return ((s.st_mode & S_IFMT) == i);

  filebit:
	return ((s.st_mode & i) != 0);
}

static arith_t primary(enum token n TA)
{
	arith_t res;

	if (n == EOI)
		syntax(NULL, "argument expected" T_A);

	if (n == LPAREN) {
		res = oexpr(t_lex(*++t_wp T_A) T_A);
		if (t_lex(*++t_wp T_A) != RPAREN)
			syntax(NULL, "closing paren expected" T_A);
		return res;
	}
	if (t_wp_op && t_wp_op->op_type == UNOP) {
		/* unary expression */
		if (*++t_wp == NULL)
			syntax(t_wp_op->op_text, "argument expected" T_A);
		switch (n) {
		case STREZ:
			return strlen(*t_wp) == 0;
		case STRNZ:
			return strlen(*t_wp) != 0;
		case FILTT:
			return isatty(getn(*t_wp T_A));
		default:
			return filstat(*t_wp, n T_A);
		}
	}

	if (t_lex(t_wp[1] T_A), t_wp_op && t_wp_op->op_type == BINOP)
		return binop(BSS_VAR);

	return strlen(*t_wp) > 0;
}

static arith_t nexpr(enum token n TA)
{
	if (n == UNOT)
		return !nexpr(t_lex(*++t_wp T_A) T_A);
	return primary(n T_A);
}


static int bb_test_int(const char *applet_name, int argc, char **argv TA)
{
	int res;

	if(applet_name[0] == '[') {
		char *e = applet_name[1] == '\0' ? "]" : "]]";
		if (strcmp(argv[--argc], e)) {
			bb_error_msg("missing %s", e);
			return 2;
		}
		argv[argc] = NULL;
	}

	if (setjmp(jmp_env))
		return 2;
	/* Implement special cases from POSIX.2, section 4.62.4 */
	switch (argc) {
	case 1:
		return 1;
	case 2:
		return *argv[1] == '\0';
	case 3:
		if (argv[1][0] == '!' && argv[1][1] == '\0')
			return !(*argv[2] == '\0');
		break;
	case 4:
		if (argv[1][0] != '!' || argv[1][1] != '\0') {
			if (t_lex(argv[2] T_A), t_wp_op && t_wp_op->op_type == BINOP) {
				t_wp = &argv[1];
				return binop(BSS_VAR) == 0;
			}
		}
		break;
	case 5:
		if (argv[1][0] == '!' && argv[1][1] == '\0') {
			if (t_lex(argv[3] T_A), t_wp_op && t_wp_op->op_type == BINOP) {
				t_wp = &argv[2];
				return !(binop(BSS_VAR) == 0);
			}
		}
		break;
	}

	t_wp = &argv[1];
	res = !oexpr(t_lex(*t_wp T_A) T_A);

	if (*t_wp != NULL && *++t_wp != NULL)
		syntax(*t_wp, "unknown operand" T_A);

	return res;
}

#ifdef CONFIG_ASH_BUILTIN_TEST
extern int bb_test(const char *applet_name, int argc, char **argv)
{
	int r;
	char **av = argv;
	TA0 = alloca(sizeof(struct BSS));

	memset(BSS_VAR, 0, sizeof(struct BSS));
	if(applet_name != bb_applet_name && applet_name[0] == '[') {
		/* protect memory leak from argv["]"]=NULL */
		av = alloca(argc * sizeof(char *));
		for(r = 0; r <= argc; r++)
			av[r] = argv[r];
	}
	r = bb_test_int(applet_name, argc, av, BSS_VAR);
	free(group_array);
	return r;
}
#endif

extern int test_main(void)
{
#ifdef CONFIG_ASH_BUILTIN_TEST
	return bb_test(bb_applet_name, Gargc, Gargv);
#else
	return bb_test_int(bb_applet_name, Gargc, Gargv);
#endif
}
#endif /* APPLET_USAGE */
