/* timeout -- run a command with bounded time
   Copyright (C) 2008-2022 Free Software Foundation, Inc.
   Written by Pádraig Brady.  */

#ifdef APPLET_USAGE
#define timeout_trivial_usage \
	"[-k kill-after-duration] [-s signal] [-v] DURATION COMMAND [ARG]..."
#define timeout_full_usage \
	"Run a command with bounded time.\n" \
	"Options:\n" \
	USE_FEATURE_LONG_OPTIONS( \
	"\t--preserve-status\n" \
	"\t\texit with the same status as COMMAND, even when the\n" \
	"\t\tcommand times out\n" \
	"\t--foreground\n" \
	"\t\twhen not running timeout directly from a shell prompt,\n" \
	"\t\t  allow COMMAND to read from the TTY and get TTY signals\n" ) \
	"\t-k" USE_FEATURE_LONG_OPTIONS(", --kill-after=") SKIP_FEATURE_LONG_OPTIONS(" ") "DURATION\n" \
	"\t\talso send a KILL signal if COMMAND is still running\n" \
	"\t\t  this long after the initial signal was sent\n" \
	"\t-s" USE_FEATURE_LONG_OPTIONS(", --signal=") SKIP_FEATURE_LONG_OPTIONS(" ") "SIGNAL\n" \
	"\t\tspecify the signal to be sent on timeout\n" \
	"\t-v" USE_FEATURE_LONG_OPTIONS(", --verbose") "\t\tdiagnose to stderr any signal sent upon timeout"
#else
#include "busybox.h"
#include "bb_sig.h"
#include <sys/wait.h>
#include <sys/resource.h>
#include <unistd.h>

CALC_BSS struct BB_COMMON_BSS {
	int timed_out;
	int term_signal;         /* same default as kill command.  */
	pid_t monitored_pid;
	double kill_after;
	int foreground;          /* whether to use another program group.  */
	int preserve_status;     /* whether to use a timeout status or not.  */
	int verbose;             /* whether to diagnose timeouts or not.  */
	char const *command;
} bb_common_bss;
#define timed_out bb_common_bss.timed_out
#define term_signal bb_common_bss.term_signal
#define monitored_pid bb_common_bss.monitored_pid
#define kill_after bb_common_bss.kill_after
#define foreground bb_common_bss.foreground
#define preserve_status bb_common_bss.preserve_status
#define verbose bb_common_bss.verbose
#define command bb_common_bss.command


#if ENABLE_FEATURE_LONG_OPTIONS
static struct option const timeout_long_options[] =
{
  {"kill-after", required_argument, NULL, 'k'},
  {"signal", required_argument, NULL, 's'},
  {"verbose", no_argument, NULL, 'v'},
  {"foreground", no_argument, NULL, 'f'},
  {"preserve-status", no_argument, NULL, 'p'},
  {NULL, 0, NULL, 0}
};
#endif

/* Start the timeout after which we'll receive a SIGALRM.
   Round DURATION up to the next representable value.
   Treat out-of-range values as if they were maximal,
   as that's more useful in practice than reporting an error.
   '0' means don't timeout.  */
static void
settimeout (double duration)
{
  unsigned int timeint;
  if (UINT_MAX <= duration)
    timeint = UINT_MAX;
  else {
      unsigned int duration_floor = duration;
      timeint = duration_floor + (duration_floor < duration);
  }
  alarm (timeint);
}

/* send SIG avoiding the current process.  */

static int
send_sig (pid_t where, int sig)
{
  /* If sending to the group, then ignore the signal,
     so we don't go into a signal loop.  Note that this will ignore any of the
     signals registered in "Setup handlers", that are sent after we
     propagate the first one, which hopefully won't be an issue.  Note this
     process can be implicitly multithreaded due to some timer_settime()
     implementations, therefore a signal sent to the group, can be sent
     multiple times to this process.  */
  if (where == 0)
    signal (sig, SIG_IGN);
  return kill (where, sig);
}

/* Signal handler which is required for sigsuspend() to be interrupted
   whenever SIGCHLD is received.  */
static void
chld (int sig)
{
}

static void
cleanup (int sig)
{
  if (sig == SIGALRM) {
      timed_out = 1;
      sig = term_signal;
  }
  if (monitored_pid) {
      if (kill_after) {
	  int saved_errno = errno; /* settimeout may reset.  */
	  /* Start a new timeout after which we'll send SIGKILL.  */
	  term_signal = SIGKILL;
	  settimeout (kill_after);
	  kill_after = 0; /* Don't let later signals reset kill alarm.  */
	  errno = saved_errno;
      }

      /* Send the signal directly to the monitored child,
	 in case it has itself become group leader,
	 or is not running in a separate group.  */
      if (verbose) {
	  char signame[16];
	  if (u_signal_num2name (sig, 1, signame) == NULL)
	    snprintf (signame, sizeof signame, "%d", sig);
	  bb_error_msg ("sending signal %s to command %s", signame, command);
      }
      send_sig (monitored_pid, sig);

      /* The normal case is the job has remained in our
	 newly created process group, so send to all processes in that.  */
      if (!foreground) {
	  send_sig (0, sig);
	  if (sig != SIGKILL && sig != SIGCONT) {
	      send_sig (monitored_pid, SIGCONT);
	      send_sig (0, SIGCONT);
	  }
      }
  } else /* we're the child or the child is not exec'd yet.  */
    _exit (128 + sig);
}

/* Given a floating point value *X, and a suffix character, SUFFIX_CHAR,
   scale *X by the multiplier implied by SUFFIX_CHAR.  SUFFIX_CHAR may
   be the NUL byte or 's' to denote seconds, 'm' for minutes, 'h' for
   hours, or 'd' for days.  If SUFFIX_CHAR is invalid, don't modify *X
   and return FALSE.  Otherwise return TRUE.  */

static int
apply_time_suffix (double *x, char suffix_char)
{
  int multiplier;

  switch (suffix_char) {
    case 0:
    case 's':
      multiplier = 1;
      break;
    case 'm':
      multiplier = 60;
      break;
    case 'h':
      multiplier = 60 * 60;
      break;
    case 'd':
      multiplier = 60 * 60 * 24;
      break;
    default:
      return FALSE;
    }

  *x *= multiplier;

  return TRUE;
}

#include <math.h>

static double
parse_duration (char const *str)
{
  double duration;
  char *ep;

  errno = 0;
  duration = strtod(str, &ep);
  if (errno != 0 || ep==str || !finite(duration)
      /* Nonnegative interval.  */
      || ! (0 <= duration)
      /* No extra chars after the number and an optional s,m,h,d char.  */
      || (*ep && *(ep + 1))
      /* Check any suffix char and update timeout based on the suffix.  */
      || !apply_time_suffix (&duration, *ep))
    {
      bb_default_error_retval = 125;
      bb_error_msg_and_die ("invalid time interval %s", str);
    }

  return duration;
}

static void
unblock_signal (int sig)
{
  sigset_t unblock_set;
  sigemptyset (&unblock_set);
  sigaddset (&unblock_set, sig);
  sigprocmask (SIG_UNBLOCK, &unblock_set, NULL);
}

static void
install_sigchld (void)
{
  bb_signal(1 | (1 << SIGCHLD), chld);
  /* We inherit the signal mask from our parent process,
     so ensure SIGCHLD is not blocked. */
  unblock_signal (SIGCHLD);
}

/* Block all signals which were registered with cleanup() as the signal
   handler, so we never kill processes after waitpid() returns.
   Also block SIGCHLD to ensure it doesn't fire between
   waitpid() polling and sigsuspend() waiting for a signal.
   Return original mask in OLD_SET.  */
static void
block_cleanup_and_chld (int sigterm, sigset_t *old_set)
{
  sigset_t block_set = bb_sigmask((1 << SIGALRM) | (1 << SIGINT) | (1 << SIGQUIT) | (1 << SIGHUP) | (1 << SIGTERM) |
				(1 << sigterm) | (1 << SIGCHLD));
  sigprocmask (SIG_BLOCK, &block_set, old_set);
}

/* Try to disable core dumps for this process.
   Return TRUE if successful, FALSE otherwise.  */
static int
disable_core_dumps (void)
{
  /* Note this doesn't disable processing by a filter in
     /proc/sys/kernel/core_pattern on Linux.  */
  if (setrlimit (RLIMIT_CORE, &(struct rlimit) {0,0}) == 0)
    return TRUE;
  bb_perror_msg ("warning: disabling core dumps failed");
  return FALSE;
}

int
timeout_main (void)
{
  double timeout;
  char *ska = NULL, *signame = NULL;
  int c;

#if ENABLE_FEATURE_LONG_OPTIONS
  bb_applet_long_options = timeout_long_options;
#endif

  term_signal = SIGTERM;
  c = bb_xgetopt_ulflags("+vfpk:s:\0-2", &ska, &signame);
  verbose = c & 1;
  foreground = c & 2;
  preserve_status = c & 4;
  if (ska)
	kill_after = parse_duration (ska);
  if (signame && u_signal_name2num (signame, &term_signal, 1, NULL) == NULL)
	    bb_error_msg_and_die("invalid signal name `%s'", signame);

  timeout = parse_duration (*Gargv);
  Gargv++;
  command = Gargv[0];

  /* Ensure we're in our own group so all subprocesses can be killed.
     Note we don't just put the child in a separate group as
     then we would need to worry about foreground and background groups
     and propagating signals between them.  */
  if (!foreground)
    setpgid (0, 0);

  /* Setup handlers before fork() so that we
     handle any signals caused by child, without races.  */
  bb_signal(1 | (1 << SIGALRM) | (1 << SIGINT) | (1 << SIGQUIT) |
		(1 << SIGHUP) | (1 << SIGTERM) | (1 << term_signal), cleanup);
  /* Don't stop if background child needs tty.  */
  bb_signal((1 << SIGTTIN) | (1 << SIGTTOU), SIG_IGN);
  signal (SIGTTIN, SIG_IGN);   /* Don't stop if background child needs tty.  */
  install_sigchld ();          /* Interrupt sigsuspend() when child exits.   */

  monitored_pid = fork ();
  if (monitored_pid == -1) {
      bb_default_error_retval = 125;
      bb_perror_msg_and_die ("fork system call failed");
  } else if (monitored_pid == 0) {
	/* child */
	/* exec doesn't reset SIG_IGN -> SIG_DFL.  */
	signal (SIGTTIN, SIG_DFL);
	signal (SIGTTOU, SIG_DFL);

	execvp (Gargv[0], Gargv);   /* FIXME: should we use "sh -c" ... here?  */

	/* exit like sh, env, nohup, ...  */
	/* 127 - Could not find program to exec.  */
	/* 126 - Program located, but not usable.  */
	int exit_status = errno == ENOENT ? 127 : 126;
	bb_perror_msg ("failed to run command %s", command);
	return exit_status;
  } else {
      pid_t wait_result;
      int status;

      /* We configure timers so that SIGALRM is sent on expiry.
	 Therefore ensure we don't inherit a mask blocking SIGALRM.  */
      unblock_signal (SIGALRM);

      settimeout (timeout);

      /* Ensure we don't cleanup() after waitpid() reaps the child,
	 to avoid sending signals to a possibly different process.  */
      sigset_t cleanup_set;
      block_cleanup_and_chld (term_signal, &cleanup_set);

      while ((wait_result = waitpid (monitored_pid, &status, WNOHANG)) == 0)
	sigsuspend (&cleanup_set);  /* Wait with cleanup signals unblocked.  */

      if (wait_result < 0) {
	  /* shouldn't happen.  */
	  bb_perror_msg ("error waiting for command");
	  status = 125;
      } else {
	  if (WIFEXITED (status))
	    status = WEXITSTATUS (status);
	  else if (WIFSIGNALED (status)) {
	      int sig = WTERMSIG (status);
	      if (WCOREDUMP (status))
		bb_error_msg ("the monitored command dumped core");
	      if (!timed_out && disable_core_dumps ()) {
		  /* exit with the signal flag set.  */
		  signal (sig, SIG_DFL);
		  unblock_signal (sig);
		  raise (sig);
	      }
	      /* Allow users to distinguish if command was forcably killed.
		 Needed with --foreground where we don't send SIGKILL to
		 the timeout process itself.  */
	      if (timed_out && sig == SIGKILL)
		preserve_status = TRUE;
	      status = sig + 128; /* what sh returns for signaled processes.  */
	  } else {
	      /* shouldn't happen.  */
	      bb_error_msg ("unknown status from command (%d)", status);
	      status = EXIT_FAILURE;
	  }
      }

      if (timed_out && !preserve_status)
	status = 124; /* Time expired before child completed.  */
      return status;
  }
}
#endif
