/*
 * Mini tr implementation for busybox
 *
 * Copyright (c) 1987,1997, Prentice Hall
 * All rights reserved.
 *
 * The name of Prentice Hall may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * Copyright (c) Michiel Huisjes
 *
 * This version of tr is adapted from Minix tr and was modified
 * by Erik Andersen <andersen@codepoet.org> to be used in busybox.
 *
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */
#ifdef APPLET_USAGE
#define tr_trivial_usage \
	"[-cds] STRING1 [STRING2]"
#define tr_full_usage \
	"Translate, squeeze, and/or delete characters from\n" \
	"standard input, writing to standard output.\n\n" \
	"Options:\n" \
	"\t-c\ttake complement of STRING1\n" \
	"\t-d\tdelete input characters coded STRING1\n" \
	"\t-s\tsqueeze multiple output characters of STRING2 into one character"
#else

#include "opt_gen.h"
OPT_GEN(opt(c, COM_FL_OPT), opt(d, DEL_FL_OPT), opt(s, SQ_FL_OPT))
#include "opt_gen.h"

#include "busybox.h"
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#define ASCII 0377

CALC_BSS struct BB_COMMON_BSS {
	/* some "globals" shared across this file */
	char com_fl;
	char del_fl;
	char sq_fl;
	int in_index;
	int out_index;
	unsigned char output[BUFSIZ];
	unsigned char vector[ASCII+1];
	unsigned char invec[ASCII+1];
	unsigned char outvec[ASCII+1];
	char input[BUFSIZ];
} bb_common_bss;
#define com_fl bb_common_bss.com_fl
#define del_fl bb_common_bss.del_fl
#define sq_fl bb_common_bss.sq_fl
#define in_index bb_common_bss.in_index
#define out_index bb_common_bss.out_index
#define output bb_common_bss.output
#define vector bb_common_bss.vector
#define invec bb_common_bss.invec
#define outvec bb_common_bss.outvec
#define input bb_common_bss.input

static void convert(void)
{
	int read_chars = 0;
	int c, coded;
	int last = -1;

	for (;;) {
		if (in_index == read_chars) {
			if ((read_chars = read(0, input, BUFSIZ)) <= 0) {
				if (write(1, (char *) output, out_index) != out_index)
					bb_error_msg("%s", bb_msg_write_error);
				return;
			}
			in_index = 0;
		}
		c = input[in_index++];
		coded = vector[c];
		if (del_fl && invec[c])
			continue;
		if (sq_fl && last == coded && (invec[c] || outvec[coded]))
			continue;
		output[out_index++] = last = coded;
		if (out_index == BUFSIZ) {
			if (write(1, (char *) output, out_index) != out_index)
				bb_error_msg_and_die("%s", bb_msg_write_error);
			out_index = 0;
		}
	}

	/* NOTREACHED */
}

static void map(unsigned char *string1, unsigned int string1_len,
		unsigned char *string2, unsigned int string2_len)
{
	unsigned char last = '0';
	unsigned int i, j;

	for (j = 0, i = 0; i < string1_len; i++) {
		if (string2_len <= j)
			vector[string1[i]] = last;
		else
			vector[string1[i]] = last = string2[j++];
	}
}

/* supported constructs:
 *   Ranges,  e.g.,  [0-9]  ==>  0123456789
 *   Escapes, e.g.,  \a     ==>  Control-G
 *       Character classes, e.g. [:upper:] ==> A ... Z
 */
static unsigned int expand(const char *arg, unsigned char *buffer)
{
	unsigned char *buffer_start = buffer;
	int i, ac;

	while (*arg) {
		if (*arg == '\\') {
			arg++;
			*buffer++ = bb_process_escape_sequence(&arg);
		} else if (*(arg+1) == '-') {
			ac = *(arg+2);
			if(ac == 0) {
				*buffer++ = *arg++;
				continue;
			}
			i = *arg;
			while (i <= ac)
				*buffer++ = i++;
			arg += 3; /* Skip the assumed a-z */
		} else if (*arg == '[') {
			arg++;
			if (ENABLE_FEATURE_TR_CLASSES && *arg++ == ':') {
				if (strncmp(arg, "alpha", 5) == 0) {
					for (i = 'A'; i <= 'Z'; i++)
						*buffer++ = i;
					for (i = 'a'; i <= 'z'; i++)
						*buffer++ = i;
				}
				else if (strncmp(arg, "alnum", 5) == 0) {
					for (i = 'A'; i <= 'Z'; i++)
						*buffer++ = i;
					for (i = 'a'; i <= 'z'; i++)
						*buffer++ = i;
					for (i = '0'; i <= '9'; i++)
						*buffer++ = i;
				}
				else if (strncmp(arg, "digit", 5) == 0)
					for (i = '0'; i <= '9'; i++)
						*buffer++ = i;
				else if (strncmp(arg, "lower", 5) == 0)
					for (i = 'a'; i <= 'z'; i++)
						*buffer++ = i;
				else if (strncmp(arg, "upper", 5) == 0)
					for (i = 'A'; i <= 'Z'; i++)
						*buffer++ = i;
				else if (strncmp(arg, "space", 5) == 0) {
					static const char spaces[] = " \f\n\r\t\v";
					for(i = 0; spaces[i] != '\0'; i++)
						*buffer++ = spaces[i];
				} else if (strncmp(arg, "blank", 5) == 0) {
					*buffer++ = '\t';
					*buffer++ = ' ';
				/* gcc gives a warning if braces aren't used here */
				} else if (strncmp(arg, "punct", 5) == 0) {
					for (i = 0; i <= ASCII; i++)
						if (isprint(i) && (!isalnum(i)) && (!isspace(i)))
							*buffer++ = i;
				}
				else if (strncmp(arg, "cntrl", 5) == 0) {
					for (i = 0; i <= ASCII; i++)
						if (iscntrl(i))
							*buffer++ = i;
				}
				else {
					strcat((char*)buffer, "[:");
					arg++;
					continue;
				}
				break;
			}
			if (ENABLE_FEATURE_TR_EQUIV && *arg++ == '=') {
				*buffer++ = *arg;
				/* skip the closing =] */
				arg += 3;
				continue;
			}
			if (*arg++ != '-') {
				*buffer++ = '[';
				arg -= 2;
				continue;
			}
			i = *arg++;
			ac = *arg++;
			while (i <= ac)
				*buffer++ = i++;
			arg++;                          /* Skip the assumed ']' */
		} else
			*buffer++ = *arg++;
	}

	return (buffer - buffer_start);
}

static int complement(unsigned char *buffer, int buffer_len)
{
	int i, j, ix;
	char conv[ASCII + 2];

	ix = 0;
	for (i = 0; i <= ASCII; i++) {
		for (j = 0; j < buffer_len; j++)
			if (buffer[j] == i)
				break;
		if (j == buffer_len)
			conv[ix++] = i & ASCII;
	}
	memcpy(buffer, conv, ix);
	return ix;
}

extern int tr_main(void)
{
	char **argv;
	int output_length=0, input_length;
	int opt, i, argc, min_operands, max_operands;

	opt = bb_getopt_ulflags(opt_str);
	com_fl = opt & COM_FL_OPT;
	del_fl = (opt & DEL_FL_OPT) != 0;
	sq_fl = (opt & SQ_FL_OPT) != 0;
	argv = Gargv;
	argc = Gargc;
	min_operands = 1 + (del_fl == sq_fl);
	max_operands = 1 + (del_fl <= sq_fl);

	if (argc < min_operands) {
		if (argc == 0)
			bb_error_msg_and_die("missing operand");
		else
			bb_error_msg_and_die("missing operand after '%s'", argv[argc - 1]);
	}
	if (max_operands < argc)
		bb_error_msg_and_die("extra operand '%s'", argv[max_operands]);

	for (i = 0; i <= ASCII; i++) {
		vector[i] = i;
		invec[i] = outvec[i] = FALSE;
	}
	input_length = expand(*argv++, (unsigned char*)input);
	if (com_fl)
		input_length = complement((unsigned char*)input, input_length);
	if (*argv != NULL) {
		if (**argv == '\0')
			bb_error_msg_and_die("STRING2 cannot be empty");
		output_length = expand(*argv++, output);
		map((unsigned char*)input, input_length, output, output_length);
	}
	for (i = 0; i < input_length; i++)
		invec[(unsigned char)input[i]] = TRUE;
	for (i = 0; i < output_length; i++)
		outvec[output[i]] = TRUE;
	convert();
	return 0;
}
#endif /* APPLET_USAGE */
