/*
 * Mini run-parts implementation for busybox
 *
 *
 * Copyright (C) 2001 by Emanuele Aina <emanuele.aina@tiscali.it>
 *
 * Based on the Debian run-parts program, version 1.15
 *   Copyright (C) 1996 Jeff Noxon <jeff@router.patch.net>,
 *   Copyright (C) 1996-1999 Guy Maor <maor@debian.org>
 *
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */

/* This is my first attempt to write a program in C (well, this is my first
 * attempt to write a program! :-) . */

/* This piece of code is heavily based on the original version of run-parts,
 * taken from debian-utils. I've only removed the long options and a the
 * report mode. As the original run-parts support only long options, I've
 * broken compatibility because the BusyBox policy doesn't allow them.
 * The supported options are:
 * -t                   test. Print the name of the files to be executed, without
 *                              execute them.
 * -a ARG               argument. Pass ARG as an argument the program executed. It can
 *                              be repeated to pass multiple arguments.
 * -u MASK              umask. Set the umask of the program executed to MASK. */

#ifdef APPLET_USAGE
#define run_parts_trivial_usage \
	"[-t] [-a ARG] [-u MASK] DIRECTORY"
#define run_parts_full_usage \
	"Run a bunch of scripts in a directory.\n\n" \
	"Options:\n" \
	"\t-t\tPrints what would be run, but does not actually run anything\n" \
	"\t-a ARG\tPass ARG as an argument for every program invoked\n" \
	"\t-u MASK\tSet the umask to MASK before executing every program"
#else

#include "busybox.h"
#include <sys/wait.h>
#include <dirent.h>
#include <unistd.h>
#include <ctype.h>

#if ENABLE_FEATURE_LONG_OPTIONS
static const struct option runparts_long_options[] = {
	{ "test",               0,              NULL,           't' },
	{ "umask",              1,              NULL,           'u' },
	{ "arg",                1,              NULL,           'a' },
	{ NULL,                 0,              NULL,            0 }
};
#endif

/* valid_name */
/* True or false? Is this a valid filename (upper/lower alpha, digits,
 * underscores, and hyphens only?)
 */
static int valid_name(const struct dirent *d)
{
	const char *c = d->d_name;

	while (*c) {
		if (!isalnum(*c) && (*c != '_') && (*c != '-')) {
			return 0;
		}
		++c;
	}
	return 1;
}

/* test mode = 1 is the same as official run_parts
 * test_mode = 2 means to fail silently on missing directories
 */

static int run_parts(char **args, const int test_mode)
{
	struct dirent **namelist = 0;
	struct stat st;
	char *filename;
	char *arg0 = args[0];
	int entries;
	int i;
	int exitstatus = 0;

#if __GNUC__
	/* Avoid longjmp clobbering */
	(void) &i;
	(void) &exitstatus;
#endif
	/* scandir() isn't POSIX, but it makes things easy. */
	entries = scandir(arg0, &namelist, valid_name, alphasort);

	if (entries == -1) {
		if (test_mode & 2) {
			return(2);
		}
		bb_perror_msg_and_die("failed to open directory %s", arg0);
	}

	for (i = 0; i < entries; i++) {

		filename = concat_path_file(arg0, namelist[i]->d_name);

		if (stat(filename, &st) < 0) {
			bb_perror_msg_and_die("failed to stat component %s", filename);
		}
		if (S_ISREG(st.st_mode) && !access(filename, X_OK)) {
			if (test_mode) {
				puts(filename);
			} else {
				/* exec_errno is common vfork variable */
				volatile int exec_errno = 0;
				int result;
				int pid;

				if ((pid = vfork()) < 0) {
					bb_perror_msg_and_die("failed to fork");
				} else if (!pid) {
					args[0] = filename;
					execv(filename, args);
					exec_errno = errno;
					_exit(1);
				}

				waitpid_no_int(pid, &result, 0, 1);
				if(exec_errno) {
					errno = exec_errno;
					bb_perror_msg("failed to exec %s", filename);
					exitstatus = 1;
				}
				if (WIFEXITED(result) && WEXITSTATUS(result)) {
					bb_perror_msg("%s exited with return code %d", filename, WEXITSTATUS(result));
					exitstatus = 1;
				} else if (WIFSIGNALED(result)) {
					bb_perror_msg("%s exited because of uncaught signal %d", filename, WTERMSIG(result));
					exitstatus = 1;
				}
			}
		}
		else if (!S_ISDIR(st.st_mode)) {
			bb_error_msg("component %s is not an executable plain file", filename);
			exitstatus = 1;
		}

		free(namelist[i]);
		free(filename);
	}
	free(namelist);

	return(exitstatus);
}

/* run_parts_main */
/* Process options */
int run_parts_main(void)
{
	char **args;
	int test_mode = 0;
	int argcount = 2;
	int opt;
	llist_t *a = NULL;
	char *su = NULL;

	umask(022);

#if ENABLE_FEATURE_LONG_OPTIONS
	bb_applet_long_options = runparts_long_options;
#endif
	/* We require exactly one argument: the directory name */
	/* count '-t' and '-a', '-a' is list */
	opt = bb_xgetopt_ulflags("u:a:t\0=1:a::aa:tt", &su, &a,
						&argcount, &test_mode);
	if(su) {
		/* Check and set the umask of the program executed. As stated in the original
		 * run-parts, the octal conversion in libc is not foolproof; it will take the
		 * 8 and 9 digits under some circumstances. We'll just have to live with it.
		 */
		umask(bb_xparse_ui_number(8 | BB_PN_SET_UPPER | BB_PN_ERR_STR1 | BB_PN_SHOW_USAGE,
						su, 07777, "invalid mask `%s'"));
	}

	/* Add an argument to the commands that we will call.
	* Called once for every argument. */
	args = xmalloc(argcount * (sizeof(char *)));
	opt = 0;
	args[opt++] = *Gargv;
	while(a) {
		args[opt++] = a->data;
		a = a->link;
	}
	args[opt] = NULL;

	return(run_parts(args, test_mode));
}
#endif /* APPLET_USAGE */
