/*
 * Mini start-stop-daemon implementation(s) for busybox
 *
 * Written by Marek Michalkiewicz <marekm@i17linuxb.ists.pwr.wroc.pl>,
 * public domain.
 * Adapted for busybox David Kimdon <dwhedon@gordian.com>
 */
#ifdef APPLET_USAGE
#define start_stop_daemon_trivial_usage \
	"[OPTIONS] [--start|--stop] ... [-- arguments...]"
#define start_stop_daemon_full_usage \
	"Program to start and stop services." \
	"\n\nOptions:" \
	"\n\t-S|--start\t\t\tstart" \
	"\n\t-K|--stop\t\t\tstop" \
	"\n\t-a|--startas <pathname>\t\tstarts process specified by pathname" \
	"\n\t-b|--background\t\t\tforce process into background" \
	"\n\t-u|--user <username>|<uid>\tstop this user's processes" \
	"\n\t-x|--exec <executable>\t\tprogram to either start or check" \
	"\n\t-m|--make-pidfile <filename>\tcreate the -p file and enter pid in it" \
	"\n\t-n|--name <process-name>\tstop processes with this name" \
	"\n\t-p|--pidfile <pid-file>\t\tsave or load pid using a pid-file" \
	"\n\t-q|--quiet\t\t\tbe quiet" \
	"\n\t-s|--signal <signal>\t\tsignal to send (default TERM)"
#else

#include "busybox.h"
#include <string.h>
#include <signal.h>
#include <dirent.h>
#include <unistd.h>
#include <syslog.h>


struct pid_list {
	struct pid_list *next;
	pid_t pid;
};

CALC_BSS struct BB_COMMON_BSS
{
	int quiet;
	char *userspec;
	char *cmdname;
	char *execname;
	char *pidfile;
	struct pid_list *found;
	int signal_nr;
	int user_id;
	char what[1024];
} bb_common_bss;
#define quiet bb_common_bss.quiet
#define userspec bb_common_bss.userspec
#define cmdname bb_common_bss.cmdname
#define execname bb_common_bss.execname
#define pidfile bb_common_bss.pidfile
#define found bb_common_bss.found
#define signal_nr bb_common_bss.signal_nr
#define user_id bb_common_bss.user_id


static inline void push(pid_t pid)
{
	struct pid_list *p;

	p = xmalloc(sizeof(*p));
	p->next = found;
	p->pid = pid;
	found = p;
}

static int pid_is_exec(pid_t pid, const char *name)
{
	char buf[32];
	struct stat sb, exec_stat;

	if (name && stat(name, &exec_stat))
		bb_perror_msg_and_die("stat %s", name);

	sprintf(buf, "/proc/%d/exe", pid);
	if (stat(buf, &sb) != 0)
		return 0;
	return (sb.st_dev == exec_stat.st_dev && sb.st_ino == exec_stat.st_ino);
}

static int pid_is_user(int pid, int uid)
{
	struct stat sb;
	char buf[32];

	sprintf(buf, "/proc/%d", pid);
	if (stat(buf, &sb) != 0)
		return 0;
	return (sb.st_uid == uid);
}

static int pid_is_cmd(pid_t pid, const char *name)
{
	char buf[32];
	FILE *f;
	int c;

	sprintf(buf, "/proc/%d/stat", pid);
	f = fopen(buf, "r");
	if (!f)
		return 0;
	while ((c = getc(f)) != EOF && c != '(')
		;
	if (c != '(') {
		fclose(f);
		return 0;
	}
	/* this hopefully handles command names containing ')' */
	while ((c = getc(f)) != EOF && c == *name)
		name++;
	fclose(f);
	return (c == ')' && *name == '\0');
}


static void check(int pid)
{
	if (execname && !pid_is_exec(pid, execname))
		return;
	if (userspec && !pid_is_user(pid, user_id))
		return;
	if (cmdname && !pid_is_cmd(pid, cmdname))
		return;
	push(pid);
}


static void do_pidfile(void)
{
	FILE *f;
	pid_t pid;

	f = fopen(pidfile, "r");
	if (f) {
		if (fscanf(f, "%d", &pid) == 1)
			check(pid);
		fclose(f);
	} else if (errno != ENOENT)
		bb_perror_msg_and_die("open pidfile %s", pidfile);

}

static void do_procinit(void)
{
	DIR *procdir;
	struct dirent *entry;
	int foundany, pid;

	if (pidfile) {
		do_pidfile();
		return;
	}

	procdir = opendir("/proc");
	if (!procdir)
		bb_perror_msg_and_die ("opendir /proc");

	foundany = 0;
	while ((entry = readdir(procdir)) != NULL) {
		if (sscanf(entry->d_name, "%d", &pid) != 1)
			continue;
		foundany++;
		check(pid);
	}
	closedir(procdir);
	if (!foundany)
		bb_error_msg_and_die ("nothing in /proc - not mounted?");
}


static void do_stop(void)
{
	struct pid_list *p;
	int killed = 0;

	do_procinit();

	if (cmdname)
		strcpy(bb_common_bss.what, cmdname);
	else if (execname)
		strcpy(bb_common_bss.what, execname);
	else if (pidfile)
		sprintf(bb_common_bss.what, "process in pidfile `%.200s'", pidfile);
	else if (userspec)
		sprintf(bb_common_bss.what, "process(es) owned by `%s'", userspec);
	else
		bb_error_msg_and_die ("internal error, please report");

	if (!found) {
		if (!quiet)
			printf("no %s found; none killed.\n", bb_common_bss.what);
		return;
	}
	for (p = found; p; p = p->next) {
		if (kill(p->pid, signal_nr) == 0) {
			p->pid = -p->pid;
			killed++;
		} else {
			bb_perror_msg("warning: failed to kill %d", p->pid);
		}
	}
	if (!quiet && killed) {
		printf("stopped %s (pid", bb_common_bss.what);
		for (p = found; p; p = p->next)
			if(p->pid < 0)
				printf(" %d", -p->pid);
		printf(").\n");
	}
}


#if ENABLE_FEATURE_LONG_OPTIONS
static const struct option ssd_long_options[] = {
	{ "stop",                       0,              NULL,           'K' },
	{ "start",                      0,              NULL,           'S' },
	{ "background",                 0,              NULL,           'b' },
	{ "quiet",                      0,              NULL,           'q' },
	{ "make-pidfile",               0,              NULL,           'm' },
	{ "startas",                    1,              NULL,           'a' },
	{ "name",                       1,              NULL,           'n' },
	{ "signal",                     1,              NULL,           's' },
	{ "user",                       1,              NULL,           'u' },
	{ "exec",                       1,              NULL,           'x' },
	{ "pidfile",                    1,              NULL,           'p' },
	{ NULL,                         0,              NULL,            0 }
};
#endif

#define SSD_CTX_STOP            1
#define SSD_CTX_START           2
#define SSD_OPT_BACKGROUND      4
#define SSD_OPT_QUIET           8
#define SSD_OPT_MAKEPID         16

int start_stop_daemon_main(void)
{
	char **argv = Gargv;
	unsigned long opt;
	char *signame = NULL;
	char *startas = NULL;

#if ENABLE_FEATURE_LONG_OPTIONS
	bb_applet_long_options = ssd_long_options;
#endif

	/* Check required one context option was given */
	opt = bb_xgetopt_ulflags("KSbqma:n:s:u:x:p:\0K:S:?K--S:S--K",
		&startas, &cmdname, &signame, &userspec, &execname, &pidfile);

	quiet = opt & SSD_OPT_QUIET;

	signal_nr = signame ? bb_xparse_i_number(10 | BB_PN_SET_UPPER, signame, NSIG) : 15;

	if (!execname && !pidfile && !userspec && !cmdname)
		bb_error_msg_and_die ("need at least one of -x, -p, -u, or -n");

	if (!startas)
		startas = execname;

	if ((opt & SSD_CTX_START) && !startas)
		bb_error_msg_and_die ("-S needs -x or -a");

	if ((opt & SSD_OPT_MAKEPID) && pidfile == NULL)
		bb_error_msg_and_die ("-m needs -p");

	if (userspec && sscanf(userspec, "%d", &user_id) != 1) {
		user_id = bb_xgetpwnam(userspec);
	} else {
		user_id = -1;
	}

	if (opt & SSD_CTX_STOP) {
		do_stop();
		return EXIT_SUCCESS;
	}

	do_procinit();

	if (found) {
		if (!quiet)
			printf("%s already running.\n%d\n", execname ,found->pid);
		return EXIT_SUCCESS;
	}
	if ((opt & SSD_OPT_BACKGROUND) != 0) {
		/* user wants _us_ to make the pidfile */
		xdaemon(opt & SSD_OPT_MAKEPID ? XDAEMON_PF_ARG : 0, pidfile);
	}

	argv = Gargv;
	*--argv = startas;
	execv(startas, argv);
	bb_perror_msg_and_die("unable to start %s", startas);
	return 1;
}
#endif /* APPLET_USAGE */
