/*
 * lsattr.c             - List file attributes on an ext2 file system
 *
 * Copyright (C) 1993, 1994  Remy Card <card@masi.ibp.fr>
 *                           Laboratoire MASI, Institut Blaise Pascal
 *                           Universite Pierre et Marie Curie (Paris VI)
 *
 * This file can be redistributed under the terms of the GNU General
 * Public License
 */

/*
 * History:
 * 93/10/30     - Creation
 * 93/11/13     - Replace stat() calls by lstat() to avoid loops
 * 94/02/27     - Integrated in Ted's distribution
 * 98/12/29     - Display version info only when -V specified (G M Sipe)
 */
#ifdef APPLET_USAGE
#define lsattr_trivial_usage \
	"[-Radlv] [files...]"
#define lsattr_full_usage \
	"list file attributes on an ext2 fs\n\n" \
	"Options:\n" \
	"\t-R\trecursively list subdirectories\n" \
	"\t-a\tdo not hide entries starting with .\n" \
	"\t-d\tlist directory entries instead of contents\n" \
	"\t-l\tprint long flag names\n" \
	"\t-v\tlist the file's version/generation number"
#else

#include "opt_gen.h"
OPT_GEN(
opt(R, OPT_RECUR),
opt(a, OPT_ALL),
opt(d, OPT_DIRS_OPT),
opt(l, OPT_PF_LONG),
opt(v, OPT_GENERATION)
)
#include "opt_gen.h"

#include <dirent.h>
#include <unistd.h>
#include <string.h>
#include <sys/param.h>

#include "ext2fs/ext2_fs.h"
#include "e2fsbb.h"
#include "e2p/e2p.h"

#ifdef __GNUC__
# define EXT2FS_ATTR(x) __attribute__(x)
#else
# define EXT2FS_ATTR(x)
#endif

static void list_attributes(const char *name)
{
	unsigned long fsflags;
	unsigned long generation;

	if (fgetflags(name, &fsflags) == -1)
		goto read_err;
	if (bss_flags & OPT_GENERATION) {
		if (fgetversion(name, &generation) == -1)
			goto read_err;
		printf("%5lu ", generation);
	}

	if (bss_flags & OPT_PF_LONG) {
		printf("%-28s ", name);
		print_flags(stdout, fsflags, PFOPT_LONG);
		printf("\n");
	} else {
		print_flags(stdout, fsflags, 0);
		printf(" %s\n", name);
	}

	return;
read_err:
	bb_perror_msg("reading %s", name);
}

static int lsattr_dir_proc(const char *dir_name, struct dirent *de)
{
	struct stat     st;
	char *path;

	path = concat_path_file(dir_name, de->d_name);

	if (lstat(path, &st) == -1)
		bb_perror_msg(path);
	else {
		if (de->d_name[0] != '.' || (bss_flags & OPT_ALL)) {
			list_attributes(path);
			if (S_ISDIR(st.st_mode) && (bss_flags & OPT_RECUR) &&
			   (de->d_name[0] != '.' && (de->d_name[1] != '\0' ||
			   (de->d_name[1] != '.' && de->d_name[2] != '\0')))) {
				printf("\n%s:\n", path);
				iterate_on_dir(path, lsattr_dir_proc);
				printf("\n");
			}
		}
	}

	free(path);

	return 0;
}

static void lsattr_args(const char *name)
{
	struct stat     st;

	if (lstat(name, &st) == -1) {
		bb_perror_msg("stating %s", name);
	} else {
		if (S_ISDIR(st.st_mode) && !(bss_flags & OPT_DIRS_OPT))
			iterate_on_dir(name, lsattr_dir_proc);
		else
			list_attributes(name);
	}
}

int lsattr_main(void)
{
	char **argv;

	bb_getopt_ulflags(opt_str);
	argv = Gargv;
	if (*argv == NULL) {
		argv--;
		*argv = ".";
	}
	for (; *argv != NULL; argv++)
		lsattr_args(*argv);

	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
