/*  GNU SED, a batch stream editor.
    Copyright (C) 1998 Free Software Foundation, Inc.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

    Copyright (c) 2016 Vladimir Oleynik <dzo@simtreas.ru>
     Busybox port
*/

#ifdef APPLET_USAGE
#define sed_trivial_usage \
	"[-nr] [-e command] pattern|[-f command_file] [files...]"
#define sed_full_usage \
	"Options:\n" \
	"\t-e script\tadd the script to the commands to be executed\n" \
	"\t-f scriptfile\tadd script-file contents to the\n" \
	"\t\t\tcommands to be executed\n" \
	"\t-n\t\tsuppress automatic printing of pattern space\n" \
	"\t-r\t\tuse extended regular expression syntax\n" \
	"\n" \
	"If no -e or -f is given, the first non-option argument is taken as the sed\n" \
	"script to interpret. All remaining arguments are names of input files; if no\n" \
	"input files are specified, then the standard input is read. Source files\n" \
	"will not be modified unless -i option is given."
#else

/* Package VERSION number */
#define VERSION "3.02"

#define INITIAL_BUFFER_SIZE     50
#define LCMD_OUT_LINE_LEN       70
#define FREAD_BUFFER_SIZE       8192

#include "busybox.h"
#include <ctype.h>
#include <unistd.h>
#include <regex.h>
#include <string.h>
#include <memory.h>

#include <sys/mman.h>
#ifndef MAP_FAILED
#  define MAP_FAILED    ((char *)-1)    /* what a stupid concept... */
#endif
# ifndef S_ISREG
#  define S_ISREG(m)    (((m)&S_IFMT) == S_IFMT)
#endif

#define MALLOC(n,t)     (xmalloc((n)*sizeof(t)))
#define REALLOC(x,n,t)  (ck_realloc((x),(n)*sizeof(t)))
#define MEMDUP(x,n,t)   (ck_memdup((x),(n)*sizeof(t)))

#ifdef isblank
# define ISBLANK(c) (isblank(c))
#else
# define ISBLANK(c) ((c) == ' ' || (c) == '\t')
#endif

#undef ISPRINT
#define ISPRINT(c) isprint(c)
#define ISDIGIT(c) isdigit(c)
#define ISALNUM(c) isalnum(c)
#define ISALPHA(c) isalpha(c)
#define ISCNTRL(c) iscntrl(c)
#define ISLOWER(c) islower(c)
#define ISPUNCT(c) ispunct(c)
#define ISSPACE(c) isspace(c)
#define ISUPPER(c) isupper(c)
#define ISXDIGIT(c) isxdigit(c)


/* type countT is used to keep track of line numbers, etc. */
typedef unsigned long countT;

/* type flagT is used for boolean values */
typedef int flagT;

/* Struct vector is used to describe a chunk of a compiled sed program.
 * There is one vector for the main program, and one for each { } pair.
 * For {} blocks, RETURN_[VI] tells where to continue execution after
 * this VECTOR. */
struct vector {
	struct sed_cmd *v;              /* a dynamically allocated array */
	size_t v_allocated;             /* ... number slots allocated */
	size_t v_length;                /* ... number of slots in use */
	struct error_info *err_info;    /* info for unclosed '{'s */
	struct vector *return_v;        /* where the matching '}' will return to */
	countT return_i;                /* ... further info */
};

/* Goto structure is used to hold both GOTO's and labels.  There are two
 * separate lists, one of goto's, called 'jumps', and one of labels, called
 * 'labels'.
 * The V element points to the descriptor for the program-chunk in which the
 * goto was encountered.
 * The v_index element counts which element of the vector actually IS the
 * goto/label.
 * The NAME element is the null-terminated name of the label.
 * "next" is the next goto/label in the linked list. */
struct sed_label {
	struct vector *v;
	countT v_index;
	char *name;
	struct sed_label *next;
};

struct text_buf {
	char *text;
	size_t text_len;
};


struct addr {
	enum {
		addr_is_null,   /* null address */
		addr_is_num,    /* a.addr_number is valid */
		addr_is_mod,    /* doing every n'th line; a.m.* are valid */
		addr_is_last,   /* address is $ */
		addr_is_regex   /* a.addr_regex is valid */
	} addr_type;

	union {
		regex_t *addr_regex;
		countT addr_number;
		struct {
			countT modulo;
			countT offset;
		} m;
	} a;
};

struct subst {
	regex_t *regx;
	char *replacement;
	size_t replace_length;
	countT numb;            /* if >0, only substitute for match number "numb" */
	FILE *wfile;            /* 'w' option given */
	char global;            /* 'g' option given */
	char print;             /* 'p' option given */
};


struct sed_cmd {
	struct addr a1;
	struct addr a2;

	/* non-zero if a1 has been matched; apply command until a2 matches */
	char a1_matched;

	/* non-zero: only apply command to non-matches */
	char addr_bang;

	/* the actual command */
	char cmd;

	/* auxiliary data for various commands */
	union {
		/* This structure is used for a, i, and c commands */
		struct text_buf cmd_txt;

		/* This is used for b and t commands */
		struct sed_label *jump;

		/* This for r command */
		const char *rfile;

		/* This for w command */
		FILE *wfile;

		/* This for the y command */
		unsigned char *translate;

		/* For {} */
		struct vector *sub;

		/* This for the hairy s command */
		struct subst cmd_regex;
	} x;
};

/* Implement a variable sized buffer of 'stuff'.  We don't know what it is,
nor do we care, as long as it doesn't mind being aligned by malloc. */
struct buffer {
	size_t allocated;
	size_t length;
	char *b;
};

/* Store information about files opened with ck_fopen
   so that error messages from ck_fread, ck_fwrite, etc. can print the
   name of the file that had the error */
struct id {
	FILE *fp;
	char *name;
	struct id *link;
};

/* Sed operates a line at a time. */
struct line
{
	char *text;     /* Pointer to line allocated by malloc. */
	char *active;   /* Pointer to non-consumed part of text. */
	size_t length;  /* Length of text (or active, if used). */
	size_t alloc;   /* Allocated space for text. */
	flagT chomped;  /* Was a trailing newline dropped? */
};

/* A queue of text to write out at the end of a cycle
   (filled by the "a" and "r" commands.) */
struct append_queue
{
	const char *rfile;
	const char *text;
	size_t textlen;
	struct append_queue *next;
};

/* State information for the input stream. */
struct input
{
	char **file_list;     /* The list of yet-to-be-opened files.
			   It is invalid for file_list to be NULL.
			   When *file_list is NULL we are
			   currently processing the last file. */
	countT bad_count;     /* count of files we failed to open */
	countT line_number;   /* Current input line number (over all files) */

	flagT (*read_fn)(struct input *);       /* read one line */
	/* If fp is NULL, read_fn better not be one which uses fp;
	   in particular, read_always_fail() is recommended. */

	FILE *fp;             /* if NULL, none of the following are valid */
	void *base;           /* if non-NULL, we are using mmap()ed input */
	const char *cur;      /* only valid if base is non-NULL */
	size_t length;        /* only valid if base is non-NULL */
	size_t left;          /* only valid if base is non-NULL */
	flagT is_tty;         /* only valid if base is NULL */
};

#define YMAP_LENGTH             256     /*XXX shouldn't this be (UCHAR_MAX+1)? */
#define VECTOR_ALLOC_INCREMENT  40

struct prog_info {
	/* When we're reading a script command from a string, 'prog.base'
	   points to the first character in the string, 'prog.cur' points
	   to the current character in the string, and 'prog.end' points
	   to the end of the string.  This allows us to compile script
	   strings that contain nulls, which might happen if we're mmap'ing
	   a file. */
	void *base;
	const unsigned char *cur;
	const unsigned char *end;

	/* This is the current script file.  If it is NULL, we are reading
	   from a string stored at 'prog.cur' instead.  If both 'prog.file'
	   and 'prog.cur' are NULL, we're in trouble! */
	FILE *file;
};

/* Information used to give out useful and informative error messages. */
struct error_info {
	/* This is the name of the current script file. */
	const char *name;

	/* This is the number of the current script line that we're compiling. */
	countT line;

	/* This is the index of the "-e" expressions on the command line. */
	countT string_expr_count;
};

/* This structure tracks files opened by the 'r', 'w', and 's///w' commands
   so that they may all be closed cleanly at normal program termination.  */
struct fp_list {
	char *name;
	FILE *fp;
	flagT readit_p;
	struct fp_list *link;
};

/* Various error messages we may want to print */
static const char ONE_ADDR[] = "Command only uses one address";
static const char NO_ADDR[] = "Command doesn't take any addresses";
static const char LINE_JUNK[] = "Extra characters after command";
static const char BAD_EOF[] = "Unexpected End-of-file";
static const char BAD_REGEX_FMT[] = "bad regexp: %s\n";
static const char EXCESS_OPEN_BRACKET[] = "Unmatched `{'";
static const char EXCESS_CLOSE_BRACKET[] = "Unexpected `}'";
static const char NO_REGEX[] = "No previous regular expression";
static const char NO_COMMAND[] = "Missing command";
static const char UNTERM_S_CMD[] = "Unterminated `s' command";
static const char UNTERM_Y_CMD[] = "Unterminated `y' command";


/* globals */
CALC_BSS struct BB_COMMON_BSS {
	/* Where we are in the processing of the input. */
	struct prog_info prog;
	struct error_info cur_input;

	/* Information about labels and jumps-to-labels.  This is used to do
	   the required backpatching after we have compiled all the scripts. */
	struct sed_label *jumps;
	struct sed_label *labels;

	/* We wish to detect #n magic only in the first input argument;
	   this flag tracks when we have consumed the first file of input. */
	flagT first_script;

	/* Allow for scripts like "sed -e 'i\' -e foo": */
	struct buffer *pending_text;
	struct text_buf *old_text_buf;

	struct fp_list *file_ptrs;

	flagT use_extended_syntax_p;
	flagT rx_testing;
	/* If set, don't write out the line unless explicitly told to */
	flagT no_default_output;
	/* Do we need to be pedantically POSIX compliant? */
	flagT POSIXLY_CORRECT;
	/* Have we done any replacements lately?  This is used by the 't' command. */
	flagT replaced;
	/* The 'current' input line. */
	struct line Gline;
	/* An input line that's been stored by later use by the program */
	struct line hold;
	/* The buffered input look-ahead. */
	struct line irl_buffer;
	struct append_queue *append_head;
	struct append_queue *append_tail;
	struct id *utils_id_s;
	/* The complete compiled SED program that we are going to run: */
	struct vector *the_program;
} bb_common_bss;
#define prog bb_common_bss.prog
#define cur_input bb_common_bss.cur_input
#define pending_text bb_common_bss.pending_text
#define Gline bb_common_bss.Gline
#define utils_id_s bb_common_bss.utils_id_s
#define the_program bb_common_bss.the_program


/* let's not confuse text editors that have only dumb bracket-matching... */
#define OPEN_BRACKET    '['
#define CLOSE_BRACKET   ']'

/* Panic on failing realloc */
static void *ck_realloc(void *ptr, size_t size)
{
	if (size == 0) {
		free(ptr);
		return NULL;
	}
	return ptr == NULL ? xmalloc(size) : xrealloc(ptr, size);
}

/* Internal routine to get a filename from utils_id_s */
static const char *utils_fp_name(FILE * fp)
{
	struct id *p;

	for (p = utils_id_s; p; p = p->link)
		if (p->fp == fp)
			return p->name;
	if (fp == stdin)
		return "{standard input}";
	else if (fp == stdout)
		return "{standard output}";
	else if (fp == stderr)
		return "{standard error}";
	return "{Unknown file pointer}";
}

/* Panic on failing fopen */
static FILE *ck_fopen(const char *name, const char *mode)
{
	FILE *fp;
	struct id *p;

	fp = bb_xfopen (name, mode);
	for (p = utils_id_s; p; p = p->link) {
		if (fp == p->fp) {
			free(p->name);
			break;
		}
	}
	if (!p) {
		p = MALLOC(1, struct id);
		p->link = utils_id_s;
		utils_id_s = p;
	}
	p->name = xstrdup(name);
	p->fp = fp;
	return fp;
}

/* Panic on failing fwrite */
static void ck_fwrite(const void *ptr, size_t size, size_t nmemb, FILE *stream)
{
	if (size && fwrite(ptr, size, nmemb, stream) != nmemb)
		bb_perror_msg_and_die("couldn't write %lu item%s to %s",
		   (unsigned long)nmemb, nmemb == 1 ? "" : "s", utils_fp_name(stream));
}

/* Panic on failing fread */
static size_t ck_fread(void *ptr, size_t size, size_t nmemb, FILE *stream)
{
	if (size && (nmemb = fread(ptr, size, nmemb, stream)) <= 0 && ferror(stream))
		bb_perror_msg_and_die("read error on %s", utils_fp_name(stream));
	return nmemb;
}

/* Panic on failing fflush */
static void ck_fflush(FILE *stream)
{
	if (fflush(stream) == EOF)
		bb_error_msg_and_die("couldn't flush %s", utils_fp_name(stream));
}

/* Panic on failing fclose */
static void ck_fclose(FILE *stream)
{
	struct id r;
	struct id *prev;
	struct id *cur;

	if (!stream)
		return;
	if (fclose(stream) == EOF)
		bb_error_msg_and_die("couldn't close %s", utils_fp_name(stream));
	r.link = utils_id_s;
	prev = &r;
	while ((cur = prev->link)) {
		if (stream == cur->fp) {
			prev->link = cur->link;
			free(cur->name);
			free(cur);
#ifdef TRUST_THAT_ID_CHAIN_IS_CLEAN
			break;
#endif
		} else {
			prev = cur;
		}
	}
	utils_id_s = r.link;
}

/* Return a malloc()'d copy of a block of memory */
static void *ck_memdup(const void *buf, size_t len)
{
	void *ret = xmalloc(len);
	return memcpy(ret, buf, len);
}

#define MIN_ALLOCATE 50

static struct buffer *init_buffer(void)
{
	struct buffer *b = MALLOC(1, struct buffer);
	b->b = MALLOC(MIN_ALLOCATE, char);
	b->allocated = MIN_ALLOCATE;
	b->length = 0;
	return b;
}

static inline char *get_buffer(struct buffer *b)
{
	return b->b;
}

static inline size_t size_buffer(struct buffer *b)
{
	return b->length;
}

static void resize_buffer(struct buffer *b, size_t newlen)
{
	char *try = NULL;
	size_t alen = b->allocated;

	if (newlen <= alen)
		return;
	alen *= 2;
	if (newlen < alen)
		try = realloc(b->b, alen);     /* Note: *not* the REALLOC() macro! */
	if (!try) {
		alen = newlen;
		try = REALLOC(b->b, alen, char);
	}
	b->allocated = alen;
	b->b = try;
}

static void add1_buffer(struct buffer *b, int c)
{
	/* This special case should be kept cheap;
	 *  don't make it just a mere convenience
	 *  wrapper for add_buffer() -- even "builtin"
	 *  versions of memcpy(a, b, 1) can become
	 *  expensive when called too often. */
	if (c != EOF) {
		if (b->allocated - b->length < 1)
			resize_buffer(b, b->length + 1);
		b->b[b->length++] = c;
	}
}

static void free_buffer(struct buffer *b)
{
	if (b)
		free(b->b);
	free(b);
}

/* dummy function to simplify read_pattern_space() */
static flagT read_always_fail(struct input *input ATTRIBUTE_UNUSED)
{
	return 0;
}

/* Attempt to munmap() a memory region. */
static void unmap_file(void *base, size_t len)
{
	if (base)
		munmap(base, len);
}

/* Clean up an input stream that we are done with. */
static void closedown(struct input *input)
{
	input->read_fn = read_always_fail;
	if (!input->fp)
		return;
	if (input->base)
		unmap_file(input->base, input->length);
	if (input->fp != stdin) /* stdin can be reused on tty and tape devices */
		ck_fclose(input->fp);
	input->fp = NULL;
}

/* increase a struct line's length, making some attempt at
   keeping realloc() calls under control by padding for future growth.  */
static void resize_line(struct line *lb, size_t len)
{
	lb->alloc *= 2;
	if (lb->alloc - lb->length < len)
		lb->alloc = lb->length + len;
	lb->text = REALLOC(lb->text, lb->alloc, char);
}

/* Attempt to mmap() a file.  On failure, just return a zero,
   otherwise set *base and *len and return non-zero.  */
static flagT map_file(FILE *fp, void **base, size_t *len)
{
	struct stat s;
	void *nbase;

	if (fstat(fileno(fp), &s) == 0 && S_ISREG(s.st_mode) &&
			(unsigned long long)s.st_size == (unsigned long long)(size_t)s.st_size) {
		/* "As if" the whole file was read into memory at this moment... */
		nbase = mmap(NULL, (size_t) s.st_size, PROT_READ, MAP_PRIVATE, fileno(fp), (off_t)0);
		if (nbase != MAP_FAILED) {
			*base = nbase;
			*len = s.st_size;
			return 1;
		}
	}
	return 0;
}

/* The quick-and-easy mmap()'d case... */
static flagT read_mem_line(struct input *input)
{
	const char *e;
	size_t l;

	if ((e = memchr(input->cur, '\n', input->left))) {
		/* common case */
		l = e++ - input->cur;
		Gline.chomped = 1;
	} else {
		/* This test is placed _after_ the memchr() fails for performance
		reasons (because this branch is the uncommon case and the
		memchr() will safely fail if input->left == 0).  Okay, so the
		savings is trivial.  I'm doing it anyway. */
		if (input->left == 0)
			return 0;
		e = input->cur + input->left;
		l = input->left;
		if (*input->file_list || bb_common_bss.POSIXLY_CORRECT)
			Gline.chomped = 1;
	}

	if (Gline.alloc - Gline.length < l)
		resize_line(&Gline, l);
	memcpy(Gline.text + Gline.length, input->cur, l);
	Gline.length += l;
	input->left -= e - input->cur;
	input->cur = e;
	return 1;
}

/* fgets() doesn't let us handle NULs and fread() buffers too much
 * for interactive use. */
static size_t slow_getline(char *buf, size_t buflen, FILE *in)
{
	size_t resultlen = 0;
	int c;

	while (resultlen < buflen && (c = getc(in)) != EOF) {
		++resultlen;
		*buf++ = c;
		if (c == '\n')
			break;
	}
	if (ferror(in))
		bb_perror_msg_and_die("%s", bb_msg_read_error);
	return resultlen;
}

static flagT read_file_line(struct input *input)
{
	char *b;
	size_t blen;
	size_t initial_length = Gline.length;

#define buffer bb_common_bss.irl_buffer
	if (!buffer.active)
		buffer.active = buffer.text;
	while (!(b = memchr(buffer.active, '\n', buffer.length))) {
		if (Gline.alloc - Gline.length < buffer.length)
			resize_line(&Gline, buffer.length);
		memcpy(Gline.text + Gline.length, buffer.active, buffer.length);
		Gline.length += buffer.length;
		buffer.length = 0;
		if (!feof(input->fp)) {
			if (input->is_tty)
				buffer.length = slow_getline(buffer.text, buffer.alloc, input->fp);
			else
				buffer.length = ck_fread(buffer.text, 1, buffer.alloc, input->fp);
		}

		if (buffer.length == 0) {
			if (*input->file_list || bb_common_bss.POSIXLY_CORRECT)
				Gline.chomped = 1;
			/* Did we hit EOF without reading anything?  If so, try
			the next file; otherwise just go with what we did get. */
			return Gline.length > initial_length;
		}
		buffer.active = buffer.text;
	}

	blen = b - buffer.active;
	if (Gline.alloc - Gline.length < blen)
		resize_line(&Gline, blen);
	memcpy(Gline.text + Gline.length, buffer.active, blen);
	Gline.length += blen;
	++blen;
	Gline.chomped = 1;
	buffer.active += blen;
	buffer.length -= blen;
	return 1;
#undef buffer
}

/* Initialize a struct input for the named file. */
static void open_next_file(const char *name, struct input *input)
{
	input->base = NULL;
	bb_common_bss.irl_buffer.length = 0;

	if (name[0] == '-' && name[1] == '\0') {
		clearerr(stdin);        /* clear any stale EOF indication */
		input->fp = stdin;
	} else if (!(input->fp = fopen(name, "r"))) {
		bb_perror_msg("can't read %s", name);
		input->read_fn = read_always_fail;      /* a redundancy */
		++input->bad_count;
		return;
	}

	input->read_fn = read_file_line;
	if (map_file(input->fp, &input->base, &input->length)) {
		input->cur = input->base;
		input->left = input->length;
		input->read_fn = read_mem_line;
	} else
		input->is_tty = isatty(fileno(input->fp));
}

static flagT last_file_with_data_p(struct input *input)
{
	/*XXX reference implementation is equivalent to:
	return !*input->file_list; */
	for (;;) {
		int ch;

		closedown(input);
		if (!*input->file_list)
			return 1;
		open_next_file(*input->file_list++, input);
		if (input->fp) {
			if (input->base) {
				if (0 < input->left)
					return 0;
			} else if ((ch = getc(input->fp)) != EOF) {
				ungetc(ch, input->fp);
				return 0;
			}
		}
	}
}

/* Determine if we match the '$' address.  */
static flagT test_dollar_EOF(struct input *input)
{
	int ch;

	if (bb_common_bss.irl_buffer.length)
		return 0;
	if (!input->fp)
		return last_file_with_data_p(input);
	if (input->base)
		return (input->left == 0 && last_file_with_data_p(input));
	if (feof(input->fp))
		return last_file_with_data_p(input);
	if ((ch = getc(input->fp)) == EOF)
		return last_file_with_data_p(input);
	ungetc(ch, input->fp);
	return 0;
}

/* Append a NUL char after the end of DEST, without resizing DEST.
   This is to permit the use of regexp routines which expect C strings
   instead of counted strings. */
static void nul_append(struct line *dest)
{
#ifndef HAVE_REGNEXEC
	if (dest->alloc - dest->length < 1)
		resize_line(dest, 1);
	dest->text[dest->length] = '\0';
#endif
}

/* Return non-zero if the current line matches the address
   pointed to by 'addr'.  */
static flagT match_an_address_p(struct addr *addr, struct input *input, flagT is_addr2_p)
{
	switch (addr->addr_type) {
	case addr_is_null:
		return 1;
	case addr_is_num:
		if (is_addr2_p)
			return (input->line_number >= addr->a.addr_number);
		return (input->line_number == addr->a.addr_number);
	case addr_is_mod:
		if (addr->a.m.offset < addr->a.m.modulo)
			return (input->line_number % addr->a.m.modulo == addr->a.m.offset);
		/* offset >= modulo implies we have an extra initial skip */
		if (input->line_number < addr->a.m.offset)
			return 0;
		/* normalize */
		addr->a.m.offset %= addr->a.m.modulo;
		return 1;
	case addr_is_last:
		return test_dollar_EOF(input);
	case addr_is_regex:
		nul_append(&Gline);
		return !regexec(addr->a.addr_regex, Gline.text, 0, NULL, 0);
	default:
		bb_error_msg_and_die("INTERNAL ERROR: bad address type");
	}
	/*NOTREACHED*/
	return 0;
}

/* return non-zero if current address is valid for cmd */
static flagT match_address_p(struct sed_cmd *cmd, struct input *input)
{
	flagT addr_matched = cmd->a1_matched;

	if (addr_matched) {
		if (match_an_address_p(&cmd->a2, input, 1))
			cmd->a1_matched = 0;
	} else if (match_an_address_p(&cmd->a1, input, 0)) {
		addr_matched = 1;
		if (cmd->a2.addr_type != addr_is_null)
			if ((cmd->a2.addr_type == addr_is_regex) || !match_an_address_p(&cmd->a2, input, 1))
				cmd->a1_matched = 1;
	}
	if (cmd->addr_bang)
		addr_matched = !addr_matched;
	return addr_matched;
}

static void do_list(void)
{
	unsigned char *p = (unsigned char *)Gline.text;
	countT len = Gline.length;
	countT width = 0;
	char obuf[180]; /* just in case we encounter a 512-bit char ;-) */
	char *o;
	size_t olen;

	for (; len--; ++p) {
		o = obuf;
		if (ISPRINT(*p)) {
			*o++ = *p;
			if (*p == '\\')
				*o++ = '\\';
		} else {
			*o++ = '\\';
			switch (*p) {
			case '\a':
				*o++ = 'a';
				break;
			case '\b':
				*o++ = 'b';
				break;
			case '\f':
				*o++ = 'f';
				break;
			case '\n':
				*o++ = 'n';
				break;
			case '\r':
				*o++ = 'r';
				break;
			case '\t':
				*o++ = 't';
				break;
			case '\v':
				*o++ = 'v';
				break;
			default:
				sprintf(o, "%03o", *p);
				o += strlen(o);
				break;
			}
		}
		olen = o - obuf;
		if (width + olen >= LCMD_OUT_LINE_LEN) {
			ck_fwrite("\\\n", 1, 2, stdout);
			width = 0;
		}
		ck_fwrite(obuf, 1, olen, stdout);
		width += olen;
	}
	ck_fwrite("$\n", 1, 2, stdout);
}

static void output_line(const char *text, size_t length, flagT nl, FILE *fp)
{
	ck_fwrite(text, 1, length, fp);
	if (nl)
		ck_fwrite("\n", 1, 1, fp);
	if (fp != stdout)
		ck_fflush(fp);
}

/* initialize a "struct line" buffer */
static void line_init(struct line *buf, size_t initial_size)
{
	buf->text = MALLOC(initial_size, char);
	buf->active = NULL;
	buf->alloc = initial_size;
	buf->length = 0;
	buf->chomped = 1;
}

/* Append 'length' bytes from 'string' to the line 'to'
   This routine *will* append NUL bytes without failing.  */
static void str_append(struct line *to, const char *string, size_t length)
{
	if (to->alloc - to->length < length)
		resize_line(to, length);
	memcpy(to->text + to->length, string, length);
	to->length += length;
}

static void do_subst(struct subst *sub)
{
	/* s_accum will accumulate the result of the s command. */
	/* static so as to keep malloc() calls down */
	static struct line s_accum;

	size_t start;           /* where to start scan for match in LINE */
	size_t offset;          /* where in LINE a match was found */
	size_t remain;          /* length after START, sans trailing newline */
	countT count;           /* number of matches found */
	char *rep;              /* the replacement string */
	char *rep_end;          /* end of the replacement string */
	flagT not_bol_p;
	flagT did_subst;
	regmatch_t regs[10];

	if (s_accum.alloc == 0)
		line_init(&s_accum, INITIAL_BUFFER_SIZE);
	s_accum.length = 0;

	count = 0;
	did_subst = 0;
	not_bol_p = 0;
	start = 0;
	remain = Gline.length - start;
	rep = sub->replacement;
	rep_end = rep + sub->replace_length;

	nul_append(&Gline);
	while (!regexec(sub->regx, Gline.text + start, sizeof(regs) / sizeof(regs[0]), regs, not_bol_p)) {
		++count;
		offset = regs[0].rm_so + start;

		/*
		 *  +- line.text   +-[offset]
		 *  V              V
		 * "blah blah blah xyzzy blah blah blah blah"
		 *       ^             ^                    ^
		 *       +-[start]     +-[end]              +-[start + remain]
		 *
		 *
		 * regs[0].rm_so == offset - start
		 * regs[0].rm_eo == end - start */

		/* Copy stuff to the left of the next match into the output string. */
		if (start < offset)
			str_append(&s_accum, Gline.text + start, offset - start);

		/* If we're counting up to the Nth match, are we there yet? */
		if (count < sub->numb) {
			/* Not there yet...so skip this match. */
			size_t matched = regs[0].rm_eo - regs[0].rm_so;

			/* If the match was vacuous, skip ahead one character
			 * anyway.   It isn't at all obvious to me that this is
			 * the right behavior for this case.    -t    XXX */
			if (matched == 0 && offset < Gline.length)
				matched = 1;
			str_append(&s_accum, Gline.text + offset, matched);
			start = offset + matched;
			remain = Gline.length - start;
			not_bol_p = 1;
			continue;
		}

		/* Expand the replacement string into the output string. */
		{
			char *rep_cur;  /* next burst being processed in replacement */
			char *rep_next; /* end of burst in replacement */

			/* Replacement strings can be viewed as a sequence of variable
			 * length commands.  A command can be a literal string or a
			 * backreference (either numeric or "&").
			 *
			 * This loop measures off the next command between
			 * REP_CUR and REP_NEXT, handles that command, and loops. */
			for (rep_next = rep_cur = rep; rep_next < rep_end; rep_next++) {
				if (*rep_next == '\\') {
					/* Preceding the backslash may be some literal text: */
					if (rep_cur < rep_next)
						str_append(&s_accum, rep_cur, (size_t)(rep_next - rep_cur));

					/* Skip the backslash itself and look for
					 * a numeric back-reference: */
					rep_next++;
					if (rep_next < rep_end && '0' <= *rep_next && *rep_next <= '9') {
						int i = *rep_next - '0';
						str_append(&s_accum, Gline.text + start + regs[i].rm_so,
								(size_t)(regs[i].rm_eo - regs[i].rm_so));
					} else
						str_append(&s_accum, rep_next, 1);
					rep_cur = rep_next + 1;
				} else if (*rep_next == '&') {
					/* Preceding the ampersand may be some literal text: */
					if (rep_cur < rep_next)
						str_append(&s_accum, rep_cur, (size_t)(rep_next - rep_cur));
					str_append(&s_accum, Gline.text + start + regs[0].rm_so,
							(size_t)(regs[0].rm_eo - regs[0].rm_so));
					rep_cur = rep_next + 1;
				}
			}

			/* There may be a trailing literal in the replacement string. */
			if (rep_cur < rep_next)
				str_append(&s_accum, rep_cur, (size_t)(rep_next - rep_cur));
		}

		did_subst = 1;
		not_bol_p = 1;
		start += regs[0].rm_eo;
		remain = Gline.length - start;
		if (!sub->global ||remain == 0)
			break;

		/* If the match was vacuous, skip over one character and add that character to the output. */
		if (regs[0].rm_so == regs[0].rm_eo) {
			str_append(&s_accum, Gline.text + offset, 1);
			++start;
			--remain;
		}
	}

	if (did_subst) {
		struct line tmp;
		if (start < Gline.length)
			str_append(&s_accum, Gline.text + start, remain);
		memcpy(&tmp, &Gline, sizeof Gline);
		memcpy(&Gline, &s_accum, sizeof Gline);
		Gline.chomped = tmp.chomped;
		memcpy(&s_accum, &tmp, sizeof Gline);
		if (sub->wfile)
			output_line(Gline.text, Gline.length, Gline.chomped, sub->wfile);
		if (sub->print)
			output_line(Gline.text, Gline.length, Gline.chomped, stdout);
		bb_common_bss.replaced = 1;
	}
}

/* Copy the contents of the line 'from' into the line 'to'.
   This destroys the old contents of 'to'.  It will still work
   if the line 'from' contains NULs. */
static void line_copy(struct line *from, struct line *to)
{
	if (to->alloc < from->length) {
		to->alloc = from->length;
		to->text = REALLOC(to->text, to->alloc, char);
	}
	memcpy(to->text, from->text, from->length);
	to->length = from->length;
	to->chomped = from->chomped;
}

/* Append the contents of the line 'from' to the line 'to'.
   This routine will work even if the line 'from' contains nulls */
static void line_append(struct line *from, struct line *to)
{
	str_append(to, "\n", 1);
	str_append(to, from->text, from->length);
	to->chomped = from->chomped;
}

static void dump_append_queue(void)
{
	struct append_queue *p, *q;

	for (p = bb_common_bss.append_head; p; p = q) {
		if (p->text)
			output_line(p->text, p->textlen, 0, stdout);
		if (p->rfile) {
			char buf[FREAD_BUFFER_SIZE];
			size_t cnt;
			FILE *fp;

			fp = fopen(p->rfile, "r");
			/* Not ck_fopen() because: "If _rfile_ does not exist or cannot be
			read, it shall be treated as if it were an empty file, causing
			no error condition."  IEEE Std 1003.2-1992 */
			if (fp) {
				while ((cnt = ck_fread(buf, 1, sizeof buf, fp)) > 0)
					ck_fwrite(buf, 1, cnt, stdout);
				fclose(fp);
			}
		}
		q = p->next;
		free(p);
	}
	bb_common_bss.append_head = bb_common_bss.append_tail = NULL;
}

/* Read in the next line of input, and store it in the pattern space.
   Return zero if there is nothing left to input. */
static flagT read_pattern_space(struct input *input, flagT append)
{
	dump_append_queue();
	bb_common_bss.replaced = 0;
	Gline.chomped = 0;
	if (!append)
		Gline.length = 0;

	while (!(*input->read_fn)(input)) {
		closedown(input);
		if (!*input->file_list)
			return 0;
		open_next_file(*input->file_list++, input);
	}

	++input->line_number;
	return 1;
}

static struct append_queue *next_append_slot(void)
{
	struct append_queue *n = MALLOC(1, struct append_queue);

	n->rfile = NULL;
	n->text = NULL;
	n->textlen = 0;
	n->next = NULL;

	if (bb_common_bss.append_tail)
		bb_common_bss.append_tail->next = n;
	else
		bb_common_bss.append_head = n;
	return bb_common_bss.append_tail = n;
}

/* Execute the program 'vec' on the current input line.
   Return non-zero if caller should quit, 0 otherwise.  */
static flagT execute_program(struct vector *vec, struct input *input)
{
	struct vector *cur_vec;
	struct sed_cmd *cur_cmd;
	size_t n;

	cur_vec = vec;
	cur_cmd = cur_vec->v;
	n = cur_vec->v_length;
	while (n) {
		if (match_address_p(cur_cmd, input)) {
			switch (cur_cmd->cmd) {
			case '{':                             /* Execute sub-program */
				if (cur_cmd->x.sub->v_length) {
					cur_vec = cur_cmd->x.sub;
					cur_cmd = cur_vec->v;
					n = cur_vec->v_length;
					continue;
				}
				break;

			case '}': {
				countT i = cur_vec->return_i;
				cur_vec = cur_vec->return_v;
				cur_cmd = cur_vec->v + i;
				n = cur_vec->v_length - i;
				}
				continue;

			case 'a': {
				struct append_queue *aq = next_append_slot();
				aq->text = cur_cmd->x.cmd_txt.text;
				aq->textlen = cur_cmd->x.cmd_txt.text_len;
				}
				break;

			case 'b':
				if (cur_cmd->x.jump) {
					struct sed_label *j = cur_cmd->x.jump;
					countT i = j->v_index;
					cur_vec = j->v;
					cur_cmd = cur_vec->v + i;
					n = cur_vec->v_length - i;
					continue;
				}
				return 0;

			case 'c':
				Gline.length = 0;
				Gline.chomped = 0;
				if (!cur_cmd->a1_matched)
					output_line(cur_cmd->x.cmd_txt.text,
						cur_cmd->x.cmd_txt.text_len, 0, stdout);
				/* POSIX.2 is silent about c starting a new cycle,
				but it seems to be expected (and make sense). */
				return 0;

			case 'd':
				Gline.length = 0;
				Gline.chomped = 0;
				return 0;

			case 'D': {
				char *p = memchr(Gline.text, '\n', Gline.length);
				if (!p) {
					Gline.length = 0;
					Gline.chomped = 0;
					return 0;
				}
				++p;
				memmove(Gline.text, p, Gline.length);
				Gline.length -= p - Gline.text;

				/* reset to start next cycle without reading a new line: */
				cur_vec = vec;
				cur_cmd = cur_vec->v;
				n = cur_vec->v_length;
				continue;
				}

			case 'g':
				line_copy(&bb_common_bss.hold, &Gline);
				break;

			case 'G':
				line_append(&bb_common_bss.hold, &Gline);
				break;

			case 'h':
				line_copy(&Gline, &bb_common_bss.hold);
				break;

			case 'H':
				line_append(&Gline, &bb_common_bss.hold);
				break;

			case 'i':
				output_line(cur_cmd->x.cmd_txt.text, cur_cmd->x.cmd_txt.text_len, 0, stdout);
				break;

			case 'l':
				do_list();
				break;

			case 'n':
				if (!bb_common_bss.no_default_output)
					output_line(Gline.text, Gline.length, Gline.chomped, stdout);
				if (!read_pattern_space(input, 0))
					return 1;
				break;

			case 'N':
				str_append(&Gline, "\n", 1);
				if (!read_pattern_space(input, 1))
					return 1;
				break;

			case 'p':
				output_line(Gline.text, Gline.length, Gline.chomped, stdout);
				break;

			case 'P': {
				char *p = memchr(Gline.text, '\n', Gline.length);
				output_line(Gline.text, (p ? (size_t)(p - Gline.text) : Gline.length),
						p ? 1 : Gline.chomped, stdout);
				}
				break;

			case 'q':
				return 1;

			case 'r':
				if (cur_cmd->x.rfile) {
					struct append_queue *aq = next_append_slot();
					aq->rfile = cur_cmd->x.rfile;
				}
				break;

			case 's':
				do_subst(&cur_cmd->x.cmd_regex);
				break;

			case 'T':
				if (bb_common_bss.replaced)
					break;
				bb_common_bss.replaced = 1;
				/* */
			case 't':
				if (bb_common_bss.replaced) {
					bb_common_bss.replaced = 0;
					if (cur_cmd->x.jump) {
						struct sed_label *j = cur_cmd->x.jump;
						countT i = j->v_index;
						cur_vec = j->v;
						cur_cmd = cur_vec->v + i;
						n = cur_vec->v_length - i;
						continue;
					}
					return 0;
				}
				break;

			case 'w':
				if (cur_cmd->x.wfile)
					output_line(Gline.text, Gline.length, Gline.chomped, cur_cmd->x.wfile);
				break;

			case 'x': {
				struct line temp;
				memcpy(&temp, &Gline, sizeof Gline);
				memcpy(&Gline, &bb_common_bss.hold, sizeof Gline);
				memcpy(&bb_common_bss.hold, &temp, sizeof Gline);
				}
				break;

			case 'y': {
				unsigned char *p, *e;
				p = (unsigned char *)Gline.text;
				for (e = p + Gline.length; p < e; ++p)
					*p = cur_cmd->x.translate[*p];
				}
				break;

			case ':':
				/* Executing labels is easy. */
				break;

			case '=':
				printf("%lu\n", (unsigned long)input->line_number);
				break;

			default:
				bb_error_msg_and_die("INTERNAL ERROR: Bad cmd %c", cur_cmd->cmd);
			}
		}

		/* these are buried down here so that a "continue" statement can skip them */
		++cur_cmd;
		--n;
	}
	return 0;
}

static void close_all_files(void)
{
	struct fp_list *p, *q;

	for (p = bb_common_bss.file_ptrs; p; p = q) {
		if (p->fp)
			ck_fclose(p->fp);
		free(p->name);
		q = p->link;
		free(p);
	}
	bb_common_bss.file_ptrs = NULL;
}

/* Complain about a programming error and exit. */
static ATTRIBUTE_NORETURN void bad_prog(const char *why)
{
	bb_default_error_retval = 1;
	if (cur_input.name)
		bb_error_msg_and_die("file %s line %lu: %s",
			 cur_input.name, (unsigned long)cur_input.line, why);
	else
		bb_error_msg_and_die("-e expression #%lu, char %lu: %s",
			 (unsigned long)cur_input.string_expr_count,
			 (unsigned long)(prog.cur - (unsigned char *)prog.base), why);
}

/* Make any checks which require the whole program to have been read.
   In particular: this backpatches the jump targets.
   Any cleanup which can be done after these checks is done here also.  */
static void check_final_program(void)
{
	struct sed_label *go;
	struct sed_label *lbl;
	struct sed_label *nxt;

	/* do all "{"s have a corresponding "}"? */
	if (the_program->return_v) {
		cur_input = *the_program->return_v->err_info;       /* for error reporting */
		bad_prog(EXCESS_OPEN_BRACKET);
	}
	free(the_program->err_info);             /* no longer need this */
	the_program->err_info = NULL;

	/* was the final command an unterminated a/c/i command? */
	if (pending_text) {
		bb_common_bss.old_text_buf->text_len = size_buffer(pending_text);
		bb_common_bss.old_text_buf->text = MEMDUP(get_buffer(pending_text), bb_common_bss.old_text_buf->text_len, char);
		free_buffer(pending_text);
		pending_text = NULL;
	}

	for (go = bb_common_bss.jumps; go; go = nxt) {
		for (lbl = bb_common_bss.labels; lbl; lbl = lbl->next)
			if (strcmp(lbl->name, go->name) == 0)
				break;
		if (*go->name && !lbl)
			bb_error_msg_and_die("can't find label for jump to '%s'", go->name);
		go->v->v[go->v_index].x.jump = lbl;
		nxt = go->next;
		free(go->name);
		free(go);
	}
	bb_common_bss.jumps = NULL;

	for (lbl = bb_common_bss.labels; lbl; lbl = lbl->next) {
		free(lbl->name);
		lbl->name = NULL;
	}
}

static struct vector *new_vector(struct error_info *errinfo, struct vector *old_vector)
{
	struct vector *vector = MALLOC(1, struct vector);
	vector->v = NULL;
	vector->v_allocated = 0;
	vector->v_length = 0;
	vector->err_info = MEMDUP(errinfo, 1, struct error_info);
	vector->return_v = old_vector;
	vector->return_i = old_vector ? old_vector->v_length : 0;
	return vector;
}

/* Read the next character from the program.  Return EOF if there isn't
   anything to read.  Keep cur_input.line up to date, so error messages
   can be meaningful. */
static int inchar(void)
{
	int ch = EOF;

	if (prog.cur) {
		if (prog.cur < prog.end)
			ch = *prog.cur++;
	} else if (prog.file) {
		if (!feof(prog.file))
			ch = getc(prog.file);
	}
	if (ch == '\n')
		++cur_input.line;
	return ch;
}

static void read_text(struct text_buf *buf, int leadin_ch)
{
	int ch;

	/* Should we start afresh (as opposed to continue a partial text)? */
	if (buf) {
		if (pending_text)
			free_buffer(pending_text);
		pending_text = init_buffer();
		buf->text = NULL;
		buf->text_len = 0;
		bb_common_bss.old_text_buf = buf;
	}
	/* assert(bb_common_bss.old_text_buf != NULL); */

	if (leadin_ch == EOF)
		return;
	while ((ch = inchar()) != EOF && ch != '\n') {
		if (ch == '\\') {
			if ((ch = inchar()) == EOF) {
				add1_buffer(pending_text, '\n');
				return;
			}
		}
		add1_buffer(pending_text, ch);
	}
	add1_buffer(pending_text, '\n');

	if (!buf)
		buf = bb_common_bss.old_text_buf;
	buf->text_len = size_buffer(pending_text);
	buf->text = MEMDUP(get_buffer(pending_text), buf->text_len, char);
	free_buffer(pending_text);
	pending_text = NULL;
}

static struct sed_cmd *next_cmd_entry(struct vector **vectorp)
{
	struct sed_cmd *cmd;
	struct vector *v;

	v = *vectorp;
	if (v->v_length == v->v_allocated) {
		v->v_allocated += VECTOR_ALLOC_INCREMENT;
		v->v = REALLOC(v->v, v->v_allocated, struct sed_cmd);
	}

	cmd = v->v + v->v_length;
	cmd->a1.addr_type = addr_is_null;
	cmd->a2.addr_type = addr_is_null;
	cmd->a1_matched = 0;
	cmd->addr_bang = 0;

	*vectorp = v;
	return cmd;
}

static int add_then_next(struct buffer *b, int ch)
{
	add1_buffer(b, ch);
	return inchar();
}

static int snarf_char_class(struct buffer *b)
{
	int ch;

	ch = inchar();
	if (ch == '^')
		ch = add_then_next(b, ch);
	if (ch == CLOSE_BRACKET)
		ch = add_then_next(b, ch);
	while (ch != EOF && ch != '\n' && ch != CLOSE_BRACKET) {
		if (ch == OPEN_BRACKET) {
			int prev;
			int delim = ch = add_then_next(b, ch);

			if (delim != '.' && delim != ':' && delim != '=')
				continue;       /* bypass the add_then_next() call at bottom of loop */
			for (prev = ch = add_then_next(b, ch); !(ch == CLOSE_BRACKET && prev == delim); ch = add_then_next(b, ch)) {
				if (ch == EOF || ch == '\n')
					return EOF;
				prev = ch;
			}
		} else if (ch == '\\') {
			ch = inchar();
			if (ch == EOF)
				break;
			if (ch != 'n' && ch != '\n') {
				add1_buffer(b, '\\');
				continue;       /* bypass the add_then_next() call at bottom of loop */
			}
			ch = '\n';
		}
		ch = add_then_next(b, ch);
	}

	/* ensure that a newline here is an error, not a "slash" match: */
	return ch == '\n' ? EOF : ch;
}

/* unget 'ch' so the next call to inchar will return it.   */
static void savchar(int ch)
{
	if (ch == EOF)
		return;
	if (ch == '\n' && cur_input.line > 0)
		--cur_input.line;
	if (prog.cur) {
		if (prog.cur <= (const unsigned char *)prog.base || *--prog.cur != ch)
			bb_error_msg_and_die("called savchar() with unexpected pushback (%x)", (unsigned char)ch);
	} else
		ungetc(ch, prog.file);
}

static struct buffer *match_slash(int slash, flagT regex, flagT keep_backwhack)
{
	struct buffer *b;
	int ch;

	b = init_buffer();
	while ((ch = inchar()) != EOF && ch != '\n' && ch != slash) {
		if (ch == '\\') {
			ch = inchar();
			if (ch == EOF)
				break;
			else if (ch == 'n' && regex)
				ch = '\n';
			else if (ch != '\n' && (ch != slash || keep_backwhack))
				add1_buffer(b, '\\');
		} else if (ch == OPEN_BRACKET && regex) {
			add1_buffer(b, ch);
			ch = snarf_char_class(b);
			if (ch != CLOSE_BRACKET)
				break;
		}
		add1_buffer(b, ch);
	}
	if (ch == slash)
		return b;

	if (ch == '\n')
		savchar(ch);   /* for proper line number in error report */
	if (regex && bb_common_bss.rx_testing) {
		/* This is slightly bogus.  Sed is noticing an ill-formed regexp,
		 * but rx is getting the credit.  Fortunately, this only affects
		 * a few spencer tests. */
		size_t sz = size_buffer(b);
		char *re_txt;
		add1_buffer(b, '\0');
		re_txt = get_buffer(b);
		if (sz > 0 && re_txt[sz] == '\n')
			re_txt[sz] = '\0';
		printf(BAD_REGEX_FMT, re_txt);
		exit(1);
	}
	free_buffer(b);
	return NULL;
}

static regex_t *compile_regex(struct buffer *b, flagT ignore_case, flagT nosub)
{
	/* ``An empty regular expression is equivalent to the last regular
	   expression read'' so we have to keep track of the last regexp read.
	   We keep track the _source_ form of the regular expression because
	   it is possible for different modifiers to be specified. */
	static char *last_re = NULL;
	regex_t *new_regex = MALLOC(1, regex_t);
	size_t sz;
	int cflags;
	int error;

	sz = size_buffer(b);
	if (sz > 0) {
		add1_buffer(b, '\0');
		free(last_re);
		last_re = xstrdup(get_buffer(b));
	} else if (!last_re)
		bad_prog(NO_REGEX);

	cflags = 0;
	if (ignore_case)
		cflags |= REG_ICASE;
	if (nosub)
		cflags |= REG_NOSUB;
	if (bb_common_bss.use_extended_syntax_p)
		cflags |= REG_EXTENDED;
	if ((error = regcomp(new_regex, last_re, cflags))) {
		char msg[1000];
		if (bb_common_bss.rx_testing) {
			printf(BAD_REGEX_FMT, last_re);
			exit(1);
		}
		regerror(error, NULL, msg, sizeof msg);
		bad_prog(msg);
	}

	return new_regex;
}

/* Read the next non-blank character from the program.  */
static int in_nonblank(void)
{
	int ch;
	do
		ch = inchar();
	while (ISBLANK(ch));
	return ch;
}

/* Read an integer value from the program.  */
static countT in_integer(int ch)
{
	countT num = 0;

	while (ISDIGIT(ch)) {
		num = num * 10 + ch - '0';
		ch = inchar();
	}
	savchar(ch);
	return num;
}

/* Try to read an address for a sed command.  If it succeeds,
   return non-zero and store the resulting address in *'addr'.
   If the input doesn't look like an address read nothing
   and return zero.  */
static flagT compile_address(struct addr *addr, int ch)
{
	if (ISDIGIT(ch)) {
		countT num = in_integer(ch);
		ch = in_nonblank();
#if 1
		if (ch != '~') {
			savchar(ch);
		} else {
			countT num2 = in_integer(in_nonblank());
			if (num2 > 0) {
				addr->addr_type = addr_is_mod;
				addr->a.m.offset = num;
				addr->a.m.modulo = num2;
				return 1;
			}
		}
		addr->addr_type = addr_is_num;
		addr->a.addr_number = num;
#else
		if (ch == '~') {
			countT num2 = in_integer(in_nonblank());
			if (num2 == 0)
				bad_prog("Zero step size for ~ addressing is not valid");
			addr->addr_type = addr_is_mod;
			addr->a.m.offset = num;
			addr->a.m.modulo = num2;
		} else {
			savchar(ch);
			addr->addr_type = addr_is_num;
			addr->a.addr_number = num;
		}
#endif
		return 1;
	} else if (ch == '/' || ch == '\\') {
		flagT caseless = 0;
		struct buffer *b;
		addr->addr_type = addr_is_regex;
		if (ch == '\\')
			ch = inchar();
		if (!(b = match_slash(ch, 1, 1)))
			bad_prog("Unterminated address regex");
		ch = in_nonblank();
		if (ch == 'I')  /* lower case i would break existing programs */
			caseless = 1;
		else
			savchar(ch);
		addr->a.addr_regex = compile_regex(b, caseless, 1);
		free_buffer(b);
		return 1;
	} else if (ch == '$') {
		addr->addr_type = addr_is_last;
		return 1;
	}
	return 0;
}

/* read in a filename for a 'r', 'w', or 's///w' command, and
   update the internal structure about files.  The file is
   opened if it isn't already open in the given mode.  */
static void compile_filename(flagT readit, const char **name, FILE **fp)
{
	struct buffer *b;
	int ch;
	char *file_name;
	struct fp_list *p;

	b = init_buffer();
	ch = in_nonblank();
	while (ch != EOF && ch != '\n') {
		add1_buffer(b, ch);
		ch = inchar();
	}
	add1_buffer(b, '\0');
	file_name = get_buffer(b);

	for (p = bb_common_bss.file_ptrs; p; p = p->link)
		if (p->readit_p == readit && strcmp(p->name, file_name) == 0)
			break;

	if (!p) {
		p = MALLOC(1, struct fp_list);
		p->name = xstrdup(file_name);
		p->readit_p = readit;
		p->fp = NULL;
		if (!readit)
			p->fp = ck_fopen(p->name, "Aw");
		p->link = bb_common_bss.file_ptrs;
		bb_common_bss.file_ptrs = p;
	}

	free_buffer(b);
	if (name)
		*name = p->name;
	if (fp)
		*fp = p->fp;
}

/* Store a label (or label reference) created by a ':', 'b', 't' or 'T'
   command so that the jump to/from the label can be backpatched after
   compilation is complete.  */
static struct sed_label *setup_jump(struct sed_label *list, struct sed_cmd *cmd, struct vector *vec)
{
	struct sed_label *ret;
	struct buffer *b;
	int ch;

	b = init_buffer();
	ch = in_nonblank();

	while (ch != EOF && ch != '\n' && !ISBLANK(ch) && ch != ';' && ch != /*{ */ '}') {
		add1_buffer(b, ch);
		ch = inchar();
	}
	savchar(ch);
	add1_buffer(b, '\0');
	ret = MALLOC(1, struct sed_label);
	ret->name = xstrdup(get_buffer(b));
	ret->v = vec;
	ret->v_index = cmd - vec->v;
	ret->next = list;
	free_buffer(b);
	return ret;
}

static flagT mark_subst_opts(struct subst *cmd)
{
	flagT caseless = 0;
	int ch;

	cmd->global = 0;
	cmd->print = 0;
	cmd->numb = 0;
	cmd->wfile = NULL;

	for (;;)
		switch ((ch = in_nonblank())) {
		case 'i':       /* GNU extension */
		case 'I':       /* GNU extension */
			caseless = 1;
			break;

		case 'p':
			cmd->print = 1;
			break;

		case 'g':
			cmd->global = 1;
			break;

		case 'w':
			compile_filename(0, NULL, &cmd->wfile);
			return caseless;

		case '0': case '1': case '2': case '3':
		case '4': case '5': case '6': case '7':
		case '8': case '9':
			if (cmd->numb)
				bad_prog("multiple number options to 's' command");
			cmd->numb = in_integer(ch);
			if (!cmd->numb)
				bad_prog("number option to 's' command may not be zero");
			break;

		case EOF:
		case '\n':
		case ';':
			return caseless;

		default:
			bad_prog("Unknown option to 's'");
/*NOTREACHED*/}
}

/* Read a program (or a subprogram within '{' '}' pairs) in and store
   the compiled form in *'vector'.  Return a pointer to the new vector.  */
static struct vector *compile_program(struct vector *vector)
{
	struct sed_cmd *cur_cmd;
	struct buffer *b;
	struct buffer *b2;
	unsigned char *ustring;
	size_t len;
	int slash;
	int ch;

	if (!vector)
		vector = new_vector(&cur_input, NULL);
	if (pending_text)
		read_text(NULL, '\n');

	for (;;) {
		while ((ch = inchar()) == ';' || ISSPACE(ch))
			;
		if (ch == EOF)
			break;

		cur_cmd = next_cmd_entry(&vector);
		if (compile_address(&cur_cmd->a1, ch)) {
			ch = in_nonblank();
			if (ch == ',') {
				if (!compile_address(&cur_cmd->a2, in_nonblank()))
					bad_prog("Unexpected ','");
				ch = in_nonblank();
			}
		}
		if (cur_cmd->a2.addr_type == addr_is_num && cur_cmd->a1.addr_type == addr_is_num
				&& cur_cmd->a2.a.addr_number < cur_cmd->a1.a.addr_number)
			cur_cmd->a2.a.addr_number = cur_cmd->a1.a.addr_number;
		if (ch == '!') {
			cur_cmd->addr_bang = 1;
			ch = in_nonblank();
			if (ch == '!')
				bad_prog("Multiple '!'s");
		}

		cur_cmd->cmd = ch;
		switch (ch) {
		case '#':
			if (cur_cmd->a1.addr_type != addr_is_null)
				bad_prog(NO_ADDR);
			ch = inchar();
			if (ch == 'n' && bb_common_bss.first_script && cur_input.line < 2)
				if ((prog.base && prog.cur == 2 + (unsigned char *)prog.base)
						|| (prog.file && !prog.base && 2 == ftell(prog.file)))
					bb_common_bss.no_default_output = 1;
			while (ch != EOF && ch != '\n')
				ch = inchar();
			continue;       /* restart the for (;;) loop */

		case '{':
			++vector->v_length;     /* be sure to count this insn */
			vector = cur_cmd->x.sub = new_vector(&cur_input, vector);
			continue;       /* restart the for (;;) loop */

		case '}':
			/* a return insn for subprograms -t */
			if (!vector->return_v)
				bad_prog(EXCESS_CLOSE_BRACKET);
			if (cur_cmd->a1.addr_type != addr_is_null)
				bad_prog( /*{ */ "} doesn't want any addresses");
			ch = in_nonblank();
			if (ch != EOF && ch != '\n' && ch != ';')
				bad_prog(LINE_JUNK);
			++vector->v_length;     /* we haven't counted this insn yet */
			free(vector->err_info); /* no longer need this */
			vector->err_info = NULL;
			vector = vector->return_v;
			continue;       /* restart the for (;;) loop */

		case 'a':
		case 'i':
			if (cur_cmd->a2.addr_type != addr_is_null)
				bad_prog(ONE_ADDR);
			/* Fall Through */
		case 'c':
			ch = in_nonblank();
			if (ch != '\\')
				bad_prog(LINE_JUNK);
			ch = inchar();
			if (ch != '\n' && ch != EOF)
				bad_prog(LINE_JUNK);
			read_text(&cur_cmd->x.cmd_txt, ch);
			break;

		case ':':
			if (cur_cmd->a1.addr_type != addr_is_null)
				bad_prog(": doesn't want any addresses");
			bb_common_bss.labels = setup_jump(bb_common_bss.labels, cur_cmd, vector);
			break;
		case 'b':
		case 't':
		case 'T':
			bb_common_bss.jumps = setup_jump(bb_common_bss.jumps, cur_cmd, vector);
			break;

		case 'q':
		case '=':
			if (cur_cmd->a2.addr_type != addr_is_null)
				bad_prog(ONE_ADDR);
			/* Fall Through */
		case 'd':
		case 'D':
		case 'g':
		case 'G':
		case 'h':
		case 'H':
		case 'l':
		case 'n':
		case 'N':
		case 'p':
		case 'P':
		case 'x':
			ch = in_nonblank();
			if (ch != EOF && ch != '\n' && ch != ';')
				bad_prog(LINE_JUNK);
			break;

		case 'r':
			if (cur_cmd->a2.addr_type != addr_is_null)
				bad_prog(ONE_ADDR);
			compile_filename(1, &cur_cmd->x.rfile, NULL);
			break;
		case 'w':
			compile_filename(0, NULL, &cur_cmd->x.wfile);
			break;

		case 's':
			slash = inchar();
			if (!(b = match_slash(slash, 1, 1)))
				bad_prog(UNTERM_S_CMD);
			if (!(b2 = match_slash(slash, 0, 1)))
				bad_prog(UNTERM_S_CMD);

			cur_cmd->x.cmd_regex.replace_length = size_buffer(b2);
			cur_cmd->x.cmd_regex.replacement = MEMDUP(get_buffer(b2), cur_cmd->x.cmd_regex.replace_length, char);
			free_buffer(b2);

			{
				flagT caseless = mark_subst_opts(&cur_cmd->x.cmd_regex);
				cur_cmd->x.cmd_regex.regx = compile_regex(b, caseless, 0);
			}
			free_buffer(b);
			break;

		case 'y':
			ustring = MALLOC(YMAP_LENGTH, unsigned char);
			for (len = 0; len < YMAP_LENGTH; len++)
				ustring[len] = len;
			cur_cmd->x.translate = ustring;

			slash = inchar();
			if (!(b = match_slash(slash, 1, 0)))
				bad_prog(UNTERM_Y_CMD);
			ustring = (unsigned char *)get_buffer(b);
			for (len = size_buffer(b); len; --len) {
				ch = inchar();
				if (ch == slash)
					bad_prog("strings for y command are different lengths");
				if (ch == '\n')
					bad_prog(UNTERM_Y_CMD);
				if (ch == '\\')
					ch = inchar();
				if (ch == EOF)
					bad_prog(BAD_EOF);
				cur_cmd->x.translate[*ustring++] = ch;
			}
			free_buffer(b);

			if (inchar() != slash || ((ch = in_nonblank()) != EOF && ch != '\n' && ch != ';'))
				bad_prog(LINE_JUNK);
			break;

		case EOF:
			bad_prog(NO_COMMAND);
/*NOTREACHED*/ default: {
			static char unknown_cmd[] = "Unknown command: ``_''";
			unknown_cmd[sizeof(unknown_cmd) - 4] = ch;
			bad_prog(unknown_cmd);
			}
/*NOTREACHED*/  }

		/* this is buried down here so that "continue" statements will miss it */
		++vector->v_length;
	}
	return vector;
}


/* Apply the compiled script to all the named files. */
static countT process_files(char **argv)
{
	static char dash[] = "-";
	static char *stdin_argv[2] = { dash, NULL };
	struct input input;

	input.file_list = stdin_argv;
	if (argv && *argv)
		input.file_list = argv;
	input.bad_count = 0;
	input.line_number = 0;
	input.read_fn = read_always_fail;
	input.fp = NULL;

	line_init(&Gline, INITIAL_BUFFER_SIZE);
	line_init(&bb_common_bss.hold, INITIAL_BUFFER_SIZE);
	line_init(&bb_common_bss.irl_buffer, FREAD_BUFFER_SIZE);

	while (read_pattern_space(&input, 0)) {
		flagT quit = execute_program(the_program, &input);
		if (!bb_common_bss.no_default_output)
			output_line(Gline.text, Gline.length, Gline.chomped, stdout);
		if (quit)
			break;
	}
	closedown(&input);
	return input.bad_count;
}

/* 'str' is a string (from the command line) that contains a sed command.
   Compile the command, and add it to the end of 'cur_program'. */
static struct vector *compile_string(struct vector *cur_program, char *str)
{
	static countT string_expr_count;
	struct vector *ret;

	prog.file = NULL;
	prog.base = str;
	prog.cur = (unsigned char *)str;
	prog.end = prog.cur + strlen(str);

	cur_input.line = 0;
	cur_input.name = NULL;
	cur_input.string_expr_count = ++string_expr_count;

	ret = compile_program(cur_program);
	prog.base = NULL;
	prog.cur = NULL;
	prog.end = NULL;

	bb_common_bss.first_script = 0;
	return ret;
}

/* 'cmdfile' is the name of a file containing sed commands.
   Read them in and add them to the end of 'cur_program'. */
static struct vector *compile_file(struct vector *cur_program, const char *cmdfile)
{
	size_t len;
	struct vector *ret;

	prog.base = NULL;
	prog.cur = NULL;
	prog.end = NULL;
	prog.file = ck_fopen(cmdfile, "A-r");
	if (map_file(prog.file, &prog.base, &len)) {
		prog.cur = prog.base;
		prog.end = prog.cur + len;
	}

	cur_input.line = 1;
	cur_input.name = cmdfile;
	cur_input.string_expr_count = 0;

	ret = compile_program(cur_program);
	if (prog.base) {
		unmap_file(prog.base, len);
		prog.base = NULL;
		prog.cur = NULL;
		prog.end = NULL;
	}
	if (prog.file != stdin)
		ck_fclose(prog.file);
	prog.file = NULL;

	bb_common_bss.first_script = 0;
	return ret;
}

#define COPYRIGHT_NOTICE "Copyright (C) 1998 Free Software Foundation, Inc."


static void process_e_opt(char *o)
{
	the_program = compile_string(the_program, o);
}

static void process_f_opt(char *o)
{
	the_program = compile_file(the_program, o);
}

int sed_main(void)
{
#if ENABLE_FEATURE_LONG_OPTIONS
	static struct option longopts[] = {
		{"rxtest", 0, NULL, 'r'},
		{"expression", 1, NULL, 'e'},
		{"file", 1, NULL, 'f'},
		{"quiet", 0, NULL, 'n'},
		{"silent", 0, NULL, 'n'},
		{"version", 0, NULL, 'V'},
		{NULL, 0, NULL, 0}
	};
#endif

	int opt;
	flagT bad_input; /* If this variable is non-zero at exit, one or
			    more of the input files couldn't be opened. */

	bb_common_bss.POSIXLY_CORRECT = (getenv("POSIXLY_CORRECT") != NULL);
	bb_common_bss.first_script = 1;

	bb_default_error_retval = 4;
#if ENABLE_FEATURE_LONG_OPTIONS
	bb_applet_long_options = longopts;
#endif
	opt = bb_xgetopt_ulflags("nrVe:f:\0=ef", process_e_opt, process_f_opt);
	if((opt & 1) && !bb_common_bss.no_default_output)       /* may be set bu process_f_opt() */
		bb_common_bss.no_default_output = 1;
	bb_common_bss.use_extended_syntax_p = bb_common_bss.rx_testing = opt & 2;
	if((opt & 4)) {
		printf("GNU sed version %s\n\n", VERSION);
		printf("%s\n\
This is free software; see the source for copying conditions.  There is NO\n\
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE,\n\
to the extent permitted by law.\n", COPYRIGHT_NOTICE);
		return 0;
	}

	if (!the_program) {
		if (*Gargv) {
			the_program = compile_string(the_program, *Gargv);
			Gargv++;
		} else {
			bb_show_usage();
		}
	}
	check_final_program();

	bad_input = process_files(Gargv);

	close_all_files();
	ck_fclose(stdout);
	if (bad_input)
		return 2;
	return 0;
}
#endif
