/* gpio_poll.c
   Public Domain
*/

#define _GNU_SOURCE

#include "libbb.h"
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <time.h>
#include <poll.h>
#include <string.h>
#include <syslog.h>
#include "bb_sig.h"
#include "config_read.h"

#define SYS_CLASS_GPIO "/sys/class/gpio"

#define CFG_FILE "/etc/gpio_buttons.cfg"

static const char *cfg_file = CFG_FILE;

struct GPIO_CFG {
	int pin;
	char *name;
	int initial;

	int fd;
	int prev;
	time_t t;
	int trigger;

	char **call;

	struct GPIO_CFG *next;
};

static struct GPIO_CFG *pins;


static int config_safe_int(const char *str, int max_val, const char *name, int line)
{
	return bb_xparse_i_number(BB_PN_SET_UPPER | BB_PN_ERR_STRV,
				  str, max_val,
				  "%s:%d: strange %s number '%s', need 0..%d",
				  cfg_file, line, name, str, max_val);
}

void bb_show_usage (void)
{
	const char *a = bb_applet_name;
	bb_applet_name = NULL;
	bb_default_error_retval = 2;
	bb_error_msg_and_die("Usage: %s [-f(foreground)] [-c config_file]", a);
}

/* SIGCHLD: one of our children has died. */
static void chld_handler(int ATTRIBUTE_UNUSED sig)
{
	int     saved_errno = errno;

	while(waitpid(-1, NULL, WNOHANG) != 0)
		if (errno == ECHILD) break;
	errno = saved_errno;
}

static void xlog_info(const char* fmt, ...)
{
	va_list args;
	int prev_log_syslog = log_syslog;

	if(log_syslog)
		log_syslog = LOG_INFO;
	va_start(args, fmt);
	bb_verror_msg(fmt, args);
	va_end(args);
	log_syslog = prev_log_syslog;
}

int main(int argc, char **argv)
{
   FILE *f;
   char str[256];
   struct pollfd *pfd;
   int i, fd, npins;
   char buf[8];
   struct GPIO_CFG *s = NULL;
   config_read_t *crt;
   int foreground;

   libbb_init(argc, argv);
   foreground = bb_xgetopt_ulflags("fc:\0?0", &cfg_file) & 1;

   crt = config_read_prepare(cfg_file, 'e');
   while( (crt = config_read(crt, "#~\0\0")) != NULL) {
	if(crt->ntokens < 2)
		bb_error_msg_and_die("%s:%d: strange line, need key and values", cfg_file, crt->num_line);
	if(!strcmp(crt->tokens[0], "pin")) {
		if(crt->ntokens != 2)
			bb_error_msg_and_die("%s:%d: strange line, need pin pinNumber", cfg_file, crt->num_line);
		i = config_safe_int(crt->tokens[1], 1023, "pin", crt->num_line);
		for(s = pins; s != NULL; s = s->next)
			if(s->pin == i)
				bb_error_msg_and_die("%s:%d: pin %s defined already", cfg_file, crt->num_line, crt->tokens[1]);
		s = zmalloc(sizeof(struct GPIO_CFG));
		s->pin = i;
		s->initial = 1;
		s->fd = 0;
		s->next = pins;
		pins = s;
		continue;
	}
	if(s == NULL)
		bb_error_msg_and_die("%s:%d: except pin pinNumber", cfg_file, crt->num_line);
	if(!strcmp(crt->tokens[0], "name")) {
		if(crt->ntokens != 2)
			bb_error_msg_and_die("%s:%d: strange line, need name word", cfg_file, crt->num_line);
		if(s->name)
			bb_error_msg_and_die("%s:%d: name for pin %d defined already", cfg_file, crt->num_line, s->pin);
		s->name = xstrdup(crt->tokens[1]);
		continue;
	}
	if(!strcmp(crt->tokens[0], "initial")) {
		if(crt->ntokens != 2)
			bb_error_msg_and_die("%s:%d: strange line, need initial 0|1", cfg_file, crt->num_line);
		if((s->fd & 1))
			bb_error_msg_and_die("%s:%d: initial for pin %d defined already", cfg_file, crt->num_line, s->pin);
		s->initial = config_safe_int(crt->tokens[1], 1, "initial", crt->num_line);
		s->fd |= 1;
		continue;
	}
	if(!strcmp(crt->tokens[0], "call")) {
		if(crt->ntokens < 2)
			bb_error_msg_and_die("%s:%d: strange line, need %s command [args]",
						cfg_file, crt->num_line, "call");
		if(s->call)
			bb_error_msg_and_die("%s:%d: call for pin %d defined already", cfg_file, crt->num_line, s->pin);
		if(crt->tokens[1][0] != '/')
			bb_error_msg_and_die("%s:%d: call for pin need absolute pathname", cfg_file, crt->num_line);
		s->call = zmalloc(sizeof(char *) * crt->ntokens);
		for(i = 0; i < crt->ntokens - 1; i++)
			s->call[i] = xstrdup(crt->tokens[i + 1]);
		s->call[i] = NULL;
		continue;
	}
	bb_error_msg_and_die("%s:%d: can not parse line", cfg_file, crt->num_line);
   }
   if(pins == NULL)
	bb_error_msg_and_die("%s: have not pins", cfg_file);

   /* Prior loop assumed. */

   xchdir(SYS_CLASS_GPIO);

   f = bb_xfopen("export", "Aw");
   for(i = 0, s = pins; s != NULL; i++, s = s->next) {
	fprintf(f, "%d\n", s->pin);
	fflush(f);

	sprintf(str, "gpio%d/edge", s->pin);
	if ((fd = open(str, O_WRONLY)) < 0)
		bb_perror_msg_and_die("Failed open %s/%s", SYS_CLASS_GPIO, str);
	write(fd, "both\n", 5);
	close(fd);
   }
   fclose(f);

   npins = i;
   pfd = xmalloc(sizeof(struct pollfd) * npins);

   for(i = 0, s = pins; s != NULL; i++, s = s->next) {
	sprintf(str, "gpio%d/value", s->pin);

	if ((fd = open(str, O_RDONLY)) < 0)
		bb_perror_msg_and_die("Failed open %s/%s", SYS_CLASS_GPIO, str);
	s->fd = pfd[i].fd = fd;
	pfd[i].events = POLLPRI;
	pfd[i].revents = 0;
	s->prev = s->initial;
	if(s->name == NULL)
		s->name = "(undefined)";
   }
   if(!foreground) {
	bb_openlog(LOG_PID, LOG_DAEMON);
	xdaemon(0);
   }
   bb_signal((1 << SIGCHLD), chld_handler);

   while(1) {
	poll(pfd, npins, -1);      /* wait for interrupt */
	for(i = 0, s = pins; i < npins; i++, s = s->next) {
		if(pfd[i].revents != 0) {
			time_t ct;

			pfd[i].revents = 0;
			fd = pfd[i].fd;

			lseek(fd, 0, SEEK_SET);    /* consume interrupt */
			read(fd, buf, sizeof buf);
			fd = buf[0] - '0';
			if(fd == s->prev)
				continue;
			s->prev = fd;
			time(&ct);
			if(fd != s->initial) {
				/* pressed */
				s->t = ct;
			} else {
				/* released */
				long long df = (long long)ct - (long long)s->t;
				const char *t;

				if(df <= 0)
					df = 1;
				sprintf(str, "%d", (int)df);
				s->trigger ^= 1;
				t = s->trigger == 0 ? "even" : "odd";
				if(s->call) {
					fd = fork();
					if(fd < 0)
						bb_perror_msg("can not fork()");
					if(fd == 0) {
						char *p = s->call[0];

						setenv("NAME", s->name, 1);
						setenv("TRIGGER", t, 1);
						setenv("RELTIME", str, 1);
						s->call[0] = strrchr(p, '/') + 1;
						xlog_info("pin[%d]=%s released=%s sec=%s call=%s",
								s->pin, t, s->name, str, s->call[0]);
						execv(p, s->call);
						bb_perror_msg_and_die("can not call %s", p);
					}
				} else {
					xlog_info("pin[%d]=%s released=%s sec=%s", s->pin, t, s->name, str);
				}
			}
		}
	}
   }

   return 0;
}
