/* find - look for files satisfying a predicate       Author: E. Baalbergen */

/* Original author: Erik Baalbergen; POSIX compliant version: Bert Laverman */

/*
	Stolen from Minix. Busybox port by
	Vladimir Oleynik (C) 2004 <dzo@simtreas.ru>
*/
#ifdef APPLET_USAGE
#define find_trivial_usage \
	"[PATH...] [EXPRESSION]"
#define find_full_usage \
	"Search for files in a directory hierarchy. The default PATH is\n" \
	"the current directory; default EXPRESSION is '-print'\n" \
	"\nEXPRESSION may consist of:\n" \
	"\t-follow\t\tDereference symbolic links\n" \
	"\t-name PATTERN\tFile name (leading directories removed) matches PATTERN\n" \
	"\t-print\t\tPrint (default and assumed)\n" \
	"\t-type X\t\tFiletype matches X (where X is one of: f,d,l,b,c,...)\n" \
	"\t-perm PERMS\tPermissions match any of (+NNN); all of (-NNN);\n\t\t\tor exactly (NNN)\n" \
	"\t-mtime DAYS\tModified time is greater than (+N); less than (-N);\n\t\t\tor exactly (N) days\n" \
	"\t-mmin MINS\tModified time is greater than (+N); less than (-N);\n\t\t\tor exactly (N) minutes\n" \
	"\t-newer FILE\tModified time is more recent than FILE's\n" \
	"\t-inum N\t\tFile has inode number N\n" \
	"\t-exec CMD\tExecute CMD with all instances of {} replaced by the\n" \
	"\t\t\tfiles matching EXPRESSION"
#else

#include "busybox.h"
#include <sys/wait.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <dirent.h>

#include "bb_xopen.h"

/*######################## DEFINITIONS ##############################*/

#define MAXARG          256     /* maximum length for an argv for -exec  */
#define BSIZE           512     /* POSIX wants 512 byte blocks           */
#define SECS_PER_DAY    (24L*60L*60L)   /* check your planet             */


/* For -perm with symbolic modes: */
#define ISWHO(c)        ((c == 'u') || (c == 'g') || (c == 'o') || (c == 'a'))
#define ISOPER(c)       ((c == '-') || (c == '=') || (c == '+'))
#define ISMODE(c)       ((c == 'r') || (c == 'w') || (c == 'x') || \
			 (c == 's') || (c == 't'))
#define MUSER           1
#define MGROUP          2
#define MOTHERS         4


struct exec {
  int e_cnt;
  char *e_vec[MAXARG];
};

struct node {
  int n_type;                   /* any OP_ or NOT */
  union {
	char *n_str;
	struct {
		long n_val;
		int n_sign;
	} n_int;
	struct exec *n_exec;
	struct {
		struct node *n_left, *n_right;
	} n_opnd;
  } n_info;
};

struct oper {
  const char *op_str;
};


static const struct oper ops[] = {
  { "name" },
#define OP_NAME          0      /* match name                            */
  { "perm" },
#define OP_PERM          1      /* check file permission bits            */
  { "type" },
#define OP_TYPE          2      /* check file type bits                  */
  { "links" },
#define OP_LINKS         3      /* check link count                      */
  { "user" },
#define OP_USER          4      /* check owner                           */
  { "group" },
#define OP_GROUP         5      /* check group ownership                 */
  { "size" },
#define OP_SIZE          6      /* check size, blocks or bytes           */
  { "inum" },
#define OP_INUM          7      /* compare inode number                  */
  { "atime" },
#define OP_ATIME         8      /* check last access time                */
  { "ctime" },
#define OP_CTIME         9      /* check creation time                   */
  { "mtime" },
#define OP_MTIME        10      /* check last modification time          */
  { "exec" },
#define OP_EXEC         11      /* execute command                       */
  { "ok" },
#define OP_OK           12      /* execute with confirmation             */
  { "print" },
#define OP_PRINT        13      /* print name                            */
  { "print0" },
#define OP_PRINT0       14      /* print name null terminated            */
  { "newer" },
#define OP_NEWER        15      /* compare modification times            */
  { "a" },
#define OP_AND          16      /* logical and (short circuit)           */
  { "o" },
#define OP_OR           17      /* logical or (short circuit)            */
  { "xdev" },
#define OP_XDEV         18      /* do not cross file-system boundaries   */
  { "depth" },
#define OP_DEPTH        19      /* descend directory before testing      */
  { "prune" },
#define OP_PRUNE        20      /* don't descend into current directory  */
  { "nouser" },
#define OP_NOUSER       21      /* check validity of user id             */
  { "nogroup" },
#define OP_NOGROUP      22      /* check validity of group id            */
  { NULL }
#define NONE            23      /* not a valid predicate                 */
};

/* Some return values: */
#define OP_SIZEC        24      /* this is a fake for -size with 'c'     */
#define LPAR            25      /* left parenthesis                      */
#define RPAR            26      /* right parenthesis                     */
#define NOT             27      /* logical not                           */
#define EOI             -1      /* end of expression                     */


CALC_BSS struct BB_COMMON_BSS {
	char **ipp;                     /* pointer to next argument during parsing       */
	long current_time;              /* for computing age                             */
	int tty;                        /* fd for /dev/tty when using -ok                */
	int xdev_flag;                  /* cross device boundaries?                      */
	int devnr;                      /* device nr of first inode                      */
	int depth_flag;                 /* descend before check?                         */
	int prune_here;                 /* This is Baaaad! Don't ever do this again!     */
	int um;                         /* current umask()                               */
	int noneedprint;                /* nop implicit -print needed?                   */
} bb_common_bss;
#define epath bb_common_bss.epath
#define ipp bb_common_bss.ipp
#define current_time bb_common_bss.current_time
#define tty bb_common_bss.tty
#define xdev_flag bb_common_bss.xdev_flag
#define devnr bb_common_bss.devnr
#define depth_flag bb_common_bss.depth_flag
#define prune_here bb_common_bss.prune_here
#define um bb_common_bss.um
#define noneedprint bb_common_bss.noneedprint


static int ichk(long val, struct node *n)
{
  switch (n->n_info.n_int.n_sign) {
    case 0:
	return val == n->n_info.n_int.n_val;
    case 1:
	return val > n->n_info.n_int.n_val;
    case -1:    return val < n->n_info.n_int.n_val;
  }
  bb_error_msg_and_die("internal: bad n_sign");
}

static int lex(const char *str)
{
  if (str == (char *) NULL) return EOI;
  if (*str == '-') {
	const struct oper *op = find_string_from_struct(ops, sizeof(struct oper), str + 1);
	return (op-ops);
  }
  if (str[1] == 0) {
	switch (*str) {
	  case '(':
		return LPAR;
	  case ')':
		return RPAR;
	  case '!':     return NOT;
	}
  }
  return NONE;
}

/*################### SMATCH #########################*/
/* Don't try to understand the following one... */
static int smatch(const char *s, const char *t)         /* shell-like matching */
{
  int n;

  if (*t == '\0') return *s == '\0';
  if (*t == '*') {
	++t;
	do
		if (smatch(s, t)) return 1;
	while (*s++ != '\0');
	return 0;
  }
  if (*s == '\0') return 0;
  if (*t == '\\') return (*s == *++t) ? smatch(++s, ++t) : 0;
  if (*t == '?') return smatch(++s, ++t);
  if (*t == '[') {
	while (*++t != ']') {
		if (*t == '\\') ++t;
		if (*(t + 1) != '-') {
			if (*t == *s) {
				while (*++t != ']')
					if (*t == '\\') ++t;
				return smatch(++s, ++t);
			} else {
				continue;
			}
		}
		if (*(t + 2) == ']') return(*s == *t || *s == '-');
		n = (*(t + 2) == '\\') ? 3 : 2;
		if (*s >= *t && *s <= *(t + n)) {
			while (*++t != ']')
				if (*t == '\\') ++t;
			return smatch(++s, ++t);
		}
		t += n;
	}
	return 0;
  }
  return(*s == *t) ? smatch(++s, ++t) : 0;
}

static int execute(int op, struct exec *e, char *path)
{
  int s, pid;
  char *argv[MAXARG];
  char **p, **q;

  for (p = e->e_vec, q = argv; *p;)     /* replace the {}s */
	if ((*q++ = *p++) == (char *) -1) q[-1] = path;
  *q = NULL;
  if (op == OP_OK) {
	char answer[10];

	for (p = &argv[1]; *p; p++) {
		dprintf(tty, "%s ", *p);
	}
	dprintf(tty, "? ");
	if (read(tty, answer, 10) < 2 || *answer != 'y') return 0;
  }
  if ((pid = fork()) == -1) bb_perror_msg_and_die("can't fork");
  if (pid == 0) {
	int i = 3;

	while (close(i++) == 0) {
	}                       /* wow!!! */
	execvp(argv[1], &argv[1]);       /* binary itself? */
	execv(DEFAULT_SHELL, &argv[0]);       /* shell command? */
	bb_perror_msg("exec failure: %s", argv[1]);       /* none of them! */
	exit(127);
  }
  return wait(&s) == pid && s == 0;
}

static int check(char *path, struct stat *st, struct node *n, const char *last)
{
  if (n == (struct node *) NULL) return 1;
  switch (n->n_type) {
    case OP_AND:
	return check(path, st, n->n_info.n_opnd.n_left, last) &&
		check(path, st, n->n_info.n_opnd.n_right, last);
    case OP_OR:
	return check(path, st, n->n_info.n_opnd.n_left, last) ||
		check(path, st, n->n_info.n_opnd.n_right, last);
    case NOT:
	return !check(path, st, n->n_info.n_opnd.n_left, last);
    case OP_NAME:
	return smatch(last, n->n_info.n_str);
    case OP_PERM:
	if (n->n_info.n_int.n_sign < 0)
		return(st->st_mode & (int) n->n_info.n_int.n_val) ==
			(int) n->n_info.n_int.n_val;
	return(st->st_mode & 07777) == (int) n->n_info.n_int.n_val;
    case OP_NEWER:
	return st->st_mtime > n->n_info.n_int.n_val;
    case OP_TYPE:
	return(st->st_mode & S_IFMT) == (mode_t) n->n_info.n_int.n_val;
    case OP_LINKS:
	return ichk((long) (st->st_nlink), n);
    case OP_USER:
	return st->st_uid == n->n_info.n_int.n_val;
    case OP_GROUP:
	return st->st_gid == n->n_info.n_int.n_val;
    case OP_SIZE:
	return ichk((st->st_size == 0) ? 0L :
		    (long) ((st->st_size - 1) / BSIZE + 1), n);
    case OP_SIZEC:
	return ichk((long) st->st_size, n);
    case OP_INUM:
	return ichk((long) (st->st_ino), n);
    case OP_ATIME:
	return ichk(st->st_atime, n);
    case OP_CTIME:
	return ichk(st->st_ctime, n);
    case OP_MTIME:
	return ichk(st->st_mtime, n);
    case OP_EXEC:
    case OP_OK:
	return execute(n->n_type, n->n_info.n_exec, path);
    case OP_PRINT:
	printf("%s\n", path);
	return 1;
    case OP_PRINT0:
	printf("%s", path); putchar(0);
	return 1;
    case OP_XDEV:
    case OP_DEPTH:
	return 1;
    case OP_PRUNE:
	prune_here = 1;
	return 1;
    case OP_NOUSER:
	return(getpwuid(st->st_uid) == (struct passwd *) NULL);
    case OP_NOGROUP:
	return(getgrgid(st->st_gid) == (struct group *) NULL);
  }
  bb_error_msg_and_die("ILLEGAL NODE");
  return 0;                     /* Never reached */
}

static struct node * newnode(int t)
{
  struct node *n = (struct node *) xmalloc(sizeof(struct node));

  n->n_type = t;
  return n;
}

static void __attribute__ ((noreturn)) serr(const char *m)
{
  bb_error_msg_and_die("syntax error: %s", m);
}

/* Isnumber checks correct number syntax. A sign is allowed, but the '+'
 * only if the number is to be in decimal.
 */
static int isnumber(const char *str, int base, int sign)
{
  if (sign && ((*str == '-') || ((base == 8) && (*str == '+')))) str++;
  while ((*str >= '0') && (*str < ('0' + base))) str++;
  return(*str == '\0' ? 1 : 0);
}

/* Convert a string to an integer, storing sign info in *ps. */
static void number(const char *str, int base, long *pl, int *ps)
{
  int up = '0' + base - 1;
  long val = 0;

  *ps = ((*str == '-' || *str == '+') ? ((*str++ == '-') ? -1 : 1) : 0);
  while (*str >= '0' && *str <= up) val = base * val + *str++ - '0';
  if (*str)
	serr("illegal numeric value");
  *pl = val;
}

static void domode(int op, int *mode, int bits)
{
  switch (op) {
    case '-':
	*mode &= ~bits;
	break;                  /* clear bits */
    case '=':
	*mode |= bits;
	break;                  /* set bits */
    case '+':
	*mode |= (bits & ~um);  /* set, but take umask in account */
  }
}

static void fmode(const char *str, long *pl, int *ps)
{
  int m = 0, w, op;
  const char *p = str;

  if (*p == '-') {
	*ps = -1;
	p++;
  } else
	*ps = 0;

  while (*p) {
	w = 0;
	if (ISOPER(*p))
		w = MUSER | MGROUP | MOTHERS;
	else if (!ISWHO(*p))
		bb_error_msg_and_die("u, g, o, or a expected: %s", p);
	else {
		while (ISWHO(*p)) {
			switch (*p) {
			  case 'u':
				w |= MUSER;
				break;
			  case 'g':
				w |= MGROUP;
				break;
			  case 'o':
				w |= MOTHERS;
				break;
			  case 'a':
				w = MUSER | MGROUP | MOTHERS;
			}
			p++;
		}
		if (!ISOPER(*p)) bb_error_msg_and_die("-, + or = expected: %s", p);
	}
	op = *p++;
	while (ISMODE(*p)) {
		switch (*p) {
		  case 'r':
			if (w & MUSER) domode(op, &m, S_IRUSR);
			if (w & MGROUP) domode(op, &m, S_IRGRP);
			if (w & MOTHERS) domode(op, &m, S_IROTH);
			break;
		  case 'w':
			if (w & MUSER) domode(op, &m, S_IWUSR);
			if (w & MGROUP) domode(op, &m, S_IWGRP);
			if (w & MOTHERS) domode(op, &m, S_IWOTH);
			break;
		  case 'x':
			if (w & MUSER) domode(op, &m, S_IXUSR);
			if (w & MGROUP) domode(op, &m, S_IXGRP);
			if (w & MOTHERS) domode(op, &m, S_IXOTH);
			break;
		  case 's':
			if (w & MUSER) domode(op, &m, S_ISUID);
			if (w & MGROUP) domode(op, &m, S_ISGID);
			break;
		  case 't':
			domode(op, &m, S_ISVTX);
		}
		p++;
	}
	if (*p) {
		if (*p == ',')
			p++;
		else
			bb_error_msg_and_die("garbage at end of mode string: %s", p);
	}
  }
  *pl = m;
}

static void find(const char *path, struct node *pred, const char *last)
{
  char spath[PATH_MAX];
  char *s_end = spath;
  struct stat st;
  DIR *dp;
  struct dirent *de;

  if (path[1] == '\0' && *path == '/') {
	*s_end++ = '/';
	*s_end = '\0';
  } else {
	while ((*s_end++ = *path++) != 0) {
	}
  }

  if (lstat(spath, &st) == -1)
	bb_perror_msg("can't get status of %s", spath);
  else {
	switch (xdev_flag) {
	  case 0:
		break;
	  case 1:
		if (st.st_dev != devnr) return;
		break;
	  case 2:               /* set current device number */
		xdev_flag = 1;
		devnr = st.st_dev;
		break;
	}

	prune_here = 0;
	if (!depth_flag && check(spath, &st, pred, last) && !noneedprint)
		printf("%s\n", spath);
	if (!prune_here && (st.st_mode & S_IFMT) == S_IFDIR) {
		if ((dp = opendir(spath)) == NULL) {
			bb_perror_msg("can't read directory %s", spath);
			return;
		}
		s_end[-1] = '/';
		while ((de = readdir(dp)) != NULL) {
			if ((de->d_name[0] != '.') || ((de->d_name[1])
					  && ((de->d_name[1] != '.')
					      || (de->d_name[2])))) {
				strcpy(s_end, de->d_name);
				find(spath, pred, s_end);
			}
		}
		closedir(dp);
	}
	if (depth_flag) {
		s_end[-1] = '\0';
		if (check(spath, &st, pred, last) && !noneedprint)
			printf("%s\n", spath);
	}
  }
}

/*########################### PARSER ###################################*/
/* Grammar:
 * expr        : primary | primary OR expr;
 * primary     : secondary | secondary AND primary | secondary primary;
 * secondary   : NOT secondary | LPAR expr RPAR | simple;
 * simple      : -OP args...
 */

static struct node *secondary(int t);

static struct node *primary(int t)
{
  struct node *nd, *p, *nd2;

  nd = secondary(t);
  if ((t = lex(*++ipp)) != OP_AND) {
	ipp--;
	if (t == EOI || t == RPAR || t == OP_OR) return nd;
  }
  nd2 = primary(lex(*++ipp));
  p = newnode(OP_AND);
  p->n_info.n_opnd.n_left = nd;
  p->n_info.n_opnd.n_right = nd2;
  return p;
}

static struct node *expr(int t)
{
  struct node *nd, *p, *nd2;

  nd = primary(t);
  if ((t = lex(*++ipp)) == OP_OR) {
	nd2 = expr(lex(*++ipp));
	p = newnode(OP_OR);
	p->n_info.n_opnd.n_left = nd;
	p->n_info.n_opnd.n_right = nd2;
	return p;
  }
  ipp--;
  return nd;
}

static void checkarg(char *arg)
{
  if (arg == 0)
	serr("argument expected");
}

static struct node * simple(int t)
{
  struct node *p = newnode(t);
  struct exec *e;
  struct stat est;
  struct passwd *pw;
  struct group *gr;
  long l;
  int i;

  switch (t) {
    case OP_TYPE:
	checkarg(*++ipp);
	switch (**ipp) {
	  case 'b':
		p->n_info.n_int.n_val = S_IFBLK;
		break;
	  case 'c':
		p->n_info.n_int.n_val = S_IFCHR;
		break;
	  case 'd':
		p->n_info.n_int.n_val = S_IFDIR;
		break;
	  case 'f':
		p->n_info.n_int.n_val = S_IFREG;
		break;
	  case 'l':
#ifdef S_IFLNK
		p->n_info.n_int.n_val = S_IFLNK;
#else
		p->n_info.n_int.n_val = ~0;     /* Always unequal. */
#endif
		break;
	  default:
		bb_error_msg_and_die("-type needs b, c, d, f or l");
	}
	break;
    case OP_USER:
	checkarg(*++ipp);
	if ((pw = getpwnam(*ipp)) == NULL && isnumber(*ipp, 10, 0))
		number(*ipp, 10, &(p->n_info.n_int.n_val),
		       &(p->n_info.n_int.n_sign));
	else {
		if (pw == NULL)
			bb_error_msg_and_die("unknown user: %s", *ipp);
		p->n_info.n_int.n_val = pw->pw_uid;
		p->n_info.n_int.n_sign = 0;
	}
	break;
    case OP_GROUP:
	checkarg(*++ipp);
	if (((gr = getgrnam(*ipp)) == NULL)
	    && isnumber(*ipp, 10, 0))
		number(*ipp, 10, &(p->n_info.n_int.n_val),
		       &(p->n_info.n_int.n_sign));
	else {
		if (gr == NULL)
			bb_error_msg_and_die("unknown group: %s", *ipp);
		p->n_info.n_int.n_val = gr->gr_gid;
		p->n_info.n_int.n_sign = 0;
	}
	break;
    case OP_SIZE:
	checkarg(*++ipp);
	i = strlen(*ipp) - 1;
	if ((*ipp)[i] == 'c') {
		p->n_type = OP_SIZEC;   /* Count in bytes i.s.o. blocks */
		(*ipp)[i] = '\0';
	}
	number(*ipp, 10, &(p->n_info.n_int.n_val),
	       &(p->n_info.n_int.n_sign));
	break;
    case OP_LINKS:
    case OP_INUM:
	checkarg(*++ipp);
	number(*ipp, 10, &(p->n_info.n_int.n_val),
	       &(p->n_info.n_int.n_sign));
	break;
    case OP_PERM:
	checkarg(*++ipp);
	if (isnumber(*ipp, 8, 1)) number(*ipp, 8, &(p->n_info.n_int.n_val),
		       &(p->n_info.n_int.n_sign));
	else
		fmode(*ipp, &(p->n_info.n_int.n_val),
		      &(p->n_info.n_int.n_sign));
	break;
    case OP_ATIME:
    case OP_CTIME:
    case OP_MTIME:
	checkarg(*++ipp);
	number(*ipp, 10, &l, &(p->n_info.n_int.n_sign));
	p->n_info.n_int.n_val = current_time - l * SECS_PER_DAY;
	/* More than n days old means less than the absolute time */
	p->n_info.n_int.n_sign *= -1;
	break;
    case OP_EXEC:
    case OP_OK:
	checkarg(*++ipp);
	e = (struct exec *) xmalloc(sizeof(struct exec));
	e->e_cnt = 1;
	e->e_vec[0] = (char *)DEFAULT_SHELL_SHORT_NAME;
	p->n_info.n_exec = e;
	while (*ipp) {
		if (**ipp == ';' && (*ipp)[1] == '\0') {
			e->e_vec[e->e_cnt] = 0;
			break;
		}
		e->e_vec[(e->e_cnt)++] =
			(**ipp == '{' && (*ipp)[1] == '}'
		       && (*ipp)[2] == '\0') ? (char *) (-1) : *ipp;
		ipp++;
	}
	if (*ipp == 0) bb_error_msg_and_die("-exec/-ok: ; missing");
	if (t == OP_OK)
		tty = bb_xopen(bb_dev_tty, O_RDWR | BB_OPEN_ERR);
	break;
    case OP_NEWER:
	checkarg(*++ipp);
	if (lstat(*ipp, &est) == -1)
		bb_perror_msg_and_die("-newer: can't get status of %s", *ipp);
	p->n_info.n_int.n_val = est.st_mtime;
	break;
    case OP_NAME:
	checkarg(*++ipp);
	p->n_info.n_str = *ipp;
	break;
    case OP_XDEV:       xdev_flag = 1;  break;
    case OP_DEPTH:      depth_flag = 1; break;
    case OP_PRUNE:
    case OP_PRINT:
    case OP_PRINT0:
    case OP_NOUSER:     case OP_NOGROUP:        break;
	  default:
		serr("operator expected");
  }
  if ((t == OP_PRINT) || (t == OP_PRINT0) || (t == OP_EXEC) || (t == OP_OK))
	noneedprint = 1;

  return p;
}

static struct node *secondary(int t)
{
  struct node *n, *p;

  if (t == LPAR) {
	n = expr(lex(*++ipp));
	if (lex(*++ipp) != RPAR)
		serr(") expected");
	return n;
  }
  if (t == NOT) {
	n = secondary(lex(*++ipp));
	p = newnode(NOT);
	p->n_info.n_opnd.n_left = n;
	return p;
  }
  return simple(t);
}


/* Main: the main body */
int find_main(void)
{
  char **pathlist, *path, *last;
  int pathcnt = 0, i, ac;
  struct node *pred;
  int argc = Gargc;
  char **argv = Gargv;

  (void) umask(um = umask(0));  /* non-destructive get-umask :-)         */
  time(&current_time);          /* get current time                      */

  argv++;
  ac = argc;
gnu_hack:
  pathlist= argv;
  while (--argc > 0 && lex(*argv) == NONE) {    /* find paths            */
	pathcnt++;
	argv++;
  }
  if (pathcnt == 0) {
	/* path-list don`t specified, GNU extension, make path-list as "." */
	argc = ac + 1;
	argv = pathlist - 1;
	*argv = ".";
	goto gnu_hack;
  }
  ipp = argv;                   /* prepare for parsing                   */
  if (argc != 0) {              /* If there is anything to parse,        */
	pred = expr(lex(*ipp)); /* then do so                            */
	if (lex(*++ipp) != EOI) /* Make sure there's nothing left        */
		serr("garbage at end of predicate");
  } else                        /* No predicate list                     */
	pred = (struct node *) NULL;

  for (i = 0; i < pathcnt; i++) {
	if (xdev_flag) xdev_flag = 2;
	path = pathlist[i];
	if ((last = strrchr(path, '/')) == NULL) last = path; else last++;
	find(path, pred, last);
  }
  return 0;
}
#endif /* APPLET_USAGE */
