/*
 * Init         A System-V Init Clone.
 *
 * Usage:       /sbin/init
 *                   init [0123456SsQqAaBbCc]
 *                telinit [0123456SsQqAaBbCc]
 *
 * Version:     @(#)init.c  2.78  11-Feb-2000  miquels@cistron.nl
 *
 *              This file is part of the sysvinit suite,
 *              Copyright 1991-2000 Miquel van Smoorenburg.
 *
 * busybox port by Vladimir Oleynik
 * (C) 2002,2008 <dzo@simtreas.ru>
 *
 * Modified:    21 Feb 1998, Al Viro:
 *              'U' flag added to telinit. It forces init to re-exec itself
 *              (passing its state through exec, certainly).
 *              May be useful for smoother (heh) upgrades.
 *              24 Feb 1998, AV:
 *              did_boot made global and added to state - thanks, Miquel.
 *              Yet another file descriptors leak - close state pipe if
 *              re_exec fails.
 */
#ifdef APPLET_USAGE
#define init_trivial_usage \
	"[0123456SsQqAaBbCc]"
#define init_full_usage \
	"Init is the parent of all processes.\n" \
	"\t0\tRunlevel for halt system\n" \
	"\t1 S or s\tRunlevel is used to get the system down into single mode\n" \
	"\t6\tRunlevel is used to reboot the system\n" \
	"\tQ or q\tComand wake up re-examine the /etc/inittab file."
#else

#include "busybox.h"
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>
#include <string.h>
#include <termios.h>
#include <ctype.h>
#include <syslog.h>
#include <sys/time.h>
#include <sys/utsname.h>
#include <utmp.h>
#include <linux/kd.h>

#ifdef MAXHOSTNAMELEN_NETDB
#include <netdb.h>
#endif
#include "bb_sig.h"

#if defined(CONFIG_FEATURE_UTMP) || defined(CONFIG_FEATURE_WTMP)
# define BB_FEATURE_WUTMP 1
#endif

/* Standard configuration */
#define CHANGE_WAIT 0                   /* Change runlevel while
					   waiting for a process to exit? */
/* Debug and test modes */
#define DEBUG      0                    /* Debug code off */
#define INITDEBUG  0                    /* Fork at startup to debug init. */

/* Some constants */
#define INITPID    1                    /* pid of first process */
#define PIPE_FD    10                   /* Fileno of initfifo. */
#define STATE_PIPE 11                   /* used to pass state through exec */

/* Failsafe configuration */
#define MAXSPAWN   10                   /* Max times respawned in.. */
#define TESTTIME   120                  /* this much seconds */
#define SLEEPTIME  300                  /* Disable time */

/* Default path inherited by every child if it's not set. */
#define PATH_DFL   "PATH=/usr/local/sbin:/sbin:/bin:/usr/sbin:/usr/bin"


#ifdef BB_FEATURE_WUTMP
/* Prototypes. */
#if defined(__BIONIC__)
#error "BIONIC have not support wtmp stuff. Reconfigure for disable UTMP/WTMP features."
#endif
static void write_utmp_wtmp(char *user, char *id, int pid, int type, char *line);
#endif

/* Actions to be taken by init */
#define RESPAWN                 1
#define WAIT                    2
#define ONCE                    3
#define BOOT                    4
#define BOOTWAIT                5
#define POWERFAIL               6
#define POWERWAIT               7
#define POWEROKWAIT             8
#define CTRLALTDEL              9
#define OFF                    10
#define ONDEMAND               11
#define INITDEFAULT            12
#define SYSINIT                13
#define POWERFAILNOW           14
#define KBREQUEST              15

/* Information about a process in the in-core inittab */
typedef struct _child_ {
  int flags;                    /* Status of this entry */
  int exstat;                   /* Exit status of process */
  int pid;                      /* Pid of this process */
  time_t tm;                    /* When respawned last */
  int count;                    /* Times respawned in the last 2 minutes */
  char id[8];                   /* Inittab id (must be unique) */
  char rlevel[12];              /* run levels */
  int action;                   /* what to do (see list below) */
  char process[128];            /* The command line */
  struct _child_ *new;          /* New entry (after inittab re-read) */
  struct _child_ *next;         /* For the linked list */
} CHILD;

/* Values for the 'flags' field */
#define RUNNING                 2       /* Process is still running */
#define KILLME                  4       /* Kill this process */
#define DEMAND                  8       /* "runlevels" a b c */
#define FAILING                 16      /* process respawns rapidly */
#define WAITING                 32      /* We're waiting for this process */
#define ZOMBIE                  64      /* This process is already dead */
#define XECUTED         128     /* Set if spawned once or more times */

/* Log levels. */
#define L_CO    1               /* Log on the console. */
#define L_SY    2               /* Log with syslog() */
#define L_VB    (L_CO|L_SY)     /* Log with both. */

/* Tokens in state parser */
#define C_VER           1
#define C_END           2
#define C_REC           3
#define C_EOR           4
#define C_LEV           5
#define C_FLAG          6
#define C_ACTION        7
#define C_PROCESS       8
#define C_PID           9
#define C_EXS          10
#define C_EOF          -1
#define D_RUNLEVEL     -2
#define D_THISLEVEL    -3
#define D_PREVLEVEL    -4
#define D_GOTSIGN      -5
#define D_WROTE_WTMP_REBOOT -6
#define D_WROTE_UTMP_REBOOT -7
#define D_SLTIME       -8
#define D_DIDBOOT      -9


#define INIT_MAGIC 0x03091969
#define INIT_FIFO  "/dev/initctl"
#define INIT_CMD_START         0
#define INIT_CMD_RUNLVL        1
#define INIT_CMD_POWERFAIL     2
#define INIT_CMD_POWERFAILNOW  3
#define INIT_CMD_POWEROK       4

struct init_request {
  int magic;                    /* Magic number                 */
  int cmd;                      /* What kind of request         */
  int runlevel;                 /* Runlevel to change to        */
  int sleeptime;                /* Time between TERM and KILL   */
  char gen_id[8];               /* Beats me.. telnetd uses "fe" */
  char tty_id[16];              /* Tty name minus /dev/tty      */
  char host[MAXHOSTNAMELEN];    /* Hostname                     */
  char term_type[16];           /* Terminal type                */
  int signal;                   /* Signal to send               */
  int pid;                      /* Process to send to           */
  char exec_name[128];          /* Program to execute           */
  char reserved[128];           /* For future expansion.        */
};

#define VT_MASTER       "/dev/tty0"             /* Virtual console master */
#define CONSOLE         "/dev/console"          /* Logical system console */
#define INITTAB        "/etc/inittab"          /* Location of inittab */
#define IOSAVE          "/tmp/ioctl.save"       /* termios settings for SU */

#include <sys/reboot.h>

#define BMAGIC_HARD     0x89ABCDEF
#define BMAGIC_SOFT     0
#define BMAGIC_REBOOT   0x01234567
#define BMAGIC_HALT     0xCDEF0123
#define BMAGIC_POWEROFF 0x4321FEDC

#define ISMEMBER(set, val) ((set) & (1 << (val)))
#define DELSET(set, val)   ((set) &= ~(1 << (val)))
#define ADDSET(set, val)   ((set) |=  (1 << (val)))


#ifndef SIGPWR
#  define SIGPWR SIGUSR2
#endif


/* Version information */
#define Version "@(#) init 2.78  11-Feb-2000  miquels@cistron.nl"
static const char bootmsg[] = "version 2.78 %s";
#define E_VERSION "INIT_VERSION=sysvinit-2.78"


CALC_BSS struct BB_COMMON_BSS
{
	int dfl_level;                  /* Default runlevel */
	sig_atomic_t got_cont;          /* Set if we received the SIGCONT signal */
	sig_atomic_t got_signals;       /* Set if we received a signal. */
	int emerg_shell;                /* Start emergency shell? */
	struct utmp utproto;            /* Only used for sizeof(utproto.ut_id) */
	const char *console_dev;        /* Console device. */
	int wrote_iosave;               /* Did we write ioctl.save yet? */
	int did_boot;                   /* Did we already do BOOT* stuff? */
	int reload;                     /* Should we do initialization stuff? */
	int oops_error;                 /* Used by some of the re-exec code. */
	CHILD *family;                  /* The linked list of all entries */
	CHILD *newFamily;               /* The list after inittab re-read */
	int      dumped;
	char *get_part_s;
	int newlevel;
	/* data segment */
	char runlevel;                  /* = 'S' The current run level */
	char thislevel;                 /* = 'S' The current runlevel */
	char prevlevel;                 /* = 'N' Previous runlevel */
	int sltime;                     /* = 5   Sleep time between TERM and KILL */
	int pipe_fd;                    /* = -1; /dev/initctl */
	const char *myname;             /* = "/sbin/init" What should we exec */
	int warn;                       /* = 1 */
	CHILD ch_emerg;                 /* Emergency shell */
#ifdef CONFIG_FEATURE_UTMP
	int wrote_utmp_reboot;          /* Set when we wrote the reboot record */
#endif
#ifdef CONFIG_FEATURE_WTMP
	int wrote_wtmp_reboot;          /* Set when we wrote the reboot record */
#endif
} bb_common_bss;
#define dfl_level bb_common_bss.dfl_level
#define got_cont bb_common_bss.got_cont
#define got_signals bb_common_bss.got_signals
#define emerg_shell bb_common_bss.emerg_shell
#define utproto bb_common_bss.utproto
#define console_dev bb_common_bss.console_dev
#define wrote_iosave bb_common_bss.wrote_iosave
#define did_boot bb_common_bss.did_boot
#define reload bb_common_bss.reload
#define oops_error bb_common_bss.oops_error
#define family bb_common_bss.family
#define newFamily bb_common_bss.newFamily
#define dumped bb_common_bss.dumped
#define get_part_s bb_common_bss.get_part_s
#define newlevel bb_common_bss.newlevel
#define pipe_fd bb_common_bss.pipe_fd
#define warn bb_common_bss.warn
#define ch_emerg bb_common_bss.ch_emerg
#define wrote_wtmp_reboot bb_common_bss.wrote_wtmp_reboot
#define wrote_utmp_reboot bb_common_bss.wrote_utmp_reboot

static const char Signature[] = "12567362";   /* Signature for re-exec fd */

/* Macro to see if this is a special action */
#define ISPOWER(i) ((i) == POWERWAIT || (i) == POWERFAIL || \
		    (i) == POWEROKWAIT || (i) == POWERFAILNOW || \
		    (i) == CTRLALTDEL)

/* ascii values for the `action' field. */
static const struct actions {
  const char *name;
  int act;
} actions[] = {
  { "respawn",     RESPAWN      },
  { "wait",        WAIT         },
  { "once",        ONCE         },
  { "boot",        BOOT         },
  { "bootwait",    BOOTWAIT     },
  { "powerfail",   POWERFAIL    },
  { "powerfailnow",POWERFAILNOW },
  { "powerwait",   POWERWAIT    },
  { "powerokwait", POWEROKWAIT  },
  { "ctrlaltdel",  CTRLALTDEL   },
  { "off",         OFF          },
  { "ondemand",    ONDEMAND     },
  { "initdefault", INITDEFAULT  },
  { "sysinit",     SYSINIT      },
  { "kbrequest",   KBREQUEST    },
  { NULL,          OFF          },
};

/*
 *      State parser token table (see receive_state)
 */
struct CMDS {
	const char *name;
	int cmd;
};

static const struct CMDS cmds[] = {
  { "VER",         C_VER        },
  { "END",         C_END        },
  { "REC",         C_REC        },
  { "EOR",         C_EOR        },
  { "LEV",         C_LEV        },
  { "FL ",         C_FLAG       },
  { "AC ",         C_ACTION     },
  { "CMD",         C_PROCESS    },
  { "PID",         C_PID        },
  { "EXS",         C_EXS        },
  { "-RL",         D_RUNLEVEL   },
  { "-TL",         D_THISLEVEL  },
  { "-PL",         D_PREVLEVEL  },
  { "-SI",         D_GOTSIGN    },
#ifdef CONFIG_FEATURE_WTMP
  { "-WR",         D_WROTE_WTMP_REBOOT},
#endif
#ifdef CONFIG_FEATURE_UTMP
  { "-WU",         D_WROTE_UTMP_REBOOT},
#endif
  { "-ST",         D_SLTIME     },
  { "-DB",         D_DIDBOOT    },
  { NULL,          0            }
};

struct FLAGS {
	const char *name;
	int mask;
};

static const struct FLAGS flags[] = {
	{ "RU", RUNNING},
	{ "DE", DEMAND},
	{ "XD", XECUTED},
	{ NULL, 0}
};

/*
 *      Open the console with retries.
 */
static int console_open(int mode)
{
	int f, fd = -1;
	int m;

	/*
	 *      Open device in nonblocking mode.
	 */
	m = mode | O_NONBLOCK;

	/*
	 *      Retry the open five times.
	 */
	for(f = 0; f < 5; f++)
		if ((fd = open(console_dev, m)) >= 0) break;

	if (fd < 0) return fd;

	/*
	 *      Set original flags.
	 */
	if (m != mode)
		fcntl(fd, F_SETFL, mode);
	return fd;
}

/*
 *      Log something to a logfile and the console.
 */
static void i_log(int loglevel, const char *s, ...)
{
	va_list va_alist;
	char buf[256];

	va_start(va_alist, s);
	vsnprintf(buf, sizeof(buf), s, va_alist);
	va_end(va_alist);

	if (loglevel & L_SY) {
		/*
		 *      Re-etablish connection with syslogd every time.
		 */
		bb_openlog(0, LOG_DAEMON);
		syslog(LOG_INFO, "%s", buf);
		/* closelog(); NOT needed with recent libc's. */
	}

	/*
	 *      And log to the console.
	 */
	if (loglevel & L_CO) {
		int fd;
		if ((fd = console_open(O_WRONLY|O_NOCTTY|O_NDELAY)) >= 0) {
			dprintf(fd, "\rINIT: %s\r\n", buf);
			close(fd);
		}
	}
}

/*
 *      Sleep a number of seconds.
 *
 *      This only works correctly because the linux select updates
 *      the elapsed time in the struct timeval passed to select!
 */
static void do_sleep(int sec)
{
	struct timeval tv;

	tv.tv_sec = sec;
	tv.tv_usec = 0;

	while(select(0, NULL, NULL, NULL, &tv) < 0 && errno == EINTR)
		;
}

/*
 *      Send the state info of the previous running init to
 *      the new one, in a version-independant way.
 */
static void send_state(int fd)
{
	FILE    *fp;
	CHILD   *p;
	int     i,val;

	fp = fdopen(fd,"w");

	fprintf(fp, "VER" Version "\n"
		"-RL%c\n-TL%c\n-PL%c\n-SI%u\n"
#ifdef CONFIG_FEATURE_WTMP
		"-WR%d\n"
#endif
#ifdef CONFIG_FEATURE_UTMP
		"-WU%d\n"
#endif
		"-ST%d\n-DB%d\n",
		bb_common_bss.runlevel, bb_common_bss.thislevel, bb_common_bss.prevlevel, got_signals,
#ifdef CONFIG_FEATURE_WTMP
		wrote_wtmp_reboot,
#endif
#ifdef CONFIG_FEATURE_UTMP
		wrote_utmp_reboot,
#endif
		bb_common_bss.sltime, did_boot);
	for (p = family; p; p = p->next) {
		fprintf(fp, "REC%s\nLEV%s\n", p->id, p->rlevel);
		for (i = 0, val = p->flags; flags[i].mask; i++)
			if (val & flags[i].mask) {
				val &= ~flags[i].mask;
				fprintf(fp, "FL %s\n",flags[i].name);
			}
		fprintf(fp, "PID%d\nEXS%u\n", p->pid, p->exstat);
		for(i = 0; actions[i].act; i++)
			if (actions[i].act == p->action) {
				fprintf(fp, "AC %s\n", actions[i].name);
				break;
			}
		fprintf(fp, "CMD%s\nEOR\n", p->process);
	}
	fprintf(fp, "END\n");
	fclose(fp);
}

/* Read a string from a file descriptor. */
static void get_string(char *p, int size, FILE *f)
{
	int     c;

	while ((c = getc(f)) != EOF && c != '\n') {
		if (--size > 0)
			*p++ = c;
	}
	*p = '\0';
}

/* Read trailing data from the state pipe until we see a newline. */
static void get_void(FILE *f)
{
	int     c;

	while ((c = getc(f)) != EOF && c != '\n')
		;
}

/* Read the next "command" from the state pipe. */
static int get_cmd(FILE *f)
{
	char    cmd[4] = "   ";

	if (fread(cmd, 1, 3, f) != 3)
		return C_EOF;

	return ((struct CMDS *)find_string_from_struct(cmds, sizeof(struct CMDS), cmd))->cmd;
}

/* Read a CHILD * from the state pipe. */
static CHILD *get_record(FILE *f)
{
	int     cmd;
	char    s[32];
	CHILD   *p;

	do {
		switch (cmd = get_cmd(f)) {
			case C_END:
				get_void(f);
				return NULL;
			case 0:
				get_void(f);
				break;
			case C_REC:
				break;
			case D_RUNLEVEL:
				fscanf(f, "%c\n", &bb_common_bss.runlevel);
				break;
			case D_THISLEVEL:
				fscanf(f, "%c\n", &bb_common_bss.thislevel);
				break;
			case D_PREVLEVEL:
				fscanf(f, "%c\n", &bb_common_bss.prevlevel);
				break;
			case D_GOTSIGN:
				fscanf(f, "%u\n", &got_signals);
				break;
#ifdef CONFIG_FEATURE_WTMP
			case D_WROTE_WTMP_REBOOT:
				fscanf(f, "%d\n", &wrote_wtmp_reboot);
				break;
#endif
#ifdef CONFIG_FEATURE_UTMP
			case D_WROTE_UTMP_REBOOT:
				fscanf(f, "%d\n", &wrote_utmp_reboot);
				break;
#endif
			case D_SLTIME:
				fscanf(f, "%d\n", &bb_common_bss.sltime);
				break;
			case D_DIDBOOT:
				fscanf(f, "%d\n", &did_boot);
				break;
			default:
				if (cmd > 0 || cmd == C_EOF) {
					oops_error = -1;
					return NULL;
				}
		}
	} while (cmd != C_REC);

	while ((p = malloc(sizeof(CHILD))) == NULL ) {
		i_log(L_VB, bb_msg_memory_exhausted);
		do_sleep(5);
	}
	memset(p, 0, sizeof(CHILD));
	get_string(p->id, sizeof(p->id), f);
	do switch(cmd = get_cmd(f)) {
		case 0:
		case C_EOR:
			get_void(f);
			break;
		case C_PID:
			fscanf(f, "%d\n", &(p->pid));
			break;
		case C_EXS:
			fscanf(f, "%u\n", &(p->exstat));
			break;
		case C_LEV:
			get_string(p->rlevel, sizeof(p->rlevel), f);
			break;
		case C_PROCESS:
			get_string(p->process, sizeof(p->process), f);
			break;
		case C_FLAG:
			get_string(s, sizeof(s), f);
			p->flags |= ((struct FLAGS *)find_string_from_struct(flags, sizeof(struct FLAGS), s))->mask;
			break;
		case C_ACTION:
			get_string(s, sizeof(s), f);
			p->action = ((struct actions *)find_string_from_struct(actions, sizeof(struct actions), s))->act;
			break;
		default:
			free(p);
			oops_error = -1;
			return NULL;
	} while( cmd != C_EOR);

	return p;
}

/*
 *      Read the complete state info from the state pipe.
 *      Returns 0 on success
 */
static int receive_state(int fd)
{
	FILE    *f;
	char    old_version[256];
	CHILD   **pp;

	f = fdopen(fd, "r");

	if (get_cmd(f) != C_VER)
		return -1;
	get_string(old_version, sizeof(old_version), f);
	oops_error = 0;
	for (pp = &family; (*pp = get_record(f)) != NULL; pp = &((*pp)->next))
		;
	fclose(f);
	return oops_error;
}

/* Reset console tries. */
static void console_init(void)
{
	int fd;
	int tried_devcons = 0;
	int tried_vtmaster = 0;
	char *s;

	if ((s = getenv("CONSOLE")) != NULL)
		console_dev = s;
	else {
		console_dev = CONSOLE;
		tried_devcons++;
	}

	while ((fd = open(console_dev, O_RDONLY|O_NONBLOCK)) < 0) {
		if (!tried_devcons) {
			tried_devcons++;
			console_dev = CONSOLE;
			continue;
		}
		if (!tried_vtmaster) {
			tried_vtmaster++;
			console_dev = VT_MASTER;
			continue;
		}
		break;
	}
	if (fd < 0)
		console_dev = bb_dev_null;
	else
		close(fd);
}

/* We got a signal (HUP PWR WINCH ALRM INT) */
static void ATTRIBUTE_UNUSED signal_handler(int sig)
{
	ADDSET(got_signals, sig);
}

/* SIGCHLD: one of our children has died. */
static void ATTRIBUTE_UNUSED chld_handler(int sig)
{
	CHILD           *ch;
	int             pid, st;
	int             saved_errno = errno;

	/*
	 *      Find out which process(es) this was (were)
	 */
	while((pid = waitpid(-1, &st, WNOHANG)) != 0) {
		if (errno == ECHILD) break;
		for( ch = family; ch; ch = ch->next )
			if ( ch->pid == pid && (ch->flags & RUNNING) ) {
#if DEBUG
				i_log(L_VB, "chld_handler: marked %d as zombie",
					ch->pid);
#endif
				ADDSET(got_signals, SIGCHLD);
				ch->exstat = st;
				ch->flags |= ZOMBIE;
				if (ch->new) {
					ch->new->exstat = st;
					ch->new->flags |= ZOMBIE;
				}
				break;
			}
#if DEBUG
		if (ch == NULL)
			i_log(L_VB, "chld_handler: unknown child %d exited.",
				pid);
#endif
	}
	errno = saved_errno;
}

/*
 *      Linux ignores all signals sent to init when the
 *      SIG_DFL handler is installed. Therefore we must catch SIGTSTP
 *      and SIGCONT, or else they won't work....
 *
 *      The SIGCONT handler
 */
static void ATTRIBUTE_UNUSED cont_handler(int sig)
{
	got_cont = 1;
}

/*
 *      Dump core. Returns 0 if we are the child, so that the caller
 *      can return if it is a signal handler - SIGSEGV is blocked in
 *      the handler, so it will be raised when the handler returns.
 */
static inline int coredump(void)
{
	struct rlimit   rlim;

	if (dumped) return 1;
	dumped = 1;

	if (fork() != 0) return 1;

	rlim.rlim_cur = RLIM_INFINITY;
	rlim.rlim_max = RLIM_INFINITY;
	setrlimit(RLIMIT_CORE, &rlim);

	chdir("/");
	bb_signal((1 << SIGSEGV), SIG_DFL);
	raise(SIGSEGV);
	return 0;
}

static void ATTRIBUTE_UNUSED segv_handler(int sig)
{
	int     saved_errno = errno;

	i_log(L_VB, "PANIC: segmentation violation! sleeping for 30 seconds.");
	if (coredump() != 0)
		do_sleep(30);
	errno = saved_errno;
}

/* The SIGSTOP & SIGTSTP handler */
static void ATTRIBUTE_UNUSED stop_handler(int sig)
{
	int     saved_errno = errno;

	got_cont = 0;
	while(!got_cont) pause();
	got_cont = 0;
	errno = saved_errno;
}

/* Set terminal settings to reasonable defaults */
static void set_term(int how)
{
	struct termios tty;
	int fd, fd2;
	int restore_ok = 0;

	if ((fd = console_open(O_RDWR|O_NOCTTY)) < 0) {
		i_log(L_VB, "can't open %s", console_dev);
		return;
	}

	/* Do we want to save or restore modes. */
	if (how > 0) {
		/* Save the terminal settings. */
		(void) tcgetattr(fd, &tty);
		if ((fd2 = open(IOSAVE, O_WRONLY|O_CREAT|O_TRUNC, 0600)) < 0) {
			if (errno != EROFS) {
#ifndef __BIONIC__
				i_log(L_VB, "can't open(%s, O_WRONLY): %m", IOSAVE);
#else
				i_log(L_VB, "can't open(%s, O_WRONLY): %s", IOSAVE, strerror(errno));
#endif
			}
			(void) close(fd);
			(void) close(fd2);
			return;
		}
		(void) write(fd2, &tty, sizeof(struct termios));
		wrote_iosave = 1;
		(void) close(fd);
		(void) close(fd2);
		return;
	}

	/* Restore the terminal settings. */
	if (wrote_iosave && (fd2 = open(IOSAVE, O_RDONLY)) >= 0) {
		if (read(fd2, &tty, sizeof(tty)) == sizeof(tty))
			restore_ok = 1;
		close(fd2);
	}
	if (restore_ok == 0) {
		/* No old settings - create some resonable defaults. */
		(void) tcgetattr(fd, &tty);

		tty.c_cflag &= CBAUD|CBAUDEX|CSIZE|CSTOPB|PARENB|PARODD;
		tty.c_cflag |= HUPCL|CLOCAL;

		tty.c_cc[VINTR]  = 3;   /* ctrl('c') */
		tty.c_cc[VQUIT]  = 28;  /* ctrl('\\') */
		tty.c_cc[VERASE] = 127;
		tty.c_cc[VKILL]  = 24;  /* ctrl('x') */
		tty.c_cc[VEOF]   = 4;   /* ctrl('d') */
		tty.c_cc[VTIME]  = 0;
		tty.c_cc[VMIN]   = 1;
		tty.c_cc[VSTART] = 17;  /* ctrl('q') */
		tty.c_cc[VSTOP]  = 19;  /* ctrl('s') */
		tty.c_cc[VSUSP]  = 26;  /* ctrl('z') */
	}

	/* Set pre and post processing */
	tty.c_iflag = IGNPAR|ICRNL|IXON|IXANY;
	tty.c_oflag = OPOST|ONLCR;
	tty.c_lflag = ISIG|ICANON|ECHO|ECHOCTL|ECHOPRT|ECHOKE;

	/* Now set the terminal line. */
	(void) tcsetattr(fd, TCSADRAIN, &tty);
	(void) close(fd);
}

/* See if one character of s2 is in s1 */
static int any(char *s1, char *s2)
{
	while(*s2)
		if (strchr(s1, *s2++) != NULL)
			return(1);
	return(0);
}

/* Fork and execute. This function is too long and indents too deep. */
static int spawn(CHILD *ch, int *res)
{
  const char *args[16];               /* Argv array */
  char buf[136];                /* Line buffer */
  int f, st;                    /* Scratch variables */
  char *ptr;                    /* Ditto */
  time_t t;                     /* System time */
  int oldAlarm;                 /* Previous alarm value */
  char *proc = ch->process;     /* Command line */
  char i_lvl[] = "RUNLEVEL=x";  /* Runlevel in environment. */
  char i_prev[] = "PREVLEVEL=x";/* Previous runlevel. */
  char i_cons[32];              /* console device. */
  pid_t pid, pgrp;              /* child, console process group. */
  sigset_t nmask;               /* For blocking SIGCHLD */

  *res = -1;
  buf[sizeof(buf) - 1] = 0;

  /* Skip '+' if it's there */
  if (proc[0] == '+') proc++;

  ch->flags |= XECUTED;

  if (ch->action == RESPAWN || ch->action == ONDEMAND) {
	/* Is the date stamp from less than 2 minutes ago? */
	time(&t);
	if (ch->tm + TESTTIME > t) {
		ch->count++;
	} else {
		ch->count = 0;
		ch->tm = t;
	}

	/* Do we try to respawn too fast? */
	if (ch->count >= MAXSPAWN) {

	  i_log(L_VB, "Id \"%s\" respawning too fast: disabled for %d minutes",
			ch->id, SLEEPTIME / 60);
	  ch->flags &= ~RUNNING;
	  ch->flags |= FAILING;

	  /* Remember the time we stopped */
	  ch->tm = t;

	  /* Try again in 5 minutes */
	  oldAlarm = alarm(0);
	  if (oldAlarm > SLEEPTIME || oldAlarm <= 0) oldAlarm = SLEEPTIME;
	  alarm(oldAlarm);
	  return(-1);
	}
  }

  if (any(proc, "~`!$^&*()=|\\{}[];\"'<>?")) {
  /* See if we need to fire off a shell for this command */
	/* Give command line to shell */
	args[1] = DEFAULT_SHELL;
	args[2] = "-c";
	strcpy(buf, "exec ");
	strncat(buf, proc, sizeof(buf) - strlen(buf) - 1);
	args[3] = buf;
	args[4] = NULL;
  } else {
	/* Split up command line arguments */
	strncpy(buf, proc, sizeof(buf) - 1);
	ptr = buf;
	for(f = 1; f < 15; f++) {
		/* Skip white space */
		while(*ptr == ' ' || *ptr == '\t') ptr++;
		args[f] = ptr;

		/* May be trailing space.. */
		if (*ptr == 0) break;

		/* Skip this `word' */
		while(*ptr && *ptr != ' ' && *ptr != '\t' && *ptr != '#')
			ptr++;

		/* If end-of-line, break */
		if (*ptr == '#' || *ptr == 0) {
			f++;
			*ptr = 0;
			break;
		}
		/* End word with \0 and continue */
		*ptr++ = 0;
	}
	args[f] = NULL;
  }
  args[0] = args[1];
  while(1) {
	/* Block sigchild while forking. */
	nmask = bb_sigblock(1 << SIGCHLD);

	if ((pid = fork()) == 0) {
		unsigned int nsig_msk;

		close(0);
		close(1);
		close(2);
		if (pipe_fd >= 0) close(pipe_fd);

		bb_sigunblock(nmask);

		/* Now set RUNLEVEL and PREVLEVEL */
		snprintf(i_cons, sizeof(i_cons), "CONSOLE=%s", console_dev);
		i_lvl[9]   = bb_common_bss.thislevel;
		i_prev[10] = bb_common_bss.prevlevel;
		putenv(i_lvl);
		putenv(i_prev);
		putenv(i_cons);
		putenv(E_VERSION);

		/*
		 *      In sysinit, boot, bootwait or single user mode:
		 *      any wait-type subprocess gets the console
		 *      as its controlling tty (if not in use yet).
		 */
		if (strchr("*#sS", bb_common_bss.runlevel) && ch->flags & WAITING) {
			/*
			 *      We fork once extra. This is so that we can
			 *      wait and change the process group and session
			 *      of the console after exit of the leader.
			 */
			setsid();
			if ((f = console_open(O_RDWR|O_NOCTTY)) >= 0) {
				(void)ioctl(f, TIOCSCTTY, 0);
				dup(f);
				dup(f);
			}
			if ((pid = fork()) < 0) {
				i_log(L_VB, "cannot fork");
				exit(1);
			}
			if (pid > 0) {
				/* Ignore keyboard signals etc. */
				bb_signal((1 << SIGINT)
					    | (1 << SIGTSTP)
					    | (1 << SIGQUIT)
					    | (1 << SIGCHLD),
						SIG_IGN);


				waitpid_no_int(pid, &st, 0, 1);
				pgrp = 0;
				pgrp = tcgetpgrp(f);
				if (pgrp != getpid())
					exit(0);
				/*
				 *      De-attach session from controlling tty.
				 *      Or actually, the other way around.
				 *      This _without_ SIGHUP etc.
				 */
				if ((pid = fork()) < 0) {
					i_log(L_VB, "cannot fork");
					exit(1);
				}
				if (pid == 0) {
					setsid();
					(void)ioctl(f, TIOCSCTTY, 1);
					exit(0);
				}
				waitpid_no_int(pid, &st, 0, 1);
				exit(0);
			}
			/* Set ioctl settings to default ones */
			set_term(0);
		} else {
			if ((f = console_open(O_RDWR|O_NOCTTY)) < 0) {
#ifndef __BIONIC__
				i_log(L_VB, "open(%s): %m", console_dev);
#else
				i_log(L_VB, "open(%s): %s", console_dev, strerror(errno));
#endif
				f = open(bb_dev_null, O_RDWR);
			}
			dup(f);
			dup(f);
			setsid();
		}

		/* Reset all the signals */
		nsig_msk = 0;
		for(f = 1; f < NSIG; f++)
			nsig_msk |= (1 << f);
		bb_signal(nsig_msk, SIG_DFL);
		execvp(args[1], ((char **)args) + 1);

		/*
		 *      Is this a bug in execvp? It does _not_ execute shell
		 *      scripts (/etc/rc !), so we try again with
		 *      'sh -c exec ...'
		 */
		if (errno == ENOEXEC) {
			args[1] = DEFAULT_SHELL;
			args[2] = "-c";
			strcpy(buf, "exec ");
			strncat(buf, proc, sizeof(buf) - strlen(buf) - 1);
			args[3] = buf;
			args[4] = NULL;
			execvp(args[1], ((char **)args) + 1);
		}
		i_log(L_VB, "cannot execute \"%s\"", args[1]);
		exit(1);
	}
	*res = pid;
	bb_sigunblock(nmask);

#if DEBUG
	i_log(L_VB, "Started id %s (pid %d)", ch->id, pid);
#endif

	if (pid == -1) {
		i_log(L_VB, "cannot fork, retry..", NULL, NULL);
		do_sleep(5);
		continue;
	}
	return(pid);
  }
}

/* Start a child running! */
static void startup(CHILD *ch)
{
	/* See if it's disabled */
	if (ch->flags & FAILING) return;

	switch(ch->action) {

		case SYSINIT:
		case BOOTWAIT:
		case WAIT:
		case POWERWAIT:
		case POWERFAILNOW:
		case POWEROKWAIT:
		case CTRLALTDEL:
			if (!(ch->flags & XECUTED)) ch->flags |= WAITING;
		case KBREQUEST:
		case BOOT:
		case POWERFAIL:
		case ONCE:
			if (ch->flags & XECUTED) break;
		case ONDEMAND:
		case RESPAWN:
			ch->flags |= RUNNING;
			if (spawn(ch, &(ch->pid)) < 0) break;
#ifdef BB_FEATURE_WUTMP
			/*
			 *      Do NOT log if process field starts with '+'
			 *      FIXME: that's for compatibility with *very*
			 *      old getties - probably it can be taken out.
			 */
			if (ch->process[0] != '+')
				write_utmp_wtmp("", ch->id, ch->pid,
					INIT_PROCESS, "");
#endif
			break;
	}
}

/* My version of strtok(3). */
static char *get_part(char *str, int tok)
{
	char *p, *q;

	if (str != NULL)
		get_part_s = str;
	if (get_part_s == NULL || *get_part_s == 0)
		return NULL;
	q = p = get_part_s;
	while(*p != tok && *p)
		p++;
	if (*p == tok)
		*p++ = 0;
	get_part_s = p;

	return q;
}

/* Read the inittab file. */
static void read_inittab(void)
{
  FILE *fp;                     /* The INITTAB file */
  char buf[256];                /* Line buffer */
  char err[64];                 /* Error message. */
  char *id, *rlevel,
       *action, *process;       /* Fields of a line */
  char *p;
  CHILD *ch, *old, *i;          /* Pointers to CHILD structure */
  CHILD *head = NULL;           /* Head of linked list */
  int lineNo = 0;               /* Line number in INITTAB file */
  int actionNo;                 /* Decoded action field */
  int f;                        /* Counter */
  int round;                    /* round 0 for SIGTERM, round 1 for SIGKILL */
  int foundOne = 0;             /* No killing no sleep */
  int talk;                     /* Talk to the user */
  int done = 0;                 /* Ready yet? */
  sigset_t nmask;               /* For blocking SIGCHLD. */

#if DEBUG
  if (newFamily != NULL) {
	i_log(L_VB, "PANIC newFamily != NULL");
	exit(1);
  }
  i_log(L_VB, "Reading inittab");
#endif

  /* Open INITTAB and real line by line. */
  if ((fp = fopen(INITTAB, "r")) == NULL)
	i_log(L_VB, "No inittab file found");

  while(!done) {
	/* Add single user shell entry at the end. */
	if (fp == NULL || fgets(buf, sizeof(buf), fp) == NULL) {
		done = 1;
		/* See if we have a single user entry. */
		for(old = newFamily; old; old = old->next)
			if (strcmp(old->rlevel, "S") == 0) break;
		if (old == NULL)
			snprintf(buf, sizeof(buf), "~~:S:wait:%s\n", DEFAULT_SHELL);
		else
			continue;
	}
	lineNo++;
	/* Skip comments and empty lines */
	for(p = buf; *p == ' ' || *p == '\t'; p++)
		;
	if (*p == '#' || *p == '\n') continue;

	/* Decode the fields */
	id =      get_part(p,  ':');
	rlevel =  get_part(NULL, ':');
	action =  get_part(NULL, ':');
	process = get_part(NULL, '\n');

	/*
	 *      Check if syntax is OK. Be very verbose here, to
	 *      avoid newbie postings on comp.os.linux.setup :)
	 */
	err[0] = 0;
	if (!id || !*id) strcpy(err, "missing id field");
	if (!rlevel)     strcpy(err, "missing runlevel field");
	if (!process)    strcpy(err, "missing process field");
	if (!action || !*action)
			strcpy(err, "missing action field");
	if (id && strlen(id) > sizeof(utproto.ut_id))
		sprintf(err, "id field too long (max %d characters)",
			(int)sizeof(utproto.ut_id));
	if (rlevel && strlen(rlevel) > 11)
		strcpy(err, "rlevel field too long (max 11 characters)");
	if (process && strlen(process) > 127)
		strcpy(err, "process field too long");
	if (action && strlen(action) > 32)
		strcpy(err, "action field too long");
	if (err[0] != 0) {
		i_log(L_VB, "%s[%d]: %s", INITTAB, lineNo, err);
#if DEBUG
		i_log(L_VB, "%s:%s:%s:%s", id, rlevel, action, process);
#endif
		continue;
	}

	/* Decode the "action" field */
	actionNo = -1;
	for(f = 0; actions[f].name; f++)
		if (strcasecmp(action, actions[f].name) == 0) {
			actionNo = actions[f].act;
			break;
		}
	if (actionNo == -1) {
		i_log(L_VB, "%s[%d]: %s: unknown action field",
			INITTAB, lineNo, action);
		continue;
	}

	/* See if the id field is unique */
	for(old = newFamily; old; old = old->next) {
		if(strcmp(old->id, id) == 0 && strcmp(id, "~~")) {
			i_log(L_VB, "%s[%d]: duplicate ID field \"%s\"",
				INITTAB, lineNo, id);
			break;
		}
	}
	if (old) continue;

	/* Allocate a CHILD structure */
	while ((ch = malloc(sizeof(CHILD))) == NULL) {
		i_log(L_VB, bb_msg_memory_exhausted);
		do_sleep(5);
	}
	memset(ch, 0, sizeof(CHILD));

	/* And fill it in. */
	ch->action = actionNo;
	strncpy(ch->id, id, sizeof(utproto.ut_id) + 1); /* Hack for different libs. */
	strncpy(ch->process, process, sizeof(ch->process) - 1);
	if (rlevel[0]) {
		for(f = 0; f < sizeof(rlevel) - 1 && rlevel[f]; f++) {
			ch->rlevel[f] = rlevel[f];
			if (ch->rlevel[f] == 's') ch->rlevel[f] = 'S';
		}
		strncpy(ch->rlevel, rlevel, sizeof(ch->rlevel) - 1);
	} else {
		strcpy(ch->rlevel, "0123456789");
		if (ISPOWER(ch->action))
			strcpy(ch->rlevel, "S0123456789");
	}
	/* We have the fake runlevel '#' for SYSINIT  and
	 *      '*' for BOOT and BOOTWAIT.
	 */
	if (ch->action == SYSINIT) strcpy(ch->rlevel, "#");
	if (ch->action == BOOT || ch->action == BOOTWAIT)
		strcpy(ch->rlevel, "*");

	/* Now add it to the linked list. Special for powerfail. */
	if (ISPOWER(ch->action)) {

		/* Disable by default */
		ch->flags |= XECUTED;

		/* Tricky: insert at the front of the list.. */
		old = NULL;
		for(i = newFamily; i; i = i->next) {
			if (!ISPOWER(i->action)) break;
			old = i;
		}
		/* Now add after entry "old" */
		if (old) {
			ch->next = i;
			old->next = ch;
			if (i == NULL) head = ch;
		} else {
			ch->next = newFamily;
			newFamily = ch;
			if (ch->next == NULL) head = ch;
		}
	} else {
		/* Just add at end of the list */
		if (ch->action == KBREQUEST) ch->flags |= XECUTED;
		ch->next = NULL;
		if (head)
			head->next = ch;
		else
			newFamily = ch;
		head = ch;
	}

	/* Walk through the old list comparing id fields */
	for(old = family; old; old = old->next)
		if (strcmp(old->id, ch->id) == 0) {
			old->new = ch;
			break;
		}
  }
  /* We're done. */
  if (fp) fclose(fp);

  /* Loop through the list of children, and see if they need to be killed. */

#if DEBUG
  i_log(L_VB, "Checking for children to kill");
#endif
  for(round = 0; round < 2; round++) {
    talk = 1;
    for(ch = family; ch; ch = ch->next) {
	ch->flags &= ~KILLME;

	/* Is this line deleted? */
	if (ch->new == NULL) ch->flags |= KILLME;

	/*
	 *      If the entry has changed, kill it anyway. Note that
	 *      we do not check ch->process, only the "action" field.
	 *      This way, you can turn an entry "off" immediately, but
	 *      changes in the command line will only become effective
	 *      after the running version has exited.
	 */
	if (ch->new && ch->action != ch->new->action) ch->flags |= KILLME;

	/* Only BOOT processes may live in all levels */
	if (ch->action != BOOT &&
	    strchr(ch->rlevel, bb_common_bss.runlevel) == NULL) {
		/*
		 *      Ondemand procedures live always,
		 *      except in single user
		 */
		if (bb_common_bss.runlevel == 'S' || !(ch->flags & DEMAND))
			ch->flags |= KILLME;
	}

	/* Now, if this process may live note so in the new list */
	if ((ch->flags & KILLME) == 0) {
		ch->new->flags  = ch->flags;
		ch->new->pid    = ch->pid;
		ch->new->exstat = ch->exstat;
		continue;
	}


	/* Is this process still around? */
	if ((ch->flags & RUNNING) == 0) {
		ch->flags &= ~KILLME;
		continue;
	}
#if DEBUG
	i_log(L_VB, "Killing \"%s\"", ch->process);
#endif
	switch(round) {
		case 0: /* Send TERM signal */
			if (talk)
				i_log(L_CO, "Sending processes the TERM signal");
			kill(-(ch->pid), SIGTERM);
			foundOne = 1;
			break;
		case 1: /* Send KILL signal and collect status */
			if (talk)
				i_log(L_CO, "Sending processes the KILL signal");
			kill(-(ch->pid), SIGKILL);
			break;
	}
	talk = 0;

    }
    /* See if we have to wait 5 seconds */
    if (foundOne && round == 0) {
	/* Yup, but check every second if we still have children. */
	for(f = 0; f < bb_common_bss.sltime; f++) {
		for(ch = family; ch; ch = ch->next) {
			if (!(ch->flags & KILLME)) continue;
			if ((ch->flags & RUNNING) && !(ch->flags & ZOMBIE))
				break;
		}
		if (ch == NULL) {
			/* No running children, skip SIGKILL */
			round = 1;
			foundOne = 0; /* Skip the sleep below. */
			break;
		}
		do_sleep(1);
	}
    }
  }

  /* Now give all processes the chance to die and collect exit statuses. */
  if (foundOne) do_sleep(1);
  for(ch = family; ch; ch = ch->next)
	if (ch->flags & KILLME) {
		if (!(ch->flags & ZOMBIE))
		    i_log(L_CO, "Pid %d [id %s] seems to hang", ch->pid,
				ch->id);
		else {
#if DEBUG
		    i_log(L_VB, "Updating utmp for pid %d [id %s]", ch->pid, ch->id);
#endif
		    ch->flags &= ~RUNNING;
#ifdef BB_FEATURE_WUTMP
		    if (ch->process[0] != '+')
			write_utmp_wtmp("", ch->id, ch->pid, DEAD_PROCESS, NULL);
#endif
		}
	}

  /* Both rounds done; clean up the list. */
  nmask = bb_sigblock(1 << SIGCHLD);
  for(ch = family; ch; ch = old) {
	old = ch->next;
	free(ch);
  }
  family = newFamily;
  for(ch = family; ch; ch = ch->next) ch->new = NULL;
  newFamily = NULL;
  bb_sigunblock(nmask);
}

/*
 *      Walk through the family list and start up children.
 *      The entries that do not belong here at all are removed
 *      from the list.
 */
static void start_if_needed(void)
{
  CHILD *ch;            /* Pointer to child */
  int delete;           /* Delete this entry from list? */

#if DEBUG
  i_log(L_VB, "Checking for children to start");
#endif

  for(ch = family; ch; ch = ch->next) {

#if DEBUG
	if (ch->rlevel[0] == 'C') {
		i_log(L_VB, "%s: flags %d", ch->process, ch->flags);
	}
#endif

	/* Are we waiting for this process? Then quit here. */
	if (ch->flags & WAITING) break;

	/* Already running? OK, don't touch it */
	if (ch->flags & RUNNING) continue;

	/* See if we have to start it up */
	delete = 1;
	if (strchr(ch->rlevel, bb_common_bss.runlevel) ||
	    ((ch->flags & DEMAND) && !strchr("#*Ss", bb_common_bss.runlevel))) {
		startup(ch);
		delete = 0;
	}

	if (delete) {
		/* FIXME: is this OK? */
		ch->flags &= ~(RUNNING|WAITING);
		if (!ISPOWER(ch->action) && ch->action != KBREQUEST)
			ch->flags &= ~XECUTED;
		ch->pid = 0;
	} else
		/* Do we have to wait for this process? */
		if (ch->flags & WAITING) break;
  }
  /* Done. */
}

/* Ask the user on the console for a runlevel */
static int ask_runlevel(void)
{
	const char      prompt[] = "\nEnter runlevel: ";
	char            buf[8];
	int             lvl = -1;
	int             fd;

	set_term(0);
	fd = console_open(O_RDWR|O_NOCTTY);

	if (fd < 0) return('S');

	while(!strchr("0123456789S", lvl)) {
		write(fd, prompt, sizeof(prompt) - 1);
		buf[0] = 0;
		read(fd, buf, sizeof(buf));
		if (buf[0] != 0 && (buf[1] == '\r' || buf[1] == '\n'))
			lvl = buf[0];
		if (islower(lvl)) lvl = toupper(lvl);
	}
	close(fd);
	return lvl;
}

/*
 *      Search the INITTAB file for the 'initdefault' field, with the default
 *      runlevel. If this fails, ask the user to supply a runlevel.
 */
static int get_init_default(void)
{
	CHILD *ch;
	int lvl = -1;
	char *p;

	/* Look for initdefault. */
	for(ch = family; ch; ch = ch->next)
		if (ch->action == INITDEFAULT) {
			p = ch->rlevel;
			while(*p) {
				if (*p > lvl) lvl = *p;
				p++;
			}
			break;
		}
	/* See if level is valid */
	if (lvl > 0) {
		if (islower(lvl)) lvl = toupper(lvl);
		if (strchr("0123456789S", lvl) == NULL) {
			i_log(L_VB, "Initdefault level '%c' is invalid", lvl);
			lvl = 0;
		}
	}
	/* Ask for runlevel on console if needed. */
	if (lvl <= 0) lvl = ask_runlevel();

	/* Log the fact that we have a runlevel now. */
	return lvl;
}


/* We got signaled. */
static int read_level(int arg)
{
  unsigned char foo = 'X';              /* Contents of INITLVL */
  CHILD *ch;                            /* Walk through list */
  int ok = 1;

  if (arg) foo = arg;

  if (islower(foo)) foo = toupper(foo);
  if (ok < 1 || ok > 2 || strchr("QS0123456789ABCU", foo) == NULL) {
	i_log(L_VB, "bad runlevel: %c", foo);
	return(bb_common_bss.runlevel);
  }

  /* Log this action */
  switch(foo) {
	case 'S':
		i_log(L_VB, "Going single user");
		break;
	case 'Q':
		i_log(L_SY, "Re-reading inittab");
		break;
	case 'A':
	case 'B':
	case 'C':
		i_log(L_SY, "Activating demand-procedures for '%c'", foo);
		break;
	case 'U':
		i_log(L_SY, "Trying to re-exec init");
		return 'U';
	default:
		i_log(L_VB, "Switching to runlevel: %c", foo);
  }

  if (foo == 'Q') return(bb_common_bss.runlevel);

  /* Check if this is a runlevel a, b or c */
  if (strchr("ABC", foo)) {
	if (bb_common_bss.runlevel == 'S') return(bb_common_bss.runlevel);

	/* Read inittab again first! */
	read_inittab();

	/* Mark those special tasks */
	for(ch = family; ch; ch = ch->next)
		if (strchr(ch->rlevel, foo) != NULL ||
		    strchr(ch->rlevel, tolower(foo)) != NULL) {
			ch->flags |= DEMAND;
			ch->flags &= ~XECUTED;
#if DEBUG
			i_log(L_VB, "Marking (%s) as ondemand, flags %d",
				ch->id, ch->flags);
#endif
		}
	return(bb_common_bss.runlevel);
  }

  /* Store both the old and the new runlevel. */
#ifdef BB_FEATURE_WUTMP
  write_utmp_wtmp("runlevel", "~~", foo + 256 * bb_common_bss.runlevel, RUN_LVL, "~");
#endif
  bb_common_bss.thislevel = foo;
  bb_common_bss.prevlevel = bb_common_bss.runlevel;
  return(foo);
}


/*
 *      This procedure is called after every signal (SIGHUP, SIGALRM..)
 *
 *      Only clear the 'failing' flag if the process is sleeping
 *      longer than 5 minutes, or inittab was read again due
 *      to user interaction.
 */
static void fail_check(void)
{
  time_t t;                     /* System time */
  CHILD *ch;                    /* Pointer to child structure */
  time_t next_alarm = 0;        /* When to set next alarm */

  time(&t);

  for(ch = family; ch; ch = ch->next) {

	if (ch->flags & FAILING) {
		/* Can we free this sucker? */
		if (ch->tm + SLEEPTIME < t) {
			ch->flags &= ~FAILING;
			ch->count = 0;
			ch->tm = 0;
		} else {
			/* No, we'll look again later */
			if (next_alarm == 0 || ch->tm + SLEEPTIME > next_alarm)
				next_alarm = ch->tm + SLEEPTIME;
		}
	}
  }
  if (next_alarm) {
	next_alarm -= t;
	if (next_alarm < 1) next_alarm = 1;
	alarm(next_alarm);
  }
}

/* Set all 'Fail' timers to 0 */
static void fail_cancel(void)
{
	CHILD *ch;

	for(ch = family; ch; ch = ch->next) {
		ch->count = 0;
		ch->tm = 0;
		ch->flags &= ~FAILING;
	}
}

/* Start up powerfail entries. */
static void do_power_fail(int pwrstat)
{
	CHILD *ch;

	/* Tell powerwait & powerfail entries to start up */
	for (ch = family; ch; ch = ch->next) {
		if (pwrstat == 'O') {
			/* The power is OK again. */
			if (ch->action == POWEROKWAIT)
				ch->flags &= ~XECUTED;
		} else if (pwrstat == 'L') {
			/* Low battery, shut down now. */
			if (ch->action == POWERFAILNOW)
				ch->flags &= ~XECUTED;
		} else {
			/* Power is failing, shutdown imminent */
			if (ch->action == POWERFAIL || ch->action == POWERWAIT)
				ch->flags &= ~XECUTED;
		}
	}
}

/* Check for state-pipe presence */
static int check_pipe(int fd)
{
	struct timeval t;
	fd_set s;
	char signature[8];

	FD_ZERO(&s);
	FD_SET(fd, &s);
	t.tv_sec = t.tv_usec = 0;

	if (select(fd+1, &s, NULL, NULL, &t) != 1)
		return 0;
	if (read(fd, signature, 8) != 8)
		 return 0;
	return strncmp(Signature, signature, 8) == 0;
}

/* Make a state-pipe. */
static int make_pipe(int fd)
{
	int fds[2];

	pipe(fds);
	dup2(fds[0], fd);
	close(fds[0]);
	fcntl(fds[1], F_SETFD, 1);
	fcntl(fd, F_SETFD, 0);
	write(fds[1], Signature, 8);

	return fds[1];
}

/* Attempt to re-exec. */
static void re_exec(void)
{
	sigset_t mask, oldset;
	pid_t   pid;
	int     fd;
	CHILD   *ch;

	if (strchr("S12345", bb_common_bss.runlevel) == NULL)
		return;
	closelog();                     /* AAAARRRGH! It _is_ needed */

	/* Reset the alarm, and block all signals. */
	alarm(0);
	sigfillset(&mask);
	sigprocmask(SIG_BLOCK, &mask, &oldset);

	/* construct a pipe fd --> STATE_PIPE and write a signature */
	fd = make_pipe(STATE_PIPE);

	/* It's a backup day today, so I'm pissed off.  Being a BOFH, however,
	 * does have it's advantages...
	 */
	fail_cancel();
	close(pipe_fd);
	pipe_fd = -1;
	DELSET(got_signals, SIGCHLD);
	DELSET(got_signals, SIGHUP);
	DELSET(got_signals, SIGUSR1);

	/* That should be cleaned. */
	for(ch = family; ch; ch = ch->next)
	    if (ch->flags & ZOMBIE) {
#if DEBUG
		i_log(L_VB, "Child died, PID= %d", ch->pid);
#endif
		ch->flags &= ~(RUNNING|ZOMBIE|WAITING);
#ifdef BB_FEATURE_WUTMP
		if (ch->process[0] != '+')
			write_utmp_wtmp("", ch->id, ch->pid, DEAD_PROCESS, NULL);
#endif
	    }

	if ((pid = fork()) > 0) {
		/* Yup. _Parent_ exec's ... */
		execl(bb_common_bss.myname, bb_common_bss.myname, NULL);
	} else if (pid == 0) {
		/* ... while child sends her the
		 * state information and dies
		 */
		send_state(fd);
		exit(0);
	}
	/*
	 *      We shouldn't be here, something failed.
	 *      Bitch, close the state pipe, unblock signals and return.
	 */
	close(fd);
	close(STATE_PIPE);
	sigprocmask(SIG_SETMASK, &oldset, NULL);
	i_log(L_CO, "Attempt to re-exec failed");
}


/*
 *      We got a change runlevel request through the
 *      init.fifo. Process it.
 */
static void fifo_new_level(int level)
{
  int oldlevel;
#if CHANGE_WAIT
  CHILD *ch;
#endif

  if (level == bb_common_bss.runlevel) return;

#if CHANGE_WAIT
  /* Are we waiting for a child? */
  for(ch = family; ch; ch = ch->next)
	if (ch->flags & WAITING) break;
  if (ch == NULL)
#endif
  {
	/* We need to go into a new runlevel */
	oldlevel = bb_common_bss.runlevel;
	bb_common_bss.runlevel = read_level(level);
	if (bb_common_bss.runlevel == 'U') {
		bb_common_bss.runlevel = oldlevel;
		re_exec();
	} else {
		if (oldlevel != 'S' && bb_common_bss.runlevel == 'S') set_term(0);
		if (bb_common_bss.runlevel == '6' || bb_common_bss.runlevel == '0' || bb_common_bss.runlevel == '1') set_term(0);
		read_inittab();
		fail_cancel();
		bb_setproctitle("init [%c]", bb_common_bss.runlevel);
	}
  }
}

/*
 *      Read from the init FIFO. Processes like telnetd and rlogind can
 *      ask us to create login processes on their behalf.
 *
 *      FIXME:  this needs to be finished. NOT that it is buggy, but we need
 *              to add the telnetd/rlogind stuff so people can start using it.
 *              Maybe move to using an AF_UNIX socket so we can use
 *              the 2.2 kernel credential stuff to see who we're talking to.
 *
 */
static void check_init_fifo(void)
{
  struct init_request request;
  int n;
  fd_set fds;
  int quit = 0;
  struct stat st, st2;
  struct timeval tv;

  /* First, try to create /dev/initctl if not present. */
  if (stat(INIT_FIFO, &st2) < 0 && errno == ENOENT)
	(void)mkfifo(INIT_FIFO, 0600);

  /*
   *    If /dev/initctl is open, stat the file to see if it
   *    is still the _same_ inode.
   */
  if (pipe_fd >= 0) {
	fstat(pipe_fd, &st);
	if (stat(INIT_FIFO, &st2) < 0 ||
	    st.st_dev != st2.st_dev ||
	    st.st_ino != st2.st_ino) {
		close(pipe_fd);
		pipe_fd = -1;
	}
  }

  /*
   *    Now finally try to open /dev/initctl
   */
  if (pipe_fd < 0) {
	if ((pipe_fd = open(INIT_FIFO, O_RDWR|O_NONBLOCK)) >= 0) {
		fstat(pipe_fd, &st);
		if (!S_ISFIFO(st.st_mode)) {
			i_log(L_VB, "%s is not a fifo", INIT_FIFO);
			close(pipe_fd);
			pipe_fd = -1;
		}
	}
	if (pipe_fd >= 0) {
		/*
		 *      Don't use fd's 0, 1 or 2.
		 */
		(void) dup2(pipe_fd, PIPE_FD);
		close(pipe_fd);
		pipe_fd = PIPE_FD;

		/*
		 *      Return to caller - we'll be back later.
		 */
	}
  }

  /* Wait for data to appear, _if_ the pipe was opened. */
  if (pipe_fd >= 0) while(!quit) {

	/* Do select, return on EINTR. */
	FD_ZERO(&fds);
	FD_SET(pipe_fd, &fds);
	tv.tv_sec = 5;
	tv.tv_usec = 0;
	n = select(pipe_fd + 1, &fds, NULL, NULL, &tv);
	if (n <= 0) {
		if (n == 0 || errno == EINTR) return;
		continue;
	}

	/* Read the data, return on EINTR. */
	n = read(pipe_fd, &request, sizeof(request));
	if (n == 0) {
		/*
		 *      End of file. This can't happen under Linux (because
		 *      the pipe is opened O_RDWR - see select() in the
		 *      kernel) but you never know...
		 */
		close(pipe_fd);
		pipe_fd = -1;
		return;
	}
	if (n <= 0) {
		if (errno == EINTR) return;
		i_log(L_VB, "error reading initrequest");
		continue;
	}

	/*
	 *      This is a convenient point to also try to
	 *      find the console device or check if it changed.
	 */
	console_init();

	/*
	 *      Process request.
	 */
	if (request.magic != INIT_MAGIC || n != sizeof(request)) {
		i_log(L_VB, "got bogus initrequest");
		continue;
	}
	switch(request.cmd) {
		case INIT_CMD_RUNLVL:
			bb_common_bss.sltime = request.sleeptime;
			fifo_new_level(request.runlevel);
			quit = 1;
			break;
		case INIT_CMD_POWERFAIL:
			bb_common_bss.sltime = request.sleeptime;
			do_power_fail('F');
			quit = 1;
			break;
		case INIT_CMD_POWERFAILNOW:
			bb_common_bss.sltime = request.sleeptime;
			do_power_fail('L');
			quit = 1;
			break;
		case INIT_CMD_POWEROK:
			bb_common_bss.sltime = request.sleeptime;
			do_power_fail('O');
			quit = 1;
			break;
		default:
			i_log(L_VB, "got unimplemented initrequest.");
			break;
	}
  }

  /*
   *    We come here if the pipe couldn't be opened.
   */
  if (pipe_fd < 0) pause();

}


/*
 *      This function is used in the transition
 *      sysinit (-> single user) boot -> multi-user.
 */
static void boot_transitions(void)
{
  CHILD *ch;
  int loglevel;
  int oldlevel;

  /* Check if there is something to wait for! */
  for( ch = family; ch; ch = ch->next )
	if ((ch->flags & RUNNING) && ch->action != BOOT) break;

  if (ch == NULL) {
	/* No processes left in this level, proceed to next level. */
	loglevel = -1;
	oldlevel = 'N';
	switch(bb_common_bss.runlevel) {
		case '#': /* SYSINIT -> BOOT */
#if DEBUG
			i_log(L_VB, "SYSINIT -> BOOT");
#endif
			/* Save tty modes. */
			set_term(1);
#ifdef BB_FEATURE_WUTMP
			/* Write a boot record. */
#ifdef CONFIG_FEATURE_UTMP
			wrote_utmp_reboot = 0;
#endif
#ifdef CONFIG_FEATURE_WTMP
			wrote_wtmp_reboot = 0;
#endif
			write_utmp_wtmp("reboot", "~~", 0, BOOT_TIME, "~");
#endif

			/* Get our run level */
			newlevel = dfl_level ? dfl_level : get_init_default();
			if (newlevel == 'S') {
				bb_common_bss.runlevel = newlevel;
				/* Not really 'S' but show anyway. */
				bb_setproctitle("init [S]");
			} else
				bb_common_bss.runlevel = '*';
			break;
		case '*': /* BOOT -> NORMAL */
#if DEBUG
			i_log(L_VB, "BOOT -> NORMAL");
#endif
			/* Save tty modes. */
			set_term(1);

			if (bb_common_bss.runlevel != newlevel)
				loglevel = newlevel;
			bb_common_bss.runlevel = newlevel;
			did_boot = 1;
			warn = 1;
			break;
		case 'S': /* Ended SU mode */
		case 's':
#if DEBUG
			i_log(L_VB, "END SU MODE");
#endif
			/* Save tty modes. */
			set_term(1);

			newlevel = get_init_default();
			if (!did_boot && newlevel != 'S')
				bb_common_bss.runlevel = '*';
			else {
				if (bb_common_bss.runlevel != newlevel)
					loglevel = newlevel;
				bb_common_bss.runlevel = newlevel;
				oldlevel = 'S';
			}
			warn = 1;
			for(ch = family; ch; ch = ch->next)
			    if (strcmp(ch->rlevel, "S") == 0)
				ch->flags &= ~(FAILING|WAITING|XECUTED);
			break;
		default:
			if (warn)
			  i_log(L_VB, "no more processes left in this runlevel");
			warn = 0;
			loglevel = -1;
			if (got_signals == 0)
				check_init_fifo();
			break;
	}
	if (loglevel > 0) {
		i_log(L_VB, "Entering runlevel: %c", bb_common_bss.runlevel);
#ifdef BB_FEATURE_WUTMP
		write_utmp_wtmp("runlevel", "~~", bb_common_bss.runlevel + 256 * oldlevel, RUN_LVL, "~");
#endif
		bb_common_bss.thislevel = bb_common_bss.runlevel;
		bb_common_bss.prevlevel = oldlevel;
		bb_setproctitle("init [%c]", bb_common_bss.runlevel);
	}
  }
}

/*
 *      Init got hit by a signal. See which signal it is,
 *      and act accordingly.
 */
static void process_signals(void)
{
  int oldlevel;
  CHILD *ch;

  if (ISMEMBER(got_signals, SIGPWR)) {
#if DEBUG
	i_log(L_VB, "got SIGPWR");
#endif
	/* See _what_ kind of SIGPWR this is. */
	do_power_fail(0);
	DELSET(got_signals, SIGPWR);
  }

  if (ISMEMBER(got_signals, SIGINT)) {
#if DEBUG
	i_log(L_VB, "got SIGINT");
#endif
	/* Tell ctrlaltdel entry to start up */
	for(ch = family; ch; ch = ch->next)
		if (ch->action == CTRLALTDEL)
			ch->flags &= ~XECUTED;
	DELSET(got_signals, SIGINT);
  }

  if (ISMEMBER(got_signals, SIGWINCH)) {
#if DEBUG
	i_log(L_VB, "got SIGWINCH");
#endif
	/* Tell kbrequest entry to start up */
	for(ch = family; ch; ch = ch->next)
		if (ch->action == KBREQUEST)
			ch->flags &= ~XECUTED;
	DELSET(got_signals, SIGWINCH);
  }

  if (ISMEMBER(got_signals, SIGALRM)) {
#if DEBUG
	i_log(L_VB, "got SIGALRM");
#endif
	/* The timer went off: check it out */
	DELSET(got_signals, SIGALRM);
  }

  if (ISMEMBER(got_signals, SIGCHLD)) {
#if DEBUG
	i_log(L_VB, "got SIGCHLD");
#endif
	/* First set flag to 0 */
	DELSET(got_signals, SIGCHLD);

	/* See which child this was */
	for(ch = family; ch; ch = ch->next)
	    if (ch->flags & ZOMBIE) {
#if DEBUG
		i_log(L_VB, "Child died, PID= %d", ch->pid);
#endif
		ch->flags &= ~(RUNNING|ZOMBIE|WAITING);
#ifdef BB_FEATURE_WUTMP
		if (ch->process[0] != '+')
			write_utmp_wtmp("", ch->id, ch->pid, DEAD_PROCESS, NULL);
#endif
	    }

  }

  if (ISMEMBER(got_signals, SIGHUP)) {
#if DEBUG
	i_log(L_VB, "got SIGHUP");
#endif
#if CHANGE_WAIT
	/* Are we waiting for a child? */
	for(ch = family; ch; ch = ch->next)
		if (ch->flags & WAITING) break;
	if (ch == NULL)
#endif
	{
		/* We need to go into a new runlevel */
		oldlevel = bb_common_bss.runlevel;
		if (bb_common_bss.runlevel == 'U') {
			bb_common_bss.runlevel = oldlevel;
			re_exec();
		} else {
			if (oldlevel != 'S' && bb_common_bss.runlevel == 'S') set_term(0);
			if (bb_common_bss.runlevel == '6' || bb_common_bss.runlevel == '0' || bb_common_bss.runlevel == '1') set_term(0);
			read_inittab();
			fail_cancel();
			bb_setproctitle("init [%c]", bb_common_bss.runlevel);
			DELSET(got_signals, SIGHUP);
		}
	}
  }
  if (ISMEMBER(got_signals, SIGUSR1)) {
	/*
	 *      SIGUSR1 means close and reopen /dev/initctl
	 */
#if DEBUG
	i_log(L_VB, "got SIGUSR1");
#endif
	close(pipe_fd);
	pipe_fd = -1;
	DELSET(got_signals, SIGUSR1);
  }
}

/*
 *      The main loop
 */
static int init_main_loop(void) __attribute__ ((noreturn));
static int init_main_loop(void)
{
  int f, st;
  CHILD *ch;
  sigset_t sgt;
  unsigned int nsig_msk;

  if (!reload) {

#if INITDEBUG
	/*
	 * Fork so we can debug the init process.
	 */
	if ((f = fork()) > 0) {
		static const char killmsg[] = "PRNT: init killed.\r\n";

		while(wait(&st) != f)
			;
		write(1, killmsg, sizeof(killmsg) - 1);
		while(1) pause();
	}
#endif

	/* Tell the kernel to send us SIGINT when CTRL-ALT-DEL
	 * is pressed, and that we want to handle keyboard signals. */
	reboot(BMAGIC_SOFT);
	if ((f = open(VT_MASTER, O_RDWR | O_NOCTTY)) >= 0) {
		(void) ioctl(f, KDSIGACCEPT, SIGWINCH);
		close(f);
	} else
		(void) ioctl(0, KDSIGACCEPT, SIGWINCH);

	/* Ignore all signals. */
	nsig_msk = 0;
	for(f = 1; f < NSIG; f++)
		nsig_msk |= (1 << f);
	bb_signal(nsig_msk, SIG_IGN);
  }

  bb_signal((1 << SIGALRM) | (1 << SIGHUP) | (1 << SIGINT) |
	(1 << SIGPWR) | (1 << SIGWINCH) | (1 << SIGUSR1), signal_handler);
  bb_signal((1 << SIGCHLD), chld_handler);
  bb_signal((1 << SIGSTOP) | (1 << SIGTSTP), stop_handler);
  bb_signal((1 << SIGCONT), cont_handler);
  bb_signal((1 << SIGSEGV), segv_handler);

  console_init();

  if (!reload) {

	/* Close whatever files are open, and reset the console. */
	close(0);
	close(1);
	close(2);
	set_term(0);
	setsid();

	/* Set default PATH variable */
	if (getenv("PATH") == NULL) putenv(PATH_DFL);

#ifdef CONFIG_FEATURE_UTMP
	/*
	 *      Initialize /var/run/utmp (only works if /var is on
	 *      root and mounted rw)
	 */
	(void) close(open(UTMP_FILE, O_WRONLY|O_CREAT|O_TRUNC, 0644));
#endif
	/* Say hello to the world */
	i_log(L_CO, bootmsg, "booting");

	/*
	 *      See if we have to start an emergency shell.
	 */
	if (emerg_shell) {
		bb_signal((1 << SIGCHLD), SIG_DFL);
		if (spawn(&ch_emerg, &f) > 0) {
			while(wait(&st) != f)
				;
		}
		bb_signal((1 << SIGCHLD), chld_handler);
	}

	/*
	 *      Start normal boot procedure.
	 */
	bb_common_bss.runlevel = '#';
	read_inittab();

  } else {
	/*
	 *      Restart: unblock signals and let the show go on
	 */
	i_log(L_CO, bootmsg, "reloading");
	sigfillset(&sgt);
	bb_sigunblock(sgt);
  }
  start_if_needed();

  while(1) {

     /* See if we need to make the boot transitions. */
     boot_transitions();
#if DEBUG
     i_log(L_VB, "init_main_loop: waiting..");
#endif

     /* Check if there are processes to be waited on. */
     for(ch = family; ch; ch = ch->next)
	if ((ch->flags & RUNNING) && ch->action != BOOT) break;

#if CHANGE_WAIT
     /* Wait until we get hit by some signal. */
     while (ch != NULL && got_signals == 0) {
	if (ISMEMBER(got_signals, SIGHUP)) {
		/* See if there are processes to be waited on. */
		for(ch = family; ch; ch = ch->next)
			if (ch->flags & WAITING) break;
	}
	if (ch != NULL) check_init_fifo();
     }
#else /* CHANGE_WAIT */
     if (ch != NULL && got_signals == 0) check_init_fifo();
#endif /* CHANGE_WAIT */

     /* Check the 'failing' flags */
     fail_check();

     /* Process any signals. */
     process_signals();

     /* See what we need to start up (again) */
     start_if_needed();
  }
  /*NOTREACHED*/
}

/*
 * Main entry for init and telinit.
 */
int init_main(void)
{
  char **argv = Gargv;
  struct init_request   request;
  int                   f, fd;

  /* setup the data segment */
  bb_common_bss.runlevel = 'S';            /* The current run level */
  bb_common_bss.thislevel = 'S';           /* The current runlevel */
  bb_common_bss.prevlevel = 'N';           /* Previous runlevel */
  bb_common_bss.sltime = 5;                /* Sleep time between TERM and KILL */
  pipe_fd = -1;                            /* /dev/initctl */
  bb_common_bss.myname = "/sbin/init";
  warn = 1;
  strcpy(ch_emerg.id, "~~");
  strcpy(ch_emerg.rlevel, "S");
  ch_emerg.action = 3;
  strcpy(ch_emerg.process, "/sbin/sulogin");
#ifdef CONFIG_FEATURE_UTMP
  wrote_utmp_reboot = 1;
#endif
#ifdef CONFIG_FEATURE_WTMP
  wrote_wtmp_reboot = 1;
#endif

  /*
   *    Is this telinit or init ?
   */
#ifndef TEST
  if (getpid() == INITPID ||
	!strcmp(bb_applet_name, "init.new") || !strcmp(bb_applet_name, "sh"))
#endif
  {
	/*
	 *      Check for re-exec
	 */
	if (check_pipe(STATE_PIPE)) {

		receive_state(STATE_PIPE);

		bb_common_bss.myname = strdup(argv[0]);
		reload = 1;
		wrote_iosave = 1;
		bb_setproctitle("init [%c]", bb_common_bss.runlevel);

		init_main_loop();
	}

	/* Check command line arguments */
	for(f = 1; argv[f] != NULL; f++) {
		if (!strcmp(argv[f], "single") || !strcmp(argv[f], "-s"))
			dfl_level = 'S';
		else if (!strcmp(argv[f], "-a") || !strcmp(argv[f], "auto"))
			putenv("AUTOBOOT=YES");
		else if (!strcmp(argv[f], "-b") || !strcmp(argv[f], "emergency"))
			emerg_shell = 1;
		else if (!strcmp(argv[f], "-z")) {
			/* Ignore -z xxx */
			if (argv[f + 1]) f++;
		} else if (strchr("0123456789sS", argv[f][0])
			&& strlen(argv[f]) == 1)
			dfl_level = argv[f][0];
		/* "init u" in the very beginning makes no sense */
		if (dfl_level == 's') dfl_level = 'S';
	}

	/* Start booting. */
	argv[1] = NULL;
	bb_setproctitle("init boot");
	init_main_loop();
  }

  /* Nope, this is a change-run-level init */
  bb_xgetopt_ulflags("d:\0:-d:=1", &bb_common_bss.sltime);
  if (geteuid() != 0)
	bb_error_msg_and_die("must be superuser.");
  argv = Gargv;
  /* Check syntax. */
  if (strlen(*argv) != 1) bb_show_usage();
  if (!strchr("0123456789SsQqAaBbCcUu", argv[0][0])) bb_show_usage();

  umask(022);

  /* Open the fifo and write a command. */
  memset(&request, 0, sizeof(request));
  request.magic     = INIT_MAGIC;
  request.cmd       = INIT_CMD_RUNLVL;
  request.runlevel  = argv[0][0];
  request.sleeptime = bb_common_bss.sltime;

  /* Make sure we don't hang on opening INIT_FIFO */
  bb_signal( (1 << SIGALRM), signal_handler);
  alarm(3);
  if ((fd = open(INIT_FIFO, O_WRONLY)) >= 0 &&
		write(fd, &request, sizeof(request)) == sizeof(request)) {
	close(fd);
	alarm(0);
	return 0;
  }
  bb_perror_msg_and_die("%s", INIT_FIFO);
}


#ifdef CONFIG_FEATURE_WTMP

#if defined(__GLIBC__)
#  if (__GLIBC__ == 2) && (__GLIBC_MINOR__ == 0) && defined(__powerpc__)
#    define HAVE_UPDWTMP 0
#  else
#    define HAVE_UPDWTMP 1
#  endif
#else
#  define HAVE_UPDWTMP 0
#endif


/*
 *      Log an event in the wtmp file (reboot, runlevel)
 */
static void write_wtmp(
char *user,                     /* name of user */
char *id,                       /* inittab ID */
int pid,                        /* PID of process */
int type,                       /* TYPE of entry */
char *line)                     /* Which line is this */
{
	int fd;
	struct utmp utmp;
	struct utsname uname_buf;
	time_t t;

	/*
	 *      Try to open the wtmp file. Note that we even try
	 *      this if we have updwtmp() so we can see if the
	 *      wtmp file is accessible.
	 */
	if ((fd = open(WTMP_FILE, O_WRONLY|O_APPEND)) < 0) return;

	/*
	 *      Note if we are going to write a boot record.
	 */
	if (type == BOOT_TIME) wrote_wtmp_reboot++;

	/*
	 *      See if we need to write a reboot record. The reason that
	 *      we are being so paranoid is that when we first tried to
	 *      write the reboot record, /var was possibly not mounted
	 *      yet. As soon as we can open WTMP we write a delayed boot record.
	 */
	if (wrote_wtmp_reboot == 0 && type != BOOT_TIME)
		write_wtmp("reboot", "~~", 0, BOOT_TIME, "~");

	/*
	 *      Zero the fields and enter new fields.
	 */
	memset(&utmp, 0, sizeof(utmp));
	time(&t);
	utmp.ut_time = t;
	utmp.ut_pid  = pid;
	utmp.ut_type = type;
	strncpy(utmp.ut_name, user, sizeof(utmp.ut_name));
	strncpy(utmp.ut_id  , id  , sizeof(utmp.ut_id  ));
	strncpy(utmp.ut_line, line, sizeof(utmp.ut_line));

	/* Put the OS version in place of the hostname */
	if (uname(&uname_buf) == 0)
		strncpy(utmp.ut_host, uname_buf.release, sizeof(utmp.ut_host));

#if HAVE_UPDWTMP
	updwtmp(WTMP_FILE, &utmp);
#else
	write(fd, (char *)&utmp, sizeof(utmp));
#endif
	close(fd);
}
#endif


#ifdef CONFIG_FEATURE_UTMP
/*
 *      Write an entry to the UTMP file. For DEAD_PROCESS, put
 *      the previous ut_line into oldline if oldline != NULL.
 */
static void write_utmp(
char *user,                     /* name of user */
char *id,                       /* inittab ID */
int pid,                        /* PID of process */
int type,                       /* TYPE of entry */
char *line,                     /* LINE if used. */
char *oldline)                  /* Line of old utmp entry. */
{
	struct utmp utmp;
	struct utmp tmp;
	struct utmp *utmptr;
	time_t t;

	/*
	 *      Can't do much if UTMP_FILE is not present.
	 */
	if (access(UTMP_FILE, F_OK) < 0)
		return;

	/*
	 *      Note if we are going to write a boot record.
	 */
	if (type == BOOT_TIME) wrote_utmp_reboot++;

	/*
	 *      See if we need to write a reboot record. The reason that
	 *      we are being so paranoid is that when we first tried to
	 *      write the reboot record, /var was possibly not mounted
	 *      yet. As soon as we can open WTMP we write a delayed boot record.
	 */
	if (wrote_utmp_reboot == 0 && type != BOOT_TIME)
		write_utmp("reboot", "~~", 0, BOOT_TIME, "~", NULL);

	/*
	 *      Fill out an utmp struct.
	 */
	memset(&utmp, 0, sizeof(utmp));
	utmp.ut_type = type;
	utmp.ut_pid = pid;
	strncpy(utmp.ut_id, id, sizeof(utmp.ut_id));
	time(&t);
	utmp.ut_time = t;
	strncpy(utmp.ut_user, user, UT_NAMESIZE);
	if (line) strncpy(utmp.ut_line, line, UT_LINESIZE);

	/*
	 *      We might need to find the existing entry first, to
	 *      find the tty of the process (for wtmp accounting).
	 */
	if (type == DEAD_PROCESS) {
		/*
		 *      Find existing entry for the tty line.
		 */
		setutent();
		tmp = utmp;
		if ((utmptr = getutid(&tmp)) != NULL) {
			strncpy(utmp.ut_line, utmptr->ut_line, UT_LINESIZE);
			if (oldline)
				strncpy(oldline, utmptr->ut_line, UT_LINESIZE);
		}
	}

	/*
	 *      Update existing utmp file.
	 */
	setutent();
	pututline(&utmp);
	endutent();
}
#endif

#ifdef BB_FEATURE_WUTMP
/*
 *      Write a record to both utmp and wtmp.
 */
static void write_utmp_wtmp(
char *user,                     /* name of user */
char *id,                       /* inittab ID */
int pid,                        /* PID of process */
int type,                       /* TYPE of entry */
char *line)                     /* LINE if used. */
{
	char    oldline[UT_LINESIZE];

	/*
	 *      For backwards compatibility we just return
	 *      if user == NULL (means : clean up utmp file).
	 */
	if (user == NULL)
		return;

	oldline[0] = 0;
#ifdef CONFIG_FEATURE_UTMP
	write_utmp(user, id, pid, type, line, oldline);
#endif
#ifdef CONFIG_FEATURE_WTMP
	write_wtmp(user, id, pid, type, line && line[0] ? line : oldline);
#endif
}
#endif /* BB_FEATURE_WUTMP */
#endif /* APPLET_USAGE */
