/*
 * password utility routines.
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 * Tito Ragusa (c) 2004 <farmatito@tiscali.it>
 * Vladimir Oleynik (c) 2005,2016 <dzo@simtreas.ru>
 *
 * Licensed under the GPL v2, see the file LICENSE in this tarball.
 */

#include "libbb.h"
#include <string.h>
#include "grp_.h"
#include <assert.h>

/*
 * if bufsize is > 0 char *name cannot be set to NULL.
 *                   On success (user|group)name is written to 'name'
 *                   (and a pointer to it is returned).
 *                   On failure id as string is written to 'name'
 *                   buffer and NULL is returned.
 * if bufsize is = 0 char *name can be set to NULL.
 *                   On success groupname is returned.
 *                   On failure pointer 'name' is returned.
 * if bufsize is < 0 equivalent as bufsize = 0, but
 *                   on failure an error message is printed and
 *                   the program exits.
 */

#ifdef L_bb_getug
/* internal function for bb_getpwuid and bb_getgrgid */
static char *bb_getug(char *buffer, char *idname, long id, int bufsize, char prefix)
{
	if(bufsize > 0 ) {
		assert(buffer!=NULL);
		if(idname) {
			return safe_strncpy(buffer, idname, bufsize);
		}
		snprintf(buffer, bufsize, "%ld", id);
	} else if(bufsize < 0 && !idname) {
		bb_error_msg_and_die("unknown %cid %ld", prefix, id);
	}
	return idname;
}

/* gets a username given a uid */
FAST_FUNC char * bb_getpwuid(char *name, long uid, int bufsize)
{
	struct passwd *myuser = getpwuid(uid);

	return bb_getug(name,
			myuser != NULL ? myuser->pw_name : (bufsize == 0 ? name : NULL),
			uid, bufsize, 'u');
}

/* gets a groupname given a gid */
FAST_FUNC char * bb_getgrgid(char *name, long gid, int bufsize)
{
	struct group *mygroup = getgrgid(gid);

	return bb_getug(name,
			mygroup != NULL ? mygroup->gr_name : (bufsize == 0 ? name : NULL),
			gid, bufsize, 'g');
}
#endif /* L_bb_getug */

#ifdef L_bb_xgetgrnam
/* returns a gid given a group name */
long bb_xgetgrnam(const char *name)
{
	struct group *mygroup = getgrnam(name);

	if (mygroup == NULL)
		bb_error_msg_and_die("unknown group name: %s", name);

	return mygroup->gr_gid;
}
#endif /* L_bb_xgetgrnam */

#ifdef L_bb_xgetpwnam
/* returns a uid given a username */
long bb_xgetpwnam(const char *name)
{
	struct passwd *myuser = getpwnam(name);

	if (myuser == NULL)
		bb_error_msg_and_die("unknown user name: %s", name);

	return myuser->pw_uid;
}
#endif /* L_bb_xgetpwnam */


#ifdef L_get_ug_id
/* indirect dispatcher for pwd helpers.  */

extern unsigned long get_ug_id(const char *s,
		long (*__bb_getxxnam)(const char *))
{
	unsigned long r;
	char *p;

	r = strtoul(s, &p, 10);
	if (*p || (s == p)) {
		r = __bb_getxxnam(s);
	}

	return r;
}
#endif /* L_get_ug_id */
