/*
 * Copyright (c) 2014-2016 may safely be consumed by a BSD or GPL license.
 * Written by:   Vladimir Oleynik <dzo@simtreas.ru>
 *
 * Documentation:
 *
 *  *config_read_prepare(const char *filename, int flg);
 * Prepare for read config file `filename'.
 *
 * `flg' is encoding: '-' or 'e' or 'w' or 'a' or 'S' or 'm' or any character:
 *  '-' - read stdin, filename unused and may be set to NULL;
 *  'e' - show error problem while open and die;
 *  'w' - show error problem while open and return NULL;
 *  'a' - already fopen, must insert FILE ptr to returned config_read_t ptr->in manually.
 *       filename unused and may be set to NULL;
 *  'S' - parse from string, input string for parse is strdup(filename)
 *  'm' - parse from string, input string for parse is not constant(!) string used
 *        by `filename`
 *  OTHER - fopen() and return NULL if found problem silent.
 *
 * if need do not close after found EOF, set to non zero into flag returned
 * config_read_t ptr->flg_no_close before call config_read()
 *
 * input string (S/m modes) must have not contain '\n'!
 *
 * config_read_t *config_read(config_read_t *crt, const char *nc);
 * call this into loop. First argument is struct returned of config_read_prepare()
 *
 * Second argument encoded:
 *  first char is start_rem_eol set non zero for detect comments from this character;
 *  next string of interested tokens, '~' for all, example \2~ encode skip first
 *       and all others;
 *  zero delimiter;
 *  number if need test minimal numbers of tokens in line or zero char;
 *  and optional add specified of special delimiter characters to harcoded posixly standart
 *      delimiters: " \t\v\f\r".
 *
 * config_read does:
 * - read a line with CR or EOF
 * - remove CR and last NL(\r)
 * - test a last char from line, if is \, then split with next a line
 * - cut line with start_rem_eol if non zero
 * - convert any '\t\v\f\r' to ' '
 * - find tokens in line separated by ' ' and specified add delimiter
 * - if tokens do not found then loop to begin internally (empty/remark line)
 * - if number need, test the minimal number of tokens in line" specified and
 *   found less this value then loop to begin internally (malformed config line)
 * - set crt->ntokens to number tokens in line, may be great of interested tokens
 * - set null-terminated array interested tokens to crt->tokens
 * - set crt->flags bitmask to found delims:
 *       CR_FLAGS_FOUND_SPACE_DELIM - one or more spaces
 *       CR_FLAGS_FOUND_SPEC_DELIM - one or more specified delim
 *       CR_FLAGS_FOUND_DOUBLE_SPEC_DELIM - more one specified delim found (may be error)
 *       CR_FLAGS_FOUND_EOL_DELIM - end of line/EOF found (may be error with previous bits set)
 * return:
 * - NULL if EOF and flg_no_close is not set or I/O error found
 * - crt with crt->ntokens=0 if EOF and flg_no_close is set (for rewind manually and rescan)
 *
 * nc may be different in any call config_read()
 *
 *
 * void config_close(config_read_t *crt);
 * This function do need call if loop config_read() is break or close for flg_no_close is set
 * config_close(NULL) usaging silent too. Example:

 config_read_t *crt = config_read_prepare(...);
 while( (crt = config_read(crt, ...)) != NULL) {
	...
	if(...)
		break;
	...
 }
 config_close(crt);

 * Examples:

 // open config file or die
 config_read_t *crt = config_read_prepare("config_file", 'e');

 // decode config_read nc encodes:
 // - config file have not remarks;
 // - interested tokes is 1 and 3 (skip the second token and others words);
 // - do not test minimal number of tokens in line;
 // - delimiters of tokens is: spaces only
 while( (crt = config_read(crt, "\0\1\3\0\0")) != NULL) {
	printf("First token is %s\n", crt->tokens[0]);
	if(crt->tokens[1] != NULL)
		printf("Third token is %s\n", crt->tokens[1]);
 }

 * Next example:
 * - format: var=value
 * - test need '=' and two tokens in line
 * - if VAR1 with any value found, rescan and save value for VAR2

 char *find_var(config_read_t *crt, const char *var)
 {
	// decode config_read nc encodes:
	// - config file may be have # as start remarks;
	// - interested tokes is 1 and 2
	// - delimiters of tokens is: spaces and '='
	while( 1 ) {
		crt = config_read(crt, "#\1\2\0\0=");
		if(crt == NULL)   // EOF
			break;
		// [ ]*#.* - valid, config_read() loop internally
		// [ ]*var[ ]*=[ ]*val[ ]* - valid
		// [ ]*var[ ]*=[ ]*val[ ]*#.* - valid, other - invalid
		if(crt->ntokens != 2 ||
			  (crt->flags[0] & (CR_FLAGS_FOUND_SPEC_DELIM|CR_FLAGS_FOUND_DOUBLE_SPEC_DELIM)) != CR_FLAGS_FOUND_SPEC_DELIM
			  || (crt->flags[1] & (CR_FLAGS_FOUND_SPEC_DELIM|CR_FLAGS_FOUND_EOL_DELIM)) != CR_FLAGS_FOUND_EOL_DELIM) {
			printf("strange syntax in line %d, need var=value\n", crt->num_line);
			continue;
		}
		if(!strcmp(crt->tokens[0], var)) {
			printf("%s=%s\n", var, crt->tokens[1]);
			return crt->tokens[1];
		}
	}
	return NULL;
 }

 config_read_t *crt = config_read_prepare("config_file", 'e');
 crt->flg_no_close = 1; // we call with rescan
 if(find_var(crt, "VAR1")) {
	rewind(crt->in);       // prepare for find next
	var2_value = find_var(crt, "VAR2");
	if(var2_value != NULL)
		var2_value = stdup(var2_value); // need save before close

 }
 config_close(crt);


 // interested tokes is all example:
 int i = 0;
 config_read_t *crt = config_read_prepare("config_file", 'w');
 while( (crt = config_read(crt, "\0~\0\0")) != NULL) {
	printf("%s =", crt->tokens[0]);
	for(i = 1; crt->tokens[i] != NULL; i++)
		printf(" %s", crt->tokens[i]);
	puts("");
 }
*/

#include "libbb.h"
#include <string.h>
#include <unistd.h>
#include "config_read.h"

#define BEGIN_SIZE 128

config_read_t *config_read_prepare(const char *filename, int flg)
{
	FILE *in = NULL;
	config_read_t *new_cr;
	static const char xmode[] = "EW-r";
	const char *mode = NULL;

	switch(flg) {
		case '-':
			if(stdin == NULL)
				return NULL;
			/* no break; */
		case 'S':
		case 'm':
			in = stdin;     /* fake flag */
			break;
		case 'a':
			/* set to NULL for set to aleady open later */
			break;
		case 'e':
			mode = xmode;           /* "EW-r" */
			break;
		case 'w':
			mode = xmode + 1;       /* "W-r" */
			break;
		default:
			mode = xmode + 2;       /* "-r" - no warn, no exit */
	}
	if(mode != NULL) {
		in = bb_xfopen(filename, mode);
		if(in == NULL)
			return NULL;
	}
	new_cr = xmalloc(sizeof(config_read_t));
	new_cr->in = in;
	new_cr->num_line = 0;
	new_cr->ntokens = 0;
	new_cr->tokens = NULL;
	new_cr->flags = NULL;
	new_cr->flg_no_close = 0;
	new_cr->max_tokens = 0;
	switch(flg) {
		case 'm':
			new_cr->in_buff_size = strlen(filename) + 1;
			new_cr->input_buff = (char *)filename;
			new_cr->flg_from_str = flg;
			break;
		case 'S':
			new_cr->in_buff_size = strlen(filename) + 1;
			new_cr->input_buff = xstrdup(filename);
			new_cr->flg_from_str = flg;
			break;
		default:
			new_cr->in_buff_size = BEGIN_SIZE;
			new_cr->input_buff = xmalloc(BEGIN_SIZE);
			new_cr->flg_from_str = 0;
			break;
	}
	new_cr->readed = 0;
	return new_cr;
}

void config_close(config_read_t *crt)
{
	if(crt) {
		free(crt->flags);
		free(crt->tokens);
		if(crt->flg_from_str) {
			if(crt->flg_from_str != 'm')
				free(crt->input_buff);
		} else {
			bb_fclose_nonstdin(crt->in);
		}
		free(crt);
	}
}

config_read_t *config_read(config_read_t *crt, const char *nc)
{
  size_t in_buff_size;
  char *input_buff;
  int readed;
  char *line;
  int token_number;
  int i;
  const char *nums_interest_tokens;
  int max_tokens;
  char *delims;
  char add_spec_delims[4];
  int test_max_tokens;
  int start_rem_eol;

  if(crt == NULL)
	return NULL;    /* can not open file */
  crt->ntokens = 0;
  in_buff_size = crt->in_buff_size;
  input_buff = crt->input_buff;
  readed = crt->readed;

  start_rem_eol = *nc++;
  nums_interest_tokens = nc;
  max_tokens = strlen(nc) + 1;  /* + NULL */
  test_max_tokens = nc[max_tokens];
  if(nc[max_tokens + 1]) {
	add_spec_delims[0] = nc[max_tokens + 1];
	i = 1;
  } else {
	i = 0;
  }
  add_spec_delims[i++] = ' ';
  add_spec_delims[i] = '\0';

  for(;;) {
    if(readed < 0) {
	  eof:
		if(crt->flg_no_close) {
			crt->readed = 0;
			return crt;
		}
		config_close(crt);
		return NULL;
    }

    for(;;) {
	if(crt->flg_from_str) {
		if(crt->in == stdin) {
			line = input_buff;      /* first */
			crt->in = NULL;
		} else {
			line = NULL;            /* "EOF" */
		}
	} else {
		line = fgets (input_buff + readed, in_buff_size - readed, crt->in);
	}
	if(line == NULL) {
		/* EOF or I/O error */
		if(readed == 0) {
			goto eof;
		} else {
			/* flag of EOF for next call */
			readed = -readed;
		}
	} else {
		line = strchr(input_buff, '\n');
		if(line == NULL) {
			/* string longest in_buff_size or EOF without '\n' */
			readed = strlen(input_buff);
			if(readed == (int)(in_buff_size - 1) && crt->flg_from_str == 0) {
				/* grow buffer */
				crt->input_buff = input_buff = xrealloc(input_buff, in_buff_size *= 2);
				crt->in_buff_size = in_buff_size;
			}
			continue;
		} else {
		 eat_tail:
			*line = '\0';
			if(line == input_buff) {
				readed = 0;
				crt->num_line++;
				continue;
			}
			--line;
			if(*line == '\r')
				goto eat_tail;
			if(*line == '\\') {
				/* string have flag of continue - tail \ */
				*line = '\0';
				readed = (line - input_buff);
				crt->num_line++;
				continue;
			}
			readed = 0;
		}
	}
	crt->num_line++;
	crt->readed = readed;
	break;
    }

    if(crt->max_tokens < max_tokens) {
	free(crt->tokens);
	free(crt->flags);
	crt->tokens = xmalloc(max_tokens * sizeof(char *));
	crt->flags = xmalloc(max_tokens);
	crt->max_tokens = max_tokens;
    }
    token_number = 0;
    line = input_buff;

    if(start_rem_eol != 0)
	strchr_set0(line, start_rem_eol);
#undef start_rem_eol

    /* posix standart delims for text configs */
    while ((line = strchr(line, '\t')) != NULL)
	*line++ = ' ';
    line = input_buff;
    while ((line = strchr(line, '\f')) != NULL)
	*line++ = ' ';
    line = input_buff;
    while ((line = strchr(line, '\v')) != NULL)
	*line++ = ' ';
    line = input_buff;
    while ((line = strchr(line, '\r')) != NULL)
	*line++ = ' ';
    line = input_buff;

    i = 0;
    delims = NULL;
    while (*line) {
	size_t token_size;
	char delim;

	token_size = strcspn (line, add_spec_delims);
	delim = line[token_size];
	line[token_size] = '\0';

	if (token_size) {
		token_number++;
		if(*nums_interest_tokens == '~') {
			if(i == (max_tokens-1)) {
				int j = i;

				max_tokens *= 2;
				crt->tokens = xrealloc(crt->tokens, max_tokens * sizeof(char *));
				crt->flags = xrealloc(crt->flags, max_tokens);
				while(j < max_tokens)
					crt->flags[j++] = '\0';
				crt->max_tokens = max_tokens;
			}
			goto set;
		} else if(token_number == *nums_interest_tokens) {
			nums_interest_tokens++;
		set:
			crt->tokens[i] = line;
			delims = crt->flags + i;
			*delims = 0;
			i++;
		} else {
			delims = NULL;
		}
	}
	if(delim == '\0') {
		if(delims)
			*delims |= CR_FLAGS_FOUND_EOL_DELIM;
		break;
	} else if(delim == ' ') {
		if(delims)
			*delims |= CR_FLAGS_FOUND_SPACE_DELIM;
	} else if(delims) {
		if((*delims & CR_FLAGS_FOUND_SPEC_DELIM) != 0)
			*delims |= CR_FLAGS_FOUND_DOUBLE_SPEC_DELIM;
		else
			*delims |= CR_FLAGS_FOUND_SPEC_DELIM;
	}
	/* shift */
	line += token_size + 1;
    }

    if(token_number != 0) {
	/* non empty line */
	if (test_max_tokens > 0 && token_number < test_max_tokens) {
		nums_interest_tokens = nc;
		continue;
	}
	crt->ntokens = token_number;
	for(; i < crt->max_tokens; i++)
		crt->tokens[i] = NULL;
	return crt;
    }
  }
}
