/*
 * universal getopt_ulflags implementation for busybox
 *
 * Copyright (C) 2003-2008  Vladimir Oleynik  <dzo@simtreas.ru>
 *
 */

#include "libbb.h"
#include <getopt.h>
#include <string.h>

/*                  Documentation !

unsigned long
bb_getopt_ulflags (const char *applet_opts, ...)
unsigned long
bb_xgetopt_ulflags (const char *applet_opts, ...)

	The first argument  (in x-function "\0" before separator) is
	compatibility with 3-th argument (optsring) in the GNU getopt(3)
	function.
	The followed arguments is a pointers for save a option's arguments.
	For example:

	flags = bb_getopt_ulflags("rnug");

	If one of the given options is found, a flag value is added to
	the return value (an unsigned long).

	The flag value is determined by the position of the char in
	applet_opts string.  For example, in the above case:

	flags = bb_getopt_ulflags("rnug");

	"r" will add 1    (bit 0)
	"n" will add 2    (bit 1)
	"u  will add 4    (bit 2)
	"g" will add 8    (bit 3)

	and so on.  You can also look at the return value as a bit
	field and each option sets one of bits.

	The syntax applet_opts you can see from GNU getopt(3) manual.
	bb_getopt_ulflags() support GNU getopt(3) extension: first
	char may be '+' or '-' with special case.

	If one of the options requires an argument, then provide
	a pointer to store the argument. For example:

	void add_option(char *o) { printf("called with -e %s\n", o); }
	char *pointer_to_arg_for_b;
	long lc = DEFAULT_VALUE_FOR_C;
	int d;
	llist_t *l = NULL;              // default empty

	flags = bb_xgetopt_ulflags("ab:c:d:e:l:\0:c-d:=e:l::",
					&pointer_to_arg_for_b,
					&lc, &d, add_option, &l);
	if(flags & 1)
		printf("called with -a\n");
	if(flags & 2)
		printf("called with -b %s\n", pointer_to_arg_for_b);
	printf("use -c %ld\n", lc);     // always have value
	if(flags & 4)
		printf("called -d %d\n", d);
	if(flags & 16) {
		while (l) {
			printf("called with -l %s\n", l->data);
			l = l->link;
		}
	}

	The type of the pointer (char*, long*, int*, (*funptr)(*) or llist_t*)
	controlled by the "complementary options groups"
	that is set in the options string in bb_xgetopt_ulflags(applet_opts)
	after "\0" separator (see below for info).

#if ENABLE_FEATURE_LONG_OPTIONS

	Struct option allows you to define long options.  The syntax for
	declaring the array is just like that of getopt's longopts.
	(see getopt(3))

	static const struct option applet_long_options[] = {
		{ "verbose", 0, 0, val },
		{ 0, 0, 0, 0 }
	};
	bb_applet_long_options = applet_long_options;

	The last integer argument (val) can undefined from applet_opts.
	If you use this, then:
	- return bit have next position after short options
	- if has_arg is not "no_argument", use ptr for arg also
	- if val is char then complementary options affects it too

	Note: a good applet will make long options configurable via the
	config process and not a required feature.  The current standard
	is to name the config option CONFIG_FEATURE_<applet>_LONG_OPTIONS
	with global enable CONFIG_FEATURE_LONG_OPTIONS
#endif // ENABLE_FEATURE_LONG_OPTIONS

complementary options groups in string after "\0" separator:

 ":"    The colon (":") is used to separate groups of two or more chars
	and/or groups of chars and special characters (stating some
	conditions to be checked). Special case for first position see below.

 "abc"  If groups of two or more chars are specified, the first char
	is the main option and the other chars are secondary options.
	Their flags will be turned on if the main option is found even
	if they are not specifed on the command line.  For example:

	flags = bb_xgetopt_ulflags("abcd\0abc")

	If getopt() finds "-a" on the command line, then
	bb_xgetopt_ulflags's return value will be as if "-a -b -c" were
	found.

 "ww"   Adjacent double options have a counter associated which indicates
	the number of occurences of the option.
	For example the ps applet needs:
	if w is given once, GNU ps sets the width to 132,
	if w is given more than once, it is "unlimited"

	int w_counter = 0;
	bb_xgetopt_ulflags("w\0ww", &w_counter);

	if(w_counter)
		width = (w_counter == 1) ? 132 : INT_MAX;
	else
		get_terminal_width(...&width...);

	w_counter is a pointer to an integer. It has to be passed to
	bb_xgetopt_ulflags() after all other option argument sinks.
	For example: accept multiple -v to indicate the level of verbosity
	and for each -b optarg, add optarg to my_b. Finally, if b is given,
	turn off c and vice versa:

	llist_t *my_b = NULL;
	int verbose_level = 0;
	f = bb_xgetopt_ulflags("vb:c\0vv:b::b-c:c-b", &my_b, &verbose_level);
	if((f & 2))     // -c after -b unset this -b flag
	  while (my_b) { dosomething_with(my_b->data); my_b = my_b->link; }
	if(my_b)        // but llist stored always if -b found
		free_llist(my_b);
	if (verbose_level) bb_printf("verbose level is %d\n", verbose_level);

Special characters:

 "-"    A dash between two options causes the second of the two
	to be unset (and ignored or triggered) if it is given on
	the command line.

	For example:
	The du applet has the options "-s" and "-d depth".  If
	bb_xgetopt_ulflags finds -s, then -d is unset or if it finds -d
	then -s is unset.  (Note:  busybox implements the GNU
	"--max-depth" option as "-d".)  To obtain this behavior, you
	set complementary as "s-d:d-s".  Only one flag value is
	added to bb_xgetopt_ulflags's return value depending on the
	position of the options on the command line.  If one of the
	two options requires an argument pointer (":" in applet_opts
	as in "d:") optarg is set accordingly.

	int max_print_depth = INT_MAX;
	opt = bb_xgetopt_ulflags("sd:x\0:-d:s-d:d-s:x-x", &max_print_depth);

	if(opt & 1)
		printf("Usaged with -s\n");
	if(opt & 4)
		printf("Detected odd -x usage\n");

 "-"    A dash as the first char in a complementary group means to
	convert the arguments as option. Next char for this case can't set
	[0-9], recomended use ':' or end of line. For example:

	bb_xgetopt_ulflags("wx\0-:w-x:x-w");

	Allows any arguments to be given without a dash (./program w x)
	as well as with a dash (./program -x). Why unset -w see above.

 "-0"   Convert zero argument usage to one argument "-"
	see "cat" applet for example.

 "-N"   A dash as the first char in a complementary group followed
	by a single digit (1-9) means that at least N non-option
	arguments must be present on the command line

 "V-"   A option with dash before colon or end line indicate: call
	bb_show_usage if this option give, for example verbose
	usage option.

 "V--"  A option with double dash before colon or end line indicate:
	reverse flag. Default returned up this bit flag if unused this
	option. If called with this option, returned down flag value.
	Do not use this option char with on/off/trigger a logic together.

 "--"   A double dash between two options, or between an option and a group
	of options, means that they are mutually exclusive.  Unlike
	the "-" case above, an error will be forced if the options
	are used together.

	For example:
	The cut applet must have only one type of list specified, so
	-b, -c and -f are mutally exclusive and should raise an error
	if specified together.  In this case you must set
	complementary as "b--cf:c--bf:f--bc".  If two of the
	mutually exclusive options are found, bb_xgetopt_ulflags's
	will have the Gargc set -1 so that we can check for it:

	if (Gargc < 0)
		bb_show_usage();

 "?"    A "?" as the first char in a complementary group give:
	if previous point set error, don't return and
	call previous example internally. Next char for this case can't
	set to [0-9], recomended use ':' or end of line.

 "?N"   A "?" as the first char in a complementary group with
	number 0-9 as one char is means check maximal arguments possible.

 "=N"   A "=" as the first char in a complementary group with
	number 0-9 as one char is means check exactly this arguments possible.

 "=F"   A "=" as the first char in a complementary group do callback
	function with option argument every found

 "==n"  A double "=" with char means allow legacy syntax of an initial
	numeric option without char option. Convert a -Number argument
	to -nNumber, special for "head" "tail" and "fold" applets.

 ":N"   A colon as the first char in a complementary string with
	options list means than options have a number argument.
	":L-i"  For all option before "-" in this complementary group pointers
	for bb_xgetopt_ulflags() must have a pointer to "long".
	After "-" pointers for bb_xgetopt_ulflags() must have a
	pointer to "int".

	For example
	The "arping" applet can have "-c count" and "-w deadline" arguments.
	This examples equivalent

	First:

	char *scount = NULL;
	char *stimeout = NULL;
	int count = -1;                // default, if not specified
	long timeout = 0;              // default, if not specified
	bb_getopt_ulflags("c:w:", &scount, &stimeout);
	if(scount)
		count = atoi(scount);
	if(stimeout)
		timeout = atol(stimeout);

	Second with use complementary group:

	int count = -1;                // default, if not specified
	long timeout = 0;              // default, if not specified
	bb_xgetopt_ulflags("c:w:\0:w-c", &count, &timeout);

	Do not mix positions. The position set by optstring,
	do not by complementary group.

 "::"   A double colon after a char in complementary means that the
	option can occur multiple times:

	For example:
	The grep applet can have one or more "-e pattern" arguments.
	In this case you should use bb_xgetopt_ulflags() as follows:

	llist_t *patterns = NULL;

	(this pointer must be initializated to NULL if the list is empty
	as required by *llist_add_to_end(llist_t *old_head, char *new_item).)

	bb_xgetopt_ulflags("e:\0e::", &patterns);
	$ grep -e user -e root /etc/passwd
	root:x:0:0:root:/root:/bin/bash
	user:x:500:500::/home/user:/bin/bash

 "--"   A double dash at the beginning of complementary means the
	argv[1] string should always be treated as options, even if it isn't
	prefixed with a "-".  This is to support the special syntax in applets
	such as "ar" and "tar":
	tar xvf foo.tar

 "?"    An "?" between main and group options causes the second of the two
	to be depending required as or if first is given on the command line.
	For example from "id" applet:

	// Don't allow -n -r -rn -ug -rug -nug -rnug
	flags = bb_xgetopt_ulflags("rnug\0r?ug:n?ug:?u--g:g--u");

	This example allowed only:
	$ id; id -u; id -g; id -ru; id -nu; id -rg; id -ng; id -rnu; id -rng

 "X"    A one options in complementary group means
	requires this option always with "or" logic if more one specified,
	checked after switch off from complementary logic.
	For example from "start-stop-daemon" applet:

	// Don't allow -KS -SK, but -S or -K is required
	flags = bb_xgetopt_ulflags("KS...\0K:S:?K--S:S--K");


 "x--x" give error if double or more used -x option, for example "cut" applet

 "??"   warning of ignoring arguments if found

 "?-"   for future use

 Don't forget ':' store. Documented usage without ':' separate is
 "-N1?N2" or "?a--b" only (example see in "cmp" applet).

 WARNING! For "ar" and "tar" applets (with a double dash at the beginning
	  of complementary group) do not use "argc", use now as example:
	  while(*Gargv != NULL) {
		foo(*Gargv);
		Gargv++;
	  }
*/

typedef void (*function_arg_t) (char *);

typedef struct {
	int opt;
	int flg;
#define FLG_IS_LLIST            1
#define FLG_IS_LONGPTR          2
#define FLG_IS_INTPTR           4
#define FLG_IS_FUNPTR           8
#define FLG_IS_SHOW_USAGE       16
	unsigned long switch_on;
	unsigned long switch_off;
	unsigned long incongruously;
	unsigned long requires;
	union {
		void *optarg;
		char **optarg_ptr;
		llist_t **optarg_list;
		long *lnum_ptr;
		int *inum_ptr;
		function_arg_t fa;
	} o;
	int *counter;
} t_complementary;

#if ENABLE_FEATURE_LONG_OPTIONS
/* You can set bb_applet_long_options for parse called long options */
static const struct option bb_default_long_options[] = {
/*      { "help", 0, NULL, '?' }, */
	{ 0, 0, 0, 0 }
};
#endif


#define MAX_OPT_CHRS 32

unsigned long bss_flags;

static unsigned long
bb_xvgetopt (const char *applet_opts, const char *is_xfunc, va_list p)
{
	int argc = Gargc;
	char **argv = Gargv;
	unsigned long flags = 0;
	t_complementary complementary[MAX_OPT_CHRS+1];
#define Requires complementary[MAX_OPT_CHRS].requires
	int c;
	const unsigned char *s, *opt_complementary;
	t_complementary *on_off;
	USE_FEATURE_LONG_OPTIONS(const struct option *l_o;)
	unsigned long trigger;
	char **pargv = NULL;
	int min_arg = 0;
	int max_arg = -1;

#define SHOW_USAGE_IF_ERROR     1
#define ALL_ARGV_IS_OPTS        2
#define FIRST_ARGV_IS_OPT       4
	int spec_flgs = 0;

	c = 0;
	on_off = complementary;
	memset(on_off, 0, sizeof(complementary));

	/* skip GNU extension */
	s = (const unsigned char *)applet_opts;
	if(*s == '+' || *s == '-')
		s++;
	for (; *s; s++) {
		on_off->opt = *s;
		on_off->switch_on = (1 << c);
		if (s[1] == ':') {
			on_off->o.optarg = va_arg (p, void *);
			do
				s++;
			while (s[1] == ':');
		}
		if(c < (MAX_OPT_CHRS-1)) {
			/* else erroneous usage in source code */
			c++;
			on_off++;
		}
	}
	opt_complementary = is_xfunc ? s + 1 : NULL;

#if ENABLE_FEATURE_LONG_OPTIONS
	if(bb_applet_long_options == NULL)
		bb_applet_long_options = bb_default_long_options;
	for(l_o = bb_applet_long_options; l_o->name; l_o++) {
		if(l_o->flag)
			continue;
		for(on_off = complementary; on_off->opt != 0; on_off++)
			if(on_off->opt == l_o->val)
				break;
		if(on_off->opt == 0 && c < MAX_OPT_CHRS) {
			on_off->opt = l_o->val;
			on_off->switch_on = (1 << c);
			if(l_o->has_arg != no_argument)
				on_off->o.optarg = va_arg (p, void *);
			c++;
		}
	}
#endif
	s = opt_complementary;
	if(s != NULL) {
	    for (; s && *s; s++) {
		t_complementary *pair;
		unsigned long *pair_switch;
		int flg_ptr = 0;
#define LONG_PTR 1
#define INT_PTR  2
#define FUN_PTR  (-1)

		c = s[1];
		if(*s == '=' && ((unsigned int)(c - '0')) <= 9) {
			min_arg = max_arg = c - '0';
			s++;
			continue;
		}
		if(*s == '=') {
#if ENABLE_HEAD || ENABLE_TAIL || ENABLE_FOLD
			if(c == '=') {
				/* -[0-9]+ -> -n[0-9]+ */
				for (pargv = argv; *pargv; pargv++) {
					char const *a = *pargv;
					if (*a++ == '-') {
						if (*a == '-' && !a[1])
							break;
						if (((unsigned int)(*a - '0')) <= 9) {
							*pargv = xasprintf("-%c%s", s[2], a);
						}
					}
				}
				pargv = NULL;
				s += 3;
			} else
#endif
			{
				s++;
				flg_ptr = FUN_PTR;
			}
		}
		if(*s == '?') {
			if(((unsigned int)(c - '0')) <= 9) {
				max_arg = c - '0';
				s++;
				continue;
			} else if(c == '?') {
				max_arg = -2;
				s++;
				continue;
			} else if(c != '-') {
				spec_flgs |= SHOW_USAGE_IF_ERROR;
				continue;
			}
			/* "?-" complementary */
		}
		if(*s == '-') {
			if(((unsigned int)(c - '0')) > 9) {
				if(c == '-') {
					spec_flgs |= FIRST_ARGV_IS_OPT;
					s++;
				} else
					spec_flgs |= ALL_ARGV_IS_OPTS;
			} else {
				min_arg = c - '0';
				if(min_arg == 0)
					min_arg--;
				s++;
			}
			continue;
		}
		if (*s == ':') {
			if(s == opt_complementary) {
				s++;
				flg_ptr = LONG_PTR;
			} else
				continue;
		}
	l_long:
		if(*s == '\0')
			break;
		for (on_off = complementary; on_off->opt; on_off++)
			if (on_off->opt == *s)
				break;
		if(flg_ptr == FUN_PTR) {
			if(on_off->opt != 0) {
				on_off->flg |= FLG_IS_FUNPTR;
				s++;
				goto l_long;
			}
			continue;
		}
		if(flg_ptr) {
			if(*s == '-') {
				if(sizeof(int) != sizeof(long))
					flg_ptr = INT_PTR;
				s++;
				goto l_long;
			}
			if(on_off->opt != 0) {
				if(sizeof(int) != sizeof(long))
				    on_off->flg |= FLG_IS_LONGPTR * flg_ptr;
				else
				    on_off->flg |= FLG_IS_LONGPTR;
				s++;
				goto l_long;
			}
			continue;
		}
		if(c == ':' && s[2] == ':') {
			on_off->flg |= FLG_IS_LLIST;
			continue;
		}
		if(c == ':' || c == '\0') {
			Requires |= on_off->switch_on;
			continue;
		}
		if(c == '-' && (s[2] == ':' || s[2] == '\0')) {
			/*  V- */
			on_off->flg |= FLG_IS_SHOW_USAGE;
			s++;
			continue;
		}
		if(c == '-' && s[2] == '-' && (s[3] == ':' || s[3] == '\0')) {
			flags |= on_off->switch_on;     /* reverse flag */
			on_off->switch_off |= on_off->switch_on;
			on_off->switch_on = 0;
			s += 2;
			continue;
		}
		if(c == *s) {
			on_off->counter = va_arg (p, int *);
			s++;
		}
		pair = on_off;
		pair_switch = &(pair->switch_on);
		for(s++; *s && *s != ':'; s++) {
			if(*s == '?') {
				pair_switch = &(pair->requires);
			} else if (*s == '-') {
				if(pair_switch == &(pair->switch_off))
					pair_switch = &(pair->incongruously);
				else
					pair_switch = &(pair->switch_off);
			} else {
			    for (on_off = complementary; on_off->opt; on_off++)
				if (on_off->opt == *s) {
				    *pair_switch |= on_off->switch_on;
				    break;
				}
			}
		}
		s--;
	    }
	}
#if defined(CONFIG_AR) || defined(CONFIG_TAR)
	if((spec_flgs & FIRST_ARGV_IS_OPT)) {
		if(argv[1] && argv[1][0] != '-' && argv[1][0] != '\0') {
			pargv = argv + 1;
			c = **pargv;
			optarg = argv[2];        // may be option
		}
	}
#endif

#if ENABLE_FEATURE_LONG_OPTIONS
	while (pargv != NULL || (c = getopt_long (argc, argv, applet_opts,
				 bb_applet_long_options, NULL)) >= 0)
#else
	while (pargv != NULL || (c = getopt(argc, argv, applet_opts)) >= 0)
#endif
									{
		for (on_off = complementary; on_off->opt != c; on_off++) {
			/* c==0 if long opt have non NULL flag */
			if(on_off->opt == 0 && c != 0)
				bb_show_usage ();
		}
		if((on_off->flg & FLG_IS_SHOW_USAGE) != 0)
			bb_show_usage ();
		if(flags & on_off->incongruously) {
			if((spec_flgs & SHOW_USAGE_IF_ERROR))
				bb_show_usage ();
			opt_complementary = (const unsigned char *)applet_opts;
		}
		trigger = on_off->switch_on & on_off->switch_off;
		flags &= ~(on_off->switch_off ^ trigger);
		flags |= on_off->switch_on ^ trigger;
		flags ^= trigger;

		if(on_off->counter)
			(*(on_off->counter))++;
		if((on_off->flg & FLG_IS_LLIST) != 0) {
			*(on_off->o.optarg_list) =
			  llist_add_to_end(*(on_off->o.optarg_list), optarg);
		} else if((on_off->flg & (FLG_IS_LONGPTR|FLG_IS_INTPTR)) != 0) {
			char eo[2];

			eo[0] = c;
			eo[1] = '\0';
			if(sizeof(int) == sizeof(long) || (on_off->flg & FLG_IS_LONGPTR))
			  *(on_off->o.lnum_ptr) = bb_xparse_l_number(10 | BB_PN_LOWER_MIN | BB_PN_ERR_STRV,
									optarg, "error number '%s' for -%s option", optarg, eo);
			else
			  *(on_off->o.inum_ptr) = bb_xparse_i_number(10 | BB_PN_LOWER_MIN | BB_PN_ERR_STRV,
									optarg, "error number '%s' for -%s option", optarg, eo);
		} else if((on_off->flg & FLG_IS_FUNPTR) != 0) {
			(*(on_off->o.fa))(optarg);
		} else if (on_off->o.optarg_ptr) {
			*(on_off->o.optarg_ptr) = optarg;
#if defined(CONFIG_AR) || defined(CONFIG_TAR)
			if(pargv != NULL) {
				if(optarg == NULL)
					bb_show_usage ();
				optarg = NULL;
			}
#endif
		}

#if defined(CONFIG_PS)
		if((spec_flgs & ALL_ARGV_IS_OPTS) != 0 && pargv != NULL) {
loop_arg_is_opt:
			c = **pargv;
			if(c == '\0') {
				pargv++;
				if(*pargv == NULL)
					break;
				c = **pargv;
			} else {
				(*pargv)++;
			}
		}
#endif

#if defined(CONFIG_AR) || defined(CONFIG_TAR)
		if((spec_flgs & FIRST_ARGV_IS_OPT) != 0 && pargv != NULL) {
			(*pargv)++;
			c = **pargv;
			if(c == '\0') {
				c = 1 + (optarg == NULL);
				argc -= c;
				pargv = argv + 1;
				do {
					*pargv = argv[++c];
					pargv++;
				} while(pargv[-1] != NULL);
				pargv = NULL;
			}
		}
#endif

	}


#ifdef CONFIG_PS
	if((spec_flgs & ALL_ARGV_IS_OPTS) != 0 && pargv == NULL) {
		pargv = argv + optind;
		if(*pargv != NULL)
			goto loop_arg_is_opt;
	}
#endif

	/* check depending requires for given options */
	for (on_off = complementary; on_off->opt; on_off++) {
		if(on_off->requires && (flags & on_off->switch_on) &&
					(flags & on_off->requires) == 0)
			bb_show_usage ();
	}
	if(Requires && (flags & Requires) == 0) {
		/* do not change to (flags & Requires) != Requires */
		bb_show_usage ();
	}
	argc -= optind;
	if(argc < min_arg || (max_arg >= 0 && argc > max_arg))
		bb_show_usage ();
	if(max_arg == -2 && argc > 0)
		bb_error_msg("ignoring all arguments");
	argv += optind;
	if(min_arg < 0 && *argv == NULL) {
		argv--;
		*argv = "-";
		argc++;
	}
	Gargc = opt_complementary == (const unsigned char *)applet_opts ? -1 : argc;
	Gargv = argv;

	bss_flags = flags;
	return flags;
}

unsigned long
bb_getopt_ulflags (const char *applet_opts, ...)
{
	va_list p;
	unsigned long r;

	va_start (p, applet_opts);
	r = bb_xvgetopt(applet_opts, NULL, p);
	va_end (p);
	return r;
}

unsigned long
bb_xgetopt_ulflags (const char *applet_opts, ...)
{
	va_list p;
	unsigned long r;

	va_start (p, applet_opts);
	r = bb_xvgetopt(applet_opts, applet_opts, p);
	va_end (p);
	return r;
}

#if ENABLE_RPM || ENABLE_OD || ENABLE_HEXDUMP
FAST_FUNC char *xgetopt(const char *applet_opts)
{
	char *p;
	int c;

	c = getopt(Gargc, Gargv, applet_opts);
	if(c == -1) {
		Gargv += optind;
		Gargc -= optind;
		return NULL;
	}
	p = strchr(applet_opts, c);
	if(p == NULL)
		bb_show_usage ();
	return p;
}
#endif
