/*
 * stolen from net-tools-1.59 and stripped down for busybox by
 *                      Erik Andersen <andersen@codepoet.org>
 *
 * Heavily modified by Manuel Novoa III       Mar 12, 2001-2004
 *
 */

#include "libbb.h"
#include "bb_inet.h"
#include <string.h>
#include <unistd.h>

#ifdef DEBUG
# include <resolv.h>
#endif

FAST_FUNC int INET_resolve(const char *name, struct sockaddr_in *s_in, int hostfirst)
{
	struct netent *np;
	int type_error = (hostfirst & 2) != 0 ? BB_GHBN_ERR_DIE : BB_GHBN_SILENT;

	hostfirst &= ~2;
	/* Grmpf. -FvK */
	s_in->sin_family = AF_INET;
	s_in->sin_port = 0;

	/* Default is special, meaning 0.0.0.0. */
	if (!strcmp(name, bb_default)) {
		s_in->sin_addr.s_addr = INADDR_ANY;
		return 1;
	}
	/* Look to see if it's a dotted quad. */
	if (inet_aton(name, &s_in->sin_addr))
		return 0;
	/* If we expect this to be a hostname, try hostname database first */
	if (hostfirst && bb_gethostbyname(BB_GHBN_IN_ADDR | type_error, name, &s_in->sin_addr) != NULL)
		return 0;
	/* Try the NETWORKS database to see if this is a known network. */
	if ((np = getnetbyname(name)) != (struct netent *) NULL) {
		s_in->sin_addr.s_addr = htonl(np->n_net);
		return 1;
	}
	if (hostfirst) {
		/* Don't try again */
		return -1;
	}
#ifdef DEBUG
	res_init();
	_res.options |= RES_DEBUG;
#endif

	if(bb_gethostbyname(BB_GHBN_IN_ADDR | type_error, name, &s_in->sin_addr) == NULL)
		return -1;
	return 0;
}

FAST_FUNC char *bb_xgethostbyaddr(const void *a, char hbuf[NI_MAXHOST], int flags)
{
	int status;
	struct sockaddr_in saddr;
#if ENABLE_FEATURE_IPV6
	struct sockaddr_in6 saddr6;
#endif
	struct sockaddr *sa;
	socklen_t len;
#if ENABLE_FEATURE_IPV6
	unsigned char buf[sizeof(struct in6_addr)];
#else
	unsigned char buf[sizeof(struct in_addr)];
#endif
	if((flags & BB_XGETHOSTBYADDR_FLG_IN_ADDR)) {
		len = sizeof(struct sockaddr_in);
		saddr.sin_family = AF_INET;
		memcpy(&saddr.sin_addr, a, sizeof(struct in_addr));
		sa = (struct sockaddr *)&saddr;
	}
#if ENABLE_FEATURE_IPV6
	else if((flags & BB_XGETHOSTBYADDR_FLG_IN6_ADDR)) {
		len = sizeof(struct sockaddr_in6);
		saddr6.sin6_family = AF_INET6;
		memcpy(&saddr6.sin6_addr, a, sizeof(struct in6_addr));
		sa = (struct sockaddr *)&saddr6;
	}
#endif
	else if (inet_pton(AF_INET, a, buf) > 0) {
		len = sizeof(struct sockaddr_in);
		saddr.sin_family = AF_INET;
		memcpy(&saddr.sin_addr, buf, sizeof(struct in_addr));
		sa = (struct sockaddr *)&saddr;
	}
#if ENABLE_FEATURE_IPV6
	else if(inet_pton(AF_INET6, a, buf) > 0) {
		len = sizeof(struct sockaddr_in6);
		saddr6.sin6_family = AF_INET6;
		memcpy(&saddr6.sin6_addr, buf, sizeof(struct in6_addr));
		sa = (struct sockaddr *)&saddr6;
	}
#endif
	else {
		hbuf[0] = '\0';
		if((flags & BB_XGETHOSTBYADDR_FLG_SHOW_ERROR))
			bb_error_msg("%s: is not a Internet address", (char *)a);
		if((flags & BB_XGETHOSTBYADDR_FLG_ERROR_DIE))
			exit(bb_default_error_retval);
		return NULL;
	}

	if ((status = getnameinfo(sa, len, hbuf, NI_MAXHOST, NULL, 0, NI_NAMEREQD)) != 0) {
		if((flags & BB_XGETHOSTBYADDR_FLG_SHOW_ERROR))
			bb_error_msg("%s: %s", (char *)a, gai_strerror(status));
		if((flags & BB_XGETHOSTBYADDR_FLG_ERROR_DIE))
			exit(bb_default_error_retval);
		if((flags & BB_XGETHOSTBYADDR_FLG_NTOP))
			return (char *)inet_ntop(sa->sa_family, &((struct sockaddr_in *)sa)->sin_addr, hbuf, NI_MAXHOST);
		return NULL;
	}
	return hbuf;
}

/* cache */
struct addr {
	struct sockaddr_in addr;
	char *name;
	int host;
	struct addr *next;
};

static struct addr *INET_nn;     /* addr-to-name cache */

/* numeric: & 0x8000: default instead of *,
 *          & 0x4000: host instead of net,
 *          & 0x0fff: don't resolve
 */
FAST_FUNC int INET_rresolve(char *name, size_t len, struct sockaddr_in *s_in,
				  int numeric, unsigned int netmask)
{
	char *ent;
	struct netent *np;
	struct addr *pn;
	unsigned long ad, host_ad;
	int host = 0;

	/* Grmpf. -FvK */
	if (s_in->sin_family != AF_INET) {
#ifdef DEBUG
		bb_error_msg("rresolve: unsupport address family %d !",
				  s_in->sin_family);
#endif
		errno = EAFNOSUPPORT;
		return -1;
	}
	ad = (unsigned long) s_in->sin_addr.s_addr;
#ifdef DEBUG
	bb_error_msg("rresolve: %08lx, mask %08x, num %08x", ad, netmask, numeric);
#endif
	if (ad == INADDR_ANY) {
		if ((numeric & 0x0FFF) == 0) {
			if (numeric & 0x8000)
				safe_strncpy(name, bb_default, len);
			else
				safe_strncpy(name, "*", len);
			return 0;
		}
	}
	if (numeric & 0x0FFF) {
		safe_strncpy(name, inet_ntoa(s_in->sin_addr), len);
		return 0;
	}

	if ((ad & (~netmask)) != 0 || netmask == 0xffffffff || (numeric & 0x4000))
		host = 1;
#if 0
	INET_nn = NULL;
#endif
	pn = INET_nn;
	while (pn != NULL) {
		if (pn->addr.sin_addr.s_addr == ad && pn->host == host) {
			safe_strncpy(name, pn->name, len);
#ifdef DEBUG
			bb_error_msg("rresolve: found %s %08lx in cache",
					  (host ? "host" : "net"), ad);
#endif
			return 0;
		}
		pn = pn->next;
	}

	host_ad = ntohl(ad);
	np = NULL;
	ent = NULL;
	if (host) {
		char hbuf[NI_MAXHOST];
#ifdef DEBUG
		bb_error_msg("gethostbyaddr (%08lx)", ad);
#endif
		ent = bb_xgethostbyaddr(&ad, hbuf, BB_XGETHOSTBYADDR_FLG_IN_ADDR);
		if (ent != NULL)
			safe_strncpy(name, hbuf, len);
	} else {
#ifdef DEBUG
		bb_error_msg("getnetbyaddr (%08lx)", host_ad);
#endif
		np = getnetbyaddr(host_ad, AF_INET);
		if (np != NULL)
			safe_strncpy(name, np->n_name, len);
	}
	if (ent == NULL && np == NULL)
		safe_strncpy(name, inet_ntoa(s_in->sin_addr), len);
	pn = (struct addr *) xmalloc(sizeof(struct addr));
	pn->addr = *s_in;
	pn->next = INET_nn;
	pn->host = host;
	pn->name = xstrdup(name);
	INET_nn = pn;

	return 0;
}

#ifdef CONFIG_FEATURE_IPV6

FAST_FUNC int INET6_resolve(const char *name, struct sockaddr_in6 *sin6)
{
	if(bb_gethostbyname(BB_GHBN_SOCKADDR_IN6, name, sin6) == NULL)
		return -1;
	return 0;
}

#ifndef IN6_IS_ADDR_UNSPECIFIED
# define IN6_IS_ADDR_UNSPECIFIED(a) \
	(((uint32_t *) (a))[0] == 0 && ((uint32_t *) (a))[1] == 0 && \
	 ((uint32_t *) (a))[2] == 0 && ((uint32_t *) (a))[3] == 0)
#endif


FAST_FUNC int
INET6_rresolve(char *name, size_t len, struct sockaddr_in6 *sin6, int numeric)
{
	char hbuf[NI_MAXHOST];

	/* Grmpf. -FvK */
	if (sin6->sin6_family != AF_INET6) {
#ifdef DEBUG
		bb_error_msg(_("rresolve: unsupport address family %d !\n"),
				  sin6->sin6_family);
#endif
		errno = EAFNOSUPPORT;
		return -1;
	}
	if (numeric & 0x7FFF) {
		inet_ntop(AF_INET6, &sin6->sin6_addr, name, len);
		return 0;
	}
	if (IN6_IS_ADDR_UNSPECIFIED(&sin6->sin6_addr)) {
		if (numeric & 0x8000) {
			strcpy(name, bb_default);
		} else {
			strcpy(name, "*");
		}
		return 0;
	}

	if(bb_xgethostbyaddr(&sin6->sin6_addr, hbuf, BB_XGETHOSTBYADDR_FLG_SHOW_ERROR | BB_XGETHOSTBYADDR_FLG_IN6_ADDR) == NULL)
		return -1;
	safe_strncpy(name, hbuf, len);
	return 0;
}

#endif                                                  /* CONFIG_FEATURE_IPV6 */

int bb_xsocket(int flags, ...)
{
	va_list p;
	int af, type, proto, s;

	if((flags & (BB_XSOCKET_FLG_PROTO_SPEC|BB_XSOCKET_FLG_AF_SPEC)) != 0)
		va_start(p, flags);

	switch(flags & (BB_XSOCKET_FLG_PF_PACKET|BB_XSOCKET_FLG_AF_INET6|BB_XSOCKET_FLG_AF_UNIX|BB_XSOCKET_FLG_AF_SPEC)) {
		case BB_XSOCKET_FLG_AF_UNIX:
			af = AF_UNIX;
			break;
		case BB_XSOCKET_FLG_PF_PACKET:
			af = PF_PACKET;
			break;
		case BB_XSOCKET_FLG_AF_SPEC:
			af = va_arg(p, int);
			break;
#ifdef CONFIG_FEATURE_IPV6
		case BB_XSOCKET_FLG_AF_INET6:
			af = AF_INET6;
			break;
#endif
		default:
			af = AF_INET;
			break;
	}
	switch(flags & (BB_XSOCKET_FLG_SOCK_DGRAM|BB_XSOCKET_FLG_SOCK_RAW)) {
		case BB_XSOCKET_FLG_SOCK_DGRAM:
			type = SOCK_DGRAM;
			break;
		case BB_XSOCKET_FLG_SOCK_RAW:
			type = SOCK_RAW;
			break;
		default:
			type = SOCK_STREAM;
			break;
	}
	if((flags & BB_XSOCKET_FLG_PROTO_SPEC) != 0) {
		proto = va_arg(p, int);
	} else {
		proto = 0;
	}
	if ((s = socket(af, type, proto)) < 0) {
		if((flags & BB_XSOCKET_FLG_NO_WARN) == 0)
			bb_perror_msg("can not create socket");
		if((flags & BB_XSOCKET_FLG_ERR_DIE))
			exit(bb_default_error_retval);
	}
	if((flags & (BB_XSOCKET_FLG_PROTO_SPEC|BB_XSOCKET_FLG_AF_SPEC)) != 0)
		va_end(p);
	return s;
}

FAST_FUNC int setsockopt_1i(int sockfd, int level, int optname, int int_opt)
{
	return setsockopt(sockfd, level, optname, (char *)&int_opt, sizeof (int));
}

void FAST_FUNC bb_xbind(int sockfd, struct sockaddr *addr, socklen_t len)
{
	if(bind(sockfd, addr, len) < 0)
		bb_perror_msg_and_die("can't bind socket");
}

#if BYTE_ORDER == LITTLE_ENDIAN
# define ODDBYTE(v)     (v)
#elif BYTE_ORDER == BIG_ENDIAN
# define ODDBYTE(v)     ((u_short)(v) << 8)
#else
# define ODDBYTE(v)     htons((u_short)(v) << 8)
#endif

unsigned short FAST_FUNC
in_cksum(const void *addr, int len, unsigned short csum)
{
	int nleft = len;
	const unsigned short *w = addr;
	unsigned short answer;
	int sum = csum;

	/*
	 *  Our algorithm is simple, using a 32 bit accumulator (sum),
	 *  we add sequential 16 bit words to it, and at the end, fold
	 *  back all the carry bits from the top 16 bits into the lower
	 *  16 bits.
	 */
	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1)
		sum += ODDBYTE(*(u_char *)w); /* le16toh() may be unavailable on old systems */

	/*
	 * add back carry outs from top 16 bits to low 16 bits
	 */
	sum = (sum >> 16) + (sum & 0xffff);     /* add hi 16 to low 16 */
	sum += (sum >> 16);                     /* add carry */
	answer = ~sum;                          /* truncate to 16 bits */
	return (answer);
}
