/*
 * universal parse number implementation for busybox
 *
 * Copyright (C) 2018  Vladimir Oleynik  <dzo@simtreas.ru>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

#include "libbb.h"
#include <string.h>
#include <ctype.h>
#include <inttypes.h>

/* The documentaion

extern int bb_xparse_i_number(int flags, const char *arg, ...);
return int converted "arg" string

extern unsigned int bb_xparse_ui_number(int flags, const char *arg, ...);
return unsigned int converted "arg" string

extern long bb_xparse_l_number(int flags, const char *arg, ...);
return long converted "arg" string

extern unsigned long bb_xparse_ul_number(int flags, const char *arg, ...);
return unsigned long converted "arg" string

extern intmax_t bb_xparse_im_number(int flags, const char *arg, ...);
return intmax_t converted "arg" string

extern uintmax_t bb_xparse_uim_number(int flags, const char *arg, ...);
return uintmax_t converted "arg" string


flags (can split by the OR operation):

0-31               - the base for converting (see man strtol);

BB_PN_LOWER_MIN    - the lower value for test bounds is:
		     * INT_MIN/LONG_MIN/INTMAX_MIN for i/l/im functions or
		     * 1 for ui/ul/uim functions
		     Without this flag default is 0;

BB_PN_SET_LOWER    - next arg is specified the lower value for test bounds.
		     WARNING: do not forgot cast to ([unsigned] long)/([u]intmax)
		     for l/ul/im/uim functions.
		     Default see above for BB_PN_LOWER_MIN;

BB_PN_SET_LOWER1   - the lower value for test bounds is 1;

BB_PN_SET_UPPER    - next arg is specified the upper value for test bounds.
		     WARNING: do not forgot cast to ([unsigned] long)/([u]intmax)
		     for l/ul/im/uim functions.
		     Default is (U)INT_MAX or (U)LONG_MAX or (U)INTMAX_MAX;

BB_PN_END_PTR      - next arg is endptr (see man strtol)
		     and also do not test tailing for empty;

BB_PN_SUFFIX       - Next arg is pointer to const struct suffix_mult.
		     Parse none/one suffix only;

BB_PN_BKM_SUFFIXES - internall default most usage struct suffix_mult. See below;

BB_PN_ALLOW_TAIL_SPACE - skip tailing whitespace before test suffix/zero;

BB_PN_ERR_STR1     - next arg is set the format string for error message. Auto added arg.
		     Default message is ("invalid number `%s'", arg);

BB_PN_ERR_STRV     - Set format string for error message and other optional args;

BB_PN_SILENT       - Do not show error message;

BB_PN_SHOW_USAGE   - call show_usage() if detect error. Its affect if BB_PN_SILENT set too;

BB_PN_RETURN_ERR   - Do not terminate with the exit code `bb_default_error_retval',
		     set errno and return -1 for signed and 0 for unsigned functions.

BB_PN_SILENT_RE      (BB_PN_SILENT | BB_PN_RETURN_ERR) - most usage
*/

static const struct suffix_mult bkm_suffixes[] = {
	{ "c", 1 },
	{ "w", 2 },
	{ "b", 512 },
	{ "kB", 1000 },
	{ "k", 1024 },
	{ "K", 1024 },
	{ "MB", 1000*1000 },
	{ "m", 1024*1024 },
	{ "M", 1024*1024 },
	{ "GB", 1000*1000*1000 },
	{ "G", 1024*1024*1024 },
	{ NULL, 0 }
};

static void error4flags(int flags, const char *arg0, va_list p)
{
	const char *e = "invalid number `%s'";

	if((flags & (BB_PN_ERR_STR1|BB_PN_ERR_STRV)) != 0)
		e = va_arg(p, const char *);
	if((flags & BB_PN_SILENT) == 0) {
		if((flags & (BB_PN_ERR_STRV)) != 0)
			bb_verror_msg(e, p);
		else
			bb_error_msg(e, arg0);
	}
	if((flags & BB_PN_SHOW_USAGE) != 0)
		bb_show_usage();
	if((flags & BB_PN_RETURN_ERR) == 0)
		exit(bb_default_error_retval);
}

static intmax_t
safe_strton(int flags, const char *arg, intmax_t lower, intmax_t upper, va_list p)
{
	intmax_t r;
	char *e;
	char **ea = NULL;
	const struct suffix_mult *suffixes = NULL;

	if (arg == NULL) {
		errno = EINVAL;
		goto err;
	}
	if((flags & BB_PN_SET_LOWER1) != 0)
		lower = 1;
	if((flags & BB_PN_END_PTR) != 0) {
		ea = va_arg(p, char **);
		*ea = (char *)arg;
	}
	if((flags & BB_PN_BKM_SUFFIXES) != 0)
		suffixes = bkm_suffixes;
	else if((flags & BB_PN_SUFFIX) != 0)
		suffixes = va_arg(p, const struct suffix_mult *);

	errno = 0;
	r = strtoimax(arg, &e, flags & (BB_PN_BASE_HIGH_BIT-1));
	/* Do the initial validity check.  Note: The standards do not
	 * guarantee that errno is set if no digits were found.  So we
	 * must test for this explicitly. */
	if (errno)
		goto err;
	if (arg == e) {
		/* error or no digits */
		errno = EINVAL;
		goto err;
	}
	if ((flags & BB_PN_ALLOW_TAIL_SPACE) != 0)
		e = bb_skip_whitespace(e);
	/* Do optional suffix parsing.  Allow 'empty' suffix tables.
	 * Note that we also all nul suffixes with associated multipliers,
	 * to allow for scaling of the arg by some default multiplier. */
	if (suffixes) {
		suffixes = find_string_from_struct(suffixes, sizeof(struct suffix_mult), e);
		if (suffixes->mult) {
			if (upper / suffixes->mult < r) {
				/* Overflow! */
				errno = ERANGE;
				goto err;
			}
			r *= suffixes->mult;
			e += strlen(suffixes->suffix);
		}
	}
	if (ea == NULL) {
		if(*e != '\0') {
			errno = EINVAL;
			goto err;
		}
	} else
		*ea = e;
	if (r < lower || r > upper) {
		errno = ERANGE;
		goto err;
	}
	return r;
err:
	error4flags(flags, arg, p);
	return -1;
}

#ifndef bb_xparse_i_number
extern int
bb_xparse_i_number(int flags, const char *arg, ...)
{
	va_list p;
	int r, lower = 0, upper = INT_MAX;

	va_start(p, arg);
	if((flags & BB_PN_SET_LOWER) != 0) {
		lower = va_arg(p, int);
	} else if((flags & BB_PN_LOWER_MIN) != 0) {
		lower = INT_MIN;
	}
	if((flags & BB_PN_SET_UPPER) != 0)
		upper = va_arg(p, int);
	r = safe_strton(flags, arg, (intmax_t)lower, (intmax_t)upper, p);
	va_end(p);
	return r;
}
#endif

extern long
bb_xparse_l_number(int flags, const char *arg, ...)
{
	va_list p;
	long r, lower = 0, upper = LONG_MAX;

	va_start(p, arg);
	if((flags & BB_PN_SET_LOWER) != 0) {
		lower = va_arg(p, long);
	} else if((flags & BB_PN_LOWER_MIN) != 0) {
		lower = LONG_MIN;
	}
	if((flags & BB_PN_SET_UPPER) != 0)
		upper = va_arg(p, long);
	r = safe_strton(flags, arg, (intmax_t)lower, (intmax_t)upper, p);
	va_end(p);
	return r;
}

#ifndef bb_xparse_im_number
extern intmax_t
bb_xparse_im_number(int flags, const char *arg, ...)
{
	va_list p;
	intmax_t r, lower = 0, upper = INTMAX_MAX;

	va_start(p, arg);
	if((flags & BB_PN_SET_LOWER) != 0) {
		lower = va_arg(p, intmax_t);
	} else if((flags & BB_PN_LOWER_MIN) != 0) {
		lower = INTMAX_MIN;
	}
	if((flags & BB_PN_SET_UPPER) != 0)
		upper = va_arg(p, intmax_t);
	r = safe_strton(flags, arg, lower, upper, p);
	va_end(p);
	return r;
}
#endif

static intmax_t
safe_strtoun(int flags, const char *arg, uintmax_t lower, uintmax_t upper, va_list p)
{
	uintmax_t r;
	char *e;
	char **ea = NULL;
	const struct suffix_mult *suffixes = NULL;

	if (arg == NULL) {
		errno = EINVAL;
		goto err;
	}
	if((flags & (BB_PN_LOWER_MIN|BB_PN_SET_LOWER1)) != 0)
		lower = 1;
	if((flags & BB_PN_END_PTR) != 0) {
		ea = va_arg(p, char **);
		*ea = (char *)arg;
	}
	if((flags & BB_PN_BKM_SUFFIXES) != 0)
		suffixes = bkm_suffixes;
	else if((flags & BB_PN_SUFFIX) != 0)
		suffixes = va_arg(p, const struct suffix_mult *);

	errno = 0;
	r = strtoumax(arg, &e, flags & (BB_PN_BASE_HIGH_BIT-1));
	/* Do the initial validity check.  Note: The standards do not
	 * guarantee that errno is set if no digits were found.  So we
	 * must test for this explicitly. */
	if (errno)
		goto err;
	if (arg == e) {
		/* error or no digits */
		errno = EINVAL;
		goto err;
	}
	if ((flags & BB_PN_ALLOW_TAIL_SPACE) != 0)
		e = bb_skip_whitespace(e);
	/* Do optional suffix parsing.  Allow 'empty' suffix tables.
	 * Note that we also all nul suffixes with associated multipliers,
	 * to allow for scaling of the arg by some default multiplier. */
	if (suffixes) {
		suffixes = find_string_from_struct(suffixes, sizeof(struct suffix_mult), e);
		if (suffixes->mult) {
			if (upper / suffixes->mult < r) {
				/* Overflow! */
				errno = ERANGE;
				goto err;
			}
			r *= suffixes->mult;
			e += strlen(suffixes->suffix);
		}
	}
	if (ea == NULL) {
		if(*e != '\0') {
			errno = EINVAL;
			goto err;
		}
	} else
		*ea = e;
	if (r < lower || r > upper) {
		errno = ERANGE;
		goto err;
	}
	return r;
err:
	error4flags(flags, arg, p);
	return 0;
}

#ifndef bb_xparse_ui_number
extern unsigned int
bb_xparse_ui_number(int flags, const char *arg, ...)
{
	va_list p;
	unsigned int u, lower = 0, upper = UINT_MAX;

	va_start(p, arg);
	if((flags & BB_PN_SET_LOWER) != 0)
		lower = va_arg(p, unsigned int);
	if((flags & BB_PN_SET_UPPER) != 0)
		upper = va_arg(p, unsigned int);
	u = safe_strtoun(flags, arg, (uintmax_t)lower, (uintmax_t)upper, p);
	va_end(p);
	return u;
}
#endif

extern unsigned long
bb_xparse_ul_number(int flags, const char *arg, ...)
{
	va_list p;
	unsigned long u, lower = 0, upper = ULONG_MAX;

	va_start(p, arg);
	if((flags & BB_PN_SET_LOWER) != 0)
		lower = va_arg(p, unsigned long);
	if((flags & BB_PN_SET_UPPER) != 0)
		upper = va_arg(p, unsigned long);
	u = safe_strtoun(flags, arg, (uintmax_t)lower, (uintmax_t)upper, p);
	va_end(p);
	return u;
}

#ifndef bb_xparse_uim_number
extern uintmax_t
bb_xparse_uim_number(int flags, const char *arg, ...)
{
	va_list p;
	uintmax_t u, lower = 0, upper = ULONG_MAX;

	va_start(p, arg);
	if((flags & BB_PN_SET_LOWER) != 0)
		lower = va_arg(p, uintmax_t);
	if((flags & BB_PN_SET_UPPER) != 0)
		upper = va_arg(p, uintmax_t);
	u = safe_strtoun(flags, arg, lower, upper, p);
	va_end(p);
	return u;
}
#endif
