/*
 * Utility routines.
 *
 * Copyright (C) many different people.
 * Copyright (C) 2003 Vladimir N. Oleynik <dzo@simtreas.ru>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

#include "libbb.h"
#include <signal.h>
#include <ctype.h>
#include <string.h>

struct signal_name {
	const char *name;
	int number;
};

static const struct signal_name signames[] = {
	/* POSIX signals */
	{ "EXIT",       0 },            /* 0 */
	{ "HUP",        SIGHUP },       /* 1 */
	{ "INT",        SIGINT },       /* 2 */
	{ "QUIT",       SIGQUIT },      /* 3 */
	{ "ILL",        SIGILL },       /* 4 */
	{ "ABRT",       SIGABRT },      /* 6 */
	{ "FPE",        SIGFPE },       /* 8 */
	{ "KILL",       SIGKILL },      /* 9 */
	{ "SEGV",       SIGSEGV },      /* 11 */
	{ "PIPE",       SIGPIPE },      /* 13 */
	{ "ALRM",       SIGALRM },      /* 14 */
	{ "TERM",       SIGTERM },      /* 15 */
	{ "USR1",       SIGUSR1 },      /* 10 (arm,i386,m68k,ppc), 30 (alpha,sparc*), 16 (mips) */
	{ "USR2",       SIGUSR2 },      /* 12 (arm,i386,m68k,ppc), 31 (alpha,sparc*), 17 (mips) */
	{ "CHLD",       SIGCHLD },      /* 17 (arm,i386,m68k,ppc), 20 (alpha,sparc*), 18 (mips) */
	{ "CONT",       SIGCONT },      /* 18 (arm,i386,m68k,ppc), 19 (alpha,sparc*), 25 (mips) */
	{ "STOP",       SIGSTOP },      /* 19 (arm,i386,m68k,ppc), 17 (alpha,sparc*), 23 (mips) */
	{ "TSTP",       SIGTSTP },      /* 20 (arm,i386,m68k,ppc), 18 (alpha,sparc*), 24 (mips) */
	{ "TTIN",       SIGTTIN },      /* 21 (arm,i386,m68k,ppc,alpha,sparc*), 26 (mips) */
	{ "TTOU",       SIGTTOU },      /* 22 (arm,i386,m68k,ppc,alpha,sparc*), 27 (mips) */
	/* Miscellaneous other signals */
#ifdef SIGTRAP
	{ "TRAP",       SIGTRAP },      /* 5 */
#endif
#ifdef SIGIOT
	{ "IOT",        SIGIOT },       /* 6, same as SIGABRT */
#endif
#ifdef SIGEMT
	{ "EMT",        SIGEMT },       /* 7 (mips,alpha,sparc*) */
#endif
#ifdef SIGBUS
	{ "BUS",        SIGBUS },       /* 7 (arm,i386,m68k,ppc), 10 (mips,alpha,sparc*) */
#endif
#ifdef SIGSYS
	{ "SYS",        SIGSYS },       /* 12 (mips,alpha,sparc*) */
#endif
#ifdef SIGSTKFLT
	{ "STKFLT",     SIGSTKFLT },    /* 16 (arm,i386,m68k,ppc) */
#endif
#ifdef SIGURG
	{ "URG",        SIGURG },       /* 23 (arm,i386,m68k,ppc), 16 (alpha,sparc*), 21 (mips) */
#endif
#ifdef SIGIO
	{ "IO",         SIGIO },        /* 29 (arm,i386,m68k,ppc), 23 (alpha,sparc*), 22 (mips) */
#endif
#ifdef SIGPOLL
	{ "POLL",       SIGPOLL },      /* same as SIGIO */
#endif
#ifdef SIGCLD
	{ "CLD",        SIGCLD },       /* same as SIGCHLD (mips) */
#endif
#ifdef SIGXCPU
	{ "XCPU",       SIGXCPU },      /* 24 (arm,i386,m68k,ppc,alpha,sparc*), 30 (mips) */
#endif
#ifdef SIGXFSZ
	{ "XFSZ",       SIGXFSZ },      /* 25 (arm,i386,m68k,ppc,alpha,sparc*), 31 (mips) */
#endif
#ifdef SIGVTALRM
	{ "VTALRM",     SIGVTALRM },    /* 26 (arm,i386,m68k,ppc,alpha,sparc*), 28 (mips) */
#endif
#ifdef SIGPROF
	{ "PROF",       SIGPROF },      /* 27 (arm,i386,m68k,ppc,alpha,sparc*), 29 (mips) */
#endif
#ifdef SIGPWR
	{ "PWR",        SIGPWR },       /* 30 (arm,i386,m68k,ppc), 29 (alpha,sparc*), 19 (mips) */
#endif
#ifdef SIGINFO
	{ "INFO",       SIGINFO },      /* 29 (alpha) */
#endif
#ifdef SIGLOST
	{ "LOST",       SIGLOST },      /* 29 (arm,i386,m68k,ppc,sparc*) */
#endif
#ifdef SIGWINCH
	{ "WINCH",      SIGWINCH },     /* 28 (arm,i386,m68k,ppc,alpha,sparc*), 20 (mips) */
#endif
#ifdef SIGUNUSED
	{ "UNUSED",     SIGUNUSED },    /* 31 (arm,i386,m68k,ppc) */
#endif
};


#define ELEMENTS(arr) ( sizeof(arr) / sizeof((arr)[0]) )
#define END (signames + ELEMENTS(signames))

static const char prefix[] = "SIG";
#define SIZE_OF_PREFIX 3

FAST_FUNC const char *
u_signal_num2name(int signo, int minsig, char *buf)
{
	const struct signal_name *s;

	for (s = signames + minsig; s < END; s++) {
		if(s->number == signo) {
			if(s == signames || buf == NULL)
				return s->name;
			memcpy(buf, prefix, SIZE_OF_PREFIX);
			strcpy(buf + SIZE_OF_PREFIX, s->name);
			return buf;
		}
	}
	return NULL;
}

FAST_FUNC const char *
u_signal_name2num(const char *str_sig, int *signo, int minsig, char *buf)
{
	const struct signal_name *s = signames + minsig;

	if (isdigit(((unsigned char)*str_sig))) {
		int sn = bb_xparse_i_number(BB_PN_SET_UPPER | BB_PN_SILENT_RE, str_sig, NSIG);
		/* test correct and overflow */
		if(errno == 0) {
			/* test for unnamed */
			for (; s < END; s++) {
				if(s->number == sn) {
					if(signo != NULL)
						*signo = sn;
					return s->name;
				}
			}
		}
	} else {
		if(strncasecmp(str_sig, prefix, 3) == 0)
			str_sig += SIZE_OF_PREFIX;
		for (; s < END; s++) {
			if (strcasecmp(s->name, str_sig) == 0) {
				if(signo != NULL)
					*signo = s->number;
				if(buf) {
					sprintf(buf, "%d", s->number);
					return buf;
				} else {
					return s->name;
				}
			}
		}
	}
	return NULL;
}
