/*
 * Utility routines.
 *
 * Connect to host at port using address resolution from getaddrinfo
 *
 */

#include "libbb.h"
#include "bb_inet.h"
#include <unistd.h>
#include <string.h>
#include <netinet/in.h>

/* Return network byte ordered port number for a service.
 * If "port" is a number use it as the port.
 * If "port" is a name it is looked up in /etc/services, if it isnt found return
 * default_port
 */
FAST_FUNC unsigned short
bb_lookup_tcp_port(const char *service, unsigned short default_port)
{
	unsigned short port_nr = htons(default_port);

	if (service) {
		int port = bb_xparse_i_number(BB_PN_SET_UPPER | BB_PN_SILENT_RE, service, 65535);
		if (port < 0) {
			struct servent *tserv;

			tserv = getservbyname(service, "tcp");
			if (tserv) {
				port_nr = tserv->s_port;
			}
		} else {
			port_nr = htons(port);
		}
	}
	return port_nr;
}

FAST_FUNC int xconnect(const struct sockaddr *s_addr)
{
#if ENABLE_FEATURE_IPV6
#define f s_addr->sa_family
	socklen_t sz;
	int ff = BB_XSOCKET_FLG_ERR_DIE;

	switch(s_addr->sa_family) {
		case AF_INET:
			sz = sizeof(struct sockaddr_in);
			break;
		case AF_INET6:
			sz = sizeof(struct sockaddr_in6);
			ff |= BB_XSOCKET_FLG_AF_INET6;
			break;
		default:
			bb_error_msg_and_die("internal error: bad family for connect");

	}
#else
#define f AF_INET
# define sz sizeof(struct sockaddr_in)
# define ff BB_XSOCKET_FLG_ERR_DIE
#endif
	int s = bb_xsocket(ff);
	if (connect(s, s_addr, sz) < 0) {
		char ibuf[USE_FEATURE_IPV6(INET6_ADDRSTRLEN) SKIP_FEATURE_IPV6(INET_ADDRSTRLEN) + 1];

		inet_ntop(f, &((const struct sockaddr_in*)s_addr)->sin_addr, ibuf, sizeof(ibuf));
		bb_perror_msg_and_die("unable to connect to remote host (%s)", ibuf);
	}
	return s;
}
