/*
 * Utility routines.
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

#include <string.h>
#include <unistd.h>

/* Since gcc always inlines strlen(), this saves a byte or two, but we need
 * the #undef here to avoid endless loop from #define strlen bb_strlen */
#ifdef L_strlen
#define BB_STRLEN_IMPLEMENTATION
#endif

#include "libbb.h"


#ifndef DMALLOC
#ifdef L_xmalloc
FAST_FUNC void *xmalloc(size_t size)
{
	void *ptr = malloc(size);
	if (ptr == NULL && size != 0)
		bb_error_msg_and_die("%s", bb_msg_memory_exhausted);
	return ptr;
}
#endif

#ifdef L_xrealloc
FAST_FUNC void *xrealloc(void *ptr, size_t size)
{
	ptr = realloc(ptr, size);
	if (ptr == NULL && size != 0)
		bb_error_msg_and_die("%s", bb_msg_memory_exhausted);
	return ptr;
}
#endif
#endif /* DMALLOC */

#ifdef L_zmalloc
FAST_FUNC void *zmalloc(size_t size)
{
	void *ptr = xmalloc(size);
	if (ptr != NULL)
		memset(ptr, 0, size);
	return ptr;
}
#endif

#ifdef L_xstrdup
FAST_FUNC char * xstrdup(const char *s) {
	char *t;

	if (s == NULL)
		return NULL;

	t = strdup(s);

	if (t == NULL)
		bb_error_msg_and_die("%s", bb_msg_memory_exhausted);

	return t;
}
#endif

#ifdef L_xstrndup
FAST_FUNC char * xstrndup(const char *s, size_t n) {
	char *t;

	if (s == NULL)
		bb_error_msg_and_die("xstrndup bug");

	t = xmalloc(++n);

	return safe_strncpy(t,s,n);
}
#endif

#ifdef L_xread
FAST_FUNC ssize_t bb_xread(int fd, void *buf, size_t count)
{
	ssize_t size;

	size = read(fd, buf, count);
	if (size == -1)
		bb_perror_msg_and_die("%s", bb_msg_read_error);
	return(size);
}
#endif

#ifdef L_xread_all
FAST_FUNC void bb_xread_all(int fd, void *buf, size_t count)
{
	ssize_t size;

	while (count) {
		if ((size = bb_xread(fd, buf, count)) == 0) /* EOF */
			bb_error_msg_and_die("Short read");
		count -= size;
		buf = ((char *) buf) + size;
	}
	return;
}
#endif

#ifdef L_xread_char
FAST_FUNC unsigned char bb_xread_char(int fd)
{
	char tmp;

	bb_xread_all(fd, &tmp, 1);

	return(tmp);
}
#endif

#ifdef L_xferror
FAST_FUNC void bb_xferror(FILE *fp, const char *fn)
{
	if (ferror(fp))
		bb_error_msg_and_die("%s", fn);
}
#endif

#ifdef L_xferror_stdout
FAST_FUNC void bb_xferror_stdout(void)
{
	bb_xferror(stdout, bb_msg_standard_output);
}
#endif

#ifdef L_xfflush_stdout
FAST_FUNC void bb_xfflush_stdout(void)
{
	if (fflush(stdout))
		bb_perror_msg_and_die("%s", bb_msg_standard_output);
}
#endif

#ifdef L_strlen
size_t bb_strlen(const char *string)
{
	    return(strlen(string));
}
#endif

#ifdef L_xchdir
void FAST_FUNC xchdir(const char *path)
{
	if (chdir(path))
		bb_perror_msg_and_die("can't change directory to '%s'", path);
}
#endif

#ifdef L_strchr_set0
char FAST_FUNC *strchr_set0(char *s, int c)
{
	char *e = strchr(s, c);
	if(e != NULL)
		*e = '\0';
	return e;
}
#endif

#ifdef L_xioctl
#include <sys/ioctl.h>
int FAST_FUNC xioctl(int d, int request, void *a, const char *msg, ...)
{
	int r = ioctl(d, request, a);
	if(r < 0) {
		int flg_exit = 0;
		va_list p;

		va_start(p, msg);
		if(*msg == '\1') {
			flg_exit = 1;
			msg++;
		}
		bb_vperror_msg(msg, p);
		va_end(p);
		if(flg_exit)
			exit(bb_default_error_retval);
	}
	return r;
}
#endif
