/*
 * Copyright (C) 2016 Vladimir Oleynik <dzo@simtreas.ru>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

#include "libbb.h"
#include "bb_inet.h"
#include <string.h>

void *bb_gethostbyname(int flags, const char *name, ...)
{
	va_list p;

	int status;
	struct addrinfo hints, *res;
	int *naddr;
	unsigned size = sizeof(struct sockaddr);
	int one = 0;
	unsigned offset = 0;
#if ENABLE_FEATURE_IPV6
	int family = AF_UNSPEC;
#else
#define family AF_INET
#endif
	const char *add_error_msg;
	char **canon_name;
	struct addrinfo *ap;
	void *ret = NULL;

	va_start(p, name);
	if((flags & BB_GHBN_IN_ADDR) != 0) {
		ret = va_arg(p, struct in_addr *);
		size = sizeof(struct in_addr);
		one = 1;
		offset = (char *)(&((const struct sockaddr_in *)ret)->sin_addr) - (char *)ret;
		USE_FEATURE_IPV6(family = AF_INET);
	} else if((flags & BB_GHBN_SOCKADDR_IN) != 0) {
		ret = va_arg(p, struct sockaddr_in *);
		size = sizeof(struct sockaddr_in);
		one = 1;
		USE_FEATURE_IPV6(family = AF_INET);
	}
#if ENABLE_FEATURE_IPV6
	else if((flags & BB_GHBN_IN6_ADDR) != 0) {
		ret = va_arg(p, struct in6_addr *);
		size = sizeof(struct in6_addr);
		one = 1;
		offset = (char *)(&((const struct sockaddr_in6 *)ret)->sin6_addr) - (char *)ret;
		family = AF_INET6;
	} else if((flags & BB_GHBN_SOCKADDR_IN6) != 0) {
		ret = va_arg(p, struct sockaddr_in6 *);
		size = sizeof(struct sockaddr_in6);
		one = 1;
		family = AF_INET6;
	}
#endif
	else if((flags & BB_GHBN_SOCKADDR) != 0) {
		ret = va_arg(p, struct sockaddr *);
		one = 1;
	}

	naddr = (flags & (BB_GHBN_NADDR|BB_GHBN_NADDR6|BB_GHBN_NSOCKADDR)) != 0 ? va_arg(p, int *) : NULL;
	if(naddr != NULL) {
		if((flags & BB_GHBN_NADDR) != 0) {
			size = sizeof(struct in_addr);
			offset = (char *)(&((const struct sockaddr_in *)ret)->sin_addr) - (char *)ret;
			USE_FEATURE_IPV6(family = AF_INET);
		}
#if ENABLE_FEATURE_IPV6
		else if((flags & BB_GHBN_NADDR6) != 0) {
			size = sizeof(struct in6_addr);
			offset = (char *)(&((const struct sockaddr_in6 *)ret)->sin6_addr) - (char *)ret;
			family = AF_INET6;
		}
#endif
	}

	add_error_msg = (flags & BB_GHBN_ADD_ERR_MSG) != 0 ? va_arg(p, const char *) : "";
	canon_name = (flags & BB_GHBN_CANON_NAME) != 0 ? va_arg(p, char **) : NULL;
	va_end(p);

	memset(&hints, 0, sizeof(hints));

#if ENABLE_FEATURE_IPV6
	hints.ai_flags = canon_name != NULL ? AI_V4MAPPED | AI_ADDRCONFIG | AI_CANONNAME: AI_V4MAPPED | AI_ADDRCONFIG;
#else
	hints.ai_flags = canon_name != NULL ? AI_ADDRCONFIG | AI_CANONNAME: AI_ADDRCONFIG;
#endif
//#ifdef AI_IDN
//        hints.ai_flags |= AI_IDN;
//#endif
	hints.ai_family = family;

	/* Needed. Or else we will get each address thrice (or more)
	 * for each possible socket type (tcp,udp,raw...): */
	hints.ai_socktype = SOCK_STREAM;

	if ((status = getaddrinfo(name, NULL, &hints, &res)) != 0) {
		if((flags & BB_GHBN_SILENT) == 0) {
			bb_error_msg("%s: %s%s%s", name, add_error_msg,
				(add_error_msg[0] == '\0' ? "" : ": "), gai_strerror(status));
		}
		if((flags & BB_GHBN_ERR_DIE))
			exit(bb_default_error_retval);
		if(naddr)
			*naddr = 0;
		return NULL;
	}

	if(!one) {
		for (ap = res; ap != NULL; ap = ap->ai_next)
			one++;
		ret = xmalloc(one * size);
	}
	if(ret) {
		for (status = 0, ap = res; ap != NULL && status < one; ap = ap->ai_next) {
			if(ap->ai_family == family USE_FEATURE_IPV6(|| family == AF_UNSPEC)) {
				memcpy(((char *)ret) + (size * status), ((char *)ap->ai_addr) + offset, size);
				status++;
			}
		}
		if(naddr != NULL)
			*naddr = status;
	}

	if((flags & BB_GHBN_NTOP) != 0) {
		char ibuf[USE_FEATURE_IPV6(INET6_ADDRSTRLEN) SKIP_FEATURE_IPV6(INET_ADDRSTRLEN) + 1];
		ret = xstrdup(inet_ntop(res->ai_family, &((const struct sockaddr_in*)res->ai_addr)->sin_addr, ibuf, sizeof(ibuf)));
	}
	if(canon_name != NULL) {
#if !defined(__UCLIBC__)
		*canon_name = xstrdup(res->ai_canonname);
#else
		/* uclibc do not convert name to fqdn */
		char hbuf[NI_MAXHOST];

		if (getnameinfo(res->ai_addr, res->ai_addrlen, hbuf, sizeof(hbuf), NULL, 0, NI_NAMEREQD) != 0)
			*canon_name = xstrdup(res->ai_canonname); /* invalid */
		else
			*canon_name = xstrdup(hbuf);
#endif
		if(ret == NULL)
		      ret = *canon_name;
	}
	freeaddrinfo(res);
	return ret;
}
