#ifdef APPLET_USAGE
#define login_trivial_usage \
	"[OPTION]... [username] [ENV=VAR ...]"
#define login_full_usage \
	"Begin a new session on the system\n\n" \
	"Options:\n" \
	"\t-f\tDo not authenticate (user already authenticated)\n" \
	"\t-h\tName of the remote host for this login\n" \
	"\t-p\tPreserve environment"
#else

#include "busybox.h"
#include <string.h>
#include <syslog.h>
#include <termios.h>
#include <unistd.h>
#include <utmp.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <ctype.h>
#include <time.h>
#include <fcntl.h>

#include "bb_sig.h"

// login defines
#define TIMEOUT       60
#define EMPTY_USERNAME_COUNT    10
#define USERNAME_SIZE 32

#ifdef CONFIG_FEATURE_UTMP
// import from utmp.c
static void checkutmp(int picky);
static void setutmp(const char *name);
/* Stuff global to this file */
#endif

/* Android-NDK suxx */
#ifndef LOGIN_PROCESS
#define LOGIN_PROCESS 6
#endif
#ifndef USER_PROCESS
#define USER_PROCESS  7
#endif

CALC_BSS struct BB_COMMON_BSS {
#ifdef CONFIG_FEATURE_UTMP
	struct utmp utent;
#endif
	char login_prompt_buf [1024];
#define TTY_NAME_SZ BUFSIZ
	char tty_name[BUFSIZ];
#define FULL_TTY_SZ (BUFSIZ+8)
	char full_tty[BUFSIZ+8];
} bb_common_bss;
#define utent bb_common_bss.utent
#define login_prompt_buf bb_common_bss.login_prompt_buf
#define full_tty bb_common_bss.full_tty
#define tty_name bb_common_bss.tty_name

#if !defined CONFIG_FEATURE_SECURETTY
static inline int check_tty ( const char *tty )  { return 1; }
#endif

static void alarm_handler ( int sig ATTRIBUTE_UNUSED)
{
	fprintf (stderr, "\nLogin timed out after %d seconds.\n", TIMEOUT );
	exit ( EXIT_SUCCESS );
}


static int login_prompt ( char *buf_name )
{
	char *sp, *ep;
	int i;

	for(i=0; i<EMPTY_USERNAME_COUNT; i++) {
		print_login_prompt();

		if ( !fgets ( login_prompt_buf, sizeof( login_prompt_buf ) - 1, stdin ))
			return 0;

		if ( !strchr ( login_prompt_buf, '\n' ))
			return 0;

		for ( sp = login_prompt_buf; isspace ( *sp ); sp++ ) { }
		for ( ep = sp; isgraph ( *ep ); ep++ ) { }

		*ep = 0;
		safe_strncpy(buf_name, sp, USERNAME_SIZE);
		if(buf_name[0])
			return 1;
	}
	return 0;
}


static int check_nologin ( int amroot )
{
	if ( access ( bb_path_nologin_file, F_OK ) == 0 ) {
		FILE *fp;
		int c;

		if (( fp = fopen ( bb_path_nologin_file, "r" ))) {
			while (( c = getc ( fp )) != EOF )
				putchar (( c == '\n' ) ? '\r' : c );

			fflush ( stdout );
			fclose ( fp );
		} else {
			puts ( "\r\nSystem closed for routine maintenance.\r" );
		}
		if ( !amroot )
			return 1;

		puts ( "\r\n[Disconnect bypassed -- root login allowed.]\r" );
	}
	return 0;
}

#ifdef CONFIG_FEATURE_SECURETTY
#include "config_read.h"

static int check_tty ( const char *tty )
{
	config_read_t *fp = config_read_prepare(bb_path_securetty_file, 0);

	if(fp != NULL) {
		while ((fp = config_read(fp, "#\1\0\0")) != NULL) {
			if ( strcmp ( fp->tokens[0], tty ) == 0 ) {
				config_close( fp );
				return 1;
			}
		}
		return 0;
	}
	/* A missing securetty file is not an error. */
	return 1;
}

#endif

/* returns 1 if true */
static int is_my_tty ( const char *tty )
{
	struct stat by_name, by_fd;

	if ( stat ( tty, &by_name ) || fstat ( 0, &by_fd ))
		return 0;

	if ( by_name. st_rdev != by_fd. st_rdev )
		return 0;
	else
		return 1;
}


static void motd (void)
{
	FILE *fp;
	int c;

	if (( fp = fopen ( bb_path_motd_file, "r" ))) {
		while (( c = getc ( fp )) != EOF )
			putchar ( c );
		fclose ( fp );
	}
}


#ifdef CONFIG_FEATURE_UTMP
// vv  Taken from tinylogin utmp.c  vv

#define NO_UTENT \
	"No utmp entry.  You must exec \"login\" from the lowest level \"sh\""
#define NO_TTY \
	"Unable to determine your tty name."

/*
 * checkutmp - see if utmp file is correct for this process
 *
 *      System V is very picky about the contents of the utmp file
 *      and requires that a slot for the current process exist.
 *      The utmp file is scanned for an entry with the same process
 *      ID.  If no entry exists the process exits with a message.
 *
 *      The "picky" flag is for network and other logins that may
 *      use special flags.  It allows the pid checks to be overridden.
 *      This means that getty should never invoke login with any
 *      command line flags.
 */

static void checkutmp(int picky)
{
	char *line;
	struct utmp *ut;
	pid_t pid = getpid();

	setutent();

	/* First, try to find a valid utmp entry for this process.  */
	while ((ut = getutent()))
		if (ut->ut_pid == pid && ut->ut_line[0] && ut->ut_id[0] &&
			(ut->ut_type == LOGIN_PROCESS || ut->ut_type == USER_PROCESS))
			break;

	/* If there is one, just use it, otherwise create a new one.  */
	if (ut) {
		utent = *ut;
	} else {
		time_t t_tmp;

		if (picky) {
			puts(NO_UTENT);
			exit(1);
		}
		line = ttyname(0);
		if (!line) {
			puts(NO_TTY);
			exit(1);
		}
		if (strncmp(line, "/dev/", 5) == 0)
			line += 5;
		memset((void *) &utent, 0, sizeof utent);
		utent.ut_type = LOGIN_PROCESS;
		utent.ut_pid = pid;
		strncpy(utent.ut_line, line, sizeof utent.ut_line);
		/* XXX - assumes /dev/tty?? */
		strncpy(utent.ut_id, utent.ut_line + 3, sizeof utent.ut_id);
		strncpy(utent.ut_user, "LOGIN", sizeof utent.ut_user);
		time(&t_tmp);
		utent.ut_time = t_tmp;
	}
}

/*
 * setutmp - put a USER_PROCESS entry in the utmp file
 *
 *      setutmp changes the type of the current utmp entry to
 *      USER_PROCESS.  the wtmp file will be updated as well.
 */

static void setutmp(const char *name)
{
	time_t t_tmp;

	utent.ut_type = USER_PROCESS;
	strncpy(utent.ut_user, name, sizeof utent.ut_user);
	time(&t_tmp);
	utent.ut_time = t_tmp;
	/* other fields already filled in by checkutmp above */
	setutent();
	pututline(&utent);
	endutent();
#ifdef CONFIG_FEATURE_WTMP
	if (access(_PATH_WTMP, R_OK|W_OK) == -1) {
		close(creat(_PATH_WTMP, 0664));
	}
	updwtmp(_PATH_WTMP, &utent);
#endif
}
#endif /* CONFIG_FEATURE_UTMP */

extern int login_main(void)
{
	char fromhost[512];
	char *s_username;
	char username[USERNAME_SIZE];
	const char *tmp;
	int amroot;
	unsigned long flags;
	int failed;
	int count=0;
	struct passwd *pw, pw_copy;
#ifdef CONFIG_WHEEL_GROUP
	struct group *grp;
#endif
	int opt_preserve;
	int opt_fflag;
	char *opt_host = NULL;
	int alarmstarted = 0;

	username[0] = 0;
	amroot = ( getuid() == 0 );
	bb_signal((1 << SIGALRM), alarm_handler);
	alarm(TIMEOUT);
	alarmstarted = 1;

	flags = bb_getopt_ulflags("pf:h:", &s_username, &opt_host);
	opt_preserve = flags & 1;
	opt_fflag = flags & 2;
	if(opt_fflag) {
		/*
		 * username must be a separate token
		 * (-f root, *NOT* -froot). --marekm
		 */
		if(s_username[-1] == 'f')
			bb_show_usage();
		if ( !amroot )  /* Auth bypass only if real UID is zero */
			bb_error_msg_and_die("-f permission denied");

		safe_strncpy(username, s_username, USERNAME_SIZE);
	}

	if (*Gargv)     // user from command line (getty)
		safe_strncpy(username, *Gargv, USERNAME_SIZE);

	if ( !isatty ( 0 ) || !isatty ( 1 ) || !isatty ( 2 ))
		return EXIT_FAILURE;            /* Must be a terminal */

#ifdef CONFIG_FEATURE_UTMP
	checkutmp ( !amroot );
#endif

	tmp = ttyname ( 0 );
	if ( tmp && ( strncmp ( tmp, "/dev/", 5 ) == 0 ))
		safe_strncpy ( tty_name, tmp + 5, TTY_NAME_SZ);
	else if ( tmp && *tmp == '/' )
		safe_strncpy ( tty_name, tmp, TTY_NAME_SZ);
	else
		safe_strncpy ( tty_name, "UNKNOWN", TTY_NAME_SZ);

#ifdef CONFIG_FEATURE_UTMP
	if ( amroot )
		memset ( utent.ut_host, 0, sizeof utent.ut_host );
#endif

	if ( opt_host ) {
#ifdef CONFIG_FEATURE_UTMP
		safe_strncpy ( utent.ut_host, opt_host, sizeof( utent.ut_host ));
#endif
		snprintf ( fromhost, sizeof( fromhost ) - 1, " on `%.100s' from `%.200s'", tty_name, opt_host );
	}
	else
		snprintf ( fromhost, sizeof( fromhost ) - 1, " on `%.100s'", tty_name );

	setpgrp();

	bb_openlog(LOG_PID | LOG_CONS | LOG_NOWAIT, LOG_AUTH );

	while ( 1 ) {
		failed = 0;

		if ( !username[0] )
			if(!login_prompt ( username ))
				return EXIT_FAILURE;

		if ( !alarmstarted && ( TIMEOUT > 0 )) {
			alarm ( TIMEOUT );
			alarmstarted = 1;
		}

		if (!( pw = getpwnam ( username ))) {
			pw_copy.pw_name   = "UNKNOWN";
			pw_copy.pw_passwd = "!";
			opt_fflag = 0;
			failed = 1;
		} else
			pw_copy = *pw;

		pw = &pw_copy;

		if (pw->pw_passwd == NULL || pw->pw_passwd[0] == '!' || pw->pw_passwd[0] == '*')
			failed = 1;

		if ( opt_fflag ) {
			opt_fflag = 0;
			goto auth_ok;
		}

		if (!failed && ( pw-> pw_uid == 0 ) && ( !check_tty ( tty_name )))
			failed = 1;

		/* Don't check the password if password entry is empty (!) */
		if ( pw->pw_passwd == NULL || !pw->pw_passwd[0] )
			goto auth_ok;

		/* authorization takes place here */
		if ( correct_password ( pw ))
			goto auth_ok;

		failed = 1;

auth_ok:
		if ( !failed)
			break;

		bb_do_delay(FAIL_DELAY);
		puts("Login incorrect");
		username[0] = 0;
		if ( ++count == 3 ) {
			xlog_warn("invalid password for `%s'%s", pw->pw_name, fromhost);
			return EXIT_FAILURE;
		}
	}

	alarm ( 0 );
	if ( check_nologin ( pw-> pw_uid == 0 ))
		return EXIT_FAILURE;

#ifdef CONFIG_FEATURE_UTMP
	setutmp ( username );
#endif

	if ( tty_name[0] != '/' )
		snprintf ( full_tty, FULL_TTY_SZ - 1, "/dev/%s", tty_name);
	else
		safe_strncpy ( full_tty, tty_name, FULL_TTY_SZ - 1 );

	if ( !is_my_tty ( full_tty )) {
		log_syslog = LOG_ERR;
		bb_perror_msg("unable to determine TTY name, got %s", full_tty);
	}

	/* Try these, but don't complain if they fail
	 * (for example when the root fs is read only) */
	chown ( full_tty, pw-> pw_uid, pw-> pw_gid );
	chmod ( full_tty, 0600 );

	change_identity ( pw );
	tmp = pw-> pw_shell;
	if(!tmp || !*tmp)
		tmp = DEFAULT_SHELL;
	setup_environment ( tmp, 1, !opt_preserve, pw );

	motd ( );
	bb_signal((1 << SIGALRM), SIG_DFL);    /* default alarm signal */

	if ( pw-> pw_uid == 0 )
		syslog ( LOG_INFO, "root login %s\n", fromhost );
	run_shell ( tmp, 0, 0);      /* exec the shell finally. */

	return EXIT_FAILURE;
}
#endif /* APPLET_USAGE */
