#ifdef APPLET_USAGE
#define passwd_trivial_usage \
	"[OPTION] [name]"
#define passwd_full_usage \
	"Change a user password. If no name is specified,\n" \
	"changes the password for the current user.\n" \
	"Options:\n" \
	"\t-a\tDefine which algorithm shall be used for the password\n" \
	"\t\t\t(Choices: des, md5)\n" \
	"\t-d\tDelete the password for the specified user account\n" \
	"\t-l\tLocks (disables) the specified user account\n" \
	"\t-u\tUnlocks (re-enables) the specified user account"
#else

#include "busybox.h"
#include <string.h>
#include <unistd.h>
#include <syslog.h>
#include <time.h>
#include <sys/resource.h>
#include <fcntl.h>
#include "bb_shadow.h"

#include "bb_sig.h"

CALC_BSS struct BB_COMMON_BSS
{
	char crypt_passwd[ENCRYPT_BUFFER_SIZE];
	unsigned long x;
	char result[3];
} bb_common_bss;
#define crypt_passwd bb_common_bss.crypt_passwd
#define x bb_common_bss.x
#define result bb_common_bss.result
#define algo bss_flags


static void get_algo(const char *a)
{
	if (a != NULL && strcasecmp(a, "des") == 0)
		algo = 0;
	else
		algo = 1;       /* standard: MD5 */
}

static int create_backup(const char *backup, FILE * fp)
{
	struct stat sb;
	FILE *bkfp;
	int c, mask;

	if (fstat(fileno(fp), &sb))
		/* return -1; */
		return 1;

	mask = umask(077);
	bkfp = fopen(backup, "w");
	umask(mask);
	if (!bkfp)
		/* return -1; */
		return 1;

	/* TODO: faster copy, not one-char-at-a-time.  --marekm */
	rewind(fp);
	while ((c = getc(fp)) != EOF) {
		if (putc(c, bkfp) == EOF)
			break;
	}
	if (c != EOF || fflush(bkfp)) {
		fclose(bkfp);
		/* return -1; */
		return 1;
	}
	if (fclose(bkfp))
		/* return -1; */
		return 1;

	bb_utime(backup, sb.st_atime, sb.st_mtime);
	return 0;
}

static int update_passwd(const struct passwd *pw, const char *crypt_pw)
{
	char filename[1024];
	char buf[1025];
	char buffer[80];
	char username[32];
	char *pw_rest;
	int mask;
	int continued;
	FILE *fp;
	FILE *out_fp;
	struct stat sb;
	struct flock lock;

#if ENABLE_FEATURE_SHADOWPASSWDS
	if (access(bb_path_shadow_file, F_OK) == 0) {
		snprintf(filename, sizeof filename, "%s", bb_path_shadow_file);
	} else
#endif
	{
		snprintf(filename, sizeof filename, "%s", bb_path_passwd_file);
	}

	if (((fp = fopen(filename, "r+")) == 0) || (fstat(fileno(fp), &sb))) {
		/* return 0; */
		return 1;
	}

	/* Lock the password file before updating */
	lock.l_type = F_WRLCK;
	lock.l_whence = SEEK_SET;
	lock.l_start = 0;
	lock.l_len = 0;
	if (fcntl(fileno(fp), F_SETLK, &lock) < 0) {
		bb_perror_msg("%s", filename);
		return 1;
	}
	lock.l_type = F_UNLCK;

	snprintf(buf, sizeof buf, "%s-", filename);
	if (create_backup(buf, fp)) {
		fcntl(fileno(fp), F_SETLK, &lock);
		fclose(fp);
		return 1;
	}
	snprintf(buf, sizeof buf, "%s+", filename);
	mask = umask(0777);
	out_fp = fopen(buf, "w");
	umask(mask);
	if ((!out_fp) || (fchmod(fileno(out_fp), sb.st_mode & 0777))
		|| (fchown(fileno(out_fp), sb.st_uid, sb.st_gid))) {
		fcntl(fileno(fp), F_SETLK, &lock);
		fclose(fp);
		fclose(out_fp);
		return 1;
	}

	continued = 0;
	snprintf(username, sizeof username, "%s:", pw->pw_name);
	rewind(fp);
	while (!feof(fp)) {
		fgets(buffer, sizeof buffer, fp);
		if (!continued) { /* Check to see if we're updating this line.  */
			if (strncmp(username, buffer, strlen(username)) == 0) {
				/* we have a match. */
				pw_rest = strchr_set0(buffer, ':');
				pw_rest = strchr(pw_rest+1, ':');
				fprintf(out_fp, "%s:%s%s", buffer, crypt_pw, pw_rest);
			} else {
				fputs(buffer, out_fp);
			}
		} else {
			fputs(buffer, out_fp);
		}
		if (buffer[strlen(buffer) - 1] == '\n') {
			continued = 0;
		} else {
			continued = 1;
		}
		memset(buffer, 0, sizeof buffer);
	}

	if (fflush(out_fp) || fsync(fileno(out_fp)) || fclose(out_fp)) {
		unlink(buf);
		fcntl(fileno(fp), F_SETLK, &lock);
		fclose(fp);
		return 1;
	}
	if (rename(buf, filename) < 0) {
		fcntl(fileno(fp), F_SETLK, &lock);
		fclose(fp);
		return 1;
	} else {
		fcntl(fileno(fp), F_SETLK, &lock);
		fclose(fp);
		return 0;
	}
}


static int i64c(int i)
{
	if (i <= 0)
		return ('.');
	if (i == 1)
		return ('/');
	if (i >= 2 && i < 12)
		return ('0' - 2 + i);
	if (i >= 12 && i < 38)
		return ('A' - 12 + i);
	if (i >= 38 && i < 63)
		return ('a' - 38 + i);
	return ('z');
}

static char *crypt_make_salt(void)
{
	time_t now;

	time(&now);
	x += now + getpid() + clock();
	result[0] = i64c(((x >> 18) ^ (x >> 6)) & 077);
	result[1] = i64c(((x >> 12) ^ x) & 077);
	result[2] = '\0';
	return result;
}


static int new_password(const struct passwd *pw, int amroot)
{
	char *clear;
	char *cp;
	char enc[ENCRYPT_BUFFER_SIZE];
	char orig[PASSWORD_BUFFER_SIZE];
	char pass[PASSWORD_BUFFER_SIZE];
	char salt[PASSWORD_BUFFER_SIZE]; /* "$N$XXXXXXXX" or "XX" */

	if (!amroot && crypt_passwd[0]) {
		if (!(clear = bb_askpass(0, "Old password:", orig))) {
			/* return -1; */
			return 1;
		}
		if (strcmp(pw_encrypt(clear, crypt_passwd, enc), crypt_passwd) != 0) {
			xlog_warn("incorrect password for `%s'", pw->pw_name);
			bb_do_delay(FAIL_DELAY);
			fprintf(stderr, "Incorrect password.\n");
			/* return -1; */
			return 1;
		}
		memset(enc, 0, sizeof(enc));
	} else {
		orig[0] = '\0';
	}
	if (! (cp=bb_askpass(0, "Enter the new password (minimum of 5, maximum of 8 characters)\n"
					  "Please use a combination of upper and lower case letters and numbers.\n"
					  "Enter new password: ", pass)))
	{
		memset(orig, 0, sizeof orig);
		/* return -1; */
		return 1;
	}

	if (obscure(orig, pass, pw)) {
		if (amroot) {
			printf("\nWarning: weak password (continuing).\n");
		} else {
			/* return -1; */
			return 1;
		}
	}
	if (!(cp = bb_askpass(0, "Re-enter new password: ", orig))) {
		/* return -1; */
		return 1;
	}
	if (strcmp(cp, pass)) {
		fprintf(stderr, "Passwords do not match.\n");
		/* return -1; */
		return 1;
	}
	memset(orig, 0, sizeof(orig));

	if (algo == 1) {
		strcpy(salt, "$1$");
		strcat(salt, crypt_make_salt());
		strcat(salt, crypt_make_salt());
		strcat(salt, crypt_make_salt());
	}

	strcat(salt, crypt_make_salt());
	pw_encrypt(pass, salt, crypt_passwd);

	memset(pass, 0, sizeof pass);
	memset(salt, 0, sizeof pass);
	return 0;
}

static void set_filesize_limit(int blocks)
{
	struct rlimit rlimit_fsize;

	rlimit_fsize.rlim_cur = rlimit_fsize.rlim_max = 512L * blocks;
	setrlimit(RLIMIT_FSIZE, &rlimit_fsize);
}

extern int passwd_main(void)
{
	int amroot;
	char *cp;
	char *np;
	char *name;
	char *myname;
	unsigned long flag;
	int lflg;                       /* -l - lock account */
	int uflg;                       /* -u - unlock account */
	int dflg;                       /* -d - delete password */
	const struct passwd *pw;
	char *s_algo = NULL;

#if ENABLE_FEATURE_SHADOWPASSWDS
	struct spwd *sp, spb;
#endif
	flag = bb_getopt_ulflags("a:dlu", s_algo);
	dflg = flag & 2;
	lflg = flag & 4;
	uflg = flag & 8;
	get_algo(s_algo);       /* -a - password algorithm */

	myname = (char *) xstrdup(bb_getpwuid(NULL, getuid(), -1));
	/* exits on error */
	if (*Gargv != NULL) {
		name = *Gargv;
	} else {
		name = myname;
	}
	amroot = (getuid() == 0);
	if ((lflg || uflg || dflg) && (0 >= Gargc || !amroot))
		bb_show_usage();
	pw = getpwnam(name);
	if (!pw)
		bb_error_msg_and_die("Unknown user %s\n", name);

	bb_openlog(LOG_PID | LOG_CONS | LOG_NOWAIT, LOG_AUTH);

	if (!amroot && pw->pw_uid != getuid()) {
		xlog_warn("can't change pwd for `%s'", name);
		bb_error_msg_and_die("Permission denied.\n");
	}
#if ENABLE_FEATURE_SHADOWPASSWDS
	sp = getspnam(name);
	if (!sp)
		sp = pwd_to_spwd(pw, &spb);
	cp = sp->sp_pwdp;
	np = sp->sp_namp;
#else
	cp = pw->pw_passwd;
	np = name;
#endif

	if(cp != NULL)
		safe_strncpy(crypt_passwd, cp, sizeof(crypt_passwd));
	if (!(dflg || lflg || uflg)) {
		if (!amroot) {
			if (cp[0] == '!') {
				xlog_warn("password locked for `%s'", np);
				bb_error_msg_and_die( "The password for `%s' cannot be changed.\n", np);
			}
		}
		printf("Changing password for %s\n", name);
		if (new_password(pw, amroot)) {
			bb_error_msg_and_die( "The password for %s is unchanged.\n", name);
		}
	} else if (lflg) {
		if (crypt_passwd[0] != '!') {
			memmove(&crypt_passwd[1], crypt_passwd,
					sizeof crypt_passwd - 1);
			crypt_passwd[sizeof crypt_passwd - 1] = '\0';
			crypt_passwd[0] = '!';
		}
	} else if (uflg) {
		if (crypt_passwd[0] == '!') {
			memmove(crypt_passwd, &crypt_passwd[1],
					sizeof crypt_passwd - 1);
		}
	} else if (dflg) {
		crypt_passwd[0] = '\0';
	}
	set_filesize_limit(30000);
	bb_signal( (1 << SIGHUP) | (1 << SIGINT) | (1 << SIGQUIT), SIG_IGN);
	umask(077);
	if (setuid(0)) {
		log_syslog = LOG_ERR; /* add syslog(LOG_ERR...) too */
		bb_error_msg_and_die( "Cannot change ID to root.");
	}
	if (!update_passwd(pw, crypt_passwd)) {
		syslog(LOG_INFO, "password for `%s' changed by user `%s'", name,
			   myname);
		printf("Password changed.\n");
	} else {
		log_syslog = LOG_WARNING; /* add syslog(LOG_WARNING...) too */
		bb_error_msg_and_die("an error occurred updating the password file");
	}
	return 0;
}
#endif /* APPLET_USAGE */
