/*
 *      Chat -- a program for automatic session establishment (i.e. dial
 *              the phone and log in).
 *
 * Standard termination codes:
 *  0 - successful completion of the script
 *  1 - invalid argument, expect string too large, etc.
 *  2 - error on an I/O operation or fatal error condition.
 *  3 - timeout waiting for a simple string.
 *  4 - the first string declared as "ABORT"
 *  5 - the second string declared as "ABORT"
 *  6 - ... and so on for successive ABORT strings.
 *
 *      This software is in the public domain.
 *
 * -----------------
 * Busybox port by Vladimir Oleynik <dzo@simtreas.ru> (c) 2003-2016
 *
 * -----------------
 *      22-May-99 added environment substitutuion, enabled with -E switch.
 *      Andreas Arens <andras@cityweb.de>.
 *
 *      12-May-99 added a feature to read data to be sent from a file,
 *      if the send string starts with @.  Idea from gpk <gpk@onramp.net>.
 *
 *      added -T and -U option and \T and \U substitution to pass a phone
 *      number into chat script. Two are needed for some ISDN TA applications.
 *      Keith Dart <kdart@cisco.com>
 *
 *
 *      Added SAY keyword to send output to stderr.
 *      This allows to turn ECHO OFF and to output specific, user selected,
 *      text to give progress messages. This best works when stderr
 *      exists (i.e.: pppd in nodetach mode).
 *
 *      Added HANGUP directives to allow for us to be called
 *      back. When HANGUP is set to NO, chat will not hangup at HUP signal.
 *      We rely on timeouts in that case.
 *
 *      Added CLR_ABORT to clear previously set ABORT string. This has been
 *      dictated by the HANGUP above as "NO CARRIER" (for example) must be
 *      an ABORT condition until we know the other host is going to close
 *      the connection for call back. As soon as we have completed the
 *      first stage of the call back sequence, "NO CARRIER" is a valid, non
 *      fatal string. As soon as we got called back (probably get "CONNECT"),
 *      we should re-arm the ABORT "NO CARRIER". Hence the CLR_ABORT command.
 *      Note that CLR_ABORT packs the abort_strings[] array so that we do not
 *      have unused entries not being reclaimed.
 *
 *      In the same vein as above, added CLR_REPORT keyword.
 *
 *      Allow for comments. Line starting with '#' are comments and are
 *      ignored. If a '#' is to be expected as the first character, the
 *      expect string must be quoted.
 *
 *
 *              Francis Demierre <Francis@SwissMail.Com>
 *              Thu May 15 17:15:40 MET DST 1997
 *
 *
 *      Added -r "report file" switch & REPORT keyword.
 *              Robert Geer <bgeer@xmission.com>
 *
 *      Added -s "use stderr" and -S "don't use syslog" switches.
 *              June 18, 1997
 *              Karl O. Pinc <kop@meme.com>
 *
 *
 *      Added -e "echo" switch & ECHO keyword
 *              Dick Streefland <dicks@tasking.nl>
 *
 *
 *      Considerable updates and modifications by
 *              Al Longyear <longyear@pobox.com>
 *              Paul Mackerras <paulus@cs.anu.edu.au>
 *
 *
 *      The original author is:
 *
 *              Karl Fox <karl@MorningStar.Com>
 *              Morning Star Technologies, Inc.
 *              1760 Zollinger Road
 *              Columbus, OH  43221
 *              (614)451-1883
 *
 */
#ifdef APPLET_USAGE
#define chat_trivial_usage \
	"[ -vEsS ] [ -T n ] [ -U n ] [ -t t ] [ -f f ] [ -r r ]\n" \
	"[...[[expect[-say[-expect...]]"
#define chat_full_usage \
	"Chat -- a program for automatic session establishment (i.e. dial\n" \
	"the phone and log in).\n\n" \
	"Options:\n" \
	"Options:\n" \
	"\t-v\tRequest that the chat script be executed in a verbose mode.\n" \
	"\t-E\tEnables environment variable substituion within chat scripts\n" \
	"\t\tusing the standard $xxx syntax.\n" \
	"\t-T number\tPass in an arbitary string, usually a phone number,\n" \
	"\t\tthat will be substituted for the \\T substitution metacharacter\n" \
	"\t\tin a send string.\n" \
	"\t-U number\tPass in a second string, usually a phone number, that\n" \
	"\t\twill be substituted for the \\U substitution metacharacter in a\n" \
	"\t\tsend string.\n" \
	"\t-t timeout\tSet the timeout for the expected string to be received.\n" \
	"\t-f file\tRead the chat script from the chat file.\n" \
	"\t-r file\tSet the file for output of the report strings.\n" \
	"\t-s\tUse stderr for log and error messages.\n" \
	"\t-S\tDo not use the SYSLOG."
#else

#define NO_VERBOSE 1

#include <ctype.h>
#include <time.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <syslog.h>
#include <termios.h>
#include "busybox.h"
#include "bb_sig.h"

#define STR_LEN 1024

#ifndef O_NONBLOCK
#define O_NONBLOCK      O_NDELAY
#endif

#define MAX_ABORTS              50
#define MAX_REPORTS             50
#define DEFAULT_CHAT_TIMEOUT    45

CALC_BSS struct BB_COMMON_BSS
{
	int echo;
#ifndef NO_VERBOSE
	int verbose;
	int Verbose;
#endif
	int quiet;
	int use_env;
	int exit_code;
	FILE* report_fp;
	const char *report_file;
	const char *chat_file;
	const char *phone_num;
	const char *phone_num2;

	int have_tty_parameters;
	struct termios saved_tty_parameters;

	char *abort_string[MAX_ABORTS], *fail_reason, fail_buffer[50];
	int n_aborts, abort_next, timeout_next, echo_next;
	int clear_abort_next;

	char *report_string[MAX_REPORTS] ;
	char  report_buffer[50] ;
	int n_reports, report_next, report_gathering;
	int clear_report_next;

	int say_next, hup_next;
	int terminating;
	int need_lf;
	int alarmed;
	char ch_string[10];
	int timeout;
	char *str;
} bb_common_bss;
#define echo bb_common_bss.echo
#define verbose bb_common_bss.verbose
#define Verbose bb_common_bss.Verbose
#define quiet bb_common_bss.quiet
#define use_env bb_common_bss.use_env
#define exit_code bb_common_bss.exit_code
#define report_fp bb_common_bss.report_fp
#define report_file bb_common_bss.report_file
#define chat_file bb_common_bss.chat_file
#define phone_num bb_common_bss.phone_num
#define phone_num2 bb_common_bss.phone_num2
#define have_tty_parameters bb_common_bss.have_tty_parameters
#define saved_tty_parameters bb_common_bss.saved_tty_parameters
#define abort_string bb_common_bss.abort_string
#define fail_reason bb_common_bss.fail_reason
#define fail_buffer bb_common_bss.fail_buffer
#define n_aborts bb_common_bss.n_aborts
#define abort_next bb_common_bss.abort_next
#define timeout_next bb_common_bss.timeout_next
#define echo_next bb_common_bss.echo_next
#define clear_abort_next bb_common_bss.clear_abort_next
#define report_string bb_common_bss.report_string
#define report_buffer bb_common_bss.report_buffer
#define n_reports bb_common_bss.n_reports
#define report_next bb_common_bss.report_next
#define report_gathering bb_common_bss.report_gathering
#define clear_report_next bb_common_bss.clear_report_next
#define say_next bb_common_bss.say_next
#define hup_next bb_common_bss.hup_next
#define terminating bb_common_bss.terminating
#define need_lf bb_common_bss.need_lf
#define alarmed bb_common_bss.alarmed
#define ch_string bb_common_bss.ch_string
#define timeout bb_common_bss.timeout
#define str bb_common_bss.str

/*
 * Translate the input character to the appropriate string for printing
 * the data.
 */

static int character(int c)
{
	char *meta;

	if (c == '\r') /* ignore '\r' */
		return 0;
	if (c == '\n') {
		ch_string[0] = '\n';
		ch_string[1] = '\0';
		return 1;
	}
	meta = (c & 0x80) ? "M-" : "";
	c &= 0x7F;

	if (c < 32)
		c = sprintf(ch_string, "%s^%c", meta, (int)c + '@');
	else if (c == 127)
		c = sprintf(ch_string, "%s^?", meta);
	else
		c = sprintf(ch_string, "%s%c", meta, c);

	return c;
}

/*
 *      Echo a character to stderr.
 *      When called with -1, a '\n' character is generated when
 *      the cursor is not at the beginning of a line.
 */
static void echo_stderr(int c)
{
	int n;

	if(c == -1) {
		if (need_lf == 0)
			return;
		c = '\n';
	}
	n = character(c);
	if(n > 0) {
		write(2, ch_string, n);
		need_lf = c != '\n';
	}
}

/*
 * Send a message to syslog and/or stderr.
 */

static void chat_logf (const char *fmt, ...) __attribute__ ((format (printf, 1, 2)));
static void chat_logf (const char *fmt, ...)
{
	va_list args;
	int prev_log_syslog = log_syslog;

	log_syslog = log_syslog ? LOG_INFO : 0;
	va_start(args, fmt);
	bb_verror_msg(fmt, args);
	va_end(args);
	log_syslog = prev_log_syslog;
}

static int get_char(void);
static void fatal (int code, const char *fmt, ...) __attribute__ ((noreturn, format (printf, 2, 3)));

static void terminate(int status) __attribute__ ((noreturn));
static void terminate(int status)
{
    if (terminating)
	exit(status);
    terminating = 1;
    echo_stderr(-1);
/*
 * Allow the last of the report string to be gathered before we terminate.
 */
    if (report_gathering && report_fp != (FILE *) NULL) {
	int c, rep_len;

	rep_len = strlen(report_buffer);
	while (rep_len + 1 <= sizeof(report_buffer)) {
	    alarm(1);
	    c = get_char();
	    alarm(0);
	    if (c < 0 || iscntrl(c))
		break;
	    report_buffer[rep_len] = c;
	    ++rep_len;
	}
	report_buffer[rep_len] = 0;
	fprintf (report_fp, "chat:  %s\n", report_buffer);
    }
    if (report_file != (char *) 0 && report_fp != (FILE *) NULL) {
#ifndef NO_VERBOSE
	if (verbose)
	    fprintf (report_fp, "Closing \"%s\".\n", report_file);
#endif
	fclose (report_fp);
	report_fp = (FILE *) NULL;
    }

    if (have_tty_parameters) {
	if (tcsetattr(0, TCSANOW, &saved_tty_parameters) < 0) {
	    bb_default_error_retval = 2;
	    bb_perror_msg_and_die("can't restore terminal parameters");
	}
    }

    exit(status);
}

/*
 *      Print an error message and terminate.
 */

static void fatal (int code, const char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    bb_verror_msg(fmt, args);
    va_end(args);
    terminate(code);
}

static int get_char(void)
{
    int status;
    char c;

    status = read(0, &c, 1);

    switch (status) {
    case 1:
	return ((int)c & 0x7F);

    default:
	chat_logf("warning: read() on stdin returned %d", status);

    case -1:
	if ((status = fcntl(0, F_GETFL, 0)) == -1)
	    fatal(2, "Can't get file mode flags on stdin: %m");

	if (fcntl(0, F_SETFL, status & ~O_NONBLOCK) == -1)
	    fatal(2, "Can't set file mode flags on stdin: %m");

	return -1;
    }
}


static void sigalrm(int signo)
{
    int flags;

    alarm(1);
    alarmed = 1;                /* Reset alarm to avoid race window */
    /* that can cause hanging in read() */
    bb_signal((1 << SIGALRM), sigalrm);

    if ((flags = fcntl(0, F_GETFL, 0)) == -1)
	fatal(2, "Can't get file mode flags on stdin: %m");

    if (fcntl(0, F_SETFL, flags | O_NONBLOCK) == -1)
	fatal(2, "Can't set file mode flags on stdin: %m");

#ifndef NO_VERBOSE
    if (verbose)
	chat_logf("alarm");
#endif
}

static void sigint(int signo)
{
    fatal(2, "SIGINT");
}

static void sigterm(int signo)
{
    fatal(2, "SIGTERM");
}

static void sighup(int signo)
{
    fatal(2, "SIGHUP");
}

static void set_tty_parameters(void)
{
    struct termios t;

    if (tcgetattr(0, &t) < 0)
	fatal(2, "Can't get terminal parameters: %m");

    saved_tty_parameters = t;
    have_tty_parameters  = 1;

    t.c_iflag     |= IGNBRK | ISTRIP | IGNPAR;
    t.c_oflag      = 0;
    t.c_lflag      = 0;
    t.c_cc[VERASE] =
    t.c_cc[VKILL]  = 0;
    t.c_cc[VMIN]   = 1;
    t.c_cc[VTIME]  = 0;

    if (tcsetattr(0, TCSANOW, &t) < 0)
	fatal(2, "Can't set terminal parameters: %m");
}

/*
    array - The address of the array of string pointers
    end -   The index of the next free entry before CLR_
*/
static void
pack_array (char **array, int end)
{
    int i, j;

    for (i = 0; i < end; i++) {
	if (array[i] == NULL) {
	    for (j = i+1; j < end; ++j)
		if (array[j] != NULL)
		    array[i++] = array[j];
	    for (; i < end; ++i)
		array[i] = NULL;
	    break;
	}
    }
}

/*
 *      'Clean up' this string.
 * sending - set to 1 when sending (putting) this string
 */
static char *clean(const char *s, int sending)
{
    char temp[STR_LEN], env_str[STR_LEN], cur_chr;
    char *s1, *phchar;
    int add_return = sending;
#define isoctal(chr)    (((chr) >= '0') && ((chr) <= '7'))
#define isalnumx(chr)   ((((chr) >= '0') && ((chr) <= '9')) \
			 || (((chr) >= 'a') && ((chr) <= 'z')) \
			 || (((chr) >= 'A') && ((chr) <= 'Z')) \
			 || (chr) == '_')

    s1 = temp;
    while (*s) {
	cur_chr = *s++;
	if (cur_chr == '^') {
	    cur_chr = *s++;
	    if (cur_chr == '\0') {
		*s1++ = '^';
		break;
	    }
	    cur_chr &= 0x1F;
	    if (cur_chr != 0) {
		*s1++ = cur_chr;
	    }
	    continue;
	}

	if (use_env && cur_chr == '$') {                /* ARI */
	    phchar = env_str;
	    while (isalnumx(*s))
		*phchar++ = *s++;
	    *phchar = '\0';
	    phchar = getenv(env_str);
	    if (phchar)
		while (*phchar)
		    *s1++ = *phchar++;
	    continue;
	}

	if (cur_chr != '\\') {
	    *s1++ = cur_chr;
	    continue;
	}

	cur_chr = *s++;
	if (cur_chr == '\0') {
	    if (sending) {
		*s1++ = '\\';
		*s1++ = '\\';
	    }
	    break;
	}

	switch (cur_chr) {
	case 'b':
	    *s1++ = '\b';
	    break;

	case 'c':
	    if (sending && *s == '\0')
		add_return = 0;
	    else
		*s1++ = cur_chr;
	    break;

	case '\\':
	case 'K':
	case 'p':
	case 'd':
	    if (sending)
		*s1++ = '\\';
	    *s1++ = cur_chr;
	    break;

	case 'T':
	    if (sending && phone_num) {
		const char *pn;

		for (pn = phone_num; *pn != '\0'; pn++)
		    *s1++ = *pn;
	    }
	    else {
		*s1++ = '\\';
		*s1++ = 'T';
	    }
	    break;

	case 'U':
	    if (sending && phone_num2) {
		const char *pn;

		for (pn = phone_num2; *pn != '\0'; pn++)
		    *s1++ = *pn;
	    }
	    else {
		*s1++ = '\\';
		*s1++ = 'U';
	    }
	    break;

	case 'q':
	    quiet = 1;
	    break;

	case 'r':
	    *s1++ = '\r';
	    break;

	case 'n':
	    *s1++ = '\n';
	    break;

	case 's':
	    *s1++ = ' ';
	    break;

	case 't':
	    *s1++ = '\t';
	    break;

	case 'N':
	    if (sending) {
		*s1++ = '\\';
		*s1++ = '\0';
	    }
	    else
		*s1++ = 'N';
	    break;

	case '$':                       /* ARI */
	    if (use_env) {
		*s1++ = cur_chr;
		break;
	    }
	    /* FALL THROUGH */

	default:
	    if (isoctal (cur_chr)) {
		cur_chr &= 0x07;
		if (isoctal (*s)) {
		    cur_chr <<= 3;
		    cur_chr |= *s++ - '0';
		    if (isoctal (*s)) {
			cur_chr <<= 3;
			cur_chr |= *s++ - '0';
		    }
		}

		if (cur_chr != 0 || sending) {
		    if (sending && (cur_chr == '\\' || cur_chr == 0))
			*s1++ = '\\';
		    *s1++ = cur_chr;
		}
		break;
	    }

	    if (sending)
		*s1++ = '\\';
	    *s1++ = cur_chr;
	    break;
	}
    }

    if (add_return)
	*s1++ = '\r';

    *s1++ = '\0'; /* guarantee closure */
    *s1++ = '\0'; /* terminate the string */
    phchar = malloc ((size_t) (s1 - temp));
    if (phchar == NULL)
	fatal(2, "memory error!");

    memcpy (phchar, temp, (size_t) (s1 - temp));
    return phchar;
}

static int put_char(int c)
{
    int status;
    char ch = c;

    xusleep(10000);             /* inter-character typing delay (?) */

    status = write(1, &ch, 1);

    switch (status) {
    case 1:
	return 0;

    default:
	chat_logf("warning: write() on stdout returned %d", status);

    case -1:
	if ((status = fcntl(0, F_GETFL, 0)) == -1)
	    fatal(2, "Can't get file mode flags on stdin, %m");

	if (fcntl(0, F_SETFL, status & ~O_NONBLOCK) == -1)
	    fatal(2, "Can't set file mode flags on stdin: %m");

	return -1;
    }
}

static int write_char (int c)
{
    if (alarmed || put_char(c) < 0) {
	alarm(0);
	alarmed = 0;

#ifndef NO_VERBOSE
	if (verbose) {
	    if (errno == EINTR || errno == EWOULDBLOCK)
		chat_logf(" -- write timed out");
	    else
		chat_logf(" -- write failed: %m");
	}
#endif
	return 0;
    }
    return 1;
}

static int put_string (char *s)
{
    quiet = 0;
    s = clean(s, 1);

#ifndef NO_VERBOSE
    if (verbose) {
	if (quiet)
	    chat_logf("send (?????\?)");
	else
	    chat_logf("send (%s)", s);
    }
#endif

    alarm(timeout); alarmed = 0;

    while (*s) {
	char c = *s++;

	if (c != '\\') {
	    if (!write_char (c))
		return 0;
	    continue;
	}

	c = *s++;
	switch (c) {
	case 'd':
	    sleep(1);
	    break;

	case 'K':
#ifdef TERMIOS
	    tcsendbreak (0, 0);
#endif
	    break;

	case 'p':
	    xusleep(10000);     /* 1/100th of a second (arg is microseconds) */
	    break;

	default:
	    if (!write_char (c))
		return 0;
	    break;
	}
    }

    alarm(0);
    alarmed = 0;
    return 1;
}

/*
 *  process the reply string
 */
static void chat_send (char *s)
{
    char file_data[STR_LEN];

    if (say_next) {
	say_next = 0;
	s = clean(s, 1);
	write(2, s, strlen(s));
	free(s);
	return;
    }

    if (hup_next) {
	hup_next = 0;
	if (strcmp(s, "OFF") == 0)
	   bb_signal((1 << SIGHUP), SIG_IGN);
	else
	   bb_signal((1 << SIGHUP), sighup);
	return;
    }

    if (echo_next) {
	echo_next = 0;
	echo = (strcmp(s, "ON") == 0);
	return;
    }

    if (abort_next) {
	char *s1;

	abort_next = 0;

	if (n_aborts >= MAX_ABORTS)
	    fatal(2, "Too many ABORT strings");

	s1 = clean(s, 0);

	if (strlen(s1) > strlen(s)
	    || strlen(s1) + 1 > sizeof(fail_buffer))
	    fatal(1, "Illegal or too-long ABORT string ('%s')", s);

	abort_string[n_aborts++] = s1;

#ifndef NO_VERBOSE
	if (verbose)
	    chat_logf("abort on (%s)", s);
#endif
	return;
    }

    if (clear_abort_next) {
	char *s1;
	int   i;
	int   old_max;
	int   pack = 0;

	clear_abort_next = 0;

	s1 = clean(s, 0);

	if (strlen(s1) > strlen(s)
	    || strlen(s1) + 1 > sizeof(fail_buffer))
	    fatal(1, "Illegal or too-long CLR_ABORT string ('%s')", s);

	old_max = n_aborts;
	for (i=0; i < n_aborts; i++) {
	    if ( strcmp(s1,abort_string[i]) == 0 ) {
		free(abort_string[i]);
		abort_string[i] = NULL;
		pack++;
		n_aborts--;
#ifndef NO_VERBOSE
		if (verbose)
		    chat_logf("clear abort on (%s)", s);
#endif
	    }
	}
	free(s1);
	if (pack)
	    pack_array(abort_string,old_max);
	return;
    }

    if (report_next) {
	char *s1;

	report_next = 0;
	if (n_reports >= MAX_REPORTS)
	    fatal(2, "Too many REPORT strings");

	s1 = clean(s, 0);

	if (strlen(s1) > strlen(s) || strlen(s1) > sizeof fail_buffer - 1)
	    fatal(1, "Illegal or too-long REPORT string ('%s')", s);

	report_string[n_reports++] = s1;

#ifndef NO_VERBOSE
	if (verbose)
	    chat_logf("report (%s)", s);
#endif
	return;
    }

    if (clear_report_next) {
	char *s1;
	int   i;
	int   old_max;
	int   pack = 0;

	clear_report_next = 0;

	s1 = clean(s, 0);

	if (strlen(s1) > strlen(s) || strlen(s1) > sizeof fail_buffer - 1)
	    fatal(1, "Illegal or too-long REPORT string ('%s')", s);

	old_max = n_reports;
	for (i=0; i < n_reports; i++) {
	    if ( strcmp(s1,report_string[i]) == 0 ) {
		free(report_string[i]);
		report_string[i] = NULL;
		pack++;
		n_reports--;
#ifndef NO_VERBOSE
		if (verbose)
		    chat_logf("clear report (%s)", s);
#endif
	    }
	}
	free(s1);
	if (pack)
	    pack_array(report_string,old_max);

	return;
    }

    if (timeout_next) {
	timeout_next = 0;
	s = clean(s, 0);
	timeout = atoi(s);

	if (timeout <= 0)
	    timeout = DEFAULT_CHAT_TIMEOUT;

#ifndef NO_VERBOSE
	if (verbose)
	    chat_logf("timeout set to %d seconds", timeout);
#endif
	return;
    }

    /*
     * The syntax @filename means read the string to send from the
     * file `filename'.
     */
    if (s[0] == '@') {
	/* skip the @ and any following white-space */
	char *fn = s;
	while (*++fn == ' ' || *fn == '\t')
	    ;

	if (*fn != 0) {
	    FILE *f;
	    int n = 0;

	    /* open the file and read until STR_LEN-1 bytes or end-of-file */
	    f = fopen(fn, "r");
	    if (f == NULL)
		fatal(1, "%s -- open failed: %m", fn);
	    while (n < STR_LEN - 1) {
		int nr = fread(&file_data[n], 1, STR_LEN - 1 - n, f);
		if (nr < 0)
		    fatal(1, "%s -- read error", fn);
		if (nr == 0)
		    break;
		n += nr;
	    }
	    fclose(f);

	    /* use the string we got as the string to send,
	       but trim off the final newline if any. */
	    if (n > 0 && file_data[n-1] == '\n')
		--n;
	    file_data[n] = 0;
	    s = file_data;
	}
    }

    if (strcmp(s, "EOT") == 0)
	s = "^D\\c";
    else if (strcmp(s, "BREAK") == 0)
	s = "\\K\\c";

    if (!put_string(s))
	fatal(1, "Failed");
}

/*
 * A modified version of 'strtok'. This version skips \ sequences.
 */

static char *expect_strtok (char *s, const char *term)
{
    int     escape_flag = 0;
    char   *result;

/*
 * If a string was specified then do initial processing.
 */
    if (s)
	str = s;

/*
 * If this is the escape flag then reset it and ignore the character.
 */
    if (*str)
	result = str;
    else
	result = (char *) 0;

    while (*str) {
	if (escape_flag) {
	    escape_flag = 0;
	    ++str;
	    continue;
	}

	if (*str == '\\') {
	    ++str;
	    escape_flag = 1;
	    continue;
	}

/*
 * If this is not in the termination string, continue.
 */
	if (strchr (term, *str) == NULL) {
	    ++str;
	    continue;
	}

/*
 * This is the terminator. Mark the end of the string and stop.
 */
	*str++ = '\0';
	break;
    }
    return (result);
}

/*
 *      'Wait for' this string to appear on this file descriptor.
 */
static int get_string(char *string)
{
    char temp[STR_LEN];
    int c, len, minlen;
#ifndef NO_VERBOSE
    printed = 0;
#endif
    char *s = temp, *end = s + STR_LEN;
    char *logged = temp;

    fail_reason = (char *)0;
    string = clean(string, 0);
    len = strlen(string);
    minlen = (len > sizeof(fail_buffer)? len: sizeof(fail_buffer)) - 1;

#ifndef NO_VERBOSE
    if (verbose)
	chat_logf("expect (%s)", string);
#endif

    if (len > STR_LEN) {
	chat_logf("expect string is too long");
	exit_code = 1;
	return 0;
    }

    if (len == 0) {
#ifndef NO_VERBOSE
	if (verbose)
	    chat_logf("got it");
#endif
	return 1;
    }

    alarm(timeout);
    alarmed = 0;

    while ( ! alarmed && (c = get_char()) >= 0) {
	int n, abort_len, report_len;

	if (echo)
	    echo_stderr(c);
#ifndef NO_VERBOSE
	if (verbose && c == '\n') {
	    if (s == logged)
		chat_logf("");       /* blank line */
	    else
		chat_logf("%0.*v", s - logged, logged);
	    logged = s + 1;
	}

#endif
	*s++ = c;

#ifndef NO_VERBOSE
	if (verbose && s >= logged + 80) {
	    chat_logf("%0.*v", s - logged, logged);
	    logged = s;
	}

	if (Verbose)
	   echo_stderr(c);
#endif
	if (!report_gathering) {
	    for (n = 0; n < n_reports; ++n) {
		if ((report_string[n] != (char*) NULL) &&
		    s - temp >= (report_len = strlen(report_string[n])) &&
		    strncmp(s - report_len, report_string[n], report_len) == 0) {
		    time_t time_now   = time ((time_t*) NULL);
		    struct tm* tm_now = localtime (&time_now);

		    strftime (report_buffer, 20, "%b %d %H:%M:%S ", tm_now);
		    strcat (report_buffer, report_string[n]);

		    report_string[n] = (char *) NULL;
		    report_gathering = 1;
		    break;
		}
	    }
	} else {
	    if (!iscntrl (c)) {
		int rep_len = strlen (report_buffer);
		report_buffer[rep_len]     = c;
		report_buffer[rep_len + 1] = '\0';
	    }
	    else {
		report_gathering = 0;
		fprintf (report_fp, "chat:  %s\n", report_buffer);
	    }
	}

	if (s - temp >= len && c == string[len - 1] &&
	    strncmp(s - len, string, len) == 0) {
#ifndef NO_VERBOSE
	    if (verbose) {
		if (s > logged)
		    chat_logf("%0.*v", s - logged, logged);
		chat_logf(" -- got it\n");
	    }
#endif

	    alarm(0);
	    alarmed = 0;
	    return 1;
	}

	for (n = 0; n < n_aborts; ++n) {
	    if (s - temp >= (abort_len = strlen(abort_string[n])) &&
		strncmp(s - abort_len, abort_string[n], abort_len) == 0) {
#ifndef NO_VERBOSE
		if (verbose) {
		    if (s > logged)
			chat_logf("%0.*v", s - logged, logged);
		    chat_logf(" -- failed");
		}
#endif

		alarm(0);
		alarmed = 0;
		exit_code = n + 4;
		strcpy(fail_reason = fail_buffer, abort_string[n]);
		return 0;
	    }
	}

	if (s >= end) {
	    if (logged < s - minlen) {
#ifndef NO_VERBOSE
		chat_logf("%0.*v", s - logged, logged);
#endif
		logged = s;
	    }
	    s -= minlen;
	    memmove(temp, s, minlen);
	    logged = temp + (logged - s);
	    s = temp + minlen;
	}

#ifndef NO_VERBOSE
	if (alarmed && verbose)
	    chat_logf("warning: alarm synchronization problem");
#endif
    }

    alarm(0);

#ifndef NO_VERBOSE
    if (verbose && printed) {
	if (alarmed)
	    chat_logf(" -- read timed out");
	else
	    chat_logf(" -- read failed: %m");
    }
#endif

    exit_code = 3;
    alarmed   = 0;
    return 0;
}

/*
 * Process the expect string
 */

static void chat_expect (char *s)
{
    char *expect;
    char *reply;

    if (strcmp(s, "HANGUP") == 0) {
	++hup_next;
	return;
    }

    if (strcmp(s, "ABORT") == 0) {
	++abort_next;
	return;
    }

    if (strcmp(s, "CLR_ABORT") == 0) {
	++clear_abort_next;
	return;
    }

    if (strcmp(s, "REPORT") == 0) {
	++report_next;
	return;
    }

    if (strcmp(s, "CLR_REPORT") == 0) {
	++clear_report_next;
	return;
    }

    if (strcmp(s, "TIMEOUT") == 0) {
	++timeout_next;
	return;
    }

    if (strcmp(s, "ECHO") == 0) {
	++echo_next;
	return;
    }

    if (strcmp(s, "SAY") == 0) {
	++say_next;
	return;
    }

/*
 * Fetch the expect and reply string.
 */
    for (;;) {
	expect = expect_strtok (s, "-");
	s      = (char *) 0;

	if (expect == (char *) 0)
	    return;

	reply = expect_strtok (s, "-");

/*
 * Handle the expect string. If successful then exit.
 */
	if (get_string (expect))
	    return;

/*
 * If there is a sub-reply string then send it. Otherwise any condition
 * is terminal.
 */
	if (reply == (char *) 0 || exit_code != 3)
	    break;

	chat_send (reply);
    }

/*
 * The expectation did not occur. This is terminal.
 */
    if (fail_reason)
	chat_logf("Failed (%s)", fail_reason);
    else
	chat_logf("Failed");
    terminate(exit_code);
}


/*
 *  Process a chat script when read from a file.
 */

static void do_file (void)
{
    int linect, sendflg;
    char *sp, *arg, quote;
    char buf [STR_LEN];
    FILE *cfp;

    cfp = fopen (chat_file, "r");
    if (cfp == NULL)
	fatal(1, "%s -- open failed: %m", chat_file);

    linect = 0;
    sendflg = 0;

    while (fgets(buf, STR_LEN, cfp) != NULL) {
	strchr_set0 (buf, '\n');

	linect++;
	sp = buf;

	/* lines starting with '#' are comments. If a real '#'
	   is to be expected, it should be quoted .... */
	if ( *sp == '#' )
	    continue;

	while (*sp != '\0') {
	    if (*sp == ' ' || *sp == '\t') {
		++sp;
		continue;
	    }

	    if (*sp == '"' || *sp == '\'') {
		quote = *sp++;
		arg = sp;
		while (*sp != quote) {
		    if (*sp == '\0')
			fatal(1, "unterminated quote (line %d)", linect);

		    if (*sp++ == '\\') {
			if (*sp != '\0')
			    ++sp;
		    }
		}
	    }
	    else {
		arg = sp;
		while (*sp != '\0' && *sp != ' ' && *sp != '\t')
		    ++sp;
	    }

	    if (*sp != '\0')
		*sp++ = '\0';

	    if (sendflg)
		chat_send (arg);
	    else
		chat_expect (arg);
	    sendflg = !sendflg;
	}
    }
    fclose (cfp);
}

/*
 * chat [ -vEsS ] [ -T number ] [ -U number ] [ -t timeout ] [ -f chat-file ] \
 * [ -r report-file ] \
 *              [...[[expect[-say[-expect...]] say expect[-say[-expect]] ...]]]
 *
 *      Perform a UUCP-dialer-like chat script on stdin and stdout.
 */
extern int
chat_main(void)
{
    char **argv;
    unsigned long flags;
    long lt = DEFAULT_CHAT_TIMEOUT;

    str   = "";
    tzset();

    flags = bb_xgetopt_ulflags("eEvVsSf:t:r:T:U:\0:t", &chat_file, &lt,
	&report_file, &phone_num, &phone_num2);

    echo = flags & 1;
    use_env = flags & 2;
#ifndef NO_VERBOSE
    verbose = flags & 4;
    Verbose = flags & 8;
#endif
    log_stderr = flags & 16;
    if((flags & 32) == 0)
	log_syslog = LOG_ERR;
    timeout = lt;

    if(report_file) {
	    report_fp = fopen (report_file, "a");
#ifndef NO_VERBOSE
	    if (report_fp != NULL) {
		if (verbose)
			fprintf (report_fp, "Opening \"%s\"...\n", report_file);
	    }
#endif
    }
/*
 * Default the report file to the stderr location
 */
    if (report_fp == NULL)
	report_fp = stderr;

    if (log_syslog) {
	bb_openlog(LOG_PID | LOG_NDELAY, LOG_LOCAL2);

#ifndef NO_VERBOSE
	if (verbose)
	    setlogmask(LOG_UPTO(LOG_INFO));
	else
#endif
	    setlogmask(LOG_UPTO(LOG_WARNING));
    }

    bb_signal((1 << SIGINT), sigint);
    bb_signal((1 << SIGTERM), sigterm);
    bb_signal((1 << SIGHUP), sighup);

    set_tty_parameters();
    bb_signal((1 << SIGALRM), sigalrm);

    argv = Gargv;

    if (chat_file != NULL) {
	if (*argv)
	    bb_show_usage();
	else
	    do_file ();
    } else {
	while (*argv) {
	    chat_expect(*argv);
	    argv++;
	    if (*argv) {
		chat_send(*argv);
		argv++;
	    }
	}
    }

    terminate(0);
    return 0;
}
#endif /* APPLET_USAGE */
