/* General purpose Linux console screen save/restore server
   Copyright (C) 1994 Janne Kukonlehto <jtklehto@stekt.oulu.fi>
   Original idea from Unix Interactive Tools version 3.2b (tty.c)
   This code requires root privileges.
   You may want to make the cons.saver setuid root.
   The code should be safe even if it is setuid but who knows?

   Busybox port by Vladimir Oleynik <dzo@simtreas.ru> (c) 2003-2007

   Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
*/
#ifdef APPLET_USAGE
#define cons_saver_trivial_usage \
	"<ttyname>"
#define cons_saver_full_usage \
	"General purpose Linux console screen save/restore server for mc"
#else

#include "busybox.h"
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <ctype.h>      /* For isdigit() */

typedef struct WINDOW WINDOW;

enum {
    CONSOLE_INIT = '1',
    CONSOLE_DONE,
    CONSOLE_SAVE,
    CONSOLE_RESTORE,
    CONSOLE_CONTENTS
};

#define cmd_input 0
#define cmd_output 1

CALC_BSS struct BB_COMMON_BSS
{
	char *tty_name;
	int len;
	char *conss_buffer;
	int buffer_size;
	int columns, rows;
	int vcs_fd;
	int console_fd;
/* Meaning of console_flag:
   -1 == to be detected,
   0  == not a console
   1  == is a console, Linux < 1.1.67 (black & white)
   2  == is a console, Linux >= 1.1.67 (color)
   3  == is a console, Linux >= 1.1.92 (color, use /dev/vcsa$num
   */
	signed char console_flag;
} bb_common_bss;
#define tty_name bb_common_bss.tty_name
#define len bb_common_bss.len
#define conss_buffer bb_common_bss.conss_buffer
#define buffer_size bb_common_bss.buffer_size
#define columns bb_common_bss.columns
#define rows bb_common_bss.rows
#define vcs_fd bb_common_bss.vcs_fd
#define console_fd bb_common_bss.console_fd
#define console_flag bb_common_bss.console_flag


static void dwrite (int fd, const char *buf)
{
    write (fd, buf, strlen (buf));
}

static void write_cmd (const char *buf)
{
    write (cmd_output, buf, 1);
}

static inline void tty_getsize (void)
{
    struct winsize winsz;

    winsz.ws_col = winsz.ws_row = 0;
    ioctl (console_fd, TIOCGWINSZ, &winsz);
    if (winsz.ws_col && winsz.ws_row){
	columns = winsz.ws_col;
	rows    = winsz.ws_row;
    } else {
	/* Never happens (I think) */
	dwrite (2, "TIOCGWINSZ failed\n");
	columns = 80;
	rows = 25;
	console_flag = 0;
    }
}

static inline void tty_cursormove(int y, int x)
{
    char buffer [20];

    /* Standard ANSI escape sequence for cursor positioning */
    sprintf (buffer,"\33[%d;%dH", y + 1, x + 1);
    dwrite (console_fd, buffer);
}

static int check_file (const char *filename, int check_console, const char **msg)
{
    int fd;
    struct stat stat_buf;

    /* Avoiding race conditions: use of fstat makes sure that
       both 'open' and 'stat' operate on the same file */

    *msg = 0;

    fd = open (filename, O_RDWR);
    if (fd == -1)
	return -1;

    if (fstat (fd, &stat_buf) == -1)
	return -1;

    /* Must be character device */
    if (!S_ISCHR (stat_buf.st_mode)){
	*msg = "Not a character device";
	return -1;
    }

#ifdef DEBUG
    fprintf (stderr, "Device: %x\n", stat_buf.st_rdev);
#endif
    if (check_console){
	/* Second time: must be console */
	if ((stat_buf.st_rdev & 0xff00) != 0x0400){
	    *msg = "Not a console";
	    return -1;
	}

	if ((stat_buf.st_rdev & 0x00ff) > 63){
	    *msg = "Minor device number too big";
	    return -1;
	}

	/* Must be owned by the user */
	if (stat_buf.st_uid != getuid ()){
	    *msg = "Not a owner";
	    return -1;
	}
    }

    /* Everything seems to be okay */
    return fd;
}

/* Detect console */
/* Because the name of the tty is supplied by the user and this
   can be a setuid program a lot of checks has to done to avoid
   creating a security hole */
static const char *detect_console (void)
{
    const char *msg;
    int  xlen;
    char vcs_name [40];

    /* Must be console */
    /* Handle the case for /dev/tty?? */
    if (tty_name[len-5] == 't')
	xlen = len - 1;
    else
	xlen = len;

    /* General: /dev/ttyn */
    if (tty_name[xlen - 5] != '/' ||
	tty_name[xlen - 4] != 't' ||
	tty_name[xlen - 3] != 't' ||
	tty_name[xlen - 2] != 'y' ||
	!isdigit(tty_name[xlen - 1]) ||
	!isdigit(tty_name[len - 1]))
	return "Doesn't look like console";

    sprintf (vcs_name, "/dev/vcsa%s", tty_name + xlen - 1);
    vcs_fd = check_file (vcs_name, 0, &msg);
    console_fd = check_file (tty_name, 1, &msg);

#ifdef DEBUG
    fprintf (stderr, "vcs_fd = %d console_fd = %d\n", vcs_fd, console_fd);
#endif

    if (vcs_fd != -1){
	console_flag = 3;
    }

    if (console_fd == -1)
	return msg;

    return NULL;
}

static void save_console (void)
{
    int i;

    if (!console_flag)
	return;
    conss_buffer [1] = tty_name [len-1] - '0';
    if (console_flag >= 2){
	/* Linux >= 1.1.67 */
	/* Get screen contents and cursor position */
	conss_buffer [0] = 8;
	if (console_flag == 2){
	    if ((i = ioctl (console_fd, TIOCLINUX, conss_buffer)) == -1){
		/* Oops, this is not Linux 1.1.67 */
		console_flag = 1;
	    }
	} else {
	    lseek (vcs_fd, 0, 0);
	    read (vcs_fd, conss_buffer, buffer_size);
	}
    }
    if (console_flag == 1){
	int cxy_index, x, y;

	/* Linux < 1.1.67 */
	/* Get screen contents */
	conss_buffer [0] = 0;
	if (ioctl(console_fd, TIOCLINUX, conss_buffer) == -1){
	    conss_buffer[0] = conss_buffer[1] = 0;

	    /* Linux bug: bad ioctl on console 8 */
	    if (ioctl(console_fd, TIOCLINUX, conss_buffer) == -1){
		/* Oops, this is not a console after all */
		console_flag = 0;
		return;
	    }
	}
	/* Select the beginning of the bottommost empty line
	   to be the cursor position */
	cxy_index = 2 + rows * columns;
	for (y = rows - 1; y >= 0; y--)
	    for (x = columns - 1; x >= 0; x--)
		if (conss_buffer[--cxy_index] != ' ')
		    goto non_space_found;
    non_space_found:
	conss_buffer[0] = y + 1;
	conss_buffer[1] = 0;
	/*tty_cursormove(y + 1, 0);*/
    }
}

static void restore_console (void)
{
    if (!console_flag)
	return;
    if (console_flag == 2){
	/* Linux >= 1.1.67 */
	/* Restore screen contents and cursor position */
	conss_buffer [0] = 9;
	conss_buffer [1] = tty_name [len-1] - '0';
	ioctl (console_fd, TIOCLINUX, conss_buffer);
    }
    if (console_flag == 3){
	lseek (vcs_fd, 0, 0);
	write (vcs_fd, conss_buffer, buffer_size);
    }
    if (console_flag == 1){
	/* Clear screen */
	dwrite(console_fd, "\033[H\033[2J");
	/* Output saved screen contents */
	write(console_fd, conss_buffer + 2, rows * columns);
	/* Move the cursor to the previously selected position */
	tty_cursormove(conss_buffer[0], conss_buffer[1]);
    }
}

static void send_contents (void)
{
    unsigned char begin_line=0, end_line=0;
    int cxy_index, x, y;
    int lastline;
    char message;
    unsigned short bytes;
    int bytes_per_char;

    bytes_per_char = console_flag == 1 ? 1 : 2;

    /* Calculate the number of used lines */
    if (console_flag == 2 || console_flag == 1 || console_flag == 3){
	cxy_index = (2 + rows * columns) * bytes_per_char;
	for (y = rows - 1; y >= 0; y--)
	    for (x = columns - 1; x >= 0; x--){
		cxy_index -= bytes_per_char;
		if (conss_buffer[cxy_index] != ' ')
		    goto non_space_found;
	    }
    non_space_found:
	lastline = y + 1;
    } else
	return;

    /* Inform the invoker that we can handle this command */
    message = CONSOLE_CONTENTS;
    write_cmd (&message);

    /* Read the range of lines wanted */
    read (cmd_input, &begin_line, 1);
    read (cmd_input, &end_line, 1);
    if (begin_line > lastline)
	begin_line = lastline;
    if (end_line > lastline)
	end_line = lastline;

    /* Tell the invoker how many bytes it will be */
    bytes = (end_line - begin_line) * columns;
    write (cmd_output, &bytes, 2);

    /* Send the contents */
    for (cxy_index = (2 + begin_line * columns) * bytes_per_char;
	 cxy_index < (2 + end_line * columns) * bytes_per_char;
	 cxy_index += bytes_per_char)
		write_cmd (conss_buffer + cxy_index);

    /* All done */
}

int cons_saver_main(void)
{
    const char *error;
    unsigned char action = 0;

    if (Gargc != 2) {
	/* Wrong number of arguments */
	dwrite (2, "Usage: cons.saver <ttyname>\n");
	write_cmd ((const char *)&console_flag);        /* '\0' */
	return 3;
    }

    /* Lose the control terminal */
    setsid ();

    /* Check that the argument is a legal console */
    tty_name = Gargv [1];
    len = strlen(tty_name);
    console_fd = -1;
    console_flag = -1;
    error = detect_console ();

    if (error) {
	/* Not a console -> no need for privileges */
	setuid (getuid ());
	dwrite (2, error);
	console_flag = 0;
	if (console_fd >= 0)
	    close (console_fd);
    } else {
	/* Console was detected */
	if (console_flag != 3)
	    console_flag = 2; /* Default to Linux >= 1.1.67 */
	/* Allocate conss_buffer for screen image */
	tty_getsize ();
	buffer_size = 4 + 2 * columns * rows;
	conss_buffer = (char*) malloc (buffer_size);
    }

    /* If using /dev/vcs*, we don't need anymore the console fd */
    if (console_flag == 3)
	close (console_fd);

    /* Inform the invoker about the result of the tests */
    write_cmd ((const char *)&console_flag);

    /* Read commands from the invoker */
    while (console_flag && read (cmd_input, &action, 1)){
	/* Handle command */
	switch (action){
	case CONSOLE_DONE:
	    console_flag = 0;
	    continue; /* Break while loop instead of switch clause */
	case CONSOLE_SAVE:
	    save_console ();
	    break;
	case CONSOLE_RESTORE:
	    restore_console ();
	    break;
	case CONSOLE_CONTENTS:
	    send_contents ();
	    break;
	} /* switch (action) */

	/* Inform the invoker that command is handled */
	write_cmd ((const char *)&console_flag);
    } /* while (read ...) */

    return 0;
}
#endif /* APPLET_USAGE */
