/*
 * CRONTAB
 *
 * usually setuid root, -c option only works if getuid() == geteuid()
 *
 * Copyright 1994 Matthew Dillon (dillon@apollo.west.oic.com)
 * May be distributed under the GNU General Public License
 *
 * Vladimir Oleynik <dzo@simtreas.ru> (C) 2002 to be used in busybox
 *
 */
#ifdef APPLET_USAGE
#define crontab_trivial_usage \
	"[-c dir] {file|-}|[-u|-l|-e|-d user]"
#define crontab_full_usage \
	"\tfile <opts>  replace crontab from file\n" \
	"\t-    <opts>  replace crontab from stdin\n" \
	"\t-u user      specify user\n" \
	"\t-l [user]    list crontab for user\n" \
	"\t-e [user]    edit crontab for user\n" \
	"\t-d [user]    delete crontab for user\n" \
	"\t-c dir       specify crontab directory"
#else

#include "busybox.h"
#include <string.h>
#include <time.h>
#include <dirent.h>
#include <unistd.h>
#include <syslog.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/resource.h>

#ifndef CRONTABS
#define CRONTABS        "/var/spool/cron/crontabs"
#endif
#ifndef TMPDIR
#define TMPDIR          "/var/spool/cron"
#endif
#ifndef CRONUPDATE
#define CRONUPDATE      "cron.update"
#endif
#ifndef PATH_VI
#define PATH_VI         "/bin/vi"   /* location of vi       */
#endif

#include "bb_xopen.h"

extern struct BB_COMMON_BSS
{
	const char *CDir;
} bb_common_bss;
#define CDir bb_common_bss.CDir

static int
ChangeUser(const char *user, short dochdir)
{
    struct passwd *pas;

    /*
     * Obtain password entry and change privileges
     */

    pas = getpwnam(user);
    bb_setenv1("USER", pas->pw_name);
    bb_setenv1("HOME", pas->pw_dir);
    bb_setenv1("SHELL", DEFAULT_SHELL);

    /*
     * Change running state to the user in question
     */
    change_identity(pas);

    if (dochdir) {
	if (chdir(pas->pw_dir) < 0) {
	    bb_perror_msg("chdir failed: %s %s", user, pas->pw_dir);
	    xchdir(TMPDIR);
	}
    }
    return(pas->pw_uid);
}

static int
GetReplaceStream(const char *user, const char *file)
{
    int filedes[2];
    int pid;
    int fd;
    int n;
    char buf[1024];

    if (pipe(filedes) < 0) {
	bb_perror_msg("pipe");
	return(-1);
    }
    if ((pid = fork()) < 0) {
	bb_perror_msg("fork");
	return(-1);
    }
    if (pid > 0) {
	/*
	 * PARENT
	 */

	close(filedes[1]);
	if (read(filedes[0], buf, 1) != 1) {
	    close(filedes[0]);
	    filedes[0] = -1;
	}
	return(filedes[0]);
    }

    /*
     * CHILD
     */

    close(filedes[0]);

    if (ChangeUser(user, 0) < 0)
	exit(0);

    fd = bb_xopen(file, O_RDONLY|BB_OPEN_ERR);
    buf[0] = 0;
    write(filedes[1], buf, 1);
    while ((n = read(fd, buf, sizeof(buf))) > 0) {
	write(filedes[1], buf, n);
    }
    exit(0);
}

static void
EditFile(const char *user, const char *file)
{
    int pid;

    if ((pid = fork()) == 0) {
	/*
	 * CHILD - change user and run editor
	 */
	char *ptr;

	if (ChangeUser(user, 1) < 0)
	    exit(0);
	ptr = getenv("VISUAL");
	if (ptr == NULL)
	    ptr = getenv("EDITOR");
	if (ptr == NULL)
	    ptr = PATH_VI;

	ptr = xasprintf("%s %s", ptr, file);
	run_shell(NULL, ptr, NULL);
    }
    if (pid < 0) {
	/*
	 * PARENT - failure
	 */
	bb_perror_msg_and_die("fork");
    }
    wait(NULL);
}


int crontab_main(void)
{
    int ac;
    char **av;
    enum { NONE, EDIT, LIST, REPLACE, DELETE } option = NONE;
    const char *pwd_name;
    const char *repFile = NULL;
    int repFd = 0;
    int i;
    char caller[256];           /* user that ran program */
    int  UserId;

    UserId = getuid();
    pwd_name = bb_getpwuid(NULL, UserId, -1);

    safe_strncpy(caller, pwd_name, sizeof(caller));

    i = 1;
    ac = Gargc;
    av = Gargv;
    if (ac > 1) {
	if (av[1][0] == '-' && av[1][1] == 0) {
	    option = REPLACE;
	    ++i;
	} else if (av[1][0] != '-') {
	    option = REPLACE;
	    ++i;
	    repFile = av[1];
	}
    }

    for (; i < ac; ++i) {
	char *ptr = av[i];

	if (*ptr != '-')
	    break;
	ptr += 2;

	switch(ptr[-1]) {
	case 'l':
	    if (ptr[-1] == 'l')
		option = LIST;
	    /* fall through */
	case 'e':
	    if (ptr[-1] == 'e')
		option = EDIT;
	    /* fall through */
	case 'd':
	    if (ptr[-1] == 'd')
		option = DELETE;
	    /* fall through */
	case 'u':
	    if (i + 1 < ac && av[i+1][0] != '-') {
		++i;
		if (getuid() == geteuid()) {
		    UserId = bb_xgetpwnam(av[i]);
		} else {
		    bb_error_msg_and_die("only the superuser may specify a user");
		}
	    }
	    break;
	case 'c':
	    if (getuid() == geteuid()) {
		CDir = (*ptr) ? ptr : av[++i];
	    } else {
		bb_error_msg_and_die("-c option: superuser only");
	    }
	    break;
	default:
	    i = ac;
	    break;
	}
    }
    if (i != ac || option == NONE)
	bb_show_usage();

    /* Get password entry */
    pwd_name = bb_getpwuid(NULL, UserId, -1);

    /* If there is a replacement file, obtain a secure descriptor to it. */
    if (repFile) {
	repFd = GetReplaceStream(caller, repFile);
	if (repFd < 0)
	    bb_error_msg_and_die("unable to read replacement file");
    }

    /* Change directory to our crontab directory */
    if(CDir == NULL)
	CDir = CRONTABS;
    xchdir(CDir);

    /*
     * Handle options as appropriate
     */

    switch(option) {
    case LIST:
	{
	    FILE *fi;
	    char buf[1024];

	    if ((fi = fopen(pwd_name, "r"))) {
		while (fgets(buf, sizeof(buf), fi) != NULL)
		    fputs(buf, stdout);
		fclose(fi);
	    } else {
		bb_error_msg("no crontab for %s", pwd_name);
	    }
	}
	break;
    case EDIT:
	{
	    FILE *fi;
	    int fd;
	    int n;
	    char tmp[128];
	    char buf[1024];

	    snprintf(tmp, sizeof(tmp), TMPDIR "/crontab.%d", getpid());
	    fd = bb_xopen(tmp, O_RDWR|O_CREAT|O_TRUNC|O_EXCL|BB_OPEN_ERR, 0600);
	    chown(tmp, getuid(), getgid());
	    if ((fi = fopen(pwd_name, "r"))) {
		    while ((n = fread(buf, 1, sizeof(buf), fi)) > 0)
			write(fd, buf, n);
	    }
	    EditFile(caller, tmp);
	    remove(tmp);
	    lseek(fd, 0L, 0);
	    repFd = fd;
	}
	option = REPLACE;
	/* fall through */
    case REPLACE:
	{
	    char path[1024];
	    int fd;

	    snprintf(path, sizeof(path), "%s.new", pwd_name);
	    fd = bb_xopen(path, O_CREAT|O_TRUNC|O_APPEND|O_WRONLY|BB_OPEN_ERR, 0600);
	    bb_copyfd_eof(repFd, fd);
	    close(fd);
	    rename(path, pwd_name);
	    close(repFd);
	}
	break;
    case DELETE:
	remove(pwd_name);
	break;
    case NONE:
    default:
	break;
    }

    /*
     *  Bump notification file.  Handle window where crond picks file up
     *  before we can write our entry out.
     */

    if (option == REPLACE || option == DELETE) {
	FILE *fo;
	struct stat st;

	while ((fo = fopen(CRONUPDATE, "a"))) {
	    fprintf(fo, "%s\n", pwd_name);
	    fflush(fo);
	    if (fstat(fileno(fo), &st) != 0 || st.st_nlink != 0) {
		fclose(fo);
		break;
	    }
	    fclose(fo);
	    /* loop */
	}
	if (fo == NULL)
	    bb_error_msg("unable to append to %s/%s", CDir, CRONUPDATE);
    }
    return 0;
}
#endif /* APPLET_USAGE */
