/*
 * a Tiny RPN calculator
 *
 * rewroted by 2016 (C) Vladimir Oleynik <dzo@simtreas.ru>
 * use new my config_read() interface
 * make compatibility with mainstream dc utility:
 * - usage [-e expr](multiple) and [scripts list...]
 * - support "q" (quit) operator
 * - do not need extra space syntax support, for example: "1 2+pq"
 * - do not segmentation failed for print empty stack
 * - add "devide by zero" warning and push to stack 0
 *
 * Licensed under GPLv2 or later, see file License in this tarball for details.
 */

#ifdef APPLET_USAGE
#define dc_trivial_usage \
	"[-e expr] [scripts_list]"
#define dc_full_usage \
	"This is a Tiny RPN calculator that understands the\n" \
	"following operations: +, add, -, sub, *, mul, /, div, %, mod, " \
	"**, exp, and, or, not, eor.\n" \
	"\nOptions:\n" \
	"p - Prints the value on the top of the stack, without altering the stack\n" \
	"f - Prints the entire contents of the stack without altering anything\n" \
	"o - Pops the value off the top of the stack and uses it to set the output radix\n" \
	"q - quit script\n" \
	"    2, 8, 10 and 16 are supported"
#else

#include "busybox.h"
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include "config_read.h"

/* Tiny RPN calculator, because "expr" didn't give me bitwise operations. */

CALC_BSS struct BB_COMMON_BSS
{
	double stack[100];
	unsigned int pointer;
	unsigned char base;
} bb_common_bss;
#define stack bb_common_bss.stack
#define pointer bb_common_bss.pointer
#define base bb_common_bss.base

static void push(double a)
{
	if (pointer >= (sizeof(stack) / sizeof(*stack)))
		bb_error_msg_and_die("stack overflow");
	stack[pointer++] = a;
}

static double pop(void)
{
	if (pointer == 0)
		bb_error_msg_and_die("stack exmpty");
	return stack[--pointer];
}

static void add(void)
{
	push(pop() + pop());
}

static void sub(void)
{
	double subtrahend = pop();

	push(pop() - subtrahend);
}

static void mul(void)
{
	push(pop() * pop());
}

static void power(void)
{
	double topower = pop();

	push(pow(pop(), topower));
}

static void divide(void)
{
	double divisor = pop();
	double a = pop();

	if(divisor == 0.0) {
		bb_error_msg("divide by zero");
		push(0);
	} else {
		push(a / divisor);
	}
}

static void mod(void)
{
	unsigned int d = pop();

	push((unsigned int) pop() % d);
}

static void and(void)
{
	push((unsigned int) pop() & (unsigned int) pop());
}

static void or(void)
{
	push((unsigned int) pop() | (unsigned int) pop());
}

static void eor(void)
{
	push((unsigned int) pop() ^ (unsigned int) pop());
}

static void not(void)
{
	push(~(unsigned int) pop());
}

static void set_output_base(void)
{
	static const char bases[] = { 2, 8, 10, 16, 0 };
	unsigned b = (unsigned)pop();

	base = *strchrnul(bases, b);
	if (base == 0)
		bb_error_msg("error, base %u is not supported", b);
}

static void print_base(double print)
{
	ubb64_t x = (ubb64_t)print;

	switch (base) {
	case 0:
	case 10:
		printf("%g\n", print);
		break;
	case 16:
		printf("%" P_BB64T "x\n", x);
		break;
	case 8:
		printf("%" P_BB64T "o\n", x);
		break;
	default: /* base 2 */
		{
			ubb64_t i = 0x1000000000000000LL;
			do {
				if (x & i) break;
				i >>= 1;
			} while (i > 1);
			do {
				putchar('1' - !(x & i));
				i >>= 1;
			} while (i);
			putchar('\n');
		}
	}
}

static void print_stack_no_pop(void)
{
	unsigned int i=pointer;
	while (i)
		print_base(stack[--i]);
}

static void print_no_pop(void)
{
	print_base(pop());      /* test stack empty */
	pointer++;
}

struct op {
	const char *name;
	void (*function) (void);
};

static const struct op operators[] = {
	{"+",   add},
	{"add", add},
	{"-",   sub},
	{"sub", sub},
	{"*",   mul},
	{"mul", mul},
	{"/",   divide},
	{"div", divide},
	{"**",  power},
	{"exp", power},
	{"pow", power},
	{"%",   mod},
	{"mod", mod},
	{"and", and},
	{"or",  or},
	{"not", not},
	{"eor", eor},
	{"xor", eor},
	{"p",   print_no_pop},
	{"f",   print_stack_no_pop},
	{"o",   set_output_base},
	{"q",   NULL},
	{NULL,  NULL}
};

static int stack_machine(const char *argument)
{
	char *endPointer;
	double d;

	for(;;) {
		d = strtod(argument, &endPointer);
		if (endPointer == argument) {
			const struct op *o = operators;

			while (o->name != 0) {
				size_t l = strlen(o->name);
				if (strncmp(o->name, argument, l) == 0) {
					if(o->function == NULL)
						return 0;
					(*(o->function)) ();
					endPointer = (char *)argument + l;
					break;
				}
				o++;
			}
			if(o->name == NULL)
				bb_error_msg_and_die("'%c' unimplemented", *argument);
		} else {
			push(d);
		}
		if(*endPointer == '\0')
			return 1;
		argument = endPointer;
	}
}

static void parse(const char *line_or_file, int mode)
{
	config_read_t *crt = config_read_prepare(line_or_file, mode);
	int i;

	while( (crt = config_read(crt, "\0~\0\0")) != NULL) {
		for(i = 0; crt->tokens[i] != NULL; i++)
			if(!stack_machine(crt->tokens[i]))
				return;
	}
}

int dc_main(void)
{
	llist_t *e = NULL;

	/* multiple -e expr */
	bb_xgetopt_ulflags("e:\0e::", &e);

	if(e == NULL) {
		if(*Gargv == NULL) {
			/* stdin for expession if -e do not use */
			Gargv--;
			*Gargv = "-";
		}
	} else {
		do {
			/* expession from multiple -e exp usage */
			parse(e->data, 'S');
			e = e->link;
		} while(e != NULL);
	}

	while(*Gargv != NULL) {
		/* expession from files */
		parse(*Gargv, 'w');
		Gargv++;
	}
	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
