/*
 * microdip     A program for handling dialup IP connecions.
 *              This program handles the connections needed for PPP dialup
 *
 * Usage:       dip [-v] -k [-l tty]
 *              dip -t [-v] [-l tty] [-s speed]
 *              dip [-v] [-m mtu] [-l tty] [-s speed] [telno | script]
 *
 * Based version: 3.3.7p-uri      12/25/96
 *
 * Author:      Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
 *              Copyright 1988-1993 MicroWalt Corporation
 *
 * Modified:    Uri Blumenthal <uri@watson.ibm.com>
 *              Copyright 1994
 *
 *              Paul Cadach, <paul@paul.east.alma-ata.su>
 *              (C) 1994
 *
 *              Pauline Middelink <middelin@calvin.iaf.nl>
 *              - Changed various paths for FSSTND
 *              - Allow exit with optional exitcode.
 *              - Removed the strange parsing of argv[0] which overwrote
 *                part of the string.
 *              - various source changes
 *              - Well, and since everybody is doing it: (C) 1994.
 *
 *              Edward Coltens (statistics, IP Acct, etc.)
 *
 *              This program is free software; you can redistribute it
 *              and/or  modify it under  the terms of  the GNU General
 *              Public  License as  published  by  the  Free  Software
 *              Foundation;  either  version 2 of the License, or  (at
 *              your option) any later version.
 *
 *              Vladimir Oleynik <dzo@simtreas.ru> (C) 2003,2005,2008
 *              - busybox port
 *              - removed support SLIP and dialIN
 *              - support [-l line] and [-s speed] options always
 *              - support make specefied speed before open port
 *              - auto open port for all requires command
 *              - perfect and simplify call pppd
 *              - paste to one file
 *              - added user_pass command for set this arguments for pppd
 *              - modify to modern GNU source
 *
 */
#if defined(APPLET_USAGE) || defined(STANDALONE)
#define dip_trivial_usage \
	"[-v] [-k|t] [-l tty] [-m mtu] [telno | script]"
#define dip_full_usage \
	"A program for handling dialup IP connecions.\n" \
	"Options:\n" \
	"\t-v\tVerbose mode\n" \
	"\t-k\tKill session for specified line (required)\n" \
	"\t-t\tIneractive terminal mode without script"
#endif

#if !defined(APPLET_USAGE) || defined(STANDALONE)

#ifdef STANDALONE
#define _GNU_SOURCE
#endif

#define VERSION "3.3.7p-micro (02 Jul 03)"

#include "busybox.h"
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <fcntl.h>
#include <dirent.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <resolv.h>

#include <sys/wait.h>

#include "speed_table.h"
#include "bb_inet.h"
#include "bb_sig.h"

#ifdef CONFIG_FEATURE_COMMAND_EDITING
#include "cmdedit.h"
#endif

/* DIP definitions. */
#define DIP_SFX         ".dip"

/* PPP initial protocol configuration. */
#define DEF_MTU         1500
#define DEF_SPEED       115200

/* Pathnames of base-level NET programs. */
#define _PATH_BIN_PPP           "/usr/sbin/pppd"

#define _PATH_LOCKD           "/var/lock"             /* lock files   */


/* Modem chat responses list */
struct response {
  struct response *next;
  char            *string;
  int              code;
};

#define TTY_BUFSIZE     1024            /* size of one serial buffer    */

/* Here are some HAYES modem commands. */
#define AT_ATTN         "+++"                   /* "attention" string   */
#define AT_RSET         "ATZ"                   /* reset the modem      */
#define AT_INIT         "ATE0 Q0 V1 X4"         /* setup for dialing    */
#define AT_DIAL         "ATD"                   /* start a dial         */
#define AT_DIAL_TONE    "ATDT"                  /* start a dial (+tone) */
#define AT_HANG         "ATH"                   /* hang up the modem    */
#define AT_EOL          "\r"                    /* AT "CRLF" mark       */


#define ARGVSIZE 32

struct variable {
  struct variable *next;
  int              type;
  char            *name;
  union V {
    char            *str_value;
    int              value;
  } vv;
};

struct keyword {
  char            *key;
  int              code;
};

struct label {
  struct label    *next;
  char            *name;
  off_t            offset;
};

struct commands {
  char  valid_n_args;
  const char  *name;
  int   (*func)(int, char **);
  const char *usage;
};


static const char modem[] = "HAYES";

#ifdef STANDALONE
#define CALC_BSS
#endif

CALC_BSS struct BB_COMMON_BSS
{
	const char *mdm_init_s;          /* current INIT string  */
	struct termios   tty_saved;      /* saved TTY device state       */
	struct termios   tty_current;    /* current TTY device state     */
	char             *in_buff,       /* line input/output buffers    */
			*out_buff,
			*in_ptr,
			*out_ptr;
	int              in_size,        /* buffer sizes and counters    */
			out_size,
			in_cnt,
			out_cnt,
			tty_mode;       /* TTY mode (login or not)      */
	uid_t uid, euid;
	volatile int timeout;           /* script "wait" timed out      */
	int      errlevel;              /* script command return code   */
	FILE     *scriptfp;             /* input script pointer */
	char     *var_port;             /* terminal line used           */
	int      port_opened;
	unsigned long var_speed;        /* terminal line speed          */
	int      var_echo;              /* enable line data echoing     */
	int      var_dcount;            /* number of times we've dialed */
	int      opt_v;                 /* debug flag                   */
	struct variable *vartable;
	struct label    *labels;
	int              alllabels;     /* all labels collected */
	struct response *chat;          /* List of modem's answers      */
	char saved_lock;
	char *tl_saved_path;
	int onexit_argc;
	char **onexit_argv; /* Warning! If not NULL, do_onexit will try to deallocate it!! */
	int forked;
	int catched;
	volatile int do_exit;
	void (*oldalrmsig)(int);
	/* Misc protocol info. */
	int mtu;                        /* MTU to use for conn. */
	/* Connection info. */
	char local[128];                /* local host name      */
	struct in_addr loc_ip;          /* mapped IP address    */
	char remote[128];               /* remote host name     */
	struct in_addr rmt_ip;          /* mapped IP address    */
	struct in_addr rmt_rt;          /* mapped IP address    */
	int rtdefault;                  /* are we the dflt gw?  */
	int proxyarp;                   /* set proxy arp?       */
	char user_pass[128];            /* ppp user and password */
	int tty_fd;                     /* TTY file descriptor */
} bb_common_bss;
#define mdm_init_s bb_common_bss.mdm_init_s
#define tty_saved bb_common_bss.tty_saved
#define tty_current bb_common_bss.tty_current
#define in_buff bb_common_bss.in_buff
#define out_buff bb_common_bss.out_buff
#define in_ptr bb_common_bss.in_ptr
#define in_size bb_common_bss.in_size
#define out_size bb_common_bss.out_size
#define in_cnt bb_common_bss.in_cnt
#define out_cnt bb_common_bss.out_cnt
#define uid bb_common_bss.uid
#define out_ptr bb_common_bss.out_ptr
#define tty_mode bb_common_bss.tty_mode
#define euid bb_common_bss.euid
#define timeout bb_common_bss.timeout
#define errlevel bb_common_bss.errlevel
#define scriptfp bb_common_bss.scriptfp
#define var_port bb_common_bss.var_port
#define port_opened bb_common_bss.port_opened
#define var_speed bb_common_bss.var_speed
#define var_echo bb_common_bss.var_echo
#define var_dcount bb_common_bss.var_dcount
#define opt_v bb_common_bss.opt_v
#define vartable bb_common_bss.vartable
#define labels bb_common_bss.labels
#define alllabels bb_common_bss.alllabels
#define chat bb_common_bss.chat
#define saved_lock bb_common_bss.saved_lock
#define tl_saved_path bb_common_bss.tl_saved_path
#define onexit_argc bb_common_bss.onexit_argc
#define onexit_argv bb_common_bss.onexit_argv
#define forked bb_common_bss.forked
#define catched bb_common_bss.catched
#define do_exit bb_common_bss.do_exit
#define oldalrmsig bb_common_bss.oldalrmsig
#define mtu bb_common_bss.mtu
#define local bb_common_bss.local
#define loc_ip bb_common_bss.loc_ip
#define remote bb_common_bss.remote
#define rmt_ip bb_common_bss.rmt_ip
#define rmt_rt bb_common_bss.rmt_rt
#define rtdefault bb_common_bss.rtdefault
#define proxyarp bb_common_bss.proxyarp
#define user_pass bb_common_bss.user_pass
#define tty_fd bb_common_bss.tty_fd

static void TimeOut(int sig)
{
  (void) sig;
  timeout = 1;
}

/* Set the state of a terminal. */
static int
tty_set_state(struct termios *tty)
{
  if (tty_fd >= 0) {
	if (xioctl(tty_fd, TCSETS, tty, "tty: set_state") < 0)
	    return -errno;
	return 0;
  }
  return -1;
}

static int
tty_already_locked(char *nam) /* nam - complete path to lock file */
{
  int  i = 0, pid = 0;
  FILE *fd = (FILE *)0;

  /* Does the lock file on our device exist? */
  if ((fd = fopen(nam, "r")) == (FILE *)0)
    return 0; /* No, return perm to continue */

  /* Yes, the lock is there.  Now let's make sure */
  /* at least there's no active process that owns */
  /* that lock.                                   */
  i = fscanf(fd, "%d", &pid);
  (void) fclose(fd);

  if (i != 1) /* Lock file format's wrong! Kill't */
    return 0;

  /* We got the pid, check if the process's alive */
  if (kill(pid, 0) == 0)      /* it found process */
      return 1;           /* Yup, it's running... */

  /* Dead, we can proceed locking this device...  */
  return 0;
}


static int put_pid_to_lock_file(const char *saved_path, const char *smode)
{
    FILE *f;

    if ((f = fopen(saved_path, "w")) == NULL) {
      bb_perror_msg("tty_lock(%s): %s", saved_path, smode);
      return -1;
    }

    fprintf(f, "%10d\n", (int)getpid());
    (void)fclose(f);
    return 0;
}

/* Lock or unlock a terminal line. */
static int tty_lock(const char *path, int mode)
{
  if (mode == 1) {      /* lock */

    tl_saved_path = xasprintf("%s/LCK..%s", _PATH_LOCKD, path);

    if (tty_already_locked(tl_saved_path) == 1) {
      bb_perror_msg("attempt to use already locked tty %s", tl_saved_path);
      return -1;
    }
    if(put_pid_to_lock_file(tl_saved_path, ""))
	return -1;

    saved_lock++;
  } else if (mode == 2) {
      /* re-acquire a lock after a fork() */

      if (saved_lock) {
	if(put_pid_to_lock_file(tl_saved_path, "reacquire: "))
	    return -1;
	(void) chmod(tl_saved_path, 0444);
	(void) chown(tl_saved_path, getuid(), getgid());
      }
  } else {
      /* unlock */

      if (saved_lock) {
	saved_lock = 0;
	if (unlink(tl_saved_path) < 0) {
	  bb_perror_msg("tty: unlock: (%s)", tl_saved_path);
	  return -1;
	}
      }
  }

  return 0;
}

/* Return the TTY link's file descriptor. */
static inline int tty_askfd(void)
{
  return tty_fd;
}

static int prephangup(void)
{
  int fd = tty_askfd();

  ioctl(fd, TIOCMBIC, TIOCM_DTR);
  xusleep(500000);
  ioctl(fd, TIOCMBIS, TIOCM_DTR);
  xusleep(250000);
  return 0;
}

/* Set the line speed of a terminal line. */
static int tty_set_speed(struct termios *tty, unsigned long speed)
{
  int code;

#ifdef STANDALONE
  if (opt_v) printf("tty: set_speed: %lu\n", speed);
#endif
  code = bb_value_to_baud(speed);
  if (code < 0) return code;
  tty->c_cflag &= ~CBAUD;
  tty->c_cflag |= code;

  return 0;
}

/* Hangup the line. */
static int hanguptty(void)
{
  struct termios tty;

  tty = tty_current;
  (void) tty_set_speed(&tty, 0);
  if (tty_set_state(&tty) < 0) {
    bb_perror_msg("tty: hangup(DROP)");
    return -errno;
  }

  (void) sleep(1);

  if (tty_set_state(&tty_current) < 0) {
    bb_perror_msg("tty: hangup(RAISE)");
    return -errno;
  }
  return 0;
}

/* Write one character (byte) to the TTY link. */
static int tty_putc(int c)
{
  int s;

  if ((out_cnt == out_size) || (c == -1)) {
    s = write(tty_fd, out_buff, out_cnt);
    out_cnt = 0;
    out_ptr = out_buff;
    if (s < 0) {
#ifdef STANDALONE
      if (opt_v) {
	bb_error_msg("Probably line disconnected!");
	bb_perror_msg("tty_puts: failed to write to tty");
      }
#endif
      return -1;
    }
  }

  if (c != -1) {
    *out_ptr = (char) c;
    out_ptr++;
    out_cnt++;
  }

  return 0;
}

/* Output a string of characters to the TTY link. */
static int tty_puts(unsigned const char *s)
{
  while(*s != '\0')
    if (tty_putc((int) *s++) < 0)
	return -1;
  if (tty_putc(-1) < 0)         /* flush */
	return -1;
  return 0;
}

/* Flush input on the terminal. */
static int tty_flush(void)
{
  int blkmode;
  char buf;

  if (tty_fd < 0)
	return -1;

  blkmode = fcntl(tty_fd, F_GETFL, NULL);
  fcntl(tty_fd, F_SETFL, (O_NDELAY | blkmode));
  while (read(tty_fd, &buf, 1) > 0)
		;
  fcntl(tty_fd, F_SETFL, blkmode);
  return 0;
}

/* Read one character (byte) from the TTY link. */
static int tty_getc(void)
{
  int s;

  if (tty_fd < 0)
	return -1;

  if (in_cnt <= 0) {
	s = read(tty_fd, in_buff, in_size);
	in_cnt = s;
	in_ptr = in_buff;
  }

  if (in_cnt < 0) {
#ifdef STANDALONE
    if (opt_v) {
	bb_perror_msg("tty_getc: I/O error");
	bb_error_msg("Probably line disconnected!");
    }
#endif
    return -1;
  }

  s = (int) *in_ptr;
  s &= 0xFF;
  in_ptr++;
  in_cnt--;
  return s;
}

/* Chat with modem, wait for one of specified responses. */
static int mdm_chatseq(const char *m_send, struct response *expect, int out)
{
  struct response *resp;
  int  minlen;
  int  maxlen;
  char buff[128];
  char *s;
  int  len;
  void (*oldsig)(int);

  /* If you want to send nothing - the modem doesn't mind */
  if ((m_send == (char *)0) || (*m_send == '\0'))
    return 0;

#if 0
  /*
   * _FIXME: this is ugly!!! Some modem's faults when its receiving
   * new command before finish of sending reply for previous command.
   */
  struct timeval   delay;

  delay.tv_sec = 0;
  delay.tv_usec = 500000;
  (void) select(0, (fd_set *)NULL, (fd_set *)NULL, (fd_set *)NULL, &delay);
#else
  (void) sleep(1);
#endif

  tty_flush();
  if((tty_puts((unsigned const char *)m_send) < 0) ||
		(tty_puts((unsigned const char *)AT_EOL) < 0))
    return -1;          /* Modem fail */

  if(expect == NULL)
    return 0;

  /* Calculating minimal and maximal length of response string */
  resp = expect;
  minlen = 9999;
  maxlen = 0;
  while(resp != NULL) {
    len = strlen(resp->string);
    if(len < minlen)
      minlen = len;
    if(len > maxlen)
      maxlen = len;
    resp = resp->next;
  }

  timeout = 0;
  oldsig = bb_signal((1 << SIGALRM), TimeOut);
  (void) alarm(out);
  s = buff;
  while(s - buff < minlen) {
    /* Terminal detached from modem */
    if(((len = tty_getc()) == -1) || timeout) {
      (void) alarm(0);
      bb_signal((1 << SIGALRM), oldsig);
      return timeout ? 3 : -1;
    }
    if(len == -2)                   /* No symbols in buffer */
      continue;
    *s++ = (char) len;
  }
  for(;;) {
    resp = expect;
    while(resp != NULL) {
      len = strlen(resp->string);

      /* Looking for wanted string in tail of buffer */
      if((s - buff >= len) && !strncmp(s - len, resp->string, len)) {
	bb_signal((1 << SIGALRM), oldsig);
	(void) alarm(0);
#ifdef STANDALONE
	if (opt_v)
		printf("Got response \"[%d] %s\"\n", resp->code, resp->string);
#endif
	return resp->code;     /* We got that */
      }

      resp = resp->next;
    }
    if((s - buff) >= maxlen) {
      memcpy(buff, &buff[1], maxlen);   /* Scroll the string in buffer */
      --s;
    }
    len = -2;
    while(len == -2) {
      if(((len = tty_getc()) == -1) || timeout)
	break;
    }
    if((len == -1) || timeout)
      break;
    *s++ = (char) len;
  }
  bb_signal((1 << SIGALRM), oldsig);
  (void) alarm(0);
  return timeout ? 3 : -1;
}


/* Close down a terminal line. */
static int tty_close(void)
{
  if (tty_fd < 0)
     return -1;

  if (tty_set_state(&tty_saved) < 0) {
	bb_perror_msg("tty_close: restore");
	(void) close(tty_fd);
	tty_fd = -1;
	sleep (1);
	(void) tty_lock(NULL, 0);
	return -errno;
  }

  (void) prephangup();
  (void) hanguptty();
  /* Make the modem listen to us. */
  (void) sleep(2);
  tty_puts((unsigned const char *)AT_ATTN);
  (void) sleep(2);
  /* Hang up the modem. */
  (void) mdm_chatseq(AT_HANG, NULL, 5);
  (void) close(tty_fd);
  tty_fd = -1;
  (void) tty_lock(NULL, 0);

  return 0;
}

static void sig_exit(int sig)
{
  tty_close();
  exit(sig);
}


/* Convert a C-style backslash sequence to ASCII. */
static char cvt_char(char c)
{
  if ( !c || c == '0') return 0x00;

  switch(c) {
	case 'a':
		return '\a';
	case 'b':
		return '\b';
	case 'e':
		return 0x1b;
	case 'f':
		return '\f';
	case 'n':
		return '\n';
	case 'r':
		return '\r';
	case 's':
		return ' ';
	case 't':
		return '\t';
	case 'v':
		return '\v';
	case '\\':
		return '\\';
	case '\'':
		return '\'';
	case '"':
		return '\"';
	case '?':
		return '\?';
	case '$':
		return '$';
	default:
		return '?';
  }
  /* NOTREACHED */
  return '?';
}


static int add_label(char *label, off_t offset)
{
  struct label *l;

  for(l = labels; l != NULL; l = l->next) {
	if(strcmp(l->name, label) == 0)        /* Duplicated labels */
		return -1;
  }
  l = xmalloc(sizeof(struct label));
  l->name = xstrdup(label);
  l->offset = offset;
  l->next = labels;     /* add reverse */
  labels = l;
  return 0;
}

static off_t lookup_label(char *label)
{
  struct label *l;

  for(l = labels; l != NULL; l = l->next) {
    if(strcmp(l->name, label) == 0)
      return l->offset;
  }
  return (off_t) -1;
}

static off_t last_label(void)
{
  struct label *l = labels;

  if(l != NULL) {       /* labels added reverse */
    return l->offset;
  }
  return (off_t) -1;
}

static struct variable * lookup(const char *varname)
{
  struct variable *var;

  ++varname;                            /* Skip leading '$' */
  for(var = vartable; var != (struct variable *)NULL; var = var->next)
    if(strcasecmp(var->name, varname) == 0)
	break;
  return var;
}

static struct variable * new_var(const char *varname)
{
  struct variable *var;

  var = lookup(varname);
  if(var)
    return var;
  ++varname;         /* Skip leading '$' */
  var = (struct variable *)xmalloc(sizeof(struct variable));
  var->name = xstrdup(varname);
  var->next = vartable;
  vartable = var;
  var->type = 'i';  /* jus in case, fill with integer */
  return var;
}

static int is_special(const char *name)
{
  name++;
  if(!strcmp(name, "locip") || !strcmp(name, "local") ||
     !strcmp(name, "rmtip") || !strcmp(name, "remote") ||
     !strcmp(name, "mtu") || !strcmp(name, "modem") ||
     !strcmp(name, "port") || !strcmp(name, "speed")) {
	bb_error_msg("Cannot use special variable %s in this place!", name);
	return 1;
  }
  return 0;
}


/* validates a decimal number as value of variable */
static int is_a_num(const char *n)
{
  if (*n == '0' && n[1]) return 0;      /* no leading zeros on decimal numbers */
  while (*n) { /* check all the characters on the value */
    if ((*n < '0') || (*n > '9'))
      return 0; /* not a decimal number */
    n++;
  }
  return 1; /* yes, a digit */
}


static inline void get_value(int *var, const char *name)
{
  struct variable *v;

  *var = -1;  /* Just a precaution... */

  if(*name != '$') {
    if (is_a_num(name))
      *var = atoi(name);
    else {
      bb_error_msg("In this place must use integer variables!");
      return;
    }
  } else {
    if(is_special(name))  {
      return;
    }
    if((v = lookup(name)) == (struct variable *)NULL) {
      bb_error_msg("Variable %s not defined", name);
      return;
    }
    *var = v->vv.value;
  }
}


/* Split the input string into multiple fields. */
static int getargs(char *string, char *arguments[])
{
  char *sp;
  int argc;
  int i;

  sp = string; i = 0;
  arguments[i] = sp;
  while (*sp && i < ARGVSIZE) {
    /* skip leading whitespaces */
    while (*sp && (*sp == ' ' || *sp == '\t')) sp++;
    /* break at " #..." comment start */
    if (*sp == '#') {
      *sp = '\0';                         /* might be end of arg */
      break;
    }
    arguments[i++] = sp;                   /* found begin of arg */
    /* find end of arg */
    while (*sp && *sp != ' ' && *sp != '\t') sp++;
    if (*sp != '\0')
      *sp++ = '\0';             /* mark end of arg */
  }
  argc = i;
  while (i < ARGVSIZE) arguments[i++] = (char *)NULL;
  return argc;
}


static int do_beep(int argc, char *argv[])
{
  int  i,times;

  if(argc==2)  times = atoi(argv[1]);
  else times=1;

  if (times > 30) times = 30; /* add some sanity */

  for(i=0;i<times;i++) {
    fprintf(stderr,"\aBEEP\n");
    if(i+1 < times)
      sleep(1);
  }
  return 0;
}


/* Enable/Disable echoing of data from terminal line. */
static int do_echo(int argc, char *argv[])
{
  var_echo = strcasecmp(argv[1], "off");

#ifdef STANDALONE
  if (opt_v) printf("Display modem output: %sabled\n",
				(var_echo == 0) ? "dis" : "en");
#endif
  return 0;
}


/* Go to some label in the script. */
static int do_goto(int argc, char *argv[])
{
  char buff[1024];
  off_t oldpos, pos;
  char *label, *xp;
  char *sp;

  if (scriptfp == stdin) {
	bb_error_msg("GOTO not possible in TEST mode!");
	return -1;
  }

  if((pos = lookup_label(label = argv[1])) == (off_t)-1) {
    if(alllabels) {
BADLABEL:
      bb_error_msg("Label `%s' not found - skipped", label);
      return -1;
    }
    oldpos = ftell(scriptfp);
    if(((pos = last_label()) != (off_t)-1) && (pos > oldpos))
      fseek(scriptfp, pos, SEEK_SET);           /* Continue search */
    do {
      if(fgets(buff, 1024, scriptfp) == (char *)NULL) {
	alllabels = 1;
	break;
      }
      strchr_set0(buff, '\n');          /* mark EOL */
      sp = buff;                        /* start search */
      while(*sp == ' ' || *sp == '\t')  /* skip whitespace */
	sp++;
      if(*sp == '#' || *sp == '\0')     /* skip empty or comment line */
	continue;
      /* A label must be the first word in the line; otherwise we find */
      /* labels in lines like ``wait login: 30''. */
      /* Labels consist of [-a-zA-Z0-9_] only. */
      for( xp=sp; isalnum(*xp) || *xp == '-' || *xp == '_'; xp++ )
	;
      if( *xp != ':' )                          /* label found? */
	continue;
      *xp = '\0';
      if(add_label(sp, pos = ftell(scriptfp)) != 0)
	bb_error_msg("Duplicated label `%s' - ignored", sp);
#ifdef STANDALONE
      if(opt_v)
	bb_error_msg("Got label `%s' at pos %08X\n", sp, (unsigned int) pos);
#endif
      if(!strcmp(label, sp))
	return 0;
    } while(1);
    (void) fseek(scriptfp, oldpos, SEEK_SET);
    goto BADLABEL;
  }
  (void) fseek(scriptfp, pos, SEEK_SET);
  return 0;
}


/* Check some error (result) code. */
static int do_if(int argc, char *argv[])
{
  char *cmd[3];
  char opcode;
  long val, var;
  int ret;
  struct variable *v;

  if (! strcmp(argv[2], "==")) opcode = '=';
    else if (! strcmp(argv[2], "!=")) opcode = '@';
    else if (! strcmp(argv[2], "<")) opcode = '<';
    else if (! strcmp(argv[2], ">")) opcode = '>';
    else if (! strcmp(argv[2], "<=")) opcode = 'L';
    else if (! strcmp(argv[2], ">=")) opcode = 'G';
    else {
	bb_error_msg("Syntax error: \"%s\" is not an opcode!", argv[2]);
	return -1;
  }
  val = (long) atol(argv[3]);

  /* Now - do we really need this check? Here? */
  if (*argv[1] != '$') {
    bb_error_msg("Variable must lead with '$'!");
    return -1;
  }

  if (! strcmp(&argv[1][1], "errlvl")) var = (long) errlevel;
    else if (! strcmp(&argv[1][1], "locip")) var = (long) loc_ip.s_addr;
    else if (! strcmp(&argv[1][1], "rmtip")) var = (long) rmt_ip.s_addr;
    else if ((v = lookup(argv[1])) != (struct variable *)NULL)
      var = (long) v->vv.value;
    else {
	bb_error_msg("Variable %s not defined", argv[1]);
	return -1;
  }

  ret = -1;
  switch(opcode) {
	case '=':       /* EQUAL */
		ret = (var == val);
		break;

	case '@':       /* NOT EQUAL */
		ret = (var != val);
		break;

	case '<':       /* LESS */
		ret = (var < val);
		break;

	case 'L':       /* LESS-EQ */
		ret = (var <= val);
		break;

	case '>':       /* GREATER */
		ret = (var > val);
		break;

	case 'G':       /* GREATER-EQ */
		ret = (var >= val);
		break;
  }

  if (strcasecmp(argv[4], "goto")) {
	bb_error_msg("Warning: keyword not \"goto\" !");
	argv[4] = "goto";
  }

  if (ret != 0) {
	cmd[1] = argv[4];
	cmd[1] = argv[5];
	cmd[2] = (char *)NULL;
	return do_goto(2, cmd);
  }
  return errlevel;
}

/* Set the shell command to be executed on program exit */

/* As things have evolved dramatically since the last time
 * I nosed around this code, I will do it in another fashion.
 * I'll just store the argv array and the argc count and we'll
 * pass'em over to do_shell upon the call of cleanup()
 */

/* I don't know which position will you prefer to put all these
 * functions, so I let them be together, so you find easier to put
 * them whereever you find they look prettier :)
 * The externs on main.c of onexit_argc, onexit_argv and
 * arg_clean() should be in main.c for use in cleanup()
 */


/* Deallocate the given double array. argc is the number of sub-arrays */

static void arg_clean (int argc, char **argv)
{
  int cnt;                          /* counter */
  if (argv != NULL && argc != 0)    /* previous data allocated? */
  {
    for (cnt = 0; cnt < argc; cnt++)    /* Ok, free old arg vectors */
      if (onexit_argv[cnt] != NULL)     /* if not NULL!! :*) */
	free (onexit_argv[cnt]);
  }
}

/* Allocate an string and print into there the argv's with spaces between
 * them. String's size is adjusted to the correct size!
 * free() the string after use!
 * Returns NULL on error
 * Remember! We have to skip argv[0], containing the function name
 */

static char *argv_sprintf (int argc, char **argv)
{
  int  cnt;               /* counter */
  char *buf;              /* allocated string */
  int  size;              /* size of string */

  if (argv == NULL || argc == 0)  /* parameter error? */
  {
    buf = xstrdup("");      /* allocate one char */
    return buf;            /* return it */
  }

  /* Add up all the sizes to get the total string length
   * (mind one more for the space)
   */
  for (size = 0, cnt = 1; cnt < argc; cnt++)
    size += strlen (argv[cnt]) + 1;
  size++;                 /* this one is for the trailing NULL */

  /* allocate the string */
  buf = xmalloc (size);

  /* Concatenate all the strings :) */

  *buf = 0;
  for (cnt = 1; cnt < argc; cnt++)
  {
     strcat (buf, argv[cnt]);   /* concatenate string #cnt */
     if (cnt != argc-1)         /* If this isn't the last one ... */
       strcat (buf, " ");       /* ... print an space to spare */
  }
  return buf;                  /* done ... return the pointer */
}

static int do_onexit (int argc, char **argv)
{
  int  cnt;      /* counter */
  char **nargv;  /* temporary array */

  arg_clean (onexit_argc, onexit_argv); /* remove old commands */
  onexit_argc = 0;                      /* and clean to ... */
  onexit_argv = NULL;                   /* ... avoid confusion */

  /* Ok, right now, the only thing we should do is to copy the argc and argv's
   * given into our array.
   */

  /* allocate the array of pointers */
  nargv = xmalloc (argc * sizeof(char *));

  for (cnt = 0; cnt < argc; cnt++)  /* clean (for arg_clean() to work ok) */
     nargv[cnt] = NULL;

  for (cnt = 0; cnt < argc; cnt++)  /* allocate and copy the argv[]'s */
  {
    nargv[cnt] = xstrdup(cnt == 0 ? "shell" : argv[cnt]);   /* allocate */
  }
  onexit_argv = nargv;    /* finished, copy the data */
  onexit_argc = argc;
  return 0;               /* and exit succesfully */
}

/* Print the contents of some variable. */
static int do_print(int argc, char *argv[])
{
  char  *sp      = (char *)0;
  int    i       = 0;
  int    nonewln = 0;
  struct variable *var;
  char   *onexit_str;

  if (argc == 1) {
	printf("\n");
	return 0;
  }
  i = 0;
  if ((argv[1][0]=='-') &&
      (argv[1][1]=='n') &&
      (argv[1][2]==0))
    {
      i++;
      nonewln=1;
    }

  while (argv[++i] != (char *)NULL) {
    sp = argv[i];
    if (i != (1+nonewln)) printf(" ");
    if (*sp == '$') {
      if (! strcmp(++sp, "errlvl")) printf("%d", errlevel);
      else if (! strcmp(sp, "locip")) printf("%s", inet_ntoa(loc_ip));
      else if (! strcmp(sp, "rmtip")) printf("%s", inet_ntoa(rmt_ip));
      else if (! strcmp(sp, "local")) printf("%s", local);
      else if (! strcmp(sp, "remote")) printf("%s", remote);
      else if (! strcmp(sp, "mtu")) printf("%d", mtu);
      else if (! strcmp(sp, "modem")) printf("%s", modem);
      else if (! strcmp(sp, "user_pass")) printf("%s", user_pass);
      else if (! strcmp(sp, "onexit"))
      {
	onexit_str = argv_sprintf (onexit_argc, onexit_argv);
	if (onexit_str)         /* If no error, we print it; otherwise, */
	{                       /* arg_sprintf() has already complained */
	  printf ("%s", onexit_str);
	  free (onexit_str);
	}
      }
      else if (! strcmp(sp, "port")) printf("%s", var_port);
      else if (! strcmp(sp, "speed")) printf("%lu", var_speed);
      else if (! strcmp(sp, "count")) printf("%d", var_dcount);
      else if ((var = lookup(sp - 1)) == (struct variable *)NULL)
	bb_error_msg("Variable %s not defined", sp);
      else {
	if (var->type == 'i')
	  printf("%d", var->vv.value);
	else
	  printf("%s", var->vv.str_value);
      }
    } else printf("%s", sp);
  }
  if (!nonewln)
    printf("\n");
  return 0;
}

/* Fetch the state of a terminal. */
static int tty_get_state(struct termios *tty)
{
  if (tty_fd >= 0) {
	if (xioctl(tty_fd, TCGETS, tty, "tty: get_state") < 0)
	  return -errno;
	return 0;
  }
  return -1;
}

/* Put a terminal line in a transparent state. */
static int tty_set_raw(struct termios *tty)
{
  int i;
  int speed;

  for(i = 0; i < NCCS; i++)
		tty->c_cc[i] = '\0';            /* no spec chr          */
  tty->c_cc[VMIN] = 1;
  tty->c_cc[VTIME] = 0;
  tty->c_iflag = (IGNBRK | IGNPAR);             /* input flags          */
  tty->c_oflag = (0);                           /* output flags         */
  tty->c_lflag = (0);                           /* local flags          */
  speed = (tty->c_cflag & CBAUD);               /* save current speed   */
  tty->c_cflag = (CRTSCTS|HUPCL|CREAD|CLOCAL);  /* UART flags           */
  tty->c_cflag |= speed;                        /* restore speed        */
  return 0;
}

/* Set the number of data bits. */
static int tty_set_databits(struct termios *tty, char *databits)
{
#ifdef STANDALONE
  if (opt_v) printf("DIP: tty: set_databits: %c\n", *databits);
#endif
  tty->c_cflag &= ~CSIZE;
  switch(*databits) {
	case '5':
		tty->c_cflag |= CS5;
		break;

	case '6':
		tty->c_cflag |= CS6;
		break;

	case '7':
		tty->c_cflag |= CS7;
		break;

	case '8':
		tty->c_cflag |= CS8;
		break;

	default:
		return -EINVAL;
  }
  return 0;
}

/* Set the number of stop bits. */
static int tty_set_stopbits(struct termios *tty, char *stopbits)
{
#ifdef STANDALONE
  if (opt_v) printf("DIP: tty: set_stopbits: %c\n", *stopbits);
#endif
  switch(*stopbits) {
	case '1':
		tty->c_cflag &= ~CSTOPB;
		break;

	case '2':
		tty->c_cflag |= CSTOPB;
		break;

	default:
		return -EINVAL;
  }
  return 0;
}

/* Set the type of parity encoding. */
static int tty_set_parity(struct termios *tty, char *parity)
{
#ifdef STANDALONE
  if (opt_v) printf("DIP: tty: set_parity: %c\n", *parity);
#endif
  switch(toupper(*parity)) {
	case 'N':
		tty->c_cflag &= ~(PARENB | PARODD);
		break;

	case 'O':
		tty->c_cflag &= ~(PARENB | PARODD);
		tty->c_cflag |= (PARENB | PARODD);
		break;

	case 'E':
		tty->c_cflag &= ~(PARENB | PARODD);
		tty->c_cflag |= (PARENB);
		break;

	default:
		return -EINVAL;
  }
  return 0;
}


/* Open and initialize a terminal line. */
static int tty_open(const char *name)
{
  const char *sp;
  int fd = -1, flags;

  tty_mode = 0;

  /* Try opening the TTY device. */
  if ((sp = bb_basename(name)) == name)
	  name = concat_path_file("/dev", name);
  /* Now - can we lock it? */
  if(tty_lock(sp, 1)) {
	  bb_perror_msg("problems with locking %s", sp);
	  return -1;
  }
  /* drop privileges here for security reasons */
  seteuid(uid); /* now can attempt to open tty */
  if ((fd = open(name, O_RDWR|O_NDELAY)) < 0) {
	  bb_perror_msg("tty: open(%s, RW)", name);
	  seteuid(euid); /* return it back, just in case */
	  tty_lock(NULL, 0);
	  return -errno;
  }
  seteuid(euid); /* get the privileges back */
  tty_fd = fd;
#ifdef STANDALONE
  if (opt_v) {
    bb_error_msg("tty_open: %s (%d)", name, fd);
    syslog(LOG_INFO, "tty_open: %s (%d)", name, fd);
  }
#endif

  /* Size and allocate the I/O buffers. */
  in_size = TTY_BUFSIZE;
  out_size = in_size;
  in_buff = (char *) xmalloc(in_size);
  out_buff = (char *) xmalloc(out_size);
  in_cnt = 0;
  out_cnt = 0;
  in_ptr = in_buff;
  out_ptr = out_buff;
  out_size -= 4; /* safety */

  /* Fetch the current state of the terminal. */
  if (tty_get_state(&tty_saved) < 0) {
	bb_error_msg("tty_open: cannot get current state!");
	return -errno;
  }

  (void) memcpy (&tty_current, &tty_saved, sizeof(struct termios));

  /* Put this terminal line in a 8-bit transparent mode. */
  if (tty_set_raw(&tty_current) < 0) {
	bb_error_msg("tty_open: cannot set RAW mode!");
	return -errno;
  }

  /* set the speed. */
  tty_set_speed(&tty_current, var_speed);

  /* Set up a completely 8-bit clean line. */
  if (tty_set_databits(&tty_current, "8") ||
      tty_set_stopbits(&tty_current, "1") ||
      tty_set_parity(&tty_current, "N")) {
	bb_error_msg("tty_open: cannot set 8N1 mode!");
	return -errno;
  }

  /* Set the new line mode. */
  if ((fd = tty_set_state(&tty_current)) < 0) return fd;
  /* Clear the NDELAY flag now (line is in CLOCAL) */
  flags = fcntl(tty_fd, F_GETFL, 0);
  fcntl(tty_fd, F_SETFL, flags & ~O_NDELAY);

  /* OK, all done.  Lock this terminal line. */
  return 0;
}


/* Set the name of the terminal port to use. */
static int do_port(int argc, char *argv[])
{
  if (port_opened) {
    bb_error_msg("PORT: terminal port already set to \"%s\".", var_port);
    return -1;
  }

  var_port = xstrdup(argv[1]);
#ifdef STANDALONE
  if (opt_v) printf("PORT: terminal port set to \"%s\".\n", var_port);
#endif

  /* Initialize the terminal line. */
  if (tty_open(var_port) < 0) {
    var_port = NULL;
    return -1;
  }
  port_opened++;
  return 0;
}

static int chk_open_port(void)
{
  char *argv[2];

  argv[0] = "port";
  argv[1] = var_port;

  if (!port_opened) {
	if(var_port == NULL) {
		bb_error_msg("Please set PORT first.");
		return -1;
	}
	return do_port(2, argv);
  }
  return 0;
}

/* Send a string to the serial driver. */
static int do_send(int argc, char *argv[])
{
  char *sp;
  char c;
  int i;
  char t_buf[32];

  if (*argv[1] == '$') { /* We're dealing with variable! */
    struct variable *v;
    char *p;

    v = lookup(argv[1]);
    if (v == (struct variable *)NULL) {
      bb_error_msg("Variable %s not defined", argv[1]);
      return -1;
    }
    switch (v->type) {
    case 'i': /* integer value */
      p = t_buf;
      (void) sprintf (t_buf, "%i", v->vv.value);
      break;
    case 's': /* ASCII string value */
      p = v->vv.str_value;
      break;
    default:
      /* wrong/unknown variable type! */
      return -1;
    } /* end switch var-type */
    argv[1] = p;
  } /* end of processing variable */

  for (i = 1; i < argc; i++) {
	sp = argv[i];
	while(*sp != '\0') {
		switch(*sp) {
			case '~':
				tty_putc('\r');
				tty_putc('\n');
				break;

			case '\\':
				sp++;
				if (isdigit(*sp)) {
				   c = atoi(sp);
				   while(isdigit(*sp))
				     sp++;
				} else
				 c = cvt_char(*sp);
				tty_putc((int) c);
				break;

			default:
				c = *sp;
				tty_putc((int) c);
		}
		sp++;
	}
	if (i < (argc - 1)) tty_putc(' ');
  }
  tty_putc(-1);         /* flush the connection */
  return 0;
}


/* Send the output of a process to the serial driver */
static int do_psend(int argc, char *argv[])
{
  pid_t pid;
  char *l, cmd[1024];
  int pipe_fd[2], i, status;
  unsigned char ch;

  l = cmd;
  for(i = 1; i < argc; ++i) {
	*l++ = ' ';
	l = stpcpy(l, argv[i]);
  }
  *l = '\0';

  /* We can't use popen because the process has to be executed without
   * special privileges
   */
  if (pipe(pipe_fd) < 0) {
	bb_perror_msg("psend: pipe");
	return -1;
  }

  if (!(pid = fork())) {        /* This is the new child-process */
	int fd;

	dup2(pipe_fd[1], 1);
	/* Close all open files, so that no unprivileged process can read/
	 * write files we left open. Anyone using dip on a non-Posix-system?
	 */
	for (fd = sysconf(_SC_OPEN_MAX) - 1; fd >= 3; fd--)
		close(fd);

	setuid(uid);
	run_shell(NULL, cmd, NULL);
  }
  close(pipe_fd[1]);
  if (pid < 0) {        /* fork() failed */
	close(pipe_fd[0]);
	bb_perror_msg("psend: fork");
	return -1;
  }

  /* Now copy the output to the serial driver. This is obnoxiously
   * inefficient, but you just want to send short strings anyway, don't you?
   */
  while (read(pipe_fd[0], &ch, 1) == 1)
	tty_putc(ch);
  tty_putc(-1);

  close(pipe_fd[0]);
  if (waitpid_no_int(pid, &status, 0, 1) >= 0
	&& WIFEXITED(status) && WEXITSTATUS(status)) {
	bb_error_msg("%s returned status %d", cmd, WEXITSTATUS(status));
	return -1;
  }
  return 0;
}


/* Wait some time. */
static int do_sleep(int argc, char *argv[])
{
  int secs;

  get_value(&secs, argv[1]);
  (void) sleep(secs);
  return 0;
}

/* Execute a shell command using DIP variable substitution */
static int do_shell(int argc, char *argv[])
{
  char *usershell, *s;
  struct variable *var;
  char **subst_argv=0, *shell_argv[4];
  int i, status, ch_count=0;

  usershell = getenv("SHELL");
  if (!usershell)
  {
    bb_error_msg("Can't get SHELL environment variable");
    return -1;
  }

  /* Parse the arguments, doing variable substitution and writing to a new
     array */
  subst_argv = (char **) xmalloc((argc-1) * sizeof(char *));
  for(i = 1; i < argc; i++) {
    if (*argv[i] == '$') {      /* DIP variable substitution */
      if (! strcmp(&argv[i][1], "errlvl")) {
	subst_argv[i-1] = NULL;
	subst_argv[i-1] = xasprintf("%d", errlevel);
      }
      else if (! strcmp(&argv[i][1], "locip")) {
	s = inet_ntoa(loc_ip);
	subst_argv[i-1] = xstrdup(s);
      }
      else if (! strcmp(&argv[i][1], "rmtip")) {
	s = inet_ntoa(rmt_ip);
	subst_argv[i-1] = xstrdup(s);
      }
      else if (! strcmp(&argv[i][1], "local")) {
	subst_argv[i-1] = xstrdup(local);
      }
      else if (! strcmp(&argv[i][1], "remote")) {
	subst_argv[i-1] = xstrdup(remote);
      }
      else if (! strcmp(&argv[i][1], "mtu")) {
	/* I guess 11 characters is long enough for ints. */
	subst_argv[i-1] = xasprintf("%d", mtu);
      }
      else if (! strcmp(&argv[i][1], "modem")) {
	subst_argv[i-1] = xstrdup(modem);
      }
      else if (! strcmp(&argv[i][1], "port")) {
	subst_argv[i-1] = xstrdup(var_port);
      }
      else if (! strcmp(&argv[i][1], "speed")) {
	subst_argv[i-1] = xasprintf("%lu", var_speed);
      }
      else if (! strcmp(&argv[i][1], "onexit")) {
	subst_argv[i-1] = argv_sprintf (onexit_argc, onexit_argv);
	if (subst_argv[i-1] == NULL)  /* arg_sprintf() complains on error */
	  return -1;
      }
      else if ((var = lookup(&argv[i][1] - 1)) == (struct variable *)NULL) {
	bb_error_msg("Variable %s not defined", &argv[i][1]);
	subst_argv[i-1] = xstrdup(&argv[i][1]);
      }
      else {
	subst_argv[i-1] = NULL;
	subst_argv[i-1] = xasprintf("%d", var->vv.value);
      }
    }
    else if ((*argv[i]=='\\') && (argv[i][1]=='$')) {
      subst_argv[i-1] = xstrdup(&argv[i][1]);
    }
    else { /* ordinary parameter */
      subst_argv[i-1] = xstrdup(argv[i]);
    }
    ch_count += strlen(subst_argv[i-1]) + 1;
  }
  /* create a new argv array to pass to execv */
  shell_argv[0] = xstrdup(usershell);
  shell_argv[1] = xstrdup("-c");
  shell_argv[2] = (char *) xmalloc(ch_count+1);
  shell_argv[2][0] = 0;
  for(i = 0; i < argc-1; i++) {
    strcat(shell_argv[2], subst_argv[i]);
    if (i<argc-2) strcat(shell_argv[2], " ");
  }
  shell_argv[3] = NULL;
  /* now execute the command using execv */
  if (!fork()) { /* child process */
    seteuid(uid);          /* don't execute as root! */
    execv(usershell, shell_argv);
    exit(0);
  }
  else { /* parent process */
    wait(&status);
  }
  /* free memory we just allocated */
  for(i=0;i<argc-1;i++)
     free(subst_argv[i]);
  free(subst_argv);
  free(shell_argv[0]);
  free(shell_argv[1]);
  free(shell_argv[2]);
  return 0;
}

/* Flush tty input */
static int do_flush(int argc, char *argv[])
{
  tty_flush();
  return 0;
}


/* Wait for some string to arrive. */
static int do_wait(int argc, char *argv[])
{
  char c, c2, *p;
  int howlong;
  void (*oldsig)(int);

  if (argc == 3)
	get_value (&howlong, argv[2]);
  else
	howlong = 0;

#ifdef STANDALONE
  if (opt_v) {
    syslog(LOG_INFO, "waiting for \"%s\" for %d sec's.",
	   argv[1], howlong);
  }
#endif

  oldsig = bb_signal((1 << SIGALRM), TimeOut);
  timeout = 0;
  (void) alarm(howlong);

  p = argv[1];
  while((!timeout) && (*p != '\0')) {
	while (((howlong = tty_getc()) == -2) && !timeout);
	if ((howlong == -1) || timeout) {
	  timeout = 1;
	  break;
	}
	c = (char) howlong;
	c &= 0177;
	if ((var_echo)
#ifdef STANDALONE
		      || (opt_v)
#endif
		      ) {
		fputc(c, stdout);
		fflush(stdout);
	}
	if (timeout == 1) break;
	if (*p == '\\') c2 = cvt_char(*++p);
	  else c2 = *p;
	if (c2 != c) p = argv[1];
	  else p++;
  }
  (void) alarm(0);
  bb_signal((1 << SIGALRM), oldsig);

#ifdef STANDALONE
  if (opt_v) {
     if (timeout == 1)
	syslog(LOG_INFO, "do_wait: timeout...");
  }
#endif

  return timeout == 1 ? 3 : 0;
}

/* Set the correct DATABITS to use. */
static int do_databits(int argc, char *argv[])
{
  if (tty_set_databits(&tty_current, argv[1]) < 0) return -1;
  return tty_set_state(&tty_current);
}

/* Define a modem "INIT" string. */
static int do_init(int argc, char *argv[])
{
  mdm_init_s = argc == 1 ? "" : xstrdup(argv[1]);
#ifdef STANDALONE
  if (opt_v) printf("Modem INIT string set to \"%s\"\n", mdm_init_s);
#endif
  return 0;
}


static int do_dial(int argc, char *argv[])
{
  int d_timeout;
  struct variable *v;
  char *number = argv[1];
  char  buff[128];
  char dd_t_buf[32];
  char *prefix = (char *)0;
  int   res    = 0;

  if (( *number == '$' ) && (v=lookup(number))) {   /* variable */
	if      (v->type == 's')        number = v->vv.str_value;
	else if (v->type == 'i') {
	    number = dd_t_buf;
	    sprintf (dd_t_buf, "%i", v->vv.value);
	}
	else    return -1;
  }

  if (argc == 3)
     get_value(&d_timeout, argv[2]);
  else
     d_timeout = 60;
  var_dcount++;
  /* Dial a phone number. */

/* R.H. is Ronald Holzloehner <ronald@marie.physik.TU-Berlin.DE> */
#define AT_DIAL_STORED "ATDS="

  /* R.H. */
  /* Numbers beginning with 'S' are treated as stored ones,
  for example
  dial S2
  will be translated in
  ATDS=2
  */

  if(*number == 'S') {
	number++;
	prefix = AT_DIAL_STORED;
  } else if ( isdigit(*number) )
	prefix = AT_DIAL_TONE;
  else
	prefix = AT_DIAL;
/* R.H. End */


  /* Setup to dial out.  Send the init command */
  if(mdm_init_s != NULL && mdm_init_s[0] != '\0') {
#ifdef STANDALONE
	  if (opt_v)
		printf("Sending INIT string \"%s\"\n", mdm_init_s);
#endif
	  if ((res = mdm_chatseq(mdm_init_s, chat, 5)) != 0)
	    return res;
  }
  /* How's our parameter length? Trying to override stack? */
  snprintf(buff, sizeof(buff), "%s%s", prefix, number);

  /* Dial the phone number.  The dialing itself */
#ifdef STANDALONE
  if (opt_v)
	printf("Dialing number \"%s\"\n", buff);
#endif
  return mdm_chatseq(buff, chat, d_timeout);
}


static int do_break(int argc, char *argv[])
{
  if (tty_fd >= 0)
	return tcsendbreak(tty_fd, 0);
  return 0;
}

/* Set the type of parity of a terminal line. */
static int tty_parity(char *type)
{
  if (tty_set_parity(&tty_current, type) < 0) return -1;
  return tty_set_state(&tty_current);
}

static int do_parity(int argc, char *argv[])
{
  return tty_parity(argv[1]);
}


static int mdm_reset(void)
{
	/* Make the modem listen to us. */
	(void) sleep(1);
	if (tty_puts((unsigned const char *)AT_ATTN) < 0)
		return -1;
	(void) sleep(2);

	/* Reset the modem. */
	return mdm_chatseq(AT_RSET, chat, 5);
}

/* Reset the modem. */
static int do_reset(int argc, char *argv[])
{
  /* Reset the modem. */
  (void) mdm_reset();
  return 0;
}


/* Set the line speed of a terminal line. */
static int tty_speed(unsigned long speed)
{
  if (tty_set_speed(&tty_current, speed) < 0) return -1;
  return tty_set_state(&tty_current);
}

/* Set the correct SPEED to use. */
static int do_speed(int argc, char *argv[])
{
  int ret_stat = 0;

  var_speed = strtoul(argv[1], NULL, 10);
  if(port_opened) {
    ret_stat = tty_speed(var_speed);
    if ( ret_stat < 0 ) {
	   bb_error_msg("speed=%lu not supported, reverting to default",
		   var_speed);
	   var_speed = DEF_SPEED;
	   ret_stat = tty_speed(var_speed);
    }
  }
  return ret_stat;
}

/* Set the number of stopbits of a terminal line. */
static int tty_stopbits(char *bits)
{
  if (tty_set_stopbits(&tty_current, bits) < 0) return -1;
  return tty_set_state(&tty_current);
}


/* Set the correct STOPBITS to use. */
static int do_stopbits(int argc, char *argv[])
{
  return tty_stopbits(argv[1]);
}


static void do_terminal(void)
{
  struct termios otty, ntty;
  char buff[1024];
  void (*old_sigs[31])(int);
  fd_set mask, rmask;
  int i;
  int tty = tty_askfd();

  printf("[ Entering TERMINAL mode.  Use CTRL-] to get back ]\n");

  /* Set up the keyboard. */
  if (xioctl(0, TCGETS, &otty, "term ioctl TCGETS") < 0)
	return;
  ntty = otty;
  ntty.c_lflag &= ~(ECHO  | ICANON | ISIG );
  ntty.c_iflag &= ~(ICRNL | IXON   | IXANY);
  ntty.c_oflag &= ~(ONLCR | OPOST);
  if (xioctl(0, TCSETS, &ntty, "term ioctl TCSETS") < 0)
	return;

  /* Set up the correct descriptor masks for select(2). */
  FD_ZERO(&mask);
  FD_SET(0, &mask);
  FD_SET(tty, &mask);

  /* Disable all signals, saving them for later. */
  for (i = 0; i < 31; i++) old_sigs[i] = bb_signal((1 << (i + 1)), SIG_IGN);

  /* Go into an endless terminal loop. */
  while(1) {
	rmask = mask;
	i = select(32, &rmask, (fd_set *)NULL, (fd_set *)NULL,
					(struct timeval *)NULL);
	if (i <= 0) break;
	if (FD_ISSET(0, &rmask)) {
		i = read(0, buff, 1);
		if (i > 0) {
			if (buff[0] == (']' & 037)) break;
			(void) write(tty, buff, i);
		}
	}

	if (FD_ISSET(tty, &rmask)) {
		i = read(tty, buff, 1024);
		if (i > 0) (void) write(1, buff, i);
	};
  }

  if (xioctl(0, TCSETS, &otty, "term ioctl TCSETS") < 0)
	return;

  /* Restore all signals. */
  for (i = 0; i < 31; i++) bb_signal((1 << (i + 1)), old_sigs[i]);

  printf("\n[ Back to LOCAL mode. ]\n");
}

/* Enter a TERMINAL mode. */
static int do_term(int argc, char *argv[])
{
  do_terminal();
  return 0;
}

/*
 * Get or ask for the value of a variable.
 * Get local IP#, Remote IP# and MTU
 */

#define VAR_IP          1
#define VAR_NUM         2
#define VAR_STR         3

static int do_get(int argc, char *argv[])
{
  char str[128] = "";
  int var = 0;
  void *v1 = NULL, *v2 = NULL;
  void (*oldsig)(int);
  struct variable *v = NULL;
  const char *svar = argv[1];

  /*
   * Check for special variables
   */
  if(*svar != '$') {
    bb_error_msg("Variable must lead with '$'");
    return -1;
  }
  svar++;

  if (! strcmp(svar, "locip") || !strcmp(svar, "local")) {
	v1 = (void *) local;
	v2 = (void *) &loc_ip;
	var = VAR_IP;
  } else if (! strcmp(svar, "rmtip") || !strcmp(svar, "remote")) {
	v1 = (void *) remote;
	v2 = (void *) &rmt_ip;
	var = VAR_IP;
  } else if (! strcmp(svar, "mtu")) {
	v1 = (void *) &mtu;
	var = VAR_NUM;
  } else { /* a new variable */
     if((v = new_var(svar-1)) == (struct variable *)NULL)
       return -1;
     v1 = (void *) &v->vv.value;
     var = VAR_NUM;
  }

  if (argc > 2) {
    if (! strcmp(argv[2], "ask")) {
      /* Ask the stdinput for the value of the variable. */
      fprintf(stderr, "Enter the value for %s: ", argv[1]);
      if (fgets(str, 128, stdin) == (char *)NULL)
	return 1;
      strchr_set0(str, '\n');
      if (is_a_num(str) == 0) { /* if String value */
	var = VAR_STR;
	v->type  = 's';
	v->vv.str_value = xstrdup(str);
	return 0;
      }
    } else if (! strcmp(argv[2], "remote")) {
      /* Get the variable string from the "remote" line */
      char c, *p;
      int howlong = 0, state = 0;

      if (argc == 4) get_value (&howlong, argv[3]);
      oldsig = bb_signal((1 << SIGALRM), TimeOut);
      (void) alarm(howlong);

      p = str;
      timeout = 0;
      state = 0;
      if (var == VAR_IP || var == VAR_NUM) state = 1;
      while(!timeout && (state >= 0)) {
	while (((howlong = tty_getc()) == -2) && !timeout);
	if ((howlong == -1) || timeout) {
	  timeout = 1;
	  break;
	}
	c = (char) howlong;
	c &= 0177;
	switch (state) {
	case 0: /* throw away "white space" */
	  if (! strchr(" \n\t", c)) break;
	  if (var == VAR_STR) {
	    *(p++) = c;
	    state = 10;
	    break;
	  }
	  state = 1;

	case 1:
	  if ( c >= '0' && c <= '9' ) {
	    *(p++) = c;
	    if (var == VAR_IP) {
	      state = 2;
	      break;
	    } else state = 5;
	    break;
	  }
	  break;

	case 2:
	case 3:
	case 4:
	  if ((c >= '0' && c <= '9') || c == '.') {
	    *(p++) = c;
	    if (c == '.') state++;
	    break;
	  }
	  p = str;
	  state = 0;
	  break;
	case 5:
	  if (c >= '0' && c <= '9') {
	    *(p++) = c;
	    break;
	  }
	  state = -1;
	  break;

	case 10:
	  if (strchr(" \n\t", c)) state = -1;
	  else *(p++) = c;
	  break;

	default:
	  break;
	}
      }
      *p = '\0';
      (void) alarm(0);
      bb_signal((1 << SIGALRM), oldsig);
      if (timeout == 1) return 1;
    } else {
      /* The third argv is the value for the variable. */
      strncpy(str, argv[2], 128);
      str[127] = 0;
      if ((is_a_num(str) == 0) && (var != VAR_IP)) {
	var = VAR_STR;
	v->type = 's';
      }
    }
  }

#ifdef STANDALONE
  if (opt_v) printf("About to set variable %s to %s\n", argv[1], str);
#endif

  if (*str) {
    switch (var) {
    case VAR_IP:
      {
	char *cn;
	if(bb_gethostbyname(BB_GHBN_IN_ADDR | BB_GHBN_CANON_NAME, str, v2, &cn) == NULL)
		return -1;
	safe_strncpy((char *) v1, cn, 128);
	free(cn);
      }
      break;

    case VAR_NUM:
      *(int *) v1 = atoi(str);
      break;

    case VAR_STR:
      v->type = 's'; /* var type is String */
      v->vv.str_value = xstrdup(str);
      break;

    default:
      break;
    }
  }

  return 0;
}


static inline void log_cleanup(void)
{
  tty_close();
}

static void sig_login_catcher(int sig)
{
  bb_signal((1 << SIGALRM), oldalrmsig);
  alarm (0);

  /* Don't let SIGHUP from terminal bother us here */
  bb_signal((1 << SIGHUP), SIG_IGN);

  if (do_exit == 1) {
    bb_error_msg("Line disconnected");
    log_cleanup();
    do_exit = 0;
    return;
  }
  log_cleanup();
  exit(0);
}

/* Cleanup before exit. */
static __attribute__ ((noreturn)) void cleanup(int exit_code)
{
  /* execute the onexit command(s) */
  if (onexit_argc != 0 && onexit_argv != NULL)
    do_shell (onexit_argc, onexit_argv);
  arg_clean (onexit_argc, onexit_argv);

  tty_close();
  exit(exit_code);
}

/* Catch any signals. */
static void sig_catcher(int sig)
{
  /* Don't let SIGHUP from terminal bother us here */
  bb_signal((1 << SIGHUP), SIG_IGN);

  bb_signal((1 << SIGALRM), oldalrmsig);
  alarm (0);

  (void) cleanup(0);
}

static void do_ppp(void)
{
  int i;
  char *argv[20];
  char string[255];
  char mtu_string[16];
  char log_string[512];

  dup2(tty_askfd(),0);
  dup2(tty_askfd(),1);    /* re-direct cua? to stdin/stdout */

  argv[0] = "pppd";
  argv[1] = "-detach";
  argv[2] = "crtscts";
  argv[3] = "modem";
  argv[4] = "mtu";
  argv[5] = mtu_string;
  argv[6] = string;
  i = 7;

  sprintf(mtu_string, "%d", mtu);
  if(rtdefault) {
	argv[i++] = "defaultroute";
  }
  if(proxyarp) {
	argv[i++] = "proxyarp";
  }
  if (opt_v) {
	argv[i++] = "debug";
  }
  if(user_pass[0]) {
	char *pc = strchr_set0(user_pass, ':');
	argv[i++] = "user";
	argv[i++] = user_pass;
	if(pc) {
		argv[i++] = "password";
		argv[i++] = pc + 1;
	}
  }
  argv[i] = NULL;

  if (loc_ip.s_addr == INADDR_ANY && rmt_ip.s_addr == INADDR_ANY) {
      /* PPP will negotiate everything by itself, automatic!     */
	strcpy(string, "noipdefault");
  } else {
	sprintf(string, "%s:%s",
	    (loc_ip.s_addr != INADDR_ANY ? local : ""),
	    (rmt_ip.s_addr != INADDR_ANY ? remote : ""));
  }
  strcpy(log_string, "Invoking");
  for(i = 0; argv[i]; i++) {
    strcat(log_string, " ");
    strcat(log_string, argv[i]);
  }
  syslog(LOG_DEBUG, "%s", log_string);
  execvp(_PATH_BIN_PPP, argv);

  bb_perror_msg("do_ppp");    /* point of no return */
}


/* Enter a specific protocol. */
static int do_mode(int argc, char *argv[])
{
  int i;

  if (argc == 2 && strcasecmp(argv[1], "PPP")) {
    bb_error_msg("protocol_name \"%s\" not found!", argv[1]);
    return -1;
  }

  /* Enter BACKGROUND mode here! */

#if 1
  /* First of all, if not forked yet - fork off a sub-process. */
  if (!forked) {
    if ((i = fork()) < 0) return 0;
    if (i != 0) exit(0);
  }
#endif

  /* Make tty the control terminal, detect DCD loss from now. */
  if (tty_fd < 0)
	return -1;

  /* Release any control terminal we might have. */
  if (setsid() < 0) {
    bb_perror_msg("tty_notlocal cannot setsid");
  }

#ifdef STANDALONE
  if (opt_v) {
    const char *tn;

    if (isatty(tty_fd))
       tn = ttyname(tty_fd);
    else
       tn = "not defined";
    syslog(LOG_INFO, "tty: flle_%d %s", tty_fd, tn);
  }
#endif

  xioctl(tty_fd, TIOCSCTTY, (void *)1, "tty_notlocal cannot TIOCSCTTY");

  /* Clear the CLOCAL bit. (detect carrier loss) */
  tty_current.c_cflag &= ~CLOCAL;
  if (tty_set_state(&tty_current) < 0) {
    bb_perror_msg("tty_notlocal cannot clr CLOCAL");
    return -1;
  }

  if (!forked) {

    for (i = 0; i < 3; i++)
	close(i);

    /* Disable screen output... */
    (void) open(bb_dev_null, O_RDONLY);  /* stdin  */
    (void) open(bb_dev_null, O_WRONLY);  /* stdout */
    (void) dup(1);                       /* stderr */
    log_syslog = LOG_ERR;
    log_stderr = 0;

    /* Re-acquire the lock */
    (void)tty_lock(NULL, 2);
    forked = 1;
  }

  if (!catched) {
    catched = 1;

    oldalrmsig = bb_signal((1 << SIGALRM), SIG_IGN);
    bb_signal((1 << SIGHUP),  sig_login_catcher);
    bb_signal((1 << SIGINT) | (1 << SIGTERM) | (1 << SIGQUIT), sig_catcher);
  }

  /* Standard BSD behaviour: change to the root dir. */
  (void) chdir("/");

  /* Fire up the protocol here. */
#ifdef STANDALONE
  if (opt_v)
    syslog (LOG_INFO, "Firing up PPP protocol");
#endif
  do_ppp();

  /* Wait forever to terminate. */
  do_exit = 1;
  while(do_exit) {
     (void) sleep(30);
  }

  return 0;
}

/* Quit upon error - reset the tty to the proper ldisc */
static int do_quit(int argc, char *argv[])
{
  if (port_opened)
	tty_close();
  exit(1);
}

/* Set routing as default */
static int do_default(int argc, char *argv[])
{
#ifdef STANDALONE
  const char *on_off = "on";
#endif

  rtdefault = 1;
  if(argc == 2 && strcasecmp(argv[1], "off") == 0) {
#ifdef STANDALONE
    on_off = "off";
#endif
    rtdefault = 0;
  }
#ifdef STANDALONE
  if (opt_v)
    printf("Destination net/address set 'default' to %s\n", on_off);
#endif

  return 0;
}


/* Request Proxy ARP to be set */
static int do_proxyarp(int argc, char *argv[])
{
#ifdef STANDALONE
  const char *on_off = "on";
#endif

  proxyarp = 1;
  if(argc == 2 && strcasecmp(argv[1], "off") == 0) {
#ifdef STANDALONE
    on_off = "off";
#endif
    proxyarp = 0;
  }

#ifdef STANDALONE
  if (opt_v) printf("Setting Proxy ARP entry to %s\n", on_off);
#endif

  return 0;
}

/* Request Proxy ARP to be set */
static int do_user_pass(int argc, char *argv[])
{
  if(argc == 2) {
    safe_strncpy(user_pass, argv[1], 128);
#ifdef STANDALONE
  if (opt_v) printf("Setting username and password to %s\n", argv[1]);
#endif
  }
  return 0;
}

/* Prompt for a password to be sent */
static int do_password(int argc, char *argv[])
{
    const char *prompt = "Password:";
    char        *pass[3];

    pass[0] = "send";           /* in case anyone cares */
    pass[2] = NULL;             /* ditto */

    if (argc > 1)
	prompt = argv[1];

    pass[1] = getpass(prompt);
    do_send(2, pass);
    return 0;
}

static int do_inc(int argc, char *argv[])
{
  struct variable *v;
  int increment = 1;

  if(*argv[1] != '$') {
    bb_error_msg("Variable must lead with '$'");
    return -1;
  }

  /*
   * Check for special variables
   */
  if(is_special(argv[1]))
    return -1;

  if((v = lookup(argv[1])) == (struct variable *)NULL) {
    bb_error_msg("Variable %s not defined", argv[1]);
    return -1;
  }

  if(argc == 3)
    get_value(&increment, argv[2]);

#ifdef STANDALONE
  if(opt_v)
    printf("Incrementing variable %s for %d\n", argv[1], increment);
#endif

  v->vv.value += increment;
  return 0;
}

static int do_dec(int argc, char *argv[])
{
  struct variable *v;
  int decrement = 1;

  if(*argv[1] != '$') {
    bb_error_msg("Variable must lead with '$'");
    return -1;
  }

  /*
   * Check for special variables
   */
  if(is_special(argv[1])) {
    return -1;
  }

  if((v = lookup(argv[1])) == (struct variable *)NULL) {
    bb_error_msg("Variable %s not defined", argv[1]);
    return -1;
  }

  if(argc == 3)
    get_value(&decrement, argv[2]);

#ifdef STANDALONE
  if(opt_v)
    bb_error_msg("Decrementing variable %s for %d", argv[1], decrement);
#endif

  v->vv.value -= decrement;
  return 0;
}

static void dip_addchat(const char *string, int code)
{
  struct response *resp;

  resp = xmalloc(sizeof(*resp));
  resp->string = xstrdup(string);
  resp->code = code;
  resp->next = chat;
  chat = resp;
}

static int do_chatkey(int argc, char *argv[])
{
  char buf[128];
  int code;

  safe_strncpy(buf, argv[1], sizeof(buf));

  if(argc == 3) {
    code = bb_xparse_i_number(BB_PN_RETURN_ERR | BB_PN_ERR_STR1,
				argv[2], "Vader's numbers not permitted");
    if(code < 0)
	return -1;
  } else {
   code = 2;
  }

  dip_addchat(buf, code);

#ifdef STANDALONE
  if (opt_v)
    bb_error_msg("Added chat response keyword '%s' with code '%d'", buf, code);
#endif
  return 0;
}


#define A1  1
#define A2  2
#define A3  4
#define A4  8
#define A5  16
#define A6  32
#define Any 64
#define CCP 128 /* command call chk_open_port firts */

static int do_help(int argc, char *argv[]);

static const struct commands commands[] = {
  { A2 ,       "beep",       do_beep,     "[times]"    },
  { A1|CCP,    "break",      do_break,    ""    },
  { A2|A3,     "chatkey",    do_chatkey,  "<keyword> [<code>]" },
  { A2|CCP,    "databits",   do_databits, "bits"    },
  { A2|A3,     "dec",        do_dec,      "$variable [<decrement value>|variable]" },
  { A1|A2,     bb_default,   do_default,  "[on|off]"    },
  { A2|A3|CCP, "dial",       do_dial,     "telno [<timeout>]"    },
  { A2 ,       "echo",       do_echo,     "on|off"    },
  { A1|CCP,    "flush",      do_flush,    ""    },
  { A2|A3|A4,  "get",        do_get,      "$variable [value | ask | remote "
					"[timeout_value | variable]]"    },
  { A2 ,       "goto",       do_goto,     "label"    },
  { A1|A2,     "help",       do_help,     "[command]"    },
  { A6 ,       "if",         do_if,       "expr goto label"    },
  { A2|A3,     "inc",        do_inc,      "$variable [<increment value>|variable]" },
  { A1|A2|CCP, "init",       do_init,     "[<init string>]"    },
  { A1|A2,     "mode",       do_mode,     "[PPP]"    },
  { 2|Any,     "onexit",     do_onexit,   "command [parameters]"    },
  { A2|CCP,    "parity",     do_parity,   "E/O/N"    },
  { A1|A2,     "password",   do_password, "[prompt]"    },
  { A1|A2,     "proxyarp",   do_proxyarp, "[on|off]"   },
  { Any,       "print",      do_print,    "[-n] [messages]"    },
  { 2|Any|CCP, "psend",      do_psend,    "cmd args..."    },
  { A2 ,       "port",       do_port,     "tty_name"    },
  { A1 ,       "quit",       do_quit,     ""    },
  { A1|CCP,    "reset",      do_reset,    ""    },
  { 2|Any|CCP, "send",       do_send,     "words of text..."    },
  { 2|Any,     "shell",      do_shell,    "command [parameters]"    },
  { A2 ,       "sleep",      do_sleep,    "time_in_secs"    },
  { A2,        "speed",      do_speed,    "baudrate"    },
  { A2|CCP,    "stopbits",   do_stopbits, "bits"    },
  { A1|CCP,    "term",       do_term,     ""    },
  { A2,        "user_pass",  do_user_pass,"user:password"  },
  { A2|A3|CCP, "wait",       do_wait,     "text [timeout_value | variable]"    },
  { 0,         (char *)NULL, NULL,        NULL    }
};

/* Show some help. */
static int do_help(int argc, char *argv[])
{
  int i = 0;
  int found = 0;

  if(argc != 2)
    printf("DIP knows about the following commands:\n\n");
  while (commands[i].name != (char *)NULL) {
	if(argc != 2 || strcasecmp(commands[i].name, argv[1]) == 0) {
	    printf("%-12.12s %s\n", commands[i].name, commands[i].usage);
	    found++;
	}
	i++;
  }
  if(argc != 2)
    printf("\n");
  else if(found == 0)
    printf("DIP don`t know command %s\n", argv[1]);
  return 0;
}


static __attribute__((noreturn)) void do_command(FILE *fp)
{
  char cline[BUFSIZ];
  char *argv[ARGVSIZE];
  int argc, i;
  int running;
  char *sp, *xp;
  off_t offs;
  int errlevel_flag = 1;
  int errlevel_temp = 0;

  /* Initialize the command level module. */
  scriptfp  = fp;
  running   = 1;
  timeout   = 0;

  bb_signal((1 << SIGINT), sig_exit);

  do {
	if (scriptfp == stdin) {
#ifdef CONFIG_FEATURE_COMMAND_EDITING
		char prompt[32];
		sprintf(prompt, (opt_v ? "DIP [%-4d]> " : "DIP> "), errlevel);
		if(cmdedit_read_input(prompt, cline) <= 0)
		    goto l_eof;
		goto l_skip_fgets;
#else
		if (opt_v)
		    printf("DIP [%-4d]> ", errlevel);
		else
		  printf("DIP> ");
		fflush(stdout);
#endif
	}
	if (fgets(cline, 1024, scriptfp) == (char *)NULL) {
#ifdef CONFIG_FEATURE_COMMAND_EDITING
l_eof:
#endif
	  if (alllabels != -1)
	    alllabels = 1;
	  break;
	}
#ifdef CONFIG_FEATURE_COMMAND_EDITING
l_skip_fgets:
#endif
	strchr_set0(cline, '\n');
	sp = cline;
	while (*sp == ' ' || *sp == '\t') sp++;
	if (*sp == '#' || *sp == '\0') continue;
#ifdef STANDALONE
	if (opt_v) bb_error_msg(">> %s", sp);
#endif
	if ((argc = getargs(sp, argv)) == 0) continue;

	/* If this is a label, skip it. */
	if ((xp = strchr(argv[0], ':')) != (char *)NULL) {
	  if (!alllabels) {
	    *xp = '\0';
	    if ((offs = ftell(scriptfp)) == (off_t) -1)
	      alllabels = -1;
	    else {
	      if ((offs > last_label()) && (add_label(sp, offs) != 0))
		bb_error_msg("Duplicated label `%s' - ignored", sp);
#ifdef STANDALONE
	      if (opt_v)
		bb_error_msg("Got label `%s' at offset %08X", sp,
			(unsigned int) offs);
#endif
	    }
	  }
	  continue;
	}

	if (
	    (strcmp(argv[0], "if")    == 0)  &&
	    (strcmp(argv[0], "goto")  == 0)  &&
	    (strcmp(argv[0], "print") == 0)
	    )
	  {
	    errlevel_flag=0; /* don't modify errlevel value */
	    errlevel  = 0; /* allow print of $errlvl after error detected */
	  }
	else {
	  errlevel_flag=1; /* allow to modify errlevel with the new rc */
	}

	/* Now, check which command it is. */
	if (strcasecmp(argv[0], "exit") != 0) {
	  i = 0;
	  while (commands[i].name != (char *)NULL) {
	    if (strcasecmp(commands[i].name, argv[0]) == 0) break;
	    i++;
	  }
	  if (commands[i].name != (char *)NULL) {
	    int num_args_ok = 1;
	    char valid_n_args = commands[i].valid_n_args & ~CCP;

	    if(valid_n_args != Any) {
		if (valid_n_args & Any) {
		    int min_a = valid_n_args & ~Any;

		    if(argc < min_a)
			num_args_ok = 0;
		} else {
		    if(((1 << (argc-1)) & ~valid_n_args))
			num_args_ok = 0;
		}
	    }
	    if(num_args_ok == 0) {
		char *h_argv[2];

		h_argv[0] = "help";
		h_argv[1] = argv[0];
		printf("Usage: ");
		do_help(2, h_argv);
		errlevel_temp = -1;
	    } else {
		if((commands[i].valid_n_args & CCP) && chk_open_port())
		    errlevel_temp = -1;
		else
		    errlevel_temp = (*commands[i].func)(argc, argv);
	    }
	    if (errlevel_flag)
		    errlevel = errlevel_temp;
	  } else printf("? Unknown command (%s).\n", argv[0]);
	} else {
	  if (argc>1 && (errlevel=atoi(argv[1]))==-1) errlevel = 0;
	  running = 0;
	}
  } while(running);

  cleanup(errlevel);
}


static void catch(int sig)
{
  (void) sig;
  bb_signal((1 << SIGHUP), SIG_IGN);
  cleanup(1);
}


static void log_catch(int sig)
{
  (void) sig;
  bb_signal((1 << SIGHUP), SIG_IGN);
  log_cleanup();
  exit(1);
}


int dip_main(void)
{
  char *path;
  FILE *fp;
  int opt;
  char *s_speed;
  char **argv;

  /* Setup. */
  mdm_init_s = AT_INIT;
  uid = getuid();
  euid = geteuid();
  tty_fd = -1;

  /* Scan command line for any arguments. */
  mtu = DEF_MTU;
  opt = (int)bb_xgetopt_ulflags("ktvl:m:s:\0:-m:?1", &var_port, &mtu, &s_speed);
  opt_v = opt & 4;
  if(opt & 32)
	var_speed = strtoul(s_speed, NULL, 10);
  else
	var_speed = DEF_SPEED;

  /* make sure we log the attachments */
  bb_openlog(LOG_PID, LOG_DAEMON);

  bb_error_msg("Dialup IP Protocol Driver version " VERSION);

  /* Are we called to kill off a DIP process? */
  if (opt & 1) {
    if(opt & 8) {
	int pid = 0;

	path = xasprintf("%s/LCK..%s", _PATH_LOCKD, bb_basename(var_port));
	fp = bb_xfopen(path, "Ar");
	(void) fscanf(fp, "%d", &pid);
	seteuid(uid);
	if (kill(pid, SIGTERM))
		bb_perror_msg_and_die("cannot kill process %d", pid);
	else
		bb_error_msg("process %d killed.", pid);
	seteuid(euid);
	return unlink(path);
    } else {
	bb_show_usage();
    }
  }

  dip_addchat("OK",          0);
  dip_addchat("CONNECT",     1);
  dip_addchat("ERROR",       2);
  dip_addchat("BUSY",        3);
  dip_addchat("NO CARRIER",  4);
  dip_addchat("NO DIALTONE", 5);

  bb_signal((1 << SIGINT) | (1 << SIGTERM) | (1 << SIGQUIT), catch);
  bb_signal((1 << SIGHUP),  log_catch);

  argv = Gargv;

  /* Are we running in TERMINAL/DIALOG mode? */
  if (opt & 2) {
    if (*argv) bb_show_usage();
    do_command(stdin);
    /*NOTREACHED*/
  }

  /* No, so we are running in dialOUT mode. */
  if (*argv == NULL) bb_show_usage();
  path = *argv;

  /* set euid to ruid */
  if (setreuid(euid, uid) != 0) {
	bb_perror_msg_and_die("setreuid(%d, %d)", euid, uid);
	return -1;
  }
  /* open file with real uid */
  if ((fp = fopen(path, "r")) == (FILE *)NULL) {
      if (strchr(path, '.') == (char *)NULL) {
	path = xasprintf("%s%s", path, DIP_SFX);
	fp = bb_xfopen(path, "Ar");
      } else {
	bb_perror_msg_and_die("%s", path);
      }
  }
  /* set uids back */
  if (setreuid(uid, euid) != 0) {
	bb_perror_msg_and_die("setreuid(%d, %d)", uid, euid);
  }

  (void) setbuf(fp, (char *)NULL);
  do_command(fp);

  /*NOTREACHED*/
}

#ifdef STANDALONE
/* require from libbb.a for linking */
void bb_show_usage(void)
{
  fprintf(stderr, "Usage: %s %s\n%s\n\n", bb_applet_name, dip_trivial_usage, dip_full_usage);
  exit(1);
}

int main(int argc, char **argv)
{
  libbb_init();
  log_stderr = 1;
  return dip_main();
}

#endif  /* STANDALONE */
#endif /* APPLET_USAGE */
