/*
 * eline - enhanced read one line
 *
 * Copyright (C) 2022 Vladimir Oleynik <dzo@simtreas.ru>
 *
 * Public Domain.
 */

#if defined(APPLET_USAGE) || defined(STANDALONE)
#define eline_trivial_usage \
	"[-d delim] [-E code] [-Ls] [-n nchar] [-T fd2#] [-t timeout] [-u fd#] [-z code] [-Z repl_zero]"
#define eline_full_usage \
	"Enhanced read one line.\n\n" \
	"Options:\n" \
	"\t-d delim\tCharacters of delim is used to terminate the input line,\n" \
	"\t\trather than newline if -n nchar unsets.  If delim is the empty string,\n" \
	"\t\tread will terminate a line when it reads a NUL character.\n" \
	"\t\tString interpretation of backslash escapes.\n" \
	"\t-E code\t\tSet exit code when detect EOF, default 0.\n" \
	"\t-L\t\tIgnore LF (\\r).\n" \
	"\t-n nchar\tRead returns after reading exactly nchars characters rather than waiting\n" \
	"\t\tfor a complete line of input, unless EOF is encountered or read times out,\n" \
	"\t\tset exit code 126 after nchar read and the delim is set and do not found.\n" \
	"\t-s\t\tSkip leading whitespace.\n" \
	"\t-T fd2#\t\tCopy stdin to the descriptor too, example: s=$(eline -n10 -T2)\n" \
	"\t-t timeout[m|h]\tSet read timeout (seconds > 0), m - in minutes, h - in hours,\n" \
	"\t\tset exit code 125 if timeout.\n" \
	"\t-u fd\t\tRead input from file descriptor fd, default 0.\n" \
	"\t-z code\t\tSet exit code when detect reads NUL byte, if -Z repl is not specified,\n" \
	"\t\treturn immediately.\n" \
	"\t-Z repl_str\tReplce NUL byte with string."
#endif

#if !defined(APPLET_USAGE) || defined(STANDALONE)

#ifdef STANDALONE
# define _GNU_SOURCE
#endif

#include "busybox.h"
#include <unistd.h>
#include <string.h>
#include "bb_sig.h"
#include <sys/ioctl.h>
#include <termios.h>

#define setTermSettings(fd, argp) tcsetattr(fd, TCSANOW, argp)
#define getTermSettings(fd, argp) tcgetattr(fd, argp);

#define COMPLEMENTARY "\0:-Enuz:=0"
#include "opt_gen.h"
OPT_GEN(
opt_simple(d, delim),
opt_simple(E, code),
opt_simple(L),
opt_simple(n, nchar),
opt_simple(s),
opt_simple(T, fd2),
opt_simple(t, timeout),
opt_simple(u, fd),
opt_simple(z, code),
opt_simple(Z, repl))
#include "opt_gen.h"

#ifndef STANDALONE
CALC_BSS struct BB_COMMON_BSS {
	struct termios initial_settings, new_settings;
} bb_common_bss;
#define initial_settings bb_common_bss.initial_settings
#define new_settings bb_common_bss.new_settings
# else
static struct termios initial_settings, new_settings;
#endif

static void th(int s)
{
	(void)&s;
}

static void out2(int fd2, char c, int fd1)
{
	if (fd2 >= 0 && safe_write(fd2, &c, 1) < 0)
		bb_perror_msg_and_die("write to #%d", fd2);
	if (fd1 == 1)
		out2(1, c, -1);
}

int
eline_main(void)
{
	int nchar = INT_MAX, fd = 0, fd2 = -1, ecode = 0, zcode = 0, r, status = 0, sws = 0;
	char c, *delim = "\n", *sfd2 = NULL, *stimeout = NULL, *repl = NULL;
	unsigned int flags, timeout;
	const char *p;

	flags = bb_xgetopt_ulflags(opt_str, &delim, &ecode, &nchar, &sfd2, &stimeout, &fd, &zcode, &repl);
	sws = flags & OPT_s;
	if ( fd < 0) {
		bb_default_error_retval = 2;
		bb_error_msg_and_die("need fd>=0");
	}
	if (sfd2 != NULL) {
		fd2 = bb_xparse_i_number(BB_PN_ERR_STR1, sfd2, "invalid descriptor '%s'");
		if (fd2 == fd) {
			bb_default_error_retval = 2;
			bb_error_msg_and_die("error: -T need do not eq %d", fd);
		}
	}
	if (nchar == 0)
		return 0;
	if (nchar < 0) {
		bb_default_error_retval = 2;
		bb_error_msg_and_die("need nchar>=0");
	}
	if ((flags & OPT_d) != 0) {
		char *q;

		p = delim;
		q = delim = xstrdup(delim);
		while (*p) {
			if (*p == '\\') {
				p++;
				*q++ = bb_process_escape_sequence(&p);
			} else {
				*q++ = *p++;
			}
		}
		*q = '\0';
	}

	if ((flags & OPT_n) != 0 && isatty(fd)) {
		getTermSettings(0, (void *) &initial_settings);
		memcpy(&new_settings, &initial_settings, sizeof(struct termios));
		new_settings.c_lflag &= ~ICANON;        /* unbuffered input */
		/* Turn off echoing and CTRL-C, so we can trap it */
		new_settings.c_lflag &= ~(ECHO | ECHONL | ISIG);
		/* Hmm, in linux c_cc[] not parsed if set ~ICANON */
		new_settings.c_cc[VMIN] = 1;
		new_settings.c_cc[VTIME] = 0;
		/* Turn off CTRL-C, so we can trap it */
#ifndef _POSIX_VDISABLE
# define _POSIX_VDISABLE '\0'
#endif
		new_settings.c_cc[VINTR] = _POSIX_VDISABLE;
		setTermSettings(fd, (void *) &new_settings);
		flags |= (1<<31);
	}

	if (stimeout != NULL) {
		static const struct suffix_mult mh_suffixes[] = {
			{ "m", 60 },
			{ "h", 60*60 },
			{ NULL, 0 }
		};
		timeout = bb_xparse_ui_number(BB_PN_SET_LOWER1 | BB_PN_SHOW_USAGE | BB_PN_SUFFIX,
						stimeout, mh_suffixes);
		bb_signal(1 | (1<<SIGALRM), th);
		alarm(timeout);
	}

	for (;;) {
		r = read(fd, &c, 1);
		if (r < 0) {
			if (errno == EAGAIN || r == EWOULDBLOCK)
				continue;
			if (errno != EINTR)
				bb_perror_msg_and_die("read");
			status = 125;
			break;
		}
		if (r == 0) {
			status = ecode;
			break;
		}
		if ((flags & OPT_d) != 0 || (flags & OPT_n) == 0) {
			for (p = delim; *p; p++)
				if (c == *p)
					break;
			if (*p) {
				out2(fd2, c, -1);
				break;
			}
		}
		if (c == 0) {
			status = zcode;
			if (*delim == '\0') {
				out2(fd2, c, -1);
				break;
			}
			if (repl != NULL) {
				p = repl;
				while (*p) {
					c = *p++;
					if (c == '\\')
						c = bb_process_escape_sequence(&p);
					out2(fd2, c, 1);
				}
				goto skip;
			} else {
				if ((flags & OPT_z))
					break;
			}
		} else if (c == '\r' && (flags & OPT_L) != 0) {
			continue;
		} else if (sws && (c == '\r' || c == ' ' || c == '\t' || c == '\f' || c == '\v' || c == '\n')) {
			out2(fd2, c, -1);
			continue;
		} else {
			sws = 0;
		}
		out2(fd2, c, 1);
	  skip:
		if (--nchar == 0) {
			if ((flags & OPT_d))
				status = 126;
			flags |= (1<<30);
			break;
		}
	}

	if (stimeout != NULL)
		alarm(0);
	if((flags & (1<<31)))
		setTermSettings(fd, (void *) &initial_settings);
	if((flags & (1<<30)) == 0) {
		c = '\n';
		out2(-1, c, 1);
	}
	return status;
}

#ifdef STANDALONE
void bb_show_usage (void)
{
	const char *a = bb_applet_name;
	bb_applet_name = NULL;
	bb_default_error_retval = 2;
	bb_error_msg_and_die("Usage: %s %s\n%s\n", a, eline_trivial_usage, eline_full_usage);
}

int main(int argc, char **argv)
{
	libbb_init(argc, argv);
	return eline_main();
}
#endif  /* STANDALONE */
#endif  /* APPLET_USAGE */
