/*
 * a stripped-down version of gettext's envsubst(1)
 * Copyright (C) 2010 Bjarmason
 *
 * Vladimir Oleynik <dzo@simtreas.ru> (C) 2019
 *
 * Original code written by Bruno Haible <bruno@clisp.org>, 2003.
 *  Copyright (C) 2003-2007 Free Software Foundation, Inc.
 *
 *   GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 */

#ifdef APPLET_USAGE
#define envsubst_trivial_usage \
	"[-v] [SHELL-FORMAT]"
#define envsubst_full_usage \
	"Substitutes the values of environment variables.\n\n" \
	"Options:\n" \
	"\t-v\toutput the variables occurring in SHELL-FORMAT"
#else

#include "busybox.h"
#include <string.h>

struct VLIST {
	const char *name;
	size_t len;
	struct VLIST *next;
};

CALC_BSS struct BB_COMMON_BSS
{
	struct VLIST *sl_head;
	char *buffer;
	size_t bufmax;
	size_t buflen;
} bb_common_bss;
#define sl_head bb_common_bss.sl_head
#define buffer bb_common_bss.buffer
#define bufmax bb_common_bss.bufmax
#define buflen bb_common_bss.buflen

/* Print/Adds a variable to variables_set.  */
static void save_variable (const char *var_ptr, size_t var_len)
{
	struct VLIST *cur = xmalloc(sizeof(struct VLIST));

	cur->name = var_ptr;
	cur->len = var_len;
	cur->next = sl_head;
	sl_head = cur;
}

static void print_variable (const char *var_ptr, size_t var_len)
{
	fwrite (var_ptr, var_len, 1, stdout);
	putchar('\n');
}

/* Parse the string with $VARIABLE or ${VARIABLE}*/
static void find_variables (const char *string, void (*callback) (const char *var_ptr, size_t var_len))
{
	for (; *string != '\0';)
		if (*string++ == '$') {
			const char *variable_start;
			const char *variable_end;
			char c;

			if (*string == '{') /* } */
				string++;
			variable_start = string;
			c = *string;
			if ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') || c == '_') {
				int valid = c;
				do
					c = *++string;
				while ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') || (c >= '0' && c <= '9') || c == '_');
				variable_end = string;
				if (variable_start[-1] == '{') {        /* } */
				/* { */ if (*string == '}') {
						string++;
					} else
						valid = 0;
				}
				if (valid)
					(*callback)(variable_start, variable_end - variable_start);
			}
		}
}

/* Test whether a sorted string list contains a given string.  */
static const char *string_list_member (const char *s, size_t len)
{
	extern char **environ;
	char **e = environ;
	struct VLIST *l;

	if(sl_head != NULL) {
		for(l = sl_head; l != NULL; l = l->next) {
			if(l->len == len && !memcmp(l->name, s, len))
				break;
		}
		if(l == NULL)
			return NULL;
	}
	if(e != NULL)
		for(; *e != NULL; e++) {
			char *p = strchr(*e, '=');
			if(p != NULL) {
				if((p - *e) == len && !memcmp(*e, s, len))
					return p+1;
			}
		}
	return "";
}

static int push_getc(void)
{
	int c;

	c = getc (stdin);
	if (c == EOF) {
		if (ferror (stdin))
			bb_error_msg_and_die ("error while reading standard input");
	} else {
		if (buflen >= bufmax) {
			bufmax = 2 * bufmax + 32;
			buffer = xrealloc (buffer, bufmax);
		}
		buffer[buflen] = c;
	}
	buflen++;
	return c;
}

/* Copies stdin to stdout, performing substitutions. */
static void subst_from_stdin (void)
{
	int c;
	size_t ready = 0;

	for(;;) {
		const char *var = NULL;

		if(!ready) {
			buflen = 0;
			c = push_getc();
		}
		if (c == EOF)
			return;
		if(ready) {
			buffer[0] = c;
			buflen = 1;
		} else
			ready = 1;
		/* Look for $VARIABLE or ${VARIABLE}.  */
		if (c == '$') {
			c = push_getc();
			if (c == '{' /* } */)
				c = push_getc();
			if ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') || c == '_') {
				/* Accumulate the VARIABLE in buffer.  */
				do
					c = push_getc();
				while ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') || (c >= '0' && c <= '9') || c == '_');
				if (buffer[1] == '{' /* } */) {
				/* { */ if (c == '}') {
						ready = 0;
						var = buffer + 2;
					}
				} else
					var = buffer + 1;
				/* Test whether the variable shall be substituted.  */
				if (var != NULL)
					var = string_list_member(var, buflen + ready - 3);
			}
		} else
			ready = 0;
		if (var == NULL)
			fwrite (buffer, buflen - ready, 1, stdout);
		else
			fputs (var, stdout);
	}
}

#if ENABLE_FEATURE_LONG_OPTIONS
static const struct option env_long_options[] = {
	{ "variables", 1, NULL, 'v' },
	{ 0, 0, 0, 0 }
};
#endif

int envsubst_main()
{
	unsigned int flags;

#if ENABLE_FEATURE_LONG_OPTIONS
	bb_applet_long_options = env_long_options;
#endif
	flags = bb_xgetopt_ulflags("v\0?1");
	if(flags && *Gargv == NULL)
		bb_error_msg_and_die("missing argument");
	if(*Gargv != NULL)
		find_variables (*Gargv, flags ? print_variable : save_variable);
	if(!flags)
		subst_from_stdin();

	if (ferror (stdout) || fflush (stdout))
		return EXIT_FAILURE;
	if (fclose (stdout) && errno != EBADF)
		return EXIT_FAILURE;
	return 0;
}
#endif /* APPLET_USAGE */
