/*
 * Dave 'Kill a Cop' Cinege <dcinege@psychosis.com>
 * Licensed under the GPL v2 or later, see the file LICENSE in this tarball.
 *
 * makedevs
 * Make ranges of device files quickly.
 * known bugs: can't deal with alpha ranges
 */

#ifdef APPLET_USAGE
#ifdef CONFIG_FEATURE_MAKEDEVS_LEAF
#define makedevs_trivial_usage \
	"NAME TYPE MAJOR MINOR FIRST LAST [s]"
#define makedevs_full_usage \
	"Creates a range of block or character special files\n\n" \
	"TYPEs include:\n" \
	"\tb:\tMake a block (buffered) device.\n" \
	"\tc or u:\tMake a character (un-buffered) device.\n" \
	"\tp:\tMake a named pipe. MAJOR and MINOR are ignored for named pipes.\n\n" \
	"FIRST specifies the number appended to NAME to create the first device.\n" \
	"LAST specifies the number of the last item that should be created\n" \
	"If 's' is the last argument, the base device is created as well.\n\n" \
	"For example:\n" \
	"\tmakedevs /dev/ttyS c 4 66 2 63   ->  ttyS2-ttyS63\n" \
	"\tmakedevs /dev/hda b 3 0 0 8 s    ->  hda,hda1-hda8"
#endif

#ifdef CONFIG_FEATURE_MAKEDEVS_TABLE
#define makedevs_trivial_usage \
	"[-d device_table] rootdir"
#define makedevs_full_usage \
	"Creates a range of special files as specified in a device table.\n" \
	"Device table entries take the form of:\n" \
	"<type> <mode> <uid> <gid> <major> <minor> <start> <inc> <count>\n" \
	"Where name is the file name,  type can be one of:\n" \
	" f\tA regular file\n" \
	" d\tDirectory\n" \
	" c\tCharacter special device file\n" \
	" b\tBlock special device file\n" \
	" p\tFifo (named pipe)\n" \
	"uid is the user id for the target file, gid is the group id for the\n" \
	"target file. The rest of the entries (major, minor, etc) apply to\n" \
	"to device special files. A '-' may be used for blank entries."
#endif
#else

#include "busybox.h"
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/sysmacros.h>     /* major() and minor() */
#include "config_read.h"

#ifdef CONFIG_FEATURE_MAKEDEVS_LEAF
int makedevs_main(void)
{
	int argc = Gargc;
	char **argv = Gargv;
	mode_t mode;
	char *basedev, *type, *nodname, buf[255];
	int Smajor, Sminor, S, E;

	if (argc < 7 || *argv[1]=='-')
		bb_show_usage();

	basedev = argv[1];
	type = argv[2];
	Smajor = atoi(argv[3]);
	Sminor = atoi(argv[4]);
	S = atoi(argv[5]);
	E = atoi(argv[6]);
	nodname = argc == 8 ? basedev : buf;

	mode = 0660;

	switch (type[0]) {
	case 'c':
		mode |= S_IFCHR;
		break;
	case 'b':
		mode |= S_IFBLK;
		break;
	case 'f':
		mode |= S_IFIFO;
		break;
	default:
		bb_show_usage();
	}

	while (S <= E) {
		int sz;

		sz = snprintf(buf, sizeof(buf), "%s%d", basedev, S);
		if(sz<0 || sz>=sizeof(buf))  /* libc different */
			bb_error_msg_and_die("%s too large", basedev);

	/* if mode != S_IFCHR and != S_IFBLK third param in mknod() ignored */

		if (mknod(nodname, mode, makedev(Smajor, Sminor)))
			bb_error_msg("Failed to create: %s", nodname);

		if (nodname == basedev) /* ex. /dev/hda - to /dev/hda1 ... */
			nodname = buf;
		S++;
		Sminor++;
	}

	return 0;
}

#elif defined CONFIG_FEATURE_MAKEDEVS_TABLE
extern int makedevs_main(void)
{
	config_read_t *table;
	char *rootdir = NULL;
	char *line = NULL;
	int ret = EXIT_SUCCESS;
#define linenum table->num_line

	bb_getopt_ulflags("d:", &line);

	if (line) {
		table = config_read_prepare(line, 'e');
	} else {
		line = "<stdin>";
		table = config_read_prepare(line, '-');
	}

	if ((rootdir = *Gargv) == NULL)
		bb_error_msg_and_die("root directory not specified");

	xchdir(rootdir);

	umask(0);

	printf("rootdir=%s\n", rootdir);
	printf("table='%s'\n", line);

	while ((table = config_read(table, "#\1\2\3\4\5\6\7\10\11\12\13\0\0")) != NULL) {
		char type;
		unsigned int mode = 0755;
		unsigned int device_major = 0;
		unsigned int device_minor = 0;
		unsigned int count = 0;
		unsigned int increment = 0;
		unsigned int start = 0;
#define name table->tokens[0]
		char *full_name;
		uid_t uid;
		gid_t gid;
		unsigned int *tokens_6_10[] = { &device_major, &device_minor, &start, &increment, &count };
		int i;

		if(table->tokens[10] != NULL || table->tokens[1] == NULL || table->tokens[1][1] != '\0')
			goto err;
		type = table->tokens[1][0];
		if(table->tokens[2]) {
			if(sscanf(table->tokens[2], "%o", &mode) != 1)
				goto err;
		}
		uid = table->tokens[3] ? get_ug_id(table->tokens[3], bb_xgetpwnam) : getuid();
		gid = table->tokens[4] ? get_ug_id(table->tokens[4], bb_xgetgrnam) : getgid();
		for(i = 5; i < 10; i++) {
			if(table->tokens[i]) {
				if(sscanf(table->tokens[i], "%u", tokens_6_10[i - 5]) != 1)
					goto err;
			}
		}
		if ((device_major | device_minor | start | count | increment) > 255) {
		  err:
			bb_error_msg("line %d invalid", linenum);
			ret = EXIT_FAILURE;
			continue;
		}

		full_name = concat_path_file(rootdir, name);

		if (type == 'd') {
			bb_make_directory(full_name, mode | S_IFDIR, FILEUTILS_RECUR);
			if (chown(full_name, uid, gid) == -1) {
				bb_perror_msg("line %d: chown failed for %s", linenum, full_name);
				ret = EXIT_FAILURE;
				goto loop;
			}
			if ((mode != -1) && (chmod(full_name, mode) < 0)){
				bb_perror_msg("line %d: chmod failed for %s", linenum, full_name);
				ret = EXIT_FAILURE;
				goto loop;
			}
		} else if (type == 'f') {
			struct stat st;
			if ((stat(full_name, &st) < 0 || !S_ISREG(st.st_mode))) {
				bb_perror_msg("line %d: regular file '%s' does not exist", linenum, full_name);
				ret = EXIT_FAILURE;
				goto loop;
			}
			if (chown(full_name, uid, gid) == -1) {
				bb_perror_msg("line %d: chown failed for %s", linenum, full_name);
				ret = EXIT_FAILURE;
				goto loop;
			}
			if ((mode != -1) && (chmod(full_name, mode) < 0)){
				bb_perror_msg("line %d: chmod failed for %s", linenum, full_name);
				ret = EXIT_FAILURE;
				goto loop;
			}
		} else {
			dev_t rdev;

			if (type == 'p') {
				mode |= S_IFIFO;
			} else if (type == 'c') {
				mode |= S_IFCHR;
			} else if (type == 'b') {
				mode |= S_IFBLK;
			} else {
				bb_error_msg("line %d: unsupported file type %c", linenum, type);
				ret = EXIT_FAILURE;
				goto loop;
			}

			if (count > 0) {
				char *full_name_inc;

				full_name_inc = xmalloc(strlen(full_name) + 4);
				for (i = start; i < count; i++) {
					sprintf(full_name_inc, "%s%d", full_name, i);
					rdev = (device_major << 8) + device_minor + (i * increment - start);
					if (mknod(full_name_inc, mode, rdev) == -1) {
						bb_perror_msg("line %d: could not create node %s", linenum, full_name_inc);
						ret = EXIT_FAILURE;
					}
					else if (chown(full_name_inc, uid, gid) == -1) {
						bb_perror_msg("line %d: chown failed for %s", linenum, full_name_inc);
						ret = EXIT_FAILURE;
					}
					if ((mode != -1) && (chmod(full_name_inc, mode) < 0)){
						bb_perror_msg("line %d: chmod failed for %s", linenum, full_name_inc);
						ret = EXIT_FAILURE;
					}
				}
				free(full_name_inc);
			} else {
				rdev = (device_major << 8) + device_minor;
				if (mknod(full_name, mode, rdev) == -1) {
					bb_perror_msg("line %d: could not create node %s", linenum, full_name);
					ret = EXIT_FAILURE;
				}
				else if (chown(full_name, uid, gid) == -1) {
					bb_perror_msg("line %d: chown failed for %s", linenum, full_name);
					ret = EXIT_FAILURE;
				}
				if ((mode != -1) && (chmod(full_name, mode) < 0)){
					bb_perror_msg("line %d: chmod failed for %s", linenum, full_name);
					ret = EXIT_FAILURE;
				}
			}
		}
loop:
		free(full_name);
	}

	return ret;
}

#else
# error makedevs configuration error, either leaf or table must be selected
#endif
#endif /* APPLET_USAGE */
