/*-------------------------------------------------------------------------
 * Filename:      xmodem.c
 * Copyright:     Copyright (C) 2001, Hewlett-Packard Company
 * Author:        Christopher Hoover <ch@hpl.hp.com>
 * Description:   xmodem functionality for uploading of kernels
 *                and the like
 * Created at:    Thu Dec 20 01:58:08 PST 2001
 *-----------------------------------------------------------------------*/
/*
 * xmodem.c: xmodem functionality for uploading of kernels and
 *            the like
 *
 * Copyright (C) 2001 Hewlett-Packard Laboratories
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 *
 * This was originally written for blob and then adapted for busybox.
 *
 */
#ifdef APPLET_USAGE
#define rx_trivial_usage \
	"FILE"
#define rx_full_usage \
	"Receive a file using the xmodem protocol."
#else

#include "busybox.h"
#include <unistd.h>
#include <termios.h>
#include <string.h>
#include "bb_xopen.h"
#include "bb_sig.h"

#define SOH 0x01
#define STX 0x02
#define EOT 0x04
#define ACK 0x06
#define NAK 0x15
#define CAN 0x18
#define BS  0x08

#ifndef HAVE_TCDRAIN
/* Wait for pending output to be written on FD.  */
static inline int tcdrain (int fd)
{
      /* With an argument of 1, TCSBRK waits for the output to drain.  */
      return ioctl(fd, TCSBRK, 1);
}
#endif

/*

Cf:

  http://www.textfiles.com/apple/xmodem
  http://www.phys.washington.edu/~belonis/xmodem/docxmodem.txt
  http://www.phys.washington.edu/~belonis/xmodem/docymodem.txt
  http://www.phys.washington.edu/~belonis/xmodem/modmprot.col

*/

#define TIMEOUT 1
#define TIMEOUT_LONG 10
#define MAXERRORS 10

static inline void write_byte(int fd, char cc) {
	write(fd, &cc, 1);
}

static inline void write_flush(int fd) {
	tcdrain(fd);
}

static inline void read_flush(int fd) {
	tcflush(fd, TCIFLUSH);
}

static int read_byte(int fd, unsigned int timeout) {
	char buf[1];
	int n;

	alarm(timeout);

	n = read(fd, &buf, 1);

	alarm(0);

	if (n == 1)
		return buf[0] & 0xff;
	else
		return -1;
}

CALC_BSS struct BB_COMMON_BSS
{
	char error_buf[256];
	struct termios orig_tty;
	int ttyfd;
} bb_common_bss;
#define error_buf bb_common_bss.error_buf
#define orig_tty bb_common_bss.orig_tty
#define ttyfd bb_common_bss.ttyfd

static int receive(int filefd)
{
	int wttyfd = ttyfd;
	char blockBuf[1024];
	unsigned int errors = 0;
	unsigned int wantBlockNo = 1;
	unsigned int length = 0;
	int docrc = 1;
	char nak = 'C';
	unsigned int timeout = TIMEOUT_LONG;

#define note_error(fmt,args...) \
	snprintf(error_buf, sizeof(error_buf), fmt,##args)

	read_flush(wttyfd);

	/* Ask for CRC; if we get errors, we will go with checksum */
	write_byte(wttyfd, nak);
	write_flush(wttyfd);

	for (;;) {
		int blockBegin;
		int blockNo, blockNoOnesCompl;
		int blockLength;
		int cksum = 0;
		int crcHi = 0;
		int crcLo = 0;

		blockBegin = read_byte(wttyfd, timeout);
		if (blockBegin < 0)
			goto timeout;

		timeout = TIMEOUT;
		nak = NAK;

		switch (blockBegin) {
		case SOH:
		case STX:
			break;

		case EOT:
			write_byte(wttyfd, ACK);
			write_flush(wttyfd);
			goto done;

		default:
			goto error;
		}

		/* block no */
		blockNo = read_byte(wttyfd, TIMEOUT);
		if (blockNo < 0)
			goto timeout;

		/* block no one's compliment */
		blockNoOnesCompl = read_byte(wttyfd, TIMEOUT);
		if (blockNoOnesCompl < 0)
			goto timeout;

		if (blockNo != (255 - blockNoOnesCompl)) {
			note_error("bad block ones compl");
			goto error;
		}

		blockLength = (blockBegin == SOH) ? 128 : 1024;

		{
			int i;

			for (i = 0; i < blockLength; i++) {
				int cc = read_byte(wttyfd, TIMEOUT);
				if (cc < 0)
					goto timeout;
				blockBuf[i] = cc;
			}
		}

		if (docrc) {
			crcHi = read_byte(wttyfd, TIMEOUT);
			if (crcHi < 0)
				goto timeout;

			crcLo = read_byte(wttyfd, TIMEOUT);
			if (crcLo < 0)
				goto timeout;
		} else {
			cksum = read_byte(wttyfd, TIMEOUT);
			if (cksum < 0)
				goto timeout;
		}

		if (blockNo == ((wantBlockNo - 1) & 0xff)) {
			/* a repeat of the last block is ok, just ignore it. */
			/* this also ignores the initial block 0 which is */
			/* meta data. */
			goto next;
		} else if (blockNo != (wantBlockNo & 0xff)) {
			note_error("unexpected block no, 0x%08x, expecting 0x%08x", blockNo, wantBlockNo);
			goto error;
		}

		if (docrc) {
			int crc = 0;
			int i, j;
			int expectedCrcHi;
			int expectedCrcLo;

			for (i = 0; i < blockLength; i++) {
				crc = crc ^ (int) blockBuf[i] << 8;
				for (j = 0; j < 8; j++)
					if (crc & 0x8000)
						crc = crc << 1 ^ 0x1021;
					else
						crc = crc << 1;
			}

			expectedCrcHi = (crc >> 8) & 0xff;
			expectedCrcLo = crc & 0xff;

			if ((crcHi != expectedCrcHi) ||
			    (crcLo != expectedCrcLo)) {
				note_error("crc error, expected 0x%02x 0x%02x, got 0x%02x 0x%02x", expectedCrcHi, expectedCrcLo, crcHi, crcLo);
				goto error;
			}
		} else {
			unsigned char expectedCksum = 0;
			int i;

			for (i = 0; i < blockLength; i++)
				expectedCksum += blockBuf[i];

			if (cksum != expectedCksum) {
				note_error("checksum error, expected 0x%02x, got 0x%02x", expectedCksum, cksum);
				goto error;
			}
		}

		wantBlockNo++;
		length += blockLength;

		if (bb_full_write(filefd, blockBuf, blockLength) < 0) {
			note_error("write to file failed: %m");
			goto fatal;
		}

	next:
		errors = 0;
		write_byte(wttyfd, ACK);
		write_flush(wttyfd);
		continue;

	error:
	timeout:
		errors++;
		if (errors == MAXERRORS) {
			/* Abort */
			int i;

			// if using crc, try again w/o crc
			if (nak == 'C') {
				nak = NAK;
				errors = 0;
				docrc = 0;
				goto timeout;
			}

			note_error("too many errors; giving up");

		fatal:
			for (i = 0; i < 5; i ++)
				write_byte(wttyfd, CAN);
			for (i = 0; i < 5; i ++)
				write_byte(wttyfd, BS);
			write_flush(wttyfd);
			return -1;
		}

		read_flush(wttyfd);
		write_byte(wttyfd, nak);
		write_flush(wttyfd);
	}

 done:
	return length;

#undef note_error
}

static void sigalrm_handler(int ATTRIBUTE_UNUSED signum)
{
	tcsetattr(ttyfd, TCSAFLUSH, &orig_tty);
}


int rx_main(void)
{
	int filefd;
	struct termios tty;
	int n;

	bb_xgetopt_ulflags("\0=1");

	ttyfd = bb_xopen(bb_dev_tty, O_RDWR|BB_OPEN_ERR);

	filefd = bb_xopen(*Gargv, O_RDWR|O_CREAT|O_TRUNC|BB_OPEN_ERR, 0666);

	if (tcgetattr(ttyfd, &orig_tty) < 0)
		bb_perror_msg_and_die("tcgetattr failed");

	tty = orig_tty;
	bb_signal( (1 << SIGALRM), sigalrm_handler);

	cfmakeraw(&tty);
	tcsetattr(ttyfd, TCSAFLUSH, &tty);

	n = receive(filefd);

	close(filefd);

	sigalrm_handler(0);     /* tcsetattr(ttyfd, TCSAFLUSH, &orig_tty); */

	if (n < 0) {
		putc('\n', stderr);
		bb_error_msg_and_die("receive failed: %s", error_buf);
	}
	bb_fflush_stdout_and_exit(EXIT_SUCCESS);
}
#endif /* APPLET_USAGE */
