/*
 * Secure Locate v3.1
 * Author: Kevin Lindsay
 * Copyright (c) 2005, 2006 Kevin Lindsay
 *
 * Modified by Vladimir Oleynik <dzo@simtreas.ru> (c) 2010
 * - very simplify code and size optimizations;
 * - to be used in busybox;
 * - can use for HP UX (do not need any other dependences library)
 * - rewrite without fts_*()
 *
 */

/*
 *
 * Secure Locate -- search database for filenames that match patterns without
 *                  showing files that the user using slocate does not have
 *                  access to.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2, or (at your option)
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 */
#ifdef APPLET_USAGE
#define slocate_trivial_usage \
	"search_string|-r regexp [-n num] [-iq] [-d db]"
#define slocate_full_usage \
	"Options:\n" \
	"\t-n num\tLimit the amount of results shown to <num>.\n" \
	"\t-i\tDoes a case insensitive search.\n" \
	"\t-q\tQuiet mode.  Error messages are suppressed.\n" \
	"\t-d db\tSpecfies the path of databases to search in."

#define updatedb_trivial_usage \
	"[-qv] [-o db] [-e dir1,...] [-f fs_type1,... ]\n" \
	"[-c file] [-u]|[-U path] [-l 0|1]"
#define updatedb_full_usage \
	"Options:\n" \
	"\t-u\tCreate slocate database starting at path /\n" \
	"\t-U path\tCreate slocate database starting at path\n" \
	"\t-c file\tParse original GNU Locate's configuration file\n" \
	"\t-e dir1,...\tExclude directories from the slocate database\n" \
	"\t-f fs_type1,...\tExclude file system types from the slocate database\n" \
	"\t-l level\tSecurity level. 0 - off, 1 - on (default)\n" \
	"\t-o db\tSpecifies the database to create.\n" \
	"\t-q\tQuiet mode.  Error messages are suppressed.\n" \
	"\t-v\tVerbose mode. Display files when creating database."
#else

#define VERSION "3.1"
#define SL_RELEASE "March 7, 2006"
#define SL_VERSION "Secure Locate " VERSION " - Released " SL_RELEASE

#define __USE_GNU
//#define _GNU_SOURCE

#include "busybox.h"
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <fnmatch.h>
#include <time.h>
#include <dirent.h>
#include "xregex.h"

#ifdef RAND_MAX
# undef RAND_MAX
#endif

#define RAND_MAX 9999

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE 1
#endif

/* Warn if a database is older than this.  8 days allows for a weekly
 *    update that takes up to a day to perform.  */
#define WARN_SECONDS (60 * 60 * 24 * 8)

/* Printable version of WARN_SECONDS.  */
#define WARN_MESSAGE "8 days"

#ifdef __hpux
#define MTAB_FILE "/etc/mnttab"
# define DEFAULT_DB "/var/opt/slocate/slocate.db"
# define DEFAULT_DB_DIR "/var/opt/slocate/"
#else
#define MTAB_FILE "/etc/mtab"
# define DEFAULT_DB "/var/lib/slocate/slocate.db"
# define DEFAULT_DB_DIR "/var/lib/slocate/"
#endif

#define UPDATEDB_FILE "/etc/updatedb.conf"

#define DB_UID 0
#define DB_GROUP "slocate"
#define DB_MODE 00640

#define SLOC_ESC -0x80

/* Number of bytes to read in at a time from DB when searching */
#define BLOCK_SIZE 4096

#define GRANT_ACCESS '0'
#define VERIFY_ACCESS '1'

CALC_BSS struct BB_COMMON_BSS
{
	int QUIET;               /* default FALSE */
	int VERBOSE;             /* default FALSE */
	int nocase;              /* default FALSE */
	char *index_path;
	uid_t uid;
	gid_t gid;
	gid_t SLOCATE_GID;
	char *output_db;
	char **exclude;
	char **input_db;
	char *reg_pattern;
	regex_t reg;
	int updatedb;
	char **cmd_search_str;
	char *cmd_updatedb_conf;
	int queries;
	char slevel;
} bb_common_bss;
#define QUIET bb_common_bss.QUIET
#define VERBOSE bb_common_bss.VERBOSE
#define VERBOSE bb_common_bss.VERBOSE
#define nocase bb_common_bss.nocase
#define index_path bb_common_bss.index_path
#define uid bb_common_bss.uid
#define gid bb_common_bss.gid
#define SLOCATE_GID bb_common_bss.SLOCATE_GID
#define output_db bb_common_bss.output_db
#define exclude bb_common_bss.exclude
#define input_db bb_common_bss.input_db
#define reg_pattern bb_common_bss.reg_pattern
#define reg bb_common_bss.reg
#define updatedb bb_common_bss.updatedb
#define cmd_search_str bb_common_bss.cmd_search_str
#define cmd_updatedb_conf bb_common_bss.cmd_updatedb_conf
#define queries bb_common_bss.queries
#define slevel bb_common_bss.slevel


/* Encoding data */
struct enc_data_s {
	char *prev_line;
	short prev_len;
};


/* Decode states */
#define DC_NONE 0
#define DC_CODE 1
#define DC_ESC 2
#define DC_ESC_INTR 3
#define DC_DATA 4
#define DC_DATA_INTR 5

static int my_isspace(char c)
{
       return c == ' ' || c == '\t' || c == '\r' || c == '\n' || c == '\v'
	       || c == '\f';
}


/* Duplicate string and make all characters lowercase */
#ifndef FNM_CASEFOLD
static char * tolower_strdup(const char *str)
{
	char *nocase_str = xstrdup(str);
	str_tolower(nocase_str);
	return nocase_str;
}
#endif

/* Check if the current path matches the search criteria */
static int
match(char *full_path, char *search_str, int globflag)
{
	int foundit = 0;
	int ret = 0;
	int nmatch = 32;
	regmatch_t pmatch[32];
#ifndef FNM_CASEFOLD
	char *nocase_str = NULL;
	char *nocase_path = NULL;
#endif

	/* If searching with regular expressions */
	if (reg_pattern != NULL) {
		foundit = ! regexec(&reg, full_path, nmatch, pmatch, 0);
	/* Case sensitive search */
	} else if (search_str && !nocase) {
		if (globflag) {
			foundit = fnmatch(search_str, full_path, 0);

			if (foundit == 0)
			    foundit = 1;
			else if (foundit == FNM_NOMATCH)
			    foundit = 0;
			else {
				bb_error_msg_and_die("match: fnmatch: unknown error.");
			}
		} else
		    foundit=(strstr(full_path, search_str) != NULL);
	/* Case insensitive search */
	} else if (search_str && nocase) {
#ifndef FNM_CASEFOLD
		nocase_str = tolower_strdup(search_str);

		nocase_path = tolower_strdup(full_path);

		if (globflag) {
			foundit = fnmatch(nocase_str,nocase_path, 0);

			if (foundit == 0)
			    foundit = 1;
			else if (foundit == FNM_NOMATCH)
			    foundit = 0;
			else {
				bb_error_msg_and_die("match: not FNM_CASEFOLD: fnmatch: nocase: unknown error.");
			}

		}
		else if (strstr(nocase_path, nocase_str))
		    foundit = 1;

#else /* FNM_CASEFOLD */
		if (globflag) {
			foundit = fnmatch(search_str, full_path, FNM_CASEFOLD);

			if (foundit == 0)
			    foundit = 1;
			else if (foundit == FNM_NOMATCH)
			    foundit = 0;
			else {
				bb_error_msg_and_die("match: FNM_CASEFOLD: fnmatch: nocase: unknown error.");
			}

		} else
		    foundit = (strcasestr(search_str, search_str) != NULL);

#endif /* FNM_CASEFOLD */
	}

	if (foundit)
	    ret = 1;

#ifndef FNM_CASEFOLD
	free(nocase_str);
	free(nocase_path);
#endif
	return ret;
}

/* Match exclude
 *
 * 1  == match
 * 0  == no match
 * -1 == error
 */
static int match_exclude(char *path)
{
	int i;

	if (!exclude || !path)
	    return 0;

	/* Compare inode numbers to check if the exclude path matches */
	for (i = 0; exclude[i]; i++) {
		if (strcmp(path, exclude[i]) == 0) {
			if (VERBOSE)
			    printf("Excluding: %s\n", path);
			return 1;
		}
	}

	return 0;
}

/* Get the GID for group slocate */
static unsigned short
get_gid(const char *group)
{
	struct group *grpres = NULL;

	if ((grpres = getgrnam(group)) == NULL)
		bb_error_msg_and_die("This is a result of the group missing or a corrupted group file.");

	return (unsigned short) grpres->gr_gid;
}

/* Verify DB
 * For now we just check if there is a '1' or '0' as the first character
 * of the DB. (security level).
 *
 * Returns:  0 == Invalid
 *          -1 == File does not exist or inaccessible
 *           1 == Valid
 */
static int
verify_slocate_db(char *file)
{
	char ch[1];
	struct stat tf_stat;
	int bytes = 0;
	int fd = -1;

	if (access(file, W_OK | R_OK) == 0) {
		if (stat(file, &tf_stat) == -1)
			bb_perror_msg_and_die("get_temp_file: stat(): %s", file);

		if (tf_stat.st_size == 0)
		    return 0;

		if ((fd = open(file, O_RDONLY)) == -1)
			bb_perror_msg_and_die("get_temp_file: open(): %s", file);

		bytes = read(fd, ch, 1);

		if (close(fd) == -1)
			bb_perror_msg_and_die("get_temp_file: close(): %s", file);

		if (bytes == 0)
			bb_error_msg_and_die("Could not read from file: %s", file);

		if (ch[0] != '1' && ch[0] != '0')
		    return 0;

	} else
	    return -2;

	return 1;
}

/* Get a temp filename which doesn't exist */
static char *
get_temp_file(void)
{
	char *tmp_file;

	/* .stf == Slocate Temporary File */
	tmp_file = xasprintf("%s.stf", output_db);

	if (verify_slocate_db(tmp_file) == 0)
		bb_error_msg_and_die("The temp file '%s' already exists and does not appear to be a valid slocate database. Please remove before creating the database.", tmp_file);

	return tmp_file;
}

/* Check if a path/file is writeable */
static int
access_path(char *path)
{
	char *ptr;
	int ret;

	if (!path)
	    return 0;

	if (!(ptr = strrchr(path, '/')))
	    return 0;

	ret = (ptr - path) + 1;
	path = xstrdup(path);
	path[ret] = '\0';

	ret = access(path, X_OK | W_OK);
	free(path);

	if (ret == 0)
	    return 1;

	return 0;
}


static void
load_file(const char *filename, char **file_data)
{
	int fd;
	struct stat fbuf_stat;

	*file_data = NULL;

	if (!filename)
		bb_error_msg_and_die("load_file: filename == NULL.");

	if ((fd = open(filename, O_RDONLY)) == -1)
		bb_perror_msg_and_die("load_file: Could not open file: %s", filename);

	if (fstat(fd, &fbuf_stat) == -1)
		bb_perror_msg_and_die("load_file: Could not stat file: %s", filename);

	*file_data = xmalloc(sizeof(char) * (fbuf_stat.st_size+1));

	if (read(fd, *file_data, fbuf_stat.st_size) != fbuf_stat.st_size)
		bb_error_msg_and_die("load_file: read: Failed to read all %lld bytes from %s.", (long long)fbuf_stat.st_size, filename);

	(*file_data)[fbuf_stat.st_size] = 0;
	close(fd);
}

/* Verify access to the file. access() follows symlinks, so we need
 * to check them separately */
static int verify_access(const char *path)
{
	struct stat path_stat;
	int ret = 0;
	char *ptr = NULL;

	if (lstat(path, &path_stat) == -1)
	    goto EXIT;

	if (!S_ISLNK(path_stat.st_mode)) {
		if (access(path, F_OK) != 0)
		    goto EXIT;
	} else if ((ptr = strrchr(path, '/'))) {
		*ptr = 0;
		if (access(path, F_OK) == 0)
		    ret = 1;
		*ptr = '/';
		goto EXIT;
	}

	ret = 1;
EXIT:
	return ret;
}


/* Parse a comma delimited string of files and dirs into an array of pointers.
 *
 * string == "file,dir,...\0"
 */
/* TODO: Clean up returns. use ret. */
static void parse_exclude(char *estr)
{
	char *token = NULL;
	char *ptr1 = NULL;
	char *ptr2 = NULL;
	int len = 0;

	if (!estr) {
		bb_error_msg("parse_exclude: String passed is NULL.");
		return;
	}

	if (strlen(estr) == 0)
	    return;

	/* Get the array length */
	for (len = 0; exclude && exclude[len]; len++);

	ptr1 = estr;

	/* Loop until we run out of string to parse.
	 * Loop ends via break; */
	while (1) {
		/* If *ptr2 == 0 then we have already parsed the last token */
		if (ptr2 && *ptr2 == 0)
		    break;

		/* Move start pointer to start of next token.
		 * If ptr2 == NULL then this is the first token and we havn't
		 * parsed it yet. */
		if (ptr2)
		    ptr1 = ptr2+1;

		/* Get first token postion.
		 * If no delimeter, most ptr2 to the end of the string. */
		if (!(ptr2 = strchr(ptr1, ','))) {
			ptr2 = ptr1;
			while(*ptr2 != 0)
			    ptr2 += 1;
		}

		/* Make token */
		token = xstrndup(ptr1, ptr2-ptr1);

		/* Make sure the file exists and is accessible, otherwise ignore it */
		if (access(token, F_OK) != 0) {
			if (errno != EACCES && errno != ENOENT)
				bb_perror_msg_and_die("parse_exclude: access: '%s'", token);

			free(token);
			token = NULL;

			continue;
		}

		/* Increment the length */
		len += 1;

		/* Allocate memory */
		exclude = xrealloc(exclude, sizeof(char *) * (len+1));

		/* Assign the token to the array */
		exclude[len] = NULL;
		exclude[len-1] = xstrdup(token);
		free(token);
		token = NULL;

	}
}

/* Init Input DB variable */
static char **init_input_db(int len)
{
	int i = 0;
	int cur_len = 0;

	if (len < 0)
		bb_error_msg_and_die("init_input_db: Initialization length must be 0 or greater.");

	if (input_db == NULL) {
		input_db = xmalloc(sizeof(char *)*(len+1));
	} else {
		cur_len = 0;
		while (input_db[cur_len])
		    cur_len += 1;

		input_db = xrealloc(input_db, sizeof(char *)*(cur_len+len+1));
	}

	/* Initialize new space */
	for (i = cur_len; i <= cur_len+len; i += 1)
	    input_db[i] = NULL;

	return input_db;
}

/* Parse Database Paths.
 *
 * Concatenate all database paths into one string and validate each database.
 *
 */
static void parse_userdb(char *dblist)
{
	char *tmp_ptr = NULL;
	struct stat db_stat;
	int last_sgid = 0;
	char *ptr1 = NULL;
	char *ptr2 = NULL;
	int i_pos = 0;
	int list_len = 0;
	gid_t db_gid = -1;
	int duplicate = 0;
	int d = 0;

	/* Make sure dblist is not empty */
	if (!dblist || strlen(dblist) == 0) {
		//bb_error_msg("parse_userdb: database string is empty.");
		return;
	}

	/* Check how many paths are currently in the string. */
	list_len = 1;
	ptr1 = dblist;
	while ((ptr1 = strchr(ptr1+1, ':')))
	    list_len += 1;

	input_db = init_input_db(list_len);

	/* Find next position in input_db */
	i_pos = 0;
	while (input_db[i_pos])
	    i_pos += 1;

	/* Parse dblist */
	ptr1 = dblist;

	while (ptr1 && *ptr1) {
		if (*ptr1 == ':')
		    ptr1 += 1;

		if (!(ptr2 = strchr(ptr1, ':'))) {
			ptr2 = dblist+strlen(dblist);
		}

		/* Don't worry about blank entries */
		if (ptr2-ptr1 > 0) {
			/* Check for duplicates */
			duplicate = 0;
			for (d = 0; input_db[d]; d += 1) {
				if (!strncmp(input_db[d], ptr1, ptr2-ptr1)) {
					duplicate = 1;
					break;
				}
			}
			if (!duplicate) {
				input_db[i_pos] = xstrndup(ptr1, ptr2-ptr1);
				i_pos += 1;
			}
		}

		if (*ptr2 == ':') {
			ptr1 = ptr2 + 1;
		} else
		    ptr1 = NULL;
	}


	db_gid = get_gid(DB_GROUP);

	last_sgid = 0;
	/* Sort sgid slocate db's to the top */
	for (i_pos = 0; input_db[i_pos]; i_pos += 1) {

		if (stat(input_db[i_pos], &db_stat) == -1)
			bb_perror_msg_and_die("Could not find user database '%s'", input_db[i_pos]);

		if (db_stat.st_gid == db_gid) {

			if (i_pos != last_sgid) {
				tmp_ptr = input_db[last_sgid];
				input_db[last_sgid] = input_db[i_pos];
				input_db[i_pos] = tmp_ptr;
			}

			last_sgid += 1;
		}

	}
}

/* Parse File System Type Exclusion */
static void parse_fs_exclude(char *data_str)
{
	char *fs_str;
	size_t exclude_str_len = 0;
	char *fbuf=NULL;
	char *head_ptr;
	char *tail_ptr;
	char *exclude_str=NULL;
	int i;
	int matched;
	char *match_ptr;

	if (!data_str)
		bb_error_msg_and_die("parse_fs_exclude: data_str == NULL");

	/* Duplicate so that we can change the case */
	fs_str = xstrdup(data_str);

	for (i = 0; fs_str[i]; i++)
	    fs_str[i] = toupper(fs_str[i]);

	/* Load the mtab file */
	load_file(MTAB_FILE, &fbuf);

	head_ptr = fbuf;
	while (head_ptr) {
		/* find filesystem type */
		if ((head_ptr = strchr(head_ptr,' '))) {
			head_ptr += 1;
			head_ptr = strchr(head_ptr,' ');
		}

		if (!head_ptr)
		    continue;

		head_ptr += 1;

		tail_ptr = strchr(head_ptr,' ');
		if (!tail_ptr) {
			head_ptr = NULL;
			continue;
		}

		*tail_ptr = 0;
		/* Upper case string */
		for (i = 0; head_ptr[i]; i++)
		    head_ptr[i] = toupper(head_ptr[i]);

		/* Check if file sytem type exists in exclude string */
		matched = 0;
		if ((match_ptr = strstr(fs_str, head_ptr))) {
			if ((match_ptr == fs_str || *(match_ptr-1) == ',') &&
			    (*(match_ptr+(tail_ptr-head_ptr)) == ',' || *(match_ptr+(tail_ptr-head_ptr)) == '\0'))
			    matched = 1;
		}

		*tail_ptr = ' ';

		if (matched) {
			/* go backwards a bit so that we can get the
			 * mount point of the filesystem */
			head_ptr -= 2;

			while (*head_ptr != ' ' && head_ptr != fbuf)
			    head_ptr -= 1;

			if (head_ptr == fbuf)
				bb_error_msg_and_die("parse_fs_exclude: File System Exclude: (1) corrupt mtab file: %s", MTAB_FILE);

			head_ptr += 1;

			if (!(tail_ptr = strchr(head_ptr,' ')))
				bb_error_msg_and_die("parse_fs_exclude: File System Exclude: (2) corrupt mtab file: %s", MTAB_FILE);

			*tail_ptr = 0;

			/* +1 for the extra "," delimiter */
			exclude_str_len += strlen(head_ptr)+1;
			exclude_str = xrealloc(exclude_str, exclude_str_len+1);

			if (exclude_str_len == strlen(head_ptr)+1)
			    *exclude_str = 0;
			else
			    strcat(exclude_str, ",");
			strcat(exclude_str, head_ptr);

			*tail_ptr = ' ';
		}


		head_ptr = strchr(head_ptr,'\n');
	}

	if (exclude_str)
		parse_exclude(exclude_str);

	free(exclude_str);
	free(fbuf);
	free(fs_str);
}



/* updatedb.conf parsing */
static char * get_next_char(char ch, char *ptr)
{
	while (*ptr && *ptr != ch && *ptr != '\n')
	    ptr++;

	if (*ptr != ch)
	    return NULL;

	return ptr;
}

static char * get_value(char *key, char *conf_data)
{
	char *ptr = NULL;
	char *val = NULL;
	char *quote_ptr = NULL;
	char *end_ptr = NULL;
	int i;
	size_t key_len = strlen(key);

	ptr = conf_data;

	/* Find a valid key */
	while ((ptr = strstr(ptr, key))) {

		/* Make sure we have a proper key */
		if (strlen(ptr) >= key_len) {
			if (!my_isspace(ptr[key_len]) && ptr[key_len] != '=') {
				ptr++;
				continue;
			}
		}

		/* Check if we are not commented out */
		for (i = 0; *(ptr-i) && (ptr-i) != conf_data; i++) {
			if (*(ptr-i) == '\n')
			    break;
			if (*(ptr-i) == '#')
			    break;
		}

		if (*(ptr-i) == '\n' || (ptr-i) == conf_data)
		    break;

		ptr++;
	}

	if (!ptr)
	    goto EXIT;

	/* Parse the valid key */
	if (!(ptr = get_next_char('=', ptr))) {
		bb_error_msg("%s: syntax error at '%s': Could not find '=' character.", UPDATEDB_FILE, key);
		goto EXIT;
	}

	if (!(quote_ptr = get_next_char('"', ptr))) {
		ptr++;
		while (*ptr && my_isspace(*ptr) && *ptr != '\n')
		    ptr++;

		if (!*ptr || *ptr == '\n') {
			bb_error_msg("%s: syntax error at '%s': No value found.", UPDATEDB_FILE, key);
			goto EXIT;
		}
	} else
	    ptr = quote_ptr+1;

	end_ptr = ptr;

	if (quote_ptr) {
		if (!(end_ptr = get_next_char('"', end_ptr))) {
			bb_error_msg("%s: syntax error at '%s': Missing second \".", UPDATEDB_FILE, key);
			goto EXIT;
		}
	} else {
		while (*end_ptr && !my_isspace(*end_ptr) && *end_ptr != '\n' )
		    end_ptr++;
	}

	if (!(end_ptr-ptr)) {
		bb_error_msg("%s: syntax error at '%s': No value found.", UPDATEDB_FILE, key);
		goto EXIT;
	}

	if (!(val = strndup(ptr, end_ptr-ptr)))
	    goto EXIT;

EXIT:

	return val;
}

static int
parse_prune(char *key, char *conf_data, const char *delim, char **data_str)
{
	char *val = NULL;
	char *ptr = NULL;
	char *end_ptr = NULL;

	if (!(val = get_value(key, conf_data)))
	    return 0;

	*data_str = xmalloc(sizeof(char) * (strlen(val)+1));
	**data_str = 0;

	ptr = val;

	/* replace all spaces with ','s, but only the
	   spaces between words and one space only */
	while (*ptr) {
		while (*ptr && my_isspace(*ptr))
		    ptr++;

		if (!*ptr)
		    break;

		end_ptr = ptr;

		while (*end_ptr && !my_isspace(*end_ptr))
		    end_ptr++;

		if (ptr != val)
		    *data_str = strcat(*data_str, delim);
		*data_str = strncat(*data_str, ptr, end_ptr-ptr);

		ptr = end_ptr;
	}

	return 1;
}

static int parse_PRUNEFS(char *conf_data)
{
	char *fs_str = NULL;
	int ret = 0;

	if (!parse_prune("PRUNEFS", conf_data, ",", &fs_str))
	    goto EXIT;

	parse_fs_exclude(fs_str);

	ret = 1;

EXIT:

	free(fs_str);
	return ret;
}

static int parse_PRUNEPATHS(char *conf_data)
{
	char *path_str = NULL;
	int ret = 0;

	if (!parse_prune("PRUNEPATHS", conf_data, ",", &path_str))
	    goto EXIT;

	parse_exclude(path_str);

	ret = 1;
EXIT:
	free(path_str);
	return ret;
}

static int parse_updatedb(const char *conf_file)
{
	char *conf_data = NULL;
	int res = 0;

	if (conf_file)
	    load_file(conf_file, &conf_data);
	else
	    load_file(UPDATEDB_FILE, &conf_data);

	if (!parse_PRUNEFS(conf_data))
	    goto EXIT;

	if (!parse_PRUNEPATHS(conf_data))
	    goto EXIT;

	res = 1;
EXIT:
	free(conf_data);
	return res;
}


/* Incremental Encoding algorithm */
static void
encode(FILE *fd, char *path, struct enc_data_s *enc_data)
{
	short code_len = 0;
	short code_num = 0;
	char *ptr1 = NULL;
	char *ptr2 = NULL;
	char *code_line = NULL;

	if (!path)
		bb_error_msg_and_die("encode: 'char *path' is NULL");
	if (VERBOSE)
	    printf("%s\n", path);
	/* Match number string */
	ptr1 = path;
	code_len = 0;
	if (enc_data->prev_line) {
		ptr2 = enc_data->prev_line;
		while (*ptr1 != '\0' && *ptr2 != '\0' && *ptr1 == *ptr2) {
			ptr1 += 1;
			ptr2 += 1;
			code_len += 1;
		}
	}
	code_num = code_len - enc_data->prev_len;
	enc_data->prev_len = code_len;
	code_line = ptr1;
	if (code_num < -127 || code_num > 127) {
		if (fputc((char)SLOC_ESC, fd) == EOF)
			bb_perror_msg_and_die("encode: fputc(): SLOC_ESC");
		if (fputc(code_num >> 8, fd) == EOF)
			bb_perror_msg_and_die("encode: fputc(): SLOC_ESC: code_num >> 8");
		if (fputc(code_num, fd) == EOF)
			bb_perror_msg_and_die("encode: fputc(): SLOC_ESC: code_num");
	} else {
		if (fputc(code_num, fd) == EOF)
			bb_perror_msg_and_die("encode: fputc(): code_num");
	}
	if (fputs(code_line, fd) == EOF)
		bb_perror_msg_and_die("encode: fprintf(): code_line");
	if (fputc('\0', fd) == EOF)
		bb_perror_msg_and_die("encode: fputc(): '\\0'");

	free(enc_data->prev_line);
	enc_data->prev_line = xstrdup(path);
}

struct DLIST_S {
	char *data;
	size_t len;
	struct DLIST_S *link;
};

static inline struct DLIST_S *
dlist_add_to(struct DLIST_S *old_head, char *new_item, size_t len)
{
	struct DLIST_S *new_head;

	new_head = xmalloc(sizeof(struct DLIST_S));
	new_head->data = new_item;
	new_head->len = len;
	new_head->link = old_head;

	return(new_head);
}

/* fast preudorecursive directory tree scaner
   by Vladimir Oleynik <dzo@simtreas.ru> */
static void my_scan_dir(char *p, FILE *fd, struct enc_data_s *penc_data)
{
    struct DLIST_S *dirs;
    struct DLIST_S *d_add;
    struct DLIST_S *d;
    size_t m0;
    char *fp0;

    if (!match_exclude(p))
	encode(fd, p, penc_data);
    else
	return;
    m0 = p[1] == '\0' ? 0 : strlen(p);
    dirs = dlist_add_to(NULL, strdup(p), m0);
    m0 = 4096;
    fp0 = xmalloc(m0);
    /* emulate recursion */
    do {
	d_add = NULL;   /* found subdirs */
	do {
	    DIR *dir = opendir(dirs->data);

	    if (dir != NULL) {
		struct dirent *de;
		size_t dirlen = dirs->len;

		while ((de = readdir(dir)) != NULL) {
			char *f = de->d_name;
			size_t fe_sz;
			size_t fsz;
			int st;

			if(*f == '.' && (!f[1] || (f[1] == '.' && !f[2])))
				continue;

			fe_sz = strlen(f) + 1;
			/* dir / file \0 */
			fsz = dirlen + 1 + fe_sz;
			if(fsz > m0) {
				free(fp0);
				m0 = fsz;
				fp0 = xmalloc(m0);
			}
			if(dirlen)
				memcpy(fp0, dirs->data, dirlen);
			fp0[dirlen] = '/';
			memcpy(fp0 + dirlen + 1, f, fe_sz);
			st = is_directory(fp0, 0, NULL);
			if(st < 0)
				continue;
			if (!match_exclude(fp0))
				encode(fd, fp0, penc_data);
			else
				continue;
			if(st > 0) {
				f = xmalloc(fsz);
				d_add = dlist_add_to(d_add, memcpy(f, fp0, fsz), fsz - 1);
			}
		}
		closedir(dir);
	    }
	    free(dirs->data);
	    d = dirs;
	    dirs = dirs->link;
	    free(d);
	} while(dirs);
	dirs = d_add;
    } while(dirs);
    free(fp0);
}

/* Create the database */
static void create_db(void)
{
	FILE *fd;
	char *tmp_file;
	uid_t db_uid;
	gid_t db_gid;
	mode_t db_mode;
	struct enc_data_s enc_data;

	/* Initialize encode data struct */
	enc_data.prev_line = NULL;
	enc_data.prev_len = 0;

	if (strcmp(output_db, DEFAULT_DB) == 0 && uid != DB_UID)
		bb_perror_msg_and_die("You are not authorized to create a default slocate database!");
	if (!access_path(output_db))
		bb_perror_msg_and_die("Unable to create database: %s", output_db);

	/* Determine the DB ownership */
	if (strcmp(output_db, DEFAULT_DB) == 0) {
		db_uid = DB_UID;
		db_gid = SLOCATE_GID;
		db_mode = DB_MODE;
	} else {
		db_uid = uid;
		db_gid = gid;
		/* Leave it to umask */
		db_mode = 0;
	}
	/* Make sure we can access the directory that we want to start
	 * searching on. */
	if (access(index_path, R_OK | X_OK) != 0)
		bb_perror_msg_and_die("Could not access index path '%s'", index_path);
	index_path = bb_simplify_path(index_path);

	tmp_file = get_temp_file();

	if (!(fd = fopen(tmp_file, "w")))
		bb_perror_msg_and_die("Could not open file for writing: %s", tmp_file);

	/* Set the mode so people can't peak by accident */
	if (db_mode) {
		if (fchmod(fileno(fd), db_mode) == -1)
			bb_perror_msg_and_die("Could not change permissions of '%u' on file: %s", db_mode, tmp_file);
	}

	/* Set the security level */
	if (putc((char)slevel, fd) == EOF)
		bb_perror_msg_and_die("create_db: Could not write to database. putc returned EOF.");

	// XXX: TODO: Support for multiple paths to index
	my_scan_dir(index_path, fd, &enc_data);

	if (fclose(fd) == -1)
		bb_perror_msg_and_die("fclose(): Could not close tmp file: %s", tmp_file);
	if (rename(tmp_file, output_db) == -1)
		bb_perror_msg_and_die("create_db(): rename(): Could not rename '%s' to '%s'", tmp_file, output_db);
	/* Only chown database to group 'slocate' if the output database
	 * is the default one. */
	if (strcmp(output_db, DEFAULT_DB) == 0) {
		if (chown(output_db, db_uid, db_gid) == -1)
			bb_perror_msg_and_die("create_db(): chown(): Could not set '%s' group on file '%s'", DB_GROUP, output_db);
	}

	free(tmp_file);
	free(enc_data.prev_line);
}


/* Set the path head.
 *
 * Construct the beginning of the full path.
 */
static char *
set_path_head(char *path_head, int code_num, char *prev_code_str)
{
	int path_len = 0;
	int len = 0;

	if (path_head)
	    path_len = strlen(path_head);

	/* If code_num > 0 then we want to add from the previous line */
	if (code_num > 0) {
		if (!prev_code_str)
			bb_perror_msg_and_die("set_path_head: prev_code_str == NULL.");
		if (!path_head) {
			path_head = xmalloc(1);
			path_head[0] = 0;
		}
		path_head = xrealloc(path_head, path_len+code_num+1);
		path_head[path_len+code_num] = 0;
		strncat(path_head, prev_code_str, code_num);
	/* If code_num < 0 then we will want to delete from the end of the
	 * previous line */
	} else if (code_num < 0) {
		/* NOTE '+code_num' is done because code_num will be < 0 thus
		 * cancelling the + */
		len = path_len+code_num;
		if (len <= 0)
			bb_perror_msg_and_die("set_path_head: path_head len <= 0: %d", len);
		path_head = xrealloc(path_head, len+1);

		if (len <= path_len)
		    path_head[len] = 0;
	}
	return path_head;
}

static int
search_path(char *full_path, char *search_str, int globflag)
{
	int ret = 0;
	int match_ret = 0;

	match_ret = match(full_path, search_str, globflag);
	if (match_ret == 1) {
		if (slevel == VERIFY_ACCESS && !verify_access(full_path))
		    match_ret = 0;
	} else if (match_ret == -1) {
		goto EXIT;
	}
	if (match_ret == 1) {
		if (queries > 0)
		    queries -= 1;
		fprintf(stdout, "%s\n", full_path);
	}
	ret = 1;
EXIT:
	return ret;
}

/* Search the database */
static int
search_db(char *database, char *search_str)
{
	int fd = -1;
	char ch[1];
	int buf_len;
	signed char buffer[BLOCK_SIZE];
	int ret = 0;
	int code_num = 0;
	char *path_head = NULL;
	char *prev_code_str = NULL;
	char *full_path = NULL;
	char *code_str = NULL;
	int b = 0;
	int b_mark;
	int STATE = DC_CODE;
	int size = 0;
	int globflag = 0;
	struct stat db_stat;
	gid_t effective_gid = 0;
	time_t now = 0;

	effective_gid = getegid();

	/* Drop privileges if the database's group is not slocate */
	if (stat(database, &db_stat) == -1)
		bb_perror_msg_and_die("Could not obtain information on database file '%s'", database);
	/* If the database's file group is not apart of the 'slocate' group,
	 * drop privileges. When multiple databases are specified, the ones
	 * apart of the 'slocate' group will be searched first before the
	 * privileges are dropped. */
	if (effective_gid == SLOCATE_GID && db_stat.st_gid != SLOCATE_GID) {
		if (setgid(gid) == -1)
			bb_perror_msg_and_die("Could not drop privileges.");
	}

	/* Warn if the database is old */
	if (!QUIET) {
		if ((now = time(&now)) == -1)
			bb_perror_msg_and_die("search_db: could not get time");

		if (now - db_stat.st_mtime > WARN_SECONDS)
			bb_error_msg("database %s' is more than %s old", database, WARN_MESSAGE);
	}
	if ((fd = open(database, O_RDONLY)) == -1)
		bb_perror_msg_and_die("search_db: open: '%s'", database);

	/* slevel */
	buf_len = read(fd, ch, 1);
	if (buf_len == 0)
		bb_perror_msg_and_die("search_db: read: '%s': Database file is empty.", database);
	else if (buf_len == -1)
		bb_perror_msg_and_die("serach_db: read: '%s'", database);

	if (search_str && (strchr(search_str,'*') != NULL || strchr(search_str,'?') ||
			   (strchr(search_str,'[') && strchr(search_str,']')))) {
		globflag = 1;
		/* Wrap search string with '*' wildcard characters
		 * since fnmatch will not match midstring */
		search_str = xasprintf("*%s*", search_str);
	}

	slevel = ch[0];
	b = 0;
	buf_len = read(fd, buffer, BLOCK_SIZE);
	while (buf_len > 0) {
		code_num = buffer[b];
		/* Escape char, read extra byte */
		if (code_num == SLOC_ESC) {
			b += 1;
			if (b == buf_len) {
				buf_len = read(fd, buffer, BLOCK_SIZE);
				b = 0;
			}
			/* A DC_ESC character indicates that we must read in two bytes
			 * for our code_num due to a long path. */
			code_num = buffer[b];
			b += 1;
			if (b == buf_len) {
				buf_len = read(fd, buffer, BLOCK_SIZE);
				b = 0;
			}
			code_num = (code_num << 8) | (buffer[b] & 0xff);
		}

		/* Data */
		b += 1;
		if (b == buf_len) {
			buf_len = read(fd, buffer, BLOCK_SIZE);
			b = 0;
		}
		/* If we are not resuming from an interrupted state then call
		 * this function to construct the beginning of the path to
		 * search */
		if (!(path_head = set_path_head(path_head, code_num, prev_code_str)) && code_num != 0)
		    goto EXIT;

		/* Mark the current location in the buffer so we can parse out
		 * the string data for the path */
		while(1) {
			b_mark = b;
			while (buffer[b] != '\0' && b < buf_len)
			    b += 1;
			/* If we are not resuming an interruption then initialize the
			 * code_str variable */
			if (STATE != DC_DATA_INTR) {
				code_str = xmalloc(b-b_mark+1);
				code_str[0] = 0;
			} else {
				size = strlen(code_str)+(b-b_mark)+1;
				if (size < 0)
				    size = 1;
				code_str = xrealloc(code_str, size);
				code_str[size-1] = 0;
				STATE = DC_NONE;
			}

			/* recast, buffer doesn't need to be signed here. */
			code_str = strncat(code_str, ((char *)buffer)+b_mark, b-b_mark);
			/* If we ran into the end of the current buffer, set the state
			 * to DC_DATA_INTR and exit so we can read more data and
			 * return to this state */
			if (b == buf_len && buffer[b-1] != '\0') {
				buf_len = read(fd, buffer, BLOCK_SIZE);
				b = 0;
				STATE = DC_DATA_INTR;
			} else
			    break;
		}

		if (!path_head) {
			full_path = xstrdup(code_str);
		} else {
			full_path = xasprintf("%s%s", path_head, code_str);
		}
		if (prev_code_str) {
			free(prev_code_str);
			prev_code_str = NULL;
		}

		/* Save a pointer to the current path, we need it for the next
		 * path we decode */
		prev_code_str = xstrdup(code_str);
		free(code_str);
		code_str = NULL;

		/* Search the current path string */
		if (!search_path(full_path, search_str, globflag))
		    goto EXIT;

		if (queries == 0)
		    break;
		if (full_path) {
			free(full_path);
			full_path = NULL;
		}

		b += 1;
		if (b == buf_len) {
			buf_len = read(fd, buffer, BLOCK_SIZE);
			b = 0;
		}
	}

	if (buf_len == -1)
		bb_perror_msg_and_die("search_db: read: '%s'", database);

	ret = 1;
EXIT:
	if (fd > -1)
	    close(fd);
	free(full_path);
	free(code_str);
	free(prev_code_str);
	free(path_head);
	return ret;
}



/* Parse command line options */
static void parse_cmds(void)
{
	int ch;
	char *ENV_locate_path;
	int add_default_db = 1;

	llist_t *d_exclude = NULL;
	int opt_s_level;
	llist_t *d_in = NULL;
	llist_t *f_exclude = NULL;

	if (strcmp(bb_applet_name, "updatedb") == 0)
	    updatedb = TRUE;

	ch = bb_xgetopt_ulflags("qve:l:n:uU:ic:r:d:o:f:h\0:-ln:e::d::f::h-:Uu",
				&d_exclude, &opt_s_level, &queries,
				&index_path, &cmd_updatedb_conf,
				&reg_pattern, &d_in, &output_db,
				&f_exclude);

	/* Quiet Mode. Don't print warnings or errors. */
	/* We set g_data since this is the only place it is called from */
	QUIET = ch & 1;
	if(QUIET)
		freopen(bb_dev_null, "w", stderr);
	/* Turn VERBOSE mode ON */
	VERBOSE = ch & 2;
	/* 'e' Exclude specified directories from database */
	while (d_exclude) {
		parse_exclude(d_exclude->data);
		d_exclude = d_exclude->link;
	}
	/* Set the security level of the database when creating or updating the
	 * database. If set to 0, security checks will not be preformed */
	if((ch & 8)) {
		if(opt_s_level != 0 && opt_s_level != 1)
			bb_perror_msg_and_die("Security level must be 0 or 1.");
		slevel = opt_s_level + '0';
	}
	if((ch & 16)) {
		/* Limit the amount of search results */
		if (queries < 0)
			bb_error_msg("Argument 'n': '%d': value should be greater than 0.", queries);
	}
	if((ch & 32)) {
		/* -u or -U path */
		/* Index from the root '/' or a specific path */
		updatedb = TRUE;
		add_default_db = 0;
	}
	/* 'i' - Make search case insensitive */
	nocase = ch & 128;

	while (d_in) {
		/* Specify the database to search in */
		add_default_db = 0;
		parse_userdb(d_in->data);
		d_in = d_in->link;
	}
	if(output_db != NULL) {
		/* Specify the database path to write to when creating or updating a database */
		output_db = bb_simplify_path(output_db);
	}
	while (f_exclude) {
		/* Exclude by filesystem */
		parse_fs_exclude(f_exclude->data);
		f_exclude = f_exclude->link;
	}

	/* Get search strings, unless a regular expression was specified */
	if (reg_pattern == NULL)
		cmd_search_str = Gargv;
	else {
		/* We compile the regexp here so that nocase will be determined */
		xregcomp(&reg, reg_pattern, nocase?REG_ICASE:0);
	}

	if (!updatedb) {
		/* Parse environment variables */
		if ((ENV_locate_path = getenv("LOCATE_PATH")))
			parse_userdb(ENV_locate_path);

		if (add_default_db)
			parse_userdb(DEFAULT_DB);
	} else
	    parse_updatedb(cmd_updatedb_conf);
}

/* Main function */
int slocate_main(void)
{
	/* Initialize global data defaults */
	index_path = "/";
	uid = getuid();
	gid = getgid();
	SLOCATE_GID = get_gid(DB_GROUP);
	queries = -1;
	slevel = VERIFY_ACCESS;

	/* Parse command line arguments */
	parse_cmds();

	if (updatedb) {
		/* Drop privileges since they are not required to
		 * create a database */
		if (setgid(gid) == -1)
			bb_perror_msg_and_die("Could not drop privileges.");

		if (!output_db)
			output_db = DEFAULT_DB;
		create_db();
		return 0;
	} else if (Gargc > 0 || reg_pattern) {
		int i;

		/* Search the database */
		if (!input_db || !input_db[0]) {
			input_db = init_input_db(1);
			input_db[0] = xstrdup(DEFAULT_DB);
		}

		for (i = 0; input_db[i]; i += 1) {
			int search_ret;

			/* Regular expression search */
			if (reg_pattern)
			    search_ret = search_db(input_db[i], NULL);
			/* Search each string */
			else {
				int s;

				search_ret = 1;
				for (s = 0; cmd_search_str && cmd_search_str[s] && search_ret; s += 1)
				    search_ret = search_db(input_db[i], cmd_search_str[s]);
			}

			if (!search_ret)
			    return 1;
		}
		return 0;
	} else {
		bb_show_usage();
	}
}
#endif /* APPLET_USAGE */
