/*
 * Virtual unmanaged IPv4 client-server Tunnel over UDP with strong
 * IP and TCP/UDP/ICMP header compression and auto calculate mtu
 * value with minimal reduce for preconfigured hosts/nets over tunnel.
 * Server must have public IP address, client connecting to it
 * may be over NAT.
 *
 * Copyright (c) 2009,2010,2018 may safely be consumed by a BSD or GPL license.
 * Written by:   Vladimir Oleynik <dzo@simtreas.ru>
 *
 *
 * For example:
 * if you need tunnel may be with one another host connected to server
 * and with one another host connected to client for icmp/tcp/udp
 * traffic only then this program set mtu a gift to 14 bytes.
 * Its great that UDP header (8)! For vu4tun over UDP encapsulation result is
 * IP(20)+UDP(8)+VU4TUN(2)-2(L)-14=14 and for original MTU=1500 need
 * set tunnel MTU=1500-14=1486
 * If unused IP-option and use predefined port for tcp and udp
 * (in each pair) set mtu a gift to 16 bytes.
*/

#define ENV_TUN_DEV "TUN_DEV"

#if defined(APPLET_USAGE) || defined(STANDALONE)
#define vu4tun_trivial_usage \
"[-p] -l local_tun_addr[/MASK] [-m mtu] [-k keyfile]\n" \
"[-A HOST,...] [-B IP1.IP2.IP3,...] [-C IP1.IP2,...]\n" \
"[-V or -E IP1] [-t spec_TCP_port] [-u spec_UDP_port]\n" \
"[-r /path/run_script] [-b bind_address] [-a allow_remote_addr]\n" \
"[client mode: to_server_address] port [-P]"
#define vu4tun_full_usage "Options\n" \
"\t-p\t- send over tunnel ICMP/TCP/UDP only packets\n" \
"\t-l ADDR\t- set tunnel local IP address and do inform remote about\n" \
"it. Also, read remote config with tunnel_address/MASK and automatic adding\n" \
"a route to remote network with MASK by kernel.\n" \
"\t-m mtu\t- set MTU for a traffic. Default is 1500.\n" \
"\t-k file_with_password\t- use password protect conection from\n" \
"content this binary file.\n" \
"\t-A HOST,... - comma separarated list by specified local another\n" \
"hosts for encoding.\n"  \
"\t-B IP1.IP2.IP3,... - comma separarated list by specified local\n" \
"another nets/24 for encoding.\n"  \
"\t-C IP1.IP2,... - comma separarated list by specified local another\n" \
"nets/16 for encoding.\n"  \
"\t-E IP1\t- specified local a net/8 for encoding\n" \
"\t-V\t- allow send over tunnel any 0.0.0.0/0 source address\n" \
"\t-t PORT\t- specified local tcp port (256-65535) for encodind.\n" \
"\t-u PORT\t- specified local udp port (256-65535) for encodind.\n" \
"\t-r /path/run_script - run script after established tunnel\n"  \
"for example sets routes and arps. You can use tunnel device name\n"  \
"as $"ENV_TUN_DEV" from enviroment variable\n" \
"\t-b bind_address - use this IP for make tunnel in multihome host\n"  \
"for the server mode\n"  \
"\t-a ADDR\t- allow make tunnel with this specified client address only\n"  \
"for the server mode\n"  \
"\t-P\t- create pidfilename with a port as /var/run/vu4tun_port.pid"
#endif

#if !defined(APPLET_USAGE) || defined(STANDALONE)

#include "busybox.h"
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <syslog.h>
#include <linux/if_tun.h>
#include <sys/time.h>
#include "bb_xopen.h"
#include "bb_inet.h"
#include "md5.h"

#define VU4TUN_FSIZE_MASK         0x07ff

#define VU4TUN_IPH_COMPRESSED     0x0800
#define VU4TUN_BAD_FRAME          0x8000
#define VU4TUN_OPT_BIT            0x1000

#define VU4TUN_CFG_OUT            0x2000
#define VU4TUN_ECHO_REQ           0x4000
#define VU4TUN_CFG_ASK            (VU4TUN_CFG_OUT|VU4TUN_OPT_BIT)
#define VU4TUN_CFG_BAD            (VU4TUN_CFG_OUT|VU4TUN_BAD_FRAME)
#define VU4TUN_ECHO_REP           (VU4TUN_ECHO_REQ|VU4TUN_OPT_BIT)
#define VU4TUN_CFG_OK             (VU4TUN_CFG_OUT|VU4TUN_ECHO_REQ)
#define VU4TUN_QUIT               (VU4TUN_BAD_FRAME|VU4TUN_OPT_BIT)

#define VU4TUN_TIMEOUT 10       /* sec */

#define ZALR 18
#define ZA   2

#define IPv4_chars 4

#define IS_ICMP 1
#define IS_TCP  6
#define IS_UDP  17

#define U_C(c) ((unsigned char) (c))

#define MD5_HASH_SIZE           16

struct LR {
	struct in_addr tun_addr;
	struct in_addr tun_mask;
	char tun_addr_s[IPv4_chars];
	char tun_mask_s[IPv4_chars];

	int var;
	char var_host_net[IPv4_chars];
	int vara;
	char vara_hosts[ZALR][IPv4_chars];
	int varb;
	char varb_nets[ZALR][IPv4_chars];
	int varc;
	char varc_nets[ZALR][IPv4_chars];
	int tcp_port;
	int udp_port;

	unsigned char challenge[MD5_HASH_SIZE];
};

CALC_BSS struct BB_COMMON_BSS {
	struct in_addr allow_remote_addr;
	struct in_addr *allow_remote_addr_set;

	int tcp_udp_icmp_proto_only;
	int mtu;

	struct LR l;
	struct LR r;

	char cfg_buf[VU4TUN_FSIZE_MASK + 1];
	char buf[VU4TUN_FSIZE_MASK + 1];
	char hc_out[VU4TUN_FSIZE_MASK + 1];
	char *run;
	char *secret;
	int secret_len;
} bb_common_bss;

static int encoder(char *in, int len);
static int decoder(char *in, int len);

#define L bb_common_bss.l
#define R bb_common_bss.r
#define run bb_common_bss.run

#define vT L.var
#define vA L.vara
#define vB L.varb
#define vC L.varc
#define rvT R.var
#define rvA R.vara
#define rvB R.varb
#define rvC R.varc

static int calc_mtu(void)
{
	int dec;

	/* minimal decrement: header compression by default */
	dec = 3;
	if(vT == 'H')
		dec += 4;
	if(rvT == 'H')
		dec += 4;
	if(vT == 'R')
		dec += 3;
	if(rvT == 'R')
		dec += 3;
	if(vT == 'E')
		dec += 1;
	if(rvT == 'E')
		dec += 1;
	if(vT == 'S')
		dec += 1 + (vC == 0);
	if(rvT == 'S')
		dec += 1 + (rvC == 0);
	if(vT == 'X')
		dec += 1 + (vB == 0 && vC == 0) + (vC == 0);
	if(rvT == 'X')
		dec += 1 + (rvB == 0 && rvC == 0) + (rvC == 0);
	if(bb_common_bss.tcp_udp_icmp_proto_only)
		dec += 3;
	dec &= 0xfffe;
	/* IP + UDP - header compression */
	return bb_common_bss.mtu - 20 - 8 + dec;
}

static int tun_open(void)
{
	struct ifreq ifr;
	struct sockaddr_in sai;
	int fd, sockfd;
	char tun_name[IFNAMSIZ + 1];

	fd = bb_xopen("/dev/net/tun", O_RDWR | BB_OPEN_ERR);

	memset(&ifr, 0, sizeof(ifr));
	ifr.ifr_flags = IFF_TUN | IFF_NO_PI;
	xioctl(fd, TUNSETIFF, (void *) &ifr, "\1TUNSETIFF");
	safe_strncpy(tun_name, ifr.ifr_name, sizeof(tun_name));
	setenv(ENV_TUN_DEV, tun_name, 1);

	sockfd = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM | BB_XSOCKET_FLG_ERR_DIE);

	ifr.ifr_flags = 0;
	memset(&sai, 0, sizeof(struct sockaddr_in));
	sai.sin_family = AF_INET;

	memcpy(&sai.sin_addr, &L.tun_addr, sizeof(L.tun_addr));
	memcpy(&ifr.ifr_addr, &sai, sizeof(ifr.ifr_addr));
	xioctl(sockfd, SIOCSIFADDR, &ifr, "\1set tunnel IP");

	xioctl(sockfd, SIOCGIFFLAGS, &ifr, "\1get tunnel flags");
	ifr.ifr_flags |= IFF_UP | IFF_RUNNING | IFF_POINTOPOINT;
	ifr.ifr_flags &= ~IFF_MULTICAST;
	xioctl(sockfd, SIOCSIFFLAGS, &ifr, "\1set tunnel flags");

	memcpy(&sai.sin_addr, &R.tun_addr, sizeof(R.tun_addr));
	memcpy(&ifr.ifr_dstaddr, &sai, sizeof(ifr.ifr_dstaddr));
	xioctl(sockfd, SIOCSIFDSTADDR, &ifr, "\1set tunnel remote IP");

	memcpy(&sai.sin_addr, &R.tun_mask, sizeof(R.tun_mask));
	memcpy(&ifr.ifr_netmask, &sai, sizeof(ifr.ifr_netmask));
	xioctl(sockfd, SIOCSIFNETMASK, &ifr, "\1set tunnel netmask");

	ifr.ifr_mtu = calc_mtu();

	xioctl(sockfd, SIOCSIFMTU, &ifr, "\1set tunnel mtu");
	close(sockfd);

	if(run != NULL) {
		volatile int e = 0;
		pid_t pid = vfork();
		if(pid < 0)
			bb_perror_msg_and_die("fork");
		if(pid == 0) {
			execl(run, bb_basename(run), NULL);
			e = errno;
			exit(1);
		} else {
			while(wait(NULL) < 0);
			if(e) {
				errno = e;
				bb_perror_msg_and_die("exec %s", run);
			}
		}
	}
	return fd;
}

static void make_printable_bytes(unsigned char *bytes, char *out_str)
{
	int i;

	for(i = 0; i < MD5_HASH_SIZE; i++)
		sprintf(out_str + i * 2, "%02x", bytes[i]);
}

static void md5_generate_challenge(void)
{
	long seed;
	struct timeval t;
	size_t len;
	unsigned char *buf = L.challenge;

	/* init random */
	gettimeofday(&t, NULL);
	seed = t.tv_sec ^ t.tv_usec ^ getpid();
	srand48(seed);

	for(len = MD5_HASH_SIZE; len >= sizeof(uint32_t); len -= sizeof(uint32_t)) {
		uint32_t l32 = mrand48();
		memcpy(buf, &l32, sizeof(uint32_t));
		buf += sizeof(uint32_t);
	}
}

static void md5_make_response(unsigned char *response, unsigned char challenge[MD5_HASH_SIZE])
{
	md5_ctx_t ctx;

	md5_begin(&ctx);
	md5_hash(bb_common_bss.secret, bb_common_bss.secret_len, &ctx);
	md5_hash(challenge, MD5_HASH_SIZE, &ctx);
	md5_end(response, &ctx);
}

static char *out4_dot(char buf[IPv4_chars], char out[16])
{
	sprintf(out, "%d.%d.%d.%d", U_C(buf[0]), U_C(buf[1]), U_C(buf[2]), U_C(buf[3]));
	return out;
}

static int make_config_message(int ask)
{
	int l = sizeof(uint16_t), i;
	char *p = bb_common_bss.cfg_buf;
	uint16_t c16;
	char outa[16], outm[MD5_HASH_SIZE * 2 + 1];

	l += sprintf(p + l, "L%s/%s\n", out4_dot(L.tun_addr_s, outa), out4_dot(L.tun_mask_s, outm));
	l += sprintf(p + l, "%c%s\n", vT, out4_dot(L.var_host_net, outa));
	for(i = 0; i < vA; i++)
		l += sprintf(p + l, "A%s\n", out4_dot(L.vara_hosts[i], outa));
	for(i = 0; i < vB; i++)
		l += sprintf(p + l, "B%s\n", out4_dot(L.varb_nets[i], outa));
	for(i = 0; i < vC; i++)
		l += sprintf(p + l, "C%s\n", out4_dot(L.varc_nets[i], outa));
	if(bb_common_bss.tcp_udp_icmp_proto_only)
		l += sprintf(p + l, "P\n");
	l += sprintf(p + l, "M%d\n", bb_common_bss.mtu);
	if(L.tcp_port)
		l += sprintf(p + l, "T%d\n", L.tcp_port);
	if(L.udp_port)
		l += sprintf(p + l, "U%d\n", L.udp_port);
	if(bb_common_bss.secret_len) {
		unsigned char response[MD5_HASH_SIZE];

		if(ask) {
			memset(response, 0, sizeof(response));
			if(!memcmp(response, L.challenge, sizeof(response)))
				md5_generate_challenge();
		} else {
			md5_generate_challenge();
		}
		make_printable_bytes(L.challenge, outm);
		l += sprintf(p + l, "N%s\n", outm);
		if(ask == 0) {
			md5_make_response(response, R.challenge);
			make_printable_bytes(response, outm);
			l += sprintf(p + l, "K%s\n", outm);
		}
	}
	l += sprintf(p + l, "Z\n");
	c16 = l | VU4TUN_CFG_OUT | ask;
	c16 = htons(c16);
	memcpy(bb_common_bss.cfg_buf, &c16, sizeof(c16));
	return l;
}

/* gethostbyname/getnetworkbyname with allow without tails
   .0 or .0.0 or .0.0.0 for /24/16/8 masks and allow
   [:spaces:]IP[.][:separator:] and check zero in tails for non /32 masks */
static char *safe_set_net(const char *arg, int msk, int sep, struct in_addr *a, char out[IPv4_chars], const char *err_msg)
{
	char *endptr;
	int value;
	int n = 0, initial_msk = msk;
	const char *initial_arg = arg;
	int need_zero = 0;
	char ibuf[16];

	while(*arg == ' ' || *arg == '\t') arg++;
	if(*arg < '0' || *arg > '9') {
		/* hostname/networkname or unexpected word */
		struct in_addr a_stack;

		if(a == NULL) a = &a_stack;
		endptr = strchr_set0((char *)arg, sep);
		if(endptr == NULL) {
			sep = 0;
			endptr = strchr_set0((char *)arg, 0);
		}

		if(msk == 32) {
			if(bb_gethostbyname(BB_GHBN_IN_ADDR | BB_GHBN_SILENT, arg, a) == NULL)
				bb_error_msg_and_die("%s %s: invalid ip address", err_msg, arg);
		} else {
			struct netent *np;

			if ((np = getnetbyname(arg)) != (struct netent *) NULL)
				a->s_addr = htonl(np->n_net);
			 else
				bb_error_msg_and_die("%s %s: invalid net", err_msg, arg);
		}

		inet_ntop(AF_INET, a, ibuf, sizeof(ibuf));
		/* recursion call with protect ibuf change */
		safe_set_net(ibuf, msk, 0, NULL, out, err_msg);
		*endptr = sep;
		return endptr;
	}

	memset(out, 0, IPv4_chars);
	for(; msk; msk -= 8) {
		value = bb_xparse_i_number(BB_PN_SILENT_RE | BB_PN_SET_UPPER |
				BB_PN_END_PTR,
				arg, (need_zero ? 0 : 255), &endptr);
		if(errno) break;
		out[n++] = value;
		if (*endptr == '.')
			arg = ++endptr;
		if (*endptr == '\0' || *endptr == sep) {
			if(msk != 8) break;
			if(a) {
				sprintf(ibuf, "%d.%d.%d.%d", U_C(out[0]), U_C(out[1]),
						U_C(out[2]),  U_C(out[3]));
				inet_pton(AF_INET, ibuf, a);
			}
			return endptr;
		}
		if(msk == 8 && (initial_msk + need_zero) != 32) {
			need_zero += 8;
			msk += 8;
		}
	}
	if(err_msg != NULL) {
		if(initial_msk == 32)
			bb_error_msg_and_die("%s %s: invalid ip address", err_msg, initial_arg);
		else
			bb_error_msg_and_die("%s %s/%d: invalid net", err_msg, initial_arg, initial_msk);
	}
	return NULL;
}

static char *get_cfg_port(const char *buf, int *p)
{
	char *endptr;
	int value;

	value = bb_xparse_i_number(BB_PN_SILENT_RE |
				BB_PN_SET_LOWER | BB_PN_SET_UPPER | BB_PN_END_PTR,
				buf, 256, 65535, &endptr);
	if (errno || *endptr != '\n')
		return NULL;
	*p = value;
	return endptr;
}

static const char *get_md5_str(const char *response, unsigned char hash_response[MD5_HASH_SIZE])
{
	int i;
	char buf[4];

	for(i = 0; i < MD5_HASH_SIZE; i++) {
		int c = *response++;

		if(c < '0') /* invalid responce format */ return NULL;
		buf[0] = c;
		c = *response++;
		if(c < '0') /* invalid responce format */ return NULL;
		buf[1] = c;
		buf[2] = '\0';
		c = bb_xparse_i_number(16 | BB_PN_SILENT_RE, buf);
		if(errno) /* invalid responce format */ return NULL;
		hash_response[i] = c;
	}
	if(*response != '\n') /* invalid responce format */ return NULL;
	return response;
}

#define KEY_UNCOMPARED 1
#define KEY_COMPARED   2

static int decode_config_message(size_t l, int po, int mtu)
{
	int o;
	size_t i;
	int L_required_specified = 0;
	int N_required_specified = bb_common_bss.secret_len == 0, K = 0;
	const char *e;
	int n = 0, sum = 0;
	const char *buf;
	unsigned char response[MD5_HASH_SIZE], cmp_response[MD5_HASH_SIZE];

	if(l < sizeof(uint16_t))
		return -1;
	buf = bb_common_bss.buf + sizeof(uint16_t);
	l -= sizeof(uint16_t);
	bb_common_bss.tcp_udp_icmp_proto_only = po;
	bb_common_bss.mtu = mtu;
	memset(&bb_common_bss.r, 0, sizeof(bb_common_bss.r));

	for(;;) {
		if(l < 2)
			return -1;
		o = *buf++;
		l--;
		for(i = 0; i < l; i++)
			if(buf[i] == '\n' || buf[i] < '.')
				break;
		if(buf[i] != '\n')
			return -1;
		e = buf;
		switch(o) {
			case 'N':
				N_required_specified = 1;
				e = get_md5_str(buf, R.challenge);
				break;
			case 'K':
				K = 1;
				e = get_md5_str(buf, response);
				break;
			case 'L':
				L_required_specified = 1;
				e = safe_set_net(e, 32, '/', &R.tun_addr, R.tun_addr_s, NULL);
				if(e == NULL)
					return -1;
				e = safe_set_net(e + 1, 32, '\n', &R.tun_mask, R.tun_mask_s, NULL);
				break;
			case 'V':
			case 'R':
			case 'E':
			case 'S':
			case 'X':
			case 'H':
				if(sum) return -1;
				rvT = o;
				sum = o == 'H' ? -1 : o == 'V' || o == 'R' ? ZALR : 2;
				e = safe_set_net(e, 32, '\n', NULL, R.var_host_net, NULL);
				break;
			case 'A':
			case 'B':
			case 'C':
				if(n >= sum) return -1;
				if(rvT == 'R' && o != 'A') return -1;
				e = safe_set_net(e, 32, '\n', NULL,
						(o == 'A' ? R.vara_hosts[n] :
						 o == 'B' ? R.varb_nets[n] : R.varc_nets[n]),
						NULL);
				n++;
				if(o == 'A')
					rvA++;
				else if(o == 'B')
					rvB++;
				else
					rvC++;
				break;
			case 'P':
				bb_common_bss.tcp_udp_icmp_proto_only = 1;
				break;
			case 'M':
				e = get_cfg_port(e, &mtu);
				if(mtu < bb_common_bss.mtu)
					bb_common_bss.mtu = mtu;
				break;
			case 'T':
				e = get_cfg_port(e, &R.tcp_port);
				break;
			case 'U':
				e = get_cfg_port(e, &R.udp_port);
				break;
			case 'Z':
				if(*e != '\n' || l != 1 || !L_required_specified || rvT == 0 || !N_required_specified)
					return -1;
				if(K) {
					md5_make_response(cmp_response, L.challenge);
					if(memcmp(cmp_response, response, MD5_HASH_SIZE))
						return KEY_UNCOMPARED;
					return KEY_COMPARED;
				}
				return 0;
			default:
				return -1;
		}
		if(e == NULL || *e++ != '\n')
			return -1;
		l -= (e - buf);
		buf = e;
	}
}

static void parse_ip_list(const char *l, const char *opt)
{
	char *vs, *ve;
	int *pnhn, nhn, msk;

	switch(opt[1]) {
		case 'A':
			pnhn = &vA;
			nhn = *pnhn;
			vs = L.vara_hosts[nhn];
			ve = L.vara_hosts[ZALR - (vB + vC)];
			msk = 32;
			break;
		case 'B':
			pnhn = &vB;
			nhn = *pnhn;
			vs = L.varb_nets[nhn];
			ve = L.varb_nets[ZALR - (vA + vC)];
			msk = 24;
			break;
		case 'C':
			pnhn = &vC;
			nhn = *pnhn;
			vs = L.varc_nets[nhn];
			ve = L.varc_nets[ZALR - (vA + vB)];
			msk = 16;
			break;
		default:
			bb_error_msg_and_die("internal error");
	}

	for( ; vs < ve; ) {
		l = safe_set_net(l, msk, ',', NULL, vs, opt);
		nhn++;
		if(*l != '\0')
			l++;
		if(*l == '\0') {
			*pnhn = nhn;
			return;
		}
		vs += IPv4_chars;
	}
	bb_error_msg_and_die("too many IP specified");
}

static void parse_ip_list_from_optA(const char *opt)
{
	parse_ip_list(opt, "-A");
}

static void parse_ip_list_from_optB(const char *opt)
{
	parse_ip_list(opt, "-B");
}

static void parse_ip_list_from_optC(const char *opt)
{
	parse_ip_list(opt, "-C");
}

static int parse_spec_port(const char *s, const char *proto)
{
	if(s == NULL)
		return 0;
	return bb_xparse_i_number(BB_PN_SET_LOWER | BB_PN_SET_UPPER |
				BB_PN_ERR_STRV | BB_PN_SHOW_USAGE,
				s, 256, 65535, "invalid port specified '%s' for %s", s, proto);
}

static void make_spec(int code, int socket, struct sockaddr_in *sa)
{
	char buf[sizeof(uint16_t) + MD5_HASH_SIZE * 2 + 1];
	int l = sizeof(uint16_t);
	uint16_t ls;

	if(code == VU4TUN_CFG_OK && bb_common_bss.secret_len) {
		md5_make_response((unsigned char *)buf + sizeof(uint16_t), R.challenge);
		l += MD5_HASH_SIZE;
	}
	ls = htons(l | code);
	memcpy(buf, &ls, sizeof(uint16_t));
	if(sa != NULL)
		sendto(socket, buf, l, 0, (struct sockaddr *) sa, sizeof(*sa));
	else
		bb_full_write(socket, buf, l);
}

static int decode_input_from_net(int l)
{
	uint16_t c;
	int lc;
	const char *buf = bb_common_bss.buf;

	if((size_t)l < sizeof(uint16_t))
		return VU4TUN_BAD_FRAME;
	memcpy(&c, buf, sizeof(uint16_t));
	c = ntohs(c);
	lc = c & VU4TUN_FSIZE_MASK;
	if(lc != l)
		return VU4TUN_BAD_FRAME;
	return (int)c - (int)l;
}

static int my_select(int maxfd, fd_set *r, struct timeval *ptv)
{
	for(;;) {
		int l = select(maxfd, r, NULL, NULL, ptv);
		if(l < 0) {
			if(errno == EINTR)
				continue;
			bb_perror_msg_and_die("select");
		}
		return l;
	}
}

static int my_select1(int maxfd, fd_set *r, struct timeval *ptv, int socket)
{
	FD_SET(socket, r);
	return my_select(maxfd, r, ptv);
}

static int my_select2(int maxfd, fd_set *r, struct timeval *ptv, int socket, int tun)
{
	FD_SET(tun, r);
	return my_select1(maxfd, r, ptv, socket);
}

static void client(int s, int po, int mtu)
{
	int code, ping_timeout, net_timeout;
	int tun, l, maxfd = s + 1;
	fd_set  rset;
	struct timeval tv;
	int lcm;

	FD_ZERO(&rset);
	memset(L.challenge, 0, sizeof(L.challenge));
	for(;;) {
		lcm = make_config_message(VU4TUN_CFG_ASK);
		/* send config */
		l = safe_write(s, bb_common_bss.cfg_buf, lcm);
		if(l > 0) {
			tv.tv_usec = 0; tv.tv_sec = VU4TUN_TIMEOUT;
			l = my_select1(maxfd, &rset, &tv, s);
			if(l == 0) continue;
			l = safe_read(s, bb_common_bss.buf, sizeof(bb_common_bss.buf));
			if(l > 0) {
				code = decode_input_from_net(l);
				if(code == VU4TUN_CFG_OUT) {
					code = decode_config_message(l, po, mtu);
					if((bb_common_bss.secret_len == 0 && code == 0) || code == KEY_COMPARED) {
						make_spec(VU4TUN_CFG_OK, s, NULL);
						break;
					} else {
						make_spec(VU4TUN_CFG_BAD, s, NULL);
					}
				}
			}
		}
		sleep(1);
	}
	tun = tun_open();
	if(s < tun)
		maxfd = tun + 1;
	for(net_timeout = ping_timeout = 0;;) {
		if(!net_timeout) {
			tv.tv_usec = 0;
			tv.tv_sec = VU4TUN_TIMEOUT;
			net_timeout = 1;
		}
		l = my_select2(maxfd, &rset, &tv, s, tun);
		if(l == 0) {
			ping_timeout++;
			if(ping_timeout == 3) {
				make_spec(VU4TUN_QUIT, s, NULL);
				close(tun);
				return;
			}
			make_spec(VU4TUN_ECHO_REQ, s, NULL);
			net_timeout = 0;
			continue;
		}
		if( FD_ISSET(s, &rset) ) {
			l = safe_read(s, bb_common_bss.buf, sizeof(bb_common_bss.buf));
			if(l < 0) {
				close(tun);
				return;
			}
			code = decode_input_from_net(l);
			if(code == VU4TUN_ECHO_REP) {
				ping_timeout = 0;
				net_timeout = 0;
			} else if(code == VU4TUN_ECHO_REQ) {
				net_timeout = 0;
				make_spec(VU4TUN_ECHO_REP, s, NULL);
			} else if(code == VU4TUN_QUIT) {
				close(tun);
				return;
			} else if(code == VU4TUN_IPH_COMPRESSED) {
				code = decoder(bb_common_bss.buf + sizeof(uint16_t), l - sizeof(uint16_t));
				if(code > 0) {
					net_timeout = 0;
					safe_write(tun, bb_common_bss.hc_out, code);
				}
			}
		}
		if( FD_ISSET(tun, &rset) ) {
			l = safe_read(tun, bb_common_bss.buf, sizeof(bb_common_bss.buf));
			code = encoder(bb_common_bss.buf, l);
			if(code > 0)
				safe_write(s, bb_common_bss.buf, code);
		}
	}
}

static int my_recvfrom(int s, struct sockaddr_in *ca, struct sockaddr_in *cmp)
{
	socklen_t slen = sizeof(*ca);
	int l;

	l = recvfrom(s, bb_common_bss.buf, sizeof(bb_common_bss.buf), 0, (struct sockaddr *) ca, &slen);
	if(l <= 0)
		return -1;
	if(cmp->sin_port == 0) {
		/* cmp address only */
		memcpy(cmp, ca, sizeof(*ca));
		if(bb_common_bss.allow_remote_addr_set != NULL)
			memcpy(&cmp->sin_addr, &bb_common_bss.allow_remote_addr, sizeof(cmp->sin_addr));
	}
	if(memcmp(ca, cmp, sizeof(*ca))) {
		make_spec(VU4TUN_QUIT, s, ca);
		return -1;
	}
	return l;
}

static void server(int s, int po, int mtu)
{
	int code, ping_timeout, net_timeout;
	int tun, l, maxfd = s + 1;
	fd_set  rset;
	struct timeval tv;
	struct sockaddr_in cliaddr, cliaddr_cmp;
	int lcm;

	memset(&cliaddr, 0, sizeof(cliaddr));
	FD_ZERO(&rset);
	for(;;) {
		memset(&cliaddr_cmp, 0, sizeof(cliaddr_cmp));
		/* read config */
		l = my_recvfrom(s, &cliaddr, &cliaddr_cmp);
		if(l < 0) {
			sleep(1);
			continue;
		}
		code = decode_input_from_net(l);
		if(code == VU4TUN_CFG_ASK) {
			code = decode_config_message(l, po, mtu);
			if(code == 0) {
				break;
			} else {
				make_spec(VU4TUN_CFG_BAD, s, &cliaddr);
				sleep(1);
			}
		}
	}

	lcm = make_config_message(0);
	/* send config */
	sendto(s, bb_common_bss.cfg_buf, lcm, 0, (struct sockaddr *) &cliaddr, sizeof(cliaddr));
	tv.tv_usec = 0; tv.tv_sec = VU4TUN_TIMEOUT;
	for(;;) {
		l = my_select1(maxfd, &rset, &tv, s);
		if(l != 0) {
			l = my_recvfrom(s, &cliaddr_cmp, &cliaddr);
			if(l < 0)
				continue;
			code = decode_input_from_net(l);
			if(code == VU4TUN_CFG_OK) {
				if(bb_common_bss.secret_len) {
					if(l == (MD5_HASH_SIZE + sizeof(uint16_t))) {
						unsigned char cmp_response[MD5_HASH_SIZE];

						md5_make_response(cmp_response, L.challenge);
						if(!memcmp(cmp_response, bb_common_bss.buf + sizeof(uint16_t), MD5_HASH_SIZE))
							break;
						bb_error_msg("Password incorrect");
						sleep(1);
					}
				} else {
					break;
				}
			}
		}
		make_spec(VU4TUN_CFG_BAD, s, &cliaddr);
		return;
	}

	tun = tun_open();
	if(s < tun)
		maxfd = tun + 1;
	for(net_timeout = ping_timeout = 0;;) {
		if(!net_timeout) {
			tv.tv_usec = 0;
			tv.tv_sec = VU4TUN_TIMEOUT;
			net_timeout = 1;
		}

		l = my_select2(maxfd, &rset, &tv, s, tun);
		if(l == 0) {
			ping_timeout++;
			if(ping_timeout == 3) {
				make_spec(VU4TUN_QUIT, s, &cliaddr);
				close(tun);
				return;
			}
			make_spec(VU4TUN_ECHO_REQ, s, &cliaddr);
			net_timeout = 0;
			continue;
		}
		if( FD_ISSET(s, &rset) ) {
			l = my_recvfrom(s, &cliaddr_cmp, &cliaddr);
			if(l < 0)
				continue;
			code = decode_input_from_net(l);
			if(code == VU4TUN_ECHO_REP) {
				ping_timeout = 0;
				net_timeout = 0;
			} else if(code == VU4TUN_ECHO_REQ) {
				net_timeout = 0;
				make_spec(VU4TUN_ECHO_REP, s, &cliaddr);
			} else if(code == VU4TUN_QUIT) {
				close(tun);
				return;
			} else if(code == VU4TUN_IPH_COMPRESSED) {
				code = decoder(bb_common_bss.buf + sizeof(uint16_t), l - sizeof(uint16_t));
				if(code > 0) {
					net_timeout = 0;
					safe_write(tun, bb_common_bss.hc_out, code);
				}
			}
		}
		if( FD_ISSET(tun, &rset) ) {
			l = safe_read(tun, bb_common_bss.buf, sizeof(bb_common_bss.buf));
			code = encoder(bb_common_bss.buf, l);
			if(code > 0)
				sendto(s, bb_common_bss.buf, code, 0, (struct sockaddr *) &cliaddr, sizeof(cliaddr));
		}
	}
}

static int conv_abc_xsrhve(int opt)
{
	vT = opt;
	if(opt == 'R' || opt == 'V') {
		return ZALR;
	} else if(opt == 'H') {
		memcpy(L.var_host_net, L.vara_hosts[0], IPv4_chars);
		vA = 0;
		return 0;
	} else if(opt == 'X') {
		memcpy(L.var_host_net, L.varb_nets[0], IPv4_chars);
		if(--vB) {
			memcpy(L.varb_nets[0], L.varb_nets[1], IPv4_chars);
			if(vB > 1)
				memcpy(L.varb_nets[1], L.varb_nets[2], IPv4_chars);
		}
	} else if(opt == 'S') {
		memcpy(L.var_host_net, L.varc_nets[0], IPv4_chars);
		if(--vC) {
			memcpy(L.varc_nets[0], L.varc_nets[1], IPv4_chars);
			if(vC > 1)
				memcpy(L.varc_nets[1], L.varc_nets[2], IPv4_chars);
		}
	}
	return ZA; /* opt == 'E' too */
}


static int parse_args(struct sockaddr_in *ba)
{
	char *bind_addr = NULL;
	int c, port;
	char *allow_remote_addr = NULL;
	char *local_tun_addr = NULL;
	char *X = NULL, *smtu = NULL, *password_file = NULL;
	int count = 0;
	char *stcp_port = NULL, *sudp_port = NULL;

#define p_BIT 1
#define V_BIT 2
#define E_BIT 4
#define A_BIT 8
#define B_BIT 16
#define C_BIT 32
#define P_BIT 64
	c = bb_xgetopt_ulflags("pVE:" "A:B:C:" "Pm:k:" "b:a:l:"
					"t:u:r:\0-1:?2:=A:=B:=C:EE",
				&X,
				parse_ip_list_from_optA,
				parse_ip_list_from_optB,
				parse_ip_list_from_optC, &smtu, &password_file,
				&bind_addr, &allow_remote_addr, &local_tun_addr,
				&stcp_port, &sudp_port, &run,
				&count);
	bb_common_bss.tcp_udp_icmp_proto_only = c & p_BIT;
	if(local_tun_addr == NULL) {
		bb_error_msg("do not found the option: -l local_tun_addr[/mask]");
		bb_show_usage();
	} else {
		char out[16];
		char *m = strchr(local_tun_addr, '/');
		if(m != NULL) {
			char *p = xstrdup(local_tun_addr);
			p[(m-local_tun_addr)] = '\0';
			local_tun_addr = p;
			m++;
			p = strchr(m, '.');
			if(p == NULL) {
				uint32_t netmask, msk;
				unsigned int netprefix = bb_xparse_ui_number(BB_PN_SET_LOWER1 | BB_PN_SET_UPPER |
								BB_PN_ERR_STR1,
							m, 32, "invalid mask `-l %s' but expect 1-32");
				netmask = 0;
				msk = 0x80000000;
				while (netprefix) {
					netmask |= msk;
					msk >>= 1;
					netprefix--;
				}
				sprintf(out, "%d.%d.%d.%d", netmask >> 24, (netmask >> 16) & 0xff,
						(netmask >> 8) & 0xff, netmask & 0xff);
				m = out;
			}
		} else {
			m = "255.255.255.255";
		}
		safe_set_net(local_tun_addr, 32, 0, &L.tun_addr, L.tun_addr_s, "-l");
		safe_set_net(m, 32, 0, &L.tun_mask, L.tun_mask_s, "invalid mask in -l");
	}

	if(count > 1)
		bb_error_msg_and_die("options -E -S -X must specified only once");
	switch(c & (V_BIT|E_BIT)) {
		case 0:
			if(vA <= 1 && vB == 0 && vC == 0) {
				count = conv_abc_xsrhve('H'); break;
			}
			if(vB == 0 && vC == 0) {
				count = conv_abc_xsrhve('R'); break;
			}
			if(vA == 2 && vB == 1 && vC == 0) {
				count = conv_abc_xsrhve('X'); break;
			}
			if(vA == 2 && vB == 0 && vC == 1) {
				count = conv_abc_xsrhve('S'); break;
			}
			if(vA == 1 && vB != 0 && vB <= 2 && vC < (2-vB)) {
				count = conv_abc_xsrhve('X'); break;
			}
			if(vA == 1 && vB == 0 && vC <= 2) {
				count = conv_abc_xsrhve('S'); break;
			}
			if(vA == 0 && vB != 0 && vB <= 3 && vC <= (3-vB)) {
				count = conv_abc_xsrhve('X'); break;
			}
			if(vA == 0 && vB == 0 && vC <= 3) {
				count = conv_abc_xsrhve('S'); break;
			}
			bb_error_msg("warning: switch on -V option");
			/* Fall Through */
		case V_BIT:
			count = conv_abc_xsrhve('V');
			break;
		case E_BIT:
			count = conv_abc_xsrhve('E');
			safe_set_net(X, 8, 0, NULL, L.var_host_net, "-E");
			break;
		default:
			bb_error_msg_and_die("options -V -E are mutually exclusive");
	}
	if((vA + vB + vC) > count)
		bb_error_msg_and_die("too many IP from options -A -B -C for -%c option, expect count is %d",
					vT, count);

	if(smtu != NULL) {
		bb_common_bss.mtu = bb_xparse_i_number(BB_PN_SET_LOWER | BB_PN_SET_UPPER | BB_PN_ERR_STR1,
					smtu, 512, 65536, "invalid %s mtu (expect [512-65536])");
	} else {
		bb_common_bss.mtu = 1500;
	}
	L.tcp_port = parse_spec_port(stcp_port, "tcp");
	L.udp_port = parse_spec_port(sudp_port, "udp");

	sudp_port = Gargv[1] != NULL ? Gargv[1] : Gargv[0];
	port = bb_xparse_i_number(BB_PN_SET_LOWER1 | BB_PN_SET_UPPER | BB_PN_ERR_STR1, sudp_port,
							65535, "invalid -p %s, must 1-65535");
	memset(ba, 0, sizeof(*ba));
	if(bind_addr != NULL) {
		if(Gargv[1] != NULL)
			bb_error_msg_and_die("can not set bind address for client mode");
		bb_gethostbyname(BB_GHBN_SOCKADDR_IN | BB_GHBN_ADD_ERR_MSG | BB_GHBN_ERR_DIE, bind_addr,
			ba, "can not resolve bind address");
	} else if(Gargv[1] == NULL) {
		ba->sin_family = AF_INET;
	} else if(Gargv[1] != NULL) {
		bb_gethostbyname(BB_GHBN_SOCKADDR_IN | BB_GHBN_ADD_ERR_MSG | BB_GHBN_ERR_DIE, Gargv[0],
			ba, "can not resolve server address");
	}
	ba->sin_port = htons(port);
	if(allow_remote_addr != NULL) {
		if(Gargv[1] != NULL)
			bb_error_msg_and_die("can not set allow address for client mode");
		bb_common_bss.allow_remote_addr_set = bb_gethostbyname(BB_GHBN_IN_ADDR | BB_GHBN_ADD_ERR_MSG | BB_GHBN_ERR_DIE,
			allow_remote_addr, &bb_common_bss.allow_remote_addr, "can not resolve allow address");
	}
	if(run != NULL && run[0] != '/')
		bb_error_msg_and_die("-r script need absolute pathname");
	if(password_file != NULL) {
		ssize_t r;
		int pf = bb_xopen(password_file, BB_OPEN_MINUS_IS_STDIN | BB_OPEN_ERR, O_RDONLY);
		r = safe_read(pf, bb_common_bss.cfg_buf, sizeof(bb_common_bss.cfg_buf));
		if(r < 0)
			bb_perror_msg_and_die("I/O error while reading %s", password_file);
		if(r == sizeof(bb_common_bss.cfg_buf))
			bb_error_msg_and_die("password from file `%s` too long", password_file);
		if(r < 6)
			bb_error_msg_and_die("password from file `%s` too short", password_file);
		close(pf);
		bb_common_bss.secret = xmalloc(r);
		memcpy(bb_common_bss.secret, bb_common_bss.cfg_buf, r);
		bb_common_bss.secret_len = r;
	}
	return c & P_BIT;
}

#ifdef DEBUG
#define set_daemon(p)
#else
static void set_daemon(int add_port)
{
	if(add_port == 0) {
		xdaemon(XDAEMON_PIDFILE);
	} else {
		char *p;

		add_port = ntohs(add_port);
		p = xasprintf("/var/run/%s_%d.pid", bb_applet_name, add_port);
		xdaemon(XDAEMON_PF_ARG, p);
	}
	bb_openlog(LOG_PID, LOG_DAEMON);
}
#endif

#ifdef STANDALONE
void bb_show_usage (void)
{
	printf("Usage: %s %s\n%s\n", bb_applet_name, vu4tun_trivial_usage, vu4tun_full_usage);
	exit (2);
}

int main(int argc, char **argv)
#else
int vu4tun_main()
#endif  /* STANDALONE */
{
	struct sockaddr_in addr;
	int s, po, P, mtu;

#ifdef STANDALONE
	libbb_init(argc, argv);
#endif
	P = parse_args(&addr);
	po = bb_common_bss.tcp_udp_icmp_proto_only;
	mtu = bb_common_bss.mtu;

	s = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM |
					BB_XSOCKET_FLG_PROTO_SPEC |
					BB_XSOCKET_FLG_ERR_DIE, IPPROTO_UDP);
	if(Gargv[1] != NULL) {
		/* client */
		if(connect(s, (struct sockaddr *) &addr, sizeof(addr)))
			bb_perror_msg_and_die("connect");
		set_daemon(P ? addr.sin_port : 0);
		for(;;)
			client(s, po, mtu);
	} else {
		/* server */
		bb_xbind(s, (struct sockaddr *) &addr, sizeof(addr));
		set_daemon(P ? addr.sin_port : 0);
		for(;;)
			server(s, po, mtu);
	}
}

/* The strong IPv4+TCP/UDP/ICMP header compressor.

This make less size IP and may be ICMP/TCP/UDP header special
for the vu4tun with tun device and any IPv4 traffic.
For this traffic, minimal less is 3 byte always (but it strange traffic)
and maximal less is 20/25/23 bytes size.
It do not send any added traffic always.
It is not like VJ TCP header compression.
It have not state and full stable with loses packets.
Maximal compress for a client<->server traffic, one byte
plus for one manual (popular or delay-sensitive) host
or two byte encode for hosts in configure-net/24 or
second popular host encode as configure-net.0 IP address.


The "other configured hosts" can have any IP address, for example,
may be from your configured_net/23.

Size less: L-H+T+t+C+S+D+P+F+c+A+U+s+d+y+e+l

The encoder do not send IP length. Use vtund protocol datagram length.
Into summarize formula is L=0(!) - it do compensation vtund porocol's addes.

If IP datagram have IP option, encoder set reserved flag of IP header
and out 1 byte real IHL. Into summarize formula is H={0,1}.

The encoder do not send TTL, harcoded: tunnel src ttl=64, other src ttl=63.
Into summarize formula is T=1.

The encoder do not send tos. Into summarize formula is t=1.

Encoder do not send IP header CRC. It is not bad idea,
we have tunnel protocol CRC. Into summarize formula is C=2.

Notes: Version, IHL, IP options, etc - see RFC791.

"Version":
0XXX - IPv4, 1XXX - IPv6 for future version of header compression.
X0XX - If fragmet offset is 0-0x1f
X1XX - If fragmet offset is 0x20-0x1fff
       Into summarize formula is F={0,1}.

"Version"  Var   Octets                 S or D
 XX S|D        z=0|a=255|l=127|r=240-254      specified by options
     1      *  Tunnel IP                 4        - (always)


     0      V  ***.***.***.***           0    -V 0/0
     0      a  zalr                      3    and -A n1=[0..18]-n1-n3 * hosts
     0      b  zalr.***                  2    and -B n2=[0..18]-n1-n3 * nets/24
     0      c  zalr.***.***              1    and -C n3=[0..18]-n1-n3 * nets/16

     0      E  ***.***.***               1    or -E 1 * nets/8
     0      a  za                        3    and -A n1=[1-2]-n1-n3 * hosts
     0      b  za.***                    2    and -B n2=[1-2]-n1-n3 * nets/24
     0      c  za.***.***                1    and -C n3=[1-2]-n1-n3 * nets/16

     0      S  ***.***                   2    or -S 1 * nets/16
     0      a  za                        3    and -A n1=[1-2]-n1-n3 * hosts
     0      b  za.***                    2    and -B n2=[1-2]-n1-n3 * nets/24
     0      c  za.***.***                1    and -C n3=[1-2]-n1-n3 * nets/16

     0      X  ***                       3    or -X 1 * nets/24
     0      a  za                        3    and -A n1=[1-2]-n1-n3 * hosts
     0      b  za.***                    2    and -B n2=[1-2]-n1-n3 * nets/24
     0      c  za.***.***                1    and -C n3=[1-2]-n1-n3 * nets/16

     0      R  zalr                      3    or -A [2-18] * host

     0      H  -                         4    or -A 1 * host

Encodes variants may be: V or Va or Vb or Vc or Vab or Vac or Vbc or Vabc or
E or Ea or Eaa or Eab or Eac or Eb or Ebb or Ebc or Ec or Ecc or
S or Sa or Saa or Sab or Sac or Sb or Sbb or Sbc or Sc or Scc or
X or Xa or Xaa or Xab or Xac or Xb or Xbb or Xbc or Xc or Xcc or R or H

  "IHL"  Protocol    P
  00XX   Other       0
  01XX   ICMP        1
  10XX   TCP         1
  11XX   UDP         1

For a ICMP/TCP/UDP datagram:
Encoder do not send ICMP/TCP/UDP CRC.
Into summarize formula is c=2 or 0 for non-ICMP/TCP/UDP.

For a ICMP datagram:
"IHL"   Type      y
XX0X    0         1
XX1X    any       0

"IHL"   Code      e
XXX0    0         1
XXX1    any       0


For a TCP datagram:

If ACK=0 (one/two packet into a TCP session) encoder do not send
Acknowledgment Number 4 bytes. Into summarize formula is A={4,0}.

If URG=0 (the Urgent pointer is 0), encoder do not send Urgent pointer
2 byte. Into summarize formula is U={2,0}.

"IHL"   Source port                             s
XX0X    0 - local specified port (1024-65535)   1
XX0X    1-1023                                  1
XX1X    1024-65535                              0
2 bites used from reserved flags of TCP header.

"IHL"   Destination port                        d
XXX0    0 - remote specified port (1024-65535)  1
XXX0    1-1023                                  1
XXX1    1024-65535                              0
2 bites used from reserved flags of TCP header.


For a UDP datagram:

"IHL"   Source port                             s
XX0X    0 - local specified port (256-65535)    1
XX0X    1-255                                   1
XX0X    256-65535                               0

"IHL"   Destination port                        d
XXX0    0 - remote specified port (256-65535)   1
XXX0    1-255                                   1
XXX1    256-65535                               0

Encoder do not send UDP header length (2 bytes) if it set really identical
with manually checker total_size-total_headers_size. If not compare,
send as is. Into summarize formula is l=2 (0 for non-UDP).

Minimal less size for non-ICMP/TCP/UDP otherIP-otherIP
with IP options and set fragments: L-H+C+T+t = 3 bytes.

Minimal less size for option preconfigures:
ICMP/TCP/UDP-only tunnel+one_host-tunnel+one_host
with IP options and set fragments:
L-H+C+T+t+S+D+P+c = 0-1+2+1+1+4+4+1+2 = 14 bytes. Its great that UDP header (8)!
For vtun over UDP encapsulation result is
IP(20)+UDP(8)+VTUND(2)-2(L)-14=14 and for original MTU=1500 need
set tunnel MTU=1500-14=1486

IP header:
"Version"+"IHL" [Tos] Id(2) ReservedFlag+DF+MF [FragmentOffset] [TTL]
[SourceIP](0|1|2|4) [DestinationIP](0|1|2|4)
[{Real_IHL}(1) {IP_options}(4*any)]

For ICMP:
[Type] [Code]

For TCP:
Seq(4) Header_lenght(1/2)+(1/2 for ports)
Source_port(1|2) Destination_port(1|2)
CWR_ECE_URG_ACK_PSH_RST_SYN_FIN [Ack_number](0|4) [Urgent_pointer](0|2)
Windows(2) [TCP_options](any)

For UDP:
Source_port(1|2) [Destination_port](0|2)

And any data:
[Data](any)

*/

static int encode_zalr(int num)
{
	if(num == 0)
		return 0;
	if(num == 1)
		return 255;
	if(num == 2)
		return 127;
	return 240 + num - 3;
}

static int encode_addr(struct LR *lr, char octets_out[IPv4_chars], int *ta)
{
#define c1 octets_out[0]
#define c2 octets_out[1]
#define c3 octets_out[2]
#define c4 octets_out[3]
	int c;
	char cmp[IPv4_chars], cmpx[IPv4_chars];

	if(c1 == 0 || c1 == 127) /* Strange IP */ return -1;

	if(!memcmp(lr->tun_addr_s, octets_out, IPv4_chars)) {
		*ta = 1;
		return 0;
	}

	if(lr->var == 'H' && !memcmp(lr->var_host_net, octets_out, IPv4_chars))
		return 0;

	for(c = 0; c < lr->vara; c++) {
		if(!memcmp(lr->vara_hosts[c], octets_out, IPv4_chars)) {
			c1 = encode_zalr(c);
			return 1;
		}
	}

	memcpy(cmp, lr->var_host_net, sizeof(cmp));
	cmp[3] = c4;
	if(lr->var == 'X' && !memcmp(cmp, octets_out, sizeof(cmp))) {
		if(c4 == 0 || U_C(c4) == 255)        /* unsupport address ip1.ip2.ip3.0|255 */
			return -1;
		c1 = c4;
		return 1;
	}
	for(c = 0; c < lr->varb; c++) {
		memcpy(cmpx, lr->varb_nets[c], sizeof(cmpx));
		cmpx[3] = c4;
		if(!memcmp(cmpx, octets_out, sizeof(cmpx))) {
			c1 = encode_zalr(c + lr->vara);
			c2 = c4;
			return 2;
		}
	}
	cmp[2] = c3;
	if(lr->var == 'S' && !memcmp(cmp, octets_out, sizeof(cmp))) {
		if(c3 == 0 || U_C(c3) == 255 || c4 == 0 || U_C(c4) == 255) /* unsupport address ip1.ip2.0|255.0|255 */
			return -1;
		c1 = c3;
		c2 = c4;
		return 2;
	}
	for(c = 0; c < lr->varc; c++) {
		memcpy(cmpx, lr->varc_nets[c], sizeof(cmpx));
		cmpx[2] = c3;
		cmpx[3] = c4;
		if(!memcmp(cmpx, octets_out, sizeof(cmpx))) {
			c1 = encode_zalr(c + lr->vara + lr->varb);
			c2 = c3;
			c3 = c4;
			return 3;
		}
	}
	cmp[1] = c2;
	if(lr->var == 'E' && !memcmp(cmp, octets_out, sizeof(cmp))) {
		/* unsupport address ip1.0|255.0|255.0|255 */
		if(c2 == 0 || U_C(c2) == 255 || c3 == 0 || U_C(c3) == 255 || c4 == 0 || U_C(c4) == 255)
			return -1;
		c1 = c2;
		c2 = c3;
		c3 = c4;
		return 3;
	}
	if(lr->var != 'V')
		return -1;      /* do not allow any another addresses */
	return IPv4_chars;
#undef c1
#undef c2
#undef c3
#undef c4
}


int encoder(char *in, int len)
{
	int c, c1, c2, c3, c4;
	int ldata, t;
	int id, ihl;

	int sta = 0;    /* detect sourse IP is tunnel IP */

	int sa_octets = 0;    /* sourse IP  */
	char sa_octets_out[IPv4_chars];

	int dta = 0;    /* detect destination IP is tunnel IP */
	int da_octets = 0;    /* destination IP  */
	char da_octets_out[IPv4_chars];

	int sp, dp;     /* TCP/UDP source and destination ports,
			   ICMP - type:code */
	int fo, rf_df_mf;  /* fragment offset, Reserved flag, DF, MF */
	int seq = 0, ack = 0, tcp_flg = 0, urg = 0, win = 0, hl = 0;
	char *start_other_data;

	int n_packet_buf = 0;

	char is_tcp_udp_icmp;
	char *start_opt = 0, *start_tcp_opts = 0;

#define getc1() U_C(in[n_packet_buf++])
#define getc2(r) do { c1 = getc1(); c2 = getc1(); r = (c1 * 256) + c2; } while(0)
#define getc4(r) do { getc2(r); c3 = getc1(); c4 = getc1(); r = r*256*256 + (c3 * 256) + c4; } while(0)
#define skip(N) n_packet_buf += N

	/* decode IPv4 header */
	c = getc1();
	c1 = c >> 4;
	if(c1 != 4) /* Is not IPv4 */ return -1;
	c2 = c & 0xf;
	if(c2 < 5 || (c2 * 4) > len) /* Strange IP header length */ return -1;

	ihl = c2;       /* real IHL */

	/* tos */
	skip(1);

	/* Total IP length */
	getc2(t);
	if(t != len) /* XXX, Total IP length do not eq vtun packet length */ return -1;

	getc2(id);

	c = getc1();
	c1 = c & 0x80;
	if(c1) /* XXX, Found set reserved IP flag */ return -1;

	/* DF = c & 0x40; MF = c & 0x20; */
	rf_df_mf = c & 0xe0;

	/* Fragment offset */
	c &= 0x1f;
	c2 = getc1();
	fo = (c * 256) + c2;

	/* ttl */
	skip(1);

	/* protocol */
	is_tcp_udp_icmp = getc1();
	if(bb_common_bss.tcp_udp_icmp_proto_only &&
			is_tcp_udp_icmp != IS_ICMP &&
			is_tcp_udp_icmp != IS_TCP &&
			is_tcp_udp_icmp != IS_UDP)
		return -1;

	/* chksum */
	skip(2);

	/* Source IP */
	sa_octets_out[0] = getc1();
	sa_octets_out[1] = getc1();
	sa_octets_out[2] = getc1();
	sa_octets_out[3] = getc1();

	sa_octets = encode_addr(&bb_common_bss.l, sa_octets_out, &sta);
	if(sa_octets < 0)
		return -1;

	/* Destination IP */
	da_octets_out[0] = getc1();
	da_octets_out[1] = getc1();
	da_octets_out[2] = getc1();
	da_octets_out[3] = getc1();

	da_octets = encode_addr(&bb_common_bss.r, da_octets_out, &dta);
	if(da_octets < 0)
		return -1;

	c = ihl - 5;
	if(c) {
		start_opt = in + n_packet_buf;
		skip( 4 * c );
	}

	ldata = len - ihl * 4;

	/* decode ICMP/TCP/UDP header */
	sp = dp = 0;
	if(is_tcp_udp_icmp == IS_TCP || is_tcp_udp_icmp == IS_UDP) {
		/* Source port */
		getc2(sp);
		if(sp == 0) /* Strange source port */ return -1;
		if(is_tcp_udp_icmp == IS_TCP && sp == L.tcp_port)
			sp = 0;
		if(is_tcp_udp_icmp == IS_UDP && sp == L.udp_port)
			sp = 0;

		/* Destination port */
		getc2(dp);
		if(dp == 0) /* Strange destination port */ return -1;
		if(is_tcp_udp_icmp == IS_TCP && dp == R.tcp_port)
			dp = 0;
		if(is_tcp_udp_icmp == IS_UDP && dp == R.udp_port)
			dp = 0;
	}
	if(is_tcp_udp_icmp == IS_TCP) {
		/* Seq number */
		getc4(seq);

		/* Ack number */
		getc4(ack);

		/* Data offset */
		c = getc1();
		hl = c >> 4;
		if(hl < 5) /* XXX, Strange TCP IP header length */ return -1;

		/* reserved flags */
		c2 = c & 0xf;
		if(c2) /* XXX, Strange, reserved flags set */ return -1;

		/* CWR ECE URG ACK PSH RST SYN FIN */
		tcp_flg = getc1();

		/* Windows */
		getc2(win);

		/* chksum */
		skip(2);

		/* Urgent pointer */
		getc2(urg);
		if(((tcp_flg & 0x20) && !urg) || ((tcp_flg & 0x20)==0 && urg)) {
			/* XXX, Strange, URG flag with Urgent pointer values */
			return -1;
		}

		/* TCP Options */
		c = hl - 5;
		start_tcp_opts = in + n_packet_buf;
		skip( 4 * c );
		/* TCP data */
		ldata -= hl * 4;
	} else if(is_tcp_udp_icmp == IS_UDP) {
		/* Length */
		getc2(t);
		if(t != ldata) {
			/* FIX ME! I do not know, it is problem or not */
			return len;
		}

		/* Checksum */
		skip(2);

		/* UDP data */
		ldata -= 8;
	} else if(is_tcp_udp_icmp == IS_ICMP) {
		/* Type */
		sp = getc1();
		/* Code */
		dp = getc1();

		/* Checksum */
		skip(2);

		/* ICMP data */
		ldata -= 4;
	}

	if(ldata < 0) /* XXX, Strange data size */ return -1;

	start_other_data = in + n_packet_buf;

	/* encode */

	c = 0;  /* "Version"+"IHL" */
	if(fo > 0x1f)
		c |= 0x40;
	if(sta)
		c |= 0x20;      /* source IP = tunnel src, [SourceIP](0) */
	if(dta)
		c |= 0x10;      /* dest IP = tunnel dst, [DestinationIP](0) */
	/* "IHL" */
	if(is_tcp_udp_icmp == IS_ICMP)
		c |= 0x04;      /* encode "Version" as proto is ICMP */
	else if(is_tcp_udp_icmp == IS_TCP)
		c |= 0x08;      /* encode "Version" as proto is TCP */
	else if(is_tcp_udp_icmp == IS_UDP)
		c |= 0x0c;      /* encode "Version" as proto is UDP */
	if(is_tcp_udp_icmp == IS_ICMP) {
		if(sp)
			c |= 2; /* encode "IHL" as Type(1) is not 0 */
		if(dp)
			c |= 1; /* encode "IHL" as Code(1) is not 0 */
	}
	/* encode "IHL" bits Source_port(1|2) Destination_port(1|2) */
	if(is_tcp_udp_icmp == IS_TCP) {
		if(sp >= 1024)
			c |= 2; /* encode "IHL" as Source_port(2) */
		if(dp >= 1024)
			c |= 1; /* encode "IHL" as Destination_port(2) */
	}
	if(is_tcp_udp_icmp == IS_UDP) {
		if(sp >= 256)
			c |= 2; /* encode "IHL" as Source_port(2) */
		if(dp >= 256)
			c |= 1; /* encode "IHL" as Destination_port(2) */
	}

	n_packet_buf = sizeof(uint16_t);  /* offset for lenght */
#define OUT(c)  in[n_packet_buf++] = (c)
#define OUT2(c) do { \
		in[n_packet_buf++] = ((c) >> 8) & 0xff; \
		in[n_packet_buf++] = (c) & 0xff; \
		} while(0)
#define OUT4(c) do { \
		in[n_packet_buf++] = (c) >> 24; \
		in[n_packet_buf++] = ((c) >> 16) & 0xff; \
		in[n_packet_buf++] = ((c) >> 8) & 0xff; \
		in[n_packet_buf++] = (c) & 0xff; \
		} while(0)

	OUT(c);         /* out "Version"+"IHL" */

	OUT2(id);

	/* ReservedFlag+DF+MF and 5 low bits of fragment offset */
	c1 = rf_df_mf | (fo & 0x1f);
	OUT(c1);
	/* high 8 bits of fragment offset if need */
	if((fo > 0x1f))
		OUT(fo >> 8);

	/* send protocol if non-ICMP/TCP/UDP */
	if(is_tcp_udp_icmp != IS_ICMP && is_tcp_udp_icmp != IS_TCP && is_tcp_udp_icmp != IS_UDP)
		OUT(is_tcp_udp_icmp);

	/* [SourceIP](0|1|2|4) */
	for(c = 0; c < sa_octets; c++)
		OUT(sa_octets_out[c]);

	/* [DestinationIP](0|1|2|4) */
	for(c = 0; c < da_octets; c++)
		OUT(da_octets_out[c]);

	if(ihl > 5) {
		/* {Real_IHL}(1) {IP_options}(4*any) */
		OUT(ihl);
		c2 = ihl - 5;
		while(c2) {
			OUT(*start_opt++);
			OUT(*start_opt++);
			OUT(*start_opt++);
			OUT(*start_opt++);
			c2--;
		}
	}

	if(is_tcp_udp_icmp == IS_ICMP) {
		/* [Type] [Code] */
		if(sp) OUT(sp);
		if(dp) OUT(dp);
	} else if(is_tcp_udp_icmp == IS_TCP) {
		OUT4(seq);

		/* Header_lenght(1/2)+(1/2 for ports) */
		c2 = hl << 4;
		c3 = 0;
		if(sp < 1024) {
			c3 = sp >> 8;
			c3 <<= 2;
		}
		if(dp < 1024)
			c3 |= dp >> 8;
		OUT(c2 | c3);

		/* Source_port(1|2) */
		if(sp >= 1024)
			OUT(sp >> 8);
		OUT(sp & 0xff);
		/* Destination_port(1|2) */
		if(dp >= 1024)
			OUT(dp >> 8);
		OUT(dp & 0xff);

		/* CWR_ECE_URG_ACK_PSH_RST_SYN_FIN */
		OUT(tcp_flg);

		if(ack)
			OUT4(ack);
		if((tcp_flg & 0x20))
			OUT2(urg);

		OUT2(win);

		c4 = hl - 5;
		if(c4) {
			/* TCP_options(any) */
			while(c4) {
				OUT(*start_tcp_opts++);
				OUT(*start_tcp_opts++);
				OUT(*start_tcp_opts++);
				OUT(*start_tcp_opts++);
				c4--;
			}
		}
	} else if(is_tcp_udp_icmp == IS_UDP) {
		/* Source_port(1|2) */
		if(sp >= 256)
			OUT2(sp);
		else
			OUT(sp);
		/* [Destination_port](1|2) */
		if(dp >= 256)
			OUT2(dp);
		else
			OUT(dp);
	}

	/* other data */
	memmove(in + n_packet_buf, start_other_data, ldata);
	/* out size */
	len = n_packet_buf + ldata;

	{
		/* added flag */
		uint16_t c16 = htons(len | VU4TUN_IPH_COMPRESSED);
		memcpy(in, &c16, sizeof(c16));
	}
	return len;
#undef OUT
#undef OUT2
#undef OUT4
}

static int decode_addr(struct LR *lr, int tun, char *in, char out[IPv4_chars])
{
	int n_packet_buf = 0;
	char *a = NULL;
	int c;

	if(tun) {
		memcpy(out, lr->tun_addr_s, IPv4_chars);
		return 0;
	}
	if(lr->var == 'H') {
		memcpy(out, lr->var_host_net, IPv4_chars);
		return 0;
	}
	c = getc1();
	if(c == 0) {
		if(lr->vara) {
			a = lr->vara_hosts[0];
		} else if(lr->varb) {
			a = memcpy(out, lr->varb_nets[0], IPv4_chars);
			a[3] = getc1();
		} else if(lr->varc) {
			a = memcpy(out, lr->varc_nets[0], IPv4_chars);
			a[2] = getc1();
			a[3] = getc1();
		} else {
			return -1;
		}
	} else if(c == 255) {
		if(lr->vara > 1) {
			a = lr->vara_hosts[1];
		} else if(lr->varb > (1 - lr->vara)) {
			a = memcpy(out, lr->varb_nets[1 - lr->vara], IPv4_chars);
			a[3] = getc1();
		} else if(lr->varc > (1 - lr->vara - lr->varb)) {
			a = memcpy(out, lr->varc_nets[1 - lr->vara - lr->varb], IPv4_chars);
			a[2] = getc1();
			a[3] = getc1();
		} else {
			return -1;
		}
	} else if(lr->var == 'X' || lr->var == 'S' || lr->var == 'E') {
		a = memcpy(out, lr->var_host_net, IPv4_chars);
		if(lr->var == 'X') {
			a[3] = c;
		} else if(lr->var == 'S') {
			a[2] = c;
			a[3] = getc1();
		} else if(lr->var == 'E') {
			a[1] = c;
			a[2] = getc1();
			a[3] = getc1();
		} else {
			return -1;
		}
	} else if(c == 127) {
		if(lr->vara > 2) {
			a = lr->vara_hosts[2];
		} else if(lr->varb > (2 - lr->vara)) {
			a = memcpy(out, lr->varb_nets[2 - lr->vara], IPv4_chars);
			a[3] = getc1();
		} else if(lr->varc > (2 - lr->vara - lr->varb)) {
			a = memcpy(out, lr->varc_nets[2 - lr->vara - lr->varb], IPv4_chars);
			a[2] = getc1();
			a[3] = getc1();
		} else {
			return -1;
		}
	} else if(c >= 240 && c <= 254) {
		if(lr->vara > (3 + c - 240)) {
			a = lr->vara_hosts[3 + c - 240];
		} else if(lr->varb > (3 - lr->vara + c - 240)) {
			a = memcpy(out, lr->varb_nets[3 + c - 240 - lr->vara], IPv4_chars);
			a[3] = getc1();
		} else if(lr->varc > (3 - lr->vara - lr->varb + c - 240)) {
			a = memcpy(out, lr->varc_nets[3 + c - 240 - lr->vara - lr->varb], IPv4_chars);
			a[2] = getc1();
			a[3] = getc1();
		} else {
			return -1;
		}
	} else if(lr->var == 'V') {
		a = out;
		a[0] = c;
		a[1] = getc1();
		a[2] = getc1();
		a[3] = getc1();
	} else {
		return -1;
	}
	if(a != out)
		memcpy(out, a, IPv4_chars);
	return n_packet_buf;
}

int decoder(char *in, int len)
{
	int need_real_ihl, fo, need_fo2, c, c1, c2, c3, c4, t;
	int ihl, id, rf_df_mf;
	int sta, dta;
	int sp, dp, hl, tcp_flg;
	int is_tcp_udp_icmp = 0;
	char *ip_opt_buf = NULL;
	char pseudo_header[12];
	char *out = bb_common_bss.hc_out;

	int n_packet_buf = 0, n_decode_buf = 0;

	c = getc1();    /* "Version"+"IHL" */
	sta = c & 0x20;
	dta = c & 0x10;

	if((c & 0x0c) == 0x0c)
		is_tcp_udp_icmp = IS_UDP;
	else if((c & 0x08))
		is_tcp_udp_icmp = IS_TCP;
	else if((c & 0x04))
		is_tcp_udp_icmp = IS_ICMP;

	need_fo2 = c & 0x40;

	sp = c & 2;
	dp = c & 1;

	getc2(id);

	c = getc1();
	need_real_ihl = c & 0x80;
	rf_df_mf = c & 0x60;
	fo = c - rf_df_mf;
	if(need_fo2) {
		c2 = getc1();
		c2 <<= 8;
		fo |= c2;
	}
	c1 = fo >> 8;
	rf_df_mf |= c1;
	fo &= 0xff;

	if(!is_tcp_udp_icmp)
		is_tcp_udp_icmp = getc1();

	c = decode_addr(&bb_common_bss.r, sta, in + n_packet_buf, pseudo_header);
	if(c < 0)
		return -1;
	n_packet_buf += c;

	c = decode_addr(&bb_common_bss.l, dta, in + n_packet_buf, pseudo_header + 4);
	if(c < 0)
		return -1;
	n_packet_buf += c;

	if(need_real_ihl) {
		ihl = getc1();
		c = ihl - 5;
		ip_opt_buf = in;
		skip(4 * c);
	} else {
		ihl = 5;
	}

	/* make IP header */
#define OUT(c)  out[n_decode_buf++] = (c)
#define OUT2(c) do { \
		out[n_decode_buf++] = ((c) >> 8) & 0xff; \
		out[n_decode_buf++] = (c) & 0xff; \
		} while(0)
#define OUT4(c) do { \
		out[n_decode_buf++] = (c) >> 24; \
		out[n_decode_buf++] = ((c) >> 16) & 0xff; \
		out[n_decode_buf++] = ((c) >> 8) & 0xff; \
		out[n_decode_buf++] = (c) & 0xff; \
		} while(0)

	/* Version+IHL */
	c = 4 << 4;     /* IPv4 */
	c |= ihl;
	OUT(c);

	/* tos */
	OUT(0);

	/* set size later */
	OUT2(0);

	OUT2(id);

	OUT(rf_df_mf);
	OUT(fo);

	/* ttl */
	OUT(sta ? 64 : 63);
	OUT(is_tcp_udp_icmp);

	/* cksum place */
	OUT2(0);

	/* Source IP */
	OUT(pseudo_header[0]);
	OUT(pseudo_header[1]);
	OUT(pseudo_header[2]);
	OUT(pseudo_header[3]);
	/* Destination IP */
	OUT(pseudo_header[4]);
	OUT(pseudo_header[5]);
	OUT(pseudo_header[6]);
	OUT(pseudo_header[7]);

	/* IP options */
	c = ihl - 5;
	while(c) {
		OUT(*ip_opt_buf++);
		OUT(*ip_opt_buf++);
		OUT(*ip_opt_buf++);
		OUT(*ip_opt_buf++);
		c--;
	}

	/* start a protocols data */
	if(is_tcp_udp_icmp == IS_TCP) {
		int seq, urg;

		getc4(seq);

		hl = getc1();   /* TCP header length */
		if(sp) {
			OUT(getc1());
			OUT(getc1());
		} else {
			sp = (hl & 0xc) >> 2;
			sp = (sp << 8) | getc1();
			if(!sp)
				sp = R.tcp_port;
			OUT2(sp);
		}
		if(dp) {
			OUT(getc1());
			OUT(getc1());
		} else {
			dp = hl & 0x3;
			dp = (dp << 8) | getc1();
			if(!dp)
				dp = L.tcp_port;
			OUT2(dp);
		}

		OUT4(seq);

		tcp_flg = getc1();
		if((tcp_flg & 0x10)) {
			/* ack */
			OUT(getc1());
			OUT(getc1());
			OUT(getc1());
			OUT(getc1());
		} else
			OUT4(0);

		hl &= 0xf0;
		OUT(hl);
		hl >>= 4;

		OUT(tcp_flg);

		/* URG */
		if((tcp_flg & 0x20)) {
			getc2(urg);
		} else {
			urg = 0;
		}

		/* Windows */
		OUT(getc1());
		OUT(getc1());

		/* TCP cksum (later) */
		OUT2(0);

		/* URG */
		OUT2(urg);

		/* TCP options */
		c = hl - 5;
		while(c) {
			OUT(getc1());
			OUT(getc1());
			OUT(getc1());
			OUT(getc1());
			c--;
		}
	}

	if(is_tcp_udp_icmp == IS_UDP) {
		if(sp) {
			c = getc1() * 256;
			c += getc1();
			OUT2(c);
		} else {
			sp = getc1();
			if(!sp)
				sp = R.udp_port;
			OUT2(sp);
		}
		if(dp) {
			OUT(getc1());
			OUT(getc1());
		} else {
			dp = getc1();
			if(!dp)
				dp = L.udp_port;
			OUT2(dp);
		}

		/* Length (later) */
		OUT2(0);
		/* chksum (later) */
		OUT2(0);
	}

	if(is_tcp_udp_icmp == IS_ICMP) {
		/* Type */
		if(sp)
			OUT(getc1());
		else
			OUT(0);
		/* Code */
		if(dp)
			OUT(getc1());
		else
			OUT(0);
		/* chksum (later) */
		OUT2(0);
	}

	/* body offset */
	c1 = n_decode_buf;
	/* protocol data */
	t = len - n_packet_buf;
	for(c = t; c > 0; c--) {
		OUT(getc1());
	}

	/* uncompressed (original) size */
	len = n_decode_buf;
	if((len & 1)) {
		/* 16 bit pad for cksum */
		OUT(0);
	}
	/* Version+IHL + ToS */
	n_decode_buf = 2;
	/* IP Total size */
	OUT2(len);

	/* make cksum of IP header */
	c = htons(in_cksum(out, ihl * 4, 0));
	/* Version+IHL + ToS + Total_IP_length(2) +
	   Id(2) + RF_DF_MF + Fragment_offset + TTL + Protocol */
	n_decode_buf = 1 + 1 + 2 + 2 + 1 + 1 + 1 + 1;
	OUT2(c);

	t = len;
	if((t & 1)) {
		/* 16 bit pad for cksum */
		t++;
	}

	if(is_tcp_udp_icmp == IS_ICMP) {
		/* make cksum into ICMP header */
		c = htons(in_cksum(out + ihl * 4, t - ihl * 4, 0));
		/* IP_header + Type + Code */
		n_decode_buf = ihl * 4 + 1 + 1;
		OUT2(c);
	}
	if(is_tcp_udp_icmp == IS_TCP || is_tcp_udp_icmp == IS_UDP) {
		unsigned short cksum_ph;
		if(is_tcp_udp_icmp == IS_UDP) {
			/* set UDP packet length */
			/* IP_header + SrcPort(2) + DstPort(2) */
			n_decode_buf = ihl * 4 + 2 * 2;
			c = len - ihl * 4;
			OUT2(c);
		}
		/* make pseudo header */
		pseudo_header[8] = 0;
		pseudo_header[9] = is_tcp_udp_icmp;
		c = len - ihl * 4;
		pseudo_header[10] = (c & 0xff00) >> 8;
		pseudo_header[11] = c & 0xff;
		cksum_ph = in_cksum(pseudo_header, 12, 0);
		/* cksum TCP/UDP packet after pseudo header */
		c = htons(in_cksum(out + ihl * 4, t - ihl * 4, ~cksum_ph));
		/* set cksum into TCP/UDP packet */
		if(is_tcp_udp_icmp == IS_TCP) {
			/* IP_header +
			   SrcPort(2) DstPort(2) SEQ(4) ACK(4)
			   hl tcp_flg Windows(2) */
			n_decode_buf = ihl * 4 + 4 * 4;
			OUT2(c);
		} else {
			/* IP_header +
			   SrcPort(2) + DstPort(2) + UDP_Length(2) */
			n_decode_buf = ihl * 4 + 3 * 2;
			OUT2(c);
		}
	}
	return len;     /* flags is clear */
}

#endif /* APPLET_USAGE */
