/* vi: set sw=4 ts=4: */
/*
 * Modprobe written from scratch for BusyBox
 *
 * Copyright (c) 2008 Timo Teras <timo.teras@iki.fi>
 * Copyright (c) 2008 Vladimir Dronnikov
 * Copyright (c) 2014 Vladimir Oleynik
 *
 * Licensed under GPLv2 or later, see file LICENSE in this source tree.
 */
#ifdef APPLET_USAGE
#define modprobe_trivial_usage \
	"[-alrDbq] MODULE [symbol=value ...]"
#define modprobe_full_usage \
	"Options:\n" \
	"\t-a\tLoad multiple MODULEs\n" \
	"\t-l\tList (MODULE is a pattern)\n" \
	"\t-D\tShow dependencies\n" \
	"\t-r\tRemove module (stacks) or do autoclean\n" \
	"\t-b\tApply blacklist to module names too" \
	"\t-q\tQuiet"
#else

#include "busybox.h"
#include "modutils.h"
#include <sys/utsname.h>
#include <fnmatch.h>
#include <string.h>
#include <fcntl.h>
#include "config_read.h"

//#define DBG(fmt, ...) bb_error_msg("%s: " fmt, __func__, ## __VA_ARGS__)
#define DBG(...) ((void)0)

/* Note that unlike older versions of modules.dep/depmod (busybox and m-i-t),
 * we expect the full dependency list to be specified in modules.dep.
 * Older versions would only export the direct dependency list.
 */


/* -a and -D _are_ in fact compatible */
#define MODPROBE_COMPLEMENTARY ("l--arD:r--alD:a--lr:D--rl")
//#define MODPROBE_OPTS  "acd:lnrt:C:b"
enum {
	OPT_INSERT_ALL   = 1,  /* a */
	OPT_LIST_ONLY    = 2,  /* l */
	OPT_REMOVE       = 4,  /* r */
	OPT_SHOW_DEPS    = 8,  /* D */
	OPT_BLACKLIST    = 16, /* b */
	OPT_SILENT       = 32, /* q */
	//OPT_DUMP_ONLY  = 0,  /* c */
	//OPT_DIRNAME    = 0,  /* d */
	//OPT_SHOW_ONLY  = 0,  /* n */
	//OPT_RESTRICT   = 0,  /* t */
	//OPT_VERONLY    = 0,  /* V */
	//OPT_CONFIGFILE = 0,  /* C */
};

#ifdef CONFIG_FEATURE_LONG_OPTIONS
static const struct option modprobe_long_options[] = {
	{ "all", 0, NULL, 'a' },
	{ "list", 0, NULL, 'l' },
	{ "remove", 0, NULL, 'r' },
	/* module-init-tools 3.11.1 has only long opt --show-depends
	 * but no short -D, we provide long opt for scripts which
	 * were written for 3.11.1: */
	{ "show-depends", 0, NULL, 'D' },
	{ "use-blacklist", 0, NULL, 'b' },
	{ "quiet", 0, NULL, 'q' },
	/*
	{ "verbose", 0, NULL, 'v' },
	{ "syslog", 0, NULL, 's' },
	*/
	{ 0, 0, 0, 0 }
};
#endif

#define MODULE_FLAG_LOADED              0x0001
#define MODULE_FLAG_NEED_DEPS           0x0002
/* "was seen in modules.dep": */
#define MODULE_FLAG_FOUND_IN_MODDEP     0x0004
#define MODULE_FLAG_BLACKLISTED         0x0008

struct module_entry { /* I'll call it ME. */
	unsigned flags;
	char *modname; /* stripped of /path/, .ext and s/-/_/g */
	const char *probed_name; /* verbatim as seen on cmdline */
	char *options; /* options from config files */
	llist_t *realnames; /* strings. if this module is an alias, */
	/* real module name is one of these. */
//Can there really be more than one? Example from real kernel?
	llist_t *deps; /* strings. modules we depend on */
};

#define DB_HASH_SIZE 256

CALC_BSS struct BB_COMMON_BSS
{
	llist_t *probes; /* MEs of module(s) requested on cmdline */
	char *cmdline_mopts; /* module options from cmdline */
	int num_unresolved_deps;
	/* bool. "Did we have 'symbol:FOO' requested on cmdline?" */
	int need_symbols;
	struct utsname uts;
	llist_t *db[DB_HASH_SIZE]; /* MEs of all modules ever seen (caching for speed) */
} bb_common_bss;
#define probes bb_common_bss.probes
#define cmdline_mopts bb_common_bss.cmdline_mopts
#define num_unresolved_deps bb_common_bss.num_unresolved_deps
#define need_symbols bb_common_bss.need_symbols
#define uts bb_common_bss.uts
#define db bb_common_bss.db

static int read_config(const char *path);

static char *gather_options_str(char *opts, const char *append)
{
	if (append) {
		if (opts == NULL) {
			opts = xstrdup(append);
		} else {
			char *new_opts = xasprintf("%s %s", opts, append);
			free(opts);
			return new_opts;
		}
	}
	return opts;
}

/* These three functions called many times, optimizing for speed.
 * Users reported minute-long delays when they runn iptables repeatedly
 * (iptables use modprobe to install needed kernel modules).
 */
static struct module_entry *helper_get_module(const char *module, int create)
{
	char modname[MODULE_NAME_LEN];
	struct module_entry *e;
	llist_t *l;
	unsigned i;
	unsigned hash;

	filename2modname(module, modname);

	hash = 0;
	for (i = 0; modname[i]; i++)
		hash = ((hash << 5) + hash) + modname[i];
	hash %= DB_HASH_SIZE;

	for (l = db[hash]; l; l = l->link) {
		e = (struct module_entry *) l->data;
		if (strcmp(e->modname, modname) == 0)
			return e;
	}
	if (!create)
		return NULL;

	e = zmalloc(sizeof(*e));
	e->modname = xstrdup(modname);
	db[hash] = llist_add_to(db[hash], e);

	return e;
}
#define get_or_add_modentry(module) helper_get_module(module, 1)

static void add_probe(const char *name)
{
	struct module_entry *m;

	m = get_or_add_modentry(name);
	if (!(bss_flags & (OPT_REMOVE | OPT_SHOW_DEPS))
			&& (m->flags & MODULE_FLAG_LOADED)) {
		DBG("skipping %s, it is already loaded", name);
		return;
	}

	DBG("queuing %s", name);
	m->probed_name = name;
	m->flags |= MODULE_FLAG_NEED_DEPS;
	probes = llist_add_to_end(probes, m);
	num_unresolved_deps++;
	if (strncmp(m->modname, "symbol:", 7) == 0)
		need_symbols = 1;
}

static int
config_file_action(const char *filename, struct stat *statbuf, void *userdata)
{
	config_read_t *p;
	struct module_entry *m;
	int rc = TRUE;

	(void)statbuf;
	(void)userdata;

	if (bb_basename(filename)[0] == '.')
		return TRUE;

	p = config_read_prepare(filename, 0);
	if (p == NULL)
		return FALSE;

	while ((p = config_read(p, "#~\0\2")) != NULL) {
//Use index_in_strings?
		if (strcmp(p->tokens[0], "alias") == 0) {
			/* alias <wildcard> <modulename> */
			llist_t *l;
			char wildcard[MODULE_NAME_LEN];
			char *rmod;

			if (p->ntokens < 3)
				continue;
			filename2modname(p->tokens[1], wildcard);

			for (l = probes; l; l = l->link) {
				m = (struct module_entry *) l->data;
				if (fnmatch(wildcard, m->modname, 0) != 0)
					continue;
				rmod = filename2modname(p->tokens[2], NULL);
				m->realnames = llist_add_to(m->realnames, rmod);

				if (m->flags & MODULE_FLAG_NEED_DEPS) {
					m->flags &= ~MODULE_FLAG_NEED_DEPS;
					num_unresolved_deps--;
				}

				m = get_or_add_modentry(rmod);
				if (!(m->flags & MODULE_FLAG_NEED_DEPS)) {
					m->flags |= MODULE_FLAG_NEED_DEPS;
					num_unresolved_deps++;
				}
			}
		} else if (strcmp(p->tokens[0], "options") == 0) {
			int i;
			/* options <modulename> <option...> */
			if (p->ntokens < 3)
				continue;
			m = get_or_add_modentry(p->tokens[1]);
			for(i = 2; i < p->ntokens; i++)
				m->options = gather_options_str(m->options, p->tokens[i]);
		} else if (strcmp(p->tokens[0], "include") == 0) {
			/* include <filename> */
			read_config(p->tokens[1]);
		} else if (strcmp(p->tokens[0], "blacklist") == 0) {
			/* blacklist <modulename> */
			get_or_add_modentry(p->tokens[1])->flags |= MODULE_FLAG_BLACKLISTED;
		}
	}
	return rc;
}

static int read_config(const char *path)
{
	return recursive_action(path, 1, 0, 1, config_file_action, NULL, NULL);
}


static const char *humanly_readable_name(struct module_entry *m)
{
	/* probed_name may be NULL. modname always exists. */
	return m->probed_name ? m->probed_name : m->modname;
}

static char *parse_and_add_kcmdline_module_options(char *options, const char *modulename)
{
	char kcmdline_buf[4096];
	FILE *p_c;
	char *kcmdline;
	char *kptr;
	int len;

	p_c = fopen("/proc/cmdline", "r");
	if(p_c == NULL)
		return options;
	len = fread(kcmdline_buf, 1, sizeof(kcmdline_buf) - 1, p_c);
	fclose(p_c);
	if(len == 0)
		return options;
	kcmdline_buf[len] = '\0';

	kcmdline = kcmdline_buf;
	len = strlen(modulename);
	while ((kptr = strsep(&kcmdline, "\n\t ")) != NULL) {
		if (strncmp(modulename, kptr, len) != 0)
			continue;
		kptr += len;
		if (*kptr != '.')
			continue;
		/* It is "modulename.xxxx" */
		kptr++;
		if (strchr(kptr, '=') != NULL) {
			/* It is "modulename.opt=[val]" */
			options = gather_options_str(options, kptr);
		}
	}
	return options;
}

/* Reverse list order. */
static llist_t* FAST_FUNC llist_rev(llist_t *list)
{
	llist_t *rev = NULL;

	while (list) {
		llist_t *next = list->link;

		list->link = rev;
		rev = list;
		list = next;
	}
	return rev;
}

/* Return: similar to bb_init_module:
 * 0 on success,
 * -errno on open/read error,
 * errno on init_module() error
 */
/* NB: OPT_SILENT bit suppresses ONLY non-existent modules,
 * not deleted ones (those are still listed in modules.dep).
 * module-init-tools version 3.4:
 * # modprobe bogus
 * FATAL: Module bogus not found. [exitcode 1]
 * # modprobe -q bogus            [silent, exitcode still 1]
 * but:
 * # rm kernel/drivers/net/dummy.ko
 * # modprobe -q dummy
 * FATAL: Could not open '/lib/modules/xxx/kernel/drivers/net/dummy.ko': No such file or directory
 * [exitcode 1]
 */
static int do_modprobe(struct module_entry *m)
{
	int rc, first;

	if (!(m->flags & MODULE_FLAG_FOUND_IN_MODDEP)) {
		if (!(bss_flags & OPT_SILENT))
			bb_error_msg("module %s not found in modules.dep",
				humanly_readable_name(m));
		return -ENOENT;
	}
	DBG("do_modprob'ing %s", m->modname);

	if (!(bss_flags & OPT_REMOVE))
		m->deps = llist_rev(m->deps);

	if (0) {
		llist_t *l;
		for (l = m->deps; l; l = l->link)
			DBG("dep: %s", l->data);
	}

	first = 1;
	rc = 0;
	while (m->deps) {
		struct module_entry *m2;
		char *fn, *options;

		rc = 0;
		fn = llist_pop(&m->deps); /* we leak it */
		m2 = get_or_add_modentry(fn);

		if (bss_flags & OPT_REMOVE) {
			/* modprobe -r */
			if (m2->flags & MODULE_FLAG_LOADED) {
				rc = bb_delete_module(m2->modname, O_EXCL);
				if (rc) {
					if (first) {
						bb_error_msg("can't unload module %s: %s",
							humanly_readable_name(m2),
							moderror(rc));
						break;
					}
				} else {
					m2->flags &= ~MODULE_FLAG_LOADED;
				}
			}
			/* do not error out if *deps* fail to unload */
			first = 0;
			continue;
		}

		options = m2->options;
		m2->options = NULL;
		options = parse_and_add_kcmdline_module_options(options, m2->modname);
		if (m == m2)
			options = gather_options_str(options, cmdline_mopts);

		if (bss_flags & OPT_SHOW_DEPS) {
			printf(options ? "insmod %s/%s/%s %s\n"
					: "insmod %s/%s/%s\n",
				"/lib/modules", uts.release, fn, options);
			free(options);
			continue;
		}

		if (m2->flags & MODULE_FLAG_LOADED) {
			DBG("%s is already loaded, skipping", fn);
			free(options);
			continue;
		}

		rc = bb_init_module(fn, options);
		DBG("loaded %s '%s', rc:%d", fn, options, rc);
		if (rc == EEXIST)
			rc = 0;
		free(options);
		if (rc) {
			bb_error_msg("can't load module %s (%s): %s",
				humanly_readable_name(m2),
				fn,
				moderror(rc)
			);
			break;
		}
		m2->flags |= MODULE_FLAG_LOADED;
	}

	return rc;
}

static void load_modules_dep(void)
{
	struct module_entry *m;
	char *colon;
	config_read_t *p;

	/* Modprobe does not work at all without modules.dep,
	 * even if the full module name is given. Returning error here
	 * was making us later confuse user with this message:
	 * "module /full/path/to/existing/file/module.ko not found".
	 * It's better to die immediately, with good message. */
	p = config_read_prepare(CONFIG_DEFAULT_DEPMOD_FILE, 'e');

	while (num_unresolved_deps && (p = config_read(p, "#~\0\0")) != NULL) {
		colon = last_char_is(p->tokens[0], ':');
		if (colon == NULL)
			continue;
		*colon = 0;

		m = helper_get_module(p->tokens[0], 0);
		if (m == NULL)
			continue;

		/* Optimization... */
		if ((m->flags & MODULE_FLAG_LOADED)
		 && !(bss_flags & (OPT_REMOVE | OPT_SHOW_DEPS))
		) {
			DBG("skip deps of %s, it's already loaded", p->tokens[0]);
			continue;
		}

		m->flags |= MODULE_FLAG_FOUND_IN_MODDEP;
		if ((m->flags & MODULE_FLAG_NEED_DEPS) && (m->deps == NULL)) {
			int i;
			num_unresolved_deps--;
			m->deps = llist_add_to(m->deps, xstrdup(p->tokens[0]));
			for(i = 1; i < p->ntokens; i++)
				m->deps = llist_add_to_end(m->deps, xstrdup(p->tokens[i]));
		} else
			DBG("skipping dep line");
	}
	config_close(p);
}

static void replace(char *s, char what, char with)
{
	while (*s) {
		if (what == *s)
			*s = with;
		++s;
	}
}

int modprobe_main(void)
{
	int rc;
	unsigned opt;
	struct module_entry *me;
	char **argv;

#ifdef CONFIG_FEATURE_LONG_OPTIONS
	bb_applet_long_options = modprobe_long_options;
#endif
	opt = bb_xgetopt_ulflags ("alrDbq\0l--arD:r--alD:a--lr:D--rl");
	argv = Gargv;

	/* Goto modules location */
	xchdir("/lib/modules");
	uname(&uts);
	xchdir(uts.release);

	if (opt & OPT_LIST_ONLY) {
		int i;
		char name[MODULE_NAME_LEN];
		char *colon;
		config_read_t *p = config_read_prepare(CONFIG_DEFAULT_DEPMOD_FILE, 'e');

		for (i = 0; argv[i]; i++)
			replace(argv[i], '-', '_');

		while ((p = config_read(p, "#\1\0\0")) != NULL) {
			colon = last_char_is(p->tokens[0], ':');
			if (!colon)
				continue;
			*colon = '\0';
			filename2modname(p->tokens[0], name);
			if (!argv[0])
				puts(p->tokens[0]);
			else {
				for (i = 0; argv[i]; i++) {
					if (fnmatch(argv[i], name, 0) == 0) {
						puts(p->tokens[0]);
					}
				}
			}
		}
		return EXIT_SUCCESS;
	}

#if 0
	/* Yes, for some reason -l ignores -s... */
	if (opt & INSMOD_OPT_SYSLOG)
		logmode = LOGMODE_SYSLOG;
#endif

	if (!argv[0]) {
		if (opt & OPT_REMOVE) {
			/* "modprobe -r" (w/o params).
			 * "If name is NULL, all unused modules marked
			 * autoclean will be removed".
			 */
			if (bb_delete_module(NULL, O_NONBLOCK | O_EXCL) != 0)
				bb_perror_msg_and_die("rmmod");
		}
		return EXIT_SUCCESS;
	}

	/* Retrieve module names of already loaded modules */
	{
		config_read_t *parser = config_read_prepare("/proc/modules", 0);
		while ((parser = config_read(parser, "\0\1\0\0 ")) != NULL)
			get_or_add_modentry(parser->tokens[0])->flags |= MODULE_FLAG_LOADED;
	}

	if (opt & (OPT_INSERT_ALL | OPT_REMOVE)) {
		/* Each argument is a module name */
		do {
			DBG("adding module %s", *argv);
			add_probe(*argv++);
		} while (*argv);
	} else {
		/* First argument is module name, rest are parameters */
		DBG("probing just module %s", *argv);
		add_probe(argv[0]);
		cmdline_mopts = parse_cmdline_module_options(argv, /*quote_spaces:*/ 1);
	}

	/* Happens if all requested modules are already loaded */
	if (probes == NULL)
		return EXIT_SUCCESS;

	read_config("/etc/modprobe.conf");
	read_config("/etc/modprobe.d");
	if (need_symbols)
		read_config("modules.symbols");
	load_modules_dep();
	if (num_unresolved_deps) {
		read_config("modules.alias");
		load_modules_dep();
	}

	rc = 0;
	while ((me = llist_pop(&probes)) != NULL) {
		if (me->realnames == NULL) {
			DBG("probing by module name");
			/* This is not an alias. Literal names are blacklisted
			 * only if '-b' is given.
			 */
			if (!(opt & OPT_BLACKLIST)
			 || !(me->flags & MODULE_FLAG_BLACKLISTED)
			) {
				rc |= do_modprobe(me);
			}
			continue;
		}

		/* Probe all real names for the alias */
		do {
			char *realname = llist_pop(&me->realnames);
			struct module_entry *m2;

			DBG("probing alias %s by realname %s", me->modname, realname);
			m2 = get_or_add_modentry(realname);
			if (!(m2->flags & MODULE_FLAG_BLACKLISTED)
			 && (!(m2->flags & MODULE_FLAG_LOADED)
			    || (opt & (OPT_REMOVE | OPT_SHOW_DEPS)))
			) {
//TODO: we can pass "me" as 2nd param to do_modprobe,
//and make do_modprobe emit more meaningful error messages
//with alias name included, not just module name alias resolves to.
				rc |= do_modprobe(m2);
			}
			free(realname);
		} while (me->realnames != NULL);
	}

	return (rc != 0);
}
#endif /* APPLET_USAGE */
