/*
 * arp                This file contains an implementation of the command
 *              that maintains the kernel's ARP cache.  It is derived
 *              from Berkeley UNIX arp(8), but cleaner and with sup-
 *              port for devices other than Ethernet.
 *
 * NET-TOOLS    A collection of programs that form the base set of the
 *              NET-3 Networking Distribution for the LINUX operating
 *              system.
 *
 * Version:     arp.c 1.12 1999/03/24 09:49:43
 *
 * Maintainer:  Bernd 'eckes' Eckenfels, <net-tools@lina.inka.de>
 *
 * Author:      Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
 *
 *              This program is free software; you can redistribute it
 *              and/or  modify it under  the terms of  the GNU General
 *              Public  License as  published  by  the  Free  Software
 *              Foundation;  either  version 2 of the License, or  (at
 *              your option) any later version.
 *
 * Busybox port by Vladimir Oleynik (C) 2008 <dzo@simtreas.ru>
 * - improve getargs():
 *   do not use temp buffer, can set <space># as begin a comment,
 *   do not break arg if start quote. Size optimization. Example:
 *   can usage a"b\"c"d# # as one arg `ab"cd#' only.
 *
 */
#ifdef APPLET_USAGE
#define arp_trivial_usage \
	"[-vn] [-D] [<HW>] [-i <if>] [-a|d|s] [<hostname>] <hwaddr>"
#define arp_full_usage \
	"Manipulate the system ARP cache.\n\n" \
	"Usage:\n" \
	" [-vn] [-i <if>] [-a] [<hostname>]  <-Display ARP cache\n" \
	" [-v] [-i <if>] -d <hostname> [pub][nopub]  <-Delete ARP entry\n" \
	" [-vnD] [-i <if>] -f <filename>     <-Add entry from file\n" \
	" [-v] [-i <if>] -s <hostname> <hwaddr> [temp][nopub]        <-Add entry\n" \
	" [-v] [-i <if>] -s  <hostname> <hwaddr> [netmask <nm>] pub  <-''-\n" \
	" [-v] [-i <if>] -Ds <hostname> <if> [netmask <nm>] pub      <-''-\n\n" \
	"Options:\n" \
	"\t-a\t\tdisplay (all) hosts in alternative (BSD) style\n" \
	"\t-s\t\tset a new ARP entry\n" \
	"\t-d\t\tdelete a specified entry\n" \
	"\t-v\t\tbe verbose\n" \
	"\t-n\t\tdont resolve names\n" \
	"\t-i\t\tspecify network interface (e.g. eth0)\n" \
	"\t-D\t\tread <hwaddr> from given device\n" \
	"\t-f\t\tread new entries from file"
#else

#include "busybox.h"
#include <sys/ioctl.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <ctype.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include "bb_inet.h"
#include "af.h"

#define PATH_PROCNET_ARP "/proc/net/arp"
#define DFLT_AF "inet"
#define DFLT_HW "ether"

#define ENOSUPP(B) bb_error_msg("feature `%s' not supported.", B);

#define E_USAGE   4
#define E_VERSION 5

#define FEATURE_ARP

CALC_BSS struct BB_COMMON_BSS {
	int opt_n;              /* do not resolve addresses     */
	int opt_v;              /* debugging output flag        */
	int opt_D;              /* HW-address is devicename     */
	int opt_e;              /* 0=BSD output, other=new linux*/
	const struct aftype *ap;      /* current address family */
	const struct hwtype *hw;      /* current hardware type  */
	int sockfd;             /* active socket descriptor     */
	char device[16];        /* current device               */
	int title;
} bb_common_bss;
#define opt_n bb_common_bss.opt_n
#define opt_v bb_common_bss.opt_v
#define opt_e bb_common_bss.opt_e
#define opt_a (bss_flags & 1)
#define opt_D (bb_common_bss.opt_D & (1<<6))
#define ap bb_common_bss.ap
#define hw bb_common_bss.hw
#define sockfd bb_common_bss.sockfd
#define device bb_common_bss.device
#define title bb_common_bss.title

#define FLAG_NUM      0x0FFF    /* do not rresolve, see inet_common.c */

static void version(void) ATTRIBUTE_NORETURN;
static void version(void)
{
    bb_default_error_retval = E_VERSION;
    bb_error_msg_and_die("net-tools 1.52\narp 1.85 (1999-01-05)\nHW: (ether) +ETHER");
}

static void usage(void) ATTRIBUTE_NORETURN;
static void usage(void)
{
    bb_default_error_retval = E_USAGE;
    bb_show_usage();
}

static void No_ARP_entry_for_host(const char *host) ATTRIBUTE_NORETURN;
static void No_ARP_entry_for_host(const char *host)
{
	printf("No ARP entry for %s\n", host);
	exit(-1);
}

/* Delete an entry from the ARP cache. */
static void arp_del(char **args)
{
    char host[128];
    struct arpreq req;
    struct sockaddr sa;
    int flags = 0;
    int err;

    memset((char *) &req, 0, sizeof(req));

    /* Resolve the host name. */
    if (*args == NULL)
	bb_error_msg_and_die("need host name");
    safe_strncpy(host, *args, sizeof host);
    ap->input(2, host, &sa);
    /* If a host has more than one address, use the correct one! */
    memcpy((char *) &req.arp_pa, (char *) &sa, sizeof(struct sockaddr));

    req.arp_flags = ATF_PERM;
    args++;
    while (*args != NULL) {
	if (ENABLE_FEATURE_ARP_VERBOSE && opt_v)
	    fprintf(stderr, "args=%s\n", *args);
	if (!strcmp(*args, "pub")) {
	    flags |= 1;
	    args++;
	    continue;
	}
	if (!strcmp(*args, "priv")) {
	    flags |= 2;
	    args++;
	    continue;
	}
	if (!strcmp(*args, "temp")) {
	    req.arp_flags &= ~ATF_PERM;
	    args++;
	    continue;
	}
	if (!strcmp(*args, "trail")) {
	    req.arp_flags |= ATF_USETRAILERS;
	    args++;
	    continue;
	}
	if (!strcmp(*args, "dontpub")) {
#ifdef HAVE_ATF_DONTPUB
	    req.arp_flags |= ATF_DONTPUB;
#else
	    ENOSUPP("ATF_DONTPUB");
#endif
	    args++;
	    continue;
	}
	if (!strcmp(*args, "auto")) {
#ifdef HAVE_ATF_MAGIC
	    req.arp_flags |= ATF_MAGIC;
#else
	    ENOSUPP("ATF_MAGIC");
#endif
	    args++;
	    continue;
	}
	if (!strcmp(*args, "dev")) {
	    if (*++args == NULL)
		usage();
	    safe_strncpy(device, *args, sizeof(device));
	    args++;
	    continue;
	}
	if (!strcmp(*args, "netmask")) {
	    if (*++args == NULL)
		usage();
	    if (strcmp(*args, "255.255.255.255") != 0) {
		safe_strncpy(host, *args, sizeof host);
		ap->input(2, host, &sa);
		memcpy((char *) &req.arp_netmask, (char *) &sa,
		       sizeof(struct sockaddr));
		req.arp_flags |= ATF_NETMASK;
	    }
	    args++;
	    continue;
	}
	usage();
    }
    if (flags == 0)
	flags = 3;

    strcpy(req.arp_dev, device);

    err = -1;

    /* Call the kernel. */
    if (flags & 2) {
	if (ENABLE_FEATURE_ARP_VERBOSE && opt_v)
	    bb_error_msg("SIOCDARP(nopub)");
	if ((err = ioctl(sockfd, SIOCDARP, &req) < 0)) {
	    if (errno == ENXIO) {
		if (flags & 1)
		    goto nopub;
		No_ARP_entry_for_host(host);
	    }
	    bb_perror_msg_and_die("SIOCDARP(priv)");
	}
    }
    if ((flags & 1) && (err)) {
      nopub:
	req.arp_flags |= ATF_PUBL;
	if (ENABLE_FEATURE_ARP_VERBOSE && opt_v)
	    bb_error_msg("SIOCDARP(pub)");
	if (ioctl(sockfd, SIOCDARP, &req) < 0) {
	    if (errno == ENXIO)
		No_ARP_entry_for_host(host);
	    bb_perror_msg_and_die("SIOCDARP(pub)");
	}
    }
}

/* Get the hardware address to a specified interface name */
static int arp_getdevhw(const char *ifname, struct sockaddr *sa)
{
    struct ifreq ifr;
    const struct hwtype *xhw;

    safe_ifname_cp0(&ifr, ifname);
    if (xioctl(sockfd, SIOCGIFHWADDR, &ifr, "can not get HW-Address for `%s'", ifname) < 0)
	return -1;
    memcpy((char *) sa, (char *) &(ifr.ifr_hwaddr), sizeof(struct sockaddr));

    if (ENABLE_FEATURE_ARP_VERBOSE && opt_v) {
	if (!(xhw = get_hwntype(ifr.ifr_hwaddr.sa_family)) || (xhw->sprint == 0)) {
	    xhw = get_hwntype(-1);
	}
	bb_error_msg("device `%s' has HW address %s `%s'.",
			ifname, xhw->name, xhw->sprint(&ifr.ifr_hwaddr));
    }
    return 0;
}

/* Set an entry in the ARP cache. */
static int arp_set(char **args)
{
    char host[128];
    struct arpreq req;
    struct sockaddr sa;
    int flags;

    memset((char *) &req, 0, sizeof(req));

    /* Resolve the host name. */
    if (*args == NULL)
	bb_error_msg_and_die("need host name");
    safe_strncpy(host, *args++, (sizeof host));
    ap->input(2, host, &sa);
    /* If a host has more than one address, use the correct one! */
    memcpy((char *) &req.arp_pa, (char *) &sa, sizeof(struct sockaddr));

    /* Fetch the hardware address. */
    if (*args == NULL)
	bb_error_msg_and_die("need hardware address");
    if (opt_D) {
	if (arp_getdevhw(*args++, &req.arp_ha) < 0)
	    return -1;
    } else {
	if (hw->input(*args++, &req.arp_ha) < 0)
	    bb_error_msg_and_die("invalid hardware address");
    }

    /* Check out any modifiers. */
    flags = ATF_PERM | ATF_COM;
    while (*args != NULL) {
	if (!strcmp(*args, "temp")) {
	    flags &= ~ATF_PERM;
	    args++;
	    continue;
	}
	if (!strcmp(*args, "pub")) {
	    flags |= ATF_PUBL;
	    args++;
	    continue;
	}
	if (!strcmp(*args, "priv")) {
	    flags &= ~ATF_PUBL;
	    args++;
	    continue;
	}
	if (!strcmp(*args, "trail")) {
	    flags |= ATF_USETRAILERS;
	    args++;
	    continue;
	}
	if (!strcmp(*args, "dontpub")) {
#ifdef HAVE_ATF_DONTPUB
	    flags |= ATF_DONTPUB;
#else
	    ENOSUPP("ATF_DONTPUB");
#endif
	    args++;
	    continue;
	}
	if (!strcmp(*args, "auto")) {
#ifdef HAVE_ATF_MAGIC
	    flags |= ATF_MAGIC;
#else
	    ENOSUPP("ATF_MAGIC");
#endif
	    args++;
	    continue;
	}
	if (!strcmp(*args, "dev")) {
	    if (*++args == NULL)
		usage();
	    safe_strncpy(device, *args, sizeof(device));
	    args++;
	    continue;
	}
	if (!strcmp(*args, "netmask")) {
	    if (*++args == NULL)
		usage();
	    if (strcmp(*args, "255.255.255.255") != 0) {
		strcpy(host, *args);
		ap->input(2, host, &sa);
		memcpy((char *) &req.arp_netmask, (char *) &sa,
		       sizeof(struct sockaddr));
		flags |= ATF_NETMASK;
	    }
	    args++;
	    continue;
	}
	usage();
    }

    /* Fill in the remainder of the request. */
    req.arp_flags = flags;

    strcpy(req.arp_dev, device);

    /* Call the kernel. */
    if (ENABLE_FEATURE_ARP_VERBOSE && opt_v)
	bb_error_msg("SIOCSARP()");
    xioctl(sockfd, SIOCSARP, &req, "\1SIOCSARP");
    return 0;
}

/* Split the input string into multiple fields. */
static int getargs(char *string, char *arguments[])
{
    char *sp = string, *ptr = string;
    int argc = 0;
    char want;

    while (argc < 31) {
	/* Ignore leading whitespace before argument */
	while (*ptr == ' ' || *ptr == '\t')
		ptr++;
	/* start comment or empty or end of line? */
	if(*ptr == '#' || *ptr == '\0')
		break;

	/* Set argument pointer. */
	arguments[argc++] = sp;

	/* Check for any delimiters. a"b\"c"d convert to ab"cd */
	while( (want = *ptr++) != '\0' && want != ' ' && want != '\t') {
	    if (want == '"' || want == '\'') {
		/* Copy the string up to the end quote if not escaped */
		while (*ptr != '\0') {
		    if (*ptr == want && *(ptr - 1) != '\\') {
			ptr++;
			break;
		    }
		    *sp++ = *ptr++;
		}
	    } else {
		/* Just copy the one string char */
		*sp++ = want;
	    }
	}
	/* restore pointer if end of line */
	if (want == '\0')
		ptr--;
	/* set first whitespace or end of line to end of arg */
	*sp++ = '\0';
    }
    arguments[argc] = NULL;
    return argc;
}

/* Process an EtherFile */
static void arp_file(char *name)
{
    char buff[1024];
    char *args[32];
    int linenr, argc;
    FILE *fp;

    fp = bb_xfopen(name, "Ar");
    /* Read the lines in the file. */
    linenr = 0;
    while (fgets(buff, sizeof(buff), fp) != (char *) NULL) {
	linenr++;
	if (ENABLE_FEATURE_ARP_VERBOSE && opt_v)
	    fprintf(stderr, ">> %s", buff);
	strchr_set0(buff, '\n');
	argc = getargs(buff, args);
	if (argc == 0)
		continue;
	if (argc < 2) {
	    bb_error_msg("format error on line %d of etherfile %s !",
		    linenr, name);
	    continue;
	}
	if (arp_set(args) != 0)
	    bb_error_msg("can not set entry on line %d of etherfile %s !",
		    linenr, name);
    }

    (void) fclose(fp);
}


/* Print the contents of an ARP request block. */
static void arp_disp_2(char *name, int type, int arp_flags, char *hwa, char *mask, char *dev)
{
    const struct hwtype *xhw;
    char flags[10];

    xhw = get_hwntype(type);
    if (xhw == NULL)
	xhw = get_hwntype(ARPHRD_ETHER);

    if (title++ == 0) {
	printf("Address\t\t\tHWtype\tHWaddress\t    Flags Mask\t\t  Iface\n");
    }
    /* Setup the flags. */
    flags[0] = '\0';
    if (arp_flags & ATF_COM)
	strcat(flags, "C");
    if (arp_flags & ATF_PERM)
	strcat(flags, "M");
    if (arp_flags & ATF_PUBL)
	strcat(flags, "P");
#ifdef HAVE_ATF_MAGIC
    if (arp_flags & ATF_MAGIC)
	strcat(flags, "A");
#endif
#ifdef HAVE_ATF_DONTPUB
    if (arp_flags & ATF_DONTPUB)
	strcat(flags, "!");
#endif
    if (arp_flags & ATF_USETRAILERS)
	strcat(flags, "T");

    if (!(arp_flags & ATF_NETMASK))
	mask = "";

    printf("%-23.23s\t", name);

    if (!(arp_flags & ATF_COM)) {
	if (arp_flags & ATF_PUBL)
	    printf("%-8.8s%-20.20s", "*", "*");
	else
	    printf("%-8.8s%-20.20s", "", "(incomplete)");
    } else {
	printf("%-8.8s%-20.20s", xhw->name, hwa);
    }

    printf("%-6.6s%-15.15s %s\n", flags, mask, dev);
}

/* Print the contents of an ARP request block. */
static void arp_disp(char *name, char *ip, int type, int arp_flags, char *hwa, char *mask, char *dev)
{
    const struct hwtype *xhw;

    xhw = get_hwntype(type);
    if (xhw == NULL)
	xhw = get_hwntype(ARPHRD_ETHER);

    printf("%s (%s) at ", name, ip);

    if (!(arp_flags & ATF_COM)) {
	if (arp_flags & ATF_PUBL)
	    printf("* ");
	else
	    printf("<incomplete> ");
    } else {
	printf("%s [%s] ", hwa, xhw->name);
    }

    if (arp_flags & ATF_NETMASK)
	printf("netmask %s ", mask);

    if (arp_flags & ATF_PERM)
	printf("PERM ");
    if (arp_flags & ATF_PUBL)
	printf("PUP ");
#ifdef HAVE_ATF_MAGIC
    if (arp_flags & ATF_MAGIC)
	printf("AUTO ");
#endif
#ifdef HAVE_ATF_DONTPUB
    if (arp_flags & ATF_DONTPUB)
	printf("DONTPUB ");
#endif
    if (arp_flags & ATF_USETRAILERS)
	printf("TRAIL ");

    printf("on %s\n", dev);
}


/* Display the contents of the ARP cache in the kernel. */
static void arp_show(char *name)
{
    char host[100];
    struct sockaddr sa;
    char ip[100];
    char hwa[100];
    char mask[100];
    char line[200];
    char dev[100];
    int type, flags;
    FILE *fp;
    char *hostname;
    int num, entries = 0, showed = 0;

    host[0] = '\0';

    if (name != NULL) {
	/* Resolve the host name. */
	safe_strncpy(host, name, (sizeof host));
	ap->input(2, host, &sa);
	strcpy(host, ap->sprint(&sa, 1));
    }
    /* Open the PROCps kernel table. */
    fp = bb_xfopen(PATH_PROCNET_ARP, "Ar");
    /* Bypass header -- read until newline */
    if (fgets(line, sizeof(line), fp) != (char *) NULL) {
	strcpy(mask, "-");
	strcpy(dev, "-");
	/* Read the ARP cache entries. */
	for (; fgets(line, sizeof(line), fp);) {
	    num = sscanf(line, "%s 0x%x 0x%x %100s %100s %100s\n",
			 ip, &type, &flags, hwa, mask, dev);
	    if (num < 4)
		break;

	    entries++;

	    /* if the user specified address differs, skip it */
	    if (host[0] && strcmp(ip, host))
		continue;

	    /* if the user specified device differs, skip it */
	    if (device[0] && strcmp(dev, device))
		continue;

	    showed++;
	    /* This IS ugly but it works -be */
	    if (opt_n)
		hostname = "?";
	    else {
		if (ap->input(0, ip, &sa) < 0)
		    hostname = ip;
		else
		    hostname = ap->sprint(&sa, opt_n | 0x8000);
		if (strcmp(hostname, ip) == 0)
		    hostname = "?";
	    }

	    if (opt_e)
		arp_disp_2(hostname[0] == '?' ? ip : hostname, type, flags, hwa, mask, dev);
	    else
		arp_disp(hostname, ip, type, flags, hwa, mask, dev);
	}
    }
    if (ENABLE_FEATURE_ARP_VERBOSE && opt_v)
	printf("Entries: %d\tSkipped: %d\tFound: %d\n", entries, entries - showed, showed);

    if (!showed) {
	if (host[0] && !opt_a)
	    printf("%s (%s) -- no entry\n", name, host);
	else if (host[0] || device[0]) {
	    printf("arp: in %d entries no match found.\n", entries);
	}
    }
    (void) fclose(fp);
}


int arp_main(void)
{
    char **argv;
    char *d;
    int what;

#if ENABLE_FEATURE_LONG_OPTIONS
    static const struct option longopts[] =
    {
	{"verbose", 0, NULL, 'v'},
	{"version", 0, NULL, 'V'},
	{"all", 0, NULL, 'a'},
	{"delete", 0, NULL, 'd'},
	{"file", 0, NULL, 'f'},
	{"numeric", 0, NULL, 'n'},
	{"set", 0, NULL, 's'},
	{"device", 1, NULL, 'i'},
	{"help", 0, NULL, 'h'},
	{"use-device", 0, NULL, 'D'},
	{"symbolic", 0, NULL, 'N'},
	{NULL, 0, NULL, 0}
    };
    bb_applet_long_options = longopts;
#endif

    /* Initialize variables... */
    bb_default_error_retval = -1;
    hw = get_hwntype(ARPHRD_ETHER);
    if ((ap = get_aftype(DFLT_AF)) == NULL)
	bb_error_msg_and_die("%s: address family not supported!", DFLT_AF);

    /* Fetch the command-line arguments. */
    what = bb_getopt_ulflags("afdsenDNvi:Vh?", &d);
    opt_e = what & (1<<4);
    if((what & (1<<5)) != 0)
	    opt_n = FLAG_NUM;
    if((what & (1<<7)) != 0)
	    bb_error_msg("-N not yet supported.");
#if ENABLE_FEATURE_ARP_VERBOSE
    opt_v = what & (1<<8);
#endif
    if((what & (1<<9)) != 0)
	    safe_strncpy(device, d, sizeof(device));
    if((what & (1<<10)) != 0)
	    version();
    if((what & ((1<<11)|(1<<12))) != 0)
	    usage();
    what &= (1|2|4|8);
    argv = Gargv;

    if (ap->af != AF_INET)
	bb_error_msg_and_die("%s: kernel only supports 'inet'.", ap->name);
    if (hw->alen <= 0)
	bb_error_msg_and_die("%s: hardware type without ARP support.", hw->name);
    sockfd = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM | BB_XSOCKET_FLG_ERR_DIE);
    /* Now see what we have to do here... */
    switch (what) {
    case 0:
	opt_e = 1;
	arp_show(*argv);
	break;

    case 1:                     /* show an ARP entry in the cache */
	arp_show(*argv);
	break;

    case 2:                     /* process an EtherFile */
	arp_file(*argv);
	break;

    case 4:                     /* delete an ARP entry from the cache */
	arp_del(argv);
	break;

    case 8:                     /* set an ARP entry in the cache */
	return arp_set(argv);

    }

    return 0;
}
#endif /* APPLET_USAGE */
