/*
 * arping.c - Ping hosts by ARP requests/replies
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 *
 * Author:      Alexey Kuznetsov <kuznet@ms2.inr.ac.ru>
 * Busybox port: Nick Fedchik <nick@fedchik.org.ua>
 */

#ifdef APPLET_USAGE
#define arping_trivial_usage \
	"[-fqbDUA] [-c count] [-w timeout] [-I device] [-s sender] target"
#define arping_full_usage \
	"Ping hosts by ARP requests/replies.\n\n" \
	"Options:\n" \
	"\t-f\t\tQuit on first ARP reply\n" \
	"\t-q\t\tBe quiet\n" \
	"\t-b\t\tKeep broadcasting, don't go unicast\n" \
	"\t-D\t\tDuplicated address detection mode\n" \
	"\t-U\t\tUnsolicited ARP mode, update your neighbours\n" \
	"\t-A\t\tARP answer mode, update your neighbours\n" \
	"\t-c count\tStop after sending count ARP request packets\n" \
	"\t-w timeout\tTime to wait for ARP reply, in seconds\n" \
	"\t-I device\tOutgoing interface name, default is eth0\n" \
	"\t-s sender\tSet specific sender IP address\n" \
	"\ttarget\t\tTarget IP address of ARP request"
#else
#include "busybox.h"
#include <sys/ioctl.h>
#include <sys/time.h>

#include <string.h>
#include <unistd.h>

#include "bb_inet.h"
#include <netinet/ether.h>
#include <netpacket/packet.h>
#include <net/if_arp.h>

#include "bb_sig.h"


CALC_BSS struct BB_COMMON_BSS {
	struct in_addr src;
	struct in_addr dst;
	struct sockaddr_ll me;
	struct sockaddr_ll he;
	struct timeval last;
	int dad;
	int unsolicited;
	int advert;
	int quiet;
	long timeout;
	int unicasting;
	int s;
	int sent;
	int brd_sent;
	int received;
	int brd_recv;
	int req_recv;
	struct timeval start;
	unsigned char packet[4096];
	int count;
} bb_common_bss;
#define src bb_common_bss.src
#define dst bb_common_bss.dst
#define me bb_common_bss.me
#define he bb_common_bss.he
#define last bb_common_bss.last
#define dad bb_common_bss.dad
#define unsolicited bb_common_bss.unsolicited
#define advert bb_common_bss.advert
#define quiet bb_common_bss.quiet
#define timeout bb_common_bss.timeout
#define unicasting bb_common_bss.unicasting
#define s bb_common_bss.s
#define sent bb_common_bss.sent
#define brd_sent bb_common_bss.brd_sent
#define received bb_common_bss.received
#define brd_recv bb_common_bss.brd_recv
#define req_recv bb_common_bss.req_recv
#define start bb_common_bss.start
#define packet bb_common_bss.packet
#define count bb_common_bss.count

#define broadcast_only (bss_flags & 2)
#define quit_on_reply (bss_flags & 4)


#define MS_TDIFF(tv1,tv2) ( ((tv1).tv_sec-(tv2).tv_sec)*1000 + \
			   ((tv1).tv_usec-(tv2).tv_usec)/1000 )

static int send_pack(int sock, struct in_addr *src_addr,
					 struct in_addr *dst_addr, struct sockaddr_ll *ME,
					 struct sockaddr_ll *HE)
{
	int err;
	struct timeval now;
	unsigned char buf[256];
	struct arphdr *ah = (struct arphdr *) buf;
	unsigned char *p = (unsigned char *) (ah + 1);

	ah->ar_hrd = htons(ARPHRD_ETHER);
	ah->ar_pro = htons(ETH_P_IP);
	ah->ar_hln = ME->sll_halen;
	ah->ar_pln = 4;
	ah->ar_op = advert ? htons(ARPOP_REPLY) : htons(ARPOP_REQUEST);

	memcpy(p, &ME->sll_addr, ah->ar_hln);
	p += ME->sll_halen;

	memcpy(p, src_addr, 4);
	p += 4;

	if (advert)
		memcpy(p, &ME->sll_addr, ah->ar_hln);
	else
		memcpy(p, &HE->sll_addr, ah->ar_hln);
	p += ah->ar_hln;

	memcpy(p, dst_addr, 4);
	p += 4;

	gettimeofday(&now, NULL);
	err = sendto(sock, buf, p - buf, 0, (struct sockaddr *) HE, sizeof(*HE));
	if (err == p - buf) {
		last = now;
		sent++;
		if (!unicasting)
			brd_sent++;
	}
	return err;
}

static void finish(void)
{
	if (!quiet) {
		printf("Sent %d probes (%d broadcast(s))\n", sent, brd_sent);
		printf("Received %d repl%s", received, (received > 1) ? "ies" : "y");
		if (brd_recv || req_recv) {
			printf(" (");
			if (req_recv)
				printf("%d request(s)", req_recv);
			if (brd_recv)
				printf("%s%d broadcast(s)", req_recv ? ", " : "", brd_recv);
			putchar(')');
		}
		putchar('\n');
		fflush(stdout);
	}
	if (dad)
		exit(!!received);
	if (unsolicited)
		exit(0);
	exit(!received);
}

static void catcher(void)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);

	if (start.tv_sec == 0)
		start = tv;

	if (count-- == 0
		|| (timeout && MS_TDIFF(tv, start) > timeout * 1000 + 500))
		finish();

	if (last.tv_sec == 0 || MS_TDIFF(tv, last) > 500) {
		send_pack(s, &src, &dst, &me, &he);
		if (count == 0 && unsolicited)
			finish();
	}
	alarm(1);
}

static int recv_pack(unsigned char *buf, int len, struct sockaddr_ll *FROM)
{
	struct timeval tv;
	struct arphdr *ah = (struct arphdr *) buf;
	unsigned char *p = (unsigned char *) (ah + 1);
	struct in_addr src_ip, dst_ip;

	gettimeofday(&tv, NULL);

	/* Filter out wild packets */
	if (FROM->sll_pkttype != PACKET_HOST &&
		FROM->sll_pkttype != PACKET_BROADCAST &&
		FROM->sll_pkttype != PACKET_MULTICAST)
		return 0;

	/* Only these types are recognised */
	if (ah->ar_op != htons(ARPOP_REQUEST) && ah->ar_op != htons(ARPOP_REPLY))
		return 0;

	/* ARPHRD check and this darned FDDI hack here :-( */
	if (ah->ar_hrd != htons(FROM->sll_hatype) &&
		(FROM->sll_hatype != ARPHRD_FDDI
		 || ah->ar_hrd != htons(ARPHRD_ETHER)))
		return 0;

	/* Protocol must be IP. */
	if (ah->ar_pro != htons(ETH_P_IP))
		return 0;
	if (ah->ar_pln != 4)
		return 0;
	if (ah->ar_hln != me.sll_halen)
		return 0;
	if (len < sizeof(*ah) + 2 * (4 + ah->ar_hln))
		return 0;
	memcpy(&src_ip, p + ah->ar_hln, 4);
	memcpy(&dst_ip, p + ah->ar_hln + 4 + ah->ar_hln, 4);
	if (!dad) {
		if (src_ip.s_addr != dst.s_addr)
			return 0;
		if (src.s_addr != dst_ip.s_addr)
			return 0;
		if (memcmp(p + ah->ar_hln + 4, &me.sll_addr, ah->ar_hln))
			return 0;
	} else {
		/* DAD packet was:
		   src_ip = 0 (or some src)
		   src_hw = ME
		   dst_ip = tested address
		   dst_hw = <unspec>

		   We fail, if receive request/reply with:
		   src_ip = tested_address
		   src_hw != ME
		   if src_ip in request was not zero, check
		   also that it matches to dst_ip, otherwise
		   dst_ip/dst_hw do not matter.
		 */
		if (src_ip.s_addr != dst.s_addr)
			return 0;
		if (memcmp(p, &me.sll_addr, me.sll_halen) == 0)
			return 0;
		if (src.s_addr && src.s_addr != dst_ip.s_addr)
			return 0;
	}
	if (!quiet) {
		int s_printed = 0;

		printf("%s ",
			   FROM->sll_pkttype == PACKET_HOST ? "Unicast" : "Broadcast");
		printf("%s from ",
			   ah->ar_op == htons(ARPOP_REPLY) ? "reply" : "request");
		printf("%s ", inet_ntoa(src_ip));
		printf("[%s]", ether_ntoa((struct ether_addr *) p));
		if (dst_ip.s_addr != src.s_addr) {
			printf("for %s ", inet_ntoa(dst_ip));
			s_printed = 1;
		}
		if (memcmp(p + ah->ar_hln + 4, me.sll_addr, ah->ar_hln)) {
			if (!s_printed)
				printf("for ");
			printf("[%s]",
				   ether_ntoa((struct ether_addr *) p + ah->ar_hln + 4));
		}
		if (last.tv_sec) {
			long usecs = (tv.tv_sec - last.tv_sec) * 1000000 +
				tv.tv_usec - last.tv_usec;
			long msecs = (usecs + 500) / 1000;

			usecs -= msecs * 1000 - 500;
			printf(" %ld.%03ldms\n", msecs, usecs);
		} else {
			printf(" UNSOLICITED?\n");
		}
		fflush(stdout);
	}
	received++;
	if (FROM->sll_pkttype != PACKET_HOST)
		brd_recv++;
	if (ah->ar_op == htons(ARPOP_REQUEST))
		req_recv++;
	if (quit_on_reply)
		finish();
	if (!broadcast_only) {
		memcpy(he.sll_addr, p, me.sll_halen);
		unicasting = 1;
	}
	return 1;
}

int arping_main(void)
{
	unsigned long ch;
	char *device = "eth0";
	int ifindex = 0;
	char *source = NULL;
	char *target;

	count = -1;

	ch = bb_xgetopt_ulflags("hbfDUAqI:c:w:s:\0:w-c:-1:h-:Df:AU",
					&device, &count, &timeout, &source);
	dad         = ch & 8;
	unsolicited = ch & 16;
	advert      = ch & 32;
	quiet       = ch & 64;
	bb_default_error_retval = 2;
	if( (ch & 128) != 0) {
	    if (strlen(device) > IF_NAMESIZE) {
		bb_error_msg_and_die("Interface name `%s' must be less than %d",
				device, IF_NAMESIZE);
	    }
	}

	target = *Gargv;

	s = bb_xsocket(BB_XSOCKET_FLG_PF_PACKET | BB_XSOCKET_FLG_SOCK_DGRAM | BB_XSOCKET_FLG_ERR_DIE);

	{
		struct ifreq ifr;

		safe_ifname_cp0(&ifr, device);
		xioctl(s, SIOCGIFINDEX, &ifr, "\1Interface %s not found", device);
		ifindex = ifr.ifr_ifindex;

		xioctl(s, SIOCGIFFLAGS, &ifr, "\1SIOCGIFFLAGS");
		if (!(ifr.ifr_flags & IFF_UP))
			bb_error_msg_and_die("Interface %s is down", device);
		if (ifr.ifr_flags & (IFF_NOARP | IFF_LOOPBACK)) {
			if(dad)
				bb_default_error_retval = 0;
			bb_error_msg_and_die("Interface %s is not ARPable", device);
		}
	}

	bb_gethostbyname(BB_GHBN_IN_ADDR | BB_GHBN_ERR_DIE, target, &dst);

	if (source && !inet_aton(source, &src))
		bb_error_msg_and_die("invalid source address %s", source);

	if (!dad && unsolicited && src.s_addr == 0)
		src = dst;

	if (!dad || src.s_addr) {
		struct sockaddr_in saddr;
		int probe_fd = bb_xsocket(BB_XSOCKET_FLG_ERR_DIE | BB_XSOCKET_FLG_SOCK_DGRAM);

		if (device) {
			if (setsockopt
				(probe_fd, SOL_SOCKET, SO_BINDTODEVICE, device,
				 strlen(device) + 1) == -1)
				bb_error_msg("WARNING: interface %s is ignored", device);
		}
		memset(&saddr, 0, sizeof(saddr));
		saddr.sin_family = AF_INET;
		if (src.s_addr) {
			saddr.sin_addr = src;
			bb_xbind(probe_fd, (struct sockaddr *) &saddr, sizeof(saddr));
		} else if (!dad) {
			socklen_t alen = sizeof(saddr);

			saddr.sin_port = htons(1025);
			saddr.sin_addr = dst;

			if (setsockopt_1i(probe_fd, SOL_SOCKET, SO_DONTROUTE, 1) == -1)
				bb_perror_msg("WARNING: setsockopt(SO_DONTROUTE)");
			if (connect(probe_fd, (struct sockaddr *) &saddr, sizeof(saddr)) == -1)
				bb_error_msg_and_die("connect");
			if (getsockname(probe_fd, (struct sockaddr *) &saddr, &alen) == -1)
				bb_error_msg_and_die("getsockname");
			src = saddr.sin_addr;
		}
		close(probe_fd);
	}

	me.sll_family = AF_PACKET;
	me.sll_ifindex = ifindex;
	me.sll_protocol = htons(ETH_P_ARP);
	bb_xbind(s, (struct sockaddr *) &me, sizeof(me));

	{
		socklen_t alen = sizeof(me);

		if (getsockname(s, (struct sockaddr *) &me, &alen) == -1)
			bb_error_msg_and_die("getsockname");
	}
	if (me.sll_halen == 0) {
		if(dad)
			bb_default_error_retval = 0;
		bb_error_msg_and_die("Interface \"%s\" is not ARPable (no ll address)", device);
	}
	he = me;
	memset(he.sll_addr, -1, he.sll_halen);

	if (!quiet) {
		printf("ARPING to %s", inet_ntoa(dst));
		printf(" from %s via %s\n", inet_ntoa(src), device ? device : "unknown");
	}

	if (!src.s_addr && !dad)
		bb_error_msg_and_die("no src address in the non-DAD mode");

	bb_signal((1 << SIGINT), (void (*)(int)) finish);
	bb_signal((1 << SIGALRM), (void (*)(int)) catcher);

	catcher();

	while (1) {
		sigset_t sset;
		struct sockaddr_ll from;
		socklen_t alen = sizeof(from);
		int cc;

		if ((cc = recvfrom(s, packet, sizeof(packet), 0,
				      (struct sockaddr *) &from, &alen)) < 0) {
			bb_perror_msg("recvfrom");
			continue;
		}
		sset = bb_sigblock( (1 << SIGALRM) | (1 << SIGINT) );
		recv_pack(packet, cc, &from);
		bb_sigunblock(sset);
	}
}
#endif /* APPLET_USAGE */
