/*
 * Copyright (C) 2000 Lennert Buytenhek
 *
 * Modified by Vladimir Oleynik <dzo@simtreas.ru> (c) 2005-2009 to be
 * used in busybox and size optimizations, reduce warning (-W)
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */
#ifdef APPLET_USAGE
#define brctl_trivial_usage \
	"[-Vh] or commands"
#define brctl_full_usage \
	"Options:\n" \
	"\t-V\tShow version\n" \
	"\t-h\tShow more help usage about commands"
#else

#define PACKAGE_NAME    "bridge-utils"
#define PACKAGE_VERSION "1.4"

#include "busybox.h"
#include "bb_inet.h"
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>

#if defined(__UCLIBC__) && !defined(__UCLIBC_HAS_IPV6__)
/* IPv6 address special for if_bridge.h */
struct in6_addr {
	union {
		uint8_t __u6_addr8[16];
		uint16_t __u6_addr16[8];
		uint32_t __u6_addr32[4];
      } __in6_u;
};
#endif

#define _LINUX_IN6_H
#include <linux/if_bridge.h>
#include <linux/sockios.h>

#define MAX_BRIDGES     1024
#define MAX_PORTS       1024

#define SYSFS_CLASS_NET "/sys/class/net/"
#define SYSFS_PATH_MAX  576

#define dprintf(fmt,arg...)

/* defined in net/if.h but that conflicts with linux/if.h... */
extern unsigned int if_nametoindex (const char *__ifname);
extern char *if_indextoname (unsigned int __ifindex, char *__ifname);


struct bridge_id
{
	unsigned char prio[2];
	unsigned char addr[6];
};

struct bridge_info
{
	struct bridge_id designated_root;
	struct bridge_id bridge_id;
	unsigned root_path_cost;
	struct timeval max_age;
	struct timeval hello_time;
	struct timeval forward_delay;
	struct timeval bridge_max_age;
	struct timeval bridge_hello_time;
	struct timeval bridge_forward_delay;
	u_int16_t root_port;
	unsigned char stp_enabled;
	unsigned char topology_change;
	unsigned char topology_change_detected;
	struct timeval ageing_time;
	struct timeval hello_timer_value;
	struct timeval tcn_timer_value;
	struct timeval topology_change_timer_value;
	struct timeval gc_timer_value;
};

struct fdb_entry
{
	u_int8_t mac_addr[6];
	u_int16_t port_no;
	unsigned char is_local;
	struct timeval ageing_timer_value;
};

struct port_info
{
	unsigned port_no;
	struct bridge_id designated_root;
	struct bridge_id designated_bridge;
	u_int16_t port_id;
	u_int16_t designated_port;
	u_int8_t priority;
	unsigned char top_change_ack;
	unsigned char config_pending;
	unsigned char state;
	unsigned path_cost;
	unsigned designated_cost;
	struct timeval message_age_timer_value;
	struct timeval forward_delay_timer_value;
	struct timeval hold_timer_value;
};

CALC_BSS struct BB_COMMON_BSS {
	int first;
	int br_socket_fd;
} bb_common_bss;
#define first bb_common_bss.first
#define br_socket_fd bb_common_bss.br_socket_fd


static inline unsigned long __tv_to_jiffies(const struct timeval *tv)
{
	unsigned long long jif;

	jif = 1000000ULL * tv->tv_sec + tv->tv_usec;

	return jif/10000;
}

static inline void __jiffies_to_tv(struct timeval *tv, unsigned long jiffies)
{
	unsigned long long tvusec;

	tvusec = 10000ULL*jiffies;
	tv->tv_sec = tvusec/1000000;
	tv->tv_usec = tvusec - 1000000 * tv->tv_sec;
}


static FILE *fpopen(const char *dir, const char *name)
{
	char path[SYSFS_PATH_MAX];

	snprintf(path, SYSFS_PATH_MAX, "%s/%s", dir, name);
	return bb_xfopen(path, "Wr");
}

static void fetch_id(const char *dev, const char *name, struct bridge_id *id)
{
	FILE *f = fpopen(dev, name);
	if(f != NULL) {
		fscanf(f, "%2hhx%2hhx.%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx",
				&id->prio[0], &id->prio[1],
				&id->addr[0], &id->addr[1], &id->addr[2],
				&id->addr[3], &id->addr[4], &id->addr[5]);
		fclose(f);
	}
}

/* Fetch an integer attribute out of sysfs. */
static int fetch_int(const char *dev, const char *name)
{
	FILE *f = fpopen(dev, name);
	int value = -1;

	if (f != NULL) {
		fscanf(f, "%i", &value);
		fclose(f);
	}
	return value;
}

/* Get a time value out of sysfs */
static void fetch_tv(const char *dev, const char *name, struct timeval *tv)
{
	__jiffies_to_tv(tv, fetch_int(dev, name));
}

/*
 * Convert device name to an index in the list of ports in bridge.
 *
 * Old API does bridge operations as if ports were an array
 * inside bridge structure.
 */
static int get_portno(const char *brname, const char *ifname)
{
	int i;
	int ifindex = if_nametoindex(ifname);
	int ifindices[MAX_PORTS];
	unsigned long args[4] = { BRCTL_GET_PORT_LIST,
				  (unsigned long)ifindices, MAX_PORTS, 0 };
	struct ifreq ifr;

	if (ifindex <= 0)
		goto error;

	memset(ifindices, 0, sizeof(ifindices));
	safe_ifname_cp(&ifr, brname);
	ifr.ifr_data = (char *) &args;

	if (ioctl(br_socket_fd, SIOCDEVPRIVATE, &ifr) < 0) {
		dprintf("get_portno: get ports of %s failed: %m\n", brname);
		goto error;
	}

	for (i = 0; i < MAX_PORTS; i++) {
		if (ifindices[i] == ifindex)
			return i;
	}

	dprintf("%s is not a in bridge %s\n", ifname, brname);
 error:
	return -1;
}

/* get information via ioctl */
static int old_get_bridge_info(const char *bridge, struct bridge_info *info)
{
	struct ifreq ifr;
	struct __bridge_info i;
	unsigned long args[4] = { BRCTL_GET_BRIDGE_INFO,
				  (unsigned long) &i, 0, 0 };

	memset(info, 0, sizeof(*info));
	safe_ifname_cp(&ifr, bridge);
	ifr.ifr_data = (char *) &args;

	if (ioctl(br_socket_fd, SIOCDEVPRIVATE, &ifr) < 0) {
		dprintf("%s: can't get info %m\n", bridge);
		return errno;
	}

	memcpy(&info->designated_root, &i.designated_root, 8);
	memcpy(&info->bridge_id, &i.bridge_id, 8);
	info->root_path_cost = i.root_path_cost;
	info->root_port = i.root_port;
	info->topology_change = i.topology_change;
	info->topology_change_detected = i.topology_change_detected;
	info->stp_enabled = i.stp_enabled;
	__jiffies_to_tv(&info->max_age, i.max_age);
	__jiffies_to_tv(&info->hello_time, i.hello_time);
	__jiffies_to_tv(&info->forward_delay, i.forward_delay);
	__jiffies_to_tv(&info->bridge_max_age, i.bridge_max_age);
	__jiffies_to_tv(&info->bridge_hello_time, i.bridge_hello_time);
	__jiffies_to_tv(&info->bridge_forward_delay, i.bridge_forward_delay);
	__jiffies_to_tv(&info->ageing_time, i.ageing_time);
	__jiffies_to_tv(&info->hello_timer_value, i.hello_timer_value);
	__jiffies_to_tv(&info->tcn_timer_value, i.tcn_timer_value);
	__jiffies_to_tv(&info->topology_change_timer_value,
			i.topology_change_timer_value);
	__jiffies_to_tv(&info->gc_timer_value, i.gc_timer_value);

	return 0;
}

/*
 * Get bridge parameters using either sysfs or old
 * ioctl.
 */
static int br_get_bridge_info(const char *bridge, struct bridge_info *info)
{
	DIR *dir;
	char path[SYSFS_PATH_MAX];

	snprintf(path, SYSFS_PATH_MAX, SYSFS_CLASS_NET "%s/bridge", bridge);
	dir = opendir(path);
	if (dir == NULL) {
		dprintf("path '%s' is not a directory\n", path);
		goto fallback;
	}

	memset(info, 0, sizeof(*info));
	fetch_id(path, "root_id", &info->designated_root);
	fetch_id(path, "bridge_id", &info->bridge_id);
	info->root_path_cost = fetch_int(path, "root_path_cost");
	fetch_tv(path, "max_age", &info->max_age);
	fetch_tv(path, "hello_time", &info->hello_time);
	fetch_tv(path, "forward_delay", &info->forward_delay);
	fetch_tv(path, "max_age", &info->bridge_max_age);
	fetch_tv(path, "hello_time", &info->bridge_hello_time);
	fetch_tv(path, "forward_delay", &info->bridge_forward_delay);
	fetch_tv(path, "ageing_time", &info->ageing_time);
	fetch_tv(path, "hello_timer", &info->hello_timer_value);
	fetch_tv(path, "tcn_timer", &info->tcn_timer_value);
	fetch_tv(path, "topology_change_timer",
		 &info->topology_change_timer_value);;
	fetch_tv(path, "gc_timer", &info->gc_timer_value);

	info->root_port = fetch_int(path, "root_port");
	info->stp_enabled = fetch_int(path, "stp_state");
	info->topology_change = fetch_int(path, "topology_change");
	info->topology_change_detected = fetch_int(path, "topology_change_detected");

	closedir(dir);
	return 0;

fallback:
	return old_get_bridge_info(bridge, info);
}

static int old_get_port_info(const char *brname, const char *port,
			     struct port_info *info)
{
	struct __port_info i;
	int port_index;

	memset(info, 0, sizeof(*info));

	port_index = get_portno(brname, port);
	if (port_index < 0)
		return errno;

	else {
		struct ifreq ifr;
		unsigned long args[4] = { BRCTL_GET_PORT_INFO,
					   (unsigned long) &i, port_index, 0 };

		safe_ifname_cp(&ifr, brname);
		ifr.ifr_data = (char *) &args;

		if (ioctl(br_socket_fd, SIOCDEVPRIVATE, &ifr) < 0) {
			dprintf("old can't get port %s(%d) info %m\n",
							brname, port_index);
			return errno;
		}
	}

	info->port_no = port_index;
	memcpy(&info->designated_root, &i.designated_root, 8);
	memcpy(&info->designated_bridge, &i.designated_bridge, 8);
	info->port_id = i.port_id;
	info->designated_port = i.designated_port;
	info->path_cost = i.path_cost;
	info->designated_cost = i.designated_cost;
	info->state = i.state;
	info->top_change_ack = i.top_change_ack;
	info->config_pending = i.config_pending;
	__jiffies_to_tv(&info->message_age_timer_value,
			i.message_age_timer_value);
	__jiffies_to_tv(&info->forward_delay_timer_value,
			i.forward_delay_timer_value);
	__jiffies_to_tv(&info->hold_timer_value, i.hold_timer_value);
	return 0;
}

/*
 * Get information about port on bridge.
 */
static int br_get_port_info(const char *brname, const char *port,
		     struct port_info *info)
{
	DIR *d;
	char path[SYSFS_PATH_MAX];

	snprintf(path, SYSFS_PATH_MAX, SYSFS_CLASS_NET "%s/brport", port);
	d = opendir(path);
	if (!d)
		goto fallback;

	memset(info, 0, sizeof(*info));

	fetch_id(path, "designated_root", &info->designated_root);
	fetch_id(path, "designated_bridge", &info->designated_bridge);
	info->port_no = fetch_int(path, "port_no");
	info->port_id = fetch_int(path, "port_id");
	info->designated_port = fetch_int(path, "designated_port");
	info->path_cost = fetch_int(path, "path_cost");
	info->designated_cost = fetch_int(path, "designated_cost");
	info->state = fetch_int(path, "state");
	info->top_change_ack = fetch_int(path, "change_ack");
	info->config_pending = fetch_int(path, "config_pending");
	fetch_tv(path, "message_age_timer", &info->message_age_timer_value);
	fetch_tv(path, "forward_delay_timer", &info->forward_delay_timer_value);
	fetch_tv(path, "hold_timer", &info->hold_timer_value);
	closedir(d);

	return 0;
fallback:
	return old_get_port_info(brname, port, info);
}

static int br_set(const char *bridge, const char *name,
		  unsigned long value, unsigned long oldcode)
{
	int ret;
	char path[SYSFS_PATH_MAX];
	FILE *f;

	snprintf(path, SYSFS_PATH_MAX, SYSFS_CLASS_NET "%s/%s", bridge, name);

	f = fopen(path, "w");
	if (f) {
		ret = fprintf(f, "%ld\n", value);
		fclose(f);
	} else {
		/* fallback to old ioctl */
		struct ifreq ifr;
		unsigned long args[4] = { oldcode, value, 0, 0 };

		safe_ifname_cp(&ifr, bridge);
		ifr.ifr_data = (char *) &args;
		ret = ioctl(br_socket_fd, SIOCDEVPRIVATE, &ifr);
	}

	return ret < 0 ? errno : 0;
}

static int br_set_bridge_forward_delay(const char *br, struct timeval *tv)
{
	return br_set(br, "forward_delay", __tv_to_jiffies(tv),
		      BRCTL_SET_BRIDGE_FORWARD_DELAY);
}

static int br_set_bridge_hello_time(const char *br, struct timeval *tv)
{
	return br_set(br, "hello_time", __tv_to_jiffies(tv),
		      BRCTL_SET_BRIDGE_HELLO_TIME);
}

static int br_set_bridge_max_age(const char *br, struct timeval *tv)
{
	return br_set(br, "max_age", __tv_to_jiffies(tv),
		      BRCTL_SET_BRIDGE_MAX_AGE);
}

static int br_set_ageing_time(const char *br, struct timeval *tv)
{
	return br_set(br, "ageing_time", __tv_to_jiffies(tv),
		      BRCTL_SET_AGEING_TIME);
}

static int br_set_stp_state(const char *br, int stp_state)
{
	return br_set(br, "stp_state", stp_state, BRCTL_SET_BRIDGE_STP_STATE);
}

static int br_set_bridge_priority(const char *br, int bridge_priority)
{
	return br_set(br, "priority", bridge_priority,
		      BRCTL_SET_BRIDGE_PRIORITY);
}

static int port_set(const char *bridge, const char *ifname,
		    const char *name, unsigned long value,
		    unsigned long oldcode)
{
	int ret;
	char path[SYSFS_PATH_MAX];
	FILE *f;

	snprintf(path, SYSFS_PATH_MAX, SYSFS_CLASS_NET "%s/brport/%s", ifname, name);
	f = fopen(path, "w");
	if (f) {
		ret = fprintf(f, "%ld\n", value);
		fclose(f);
	} else {
		int port_index = get_portno(bridge, ifname);

		if (port_index < 0)
			ret = port_index;
		else {
			struct ifreq ifr;
			unsigned long args[4] = { oldcode, port_index, value, 0 };

			safe_ifname_cp(&ifr, bridge);
			ifr.ifr_data = (char *) &args;
			ret = ioctl(br_socket_fd, SIOCDEVPRIVATE, &ifr);
		}
	}

	return ret < 0 ? errno : 0;
}

static int br_set_port_priority(const char *bridge, const char *port, int priority)
{
	return port_set(bridge, port, "priority", priority, BRCTL_SET_PORT_PRIORITY);
}

static int br_set_path_cost(const char *bridge, const char *port, int cost)
{
	return port_set(bridge, port, "path_cost", cost, BRCTL_SET_PATH_COST);
}

static inline void __copy_fdb(struct fdb_entry *ent,
			      const struct __fdb_entry *f)
{
	memcpy(ent->mac_addr, f->mac_addr, 6);
	ent->port_no = f->port_no;
	ent->is_local = f->is_local;
	__jiffies_to_tv(&ent->ageing_timer_value, f->ageing_timer_value);
}

static int br_read_fdb(const char *bridge, struct fdb_entry *fdbs,
		unsigned long offset, int num)
{
	FILE *f;
	int i, n;
	struct __fdb_entry fe[num];
	char path[SYSFS_PATH_MAX];

	/* open /sys/class/net/brXXX/brforward */
	snprintf(path, SYSFS_PATH_MAX, SYSFS_CLASS_NET "%s/brforward", bridge);
	f = fopen(path, "r");
	if (f) {
		fseek(f, offset*sizeof(struct __fdb_entry), SEEK_SET);
		n = fread(fe, sizeof(struct __fdb_entry), num, f);
		fclose(f);
	} else {
		/* old kernel, use ioctl */
		unsigned long args[4] = { BRCTL_GET_FDB_ENTRIES,
					  (unsigned long) fe,
					  num, offset };
		struct ifreq ifr;
		int retries = 0;

		safe_ifname_cp(&ifr, bridge);
		ifr.ifr_data = (char *) args;

	retry:
		n = ioctl(br_socket_fd, SIOCDEVPRIVATE, &ifr);

		/* table can change during ioctl processing */
		if (n < 0 && errno == EAGAIN && ++retries < 10) {
			sleep(0);
			goto retry;
		}
	}

	for (i = 0; i < n; i++)
		__copy_fdb(fdbs+i, fe+i);

	return n;
}


static int br_add_bridge(const char *brname)
{
	int ret;

#ifdef SIOCBRADDBR
	ret = ioctl(br_socket_fd, SIOCBRADDBR, brname);
	if (ret < 0)
#endif
	{
		char _br[IFNAMSIZ];
		unsigned long arg[3]
			= { BRCTL_ADD_BRIDGE, (unsigned long) _br };

		strncpy(_br, brname, IFNAMSIZ);
		ret = ioctl(br_socket_fd, SIOCSIFBR, arg);
	}

	return ret < 0 ? errno : 0;
}

static int br_del_bridge(const char *brname)
{
	int ret;

#ifdef SIOCBRDELBR
	ret = ioctl(br_socket_fd, SIOCBRDELBR, brname);
	if (ret < 0)
#endif
	{
		char _br[IFNAMSIZ];
		unsigned long arg[3]
			= { BRCTL_DEL_BRIDGE, (unsigned long) _br };

		strncpy(_br, brname, IFNAMSIZ);
		ret = ioctl(br_socket_fd, SIOCSIFBR, arg);
	}
	return  ret < 0 ? errno : 0;
}

static int br_add_interface(const char *bridge, const char *dev)
{
	struct ifreq ifr;
	int err;
	int ifindex = if_nametoindex(dev);

	if (ifindex == 0)
		return ENODEV;

	safe_ifname_cp(&ifr, bridge);
#ifdef SIOCBRADDIF
	ifr.ifr_ifindex = ifindex;
	err = ioctl(br_socket_fd, SIOCBRADDIF, &ifr);
	if (err < 0)
#endif
	{
		unsigned long args[4] = { BRCTL_ADD_IF, ifindex, 0, 0 };

		ifr.ifr_data = (char *) args;
		err = ioctl(br_socket_fd, SIOCDEVPRIVATE, &ifr);
	}

	return err < 0 ? errno : 0;
}

static int br_del_interface(const char *bridge, const char *dev)
{
	struct ifreq ifr;
	int err;
	int ifindex = if_nametoindex(dev);

	if (ifindex == 0)
		return ENODEV;

	safe_ifname_cp(&ifr, bridge);
#ifdef SIOCBRDELIF
	ifr.ifr_ifindex = ifindex;
	err = ioctl(br_socket_fd, SIOCBRDELIF, &ifr);
	if (err < 0)
#endif
	{
		unsigned long args[4] = { BRCTL_DEL_IF, ifindex, 0, 0 };

		ifr.ifr_data = (char *) args;
		err = ioctl(br_socket_fd, SIOCDEVPRIVATE, &ifr);
	}

	return err < 0 ? errno : 0;
}

/* If /sys/class/net/XXX/bridge exists then it must be a bridge */
static int isbridge(const struct dirent *entry)
{
	char path[SYSFS_PATH_MAX];

	snprintf(path, SYSFS_PATH_MAX, SYSFS_CLASS_NET "%s/bridge", entry->d_name);
	return is_directory(path, 1, NULL) > 0;
}

/* New interface uses sysfs to find bridges */
static int new_foreach_bridge(int (*iterator)(const char *name))
{
	struct dirent **namelist;
	int i, count = 0;

	count = scandir(SYSFS_CLASS_NET, &namelist, isbridge, alphasort);
	if (count < 0)
		return -1;

	for (i = 0; i < count; i++) {
		if (iterator(namelist[i]->d_name))
			break;
	}

	for (i = 0; i < count; i++)
		free(namelist[i]);
	free(namelist);

	return count;
}

/* Old interface uses ioctl */
static int old_foreach_bridge(int (*iterator)(const char *))
{
	int i, ret=0, num;
	char ifname[IFNAMSIZ];
	int ifindices[MAX_BRIDGES];
	unsigned long args[3] = { BRCTL_GET_BRIDGES,
				 (unsigned long)ifindices, MAX_BRIDGES };

	num = ioctl(br_socket_fd, SIOCGIFBR, args);
	if (num < 0) {
		dprintf("Get bridge indices failed: %m\n");
		return -errno;
	}

	for (i = 0; i < num; i++) {
		if (!if_indextoname(ifindices[i], ifname)) {
			dprintf("get find name for ifindex %d\n", ifindices[i]);
			return -errno;
		}

		++ret;
		if(iterator(ifname))
			break;

	}

	return ret;

}

/*
 * Go over all bridges and call iterator function.
 * if iterator returns non-zero then stop.
 */
static int br_foreach_bridge(int (*iterator)(const char *))
{
	int ret;

	ret = new_foreach_bridge(iterator);
	if (ret <= 0)
		ret = old_foreach_bridge(iterator);

	return ret;
}

/* Only used if sysfs is not available. */
static int old_foreach_port(const char *brname,
			    int (*iterator)(const char *br, const char *port))
{
	int i, err, count;
	struct ifreq ifr;
	char ifname[IFNAMSIZ];
	int ifindices[MAX_PORTS];
	unsigned long args[4] = { BRCTL_GET_PORT_LIST,
				  (unsigned long)ifindices, MAX_PORTS, 0 };

	memset(ifindices, 0, sizeof(ifindices));
	safe_ifname_cp(&ifr, brname);
	ifr.ifr_data = (char *) &args;

	err = ioctl(br_socket_fd, SIOCDEVPRIVATE, &ifr);
	if (err < 0) {
		dprintf("list ports for bridge:'%s' failed: %m\n", brname);
		return -errno;
	}

	count = 0;
	for (i = 0; i < MAX_PORTS; i++) {
		if (!ifindices[i])
			continue;

		if (!if_indextoname(ifindices[i], ifname)) {
			dprintf("can't find name for ifindex:%d\n", ifindices[i]);
			continue;
		}

		++count;
		if (iterator(brname, ifname))
			break;
	}

	return count;
}

/* Iterate over all ports in bridge (using sysfs). */
static int br_foreach_port(const char *brname,
		    int (*iterator)(const char *br, const char *port))
{
	int i, count;
	struct dirent **namelist;
	char path[SYSFS_PATH_MAX];

	snprintf(path, SYSFS_PATH_MAX, SYSFS_CLASS_NET "%s/brif", brname);
	count = scandir(path, &namelist, 0, alphasort);
	if (count < 0)
		return old_foreach_port(brname, iterator);

	for (i = 0; i < count; i++) {
		if (namelist[i]->d_name[0] == '.'
		    && (namelist[i]->d_name[1] == '\0'
			|| (namelist[i]->d_name[1] == '.'
			    && namelist[i]->d_name[2] == '\0')))
			continue;

		if (iterator(brname, namelist[i]->d_name))
			break;
	}
	for (i = 0; i < count; i++)
		free(namelist[i]);
	free(namelist);

	return count;
}


static const char * const state_names[5] = {
	[BR_STATE_DISABLED] = "disabled",
	[BR_STATE_LISTENING] = "listening",
	[BR_STATE_LEARNING] = "learning",
	[BR_STATE_FORWARDING] = "forwarding",
	[BR_STATE_BLOCKING] = "blocking",
};

static inline const char *br_get_state_name(int state)
{
	if (state >= 0 && state <= 4)
		return state_names[state];

	return "<INVALID STATE>";
}

struct command
{
	int             nargs;
	const char      *name;
	int             (*func)(char *const* argv);
	const char      *help;
};


static int strtotimeval(struct timeval *tv, const char *time)
{
	double secs;
	if (sscanf(time, "%lf", &secs) != 1)
		return -1;
	tv->tv_sec = secs;
	tv->tv_usec = 1000000 * (secs - tv->tv_sec);
	return 0;
}

static int br_cmd_addbr(char * const *argv)
{
	int err;

	switch (err = br_add_bridge(*argv)) {
	case 0:
		return 0;
	case EEXIST:
		bb_error_msg("device %s already exists; can't create "
			"bridge with the same name", *argv);
		return 1;
	default:
		bb_perror_msg("add bridge failed");
		return 1;
	}
}

static int br_cmd_delbr(char * const *argv)
{
	int err;

	switch (err = br_del_bridge(*argv)){
	case 0:
		return 0;
	case ENXIO:
		bb_error_msg("bridge %s doesn't exist; can't delete it",
			*argv);
		return 1;
	case EBUSY:
		bb_error_msg("bridge %s is still up; can't delete it",
			*argv);
		return 1;
	default:
		bb_perror_msg("can't delete bridge %s", *argv);
		return 1;
	}
}

static int br_cmd_addif(char * const *argv)
{
	const char *brname;
	int err;

	brname = *argv++;

	while (*argv != NULL) {
		const char *ifname = *argv++;
		err = br_add_interface(brname, ifname);

		switch(err) {
		case 0:
			continue;
		case ENODEV:
			if (if_nametoindex(ifname) == 0)
				bb_error_msg("interface %s does not exist!", ifname);
			else
				bb_error_msg("bridge %s does not exist!", brname);
			break;
		case EBUSY:
			bb_error_msg("device %s is already a member of a bridge; "
				"can't enslave it to bridge %s.", ifname,
				brname);
			break;
		case ELOOP:
			bb_error_msg("device %s is a bridge device itself; "
				"can't enslave a bridge device to a bridge device.",
				ifname);
			break;
		default:
			bb_perror_msg("can't add %s to bridge %s", ifname, brname);
		}
		return 1;
	}
	return 0;
}

static int br_cmd_delif(char * const *argv)
{
	const char *brname;
	int err;

	brname = *argv++;

	while (*argv != NULL) {
		const char *ifname = *argv++;
		err = br_del_interface(brname, ifname);
		switch (err) {
		case 0:
			continue;
		case ENODEV:
			if (if_nametoindex(ifname) == 0)
				bb_error_msg("interface %s does not exist!", ifname);
			else
				bb_error_msg("bridge %s does not exist!", brname);
			break;
		case EINVAL:
			bb_error_msg("device %s is not a slave of %s",
				ifname, brname);
			break;
		default:
			bb_perror_msg("can't delete %s from %s", ifname, brname);
		}
		return 1;
	}
	return 0;
}

static int br_cmd_setageing(char * const *argv)
{
	int err;
	struct timeval tv;

	if (strtotimeval(&tv, argv[1])) {
		bb_error_msg("bad ageing time value");
		return 1;
	}

	err = br_set_ageing_time(*argv, &tv);
	if (err)
		bb_perror_msg("set ageing time failed");

	return err != 0;
}

static int br_cmd_setbridgeprio(char * const *argv)
{
	int prio;
	int err;

	if (sscanf(argv[1], "%i", &prio) != 1) {
		bb_error_msg("bad priority");
		return 1;
	}

	err = br_set_bridge_priority(*argv, prio);
	if (err)
		bb_perror_msg("set bridge priority failed");
	return err != 0;
}

static int br_cmd_setfd(char * const *argv)
{
	struct timeval tv;
	int err;

	if (strtotimeval(&tv, argv[1])) {
		bb_error_msg("bad forward delay value");
		return 1;
	}

	err = br_set_bridge_forward_delay(*argv, &tv);
	if (err)
		bb_perror_msg("set forward delay failed");

	return err != 0;
}

static int br_cmd_sethello(char * const *argv)
{
	struct timeval tv;
	int err;

	if (strtotimeval(&tv, argv[1])) {
		bb_error_msg("bad hello timer value");
		return 1;
	}

	err = br_set_bridge_hello_time(*argv, &tv);
	if (err)
		bb_perror_msg("set hello timer failed");

	return err != 0;
}

static int br_cmd_setmaxage(char * const *argv)
{
	struct timeval tv;
	int err;

	if (strtotimeval(&tv, argv[1])) {
		bb_error_msg("bad max age value");
		return 1;
	}
	err = br_set_bridge_max_age(*argv, &tv);
	if (err)
		bb_perror_msg("set max age failed");

	return err != 0;
}

static int br_cmd_setpathcost(char * const *argv)
{
	int cost, err;

	if (sscanf(argv[3], "%i", &cost) != 1) {
		bb_error_msg("bad path cost value");
		return 1;
	}

	err = br_set_path_cost(*argv, argv[1], cost);
	if (err)
		bb_perror_msg("set path cost failed");
	return err != 0;
}

static int br_cmd_setportprio(char * const *argv)
{
	int cost, err;

	if (sscanf(argv[3], "%i", &cost) != 1) {
		bb_error_msg("bad path priority value");
		return 1;
	}

	err = br_set_port_priority(*argv, argv[1], cost);
	if (err)
		bb_perror_msg("set port priority failed");

	return err != 0;
}

static int br_cmd_stp(char * const *argv)
{
	int stp, err;

	if (!strcmp(argv[1], "on") || !strcmp(argv[1], "yes")
	    || !strcmp(argv[1], "1"))
		stp = 1;
	else if (!strcmp(argv[1], "off") || !strcmp(argv[1], "no")
		 || !strcmp(argv[1], "0"))
		stp = 0;
	else {
		bb_error_msg("expect on/off for argument");
		return 1;
	}

	err = br_set_stp_state(*argv, stp);
	if (err)
		bb_perror_msg("set stp status failed");
	return err != 0;
}

static void br_dump_bridge_id(const unsigned char *x)
{
	printf("%.2x%.2x.%.2x%.2x%.2x%.2x%.2x%.2x", x[0], x[1], x[2], x[3],
	       x[4], x[5], x[6], x[7]);
}

static void br_show_timer(const struct timeval *tv)
{
	printf("%4i.%.2i", (int)tv->tv_sec, (int)tv->tv_usec/10000);
}

static int dump_port_info(const char *br, const char *p)
{
	struct port_info pinfo;

	if (br_get_port_info(br, p, &pinfo)) {
		printf("Can't get info for %p",p);
		return 1;
	}

	printf("%s (%d)\n", p, pinfo.port_no);
	printf(" port id\t\t%.4x",  pinfo.port_id);
	printf("\t\t\tstate\t\t%15s\n", br_get_state_name(pinfo.state));
	printf(" designated root\t");
	br_dump_bridge_id((unsigned char *)&pinfo.designated_root);
	printf("\tpath cost\t\t%4i\n", pinfo.path_cost);

	printf(" designated bridge\t");
	br_dump_bridge_id((unsigned char *)&pinfo.designated_bridge);
	printf("\tmessage age timer\t");
	br_show_timer(&pinfo.message_age_timer_value);
	printf("\n designated port\t%.4x", pinfo.designated_port);
	printf("\t\t\tforward delay timer\t");
	br_show_timer(&pinfo.forward_delay_timer_value);
	printf("\n designated cost\t%4i", pinfo.designated_cost);
	printf("\t\t\thold timer\t\t");
	br_show_timer(&pinfo.hold_timer_value);
	printf("\n flags\t\t\t");
	if (pinfo.config_pending)
		printf("CONFIG_PENDING ");
	if (pinfo.top_change_ack)
		printf("TOPOLOGY_CHANGE_ACK ");
	printf("\n\n");
	return 0;
}


static void br_dump_info(const char *br, const struct bridge_info *bri)
{
	int err;

	printf("%s\n", br);
	printf(" bridge id\t\t");
	br_dump_bridge_id((unsigned char *)&bri->bridge_id);
	printf("\n designated root\t");
	br_dump_bridge_id((unsigned char *)&bri->designated_root);
	printf("\n root port\t\t%4i\t\t\t", bri->root_port);
	printf("path cost\t\t%4i\n", bri->root_path_cost);
	printf(" max age\t\t");
	br_show_timer(&bri->max_age);
	printf("\t\t\tbridge max age\t\t");
	br_show_timer(&bri->bridge_max_age);
	printf("\n hello time\t\t");
	br_show_timer(&bri->hello_time);
	printf("\t\t\tbridge hello time\t");
	br_show_timer(&bri->bridge_hello_time);
	printf("\n forward delay\t\t");
	br_show_timer(&bri->forward_delay);
	printf("\t\t\tbridge forward delay\t");
	br_show_timer(&bri->bridge_forward_delay);
	printf("\n ageing time\t\t");
	br_show_timer(&bri->ageing_time);
	printf("\n hello timer\t\t");
	br_show_timer(&bri->hello_timer_value);
	printf("\t\t\ttcn timer\t\t");
	br_show_timer(&bri->tcn_timer_value);
	printf("\n topology change timer\t");
	br_show_timer(&bri->topology_change_timer_value);
	printf("\t\t\tgc timer\t\t");
	br_show_timer(&bri->gc_timer_value);
	printf("\n flags\t\t\t");
	if (bri->topology_change)
		printf("TOPOLOGY_CHANGE ");
	if (bri->topology_change_detected)
		printf("TOPOLOGY_CHANGE_DETECTED ");
	printf("\n\n\n");

	err = br_foreach_port(br, dump_port_info);
	if (err < 0)
		printf("can't get ports: %m\n");
}


static int br_cmd_showstp(char * const *argv)
{
	struct bridge_info info;

	if (br_get_bridge_info(*argv, &info)) {
		bb_perror_msg("%s: can't get info", *argv);
		return 1;
	}

	br_dump_info(*argv, &info);
	return 0;
}


static int dump_interface(const char *b, const char *p)
{
	(void)&b;

	if (first)
		first = 0;
	else
		printf("\n\t\t\t\t\t\t\t");

	printf("%s", p);

	return 0;
}

static void br_dump_interface_list(const char *br)
{
	int err;

	first = 1;
	err = br_foreach_port(br, dump_interface);
	if (err < 0)
		printf(" can't get port info: %m\n");
	else
		printf("\n");
}


static int show_bridge(const char *name)
{
	struct bridge_info info;

	printf("%s\t\t", name);
	fflush(stdout);

	if (br_get_bridge_info(name, &info)) {
		bb_perror_msg("can't get info");
		return 1;
	}

	br_dump_bridge_id((unsigned char *)&info.bridge_id);
	printf("\t%s\t\t", info.stp_enabled?"yes":"no");

	br_dump_interface_list(name);
	return 0;
}

static int br_cmd_show(char * const *argv)
{
	(void)&argv;

	printf("bridge name\tbridge id\t\tSTP enabled\tinterfaces\n");
	br_foreach_bridge(show_bridge);
	return 0;
}

static int compare_fdbs(const void *_f0, const void *_f1)
{
	const struct fdb_entry *f0 = _f0;
	const struct fdb_entry *f1 = _f1;

	return memcmp(f0->mac_addr, f1->mac_addr, 6);
}


static int br_cmd_showmacs(char * const *argv)
{
	const char *brname = *argv;
#define CHUNK 128
	int i, n;
	struct fdb_entry *fdb = NULL;
	int offset = 0;

	for(;;) {
		fdb = xrealloc(fdb, (offset + CHUNK) * sizeof(struct fdb_entry));

		n = br_read_fdb(brname, fdb+offset, offset, CHUNK);
		if (n == 0)
			break;

		if (n < 0)
			bb_perror_msg_and_die("read of forward table failed");

		offset += n;
	}

	qsort(fdb, offset, sizeof(struct fdb_entry), compare_fdbs);

	printf("port no\tmac addr\t\tis local?\tageing timer\n");
	for (i = 0; i < offset; i++) {
		const struct fdb_entry *f = fdb + i;
		printf("%3i\t", f->port_no);
		printf("%.2x:%.2x:%.2x:%.2x:%.2x:%.2x\t",
		       f->mac_addr[0], f->mac_addr[1], f->mac_addr[2],
		       f->mac_addr[3], f->mac_addr[4], f->mac_addr[5]);
		printf("%s\t\t", f->is_local?"yes":"no");
		br_show_timer(&f->ageing_timer_value);
		printf("\n");
	}
	return 0;
}

static const struct command commands[] = {
	{ 1, "addbr", br_cmd_addbr, "<bridge>\t\tadd bridge" },
	{ 1, "delbr", br_cmd_delbr, "<bridge>\t\tdelete bridge" },
	{ 2, "addif", br_cmd_addif,
	  "<bridge> <device>\tadd interface to bridge" },
	{ 2, "delif", br_cmd_delif,
	  "<bridge> <device>\tdelete interface from bridge" },
	{ 2, "setageing", br_cmd_setageing,
	  "<bridge> <time>\t\tset ageing time" },
	{ 2, "setbridgeprio", br_cmd_setbridgeprio,
	  "<bridge> <prio>\t\tset bridge priority" },
	{ 2, "setfd", br_cmd_setfd,
	  "<bridge> <time>\t\tset bridge forward delay" },
	{ 2, "sethello", br_cmd_sethello,
	  "<bridge> <time>\t\tset hello time" },
	{ 2, "setmaxage", br_cmd_setmaxage,
	  "<bridge> <time>\t\tset max message age" },
	{ 3, "setpathcost", br_cmd_setpathcost,
	  "<bridge> <port> <cost>\tset path cost" },
	{ 3, "setportprio", br_cmd_setportprio,
	  "<bridge> <port> <prio>\tset port priority" },
	{ 0, "show", br_cmd_show, "\t\t\tshow a list of bridges" },
	{ 1, "showmacs", br_cmd_showmacs,
	  "<bridge>\t\tshow a list of mac addrs"},
	{ 1, "showstp", br_cmd_showstp,
	  "<bridge>\t\tshow bridge stp info"},
	{ 2, "stp", br_cmd_stp,
	  "<bridge> {on|off}\tturn stp on/off" },
};

static const struct command *command_lookup(const char *cmd)
{
	size_t i;

	for (i = 0; i < sizeof(commands)/sizeof(commands[0]); i++) {
		if (!strcmp(cmd, commands[i].name))
			return &commands[i];
	}

	return NULL;
}


static void help(void)
{
	size_t i;

	printf("Usage: brctl commands\ncommands:\n");
	for (i = 0; i < sizeof(commands)/sizeof(commands[0]); i++)
		printf("\t%-10s\t%s\n", commands[i].name, commands[i].help);
}

int brctl_main(void)
{
	const struct command *cmd;
	int f;
#if ENABLE_FEATURE_LONG_OPTIONS
	static const struct option longopts[] = {
		{ .name = "help", .val = 'h' },
		{ .name = "version", .val = 'V' },
		{ .name = 0 }
	};
	bb_applet_long_options = longopts;
#endif
	f = bb_xgetopt_ulflags("hV");
	if((f & 1)) {
		help();
		return 0;
	}
	if((f & 2)) {
		printf("%s, %s\n", PACKAGE_NAME, PACKAGE_VERSION);
		return 0;
	}

	if(*Gargv == NULL) {
		help();
		return 1;
	}

	if ((br_socket_fd = socket(AF_LOCAL, SOCK_STREAM, 0)) < 0)
		bb_perror_msg_and_die("can't setup bridge control");

	if ((cmd = command_lookup(*Gargv)) == NULL) {
		bb_error_msg("never heard of command [%s]", *Gargv);
		help();
		return 1;
	}

	if (Gargc < cmd->nargs + 1) {
		printf("Incorrect number of arguments for command\n");
		printf("Usage: brctl %s %s\n", cmd->name, cmd->help);
		return 1;
	}

	return cmd->func(++Gargv);
}
#endif /* APPLET_USAGE */
