/*
 * A fake identd server
 *
 * Adapted to busybox by Thomas Lundquist <thomasez@zelow.no>
 * Original Author: Tomi Ollila <too@iki.fi>
 *                  http://www.guru-group.fi/~too/sw/
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */
#ifdef APPLET_USAGE
#define fakeidentd_trivial_usage \
	"[-b ip] [STRING]"
#define fakeidentd_full_usage \
	"Returns a set string to auth requests\n\n" \
	"\t-b\tBind to ip address\n" \
	"\tSTRING\tThe ident answer string (default is nobody)"
#else

#include "busybox.h"
#include "bb_sig.h"
#include "bb_inet.h"
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <sys/time.h>
#include <time.h>
#include <netinet/in.h>
#include <sys/uio.h>
#include <fcntl.h>


#define IDENT_PORT  113
#define MAXCONNS    20
#define MAXIDLETIME 45

static const char ident_substr[] = " : USERID : UNIX : ";
static const int ident_substr_len = sizeof(ident_substr) - 1;

/*
 * We have to track the 'first connection socket' so that we
 * don't go around closing file descriptors for non-clients.
 *
 * descriptor setup normally
 *  0 = server socket
 *  1 = syslog fd (hopefully -- otherwise this won't work)
 *  2 = connection socket after detached from tty. standard error before that
 *  3 - 2 + MAXCONNS = rest connection sockets
 *
 * To try to make sure that syslog fd is what is "requested", the that fd
 * is closed before openlog() call.  It can only severely fail if fd 0
 * is initially closed.
 */
#define FCS 2

typedef struct CONN {
	char buf[20];
	int len;
	time_t lasttime;
} conn_t;

CALC_BSS struct BB_COMMON_BSS {
/* FD of the connection is always the index of the connection structure
 * in `conns' array + FCS */
	conn_t conns[MAXCONNS];
	const char *identuser;
	fd_set readfds;
	int conncnt;
	char *pidfile;
} bb_common_bss;
#define conns bb_common_bss.conns
#define identuser bb_common_bss.identuser
#define readfds bb_common_bss.readfds
#define conncnt bb_common_bss.conncnt
#define pidfile bb_common_bss.pidfile

static const char nobodystr[] = "nobody"; /* this needs to be declared like this */

static inline void movefd(int from, int to)
{
	if (from != to) {
		dup2(from, to);
		close(from);
	}
}

static void inetbind(const char *bind_ip_address)
{
	int s;
	struct sockaddr_in addr;

	s = bb_xsocket(BB_XSOCKET_FLG_ERR_DIE);

	setsockopt_1i(s, SOL_SOCKET, SO_REUSEADDR, 1);

	bb_gethostbyname(BB_GHBN_SOCKADDR_IN | BB_GHBN_ERR_DIE, bind_ip_address, &addr);
	addr.sin_port = bb_lookup_tcp_port("identd", IDENT_PORT);

	bb_xbind(s, (struct sockaddr *)&addr, sizeof(addr));

	if (listen(s, 5) < 0)
		bb_perror_msg_and_die("Cannot listen() on identd port");

	movefd(s, 0);
}

static void handlexitsigs(int signum)
{
	if(pidfile)
		unlink(pidfile);
	exit(0);
}

static void deleteConn(int s)
{
	int i = s - FCS;

	close(s);

	conncnt--;

	/*
	 * Most of the time there is 0 connections. Most often that there
	 * is connections, there is just one connection. When this one connection
	 * closes, i == conncnt = 0 -> no copying.
	 * When there is more than one connection, the oldest connections closes
	 * earlier on average. When this happens, the code below starts copying
	 * the connection structure w/ highest index to the place which which is
	 * just deleted. This means that the connection structures are no longer
	 * in chronological order. I'd quess this means that when there is more
	 * than 1 connection, on average every other connection structure needs
	 * to be copied over the time all these connections are deleted.
	 */
	if (i != conncnt) {
		memcpy(&conns[i], &conns[conncnt], sizeof(conns[0]));
		movefd(conncnt + FCS, s);
	}

	FD_CLR(conncnt + FCS, &readfds);
}

static void replyError(int s, char *buf)
{
	struct iovec iv[3];

	iv[0].iov_base = "0, 0 : ERROR : ";   iv[0].iov_len = 15;
	iv[1].iov_base = buf;                 iv[1].iov_len = strlen(buf);
	iv[2].iov_base = "\r\n";              iv[2].iov_len = 2;
	writev(s, iv, 3);
}

static int closeOldest(void)
{
	time_t min = conns[0].lasttime;
	int idx = 0;
	int i;

	for (i = 1; i < MAXCONNS; i++)
		if (conns[i].lasttime < min)
			idx = i;

	replyError(idx + FCS, "X-SERVER-TOO-BUSY");
	close(idx + FCS);

	return idx;
}

static int checkInput(char *buf, int len, int l)
{
	int i;
	for (i = len; i < len + l; ++i)
		if (buf[i] == '\n')
			return 1;
	return 0;
}

static int chmatch(char c, const char *chars)
{
	for (; *chars; chars++)
		if (c == *chars)
			return 1;
	return 0;
}

static int skipchars(char **p, const char *chars)
{
	while (chmatch(**p, chars))
		(*p)++;
	if (**p == '\r' || **p == '\n')
		return 0;
	return 1;
}

static int parseAddrs(char *ptr, char **myaddr, char **heraddr)
{
	/* parse <port-on-server> , <port-on-client> */

	if (!skipchars(&ptr, bb_spacestok))
		return -1;

	*myaddr = ptr;

	if (!skipchars(&ptr, "1234567890"))
		return -1;

	if (!chmatch(*ptr, " \t,"))
		return -1;

	*ptr++ = '\0';

	if (!skipchars(&ptr, " \t,") )
		return -1;

	*heraddr = ptr;

	skipchars(&ptr, "1234567890");

	if (!chmatch(*ptr, " \n\r"))
		return -1;

	*ptr = '\0';

	return 0;
}

static void reply(int s, char *buf)
{
	char *myaddr, *heraddr;

	myaddr = heraddr = NULL;

	if (parseAddrs(buf, &myaddr, &heraddr))
		replyError(s, "X-INVALID-REQUEST");
	else {
		struct iovec iv[6];
		iv[0].iov_base = myaddr;               iv[0].iov_len = strlen(myaddr);
		iv[1].iov_base = ", ";                 iv[1].iov_len = 2;
		iv[2].iov_base = heraddr;              iv[2].iov_len = strlen(heraddr);
		iv[3].iov_base = (void *)ident_substr; iv[3].iov_len = ident_substr_len;
		iv[4].iov_base = (void *)identuser;    iv[4].iov_len = strlen(identuser);
		iv[5].iov_base = "\r\n";               iv[5].iov_len = 2;
		writev(s, iv, 6);
	}
}

int fakeidentd_main(void)
{
	struct passwd *pw;

	/* handle -b <ip> parameter */
	char *bind_ip_address = "0.0.0.0";
#define OPT_FOREGROUND 2
	bb_getopt_ulflags("b:", &bind_ip_address);
	/* handle optional REPLY STRING */
	if (*Gargv)
		identuser = *Gargv;
	else
		identuser = nobodystr;

	pw = getpwnam(nobodystr);
	if (pw == NULL)
		bb_error_msg_and_die("Cannot find uid/gid of user '%s'", nobodystr);

	bb_signal((1 << SIGTERM) | (1 << SIGINT) | (1 << SIGQUIT), handlexitsigs);

	inetbind(bind_ip_address);
	setgid(pw->pw_gid);
	setuid(pw->pw_uid);

	/* SIGPIPE - connection closed when writing (raises ???) */
	bb_signal((1 << SIGHUP) | (1 << SIGPIPE), SIG_IGN);

	bb_openlog(0, LOG_DAEMON);

	pidfile = xdaemon(XDAEMON_PIDFILE);

	FD_ZERO(&readfds);
	FD_SET(0, &readfds);
	/* main loop where we process all events and never exit */
	while (1) {
	   fd_set rfds = readfds;
	   struct timeval tv = { 15, 0 };
	   int i;
	   int tim = time(NULL);

	   select(conncnt + FCS, &rfds, NULL, NULL, conncnt? &tv: NULL);

	   for (i = conncnt - 1; i >= 0; i--) {
		   int s = i + FCS;

		   if (FD_ISSET(s, &rfds)) {
			   char *buf = conns[i].buf;
			   unsigned int len = conns[i].len;
			   unsigned int l;

			   if ((l = read(s, buf + len, sizeof(conns[0].buf) - len)) > 0) {
				   if (checkInput(buf, len, l)) {
					   reply(s, buf);
					   goto deleteconn;
				   } else if (len + l >= sizeof(conns[0].buf)) {
					   replyError(s, "X-INVALID-REQUEST");
					   goto deleteconn;
				   } else {
					   conns[i].len += l;
				   }
			   } else {
				   goto deleteconn;
			   }

			   conns[i].lasttime = tim;
			   continue;

   deleteconn:
			   deleteConn(s);
		   } else {
			   /* implement as time_after() in linux kernel sources ... */
			   if (conns[i].lasttime + MAXIDLETIME <= tim) {
				   replyError(s, "X-TIMEOUT");
				   deleteConn(s);
			   }
		   }
	   }

	   if (FD_ISSET(0, &rfds)) {
		   int s = accept(0, NULL, 0);

		   if (s < 0) {
			   if (errno != EINTR)
				   bb_perror_msg("accept");
		   } else {
			   if (conncnt == MAXCONNS)
				   i = closeOldest();
			   else
				   i = conncnt++;

			   movefd(s, i + FCS); /* move if not already there */
			   FD_SET(i + FCS, &readfds);

			   conns[i].len = 0;
			   conns[i].lasttime = time(NULL);
		   }
	   }
	} /* end of while(1) */

	return 0;
}
#endif /* APPLET_USAGE */
