/*
 * ftp
 *
 * Mini implementation of FTP commands (retrieve, store, delete, append,
 * list).
 *
 * Copyright (C) 2002 Jeff Angielski, The PTR Group <jeff@theptrgroup.com>
 * Copyright (C) 2002 Glenn McGrath <bug1@iinet.net.au>
 * Copyright (C) 2005 Rainer Weikusat <rainer.weikusat@sncag.com>
 *
 * Based on wget.c by Chip Rosenthal Covad Communications
 * <chip@laserlink.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifdef APPLET_USAGE
#define ftpgetput_trivial_usage \
	"[-c] [-v] [-u user] [-p password] [-P port] remote-host -s|r|d|l|n file1 file2"
#define ftpgetput_full_usage \
	"Minimalistic ftp client.\n\n" \
	"Options:\n" \
	USE_FEATURE_LONG_OPTIONS(\
	"\t-c, --continue\tContinue a previous retrieve\n" \
	"\t-v, --verbose\tVerbose\n" \
	"\t-p, --password\tPassword to be used\n" \
	"\t-P, --port\tPort number to be used\n" \
	"\t-s, --send\tPut local-file1 -> remote-file2\n" \
	"\t-r, --recieve\tGet remote-file1 -> local-file2\n" \
	"\t-d, --delete\tDelete remote-dir1 remote-file2\n" \
	"\t-l, --list\tList files in remote-path1 (verbose) to out-file2\n" \
	"\t-n, --nlist\tList files in remote-path1 (sparse) to out-file2") \
	SKIP_FEATURE_LONG_OPTIONS(\
	"\t-c\tContinue a previous retrieve\n" \
	"\t-v\tVerbose\n" \
	"\t-p\tPassword to be used\n" \
	"\t-P\tPort number to be used\n" \
	"\t-s\tPut local-file1 -> remote-file2\n" \
	"\t-r\tGet remote-file1 -> local-file2\n" \
	"\t-d\tDelete remote-dir1 remote-file2\n" \
	"\t-l\tList files in remote-path1 (verbose) to out-file2\n" \
	"\t-n\tList files in remote-path1 (sparse) to out-file2")
#else

#include "busybox.h"
#include <sys/ioctl.h>
#include <sys/time.h>

#include <ctype.h>
#include <string.h>
#include <unistd.h>

#include <netinet/in.h>

#include "bb_xopen.h"
#include "bb_inet.h"

enum {
	PROTOBUF_SIZE = 512
};

typedef struct ftp_host_info_s {
	char *user;
	char *password;
	struct sockaddr s_addr;
} ftp_host_info_t;

#define PORT_FIELD(pserver) ((struct sockaddr_in *)&((pserver)->s_addr))->sin_port

typedef int ftp_action_routine(ftp_host_info_t *, char *, char *);

CALC_BSS struct BB_COMMON_BSS {
	FILE *control_stream;
	int verbose_flag;
	int do_continue;
	int want_allo;
} bb_common_bss;

static int read_ftp_reply(char *buf)
{
	char *p;

	do {
		p = fgets(buf, PROTOBUF_SIZE, bb_common_bss.control_stream);
		if (!p)
			bb_perror_msg_and_die("fgets");
		if (bb_common_bss.verbose_flag)
			fputs(buf, stderr);
		p[strcspn(p, "\r\n")] = 0;      /* should work with \r\n and bare \n */
	}
	while (!isdigit(*buf) || buf[3] != ' ');

	return atoi(buf);
}

static int send_ftp_cmd(char *buf, char const *fmt, ...)
{
	va_list val;
	int rc;

	va_start(val, fmt);
	rc = vsnprintf(buf, PROTOBUF_SIZE - 3, fmt, val);
	va_end(val);
	buf[rc++] = '\r';
	buf[rc++] = '\n';
	buf[rc] = 0;

	fputs(buf, bb_common_bss.control_stream);

	if (bb_common_bss.verbose_flag)
		fputs(buf, stderr);
	return read_ftp_reply(buf);
}

static void ftp_login(const ftp_host_info_t *server)
{
	char buf[PROTOBUF_SIZE];

	/* Connect to the command socket */
	bb_common_bss.control_stream = fdopen(xconnect(&server->s_addr), "r+");
	if (!bb_common_bss.control_stream)
		bb_perror_msg_and_die("Couldnt open control stream");

	if (read_ftp_reply(buf) != 220)
		bb_error_msg_and_die("%s", buf + 4);

	/*  Login to the server */
	switch (send_ftp_cmd(buf, "USER %s", server->user)) {
	case 230:
		break;
	case 331:
		if (send_ftp_cmd(buf, "PASS %s", server->password) != 230)
			bb_error_msg_and_die("PASS error: %s", buf + 4);
		break;
	default:
		bb_error_msg_and_die("USER error: %s", buf + 4);
	}

	send_ftp_cmd(buf, "TYPE I");
}

static int get_ftp_dataconn(ftp_host_info_t * server, char *buf)
{
	char *buf_ptr;
	int rc;
	unsigned port_num;

	rc = send_ftp_cmd(buf, "PASV");
	if (rc != 227)
		bb_error_msg_and_die("PASV error: %s", buf + 4);

	buf_ptr = strrchr(buf, ',');
	*buf_ptr = 0;
	port_num = atoi(buf_ptr + 1);

	buf_ptr = strrchr(buf, ',');
	*buf_ptr = 0;
	port_num += atoi(buf_ptr + 1) * 256;

	PORT_FIELD(server) = htons(port_num);
	return xconnect(&server->s_addr);
}

static int terminate_ftp_session(int fd, char *buf)
{
	close(fd);
	if (read_ftp_reply(buf) != 226)
		bb_error_msg_and_die("ftp error: %s", buf + 4);

	send_ftp_cmd(buf, "QUIT");
	return EXIT_SUCCESS;
}

#define is_stdio_path(path) (*path == '-' && !path[1])

static int
rest_list_nlist(ftp_host_info_t * server, char *server_path, char *local_path,
						int flg_rest, const char *cmd)
{
	int f;
	int fd_data;
	char buf[PROTOBUF_SIZE];

	if (is_stdio_path(local_path)) {
		f = STDOUT_FILENO;
	} else {
		f = bb_common_bss.do_continue != 0 ? O_APPEND : O_CREAT | O_TRUNC | BB_OPEN_CREAT_0777;
		f = bb_xopen(local_path, f | O_WRONLY | BB_OPEN_ERR);
	}

	fd_data = get_ftp_dataconn(server, buf);

	if(flg_rest) {
		if(bb_common_bss.do_continue) {
			struct stat sbuf;

			if (fstat(f, &sbuf) < 0)
				bb_perror_msg_and_die("%s", local_path);
			if(sbuf.st_size > 0) {
				if (send_ftp_cmd(buf, "REST %ld", sbuf.st_size) != 350)
					bb_error_msg_and_die("server won't restart -- try without -c");
			}
		}
	}

	if (send_ftp_cmd(buf, "%s %s", cmd, server_path) > 150)
		bb_error_msg_and_die("%s error: %s", cmd, buf + 4);

	if (bb_copyfd_eof(fd_data, f) == -1)
		return EXIT_FAILURE;

	return terminate_ftp_session(fd_data, buf);
}

static int
ftp_receive(ftp_host_info_t * server, char *server_path, char *local_path)
{
	return rest_list_nlist(server, server_path, local_path, 1, "RETR");
}

static int
ftp_send(ftp_host_info_t * server, char *local_path, char *server_path)
{
	struct stat sbuf;
	char buf[PROTOBUF_SIZE];
	int fd_data;
	int fd_local;

	fd_data = get_ftp_dataconn(server, buf);

	/* get the local file */
	fd_local = bb_xopen(local_path, O_RDONLY | BB_OPEN_MINUS_IS_STDIN | BB_OPEN_ERR);
	if (!is_stdio_path(local_path) && bb_common_bss.want_allo) {
		fstat(fd_local, &sbuf);

		switch (send_ftp_cmd(buf, "ALLO %lu", sbuf.st_size)) {
		case 200:
		case 202:
			break;

		default:
			bb_error_msg_and_die("ALLO error: %s", buf + 4);
		}
	}

	switch (send_ftp_cmd(buf, "STOR %s", server_path)) {
	case 125:
	case 150:
		break;

	default:
		bb_error_msg_and_die("STOR error: %s", buf + 4);
	}

	/* transfer the file  */
	if (bb_copyfd_eof(fd_local, fd_data) == -1)
		return EXIT_FAILURE;

	return terminate_ftp_session(fd_data, buf);
}

static int
ftp_delete(ftp_host_info_t * server, char *server_path, char *server_file)
{
	char buf[PROTOBUF_SIZE];
	int fd_data;

	fd_data = get_ftp_dataconn(server, buf);

	switch (send_ftp_cmd(buf, "CWD %s", server_path)) {
	case 250:
		break;

	default:
		bb_error_msg_and_die("CWD error: %s", buf + 4);
	}

	switch (send_ftp_cmd(buf, "DELE %s", server_file))
	{
	case 200:
	case 250:
		break;
	default:
		bb_error_msg_and_die("DELE error: %s", buf + 4);
		break;
	}

	close(fd_data);
	send_ftp_cmd(buf, "QUIT");
	return EXIT_SUCCESS;
}

static int
ftp_list(ftp_host_info_t * server, char *server_path, char *local_path)
{
	return rest_list_nlist(server, server_path, local_path, 0, "LIST");
}

static int
ftp_nlist(ftp_host_info_t * server, char *server_path, char *local_path)
{
	return rest_list_nlist(server, server_path, local_path, 0, "NLST");
}

#define FTPGETPUT_OPT_CONTINUE  1
#define FTPGETPUT_OPT_VERBOSE   2
#define FTPGETPUT_OPT_USER      4
#define FTPGETPUT_OPT_PASSWORD  8
#define FTPGETPUT_OPT_PORT      16

#define FTPGETPUT_OPT_ALLO      32

#define FTPGETPUT_OPT_STOR      64
#define FTPGETPUT_OPT_RETR      128
#define FTPGETPUT_OPT_DELE      256
#define FTPGETPUT_OPT_LIST      512
#define FTPGETPUT_OPT_NLST      1024


static ftp_action_routine *determine_ftp_action(unsigned long opt)
{

	if ((opt & FTPGETPUT_OPT_STOR) == FTPGETPUT_OPT_STOR)
		return ftp_send;
	if ((opt & FTPGETPUT_OPT_RETR) == FTPGETPUT_OPT_RETR)
		return ftp_receive;
	if ((opt & FTPGETPUT_OPT_DELE) == FTPGETPUT_OPT_DELE)
		return ftp_delete;
	if ((opt & FTPGETPUT_OPT_LIST) == FTPGETPUT_OPT_LIST)
		return ftp_list;
	if ((opt & FTPGETPUT_OPT_NLST) == FTPGETPUT_OPT_NLST)
		return ftp_nlist;

	bb_error_msg_and_die("cannot determine FTP action");
}

int ftpgetput_main(void)
{
	ftp_host_info_t server;
	ftp_action_routine *ftp_action;
	unsigned long opt;
	char *port = "ftp";
	char *a;

#if ENABLE_FEATURE_LONG_OPTIONS
	static const struct option longopts[] = {
		{"continue", 1, NULL, 'c'},
		{"verbose",  0, NULL, 'v'},
		{"username", 1, NULL, 'u'},
		{"password", 1, NULL, 'p'},
		{"port",     1, NULL, 'P'},
		{"send",     0, NULL, 's'},
		{"recieve",  0, NULL, 'r'},
		{"delete",   0, NULL, 'd'},
		{"list",     0, NULL, 'l'},
		{"nlist",    0, NULL, 'n'},
		{"allo",     0, NULL, 'A'},
		{NULL,       0, NULL,  0}
	};
	bb_applet_long_options = longopts;
#endif

	/* Set default values */
	server.user = "anonymous";
	server.password = "busybox@";

	/* Decipher the command line */
	opt = bb_xgetopt_ulflags("cvu:p:P:Asrdln\0=3:?:s--rdln:r--sdln:d--srln:l--srdn:n--srdl:s:r:d:l:n",
					&server.user, &server.password, &port);

	ftp_action = determine_ftp_action(opt);

	bb_common_bss.do_continue = opt & FTPGETPUT_OPT_CONTINUE;
	bb_common_bss.verbose_flag = opt & FTPGETPUT_OPT_VERBOSE;
	bb_common_bss.want_allo = opt & FTPGETPUT_OPT_ALLO;

	/* We want to do exactly _one_ DNS lookup, since some
	 * sites (i.e. ftp.us.debian.org) use round-robin DNS
	 * and we want to connect to only one IP... */
	a = bb_gethostbyname(BB_GHBN_SOCKADDR | BB_GHBN_ERR_DIE |
			(bb_common_bss.verbose_flag ? BB_GHBN_NTOP: 0),
			*Gargv++, &server.s_addr);
	PORT_FIELD(&server) = bb_lookup_tcp_port(port, 21);
	if (bb_common_bss.verbose_flag)
		fprintf(stderr, "Connecting to %s:%d\n", a, ntohs(PORT_FIELD(&server)));

	/*  Connect/Setup/Configure the FTP session */
	ftp_login(&server);
	return ftp_action(&server, *Gargv, Gargv[1]);
}
#endif /* APPLET_USAGE */
