/*
 * Mini hostname implementation for busybox
 *
 * Copyright (C) 1999 by Randolph Chung <tausq@debian.org>
 *
 * adjusted by Erik Andersen <andersen@codepoet.org> to remove
 * use of long options and GNU getopt.  Improved the usage info.
 *
 * Vladimir Oleynik <dzo@simtreas.ru> (C) 2008
 * size optimization
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */
#ifdef APPLET_USAGE
#define hostname_trivial_usage \
	"[OPTION] {hostname | -F FILE}"
#define hostname_full_usage \
	"Get or set the hostname or DNS domain name. If a hostname is given\n" \
	"(or FILE with the -F parameter), the host name will be set.\n\n" \
	"Options:\n" \
	"\t-s\tShort\n" \
	"\t-i\tAddresses for the hostname\n" \
	"\t-d\tDNS domain name\n" \
	"\t-f\tFully qualified domain name\n" \
	"\t-F FILE\tUse the contents of FILE to specify the hostname"
#else

#include "busybox.h"
#include <arpa/inet.h>
#include <unistd.h>
#include <string.h>
#include "bb_inet.h"
#include "config_read.h"

#ifndef HAVE_SETHOSTNAME
int sethostname(const char *name, size_t len);
#endif

CALC_BSS struct BB_COMMON_BSS {
	char buf[256];
} bb_common_bss;
#define buf bb_common_bss.buf


static void do_sethostname(const char *s)
{
	if (sethostname(s, strlen(s)) < 0) {
	    if (errno == EPERM)
		bb_error_msg_and_die("you must be root to change the host name");
	    else
		bb_perror_msg_and_die("sethostname");
	}
}

static void getaddrinfo_canon(const char *name, int type)
{
	char *out;

	if((type & (1|2|8)) != 0)
		bb_gethostbyname(BB_GHBN_ERR_DIE | BB_GHBN_CANON_NAME, name, &out);
	if ((type & 8 ) /* == 's' */) {
		strchr_set0(out, '.');
	} else if ((type & 1 ) /* == 'd' */) {
		out = strchr(out, '.');
		if(out == NULL)
			return;
		out++;
	} else {
		int n, i;
#if ENABLE_FEATURE_IPV6
		char ibuf[INET6_ADDRSTRLEN+1];
#else
		char ibuf[INET_ADDRSTRLEN+1];
#endif
		struct sockaddr *addr_buf = bb_gethostbyname(BB_GHBN_ERR_DIE | BB_GHBN_NSOCKADDR, name, &n);
		for (i = 0; i < n; i++)
			printf("%s%s", &(" "[(i == 0)]),
				inet_ntop(addr_buf[i].sa_family,
					&((struct sockaddr_in*)&addr_buf[i])->sin_addr, ibuf, sizeof(ibuf)));
		/* free(addr_buf); */
		out = "";
	}
	printf("%s\n", out);
}


int hostname_main(void)
{
	char **argv;
	int type;
	char *filename = NULL;

	type = bb_xgetopt_ulflags("dfisF:\0d-fis:f-dis:i-dfs:s-dfi",
					&filename) & (1|2|4|8);
	argv = Gargv;

	/* Output in desired format */
	if ((type & (1|2|4|8)) != 0) {
		gethostname(buf, 255);
		buf[255] = '\0';    /* safe, hostname may be 255 long */
		getaddrinfo_canon(buf, type);
	}
	/* Set the hostname */
	else if (filename != NULL) {
		config_read_t *crt = config_read_prepare(filename, 'e');

		if( (crt = config_read(crt, "#\1\0\0")) != NULL) {
			do_sethostname(crt->tokens[0]);
			/* config_close(crt); */
		}
	} else if (*argv) {
		do_sethostname(*argv);
	}
	/* Or if all else fails,
	 * just print the current hostname */
	 else {
		gethostname(buf, 255);
		buf[255] = '\0';    /* safe, hostname may be 255 long */
		puts(buf);
	}
	return 0;
}
#endif /* APPLET_USAGE */
