/*
 * httpd implementation for busybox
 *
 * Copyright (C) 2002,2003 Glenn Engel <glenne@engel.org>
 * Copyright (C) 2003-2006 Vladimir Oleynik <dzo@simtreas.ru>
 *
 * simplify patch stolen from libbb without using strdup
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 *
 *****************************************************************************
 *
 * Typical usage:
 *   for non root user
 * httpd -p 8080 -h $HOME/public_html
 *   or for daemon start from rc script with uid=0:
 * httpd -u www
 * This is equivalent if www user have uid=80 to
 * httpd -p 80 -u 80 -h /www -c /etc/httpd.conf -r "Web Server Authentication"
 *
 *
 * When a url contains "cgi-bin" it is assumed to be a cgi script.  The
 * server changes directory to the location of the script and executes it
 * after setting QUERY_STRING and other environment variables.
 *
 * The server can also be invoked as a url arg decoder and html text encoder
 * as follows:
 *  foo=`httpd -d $foo`           # decode "Hello%20World" as "Hello World"
 *  bar=`httpd -e "<Hello World>"`  # encode as "&#60Hello&#32World&#62"
 * Note that url encoding for arguments is not the same as html encoding for
 * presentation.  -d decodes a url-encoded argument while -e encodes in html
 * for page display.
 *
 * httpd.conf has the following format:
 *
 * A:172.20.         # Allow address from 172.20.0.0/16
 * D:*               # Deny from other IP connections
 * /cgi-bin:foo:bar  # Require user foo, pwd bar on urls starting with /cgi-bin/
 * /adm:admin:setup  # Require user admin, pwd setup on urls starting with /adm/
 * /adm:toor:PaSsWd  # or user toor, pwd PaSsWd on urls starting with /adm/
 * .au:audio/basic   # Additional mime type for audio.au files
 * *.php:/path/php   # running cgi.php scripts through an interpreter
 * E:error404.html   # Name of the file to use as a error 404 page.
 * I:index.shtml     # Name of the file to use as a pre-written HTML
 *                   # directory index. Default index.html
 *                   # E and I priority set by reverse!
 *
 * A/D may be as a/d or allow/deny - first char case insensitive
 * Acepted with mask too with /N (N=0-32) or /M.M.M.M a form.
 * Example a IP access logic:
 *   1. Allow only specified addresses
 *     A:172.20               # Allow any address that begins with 172.20.
 *     A:1.1.0.0/255.255.128  # Allow address 1.1.0.0 - 1.1.1.255
 *     A:127.0.0.1            # Allow local loopback connections
 *     D:*                    # Deny from other IP connections
 *
 *   2. Only deny specified addresses
 *     D:1.2.3.        # deny from 1.2.3.0 - 1.2.3.255
 *     D:2.3.4.0/25    # deny from 2.3.4.0 - 2.3.4.127
 *     A:*             # (optional line added for clarity)
 *
 * If a sub directory contains a config file it is parsed and merged with
 * any existing settings as if it was appended to the original configuration.
 *
 * subdir paths are relative to the containing subdir and thus cannot
 * affect the parent rules.
 *
 * Note that since the sub dir is parsed in the forked thread servicing the
 * subdir http request, any merge is discarded when the process exits.  As a
 * result, the subdir settings only have a lifetime of a single request.
 *
 *
 * If -c is not set, an attempt will be made to open the default
 * root configuration file.  If -c is set and the file is not found, the
 * server exits with an error.
 *
*/
#ifdef APPLET_USAGE
#define httpd_trivial_usage \
	"[-c <conf file>]" \
	SKIP_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY(" [-p <port>]") \
	SKIP_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY(" [-u user]") \
	USE_FEATURE_HTTPD_BASIC_AUTH(" [-r <realm>]") \
	USE_FEATURE_HTTPD_AUTH_MD5(" [-m pass]") \
	" [-h home]" \
	" [-d/-e <string>]"
#define httpd_full_usage \
	"Listens for incoming http server requests.\n\n" \
	"Options:\n" \
	"\t-c FILE\t\tSpecifies configuration file. (default httpd.conf)\n" \
	SKIP_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY("\t-p PORT\tServer port (default 80)\n") \
	SKIP_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY("\t-u USER\tSet uid to USER after listening privileges port\n") \
	USE_FEATURE_HTTPD_BASIC_AUTH("\t-r REALM\tAuthentication Realm for Basic Authentication\n") \
	USE_FEATURE_HTTPD_AUTH_MD5("\t-m PASS\t\tCrypt PASS with md5 algorithm\n") \
	"\t-h HOME\tSpecifies http HOME directory (default ./)\n" \
	"\t-e STRING\tHtml encode STRING\n" \
	"\t-d STRING\tURL decode STRING"
#else

/* You can use this server as standalone, require libbb.a for linking */
//#define HTTPD_STANDALONE

#ifdef HTTPD_STANDALONE
#define _GNU_SOURCE
#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64
#endif

#include "busybox.h"

#include <ctype.h>         /* for isspace           */
#include <string.h>
#include <time.h>
#include <unistd.h>        /* for close             */
#include <netinet/in.h>    /* for sockaddr_in       */
#include <sys/time.h>
#include <sys/wait.h>
#include <fcntl.h>         /* for open modes        */
#include "bb_sig.h"
#include "bb_inet.h"


static const char httpdVersion[] = "busybox httpd/2.2 22-Apr-2006";
static const char default_path_httpd_conf[] = "/etc";
static const char httpd_conf[] = "httpd.conf";
static const char home[] = "./";
static const char dot[] = ".";

#define TIMEOUT 60

// Note: busybox xfuncs are not used because we want the server to keep running
//       if something bad happens due to a malformed user request.
//       As a result, all memory allocation after daemonize
//       is checked rigorously

//#define DEBUG 1

/* Configure options, disabled by default as custom httpd feature */

/* If set, use this server from internet superserver only */
//#define CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY


#ifdef HTTPD_STANDALONE
/* standalone, enable all features */
#undef CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY
/* unset config option for remove warning as redefined */
#undef CONFIG_FEATURE_HTTPD_BASIC_AUTH
#undef CONFIG_FEATURE_HTTPD_AUTH_MD5
#undef CONFIG_FEATURE_HTTPD_ENCODE_URL_STR
#undef CONFIG_FEATURE_HTTPD_SET_REMOTE_PORT_TO_ENV
#undef CONFIG_FEATURE_HTTPD_CONFIG_WITH_MIME_TYPES
#undef CONFIG_FEATURE_HTTPD_CGI
#undef CONFIG_FEATURE_HTTPD_SETUID
#undef CONFIG_FEATURE_HTTPD_BIND
#undef CONFIG_FEATURE_HTTPD_RELOAD_CONFIG_SIGHUP
#undef CONFIG_FEATURE_HTTPD_SSI
#undef CONFIG_FEATURE_HTTPD_SSI_EXEC
#undef CONFIG_FEATURE_HTTPD_CONFIG_WITH_SCRIPT_INTERPR
#undef CONFIG_FEATURE_HTTPD_DIR_LIST
/* enable all features now */
#define CONFIG_FEATURE_HTTPD_BASIC_AUTH
#define CONFIG_FEATURE_HTTPD_AUTH_MD5
#define CONFIG_FEATURE_HTTPD_ENCODE_URL_STR
#define CONFIG_FEATURE_HTTPD_SET_REMOTE_PORT_TO_ENV
#define CONFIG_FEATURE_HTTPD_CONFIG_WITH_MIME_TYPES
#define CONFIG_FEATURE_HTTPD_CGI
#define CONFIG_FEATURE_HTTPD_SETUID
#define CONFIG_FEATURE_HTTPD_BIND
#define CONFIG_FEATURE_HTTPD_RELOAD_CONFIG_SIGHUP
#define CONFIG_FEATURE_HTTPD_SSI
#define CONFIG_FEATURE_HTTPD_SSI_EXEC
#define CONFIG_FEATURE_HTTPD_CONFIG_WITH_SCRIPT_INTERPR
#define CONFIG_FEATURE_HTTPD_DIR_LIST

#endif

#ifdef CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY
#undef CONFIG_FEATURE_HTTPD_SETUID  /* use inetd user.group config settings */
#undef CONFIG_FEATURE_HTTPD_RELOAD_CONFIG_SIGHUP    /* so is not daemon */
#undef CONFIG_FEATURE_HTTPD_BIND        /* use inetd with tcpwrapper */
/* inetd set stderr to accepted socket and we can`t true see debug messages */
#define DEBUG 0
#else
# ifndef DEBUG
#  define DEBUG 0
# endif
#endif

#define MAX_MEMORY_BUFF 8192    /* IO buffer */

typedef struct Htaccess {
	char *after_colon;
	struct Htaccess *next;
	char before_colon[1];         /* really bigger, must be last */
} Htaccess;

typedef struct Htaccess_IP {
	unsigned int ip;
	unsigned int mask;
	int allow_deny;
	struct Htaccess_IP *next;
} Htaccess_IP;

typedef struct
{
    /* /path/file/path_info?query */
    char *url;          /* /path/file */
    char *path_info;    /* path_info or NULL */
    char *query;        /* query or NULL */
    int   found_e404;   /* found error 404 page */
} URI;

CALC_BSS struct BB_COMMON_BSS
{
  char buf[MAX_MEMORY_BUFF];

#ifdef CONFIG_FEATURE_HTTPD_BASIC_AUTH
  const char *realm;
  Htaccess *auth;               /* config user:password lines */
  char *usr_pass;
#endif

#if defined(CONFIG_FEATURE_HTTPD_CGI) || defined(CONFIG_FEATURE_HTTPD_SSI)
  char *referer;
  char *cookie;
  char *user_agent;
  char *content_type;
  char *document_root;
#endif

  const char *configFile;

  unsigned int rmt_ip;
#if defined(CONFIG_FEATURE_HTTPD_CGI) || defined(CONFIG_FEATURE_HTTPD_SSI) || DEBUG
  char rmt_ip_str[16];     /* for set env REMOTE_ADDR */
#endif
  unsigned port;           /* server initial port and for
			      set env REMOTE_PORT */
  union HTTPD_FOUND {
	const char *found_mime_type;
	const char *found_moved_temporarily;
  } httpd_found;

  bbst_size_t ContentLength;    /* -1 - unknown */
  time_t last_mod;

  Htaccess_IP *ip_a_d;          /* config allow/deny lines */
  Htaccess    *htindex;         /* pre-written HTML index file */
  Htaccess    *e404;            /* pre-written HTML eror 404 page file */

#ifdef CONFIG_FEATURE_HTTPD_CONFIG_WITH_MIME_TYPES
  Htaccess *mime_a;             /* config mime types */
#endif

#ifndef CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY
  int accepted_socket;
#define a_c_r config.accepted_socket
#define a_c_w config.accepted_socket
#else
#define a_c_r 0
#define a_c_w 1
#endif
  volatile int alarm_signaled;
#ifdef CONFIG_FEATURE_HTTPD_CONFIG_WITH_SCRIPT_INTERPR
  Htaccess *script_i;           /* config script interpreters */
#endif
#ifdef CONFIG_FEATURE_HTTPD_SSI
  FILE *s_out;
  const char *shtml_url;
#endif
} bb_common_bss;
#define config bb_common_bss
#ifdef HTTPD_STANDALONE
struct BB_COMMON_BSS bb_common_bss;
#endif

static const char request_GET[] = "GET";    /* size algorithmic optimize */

static const char* const suffixTable [] = {
/* Warning: shorted equivalent suffix in one line must be first */
  ".htm.html", "text/html",
#if defined(CONFIG_FEATURE_HTTPD_SSI)
  ".shtm.shtml", "text/x-server-parsed-html",
#endif
  ".jpg.jpeg", "image/jpeg",
  ".gif", "image/gif",
  ".pdf", "application/pdf",
  ".png", "image/png",
  ".tgz", "application/x-gtar",
  ".txt.h.c.cc.cpp", "text/plain",
  ".tiff", "image/tiff",
  ".css", "text/css",
  ".wav", "audio/wav",
  ".avi", "video/x-msvideo",
  ".qt.mov", "video/quicktime",
  ".mpe.mpeg", "video/mpeg",
  ".mid.midi", "audio/midi",
  ".mp3", "audio/mpeg",
#if 0                        /* unpopular */
  ".au", "audio/basic",
  ".pac", "application/x-ns-proxy-autoconfig",
  ".vrml.wrl", "model/vrml",
#endif
  0, "application/octet-stream" /* default */
  };

typedef enum
{
  HTTP_OK = 200,
  HTTP_MOVED_TEMPORARILY = 302,
  HTTP_BAD_REQUEST = 400,       /* malformed syntax */
  HTTP_UNAUTHORIZED = 401, /* authentication needed, respond with auth hdr */
  HTTP_NOT_FOUND = 404,
  HTTP_FORBIDDEN = 403,
  HTTP_REQUEST_TIMEOUT = 408,
  HTTP_NOT_IMPLEMENTED = 501,   /* used for unrecognized requests */
  HTTP_INTERNAL_SERVER_ERROR = 500,
#if 0 /* future use */
  HTTP_CONTINUE = 100,
  HTTP_SWITCHING_PROTOCOLS = 101,
  HTTP_CREATED = 201,
  HTTP_ACCEPTED = 202,
  HTTP_NON_AUTHORITATIVE_INFO = 203,
  HTTP_NO_CONTENT = 204,
  HTTP_MULTIPLE_CHOICES = 300,
  HTTP_MOVED_PERMANENTLY = 301,
  HTTP_NOT_MODIFIED = 304,
  HTTP_PAYMENT_REQUIRED = 402,
  HTTP_BAD_GATEWAY = 502,
  HTTP_SERVICE_UNAVAILABLE = 503, /* overload, maintenance */
  HTTP_RESPONSE_SETSIZE=0xffffffff,
#endif
  HTTP_SEND_E404 = 604  /* internally, OK+404 */
} HttpResponseNum;

typedef struct
{
  HttpResponseNum type;
  const char *name;
  const char *info;
} HttpEnumString;

static const HttpEnumString httpResponseNames[] = {
  { HTTP_OK, "OK", NULL },
  { HTTP_MOVED_TEMPORARILY, "Found", "Directories must end with a slash." },
  { HTTP_REQUEST_TIMEOUT, "Request Timeout",
    "No request appeared within a reasonable time period." },
  { HTTP_NOT_IMPLEMENTED, "Not Implemented",
    "The requested method is not recognized by this server." },
#ifdef CONFIG_FEATURE_HTTPD_BASIC_AUTH
  { HTTP_UNAUTHORIZED, "Unauthorized", "" },
#endif
  { HTTP_NOT_FOUND, "Not Found",
    "The requested URL was not found on this server." },
  { HTTP_BAD_REQUEST, "Bad Request", "Unsupported method." },
  { HTTP_FORBIDDEN, "Forbidden", "" },
  { HTTP_INTERNAL_SERVER_ERROR, "Internal Server Error",
    "Internal Server Error" },
#if 0                               /* not implemented */
  { HTTP_CREATED, "Created" },
  { HTTP_ACCEPTED, "Accepted" },
  { HTTP_NO_CONTENT, "No Content" },
  { HTTP_MULTIPLE_CHOICES, "Multiple Choices" },
  { HTTP_MOVED_PERMANENTLY, "Moved Permanently" },
  { HTTP_NOT_MODIFIED, "Not Modified" },
  { HTTP_BAD_GATEWAY, "Bad Gateway", "" },
  { HTTP_SERVICE_UNAVAILABLE, "Service Unavailable", "" },
#endif
  { HTTP_SEND_E404, "Not Found", NULL }
};


static const char RFC1123FMT[] = "%a, %d %b %Y %H:%M:%S GMT";
static const char Content_length[] = "Content-length:";


static int
scan_ip (const char **ep, unsigned int *ip, unsigned char endc)
{
  const char *p = *ep;
  int auto_mask = 8;
  int j;

  *ip = 0;
  for (j = 0; j < 4; j++) {
	unsigned int octet;

	if ((*p < '0' || *p > '9') && (*p != '/' || j == 0) && *p != 0)
	  return -auto_mask;
	octet = 0;
	while (*p >= '0' && *p <= '9') {
	  octet *= 10;
	  octet += *p - '0';
	  if (octet > 255)
		return -auto_mask;
	  p++;
	}
	if (*p == '.')
	  p++;
	if (*p != '/' && *p != 0)
	  auto_mask += 8;
	*ip = ((*ip) << 8) | octet;
  }
  if (*p != 0) {
	if (*p != endc)
		return -auto_mask;
	p++;
	if(*p == 0)
		return -auto_mask;
  }
  *ep = p;
  return auto_mask;
}

static int
scan_ip_mask (const char *ipm, unsigned int *ip, unsigned int *mask)
{
  int i;
  unsigned int msk;

  i = scan_ip(&ipm, ip, '/');
  if(i < 0)
	return i;
  if(*ipm) {
	const char *p = ipm;

	i = 0;
	while (*p) {
		if (*p < '0' || *p > '9') {
			if (*p == '.') {
				i = scan_ip (&ipm, mask, 0);
				return i != 32;
			}
			return -1;
		}
		i *= 10;
		i += *p - '0';
		p++;
	}
  }
  if (i > 32 || i < 0)
	  return -1;
  msk = 0x80000000;
  *mask = 0;
  while (i > 0) {
	*mask |= msk;
	msk >>= 1;
	i--;
  }
  return 0;
}

static void free_config_lines(Htaccess **pprev)
{
    Htaccess *prev = *pprev;

    while( prev ) {
	Htaccess *cur = prev;

	prev = cur->next;
	free(cur);
    }
    *pprev = NULL;
}

/* flag */
#define FIRST_PARSE          0
#define SUBDIR_PARSE         1
#define SIGNALED_PARSE       2
#define FIND_FROM_HTTPD_ROOT 3
/****************************************************************************
 *
 > $Function: parse_conf()
 *
 * $Description: parse configuration file into in-memory linked list.
 *
 * The first non-white character is examined to determine
 *
 * Any previous IP rules are discarded.
 * If the flag argument is not SUBDIR_PARSE then all /path and mime rules
 * are also discarded.  That is, previous settings are retained if flag is
 * SUBDIR_PARSE.
 *
 * $Parameters:
 *      (const char *) path . . null for ip address checks, path for password
 *                              checks.
 *      (int) flag  . . . . . . the source of the parse request.
 *
 * $Return: (None)
 *
 ****************************************************************************/
static void parse_conf(const char *path, int flag)
{
    FILE *f;
    Htaccess *cur;
#ifdef CONFIG_FEATURE_HTTPD_BASIC_AUTH
    Htaccess *prev;
#endif

    const char *cf = config.configFile;
    char buf[160];
    char *p0 = NULL;
    char *c, *p;

    /* free previous ip setup if present */
    Htaccess_IP *pip = config.ip_a_d;

    while( pip ) {
	Htaccess_IP *cur_ipl = pip;

	pip = cur_ipl->next;
	free(cur_ipl);
    }
    config.ip_a_d = NULL;

    /* retain previous config only for subdir parse */
    if(flag != SUBDIR_PARSE) {
	free_config_lines(&config.htindex);
	free_config_lines(&config.e404);
#ifdef CONFIG_FEATURE_HTTPD_BASIC_AUTH
	free_config_lines(&config.auth);
#endif
#ifdef CONFIG_FEATURE_HTTPD_CONFIG_WITH_MIME_TYPES
	free_config_lines(&config.mime_a);
#endif
#ifdef CONFIG_FEATURE_HTTPD_CONFIG_WITH_SCRIPT_INTERPR
	free_config_lines(&config.script_i);
#endif
    }

    if(flag == SUBDIR_PARSE || cf == NULL) {
	cf = alloca(strlen(path) + sizeof(httpd_conf) + 2);
	if(cf == NULL) {
	    if(flag == FIRST_PARSE)
		bb_error_msg_and_die("%s", bb_msg_memory_exhausted);
	    return;
	}
	sprintf((char *)cf, "%s/%s", path, httpd_conf);
    }

    while((f = fopen(cf, "r")) == NULL) {
#if DEBUG
	fprintf(stderr, "parse_conf fopen %s == NULL\n", cf);
#endif
	if(flag == SUBDIR_PARSE) {
	    /* config file not found, no changes to config */
	    return;
	}
	if(flag == FIND_FROM_HTTPD_ROOT) {
	    /* goto set default autoindexied file */
	    goto set_aif;
	}
	if(config.configFile && flag == FIRST_PARSE) /* if -c option given */
	    bb_perror_msg_and_die("%s", cf);
	flag = FIND_FROM_HTTPD_ROOT;
	cf = httpd_conf;
    }

#ifdef CONFIG_FEATURE_HTTPD_BASIC_AUTH
    prev = config.auth;
#endif
    /* This could stand some work */
    while ( (p0 = fgets(buf, sizeof(buf), f)) != NULL) {
	c = NULL;
	for(p = p0; *p0 != 0 && *p0 != '#'; p0++) {
		if(!isspace(*p0)) {
		    *p++ = *p0;
		    if(*p0 == ':' && c == NULL)
			c = p;
		}
	}
	*p = 0;

	/* test for empty or strange line */
	if (c == NULL || *c == 0)
	    continue;
	p0 = buf;
	if(*p0 == 'd')
	    *p0 = 'D';
	if(*c == '*') {
	    if(*p0 == 'D') {
		/* deny all */
		c = "0.0.0.0/0";
	    }
	    /* skip default other "word:*" config lines */
	    continue;
	}

	if(*p0 == 'a')
	    *p0 = 'A';
	else if(*p0 == 'i')
	    *p0 = 'I';
	else if(*p0 == 'e')
	    *p0 = 'E';
	else if(*p0 != 'D' && *p0 != 'A' && *p0 != 'I' && *p0 != 'E'
#ifdef CONFIG_FEATURE_HTTPD_BASIC_AUTH
	   && *p0 != '/'
#endif
#ifdef CONFIG_FEATURE_HTTPD_CONFIG_WITH_MIME_TYPES
	   && *p0 != '.'
#endif
#ifdef CONFIG_FEATURE_HTTPD_CONFIG_WITH_SCRIPT_INTERPR
	   && *p0 != '*'
#endif
	  )
	       continue;
	if(*p0 == 'A' || *p0 == 'D') {
		/* storing current config IP line */
		pip = calloc(1, sizeof(Htaccess_IP));
		if(pip) {
		    /* add to bottom A|D:form_IP{/mask} config line */
		    Htaccess_IP *prev_IP = config.ip_a_d;
		    if(scan_ip_mask (c, &(pip->ip), &(pip->mask))) {
			/* syntax IP{/mask} error detected, protect all */
			*p0 = 'D';
			pip->mask = 0;
		    }
		    pip->allow_deny = *p0;
		    if(prev_IP == NULL) {
			config.ip_a_d = pip;
		    } else {
			while(prev_IP->next)
			    prev_IP = prev_IP->next;
			prev_IP->next = pip;
		    }
		}
		continue;
	}
#ifdef CONFIG_FEATURE_HTTPD_BASIC_AUTH
	if(*p0 == '/') {
	    /* make full path from httpd root / curent_path / config_line_path */
	    size_t lp0;
	    cf = flag == SUBDIR_PARSE ? path : "";
	    p0 = malloc(strlen(cf) + (c - buf) + 2 + strlen(c));
	    if(p0 == NULL)
		continue;
	    c[-1] = 0;
	    sprintf(p0, "/%s%s", cf, buf);

	    /* another call bb_simplify_path */
	    cf = p = p0;

	    do {
		    if (*p == '/') {
			if (*cf == '/') {    /* skip duplicate (or initial) slash */
			    continue;
			} else if (*cf == '.') {
			    if (cf[1] == '/' || cf[1] == 0) { /* remove extra '.' */
				continue;
			    } else if ((cf[1] == '.') && (cf[2] == '/' || cf[2] == 0)) {
				++cf;
				if (p > p0) {
				    while (*--p != '/');    /* omit previous dir */
				}
				continue;
			    }
			}
		    }
		    *++p = *cf;
	    } while (*++cf);

	    if ((p == p0) || (*p != '/')) {      /* not a trailing slash */
		++p;                             /* so keep last character */
	    }
	    *p = 0;
	    lp0 = strlen(p0);
	    p0[lp0] = ':';
	    strcpy(p0 + lp0 + 1, c);
	}
#endif

	/* storing current config line */
	cur = calloc(1, sizeof(Htaccess) + strlen(p0));
	if(cur) {
	    cf = strcpy(cur->before_colon, p0);
	    cur->after_colon = strchr_set0(cur->before_colon, ':') + 1;

	    if(*cf == 'I') {
		/* config pre-HTML index file move top for overwrite previous */
		cur->next = config.htindex;
		config.htindex = cur;
		continue;
	    }

	    if(*cf == 'E') {
		/* config error 404 page file move top for overwrite previous */
		cur->next = config.e404;
		config.e404 = cur;
		continue;
	    }

#ifdef CONFIG_FEATURE_HTTPD_CONFIG_WITH_MIME_TYPES
	    if(*cf == '.') {
		/* config .mime line move top for overwrite previous */
		cur->next = config.mime_a;
		config.mime_a = cur;
		continue;
	    }
#endif
#ifdef CONFIG_FEATURE_HTTPD_CONFIG_WITH_SCRIPT_INTERPR
	    if(*cf == '*') {
		/* config script interpreter line move top for overwrite previous */
		cur->next = config.script_i;
		config.script_i = cur;
		continue;
	    }
#endif
#ifdef CONFIG_FEATURE_HTTPD_BASIC_AUTH
	    free(p0);
	    if(prev == NULL) {
		/* first line */
		config.auth = prev = cur;
	    } else {
		/* sort path, if current lenght eq or bigger then move up */
		Htaccess *prev_hti = config.auth;
		size_t l = strlen(cf);
		Htaccess *hti;

		for(hti = prev_hti; hti; hti = hti->next) {
		    if(l >= strlen(hti->before_colon)) {
			/* insert before hti */
			cur->next = hti;
			if(prev_hti != hti) {
			    prev_hti->next = cur;
			} else {
			    /* insert as top */
			    config.auth = cur;
			}
			break;
		    }
		    if(prev_hti != hti)
			    prev_hti = prev_hti->next;
		}
		if(!hti)  {       /* not inserted, add to bottom */
		    prev->next = cur;
		    prev = cur;
		}
	    }
#endif
	}
   }
   fclose(f);
set_aif:
   if(config.htindex == NULL) {
    /* set default */
    cur = calloc(1, sizeof(Htaccess) + sizeof("index.html"));
    if(cur != NULL) {
	cur->after_colon = cur->before_colon;
	strcpy(cur->before_colon, "index.html");    /* unused */
	config.htindex = cur;
    }
   }
}

#ifdef CONFIG_FEATURE_HTTPD_ENCODE_URL_STR
/****************************************************************************
 *
 > $Function: encodeString()
 *
 * $Description: Given a string, html encode special characters.
 *   This is used for the -e command line option to provide an easy way
 *   for scripts to encode result data without confusing browsers.  The
 *   returned string pointer is memory allocated by malloc().
 *
 * $Parameters:
 *      (const char *) string . . The first string to encode.
 *
 * $Return: (char *) . . . .. . . A pointer to the encoded string.
 *
 * $Errors: Returns a null string ("") if memory is not available.
 *
 ****************************************************************************/
static char *encodeString(const char *string)
{
  /* take the simple route and encode everything */
  /* could possibly scan once to get length.     */
  int len = strlen(string);
  char *out = malloc(len * 6 + 1);
  char *p=out;
  char ch;

  if (out == NULL) return "";
  while ((ch = *string++)) {
    /* very simple check for what to encode */
    /* XXX: UTF-8? */
    if (isalnum(ch)) *p++ = ch;
    else p += sprintf(p, "&#%d;", (unsigned char) ch);
  }
  *p=0;
  return out;
}
#endif          /* CONFIG_FEATURE_HTTPD_ENCODE_URL_STR */

/****************************************************************************
 *
 > $Function: decodeString()
 *
 * $Description: Given a URL encoded string, convert it to plain ascii.
 *   Since decoding always makes strings smaller, the decode is done in-place.
 *   Thus, callers should strdup() the argument if they do not want the
 *   argument modified.  The return is the original pointer, allowing this
 *   function to be easily used as arguments to other functions.
 *
 * $Parameters:
 *      (char *) string . . . The first string to decode.
 *      (int)    flag   . . . 1 if require decode '+' as ' ' for CGI
 *
 * $Return: (char *)  . . . . A pointer to the decoded string (same as input).
 *
 * $Errors: None
 *
 ****************************************************************************/
#define DECODESTRING_PARANOIA_CHECK 2
static char *decodeString(char *orig, int flag_plus_to_space)
{
  /* note that decoded string is always shorter than original */
  char *string = orig;
  char *ptr = string;

  while (*ptr)
  {
    if (*ptr == '+' && flag_plus_to_space)    { *string++ = ' '; ptr++; }
    else if (*ptr != '%') *string++ = *ptr++;
    else  {
      unsigned int value1, value2;

      ptr++;
      if(sscanf(ptr, "%1X", &value1) != 1 || sscanf(ptr+1, "%1X", &value2) != 1) {
	if((flag_plus_to_space & DECODESTRING_PARANOIA_CHECK))
	    return NULL;
	*string++ = '%';
      } else {
	value1 = value1 * 16 + value2;
	if((flag_plus_to_space & DECODESTRING_PARANOIA_CHECK) &&
		(value1 == '/' || value1 == 0))
	    return orig + 1;
	*string++ = value1;
	ptr += 2;
      }
    }
  }
  *string = '\0';
  return orig;
}


#if defined(CONFIG_FEATURE_HTTPD_CGI) || defined(CONFIG_FEATURE_HTTPD_SSI)
/****************************************************************************
 *
 > $Function: addEnv()
 *
 * $Description: Add an environment variable setting to the global list.
 *    A NAME=VALUE string is allocated, filled, and added to the list of
 *    environment settings passed to the cgi execution script.
 *
 * $Parameters:
 *  (char *) name_before_underline - The first part environment variable name.
 *  (char *) name_after_underline  - The second part environment variable name.
 *  (char *) value  . . The value to which the env variable is set.
 *
 * $Return: (void)
 *
 * $Errors: Silently returns if the env runs out of space to hold the new item
 *
 ****************************************************************************/
static void addEnv(const char *name_before_underline,
			const char *name_after_underline, const char *value)
{
  char *s = NULL;
  const char *underline;

  if (!value)
	value = "";
  underline = *name_after_underline ? "_" : "";
  asprintf(&s, "%s%s%s=%s", name_before_underline, underline,
					name_after_underline, value);
  if(s) {
    putenv(s);
  }
}

#if defined(CONFIG_FEATURE_HTTPD_SET_REMOTE_PORT_TO_ENV) || !defined(CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY)
/* set environs SERVER_PORT and REMOTE_PORT */
static void addEnvPort(const char *port_name)
{
      char buf[16];

      sprintf(buf, "%u", config.port);
      addEnv(port_name, "PORT", buf);
}
#endif

#endif          /* CONFIG_FEATURE_HTTPD_CGI || CONFIG_FEATURE_HTTPD_SSI */


#ifdef CONFIG_FEATURE_HTTPD_BASIC_AUTH
static const char empty_auth[] = ":";
/****************************************************************************
 *
 > $Function: decodeBase64()
 *
 > $Description: Decode a base 64 data stream as per rfc1521.
 *    Note that the rfc states that none base64 chars are to be ignored.
 *    Since the decode always results in a shorter size than the input, it is
 *    OK to pass the input arg as an output arg.
 *
 * $Parameter:
 *      (char *) Data . . . . A pointer to a base64 encoded string.
 *                            Where to place the decoded data.
 *
 * $Return: void
 *
 * $Errors: None
 *
 ****************************************************************************/
static void decodeBase64(char *Data)
{

  const unsigned char *in = (const unsigned char *)Data;
  // The decoded size will be at most 3/4 the size of the encoded
  unsigned long ch = 0;
  int i = 0;

  while (*in) {
    int t = *in++;

    if(t >= '0' && t <= '9')
	t = t - '0' + 52;
    else if(t >= 'A' && t <= 'Z')
	t = t - 'A';
    else if(t >= 'a' && t <= 'z')
	t = t - 'a' + 26;
    else if(t == '+')
	t = 62;
    else if(t == '/')
	t = 63;
    else if(t == '=')
	t = 0;
    else
	continue;

    ch = (ch << 6) | t;
    i++;
    if (i == 4) {
	*Data++ = (char) (ch >> 16);
	*Data++ = (char) (ch >> 8);
	*Data++ = (char) ch;
	i = 0;
    }
  }
  *Data = 0;
}
#endif

/****************************************************************************
 *
 > $Function: sendHeaders()
 *
 * $Description: Create and send HTTP response headers.
 *   The arguments are combined and sent as one write operation.  Note that
 *   IE will puke big-time if the headers are not sent in one packet and the
 *   second packet is delayed for any reason.
 *
 * $Parameter:
 *      (HttpResponseNum) responseNum . . . The result code to send.
 *
 * $Return: (int)  . . . . writing errors
 *
 ****************************************************************************/
static int sendHeaders(HttpResponseNum responseNum)
{
  char *buf = config.buf;
  const char *responseString = "";
  const char *infoString = 0;
  const char *mime_type;
  unsigned int i;
  time_t timer = time(0);
  char timeStr[80];
  int len;

  for (i = 0;
	i < (sizeof(httpResponseNames)/sizeof(httpResponseNames[0])); i++) {
		if (httpResponseNames[i].type == responseNum) {
			responseString = httpResponseNames[i].name;
			infoString = httpResponseNames[i].info;
			break;
		}
  }
  /* error message is HTML */
  mime_type = responseNum == HTTP_OK || responseNum == HTTP_SEND_E404 ?
		config.httpd_found.found_mime_type : "text/html";

  /* emit the current date */
  strftime(timeStr, sizeof(timeStr), RFC1123FMT, gmtime(&timer));
  len = sprintf(buf,
	"HTTP/1.0 %d %s\r\nContent-type: %s\r\n"
	"Date: %s\r\nConnection: close\r\n",
	  (responseNum == HTTP_SEND_E404 ? HTTP_NOT_FOUND : responseNum),
	  responseString, mime_type, timeStr);

#ifdef CONFIG_FEATURE_HTTPD_BASIC_AUTH
  if (responseNum == HTTP_UNAUTHORIZED) {
    len += sprintf(buf+len, "WWW-Authenticate: Basic realm=\"%s\"\r\n",
							    config.realm);
  }
#endif
  if(responseNum == HTTP_MOVED_TEMPORARILY) {
	len += sprintf(buf+len, "Location: %s\r\n",
		config.httpd_found.found_moved_temporarily);
  }

  if (config.ContentLength != -1) {    /* file */
    strftime(timeStr, sizeof(timeStr), RFC1123FMT, gmtime(&config.last_mod));
    len += sprintf(buf+len, "Last-Modified: %s\r\n%s %" P_ST_SIZE "d\r\n",
			      timeStr, Content_length, config.ContentLength);
  }
  strcat(buf, "\r\n");
  len += 2;
  if (infoString) {
    len += sprintf(buf+len,
	    "<HEAD><TITLE>%d %s</TITLE></HEAD>\n"
	    "<BODY><H1>%d %s</H1>\n%s\n</BODY>\n",
	    responseNum, responseString,
	    responseNum, responseString, infoString);
  }
#if DEBUG
  fprintf(stderr, "Headers: '%s'", buf);
#endif
  return bb_full_write(a_c_w, buf, len);
}

#ifdef CONFIG_FEATURE_HTTPD_DIR_LIST
#include <dirent.h>
typedef struct namelist {       /* Obviously a list of names. */
	struct namelist *next;
	char            name[1];
} namelist_t;

static namelist_t *collect(DIR *dp)
/* Return a sorted list of directory entries. */
{
    namelist_t *pn;
    namelist_t *s, *prev;
    char *n;
    struct dirent *entry;
    size_t lcur;
    static namelist_t parent_directory; /* name[0]=0; next=NULL; */
    namelist_t *top = &parent_directory;

    while ((entry = readdir(dp)) != NULL) {
	n = entry->d_name;
	if (n[0] == '.')
	    continue;           /* do not show any .name */
	if(strcmp(n, httpd_conf) == 0)
	    continue;           /* do not show httpd.conf */
	lcur = strlen(n);
	pn = malloc(sizeof(namelist_t) + lcur);
	if (pn == NULL)
	    break;
	strcpy(pn->name, n);
	/* very slow sort */
	for(s = prev = top; s; s = s->next) {
		if(strcmp(n, s->name) < 0)
			break;
		if(s != prev)
			prev = prev->next;
	}
	s = prev->next;
	prev->next = pn;
	pn->next = s;
    }
    closedir(dp);
    return top;
}

static void write_string_escape(const char *s, int flg_href_escape)
{
    const char *start;
    char f[6];
    static const char must_escape[]="\"#%&<>?[\\]^`{|}\127";
    char *pout;

    while(*s) {
	pout = NULL;
	start = s;
	while(*s) {
	    if(*s == '<') {
		pout = "\004&lt;";
		break;
	    } else if(*s == '>') {
		pout = "\004&gt;";
		break;
	    } else if(*s == '&') {
		pout = "\005&apm;";
		break;
	    }
	    if(flg_href_escape) {
		if((signed char)*s <= ' ' || strchr(must_escape, *s)) {
		    sprintf(f, "\003%%%2x", (unsigned)(unsigned char)*s);
		    pout = f;
		    break;
		}
	    }
	    s++;
	}
	if(start != s)
	    bb_full_write(a_c_w, start, (s-start));
	if(pout) {
	    bb_full_write(a_c_w, pout + 1, (int)pout[0]);
	    s++;
	}
    }
}

static int dir2html(const char *vpath
#ifdef CONFIG_FEATURE_HTTPD_SSI
			, int from_ssi
#endif
			)
{
    DIR *dp;
    namelist_t *np;
    char work[128];
    char work2[32];
    char *filename;
    struct stat st;
    struct tm *tmp;
    char *n;
    static const char month[][4] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec",
    };
    const char *pout;
    size_t l;

    if(vpath[1] != '\0') {
	if(chdir(vpath+1))
	    return HTTP_NOT_FOUND;
    }
    if ((dp = opendir(dot)) == NULL)
	return HTTP_INTERNAL_SERVER_ERROR;
    np = collect(dp);
#ifdef CONFIG_FEATURE_HTTPD_SSI
    if(!from_ssi)
#endif
    {
	sendHeaders(HTTP_OK);
	pout = "<HTML><HEAD><TITLE>Index of ";
	bb_full_write(a_c_w, pout, strlen(pout));
	write_string_escape(vpath, 0);
	pout = "</TITLE></HEAD>\n<BODY>\n";
	bb_full_write(a_c_w, pout, strlen(pout));
    }
    pout = "<H1>Index of ";
    bb_full_write(a_c_w, pout, strlen(pout));
    write_string_escape(vpath, 0);
    sprintf(work, "</H1>\n<PRE>\n%-22s  %-17s  Size/Type\n<HR>\n",
			"Name", "Last modified");
    bb_full_write(a_c_w, work, strlen(work));
    while (np != NULL) {
	pout = filename = np->name;
	np = np->next;

	if(*filename == '\0') {
		/* .. */
		if(vpath[1] == '\0') {
		    /* root/.. is root */
		    filename = dot;
		} else {
		    filename = "..";
		}
	}
	if (stat(filename, &st) < 0)
		continue;
	if(filename != pout) {
	    /* .. */
	    if(!S_ISDIR(st.st_mode))
		continue;               /* hmm */
	    filename = strdup(vpath);
	    if(filename == NULL)
		continue;
	    n = strrchr(filename, '/');
	    *n = '\0';
	    n = strrchr(filename, '/');
	    if(n)
		*n = '\0';
	}
	bb_full_write(a_c_w, "<A HREF=\"", 9);
	write_string_escape(filename, 1);
	if(S_ISDIR(st.st_mode))
		bb_full_write(a_c_w, "/\">", 3);
	else
		bb_full_write(a_c_w, "\">", 2);
	l = 0;
	if (filename != pout) {
		free(filename);
		bb_full_write(a_c_w, "Parent Directory", 16);
	} else {
		if(strlen(filename) <= 22) {
		    l = sprintf(work2, "%s%s",
			filename, S_ISDIR(st.st_mode) ? "/" : "");
		} else {
		    sprintf(work2, "%.20s..>", filename);
		}
		write_string_escape(work2, 0);
	}
	if(l > 0 && l < 23)
	    sprintf(work2, "%*c", (l-23), ' ');
	else
	    work2[0] = '\0';
	tmp = localtime(&st.st_mtime);
	l = sprintf(work, "</A>%s %02d %s %d %02d:%02d ", work2,
	    tmp->tm_mday, month[tmp->tm_mon], 1900+tmp->tm_year,
	    tmp->tm_hour, tmp->tm_min);
	bb_full_write(a_c_w, work, l);
	if (S_ISREG(st.st_mode)) {
	    if (st.st_size < 10240) {
		sprintf(work2, "%d", (int) st.st_size);
	    } else if (st.st_size < 10240L * 1024) {
		sprintf(work2, "%dK", (int)((st.st_size - 1) / 1024) + 1);
	    } else
#if defined(CONFIG_LFS) || defined(NATIVE_64)
		    if (st.st_size < 10240LL * 1024 * 1024)
#endif
	    {
		sprintf(work2, "%dM", (int)((st.st_size-1)/(1024L * 1024))+1);
	    }
#if defined(CONFIG_LFS) || defined(NATIVE_64)
	      else {
		sprintf(work2, "%luG", (unsigned long)
		    ((st.st_size - 1) / (1024LL * 1024 * 1024) + 1));
	    }
#endif
	} else {
	    strcpy(work2,
		S_ISDIR(st.st_mode) ? "[dir]" :
		S_ISBLK(st.st_mode) ? "[block]" :
		S_ISCHR(st.st_mode) ? "[char]" :
		S_ISFIFO(st.st_mode) ? "[pipe]" :
					"[???]");
	}
	l = sprintf(work, "%s\n", work2);
	bb_full_write(a_c_w, work, l);
    }

    l = sprintf(work, "</PRE>\n<HR>\n<SMALL><i>%s</i></SMALL>\n", httpdVersion);
    bb_full_write(a_c_w, work, l);
#ifdef CONFIG_FEATURE_HTTPD_SSI
    if(!from_ssi)
#endif
	bb_full_write(a_c_w, "</BODY>\n</HTML>\n", 16);
    return HTTP_OK;
}
#endif /* CONFIG_FEATURE_HTTPD_DIR_LIST */

#ifndef CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY
/****************************************************************************
 *
 > $Function: openServer()
 *
 * $Description: create a listen server socket on the designated port.
 *
 * $Return: (int)  . . . A connection socket. -1 for errors.
 *
 * $Errors: None
 *
 ****************************************************************************/
#ifdef CONFIG_FEATURE_HTTPD_BIND
static int openServer(struct sockaddr_in *lsocket)
#else
static int openServer(void)
#endif
{
  int fd;

#ifndef CONFIG_FEATURE_HTTPD_BIND
  struct sockaddr_in lsocket;
  /* inet_addr() returns a value that is already in network order */
  memset(&lsocket, 0, sizeof(lsocket));
  lsocket.sin_family = AF_INET;
  lsocket.sin_addr.s_addr = INADDR_ANY;
  lsocket.sin_port = htons(config.port);
#endif
  /* create the socket right now */
  fd = bb_xsocket(BB_XSOCKET_FLG_ERR_DIE);
  /* tell the OS it's OK to reuse a previous address even though */
  /* it may still be in a close down state.  Allows bind to succeed. */
#ifdef SO_REUSEPORT
  setsockopt_1i(fd, SOL_SOCKET, SO_REUSEPORT, 1);
#else
  setsockopt_1i(fd, SOL_SOCKET, SO_REUSEADDR, 1);
#endif

#ifdef CONFIG_FEATURE_HTTPD_BIND
  bb_xbind(fd, (struct sockaddr *)lsocket, sizeof(struct sockaddr_in));
#else
  bb_xbind(fd, (struct sockaddr *)&lsocket, sizeof(lsocket));
#endif
  listen(fd, 9);
  /* prevent zombie (defunct) processes */
  bb_signal((1 << SIGCHLD), SIG_IGN);
  return fd;
}
#endif  /* CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY */


/****************************************************************************
 *
 > $Function: getLine()
 *
 * $Description: Read from the socket until an end of line char found.
 *
 *   Characters are read one at a time until an eol sequence is found.
 *
 * $Return: (int) . . . . number of characters read.  -1 if error.
 *
 ****************************************************************************/
static int getLine(void)
{
  int  count = 0;
  char *buf = config.buf;

  while (read(a_c_r, buf + count, 1) == 1) {
    if (buf[count] == '\r') continue;
    if (buf[count] == '\n') {
      buf[count] = 0;
      return count;
    }
    if(count < (MAX_MEMORY_BUFF-1))      /* check owerflow */
	count++;
  }
  if (count) return count;
  else return -1;
}

#if defined(CONFIG_FEATURE_HTTPD_SSI)
static char* ssi_timefmt;
#define TIMES_BUF_SIZE 256

static int ssi_show_time( time_t t, int gmt, char *buf )
{
  struct tm* tmP;

  if ( gmt )
      tmP = gmtime( &t );
  else
      tmP = localtime( &t );
  return strftime( buf, TIMES_BUF_SIZE, ssi_timefmt, tmP );
}

typedef enum
{
 UNKNOWN_DIRECTIVE = 1,
 UNKNOWN_TAG,
 UNKNOWN_VALUE,
 NOT_FOUND,
 NOT_PERMITTED,
 RECURSIVE_INCLUDE,
 INTERNAL_ERROR,
} ERR;

static void ssi_directive_err(ERR err, char* fmt, ...)
{
  va_list p;
  const char *emsg;

  switch( err ) {
       case UNKNOWN_DIRECTIVE: emsg = "Unknown directive";     break;
       case UNKNOWN_TAG:       emsg = "Unknown tag";           break;
       case UNKNOWN_VALUE:     emsg = "Unknown value";         break;
       case NOT_FOUND:         emsg = "File not found";        break;
       case NOT_PERMITTED:     emsg = "Not permitted";         break;
       case RECURSIVE_INCLUDE: emsg = "Recursive include of "; break;
       default:                emsg = "Internal error";        break;
  }

  fprintf(config.s_out, "<font color=red>&lt;[%s] %s ", config.shtml_url, emsg);
  va_start(p, fmt);
  vfprintf(config.s_out, fmt, p);
  va_end(p);
  fprintf(config.s_out, "&gt;</font>");
}

static void form_QUERY_STRING_UNESCAPED(const char *cp)
{
    if (cp) {
	char *dcp0;
	char *dcp;

	dcp0 = alloca(strlen(cp) + 1);
	if(dcp0 == NULL)
	    return;
	cp = decodeString(strcpy(dcp0, cp), 0);
	dcp0 = dcp = malloc(strlen(cp) * 2 + 1);
	if(dcp == NULL)
	    return;

	for(;*cp;) {
	    const char *p = strpbrk(cp, "&;`'\"|*?~<>^()[]{}$\\\n\r%");

	    if(p == NULL && dcp == dcp0)
		break;
	    while(*cp) {
		if(p) {
		    /* present for escape */
		    while(cp != p)
			*dcp++ = *cp++;
		    *dcp++ = '\\';
		    *dcp++ = *cp++;
		    break;
		} else {
		    *dcp++ = *cp++;
		}
	    }
	}
	if(dcp != dcp0)
	    *dcp = 0;
	addEnv("QUERY_STRING", "UNESCAPED", dcp != dcp0 ? dcp0 : cp);
	free(dcp0);
    }
}

static void form_env_conf(const char *conf_ssi_timefmt, const char *file)
{
	size_t p;
	struct stat sb;
	char timesbuf[TIMES_BUF_SIZE];

	/* The current date, local time zone. */
	if(conf_ssi_timefmt) {
		time_t t = time( (time_t*) 0 );

		p = ssi_show_time( t, 0, timesbuf );
		if(p > 0)
			addEnv("DATE", "LOCAL", timesbuf );
		p = ssi_show_time( t, 1, timesbuf );
		if(p > 0)
			addEnv("DATE", "GMT", timesbuf );
	}
	if(file) {
		if ( stat(file, &sb ) >= 0 ) {
			p = ssi_show_time( sb.st_mtime, 0, timesbuf );
			if(p > 0)
				addEnv("LAST_MODIFIED", "", timesbuf );
		}
	}
}

#endif

#if defined(CONFIG_FEATURE_HTTPD_CGI) || defined(CONFIG_FEATURE_HTTPD_SSI)

static void form_env_CGI_SSI(const char *request, const URI *uri)
{
	char *purl;
	unsigned int ms;

	if(request) {
	    /* set common env */
	    addEnv("DOCUMENT",       "ROOT",     config.document_root);
	    addEnv("REQUEST",        "METHOD",   request);
	    addEnv("SERVER",         "SOFTWARE", httpdVersion);
	    addEnv("SERVER",         "PROTOCOL", "HTTP/1.0");
	    addEnv("GATEWAY_INTERFACE", "",      "CGI/1.1");
	    addEnv("REMOTE",         "ADDR",     config.rmt_ip_str);
#ifdef CONFIG_FEATURE_HTTPD_SET_REMOTE_PORT_TO_ENV
	    addEnvPort("REMOTE");
#endif
	    if(config.cookie)
		addEnv("HTTP", "COOKIE", config.cookie);
	    if(config.referer)
		addEnv("HTTP", "REFERER", config.referer);
	    if(config.user_agent)
		addEnv("HTTP", "USER_AGENT", config.user_agent);
#ifdef CONFIG_FEATURE_HTTPD_BASIC_AUTH
	    if(config.usr_pass != empty_auth) {
		char *u = strchr_set0(config.usr_pass, ':');

		addEnv("REMOTE", "USER", config.usr_pass);
		*u = ':';
		addEnv("AUTH", "TYPE", "Basic");
	    }
#endif
	    if(config.content_type)
		addEnv("CONTENT", "TYPE", config.content_type);
	    addEnv("SCRIPT",       "NAME",      uri->url);
	}
	/* env dependece of url */
	addEnv("QUERY_STRING", "",          uri->query);
	ms = strlen(config.document_root) + strlen(uri->url) + 1;
	if(uri->path_info)
	    ms += strlen(uri->path_info) + 1;
	if(uri->query)
	    ms += strlen(uri->query) + 1;
	purl = alloca(ms);
	if(purl == NULL)
		return;
	sprintf(purl, "%s%s", config.document_root, uri->url);
	addEnv("SCRIPT",   "FILENAME", purl);
	if(uri->path_info) {
	    sprintf(purl, "/%s", uri->path_info);
	    addEnv("PATH", "INFO", purl);
	    sprintf(purl, "%s/%s", config.document_root, uri->path_info);
	    addEnv("PATH", "TRANSLATED", purl);
	}
	/* no rewrite from SSI #include virtual="/cgi-bin/cgi" */
	if(request) {
	    if(uri->path_info)
		sprintf(purl, "%s/%s", uri->url, uri->path_info);
	    else
		strcpy(purl, uri->url);
	    if(uri->query)
		strcpy(stpcpy(purl, "?"), uri->query);       /* do not overlap purl */
	    addEnv("REQUEST", "URI",   purl);
	}
}
#endif

#ifdef CONFIG_FEATURE_HTTPD_CGI

#ifndef PIPE_BUF
# define PIPESIZE 4096          /* amount of buffering in a pipe */
#else
# define PIPESIZE PIPE_BUF
#endif
#if PIPESIZE >= MAX_MEMORY_BUFF
# error "PIPESIZE >= MAX_MEMORY_BUFF"
#endif

static int exec_cgi(URI *uri)
{
    char *script;
    char **argp;
    int na = 2;
    char *plus;

    /* If a request includes query info in the URL (stuff after "?"), and
     * the query info does not contain "=" (indicative of a FORM submission),
     * then this routine is called to create the argument list to be passed
     * to the CGI script.
     */
    if(uri->query) {
	na++;
	if(strchr(uri->query, '=') == NULL) {
	    plus = uri->query;
	    while((plus = strchr(plus, '+')) != NULL) {
		na++;
		plus++;
	    }
	}
    }
    argp = alloca(na * sizeof(char *));
    if(argp == NULL)
	return 0;
    na = 1;
    if(uri->query) {
	if(strchr(uri->query, '=') == NULL) {
	    plus = uri->query;
	    while(1) {
		argp[na] = plus;
		plus = strchr_set0(plus, '+');
		if(plus)
		    plus++;
		decodeString(argp[na], 0);
		na++;
		if(plus == NULL)
		    break;
	    }
	} else {
	    decodeString(uri->query, 0);
	    argp[na++] = uri->query;
	}
    }
    argp[na] = NULL;
    /* set execve argp[0] without path */
    script = strrchr(uri->url + 1, '/');  /* cgi-bin/[path] */
    *script++ = '\0';
    argp[0] = script;
    if(chdir(uri->url + 1) == 0) {
#ifdef CONFIG_FEATURE_HTTPD_CONFIG_WITH_SCRIPT_INTERPR
	char *interpr = NULL;
	char *suffix = strrchr(uri->url, '.');

	if(suffix) {
		Htaccess * cur;
		for (cur = config.script_i; cur; cur = cur->next)
			if(strcmp(cur->before_colon + 1, suffix) == 0) {
				interpr = cur->after_colon;
				break;
			}
	}
	if (interpr)
		execv(interpr, argp);
	else
#endif
		execv(script, argp);
    }
#ifndef CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY
    config.accepted_socket = 1;      /* send to stdout */
#endif
    return 1;
}

#if defined(CONFIG_FEATURE_HTTPD_SSI)
/* execute CGI from SSI, SERVER_PROTOCOL=INCLUDED, remove Content-type: * */
static void sendCgi_included(URI *uri)
{
  int fromCgi[2];  /* pipe for reading data from CGI */
  int pid = 0;
  int inFd;
  int outFd;

  fflush(config.s_out);
  if (pipe(fromCgi) != 0 || (pid = fork()) < 0) {
	ssi_directive_err(INTERNAL_ERROR, "%m");
	return;
  }

  if (!pid) {
      /* child process */

      form_env_CGI_SSI(NULL, uri);
      addEnv("SERVER", "PROTOCOL", "INCLUDED");

      outFd = fromCgi[1];
      close(0);
      open(bb_dev_null, O_RDONLY);

      dup2(outFd, 1);  // replace stdout with the pipe
      if(!DEBUG)
	dup2(outFd, 2);  // replace stderr with the pipe

      close(fromCgi[0]);
      close(fromCgi[1]);
      if(exec_cgi(uri) == 0)
	ssi_directive_err(INTERNAL_ERROR, "%m");
      else
	ssi_directive_err(NOT_FOUND, "%s: %m", uri->url);
      fflush(config.s_out);
      _exit(242);
  } /* end child */

  if (pid) {
    /* parent process */
#if DEBUG
    int status;
#endif
    int firstLine = 1;

    inFd  = fromCgi[0];
    close(fromCgi[1]);
    bb_signal((1 << SIGPIPE), SIG_IGN);

    while (1) {
	char *rbuf = config.buf;
	int count;

	count = safe_read(inFd, rbuf, PIPESIZE);
	if (count == 0)
	    break;  /* closed */
	if (count > 0) {
#if DEBUG
		fprintf(stderr, "CGI read %d bytes\n", count);
#endif
	  if (firstLine) {
	    char *cr;

	    rbuf[count] = 0;
	    /* check to see if the user script have header for remove */
	    cr = strstr(rbuf, "ontent-");
	    if (cr) {
	      cr = strchr(cr, '\n');
	      if(cr)
		cr++;
		count -= (cr - rbuf);
		rbuf = cr;
	    }
	    firstLine = 0;
	  }
	  if (fwrite(rbuf, 1, count, config.s_out) != (size_t)count)
	      break;
	}
    }
    close(inFd);

#if DEBUG
    waitpid(pid, &status, 0);
    if (WIFEXITED(status))
	bb_error_msg("piped has exited with status=%d", WEXITSTATUS(status));
    if (WIFSIGNALED(status))
	bb_error_msg("piped has exited with signal=%d", WTERMSIG(status));
#else
    waitpid_no_int(pid, NULL, 0, 0);
#endif
  }
}
#endif

/****************************************************************************
 *
 > $Function: sendCgi()
 *
 * $Description: Execute a CGI script and send it's stdout back
 *
 *   Environment variables are set up and the script is invoked with pipes
 *   for stdin/stdout.  If a post is being done the script is fed the POST
 *   data in addition to setting the QUERY_STRING variable (for GETs or POSTs).
 *
 * $Parameters:
 *      (const char *) url . . . . . . The requested URL (with leading /).
 *      (int bodyLen)  . . . . . . . . Length of the post body.

 *
 * $Return: (char *)  . . . . A pointer to the decoded string (same as input).
 *
 * $Errors: None
 *
 ****************************************************************************/
static int sendCgi(URI *uri, const char *request, int bodyLen)
{
  int fromCgi[2];  /* pipe for reading data from CGI */
  int toCgi[2];    /* pipe for sending data to CGI */

  int pid = 0;
  int inFd;
  int outFd;
  int firstLine = 1;

  do {
    if (pipe(fromCgi) != 0) {
      break;
    }
    if (pipe(toCgi) != 0) {
      break;
    }

    pid = fork();
    if (pid < 0) {
	pid = 0;
	break;
    }

    if (!pid) {
      /* child process */

      form_env_CGI_SSI(request, uri);
      if(bodyLen) {
	char sbl[32];

	sprintf(sbl, "%d", bodyLen);
	addEnv("CONTENT", "LENGTH", sbl);
      }

      inFd  = toCgi[0];
      outFd = fromCgi[1];

      dup2(inFd, 0);  // replace stdin with the pipe
      dup2(outFd, 1);  // replace stdout with the pipe
      if(!DEBUG)
	dup2(outFd, 2);  // replace stderr with the pipe

      close(toCgi[0]);
      close(toCgi[1]);
      close(fromCgi[0]);
      close(fromCgi[1]);
      if(exec_cgi(uri) == 0)
	sendHeaders(HTTP_INTERNAL_SERVER_ERROR);
      else
	sendHeaders(HTTP_NOT_FOUND);
      _exit(242);
    } /* end child */

  } while (0);

  if (pid) {
    /* parent process */
#if DEBUG
    int status;
#endif
    size_t post_readed_size = 0, post_readed_idx = 0;

    inFd  = fromCgi[0];
    outFd = toCgi[1];
    close(fromCgi[1]);
    close(toCgi[0]);
    bb_signal((1 << SIGPIPE), SIG_IGN);

    while (1) {
      fd_set readSet;
      fd_set writeSet;
      char wbuf[128];
      int nfound;
      int count;

      FD_ZERO(&readSet);
      FD_ZERO(&writeSet);
      FD_SET(inFd, &readSet);
      if(bodyLen > 0 || post_readed_size > 0) {
	FD_SET(outFd, &writeSet);
	nfound = outFd > inFd ? outFd : inFd;
	if(post_readed_size == 0) {
		FD_SET(a_c_r, &readSet);
		if(nfound < a_c_r)
			nfound = a_c_r;
	}
      /* Now wait on the set of sockets! */
	nfound = select(nfound + 1, &readSet, &writeSet, 0, NULL);
      } else {
	if(!bodyLen) {
		close(outFd);
		bodyLen = -1;
	}
	nfound = select(inFd + 1, &readSet, 0, 0, NULL);
      }

      if (nfound <= 0) {
#if DEBUG
	if (waitpid(pid, &status, WNOHANG) > 0) {
	  close(inFd);
	  if (WIFEXITED(status))
	      bb_error_msg("piped has exited with status=%d", WEXITSTATUS(status));
	  if (WIFSIGNALED(status))
	      bb_error_msg("piped has exited with signal=%d", WTERMSIG(status));
	  break;
	}
#else
	if (waitpid_no_int(pid, NULL, WNOHANG, 0) > 0) {
	  close(inFd);
	  break;
	}
#endif
      } else if(post_readed_size > 0 && FD_ISSET(outFd, &writeSet)) {
		count = bb_full_write(outFd, wbuf + post_readed_idx, post_readed_size);
		if(count > 0) {
			post_readed_size -= count;
			post_readed_idx += count;
			if(post_readed_size == 0)
				post_readed_idx = 0;
		} else {
			post_readed_size = post_readed_idx = bodyLen = 0;
			/* broken pipe to CGI */
		}
      } else if(bodyLen > 0 && post_readed_size == 0 && FD_ISSET(a_c_r, &readSet)) {
		count = bodyLen > (int)sizeof(wbuf) ? (int)sizeof(wbuf) : bodyLen;
		count = safe_read(a_c_r, wbuf, count);
		if(count > 0) {
			post_readed_size += count;
			bodyLen -= count;
      } else {
			bodyLen = 0;    /* closed */
		}
      }
      if(FD_ISSET(inFd, &readSet)) {
	int s = a_c_w;
	char *rbuf = config.buf;

#ifndef PIPE_BUF
# define PIPESIZE 4096          /* amount of buffering in a pipe */
#else
# define PIPESIZE PIPE_BUF
#endif
#if PIPESIZE >= MAX_MEMORY_BUFF
# error "PIPESIZE >= MAX_MEMORY_BUFF"
#endif

	// There is something to read
	count = safe_read(inFd, rbuf, PIPESIZE);
	if (count == 0)
		break;  /* closed */
	if (count > 0) {
	  if (firstLine) {
	    rbuf[count] = 0;
	    /* check to see if the user script added headers */
	    if(strncmp(rbuf, "HTTP/1.0 200 OK\r\n", 4) != 0) {
	      bb_full_write(s, "HTTP/1.0 200 OK\r\n", 16);
	    }
	    firstLine = 0;
	  }
	  if (bb_full_write(s, rbuf, count) != count)
	      break;

#if DEBUG
	  fprintf(stderr, "CGI read %d bytes\n", count);
#endif
	}
      }
    }
  }
  return 0;
}
#endif          /* CONFIG_FEATURE_HTTPD_CGI */

static const char *find_mime_type(const char *url)
{
  const char * suffix;
  const char * const * table;
  const char * try_suffix;

  suffix = strrchr(url, '/');
  suffix = (suffix && suffix[1]) ? suffix + 1 : url;
  suffix = strrchr(suffix, '.');

  for (table = suffixTable; *table; table += 2)
	if(suffix != NULL && (try_suffix = strstr(*table, suffix)) != 0) {
		try_suffix += strlen(suffix);
		if(*try_suffix == 0 || *try_suffix == '.')
			break;
	}
  /* also, if not found, set default as "application/octet-stream";  */
  try_suffix = *(table+1);
#ifdef CONFIG_FEATURE_HTTPD_CONFIG_WITH_MIME_TYPES
  if (suffix) {
    Htaccess * cur;

    for (cur = config.mime_a; cur; cur = cur->next) {
	if(strcmp(cur->before_colon, suffix) == 0) {
		try_suffix = cur->after_colon;
		break;
	}
    }
  }
#endif  /* CONFIG_FEATURE_HTTPD_CONFIG_WITH_MIME_TYPES */
  config.httpd_found.found_mime_type = try_suffix;
  return try_suffix;
}

/****************************************************************************
 *
 > $Function: checkPerm()
 *
 * $Description: Check the permission file for access password protected.
 *
 *   If config file isn't present, everything is allowed.
 *   Entries are of the form you can see example from header source
 *
 * $Parameters:
 *      (const char *) path  . . . . The file path.
 *      (const char *) request . . . User information to validate.
 *
 * $Return: (int)  . . . . . . . . . 1 if request OK, 0 otherwise.
 *
 ****************************************************************************/

#ifdef CONFIG_FEATURE_HTTPD_BASIC_AUTH

static int checkPerm(const char *path, const char *request)
{
    Htaccess * cur;
    const char *p;
    const char *p0;

    const char *prev = NULL;

    /* This could stand some work */
    for (cur = config.auth; cur; cur = cur->next) {
	p0 = cur->before_colon;
	if(prev != NULL && strcmp(prev, p0) != 0)
	    continue;       /* find next identical */
	p = cur->after_colon;
#if DEBUG
	fprintf(stderr,"checkPerm: '%s' ? '%s'\n", p0, request);
#endif
	{
	    size_t l = strlen(p0);

	    if(strncmp(p0, path, l) == 0 &&
			    (l == 1 || path[l] == '/' || path[l] == 0)) {
		/* path match found.  Check request */
		/* for check next /path:user:password */
		prev = p0;
#ifdef CONFIG_FEATURE_HTTPD_AUTH_MD5
		{
			char *pp;
			char *u;

			u = strchr(request, ':');
			if(u == NULL || request == u) {
			    /* bad request, 'user:' required */
			    break;
			}
			if(strncmp(p, request, u-request) != 0) {
				/* user uncompared */
				continue;
			}
			pp = strchr(p, ':');
			if(pp && pp[1] == '$' && pp[2] == '1' &&
						 pp[3] == '$' && pp[4]) {
				char enc[ENCRYPT_BUFFER_SIZE];

				pp++;
				if (strcmp(pw_encrypt(u+1, pp, enc), pp) == 0)
					return 1;   /* Ok */
				/* unauthorized */
				continue;
			}
		}
#endif
		if (strcmp(p, request) == 0) {
		    return 1;   /* Ok */
		}
		/* unauthorized */
	    }
	}
    }   /* for */

    return prev == NULL;
}

#endif  /* CONFIG_FEATURE_HTTPD_BASIC_AUTH */

static int checkPermIP(void)
{
    Htaccess_IP * cur;

    /* This could stand some work */
    for (cur = config.ip_a_d; cur; cur = cur->next) {
#if DEBUG
	fprintf(stderr, "checkPermIP: '%s' ? ", config.rmt_ip_str);
	fprintf(stderr, "'%u.%u.%u.%u/%u.%u.%u.%u'\n",
		(unsigned char)(cur->ip >> 24),
		(unsigned char)(cur->ip >> 16),
		(unsigned char)(cur->ip >> 8),
				cur->ip & 0xff,
		(unsigned char)(cur->mask >> 24),
		(unsigned char)(cur->mask >> 16),
		(unsigned char)(cur->mask >> 8),
				cur->mask & 0xff);
#endif
	if((config.rmt_ip & cur->mask) == cur->ip)
	    return cur->allow_deny == 'A';   /* Allow/Deny */
    }

    /* if unconfigured, access from all */
    return 1;
}


static HttpResponseNum parse_uri(char *url, URI *uri, struct stat *psb)
{
    char        *cp, *purl, *q;
    Htaccess    *cur;
    size_t      il = 0, lcur = 0;
#ifdef CONFIG_FEATURE_HTTPD_DIR_LIST
    struct stat st;
#endif
    uri->found_e404 = 0;
    /* algorithm stolen from libbb bb_simplify_path(),
       but don`t strdup and reducing trailing slash and protect out root */
    purl = cp = url;
    do {
	if (*purl == '/') {
	    if (*cp == '/') {        /* skip duplicate (or initial) slash */
		continue;
	    } else if (*cp == '.') {
		if (cp[1] == '/' || cp[1] == 0) { /* skip extra '.' */
		    continue;
		} else if ((cp[1] == '.') && (cp[2] == '/' || cp[2] == 0)) {
		    ++cp;
		    if (purl == url) {
			/* deny out root */
			return HTTP_FORBIDDEN;
		    }
		    while (*--purl != '/');    /* omit previous dir */
		    continue;
		}
	    }
	}
	*++purl = *cp;
    } while (*++cp);

    *++purl = 0;        /* so keep last character */

    /* find query */
    q = strchr_set0(url, '?');
    if(q) {
	purl = q;
	q++;
    }
    cp = decodeString(url, DECODESTRING_PARANOIA_CHECK);
    if(cp == NULL)
	return HTTP_BAD_REQUEST;
    if(cp == (url+1))
	return HTTP_NOT_FOUND;

    if(!checkPermIP())
	return HTTP_FORBIDDEN;

    /* find path_info */
    /* check permission of client IP after parse found subdir config file */
    cp = url;
    while((cp = strchr_set0( cp + 1, '/' )) != NULL) {
	/* have script.cgi/PATH_INFO or dirs/script.cgi[/PATH_INFO] */
	if(is_directory(url + 1, 1, NULL) <= 0) {
	    /* not directory, found script.cgi/PATH_INFO */
	    *cp = '/';
	    break;
	}
	/* may be having subdir config */
	parse_conf(url + 1, SUBDIR_PARSE);
	if(!checkPermIP())
	    return HTTP_FORBIDDEN;
	*cp = '/';          /* is directory, find next '/' */
    }
    if(cp) {
	purl = cp;
	*cp++ = 0;
    }
    uri->path_info = cp;
    cp = url + 1;
    if(*cp == 0)
	cp = (char *)dot;
#if DEBUG
    fprintf(stderr, "lstat(%s)\n", cp);
#endif
    if (lstat( cp, psb ) < 0) {
	if(cp == dot)
		return HTTP_NOT_FOUND;
#if DEBUG
	fprintf(stderr, "start find e404 = %s\n", (config.e404 == NULL ? "NULL" : "setuped"));
#endif
	for(cur = config.e404; cur; cur = cur->next) {
	    char *url_dup = strdup(url);
	    if(url_dup == NULL)
		return HTTP_INTERNAL_SERVER_ERROR;
#if DEBUG
	    fprintf(stderr, "e404 = %s/%s\n", url_dup, cur->after_colon);
#endif
	    lcur = strlen(cur->after_colon);
	    while ( (cp = strrchr(url_dup, '/')) != NULL) {
		*cp = '\0';
		cp = malloc(strlen(url_dup) + lcur + 1);
		if(cp == NULL)
			return HTTP_INTERNAL_SERVER_ERROR;
		sprintf(cp, "%s/%s", url_dup, cur->after_colon);
#if DEBUG
		fprintf(stderr, "e404 lstat(%s)\n", cp + 1);
#endif
		if (lstat( cp + 1, psb )==0 && S_ISREG(psb->st_mode))
			break;
		free(cp);
#if DEBUG
		fprintf(stderr, "e404 lstat failed\n");
#endif
	    }
	    free(url_dup);
	    if(cp != NULL) {
		url = cp;
		uri->found_e404 = 1;
		break;
	    }
	}
	if(cur == NULL)
		return HTTP_NOT_FOUND;
    }
#if DEBUG
    fprintf(stderr, "lstat ok\n");
#endif

#ifdef CONFIG_FEATURE_HTTPD_BASIC_AUTH
    /* check access authentication (after load subdirs config) */
    if(checkPerm(url, config.usr_pass) == 0)
	return HTTP_UNAUTHORIZED;
#endif

    if(S_ISDIR(psb->st_mode)) {
#ifdef CONFIG_FEATURE_HTTPD_CGI
	if(strncmp(url, "/cgi-bin/", 8) == 0) {
	    if(url[8] == 0 || url[8] == '/') {
		/* deny listing any /cgi-bin[/[dirs]] */
		return HTTP_FORBIDDEN;
	    }
	    /* but for example /cgi-bin2 is not /cgi-bin and allow */
	}
#endif
	/* find max strlen index file name */
	for(cur = config.htindex; cur; cur = cur->next) {
	    lcur = strlen(cur->after_colon);
	    if(lcur > il)
		il = lcur;
	}
    }
    /* restore URI, allocate memory for save */
    if(q)
	il += strlen(q) + 1;
    cp = uri->path_info;
    if(cp)
	il += strlen(cp) + 1;
    il += strlen(url) + 2;          /* /\0 */
    cp = malloc(il);
    if(cp == NULL)
	return HTTP_INTERNAL_SERVER_ERROR;
    uri->url = strcpy(cp, url);
    il = strlen(url);
    if(S_ISDIR(psb->st_mode)) {
	if(purl[-1] != '/') {
	    cp[il++] = '/';
	    if(q) {
		cp[il] = '?';
		strcpy(cp + il + 1, q);
	    } else
		cp[il] = 0;
	    config.httpd_found.found_moved_temporarily = cp;
	    return HTTP_MOVED_TEMPORARILY;
	}
#ifdef CONFIG_FEATURE_HTTPD_DIR_LIST
	memcpy(&st, psb, sizeof(st));
#endif
	for(cur = config.htindex; cur; cur = cur->next) {
	    strcpy(cp + il, cur->after_colon);
	    if (lstat( cp + 1, psb ) == 0)
		break;
	}
#if DEBUG
	fprintf(stderr, "new cp=%s\n", cp);
#endif

	if(cur == NULL) {
#ifdef CONFIG_FEATURE_HTTPD_DIR_LIST
	    cp[il] = '\0';
	    memcpy(psb, &st, sizeof(st));
#else
# if DEBUG
	    fprintf(stderr, "cur = NULL for new cp\n");
# endif
	    return HTTP_NOT_FOUND;
#endif
	}
	il = strlen(cp);
    }
    if(!S_ISREG(psb->st_mode)) {
	/* this protected index.html as directory also */
#ifdef CONFIG_FEATURE_HTTPD_DIR_LIST
	if(!S_ISDIR(psb->st_mode))
#endif
	    return HTTP_FORBIDDEN;
    }
    if(strcmp(strrchr(url, '/') + 1, httpd_conf) == 0)
	return HTTP_FORBIDDEN;
    purl = uri->path_info;
    if(purl) {
	uri->path_info = strcpy(cp + il + 1, purl);
	il += strlen(purl) + 1;
    }
    if(q)
	uri->query = strcpy(cp + il + 1, q);
    else
	uri->query = q;
    return HTTP_OK;
}

#ifdef CONFIG_FEATURE_HTTPD_SSI

static void sendSsi(FILE *in);

static const char ssi_timefmt0[] = "%a %b %e %T %Z %Y";
static enum SSI_SIZE_FMT { BYTES, ABBREV } ssi_sizefmt;
static const char ssiheader[] = "HTTP/1.0 200 OK\r\n"
			       "Content-type: text/html\r\n"
			       "Pragma: no-cache\r\n"
			       "Cache-Control: no-cache\r\n"
			       "Expires: 0\r\n"
			       "\r\n";
static const char ssi_empty_val[1];
#define empty_val ((char *)ssi_empty_val)


static HttpResponseNum ssi_get_filename(const char *tag, char *val,
					struct stat *psb, URI *uri)
{
    /* Used for the various commands that accept a file name.
    ** These commands accept two tags:
    **   virtual
    **     Gives a virtual path to a document on the server.
    **   file
    **     Gives a pathname relative to the current directory. ../ cannot
    **     be used in this pathname, nor can absolute paths be used.
    */

  char *cp;
  char *v;
  HttpResponseNum htrn;

  if (strcmp( tag, "file" ) != 0 && strcmp( tag, "virtual" ) != 0) {
    ssi_directive_err(UNKNOWN_TAG, "%s", tag);
    return HTTP_BAD_REQUEST;
  }
  /*
   * make virtual url
   */
  if (tag[0] == 'f' || *val != '/') {
    if(*val == '/') {
	    /* deny unrelative file="/file" */
not_permitted:
	    ssi_directive_err(NOT_PERMITTED, "%s=\"%s\"", tag, val);
	    return HTTP_FORBIDDEN;
    }
    v = alloca(strlen(config.shtml_url) + strlen(val) + 2);
    if(v == NULL) {
internal_server_error:
	ssi_directive_err(INTERNAL_ERROR, "%m");
	return HTTP_INTERNAL_SERVER_ERROR;
    }
    /* relative path, find last dir and add value */
    cp = bb_basename(strcpy(v, config.shtml_url) + 1);
    strcpy(cp, val);
  } else {
     v = val;
  }

  htrn = parse_uri(v, uri, psb);
  if(htrn == HTTP_MOVED_TEMPORARILY) {
    cp = v = uri->url;
    htrn = parse_uri(v, uri, psb);
    free(cp);
  }
  switch(htrn) {
    case HTTP_OK:
#ifdef CONFIG_FEATURE_HTTPD_CGI
	if(strncmp(uri->url, "/cgi-bin/", 9) == 0)
	    return HTTP_MOVED_TEMPORARILY;
#endif
	return HTTP_OK;
    case HTTP_FORBIDDEN:
	goto not_permitted;
    case HTTP_NOT_FOUND:
#if DEBUG
	fprintf(stderr, "NOT_FOUND %s=%s\n", tag, val);
#endif
       ssi_directive_err(NOT_FOUND, "%s=%s", tag, val);
       return HTTP_NOT_FOUND;
    case HTTP_UNAUTHORIZED:
	ssi_directive_err(NOT_PERMITTED, "Unauthorized %s=%s", tag, val);
	return HTTP_UNAUTHORIZED;
    case HTTP_BAD_REQUEST:
	return HTTP_BAD_REQUEST;
    default:
	goto internal_server_error;
  }
}


static void ssi_show_size (bbst_size_t size)
{
  static const char size_suf[] = "\0KMG";
  const char *pss = size_suf;
  char str_size_suf[2];

  switch ( ssi_sizefmt )
     {
       case BYTES:
	    fprintf(config.s_out, "%lld", (long long) size );
	    break;
       case ABBREV:
	    while(size > 1024 && *pss != 'G') {
		size /= 1024;
		pss++;
	    }
	    str_size_suf[0] = *pss;
	    str_size_suf[1] = 0;
	    fprintf(config.s_out, "%ld%s",  (long)size, str_size_suf);
	    break;
     }
}

static char *ssi_get_val(char *parg)
{
	char *val;

	if(parg == NULL) {
	    /* tag without '=' */
	    ssi_directive_err(UNKNOWN_TAG, "must have '='");
	}
	val = parg;
	/* TODO: expand $VAR ${VAR} &lt; &gt; &amp; */
	return val;
}

static void ssi_config(char **argv)
{
    char *tag;
    /*
    ** SSI "config" directive, to controls output format.
    ** valid tags:
    **   - timefmt: strftime format string
    **   - sizefmt: bytes or abbrev
    */

    while((tag = *++argv)) {
	 char *val;

	 val = ssi_get_val(*++argv);
	 if (val == NULL)
	      break;

	 if (!strcmp(tag, "timefmt")) {
	      if(ssi_timefmt != ssi_timefmt0)
		free(ssi_timefmt);
	      ssi_timefmt = strdup( val );
	      form_env_conf(ssi_timefmt, NULL);
	 } else if (!strcmp(tag, "sizefmt")) {
	      if (!strcmp(val, "bytes")) {
		ssi_sizefmt = BYTES;
	      } else if (!strcmp(val, "abbrev")) {
		ssi_sizefmt = ABBREV;
	      } else
		 ssi_directive_err(UNKNOWN_VALUE, "%s=%s", tag, val);
	      form_env_conf(NULL, config.shtml_url + 1);
	 } else
	    ssi_directive_err(UNKNOWN_TAG, "%s", tag);
    }
}

typedef struct CHK_FILE_RECURSIVE_LOOPED {
	const char *file;
	struct CHK_FILE_RECURSIVE_LOOPED *next;
} chk_file_recursive_looped_t;

static void ssi_include(char** argv)
{
    char *fname;
    char *tag;

    /* Inserts the another document into the parsed document. */
    while((tag = *++argv)) {
	char *val;
	struct stat sb;
	URI uri;
	HttpResponseNum htrn;

	val = ssi_get_val(*++argv);
	if (val == NULL)
	    break;
	htrn = ssi_get_filename(tag, val, &sb, &uri);
	fname = uri.url;
#ifdef CONFIG_FEATURE_HTTPD_CGI
	if(htrn == HTTP_MOVED_TEMPORARILY) {
	    sendCgi_included(&uri);
	} else
#endif
	if (htrn == HTTP_OK) {
	      const char *mime;
#ifdef CONFIG_FEATURE_HTTPD_DIR_LIST
	      if(S_ISDIR(sb.st_mode)) {
		if(dir2html(fname, 1) != HTTP_OK)
		    ssi_directive_err(INTERNAL_ERROR, "%m");
		goto loop;
	      }
#endif
	      /*
	      ** Get the mime type of the requested file
	      ** if it is our magic, recursivly SSI read it
	      ** else, it should match "text/". We don't want
	      ** to send garbbage.
	      */
	      mime = find_mime_type(fname);
	      if (!strncmp(mime, "text/", 5)) {
		   int  fd_new = open(fname + 1, O_RDONLY );

		   if (fd_new > 0) {
			if (!strcmp(mime, "text/x-server-parsed-html")) {
			    static chk_file_recursive_looped_t *prev_chk_file_recursive;
			    chk_file_recursive_looped_t *cur;
			    chk_file_recursive_looped_t cur_save;
			    FILE *in;

			    /* recursive try as incluse */
			    for(cur = prev_chk_file_recursive; cur; cur = cur->next)
				if(strcmp(cur->file, fname) == 0)
				    break;
			    if(cur || strcmp(config.shtml_url, fname) == 0) {
				ssi_directive_err(RECURSIVE_INCLUDE, "\"%s\"", fname);
				goto loop;
			    }
			    /* save current file name */
			    cur = prev_chk_file_recursive;
			    cur_save.file = fname;
			    cur_save.next = cur;
			    prev_chk_file_recursive = &cur_save;

			    in = fdopen(fd_new, "r");
			    if(in) {
				const char *chg_ssi_timefmt;
				const char *prev_ssi_timefmt = ssi_timefmt;
				enum SSI_SIZE_FMT prev_ssi_sizefmt = ssi_sizefmt;
				const char *cur_shtml = config.shtml_url;

				config.shtml_url = uri.url;
				form_env_CGI_SSI(NULL, &uri);
				form_QUERY_STRING_UNESCAPED(uri.query);
				ssi_timefmt = (char *)ssi_timefmt0;
				ssi_sizefmt = BYTES;
				chg_ssi_timefmt =
				    prev_ssi_timefmt == ssi_timefmt0 ?
				    NULL : ssi_timefmt0;
				form_env_conf(chg_ssi_timefmt, uri.url + 1);

				sendSsi(in);

				config.shtml_url = cur_shtml;
				ssi_sizefmt = prev_ssi_sizefmt;
				if(ssi_timefmt != (char *)ssi_timefmt0)
					free(ssi_timefmt);
				ssi_timefmt = (char *)prev_ssi_timefmt;
				if(chg_ssi_timefmt)
				    chg_ssi_timefmt = ssi_timefmt;
				form_env_conf(chg_ssi_timefmt, config.shtml_url + 1);
			    }
			    prev_chk_file_recursive = cur;
			} else {
			    int count;
			    char* buf = config.buf;

			    while ((count = bb_full_read(fd_new, buf, MAX_MEMORY_BUFF)) > 0) {
				if (fwrite(buf, 1, count, config.s_out) != (size_t)count)
				    break;
			    }
			}
			close(fd_new);
		   }
	      }
loop:
	free(fname);
	}
    }
}


static void ssi_echo(char** argv)
{
    const char *cp;
    int i = 0;
    char *tag;

    /* Prints the value of one of the include variables.  Any dates are
    ** printed subject to the currently configured timefmt.  The only valid
    ** tag is var, whose value is the name of the variable you wish to echo.
    */

    while((tag = *++argv)) {
	char *val;

	if (i++) // insert a blank between echo'ed fields
	    putc(' ', config.s_out);

	val = ssi_get_val(*++argv);
	if (val == NULL)
	    break;

	if ( strcmp( tag, "var" ) ) {
	    ssi_directive_err(UNKNOWN_TAG, "%s=%s", tag, val);
	} else {
	    /* Try an environment variable. */
	    cp = getenv( val );
	    if ( cp == NULL )
		ssi_directive_err(UNKNOWN_VALUE, "%s", val);
	    else
		fputs(cp, config.s_out);
	}
    }
}

static void ssi_fstat(char **argv)
{
  struct stat sb;
  int i = 0;
  const char *directive = *argv;
  char *tag;

    /* Prints the size of the specified file. */
  while((tag = *++argv)) {
	 char *val;
	 URI uri;

	 if (i++) // insert a blank between echo'ed fields
	    putc(' ', config.s_out);
	 val = ssi_get_val(*++argv);
	 if (val == NULL)
	      break;

	 if ( ssi_get_filename(tag, val, &sb, &uri) == HTTP_OK ) {
		 if (!strcmp(directive, "fsize")) {
		    ssi_show_size(sb.st_size);
		 } else {
		    char timesbuf[TIMES_BUF_SIZE];
		    size_t p = ssi_show_time(sb.st_mtime, 0, timesbuf);

		    if(p > 0)
			fputs(timesbuf, config.s_out);
		 }
	 }
  }
}

#ifdef CONFIG_FEATURE_HTTPD_SSI_EXEC
static void
ssi_handle_sigalrm( int sig )
{
    config.alarm_signaled = sig;
}
#endif

static void ssi_exec(char **argv)
{
#ifndef CONFIG_FEATURE_HTTPD_SSI_EXEC
    ssi_directive_err(UNKNOWN_DIRECTIVE,
		       "#exec %s (feature not implemented - recompile)",
		       *argv);
#else
    char *tag;

    while((tag = *++argv)) {
	char *val;
	pid_t pid;

	val = ssi_get_val(*++argv);
	if (val == NULL)
	    break;

#ifdef CONFIG_FEATURE_HTTPD_CGI
	if ( !strcmp( tag, "cgi" ) ) {
	    struct stat sb;
	    URI uri;
	    HttpResponseNum htrn;

	    htrn = ssi_get_filename("virtual", val, &sb, &uri);
	    if(htrn == HTTP_OK)
		ssi_directive_err(NOT_PERMITTED, "cgi=%s", val);
	    else if(htrn == HTTP_MOVED_TEMPORARILY)
		sendCgi_included(&uri);
	    continue;
	}
#endif
	if ( strcmp( tag, "cmd" ) ) {
	    ssi_directive_err(UNKNOWN_TAG, "%s", tag);
	    continue;
	}

	bb_signal(1 | (1 << SIGALRM), ssi_handle_sigalrm);  /* no SA_RESTART */
	fflush(config.s_out);

	pid = fork();
	if(pid < 0) {
	    ssi_directive_err(INTERNAL_ERROR, "fork: %m");
	}
	if(pid == 0) {
#if DEBUG
	    fprintf(stderr,"about to exec: [%s]\n", val);
#endif
	    close(0);
	    open(bb_dev_null, O_RDONLY);
	    dup2(a_c_w, 1);
	    close(2);
	    open(bb_dev_null, O_WRONLY);
	    fsync(a_c_w);
	    execl(DEFAULT_SHELL, DEFAULT_SHELL_SHORT_NAME, "-c", val, NULL);
	    ssi_directive_err(NOT_PERMITTED, "exec %s: %m", val);
	    _exit(1);
	}
	if(pid > 0) {
	    (void) alarm( TIMEOUT );
	    while( wait(NULL) != pid)
		;
	    fsync(a_c_w);
	}
    }
#endif
}

static void ssi_set(char **argv)
{
    char *tag;

    while((tag = *++argv)) {
	 char *val;

	 val = ssi_get_val(*++argv);
	 if (val == NULL)
	      break;
	 addEnv(tag, "", val);
   }
}

/****************************************************************************
 *
 > $Function: sendSsi()
 *
 * $Description: perform Server-Side-Include parsing while sending the file.
 *
 *   It is done the hard way, copy it char by char to the output
 *   while running a state machine to look for ssi directive.
 *
 ****************************************************************************/

typedef void (*ssi_fun_t)(char **argv);

typedef struct SSI_DIRS_FUNC {
	const char *name;
	ssi_fun_t fun;
} ssi_dirs_func_t;

#if DEBUG
void debug_show_ssi_cmd(char **tags)
{
	char **argv;

	fflush(config.s_out);
	fprintf(stderr,"About to run");
	for(argv = tags; *argv; argv++) {
	    fprintf(stderr," %s", *argv);
	}
	fprintf(stderr,"\n");
}
#else
#define debug_show_ssi_cmd(tags)
#endif

static void sendSsi(FILE *in)
{
  enum SSI_FSM { NONE = 0, POUND, HSAD, GNAB } state = NONE;
  int  ch;
  char req[1024];
  char *ptr_tag = req;
  char *tags[512];    /* absolutely max is sizeof(req-"-->")/2 + 1 (NULL) */
  int   ntag = 0;     /* found tags, initializing for make compiler happy only */
  int   q = 0;        /* quote trigger, initializing for make compiler happy only */
  int   bkslash = 0;
  static const char html_rem_SSI[] = "<!--#";
  const char *phtml_rem = html_rem_SSI;
  static const ssi_dirs_func_t ssi_dirs_func[] = {
	{ "config",       ssi_config  },
	{ "include",      ssi_include },
	{ "echo",         ssi_echo    },
	{ "fsize",        ssi_fstat   },
	{ "flastmod",     ssi_fstat   },
	{ "exec",         ssi_exec    },
	{ "set",          ssi_set     },
	{ "break",        NULL        },
	{ NULL,           NULL }
  };

#if DEBUG
  fprintf(stderr, "SSI parser entered\n");
#endif

  while( (ch = getc(in)) != EOF ) {
	if (ptr_tag > &req[sizeof(req) - 2]) ptr_tag--;
	*ptr_tag++ = ch;
	switch ( state ) {
	     case NONE:
		 if ( ch == *phtml_rem++ ) {
		    if(*phtml_rem == 0) {
			phtml_rem = html_rem_SSI;
			ptr_tag = req;
			ntag = 0;
			bkslash = q = 0;
			state = POUND;
		    }
		 } else {
		    fwrite(req, (ptr_tag - req), 1, config.s_out);
		    ptr_tag = req;
		    phtml_rem = html_rem_SSI;
		 }
		 break;
	     case HSAD:
		 if ( ch == '-' ) {
		    state = GNAB;
		    break;
		 } else {
double_minus_in_req:
		    state = POUND;  /* and do not break */
		 }
	     case POUND:
		 if ( ch == '-' && q == 0) {
		    state = HSAD;
		    break;
		 }
		 if((ch == '"' || ch == '\'') && !bkslash) {
			if(q == 0) {
			    q = ch;
			} else if(q == ch) {
			    q = 0;
			}
		 }
		 bkslash = (ch == '\\');
		 if((ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n') && q == 0) {
			if((ptr_tag-1) == req) {
			    ptr_tag--;
			} else {
			    ptr_tag[-1] = ' ';
			}
		 }
		 break;
	     case GNAB:
		 if ( ch == '>' ) {
		    const ssi_dirs_func_t *p;
		    char *s;

		    ptr_tag[-3] = '\0';
		    /* process directive */
		    tags[ntag++] = req;
		    s = req;
		    while(*s && *s != ' ')
			s++;
		    if(*s)
			*s++ = 0;
		    /* process tags */
		    while(1) {
			char *sp;

			while(*s == ' ')
			    s++;
			if(*s == 0)
			    break;
			/* save start new tag */
			tags[ntag++] = s;

			while(*s && *s != ' ' && *s != '=')
			    s++;
			sp = s;             /* sp is tag[<space>][=]... */
			while(*s == ' ')
			    s++;
			if(*s != '=') {
			    /* hmm, tag without '=' */
			    break;
			}
			*sp = 0;
			while(*++s == ' ')
			    ;
			/* set terminator */
			ch = *s;
			if(ch == '"' || ch == '\'') {
			    s++;
			} else {
			    ch = ' ';
			}
			/* save start val */
			tags[ntag++] = sp = s;
			/* find \0 or val\0 or val<space> or "val" or 'val' */
			while(*s) {
			    if(*s == ch) {
				s++;
				break;
			    }
			    if(*s == '\\') {
				/* if escaped a terminator? */
				if(s[1] == ch)
				    s++;
			    }
			    *sp++ = *s++;
			}
			*sp = 0;            /* is [val] now */
		    }
		    tags[ntag] = NULL;
		    p = find_string_from_struct(ssi_dirs_func, sizeof(ssi_dirs_func_t), req);
		    if(p->name == NULL) {
			ssi_directive_err(UNKNOWN_DIRECTIVE, "%s", req);
		    } else {
			/* execute ssi directive */
			debug_show_ssi_cmd(tags);
			/* check "break" directive */
			if(p->fun == NULL)
			    goto lbreak;
			(*p->fun)(tags);
		    }
		    ptr_tag = req;
		    state = NONE;
		 } else if(ch != '-') {
			goto double_minus_in_req;
		 }
		 break;
       }
  }
lbreak:
  fclose(in);
}

#endif            /* CONFIG_FEATURE_HTTPD_SSI */

static void
handle_sigalrm( int sig )
{
    sendHeaders(HTTP_REQUEST_TIMEOUT);
    config.alarm_signaled = sig;
}

/****************************************************************************
 *
 > $Function: handleIncoming()
 *
 * $Description: Handle an incoming http request.
 *
 ****************************************************************************/
static void handleIncoming(void)
{
  char *buf = config.buf;
  URI uri;
  char *url;
  char *purl;
  int  blank = -1;
  struct stat sb;
#ifdef CONFIG_FEATURE_HTTPD_CGI
  const char *prequest = request_GET;
  long length = 0;
#endif

  fd_set s_fd;
  struct timeval tv;

  int f;
  HttpResponseNum httpresponsenum;

  bb_signal(1 | (1 << SIGALRM), handle_sigalrm);    /* no SA_RESTART */

  do {
    (void) alarm( TIMEOUT );
    if (getLine() <= 0)
	break;  /* closed */

    purl = strpbrk(buf, bb_spacestok);
    if(purl == NULL) {
BAD_REQUEST:
      sendHeaders(HTTP_BAD_REQUEST);
      break;
    }
    *purl = 0;
#ifdef CONFIG_FEATURE_HTTPD_CGI
    if(strcasecmp(buf, prequest) != 0) {
	prequest = "POST";
	if(strcasecmp(buf, prequest) != 0) {
	    sendHeaders(HTTP_NOT_IMPLEMENTED);
	    break;
	}
    }
#else
    if(strcasecmp(buf, request_GET) != 0) {
	sendHeaders(HTTP_NOT_IMPLEMENTED);
	break;
    }
#endif
    *purl = ' ';

    if (sscanf(purl, " %[^ ] HTTP/%d.%*d", buf, &blank) < 1 || buf[0] != '/') {
      /* Garbled request/URL */
      goto BAD_REQUEST;
    }
    url = alloca(strlen(buf));
    if(url == NULL) {
	sendHeaders(HTTP_INTERNAL_SERVER_ERROR);
	break;
    }
    strcpy(url, buf);


    // read until blank line for HTTP version specified, else parse immediate
    if(blank >= 0) {
     while(1) {
	alarm(TIMEOUT);
	if(getLine() <= 0)
	  break;

#if DEBUG
      fprintf(stderr, "Header: '%s'\n", buf);
#endif

#ifdef CONFIG_FEATURE_HTTPD_CGI
      /* try and do our best to parse more lines */
      if ((strncasecmp(buf, Content_length, 15) == 0)) {
	if(prequest != request_GET)
		length = strtol(buf + 15, 0, 0); // extra read only for POST
      } else
#endif
# if defined(CONFIG_FEATURE_HTTPD_CGI) || defined(CONFIG_FEATURE_HTTPD_SSI)
      if ((strncasecmp(buf, "Content-Type:", 13) == 0)) {
		config.content_type = strdup(bb_skip_whitespace(buf + 13));
      } else if ((strncasecmp(buf, "Cookie:", 7) == 0)) {
		config.cookie = strdup(bb_skip_whitespace(buf + 7));
      } else if ((strncasecmp(buf, "Referer:", 8) == 0)) {
		config.referer = strdup(bb_skip_whitespace(buf + 8));
      } else if ((strncasecmp(buf, "User-Agent:", 11) == 0)) {
		config.user_agent = strdup(bb_skip_whitespace(buf + 11));
      } else
#endif

#ifdef CONFIG_FEATURE_HTTPD_BASIC_AUTH
      if (strncasecmp(buf, "Authorization:", 14) == 0) {
	/* We only allow Basic credentials.
	 * It shows up as "Authorization: Basic <userid:password>" where
	 * the userid:password is base64 encoded.
	 */
	char *test = bb_skip_whitespace(buf + 14);
	if (strncasecmp(test, "Basic", 5) != 0)
		continue;

	test += 5;  /* decodeBase64() skiping space self */
	decodeBase64(test);
	if(test[0] != 0 && test[0] != ':' && test[1] != 0) {
	    if(config.usr_pass != empty_auth)
		free(config.usr_pass);
	    config.usr_pass = strdup(test);
	}
      }
#endif          /* CONFIG_FEATURE_HTTPD_BASIC_AUTH */
    }   /* while extra header reading */
   }
    (void) alarm( 0 );
    if(config.alarm_signaled)
	break;

    httpresponsenum = parse_uri(url, &uri, &sb);
    if(httpresponsenum != HTTP_OK) {
	sendHeaders(httpresponsenum);
	break;
    }

    url = uri.url;

#if DEBUG
    fprintf(stderr, "url='%s', query=%s path_info=%s\n", url, uri.query, uri.path_info);
#endif

#ifdef CONFIG_FEATURE_HTTPD_CGI
    /* if strange Content-Length */
    if (length < 0)
	break;

    if (strncmp(url, "/cgi-bin/", 9) == 0) {
	sendCgi(&uri, prequest, length);
	break;
    }
#endif  /* CONFIG_FEATURE_HTTPD_CGI */

#ifdef CONFIG_FEATURE_HTTPD_DIR_LIST
    if(S_ISDIR(sb.st_mode)) {
	config.httpd_found.found_mime_type = "text/html";
	goto call_dirlist;
    }
#endif

#ifdef CONFIG_FEATURE_HTTPD_SSI
    if (!strcmp(find_mime_type(url), "text/x-server-parsed-html")) {
	FILE *in, *out;

	in = fopen(url + 1, "r");
	if(in == NULL) {
	    sendHeaders(HTTP_NOT_FOUND);
	    break;
	}
	out = fdopen(a_c_w, "w");
	if(out == NULL) {
	    sendHeaders(HTTP_INTERNAL_SERVER_ERROR);
	    break;
	}
	config.s_out = out;
	config.shtml_url = uri.url;
	/* Send a header with mime type browsers expect. */
	fputs(ssiheader, out);

	form_env_CGI_SSI(request_GET, &uri);
	addEnv("DOCUMENT",     "NAME",      uri.url + 1);
	if(uri.path_info) {
	    purl = alloca(strlen(uri.url) + strlen(uri.path_info) + 2);
	    sprintf(purl, "/%s", uri.path_info);
	    addEnv("DOCUMENT",     "PATH_INFO", purl);
	    sprintf(purl, "%s/%s", uri.url, uri.path_info);
	} else {
	    purl = uri.url;
	    addEnv("DOCUMENT",     "PATH_INFO", uri.path_info);
	}
	addEnv("DOCUMENT", "URI", purl);

	form_QUERY_STRING_UNESCAPED(uri.query);
	/* Setup SSI config to the defined state */
	ssi_timefmt = (char *)ssi_timefmt0;
	ssi_sizefmt = BYTES;
	form_env_conf(ssi_timefmt0, url + 1);
	/* Run the SSI processor */
	sendSsi(in);
	fclose(out);
	break;
    }
#else
    find_mime_type(url);
#endif

#ifdef CONFIG_FEATURE_HTTPD_DIR_LIST
call_dirlist:
#endif
    if(uri.path_info && uri.found_e404 == 0) {
	sendHeaders(HTTP_NOT_FOUND);
	break;
    }
    config.ContentLength = sb.st_size;
    config.last_mod = sb.st_mtime;
#if DEBUG
    fprintf(stderr, "Sending file '%s' Content-type: %s\n",
				url + 1, config.httpd_found.found_mime_type);
#endif

#ifdef CONFIG_FEATURE_HTTPD_DIR_LIST
    if(S_ISDIR(sb.st_mode)) {
	if(dir2html(url
#ifdef CONFIG_FEATURE_HTTPD_SSI
			    , 0
#endif
			    ) != HTTP_OK)
	    sendHeaders(HTTP_INTERNAL_SERVER_ERROR);
    } else {
#endif

	f = open(url + 1, O_RDONLY);
	if (f >= 0) {
	    ssize_t count;

	    sendHeaders(uri.found_e404 == 0 ? HTTP_OK : HTTP_SEND_E404);
	    while ((count = bb_full_read(f, buf, MAX_MEMORY_BUFF)) > 0) {
		if (bb_full_write(a_c_w, buf, count) != count)
		    break;
	    }
	    close(f);
	} else {
#if DEBUG
	    bb_perror_msg("Unable to open '%s'", url);
#endif
	    sendHeaders(HTTP_NOT_FOUND);
	}

#ifdef CONFIG_FEATURE_HTTPD_DIR_LIST
    }
#endif

#ifndef CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY
    /* unset if non inetd looped */
    config.ContentLength = -1;
#endif

  } while (0);


#ifndef CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY
/* from inetd don`t looping: freeing, closing automatic from exit always */
# if DEBUG
  fprintf(stderr, "closing socket\n");
#  if defined(CONFIG_FEATURE_HTTPD_BASIC_AUTH)
  if(config.usr_pass != empty_auth) {
	free(config.usr_pass);
	config.usr_pass = (char *)empty_auth;
  }
#  endif
# endif
# if defined(CONFIG_FEATURE_HTTPD_CGI) || defined(CONFIG_FEATURE_HTTPD_SSI)
#  if DEBUG
  free(config.cookie);
  config.cookie = 0;
  free(config.referer);
  config.referer = 0;
  free(config.content_type);
  config.content_type = 0;
#  endif
# endif
#endif  /* CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY */
  shutdown(a_c_w, SHUT_WR);

  /* Properly wait for remote to closed */
  FD_ZERO (&s_fd);
  FD_SET (a_c_r, &s_fd);

  do {
    tv.tv_sec = 2;
    tv.tv_usec = 0;
    f = select (a_c_r + 1, &s_fd, NULL, NULL, &tv);
  } while (f > 0 && (read (a_c_r, buf, sizeof (config.buf)) > 0));

  shutdown(a_c_r, SHUT_RD);
#ifndef CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY
  close(config.accepted_socket);
#endif  /* CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY */
}

/****************************************************************************
 *
 > $Function: miniHttpd()
 *
 * $Description: The main http server function.
 *
 *   Given an open socket fildes, listen for new connections and farm out
 *   the processing as a forked process.
 *
 * $Parameters:
 *      (int) server. . . The server socket fildes.
 *
 * $Return: (int) . . . . Always 0.
 *
 ****************************************************************************/
#ifndef CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY
static void miniHttpd(int server) ATTRIBUTE_NORETURN;

static void miniHttpd(int server)
{
  fd_set readfd, portfd;

  FD_ZERO(&portfd);
  FD_SET(server, &portfd);

  /* copy the ports we are watching to the readfd set */
  while (1) {
    readfd = portfd;

    /* Now wait INDEFINITELY on the set of sockets! */
    if (select(server + 1, &readfd, 0, 0, 0) > 0) {
      if (FD_ISSET(server, &readfd)) {
	struct sockaddr_in fromAddr;

	socklen_t fromAddrLen = sizeof(fromAddr);
	int s = accept(server,
		       (struct sockaddr *)&fromAddr, &fromAddrLen);

	if (s < 0) {
	    continue;
	}
	config.accepted_socket = s;
	config.rmt_ip = ntohl(fromAddr.sin_addr.s_addr);
#if defined(CONFIG_FEATURE_HTTPD_CGI) || defined(CONFIG_FEATURE_HTTPD_SSI) || DEBUG
	sprintf(config.rmt_ip_str, "%u.%u.%u.%u",
		(unsigned char)(config.rmt_ip >> 24),
		(unsigned char)(config.rmt_ip >> 16),
		(unsigned char)(config.rmt_ip >> 8),
				config.rmt_ip & 0xff);
	config.port = ntohs(fromAddr.sin_port);
#if DEBUG
	bb_error_msg("connection from IP=%s, port %u\n",
					config.rmt_ip_str, config.port);
#endif
#endif /* CONFIG_FEATURE_HTTPD_CGI */

	/*  set the KEEPALIVE option to cull dead connections */
	setsockopt_1i(s, SOL_SOCKET, SO_KEEPALIVE, 1);

#if !DEBUG
	if(fork() == 0)
#endif
	{
	    /* This is the spawned thread */
#ifdef CONFIG_FEATURE_HTTPD_RELOAD_CONFIG_SIGHUP
	    /* protect reload config, may be confuse checking */
	    bb_signal((1 << SIGHUP), SIG_IGN);
#endif
	    handleIncoming();
	    if(!DEBUG)
		exit(0);
	}
	close(s);
      }
    }
  } // while (1)
}

#else
    /* from inetd */

static void miniHttpd(void)
{
  struct sockaddr_in fromAddrLen;
  socklen_t sinlen = sizeof (struct sockaddr_in);

  getpeername (0, (struct sockaddr *)&fromAddrLen, &sinlen);
  config.rmt_ip = ntohl(fromAddrLen.sin_addr.s_addr);
#if defined(CONFIG_FEATURE_HTTPD_CGI) || defined(CONFIG_FEATURE_HTTPD_SSI) || DEBUG
  sprintf(config.rmt_ip_str, "%u.%u.%u.%u",
		(unsigned char)(config.rmt_ip >> 24),
		(unsigned char)(config.rmt_ip >> 16),
		(unsigned char)(config.rmt_ip >> 8),
				config.rmt_ip & 0xff);
#endif
  config.port = ntohs(fromAddrLen.sin_port);
  handleIncoming();
}
#endif  /* CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY */

#ifdef CONFIG_FEATURE_HTTPD_RELOAD_CONFIG_SIGHUP
static void sighup_handler(int sig)
{
	/* set and reset */
	parse_conf(default_path_httpd_conf,
		    sig == SIGHUP ? SIGNALED_PARSE : FIRST_PARSE);
	bb_signal((1 << SIGHUP), sighup_handler);
}
#endif

#include "opt_gen.h"
OPT_GEN(
opt(c, OPT_CONFIG_FILE, arg),
opt(d, OPT_DECODE_URL, arg),
opt(h, OPT_HOME_HTTPD, arg),
opt_opt(ENABLE_FEATURE_HTTPD_ENCODE_URL_STR, e, OPT_ENCODE_URL, arg),
opt_opt(ENABLE_FEATURE_HTTPD_BASIC_AUTH, r, OPT_REALM, arg),
opt_opt(ENABLE_FEATURE_HTTPD_AUTH_MD5, m, OPT_MD5, arg),
opt_opt(NOT(ENABLE_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY), p, OPT_PORT, arg),
opt_opt(ENABLE_FEATURE_HTTPD_SETUID, u, OPT_SETUID, arg),
opt_opt(ENABLE_FEATURE_HTTPD_BIND, b, OPT_BIND, arg),
opt_opt(NOT(ENABLE_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY), P, OPT_PID, arg),
opt_opt(NOT(ENABLE_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY), f, OPT_FOREGROUND)
)
#include "opt_gen.h"

#ifdef HTTPD_STANDALONE
/* require from libbb.a for linking */
void bb_show_usage(void)
{
  fprintf(stderr, "Usage: %s [-p <port>] [-c configFile] [-d/-e <string>] "
		  "[-r realm] [-u user] [-h homedir] [-f] [-P pidfile]\n",
		  bb_applet_name);
  exit(1);
}

int main(int argc, char *argv[])
#else
int httpd_main(void)
#endif
{
  unsigned long opt;
  const char *home_httpd = home;
  char *url_for_decode;
  USE_FEATURE_HTTPD_ENCODE_URL_STR(const char *url_for_encode;)

  SKIP_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY(int server;)
  SKIP_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY(const char *s_port;)
  USE_FEATURE_HTTPD_BIND(const char *s_bind;)
  SKIP_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY(const char *s_pid = NULL;)

  USE_FEATURE_HTTPD_SETUID(const char *s_uid;)
  USE_FEATURE_HTTPD_SETUID(long uid = -1;)

  USE_FEATURE_HTTPD_AUTH_MD5(const char *pass;)

  USE_FEATURE_HTTPD_BASIC_AUTH(config.realm = "Web Server Authentication";)
  USE_FEATURE_HTTPD_BASIC_AUTH(config.usr_pass = (char *)empty_auth;)

  SKIP_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY(config.port = 80;)

  config.ContentLength = -1;

#ifdef HTTPD_STANDALONE
  libbb_init(argc, argv);
#endif

  opt = bb_getopt_ulflags(opt_str,
	&(config.configFile), &url_for_decode, &home_httpd
	USE_FEATURE_HTTPD_ENCODE_URL_STR(, &url_for_encode)
	USE_FEATURE_HTTPD_BASIC_AUTH(, &(config.realm))
	USE_FEATURE_HTTPD_AUTH_MD5(, &pass)
	SKIP_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY(, &s_port)
	USE_FEATURE_HTTPD_SETUID(, &s_uid)
	USE_FEATURE_HTTPD_BIND(, &s_bind)
	SKIP_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY(, &s_pid)
    );

  if(opt & OPT_DECODE_URL) {
      printf("%s", decodeString(url_for_decode, 1));
      return 0;
  }
#ifdef CONFIG_FEATURE_HTTPD_ENCODE_URL_STR
  if(opt & OPT_ENCODE_URL) {
      printf("%s", encodeString(url_for_encode));
      return 0;
  }
#endif
#ifdef CONFIG_FEATURE_HTTPD_AUTH_MD5
  if(opt & OPT_MD5) {
      char enc[ENCRYPT_BUFFER_SIZE];

      printf("%s\n", pw_encrypt(pass, "$1$", enc));
      return 0;
  }
#endif
#ifndef CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY
  if(opt & OPT_PORT)
	config.port = bb_xparse_ui_number(10 | BB_PN_SET_LOWER1 | BB_PN_SET_UPPER, s_port, 0xffff);
#ifdef CONFIG_FEATURE_HTTPD_SETUID
  if(opt & OPT_SETUID) {
	uid = bb_xparse_i_number(BB_PN_SILENT_RE, s_uid);
	if(errno) {
		/* not integer */
		uid = bb_xgetpwnam(s_uid);
	}
  }
#endif
# ifdef CONFIG_FEATURE_HTTPD_BIND
  {
    struct sockaddr_in usa;

    (void) memset(&usa, 0, sizeof(usa));
    usa.sin_family = AF_INET;
    if(opt & OPT_BIND) {
	    bb_gethostbyname(BB_GHBN_IN_ADDR | BB_GHBN_ERR_DIE, s_bind, &usa.sin_addr);
    } else {
	    usa.sin_addr.s_addr = htonl(INADDR_ANY);
    }
    usa.sin_port = htons(config.port);
    server = openServer(&usa);
  }
# else
  server = openServer();
# endif
#endif /* CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY */
  xchdir(home_httpd);
#if defined(CONFIG_FEATURE_HTTPD_CGI) || defined(CONFIG_FEATURE_HTTPD_SSI)
   {
	char *p = xstrdup(getenv("PATH"));      /* if PATH NULL, returns NULL */
	char realpath_buff[MAXPATHLEN];

	clearenv();
	if(p)
		bb_setenv1("PATH", p);
	if(realpath(dot, realpath_buff) == NULL)
	    bb_perror_msg_and_die("realpath of %s", home_httpd);
	config.document_root = xstrdup(realpath_buff);

# ifndef CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY
	addEnvPort("SERVER");
# endif
   }
#endif

  USE_FEATURE_HTTPD_RELOAD_CONFIG_SIGHUP(sighup_handler(0);)
  SKIP_FEATURE_HTTPD_RELOAD_CONFIG_SIGHUP(parse_conf(default_path_httpd_conf, FIRST_PARSE);)

#ifndef CONFIG_FEATURE_HTTPD_USAGE_FROM_INETD_ONLY
# if !DEBUG
  xdaemon(XDAEMON_NO_CHDIR | XDAEMON_PF_ARG | XDAEMON_PIDFILE, s_pid);
# endif
# ifdef CONFIG_FEATURE_HTTPD_SETUID
  /* drop privileges after bind and create pidfile */
  if(uid > 0)
	setuid(uid);
# endif
  miniHttpd(server);
#else
  miniHttpd();
  return 0;
#endif
}
#endif /* APPLET_USAGE */
