/* micro_inetd - simple network service spawner
**
** Copyright (C)1996,2000 by Jef Poskanzer <jef@acme.com>. All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
** OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
*/

/*
 * Busybox style by Vladimir Oleynik (C) 2003, 2004 <dzo@simtreas.ru>
*/

#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <memory.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <syslog.h>
#include <sys/ioctl.h>
#include "bb_inet.h"

static void child_handler (int sig)
{
  /* Set up the signal handler again.  Don't need to do this on BSD
     ** systems, but it doesn't hurt.
   */
  (void) signal (SIGCHLD, child_handler);
  /* Reap defunct children until there aren't any more. */
  waitpid_no_int((pid_t) - 1, NULL, WNOHANG, 0);
}

int inetd_main (void)
{
  char **child_argv;
  int listen_fd;
  int on;
  char *user;
  char *sbind;
  long uid = -1, gid = -1;
  struct sockaddr_in sa;
  short port;

  on = bb_xgetopt_ulflags("+b:u:\0-2", &sbind, &user);
#define foreground (on & 4)

  if(on & 2) {
	char *group = strchr(user, '.');

	if(group != NULL) {
		size_t sz = (group - user);
		user = xstrdup(user);
		user[sz] = 0;
		if(*++group == '\0')
			group = NULL;
	}
	uid = get_ug_id(user, bb_xgetpwnam);
	if(group != NULL)
		gid = get_ug_id(group, bb_xgetgrnam);
  }

  port = bb_xparse_i_number(BB_PN_SET_LOWER1 | BB_PN_SET_UPPER | BB_PN_ERR_STR1,
				Gargv[0], 65535, "strange port number %s");
  if(on & 1) {
	  bb_gethostbyname(BB_GHBN_SOCKADDR_IN | BB_GHBN_ERR_DIE, sbind, &sa);
  } else {
	  memset (&sa, 0, sizeof (sa));
	  sa.sin_family = AF_INET;
	  sa.sin_addr.s_addr = htonl( INADDR_ANY );
  }

  /* Create socket. */
  listen_fd = bb_xsocket(BB_XSOCKET_FLG_ERR_DIE);

  /* Allow reuse of local addresses. */
  setsockopt_1i(listen_fd, SOL_SOCKET, SO_REUSEADDR, 1);

  /* Bind to it. */
  sa.sin_port = htons(port);
  bb_xbind(listen_fd, (struct sockaddr *) &sa, sizeof (sa));

  /* Start a listen going. */
  if (listen(listen_fd, 1024) < 0)
	  bb_perror_msg_and_die("listen");

  child_argv = Gargv + 1;
  bb_openlog(LOG_PERROR | LOG_PID, LOG_DAEMON);

  /* drop privilegies */
  if(gid > 0)
	setgid(gid);
  if(uid > 0)
	setuid(uid);

  xdaemon(0);

  /* Set up a signal handler for child reaping. */
  (void) signal (SIGCHLD, child_handler);

  for (;;) {
	/* Accept a new connection. */
	int conn_fd = accept(listen_fd, NULL, NULL);
	if (conn_fd < 0) {
	  if (errno == EINTR)           /* because of SIGCHLD (or ptrace) */
		continue;
	  bb_perror_msg_and_die("accept");
	}

	/* Fork a sub-process. */
	if (vfork () == 0) {
	  (void) close(listen_fd);
	  /* Dup the connection onto the standard descriptors. */
	  (void) dup2(conn_fd, 0);
	  (void) dup2(conn_fd, 1);
	  (void) dup2(conn_fd, 2);
	  (void) close(conn_fd);
	  /* Run the program. */
	  (void) execvp(child_argv[0], child_argv);
	  /* Something went wrong. */
	  bb_perror_msg_and_die("execl");
	}
	/* Parent process. */
	(void) close(conn_fd);
  }
}
