/*
 * nameif.c - Naming Interfaces based on MAC address for busybox.
 *
 * Written 2000 by Andi Kleen.
 * Busybox port 2002 by Nick Fedchik <nick@fedchik.org.ua>
 *                      Glenn McGrath <bug1@iinet.net.au>
 * Optimization by Vladimir Oleynik (C) 2016 <dzo@simtreas.ru>
 *
 * Licensed under the GPL v2 or later, see the file LICENSE in this tarball.
 */
#ifdef APPLET_USAGE
#define nameif_trivial_usage \
	"[-s] [-c FILE] [{IFNAME MACADDR}]"
#define nameif_full_usage \
	"Nameif renaming network interface while it in the down state.\n\n" \
	"Options:\n" \
	"\t-c FILE\t\tUse configuration file (default is /etc/mactab)\n" \
	"\t-s\t\tUse syslog (LOCAL0 facility)\n" \
	"\tIFNAME MACADDR\tnew_interface_name interface_mac_address"
#else

#include "busybox.h"
#include <syslog.h>
#include <sys/ioctl.h>
#include <string.h>
#include <unistd.h>
#include <netinet/ether.h>

#include "config_read.h"
#include "bb_inet.h"

#ifndef HAVE_ETHER_ATON
extern struct ether_addr *ether_aton(const char *asc);
#endif

/* Older versions of net/if.h do not appear to define IF_NAMESIZE. */
#ifndef IF_NAMESIZE
#  ifdef IFNAMSIZ
#    define IF_NAMESIZE IFNAMSIZ
#  else
#    define IF_NAMESIZE 16
#  endif
#endif

/* take from linux/sockios.h */
#define SIOCSIFNAME     0x8923  /* set interface name */

#ifndef ifr_newname
#define ifr_newname ifr_ifru.ifru_slave
#endif

typedef struct mactable_s {
	struct mactable_s *next;
	const char *ifname;
	struct ether_addr mac;
} mactable_t;

CALC_BSS struct BB_COMMON_BSS {
	mactable_t *clist;
	int nclist;
} bb_common_bss;

/* Check ascii str_macaddr, convert and setup mactable_t list */
static void cc_macaddr(const char *ifname, const char *str_macaddr)
{
	struct ether_addr *lmac;
	mactable_t *ch;
	size_t l = strlen(ifname) + 1;

	if (l >= IF_NAMESIZE)
		bb_error_msg_and_die("interface name `%s' too long", ifname);
	ch = xmalloc(*Gargv == NULL ? sizeof(mactable_t) + l : sizeof(mactable_t));
	ch->ifname = *Gargv == NULL ? strcpy((char *)(ch + 1), ifname) : ifname;
	lmac = ether_aton(str_macaddr);
	if (lmac == NULL)
		bb_error_msg_and_die("cannot parse MAC %s", str_macaddr);
	memcpy(&ch->mac, lmac, sizeof(struct ether_addr));
	ch->next = bb_common_bss.clist;
	bb_common_bss.clist = ch;
	bb_common_bss.nclist++;
}

int nameif_main(void)
{
	const char *fname = "/etc/mactab";
	int ctl_sk;
	int if_index = 1;

	ctl_sk = bb_getopt_ulflags("sc:", &fname);

	if (Gargc & 1)
		bb_show_usage();
	if(ctl_sk & 1) {
		bb_openlog(0, LOG_LOCAL0);
		log_syslog = LOG_ERR;
		log_stderr = 0;
	}

	if (*Gargv != NULL) {
		char **a = Gargv;

		while (*a) {
			cc_macaddr(a[0], a[1]);
			a += 2;
		}
	} else {
		config_read_t *crt = config_read_prepare(fname, 'e');

		while( (crt = config_read(crt, "#\1\2\0\2")) != NULL)
			cc_macaddr(crt->tokens[0], crt->tokens[1]);
	}

	ctl_sk = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM | BB_XSOCKET_FLG_ERR_DIE);

	while (bb_common_bss.nclist) {
		mactable_t *ch;
		struct ifreq ifr;

		memset(&ifr, 0, sizeof(struct ifreq));
		if_index++;
		ifr.ifr_ifindex = if_index;

		/* Get ifname by index or die */
		if (ioctl(ctl_sk, SIOCGIFNAME, &ifr))
			break;

		/* Has this device hwaddr? */
		if (ioctl(ctl_sk, SIOCGIFHWADDR, &ifr))
			continue;

		/* Search for mac like in ifr.ifr_hwaddr.sa_data */
		for (ch = bb_common_bss.clist; ch; ch = ch->next) {
			if (ch->ifname != NULL &&
			    !memcmp(&ch->mac, ifr.ifr_hwaddr.sa_data, sizeof(struct ether_addr))) {
				strcpy(ifr.ifr_newname, ch->ifname);
				xioctl(ctl_sk, SIOCSIFNAME, &ifr, "\1cannot change ifname %s to %s",
						ifr.ifr_name, ch->ifname);

				/* set flag of already renamed interface */
				ch->ifname = NULL;
				bb_common_bss.nclist--;
				break;
			}
		}
	}

	return 0;
}
#endif /* APPLET_USAGE */
