/*
 * Mini netstat implementation(s) for busybox
 * based in part on the netstat implementation from net-tools.
 *
 * Copyright (C) 2002 by Bart Visscher <magick@linux-fan.com>
 *
 * 2002-04-20
 * IPV6 support added by Bart Visscher <magick@linux-fan.com>
 *
 * Modified by Vladimir Oleynik <dzo@simtreas.ru> (C) 2008
 * - size optimization
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */
#ifdef APPLET_USAGE
#define netstat_trivial_usage \
	"[-laenrtuwx]"
#define netstat_full_usage \
	"Netstat displays Linux networking information.\n\n" \
	"Options:\n" \
	"\t-l display listening server sockets\n" \
	"\t-a display all sockets (default: connected)\n" \
	"\t-e display other/more information\n" \
	"\t-n don't resolve names\n" \
	"\t-r display routing table\n" \
	"\t-t tcp sockets\n" \
	"\t-u udp sockets\n" \
	"\t-w raw sockets\n" \
	"\t-x unix sockets"
#else

#include "busybox.h"
#include <string.h>
#include <dirent.h>
#include <unistd.h>
#include "bb_inet.h"

#ifdef CONFIG_ROUTE
extern void displayroutes(int noresolve, int netstatfmt);
#endif

#define NETSTAT_CONNECTED       0x01
#define NETSTAT_LISTENING       0x02
#define NETSTAT_NUMERIC         0x04
#define NETSTAT_TCP             0x10
#define NETSTAT_UDP             0x20
#define NETSTAT_RAW             0x40
#define NETSTAT_UNIX            0x80


#define PROGNAME_WIDTHs PROGNAME_WIDTH1(PROGNAME_WIDTH)
#define PROGNAME_WIDTH1(s) PROGNAME_WIDTH2(s)
#define PROGNAME_WIDTH2(s) #s

#define PRG_HASH_SIZE 211

enum {
    TCP_ESTABLISHED = 1,
    TCP_SYN_SENT,
    TCP_SYN_RECV,
    TCP_FIN_WAIT1,
    TCP_FIN_WAIT2,
    TCP_TIME_WAIT,
    TCP_CLOSE,
    TCP_CLOSE_WAIT,
    TCP_LAST_ACK,
    TCP_LISTEN,
    TCP_CLOSING                 /* now a valid state */
};

static const char * const tcp_state[] =
{
    "",
    "ESTABLISHED",
    "SYN_SENT",
    "SYN_RECV",
    "FIN_WAIT1",
    "FIN_WAIT2",
    "TIME_WAIT",
    "CLOSE",
    "CLOSE_WAIT",
    "LAST_ACK",
    "LISTEN",
    "CLOSING"
};

typedef enum {
    SS_FREE = 0,                /* not allocated                */
    SS_UNCONNECTED,             /* unconnected to any socket    */
    SS_CONNECTING,              /* in process of connecting     */
    SS_CONNECTED,               /* connected to socket          */
    SS_DISCONNECTING            /* in process of disconnecting  */
} socket_state;

#define SO_ACCEPTCON    (1<<16) /* performed a listen           */
#define SO_WAITDATA     (1<<17) /* wait data to read            */
#define SO_NOSPACE      (1<<18) /* no space to write            */

CALC_BSS struct BB_COMMON_BSS {
	/* 21 digits plus null terminator, good for 64-bit or smaller ints */
	char itoa_local[22];
	int has;
	char buffer[8192];
	int opt_flags;
} bb_common_bss;
#define itoa_local bb_common_bss.itoa_local
#define has bb_common_bss.has
#define buffer bb_common_bss.buffer
#define opt_flags bb_common_bss.opt_flags


static char *itoa(unsigned int i)
{
	/* 21 digits plus null terminator, good for 64-bit or smaller ints */
	char *p = &itoa_local[21];
	*p-- = '\0';
	do {
		*p-- = '0' + i % 10;
		i /= 10;
	} while (i > 0);
	return p + 1;
}

#define SZ 64

static void snprint_ip_port(char *ip_port, struct sockaddr *addr, int port, char *proto)
{
	char *port_name;
	char ip_addr[SZ];
	int numeric = opt_flags & NETSTAT_NUMERIC;

#ifdef CONFIG_FEATURE_IPV6
	if (addr->sa_family == AF_INET6) {
		INET6_rresolve(ip_addr, SZ, (struct sockaddr_in6 *)addr,
					   numeric ? 0x0fff : 0);
	} else
#endif
	{
		INET_rresolve(ip_addr, SZ, (struct sockaddr_in *)addr,
					0x4000 | (numeric ? 0x0fff : 0),
					0xffffffff);
	}
	if (!port)
		port_name = "*";
	else {
		port_name = NULL;
		if (numeric == 0) {
			struct servent *se = getservbyport(htons(port), proto);
			if (se)
				port_name = se->s_name;
		}
		if(port_name == NULL)
			port_name = itoa(port);
	}
	numeric = 22 - strlen(port_name);
	snprintf(ip_port, SZ, "%.*s:%s", numeric, ip_addr, port_name);
}

static void tcp_do_one(int lnr)
{
	char local_addr[SZ], rem_addr[SZ];
	const char *state_str;
	int local_port, rem_port, d, state;
#ifdef CONFIG_FEATURE_IPV6
	struct sockaddr_in6 localaddr, remaddr;
	char addr6[INET6_ADDRSTRLEN];
	struct in6_addr in6;
#else
	struct sockaddr_in localaddr, remaddr;
#endif
	unsigned long rxq, txq;

	if (lnr == 0)
		return;

	sscanf(buffer, "%d: %64[0-9A-Fa-f]:%X "
				"%64[0-9A-Fa-f]:%X %X "
				"%lX:%lX %*s\n",
			      &d, local_addr, &local_port,
				rem_addr, &rem_port, &state,
				&txq, &rxq);

	if (strlen(local_addr) > 8) {
#ifdef CONFIG_FEATURE_IPV6
		sscanf(local_addr, "%08X%08X%08X%08X",
			   &in6.s6_addr32[0], &in6.s6_addr32[1],
			   &in6.s6_addr32[2], &in6.s6_addr32[3]);
		inet_ntop(AF_INET6, &in6, addr6, sizeof(addr6));
		inet_pton(AF_INET6, addr6, (struct sockaddr *) &localaddr.sin6_addr);
		sscanf(rem_addr, "%08X%08X%08X%08X",
			   &in6.s6_addr32[0], &in6.s6_addr32[1],
			   &in6.s6_addr32[2], &in6.s6_addr32[3]);
		inet_ntop(AF_INET6, &in6, addr6, sizeof(addr6));
		inet_pton(AF_INET6, addr6, (struct sockaddr *) &remaddr.sin6_addr);
		localaddr.sin6_family = AF_INET6;
		remaddr.sin6_family = AF_INET6;
#endif
	} else {
		sscanf(local_addr, "%X",
			   &((struct sockaddr_in *) &localaddr)->sin_addr.s_addr);
		sscanf(rem_addr, "%X",
			   &((struct sockaddr_in *) &remaddr)->sin_addr.s_addr);
		((struct sockaddr *) &localaddr)->sa_family = AF_INET;
		((struct sockaddr *) &remaddr)->sa_family = AF_INET;
	}
	state_str = tcp_state[state];
	if ((rem_port && (opt_flags&NETSTAT_CONNECTED)) ||
		(!rem_port && (opt_flags&NETSTAT_LISTENING)))
	{
		snprint_ip_port(local_addr, (struct sockaddr *) &localaddr,
						local_port, "tcp");

		snprint_ip_port(rem_addr, (struct sockaddr *) &remaddr,
						rem_port, "tcp");

		printf("tcp   %6ld %6ld %-23s %-23s %-12s\n",
			   rxq, txq, local_addr, rem_addr, state_str);

	}
}

static void udp_do_one(int lnr)
{
	char local_addr[SZ], rem_addr[SZ];
	char *state_str;
	int local_port, rem_port, d, state;
#ifdef CONFIG_FEATURE_IPV6
	struct sockaddr_in6 localaddr, remaddr;
	char addr6[INET6_ADDRSTRLEN];
	struct in6_addr in6;
#else
	struct sockaddr_in localaddr, remaddr;
#endif
	unsigned long rxq, txq;

	if (lnr == 0)
		return;

	sscanf(buffer, "%d: %64[0-9A-Fa-f]:%X "
				 "%64[0-9A-Fa-f]:%X %X "
				 "%lX:%lX %*s\n",
			   &d, local_addr, &local_port,
				 rem_addr, &rem_port, &state,
				 &txq, &rxq);

	if (strlen(local_addr) > 8) {
#ifdef CONFIG_FEATURE_IPV6
	/* Demangle what the kernel gives us */
		sscanf(local_addr, "%08X%08X%08X%08X",
			   &in6.s6_addr32[0], &in6.s6_addr32[1],
			   &in6.s6_addr32[2], &in6.s6_addr32[3]);
		inet_ntop(AF_INET6, &in6, addr6, sizeof(addr6));
		inet_pton(AF_INET6, addr6, (struct sockaddr *) &localaddr.sin6_addr);
		sscanf(rem_addr, "%08X%08X%08X%08X",
			   &in6.s6_addr32[0], &in6.s6_addr32[1],
			   &in6.s6_addr32[2], &in6.s6_addr32[3]);
		inet_ntop(AF_INET6, &in6, addr6, sizeof(addr6));
		inet_pton(AF_INET6, addr6, (struct sockaddr *) &remaddr.sin6_addr);
		localaddr.sin6_family = AF_INET6;
		remaddr.sin6_family = AF_INET6;
#endif
	} else {
		sscanf(local_addr, "%X",
			   &((struct sockaddr_in *) &localaddr)->sin_addr.s_addr);
		sscanf(rem_addr, "%X",
			   &((struct sockaddr_in *) &remaddr)->sin_addr.s_addr);
		((struct sockaddr *) &localaddr)->sa_family = AF_INET;
		((struct sockaddr *) &remaddr)->sa_family = AF_INET;
	}

	switch (state) {
		case TCP_ESTABLISHED:
			state_str = "ESTABLISHED";
			break;

		case TCP_CLOSE:
			state_str = "";
			break;

		default:
			state_str = "UNKNOWN";
			break;
	}

#ifdef CONFIG_FEATURE_IPV6
# define notnull(A) (((A.sin6_family == AF_INET6) &&            \
					 ((A.sin6_addr.s6_addr32[0]) ||            \
					  (A.sin6_addr.s6_addr32[1]) ||            \
					  (A.sin6_addr.s6_addr32[2]) ||            \
					  (A.sin6_addr.s6_addr32[3]))) ||          \
					((A.sin6_family == AF_INET) &&             \
					 ((struct sockaddr_in *) &A)->sin_addr.s_addr))
#else
# define notnull(A) (A.sin_addr.s_addr)
#endif
	if ((notnull(remaddr) && (opt_flags&NETSTAT_CONNECTED)) ||
		(!notnull(remaddr) && (opt_flags&NETSTAT_LISTENING)))
	{
		snprint_ip_port(local_addr, (struct sockaddr *) &localaddr,
						local_port, "udp");

		snprint_ip_port(rem_addr, (struct sockaddr *) &remaddr,
						rem_port, "udp");

		printf("udp   %6ld %6ld %-23s %-23s %-12s\n",
			   rxq, txq, local_addr, rem_addr, state_str);

	}
}

static void raw_do_one(int lnr)
{
	char local_addr[SZ], rem_addr[SZ];
	char *state_str;
	int local_port, rem_port, d, state;
#ifdef CONFIG_FEATURE_IPV6
	struct sockaddr_in6 localaddr, remaddr;
	char addr6[INET6_ADDRSTRLEN];
	struct in6_addr in6;
#else
	struct sockaddr_in localaddr, remaddr;
#endif
	unsigned long rxq, txq;

	if (lnr == 0)
		return;

	sscanf(buffer, "%d: %64[0-9A-Fa-f]:%X "
				 "%64[0-9A-Fa-f]:%X %X "
				 "%lX:%lX %*s\n",
			   &d, local_addr, &local_port,
				 rem_addr, &rem_port, &state,
				 &txq, &rxq);

	if (strlen(local_addr) > 8) {
#ifdef CONFIG_FEATURE_IPV6
		sscanf(local_addr, "%08X%08X%08X%08X",
			   &in6.s6_addr32[0], &in6.s6_addr32[1],
			   &in6.s6_addr32[2], &in6.s6_addr32[3]);
		inet_ntop(AF_INET6, &in6, addr6, sizeof(addr6));
		inet_pton(AF_INET6, addr6, (struct sockaddr *) &localaddr.sin6_addr);
		sscanf(rem_addr, "%08X%08X%08X%08X",
			   &in6.s6_addr32[0], &in6.s6_addr32[1],
			   &in6.s6_addr32[2], &in6.s6_addr32[3]);
		inet_ntop(AF_INET6, &in6, addr6, sizeof(addr6));
		inet_pton(AF_INET6, addr6, (struct sockaddr *) &remaddr.sin6_addr);
		localaddr.sin6_family = AF_INET6;
		remaddr.sin6_family = AF_INET6;
#endif
	} else {
		sscanf(local_addr, "%X",
			   &((struct sockaddr_in *) &localaddr)->sin_addr.s_addr);
		sscanf(rem_addr, "%X",
			   &((struct sockaddr_in *) &remaddr)->sin_addr.s_addr);
		((struct sockaddr *) &localaddr)->sa_family = AF_INET;
		((struct sockaddr *) &remaddr)->sa_family = AF_INET;
	}

	state_str = itoa(state);

#ifdef CONFIG_FEATURE_IPV6
# define notnull(A) (((A.sin6_family == AF_INET6) &&            \
					 ((A.sin6_addr.s6_addr32[0]) ||            \
					  (A.sin6_addr.s6_addr32[1]) ||            \
					  (A.sin6_addr.s6_addr32[2]) ||            \
					  (A.sin6_addr.s6_addr32[3]))) ||          \
					((A.sin6_family == AF_INET) &&             \
					 ((struct sockaddr_in *) &A)->sin_addr.s_addr))
#else
# define notnull(A) (A.sin_addr.s_addr)
#endif
	if ((notnull(remaddr) && (opt_flags&NETSTAT_CONNECTED)) ||
		(!notnull(remaddr) && (opt_flags&NETSTAT_LISTENING)))
	{
		snprint_ip_port(local_addr, (struct sockaddr *) &localaddr,
						local_port, "raw");

		snprint_ip_port(rem_addr, (struct sockaddr *) &remaddr,
						rem_port, "raw");

		printf("raw   %6ld %6ld %-23s %-23s %-12s\n",
			   rxq, txq, local_addr, rem_addr, state_str);

	}
}

#define HAS_INODE 1

static void unix_do_one(int nr)
{
	char path[PATH_MAX], ss_flags[32];
	char *ss_proto, *ss_state, *ss_type;
	int num, state, type, inode;
	void *d;
	unsigned long refcnt, proto, unix_flags;

	if (nr == 0) {
		if (strstr(buffer, "Inode"))
			has |= HAS_INODE;
		return;
	}
	path[0] = '\0';
	num = sscanf(buffer, "%p: %lX %lX %lX %X %X %d %s",
			   &d, &refcnt, &proto, &unix_flags, &type, &state,
			   &inode, path);
	if (num < 6) {
		bb_error_msg("warning, got bogus unix line.");
		return;
	}
	if (!(has & HAS_INODE))
		snprintf(path, sizeof(path), "%d", inode);

	if ((opt_flags&(NETSTAT_LISTENING|NETSTAT_CONNECTED))!=(NETSTAT_LISTENING|NETSTAT_CONNECTED)) {
		if ((state == SS_UNCONNECTED) && (unix_flags & SO_ACCEPTCON)) {
			if (!(opt_flags&NETSTAT_LISTENING))
				return;
		} else {
			if (!(opt_flags&NETSTAT_CONNECTED))
				return;
		}
	}

	switch (proto) {
		case 0:
			ss_proto = "unix";
			break;

		default:
			ss_proto = "??";
	}

	switch (type) {
		case SOCK_STREAM:
			ss_type = "STREAM";
			break;

		case SOCK_DGRAM:
			ss_type = "DGRAM";
			break;

		case SOCK_RAW:
			ss_type = "RAW";
			break;

		case SOCK_RDM:
			ss_type = "RDM";
			break;

		case SOCK_SEQPACKET:
			ss_type = "SEQPACKET";
			break;

		default:
			ss_type = "UNKNOWN";
	}

	switch (state) {
		case SS_FREE:
			ss_state = "FREE";
			break;

		case SS_UNCONNECTED:
			/*
			 * Unconnected sockets may be listening
			 * for something.
			 */
			if (unix_flags & SO_ACCEPTCON) {
				ss_state = "LISTENING";
			} else {
				ss_state = "";
			}
			break;

		case SS_CONNECTING:
			ss_state = "CONNECTING";
			break;

		case SS_CONNECTED:
			ss_state = "CONNECTED";
			break;

		case SS_DISCONNECTING:
			ss_state = "DISCONNECTING";
			break;

		default:
			ss_state = "UNKNOWN";
	}

	strcpy(ss_flags, "[ ");
	if (unix_flags & SO_ACCEPTCON)
		strcat(ss_flags, "ACC ");
	if (unix_flags & SO_WAITDATA)
		strcat(ss_flags, "W ");
	if (unix_flags & SO_NOSPACE)
		strcat(ss_flags, "N ");

	strcat(ss_flags, "]");

	printf("%-5s %-6ld %-11s %-10s %-13s ",
		   ss_proto, refcnt, ss_flags, ss_type, ss_state);
	if (has & HAS_INODE)
		printf("%-6d ",inode);
	else
		printf("-      ");
	puts(path);
}

#define _PATH_PROCNET_UDP "/proc/net/udp"
#define _PATH_PROCNET_UDP6 "/proc/net/udp6"
#define _PATH_PROCNET_TCP "/proc/net/tcp"
#define _PATH_PROCNET_TCP6 "/proc/net/tcp6"
#define _PATH_PROCNET_RAW "/proc/net/raw"
#define _PATH_PROCNET_RAW6 "/proc/net/raw6"
#define _PATH_PROCNET_UNIX "/proc/net/unix"

static const struct DO_INFO_S {
	int netstat_flag;
	const char *patch;
	const char *type;
	void (*proc)(int);
} do_info_s[] = {
	{ NETSTAT_TCP, _PATH_PROCNET_TCP, "AF INET (tcp)", tcp_do_one },
#ifdef CONFIG_FEATURE_IPV6
	{ NETSTAT_TCP, _PATH_PROCNET_TCP6, "AF INET6 (tcp)", tcp_do_one },
#endif
	{NETSTAT_UDP, _PATH_PROCNET_UDP, "AF INET (udp)", udp_do_one},
#ifdef CONFIG_FEATURE_IPV6
	{NETSTAT_UDP, _PATH_PROCNET_UDP6, "AF INET6 (udp)", udp_do_one},
#endif
	{NETSTAT_RAW, _PATH_PROCNET_RAW, "AF INET (raw)", raw_do_one},
#ifdef CONFIG_FEATURE_IPV6
	{NETSTAT_RAW, _PATH_PROCNET_RAW6, "AF INET6 (raw)", raw_do_one},
#endif
	{NETSTAT_UNIX, _PATH_PROCNET_UNIX, "AF UNIX", unix_do_one},
	{0, NULL, NULL, NULL }
};

/*
 * Our main function.
 */

int netstat_main(void)
{
	int opt;
	const char *top_msg = NULL;
	const char *displ_serv_types;

	const struct DO_INFO_S *d;

	opt_flags = NETSTAT_CONNECTED |
			NETSTAT_TCP | NETSTAT_UDP | NETSTAT_RAW | NETSTAT_UNIX;
	/* unset '-l' if '-a' setted */
	opt = (int)bb_xgetopt_ulflags("lanrtuwxe\0a-l");
#define showroute (opt & 8)     /* "-r" */
#define extended  (opt & 256)   /* "-e" */

	if ( showroute ) {
#ifdef CONFIG_ROUTE                     /* "-n" = 4  */
		displayroutes ( opt & NETSTAT_NUMERIC, !extended );
		return 0;
#else
		bb_error_msg_and_die( "-r (display routing table) is not compiled in." );
#endif
	}

	if(opt & 1) {
		/* 'l' */
		opt_flags &= ~NETSTAT_CONNECTED;
		opt_flags |= NETSTAT_LISTENING;
	}
			   /* & 2 '-a' */    /* & 4 '-n' */
	opt_flags |= opt & (NETSTAT_LISTENING | NETSTAT_NUMERIC);
	/*           16              32      64              128 */
	opt &= (NETSTAT_TCP | NETSTAT_UDP | NETSTAT_RAW | NETSTAT_UNIX);
	if (opt) {
		opt_flags &= ~(NETSTAT_TCP|NETSTAT_UDP|NETSTAT_RAW|NETSTAT_UNIX);
		opt_flags |= opt;
	}
	opt = opt_flags;
	if ((opt & (NETSTAT_LISTENING | NETSTAT_CONNECTED))==
				(NETSTAT_LISTENING | NETSTAT_CONNECTED))
		displ_serv_types = "(servers and established)";
	else {
		if (opt & NETSTAT_LISTENING)
			displ_serv_types = "(only servers)";
		else
			displ_serv_types = "(w/o servers)";
	}

	for(d = do_info_s; d->patch != NULL; d++) {
		const char *table_head_msg;
		FILE *procinfo;

		if((opt & d->netstat_flag) == 0)
			continue;

		if(d->netstat_flag != NETSTAT_UNIX) {
			if(top_msg != NULL)
				goto l_do;
			top_msg = "Internet connection";
			table_head_msg =
"cv-Q Send-Q Local Address           Foreign Address         State";
		} else {
			top_msg = "UNIX domain socket";
			table_head_msg =
"fCnt Flags       Type       State         I-Node Path";
		}
		printf("Active %ss %s\nProto Re%s\n",
				top_msg,
				    displ_serv_types,
					table_head_msg);

	    l_do:
		procinfo = fopen(d->patch, "r");
		if (procinfo == NULL) {
			if (errno != ENOENT) {
				bb_perror_msg("%s", d->patch);
			} else {
				bb_error_msg("no support for `%s' on this system.", d->type);
			}
		} else {
			int lnr = 0;
			while(fgets(buffer, sizeof(buffer), procinfo) != NULL)
				(d->proc)(lnr++);
			fclose(procinfo);
		}
	}

	return 0;
}
#endif /* APPLET_USAGE */
