/*
 * busybox nslookup
 *
 * Copyright (C) 2017 Jo-Philipp Wich <jo@mein.io>
 *
 * Copyright (C) 2018 Vladimir Oleynik <dzo@simtreas.ru>
 *  bug fixes
 *
 */

#ifdef APPLET_USAGE
#define nslookup_trivial_usage \
	"[HOST] [SERVER]"
#define nslookup_full_usage \
	"Query the nameserver for the IP address of the given HOST\n" \
	"optionally using a specified DNS server"
#else

#include "busybox.h"
#include <resolv.h>
#include <string.h>
#include <time.h>
#include <poll.h>
#include <unistd.h>
#include "bb_inet.h"
#include "config_read.h"

typedef struct len_and_sockaddr {
	socklen_t len;
	union {
		struct sockaddr sa;
		struct sockaddr_in sin;
#if ENABLE_FEATURE_IPV6
		struct sockaddr_in6 sin6;
#endif
	} u;
} len_and_sockaddr;

#ifndef INET6_ADDRSTRLEN
# define INET6_ADDRSTRLEN 46
#endif

struct ns {
	const char *name;
	len_and_sockaddr addr;
	int failures;
	int replies;
};

struct query {
	const char *name;
	size_t qlen, rlen;
	unsigned char query[512], reply[512];
	unsigned long latency;
	int rcode, n_ns;
};

struct QTYPES {
	const char *name;
	int type;
};

static const struct QTYPES qtypes[] = {
	{ "SOA",  ns_t_soa   },
	{ "NS",   ns_t_ns    },
	{ "A",    ns_t_a     },
	{ "AAAA", ns_t_aaaa  },
	{ "CNAME",ns_t_cname },
	{ "MX",   ns_t_mx    },
	{ "TXT",  ns_t_txt   },
	{ "PTR",  ns_t_ptr   },
	{ "ANY",  ns_t_any   },
	{ NULL, 0 }
};

static const char * const rcodes[] = {
	"NOERROR",
	"FORMERR",
	"SERVFAIL",
	"NXDOMAIN",
	"NOTIMP",
	"REFUSED",
	"YXDOMAIN",
	"YXRRSET",
	"NXRRSET",
	"NOTAUTH",
	"NOTZONE",
	"RESERVED11",
	"RESERVED12",
	"RESERVED13",
	"RESERVED14",
	"RESERVED15",
	"BADVERS"
};

CALC_BSS struct BB_COMMON_BSS {
	int default_port;
	int default_retry;
	int default_timeout;
} bb_common_bss;
#define default_port bb_common_bss.default_port
#define default_retry bb_common_bss.default_retry
#define default_timeout bb_common_bss.default_timeout


static int parse_reply(const unsigned char *msg, size_t len)
{
	ns_msg handle;
	ns_rr rr;
	int i, n, rdlen;
	const char *format = NULL;
	char astr[INET6_ADDRSTRLEN], dname[MAXDNAME];
	const unsigned char *cp;

	if (ns_initparse(msg, len, &handle) != 0) {
		//fprintf(stderr, "Unable to parse reply: %s\n", strerror(errno));
		return -1;
	}

	for (i = 0; i < ns_msg_count(handle, ns_s_an); i++) {
		if (ns_parserr(&handle, ns_s_an, i, &rr) != 0) {
			//fprintf(stderr, "Unable to parse resource record: %s\n", strerror(errno));
			return -1;
		}

		rdlen = ns_rr_rdlen(rr);

		switch (ns_rr_type(rr))
		{
		case ns_t_a:
			if (rdlen != 4) {
				//fprintf(stderr, "Unexpected A record length\n");
				return -1;
			}
			inet_ntop(AF_INET, ns_rr_rdata(rr), astr, sizeof(astr));
			printf("Name:\t%s\nAddress: %s\n", ns_rr_name(rr), astr);
			break;

		case ns_t_aaaa:
			if (rdlen != 16) {
				//fprintf(stderr, "Unexpected AAAA record length\n");
				return -1;
			}
			inet_ntop(AF_INET6, ns_rr_rdata(rr), astr, sizeof(astr));
			printf("%s\thas AAAA address %s\n", ns_rr_name(rr), astr);
			break;

		case ns_t_ns:
			if (!format)
				format = "%s\tnameserver = %s\n";
			/* fall through */

		case ns_t_cname:
			if (!format)
				format = "%s\tcanonical name = %s\n";
			/* fall through */

		case ns_t_ptr:
			if (!format)
				format = "%s\tname = %s\n";
			if (ns_name_uncompress(ns_msg_base(handle), ns_msg_end(handle),
				ns_rr_rdata(rr), dname, sizeof(dname)) < 0) {
				//fprintf(stderr, "Unable to uncompress domain: %s\n", strerror(errno));
				return -1;
			}
			printf(format, ns_rr_name(rr), dname);
			break;

		case ns_t_mx:
			if (rdlen < 2) {
				//fprintf(stderr, "MX record too short\n");
				return -1;
			}
			n = ns_get16(ns_rr_rdata(rr));
			if (ns_name_uncompress(ns_msg_base(handle), ns_msg_end(handle),
				ns_rr_rdata(rr) + 2, dname, sizeof(dname)) < 0) {
				//fprintf(stderr, "Cannot uncompress MX domain: %s\n", strerror(errno));
				return -1;
			}
			printf("%s\tmail exchanger = %d %s\n", ns_rr_name(rr), n, dname);
			break;

		case ns_t_txt:
			if (rdlen < 1) {
				//fprintf(stderr, "TXT record too short\n");
				return -1;
			}
			n = *(unsigned char *)ns_rr_rdata(rr);
			if (n > 0) {
				memset(dname, 0, sizeof(dname));
				memcpy(dname, ns_rr_rdata(rr) + 1, n);
				printf("%s\ttext = \"%s\"\n", ns_rr_name(rr), dname);
			}
			break;

		case ns_t_soa:
			if (rdlen < 20) {
				//fprintf(stderr, "SOA record too short\n");
				return -1;
			}

			printf("%s\n", ns_rr_name(rr));

			cp = ns_rr_rdata(rr);
			n = ns_name_uncompress(ns_msg_base(handle), ns_msg_end(handle),
					       cp, dname, sizeof(dname));

			if (n < 0) {
				//fprintf(stderr, "Unable to uncompress domain: %s\n", strerror(errno));
				return -1;
			}

			printf("\torigin = %s\n", dname);
			cp += n;

			n = ns_name_uncompress(ns_msg_base(handle), ns_msg_end(handle),
					       cp, dname, sizeof(dname));

			if (n < 0) {
				//fprintf(stderr, "Unable to uncompress domain: %s\n", strerror(errno));
				return -1;
			}

			printf("\tmail addr = %s\n", dname);
			cp += n;

			printf("\tserial = %lu\n", (unsigned long)ns_get32(cp));
			cp += 4;

			printf("\trefresh = %lu\n", (unsigned long)ns_get32(cp));
			cp += 4;

			printf("\tretry = %lu\n", (unsigned long)ns_get32(cp));
			cp += 4;

			printf("\texpire = %lu\n", (unsigned long)ns_get32(cp));
			cp += 4;

			printf("\tminimum = %lu\n", (unsigned long)ns_get32(cp));
			break;

		default:
			break;
		}
	}

	return i;
}

static int parse_nsaddr(const char *addrstr, len_and_sockaddr *lsa)
{
	char *hash;
	int port = default_port;
#if ENABLE_FEATURE_IPV6
	char *eptr, ifname[IFNAMSIZ];
	unsigned int scope = 0;
#else
# define scope 0
#endif

	hash = strchr(addrstr, '#');

	if (hash) {
		int p = bb_xparse_i_number(BB_PN_SET_UPPER | BB_PN_SILENT_RE, hash + 1, 65535);
		if (p < 0)
			return -1;
		port = p;
	}

#if ENABLE_FEATURE_IPV6
	hash = strchr(addrstr, '%');

	if (hash) {
		for (eptr = ++hash; *eptr != '\0' && *eptr != '#'; eptr++) {
			if ((eptr - hash) >= IFNAMSIZ) {
				errno = ENODEV;
				return -1;
			}

			ifname[eptr - hash] = *eptr;
		}

		ifname[eptr - hash] = '\0';
		scope = if_nametoindex(ifname);

		if (scope == 0) {
			errno = ENODEV;
			return -1;
		}
	}

	if (inet_pton(AF_INET6, addrstr, &lsa->u.sin6.sin6_addr) > 0) {
		lsa->u.sin6.sin6_family = AF_INET6;
		lsa->u.sin6.sin6_port = htons(port);
		lsa->u.sin6.sin6_scope_id = scope;
		lsa->len = sizeof(lsa->u.sin6);
		return 0;
	}
#endif

	if (!scope && inet_pton(AF_INET, addrstr, &lsa->u.sin.sin_addr) > 0) {
		lsa->u.sin.sin_family = AF_INET;
		lsa->u.sin.sin_port = htons(port);
		lsa->len = sizeof(lsa->u.sin);
		return 0;
	}

	errno = EINVAL;
	return -1;
}

static char *make_ptr(const char *addrstr)
{
	char ptrstr[80];
	unsigned char addr[16];
	char *ptr = ptrstr;

#if ENABLE_FEATURE_IPV6
	if (inet_pton(AF_INET6, addrstr, addr) > 0) {
		static const char hexdigit[] = "0123456789abcdef";
		int i;

		if (memcmp(addr, "\0\0\0\0\0\0\0\0\0\0\xff\xff", 12) != 0) {
			for (i = 0; i < 16; i++) {
				*ptr++ = hexdigit[(unsigned char)addr[15 - i] & 0xf];
				*ptr++ = '.';
				*ptr++ = hexdigit[(unsigned char)addr[15 - i] >> 4];
				*ptr++ = '.';
			}
			strcpy(ptr, "ip6.arpa");
		}
		else {
			sprintf(ptr, "%u.%u.%u.%u.in-addr.arpa", addr[15], addr[14], addr[13], addr[12]);
		}

		return xstrdup(ptrstr);
	}
#endif
	if (inet_pton(AF_INET, addrstr, addr) > 0) {
		sprintf(ptr, "%u.%u.%u.%u.in-addr.arpa", addr[3], addr[2], addr[1], addr[0]);
		return xstrdup(ptrstr);
	}

	return NULL;
}

static unsigned long mtime(void)
{
	struct timespec ts;
	clock_gettime(CLOCK_REALTIME, &ts);
	return (unsigned long)ts.tv_sec * 1000 + ts.tv_nsec / 1000000;
}

#if ENABLE_FEATURE_IPV6
static void to_v4_mapped(len_and_sockaddr *a)
{
	if (a->u.sa.sa_family != AF_INET)
		return;

	memcpy(a->u.sin6.sin6_addr.s6_addr + 12, &a->u.sin.sin_addr, 4);
	memcpy(a->u.sin6.sin6_addr.s6_addr, "\0\0\0\0\0\0\0\0\0\0\xff\xff", 12);
	a->u.sin6.sin6_family = AF_INET6;
	a->u.sin6.sin6_flowinfo = 0;
	a->u.sin6.sin6_scope_id = 0;
	a->len = sizeof(a->u.sin6);
}
#endif

/*
 * Function logic borrowed & modified from musl libc, res_msend.c
 */

static int send_queries(struct ns *ns, int n_ns, struct query *queries, int n_queries)
{
	int fd;
	int timeout = default_timeout * 1000, retry_interval, servfail_retry = 0;
	len_and_sockaddr from = { };
	int recvlen = 0;
	int n_replies = 0;
	struct pollfd pfd;
	unsigned long t0, t1, t2;
	int nn, qn, next_query = 0;

	from.u.sa.sa_family = AF_INET;
	from.len = sizeof(from.u.sin);

#if ENABLE_FEATURE_IPV6
	for (nn = 0; nn < n_ns; nn++) {
		if (ns[nn].addr.u.sa.sa_family == AF_INET6) {
			from.u.sa.sa_family = AF_INET6;
			from.len = sizeof(from.u.sin6);
			break;
		}
	}
#endif

	/* Get local address and open/bind a socket */
	fd = socket(from.u.sa.sa_family, SOCK_DGRAM|SOCK_CLOEXEC|SOCK_NONBLOCK, 0);

#if ENABLE_FEATURE_IPV6
	/* Handle case where system lacks IPv6 support */
	if (fd < 0 && from.u.sa.sa_family == AF_INET6 && errno == EAFNOSUPPORT) {
		fd = socket(AF_INET, SOCK_DGRAM|SOCK_CLOEXEC|SOCK_NONBLOCK, 0);
		from.u.sa.sa_family = AF_INET;
	}
#endif

	if (fd < 0)
		return -1;

	if (bind(fd, &from.u.sa, from.len) < 0) {
		close(fd);
		return -1;
	}

#if ENABLE_FEATURE_IPV6
	/* Convert any IPv4 addresses in a mixed environment to v4-mapped */
	if (from.u.sa.sa_family == AF_INET6) {
		setsockopt_1i(fd, IPPROTO_IPV6, IPV6_V6ONLY, 1);

		for (nn = 0; nn < n_ns; nn++)
			to_v4_mapped(&ns[nn].addr);
	}
#endif

	pfd.fd = fd;
	pfd.events = POLLIN;
	retry_interval = timeout / default_retry;
	t0 = t2 = mtime();
	t1 = t2 - retry_interval;

	for (; t2 - t0 < timeout; t2 = mtime()) {
		if (t2 - t1 >= retry_interval) {
			for (qn = 0; qn < n_queries; qn++) {
				if (queries[qn].rlen)
					continue;

				for (nn = 0; nn < n_ns; nn++) {
					sendto(fd, queries[qn].query, queries[qn].qlen,
					       MSG_NOSIGNAL, &ns[nn].addr.u.sa, ns[nn].addr.len);
				}
			}

			t1 = t2;
			servfail_retry = 2 * n_queries;
		}

		/* Wait for a response, or until time to retry */
		if (poll(&pfd, 1, t1+retry_interval-t2) <= 0)
			continue;

		while (1) {
			recvlen = recvfrom(fd, queries[next_query].reply,
					   sizeof(queries[next_query].reply), 0,
					   &from.u.sa, &from.len);

			/* read error */
			if (recvlen < 0)
				break;

			/* Ignore non-identifiable packets */
			if (recvlen < 4)
				continue;

			/* Ignore replies from addresses we didn't send to */
			for (nn = 0; nn < n_ns; nn++)
				if (memcmp(&from.u.sa, &ns[nn].addr.u.sa, from.len) == 0)
					break;

			if (nn >= n_ns)
				continue;

			/* Find which query this answer goes with, if any */
			for (qn = next_query; qn < n_queries; qn++)
				if (!memcmp(queries[next_query].reply, queries[qn].query, 2))
					break;

			if (qn >= n_queries || queries[qn].rlen)
				continue;

			queries[qn].rcode = queries[next_query].reply[3] & 15;
			queries[qn].latency = mtime() - t0;
			queries[qn].n_ns = nn;

			ns[nn].replies++;

			/* Only accept positive or negative responses;
			 * retry immediately on server failure, and ignore
			 * all other codes such as refusal. */
			switch (queries[qn].rcode) {
			case 0:
			case 3:
				break;

			case 2:
				if (servfail_retry && servfail_retry--) {
					ns[nn].failures++;
					sendto(fd, queries[qn].query, queries[qn].qlen,
					       MSG_NOSIGNAL, &ns[nn].addr.u.sa, ns[nn].addr.len);
				}
				/* fall through */

			default:
				continue;
			}

			/* Store answer */
			n_replies++;

			queries[qn].rlen = recvlen;

			if (qn == next_query) {
				while (next_query < n_queries) {
					if (!queries[next_query].rlen)
						break;

					next_query++;
				}
			}
			else {
				memcpy(queries[qn].reply, queries[next_query].reply, recvlen);
			}

			if (next_query >= n_queries)
				return n_replies;
		}
	}

	return n_replies;
}

static struct ns *add_ns(struct ns **ns, int *n_ns, const char *addr)
{
	len_and_sockaddr a = { };
	struct ns *tmp;

	if (parse_nsaddr(addr, &a))
		return NULL;

	tmp = realloc(*ns, sizeof(**ns) * (*n_ns + 1));

	if (!tmp)
		return NULL;

	*ns = tmp;

	(*ns)[*n_ns].addr = a;
	(*ns)[*n_ns].name = addr;
	(*ns)[*n_ns].replies = 0;
	(*ns)[*n_ns].failures = 0;

	return &(*ns)[(*n_ns)++];
}

static int parse_resolvconf(struct ns **ns, int *n_ns)
{
	int prev_n_ns = *n_ns;
	config_read_t *crt = config_read_prepare("/etc/resolv.conf", 0);

	while( (crt = config_read(crt, "\0\1\2\0\0")) != NULL) {
		char *p = crt->tokens[1];
		if(p != NULL && !strcmp(crt->tokens[0], "nameserver")) {
			if (!add_ns(ns, n_ns, p = strdup(p)))
				free(p);
		}
	}

	return *n_ns - prev_n_ns;
}

static struct query *add_query(struct query **queries, int *n_queries,
			       int type, const char *dname)
{
	struct query *tmp;
	ssize_t qlen;

	tmp = realloc(*queries, sizeof(**queries) * (*n_queries + 1));

	if (!tmp)
		return NULL;

	memset(&tmp[*n_queries], 0, sizeof(*tmp));

	qlen = res_mkquery(QUERY, dname, C_IN, type, NULL, 0, NULL,
			   tmp[*n_queries].query, sizeof(tmp[*n_queries].query));

	tmp[*n_queries].qlen = qlen;
	tmp[*n_queries].name = dname;
	*queries = tmp;

	return &tmp[(*n_queries)++];
}

#define sizeof_sal2str_buf (INET6_ADDRSTRLEN + 1 + IFNAMSIZ + 1 + 5 + 1)

static char *sal2str(len_and_sockaddr *a, char buf[])
{
	char *p = buf;

#if ENABLE_FEATURE_IPV6
	if (a->u.sa.sa_family == AF_INET6) {
		inet_ntop(AF_INET6, &a->u.sin6.sin6_addr, buf, sizeof_sal2str_buf);
		p += strlen(p);

		if (a->u.sin6.sin6_scope_id) {
			if (if_indextoname(a->u.sin6.sin6_scope_id, p + 1)) {
				*p++ = '%';
				p += strlen(p);
			}
		}
		sprintf(p, "#%hu", ntohs(a->u.sin6.sin6_port));
	} else
#endif
		{
		inet_ntop(AF_INET, &a->u.sin.sin_addr, buf, sizeof_sal2str_buf);
		p += strlen(p);
		sprintf(p, "#%hu", ntohs(a->u.sin.sin_port));
	}

	return buf;
}


struct OPT_ABBR {
	const char *abbr;
	int type;
};

#define OPT_Q           0
#define OPT_PORT        1
#define OPT_RETRY       2
#define OPT_TIMEOUT     3
#define OPT_DEBUG       4
#define OPT_NO_DEBUG    5

static const struct OPT_ABBR opt_abbr[] = {
	{ "type", OPT_Q },
	{ "ty", OPT_Q },
	{ "querytype", OPT_Q },
	{ "query", OPT_Q },
	{ "qu", OPT_Q },
	{ "q", OPT_Q },
	{ "port", OPT_PORT },
	{ "po", OPT_PORT },
	{ "retry", OPT_RETRY },
	{ "ret", OPT_RETRY },
	{ "timeout", OPT_TIMEOUT },
	{ "t", OPT_TIMEOUT },
	{ "debug", OPT_DEBUG },
	{ "deb", OPT_DEBUG },
	{ "nodebug", OPT_NO_DEBUG },
	{ "nodeb", OPT_NO_DEBUG },
	{ NULL, 0 }
};

static void *
find_case_string_from_struct(const void *p, unsigned offset, const char *key)
{
	while (*(char **)p != NULL) {
		if (strcasecmp(*(char **)p, key) == 0)
			break;
		p = ((char *)p) + offset;
	}
	return (void *)p;
}


int nslookup_main()
{
	int rc;
	char *ptr;
	struct ns *ns = NULL;
	struct query *queries = NULL;
	int n_ns = 0, n_queries = 0;
	int c;
	int debug = 0;
	unsigned int types = 0;
	HEADER *header;
	char buf[sizeof_sal2str_buf];

	default_port = 53;
	default_retry = 2;
	default_timeout = 5;

	Gargv++;
	for (;;) {
		char *arg = *Gargv;
		char *val;
		const struct OPT_ABBR *a;

		if (arg == NULL)
			bb_show_usage();
		if (*arg != '-')
			break;
		arg++;
		if(*arg == '\0') {
			Gargv++;
			break;
		}
		/* Separate out "=val" part */
		val = strchr_set0(arg, '=');
		if(val == NULL)
			bb_error_msg_and_die("Invalid option %s", arg);
		val++;

		a = find_case_string_from_struct(opt_abbr, sizeof(struct OPT_ABBR), arg);
		if(a->abbr == NULL)
			bb_show_usage();
		if(a->type == OPT_Q) {
			const struct QTYPES *q = find_case_string_from_struct(qtypes, sizeof(struct QTYPES), val);
			if (q->name == NULL)
				bb_error_msg_and_die("Invalid query type \"%s\"", val);
			types |= (1 << (q-qtypes));
		}
		if(a->type == OPT_PORT || a->type == OPT_RETRY || a->type == OPT_TIMEOUT) {
			if(a->type == OPT_PORT) {
				default_port = bb_xparse_i_number(BB_PN_SET_LOWER1 | BB_PN_SET_UPPER | BB_PN_ERR_STR1,
									val, 65535,"invalid server port");
			} else if(a->type == OPT_TIMEOUT) {
				default_timeout = bb_xparse_i_number(BB_PN_SET_LOWER1 | BB_PN_ERR_STR1,
									val, "invalid timeout value");
			} else {
				default_retry = bb_xparse_i_number(BB_PN_SET_LOWER1 | BB_PN_SET_UPPER | BB_PN_ERR_STR1,
									val, INT_MAX / 1000, "invalid retry value");
			}
		}
		if(a->type == OPT_DEBUG)
			debug = 1;
		if(a->type == OPT_NO_DEBUG)
			debug = 0;
		Gargv++;
	}
	if (types == 0) {
		/* No explicit type given, guess query type.
		 * If we can convert the domain argument into a ptr (means that
		 * inet_pton() could read it) we assume a PTR request, else
		 * we issue A queries. */
		ptr = make_ptr(*Gargv);

		if (ptr)
			add_query(&queries, &n_queries, T_PTR, ptr);
		else
			add_query(&queries, &n_queries, T_A, *Gargv);
	} else {
		for (c = 0; qtypes[c].name; c++)
			if (types & (1 << c))
				add_query(&queries, &n_queries, qtypes[c].type, *Gargv);
	}

	/* Use given DNS server if present */
	Gargv++;
	if (*Gargv != NULL) {
		if(Gargv[1] != NULL)
			bb_show_usage();
		if (!add_ns(&ns, &n_ns, *Gargv))
			bb_perror_msg_and_die("Invalid NS server address \"%s\"", *Gargv);
	} else {
		parse_resolvconf(&ns, &n_ns);
	}

	/* Fall back to localhost if we could not find NS in resolv.conf */
	if (n_ns == 0) {
		add_ns(&ns, &n_ns, "127.0.0.1");
	}

	for (c = 0; c < n_ns; c++) {
		rc = send_queries(&ns[c], 1, queries, n_queries);

		if (rc < 0)
			bb_perror_msg_and_die("Failed to send queries");
		else if (rc > 0)
			break;
	}

	if (c >= n_ns) {
		printf(";; connection timed out; no servers could be reached\n\n");

		return 1;
	}

	printf("Server:\t\t%s\n", ns[c].name);
	printf("Address:\t%s\n", sal2str(&ns[c].addr, buf));

	printf("\n");

	for (rc = 0; rc < n_queries; rc++) {
		if (debug)
			printf("Query #%d completed in %lums:\n", rc, queries[rc].latency);

		if (queries[rc].rcode != 0) {
			printf("** server can't find %s: %s\n", queries[rc].name, rcodes[queries[rc].rcode]);
			continue;
		}

		c = 0;

		if (queries[rc].rlen) {
			header = (HEADER *)queries[rc].reply;

			if (!header->aa)
				printf("Non-authoritative answer:\n");

			c = parse_reply(queries[rc].reply, queries[rc].rlen);
		}

		if (c == 0)
			printf("*** Can't find %s: No answer\n", queries[rc].name);
		else if (c < 0)
			printf("*** Can't find %s: Parse error\n", queries[rc].name);

		printf("\n");
	}

	return 0;
}
#endif
