/*
 * ntpclient.c - NTP client
 *
 * Copyright (C) 1997, 1999, 2000, 2003, 2006, 2007, 2010, 2015  Larry Doolittle  <larry@doolittle.boa.org>
 * Modified by Vladimir Oleynik (vodz) <dzo@simtreas.ru> (c) 2022 to be used in busybox
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License (Version 2,
 *  June 1991) as published by the Free Software Foundation.  At the
 *  time of writing, that license was published by the FSF with the URL
 *  http://www.gnu.org/copyleft/gpl.html, and is incorporated herein by
 *  reference.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  Possible future improvements:
 *      - Write more documentation  :-(
 *      - Support leap second processing
 *      - Support IPv6
 *      - Support multiple (interleaved) servers
 *
 */

/* phaselock.c - Phase locking for NTP client
 *  Possible future improvements:
 *      - Subtract configurable amount from errorbar
 *      - Build in general run-time access to tune parameters
 *      - Sculpt code so it's legible, this version is out of control
 */

#ifdef APPLET_USAGE
#define ntpclient_trivial_usage \
	"[-c c] [-f f] [-g g] [-i i] [-p p] [-q md] [-lst] -h host"
#define ntpclient_full_usage \
	"NTP (RFC-1305, RFC-4330) client.\n\n" \
	"Options:\n" \
	"\t-c count\tstop after count time measurements (default 0 means go forever)\n" \
	"\t-d\t\tprint diagnostics (feature can be disabled at compile time)\n" \
	"\t-g goodness\tcauses ntpclient to stop after getting a result more accurate\n" \
	"\t\t\t\tthan goodness (microseconds, default 0 means go forever)\n" \
	"\t-i interval\tcheck time every interval seconds (default 600)\n" \
	"\t-l\t\tattempt to lock local clock to server using adjtimex(2)\n" \
	"\t-p port\t\tlocal NTP client UDP port (default 0 means 'any available')\n" \
	"\t-q min_delay\tminimum packet delay for transaction (default 800 microseconds)\n" \
	"\t-r\t\treplay analysis code based on stdin\n" \
	"\t-s\t\tsimple clock set (implies -c 1)\n" \
	"\t-t\t\ttrust network and server, no RFC-4330 recommended cross-checks\n" \
	"\t-h hostname\t(mandatory) NTP server, against which to measure system time"
#else

#define PRECISION_SIOCGSTAMP

#include "libbb.h"
#include "bb_inet.h"

#include <string.h>
#include <netinet/in.h>
#include <time.h>
#include <unistd.h>
#ifdef PRECISION_SIOCGSTAMP
#include <sys/ioctl.h>
#include <linux/sockios.h>      /* SIOCGSTAMP */
#endif

/* Default to the RFC-4330 specified value */
#ifndef MIN_INTERVAL
#define MIN_INTERVAL 15
#endif


/* XXX fixme - non-automatic build configuration */
#ifdef __linux__
#include <sys/utsname.h>
#include <sys/time.h>
#include <sys/timex.h>
#endif
/* end configuration for host systems */

#define JAN_1970        0x83aa7e80      /* 2208988800 1970 - 1900 in seconds */
#define NTP_PORT (123)

/* How to multiply by 4294.967296 quickly (and not quite exactly)
 * without using floating point or greater than 32-bit integers.
 * If you want to fix the last 12 microseconds of error, add in
 * (2911*(x))>>28)
 */
#define NTPFRAC(x) ( 4294*(x) + ( (1981*(x))>>11 ) )

/* The reverse of the above, needed if we want to set our microsecond
 * clock (via clock_settime) based on the incoming time in NTP format.
 * Basically exact.
 */
#define USEC(x) ( ( (x) >> 12 ) - 759 * ( ( ( (x) >> 10 ) + 32768 ) >> 16 ) )

/* Converts NTP delay and dispersion, apparently in seconds scaled
 * by 65536, to microseconds.  RFC-1305 states this time is in seconds,
 * doesn't mention the scaling.
 * Should somehow be the same as 1000000 * x / 65536
 */
#define sec2u(x) ( (x) * 15.2587890625 )

#define RING_SIZE 16
#define MAX_CORRECT 250   /* ppm change to system clock */
#define MAX_C (MAX_CORRECT)*65536

struct datum {
	unsigned int absolute;
	double skew;
	double errorbar;
	int freq;
	/* s.s.min and s.s.max (skews) are "corrected" to what they would
	 * have been if freq had been constant at its current value during
	 * the measurements.
	 */
	union {
		struct { double min; double max; } s;
		double ss[2];
	} s;
};

struct _seg {
	double slope;
	double offset;
};

/* Pseudo-class for finding consistent frequency shift */
#define MIN_INIT 20
struct _polygon {
	double l_min;
	double r_min;
};

CALC_BSS struct BB_COMMON_BSS {
	double min_delay;       /* user-changeable, units are microseconds */
	struct datum d_ring[RING_SIZE];
	struct _seg maxseg[RING_SIZE+1], minseg[RING_SIZE+1];
	struct _polygon df;
	uint32_t incoming_word[325];
	int d_ring_rp, valid;
	uint32_t time_of_send[2];
	int live;
	int set_clock;   /* non-zero presumably needs root privs */
	int probe_count; /* default of 0 means loop forever */
	int cycle_time;
	int goodness;
	int cross_check;
} bb_common_bss;
#define min_delay bb_common_bss.min_delay
#define d_ring bb_common_bss.d_ring
#define maxseg bb_common_bss.maxseg
#define minseg bb_common_bss.minseg
#define df bb_common_bss.df
#define incoming_word bb_common_bss.incoming_word
#define d_ring_rp bb_common_bss.d_ring_rp
#define valid bb_common_bss.valid

static int next_up(int i) { int r = i + 1; if (r >= RING_SIZE) r = 0; return r;}
static int next_dn(int i) { int r = i - 1; if (r < 0) r = RING_SIZE - 1; return r;}

/* Looks at the line segments that start at point j, that end at
 * all following points (ending at index rp).  The initial point
 * is on curve s0, the ending point is on curve s1.  The curve choice
 * (s.min vs. s.max) is based on the index in ss[].  The scan
 * looks for the largest (sign=0) or smallest (sign=1) slope.
 */
static int search(int rp, int j, int s0, int s1, int sign, struct _seg *answer)
{
	double dt, slope;
	int n, nextj=0, cinit=1;
	for (n=next_up(j); n!=next_up(rp); n=next_up(n)) {
		dt = d_ring[n].absolute - d_ring[j].absolute;
		slope = (d_ring[n].s.ss[s0] - d_ring[j].s.ss[s1]) / dt;
		if (cinit || (slope < answer->slope) ^ sign) {
			answer->slope = slope;
			answer->offset = d_ring[n].s.ss[s0] +
				slope*(d_ring[rp].absolute - d_ring[n].absolute);
			cinit = 0;
			nextj = n;
		}
	}
	return nextj;
}

static void polygon_reset(void)
{
	df.l_min = (MIN_INIT);
	df.r_min = (MIN_INIT);
}

static double find_df(int *flag)
{
	if (df.l_min == 0.0) {
		if (df.r_min == 0.0) {
			return 0.0;   /* every point was OK */
		} else {
			return -df.r_min;
		}
	} else {
		if (df.r_min == 0.0) {
			return df.l_min;
		} else {
			if (flag) *flag=1;
			return 0.0;   /* some points on each side,
				       * or no data at all */
		}
	}
}

/* Finds the amount of delta-f required to move a point onto a
 * target line in delta-f/delta-t phase space.  Any line is OK
 * as long as it's not convex and never returns greater than
 * MIN_INIT. */
static double find_shift(double slope, double offset)
{
	double shift  = slope - offset/600.0;
	double shift2 = slope + 0.3 - offset/6000.0;
	if (shift2 < shift) shift = shift2;
	if (shift  < 0) return 0.0;
	return shift;
}

static void polygon_point(struct _seg *s)
{
	double l, r;
	l = find_shift(- s->slope,   s->offset);
	r = find_shift(  s->slope, - s->offset);
	if (l < df.l_min) df.l_min = l;
	if (r < df.r_min) df.r_min = r;
}

/* Something like linear feedback to be used when we are "close" to
 * phase lock.  Not really used at the moment:  the logic in find_df()
 * never sets the flag. */
static double find_df_center(struct _seg *min, struct _seg *max, double gross_df)
{
	const double crit_time=1000.0;
	double slope  = 0.5 * (max->slope  + min->slope)+gross_df;
	double dslope =       (max->slope  - min->slope);
	double offset = 0.5 * (max->offset + min->offset);
	double doffset =      (max->offset - min->offset);
	double delta1 = -offset/ 600.0 - slope;
	double delta2 = -offset/1800.0 - slope;
	double delta  = 0.0;
	double factor = crit_time/(crit_time+doffset+dslope*1200.0);
	if (offset <  0 && delta2 > 0) delta = delta2;
	if (offset <  0 && delta1 < 0) delta = delta1;
	if (offset >= 0 && delta1 > 0) delta = delta1;
	if (offset >= 0 && delta2 < 0) delta = delta2;
	if (max->offset < -crit_time || min->offset > crit_time) return 0.0;
	return factor*delta;
}

static int contemplate_data(unsigned int absolute, double skew, double errorbar, int freq)
{
	/*  Here is the actual phase lock loop.
	 *  Need to keep a ring buffer of points to make a rational
	 *  decision how to proceed.
	 */
	int both_sides_now = 0;
	int j, n, c, max_avail, min_avail, dinit;
	int nextj = 0;  /* initialization not needed; but gcc can't figure out my logic */
	double cum;
	struct _seg check, save_min, save_max;
	double last_slope;
	int delta_freq;
	double delta_f;
	int inconsistent = 0, max_imax, max_imin = 0, min_imax, min_imin = 0;
	int computed_freq = freq;

	d_ring[d_ring_rp].absolute = absolute;
	d_ring[d_ring_rp].skew     = skew;
	d_ring[d_ring_rp].errorbar = errorbar - min_delay;   /* quick hack to speed things up */
	d_ring[d_ring_rp].freq     = freq;

	if (valid < RING_SIZE) ++valid;
	if (valid == RING_SIZE) {
		/*
		 * Pass 1: correct for wandering freq's */
		cum = 0.0;
		for (j=d_ring_rp; ; j=n) {
			d_ring[j].s.s.max = d_ring[j].skew - cum + d_ring[j].errorbar;
			d_ring[j].s.s.min = d_ring[j].skew - cum - d_ring[j].errorbar;
			n=next_dn(j);
			if (n == d_ring_rp) break;
			/* Assume the freq change took place immediately after
			 * the data was taken; this is valid for the case where
			 * this program was responsible for the change.
			 */
			cum = cum + (d_ring[j].absolute-d_ring[n].absolute) *
				(double)(d_ring[j].freq-freq)/65536;
		}
		/*
		 * Pass 2: find the convex down envelope of s.max, composed of
		 * line segments in s.max vs. absolute space, which are
		 * points in freq vs. dt space.  Find points in order of increasing
		 * slope == freq */
		dinit=1; last_slope=-2*(MAX_CORRECT);
		for (c=1, j=next_up(d_ring_rp); ; j=nextj) {
			nextj = search(d_ring_rp, j, 1, 1, 0, &maxseg[c]);
				search(d_ring_rp, j, 0, 1, 1, &check);
			if (check.slope < maxseg[c].slope && check.slope > last_slope &&
			    (dinit || check.slope < save_min.slope)) {dinit=0; save_min=check; }
			last_slope = maxseg[c].slope;
			c++;
			if (nextj == d_ring_rp) break;
		}
		if (dinit==1) inconsistent=1;
		max_avail=c;
		/*
		 * Pass 3: find the convex up envelope of s.min, composed of
		 * line segments in s.min vs. absolute space, which are
		 * points in freq vs. dt space.  These points are found in
		 * order of decreasing slope. */
		dinit=1; last_slope=+2*(MAX_CORRECT);
		for (c=1, j=next_up(d_ring_rp); ; j=nextj) {
			nextj = search(d_ring_rp, j, 0, 0, 1, &minseg[c]);
				search(d_ring_rp, j, 1, 0, 0, &check);
			if (check.slope > minseg[c].slope && check.slope < last_slope &&
			    (dinit || check.slope < save_max.slope)) {dinit=0; save_max=check; }
			last_slope = minseg[c].slope;
			c++;
			if (nextj == d_ring_rp) break;
		}
		if (dinit==1) inconsistent=1;
		min_avail=c;
		/*
		 * Pass 4: splice together the convex polygon that forms
		 * the envelope of slope/offset coordinates that are consistent
		 * with the observed data.  The order of calls to polygon_point
		 * doesn't matter for the frequency shift determination, but
		 * the order chosen is nice for visual display. */
		if (!inconsistent) {
		polygon_reset();
		polygon_point(&save_min);
		for (dinit=1, c=1; c<max_avail; c++) {
			if (dinit && maxseg[c].slope > save_min.slope) {
				max_imin = c-1;
				maxseg[max_imin] = save_min;
				dinit = 0;
			}
			if (maxseg[c].slope > save_max.slope)
				break;
			if (dinit==0) polygon_point(&maxseg[c]);
		}
		if (dinit) inconsistent=1;
		max_imax = c;
		maxseg[max_imax] = save_max;

		polygon_point(&save_max);
		for (dinit=1, c=1; c<min_avail; c++) {
			if (dinit && minseg[c].slope < save_max.slope) {
				max_imin = c-1;
				minseg[min_imin] = save_max;
				dinit = 0;
			}
			if (minseg[c].slope < save_min.slope)
				break;
			if (dinit==0) polygon_point(&minseg[c]);
		}
		if (dinit) inconsistent=1;
		min_imax = c;
		minseg[min_imax] = save_max;

		} /* !inconsistent */

		/*
		 * Pass 5: decide on a new freq */
		if (inconsistent) {
			printf("# inconsistent\n");
		} else {
			delta_f = find_df(&both_sides_now);
			delta_f += find_df_center(&save_min,&save_max, delta_f);
			delta_freq = delta_f*65536+.5;
			computed_freq -= delta_freq;
			printf ("# box [( %.3f , %.1f ) ",  save_min.slope, save_min.offset);
			printf (      " ( %.3f , %.1f )] ", save_max.slope, save_max.offset);
			printf (" delta_f %.3f  computed_freq %d\n", delta_f, computed_freq);

			if (computed_freq < -(MAX_C)) computed_freq=-(MAX_C);
			if (computed_freq >  (MAX_C)) computed_freq= (MAX_C);
		}
	}
	d_ring_rp = (d_ring_rp+1) % RING_SIZE;
	return computed_freq;
}

struct ntptime {
	unsigned int coarse;
	unsigned int fine;
};

static int get_current_freq(void)
{
	/* OS dependent routine to get the current value of clock frequency. */
#ifdef __linux__
	struct timex txc;
	txc.modes=0;
	if (adjtimex(&txc) < 0)
		bb_perror_msg_and_die("adjtimex");
	return txc.freq;
#else
	return 0;
#endif
}

static int set_freq(int new_freq)
{
	/* OS dependent routine to set a new value of clock frequency. */
#ifdef __linux__
	struct timex txc;
	txc.modes = ADJ_FREQUENCY;
	txc.freq = new_freq;
	if (adjtimex(&txc) < 0)
		bb_perror_msg_and_die("adjtimex");
	return txc.freq;
#else
	return 0;
#endif
}

static void set_time(struct ntptime *new)
{
	/* POSIX 1003.1-2001 way to set the system clock */
	struct timespec tv_set;

	/* it would be even better to subtract half the slop */
	tv_set.tv_sec  = new->coarse - JAN_1970;
	/* divide xmttime.fine by 4294.967296 */
	tv_set.tv_nsec = USEC(new->fine) * 1000;
	if (clock_settime(CLOCK_REALTIME, &tv_set)<0)
		bb_perror_msg_and_die("clock_settime");
}

static void ntpc_gettime(uint32_t *time_coarse, uint32_t *time_fine)
{
	/* POSIX 1003.1-2001 way to get the system time */
	struct timespec now;

	clock_gettime(CLOCK_REALTIME, &now);
	*time_coarse = now.tv_sec + JAN_1970;
	*time_fine   = NTPFRAC(now.tv_nsec/1000);
}

static void send_packet(int usd, uint32_t time_sent[2])
{
	uint32_t data[12];
#define LI 0
#define VN 3
#define MODE 3
#define STRATUM 0
#define POLL 4
#define PREC -6

	memset(data, 0, sizeof data);
	data[0] = htonl (
		( LI << 30 ) | ( VN << 27 ) | ( MODE << 24 ) |
		( STRATUM << 16) | ( POLL << 8 ) | ( PREC & 0xff ) );
	data[1] = htonl(1<<16);  /* Root Delay (seconds) */
	data[2] = htonl(1<<16);  /* Root Dispersion (seconds) */
	ntpc_gettime(time_sent, time_sent+1);
	data[10] = htonl(time_sent[0]); /* Transmit Timestamp coarse */
	data[11] = htonl(time_sent[1]); /* Transmit Timestamp fine   */
	send(usd,data,48,0);
}

static void get_packet_timestamp(int usd, struct ntptime *udp_arrival_ntp)
{
#ifdef PRECISION_SIOCGSTAMP
	struct timeval udp_arrival;
	if ( xioctl(usd, SIOCGSTAMP, &udp_arrival, "ioctl-SIOCGSTAMP") < 0 ) {
		ntpc_gettime(&udp_arrival_ntp->coarse, &udp_arrival_ntp->fine);
	} else {
		udp_arrival_ntp->coarse = udp_arrival.tv_sec + JAN_1970;
		udp_arrival_ntp->fine   = NTPFRAC(udp_arrival.tv_usec);
	}
#else
	(void) usd;  /* not used */
	ntpc_gettime(&udp_arrival_ntp->coarse, &udp_arrival_ntp->fine);
#endif
}

static int check_source(struct sockaddr_in *sa_in)
{
	if (NTP_PORT != ntohs(sa_in->sin_port))
		return 1;  /* fault */
	return 0;
}

static double ntpdiff( struct ntptime *start, struct ntptime *stop)
{
	int a = stop->coarse - start->coarse;
	unsigned int b;

	if (stop->fine >= start->fine) {
		b = stop->fine - start->fine;
	} else {
		b = start->fine - stop->fine;
		b = ~b;
		a -= 1;
	}

	return a*1.e6 + b * (1.e6/4294967296.0);
}

/* Does more than print, so this name is bogus.
 * It also makes time adjustments, both sudden (-s)
 * and phase-locking (-l).
 * sets *error to the number of microseconds uncertainty in answer
 * returns 0 normally, 1 if the message fails sanity checks
 */
static int rfc1305print(uint32_t *data, struct ntptime *arrival, int *error)
{
/* straight out of RFC-1305 Appendix A */
	int li, vn, mode, stratum, prec;
	int delay, disp;
	struct ntptime orgtime, rectime, xmttime;
	double el_time,st_time,skew1,skew2;
	int freq;

#define Data(i) ntohl(((uint32_t *)data)[i])
	li      = Data(0) >> 30 & 0x03;
	vn      = Data(0) >> 27 & 0x07;
	mode    = Data(0) >> 24 & 0x07;
	stratum = Data(0) >> 16 & 0xff;
	prec    = Data(0)       & 0xff;
	if (prec & 0x80) prec |= 0xffffff00;
	delay   = Data(1);
	disp    = Data(2);
	orgtime.coarse = Data(6);
	orgtime.fine   = Data(7);
	rectime.coarse = Data(8);
	rectime.fine   = Data(9);
	xmttime.coarse = Data(10);
	xmttime.fine   = Data(11);
#undef Data

	el_time = ntpdiff(&orgtime,arrival);   /* elapsed */
	st_time = ntpdiff(&rectime,&xmttime);  /* stall */
	skew1 = ntpdiff(&orgtime,&rectime);
	skew2 = ntpdiff(&xmttime,arrival);
	freq = get_current_freq();

	/* error checking, see RFC-4330 section 5 */
	if (bb_common_bss.cross_check) {
		if (li == 3) goto fail;  /* unsynchronized */
		if (vn < 3) goto fail;   /* RFC-4330 documents SNTP v4, but we interoperate with NTP v3 */
		if (mode != 4) goto fail;
		if (orgtime.coarse != bb_common_bss.time_of_send[0] ||
		    orgtime.fine   != bb_common_bss.time_of_send[1] ) goto fail;
		if (xmttime.coarse == 0 && xmttime.fine == 0) goto fail;
		if (delay > 65536 || delay < -65536) goto fail;
		if (disp  > 65536 || disp  < -65536) goto fail;
		if (stratum == 0) goto fail;  /* kiss o' death */
	}

	/* XXX should I do this if debug flag is set? */
	if (bb_common_bss.set_clock) { /* you'd better be root, or ntpclient will exit here! */
		set_time(&xmttime);
	}

	/* Not the ideal order for printing, but we want to be sure
	 * to do all the time-sensitive thinking (and time setting)
	 * before we start the output, especially fflush() (which
	 * could be slow).  Of course, if debug is turned on, speed
	 * has gone down the drain anyway. */
	if (bb_common_bss.live) {
		int new_freq = contemplate_data(arrival->coarse, (skew1-skew2)/2,
			el_time+sec2u(disp), freq);
		if (new_freq != freq) set_freq(new_freq);
	}
	printf("%d %.5d.%.3d  %8.1f %8.1f  %8.1f %8.1f %9d\n",
		arrival->coarse/86400, arrival->coarse%86400,
		arrival->fine/4294967, el_time, st_time,
		(skew1-skew2)/2, sec2u(disp), freq);
	fflush(stdout);
	*error = el_time-st_time;

	return 0;
fail:
	printf("%d %.5d.%.3d  rejected packet\n",
		arrival->coarse/86400, arrival->coarse%86400,
		arrival->fine/4294967);
	return 1;
}

static void setup_receive(int usd, unsigned int interface, short port)
{
	struct sockaddr_in sa_rcvr;

	memset(&sa_rcvr, 0, sizeof sa_rcvr);
	sa_rcvr.sin_family = AF_INET;
	sa_rcvr.sin_addr.s_addr = htonl(interface);
	sa_rcvr.sin_port = htons(port);
	bb_xbind(usd, (struct sockaddr *) &sa_rcvr, sizeof sa_rcvr);
}

static void setup_transmit(int usd, char *host, short port)
{
	struct sockaddr_in sa_dest;

	memset(&sa_dest, 0, sizeof sa_dest);
	sa_dest.sin_family = AF_INET;
	bb_gethostbyname(BB_GHBN_IN_ADDR|BB_GHBN_ERR_DIE, host, &(sa_dest.sin_addr));
	sa_dest.sin_port = htons(port);
	if (connect(usd, (struct sockaddr *)&sa_dest, sizeof sa_dest) == -1)
		bb_perror_msg_and_die("connect");
}

static void primary_loop(int usd)
{
	fd_set fds;
	struct sockaddr_in sa_xmit_in;
	int i, pack_len, probes_sent = 0, error;
	socklen_t sa_xmit_len;
	struct timeval to;
	struct ntptime udp_arrival_ntp;
#define incoming ((char *) incoming_word)
#define sizeof_incoming (sizeof incoming_word)

	sa_xmit_len = sizeof sa_xmit_in;
	to.tv_sec = 0;
	to.tv_usec = 0;
	for (;;) {
		FD_ZERO(&fds);
		FD_SET(usd, &fds);
		i = select(usd + 1, &fds, NULL, NULL, &to);  /* Wait on read or error */
		if ((i != 1) || (!FD_ISSET(usd, &fds))) {
			if (i < 0) {
				if (errno != EINTR) bb_perror_msg("select");
				continue;
			}
			if (to.tv_sec == 0) {
				if (probes_sent >= bb_common_bss.probe_count && bb_common_bss.probe_count != 0)
					break;
				send_packet(usd, bb_common_bss.time_of_send);
				++probes_sent;
				to.tv_sec = bb_common_bss.cycle_time;
				to.tv_usec = 0;
			}
			continue;
		}
		pack_len = recvfrom(usd,incoming,sizeof_incoming,0,
				  (struct sockaddr *) &sa_xmit_in,&sa_xmit_len);
		error = bb_common_bss.goodness;
		if (pack_len < 0) {
			bb_perror_msg("recvfrom");
		} else if (pack_len > 0 && (unsigned)pack_len < sizeof_incoming) {
			get_packet_timestamp(usd, &udp_arrival_ntp);
			if (check_source(&sa_xmit_in)!=0) continue;
			if (rfc1305print(incoming_word, &udp_arrival_ntp, &error)!=0) continue;
			/* udp_handle(usd,incoming,pack_len,&sa_xmit,sa_xmit_len); */
		} else {
			printf("Ooops.  pack_len=%d\n", pack_len);
			fflush(stdout);
		}
		/* best rollover option: specify -g, -s, and -l.
		 * simpler rollover option: specify -s and -l, which
		 * triggers a magic -c 1 */
		if ((error < bb_common_bss.goodness && bb_common_bss.goodness != 0) ||
		    (probes_sent >= bb_common_bss.probe_count && bb_common_bss.probe_count != 0)) {
			bb_common_bss.set_clock = 0;
			if (!bb_common_bss.live) break;
		}
	}
#undef incoming
#undef sizeof_incoming
}

int ntpclient_main(void)
{
	int usd;  /* socket */
	int c;
	/* These parameters are settable from the command line
	   the initializations here provide default behavior */
	short int udp_local_port = 0;   /* default of 0 means kernel chooses */
	char *hostname = NULL;          /* must be set */
	int initial_freq = 0;           /* initial freq value to use */
	char *s_min_delay = NULL;

	bb_common_bss.cycle_time = 600;          /* seconds */

	c = bb_xgetopt_ulflags("stlf:c:g:h:i:p:q:\0:-fcgip",
		&initial_freq, &bb_common_bss.probe_count, &bb_common_bss.goodness, &hostname,
		&bb_common_bss.cycle_time, &udp_local_port, &s_min_delay);
	bb_common_bss.set_clock = c & 1;
	bb_common_bss.cross_check = (c & 2) ^ 2;        /* reverse */
	bb_common_bss.live = c & 4;
	if ((c & 8))
		set_freq(initial_freq);
	min_delay = s_min_delay != NULL ? atof(s_min_delay) : 800.0;
	if (hostname == NULL)
		bb_show_usage();

	if (bb_common_bss.set_clock && !bb_common_bss.live && !bb_common_bss.goodness && !bb_common_bss.probe_count)
		bb_common_bss.probe_count = 1;

	/* respect only applicable MUST of RFC-4330 */
	if (bb_common_bss.probe_count != 1 && bb_common_bss.cycle_time < MIN_INTERVAL)
		bb_common_bss.cycle_time = MIN_INTERVAL;

	/* Startup sequence */
	usd = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM|BB_XSOCKET_FLG_ERR_DIE|BB_XSOCKET_FLG_PROTO_SPEC, IPPROTO_UDP);

	setup_receive(usd, INADDR_ANY, udp_local_port);

	setup_transmit(usd, hostname, NTP_PORT);

	primary_loop(usd);

	close(usd);
	return 0;
}
#endif /* APPLET_USAGE */
