/*
 *                      P I N G . C
 *
 * Using the InterNet Control Message Protocol (ICMP) "ECHO" facility,
 * measure round-trip-delays and packet loss across network paths.
 *
 * Author -
 *      Mike Muuss
 *      U. S. Army Ballistic Research Laboratory
 *      December, 1983
 *
 * Status -
 *      Public Domain.  Distribution Unlimited.
 * Bugs -
 *      More statistics could always be gathered.
 *      This program has to run SUID to ROOT to access the ICMP socket.
 *
 * Modified by Vladimir Oleynik <dzo@simtreas.ru> (c) 2018 to be
 * used in busybox
 *
 */

#ifdef APPLET_USAGE
#define ping_trivial_usage \
		"[-aAbBDfLnOqrRUvV] [-c count] [-i intrvl] [-I interf]" \
		" [-m mark] [-M pmtudisc] [-l preload] [-p pattern]" \
		" [-Q tos] [-s pktsz] [-S sndbuf] [-t ttl]" \
		" [-T timestamp] [-w dl] [-W timeout] [hop1 ...] dest"
#define ping_full_usage \
	"send ICMP ECHO_REQUEST to network hosts.\n\n" \
	"Options:\n" \
	"\t-a\t\tAudible ping.\n" \
	"\t-A\t\tAdaptive ping.\n" \
	"\t-b\t\tAllow pinging a broadcast address.\n" \
	"\t-B\t\tDo not allow ping to change source address of probes.\n" \
	"\t-c count\tStop after sending count ECHO_REQUEST packets.\n" \
	"\t-D\t\tPrint timestamp\n" \
	"\t-f\t\tFlood ping.\n" \
	"\t-i intrvl\tWait interval seconds between sending each packet.\n" \
	"\t-l preload\tSends that many packets not waiting for reply.\n" \
	"\t-L\t\tSuppress  loopback  of  multicast  packets.\n" \
	"\t-m mark\t\tUse mark to tag the packets going out.\n" \
	"\t-M pmtudisc\tSelect Path MTU Discovery strategy: do/want/dont.\n" \
	"\t-n\t\tNumeric output only.\n" \
	"\t-O\t\tReport outstanding ICMP ECHO reply before sending next packet.\n" \
	"\t-p pattern\tSpecify up to 16 pad bytes to fill out the packet send.\n" \
	"\t-q\t\tQuiet output.\n" \
	"\t-Q tos\t\tSet Quality of Service -related bits in ICMP datagrams.\n" \
	"\t-r\t\tBypass the normal routing tables and send directly.\n" \
	"\t-R\t\tRecord route.\n" \
	"\t-s packetsize\tSpecifies the number of data bytes to be sent.\n" \
	"\t-S sndbuf\tSet socket sndbuf.\n" \
	"\t-t ttl\t\tSet the IP Time to Live.\n" \
	"\t-v\t\tVerbose output.\n" \
	"\t-w deadline\tSpecify a timeout, in seconds.\n" \
	"\t-W timeout\tTime to wait for a response, in seconds."
#define ping6_trivial_usage \
		"[-aAbBDfFLnOqrUvV] [-c c] [-F flwlbl] [-i intrvl] [-I interf]" \
		" [-m mark] [-M pmtudisc] [-l preload] [-p pattern]" \
		" [-Q tos] [-s packetsize] [-S sndbuf]" \
		" [-T timestamp] [-w dl] [-W timeout] [hop1 ...] dest"
#define ping6_full_usage \
	"send ICMP ECHO_REQUEST to network hosts.\n\n" \
	"Options:\n" \
	"\t-a\t\tAudible ping.\n" \
	"\t-A\t\tAdaptive ping.\n" \
	"\t-b\t\tAllow pinging a broadcast address.\n" \
	"\t-B\t\tDo not allow ping to change source address of probes.\n" \
	"\t-c count\tStop after sending count ECHO_REQUEST packets.\n" \
	"\t-D\t\tPrint timestamp\n" \
	"\t-f\t\tFlood ping.\n" \
	"\t-F flow_label\tSet 20 bit flow label (hex) on echo request packets, zero - random.\n" \
	"\t-i intrvl\tWait interval seconds between sending each packet.\n" \
	"\t-l preload\tSends that many packets not waiting for reply.\n" \
	"\t-L\t\tSuppress  loopback  of  multicast  packets.\n" \
	"\t-m mark\t\tUse mark to tag the packets going out.\n" \
	"\t-M pmtudisc\tSelect Path MTU Discovery strategy: do/want/dont.\n" \
	"\t-n\t\tNumeric output only.\n" \
	"\t-O\t\tReport outstanding ICMP ECHO reply before sending next packet.\n" \
	"\t-p pattern\tSpecify up to 16 pad bytes to fill out the packet send.\n" \
	"\t-q\t\tQuiet output.\n" \
	"\t-Q tos\t\tSet Quality of Service -related bits in ICMP datagrams.\n" \
	"\t-r\t\tBypass the normal routing tables and send directly.\n" \
	"\t-s packetsize\tSpecifies the number of data bytes to be sent.\n" \
	"\t-S sndbuf\tSet socket sndbuf.\n" \
	"\t-v\t\tVerbose output.\n" \
	"\t-w deadline\tSpecify a timeout, in seconds.\n" \
	"\t-W timeout\tTime to wait for a response, in seconds."
#else

static const char SNAPSHOT[] = "s20121221";

#include "busybox.h"
#if defined(__MUSL__)
# include <linux/param.h>       /* HZ */
#endif
#include <unistd.h>
#include <time.h>
#include <linux/sockios.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <poll.h>
#include <ctype.h>
#include <string.h>
#include <setjmp.h>
#include <ctype.h>
#include <sched.h>
#include <math.h>
#include "bb_sig.h"
#include "bb_inet.h"

#ifdef CAPABILITIES
#include <sys/prctl.h>
#include <sys/capability.h>
#endif

#ifdef USE_IDN
#include <locale.h>
#include <idna.h>
#endif

#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/types.h>
#include <linux/errqueue.h>
#include <linux/filter.h>

#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#ifndef WITHOUT_IFADDRS
#include <ifaddrs.h>
#endif

#ifdef CONFIG_PING6
#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#endif

#define DEFDATALEN      (64 - 8)        /* default data length */

#define MAXWAIT         10              /* max seconds to wait for response */
#define MININTERVAL     10              /* Minimal interpacket gap */
#define MINUSERINTERVAL 200             /* Minimal allowed interval for non-root */

#define SCHINT(a)       (((a) <= MININTERVAL) ? MININTERVAL : (a))

#define F_FLOOD         0x001
#define F_INTERVAL      0x002
#define F_NUMERIC       0x004
#define F_PINGFILLED    0x008
#define F_QUIET         0x010
#define F_RROUTE        0x020
#define F_SO_DEBUG      0x040
#define F_SO_DONTROUTE  0x080
#define F_VERBOSE       0x100
#define F_TIMESTAMP     0x200
#define F_FLOWINFO      0x200
#define F_SOURCEROUTE   0x400
#define F_TCLASS        0x400
#define F_FLOOD_POLL    0x800
#define F_LATENCY       0x1000
#define F_AUDIBLE       0x2000
#define F_ADAPTIVE      0x4000
#define F_STRICTSOURCE  0x8000
#define F_NOLOOP        0x10000
#define F_TTL           0x20000
#define F_MARK          0x40000
#define F_PTIMEOFDAY    0x80000
#define F_OUTSTANDING   0x100000

/*
 * MAX_DUP_CHK is the number of bits in received table, i.e. the maximum
 * number of received sequence numbers we can keep track of.
 */
#define MAX_DUP_CHK     0x10000

#if defined(__WORDSIZE) && __WORDSIZE == 64
# define USE_BITMAP64
#endif

#ifdef USE_BITMAP64
typedef __u64   bitmap_t;
# define BITMAP_SHIFT   6
#else
typedef __u32   bitmap_t;
# define BITMAP_SHIFT   5
#endif

#if ((MAX_DUP_CHK >> (BITMAP_SHIFT + 3)) << (BITMAP_SHIFT + 3)) != MAX_DUP_CHK
# error Please MAX_DUP_CHK and/or BITMAP_SHIFT
#endif

struct rcvd_table {
	bitmap_t bitmap[MAX_DUP_CHK / (sizeof(bitmap_t) * 8)];
};



#ifndef MSG_CONFIRM
#define MSG_CONFIRM 0
#endif

typedef void (*pf_install_filter)(void);
typedef int (*pf_int__void)(void);
typedef int (*pf_parse_reply)(struct msghdr *msg, int cc, void *addr, struct timeval *tv);

#ifdef CONFIG_PING6
#define INSNS_PING6_ST_NUM 7
static const struct sock_filter insns_ping6_st[INSNS_PING6_ST_NUM] = {
	BPF_STMT(BPF_LD|BPF_H|BPF_ABS, 4),  /* Load icmp echo ident */
	BPF_JUMP(BPF_JMP|BPF_JEQ|BPF_K, 0xAAAA, 0, 1),  /* Ours? */
	BPF_STMT(BPF_RET|BPF_K, ~0U),  /* Yes, it passes. */
	BPF_STMT(BPF_LD|BPF_B|BPF_ABS, 0),  /* Load icmp type */
	BPF_JUMP(BPF_JMP|BPF_JEQ|BPF_K, ICMP6_ECHO_REPLY, 1, 0), /* Echo? */
	BPF_STMT(BPF_RET|BPF_K, ~0U), /* No. It passes. This must not happen. */
	BPF_STMT(BPF_RET|BPF_K, 0), /* Echo with wrong ident. Reject. */
};
#endif

#ifdef CONFIG_PING
#define INSNS_PING_ST_NUM 8
static const struct sock_filter insns_ping_st[INSNS_PING_ST_NUM] = {
	BPF_STMT(BPF_LDX|BPF_B|BPF_MSH, 0), /* Skip IP header. F..g BSD... Look into ping6. */
	BPF_STMT(BPF_LD|BPF_H|BPF_IND, 4), /* Load icmp echo ident */
	BPF_JUMP(BPF_JMP|BPF_JEQ|BPF_K, 0xAAAA, 0, 1), /* Ours? */
	BPF_STMT(BPF_RET|BPF_K, ~0U), /* Yes, it passes. */
	BPF_STMT(BPF_LD|BPF_B|BPF_IND, 0), /* Load icmp type */
	BPF_JUMP(BPF_JMP|BPF_JEQ|BPF_K, ICMP_ECHOREPLY, 1, 0), /* Echo? */
	BPF_STMT(BPF_RET|BPF_K, 0xFFFFFFF), /* No. It passes. */
	BPF_STMT(BPF_RET|BPF_K, 0) /* Echo with wrong ident. Reject. */
};
struct G_CMSG {
	struct cmsghdr cm;
	struct in_pktinfo ipi;
};

#endif

CALC_BSS struct BB_COMMON_BSS {
	int g_options; /* various options */

	int mark;
	int sndbuf;
	int g_ttl;
	int rtt;
	int rtt_addend;

	char *device;
	int icmp_sock;                  /* socket file descriptor */
	int maxpacket;
	pf_install_filter install_filter;

#define MAXPACKET       128000          /* max packet size */
	u_char outpack[MAXPACKET];

	int pmtudisc;
	pf_int__void receive_error_msg;
	pf_int__void send_probe;
	pf_parse_reply parse_reply;

	/* counters */
	long npackets;           /* max packets to transmit */
	long nreceived;          /* # of packets we got back */
	long nrepeats;           /* number of duplicates */
	long ntransmitted;       /* sequence # for outbound packets = #sent */
	long nchecksum;          /* replies with bad checksum */
	long nerrors;            /* icmp errors */
	int interval;            /* interval between packets (msec) */
	int preload;
	int deadline;            /* time to die */
	int lingertime;
	struct timeval start_time, cur_time;
	volatile int exiting;
	volatile int status_snapshot;
	int confirm;
	volatile int in_pr_addr;    /* pr_addr() is executing */
	jmp_buf pr_addr_jmp;

	/* Stupid workarounds for bugs/missing functionality in older linuces.
	 * confirm_flag fixes refusing service of kernels without MSG_CONFIRM.
	 * i.e. for linux-2.2 */
	int confirm_flag;
	/* And this is workaround for bug in IP_RECVERR on raw sockets which is present
	 * in linux-2.2.[0-19], linux-2.4.[0-7] */
	int working_recverr;

	int timing;                      /* flag to do timing */
	__u16 acked;
	long tmin;           /* minimum round trip time */
	long tmax;                      /* maximum round trip time */
	/* Message for rpm maintainers: have _shame_. If you want
	 * to fix something send the patch to me for sanity checking.
	 * "sparcfix" patch is a complete non-sense, apparenly the person
	 * prepared it was stoned.
	 */
	long long tsum;                 /* sum of all times, for doing average */
	long long tsum2;
	int  pipesize;

	int datalen;

	char *hostname;
	int uid;
	uid_t euid;
	int ident;                      /* process id to identify our packets */

	int screen_width;

#ifdef CONFIG_PING
	struct sock_filter insns_ping[INSNS_PING_ST_NUM];
	struct sock_fprog filter_ping;
	char pr_addr_buf[4096];
	struct sockaddr_in whereto;     /* who to ping */
	int ping_cmsg_len;
	int old_rrlen;
	char old_rr[MAX_IPOPTLEN];
	struct G_CMSG g_cmsg;
	int once;
#endif

#ifdef CONFIG_PING6
	struct sock_filter insns_ping6[INSNS_PING6_ST_NUM];
	struct sock_fprog filter_ping6;
	char *g_pr_addr6_s;
	struct sockaddr_in6 whereto6;    /* who to ping */
	unsigned char cmsgbuf[4096];
	int cmsglen;
#endif

	unsigned long waittime;
	struct rcvd_table rcvd_tbl;
	int oom_count;
	int tokens;
} bb_common_bss;

#define g_options bb_common_bss.g_options
#define mark bb_common_bss.mark
#define sndbuf bb_common_bss.sndbuf
#define g_ttl bb_common_bss.g_ttl
#define rtt bb_common_bss.rtt
#define rtt_addend bb_common_bss.rtt_addend
#define device bb_common_bss.device
#define icmp_sock bb_common_bss.icmp_sock
#define maxpacket bb_common_bss.maxpacket
#define install_filter bb_common_bss.install_filter
#define outpack bb_common_bss.outpack
#define pmtudisc bb_common_bss.pmtudisc
#define receive_error_msg bb_common_bss.receive_error_msg
#define send_probe bb_common_bss.send_probe
#define parse_reply bb_common_bss.parse_reply
#define npackets bb_common_bss.npackets
#define nreceived bb_common_bss.nreceived
#define nrepeats bb_common_bss.nrepeats
#define ntransmitted bb_common_bss.ntransmitted
#define nchecksum bb_common_bss.nchecksum
#define nerrors bb_common_bss.nerrors
#define interval bb_common_bss.interval
#define preload bb_common_bss.preload
#define deadline bb_common_bss.deadline
#define lingertime bb_common_bss.lingertime
#define start_time bb_common_bss.start_time
#define cur_time bb_common_bss.cur_time
#define exiting bb_common_bss.exiting
#define status_snapshot bb_common_bss.status_snapshot
#define confirm bb_common_bss.confirm
#define in_pr_addr bb_common_bss.in_pr_addr
#define pr_addr_jmp bb_common_bss.pr_addr_jmp
#define confirm_flag bb_common_bss.confirm_flag
#define working_recverr bb_common_bss.working_recverr
#define timing bb_common_bss.timing
#define acked bb_common_bss.acked
#define tmin bb_common_bss.tmin
#define tmax bb_common_bss.tmax
#define tsum bb_common_bss.tsum
#define tsum2 bb_common_bss.tsum2
#define pipesize bb_common_bss.pipesize
#define datalen bb_common_bss.datalen
#define hostname bb_common_bss.hostname
#define uid bb_common_bss.uid
#define euid bb_common_bss.euid
#define ident bb_common_bss.ident
#define screen_width bb_common_bss.screen_width
#define insns_ping bb_common_bss.insns_ping
#define filter_ping bb_common_bss.filter_ping
#define insns_ping6 bb_common_bss.insns_ping6
#define filter_ping6 bb_common_bss.filter_ping6
#define pr_addr_buf bb_common_bss.pr_addr_buf
#define whereto bb_common_bss.whereto
#define ping_cmsg_len bb_common_bss.ping_cmsg_len
#define cmsgbuf bb_common_bss.cmsgbuf
#define cmsglen bb_common_bss.cmsglen
#define waittime bb_common_bss.waittime
#define old_rrlen bb_common_bss.old_rrlen
#define old_rr bb_common_bss.old_rr
#define g_cmsg bb_common_bss.g_cmsg
#define g_pr_addr6_s bb_common_bss.g_pr_addr6_s
#define whereto6 bb_common_bss.whereto6
#define rcvd_tbl bb_common_bss.rcvd_tbl
#define oom_count bb_common_bss.oom_count
#define tokens bb_common_bss.tokens
#define once bb_common_bss.once

#define A(bit)  (rcvd_tbl.bitmap[(bit) >> BITMAP_SHIFT])        /* identify word in array */
#define B(bit)  (((bitmap_t)1) << ((bit) & ((1 << BITMAP_SHIFT) - 1)))  /* identify bit in word */

static inline void rcvd_set(__u16 seq)
{
	unsigned bit = seq % MAX_DUP_CHK;
	A(bit) |= B(bit);
}

static inline void rcvd_clear(__u16 seq)
{
	unsigned bit = seq % MAX_DUP_CHK;
	A(bit) &= ~B(bit);
}

static inline bitmap_t rcvd_test(__u16 seq)
{
	unsigned bit = seq % MAX_DUP_CHK;
	return A(bit) & B(bit);
}

static void common_data_init(void)
{
	maxpacket = sizeof(outpack);
	confirm_flag = MSG_CONFIRM;
	tmin = LONG_MAX;
	pipesize = -1;
	datalen = DEFDATALEN;
	pmtudisc = -1;
	interval = 1000;
	lingertime = MAXWAIT*1000;
	screen_width = INT_MAX;
	/* for any error, 1 - do not reply, 0 - full Ok */
	bb_default_error_retval = 2;
}

#define COMMON_OPTIONS \
case 'a': case 'U': case 'c': case 'd': \
case 'f': case 'i': case 'w': case 'l': \
case 'S': case 'n': case 'p': case 'q': \
case 'r': case 's': case 'v': case 'L': \
case 't': case 'A': case 'W': case 'B': case 'm': \
case 'D': case 'O':

#define COMMON_OPTSTR "h?VQ:I:M:aUc:fi:w:l:S:np:qrs:vLt:AW:Bm:DO"

/* Write to stdout */
static inline void write_stdout(const char *str, size_t len)
{
	bb_full_write(STDOUT_FILENO, str, len);
}

/*
 * tvsub --
 *      Subtract 2 timeval structs:  out = out - in.  Out is assumed to
 * be >= in.
 */
static inline void tvsub(struct timeval *out, struct timeval *in)
{
	if ((out->tv_usec -= in->tv_usec) < 0) {
		--out->tv_sec;
		out->tv_usec += 1000000;
	}
	out->tv_sec -= in->tv_sec;
}

static int next_schedule_exit(int next)
{
	struct itimerval it;

	if (waittime)
		return next;

	if (nreceived) {
		waittime = 2 * tmax;
		if (waittime < 1000UL*interval)
			waittime = 1000UL*interval;
	} else
		waittime = 1000UL*lingertime;

	if (next < 0 || (unsigned long)next < waittime/1000)
		next = waittime/1000;

	it.it_interval.tv_sec = 0;
	it.it_interval.tv_usec = 0;
	it.it_value.tv_sec = waittime/1000000;
	it.it_value.tv_usec = waittime%1000000;
	setitimer(ITIMER_REAL, &it, NULL);
	return next;
}

static inline int schedule_exit(int next)
{
	if (npackets && ntransmitted >= npackets && !deadline)
		next = next_schedule_exit(next);
	return next;
}

static inline int in_flight(void)
{
	__u16 diff = (__u16)ntransmitted - acked;
	return (diff<=0x7FFF) ? diff : ntransmitted-nreceived-nerrors;
}

static inline void acknowledge(__u16 seq)
{
	__u16 diff = (__u16)ntransmitted - seq;
	if (diff <= 0x7FFF) {
		if ((int)diff+1 > pipesize)
			pipesize = (int)diff+1;
		if ((__s16)(seq - acked) > 0 ||
		    (__u16)ntransmitted - acked > 0x7FFF)
			acked = seq;
	}
}

static inline void advance_ntransmitted(void)
{
	ntransmitted++;
	/* Invalidate acked, if 16 bit seq overflows. */
	if ((__u16)ntransmitted - acked > 0x7FFF)
		acked = (__u16)ntransmitted + 1;
}

#ifdef CAPABILITIES
static int modify_capability(cap_value_t, cap_flag_value_t);
static inline int enable_capability_raw(void)           { return modify_capability(CAP_NET_RAW,   CAP_SET);   };
static inline int disable_capability_raw(void)          { return modify_capability(CAP_NET_RAW,   CAP_CLEAR); };
static inline int enable_capability_admin(void)         { return modify_capability(CAP_NET_ADMIN, CAP_SET);   };
static inline int disable_capability_admin(void)        { return modify_capability(CAP_NET_ADMIN, CAP_CLEAR); };
#else
static int modify_capability(int);
static inline int enable_capability_raw(void)           { return modify_capability(1); };
static inline int disable_capability_raw(void)          { return modify_capability(0); };
static inline int enable_capability_admin(void)         { return modify_capability(1); };
static inline int disable_capability_admin(void)        { return modify_capability(0); };
#endif

static void main_loop(__u8 *buf, int buflen) __attribute__((noreturn));
static  void finish(void) __attribute__((noreturn));


#define ARRAY_SIZE(a)   (sizeof(a) / sizeof(a[0]))

#ifdef CAPABILITIES
static cap_value_t cap_raw = CAP_NET_RAW;
static cap_value_t cap_admin = CAP_NET_ADMIN;
#endif

static void limit_capabilities(void)
{
#ifdef CAPABILITIES
	cap_t cap_cur_p;
	cap_t cap_p;
	cap_flag_value_t cap_ok;

	cap_cur_p = cap_get_proc();
	if (!cap_cur_p)
		bb_perror_msg_and_die("cap_get_proc");

	cap_p = cap_init();
	if (!cap_p)
		bb_perror_msg_and_die("cap_init");

	cap_ok = CAP_CLEAR;
	cap_get_flag(cap_cur_p, CAP_NET_ADMIN, CAP_PERMITTED, &cap_ok);

	if (cap_ok != CAP_CLEAR)
		cap_set_flag(cap_p, CAP_PERMITTED, 1, &cap_admin, CAP_SET);

	cap_ok = CAP_CLEAR;
	cap_get_flag(cap_cur_p, CAP_NET_RAW, CAP_PERMITTED, &cap_ok);

	if (cap_ok != CAP_CLEAR)
		cap_set_flag(cap_p, CAP_PERMITTED, 1, &cap_raw, CAP_SET);

	if (cap_set_proc(cap_p) < 0)
		bb_perror_msg_and_die("cap_set_proc");

	if (prctl(PR_SET_KEEPCAPS, 1) < 0)
		bb_perror_msg_and_die("prctl");

	if (setuid(getuid()) < 0)
		bb_perror_msg_and_die("setuid");

	if (prctl(PR_SET_KEEPCAPS, 0) < 0)
		bb_perror_msg_and_die("prctl");

	cap_free(cap_p);
	cap_free(cap_cur_p);
#endif
	uid = getuid();
	euid = geteuid();
#ifndef CAPABILITIES
	if (seteuid(uid))
		bb_perror_msg_and_die("setuid");
#endif
}

#ifdef CAPABILITIES
static int modify_capability(cap_value_t cap, cap_flag_value_t on)
{
	cap_t cap_p = cap_get_proc();
	cap_flag_value_t cap_ok;
	int rc = -1;

	if (!cap_p) {
		bb_perror_msg("cap_get_proc");
		goto out;
	}

	cap_ok = CAP_CLEAR;
	cap_get_flag(cap_p, cap, CAP_PERMITTED, &cap_ok);
	if (cap_ok == CAP_CLEAR) {
		rc = on ? -1 : 0;
		goto out;
	}

	cap_set_flag(cap_p, CAP_EFFECTIVE, 1, &cap, on);

	if (cap_set_proc(cap_p) < 0) {
		bb_perror_msg("cap_set_proc");
		goto out;
	}

	cap_free(cap_p);

	rc = 0;
out:
	if (cap_p)
		cap_free(cap_p);
	return rc;
}
#else
static int modify_capability(int on)
{
	if (seteuid(on ? euid : getuid())) {
		bb_perror_msg("seteuid");
		return -1;
	}

	return 0;
}
#endif

static void drop_capabilities(void)
{
#ifdef CAPABILITIES
	cap_t cap = cap_init();
	if (cap_set_proc(cap) < 0)
		bb_perror_msg_and_die("cap_set_proc");
	cap_free(cap);
#else
	if (setuid(getuid()))
		bb_perror_msg_and_die("setuid");
#endif
}

/* Fills all the outpack, excluding ICMP header, but _including_
 * timestamp area with supplied pattern.
 */
static void fill(char *patp)
{
	int ii, jj, kk;
	int pat[16];
	char *cp;
	u_char *bp = outpack+8;

#ifdef USE_IDN
	setlocale(LC_ALL, "C");
#endif

	for (cp = patp; *cp; cp++) {
		if (!isxdigit(*cp))
			bb_error_msg_and_die("patterns must be specified as hex digits.");
	}
	ii = sscanf(patp, "%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x",
	    &pat[0], &pat[1], &pat[2], &pat[3], &pat[4], &pat[5], &pat[6],
	    &pat[7], &pat[8], &pat[9], &pat[10], &pat[11], &pat[12],
	    &pat[13], &pat[14], &pat[15]);

	if (ii > 0) {
		for (kk = 0; kk <= maxpacket - (8 + ii); kk += ii)
			for (jj = 0; jj < ii; ++jj)
				bp[jj + kk] = pat[jj];
	}
	if (!(g_options & F_QUIET)) {
		printf("PATTERN: 0x");
		for (jj = 0; jj < ii; ++jj)
			printf("%02x", bp[jj] & 0xFF);
		printf("\n");
	}

#ifdef USE_IDN
	setlocale(LC_ALL, "");
#endif
}

static void common_options(int ch)
{
	switch(ch) {
	case 'a':
		g_options |= F_AUDIBLE;
		break;
	case 'A':
		g_options |= F_ADAPTIVE;
		break;
	case 'c':
		npackets = bb_xparse_l_number(10 | BB_PN_ERR_STR1 | BB_PN_SET_LOWER1,
						optarg, "bad number of packets to transmit.");
		break;
	case 'D':
		g_options |= F_PTIMEOFDAY;
		break;
	case 'i':               /* wait between sending packets */
	{
		double dbl;

		if (safe_strtod(optarg, &dbl) || dbl < 0.0 || dbl >= (double)INT_MAX / 1000 - 1.0)
			bb_error_msg_and_die("bad timing interval");
		interval = (int)(dbl * 1000);
		g_options |= F_INTERVAL;
		break;
	}
	case 'm':
		mark = bb_xparse_i_number(10 | BB_PN_ERR_STR1, optarg, "'%s' bad mark.");
		g_options |= F_MARK;
		break;
	case 'w':
		deadline = bb_xparse_i_number(10 | BB_PN_ERR_STR1, optarg, "'%s' wait time.");
		break;
	case 'l':
		preload = bb_xparse_i_number(10 | BB_PN_SET_LOWER1 | BB_PN_SET_UPPER | BB_PN_ERR_STRV,
						optarg, MAX_DUP_CHK,
						"bad preload value, should be 1..%d", MAX_DUP_CHK);
		if (uid && preload > 3)
			bb_error_msg_and_die("cannot set preload to value > 3");
		break;
	case 'O':
		g_options |= F_OUTSTANDING;
		break;
	case 'S':
		sndbuf = bb_xparse_i_number(10 | BB_PN_ERR_STR1 | BB_PN_SET_LOWER1, optarg, "bad sndbuf value.");
		break;
	case 'f':
		g_options |= F_FLOOD;
		setbuf(stdout, (char *)NULL);
		/* fallthrough to numeric - avoid gethostbyaddr during flood */
	case 'n':
		g_options |= F_NUMERIC;
		break;
	case 'p':               /* fill buffer with user pattern */
		g_options |= F_PINGFILLED;
		fill(optarg);
		break;
	case 'q':
		g_options |= F_QUIET;
		break;
	case 'r':
		g_options |= F_SO_DONTROUTE;
		break;
	case 's':               /* size of packet to send */
		datalen = bb_xparse_i_number(10 | BB_PN_ERR_STR1, optarg, "'%s' illegal packet size.");
		if (datalen > maxpacket - 8)
			bb_error_msg_and_die("packet size too large: %d", datalen);
		break;
	case 'v':
		g_options |= F_VERBOSE;
		break;
	case 'L':
		g_options |= F_NOLOOP;
		break;
	case 't':
		g_options |= F_TTL;
		g_ttl = bb_xparse_i_number(10 | BB_PN_ERR_STR1 | BB_PN_SET_UPPER,
						optarg, 255, "ttl %s out of range");
		break;
	case 'U':
		g_options |= F_LATENCY;
		break;
	case 'B':
		g_options |= F_STRICTSOURCE;
		break;
	case 'W':
		lingertime = bb_xparse_i_number(10 | BB_PN_ERR_STR1 | BB_PN_SET_UPPER,
					optarg, (int)(INT_MAX/1000000), "bad linger time.");
		lingertime *= 1000;
		break;
	case 'V':
		printf("%s utility, iputils-%s\n", bb_applet_name, SNAPSHOT);
		exit(0);
	default:
		abort();
	}
}

static void sigexit(int signo)
{
	exiting = signo;
	if (in_pr_addr)
		longjmp(pr_addr_jmp, 0);
}

static void sigstatus(int signo)
{
	status_snapshot = signo;
}



static inline void update_interval(void)
{
	int est = rtt ? rtt/8 : interval*1000;

	interval = (est+rtt_addend+500)/1000;
	if (uid && interval < MINUSERINTERVAL)
		interval = MINUSERINTERVAL;
}

/*
 * Print timestamp
 */
static void print_timestamp(void)
{
	if (g_options & F_PTIMEOFDAY) {
		struct timeval tv;
		gettimeofday(&tv, NULL);
		printf("[%lu.%06lu] ",
		       (unsigned long)tv.tv_sec, (unsigned long)tv.tv_usec);
	}
}

/*
 * pinger --
 *      Compose and transmit an ICMP ECHO REQUEST packet.  The IP packet
 * will be added on by the kernel.  The ID field is our UNIX process ID,
 * and the sequence number is an ascending integer.  The first 8 bytes
 * of the data portion are used to hold a UNIX "timeval" struct in VAX
 * byte-order, to compute the round-trip time.
 */
static int pinger(void)
{
	int i;

	/* Have we already sent enough? If we have, return an arbitrary positive value. */
	if (exiting || (npackets && ntransmitted >= npackets && !deadline))
		return 1000;

	/* Check that packets < rate*time + preload */
	if (cur_time.tv_sec == 0) {
		gettimeofday(&cur_time, NULL);
		tokens = interval*(preload-1);
	} else {
		long ntokens;
		struct timeval tv;

		gettimeofday(&tv, NULL);
		ntokens = (tv.tv_sec - cur_time.tv_sec)*1000 +
			(tv.tv_usec-cur_time.tv_usec)/1000;
		if (!interval) {
			/* Case of unlimited flood is special;
			 * if we see no reply, they are limited to 100pps */
			if (ntokens < MININTERVAL && in_flight() >= preload)
				return MININTERVAL-ntokens;
		}
		ntokens += tokens;
		if (ntokens > interval*preload)
			ntokens = interval*preload;
		if (ntokens < interval)
			return interval - ntokens;

		cur_time = tv;
		tokens = ntokens - interval;
	}

	if (g_options & F_OUTSTANDING) {
		if (ntransmitted > 0 && !rcvd_test(ntransmitted)) {
			print_timestamp();
			printf("no answer yet for icmp_seq=%lu\n", (ntransmitted % MAX_DUP_CHK));
			fflush(stdout);
		}
	}

resend:
	i = (*send_probe)();

	if (i == 0) {
		oom_count = 0;
		advance_ntransmitted();
		if (!(g_options & F_QUIET) && (g_options & F_FLOOD)) {
			/* Very silly, but without this output with
			 * high preload or pipe size is very confusing. */
			if ((preload < screen_width && pipesize < screen_width) ||
			    in_flight() < screen_width)
				write_stdout(".", 1);
		}
		return interval - tokens;
	}

	/* And handle various errors... */
	if (i > 0) {
		/* Apparently, it is some fatal bug. */
		abort();
	} else if (errno == ENOBUFS || errno == ENOMEM) {
		int nores_interval;

		/* Device queue overflow or OOM. Packet is not sent. */
		tokens = 0;
		/* Slowdown. This works only in adaptive mode (option -A) */
		rtt_addend += (rtt < 8*50000 ? rtt/8 : 50000);
		if (g_options&F_ADAPTIVE)
			update_interval();
		nores_interval = SCHINT(interval/2);
		if (nores_interval > 500)
			nores_interval = 500;
		oom_count++;
		if (oom_count*nores_interval < lingertime)
			return nores_interval;
		i = 0;
		/* Fall to hard error. It is to avoid complete deadlock
		 * on stuck output device even when dealine was not requested.
		 * Expected timings are screwed up in any case, but we will
		 * exit some day. :-) */
	} else if (errno == EAGAIN) {
		/* Socket buffer is full. */
		tokens += interval;
		return MININTERVAL;
	} else {
		if ((i = (*receive_error_msg)()) > 0) {
			/* An ICMP error arrived. */
			tokens += interval;
			return MININTERVAL;
		}
		/* Compatibility with old linuces. */
		if (i == 0 && confirm_flag && errno == EINVAL) {
			confirm_flag = 0;
			errno = 0;
		}
		if (!errno)
			goto resend;
	}

	/* Hard local error. Pretend we sent packet. */
	advance_ntransmitted();

	if (i == 0 && !(g_options & F_QUIET)) {
		if (g_options & F_FLOOD)
			write_stdout("E", 1);
		else
			bb_perror_msg("sendmsg");
	}
	tokens = 0;
	return SCHINT(interval);
}

/* Set socket buffers, "alloc" is an estimate of memory taken by single packet. */

static void sock_setbufs(int alloc)
{
	int rcvbuf, hold;
	socklen_t tmplen = sizeof(hold);

	if (!sndbuf)
		sndbuf = alloc;
	setsockopt(icmp_sock, SOL_SOCKET, SO_SNDBUF, (char *)&sndbuf, sizeof(sndbuf));

	rcvbuf = hold = alloc * preload;
	if (hold < 65536)
		hold = 65536;
	setsockopt(icmp_sock, SOL_SOCKET, SO_RCVBUF, (char *)&hold, sizeof(hold));
	if (getsockopt(icmp_sock, SOL_SOCKET, SO_RCVBUF, (char *)&hold, &tmplen) == 0) {
		if (hold < rcvbuf)
			bb_error_msg("WARNING: probably, rcvbuf is not enough to hold preload.");
	}
}

/* Protocol independent setup and parameter checks. */

static void setup(void)
{
	struct timeval tv;
	sigset_t sset;

	if ((g_options & F_FLOOD) && !(g_options & F_INTERVAL))
		interval = 0;

	if (uid && interval < MINUSERINTERVAL)
		bb_error_msg_and_die("cannot flood; minimal interval, allowed for user, is %dms", MINUSERINTERVAL);

	if (interval >= INT_MAX/preload)
		bb_error_msg_and_die("illegal preload and/or interval");

	if (g_options & F_SO_DONTROUTE)
		setsockopt_1i(icmp_sock, SOL_SOCKET, SO_DONTROUTE, 1);

#ifdef SO_TIMESTAMP
	if (!(g_options & F_LATENCY)) {
		if (setsockopt_1i(icmp_sock, SOL_SOCKET, SO_TIMESTAMP, 1))
			bb_error_msg("Warning: no SO_TIMESTAMP support, falling back to SIOCGSTAMP");
	}
#endif
	if (g_options & F_MARK) {
		int ret;

		enable_capability_admin();
		ret = setsockopt_1i(icmp_sock, SOL_SOCKET, SO_MARK, mark);
		disable_capability_admin();

		if (ret == -1) {
			/* we probably dont wanna exit since old kernels
			 * dont support mark ..
			*/
			bb_error_msg("Warning: Failed to set mark %d", mark);
		}
	}

	/* Set some SNDTIMEO to prevent blocking forever
	 * on sends, when device is too slow or stalls. Just put limit
	 * of one second, or "interval", if it is less.
	 */
	tv.tv_sec = 1;
	tv.tv_usec = 0;
	if (interval < 1000) {
		tv.tv_sec = 0;
		tv.tv_usec = 1000 * SCHINT(interval);
	}
	setsockopt(icmp_sock, SOL_SOCKET, SO_SNDTIMEO, (char*)&tv, sizeof(tv));

	/* Set RCVTIMEO to "interval". Note, it is just an optimization
	 * allowing to avoid redundant poll(). */
	tv.tv_sec = SCHINT(interval)/1000;
	tv.tv_usec = 1000*(SCHINT(interval)%1000);
	if (setsockopt(icmp_sock, SOL_SOCKET, SO_RCVTIMEO, (char*)&tv, sizeof(tv)))
		g_options |= F_FLOOD_POLL;

	if (!(g_options & F_PINGFILLED)) {
		int i;
		u_char *p = outpack+8;

		/* Do not forget about case of small datalen,
		 * fill timestamp area too!
		 */
		for (i = 0; i < datalen; ++i)
			*p++ = i;
	}

	ident = htons(getpid() & 0xFFFF);

#ifdef SA_INTERRUPT
# define BB_SIGNAL_SA_INTERRUPT 1
#else
# define BB_SIGNAL_SA_INTERRUPT 0
#endif
	bb_signal((1<<SIGINT)|(1<<SIGALRM)|BB_SIGNAL_SA_INTERRUPT, sigexit);
	bb_signal((1<<SIGQUIT)|BB_SIGNAL_SA_INTERRUPT, sigstatus);

	sigemptyset(&sset);
	sigprocmask(SIG_SETMASK, &sset, NULL);

	gettimeofday(&start_time, NULL);

	if (deadline) {
		struct itimerval it;

		it.it_interval.tv_sec = 0;
		it.it_interval.tv_usec = 0;
		it.it_value.tv_sec = deadline;
		it.it_value.tv_usec = 0;
		setitimer(ITIMER_REAL, &it, NULL);
	}

	if (isatty(STDOUT_FILENO))
		get_terminal_width_height(&screen_width, NULL);
}


static int gather_statistics(__u8 *icmph, int icmplen,
		      int cc, __u16 seq, int hops,
		      int csfailed, struct timeval *tv, char *from,
		      void (*pr_reply)(__u8 *icmph))
{
	int dupflag = 0;
	long triptime = 0;
	__u8 *ptr = icmph + icmplen;

	++nreceived;
	if (!csfailed)
		acknowledge(seq);

	if (timing && cc >= (int)(8+sizeof(struct timeval))) {
		struct timeval tmp_tv;
		memcpy(&tmp_tv, ptr, sizeof(tmp_tv));

restamp:
		tvsub(tv, &tmp_tv);
		triptime = tv->tv_sec * 1000000 + tv->tv_usec;
		if (triptime < 0) {
			bb_error_msg("Warning: time of day goes back (%ldus), taking countermeasures.", triptime);
			triptime = 0;
			if (!(g_options & F_LATENCY)) {
				gettimeofday(tv, NULL);
				g_options |= F_LATENCY;
				goto restamp;
			}
		}
		if (!csfailed) {
			tsum += triptime;
			tsum2 += (long long)triptime * (long long)triptime;
			if (triptime < tmin)
				tmin = triptime;
			if (triptime > tmax)
				tmax = triptime;
			if (!rtt)
				rtt = triptime*8;
			else
				rtt += triptime-rtt/8;
			if (g_options & F_ADAPTIVE)
				update_interval();
		}
	}

	if (csfailed) {
		++nchecksum;
		--nreceived;
	} else if (rcvd_test(seq)) {
		++nrepeats;
		--nreceived;
		dupflag = 1;
	} else {
		rcvd_set(seq);
		dupflag = 0;
	}
	confirm = confirm_flag;

	if (g_options & F_QUIET)
		return 1;

	if (g_options & F_FLOOD) {
		if (!csfailed)
			write_stdout("\b \b", 3);
		else
			write_stdout("\bC", 2);
	} else {
		int i;
		__u8 *cp, *dp;

		print_timestamp();
		printf("%d bytes from %s:", cc, from);

		if (pr_reply)
			pr_reply(icmph);

		if (hops >= 0)
			printf(" ttl=%d", hops);

		if (cc < datalen+8) {
			printf(" (truncated)\n");
			return 1;
		}
		if (timing) {
			if (triptime >= 100000)
				printf(" time=%ld ms", triptime/1000);
			else if (triptime >= 10000)
				printf(" time=%ld.%01ld ms", triptime/1000,
				       (triptime%1000)/100);
			else if (triptime >= 1000)
				printf(" time=%ld.%02ld ms", triptime/1000,
				       (triptime%1000)/10);
			else
				printf(" time=%ld.%03ld ms", triptime/1000,
				       triptime%1000);
		}
		if (dupflag)
			printf(" (DUP!)");
		if (csfailed)
			printf(" (BAD CHECKSUM!)");

		/* check the data */
		cp = ((u_char*)ptr) + sizeof(struct timeval);
		dp = &outpack[8 + sizeof(struct timeval)];
		for (i = sizeof(struct timeval); i < datalen; ++i, ++cp, ++dp) {
			if (*cp != *dp) {
				printf("\nwrong data byte #%d should be 0x%x but was 0x%x",
				       i, *dp, *cp);
				cp = (u_char*)ptr + sizeof(struct timeval);
				for (i = sizeof(struct timeval); i < datalen; ++i, ++cp) {
					if ((i % 32) == sizeof(struct timeval))
						printf("\n#%d\t", i);
					printf("%x ", *cp);
				}
				break;
			}
		}
	}
	return 0;
}

static long llsqrt(long long a)
{
	long long prev = ~((long long)1 << 63);
	long long x = a;

	if (x > 0) {
		while (x < prev) {
			prev = x;
			x = (x+(a/x))/2;
		}
	}

	return (long)x;
}

/*
 * finish --
 *      Print out statistics, and give up.
 */
static void finish(void)
{
	struct timeval tv = cur_time;
	char *comma = "";

	tvsub(&tv, &start_time);

	putchar('\n');
	fflush(stdout);
	printf("--- %s ping statistics ---\n", hostname);
	printf("%ld packets transmitted, ", ntransmitted);
	printf("%ld received", nreceived);
	if (nrepeats)
		printf(", +%ld duplicates", nrepeats);
	if (nchecksum)
		printf(", +%ld corrupted", nchecksum);
	if (nerrors)
		printf(", +%ld errors", nerrors);
	if (ntransmitted) {
		printf(", %d%% packet loss",
		       (int) ((((long long)(ntransmitted - nreceived)) * 100) /
			      ntransmitted));
		printf(", time %ldms", 1000*tv.tv_sec+tv.tv_usec/1000);
	}
	putchar('\n');

	if (nreceived && timing) {
		long tmdev;

		tsum /= nreceived + nrepeats;
		tsum2 /= nreceived + nrepeats;
		tmdev = llsqrt(tsum2 - tsum * tsum);

		printf("rtt min/avg/max/mdev = %ld.%03ld/%lu.%03ld/%ld.%03ld/%ld.%03ld ms",
		       (long)tmin/1000, (long)tmin%1000,
		       (unsigned long)(tsum/1000), (long)(tsum%1000),
		       (long)tmax/1000, (long)tmax%1000,
		       (long)tmdev/1000, (long)tmdev%1000
		       );
		comma = ", ";
	}
	if (pipesize > 1) {
		printf("%spipe %d", comma, pipesize);
		comma = ", ";
	}
	if (nreceived && (!interval || (g_options&(F_FLOOD|F_ADAPTIVE))) && ntransmitted > 1) {
		int ipg = (1000000*(long long)tv.tv_sec+tv.tv_usec)/(ntransmitted-1);
		printf("%sipg/ewma %d.%03d/%d.%03d ms",
		       comma, ipg/1000, ipg%1000, rtt/8000, (rtt/8)%1000);
	}
	putchar('\n');
	exit(!nreceived || (deadline && nreceived < npackets));
}


static void status(void)
{
	int loss = 0;
	long tavg = 0;

	status_snapshot = 0;

	if (ntransmitted)
		loss = (((long long)(ntransmitted - nreceived)) * 100) / ntransmitted;

	fprintf(stderr, "\r%ld/%ld packets, %d%% loss", ntransmitted, nreceived, loss);

	if (nreceived && timing) {
		tavg = tsum / (nreceived + nrepeats);

		fprintf(stderr, ", min/avg/ewma/max = %ld.%03ld/%lu.%03ld/%d.%03d/%ld.%03ld ms",
		       (long)tmin/1000, (long)tmin%1000,
		       tavg/1000, tavg%1000,
		       rtt/8000, (rtt/8)%1000,
		       (long)tmax/1000, (long)tmax%1000
		       );
	}
	fprintf(stderr, "\n");
}

#ifdef CONFIG_PING

#ifndef ICMP_FILTER
#define ICMP_FILTER     1
struct icmp_filter {
	__u32   data;
};
#endif


#define MAXIPLEN        60
#define MAXICMPLEN      76
#define NROUTES         9               /* number of record route slots */
#define TOS_MAX         255             /* 8-bit TOS field */


/*
 * pr_addr --
 *      Return an ascii host address as a dotted quad and optionally with
 * a hostname.
 */
static char *pr_addr(__u32 addr)
{
	char hbuf[NI_MAXHOST];

	in_pr_addr = !setjmp(pr_addr_jmp);

	if (exiting || (g_options & F_NUMERIC) ||
	    !(bb_xgethostbyaddr(&addr, hbuf, BB_XGETHOSTBYADDR_FLG_IN_ADDR)))
		sprintf(pr_addr_buf, "%s", inet_ntoa(*(struct in_addr *)&addr));
	else {
#if USE_IDN
		char *s;

		if (idna_to_unicode_lzlz(hbuf, &s, 0) != IDNA_SUCCESS)
			s = NULL;
		snprintf(pr_addr_buf, sizeof(pr_addr_buf), "%s (%s)", s ? s : hbuf,
			 inet_ntoa(*(struct in_addr *)&addr));
		free(s);
#else
		snprintf(pr_addr_buf, sizeof(pr_addr_buf), "%s (%s)", hbuf, inet_ntoa(*(struct in_addr *)&addr));
#endif
	}

	in_pr_addr = 0;

	return pr_addr_buf;
}

static void pr_options(unsigned char * cp, int hlen)
{
	int i, j;
	int opt_len, totlen;
	unsigned char * optptr;

	totlen = hlen-sizeof(struct iphdr);
	optptr = cp;

	while (totlen > 0) {
		if (*optptr == IPOPT_EOL)
			break;
		if (*optptr == IPOPT_NOP) {
			totlen--;
			optptr++;
			printf("\nNOP");
			continue;
		}
		cp = optptr;
		opt_len = optptr[1];
		if (opt_len < 2 || opt_len > totlen)
			break;

		switch (*cp) {
		case IPOPT_SSRR:
		case IPOPT_LSRR:
			printf("\n%cSRR: ", *cp==IPOPT_SSRR ? 'S' : 'L');
			j = *++cp;
			i = *++cp;
			i -= 4;
			cp++;
			if (j > IPOPT_MINOFF) {
				for (;;) {
					__u32 address;
					memcpy(&address, cp, 4);
					cp += 4;
					if (address == 0)
						printf("\t0.0.0.0");
					else
						printf("\t%s", pr_addr(address));
					j -= 4;
					putchar('\n');
					if (j <= IPOPT_MINOFF)
						break;
				}
			}
			break;
		case IPOPT_RR:
			j = *++cp;              /* get length */
			i = *++cp;              /* and pointer */
			if (i > j)
				i = j;
			i -= IPOPT_MINOFF;
			if (i <= 0)
				break;
			if (i == old_rrlen
			    && !memcmp(cp, old_rr, i)
			    && !(g_options & F_FLOOD)) {
				printf("\t(same route)");
				i = ((i + 3) / 4) * 4;
				cp += i;
				break;
			}
			old_rrlen = i;
			memcpy(old_rr, (char *)cp, i);
			printf("\nRR: ");
			cp++;
			for (;;) {
				__u32 address;
				memcpy(&address, cp, 4);
				cp += 4;
				if (address == 0)
					printf("\t0.0.0.0");
				else
					printf("\t%s", pr_addr(address));
				i -= 4;
				putchar('\n');
				if (i <= 0)
					break;
			}
			break;
		case IPOPT_TS:
		{
			int stdtime = 0, nonstdtime = 0;
			__u8 flags;
			j = *++cp;              /* get length */
			i = *++cp;              /* and pointer */
			if (i > j)
				i = j;
			i -= 5;
			if (i <= 0)
				break;
			flags = *++cp;
			printf("\nTS: ");
			cp++;
			for (;;) {
				long l;

				if ((flags&0xF) != IPOPT_TS_TSONLY) {
					__u32 address;
					memcpy(&address, cp, 4);
					cp += 4;
					if (address == 0)
						printf("\t0.0.0.0");
					else
						printf("\t%s", pr_addr(address));
					i -= 4;
					if (i <= 0)
						break;
				}
				l = *cp++;
				l = (l<<8) + *cp++;
				l = (l<<8) + *cp++;
				l = (l<<8) + *cp++;

				if  (l & 0x80000000) {
					if (nonstdtime==0)
						printf("\t%ld absolute not-standard", l&0x7fffffff);
					else
						printf("\t%ld not-standard", (l&0x7fffffff) - nonstdtime);
					nonstdtime = l&0x7fffffff;
				} else {
					if (stdtime==0)
						printf("\t%ld absolute", l);
					else
						printf("\t%ld", l - stdtime);
					stdtime = l;
				}
				i -= 4;
				putchar('\n');
				if (i <= 0)
					break;
			}
			if (flags>>4)
				printf("Unrecorded hops: %d\n", flags>>4);
			break;
		}
		default:
			printf("\nunknown option %x", *cp);
			break;
		}
		totlen -= opt_len;
		optptr += opt_len;
	}
}

/*
 * pr_iph --
 *      Print an IP header with options.
 */
static void pr_iph(struct iphdr *ip)
{
	int hlen;
	u_char *cp;

	hlen = ip->ihl << 2;
	cp = (u_char *)ip + 20;         /* point to options */

	printf("Vr HL TOS  Len   ID Flg  off TTL Pro  cks      Src      Dst Data\n");
	printf(" %1x  %1x  %02x %04x %04x",
	       ip->version, ip->ihl, ip->tos, ip->tot_len, ip->id);
	printf("   %1x %04x", ((ip->frag_off) & 0xe000) >> 13,
	       (ip->frag_off) & 0x1fff);
	printf("  %02x  %02x %04x", ip->ttl, ip->protocol, ip->check);
	printf(" %s ", inet_ntoa(*(struct in_addr *)&ip->saddr));
	printf(" %s ", inet_ntoa(*(struct in_addr *)&ip->daddr));
	printf("\n");
	pr_options(cp, hlen);
}


/*
 * pr_icmph --
 *      Print a descriptive string about an ICMP header.
 */
static void pr_icmph(__u8 type, __u8 code, __u32 info, struct icmphdr *icp)
{
	switch(type) {
	case ICMP_ECHOREPLY:
		printf("Echo Reply\n");
		/* XXX ID + Seq + Data */
		break;
	case ICMP_DEST_UNREACH:
		switch(code) {
		case ICMP_NET_UNREACH:
			printf("Destination Net Unreachable\n");
			break;
		case ICMP_HOST_UNREACH:
			printf("Destination Host Unreachable\n");
			break;
		case ICMP_PROT_UNREACH:
			printf("Destination Protocol Unreachable\n");
			break;
		case ICMP_PORT_UNREACH:
			printf("Destination Port Unreachable\n");
			break;
		case ICMP_FRAG_NEEDED:
			printf("Frag needed and DF set (mtu = %u)\n", info);
			break;
		case ICMP_SR_FAILED:
			printf("Source Route Failed\n");
			break;
		case ICMP_NET_UNKNOWN:
			printf("Destination Net Unknown\n");
			break;
		case ICMP_HOST_UNKNOWN:
			printf("Destination Host Unknown\n");
			break;
		case ICMP_HOST_ISOLATED:
			printf("Source Host Isolated\n");
			break;
		case ICMP_NET_ANO:
			printf("Destination Net Prohibited\n");
			break;
		case ICMP_HOST_ANO:
			printf("Destination Host Prohibited\n");
			break;
		case ICMP_NET_UNR_TOS:
			printf("Destination Net Unreachable for Type of Service\n");
			break;
		case ICMP_HOST_UNR_TOS:
			printf("Destination Host Unreachable for Type of Service\n");
			break;
		case ICMP_PKT_FILTERED:
			printf("Packet filtered\n");
			break;
		case ICMP_PREC_VIOLATION:
			printf("Precedence Violation\n");
			break;
		case ICMP_PREC_CUTOFF:
			printf("Precedence Cutoff\n");
			break;
		default:
			printf("Dest Unreachable, Bad Code: %d\n", code);
			break;
		}
		if (icp && (g_options & F_VERBOSE))
			pr_iph((struct iphdr*)(icp + 1));
		break;
	case ICMP_SOURCE_QUENCH:
		printf("Source Quench\n");
		if (icp && (g_options & F_VERBOSE))
			pr_iph((struct iphdr*)(icp + 1));
		break;
	case ICMP_REDIRECT:
		switch(code) {
		case ICMP_REDIR_NET:
			printf("Redirect Network");
			break;
		case ICMP_REDIR_HOST:
			printf("Redirect Host");
			break;
		case ICMP_REDIR_NETTOS:
			printf("Redirect Type of Service and Network");
			break;
		case ICMP_REDIR_HOSTTOS:
			printf("Redirect Type of Service and Host");
			break;
		default:
			printf("Redirect, Bad Code: %d", code);
			break;
		}
		if (icp)
			printf("(New nexthop: %s)\n", pr_addr(icp->un.gateway));
		if (icp && (g_options & F_VERBOSE))
			pr_iph((struct iphdr*)(icp + 1));
		break;
	case ICMP_ECHO:
		printf("Echo Request\n");
		/* XXX ID + Seq + Data */
		break;
	case ICMP_TIME_EXCEEDED:
		switch(code) {
		case ICMP_EXC_TTL:
			printf("Time to live exceeded\n");
			break;
		case ICMP_EXC_FRAGTIME:
			printf("Frag reassembly time exceeded\n");
			break;
		default:
			printf("Time exceeded, Bad Code: %d\n", code);
			break;
		}
		if (icp && (g_options & F_VERBOSE))
			pr_iph((struct iphdr*)(icp + 1));
		break;
	case ICMP_PARAMETERPROB:
		printf("Parameter problem: pointer = %u\n", icp ? (ntohl(icp->un.gateway)>>24) : info);
		if (icp && (g_options & F_VERBOSE))
			pr_iph((struct iphdr*)(icp + 1));
		break;
	case ICMP_TIMESTAMP:
		printf("Timestamp\n");
		/* XXX ID + Seq + 3 timestamps */
		break;
	case ICMP_TIMESTAMPREPLY:
		printf("Timestamp Reply\n");
		/* XXX ID + Seq + 3 timestamps */
		break;
	case ICMP_INFO_REQUEST:
		printf("Information Request\n");
		/* XXX ID + Seq */
		break;
	case ICMP_INFO_REPLY:
		printf("Information Reply\n");
		/* XXX ID + Seq */
		break;
#ifdef ICMP_MASKREQ
	case ICMP_MASKREQ:
		printf("Address Mask Request\n");
		break;
#endif
#ifdef ICMP_MASKREPLY
	case ICMP_MASKREPLY:
		printf("Address Mask Reply\n");
		break;
#endif
	default:
		printf("Bad ICMP type: %d\n", type);
	}
}

static int receive_error_msg_ping(void)
{
	int res;
	char cbuf[512];
	struct iovec  iov;
	struct msghdr msg;
	struct cmsghdr *cmsg;
	struct sock_extended_err *e;
	struct icmphdr icmph;
	struct sockaddr_in target;
	int net_errors = 0;
	int local_errors = 0;
	int saved_errno = errno;

	iov.iov_base = &icmph;
	iov.iov_len = sizeof(icmph);
	msg.msg_name = (void*)&target;
	msg.msg_namelen = sizeof(target);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_flags = 0;
	msg.msg_control = cbuf;
	msg.msg_controllen = sizeof(cbuf);

	res = recvmsg(icmp_sock, &msg, MSG_ERRQUEUE|MSG_DONTWAIT);
	if (res < 0)
		goto out;

	e = NULL;
	for (cmsg = CMSG_FIRSTHDR(&msg); cmsg; cmsg = CMSG_NXTHDR(&msg, cmsg)) {
		if (cmsg->cmsg_level == SOL_IP) {
			if (cmsg->cmsg_type == IP_RECVERR)
				e = (struct sock_extended_err *)CMSG_DATA(cmsg);
		}
	}
	if (e == NULL)
		abort();

	if (e->ee_origin == SO_EE_ORIGIN_LOCAL) {
		local_errors++;
		if (g_options & F_QUIET)
			goto out;
		if (g_options & F_FLOOD)
			write_stdout("E", 1);
		else if (e->ee_errno != EMSGSIZE)
			bb_error_msg("local error: %s", strerror(e->ee_errno));
		else
			bb_error_msg("local error: Message too long, mtu=%u", e->ee_info);
		nerrors++;
	} else if (e->ee_origin == SO_EE_ORIGIN_ICMP) {
		struct sockaddr_in *sin = (struct sockaddr_in*)(e+1);

		if (res < (int)sizeof(icmph) ||
		    target.sin_addr.s_addr != whereto.sin_addr.s_addr ||
		    icmph.type != ICMP_ECHO ||
		    icmph.un.echo.id != ident) {
			/* Not our error, not an error at all. Clear. */
			saved_errno = 0;
			goto out;
		}

		acknowledge(ntohs(icmph.un.echo.sequence));

		if (!working_recverr) {
			struct icmp_filter filt;
			working_recverr = 1;
			/* OK, it works. Add stronger filter. */
			filt.data = ~((1<<ICMP_SOURCE_QUENCH)|
				      (1<<ICMP_REDIRECT)|
				      (1<<ICMP_ECHOREPLY));
			if (setsockopt(icmp_sock, SOL_RAW, ICMP_FILTER, (char*)&filt, sizeof(filt)) == -1)
				bb_perror_msg("WARNING: setsockopt(ICMP_FILTER)");
		}

		net_errors++;
		nerrors++;
		if (g_options & F_QUIET)
			goto out;
		if (g_options & F_FLOOD) {
			write_stdout("\bE", 2);
		} else {
			print_timestamp();
			printf("From %s icmp_seq=%u ", pr_addr(sin->sin_addr.s_addr), ntohs(icmph.un.echo.sequence));
			pr_icmph(e->ee_type, e->ee_code, e->ee_info, NULL);
			fflush(stdout);
		}
	}

out:
	errno = saved_errno;
	return net_errors ? : -local_errors;
}

static int send_probe_ping(void)
{
	struct icmphdr *icp;
	int cc;
	int i;

	icp = (struct icmphdr *)outpack;
	icp->type = ICMP_ECHO;
	icp->code = 0;
	icp->checksum = 0;
	icp->un.echo.sequence = htons(ntransmitted+1);
	icp->un.echo.id = ident;                        /* ID */

	rcvd_clear(ntransmitted+1);

	if (timing) {
		if (g_options & F_LATENCY) {
			struct timeval tmp_tv;
			gettimeofday(&tmp_tv, NULL);
			memcpy(icp+1, &tmp_tv, sizeof(tmp_tv));
		} else {
			memset(icp+1, 0, sizeof(struct timeval));
		}
	}

	cc = datalen + 8;                       /* skips ICMP portion */

	/* compute ICMP checksum here */
	icp->checksum = in_cksum((u_short *)icp, cc, 0);

	if (timing && !(g_options & F_LATENCY)) {
		struct timeval tmp_tv;
		gettimeofday(&tmp_tv, NULL);
		memcpy(icp+1, &tmp_tv, sizeof(tmp_tv));
		icp->checksum = in_cksum((u_short *)&tmp_tv, sizeof(tmp_tv), ~icp->checksum);
	}

	{
		struct iovec iov = {outpack, cc};
		struct msghdr m = { .msg_name = &whereto, .msg_namelen = sizeof(whereto),
					.msg_iov = &iov, .msg_iovlen = 1,
					.msg_control = &g_cmsg, .msg_controllen = ping_cmsg_len,
					.msg_flags = 0 };

		i = sendmsg(icmp_sock, &m, confirm);
		confirm = 0;
	}

	return (cc == i ? 0 : i);
}

/*
 * parse_reply --
 *      Print out the packet, if it came from us.  This logic is necessary
 * because ALL readers of the ICMP socket get a copy of ALL ICMP packets
 * which arrive ('tis only fair).  This permits multiple copies of this
 * program to be run without having intermingled output (or statistics!).
 */
static void pr_echo_reply(__u8 *_icp)
{
	struct icmphdr *icp = (struct icmphdr *)_icp;
	printf(" icmp_seq=%u", ntohs(icp->un.echo.sequence));
}

static int
parse_reply_ping(struct msghdr *msg, int cc, void *addr, struct timeval *tv)
{
	struct sockaddr_in *from = addr;
	__u8 *buf = msg->msg_iov->iov_base;
	struct icmphdr *icp;
	struct iphdr *ip;
	int hlen;
	int csfailed;

	/* Check the IP header */
	ip = (struct iphdr *)buf;
	hlen = ip->ihl*4;
	if (cc < hlen + 8 || ip->ihl < 5) {
		if (g_options & F_VERBOSE)
			bb_error_msg("packet too short (%d bytes) from %s", cc, pr_addr(from->sin_addr.s_addr));
		return 1;
	}

	/* Now the ICMP part */
	cc -= hlen;
	icp = (struct icmphdr *)(buf + hlen);
	csfailed = in_cksum((u_short *)icp, cc, 0);

	if (icp->type == ICMP_ECHOREPLY) {
		if (icp->un.echo.id != ident)
			return 1;                       /* 'Twas not our ECHO */
		if (gather_statistics((__u8*)icp, sizeof(*icp), cc,
				      ntohs(icp->un.echo.sequence),
				      ip->ttl, 0, tv, pr_addr(from->sin_addr.s_addr),
				      pr_echo_reply))
			return 0;
	} else {
		/* We fall here when a redirect or source quench arrived.
		 * Also this branch processes icmp errors, when IP_RECVERR
		 * is broken. */

		switch (icp->type) {
		case ICMP_ECHO:
			/* MUST NOT */
			return 1;
		case ICMP_SOURCE_QUENCH:
		case ICMP_REDIRECT:
		case ICMP_DEST_UNREACH:
		case ICMP_TIME_EXCEEDED:
		case ICMP_PARAMETERPROB:
			{
				struct iphdr * iph = (struct  iphdr *)(&icp[1]);
				struct icmphdr *icp1 = (struct icmphdr*)((unsigned char *)iph + iph->ihl*4);
				int error_pkt;
				if (cc < (int)(8+sizeof(struct iphdr)+8) ||
				    cc < 8+iph->ihl*4+8)
					return 1;
				if (icp1->type != ICMP_ECHO ||
				    iph->daddr != whereto.sin_addr.s_addr ||
				    icp1->un.echo.id != ident)
					return 1;
				error_pkt = (icp->type != ICMP_REDIRECT &&
					     icp->type != ICMP_SOURCE_QUENCH);
				if (error_pkt) {
					acknowledge(ntohs(icp1->un.echo.sequence));
					if (working_recverr) {
						return 0;
					} else {
						/* Sigh, IP_RECVERR for raw socket
						 * was broken until 2.4.9. So, we ignore
						 * the first error and warn on the second.
						 */
						if (once++ == 1)
							bb_error_msg("WARNING: kernel is not very fresh, upgrade is recommended.");
						if (once == 1)
							return 0;
					}
				}
				nerrors+=error_pkt;
				if (g_options & F_QUIET)
					return !error_pkt;
				if (g_options & F_FLOOD) {
					if (error_pkt)
						write_stdout("\bE", 2);
					return !error_pkt;
				}
				print_timestamp();
				printf("From %s: icmp_seq=%u ", pr_addr(from->sin_addr.s_addr),
				       ntohs(icp1->un.echo.sequence));
				if (csfailed)
					printf("(BAD CHECKSUM)");
				pr_icmph(icp->type, icp->code, ntohl(icp->un.gateway), icp);
				return !error_pkt;
			}
		default:
			/* MUST NOT */
			break;
		}
		if ((g_options & F_FLOOD) && !(g_options & (F_VERBOSE|F_QUIET))) {
			if (!csfailed)
				write_stdout("!E", 2);
			else
				write_stdout("!EC", 3);
			return 0;
		}
		if (!(g_options & F_VERBOSE) || uid)
			return 0;
		if (g_options & F_PTIMEOFDAY) {
			struct timeval recv_time;
			gettimeofday(&recv_time, NULL);
			printf("%lu.%06lu ", (unsigned long)recv_time.tv_sec, (unsigned long)recv_time.tv_usec);
		}
		printf("From %s: ", pr_addr(from->sin_addr.s_addr));
		if (csfailed) {
			printf("(BAD CHECKSUM)\n");
			return 0;
		}
		pr_icmph(icp->type, icp->code, ntohl(icp->un.gateway), icp);
		return 0;
	}

	if (!(g_options & F_FLOOD)) {
		pr_options(buf + sizeof(struct iphdr), hlen);

		if (g_options & F_AUDIBLE)
			putchar('\a');
		putchar('\n');
		fflush(stdout);
	} else {
		putchar('\a');
		fflush(stdout);
	}
	return 0;
}

/* Set Type of Service (TOS) and other Quality of Service relating bits */
static int parsetos(char *str)
{
	const char *cp;
	int tos;
	char *ep;

	/* handle both hex and decimal values */
	if (str[0] == '0' && (str[1] == 'x' || str[1] == 'X')) {
		cp = str + 2;
		tos = (int)strtol(cp, &ep, 16);
	} else
		tos = (int)strtol(str, &ep, 10);

	/* doesn't look like decimal or hex, eh? */
	if (*ep != '\0')
		bb_error_msg_and_die("\"%s\" bad value for TOS", str);

	if (tos > TOS_MAX)
		bb_error_msg_and_die("the decimal value of TOS bits must be 0-254 (or zero)");
	return(tos);
}

static void install_filter_ping(void)
{
	if (filter_ping.filter)
		return;
	memcpy(insns_ping, insns_ping_st, sizeof(insns_ping));
	filter_ping.len = sizeof insns_ping / sizeof(insns_ping[0]);
	filter_ping.filter = insns_ping;

	/* Patch bpflet for current identifier. */
	insns_ping[2] = (struct sock_filter)BPF_JUMP(BPF_JMP|BPF_JEQ|BPF_K, htons(ident), 0, 1);

	if (setsockopt(icmp_sock, SOL_SOCKET, SO_ATTACH_FILTER, &filter_ping, sizeof(filter_ping)))
		bb_perror_msg("WARNING: failed to install socket filter");
}

#endif /* CONFIG_PING */

static void main_loop(__u8 *packet, int packlen)
{
	char addrbuf[128];
	char ans_data[4096];
	struct iovec iov;
	struct msghdr msg;
	struct cmsghdr *c;
	int cc;
	int next;
	int polling;

	iov.iov_base = (char *)packet;

	setup();
	drop_capabilities();

	for (;;) {
		/* Check exit conditions. */
		if (exiting)
			break;
		if (npackets && nreceived + nerrors >= npackets)
			break;
		if (deadline && nerrors)
			break;
		/* Check for and do special actions. */
		if (status_snapshot)
			status();

		/* Send probes scheduled to this time. */
		do {
			next = pinger();
			next = schedule_exit(next);
		} while (next <= 0);

		/* "next" is time to send next probe, if positive.
		 * If next<=0 send now or as soon as possible. */

		/* Technical part. Looks wicked. Could be dropped,
		 * if everyone used the newest kernel. :-)
		 * Its purpose is:
		 * 1. Provide intervals less than resolution of scheduler.
		 *    Solution: spinning.
		 * 2. Avoid use of poll(), when recvmsg() can provide
		 *    timed waiting (SO_RCVTIMEO). */
		polling = 0;
		if ((g_options & (F_ADAPTIVE|F_FLOOD_POLL)) || next<SCHINT(interval)) {
			int recv_expected = in_flight();

			/* If we are here, recvmsg() is unable to wait for
			 * required timeout. */
			if (1000 % HZ == 0 ? next <= 1000 / HZ : (next < INT_MAX / HZ && next * HZ <= 1000)) {
				/* Very short timeout... So, if we wait for
				 * something, we sleep for MININTERVAL.
				 * Otherwise, spin! */
				if (recv_expected) {
					next = MININTERVAL;
				} else {
					next = 0;
					/* When spinning, no reasons to poll.
					 * Use nonblocking recvmsg() instead. */
					polling = MSG_DONTWAIT;
					/* But yield yet. */
					sched_yield();
				}
			}

			if (!polling &&
			    ((g_options & (F_ADAPTIVE|F_FLOOD_POLL)) || interval)) {
				struct pollfd pset;
				pset.fd = icmp_sock;
				pset.events = POLLIN|POLLERR;
				pset.revents = 0;
				if (poll(&pset, 1, next) < 1 ||
				    !(pset.revents&(POLLIN|POLLERR)))
					continue;
				polling = MSG_DONTWAIT;
			}
		}

		for (;;) {
			struct timeval *recv_timep = NULL;
			struct timeval recv_time;
			int not_ours = 0; /* Raw socket can receive messages
					   * destined to other running pings. */

			iov.iov_len = packlen;
			memset(&msg, 0, sizeof(msg));
			msg.msg_name = addrbuf;
			msg.msg_namelen = sizeof(addrbuf);
			msg.msg_iov = &iov;
			msg.msg_iovlen = 1;
			msg.msg_control = ans_data;
			msg.msg_controllen = sizeof(ans_data);

			cc = recvmsg(icmp_sock, &msg, polling);
			polling = MSG_DONTWAIT;

			if (cc < 0) {
				if (errno == EAGAIN || errno == EINTR)
					break;
				if (!(*receive_error_msg)()) {
					if (errno) {
						bb_perror_msg("recvmsg");
						break;
					}
					not_ours = 1;
				}
			} else {

#ifdef SO_TIMESTAMP
				for (c = CMSG_FIRSTHDR(&msg); c; c = CMSG_NXTHDR(&msg, c)) {
					if (c->cmsg_level != SOL_SOCKET ||
					    c->cmsg_type != SO_TIMESTAMP)
						continue;
					if (c->cmsg_len < CMSG_LEN(sizeof(struct timeval)))
						continue;
					recv_timep = (struct timeval*)CMSG_DATA(c);
				}
#endif

				if ((g_options & F_LATENCY) || recv_timep == NULL) {
					if ((g_options & F_LATENCY) ||
					    ioctl(icmp_sock, SIOCGSTAMP, &recv_time))
						gettimeofday(&recv_time, NULL);
					recv_timep = &recv_time;
				}

				not_ours = (*parse_reply)(&msg, cc, addrbuf, recv_timep);
			}

			/* See? ... someone runs another ping on this host. */
			if (not_ours)
				(*install_filter)();

			/* If nothing is in flight, "break" returns us to pinger. */
			if (in_flight() == 0)
				break;

			/* Otherwise, try to recvmsg() again. recvmsg()
			 * is nonblocking after the first iteration, so that
			 * if nothing is queued, it will receive EAGAIN
			 * and return to pinger. */
		}
	}
	finish();
}

#ifdef CONFIG_PING
int
ping_main(void)
{
	int argc = Gargc;
	char **argv = Gargv;
	int ch, hold, packlen;
	int socket_errno;
	u_char *packet;
	char *target;
	char *hnamebuf = NULL;
	char rspace[3 + 4 * NROUTES + 1];       /* record route space */
	__u32 route[10];
	int nroute = 0;
	int ts_type = 0;
	struct sockaddr_in source;
	int optlen = 0;
	int settos = 0;         /* Set TOS, Precendence or other QOS options */
	int broadcast_pings = 0;

	install_filter = install_filter_ping;
	receive_error_msg = receive_error_msg_ping;
	send_probe = send_probe_ping;
	parse_reply = parse_reply_ping;
	g_cmsg.cm.cmsg_len = sizeof(struct cmsghdr) + sizeof(struct in_pktinfo);
	g_cmsg.cm.cmsg_level = SOL_IP;
	g_cmsg.cm.cmsg_type = IP_PKTINFO;

	common_data_init();

	limit_capabilities();

#ifdef USE_IDN
	setlocale(LC_ALL, "");
#endif

	enable_capability_raw();

	icmp_sock = bb_xsocket(BB_XSOCKET_FLG_SOCK_RAW | BB_XSOCKET_FLG_PROTO_SPEC | BB_XSOCKET_FLG_NO_WARN, IPPROTO_ICMP);
	socket_errno = errno;

	disable_capability_raw();

	memset(&source, 0, sizeof(source));
	source.sin_family = AF_INET;

	preload = 1;
	while ((ch = getopt(argc, argv, COMMON_OPTSTR "bRT:")) != EOF) {
		switch(ch) {
		case 'b':
			broadcast_pings = 1;
			break;
		case 'Q':
			settos = parsetos(optarg);
			if (settos && setsockopt_1i(icmp_sock, IPPROTO_IP, IP_TOS, settos) < 0)
				bb_perror_msg_and_die("error setting QOS sockopts");
			break;
		case 'R':
			if (g_options & F_TIMESTAMP)
				bb_error_msg_and_die("Only one of -T or -R may be used");
			g_options |= F_RROUTE;
			break;
		case 'T':
			if (g_options & F_RROUTE)
				bb_error_msg_and_die("Only one of -T or -R may be used");
			g_options |= F_TIMESTAMP;
			if (strcmp(optarg, "tsonly") == 0)
				ts_type = IPOPT_TS_TSONLY;
			else if (strcmp(optarg, "tsandaddr") == 0)
				ts_type = IPOPT_TS_TSANDADDR;
			else if (strcmp(optarg, "tsprespec") == 0)
				ts_type = IPOPT_TS_PRESPEC;
			else
				bb_error_msg_and_die("Invalid timestamp type");
			break;
		case 'I':
		{
			if (inet_pton(AF_INET, optarg, &source.sin_addr) > 0)
				g_options |= F_STRICTSOURCE;
			else
				device = optarg;
			break;
		}
		case 'M':
			if (strcmp(optarg, "do") == 0)
				pmtudisc = IP_PMTUDISC_DO;
			else if (strcmp(optarg, "dont") == 0)
				pmtudisc = IP_PMTUDISC_DONT;
			else if (strcmp(optarg, "want") == 0)
				pmtudisc = IP_PMTUDISC_WANT;
			else
				bb_error_msg_and_die("wrong value for -M: do, dont, want are valid ones.");
			break;
		COMMON_OPTIONS
			common_options(ch);
			break;
		default:
			bb_show_usage();
		}
	}
	argc -= optind;
	argv += optind;

	if (argc == 0)
		bb_show_usage();
	if (argc > 1) {
		if (g_options & F_RROUTE)
			bb_show_usage();
		else if (g_options & F_TIMESTAMP) {
			if (ts_type != IPOPT_TS_PRESPEC)
				bb_show_usage();
			if (argc > 5)
				bb_show_usage();
		} else {
			if (argc > 10)
				bb_show_usage();
			g_options |= F_SOURCEROUTE;
		}
	}
	while (argc > 0) {
		target = *argv;

		memset((char *)&whereto, 0, sizeof(whereto));
		whereto.sin_family = AF_INET;
		if (inet_aton(target, &whereto.sin_addr) == 1) {
			hostname = target;
			if (argc == 1)
				g_options |= F_NUMERIC;
		} else {
#ifdef USE_IDN
			char *idn;
			char *canon_name;
			int rc;
#endif
			if (hnamebuf) {
				free(hnamebuf);
				hnamebuf = NULL;
			}

#ifdef USE_IDN
			rc = idna_to_ascii_lz(target, &idn, 0);
			if (rc != IDNA_SUCCESS)
				bb_error_msg_and_die("IDN encoding failed: %s", idna_strerror(rc));

			bb_gethostbyname(BB_GHBN_SOCKADDR_IN | BB_GHBN_ERR_DIE | BB_GHBN_CANON_NAME, idn, &whereto, &canon_name);
			free(idn);
			if (idna_to_unicode_lzlz(canon_name, &hnamebuf, 0) != IDNA_SUCCESS)
				hnamebuf = canon_name;
			else
				free(canon_name);
#else
			bb_gethostbyname(BB_GHBN_SOCKADDR_IN | BB_GHBN_ERR_DIE | BB_GHBN_CANON_NAME, target, &whereto, &hnamebuf);
#endif
			hostname = hnamebuf;
		}
		if (argc > 1)
			route[nroute++] = whereto.sin_addr.s_addr;
		argc--;
		argv++;
	}

	if (source.sin_addr.s_addr == 0) {
		socklen_t alen;
		struct sockaddr_in dst = whereto;
		int probe_fd = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM | BB_XSOCKET_FLG_ERR_DIE);

		if (device) {
			struct ifreq ifr;
			int rc;

			safe_ifname_cp0(&ifr, device);

			enable_capability_raw();
			rc = setsockopt(probe_fd, SOL_SOCKET, SO_BINDTODEVICE, device, strlen(device)+1);
			disable_capability_raw();

			if (rc == -1) {
				if (IN_MULTICAST(ntohl(dst.sin_addr.s_addr))) {
					struct ip_mreqn imr;

					xioctl(probe_fd, SIOCGIFINDEX, &ifr, "\1unknown iface %s", device);
					memset(&imr, 0, sizeof(imr));
					imr.imr_ifindex = ifr.ifr_ifindex;
					if (setsockopt(probe_fd, SOL_IP, IP_MULTICAST_IF, &imr, sizeof(imr)) == -1)
						bb_perror_msg_and_die("IP_MULTICAST_IF");
				} else {
					bb_perror_msg_and_die("SO_BINDTODEVICE");
				}
			}
		}

		if (settos && setsockopt_1i(probe_fd, IPPROTO_IP, IP_TOS, settos) < 0)
			bb_perror_msg("Warning: error setting QOS sockopts");

		dst.sin_port = htons(1025);
		if (nroute)
			dst.sin_addr.s_addr = route[0];
		if (connect(probe_fd, (struct sockaddr*)&dst, sizeof(dst)) == -1) {
			if (errno == EACCES) {
				if (broadcast_pings == 0)
					bb_error_msg_and_die("Do you want to ping broadcast? Then -b");
				bb_error_msg("WARNING: pinging broadcast address");
				if (setsockopt_1i(probe_fd, SOL_SOCKET, SO_BROADCAST, broadcast_pings) < 0)
					bb_perror_msg_and_die ("can't set broadcasting");
				if (connect(probe_fd, (struct sockaddr*)&dst, sizeof(dst)) == -1)
					bb_perror_msg_and_die("connect");
			} else {
				bb_perror_msg_and_die("connect");
			}
		}
		alen = sizeof(source);
		if (getsockname(probe_fd, (struct sockaddr*)&source, &alen) == -1)
			bb_perror_msg_and_die("getsockname");
		source.sin_port = 0;

#ifndef WITHOUT_IFADDRS
		if (device) {
			struct ifaddrs *ifa0, *ifa;
			int ret;

			ret = getifaddrs(&ifa0);
			if (ret)
				bb_error_msg_and_die("gatifaddrs() failed.");
			for (ifa = ifa0; ifa; ifa = ifa->ifa_next) {
				if (!ifa->ifa_addr || ifa->ifa_addr->sa_family != AF_INET)
					continue;
				if (!strncmp(ifa->ifa_name, device, sizeof(device) - 1) &&
				    !memcmp(&((struct sockaddr_in *)ifa->ifa_addr)->sin_addr,
					    &source.sin_addr, sizeof(source.sin_addr)))
					break;
			}
			freeifaddrs(ifa0);
			if (!ifa)
				bb_error_msg("Warning: source address might be selected on device other than %s.", device);
		}
#endif
		close(probe_fd);
	} while (0);

	if (whereto.sin_addr.s_addr == 0)
		whereto.sin_addr.s_addr = source.sin_addr.s_addr;

	if (icmp_sock < 0) {
		errno = socket_errno;
		bb_perror_msg_and_die("icmp open socket");
	}

	if (device) {
		struct ifreq ifr;

		safe_ifname_cp0(&ifr, device);
		xioctl(icmp_sock, SIOCGIFINDEX, &ifr, "\1unknown iface %s", device);
		g_cmsg.ipi.ipi_ifindex = ifr.ifr_ifindex;
		ping_cmsg_len = sizeof(g_cmsg);
	}

	if (broadcast_pings || IN_MULTICAST(ntohl(whereto.sin_addr.s_addr))) {
		if (uid) {
			if (interval < 1000)
				bb_error_msg_and_die("broadcast ping with too short interval.");
			if (pmtudisc >= 0 && pmtudisc != IP_PMTUDISC_DO)
				bb_error_msg_and_die("broadcast ping does not fragment.");
		}
		if (pmtudisc < 0)
			pmtudisc = IP_PMTUDISC_DO;
	}

	if (pmtudisc >= 0) {
		if (setsockopt_1i(icmp_sock, SOL_IP, IP_MTU_DISCOVER, pmtudisc) == -1)
			bb_perror_msg_and_die("IP_MTU_DISCOVER");
	}

	if ((g_options & F_STRICTSOURCE))
		bb_xbind(icmp_sock, (struct sockaddr*)&source, sizeof(source));

	if (1) {
		struct icmp_filter filt;
		filt.data = ~((1<<ICMP_SOURCE_QUENCH)|
			      (1<<ICMP_DEST_UNREACH)|
			      (1<<ICMP_TIME_EXCEEDED)|
			      (1<<ICMP_PARAMETERPROB)|
			      (1<<ICMP_REDIRECT)|
			      (1<<ICMP_ECHOREPLY));
		if (setsockopt(icmp_sock, SOL_RAW, ICMP_FILTER, (char*)&filt, sizeof(filt)) == -1)
			bb_perror_msg("WARNING: setsockopt(ICMP_FILTER)");
	}

	if (setsockopt_1i(icmp_sock, SOL_IP, IP_RECVERR, 1))
		bb_error_msg("WARNING: your kernel is veeery old. No problems.");

	/* record route option */
	if (g_options & F_RROUTE) {
		memset(rspace, 0, sizeof(rspace));
		rspace[0] = IPOPT_NOP;
		rspace[1+IPOPT_OPTVAL] = IPOPT_RR;
		rspace[1+IPOPT_OLEN] = sizeof(rspace)-1;
		rspace[1+IPOPT_OFFSET] = IPOPT_MINOFF;
		optlen = 40;
		if (setsockopt(icmp_sock, IPPROTO_IP, IP_OPTIONS, rspace, sizeof(rspace)) < 0)
			bb_perror_msg_and_die("record route");
	}
	if (g_options & F_TIMESTAMP) {
		memset(rspace, 0, sizeof(rspace));
		rspace[0] = IPOPT_TIMESTAMP;
		rspace[1] = (ts_type==IPOPT_TS_TSONLY ? 40 : 36);
		rspace[2] = 5;
		rspace[3] = ts_type;
		if (ts_type == IPOPT_TS_PRESPEC) {
			int i;
			rspace[1] = 4+nroute*8;
			for (i=0; i<nroute; i++)
				memcpy(&rspace[4+i*8], route + i, sizeof(__u32));
		}
		if (setsockopt(icmp_sock, IPPROTO_IP, IP_OPTIONS, rspace, rspace[1]) < 0) {
			rspace[3] = 2;
			if (setsockopt(icmp_sock, IPPROTO_IP, IP_OPTIONS, rspace, rspace[1]) < 0)
				bb_perror_msg_and_die("ts option");
		}
		optlen = 40;
	}
	if (g_options & F_SOURCEROUTE) {
		int i;
		memset(rspace, 0, sizeof(rspace));
		rspace[0] = IPOPT_NOOP;
		rspace[1+IPOPT_OPTVAL] = (g_options & F_SO_DONTROUTE) ? IPOPT_SSRR
			: IPOPT_LSRR;
		rspace[1+IPOPT_OLEN] = 3 + nroute*4;
		rspace[1+IPOPT_OFFSET] = IPOPT_MINOFF;
		for (i=0; i<nroute; i++)
			memcpy(&rspace[4+i*4], route + i, sizeof(__u32));

		if (setsockopt(icmp_sock, IPPROTO_IP, IP_OPTIONS, rspace, 4 + nroute*4) < 0)
			bb_perror_msg_and_die("record route");
		optlen = 40;
	}

	/* Estimate memory eaten by single packet. It is rough estimate.
	 * Actually, for small datalen's it depends on kernel side a lot. */
	hold = datalen + 8;
	hold += ((hold+511)/512)*(optlen + 20 + 16 + 64 + 160);
	sock_setbufs(hold);

	if (broadcast_pings) {
		if (setsockopt_1i(icmp_sock, SOL_SOCKET, SO_BROADCAST, 1) < 0)
			bb_perror_msg_and_die ("can't set broadcasting");
	}

	if (g_options & F_NOLOOP) {
		if (setsockopt_1i(icmp_sock, IPPROTO_IP, IP_MULTICAST_LOOP, 0) == -1)
			bb_perror_msg_and_die ("can't disable multicast loopback");
	}
	if (g_options & F_TTL) {
		if (setsockopt_1i(icmp_sock, IPPROTO_IP, IP_MULTICAST_TTL, g_ttl) == -1)
			bb_perror_msg_and_die ("can't set multicast time-to-live");
		if (setsockopt_1i(icmp_sock, IPPROTO_IP, IP_TTL, g_ttl) == -1)
			bb_perror_msg_and_die ("can't set unicast time-to-live");
	}

	if (datalen > 0xFFFF - 8 - optlen - 20) {
		if (uid || datalen > (int)(sizeof(outpack)-8))
			bb_error_msg_and_die("Error: packet size %d is too large. Maximum is %d", datalen, 0xFFFF-8-20-optlen);
		/* Allow small oversize to root yet. It will cause EMSGSIZE. */
		bb_error_msg("WARNING: packet size %d is too large. Maximum is %d", datalen, 0xFFFF-8-20-optlen);
	}

	if (datalen >= (int)sizeof(struct timeval))  /* can we time transfer */
		timing = 1;
	packlen = datalen + MAXIPLEN + MAXICMPLEN;
	packet = xmalloc((u_int)packlen);

	printf("PING %s (%s) ", hostname, inet_ntoa(whereto.sin_addr));
	if (device || (g_options & F_STRICTSOURCE))
		printf("from %s %s: ", inet_ntoa(source.sin_addr), device ?: "");
	printf("%d(%d) bytes of data.\n", datalen, datalen+8+optlen+20);

	main_loop(packet, packlen);
}
#endif /* CONFIG_PING */

#ifdef CONFIG_PING6
/*
   It is just a stripped copy of the kernel header "linux/in6.h"

   "Flow label" things are still not defined in "netinet/in*.h" headers,
   but we cannot use "linux/in6.h" immediately because it currently
   conflicts with "netinet/in.h" .
*/

struct in6_flowlabel_req
{
	struct in6_addr flr_dst;
	__u32   flr_label;
	__u8    flr_action;
	__u8    flr_share;
	__u16   flr_flags;
	__u16   flr_expires;
	__u16   flr_linger;
	__u32   __flr_pad;
	/* Options in format of IPV6_PKTOPTIONS */
};

#define IPV6_FL_A_GET   0
#define IPV6_FL_A_PUT   1
#define IPV6_FL_A_RENEW 2

#define IPV6_FL_F_CREATE        1
#define IPV6_FL_F_EXCL          2

#define IPV6_FL_S_NONE          0
#define IPV6_FL_S_EXCL          1
#define IPV6_FL_S_PROCESS       2
#define IPV6_FL_S_USER          3
#define IPV6_FL_S_ANY           255

#define IPV6_FLOWINFO_FLOWLABEL         0x000fffff
#define IPV6_FLOWINFO_PRIORITY          0x0ff00000

#define IPV6_FLOWLABEL_MGR      32
#define IPV6_FLOWINFO_SEND      33
/*
 *
 *      Modified for AF_INET6 by Pedro Roque
 *
 *      <roque@di.fc.ul.pt>
 *
 *      Original copyright notice included bellow
 */


#include <resolv.h>

#ifdef USE_IDN
#include <stringprep.h>
#endif

#ifndef SOL_IPV6
#define SOL_IPV6 IPPROTO_IPV6
#endif

#ifndef SOL_ICMPV6
#define SOL_ICMPV6 IPPROTO_ICMPV6
#endif

/* RFC3542 */
#ifndef ICMP6_DST_UNREACH_BEYONDSCOPE
#define ICMP6_DST_UNREACH_BEYONDSCOPE ICMP6_DST_UNREACH_NOTNEIGHBOR
#endif

#if defined(ENABLE_PING6_RTHDR) && !defined(ENABLE_PING6_RTHDR_RFC3542)
#ifndef IPV6_SRCRT_TYPE_0
#define IPV6_SRCRT_TYPE_0       0
#endif
#endif

#ifndef MLD_LISTENER_QUERY
#define MLD_LISTENER_QUERY      130
#define MLD_LISTENER_REPORT     131
#define MLD_LISTENER_REDUCTION  132
#endif

#define BIT_CLEAR(nr, addr) do { ((__u32 *)(addr))[(nr) >> 5] &= ~(1U << ((nr) & 31)); } while(0)
#define BIT_SET(nr, addr) do { ((__u32 *)(addr))[(nr) >> 5] |= (1U << ((nr) & 31)); } while(0)
#define BIT_TEST(nr, addr) do { (__u32 *)(addr))[(nr) >> 5] & (1U << ((nr) & 31)); } while(0)

#ifndef ICMP6_FILTER_WILLPASS
#define ICMP6_FILTER_WILLPASS(type, filterp) \
	(BIT_TEST((type), filterp) == 0)

#define ICMP6_FILTER_WILLBLOCK(type, filterp) \
	BIT_TEST((type), filterp)

#define ICMP6_FILTER_SETPASS(type, filterp) \
	BIT_CLEAR((type), filterp)

#define ICMP6_FILTER_SETBLOCK(type, filterp) \
	BIT_SET((type), filterp)

#define ICMP6_FILTER_SETPASSALL(filterp) \
	memset(filterp, 0, sizeof(struct icmp6_filter));

#define ICMP6_FILTER_SETBLOCKALL(filterp) \
	memset(filterp, 0xFF, sizeof(struct icmp6_filter));
#endif


#ifdef SO_TIMESTAMP
#define HAVE_SIN6_SCOPEID 1
#endif

#ifndef SCOPE_DELIMITER
# define SCOPE_DELIMITER '%'
#endif

static int pr_icmph6(__u8 type, __u8 code, __u32 info);


static inline int ntohsp(__u16 *p)
{
	__u16 v;
	memcpy(&v, p, sizeof(v));
	return ntohs(v);
}

#if defined(ENABLE_PING6_RTHDR) && !defined(ENABLE_PING6_RTHDR_RFC3542)
size_t inet6_srcrt_space(int type, int segments)
{
	if (type != 0 || segments > 24)
		return 0;

	return (sizeof(struct cmsghdr) + sizeof(struct ip6_rthdr0) +
		segments * sizeof(struct in6_addr));
}

extern struct cmsghdr * inet6_srcrt_init(void *bp, int type)
{
	struct cmsghdr *cmsg;

	if (type)
		return NULL;

	memset(bp, 0, sizeof(struct cmsghdr) + sizeof(struct ip6_rthdr0));
	cmsg = (struct cmsghdr *) bp;

	cmsg->cmsg_len = sizeof(struct cmsghdr) + sizeof(struct ip6_rthdr0);
	cmsg->cmsg_level = SOL_IPV6;
	cmsg->cmsg_type = IPV6_RTHDR;

	return cmsg;
}

int inet6_srcrt_add(struct cmsghdr *cmsg, const struct in6_addr *addr)
{
	struct ip6_rthdr0 *hdr;

	hdr = (struct ip6_rthdr0 *) CMSG_DATA(cmsg);

	cmsg->cmsg_len += sizeof(struct in6_addr);
	hdr->ip6r0_len += sizeof(struct in6_addr) / 8;

	memcpy(&hdr->ip6r0_addr[hdr->ip6r0_segleft++], addr,
	       sizeof(struct in6_addr));

	return 0;
}
#endif

static unsigned int if_name2index(const char *ifname)
{
	unsigned int i = if_nametoindex(ifname);
	if (!i)
		bb_error_msg_and_die("unknown iface %s", ifname);
	return i;
}

#define OFFSET_OF(type,elem)    ((size_t)&((type *)0)->elem)

static int hextoui(const char *str)
{
	unsigned long val;
	char *ep;

	errno = 0;
	val = strtoul(str, &ep, 16);
	if (*ep) {
		if (!errno)
			errno = EINVAL;
		return -1;
	}

	if (val > UINT_MAX) {
		errno = ERANGE;
		return UINT_MAX;
	}

	return val;
}

static char *pr_addr_n(struct in6_addr *addr, char hbuf[NI_MAXHOST])
{
	inet_ntop(AF_INET6, addr, hbuf, NI_MAXHOST);
	return hbuf;
}

/*
 * pr_addr6 --
 *      Return an ascii host address as a dotted quad and optionally with
 * a hostname.
 */
static char * pr_addr6(struct in6_addr *addr)
{
	char *hp = NULL;
	char hbuf[NI_MAXHOST];

	free(g_pr_addr6_s);
	in_pr_addr = !setjmp(pr_addr_jmp);
#ifdef USE_IDN

	if (!(exiting || g_options & F_NUMERIC))
		hp = bb_xgethostbyaddr(addr, hbuf, BB_XGETHOSTBYADDR_FLG_IN6_ADDR);

	in_pr_addr = 0;

	if (!hp || idna_to_unicode_lzlz(hp, &g_pr_addr6_s, 0) != IDNA_SUCCESS)
		g_pr_addr6_s = NULL;

	if(hp) {
		if(g_pr_addr6_s == NULL)
			g_pr_addr6_s = xstrdup(hp);
	}
#else
	if (!(exiting || g_options & F_NUMERIC))
		hp = bb_xgethostbyaddr(addr, hbuf, BB_XGETHOSTBYADDR_FLG_IN6_ADDR);

	in_pr_addr = 0;

	if(hp)
		g_pr_addr6_s = xstrdup(hp);
#endif
	else
		g_pr_addr6_s = xstrdup(pr_addr_n(addr, hbuf));
	return g_pr_addr6_s;
}

static int receive_error_msg_ping6(void)
{
	int res;
	char cbuf[512];
	struct iovec  iov;
	struct msghdr msg;
	struct cmsghdr *cmsg;
	struct sock_extended_err *e;
	struct icmp6_hdr icmph;
	struct sockaddr_in6 target;
	int net_errors = 0;
	int local_errors = 0;
	int saved_errno = errno;

	iov.iov_base = &icmph;
	iov.iov_len = sizeof(icmph);
	msg.msg_name = (void*)&target;
	msg.msg_namelen = sizeof(target);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_flags = 0;
	msg.msg_control = cbuf;
	msg.msg_controllen = sizeof(cbuf);

	res = recvmsg(icmp_sock, &msg, MSG_ERRQUEUE|MSG_DONTWAIT);
	if (res < 0)
		goto out;

	e = NULL;
	for (cmsg = CMSG_FIRSTHDR(&msg); cmsg; cmsg = CMSG_NXTHDR(&msg, cmsg)) {
		if (cmsg->cmsg_level == SOL_IPV6) {
			if (cmsg->cmsg_type == IPV6_RECVERR)
				e = (struct sock_extended_err *)CMSG_DATA(cmsg);
		}
	}
	if (e == NULL)
		abort();

	if (e->ee_origin == SO_EE_ORIGIN_LOCAL) {
		local_errors++;
		if (g_options & F_QUIET)
			goto out;
		if (g_options & F_FLOOD)
			write_stdout("E", 1);
		else if (e->ee_errno != EMSGSIZE)
			bb_error_msg("local error: %s", strerror(e->ee_errno));
		else
			bb_error_msg("local error: Message too long, mtu=%u", e->ee_info);
		nerrors++;
	} else if (e->ee_origin == SO_EE_ORIGIN_ICMP6) {
		struct sockaddr_in6 *sin6 = (struct sockaddr_in6*)(e+1);

		if (res < (int)sizeof(icmph) ||
		    memcmp(&target.sin6_addr, &whereto6.sin6_addr, 16) ||
		    icmph.icmp6_type != ICMP6_ECHO_REQUEST ||
		    icmph.icmp6_id != ident) {
			/* Not our error, not an error at all. Clear. */
			saved_errno = 0;
			goto out;
		}

		net_errors++;
		nerrors++;
		if (g_options & F_QUIET)
			goto out;
		if (g_options & F_FLOOD) {
			write_stdout("\bE", 2);
		} else {
			print_timestamp();
			printf("From %s icmp_seq=%u ", pr_addr6(&sin6->sin6_addr), ntohs(icmph.icmp6_seq));
			pr_icmph6(e->ee_type, e->ee_code, e->ee_info);
			putchar('\n');
			fflush(stdout);
		}
	}

out:
	errno = saved_errno;
	return net_errors ? : -local_errors;
}

static int build_echo(__u8 *_icmph)
{
	struct icmp6_hdr *icmph;
	int cc;

	icmph = (struct icmp6_hdr *)_icmph;
	icmph->icmp6_type = ICMP6_ECHO_REQUEST;
	icmph->icmp6_code = 0;
	icmph->icmp6_cksum = 0;
	icmph->icmp6_seq = htons(ntransmitted+1);
	icmph->icmp6_id = ident;

	if (timing)
		gettimeofday((struct timeval *)&outpack[8],
		    (struct timezone *)NULL);

	cc = datalen + 8;                       /* skips ICMP portion */

	return cc;
}

static int send_probe_ping6(void)
{
	int len, cc;

	rcvd_clear(ntransmitted + 1);

	len = build_echo(outpack);

	if (cmsglen == 0) {
		cc = sendto(icmp_sock, (char *)outpack, len, confirm,
			    (struct sockaddr *) &whereto6,
			    sizeof(struct sockaddr_in6));
	} else {
		struct msghdr mhdr;
		struct iovec iov;

		iov.iov_len  = len;
		iov.iov_base = outpack;

		memset(&mhdr, 0, sizeof(mhdr));
		mhdr.msg_name = &whereto6;
		mhdr.msg_namelen = sizeof(struct sockaddr_in6);
		mhdr.msg_iov = &iov;
		mhdr.msg_iovlen = 1;
		mhdr.msg_control = cmsgbuf;
		mhdr.msg_controllen = cmsglen;

		cc = sendmsg(icmp_sock, &mhdr, confirm);
	}
	confirm = 0;

	return (cc == len ? 0 : cc);
}

static void pr_echo_reply6(__u8 *_icmph)
{
	struct icmp6_hdr *icmph = (struct icmp6_hdr *) _icmph;
	printf(" icmp_seq=%u", ntohs(icmph->icmp6_seq));
};

/*
 * parse_reply --
 *      Print out the packet, if it came from us.  This logic is necessary
 * because ALL readers of the ICMP socket get a copy of ALL ICMP packets
 * which arrive ('tis only fair).  This permits multiple copies of this
 * program to be run without having intermingled output (or statistics!).
 */
static int
parse_reply_ping6(struct msghdr *msg, int cc, void *addr, struct timeval *tv)
{
	struct sockaddr_in6 *from = addr;
	__u8 *buf = msg->msg_iov->iov_base;
	struct cmsghdr *c;
	struct icmp6_hdr *icmph;
	int hops = -1;

	for (c = CMSG_FIRSTHDR(msg); c; c = CMSG_NXTHDR(msg, c)) {
		if (c->cmsg_level != SOL_IPV6)
			continue;
		switch(c->cmsg_type) {
		case IPV6_HOPLIMIT:
#ifdef IPV6_2292HOPLIMIT
		case IPV6_2292HOPLIMIT:
#endif
			if (c->cmsg_len < CMSG_LEN(sizeof(int)))
				continue;
			memcpy(&hops, CMSG_DATA(c), sizeof(hops));
		}
	}


	/* Now the ICMP part */

	icmph = (struct icmp6_hdr *) buf;
	if (cc < 8) {
		if (g_options & F_VERBOSE)
			bb_error_msg("packet too short (%d bytes)", cc);
		return 1;
	}

	if (icmph->icmp6_type == ICMP6_ECHO_REPLY) {
		if (icmph->icmp6_id != ident)
			return 1;
		if (gather_statistics((__u8*)icmph, sizeof(*icmph), cc,
				      ntohs(icmph->icmp6_seq),
				      hops, 0, tv, pr_addr6(&from->sin6_addr),
				      pr_echo_reply6))
			return 0;
	} else {
		int nexthdr;
		struct ip6_hdr *iph1 = (struct ip6_hdr*)(icmph+1);
		struct icmp6_hdr *icmph1 = (struct icmp6_hdr *)(iph1+1);

		/* We must not ever fall here. All the messages but
		 * echo reply are blocked by filter and error are
		 * received with IPV6_RECVERR. Ugly code is preserved
		 * however, just to remember what crap we avoided
		 * using RECVRERR. :-)
		 */

		if (cc < (int)(8+sizeof(struct ip6_hdr)+8))
			return 1;

		if (memcmp(&iph1->ip6_dst, &whereto6.sin6_addr, 16))
			return 1;

		nexthdr = iph1->ip6_nxt;

		if (nexthdr == 44) {
			nexthdr = *(__u8*)icmph1;
			icmph1++;
		}
		if (nexthdr == IPPROTO_ICMPV6) {
			if (icmph1->icmp6_type != ICMP6_ECHO_REQUEST ||
			    icmph1->icmp6_id != ident)
				return 1;
			acknowledge(ntohs(icmph1->icmp6_seq));
			if (working_recverr)
				return 0;
			nerrors++;
			if (g_options & F_FLOOD) {
				write_stdout("\bE", 2);
				return 0;
			}
			print_timestamp();
			printf("From %s: icmp_seq=%u ", pr_addr6(&from->sin6_addr), ntohs(icmph1->icmp6_seq));
		} else {
			/* We've got something other than an ECHOREPLY */
			if (!(g_options & F_VERBOSE) || uid)
				return 1;
			print_timestamp();
			printf("From %s: ", pr_addr6(&from->sin6_addr));
		}
		pr_icmph6(icmph->icmp6_type, icmph->icmp6_code, ntohl(icmph->icmp6_mtu));
	}

	if (!(g_options & F_FLOOD)) {
		if (g_options & F_AUDIBLE)
			putchar('\a');
		putchar('\n');
		fflush(stdout);
	} else {
		putchar('\a');
		fflush(stdout);
	}
	return 0;
}


static int pr_icmph6(__u8 type, __u8 code, __u32 info)
{
	switch(type) {
	case ICMP6_DST_UNREACH:
		printf("Destination unreachable: ");
		switch (code) {
		case ICMP6_DST_UNREACH_NOROUTE:
			printf("No route");
			break;
		case ICMP6_DST_UNREACH_ADMIN:
			printf("Administratively prohibited");
			break;
		case ICMP6_DST_UNREACH_BEYONDSCOPE:
			printf("Beyond scope of source address");
			break;
		case ICMP6_DST_UNREACH_ADDR:
			printf("Address unreachable");
			break;
		case ICMP6_DST_UNREACH_NOPORT:
			printf("Port unreachable");
			break;
		default:
			printf("Unknown code %d", code);
			break;
		}
		break;
	case ICMP6_PACKET_TOO_BIG:
		printf("Packet too big: mtu=%u", info);
		if (code)
			printf(", code=%d", code);
		break;
	case ICMP6_TIME_EXCEEDED:
		printf("Time exceeded: ");
		if (code == ICMP6_TIME_EXCEED_TRANSIT)
			printf("Hop limit");
		else if (code == ICMP6_TIME_EXCEED_REASSEMBLY)
			printf("Defragmentation failure");
		else
			printf("code %d", code);
		break;
	case ICMP6_PARAM_PROB:
		printf("Parameter problem: ");
		if (code == ICMP6_PARAMPROB_HEADER)
			printf("Wrong header field ");
		else if (code == ICMP6_PARAMPROB_NEXTHEADER)
			printf("Unknown header ");
		else if (code == ICMP6_PARAMPROB_OPTION)
			printf("Unknown option ");
		else
			printf("code %d ", code);
		printf ("at %u", info);
		break;
	case ICMP6_ECHO_REQUEST:
		printf("Echo request");
		break;
	case ICMP6_ECHO_REPLY:
		printf("Echo reply");
		break;
	case MLD_LISTENER_QUERY:
		printf("MLD Query");
		break;
	case MLD_LISTENER_REPORT:
		printf("MLD Report");
		break;
	case MLD_LISTENER_REDUCTION:
		printf("MLD Reduction");
		break;
	default:
		printf("unknown icmp type: %u", type);

	}
	return 0;
}

static void install_filter_ping6(void)
{
	if (filter_ping6.filter)
		return;

	memcpy(insns_ping6, insns_ping6_st, sizeof(insns_ping6));
	filter_ping6.len = sizeof insns_ping6 / sizeof(insns_ping6[0]);
	filter_ping6.filter = insns_ping6;

	/* Patch bpflet for current identifier. */
	insns_ping6[1] = (struct sock_filter)BPF_JUMP(BPF_JMP|BPF_JEQ|BPF_K, htons(ident), 0, 1);

	if (setsockopt(icmp_sock, SOL_SOCKET, SO_ATTACH_FILTER, &filter_ping6, sizeof(filter_ping6)))
		bb_perror_msg("WARNING: failed to install socket filter");
}


int ping6_main(void)
{
	int argc = Gargc;
	char **argv = Gargv;
	int ch, hold, packlen;
	u_char *packet;
	char *target;
	struct sockaddr_in6 firsthop;
#ifdef ENABLE_PING6_RTHDR
	struct sockaddr_in6 ghbn;
#endif
	int socket_errno;
	struct icmp6_filter filter;
	int err;
	uint32_t scope_id = 0;
	__u32 flowlabel = 0;
	__u32 tclass = 0;
	struct sockaddr_in6 source6;
#ifdef ENABLE_PING6_RTHDR
	struct cmsghdr *srcrt = NULL;
#endif

	install_filter = install_filter_ping6;
	receive_error_msg = receive_error_msg_ping6;
	send_probe = send_probe_ping6;
	parse_reply = parse_reply_ping6;

	common_data_init();

	limit_capabilities();

#ifdef USE_IDN
	setlocale(LC_ALL, "");
#endif

	enable_capability_raw();

	icmp_sock = bb_xsocket(BB_XSOCKET_FLG_AF_INET6 | BB_XSOCKET_FLG_SOCK_RAW | BB_XSOCKET_FLG_PROTO_SPEC | BB_XSOCKET_FLG_NO_WARN,
				IPPROTO_ICMPV6);
	socket_errno = errno;

	disable_capability_raw();

	memset(&source6, 0, sizeof(source6));
	source6.sin6_family = AF_INET6;
	memset(&firsthop, 0, sizeof(firsthop));
	firsthop.sin6_family = AF_INET6;

	preload = 1;
	while ((ch = getopt(argc, argv, COMMON_OPTSTR "F:")) != EOF) {
		switch(ch) {
		case 'F':
			flowlabel = hextoui(optarg);
			if (errno || (flowlabel & ~IPV6_FLOWINFO_FLOWLABEL))
				bb_perror_msg_and_die("Invalid flowinfo %s", optarg);
			g_options |= F_FLOWINFO;
			break;
		case 'Q':
			tclass = hextoui(optarg);
			if (errno || (tclass & ~0xff))
				bb_perror_msg_and_die("Invalid tclass %s", optarg);
			g_options |= F_TCLASS;
			break;
		case 'I':
			if (strchr(optarg, ':')) {
				char *p, *addr = xstrdup(optarg);

				p = strchr_set0(addr, SCOPE_DELIMITER);
				if (p)
					device = optarg + (p - addr) + 1;

				if (inet_pton(AF_INET6, addr, (char*)&source6.sin6_addr) <= 0)
					bb_perror_msg_and_die("invalid source address %s", optarg);

				g_options |= F_STRICTSOURCE;

				free(addr);
			} else {
				device = optarg;
			}
			break;
		case 'M':
			if (strcmp(optarg, "do") == 0)
				pmtudisc = IPV6_PMTUDISC_DO;
			else if (strcmp(optarg, "dont") == 0)
				pmtudisc = IPV6_PMTUDISC_DONT;
			else if (strcmp(optarg, "want") == 0)
				pmtudisc = IPV6_PMTUDISC_WANT;
			else
				bb_perror_msg_and_die("wrong value for -M: do, dont, want are valid ones.");
			break;
		COMMON_OPTIONS
			common_options(ch);
			break;
		default:
			bb_show_usage();
		}
	}
	argc -= optind;
	argv += optind;

#ifdef ENABLE_PING6_RTHDR
	while (argc > 1) {
		struct in6_addr *addr;

		if (srcrt == NULL) {
			int space;

			bb_perror_msg("Warning: Source routing is deprecated by RFC5095.");

#ifdef ENABLE_PING6_RTHDR_RFC3542
			space = inet6_rth_space(IPV6_RTHDR_TYPE_0, argc - 1);
#else
			space = inet6_srcrt_space(IPV6_SRCRT_TYPE_0, argc - 1);
#endif
			if (space == 0)
				bb_perror_msg_and_die("srcrt_space failed");
#ifdef ENABLE_PING6_RTHDR_RFC3542
			if (cmsglen + CMSG_SPACE(space) > sizeof(cmsgbuf))
				bb_perror_msg_and_die("no room for options");
#else
			if (space + cmsglen > sizeof(cmsgbuf))
				bb_perror_msg_and_die("no room for options");
#endif
			srcrt = (struct cmsghdr*)(cmsgbuf+cmsglen);
#ifdef ENABLE_PING6_RTHDR_RFC3542
			memset(srcrt, 0, CMSG_SPACE(0));
			srcrt->cmsg_len = CMSG_LEN(space);
			srcrt->cmsg_level = IPPROTO_IPV6;
			srcrt->cmsg_type = IPV6_RTHDR;
			inet6_rth_init(CMSG_DATA(srcrt), space, IPV6_RTHDR_TYPE_0, argc - 1);
			cmsglen += CMSG_SPACE(space);
#else
			cmsglen += CMSG_ALIGN(space);
			inet6_srcrt_init(srcrt, IPV6_SRCRT_TYPE_0);
#endif
		}

		target = *argv;

		bb_gethostbyname(BB_GHBN_SOCKADDR_IN6 | BB_GHBN_ERR_DIE, target, &ghbn);
		addr = &ghbn.sin6_addr;
#ifdef ENABLE_PING6_RTHDR_RFC3542
		inet6_rth_add(CMSG_DATA(srcrt), addr);
#else
		inet6_srcrt_add(srcrt, addr);
#endif
		if (IN6_IS_ADDR_UNSPECIFIED(&firsthop.sin6_addr)) {
			memcpy(&firsthop.sin6_addr, addr, 16);
#ifdef HAVE_SIN6_SCOPEID
			firsthop.sin6_scope_id = ghbn.sin6_scope_id;
			/* Verify scope_id is the same as previous nodes */
			if (firsthop.sin6_scope_id && scope_id && firsthop.sin6_scope_id != scope_id)
				bb_perror_msg_and_die("scope discrepancy among the nodes");
			 else if (!scope_id)
				scope_id = firsthop.sin6_scope_id;
#endif
		}

		argv++;
		argc--;
	}
#endif

	if (argc > 1) {
#ifndef ENABLE_PING6_RTHDR
		bb_perror_msg("Source routing is deprecated by RFC5095.");
#endif
		bb_show_usage();
	} else if (argc == 1) {
		target = *argv;
	} else {
		bb_show_usage();
	}
	bb_gethostbyname(BB_GHBN_SOCKADDR_IN6 | BB_GHBN_ERR_DIE, target, &whereto6);
	whereto6.sin6_port = htons(IPPROTO_ICMPV6);

	if (strchr(target, ':'))
		g_options |= F_NUMERIC;

	if (IN6_IS_ADDR_UNSPECIFIED(&firsthop.sin6_addr)) {
		memcpy(&firsthop.sin6_addr, &whereto6.sin6_addr, 16);
#ifdef HAVE_SIN6_SCOPEID
		firsthop.sin6_scope_id = whereto6.sin6_scope_id;
		/* Verify scope_id is the same as intermediate nodes */
		if (firsthop.sin6_scope_id && scope_id && firsthop.sin6_scope_id != scope_id)
			bb_perror_msg_and_die("scope discrepancy among the nodes");
		else if (!scope_id)
			scope_id = firsthop.sin6_scope_id;
#endif
	}

	hostname = target;

	if (IN6_IS_ADDR_UNSPECIFIED(&source6.sin6_addr)) {
		socklen_t alen;
		int probe_fd = bb_xsocket(BB_XSOCKET_FLG_AF_INET6 | BB_XSOCKET_FLG_SOCK_DGRAM | BB_XSOCKET_FLG_ERR_DIE);

		if (device) {
#if defined(IPV6_RECVPKTINFO) || defined(HAVE_SIN6_SCOPEID)
			unsigned int iface = if_name2index(device);
#endif
#ifdef IPV6_RECVPKTINFO
			struct in6_pktinfo ipi;

			memset(&ipi, 0, sizeof(ipi));
			ipi.ipi6_ifindex = iface;
#endif

#ifdef HAVE_SIN6_SCOPEID
			if (IN6_IS_ADDR_LINKLOCAL(&firsthop.sin6_addr) ||
			    IN6_IS_ADDR_MC_LINKLOCAL(&firsthop.sin6_addr))
				firsthop.sin6_scope_id = iface;
#endif
			enable_capability_raw();
			if (
#ifdef IPV6_RECVPKTINFO
			    setsockopt(probe_fd, IPPROTO_IPV6, IPV6_PKTINFO, &ipi, sizeof(ipi)) == -1 &&
#endif
			    setsockopt(probe_fd, SOL_SOCKET, SO_BINDTODEVICE, device, strlen(device)+1) == -1)
				bb_perror_msg_and_die("setsockopt(SO_BINDTODEVICE)");
			disable_capability_raw();
		}
		firsthop.sin6_port = htons(1025);
		if (connect(probe_fd, (struct sockaddr*)&firsthop, sizeof(firsthop)) == -1)
			bb_perror_msg_and_die("connect");
		alen = sizeof(source6);
		if (getsockname(probe_fd, (struct sockaddr*)&source6, &alen) == -1)
			bb_perror_msg_and_die("getsockname");
		source6.sin6_port = 0;
		close(probe_fd);

#ifndef WITHOUT_IFADDRS
		if (device) {
			struct ifaddrs *ifa0, *ifa;

			if (getifaddrs(&ifa0))
				bb_perror_msg_and_die("getifaddrs");

			for (ifa = ifa0; ifa; ifa = ifa->ifa_next) {
				if (!ifa->ifa_addr || ifa->ifa_addr->sa_family != AF_INET6)
					continue;
				if (!strncmp(ifa->ifa_name, device, sizeof(device) - 1) &&
				    IN6_ARE_ADDR_EQUAL(&((struct sockaddr_in6 *)ifa->ifa_addr)->sin6_addr,
						       &source6.sin6_addr))
					break;
			}
			if (!ifa)
				bb_perror_msg("Warning: source address might be selected on device other than %s.", device);

			freeifaddrs(ifa0);
		}
#endif
	}
#ifdef HAVE_SIN6_SCOPEID
	else if (device && (IN6_IS_ADDR_LINKLOCAL(&source6.sin6_addr) ||
			    IN6_IS_ADDR_MC_LINKLOCAL(&source6.sin6_addr)))
		source6.sin6_scope_id = if_name2index(device);
#endif

	if (icmp_sock < 0) {
		errno = socket_errno;
		bb_perror_msg_and_die("icmp open socket");
	}

	if (device) {
		struct cmsghdr *cmsg;
		struct in6_pktinfo *ipi;

		cmsg = (struct cmsghdr*)(cmsgbuf+cmsglen);
		cmsglen += CMSG_SPACE(sizeof(*ipi));
		cmsg->cmsg_len = CMSG_LEN(sizeof(*ipi));
		cmsg->cmsg_level = SOL_IPV6;
		cmsg->cmsg_type = IPV6_PKTINFO;

		ipi = (struct in6_pktinfo*)CMSG_DATA(cmsg);
		memset(ipi, 0, sizeof(*ipi));
		ipi->ipi6_ifindex = if_name2index(device);
	}

	if ((whereto6.sin6_addr.s6_addr16[0]&htons(0xff00)) == htons (0xff00)) {
		if (uid) {
			if (interval < 1000)
				bb_perror_msg_and_die("multicast ping with too short interval.");
			if (pmtudisc >= 0 && pmtudisc != IPV6_PMTUDISC_DO)
				bb_perror_msg_and_die("multicast ping does not fragment.");
		}
		if (pmtudisc < 0)
			pmtudisc = IPV6_PMTUDISC_DO;
	}

	if (pmtudisc >= 0) {
		if (setsockopt_1i(icmp_sock, SOL_IPV6, IPV6_MTU_DISCOVER, pmtudisc) == -1)
			bb_perror_msg_and_die("IPV6_MTU_DISCOVER");
	}

	if ((g_options & F_STRICTSOURCE))
		bb_xbind(icmp_sock, (struct sockaddr*)&source6, sizeof(source6));

	if (datalen >= (int)(sizeof(struct timeval))) {
		/* can we time transfer */
		timing = 1;
	}
	packlen = datalen + 8 + 4096 + 40 + 8; /* 4096 for rthdr */
	packet = (u_char *)xmalloc((u_int)packlen);

	working_recverr = 1;
	if (setsockopt_1i(icmp_sock, SOL_IPV6, IPV6_RECVERR, 1)) {
		bb_error_msg("WARNING: your kernel is veeery old. No problems.");
		working_recverr = 0;
	}

	/* Estimate memory eaten by single packet. It is rough estimate.
	 * Actually, for small datalen's it depends on kernel side a lot. */
	hold = datalen+8;
	hold += ((hold+511)/512)*(40+16+64+160);
	sock_setbufs(hold);

#ifdef __linux__
	if (setsockopt_1i(icmp_sock, SOL_RAW, IPV6_CHECKSUM, 2) < 0) {
		/* checksum should be enabled by default and setting this
		 * option might fail anyway.
		 */
		bb_perror_msg("setsockopt(RAW_CHECKSUM) failed - try to continue.");
	}
#endif

	/*
	 *      select icmp echo reply as icmp type to receive
	 */

	ICMP6_FILTER_SETBLOCKALL(&filter);

	if (!working_recverr) {
		ICMP6_FILTER_SETPASS(ICMP6_DST_UNREACH, &filter);
		ICMP6_FILTER_SETPASS(ICMP6_PACKET_TOO_BIG, &filter);
		ICMP6_FILTER_SETPASS(ICMP6_TIME_EXCEEDED, &filter);
		ICMP6_FILTER_SETPASS(ICMP6_PARAM_PROB, &filter);
	}

	ICMP6_FILTER_SETPASS(ICMP6_ECHO_REPLY, &filter);

	err = setsockopt(icmp_sock, IPPROTO_ICMPV6, ICMP6_FILTER, &filter,
			 sizeof(struct icmp6_filter));

	if (err < 0)
		bb_perror_msg_and_die("setsockopt(ICMP6_FILTER)");

	if (g_options & F_NOLOOP) {
		if (setsockopt_1i(icmp_sock, IPPROTO_IPV6, IPV6_MULTICAST_LOOP, 0) == -1)
			bb_perror_msg_and_die ("can't disable multicast loopback");
	}
	if (g_options & F_TTL) {
		if (setsockopt_1i(icmp_sock, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, g_ttl) == -1)
			bb_perror_msg_and_die ("can't set multicast hop limit");
		if (setsockopt_1i(icmp_sock, IPPROTO_IPV6, IPV6_UNICAST_HOPS, g_ttl) == -1)
			bb_perror_msg_and_die ("can't set unicast hop limit");
	}

	if (1) {
		if (
#ifdef IPV6_RECVHOPLIMIT
		    setsockopt_1i(icmp_sock, IPPROTO_IPV6, IPV6_RECVHOPLIMIT, 1) == -1 &&
		    setsockopt_1i(icmp_sock, IPPROTO_IPV6, IPV6_2292HOPLIMIT, 1) == -1
#else
		    setsockopt_1i(icmp_sock, IPPROTO_IPV6, IPV6_HOPLIMIT, 1) == -1
#endif
		   )
			bb_perror_msg_and_die ("can't receive hop limit");
	}

	if (g_options & F_TCLASS) {
#ifdef IPV6_TCLASS
		if (setsockopt(icmp_sock, IPPROTO_IPV6, IPV6_TCLASS,
			       &tclass, sizeof(tclass)) == -1)
			bb_perror_msg_and_die ("setsockopt(IPV6_TCLASS)");
#else
		bb_perror_msg("Traffic class is not supported.");
#endif
	}

	if (g_options & F_FLOWINFO) {
#ifdef IPV6_FLOWLABEL_MGR
		char freq_buf[CMSG_ALIGN(sizeof(struct in6_flowlabel_req)) + cmsglen];
		struct in6_flowlabel_req *freq = (struct in6_flowlabel_req *)freq_buf;
		int freq_len = sizeof(*freq);
#ifdef ENABLE_PING6_RTHDR
		if (srcrt)
			freq_len = CMSG_ALIGN(sizeof(*freq)) + srcrt->cmsg_len;
#endif
		memset(freq, 0, sizeof(*freq));
		freq->flr_label = htonl(flowlabel & IPV6_FLOWINFO_FLOWLABEL);
		freq->flr_action = IPV6_FL_A_GET;
		freq->flr_flags = IPV6_FL_F_CREATE;
		freq->flr_share = IPV6_FL_S_EXCL;
		memcpy(&freq->flr_dst, &whereto6.sin6_addr, 16);
#ifdef ENABLE_PING6_RTHDR
		if (srcrt)
			memcpy(freq_buf + CMSG_ALIGN(sizeof(*freq)), srcrt, srcrt->cmsg_len);
#endif
		if (setsockopt(icmp_sock, IPPROTO_IPV6, IPV6_FLOWLABEL_MGR, freq, freq_len) == -1)
			bb_perror_msg_and_die ("can't set flowlabel");
		flowlabel = freq->flr_label;
#ifdef ENABLE_PING6_RTHDR
		if (srcrt) {
			cmsglen = (char*)srcrt - (char*)cmsgbuf;
			srcrt = NULL;
		}
#endif
#else
		bb_perror_msg_and_die("Flow labels are not supported.");
#endif

#ifdef IPV6_FLOWINFO_SEND
		whereto6.sin6_flowinfo = flowlabel;
		if (setsockopt_1i(icmp_sock, IPPROTO_IPV6, IPV6_FLOWINFO_SEND, 1) == -1)
			bb_perror_msg_and_die ("can't send flowinfo");
#else
		bb_perror_msg_and_die("Flowinfo is not supported.");
#endif
	}

	printf("PING %s(%s) ", hostname, pr_addr6(&whereto6.sin6_addr));
	if (flowlabel)
		printf(", flow 0x%05x, ", (unsigned)ntohl(flowlabel));
	if (device || (g_options & F_STRICTSOURCE)) {
		char hbuf[NI_MAXHOST];
		printf("from %s %s: ", pr_addr_n(&source6.sin6_addr, hbuf), device ? : "");
	}
	printf("%d data bytes\n", datalen);

	main_loop(packet, packlen);
}
#endif /* CONFIG_PING6 */

#endif /* APPLET_USAGE */

/*
 * Copyright (c) 1989 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Mike Muuss.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
