/*
     pptp-linux is an implementation of the PPTP protocol for Linux and
     other Unix systems.

     New versions and additional documentation can be found on the PPTP
     Client project page, at http://pptpclient.sourceforge.net/

     (C) 200-2003   C. Scott Ananian <cananian@alumni.princeton.edu>
     GNU GENERAL PUBLIC LICENSE Version 2, June 1991

     Busybox port by Vladimir Oleynik (C) 2004 - 2008 <dzo@simtreas.ru>

*/
#ifdef APPLET_USAGE
#define pptp_trivial_usage \
	"<hostname> [<pptp options>] [[--] <pppd options>]"
#define pptp_full_usage \
	"PPTP client daemon.\n" \
	"Call pptp -h for more options"
#else

#define PPTP_LINUX_VERSION "1.4.0"
#define PPPD_BINARY "/usr/sbin/pppd"

//#define SUPPORT_BEZEQ_ISRAEL

#define PPTP_SOCKET_PREFIX "/var/run/pptp/"

#include "busybox.h"
#include "bb_inet.h"
#include "bb_sig.h"
#include <sys/time.h>
#include <sys/un.h>
#include <sys/wait.h>

#include <fcntl.h>
#include <assert.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <syslog.h>
#include <setjmp.h>
#include <limits.h>

#include "pty_.h"

#include <netinet/in.h>

#ifdef DEBUG_PQUEUE
#define DEBUG_ON 1
#else
#define DEBUG_ON 0
#endif

#define DEBUG_CMD(_a) if (DEBUG_ON) { _a }

#ifdef DEBUG
#define DEBUG_LOG(l)    l
#else
#define DEBUG_LOG(l)
#endif

/* pptp_options.h ...... various constants used in the PPTP protocol.
 *                       #define STANDARD to emulate NT 4.0 exactly.
 */

#ifndef STANDARD
#define PPTP_MAX_CHANNELS 65535
#define PPTP_FIRMWARE_VERSION 0x001
#define PPTP_HOSTNAME {'l','o','c','a','l',0}
#define PPTP_VENDOR   {'c','a','n','a','n','i','a','n',0}
#define PPTP_FRAME_CAP  PPTP_FRAME_ANY
#define PPTP_BEARER_CAP PPTP_BEARER_ANY
#else
#define PPTP_MAX_CHANNELS 5
#define PPTP_FIRMWARE_VERSION 0
#define PPTP_HOSTNAME {'l','o','c','a','l',0}
#define PPTP_VENDOR   {'N','T',0}
#define PPTP_FRAME_CAP  2
#define PPTP_BEARER_CAP 1
#endif

/*  pptp.h:  packet structures and magic constants for the PPTP protocol
 *
 * pptp_msg.h v. 1.3 2003/02/15 10:37:21 quozl
 */

/* define "portable" htons, etc. */
#define hton8(x)  (x)
#define ntoh8(x)  (x)
#define hton16(x) htons(x)
#define ntoh16(x) ntohs(x)
#define hton32(x) htonl(x)
#define ntoh32(x) ntohl(x)

/* PPTP magic numbers: ----------------------------------------- */

#define PPTP_MAGIC 0x1A2B3C4D /* Magic cookie for PPTP datagrams */
#define PPTP_PORT  1723       /* PPTP TCP port number            */
#define PPTP_PROTO 47         /* PPTP IP protocol number         */

/* Control Connection Message Types: --------------------------- */

#define PPTP_MESSAGE_CONTROL            1
#define PPTP_MESSAGE_MANAGE             2

/* Control Message Types: -------------------------------------- */

/* (Control Connection Management) */
#define PPTP_START_CTRL_CONN_RQST       1
#define PPTP_START_CTRL_CONN_RPLY       2
#define PPTP_STOP_CTRL_CONN_RQST        3
#define PPTP_STOP_CTRL_CONN_RPLY        4
#define PPTP_ECHO_RQST                  5
#define PPTP_ECHO_RPLY                  6

/* (Call Management) */
#define PPTP_OUT_CALL_RQST              7
#define PPTP_OUT_CALL_RPLY              8
#define PPTP_IN_CALL_RQST               9
#define PPTP_IN_CALL_RPLY               10
#define PPTP_IN_CALL_CONNECT            11
#define PPTP_CALL_CLEAR_RQST            12
#define PPTP_CALL_CLEAR_NTFY            13

/* (Error Reporting) */
#define PPTP_WAN_ERR_NTFY               14

/* (PPP Session Control) */
#define PPTP_SET_LINK_INFO              15

/* PPTP version information: --------------------------------------*/
#define PPTP_VERSION            0x100

/* PPTP capabilities: ---------------------------------------------*/

/* (Framing capabilities for msg sender) */
#define PPTP_FRAME_ASYNC        1
#define PPTP_FRAME_SYNC         2
#define PPTP_FRAME_ANY          3

/* (Bearer capabilities for msg sender) */
#define PPTP_BEARER_ANALOG      1
#define PPTP_BEARER_DIGITAL     2
#define PPTP_BEARER_ANY         3

#define PPTP_RESULT_GENERAL_ERROR 2

/* (Reasons to close a connection) */
#define PPTP_STOP_NONE            1 /* no good reason                        */
#define PPTP_STOP_PROTOCOL        2 /* can't support peer's protocol version */
#define PPTP_STOP_LOCAL_SHUTDOWN  3 /* requester is being shut down          */

/* PPTP datagram structures (all data in network byte order): ----------*/

struct pptp_header {
  u_int16_t length;       /* message length in octets, including header */
  u_int16_t pptp_type;    /* PPTP message type. 1 for control message.  */
  u_int32_t magic;        /* this should be PPTP_MAGIC.                 */
  u_int16_t ctrl_type;    /* Control message type (0-15)                */
  u_int16_t reserved0;    /* reserved.  MUST BE ZERO.                   */
};

struct pptp_start_ctrl_conn { /* for control message types 1 and 2 */
  struct pptp_header header;

  u_int16_t version;      /* PPTP protocol version.  = PPTP_VERSION     */
  u_int8_t  result_code;  /* these two fields should be zero on rqst msg*/
  u_int8_t  error_code;   /* 0 unless result_code==2 (General Error)    */
  u_int32_t framing_cap;  /* Framing capabilities                       */
  u_int32_t bearer_cap;   /* Bearer Capabilities                        */
  u_int16_t max_channels; /* Maximum Channels (=0 for PNS, PAC ignores) */
  u_int16_t firmware_rev; /* Firmware or Software Revision              */
  u_int8_t  hostname[64]; /* Host Name (64 octets, zero terminated)     */
  u_int8_t  vendor[64];   /* Vendor string (64 octets, zero term.)      */
  /* MS says that end of hostname/vendor fields should be filled with   */
  /* octets of value 0, but Win95 PPTP driver doesn't do this.          */
};

struct pptp_stop_ctrl_conn { /* for control message types 3 and 4 */
  struct pptp_header header;

  u_int8_t reason_result; /* reason for rqst, result for rply          */
  u_int8_t error_code;    /* MUST be 0, unless rply result==2 (general err)*/
  u_int16_t reserved1;    /* MUST be 0                                */
};

struct pptp_echo_rqst { /* for control message type 5 */
  struct pptp_header header;
  u_int32_t identifier;   /* arbitrary value set by sender which is used */
			  /* to match up reply and request               */
};

struct pptp_echo_rply { /* for control message type 6 */
  struct pptp_header header;
  u_int32_t identifier;   /* should correspond to id of rqst             */
  u_int8_t result_code;
  u_int8_t error_code;    /* =0, unless result_code==2 (general error)   */
  u_int16_t reserved1;    /* MUST BE ZERO                                */
};

struct pptp_out_call_rqst { /* for control message type 7 */
  struct pptp_header header;
  u_int16_t call_id;      /* Call ID (unique id used to multiplex data)  */
  u_int16_t call_sernum;  /* Call Serial Number (used for logging)       */
  u_int32_t bps_min;      /* Minimum BPS (lowest acceptable line speed)  */
  u_int32_t bps_max;      /* Maximum BPS (highest acceptable line speed) */
  u_int32_t bearer;       /* Bearer type                                 */
  u_int32_t framing;      /* Framing type                                */
  u_int16_t recv_size;    /* Recv. Window Size (no. of buffered packets) */
  u_int16_t delay;        /* Packet Processing Delay (in 1/10 sec)       */
  u_int16_t phone_len;    /* Phone Number Length (num. of valid digits)  */
  u_int16_t reserved1;    /* MUST BE ZERO                                */
  char      phone_num[64]; /* Phone Number (64 octets, null term.)       */
  u_int8_t subaddress[64]; /* Subaddress (64 octets, null term.)         */
};

struct pptp_out_call_rply { /* for control message type 8 */
  struct pptp_header header;
  u_int16_t call_id;      /* Call ID (used to multiplex data over tunnel)*/
  u_int16_t call_id_peer; /* Peer's Call ID (call_id of pptp_out_call_rqst)*/
  u_int8_t  result_code;  /* Result Code (1 is no errors)                */
  u_int8_t  error_code;   /* Error Code (=0 unless result_code==2)       */
  u_int16_t cause_code;   /* Cause Code (addt'l failure information)     */
  u_int32_t speed;        /* Connect Speed (in BPS)                      */
  u_int16_t recv_size;    /* Recv. Window Size (no. of buffered packets) */
  u_int16_t delay;        /* Packet Processing Delay (in 1/10 sec)       */
  u_int32_t channel;      /* Physical Channel ID (for logging)           */
};

struct pptp_in_call_rqst { /* for control message type 9 */
  struct pptp_header header;
  u_int16_t call_id;      /* Call ID (unique id used to multiplex data)  */
  u_int16_t call_sernum;  /* Call Serial Number (used for logging)       */
  u_int32_t bearer;       /* Bearer type                                 */
  u_int32_t channel;      /* Physical Channel ID (for logging)           */
  u_int16_t dialed_len;   /* Dialed Number Length (# of valid digits)    */
  u_int16_t dialing_len;  /* Dialing Number Length (# of valid digits)   */
  u_int8_t dialed_num[64]; /* Dialed Number (64 octets, zero term.)      */
  u_int8_t dialing_num[64]; /* Dialing Number (64 octets, zero term.)    */
  u_int8_t subaddress[64];  /* Subaddress (64 octets, zero term.)        */
};

struct pptp_in_call_rply { /* for control message type 10 */
  struct pptp_header header;
  u_int16_t call_id;      /* Call ID (used to multiplex data over tunnel)*/
  u_int16_t call_id_peer; /* Peer's Call ID (call_id of pptp_out_call_rqst)*/
  u_int8_t  result_code;  /* Result Code (1 is no errors)                */
  u_int8_t  error_code;   /* Error Code (=0 unless result_code==2)       */
  u_int16_t recv_size;    /* Recv. Window Size (no. of buffered packets) */
  u_int16_t delay;        /* Packet Processing Delay (in 1/10 sec)       */
  u_int16_t reserved1;    /* MUST BE ZERO                                */
};

struct pptp_in_call_connect { /* for control message type 11 */
  struct pptp_header header;
  u_int16_t call_id_peer; /* Peer's Call ID (call_id of pptp_out_call_rqst)*/
  u_int16_t reserved1;    /* MUST BE ZERO                                */
  u_int32_t speed;        /* Connect Speed (in BPS)                      */
  u_int16_t recv_size;    /* Recv. Window Size (no. of buffered packets) */
  u_int16_t delay;        /* Packet Processing Delay (in 1/10 sec)       */
  u_int32_t framing;      /* Framing type                                */
};

struct pptp_call_clear_rqst { /* for control message type 12 */
  struct pptp_header header;
  u_int16_t call_id;      /* Call ID (used to multiplex data over tunnel)*/
  u_int16_t reserved1;    /* MUST BE ZERO                                */
};

struct pptp_call_clear_ntfy { /* for control message type 13 */
  struct pptp_header header;
  u_int16_t call_id;      /* Call ID (used to multiplex data over tunnel)*/
  u_int8_t  result_code;  /* Result Code                                 */
  u_int8_t  error_code;   /* Error Code (=0 unless result_code==2)       */
  u_int16_t cause_code;   /* Cause Code (for ISDN, is Q.931 cause code)  */
  u_int16_t reserved1;    /* MUST BE ZERO                                */
  u_int8_t call_stats[128]; /* Call Statistics: 128 octets, ascii, 0-term */
};

struct pptp_wan_err_ntfy {    /* for control message type 14 */
  struct pptp_header header;
  u_int16_t call_id_peer; /* Peer's Call ID (call_id of pptp_out_call_rqst)*/
  u_int16_t reserved1;    /* MUST BE ZERO                                */
  u_int32_t crc_errors;   /* CRC errors                                  */
  u_int32_t frame_errors; /* Framing errors                              */
  u_int32_t hard_errors;  /* Hardware overruns                           */
  u_int32_t buff_errors;  /* Buffer overruns                             */
  u_int32_t time_errors;  /* Time-out errors                             */
  u_int32_t align_errors; /* Alignment errors                            */
};

struct pptp_set_link_info {   /* for control message type 15 */
  struct pptp_header header;
  u_int16_t call_id_peer; /* Peer's Call ID (call_id of pptp_out_call_rqst) */
  u_int16_t reserved1;    /* MUST BE ZERO                                   */
  u_int32_t send_accm;    /* Send ACCM (for PPP packets; default 0xFFFFFFFF)*/
  u_int32_t recv_accm;    /* Receive ACCM (for PPP pack.;default 0xFFFFFFFF)*/
};

/* helpful #defines: -------------------------------------------- */
#define pptp_isvalid_ctrl(header, type, length) \
 (!( ( ntoh16(((struct pptp_header *)header)->length)    < (length)  ) ||   \
     ( ntoh16(((struct pptp_header *)header)->pptp_type) !=(type)    ) ||   \
     ( ntoh32(((struct pptp_header *)header)->magic)     !=PPTP_MAGIC) ||   \
     ( ntoh16(((struct pptp_header *)header)->ctrl_type) > PPTP_SET_LINK_INFO) || \
     ( ntoh16(((struct pptp_header *)header)->reserved0) !=0         ) ))

#define PPTP_HEADER_CTRL(type)  \
{ hton16(PPTP_CTRL_SIZE(type)), \
  hton16(PPTP_MESSAGE_CONTROL), \
  hton32(PPTP_MAGIC),           \
  hton16(type), 0 }

#define PPTP_CTRL_SIZE(type) ( \
(type==PPTP_START_CTRL_CONN_RQST)?sizeof(struct pptp_start_ctrl_conn):  \
(type==PPTP_START_CTRL_CONN_RPLY)?sizeof(struct pptp_start_ctrl_conn):  \
(type==PPTP_STOP_CTRL_CONN_RQST )?sizeof(struct pptp_stop_ctrl_conn):   \
(type==PPTP_STOP_CTRL_CONN_RPLY )?sizeof(struct pptp_stop_ctrl_conn):   \
(type==PPTP_ECHO_RQST           )?sizeof(struct pptp_echo_rqst):        \
(type==PPTP_ECHO_RPLY           )?sizeof(struct pptp_echo_rply):        \
(type==PPTP_OUT_CALL_RQST       )?sizeof(struct pptp_out_call_rqst):    \
(type==PPTP_OUT_CALL_RPLY       )?sizeof(struct pptp_out_call_rply):    \
(type==PPTP_IN_CALL_RQST        )?sizeof(struct pptp_in_call_rqst):     \
(type==PPTP_IN_CALL_RPLY        )?sizeof(struct pptp_in_call_rply):     \
(type==PPTP_IN_CALL_CONNECT     )?sizeof(struct pptp_in_call_connect):  \
(type==PPTP_CALL_CLEAR_RQST     )?sizeof(struct pptp_call_clear_rqst):  \
(type==PPTP_CALL_CLEAR_NTFY     )?sizeof(struct pptp_call_clear_ntfy):  \
(type==PPTP_WAN_ERR_NTFY        )?sizeof(struct pptp_wan_err_ntfy):     \
(type==PPTP_SET_LINK_INFO       )?sizeof(struct pptp_set_link_info):    \
0)
#define max(a,b) (((a)>(b))?(a):(b))
#define PPTP_CTRL_SIZE_MAX (                    \
max(sizeof(struct pptp_start_ctrl_conn),        \
max(sizeof(struct pptp_echo_rqst),              \
max(sizeof(struct pptp_echo_rply),              \
max(sizeof(struct pptp_out_call_rqst),          \
max(sizeof(struct pptp_out_call_rply),          \
max(sizeof(struct pptp_in_call_rqst),           \
max(sizeof(struct pptp_in_call_rply),           \
max(sizeof(struct pptp_in_call_connect),        \
max(sizeof(struct pptp_call_clear_rqst),        \
max(sizeof(struct pptp_call_clear_ntfy),        \
max(sizeof(struct pptp_wan_err_ntfy),           \
max(sizeof(struct pptp_set_link_info), 0)))))))))))))


/* gre header structure: -------------------------------------------- */

#define PPTP_GRE_PROTO  0x880B
#define PPTP_GRE_VER    0x1

#define PPTP_GRE_FLAG_C 0x80
#define PPTP_GRE_FLAG_R 0x40
#define PPTP_GRE_FLAG_K 0x20
#define PPTP_GRE_FLAG_S 0x10
#define PPTP_GRE_FLAG_A 0x80

#define PPTP_GRE_IS_C(f) ((f)&PPTP_GRE_FLAG_C)
#define PPTP_GRE_IS_R(f) ((f)&PPTP_GRE_FLAG_R)
#define PPTP_GRE_IS_K(f) ((f)&PPTP_GRE_FLAG_K)
#define PPTP_GRE_IS_S(f) ((f)&PPTP_GRE_FLAG_S)
#define PPTP_GRE_IS_A(f) ((f)&PPTP_GRE_FLAG_A)

struct pptp_gre_header {
  u_int8_t flags;               /* bitfield */
  u_int8_t ver;                 /* should be PPTP_GRE_VER (enhanced GRE) */
  u_int16_t protocol;           /* should be PPTP_GRE_PROTO (ppp-encaps) */
  u_int16_t payload_len;        /* size of ppp payload, not inc. gre header */
  u_int16_t call_id;            /* peer's call_id for this session */
  u_int32_t seq;                /* sequence number.  Present if S==1 */
  u_int32_t ack;                /* seq number of highest packet recieved by */
				/*  sender in this session */
};


#ifdef SUPPORT_BEZEQ_ISRAEL
/* orckit_quirks.h ...... fix quirks in orckit adsl modems
 *                        mulix <mulix@actcom.co.il>
 */

/* return 0 on success, non zero otherwise */
static int
orckit_atur3_build_hook(struct pptp_out_call_rqst* packt);

/* return 0 on success, non zero otherwise */
static int
orckit_atur3_set_link_hook(struct pptp_set_link_info* packet,
			   int peer_call_id);

/* return 0 on success, non zero otherwise */
static int
orckit_atur3_start_ctrl_conn_hook(struct pptp_start_ctrl_conn* packet);

#endif /* SUPPORT_BEZEQ_ISRAEL */

/* ppp_fcs.h ... header file for PPP-HDLC FCS
 */

#define PPPINITFCS16    0xffff  /* Initial FCS value */
#define PPPGOODFCS16    0xf0b8  /* Good final FCS value */

/* pptp_callmgr.h ... Call manager for PPTP connections.
 *                    Handles TCP port 1723 protocol.
 */

enum call_state { CALL_OPEN_RQST,  CALL_OPEN_DONE, CALL_OPEN_FAIL,
		  CALL_CLOSE_RQST, CALL_CLOSE_DONE };
enum conn_state { CONN_OPEN_RQST,  CONN_OPEN_DONE, CONN_OPEN_FAIL,
		  CONN_CLOSE_RQST, CONN_CLOSE_DONE };

typedef struct PPTP_CONN PPTP_CONN;
typedef void (*pptp_conn_cb)(PPTP_CONN *, enum conn_state);
typedef struct PPTP_CALL PPTP_CALL;
typedef void (*pptp_call_cb)(PPTP_CONN*, PPTP_CALL*, enum call_state);

/* vector.h ..... store a vector of PPTP_CALL information and search it
 *                efficiently.
 */

struct vector_item {
    int key;
    PPTP_CALL *call;
};

struct vector_struct {
    struct vector_item *item;
    int size;
    int alloc;
};

typedef struct vector_struct VECTOR;

/* pptp_ctrl.h ... handle PPTP control connection. */

/* This structure contains connection-specific information that the
 * signal handler needs to see.  Thus, it needs to be in a global
 * variable.  If you end up using pthreads or something (why not
 * just processes?), this would have to be placed in a thread-specific
 * data area, using pthread_get|set_specific, etc., so I've
 * conveniently encapsulated it for you.
 * [linux threads will have to support thread-specific signals
 *  before this would work at all, which, as of this writing
 *  (linux-threads v0.6, linux kernel 2.1.72), it does not.]
 */

#define INITIAL_BUFSIZE 512 /* initial I/O buffer size. */


struct PPTP_CONN {
    int inet_sock;
    /* Connection States */
    enum {
	CONN_IDLE, CONN_WAIT_CTL_REPLY, CONN_WAIT_STOP_REPLY, CONN_ESTABLISHED
    } conn_state; /* on startup: CONN_IDLE */
    /* Keep-alive states */
    enum {
	KA_NONE, KA_OUTSTANDING
    } ka_state;  /* on startup: KA_NONE */
    /* Keep-alive ID; monotonically increasing (watch wrap-around!) */
    u_int32_t ka_id; /* on startup: 1 */
    /* Other properties. */
    u_int16_t version;
    u_int16_t firmware_rev;
    u_int8_t  hostname[64], vendor[64];
    /* XXX these are only PNS properties, currently XXX */
    /* Call assignment information. */
    u_int16_t call_serial_number;
    VECTOR *call;
    void * closure;
    pptp_conn_cb callback;
    /******* IO buffers ******/
    char * read_buffer, *write_buffer;
    size_t read_alloc,   write_alloc;
    size_t read_size,    write_size;
};


struct PPTP_CALL {
    /* Call properties */
    enum {
	PPTP_CALL_PAC, PPTP_CALL_PNS
    } call_type;
    union {
	enum pptp_pac_state {
	    PAC_IDLE, PAC_WAIT_REPLY, PAC_ESTABLISHED, PAC_WAIT_CS_ANS
	} pac;
	enum pptp_pns_state {
	    PNS_IDLE, PNS_WAIT_REPLY, PNS_ESTABLISHED, PNS_WAIT_DISCONNECT
	} pns;
    } state;
    u_int16_t call_id, peer_call_id;
    u_int16_t sernum;
    u_int32_t speed;
    /* For user data: */
    pptp_call_cb callback;
    void * closure;
};


struct thread_specific {
    struct sigaction old_sigaction; /* evil signals */
    PPTP_CONN * conn;
};


/* pptp_gre.h -- encapsulate PPP in PPTP-GRE.
 *               Handle the IP Protocol 47 portion of PPTP.
 */

typedef struct pack_track {
  uint32_t seq;       // seq no of this tracked packet
  uint64_t time;      // time when this tracked packet was sent (in usecs)
} pack_track_t;

typedef struct gre_stats {
  /* statistics for GRE receive */

  uint32_t rx_accepted;  // data packet was passed to pppd
  uint32_t rx_lost;      // data packet did not arrive before timeout
  uint32_t rx_underwin;  // data packet was under window (arrived too late
			 // or duplicate packet)
  uint32_t rx_overwin;   // data packet was over window
			 // (too many packets lost?)
  uint32_t rx_buffered;  // data packet arrived earlier than expected,
			 // packet(s) before it were lost or reordered
  uint32_t rx_errors;    // OS error on receive
  uint32_t rx_truncated; // truncated packet
  uint32_t rx_invalid;   // wrong protocol or invalid flags
  uint32_t rx_acks;      // acknowledgement only

  /* statistics for GRE transmit */

  uint32_t tx_sent;      // data packet write() to GRE socket succeeded
  uint32_t tx_failed;    // data packet write() to GRE socket returned error
  uint32_t tx_short;     // data packet write() to GRE socket underflowed
  uint32_t tx_acks;      // sent packet with just ACK
  uint32_t tx_oversize;  // data packet dropped because it was too large

  /* statistics for packet tracking, for RTT calculation */

  pack_track_t pt;       // last data packet seq/time
  int rtt;               // estimated round-trip time in us

} gre_stats_t;


#define PPTP_BUF_MAX 65536
#define PPTP_TIMEOUT 60 /* seconds */
#define PPTP_CONNECT_SPEED 10000000
#define PPTP_WINDOW 3
#define PPTP_DELAY  0
#define PPTP_BPS_MIN 2400
#define PPTP_BPS_MAX 10000000


#ifdef SUPPORT_BEZEQ_ISRAEL

/* pptp_quirks.h ...... various options to fix quirks found in buggy adsl modems
 *                      mulix <mulix@actcom.co.il>
 */


/* isp defs - correspond to slots in the fixups table */
#define BEZEQ_ISRAEL "BEZEQ_ISRAEL"

/* vendor defs */

#define ORCKIT 1
#define ALCATEL 2

/* device defs */

#define ORCKIT_ATUR2 1
#define ORCKIT_ATUR3 2

struct pptp_fixup {
    const char* isp;    /* which isp? e.g. Bezeq in Israel */
    int vendor; /* which vendor? e.g. Orckit */
    int device; /* which device? e.g. Orckit Atur3 */

    /* use this hook to build your own out call request packet */
    int (*out_call_rqst_hook)(struct pptp_out_call_rqst* packet);

    /* use this hook to build your own start control connection packet */
    /* note that this hook is called from two different places, depending
       on whether this is a request or reply */
    int (*start_ctrl_conn)(struct pptp_start_ctrl_conn* packet);

    /* use this hook if you need to send a 'set_link' packet once
       the connection is established */
    int (*set_link_hook)(struct pptp_set_link_info* packet,
			 int peer_call_id);
};

static int quirk_index = -1;

static struct pptp_fixup pptp_fixups[] = {
    {BEZEQ_ISRAEL, ORCKIT, ORCKIT_ATUR3,
     orckit_atur3_build_hook,
     orckit_atur3_start_ctrl_conn_hook,
     orckit_atur3_set_link_hook}
};

/* find the index for this isp in the quirks table */
/* return the index on success, -1 if not found */
static int find_quirk(const char* isp_name);

/* set the global quirk index. return 0 on success, non 0 otherwise */
static int set_quirk_index(int index);

/* get the global quirk index. return the index on success,
   -1 if no quirk is defined */

static inline int get_quirk_index(void)
{
    return quirk_index;
}


#endif /* SUPPORT_BEZEQ_ISRAEL */




/* assume packet is bad/spoofed if it's more than this many seqs ahead */
#define MISSING_WINDOW 300

/* Packet queue structure: linked list of packets received out-of-order */
typedef struct pqueue {
  struct pqueue *next;
  struct pqueue *prev;
  int seq;
  struct timeval expires;
  unsigned char *packet;
  int packlen;
  int capacity;
} pqueue_t;

static int       pqueue_add  (int seq, unsigned char *packet, int packlen);
static int       pqueue_del  (pqueue_t *pq_point);
static int       pqueue_expiry_time (pqueue_t *entry);


/* util.h ....... error message utilities. */

#ifdef DEBUG
static void pptp_log(const char *func, int line, char *format, ...)
     __attribute__ ((format (printf, 3, 4)));
static void pptp_warn(const char *func, int line, char *format, ...)
     __attribute__ ((format (printf, 3, 4)));
static void pptp_fatal(const char *func, int line, char *format, ...)
     __attribute__ ((format (printf, 3, 4))) __attribute__ ((noreturn));

#define logF(format, args...) \
	pptp_log(__FUNCTION__,__LINE__, format , ## args)
#define warn(format, args...) \
	pptp_warn(__FUNCTION__,__LINE__, format , ## args)
#define fatal(format, args...) \
	pptp_fatal(__FUNCTION__,__LINE__, format , ## args)
#else
static void logF(const char *format, ...)
     __attribute__ ((format (printf, 1, 2)));
static void warn(const char *format, ...)
     __attribute__ ((format (printf, 1, 2)));
static void fatal(const char *format, ...)
     __attribute__ ((format (printf, 1, 2))) __attribute__ ((noreturn));

#endif

static const char version[] = "pptp-linux version " PPTP_LINUX_VERSION;

#define PACKET_MAX 8196

CALC_BSS struct BB_COMMON_BSS
{
	u_int32_t ack_sent, ack_recv;
	u_int32_t seq_sent, seq_recv;
	u_int16_t pptp_gre_call_id, pptp_gre_peer_call_id;
	sigjmp_buf callmgr_env;
	unsigned int decaps_len, escape;
	unsigned char copy[PACKET_MAX];
	int checkedsync;
	sigjmp_buf env;
	struct thread_specific global;
	/* contains a list of free queue elements. */
	pqueue_t *pq_freelist_head;
	pqueue_t *pq_head, *pq_tail;
	volatile int signaled;
	int syncppp;
	gre_stats_t stats;
	int decaps_gre_no_first;
	char *Gargv0;
	/* control the number of times echo packets will be logged
	   default 10 */
	int nlogecho;
	char ttydev[PATH_MAX];

	/* wait this many seconds for missing packets before forgetting about them */
#define DEFAULT_PACKET_TIMEOUT 0.3
	int packet_timeout_usecs /* = DEFAULT_PACKET_TIMEOUT * 1000000 */;
#ifdef CONFIG_FEATURE_VERBOSITY_PPTP
	/* log_level sets the logging verbosity. Values range from 0 (errors only)
	   to 1 (errors and warnings) to 2 (high verbosity, for debugging) */
	int log_level;
#define log_level bb_common_bss.log_level
#else
#define log_level 0
#endif
	/* log_string is an identifier for this pptp process, passed from
	   command line using --log-string=X, and included with every log message.
	   Useful for people with multiple pptp sessions open at a time */
	const char *log_string;
	u_int32_t encaps_gre_seq; /* first sequence number sent must be 1 */
	struct in_addr localbind;
} bb_common_bss;
#define ack_sent bb_common_bss.ack_sent
#define ack_recv bb_common_bss.ack_recv
#define seq_sent bb_common_bss.seq_sent
#define seq_recv bb_common_bss.seq_recv
#define pptp_gre_call_id bb_common_bss.pptp_gre_call_id
#define pptp_gre_peer_call_id bb_common_bss.pptp_gre_peer_call_id
#define callmgr_env bb_common_bss.callmgr_env
#define decaps_len bb_common_bss.decaps_len
#define escape bb_common_bss.escape
#define copy bb_common_bss.copy
#define checkedsync bb_common_bss.checkedsync
#define env bb_common_bss.env
#define global bb_common_bss.global
#define pq_freelist_head bb_common_bss.pq_freelist_head
#define pq_head bb_common_bss.pq_head
#define pq_tail bb_common_bss.pq_tail
#define signaled bb_common_bss.signaled
#define syncppp bb_common_bss.syncppp
#define stats bb_common_bss.stats
#define decaps_gre_no_first bb_common_bss.decaps_gre_no_first
#define Gargv0 bb_common_bss.Gargv0
#define nlogecho bb_common_bss.nlogecho
#define ttydev bb_common_bss.ttydev
#define packet_timeout_usecs bb_common_bss.packet_timeout_usecs
#define log_string bb_common_bss.log_string
#define encaps_gre_seq bb_common_bss.encaps_gre_seq
#define localbind bb_common_bss.localbind

static inline pqueue_t *pqueue_head (void) {
  return pq_head;
}

/* dirutil.c ... directory utilities. */


/* In-place modify a string to remove trailing slashes.  Returns arg.
 * stripslash("/") returns "/";
 */
static char *stripslash(char *pathname) {
    int len = strlen(pathname);

    while (len > 1 && pathname[len - 1] == '/')
	pathname[--len] = '\0';
    return pathname;
}

/* Return malloc'ed string representing directory name (no trailing slash) */
static char *pptp_dirname(const char *pathname)
{
    char *dup_pn = strdup(pathname);
    char *ptr = strrchr(stripslash(dup_pn), '/');

    if (ptr == NULL) { free(dup_pn); return strdup("."); }
    if (ptr == dup_pn && dup_pn[0] == '/') ptr++;
    *ptr = '\0';
    return dup_pn;
}


/* ensure dirname exists, creating it if necessary. */
static int make_valid_path(const char *dir, mode_t mode)
{
    char *tmp = NULL, *path = stripslash(strdup(dir));
    int retval = is_directory(path, 1, NULL);

    if (retval >= 0) {
	/* file exists */
	/* retval is flag of directory or not a directory (Oops.) */
	goto end;
    }
    retval = 0;
    /* Directory doesn't exist.  Let's make it. */
    /*   Make parent first. */
    if (!make_valid_path(tmp = pptp_dirname(path), mode)) goto end;
    /*   Now make this 'un. */
    if (mkdir(path, mode) < 0) goto end;
    /* Success. */
    retval = 1;

end:
    free(tmp);
    free(path);
    return retval;
}


#ifdef SUPPORT_BEZEQ_ISRAEL
/* orckit_quirks.c ...... fix quirks in orckit adsl modems
 *                        mulix <mulix@actcom.co.il>
 */

/* return 0 on success, non zero otherwise */
static int
orckit_atur3_build_hook(struct pptp_out_call_rqst* packet)
{
    unsigned int name_length = 10;

    struct pptp_out_call_rqst fixed_packet = {
	PPTP_HEADER_CTRL(PPTP_OUT_CALL_RQST),
	0, /* hton16(call->callid) */
	0, /* hton16(call->sernum) */
	hton32(PPTP_BPS_MIN), hton32(PPTP_BPS_MAX),
	hton32(PPTP_BEARER_DIGITAL), hton32(PPTP_FRAME_ANY),
	hton16(PPTP_WINDOW), 0, hton16(name_length), 0,
	{'R','E','L','A','Y','_','P','P','P','1',0}, {0}
    };

    if (!packet)
	return -1;

    memcpy(packet, &fixed_packet, sizeof(*packet));

    return 0;
}

/* return 0 on success, non zero otherwise */
static int
orckit_atur3_set_link_hook(struct pptp_set_link_info* packet,
			   int peer_call_id)
{
    struct pptp_set_link_info fixed_packet = {
	PPTP_HEADER_CTRL(PPTP_SET_LINK_INFO),
	hton16(peer_call_id),
	0,
	0xffffffff,
	0xffffffff};

    if (!packet)
	return -1;

    memcpy(packet, &fixed_packet, sizeof(*packet));
    return 0;
}

/* return 0 on success, non 0 otherwise */
static int
orckit_atur3_start_ctrl_conn_hook(struct pptp_start_ctrl_conn* packet)
{
    struct pptp_start_ctrl_conn fixed_packet = {
	{0}, /* we'll set the header later */
	hton16(PPTP_VERSION), 0, 0,
	hton32(PPTP_FRAME_ASYNC), hton32(PPTP_BEARER_ANALOG),
	hton16(0) /* max channels */,
	hton16(0x6021),
	{'R','E','L','A','Y','_','P','P','P','1',0}, /* hostname */
	{'M','S',' ','W','i','n',' ','N','T',0} /* vendor */
    };

    if (!packet)
	return -1;

    /* grab the header from the original packet, since we dont
       know if this is a request or a reply */
    memcpy(&fixed_packet.header, &packet->header, sizeof(struct pptp_header));

    /* and now overwrite the full packet, effectively preserving the header */
    memcpy(packet, &fixed_packet, sizeof(*packet));
    return 0;
}
#endif /* SUPPORT_BEZEQ_ISRAEL */

/* Fast Frame Check Sequence (FCS) Implementation, for HDLC-like framing of
 * PPP.
 * from RFC1662:
 *
 * C.2.  16-bit FCS Computation Method
 *
 *  The following code provides a table lookup computation for
 *  calculating the Frame Check Sequence as data arrives at the
 *  interface.  This implementation is based on [7], [8], and [9].
 *
 *  [7]   Perez, "Byte-wise CRC Calculations", IEEE Micro, June 1983.
 *
 *  [8]   Morse, G., "Calculating CRC's by Bits and Bytes", Byte,
 *        September 1986.
 *
 *  [9]   LeVan, J., "A Fast CRC", Byte, November 1987.
 */


/*
 * u16 represents an unsigned 16-bit number.  Adjust the typedef for
 * your hardware.
 */
typedef u_int16_t u16;

/*
 * FCS lookup table as calculated by the table generator.
 */
static const u16 fcstab[256] = {
   0x0000, 0x1189, 0x2312, 0x329b, 0x4624, 0x57ad, 0x6536, 0x74bf,
   0x8c48, 0x9dc1, 0xaf5a, 0xbed3, 0xca6c, 0xdbe5, 0xe97e, 0xf8f7,
   0x1081, 0x0108, 0x3393, 0x221a, 0x56a5, 0x472c, 0x75b7, 0x643e,
   0x9cc9, 0x8d40, 0xbfdb, 0xae52, 0xdaed, 0xcb64, 0xf9ff, 0xe876,
   0x2102, 0x308b, 0x0210, 0x1399, 0x6726, 0x76af, 0x4434, 0x55bd,
   0xad4a, 0xbcc3, 0x8e58, 0x9fd1, 0xeb6e, 0xfae7, 0xc87c, 0xd9f5,
   0x3183, 0x200a, 0x1291, 0x0318, 0x77a7, 0x662e, 0x54b5, 0x453c,
   0xbdcb, 0xac42, 0x9ed9, 0x8f50, 0xfbef, 0xea66, 0xd8fd, 0xc974,
   0x4204, 0x538d, 0x6116, 0x709f, 0x0420, 0x15a9, 0x2732, 0x36bb,
   0xce4c, 0xdfc5, 0xed5e, 0xfcd7, 0x8868, 0x99e1, 0xab7a, 0xbaf3,
   0x5285, 0x430c, 0x7197, 0x601e, 0x14a1, 0x0528, 0x37b3, 0x263a,
   0xdecd, 0xcf44, 0xfddf, 0xec56, 0x98e9, 0x8960, 0xbbfb, 0xaa72,
   0x6306, 0x728f, 0x4014, 0x519d, 0x2522, 0x34ab, 0x0630, 0x17b9,
   0xef4e, 0xfec7, 0xcc5c, 0xddd5, 0xa96a, 0xb8e3, 0x8a78, 0x9bf1,
   0x7387, 0x620e, 0x5095, 0x411c, 0x35a3, 0x242a, 0x16b1, 0x0738,
   0xffcf, 0xee46, 0xdcdd, 0xcd54, 0xb9eb, 0xa862, 0x9af9, 0x8b70,
   0x8408, 0x9581, 0xa71a, 0xb693, 0xc22c, 0xd3a5, 0xe13e, 0xf0b7,
   0x0840, 0x19c9, 0x2b52, 0x3adb, 0x4e64, 0x5fed, 0x6d76, 0x7cff,
   0x9489, 0x8500, 0xb79b, 0xa612, 0xd2ad, 0xc324, 0xf1bf, 0xe036,
   0x18c1, 0x0948, 0x3bd3, 0x2a5a, 0x5ee5, 0x4f6c, 0x7df7, 0x6c7e,
   0xa50a, 0xb483, 0x8618, 0x9791, 0xe32e, 0xf2a7, 0xc03c, 0xd1b5,
   0x2942, 0x38cb, 0x0a50, 0x1bd9, 0x6f66, 0x7eef, 0x4c74, 0x5dfd,
   0xb58b, 0xa402, 0x9699, 0x8710, 0xf3af, 0xe226, 0xd0bd, 0xc134,
   0x39c3, 0x284a, 0x1ad1, 0x0b58, 0x7fe7, 0x6e6e, 0x5cf5, 0x4d7c,
   0xc60c, 0xd785, 0xe51e, 0xf497, 0x8028, 0x91a1, 0xa33a, 0xb2b3,
   0x4a44, 0x5bcd, 0x6956, 0x78df, 0x0c60, 0x1de9, 0x2f72, 0x3efb,
   0xd68d, 0xc704, 0xf59f, 0xe416, 0x90a9, 0x8120, 0xb3bb, 0xa232,
   0x5ac5, 0x4b4c, 0x79d7, 0x685e, 0x1ce1, 0x0d68, 0x3ff3, 0x2e7a,
   0xe70e, 0xf687, 0xc41c, 0xd595, 0xa12a, 0xb0a3, 0x8238, 0x93b1,
   0x6b46, 0x7acf, 0x4854, 0x59dd, 0x2d62, 0x3ceb, 0x0e70, 0x1ff9,
   0xf78f, 0xe606, 0xd49d, 0xc514, 0xb1ab, 0xa022, 0x92b9, 0x8330,
   0x7bc7, 0x6a4e, 0x58d5, 0x495c, 0x3de3, 0x2c6a, 0x1ef1, 0x0f78
};

/*
 * Calculate a new fcs given the current fcs and the new data.
 */
static u16 pppfcs16(u16 fcs, void *_cp, int len)
{
    unsigned char *cp = (unsigned char *)_cp;
    /* don't worry about the efficiency of these asserts here.  gcc will
     * recognise that the asserted expressions are constant and remove them.
     * Whether they are usefull is another question.
     */
#define ASSERT(x) assert(x)

    ASSERT(sizeof (u16) == 2);
    ASSERT(((u16) -1) > 0);
#undef ASSERT

    while (len--)
	fcs = (fcs >> 8) ^ fcstab[(fcs ^ *cp++) & 0xff];
    return (fcs);
}


/* pptp.c ... client shell to launch call managers, data handlers, and
 *            the pppd from the command line.
 */



static struct in_addr get_ip_address(const char *name);
static int open_callmgr(struct in_addr inetaddr, char *phonenr, int pty_fd);
static void launch_callmgr(struct in_addr inetaddr, char *phonenr)
					__attribute__ ((noreturn));
static int get_call_id(int sock, pid_t gre, pid_t pppd,
			u_int16_t *call_id, u_int16_t *peer_call_id);
static void launch_pppd(int argc, char **argv) __attribute__ ((noreturn));

/*** print usage and exit *****************************************************/
static void usage(void) __attribute__ ((noreturn));
static void usage(void)
{
    fprintf(stderr,
	    "%s\n"
	    "Usage:\n"
	    "  %s <hostname> [<pptp options>] [[--] <pppd options>]\n"
	    "\n"
	    "Or using pppd's pty option: \n"
	    "  pppd pty \"%s <hostname> --nolaunchpppd <pptp options>\"\n"
	    "\n"
	    "Available pptp options:\n"
	    "  --phone <number> Pass <number> to remote host as phone number\n"
	    "  --nolaunchpppd   Do not launch pppd, for use as a pppd pty\n"
#ifdef SUPPORT_BEZEQ_ISRAEL
	    "  --quirks <quirk> Work around a buggy PPTP implementation\n"
	    "                   Currently recognised values are BEZEQ_ISRAEL only\n"
#endif
	    "  --debug          Run in foreground (for debugging with gdb)\n"
	    "  --sync           Enable Synchronous HDLC (pppd must use it too)\n"
	    "  --timeout <secs> Time to wait for reordered packets (0.01 to 10 secs)\n"
	    "  --logstring <name>       Use <name> instead of 'anon' in syslog messages\n"
	    "  --localbind <addr>       Bind to specified IP address instead of wildcard\n"
	    "  --loglevel <level>       Sets the debugging level (0=low, 1=default, 2=high)\n",

	    version, bb_applet_name, bb_applet_name);
    exit(1);
}


/*** do nothing signal handler ************************************************/
static void do_nothing(int sig)
{
    /* do nothing signal handler. Better than SIG_IGN. */
    signaled = 1;
}


/*** signal handler ***********************************************************/
static void sighandler(int sig)
{
    siglongjmp(env, 1);
}

/*** report statistics signal handler (SIGUSR1) *******************************/
static void sigstats(int sig)
{
    syslog(LOG_NOTICE, "GRE statistics:\n");
#define LOG(name,value) syslog(LOG_NOTICE, name "\n", stats .value)
    LOG("rx accepted  = %d", rx_accepted);
    LOG("rx lost      = %d", rx_lost);
    LOG("rx under win = %d", rx_underwin);
    LOG("rx over  win = %d", rx_overwin);
    LOG("rx buffered  = %d", rx_buffered);
    LOG("rx OS errors = %d", rx_errors);
    LOG("rx truncated = %d", rx_truncated);
    LOG("rx invalid   = %d", rx_invalid);
    LOG("rx acks      = %d", rx_acks);
    LOG("tx sent      = %d", tx_sent);
    LOG("tx failed    = %d", tx_failed);
    LOG("tx short     = %d", tx_short);
    LOG("tx acks      = %d", tx_acks);
    LOG("tx oversize  = %d", tx_oversize);
    LOG("round trip   = %d usecs", rtt);
#undef LOG
}

/*** get the ipaddress coming from the command line ***************************/
static struct in_addr get_ip_address(const char *name)
{
    struct in_addr retval;

    if(bb_gethostbyname(BB_GHBN_IN_ADDR | BB_GHBN_SILENT, name, &retval) == NULL)
	fatal("Host has non-internet IPv4 address");
    return retval;
}

/*** make a unix socket address ***********************************************/
static void callmgr_name_unixsock(struct sockaddr_un *where,
					struct in_addr inetaddr,
					struct in_addr local_bind)
{
    char localaddr[16];
    where->sun_family = AF_UNIX;
    strncpy(localaddr,  inet_ntoa(local_bind), 16);
    snprintf(where->sun_path, sizeof(where->sun_path),
	    PPTP_SOCKET_PREFIX "%s:%s", localaddr, inet_ntoa(inetaddr));
}


/*** start the call manager ***************************************************/
static int open_callmgr(struct in_addr inetaddr, char *phonenr, int pty_fd)
{
    /* Try to open unix domain socket to call manager. */
    struct sockaddr_un where;
    const int NUM_TRIES = 3;
    int i, fd;
    pid_t pid;
    int status;
    /* Open socket */
    if ((fd = bb_xsocket(BB_XSOCKET_FLG_NO_WARN)) < 0) {
	fatal("Could not create unix domain socket: %m");
    }
    /* Make address */
    callmgr_name_unixsock(&where, inetaddr, localbind);
    for (i = 0; i < NUM_TRIES; i++) {
	if (connect(fd, (struct sockaddr *) &where, sizeof(where)) < 0) {
	    /* couldn't connect.  We'll have to launch this guy. */

	    unlink (where.sun_path);

	    /* fork and launch call manager process */
	    switch (pid = fork()) {
		case -1: /* failure */
		    fatal("fork() to launch call manager failed.");
		case 0: /* child */
		{
		    close (fd);
		    /* close the pty in the call manager */
		    close(pty_fd);
		    launch_callmgr(inetaddr, phonenr);
		}
		default: /* parent */
		    waitpid_no_int(pid, &status, 0, 0);
		    if(status != 0)
			fatal("Call manager exited with error %d", status);
		    break;
	    }
	    sleep(1);
	}
	else return fd;
    }
    close(fd);
    fatal("Could not launch call manager after %d tries.", i);
    return -1;   /* make gcc happy */
}


/*** exchange data with the call manager  *************************************/
/* XXX need better error checking XXX */
static int get_call_id(int sock, pid_t gre, pid_t pppd,
		 u_int16_t *call_id, u_int16_t *peer_call_id)
{
    u_int16_t m_call_id, m_peer_call_id;
    /* write pid's to socket */
    /* don't bother with network byte order, because pid's are meaningless
     * outside the local host.
     */
    int rc;
    rc = write(sock, &gre, sizeof(gre));
    if (rc != sizeof(gre))
	return -1;
    rc = write(sock, &pppd, sizeof(pppd));
    if (rc != sizeof(pppd))
	return -1;
    rc = read(sock,  &m_call_id, sizeof(m_call_id));
    if (rc != sizeof(m_call_id))
	return -1;
    rc = read(sock,  &m_peer_call_id, sizeof(m_peer_call_id));
    if (rc != sizeof(m_peer_call_id))
	return -1;
    /*
     * XXX FIXME ... DO ERROR CHECKING & TIME-OUTS XXX
     * (Rhialto: I am assuming for now that timeouts are not relevant
     * here, because the read and write calls would return -1 (fail) when
     * the peer goes away during the process. We know it is (or was)
     * running because the connect() call succeeded.)
     * (James: on the other hand, if the route to the peer goes away, we
     * wouldn't get told by read() or write() for quite some time.)
     */
    *call_id = m_call_id;
    *peer_call_id = m_peer_call_id;
    return 0;
}

/*** execvp pppd **************************************************************/
static void launch_pppd(int argc, char **argv)
{
    char *new_argv[argc + 4];  /* XXX if not using GCC, hard code a limit here. */
    int i = 0, j;

    new_argv[i++] = PPPD_BINARY;
#ifdef USER_PPP
    new_argv[i++] = "-direct";
    /* ppp expects to have stdin connected to ttydev */
    if ((j = open(ttydev, O_RDWR)) == -1)
	fatal("Cannot open %s: %m", ttydev);
    if (dup2(j, 0) == -1)
	fatal("dup2 failed: %m");
    close(j);
#else
    new_argv[i++] = ttydev;
    new_argv[i++] = "38400";
#endif
    for (j = 0; j < argc; j++)
	new_argv[i++] = argv[j];
    new_argv[i] = NULL;
    execvp(new_argv[0], new_argv);
    bb_perror_msg("Error");
    fatal("Could not launch pppd");
}

/* pptp_callmgr.c ... Call manager for PPTP connections.
 *                    Handles TCP port 1723 protocol.
 */

static void callmgr_sighandler(int sig) {
    siglongjmp (callmgr_env, 1);
}

static void callmgr_do_nothing(int sig) {
    /* do nothing signal handler */
}

struct local_callinfo {
    int unix_sock;
    pid_t pid[2];
};

struct local_conninfo {
    VECTOR * call_list;
    fd_set * call_set;
};

/* vector.c ..... store a vector of PPTP_CALL information and search it
 *                efficiently.
 */


/*** vector_item **************************************************************/
static struct vector_item *binary_search(VECTOR *v, int key)
{
    int l,r,x;
    l = 0;
    r = v->size - 1;
    while (r >= l) {
	x = (l + r)/2;
	if (key <  v->item[x].key) r = x - 1; else l = x + 1;
	if (key == v->item[x].key) return &(v->item[x]);
    }
    return NULL;
}

/*** vector_contains **********************************************************/
static int vector_contains(VECTOR *v, int key)
{
    assert(v != NULL);
    return (binary_search(v, key) != NULL);
}

/*** vector_create ************************************************************/
static VECTOR *vector_create(void)
{
    const int INITIAL_SIZE = 4;

    VECTOR *v = malloc(sizeof(*v));
    if (v == NULL) return v;

    v->size = 0;
    v->alloc = INITIAL_SIZE;
    v->item = malloc(sizeof(*(v->item)) * (v->alloc));
    if (v->item == NULL) { free(v); return NULL; }
    else return v;
}

/*** vector_destroy ***********************************************************/
static void vector_destroy(VECTOR *v)
{
    free(v->item);
    free(v);
}

/*** vector_size **************************************************************/
static int vector_size(VECTOR *v)
{
    assert(v != NULL);
    return v->size;
}

/*** vector_insert*************************************************************
 * nice thing about file descriptors is that we are assured by POSIX
 * that they are monotonically increasing.
 */
static int vector_insert(VECTOR *v, int key, PPTP_CALL * call)
{
    int i;

    assert(v != NULL && call != NULL);
    assert(!vector_contains(v, key));
    if (!(v->size < v->alloc)) {
	void *tmp = realloc(v->item, sizeof(*(v->item)) * 2 * v->alloc);
	if (tmp != NULL) {
	    v->alloc *= 2;
	    v->item = tmp;
	} else return 0; /* failed to alloc memory. */
    }
    assert(v->size < v->alloc);
    /* for safety, we make this work in the general case;
     * but this is optimized for adding call to the end of the vector.
     */
    for(i = v->size - 1; i >= 0; i--)
	if (v->item[i].key < key)
	    break;
    /* insert after item i */
    memmove(&v->item[i + 2], &v->item[i + 1],
	    (v->size - i - 1) * sizeof(*(v->item)));
    v->item[i + 1].key  = key;
    v->item[i + 1].call = call;
    v->size++;
    return 1;
}

/*** vector_remove ************************************************************/
static int  vector_remove(VECTOR *v, int key)
{
    struct vector_item *tmp;
    assert(v != NULL);
    if ((tmp =binary_search(v,key)) == NULL) return 0;
    assert(tmp >= v->item && tmp < v->item + v->size);
    memmove(tmp, tmp + 1, (v->size - (v->item - tmp) - 1) * sizeof(*(v->item)));
    v->size--;
    return 1;
}

/*** vector_search ************************************************************/
static int vector_search(VECTOR *v, int key, PPTP_CALL **call)
{
    struct vector_item *tmp;
    assert(v != NULL);
    tmp = binary_search(v, key);
    if (tmp ==NULL) return 0;
    *call = tmp->call;
    return 1;
}

/*** vector_scan ***************************************************************
 * Hmm.  Let's be fancy and use a binary search for the first
 * unused key, taking advantage of the list is stored sorted; ie
 * we can look at pointers and keys at two different locations,
 * and if (ptr1 - ptr2) = (key1 - key2) then all the slots
 * between ptr1 and ptr2 are filled.  Note that ptr1-ptr2 should
 * never be greater than key1-key2 (no duplicate keys!)... we
 * check for this.
 */
/* find first unused key. Returns 1 on success, 0 if no. */
static int vector_scan(VECTOR *v, int lo, int hi, int *key)
{
    int l,r,x;
    assert(v != NULL);
    assert(key != NULL);
    if ((v->size<1) || (lo < v->item[0].key)) { *key = lo; return 1; }
    /* our array bounds */
    l = 0;  r = v->size - 1;
    while (r > l) {
	/* check for a free spot right after l */
	if (v->item[l].key + 1 < v->item[l + 1].key) { /* found it! */
	    *key = v->item[l].key + 1;
	    return 1;
	}
	/* no dice. Let's see if the free spot is before or after the midpoint */
	x = (l + r)/2;
	/* Okay, we have right (r), left (l) and the probe (x). */
	assert(x - l <= v->item[x].key - v->item[l].key);
	assert(r - x <= v->item[r].key - v->item[x].key);
	if (x - l < v->item[x].key - v->item[l].key)
	    /* room between l and x */
	    r = x;
	else /* no room between l and x */
	    if (r - x < v->item[r].key - v->item[x].key)
		/* room between x and r */
		l = x;
	    else /* no room between x and r, either */
		break; /* game over, man. */
    }
    /* no room found in already allocated space.  Check to see if
     * there's free space above allocated entries. */
    if (v->item[v->size - 1].key < hi) {
	*key = v->item[v->size - 1].key + 1;
	return 1;
    }
    /* outta luck */
    return 0;
}

/*** Get info from call structure *********************************************/
/* NOTE: The peer_call_id is undefined until we get a server response. */
static void pptp_call_get_ids(PPTP_CONN * conn, PPTP_CALL * call,
		       u_int16_t * call_id, u_int16_t * peer_call_id)
{
    assert(conn != NULL); assert(call != NULL);
    *call_id = call->call_id;
    *peer_call_id = call->peer_call_id;
}

/* Arbitrary user data about this call/connection.
 * It is the caller's responsibility to free this data before calling
 * pptp_call|conn_close()
 */
/*** pptp_call_closure_put ****************************************************/
static void
pptp_call_closure_put(PPTP_CONN * conn, PPTP_CALL * call, void *cl)
{
    assert(conn != NULL); assert(call != NULL);
    call->closure = cl;
}

/*** pptp_call_closure_get ****************************************************/
static void * pptp_call_closure_get(PPTP_CONN * conn, PPTP_CALL * call)
{
    assert(conn != NULL); assert(call != NULL);
    return call->closure;
}

/*** pptp_conn_closure_put ****************************************************/
static void pptp_conn_closure_put(PPTP_CONN * conn, void *cl)
{
    assert(conn != NULL);
    conn->closure = cl;
}

/*** pptp_conn_closure_get ****************************************************/
static void * pptp_conn_closure_get(PPTP_CONN * conn)
{
    assert(conn != NULL);
    return conn->closure;
}


/* Call callback */
static void
call_callback(PPTP_CONN *conn, PPTP_CALL *call, enum call_state state)
{
    struct local_callinfo *lci;
    struct local_conninfo *conninfo;
    u_int16_t call_id[2];
    switch(state) {
	case CALL_OPEN_DONE:
	    /* okey dokey.  This means that the call_id and peer_call_id are
	     * now valid, so lets send them on to our friends who requested
	     * this call.  */
	    lci = pptp_call_closure_get(conn, call); assert(lci != NULL);
	    pptp_call_get_ids(conn, call, &call_id[0], &call_id[1]);
	    write(lci->unix_sock, &call_id, sizeof(call_id));
	    /* Our duty to the fatherland is now complete. */
	    break;
	case CALL_OPEN_FAIL:
	case CALL_CLOSE_RQST:
	case CALL_CLOSE_DONE:
	    /* don't need to do anything here, except make sure tables
	     * are sync'ed */
	    logF("Closing connection");
	    conninfo = pptp_conn_closure_get(conn);
	    lci = pptp_call_closure_get(conn, call);
	    assert(lci != NULL && conninfo != NULL);
	    if (vector_contains(conninfo->call_list, lci->unix_sock)) {
		vector_remove(conninfo->call_list, lci->unix_sock);
		close(lci->unix_sock);
		FD_CLR(lci->unix_sock, conninfo->call_set);
		if(lci->pid[0]) kill(lci->pid[0], SIGTERM);
		if(lci->pid[1]) kill(lci->pid[1], SIGTERM);
	    }
	    break;
	default:
	    logF("Unhandled call callback state [%d].", (int) state);
	    break;
    }
}

/*** close_inetsock ***********************************************************/
static inline void close_inetsock(int fd, struct in_addr inetaddr)
{
    close(fd);
}

static inline int pptp_conn_established(PPTP_CONN *conn) {
  return (conn->conn_state == CONN_ESTABLISHED);
}

/* get a specific PPTP_CALL ... useful only when iterating. */
static PPTP_CALL * vector_get_Nth(VECTOR *v, int n)
{
    assert(v != NULL);
    assert(0 <= n && n < vector_size(v));
    return v->item[n].call;
}


/*** open_inetsock ************************************************************/
static int open_inetsock(struct in_addr inetaddr)
{
    struct sockaddr_in dest, src;
    int s;

    dest.sin_family = AF_INET;
    dest.sin_port   = htons(PPTP_PORT);
    dest.sin_addr   = inetaddr;
    if ((s = bb_xsocket(BB_XSOCKET_FLG_NO_WARN)) < 0) {
	warn("socket: %m");
	return s;
    }
    if (localbind.s_addr != INADDR_NONE) {
	memset(&src, 0, sizeof(src));
	src.sin_family = AF_INET;
	src.sin_addr   = localbind;
	if (bind(s, (struct sockaddr *) &src, sizeof(src)) != 0) {
	    warn("bind: %m");
	    close(s); return -1;
	}
    }
    if (connect(s, (struct sockaddr *) &dest, sizeof(dest)) < 0) {
	warn("connect: %m");
	close(s); return -1;
    }
    return s;
}

/*** open_unixsock ************************************************************/
static int open_unixsock(struct in_addr inetaddr)
{
    struct sockaddr_un where;
    struct stat st;
    char *dir;
    int s;
    if ((s = bb_xsocket(BB_XSOCKET_FLG_NO_WARN)) < 0) {
	warn("socket: %m");
	return s;
    }
    callmgr_name_unixsock(&where, inetaddr, localbind);
    if (stat(where.sun_path, &st) >= 0) {
	warn("Call manager for %s is already running.", inet_ntoa(inetaddr));
	close(s); return -1;
    }
    /* Make sure path is valid. */
    dir = pptp_dirname(where.sun_path);
    if (!make_valid_path(dir, 0770))
	fatal("Could not make path to %s: %m", where.sun_path);
    free(dir);
    if (bind(s, (struct sockaddr *) &where, sizeof(where)) < 0) {
	warn("bind: %m");
	close(s); return -1;
    }
    chmod(where.sun_path, 0777);
    listen(s, 127);
    return s;
}


/*** close_unixsock ***********************************************************/
static void close_unixsock(int fd, struct in_addr inetaddr)
{
    struct sockaddr_un where;

    close(fd);
    callmgr_name_unixsock(&where, inetaddr, localbind);
    unlink(where.sun_path);
}

/*** Deal with messages, in a non-blocking manner
 * Add file descriptors used by pptp to fd_set.
 */
static void
pptp_fd_set(PPTP_CONN * conn, fd_set * read_set, fd_set * write_set, int * max_fd)
{
    assert(conn && conn->call);
    /* Add fd to write_set if there are outstanding writes. */
    if (conn->write_size > 0)
	FD_SET(conn->inet_sock, write_set);
    /* Always add fd to read_set. (always want something to read) */
    FD_SET(conn->inet_sock, read_set);
    if (*max_fd < conn->inet_sock) *max_fd = conn->inet_sock;
}

static const char * const ctrl_msg_types[] = {
	 "invalid control message type",
/*         (Control Connection Management) */
	 "Start-Control-Connection-Request",            /* 1 */
	 "Start-Control-Connection-Reply",              /* 2 */
	 "Stop-Control-Connection-Request",             /* 3 */
	 "Stop-Control-Connection-Reply",               /* 4 */
	 "Echo-Request",                                /* 5 */
	 "Echo-Reply",                                  /* 6 */
/*         (Call Management) */
	 "Outgoing-Call-Request",                       /* 7 */
	 "Outgoing-Call-Reply",                         /* 8 */
	 "Incoming-Call-Request",                       /* 9 */
	 "Incoming-Call-Reply",                        /* 10 */
	 "Incoming-Call-Connected",                    /* 11 */
	 "Call-Clear-Request",                         /* 12 */
	 "Call-Disconnect-Notify",                     /* 13 */
/*         (Error Reporting) */
	 "WAN-Error-Notify",                           /* 14 */
/*         (PPP Session Control) */
	 "Set-Link-Info"                              /* 15 */
};
#define MAX_CTRLMSG_TYPE 15

/*** report a sent packet ****************************************************/
static void ctrlp_rep( void * buffer, int size, int isbuff)
{
    struct pptp_header *packet = buffer;
    unsigned int type;
    if(size < sizeof(struct pptp_header)) return;
    type = ntoh16(packet->ctrl_type);
    /* don't keep reporting sending of echo's */
    if( (type == PPTP_ECHO_RQST || type == PPTP_ECHO_RPLY) && nlogecho <= 0 ) return;
    logF("%s control packet type is %d '%s'\n",isbuff ? "Buffered" : "Sent",
	    type, ctrl_msg_types[type <= MAX_CTRLMSG_TYPE ? type : 0]);
}

/*** hard close ***************************************************************/
static void pptp_call_destroy(PPTP_CONN *conn, PPTP_CALL *call)
{
    assert(conn && conn->call); assert(call);
    assert(vector_contains(conn->call, call->call_id));
    /* notify */
    if (call->callback != NULL) call->callback(conn, call, CALL_CLOSE_DONE);
    /* deallocate */
    vector_remove(conn->call, call->call_id);
    free(call);
}

/*** this is a hard close *****************************************************/
static void pptp_conn_destroy(PPTP_CONN * conn)
{
    int i;
    assert(conn != NULL); assert(conn->call != NULL);
    /* destroy all open calls */
    for (i = 0; i < vector_size(conn->call); i++)
	pptp_call_destroy(conn, vector_get_Nth(conn->call, i));
    /* notify */
    if (conn->callback != NULL) conn->callback(conn, CONN_CLOSE_DONE);
    sigaction(SIGALRM, &global.old_sigaction, NULL);
    close(conn->inet_sock);
    /* deallocate */
    vector_destroy(conn->call);
    free(conn);
}

/*** Non-blocking write *******************************************************/
static void pptp_write_some(PPTP_CONN * conn)
{
    ssize_t retval;

    assert(conn && conn->call);
    retval = write(conn->inet_sock, conn->write_buffer, conn->write_size);
    if (retval < 0) { /* error. */
	if (errno == EAGAIN || errno == EINTR) {
	    /* ignore */;
	} else { /* a real error */
	    logF("write error: %m");
	    pptp_conn_destroy(conn); /* shut down fast. */
	}
	return;
    }
    assert(retval <= conn->write_size);
    conn->write_size -= retval;
    memmove(conn->write_buffer, conn->write_buffer + retval, conn->write_size);
    ctrlp_rep(conn->write_buffer, retval, 0);
}

/* Add packet to write_buffer */
static int
pptp_send_ctrl_packet(PPTP_CONN * conn, void * buffer, size_t size)
{
    assert(conn && conn->call); assert(buffer);
    if( conn->write_size > 0)
	pptp_write_some(conn);
    if( conn->write_size == 0) {
	ssize_t retval;
	retval = write(conn->inet_sock, buffer, size);
	if (retval < 0) { /* error. */
	    if (errno == EAGAIN || errno == EINTR) {
		/* ignore */;
		retval = 0;
	    } else { /* a real error */
		logF("write error: %m");
		pptp_conn_destroy(conn); /* shut down fast. */
		return 0;
	    }
	}
	ctrlp_rep( buffer, retval, 0);
	size -= retval;
	if( size <= 0) return 1;
    }
    /* Shove anything not written into the write buffer */
    if (conn->write_size + size > conn->write_alloc) { /* need more memory */
	char *new_buffer = realloc(conn->write_buffer,
		sizeof(*(conn->write_buffer)) * conn->write_alloc * 2);
	if (new_buffer == NULL) {
	    logF(bb_msg_memory_exhausted); return 0;
	}
	conn->write_alloc *= 2;
	conn->write_buffer = new_buffer;
    }
    memcpy(conn->write_buffer + conn->write_size, buffer, size);
    conn->write_size += size;
    ctrlp_rep( buffer,size,1);
    return 1;
}

/*** Reset keep-alive timer ***************************************************/
static void pptp_reset_timer(void)
{
    const struct itimerval tv = { {  0, 0 },   /* stop on time-out */
	{ PPTP_TIMEOUT, 0 } }; /* 60-second timer */
	setitimer(ITIMER_REAL, &tv, NULL);
}

/* soft close.  Will callback on completion. */
static void pptp_call_close(PPTP_CONN * conn, PPTP_CALL * call)
{
    struct pptp_call_clear_rqst rqst = {
	PPTP_HEADER_CTRL(PPTP_CALL_CLEAR_RQST), 0, 0
    };
    assert(conn && conn->call); assert(call);
    assert(vector_contains(conn->call, call->call_id));
    /* haven't thought about PAC yet */
    assert(call->call_type == PPTP_CALL_PNS);
    assert(call->state.pns != PNS_IDLE);
    rqst.call_id = hton16(call->call_id);
    /* don't check state against WAIT_DISCONNECT... allow multiple disconnect
     * requests to be made.
     */
    pptp_send_ctrl_packet(conn, &rqst, sizeof(rqst));
    pptp_reset_timer();
    call->state.pns = PNS_WAIT_DISCONNECT;
    /* call structure will be freed when we have confirmation of disconnect. */
}

/* soft close.  Will callback on completion. */
static void pptp_conn_close(PPTP_CONN * conn, u_int8_t close_reason)
{
    struct pptp_stop_ctrl_conn rqst = {
	PPTP_HEADER_CTRL(PPTP_STOP_CTRL_CONN_RQST),
	hton8(close_reason), 0, 0
    };
    int i;
    assert(conn && conn->call);
    /* avoid repeated close attempts */
    if (conn->conn_state == CONN_IDLE || conn->conn_state == CONN_WAIT_STOP_REPLY)
	return;
    /* close open calls, if any */
    for (i = 0; i < vector_size(conn->call); i++)
	pptp_call_close(conn, vector_get_Nth(conn->call, i));
    /* now close connection */
    logF("Closing PPTP connection");
    pptp_send_ctrl_packet(conn, &rqst, sizeof(rqst));
    pptp_reset_timer(); /* wait 60 seconds for reply */
    conn->conn_state = CONN_WAIT_STOP_REPLY;
    return;
}

/*** Handle keep-alive timer **************************************************/
static void pptp_handle_timer(int sig)
{
    int i;
    /* "Keep Alives and Timers, 1": check connection state */
    if (global.conn->conn_state != CONN_ESTABLISHED) {
	if (global.conn->conn_state == CONN_WAIT_STOP_REPLY)
	    /* hard close. */
	    pptp_conn_destroy(global.conn);
	else /* soft close */
	    pptp_conn_close(global.conn, PPTP_STOP_NONE);
    }
    /* "Keep Alives and Timers, 2": check echo status */
    if (global.conn->ka_state == KA_OUTSTANDING) /*no response to keep-alive*/
	pptp_conn_close(global.conn, PPTP_STOP_NONE);
    else { /* ka_state == NONE */ /* send keep-alive */
	struct pptp_echo_rqst rqst = {
	    PPTP_HEADER_CTRL(PPTP_ECHO_RQST), hton32(global.conn->ka_id) };
	pptp_send_ctrl_packet(global.conn, &rqst, sizeof(rqst));
	global.conn->ka_state = KA_OUTSTANDING;
	/* XXX FIXME: wake up ctrl thread -- or will the SIGALRM do that
	 * automagically? XXX
	 */
    }
    /* check incoming/outgoing call states for !IDLE && !ESTABLISHED */
    for (i = 0; i < vector_size(global.conn->call); i++) {
	PPTP_CALL * call = vector_get_Nth(global.conn->call, i);
	if (call->call_type == PPTP_CALL_PNS) {
	    if (call->state.pns == PNS_WAIT_REPLY) {
		/* send close request */
		pptp_call_close(global.conn, call);
		assert(call->state.pns == PNS_WAIT_DISCONNECT);
	    } else if (call->state.pns == PNS_WAIT_DISCONNECT) {
		/* hard-close the call */
		pptp_call_destroy(global.conn, call);
	    }
	} else if (call->call_type == PPTP_CALL_PAC) {
	    if (call->state.pac == PAC_WAIT_REPLY) {
		/* XXX FIXME -- drop the PAC connection XXX */
	    } else if (call->state.pac == PAC_WAIT_CS_ANS) {
		/* XXX FIXME -- drop the PAC connection XXX */
	    }
	}
    }
    pptp_reset_timer();
}

/* Open new pptp_connection.  Returns NULL on failure. */
/* if 'isclient' is true, then will send 'conn open' packet to other host.
 * not necessary if this is being opened by a server process after
 * receiving a conn_open packet from client.
 */
static PPTP_CONN *
pptp_conn_open(int inet_sock, int isclient, pptp_conn_cb callback)
{
    struct sigaction sigact;
    PPTP_CONN *conn;
    /* Allocate structure */
    if ((conn = malloc(sizeof(*conn))) == NULL) return NULL;
    if ((conn->call = vector_create()) == NULL) { free(conn); return NULL; }
    /* Initialize */
    conn->inet_sock = inet_sock;
    conn->conn_state = CONN_IDLE;
    conn->ka_state  = KA_NONE;
    conn->ka_id     = 1;
    conn->call_serial_number = 0;
    conn->callback  = callback;
    /* Create I/O buffers */
    conn->read_size = conn->write_size = 0;
    conn->read_alloc = conn->write_alloc = INITIAL_BUFSIZE;
    conn->read_buffer =
	malloc(sizeof(*(conn->read_buffer)) * conn->read_alloc);
    conn->write_buffer =
	malloc(sizeof(*(conn->write_buffer)) * conn->write_alloc);
    if (conn->read_buffer == NULL || conn->write_buffer == NULL) {
	if (conn->read_buffer  != NULL) free(conn->read_buffer);
	if (conn->write_buffer != NULL) free(conn->write_buffer);
	vector_destroy(conn->call); free(conn); return NULL;
    }
    /* Make this socket non-blocking. */
    fcntl(conn->inet_sock, F_SETFL, O_NONBLOCK);
    /* Request connection from server, if this is a client */
    if (isclient) {
	struct pptp_start_ctrl_conn packet = {
	    PPTP_HEADER_CTRL(PPTP_START_CTRL_CONN_RQST),
	    hton16(PPTP_VERSION), 0, 0,
	    hton32(PPTP_FRAME_CAP), hton32(PPTP_BEARER_CAP),
	    hton16(PPTP_MAX_CHANNELS), hton16(PPTP_FIRMWARE_VERSION),
	    PPTP_HOSTNAME, PPTP_VENDOR
	};

#ifdef SUPPORT_BEZEQ_ISRAEL
	/* fix this packet, if necessary */
	int idx, rc;
	idx = get_quirk_index();
	if (idx != -1 && pptp_fixups[idx].start_ctrl_conn) {
	    if ((rc = pptp_fixups[idx].start_ctrl_conn(&packet)))
		warn("calling the start_ctrl_conn hook failed (%d)", rc);
	}
#endif

	if (pptp_send_ctrl_packet(conn, &packet, sizeof(packet)))
	    conn->conn_state = CONN_WAIT_CTL_REPLY;
	else
	    return NULL; /* could not send initial start request. */
    }
    /* Set up interval/keep-alive timer */
    /*   First, register handler for SIGALRM */
    sigact.sa_handler = pptp_handle_timer;
    sigemptyset(&sigact.sa_mask);
    sigact.sa_flags = SA_RESTART;
    global.conn = conn;
    sigaction(SIGALRM, &sigact, &global.old_sigaction);
    /* Reset event timer */
    pptp_reset_timer();
    /* all done. */
    return conn;
}

/* This currently *only* works for client call requests.
 * We need to do something else to allocate calls for incoming requests.
 */
static PPTP_CALL * pptp_call_open(PPTP_CONN * conn, pptp_call_cb callback,
	char *phonenr)
{
    PPTP_CALL * call;
    int i;

    /* Send off the call request */
    struct pptp_out_call_rqst packet = {
	PPTP_HEADER_CTRL(PPTP_OUT_CALL_RQST),
	0,0, /*call_id, sernum */
	hton32(PPTP_BPS_MIN), hton32(PPTP_BPS_MAX),
	hton32(PPTP_BEARER_CAP), hton32(PPTP_FRAME_CAP),
	hton16(PPTP_WINDOW), 0, 0, 0, {0}, {0}
    };
    assert(conn && conn->call);
    assert(conn->conn_state == CONN_ESTABLISHED);
    /* Assign call id */
    if (!vector_scan(conn->call, 0, PPTP_MAX_CHANNELS - 1, &i))
	/* no more calls available! */
	return NULL;
    /* allocate structure. */
    if ((call = malloc(sizeof(*call))) == NULL) return NULL;
    /* Initialize call structure */
    call->call_type = PPTP_CALL_PNS;
    call->state.pns = PNS_IDLE;
    call->call_id   = (u_int16_t) i;
    call->sernum    = conn->call_serial_number++;
    call->callback  = callback;
    call->closure   = NULL;
    packet.call_id = htons(call->call_id);
    packet.call_sernum = htons(call->sernum);

#ifdef SUPPORT_BEZEQ_ISRAEL
    {
	int idx, rc;

	/* if we have a quirk, build a new packet to fit it */
	idx = get_quirk_index();
	if (idx != -1 && pptp_fixups[idx].out_call_rqst_hook) {
		if ((rc = pptp_fixups[idx].out_call_rqst_hook(&packet)))
			warn("calling the out_call_rqst hook failed (%d)", rc);
	}
    }
#endif

    /* fill in the phone number if it was specified */
    if (phonenr) {
	strcpy(packet.phone_num, phonenr);
	packet.phone_len = strlen(phonenr);
	if( packet.phone_len > sizeof(packet.phone_num))
	    packet.phone_len = sizeof(packet.phone_num);
	packet.phone_len = hton16 (packet.phone_len);
    }
    if (pptp_send_ctrl_packet(conn, &packet, sizeof(packet))) {
	pptp_reset_timer();
	call->state.pns = PNS_WAIT_REPLY;
	/* and add it to the call vector */
	vector_insert(conn->call, i, call);
	return call;
    } else { /* oops, unsuccessful. Deallocate. */
	free(call);
	return NULL;
    }
}

/*** Non-blocking read ********************************************************/
static void pptp_read_some(PPTP_CONN * conn)
{
    ssize_t retval;

    assert(conn && conn->call);
    if (conn->read_size == conn->read_alloc) { /* need to alloc more memory */
	char *new_buffer = realloc(conn->read_buffer,
		sizeof(*(conn->read_buffer)) * conn->read_alloc * 2);
	if (new_buffer == NULL) {
	    logF(bb_msg_memory_exhausted); return;
	}
	conn->read_alloc *= 2;
	conn->read_buffer = new_buffer;
    }
    retval = read(conn->inet_sock, conn->read_buffer + conn->read_size,
	    conn->read_alloc  - conn->read_size);
    if (retval < 0) {
	if (errno == EINTR || errno == EAGAIN)
	    /* ignore */ ;
	else { /* a real error */
	    logF("read error: %m");
	    pptp_conn_destroy(conn); /* shut down fast. */
	}
	return;
    }
    conn->read_size += retval;
    assert(conn->read_size <= conn->read_alloc);
}

/* Make valid packets from read_buffer */
static int pptp_make_packet(PPTP_CONN * conn, void **buf, size_t *size)
{
    struct pptp_header *header;
    size_t bad_bytes = 0;
    assert(conn && conn->call); assert(buf != NULL); assert(size != NULL);
    /* Give up unless there are at least sizeof(pptp_header) bytes */
    while ((conn->read_size-bad_bytes) >= sizeof(struct pptp_header)) {
	/* Throw out bytes until we have a valid header. */
	header = (struct pptp_header *) (conn->read_buffer + bad_bytes);
	if (ntoh32(header->magic) != PPTP_MAGIC) goto throwitout;
	if (ntoh16(header->reserved0) != 0)
	    logF("reserved0 field is not zero! (0x%x) Cisco feature? \n",
		    ntoh16(header->reserved0));
	if (ntoh16(header->length) < sizeof(struct pptp_header)) goto throwitout;
	if (ntoh16(header->length) > PPTP_CTRL_SIZE_MAX) goto throwitout;
	/* well.  I guess it's good. Let's see if we've got it all. */
	if (ntoh16(header->length) > (conn->read_size-bad_bytes))
	    /* nope.  Let's wait until we've got it, then. */
	    goto flushbadbytes;
	/* One last check: */
	if ((ntoh16(header->pptp_type) == PPTP_MESSAGE_CONTROL) &&
		(ntoh16(header->length) !=
			 PPTP_CTRL_SIZE(ntoh16(header->ctrl_type))))
	    goto throwitout;
	/* well, I guess we've got it. */
	*size = ntoh16(header->length);
	*buf = malloc(*size);
	if (*buf == NULL) { logF(bb_msg_memory_exhausted); return 0; /* ack! */ }
	memcpy(*buf, conn->read_buffer + bad_bytes, *size);
	/* Delete this packet from the read_buffer. */
	conn->read_size -= (bad_bytes + *size);
	memmove(conn->read_buffer, conn->read_buffer + bad_bytes + *size,
		conn->read_size);
	if (bad_bytes > 0)
	    logF("%lu bad bytes thrown away.", (unsigned long) bad_bytes);
	return 1;
throwitout:
	bad_bytes++;
    }
flushbadbytes:
    /* no more packets.  Let's get rid of those bad bytes */
    conn->read_size -= bad_bytes;
    memmove(conn->read_buffer, conn->read_buffer + bad_bytes, conn->read_size);
    if (bad_bytes > 0)
	logF("%lu bad bytes thrown away.", (unsigned long) bad_bytes);
    return 0;
}

/* pptp_ctrl.c ... handle PPTP control connection.
 */


/* BECAUSE OF SIGNAL LIMITATIONS, EACH PROCESS CAN ONLY MANAGE ONE
 * CONNECTION.  SO THIS 'PPTP_CONN' STRUCTURE IS A BIT MISLEADING.
 * WE'LL KEEP CONNECTION-SPECIFIC INFORMATION IN THERE ANYWAY (AS
 * OPPOSED TO USING GLOBAL VARIABLES), BUT BEWARE THAT THE ENTIRE
 * UNIX SIGNAL-HANDLING SEMANTICS WOULD HAVE TO CHANGE (OR THE
 * TIME-OUT CODE DRASTICALLY REWRITTEN) BEFORE YOU COULD DO A
 * PPTP_CONN_OPEN MORE THAN ONCE PER PROCESS AND GET AWAY WITH IT.
 */


/* PPTP error codes: ----------------------------------------------*/

/* (General Error Codes) */
static const struct {
    const char *name, *desc;
} pptp_general_errors[] = {
#define PPTP_GENERAL_ERROR_NONE                 0
    { "(None)", "No general error" },
#define PPTP_GENERAL_ERROR_NOT_CONNECTED        1
    { "(Not-Connected)", "No control connection exists yet for this "
	"PAC-PNS pair" },
#define PPTP_GENERAL_ERROR_BAD_FORMAT           2
    { "(Bad-Format)", "Length is wrong or Magic Cookie value is incorrect" },
#define PPTP_GENERAL_ERROR_BAD_VALUE            3
    { "(Bad-Value)", "One of the field values was out of range or "
	    "reserved field was non-zero" },
#define PPTP_GENERAL_ERROR_NO_RESOURCE          4
    { "(No-Resource)", "Insufficient resources to handle this command now" },
#define PPTP_GENERAL_ERROR_BAD_CALLID           5
    { "(Bad-Call ID)", "The Call ID is invalid in this context" },
#define PPTP_GENERAL_ERROR_PAC_ERROR            6
    { "(PAC-Error)", "A generic vendor-specific error occured in the PAC" }
};

#define  MAX_GENERAL_ERROR ( sizeof(pptp_general_errors) / \
	sizeof(pptp_general_errors[0]) - 1)

/* Outgoing Call Reply Result Codes */
static const char * const pptp_out_call_reply_result[] = {
/* 0 */ "Unknown Result Code",
/* 1 */ "Connected",
/* 2 */ "General Error",
/* 3 */ "No Carrier Detected",
/* 4 */ "Busy Signal",
/* 5 */ "No Dial Tone",
/* 6 */ "Time Out",
/* 7 */ "Not Accepted, Call is administratively prohibited" };

#define MAX_OUT_CALL_REPLY_RESULT 7

/* Call Disconnect Notify  Result Codes */
static const char * const pptp_call_disc_ntfy[] = {
/* 0 */ "Unknown Result Code",
/* 1 */ "Lost Carrier",
/* 2 */ "General Error",
/* 3 */ "Administrative Shutdown",
/* 4 */ "(your) Request" };

#define MAX_CALL_DISC_NTFY 4

/* Call Disconnect Notify  Result Codes */
static const char * const pptp_start_ctrl_conn_rply[] = {
/* 0 */ "Unknown Result Code",
/* 1 */ "Successful Channel Establishment",
/* 2 */ "General Error",
/* 3 */ "Command Channel Already Exists",
/* 4 */ "Requester is not Authorized" };

#define MAX_START_CTRL_CONN_REPLY 4


/*** log echo request/replies *************************************************/
static void logecho(int type)
{
    /* hack to stop flooding the log files (the most interesting part is right
     * after the connection built-up) */
    if( nlogecho > 0) {
	logF( "Echo Re%s received.", type == PPTP_ECHO_RQST ? "quest" :"ply");
	if( --nlogecho == 0)
	    logF("no more Echo Reply/Request packets will be reported.");
    }
}

/*** log error information in control packets *********************************/
static void ctrlp_error( int result, int error, int cause,
	const char * const result_text[], int max_result)
{
    if( cause >= 0)
	logF("Result code is %d '%s'. Error code is %d, Cause code is %d",
		result, result_text[result <= max_result ?  result : 0], error,
		cause );
    else
	logF("Reply result code is %d '%s'. Error code is %d",
		result, result_text[result <= max_result ?  result : 0], error);
    if ((error > 0) && (error <= MAX_GENERAL_ERROR)){
	if( result != PPTP_RESULT_GENERAL_ERROR )
	    logF("Result code is something else then \"general error\", "
		    "so the following error is probably bogus.");
	logF("Error is '%s', Error message: '%s'",
		pptp_general_errors[error].name,
		pptp_general_errors[error].desc);
    }
}

#ifdef SUPPORT_BEZEQ_ISRAEL
/* Set link info, for pptp servers that need it.
   this is a noop, unless the user specified a quirk and
   there's a set_link hook defined in the quirks table
   for that quirk */
static void pptp_set_link(PPTP_CONN* conn, int peer_call_id)
{
    int idx, rc;
    /* if we need to send a set_link packet because of buggy
       hardware or pptp server, do it now */
    if ((idx = get_quirk_index()) != -1 && pptp_fixups[idx].set_link_hook) {
	struct pptp_set_link_info packet;
	if ((rc = pptp_fixups[idx].set_link_hook(&packet, peer_call_id)))
	    warn("calling the set_link hook failed (%d)", rc);
	if (pptp_send_ctrl_packet(conn, &packet, sizeof(packet))) {
	    pptp_reset_timer();
	}
    }
}
#endif


/* Dispatch packets (control messages) */
static void ctrlp_disp(PPTP_CONN * conn, void * buffer, size_t size)
{
    struct pptp_header *header = (struct pptp_header *)buffer;
    u_int8_t close_reason = PPTP_STOP_NONE;
    assert(conn && conn->call); assert(buffer);
    assert(ntoh32(header->magic) == PPTP_MAGIC);
    assert(ntoh16(header->length) == size);
    assert(ntoh16(header->pptp_type) == PPTP_MESSAGE_CONTROL);
    if (size < PPTP_CTRL_SIZE(ntoh16(header->ctrl_type))) {
	logF("Invalid packet received [type: %d; length: %d].",
		(int) ntoh16(header->ctrl_type), (int) size);
	return;
    }
    switch (ntoh16(header->ctrl_type)) {
	/* ----------- STANDARD Start-Session MESSAGES ------------ */
	case PPTP_START_CTRL_CONN_RQST:
	{
	    struct pptp_start_ctrl_conn *packet =
		(struct pptp_start_ctrl_conn *) buffer;
	    struct pptp_start_ctrl_conn reply = {
		PPTP_HEADER_CTRL(PPTP_START_CTRL_CONN_RPLY),
		hton16(PPTP_VERSION), 0, 0,
		hton32(PPTP_FRAME_CAP), hton32(PPTP_BEARER_CAP),
		hton16(PPTP_MAX_CHANNELS), hton16(PPTP_FIRMWARE_VERSION),
		PPTP_HOSTNAME, PPTP_VENDOR };

	    logF("Received Start Control Connection Request");

#ifdef SUPPORT_BEZEQ_ISRAEL
	    {
		int idx, rc;

		/* fix this packet, if necessary */
		idx = get_quirk_index();
		if (idx != -1 && pptp_fixups[idx].start_ctrl_conn) {
		    if ((rc = pptp_fixups[idx].start_ctrl_conn(&reply)))
			warn("calling the start_ctrl_conn hook failed (%d)", rc);
		}
	    }
#endif

	    if (conn->conn_state == CONN_IDLE) {
		if (ntoh16(packet->version) < PPTP_VERSION) {
		    /* Can't support this (earlier) PPTP_VERSION */
		    reply.version = packet->version;
		    /* protocol version not supported */
		    reply.result_code = hton8(5);
		    pptp_send_ctrl_packet(conn, &reply, sizeof(reply));
		    pptp_reset_timer(); /* give sender a chance for a retry */
		} else { /* same or greater version */
		    if (pptp_send_ctrl_packet(conn, &reply, sizeof(reply))) {
			conn->conn_state = CONN_ESTABLISHED;
			logF("server connection ESTABLISHED.");
			pptp_reset_timer();
		    }
		}
	    }
	    break;
	}
	case PPTP_START_CTRL_CONN_RPLY:
	{
	    struct pptp_start_ctrl_conn *packet =
		(struct pptp_start_ctrl_conn *) buffer;
	    logF("Received Start Control Connection Reply");
	    if (conn->conn_state == CONN_WAIT_CTL_REPLY) {
		/* XXX handle collision XXX [see rfc] */
		if (ntoh16(packet->version) != PPTP_VERSION) {
		    if (conn->callback != NULL)
			conn->callback(conn, CONN_OPEN_FAIL);
		    close_reason = PPTP_STOP_PROTOCOL;
		    goto pptp_conn_close;
		}
		if (ntoh8(packet->result_code) != 1 &&
		    /* J'ai change le if () afin que la connection ne se ferme
		     * pas pour un "rien" :p adel@cybercable.fr -
		     *
		     * Don't close the connection if the result code is zero
		     * (feature found in certain ADSL modems)
		     */
			ntoh8(packet->result_code) != 0) {
		    logF("Negative reply received to our Start Control "
			    "Connection Request");
		    ctrlp_error(packet->result_code, packet->error_code,
			    -1, pptp_start_ctrl_conn_rply,
			    MAX_START_CTRL_CONN_REPLY);
		    if (conn->callback != NULL)
			conn->callback(conn, CONN_OPEN_FAIL);
		    close_reason = PPTP_STOP_PROTOCOL;
		    goto pptp_conn_close;
		}
		conn->conn_state = CONN_ESTABLISHED;
		/* log session properties */
		conn->version      = ntoh16(packet->version);
		conn->firmware_rev = ntoh16(packet->firmware_rev);
		memcpy(conn->hostname, packet->hostname, sizeof(conn->hostname));
		memcpy(conn->vendor, packet->vendor, sizeof(conn->vendor));
		pptp_reset_timer(); /* 60 seconds until keep-alive */
		logF("Client connection established.");
		if (conn->callback != NULL)
		    conn->callback(conn, CONN_OPEN_DONE);
	    } /* else goto pptp_conn_close; */
	    break;
	}
	    /* ----------- STANDARD Stop-Session MESSAGES ------------ */
	case PPTP_STOP_CTRL_CONN_RQST:
	{
	    /* conn_state should be CONN_ESTABLISHED, but it could be
	     * something else */
	    struct pptp_stop_ctrl_conn reply = {
		PPTP_HEADER_CTRL(PPTP_STOP_CTRL_CONN_RPLY),
		hton8(1), hton8(PPTP_GENERAL_ERROR_NONE), 0
	    };
	    logF("Received Stop Control Connection Request.");
	    if (conn->conn_state == CONN_IDLE) break;
	    if (pptp_send_ctrl_packet(conn, &reply, sizeof(reply))) {
		if (conn->callback != NULL)
		    conn->callback(conn, CONN_CLOSE_RQST);
		conn->conn_state = CONN_IDLE;
		pptp_conn_destroy(conn);
	    }
	    break;
	}
	case PPTP_STOP_CTRL_CONN_RPLY:
	{
	    logF("Received Stop Control Connection Reqply.");
	    /* conn_state should be CONN_WAIT_STOP_REPLY, but it
	     * could be something else */
	    if (conn->conn_state == CONN_IDLE) break;
	    conn->conn_state = CONN_IDLE;
	    pptp_conn_destroy(conn);
	    break;
	}
	    /* ----------- STANDARD Echo/Keepalive MESSAGES ------------ */
	case PPTP_ECHO_RPLY:
	{
	    struct pptp_echo_rply *packet =
		(struct pptp_echo_rply *) buffer;
	    logecho(PPTP_ECHO_RPLY);
	    if ((conn->ka_state == KA_OUTSTANDING) &&
		    (ntoh32(packet->identifier) == conn->ka_id)) {
		conn->ka_id++;
		conn->ka_state = KA_NONE;
		pptp_reset_timer();
	    }
	    break;
	}
	case PPTP_ECHO_RQST:
	{
	    struct pptp_echo_rqst *packet =
		(struct pptp_echo_rqst *) buffer;
	    struct pptp_echo_rply reply = {
		PPTP_HEADER_CTRL(PPTP_ECHO_RPLY),
		packet->identifier, /* skip hton32(ntoh32(id)) */
		hton8(1), hton8(PPTP_GENERAL_ERROR_NONE), 0
	    };
	    logecho(PPTP_ECHO_RQST);
	    pptp_send_ctrl_packet(conn, &reply, sizeof(reply));
	    pptp_reset_timer();
	    break;
	}
	    /* ----------- OUTGOING CALL MESSAGES ------------ */
	case PPTP_OUT_CALL_RQST:
	{
	    struct pptp_out_call_rqst *packet =
		(struct pptp_out_call_rqst *)buffer;
	    struct pptp_out_call_rply reply = {
		PPTP_HEADER_CTRL(PPTP_OUT_CALL_RPLY),
		0 /* callid */, packet->call_id, 1, PPTP_GENERAL_ERROR_NONE, 0,
		hton32(PPTP_CONNECT_SPEED),
		hton16(PPTP_WINDOW), hton16(PPTP_DELAY), 0
	    };
	    logF("Received Outgoing Call Request.");
	    /* XXX PAC: eventually this should make an outgoing call. XXX */
	    reply.result_code = hton8(7); /* outgoing calls verboten */
	    pptp_send_ctrl_packet(conn, &reply, sizeof(reply));
	    break;
	}
	case PPTP_OUT_CALL_RPLY:
	{
	    struct pptp_out_call_rply *packet =
		(struct pptp_out_call_rply *)buffer;
	    PPTP_CALL * call;
	    u_int16_t callid = ntoh16(packet->call_id_peer);
	    logF("Received Outgoing Call Reply.");
	    if (!vector_search(conn->call, (int) callid, &call)) {
		logF("PPTP_OUT_CALL_RPLY received for non-existant call: "
			"peer call ID (us)  %d call ID (them) %d.",
			callid, ntoh16(packet->call_id));
		break;
	    }
	    if (call->call_type != PPTP_CALL_PNS) {
		logF("Ack!  How did this call_type get here?"); /* XXX? */
		break;
	    }
	    if (call->state.pns != PNS_WAIT_REPLY) {
		warn("Unexpected(?) Outgoing Call Reply will be ignored.");
		break;
	    }
	    /* check for errors */
	    if (packet->result_code != 1) {
		/* An error.  Log it verbosely. */
		logF("Our outgoing call request [callid %d] has not been "
			"accepted.", (int) callid);
		ctrlp_error(packet->result_code, packet->error_code,
			packet->cause_code, pptp_out_call_reply_result,
			MAX_OUT_CALL_REPLY_RESULT);
		call->state.pns = PNS_IDLE;
		if (call->callback != NULL)
		    call->callback(conn, call, CALL_OPEN_FAIL);
		pptp_call_destroy(conn, call);
	    } else {
		/* connection established */
		call->state.pns = PNS_ESTABLISHED;
		call->peer_call_id = ntoh16(packet->call_id);
		call->speed        = ntoh32(packet->speed);
		pptp_reset_timer();

#ifdef SUPPORT_BEZEQ_ISRAEL
		/* call pptp_set_link. unless the user specified a quirk
		   and this quirk has a set_link hook, this is a noop */
		pptp_set_link(conn, call->peer_call_id);
#endif
		if (call->callback != NULL)
		    call->callback(conn, call, CALL_OPEN_DONE);
		logF("Outgoing call established (call ID %u, peer's "
			"call ID %u).\n", call->call_id, call->peer_call_id);
	    }
	    break;
	}
	    /* ----------- INCOMING CALL MESSAGES ------------ */
	    /* XXX write me XXX */
	    /* ----------- CALL CONTROL MESSAGES ------------ */
	case PPTP_CALL_CLEAR_RQST:
	{
	    struct pptp_call_clear_rqst *packet =
		(struct pptp_call_clear_rqst *)buffer;
	    struct pptp_call_clear_ntfy reply = {
		PPTP_HEADER_CTRL(PPTP_CALL_CLEAR_NTFY), packet->call_id,
		1, PPTP_GENERAL_ERROR_NONE, 0, 0, {0}
	    };
	    logF("Received Call Clear Request.");
	    if (vector_contains(conn->call, ntoh16(packet->call_id))) {
		PPTP_CALL * call;
		vector_search(conn->call, ntoh16(packet->call_id), &call);
		if (call->callback != NULL)
		    call->callback(conn, call, CALL_CLOSE_RQST);
		pptp_send_ctrl_packet(conn, &reply, sizeof(reply));
		pptp_call_destroy(conn, call);
		logF("Call closed (RQST) (call id %d)", (int) call->call_id);
	    }
	    break;
	}
	case PPTP_CALL_CLEAR_NTFY:
	{
	    struct pptp_call_clear_ntfy *packet =
		(struct pptp_call_clear_ntfy *)buffer;
	    logF("Call disconnect notification received (call id %d)",
		    ntoh16(packet->call_id));
	    if (vector_contains(conn->call, ntoh16(packet->call_id))) {
		PPTP_CALL * call;
		ctrlp_error(packet->result_code, packet->error_code,
			packet->cause_code, pptp_call_disc_ntfy,
			MAX_CALL_DISC_NTFY);
		vector_search(conn->call, ntoh16(packet->call_id), &call);
		pptp_call_destroy(conn, call);
	    }
	    /* XXX we could log call stats here XXX */
	    /* XXX not all servers send this XXX */
	    break;
	}
	case PPTP_SET_LINK_INFO:
	{
	    /* I HAVE NO CLUE WHAT TO DO IF send_accm IS NOT 0! */
	    /* this is really dealt with in the HDLC deencapsulation, anyway. */
	    struct pptp_set_link_info *packet =
		(struct pptp_set_link_info *)buffer;
	    /* log it. */
	    logF("PPTP_SET_LINK_INFO received from peer_callid %u",
		    (unsigned int) ntoh16(packet->call_id_peer));
	    logF("  send_accm is %08lX, recv_accm is %08lX",
		    (unsigned long) ntoh32(packet->send_accm),
		    (unsigned long) ntoh32(packet->recv_accm));
	    if (!(ntoh32(packet->send_accm) == 0 &&
		    ntoh32(packet->recv_accm) == 0))
		warn("Non-zero Async Control Character Maps are not supported!");
	    break;
	}
	default:
	    logF("Unrecognized Packet %d received.",
		    (int) ntoh16(((struct pptp_header *)buffer)->ctrl_type));
	    /* goto pptp_conn_close; */
	    break;
    }
    return;
pptp_conn_close:
    warn("pptp_conn_close(%d)", (int) close_reason);
    pptp_conn_close(conn, close_reason);
}

/*** Packet Dispatch (general) ************************************************/
static void pptp_dispatch_packet(PPTP_CONN * conn, void * buffer, size_t size)
{
    struct pptp_header *header = (struct pptp_header *)buffer;

    assert(conn && conn->call); assert(buffer);
    assert(ntoh32(header->magic) == PPTP_MAGIC);
    assert(ntoh16(header->length) == size);
    switch (ntoh16(header->pptp_type)) {
	case PPTP_MESSAGE_CONTROL:
	    ctrlp_disp(conn, buffer, size);
	    break;
	case PPTP_MESSAGE_MANAGE:
	    /* MANAGEMENT messages aren't even part of the spec right now. */
	    logF("PPTP management message received, but not understood.");
	    break;
	default:
	    logF("Unknown PPTP control message type received: %u",
		    (unsigned int) ntoh16(header->pptp_type));
	    break;
    }
}

/*** handle any pptp file descriptors set in fd_set, and clear them ***********/
static void
pptp_dispatch(PPTP_CONN * conn, fd_set * read_set, fd_set * write_set)
{
    assert(conn && conn->call);
    /* Check write_set could be set. */
    if (FD_ISSET(conn->inet_sock, write_set)) {
	FD_CLR(conn->inet_sock, write_set);
	if (conn->write_size > 0)
	    pptp_write_some(conn);/* write as much as we can without blocking */
    }
    /* Check read_set */
    if (FD_ISSET(conn->inet_sock, read_set)) {
	void *buffer; size_t size;
	FD_CLR(conn->inet_sock, read_set);
	pptp_read_some(conn); /* read as much as we can without blocking */
	/* make packets of the buffer, while we can. */
	while (pptp_make_packet(conn, &buffer, &size)) {
	    pptp_dispatch_packet(conn, buffer, size);
	    free(buffer);
	}
    }
    /* That's all, folks.  Simple, eh? */
}

/* re-open stderr as /dev/null to release it */
static int null2fd(void)
{
    int ok = 0;
    FILE *file;

    file = fopen(bb_dev_null, "w");
    if (file != NULL && dup2(fileno(file), STDERR_FILENO) != -1)
	ok = 1;
    if (file) fclose(file);
    return ok;
}

/*** Call Manager *************************************************************/
static int callmgr_main(int argc, char **argv)
{
    struct in_addr inetaddr;
    int inet_sock, unix_sock;
    fd_set call_set;
    PPTP_CONN * conn;
    VECTOR * call_list;
    int max_fd = 0;
    volatile int first = 1;
    int retval;
    int i;
    char *phonenr;

    /* Step 0: Check arguments */
    if (argc < 2)
	fatal("Usage: %s ip.add.ress.here [--phone <phone number>]", argv[0]);
    phonenr = argc == 3 ? argv[2] : NULL;
    if (inet_aton(argv[1], &inetaddr) == 0)
	fatal("Invalid IP address: %s", argv[1]);
    /* Step 1: Open sockets. */
    if ((inet_sock = open_inetsock(inetaddr)) < 0)
	fatal("Could not open control connection to %s", argv[1]);
    if ((unix_sock = open_unixsock(inetaddr)) < 0)
	fatal("Could not open unix socket for %s", argv[1]);
    /* Step 1b: FORK and return status to calling process. */
    switch (fork()) {
	case 0: /* child. stick around. */
	    break;
	case -1: /* failure.  Fatal. */
	    fatal("Could not fork.");
	default: /* Parent. Return status to caller. */
	    exit(0);
    }
    /* re-open stderr as /dev/null to release it */
    null2fd();
    /* Step 1c: Clean up unix socket on TERM */
    if (sigsetjmp(callmgr_env, 1) != 0)
	goto cleanup;
    bb_signal((1<<SIGINT)|(1<<SIGTERM), callmgr_sighandler);
    /* SIGUSR1 - signal state change wake up accept */
    bb_signal((1<<SIGPIPE)|(1<<SIGUSR1), callmgr_do_nothing);

    /* Step 2: Open control connection and register callback */
    if ((conn = pptp_conn_open(inet_sock, 1, NULL/* callback */)) == NULL) {
	close(unix_sock); close(inet_sock); fatal("Could not open connection.");
    }
    FD_ZERO(&call_set);
    call_list = vector_create();
    {
	struct local_conninfo *conninfo = malloc(sizeof(*conninfo));
	if (conninfo == NULL) {
	    close(unix_sock); close(inet_sock); fatal(bb_msg_memory_exhausted);
	}
	conninfo->call_list = call_list;
	conninfo->call_set  = &call_set;
	pptp_conn_closure_put(conn, conninfo);
    }
    if (sigsetjmp(callmgr_env, 1) != 0) goto shutdown;

    /* Step 3: Get FD_SETs */
    do {
	int rc;
	fd_set read_set = call_set, write_set;
	FD_ZERO (&write_set);
	max_fd = 0;
	if (pptp_conn_established(conn)) {
	  FD_SET (unix_sock, &read_set);
	  if (unix_sock > max_fd) max_fd = unix_sock;
	}
	pptp_fd_set(conn, &read_set, &write_set, &max_fd);
	for (; max_fd > 0 ; max_fd--) {
	    if (FD_ISSET (max_fd, &read_set) ||
		    FD_ISSET (max_fd, &write_set))
		break;
	}
	/* Step 4: Wait on INET or UNIX event */
	if ((rc = select(max_fd + 1, &read_set, &write_set, NULL, NULL)) <0) {
	  if (errno == EBADF) break;
	  /* a signal or somesuch. */
	  continue;
	}
	/* Step 5a: Handle INET events */
	pptp_dispatch(conn, &read_set, &write_set);
	/* Step 5b: Handle new connection to UNIX socket */
	if (FD_ISSET(unix_sock, &read_set)) {
	    /* New call! */
	    struct sockaddr_un from;
	    socklen_t len = sizeof(from);
	    PPTP_CALL * call;
	    struct local_callinfo *lci;
	    int s;
	    /* Accept the socket */
	    FD_CLR (unix_sock, &read_set);
	    if ((s = accept(unix_sock, (struct sockaddr *) &from, &len)) < 0) {
		warn("Socket not accepted: %m");
		goto skip_accept;
	    }
	    /* Allocate memory for local call information structure. */
	    if ((lci = malloc(sizeof(*lci))) == NULL) {
		warn(bb_msg_memory_exhausted); close(s); goto skip_accept;
	    }
	    lci->unix_sock = s;
	    /* Give the initiator time to write the PIDs while we open
	     * the call */
	    call = pptp_call_open(conn, call_callback, phonenr);
	    /* Read and store the associated pids */
	    read(s, &lci->pid[0], sizeof(lci->pid[0]));
	    read(s, &lci->pid[1], sizeof(lci->pid[1]));
	    /* associate the local information with the call */
	    pptp_call_closure_put(conn, call, (void *) lci);
	    /* The rest is done on callback. */
	    /* Keep alive; wait for close */
	    retval = vector_insert(call_list, s, call); assert(retval);
	    if (s > max_fd) max_fd = s;
	    FD_SET(s, &call_set);
	    first = 0;
	}
skip_accept: /* Step 5c: Handle socket close */
	for (i = 0; i < max_fd + 1; i++)
	    if (FD_ISSET(i, &read_set)) {
		/* close it */
		PPTP_CALL * call;
		retval = vector_search(call_list, i, &call);
		if (retval) {
		    struct local_callinfo *lci =
			pptp_call_closure_get(conn, call);
		    logF("Closing connection");
		    if(lci->pid[0]) kill(lci->pid[0], SIGTERM);
		    if(lci->pid[1]) kill(lci->pid[1], SIGTERM);
		    free(lci);
		    /* soft shutdown.  Callback will do hard shutdown later */
		    pptp_call_close(conn, call);
		    vector_remove(call_list, i);
		}
		FD_CLR(i, &call_set);
		close(i);
	    }
    } while (vector_size(call_list) > 0 || first);
shutdown:
    {
	fd_set read_set, write_set;
	struct timeval tv;
	/* warn("Shutdown"); */
	/* kill all open calls */
	for (i = 0; i < vector_size(call_list); i++) {
	    PPTP_CALL *call = vector_get_Nth(call_list, i);
	    struct local_callinfo *lci = pptp_call_closure_get(conn, call);
	    logF("Closing connection");
	    pptp_call_close(conn, call);
	    if(lci->pid[0]) kill(lci->pid[0], SIGTERM);
	    if(lci->pid[1]) kill(lci->pid[1], SIGTERM);
	}
	/* attempt to dispatch these messages */
	FD_ZERO(&read_set);
	FD_ZERO(&write_set);
	pptp_fd_set(conn, &read_set, &write_set, &max_fd);
	pptp_dispatch(conn, &read_set, &write_set);
	/* wait for a respond, a timeout because there might not be one */
	FD_ZERO(&read_set);
	FD_ZERO(&write_set);
	pptp_fd_set(conn, &read_set, &write_set, &max_fd);
	tv.tv_sec = 2;
	tv.tv_usec = 0;
	select(max_fd + 1, &read_set, &write_set, NULL, &tv);
	pptp_dispatch(conn, &read_set, &write_set);
	if (i > 0) sleep(2);
	/* no more open calls.  Close the connection. */
	pptp_conn_close(conn, PPTP_STOP_LOCAL_SHUTDOWN);
	/* wait for a respond, a timeout because there might not be one */
	FD_ZERO(&read_set);
	FD_ZERO(&write_set);
	pptp_fd_set(conn, &read_set, &write_set, &max_fd);
	tv.tv_sec = 2;
	tv.tv_usec = 0;
	select(max_fd + 1, &read_set, &write_set, NULL, &tv);
	pptp_dispatch(conn, &read_set, &write_set);
	sleep(2);
	/* with extreme prejudice */
	pptp_conn_destroy(conn);
	vector_destroy(call_list);
    }
cleanup:
    close_inetsock(inet_sock, inetaddr);
    close_unixsock(unix_sock, inetaddr);
    return 0;
}

/*** call the call manager main ***********************************************/
static void
launch_callmgr(struct in_addr inetaddr, char *phonenr)
{
      char *my_argv[3] = { Gargv0, inet_ntoa(inetaddr), phonenr };

      bb_setproctitle("pptp: call manager for %s", my_argv[1]);
      exit(callmgr_main(3, my_argv));
}


/* pptp_gre.c  -- encapsulate PPP in PPTP-GRE.
 *                Handle the IP Protocol 47 portion of PPTP.
 */

/* test for a 32 bit counter overflow */
#define WRAPPED( curseq, lastseq) \
    ((((curseq) & 0xffffff00) == 0) && \
     (((lastseq) & 0xffffff00 ) == 0xffffff00))


typedef int (*callback_t)(int cl, void *pack, unsigned int len);

/* decaps_hdlc gets all the packets possible with ONE blocking read */
/* returns <0 if read() call fails */
static int decaps_hdlc(int fd, callback_t callback, int cl);
static int encaps_hdlc(int fd, void *pack, unsigned int len);
static int decaps_gre (int fd, callback_t callback, int cl);
static int encaps_gre (int fd, void *pack, unsigned int len);
static int dequeue_gre(callback_t callback, int cl);

/*** time_now_usecs ***********************************************************/
static uint64_t time_now_usecs(void)
{
    struct timeval tv;

    gettimeofday(&tv, NULL);
    return (tv.tv_sec * 1000000) + tv.tv_usec;
}

/*** Open IP protocol socket **************************************************/
static int pptp_gre_bind(struct in_addr inetaddr)
{
    struct sockaddr_in src_addr, loc_addr;
    int s = bb_xsocket(BB_XSOCKET_FLG_NO_WARN | BB_XSOCKET_FLG_SOCK_RAW | BB_XSOCKET_FLG_PROTO_SPEC, PPTP_PROTO);

    if (s < 0) { warn("socket: %m"); return -1; }
    if (localbind.s_addr != INADDR_NONE) {
	memset(&loc_addr, 0, sizeof(loc_addr));
	loc_addr.sin_family = AF_INET;
	loc_addr.sin_addr   = localbind;
	if (bind(s, (struct sockaddr *) &loc_addr, sizeof(loc_addr)) != 0) {
	    warn("bind: %m"); close(s); return -1;
	}
    }
    src_addr.sin_family = AF_INET;
    src_addr.sin_addr   = inetaddr;
    src_addr.sin_port   = 0;
    if (connect(s, (struct sockaddr *) &src_addr, sizeof(src_addr)) < 0) {
	warn("connect: %m"); close(s); return -1;
    }
    return s;
}

/*** pptp_gre_copy ************************************************************/
static void pptp_gre_copy(u_int16_t call_id, u_int16_t peer_call_id,
		   int pty_fd, int gre_fd)
{
    int max_fd;
    pptp_gre_call_id = call_id;
    pptp_gre_peer_call_id = peer_call_id;
    /* Pseudo-terminal already open. */
    ack_sent = ack_recv = seq_sent = seq_recv = 0;
    /* weird select semantics */
    max_fd = gre_fd;
    if (pty_fd > max_fd) max_fd = pty_fd;
    /* Dispatch loop */
    for (;;) { /* until error happens on gre_fd or pty_fd */
	struct timeval tv = {0, 0};
	struct timeval *tvp;
	fd_set rfds;
	int retval;
	pqueue_t *head;
	int block_usecs = -1; /* wait forever */
	/* watch terminal and socket for input */
	FD_ZERO(&rfds);
	FD_SET(gre_fd, &rfds);
	FD_SET(pty_fd, &rfds);
	/*
	 * if there are multiple pending ACKs, then do a minimal timeout;
	 * else if there is a single pending ACK then timeout after 0,5 seconds;
	 * else block until data is available.
	 */
	if (ack_sent != seq_recv) {
	    if (ack_sent + 1 == seq_recv)  /* u_int wrap-around safe */
		block_usecs = 500000;
	    else
		/* don't use zero, this will force a resceduling */
		/* when calling select(), giving pppd a chance to */
		/* run. */
		block_usecs = 1;
	}
	/* otherwise block_usecs == -1, which means wait forever */
	/*
	 * If there is a packet in the queue, then don't wait longer than
	 * the time remaining until it expires.
	 */
	head = pqueue_head();
	if (head != NULL) {
	    int expiry_time = pqueue_expiry_time(head);
	    if (block_usecs == -1 || expiry_time < block_usecs)
		block_usecs = expiry_time;
	}
	if (block_usecs == -1) {
	    tvp = NULL;
	} else {
	    tvp = &tv;
	    tv.tv_usec = block_usecs;
	    tv.tv_sec  = tv.tv_usec / 1000000;
	    tv.tv_usec %= 1000000;
	}
	retval = select(max_fd + 1, &rfds, NULL, NULL, tvp);
	if (FD_ISSET(pty_fd, &rfds)) {
	    if (decaps_hdlc(pty_fd, encaps_gre,  gre_fd) < 0)
		break;
	} else if (retval == 0 && ack_sent != seq_recv) {
	    /* if outstanding ack */
	    /* send ack with no payload */
	    encaps_gre(gre_fd, NULL, 0);
	}
	if (FD_ISSET(gre_fd, &rfds)) {
	    if (decaps_gre (gre_fd, encaps_hdlc, pty_fd) < 0)
		break;
	}
	if (dequeue_gre (encaps_hdlc, pty_fd) < 0)
	    break;
    }
    /* Close up when done. */
    close(gre_fd);
    close(pty_fd);
}

#define HDLC_FLAG         0x7E
#define HDLC_ESCAPE       0x7D
#define HDLC_TRANSPARENCY 0x20

/* ONE blocking read per call; dispatches all packets possible */
/* returns 0 on success, or <0 on read failure                 */
static int
decaps_hdlc(int fd, int (*cb)(int cl, void *pack, unsigned int len), int cl)
{
    unsigned char buffer[PACKET_MAX];
    unsigned int start = 0;
    int end;
    int status;
    /* start is start of packet.  end is end of buffer data */
    /*  this is the only blocking read we will allow */
    if ((end = read (fd, buffer, sizeof(buffer))) <= 0) {
	warn("short read (%d): %m", end);
	if (errno == EIO) warn("pppd may have shutdown, see pppd log");
	return -1;
    }
    /* warn if the sync options of ppp and pptp don't match */
    if( !checkedsync) {
	checkedsync = 1;
	if( buffer[0] == HDLC_FLAG){
	    if( syncppp )
		warn( "pptp --sync option is active, "
			"yet the ppp mode is asynchronous!\n");
	}
	else if( !syncppp )
	    warn( "The ppp mode is synchronous, "
		    "yet no pptp --sync option is specified!\n");
    }
    /* in synchronous mode there are no hdlc control characters nor checksum
     * bytes. Find end of packet with the length information in the PPP packet
     */
    if ( syncppp ){
	while ( start + 8 < end) {
	    decaps_len = ntoh16(*(short int *)(buffer + start + 6)) + 4;
	    /* note: the buffer may contain an incomplete packet at the end
	     * this packet will be read again at the next read() */
	    if ( start + decaps_len <= end)
		if ((status = cb (cl, buffer + start, decaps_len)) < 0)
		    return status; /* error-check */
	    start += decaps_len;
	}
	return 0;
    }
    /* asynchronous mode */
    while (start < end) {
	/* Copy to 'copy' and un-escape as we go. */
	while (buffer[start] != HDLC_FLAG) {
	    if ((escape == 0) && buffer[start] == HDLC_ESCAPE) {
		escape = HDLC_TRANSPARENCY;
	    } else {
		if (decaps_len < PACKET_MAX)
		    copy [decaps_len++] = buffer[start] ^ escape;
		escape = 0;
	    }
	    start++;
	    if (start >= end)
		return 0; /* No more data, but the frame is not complete yet. */
	}
	/* found flag.  skip past it */
	start++;
	/* check for over-short packets and silently discard, as per RFC1662 */
	if ((decaps_len < 4) || (escape != 0)) {
	    decaps_len = 0; escape = 0;
	    continue;
	}
	/* check, then remove the 16-bit FCS checksum field */
	if (pppfcs16 (PPPINITFCS16, copy, decaps_len) != PPPGOODFCS16)
	    warn("Bad Frame Check Sequence during PPP to GRE decapsulation");
	decaps_len -= sizeof(u_int16_t);
	/* so now we have a packet of length 'decaps_len' in 'copy' */
	if ((status = cb (cl, copy, decaps_len)) < 0)
	    return status; /* error-check */
	/* Great!  Let's do more! */
	decaps_len = 0; escape = 0;
    }
    return 0;
    /* No more data to process. */
}

/*** Make stripped packet into HDLC packet ************************************/
static int encaps_hdlc(int fd, void *pack, unsigned int len)
{
    unsigned char *source = (unsigned char *)pack;
    unsigned char dest[2 * PACKET_MAX + 2]; /* largest expansion possible */
    unsigned int pos = 0, i;
    u_int16_t fcs;
    /* in synchronous mode there is little to do */
    if ( syncppp )
	return write(fd, source, len);
    /* asynchronous mode */
    /* Compute the FCS */
    fcs = pppfcs16(PPPINITFCS16, source, len) ^ 0xFFFF;
    /* start character */
    dest[pos++] = HDLC_FLAG;
    /* escape the payload */
    for (i = 0; i < len + 2; i++) {
	/* wacked out assignment to add FCS to end of source buffer */
	unsigned char c =
	    (i < len)?source[i]:(i == len)?(fcs & 0xFF):((fcs >> 8) & 0xFF);
	if (pos >= sizeof(dest)) break; /* truncate on overflow */
	if ( (c < 0x20) || (c == HDLC_FLAG) || (c == HDLC_ESCAPE) ) {
	    dest[pos++] = HDLC_ESCAPE;
	    if (pos < sizeof(dest))
		dest[pos++] = c ^ 0x20;
	} else
	    dest[pos++] = c;
    }
    /* tack on the end-flag */
    if (pos < sizeof(dest))
	dest[pos++] = HDLC_FLAG;
    /* now write this packet */
    return write(fd, dest, pos);
}

/*** decaps_gre ***************************************************************/
static int decaps_gre (int fd, callback_t callback, int cl)
{
    unsigned char buffer[PACKET_MAX + 64 /*ip header*/];
    struct pptp_gre_header *header;
    int status, ip_len = 0;
    unsigned int headersize;
    unsigned int payload_len;
    u_int32_t seq;

    if ((status = read (fd, buffer, sizeof(buffer))) <= 0) {
	warn("short read (%d): %m", status);
	stats.rx_errors++;
	return -1;
    }
    /* strip off IP header, if present */
    if ((buffer[0] & 0xF0) == 0x40)
	ip_len = (buffer[0] & 0xF) * 4;
    header = (struct pptp_gre_header *)(buffer + ip_len);
    /* verify packet (else discard) */
    if (    /* version should be 1 */
	    ((ntoh8(header->ver) & 0x7F) != PPTP_GRE_VER) ||
	    /* PPTP-GRE protocol for PPTP */
	    (ntoh16(header->protocol) != PPTP_GRE_PROTO)||
	    /* flag C should be clear   */
	    PPTP_GRE_IS_C(ntoh8(header->flags)) ||
	    /* flag R should be clear   */
	    PPTP_GRE_IS_R(ntoh8(header->flags)) ||
	    /* flag K should be set     */
	    (!PPTP_GRE_IS_K(ntoh8(header->flags))) ||
	    /* routing and recursion ctrl = 0  */
	    ((ntoh8(header->flags)&0xF) != 0)) {
	/* if invalid, discard this packet */
	warn("Discarding GRE: %X %X %X %X %X %X",
		ntoh8(header->ver)&0x7F, ntoh16(header->protocol),
		PPTP_GRE_IS_C(ntoh8(header->flags)),
		PPTP_GRE_IS_R(ntoh8(header->flags)),
		PPTP_GRE_IS_K(ntoh8(header->flags)),
		ntoh8(header->flags) & 0xF);
	stats.rx_invalid++;
	return 0;
    }
    if (PPTP_GRE_IS_A(ntoh8(header->ver))) { /* acknowledgement present */
	u_int32_t ack = (PPTP_GRE_IS_S(ntoh8(header->flags)))?
	    header->ack:header->seq; /* ack in different place if S = 0 */
	ack = ntoh32( ack);
	if (ack > ack_recv) ack_recv = ack;
	/* also handle sequence number wrap-around  */
	if (WRAPPED(ack,ack_recv)) ack_recv = ack;
	if (ack_recv == stats.pt.seq) {
	    int rtt = time_now_usecs() - stats.pt.time;
	    stats.rtt = (stats.rtt + rtt) / 2;
	}
    }
    /* test if payload present */
    if (!PPTP_GRE_IS_S(ntoh8(header->flags)))
	return 0; /* ack, but no payload */
    headersize  = sizeof(*header);
    payload_len = ntoh16(header->payload_len);
    seq         = ntoh32(header->seq);
    /* no ack present? */
    if (!PPTP_GRE_IS_A(ntoh8(header->ver))) headersize -= sizeof(header->ack);
    /* check for incomplete packet (length smaller than expected) */
    if (status - headersize < payload_len) {
	warn("discarding truncated packet (expected %d, got %d bytes)",
		payload_len, status - headersize);
	stats.rx_truncated++;
	return 0;
    }
    /* check for expected sequence number */
    if ( !decaps_gre_no_first || (seq == seq_recv + 1)) { /* wrap-around safe */
	if ( log_level >= 2 )
	    logF("accepting packet %d", seq);
	stats.rx_accepted++;
	decaps_gre_no_first = 1;
	seq_recv = seq;
	return callback(cl, buffer + ip_len + headersize, payload_len);
    /* out of order, check if the number is too low and discard thepacket.
     * (handle sequence number wrap-around, and try to do it right) */
    } else if ( seq < seq_recv + 1 || WRAPPED(seq_recv, seq) ) {
	if ( log_level >= 1 )
	    logF("discarding duplicate or old packet %d (expecting %d)",
		seq, seq_recv + 1);
	stats.rx_underwin++;
    /* sequence number too high, is it reasonably close? */
    } else if ( seq < seq_recv + MISSING_WINDOW ||
		WRAPPED(seq, seq_recv + MISSING_WINDOW) ) {
	if ( log_level >= 1 )
	    logF("buffering packet %d (expecting %d, lost or reordered)",
		seq, seq_recv+1);
	pqueue_add(seq, buffer + ip_len + headersize, payload_len);
	stats.rx_buffered++;
    /* no, packet must be discarded */
    } else {
	if ( log_level >= 1 )
	    warn("discarding bogus packet %d (expecting %d)",
		 seq, seq_recv + 1);
	stats.rx_overwin++;
    }
    return 0;
}

/*** dequeue_gre **************************************************************/
static int dequeue_gre (callback_t callback, int cl)
{
    pqueue_t *head;
    int status;
    /* process packets in the queue that either are expected or have
     * timed out. */
    head = pqueue_head();
    while ( head != NULL &&
	    ( (head->seq == seq_recv + 1) || /* wrap-around safe */
	      (pqueue_expiry_time(head) <= 0)
	    )
	  ) {
	/* if it is timed out... */
	if (head->seq != seq_recv + 1 ) {  /* wrap-around safe */
	    stats.rx_lost += head->seq - seq_recv - 1;
	    if (log_level >= 2)
		logF("timeout waiting for %d packets", head->seq - seq_recv - 1);
	}
	if (log_level >= 2)
	    logF("accepting %d from queue", head->seq);
	seq_recv = head->seq;
	status = callback(cl, head->packet, head->packlen);
	pqueue_del(head);
	if (status < 0)
	    return status;
	head = pqueue_head();
    }
    return 0;
}

/*** encaps_gre ***************************************************************/
static int encaps_gre (int fd, void *pack, unsigned int len)
{
    union {
	struct pptp_gre_header header;
	unsigned char buffer[PACKET_MAX + sizeof(struct pptp_gre_header)];
    } u;
    unsigned int header_len;
    int rc;
    /* package this up in a GRE shell. */
    u.header.flags       = hton8 (PPTP_GRE_FLAG_K);
    u.header.ver         = hton8 (PPTP_GRE_VER);
    u.header.protocol    = hton16(PPTP_GRE_PROTO);
    u.header.payload_len = hton16(len);
    u.header.call_id     = hton16(pptp_gre_peer_call_id);
    /* special case ACK with no payload */
    if (pack == NULL) {
	if (ack_sent != seq_recv) {
	    u.header.ver |= hton8(PPTP_GRE_FLAG_A);
	    u.header.payload_len = hton16(0);
	    /* ack is in odd place because S == 0 */
	    u.header.seq = hton32(seq_recv);
	    ack_sent = seq_recv;
	    rc = write(fd, &u.header, sizeof(u.header) - sizeof(u.header.seq));
	    if (rc < 0) {
		stats.tx_failed++;
	    } else if (rc < sizeof(u.header) - sizeof(u.header.seq)) {
		stats.tx_short++;
	    } else {
		stats.tx_acks++;
	    }
	    return rc;
	} else return 0; /* we don't need to send ACK */
    } /* explicit brace to avoid ambiguous `else' warning */
    /* send packet with payload */
    u.header.flags |= hton8(PPTP_GRE_FLAG_S);
    u.header.seq    = hton32(encaps_gre_seq);
    if (ack_sent != seq_recv) { /* send ack with this message */
	u.header.ver |= hton8(PPTP_GRE_FLAG_A);
	u.header.ack  = hton32(seq_recv);
	ack_sent = seq_recv;
	header_len = sizeof(u.header);
    } else { /* don't send ack */
	header_len = sizeof(u.header) - sizeof(u.header.ack);
    }
    if (header_len + len >= sizeof(u.buffer)) {
	stats.tx_oversize++;
	return 0; /* drop this, it's too big */
    }
    /* copy payload into buffer */
    memcpy(u.buffer + header_len, pack, len);
    /* record and increment sequence numbers */
    seq_sent = encaps_gre_seq; encaps_gre_seq++;
    /* write this baby out to the net */
    rc = write(fd, u.buffer, header_len + len);
    if (rc < 0) {
	stats.tx_failed++;
    } else if (rc < header_len + len) {
	stats.tx_short++;
    } else {
	stats.tx_sent++;
	stats.pt.seq  = seq_sent;
	stats.pt.time = time_now_usecs();
    }
    return rc;
}


#ifdef SUPPORT_BEZEQ_ISRAEL
/* pptp_quirks.c ...... various options to fix quirks found in buggy adsl modems
 *                      mulix <mulix@actcom.co.il>
 */

static int fixups_sz = sizeof(pptp_fixups)/sizeof(pptp_fixups[0]);

/* return 0 on success, non 0 otherwise */
static int set_quirk_index(int index)
{
    if (index >= 0 && index < fixups_sz) {
	quirk_index = index;
	return 0;
    }

    return -1;
}


/* return the index for this isp in the quirks table, -1 if not found */
static int find_quirk(const char* isp_name)
{
    int i = 0;
    if (isp_name) {
	while (i < fixups_sz && pptp_fixups[i].isp) {
	    if (!strcmp(pptp_fixups[i].isp, isp_name)) {
		return i;
	    }
	    ++i;
	}
    }

    return -1;
}
#endif /* SUPPORT_BEZEQ_ISRAEL */


#define MIN_CAPACITY 128 /* min allocated buffer for a packet */

static int pqueue_alloc (int seq, unsigned char *packet, int packlen, pqueue_t **new);



static int
pqueue_alloc(int seq, unsigned char *packet, int packlen, pqueue_t **new)
{
  pqueue_t *newent;

  DEBUG_CMD(logF("seq=%d, packlen=%d", seq, packlen););

  /* search the freelist for one that has sufficient space  */
  if (pq_freelist_head) {

    for (newent = pq_freelist_head; newent; newent = newent->next) {

      if (newent->capacity >= packlen) {

	/* unlink from freelist */
	if (pq_freelist_head == newent)
	  pq_freelist_head = newent->next;

	if (newent->prev)
	  newent->prev->next = newent->next;

	if (newent->next)
	  newent->next->prev = newent->prev;

	if (pq_freelist_head)
	  pq_freelist_head->prev = NULL;

	break;
      } /* end if capacity >= packlen */
    } /* end for */

    /* nothing found? Take first and reallocate it */
    if (NULL == newent) {

      newent = pq_freelist_head;
      pq_freelist_head = pq_freelist_head->next;

      if (pq_freelist_head)
	pq_freelist_head->prev = NULL;

      DEBUG_CMD(logF("realloc capacity %d to %d",newent->capacity, packlen););

      newent->packet = (unsigned char *)realloc(newent->packet, packlen);

      if (!newent->packet) {
	warn("error reallocating packet: %m");
	return -1;
      }
      newent->capacity = packlen;
    }

    DEBUG_CMD(logF("Recycle entry from freelist. Capacity: %d", newent->capacity););

  } else {

    /* allocate a new one */
    newent = malloc( sizeof(pqueue_t) );
    if (!newent) {
      warn("error allocating newent: %m");
      return -1;
    }
    newent->capacity = 0;

    DEBUG_CMD(logF("Alloc new queue entry"););
  }

  if ( ! newent->capacity  ) {
    /* a new queue entry was allocated. Allocate the packet buffer */
    int size = packlen < MIN_CAPACITY ? MIN_CAPACITY : packlen;
    /* Allocate at least MIN_CAPACITY */
    DEBUG_CMD(logF("allocating for packet size %d", size););
    newent->packet = malloc(size);
    if (!newent->packet) {
      warn("error allocating packet: %m");
      return -1;
    }
    newent->capacity = size;
  } /* endif ! capacity */
  assert( newent->capacity >= packlen );
  /* store the contents into the buffer */
  memcpy(newent->packet, packet, packlen);

  newent->next = newent->prev = NULL;
  newent->seq     = seq;
  newent->packlen = packlen;

  gettimeofday(&newent->expires, NULL);
  newent->expires.tv_usec += packet_timeout_usecs;
  newent->expires.tv_sec  += (newent->expires.tv_usec / 1000000);
  newent->expires.tv_usec %= 1000000;

  *new = newent;
  return 0;
}


static int pqueue_add (int seq, unsigned char *packet, int packlen) {
  pqueue_t *newent, *point;

  /* get a new entry */
  if ( 0 != pqueue_alloc(seq, packet, packlen, &newent) ) {
    return -1;
  }

  for (point = pq_head; point != NULL; point = point->next) {
    if (point->seq == seq) {
      // queue already contains this packet
      warn("discarding duplicate packet %d", seq);
      return -1;
    }
    if (point->seq > seq) {
      // gone too far: point->seq > seq and point->prev->seq < seq
      if (point->prev) {
	// insert between point->prev and point
	DEBUG_CMD(logF("adding %d between %d and %d",
		      seq, point->prev->seq, point->seq););

	point->prev->next = newent;
      } else {
	// insert at head of queue, before point
	DEBUG_CMD(logF("adding %d before %d", seq, point->seq););
	pq_head = newent;
      }
      newent->prev = point->prev; // will be NULL, at head of queue
      newent->next = point;
      point->prev = newent;
      return 0;
    }
  }

  /* We didn't find anywhere to insert the packet,
   * so there are no packets in the queue with higher sequences than this one,
   * so all the packets in the queue have lower sequences,
   * so this packet belongs at the end of the queue (which might be empty)
   */

  if (pq_head == NULL) {
    DEBUG_CMD(logF("adding %d to empty queue", seq););
    pq_head = newent;
  } else {
    DEBUG_CMD(logF("adding %d as tail, after %d", seq, pq_tail->seq););
    pq_tail->next = newent;
  }
  newent->prev = pq_tail;
  pq_tail = newent;

  return 0;
}


static int pqueue_del (pqueue_t *pq_point) {

  DEBUG_CMD(logF("Move seq %d to freelist", pq_point->seq););

  /* unlink from pq */
  if (pq_head == pq_point) pq_head = pq_point->next;
  if (pq_tail == pq_point) pq_tail = pq_point->prev;
  if (pq_point->prev) pq_point->prev->next = pq_point->next;
  if (pq_point->next) pq_point->next->prev = pq_point->prev;

  /* add pq_point to the freelist */
  pq_point->next = pq_freelist_head;
  pq_point->prev = NULL;

  if (pq_point->next)
    pq_point->next->prev = pq_point;
  pq_freelist_head = pq_point;

  DEBUG_CMD(
    int pq_count = 0;
    int pq_freelist_count = 0;
    pqueue_t *p;
    for ( p = pq_head; p ; p = p->next) {
      ++pq_count;
    }

    for ( p = pq_freelist_head; p ; p = p->next) {
      ++pq_freelist_count;
    }
    logF("queue length is %d, freelist length is %d", pq_count, pq_freelist_count);
    );

  return 0;
}


static int pqueue_expiry_time (pqueue_t *entry) {
  struct timeval tv;
  int expiry_time;

  gettimeofday(&tv, NULL);
  expiry_time  = (entry->expires.tv_sec  - tv.tv_sec) * 1000000;
  expiry_time += (entry->expires.tv_usec - tv.tv_usec);
  return expiry_time;
}

/* util.c ....... error message utilities. */
#ifdef DEBUG
#define MAKE_STRING(label)                              \
va_list ap;                                             \
char buf[256], string[256];                             \
va_start(ap, format);                                   \
vsnprintf(buf, sizeof(buf), format, ap);                \
snprintf(string, sizeof(string), "%s %s[%s:%d]: %s",    \
	 log_string, label, func, line, buf);           \
va_end(ap)
#else
#define MAKE_STRING(label)                              \
va_list ap;                                             \
char buf[256], string[256];                             \
va_start(ap, format);                                   \
vsnprintf(buf, sizeof(buf), format, ap);                \
snprintf(string, sizeof(string), "%s: %s", log_string, buf);   \
va_end(ap)
#endif

/*** print a message to syslog ************************************************/
#ifdef DEBUG
static void pptp_log(const char *func, int line, char *format, ...)
#else
static void logF(const char *format, ...)
#endif
{
    MAKE_STRING("log");
    log_syslog = LOG_NOTICE;
    bb_error_msg("%s", string);
    log_syslog = LOG_ERR;
}

/*** print a warning to syslog ************************************************/
#ifdef DEBUG
static void pptp_warn(const char *func, int line, char *format, ...)
#else
static void warn(const char *format, ...)
#endif
{
    MAKE_STRING("warn");
    xlog_warn("%s", string);
}

/*** print a fatal warning to syslog and exit *********************************/
#ifdef DEBUG
static void pptp_fatal(const char *func, int line, char *format, ...)
#else
static void fatal(const char *format, ...)
#endif
{
    MAKE_STRING("fatal");
    log_syslog = LOG_CRIT;
    bb_error_msg_and_die("%s", string);
}

#if ENABLE_FEATURE_LONG_OPTIONS
/* structure with all recognised options for pptp */
static const struct option long_options[] = {
	    {"phone", 1, 0, 'p'},
	    {"nolaunchpppd", 0, 0, 'n'},
	    {"debug", 0, 0, 'd'},
	    {"sync", 0, 0, 's'},
	    {"timeout", 1, 0, 't'},
	    {"logstring", 1, 0, 'l'},
	    {"localbind", 1, 0, 'b'},
#ifdef SUPPORT_BEZEQ_ISRAEL
	    {"quirks", 1, 0, 'q'},
#endif
#ifdef CONFIG_FEATURE_VERBOSITY_PPTP
	    {"loglevel", 1, 0, 'v'},
#endif
	    {0, 0, 0, 0}
};
#endif /* long option */

int pptp_main(void)
{
    struct in_addr inetaddr;
    volatile int callmgr_sock = -1;

    int pty_fd, tty_fd, gre_fd, rc;
    volatile pid_t parent_pid, child_pid;
    u_int16_t call_id, peer_call_id;
    int pppdargc;
    char *phonenr = NULL;
    char *stimeout;
    char *slocalbind;
#ifdef SUPPORT_BEZEQ_ISRAEL
    char *quirks;
#endif
    int launchpppd = 1, debug = 0;

    /* set data to global bss */
    Gargv0 = Gargv[0];
    nlogecho = 10;
    packet_timeout_usecs = DEFAULT_PACKET_TIMEOUT * 1000000;
    USE_FEATURE_VERBOSITY_PPTP(log_level = 1;)
    log_string = "anon";
    encaps_gre_seq = 1;
    localbind.s_addr = INADDR_NONE;

#if ENABLE_FEATURE_LONG_OPTIONS
    bb_applet_long_options = long_options;
#endif

    pppdargc = bb_xgetopt_ulflags("hp:t:l:b:"
#ifdef SUPPORT_BEZEQ_ISRAEL
				 "q:"
#endif
				USE_FEATURE_VERBOSITY_PPTP("v:\0:-v:")
    /* at least one argument is required */
#ifdef CONFIG_FEATURE_VERBOSITY_PPTP
				"-1"
#else
				"\0-1"
#endif
				, &phonenr, &stimeout,
				&log_string, &slocalbind
#ifdef SUPPORT_BEZEQ_ISRAEL
				, &quirks
#endif
				USE_FEATURE_VERBOSITY_PPTP(, &log_level)
				);
    bb_openlog(LOG_PID, LOG_DAEMON);
    if((pppdargc & 1)) {
	/* -h, do not support from original program */
	usage();
    }
    if((pppdargc & 2)) {
	/* --phone specified */
	/* copy it to a buffer, as the argv's will be overwritten
	 * by setproctitle() */
	phonenr = xstrdup(phonenr);
    }
    if((pppdargc & 4)) {
	/* --nolaunchpppd specified */
	launchpppd = 0;
    }
    debug = pppdargc & 8;       /* --debug */
    syncppp = pppdargc & 16;    /* --sync specified */
    if((pppdargc & 32)) {
	/* --timeout */
	float new_packet_timeout = atof(stimeout);
	if (new_packet_timeout < 0.0099 || new_packet_timeout > 10) {
		bb_default_error_retval = 2;
		bb_error_msg_and_die("Packet timeout %s (%f) out of range: "
				"should be between 0.01 and 10 seconds",
				stimeout, new_packet_timeout);
	} else {
		packet_timeout_usecs = new_packet_timeout * 1000000;
	}
    }
    if((pppdargc & 64)) {
	/* --logstring */
	/* strdup need! as the argv's will be overwritten by inststr() */
	log_string = xstrdup(log_string);
    }
    if((pppdargc & 128)) {
	/* --localbind */
	if (inet_pton(AF_INET, slocalbind, (void *) &localbind) < 1) {
		bb_default_error_retval = 2;
		bb_error_msg_and_die("Local bind address %s invalid", slocalbind);
	}
    }

#ifdef SUPPORT_BEZEQ_ISRAEL
    if((pppdargc & 256)) {
	/* --quirks specified */
	if (set_quirk_index(find_quirk(quirks)))
		usage();
    }
# define NEXT_OPT 512
#else
# define NEXT_OPT 256
#endif  /* SUPPORT_BEZEQ_ISRAEL */

#ifdef CONFIG_FEATURE_VERBOSITY_PPTP
	if((pppdargc & NEXT_OPT)) {
		/* --loglevel */
		if (log_level < 0 || log_level > 2)
			usage();
	}
#endif

    /* Get IP address for the hostname in argv[1] */
    inetaddr = get_ip_address(*Gargv);
    Gargv++;
    Gargc--;

    /* Now we have the peer address, bind the GRE socket early,
       before starting pppd. This prevents the ICMP Unreachable bug
       documented in <1026868263.2855.67.camel@jander> */
    gre_fd = pptp_gre_bind(inetaddr);
    if (gre_fd < 0)
	bb_error_msg_and_die("Cannot bind GRE socket, aborting.");
    log_syslog = LOG_ERR;
    log_stderr = 0;
    logF("The synchronous pptp option is %sactivated\n", syncppp ? "" : "NOT ");

    /* Find an open pty/tty pair. */
    if(launchpppd) {
	rc = openpty (&pty_fd, &tty_fd, ttydev, NULL, NULL);
	if (rc < 0) {
	    fatal("Could not find free pty.");
	}

	/* fork and wait. */
	bb_signal((1<<SIGUSR1), do_nothing); /* don't die */
	parent_pid = getpid();
	switch (child_pid = fork()) {
	    case -1:
		fatal("Could not fork pppd process");
	    case 0: /* I'm the child! */
		close (tty_fd);
		bb_signal((1<<SIGUSR1), SIG_DFL);
		child_pid = getpid();
		break;
	    default: /* parent */
		close (pty_fd);
		/*
		 * There is still a very small race condition here.  If a signal
		 * occurs after signaled is checked but before pause is called,
		 * things will hang.
		 */
		if (!signaled) {
		    pause(); /* wait for the signal */
		}
		launch_pppd(Gargc, Gargv); /* launch pppd */
	}
    } else { /* ! launchpppd */
	pty_fd = tty_fd = STDIN_FILENO;
	/* close unused file descriptor, that is redirected to the pty */
	close(STDOUT_FILENO);
	child_pid = getpid();
	parent_pid = 0; /* don't kill pppd */
    }

    do {
	/*
	 * Open connection to call manager (Launch call manager if necessary.)
	 */
	callmgr_sock = open_callmgr(inetaddr, phonenr, pty_fd);
	/* Exchange PIDs, get call ID */
    } while (get_call_id(callmgr_sock, parent_pid, child_pid,
		&call_id, &peer_call_id) < 0);

    /* Send signal to wake up pppd task */
    if (launchpppd) {
	kill(parent_pid, SIGUSR1);
	sleep(2);
	/* become a daemon */
	if (!debug && daemon(0, 0) != 0) {
	    bb_perror_msg("daemon");
	}
    } else {
	/* re-open stderr as /dev/null to release it */
	null2fd();
    }
    bb_setproctitle("pptp: GRE-to-PPP gateway on %s", ttyname(tty_fd));
    if (sigsetjmp(env, 1)!= 0) goto shutdown;

    bb_signal((1<<SIGINT)|(1<<SIGTERM)|(1<<SIGKILL)|(1<<SIGCHLD), sighandler);
    bb_signal(1<<SIGUSR1, sigstats);

    /* Do GRE copy until close. */
    pptp_gre_copy(call_id, peer_call_id, pty_fd, gre_fd);

shutdown:
    /* on close, kill all. */
    if(launchpppd)
	kill(parent_pid, SIGTERM);
    close(pty_fd);
    close(callmgr_sock);
    sleep(3);     /* give ctrl manager a chance to exit */
    return 0;
}
#endif /* APPLET_USAGE */
