/*
 * The Rdate command will ask a time server for the RFC 868 time
 *  and optionally set the system time.
 *
 * by Sterling Huxley <sterling@europa.com>
 *
 * Licensed under GPL v2 or later, see file License for details.
*/
#ifdef APPLET_USAGE
#define rdate_trivial_usage \
	"[-sp] HOST"
#define rdate_full_usage \
	"Get and possibly set the system date and time from a remote HOST.\n\n" \
	"Options:\n" \
	"\t-s\tSet the system date and time (default)\n" \
	"\t-p\tPrint the date and time"
#else

#include "busybox.h"
#include <sys/time.h>
#include <netinet/in.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "bb_sig.h"
#include "bb_inet.h"

static const int RFC_868_BIAS = 2208988800UL;

static void socket_timeout(int sig)
{
	bb_error_msg_and_die("timeout connecting to time server");
}

static time_t askremotedate(const char *host)
{
	unsigned long nett;
	struct sockaddr s;
	int fd;

	bb_gethostbyname(BB_GHBN_SOCKADDR | BB_GHBN_ERR_DIE, host, &s);
	((struct sockaddr_in *)&s)->sin_port = bb_lookup_tcp_port("time", 37);

	/* Add a timeout for dead or inaccessible servers */
	alarm(10);
	bb_signal((1 << SIGALRM), socket_timeout);

	fd = xconnect(&s);

	if (safe_read(fd, (void *)&nett, 4) != 4)    /* read time from server */
		bb_error_msg_and_die("%s did not send the complete time", host);
	close(fd);

	/* convert from network byte order to local byte order.
	 * RFC 868 time is the number of seconds
	 *  since 00:00 (midnight) 1 January 1900 GMT
	 *  the RFC 868 time 2,208,988,800 corresponds to 00:00  1 Jan 1970 GMT
	 * Subtract the RFC 868 time  to get Linux epoch
	 */

	return(ntohl(nett) - RFC_868_BIAS);
}

int rdate_main(void)
{
	time_t remote_time;
	int flags;

	flags = (int)bb_xgetopt_ulflags("sp\0=1");

	remote_time = askremotedate(*Gargv);

	if (flags & 1) {
		time_t current_time;

		time(&current_time);
		if (current_time == remote_time) {
			bb_error_msg("Current time matches remote time.");
		} else {
			struct timespec ts;
			memset(&ts, 0, sizeof(ts));
			ts.tv_sec = remote_time;
			if(clock_settime (CLOCK_REALTIME, &ts) < 0)
				bb_perror_msg_and_die("Could not set time of day");
		}

	/* No need to check for the -p flag as it's the only option left */

	} else printf("%s", ctime(&remote_time));

	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
