/*
 * Busybox port by Vladimir Oleynik (C) 2006-2008 <dzo@simtreas.ru>
 *
 * Copyright (c) 1983, 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * rlogin - remote login
 */
#ifdef APPLET_USAGE
#define rlogin_trivial_usage \
	"[-8E] [-e escapechar] [-l user] host"
#define rlogin_full_usage \
	"rlogin - remote login\n\n" \
	"Options:\n" \
	"\t-8\tallows an eight-bit input data\n" \
	"\t-E\tstops any character from being recognized as an\n" \
	"\t\tescape character.\n" \
	"\t-e escape\tallows user specification of the escape character."
#else

#include "busybox.h"
#include "bb_inet.h"
#include <sys/file.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <termios.h>
#include <setjmp.h>

#include <unistd.h>
#include <string.h>
#include <pty.h>

#include "speed_table.h"
#include "bb_sig.h"

/* This is just a warning of a common mistake.  Possibly this should be a
 * uclibc faq entry rather than in busybox... */
#if defined(__UCLIBC__) && ! defined(__UCLIBC_HAS_RPC__)
#error "You need to build uClibc with UCLIBC_HAS_RPC for rlogin build."
#endif
#if defined(__MUSL__)
#error "MUSL libc have not RPC support. We can not compile rlogin applet. :("
#endif

#ifndef TIOCPKT_WINDOW
#define TIOCPKT_WINDOW  0x80
#endif

#ifndef TIOCPKT_FLUSHWRITE
#define TIOCPKT_FLUSHWRITE 0x02
#define TIOCPKT_NOSTOP 0x10
#define TIOCPKT_DOSTOP 0x20
#endif

/* concession to Sun */
#ifndef SIGUSR1
#define SIGUSR1 30
#endif

CALC_BSS struct BB_COMMON_BSS
{
	struct termios defmodes;
	struct termios ixon_state;
	int eight, rem;

	u_char escapechar;

	int childpid;

	char defkill, defquit, defstart, defstop, defeol, defeof, defintr;
	char defsusp, defdsusp, defreprint, defdiscard, defwerase, deflnext;
	struct winsize win_size;
	int dosigwinch;
	sigjmp_buf rcvtop;
	int ppid, rcvcnt, rcvstate;
	char rcvbuf[8 * 1024];
} bb_common_bss;
#define defmodes bb_common_bss.defmodes
#define ixon_state bb_common_bss.ixon_state
#define eight bb_common_bss.eight
#define rem bb_common_bss.rem
#define escapechar bb_common_bss.escapechar
#define childpid bb_common_bss.childpid
#define defkill bb_common_bss.defkill
#define defquit bb_common_bss.defquit
#define defstart bb_common_bss.defstart
#define defstop bb_common_bss.defstop
#define defeol bb_common_bss.defeol
#define defeof bb_common_bss.defeof
#define defintr bb_common_bss.defintr
#define defsusp bb_common_bss.defsusp
#define defdsusp bb_common_bss.defdsusp
#define defreprint bb_common_bss.defreprint
#define defdiscard bb_common_bss.defdiscard
#define defwerase bb_common_bss.defwerase
#define deflnext bb_common_bss.deflnext
#define win_size bb_common_bss.win_size
#define dosigwinch bb_common_bss.dosigwinch
#define rcvtop bb_common_bss.rcvtop
#define ppid bb_common_bss.ppid
#define rcvcnt bb_common_bss.rcvcnt
#define rcvstate bb_common_bss.rcvstate
#define rcvbuf bb_common_bss.rcvbuf

#define noescape (bss_flags & 2)

#define get_window_size(fd, wp) ioctl(fd, TIOCGWINSZ, wp)


static void
msg (const char *str)
{
  bb_error_msg ("%s\r\n", str);
}

static u_char
getescape (const char *p)
{
  long val;
  int len;

  if(p == NULL)
    return '~';
  if ((len = strlen (p)) == 1)  /* use any single char, including '\' */
	return (u_char) * p;
  /* otherwise, \nnn */
  if (*p == '\\' && len >= 2 && len <= 4) {
	val = strtol (++p, NULL, 8);
	for (;;) {
	  if (!*++p)
		return (u_char) val;
	  if (*p < '0' || *p > '8')
		break;
	}
  }
  msg ("illegal option value -- e");
  bb_show_usage ();
  /* NOTREACHED */
  return 0;
}


static void
mode (int f)
{
  struct termios tios;
  tcgetattr (0, &tios);

  switch (f) {
  case 0:
	/*
	 * remember whether IXON was set, so it can be restored
	 * when mode(1) is next done
	 */
	tcgetattr (0, &ixon_state);
	/*
	 * copy the initial modes we saved into sb; this is
	 * for restoring to the initial state
	 */
	memcpy (&tios, &defmodes, sizeof (defmodes));
	break;
  case 1:
	/* turn off output mappings */
	tios.c_oflag &= ~(ONLCR | OCRNL);
	/*
	 * turn off canonical processing and character echo;
	 * also turn off signal checking -- ICANON might be
	 * enough to do this, but we're being careful
	 */
	tios.c_lflag &= ~(ECHO | ICANON | ISIG);
	tios.c_iflag &= ~(ICRNL);
	tios.c_cc[VTIME] = 1;
	tios.c_cc[VMIN] = 1;
	if (eight)
	  tios.c_iflag &= ~(ISTRIP);
	/* preserve tab delays, but turn off tab-to-space expansion */
	if ((tios.c_oflag & TABDLY) == TAB3)
	  tios.c_oflag &= ~TAB3;
	/*
	 *  restore current flow control state
	 */
	if ((ixon_state.c_iflag & IXON) && !eight) {
	  tios.c_iflag |= IXON;
	} else {
	  tios.c_iflag &= ~IXON;
	}
	tios.c_cc[VSUSP] = 255;
	tios.c_cc[VEOL] = 255;
	tios.c_cc[VREPRINT] = 255;
	tios.c_cc[VDISCARD] = 255;
	tios.c_cc[VWERASE] = 255;
	tios.c_cc[VLNEXT] = 255;
	tios.c_cc[VEOL2] = 255;
	break;
  default:
	return;
  }
  tcsetattr (0, TCSADRAIN, &tios);
}

static void done (int status) __attribute__ ((noreturn));

static void
done (int status)
{
  int w, wstatus;

  mode (0);
  if (childpid > 0) {
	/* make sure catch_child does not snap it up */
	bb_signal((1 << SIGCHLD), SIG_DFL);
	if (kill (childpid, SIGKILL) >= 0)
	  while ((w = wait (&wstatus)) > 0 && w != childpid);
  }
  exit (status);
}

/* copy SIGURGs to the child process. */
static void ATTRIBUTE_UNUSED
copytochild (int ignore)
{
  kill (childpid, SIGURG);
}


/* trap a signal, unless it is being ignored. */
static void
setsignal (int sig, void (*act) (int))
{
  sigset_t nmask;
  int sigm = 1 << sig;

  nmask = bb_sigblock(sigm);

  if (bb_signal(sigm, act) == SIG_IGN)
	bb_signal(sigm, SIG_IGN);
  bb_sigunblock(nmask);
}


/*
 * Send the window size to the server via the magic escape
 */
static void
sendwindow (void)
{
  struct winsize *wp;
  char obuf[4 + sizeof (struct winsize)];

  wp = (struct winsize *) (obuf + 4);
  obuf[0] = (char) 0377;    /* (char) casts added for 32/64 bit machines */
  obuf[1] = (char) 0377;
  obuf[2] = (char) 's';
  obuf[3] = (char) 's';
  wp->ws_row = htons (win_size.ws_row);
  wp->ws_col = htons (win_size.ws_col);
  wp->ws_xpixel = htons (win_size.ws_xpixel);
  wp->ws_ypixel = htons (win_size.ws_ypixel);

  write (rem, obuf, sizeof (obuf));
}

static void ATTRIBUTE_UNUSED
sigwinch (int signum)
{
  struct winsize ws;

  if (dosigwinch && get_window_size (0, &ws) == 0 &&
	  memcmp (&ws, &win_size, sizeof (ws))) {
	win_size = ws;
	sendwindow ();
  }
}

/*
 * This is called when the reader process gets the out-of-band (urgent)
 * request to turn on the window-changing protocol.
 */
static void ATTRIBUTE_UNUSED
writeroob (int ignore)
{
  if (dosigwinch == 0) {
	sendwindow ();
	bb_signal((1 << SIGWINCH), sigwinch);
  }
  dosigwinch = 1;
}

static void ATTRIBUTE_UNUSED
catch_child (int ignore)
{
  int status;
  int pid;

  for (;;) {
	pid = wait3 (&status, WNOHANG | WUNTRACED, (struct rusage *) 0);
	if (pid == 0)
	  return;
	/* if the child (reader) dies, just quit */
	if (pid < 0 || (pid == childpid && !WIFSTOPPED (status)))
	  done ((int) (WTERMSIG(status) | WEXITSTATUS(status)));
  }
  /* NOTREACHED */
}

static void
echo (char c)
{
  char *p;
  char buf[8];

  p = buf;
  c &= 0177;
  *p++ = escapechar;
  if (c < ' ') {
	*p++ = '^';
	*p++ = c + '@';
  } else if (c == 0177) {
	*p++ = '^';
	*p++ = '?';
  } else
	*p++ = c;
  *p++ = '\r';
  *p++ = '\n';
  write (STDOUT_FILENO, buf, p - buf);
}

static void
stop (char cmdc)
{
  mode (0);
  bb_signal((1 << SIGCHLD), SIG_IGN);
  kill (cmdc == defsusp ? 0 : getpid (), SIGTSTP);
  bb_signal((1 << SIGCHLD), catch_child);
  mode (1);
  sigwinch (SIGWINCH);                  /* check for size changes */
}

/*
 * writer: write to remote: 0 -> line.
 * ~.                           terminate
 * ~^Z                          suspend rlogin process.
 * ~<delayed-suspend char>      suspend rlogin process, but leave reader alone.
 */
static void
writer (void)
{
  int bol, local, n;
  char c;

  bol = 1;      /* beginning of line */
  local = 0;
  for (;;) {
	n = read (STDIN_FILENO, &c, 1);
	if (n <= 0) {
	  if (n < 0 && errno == EINTR)
		continue;
	  break;
	}
	/*
	 * If we're at the beginning of the line and recognize a
	 * command character, then we echo locally.  Otherwise,
	 * characters are echo'd remotely.  If the command character
	 * is doubled, this acts as a force and local echo is
	 * suppressed.
	 */
	if (bol) {
	  bol = 0;
	  if (noescape == 0 && c == escapechar) {
		local = 1;
		continue;
	  }
	} else if (local) {
	  local = 0;
	  if (c == '.' || c == defeof) {
		echo (c);
		break;
	  }
	  if (c == defsusp || c == defdsusp) {
		bol = 1;
		echo (c);
		stop (c);
		continue;
	  }
	  if (c != escapechar)
		write (rem, &escapechar, 1);
	}

	if (write (rem, &c, 1) == 0) {
	  msg ("line gone");
	  break;
	}
	bol = c == defkill || c == defeof ||
	  c == defintr || c == defsusp || c == '\r' || c == '\n';
  }
}

/*
 * reader: read from remote: line -> 1
 */
#define READING 1
#define WRITING 2

static void ATTRIBUTE_UNUSED
oob_real (int ignore)
{
  struct termios tios;

  int atmark, n, rcvd;
  unsigned char waste[BUFSIZ], mark;

  rcvd = 0;
  while (recv (rem, &mark, 1, MSG_OOB) < 0) {
	sleep (1);
	switch (errno) {
	case EWOULDBLOCK:
	  /*
	   * Urgent data not here yet.  It may not be possible
	   * to send it yet if we are blocked for output and
	   * our input buffer is full.
	   */
	  if (rcvcnt < (int) sizeof (rcvbuf)) {
		n = read (rem, rcvbuf + rcvcnt, sizeof (rcvbuf) - rcvcnt);
		if (n <= 0)
		  return;
		rcvd += n;
	  } else {
		n = read (rem, waste, sizeof (waste));
		if (n <= 0)
		  return;
	  }
	  continue;
	default:
	  return;
	}
  }
  if (mark & TIOCPKT_WINDOW) {
	/* Let server know about window size changes */
	kill (ppid, SIGUSR1);
  }
  if (!eight && (mark & TIOCPKT_NOSTOP)) {
	tcgetattr (0, &tios);
	tios.c_iflag &= ~IXON;
/*              tios.c_lflag &= ~ICANON;        */
	tcsetattr (0, TCSADRAIN, &tios);
  }
  if (!eight && (mark & TIOCPKT_DOSTOP)) {
	tcgetattr (0, &tios);
/*              tios.c_lflag  |= ICANON;*/
	tios.c_iflag |= IXON;
	tcsetattr (0, TCSADRAIN, &tios);
  }
  if (mark & TIOCPKT_FLUSHWRITE) {
	tcflush (1, TCOFLUSH);

	for (;;) {
	  if (xioctl (rem, SIOCATMARK, &atmark, "ioctl") < 0)
		break;
	  if (atmark)
		break;
	  n = read (rem, waste, sizeof (waste));
	  if (n <= 0)
		break;
	}
	/*
	 * Don't want any pending data to be output, so clear the recv
	 * buffer.  If we were hanging on a write when interrupted,
	 * don't want it to restart.  If we were reading, restart
	 * anyway.
	 */
	rcvcnt = 0;
	siglongjmp (rcvtop, 1);
  }

  /* oob_real does not do FLUSHREAD (alas!) */

  /*
   * If we filled the receive buffer while a read was pending, longjmp
   * to the top to restart appropriately.  Don't abort a pending write,
   * however, or we won't know how much was written.
   */
  if (rcvd && rcvstate == READING) {
	siglongjmp (rcvtop, 1);
  }
}

/* reader: read from remote: line -> 1 */
static int
reader (sigset_t omask)
{
  int pid = getpid ();
  int n, remaining;
  char *volatile bufp = rcvbuf;

  bb_signal((1 << SIGTTOU), SIG_IGN);
  bb_signal((1 << SIGURG), oob_real);
  ppid = getppid ();
/*      fcntl(rem, F_SETOWN, pid); */
  ioctl (rem, SIOCSPGRP, &pid); /* @@@ */
  sigsetjmp (rcvtop, 1);
  bb_sigunblock(omask);
  for (;;) {
	while ((remaining = rcvcnt - (bufp - rcvbuf)) > 0) {
	  rcvstate = WRITING;
	  n = write (STDOUT_FILENO, bufp, remaining);
	  if (n < 0) {
		if (errno != EINTR)
		  return -1;
		continue;
	  }
	  bufp += n;
	}
	bufp = rcvbuf;
	rcvcnt = 0;
	rcvstate = READING;

	rcvcnt = read (rem, rcvbuf, sizeof (rcvbuf));

/*
* If we get a EIO from a read then it may mean that we have unread ungent data
* waiting that is getting in the way. We probably have got more then one lot of
* urgent data but we only got one SIGURG due to a problem in the kernel tcp.
* We can try and fix this by sending ourself a SIGURG and pretending the error
* never occurred. This might be a problem if we really _should_ be getting
* a EIO for some unrelated reason. (AJT 3/93)
*
* Hmm, I just checked this with the current (NET-2e BETA-1) kernel, and
* it seems that this patch isn't needed anymore.  FvK 09/20/93
*/
#ifdef XX_SUN_KLUDGE
	if (rcvcnt < 0 && errno == EIO) {
	  errno = 0;
	  kill (getpid (), SIGURG);
	  continue;
	}
#endif
	if (rcvcnt == 0)
	  return 0;
	if (rcvcnt < 0) {
	  if (errno == EINTR)
		continue;
	  bb_perror_msg ("read");
	  return -1;
	}
  }
}

static void ATTRIBUTE_UNUSED
lostpeer (int ignore)
{
  bb_signal((1 << SIGPIPE), SIG_IGN);
  msg ("\007connection closed.");
  done (1);
}

static void doit (sigset_t omask) __attribute__ ((noreturn));

static void
doit (sigset_t omask)
{
  struct termios tios;

  tcgetattr (0, &tios);

  tios.c_cc[VMIN] = 1;
  tios.c_cc[VTIME] = 1;

  defkill = tios.c_cc[VKILL];
  defquit = tios.c_cc[VQUIT];
  defstart = tios.c_cc[VSTART];
  defstop = tios.c_cc[VSTOP];
  defeol = tios.c_cc[VEOL];
  defeof = tios.c_cc[VEOF];
  defintr = tios.c_cc[VINTR];
  defsusp = tios.c_cc[VSUSP];   /* stop process */
#ifdef VDSUSP
  defdsusp = tios.c_cc[VDSUSP]; /* delayed stop process */
#else
  defdsusp = (char) 0xFF;               /* cast 0xFF for 32/64 bit platforms */
#endif
  defreprint = tios.c_cc[VREPRINT];     /* rprint line */
  defdiscard = tios.c_cc[VDISCARD];     /* flush output */
  defwerase = tios.c_cc[VWERASE];       /* word erase */
  deflnext = tios.c_cc[VLNEXT]; /* literal next char */

  bb_signal((1 << SIGINT), SIG_IGN);
  setsignal (SIGHUP, exit);
  setsignal (SIGQUIT, exit);
  /*
   * Do this *before* forking...
   */
  bb_signal((1 << SIGCHLD), catch_child);

  childpid = fork ();
  if (childpid == -1) {
	bb_perror_msg("fork");
	done (1);
  }
  if (childpid == 0) {
	mode (1);
	if (reader (omask) == 0) {
	  msg ("connection closed.");
	  exit (0);
	}
	sleep (1);
	msg ("\007connection closed.");
	exit (1);
  }

  /*
   * We may still own the socket, and may have a pending SIGURG (or might
   * receive one soon) that we really want to send to the reader.  Set a
   * trap that simply copies such signals to the child.
   */
  bb_signal((1 << SIGURG), copytochild);
  bb_signal((1 << SIGUSR1), writeroob);
  bb_sigunblock(omask);
#ifdef  __linux__
  /*sleep(1); *//*  why?!? */
#endif
  writer ();
  msg ("closed connection.");
  done (0);
}

int rlogin_main(void)
{
  char *pw;
  struct termios tios;

  sigset_t omask;
  int one = 1, uid;
  char *host, *user = NULL, term[300];
  const char *t;
  char *s_escapechar = NULL;

  eight = bb_xgetopt_ulflags("8Ee:l:\0-1:?1", &s_escapechar, &user) & 1;
  escapechar = getescape (s_escapechar);
  host = *Gargv;

  pw = bb_getpwuid (NULL, uid = getuid (), -1);

  if (!user)
	user = pw;

  t = getenv ("TERM");
  if (!t)
	t = "network";
  if (tcgetattr (0, &tios) == 0) {
	speed_t speed = cfgetispeed (&tios);
	sprintf (term, "%.256s/%lu", t, bb_baud_to_value(speed));
  } else
	sprintf (term, "%.256s", t);

  clearenv();

  get_window_size (0, &win_size);

  /*
   * Moved before rcmd call so that if get a SIGPIPE in rcmd
   * we will have the defmodes set already.
   */
  tcgetattr (0, &defmodes);
  tcgetattr (0, &ixon_state);

  bb_signal((1 << SIGPIPE), lostpeer);
  /* will use SIGUSR1 for window size hack, so hold it off */
  omask = bb_sigblock( (1 << SIGURG) | (1 << SIGUSR1) );

  rem = rcmd (&host, bb_lookup_tcp_port ("login", 513), pw, user,
		term, 0);

  if (rem < 0)
	exit (1);

#ifdef IP_TOS
  one = IPTOS_LOWDELAY;
  if (setsockopt (rem, IPPROTO_IP, IP_TOS, (char *) &one, sizeof (one)) < 0)
	bb_perror_msg ("setsockopt(TOS)");
#endif
  if (setuid (uid)) {
	bb_perror_msg_and_die ("setuid");
  }

  doit (omask);
   /*NOTREACHED*/ return 0;
}
#endif /* APPLET_USAGE */
