/*
 * Busybox port by Vladimir Oleynik (C) 2009 <dzo@simtreas.ru>
 * Reduce code and stack usage, move the data segment to common bss+data
 *
 * Copyright (c) 1989 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
#ifdef APPLET_USAGE
#define rlogind_trivial_usage \
	"[-ahLln]"
#define rlogind_full_usage \
	"rlogind - remote login server\n\n" \
	"Options:\n" \
	"\t-a\task hostname for verifications\n" \
	"\t-h\tpermit use of superuser .rhosts file\n" \
	"\t-L\tprevent any authentication base on .rhosts or hosts.equiv\n" \
	"\t-l\tprevent any authentication base on .rhosts file\n" \
	"\t-n\tdisable keep-alive messages"
#else

#include "busybox.h"
#include "bb_inet.h"
#include <sys/ioctl.h>          /* for TIOCPKT */
#include <sys/time.h>           /* for FD_SET() et al. */
#include <netinet/in.h>
#include <syslog.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <termios.h>            /* for tcsetattr() */
#include <pty.h>
#ifdef CONFIG_FEATURE_UTMP
#include <utmp.h>
#endif
#include "bb_sig.h"

#if defined(__GLIBC__) && (__GLIBC__ >= 2)
#define _check_rhosts_file  __check_rhosts_file
#endif

/* This is just a warning of a common mistake.  Possibly this should be a
 * uclibc faq entry rather than in busybox... */
#if defined(__UCLIBC__) && ! defined(__UCLIBC_HAS_RPC__)
#error "You need to build uClibc with UCLIBC_HAS_RPC for rlogind build."
#endif
#if defined(__MUSL__)
#error "MUSL libc have not RPC support. We can not compile rlogind applet. :("
#endif

/*
 * Global flag variables
 */

CALC_BSS struct BB_COMMON_BSS {
	/* Do paranoid DNS confirmation on all hosts? */
	int check_all;
	/* Use TCP keepalive messages on connection? */
	int keepalive;
	/* Check ~/.rhosts? */
	int use_rhosts;
	/* Check ~root/.rhosts? */
	int allow_root_rhosts;
	/* Ignore all ~/.rhosts and /etc/hosts_equiv? */
	int deny_all_rhosts_hequiv;
	int confirmed;
	char oobdata[1];
	char line[MAXPATHLEN];
	struct winsize win;
	char lusername[32], rusername[32];
	char termtype[256];
	char pibuf[1024+1], fibuf[1024];
	int hostok;
} bb_common_bss;
#define check_all bb_common_bss.check_all
#define keepalive bb_common_bss.keepalive
#define use_rhosts bb_common_bss.use_rhosts
#define allow_root_rhosts bb_common_bss.allow_root_rhosts
#define deny_all_rhosts_hequiv bb_common_bss.deny_all_rhosts_hequiv
#define confirmed bb_common_bss.confirmed
#define oobdata bb_common_bss.oobdata
#define line bb_common_bss.line
#define win bb_common_bss.win
#define termtype bb_common_bss.termtype
#define lusername bb_common_bss.lusername
#define rusername bb_common_bss.rusername
#define pibuf bb_common_bss.pibuf
#define fibuf bb_common_bss.fibuf
#define hostok bb_common_bss.hostok

#define netf 0

/*
 * Standard rlogin processing...
 */

extern int _check_rhosts_file;


/*
 * This function must either die, return -1 on authentication failure,
 * or return 0 on authentication success. Dying is discouraged.
 */
static int auth_checkauth(const char *host)
{
    struct passwd *pwd;

    pwd = getpwnam(lusername);
    if (pwd == NULL) return -1;

    /*
     * The possibilities here are:
     *    user == root
     *          (1) .rhosts entry exists
     *          (2) hosts_equiv entry exists
     *          (3) neither .rhosts nor hosts_equiv entries exist
     *    user != root
     *          (4) .rhosts entry exists
     *          (5) hosts_equiv entry exists
     *          (6) neither .rhosts nor hosts_equiv entries exist
     *
     * ruserok() by itself will fail on (2), (3), and (6).
     *
     * Turning off use_rhosts will (or should) prevent (1) and (4).
     * Leaving allow_root_rhosts off will prevent (1).
     * Setting deny_all_rhosts_hequiv prevents all cases from succeeding.
     */

    if (deny_all_rhosts_hequiv) return -1;
    if (!allow_root_rhosts && pwd->pw_uid == 0) return -1;

    _check_rhosts_file = use_rhosts;

    return ruserok(host, pwd->pw_uid==0, rusername, lusername);
}


static const char *
topdomain(const char *h)
{
	const char *p;
	const char *maybe = NULL;
	int dots = 0;

	for (p = h + strlen(h); p >= h; p--) {
		if (*p == '.') {
			if (++dots == 2)
				return (p);
			maybe = p;
		}
	}
	return (maybe);
}

/*
 * Check whether host h is in our local domain,
 * defined as sharing the last two components of the domain part,
 * or the entire domain part if the local domain has only one component.
 * If either name is unqualified (contains no '.'),
 * assume that the host is local, as it will be
 * interpreted as such.
 */
static int
local_domain(const char *h)
{
	char localhost[MAXHOSTNAMELEN];
	const char *p1, *p2;

	localhost[0] = 0;
	(void) gethostname(localhost, sizeof(localhost));
	p1 = topdomain(localhost);
	p2 = topdomain(h);
	if (p1 == NULL || p2 == NULL || !strcasecmp(p1, p2))
		return(1);
	return(0);
}


static char *
find_hostname(const struct sockaddr_in *fromp)
{
	char *hop;
	char *hname;
	char hbuf[NI_MAXHOST];

	hop = bb_xgethostbyaddr(&fromp->sin_addr, hbuf, BB_XGETHOSTBYADDR_FLG_IN_ADDR);
	if (hop == NULL) {
		hname = xstrdup(inet_ntoa(fromp->sin_addr));
		hostok = 1;
	} else if (check_all || local_domain(hop)) {
		/*
		 * If name returned by gethostbyaddr is in our domain,
		 * attempt to verify that we haven't been fooled by someone
		 * in a remote net; look up the name and check that this
		 * address corresponds to the name.
		 */
		struct in_addr *addr_buf;
		int naddr;

		addr_buf = bb_gethostbyname(BB_GHBN_NADDR | BB_GHBN_SILENT | BB_GHBN_CANON_NAME,
						hop, &naddr, &hname);
		if (addr_buf != NULL) {
			int i;

			for (i = 0; i < naddr; i++) {
				if (!memcmp(addr_buf + i, &fromp->sin_addr, sizeof(fromp->sin_addr))) {
					hostok = 1;
					break;
				}
			}
			free(addr_buf);
		} else {
			hname = xstrdup(hop);
		}
	} else {
		hname = xstrdup(hop);
		hostok = 1;
	}

	return hname;
}

#ifndef TIOCPKT_WINDOW
#define TIOCPKT_WINDOW 0x80
#endif


static void __attribute__((noreturn))
fatal(int f, const char *msg, int syserr)
{
    char c;

    /*
     * Send out a binary one if we haven't sent the magic null as confirmation.
     */
    c = '\01';               /* error indicator */

    if (!confirmed) {
	write(netf, &c, 1);
	/*
	 * Still not confirmed, but we shouldn't ever get here again.
	 */
	confirmed = -1;
    }

    if (!syserr) {
	dprintf(f, "rlogind: %s.\r\n", msg);
    }
    else {
	dprintf(f, "rlogind: %s: %s.\r\n", msg, strerror(errno));
    }
    exit(1);
}


static char * network_init(void)
{
	struct sockaddr_in from, *fromp;
	socklen_t fromlen;
	char c;
	char *hname;
	int port;

	from.sin_family = AF_INET;
	fromlen = sizeof (from);
	if (getpeername(netf, (struct sockaddr *)&from, &fromlen) < 0) {
		bb_perror_msg("Can't get peer name of remote host");
		fatal(STDERR_FILENO, "Can't get peer name of remote host", 1);
	}
	if (keepalive &&
	    setsockopt_1i(netf, SOL_SOCKET, SO_KEEPALIVE, 1) < 0)
		xlog_warn("setsockopt (SO_KEEPALIVE): %m");
#ifdef IP_TOS
	#define IPTOS_LOWDELAY          0x10
	if (setsockopt_1i(netf, IPPROTO_IP, IP_TOS, IPTOS_LOWDELAY) < 0)
		xlog_warn("setsockopt (IP_TOS): %m");
#endif
	fromp = &from;

	alarm(60);
	read(netf, &c, 1);

	if (c != 0)
		exit(1);

	alarm(0);

	hname = find_hostname(fromp);

	port = ntohs(fromp->sin_port);
	if (fromp->sin_family != AF_INET ||
	    port >= IPPORT_RESERVED || port < IPPORT_RESERVED/2) {
	    syslog(LOG_NOTICE, "Connection from %s on illegal port",
		   inet_ntoa(fromp->sin_addr));
	    fatal(netf, "Permission denied", 0);
	}

#ifdef IP_OPTIONS
	{
	    u_char optbuf[BUFSIZ/3], *cp;
	    char lbuf[BUFSIZ];
	    int lboff;
	    socklen_t optsize = sizeof(optbuf);
	    int ipproto;
	    struct protoent *ip;

	    if ((ip = getprotobyname("ip")) != NULL)
		    ipproto = ip->p_proto;
	    else
		    ipproto = IPPROTO_IP;
	    if (getsockopt(0, ipproto, IP_OPTIONS, (char *)optbuf,
		&optsize) == 0 && optsize != 0) {
		    lboff=0;
		    for (cp = optbuf; optsize > 0; cp++, optsize--, lboff += 3)
			    snprintf(lbuf+lboff, sizeof(lbuf)-lboff,
				     " %2.2x", *cp);
		    syslog(LOG_NOTICE,
			"Connection received using IP options (ignored):%s",
			lbuf);
		    if (setsockopt(0, ipproto, IP_OPTIONS, NULL, optsize) != 0)
			    bb_perror_msg_and_die("setsockopt IP_OPTIONS NULL");
	    }
	}
#endif

	return hname;
}


/*
 * Handle a "control" request (signaled by magic being present)
 * in the data stream.  For now, we are only willing to handle
 * window size changes.
 */
static int control(int pty, char *cp, int n)
{
	if (n < 4+(int)sizeof(win) || cp[2] != 's' || cp[3] != 's') {
		return 0;
	}
	oobdata[0] &= ~TIOCPKT_WINDOW;  /* we know he heard */
	memcpy(&win, cp+4, sizeof(win));
	win.ws_row = ntohs(win.ws_row);
	win.ws_col = ntohs(win.ws_col);
	win.ws_xpixel = ntohs(win.ws_xpixel);
	win.ws_ypixel = ntohs(win.ws_ypixel);
	ioctl(pty, TIOCSWINSZ, &win);
	return 4+sizeof(win);
}

/* rlogin "protocol" machine. */
static void protocol(int p)
{
	static const char magic[2] = { (char)0377, (char)0377 };

	char *pbp = NULL, *fbp = NULL;
	int pcc = 0, fcc = 0;
	int cc, m;
	char cntl;

	/*
	 * Must ignore SIGTTOU, otherwise we'll stop
	 * when we try and set slave pty's window shape
	 * (our controlling tty is the master pty).
	 */
	(void) bb_signal((1 << SIGTTOU), SIG_IGN);
	send(netf, oobdata, 1, MSG_OOB);   /* indicate new rlogin */
	for (;;) {
		fd_set ibits, obits, ebits, *omask;

		FD_ZERO(&ebits);
		FD_ZERO(&ibits);
		FD_ZERO(&obits);
		omask = (fd_set *)NULL;
		if (fcc) {
			FD_SET(p, &obits);
			omask = &obits;
		} else
			FD_SET(netf, &ibits);
		if (pcc >= 0) {
			if (pcc) {
				FD_SET(netf, &obits);
				omask = &obits;
			} else {
				FD_SET(p, &ibits);
			}
		}
		FD_SET(p, &ebits);
		if ((m = select(p + 1, &ibits, omask, &ebits, 0)) < 0) {
			if (errno == EINTR)
				continue;
			fatal(netf, "select", 1);
		}
		if (m == 0) {
			/* shouldn't happen... */
			sleep(5);
			continue;
		}
#define pkcontrol(c)    ((c)&(TIOCPKT_FLUSHWRITE|TIOCPKT_NOSTOP|TIOCPKT_DOSTOP))
		if (FD_ISSET(p, &ebits)) {
			cc = read(p, &cntl, 1);
			if (cc == 1 && pkcontrol(cntl)) {
				cntl |= oobdata[0];
				send(netf, &cntl, 1, MSG_OOB);
				if (cntl & TIOCPKT_FLUSHWRITE) {
					pcc = 0;
					FD_CLR(p, &ibits);
				}
			}
		}
		if (FD_ISSET(netf, &ibits)) {
			fcc = read(netf, fibuf, sizeof(fibuf));
			if (fcc < 0 && errno == EWOULDBLOCK)
				fcc = 0;
			else {
				char *cp;
				int left, nn;

				if (fcc <= 0)
					break;
				fbp = fibuf;

			top:
				for (cp = fibuf; cp < fibuf+fcc-1; cp++)
					if (cp[0] == magic[0] &&
					    cp[1] == magic[1]) {
						left = fcc - (cp-fibuf);
						nn = control(p, cp, left);
						if (nn) {
							left -= nn;
							if (left > 0)
								memmove(cp, cp+nn, left);
							fcc -= nn;
							goto top; /* n^2 */
						}
					}
				FD_SET(p, &obits);              /* try write */
			}
		}

		if (FD_ISSET(p, &obits) && fcc > 0) {
			cc = write(p, fbp, fcc);
			if (cc > 0) {
				fcc -= cc;
				fbp += cc;
			}
		}

		if (FD_ISSET(p, &ibits)) {
			pcc = read(p, pibuf, sizeof (pibuf));
			pbp = pibuf;
			if (pcc < 0 && errno == EWOULDBLOCK)
				pcc = 0;
			else if (pcc <= 0)
				break;
			else if (pibuf[0] == 0) {
				pbp++, pcc--;
				FD_SET(netf, &obits);      /* try write */
			} else {
				if (pkcontrol(pibuf[0])) {
					pibuf[0] |= oobdata[0];
					send(netf, &pibuf[0], 1, MSG_OOB);
				}
				pcc = 0;
			}
		}
		if ((FD_ISSET(netf, &obits)) && pcc > 0) {
			cc = write(netf, pbp, pcc);
			if (cc < 0 && errno == EWOULDBLOCK) {
				/*
				 * This happens when we try write after read
				 * from p, but some old kernels balk at large
				 * writes even when select returns true.
				 */
				if (!FD_ISSET(p, &ibits))
					sleep(5);
				continue;
			}
			if (cc > 0) {
				pcc -= cc;
				pbp += cc;
			}
		}
	}
}

static void cleanup(void)
{
    bb_wutmp_logout(line);
    /* close network  */
    shutdown(netf, 2);
}

static void sigchld_handler(int sig)
{
	(void)sig;
	cleanup();
	/* all done */
	exit(0);
}

static void setup_term(void)
{
    char *x;
    struct termios tt;

    tcgetattr(0, &tt);
    if ((x=strchr_set0(termtype, '/')) != NULL) {
	x++;
	cfsetispeed(&tt, atoi(x));
	cfsetospeed(&tt, atoi(x));
    }
    tcsetattr(0, TCSAFLUSH, &tt);
}

/*
 * Close all fds, in case libc has left fun stuff like /etc/shadow open.
 */
static void closeall(void)
{
    int i;
    for (i = getdtablesize()-1; i > 2; i--) close(i);
}


static void __attribute__((noreturn))
child(const char *hname, int authenticated)
{
    char *termenv[2];

    setup_term();

    termenv[0] = malloc(strlen(termtype)+6);
    if (termenv[0]) {   /* shouldn't ever fail, mind you */
	strcpy(termenv[0], "TERM=");
	strcat(termenv[0], termtype);
    }
    termenv[1] = NULL;

    if (authenticated) {
	closeall();
	execle(bb_login_path, LOGIN_SHORT_NAME, "-p",
	       "-h", hname, "-f", lusername, NULL, termenv);
    }
    else {
	if (lusername[0] == '-') {
	    syslog(LOG_AUTH|LOG_INFO, "rlogin with an option as a name!");
	    exit(1);
	}
	closeall();
	execle(bb_login_path, LOGIN_SHORT_NAME, "-p",
	       "-h", hname, lusername, NULL, termenv);
    }
    /* Can't exec login, croak */
    fatal(STDERR_FILENO, bb_login_path, 1);
}


static void getstr(char *buf, int cnt, const char *errmsg)
{
    char c;
    do {
	if (read(STDIN_FILENO, &c, 1) != 1) exit(1);
	if (--cnt < 0) fatal(STDOUT_FILENO, errmsg, 0);
	*buf++ = c;
    } while (c != 0);
}

static int rlogind_login_tty(int fd)
{
	(void) setsid();
	if (ioctl(fd, TIOCSCTTY, (char *)NULL) == -1)
		return (-1);
	(void) dup2(fd, 0);
	(void) dup2(fd, 1);
	(void) dup2(fd, 2);
	if (fd > 2)
		(void) close(fd);
	return (0);
}

static int rlogind_forkpty (int *amaster, char *name, struct termios *termp, struct winsize *winp)
{
  int master, slave, pid;

  if (openpty (&master, &slave, name, termp, winp) == -1)
    return -1;

  switch (pid = fork ())
    {
    case -1:
      return -1;
    case 0:
      /* Child.  */
      close (master);
      if (rlogind_login_tty (slave))
	_exit (1);

      return 0;
    default:
      /* Parent.  */
      *amaster = master;
      close (slave);

      return pid;
    }
}

int rlogind_main(void)
{
    int master, pid, on;
    char *hname;
#define authenticated on

    bb_openlog(LOG_PID | LOG_CONS, LOG_AUTH);
    log_stderr = 0;
    log_syslog = LOG_ERR;

    opterr = 0;
    /* reverse flag l and n */
    on = bb_xgetopt_ulflags("ahLln?\0l--:n--");
    check_all = on & 1;
    allow_root_rhosts = on & 2;
    deny_all_rhosts_hequiv = on & 4;
    use_rhosts = (on & 8) ? 1 : 0;      /* 1 - default */
    keepalive = on & 16;
    if(on & 32)
	bb_error_msg_and_die("usage: %s [-ahLln]", bb_applet_name);

    oobdata[0] = (char)TIOCPKT_WINDOW;

    hname = network_init();

    getstr(rusername, sizeof(rusername), "remuser too long");
    getstr(lusername, sizeof(lusername), "locuser too long");
    getstr(termtype, sizeof(termtype), "Terminal type too long");

    /*
     * This function will either die, return -1 if authentication failed,
     * or return 0 if authentication succeeded.
     *
     * dholland 6/13/97 I've changed this so it doesn't even call
     * auth_checkauth if the hostname was bogus. I don't *think*
     * this will break anything or give away state secrets.
     */
    authenticated = 0;
    if (hostok) {
	if (auth_checkauth(hname) == 0) {
	   authenticated=1;
	}
    }
    if (confirmed == 0) {       /* do_rlogin may do this */
	write(netf, "", 1);
	confirmed = 1;          /* we sent the null! */
    }

    if (!hostok) {
	write(netf, "rlogind: Host address mismatch.\r\n", 33);
    }

    pid = rlogind_forkpty(&master, line, NULL, &win);
    if (pid < 0) {
	if (errno == ENOENT) fatal(netf, "Out of ptys", 0);
	fatal(netf, "Forkpty", 1);
    }
    if (pid == 0)
	child(hname, authenticated);

    on = 1;
    ioctl(netf, FIONBIO, &on);
    ioctl(master, FIONBIO, &on);
    ioctl(master, TIOCPKT, &on);
    bb_signal((1 << SIGCHLD), sigchld_handler);
    protocol(master);
    bb_signal((1 << SIGCHLD), SIG_IGN);
    cleanup();
    return 0;
}
#endif /* APPLET_USAGE */
