/*
 *  Author: Paul Sheer
 *  Modifications: Krzysztof Taraszka
 *
 *  Modifications: Vladimit Oleynik
 *  busybox port, size reduction, many signed/unsigned/int/long corrections
 *  -p port, usage for nonroot user, reduce data segment to zero
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street - Fifth Floor, Boston, MA
 *  02110-1301, USA.
 */
#ifdef APPLET_USAGE
#define sheerdns_trivial_usage \
	"[-D dir] [-U user] [-d] [-t ttl] [-p port] [-i iface-ip]...\n" \
	"or [-D dir] -u domain -(A|P|N|S|C|M) record..."
#define sheerdns_full_usage \
	"Small nonrecursive local DNS server daemon.\n\n" \
	"Options:\n" \
	"\t-D sheerdns_path\n" \
	"\t-U user for drop privileges\n" \
	"\t-d\t\tstay daemon\n" \
	"\t-t ttl_seconds\n" \
	"\t-u domain\tupdate domain with next options record\n" \
	"\t-A\t\tadd A record to domain\n" \
	"\t-P\t\tadd PTR record to domain\n" \
	"\t-N\t\tadd NS record to domain\n" \
	"\t-S\t\tadd SOA record to domain\n" \
	"\t-C\t\tadd CNAME record to domain\n" \
	"\t-M\t\tadd MX record to domain"
#else

#define COMPLEMENTARY \
 "\0d--u:t--u:p--u:i::i--u:u--dtpi:A--dtpi:P--dtpi:N--dtpi:S--dtpi:M--dtpi:C--dtpi:=u:=A:=P:=N:=S:=M:=C"
#include "opt_gen.h"
OPT_GEN(
opt(D, OPT_SHEERDNS_PATH, arg),
opt(d, OPT_DAEMON),
opt(t, OPT_TTL, arg),
opt(p, OPT_PORT, arg),
opt(i, OPT_IFACE_IP, arg),
opt(u, OPT_DOMAIN, arg),
opt(A, OPT_ADD_A, arg),
opt(P, OPT_ADD_PTR, arg),
opt(N, OPT_ADD_NS, arg),
opt(S, OPT_ADD_SOA, arg),
opt(M, OPT_ADD_MX, arg),
opt(C, OPT_ADD_CNAME, arg),
opt(U, OPT_USER, arg)
)
#include "opt_gen.h"

#include "busybox.h"
#include <sys/time.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <syslog.h>
#include "bb_inet.h"
#include <netinet/in.h>
#include "bb_xopen.h"

#ifndef SHEERDNS_DIR
# define SHEERDNS_DIR    "/var/sheerdns"
#endif

//#define DEBUG 1

/* See section 2.3.1, "Preferred name syntax", of RFC-1035 : */
#define MAX_RECORD_LEN          255
#define isevil(c)                                               \
	(!(                                                     \
		   ((c) >= 'a' && (c) <= 'z')                   \
		|| ((c) >= 'A' && (c) <= 'Z')                   \
		|| ((c) >= '0' && (c) <= '9')                   \
		||  (c) == '.' ||  (c) == '-'                   \
	))

#define issuspect(c)    ((c) < ' ' || (c) > '~' || (c) == '/')

#define REQ_A           1
#define REQ_NS          2
#define REQ_CNAME       5
#define REQ_SOA         6
#define REQ_PTR         12
#define REQ_MX          15
#define REQ_TXT         16

static void string_wash(char *s)
{
	unsigned char *p = (unsigned char *) s;

	unsigned char *q;

	int n;

	for (n = 0, q = p; *q && n < MAX_RECORD_LEN; q++, n++)
		if (!issuspect(*q))
			*p++ = *q;
	*p = '\0';
}

static void string_purify(char *s)
{
	unsigned char *p = (unsigned char *) s;
	unsigned char *q;
	int n;

	for (n = 0, q = p; *q && n < MAX_RECORD_LEN; q++, n++)
		if (!isevil(*q))
			*p++ = *q;
	*p = '\0';
}

static char **string_split_and_chomp(const char *s)
{
	char *p, **a;
	int n;
#define space_for_split(c) (c == ' ' || c == '\t' || c == '\r' || c == '\n' || c == '\f')

	for (n = 0, p = (char *) s; *p; n++) {
		while (*p && space_for_split(*p))
			p++;
		if(*p == '\0')
			break;
		while (*p && !space_for_split(*p))
			p++;
	}
	if (n == 0 || !(a = malloc((n + 1) * sizeof(char *) + strlen(s) + 1)))
		return NULL;
	for (n = -1, p = (char *) s; *p; ) {
		while (*p && space_for_split(*p))
			p++;
		if(*p == '\0')
			break;
		a[++n] = p;
		while (*p && !space_for_split(*p))
			p++;
		if(*p == '\0')
			break;
		*p++ = '\0';
	}
	a[n+1] = NULL;
	return a;
}

static int string_present(char *s, char **a)
{
	int i;

	for (i = 0; a[i]; i++)
		if (!strcmp(a[i], s))
			return 1;
	return 0;
}


static int listen_socket(const char *iface_addr, int listen_port, int type)
{
	struct sockaddr_in a;
	int s;

	s = bb_xsocket(type | BB_XSOCKET_FLG_ERR_DIE);
#ifdef SO_REUSEADDR
	if (setsockopt_1i(s, SOL_SOCKET, SO_REUSEADDR, 1) < 0)
		bb_perror_msg_and_die("setsockopt() failed");
#endif
	memset(&a, 0, sizeof(a));
	a.sin_port = htons(listen_port);
	a.sin_family = AF_INET;
	if (iface_addr != NULL && !inet_aton(iface_addr, &a.sin_addr))
		bb_error_msg_and_die("bad iface address %s", iface_addr);
	bb_xbind(s, (struct sockaddr *)&a, sizeof(a));
	if(type == 0) {
		listen(s, 50);
		printf("accepting packets on %s:%d\n", (iface_addr == NULL ? "0.0.0.0" : iface_addr), listen_port);
	}
	return s;
}

static char *hex_hash(const char *s, char *h)
{
	unsigned long r = 0;

	for (; *s; s++)
		r += (unsigned char)*s;
	sprintf(h, "%02X", (unsigned int)(r & 0xFF));
	return h;
}


static const char *const qtype_name[] = {
	NULL, "A", "NS", "MD", "MF", "CNAME", "SOA", "MB", "MG", "MR", "NULL",
	"WKS", "PTR", "HINFO", "MINFO", "MX", "TXT", "RP", "AFSDB", "X25",
	"ISDN",
	"RT", "NSAP", "NSAP_PTR", "SIG", "KEY", "PX", "GPOS", "AAAA", "LOC",
	"NXT", "EID", "NIMLOC", "SRV", "ATMA", "NAPTR", "KX", "CERT", "A6",
	"DNAME", "SINK", "OPT",
};

static int make_and_open_filename(int qtype, char *s)
{
	char str[1024];
	char hash_buf[4];
	int fd;
	char *t;
	char new_query[16];

	if (qtype < 1 || qtype > 41)
		return -1;
	/* try PTR, remove .in-addr.arpa */
#ifdef DEBUG
	fprintf(stderr, "DEBUG: make_and_open_filename(%s)\n", s);
#endif
	if ((t = strstr(s, ".in-addr.arpa")) != NULL && (t - s) <= 15) {
		fd = (t - s);
		s = memcpy(new_query, s, fd);
		new_query[fd] = '\0';
#ifdef DEBUG
		fprintf(stderr, "DEBUG: found .in-addr.arpa, strip to=%s\n", s);
#endif
	}
	snprintf(str, sizeof(str), "%s/%s/%s", hex_hash(s, hash_buf), s, qtype_name[qtype]);
	fd = open(str, O_RDONLY);
#ifdef DEBUG
	fprintf(stderr, "DEBUG: open %s, fd=%d\n", str, fd);
#endif
	return fd;
}

static void round_robin(char **s)
{
	int n, i;
	char *t;

	for (n = 0; s[n]; n++)
		;
	for (i = 0; i < n * 2; i++) {
		int e = (long)random() % n;

		t = s[i % n];
		s[i % n] = s[e];
		s[e] = t;
	}
}

static char **directory_lookup(int qtype, unsigned char *s)
{
	int fd = -1, c, i;
	char buf[1024];
	char **r, *q = (char *)s;

	/* if this is an SOA or NS lookup, we descend through the
	   domains for any that we are an authority for. this allows
	   the user to create one NS entry for say test.com and have
	   that returned as the NS for www.test.com,
	   ftp.henry.test.com, user1.lab.chemsitry.test.com, etc. : */

	while ((fd = make_and_open_filename(qtype, q)) < 0
	       && (q = strchr(q, '.')) && q++
	       && (qtype == REQ_SOA || qtype == REQ_NS));
	if (fd == -1) {
		/* try lookup "*.example.com" where asking for "nonexistant.example.com": */
		char *t;

		if (qtype == REQ_SOA || qtype == REQ_NS)        /* probably a bad idea to wildcard these, so return */
			return NULL;
#ifdef DEBUG
		fprintf(stderr, "DEBUG: try lookup * for %s\n", s);
#endif
		q = t = strdup((char *)s);
		while (*t && *t != '.')
			t++;
		if (!*t || t == q) {
			free(q);
			return NULL;
		}
		*--t = '*';
#ifdef DEBUG
		fprintf(stderr, "DEBUG: try lookup * for %s\n", t);
#endif
		if ((fd = make_and_open_filename(qtype, t)) < 0) {
#ifdef DEBUG
			fprintf(stderr, "DEBUG: not found try lookup * for %s\n", t);
#endif
			free(q);
			return NULL;
		}
		free(q);
	}
	/* done finding and opening the file, now read the contents: */
	c = read(fd, buf, sizeof(buf));
	close(fd);
	if (c <= 0 || c == sizeof(buf))
		return NULL;
	buf[c] = '\0';
	if (!(r = string_split_and_chomp(buf)))
		return NULL;
	for (i = 0; r[i]; i++) {
#ifdef DEBUG
		fprintf(stderr, "DEBUG: compare with %s\n", r[i]);
#endif
		if (qtype == REQ_TXT)
			string_wash(r[i]);
		else
			string_purify(r[i]);
		if (!*(r[i])) {
			free(r);
#ifdef DEBUG
			fprintf(stderr, "DEBUG: string reduce to empty\n");
#endif
			return NULL;
		}
	}
	if (qtype == REQ_A || qtype == REQ_NS)
		round_robin(r);
	return r;
}

typedef unsigned char u1;
typedef unsigned short u2;
typedef unsigned int u4;

static u4 get_serial(char *s)
{
	struct stat st;
	time_t t;
	u4 serial;
	struct tm *lt;
	int fd = make_and_open_filename(REQ_SOA, s);

	if (fd < 0 || fstat(fd, &st))
		t = (time_t)1000000000UL;
	else
		t = st.st_mtime;
	close(fd);
	lt = localtime(&t);
	/* hack a serial number from the modified time of the file - this
	   has a resolution of 32 seconds and looks like YYYYppnnnn where
	   pp.nnnn is the percentage of the year passed: */
	serial = ((u4) lt->tm_year + 1900) * 1000000 + ((u4)
							((u4) lt->tm_yday *
							 24 * 3600 +
							 lt->tm_hour * 3600 +
							 lt->tm_min * 60 +
							 lt->tm_sec) >> 5);
	return serial;
#ifdef DEBUG
	fprintf(stderr, "DEBUG: serial=%lu\n", (unsigned long)serial);
#endif
}

enum
{
	ANSWERS = 0,
	SERVERS = 1,
	EXTRAS = 2,
	ANS_BLOCKS = 3
};

#define ONE_DAY                 (24 * 60 * 60)
#define THREE_DAYS              (3 * 24 * 60 * 60)

CALC_BSS struct BB_COMMON_BSS {
	u4 short_time;
	const char *sheerdns_dir;
	int dir_created;
	const char *create_domain;
	int mode;
} bb_common_bss;
#define short_time bb_common_bss.short_time
#define sheerdns_dir bb_common_bss.sheerdns_dir
#define dir_created bb_common_bss.dir_created
#define create_domain bb_common_bss.create_domain
#define mode bb_common_bss.mode

#define TTL_POLICY(q)           (((q) == REQ_SOA || (q) == REQ_NS) ? THREE_DAYS : short_time)


/* the following values are meaningless, since they all apply to
secondary DNS servers, and we are not using them: */
#define REFRESH                 (10 * 60)
#define RETRY                   (10 * 60)
#define EXPIRE                  (THREE_DAYS)
#define MINTTL                  short_time

static int put_name(u1 *buf, u1 **here, u1 *name)
{
	unsigned int c;
	int i, k = 0, n, offset;

	offset = *here - buf;
	while ((c = name[k]) != 0) {
		n = 0;
		while ((c = name[k + n]) != 0 && c != '.')
			n++;
		if (n == 0)
			break;
		*(*here)++ = (unsigned char)(n & ~0xc0);
		for (i = 0; i < n; i++)
			*(*here)++ = name[k++];
		if (name[k] == '.')
			k++;
	}
	*(*here)++ = 0;
	return offset;
}

#define lower_case(c) (((c) >= 'A' && (c) <= 'Z') ? ((c) | 0x20) : (c))

static int get_name(u1 *msg, u1 **here, u1 *end, unsigned char *string, int string_len, int k)
{
	unsigned int len;

	if (*here >= end)
		return -1;
	while ((len = **here) != 0) {
		*here += 1;
		if ((len & 0xc0) == 0xc0) {
			unsigned int offset;
			u1 *p;

			if (*here >= end)
				return -1;
			offset = ((len & ~0xc0) << 8) + **here;
			p = &msg[offset];
			k = get_name(msg, &p, end, string, string_len, k);
			if (k == -1)
				return k;
			break;
		} else if ((len & 0xc0)) {
			return -1;
		} else {
			unsigned int i;

			for (i = 0; i < len; i++) {
				if (*here >= end)
					return -1;
				if (k < string_len - 1) {
					if (isevil(**here))
						return -1;
					string[k++] = lower_case(**here);
				}
				*here += 1;
			}
			if (k < string_len - 1)
				string[k++] = '.';
		}
	}
	*here += 1;
	string[k] = '\0';
	return k;
}


static u1 *fill_u2(u1 *p, u2 val)
{
	*p++ = val >> 8;
	*p++ = val & 0xFF;
	return p;
}

static u1 *fill_u4(u1 *p, u4 val)
{
	*p++ = (val >> 24) & 0xFF;
	*p++ = (val >> 16) & 0xFF;
	return fill_u2(p, (u2)(val & 0xFFFF));
}

#define er(s)   do { errno = 0; bb_error_msg("%s", #s); return -1; } while (0)
#define xfree(x)        do { if ((x) && ((char *)(x)) != (char *)extra_results) free(x); (x) = NULL; } while (0)
#define MAX_EXTRA_RESULTS       64

static int
process_packet(int s, u1 *in_buf, int packet_len, struct sockaddr *from, socklen_t fromlen)
{
	u1 _out_buf[1536 + 2];  /* cannot get packets bigger than this on ethernet */
	u1 *out_buf = &_out_buf[2];
	u1 query[MAX_RECORD_LEN + 1];
	char *extra_results[MAX_EXTRA_RESULTS];
	char *extra_queries[MAX_EXTRA_RESULTS];
	u4 extra_ttls[MAX_EXTRA_RESULTS];
	int query_len, name_pointer, i, j, extra_results_len = 0;
	u2 id, flags, nqueries, nanswers, nservers, nextras, qtype = 0, orig_qtype = 0, qclass = 0, name_len, *p2;
	u1 *p1, *len_pointer;

	/* RFC-1035 says size limit of 512 - we make ours 1024 for both UDP and TCP */
	if (packet_len < 12 || packet_len > 1024)
		er(invalid packet size);

	p2 = (u2 *)in_buf;

	id = ntohs(p2[0]);
	flags = ntohs(p2[1]);
	/* get number of queries */
	if (!(nqueries = ntohs(p2[2])))
		er(no queries);
	/* get the query string */
	p1 = in_buf + 12;
	memset(query, 0, sizeof(query));
	if (get_name(in_buf, &p1, in_buf + packet_len, query, sizeof(query), 0) == -1)
		er(bad query format);
	query_len = strlen((char *)query);
	if (*query && query[query_len - 1] == '.')
		query[query_len - 1] = '\0';
	/* get the query type */
	qtype = *p1++;
	qtype = (qtype << 8) | *p1++;
	orig_qtype = qtype;
	qclass = *p1++;
	qclass = (qclass << 8) | *p1++;

	nqueries = 1;
	nanswers = 0;
	nservers = 0;
	nextras = 0;

	memset (_out_buf, 0, sizeof (_out_buf));
	p1 = out_buf + 12;

	/* query section starts here ----> */
	name_pointer = put_name(out_buf, &p1, query);
	p1 = fill_u2(p1, qtype);
	p1 = fill_u2(p1, qclass);

	if ((flags & 0x8000))   /* response packet - ignoring */
		er(ignoring response packet);

	/* now check for valid flags. if not valid, jump over the
	   part where we fill in the packet: */
	flags &= 0x7900;        /* copy recursive-query-bit, copy opcode-bits */
	flags |= 0x8000;        /* response-bit */

	if (!(qtype == REQ_A || qtype == REQ_PTR || qtype == REQ_MX
	      || qtype == REQ_CNAME || qtype == REQ_NS || qtype == REQ_SOA
	      || qtype == REQ_TXT)) {
#ifdef DEBUG
		fprintf(stderr, "DEBUG: process_packet unknow qtype=%lu\n", (unsigned long)qtype);
#endif
		goto empty_packet;
	}                       /* we have none of these records */
	if (qclass != 1 /* class INET */ ) {
		flags |= 4;     /* not supported */
		goto empty_packet;
#ifdef DEBUG
		fprintf(stderr, "DEBUG: process_packet unknow qclass=%lu\n", (unsigned long)qclass);
#endif
	}
	if ((flags & 0x7800) != 0) {
#ifdef DEBUG
		fprintf(stderr, "DEBUG: process_packet unknow flags=0x%lx\n", (unsigned long)flags);
#endif
		goto empty_packet;
	}
	/* answer section starts here ----> */
	extra_results[0] = NULL;
	extra_queries[0] = NULL;

	for (j = 0; j < ANS_BLOCKS; j++) {      /* there are three answer blocks */
		char **lookup_results = NULL;
		u2 *to_incr = NULL;

		switch (j) {
		case ANSWERS:   /* section 1 */
#ifdef DEBUG
			fprintf(stderr, "DEBUG: process_packet section 1\n");
#endif
			to_incr = &nanswers;
			lookup_results = directory_lookup(qtype, query);

			/* a missing a record may be a CNAME - try find it and
			   then proceed as though we had a CNAME lookup: */
			if (!lookup_results) {
				qtype = REQ_CNAME;
				xfree(lookup_results);
				lookup_results = directory_lookup(REQ_CNAME, query);
			}

			/* CNAME requires a directly adjacent A record within the
			   answer section (as per RFC-1034, 3.6.2). This is a
			   special case, and breaks the elegance of this algorithm.
			   So we prepend the CNAME answer, and then proceed as though
			   it were a regular A record (or other record) by replacing
			   the query with a new query of the canonical name: */
			if (qtype == REQ_CNAME) {
				if (lookup_results) {
					nanswers++;
					put_name(out_buf, &p1, (u1 *)query);
					p1 = fill_u2(p1, qtype);
					p1 = fill_u2(p1, qclass);
					p1 = fill_u4(p1, short_time);
					len_pointer = p1;       /* store length of string here */
					p1 += 2;
					name_pointer = put_name(out_buf, &p1, (u1 *)lookup_results[0]);       /* <--- replacement of query string with cname(2) */
					name_len = (p1 - len_pointer) - 2;
					len_pointer = fill_u2(len_pointer, name_len);  /* back insert the length */

					/* now proceed as though this were a regular query: */
					qtype = orig_qtype;
					strcpy((char *)query, lookup_results[0]);     /* <--- replacement of query string with cname(2) */
					xfree(lookup_results);
					lookup_results = directory_lookup(orig_qtype, query);
				}
			}
#ifdef DEBUG
			fprintf(stderr, "DEBUG: process_packet section 1 end\n");
#endif
			break;
		case SERVERS:   /* section 2 */
#ifdef DEBUG
			fprintf(stderr, "DEBUG: process_packet section 2\n");
#endif
			to_incr = &nservers;
			/* append any SOA records possibly present in the
			   database, but not if the original query was an SOA query,
			   in which case it would already appear in the section 1: */
			if (orig_qtype != REQ_SOA) {
				qtype = REQ_SOA;
				lookup_results = directory_lookup(REQ_SOA, query);
				/* if we know we are the authority, then we want to be
				   explicit that the domain really does not exist: */
				if (!nanswers && lookup_results)
					flags = (flags & ~0xf) | 3;
			}
			/* authoritatively no such domain  */
			/* append any NS records possibly present in the database,
			   but not if the original query was an NS query (in which
			   case it would already appear in the section 1), or if we
			   have already added SOA records (because its not really
			   necessary to add NS records along with NS records): */
			if (!lookup_results
			    && (orig_qtype !=
				REQ_NS /* don't repeat stuff */ )) {
				xfree(lookup_results);
				qtype = REQ_NS;
				lookup_results = directory_lookup(REQ_NS, query);
			}
			break;
		case EXTRAS:    /* section 3 */
#ifdef DEBUG
			fprintf(stderr, "DEBUG: process_packet section 3\n");
#endif
			to_incr = &nextras;
			/* any hostnames referenced in the previous two sections
			   are appened with their proper IP addresses to a list for
			   inclusion in the last section: */
			qtype = REQ_A;
			lookup_results = extra_results;
			break;
		}

		/* loop through each found record: */
		for (i = 0; lookup_results && lookup_results[i] && i < 256; i++, (*to_incr)++) {
			u1 *p1_within_512_bytes;
			u4 ttl;

#ifdef DEBUG
			fprintf(stderr, "DEBUG: loop through each found record\n");
#endif
			p1_within_512_bytes = p1;       /* psuh the offset in case we overrun */
			if (j == EXTRAS) {
				ttl = extra_ttls[i];    /* extra section has myriad TTLs */
				put_name(out_buf, &p1, (u1 *)extra_queries[i]);
			} /* as well as myriad names */
			else {
				ttl = TTL_POLICY(qtype);       /* other sections have stock TTLs */
				/* pointer to original query string or CNAME result(2), as the case may be: */
				*p1++ = 0xc0 | ((name_pointer >> 8) & ~0xc0);
				*p1++ = (name_pointer & 0xFF);
			}
#ifdef DEBUG
			fprintf(stderr, "DEBUG: fill qtype=%d, qclass=%d, ttl=%d\n", qtype, qclass, ttl);
#endif
			p1 = fill_u2(p1, qtype);
			p1 = fill_u2(p1, qclass);
			p1 = fill_u4(p1, ttl);
			len_pointer = p1;       /* store length of string here */
			p1 += 2;
			if (qtype == REQ_MX) {  /* for mail servers, prefix with priority number */
#ifdef DEBUG
				fprintf(stderr, "DEBUG: fill MX=%d\n", 10 + i * 10);
#endif
				p1 = fill_u2(p1, 10 + i * 10);
			}
			if (qtype == REQ_A) {   /* for A records just store the IP */
				struct in_addr a;

				a.s_addr = 0;
				if (!inet_aton(lookup_results[i], &a))
					flags = (flags & ~0xf) | 1;     /* Format error */
				/* already in network byte order */
				p1 = memcpy(p1, &a.s_addr, 4) + 4;
#ifdef DEBUG
				fprintf(stderr, "DEBUG: fill REQ_A=%s\n", lookup_results[i]);
#endif
			} else if (qtype == REQ_TXT) {  /* for TXT records just store plain text */
				int y;

				for (y = 0; y < 64 && lookup_results[i][y]; y++)
					p1[y + 1] = (u1)lookup_results[i][y];
				*p1 = y, p1 += y + 1;
			} else {
				put_name(out_buf, &p1, (u1 *)lookup_results[i]);      /* store string result */
#ifdef DEBUG
				fprintf(stderr, "DEBUG: fill str=%s\n", lookup_results[i]);
#endif
				if (j != EXTRAS) {
					char **t_results = NULL;

					if (qtype != REQ_CNAME) /* CNAMEs IPs are specially appended in the answer section */
						if (!string_present(lookup_results[i], extra_queries)) /* did we already look this one up? */
							t_results = directory_lookup(REQ_A, (u1 *)lookup_results[i]); /* lookup any A records available */
					if (t_results && extra_results_len < MAX_EXTRA_RESULTS - 1) {
						extra_results[extra_results_len] = strdup((char *)t_results[0]);
						extra_queries[extra_results_len] = strdup((char *)lookup_results[i]);

						/* now we need to calculate the ttl's of the extra
						   section. Our extra section contains only A records of
						   either an NS, PTR, CNAME, MX, or SOA lookup. NS records
						   have fixed IP address, and CNAME and SOA records are
						   fixed. Its only "A" records of REQ_MX, REQ_CNAME, and
						   REQ_PTR that need a short TTL. This is not exactly
						   correct, because a CNAME or PTR could actually be a
						   nameserver, but hay: */

						extra_ttls[extra_results_len] = TTL_POLICY (qtype);
						extra_results_len++;
						extra_results[extra_results_len] = NULL;
						extra_queries[extra_results_len] = NULL;
					}
					xfree(t_results);
				}
			}

			/* for soa records servers we suffix the entry with: mailbox,
			   serial, refresh, retry, expire, min-ttl: */
			if (qtype == REQ_SOA) {
				lookup_results[i + 1] = NULL;   /* force only one SOA record */
				p1 = (u1 *)strcpy((char *)p1, "\012hostmaster") + 11;
				put_name(out_buf, &p1, (u1 *)lookup_results[i]);      /* store string result */
				p1 = fill_u4(p1, get_serial((char *)query));
				p1 = fill_u4(p1, REFRESH);
				p1 = fill_u4(p1, RETRY);
				p1 = fill_u4(p1, EXPIRE);
				p1 = fill_u4(p1, MINTTL);
#ifdef DEBUG
				fprintf(stderr, "DEBUG: fill SOA=%s\n", lookup_results[i]);
#endif
			}

			name_len = (p1 - len_pointer) - 2;
			len_pointer = fill_u2(len_pointer, name_len);  /* back insert the length */
			if ((p1 - out_buf) > (from == NULL ? 1024 : 512)) { /* limit to 1024 bytes for TCP, 512 for UDP */
				p1 = p1_within_512_bytes;
				if(from)
					flags |= 0x0200; /* truncation bit. For TCP we are silent about the truncation */
				/* do not increment (*to_incr), and then */
				xfree(lookup_results);
				goto break_2;
			}
		}
		xfree(lookup_results);
	}
      break_2:
	while (--extra_results_len >= 0) {
		free(extra_results[extra_results_len]);
		extra_results[extra_results_len] = NULL;
		free(extra_queries[extra_results_len]);
		extra_queries[extra_results_len] = NULL;
	}
	packet_len = (p1 - out_buf);

	if (nanswers || nservers)
		flags |= 0x0400;  /* authority-bit 0x0400 */

      empty_packet:
	p1 = out_buf;           /* goto start of packet to fill in total counts */
	p1 = fill_u2(p1, id);
	p1 = fill_u2(p1, flags);
	p1 = fill_u2(p1, nqueries);
	p1 = fill_u2(p1, nanswers);
	p1 = fill_u2(p1, nservers);
	p1 = fill_u2(p1, nextras);

	/* finally... */
	if (from == NULL) {     /* TCP */
		out_buf -= 2;
		fill_u2(out_buf, packet_len);
		/* this assumes the OS queue can hold our data without
		   blocking. for 1026 bytes or less, we are pretty sure it can. */
	}
#ifdef DEBUG
	fprintf(stderr, "\n");
#endif
	return sendto(s, out_buf, packet_len, 0, from, fromlen);
}

struct TCP_CONNECTION {
	u1 *buf;
	time_t last_activity;
	int len, alloced;
};

struct connection
{
	struct connection *next;
	int type;
	int fd;
	struct TCP_CONNECTION *tcp;
};

enum {
	TYPE_LISTENS_UDP = 0,
	TYPE_LISTENS_TCP = 1,
	TYPE_TCP = 2
};

static void new_connection(struct connection **pc, int type, int fd)
{
	struct connection *c;

	for (c = *pc; c; c = c->next)
		if(c->fd == -1) /* leaved */
			break;
	if(c == NULL) {
		/* allocate for struct + 1024 bytes for TCP packet + 2 for packet's len */
		c = malloc(sizeof(struct connection) + (type == TYPE_TCP ? (sizeof(struct TCP_CONNECTION) + 1026) : 0));
		c->next = *pc;
		*pc = c;
		if(type == TYPE_TCP) {
			void *p = (c + 1);
			c->tcp = p;
			p = (((struct TCP_CONNECTION *)p) + 1);
			c->tcp->buf = p;
		}
	}
	c->type = type;
	c->fd = fd;
}

static void new_listens(struct connection **pc, const char *l, int port)
{
	new_connection(pc, TYPE_LISTENS_UDP, listen_socket(l, port, BB_XSOCKET_FLG_SOCK_DGRAM));
	new_connection(pc, TYPE_LISTENS_TCP, listen_socket(l, port, 0));
}

static void sheerdns_mkdir(const char *dirx, const char *fname, const char *out_value, int trunc_append)
{
	if (mkdir(dirx, 0755) && errno != EEXIST)
		bb_perror_msg_and_die("unable make directory %s", dirx);
	if (fname != NULL) {
		char *path;
		int fd;

		path = xasprintf("%s/%s", dirx, fname);
		fd = bb_xopen(path, O_WRONLY | O_CREAT | trunc_append | BB_OPEN_ERR, 0644);
		dprintf(fd, "%s\n", out_value);
		close(fd);
		free(path);
	}
}

static void make_directories(void)
{
	umask(022);
	sheerdns_mkdir(sheerdns_dir, NULL, NULL, 0);
	xchdir(sheerdns_dir);
	sheerdns_mkdir("C9", NULL, NULL, 0);
	sheerdns_mkdir("C9/localhost", "A", "127.0.0.1", O_TRUNC);
	sheerdns_mkdir("7A", NULL, NULL, 0);
	sheerdns_mkdir("7A/localhost.localdomain", "A", "127.0.0.1", O_TRUNC);
	sheerdns_mkdir("B5", NULL, NULL, 0);
	sheerdns_mkdir("B5/1.0.0.127", "PTR", "localhost", O_TRUNC);
	dir_created = 1;
}

static void store_record(const char *value, const char *type, const char *record)
{
	char hash_buf[4];
	char *buf;

	if(!dir_created)
		make_directories();
	hex_hash(value, hash_buf);
	sheerdns_mkdir(hash_buf, NULL, NULL, 0);
	buf = xasprintf("%s/%s", hash_buf, value);
	sheerdns_mkdir(buf, type, record, O_APPEND);
	printf("%s >> %s/%s/%s\n", record, sheerdns_dir, buf, type);
	free(buf);
}

static void set_u(char *arg)
{
	if(mode == 1)
		bb_show_usage();
	mode = 1;
	create_domain = arg;
}

static void store_records(const char *type, const char *record)
{
	if(create_domain == NULL)
		bb_show_usage();
	if(strchr(record, '.') != NULL) {
		store_record(create_domain, type, record);
	} else {
		char *r = xasprintf("%s.%s", record, create_domain);
		store_record(create_domain, type, r);
		free(r);
	}
	mode = 2;
}

static void set_A(char *arg)
{
	u4 tst;

	if(!inet_aton(arg, (void *)&tst))
		bb_error_msg_and_die("bad ip address %s for A record", arg);
	store_records("A", arg);
}

static void set_N(char *arg)
{
	store_records("NS", arg);
}

static void set_S(char *arg)
{
	store_records("SOA", arg);
}

static void set_M(char *arg)
{
	store_records("MX", arg);
}

static void set_C(char *arg)
{
	if(strchr(arg, '.') == NULL)
		bb_error_msg_and_die("%s: need the fqdn name for store a CNAME record", arg);
	store_records("CNAME", arg);
}

static void set_P(char *arg)
{
	u4 tst;
	int dot = 0;
	char tst_dot[16];
	const char *d = create_domain;

	if(mode != 1)
		bb_error_msg_and_die("need the individual -u domain for store a PTR");
	if(strchr(arg, '.') == NULL)
		bb_error_msg_and_die("%s: need the fqdn name for store a PTR to -u %s", arg, d);
	if(!inet_aton(d, (void *)&tst)) {
		unsigned int l = strlen(d);
		if(l > 15 || l < 7) {
		  ec:
			bb_error_msg_and_die("bad ip address %s for make PTR", d);
		}
		strcpy(tst_dot, d);
		if(tst_dot[l-1] == '*' && tst_dot[l-2] == '.') {
			tst_dot[l-1] = '0';     /* 1.2.3.* to 1.2.3.0 */
			if(!inet_aton(tst_dot, (void *)&tst))
				goto ec;
			dot = 1;
		} else {
			goto ec;
		}
	}
	/* reverse 1.2.3.X to X.3.2.1 */
	sprintf(tst_dot, "%d.%d.%d.%d", (int)((tst >> 24)), (int)((tst >> 16) & 0xff),
			    (int)((tst >> 8) & 0xff), (int)(tst & 0xff));
	if(dot)
		tst_dot[0] = '*';
	store_record(tst_dot, "PTR", arg);
	mode = 3;
}

int
sheerdns_main(void)
{
	int s;
	u2 port = 53;
	u1 in_buf[1536];        /* cannot get packets bigger than this on ethernet */
	llist_t *listen_interface = NULL;
	time_t current_time = current_time; /* make gcc happy: before set it unused, really! */
	struct connection *connections = NULL;
	struct passwd *pw;
	char *nobody = "nobody";
	char *sshort_time, *sport;

	short_time = ONE_DAY;
	sheerdns_dir = SHEERDNS_DIR;

	s = bb_xgetopt_ulflags(opt_str, &sheerdns_dir, &sshort_time, &sport,
				&listen_interface, set_u,
				set_A, set_P, set_N, set_S, set_M, set_C,
				&nobody);
	if(s & OPT_TTL) {
		short_time = bb_xparse_ui_number(BB_PN_SET_LOWER1 | BB_PN_SET_UPPER | BB_PN_ERR_STR1,
						sshort_time, 0xffff, "invalid value %s for option -t");
	}
	if(s & OPT_PORT) {
		port = bb_xparse_ui_number(BB_PN_SET_LOWER1 | BB_PN_SET_UPPER | BB_PN_ERR_STR1,
						sport, 0xffff, "invalid value %s for option -p");
	}
	if(listen_interface != NULL) {
		llist_t *l;
		for(l = listen_interface; l != NULL; l = l->link)
			new_listens(&connections, l->data, port);
	}

	if (mode == 1)
		bb_show_usage();
	if (mode > 1)
		return 0;

	signal(SIGHUP, SIG_IGN);

	if(!dir_created)
		make_directories();

	pw = getpwnam(nobody);
	if (pw == NULL)
		bb_error_msg_and_die("Cannot find uid/gid of user '%s'", nobody);

	if((s & OPT_DAEMON)) {
		bb_openlog(0, LOG_DAEMON);
		xdaemon(XDAEMON_NO_CHDIR | XDAEMON_PIDFILE);
	}

	xchdir(sheerdns_dir);
	if (chroot(".") != 0 && port < 1024)
		bb_perror_msg_and_die("unable chroot into %s", sheerdns_dir);
	if(listen_interface == NULL)
		new_listens(&connections, NULL, port);

	// drop privileges
	setgid(pw->pw_gid);
	setuid(pw->pw_uid);

	while (1) {
		struct connection *c;
		fd_set fdset;
		int r, f_max = 0;

		FD_ZERO(&fdset);

#define killtcp(c) do { shutdown(c->fd, 2); close(c->fd); c->fd = -1; } while (0)
#define max(a,b)   (a > b ? a : b)

		for (c = connections; c; c = c->next) {
			if(c->fd == -1) {
				continue;
			} else if (c->type == TYPE_TCP && (current_time - c->tcp->last_activity) > 30) {
				/* idle to long ? */
				killtcp(c);
			} else {
				f_max = max(f_max, c->fd);
				FD_SET(c->fd, &fdset);
			}
		}

		r = select(f_max + 1, &fdset, NULL, NULL, NULL);
		if (r < 0)
			bb_perror_msg_and_die("select");
		if (r == 0) {   /* should not be possible */
			FD_ZERO(&fdset);
		}

		time(&current_time);

		/* check all connections */
		for (c = connections; c; c = c->next) {
			if(c->fd == -1 || FD_ISSET(c->fd, &fdset) == 0)
				continue;

			if(c->type == TYPE_LISTENS_TCP) {
				/* add a new TCP connection: */
				struct sockaddr_in from;
				socklen_t fromlen = sizeof(from);
				int fd;

				memset(&from, 0, fromlen);
				fd = accept(c->fd, (struct sockaddr *)&from, &fromlen);
				if (fd < 0) {
					bb_perror_msg("accept");
				} else {
					new_connection(&connections, TYPE_TCP, fd);
					connections->tcp->last_activity = current_time;
					connections->tcp->len = 0;
					connections->tcp->alloced = 2;
				}
				continue;
			}

			if(c->type == TYPE_LISTENS_UDP) {
				struct sockaddr_in from;
				socklen_t fromlen = sizeof(from);

				r = recvfrom(c->fd, in_buf, sizeof(in_buf), 0, (struct sockaddr *)&from, &fromlen);
				if (r > 0) {
					memset(in_buf + r, 0, sizeof(in_buf) - r);
					process_packet(c->fd, in_buf, r, (struct sockaddr *)&from, fromlen);
				}
				continue;
			}

			/* processing exiting TCP connection */
			r = read(c->fd, c->tcp->buf + c->tcp->len, c->tcp->alloced - c->tcp->len);
			if (r <= 0 || (!c->tcp->len && r != 2)) {
				/* first read must be exactly 2 bytes */
				killtcp(c);
				continue;
			}
			c->tcp->len += r;
			c->tcp->last_activity = current_time;
			if (c->tcp->len == 2) {      /* first process the packet size bytes */
				int l;

				l = (((u2)c->tcp->buf[0]) << 8) | c->tcp->buf[1];
				if (l > 1024) {
					killtcp(c);
					continue;
				}
				c->tcp->alloced = l + 2;     /* add two bytes for the packet length */
			} else if (c->tcp->len == c->tcp->alloced) {
				if (process_packet(c->fd, c->tcp->buf + 2, c->tcp->len - 2, NULL, 0) < c->tcp->len) {
					killtcp(c);
					continue;
				}
				c->tcp->len = 0;
				c->tcp->alloced = 2;
			}
		}
	}
}
#endif /* USAGE */
