/*

 $Id: ssmtp.c,v 2.48 2001/11/02 10:57:37 matt Exp $

 sSMTP -- send messages via SMTP to a mailhub for local delivery or forwarding.
 This program is used in place of /usr/sbin/sendmail, called by "mail" (et all).
 sSMTP does a selected subset of sendmail's standard tasks (including exactly
 one rewriting task), and explains if you ask it to do something it can't. It
 then sends the mail to the mailhub via an SMTP connection. Believe it or not,
 this is nothing but a filter

 GNU GENERAL PUBLIC LICENSE Version 2

*/

/*
 *  get_arpadate() is a function returning the date in the
 *               ARPANET format (see RFC822 and RFC1123)
 *  Copyright (C) 1998 Hugo Haas
 *
 *  Inspired by smail source code by Ronald S. Karr and Landon Curt Noll
 */

/*
 * base64 code borrowed from fetchmail sources by
 * Eric S. Raymond <esr@snark.thyrsus.com>.
 */

/*
 * Modified by Vladimir Oleynik <dzo@simtreas.ru> (c) 2014 to be
 * used in busybox and size optimizations
*/
#ifdef APPLET_USAGE
#define sendmail_trivial_usage \
	"[-auusername] [-appassword] [-dX] [-Ffullname] [-fname|-rname] [-tvV]"
#define sendmail_full_usage \
	"/usr/sbin/sendmail replacement.\n" \
	"List options (other sendmail's options ignored or default):\n" \
	"\t-auuser\tSpecifies username for login authorization\n" \
	"\t-appas\tSpecifies password for  login  authorization\n" \
	"\t-dX\tSet debugging value to X\n" \
	"\t-Ffname\tSet the full name of the sender\n" \
	"\t-fname\tSets the name of the ``from'' person\n" \
	"\t-rname\tSame as -f\n" \
	"\t-t\tRead message, searching for recipients. To:, Cc: Bcc: lines\n" \
	"\t-v\tGo into verbose mode\n" \
	"\t-V\tSay version and quit"
#else

#include "busybox.h"
#include <unistd.h>
#include <syslog.h>
#include <time.h>
#include <ctype.h>
#include <string.h>
#include <setjmp.h>
#include "bb_sig.h"
#include "bb_inet.h"
#include <netinet/in.h>

#ifdef HAVE_SSL
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#endif

#define Version "2.48"
#define REWRITE_DOMAIN 1
#define SSMTPCONFDIR "/etc/ssmtp"
#define CONFIGURATION_FILE "/etc/ssmtp/ssmtp.conf"
#define REVALIASES_FILE "/etc/ssmtp/revaliases"

#define YES 1
#define NO 0
typedef enum {False = 0, True} Bool;

#define BUF_SZ  (1024 * 2)      /* A pretty large buffer, but not outrageous */

#define MAXWAIT (10 * 60)       /* Maximum wait between commands, in seconds */
#define MEDWAIT (5 * 60)

#define MAXSYSUID 999           /* Highest UID which is a system account */

#ifndef _POSIX_ARG_MAX
#define MAXARGS 4096
#else
#define MAXARGS  _POSIX_ARG_MAX
#endif

#ifndef MAILHUB
#define MAILHUB "mailhost"      /* A surprisingly useful default. */
#endif

CALC_BSS struct BB_COMMON_BSS {
	char *MailHub;                   /* The place to send the mail. */
	int smtp_port;                   /* SMTP port number */
	char HostName[MAXHOSTNAMELEN];   /* Our name, unless overridden. */
#ifdef REWRITE_DOMAIN
	char *RewriteDomain;                    /* Place to claim to be. */
	Bool UseRD;                             /* Do we have to rewrite the domain? */
#endif
	Bool from_override;                     /* Should we use same address in the
			    from-line of the envelope as in the
			    From:-line? */
	char *msg_from;                 /* From:-line from message */
	char *mail_from;                /* From:-line we actually use */
	char *minus_f;                  /* Content of the From: field if specified with the -f option */
	char *minus_F;                  /* Same for -F option */
	char *full_name;                /* Sending user's full name */
	char *Root;                     /* Person to send root's mail to. */
	jmp_buf TimeoutJmpBuf;          /* Timeout waiting for input from network. */
	Bool minus_v;                   /* Tell the user what's happening. */
	Bool minus_t;                   /* Was a T option given? */

	char *RCPT_list[100];
	char **rec;
	Bool from_seen;
#ifdef HASTO_OPTION
	Bool to_seen;
#endif
	Bool date_seen;

	int log_level;

	char *authUsername;             /* user name for SMTP authentication */
	char *authPassword;             /* password for SMTP authentication */

	void *headerp;                  /* Pointer to beginning of headers */

#ifdef HAVE_SSL
	Bool UseTLS;                    /* Use SSL to transfer mail to HUB */
	Bool UseTLSCert;                /* Use a certificate to transfer SSL mail */
	char *TLSCert                   /* Default Certificate */
#endif

	char append_domain_buf[(BUF_SZ + 1)];
	char *newArgV[MAXARGS];

#define ARPADATE_LENGTH 32
	char date_str[ARPADATE_LENGTH];  /* Current date in RFC format. */
} bb_common_bss;

#define MailHub bb_common_bss.MailHub
#define smtp_port bb_common_bss.smtp_port
#define HostName bb_common_bss.HostName
#define RewriteDomain bb_common_bss.RewriteDomain
#define UseRD bb_common_bss.UseRD
#define from_override bb_common_bss.from_override
#define msg_from bb_common_bss.msg_from
#define mail_from bb_common_bss.mail_from
#define minus_f bb_common_bss.minus_f
#define minus_F bb_common_bss.minus_F
#define full_name bb_common_bss.full_name
#define Root bb_common_bss.Root
#define TimeoutJmpBuf bb_common_bss.TimeoutJmpBuf
#define minus_v bb_common_bss.minus_v
#define minus_t bb_common_bss.minus_t
#define RCPT_list bb_common_bss.RCPT_list
#define rec bb_common_bss.rec
#define from_seen bb_common_bss.from_seen
#define to_seen bb_common_bss.to_seen
#define date_seen bb_common_bss.date_seen
#define log_level bb_common_bss.log_level
#define authUsername bb_common_bss.authUsername
#define authPassword bb_common_bss.authPassword
#define headerp bb_common_bss.headerp
#define UseTLS bb_common_bss.UseTLS
#define UseTLSCert bb_common_bss.UseTLSCert
#define TLSCert bb_common_bss.TLSCert
#define append_domain_buf bb_common_bss.append_domain_buf
#define newArgV bb_common_bss.newArgV
#define date_str bb_common_bss.date_str


static void get_arpadate(char *d_string)
{
	static const char * const week_day[] =
		{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
	static const char * const month[] =
		{"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
		"Aug", "Sep", "Oct", "Nov", "Dec"};
	char timezone_offs[6];
	time_t current;
	struct tm *date;
	int offset, gmt_yday, gmt_hour, gmt_min, yday;

	/* Get current time */
	(void) time (&current);

	/* Get GMT and then local dates */
	date = gmtime ((const time_t *) &current);
	gmt_yday = date->tm_yday;
	gmt_hour = date->tm_hour;
	gmt_min = date->tm_min;
	date = localtime ((const time_t *) &current);

	/* Calculates offset */
	offset = ((date->tm_hour - gmt_hour) * 60) + (date->tm_min - gmt_min);

	if ((yday = (date->tm_yday - gmt_yday))) {
		/* This code is optimised to negative timezones (West of Greenwich) */
		if (yday == -1 ||       /* UTC passed midnight before localtime */
				yday > 1)       /* UTC passed new year before localtime */
			offset -= 24 * 60;
		else
			offset += 24 * 60;
	}

	if (offset >= 0)
		sprintf (timezone_offs, "+%02d%02d", offset / 60, offset % 60);
	else
		sprintf (timezone_offs, "-%02d%02d", -offset / 60, -offset % 60);

	sprintf (d_string, "%s, %d %s %04d %02d:%02d:%02d %s",
		week_day[date->tm_wday],
		date->tm_mday, month[date->tm_mon], date->tm_year + 1900,
		date->tm_hour, date->tm_min, date->tm_sec, timezone_offs);
}


static const char base64digits[] =
   "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static void to64frombits(char *out, const char *in_signed, int inlen)
/* raw bytes in quasi-big-endian order to base 64 string (NUL-terminated) */
{
    const unsigned char *in = (const unsigned char *)in_signed;
    for (; inlen >= 3; inlen -= 3) {
	*out++ = base64digits[in[0] >> 2];
	*out++ = base64digits[((in[0] << 4) & 0x30) | (in[1] >> 4)];
	*out++ = base64digits[((in[1] << 2) & 0x3c) | (in[2] >> 6)];
	*out++ = base64digits[in[2] & 0x3f];
	in += 3;
    }
    if (inlen > 0) {
	unsigned char fragment;

	*out++ = base64digits[in[0] >> 2];
	fragment = (in[0] << 4) & 0x30;
	if (inlen > 1)
	    fragment |= in[1] >> 4;
	*out++ = base64digits[fragment];
	*out++ = (inlen < 2) ? '=' : base64digits[(in[1] << 2) & 0x3c];
	*out++ = '=';
    }
    *out = '\0';
}

/* logging */
static void log_event(int priority, char *format, ...)
{
	int type_err_saved = log_stderr;
	int type_log_saved = log_syslog;
	va_list ap;

	log_stderr = 0;
	log_syslog = priority;
	va_start(ap, format);
	bb_verror_msg(format, ap);
	va_end(ap);
	log_syslog = type_log_saved;
	log_stderr = type_err_saved;
}

/* dead_letter() -- save stdin to ~/dead.letter, if possible */
static void dead_letter(void)
{
	char path[(MAXPATHLEN + 1)], buf[(BUF_SZ + 1)];
	struct passwd *pw;
	uid_t uid;
	FILE *fp;
	int r;

	uid = getuid();
	pw = getpwuid(uid);

	if(pw == NULL) {
		/* Far to early to save things. */
		if(log_level > 0)
			log_event(LOG_ERR, "No sender (can't happen), failing horribly");
		return;
	}

	if(isatty(fileno(stdin))) {
		if(log_level > 0)
			log_event(LOG_ERR, "stdin is a TTY - not saving to %s/dead.letter", pw->pw_dir);
		return;
	}

	r = snprintf(path, BUF_SZ, "%s/dead.letter", pw->pw_dir);
	if(r == -1 || r > BUF_SZ) {
		/* can't use die() here since dead_letter() is called from die() */
		exit(1);
	}

	if((fp = fopen(path, "a")) == NULL) {
		/* Perhaps the person doesn't have a homedir... */
		if(log_level > 0)
			log_event(LOG_ERR, "Can't open %s, failing horribly!", path);
		return;
	}

	/* Make sure we start on a new line, */
	/* with a blank line separating messages. */
	(void)fprintf(fp, "\n\n");

	while(fgets(buf, sizeof(buf), stdin))
		(void)fputs(buf, fp);

	if(fclose(fp) == -1) {
		if(log_level > 0)
			log_event(LOG_ERR, "Can't close %s/dead.letter, possibly truncated", pw->pw_dir);
	}
}

static void die(char *format, ...) ATTRIBUTE_NORETURN;
static void die(char *format, ...)
{
	va_list ap;

	va_start(ap, format);
	bb_verror_msg(format, ap);
	va_end(ap);

	/* Send message to dead.letter */
	(void)dead_letter();

	exit(1);
}

static void paq(char *format, ...) ATTRIBUTE_NORETURN;
static void paq(char *format, ...)
{
	va_list ap;

	va_start(ap, format);
	log_syslog = 0;
	bb_verror_msg(format, ap);
	va_end(ap);

	exit(0);
}


/* net stuff */

#ifdef HAVE_SSL
static SSL *ssl;
#endif


/* fd_getc() -- read a character from an fd */
static ssize_t fd_getc(int fd, void *c)
{
#ifdef HAVE_SSL
	if(UseTLS==YES)
		return(SSL_read(ssl, c, 1));
#endif
	return(read(fd, c, 1));
}

/* fd_gets() -- get a characters from a fd instead of an fp */
static char *fd_gets(char *buf, int size, int fd)
{
	int i = 0;
	char c;

	while((i < size) && (fd_getc(fd, &c) == 1)) {
		if(c == '\r');  /* Strip */
		else if(c == '\n') {
			break;
		} else {
			buf[i++] = c;
		}
	}
	buf[i] = '\0';

	return(buf);
}

/* SMTP_read() -- get a line and return the initial digit.
   Deal with continuation lines by reading to the last (non-continuation) line */
static int SMTP_read(int fd, char *response)
{
	do {
		if(fd_gets(response, BUF_SZ, fd) == NULL)
			return(NO);
	} while(response[3] == '-');

	if(log_level > 0)
		log_event(LOG_INFO, "%s\n", response);

	if(minus_v)
		(void)fprintf(stderr, "[<-] %s\n", response);

	return(atoi(response) / 100);
}

/* fd_puts() -- write to fd */
static ssize_t fd_puts(int fd, const void *buf, size_t count)
{
#ifdef HAVE_SSL
	if(UseTLS == YES)
		return(SSL_write(ssl, buf, count));
#endif
	return(write(fd, buf, count));
}

/* SMTP_write() -- a printf to an fd, which appends TCP/IP <CR/LF> */
static void SMTP_write(int fd, char *format, ...)
{
	char buf[(BUF_SZ + 1)];
	va_list ap;

	va_start(ap, format);
	if(vsnprintf(buf, (BUF_SZ - 2), format, ap) == -1)
		die("Buffer error in SMTP_write()");
	va_end(ap);

	if(log_level > 0)
		log_event(LOG_INFO, "%s\n", buf);

	if(minus_v)
		(void)fprintf(stderr, "[->] %s\n", buf);

	(void)strcat(buf, "\r\n");
	(void)fd_puts(fd, buf, strlen(buf));
}

static int SMTP_open(char *host, int port)
{
	struct sockaddr name;
	int s;
#ifdef HAVE_SSL
	int err;

	/* Init SSL stuff */
	SSL_CTX *ctx;
	SSL_METHOD *meth;
	X509 *server_cert;

	SSL_load_error_strings();
	SSLeay_add_ssl_algorithms();
	meth=SSLv23_client_method();
	ctx = SSL_CTX_new(meth);
	if(!ctx) {
		log_event(LOG_ERR, "No SSL support inited");
		return(-1);
	}

	if(UseTLSCert==YES) {
		if(SSL_CTX_use_certificate_chain_file(ctx, TLSCert) <= 0) {
			bb_perror_msg("Use certfile");
			return(-1);
		}

		if(SSL_CTX_use_PrivateKey_file(ctx, TLSCert, SSL_FILETYPE_PEM) <= 0) {
			bb_perror_msg("Use PrivateKey");
			return(-1);
		}

		if(!SSL_CTX_check_private_key(ctx)) {
			log_event(LOG_ERR, "Private key does not match the certificate public key");
			return(-1);
		}
	}
#endif

	/* Check we can reach the host */
	bb_gethostbyname(BB_GHBN_SOCKADDR | BB_GHBN_ERR_DIE, host, &name);
	((struct sockaddr_in *)&name)->sin_port = htons(port);
	s = xconnect(&name);

#ifdef HAVE_SSL
	if(UseTLS == YES) {
		log_event(LOG_INFO, "Creating SSL connection to host");

		ssl = SSL_new(ctx);
		if(!ssl) {
			log_event(LOG_ERR, "SSL not working");
			return(-1);
		}
		SSL_set_fd(ssl, s);

		err = SSL_connect(ssl);
		if(err < 0) {
			bb_perror_msg("SSL_connect");
			return(-1);
		}

		if(log_level > 0)
			log_event(LOG_INFO, "SSL connection using %s", SSL_get_cipher(ssl));

		server_cert = SSL_get_peer_certificate(ssl);
		if(!server_cert)
			return(-1);
		X509_free(server_cert);

		/* TODO: Check server cert if changed! */
	}
#endif

	return(s);
}

/* SMTP_OK() -- get a line and test the three-number string at the beginning.  If it starts with a 2, it's OK */
static int SMTP_OK(int fd, char *response)
{
	return(SMTP_read(fd, response) == 2) ? YES : NO;
}


/* get_addr() -- parse user@domain,com from email address */
static void get_addr(char *str, char *buf, size_t sz)
{
	char *p, *sd;
	size_t i;

	i = 0;
	if((p = strchr(str, '<'))) {
		p++;
		while(*p && (*p != '>') && (i < sz))
			buf[i++] = *p++;
		buf[i] = '\0';
	} else {
		p = str;
		while(*p && (i < sz)) {
			switch(*p) {
				case '\n':
				case '\t':
						buf[i] = ' ';

				case '(':
						while(*p && (*p != ')')) p++;
						p++;
						continue;
/*
				case '"':
						while(*p && (*p != '"')) p++;
						p++;
						continue;
*/
				default:
						buf[i] = *p++;
			}
			i++;
		}
		buf[i] = '\0';

		/* Strip leading spaces */
		if((sd = strdup(buf)) == NULL)
			die("strup() failed in get_addr()");
		while(*sd && isspace(*sd)) sd++;

		if(strncpy(buf, sd, sz) == NULL)
			die("strncpy() failed in get_addr()");

		/* Strip trailing spaces */
		i = strlen(buf);
		while(i-- > 0) {
			if(isspace(buf[i]) == 0) break;
		}
		buf[(i + 1)] = '\0';
	}
}

/* RCPT_record() -- Record the email RCPT addresses */
static void RCPT_record(char *str)
{
	char buf[(BUF_SZ + 1)], *p;
	Bool in_quotes = False;

	/* Convert <CR> and <TAB> to space */
	p = str;
	while(*p) {
		if((*p == '\n') || (*p == '\t')) {
			*p = ' ';
		}
		p++;
	}
	p = str;

	while(*p) {
		if(*p == '"') {
			if(in_quotes == False)
				in_quotes = True;
			 else
				in_quotes = False;
		}

		if(((*p == ',') || (*(p + 1) == '\0')) && (!in_quotes)) {
			if(*(p + 1) != '\0') {
				*p = '\0';
			}

			while(*str && isspace(*str)) str++;
			get_addr(str, buf, BUF_SZ);

			str = (p + 1);

			if(strlen(buf) > 0) {
				if((rec - RCPT_list) > (100 - 1))
					die("Too many recipients (>= 100)");
			}

			if((*rec++ = strdup(buf)) == NULL)
				die("strdup() failed in X()");
		}
		p++;
	}
}

/* record_headers() -- note which ones we've seen */
static void record_headers(char *str)
{
	if(strncasecmp(str, "From:", 5) == 0)
		from_seen = True;
#ifdef HASTO_OPTION
	else if(strncasecmp(str, "To:" ,3) == 0)
		to_seen = True;
#endif
	else if(strncasecmp(str, "Date:", 5) == 0)
		date_seen = True;

	if(minus_t) {
		/* Need to figure out recipients from the e-mail */
		if(strncasecmp(str, "To:", 3) == 0)
			RCPT_record((str + 4));
		else if(strncasecmp(str, "Bcc:", 4) == 0)
			RCPT_record((str + 5));
		else if(strncasecmp(str, "CC:", 3) == 0)
			RCPT_record((str + 4));
	}
}

/* append_domain() -- Fix up address with @domain.name */
static char *append_domain(char *str)
{
#ifdef REWRITE_DOMAIN
	if(strchr(str, '@') == NULL) {
		int r = snprintf(append_domain_buf, BUF_SZ, "%s@%s", str, (UseRD == True) ? RewriteDomain : HostName);
		if(r == -1 || r > BUF_SZ)
			die("Buffer error in append_domain()");
		return(append_domain_buf);
	}
#endif

	return(str);
}

/* strip_from() -- transforms "Name <login@host>" into "login@host" or "login@host (Real name)" */
static char *strip_from(char *str)
{
	char buf[(BUF_SZ + 1)], *sd;

	if(strncmp("From: ", str, 6) == 0)
		str += 6;

	/* Remove the real name if necessary - just send the address */
	get_addr(str, buf, BUF_SZ);

	if((sd = strdup(buf)) == NULL)
		die("strdup() failed in strip_from()");
	return(sd);
}

#ifdef REWRITE_DOMAIN
/* fix_from() -- replace whole From: header with standardised pattern
		Evil, nasty, immoral header-rewriting code (:-)) */
static void fix_from(char *buf, int sz)
{
	if(strncasecmp(buf, "From:", 5) == 0) {
		if(from_override) {
			if((msg_from = strdup(buf)) == NULL)
				die("strdup() failed in fix_from()");
		} else {
			int r = snprintf(buf, sz, "From: %s", mail_from);
			if(r == -1 || r > sz)
				die("Buffer error in fix_from()");
		}
	}
}
#endif

/* RCPT_remap() -- alias systems-level users to the person who
	reads their mail.  This is variously the owner of a workstation,
	the sysadmin of a group of stations and the postmaster otherwise.
	We don't just mail stuff off to root on the mailhub (:-)) */
static char *RCPT_remap(char *str)
{
	struct passwd *pw;

	if(strchr(str, '@') ||
		((pw = getpwnam(str)) == NULL) || (pw->pw_uid > MAXSYSUID)) {
		/* It's not a local systems-level user */
		return(append_domain(str));
	} else {
		return(append_domain(Root));
	}
}

/* do_from() -- generate a from line in standard format
 ("Real Name <id@site>" or "id@site") based on local user and host names */
static void do_from(struct passwd *pw)
{
	char buf[(BUF_SZ + 1)];

	if(full_name) {
		int r = snprintf(buf, BUF_SZ, "%s <%s@%s>", full_name, pw->pw_name,
#ifdef REWRITE_DOMAIN
			UseRD == True ? RewriteDomain : HostName
#else
			HostName
#endif
			);
		if(r == -1 || r > BUF_SZ)
			die("Buffer error in do_from()");
	} else {
		int r = snprintf(buf, BUF_SZ, "%s@%s", pw->pw_name,
#ifdef REWRITE_DOMAIN
			UseRD == True ? RewriteDomain : HostName
#else
			HostName
#endif
			);
		if(r == -1 || r > BUF_SZ)
			die("Buffer error in do_from()");
	}

	if((mail_from = strdup(buf)) == NULL)
		die("strdup() failed in X()");
}

/* handler() -- a ``normal'' non-portable version of an alarm handler
      Alas, setting a flag and returning is not fully functional in
      BSD: system calls don't fail when reading from a ``slow'' device
      like a socket. So we longjump instead, which is erronious on
      a small number of machines and ill-defined in the language */
static void handler(int sig)
{
	(void)&sig;
	longjmp(TimeoutJmpBuf, (int)1);
}

/* standardise() -- trim off '\n's, double leading dots */
static void standardise(char *str)
{
	size_t sl;

	strchr_set0(str, '\n');

	/* Any line beginning with a dot has an additional dot inserted;
	not just a line consisting solely of a dot.  Thus we have to
	slide the buffer down one.  */
	sl = strlen(str);

	if(*str == '.') {
		if((sl + 2) > BUF_SZ)
			die("Buffer overflow in standardise()");
		(void)memmove((str + 1), str, (sl + 1));  /* Copy trailing \0 */

		*str = '.';
	}
}

#include "config_read.h"

/* Config file access routines */
static int read_config(void)
{
	config_read_t *crt;
	char *p, *q, *r;

	if((crt = config_read_prepare(CONFIGURATION_FILE, 0)) != NULL)
	    return(False);


	while( (crt = config_read(crt, "#\1\2\0\2=")) != NULL) {
		p = crt->tokens[0];
		q = crt->tokens[1];

		if(strcasecmp(p, "Root") == 0) {
			if((Root = strdup(q)) == NULL)
				die("strdup() failed in parse_config()");

			if(log_level > 0)
				log_event(LOG_INFO, "Set Root=\"%s\"", Root);
		} else if(strcasecmp(p, "MailHub") == 0) {
			if((MailHub = strdup(q)) == NULL)
				die("strdup() failed in parse_config()");
			r = strchr_set0(q, ':');
			if(r != NULL) {
				smtp_port = atoi(r+1);
				if(log_level > 0)
					log_event(LOG_INFO, "Set RemotePort=\"%d\"", smtp_port);
			}

			if(log_level > 0)
				log_event(LOG_INFO, "Set MailHub=\"%s\"", MailHub);
		} else if(strcasecmp(p, "HostName") == 0) {
			if(strncpy(HostName, q, MAXHOSTNAMELEN) == NULL)
				die("strncpy() failed in X()");

			if(log_level > 0)
				log_event(LOG_INFO, "Set HostName=\"%s\"", HostName);
		}
#ifdef REWRITE_DOMAIN
		else if(strcasecmp(p, "RewriteDomain") == 0) {
			if((RewriteDomain = strdup(q)) == NULL)
				die("strdup() failed in parse_config()");

			UseRD = True;
			if(log_level > 0)
				log_event(LOG_INFO, "Set RewriteDomain=\"%s\"", RewriteDomain);
		}
#endif
		else if(strcasecmp(p, "FromLineOverride") == 0) {
			if(strcasecmp(q, "yes") == 0)
				from_override = True;
			 else
				from_override = False;

			if(log_level > 0)
				log_event(LOG_INFO, "Set FromLineOverride=\"%s\"", from_override ? "True" : "False");
		} else if(strcasecmp(p, "RemotePort") == 0) {
			smtp_port = atoi(q);

			if(log_level > 0)
				log_event(LOG_INFO, "Set RemotePort=\"%d\"", smtp_port);
		}
#ifdef HAVE_SSL
		else if(strcasecmp(p, "UseTLS") == 0) {
			if(strcasecmp(q, "yes") == 0)
				UseTLS = True;
			else
				UseTLS = False;

			if(log_level > 0)
				log_event(LOG_INFO, "Set UseTLS=\"%s\"", UseTLS ? "True" : "False");
		} else if(strcasecmp(p, "UseTLSCert") == 0) {
			if(strcasecmp(q, "yes") == 0)
				UseTLSCert = True;
			else
				UseTLSCert = False;

			if(log_level > 0)
				log_event(LOG_INFO, "Set UseTLSCert=\"%s\"", UseTLSCert ? "True" : "False");
		} else if(strcasecmp(p, "TLSCert") == 0) {
			if((TLSCert = strdup(q)) == NULL)
				die("strdup() failed in parse_config()");

			if(log_level > 0)
				log_event(LOG_INFO, "Set TLSCert=\"%s\"", TLSCert);
		}
#endif
		else {
			log_event(LOG_INFO, "Unable to set %s=\"%s\"", p, q);
		}
	}
	return(True);
}

/* rev_aliases() -- parse the reverse alias file, fix globals to use any entry for sender */
static void rev_aliases(struct passwd *pw)
{
	char *p, *r;
	config_read_t *crt;

	/* Try to open the reverse aliases file */
	crt = config_read_prepare(REVALIASES_FILE, 0);
	/* Search if a reverse alias is definied for the sender */
	/* format: local_account:outgoing_address:mailhub[:port] */
	while( (crt = config_read(crt, "#\1\2\3\4\0\0:")) != NULL) {
		p = crt->tokens[0];
		r = crt->tokens[1];
		if(r == NULL) continue;

		/* Parse the alias */
		if(!strcmp(p, pw->pw_name)) {
			if(full_name)
				mail_from = xasprintf("\"%s\" <%s>", full_name, r);
			 else
				if((mail_from = strdup(r)) == NULL)
					die("strdup() failed in X()");

			r = crt->tokens[2];
			if(r == NULL) break;

			MailHub = strdup(r);
			if(log_level > 0)
				log_event(LOG_INFO, "Set MailHub=\"%s\"", MailHub);
			r = crt->tokens[3];
			if(r == NULL) break;
			smtp_port = atoi(r);
			if(log_level > 0)
				log_event(LOG_INFO, "via SMTP Port Number=\"%d\"", smtp_port);
			break;
		}
	}
	config_close(crt);
}

/* doOptions() -- pull the options out of the command-line, process them
   (and special-case calls to mailq, etc), and return the rest */
static char **doOptions(int argc, char *argv[])
{
	int i, j, add, newArgC;

	newArgV[0] = argv[0];
	newArgC = 1;

	i = 1;
	while(i < argc) {
		if(argv[i][0] != '-') {
			newArgV[newArgC++] = argv[i++];
			continue;
		}
		j = 0;

		add = 1;
		while(argv[i][++j] != '\0') {
			switch(argv[i][j]) {
			case 'a':
				switch(argv[i][++j]) {
				case 'u':
					if(!argv[i][j+1]) {
						authUsername = strdup(argv[i+1]);
						add++;
					} else {
						authUsername = strdup(argv[i]+j+1);
					}
					goto exit;

				case 'p':
					if(!argv[i][j+1]) {
						authPassword = strdup(argv[i+1]);
						add++;
					} else {
						authPassword = strdup(argv[i]+j+1);
					}
					goto exit;
				}
				goto exit;

			case 'b':
				switch(argv[i][++j]) {

				case 'a':       /* ARPANET mode. */
						paq("-ba is not supported");
				case 'd':       /* Run as a daemon. */
						paq("-bd is not supported");
				case 'i':       /* Initialise aliases. */
						paq("Aliases are not used");
				case 'm':       /*  Default addr processing. */
						continue;
				case 'p':       /* Print mailqueue. */
						paq("Mail queue is empty");
				case 's':       /* Read SMTP from stdin. */
						paq("-bs is not supported");
				case 't':       /* Test mode. */
						paq("-bt is meaningless");
				case 'v':       /*  Verify names only. */
						paq("-bv is meaningless");
				case 'z':       /* Create  freeze file. */
						paq("-bz is meaningless");
				}

			case 'C':       /* Configfile name. */
					goto exit;
			case 'd':       /* Debug */
					log_level = 1;
					minus_v = True; /* Almost the same thing... */
					continue;
			case 'E':       /* insecure channel, don't trust userid. */
					continue;
			case 'R':
				if(!argv[i][j+1]) {     /* amount of the message to be returned */
					add++;
					goto exit;
				} else {  /* Process queue for recipient */
					continue;
				}
			case 'F':       /* Fullname of sender */
				if(!argv[i][(j + 1)]) {
					minus_F = strdup(argv[(i + 1)]);
					add++;
				} else {
					minus_F = strdup(argv[i]+j+1);
				}
				goto exit;
			case 'f':       /* Set from/sender address. */
			case 'r':       /* Obsolete -f flag. */
				if(!argv[i][(j + 1)]) {
					if((minus_f = strdup(argv[(i + 1)])) == NULL)
						die("strdup() failed in X()");
					add++;
				} else {
					if((minus_f = strdup(argv[i]+j+1)) == NULL)
						die("strdup failed in X()");
				}

				strcpy(minus_f, strip_from(minus_f));
				goto exit;
			case 'h':       /* Set hopcount. */
				continue;
			case 'm':       /* Ignore originator in adress list */
				continue;
			case 'M':       /* Use specified message-id */
				goto exit;
			case 'N':       /* dsn options */
				add++;
				goto exit;
			case 'n':       /* No aliasing */
				continue;
			case 'o':
				switch(argv[i][++j]) {

				case 'A':       /* Alternate aliases file. */
						goto exit;
				case 'c':       /* Delay connections. */
						continue;
				case 'D':       /* Run newaliases if rqd. */
						paq("Aliases are not used");
				case 'd':       /* Deliver now, in background or queue. */
						/* This may warrant a diagnostic for b or q. */
						continue;
				case 'e':       /* Errors: mail, write or none. */
						j++;
						continue;
				case 'F':       /* Set tempfile mode. */
						goto exit;
				case 'f':       /* Save ``From ' lines. */
						continue;
				case 'g':       /* Set group id. */
						goto exit;
				case 'H':       /* Helpfile name. */
						continue;
				case 'i':       /* DATA ends at EOF, not \n.\n */
						continue;
				case 'L':       /* Log level. */
						goto exit;
				case 'm':       /* Send to me if in the list. */
						continue;
				case 'o':       /* Old headers, spaces between adresses. */
						paq("-oo is not supported");
				case 'Q':       /* Queue dir. */
						goto exit;
				case 'r':       /* Read timeout. */
						goto exit;
				case 's':       /* Always init the queue. */
						continue;
				case 'S':       /* Stats file. */
						goto exit;
				case 'T':       /* Queue timeout. */
						goto exit;
				case 't':       /* Set timezone. */
						goto exit;
				case 'u':       /* Set uid. */
						goto exit;
				case 'v':       /* Set verbose flag. */
						minus_v = True;
						continue;
				}
				break;
			case 'q':       /* Process the queue [at time] */
					paq("Mail queue is empty");
			case 't':       /* Read message's To/Cc/Bcc lines. */
					minus_t = True;
					continue;
			case 'v':       /* minus_v (ditto -ov). */
					minus_v = True;
					break;
			case 'V':       /*  Say version and quit. */
					/* Similar as die, but no logging */
					paq("sSMTP-%s (not sendmail at all)", Version);
			}
		}
		exit:
		i += add;
	}
	newArgV[newArgC] = NULL;

	if(newArgC <= 1 && !minus_t)
		paq("No recipients supplied, mail will not be sent");

	if(newArgC > 1 && minus_t)
		paq("recipients with -t option not supported");

	return(&newArgV[0]);
}

/* ssmtp() -- send the message (exactly one) from stdin to the mailhub SMTP port */
static int ssmtp(char *argv[])
{
	char buf[(BUF_SZ + 1)], *p, *q, *sd;
	size_t sl, hs, header_size = 0;
	struct passwd *pw;
	int i, sock;
	uid_t uid;

	uid = getuid();
	if((pw = getpwuid(uid)) == NULL)
		die("Could not find password entry for sender (UID %d)", uid);
	get_arpadate(date_str);

	if(read_config() == False)
		log_event(LOG_INFO, "No ssmtp.conf in %s", SSMTPCONFDIR);

#ifdef HAVE_GECOS
	p = strtok(pw->pw_gecos, ";,");
	if(p)
		full_name = strdup(p);
#endif
	rev_aliases(pw);

	if(mail_from == NULL)
		do_from(pw);

	headerp = NULL;
	while(fgets(buf, BUF_SZ, stdin) != NULL) {
		if(*buf == '\n') break;

		sl = strlen(buf);
		hs = header_size;
		header_size += sl;
		headerp = realloc(headerp, header_size);
		(void)memcpy((headerp + hs), buf, sl);
	}
	headerp = realloc(headerp, (header_size + 1));

	p = (char *)headerp;
	p[header_size] = '\0';

	q = buf;
	while(*p) {
		if(*p == '\n') {
			switch(*++p) {
				case ' ':
				case '\t':      continue;

				default:
						*q = '\0';

					/* Trim off \n, double leading .'s */
						standardise(buf);

						record_headers(buf);

#ifdef REWRITE_DOMAIN
						fix_from(buf, BUF_SZ);
#endif
						q = buf;
			}
		}
		*q++ = *p++;
	}
	/* Finished header processing */

	if(minus_t) {
		/* Sorry no support for the exclusions on the command line */
		/*      argv=recipients;       */
		*rec++ = NULL;
	}

	/* Now to the delivery of the message */
	(void)signal(SIGALRM, handler);         /* Catch SIGALRM */
	(void)alarm((unsigned) MAXWAIT);        /* Set initial timer */
	if(setjmp(TimeoutJmpBuf) != 0) {
		/* Then the timer has gone off and we bail out. */
		die("Connection lost in middle of processing");
	}

	if((sock = SMTP_open(MailHub, smtp_port)) == -1)
		die("Cannot open %s:%d", MailHub, smtp_port);
	else if(SMTP_OK(sock, buf) == False)
		die("Didn't get initial OK message from SMTP server");

	/* If user supplied username and password, then try ELHO */
	/* do not really know if this is required or not...      */
	if(authUsername)
		SMTP_write(sock, "EHLO %s", HostName);
	else
		SMTP_write(sock, "HELO %s", HostName);
	(void)alarm((unsigned) MEDWAIT);

	if(SMTP_OK(sock, buf) == False) die("%s (%s)", buf, HostName);

	/* Try to log in if username was supplied */
	if(authUsername) {
		to64frombits(buf, authUsername, strlen(authUsername));
		SMTP_write(sock, "AUTH LOGIN %s", buf);

		(void)alarm((unsigned) MEDWAIT);
		if(SMTP_read(sock, buf) != 3)
			die("Server didn't accept AUTH LOGIN (%s)", buf);

		to64frombits(buf, authPassword, strlen(authPassword));
		SMTP_write(sock, "%s", buf);

		(void)alarm((unsigned) MEDWAIT);
		if(SMTP_OK(sock, buf) == False)
			die("Authorization failed (%s)", buf);
	}

	/* Send "MAIL FROM:" line */
	if(msg_from && from_override)
		(void)snprintf(buf, BUF_SZ, "MAIL FROM:<%s>", append_domain(strip_from(msg_from)));
	else
		(void)snprintf(buf, BUF_SZ, "MAIL FROM:<%s>", (minus_f != NULL) ? append_domain(minus_f) : strip_from(mail_from));

	SMTP_write(sock, buf);
	(void)alarm((unsigned) MEDWAIT);

	if(SMTP_OK(sock, buf) == NO) die("%s", buf);

	/* Send all the To: adresses. */
	/* Either we're using the -t option, or we're using the arguments */
	if(minus_t) {
		if(*RCPT_list == NULL)
			die("No recipients specified although -t option used");

		i = 0;
		do {
			sd = strdup(RCPT_remap(RCPT_list[i]));
			SMTP_write(sock, "RCPT TO:<%s>", sd);
			(void)alarm((unsigned)MEDWAIT);

			if(SMTP_OK(sock, buf) == NO)
				die("RCPT TO:<%s> (%s)", sd, buf);
		}
		while(RCPT_list[++i]);
	} else {
		for(i = 1; argv[i] != NULL; i++) {
			p = strtok(argv[i], ",");
			while(p) {
				/* RFC822 Address  -> "foo@bar" */
				get_addr(p, buf, BUF_SZ);

				sd = strdup(RCPT_remap(buf));
				SMTP_write(sock, "RCPT TO:<%s>", sd);
				(void)alarm((unsigned) MEDWAIT);

				if(SMTP_OK(sock, buf) == NO)
					die("RCPT TO:<%s> (%s)", sd, buf);
				p = strtok(NULL, ",");
			}
		}
	}

	/* Send DATA. */
	SMTP_write(sock, "DATA");
	(void)alarm((unsigned) MEDWAIT);

	if(SMTP_read(sock, buf) != 3) {
		/* Oops, we were expecting "354 send your data" */
		die("%s", buf);
	}

	SMTP_write(sock, "Received: by %s (sSMTP sendmail emulation); %s", HostName, date_str);

	if(from_seen == False) {
		/* No From: line; add the sender specified in the command line or generate one */
		if(minus_F == NULL)
			SMTP_write(sock, "From: %s", (minus_f == NULL) ? mail_from : minus_f);
		else
			SMTP_write(sock, "From: %s <%s>", minus_F, (minus_f == NULL) ? strip_from(mail_from) : minus_f);
	}

	if(date_seen == False)
		SMTP_write(sock, "Date: %s", date_str);

#ifdef HASTO_OPTION
	if(to_seen == False)
		SMTP_write(sock, "To: postmaster");
#endif

	p = (char *)headerp;
	while(*p) {
		i = 0;
		while(*p && (*p != '\n') && (i < BUF_SZ))
			buf[i++] = *p++;
		buf[i] = '\0';

		if(*p) p++;

		SMTP_write(sock, "%s", buf);
	}
	free(headerp);
	(void)alarm((unsigned) MEDWAIT);

	/* End of headers, start body. */
	SMTP_write(sock, "");

	while(fgets(buf, sizeof(buf), stdin) != NULL) {
		/* Trim off \n, double leading .'s */
		standardise(buf);

		SMTP_write(sock, "%s", buf);
		(void)alarm((unsigned) MEDWAIT);
	}
	/* End of body. */

	SMTP_write(sock, ".");
	(void)alarm((unsigned) MAXWAIT);

	if(SMTP_OK(sock, buf) == NO) die("%s", buf);

	/* Close conection. */
	(void)signal(SIGALRM, SIG_IGN);
	SMTP_write(sock, "QUIT");
	(void)SMTP_OK(sock, buf);
	(void)close(sock);

	log_event(LOG_INFO, "Sent mail for %s (%s)", strip_from(mail_from), buf);

	return(0);
}

/* main() -- make the program behave like sendmail, then call ssmtp */
int sendmail_main(void)
{
	int argc = Gargc;
	char **argv = Gargv;
	char **newArgv;

#ifdef DEBUG
	log_level = 1;
#endif

	bb_openlog(LOG_PID, LOG_MAIL);
	log_syslog = LOG_ERR;      /* and stderr by default */

	/* Try to be bulletproof :-) */
	bb_signal((1<<SIGHUP) | (1<<SIGINT) | (1<<SIGTTIN) | (1<<SIGTTOU), SIG_IGN);

	if(gethostname(HostName, sizeof(HostName)) == -1)
		die("Cannot get the name of this machine");

	/* set the DATA segment */
	MailHub = MAILHUB;
	smtp_port = 25;
	RewriteDomain = "localhost";
	Root = "postmaster";
	rec = RCPT_list;
#ifdef HAVE_SSL
	TLSCert = "/etc/ssl/certs/ssmtp.pem";
#endif

	newArgv = doOptions(argc, argv);

	return ssmtp(newArgv);
}
#endif /* APPLET_USAGE */
