/*
 * TFTP User Program
 *
 *
 * Copyright (c) 1983, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * Many bug fixes are from Jim Guyton <guyton@rand-unix>
 *
 * busybox port by Vladimir Oleynik <dzo@simtreas.ru> (C) 2009
 * - size optimization, memory usage reducind
 */
#ifdef APPLET_USAGE
#define tftp_trivial_usage \
	"[-v][-l][-m mode] [host [port]] [-c command]"
#define tftp_full_usage \
	"\t-c\t\tcommand Execute command as if it had been entered on the\n" \
	"\t\t\ttftp prompt. Must be specified last on the command line.\n" \
	"\t-l\t\tDefault to literal mode. Used to avoid special proces-\n" \
	"\t\t\tsing of ':' in a file name.\n" \
	"\t-m mode\t\tSet the default transfer mode to mode. This is\n" \
	"\t\t\tusually used with -c.\n" \
	"\t-R port:port\tForce the originating port number to be in\n" \
	"\t\t\tthe specified range of port numbers.\n" \
	"\t-v\t\tDefault to verbose mode.\n" \
	"\t-V\t\tPrint the version number and configuration\n" \
	"\t\t\tto standard output, then exit gracefully."
#else

#include "busybox.h"
#include "tftp_common.h"
#include <sys/file.h>
#include <ctype.h>

#ifdef CONFIG_FEATURE_COMMAND_EDITING
#include "cmdedit.h"
#define WITH_READLINE_STR ", with readline"
#else
#define WITH_READLINE_STR ", without readline"
#endif

#define TFTP_CONFIG_STR VERSION WITH_READLINE_STR

#define TFTP_PKTSIZE    SEGSIZE+4

#define LBUFLEN         BUFSIZ  /* size of input buffer */

#define TIMEOUT         5               /* secs between rexmt's */

CALC_BSS struct BB_COMMON_BSS {
	/* to stubs */
	struct bf bfs[2];
	int nextone;
	int current;
	int newline;
	int tftp_segsize;
	int prevchar;       /* putbuf: previous char (cr check) */
	/* ******** */
	char *hostname;
	struct  sockaddr_in peeraddr;
	int     f;
	u_short port;
	int     trace;
	int     verbose;
	int     literal;
	int     connected;
	const struct modes *mode;
	char    line[LBUFLEN];
	int     margc;
	char    *margv[20];
	sigjmp_buf      toplevel;
	struct  servent *sp;
	int portrange;
	unsigned int portrange_from;
	unsigned int portrange_to;
	char    ackbuf[TFTP_PKTSIZE];
	int     timeout;
	sigjmp_buf      timeoutbuf;
	struct timeval tstart;
	struct timeval tstop;
	int maxtimeout;
	int rexmtval;
} bb_common_bss;
#define hostname bb_common_bss.hostname
#define peeraddr bb_common_bss.peeraddr
#define f bb_common_bss.f
#define port bb_common_bss.port
#define trace bb_common_bss.trace
#define verbose bb_common_bss.verbose
#define literal bb_common_bss.literal
#define connected bb_common_bss.connected
#define mode bb_common_bss.mode
#define line bb_common_bss.line
#define margc bb_common_bss.margc
#define margv bb_common_bss.margv
#define toplevel bb_common_bss.toplevel
#define sp bb_common_bss.sp
#define portrange bb_common_bss.portrange
#define portrange_from bb_common_bss.portrange_from
#define portrange_to bb_common_bss.portrange_to
#define ackbuf bb_common_bss.ackbuf
#define timeout bb_common_bss.timeout
#define timeoutbuf bb_common_bss.timeoutbuf
#define tstart bb_common_bss.tstart
#define tstop bb_common_bss.tstop
#define maxtimeout bb_common_bss.maxtimeout
#define rexmtval bb_common_bss.rexmtval

static const char prompt[] = "tftp> ";


static int
makerequest(int request, const char *name,
	    struct tftphdr *tp, const char *req_mode)
{
	char *cp;

	tp->th_opcode = htons((u_short)request);
	cp = (char *) &(tp->th_stuff);
	strcpy(cp, name);
	cp += strlen(name);
	*cp++ = '\0';
	strcpy(cp, req_mode);
	cp += strlen(req_mode);
	*cp++ = '\0';
	return (cp - (char *)tp);
}


static void
printstats(const char *direction, unsigned long amount)
{
	double delta;

	delta = (tstop.tv_sec+(tstop.tv_usec/100000.0)) -
		(tstart.tv_sec+(tstart.tv_usec/100000.0));
	if (verbose) {
	  printf("%s %lu bytes in %.1f seconds", direction, amount, delta);
	  printf(" [%.0f bit/s]", (amount*8.)/delta);
	  putchar('\n');
	}
}


static void
startclock(void)
{
	(void)gettimeofday(&tstart, NULL);
}

static void
stopclock(void)
{

	(void)gettimeofday(&tstop, NULL);
}

static void
timer(int sig)
{
	int save_errno = errno;

	(void)sig;              /* Shut up unused warning */

	timeout += rexmtval;
	if (timeout >= maxtimeout) {
		printf("Transfer timed out.\n");
		errno = save_errno;
		siglongjmp(toplevel, -1);
	}
	errno = save_errno;
	siglongjmp(timeoutbuf, 1);
}

static const char * const errmsgs[] =
{
  "Undefined error code",                       /* 0 - EUNDEF */
  "File not found",                             /* 1 - ENOTFOUND */
  "Access denied",                              /* 2 - EACCESS */
  "Disk full or allocation exceeded",           /* 3 - ENOSPACE */
  "Illegal TFTP operation",                     /* 4 - EBADOP */
  "Unknown transfer ID",                        /* 5 - EBADID */
  "File already exists",                        /* 6 - EEXISTS */
  "No such user",                               /* 7 - ENOUSER */
  "Failure to negotiate RFC2347 options"        /* 8 - EOPTNEG */
};
#define ERR_CNT (sizeof(errmsgs)/sizeof(const char *))

static void
tpacket(const char *s, struct tftphdr *tp, int n)
{
	static const char * const opcodes[] =
	   { "#0", "RRQ", "WRQ", "DATA", "ACK", "ERROR", "OACK" };
	char *cp, *file;
	u_short op = ntohs((u_short)tp->th_opcode);

	if (op < RRQ || op > ERROR)
		printf("%s opcode=%x ", s, op);
	else
		printf("%s %s ", s, opcodes[op]);
	switch (op) {

	case RRQ:
	case WRQ:
		n -= 2;
		file = cp = (char *) &(tp->th_stuff);
		cp = strchr(cp, '\0');
		printf("<file=%s, mode=%s>\n", file, cp + 1);
		break;

	case DATA:
		printf("<block=%d, %d bytes>\n", ntohs(tp->th_block), n - 4);
		break;

	case ACK:
		printf("<block=%d>\n", ntohs(tp->th_block));
		break;

	case ERROR:
		printf("<code=%d, msg=%s>\n", ntohs(tp->th_code), tp->th_msg);
		break;
	}
}


/*
 * Send a nak packet (error message).
 * Error code passed in is one of the
 * standard TFTP codes, or a UNIX errno
 * offset by 100.
 */
static void
nak(int error, const char *msg)
{
  struct tftphdr *tp;
  int length;

  tp = (struct tftphdr *)ackbuf;
  tp->th_opcode = htons((u_short)ERROR);
  tp->th_code = htons((u_short)error);

  if ( error >= 100 ) {
    /* This is a Unix errno+100 */
    if ( !msg )
      msg = strerror(error - 100);
    error = EUNDEF;
  } else {
    if ( (unsigned)error >= ERR_CNT )
      error = EUNDEF;

    if ( !msg )
      msg = errmsgs[error];
  }

  tp->th_code = htons((u_short)error);

  length = strlen(msg)+1;
  memcpy(tp->th_msg, msg, length);
  length += 4;                  /* Add space for header */

  if (trace)
    tpacket("sent", tp, length);
  if (sendto(f, ackbuf, length, 0, (struct sockaddr *)&peeraddr,
	     sizeof(peeraddr)) != length)
    bb_perror_msg("nak");
}

/*
 * Send the requested file.
 */
static void
tftp_sendfile(int fd, const char *name, const char *req_mode)
{
	struct tftphdr *ap;        /* data and ack packets */
	struct tftphdr *dp;
	int n;
	volatile int is_request;
	volatile u_short block;
	volatile int size, convert;
	volatile off_t amount;
	struct sockaddr_in from;
	socklen_t fromlen;
	FILE *file;
	u_short ap_opcode, ap_block;

	startclock();           /* start stat's clock */
	dp = tftp_r_init();     /* reset fillbuf/read-ahead code */
	ap = (struct tftphdr *)ackbuf;
	convert = !strcmp(req_mode, "netascii");
	file = fdopen(fd, convert ? "rt" : "rb");
	block = 0;
	is_request = 1;         /* First packet is the actual WRQ */
	amount = 0;

	signal(SIGALRM, timer);
	do {
		if (is_request) {
			size = makerequest(WRQ, name, dp, req_mode) - 4;
		} else {
		/*      size = read(fd, dp->th_data, SEGSIZE);   */
			size = tftp_readit(file, &dp, convert);
			if (size < 0) {
				nak(errno + 100, NULL);
				break;
			}
			dp->th_opcode = htons((u_short)DATA);
			dp->th_block = htons((u_short)block);
		}
		timeout = 0;
		(void) sigsetjmp(timeoutbuf,1);

		if (trace)
			tpacket("sent", dp, size + 4);
		n = sendto(f, dp, size + 4, 0,
		    (struct sockaddr *)&peeraddr, sizeof(peeraddr));
		if (n != size + 4) {
			bb_perror_msg("sendto");
			goto abort;
		}
		tftp_read_ahead(file, convert);
		for ( ; ; ) {
			alarm(rexmtval);
			do {
				fromlen = sizeof(from);
				n = recvfrom(f, ackbuf, sizeof(ackbuf), 0,
				    (struct sockaddr *)&from, &fromlen);
			} while (n <= 0);
			alarm(0);
			if (n < 0) {
				bb_perror_msg("recvfrom");
				goto abort;
			}
			peeraddr.sin_port = from.sin_port;      /* added */
			if (trace)
				tpacket("received", ap, n);
			/* should verify packet came from server */
			ap_opcode = ntohs((u_short)ap->th_opcode);
			ap_block = ntohs((u_short)ap->th_block);
			if (ap_opcode == ERROR) {
				printf("Error code %d: %s\n", ap_block,
					ap->th_msg);
				goto abort;
			}
			if (ap_opcode == ACK) {
				int j;

				if (ap_block == block) {
					break;
				}
				/* On an error, try to synchronize
				 * both sides.
				 */
				j = tftp_synchnet(f);
				if (j && trace) {
					printf("discarded %d packets\n",
							j);
				}
				/*
				 * RFC1129/RFC1350: We MUST NOT re-send the DATA
				 * packet in response to an invalid ACK.  Doing so
				 * would cause the Sorcerer's Apprentice bug.
				 */
			}
		}
		if ( !is_request )
			amount += size;
		is_request = 0;
		block++;
	} while (size == SEGSIZE || block == 1);
abort:
	fclose(file);
	stopclock();
	if (amount > 0)
		printstats("Sent", amount);
}

/*
 * Receive a file.
 */
static void
tftp_recvfile(int fd, const char *name, const char *req_mode)
{
	struct tftphdr *ap;
	struct tftphdr *dp;
	int n;
	volatile u_short block;
	volatile int size, firsttrip;
	volatile unsigned long amount;
	struct sockaddr_in from;
	socklen_t fromlen;
	FILE *file;
	volatile int convert;           /* true if converting crlf -> lf */
	u_short dp_opcode, dp_block;

	startclock();
	dp = tftp_w_init();
	ap = (struct tftphdr *)ackbuf;
	convert = !strcmp(req_mode, "netascii");
	file = fdopen(fd, convert ?"wt":"wb");
	block = 1;
	firsttrip = 1;
	amount = 0;

	signal(SIGALRM, timer);
	do {
		if (firsttrip) {
			size = makerequest(RRQ, name, ap, req_mode);
			firsttrip = 0;
		} else {
			ap->th_opcode = htons((u_short)ACK);
			ap->th_block = htons((u_short)block);
			size = 4;
			block++;
		}
		timeout = 0;
		(void) sigsetjmp(timeoutbuf,1);
send_ack:
		if (trace)
			tpacket("sent", ap, size);
		if (sendto(f, ackbuf, size, 0, (struct sockaddr *)&peeraddr,
		    sizeof(peeraddr)) != size) {
			alarm(0);
			bb_perror_msg("sendto");
			goto abort;
		}
		tftp_write_behind(file, convert);
		for ( ; ; ) {
			alarm(rexmtval);
			do  {
				fromlen = sizeof(from);
				n = recvfrom(f, dp, TFTP_PKTSIZE, 0,
				    (struct sockaddr *)&from, &fromlen);
			} while (n <= 0);
			alarm(0);
			if (n < 0) {
				bb_perror_msg("recvfrom");
				goto abort;
			}
			peeraddr.sin_port = from.sin_port;      /* added */
			if (trace)
				tpacket("received", dp, n);
			/* should verify client address */
			dp_opcode = ntohs((u_short)dp->th_opcode);
			dp_block = ntohs((u_short)dp->th_block);
			if (dp_opcode == ERROR) {
			  printf("Error code %d: %s\n", dp_block, dp->th_msg);
			  goto abort;
			}
			if (dp_opcode == DATA) {
				int j;

				if (dp_block == block) {
					break;          /* have next packet */
				}
				/* On an error, try to synchronize
				 * both sides.
				 */
				j = tftp_synchnet(f);
				if (j && trace) {
					printf("discarded %d packets\n", j);
				}
				if (dp_block == (block-1)) {
					goto send_ack;  /* resend ack */
				}
			}
		}
	/*      size = write(fd, dp->th_data, n - 4); */
		size = tftp_writeit(file, &dp, n - 4, convert);
		if (size < 0) {
			nak(errno + 100, NULL);
			break;
		}
		amount += size;
	} while (size == SEGSIZE);
abort:                                          /* ok to ack, since user */
	ap->th_opcode = htons((u_short)ACK);    /* has seen err msg */
	ap->th_block = htons((u_short)block);
	(void) sendto(f, ackbuf, 4, 0, (struct sockaddr *)&peeraddr,
	    sizeof(peeraddr));
	tftp_write_behind(file, convert);       /* flush last buffer */
	fclose(file);
	stopclock();
	if (amount > 0)
		printstats("Received", amount);
}


/*
 * TFTP User Program -- Command Interface.
 */


struct modes {
  const char *m_name;
  const char *m_mode;
  int m_openflags;
};

static const struct modes modes[] = {
	{ "netascii", "netascii", O_TEXT },
	{ "ascii",    "netascii", O_TEXT },
	{ "octet",    "octet",    O_BINARY },
	{ "binary",   "octet",    O_BINARY },
	{ "image",    "octet",    O_BINARY },
	{ 0, 0, 0 }
};
#define MODE_OCTET    (&modes[2])
#define MODE_NETASCII (&modes[0])
#define MODE_DEFAULT  MODE_NETASCII

#define HELPINDENT (sizeof("connect"))

struct cmd {
	const char *name;
	const char *help;
	void    (*handler) (int, char **);
};


static void usage(int errcode)
{
  fprintf(stderr, "Usage: %s [-v][-l][-m mode] [host [port]] [-c command]\n", bb_applet_name);
  exit(errcode);
}

/* Called when a command is incomplete; modifies
   the global variable "line" */
static void
getmoreargs(const char *partial, const char *mprompt)
{
#ifdef CONFIG_FEATURE_COMMAND_EDITING
  strcpy(line, partial);
  if(cmdedit_read_input2(mprompt, line)<=0)
	exit(0);              /* EOF */
#else
  int len = strlen(partial);

  strcpy(line, partial);
  fputs(mprompt, stdout);
  if ( fgets(line+len, LBUFLEN-len, stdin) == 0 )
	  if ( feof(stdin) )
		  exit(0);      /* EOF */
#endif
}

/*
 * Slice a string up into argc/argv.
 */
static void
makeargv(void)
{
	char *cp;
	char **argp = margv;

	margc = 0;
	for (cp = line; *cp;) {
		while (isspace(*cp))
			cp++;
		if (*cp == '\0')
			break;
		*argp++ = cp;
		margc += 1;
		while (*cp != '\0' && !isspace(*cp))
			cp++;
		if (*cp == '\0')
			break;
		*cp++ = '\0';
	}
	*argp++ = 0;
}

static void
setpeer(int argc, char *argv[])
{
	if (argc < 2) {
		getmoreargs("connect ", "(to) ");
		makeargv();
		argc = margc;
		argv = margv;
	}
	if ((argc < 2) || (argc > 3)) {
		printf("usage: %s host-name [port]\n", argv[0]);
		return;
	}

	if(bb_gethostbyname(BB_GHBN_SOCKADDR_IN | BB_GHBN_SILENT | BB_GHBN_CANON_NAME, argv[1], &peeraddr, &hostname) == NULL) {
		connected = 0;
		printf("%s: unknown host\n", argv[1]);
		return;
	}

	port = sp->s_port;
	if (argc == 3) {
		struct servent *usp;
		usp = getservbyname(argv[2], "udp");
		if ( usp ) {
			port = usp->s_port;
		} else {
			int myport = bb_xparse_i_number(BB_PN_SET_UPPER | BB_PN_SILENT_RE, argv[2], 65535);
			if (myport < 0) {
				printf("%s: bad port number\n", argv[2]);
				connected = 0;
				return;
			}
			port = htons((u_short)myport);
		}
	}

	if (verbose) {
		printf("Connected to %s (%s), port %u\n",
		       hostname, inet_ntoa(peeraddr.sin_addr),
		       (unsigned int)ntohs(port));
	}
	connected = 1;
}

static void
settftpmode(const struct modes *newmode)
{
	mode = newmode;
	if (verbose)
		printf("mode set to %s\n", mode->m_mode);
}

static void
modecmd(int argc, char *argv[])
{
	const struct modes *p;
	const char *sep;

	if (argc < 2) {
		printf("Using %s mode to transfer files.\n", mode->m_mode);
		return;
	}
	if (argc == 2) {
		for (p = modes; p->m_name; p++)
			if (strcmp(argv[1], p->m_name) == 0)
				break;
		if (p->m_name) {
			settftpmode(p);
			return;
		}
		printf("%s: unknown mode\n", argv[1]);
		/* drop through and print usage message */
	}

	printf("usage: %s [", argv[0]);
	sep = " ";
	for (p = modes; p->m_name; p++) {
		printf("%s%s", sep, p->m_name);
		if (*sep == ' ')
			sep = " | ";
	}
	printf(" ]\n");
	return;
}

static void
setbinary(int argc, char *argv[])
{
	(void)argc; (void)argv; /* Quiet unused warning */
	settftpmode(MODE_OCTET);
}

static void
setascii(int argc, char *argv[])
{
	(void)argc; (void)argv; /* Quiet unused warning */
	settftpmode(MODE_NETASCII);
}


static char *
tail(char *filename)
{
	char *s;

	while (*filename) {
		s = strrchr(filename, '/');
		if (s == NULL)
			break;
		if (s[1])
			return (s + 1);
		*s = '\0';
	}
	return (filename);
}

static void
putusage(char *s)
{
	printf("usage: %s file ... host:target, or\n", s);
	printf("       %s file ... target (when already connected)\n", s);
}

/*
 * Send file(s).
 */
static void
put(int argc, char *argv[])
{
	int fd;
	int n;
	char *cp, *targ;

	if (argc < 2) {
		getmoreargs("send ", "(file) ");
		makeargv();
		argc = margc;
		argv = margv;
	}
	if (argc < 2) {
		putusage(argv[0]);
		return;
	}
	targ = argv[argc - 1];
	if (!literal && strchr(argv[argc - 1], ':')) {
		for (n = 1; n < argc - 1; n++)
			if (strchr(argv[n], ':')) {
				putusage(argv[0]);
				return;
			}
		cp = argv[argc - 1];
		targ = strchr_set0(cp, ':') + 1;
		if(bb_gethostbyname(BB_GHBN_SOCKADDR_IN | BB_GHBN_CANON_NAME, cp, &peeraddr, &hostname) == NULL)
			return;
		connected = 1;
	}
	if (!connected) {
		printf("No target machine specified.\n");
		return;
	}
	if (argc < 4) {
		cp = argc == 2 ? tail(targ) : argv[1];
		fd = open(cp, O_RDONLY|mode->m_openflags);
		if (fd < 0) {
			bb_perror_msg("%s", cp);
			return;
		}
		if (verbose)
			printf("putting %s to %s:%s [%s]\n",
				cp, hostname, targ, mode->m_mode);
		peeraddr.sin_port = port;
		tftp_sendfile(fd, targ, mode->m_mode);
		return;
	}
				/* this assumes the target is a directory */
				/* on a remote unix system.  hmmmm.  */
	cp = strchr(targ, '\0');
	*cp++ = '/';
	for (n = 1; n < argc - 1; n++) {
		strcpy(cp, tail(argv[n]));
		fd = open(argv[n], O_RDONLY|mode->m_openflags);
		if (fd < 0) {
			bb_perror_msg("%s", argv[n]);
			continue;
		}
		if (verbose)
			printf("putting %s to %s:%s [%s]\n",
				argv[n], hostname, targ, mode->m_mode);
		peeraddr.sin_port = port;
		tftp_sendfile(fd, targ, mode->m_mode);
	}
}


static void
getusage(char *s)
{
	printf("usage: %s host:file host:file ... file, or\n", s);
	printf("       %s file file ... file if connected\n", s);
}

/*
 * Receive file(s).
 */
static void
get(int argc, char *argv[])
{
	int fd;
	int n;
	char *cp;
	char *src;

	if (argc < 2) {
		getmoreargs("get ", "(files) ");
		makeargv();
		argc = margc;
		argv = margv;
	}
	if (argc < 2) {
		getusage(argv[0]);
		return;
	}
	if (!connected) {
		for (n = 1; n < argc ; n++)
			if (literal || strchr(argv[n], ':') == 0) {
				getusage(argv[0]);
				return;
			}
	}
	for (n = 1; n < argc ; n++) {
		src = strchr(argv[n], ':');
		if (literal || src == NULL)
			src = argv[n];
		else {
			*src++ = 0;
			if(bb_gethostbyname(BB_GHBN_SOCKADDR_IN | BB_GHBN_CANON_NAME, argv[n], &peeraddr, &hostname) == NULL)
				continue;
			connected = 1;
		}
		if (argc < 4) {
			cp = argc == 3 ? argv[2] : tail(src);
			fd = open(cp, O_WRONLY|O_CREAT|O_TRUNC|mode->m_openflags, 0666);
			if (fd < 0) {
				bb_perror_msg("%s", cp);
				return;
			}
			if (verbose)
				printf("getting from %s:%s to %s [%s]\n",
					hostname, src, cp, mode->m_mode);
			peeraddr.sin_port = port;
			tftp_recvfile(fd, src, mode->m_mode);
			break;
		}
		cp = tail(src);         /* new .. jdg */
		fd = open(cp, O_WRONLY|O_CREAT|O_TRUNC|mode->m_openflags, 0666);
		if (fd < 0) {
			bb_perror_msg("%s", cp);
			continue;
		}
		if (verbose)
			printf("getting from %s:%s to %s [%s]\n",
				hostname, src, cp, mode->m_mode);
		peeraddr.sin_port = port;
		tftp_recvfile(fd, src, mode->m_mode);
	}
}


static void
setrexmt(int argc, char *argv[])
{
	int t;

	if (argc < 2) {
		getmoreargs("rexmt-timeout ", "(value) ");
		makeargv();
		argc = margc;
		argv = margv;
	}
	if (argc != 2) {
		printf("usage: %s value\n", argv[0]);
		return;
	}
	t = atoi(argv[1]);
	if (t < 0)
		printf("%s: bad value\n", argv[1]);
	else
		rexmtval = t;
}


static void
settimeout(int argc, char *argv[])
{
	int t;

	if (argc < 2) {
		getmoreargs("maximum-timeout ", "(value) ");
		makeargv();
		argc = margc;
		argv = margv;
	}
	if (argc != 2) {
		printf("usage: %s value\n", argv[0]);
		return;
	}
	t = atoi(argv[1]);
	if (t < 0)
		printf("%s: bad value\n", argv[1]);
	else
		maxtimeout = t;
}

static void
setliteral(int argc, char *argv[])
{
	(void)argc; (void)argv; /* Quiet unused warning */
	literal = !literal;
	printf("Literal mode %s.\n", literal ? "on" : "off");
}

static void
status(int argc, char *argv[])
{
	(void)argc; (void)argv; /* Quiet unused warning */
	if (connected)
		printf("Connected to %s.\n", hostname);
	else
		printf("Not connected.\n");
	printf("Mode: %s Verbose: %s Tracing: %s Literal: %s\n", mode->m_mode,
		verbose ? "on" : "off", trace ? "on" : "off", literal ? "on" : "off");
	printf("Rexmt-interval: %d seconds, Max-timeout: %d seconds\n",
		rexmtval, maxtimeout);
}

static void
intr(int sig)
{
	(void)sig;      /* Quiet unused warning */

	signal(SIGALRM, SIG_IGN);
	alarm(0);
	siglongjmp(toplevel, -1);
}


/*
 * Command parser.
 */
static void
quit(int argc, char *argv[])
{
	(void)argc; (void)argv; /* Quiet unused warning */
	exit(0);
}

static void
settrace(int argc, char *argv[])
{
	(void)argc; (void)argv; /* Quiet unused warning */

	trace = !trace;
	printf("Packet tracing %s.\n", trace ? "on" : "off");
}

static void
setverbose(int argc, char *argv[])
{
	(void)argc; (void)argv; /* Quiet unused warning */

	verbose = !verbose;
	printf("Verbose mode %s.\n", verbose ? "on" : "off");
}

static void help(int argc, char *argv[]);

static const struct cmd cmdtab[] = {
  { "connect",
    "connect to remote tftp",
    setpeer },
  { "mode",
    "set file transfer mode",
    modecmd },
  { "put",
    "send file",
    put },
  { "get",
    "receive file",
    get },
  { "quit",
    "exit tftp",
    quit },
  { "verbose",
    "toggle verbose mode",
    setverbose },
  { "trace",
    "toggle packet tracing",
    settrace },
  { "literal",
    "toggle literal mode, ignore ':' in file name",
    setliteral },
  { "status",
    "show current status",
    status },
  { "binary",
    "set mode to octet",
    setbinary },
  { "ascii",
    "set mode to netascii",
    setascii },
  { "rexmt",
    "set per-packet transmission timeout",
    setrexmt },
  { "timeout",
    "set total retransmission timeout",
    settimeout },
  { "?",
    "print help information",
    help },
  { "help",
    "print help information",
    help },
  { 0, 0, 0 }
};

static const struct cmd *
getcmd(char *name)
{
	const char *p;
	char *q;
	const struct cmd *c, *found;
	int nmatches, longest;

	longest = 0;
	nmatches = 0;
	found = 0;
	for (c = cmdtab; (p = c->name) != NULL; c++) {
		for (q = name; *q == *p++; q++)
			if (*q == 0)            /* exact match? */
				return (c);
		if (!*q) {                      /* the name was a prefix */
			if (q - name > longest) {
				longest = q - name;
				nmatches = 1;
				found = c;
			} else if (q - name == longest)
				nmatches++;
		}
	}
	if (nmatches > 1)
		return ((struct cmd *)-1);
	return (found);
}

/*
 * Help command.
 */
static void
help(int argc, char *argv[])
{
	const struct cmd *c;

	printf("%s\n", VERSION);

	if (argc == 1) {
		printf("Commands may be abbreviated.  Commands are:\n\n");
		for (c = cmdtab; c->name; c++)
			printf("%-*s\t%s\n", (int)HELPINDENT, c->name, c->help);
		return;
	}
	while (--argc > 0) {
		char *arg;
		arg = *++argv;
		c = getcmd(arg);
		if (c == (struct cmd *)-1)
			printf("?Ambiguous help command %s\n", arg);
		else if (c == (struct cmd *)0)
			printf("?Invalid help command %s\n", arg);
		else
			printf("%s\n", c->help);
	}
}


static void
command(void)
{
	const struct cmd *c;

	for (;;) {
#ifdef CONFIG_FEATURE_COMMAND_EDITING
		if(cmdedit_read_input(prompt, line)<=0)
			exit(0); /* EOF */
#else
		fputs(prompt, stdout);
		if (fgets(line, LBUFLEN, stdin) == 0) {
			if (feof(stdin)) {
				exit(0);
			} else {
				continue;
			}
		}
#endif
		if ((line[0] == 0) || (line[0] == '\n'))
			continue;
		makeargv();
		if (margc == 0)
			continue;

		c = getcmd(margv[0]);
		if (c == (struct cmd *)-1) {
			printf("?Ambiguous command\n");
			continue;
		}
		if (c == 0) {
			printf("?Invalid command\n");
			continue;
		}
		(*c->handler)(margc, margv);
	}
}

int
tftp_main(void)
{
  int argc = Gargc;
  char **argv = Gargv;
  struct sockaddr_in s_in;
  int arg;
  int pargc, peerargc;
  int iscmd = 0;
  char **pargv;
  const char *optx;
  char *peerargv[3];

  mode = MODE_DEFAULT;
  maxtimeout = 5 * TIMEOUT;
  rexmtval = TIMEOUT;

  peerargv[0] = argv[0];
  peerargc = 1;

  for ( arg = 1 ; !iscmd && arg < argc ; arg++ ) {
    if ( argv[arg][0] == '-' ) {
      for ( optx = &argv[arg][1] ; *optx ; optx++ ) {
	switch ( *optx ) {
	case 'v':
	  verbose = 1;
	  break;
	case 'V':
	  /* Print version and configuration to stdout and exit */
	  printf("%s\n", TFTP_CONFIG_STR);
	  return 0;
	case 'l':
	  literal = 1;
	  break;
	case 'm':
	  if ( ++arg >= argc )
	    usage(EX_USAGE);
	  {
	    const struct modes *p;

	    for ( p = modes ; p->m_name ; p++ ) {
	      if (!strcmp(argv[arg], p->m_name))
		break;
	    }
	    if (p->m_name) {
	      settftpmode(p);
	    } else {
	      bb_default_error_retval = EX_USAGE;
	      bb_error_msg_and_die("invalid mode: %s", argv[arg]);
	    }
	  }
	  break;
	case 'c':
	  iscmd = 1;
	  break;
    case 'R':
      if ( ++arg >= argc )
	   usage(EX_USAGE);
      if ( sscanf(argv[arg], "%u:%u", &portrange_from, &portrange_to) != 2 ||
	    portrange_from > portrange_to || portrange_to > 65535 ) {
		bb_default_error_retval = EX_USAGE;
		bb_error_msg_and_die("bad port range: %s", argv[arg]);
      }
      portrange = 1;
      break;
	case 'h':
	default:
	  usage(*optx == 'h' ? 0 : EX_USAGE);
	}
      }
    } else {
      if ( peerargc >= 3 )
	usage(EX_USAGE);

      peerargv[peerargc++] = argv[arg];
    }
  }

  pargv = argv + arg;
  pargc = argc - arg;

  sp = getservbyname("tftp", "udp");
  if (sp == 0) {
    /* Use canned values */
    if (verbose)
      bb_error_msg("tftp/udp: unknown service, faking it...");
    sp = xmalloc(sizeof(struct servent));
    sp->s_name    = (char *)"tftp";
    sp->s_aliases = NULL;
    sp->s_port    = htons(IPPORT_TFTP);
    sp->s_proto   = (char *)"udp";
  }
  port = sp->s_port;            /* Default port */
  bb_default_error_retval = EX_OSERR;
  f = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM | BB_XSOCKET_FLG_ERR_DIE);
  memset(&s_in, 0, sizeof (s_in));
  s_in.sin_family = AF_INET;
  if (tftp_pick_port_bind(f, &s_in, portrange_from, portrange_to)) {
    bb_default_error_retval = EX_OSERR;
    bb_perror_msg_and_die("bind");
  }
  signal(SIGINT, intr);

  if ( peerargc ) {
    /* Set peer */
    if (sigsetjmp(toplevel,1) != 0)
      return EX_NOHOST;
    setpeer(peerargc, peerargv);
  }

  if ( iscmd && pargc ) {
    /* -c specified; execute command and exit */
    const struct cmd *c;

    if (sigsetjmp(toplevel,1) != 0)
      return EX_UNAVAILABLE;

    c = getcmd(pargv[0]);
    if ( c == (struct cmd *)-1 || c == (struct cmd *)0 ) {
      bb_default_error_retval = EX_USAGE;
      bb_error_msg_and_die("invalid command: %s", pargv[1]);
    }
    (*c->handler)(pargc, pargv);
    return 0;
  }

  if (sigsetjmp(toplevel,1) != 0)
    (void)putchar('\n');
  command();

  return 0;             /* Never reached */
}
#endif /* APPLET_USAGE */
