/*
 * Trivial file transfer protocol server.
 *
 *   Copyright 2001-2006 H. Peter Anvin - All Rights Reserved
 *      many modifications by Jim Guyton <guyton@rand-unix>
 *   This program is free software available under the same license
 *   as the "OpenBSD" operating system, distributed at
 *   http://www.openbsd.org/.
 *
 * busybox port by Vladimir Oleynik <dzo@simtreas.ru> (C) 2009
 * - size optimization, memory usage reducind
 *
 */
#ifdef APPLET_USAGE
#define tftpd_trivial_usage \
	"[-vcl][-a ADDR][-m MAP][-u USER][-t INETD_T][-T PKT_T][-r option...] [-s] [directory ...]"
#define tftpd_full_usage \
       "\t-l\t\tRun the server in standalone (listen) mode\n" \
       "\t-L\t\tSimilar to -l but do not detach from the foreground\n" \
       "\t\t\tprocess.\n" \
       "\t-a [address][:port]\tSpecify a specific address and port to listen to\n" \
       "\t\t\twhen called with the -l or -L option.\n" \
       "\t-c\t\tAllow new files to be created.\n" \
       "\t-s\t\tChange root directory on startup.\n" \
       "\t-u username\tSpecify the username which tftpd will run as;\n" \
       "\t-U umask\tSets the umask for newly created files to the\n" \
       "\t\t\tspecified value.\n" \
       "\t-p\t\tPerform no additional permissions checks above the\n" \
       "\t\t\tnormal system-provided access controls for the user\n" \
       "\t\t\tspecified via the -u option.\n" \
       "\t-t timeout\tWhen run from inetd this specifies how long, in\n" \
       "\t\t\tseconds, to wait for a second connection before\n" \
       "\t\t\tterminating the server.\n" \
       "\t-T timeout\tDetermine the default timeout, in microseconds,\n" \
       "\t\t\tbefore the first packet is retransmitted.\n" \
       USE_FEATURE_TFTPD_REGEX("\t-m remap-file\tSpecify the use of filename remapping.\n") \
       "\t-v\t\tIncrease the logging verbosity.\n" \
       "\t-r tftp-option\tIndicate that a specific RFC 2347 TFTP option\n" \
       "\t\t\tshould never be accepted.\n" \
       "\t-B max-block-sz\tSpecifies the maximum permitted block size.\n" \
       "\t-R port:port\tForce the server port number (the Transaction ID)\n" \
       "\t\t\tto be in the specified range of port numbers.\n" \
       "\t-V\t\tPrint the version number and configuration,\n" \
       "\t\t\tthen exit gracefully."
#else

/* Define if we have tcpwrappers (-lwrap) and <tcpd.h>. */
//#define HAVE_TCPWRAPPERS 1

#include "busybox.h"
#include "bb_sig.h"
#include "tftp_common.h"
#include <syslog.h>
#include <sys/ioctl.h>
#include <ctype.h>
#include <limits.h>
#include "bb_inet.h"

#include "xregex.h"

#ifdef CONFIG_FEATURE_TFTPD_REGEX
#define WITH_REGEX_STR ", with remap"
#else
#define WITH_REGEX_STR ", without remap"
#endif

#ifdef HAVE_TCPWRAPPERS
#define HAVE_LIBWRAP_STR ", with tcpwrappers"
#else
#define HAVE_LIBWRAP_STR ", without tcpwrappers"
#endif
#define TFTPD_CONFIG_STR VERSION WITH_REGEX_STR HAVE_LIBWRAP_STR


#ifndef S_IREAD
#define S_IREAD 0400
#endif
#ifndef S_IWRITE
#define S_IWRITE 0200
#endif

#include <sys/uio.h>

#ifdef HAVE_TCPWRAPPERS
#include <tcpd.h>

static int deny_severity       = LOG_WARNING;
static int allow_severity      = -1;   /* Don't log at all */

static struct request_info wrap_request;
#endif

#ifdef CONFIG_FEATURE_TFTPD_REGEX
struct rule {
  struct rule *next;
  int nrule;
  int rule_flags;
  regex_t rx;
  const char *pattern;
};
#endif

#define TIMEOUT 1000000         /* Default timeout (us) */
#define TRIES   6               /* Number of attempts to send each packet */
#define TIMEOUT_LIMIT ((1 << TRIES)-1)

CALC_BSS struct BB_COMMON_BSS {
	/* to stubs */
	struct bf bfs[2];
	int nextone;
	int current;
	int newline;
	int tftp_segsize;
	/* ******** */
#ifdef CONFIG_FEATURE_TFTPD_REGEX
	char *rewrite_file;
	struct rule *rewrite_rules;
#endif
	char buf[PKTSIZE];
	char ackbuf[PKTSIZE];
	struct tftphdr *recv_ap;    /* ack buffer */
	u_short recv_block;
	int recv_acksize;
	char blksize_ret[6];
	char blksize2_ret[6];
	char tsize_ret[sizeof(ubb64_t)*CHAR_BIT/3+2];
	char timeout_ret[4];
	char utimeout_ret[4];
	int blksize_set;
	struct sockaddr_in from;
	socklen_t fromlen;
	int timeout_quit;
	sigjmp_buf timeoutbuf;
	int peer;
	off_t tsize;
	int tsize_ok;
	const char **dirs;
	int secure;
	int unixperms;
	unsigned int portrange_from, portrange_to;
	int verbosity;
	volatile sig_atomic_t caught_sighup;
	FILE *file;
	int parse_nrule;
	unsigned int max_blksize;
	unsigned long maxtimeout;
	u_short block;                  /* Static to avoid longjmp funnies */
	unsigned long timeout;          /* Current timeout value */
	unsigned long rexmtval;         /* Basic timeout value */
	int remove_opt_flg[5];
} bb_common_bss;
#define tftp_segsize bb_common_bss.tftp_segsize
#define rewrite_file bb_common_bss.rewrite_file
#define rewrite_rules bb_common_bss.rewrite_rules
#define buf bb_common_bss.buf
#define ackbuf bb_common_bss.ackbuf
#define recv_ap bb_common_bss.recv_ap
#define recv_block bb_common_bss.recv_block
#define recv_acksize bb_common_bss.recv_acksize
#define blksize_ret bb_common_bss.blksize_ret
#define blksize2_ret bb_common_bss.blksize2_ret
#define tsize_ret bb_common_bss.tsize_ret
#define timeout_ret bb_common_bss.timeout_ret
#define utimeout_ret bb_common_bss.utimeout_ret
#define blksize_set bb_common_bss.blksize_set
#define from bb_common_bss.from
#define fromlen bb_common_bss.fromlen
#define timeout_quit bb_common_bss.timeout_quit
#define timeoutbuf bb_common_bss.timeoutbuf
#define peer bb_common_bss.peer
#define tsize bb_common_bss.tsize
#define tsize_ok bb_common_bss.tsize_ok
#define dirs bb_common_bss.dirs
#define secure bb_common_bss.secure
#define unixperms bb_common_bss.unixperms
#define portrange_from bb_common_bss.portrange_from
#define portrange_to bb_common_bss.portrange_to
#define verbosity bb_common_bss.verbosity
#define caught_sighup bb_common_bss.caught_sighup
#define file bb_common_bss.file
#define parse_nrule bb_common_bss.parse_nrule
#define max_blksize bb_common_bss.max_blksize
#define maxtimeout bb_common_bss.maxtimeout
#define block bb_common_bss.block
#define timeout bb_common_bss.timeout
#define rexmtval bb_common_bss.rexmtval
#define remove_opt_flg bb_common_bss.remove_opt_flg

#define cancreate (bss_flags & 1)


/* This means we don't want to wait() for children */
static void set_signal(void)
{
  struct sigaction sa;

  memset(&sa, 0, sizeof sa);
  sa.sa_handler = SIG_IGN;
  sigemptyset(&sa.sa_mask);
#ifdef SA_NOCLDWAIT
  sa.sa_flags = SA_NOCLDSTOP|SA_NOCLDWAIT;
#else
  sa.sa_flags = SA_NOCLDSTOP;
#endif
  while ( sigaction(SIGCHLD, &sa, NULL) ) ;     /* EINTR */
}


/*
 * strdup() that does the equivalent
 */
static char *tfstrdup(const char *str)
{
  char *p;
  int de = bb_default_error_retval;

  bb_default_error_retval = EX_OSERR;
  p = xstrdup(str);
  bb_default_error_retval = de;
  return p;
}


/*
 * recvfrom.c
 *
 * Emulate recvfrom() using recvmsg(), but try to capture the local address
 * since some TFTP clients consider it an error to get the reply from another
 * IP address than the request was sent to.
 *
 */

#ifndef CMSG_LEN
# define CMSG_LEN(size)  (sizeof(struct cmsghdr) + (size))
#endif
#ifndef CMSG_SPACE
# define CMSG_SPACE(size) (sizeof(struct cmsghdr) + (size))
#endif

/*
 * Check to see if this is a valid local address.  If so, we should
 * end up having the same local and remote address when trying to
 * bind to it.
 */
static int address_is_local(const struct sockaddr_in *addr)
{
  struct sockaddr_in s_in;
  int sockfd = -1;
  int e;
  int rv = 0;
  socklen_t addrlen;

  /* Multicast or universal broadcast address? */
  if (ntohl(addr->sin_addr.s_addr) >= (224UL << 24))
    return 0;

  sockfd = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM | BB_XSOCKET_FLG_NO_WARN);
  if (sockfd < 0)
    goto err;

  if (connect(sockfd, (const struct sockaddr *)addr, sizeof *addr))
    goto err;

  addrlen = sizeof s_in;
  if (getsockname(sockfd, (struct sockaddr *)&s_in, &addrlen))
    goto err;

  rv = s_in.sin_addr.s_addr == addr->sin_addr.s_addr;

 err:
  e = errno;

  if (sockfd >= 0)
    close(sockfd);

  errno = e;
  return rv;
}


static int myrecvfrom(int s, unsigned int flags, struct sockaddr_in *myaddr)
{
  struct msghdr msg;
  struct iovec iov;
  int n;
  struct cmsghdr *cmptr;
  union {
    struct cmsghdr cm;
#ifdef IP_PKTINFO
    char control[CMSG_SPACE(sizeof(struct in_addr)) +
		CMSG_SPACE(sizeof(struct in_pktinfo))];
#else
    char control[CMSG_SPACE(sizeof(struct in_addr))];
#endif
  } control_un;
#ifdef IP_PKTINFO
  struct in_pktinfo pktinfo;
#endif

  /* Try to enable getting the return address */
#ifdef IP_RECVDSTADDR
  setsockopt_1i(s, IPPROTO_IP, IP_RECVDSTADDR, 1);
#endif
#ifdef IP_PKTINFO
  setsockopt_1i(s, IPPROTO_IP, IP_PKTINFO, 1);
#endif

  memset(&msg, 0, sizeof msg);  /* Clear possible system-dependent fields */
  msg.msg_control = control_un.control;
  msg.msg_controllen = sizeof(control_un.control);
  msg.msg_flags = 0;

  msg.msg_name = (struct sockaddr *)&from;
  msg.msg_namelen = fromlen;
  iov.iov_base = buf;
  iov.iov_len  = sizeof (buf);
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;

  if ( (n = recvmsg(s, &msg, flags)) < 0 )
    return n;                   /* Error */

  fromlen = msg.msg_namelen;

  if ( myaddr ) {
    memset(myaddr, 0, sizeof(struct sockaddr_in));
    myaddr->sin_family = AF_INET;

    if ( msg.msg_controllen < sizeof(struct cmsghdr) ||
	 (msg.msg_flags & MSG_CTRUNC) )
      return n;                 /* No information available */

    for ( cmptr = CMSG_FIRSTHDR(&msg) ; cmptr != NULL ;
	  cmptr = CMSG_NXTHDR(&msg, cmptr) ) {

#ifdef IP_RECVDSTADDR
      if ( cmptr->cmsg_level == IPPROTO_IP &&
	   cmptr->cmsg_type == IP_RECVDSTADDR ) {
	memcpy(&myaddr->sin_addr, CMSG_DATA(cmptr),
	       sizeof(struct in_addr));
      }
#endif

#ifdef IP_PKTINFO
      if ( cmptr->cmsg_level == IPPROTO_IP &&
	   cmptr->cmsg_type == IP_PKTINFO ) {
	memcpy(&pktinfo, CMSG_DATA(cmptr), sizeof(struct in_pktinfo));
	memcpy(&myaddr->sin_addr, &pktinfo.ipi_addr, sizeof(struct in_addr));
      }
#endif

    }
  }

  /* If the address is not a valid local address, then bind to any address... */
  if (address_is_local(myaddr) != 1)
    myaddr->sin_addr.s_addr = INADDR_ANY;

  return n;
}

#ifdef CONFIG_FEATURE_TFTPD_REGEX
/*
 * Perform regular-expression based filename remapping.
 */

#define DEADMAN_MAX_STEPS       1024    /* Timeout after this many steps */
#define MAXLINE                 16384   /* Truncate a line at this many bytes */

#define RULE_REWRITE    0x01    /* This is a rewrite rule */
#define RULE_GLOBAL     0x02    /* Global rule (repeat until no match) */
#define RULE_EXIT       0x04    /* Exit after matching this rule */
#define RULE_RESTART    0x08    /* Restart at the top after matching this rule */
#define RULE_ABORT      0x10    /* Terminate processing with an error */
#define RULE_GETONLY    0x20    /* Applicable to GET only */
#define RULE_PUTONLY    0x40    /* Applicable to PUT only */
#define RULE_INVERSE    0x80    /* Execute if regex *doesn't* match */

/* Prototypes for regular-expression based filename remapping. */
/* This is called when we encounter a substitution like \i.  The
   macro character is passed as the first argument; the output buffer,
   if any, is passed as the second argument.  The function should return
   the number of characters output, or -1 on failure. */
typedef int (*match_pattern_callback)(char, char *);

static int xform_null(int c)
{
  return c;
}

static int xform_toupper(int c)
{
  return toupper(c);
}

static int xform_tolower(int c)
{
  return tolower(c);
}

/* Do \-substitution.  Call with string == NULL to get length only. */
static int genmatchstring(char *string, const char *pattern, const char *input,
			  const regmatch_t *pmatch, match_pattern_callback macrosub)
{
  int (*xform)(int) = xform_null;
  int len = 0;
  int n, mlen, sublen;
  int endbytes;

  /* Get section before match; note pmatch[0] is the whole match */
  endbytes = strlen(input) - pmatch[0].rm_eo;
  len = pmatch[0].rm_so + endbytes;
  if ( string ) {
    memcpy(string, input, pmatch[0].rm_so);
    string += pmatch[0].rm_so;
  }

  /* Transform matched section */
  while ( *pattern ) {
    mlen = 0;

    if ( *pattern == '\\' && pattern[1] != '\0' ) {
      char macro = pattern[1];
      switch ( macro ) {
      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
	n = pattern[1] - '0';

	if ( pmatch[n].rm_so != -1 ) {
	  mlen = pmatch[n].rm_eo - pmatch[n].rm_so;
	  len += mlen;
	  if ( string ) {
	    const char *p = input+pmatch[n].rm_so;
	    while ( mlen-- )
	      *string++ = xform(*p++);
	  }
	}
	break;

      case 'L':
	xform = xform_tolower;
	break;

      case 'U':
	xform = xform_toupper;
	break;

      case 'E':
	xform = xform_null;
	break;

      default:
	if ( macrosub &&
	     (sublen = macrosub(macro, string)) >= 0 ) {
	  while ( sublen-- ) {
	    len++;
	    if ( string ) {
	      *string = xform(*string);
	      string++;
	    }
	  }
	} else {
	  len++;
	  if ( string )
	    *string++ = xform(pattern[1]);
	}
      }
      pattern += 2;
    } else {
      len++;
      if ( string )
	*string++ = xform(*pattern);
      pattern++;
    }
  }

  /* Copy section after match */
  if ( string ) {
    memcpy(string, input+pmatch[0].rm_eo, endbytes);
    string[endbytes] = '\0';
  }

  return len;
}

/*
 * Extract a string terminated by non-escaped whitespace; ignoring
 * leading whitespace.  Consider an unescaped # to be a comment marker,
 * functionally \n.
 */
static int readescstring(char *buffer, char **str)
{
  char *p = *str;
  int wasbs = 0, len = 0;

  while ( *p && isspace(*p) )
    p++;

  if ( ! *p ) {
    *buffer = '\0';
    *str = p;
    return 0;
  }

  while ( *p ) {
    if ( !wasbs && (isspace(*p) || *p == '#') ) {
      *buffer = '\0';
      *str = p;
      return len;
    }
    /* Important: two backslashes leave us in the !wasbs state! */
    wasbs = !wasbs && ( *p == '\\' );
    *buffer++ = *p++;
    len++;
  }

  *buffer = '\0';
  *str = p;
  return len;
}

/* Parse a line into a set of instructions */
static int parseline(char *line, struct rule *r, int lineno)
{
  char buffer[MAXLINE];
  char *p;
  int rv;
  int rxflags = REG_EXTENDED;

  memset(r, 0, sizeof *r);
  r->nrule = parse_nrule;

  if ( !readescstring(buffer, &line) )
    return 0;                   /* No rule found */

  for ( p = buffer ; *p ; p++ ) {
    switch(*p) {
    case 'r':
      r->rule_flags |= RULE_REWRITE;
      break;
    case 'g':
      r->rule_flags |= RULE_GLOBAL;
      break;
    case 'e':
      r->rule_flags |= RULE_EXIT;
      break;
    case 's':
      r->rule_flags |= RULE_RESTART;
      break;
    case 'a':
      r->rule_flags |= RULE_ABORT;
      break;
    case 'i':
      rxflags |= REG_ICASE;
      break;
    case 'G':
      r->rule_flags |= RULE_GETONLY;
      break;
    case 'P':
      r->rule_flags |= RULE_PUTONLY;
      break;
    case '~':
      r->rule_flags |= RULE_INVERSE;
      break;
    default:
      bb_perror_msg("Remap command \"%s\" on line %d contains invalid char \"%c\"",
	     buffer, lineno, *p);
      return -1;                /* Error */
      break;
    }
  }

  /* RULE_GLOBAL only applies when RULE_REWRITE specified */
  if ( !(r->rule_flags & RULE_REWRITE) )
    r->rule_flags &= ~RULE_GLOBAL;

  if ( (r->rule_flags & (RULE_INVERSE|RULE_REWRITE)) ==
       (RULE_INVERSE|RULE_REWRITE) ) {
    bb_perror_msg("r rules cannot be inverted, line %d: %s", lineno, line);
    return -1;                  /* Error */
  }

  /* Read and compile the regex */
  if ( !readescstring(buffer, &line) ) {
    bb_perror_msg("No regex on remap line %d: %s", lineno, line);
    return -1;                  /* Error */
  }

  if ( (rv = regcomp(&r->rx, buffer, rxflags)) != 0 ) {
    char errbuf[BUFSIZ];
    regerror(rv, &r->rx, errbuf, BUFSIZ);
    bb_perror_msg("Bad regex in remap line %d: %s", lineno, errbuf);
    return -1;                  /* Error */
  }

  /* Read the rewrite pattern, if any */
  if ( readescstring(buffer, &line) ) {
    r->pattern = tfstrdup(buffer);
  } else {
    r->pattern = "";
  }

  parse_nrule++;
  return 1;                     /* Rule found */
}

/*
 * malloc() that syslogs an error message and bails if it fails.
 */
static void *tfmalloc(size_t size)
{
  void *p;
  int de = bb_default_error_retval;

  bb_default_error_retval = EX_OSERR;
  p = xmalloc(size);
  bb_default_error_retval = de;
  return p;
}

/* Read a rule file */
static struct rule *parserulefile(FILE *f)
{
  char line[MAXLINE];
  struct rule *first_rule = NULL;
  struct rule **last_rule = &first_rule;
  struct rule *this_rule  = tfmalloc(sizeof(struct rule));
  int rv;
  int lineno = 0;
  int err = 0;

  while ( lineno++, fgets(line, MAXLINE, f) ) {
    rv = parseline(line, this_rule, lineno);
    if ( rv < 0 )
      err = 1;
    if ( rv > 0 ) {
      *last_rule = this_rule;
      last_rule = &this_rule->next;
      this_rule = tfmalloc(sizeof(struct rule));
    }
  }

  free(this_rule);              /* Last one is always unused */

  if ( err ) {
    /* Bail on error, we have already logged an error message */
    exit(EX_CONFIG);
  }

  return first_rule;
}

/* Destroy a rule file data structure */
static void freerules(struct rule *r)
{
  struct rule *next;

  while ( r ) {
    next = r->next;

    regfree(&r->rx);

    /* "" patterns aren't allocated by malloc() */
    if ( r->pattern && *r->pattern )
      free((void *)r->pattern);

    free(r);

    r = next;
  }
}

/* Execute a rule set on a string; returns a malloc'd new string. */
static char *rewrite_string(const char *input, const struct rule *rules,
		     int is_put, match_pattern_callback macrosub,
		     const char **errmsg)
{
  char *current = tfstrdup(input);
  char *newstr;
  const struct rule *ruleptr = rules;
  regmatch_t pmatch[10];
  int len;
  int was_match = 0;
  int deadman = DEADMAN_MAX_STEPS;

  /* Default error */
  *errmsg = "Remap table failure";

  if ( verbosity >= 3 ) {
    syslog(LOG_INFO, "remap: input: %s", current);
  }

  for ( ruleptr = rules ; ruleptr ; ruleptr = ruleptr->next ) {
    if ( ((ruleptr->rule_flags & RULE_GETONLY) && is_put) ||
	 ((ruleptr->rule_flags & RULE_PUTONLY) && !is_put) ) {
      continue;                 /* Rule not applicable, try next */
    }

    if ( ! deadman-- ) {
      xlog_warn("remap: Breaking loop, input = %s, last = %s", input, current);
      free(current);
      return NULL;              /* Did not terminate! */
    }

    do {
      if ( regexec(&ruleptr->rx, current, 10, pmatch, 0) ==
	   (ruleptr->rule_flags & RULE_INVERSE ? REG_NOMATCH : 0) ) {
	/* Match on this rule */
	was_match = 1;

	if ( ruleptr->rule_flags & RULE_INVERSE ) {
	  /* No actual match, so clear out the pmatch array */
	  int i;
	  for ( i = 0 ; i < 10 ; i++ )
	    pmatch[i].rm_so = pmatch[i].rm_eo = -1;
	}

	if ( ruleptr->rule_flags & RULE_ABORT ) {
	  if ( verbosity >= 3 ) {
	    syslog(LOG_INFO, "remap: rule %d: abort: %s",
		   ruleptr->nrule, current);
	  }
	  if ( ruleptr->pattern[0] ) {
	    /* Custom error message */
	    len = genmatchstring(NULL, ruleptr->pattern, current,
				 pmatch, macrosub);
	    newstr = tfmalloc(len+1);
	    genmatchstring(newstr, ruleptr->pattern, current,
			   pmatch, macrosub);
	    *errmsg = newstr;
	  } else {
	    *errmsg = NULL;
	  }
	  free(current);
	  return NULL;
	}

	if ( ruleptr->rule_flags & RULE_REWRITE ) {
	  len = genmatchstring(NULL, ruleptr->pattern, current,
			       pmatch, macrosub);
	  newstr = tfmalloc(len+1);
	  genmatchstring(newstr, ruleptr->pattern, current,
			 pmatch, macrosub);
	  free(current);
	  current = newstr;
	  if ( verbosity >= 3 ) {
	    syslog(LOG_INFO, "remap: rule %d: rewrite: %s",
		   ruleptr->nrule, current);
	  }
	}
      } else {
	break;                  /* No match, terminate unconditionally */
      }
      /* If the rule is global, keep going until no match */
    } while ( ruleptr->rule_flags & RULE_GLOBAL );

    if ( was_match ) {
      was_match = 0;

      if ( ruleptr->rule_flags & RULE_EXIT ) {
	if ( verbosity >= 3 ) {
	  syslog(LOG_INFO, "remap: rule %d: exit", ruleptr->nrule);
	}
	return current;         /* Exit here, we're done */
      } else if ( ruleptr->rule_flags & RULE_RESTART ) {
	ruleptr = rules;        /* Start from the top */
	if ( verbosity >= 3 ) {
	  syslog(LOG_INFO, "remap: rule %d: restart", ruleptr->nrule);
	}
      }
    }
  }

  if ( verbosity >= 3 ) {
    syslog(LOG_INFO, "remap: done");
  }
  return current;
}
#endif /* CONFIG_FEATURE_TFTPD_REGEX */




struct options {
  const char    *o_opt;
  int     (*o_fnc)(char *, char **);
  int *ptr_flg_remove_opt;
};


/* Simple handler for SIGHUP */

static void handle_sighup(int sig)
{
  (void)sig;                    /* Suppress unused warning */
  caught_sighup = 1;
}


/* Handle timeout signal or timeout event */
static void
timer(int sig)
{
  (void)sig;                    /* Suppress unused warning */
  timeout <<= 1;
  if (timeout >= maxtimeout || timeout_quit)
    exit(0);
  siglongjmp(timeoutbuf, 1);
}

#ifdef CONFIG_FEATURE_TFTPD_REGEX
static struct rule *
read_remap_rules(void)
{
  FILE *f;
  struct rule *rulep;

  f = fopen(rewrite_file, "rt");
  if ( !f ) {
    bb_default_error_retval = EX_NOINPUT;
    bb_perror_msg_and_die("Cannot open map file: %s", rewrite_file);
  }
  rulep = parserulefile(f);
  fclose(f);

  return rulep;
}
#endif

static void set_socket_nonblock(int fd, int flag)
{
  int err;
  int flags;
  /* Posixly correct */
  err = ((flags = fcntl(fd, F_GETFL, 0)) < 0) ||
    (fcntl(fd, F_SETFL, flag ? flags|O_NONBLOCK : flags&~O_NONBLOCK) < 0);
  if ( err ) {
    bb_default_error_retval = EX_OSERR;
    bb_perror_msg_and_die("Cannot set nonblock flag on socket");
  }
}

static void pmtu_discovery_off(int fd)
{
#if defined(IP_MTU_DISCOVER) && defined(IP_PMTUDISC_DONT)
  setsockopt_1i(fd, IPPROTO_IP, IP_MTU_DISCOVER, IP_PMTUDISC_DONT);
#endif
}

/* Receive packet with synchronous timeout; timeout is adjusted
 * to account for time spent waiting. */
static int recv_time(int s, void *rbuf, int len, unsigned int flags,
		     unsigned long *timeout_us_p)
{
  fd_set fdset;
  struct timeval tmv, t0, t1;
  int rv, err;
  unsigned long timeout_us = *timeout_us_p;
  unsigned long timeout_left, dt;

  gettimeofday(&t0, NULL);
  timeout_left = timeout_us;

  for ( ; ; ) {
    FD_ZERO(&fdset);
    FD_SET(s, &fdset);

    do {
      tmv.tv_sec  = timeout_left / 1000000;
      tmv.tv_usec = timeout_left % 1000000;

      rv = select(s+1, &fdset, NULL, NULL, &tmv);
      err = errno;

      gettimeofday(&t1, NULL);

      dt = (t1.tv_sec - t0.tv_sec)*1000000 + (t1.tv_usec - t0.tv_usec);
      *timeout_us_p = timeout_left = ( dt >= timeout_us ) ? 1 : (timeout_us - dt);
    } while ( rv == -1 && err == EINTR );

    if ( rv == 0 ) {
      timer(0);                 /* Should not return */
      return -1;
    }

    set_socket_nonblock(s, 1);
    rv = recv(s, rbuf, len, flags);
    err = errno;
    set_socket_nonblock(s, 0);

    if ( rv < 0 ) {
      if ( E_WOULD_BLOCK(err) || err == EINTR ) {
	continue;               /* Once again, with feeling... */
      } else {
	errno = err;
	return rv;
      }
    } else {
      return rv;
    }
  }
}

/* Set a non-standard block size (c.f. RFC2348) */
static int set_blksize(char *val, char **ret)
{
  unsigned int sz;
  char *vend;

  sz = (unsigned int)strtoul(val, &vend, 10);

  if ( blksize_set || *vend )
    return 0;

  if (sz < 8)
    return(0);
  else if (sz > max_blksize)
    sz = max_blksize;

  tftp_segsize = sz;
  sprintf(*ret = blksize_ret, "%u", sz);

  blksize_set = 1;

  return(1);
}


/* Set a power-of-two block size (nonstandard) */
static int set_blksize2(char *val, char **ret)
{
  unsigned int sz;
  char *vend;

  sz = (unsigned int)strtoul(val, &vend, 10);

  if ( blksize_set || *vend )
    return 0;

  if (sz < 8)
    return(0);
  else if (sz > max_blksize)
    sz = max_blksize;

  /* Convert to a power of two */
  if ( sz & (sz-1) ) {
    unsigned int sz1 = 1;
    /* Not a power of two - need to convert */
    while ( sz >>= 1 )
      sz1 <<= 1;
    sz = sz1;
  }

  tftp_segsize = sz;
  sprintf(*ret = blksize2_ret, "%u", sz);

  blksize_set = 1;

  return(1);
}

/* Return a file size (c.f. RFC2349)
 * For netascii mode, we don't know the size ahead of time;
 * so reject the option. */
static int set_tsize(char *val, char **ret)
{
  ubb64_t sz;
  char *vend;

  sz = BB_strtoll(val, &vend, 10);

  if ( !tsize_ok || *vend )
    return 0;

  if (sz == 0)
    sz = (ubb64_t)tsize;

  sprintf(*ret = tsize_ret, "%" P_BB64T "u", sz);
  return(1);
}

/* Set the timeout (c.f. RFC2349).  This is supposed
 * to be the (default) retransmission timeout, but being an
 * integer in seconds it seems a bit limited. */
static int set_timeout(char *val, char **ret)
{
  unsigned long to;
  char *vend;

  to = strtoul(val, &vend, 10);

  if ( to < 1 || to > 255 || *vend )
    return 0;

  rexmtval = timeout = to*1000000UL;
  maxtimeout = rexmtval*TIMEOUT_LIMIT;

  sprintf(*ret = timeout_ret, "%lu", to);
  return(1);
}

/* Similar, but in microseconds.  We allow down to 10 ms. */
static int set_utimeout(char *val, char **ret)
{
  unsigned long to;
  char *vend;

  to = strtoul(val, &vend, 10);

  if ( to < 10000UL || to > 255000000UL || *vend )
    return 0;

  rexmtval = timeout = to;
  maxtimeout = rexmtval*TIMEOUT_LIMIT;

  sprintf(*ret = utimeout_ret, "%lu", to);
  return(1);
}

static const struct options options[] = {
  { "blksize",    set_blksize, &remove_opt_flg[0]  },
  { "blksize2",   set_blksize2, &remove_opt_flg[1]  },
  { "tsize",      set_tsize, &remove_opt_flg[2] },
  { "timeout",    set_timeout, &remove_opt_flg[3]  },
  { "utimeout",   set_utimeout, &remove_opt_flg[4]  },
  { NULL,         NULL, NULL }
};

static const char * const errmsgs[] =
{
  "Undefined error code",                       /* 0 - EUNDEF */
  "File not found",                             /* 1 - ENOTFOUND */
  "Access denied",                              /* 2 - EACCESS */
  "Disk full or allocation exceeded",           /* 3 - ENOSPACE */
  "Illegal TFTP operation",                     /* 4 - EBADOP */
  "Unknown transfer ID",                        /* 5 - EBADID */
  "File already exists",                        /* 6 - EEXISTS */
  "No such user",                               /* 7 - ENOUSER */
  "Failure to negotiate RFC2347 options"        /* 8 - EOPTNEG */
};
#define ERR_CNT (sizeof(errmsgs)/sizeof(const char *))

/* Send a nak packet (error message).
 * Error code passed in is one of the
 * standard TFTP codes, or a UNIX errno
 * offset by 100. */
static void nak(int error, const char *msg)
{
  struct tftphdr *tp;
  int length;

  tp = (struct tftphdr *)buf;
  tp->th_opcode = htons((u_short)ERROR);

  if ( error >= 100 ) {
    /* This is a Unix errno+100 */
    if ( !msg )
      msg = strerror(error - 100);
    error = EUNDEF;
  } else {
    if ( (unsigned)error >= ERR_CNT )
      error = EUNDEF;

    if ( !msg )
      msg = errmsgs[error];
  }

  tp->th_code = htons((u_short)error);

  length = strlen(msg)+1;
  memcpy(tp->th_msg, msg, length);
  length += 4;                  /* Add space for header */

  if ( verbosity >= 2 ) {
    syslog(LOG_INFO, "sending NAK (%d, %s) to %s",
	   error, tp->th_msg, inet_ntoa(from.sin_addr));
  }

  if (send(peer, buf, length, 0) != length)
    xlog_warn("nak: %m");
}

/* Parse RFC2347 style options */
static void do_opt(char *opt, char *val, char **ap)
{
  const struct options *po;
  char *ret;

  /* Global option-parsing variables initialization */
  blksize_set = 0;

  if ( !*opt )
    return;

  for (po = options; po->o_opt; po++)
    if (!*po->ptr_flg_remove_opt && !strcasecmp(po->o_opt, opt)) {
      if (po->o_fnc(val, &ret)) {
	if (*ap + strlen(opt) + strlen(ret) + 2 >=
	    ackbuf + sizeof(ackbuf)) {
	  nak(EOPTNEG, "Insufficient space for options");
	  exit(0);
	}
	*ap = strrchr(strcpy(strrchr(strcpy(*ap, opt),'\0') + 1,
			     ret),'\0') + 1;
      } else {
	nak(EOPTNEG, "Unsupported option(s) requested");
	exit(0);
      }
      break;
    }
  return;
}

#ifdef CONFIG_FEATURE_TFTPD_REGEX

/* This is called by the remap engine when it encounters macros such
 * as \i.  It should write the output in "output" if non-NULL, and
 * return the length of the output (generated or not).
 *
 * Return -1 on failure. */
static int rewrite_macros(char macro, char *output)
{
  char *p;

  switch (macro) {
  case 'i':
    p = inet_ntoa(from.sin_addr);
    if ( output )
      strcpy(output, p);
    return strlen(p);

  case 'x':
    if ( output )
      sprintf(output, "%08lX", (unsigned long)ntohl(from.sin_addr.s_addr));
    return 8;

  default:
    return -1;
  }
}

/* Modify the filename, if applicable.  If it returns NULL, deny the access. */
static char * rewrite_access(char *filename, int mode, const char **msg)
{
  if ( rewrite_rules ) {
    char *newname = rewrite_string(filename, rewrite_rules, mode != RRQ,
				   rewrite_macros, msg);
    filename = newname;
  }
  return filename;
}

#else
static char * rewrite_access(char *filename, int mode, const char **msg)
{
  (void)mode;                   /* Avoid warning */
  (void)msg;
  return filename;
}
#endif

/* Validate file access.  Since we
 * have no uid or gid, for now require
 * file to exist and be publicly
 * readable/writable, unless -p specified.
 * If we were invoked with arguments
 * from inetd then the file must also be
 * in one of the given directory prefixes.
 * Note also, full path name must be
 * given as we have no login directory. */
static int validate_access(char *filename, int mode,
		int f_convert, const char **errmsg)
{
  struct stat stbuf;
  int   i, len;
  int   fd, wmode, rmode;
  char *cp;
  const char **dirp;
  char stdio_mode[3];

  tsize_ok = 0;
  *errmsg = NULL;

  if (!secure) {
    if (*filename != '/') {
      *errmsg = "Only absolute filenames allowed";
      return (EACCESS);
    }

    /* prevent tricksters from getting around the directory restrictions */
    len = strlen(filename);
    for ( i = 1 ; i < len-3 ; i++ ) {
      cp = filename + i;
      if ( *cp == '.' && memcmp(cp-1, "/../", 4) == 0 ) {
	*errmsg = "Reverse path not allowed";
	return(EACCESS);
      }
    }

    for (dirp = dirs; *dirp; dirp++)
      if (strncmp(filename, *dirp, strlen(*dirp)) == 0)
	break;
    if (*dirp==0 && dirp!=dirs) {
      *errmsg = "Forbidden directory";
      return (EACCESS);
    }
  }

  /* We use different a different permissions scheme if `cancreate' is set. */
  wmode = O_WRONLY |
    (cancreate ? O_CREAT : 0) |
    (unixperms ? O_TRUNC : 0) |
    (f_convert ? O_TEXT : O_BINARY);
  rmode = O_RDONLY |
    (f_convert ? O_TEXT : O_BINARY);

  fd = open(filename, mode == RRQ ? rmode : wmode, 0666);
  if (fd < 0) {
    switch (errno) {
    case ENOENT:
    case ENOTDIR:
      return ENOTFOUND;
    case ENOSPC:
      return ENOSPACE;
    case EEXIST:
      return EEXISTS;
    default:
      return errno+100;
    }
  }

  if ( fstat(fd, &stbuf) < 0 )
    exit(EX_OSERR);             /* This shouldn't happen */

  if (mode == RRQ) {
    if ( !unixperms && (stbuf.st_mode & (S_IREAD >> 6)) == 0 ) {
      *errmsg = "File must have global read permissions";
      return (EACCESS);
    }
    tsize = stbuf.st_size;
    /* We don't know the tsize if conversion is needed */
    tsize_ok = !f_convert;
  } else {
    if ( !unixperms ) {
      if ( (stbuf.st_mode & (S_IWRITE >> 6)) == 0 ) {
	*errmsg = "File must have global write permissions";
	return (EACCESS);
      }

      /* We didn't get to truncate the file at open() time */
      if ( ftruncate(fd, (off_t)0) ) {
	*errmsg = "Cannot reset file size";
	return(EACCESS);
      }
    }
    tsize = 0;
    tsize_ok = 1;
  }

  stdio_mode[0] = (mode == RRQ) ? 'r':'w';
  stdio_mode[1] = (f_convert) ? 't':'b';
  stdio_mode[2] = '\0';

  file = fdopen(fd, stdio_mode);
  if (file == NULL)
    exit(EX_OSERR);             /* Internal error */

  return 0;
}

/* Send the requested file. */
static void tftp_sendfile(int f_convert, struct tftphdr *oap, int oacklen)
{
  struct tftphdr *dp;
  struct tftphdr *ap;           /* ack packet */
  u_short ap_opcode, ap_block;
  unsigned long r_timeout;
  int size, n;

  if (oap) {
    timeout = rexmtval;
    (void)sigsetjmp(timeoutbuf,1);
  oack:
    r_timeout = timeout;
    if (send(peer, oap, oacklen, 0) != oacklen) {
      xlog_warn("tftpd: oack: %m");
      goto abort;
    }
    for ( ; ; ) {
      n = recv_time(peer, ackbuf, sizeof(ackbuf), 0, &r_timeout);
      if (n < 0) {
	xlog_warn("tftpd: read: %m");
	goto abort;
      }
      ap = (struct tftphdr *)ackbuf;
      ap_opcode = ntohs((u_short)ap->th_opcode);
      ap_block = ntohs((u_short)ap->th_block);

      if (ap_opcode == ERROR) {
	xlog_warn("tftp: client does not accept options");
	goto abort;
      }
      if (ap_opcode == ACK) {
	if (ap_block == 0)
	  break;
	/* Resynchronize with the other side */
	(void)tftp_synchnet(peer);
	goto oack;
      }
    }
  }

  dp = tftp_r_init();
  do {
    size = tftp_readit(file, &dp, f_convert);
    if (size < 0) {
      nak(errno + 100, NULL);
      goto abort;
    }
    dp->th_opcode = htons((u_short)DATA);
    dp->th_block = htons((u_short)block);
    timeout = rexmtval;
    (void) sigsetjmp(timeoutbuf,1);

    r_timeout = timeout;
    if (send(peer, dp, size + 4, 0) != size + 4) {
      xlog_warn("tftpd: write: %m");
      goto abort;
    }
    tftp_read_ahead(file, f_convert);
    for ( ; ; ) {
      n = recv_time(peer, ackbuf, sizeof (ackbuf), 0, &r_timeout);
      if (n < 0) {
	xlog_warn("tftpd: read(ack): %m");
	goto abort;
      }
      ap = (struct tftphdr *)ackbuf;
      ap_opcode = ntohs((u_short)ap->th_opcode);
      ap_block = ntohs((u_short)ap->th_block);

      if (ap_opcode == ERROR)
	goto abort;

      if (ap_opcode == ACK) {
	if (ap_block == block) {
	  break;
	}
				/* Re-synchronize with the other side */
	(void) tftp_synchnet(peer);
	/* RFC1129/RFC1350: We MUST NOT re-send the DATA
	 * packet in response to an invalid ACK.  Doing so
	 * would cause the Sorcerer's Apprentice bug. */
      }

    }
    block++;
  } while (size == tftp_segsize);
 abort:
  (void) fclose(file);
}

/* Receive a file. */
static void
tftp_recvfile(int f_convert, struct tftphdr *oap, int oacklen)
{
  struct tftphdr *dp;
  int n, size;
  /* These are "static" to avoid longjmp funnies */
  u_short dp_opcode, dp_block;
  unsigned long r_timeout;

  dp = tftp_w_init();
  do {
    timeout = rexmtval;

    if (!recv_block && oap) {
      recv_ap = (struct tftphdr *)ackbuf;
      recv_acksize = oacklen;
    } else {
      recv_ap = (struct tftphdr *)ackbuf;
      recv_ap->th_opcode = htons((u_short)ACK);
      recv_ap->th_block = htons((u_short)recv_block);
      recv_acksize = 4;
    }
    recv_block++;
    (void) sigsetjmp(timeoutbuf,1);
  send_ack:
    r_timeout = timeout;
    if (send(peer, ackbuf, recv_acksize, 0) != recv_acksize) {
      xlog_warn("tftpd: write(ack): %m");
      goto abort;
    }
    tftp_write_behind(file, f_convert);
    for ( ; ; ) {
      n = recv_time(peer, dp, PKTSIZE, 0, &r_timeout);
      if (n < 0) {              /* really? */
	xlog_warn("tftpd: read: %m");
	goto abort;
      }
      dp_opcode = ntohs((u_short)dp->th_opcode);
      dp_block = ntohs((u_short)dp->th_block);
      if (dp_opcode == ERROR)
	goto abort;
      if (dp_opcode == DATA) {
	if (dp_block == recv_block) {
	  break;   /* normal */
	}
				/* Re-synchronize with the other side */
	(void) tftp_synchnet(peer);
	if (dp_block == (recv_block-1))
	  goto send_ack;                /* rexmit */
      }
    }
    /*  size = write(file, dp->th_data, n - 4); */
    size = tftp_writeit(file, &dp, n - 4, f_convert);
    if (size != (n-4)) {                        /* ahem */
      if (size < 0) nak(errno + 100, NULL);
      else nak(ENOSPACE, NULL);
      goto abort;
    }
  } while (size == tftp_segsize);
  tftp_write_behind(file, f_convert);
  (void) fclose(file);          /* close data file */

  recv_ap->th_opcode = htons((u_short)ACK);    /* send the "final" ack */
  recv_ap->th_block = htons((u_short)(recv_block));
  (void) send(peer, ackbuf, 4, 0);

  timeout_quit = 1;             /* just quit on timeout */
  n = recv_time(peer, buf, sizeof (buf), 0, &timeout); /* normally times out and quits */
  timeout_quit = 0;

  if (n >= 4 &&                 /* if read some data */
      dp_opcode == DATA &&      /* and got a data recv_block */
      recv_block == dp_block) {      /* then my last ack was lost */
    (void) send(peer, ackbuf, 4, 0);     /* resend final ack */
  }
 abort:
  return;
}


static void set_max_blksize(char *arg)
{
	char *vp;
	max_blksize = (unsigned int)strtoul(arg, &vp, 10);
	if ( max_blksize < 512 || max_blksize > MAX_SEGSIZE || *vp ) {
	  bb_default_error_retval = EX_USAGE;
	  bb_error_msg_and_die("Bad maximum blocksize value (range 512-%d): %s",
		 MAX_SEGSIZE, arg);
	}
}

static void set_max_timeout(char *arg)
{
	char *vp;
	unsigned long tov = strtoul(arg, &vp, 10);
	if ( tov < 10000UL || tov > 255000000UL || *vp ) {
	  bb_default_error_retval = EX_USAGE;
	  bb_error_msg_and_die("Bad timeout value: %s", arg);
	}
	rexmtval = timeout = tov;
	maxtimeout = rexmtval*TIMEOUT_LIMIT;
}

static void set_port_range(char *arg)
{
	if ( sscanf(arg, "%u:%u", &portrange_from, &portrange_to) != 2 ||
	     portrange_from > portrange_to || portrange_to >= 65535 ) {
	  bb_default_error_retval = EX_USAGE;
	  bb_error_msg_and_die("Bad port range: %s", arg);
	}
}

#ifdef CONFIG_FEATURE_TFTPD_REGEX
static void set_rewrite_file(char *arg)
{
      if ( rewrite_file != NULL ) {
	bb_default_error_retval = EX_USAGE;
	bb_error_msg_and_die("Multiple -m options");
      }
      rewrite_file = arg;
}
#endif

/* Handle initial connection protocol. */
static int tftp(struct tftphdr *tp, int size)
{
  char *cp, *end;
  int argn, ecode;
  int   f_convert = -1;
  char *origfilename;
  char *filename, *mode = NULL;
  const char *errmsgptr;
  u_short tp_opcode = ntohs(tp->th_opcode);

  char *val = NULL, *opt = NULL;
  char *ap = ackbuf + 2;

  ((struct tftphdr *)ackbuf)->th_opcode = htons(OACK);

  origfilename = cp = (char *) &(tp->th_stuff);
  argn = 0;

  end = (char *)tp + size;

  while ( cp < end && *cp ) {
    do {
      cp++;
    } while (cp < end && *cp);

    if ( *cp ) {
      nak(EBADOP, "Request not null-terminated");
      exit(0);
    }

    argn++;
    if (argn == 1) {
      mode = ++cp;
    } else if (argn == 2) {

      for (cp = mode; *cp; cp++)
	*cp = tolower(*cp);
      if (!strcmp("netascii", mode))
	f_convert = 1;
      else if (!strcmp("octet", mode))
	f_convert = 0;
      else {
	nak(EBADOP, "Unknown mode");
	exit(0);
      }
      if ( !(filename =
	     rewrite_access(origfilename, tp_opcode, &errmsgptr)) ) {
	nak(EACCESS, errmsgptr); /* File denied by mapping rule */
	exit(0);
      }
      if ( verbosity >= 1 ) {
	if ( filename == origfilename || !strcmp(filename, origfilename) )
	  syslog(LOG_NOTICE, "%s from %s filename %s\n",
		 tp_opcode == WRQ ? "WRQ" : "RRQ",
		 inet_ntoa(from.sin_addr), filename);
	else
	  syslog(LOG_NOTICE, "%s from %s filename %s remapped to %s\n",
		 tp_opcode == WRQ ? "WRQ" : "RRQ",
		 inet_ntoa(from.sin_addr), origfilename, filename);
      }
      ecode = validate_access(filename, tp_opcode, f_convert, &errmsgptr);
      if (ecode) {
	nak(ecode, errmsgptr);
	exit(0);
      }
      opt = ++cp;
    } else if ( argn & 1 ) {
      val = ++cp;
    } else {
      do_opt(opt, val, &ap);
      opt = ++cp;
    }
  }

  if (f_convert == -1) {
    nak(EBADOP, "Missing mode");
    exit(0);
  }

  if ( ap != (ackbuf+2) ) {
    if ( tp_opcode == WRQ )
      tftp_recvfile(f_convert, (struct tftphdr *)ackbuf, ap-ackbuf);
    else
      tftp_sendfile(f_convert, (struct tftphdr *)ackbuf, ap-ackbuf);
  } else {
    if (tp_opcode == WRQ)
      tftp_recvfile(f_convert, NULL, 0);
    else
      tftp_sendfile(f_convert, NULL, 0);
  }
  exit(0);                      /* Request completed */
}

static void remove_opt(char *arg)
{
	const struct options *opt;

	for ( opt = options ; opt->o_opt ; opt++ ) {
		if ( !strcasecmp(arg, opt->o_opt) ) {
			/* Don't support this option */
			*opt->ptr_flg_remove_opt = 1;
			return;
		}
	}
	bb_default_error_retval = EX_USAGE;
	bb_error_msg_and_die("Unknown option: %s", arg);
}

int tftpd_main(void)
{
  struct tftphdr *tp;
  struct passwd *pw;
  struct sockaddr_in myaddr;
  struct sockaddr_in bindaddr;
  int n;
  int fd = 0;
  int standalone;               /* Standalone (listen) mode */
  int nodaemon;                 /* Do not detach process */
  char *address = NULL;         /* Address to listen to */
  pid_t pid;
  char *umask_arg = NULL;
  mode_t my_umask = 0;
  int spec_umask = 0;
  int c;
  int setrv;
  int waittime = 900;           /* Default time to wait for a connect*/
  const char *user = "nobody";  /* Default user */
  char *ep;
  u_short tp_opcode;
  int ndirs;

  tftp_segsize = SEGSIZE;          /* Default segsize */
  block = 1;

  bb_openlog(LOG_PID|LOG_NDELAY, LOG_DAEMON);
  log_syslog = LOG_ERR; /* add sysloging */

  srand(time(NULL) ^ getpid());
  max_blksize = MAX_SEGSIZE;
  maxtimeout = TIMEOUT_LIMIT*TIMEOUT;
  timeout  = TIMEOUT;
  rexmtval = TIMEOUT;

  c = bb_xgetopt_ulflags ("csplLVa:t:B:T:R:u:U:r:"
			USE_FEATURE_TFTPD_REGEX("m:")
			"v\0:-t:Ll:=BTRr"
			USE_FEATURE_TFTPD_REGEX("m")
			":vv",
			&address, &waittime, set_max_blksize, set_max_timeout,
			set_port_range, &user, &umask_arg, remove_opt,
#ifdef CONFIG_FEATURE_TFTPD_REGEX
			set_rewrite_file,
#endif
			&verbosity);
  /* cancreate = c & 1; */
  secure = c & 2;
  unixperms = c & 4;
  standalone = c & 8;
  nodaemon = c & 16;
  if( (c & 32) ) {      /* -V */
      /* Print configuration to stdout and exit */
      printf("%s\n", TFTPD_CONFIG_STR);
      return 0;
  }
  if(umask_arg != NULL) {       /* -U umask */
      my_umask = strtoul(umask_arg, &ep, 8);
      if ( *ep ) {
	bb_default_error_retval = EX_USAGE;
	bb_error_msg_and_die("Invalid umask: %s", umask_arg);
      }
      spec_umask = 1;
  }

  dirs = xmalloc((Gargc+1)*sizeof(char *));
  for ( ndirs = 0 ; Gargc != 0 ; Gargc-- )
    dirs[ndirs++] = *Gargv++;

  dirs[ndirs] = NULL;

  if (secure) {
    if (ndirs == 0) {
      bb_default_error_retval = EX_USAGE;
      bb_error_msg_and_die("no -s directory");
    }
    if (ndirs > 1) {
      bb_default_error_retval = EX_USAGE;
      bb_error_msg_and_die("too many -s directories");
    }
    if (chdir(dirs[0])) {
      bb_default_error_retval = EX_NOINPUT;
      bb_perror_msg_and_die("%s", dirs[0]);
    }
  }

  pw = getpwnam(user);
  if (!pw) {
    bb_default_error_retval = EX_NOUSER;
    bb_perror_msg_and_die("no user %s", user);
  }

  if ( spec_umask || !unixperms )
    umask(my_umask);

  /* Note: on Cygwin, select() on a nonblocking socket becomes
     a nonblocking select. */
#ifndef __CYGWIN__
  set_socket_nonblock(fd, 1);
#endif

#ifdef CONFIG_FEATURE_TFTPD_REGEX
  if ( rewrite_file )
    rewrite_rules = read_remap_rules();
#endif

  /* If we're running standalone, set up the input port */
  if ( standalone ) {
    fd = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM | BB_XSOCKET_FLG_ERR_DIE);

    memset(&bindaddr, 0, sizeof bindaddr);
    bindaddr.sin_family = AF_INET;
    bindaddr.sin_addr.s_addr = INADDR_ANY;
    bindaddr.sin_port = htons(IPPORT_TFTP);

    if ( address ) {
      char *portptr, *eportptr;
      struct servent *servent;
      unsigned long port;

      address = tfstrdup(address);
      portptr = strchr_set0(address, ':');
      if ( portptr )
	portptr++;

      if ( *address ) {
	if(bb_gethostbyname(BB_GHBN_IN_ADDR | BB_GHBN_ADD_ERR_MSG, address,
		&bindaddr.sin_addr, "cannot resolve local bind address") == NULL)
	  return EX_NOINPUT;
      } else {
	/* Default to using INADDR_ANY */
      }

      if ( portptr && *portptr ) {
	servent = getservbyname(portptr, "udp");
	if ( servent ) {
	  bindaddr.sin_port = servent->s_port;
	} else if ( (port = strtoul(portptr, &eportptr, 0)) && !*eportptr ) {
	  bindaddr.sin_port = htons(port);
	} else if ( !strcmp(portptr, "tftp") ) {
	  /* It's TFTP, we're OK */
	} else {
	  bb_default_error_retval = EX_NOINPUT;
	  bb_error_msg_and_die("cannot resolve local bind port: %s", portptr);
	}
      }
    }

    bb_default_error_retval = EX_OSERR;
    bb_xbind(fd, (struct sockaddr *)&bindaddr, sizeof bindaddr);

    /* Daemonize this process */
    /* Note: when running in secure mode (-s), we must not chroot, since
       we are already in the proper directory. */
    if (!nodaemon && daemon(secure, 0) < 0)
      bb_perror_msg_and_die("cannot daemonize");
    if (!nodaemon)
	log_stderr = 0;
  } else {
    /* 0 is our socket descriptor */
    close(1); close(2);
    log_stderr = 0;
  }

  /* Disable path MTU discovery */
  pmtu_discovery_off(0);

  /* This means we don't want to wait() for children */
  set_signal();

  /* Take SIGHUP and use it to set a variable.  This
     is polled synchronously to make sure we don't
     lose packets as a result. */
  bb_signal((1 << SIGHUP), handle_sighup);

  while ( 1 ) {
    fd_set readset;
    struct timeval tv_waittime;
    int rv;

    if ( caught_sighup ) {
      caught_sighup = 0;
      if ( standalone ) {
#ifdef CONFIG_FEATURE_TFTPD_REGEX
	if ( rewrite_file ) {
	  freerules(rewrite_rules);
	  rewrite_rules = read_remap_rules();
	}
#endif
      } else {
	/* Return to inetd for respawn */
	return 0;
      }
    }

    FD_ZERO(&readset);
    FD_SET(fd, &readset);
    tv_waittime.tv_sec = waittime;
    tv_waittime.tv_usec = 0;

#ifdef __CYGWIN__
    /* On Cygwin, select() on a nonblocking socket returns immediately,
       with a rv of 0! */
    set_socket_nonblock(fd, 0);
#endif

    /* Never time out if we're in standalone mode */
    rv = select(fd+1, &readset, NULL, NULL, standalone ? NULL : &tv_waittime);
    if ( rv == -1 && errno == EINTR )
      continue;         /* Signal caught, reloop */
    if ( rv == -1 ) {
      bb_default_error_retval = EX_IOERR;
      bb_perror_msg_and_die("select loop");
    } else if ( rv == 0 ) {
      return 0;          /* Timeout, return to inetd */
    }

#ifdef __CYGWIN__
    set_socket_nonblock(fd, 1);
#endif

    fromlen = sizeof (from);
    n = myrecvfrom(fd, 0, &myaddr);

    if ( n < 0 ) {
      if ( E_WOULD_BLOCK(errno) || errno == EINTR ) {
	continue;               /* Again, from the top */
      } else {
	bb_default_error_retval = EX_IOERR;
	bb_perror_msg_and_die("recvfrom");
      }
    }

    if ( from.sin_family != AF_INET ) {
      bb_default_error_retval = EX_PROTOCOL;
      bb_error_msg_and_die("received address was not AF_INET, please check your inetd config");
    }

    if ( standalone && myaddr.sin_addr.s_addr == INADDR_ANY ) {
      /* myrecvfrom() didn't capture the source address; but we might
	 have bound to a specific address, if so we should use it */
      memcpy(&myaddr.sin_addr, &bindaddr.sin_addr, sizeof bindaddr.sin_addr);
    }

    /*
     * Now that we have read the request packet from the UDP
     * socket, we fork and go back to listening to the socket.
     */
    pid = fork();
    if (pid < 0) {
      bb_default_error_retval = EX_OSERR;
      bb_perror_msg_and_die("fork");
      /* Return to inetd, just in case */
    } else if ( pid == 0 )
      break;                    /* Child exit, parent loop */
  }

  /* Child process: handle the actual request here */

  /* Ignore SIGHUP */
  bb_signal((1 << SIGHUP), SIG_IGN);

#ifdef HAVE_TCPWRAPPERS
  /* Verify if this was a legal request for us.  This has to be
     done before the chroot, while /etc is still accessible. */
  request_init(&wrap_request,
	       RQ_DAEMON, bb_applet_name,
	       RQ_FILE, fd,
	       RQ_CLIENT_SIN, &from,
	       RQ_SERVER_SIN, &myaddr,
	       0);
  sock_methods(&wrap_request);
  if ( hosts_access(&wrap_request) == 0 ) {
    if ( deny_severity != -1 )
      syslog(deny_severity, "connection refused from %s", inet_ntoa(from.sin_addr));
    return EX_NOPERM;    /* Access denied */
  } else if ( allow_severity != -1 ) {
    syslog(allow_severity, "connect from %s", inet_ntoa(from.sin_addr));
  }
#endif

  /* Close file descriptors we don't need */
  close(fd);

  /* Get a socket.  This has to be done before the chroot(), since
     some systems require access to /dev to create a socket. */

  bb_default_error_retval = EX_IOERR;
  peer = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM | BB_XSOCKET_FLG_ERR_DIE);

  /* Set up the supplementary group access list if possible */
  /* /etc/group still need to be accessible at this point */
  setrv = initgroups(user, pw->pw_gid);
  if ( setrv ) {
    bb_default_error_retval = EX_OSERR;
    bb_error_msg_and_die("cannot set groups for user %s", user);
  }

  /* Chroot and drop privileges */
  if (secure) {
    if (chroot(".")) {
      bb_default_error_retval = EX_OSERR;
      bb_error_msg_and_die("chroot");
    }
#ifdef __CYGWIN__
    chdir("/");                 /* Cygwin chroot() bug workaround */
#endif
  }

  setrv = setregid(pw->pw_gid, pw->pw_gid);

  setrv = setrv || setreuid(pw->pw_uid, pw->pw_uid);

  if ( setrv ) {
    bb_default_error_retval = EX_OSERR;
    bb_perror_msg_and_die("cannot drop privileges");
  }

  /* Other basic setup */
  from.sin_family = AF_INET;

  /* Process the request... */
  if (tftp_pick_port_bind(peer, &myaddr, portrange_from, portrange_to) < 0) {
    bb_default_error_retval = EX_IOERR;
    bb_perror_msg_and_die("bind");
  }

  if (connect(peer, (struct sockaddr *)&from, sizeof from) < 0) {
    bb_default_error_retval = EX_IOERR;
    bb_perror_msg_and_die("connect");
  }

  /* Disable path MTU discovery */
  pmtu_discovery_off(peer);

  tp = (struct tftphdr *)buf;
  tp_opcode = ntohs(tp->th_opcode);
  if (tp_opcode == RRQ || tp_opcode == WRQ)
    tftp(tp, n);
  return 0;
}
#endif /* APPLET_USAGE */
