/*
 * Sysctl 1.01 - A utility to read and manipulate the sysctl parameters
 *
 *
 * "Copyright 1999 George Staikos
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 *
 * Changelog:
 *      v1.01:
 *              - added -p <preload> to preload values from a file
 *      v1.01.1
 *              - busybox applet aware by <solar@gentoo.org>
 *
 *  use config_read() getopt_ulflags() is_directory() and other optimization
 *  by Vladimir Oleynik <dzo@simtreas.ru>
 */
#ifdef APPLET_USAGE
#define sysctl_trivial_usage \
	"[OPTIONS]... [VALUE]..."
#define sysctl_full_usage \
	"configure kernel parameters at runtime\n\n" \
	"Options:\n" \
	"\t-n\tUse this option to disable printing of the key name when printing values\n" \
	"\t-w\tUse this option when you want to change a sysctl setting\n" \
	"\t-p\tLoad in sysctl settings from the file specified or /etc/sysctl.conf if none given\n" \
	"\t-a\tDisplay all values currently available\n" \
	"\t-A\tDisplay all values currently available in table form"
#else

#include "busybox.h"
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <fcntl.h>
#include "config_read.h"

/* Globals... */
#define DEFAULT_PRELOAD "/etc/sysctl.conf"
static const char PROC_PATH[] = "/proc/sys/";

/* error messages */
static const char ERR_INVALID_KEY[] = "error: '%s' is an unknown key";
static const char ERR_PERMISSION_DENIED[] = "error: permission denied on key '%s'";


static void dwrite_str(int fd, const char *buf)
{
	write(fd, buf, strlen(buf));
}

extern struct BB_COMMON_BSS {
	int output;
} bb_common_bss;
#define output bb_common_bss.output


/* Write a single sysctl setting */
static int sysctl_write_setting(const char *setting)
{
	int retval = 0;
	const char *name = setting;
	const char *value;
	const char *equals;
	char *tmpname, *outname, *cptr;
	int fd;

	if (!name)                      /* probably dont' want to display this  err */
		return 0;

	if (!(equals = strchr(setting, '='))) {
		bb_error_msg("error: '%s' must be of the form name=value", setting);
		return -1;
	}

	value = equals + 1;     /* point to the value in name=value */

	if (!*name || !*value || name == equals) {
		bb_error_msg("error: Malformed setting '%s'", setting);
		return -2;
	}

	tmpname = xasprintf("%s%.*s", PROC_PATH, (int)(equals - name), name);
	outname = xstrdup(tmpname + strlen(PROC_PATH));

	while ((cptr = strchr(tmpname, '.')) != NULL)
		*cptr = '/';

	while ((cptr = strchr(outname, '/')) != NULL)
		*cptr = '.';

	if ((fd = open(tmpname, O_WRONLY | O_CREAT | O_TRUNC, 0666)) < 0) {
		switch (errno) {
		case ENOENT:
			bb_error_msg(ERR_INVALID_KEY, outname);
			break;
		case EACCES:
			bb_perror_msg(ERR_PERMISSION_DENIED, outname);
			break;
		default:
			bb_perror_msg("error setting key '%s'", outname);
			break;
		}
		retval = -1;
	} else {
		dwrite_str(fd, value);
		close(fd);
		if (output) {
			dwrite_str(STDOUT_FILENO, outname);
			dwrite_str(STDOUT_FILENO, " = ");
		}
		dwrite_str(STDOUT_FILENO, value);
		dwrite_str(STDOUT_FILENO, "\n");
	}

	/* cleanup */
	free(tmpname);
	free(outname);
	return retval;
}                                               /* end sysctl_write_setting() */


/*
 *     sysctl_preload_file
 *      preload the sysctl's from a conf file
 *           - we parse the file and then reform it (strip out whitespace)
 */
static int sysctl_preload_file(const char *filename)
{
	config_read_t *crt = config_read_prepare(filename, 'w');

	if (crt == NULL)
		return 1;

	while ((crt = config_read(crt, "#\1\2\0\0=")) != NULL) {
		char *name = crt->tokens[0];
		char *value = crt->tokens[1];
		if (name[0] == ';')
			continue;

		if (value == NULL ||
				(crt->flags[0] & CR_FLAGS_FOUND_SPEC_DELIM) == 0 ||
				(crt->flags[1] & CR_FLAGS_FOUND_EOL_DELIM) == 0 ||
				(crt->flags[1] & CR_FLAGS_FOUND_SPEC_DELIM) != 0 ||
				((crt->flags[0]|crt->flags[1]) & CR_FLAGS_FOUND_DOUBLE_SPEC_DELIM) != 0) {
			bb_error_msg("warning: %s(%d): invalid syntax, continuing...", filename, crt->num_line);
			continue;
		}

		/* yes, its hack, use internal variable */
		sprintf(crt->input_buff, "%s=%s", name, value);
		sysctl_write_setting(crt->input_buff);
	}
	return 0;
}                       /* end sysctl_preload_file() */


/* Read a sysctl setting */
static int sysctl_read_setting(const char *setting)
{
	int retval = 0;
	char *tmpname, *outname, *cptr;
	char inbuf[1025];
	const char *name = setting;
	FILE *fp;

	if (!*setting)
		bb_error_msg(ERR_INVALID_KEY, setting);

	tmpname = concat_path_file(PROC_PATH, name);
	outname = xstrdup(tmpname + strlen(PROC_PATH));

	while ((cptr = strchr(tmpname, '.')) != NULL)
		*cptr = '/';
	while ((cptr = strchr(outname, '/')) != NULL)
		*cptr = '.';

	if ((fp = fopen(tmpname, "r")) == NULL) {
		switch (errno) {
		case ENOENT:
			bb_error_msg(ERR_INVALID_KEY, outname);
			break;
		case EACCES:
			bb_error_msg(ERR_PERMISSION_DENIED, outname);
			break;
		default:
			bb_perror_msg("error: reading key '%s'", outname);
			break;
		}
		retval = 1;
	} else {
		while (fgets(inbuf, sizeof(inbuf) - 1, fp)) {
			if (output) {
				dwrite_str(STDOUT_FILENO, outname);
				dwrite_str(STDOUT_FILENO, " = ");
			}
			dwrite_str(STDOUT_FILENO, inbuf);
		}
		fclose(fp);
	}

	free(tmpname);
	free(outname);
	return retval;
}                               /* end sysctl_read_setting() */

/* Display all the sysctl settings */
static int sysctl_display_all(const char *path)
{
	int retval = 0;
	DIR *dp;
	struct dirent *de;
	int isdir;

	if (!(dp = opendir(path))) {
		bb_perror_msg("error: unable to open directory '%s'", path);
		retval = 1;
	} else {
		while ((de = readdir(dp)) != NULL) {
			char *tmpdir = concat_subpath_file(path, de->d_name);
			if(tmpdir == NULL)
				continue;
			isdir = is_directory(tmpdir, 1, NULL);
			if (isdir < 0)
				bb_perror_msg("%s", tmpdir);
			else {
				if (isdir > 0)
					sysctl_display_all(tmpdir);
				 else
					retval |= sysctl_read_setting(tmpdir + strlen(PROC_PATH));

			}
			free(tmpdir);
		}                               /* end while */
		closedir(dp);
	}

	return retval;
}                                       /* end sysctl_display_all() */

/* sysctl_main()... */
int sysctl_main(void)
{
	int retval = 0;
	int what = bb_getopt_ulflags("npwaA");

	output = (what & 1) ^ 1;    /* reverse */
	if((what & (2|4|8|16)) == 0) {
		if(*Gargv == NULL)
			bb_show_usage();
		while (*Gargv != NULL)
			retval |= sysctl_read_setting(*Gargv++);
		return retval;
	}
	if((what & 2) != 0)
		return sysctl_preload_file(*Gargv == NULL ? DEFAULT_PRELOAD : *Gargv);
	if((what & 4) != 0) {
		while (*Gargv != NULL)
			retval |= sysctl_write_setting(*Gargv++);
		return retval;
	}
	/* -a or -A, have not different */
	return sysctl_display_all(PROC_PATH);
}
#endif /* APPLET_USAGE */
