/* from zpipe.c: example of proper use of zlib */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "zlib.h"


/* report a zlib or i/o error */
static void zerr(int ret)
{
    fputs("zpipe: ", stderr);
    switch (ret) {
    case Z_ERRNO:
	if (ferror(stdin))
	    fputs("error reading stdin\n", stderr);
	if (ferror(stdout))
	    fputs("error writing stdout\n", stderr);
	break;
    case Z_STREAM_ERROR:
	fputs("invalid compression level\n", stderr);
	break;
    case Z_DATA_ERROR:
	fputs("invalid or incomplete deflate data\n", stderr);
	break;
    case Z_MEM_ERROR:
	fputs("out of memory\n", stderr);
	break;
    case Z_VERSION_ERROR:
	fputs("zlib version mismatch!\n", stderr);
    }
}

#define CHUNK 16384

/* Compress from file source to file dest until EOF on source.
   def() returns Z_OK on success, Z_MEM_ERROR if memory could not be
   allocated for processing, Z_STREAM_ERROR if an invalid compression
   level is supplied, Z_VERSION_ERROR if the version of zlib.h and the
   version of the library linked do not match, or Z_ERRNO if there is
   an error reading or writing the files. */
/* compress or decompress from stdin to stdout */
int main(void)
{
    int ret, flush;
    unsigned have;
    z_stream strm;
    unsigned char in[CHUNK];
    unsigned char out[CHUNK];

    /* allocate deflate state */
    strm.zalloc = Z_NULL;
    strm.zfree = Z_NULL;
    strm.opaque = Z_NULL;
    ret = deflateInit(&strm, Z_DEFAULT_COMPRESSION);
    if (ret != Z_OK) {
gret:
	zerr(ret);
	return ret;
    }

    /* compress until end of file */
    do {
	strm.avail_in = fread(in, 1, CHUNK, stdin);
	if (ferror(stdin)) {
	    (void)deflateEnd(&strm);
	    ret = Z_ERRNO;
	    goto gret;
	}
	flush = feof(stdin) ? Z_FINISH : Z_NO_FLUSH;
	strm.next_in = in;

	/* run deflate() on input until output buffer not full, finish
	   compression if all of stdin has been read in */
	do {
	    strm.avail_out = CHUNK;
	    strm.next_out = out;
	    ret = deflate(&strm, flush);    /* no bad return value */
	    assert(ret != Z_STREAM_ERROR);  /* state not clobbered */
	    have = CHUNK - strm.avail_out;
	    if (fwrite(out, 1, have, stdout) != have || ferror(stdout)) {
		(void)deflateEnd(&strm);
		ret = Z_ERRNO;
		goto gret;
	    }
	} while (strm.avail_out == 0);
	assert(strm.avail_in == 0);     /* all input will be used */

	/* done when last data in file processed */
    } while (flush != Z_FINISH);
    assert(ret == Z_STREAM_END);        /* stream will be complete */

    /* clean up and return */
    (void)deflateEnd(&strm);
    return Z_OK;
}
