/*
 * Mini klogd implementation for busybox
 *
 * Copyright (C) 2001 by Gennady Feldman <gfeldman@gena01.com>.
 * Changes: Made this a standalone busybox module which uses standalone
 *                                      syslog() client interface.
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 *
 * Copyright (C) 2000 by Karl M. Hegbloom <karlheg@debian.org>
 *
 * "circular buffer" Copyright (C) 2000 by Gennady Feldman <gfeldman@gena01.com>
 *
 * Maintainer: Gennady Feldman <gfeldman@gena01.com> as of Mar 12, 2001
 *
 * Licensed under the GPL v2 or later, see the file LICENSE in this tarball.
 */
#ifdef APPLET_USAGE
#define klogd_trivial_usage \
	"[-c n] [-n]"
#define klogd_full_usage \
	"Kernel logger.\n" \
	"Options:\n" \
	"\t-c n\tSets the default log level of console messages to n\n" \
	"\t-n\tRun as a foreground process"
#else

#include "busybox.h"
#include <string.h>             /* strncpy() */
#include <unistd.h>
#include <ctype.h>
#include <syslog.h>
#include <sys/klog.h>

#include "bb_sig.h"

static void klogd_signal(int sig)
{
	klogctl(7, NULL, 0);
	klogctl(0, 0, 0);
	/* logMessage(0, "Kernel log daemon exiting."); */
	log_syslog = LOG_NOTICE;
	bb_default_error_retval = EXIT_SUCCESS;
	bb_error_msg_and_die("Kernel log daemon exiting.");
}

#define KLOGD_LOGBUF_SIZE 4096

CALC_BSS struct BB_COMMON_BSS
{
	char log_buffer[KLOGD_LOGBUF_SIZE];
} bb_common_bss;
#define log_buffer bb_common_bss.log_buffer


#define OPT_LEVEL        1
#define OPT_FOREGROUND   2

extern int klogd_main(void)
{
	unsigned long opt;
	char *c_arg;
	int console_log_level = -1;
	int priority = LOG_INFO;
	int i, n, lastc;
	char *start;

	/* do normal option parsing */
	opt = bb_getopt_ulflags("c:n", &c_arg);

	if (opt & OPT_LEVEL) {
		/* Valid levels are between 1 and 8 */
		console_log_level = bb_xparse_i_number(10 | BB_PN_SET_LOWER1 | BB_PN_SET_UPPER, c_arg, 8);
	}

	if((opt & OPT_FOREGROUND) == 0)
		xdaemon(XDAEMON_PIDFILE);

	openlog("kernel", 0, LOG_KERN);

	/* Set up sig handlers */
	bb_signal( (1 << SIGINT) | (1 << SIGTERM) | (1 << SIGQUIT), klogd_signal);
	bb_signal((1 << SIGHUP), SIG_IGN);

	/* "Open the log. Currently a NOP." */
	klogctl(1, NULL, 0);

	/* Set level of kernel console messaging.. */
	if (console_log_level != -1)
		klogctl(8, NULL, console_log_level);

	syslog(LOG_NOTICE, "klogd started: %s", BB_BANNER);

	while (1) {
		/* Use kernel syscalls */
		memset(log_buffer, 0, KLOGD_LOGBUF_SIZE);
		n = klogctl(2, log_buffer, KLOGD_LOGBUF_SIZE);
		if (n < 0) {
			if (errno == EINTR)
				continue;
			bb_perror_msg_and_die("Error return from sys_sycall: %d", errno);
		}

		/* klogctl buffer parsing modelled after code in dmesg.c */
		start = &log_buffer[0];
		lastc = '\0';
		for (i = 0; i < n; i++) {
			if (lastc == '\0' && log_buffer[i] == '<') {
				priority = 0;
				i++;
				while (log_buffer[i] >= '0' && log_buffer[i] <= '9') {
					priority = priority * 10 + (log_buffer[i] - '0');
					i++;
				}
				if (log_buffer[i] == '>')
					i++;
				start = &log_buffer[i];
			}
			if (log_buffer[i] == '\n') {
				log_buffer[i] = '\0';   /* zero terminate this message */
				syslog(priority, "%s", start);
				start = &log_buffer[i + 1];
				priority = LOG_INFO;
			}
			lastc = log_buffer[i];
		}
	}
}
#endif /* APPLET_USAGE */
