/*
 * Mini syslogd implementation for busybox
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 *
 * Copyright (C) 2000 by Karl M. Hegbloom <karlheg@debian.org>
 *
 * "circular buffer" Copyright (C) 2001 by Gennady Feldman <gfeldman@gena01.com>
 *
 * Maintainer: Gennady Feldman <gfeldman@gena01.com> as of Mar 12, 2001
 *
 * Licensed under the GPL v2 or later, see the file LICENSE in this tarball.
 */
#ifdef APPLET_USAGE
#define syslogd_trivial_usage \
	"[OPTION]..."
#define syslogd_full_usage \
	"Linux system and kernel logging utility.\n" \
	"Note that this version of syslogd ignores /etc/syslog.conf.\n\n" \
	"Options:\n" \
	"\t-m MIN\t\tMinutes between MARK lines (default=20, 0=off)\n" \
	"\t-n\t\tRun as a foreground process\n" \
	"\t-O FILE\t\tUse an alternate log file (default=/var/log/messages)\n" \
	"\t-S\t\tMake logging output smaller." \
	USE_FEATURE_ROTATE_LOGFILE( \
	"\n\t-s SIZE\t\tMax size (KB) before rotate (default=200KB, 0=off)\n" \
	"\t-b NUM\t\tNumber of rotated logs to keep (default=1, max=99, 0=purge)") \
	USE_FEATURE_REMOTE_LOG( \
	"\n\t-R HOST[:PORT]\tLog to IP or hostname on PORT (default PORT=514/UDP)\n" \
	"\t-L\t\tLog locally and via network logging (default is network only)") \
	USE_FEATURE_IPC_SYSLOG( \
	"\n\t-C [size(KiB)]\tLog to a circular buffer (read the buffer using logread)")
#else

#include "busybox.h"
#include <ctype.h>
#include <paths.h>
#include <stdbool.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <sys/un.h>
#include <fcntl.h>
#include "bb_inet.h"
#include "bb_sig.h"

/* SYSLOG_NAMES defined to pull some extra junk from syslog.h */
#define SYSLOG_NAMES
#include <syslog.h>
#include <sys/uio.h>

#if defined(__BIONIC__)
#error "Android have special syslog. Reconfigure for disable the 'syslog' applet."
#endif

#ifdef CONFIG_FEATURE_REMOTE_LOG
#include <netinet/in.h>
#endif

#define MAXLINE         1024    /* maximum line length */

#if MAXLINE > BUFSIZ
# define TMP_BUF_SZ BUFSIZ
#else
# define TMP_BUF_SZ MAXLINE
#endif

CALC_BSS struct BB_COMMON_BSS
{
	char tmpbuf[TMP_BUF_SZ];
/* Path to the unix socket */
	char lfile[MAXPATHLEN];
/* localhost's name */
	char LocalHostName[64];
#ifdef CONFIG_FEATURE_REMOTE_LOG
	struct sockaddr remoteaddr;
/* where do we log? */
	char *RemoteHost;
/* To remote log or not to remote log, that is the question. */
	int doRemoteLog;
	int local_logging;
#endif
/* Make loging output smaller. */
	bool small;
#ifdef CONFIG_FEATURE_IPC_SYSLOG
	int circular_logging;
	struct shbuf_ds *buf;
#endif
	char res[20];
#ifdef CONFIG_FEATURE_REMOTE_LOG
	char lm_line[MAXLINE + 1];
#endif

	const char *logFilePath;
#ifdef CONFIG_FEATURE_ROTATE_LOGFILE
	/* max size of message file before being rotated in bytes (initial value in kb) */
	long logFileSize;
	int logFileRotate; /* number of rotated message files */
#endif
	/* interval between marks in seconds (initial value in min) */
	int MarkInterval;
#ifdef CONFIG_FEATURE_REMOTE_LOG
	int remotefd; /* udp socket for logging to remote host */
	int RemotePort; /* what port to log to? */
#endif
#ifdef CONFIG_FEATURE_IPC_SYSLOG
	int shmid;      // ipc shared memory id
	int s_semid;    // ipc semaphore id
	int shm_size;
#endif
} bb_common_bss;
#define lfile bb_common_bss.lfile
#define LocalHostName bb_common_bss.LocalHostName
#define remoteaddr bb_common_bss.remoteaddr
#define RemoteHost bb_common_bss.RemoteHost
#define doRemoteLog bb_common_bss.doRemoteLog
#define local_logging bb_common_bss.local_logging
#define small bb_common_bss.small
#define circular_logging bb_common_bss.circular_logging
#define buf bb_common_bss.buf
#define res bb_common_bss.res
#define lm_line bb_common_bss.lm_line
#define logFilePath bb_common_bss.logFilePath
#define logFileSize bb_common_bss.logFileSize
#define logFileRotate bb_common_bss.logFileRotate
#define MarkInterval bb_common_bss.MarkInterval
#define remotefd bb_common_bss.remotefd
#define RemotePort bb_common_bss.RemotePort
#define shmid bb_common_bss.shmid
#define s_semid bb_common_bss.s_semid
#define shm_size bb_common_bss.shm_size

/* circular buffer variables/structures */
#ifdef CONFIG_FEATURE_IPC_SYSLOG

#if CONFIG_FEATURE_IPC_SYSLOG_BUFFER_SIZE < 4
#error Sorry, you must set the syslogd buffer size to at least 4KB.
#error Please check CONFIG_FEATURE_IPC_SYSLOG_BUFFER_SIZE
#endif

#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>

/* our shared key */
static const long KEY_ID = 0x414e4547;  /*"GENA" */

// Semaphore operation structures
struct shbuf_ds {
	int size;                       // size of data written
	int head;                       // start of message list
	int tail;                       // end of message list
	char data[1];                   // data/messages
};                                 // shared memory pointer

/*
 * sem_up - up()'s a semaphore.
 */
static inline void sem_up(int semid)
{
	struct sembuf SMwup = { 1, -1, IPC_NOWAIT };
	if (semop(semid, &SMwup, 1) == -1)
		bb_perror_msg_and_die("semop[SMwup]");
}

/*
 * sem_down - down()'s a semaphore
 */
static inline void sem_down(int semid)
{
	struct sembuf SMwdn[3] = { {0, 0, 0}, {1, 0, 0}, {1, +1, 0} };
	if (semop(semid, SMwdn, 3) == -1)
		bb_perror_msg_and_die("semop[SMwdn]");
}


static void ipcsyslog_cleanup(void)
{
	printf("Exiting Syslogd!\n");
	if (shmid != -1) {
		shmdt(buf);
	}

	if (shmid != -1) {
		shmctl(shmid, IPC_RMID, NULL);
	}
	if (s_semid != -1) {
		semctl(s_semid, 0, IPC_RMID, 0);
	}
}

static void ipcsyslog_init(void)
{
	if (buf == NULL) {
		if ((shmid = shmget(KEY_ID, shm_size, IPC_CREAT | 1023)) == -1)
			bb_perror_msg_and_die("shmget");

		if ((buf = shmat(shmid, NULL, 0)) == (void *) -1)
			bb_perror_msg_and_die("shmat");

		buf->size = shm_size - sizeof(*buf);
		buf->head = buf->tail = 0;

		// we'll trust the OS to set initial semval to 0 (let's hope)
		if ((s_semid = semget(KEY_ID, 2, IPC_CREAT | IPC_EXCL | 1023)) == -1) {
			if (errno == EEXIST) {
				if ((s_semid = semget(KEY_ID, 2, 0)) == -1)
					bb_perror_msg_and_die("semget");
			} else {
				bb_perror_msg_and_die("semget");
			}
		}
	} else {
		printf("Buffer already allocated just grab the semaphore?");
	}
}

/* write message to buffer */
static void circ_message(const char *msg)
{
	int l = strlen(msg) + 1;        /* count the whole message w/ '\0' included */

	sem_down(s_semid);

	/*
	 * Circular Buffer Algorithm:
	 * --------------------------
	 *
	 * Start-off w/ empty buffer of specific size SHM_SIZ
	 * Start filling it up w/ messages. I use '\0' as separator to break up messages.
	 * This is also very handy since we can do printf on message.
	 *
	 * Once the buffer is full we need to get rid of the first message in buffer and
	 * insert the new message. (Note: if the message being added is >1 message then
	 * we will need to "remove" >1 old message from the buffer). The way this is done
	 * is the following:
	 *      When we reach the end of the buffer we set a mark and start from the beginning.
	 *      Now what about the beginning and end of the buffer? Well we have the "head"
	 *      index/pointer which is the starting point for the messages and we have "tail"
	 *      index/pointer which is the ending point for the messages. When we "display" the
	 *      messages we start from the beginning and continue until we reach "tail". If we
	 *      reach end of buffer, then we just start from the beginning (offset 0). "head" and
	 *      "tail" are actually offsets from the beginning of the buffer.
	 *
	 * Note: This algorithm uses Linux IPC mechanism w/ shared memory and semaphores to provide
	 *       a threadsafe way of handling shared memory operations.
	 */
	if ((buf->tail + l) < buf->size) {
		/* before we append the message we need to check the HEAD so that we won't
		   overwrite any of the message that we still need and adjust HEAD to point
		   to the next message! */
		if (buf->tail < buf->head) {
			if ((buf->tail + l) >= buf->head) {
				/* we need to move the HEAD to point to the next message
				 * Theoretically we have enough room to add the whole message to the
				 * buffer, because of the first outer IF statement, so we don't have
				 * to worry about overflows here!
				 */
				int k = buf->tail + l - buf->head;      /* we need to know how many bytes
													   we are overwriting to make
													   enough room */
				char *c =
					memchr(buf->data + buf->head + k, '\0',
						   buf->size - (buf->head + k));
				if (c != NULL) {        /* do a sanity check just in case! */
					buf->head = c - buf->data + 1;  /* we need to convert pointer to
													   offset + skip the '\0' since
													   we need to point to the beginning
													   of the next message */
					/* Note: HEAD is only used to "retrieve" messages, it's not used
					   when writing messages into our buffer */
				} else {        /* show an error message to know we messed up? */
					printf("Weird! Can't find the terminator token?\n");
					buf->head = 0;
				}
			}
		}

		/* in other cases no overflows have been done yet, so we don't care! */
		/* we should be ok to append the message now */
		strncpy(buf->data + buf->tail, msg, l); /* append our message */
		buf->tail += l; /* count full message w/ '\0' terminating char */
	} else {
		/* we need to break up the message and "circle" it around */
		char *c;
		int k = buf->tail + l - buf->size;      /* count # of bytes we don't fit */

		/* We need to move HEAD! This is always the case since we are going
		 * to "circle" the message.
		 */
		c = memchr(buf->data + k, '\0', buf->size - k);

		if (c != NULL) {        /* if we don't have '\0'? weird!!! */
			/* move head pointer */
			buf->head = c - buf->data + 1;

			/* now write the first part of the message */
			strncpy(buf->data + buf->tail, msg, l - k - 1);

			/* ALWAYS terminate end of buffer w/ '\0' */
			buf->data[buf->size - 1] = '\0';

			/* now write out the rest of the string to the beginning of the buffer */
			strcpy(buf->data, &msg[l - k - 1]);

			/* we need to place the TAIL at the end of the message */
			buf->tail = k + 1;
		} else {
			printf("Weird! Can't find the terminator token from the beginning?\n");
			buf->head = buf->tail = 0;      /* reset buffer, since it's probably corrupted */
		}

	}
	sem_up(s_semid);
}
#endif                                                  /* CONFIG_FEATURE_IPC_SYSLOG */

/* Note: There is also a function called "message()" in init.c */
/* Print a message to the log file. */
static void message(char *fmt, ...) __attribute__ ((format(printf, 1, 2)));
static void message(char *fmt, ...)
{
	int fd;
	struct flock fl;
	va_list arguments;

	fl.l_whence = SEEK_SET;
	fl.l_start = 0;
	fl.l_len = 1;

#ifdef CONFIG_FEATURE_IPC_SYSLOG
	if ((circular_logging) && (buf != NULL)) {
		char b[1024];

		va_start(arguments, fmt);
		vsnprintf(b, sizeof(b) - 1, fmt, arguments);
		va_end(arguments);
		circ_message(b);

	} else
#endif
	if ((fd = device_open(logFilePath,
					O_WRONLY | O_CREAT | O_NOCTTY | O_APPEND |
							 O_NONBLOCK)) >= 0) {
		fl.l_type = F_WRLCK;
		fcntl(fd, F_SETLKW, &fl);
#ifdef CONFIG_FEATURE_ROTATE_LOGFILE
		if ( logFileSize > 0 ) {
			struct stat statf;
			int r = fstat(fd, &statf);
			if( !r && (statf.st_mode & S_IFREG)
				&& (lseek(fd,0,SEEK_END) > logFileSize) ) {
				if(logFileRotate > 0) {
					int i;
					char oldFile[(strlen(logFilePath)+4)], newFile[(strlen(logFilePath)+4)];
					for(i=logFileRotate-1;i>0;i--) {
						sprintf(oldFile, "%s.%d", logFilePath, i-1);
						sprintf(newFile, "%s.%d", logFilePath, i);
						rename(oldFile, newFile);
					}
					sprintf(newFile, "%s.%d", logFilePath, 0);
					fl.l_type = F_UNLCK;
					fcntl (fd, F_SETLKW, &fl);
					close(fd);
					rename(logFilePath, newFile);
					fd = device_open (logFilePath,
						   O_WRONLY | O_CREAT | O_NOCTTY | O_APPEND |
						   O_NONBLOCK);
					fl.l_type = F_WRLCK;
					fcntl (fd, F_SETLKW, &fl);
				} else {
					ftruncate( fd, 0 );
				}
			}
		}
#endif
		va_start(arguments, fmt);
		vdprintf(fd, fmt, arguments);
		va_end(arguments);
		fl.l_type = F_UNLCK;
		fcntl(fd, F_SETLKW, &fl);
		close(fd);
	} else {
		/* Always send console messages to /dev/console so people will see them. */
		if ((fd = device_open(_PATH_CONSOLE,
						 O_WRONLY | O_NOCTTY | O_NONBLOCK)) >= 0) {
			va_start(arguments, fmt);
			vdprintf(fd, fmt, arguments);
			va_end(arguments);
			close(fd);
		} else {
			fprintf(stderr, "Bummer, can't print: ");
			va_start(arguments, fmt);
			vfprintf(stderr, fmt, arguments);
			fflush(stderr);
			va_end(arguments);
		}
	}
}

#ifdef CONFIG_FEATURE_REMOTE_LOG
static void init_RemoteLog(void)
{
	if(bb_gethostbyname(BB_GHBN_SOCKADDR, RemoteHost, &remoteaddr) != NULL) {
		remotefd = bb_xsocket(
#if ENABLE_FEATURE_IPV6
			(remoteaddr.sa_family == AF_INET6 ? BB_XSOCKET_FLG_AF_INET6 : 0) |
#endif
						BB_XSOCKET_FLG_SOCK_DGRAM| BB_XSOCKET_FLG_ERR_DIE);

		((struct sockaddr_in *)&remoteaddr)->sin_port = htons(RemotePort);
	}
}
#endif

static void logMessage(int pri, char *msg)
{
	time_t now;
	char *timestamp;
	CODE *c_pri, *c_fac;

	if (pri != 0) {
		for (c_fac = facilitynames;
			 c_fac->c_name && !(c_fac->c_val == LOG_FAC(pri) << 3); c_fac++);
		for (c_pri = prioritynames;
			 c_pri->c_name && !(c_pri->c_val == LOG_PRI(pri)); c_pri++);
		if (c_fac->c_name == NULL || c_pri->c_name == NULL) {
			snprintf(res, sizeof(res), "<%d>", pri);
		} else {
			snprintf(res, sizeof(res), "%s.%s", c_fac->c_name, c_pri->c_name);
		}
	}

	if (strlen(msg) < 16 || msg[3] != ' ' || msg[6] != ' ' ||
		msg[9] != ':' || msg[12] != ':' || msg[15] != ' ') {
		time(&now);
		timestamp = ctime(&now) + 4;
		timestamp[15] = '\0';
	} else {
		timestamp = msg;
		timestamp[15] = '\0';
		msg += 16;
	}

	/* todo: supress duplicates */

#ifdef CONFIG_FEATURE_REMOTE_LOG
	if (doRemoteLog == TRUE) {
		/* trying connect the socket */
		if (-1 == remotefd)
			init_RemoteLog();

		/* if we have a valid socket, send the message */
		if (-1 != remotefd) {
			now = 1;
			snprintf(lm_line, sizeof(lm_line), "<%d>%s", pri, msg);

		retry:
			/* send message to remote logger */
			if(( -1 == sendto(remotefd, lm_line, strlen(lm_line), 0,
							&remoteaddr,
							sizeof(remoteaddr))) && (errno == EINTR)) {
				/* sleep now seconds and retry (with now * 2) */
				sleep(now);
				now *= 2;
				goto retry;
			}
		}
	}

	if (local_logging)
#endif
	{
		/* now spew out the message to wherever it is supposed to go */
		if (small)
			message("%s %s\n", timestamp, msg);
		else
			message("%s %s %s %s\n", timestamp, LocalHostName, res, msg);
	}
}

static void quit_signal(int sig)
{
	logMessage(LOG_SYSLOG | LOG_INFO, "System log daemon exiting.");
	unlink(lfile);
#ifdef CONFIG_FEATURE_IPC_SYSLOG
	ipcsyslog_cleanup();
#endif

	exit(TRUE);
}

static void domark(int sig)
{
	if (MarkInterval > 0) {
		logMessage(LOG_SYSLOG | LOG_INFO, "-- MARK --");
		alarm(MarkInterval);
	}
}

/* This must be a #define, since when CONFIG_DEBUG and BUFFERS_GO_IN_BSS are
 * enabled, we otherwise get a "storage size isn't constant error. */
static int serveConnection(char *tmpbuf, int n_read)
{
	char *p = tmpbuf;

	while (p < tmpbuf + n_read) {

		int pri = (LOG_USER | LOG_NOTICE);
		int num_lt = 0;
		char line[MAXLINE + 1];
		unsigned char c;
		char *q = line;

		while ((c = *p) && q < &line[sizeof(line) - 1]) {
			if (c == '<' && num_lt == 0) {
				/* Parse the magic priority number. */
				num_lt++;
				pri = 0;
				while (isdigit(*(++p))) {
					pri = 10 * pri + (*p - '0');
				}
				if (pri & ~(LOG_FACMASK | LOG_PRIMASK)) {
					pri = (LOG_USER | LOG_NOTICE);
				}
			} else if (c == '\n') {
				*q++ = ' ';
			} else if (iscntrl(c) && (c < 0177)) {
				*q++ = '^';
				*q++ = c ^ 0100;
			} else {
				*q++ = c;
			}
			p++;
		}
		*q = '\0';
		p++;
		/* Now log it */
		logMessage(pri, line);
	}
	return n_read;
}


static void doSyslogd(void) ATTRIBUTE_NORETURN;
static void doSyslogd(void)
{
	struct sockaddr_un sunx;
	socklen_t addrLength;

	int sock_fd;
	fd_set fds;

	/* Set up signal handlers. */
	bb_signal( (1 << SIGINT) | (1 << SIGTERM) | (1 << SIGQUIT), quit_signal);
	bb_signal( (1 << SIGHUP) | (1 << SIGCHLD)
#ifdef SIGCLD
			| (1 << SIGCLD)
#endif
			, SIG_IGN);
	bb_signal((1 << SIGALRM), domark);
	alarm(MarkInterval);

	/* Create the syslog file so realpath() can work. */
	if (realpath(_PATH_LOG, lfile) != NULL)
		unlink(lfile);
	else
		strcpy(lfile, _PATH_LOG);

	memset(&sunx, 0, sizeof(sunx));
	sunx.sun_family = AF_UNIX;
	strncpy(sunx.sun_path, lfile, sizeof(sunx.sun_path));
	sock_fd = bb_xsocket(BB_XSOCKET_FLG_AF_UNIX | BB_XSOCKET_FLG_SOCK_DGRAM | BB_XSOCKET_FLG_ERR_DIE);

	addrLength = sizeof(sunx.sun_family) + strlen(sunx.sun_path);
	bb_xbind(sock_fd, (struct sockaddr *) &sunx, addrLength);

	if (chmod(lfile, 0666) < 0)
		bb_perror_msg_and_die("Could not set permission on " _PATH_LOG);

#ifdef CONFIG_FEATURE_IPC_SYSLOG
	if (circular_logging) {
		ipcsyslog_init();
	}
#endif

#ifdef CONFIG_FEATURE_REMOTE_LOG
	if (doRemoteLog == TRUE)
		init_RemoteLog();
#endif

	logMessage(LOG_SYSLOG | LOG_INFO, "syslogd started: " "BusyBox v" BB_VER );

	for (;;) {

		FD_ZERO(&fds);
		FD_SET(sock_fd, &fds);

		if (select(sock_fd + 1, &fds, NULL, NULL, NULL) < 0) {
			if (errno == EINTR) {
				/* alarm may have happened. */
				continue;
			}
			bb_perror_msg_and_die("select error");
		}

		if (FD_ISSET(sock_fd, &fds)) {
			int i;

#define tmpbuf bb_common_bss.tmpbuf
			if ((i = recv(sock_fd, tmpbuf, TMP_BUF_SZ, 0)) > 0) {
				tmpbuf[i] = '\0';
				serveConnection(tmpbuf, i);
			} else {
				bb_perror_msg_and_die("UNIX socket error");
			}
		}                               /* FD_ISSET() */
	}                                       /* for main loop */
}

#define __LOG_FILE "/var/log/messages"

extern int syslogd_main(void)
{
	unsigned long int opt;

#ifdef CONFIG_FEATURE_REMOTE_LOG
	char *p;
#endif
#ifdef CONFIG_FEATURE_IPC_SYSLOG
	char *scircular = NULL;
#endif

	logFilePath = __LOG_FILE;
#ifdef CONFIG_FEATURE_ROTATE_LOGFILE
	logFileSize = 200;
	logFileRotate = 1;
#endif
	MarkInterval = 20;
#ifdef CONFIG_FEATURE_REMOTE_LOG
	remotefd = -1;
	RemotePort = 514;
#endif
#ifdef CONFIG_FEATURE_IPC_SYSLOG
	shmid = -1;
	s_semid = -1;
	shm_size = ((CONFIG_FEATURE_IPC_SYSLOG_BUFFER_SIZE)*1024);
#endif
	/* do normal option parsing */
#if ENABLE_FEATURE_ROTATE_LOGFILE
	opt = bb_xgetopt_ulflags("m:nSO:"
		USE_FEATURE_REMOTE_LOG("R:L")
		USE_FEATURE_IPC_SYSLOG("C::")
		"s:b:\0:s-mb"
		  ,&MarkInterval, &logFilePath
		  USE_FEATURE_REMOTE_LOG(,&RemoteHost)
		  USE_FEATURE_IPC_SYSLOG(,&scircular)
		  ,&logFileSize, &logFileRotate);
#else
	opt = bb_getopt_ulflags("m:nSO:"
		USE_FEATURE_REMOTE_LOG("R:L")
		USE_FEATURE_IPC_SYSLOG("C::")
		  ,&MarkInterval, &logFilePath
		  USE_FEATURE_REMOTE_LOG(,&RemoteHost)
		  USE_FEATURE_IPC_SYSLOG(,&scircular));
#endif

#define foreground (opt & 2)
#define opt_small (opt & 4)
#define opt_local_logging (opt & 32)
#ifdef CONFIG_FEATURE_REMOTE_LOG
# define opt_circular_logging (opt & 64)
#else
# define opt_circular_logging (opt & 16)
#endif

	MarkInterval *= 60;
	small = opt_small;
#ifdef CONFIG_FEATURE_ROTATE_LOGFILE
	logFileSize *= 1024;
	if( logFileRotate > 99 ) logFileRotate = 99;
#endif
#ifdef CONFIG_FEATURE_REMOTE_LOG
	if(RemoteHost) {
		RemoteHost = xstrdup(RemoteHost);
		if ((p = strchr_set0(RemoteHost, ':')))
			RemotePort = atoi(p + 1);
		doRemoteLog = TRUE;
	}
	local_logging = opt_local_logging;
#endif
#ifdef CONFIG_FEATURE_IPC_SYSLOG
	if(scircular) {
		int buf_size = atoi(scircular);
		if (buf_size >= 4)
			shm_size = buf_size * 1024;
	}
	circular_logging = opt_circular_logging;
#endif

#ifdef CONFIG_FEATURE_REMOTE_LOG
	/* If they have not specified remote logging, then log locally */
	if (doRemoteLog == FALSE)
		local_logging = TRUE;
#endif

	/* Store away localhost's name before the fork */
	gethostname(LocalHostName, sizeof(LocalHostName));
	strchr_set0(LocalHostName, '.');

	umask(0);

	if(!foreground)
		xdaemon(XDAEMON_PIDFILE);

	doSyslogd();

	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
