/*
 *  ACPI daemon
 *
 *  Copyright (C) 2008-2009 Ted Felix (www.tedfelix.com)
 *
 *  Portions Copyright (C) 1999-2000 Andrew Henroid
 *  Portions Copyright (C) 2001 Sun Microsystems
 *  Portions Copyright (C) 2004 Tim Hockin (thockin@hockin.org)
 *  Portions Copyright (C) 2001 Sun Microsystems
 *  Portions from acpi_genl Copyright (C) Zhang Rui <rui.zhang@intel.com>
 *  Inspired by (and in some cases blatantly lifted from) Zhang Rui's
 *   acpi_genl and Alexey Kuznetsov's libnetlink.  Thanks also to Yi Yang
 *   at intel.
 *  Inspired by (and in some cases blatantly lifted from) Vojtech Pavlik's
 *   evtest.c.
 *
 *  Busybox port by Vladimir Oleynik (c) 2018 <dzo@simtreas.ru>
 */


#ifdef APPLET_USAGE
#define acpid_trivial_usage \
	"[-f] [-c CONFDIR] [-e EVENTFILE] [-s SOCKFILE]"
#define acpid_full_usage \
	"Advanced Configuration and Power Interface event daemon\n\n" \
	"Options:\n" \
	USE_FEATURE_LONG_OPTIONS( \
	"\t-c, --confdir CONFDIR       Set the configuration directory.\n" \
	"\t-C, --clientmax number      Set the limit on non-root socket connections.\n" \
	"\t-d, --debug                 Increase debugging level (implies -f).\n" \
	"\t-e, --eventfile EVENTFILE   Use the specified file for events.\n" \
	"\t-f, --foreground            Run in the foreground.\n" \
	"\t-g, --socketgroup SOCKETGRP Set the group on the socket file.\n" \
	"\t-l, --logevents             Log all event activity.\n" \
	"\t-L, --lockfile filename     Use the specified lockfile to stop processing.\n" \
	"\t-m, --socketmode SOCKETMODE Set the permissions on the socket file.\n" \
	"\t-n, --netlink               Force netlink/input layer mode. (overrides -e).\n" \
	"\t-p, --pidfile filename      Use the specified PID file.\n" \
	"\t-r, --dropaction DROPACTION Define the pseudo-action to drop an event.\n" \
	"\t-s, --socketfile SOCKETFILE Use the specified socket file.\n" \
	"\t-S, --nosocket              Do not listen on a UNIX socket (overrides -s).\n" \
	"\t-t, --tpmutefix             Fixup for ThinkPad mute-repeat behaviour.\n" \
	"\t-v, --version               Print version information." ) \
	SKIP_FEATURE_LONG_OPTIONS( \
	"\t-c CONFDIR    Set the configuration directory.\n" \
	"\t-C number     Set the limit on non-root socket connections.\n" \
	"\t-d            Increase debugging level (implies -f).\n" \
	"\t-e EVENTFILE  Use the specified file for events.\n" \
	"\t-f            Run in the foreground.\n" \
	"\t-g SOCKETGRP  Set the group on the socket file.\n" \
	"\t-l            Log all event activity.\n" \
	"\t-L filename   Use the specified lockfile to stop processing.\n" \
	"\t-m SOCKETMODE Set the permissions on the socket file.\n" \
	"\t-n            Force netlink/input layer mode. (overrides -e).\n" \
	"\t-p filename   Use the specified PID file.\n" \
	"\t-r DROPACTION Define the pseudo-action to drop an event.\n" \
	"\t-s SOCKETFILE Use the specified socket file.\n" \
	"\t-S            Do not listen on a UNIX socket (overrides -s).\n" \
	"\t-t            Fixup for ThinkPad mute-repeat behaviour.\n" \
	"\t-v            Print version information." )
#else

/* Version number of package */
#define VERSION "2.0.29"

#include "busybox.h"

/* no arg, mode - long, clientmax - int, debug++ */
#define COMPLEMENTARY "\0?0::m-C:dd"
#include "opt_gen.h"
OPT_GEN(
opt(c, CONFDIR_OPT, confdir),
opt(C, CLIENTMAX_OPT, number),
opt(d, DEBUG_OPT),
opt(e, EVENTFILE_OPT, eventfile),
opt(f, FOREGROUND_OPT),
opt(l, LOGEVENTS_OPT),
opt(g, SOCKETGROUP_OPT, socketgrp),
opt(m, SOCKETMODE_OPT, socketmode),
opt(s, SOCKETFILE_OPT, socketfile),
opt(S, NOSOCKET_OPT),
opt(p, PIDFILE_OPT, pidfile),
opt(L, LOCKFILE_OPT, lockfile),
opt(n, NETLINK_OPT),
opt(r, DROPACTION_OPT, dropaction),
opt(t, TPMUTEFIX_OPT),
opt(v, VERSION_OPT),
opt(h, HELP_OPT),
)
#include "opt_gen.h"

#include "grp_.h"
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
/* for LOG_ERR, LOG_INFO, etc... */
#include <syslog.h>
#include <poll.h>
#include <dirent.h>
#include <ctype.h>
#include <regex.h>
#include <malloc.h>
#include <glob.h>

#include <sys/select.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/un.h>
#include <sys/uio.h>
#include <sys/inotify.h>

#include <net/if_arp.h>
#include <netinet/in.h>
/* needed by netlink.h, should be in there */
#include <arpa/inet.h>
#include <asm/types.h>
#include <linux/types.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <linux/input.h>

#include "bb_xopen.h"
#include "bb_sig.h"

/*
 *  libc_compat.h - implement defs/macros missing from some libcs
 *
 */

/* Evaluate EXPRESSION, and repeat as long as it returns -1 with `errno'
   set to EINTR. This macro is present on glibc/uclibc but may not be in other cases.  */

#ifndef ____GLIBC__
#ifndef TEMP_FAILURE_RETRY
#define TEMP_FAILURE_RETRY(expression) \
  (__extension__                                                             \
    ({ long int __result;                                                    \
       do __result = (long int) (expression);                                \
       while (__result == -1L && errno == EINTR);                            \
       __result; }))
#endif
#else
/* Define to 1 if you have the `isfdtype' function. */
#define HAVE_ISFDTYPE 1
#endif /* __GLIBC__ */

#ifndef max
#define max(a, b)  (((a)>(b))?(a):(b))
#endif

struct rtnl_handle
{
	int                     fd;
	struct sockaddr_nl      local;
	struct sockaddr_nl      peer;
	__u32                   seq;
	__u32                   dump;
};

typedef int (*rtnl_filter_t)(const struct sockaddr_nl *,
			     struct nlmsghdr *n, void *);

#define parse_rtattr_nested(tb, max, rta) \
	(parse_rtattr((tb), (max), RTA_DATA(rta), RTA_PAYLOAD(rta)))

#define NLMSG_TAIL(nmsg) \
	((struct rtattr *) (((void *) (nmsg)) + NLMSG_ALIGN((nmsg)->nlmsg_len)))

#ifndef IFA_RTA
#define IFA_RTA(r) \
	((struct rtattr*)(((char*)(r)) + NLMSG_ALIGN(sizeof(struct ifaddrmsg))))
#endif
#ifndef IFA_PAYLOAD
#define IFA_PAYLOAD(n)  NLMSG_PAYLOAD(n,sizeof(struct ifaddrmsg))
#endif

#ifndef IFLA_RTA
#define IFLA_RTA(r) \
	((struct rtattr*)(((char*)(r)) + NLMSG_ALIGN(sizeof(struct ifinfomsg))))
#endif
#ifndef IFLA_PAYLOAD
#define IFLA_PAYLOAD(n) NLMSG_PAYLOAD(n,sizeof(struct ifinfomsg))
#endif

#ifndef NDA_RTA
#define NDA_RTA(r) \
	((struct rtattr*)(((char*)(r)) + NLMSG_ALIGN(sizeof(struct ndmsg))))
#endif
#ifndef NDA_PAYLOAD
#define NDA_PAYLOAD(n)  NLMSG_PAYLOAD(n,sizeof(struct ndmsg))
#endif

#ifndef NDTA_RTA
#define NDTA_RTA(r) \
	((struct rtattr*)(((char*)(r)) + NLMSG_ALIGN(sizeof(struct ndtmsg))))
#endif
#ifndef NDTA_PAYLOAD
#define NDTA_PAYLOAD(n) NLMSG_PAYLOAD(n,sizeof(struct ndtmsg))
#endif


/*****************************************************************
 *  Connection List Public Members
 *****************************************************************/

struct connection
{
	/* file descriptor */
	int fd;

	/* process incoming data on the connection */
	/* ??? suggest passing a pointer to this connection struct */
	void (* process)(int fd);

	/* Optional.  Used by find_connection_name() to find the connection for a
	   specific file.  Set to NULL if not specified.  Memory will be freed
	   with free() when connection is deleted. */
	char *pathname;

	/* 0 indicates this is probably not a keyboard device */
	int kybd;
};

#define GENL_NAMSIZ     16      /* length of family name */

#define GENL_MIN_ID     NLMSG_MIN_TYPE
#define GENL_MAX_ID     1023

struct genlmsghdr {
	__u8    cmd;
	__u8    version;
	__u16   reserved;
};

#define GENL_HDRLEN     NLMSG_ALIGN(sizeof(struct genlmsghdr))

#define GENL_ADMIN_PERM         0x01
#define GENL_CMD_CAP_DO         0x02
#define GENL_CMD_CAP_DUMP       0x04
#define GENL_CMD_CAP_HASPOL     0x08

/*
 * List of reserved static generic netlink identifiers:
 */
#define GENL_ID_GENERATE        0
#define GENL_ID_CTRL            NLMSG_MIN_TYPE

/**************************************************************************
 * Controller
 **************************************************************************/

enum {
	CTRL_CMD_UNSPEC,
	CTRL_CMD_NEWFAMILY,
	CTRL_CMD_DELFAMILY,
	CTRL_CMD_GETFAMILY,
	CTRL_CMD_NEWOPS,
	CTRL_CMD_DELOPS,
	CTRL_CMD_GETOPS,
	CTRL_CMD_NEWMCAST_GRP,
	CTRL_CMD_DELMCAST_GRP,
	CTRL_CMD_GETMCAST_GRP, /* unused */
	__CTRL_CMD_MAX,
};

#define CTRL_CMD_MAX (__CTRL_CMD_MAX - 1)

enum {
	CTRL_ATTR_UNSPEC,
	CTRL_ATTR_FAMILY_ID,
	CTRL_ATTR_FAMILY_NAME,
	CTRL_ATTR_VERSION,
	CTRL_ATTR_HDRSIZE,
	CTRL_ATTR_MAXATTR,
	CTRL_ATTR_OPS,
	CTRL_ATTR_MCAST_GROUPS,
	__CTRL_ATTR_MAX,
};

#define CTRL_ATTR_MAX (__CTRL_ATTR_MAX - 1)

enum {
	CTRL_ATTR_OP_UNSPEC,
	CTRL_ATTR_OP_ID,
	CTRL_ATTR_OP_FLAGS,
	__CTRL_ATTR_OP_MAX,
};

#define CTRL_ATTR_OP_MAX (__CTRL_ATTR_OP_MAX - 1)

enum {
	CTRL_ATTR_MCAST_GRP_UNSPEC,
	CTRL_ATTR_MCAST_GRP_NAME,
	CTRL_ATTR_MCAST_GRP_ID,
	__CTRL_ATTR_MCAST_GRP_MAX,
};
#define CTRL_ATTR_MCAST_GRP_MAX (__CTRL_ATTR_MCAST_GRP_MAX - 1)

#define DROP_ACTION "<drop>"
#define DROP_EVENT 1

struct acpi_genl_event {
	char device_class[20];
	char bus_id[15];
	__u32 type;
	__u32 data;
};

/* attributes of acpi_genl_family */
enum {
	ACPI_GENL_ATTR_UNSPEC,
	ACPI_GENL_ATTR_EVENT,   /* ACPI event info needed by user space */
	__ACPI_GENL_ATTR_MAX,
};
#define ACPI_GENL_ATTR_MAX (__ACPI_GENL_ATTR_MAX - 1)

/* commands supported by the acpi_genl_family */
enum {
	ACPI_GENL_CMD_UNSPEC,
	ACPI_GENL_CMD_EVENT,    /* kernel->user notifications for ACPI events */        __ACPI_GENL_CMD_MAX,
};
#define ACPI_GENL_CMD_MAX (__ACPI_GENL_CMD_MAX - 1)
#define GENL_MAX_FAM_OPS        256
#define GENL_MAX_FAM_GRPS       256

#define ACPI_EVENT_FAMILY_NAME          "acpi_event"
#define ACPI_EVENT_MCAST_GROUP_NAME     "acpi_mc_group"

#define ACPI_PROCDIR            "/proc/acpi"
#define ACPID_EVENTFILE         ACPI_PROCDIR "/event"
#define ACPID_CONFDIR           "/etc/acpi/events"
#define ACPID_SOCKETFILE        "/var/run/acpid.socket"
#define ACPID_SOCKETMODE        0666
#define ACPID_CLIENTMAX         256
#define ACPID_LOCKFILE          "/var/lock/acpid"
#define ACPID_MAX_ERRS          5

/* ??? make these changeable by commandline option? */
#define ACPID_INPUTLAYERDIR   "/dev/input"
#define ACPID_INPUTLAYERFILES ACPID_INPUTLAYERDIR "/event*"

#define PACKAGE                 "acpid"

/* What is a rule?  It's polymorphic, pretty much. */
#define RULE_REGEX_FLAGS (REG_EXTENDED | REG_ICASE | REG_NOSUB | REG_NEWLINE)
struct rule {
	enum {
		RULE_NONE = 0,
		RULE_CMD,
		RULE_CLIENT,
	} type;
	char *origin;
	regex_t *event;
	union {
		char *cmd;
		int fd;
	} action;
	struct rule *next;
	struct rule *prev;
};
struct rule_list {
	struct rule *head;
	struct rule *tail;
};

CALC_BSS struct BB_COMMON_BSS
{
	const char *eventfile;
	const char *lockfile;
	/* do we log event info? */
	int logevents;
	int tpmutefix;
	/* Global debug level. */
	int debug_level;
	int capacity;
	struct connection *connection_list;
	int nconnections;
	/* fd_set containing all the fd's that come in */
	fd_set allfds;
	/* highest fd that is opened */
	/* (-2 + 1) causes select() to return immediately */
	int highestfd;
	mode_t socketmode;
	int initialized;
	__u16 acpi_event_family_id;
	__u32 acpi_event_mcast_group_id;
	const char *socketfile;
	const char *socketgroup;
	int clientmax;
	/* the number of non-root clients that are connected */
	int non_root_clients;
	struct rule_list drop_list;
	struct rule_list client_list;
	struct rule_list cmd_list;
	const char *dropaction;
	const char *confdir;
	int nosocket;
	int foreground;
	const char *pidfile;
	int netlink;
	char parse_cmd_buf[4096];
	/* This depends on fixes in linux ACPI after 2.4.8 */
#define BUFLEN 1024
	char read_line_buf[BUFLEN];
	int process_proc_nerrs;
	int process_netlink_nerrs;
	size_t repeat_count;
	int accept_errors;
	int process_input_nerrs;
} bb_common_bss;
#define eventfile bb_common_bss.eventfile
#define lockfile bb_common_bss.lockfile
#define logevents bb_common_bss.logevents
#define tpmutefix bb_common_bss.tpmutefix
#define debug_level bb_common_bss.debug_level
#define capacity bb_common_bss.capacity
#define connection_list bb_common_bss.connection_list
#define nconnections bb_common_bss.nconnections
#define allfds bb_common_bss.allfds
#define highestfd bb_common_bss.highestfd
#define socketmode bb_common_bss.socketmode
#define initialized bb_common_bss.initialized
#define acpi_event_family_id bb_common_bss.acpi_event_family_id
#define acpi_event_mcast_group_id bb_common_bss.acpi_event_mcast_group_id
#define socketfile bb_common_bss.socketfile
#define socketgroup bb_common_bss.socketgroup
#define clientmax bb_common_bss.clientmax
#define non_root_clients bb_common_bss.non_root_clients
#define drop_list bb_common_bss.drop_list
#define client_list bb_common_bss.client_list
#define cmd_list bb_common_bss.cmd_list
#define dropaction bb_common_bss.dropaction
#define confdir bb_common_bss.confdir
#define nosocket bb_common_bss.nosocket
#define foreground bb_common_bss.foreground
#define pidfile bb_common_bss.pidfile
#define netlink bb_common_bss.netlink
#define parse_cmd_buf bb_common_bss.parse_cmd_buf
#define read_line_buf bb_common_bss.read_line_buf
#define process_proc_nerrs bb_common_bss.process_proc_nerrs
#define process_netlink_nerrs bb_common_bss.process_netlink_nerrs
#define repeat_count bb_common_bss.repeat_count
#define accept_errors bb_common_bss.accept_errors
#define process_input_nerrs bb_common_bss.process_input_nerrs

static void xlog(int type, const char* fmt, ...)
{
	va_list args;
	int prev_log_syslog = log_syslog;

	if(prev_log_syslog)
		log_syslog = type;
	va_start(args, fmt);
	bb_verror_msg(fmt, args);
	va_end(args);
	log_syslog = prev_log_syslog;
}

static void rtnl_close(struct rtnl_handle *rth)
{
	if (rth->fd >= 0) {
		close(rth->fd);
		rth->fd = -1;
	}
}

#ifndef MSG_CMSG_CLOEXEC
#define MSG_CMSG_CLOEXEC 0x40000000
#endif

static int rtnl_open_byproto(struct rtnl_handle *rth, unsigned subscriptions,
		      int protocol)
{
	socklen_t addr_len;
	int sndbuf = 32768;
	int rcvbuf = 32768;

	memset(rth, 0, sizeof(struct rtnl_handle));

	rth->fd = socket(AF_NETLINK, SOCK_RAW | SOCK_CLOEXEC, protocol);
	if (rth->fd < 0) {
		bb_perror_msg("Cannot open netlink socket");
		return -1;
	}

	if (setsockopt(rth->fd,SOL_SOCKET,SO_SNDBUF,&sndbuf,sizeof(sndbuf)) < 0) {
		bb_perror_msg("SO_SNDBUF");
		return -1;
	}

	if (setsockopt(rth->fd,SOL_SOCKET,SO_RCVBUF,&rcvbuf,sizeof(rcvbuf)) < 0) {
		bb_perror_msg("SO_RCVBUF");
		return -1;
	}

	memset(&rth->local, 0, sizeof(rth->local));
	rth->local.nl_family = AF_NETLINK;
	rth->local.nl_groups = subscriptions;

	if (bind(rth->fd, (struct sockaddr*)&rth->local, sizeof(rth->local)) < 0) {
		bb_perror_msg("Cannot bind netlink socket");
		return -1;
	}
	addr_len = sizeof(rth->local);
	if (getsockname(rth->fd, (struct sockaddr*)&rth->local, &addr_len) < 0) {
		bb_perror_msg("Cannot getsockname");
		return -1;
	}
	if (addr_len != sizeof(rth->local)) {
		bb_error_msg("Wrong address length %d", addr_len);
		return -1;
	}
	if (rth->local.nl_family != AF_NETLINK) {
		bb_error_msg("Wrong address family %d", rth->local.nl_family);
		return -1;
	}
	rth->seq = time(NULL);
	return 0;
}

static int rtnl_talk(struct rtnl_handle *rtnl, struct nlmsghdr *n, pid_t peer,
	      unsigned groups, struct nlmsghdr *answer,
	      rtnl_filter_t junk,
	      void *jarg)
{
	int status;
	unsigned seq;
	struct nlmsghdr *h;
	struct sockaddr_nl nladdr;
	struct iovec iov = {
		.iov_base = (void*) n,
		.iov_len = n->nlmsg_len
	};
	struct msghdr msg = {
		.msg_name = &nladdr,
		.msg_namelen = sizeof(nladdr),
		.msg_iov = &iov,
		.msg_iovlen = 1,
	};
	char   buf[16384];

	memset(&nladdr, 0, sizeof(nladdr));
	nladdr.nl_family = AF_NETLINK;
	nladdr.nl_pid = peer;
	nladdr.nl_groups = groups;

	n->nlmsg_seq = seq = ++rtnl->seq;

	if (answer == NULL)
		n->nlmsg_flags |= NLM_F_ACK;

	status = sendmsg(rtnl->fd, &msg, 0);

	if (status < 0) {
		bb_perror_msg("Cannot talk to rtnetlink");
		return -1;
	}

	memset(buf,0,sizeof(buf));

	iov.iov_base = buf;

	while (1) {
		iov.iov_len = sizeof(buf);
		status = TEMP_FAILURE_RETRY ( recvmsg(rtnl->fd, &msg, MSG_CMSG_CLOEXEC) );

		if (status < 0) {
			bb_error_msg("OVERRUN");
			continue;
		}
		if (status == 0) {
			bb_error_msg("EOF on netlink");
			return -1;
		}
		if (msg.msg_namelen != sizeof(nladdr))
			bb_error_msg_and_die("sender address length == %d", msg.msg_namelen);
		for (h = (struct nlmsghdr*)buf; (unsigned)status >= sizeof(*h); ) {
			int err;
			int len = h->nlmsg_len;
			int l = len - sizeof(*h);

			if (l<0 || len>status) {
				if (msg.msg_flags & MSG_TRUNC) {
					bb_error_msg("Truncated message");
					return -1;
				}
				bb_error_msg_and_die("!!!malformed message: len=%d", len);
			}

			if (nladdr.nl_pid != (unsigned)peer ||
			    h->nlmsg_pid != rtnl->local.nl_pid ||
			    h->nlmsg_seq != seq) {
				if (junk) {
					err = junk(&nladdr, h, jarg);
					if (err < 0)
						return err;
				}
				/* Don't forget to skip that message. */
				status -= NLMSG_ALIGN(len);
				h = (struct nlmsghdr*)((char*)h + NLMSG_ALIGN(len));
				continue;
			}

			if (h->nlmsg_type == NLMSG_ERROR) {
				struct nlmsgerr *msgerr = (struct nlmsgerr*)NLMSG_DATA(h);
				if ((unsigned)l < sizeof(struct nlmsgerr)) {
					bb_error_msg("ERROR truncated");
				} else {
					errno = -msgerr->error;
					if (errno == 0) {
						if (answer)
							memcpy(answer, h, h->nlmsg_len);
						return 0;
					}
					bb_perror_msg("RTNETLINK1 answers");
				}
				return -1;
			}
			if (answer) {
				memcpy(answer, h, h->nlmsg_len);
				return 0;
			}

			bb_error_msg("Unexpected reply!!!");

			status -= NLMSG_ALIGN(len);
			h = (struct nlmsghdr*)((char*)h + NLMSG_ALIGN(len));
		}
		if (msg.msg_flags & MSG_TRUNC) {
			bb_error_msg("Message truncated");
			continue;
		}
		if (status)
			bb_error_msg_and_die("!!!Remnant of size %d", status);
	}
}

static int addattr_l(struct nlmsghdr *n, int maxlen, int type, const void *data,
	      int alen)
{
	int len = RTA_LENGTH(alen);
	struct rtattr *rta;

	if ((int)NLMSG_ALIGN(n->nlmsg_len) + (int)RTA_ALIGN(len) > maxlen) {
		bb_error_msg("addattr_l ERROR: message exceeded bound of %d", maxlen);
		return -1;
	}
	rta = NLMSG_TAIL(n);
	rta->rta_type = type;
	rta->rta_len = len;
	memcpy(RTA_DATA(rta), data, alen);
	n->nlmsg_len = NLMSG_ALIGN(n->nlmsg_len) + RTA_ALIGN(len);
	return 0;
}

static int parse_rtattr(struct rtattr *tb[], int max, struct rtattr *rta, int len)
{
	memset(tb, 0, sizeof(struct rtattr *) * (max + 1));
	while (RTA_OK(rta, len)) {
		if (rta->rta_type <= max)
			tb[rta->rta_type] = rta;
		rta = RTA_NEXT(rta,len);
	}
	if (len)
		bb_error_msg("!!!Deficit %d, rta_len=%d", len, rta->rta_len);
	return 0;
}


/* add a connection to the list */
static int add_connection(struct connection *p)
{
	if (nconnections < 0)
		return -1;

	/* if the list is full, allocate more space */
	if (nconnections >= capacity) {
		/* no more than 1024 */
		if (capacity > 1024) {
			bb_error_msg("Too many connections.");
			return -1;
		}

		/* another 20 */
		capacity += 20;
		connection_list = realloc(connection_list, sizeof(struct connection) * capacity);
	}

	if (nconnections == 0)
		FD_ZERO(&allfds);

	/* add the connection to the connection list */
	connection_list[nconnections] = *p;
	++nconnections;

	/* add to the fd set */
	FD_SET(p->fd, &allfds);
	highestfd = max(highestfd, p->fd);

	return 0;
}

/* delete a connection from the list */
static void delete_connection(int fd)
{
	int i;

	/* close anything other than stdin/stdout/stderr */
	if (fd > 2)
		close(fd);

	/* remove from the fd set */
	FD_CLR(fd, &allfds);

	for (i = 0; i < nconnections; ++i) {
		/* if the file descriptors match, delete the connection */
		if (connection_list[i].fd == fd) {
			free(connection_list[i].pathname);

			--nconnections;
			connection_list[i] = connection_list[nconnections];

			break;
		}
	}

	/* prepare for recalculation of highestfd */
	highestfd = -2;

	/* recalculate highestfd */
	for (i = 0; i < nconnections; ++i) {
		highestfd = max(highestfd, connection_list[i].fd);
	}
}



/* find a connection in the list by file descriptor */
/* ??? This routine is unnecessary.  When we call the connection's process
 *     routine, we should pass a pointer to the connection.  That will have
 *     the usual fd along with everything else. */
static struct connection *find_connection(int fd)
{
	int i;

	/* for each connection */
	for (i = 0; i < nconnections; ++i) {
		/* if the file descriptors match, return the connection */
		if (connection_list[i].fd == fd)
			return &connection_list[i];
	}

	return NULL;
}

/* get a specific connection by index from the list */
static struct connection *get_connection(int i)
{
	if (i < 0  ||  i >= nconnections)
		return NULL;

	return &connection_list[i];
}

/* get an fd_set with all the fd's that have been added to the list */
static inline const fd_set *get_fdset(void)
{
	return &allfds;
}

static int ud_create_socket(const char *name)
{
	int fd;
	int r;
	struct sockaddr_un uds_addr;

	if (strnlen(name, sizeof(uds_addr.sun_path)) > sizeof(uds_addr.sun_path) - 1) {
		bb_error_msg("ud_create_socket(): "
			"socket filename longer than %zu characters: %s",
			sizeof(uds_addr.sun_path) - 1, name);
		errno = EINVAL;
		return -1;
	}

	/* JIC */
	unlink(name);

	fd = socket(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC | SOCK_NONBLOCK, 0);
	if (fd < 0)
		return fd;

	/* Clear the umask to guarantee predictable results from fchmod(). */
	umask(0);

	if (fchmod(fd, socketmode) < 0) {
		close(fd);
		bb_perror_msg("fchmod() on socket %s", name);
		return -1;
	}

	/* setup address struct */
	memset(&uds_addr, 0, sizeof(uds_addr));
	uds_addr.sun_family = AF_UNIX;
	strncpy(uds_addr.sun_path, name, sizeof(uds_addr.sun_path) - 1);

	/* bind it to the socket */
	r = bind(fd, (struct sockaddr *)&uds_addr, sizeof(uds_addr));
	if (r < 0) {
		close (fd);
		return r;
	}

	/* listen - allow 10 to queue */
	r = listen(fd, 10);
	if (r < 0) {
		close(fd);
		return r;
	}

	return fd;
}

static int ud_accept(int listenfd, struct ucred *cred)
{
	while (1) {
		int newsock = 0;
		struct sockaddr_un cliaddr;
		socklen_t len = sizeof(struct sockaddr_un);

		newsock = TEMP_FAILURE_RETRY (accept4(listenfd, (struct sockaddr *)&cliaddr, &len, SOCK_CLOEXEC|SOCK_NONBLOCK));
		if (newsock < 0) {
			return newsock;
		}
		if (cred) {
			len = sizeof(struct ucred);
			getsockopt(newsock,SOL_SOCKET,SO_PEERCRED,cred,&len);
		}

		return newsock;
	}
}

static int ud_get_peercred(int fd, struct ucred *cred)
{
	socklen_t len = sizeof(struct ucred);
	getsockopt(fd, SOL_SOCKET, SO_PEERCRED, cred, &len);
	return 0;
}


/*
 *  A CTRL_CMD_GETFAMILY message returns an attribute table that looks
 *    like this:
 *
 *  CTRL_ATTR_FAMILY_ID         Use this to make sure we get the proper msgs
 *  CTRL_ATTR_MCAST_GROUPS
 *    CTRL_ATTR_MCAST_GRP_NAME
 *    CTRL_ATTR_MCAST_GRP_ID    Need this for the group mask
 *    ...
 */
static int get_ctrl_grp_id(struct rtattr *arg)
{
	struct rtattr *tb[CTRL_ATTR_MCAST_GRP_MAX + 1];
	char *name;

	if (arg == NULL)
		return -1;

	/* nested within the CTRL_ATTR_MCAST_GROUPS attribute are the  */
	/* group name and ID  */
	parse_rtattr_nested(tb, CTRL_ATTR_MCAST_GRP_MAX, arg);

	/* if either of the entries needed cannot be found, bail */
	if (!tb[CTRL_ATTR_MCAST_GRP_NAME] || !tb[CTRL_ATTR_MCAST_GRP_ID])
		return -1;

	/* get the name of this multicast group we've found */
	name = RTA_DATA(tb[CTRL_ATTR_MCAST_GRP_NAME]);

	/* if it does not match the ACPI event multicast group name, bail */
	if (strcmp(name, ACPI_EVENT_MCAST_GROUP_NAME))
		return -1;

	/* At this point, we've found what we were looking for.  We now  */
	/* have the multicast group ID for ACPI events over generic netlink. */
	acpi_event_mcast_group_id =
		*((__u32 *)RTA_DATA(tb[CTRL_ATTR_MCAST_GRP_ID]));

	return 0;
}

/* n = the response to a CTRL_CMD_GETFAMILY message */
static int genl_get_mcast_group_id(struct nlmsghdr *n)
{
	/*
	 *  Attribute table.  Note the type name "rtattr" which means "route
	 *  attribute".  This is a vestige of one of netlink's main uses:
	 *  routing.
	 */
	struct rtattr *tb[CTRL_ATTR_MAX + 1];
	/* place for the generic netlink header in the incoming message */
	struct genlmsghdr ghdr;
	/* length of the attribute and payload */
	int len = n->nlmsg_len - NLMSG_LENGTH(GENL_HDRLEN);
	/* Pointer to the attribute portion of the message */
	struct rtattr *attrs;

	if (len < 0) {
		bb_error_msg("netlink CTRL_CMD_GETFAMILY response, "
			"wrong controller message len: %d", len);
		return -1;
	}

	if (n->nlmsg_type != GENL_ID_CTRL) {
		bb_error_msg("not a netlink controller message, "
			"nlmsg_len=%d nlmsg_type=0x%x",
			n->nlmsg_len, n->nlmsg_type);
		return 0;
	}

	/* copy generic netlink header into structure */
	memcpy(&ghdr, NLMSG_DATA(n), GENL_HDRLEN);

	if (ghdr.cmd != CTRL_CMD_GETFAMILY &&
	    ghdr.cmd != CTRL_CMD_DELFAMILY &&
	    ghdr.cmd != CTRL_CMD_NEWFAMILY &&
	    ghdr.cmd != CTRL_CMD_NEWMCAST_GRP &&
	    ghdr.cmd != CTRL_CMD_DELMCAST_GRP) {
		bb_error_msg("unknown netlink controller command %d", ghdr.cmd);
		return 0;
	}

	/* set attrs to point to the attribute */
	attrs = (struct rtattr *)(NLMSG_DATA(n) + GENL_HDRLEN);
	/* Read the table from the message into "tb".  This actually just  */
	/* places pointers into the message into tb[].  */
	parse_rtattr(tb, CTRL_ATTR_MAX, attrs, len);

	/* if a family ID attribute is present, get it */
	if (tb[CTRL_ATTR_FAMILY_ID])
		acpi_event_family_id = *((__u32 *)RTA_DATA(tb[CTRL_ATTR_FAMILY_ID]));

	/* if a "multicast groups" attribute is present... */
	if (tb[CTRL_ATTR_MCAST_GROUPS]) {
		struct rtattr *tb2[GENL_MAX_FAM_GRPS + 1];
		int i;

		/* get the group table within this attribute  */
		parse_rtattr_nested(tb2, GENL_MAX_FAM_GRPS,
			tb[CTRL_ATTR_MCAST_GROUPS]);

		/* for each group */
		for (i = 0; i < GENL_MAX_FAM_GRPS; i++)
			/* if this group is valid */
			if (tb2[i])
				/* Parse the ID.  If successful, we're done. */
				if (!get_ctrl_grp_id(tb2[i]))
					return 0;
	}

	return -1;
}

static int genl_get_ids(char *family_name)
{
	/* handle to the netlink connection */
	struct rtnl_handle rth;
	/* holds the request we are going to send and the reply */
	struct {
		struct nlmsghdr n;
		char buf[4096];    /* ??? Is this big enough for all cases? */
	} req;
	/* pointer to the nlmsghdr in req */
	struct nlmsghdr *nlh;
	/* place for the generic netlink header before copied into req */
	struct genlmsghdr ghdr;
	/* return value */
	int ret = -1;

	/* clear out the request */
	memset(&req, 0, sizeof(req));

	/* set up nlh to point to the netlink header in req */
	nlh = &req.n;
	/* set up the netlink header */
	nlh->nlmsg_len = NLMSG_LENGTH(GENL_HDRLEN);
	nlh->nlmsg_flags = NLM_F_REQUEST | NLM_F_ACK;
	nlh->nlmsg_type = GENL_ID_CTRL;

	/* clear out the generic netlink message header */
	memset(&ghdr, 0, sizeof(struct genlmsghdr));
	/* set the command we want to run: "GETFAMILY" */
	ghdr.cmd = CTRL_CMD_GETFAMILY;
	/* copy it into req */
	memcpy(NLMSG_DATA(&req.n), &ghdr, GENL_HDRLEN);

	/* the message payload is the family name */
	addattr_l(nlh, 128, CTRL_ATTR_FAMILY_NAME,
			  family_name, strlen(family_name) + 1);

	/* open a generic netlink connection */
	if (rtnl_open_byproto(&rth, 0, NETLINK_GENERIC) < 0) {
		bb_error_msg("cannot open generic netlink socket");
		return -1;
	}

	/*
	 *  Send CTRL_CMD_GETFAMILY message (in nlh) to the generic
	 *  netlink controller.  Reply will be in nlh upon return.
	 */
	if (rtnl_talk(&rth, nlh, 0, 0, nlh, NULL, NULL) < 0) {
		bb_error_msg("error talking to the kernel via netlink");
		goto ctrl_done;
	}

	/* process the response */
	if (genl_get_mcast_group_id(nlh) < 0) {
		bb_error_msg("failed to get acpi_event netlink multicast group");
		goto ctrl_done;
	}

	ret = 0;

ctrl_done:
	rtnl_close(&rth);
	return ret;
}

/* initialize the ACPI IDs */
static void acpi_ids_init(void)
{
	genl_get_ids(ACPI_EVENT_FAMILY_NAME);

	initialized = 1;
}

/* returns the netlink family ID for ACPI event messages */
static __u16 acpi_ids_getfamily(void)
{
	/* if the IDs haven't been initialized, initialize them */
	if (initialized == 0)
		acpi_ids_init();

	return acpi_event_family_id;
}

/* returns the netlink multicast group ID for ACPI event messages */
static __u32 acpi_ids_getgroup(void)
{
	/* if the IDs haven't been initialized, initialize them */
	if (initialized == 0)
		acpi_ids_init();

	return acpi_event_mcast_group_id;
}

#ifndef HAVE_ISFDTYPE
static int acpid_isfdtype(int fd, int fdtype)
{
	struct stat64 st;
	if (fstat64(fd, &st) != 0)
		return -1;
	return ((st.st_mode & S_IFMT) == (mode_t)fdtype);
}
#else
#define acpid_isfdtype(fd, fdtype) isfdtype(fd, fdtype)
#endif

/* determine if a file descriptor is in fact a socket */
static int is_socket(int fd)
{
    return (acpid_isfdtype(fd, S_IFSOCK) == 1);
}

static void lock_rules(void)
{
	bb_sigblock((1<<SIGHUP)|(1<<SIGTERM)|(1<<SIGQUIT)|(1<<SIGINT));
}

static void unlock_rules(void)
{
	bb_sigunblock(bb_sigmask((1<<SIGHUP)|(1<<SIGTERM)|(1<<SIGQUIT)|(1<<SIGINT)));
}


/* I hope you delisted the rule before you free() it */
static void free_rule(struct rule *r)
{
	if (r->type == RULE_CMD) {
		if (r->action.cmd)
			free(r->action.cmd);
	}

	if (r->origin)
		free(r->origin);
	if (r->event) {
		regfree(r->event);
		free(r->event);
	}

	free(r);
}

static void delist_rule(struct rule_list *list, struct rule *r)
{
	if (r->next)
		r->next->prev = r->prev;
	 else
		list->tail = r->prev;

	if (r->prev)
		r->prev->next = r->next;
	 else
		list->head = r->next;;

	r->next = r->prev = NULL;
}

/* cleanup all rules */
static int acpid_cleanup_rules(int do_detach)
{
	struct rule *p;
	struct rule *next;

	lock_rules();

	if (do_detach) {
		/* tell our clients to buzz off */
		p = client_list.head;
		while (p) {
			next = p->next;
			delist_rule(&client_list, p);
			close(p->action.fd);
			free_rule(p);
			p = next;
		}
	}

	/* clear out our conf rules */
	p = cmd_list.head;
	while (p) {
		next = p->next;
		delist_rule(&cmd_list, p);
		free_rule(p);
		p = next;
	}

	/* drop the drop rules */
	p = drop_list.head;
	while (p) {
		next = p->next;
		delist_rule(&drop_list, p);
		free_rule(p);
		p = next;
	}

	unlock_rules();

	return 0;
}

static int check_escapes(const char *str)
{
	const char *p;
	int r = 0;

	p = str;
	while (*p) {
		/* found an escape */
		if (*p == '%') {
			p++;
			if (!*p) {
				xlog(LOG_WARNING, "invalid escape at EOL");
				return -1;
			} else if (*p != '%' && *p != 'e') {
				xlog(LOG_WARNING, "invalid escape \"%%%c\"", *p);
				r = -1;
			}
		}
		p++;
	}
	return r;
}

static struct rule *new_rule(void)
{
	struct rule *r;

	r = malloc(sizeof(*r));
	if (!r) {
		bb_perror_msg("malloc()");
		return NULL;
	}

	r->type = RULE_NONE;
	r->origin = NULL;
	r->event = NULL;
	r->action.cmd = NULL;
	r->prev = r->next = NULL;

	return r;
}

static char *read_line(int fd)
{
	size_t i = 0;
	int r;

	while (1) {
		r = safe_read(fd, read_line_buf+i, 1);
		if (r < 0) {
			/* we should do something with the data */
			bb_perror_msg("read()");
			break;
		} else if (r == 0) {
			/* signal this in an almost standard way */
			errno = EPIPE;
			break;
		} else {
			/* scan for a newline */
			if (read_line_buf[i] == '\n')
				break;
			i++;
			/* only go to BUFLEN-1 so there will always be a 0 at the end */
			if (i >= BUFLEN - 1)
				break;
		}
	}

	read_line_buf[i] = '\0';
	if(r <= 0)
		return NULL;
	return read_line_buf;
}

static struct rule *parse_file(int fd_rule, const char *file)
{
	int line = 0;
	struct rule *r;
	char *p;

	/* make a new rule */
	r = new_rule();
	if (!r)
		return NULL;
	r->type = RULE_CMD;
	r->origin = strdup(file);
	if (!r->origin) {
		bb_perror_msg("strdup()");
		free_rule(r);
		return NULL;
	}

	/* read each line */
	while ((p = read_line(fd_rule)) != NULL) {
		char key[64];
		char val[512];
		int n;

		line++;
		memset(key, 0, sizeof(key));
		memset(val, 0, sizeof(val));

		/* skip leading whitespace */
		p = bb_skip_whitespace(p);
		/* blank lines and comments get ignored */
		if (!*p || *p == '#')
			continue;

		/* quick parse */
		n = sscanf(p, "%63[^=\n]=%255[^\n]", key, val);
		if (n != 2) {
			xlog(LOG_WARNING, "can't parse %s at line %d", file, line);
			continue;
		}
		/* handle the parsed line */
		if (!strcasecmp(key, "event")) {
			int rv;
			r->event = malloc(sizeof(regex_t));
			if (!r->event) {
				bb_perror_msg("malloc()");
				free_rule(r);
				return NULL;
			}
			rv = regcomp(r->event, val, RULE_REGEX_FLAGS);
			if (rv) {
				char rbuf[128];
				regerror(rv, r->event, rbuf, sizeof(rbuf));
				bb_error_msg("regcomp(): %s", rbuf);
				free_rule(r);
				return NULL;
			}
		} else if (!strcasecmp(key, "action")) {
			if (check_escapes(val) < 0) {
				bb_error_msg("can't load file %s", file);
				free_rule(r);
				return NULL;
			}
			r->action.cmd = strdup(val);
			if (!r->action.cmd) {
				bb_perror_msg("strdup()");
				free_rule(r);
				return NULL;
			}
		} else {
			xlog(LOG_WARNING, "unknown option '%s' in %s at line %d", key, file, line);
			continue;
		}
	}
	if (!r->event || !r->action.cmd) {
		xlog(LOG_INFO, "skipping incomplete file %s", file);
		free_rule(r);
		return NULL;
	}

	return r;
}

static struct rule *parse_client(int client)
{
	struct rule *r;
	int rv;

	/* make a new rule */
	r = new_rule();
	if (!r)
		return NULL;
	r->type = RULE_CLIENT;
	r->action.fd = client;
	r->event = malloc(sizeof(regex_t));
	if (!r->event) {
		bb_perror_msg("malloc()");
		free_rule(r);
		return NULL;
	}
	rv = regcomp(r->event, ".*", RULE_REGEX_FLAGS);
	if (rv) {
		char buf[128];
		regerror(rv, r->event, buf, sizeof(buf));
		bb_error_msg("regcomp(): %s", buf);
		free_rule(r);
		return NULL;
	}

	return r;
}

static void enlist_rule(struct rule_list *list, struct rule *r)
{
	r->next = r->prev = NULL;
	if (!list->head) {
		list->head = list->tail = r;
	} else {
		list->tail->next = r;
		r->prev = list->tail;
		list->tail = r;
	}
}

static int client_is_dead(int fd)
{
	struct pollfd pfd;
	int r;

	/* check the fd to see if it is dead */
	pfd.fd = fd;
	pfd.events = POLLERR | POLLHUP;
	r = poll(&pfd, 1, 0);

	if (r < 0) {
		bb_perror_msg("poll()");
		return 0;
	}

	return pfd.revents;
}

static void acpid_close_dead_clients(void)
{
	struct rule *p;

	lock_rules();

	/* scan our client list */
	p = client_list.head;
	while (p) {
		struct rule *next = p->next;
		if (client_is_dead(p->action.fd)) {
			struct ucred cred;
			/* closed */
			xlog(LOG_NOTICE, "client %s has disconnected", p->origin);
			delist_rule(&client_list, p);
			ud_get_peercred(p->action.fd, &cred);
			if (cred.uid != 0)
				non_root_clients--;
			close(p->action.fd);
			free_rule(p);
		}
		p = next;
	}

	unlock_rules();
}

static char *parse_cmd(const char *cmd, const char *event)
{
	size_t i;
	const char *p;

	p = cmd;
	i = 0;

	memset(parse_cmd_buf, 0, sizeof(parse_cmd_buf));
	while (i < (sizeof(parse_cmd_buf)-1)) {
		if (*p == '%') {
			p++;
			if (*p == 'e') {
				/* handle an event expansion */
				size_t size = sizeof(parse_cmd_buf) - i;
				size = snprintf(parse_cmd_buf+i, size, "%s", event);
				i += size;
				p++;
				continue;
			}
		}
		if (!*p) {
			break;
		}
		parse_cmd_buf[i++] = *p++;
	}

	return parse_cmd_buf;
}

/* the meat of the rules */
static int do_cmd_rule(struct rule *rule, const char *event)
{
	pid_t pid;
	int status;
	const char *action;

	if (!strcmp(rule->action.cmd, dropaction))
		return DROP_EVENT;

	pid = fork();
	switch (pid) {
	case -1:
		bb_perror_msg("fork()");
		return -1;
	case 0: /* child */
		/* parse the commandline, doing any substitutions needed */
		action = parse_cmd(rule->action.cmd, event);
		if (logevents)
			xlog(LOG_INFO, "executing action \"%s\"", action);

		/* reset signals */
		bb_signal((1<<SIGHUP)|(1<<SIGTERM)|(1<<SIGQUIT)|(1<<SIGINT)|(1<<SIGPIPE), SIG_DFL);
		bb_sigunblock(bb_sigmask((1<<SIGHUP)|(1<<SIGTERM)|(1<<SIGQUIT)|(1<<SIGINT)));

		if (debug_level && logevents)
			printf("BEGIN HANDLER MESSAGES\n");
		umask(0077);
		execl("/bin/sh", "/bin/sh", "-c", action, NULL);
		/* should not get here */
		bb_perror_msg("execl()");
		_exit(EXIT_FAILURE);
	}

	/* parent */
	waitpid(pid, &status, 0);
	if (debug_level && logevents)
		printf("END HANDLER MESSAGES\n");

	if (logevents) {
		if (WIFEXITED(status)) {
			xlog(LOG_INFO, "action exited with status %d", WEXITSTATUS(status));
		} else if (WIFSIGNALED(status)) {
			xlog(LOG_INFO, "action exited on signal %d", WTERMSIG(status));
		} else {
			xlog(LOG_INFO, "action exited with status %d", status);
		}
	}

	return 0;
}

#define NTRIES 100
static int acpid_safe_write(int fd, const char *buf, int len)
{
	int r;
	int ttl = 0;
	int ntries = NTRIES;

	do {
		r = TEMP_FAILURE_RETRY (write(fd, buf+ttl, len-ttl) );
		if (r < 0) {
			if (errno != EAGAIN) {
				/* a legit error */
				return r;
			}
			ntries--;
		} else if (r > 0) {
			/* as long as we make forward progress, reset ntries */
			ntries = NTRIES;
			ttl += r;
		}
	} while (ttl < len && ntries);

	if (!ntries) {
		if (debug_level >= 2)
			bb_error_msg("acpid_safe_write() timed out");
		return r;
	}

	return ttl;
}

static int do_client_rule(struct rule *rule, const char *event)
{
	int r;
	int client = rule->action.fd;

	if (logevents)
		xlog(LOG_INFO, "notifying client %s", rule->origin);

	r = acpid_safe_write(client, event, strlen(event));
	if (r < 0 && errno == EPIPE) {
		struct ucred cred;
		/* closed */
		xlog(LOG_NOTICE, "client %s has disconnected", rule->origin);
		delist_rule(&client_list, rule);
		ud_get_peercred(rule->action.fd, &cred);
		if (cred.uid != 0)
			non_root_clients--;
		close(rule->action.fd);
		free_rule(rule);
		return -1;
	}
	acpid_safe_write(client, "\n", 1);

	return 0;
}

static void print_nrules(int nrules, const char *type, const char *act)
{
	xlog(LOG_INFO, "%d %s%srule%s %s", nrules,
		(type != NULL ? type : ""),
		(type != NULL ? " " : ""),
		(nrules == 1 ? "" : "s"), act);
}

static int locked(void)
{
	struct stat trash;

	/* check for existence of a lockfile */
	return (stat(lockfile, &trash) == 0);
}

/* the main hook for propagating events */
static int acpid_handle_event(const char *event)
{
	struct rule *p;
	int nrules = 0;
	struct rule_list *ar[] = { &drop_list, &client_list, &cmd_list, NULL };
	struct rule_list **lp;

	/* make an event atomic wrt known signals */
	lock_rules();

	/* scan each rule list for any rules that care about this event */
	for (lp = ar; *lp; lp++) {
		struct rule_list *l = *lp;
		p = l->head;
		while (p) {
			/* the list can change underneath us */
			struct rule *pnext = p->next;
			if (!regexec(p->event, event, 0, NULL, 0)) {
				/* a match! */
				if (logevents)
					xlog(LOG_INFO, "rule from %s matched", p->origin);
				nrules++;
				if (p->type == RULE_CMD) {
					if (do_cmd_rule(p, event) == DROP_EVENT) {
						/* Abort processing if event matches drop rule */
						if (logevents)
							xlog(LOG_INFO, "event dropped");
						/* Skip the remaining rules. */
						while (*++lp)
							;
						--lp;
						break;
					}
				} else if (p->type == RULE_CLIENT) {
					do_client_rule(p, event);
				} else {
					xlog(LOG_WARNING, "unknown rule type: %d", p->type);
				}
			} else {
				if (debug_level >= 3 && logevents)
					xlog(LOG_INFO, "rule from %s did not match", p->origin);
			}
			p = pnext;
		}
	}

	unlock_rules();

	if (logevents)
		print_nrules(nrules, "total", "matched");

	return 0;
}


static void process_proc(int fd)
{
	char *event;

	/* read an event */
	event = read_line(fd);

	/* if we're locked, don't process the event */
	if (locked()) {
		if (logevents  &&  event != NULL)
			xlog(LOG_INFO, "lockfile present, not processing event \"%s\"", event);
		return;
	}

	/* handle the event */
	if (event) {
		if (logevents)
			xlog(LOG_INFO, "procfs received event \"%s\"", event);
		acpid_handle_event(event);
		if (logevents)
			xlog(LOG_INFO, "procfs completed event \"%s\"", event);
	} else if (errno == EPIPE) {
		if(log_syslog)
			log_syslog = LOG_WARNING;
		bb_error_msg_and_die("events file connection closed");
	} else {
		if (++process_proc_nerrs >= ACPID_MAX_ERRS)
			bb_error_msg_and_die("too many errors reading events file - aborting");
	}
}

static int open_proc(void)
{
	int fd;
	struct connection c;

	/* O_CLOEXEC: Make sure scripts we exec() (in event.c) don't get our file descriptors. */
	fd = open(eventfile, O_RDONLY | O_CLOEXEC);
	if (fd < 0) {
		if (errno != ENOENT)
			bb_perror_msg("can't open %s", eventfile);
		return -1;

	}

	/* add a connection to the list */
	c.fd = fd;
	c.process = process_proc;
	c.pathname = NULL;
	c.kybd = 0;

	if (add_connection(&c) < 0) {
		close(fd);
		bb_error_msg("can't add connection for %s", eventfile);
		return -1;
	}

	return 0;
}


static void format_netlink(struct nlmsghdr *msg)
{
	struct rtattr *tb[ACPI_GENL_ATTR_MAX + 1];
	struct genlmsghdr *ghdr = NLMSG_DATA(msg);
	int len;
	struct rtattr *attrs;

	len = msg->nlmsg_len;

	/* if this message doesn't have the proper family ID, drop it */
	if (msg->nlmsg_type != acpi_ids_getfamily()) {
		if (logevents)
			xlog(LOG_INFO, "wrong netlink family ID.");
		return;
	}

	len -= NLMSG_LENGTH(GENL_HDRLEN);

	if (len < 0) {
		xlog(LOG_WARNING, "wrong netlink controller message len: %d", len);
		return;
	}

	attrs = (struct rtattr *)((char *)ghdr + GENL_HDRLEN);
	/* parse the attributes in this message */
	parse_rtattr(tb, ACPI_GENL_ATTR_MAX, attrs, len);

	/* if there's an ACPI event attribute... */
	if (tb[ACPI_GENL_ATTR_EVENT]) {
		/* get the actual event struct */
		struct acpi_genl_event *event =
				RTA_DATA(tb[ACPI_GENL_ATTR_EVENT]);
		char buf[64];

		/* format it */
		snprintf(buf, sizeof(buf), "%s %s %08x %08x",
			event->device_class, event->bus_id, event->type, event->data);

		/* if we're locked, don't process the event */
		if (locked()) {
			if (logevents) {
				xlog(LOG_INFO, "lockfile present, not processing "
					"netlink event \"%s\"", buf);
			}
			return;
		}

		if (logevents)
			xlog(LOG_INFO, "received netlink event \"%s\"", buf);

		/* send the event off to the handler */
		acpid_handle_event(buf);

		if (logevents)
			xlog(LOG_INFO, "completed netlink event \"%s\"", buf);
	}
}

/* (based on rtnl_listen() in libnetlink.c) */
static void process_netlink(int fd)
{
	int status;
	struct nlmsghdr *h;
	/* the address for recvmsg() */
	struct sockaddr_nl nladdr;
	/* the io vector for recvmsg() */
	struct iovec iov;
	/* recvmsg() parameters */
	struct msghdr msg = {
		.msg_name = &nladdr,
		.msg_namelen = sizeof(nladdr),
		.msg_iov = &iov,
		.msg_iovlen = 1,
	};
	/* buffer for the incoming data */
	char buf[8192];

	/* set up the netlink address */
	memset(&nladdr, 0, sizeof(nladdr));
	nladdr.nl_family = AF_NETLINK;
	nladdr.nl_pid = 0;
	nladdr.nl_groups = 0;

	/* set up the I/O vector */
	iov.iov_base = buf;
	iov.iov_len = sizeof(buf);

	/* read the data into the buffer */
	status = TEMP_FAILURE_RETRY ( recvmsg(fd, &msg, MSG_CMSG_CLOEXEC) );

	/* if there was a problem, print a message and keep trying */
	if (status < 0) {
		bb_perror_msg("netlink read error");
		if (++process_netlink_nerrs >= ACPID_MAX_ERRS)
			bb_perror_msg_and_die("too many errors reading via "
				"netlink - aborting");
		return;
	}
	/* if an orderly shutdown has occurred, we're done */
	if (status == 0) {
		if(log_syslog)
			log_syslog = LOG_WARNING;
		bb_error_msg_and_die("netlink connection closed");
	}
	/* check to see if the address length has changed */
	if (msg.msg_namelen != sizeof(nladdr)) {
		xlog(LOG_WARNING, "netlink unexpected length: "
			"%d   expected: %zd", msg.msg_namelen, sizeof(nladdr));
		return;
	}

	/* for each message received */
	for (h = (struct nlmsghdr*)buf; (unsigned)status >= sizeof(*h); ) {
		int len = h->nlmsg_len;
		int l = len - sizeof(*h);

		if (l < 0  ||  len > status) {
			if (msg.msg_flags & MSG_TRUNC)
				xlog(LOG_WARNING, "netlink msg truncated (1)");
			 else
				xlog(LOG_WARNING, "malformed netlink msg, length %d", len);
			return;
		}

		/* format the message */
		format_netlink(h);

		status -= NLMSG_ALIGN(len);
		h = (struct nlmsghdr*)((char*)h + NLMSG_ALIGN(len));
	}
	if (msg.msg_flags & MSG_TRUNC) {
		xlog(LOG_WARNING, "netlink msg truncated (2)");
		return;
	}
	if (status)
		xlog(LOG_WARNING, "netlink remnant of size %d", status);
}

/* convert the netlink multicast group number into a bit map */
/* (e.g. 4 => 16, 5 => 32) */
static __u32 nl_mgrp(__u32 group)
{
	if (group > 31) {
		bb_error_msg("Unexpected group number %d", group);
		return 0;
	}
	return group ? (1 << (group - 1)) : 0;
}

/* open the netlink connection */
static void open_netlink(void)
{
	struct rtnl_handle rth;
	struct connection c;

	/* open the appropriate netlink socket for input */
	if (rtnl_open_byproto(&rth, nl_mgrp(acpi_ids_getgroup()), NETLINK_GENERIC) < 0) {
		bb_error_msg("cannot open generic netlink socket");
		return;
	}

	/* add a connection to the list */
	c.fd = rth.fd;
	c.process = process_netlink;
	c.pathname = NULL;
	c.kybd = 0;

	if (add_connection(&c) < 0) {
		rtnl_close(&rth);
		bb_error_msg("can't add connection for generic netlink socket");
	}
}

#define DIM(a)  (sizeof(a) / sizeof(a[0]))

struct evtab_entry {
	struct input_event event;
	const char *str;
};

/* Event Table: Events we are interested in and their strings.  Use
   evtest.c, acpi_genl, or kacpimon to find new events to add to this
   table. */

/*
 * The two numbers (e.g. "00000080 00000000") in each string is a format
 * that Xorg and maybe others expect.
 *
 * See hw/xfree86/os-support/linux/lnx_acpi.c in xserver and specifically
 * lnxACPIGetEventFromOs().
 */

static const struct evtab_entry evtab[] = {

	/*** COMMON EVENTS ***/

	{{{0,0}, EV_KEY, KEY_POWER, 1}, "button/power PBTN 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_SUSPEND, 1}, "button/suspend SUSP 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_SLEEP, 1}, "button/sleep SBTN 00000080 00000000"},
	{{{0,0}, EV_SW, SW_LID, 1}, "button/lid LID close"},
	{{{0,0}, EV_SW, SW_LID, 0}, "button/lid LID open"},
	{{{0,0}, EV_SW, SW_TABLET_MODE, 0}, "video/tabletmode TBLT 0000008A 00000000"},
	{{{0,0}, EV_SW, SW_TABLET_MODE, 1}, "video/tabletmode TBLT 0000008A 00000001"},


	/*** VIDEO ***/

	{{{0,0}, EV_KEY, KEY_ZOOM, 1}, "button/zoom ZOOM 00000080 00000000"},
	/* typical events file has "video.* 00000087" */
	{{{0,0}, EV_KEY, KEY_BRIGHTNESSDOWN, 1},
		"video/brightnessdown BRTDN 00000087 00000000"},
	/* typical events file has "video.* 00000086" */
	{{{0,0}, EV_KEY, KEY_BRIGHTNESSUP, 1},
		"video/brightnessup BRTUP 00000086 00000000"},
	/* additional events divined from the kernel's video.c */
	{{{0,0}, EV_KEY, KEY_VIDEO_NEXT, 1}, "video/next NEXT 00000083 00000000"},
	{{{0,0}, EV_KEY, KEY_VIDEO_PREV, 1}, "video/prev PREV 00000084 00000000"},
	{{{0,0}, EV_KEY, KEY_BRIGHTNESS_CYCLE, 1},
		"video/brightnesscycle BCYC 00000085 00000000"},
	{{{0,0}, EV_KEY, KEY_BRIGHTNESS_ZERO, 1},
		"video/brightnesszero BZRO 00000088 00000000"},
	{{{0,0}, EV_KEY, KEY_DISPLAY_OFF, 1},
			"video/displayoff DOFF 00000089 00000000"},
	/* procfs on Thinkpad 600X reports "video VID0 00000080 00000000" */
	/* typical events file has "video.* 00000080" */
	{{{0,0}, EV_KEY, KEY_SWITCHVIDEOMODE, 1},
		"video/switchmode VMOD 00000080 00000000"},


	/*** AUDIO ***/

	{{{0,0}, EV_KEY, KEY_VOLUMEDOWN, 1}, "button/volumedown VOLDN 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_VOLUMEDOWN, 2}, "button/volumedown VOLDN 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_VOLUMEUP, 1}, "button/volumeup VOLUP 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_VOLUMEUP, 2}, "button/volumeup VOLUP 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_MUTE, 1}, "button/mute MUTE 00000080 00000000"},
/* Kernel 3.1 or later required for KEY_MICMUTE */
#ifdef KEY_MICMUTE
	{{{0,0}, EV_KEY, KEY_MICMUTE, 1}, "button/micmute MICMUTE 00000080 00000000"},
#endif
	/* cd play/pause buttons */
	{{{0,0}, EV_KEY, KEY_NEXTSONG, 1}, "cd/next CDNEXT 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_PREVIOUSSONG, 1}, "cd/prev CDPREV 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_PLAYPAUSE, 1}, "cd/play CDPLAY 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_STOPCD, 1}, "cd/stop CDSTOP 00000080 00000000"},


	/*** JACKS ***/

/* This test probably belongs in configure.ac. */
#ifdef SW_HEADPHONE_INSERT
 #ifndef SW_LINEIN_INSERT
  #define SW_LINEIN_INSERT 0x0d
 #endif
	{{{0,0}, EV_SW, SW_HEADPHONE_INSERT, 0}, "jack/headphone HEADPHONE unplug"},
	{{{0,0}, EV_SW, SW_HEADPHONE_INSERT, 1}, "jack/headphone HEADPHONE plug"},
	{{{0,0}, EV_SW, SW_MICROPHONE_INSERT, 0}, "jack/microphone MICROPHONE unplug"},
	{{{0,0}, EV_SW, SW_MICROPHONE_INSERT, 1}, "jack/microphone MICROPHONE plug"},
	{{{0,0}, EV_SW, SW_LINEOUT_INSERT, 0}, "jack/lineout LINEOUT unplug"},
	{{{0,0}, EV_SW, SW_LINEOUT_INSERT, 1}, "jack/lineout LINEOUT plug"},
#ifdef SW_VIDEOOUT_INSERT
	{{{0,0}, EV_SW, SW_VIDEOOUT_INSERT, 0}, "jack/videoout VIDEOOUT unplug"},
	{{{0,0}, EV_SW, SW_VIDEOOUT_INSERT, 1}, "jack/videoout VIDEOOUT plug"},
#endif
	{{{0,0}, EV_SW, SW_LINEIN_INSERT, 0}, "jack/linein LINEIN unplug"},
	{{{0,0}, EV_SW, SW_LINEIN_INSERT, 1}, "jack/linein LINEIN plug"},
#else
 #warning SW_HEADPHONE_INSERT not found in input_layer.h. Support for plug/unplug events will be disabled. Please upgrade your kernel headers to Linux-3.2 or newer.
#endif


	/*** MISCELLANEOUS ***/

	/* blue access IBM button on Thinkpad T42p*/
	{{{0,0}, EV_KEY, KEY_PROG1, 1}, "button/prog1 PROG1 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_VENDOR, 1}, "button/vendor VNDR 00000080 00000000"},
	/* Fn-F2 produces KEY_BATTERY on Thinkpad T42p */
	{{{0,0}, EV_KEY, KEY_BATTERY, 1}, "button/battery BAT 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_SCREENLOCK, 1}, "button/screenlock SCRNLCK 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_COFFEE, 1}, "button/coffee CFEE 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_WLAN, 1}, "button/wlan WLAN 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_FN_F1, 1}, "button/fnf1 FNF1 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_FN_F2, 1}, "button/fnf2 FNF2 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_FN_F6, 1}, "button/fnf6 FNF6 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_FN_F9, 1}, "button/fnf9 FNF9 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_FN_F10, 1}, "button/fnf10 FF10 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_FN_F11, 1}, "button/fnf11 FF11 00000080 00000000"},
	/* F20 is sometimes used for micmute */
	{{{0,0}, EV_KEY, KEY_F20, 1}, "button/f20 F20 00000080 00000000"},
	/* Fn-F9 produces KEY_F24 on Thinkpad T42p */
	{{{0,0}, EV_KEY, KEY_F24, 1}, "button/f24 F24 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_KBDILLUMTOGGLE, 1},
		"button/kbdillumtoggle KBILLUM 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_COPY, 1}, "button/copy COPY 00000080 00000000"},
	{{{0,0}, EV_KEY, KEY_RESTART, 1}, "button/restart RSTR 00000080 00000000"},

};

/* special support for the MUTE key, as the key toggles we want to
 * consider repeated keys but don't report them all the time. We just
 * ensure that the number of key presses (MOD 2) is correct.
 */
static const char *mute_string(struct input_event event)
{
	if (event.type == EV_KEY && event.code == KEY_MUTE) {
		if (event.value == 1) {
			repeat_count = 1;
			return "button/mute MUTE (key pressed)";
		} else if (event.value == 2) {
			repeat_count++;
			return NULL;
		} else if (event.value == 0) {
			if (repeat_count % 2 == 0) {
				return "button/mute MUTE (key released)";
			}
		}
	}
	return NULL;
}

/* Given an input event, returns the string corresponding to that event.
   If there is no corresponding string, NULL is returned.  */
static const char *event_string(struct input_event event)
{
	unsigned i;

	/* for each entry in the event table */
	/* ??? Is there a faster way?  This is triggered every time the user
	 *     presses a key.  Maybe a simple hash algorithm?  Or a simple check
	 *     for very common keys (alphanumeric) and bail before this?  */
	for (i = 0; i < DIM(evtab); ++i)
	{
		/* if this is a matching event, return its string */
		if (event.type == evtab[i].event.type  &&
			event.code == evtab[i].event.code  &&
			event.value == evtab[i].event.value) {
			return evtab[i].str;
		}
	}

	return NULL;
}

/* returns non-zero if the event type/code is one we need */
static int need_event(int type, int code)
{
	unsigned i;

	/* for each entry in the event table */
	for (i = 0; i < DIM(evtab); ++i) {
		/* if we found a matching event */
		if (type == evtab[i].event.type  &&
			code == evtab[i].event.code) {
			return 1;
		}
	}

	return 0;
}

/* called when an input layer event is received */
static void process_input(int fd)
{
	struct input_event event;
	ssize_t nbytes;
	const char *str;
	struct connection *c;
	char str2[100];

	nbytes = TEMP_FAILURE_RETRY ( read(fd, &event, sizeof(event)) );

	if (nbytes == 0) {
		if(log_syslog)
			log_syslog = LOG_WARNING;
		bb_error_msg_and_die("input layer connection closed");
	}

	if (nbytes < 0) {
		if (errno == ENODEV) {
			xlog(LOG_WARNING, "input device has been disconnected, fd %d", fd);
			delete_connection(fd);
			return;
		}
		bb_perror_msg("input layer read error");
		if (++process_input_nerrs >= ACPID_MAX_ERRS) {
			bb_error_msg_and_die("too many errors reading "
				"input layer - aborting");
		}
		return;
	}

	/* ??? Is it possible for a partial message to come across? */
	/*   If so, we've got more code to write... */

	if (nbytes != sizeof(event)) {
		xlog(LOG_WARNING, "input layer unexpected length: "
			"%zd   expected: %zd", nbytes, sizeof(event));
		return;
	}

	c = find_connection(fd);

	/* if we're getting scancodes, we probably have a keyboard */
	if (event.type == EV_MSC  &&  event.code == MSC_SCAN) {
		if (c)
			c->kybd = 1;  /* appears to be a keyboard device */
	}

	/* convert the event into a string */
	if (tpmutefix) {
		str = mute_string(event);
		if (str ==  NULL)
			str = event_string(event);
	} else
		str = event_string(event);
	/* if this is not an event we care about, bail */
	if (str == NULL)
		return;

	/* If we suspect this is a keyboard, and we have enough space, tack a
	 * "K" on to the end of the event string. */
	if (c  &&  c->kybd  &&  strnlen(str, sizeof(str2)) <= sizeof(str2) - 3) {
		strcpy(str2, str);
		strcat(str2, " K");
		str = str2;
	}

	/* if we're locked, don't process the event */
	if (locked()) {
		if (logevents) {
			xlog(LOG_INFO, "lockfile present, not processing "
				"input layer event \"%s\"", str);
		}
		return;
	}

	if (logevents)
		xlog(LOG_INFO, "received input layer event \"%s\"", str);

	/* send the event off to the handler */
	acpid_handle_event(str);

	if (logevents)
		xlog(LOG_INFO, "completed input layer event \"%s\"", str);
}

#define BITS_PER_LONG (sizeof(long) * 8)
#define NBITS(x) ((((x)-1)/BITS_PER_LONG)+1)
#define OFF(x)  ((x)%BITS_PER_LONG)
#define LONG(x) ((x)/BITS_PER_LONG)
#define test_bit(bit, array)    ((array[LONG(bit)] >> OFF(bit)) & 1)

/* returns non-zero if the file descriptor supports one of the events */
/* supported by event_string().  */
static int has_event(int fd)
{
	int type, code;
	unsigned long bit[EV_MAX][NBITS(KEY_MAX)];

	memset(bit, 0, sizeof(bit));
	/* get the event type bitmap */
	ioctl(fd, EVIOCGBIT(0, sizeof(bit[0])), bit[0]);

	/* for each event type */
	for (type = 0; type < EV_MAX; type++) {
		/* if this event type is supported */
		if (test_bit(type, bit[0])) {
			/* skip sync */
			if (type == EV_SYN) continue;
			/* get the event code mask */
			ioctl(fd, EVIOCGBIT(type, sizeof(bit[type])), bit[type]);
			/* for each event code */
			for (code = 0; code < KEY_MAX; code++) {
				/* if this event code is supported */
				if (test_bit(code, bit[type])) {
					/* if we need this event */
					if (need_event(type, code) != 0)
						return 1;
				}
			}
		}
	}
	return 0;
}

/* open a single input layer file for input  */
static int open_inputfile(const char *filename)
{
	int fd;
	struct connection c;

	/* O_CLOEXEC: Make sure scripts we exec() (in event.c) don't get our file descriptors. */
	fd = open(filename, O_RDONLY | O_NONBLOCK | O_CLOEXEC);

	if (fd >= 0) {
		char evname[256];

		/* if this file doesn't have events we need, indicate failure */
		if (!has_event(fd)) {
			close(fd);
			return -1;
		}

		/* get this event file's name for debugging */
		strcpy(evname, "Unknown");
		ioctl(fd, EVIOCGNAME(sizeof(evname)), evname);

		/* add a connection to the list */
		c.fd = fd;
		c.process = process_input;
		/* delete_connection() will free */
		c.pathname = strdup(filename);
		/* assume not a keyboard until we see a scancode */
		c.kybd = 0;

		if (add_connection(&c) < 0) {
			close(fd);
			bb_error_msg("can't add connection for input layer %s (%s)",
				filename, evname);
			return -1;
		}

		return 0;  /* success */
	}

	/* open unsuccessful */
	return -1;
}

/* open each of the appropriate /dev/input/event* files for input  */
static void open_input(void)
{
	char *filename = NULL;
	glob_t globbuf;
	unsigned i;
	int success = 0;

	/* get all the matching event filenames */
	glob(ACPID_INPUTLAYERFILES, 0, NULL, &globbuf);

	/* for each event file */
	for (i = 0; i < globbuf.gl_pathc; ++i) {
		filename = globbuf.gl_pathv[i];

		/* open this input layer device file */
		if (open_inputfile(filename) == 0)
			success = 1;
	}

	if (!success)
		bb_error_msg("cannot open input layer");

	globfree(&globbuf);
}

static void close_fds(void)
{
    struct dirent *dent;
    DIR *dirp;

    if ((dirp = opendir("/proc/self/fd")) != NULL) {
	while ((dent = readdir(dirp)) != NULL) {
	    int fd = bb_xparse_i_number(10 | BB_PN_SILENT_RE | BB_PN_SET_LOWER, dent->d_name, 3);
	    if (fd > 0 && fd != dirfd(dirp))
		close(fd);
	}
	closedir(dirp);
    }
}

static void std2null(void)
{
	/* open /dev/null */
	int nullfd = bb_xopen(bb_dev_null, O_RDWR | BB_OPEN_ERR);

	/* set up stdin, stdout, stderr to /dev/null */

	/* don't redirect stdin if we're being sent a socket by systemd */
	if (!is_socket(STDIN_FILENO)  &&
			dup2(nullfd, STDIN_FILENO) != STDIN_FILENO)
		bb_perror_msg_and_die("dup2() stdin");

	if (!log_stderr && dup2(nullfd, STDOUT_FILENO) != STDOUT_FILENO)
		bb_perror_msg_and_die("dup2() stdout");
	if (!log_stderr && dup2(nullfd, STDERR_FILENO) != STDERR_FILENO)
		bb_perror_msg_and_die("dup2() stderr");

	close(nullfd);
}

static void clean_exit_with_status(int status)
{
	acpid_cleanup_rules(1);
	/* delete all connections, closing the fds */
	while (nconnections) {
		/* delete the connection at the end of the list */
		delete_connection(connection_list[nconnections-1].fd);
	}
#if 0
	free(connection_list);
	connection_list = NULL;
#endif
	xlog(LOG_NOTICE, "exiting");
	if(pidfile != NULL)
		unlink(pidfile);
	exit(status);
}

static void clean_exit(int sig __attribute__((unused)))
{
	clean_exit_with_status(EXIT_SUCCESS);
}

static char *acpid_asprintf(const char *format, ...)
{
	va_list p;
	int r;
	char *string_ptr;

	va_start(p, format);
	r = vasprintf(&string_ptr, format, p);
	va_end(p);

	if (r < 0) {
		bb_perror_msg("asprintf");
		string_ptr = NULL;
	}
	return string_ptr;
}

/* read in all the configuration files */
static int acpid_read_conf(void)
{
	DIR *dir;
	struct dirent *dirent;
	int nrules = 0;
	regex_t preg;
	int rc = 0;

	lock_rules();

	dir = opendir(confdir);
	if (!dir) {
		bb_perror_msg("opendir(%s)", confdir);
		unlock_rules();
		return -1;
	}

	/* Compile the regular expression.  This is based on run-parts(8). */
	rc = regcomp(&preg, "^[a-zA-Z0-9_-]+$", RULE_REGEX_FLAGS);
	if (rc) {
		closedir(dir);
		bb_error_msg("regcomp(): %d", rc);
		unlock_rules();
		return -1;
	}

	/* scan all the files */
	while ((dirent = readdir(dir))) {
		struct rule *r;
		struct stat stat_buf;
		char *file;  /* rename: filename */
		int fd_rule;

		/* skip "." and ".." */
		if (strcmp(dirent->d_name, ".") == 0)
			continue;
		if (strcmp(dirent->d_name, "..") == 0)
			continue;

		if ((file = acpid_asprintf("%s/%s", confdir, dirent->d_name)) == NULL) {
			regfree(&preg);
			closedir(dir);
			unlock_rules();
			return -1;
		}

		/* skip any files that don't match the run-parts convention */
		if (regexec(&preg, dirent->d_name, 0, NULL, 0) != 0) {
			xlog(LOG_INFO, "skipping conf file %s", file);
			free(file);
			continue;
		}

		/* ??? Check for DT_UNKNOWN and do this, then "else if" on not DT_REG
			and do the same as the !S_ISREG() branch below. */
		if (dirent->d_type != DT_REG) { /* may be DT_UNKNOWN ...*/
			/* allow only regular files and symlinks to files */
			if (fstatat(dirfd(dir), dirent->d_name, &stat_buf, 0) != 0) {
				bb_perror_msg("fstatat(%s)", file);
				free(file);
				continue; /* keep trying the rest of the files */
			}
			if (!S_ISREG(stat_buf.st_mode)) {
				xlog(LOG_INFO, "skipping non-file %s", file);
				free(file);
				continue; /* skip non-regular files */
			}
		}

		/* open the rule file (might want to move this into parse_file()?) */
		if ((fd_rule = openat(dirfd(dir), dirent->d_name,
			      O_RDONLY|O_CLOEXEC|O_NONBLOCK)) == -1) {
			/* something went _really_ wrong.. Not Gonna Happen(tm) */
			bb_perror_msg("openat(%s)", file);
			free(file);
			/* ??? Too extreme?  Why not just continue? */
			closedir(dir);
			regfree(&preg);
			unlock_rules();
			return -1;
		}

		r = parse_file(fd_rule, file);
		close(fd_rule);
		if (r) {
			/* if this is a drop rule */
			if (!strcmp(r->action.cmd, dropaction))
				enlist_rule(&drop_list, r);
			else
				enlist_rule(&cmd_list, r);
			nrules++;
		}
		free(file);
	}

	regfree(&preg);
	closedir(dir);
	unlock_rules();

	print_nrules(nrules, NULL, "loaded");

	return 0;
}

static void reload_conf(int sig __attribute__((unused)))
{
	xlog(LOG_NOTICE, "reloading configuration");
	acpid_cleanup_rules(0);
	acpid_read_conf();
}


/* Parse command line arguments */

static int acpid_add_client(int clifd, const char *origin)
{
	struct rule *r;
	int nrules = 0;

	xlog(LOG_NOTICE, "client connected from %s", origin);

	r = parse_client(clifd);
	if (r) {
		r->origin = strdup(origin);
		enlist_rule(&client_list, r);
		nrules++;
	}

	print_nrules(nrules, "client", "loaded");

	return 0;
}

/* accept a new client connection */
static void process_sock(int fd)
{
	int cli_fd;
	struct ucred creds;
	char *buf;

	/* accept and add to our lists */
	cli_fd = ud_accept(fd, &creds);
	if (cli_fd < 0) {
		bb_perror_msg("can't accept client");
		accept_errors++;
		if (accept_errors >= 5) {
			bb_error_msg("giving up");
			clean_exit_with_status(EXIT_FAILURE);
		}
		return;
	}
	accept_errors = 0;

	/* don't allow too many non-root clients  */
	if (creds.uid != 0 && non_root_clients >= clientmax) {
		close(cli_fd);
		bb_error_msg("too many non-root clients");
		return;
	}
	if (creds.uid != 0)
		non_root_clients++;

	if((buf = acpid_asprintf("%d[%d:%d]", creds.pid, creds.uid, creds.gid)) == NULL) {
		close(cli_fd);
		return;
	}
	acpid_add_client(cli_fd, buf);
	free(buf);
}

/* set up the socket for client connections */
static void open_sock(void)
{
	int fd;
	struct connection c;

	/* if this is a socket passed in via stdin by systemd */
	if (is_socket(STDIN_FILENO)) {
		fd = STDIN_FILENO;
		/* ??? Move CLOEXEC and NONBLOCK settings below up to here? */
	} else {
		/* create our own socket */
		fd = ud_create_socket(socketfile);
		if (fd < 0)
			bb_perror_msg_and_die("can't open socket %s", socketfile);

		/* if we need to change the socket's group, do so */
		if (socketgroup) {
			struct group *gr;
			struct stat buf;

			gr = getgrnam(socketgroup);
			if (!gr)
				bb_error_msg_and_die("group %s does not exist", socketgroup);
			if (fstat(fd, &buf) < 0)
				bb_perror_msg_and_die("can't stat %s", socketfile);
			/* ??? I've tried using fchown(), however it doesn't work here.
			 *     It also doesn't work before bind().  The GNU docs seem to
			 *     indicate it isn't supposed to work with sockets. */
			/* if (fchown(fd, buf.st_uid, gr->gr_gid) < 0)  */
			if (chown(socketfile, buf.st_uid, gr->gr_gid) < 0)
				bb_perror_msg_and_die("can't chown %s", socketfile);
		}
	}

	/* Don't leak fds when execing.
	 * ud_create_socket() already does this, but there is no guarantee that
	 * a socket sent in via STDIN will have this set. */
	if (fcntl(fd, F_SETFD, FD_CLOEXEC) < 0) {
		close(fd);
		bb_perror_msg("fcntl() on socket %s for FD_CLOEXEC", socketfile);
		return;
	}

	/* Avoid a potential hang.
	 * ud_create_socket() already does this, but there is no guarantee that
	 * a socket sent in via STDIN will have this set. */
	if (fcntl(fd, F_SETFL, O_NONBLOCK) < 0) {
		close(fd);
		bb_perror_msg("fcntl() on socket %s for O_NONBLOCK", socketfile);
		return;
	}

	/* add a connection to the list */
	c.fd = fd;
	c.process = process_sock;
	c.pathname = NULL;
	c.kybd = 0;

	if (add_connection(&c) < 0) {
		close(fd);
		bb_error_msg("can't add connection for socket %s", socketfile);
		return;
	}
}

static void handle_cmdline(void)
{
	int i;
	long mode;

#if ENABLE_FEATURE_LONG_OPTIONS
	static const struct option opts[] = {
		{"confdir", 1, 0, 'c'},
		{"clientmax", 1, 0, 'C'},
		{"debug", 0, 0, 'd'},
		{"eventfile", 1, 0, 'e'},
		{"foreground", 0, 0, 'f'},
		{"logevents", 0, 0, 'l'},
		{"socketgroup", 1, 0, 'g'},
		{"socketmode", 1, 0, 'm'},
		{"socketfile", 1, 0, 's'},
		{"nosocket", 0, 0, 'S'},
		{"pidfile", 1, 0, 'p'},
		{"lockfile", 1, 0, 'L'},
		{"netlink", 0, 0, 'n'},
		{"dropaction", 1, 0, 'r'},
		{"tpmutefix", 0, 0, 't'},
		{"version", 0, 0, 'v'},
		{"help", 0, 0, 'h'},
		{NULL, 0, 0, 0},
	};
	bb_applet_long_options = opts;
#endif
	eventfile = ACPID_EVENTFILE;
	lockfile = ACPID_LOCKFILE;
	socketmode = ACPID_SOCKETMODE;
	socketfile = ACPID_SOCKETFILE;
	clientmax = ACPID_CLIENTMAX;
	dropaction = DROP_ACTION;
	confdir = ACPID_CONFDIR;
	i = bb_xgetopt_ulflags(opt_str,
				&confdir, &clientmax, &eventfile, &socketgroup,
				&mode, &socketfile, &pidfile, &lockfile, &dropaction,
				&debug_level);
	foreground = i & FOREGROUND_OPT;
	logevents = i & LOGEVENTS_OPT;
	if((i & SOCKETMODE_OPT))
		socketmode = mode;
	nosocket = i & NOSOCKET_OPT;
	netlink = i & NETLINK_OPT;
	tpmutefix = i & TPMUTEFIX_OPT;
	if((i & VERSION_OPT)) {
		printf(PACKAGE "-" VERSION "\n");
		exit(EXIT_SUCCESS);
	}
	if((i & HELP_OPT))
		bb_show_usage();
}

/* called when an inotify event is received */
static void process_inotify(int fd)
{
	int bytes;
	int processed_bytes = 0;
	char eventbuf[sizeof(struct inotify_event) + NAME_MAX + 1];
	const int dnsize = NAME_MAX + 1;
	char devname[dnsize];

	bytes = read(fd, &eventbuf, sizeof(eventbuf));

	/* eof is not expected */
	if (bytes == 0) {
		xlog(LOG_WARNING, "inotify fd eof encountered");
		return;
	}
	else if (bytes < 0) {
		/* EINVAL means buffer wasn't big enough.  See inotify(7). */
		bb_perror_msg("inotify read error");
		bb_error_msg("disconnecting from inotify");
		delete_connection(fd);
		return;
	}

	/* while there are still messages in eventbuf */
	while (processed_bytes < bytes) {
		struct inotify_event* curevent = (struct inotify_event *)
			&eventbuf[processed_bytes];

		/* if a name is included */
		if (curevent->len > 0) {
			/* devname = ACPID_INPUTLAYERDIR + "/" + pevent -> name */
			strcpy(devname, ACPID_INPUTLAYERDIR);
			strcat(devname, "/");
			strncat(devname, curevent->name, dnsize - strlen(devname) - 1);
		}

		/* if this is a create */
		if (curevent->mask & IN_CREATE) {
			open_inputfile(devname);
		}

		processed_bytes += sizeof(struct inotify_event) + curevent->len;
	}
}

/* Set up an inotify watch on /dev/input. */
static void open_inotify(void)
{
	int fd = -1;
	int wd = -1;
	struct connection c;

	/* set up inotify */
	fd = inotify_init1(IN_CLOEXEC);

	if (fd < 0) {
		bb_perror_msg("inotify_init() failed");
		return;
	}

	/* watch for files being created or deleted in /dev/input */
	wd = inotify_add_watch(fd, ACPID_INPUTLAYERDIR, IN_CREATE | IN_DELETE);

	if (wd < 0) {
		if (errno == ENOENT) {
			/* Common with headless devices. */
			xlog(LOG_WARNING, "inotify_add_watch(): input layer not found");
		} else {
			bb_perror_msg("inotify_add_watch() failed");
		}

		close(fd);
		return;
	}

	/* add a connection to the list */
	c.fd = fd;
	c.process = process_inotify;
	c.pathname = NULL;
	c.kybd = 0;

	if (add_connection(&c) < 0) {
		close(fd);
		bb_error_msg("can't add connection for inotify");
		return;
	}
}

int acpid_main(void)
{
	highestfd = -2;

	/* handle the commandline  */
	handle_cmdline();

	/* close any extra file descriptors */
	close_fds();

	if (!foreground) {
		/* open the syslog */
		bb_openlog(LOG_CONS|LOG_NDELAY, LOG_DAEMON);
		log_syslog = LOG_ERR;
		/* if we're running in the background, and we're not being started */
		/* by systemd */
		if(!is_socket(STDIN_FILENO))
			xdaemon(XDAEMON_NO_CLOSE);
		log_stderr = 0;
	}
	/* create pidfile only */
	pidfile = xdaemon(XDAEMON_NOFORK | XDAEMON_PF_ARG | XDAEMON_PIDFILE, pidfile);

	if (!netlink) {
		/* open the acpi event file in the proc fs */
		/* if the open fails, try netlink */
		if (open_proc())
			netlink = 1;
	}

	if (netlink) {
		/* open the input layer */
		open_input();

		/* watch for new input layer devices */
		open_inotify();

		/* open netlink */
		open_netlink();
	}

	/* open our socket */
	if (!nosocket)
		open_sock();

	/* redirect standard files to /dev/null */
	std2null();

	xlog(LOG_INFO, "starting up with %s",
		netlink ? "netlink and the input layer" : "proc fs");

	/* trap key signals */
	signal(SIGHUP, reload_conf);
	bb_signal((1<<SIGTERM)|(1<<SIGQUIT)|(1<<SIGINT), clean_exit);
	signal(SIGPIPE, SIG_IGN);

	/* read in our configuration */
	if (acpid_read_conf())
		return EXIT_FAILURE;

	xlog(LOG_INFO, "waiting for events: event logging is %s", logevents ? "on" : "off");

	/* main loop */
	while (1) {
		fd_set readfds;
		int nready;
		int i;
		struct connection *p;

		/* it's going to get clobbered, so use a copy */
		readfds = *get_fdset();

		/* wait on data */
		nready = TEMP_FAILURE_RETRY(select(highestfd + 1, &readfds, NULL, NULL, NULL));

		if (nready < 0) {
			bb_perror_msg("select()");
			continue;
		}

		/* house keeping */
		acpid_close_dead_clients();

		/* for each connection */
		for (i = 0; i <= nconnections; ++i) {
			int fd;

			p = get_connection(i);

			/* if this connection is invalid, bail */
			if (!p)
				break;

			/* get the file descriptor */
			fd = p->fd;

			/* if this file descriptor has data waiting */
			if (FD_ISSET(fd, &readfds)) {
				/* delegate to this connection's process function */
				p->process(fd);
			}
		}
	}

	clean_exit_with_status(EXIT_SUCCESS);

	return 0;
}
#endif
