/*
 * fstrim.c -- discard the part (or whole) of mounted filesystem.
 *
 * Copyright (C) 2010 Red Hat, Inc. All rights reserved.
 * Written by Lukas Czerner <lczerner@redhat.com>
 *            Karel Zak <kzak@redhat.com>
 *
 * Busybox port by Vladimir Oleynik (c) 2016 <dzo@simtreas.ru>
 *
 * This program uses FITRIM ioctl to discard parts or the whole filesystem
 * online (mounted). You can specify range (start and length) to be
 * discarded, or simply discard whole filesystem.
 */
#ifdef APPLET_USAGE
#define fstrim_trivial_usage \
	"[-o offset] [-l length] [-m minimum] [-v] mountpoint"
#define fstrim_full_usage \
	"discard the part (or whole) of mounted filesystem.\n\n" \
	"Options:\n" \
	USE_FEATURE_LONG_OPTIONS( \
	"o, --offset offset\tlength of bytes to discard from the offset\n" \
	"l, --length length\tnumber of bytes after starting point to search\n" \
				"\t\t\tfor free blocks to discard\n" \
	"m, --minimum minimum\tminimum extent length to discard\n" \
	"v, --verbose\t\tprint number of discarded bytes" ) \
	SKIP_FEATURE_LONG_OPTIONS( \
	"\t-o offset\tlength of bytes to discard from the offset\n" \
	"\t-l length\tnumber of bytes after starting point to search for free blocks to discard\n" \
	"\t-m minimum\tminimum extent length to discard\n" \
	"\t-v\t\tprint number of discarded bytes" )
#else

#define UTIL_LINUX_VERSION "util-linux-2.23"

#include "busybox.h"
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include "bb_xopen.h"

#include <sys/ioctl.h>
#include <linux/fs.h>

#ifndef FITRIM
struct fstrim_range {
	uint64_t start;
	uint64_t len;
	uint64_t minlen;
};
#define FITRIM          _IOWR('X', 121, struct fstrim_range)
#endif


static int do_scale_by_power (uintmax_t *x, int base, int power)
{
	while (power--) {
		if (UINTMAX_MAX / base < *x)
			return -2;
		*x *= base;
	}
	return 0;
}

/*
 * strtosize() - convert string to size (uintmax_t).
 *
 * Supported suffixes:
 *
 * XiB or X for 2^N
 *     where X = {K,M,G,T,P,E,Z,Y}
 *        or X = {k,m,g,t,p,e}  (undocumented for backward compatibility only)
 * for example:
 *              10KiB   = 10240
 *              10K     = 10240
 *
 * XB for 10^N
 *     where X = {K,M,G,T,P,E,Z,Y}
 * for example:
 *              10KB    = 10000
 *
 * The optinal 'power' variable returns number associated with used suffix
 * {K,M,G,T,P,E,Z,Y}  = {1,2,3,4,5,6,7,8}.
 *
 * Note that the function does not accept numbers with '-' (negative sign)
 * prefix.
 */
static int parse_size(const char *str, uintmax_t *res)
{
	char *p;
	uintmax_t x;
	int base = 1024, rc = 0, pwr = 0;

	static const char *suf  = "KMGTPEYZ";
	static const char *suf2 = "kmgtpeyz";
	const char *sp;

	*res = 0;

	errno = 0;
	if (!*str)
		goto err;

	/* Only positive numbers are acceptable
	 *
	 * Note that this check is not perfect, it would be better to
	 * use lconv->negative_sign. But coreutils use the same solution,
	 * so it's probably good enough...
	 */

	p = bb_skip_whitespace(str);
	if (*p == '-')
		goto err;
	p = NULL;

	x = strtoull(str, &p, 0);

	if (p == str ||
	    (errno != 0 && (x == UINTMAX_MAX || x == 0)))
		goto err;

	if (!p || !*p)
		goto done;                      /* without suffix */

	/*
	 * Check size suffixes
	 */
	if (*(p + 1) == 'i' && *(p + 2) == 'B' && !*(p + 3))
		base = 1024;                    /* XiB, 2^N */
	else if (*(p + 1) == 'B' && !*(p + 2))
		base = 1000;                    /* XB, 10^N */
	else if (*(p + 1))
		goto err;                       /* unexpected suffix */

	sp = strchr(suf, *p);
	if (sp)
		pwr = (sp - suf) + 1;
	else {
		sp = strchr(suf2, *p);
		if (sp)
			pwr = (sp - suf2) + 1;
		else
			goto err;
	}

	rc = do_scale_by_power(&x, base, pwr);
done:
	*res = x;
	return rc;
err:
	return -1;
}

static uintmax_t strtosize_or_err(const char *str, const char *errmesg, uintmax_t def_val)
{
	uintmax_t num;

	if(str == NULL)
		return def_val;
	if (parse_size(str, &num) == 0)
		return num;

	if (errno)
		bb_perror_msg_and_die("%s: '%s'", errmesg, str);

	bb_error_msg_and_die("%s: '%s'", errmesg, str);
}

int fstrim_main(void)
{
	char *path;
	int verbose = 0;
	struct fstrim_range range;
	char *sstart = NULL, *slen = NULL, *minlen = NULL;

#ifdef CONFIG_FEATURE_LONG_OPTIONS
	static const struct option longopts[] = {
	    { "version",   0, 0, 'V' },
	    { "offset",    1, 0, 'o' },
	    { "length",    1, 0, 'l' },
	    { "minimum",   1, 0, 'm' },
	    { "verbose",   0, 0, 'v' },
	    { NULL,        0, 0, 0 }
	};
	bb_applet_long_options = longopts;
#endif
	if((bb_xgetopt_ulflags("Vvo:l:m:\0=1:vv", &sstart, &slen, &minlen, &verbose) & 1)) {
		printf(UTIL_LINUX_VERSION);
		return EXIT_SUCCESS;
	}
	range.start = strtosize_or_err(sstart, "failed to parse offset", 0);
	range.len = strtosize_or_err(slen, "failed to parse length", ULLONG_MAX);
	range.minlen = strtosize_or_err(minlen, "failed to parse minimum extent length", 0);

	path = *Gargv;

	if(is_directory(path, 1, NULL) <= 0)
		bb_perror_msg_and_die("%s", path);

	xioctl(bb_xopen(path, O_RDONLY | BB_OPEN_ERR), FITRIM, &range,
					"\1%s: FITRIM ioctl failed", path);

	if (verbose) {
		char buf[SIZE_TO_HUMAN_STRING_BUFSIZ];
		size_to_human_string(SIZE_SUFFIX_3LETTER | SIZE_SUFFIX_SPACE,
						 (uint64_t) range.len, buf);
		/* TRANSLATORS: The standard value here is a very large number. */
		printf("%s: %s (%" P_BB64T "u bytes) trimmed\n",
						path, buf,
						(ubb64_t) range.len);
	}
	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
