/*
 * mdev - Mini udev for busybox
 *
 * Copyright 2005 Rob Landley <rob@landley.net>
 * Copyright 2005 Frank Sorenson <frank@tuxrocks.com>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */
#ifdef APPLET_USAGE
#define mdev_trivial_usage \
	"[-s]"
#define mdev_full_usage \
	"\t-s\tScan /sys and populate /dev during system boot\n\n" \
	"Called with no options (via hotplug) it uses environment variables\n" \
	"to determine which device to add/remove."
#else

#include "busybox.h"
#include <ctype.h>
#include <dirent.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/sysmacros.h>
#include <unistd.h>
#include "xregex.h"
#include "bb_xopen.h"

#define DEV_PATH        "/dev"
#define MDEV_CONF       "/etc/mdev.conf"

#include <busybox.h>

#ifndef strndupa
/* Return an alloca'd copy of at most N bytes of string.  */
# define strndupa(s, n)                                                       \
  (__extension__                                                              \
({                                                                        \
  __const char *__old = (s);                                              \
  size_t __len = strnlen (__old, (n));                                    \
  char *__new = (char *) __builtin_alloca (__len + 1);                    \
  __new[__len] = '\0';                                                    \
  (char *) memcpy (__new, __old, __len);                                  \
  }))
#endif

CALC_BSS struct BB_COMMON_BSS
{
	int root_major, root_minor;
	char temp_md[PATH_MAX];
	char temp[PATH_MAX];
} bb_common_bss;
#define root_major bb_common_bss.root_major
#define root_minor bb_common_bss.root_minor
#define temp_md bb_common_bss.temp_md
#define temp bb_common_bss.temp


/* mknod in /dev based on a path like "/sys/block/hda/hda1" */
static void make_device(char *path)
{
	const char *device_name;
	char *s;
	int device_major, device_minor, type, len, fd;
	int mode = 0660;
	uid_t uid = 0;
	gid_t gid = 0;

	/* Try to read major/minor string */

	snprintf(temp_md, PATH_MAX, "%s/dev", path);
	fd = bb_xopen(temp_md, O_RDONLY | BB_OPEN_ERR);
	len = read(fd, temp_md, PATH_MAX-1);
	close(fd);
	if (len < 1)
		return;

	/* Determine device name, type, major and minor */

	device_name = bb_basename(path);
	type = strncmp(path+5, "block/", 6) ? S_IFCHR : S_IFBLK;
	if (sscanf(temp_md, "%d:%d", &device_major, &device_minor) != 2)
		return;

	/* If we have a config file, look up permissions for this device */

	if (ENABLE_FEATURE_MDEV_CONF) {
		char *conf, *pos, *end;

		/* mmap the config file */
		if (-1 != (fd=open(MDEV_CONF,O_RDONLY))) {
			len = lseek(fd, 0, SEEK_END);
			conf = mmap(NULL, len, PROT_READ, MAP_PRIVATE, fd, 0);
			if (conf) {
				int line = 0;

				/* Loop through lines in mmaped file*/
				for (pos = conf; pos-conf < len; ) {
					int field;
					char *end2;

					line++;
					/* find end of this line */
					for(end=pos; end-conf<len && *end!='\n'; end++)
						;

					/* Three fields: regex, uid:gid, mode */
					for (field=3; field; field--) {
						/* Skip whitespace */
						while (pos<end && isspace(*pos))
							pos++;
						if (pos==end || *pos=='#')
							break;
						for (end2=pos; end2<end && !isspace(*end2) && *end2!='#'; end2++)
							;

						switch (field) {
							/* Regex to match this device */
							case 3:
							{
								char *regex = strndupa(pos,end2-pos);
								regex_t match;
								regmatch_t off;
								int result;

								/* Is this it? */
								xregcomp(&match,regex,REG_EXTENDED);
								result = regexec(&match, device_name,1,&off,0);
								regfree(&match);

								/* If not this device, skip rest of line */
								if (result || off.rm_so || off.rm_eo!=strlen(device_name))
									goto end_line;

								break;
							}
							/* uid:gid */
							case 2:
							{
								char *s2;

								/* Find : */
								for(s=pos; s<end2 && *s!=':'; s++)
									;
								if (s == end2)
									goto end_line;

								/* Parse UID */
								uid = strtoul(pos,&s2,10);
								if (s != s2) {
									struct passwd *pass;
									pass = getpwnam(strndupa(pos,s-pos));
									if (!pass)
										goto end_line;
									uid = pass->pw_uid;
								}
								s++;
								/* parse GID */
								gid = strtoul(s,&s2,10);
								if (end2 != s2) {
									struct group *grp;
									grp = getgrnam(strndupa(s,end2-s));
									if (!grp)
										goto end_line;
									gid = grp->gr_gid;
								}
								break;
							}
							/* mode */
							case 1:
							{
								mode = strtoul(pos,&pos,8);
								if (pos != end2)
									goto end_line;
								else
									goto found_device;
							}
						}
						pos = end2;
					}
end_line:
					/* Did everything parse happily? */
					if (field && field!=3)
						bb_error_msg_and_die("Bad line %d",line);

					/* Next line */
					pos = ++end;
				}
found_device:
				munmap(conf, len);
			}
			close(fd);
		}
	}

	sprintf(temp_md, "%s/%s", DEV_PATH, device_name);
	umask(0);
	if (mknod(temp_md, mode | type, makedev(device_major, device_minor)) && errno != EEXIST)
		bb_perror_msg_and_die("mknod %s failed", temp_md);

	if (device_major==root_major && device_minor==root_minor)
		symlink(temp_md, DEV_PATH "/root");

	if (ENABLE_FEATURE_MDEV_CONF) chown(temp_md,uid,gid);
}

/* Recursive search of /sys/block or /sys/class.  path must be a writeable
 * buffer of size PATH_MAX containing the directory string to start at. */

static void find_dev(char *path)
{
	DIR *dir;
	size_t len = strlen(path);
	struct dirent *entry;

	if ((dir = opendir(path)) == NULL)
		return;

	while ((entry = readdir(dir)) != NULL) {

		/* Skip "." and ".." (also skips hidden files, which is ok) */

		if (entry->d_name[0] == '.')
			continue;

		if (entry->d_type == DT_DIR) {
			snprintf(path+len, PATH_MAX-len, "/%s", entry->d_name);
			find_dev(path);
			path[len] = 0;
		}

		/* If there's a dev entry, mknod it */

		if (!strcmp(entry->d_name, "dev")) make_device(path);
	}

	closedir(dir);
}

int mdev_main(void)
{
	char *action;
	char *env_path;

	/* Scan */

	if (Gargc == 2 && !strcmp(Gargv[1], "-s")) {
		struct stat st;

		stat("/", &st);  // If this fails, we have bigger problems.
		root_major = major(st.st_dev);
		root_minor = minor(st.st_dev);
		strcpy(temp,"/sys/block");
		find_dev(temp);
		strcpy(temp,"/sys/class");
		find_dev(temp);

	/* Hotplug */

	} else {
		action = getenv("ACTION");
		env_path = getenv("DEVPATH");
	    if (!action || !env_path)
			bb_show_usage();

		if (!strcmp(action, "add")) {
			sprintf(temp, "/sys%s", env_path);
			make_device(temp);
		} else if (!strcmp(action, "remove")) {
			sprintf(temp, "%s/%s", DEV_PATH, bb_basename(env_path));
			unlink(temp);
		}
	}

	return 0;
}
#endif /* APPLET_USAGE */
