/*
 * mkswap.c - set up a linux swap device
 *
 * (C) 1991 Linus Torvalds. This file may be redistributed as per
 * the Linux copyright.
 */

/*
 * 20.12.91  -  time began. Got VM working yesterday by doing this by hand.
 *
 * Usage: mkswap [-c] [-vN] [-f] device [size-in-blocks]
 *
 *      -c   for readability checking. (Use it unless you are SURE!)
 *      -vN  for swap areas version N. (Only N=0,1 known today.)
 *      -f   for forcing swap creation even if it would smash partition table.
 *
 * The device may be a block device or an image of one, but this isn't
 * enforced (but it's not much fun on a character device :-).
 *
 * Patches from jaggy@purplet.demon.co.uk (Mike Jagdis) to make the
 * size-in-blocks parameter optional added Wed Feb  8 10:33:43 1995.
 *
 * Version 1 swap area code (for kernel 2.1.117), aeb, 981010.
 *
 * Sparc fixes, jj@ultra.linux.cz (Jakub Jelinek), 981201 - mangled by aeb.
 * V1_MAX_PAGES fixes, jj, 990325.
 *
 * 1999-02-22 Arkadiusz Mikiewicz <misiek@misiek.eu.org>
 * - added Native Language Support
 *
 *  from util-linux -- adapted for busybox by
 *  Erik Andersen <andersen@codepoet.org>. I ripped out Native Language
 *  Support, made some stuff smaller, and fitted for life in busybox.
 *
 */
#ifdef APPLET_USAGE
#define mkswap_trivial_usage \
	"[-c] [-v0|-v1] device [block-count]"
#define mkswap_full_usage \
	"Prepare a disk partition to be used as a swap partition.\n\n" \
	"Options:\n" \
	"\t-c\t\tCheck for read-ability\n" \
	"\t-v0\t\tMake version 0 swap [max 128 Megs]\n" \
	"\t-v1\t\tMake version 1 swap [big!] (default for kernels >\n\t\t\t2.1.117)\n" \
	"\tblock-count\tNumber of block to use (default is entire partition)"
#else

#include "busybox.h"
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mount.h> /* BLKGETSIZE */
#include "bb_xopen.h"

/*
 * The definition of the union swap_header uses the constant PAGE_SIZE.
 * Unfortunately, on some architectures this depends on the hardware model,
 * and can only be found at run time -- we use getpagesize().
 */

struct swap_header_v1 {
	char bootbits[1024];            /* Space for disklabel etc. */
	unsigned int version;
	unsigned int last_page;
	unsigned int nr_badpages;
	unsigned int padding[125];
	unsigned int badpages[1];
};

CALC_BSS struct BB_COMMON_BSS
{
	char *device_name;
	long PAGES;
	int check;
	int g_badpages;
	int pagesize;
	unsigned int *signature_page;
	struct swap_header_v1 *p;
	int DEV;
	int comm_version;
} bb_common_bss;
#define device_name bb_common_bss.device_name
#define PAGES bb_common_bss.PAGES
#define check bb_common_bss.check
#define g_badpages bb_common_bss.g_badpages
#define pagesize bb_common_bss.pagesize
#define signature_page bb_common_bss.signature_page
#define p bb_common_bss.p
#define DEV bb_common_bss.DEV
#define comm_version bb_common_bss.comm_version


#define MAKE_VERSION(p,q,r)     (65536*(p) + 256*(q) + (r))


static inline void init_signature_page(void)
{
	pagesize = getpagesize();

#ifdef PAGE_SIZE
	if (pagesize != PAGE_SIZE)
		bb_error_msg("Assuming pages of size %d", pagesize);
#endif
	signature_page = (unsigned int *) xmalloc(pagesize);
	memset(signature_page, 0, pagesize);
	p = (struct swap_header_v1 *) signature_page;
}

static inline void write_signature(char *sig)
{
	char *sp = (char *) signature_page;

	strncpy(sp + pagesize - 10, sig, 10);
}

#define V0_MAX_PAGES    (8 * (pagesize - 10))
/* Before 2.2.0pre9 */
#define V1_OLD_MAX_PAGES        ((0x7fffffff / pagesize) - 1)
/* Since 2.2.0pre9:
   error if nr of pages >= SWP_OFFSET(SWP_ENTRY(0,~0UL))
   with variations on
	#define SWP_ENTRY(type,offset) (((type) << 1) | ((offset) << 8))
	#define SWP_OFFSET(entry) ((entry) >> 8)
   on the various architectures. Below the result - yuk.

   Machine      pagesize        SWP_ENTRY       SWP_OFFSET      bound+1 oldbound+2
   i386         2^12            o<<8            e>>8            1<<24   1<<19
   mips         2^12            o<<15           e>>15           1<<17   1<<19
   alpha        2^13            o<<40           e>>40           1<<24   1<<18
   m68k         2^12            o<<12           e>>12           1<<20   1<<19
   sparc        2^{12,13}       (o&0x3ffff)<<9  (e>>9)&0x3ffff  1<<18   1<<{19,18}
   sparc64      2^13            o<<13           e>>13           1<<51   1<<18
   ppc          2^12            o<<8            e>>8            1<<24   1<<19
   armo         2^{13,14,15}    o<<8            e>>8            1<<24   1<<{18,17,16}
   armv         2^12            o<<9            e>>9            1<<23   1<<19

   assuming that longs have 64 bits on alpha and sparc64 and 32 bits elsewhere.

   The bad part is that we need to know this since the kernel will
   refuse a swap space if it is too large.
*/
/* patch from jj - why does this differ from the above? */
#if defined(__alpha__)
#define V1_MAX_PAGES           ((1 << 24) - 1)
#elif defined(__mips__)
#define V1_MAX_PAGES           ((1 << 17) - 1)
#elif defined(__sparc_v9__)
#define V1_MAX_PAGES           ((3 << 29) - 1)
#elif defined(__sparc__)
#define V1_MAX_PAGES           (pagesize == 8192 ? ((3 << 29) - 1) : ((1 << 18) - 1))
#else
#define V1_MAX_PAGES           V1_OLD_MAX_PAGES
#endif
/* man page now says:
The maximum useful size of a swap area now depends on the architecture.
It is roughly 2GB on i386, PPC, m68k, ARM, 1GB on sparc, 512MB on mips,
128GB on alpha and 3TB on sparc64.
*/

#define MAX_BADPAGES    ((pagesize-1024-128*sizeof(int)-10)/sizeof(int))

static inline void bit_set(unsigned int *addr, unsigned int nr)
{
	unsigned int r, m;

	addr += nr / (8 * sizeof(int));

	r = *addr;
	m = 1 << (nr & (8 * sizeof(int) - 1));

	*addr = r | m;
}

static int bit_test_and_clear(unsigned int *addr, unsigned int nr)
{
	unsigned int r, m;

	addr += nr / (8 * sizeof(int));

	r = *addr;
	m = 1 << (nr & (8 * sizeof(int) - 1));

	*addr = r & ~m;
	return (r & m) != 0;
}


static void page_ok(int page)
{
	if (comm_version == 0)
		bit_set(signature_page, page);
}

static inline void page_bad(int page)
{
	if (comm_version == 0)
		bit_test_and_clear(signature_page, page);
	else {
		if (g_badpages == MAX_BADPAGES)
			bb_error_msg_and_die("too many bad pages");
		p->badpages[g_badpages] = page;
	}
	g_badpages++;
}

static void check_blocks(void)
{
	unsigned int current_page;
	int do_seek = 1;
	char *buffer;

	buffer = xmalloc(pagesize);
	current_page = 0;
	while (current_page < PAGES) {
		if (!check) {
			page_ok(current_page++);
			continue;
		}
		if (do_seek && lseek(DEV, current_page * pagesize, SEEK_SET) !=
			current_page * pagesize)
			bb_error_msg_and_die("seek failed in check_blocks");
		if ((do_seek = (pagesize != read(DEV, buffer, pagesize)))) {
			page_bad(current_page++);
			continue;
		}
		page_ok(current_page++);
	}
	if (g_badpages == 1)
		printf("one bad page\n");
	else if (g_badpages > 1)
		printf("%d bad pages\n", g_badpages);
}

static long valid_offset(int fd, int offset)
{
	char ch;

	if (lseek(fd, offset, 0) < 0)
		return 0;
	if (read(fd, &ch, 1) < 1)
		return 0;
	return 1;
}

static int find_size(int fd)
{
	unsigned int high, low;

	low = 0;
	for (high = 1; high > 0 && valid_offset(fd, high); high *= 2)
		low = high;
	while (low < high - 1) {
		const int mid = (low + high) / 2;

		if (valid_offset(fd, mid))
			low = mid;
		else
			high = mid;
	}
	return (low + 1);
}

/* return size in pages, to avoid integer overflow */
static long get_size(const char *file)
{
	int fd;
	long size;

	fd = open(file, O_RDONLY | BB_OPEN_ERR);
	if (ioctl(fd, BLKGETSIZE, &size) >= 0) {
		int sectors_per_page = pagesize / 512;

		size /= sectors_per_page;
	} else {
		size = find_size(fd) / pagesize;
	}
	close(fd);
	return size;
}

int mkswap_main(void)
{
	struct stat statbuf;
	int sz;
	int maxpages;
	int goodpages;
	int offset;
	int force;
	char **argv;

	comm_version = -1;
	force = bb_xgetopt_ulflags("cfv:\0:-v:-1?2", &comm_version);
	check = force & 1;
	force &= 2;
	init_signature_page();          /* get pagesize */
	argv = Gargv;
	device_name = *argv++;
	if(*argv != NULL) {
		int blocks_per_page = pagesize / 1024;

		PAGES = bb_xparse_ul_number(10, *argv) / blocks_per_page;
	}
	sz = get_size(device_name);
	if (!PAGES) {
		PAGES = sz;
	} else if (PAGES > sz && !force) {
		bb_error_msg("error: size %ld is larger than device size %d",
				PAGES * (pagesize / 1024), sz * (pagesize / 1024));
		return EXIT_FAILURE;
	}

	if (comm_version == -1) {
		if (get_kernel_revision() < MAKE_VERSION(2, 1, 117))
			comm_version = 0;
		else
			comm_version = 1;
	}
	if (comm_version != 0 && comm_version != 1) {
		bb_error_msg("error: unknown version %d", comm_version);
		bb_show_usage();
	}
	if (PAGES < 10) {
		bb_error_msg("error: swap area needs to be at least %ldkB",
				(long) (10 * pagesize / 1024));
		bb_show_usage();
	}
#if 0
	maxpages = ((comm_version == 0) ? V0_MAX_PAGES : V1_MAX_PAGES);
#else
	if (!comm_version)
		maxpages = V0_MAX_PAGES;
	else if (get_kernel_revision() >= MAKE_VERSION(2, 2, 1))
		maxpages = V1_MAX_PAGES;
	else {
		maxpages = V1_OLD_MAX_PAGES;
		if (maxpages > V1_MAX_PAGES)
			maxpages = V1_MAX_PAGES;
	}
#endif
	if (PAGES > maxpages) {
		PAGES = maxpages;
		bb_error_msg("warning: truncating swap area to %ldkB",
				PAGES * pagesize / 1024);
	}

	DEV = bb_xopen(device_name, O_RDWR | BB_OPEN_ERR);
	if (fstat(DEV, &statbuf) < 0)
		bb_perror_msg_and_die("%s", device_name);
	if (!S_ISBLK(statbuf.st_mode))
		check = 0;
	else if (statbuf.st_rdev == 0x0300 || statbuf.st_rdev == 0x0340)
		bb_error_msg_and_die("Will not try to make swapdevice on '%s'", device_name);

#ifdef __sparc__
	if (!force && comm_version == 0) {
		/* Don't overwrite partition table unless forced */
		unsigned char *buffer = (unsigned char *) signature_page;
		unsigned short *q, sum;

		if (read(DEV, buffer, 512) != 512)
			bb_error_msg_and_die("fatal: first page unreadable");
		if (buffer[508] == 0xDA && buffer[509] == 0xBE) {
			q = (unsigned short *) (buffer + 510);
			for (sum = 0; q >= (unsigned short *) buffer;)
				sum ^= *q--;
			if (!sum) {
				bb_error_msg("Device '%s' contains a valid Sun disklabel.\n"
"This probably means creating v0 swap would destroy your partition table\n"
"No swap created. If you really want to create swap v0 on that device, use\n"
"the -f option to force it.", device_name);
				return EXIT_FAILURE;
			}
		}
	}
#endif

	if (comm_version == 0 || check)
		check_blocks();
	if (comm_version == 0 && !bit_test_and_clear(signature_page, 0))
		bb_error_msg_and_die("fatal: first page unreadable");
	if (comm_version == 1) {
		p->version = comm_version;
		p->last_page = PAGES - 1;
		p->nr_badpages = g_badpages;
	}

	goodpages = PAGES - g_badpages - 1;
	if (goodpages <= 0)
		bb_error_msg_and_die("Unable to set up swap-space: unreadable");
	printf("Setting up swapspace version %d, size = %ld bytes\n",
		   comm_version, (long) (goodpages * pagesize));
	write_signature((comm_version == 0) ? "SWAP-SPACE" : "SWAPSPACE2");

	offset = ((comm_version == 0) ? 0 : 1024);
	if (lseek(DEV, offset, SEEK_SET) != offset)
		bb_error_msg_and_die("unable to rewind swap-device");
	if (write_all(DEV, (char *) signature_page + offset, pagesize - offset) != pagesize - offset)
		bb_error_msg_and_die("unable to write signature page");

	/* A subsequent swapon() will fail if the signature
	 * is not actually on disk. (This is a kernel bug.) */
	if (fsync_close(DEV))
		bb_error_msg_and_die("write failed");
	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
