/*
 * Mini more implementation for busybox
 *
 * Copyright (C) 1995, 1996 by Bruce Perens <bruce@pixar.com>.
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 * Copyright (C) 2005, 2008 by Vladimir Oleynik <dzo@simtreas.ru>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */
#ifdef APPLET_USAGE
#define more_trivial_usage \
	"[FILE ...]"
#define more_full_usage \
	"More is a filter for viewing FILE one screenful at a time."
#else

#include "busybox.h"
#include <unistd.h>
#include <sys/ioctl.h>
#ifdef CONFIG_FEATURE_USE_TERMIOS
#include <termios.h>
#include "bb_sig.h"
#endif


CALC_BSS struct BB_COMMON_BSS {
#ifdef CONFIG_FEATURE_USE_TERMIOS
	int cin_fileno;
	struct termios initial_settings, new_settings;
#endif
} bb_common_bss;
#define cin_fileno bb_common_bss.cin_fileno
#define initial_settings bb_common_bss.initial_settings
#define new_settings bb_common_bss.new_settings

#ifdef CONFIG_FEATURE_USE_TERMIOS
#define setTermSettings(fd,argp) tcsetattr(fd,TCSANOW,argp)
#define getTermSettings(fd,argp) tcgetattr(fd, argp);

static void set_tty_to_initial_mode(void)
{
	setTermSettings(cin_fileno, &initial_settings);
}

static void gotsig(int sig)
{
	putchar('\n');
	exit(EXIT_FAILURE);
}
#endif /* CONFIG_FEATURE_USE_TERMIOS */

static int input_after_prompt(FILE *cin, int len)
{
	/*
	* We've just displayed the "--More--" prompt, so now we need
	* to get input from the user.
	*/
	fflush(stdout);

	int input = getc(cin);
#ifndef CONFIG_FEATURE_USE_TERMIOS
	printf("\033[A"); /* up cursor */
#endif
	/* Erase the "More" message */
	putc('\r', stdout);
	while (--len >= 0)
		putc(' ', stdout);
	putc('\r', stdout);
	return input;
}

extern int more_main(void)
{
	char **argv = Gargv;
	int c, lines, input = 0;
	int please_display_more_prompt = 0;
	struct stat st;
	FILE *file;
	FILE *cin = NULL;
	int len, page_height;
	int terminal_width;
	int terminal_height;
	int next_file = 0;
#ifndef CONFIG_FEATURE_USE_TERMIOS
	int cin_fileno;
#endif

	argv++;

	/* not use inputing from terminal if usage: more > outfile */
	if(isatty(STDOUT_FILENO)) {
		cin = fopen(bb_dev_tty, "r");
		if (cin == NULL)
			cin = bb_xfopen(CONSOLE_DEV, "Ar");
		cin_fileno = fileno(cin);
#ifdef CONFIG_FEATURE_USE_TERMIOS
		getTermSettings(cin_fileno, &initial_settings);
		new_settings = initial_settings;
		new_settings.c_lflag &= ~ICANON;
		new_settings.c_lflag &= ~ECHO;
		new_settings.c_cc[VMIN] = 1;
		new_settings.c_cc[VTIME] = 0;
		setTermSettings(cin_fileno, &new_settings);
		atexit(set_tty_to_initial_mode);
		bb_signal( (1 << SIGINT) | (1 << SIGTERM)
				| (1 << SIGQUIT), gotsig);
#endif
	}

	if(Gargc == 1 && isatty(STDIN_FILENO) == 1) {
		/* do not call "more < /dev/tty" */
		bb_show_usage();
	}
	if(isatty(STDIN_FILENO) == 0) {
		/* if use "more [file2...] < file1"
		   and file1 is not tty, then show file1 first */
		argv--;
		file = stdin;
		goto show_stdin;
	}
	do {
		file = bb_xfopen(*argv, "Ar");
show_stdin:
		if(file == NULL)
			goto loop;
		next_file++;
		get_terminal_width_height(&terminal_width, &terminal_height);
		if(next_file > 1 && cin != NULL) {
			len = printf("--More--(Next file: %s)", *argv);
			input = input_after_prompt(cin, len % terminal_width);
			if (input == 'q')
				goto end;
		}
		st.st_size = 0;
		fstat(fileno(file), &st);

		please_display_more_prompt = 0;

		if (terminal_height > 4)
			terminal_height -= 2;
		if (terminal_width > 0)
			terminal_width -= 1;

		len = 0;
		lines = 0;
		page_height = terminal_height;
		while ((c = getc(file)) != EOF) {

			if (cin != NULL && please_display_more_prompt) {
				len = printf("--More-- ");
				if (file != stdin && st.st_size > 0) {
					len += printf("(%d%% of %" P_ST_SIZE "d bytes)",
						   (int) (100 * ((double) BB_ftello(file) /
						   (double) st.st_size)), st.st_size);
				}

				input = input_after_prompt(cin, len);
				len = 0;
				lines = 0;
				page_height = terminal_height;
				please_display_more_prompt = 0;

				if (input == 'q')
					goto end;
			}

			/*
			 * There are two input streams to worry about here:
			 *
			 * c     : the character we are reading from the file being "mored"
			 * input : a character received from the keyboard
			 *
			 * If we hit a newline in the _file_ stream, we want to test and
			 * see if any characters have been hit in the _input_ stream. This
			 * allows the user to quit while in the middle of a file.
			 */
			if (c == '\n') {
				/* increment by just one line if we are at
				 * the end of this line */
				if (input == '\n')
					please_display_more_prompt++;
				/* Adjust the terminal height for any overlap, so that
				 * no lines get lost off the top. */
				if (len >= terminal_width) {
					int quot, rem;
					quot = len / terminal_width;
					rem  = len - (quot * terminal_width);
					if (quot) {
						if (rem)
							page_height-=quot;
						else
							page_height-=(quot-1);
					}
				}
				if (++lines >= page_height) {
					please_display_more_prompt++;
				}
				len = 0;
			}
			/*
			 * If we just read a newline from the file being 'mored' and any
			 * key other than a return is hit, scroll by one page
			 */
			putc(c, stdout);
			len++;
		}
		fclose(file);
		fflush(stdout);
loop:
		argv++;
	} while (*argv != NULL);
  end:
	return 0;
}
#endif /* APPLET_USAGE */
