/*
 * Mini mount implementation for busybox
 *
 * Copyright (C) 1995, 1996 by Bruce Perens <bruce@pixar.com>.
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 * Copyright (C) 2005 by Rob Landley <rob@landley.net>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */
#ifdef APPLET_USAGE
#define mount_trivial_usage \
	"[flags] DEVICE NODE [-o options,more-options]"
#define mount_full_usage \
	"Mount a filesystem. Filesystem autodetection requires /proc be mounted.\n\n" \
	"Flags:\n" \
	"\t-a:\t\tMount all filesystems in fstab\n" \
	USE_FEATURE_MTAB_SUPPORT( \
	"\t-f:\t\t\"Fake\" Add entry to mount table but don't mount it\n" \
	"\t-n:\t\tDon't write a mount table entry\n" \
	) \
	"\t-o option:\tOne of many filesystem options, listed below\n" \
	"\t-r:\t\tMount the filesystem read-only\n" \
	"\t-t fs-type:\tSpecify the filesystem type\n" \
	"\t-w:\t\tMount for reading and writing (default)"
#else

#define NEED_MNTENT
#include "busybox.h"
#include <unistd.h>
#include <string.h>
#include <mntent.h>
#include <ctype.h>
#include <sys/mount.h>
#include <sys/ioctl.h>  // for CONFIG_FEATURE_MOUNT_LOOP
#include <sys/wait.h>
#include "config_read.h"

#ifndef HAVE_ADDMNTENT
int addmntent(FILE * filep, const struct mntent *mnt);
#endif

// These two aren't always defined in old headers
#ifndef MS_BIND
#define MS_BIND         4096
#endif
#ifndef MS_MOVE
#define MS_MOVE         8192
#endif

/* Consume standard mount options (from -o options or --options).
 * Set appropriate flags and collect unrecognized ones as a comma separated
 * string to pass to kernel */

struct {
	const char *name;
	long flags;
} static const  mount_options[] = {
	{"loop", 0},
	{"defaults", 0},
	{"noauto", 0},
	{"ro", MS_RDONLY},
	{"rw", ~MS_RDONLY},
	{"nosuid", MS_NOSUID},
	{"suid", ~MS_NOSUID},
	{"dev", ~MS_NODEV},
	{"nodev", MS_NODEV},
	{"exec", ~MS_NOEXEC},
	{"noexec", MS_NOEXEC},
	{"sync", MS_SYNCHRONOUS},
	{"async", ~MS_SYNCHRONOUS},
	{"remount", MS_REMOUNT},
	{"atime", ~MS_NOATIME},
	{"noatime", MS_NOATIME},
	{"diratime", ~MS_NODIRATIME},
	{"nodiratime", MS_NODIRATIME},
	{"bind", MS_BIND},
	{"move", MS_MOVE}
};

/* Uses the mount_options list above */
static int parse_mount_options(int flags, char *options, char **strflags)
{
	if(options == NULL)
		return 0;
	// Loop through options
	for(;;) {
		int i;
		char *comma = strchr_set0(options, ',');

		// Find this option in mount_options
		for(i = 0; i < (sizeof(mount_options) / sizeof(*mount_options)); i++) {
			if(!strcasecmp(mount_options[i].name, options)) {
				long fl = mount_options[i].flags;
				if(fl < 0) flags &= fl;
				else flags |= fl;
				break;
			}
		}
		// Unrecognized mount option?
		if(i == (sizeof(mount_options) / sizeof(*mount_options))) {
			// Add it to strflags, to pass on to kernel
			i = *strflags ? strlen(*strflags) : 0;
			*strflags = xrealloc(*strflags, i+strlen(options)+2);
			// Comma separated if it's not the first one
			if(i) (*strflags)[i++] = ',';
			strcpy((*strflags)+i, options);
		}
		// Advance to next option, or finish
		if(comma) {
			*comma = ',';
			options = ++comma;
		} else break;
	}
	return flags;
}

CALC_BSS struct BB_COMMON_BSS
{
	char *minus_o;
} bb_common_bss;
#define minus_o bb_common_bss.minus_o

static void split_minus_o(const char *opt_arg)
{
	if(minus_o == NULL) {
		minus_o = xstrdup(opt_arg);
	} else {
		char *new_o = xasprintf("%s,%s", minus_o, opt_arg);
		free(minus_o);
		minus_o = new_o;
	}
}

static void exec_nfsmount(char *blockDevice, char *directory, int opt
#if ENABLE_FEATURE_MTAB_SUPPORT
		, int useMtab
#endif
		)
{
	char *arg[12];
	int i = 4;
#ifndef CONFIG_INSTALL_NO_USR
	arg[0] = "/usr/sbin/mount.nfs";
#else
	arg[0] = "/sbin/mount.nfs";
#endif
	arg[1] = "mount.nfs";
	arg[2] = blockDevice;
	arg[3] = directory;
	if(opt & 4)
		arg[i++] = "-r";
	if(opt & 8)
		arg[i++] = "-w";
	if(opt & 16)
		arg[i++] = "-f";
	if(opt & 128)
		arg[i++] = "-v";
#if ENABLE_FEATURE_MTAB_SUPPORT
	if(!useMtab)
#endif
		arg[i++] = "-n";
	if(minus_o != NULL) {
		arg[i++] = "-o";
		arg[i++] = minus_o;
	}
	arg[i] = NULL;
	i = vfork();
	if(i < 0)
		bb_perror_msg_and_die("vfork()");
	if(i == 0) {
	     execv(arg[0], arg + 1);
	     bb_perror_msg_and_die("%s", arg[0]);
	}
	wait (NULL);
}

#define FLAGS0 0xc0ed0000   /* Needed for linux 2.2, ignored by 2.4 and 2.6. */

extern int mount_main(void)
{
	char **argv = Gargv;
	static const char * const files[] = {"/etc/filesystems", "/proc/filesystems", 0};
	char *string_flags = 0, *fsType = 0, *blockDevice = 0, *directory = 0,
		 *loopFile = 0;
	int i, opt, all, fakeIt, rc = 1;
	int flags = 0;
	FILE *file = 0;
	char path[PATH_MAX*2];
	struct mntent m;
	struct stat statbuf;
	config_read_t *f = NULL;
#if ENABLE_FEATURE_MTAB_SUPPORT
	int useMtab = TRUE;
#else
# define useMtab 0
#endif
	void *config_opened = NULL;

	/* parse long options, like --bind and --move.  Note that -o option
	 * and --option are synonymous.  Yes, this means --remount,rw works. */

	for(i = opt = 0; argv[i] != NULL; i++) {
		if(argv[i][0] == '-' && argv[i][1] == '-')
			split_minus_o(argv[i]+2);
		else argv[opt++] = argv[i];
	}
	argv[opt] = NULL;

	// Parse remaining options
	opt = bb_xgetopt_ulflags("o:t:rwfanv\0=o:r-w:w-r", split_minus_o, &fsType);
	fakeIt = opt & 16;
	all = opt & 32;
#if ENABLE_FEATURE_MTAB_SUPPORT
	if(opt & 64)
		useMtab = FALSE;
#endif
	argv = Gargv;
	// If we have no arguments, show currently mounted filesystems
	if(!all && argv[0] == NULL) {
		FILE *mountTable = setmntent(bb_path_mtab_file, "r");

		if(!mountTable) bb_perror_msg_and_die("setmntent(%s)", bb_path_mtab_file);

		while (getmntent_r(mountTable, &m, path, sizeof(path))) {
			blockDevice = m.mnt_fsname;

			// Clean up display a little bit regarding root device
			if(!strcmp(blockDevice, "rootfs")) continue;
			if(!strcmp(blockDevice, "/dev/root"))
				blockDevice = find_block_device("/");

			if(!fsType || !strcmp(m.mnt_type, fsType))
				printf("%s on %s type %s (%s)\n", blockDevice, m.mnt_dir,
					   m.mnt_type, m.mnt_opts);
		}
		endmntent(mountTable);
		return EXIT_SUCCESS;
	}

	/* The next argument is what to mount.  if there's an argument after that
	 * it's where to mount it.  If we're not mounting all, and we have both
	 * of these arguments, jump straight to the actual mount. */
	statbuf.st_mode = 0;
	if(argv[0] != NULL) {
		blockDevice = !stat(*argv, &statbuf) ?
				bb_simplify_path(*argv) : *argv;
		if(argv[1] != NULL) {
			directory = bb_simplify_path(argv[1]);
			/* If we don't have to loop through fstab,
			   skip ahead a bit. */
			if(!all)
				goto singlemount;
		}
	}

	// Loop through /etc/fstab entries to look up this entry.
	if(!(file = setmntent("/etc/fstab", "r")))
		bb_perror_msg_and_die("Cannot read /etc/fstab");
	for(;;) {
		// Get next fstab entry
		if(!getmntent_r(file, &m, path, sizeof(path))) {
			if(!all)
				bb_perror_msg("Can't find %s in /etc/fstab", blockDevice);
			break;
		}

		// If we're mounting all and all doesn't mount this one, skip it.
		if(all) {
			if(strstr(m.mnt_opts, "noauto") || strstr(m.mnt_type, "swap"))
				continue;

		/* If we're mounting something specific and this isn't it, skip it.
		 * Note we must match both the exact text in fstab (ala "proc") or
		 * a full path from root */
		} else if(strcmp(blockDevice, m.mnt_fsname) &&
				  strcmp(*argv, m.mnt_fsname) &&
				  strcmp(blockDevice, m.mnt_dir) &&
				  strcmp(*argv, m.mnt_dir))
					continue;

		/* Parse flags from /etc/fstab */
		flags = parse_mount_options(0, m.mnt_opts, &string_flags);

		/* Fill out remaining fields with info from mtab */
		blockDevice = m.mnt_fsname;
		directory = m.mnt_dir;
		fsType = m.mnt_type;

		/* Ok, we're ready to actually mount a specific source on a specific
		 * directory now. */
singlemount:
		flags = parse_mount_options(flags | FLAGS0, minus_o, &string_flags);
		// If they said -r or -w, override fstab
		if(opt & 4)
			flags |= MS_RDONLY;
		if(opt & 8)
			flags &= ~MS_RDONLY;

		// Might this be an NFS filesystem?
		if((!fsType || !strcmp(fsType,"nfs")) && strchr(blockDevice, ':') != NULL) {
			rc = 0;
			exec_nfsmount(blockDevice, directory,
					opt & (4|8|16|128)
#if ENABLE_FEATURE_MTAB_SUPPORT
					, useMtab
#endif
					);
			if(!all)
				break;
			continue;
		} else {
			// Do we need to allocate a loopback device?
			if(ENABLE_FEATURE_MOUNT_LOOP && !fakeIt && S_ISREG(statbuf.st_mode))
			{
				loopFile = blockDevice;
				blockDevice = 0;
				switch(set_loop(&blockDevice, loopFile, 0)) {
					case 0:
					case 1:
						break;
					default:
						bb_error_msg_and_die("%s",
							errno == EPERM || errno == EACCES ?
							bb_msg_perm_denied_are_you_root :
							"Couldn't setup loop device");
						break;
				}
			}

			/* If we know the fstype (or don't need to), jump straight
			 * to the actual mount. */
			if(fsType || (flags & (MS_REMOUNT | MS_BIND | MS_MOVE)))
				goto mount_it_now;
		}

		// Loop through filesystem types until mount succeeds or we run out
		for(i = 0; files[i] && rc; i++) {
			f = config_opened = config_read_prepare(files[i], 0);

			if(f == NULL) continue;
			// Get next block device backed filesystem
			while( (f = config_read(f, "#\1\2\0\0")) != NULL) {
				fsType = f->tokens[0];
				if(*fsType == '*') continue;
				// Skip funky entries in /proc
				if(!strcmp(fsType, "nodev") && f->ntokens > 1) continue;
mount_it_now:
				// Okay, try to mount
				if (!fakeIt) {
					for(;;) {
						rc = mount(blockDevice, directory, fsType, flags, string_flags);
						if(!rc || (flags & MS_RDONLY) || (errno!=EACCES && errno!=EROFS))
							break;
						bb_error_msg("%s is write-protected, mounting read-only", blockDevice);
						flags |= MS_RDONLY;
					}
				}
				if(!rc || !f)
					break;
			}
			if(!config_opened)
				break;
			// goto mount_it_now with -a can jump past the initialization
			config_close(f);
			f = config_opened = 0;
		}

		/* If the mount was successful, and we're maintaining an old-style
		 * mtab file by hand, add new entry to it now. */
		if((!rc || fakeIt) && useMtab) {
			FILE *mountTable = setmntent(bb_path_mtab_file, "a+");

			if(!mountTable) bb_perror_msg("setmntent(%s)", bb_path_mtab_file);
			else {
				// Remove trailing / (if any) from directory we mounted on
				int length = strlen(directory);
				if(length > 1 && directory[length-1] == '/')
					directory[length-1] = 0;

				// Fill out structure (should be ok to re-use existing one).
				m.mnt_fsname = blockDevice;
				m.mnt_dir = directory;
				m.mnt_type = fsType ? : "--bind";
				m.mnt_opts = string_flags ? :
					((flags & MS_RDONLY) ? "ro" : "rw");
				m.mnt_freq = 0;
				m.mnt_passno = 0;

				// Write and close
				addmntent(mountTable, &m);
				endmntent(mountTable);
			}
		} else {
			// Mount failed.  Clean up
			if(loopFile) {
				del_loop(blockDevice);
			}
			// Don't whine about already mounted fs when mounting all.
			if(rc < 0 && errno == EBUSY && all) rc = 0;
			else if (errno == EPERM)
				bb_error_msg_and_die("%s", bb_msg_perm_denied_are_you_root);
		}
		// We couldn't free this earlier becase fsType could be in buf.
		if(!all)
			break;
		free(string_flags);
		string_flags = NULL;
	}

	if(file)
		endmntent(file);
	if(rc)
		bb_perror_msg("Mounting %s on %s failed", blockDevice, directory);

	return rc;
}
#endif /* APPLET_USAGE */
