/* rename.c - aeb 2000-01-01 from util-linux 2.33.1
 *
 *  Busybox port by Vladimir Oleynik (c) 2019 <dzo@simtreas.ru>
 */

#ifdef APPLET_USAGE
#define rename_trivial_usage \
	"[-vsnoi] expression replacement file..."
#define rename_full_usage \
	"Rename files.\n" \
	"Options:\n" \
	USE_FEATURE_LONG_OPTIONS( \
	"\t-v, --verbose\t\texplain what is being done\n" \
	"\t-s, --symlink\t\tact on the target of symlinks\n" \
	"\t-n, --no-act\t\tdo not make any changes\n" \
	"\t-o, --no-overwrite\tdon't overwrite existing files\n" \
	"\t-i, --interactive \tprompt before overwrite") \
	SKIP_FEATURE_LONG_OPTIONS( \
	"\t-v\texplain what is being done\n" \
	"\t-s\tact on the target of symlinks\n" \
	"\t-n\tdo not make any changes\n" \
	"\t-o\tdon't overwrite existing files\n" \
	"\t-i\tprompt before overwrite")

#else

#define PACKAGE_STRING "util-linux 2.33.1"

#include "busybox.h"

#define HAVE___FPURGE

#include <stdio_ext.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>

#define RENAME_EXIT_SOMEOK      2
#define RENAME_EXIT_NOTHING     4
#define RENAME_EXIT_UNEXPLAINED 64

#define UTIL_LINUX_VERSION "%s from %s\n", bb_applet_name, PACKAGE_STRING


static int string_replace(char *from, char *to, char *s, char *orig, char **newname)
{
	char *p, *q, *where;

	where = strstr(s, from);
	if (where == NULL)
		return 1;
	p = orig;
	*newname = malloc(strlen(orig) + strlen(to) + 1);
	q = *newname;
	while (p < where)
		*q++ = *p++;
	p = to;
	while (*p)
		*q++ = *p++;
	p = where + strlen(from);
	while (*p)
		*q++ = *p++;
	*q = 0;
	return 0;
}

CALC_BSS struct BB_COMMON_BSS
{
	int tty_cbreak;
} bb_common_bss;
#define tty_cbreak bb_common_bss.tty_cbreak

static int ask(char *name)
{
	int c;
	char buf[2];
	printf("%s: overwrite `%s'? ", bb_applet_name, name);
	fflush(stdout);
	if ((c = fgetc(stdin)) == EOF) {
		buf[0] = 'n';
		printf("n\n");
	}
	else {
		buf[0] = c;
		if (c != '\n' && tty_cbreak) {
#ifdef HAVE___FPURGE
			/* Possibly purge a multi-byte character; or do a
			   required purge of the rest of the line (including
			   the newline) if the tty has been put back in
			   canonical mode (for example by a shell after a
			   SIGTSTP signal). */
			__fpurge(stdin);
#endif
			printf("\n");
		}
		else if (c != '\n')
			while ((c = fgetc(stdin)) != '\n' && c != EOF);
	}
	buf[1] = '\0';
	if (rpmatch(buf) == 1)
		return 0;
	else
		return 1;
}

static int do_symlink(char *from, char *to, char *s, int verbose, int noact,
		      int nooverwrite, int interactive)
{
	char *newname = NULL, *target = NULL;
	int ret = 1;
	struct stat sb;

	if ( faccessat(AT_FDCWD, s, F_OK, AT_SYMLINK_NOFOLLOW) != 0 &&
	     errno != EINVAL )
	   /* Skip if AT_SYMLINK_NOFOLLOW is not supported; lstat() below will
	      detect the access error */
	{
		bb_perror_msg("%s: not accessible", s);
		return 2;
	}

	if (lstat(s, &sb) == -1) {
		bb_perror_msg("stat of %s failed", s);
		return 2;
	}
	if (!S_ISLNK(sb.st_mode)) {
		bb_error_msg("%s: not a symbolic link", s);
		return 2;
	}
	target = malloc(sb.st_size + 1);
	if (readlink(s, target, sb.st_size + 1) < 0) {
		bb_perror_msg("%s: readlink failed", s);
		free(target);
		return 2;
	}
	target[sb.st_size] = '\0';
	if (string_replace(from, to, target, target, &newname) != 0)
		ret = 0;

	if (ret == 1 && (nooverwrite || interactive) && lstat(newname, &sb) != 0)
		nooverwrite = interactive = 0;

	if ( ret == 1 &&
	     (nooverwrite || (interactive && (noact || ask(newname) != 0))) )
	{
		if (verbose)
			printf("Skipping existing link: `%s' -> `%s'\n", s, target);
		ret = 0;
	}

	if (ret == 1) {
		if (!noact && 0 > unlink(s)) {
			bb_perror_msg("%s: unlink failed", s);
			ret = 2;
		}
		else if (!noact && symlink(newname, s) != 0) {
			bb_perror_msg("%s: symlinking to %s failed", s, newname);
			ret = 2;
		}
	}
	if (verbose && (noact || ret == 1))
		printf("%s: `%s' -> `%s'\n", s, target, newname);
	free(newname);
	free(target);
	return ret;
}

static int do_file(char *from, char *to, char *s, int verbose, int noact,
		   int nooverwrite, int interactive)
{
	char *newname = NULL, *file=NULL;
	int ret = 1;

	if (access(s, F_OK) != 0) {
		bb_perror_msg("%s: not accessible", s);
		return 2;
	}

	if (strchr(from, '/') == NULL && strchr(to, '/') == NULL)
		file = strrchr(s, '/');
	if (file == NULL)
		file = s;
	if (string_replace(from, to, file, s, &newname) != 0)
		return 0;

	if ((nooverwrite || interactive) && access(newname, F_OK) != 0)
		nooverwrite = interactive = 0;

	if (nooverwrite || (interactive && (noact || ask(newname) != 0))) {
		if (verbose)
			printf("Skipping existing file: `%s'\n", newname);
		ret = 0;
	}
	else if (!noact && rename(s, newname) != 0) {
		bb_perror_msg("%s: rename to %s failed", s, newname);
		ret = 2;
	}
	if (verbose && (noact || ret == 1))
		printf("`%s' -> `%s'\n", s, newname);
	free(newname);
	return ret;
}

#define COMPLEMENTARY "\0o-i:i-o"
#include "opt_gen.h"
OPT_GEN(
opt(v, VERBOSE_OPT),
opt(s, SYMLINK_OPT),
opt(n, NO_ACT_OPT),
opt(o, NO_OVERWRITE_OPT),
opt(i, INTERACTIVE_OPT),
opt(V, VERSION_OPT),
opt(h, HELP_OPT))
#include "opt_gen.h"

int rename_main(void)
{
	char *from, *to;
	int i, c, ret = 0, verbose, noact, nooverwrite, interactive;
	struct termios tio;
	int (*do_rename)(char *from, char *to, char *s, int verbose, int noact,
			 int nooverwrite, int interactive) = do_file;

#if ENABLE_FEATURE_LONG_OPTIONS
	static const struct option longopts[] = {
		{"verbose", no_argument, NULL, 'v'},
		{"version", no_argument, NULL, 'V'},
		{"help", no_argument, NULL, 'h'},
		{"no-act", no_argument, NULL, 'n'},
		{"no-overwrite", no_argument, NULL, 'o'},
		{"interactive", no_argument, NULL, 'i'},
		{"symlink", no_argument, NULL, 's'},
		{NULL, 0, NULL, 0}
	};
	bb_applet_long_options = longopts;
#endif
	c = bb_xgetopt_ulflags(opt_str);
	verbose = c & VERBOSE_OPT;
	if((c & SYMLINK_OPT))
		do_rename = do_symlink;
	noact = c & NO_ACT_OPT;
	nooverwrite = c & NO_OVERWRITE_OPT;
	interactive = c & INTERACTIVE_OPT;
	if((c & VERSION_OPT)) {
		printf(UTIL_LINUX_VERSION);
		return EXIT_SUCCESS;
	}
	if((c & HELP_OPT)) {
		bb_default_error_retval = EXIT_SUCCESS;
		bb_show_usage();
	}

	if (Gargc < 3) {
		bb_error_msg("not enough arguments");
		bb_show_usage();
	}

	from = Gargv[0];
	to = Gargv[1];

	if (!strcmp(from, to))
		return RENAME_EXIT_NOTHING;

	if (interactive && isatty(STDIN_FILENO) != 0) {
		if (tcgetattr(STDIN_FILENO, &tio) != 0)
			bb_perror_msg("failed to get terminal attributes");
		else if (!(tio.c_lflag & ICANON) && tio.c_cc[VMIN] == 1)
			tty_cbreak = 1;
	}

	for (i = 2; i < Gargc; i++)
		ret |= do_rename(from, to, Gargv[i], verbose, noact, nooverwrite, interactive);

	switch (ret) {
	case 0:
		return RENAME_EXIT_NOTHING;
	case 1:
		return EXIT_SUCCESS;
	case 2:
		return EXIT_FAILURE;
	case 3:
		return RENAME_EXIT_SOMEOK;
	default:
		return RENAME_EXIT_UNEXPLAINED;
	}
}
#endif
