/*
 *  This is the distribution of less, version 418, released 02 Jan 2008.
 *  This program is part of the GNU project (http://www.gnu.org).
 *
 * Copyright (C) 1984-2007  Mark Nudelman
 *
 * You may distribute under the terms of either the GNU General Public
 * License or the Less License, as specified in the README file.
 *
 * Modified by Vladimir Oleynik <dzo@simtreas.ru> (c) 2009-2010
 * - to be used in busybox and size optimizations,
 * - remove termcap library usage, use minimal harcoded vt100-like TERM
 * - can use for HP UX (do not need any other dependences library)
 *
 *
 */
#ifdef APPLET_USAGE
#define less_trivial_usage \
	"[-EMNmh~?] FILE1 FILE2..."
#define less_full_usage \
	"View a file or list of files. The position within files can be\n" \
	"changed, and files can be manipulated in various ways with the\n" \
	"following options:\n\n" \
	"\t-E\tQuit once the end of a file is reached\n" \
	"\t-M\tDisplay a status line containing the current line numbers\n" \
	"\t\tand the percentage through the file\n" \
	"\t-N\tPrefix line numbers to each line\n" \
	"\t-m\tDisplay a status line containing the percentage through the\n" \
	"\t\tfile\n" \
	"\t-~\tSuppress ~s displayed when input past the end of the file is\n" \
	"\t\treached.\n" \
	"\t-h, -?\tDisplay this help message"
#else

#define VERSION "418"

#define BINDIR "/usr/bin"
#define SYSDIR "/etc"

/* User preferences. */

/* SECURE is 1 if you wish to disable a bunch of features in order to
 * be safe to run by unprivileged users. */
#define SECURE          1

/* SHELL_ESCAPE is 1 if you wish to allow shell escapes.
 * (This is possible only if your system supplies the system() function.) */
#define SHELL_ESCAPE    (!SECURE)

/* EXAMINE is 1 if you wish to allow examining files by name
 * from within less. */
#define EXAMINE         (!SECURE)

/* TAB_COMPLETE_FILENAME is 1 if you wish to allow the TAB key
 * to complete filenames at prompts. */
#define TAB_COMPLETE_FILENAME   (!SECURE)

/* CMD_HISTORY is 1 if you wish to allow keys to cycle through
 * previous commands at prompts. */
#define CMD_HISTORY     1

/* HILITE_SEARCH is 1 if you wish to have search targets to be
 * displayed in standout mode. */
#define HILITE_SEARCH   1

/* EDITOR is 1 if you wish to allow editor invocation (the "v" command).
 * (This is possible only if your system supplies the system() function.)
 * EDIT_PGM is the name of the (default) editor to be invoked. */
#define EDITOR          (!SECURE)

/* TAGS is 1 if you wish to support tag files. */
#define TAGS            (!SECURE)

/* USERFILE is 1 if you wish to allow a .less file to specify
 * user-defined key bindings. */
#define USERFILE        (!SECURE)

/* GLOB is 1 if you wish to have shell metacharacters expanded in filenames.
 * This will generally work if your system provides the "popen" function
 * and the "echo" shell command. */
#define GLOB            (!SECURE)

/* PIPEC is 1 if you wish to have the "|" command
 * which allows the user to pipe data into a shell command. */
#define PIPEC           (!SECURE)

/* LOGFILE is 1 if you wish to allow the -l option (to create log files). */
#define LOGFILE         (!SECURE)

/* LESSKEYFILE is the filename of the default lesskey output file
 * (in the HOME directory).
 * LESSKEYFILE_SYS is the filename of the system-wide lesskey output file.
 * DEF_LESSKEYINFILE is the filename of the default lesskey input
 * (in the HOME directory).
 * LESSHISTFILE is the filename of the history file
 * (in the HOME directory). */
#define LESSKEYFILE             ".less"
#define LESSKEYFILE_SYS         SYSDIR "/sysless"
#define DEF_LESSKEYINFILE       ".lesskey"
#define LESSHISTFILE            ".lesshst"


/* HAVE_SHELL is 1 if your system supports a SHELL command interpreter. */
#define HAVE_SHELL      1

/* Default shell metacharacters and meta-escape character. */
#define DEF_METACHARS   "; *?\t\n'\"()<>[]|&^`#\\$%=~"
#define DEF_METAESCAPE  "\\"

/* Sizes of various buffers. */
#define CMDBUF_SIZE     512     /* Buffer for multichar commands */
#define UNGOT_SIZE      100     /* Max chars to unget() */
#define LINEBUF_SIZE    1024    /* Max size of line in input file */
#define OUTBUF_SIZE     1024    /* Output buffer */
#define PROMPT_SIZE     200     /* Max size of prompt string */
#define TAGLINE_SIZE    512     /* Max size of line in tags file */
#define TABSTOP_MAX     32      /* Max number of custom tab stops */

/* Define EDIT_PGM to your editor. */
#define EDIT_PGM "vi"

/* Define HAVE_LOCALE if you have locale.h and setlocale. */
#define HAVE_LOCALE 1

#define SUPPORT_UTF8 0
#if SUPPORT_UTF8
/* Define HAVE_WCTYPE if you have iswupper, iswlower, towupper, towlower. */
#define HAVE_WCTYPE 1
#endif

#include "busybox.h"

#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#if SUPPORT_UTF8
#include <wctype.h>
#endif
#include <string.h>
#include <signal.h>
#include <time.h>
#include <setjmp.h>
#include <sys/ioctl.h>
#include <termios.h>

#include "xregex.h"
//# define NO_REGEX 1

#ifdef __BIONIC__
#undef HAVE_LOCALE
#endif
#if HAVE_LOCALE
#include <locale.h>
#include <langinfo.h>
#endif

/* BSD setjmp() saves (and longjmp() restores) the signal mask.
 * This costs a system call or two per setjmp(), so if possible we clear the
 * signal mask with sigsetmask(), and use _setjmp()/_longjmp() instead.
 * On other systems, setjmp() doesn't affect the signal mask and so
 * _setjmp() does not exist; we just use setjmp(). */
#define HAVE__SETJMP 1
#if HAVE__SETJMP
#define SET_JUMP        _setjmp
#define LONG_JUMP       _longjmp
#else
#define SET_JUMP        setjmp
#define LONG_JUMP       longjmp
#endif

/* Simple lowercase test which can be used during option processing
 * (before options are parsed which might tell us what charset to use). */
#define ASCII_IS_UPPER(c)       ((c) >= 'A' && (c) <= 'Z')
#define ASCII_IS_LOWER(c)       ((c) >= 'a' && (c) <= 'z')
#define ASCII_TO_UPPER(c)       ((c) - 'a' + 'A')
#define ASCII_TO_LOWER(c)       ((c) - 'A' + 'a')

#undef IS_UPPER
#undef IS_LOWER
#undef TO_UPPER
#undef TO_LOWER
#undef IS_SPACE
#undef IS_DIGIT

#if HAVE_WCTYPE
#define IS_UPPER(c)     iswupper(c)
#define IS_LOWER(c)     iswlower(c)
#define TO_UPPER(c)     towupper(c)
#define TO_LOWER(c)     towlower(c)
#else
#define IS_UPPER(c)     isupper((unsigned char) (c))
#define IS_LOWER(c)     islower((unsigned char) (c))
#define TO_UPPER(c)     toupper((unsigned char) (c))
#define TO_LOWER(c)     tolower((unsigned char) (c))
#endif

#ifdef isspace
#define IS_SPACE(c)     isspace((unsigned char)(c))
#else
#define IS_SPACE(c)     ((c) == ' ' || (c) == '\t' || (c) == '\n' || (c) == '\r' || (c) == '\f')
#endif

#ifdef isdigit
#define IS_DIGIT(c)     isdigit((unsigned char)(c))
#else
#define IS_DIGIT(c)     ((c) >= '0' && (c) <= '9')
#endif

#define IS_CSI_START(c) ((c) == ESC || (!utf_mode && ((unsigned char)(c)) == CSI))

#ifndef TRUE
#define TRUE            1
#endif
#ifndef FALSE
#define FALSE           0
#endif

#define OPT_OFF         0
#define OPT_ON          1
#define OPT_ONPLUS      2

#define BAD_LSEEK       ((off_t)-1)

#ifndef CHAR_BIT
#define CHAR_BIT 8
#endif

/* Upper bound on the string length of an integer converted to string.
 * 302 / 1000 is ceil (log10 (2.0)).  Subtract 1 for the sign bit;
 * add 1 for integer division truncation; add 1 more for a minus sign. */
#define INT_STRLEN_BOUND(t) ((sizeof(t) * CHAR_BIT - 1) * 302 / 1000 + 1 + 1)

/* Special types and constants. */
#if SUPPORT_UTF8
typedef unsigned long LWCHAR;
#else
typedef unsigned char LWCHAR;
#endif
typedef off_t           POSITION;
typedef off_t           LINENUM;
#define MIN_LINENUM_WIDTH  7    /* Min printing width of a line number */
#define MAX_UTF_CHAR_LEN   6    /* Max bytes in one UTF-8 char */

#define NULL_POSITION   ((POSITION)(-1))

/* Does the shell treat "?" as a metacharacter? */
#define SHELL_META_QUEST 1

/* An IFILE represents an input file. */
#define IFILE           void *
#define NULL_IFILE      ((IFILE)NULL)

/* The structure used to represent a "screen position".
 * This consists of a file position, and a screen line number.
 * The meaning is that the line starting at the given file
 * position is displayed on the ln-th line of the screen.
 * (Screen lines before ln are empty.) */
struct scrpos
{
	POSITION pos;
	int ln;
};

typedef union parg
{
	const char *p_string;
	int p_int;
	LINENUM p_linenum;
} PARG;

#define NULL_PARG       ((PARG *)NULL)

struct textlist
{
	char *string;
	char *endstring;
};

#define EOI             (-1)

#define READ_INTR       (-2)

/* A fraction is represented by an int n; the fraction is n/NUM_FRAC_DENOM */
#define NUM_FRAC_DENOM                  1000000
#define NUM_LOG_FRAC_DENOM              6

/* How quiet should we be? */
#define NOT_QUIET       0       /* Ring bell at eof and for errors */
#define LITTLE_QUIET    1       /* Ring bell only for errors */
#define VERY_QUIET      2       /* Never ring bell */

/* How should we prompt? */
#define PR_SHORT        0       /* Prompt with colon */
#define PR_MEDIUM       1       /* Prompt with message */
#define PR_LONG         2       /* Prompt with longer message */

/* How should we handle backspaces? */
#define BS_SPECIAL      0       /* Do special things for underlining and bold */
#define BS_NORMAL       1       /* \b treated as normal char; actually output */
#define BS_CONTROL      2       /* \b treated as control char; prints as ^H */

/* How should we search? */
#define SRCH_FORW       (1 << 0)  /* Search forward from current position */
#define SRCH_BACK       (1 << 1)  /* Search backward from current position */
#define SRCH_NO_MOVE    (1 << 2)  /* Highlight, but don't move */
#define SRCH_FIND_ALL   (1 << 4)  /* Find and highlight all matches */
#define SRCH_NO_MATCH   (1 << 8)  /* Search for non-matching lines */
#define SRCH_PAST_EOF   (1 << 9)  /* Search past end-of-file, into next file */
#define SRCH_FIRST_FILE (1 << 10) /* Search starting at the first file */
#define SRCH_NO_REGEX   (1 << 12) /* Don't use regular expressions */

#define SRCH_REVERSE(t) (((t) & SRCH_FORW) ? \
				(((t) & ~SRCH_FORW) | SRCH_BACK) : \
				(((t) & ~SRCH_BACK) | SRCH_FORW))

/* */
#define NO_MCA          0
#define MCA_DONE        1
#define MCA_MORE        2

#define CC_OK           0       /* Char was accepted & processed */
#define CC_QUIT         1       /* Char was a request to abort current cmd */
#define CC_ERROR        2       /* Char could not be accepted due to error */
#define CC_PASS         3       /* Char was rejected (internal) */

#define CF_QUIT_ON_ERASE 0001   /* Abort cmd if its entirely erased */

/* Special char bit-flags used to tell put_line() to do something special */
#define AT_NORMAL       (0)
#define AT_UNDERLINE    (1 << 0)
#define AT_BOLD         (1 << 1)
#define AT_BLINK        (1 << 2)
#define AT_STANDOUT     (1 << 3)
#define AT_ANSI         (1 << 4)  /* Content-supplied "ANSI" escape sequence */
#define AT_BINARY       (1 << 5)  /* LESS*BINFMT representation */
#define AT_HILITE       (1 << 6)  /* Internal highlights (e.g., for search) */

#define CONTROL(c)      ((c)&037)

#define ESC             CONTROL('[')
#define CSI             ((unsigned char)'\233')

#define S_INTERRUPT     01
#define S_STOP          02
#define S_WINCH         04
#define ABORT_SIGS()    (sigs & (S_INTERRUPT|S_STOP))

#define QUIT_OK         0
#define QUIT_ERROR      1
#define QUIT_SAVED_STATUS (-1)

#define FOLLOW_DESC     0
#define FOLLOW_NAME     1

/* filestate flags */
#define CH_CANSEEK      001
#define CH_KEEPOPEN     002
#define CH_POPENED      004
#define CH_HELPFILE     010

#define ch_zero()       ((POSITION)0)

#define FAKE_HELPFILE   "@/\\less/\\help/\\file/\\@"

#if HILITE_SEARCH
struct hilite
{
	struct hilite *hl_next;
	POSITION hl_startpos;
	POSITION hl_endpos;
};
#define hl_first        hl_next
#endif

/* Low level character input from the input file.
 * We use these special purpose routines which optimize moving
 * both forward and backward from the current read pointer. */

typedef POSITION BLOCKNUM;

#define LBUFSIZE        8192

/* Pool of buffers holding the most recently used blocks of the input file.
 * The buffer pool is kept as a doubly-linked circular list,
 * in order from most- to least-recently used.
 * The circular list is anchored by the file state "thisfile". */
struct buf {
	struct buf *next, *prev;
	struct buf *hnext, *hprev;
	BLOCKNUM block;
	unsigned int datasize;
	unsigned char data[LBUFSIZE];
};

struct buflist {
	/* -- Following members must match struct buf */
	struct buf *buf_next, *buf_prev;
	struct buf *buf_hnext, *buf_hprev;
};

#define BUFHASH_SIZE    64
struct filestate {
	struct buf *buf_next, *buf_prev;
	struct buflist hashtbl[BUFHASH_SIZE];
	int file;
	int flags;
	POSITION fpos;
	int nbufs;
	BLOCKNUM block;
	unsigned int offset;
	POSITION fsize;
};

#define ch_bufhead      thisfile->buf_next
#define ch_buftail      thisfile->buf_prev
#define ch_nbufs        thisfile->nbufs
#define ch_block        thisfile->block
#define ch_offset       thisfile->offset
#define ch_fpos         thisfile->fpos
#define ch_fsize        thisfile->fsize
#define ch_flags        thisfile->flags
#define ch_file         thisfile->file

#define END_OF_CHAIN    ((struct buf *)&thisfile->buf_next)
#define END_OF_HCHAIN(h) ((struct buf *)&thisfile->hashtbl[h])
#define BUFHASH(blk)    ((blk) & (BUFHASH_SIZE-1))

#define FOR_BUFS_IN_CHAIN(h,bp) \
	for (bp = thisfile->hashtbl[h].buf_hnext;  \
	     bp != END_OF_HCHAIN(h);  bp = bp->hnext)

#define HASH_RM(bp) \
	(bp)->hnext->hprev = (bp)->hprev; \
	(bp)->hprev->hnext = (bp)->hnext;

#define HASH_INS(bp,h) \
	(bp)->hnext = thisfile->hashtbl[h].buf_hnext; \
	(bp)->hprev = END_OF_HCHAIN(h); \
	thisfile->hashtbl[h].buf_hnext->hprev = (bp); \
	thisfile->hashtbl[h].buf_hnext = (bp);

/* A mark is an ifile (input file) plus a position within the file. */
struct mark {
	IFILE m_ifile;
	struct scrpos m_scrpos;
};

/* The table of marks.
 * Each mark is identified by a lowercase or uppercase letter.
 * The final one is lmark, for the "last mark"; addressed by the apostrophe. */
#define NMARKS          ((2*26)+1)      /* a-z, A-Z, lastmark */
#define LASTMARK        (NMARKS-1)


/* Structure to keep track of a line number and the associated file position.
 * A doubly-linked circular list of line numbers is kept ordered by line number. */
struct linenum_info
{
	struct linenum_info *next;      /* Link to next in the list */
	struct linenum_info *prev;      /* Line to previous in the list */
	POSITION pos;                   /* File position */
	POSITION gap;                   /* Gap between prev and next */
	LINENUM line;                   /* Line number */
};

/* "gap" needs some explanation: the gap of any particular line number
 * is the distance between the previous one and the next one in the list.
 * ("Distance" means difference in file position.)  In other words, the
 * gap of a line number is the gap which would be introduced if this
 * line number were deleted.  It is used to decide which one to replace
 * when we have a new one to insert and the table is full. */

#define NPOOL   50                      /* Size of line number pool */

#define LONGTIME        (2)             /* In seconds */


/* This file was generated by mkhelp from less.hlp */
static const char helpdata[] = {
'\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','S','\b','S','U','\b','U','M','\b','M','M','\b','M','A','\b','A','R','\b','R','Y','\b','Y',' ','O','\b','O','F','\b','F',' ','L','\b','L','E','\b','E','S','\b','S','S','\b','S',' ','C','\b','C','O','\b','O','M','\b','M','M','\b','M','A','\b','A','N','\b','N','D','\b','D','S','\b','S','\n',
'\n',
' ',' ',' ',' ',' ',' ','C','o','m','m','a','n','d','s',' ','m','a','r','k','e','d',' ','w','i','t','h',' ','*',' ','m','a','y',' ','b','e',' ','p','r','e','c','e','d','e','d',' ','b','y',' ','a',' ','n','u','m','b','e','r',',',' ','_','\b','N','.','\n',
' ',' ',' ',' ',' ',' ','N','o','t','e','s',' ','i','n',' ','p','a','r','e','n','t','h','e','s','e','s',' ','i','n','d','i','c','a','t','e',' ','t','h','e',' ','b','e','h','a','v','i','o','r',' ','i','f',' ','_','\b','N',' ','i','s',' ','g','i','v','e','n','.','\n',
'\n',
' ',' ','h',' ',' ','H',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','i','s','p','l','a','y',' ','t','h','i','s',' ','h','e','l','p','.','\n',
' ',' ','q',' ',' ',':','q',' ',' ','Q',' ',' ',':','Q',' ',' ','Z','Z',' ',' ',' ',' ',' ','E','x','i','t','.','\n',
' ','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','\n',
'\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','M','\b','M','O','\b','O','V','\b','V','I','\b','I','N','\b','N','G','\b','G','\n',
'\n',
' ',' ','e',' ',' ','^','E',' ',' ','j',' ',' ','^','N',' ',' ','C','R',' ',' ','*',' ',' ','F','o','r','w','a','r','d',' ',' ','o','n','e',' ','l','i','n','e',' ',' ',' ','(','o','r',' ','_','\b','N',' ','l','i','n','e','s',')','.','\n',
' ',' ','y',' ',' ','^','Y',' ',' ','k',' ',' ','^','K',' ',' ','^','P',' ',' ','*',' ',' ','B','a','c','k','w','a','r','d',' ','o','n','e',' ','l','i','n','e',' ',' ',' ','(','o','r',' ','_','\b','N',' ','l','i','n','e','s',')','.','\n',
' ',' ','f',' ',' ','^','F',' ',' ','^','V',' ',' ','S','P','A','C','E',' ',' ','*',' ',' ','F','o','r','w','a','r','d',' ',' ','o','n','e',' ','w','i','n','d','o','w',' ','(','o','r',' ','_','\b','N',' ','l','i','n','e','s',')','.','\n',
' ',' ','b',' ',' ','^','B',' ',' ','E','S','C','-','v',' ',' ',' ',' ',' ',' ','*',' ',' ','B','a','c','k','w','a','r','d',' ','o','n','e',' ','w','i','n','d','o','w',' ','(','o','r',' ','_','\b','N',' ','l','i','n','e','s',')','.','\n',
' ',' ','z',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','F','o','r','w','a','r','d',' ',' ','o','n','e',' ','w','i','n','d','o','w',' ','(','a','n','d',' ','s','e','t',' ','w','i','n','d','o','w',' ','t','o',' ','_','\b','N',')','.','\n',
' ',' ','w',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','B','a','c','k','w','a','r','d',' ','o','n','e',' ','w','i','n','d','o','w',' ','(','a','n','d',' ','s','e','t',' ','w','i','n','d','o','w',' ','t','o',' ','_','\b','N',')','.','\n',
' ',' ','E','S','C','-','S','P','A','C','E',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','F','o','r','w','a','r','d',' ',' ','o','n','e',' ','w','i','n','d','o','w',',',' ','b','u','t',' ','d','o','n','\'','t',' ','s','t','o','p',' ','a','t',' ','e','n','d','-','o','f','-','f','i','l','e','.','\n',
' ',' ','d',' ',' ','^','D',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','F','o','r','w','a','r','d',' ',' ','o','n','e',' ','h','a','l','f','-','w','i','n','d','o','w',' ','(','a','n','d',' ','s','e','t',' ','h','a','l','f','-','w','i','n','d','o','w',' ','t','o',' ','_','\b','N',')','.','\n',
' ',' ','u',' ',' ','^','U',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','B','a','c','k','w','a','r','d',' ','o','n','e',' ','h','a','l','f','-','w','i','n','d','o','w',' ','(','a','n','d',' ','s','e','t',' ','h','a','l','f','-','w','i','n','d','o','w',' ','t','o',' ','_','\b','N',')','.','\n',
' ',' ','E','S','C','-',')',' ',' ','R','i','g','h','t','A','r','r','o','w',' ','*',' ',' ','L','e','f','t',' ',' ','o','n','e',' ','h','a','l','f',' ','s','c','r','e','e','n',' ','w','i','d','t','h',' ','(','o','r',' ','_','\b','N',' ','p','o','s','i','t','i','o','n','s',')','.','\n',
' ',' ','E','S','C','-','(',' ',' ','L','e','f','t','A','r','r','o','w',' ',' ','*',' ',' ','R','i','g','h','t',' ','o','n','e',' ','h','a','l','f',' ','s','c','r','e','e','n',' ','w','i','d','t','h',' ','(','o','r',' ','_','\b','N',' ','p','o','s','i','t','i','o','n','s',')','.','\n',
' ',' ','F',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','F','o','r','w','a','r','d',' ','f','o','r','e','v','e','r',';',' ','l','i','k','e',' ','"','t','a','i','l',' ','-','f','"','.','\n',
' ',' ','r',' ',' ','^','R',' ',' ','^','L',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','R','e','p','a','i','n','t',' ','s','c','r','e','e','n','.','\n',
' ',' ','R',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','R','e','p','a','i','n','t',' ','s','c','r','e','e','n',',',' ','d','i','s','c','a','r','d','i','n','g',' ','b','u','f','f','e','r','e','d',' ','i','n','p','u','t','.','\n',
' ',' ',' ',' ',' ',' ',' ',' ','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','\n',
' ',' ',' ',' ',' ',' ',' ',' ','D','e','f','a','u','l','t',' ','"','w','i','n','d','o','w','"',' ','i','s',' ','t','h','e',' ','s','c','r','e','e','n',' ','h','e','i','g','h','t','.','\n',
' ',' ',' ',' ',' ',' ',' ',' ','D','e','f','a','u','l','t',' ','"','h','a','l','f','-','w','i','n','d','o','w','"',' ','i','s',' ','h','a','l','f',' ','o','f',' ','t','h','e',' ','s','c','r','e','e','n',' ','h','e','i','g','h','t','.','\n',
' ','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','\n',
'\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','S','\b','S','E','\b','E','A','\b','A','R','\b','R','C','\b','C','H','\b','H','I','\b','I','N','\b','N','G','\b','G','\n',
'\n',
' ',' ','/','_','\b','p','_','\b','a','_','\b','t','_','\b','t','_','\b','e','_','\b','r','_','\b','n',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','S','e','a','r','c','h',' ','f','o','r','w','a','r','d',' ','f','o','r',' ','(','_','\b','N','-','t','h',')',' ','m','a','t','c','h','i','n','g',' ','l','i','n','e','.','\n',
' ',' ','?','_','\b','p','_','\b','a','_','\b','t','_','\b','t','_','\b','e','_','\b','r','_','\b','n',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','S','e','a','r','c','h',' ','b','a','c','k','w','a','r','d',' ','f','o','r',' ','(','_','\b','N','-','t','h',')',' ','m','a','t','c','h','i','n','g',' ','l','i','n','e','.','\n',
' ',' ','n',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','R','e','p','e','a','t',' ','p','r','e','v','i','o','u','s',' ','s','e','a','r','c','h',' ','(','f','o','r',' ','_','\b','N','-','t','h',' ','o','c','c','u','r','r','e','n','c','e',')','.','\n',
' ',' ','N',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','R','e','p','e','a','t',' ','p','r','e','v','i','o','u','s',' ','s','e','a','r','c','h',' ','i','n',' ','r','e','v','e','r','s','e',' ','d','i','r','e','c','t','i','o','n','.','\n',
' ',' ','E','S','C','-','n',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','R','e','p','e','a','t',' ','p','r','e','v','i','o','u','s',' ','s','e','a','r','c','h',',',' ','s','p','a','n','n','i','n','g',' ','f','i','l','e','s','.','\n',
' ',' ','E','S','C','-','N',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','R','e','p','e','a','t',' ','p','r','e','v','i','o','u','s',' ','s','e','a','r','c','h',',',' ','r','e','v','e','r','s','e',' ','d','i','r','.',' ','&',' ','s','p','a','n','n','i','n','g',' ','f','i','l','e','s','.','\n',
' ',' ','E','S','C','-','u',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','U','n','d','o',' ','(','t','o','g','g','l','e',')',' ','s','e','a','r','c','h',' ','h','i','g','h','l','i','g','h','t','i','n','g','.','\n',
' ',' ',' ',' ',' ',' ',' ',' ','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','\n',
' ',' ',' ',' ',' ',' ',' ',' ','S','e','a','r','c','h',' ','p','a','t','t','e','r','n','s',' ','m','a','y',' ','b','e',' ','m','o','d','i','f','i','e','d',' ','b','y',' ','o','n','e',' ','o','r',' ','m','o','r','e',' ','o','f',':','\n',
' ',' ',' ',' ',' ',' ',' ',' ','^','N',' ','o','r',' ','!',' ',' ','S','e','a','r','c','h',' ','f','o','r',' ','N','O','N','-','m','a','t','c','h','i','n','g',' ','l','i','n','e','s','.','\n',
' ',' ',' ',' ',' ',' ',' ',' ','^','E',' ','o','r',' ','*',' ',' ','S','e','a','r','c','h',' ','m','u','l','t','i','p','l','e',' ','f','i','l','e','s',' ','(','p','a','s','s',' ','t','h','r','u',' ','E','N','D',' ','O','F',' ','F','I','L','E',')','.','\n',
' ',' ',' ',' ',' ',' ',' ',' ','^','F',' ','o','r',' ','@',' ',' ','S','t','a','r','t',' ','s','e','a','r','c','h',' ','a','t',' ','F','I','R','S','T',' ','f','i','l','e',' ','(','f','o','r',' ','/',')',' ','o','r',' ','l','a','s','t',' ','f','i','l','e',' ','(','f','o','r',' ','?',')','.','\n',
' ',' ',' ',' ',' ',' ',' ',' ','^','K',' ',' ',' ',' ',' ',' ',' ','H','i','g','h','l','i','g','h','t',' ','m','a','t','c','h','e','s',',',' ','b','u','t',' ','d','o','n','\'','t',' ','m','o','v','e',' ','(','K','E','E','P',' ','p','o','s','i','t','i','o','n',')','.','\n',
' ',' ',' ',' ',' ',' ',' ',' ','^','R',' ',' ',' ',' ',' ',' ',' ','D','o','n','\'','t',' ','u','s','e',' ','R','E','G','U','L','A','R',' ','E','X','P','R','E','S','S','I','O','N','S','.','\n',
' ','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','\n',
'\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','J','\b','J','U','\b','U','M','\b','M','P','\b','P','I','\b','I','N','\b','N','G','\b','G','\n',
'\n',
' ',' ','g',' ',' ','<',' ',' ','E','S','C','-','<',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','G','o',' ','t','o',' ','f','i','r','s','t',' ','l','i','n','e',' ','i','n',' ','f','i','l','e',' ','(','o','r',' ','l','i','n','e',' ','_','\b','N',')','.','\n',
' ',' ','G',' ',' ','>',' ',' ','E','S','C','-','>',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','G','o',' ','t','o',' ','l','a','s','t',' ','l','i','n','e',' ','i','n',' ','f','i','l','e',' ','(','o','r',' ','l','i','n','e',' ','_','\b','N',')','.','\n',
' ',' ','p',' ',' ','%',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','G','o',' ','t','o',' ','b','e','g','i','n','n','i','n','g',' ','o','f',' ','f','i','l','e',' ','(','o','r',' ','_','\b','N',' ','p','e','r','c','e','n','t',' ','i','n','t','o',' ','f','i','l','e',')','.','\n',
' ',' ','t',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','G','o',' ','t','o',' ','t','h','e',' ','(','_','\b','N','-','t','h',')',' ','n','e','x','t',' ','t','a','g','.','\n',
' ',' ','T',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','G','o',' ','t','o',' ','t','h','e',' ','(','_','\b','N','-','t','h',')',' ','p','r','e','v','i','o','u','s',' ','t','a','g','.','\n',
' ',' ','{',' ',' ','(',' ',' ','[',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','F','i','n','d',' ','c','l','o','s','e',' ','b','r','a','c','k','e','t',' ','}',' ',')',' ',']','.','\n',
' ',' ','}',' ',' ',')',' ',' ',']',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','F','i','n','d',' ','o','p','e','n',' ','b','r','a','c','k','e','t',' ','{',' ','(',' ','[','.','\n',
' ',' ','E','S','C','-','^','F',' ','_','\b','<','_','\b','c','_','\b','1','_','\b','>',' ','_','\b','<','_','\b','c','_','\b','2','_','\b','>',' ',' ','*',' ',' ','F','i','n','d',' ','c','l','o','s','e',' ','b','r','a','c','k','e','t',' ','_','\b','<','_','\b','c','_','\b','2','_','\b','>','.','\n',
' ',' ','E','S','C','-','^','B',' ','_','\b','<','_','\b','c','_','\b','1','_','\b','>',' ','_','\b','<','_','\b','c','_','\b','2','_','\b','>',' ',' ','*',' ',' ','F','i','n','d',' ','o','p','e','n',' ','b','r','a','c','k','e','t',' ','_','\b','<','_','\b','c','_','\b','1','_','\b','>',' ','\n',
' ',' ',' ',' ',' ',' ',' ',' ','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','\n',
' ',' ',' ',' ',' ',' ',' ',' ','E','a','c','h',' ','"','f','i','n','d',' ','c','l','o','s','e',' ','b','r','a','c','k','e','t','"',' ','c','o','m','m','a','n','d',' ','g','o','e','s',' ','f','o','r','w','a','r','d',' ','t','o',' ','t','h','e',' ','c','l','o','s','e',' ','b','r','a','c','k','e','t',' ','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','m','a','t','c','h','i','n','g',' ','t','h','e',' ','(','_','\b','N','-','t','h',')',' ','o','p','e','n',' ','b','r','a','c','k','e','t',' ','i','n',' ','t','h','e',' ','t','o','p',' ','l','i','n','e','.','\n',
' ',' ',' ',' ',' ',' ',' ',' ','E','a','c','h',' ','"','f','i','n','d',' ','o','p','e','n',' ','b','r','a','c','k','e','t','"',' ','c','o','m','m','a','n','d',' ','g','o','e','s',' ','b','a','c','k','w','a','r','d',' ','t','o',' ','t','h','e',' ','o','p','e','n',' ','b','r','a','c','k','e','t',' ','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','m','a','t','c','h','i','n','g',' ','t','h','e',' ','(','_','\b','N','-','t','h',')',' ','c','l','o','s','e',' ','b','r','a','c','k','e','t',' ','i','n',' ','t','h','e',' ','b','o','t','t','o','m',' ','l','i','n','e','.','\n',
'\n',
' ',' ','m','_','\b','<','_','\b','l','_','\b','e','_','\b','t','_','\b','t','_','\b','e','_','\b','r','_','\b','>',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','M','a','r','k',' ','t','h','e',' ','c','u','r','r','e','n','t',' ','p','o','s','i','t','i','o','n',' ','w','i','t','h',' ','<','l','e','t','t','e','r','>','.','\n',
' ',' ','\'','_','\b','<','_','\b','l','_','\b','e','_','\b','t','_','\b','t','_','\b','e','_','\b','r','_','\b','>',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','G','o',' ','t','o',' ','a',' ','p','r','e','v','i','o','u','s','l','y',' ','m','a','r','k','e','d',' ','p','o','s','i','t','i','o','n','.','\n',
' ',' ','\'','\'',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','G','o',' ','t','o',' ','t','h','e',' ','p','r','e','v','i','o','u','s',' ','p','o','s','i','t','i','o','n','.','\n',
' ',' ','^','X','^','X',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','S','a','m','e',' ','a','s',' ','\'','.','\n',
' ',' ',' ',' ',' ',' ',' ',' ','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','\n',
' ',' ',' ',' ',' ',' ',' ',' ','A',' ','m','a','r','k',' ','i','s',' ','a','n','y',' ','u','p','p','e','r','-','c','a','s','e',' ','o','r',' ','l','o','w','e','r','-','c','a','s','e',' ','l','e','t','t','e','r','.','\n',
' ',' ',' ',' ',' ',' ',' ',' ','C','e','r','t','a','i','n',' ','m','a','r','k','s',' ','a','r','e',' ','p','r','e','d','e','f','i','n','e','d',':','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','^',' ',' ','m','e','a','n','s',' ',' ','b','e','g','i','n','n','i','n','g',' ','o','f',' ','t','h','e',' ','f','i','l','e','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','$',' ',' ','m','e','a','n','s',' ',' ','e','n','d',' ','o','f',' ','t','h','e',' ','f','i','l','e','\n',
' ','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','\n',
'\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','C','\b','C','H','\b','H','A','\b','A','N','\b','N','G','\b','G','I','\b','I','N','\b','N','G','\b','G',' ','F','\b','F','I','\b','I','L','\b','L','E','\b','E','S','\b','S','\n',
'\n',
' ',' ',':','e',' ','[','_','\b','f','_','\b','i','_','\b','l','_','\b','e',']',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','E','x','a','m','i','n','e',' ','a',' ','n','e','w',' ','f','i','l','e','.','\n',
' ',' ','^','X','^','V',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','S','a','m','e',' ','a','s',' ',':','e','.','\n',
' ',' ',':','n',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','E','x','a','m','i','n','e',' ','t','h','e',' ','(','_','\b','N','-','t','h',')',' ','n','e','x','t',' ','f','i','l','e',' ','f','r','o','m',' ','t','h','e',' ','c','o','m','m','a','n','d',' ','l','i','n','e','.','\n',
' ',' ',':','p',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','E','x','a','m','i','n','e',' ','t','h','e',' ','(','_','\b','N','-','t','h',')',' ','p','r','e','v','i','o','u','s',' ','f','i','l','e',' ','f','r','o','m',' ','t','h','e',' ','c','o','m','m','a','n','d',' ','l','i','n','e','.','\n',
' ',' ',':','x',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','*',' ',' ','E','x','a','m','i','n','e',' ','t','h','e',' ','f','i','r','s','t',' ','(','o','r',' ','_','\b','N','-','t','h',')',' ','f','i','l','e',' ','f','r','o','m',' ','t','h','e',' ','c','o','m','m','a','n','d',' ','l','i','n','e','.','\n',
' ',' ',':','d',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','e','l','e','t','e',' ','t','h','e',' ','c','u','r','r','e','n','t',' ','f','i','l','e',' ','f','r','o','m',' ','t','h','e',' ','c','o','m','m','a','n','d',' ','l','i','n','e',' ','l','i','s','t','.','\n',
' ',' ','=',' ',' ','^','G',' ',' ',':','f',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','P','r','i','n','t',' ','c','u','r','r','e','n','t',' ','f','i','l','e',' ','n','a','m','e','.','\n',
' ','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','\n',
'\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','M','\b','M','I','\b','I','S','\b','S','C','\b','C','E','\b','E','L','\b','L','L','\b','L','A','\b','A','N','\b','N','E','\b','E','O','\b','O','U','\b','U','S','\b','S',' ','C','\b','C','O','\b','O','M','\b','M','M','\b','M','A','\b','A','N','\b','N','D','\b','D','S','\b','S','\n',
'\n',
' ',' ','-','_','\b','<','_','\b','f','_','\b','l','_','\b','a','_','\b','g','_','\b','>',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','T','o','g','g','l','e',' ','a',' ','c','o','m','m','a','n','d',' ','l','i','n','e',' ','o','p','t','i','o','n',' ','[','s','e','e',' ','O','P','T','I','O','N','S',' ','b','e','l','o','w',']','.','\n',
' ',' ','-','-','_','\b','<','_','\b','n','_','\b','a','_','\b','m','_','\b','e','_','\b','>',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','T','o','g','g','l','e',' ','a',' ','c','o','m','m','a','n','d',' ','l','i','n','e',' ','o','p','t','i','o','n',',',' ','b','y',' ','n','a','m','e','.','\n',
' ',' ','_','_','\b','<','_','\b','f','_','\b','l','_','\b','a','_','\b','g','_','\b','>',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','i','s','p','l','a','y',' ','t','h','e',' ','s','e','t','t','i','n','g',' ','o','f',' ','a',' ','c','o','m','m','a','n','d',' ','l','i','n','e',' ','o','p','t','i','o','n','.','\n',
' ',' ','_','_','_','\b','<','_','\b','n','_','\b','a','_','\b','m','_','\b','e','_','\b','>',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','i','s','p','l','a','y',' ','t','h','e',' ','s','e','t','t','i','n','g',' ','o','f',' ','a','n',' ','o','p','t','i','o','n',',',' ','b','y',' ','n','a','m','e','.','\n',
' ',' ','+','_','\b','c','_','\b','m','_','\b','d',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','E','x','e','c','u','t','e',' ','t','h','e',' ','l','e','s','s',' ','c','m','d',' ','e','a','c','h',' ','t','i','m','e',' ','a',' ','n','e','w',' ','f','i','l','e',' ','i','s',' ','e','x','a','m','i','n','e','d','.','\n',
'\n',
' ',' ','!','_','\b','c','_','\b','o','_','\b','m','_','\b','m','_','\b','a','_','\b','n','_','\b','d',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','E','x','e','c','u','t','e',' ','t','h','e',' ','s','h','e','l','l',' ','c','o','m','m','a','n','d',' ','w','i','t','h',' ','$','S','H','E','L','L','.','\n',
' ',' ','|','X','\b','X','_','\b','c','_','\b','o','_','\b','m','_','\b','m','_','\b','a','_','\b','n','_','\b','d',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','P','i','p','e',' ','f','i','l','e',' ','b','e','t','w','e','e','n',' ','c','u','r','r','e','n','t',' ','p','o','s',' ','&',' ','m','a','r','k',' ','X','\b','X',' ','t','o',' ','s','h','e','l','l',' ','c','o','m','m','a','n','d','.','\n',
' ',' ','v',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','E','d','i','t',' ','t','h','e',' ','c','u','r','r','e','n','t',' ','f','i','l','e',' ','w','i','t','h',' ','$','V','I','S','U','A','L',' ','o','r',' ','$','E','D','I','T','O','R','.','\n',
' ',' ','V',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','P','r','i','n','t',' ','v','e','r','s','i','o','n',' ','n','u','m','b','e','r',' ','o','f',' ','"','l','e','s','s','"','.','\n',
' ','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','\n',
'\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','O','\b','O','P','\b','P','T','\b','T','I','\b','I','O','\b','O','N','\b','N','S','\b','S','\n',
'\n',
' ',' ',' ',' ',' ',' ',' ',' ','M','o','s','t',' ','o','p','t','i','o','n','s',' ','m','a','y',' ','b','e',' ','c','h','a','n','g','e','d',' ','e','i','t','h','e','r',' ','o','n',' ','t','h','e',' ','c','o','m','m','a','n','d',' ','l','i','n','e',',','\n',
' ',' ',' ',' ',' ',' ',' ',' ','o','r',' ','f','r','o','m',' ','w','i','t','h','i','n',' ','l','e','s','s',' ','b','y',' ','u','s','i','n','g',' ','t','h','e',' ','-',' ','o','r',' ','-','-',' ','c','o','m','m','a','n','d','.','\n',
' ',' ',' ',' ',' ',' ',' ',' ','O','p','t','i','o','n','s',' ','m','a','y',' ','b','e',' ','g','i','v','e','n',' ','i','n',' ','o','n','e',' ','o','f',' ','t','w','o',' ','f','o','r','m','s',':',' ','e','i','t','h','e','r',' ','a',' ','s','i','n','g','l','e','\n',
' ',' ',' ',' ',' ',' ',' ',' ','c','h','a','r','a','c','t','e','r',' ','p','r','e','c','e','d','e','d',' ','b','y',' ','a',' ','-',',',' ','o','r',' ','a',' ','n','a','m','e',' ','p','r','e','c','e','e','d','e','d',' ','b','y',' ','-','-','.','\n',
'\n',
' ',' ','-','?',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','h','e','l','p','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','i','s','p','l','a','y',' ','h','e','l','p',' ','(','f','r','o','m',' ','c','o','m','m','a','n','d',' ','l','i','n','e',')','.','\n',
' ',' ','-','a',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','s','e','a','r','c','h','-','s','k','i','p','-','s','c','r','e','e','n','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','F','o','r','w','a','r','d',' ','s','e','a','r','c','h',' ','s','k','i','p','s',' ','c','u','r','r','e','n','t',' ','s','c','r','e','e','n','.','\n',
' ',' ','-','b',' ','[','_','\b','N',']',' ',' ','.','.','.','.',' ',' ','-','-','b','u','f','f','e','r','s','=','[','_','\b','N',']','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','N','u','m','b','e','r',' ','o','f',' ','b','u','f','f','e','r','s','.','\n',
' ',' ','-','B',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','a','u','t','o','-','b','u','f','f','e','r','s','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','o','n','\'','t',' ','a','u','t','o','m','a','t','i','c','a','l','l','y',' ','a','l','l','o','c','a','t','e',' ','b','u','f','f','e','r','s',' ','f','o','r',' ','p','i','p','e','s','.','\n',
' ',' ','-','c',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','c','l','e','a','r','-','s','c','r','e','e','n','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','R','e','p','a','i','n','t',' ','b','y',' ','c','l','e','a','r','i','n','g',' ','r','a','t','h','e','r',' ','t','h','a','n',' ','s','c','r','o','l','l','i','n','g','.','\n',
' ',' ','-','d',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','d','u','m','b','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','u','m','b',' ','t','e','r','m','i','n','a','l','.','\n',
' ',' ','-','D',' ','[','_','\b','x','_','\b','n','_','\b','.','_','\b','n',']',' ',' ','.',' ',' ','-','-','c','o','l','o','r','=','_','\b','x','_','\b','n','_','\b','.','_','\b','n','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','S','e','t',' ','s','c','r','e','e','n',' ','c','o','l','o','r','s','.',' ','(','M','S','-','D','O','S',' ','o','n','l','y',')','\n',
' ',' ','-','e',' ',' ','-','E',' ',' ','.','.','.','.',' ',' ','-','-','q','u','i','t','-','a','t','-','e','o','f',' ',' ','-','-','Q','U','I','T','-','A','T','-','E','O','F','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','Q','u','i','t',' ','a','t',' ','e','n','d',' ','o','f',' ','f','i','l','e','.','\n',
' ',' ','-','f',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','f','o','r','c','e','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','F','o','r','c','e',' ','o','p','e','n',' ','n','o','n','-','r','e','g','u','l','a','r',' ','f','i','l','e','s','.','\n',
' ',' ','-','F',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','q','u','i','t','-','i','f','-','o','n','e','-','s','c','r','e','e','n','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','Q','u','i','t',' ','i','f',' ','e','n','t','i','r','e',' ','f','i','l','e',' ','f','i','t','s',' ','o','n',' ','f','i','r','s','t',' ','s','c','r','e','e','n','.','\n',
' ',' ','-','g',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','h','i','l','i','t','e','-','s','e','a','r','c','h','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','H','i','g','h','l','i','g','h','t',' ','o','n','l','y',' ','l','a','s','t',' ','m','a','t','c','h',' ','f','o','r',' ','s','e','a','r','c','h','e','s','.','\n',
' ',' ','-','G',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','H','I','L','I','T','E','-','S','E','A','R','C','H','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','o','n','\'','t',' ','h','i','g','h','l','i','g','h','t',' ','a','n','y',' ','m','a','t','c','h','e','s',' ','f','o','r',' ','s','e','a','r','c','h','e','s','.','\n',
' ',' ','-','h',' ','[','_','\b','N',']',' ',' ','.','.','.','.',' ',' ','-','-','m','a','x','-','b','a','c','k','-','s','c','r','o','l','l','=','[','_','\b','N',']','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','B','a','c','k','w','a','r','d',' ','s','c','r','o','l','l',' ','l','i','m','i','t','.','\n',
' ',' ','-','i',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','i','g','n','o','r','e','-','c','a','s','e','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','I','g','n','o','r','e',' ','c','a','s','e',' ','i','n',' ','s','e','a','r','c','h','e','s',' ','t','h','a','t',' ','d','o',' ','n','o','t',' ','c','o','n','t','a','i','n',' ','u','p','p','e','r','c','a','s','e','.','\n',
' ',' ','-','I',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','I','G','N','O','R','E','-','C','A','S','E','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','I','g','n','o','r','e',' ','c','a','s','e',' ','i','n',' ','a','l','l',' ','s','e','a','r','c','h','e','s','.','\n',
' ',' ','-','j',' ','[','_','\b','N',']',' ',' ','.','.','.','.',' ',' ','-','-','j','u','m','p','-','t','a','r','g','e','t','=','[','_','\b','N',']','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','S','c','r','e','e','n',' ','p','o','s','i','t','i','o','n',' ','o','f',' ','t','a','r','g','e','t',' ','l','i','n','e','s','.','\n',
' ',' ','-','J',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','s','t','a','t','u','s','-','c','o','l','u','m','n','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','i','s','p','l','a','y',' ','a',' ','s','t','a','t','u','s',' ','c','o','l','u','m','n',' ','a','t',' ','l','e','f','t',' ','e','d','g','e',' ','o','f',' ','s','c','r','e','e','n','.','\n',
' ',' ','-','k',' ','[','_','\b','f','_','\b','i','_','\b','l','_','\b','e',']',' ',' ','.',' ',' ','-','-','l','e','s','s','k','e','y','-','f','i','l','e','=','[','_','\b','f','_','\b','i','_','\b','l','_','\b','e',']','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','U','s','e',' ','a',' ','l','e','s','s','k','e','y',' ','f','i','l','e','.','\n',
' ',' ','-','L',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','n','o','-','l','e','s','s','o','p','e','n','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','I','g','n','o','r','e',' ','t','h','e',' ','L','E','S','S','O','P','E','N',' ','e','n','v','i','r','o','n','m','e','n','t',' ','v','a','r','i','a','b','l','e','.','\n',
' ',' ','-','m',' ',' ','-','M',' ',' ','.','.','.','.',' ',' ','-','-','l','o','n','g','-','p','r','o','m','p','t',' ',' ','-','-','L','O','N','G','-','P','R','O','M','P','T','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','S','e','t',' ','p','r','o','m','p','t',' ','s','t','y','l','e','.','\n',
' ',' ','-','n',' ',' ','-','N',' ',' ','.','.','.','.',' ',' ','-','-','l','i','n','e','-','n','u','m','b','e','r','s',' ',' ','-','-','L','I','N','E','-','N','U','M','B','E','R','S','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','o','n','\'','t',' ','u','s','e',' ','l','i','n','e',' ','n','u','m','b','e','r','s','.','\n',
' ',' ','-','o',' ','[','_','\b','f','_','\b','i','_','\b','l','_','\b','e',']',' ',' ','.',' ',' ','-','-','l','o','g','-','f','i','l','e','=','[','_','\b','f','_','\b','i','_','\b','l','_','\b','e',']','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','C','o','p','y',' ','t','o',' ','l','o','g',' ','f','i','l','e',' ','(','s','t','a','n','d','a','r','d',' ','i','n','p','u','t',' ','o','n','l','y',')','.','\n',
' ',' ','-','O',' ','[','_','\b','f','_','\b','i','_','\b','l','_','\b','e',']',' ',' ','.',' ',' ','-','-','L','O','G','-','F','I','L','E','=','[','_','\b','f','_','\b','i','_','\b','l','_','\b','e',']','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','C','o','p','y',' ','t','o',' ','l','o','g',' ','f','i','l','e',' ','(','u','n','c','o','n','d','i','t','i','o','n','a','l','l','y',' ','o','v','e','r','w','r','i','t','e',')','.','\n',
' ',' ','-','p',' ','[','_','\b','p','_','\b','a','_','\b','t','_','\b','t','_','\b','e','_','\b','r','_','\b','n',']',' ',' ','-','-','p','a','t','t','e','r','n','=','[','_','\b','p','_','\b','a','_','\b','t','_','\b','t','_','\b','e','_','\b','r','_','\b','n',']','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','S','t','a','r','t',' ','a','t',' ','p','a','t','t','e','r','n',' ','(','f','r','o','m',' ','c','o','m','m','a','n','d',' ','l','i','n','e',')','.','\n',
' ',' ','-','P',' ','[','_','\b','p','_','\b','r','_','\b','o','_','\b','m','_','\b','p','_','\b','t',']',' ',' ',' ','-','-','p','r','o','m','p','t','=','[','_','\b','p','_','\b','r','_','\b','o','_','\b','m','_','\b','p','_','\b','t',']','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','e','f','i','n','e',' ','n','e','w',' ','p','r','o','m','p','t','.','\n',
' ',' ','-','q',' ',' ','-','Q',' ',' ','.','.','.','.',' ',' ','-','-','q','u','i','e','t',' ',' ','-','-','Q','U','I','E','T',' ',' ','-','-','s','i','l','e','n','t',' ','-','-','S','I','L','E','N','T','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','Q','u','i','e','t',' ','t','h','e',' ','t','e','r','m','i','n','a','l',' ','b','e','l','l','.','\n',
' ',' ','-','r',' ',' ','-','R',' ',' ','.','.','.','.',' ',' ','-','-','r','a','w','-','c','o','n','t','r','o','l','-','c','h','a','r','s',' ',' ','-','-','R','A','W','-','C','O','N','T','R','O','L','-','C','H','A','R','S','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','O','u','t','p','u','t',' ','"','r','a','w','"',' ','c','o','n','t','r','o','l',' ','c','h','a','r','a','c','t','e','r','s','.','\n',
' ',' ','-','s',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','s','q','u','e','e','z','e','-','b','l','a','n','k','-','l','i','n','e','s','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','S','q','u','e','e','z','e',' ','m','u','l','t','i','p','l','e',' ','b','l','a','n','k',' ','l','i','n','e','s','.','\n',
' ',' ','-','S',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','c','h','o','p','-','l','o','n','g','-','l','i','n','e','s','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','C','h','o','p',' ','l','o','n','g',' ','l','i','n','e','s','.','\n',
' ',' ','-','t',' ','[','_','\b','t','_','\b','a','_','\b','g',']',' ',' ','.','.',' ',' ','-','-','t','a','g','=','[','_','\b','t','_','\b','a','_','\b','g',']','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','F','i','n','d',' ','a',' ','t','a','g','.','\n',
' ',' ','-','T',' ','[','_','\b','t','_','\b','a','_','\b','g','_','\b','s','_','\b','f','_','\b','i','_','\b','l','_','\b','e',']',' ','-','-','t','a','g','-','f','i','l','e','=','[','_','\b','t','_','\b','a','_','\b','g','_','\b','s','_','\b','f','_','\b','i','_','\b','l','_','\b','e',']','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','U','s','e',' ','a','n',' ','a','l','t','e','r','n','a','t','e',' ','t','a','g','s',' ','f','i','l','e','.','\n',
' ',' ','-','u',' ',' ','-','U',' ',' ','.','.','.','.',' ',' ','-','-','u','n','d','e','r','l','i','n','e','-','s','p','e','c','i','a','l',' ',' ','-','-','U','N','D','E','R','L','I','N','E','-','S','P','E','C','I','A','L','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','C','h','a','n','g','e',' ','h','a','n','d','l','i','n','g',' ','o','f',' ','b','a','c','k','s','p','a','c','e','s','.','\n',
' ',' ','-','V',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','v','e','r','s','i','o','n','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','i','s','p','l','a','y',' ','t','h','e',' ','v','e','r','s','i','o','n',' ','n','u','m','b','e','r',' ','o','f',' ','"','l','e','s','s','"','.','\n',
' ',' ','-','w',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','h','i','l','i','t','e','-','u','n','r','e','a','d','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','H','i','g','h','l','i','g','h','t',' ','f','i','r','s','t',' ','n','e','w',' ','l','i','n','e',' ','a','f','t','e','r',' ','f','o','r','w','a','r','d','-','s','c','r','e','e','n','.','\n',
' ',' ','-','W',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','H','I','L','I','T','E','-','U','N','R','E','A','D','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','H','i','g','h','l','i','g','h','t',' ','f','i','r','s','t',' ','n','e','w',' ','l','i','n','e',' ','a','f','t','e','r',' ','a','n','y',' ','f','o','r','w','a','r','d',' ','m','o','v','e','m','e','n','t','.','\n',
' ',' ','-','x',' ','[','_','\b','N','[',',','.','.','.',']',']',' ',' ','-','-','t','a','b','s','=','[','_','\b','N','[',',','.','.','.',']',']','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','S','e','t',' ','t','a','b',' ','s','t','o','p','s','.','\n',
' ',' ','-','X',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','n','o','-','i','n','i','t','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','o','n','\'','t',' ','u','s','e',' ','t','e','r','m','c','a','p',' ','i','n','i','t','/','d','e','i','n','i','t',' ','s','t','r','i','n','g','s','.','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','-','-','n','o','-','k','e','y','p','a','d','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','o','n','\'','t',' ','u','s','e',' ','t','e','r','m','c','a','p',' ','k','e','y','p','a','d',' ','i','n','i','t','/','d','e','i','n','i','t',' ','s','t','r','i','n','g','s','.','\n',
' ',' ','-','y',' ','[','_','\b','N',']',' ',' ','.','.','.','.',' ',' ','-','-','m','a','x','-','f','o','r','w','-','s','c','r','o','l','l','=','[','_','\b','N',']','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','F','o','r','w','a','r','d',' ','s','c','r','o','l','l',' ','l','i','m','i','t','.','\n',
' ',' ','-','z',' ','[','_','\b','N',']',' ',' ','.','.','.','.',' ',' ','-','-','w','i','n','d','o','w','=','[','_','\b','N',']','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','S','e','t',' ','s','i','z','e',' ','o','f',' ','w','i','n','d','o','w','.','\n',
' ',' ','-','"',' ','[','_','\b','c','[','_','\b','c',']',']',' ',' ','.',' ',' ','-','-','q','u','o','t','e','s','=','[','_','\b','c','[','_','\b','c',']',']','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','S','e','t',' ','s','h','e','l','l',' ','q','u','o','t','e',' ','c','h','a','r','a','c','t','e','r','s','.','\n',
' ',' ','-','~',' ',' ','.','.','.','.','.','.','.','.',' ',' ','-','-','t','i','l','d','e','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','o','n','\'','t',' ','d','i','s','p','l','a','y',' ','t','i','l','d','e','s',' ','a','f','t','e','r',' ','e','n','d',' ','o','f',' ','f','i','l','e','.','\n',
' ',' ','-','#',' ','[','_','\b','N',']',' ',' ','.','.','.','.',' ',' ','-','-','s','h','i','f','t','=','[','_','\b','N',']','\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','H','o','r','i','z','o','n','t','a','l',' ','s','c','r','o','l','l',' ','a','m','o','u','n','t',' ','(','0',' ','=',' ','o','n','e',' ','h','a','l','f',' ','s','c','r','e','e','n',' ','w','i','d','t','h',')','\n',
'\n',
' ','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','-','\n',
'\n',
' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','L','\b','L','I','\b','I','N','\b','N','E','\b','E',' ','E','\b','E','D','\b','D','I','\b','I','T','\b','T','I','\b','I','N','\b','N','G','\b','G','\n',
'\n',
' ',' ',' ',' ',' ',' ',' ',' ','T','h','e','s','e',' ','k','e','y','s',' ','c','a','n',' ','b','e',' ','u','s','e','d',' ','t','o',' ','e','d','i','t',' ','t','e','x','t',' ','b','e','i','n','g',' ','e','n','t','e','r','e','d',' ','\n',
' ',' ',' ',' ',' ',' ',' ',' ','o','n',' ','t','h','e',' ','"','c','o','m','m','a','n','d',' ','l','i','n','e','"',' ','a','t',' ','t','h','e',' ','b','o','t','t','o','m',' ','o','f',' ','t','h','e',' ','s','c','r','e','e','n','.','\n',
'\n',
' ','R','i','g','h','t','A','r','r','o','w',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','E','S','C','-','l',' ',' ',' ',' ',' ','M','o','v','e',' ','c','u','r','s','o','r',' ','r','i','g','h','t',' ','o','n','e',' ','c','h','a','r','a','c','t','e','r','.','\n',
' ','L','e','f','t','A','r','r','o','w',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','E','S','C','-','h',' ',' ',' ',' ',' ','M','o','v','e',' ','c','u','r','s','o','r',' ','l','e','f','t',' ','o','n','e',' ','c','h','a','r','a','c','t','e','r','.','\n',
' ','C','N','T','L','-','R','i','g','h','t','A','r','r','o','w',' ',' ','E','S','C','-','R','i','g','h','t','A','r','r','o','w',' ',' ','E','S','C','-','w',' ',' ',' ',' ',' ','M','o','v','e',' ','c','u','r','s','o','r',' ','r','i','g','h','t',' ','o','n','e',' ','w','o','r','d','.','\n',
' ','C','N','T','L','-','L','e','f','t','A','r','r','o','w',' ',' ',' ','E','S','C','-','L','e','f','t','A','r','r','o','w',' ',' ',' ','E','S','C','-','b',' ',' ',' ',' ',' ','M','o','v','e',' ','c','u','r','s','o','r',' ','l','e','f','t',' ','o','n','e',' ','w','o','r','d','.','\n',
' ','H','O','M','E',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','E','S','C','-','0',' ',' ',' ',' ',' ','M','o','v','e',' ','c','u','r','s','o','r',' ','t','o',' ','s','t','a','r','t',' ','o','f',' ','l','i','n','e','.','\n',
' ','E','N','D',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','E','S','C','-','$',' ',' ',' ',' ',' ','M','o','v','e',' ','c','u','r','s','o','r',' ','t','o',' ','e','n','d',' ','o','f',' ','l','i','n','e','.','\n',
' ','B','A','C','K','S','P','A','C','E',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','e','l','e','t','e',' ','c','h','a','r',' ','t','o',' ','l','e','f','t',' ','o','f',' ','c','u','r','s','o','r','.','\n',
' ','D','E','L','E','T','E',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','E','S','C','-','x',' ',' ',' ',' ',' ','D','e','l','e','t','e',' ','c','h','a','r',' ','u','n','d','e','r',' ','c','u','r','s','o','r','.','\n',
' ','C','N','T','L','-','B','A','C','K','S','P','A','C','E',' ',' ',' ','E','S','C','-','B','A','C','K','S','P','A','C','E',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','e','l','e','t','e',' ','w','o','r','d',' ','t','o',' ','l','e','f','t',' ','o','f',' ','c','u','r','s','o','r','.','\n',
' ','C','N','T','L','-','D','E','L','E','T','E',' ',' ',' ',' ',' ',' ','E','S','C','-','D','E','L','E','T','E',' ',' ',' ',' ',' ',' ','E','S','C','-','X',' ',' ',' ',' ',' ','D','e','l','e','t','e',' ','w','o','r','d',' ','u','n','d','e','r',' ','c','u','r','s','o','r','.','\n',
' ','C','N','T','L','-','U',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','E','S','C',' ','(','M','S','-','D','O','S',' ','o','n','l','y',')',' ',' ',' ',' ',' ',' ',' ',' ',' ','D','e','l','e','t','e',' ','e','n','t','i','r','e',' ','l','i','n','e','.','\n',
' ','U','p','A','r','r','o','w',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','E','S','C','-','k',' ',' ',' ',' ',' ','R','e','t','r','i','e','v','e',' ','p','r','e','v','i','o','u','s',' ','c','o','m','m','a','n','d',' ','l','i','n','e','.','\n',
' ','D','o','w','n','A','r','r','o','w',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','E','S','C','-','j',' ',' ',' ',' ',' ','R','e','t','r','i','e','v','e',' ','n','e','x','t',' ','c','o','m','m','a','n','d',' ','l','i','n','e','.','\n',
' ','T','A','B',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','C','o','m','p','l','e','t','e',' ','f','i','l','e','n','a','m','e',' ','&',' ','c','y','c','l','e','.','\n',
' ','S','H','I','F','T','-','T','A','B',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','E','S','C','-','T','A','B',' ',' ',' ','C','o','m','p','l','e','t','e',' ','f','i','l','e','n','a','m','e',' ','&',' ','r','e','v','e','r','s','e',' ','c','y','c','l','e','.','\n',
' ','C','N','T','L','-','L',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','C','o','m','p','l','e','t','e',' ','f','i','l','e','n','a','m','e',',',' ','l','i','s','t',' ','a','l','l','.','\n',
'\n',
'\n',
 0 };
static const int size_helpdata = sizeof(helpdata) - 1;

/* Prototypes for the three flavors of prompts.
 * These strings are expanded by pr_expand(). */
static const char s_proto[] =
  "?n?f%f .?m(%T %i of %m) ..?e(END) ?x- Next\\: %x..%t";
static const char m_proto[] =
  "?n?f%f .?m(%T %i of %m) ..?e(END) ?x- Next\\: %x.:?pB%pB\\%:byte %bB?s/%s...%t";
static const char M_proto[] =
  "?f%f .?n?m(%T %i of %m) ..?ltlines %lt-%lb?L/%L. :byte %bB?s/%s. .?e(END) ?x- Next\\: %x.:?pB%pB\\%..%t";
static const char e_proto[] =
  "?f%f .?m(%T %i of %m) .?ltlines %lt-%lb?L/%L. .byte %bB?s/%s. ?e(END) :?pB%pB\\%..%t";
static const char h_proto[] =
  "HELP -- ?eEND -- Press g to see it again:Press RETURN for more., or q when done";
static const char w_proto[] =
  "Waiting for data";
static const char more_proto[] =
  "--More--(?eEND ?x- Next\\: %x.:?pB%pB\\%:byte %bB?s/%s...%t)";


/* Get the character pointed to by the read pointer.
 * ch_get() is a macro which is more efficient to call
 * than fch_get (the function), in the usual case
 * that the block desired is at the head of the chain. */
#define ch_get()   ((ch_block == ch_bufhead->block && \
		     ch_offset < ch_bufhead->datasize) ? \
			ch_bufhead->data[ch_offset] : fch_get())

static void repaint(void);
static int search_range(POSITION pos, POSITION endpos, int search_type_r,
	    int matches, int maxlines, POSITION *plinepos, POSITION *pendpos);
static int cond(char c, int where);
#if NO_REGEX
static int
match_pattern(char *line, int line_len, char **sp, char **ep);
#define MATCH_PATTERN(line, line_len, sp, ep, notbol) match_pattern(line, line_len, sp, ep)
#else
static int
match_pattern(char *line, int line_len, char **sp, char **ep, int notbol);
#define MATCH_PATTERN(line, line_len, sp, ep, notbol) match_pattern(line, line_len, sp, ep, notbol)
#endif
static int iread(int fd, char *buf, unsigned int len);
static int search(char *pattern, int n);

/* Include file for interfacing to position.c modules. */
#define TOP             (0)
#define TOP_PLUS_ONE    (1)
#define BOTTOM          (-1)
#define BOTTOM_PLUS_ONE (-2)
#define MIDDLE          (-3)

#define END_OPTION_STRING       ('$')

/* Types of options. */
#define BOOL            01      /* Boolean option: 0 or 1 */
#define TRIPLE          02      /* Triple-valued option: 0, 1 or 2 */
#define NUMBER          04      /* Numeric option */
#define STRING          010     /* String-valued option */
#define NOVAR           020     /* No associated variable */
#define REPAINT         040     /* Repaint screen after toggling option */
#define NO_TOGGLE       0100    /* Option cannot be toggled with "-" cmd */
#define HL_REPAINT      0200    /* Repaint hilites after toggling option */
#define NO_QUERY        0400    /* Option cannot be queried with "_" cmd */
#define INIT_HANDLER    01000   /* Call option handler function at startup */

#define OTYPE           (BOOL|TRIPLE|NUMBER|STRING|NOVAR)

/* Argument to a handling function tells what type of activity: */
#define INIT    0       /* Initialization (from command line) */
#define QUERY   1       /* Query (from _ or - command) */
#define TOGGLE  2       /* Change value (from - command) */

/* Flag to toggle_option to specify how to "toggle" */
#define OPT_NO_TOGGLE   0
#define OPT_TOGGLE      1
#define OPT_UNSET       2
#define OPT_SET         3
#define OPT_NO_PROMPT   0100

/* Error code from findopt_name */
#define OPT_AMBIG       1

struct loption
{
	char oletter;           /* The controlling letter (a-z) */
	const char *onames;     /* Long (GNU-style) option name */
	int otype;              /* Type of the option */
	int odefault;           /* Default value */
	int *ovar;              /* Pointer to the associated variable */
	/* Pointer to special handling function */
	void (*ofunc)(int type, const char *s);
	char *odesc[3];         /* Description of each value */
};

#define MAX_USERCMD             500
#define MAX_CMDLEN              16

#define A_B_LINE                2
#define A_B_SCREEN              3
#define A_B_SCROLL              4
#define A_B_SEARCH              5
#define A_DIGIT                 6
#define A_DISP_OPTION           7
#define A_DEBUG                 8
#define A_EXAMINE               9
#define A_FIRSTCMD              10
#define A_FREPAINT              11
#define A_F_LINE                12
#define A_F_SCREEN              13
#define A_F_SCROLL              14
#define A_F_SEARCH              15
#define A_GOEND                 16
#define A_GOLINE                17
#define A_GOMARK                18
#define A_HELP                  19
#define A_NEXT_FILE             20
#define A_PERCENT               21
#define A_PREFIX                22
#define A_PREV_FILE             23
#define A_QUIT                  24
#define A_REPAINT               25
#define A_SETMARK               26
#define A_SHELL                 27
#define A_STAT                  28
#define A_FF_LINE               29
#define A_BF_LINE               30
#define A_VERSION               31
#define A_VISUAL                32
#define A_F_WINDOW              33
#define A_B_WINDOW              34
#define A_F_BRACKET             35
#define A_B_BRACKET             36
#define A_PIPE                  37
#define A_INDEX_FILE            38
#define A_UNDO_SEARCH           39
#define A_FF_SCREEN             40
#define A_LSHIFT                41
#define A_RSHIFT                42
#define A_AGAIN_SEARCH          43
#define A_T_AGAIN_SEARCH        44
#define A_REVERSE_SEARCH        45
#define A_T_REVERSE_SEARCH      46
#define A_OPT_TOGGLE            47
#define A_OPT_SET               48
#define A_OPT_UNSET             49
#define A_F_FOREVER             50
#define A_GOPOS                 51
#define A_REMOVE_FILE           52
#define A_NEXT_TAG              53
#define A_PREV_TAG              54

#define A_INVALID               100
#define A_NOACTION              101
#define A_UINVALID              102
#define A_END_LIST              103
#define A_SPECIAL_KEY           104

#define A_SKIP                  127

#define A_EXTRA                 0200

/* Line editting characters */
#define EC_BACKSPACE    1
#define EC_LINEKILL     2
#define EC_RIGHT        3
#define EC_LEFT         4
#define EC_W_LEFT       5
#define EC_W_RIGHT      6
#define EC_INSERT       7
#define EC_DELETE       8
#define EC_HOME         9
#define EC_END          10
#define EC_W_BACKSPACE  11
#define EC_W_DELETE     12
#define EC_UP           13
#define EC_DOWN         14
#define EC_EXPAND       15
#define EC_F_COMPLETE   17
#define EC_B_COMPLETE   18
#define EC_LITERAL      19

#define EC_NOACTION     101
#define EC_UINVALID     102

/* Flags for editchar() */
#define EC_PEEK         01
#define EC_NOHISTORY    02
#define EC_NOCOMPLETE   04
#define EC_NORIGHTLEFT  010

/* Environment variable stuff */
#define EV_OK           01

/* Special keys (keys which output different strings on different terminals) */
#define SK_SPECIAL_KEY          CONTROL('K')
#define SK_RIGHT_ARROW          1
#define SK_LEFT_ARROW           2
#define SK_UP_ARROW             3
#define SK_DOWN_ARROW           4
#define SK_PAGE_UP              5
#define SK_PAGE_DOWN            6
#define SK_HOME                 7
#define SK_END                  8
#define SK_DELETE               9
#define SK_INSERT               10
#define SK_CTL_LEFT_ARROW       11
#define SK_CTL_RIGHT_ARROW      12
#define SK_CTL_DELETE           13
#define SK_F1                   14
#define SK_BACKTAB              15
#define SK_CTL_BACKSPACE        16
#define SK_CONTROL_K            40

/* charset.h */
#define IS_ASCII_OCTET(c)   (((c) & 0x80) == 0)
#define IS_UTF8_TRAIL(c)    (((c) & 0xC0) == 0x80)
#define IS_UTF8_INVALID(c)  (((c) & 0xFE) == 0xFE)
#define IS_UTF8_LEAD(c)     (((c) & 0xC0) == 0xC0 && !IS_UTF8_INVALID(c))

#if HILITE_SEARCH
static int do_append(LWCHAR ch, char *rep, POSITION pos);
#define DO_APPEND(ch, rep, pos) do_append((ch), (rep), (pos))
#else
static int do_append(LWCHAR ch, char *rep);
#define DO_APPEND(ch, rep, pos) do_append((ch), (rep))
#endif


/* An IFILE represents an input file.
 *
 * It is actually a pointer to an ifile structure,
 * but is opaque outside this module.
 * Ifile structures are kept in a linked list in the order they
 * appear on the command line.
 * Any new file which does not already appear in the list is
 * inserted after the current file. */
struct ifile {
	struct ifile *h_next;           /* Links for command line list */
	struct ifile *h_prev;
	char *h_filename;               /* Name of the file */
	void *h_filestate;              /* File state (used in ch.c) */
	int h_index;                    /* Index within command line list */
	int h_hold;                     /* Hold count */
	char h_opened;                  /* Has this ifile been opened? */
	struct scrpos h_scrpos;         /* Saved position within the file */
};

/* Convert an IFILE (external representation)
 * to a struct file (internal representation), and vice versa. */
#define int_ifile(h)    ((struct ifile *)(h))
#define ext_ifile(h)    ((IFILE)(h))

#define auto_wrap 1             /* Terminal does \r\n when write past margin */
static const char version[] = VERSION;

#if LOGFILE
static int      logfile = -1;
static int      force_logfile = FALSE;
static char *   namelogfile;
#endif

#if TAGS
static char *tags = "tags";
#endif

/* A mlist structure represents a command history. */
struct mlist
{
	struct mlist *next;
	struct mlist *prev;
	struct mlist *curr_mp;
	char *string;
	int modified;
};

CALC_BSS struct BB_COMMON_BSS
{
	int any_display;
	int plusoption;
	int fd0;
	dev_t curr_dev;
	ino_t curr_ino;
	char *curr_altfilename;
	void *curr_altpipe;
	int is_ucase_pattern;
	int is_caseless;
	jmp_buf read_label;
	/* These are the static variables that represent the "remembered"
	 * search pattern. */
	int last_search_type;
	char *last_pattern;
#ifndef NO_REGEX
	regex_t *regpattern;
#endif
	char *prproto[3];
	char *eqproto;
	char *hproto;
	char *wproto;

	int ifiles;
	/* "sigs" contains bits indicating signals which need to be processed. */
	int sigs;
	int bs_mode;             /* How to process backspaces */
	int linenums;            /* Use line numbers */
	int ctldisp;             /* Send control chars to screen untranslated */
	int twiddle;             /* Show tildes after EOF */
	int status_col;          /* Display a status column */

	/* make minimal harcoded vt100-like TERM by dzo@ */
	int ignaw;               /* Terminal ignores \n immediately after wrap */
	int erase_char;          /* The user's erase char */
	int erase2_char;         /* The user's other erase char */
	int kill_char;           /* The user's line-kill char */
	int werase_char;         /* The user's word-erase char */
	int sc_width, sc_height; /* Height & width of screen */
	char *   every_first_cmd;
	int      new_file;
	int      is_tty;
	IFILE    curr_ifile;
	IFILE    old_ifile;
	struct scrpos initial_scrpos;
	int      wscroll;
	int      quitting;
	int      secure;
	int      dohelp;
	int less_is_more;        /* Make compatible with POSIX more */

#if SUPPORT_UTF8
	int utf_mode;
	char mbc_buf[MAX_UTF_CHAR_LEN];
	int mbc_buf_len;
	int mbc_buf_index;
	POSITION mbc_pos;
#endif

	int hit_eof;         /* Keeps track of how many times we hit end of file */
	int screen_trashed;  /* The screen has been overwritten */
	int squished;
	int forw_prompt;

#if TAGS
	char *tagoption;
#endif

	int ignore_eoi;

	/* Variables controlled by command line options. */
	int quiet;               /* Should we suppress the audible bell? */
				/* If VERY_QUIET, use visual bell for bell */
	int how_search;          /* Where should forward searches start? */
	int top_scroll;          /* Repaint screen from top?
				   (alternative is scroll from bottom) */
	int pr_type;             /* Type of prompt (short, medium, long) */
	int quit_at_eof;         /* Quit after hitting end of file twice */
	int quit_if_one_screen;  /* Quit if EOF on first screen */
	int squeeze;             /* Squeeze multiple blank lines into one */
	int back_scroll;         /* Repaint screen on backwards movement */
	int forw_scroll;         /* Repaint screen on forward movement */
	int caseless;            /* Do "caseless" searches */
	int autobuf;             /* Automatically allocate buffers as needed */
	int bufspace;            /* Max buffer space per file (K) */
	int force_open;          /* Open the file even if not regular file */
	int swindow;             /* Size of scrolling window */
	int chopline;            /* Truncate displayed lines at screen width */
	int show_attn;           /* Hilite first unread line */
	int shift_count;         /* Number of positions to shift horizontally */
	int use_lessopen;        /* Use the LESSOPEN filter */
	int quit_on_intr;        /* Quit on interrupt */
	int follow_mode;         /* F cmd Follows file desc or file name? */
#if HILITE_SEARCH
	int hilite_search;       /* Highlight matched search patterns? */
#endif
#if EDITOR
	const char *   editor;
	const char *   editproto;
#endif

	const struct loption *pendopt;
	int need_clr;
	char obuf[OUTBUF_SIZE];
	char *ob;
	int at_prompt;
	int attrmode; /* = AT_NORMAL */
	int tty;
	/* Routines to manipulate the "line buffer".
	 * The line buffer holds a line of output as it is being built
	 * in preparation for output to the screen. */
	char *linebuf;           /* Buffer which holds the current output line */
	char *attr;              /* Extension of linebuf to hold attributes */
	int size_linebuf;        /* Size of line buffer (and attr buffer) */

	int cshift;              /* Current left-shift of output line buffer */
	int hshift;              /* Desired left-shift of output line buffer */
	int tabstops[TABSTOP_MAX]; /* Custom tabstops */

	int curr;                /* Index into linebuf */
	int column;              /* Printable length, accounting for
				   backspaces, etc. */
	int overstrike;          /* Next char should overstrike previous char */
	int last_overstrike;     /*  = AT_NORMAL */
	int is_null_line;        /* There is no current line */
	int lmargin;             /* Left margin */
	LWCHAR pendc;
	POSITION pendpos;
	const char *end_ansi_chars;
	const char *mid_ansi_chars;
	int reading;
	int curr_on;
	struct termios save_term;
	int saved_term;
	/* ch_ungetchar is a rather kludgy and limited way to push
	 * a single char onto an input file descriptor. */
	int ch_ungotchar;
	const char *binfmt;
	const char *utfbinfmt;

	int binattr;
	POSITION start_attnpos;
	POSITION end_attnpos;
	long jump_sline_fraction;
	char openquote;
	char closequote;
	int first_time;
	struct hilite hilite_anchor;
	/* Anchor for linked list. */
	struct ifile ifile_anchor;
	int maxbufs;
	int ntabstops;  /* = 1 Number of tabstops */
	int tabdefault; /* = 8 Default repeated tabstops */
#if CMD_HISTORY
/* These are the various command histories that exist. */
	struct mlist mlist_search;
	struct mlist mlist_examine;
#if SHELL_ESCAPE || PIPEC
	struct mlist mlist_shell;
#endif
#endif /* CMD_HISTORY */
	int lnloop;                      /* Are we in the line num loop? */
	struct linenum_info anchor;      /* Anchor of the list */
	struct linenum_info *freelist;   /* Anchor of the unused entries */
	struct linenum_info pool[NPOOL]; /* The pool itself */
	struct linenum_info *spare;              /* We always keep one spare entry */
	struct filestate *thisfile;
	int loopcount;
	long startime;
#if SHELL_ESCAPE
	char *shellcmd;   /* For holding last shell command for "!!" */
#endif
	int mca;                 /* The multicharacter command (action) */
	int search_type;         /* The previous type of search */
	LINENUM number;          /* The number typed by the user */
	char optchar;
	int optflag;
	int optgetname;
	POSITION bottompos;
	int save_hshift;
#if PIPEC
	char pipec;
#endif
	char cmdbuf[CMDBUF_SIZE]; /* Buffer for holding a multi-char command */
	int cmd_col;             /* Current column of the cursor */
	int prompt_col;          /* Column of cursor just after prompt */
	const char *cp;          /* Pointer into cmdbuf */
	int cmd_offset;          /* Index into cmdbuf of first displayed char */
	int literal;             /* Next input char should not be interpreted */
	char chardef[256];
	struct mlist *curr_mlist;
	int curr_cmdflags;
	char cmd_mbc_buf[MAX_UTF_CHAR_LEN];
#if SUPPORT_UTF8
	int cmd_mbc_buf_len;
#endif
	int errmsgs;     /* Count of messages displayed by error() */
	int final_attr;
	long fraction;           /* The fractional part of the number */
#if HILITE_SEARCH
	int hide_hilite;
	int line_matches;        /* Number of search matches in this line */
	POSITION prep_startpos;
	POSITION prep_endpos;
#endif
	int jump_sline;          /* Screen line of "jump target" */
	char *mp;
	char message[PROMPT_SIZE];

	POSITION *table;         /* The position table */
	/* List of command tables and list of line-edit tables. */
	struct tablelist *list_fcmd_tables;
	struct tablelist *list_ecmd_tables;
	struct tablelist *list_var_tables;
	struct tablelist *list_sysvar_tables;
	int consecutive_nulls;
	struct mark marks[NMARKS];
	/* Strings passed to do various terminal functions. */
	char *sc_u_in;               /* Enter underline mode */
	char *sc_bl_in;              /* Enter blink mode */
	char *sc_visual_bell;        /* Visual bell (flash screen) sequence */
	char *sc_backspace;          /* Backspace cursor */
	char *ungotp;
	char ungot[UNGOT_SIZE];
	int save_status;
	int table_size;
	const char *mchars;
	struct mark sm;
	char buf2[8];
} bb_common_bss;
#define any_display bb_common_bss.any_display
#define plusoption bb_common_bss.plusoption
#define fd0 bb_common_bss.fd0
#define curr_dev bb_common_bss.curr_dev
#define curr_ino bb_common_bss.curr_ino
#define curr_altfilename bb_common_bss.curr_altfilename
#define curr_altpipe bb_common_bss.curr_altpipe
#define is_ucase_pattern bb_common_bss.is_ucase_pattern
#define is_caseless bb_common_bss.is_caseless
#define read_label bb_common_bss.read_label
#define last_search_type bb_common_bss.last_search_type
#define last_pattern bb_common_bss.last_pattern
#define regpattern bb_common_bss.regpattern
#define prproto bb_common_bss.prproto
#define eqproto bb_common_bss.eqproto
#define hproto bb_common_bss.hproto
#define wproto bb_common_bss.wproto
#define ifiles bb_common_bss.ifiles
#define sigs bb_common_bss.sigs
#define bs_mode bb_common_bss.bs_mode
#define linenums bb_common_bss.linenums
#define ctldisp bb_common_bss.ctldisp
#define twiddle bb_common_bss.twiddle
#define status_col bb_common_bss.status_col
#define ignaw bb_common_bss.ignaw
#define erase_char bb_common_bss.erase_char
#define erase2_char bb_common_bss.erase2_char
#define kill_char bb_common_bss.kill_char
#define werase_char bb_common_bss.werase_char
#define sc_width bb_common_bss.sc_width
#define sc_height bb_common_bss.sc_height
#define every_first_cmd bb_common_bss.every_first_cmd
#define new_file bb_common_bss.new_file
#define is_tty bb_common_bss.is_tty
#define curr_ifile bb_common_bss.curr_ifile
#define old_ifile bb_common_bss.old_ifile
#define initial_scrpos bb_common_bss.initial_scrpos
#define wscroll bb_common_bss.wscroll
#define quitting bb_common_bss.quitting
#define secure bb_common_bss.secure
#define dohelp bb_common_bss.dohelp
#define less_is_more bb_common_bss.less_is_more
#define mbc_buf bb_common_bss.mbc_buf
#define mbc_buf_len bb_common_bss.mbc_buf_len
#define mbc_buf_index bb_common_bss.mbc_buf_index
#define mbc_pos bb_common_bss.mbc_pos
#if SUPPORT_UTF8
#define utf_mode bb_common_bss.utf_mode
#else
#define utf_mode 0
#endif
#define hit_eof bb_common_bss.hit_eof
#define screen_trashed bb_common_bss.screen_trashed
#define squished bb_common_bss.squished
#define forw_prompt bb_common_bss.forw_prompt
#define tagoption bb_common_bss.tagoption
#define ignore_eoi bb_common_bss.ignore_eoi
#define quiet bb_common_bss.quiet
#define how_search bb_common_bss.how_search
#define top_scroll bb_common_bss.top_scroll
#define pr_type bb_common_bss.pr_type
#define quit_at_eof bb_common_bss.quit_at_eof
#define quit_if_one_screen bb_common_bss.quit_if_one_screen
#define squeeze bb_common_bss.squeeze
#define back_scroll bb_common_bss.back_scroll
#define forw_scroll bb_common_bss.forw_scroll
#define caseless bb_common_bss.caseless
#define autobuf bb_common_bss.autobuf
#define bufspace bb_common_bss.bufspace
#define force_open bb_common_bss.force_open
#define swindow bb_common_bss.swindow
#define chopline bb_common_bss.chopline
#define show_attn bb_common_bss.show_attn
#define shift_count bb_common_bss.shift_count
#define use_lessopen bb_common_bss.use_lessopen
#define quit_on_intr bb_common_bss.quit_on_intr
#define follow_mode bb_common_bss.follow_mode
#define hilite_search bb_common_bss.hilite_search
#define editor bb_common_bss.editor
#define editproto bb_common_bss.editproto
#define pendopt bb_common_bss.pendopt
#define need_clr bb_common_bss.need_clr
#define obuf bb_common_bss.obuf
#define ob bb_common_bss.ob
#define at_prompt bb_common_bss.at_prompt
#define attrmode bb_common_bss.attrmode
#define tty bb_common_bss.tty
#define linebuf bb_common_bss.linebuf
#define attr bb_common_bss.attr
#define size_linebuf bb_common_bss.size_linebuf
#define cshift bb_common_bss.cshift
#define hshift bb_common_bss.hshift
#define tabstops bb_common_bss.tabstops
#define curr bb_common_bss.curr
#define column bb_common_bss.column
#define overstrike bb_common_bss.overstrike
#define last_overstrike bb_common_bss.last_overstrike
#define is_null_line bb_common_bss.is_null_line
#define lmargin bb_common_bss.lmargin
#define pendc bb_common_bss.pendc
#define pendpos bb_common_bss.pendpos
#define end_ansi_chars bb_common_bss.end_ansi_chars
#define mid_ansi_chars bb_common_bss.mid_ansi_chars
#define reading bb_common_bss.reading
#define curr_on bb_common_bss.curr_on
#define save_term bb_common_bss.save_term
#define saved_term bb_common_bss.saved_term
#define ch_ungotchar bb_common_bss.ch_ungotchar
#define binfmt bb_common_bss.binfmt
#define utfbinfmt bb_common_bss.utfbinfmt
#define binattr bb_common_bss.binattr
#define start_attnpos bb_common_bss.start_attnpos
#define end_attnpos bb_common_bss.end_attnpos
#define jump_sline_fraction bb_common_bss.jump_sline_fraction
#define openquote bb_common_bss.openquote
#define closequote bb_common_bss.closequote
#define first_time bb_common_bss.first_time
#define hilite_anchor bb_common_bss.hilite_anchor
#define ifile_anchor bb_common_bss.ifile_anchor
#define maxbufs bb_common_bss.maxbufs
#define ntabstops bb_common_bss.ntabstops
#define tabdefault bb_common_bss.tabdefault

#define lnloop bb_common_bss.lnloop
#define anchor bb_common_bss.anchor
#define freelist bb_common_bss.freelist
#define pool bb_common_bss.pool
#define spare bb_common_bss.spare
#define thisfile bb_common_bss.thisfile
#define loopcount bb_common_bss.loopcount
#define startime bb_common_bss.startime
#define shellcmd bb_common_bss.shellcmd
#define mca bb_common_bss.mca
#define search_type bb_common_bss.search_type
#define number bb_common_bss.number
#define optchar bb_common_bss.optchar
#define optflag bb_common_bss.optflag
#define optgetname bb_common_bss.optgetname
#define bottompos bb_common_bss.bottompos
#define save_hshift bb_common_bss.save_hshift
#define pipec bb_common_bss.pipec
#define cmdbuf bb_common_bss.cmdbuf
#define cmd_col bb_common_bss.cmd_col
#define prompt_col bb_common_bss.prompt_col
#define cp bb_common_bss.cp
#define cmd_offset bb_common_bss.cmd_offset
#define literal bb_common_bss.literal
#define chardef bb_common_bss.chardef
#define curr_mlist bb_common_bss.curr_mlist
#define curr_cmdflags bb_common_bss.curr_cmdflags
#define cmd_mbc_buf bb_common_bss.cmd_mbc_buf
#define cmd_mbc_buf_len bb_common_bss.cmd_mbc_buf_len
#define errmsgs bb_common_bss.errmsgs
#define final_attr bb_common_bss.final_attr
#define fraction bb_common_bss.fraction
#define hide_hilite bb_common_bss.hide_hilite
#define line_matches bb_common_bss.line_matches
#define prep_startpos bb_common_bss.prep_startpos
#define prep_endpos bb_common_bss.prep_endpos
#define jump_sline bb_common_bss.jump_sline
#define mp bb_common_bss.mp
#define message bb_common_bss.message
#define table bb_common_bss.table
#define list_fcmd_tables bb_common_bss.list_fcmd_tables
#define list_ecmd_tables bb_common_bss.list_ecmd_tables
#define list_var_tables bb_common_bss.list_var_tables
#define list_sysvar_tables bb_common_bss.list_sysvar_tables
#define consecutive_nulls bb_common_bss.consecutive_nulls
#define marks bb_common_bss.marks
#define sc_u_in bb_common_bss.sc_u_in
#define sc_bl_in bb_common_bss.sc_bl_in
#define sc_visual_bell bb_common_bss.sc_visual_bell
#define sc_backspace bb_common_bss.sc_backspace
#define ungotp bb_common_bss.ungotp
#define ungot bb_common_bss.ungot
#define save_status bb_common_bss.save_status
#define table_size bb_common_bss.table_size
#define mchars bb_common_bss.mchars
#define sm bb_common_bss.sm
#define buf2 bb_common_bss.buf2


#if CMD_HISTORY
#define mlist_search bb_common_bss.mlist_search
#define mlist_examine bb_common_bss.mlist_examine
#define mlist_shell bb_common_bss.mlist_shell

/* History file */
#define HISTFILE_FIRST_LINE      ".less-history-file:"
#define HISTFILE_SEARCH_SECTION  ".search"
#define HISTFILE_SHELL_SECTION   ".shell"
static void * const ml_search = (void *) &mlist_search;
static void * const ml_examine = (void *) &mlist_examine;

#if SHELL_ESCAPE || PIPEC
static void * const ml_shell = (void *) &mlist_shell;
#endif

#else /* CMD_HISTORY */

/* If CMD_HISTORY is off, these are just flags. */
static void * const ml_search = (void *)1;
static void * const ml_examine = (void *)2;
#if SHELL_ESCAPE || PIPEC
static void * const ml_shell = (void *)3;
#endif

#endif /* CMD_HISTORY */

static int edit_ifile(IFILE ifile);
static void error(const char *fmt, PARG *parg);
static void save_cmdhist(void);
static void clear_bot(void);
static void ierror(const char *fmt, PARG *parg);
static void get_term(void);
static const struct loption *findopt_name(const char **p_optname, const char **p_oname, int *p_err);

static void *emalloc(size_t count);

/* Convert an integral type to a string. */
#define TYPE_TO_A_FUNC(funcname, type) \
static void funcname(type num, char *buf) \
{ \
	int neg = (num < 0); \
	char tbuf[INT_STRLEN_BOUND(num)+2]; \
	char *s = tbuf + sizeof(tbuf); \
	if (neg) num = -num; \
	*--s = '\0'; \
	do { \
		*--s = (num % 10) + '0'; \
	} while ((num /= 10) != 0); \
	if (neg) *--s = '-'; \
	strcpy(buf, s); \
}

TYPE_TO_A_FUNC(postoa, POSITION)
TYPE_TO_A_FUNC(linenumtoa, LINENUM)
TYPE_TO_A_FUNC(inttoa, int)

/* Find an ifile structure, given a filename. */
static struct ifile *
find_ifile(const char *filename)
{
	struct ifile *p;

	for (p = ifile_anchor.h_next;  p != &ifile_anchor;  p = p->h_next)
		if (strcmp(filename, p->h_filename) == 0)
			return p;
	return NULL;
}

static void
incr_index(struct ifile *p, int incr)
{
	for (;  p != &ifile_anchor;  p = p->h_next)
		p->h_index += incr;
}

/* Link an ifile into the ifile list. */
static void
link_ifile(struct ifile *p, struct ifile *prev)
{
	/* Link into list. */
	if (prev == NULL)
		prev = &ifile_anchor;
	p->h_next = prev->h_next;
	p->h_prev = prev;
	prev->h_next->h_prev = p;
	prev->h_next = p;
	/* Calculate index for the new one,
	 * and adjust the indexes for subsequent ifiles in the list. */
	p->h_index = prev->h_index + 1;
	incr_index(p->h_next, 1);
	ifiles++;
}

/* Copy a string to a "safe" place
 * (that is, to a buffer allocated by calloc). */
static char *
save(const char *s)
{
	char *p;

	p = emalloc(strlen(s)+1);
	strcpy(p, s);
	return p;
}

/* Allocate a new ifile structure and stick a filename in it.
 * It should go after "prev" in the list
 * (or at the beginning of the list if "prev" is NULL).
 * Return a pointer to the new ifile structure. */
static struct ifile *
new_ifile(const char *filename, struct ifile *prev)
{
	struct ifile *p;

	/* Allocate and initialize structure. */
	p = emalloc(sizeof(struct ifile));
	p->h_filename = save(filename);
	p->h_scrpos.pos = NULL_POSITION;
	p->h_opened = 0;
	p->h_hold = 0;
	p->h_filestate = NULL;
	link_ifile(p, prev);
	return p;
}

/* Get the ifile associated with a filename.
 * If the filename has not been seen before,
 * insert the new ifile after "prev" in the list. */
static IFILE
get_ifile(const char *filename, IFILE prev)
{
	struct ifile *p;

	if ((p = find_ifile(filename)) == NULL)
		p = new_ifile(filename, int_ifile(prev));
	return ext_ifile(p);
}

/* Edit a new file (given its name).
 * Filename == "-" means standard input.
 * Filename == NULL means just close the current file. */
static int
edit(const char *filename)
{
	if (filename == NULL)
		return edit_ifile(NULL_IFILE);
	return edit_ifile(get_ifile(filename, curr_ifile));
}

/* make minimal harcoded vt100-like TERM by dzo@ */
#define sc_eol_clear "\033[K"           /* Clear to end of line */
#define sc_eos_clear "\033[J"           /* Clear to end of screen */
#define sc_clear     "\033[H\033[J"     /* Clear screen */
#define sc_home      "\033[H"           /* Cursor home */
#define sc_s_in      "\033[7m"  /* Enter standout (highlighted) mode */
#define sc_s_out     "\033[m"           /* Exit standout mode */
#define sc_u_out     "\033[m"           /* Exit underline mode */
#define sc_b_in      "\033[1m"          /* Enter bold mode */
#define sc_b_out     "\033[m"           /* Exit bold mode */
#define sc_bl_out    "\033[m"           /* Exit blink mode */
#define sc_return    "\r"       /* Cursor to beginning of current line */
#define sc_addline   "\033[L"   /* Add line, scroll down following lines */


/* Flush buffered output.
 *
 * If we haven't displayed any file data yet,
 * output messages on error output (file descriptor 2),
 * otherwise output on standard output (file descriptor 1).
 *
 * This has the desirable effect of producing all
 * error messages on error output if standard output
 * is directed to a file.  It also does the same if
 * we never produce any real output; for example, if
 * the input file(s) cannot be opened.  If we do
 * eventually produce output, code in edit() makes
 * sure these messages can be seen before they are
 * overwritten or scrolled away. */
static void
flush(void)
{
	int n;
	int fd;

	n = ob - obuf;
	if (n == 0)
		return;

	fd = (any_display) ? 1 : 2;
	if (write(fd, obuf, n) != n)
		screen_trashed = 1;
	ob = obuf;
}

/* Output a character. */
static int
putchr(int c)
{
	if (need_clr) {
		need_clr = 0;
		clear_bot();
	}
	/* Some versions of flush() write to *ob, so we must flush
	 * when we are still one char from the end of obuf. */
	if (ob >= &obuf[sizeof(obuf)-1])
		flush();
	*ob++ = c;
	at_prompt = 0;
	return c;
}

#define DEFAULT_TERM            "ansi"

/* Change terminal to "raw mode", or restore to "normal" mode.
 * "Raw mode" means
 *      1. An outstanding read will complete on receipt of a single keystroke.
 *      2. Input is not echoed.
 *      3. On output, \n is mapped to \r\n.
 *      4. \t is NOT expanded into spaces.
 *      5. Signal-causing characters such as ctrl-C (interrupt),
 *         etc. are NOT disabled.
 * It doesn't matter whether an input \n is mapped to \r, or vice versa. */
static void
raw_mode(int on)
{
	struct termios s;

	if (on == curr_on)
		return;
	erase2_char = '\b'; /* in case OS doesn't know about erase2 */

	if (on) {
		/* Get terminal modes. */
		tcgetattr(tty, &s);

		/* Save modes and set certain variables dependent on modes. */
		if (!saved_term) {
			save_term = s;
			saved_term = 1;
		}
		erase_char = s.c_cc[VERASE];
#ifdef VERASE2
		erase2_char = s.c_cc[VERASE2];
#endif
		kill_char = s.c_cc[VKILL];
#ifdef VWERASE
		werase_char = s.c_cc[VWERASE];
#else
		werase_char = CONTROL('W');
#endif

		/* Set the modes to the way we want them. */
		s.c_lflag &= ~(0
#ifdef ICANON
			| ICANON
#endif
#ifdef ECHO
			| ECHO
#endif
#ifdef ECHOE
			| ECHOE
#endif
#ifdef ECHOK
			| ECHOK
#endif
#if ECHONL
			| ECHONL
#endif
		);

		s.c_oflag |= (0
#ifdef OXTABS
			| OXTABS
#else
#ifdef TAB3
			| TAB3
#else
#ifdef XTABS
			| XTABS
#endif
#endif
#endif
#ifdef OPOST
			| OPOST
#endif
#ifdef ONLCR
			| ONLCR
#endif
		);

		s.c_oflag &= ~(0
#ifdef ONOEOT
			| ONOEOT
#endif
#ifdef OCRNL
			| OCRNL
#endif
#ifdef ONOCR
			| ONOCR
#endif
#ifdef ONLRET
			| ONLRET
#endif
		);
		s.c_cc[VMIN] = 1;
		s.c_cc[VTIME] = 0;
#ifdef VLNEXT
		s.c_cc[VLNEXT] = 0;
#endif
#ifdef VDSUSP
		s.c_cc[VDSUSP] = 0;
#endif
	} else {
		/* Restore saved modes. */
		s = save_term;
	}
	fsync(tty);
	tcsetattr(tty, TCSADRAIN, &s);
	curr_on = on;
}

/* Exit the program. */
static void quit(int status) __attribute__ ((noreturn));

static void
quit(int status)
{
	/* Put cursor at bottom left corner, clear the line,
	 * reset the terminal modes, and exit. */
	if (status < 0)
		status = save_status;
	else
		save_status = status;
	quitting = 1;
	edit((char*)NULL);
	save_cmdhist();
	if (any_display && is_tty)
		clear_bot();
	flush();
	raw_mode(0);
	exit(status);
}

/* Allocate memory.
 * Like calloc(), but never returns an error (NULL). */
static void *
emalloc(size_t count)
{
	void * p;

	p = malloc(count);
	if (p != NULL) {
		memset(p, 0, count);
		return p;
	}
	error("Cannot allocate memory", NULL_PARG);
	quit(QUIT_ERROR);
	return NULL;
}

/* Structure to support a list of command tables. */
struct tablelist
{
	struct tablelist *t_next;
	char *t_start;
	char *t_end;
};

/* Search a single command table for the command string in cmd. */
static int
cmd_search(const char *cmd, const char *starttable, const char *endtable, const char **sp)
{
	const char *p;
	const char *q;
	int a;

	*sp = NULL;
	for (p = starttable, q = cmd;  p < endtable;  p++, q++) {
		if (*p == *q) {
			/* Current characters match.
			 * If we're at the end of the string, we've found it.
			 * Return the action code, which is the character
			 * after the null at the end of the string
			 * in the command table. */
			if (*p == '\0') {
				a = *++p & 0377;
				while (a == A_SKIP)
					a = *++p & 0377;
				if (a == A_END_LIST) {
					/* We get here only if the original
					 * cmd string passed in was empty ("").
					 * I don't think that can happen,
					 * but just in case ... */
					return A_UINVALID;
				}
				/* Check for an "extra" string. */
				if (a & A_EXTRA) {
					*sp = ++p;
					a &= ~A_EXTRA;
				}
				return a;
			}
		} else if (*q == '\0') {
			/* Hit the end of the user's command,
			 * but not the end of the string in the command table.
			 * The user's command is incomplete. */
			return A_PREFIX;
		} else {
			/* Not a match.
			 * Skip ahead to the next command in the
			 * command table, and reset the pointer
			 * to the beginning of the user's command. */
			if (*p == '\0' && p[1] == A_END_LIST) {
				/* A_END_LIST is a special marker that tells
				 * us to abort the cmd search. */
				return A_UINVALID;
			}
			while (*p++ != '\0')
				continue;
			while (*p == A_SKIP)
				p++;
			if (*p & A_EXTRA)
				while (*++p != '\0')
					continue;
			q = cmd-1;
		}
	}
	/* No match found in the entire command table. */
	return A_INVALID;
}

/* Decode a command character and return the associated action.
 * The "extra" string, if any, is returned in sp. */
static int
cmd_decode(const struct tablelist *tlist, const char *cmd, const char **sp)
{
	const struct tablelist *t;
	int action = A_INVALID;

	/* Search thru all the command tables.
	 * Stop when we find an action which is not A_INVALID. */
	for (t = tlist;  t != NULL;  t = t->t_next) {
		action = cmd_search(cmd, t->t_start, t->t_end, sp);
		if (action != A_INVALID)
			break;
	}
	if (action == A_UINVALID)
		action = A_INVALID;
	return action;
}

/* Get the value of an environment variable.
 * Looks first in the lesskey file, then in the real environment. */
static const char *
lgetenv(const char *var)
{
	int a;
	const char *s;

	a = cmd_decode(list_var_tables, var, &s);
	if (a == EV_OK)
		return s;
	s = getenv(var);
	if (s != NULL && *s != '\0')
		return s;
	a = cmd_decode(list_sysvar_tables, var, &s);
	if (a == EV_OK)
		return s;
	return NULL;
}


/* Expand the line buffer. */
static int
expand_linebuf(void)
{
	/* Double the size of the line buffer. */
	int new_size = size_linebuf * 2;

	char *new_buf = (char *) calloc(new_size, sizeof(char));
	char *new_attr = (char *) calloc(new_size, sizeof(char));

	if (new_buf == NULL || new_attr == NULL) {
		if (new_attr != NULL)
			free(new_attr);
		if (new_buf != NULL)
			free(new_buf);
		return 1;
	}
	/* We just calloc'd the buffers; copy the old contents. */
	memcpy(new_buf, linebuf, size_linebuf * sizeof(char));
	memcpy(new_attr, attr, size_linebuf * sizeof(char));
	free(attr);
	free(linebuf);

	linebuf = new_buf;
	attr = new_attr;
	size_linebuf = new_size;
	return 0;
}

/* Is a character ASCII? */
static int
is_ascii_char(LWCHAR ch)
{
	return (ch <= 0x7F);
}

/* Rewind the line buffer. */
static void
prewind(void)
{
	curr = 0;
	column = 0;
	cshift = 0;
	overstrike = 0;
	last_overstrike = AT_NORMAL;
#if SUPPORT_UTF8
	mbc_buf_len = 0;
#endif
	is_null_line = 0;
	pendc = '\0';
	lmargin = 0;
	if (status_col)
		lmargin += 1;
#if HILITE_SEARCH
	line_matches = 0;
#endif
}

/* Return the current time. */
static long
get_time(void)
{
	time_t t;

	time(&t);
	return t;
}

/* If we get stuck in a long loop trying to figure out the
 * line number, print a message to tell the user what we're doing. */
static void
longloopmessage(void)
{
	ierror("Calculating line numbers", NULL_PARG);
	/* Set the lnloop flag here, so if the user interrupts while
	 * we are calculating line numbers, the signal handler will
	 * turn off line numbers (linenums=0). */
	lnloop = 1;
}

static void
longish(void)
{
	if (loopcount >= 0 && ++loopcount > 100) {
		loopcount = 0;
		if (get_time() >= startime + LONGTIME) {
			longloopmessage();
			loopcount = -1;
		}
	}
}

/* Return the length of the file, if known. */
static POSITION
ch_length(void)
{
	if (thisfile == NULL)
		return NULL_POSITION;
	if (ignore_eoi)
		return NULL_POSITION;
	if (ch_flags & CH_HELPFILE)
		return size_helpdata;
	return ch_fsize;
}

/* Determine if a specific block is currently in one of the buffers. */
static int
buffered(BLOCKNUM block)
{
	struct buf *bp;
	int h;

	h = BUFHASH(block);
	FOR_BUFS_IN_CHAIN(h, bp) {
		if (bp->block == block)
			return TRUE;
	}
	return FALSE;
}

/* Allocate a new buffer.
 * The buffer is added to the tail of the buffer chain. */
static int
ch_addbuf(void)
{
	struct buf *bp;

	/* Allocate and initialize a new buffer and link it
	 * onto the tail of the buffer list. */
	bp = (struct buf *) calloc(1, sizeof(struct buf));
	if (bp == NULL)
		return 1;
	ch_nbufs++;
	bp->block = -1;
	bp->next = END_OF_CHAIN;
	bp->prev = ch_buftail;
	ch_buftail->next = bp;
	ch_buftail = bp;
	HASH_INS(bp, 0);
	return 0;
}

/* Output a string. */
static void
putstr(const char *s)
{
	while (*s != '\0')
		putchr(*s++);
}

/* Clear from the cursor to the end of the cursor's line.
 * {{ This must not move the cursor. }} */
static void
clear_eol(void)
{
	putstr(sc_eol_clear);
}

static void
ch_ungetchar(int c)
{
	if (c != -1 && ch_ungotchar != -1)
		error("ch_ungetchar overrun", NULL_PARG);
	ch_ungotchar = c;
}

/* Get the filename associated with a ifile. */
static char *
get_filename(IFILE ifile)
{
	if (ifile == NULL)
		return NULL;
	return int_ifile(ifile)->h_filename;
}

/* Adjust a screen line number to be a simple positive integer
 * in the range { 0 .. sc_height-2 }.
 * (The bottom line, sc_height-1, is reserved for prompts, etc.)
 * The given "sline" may be in the range { 1 .. sc_height-1 }
 * to refer to lines relative to the top of the screen (starting from 1),
 * or it may be in { -1 .. -(sc_height-1) } to refer to lines
 * relative to the bottom of the screen. */
static int
adjsline(int sline)
{
	/* Negative screen line number means
	 * relative to the bottom of the screen. */
	if (sline < 0)
		sline += sc_height;
	/* Can't be less than 1 or greater than sc_height-1. */
	if (sline <= 0)
		sline = 1;
	if (sline >= sc_height)
		sline = sc_height - 1;
	/* Return zero-based line number, not one-based. */
	return sline-1;
}


/* Decode a char that represents a position on the screen. */
static char *
wherechar(char *p, int *wp)
{
	switch (*p) {
	case 'b': case 'd': case 'l': case 'p': case 'P':
		switch (*++p) {
		case 't':   *wp = TOP;                  break;
		case 'm':   *wp = MIDDLE;               break;
		case 'b':   *wp = BOTTOM;               break;
		case 'B':   *wp = BOTTOM_PLUS_ONE;      break;
		case 'j':   *wp = adjsline(jump_sline); break;
		default:    *wp = TOP;  p--;            break;
		}
	}
	return p;
}


/* Append a string to the end of the message. */
static void
ap_str(char *s)
{
	int len;

	len = strlen(s);
	if (mp + len >= message + PROMPT_SIZE)
		len = message + PROMPT_SIZE - mp - 1;
	strncpy(mp, s, len);
	mp += len;
	*mp = '\0';
}

/* Append a character to the end of the message. */
static void
ap_char(char c)
{
	char buf[2];

	buf[0] = c;
	buf[1] = '\0';
	ap_str(buf);
}

/* Skip a false conditional.
 * When a false condition is found (either a false IF or the ELSE part
 * of a true IF), this routine scans the prototype string to decide
 * where to resume parsing the string.
 * We must keep track of nested IFs and skip them properly. */
static char *
skipcond(char *p)
{
	int iflevel;

	/* We came in here after processing a ? or :,
	 * so we start nested one level deep. */
	iflevel = 1;

	for (;;) switch (*++p) {
	case '?':
		/* Start of a nested IF. */
		iflevel++;
		break;
	case ':':
		/* Else.
		 * If this matches the IF we came in here with,
		 * then we're done. */
		if (iflevel == 1)
			return p;
		break;
	case '.':
		/* Endif.
		 * If this matches the IF we came in here with,
		 * then we're done. */
		if (--iflevel == 0)
			return p;
		break;
	case '\\':
		/* Backslash escapes the next character. */
		++p;
		break;
	case '\0':
		/* Whoops.  Hit end of string.
		 * This is a malformed conditional, but just treat it
		 * as if all active conditionals ends here. */
		return p-1;
	}
	/*NOTREACHED*/
}

/* Return the starting file position of a line displayed on the screen.
 * The line may be specified as a line number relative to the top
 * of the screen, but is usually one of these special cases:
 *      the top (first) line on the screen
 *      the second line on the screen
 *      the bottom line on the screen
 *      the line after the bottom line on the screen */
static POSITION
position(int where)
{
	switch (where) {
	case BOTTOM:
		where = sc_height - 2;
		break;
	case BOTTOM_PLUS_ONE:
		where = sc_height - 1;
		break;
	case MIDDLE:
		where = (sc_height - 1) / 2;
	}
	return table[where];
}

/* Return the "current" byte offset in the file. */
static POSITION
curr_byte(int where)
{
	POSITION pos;

	pos = position(where);
	while (pos == NULL_POSITION && where >= 0 && where < sc_height-1)
		pos = position(++where);
	if (pos == NULL_POSITION)
		pos = ch_length();
	return pos;
}

/* Append a POSITION (as a decimal integer) to the end of the message. */
static void
ap_pos(POSITION pos)
{
	char buf[INT_STRLEN_BOUND(pos) + 2];

	postoa(pos, buf);
	ap_str(buf);
}

/* Append a question mark to the end of the message. */
static void
ap_quest(void)
{
	ap_str("?");
}

/* Append an integer to the end of the message. */
static void
ap_int(int num)
{
	char buf[INT_STRLEN_BOUND(num) + 2];

	inttoa(num, buf);
	ap_str(buf);
}

/* Calculate the gap for an entry. */
static void
calcgap(struct linenum_info *p)
{
	/* Don't bother to compute a gap for the anchor.
	 * Also don't compute a gap for the last one in the list.
	 * The gap for that last one should be considered infinite,
	 * but we never look at it anyway. */
	if (p == &anchor || p->next == &anchor)
		return;
	p->gap = p->next->pos - p->prev->pos;
}

/* Add a new line number to the cache.
 * The specified position (pos) should be the file position of the
 * FIRST character in the specified line. */
static void
add_lnum(LINENUM linenum, POSITION pos)
{
	struct linenum_info *p;
	struct linenum_info *new;
	struct linenum_info *nextp;
	struct linenum_info *prevp;
	POSITION mingap;

	/* Find the proper place in the list for the new one.
	 * The entries are sorted by position. */
	for (p = anchor.next;  p != &anchor && p->pos < pos;  p = p->next)
		if (p->line == linenum)
			/* We already have this one. */
			return;
	nextp = p;
	prevp = p->prev;

	if (freelist != NULL) {
		/* We still have free (unused) entries. Use one of them. */
		new = freelist;
		freelist = freelist->next;
	} else {
		/* No free entries. Use the "spare" entry. */
		new = spare;
		spare = NULL;
	}

	/* Fill in the fields of the new entry,
	 * and insert it into the proper place in the list. */
	new->next = nextp;
	new->prev = prevp;
	new->pos = pos;
	new->line = linenum;

	nextp->prev = new;
	prevp->next = new;

	/* Recalculate gaps for the new entry and the neighboring entries. */
	calcgap(new);
	calcgap(nextp);
	calcgap(prevp);

	if (spare == NULL) {
		/* We have used the spare entry.
		 * Scan the list to find the one with the smallest
		 * gap, take it out and make it the spare.
		 * We should never remove the last one, so stop when
		 * we get to p->next == &anchor.  This also avoids
		 * looking at the gap of the last one, which is
		 * not computed by calcgap. */
		mingap = anchor.next->gap;
		for (p = anchor.next;  p->next != &anchor;  p = p->next) {
			if (p->gap <= mingap) {
				spare = p;
				mingap = p->gap;
			}
		}
		spare->next->prev = spare->prev;
		spare->prev->next = spare->next;
	}
}

static char *pr_expand(char *proto, int maxwidth);

/* Return a message suitable for printing while waiting in the F command. */
static char *
wait_message(void)
{
	return pr_expand(wproto, sc_width-2);
}

static int
fch_get(void)
{
	struct buf *bp;
	int n;
	int slept;
	int h;
	POSITION pos;
	POSITION len;

	if (thisfile == NULL)
		return EOI;

	slept = FALSE;

	/* Look for a buffer holding the desired block. */
	h = BUFHASH(ch_block);
	FOR_BUFS_IN_CHAIN(h, bp) {
		if (bp->block == ch_block) {
			if (ch_offset >= bp->datasize)
				/* Need more data in this buffer. */
				goto read_more;
			goto found;
		}
	}
	/* Block is not in a buffer.
	 * Take the least recently used buffer
	 * and read the desired block into it.
	 * If the LRU buffer has data in it,
	 * then maybe allocate a new buffer. */
	if (ch_buftail == END_OF_CHAIN || ch_buftail->block != -1) {
		/* There is no empty buffer to use.
		 * Allocate a new buffer if:
		 * 1. We can't seek on this file and -b is not in effect; or
		 * 2. We haven't allocated the max buffers for this file yet. */
		if ((autobuf && !(ch_flags & CH_CANSEEK)) ||
		    (maxbufs < 0 || ch_nbufs < maxbufs))
			if (ch_addbuf())
				/* Allocation failed: turn off autobuf. */
				autobuf = OPT_OFF;
	}
	bp = ch_buftail;
	HASH_RM(bp); /* Remove from old hash chain. */
	bp->block = ch_block;
	bp->datasize = 0;
	HASH_INS(bp, h); /* Insert into new hash chain. */

    read_more:
	pos = (ch_block * LBUFSIZE) + bp->datasize;
	if ((len = ch_length()) != NULL_POSITION && pos >= len)
		/* At end of file. */
		return EOI;

	if (pos != ch_fpos) {
		/* Not at the correct position: must seek.
		 * If input is a pipe, we're in trouble (can't seek on a pipe).
		 * Some data has been lost: just return "?". */
		if (!(ch_flags & CH_CANSEEK))
			return '?';
		if (lseek(ch_file, (off_t)pos, SEEK_SET) == BAD_LSEEK) {
			error("seek error", NULL_PARG);
			clear_eol();
			return EOI;
		}
		ch_fpos = pos;
	}

	/* Read the block.
	 * If we read less than a full block, that's ok.
	 * We use partial block and pick up the rest next time. */
	if (ch_ungotchar != -1) {
		bp->data[bp->datasize] = ch_ungotchar;
		n = 1;
		ch_ungotchar = -1;
	} else if (ch_flags & CH_HELPFILE) {
		bp->data[bp->datasize] = helpdata[ch_fpos];
		n = 1;
	} else {
		n = iread(ch_file, (char *)&bp->data[bp->datasize],
			(unsigned int)(LBUFSIZE - bp->datasize));
	}

	if (n == READ_INTR)
		return EOI;
	if (n < 0) {
		error("read error", NULL_PARG);
		clear_eol();
		n = 0;
	}

#if LOGFILE
	/* If we have a log file, write the new data to it. */
	if (!secure && logfile >= 0 && n > 0)
		write(logfile, (char *) &bp->data[bp->datasize], n);
#endif

	ch_fpos += n;
	bp->datasize += n;

	/* If we have read to end of file, set ch_fsize to indicate
	 * the position of the end of file. */
	if (n == 0) {
		ch_fsize = pos;
		if (ignore_eoi) {
			/* We are ignoring EOF.
			 * Wait a while, then try again. */
			if (!slept) {
				PARG parg;
				parg.p_string = wait_message();
				ierror("%s", &parg);
			}
			sleep(1);
			slept = TRUE;

			if (follow_mode == FOLLOW_NAME) {
				/* See whether the file's i-number has changed.
				 * If so, force the file to be closed and
				 * reopened. */
				struct stat st;
				int r = stat(get_filename(curr_ifile), &st);
				if (r == 0 && (st.st_ino != curr_ino ||
					st.st_dev != curr_dev)) {
					/* screen_trashed=2 causes
					 * make_display to reopen the file. */
					screen_trashed = 2;
					return EOI;
				}
			}
		}
		if (sigs)
			return EOI;
	}

    found:
	if (ch_bufhead != bp) {
		/* Move the buffer to the head of the buffer chain.
		 * This orders the buffer chain, most- to least-recently used. */
		bp->next->prev = bp->prev;
		bp->prev->next = bp->next;
		bp->next = ch_bufhead;
		bp->prev = END_OF_CHAIN;
		ch_bufhead->prev = bp;
		ch_bufhead = bp;

		/* Move to head of hash chain too. */
		HASH_RM(bp);
		HASH_INS(bp, h);
	}

	if (ch_offset >= bp->datasize)
		/* After all that, we still don't have enough data.
		 * Go back and try again. */
		goto read_more;

	return bp->data[ch_offset];
}

/* Get the current char and post-increment the read pointer. */
static int
ch_forw_get(void)
{
	int c;

	if (thisfile == NULL)
		return EOI;
	c = ch_get();
	if (c == EOI)
		return EOI;
	if (ch_offset < LBUFSIZE-1)
		ch_offset++;
	else {
		ch_block ++;
		ch_offset = 0;
	}
	return c;
}

/* Seek to a specified position in the file.
 * Return 0 if successful, non-zero if can't seek there. */
static int
ch_seek(POSITION pos)
{
	BLOCKNUM new_block;
	POSITION len;

	if (thisfile == NULL)
		return 0;

	len = ch_length();
	if (pos < ch_zero() || (len != NULL_POSITION && pos > len))
		return 1;

	new_block = pos / LBUFSIZE;
	if (!(ch_flags & CH_CANSEEK) && pos != ch_fpos && !buffered(new_block)) {
		if (ch_fpos > pos)
			return 1;
		while (ch_fpos < pos) {
			if (ch_forw_get() == EOI)
				return 1;
			if (ABORT_SIGS())
				return 1;
		}
		return 0;
	}
	/* Set read pointer. */
	ch_block = new_block;
	ch_offset = pos % LBUFSIZE;
	return 0;
}

/* Return the current position in the file. */
static POSITION
ch_tell(void)
{
	if (thisfile == NULL)
		return NULL_POSITION;
	return (ch_block * LBUFSIZE) + ch_offset;
}

/* Analogous to forw_line(), but deals with "raw lines":
 * lines which are not split for screen width.
 * {{ This is supposed to be more efficient than forw_line(). }} */
static POSITION
forw_raw_line(POSITION curr_pos, const char **linep, int *line_lenp)
{
	int n;
	int c;
	POSITION new_pos;

	if (curr_pos == NULL_POSITION || ch_seek(curr_pos) ||
		(c = ch_forw_get()) == EOI)
		return NULL_POSITION;

	n = 0;
	for (;;) {
		if (c == '\n' || c == EOI || ABORT_SIGS()) {
			new_pos = ch_tell();
			break;
		}
		if (n >= size_linebuf-1) {
			if (expand_linebuf()) {
				/* Overflowed the input buffer.
				 * Pretend the line ended here. */
				new_pos = ch_tell() - 1;
				break;
			}
		}
		linebuf[n++] = c;
		c = ch_forw_get();
	}
	linebuf[n] = '\0';
	if (linep != NULL)
		*linep = linebuf;
	if (line_lenp != NULL)
		*line_lenp = n;
	return new_pos;
}

/* Pre-decrement the read pointer and get the new current char. */
static int
ch_back_get(void)
{
	if (thisfile == NULL)
		return EOI;
	if (ch_offset > 0)
		ch_offset --;
	else {
		if (ch_block <= 0)
			return EOI;
		if (!(ch_flags & CH_CANSEEK) && !buffered(ch_block-1))
			return EOI;
		ch_block--;
		ch_offset = LBUFSIZE-1;
	}
	return ch_get();
}

/* Analogous to back_line(), but deals with "raw lines".
 * {{ This is supposed to be more efficient than back_line(). }} */
static POSITION
back_raw_line(POSITION curr_pos, const char **linep, int *line_lenp)
{
	int n;
	int c;
	POSITION new_pos;

	if (curr_pos == NULL_POSITION || curr_pos <= ch_zero() ||
		ch_seek(curr_pos-1))
		return NULL_POSITION;

	n = size_linebuf;
	linebuf[--n] = '\0';
	for (;;) {
		c = ch_back_get();
		if (c == '\n' || ABORT_SIGS()) {
			/* This is the newline ending the previous line.
			 * We have hit the beginning of the line. */
			new_pos = ch_tell() + 1;
			break;
		}
		if (c == EOI) {
			/* We have hit the beginning of the file.
			 * This must be the first line in the file.
			 * This must, of course, be the beginning of the line. */
			new_pos = ch_zero();
			break;
		}
		if (n <= 0) {
			int old_size_linebuf = size_linebuf;
			char *fm;
			char *to;
			if (expand_linebuf()) {
				/* Overflowed the input buffer.
				 * Pretend the line ended here. */
				new_pos = ch_tell() + 1;
				break;
			}
			/* Shift the data to the end of the new linebuf. */
			for (fm = linebuf + old_size_linebuf - 1,
			      to = linebuf + size_linebuf - 1;
			     fm >= linebuf;  fm--, to--)
				*to = *fm;
			n = size_linebuf - old_size_linebuf;
		}
		linebuf[--n] = c;
	}
	if (linep != NULL)
		*linep = &linebuf[n];
	if (line_lenp != NULL)
		*line_lenp = size_linebuf - 1 - n;
	return new_pos;
}

/* Find the line number associated with a given position.
 * Return 0 if we can't figure it out. */
static LINENUM
find_linenum(POSITION pos)
{
	struct linenum_info *p;
	LINENUM linenum;
	POSITION cpos;

	if (!linenums)
		/* We're not using line numbers. */
		return 0;
	if (pos == NULL_POSITION)
		/* Caller doesn't know what he's talking about. */
		return 0;
	if (pos <= ch_zero())
		/* Beginning of file is always line number 1. */
		return 1;

	/* Find the entry nearest to the position we want. */
	for (p = anchor.next;  p != &anchor && p->pos < pos;  p = p->next)
		continue;
	if (p->pos == pos)
		/* Found it exactly. */
		return p->line;

	/* This is the (possibly) time-consuming part.
	 * We start at the line we just found and start
	 * reading the file forward or backward till we
	 * get to the place we want.
	 *
	 * First decide whether we should go forward from the
	 * previous one or backwards from the next one.
	 * The decision is based on which way involves
	 * traversing fewer bytes in the file. */
	startime = get_time();
	if (p == &anchor || pos - p->prev->pos < p->pos - pos) {
		/* Go forward. */
		p = p->prev;
		if (ch_seek(p->pos))
			return 0;
		loopcount = 0;
		for (linenum = p->line, cpos = p->pos;  cpos < pos;  linenum++) {
			/* Allow a signal to abort this loop. */
			cpos = forw_raw_line(cpos, (const char **)NULL, (int *)NULL);
			if (ABORT_SIGS() || cpos == NULL_POSITION)
				return 0;
			longish();
		}
		lnloop = 0;
		/* We might as well cache it. */
		add_lnum(linenum, cpos);
		/* If the given position is not at the start of a line,
		 * make sure we return the correct line number. */
		if (cpos > pos)
			linenum--;
	} else {
		/* Go backward. */
		if (ch_seek(p->pos))
			return 0;
		loopcount = 0;
		for (linenum = p->line, cpos = p->pos;  cpos > pos;  linenum--) {
			/* Allow a signal to abort this loop. */
			cpos = back_raw_line(cpos, (const char **)NULL, (int *)NULL);
			if (ABORT_SIGS() || cpos == NULL_POSITION)
				return 0;
			longish();
		}
		lnloop = 0;
		/* We might as well cache it. */
		add_lnum(linenum, cpos);
	}

	return linenum;
}

/* Return the line number of the "current" line.
 * The argument "where" tells which line is to be considered
 * the "current" line (e.g. TOP, BOTTOM, MIDDLE, etc). */
static LINENUM
currline(int where)
{
	POSITION pos;
	POSITION len;
	LINENUM linenum;

	pos = position(where);
	len = ch_length();
	while (pos == NULL_POSITION && where >= 0 && where < sc_height)
		pos = position(++where);
	if (pos == NULL_POSITION)
		pos = len;
	linenum = find_linenum(pos);
	if (pos == len)
		linenum--;
	return linenum;
}

/* Append a line number to the end of the message. */
static void
ap_linenum(LINENUM linenum)
{
	char buf[INT_STRLEN_BOUND(linenum) + 2];

	linenumtoa(linenum, buf);
	ap_str(buf);
}

#if TAGS
static int total;
static int curseq;
/* Return the total number of tags. */
#define ntags() total
/* Return the sequence number of current tag. */
#define curr_tag() curseq
#endif

/* Return the number of ifiles. */
#define nifile() ifiles


/* Get the index of the file associated with a ifile. */
static inline int
get_index(IFILE ifile)
{
	return int_ifile(ifile)->h_index;
}

/* Return the ratio of two POSITIONS, as a percentage.
 * {{ Assumes a POSITION is a long int. }} */
static int
percentage(POSITION num, POSITION den)
{
	POSITION num100 = num * 100;

	if (num100 / 100 == num)
		return num100 / den;
	else
		return num / (den / 100);
}

/* Get the ifile after a given one in the list. */
static IFILE
next_ifile(IFILE h)
{
	struct ifile *p;

	p = (h == NULL_IFILE) ? &ifile_anchor : int_ifile(h);
	if (p->h_next == &ifile_anchor)
		return NULL_IFILE;
	return ext_ifile(p->h_next);
}

/* Decode a "percent" prototype character.
 * A prototype string may include various "percent" escapes;
 * that is, a percent sign followed by a single letter.
 * Here we decode that letter and take the appropriate action,
 * usually by appending something to the message being built. */
static void
protochar(int c, int where)
{
	POSITION pos;
	POSITION len;
#if TAGS
	int n;
#endif
	LINENUM linenum;
	LINENUM last_linenum;
	IFILE h;

#undef  PAGE_NUM
#define PAGE_NUM(linenum)  ((((linenum) - 1) / (sc_height - 1)) + 1)

	switch (c) {
	case 'b':       /* Current byte offset */
		pos = curr_byte(where);
		if (pos != NULL_POSITION)
			ap_pos(pos);
		else
			ap_quest();
		break;
	case 'c':
		ap_int(hshift);
		break;
	case 'd':       /* Current page number */
		linenum = currline(where);
		if (linenum > 0 && sc_height > 1)
			ap_linenum(PAGE_NUM(linenum));
		else
			ap_quest();
		break;
	case 'D':       /* Final page number */
		/* Find the page number of the last byte in the file (len-1). */
		len = ch_length();
		if (len == NULL_POSITION)
			ap_quest();
		else if (len == 0)
			/* An empty file has no pages. */
			ap_linenum(0);
		else {
			linenum = find_linenum(len - 1);
			if (linenum <= 0)
				ap_quest();
			else
				ap_linenum(PAGE_NUM(linenum));
		}
		break;
#if EDITOR
	case 'E':       /* Editor name */
		ap_str(editor);
		break;
#endif
	case 'f':       /* File name */
		ap_str(get_filename(curr_ifile));
		break;
	case 'i':       /* Index into list of files */
#if TAGS
		if (ntags())
			ap_int(curr_tag());
		else
#endif
			ap_int(get_index(curr_ifile));
		break;
	case 'l':       /* Current line number */
		linenum = currline(where);
		if (linenum != 0)
			ap_linenum(linenum);
		else
			ap_quest();
		break;
	case 'L':       /* Final line number */
		len = ch_length();
		if (len == NULL_POSITION || len == ch_zero() ||
		    (linenum = find_linenum(len)) <= 0)
			ap_quest();
		else
			ap_linenum(linenum-1);
		break;
	case 'm':       /* Number of files */
#if TAGS
		n = ntags();
		if (n)
			ap_int(n);
		else
#endif
			ap_int(nifile());
		break;
	case 'p':       /* Percent into file (bytes) */
		pos = curr_byte(where);
		len = ch_length();
		if (pos != NULL_POSITION && len > 0)
			ap_int(percentage(pos,len));
		else
			ap_quest();
		break;
	case 'P':       /* Percent into file (lines) */
		linenum = currline(where);
		if (linenum == 0 ||
		    (len = ch_length()) == NULL_POSITION || len == ch_zero() ||
		    (last_linenum = find_linenum(len)) <= 0)
			ap_quest();
		else
			ap_int(percentage(linenum, last_linenum));
		break;
	case 's':       /* Size of file */
	case 'B':
		len = ch_length();
		if (len != NULL_POSITION)
			ap_pos(len);
		else
			ap_quest();
		break;
	case 't':       /* Truncate trailing spaces in the message */
		while (mp > message && mp[-1] == ' ')
			mp--;
		break;
	case 'T':       /* Type of list */
#if TAGS
		if (ntags())
			ap_str("tag");
		else
#endif
			ap_str("file");
		break;
	case 'x':       /* Name of next file */
		h = next_ifile(curr_ifile);
		if (h != NULL_IFILE)
			ap_str(get_filename(h));
		else
			ap_quest();
		break;
	}
}

/* Construct a message based on a prototype string. */
static char *
pr_expand(char *proto, int maxwidth)
{
	char *p;
	int c;
	int where;

	mp = message;

	if (*proto == '\0')
		return "";

	for (p = proto;  *p != '\0';  p++) {
		switch (*p) {
		default:        /* Just put the character in the message */
			ap_char(*p);
			break;
		case '\\':      /* Backslash escapes the next character */
			p++;
			ap_char(*p);
			break;
		case '?':       /* Conditional (IF) */
			if ((c = *++p) == '\0')
				--p;
			else {
				where = 0;
				p = wherechar(p, &where);
				if (!cond(c, where))
					p = skipcond(p);
			}
			break;
		case ':':       /* ELSE */
			p = skipcond(p);
			break;
		case '.':       /* ENDIF */
			break;
		case '%':       /* Percent escape */
			if ((c = *++p) == '\0')
				--p;
			else {
				where = 0;
				p = wherechar(p, &where);
				protochar(c, where);
			}
			break;
		}
	}

	if (mp == message)
		return "";
	if (maxwidth > 0 && mp >= message + maxwidth) {
		/* Message is too long.
		 * Return just the final portion of it. */
		return mp - maxwidth;
	}
	return message;
}

/* Insert the line number (of the given position) into the line buffer. */
static void
plinenum(POSITION pos)
{
	LINENUM linenum = 0;
	int i;

	if (linenums == OPT_ONPLUS) {
		/* Get the line number and put it in the current line.
		 * {{ Note: since find_linenum calls forw_raw_line,
		 *    it may seek in the input file, requiring the caller
		 *    of plinenum to re-seek if necessary. }}
		 * {{ Since forw_raw_line modifies linebuf, we must
		 *    do this first, before storing anything in linebuf. }} */
		linenum = find_linenum(pos);
	}

	/* Display a status column if the -J option is set. */
	if (status_col) {
		linebuf[curr] = ' ';
		if (start_attnpos != NULL_POSITION &&
		    pos >= start_attnpos && pos < end_attnpos)
			attr[curr] = AT_NORMAL|AT_HILITE;
		else
			attr[curr] = AT_NORMAL;
		curr++;
		column++;
	}
	/* Display the line number at the start of each line
	 * if the -N option is set. */
	if (linenums == OPT_ONPLUS) {
		char buf[INT_STRLEN_BOUND(pos) + 2];
		int n;

		linenumtoa(linenum, buf);
		n = strlen(buf);
		if (n < MIN_LINENUM_WIDTH)
			n = MIN_LINENUM_WIDTH;
		sprintf(linebuf+curr, "%*s ", n, buf);
		n++;  /* One space after the line number. */
		for (i = 0; i < n; i++)
			attr[curr+i] = AT_NORMAL;
		curr += n;
		column += n;
		lmargin += n;
	}

	/* Append enough spaces to bring us to the lmargin. */
	while (column < lmargin) {
		linebuf[curr] = ' ';
		attr[curr++] = AT_NORMAL;
		column++;
	}
}

/* Support "locale charmap"/nl_langinfo(CODESET) values, as well as others. */
struct cs_alias {
	const char *name;
	const char *oname;
};

static const struct cs_alias cs_aliases[] = {
	{ "UTF-8",              "utf-8" },
	{ "ANSI_X3.4-1968",     "ascii" },
	{ "US-ASCII",           "ascii" },
	{ "latin1",             "iso8859" },
	{ "ISO-8859-1",         "iso8859" },
	{ "latin9",             "iso8859" },
	{ "ISO-8859-15",        "iso8859" },
	{ "latin2",             "iso8859" },
	{ "ISO-8859-2",         "iso8859" },
	{ "ISO-8859-3",         "latin3" },
	{ "latin4",             "iso8859" },
	{ "ISO-8859-4",         "iso8859" },
	{ "cyrillic",           "iso8859" },
	{ "ISO-8859-5",         "iso8859" },
	{ "ISO-8859-6",         "arabic" },
	{ "ISO-8859-7",         "greek" },
	{ "IBM9005",            "greek2005" },
	{ "ISO-8859-8",         "hebrew" },
	{ "latin5",             "iso8859" },
	{ "ISO-8859-9",         "iso8859" },
	{ "latin6",             "iso8859" },
	{ "ISO-8859-10",        "iso8859" },
	{ "latin7",             "iso8859" },
	{ "ISO-8859-13",        "iso8859" },
	{ "latin8",             "iso8859" },
	{ "ISO-8859-14",        "iso8859" },
	{ "latin10",            "iso8859" },
	{ "ISO-8859-16",        "iso8859" },
	{ "IBM437",             "dos" },
	{ "EBCDIC-US",          "ebcdic" },
	{ "IBM1047",            "IBM-1047" },
	{ "KOI8-R",             "koi8-r" },
	{ "KOI8-U",             "koi8-r" },
	{ "GEORGIAN-PS",        "georgianps" },
	{ "TCVN5712-1",         "tcvn" },
	{ "NEXTSTEP",           "next" },
	{ "windows",            "windows-1252" }, /* backward compatibility */
	{ "CP1251",             "windows-1251" },
	{ "CP1252",             "windows-1252" },
	{ "CP1255",             "windows-1255" },
	{ NULL, NULL }
};

#define IS_BINARY_CHAR  01
#define IS_CONTROL_CHAR 02

/* Is a given character a "control" character? */
static int
control_char(int c)
{
	c &= 0377;
	return chardef[c] & IS_CONTROL_CHAR;
}

#if SUPPORT_UTF8
/* Get the length of a UTF-8 character in bytes. */
static int
utf_len(char ch)
{
	/* XXX, may be if (!utf_mode || (ch & 0x80) == 0) ? dzo@ */
	if ((ch & 0x80) == 0)
		return 1;
	if ((ch & 0xE0) == 0xC0)
		return 2;
	if ((ch & 0xF0) == 0xE0)
		return 3;
	if ((ch & 0xF8) == 0xF0)
		return 4;
	if ((ch & 0xFC) == 0xF8)
		return 5;
	if ((ch & 0xFE) == 0xFC)
		return 6;
	/* Invalid UTF-8 encoding. */
	return 1;
}
#else
#define utf_len(ch) 1
#endif

#if SUPPORT_UTF8
/* Get the value of a UTF-8 character. */
static LWCHAR
get_wchar(char *p)
{
	switch (utf_len(p[0])) {
	case 1:
	default:
		/* 0xxxxxxx */
		return (LWCHAR) (p[0] & 0xFF);
	case 2:
		/* 110xxxxx 10xxxxxx */
		return (LWCHAR) (
			((p[0] & 0x1F) << 6) |
			(p[1] & 0x3F));
	case 3:
		/* 1110xxxx 10xxxxxx 10xxxxxx */
		return (LWCHAR) (
			((p[0] & 0x0F) << 12) |
			((p[1] & 0x3F) << 6) |
			(p[2] & 0x3F));
	case 4:
		/* 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx */
		return (LWCHAR) (
			((p[0] & 0x07) << 18) |
			((p[1] & 0x3F) << 12) |
			((p[2] & 0x3F) << 6) |
			(p[3] & 0x3F));
	case 5:
		/* 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx */
		return (LWCHAR) (
			((p[0] & 0x03) << 24) |
			((p[1] & 0x3F) << 18) |
			((p[2] & 0x3F) << 12) |
			((p[3] & 0x3F) << 6) |
			(p[4] & 0x3F));
	case 6:
		/* 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx */
		return (LWCHAR) (
			((p[0] & 0x01) << 30) |
			((p[1] & 0x3F) << 24) |
			((p[2] & 0x3F) << 18) |
			((p[3] & 0x3F) << 12) |
			((p[4] & 0x3F) << 6) |
			(p[5] & 0x3F));
	}
}
#else
static LWCHAR
get_wchar(char *p)
{
	return (LWCHAR) (p[0] & 0xFF);
}
#endif

#if SUPPORT_UTF8
/* Unicode characters data */
struct wchar_range { LWCHAR first, last; };

/* Characters with general category values
 *      Mn: Mark, Nonspacing
 *      Me: Mark, Enclosing
 * Last synched with
 *      <http://www.unicode.org/Public/5.0.0/ucd/UnicodeData-5.0.0d7.txt>
 *      dated 2005-11-30T00:58:48Z */
static const struct wchar_range comp_table[] = {
	{  0x0300,  0x036F} /* Mn */, {  0x0483,  0x0486} /* Mn */,
	{  0x0488,  0x0489} /* Me */,
	{  0x0591,  0x05BD} /* Mn */, {  0x05BF,  0x05BF} /* Mn */,
	{  0x05C1,  0x05C2} /* Mn */, {  0x05C4,  0x05C5} /* Mn */,
	{  0x05C7,  0x05C7} /* Mn */, {  0x0610,  0x0615} /* Mn */,
	{  0x064B,  0x065E} /* Mn */, {  0x0670,  0x0670} /* Mn */,
	{  0x06D6,  0x06DC} /* Mn */,
	{  0x06DE,  0x06DE} /* Me */,
	{  0x06DF,  0x06E4} /* Mn */, {  0x06E7,  0x06E8} /* Mn */,
	{  0x06EA,  0x06ED} /* Mn */, {  0x0711,  0x0711} /* Mn */,
	{  0x0730,  0x074A} /* Mn */, {  0x07A6,  0x07B0} /* Mn */,
	{  0x07EB,  0x07F3} /* Mn */, {  0x0901,  0x0902} /* Mn */,
	{  0x093C,  0x093C} /* Mn */, {  0x0941,  0x0948} /* Mn */,
	{  0x094D,  0x094D} /* Mn */, {  0x0951,  0x0954} /* Mn */,
	{  0x0962,  0x0963} /* Mn */, {  0x0981,  0x0981} /* Mn */,
	{  0x09BC,  0x09BC} /* Mn */, {  0x09C1,  0x09C4} /* Mn */,
	{  0x09CD,  0x09CD} /* Mn */, {  0x09E2,  0x09E3} /* Mn */,
	{  0x0A01,  0x0A02} /* Mn */, {  0x0A3C,  0x0A3C} /* Mn */,
	{  0x0A41,  0x0A42} /* Mn */, {  0x0A47,  0x0A48} /* Mn */,
	{  0x0A4B,  0x0A4D} /* Mn */, {  0x0A70,  0x0A71} /* Mn */,
	{  0x0A81,  0x0A82} /* Mn */, {  0x0ABC,  0x0ABC} /* Mn */,
	{  0x0AC1,  0x0AC5} /* Mn */, {  0x0AC7,  0x0AC8} /* Mn */,
	{  0x0ACD,  0x0ACD} /* Mn */, {  0x0AE2,  0x0AE3} /* Mn */,
	{  0x0B01,  0x0B01} /* Mn */, {  0x0B3C,  0x0B3C} /* Mn */,
	{  0x0B3F,  0x0B3F} /* Mn */, {  0x0B41,  0x0B43} /* Mn */,
	{  0x0B4D,  0x0B4D} /* Mn */, {  0x0B56,  0x0B56} /* Mn */,
	{  0x0B82,  0x0B82} /* Mn */, {  0x0BC0,  0x0BC0} /* Mn */,
	{  0x0BCD,  0x0BCD} /* Mn */, {  0x0C3E,  0x0C40} /* Mn */,
	{  0x0C46,  0x0C48} /* Mn */, {  0x0C4A,  0x0C4D} /* Mn */,
	{  0x0C55,  0x0C56} /* Mn */, {  0x0CBC,  0x0CBC} /* Mn */,
	{  0x0CBF,  0x0CBF} /* Mn */, {  0x0CC6,  0x0CC6} /* Mn */,
	{  0x0CCC,  0x0CCD} /* Mn */, {  0x0CE2,  0x0CE3} /* Mn */,
	{  0x0D41,  0x0D43} /* Mn */, {  0x0D4D,  0x0D4D} /* Mn */,
	{  0x0DCA,  0x0DCA} /* Mn */, {  0x0DD2,  0x0DD4} /* Mn */,
	{  0x0DD6,  0x0DD6} /* Mn */, {  0x0E31,  0x0E31} /* Mn */,
	{  0x0E34,  0x0E3A} /* Mn */, {  0x0E47,  0x0E4E} /* Mn */,
	{  0x0EB1,  0x0EB1} /* Mn */, {  0x0EB4,  0x0EB9} /* Mn */,
	{  0x0EBB,  0x0EBC} /* Mn */, {  0x0EC8,  0x0ECD} /* Mn */,
	{  0x0F18,  0x0F19} /* Mn */, {  0x0F35,  0x0F35} /* Mn */,
	{  0x0F37,  0x0F37} /* Mn */, {  0x0F39,  0x0F39} /* Mn */,
	{  0x0F71,  0x0F7E} /* Mn */, {  0x0F80,  0x0F84} /* Mn */,
	{  0x0F86,  0x0F87} /* Mn */, {  0x0F90,  0x0F97} /* Mn */,
	{  0x0F99,  0x0FBC} /* Mn */, {  0x0FC6,  0x0FC6} /* Mn */,
	{  0x102D,  0x1030} /* Mn */, {  0x1032,  0x1032} /* Mn */,
	{  0x1036,  0x1037} /* Mn */, {  0x1039,  0x1039} /* Mn */,
	{  0x1058,  0x1059} /* Mn */, {  0x135F,  0x135F} /* Mn */,
	{  0x1712,  0x1714} /* Mn */, {  0x1732,  0x1734} /* Mn */,
	{  0x1752,  0x1753} /* Mn */, {  0x1772,  0x1773} /* Mn */,
	{  0x17B7,  0x17BD} /* Mn */, {  0x17C6,  0x17C6} /* Mn */,
	{  0x17C9,  0x17D3} /* Mn */, {  0x17DD,  0x17DD} /* Mn */,
	{  0x180B,  0x180D} /* Mn */, {  0x18A9,  0x18A9} /* Mn */,
	{  0x1920,  0x1922} /* Mn */, {  0x1927,  0x1928} /* Mn */,
	{  0x1932,  0x1932} /* Mn */, {  0x1939,  0x193B} /* Mn */,
	{  0x1A17,  0x1A18} /* Mn */, {  0x1B00,  0x1B03} /* Mn */,
	{  0x1B34,  0x1B34} /* Mn */, {  0x1B36,  0x1B3A} /* Mn */,
	{  0x1B3C,  0x1B3C} /* Mn */, {  0x1B42,  0x1B42} /* Mn */,
	{  0x1B6B,  0x1B73} /* Mn */, {  0x1DC0,  0x1DCA} /* Mn */,
	{  0x1DFE,  0x1DFF} /* Mn */, {  0x20D0,  0x20DC} /* Mn */,
	{  0x20DD,  0x20E0} /* Me */,
	{  0x20E1,  0x20E1} /* Mn */,
	{  0x20E2,  0x20E4} /* Me */,
	{  0x20E5,  0x20EF} /* Mn */, {  0x302A,  0x302F} /* Mn */,
	{  0x3099,  0x309A} /* Mn */, {  0xA806,  0xA806} /* Mn */,
	{  0xA80B,  0xA80B} /* Mn */, {  0xA825,  0xA826} /* Mn */,
	{  0xFB1E,  0xFB1E} /* Mn */, {  0xFE00,  0xFE0F} /* Mn */,
	{  0xFE20,  0xFE23} /* Mn */, { 0x10A01, 0x10A03} /* Mn */,
	{ 0x10A05, 0x10A06} /* Mn */, { 0x10A0C, 0x10A0F} /* Mn */,
	{ 0x10A38, 0x10A3A} /* Mn */, { 0x10A3F, 0x10A3F} /* Mn */,
	{ 0x1D167, 0x1D169} /* Mn */, { 0x1D17B, 0x1D182} /* Mn */,
	{ 0x1D185, 0x1D18B} /* Mn */, { 0x1D1AA, 0x1D1AD} /* Mn */,
	{ 0x1D242, 0x1D244} /* Mn */, { 0xE0100, 0xE01EF} /* Mn */,
};

/* Special pairs, not ranges. */
static struct wchar_range comb_table[] = {
	{0x0644,0x0622}, {0x0644,0x0623}, {0x0644,0x0625}, {0x0644,0x0627}, };

/* Characters with general category values
 *      Cc: Other, Control
 *      Cf: Other, Format
 *      Cs: Other, Surrogate
 *      Co: Other, Private Use
 *      Cn: Other, Not Assigned
 *      Zl: Separator, Line
 *      Zp: Separator, Paragraph
 * Last synched with
 *      <http://www.unicode.org/Public/5.0.0/ucd/UnicodeData-5.0.0d7.txt>
 *      dated 2005-11-30T00:58:48Z */
static struct wchar_range ubin_table[] = {
	{  0x0000,  0x001F} /* Cc */, {  0x007F,  0x009F} /* Cc */,
	{  0x0370,  0x0373} /* Cn */, {  0x0376,  0x0379} /* Cn */,
	{  0x037F,  0x0383} /* Cn */, {  0x038B,  0x038B} /* Cn */,
	{  0x038D,  0x038D} /* Cn */, {  0x03A2,  0x03A2} /* Cn */,
	{  0x03CF,  0x03CF} /* Cn */, {  0x0487,  0x0487} /* Cn */,
	{  0x0514,  0x0530} /* Cn */, {  0x0557,  0x0558} /* Cn */,
	{  0x0560,  0x0560} /* Cn */, {  0x0588,  0x0588} /* Cn */,
	{  0x058B,  0x0590} /* Cn */, {  0x05C8,  0x05CF} /* Cn */,
	{  0x05EB,  0x05EF} /* Cn */, {  0x05F5,  0x05FF} /* Cn */,
	{  0x0604,  0x060A} /* Cn */, {  0x0616,  0x061A} /* Cn */,
	{  0x061C,  0x061D} /* Cn */, {  0x0620,  0x0620} /* Cn */,
	{  0x063B,  0x063F} /* Cn */, {  0x065F,  0x065F} /* Cn */,
	{  0x070E,  0x070E} /* Cn */,
	{  0x074B,  0x074C} /* Cn */, {  0x076E,  0x077F} /* Cn */,
	{  0x07B2,  0x07BF} /* Cn */, {  0x07FB,  0x0900} /* Cn */,
	{  0x093A,  0x093B} /* Cn */, {  0x094E,  0x094F} /* Cn */,
	{  0x0955,  0x0957} /* Cn */, {  0x0971,  0x097A} /* Cn */,
	{  0x0980,  0x0980} /* Cn */, {  0x0984,  0x0984} /* Cn */,
	{  0x098D,  0x098E} /* Cn */, {  0x0991,  0x0992} /* Cn */,
	{  0x09A9,  0x09A9} /* Cn */, {  0x09B1,  0x09B1} /* Cn */,
	{  0x09B3,  0x09B5} /* Cn */, {  0x09BA,  0x09BB} /* Cn */,
	{  0x09C5,  0x09C6} /* Cn */, {  0x09C9,  0x09CA} /* Cn */,
	{  0x09CF,  0x09D6} /* Cn */, {  0x09D8,  0x09DB} /* Cn */,
	{  0x09DE,  0x09DE} /* Cn */, {  0x09E4,  0x09E5} /* Cn */,
	{  0x09FB,  0x0A00} /* Cn */, {  0x0A04,  0x0A04} /* Cn */,
	{  0x0A0B,  0x0A0E} /* Cn */, {  0x0A11,  0x0A12} /* Cn */,
	{  0x0A29,  0x0A29} /* Cn */, {  0x0A31,  0x0A31} /* Cn */,
	{  0x0A34,  0x0A34} /* Cn */, {  0x0A37,  0x0A37} /* Cn */,
	{  0x0A3A,  0x0A3B} /* Cn */, {  0x0A3D,  0x0A3D} /* Cn */,
	{  0x0A43,  0x0A46} /* Cn */, {  0x0A49,  0x0A4A} /* Cn */,
	{  0x0A4E,  0x0A58} /* Cn */, {  0x0A5D,  0x0A5D} /* Cn */,
	{  0x0A5F,  0x0A65} /* Cn */, {  0x0A75,  0x0A80} /* Cn */,
	{  0x0A84,  0x0A84} /* Cn */, {  0x0A8E,  0x0A8E} /* Cn */,
	{  0x0A92,  0x0A92} /* Cn */, {  0x0AA9,  0x0AA9} /* Cn */,
	{  0x0AB1,  0x0AB1} /* Cn */, {  0x0AB4,  0x0AB4} /* Cn */,
	{  0x0ABA,  0x0ABB} /* Cn */, {  0x0AC6,  0x0AC6} /* Cn */,
	{  0x0ACA,  0x0ACA} /* Cn */, {  0x0ACE,  0x0ACF} /* Cn */,
	{  0x0AD1,  0x0ADF} /* Cn */, {  0x0AE4,  0x0AE5} /* Cn */,
	{  0x0AF0,  0x0AF0} /* Cn */, {  0x0AF2,  0x0B00} /* Cn */,
	{  0x0B04,  0x0B04} /* Cn */, {  0x0B0D,  0x0B0E} /* Cn */,
	{  0x0B11,  0x0B12} /* Cn */, {  0x0B29,  0x0B29} /* Cn */,
	{  0x0B31,  0x0B31} /* Cn */, {  0x0B34,  0x0B34} /* Cn */,
	{  0x0B3A,  0x0B3B} /* Cn */, {  0x0B44,  0x0B46} /* Cn */,
	{  0x0B49,  0x0B4A} /* Cn */, {  0x0B4E,  0x0B55} /* Cn */,
	{  0x0B58,  0x0B5B} /* Cn */, {  0x0B5E,  0x0B5E} /* Cn */,
	{  0x0B62,  0x0B65} /* Cn */, {  0x0B72,  0x0B81} /* Cn */,
	{  0x0B84,  0x0B84} /* Cn */, {  0x0B8B,  0x0B8D} /* Cn */,
	{  0x0B91,  0x0B91} /* Cn */, {  0x0B96,  0x0B98} /* Cn */,
	{  0x0B9B,  0x0B9B} /* Cn */, {  0x0B9D,  0x0B9D} /* Cn */,
	{  0x0BA0,  0x0BA2} /* Cn */, {  0x0BA5,  0x0BA7} /* Cn */,
	{  0x0BAB,  0x0BAD} /* Cn */, {  0x0BBA,  0x0BBD} /* Cn */,
	{  0x0BC3,  0x0BC5} /* Cn */, {  0x0BC9,  0x0BC9} /* Cn */,
	{  0x0BCE,  0x0BD6} /* Cn */, {  0x0BD8,  0x0BE5} /* Cn */,
	{  0x0BFB,  0x0C00} /* Cn */, {  0x0C04,  0x0C04} /* Cn */,
	{  0x0C0D,  0x0C0D} /* Cn */, {  0x0C11,  0x0C11} /* Cn */,
	{  0x0C29,  0x0C29} /* Cn */, {  0x0C34,  0x0C34} /* Cn */,
	{  0x0C3A,  0x0C3D} /* Cn */, {  0x0C45,  0x0C45} /* Cn */,
	{  0x0C49,  0x0C49} /* Cn */, {  0x0C4E,  0x0C54} /* Cn */,
	{  0x0C57,  0x0C5F} /* Cn */, {  0x0C62,  0x0C65} /* Cn */,
	{  0x0C70,  0x0C81} /* Cn */, {  0x0C84,  0x0C84} /* Cn */,
	{  0x0C8D,  0x0C8D} /* Cn */, {  0x0C91,  0x0C91} /* Cn */,
	{  0x0CA9,  0x0CA9} /* Cn */, {  0x0CB4,  0x0CB4} /* Cn */,
	{  0x0CBA,  0x0CBB} /* Cn */, {  0x0CC5,  0x0CC5} /* Cn */,
	{  0x0CC9,  0x0CC9} /* Cn */, {  0x0CCE,  0x0CD4} /* Cn */,
	{  0x0CD7,  0x0CDD} /* Cn */, {  0x0CDF,  0x0CDF} /* Cn */,
	{  0x0CE4,  0x0CE5} /* Cn */, {  0x0CF0,  0x0CF0} /* Cn */,
	{  0x0CF3,  0x0D01} /* Cn */, {  0x0D04,  0x0D04} /* Cn */,
	{  0x0D0D,  0x0D0D} /* Cn */, {  0x0D11,  0x0D11} /* Cn */,
	{  0x0D29,  0x0D29} /* Cn */, {  0x0D3A,  0x0D3D} /* Cn */,
	{  0x0D44,  0x0D45} /* Cn */, {  0x0D49,  0x0D49} /* Cn */,
	{  0x0D4E,  0x0D56} /* Cn */, {  0x0D58,  0x0D5F} /* Cn */,
	{  0x0D62,  0x0D65} /* Cn */, {  0x0D70,  0x0D81} /* Cn */,
	{  0x0D84,  0x0D84} /* Cn */, {  0x0D97,  0x0D99} /* Cn */,
	{  0x0DB2,  0x0DB2} /* Cn */, {  0x0DBC,  0x0DBC} /* Cn */,
	{  0x0DBE,  0x0DBF} /* Cn */, {  0x0DC7,  0x0DC9} /* Cn */,
	{  0x0DCB,  0x0DCE} /* Cn */, {  0x0DD5,  0x0DD5} /* Cn */,
	{  0x0DD7,  0x0DD7} /* Cn */, {  0x0DE0,  0x0DF1} /* Cn */,
	{  0x0DF5,  0x0E00} /* Cn */, {  0x0E3B,  0x0E3E} /* Cn */,
	{  0x0E5C,  0x0E80} /* Cn */, {  0x0E83,  0x0E83} /* Cn */,
	{  0x0E85,  0x0E86} /* Cn */, {  0x0E89,  0x0E89} /* Cn */,
	{  0x0E8B,  0x0E8C} /* Cn */, {  0x0E8E,  0x0E93} /* Cn */,
	{  0x0E98,  0x0E98} /* Cn */, {  0x0EA0,  0x0EA0} /* Cn */,
	{  0x0EA4,  0x0EA4} /* Cn */, {  0x0EA6,  0x0EA6} /* Cn */,
	{  0x0EA8,  0x0EA9} /* Cn */, {  0x0EAC,  0x0EAC} /* Cn */,
	{  0x0EBA,  0x0EBA} /* Cn */, {  0x0EBE,  0x0EBF} /* Cn */,
	{  0x0EC5,  0x0EC5} /* Cn */, {  0x0EC7,  0x0EC7} /* Cn */,
	{  0x0ECE,  0x0ECF} /* Cn */, {  0x0EDA,  0x0EDB} /* Cn */,
	{  0x0EDE,  0x0EFF} /* Cn */, {  0x0F48,  0x0F48} /* Cn */,
	{  0x0F6B,  0x0F70} /* Cn */, {  0x0F8C,  0x0F8F} /* Cn */,
	{  0x0F98,  0x0F98} /* Cn */, {  0x0FBD,  0x0FBD} /* Cn */,
	{  0x0FCD,  0x0FCE} /* Cn */, {  0x0FD2,  0x0FFF} /* Cn */,
	{  0x1022,  0x1022} /* Cn */, {  0x1028,  0x1028} /* Cn */,
	{  0x102B,  0x102B} /* Cn */, {  0x1033,  0x1035} /* Cn */,
	{  0x103A,  0x103F} /* Cn */, {  0x105A,  0x109F} /* Cn */,
	{  0x10C6,  0x10CF} /* Cn */, {  0x10FD,  0x10FF} /* Cn */,
	{  0x115A,  0x115E} /* Cn */, {  0x11A3,  0x11A7} /* Cn */,
	{  0x11FA,  0x11FF} /* Cn */, {  0x1249,  0x1249} /* Cn */,
	{  0x124E,  0x124F} /* Cn */, {  0x1257,  0x1257} /* Cn */,
	{  0x1259,  0x1259} /* Cn */, {  0x125E,  0x125F} /* Cn */,
	{  0x1289,  0x1289} /* Cn */, {  0x128E,  0x128F} /* Cn */,
	{  0x12B1,  0x12B1} /* Cn */, {  0x12B6,  0x12B7} /* Cn */,
	{  0x12BF,  0x12BF} /* Cn */, {  0x12C1,  0x12C1} /* Cn */,
	{  0x12C6,  0x12C7} /* Cn */, {  0x12D7,  0x12D7} /* Cn */,
	{  0x1311,  0x1311} /* Cn */, {  0x1316,  0x1317} /* Cn */,
	{  0x135B,  0x135E} /* Cn */, {  0x137D,  0x137F} /* Cn */,
	{  0x139A,  0x139F} /* Cn */, {  0x13F5,  0x1400} /* Cn */,
	{  0x1677,  0x167F} /* Cn */, {  0x169D,  0x169F} /* Cn */,
	{  0x16F1,  0x16FF} /* Cn */, {  0x170D,  0x170D} /* Cn */,
	{  0x1715,  0x171F} /* Cn */, {  0x1737,  0x173F} /* Cn */,
	{  0x1754,  0x175F} /* Cn */, {  0x176D,  0x176D} /* Cn */,
	{  0x1771,  0x1771} /* Cn */, {  0x1774,  0x177F} /* Cn */,
	{  0x17DE,  0x17DF} /* Cn */, {  0x17EA,  0x17EF} /* Cn */,
	{  0x17FA,  0x17FF} /* Cn */, {  0x180F,  0x180F} /* Cn */,
	{  0x181A,  0x181F} /* Cn */, {  0x1878,  0x187F} /* Cn */,
	{  0x18AA,  0x18FF} /* Cn */, {  0x191D,  0x191F} /* Cn */,
	{  0x192C,  0x192F} /* Cn */, {  0x193C,  0x193F} /* Cn */,
	{  0x1941,  0x1943} /* Cn */, {  0x196E,  0x196F} /* Cn */,
	{  0x1975,  0x197F} /* Cn */, {  0x19AA,  0x19AF} /* Cn */,
	{  0x19CA,  0x19CF} /* Cn */, {  0x19DA,  0x19DD} /* Cn */,
	{  0x1A1C,  0x1A1D} /* Cn */, {  0x1A20,  0x1AFF} /* Cn */,
	{  0x1B4C,  0x1B4F} /* Cn */, {  0x1B7D,  0x1CFF} /* Cn */,
	{  0x1DCB,  0x1DFD} /* Cn */, {  0x1E9C,  0x1E9F} /* Cn */,
	{  0x1EFA,  0x1EFF} /* Cn */, {  0x1F16,  0x1F17} /* Cn */,
	{  0x1F1E,  0x1F1F} /* Cn */, {  0x1F46,  0x1F47} /* Cn */,
	{  0x1F4E,  0x1F4F} /* Cn */, {  0x1F58,  0x1F58} /* Cn */,
	{  0x1F5A,  0x1F5A} /* Cn */, {  0x1F5C,  0x1F5C} /* Cn */,
	{  0x1F5E,  0x1F5E} /* Cn */, {  0x1F7E,  0x1F7F} /* Cn */,
	{  0x1FB5,  0x1FB5} /* Cn */, {  0x1FC5,  0x1FC5} /* Cn */,
	{  0x1FD4,  0x1FD5} /* Cn */, {  0x1FDC,  0x1FDC} /* Cn */,
	{  0x1FF0,  0x1FF1} /* Cn */, {  0x1FF5,  0x1FF5} /* Cn */,
	{  0x1FFF,  0x1FFF} /* Cn */,
	{  0x200B,  0x200F} /* Cf */,
	{  0x2028,  0x2028} /* Zl */,
	{  0x2029,  0x2029} /* Zp */,
	{  0x202A,  0x202E} /* Cf */,
	{  0x2060,  0x2063} /* Cf */,
	{  0x2064,  0x2069} /* Cn */,
	{  0x206A,  0x206F} /* Cf */,
	{  0x2072,  0x2073} /* Cn */, {  0x208F,  0x208F} /* Cn */,
	{  0x2095,  0x209F} /* Cn */, {  0x20B6,  0x20CF} /* Cn */,
	{  0x20F0,  0x20FF} /* Cn */, {  0x214F,  0x2152} /* Cn */,
	{  0x2185,  0x218F} /* Cn */, {  0x23E8,  0x23FF} /* Cn */,
	{  0x2427,  0x243F} /* Cn */, {  0x244B,  0x245F} /* Cn */,
	{  0x269D,  0x269F} /* Cn */, {  0x26B3,  0x2700} /* Cn */,
	{  0x2705,  0x2705} /* Cn */, {  0x270A,  0x270B} /* Cn */,
	{  0x2728,  0x2728} /* Cn */, {  0x274C,  0x274C} /* Cn */,
	{  0x274E,  0x274E} /* Cn */, {  0x2753,  0x2755} /* Cn */,
	{  0x2757,  0x2757} /* Cn */, {  0x275F,  0x2760} /* Cn */,
	{  0x2795,  0x2797} /* Cn */, {  0x27B0,  0x27B0} /* Cn */,
	{  0x27BF,  0x27BF} /* Cn */, {  0x27CB,  0x27CF} /* Cn */,
	{  0x27EC,  0x27EF} /* Cn */, {  0x2B1B,  0x2B1F} /* Cn */,
	{  0x2B24,  0x2BFF} /* Cn */, {  0x2C2F,  0x2C2F} /* Cn */,
	{  0x2C5F,  0x2C5F} /* Cn */, {  0x2C6D,  0x2C73} /* Cn */,
	{  0x2C78,  0x2C7F} /* Cn */, {  0x2CEB,  0x2CF8} /* Cn */,
	{  0x2D26,  0x2D2F} /* Cn */, {  0x2D66,  0x2D6E} /* Cn */,
	{  0x2D70,  0x2D7F} /* Cn */, {  0x2D97,  0x2D9F} /* Cn */,
	{  0x2DA7,  0x2DA7} /* Cn */, {  0x2DAF,  0x2DAF} /* Cn */,
	{  0x2DB7,  0x2DB7} /* Cn */, {  0x2DBF,  0x2DBF} /* Cn */,
	{  0x2DC7,  0x2DC7} /* Cn */, {  0x2DCF,  0x2DCF} /* Cn */,
	{  0x2DD7,  0x2DD7} /* Cn */, {  0x2DDF,  0x2DFF} /* Cn */,
	{  0x2E18,  0x2E1B} /* Cn */, {  0x2E1E,  0x2E7F} /* Cn */,
	{  0x2E9A,  0x2E9A} /* Cn */, {  0x2EF4,  0x2EFF} /* Cn */,
	{  0x2FD6,  0x2FEF} /* Cn */, {  0x2FFC,  0x2FFF} /* Cn */,
	{  0x3040,  0x3040} /* Cn */, {  0x3097,  0x3098} /* Cn */,
	{  0x3100,  0x3104} /* Cn */, {  0x312D,  0x3130} /* Cn */,
	{  0x318F,  0x318F} /* Cn */, {  0x31B8,  0x31BF} /* Cn */,
	{  0x31D0,  0x31EF} /* Cn */, {  0x321F,  0x321F} /* Cn */,
	{  0x3244,  0x324F} /* Cn */, {  0x32FF,  0x32FF} /* Cn */,
	{  0x4DB6,  0x4DBF} /* Cn */, {  0x9FBC,  0x9FFF} /* Cn */,
	{  0xA48D,  0xA48F} /* Cn */, {  0xA4C7,  0xA6FF} /* Cn */,
	{  0xA71B,  0xA71F} /* Cn */, {  0xA722,  0xA7FF} /* Cn */,
	{  0xA82C,  0xA83F} /* Cn */, {  0xA878,  0xABFF} /* Cn */,
	{  0xD7A4,  0xD7FF} /* Cn */,
	{  0xD800,  0xDFFF} /* Cs */,
	{  0xE000,  0xF8FF} /* Co */,
	{  0xFA2E,  0xFA2F} /* Cn */, {  0xFA6B,  0xFA6F} /* Cn */,
	{  0xFADA,  0xFAFF} /* Cn */, {  0xFB07,  0xFB12} /* Cn */,
	{  0xFB18,  0xFB1C} /* Cn */, {  0xFB37,  0xFB37} /* Cn */,
	{  0xFB3D,  0xFB3D} /* Cn */, {  0xFB3F,  0xFB3F} /* Cn */,
	{  0xFB42,  0xFB42} /* Cn */, {  0xFB45,  0xFB45} /* Cn */,
	{  0xFBB2,  0xFBD2} /* Cn */, {  0xFD40,  0xFD4F} /* Cn */,
	{  0xFD90,  0xFD91} /* Cn */, {  0xFDC8,  0xFDEF} /* Cn */,
	{  0xFDFE,  0xFDFF} /* Cn */, {  0xFE1A,  0xFE1F} /* Cn */,
	{  0xFE24,  0xFE2F} /* Cn */, {  0xFE53,  0xFE53} /* Cn */,
	{  0xFE67,  0xFE67} /* Cn */, {  0xFE6C,  0xFE6F} /* Cn */,
	{  0xFE75,  0xFE75} /* Cn */, {  0xFEFD,  0xFEFE} /* Cn */,
	{  0xFEFF,  0xFEFF} /* Cf */,
	{  0xFF00,  0xFF00} /* Cn */, {  0xFFBF,  0xFFC1} /* Cn */,
	{  0xFFC8,  0xFFC9} /* Cn */, {  0xFFD0,  0xFFD1} /* Cn */,
	{  0xFFD8,  0xFFD9} /* Cn */, {  0xFFDD,  0xFFDF} /* Cn */,
	{  0xFFE7,  0xFFE7} /* Cn */, {  0xFFEF,  0xFFF8} /* Cn */,
	{  0xFFF9,  0xFFFB} /* Cf */,
	{  0xFFFE,  0xFFFF} /* Cn */, { 0x1000C, 0x1000C} /* Cn */,
	{ 0x10027, 0x10027} /* Cn */, { 0x1003B, 0x1003B} /* Cn */,
	{ 0x1003E, 0x1003E} /* Cn */, { 0x1004E, 0x1004F} /* Cn */,
	{ 0x1005E, 0x1007F} /* Cn */, { 0x100FB, 0x100FF} /* Cn */,
	{ 0x10103, 0x10106} /* Cn */, { 0x10134, 0x10136} /* Cn */,
	{ 0x1018B, 0x102FF} /* Cn */, { 0x1031F, 0x1031F} /* Cn */,
	{ 0x10324, 0x1032F} /* Cn */, { 0x1034B, 0x1037F} /* Cn */,
	{ 0x1039E, 0x1039E} /* Cn */, { 0x103C4, 0x103C7} /* Cn */,
	{ 0x103D6, 0x103FF} /* Cn */,
	{ 0x1049E, 0x1049F} /* Cn */, { 0x104AA, 0x107FF} /* Cn */,
	{ 0x10806, 0x10807} /* Cn */, { 0x10809, 0x10809} /* Cn */,
	{ 0x10836, 0x10836} /* Cn */, { 0x10839, 0x1083B} /* Cn */,
	{ 0x1083D, 0x1083E} /* Cn */, { 0x10840, 0x108FF} /* Cn */,
	{ 0x1091A, 0x1091E} /* Cn */, { 0x10920, 0x109FF} /* Cn */,
	{ 0x10A04, 0x10A04} /* Cn */, { 0x10A07, 0x10A0B} /* Cn */,
	{ 0x10A14, 0x10A14} /* Cn */, { 0x10A18, 0x10A18} /* Cn */,
	{ 0x10A34, 0x10A37} /* Cn */, { 0x10A3B, 0x10A3E} /* Cn */,
	{ 0x10A48, 0x10A4F} /* Cn */, { 0x10A59, 0x11FFF} /* Cn */,
	{ 0x1236F, 0x123FF} /* Cn */, { 0x12463, 0x1246F} /* Cn */,
	{ 0x12474, 0x1CFFF} /* Cn */, { 0x1D0F6, 0x1D0FF} /* Cn */,
	{ 0x1D127, 0x1D129} /* Cn */,
	{ 0x1D173, 0x1D17A} /* Cf */,
	{ 0x1D1DE, 0x1D1FF} /* Cn */, { 0x1D246, 0x1D2FF} /* Cn */,
	{ 0x1D357, 0x1D35F} /* Cn */, { 0x1D372, 0x1D3FF} /* Cn */,
	{ 0x1D455, 0x1D455} /* Cn */, { 0x1D49D, 0x1D49D} /* Cn */,
	{ 0x1D4A0, 0x1D4A1} /* Cn */, { 0x1D4A3, 0x1D4A4} /* Cn */,
	{ 0x1D4A7, 0x1D4A8} /* Cn */, { 0x1D4AD, 0x1D4AD} /* Cn */,
	{ 0x1D4BA, 0x1D4BA} /* Cn */, { 0x1D4BC, 0x1D4BC} /* Cn */,
	{ 0x1D4C4, 0x1D4C4} /* Cn */, { 0x1D506, 0x1D506} /* Cn */,
	{ 0x1D50B, 0x1D50C} /* Cn */, { 0x1D515, 0x1D515} /* Cn */,
	{ 0x1D51D, 0x1D51D} /* Cn */, { 0x1D53A, 0x1D53A} /* Cn */,
	{ 0x1D53F, 0x1D53F} /* Cn */, { 0x1D545, 0x1D545} /* Cn */,
	{ 0x1D547, 0x1D549} /* Cn */, { 0x1D551, 0x1D551} /* Cn */,
	{ 0x1D6A6, 0x1D6A7} /* Cn */, { 0x1D7CC, 0x1D7CD} /* Cn */,
	{ 0x1D800, 0x1FFFF} /* Cn */, { 0x2A6D7, 0x2F7FF} /* Cn */,
	{ 0x2FA1E, 0xE0000} /* Cn */,
	{ 0xE0001, 0xE0001} /* Cf */,
	{ 0xE0002, 0xE001F} /* Cn */,
	{ 0xE0020, 0xE007F} /* Cf */,
	{ 0xE0080, 0xE00FF} /* Cn */, { 0xE01F0, 0xEFFFF} /* Cn */,
	{ 0xF0000, 0xFFFFD} /* Co */,
	{ 0xFFFFE, 0xFFFFF} /* Cn */,
	{0x100000,0x10FFFD} /* Co */,
	{0x10FFFE,0x10FFFF} /* Cn */,
	{0x110000,0x7FFFFFFF} /* ISO 10646?? */
};

/* Double width characters
 *      W: East Asian Wide
 *      F: East Asian Full-width
 * Unassigned code points may be included when they allow ranges to be merged.
 * Last synched with
 *      <http://www.unicode.org/Public/5.0.0/ucd/EastAsianWidth-5.0.0d2.txt>
 *      dated 2005-11-08T01:32:56Z */
static const struct wchar_range wide_table[] = {
	{  0x1100,  0x115F} /* W */, {  0x2329,  0x232A} /* W */,
	{  0x2E80,  0x2FFB} /* W */,
	{  0x3000,  0x3000} /* F */,
	{  0x3001,  0x303E} /* W */, {  0x3041,  0x4DB5} /* W */,
	{  0x4E00,  0x9FBB} /* W */, {  0xA000,  0xA4C6} /* W */,
	{  0xAC00,  0xD7A3} /* W */, {  0xF900,  0xFAD9} /* W */,
	{  0xFE10,  0xFE19} /* W */, {  0xFE30,  0xFE6B} /* W */,
	{  0xFF01,  0xFF60} /* F */, {  0xFFE0,  0xFFE6} /* F */,
	{ 0x20000, 0x2FFFD} /* W */, { 0x30000, 0x3FFFD} /* W */,
};

static int
is_in_table(LWCHAR ch, const struct wchar_range table[], int tsize)
{
	int hi;
	int lo;

	/* Binary search in the table. */
	if (ch < table[0].first)
		return 0;
	lo = 0;
	hi = tsize - 1;
	while (lo <= hi) {
		int mid = (lo + hi) / 2;
		if (ch > table[mid].last)
			lo = mid + 1;
		else if (ch < table[mid].first)
			hi = mid - 1;
		else
			return 1;
	}
	return 0;
}


/* Should this UTF-8 character be treated as binary? */
static int
is_ubin_char(LWCHAR ch)
{
	return is_in_table(ch, ubin_table, (sizeof(ubin_table) / sizeof(*ubin_table)));
}

/* Is a character a UTF-8 composing character?
 * If a composing character follows any char, the two combine into one glyph. */
static int
is_composing_char(LWCHAR ch)
{
	return is_in_table(ch, comp_table, (sizeof(comp_table) / sizeof(*comp_table)));
}

/* Is this a double width UTF-8 character? */
static int
is_wide_char(LWCHAR ch)
{
	return is_in_table(ch, wide_table, (sizeof(wide_table) / sizeof(*wide_table)));
}

/* Is a character a UTF-8 combining character?
 * A combining char acts like an ordinary char, but if it follows
 * a specific char (not any char), the two combine into one glyph. */
static int
is_combining_char(LWCHAR ch1, LWCHAR ch2)
{
	/* The table is small; use linear search. */
	size_t i;
	for (i = 0;  i < sizeof(comb_table)/sizeof(*comb_table);  i++) {
		if (ch1 == comb_table[i].first &&
		    ch2 == comb_table[i].last)
			return 1;
	}
	return 0;
}
#else
static int is_ubin_char(LWCHAR ch)
{
	return (ch <= 0x1f || ch == 0x7f || ch == 0233);
}
#define is_composing_char(ch) 0
#define is_wide_char(ch) 0
#define is_combining_char(ch1, ch2) 0
#endif /* SUPPORT_UTF8 */

static int
apply_at_specials(int sp_attr)
{
	if (sp_attr & AT_BINARY)
		sp_attr |= binattr;
	if (sp_attr & AT_HILITE)
		sp_attr |= AT_STANDOUT;
	sp_attr &= ~(AT_BINARY|AT_HILITE);

	return sp_attr;
}

/* Is a character the end of an ANSI escape sequence? */
static int
is_ansi_end(LWCHAR ch)
{
	if (!is_ascii_char(ch))
		return 0;
	return (strchr(end_ansi_chars, (char) ch) != NULL);
}

static int
is_ansi_middle(LWCHAR ch)
{
	if (!is_ascii_char(ch))
		return 0;
	if (is_ansi_end(ch))
		return 0;
	return (strchr(mid_ansi_chars, (char) ch) != NULL);
}

/* Shift the input line left.
 * This means discarding N printable chars at the start of the buffer. */
static void
pshift(int shift)
{
#if SUPPORT_UTF8
	LWCHAR prev_ch = 0;
#endif
	unsigned char c;
	int shifted = 0;
	int to;
	int from;
	int len;
	int width;

	if (shift > column - lmargin)
		shift = column - lmargin;
	if (shift > curr - lmargin)
		shift = curr - lmargin;

	to = from = lmargin;
	/* We keep on going when shifted == shift
	 * to get all combining chars.*/
	while (shifted <= shift && from < curr) {
		c = linebuf[from];
		if (ctldisp == OPT_ONPLUS && IS_CSI_START(c)) {
			/* Keep cumulative effect. */
			linebuf[to] = c;
			attr[to++] = attr[from++];
			while (from < curr && linebuf[from]) {
				linebuf[to] = linebuf[from];
				attr[to++] = attr[from];
				if (!is_ansi_middle(linebuf[from++]))
					break;
			}
			continue;
		}

		width = 0;
#if SUPPORT_UTF8
		if (!IS_ASCII_OCTET(c) && utf_mode) {
			/* Assumes well-formedness validation already done. */
			LWCHAR ch;

			len = utf_len(c);
			if (from + len > curr)
				break;
			ch = get_wchar(linebuf + from);
			if (!is_composing_char(ch) && !is_combining_char(prev_ch, ch))
				width = is_wide_char(ch) ? 2 : 1;
			prev_ch = ch;
		} else
#endif
		{
			len = 1;
			if (c == '\b')
				/* XXX - Incorrect if several '\b' in a row. */
#if SUPPORT_UTF8
				width = (utf_mode && is_wide_char(prev_ch)) ? -2 : -1;
#else
				width = -1;
#endif
			else if (!control_char(c))
				width = 1;
#if SUPPORT_UTF8
			prev_ch = 0;
#endif
		}

		if (width == 2 && shift - shifted == 1) {
			/* Should never happen when called by pshift_all(). */
			attr[to] = attr[from];
			/* Assume a wide_char will never be the first half of a
			 * combining_char pair, so reset prev_ch in case we're
			 * followed by a '\b'. */
#if SUPPORT_UTF8
			prev_ch = linebuf[to++] = ' ';
#else
			linebuf[to++] = ' ';
#endif
			from += len;
			shifted++;
			continue;
		}

		if (shift - shifted < width)
			break;
		from += len;
		shifted += width;
		if (shifted < 0)
			shifted = 0;
	}
	while (from < curr) {
		linebuf[to] = linebuf[from];
		attr[to++] = attr[from++];
	}
	curr = to;
	column -= shifted;
	cshift += shifted;
}

static void
pshift_all(void)
{
	pshift(column);
}


/* Return the printing width of a given character and attribute,
 * if the character were added to the current position in the line buffer.
 * Adding a character with a given attribute may cause an enter or exit
 * attribute sequence to be inserted, so this must be taken into account. */
static int
#if SUPPORT_UTF8
pwidth(LWCHAR ch, LWCHAR prev_ch)
#define PWIDTH(ch, prev_ch) pwidth(ch, prev_ch)
#else
pwidth(LWCHAR ch)
#define PWIDTH(ch, prev_ch) pwidth(ch)
#endif
{
	int w;

	if (ch == '\b')
		/* Backspace moves backwards one or two positions.
		 * XXX - Incorrect if several '\b' in a row. */
		return (utf_mode && is_wide_char(prev_ch)) ? -2 : -1;

	if (!utf_mode || is_ascii_char(ch)) {
		if (control_char((char)ch)) {
			/* Control characters do unpredictable things,
			 * so we don't even try to guess; say it doesn't move.
			 * This can only happen if the -r flag is in effect. */
			return 0;
		}
	} else {
		if (is_composing_char(ch) || is_combining_char(prev_ch, ch)) {
			/* Composing and combining chars take up no space.
			 *
			 * Some terminals, upon failure to compose a
			 * composing character with the character(s) that
			 * precede(s) it will actually take up one column
			 * for the composing character; there isn't much
			 * we could do short of testing the (complex)
			 * composition process ourselves and printing
			 * a binary representation when it fails. */
			return 0;
		}
	}

	/* Other characters take one or two columns */
	w = 1;
	if (is_wide_char(ch))
		w++;
	return w;
}

/* Step forward or backward one character in a string. */
static LWCHAR
step_char(const char **pp, signed int dir, const char *limit)
{
	LWCHAR ch;
	const char *p = *pp;

	if (!utf_mode) {
		/* It's easy if chars are one byte. */
		if (dir > 0)
			ch = (LWCHAR) ((p < limit) ? *p++ : 0);
		else
			ch = (LWCHAR) ((p > limit) ? *--p : 0);
		ch = ch & 0xFF;
	}
#if SUPPORT_UTF8
	else if (dir > 0) {
		int len = utf_len(*p);
		if (p + len > limit) {
			ch = 0;
			p = limit;
		} else {
			ch = get_wchar(p);
			p += len;
		}
	} else {
		while (p > limit && IS_UTF8_TRAIL(p[-1]))
			p--;
		if (p > limit)
			ch = get_wchar(--p);
		else
			ch = 0;
	}
#endif
	*pp = p;
	return ch;
}

/* Delete to the previous base character in the line buffer.
 * Return 1 if one is found. */
static int
backc(void)
{
	LWCHAR prev_ch;
	const char *p = linebuf + curr;
	LWCHAR ch = step_char(&p, -1, linebuf + lmargin);
	int width;

	/* This assumes that there is no '\b' in linebuf. */
	while (curr > lmargin
	       && column > lmargin
	       && (!(attr[curr - 1] & (AT_ANSI|AT_BINARY)))) {
		curr = p - linebuf;
		prev_ch = step_char(&p, -1, linebuf + lmargin);
		width = PWIDTH(ch, prev_ch);
		column -= width;
		if (width > 0)
			return 1;
		ch = prev_ch;
	}

	return 0;
}

/* Are we currently within a recognized ANSI escape sequence? */
static int
in_ansi_esc_seq(void)
{
	const char *p;

	/* Search backwards for either an ESC (which means we ARE in a seq);
	 * or an end char (which means we're NOT in a seq). */
	for (p = &linebuf[curr];  p > linebuf; ) {
		LWCHAR ch = step_char(&p, -1, linebuf);
		if (IS_CSI_START(ch))
			return 1;
		if (!is_ansi_middle(ch))
			return 0;
	}
	return 0;
}


#define CVT_TO_LC       01      /* Convert upper-case to lower-case */
#define CVT_BS          02      /* Do backspace processing */
#define CVT_CRLF        04      /* Remove CR after LF */
#define CVT_ANSI        010     /* Remove ANSI escape sequences */

#if HILITE_SEARCH

/* Should any characters in a specified range be highlighted? */
static int
is_hilited_range(POSITION pos, POSITION epos)
{
	struct hilite *hl;

	/* Look at each highlight and see if any part of it falls in the range. */
	for (hl = hilite_anchor.hl_first;  hl != NULL;  hl = hl->hl_next) {
		if (hl->hl_endpos > pos &&
		    (epos == NULL_POSITION || epos > hl->hl_startpos))
			return 1;
	}
	return 0;
}

#define MINPOS(a,b)     (((a) < (b)) ? (a) : (b))
#define MAXPOS(a,b)     (((a) > (b)) ? (a) : (b))

/* Add a new hilite to a hilite list. */
static void
add_hilite(struct hilite *h_anchor, struct hilite *hl)
{
	struct hilite *ihl;

	/* Hilites are sorted in the list; find where new one belongs.
	 * Insert new one after ihl. */
	for (ihl = h_anchor;  ihl->hl_next != NULL;  ihl = ihl->hl_next) {
		if (ihl->hl_next->hl_startpos > hl->hl_startpos)
			break;
	}

	/* Truncate hilite so it doesn't overlap any existing ones
	 * above and below it. */
	if (ihl != h_anchor)
		hl->hl_startpos = MAXPOS(hl->hl_startpos, ihl->hl_endpos);
	if (ihl->hl_next != NULL)
		hl->hl_endpos = MINPOS(hl->hl_endpos, ihl->hl_next->hl_startpos);
	if (hl->hl_startpos >= hl->hl_endpos) {
		/* Hilite was truncated out of existence. */
		free(hl);
		return;
	}
	hl->hl_next = ihl->hl_next;
	ihl->hl_next = hl;
}
#endif /* HILITE_SEARCH */

/* Store a character into a UTF-8 string. */
static void
put_wchar(char **pp, LWCHAR ch)
{
#if SUPPORT_UTF8
	if (!utf_mode || ch < 0x80) {
		/* 0xxxxxxx */
		*(*pp)++ = (char) ch;
	} else if (ch < 0x800) {
		/* 110xxxxx 10xxxxxx */
		*(*pp)++ = (char) (0xC0 | ((ch >> 6) & 0x1F));
		*(*pp)++ = (char) (0x80 | (ch & 0x3F));
	} else if (ch < 0x10000) {
		/* 1110xxxx 10xxxxxx 10xxxxxx */
		*(*pp)++ = (char) (0xE0 | ((ch >> 12) & 0x0F));
		*(*pp)++ = (char) (0x80 | ((ch >> 6) & 0x3F));
		*(*pp)++ = (char) (0x80 | (ch & 0x3F));
	} else if (ch < 0x200000) {
		/* 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx */
		*(*pp)++ = (char) (0xF0 | ((ch >> 18) & 0x07));
		*(*pp)++ = (char) (0x80 | ((ch >> 12) & 0x3F));
		*(*pp)++ = (char) (0x80 | ((ch >> 6) & 0x3F));
		*(*pp)++ = (char) (0x80 | (ch & 0x3F));
	} else if (ch < 0x4000000) {
		/* 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx */
		*(*pp)++ = (char) (0xF0 | ((ch >> 24) & 0x03));
		*(*pp)++ = (char) (0x80 | ((ch >> 18) & 0x3F));
		*(*pp)++ = (char) (0x80 | ((ch >> 12) & 0x3F));
		*(*pp)++ = (char) (0x80 | ((ch >> 6) & 0x3F));
		*(*pp)++ = (char) (0x80 | (ch & 0x3F));
	} else {
		/* 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx */
		*(*pp)++ = (char) (0xF0 | ((ch >> 30) & 0x01));
		*(*pp)++ = (char) (0x80 | ((ch >> 24) & 0x3F));
		*(*pp)++ = (char) (0x80 | ((ch >> 18) & 0x3F));
		*(*pp)++ = (char) (0x80 | ((ch >> 12) & 0x3F));
		*(*pp)++ = (char) (0x80 | ((ch >> 6) & 0x3F));
		*(*pp)++ = (char) (0x80 | (ch & 0x3F));
	}
#else
	*(*pp)++ = (char) ch;
#endif
}

#if HILITE_SEARCH
/* Adjust hl_startpos & hl_endpos to account for processing by cvt_text. */
static void
adj_hilite(struct hilite *h_anchor, POSITION linepos, int cvt_ops)
{
	const char *line;
	const char *oline;
	int line_len;
	const char *line_end;
	struct hilite *hl;
	int checkstart;
	POSITION opos;
	POSITION npos;
	LWCHAR ch;
	int ncwidth;

	/* The line was already scanned and hilites were added (in hilite_line).
	 * But it was assumed that each char position in the line
	 * correponds to one char position in the file.
	 * This may not be true if cvt_text modified the line.
	 * Get the raw line again.  Look at each character. */
	(void) forw_raw_line(linepos, &line, &line_len);
	line_end = line + line_len;
	opos = npos = linepos;
	hl = h_anchor->hl_first;
	checkstart = TRUE;
	while (hl != NULL) {
		/* See if we need to adjust the current hl_startpos or
		 * hl_endpos.  After adjusting startpos[i], move to endpos[i].
		 * After adjusting endpos[i], move to startpos[i+1].
		 * The hilite list must be sorted thus:
		 * startpos[0] < endpos[0] <= startpos[1] < endpos[1] <= etc. */
		if (checkstart && hl->hl_startpos == opos) {
			hl->hl_startpos = npos;
			checkstart = FALSE;
			continue; /* {{ not really necessary }} */
		} else if (!checkstart && hl->hl_endpos == opos) {
			hl->hl_endpos = npos;
			checkstart = TRUE;
			hl = hl->hl_next;
			continue; /* {{ necessary }} */
		}
		if (line == line_end)
			break;

		/* Get the next char from the line. */
		oline = line;
		ch = step_char(&line, +1, line_end);
		ncwidth = line - oline;
		npos += ncwidth;

		/* Figure out how this char was processed by cvt_text. */
		if ((cvt_ops & CVT_BS) && ch == '\b') {
			/* Skip the backspace and the following char. */
			oline = line;
			ch = step_char(&line, +1, line_end);
			ncwidth = line - oline;
			npos += ncwidth;
		} else if ((cvt_ops & CVT_TO_LC) && IS_UPPER(ch)) {
			/* Converted uppercase to lower.
			 * Note that this may have changed the number of bytes
			 * that the character occupies. */
			char dbuf[6];
			char *dst = dbuf;
			put_wchar(&dst, TO_LOWER(ch));
			opos += dst - dbuf;
		} else if ((cvt_ops & CVT_ANSI) && IS_CSI_START(ch)) {
			/* Skip to end of ANSI escape sequence. */
			line++;  /* skip the CSI start char */
			npos++;
			while (line < line_end) {
				npos++;
				if (!is_ansi_middle(*line++))
					break;
			}
		} else {
			/* Ordinary unprocessed character. */
			opos += ncwidth;
		}
	}
}

/* Make a hilite for each string in a physical line which matches
 * the current pattern.
 * sp,ep delimit the first match already found. */
static void
hilite_line(POSITION linepos, char *line, int line_len, char *sp, char *ep,
								int cvt_ops)
{
	char *searchp;
	char *line_end = line + line_len;
	struct hilite *hl;
	struct hilite hilites;

	if (sp == NULL || ep == NULL)
		return;
	/* sp and ep delimit the first match in the line.
	 * Mark the corresponding file positions, then
	 * look for further matches and mark them.
	 * {{ This technique, of calling match_pattern on subsequent
	 *    substrings of the line, may mark more than is correct
	 *    if the pattern starts with "^".  This bug is fixed
	 *    for those regex functions that accept a notbol parameter
	 *    (currently POSIX, PCRE and V8-with-regexec2). }} */
	searchp = line;
	/* Put the hilites into a temporary list until they're adjusted. */
	hilites.hl_first = NULL;
	do {
		if (ep > sp) {
			/* Assume that each char position in the "line"
			 * buffer corresponds to one char position in the file.
			 * This is not quite true; we need to adjust later. */
			hl = emalloc(sizeof(struct hilite));
			hl->hl_startpos = linepos + (sp-line);
			hl->hl_endpos = linepos + (ep-line);
			add_hilite(&hilites, hl);
		}
		/* If we matched more than zero characters,
		 * move to the first char after the string we matched.
		 * If we matched zero, just move to the next char. */
		if (ep > searchp)
			searchp = ep;
		else if (searchp != line_end)
			searchp++;
		else /* end of line */
			break;
	} while (MATCH_PATTERN(searchp, line_end - searchp, &sp, &ep, 1));

	/* If there were backspaces in the original line, they
	 * were removed, and hl_startpos/hl_endpos are not correct.
	 * {{ This is very ugly. }} */
	adj_hilite(&hilites, linepos, cvt_ops);

	/* Now put the hilites into the real list. */
	while ((hl = hilites.hl_next) != NULL) {
		hilites.hl_next = hl->hl_next;
		add_hilite(&hilite_anchor, hl);
	}
}

/* Should any characters in a specified range be highlighted?
 * If nohide is nonzero, don't consider hide_hilite. */
static int
is_hilited(POSITION pos, POSITION epos, int nohide, int *p_matches)
{
	int match;

	if (p_matches != NULL)
		*p_matches = 0;

	if (!status_col &&
	    start_attnpos != NULL_POSITION &&
	    pos < end_attnpos &&
	     (epos == NULL_POSITION || epos > start_attnpos))
		/* The attn line overlaps this range. */
		return 1;

	match = is_hilited_range(pos, epos);
	if (!match)
		return 0;

	if (p_matches != NULL)
		/* Report matches, even if we're hiding highlights. */
		*p_matches = 1;

	if (hilite_search == 0)
		/* Not doing highlighting. */
		return 0;

	if (!nohide && hide_hilite)
		/* Highlighting is hidden. */
		return 0;

	return 1;
}
#endif

/* Append a character and attribute to the line buffer. */
#if HILITE_SEARCH
#define STORE_CHAR(ch,a,rep,pos) \
	do { \
		if (store_char((ch),(a),(rep),(pos))) return 1; \
	} while (0)
#else
#define STORE_CHAR(ch,a,rep,pos) \
	do { \
		if (store_char((ch),(a),(rep))) return 1; \
	} while (0)
#endif

#if HILITE_SEARCH
static int
store_char(LWCHAR ch, int a, char *rep, POSITION pos)
#else
static int
store_char(LWCHAR ch, int a, char *rep)
#endif
{
	int w;
	int replen;
	char cs;

	w = (a & (AT_UNDERLINE|AT_BOLD));       /* Pre-use w. */
	if (w != AT_NORMAL)
		last_overstrike = w;

#if HILITE_SEARCH
	{
		int matches;
		if (is_hilited(pos, pos+1, 0, &matches)) {
			/* This character should be highlighted.
			 * Override the attribute passed in. */
			if (a != AT_ANSI)
				a |= AT_HILITE;
		}
		line_matches += matches;
	}
#endif

	if (ctldisp == OPT_ONPLUS && in_ansi_esc_seq()) {
		if (!is_ansi_end(ch) && !is_ansi_middle(ch)) {
			/* Remove whole unrecognized sequence. */
			do {
				if (curr == 0)
					break;
				--curr;
			} while (!IS_CSI_START(linebuf[curr]));
			return 0;
		}
		a = AT_ANSI;    /* Will force re-AT_'ing around it. */
		w = 0;
	} else if (ctldisp == OPT_ONPLUS && IS_CSI_START(ch)) {
		a = AT_ANSI;    /* Will force re-AT_'ing around it. */
		w = 0;
	} else {
		const char *p = &linebuf[curr];
#if SUPPORT_UTF8
		LWCHAR prev_ch = step_char(&p, -1, linebuf);
		w = pwidth(ch, prev_ch);
#else
		step_char(&p, -1, linebuf);
		w = pwidth(ch);
#endif
	}

	if (ctldisp != OPT_ON && column + w > sc_width)
		/* Won't fit on screen. */
		return 1;

	if (rep == NULL) {
		cs = (char) ch;
		rep = &cs;
		replen = 1;
	} else {
		replen = utf_len(rep[0]);
	}
	if (curr + replen >= size_linebuf-6) {
		/* Won't fit in line buffer. Try to expand it. */
		if (expand_linebuf())
			return 1;
	}

	while (replen-- > 0) {
		linebuf[curr] = *rep++;
		attr[curr] = a;
		curr++;
	}
	column += w;
	return 0;
}

/* Append a tab to the line buffer. Store spaces to represent the tab. */
#if HILITE_SEARCH
#define STORE_TAB(a,pos) \
	do { if (store_tab((a),(pos))) return 1; } while (0)
#else
#define STORE_TAB(a,pos) \
	do { if (store_tab((a))) return 1; } while (0)
#endif

static int
#if HILITE_SEARCH
store_tab(int tab_attr, POSITION pos)
#else
store_tab(int tab_attr)
#endif
{
	int to_tab = column + cshift - lmargin;
	int i;

	if (ntabstops < 2 || to_tab >= tabstops[ntabstops-1])
		to_tab = tabdefault -
		     ((to_tab - tabstops[ntabstops-1]) % tabdefault);
	else {
		for (i = ntabstops - 2;  i >= 0;  i--)
			if (to_tab >= tabstops[i])
				break;
		to_tab = tabstops[i+1] - to_tab;
	}

	if (column + to_tab - 1 + PWIDTH(' ', 0) > sc_width)
		return 1;

	do {
		STORE_CHAR(' ', tab_attr, " ", pos);
	} while (--to_tab > 0);
	return 0;
}

/* Return the printable form of a character.
 * For example, in the "ascii" charset '\3' is printed as "^C". */
static char *
prchar(int c, char buf[32])
{
	c &= 0377;
	if ((c < 128 || !utf_mode) && !control_char(c))
		sprintf(buf, "%c", c);
	else if (c == ESC)
		strcpy(buf, "ESC");
	else if (c < 128 && !control_char(c ^ 0100))
		sprintf(buf, "^%c", c ^ 0100);
	else
		snprintf(buf, 32, binfmt, c);
	return buf;
}

#if HILITE_SEARCH
#define STORE_PRCHAR(c, pos) \
	do { if (store_prchar((c), (pos))) return 1; } while (0)
#else
#define STORE_PRCHAR(c, pos) \
	do { if (store_prchar((c))) return 1; } while (0)
#endif

static int
#if HILITE_SEARCH
store_prchar(char c, POSITION pos)
#else
store_prchar(char c)
#endif
{
	char *s;
	char buf[32];

	/* Convert to printable representation. */
	s = prchar(c, buf);

	/* Make sure we can get the entire representation
	 * of the character on this line. */
	if (column + (int) strlen(s) - 1 + PWIDTH(' ', 0) > sc_width)
		return 1;

	for ( ;  *s != 0;  s++)
		STORE_CHAR(*s, AT_BINARY, NULL, pos);

	return 0;
}

#if SUPPORT_UTF8
#if HILITE_SEARCH
#define FLUSH_MBC_BUF(pos) flush_mbc_buf((pos))
#else
#define FLUSH_MBC_BUF(pos) flush_mbc_buf()
#endif

static int
#if HILITE_SEARCH
flush_mbc_buf(POSITION pos)
#else
flush_mbc_buf(void)
#endif
{
	int i;

	for (i = 0; i < mbc_buf_index; i++) {
#if HILITE_SEARCH
		if (store_prchar(mbc_buf[i], pos))
#else
		if (store_prchar(mbc_buf[i]))
#endif
			return mbc_buf_index - i;
	}

	return 0;
}

/* Is a UTF-8 character well-formed? */
static int
is_utf8_well_formed(char *ss)
{
	int i;
	int len;
	unsigned char *s = (unsigned char *)ss;

	if (IS_UTF8_INVALID(s[0]))
		return 0;

	len = utf_len((char) s[0]);
	if (len == 1)
		return 1;
	if (len == 2) {
		if (s[0] < 0xC2)
		    return 0;
	} else {
		unsigned char mask;
		mask = (~((1 << (8-len)) - 1)) & 0xFF;
		if (s[0] == mask && (s[1] & mask) == 0x80)
			return 0;
	}

	for (i = 1;  i < len;  i++)
		if (!IS_UTF8_TRAIL(s[i]))
			return 0;
	return 1;
}
#else
#define is_utf8_well_formed(s) 1
#endif /* SUPPORT_UTF8 */

/* Append a character to the line buffer.
 * Expand tabs into spaces, handle underlining, boldfacing, etc.
 * Returns 0 if ok, 1 if couldn't fit in buffer. */
static int
pappend(char c, POSITION pos)
{
	int r;

	if (pendc) {
		if (DO_APPEND(pendc, NULL, pendpos))
			/* Oops.  We've probably lost the char which
			 * was in pendc, since caller won't back up. */
			return 1;
		pendc = '\0';
	}

	if (c == '\r' && bs_mode == BS_SPECIAL) {
#if SUPPORT_UTF8
		if (mbc_buf_len > 0) { /* utf_mode must be on. */
			/* Flush incomplete (truncated) sequence. */
			r = FLUSH_MBC_BUF(mbc_pos);
			mbc_buf_index = r + 1;
			mbc_buf_len = 0;
			if (r)
				return mbc_buf_index;
		}
#endif
		/* Don't put the CR into the buffer until we see
		 * the next char.  If the next char is a newline,
		 * discard the CR. */
		pendc = c;
		pendpos = pos;
		return 0;
	}

	if (!utf_mode) {
		r = DO_APPEND((LWCHAR)c & 0xff, NULL, pos);
	}
#if SUPPORT_UTF8
	else {
		/* Perform strict validation in all possible cases. */
		if (mbc_buf_len == 0) {
		retry:
			mbc_buf_index = 1;
			*mbc_buf = c;
			if (IS_ASCII_OCTET(c))
				r = DO_APPEND((LWCHAR) c & 0xff, NULL, pos);
			else if (IS_UTF8_LEAD(c)) {
				mbc_buf_len = utf_len(c);
				mbc_pos = pos;
				return 0;
			} else
				/* UTF8_INVALID or stray UTF8_TRAIL */
				r = FLUSH_MBC_BUF(pos);
		} else if (IS_UTF8_TRAIL(c)) {
			mbc_buf[mbc_buf_index++] = c;
			if (mbc_buf_index < mbc_buf_len)
				return 0;
			if (is_utf8_well_formed(mbc_buf))
				r = DO_APPEND(get_wchar(mbc_buf), mbc_buf, mbc_pos);
			else
				/* Complete, but not shortest form, sequence. */
				mbc_buf_index = r = FLUSH_MBC_BUF(mbc_pos);
			mbc_buf_len = 0;
		} else {
			/* Flush incomplete (truncated) sequence. */
			r = FLUSH_MBC_BUF(mbc_pos);
			mbc_buf_index = r + 1;
			mbc_buf_len = 0;
			/* Handle new char. */
			if (!r)
				goto retry;
		}
	}
#endif /* SUPPORT_UTF8 */

	/* If we need to shift the line, do it.
	 * But wait until we get to at least the middle of the screen,
	 * so shifting it doesn't affect the chars we're currently
	 * pappending.  (Bold & underline can get messed up otherwise.) */
	if (cshift < hshift && column > sc_width / 2) {
		linebuf[curr] = '\0';
		pshift(hshift - cshift);
	}
	if (r) {
		/* How many chars should caller back up? */
#if SUPPORT_UTF8
		r = (!utf_mode) ? 1 : mbc_buf_index;
#else
		r = 1;
#endif
	}
	return r;
}

/* Return the printable form of a UTF-8 character. */
static char *
prutfchar(LWCHAR ch, char buf[32])
{
	if (ch == ESC)
		strcpy(buf, "ESC");
	else if (ch < 128 && control_char(ch)) {
		if (!control_char(ch ^ 0100))
			sprintf(buf, "^%c", ((char) ch) ^ 0100);
		else
			snprintf(buf, 32, binfmt, (char) ch);
	} else if (is_ubin_char(ch))
		snprintf(buf, 32, utfbinfmt, ch);
	else {
#if SUPPORT_UTF8
		int len;
		if (ch >= 0x80000000) {
			len = 3;
			ch = 0xFFFD;
		} else {
			len =   (ch < 0x80) ? 1
			      : (ch < 0x800) ? 2
			      : (ch < 0x10000) ? 3
			      : (ch < 0x200000) ? 4
			      : (ch < 0x4000000) ? 5
			      : 6;
		}
		buf[len] = '\0';
		if (len == 1)
			*buf = (char) ch;
		else {
			*buf = ((1 << len) - 1) << (8 - len);
			while (--len > 0) {
				buf[len] = (char) (0x80 | (ch & 0x3F));
				ch >>= 6;
			}
			*buf |= ch;
		}
#else
		*buf = (char) ch;
		buf[1] = '\0';
#endif
	}
	return buf;
}

static int
#if HILITE_SEARCH
do_append(LWCHAR ch, char *rep, POSITION pos)
#else
do_append(LWCHAR ch, char *rep)
#endif
{
	int a;
	LWCHAR prev_ch;

	a = AT_NORMAL;

	if (ch == '\b') {
		if (bs_mode == BS_CONTROL)
			goto do_control_char;

		/* A better test is needed here so we don't
		 * backspace over part of the printed
		 * representation of a binary character. */
		if (   curr <= lmargin
		    || column <= lmargin
		    || (attr[curr - 1] & (AT_ANSI|AT_BINARY)))
			STORE_PRCHAR('\b', pos);
		else if (bs_mode == BS_NORMAL)
			STORE_CHAR(ch, AT_NORMAL, NULL, pos);
		else if (bs_mode == BS_SPECIAL)
			overstrike = backc();

		return 0;
	}

	if (overstrike > 0) {
		/* Overstrike the character at the current position
		 * in the line buffer.  This will cause either
		 * underline (if a "_" is overstruck),
		 * bold (if an identical character is overstruck),
		 * or just deletion of the character in the buffer. */
		overstrike = utf_mode ? -1 : 0;
		/* To be correct, this must be a base character. */
		prev_ch = get_wchar(linebuf + curr);
		a = attr[curr];
		if (ch == prev_ch) {
			/* Overstriking a char with itself means make it bold.
			 * But overstriking an underscore with itself is
			 * ambiguous.  It could mean make it bold, or
			 * it could mean make it underlined.
			 * Use the previous overstrike to resolve it. */
			if (ch == '_') {
				if ((a & (AT_BOLD|AT_UNDERLINE)) != AT_NORMAL)
					a |= (AT_BOLD|AT_UNDERLINE);
				else if (last_overstrike != AT_NORMAL)
					a |= last_overstrike;
				else
					a |= AT_BOLD;
			} else
				a |= AT_BOLD;
		} else if (ch == '_') {
			a |= AT_UNDERLINE;
			ch = prev_ch;
			rep = linebuf + curr;
		} else if (prev_ch == '_') {
			a |= AT_UNDERLINE;
		}
		/* Else we replace prev_ch, but we keep its attributes. */
	} else if (overstrike < 0) {
		if (   is_composing_char(ch)
		    || is_combining_char(get_wchar(linebuf + curr), ch))
			/* Continuation of the same overstrike. */
			a = last_overstrike;
		else
			overstrike = 0;
	}

	if (ch == '\t') {
		/* Expand a tab into spaces. */
		switch (bs_mode) {
		case BS_CONTROL:
			goto do_control_char;
		case BS_NORMAL:
		case BS_SPECIAL:
			STORE_TAB(a, pos);
			break;
		}
	} else if ((!utf_mode || is_ascii_char(ch)) && control_char((char)ch)) {
	do_control_char:
		if (ctldisp == OPT_ON || (ctldisp == OPT_ONPLUS && IS_CSI_START(ch))) {
			/* Output as a normal character. */
			STORE_CHAR(ch, AT_NORMAL, rep, pos);
		} else {
			STORE_PRCHAR((char) ch, pos);
		}
	} else if (utf_mode && ctldisp != OPT_ON && is_ubin_char(ch)) {
		char *s;
		char buf[32];

		s = prutfchar(ch, buf);

		if (column + (int) strlen(s) - 1 + PWIDTH(' ', 0) > sc_width)
			return 1;

		for ( ;  *s != 0;  s++)
			STORE_CHAR(*s, AT_BINARY, NULL, pos);
	} else {
		STORE_CHAR(ch, a, rep, pos);
	}
	return 0;
}

#if SUPPORT_UTF8
static int
pflushmbc(void)
{
	int r = 0;

	if (mbc_buf_len > 0) {
		/* Flush incomplete (truncated) sequence. */
		r = FLUSH_MBC_BUF(mbc_pos);
		mbc_buf_len = 0;
	}
	return r;
}
#else
#define pflushmbc() 0
#endif

/* Terminate the line in the line buffer. */
static void
pdone(int endline)
{
	int nl;

	(void) pflushmbc();

	if (pendc && (pendc != '\r' || !endline))
		/* If we had a pending character, put it in the buffer.
		 * But discard a pending CR if we are at end of line
		 * (that is, discard the CR in a CR/LF sequence). */
		(void) DO_APPEND(pendc, NULL, pendpos);

	/* Make sure we've shifted the line, if we need to. */
	if (cshift < hshift)
		pshift(hshift - cshift);

	if (ctldisp == OPT_ONPLUS && is_ansi_end('m')) {
		/* Switch to normal attribute at end of line. */
		char *p = "\033[m";
		for ( ;  *p != '\0';  p++) {
			linebuf[curr] = *p;
			attr[curr++] = AT_ANSI;
		}
	}

	/* Add a newline if necessary,
	 * and append a '\0' to the end of the line.
	 * We output a newline if we're not at the right edge of the screen,
	 * or if the terminal doesn't auto wrap,
	 * or if this is really the end of the line AND the terminal ignores
	 * a newline at the right edge.
	 * (In the last case we don't want to output a newline if the terminal
	 * doesn't ignore it since that would produce an extra blank line.
	 * But we do want to output a newline if the terminal ignores it in case
	 * the next line is blank.  In that case the single newline output for
	 * that blank line would be ignored!) */
	nl = (column < sc_width || !auto_wrap || (endline && ignaw) || ctldisp == OPT_ON);
	if (nl) {
		linebuf[curr] = '\n';
		attr[curr] = AT_NORMAL;
		curr++;
	}
	else if (ignaw && !auto_wrap && column >= sc_width) {
		/* Big horrible kludge.
		 * No-wrap terminals are too hard to deal with when they get in
		 * the state where a full screen width of characters have been
		 * output but the cursor is sitting on the right edge instead
		 * of at the start of the next line.
		 * So after we output a full line, we output an extra
		 * space and backspace to force the cursor to the
		 * beginning of the next line, like a sane terminal. */
		linebuf[curr] = ' ';
		attr[curr++] = AT_NORMAL;
		linebuf[curr] = '\b';
		attr[curr++] = AT_NORMAL;
	}
	linebuf[curr] = '\0';
	attr[curr] = AT_NORMAL;

#if HILITE_SEARCH
	if (status_col && line_matches > 0) {
		linebuf[0] = '*';
		attr[0] = AT_NORMAL|AT_HILITE;
	}
#endif
}

/* Get a character from the current line.
 * Return the character as the function return value,
 * and the character attribute in *ap. */
static int
gline(int i, int *ap)
{
	if (is_null_line) {
		/* If there is no current line, we pretend the line is
		 * either "~" or "", depending on the "twiddle" flag. */
		if (twiddle) {
			if (i == 0) {
				*ap = AT_BOLD;
				return '~';
			}
			--i;
		}
		/* Make sure we're back to AT_NORMAL before the '\n'. */
		*ap = AT_NORMAL;
		return i ? '\0' : '\n';
	}

	*ap = attr[i];
	return (linebuf[i] & 0xFF);
}

/* Indicate that there is no current line. */
static void
null_line(void)
{
	is_null_line = 1;
	cshift = 0;
}


/* High level routines dealing with getting lines of input
 * from the file being viewed.
 *
 * When we speak of "lines" here, we mean PRINTABLE lines;
 * lines processed with respect to the screen width.
 * We use the term "raw line" to refer to lines simply
 * delimited by newlines; not processed with respect to screen width. */

/* Is there a previous (remembered) search pattern? */
static int
prev_pattern(void)
{
	if (last_search_type & SRCH_NO_REGEX)
		return (last_pattern != NULL);
#if NO_REGEX
	return (last_pattern != NULL);
#else
	return (regpattern != NULL);
#endif
}

#if HILITE_SEARCH
/* Clear the hilite list. */
static void
clr_hilite(void)
{
	struct hilite *hl;
	struct hilite *nexthl;

	for (hl = hilite_anchor.hl_first;  hl != NULL;  hl = nexthl) {
		nexthl = hl->hl_next;
		free((void*)hl);
	}
	hilite_anchor.hl_first = NULL;
	prep_startpos = prep_endpos = NULL_POSITION;
}


/* Prepare hilites in a given range of the file.
 *
 * The pair (prep_startpos,prep_endpos) delimits a contiguous region
 * of the file that has been "prepared"; that is, scanned for matches for
 * the current search pattern, and hilites have been created for such matches.
 * If prep_startpos == NULL_POSITION, the prep region is empty.
 * If prep_endpos == NULL_POSITION, the prep region extends to EOF.
 * prep_hilite asks that the range (spos,epos) be covered by the prep region. */
static void
prep_hilite(POSITION spos, POSITION epos, int maxlines)
{
	POSITION nprep_startpos = prep_startpos;
	POSITION nprep_endpos = prep_endpos;
	POSITION new_epos;
	POSITION max_epos;
	int result;
	int i;
/* Search beyond where we're asked to search, so the prep region covers
 * more than we need.  Do one big search instead of a bunch of small ones. */
#define SEARCH_MORE (3*size_linebuf)

	if (!prev_pattern())
		return;

	/* If we're limited to a max number of lines, figure out the
	 * file position we should stop at. */
	if (maxlines < 0)
		max_epos = NULL_POSITION;
	else {
		max_epos = spos;
		for (i = 0;  i < maxlines;  i++)
			max_epos = forw_raw_line(max_epos, (const char **)NULL, (int *)NULL);
	}

	/* Find two ranges:
	 * The range that we need to search (spos,epos); and the range that
	 * the "prep" region will then cover (nprep_startpos,nprep_endpos). */

	if (prep_startpos == NULL_POSITION ||
	    (epos != NULL_POSITION && epos < prep_startpos) ||
	    spos > prep_endpos) {
		/* New range is not contiguous with old prep region.
		 * Discard the old prep region and start a new one. */
		clr_hilite();
		if (epos != NULL_POSITION)
			epos += SEARCH_MORE;
		nprep_startpos = spos;
	} else {
		/* New range partially or completely overlaps old prep region. */
		if (epos == NULL_POSITION) {
			/* New range goes to end of file. */
			;
		} else if (epos > prep_endpos) {
			/* New range ends after old prep region.
			 * Extend prep region to end at end of new range. */
			epos += SEARCH_MORE;
		} else /* (epos <= prep_endpos) */ {
			/* New range ends within old prep region.
			 * Truncate search to end at start of old prep region. */
			epos = prep_startpos;
		}

		if (spos < prep_startpos) {
			/* New range starts before old prep region.
			 * Extend old prep region backwards to start at
			 * start of new range. */
			if (spos < SEARCH_MORE)
				spos = 0;
			else
				spos -= SEARCH_MORE;
			nprep_startpos = spos;
		} else /* (spos >= prep_startpos) */ {
			/* New range starts within or after old prep region.
			 * Trim search to start at end of old prep region. */
			spos = prep_endpos;
		}
	}

	if (epos != NULL_POSITION && max_epos != NULL_POSITION &&
	    epos > max_epos)
		/* Don't go past the max position we're allowed. */
		epos = max_epos;

	if (epos == NULL_POSITION || epos > spos) {
		result = search_range(spos, epos, SRCH_FORW|SRCH_FIND_ALL, 0,
				maxlines, (POSITION*)NULL, &new_epos);
		if (result < 0)
			return;
		if (prep_endpos == NULL_POSITION || new_epos > prep_endpos)
			nprep_endpos = new_epos;
	}
	prep_startpos = nprep_startpos;
	prep_endpos = nprep_endpos;
}
#endif /* HILITE_SEARCH */

/* Get the next line.
 * A "current" position is passed and a "new" position is returned.
 * The current position is the position of the first character of
 * a line.  The new position is the position of the first character
 * of the NEXT line.  The line obtained is the line starting at curr_pos. */
static POSITION
forw_line(POSITION curr_pos)
{
	POSITION base_pos;
	POSITION new_pos;
	int c;
	int blankline;
	int endline;
	int backchars;

	if (curr_pos == NULL_POSITION) {
		null_line();
		return NULL_POSITION;
	}
#if HILITE_SEARCH
	if (hilite_search == OPT_ONPLUS || status_col)
		/* If we are ignoring EOI (command F), only prepare
		 * one line ahead, to avoid getting stuck waiting for
		 * slow data without displaying the data we already have.
		 * If we're not ignoring EOI, we *could* do the same, but
		 * for efficiency we prepare several lines ahead at once. */
		prep_hilite(curr_pos, curr_pos + 3*size_linebuf,
				ignore_eoi ? 1 : -1);
#endif
	if (ch_seek(curr_pos)) {
		null_line();
		return NULL_POSITION;
	}

	base_pos = curr_pos;
	for (;;) {
		if (ABORT_SIGS()) {
			null_line();
			return NULL_POSITION;
		}
		c = ch_back_get();
		if (c == EOI)
			break;
		if (c == '\n') {
			(void) ch_forw_get();
			break;
		}
		--base_pos;
	}

	prewind();
	plinenum(base_pos);
	(void) ch_seek(base_pos);
	while (base_pos < curr_pos) {
		if (ABORT_SIGS()) {
			null_line();
			return NULL_POSITION;
		}
		c = ch_forw_get();
		backchars = pappend(c, base_pos);
		base_pos++;
		if (backchars > 0) {
			pshift_all();
			base_pos -= backchars;
			while (--backchars >= 0)
				(void) ch_back_get();
		}
	}
	(void) pflushmbc();
	pshift_all();

	c = ch_forw_get();
	if (c == EOI) {
		null_line();
		return NULL_POSITION;
	}
	blankline = (c == '\n' || c == '\r');

	for (;;) {
		if (ABORT_SIGS()) {
			null_line();
			return NULL_POSITION;
		}
		if (c == '\n' || c == EOI) {
			/* End of the line. */
			backchars = pflushmbc();
			new_pos = ch_tell();
			if (backchars > 0 && !chopline && hshift == 0) {
				new_pos -= backchars + 1;
				endline = FALSE;
			} else
				endline = TRUE;
			break;
		}
		if (c != '\r')
			blankline = 0;

		/* Append the char to the line and get the next char. */
		backchars = pappend(c, ch_tell()-1);
		if (backchars > 0) {
			/* The char won't fit in the line; the line
			 * is too long to print in the screen width.
			 * End the line here. */
			if (chopline || hshift > 0) {
				do {
					c = ch_forw_get();
				} while (c != '\n' && c != EOI);
				new_pos = ch_tell();
				endline = TRUE;
				quit_if_one_screen = FALSE;
			} else {
				new_pos = ch_tell() - backchars;
				endline = FALSE;
			}
			break;
		}
		c = ch_forw_get();
	}
	pdone(endline);

	if (squeeze && blankline) {
		/* This line is blank.
		 * Skip down to the last contiguous blank line
		 * and pretend it is the one which we are returning. */
		while ((c = ch_forw_get()) == '\n' || c == '\r')
			if (ABORT_SIGS()) {
				null_line();
				return NULL_POSITION;
			}
		if (c != EOI)
			(void) ch_back_get();
		new_pos = ch_tell();
	}

	return new_pos;
}

/* Get the previous line.
 * A "current" position is passed and a "new" position is returned.
 * The current position is the position of the first character of
 * a line.  The new position is the position of the first character
 * of the PREVIOUS line.  The line obtained is the one starting at new_pos. */
static POSITION
back_line(POSITION curr_pos)
{
	POSITION new_pos, begin_new_pos;
	int c;
	int endline;
	int backchars;

	if (curr_pos == NULL_POSITION || curr_pos <= ch_zero()) {
		null_line();
		return NULL_POSITION;
	}
#if HILITE_SEARCH
	if (hilite_search == OPT_ONPLUS || status_col)
		prep_hilite((curr_pos < 3*size_linebuf) ?
				0 : curr_pos - 3*size_linebuf, curr_pos, -1);
#endif
	if (ch_seek(curr_pos-1)) {
		null_line();
		return NULL_POSITION;
	}

	if (squeeze) {
		/* Find out if the "current" line was blank. */
		(void) ch_forw_get();   /* Skip the newline */
		c = ch_forw_get();      /* First char of "current" line */
		(void) ch_back_get();   /* Restore our position */
		(void) ch_back_get();

		if (c == '\n' || c == '\r') {
			/* The "current" line was blank.
			 * Skip over any preceding blank lines,
			 * since we skipped them in forw_line(). */
			while ((c = ch_back_get()) == '\n' || c == '\r')
				if (ABORT_SIGS()) {
					null_line();
					return NULL_POSITION;
				}
			if (c == EOI) {
				null_line();
				return NULL_POSITION;
			}
			(void) ch_forw_get();
		}
	}

	/* Scan backwards until we hit the beginning of the line. */
	for (;;) {
		if (ABORT_SIGS()) {
			null_line();
			return NULL_POSITION;
		}
		c = ch_back_get();
		if (c == '\n') {
			/* This is the newline ending the previous line.
			 * We have hit the beginning of the line. */
			new_pos = ch_tell() + 1;
			break;
		}
		if (c == EOI) {
			/* We have hit the beginning of the file.
			 * This must be the first line in the file.
			 * This must, of course, be the beginning of the line. */
			new_pos = ch_tell();
			break;
		}
	}

	/* Now scan forwards from the beginning of this line.
	 * We keep discarding "printable lines" (based on screen width)
	 * until we reach the curr_pos.
	 *
	 * {{ This algorithm is pretty inefficient if the lines
	 *    are much longer than the screen width,
	 *    but I don't know of any better way. }} */
	if (ch_seek(new_pos)) {
		null_line();
		return NULL_POSITION;
	}
	endline = FALSE;
	prewind();
	plinenum(new_pos);
    loop:
	begin_new_pos = new_pos;
	(void) ch_seek(new_pos);

	do {
		c = ch_forw_get();
		if (c == EOI || ABORT_SIGS()) {
			null_line();
			return NULL_POSITION;
		}
		new_pos++;
		if (c == '\n') {
			backchars = pflushmbc();
			if (backchars > 0 && !chopline && hshift == 0) {
				backchars++;
				goto shift;
			}
			endline = TRUE;
			break;
		}
		backchars = pappend(c, ch_tell()-1);
		if (backchars > 0) {
			/* Got a full printable line, but we haven't
			 * reached our curr_pos yet.  Discard the line
			 * and start a new one. */
			if (chopline || hshift > 0) {
				endline = TRUE;
				quit_if_one_screen = FALSE;
				break;
			}
		shift:
			pshift_all();
			while (backchars-- > 0) {
				(void) ch_back_get();
				new_pos--;
			}
			goto loop;
		}
	} while (new_pos < curr_pos);

	pdone(endline);

	return begin_new_pos;
}


/* Code to handle displaying line numbers.
 *
 * Finding the line number of a given file position is rather tricky.
 * We don't want to just start at the beginning of the file and
 * count newlines, because that is slow for large files (and also
 * wouldn't work if we couldn't get to the start of the file; e.g.
 * if input is a long pipe).
 *
 * So we use the function add_lnum to cache line numbers.
 * We try to be very clever and keep only the more interesting
 * line numbers when we run out of space in our table.  A line
 * number is more interesting than another when it is far from
 * other line numbers.   For example, we'd rather keep lines
 * 100,200,300 than 100,101,300.  200 is more interesting than
 * 101 because 101 can be derived very cheaply from 100, while
 * 200 is more expensive to derive from 100. */

/* Initialize the line number structures. */
static void
clr_linenum(void)
{
	struct linenum_info *p;

	/* Put all the entries on the free list.
	 * Leave one for the "spare". */
	for (p = pool;  p < &pool[NPOOL-2];  p++)
		p->next = p+1;
	pool[NPOOL-2].next = NULL;
	freelist = pool;

	spare = &pool[NPOOL-1];

	/* Initialize the anchor. */
	anchor.next = anchor.prev = &anchor;
	anchor.gap = 0;
	anchor.pos = (POSITION)0;
	anchor.line = 1;
}

/* Find the position of a given line number.
 * Return NULL_POSITION if we can't figure it out. */
static POSITION
find_pos(LINENUM linenum)
{
	struct linenum_info *p;
	POSITION cpos;
	LINENUM clinenum;

	if (linenum <= 1)
		/* Line number 1 is beginning of file. */
		return ch_zero();

	/* Find the entry nearest to the line number we want. */
	for (p = anchor.next;  p != &anchor && p->line < linenum;  p = p->next)
		continue;
	if (p->line == linenum)
		/* Found it exactly. */
		return p->pos;

	if (p == &anchor || linenum - p->prev->line < p->line - linenum) {
		/* Go forward. */
		p = p->prev;
		if (ch_seek(p->pos))
			return NULL_POSITION;
		for (clinenum = p->line, cpos = p->pos;  clinenum < linenum;  clinenum++) {
			/* Allow a signal to abort this loop. */
			cpos = forw_raw_line(cpos, (const char **)NULL, (int *)NULL);
			if (ABORT_SIGS() || cpos == NULL_POSITION)
				return NULL_POSITION;
		}
	} else {
		/* Go backward. */
		if (ch_seek(p->pos))
			return NULL_POSITION;
		for (clinenum = p->line, cpos = p->pos;  clinenum > linenum;  clinenum--) {
			/* Allow a signal to abort this loop. */
			cpos = back_raw_line(cpos, (const char **)NULL, (int *)NULL);
			if (ABORT_SIGS() || cpos == NULL_POSITION)
				return NULL_POSITION;
		}
	}
	/* We might as well cache it. */
	add_lnum(clinenum, cpos);
	return cpos;
}


/* High level routines dealing with the output to the screen. */

/* Initialize terminal */
static void
init(void)
{
	if (top_scroll) {
		int i;

		/* This is nice to terminals with no alternate screen,
		 * but with saved scrolled-off-the-top lines.  This way,
		 * no previous line is lost, but we start with a whole
		 * screen to ourself. */
		for (i = 1; i < sc_height; i++)
			putchr('\n');
	}
}

/* Home cursor (move to upper left corner of screen). */
static void
home(void)
{
	putstr(sc_home);
}

/* Add a blank line (called with cursor at home).
 * Should scroll the display down. */
static void
add_line(void)
{
	putstr(sc_addline);
}

/* make minimal harcoded vt100-like TERM by dzo@ */
static char *less_tgoto(int y, char xy_out[32])
{
	sprintf(xy_out, "\033[%d;1H", y + 1);
	return xy_out;
}

/* Move cursor to lower left corner of screen. */
static void
lower_left(void)
{
	char xy_out[32];
	putstr(less_tgoto(sc_height-1, xy_out));
}

/* Move cursor to left position of current line. */
static void
line_left(void)
{
	putstr(sc_return);
}

/* Goto a specific line on the screen. */


static void
goto_line(int slinenum)
{
	char xy_out[32];
	putstr(less_tgoto(slinenum, xy_out));
}

/* Output the "visual bell", if there is one. */
static void
vbell(void)
{
	if (*sc_visual_bell == '\0')
		return;
	putstr(sc_visual_bell);
}

/* Make a noise. */
static void
beep(void)
{
	putchr(CONTROL('G'));
}

/* Ring the terminal bell. */
static void
bell(void)
{
	if (quiet == VERY_QUIET)
		vbell();
	else
		beep();
}

/* Clear the screen. */
static void
clear(void)
{
	putstr(sc_clear);
}


/* Clear the current line.
 * Clear the screen if there's off-screen memory below the display. */
static void
clear_eol_bot(void)
{
	putstr(sc_eol_clear);
}

static void
at_enter(int at)
{
	at = apply_at_specials(at);

	/* The one with the most priority is last. */
	if (at & AT_UNDERLINE)
		putstr(sc_u_in);
	if (at & AT_BOLD)
		putstr(sc_b_in);
	if (at & AT_BLINK)
		putstr(sc_bl_in);
	if (at & AT_STANDOUT)
		putstr(sc_s_in);
	attrmode = at;
}

static void
at_exit(void)
{
	/* Undo things in the reverse order we did them. */
	if (attrmode & AT_STANDOUT)
		putstr(sc_s_out);
	if (attrmode & AT_BLINK)
		putstr(sc_bl_out);
	if (attrmode & AT_BOLD)
		putstr(sc_b_out);
	if (attrmode & AT_UNDERLINE)
		putstr(sc_u_out);
	attrmode = AT_NORMAL;
}

/* Clear the bottom line of the display.
 * Leave the cursor at the beginning of the bottom line. */
static void
clear_bot(void)
{
	/* If we're in a non-normal attribute mode, temporarily exit
	 * the mode while we do the clear.  Some terminals fill the
	 * cleared area with the current attribute. */
	line_left();

	if (attrmode == AT_NORMAL)
		clear_eol_bot();
	else {
		int saved_attrmode = attrmode;

		at_exit();
		clear_eol_bot();
		at_enter(saved_attrmode);
	}
}

static void
at_switch(int at)
{
	int new_attrmode = apply_at_specials(at);
	int ignore_modes = AT_ANSI;

	if ((new_attrmode & ~ignore_modes) != (attrmode & ~ignore_modes)) {
		at_exit();
		at_enter(at);
	}
}

/* Output a plain backspace, without erasing the previous char. */
static void
putbs(void)
{
	putstr(sc_backspace);
}

/* Display the line which is in the line buffer. */
static void
put_line(void)
{
	int c;
	int i;
	int a;

	if (ABORT_SIGS()) {
		/* Don't output if a signal is pending. */
		screen_trashed = 1;
		return;
	}

	final_attr = AT_NORMAL;

	for (i = 0;  (c = gline(i, &a)) != '\0';  i++) {
		at_switch(a);
		final_attr = a;
		if (c == '\b')
			putbs();
		else
			putchr(c);
	}

	at_exit();
}


/* "Unget" a command character. The next getcc() will return this character */
static void
ungetcc(int c)
{
	if (ungotp == NULL)
		ungotp = ungot;
	if (ungotp >= ungot + sizeof(ungot)) {
		error("ungetcc overflow", NULL_PARG);
		quit(QUIT_ERROR);
	}
	*ungotp++ = c;
}

/* Get a character from the keyboard. */
static int
getchr(void)
{
	char c;
	int result;

	do {
		result = iread(tty, &c, sizeof(char));
		if (result == READ_INTR)
			return READ_INTR;
		if (result < 0) {
			/* Don't call error() here,
			 * because error calls getchr! */
			quit(QUIT_ERROR);
		}
		/* Various parts of the program cannot handle
		 * an input character of '\0'.
		 * If a '\0' was actually typed, convert it to '\340' here. */
		if (c == '\0')
			c = '\340';
	} while (result != 1);

	return (c & 0xFF);
}

/* Get a RETURN.
 * If some other non-trivial char is pressed, unget it, so it will
 * become the next command. */
static void
get_return(void)
{
	int c;

	c = getchr();
	if (c != '\n' && c != '\r' && c != ' ' && c != READ_INTR)
		ungetcc(c);
}

/* Primitives for displaying the file on the screen,
 * scrolling either forward or backward. */

/* Sound the bell to indicate user is trying to move past end of file. */
static void
eof_bell(void)
{
	if (quiet == NOT_QUIET)
		bell();
	else
		vbell();
}

/* Check to see if the end of file is currently "displayed". */
static void
eof_check(void)
{
	POSITION pos;

	if (ignore_eoi)
		return;
	if (ABORT_SIGS())
		return;
	/* If the bottom line is empty, we are at EOF.
	 * If the bottom line ends at the file length,
	 * we must be just at EOF. */
	pos = position(BOTTOM_PLUS_ONE);
	if (pos == NULL_POSITION || pos == ch_length())
		hit_eof++;
}

/* Get the current screen position.
 * The screen position consists of both a file position and
 * a screen line number where the file position is placed on the screen.
 * Normally the screen line number is 0, but if we are positioned
 * such that the top few lines are empty, we may have to set
 * the screen line to a number > 0. */
static void
get_scrpos(struct scrpos *scrpos)
{
	int i;

	/* Find the first line on the screen which has something on it,
	 * and return the screen line number and the file position. */
	for (i = 0; i < sc_height;  i++)
		if (table[i] != NULL_POSITION) {
			scrpos->ln = i+1;
			scrpos->pos = table[i];
			return;
		}
	/* The screen is empty. */
	scrpos->pos = NULL_POSITION;
}

/* Initialize the position table, done whenever we clear the screen. */
static void
pos_clear(void)
{
	int i;

	for (i = 0;  i < sc_height;  i++)
		table[i] = NULL_POSITION;
}

/* Add a new file position to the top of the position table. */
static void
add_back_pos(POSITION pos)
{
	int i;

	/* Scroll the position table down. */
	for (i = sc_height - 1;  i > 0;  i--)
		table[i] = table[i-1];
	table[0] = pos;
}

/* If the screen is "squished", repaint it.
 * "Squished" means the first displayed line is not at the top
 * of the screen; this can happen when we display a short file
 * for the first time. */
static void
squish_check(void)
{
	if (!squished)
		return;
	squished = 0;
	repaint();
}

/* Get the backwards scroll limit.
 * Must call this function instead of just using the value of
 * back_scroll, because the default case depends on sc_height and
 * top_scroll, as well as back_scroll. */
static int
get_back_scroll(void)
{
	if (back_scroll >= 0)
		return back_scroll;
	if (top_scroll)
		return sc_height - 2;
	return 0x7fffffff; /* infinity */
}

/* Display n lines, scrolling backward. */
static void
back(int n, POSITION pos, int force, int only_last)
{
	int nlines = 0;
	int do_repaint;

	squish_check();
	do_repaint = (n > get_back_scroll() || (only_last && n > sc_height-1));
	hit_eof = 0;
	while (--n >= 0) {
		/* Get the previous line of input. */
		pos = back_line(pos);
		if (pos == NULL_POSITION) {
			/* Beginning of file: stop here unless "force" is true. */
			if (!force)
				break;
		}
		/* Add the position of the previous line to the position table.
		 * Display the line on the screen. */
		add_back_pos(pos);
		nlines++;
		if (!do_repaint) {
			home();
			add_line();
			put_line();
		}
	}

	eof_check();
	if (nlines == 0)
		eof_bell();
	else if (do_repaint)
		repaint();
	else
		lower_left();
	(void) currline(BOTTOM);
}

/* Add a new file position to the bottom of the position table. */
static void
add_forw_pos(POSITION pos)
{
	int i;

	/* Scroll the position table up. */
	for (i = 1;  i < sc_height;  i++)
		table[i-1] = table[i];
	table[sc_height - 1] = pos;
}

static int
empty_lines(int s, int e)
{
	int i;

	for (i = s;  i <= e;  i++)
		if (table[i] != NULL_POSITION)
			return 0;
	return 1;
}

/* See if the entire screen is empty. */
static int
empty_screen(void)
{
	return empty_lines(0, sc_height-1);
}

/* Display n lines, scrolling forward,
 * starting at position pos in the input file.
 * "force" means display the n lines even if we hit end of file.
 * "only_last" means display only the last screenful if n > screen size.
 * "nblank" is the number of blank lines to draw before the first
 *   real line.  If nblank > 0, the pos must be NULL_POSITION.
 *   The first real line after the blanks will start at ch_zero(). */
static void
forw(int n, POSITION pos, int force, int only_last, int nblank)
{
	int eof = 0;
	int nlines = 0;
	int do_repaint;

	squish_check();

	/* do_repaint tells us not to display anything till the end,
	 * then just repaint the entire screen.
	 * We repaint if we are supposed to display only the last
	 * screenful and the request is for more than a screenful.
	 * Also if the request exceeds the forward scroll limit
	 * (but not if the request is for exactly a screenful, since
	 * repainting itself involves scrolling forward a screenful). */
	do_repaint = (only_last && n > sc_height-1) ||
		(forw_scroll >= 0 && n > forw_scroll && n != sc_height-1);

	if (!do_repaint) {
		if (top_scroll && n >= sc_height - 1 && pos != ch_length()) {
			/* Start a new screen.
			 * {{ This is not really desirable if we happen
			 *    to hit eof in the middle of this screen,
			 *    but we don't yet know if that will happen. }} */
			pos_clear();
			add_forw_pos(pos);
			force = 1;
			clear();
			home();
		}

		if (pos != position(BOTTOM_PLUS_ONE) || empty_screen()) {
			/* This is not contiguous with what is
			 * currently displayed.  Clear the screen image
			 * (position table) and start a new screen. */
			pos_clear();
			add_forw_pos(pos);
			force = 1;
			if (top_scroll) {
				clear();
				home();
			} else if (!first_time) {
				putstr("...skipping...\n");
			}
		}
	}

	while (--n >= 0) {
		/* Read the next line of input. */
		if (nblank > 0) {
			/* Still drawing blanks; don't get a line
			 * from the file yet.
			 * If this is the last blank line, get ready to
			 * read a line starting at ch_zero() next time. */
			if (--nblank == 0)
				pos = ch_zero();
		} else {
			/* Get the next line from the file. */
			pos = forw_line(pos);
			if (pos == NULL_POSITION) {
				/* End of file: stop here unless the top line
				 * is still empty, or "force" is true.
				 * Even if force is true, stop when the last
				 * line in the file reaches the top of screen. */
				eof = 1;
				if (!force && position(TOP) != NULL_POSITION)
					break;
				if (!empty_lines(0, 0) &&
				    !empty_lines(1, 1) &&
				     empty_lines(2, sc_height-1))
					break;
			}
		}
		/* Add the position of the next line to the position table.
		 * Display the current line on the screen. */
		add_forw_pos(pos);
		nlines++;
		if (do_repaint)
			continue;
		/* If this is the first screen displayed and
		 * we hit an early EOF (i.e. before the requested
		 * number of lines), we "squish" the display down
		 * at the bottom of the screen.
		 * But don't do this if a + option or a -t option
		 * was given.  These options can cause us to
		 * start the display after the beginning of the file,
		 * and it is not appropriate to squish in that case. */
		if (first_time && pos == NULL_POSITION && !top_scroll &&
#if TAGS
		    tagoption == NULL &&
#endif
		    !plusoption) {
			squished = 1;
			continue;
		}
		put_line();
		forw_prompt = 1;
	}

	if (ignore_eoi)
		hit_eof = 0;
	else if (eof && !ABORT_SIGS())
		hit_eof++;
	else
		eof_check();
	if (nlines == 0)
		eof_bell();
	else if (do_repaint)
		repaint();
	first_time = 0;
	(void) currline(BOTTOM);
}

#if HILITE_SEARCH
/* Repaint the hilites currently displayed on the screen.
 * Repaint each line which contains highlighted text.
 * If on==0, force all hilites off. */
static void
repaint_hilite(int on)
{
	int slinenum;
	POSITION pos;
	int save_hide_hilite;

	if (squished)
		repaint();

	save_hide_hilite = hide_hilite;
	if (!on) {
		if (hide_hilite)
			return;
		hide_hilite = 1;
	}

	for (slinenum = TOP;  slinenum < TOP + sc_height-1;  slinenum++) {
		pos = position(slinenum);
		if (pos == NULL_POSITION)
			continue;
		position(slinenum+1);
		(void) forw_line(pos);
		goto_line(slinenum);
		put_line();
	}
	lower_left();
	hide_hilite = save_hide_hilite;
}

/* Find matching text which is currently on screen and highlight it. */
static void
hilite_screen(void)
{
	struct scrpos scrpos;

	get_scrpos(&scrpos);
	if (scrpos.pos == NULL_POSITION)
		return;
	prep_hilite(scrpos.pos, position(BOTTOM_PLUS_ONE), -1);
	repaint_hilite(1);
}


/* Change highlighting parameters. */
static void
chg_hilite(void)
{
	/* Erase any highlights currently on screen. */
	clr_hilite();
	hide_hilite = 0;

	if (hilite_search == OPT_ONPLUS)
		/* Display highlights. */
		hilite_screen();
}
#endif /* HILITE_SEARCH */

/* Return ch_flags for the current file. */
static int
ch_getflags(void)
{
	if (thisfile == NULL)
		return 0;
	return ch_flags;
}

/* Set lmark (the mark named by the apostrophe). */
static void
lastmark(void)
{
	struct scrpos scrpos;

	if (ch_getflags() & CH_HELPFILE)
		return;
	get_scrpos(&scrpos);
	if (scrpos.pos == NULL_POSITION)
		return;
	marks[LASTMARK].m_scrpos = scrpos;
	marks[LASTMARK].m_ifile = curr_ifile;
}

/* See if the byte at a specified position is currently on the screen.
 * Check the position table to see if the position falls within its range.
 * Return the position table entry if found, -1 if not. */
static int
onscreen(POSITION pos)
{
	int i;

	if (pos < table[0])
		return -1;
	for (i = 1;  i < sc_height;  i++)
		if (pos < table[i])
			return i-1;
	return -1;
}


/* Jump to a specified position in the file.
 * The position must be the first character in a line.
 * Place the target line on a specified line on the screen. */
static void
jump_loc(POSITION pos, int sline)
{
	int nline;
	POSITION tpos;
	POSITION bpos;

	/* Normalize sline. */
	sline = adjsline(sline);

	if ((nline = onscreen(pos)) >= 0) {
		/* The line is currently displayed. Just scroll there. */
		nline -= sline;
		if (nline > 0)
			forw(nline, position(BOTTOM_PLUS_ONE), 1, 0, 0);
		else
			back(-nline, position(TOP), 1, 0);
#if HILITE_SEARCH
		if (show_attn)
			repaint_hilite(1);
#endif
		return;
	}

	/* Line is not on screen. Seek to the desired location. */
	if (ch_seek(pos)) {
		error("Cannot seek to that file position", NULL_PARG);
		return;
	}

	/* See if the desired line is before or after
	 * the currently displayed screen. */
	tpos = position(TOP);
	bpos = position(BOTTOM_PLUS_ONE);
	if (tpos == NULL_POSITION || pos >= tpos) {
		/* The desired line is after the current screen.
		 * Move back in the file far enough so that we can
		 * call forw() and put the desired line at the
		 * sline-th line on the screen. */
		for (nline = 0;  nline < sline;  nline++) {
			if (bpos != NULL_POSITION && pos <= bpos) {
				/* Surprise!  The desired line is
				 * close enough to the current screen
				 * that we can just scroll there after all. */
				forw(sc_height-sline+nline-1, bpos, 1, 0, 0);
#if HILITE_SEARCH
				if (show_attn)
					repaint_hilite(1);
#endif
				return;
			}
			pos = back_line(pos);
			if (pos == NULL_POSITION) {
				/* Oops.  Ran into the beginning of the file.
				 * Exit the loop here and rely on forw()
				 * below to draw the required number of
				 * blank lines at the top of the screen. */
				break;
			}
		}
		lastmark();
		hit_eof = 0;
		squished = 0;
		screen_trashed = 0;
		forw(sc_height-1, pos, 1, 0, sline-nline);
	} else {
		/* The desired line is before the current screen.
		 * Move forward in the file far enough so that we
		 * can call back() and put the desired line at the
		 * sline-th line on the screen. */
		for (nline = sline;  nline < sc_height - 1;  nline++) {
			pos = forw_line(pos);
			if (pos == NULL_POSITION) {
				/* Ran into end of file.
				 * This shouldn't normally happen,
				 * but may if there is some kind of read error. */
				break;
			}
			if (pos >= tpos) {
				/* Surprise!  The desired line is
				 * close enough to the current screen
				 * that we can just scroll there after all. */
				back(nline+1, tpos, 1, 0);
#if HILITE_SEARCH
				if (show_attn)
					repaint_hilite(1);
#endif
				return;
			}
		}
		lastmark();
		if (!top_scroll)
			clear();
		else
			home();
		screen_trashed = 0;
		add_back_pos(pos);
		back(sc_height-1, pos, 1, 0);
	}
}

/* Repaint the screen. */
static void
repaint(void)
{
	struct scrpos scrpos;
	/* Start at the line currently at the top of the screen
	 * and redisplay the screen. */
	get_scrpos(&scrpos);
	pos_clear();
	jump_loc(scrpos.pos, scrpos.ln);
}

/* Get the ifile before a given one in the list. */
static IFILE
prev_ifile(IFILE h)
{
	struct ifile *p;

	p = (h == NULL_IFILE) ? &ifile_anchor : int_ifile(h);
	if (p->h_prev == &ifile_anchor)
		return NULL_IFILE;
	return ext_ifile(p->h_prev);
}

/* Edit the n-th next or previous file in the command line (ifile) list. */
static int
edit_istep(IFILE h, int n, int dir)
{
	IFILE next;

	/* Skip n filenames, then try to edit each filename. */
	for (;;) {
		next = (dir > 0) ? next_ifile(h) : prev_ifile(h);
		if (--n < 0) {
			if (edit_ifile(h) == 0)
				break;
		}
		if (next == NULL_IFILE) {
			/* Reached end of the ifile list. */
			return 1;
		}
		if (ABORT_SIGS()) {
			/* Interrupt breaks out, if we're in a long
			 * list of files that can't be opened. */
			return 1;
		}
		h = next;
	}
	/* Found a file that we can edit. */
	return 0;
}

static int
edit_next(int n)
{
	return edit_istep(curr_ifile, n, +1);
}

/* Get the value of the -e flag. */
static int
get_quit_at_eof(void)
{
	if (!less_is_more)
		return quit_at_eof;
	/* When less_is_more is set, the -e flag semantics are different. */
	return quit_at_eof ? OPT_ON : OPT_ONPLUS;
}

/* Display n more lines, forward.
 * Start just after the line currently displayed at the bottom of the screen.
 */
static void
forward(int n, int force, int only_last)
{
	POSITION pos;

	if (get_quit_at_eof() && hit_eof && !(ch_getflags() & CH_HELPFILE)) {
		/* If the -e flag is set and we're trying to go
		 * forward from end-of-file, go on to the next file. */
		if (edit_next(1))
			quit(QUIT_OK);
		return;
	}

	pos = position(BOTTOM_PLUS_ONE);
	if (pos == NULL_POSITION && (!force || empty_lines(2, sc_height-1))) {
		if (ignore_eoi) {
			/* ignore_eoi is to support A_F_FOREVER.
			 * Back up until there is a line at the bottom
			 * of the screen. */
			if (empty_screen())
				pos = ch_zero();
			else {
				do {
					back(1, position(TOP), 1, 0);
					pos = position(BOTTOM_PLUS_ONE);
				} while (pos == NULL_POSITION);
			}
		} else {
			eof_bell();
			hit_eof++;
			return;
		}
	}
	forw(n, pos, force, only_last, 0);
}

/* Display n more lines, backward.
 * Start just before the line currently displayed at the top of the screen. */
static void
backward(int n, int force, int only_last)
{
	POSITION pos;

	pos = position(TOP);
	if (pos == NULL_POSITION && (!force || position(BOTTOM) == 0)) {
		eof_bell();
		return;
	}
	back(n, pos, force, only_last);
}

/* Output an integer in a given radix. */
static int
iprint_int(int num)
{
	char buf[INT_STRLEN_BOUND(num)];

	inttoa(num, buf);
	putstr(buf);
	return strlen(buf);
}

/* Output a line number in a given radix. */
static int
iprint_linenum(LINENUM num)
{
	char buf[INT_STRLEN_BOUND(num)];

	linenumtoa(num, buf);
	putstr(buf);
	return strlen(buf);
}

/* This function implements printf-like functionality
 * using a more portable argument list mechanism than printf's. */
static int
less_printf(const char *fmt, PARG *parg)
{
	const char *s;
	int col;

	col = 0;
	while (*fmt != '\0') {
		if (*fmt != '%') {
			putchr(*fmt++);
			col++;
		} else {
			++fmt;
			switch (*fmt++) {
			case 's':
				s = parg->p_string;
				parg++;
				while (*s != '\0') {
					putchr(*s++);
					col++;
				}
				break;
			case 'd':
				col += iprint_int(parg->p_int);
				parg++;
				break;
			case 'n':
				col += iprint_linenum(parg->p_linenum);
				parg++;
				break;
			}
		}
	}
	return col;
}

static const char intr_to_abort[] = "... (interrupt to abort)";

/* Output a message in the lower left corner of the screen
 * and don't wait for carriage return.
 * Usually used to warn that we are beginning a potentially
 * time-consuming operation. */
static void
ierror(const char *fmt, PARG *parg)
{
	at_exit();
	clear_bot();
	at_enter(AT_STANDOUT);
	(void) less_printf(fmt, parg);
	putstr(intr_to_abort);
	at_exit();
	flush();
	need_clr = 1;
}

/* Output a message in the lower left corner of the screen
 * and return a single-character response. */
static int
query(const char *fmt, PARG *parg)
{
	int c;
	int col = 0;

	if (any_display && is_tty)
		clear_bot();

	(void) less_printf(fmt, parg);
	c = getchr();

	if (!(any_display && is_tty)) {
		putchr('\n');
		return c;
	}

	lower_left();
	if (col >= sc_width)
		screen_trashed = 1;
	flush();

	return c;
}

/* Output a message in the lower left corner of the screen
 * and wait for carriage return. */
static void
error(const char *fmt, PARG *parg)
{
	int col = 0;
	static const char return_to_continue[] = "  (press RETURN)";

	errmsgs++;

	if (any_display && is_tty) {
		squish_check();
		at_exit();
		clear_bot();
		at_enter(AT_STANDOUT);
	}

	col += less_printf(fmt, parg);

	if (!(any_display && is_tty)) {
		putchr('\n');
		return;
	}

	putstr(return_to_continue);
	at_exit();
	col += sizeof(return_to_continue);

	get_return();
	lower_left();

	if (col >= sc_width)
		/* Printing the message has probably scrolled the screen.
		 * {{ Unless the terminal doesn't have auto margins,
		 *    in which case we just hammered on the right margin. }} */
		screen_trashed = 1;

	flush();
}

static void
hold_ifile(IFILE ifile, int incr)
{
	int_ifile(ifile)->h_hold += incr;
}

static void
unsave_ifile(IFILE save_ifile)
{
	if (save_ifile != NULL_IFILE)
		hold_ifile(save_ifile, -1);
}

static int
edit_inext(IFILE h, int n)
{
	return edit_istep(h, n, +1);
}

static int
edit_iprev(IFILE h, int n)
{
	return edit_istep(h, n, -1);
}

/* Reedit the ifile which was previously open. */
static void
reedit_ifile(IFILE save_ifile)
{
	IFILE next;
	IFILE prev;

	/* Try to reopen the ifile.
	 * Note that opening it may fail (maybe the file was removed),
	 * in which case the ifile will be deleted from the list.
	 * So save the next and prev ifiles first. */
	unsave_ifile(save_ifile);
	next = next_ifile(save_ifile);
	prev = prev_ifile(save_ifile);
	if (edit_ifile(save_ifile) == 0)
		return;
	/* If can't reopen it, open the next input file in the list. */
	if (next != NULL_IFILE && edit_inext(next, 0) == 0)
		return;
	/* If can't open THAT one, open the previous input file in the list. */
	if (prev != NULL_IFILE && edit_iprev(prev, 0) == 0)
		return;
	/* If can't even open that, we're stuck.  Just quit. */
	quit(QUIT_ERROR);
}

/* Routines dealing with signals.
 *
 * A signal usually merely causes a bit to be set in the "signals" word.
 * At some convenient time, the mainline code checks to see if any
 * signals need processing by calling psignal().
 * If we happen to be reading from a file [in iread()] at the time
 * the signal is received, we call intread to interrupt the iread. */

/* Interrupt a pending iread(). */
static void
intread(void)
{
	LONG_JUMP(read_label, 1);
}

/* Interrupt signal handler. */
static void
u_interrupt(int type)
{
	(void)&type;
	signal(SIGINT, u_interrupt);
	sigs |= S_INTERRUPT;
	if (reading)
		intread();
}

#ifdef SIGTSTP
/* "Stop" (^Z) signal handler. */
static void
stop(int type)
{
	(void)&type;
	signal(SIGTSTP, stop);
	sigs |= S_STOP;
	if (reading)
		intread();
}
#endif

#ifdef SIGWINCH
/* "Window" change handler */
static void
winch(int type)
{
	(void)&type;
	signal(SIGWINCH, winch);
	sigs |= S_WINCH;
	if (reading)
		intread();
}
#endif

static void
calc_jump_sline(void)
{
	if (jump_sline_fraction < 0)
		return;
	jump_sline = sc_height * jump_sline_fraction / NUM_FRAC_DENOM;
}

/* Process any signals we have received.
 * A received signal cause a bit to be set in "sigs". */
static void
psignals(void)
{
	int tsignals;

	if ((tsignals = sigs) == 0)
		return;
	sigs = 0;

#ifdef SIGTSTP
	if (tsignals & S_STOP) {
		/* Clean up the terminal. */
#ifdef SIGTTOU
		signal(SIGTTOU, SIG_IGN);
#endif
		clear_bot();
		flush();
		raw_mode(0);
#ifdef SIGTTOU
		signal(SIGTTOU, SIG_DFL);
#endif
		signal(SIGTSTP, SIG_DFL);
		kill(getpid(), SIGTSTP);
		/* ... Bye bye. ...
		 * Hopefully we'll be back later and resume here...
		 * Reset the terminal and arrange to repaint the
		 * screen when we get back to the main command loop. */
		signal(SIGTSTP, stop);
		raw_mode(1);
		init();
		screen_trashed = 1;
		tsignals |= S_WINCH;
	}
#endif
#ifdef S_WINCH
	if (tsignals & S_WINCH) {
		int old_width, old_height;
		/* Re-execute scrsize() to read the new window size. */
		old_width = sc_width;
		old_height = sc_height;
		get_term();
		if (sc_width != old_width || sc_height != old_height) {
			wscroll = (sc_height + 1) / 2;
			calc_jump_sline();
			screen_trashed = 1;
		}
	}
#endif
	if (tsignals & S_INTERRUPT) {
		if (quit_on_intr)
			quit(QUIT_OK);
		bell();
		/* {{ You may wish to replace the bell() with
		 *    error("Interrupt", NULL_PARG); }} */

		/* If we were interrupted while in the "calculating
		 * line numbers" loop, turn off line numbers. */
		if (lnloop) {
			lnloop = 0;
			if (linenums == 2)
				screen_trashed = 1;
			linenums = 0;
			error("Line numbers turned off", NULL_PARG);
		}

	}
}

/* Set up the signal handlers. */
static void
init_signals(int on)
{
	if (on) {
		/* Set signal handlers. */
		(void) signal(SIGINT, u_interrupt);
#ifdef SIGTSTP
		(void) signal(SIGTSTP, stop);
#endif
#ifdef SIGWINCH
		(void) signal(SIGWINCH, winch);
#endif
#ifdef SIGWIND
		(void) signal(SIGWIND, winch);
#endif
#ifdef SIGQUIT
		(void) signal(SIGQUIT, SIG_IGN);
#endif
	} else {
		/* Restore signals to defaults. */
		(void) signal(SIGINT, SIG_DFL);
#ifdef SIGTSTP
		(void) signal(SIGTSTP, SIG_DFL);
#endif
#ifdef SIGWINCH
		(void) signal(SIGWINCH, SIG_IGN);
#endif
#ifdef SIGWIND
		(void) signal(SIGWIND, SIG_IGN);
#endif
#ifdef SIGQUIT
		(void) signal(SIGQUIT, SIG_DFL);
#endif
	}
}

static IFILE
save_curr_ifile(void)
{
	if (curr_ifile != NULL_IFILE)
		hold_ifile(curr_ifile, 1);
	return curr_ifile;
}


/* Get the characters which the shell considers to be "metacharacters". */
static const char *
metachars(void)
{
	if (mchars == NULL) {
		mchars = lgetenv("LESSMETACHARS");
		if (mchars == NULL)
			mchars = DEF_METACHARS;
	}
	return mchars;
}

/* Is this a shell metacharacter? */
static int
metachar(char c)
{
	return (strchr(metachars(), c) != NULL);
}

/* Get the shell's escape character. */
static const char *
get_meta_escape(void)
{
	const char *s;

	s = lgetenv("LESSMETAESCAPE");
	if (s == NULL)
		s = DEF_METAESCAPE;
	return s;
}

/* Insert a backslash before each metacharacter in a string. */
static char *
shell_quote(char *s)
{
	char *p;
	char *newstr;
	int len;
	const char *esc = get_meta_escape();
	int esclen = strlen(esc);
	int use_quotes = 0;
	int have_quotes = 0;

	/* Determine how big a string we need to allocate. */
	len = 1; /* Trailing null byte */
	for (p = s;  *p != '\0';  p++) {
		len++;
		if (*p == openquote || *p == closequote)
			have_quotes = 1;
		if (metachar(*p)) {
			if (esclen == 0) {
				/* We've got a metachar, but this shell
				 * doesn't support escape chars.  Use quotes. */
				use_quotes = 1;
			} else {
				/* Allow space for the escape char. */
				len += esclen;
			}
		}
	}
	if (use_quotes) {
		if (have_quotes)
			/* We can't quote a string that contains quotes. */
			return NULL;
		len = strlen(s) + 3;
	}
	/* Allocate and construct the new string. */
	newstr = p = emalloc(len);
	if (use_quotes) {
		snprintf(newstr, len, "%c%s%c", openquote, s, closequote);
	} else {
		while (*s != '\0') {
			if (metachar(*s)) {
				/* Add the escape char. */
				strcpy(p, esc);
				p += esclen;
			}
			*p++ = *s++;
		}
		*p = '\0';
	}
	return newstr;
}

#if EDITOR || SHELL_ESCAPE
/* Pass the specified command to a shell to be executed.
 * Like plain "system()", but handles resetting terminal modes, etc. */
static void
lsystem(char *cmd, char *donemsg)
{
	int inp;
#if HAVE_SHELL
	const char *shell;
	char *p;
#endif
	IFILE save_ifile;

	/* Print the command which is to be executed,
	 * unless the command starts with a "-". */
	if (cmd[0] == '-')
		cmd++;
	else {
		clear_bot();
		putstr("!");
		putstr(cmd);
		putstr("\n");
	}

	/* Close the current input file. */
	save_ifile = save_curr_ifile();
	(void) edit_ifile(NULL_IFILE);

	/* De-initialize the terminal and take out of raw mode. */
	flush();        /* Make sure the deinit chars get out */
	raw_mode(0);

	/* Restore signals to their defaults. */
	init_signals(0);

	/* Force standard input to be the user's terminal
	 * (the normal standard input), even if less's standard input
	 * is coming from a pipe. */
	inp = dup(0);
	close(0);
	if (open(bb_dev_tty, O_RDONLY) < 0)
		dup(inp);

	/* Pass the command to the system to be executed.
	 * If we have a SHELL environment variable, use
	 * <$SHELL -c "command"> instead of just <command>.
	 * If the command is empty, just invoke a shell. */
#if HAVE_SHELL
	p = NULL;
	if ((shell = lgetenv("SHELL")) != NULL && *shell != '\0') {
		if (*cmd == '\0')
			p = save(shell);
		else {
			char *esccmd = shell_quote(cmd);
			if (esccmd != NULL) {
				int len = strlen(shell) + strlen(esccmd) + 5;
				p = emalloc(len);
				snprintf(p, len, "%s -c %s", shell, esccmd);
				free(esccmd);
			}
		}
	}
	if (p == NULL) {
		if (*cmd == '\0')
			p = save(DEFAULT_SHELL_SHORT_NAME);
		else
			p = save(cmd);
	}
	system(p);
	free(p);
#else
	system(cmd);
#endif

	/* Restore standard input, reset signals, raw mode, etc. */
	close(0);
	dup(inp);
	close(inp);

	init_signals(1);
	raw_mode(1);
	if (donemsg != NULL) {
		putstr(donemsg);
		putstr("  (press RETURN)");
		get_return();
		putchr('\n');
		flush();
	}
	init();
	screen_trashed = 1;

	/* Reopen the current input file. */
	reedit_ifile(save_ifile);

#if defined(SIGWINCH) || defined(SIGWIND)
	/* Since we were ignoring window change signals while we executed
	 * the system command, we must assume the window changed.
	 * Warning: this leaves a signal pending (in "sigs"),
	 * so psignals() should be called soon after lsystem(). */
	winch(0);
#endif
}
#endif /* EDITOR || SHELL_ESCAPE */

/* See if a mark letter is valid (between a and z). */
static struct mark *
getumark(int c)
{
	if (c >= 'a' && c <= 'z')
		return &marks[c-'a'];

	if (c >= 'A' && c <= 'Z')
		return &marks[c-'A'+26];

	error("Invalid mark letter", NULL_PARG);
	return NULL;
}

/* Try to determine the size of a file by seeking to the end. */
static POSITION
seek_filesize(int f)
{
	off_t spos;

	spos = lseek(f, (off_t)0, SEEK_END);
	if (spos == BAD_LSEEK)
		return NULL_POSITION;
	return (POSITION) spos;
}

/* Return the size of a file, as cheaply as possible. */
static POSITION
filesize(int f)
{
	struct stat statbuf;

	if (fstat(f, &statbuf) >= 0)
		return (POSITION) statbuf.st_size;
	return seek_filesize(f);
}

/* Seek to the end of the file. */
static int
ch_end_seek(void)
{
	POSITION len;

	if (thisfile == NULL)
		return 0;

	if (ch_flags & CH_CANSEEK)
		ch_fsize = filesize(ch_file);

	len = ch_length();
	if (len != NULL_POSITION)
		return ch_seek(len);

	/* Do it the slow way: read till end of data. */
	while (ch_forw_get() != EOI)
		if (ABORT_SIGS())
			return 1;
	return 0;
}

/* Get the mark structure identified by a character.
 * The mark struct may come either from the mark table
 * or may be constructed on the fly for certain characters like ^, $. */
static struct mark *
getmark(int c)
{
	struct mark *m;

	switch (c) {
	case '^':
		/* Beginning of the current file. */
		m = &sm;
		m->m_scrpos.pos = ch_zero();
		m->m_scrpos.ln = 0;
		m->m_ifile = curr_ifile;
		break;
	case '$':
		/* End of the current file. */
		if (ch_end_seek()) {
			error("Cannot seek to end of file", NULL_PARG);
			return NULL;
		}
		m = &sm;
		m->m_scrpos.pos = ch_tell();
		m->m_scrpos.ln = sc_height-1;
		m->m_ifile = curr_ifile;
		break;
	case '.':
		/* Current position in the current file. */
		m = &sm;
		get_scrpos(&m->m_scrpos);
		m->m_ifile = curr_ifile;
		break;
	case '\'':
		/* The "last mark". */
		m = &marks[LASTMARK];
		break;
	default:
		/* Must be a user-defined mark. */
		m = getumark(c);
		if (m == NULL)
			break;
		if (m->m_scrpos.pos == NULL_POSITION) {
			error("Mark not set", NULL_PARG);
			return NULL;
		}
		break;
	}
	return m;
}

#if PIPEC

/* Create a pipe to the given shell command.
 * Feed it the file contents between the positions spos and epos. */
static int
pipe_data(char *cmd, POSITION spos, POSITION epos)
{
	FILE *f;
	int c;

	/* This is structured much like lsystem().
	 * Since we're running a shell program, we must be careful
	 * to perform the necessary deinitialization before running
	 * the command, and reinitialization after it. */
	if (ch_seek(spos) != 0) {
		error("Cannot seek to start position", NULL_PARG);
		return -1;
	}

	if ((f = popen(cmd, "w")) == NULL) {
		error("Cannot create pipe", NULL_PARG);
		return -1;
	}
	clear_bot();
	putstr("!");
	putstr(cmd);
	putstr("\n");

	flush();
	raw_mode(0);
	init_signals(0);
#ifdef SIGPIPE
	signal(SIGPIPE, SIG_IGN);
#endif

	c = EOI;
	while (epos == NULL_POSITION || spos++ <= epos) {
		/* Read a character from the file and give it to the pipe. */
		c = ch_forw_get();
		if (c == EOI)
			break;
		if (putc(c, f) == EOF)
			break;
	}

	/* Finish up the last line. */
	while (c != '\n' && c != EOI ) {
		c = ch_forw_get();
		if (c == EOI)
			break;
		if (putc(c, f) == EOF)
			break;
	}

	pclose(f);

#ifdef SIGPIPE
	signal(SIGPIPE, SIG_DFL);
#endif
	init_signals(1);
	raw_mode(1);
	init();
	screen_trashed = 1;
#if defined(SIGWINCH) || defined(SIGWIND)
	/* {{ Probably don't need this here. }} */
	winch(0);
#endif
	return 0;
}


/* Return the position associated with a given mark letter.
 *
 * We don't return which screen line the position
 * is associated with, but this doesn't matter much,
 * because it's always the first non-blank line on the screen. */
static POSITION
markpos(int c)
{
	struct mark *m;

	m = getmark(c);
	if (m == NULL)
		return NULL_POSITION;

	if (m->m_ifile != curr_ifile) {
		error("Mark not in current file", NULL_PARG);
		return NULL_POSITION;
	}
	return m->m_scrpos.pos;
}

/* Pipe a section of the input file into the given shell command.
 * The section to be piped is the section "between" the current
 * position and the position marked by the given letter.
 *
 * If the mark is after the current screen, the section between
 * the top line displayed and the mark is piped.
 * If the mark is before the current screen, the section between
 * the mark and the bottom line displayed is piped.
 * If the mark is on the current screen, or if the mark is ".",
 * the whole current screen is piped. */
static int
pipe_mark(int c, char *cmd)
{
	POSITION mpos, tpos, bpos;

	/* mpos = the marked position.
	 * tpos = top of screen.
	 * bpos = bottom of screen. */
	mpos = markpos(c);
	if (mpos == NULL_POSITION)
		return -1;
	tpos = position(TOP);
	if (tpos == NULL_POSITION)
		tpos = ch_zero();
	bpos = position(BOTTOM);

	if (c == '.')
		return pipe_data(cmd, tpos, bpos);
	else if (mpos <= tpos)
		return pipe_data(cmd, mpos, bpos);
	else if (bpos == NULL_POSITION)
		return pipe_data(cmd, tpos, bpos);
	else
		return pipe_data(cmd, tpos, mpos);
}


#endif


#if PIPEC
/* Is a mark letter is invalid? */
static int
badmark(int c)
{
	return (getmark(c) == NULL);
}
#endif

/* Set a user-defined mark. */
static void
setmark(int c)
{
	struct mark *m;
	struct scrpos scrpos;

	m = getumark(c);
	if (m == NULL)
		return;
	get_scrpos(&scrpos);
	m->m_scrpos = scrpos;
	m->m_ifile = curr_ifile;
}

/* Go to a mark. */
static void
gomark(int c)
{
	struct mark *m;
	struct scrpos scrpos;

	m = getmark(c);
	if (m == NULL)
		return;

	/* If we're trying to go to the lastmark and
	 * it has not been set to anything yet,
	 * set it to the beginning of the current file. */
	if (m == &marks[LASTMARK] && m->m_scrpos.pos == NULL_POSITION) {
		m->m_ifile = curr_ifile;
		m->m_scrpos.pos = ch_zero();
		m->m_scrpos.ln = jump_sline;
	}

	/* If we're using lmark, we must save the screen position now,
	 * because if we call edit_ifile() below, lmark will change.
	 * (We save the screen position even if we're not using lmark.) */
	scrpos = m->m_scrpos;
	if (m->m_ifile != curr_ifile) {
		/* Not in the current file; edit the correct file. */
		if (edit_ifile(m->m_ifile))
			return;
	}

	jump_loc(scrpos.pos, scrpos.ln);
}


/* Clear the marks associated with a specified ifile. */
static void
unmark(IFILE ifile)
{
	int i;

	for (i = 0;  i < NMARKS;  i++)
		if (marks[i].m_ifile == ifile)
			marks[i].m_scrpos.pos = NULL_POSITION;
}

/* Handling functions for command line options.
 *
 * Most options are handled by the generic code in option.c.
 * But all string options, and a few non-string options, require
 * special handling specific to the particular option.
 * This special processing is done by the "handling functions" in this file.
 *
 * Each handling function is passed a "type" and, if it is a string
 * option, the string which should be "assigned" to the option.
 * The type may be one of:
 *      INIT    The option is being initialized from the command line.
 *      TOGGLE  The option is being changed from within the program.
 *      QUERY   The setting of the option is merely being queried. */

/* Remove quotes around a filename. */
static char *
shell_unquote(char *str)
{
	char *name;
	char *p;

	name = p = emalloc(strlen(str)+1);
	if (*str == openquote) {
		str++;
		while (*str != '\0') {
			if (*str == closequote) {
				if (str[1] != closequote)
					break;
				str++;
			}
			*p++ = *str++;
		}
	} else {
		const char *esc = get_meta_escape();
		int esclen = strlen(esc);
		while (*str != '\0') {
			if (esclen > 0 && strncmp(str, esc, esclen) == 0)
				str += esclen;
			*p++ = *str++;
		}
	}
	*p = '\0';
	return name;
}

#if LOGFILE
/* If the user asked for a log file and our input file
 * is standard input, create the log file.
 * We take care not to blindly overwrite an existing file. */
static void
use_logfile(char *filename)
{
	int exists;
	int answer;
	PARG parg;

	if (ch_getflags() & CH_CANSEEK)
		/* Can't currently use a log file on a file that can seek. */
		return;

	/* {{ We could use access() here. }} */
	filename = shell_unquote(filename);
	exists = open(filename, O_RDONLY);
	close(exists);
	exists = (exists >= 0);

	/* Decide whether to overwrite the log file or append to it.
	 * If it doesn't exist we "overwrite" it. */
	if (!exists || force_logfile) {
		/* Overwrite (or create) the log file. */
		answer = 'O';
	} else {
		/* Ask user what to do. */
		parg.p_string = filename;
		answer = query("Warning: \"%s\" exists; Overwrite, Append or Don't log? ", &parg);
	}

loop:
	switch (answer) {
	case 'O': case 'o':
		/* Overwrite: create the file. */
		logfile = creat(filename, 0644);
		break;
	case 'A': case 'a':
		/* Append: open the file and seek to the end. */
		logfile = open(filename, O_APPEND | O_WRONLY);
		if (lseek(logfile, (off_t)0, SEEK_END) == BAD_LSEEK) {
			close(logfile);
			logfile = -1;
		}
		break;
	case 'D': case 'd':
		/* Don't do anything. */
		free(filename);
		return;
	case 'q':
		quit(QUIT_OK);
	default:
		/* Eh? */
		answer = query("Overwrite, Append, or Don't log? (Type \"O\", \"A\", \"D\" or \"q\") ", NULL_PARG);
		goto loop;
	}

	if (logfile < 0) {
		/* Error in opening logfile. */
		parg.p_string = filename;
		error("Cannot write to \"%s\"", &parg);
		free(filename);
		return;
	}
	free(filename);
}


#endif

/* Skip leading spaces in a string. */
static const char *
skipsp(const char *s)
{
	while (*s == ' ' || *s == '\t')
		s++;
	return s;
}


/* Unget a whole string of command characters.
 * The next sequence of getcc()'s will return this string. */
static void
ungetsc(const char *s)
{
	const char *p;

	for (p = s + strlen(s) - 1;  p >= s;  p--)
		ungetcc(*p);
}

#if TAGS
/* Tag type */
enum {
	T_CTAGS,        /* 'tags': standard and extended format (ctags) */
	T_CTAGS_X,      /* stdin: cross reference format (ctags) */
	T_GTAGS,        /* 'GTAGS': function defenition (global) */
	T_GRTAGS,       /* 'GRTAGS': function reference (global) */
	T_GSYMS,        /* 'GSYMS': other symbols (global) */
	T_GPATH         /* 'GPATH': path name (global) */
};

#define WHITESP(c)      ((c)==' ' || (c)=='\t')


enum tag_result {
	TAG_FOUND,
	TAG_NOFILE,
	TAG_NOTAG,
	TAG_NOTYPE,
	TAG_INTR
};


/* The list of tags generated by the last findgtag() call.
 *
 * Use either pattern or line number.
 * findgtag() always uses line number, so pattern is always NULL.
 * findctag() uses either pattern (in which case line number is 0),
 * or line number (in which case pattern is NULL). */
struct tag {
	struct tag *next, *prev; /* List links */
	char *tag_file;         /* Source file containing the tag */
	LINENUM tag_linenum;    /* Appropriate line number in source file */
	char *tag_pattern;      /* Pattern used to find the tag */
	char tag_endline;       /* True if the pattern includes '$' */
};
static struct tag *curtag;
struct taglist {
	struct tag *tl_first;
	struct tag *tl_last;
};
#define TAG_END  ((struct tag *) &taglist)
static struct taglist taglist = { TAG_END, TAG_END };

#define TAG_INS(tp) \
	(tp)->next = TAG_END; \
	(tp)->prev = taglist.tl_last; \
	taglist.tl_last->next = (tp); \
	taglist.tl_last = (tp);

#define TAG_RM(tp) \
	(tp)->next->prev = (tp)->prev; \
	(tp)->prev->next = (tp)->next;

/* Delete tag structures. */
static void
cleantags(void)
{
	struct tag *tp;

	/* Delete any existing tag list.
	 * {{ Ideally, we wouldn't do this until after we know that we
	 *    can load some other tag information. }} */
	while ((tp = taglist.tl_first) != TAG_END) {
		TAG_RM(tp);
		free(tp);
	}
	curtag = NULL;
	total = curseq = 0;
}

/* Create a new tag entry. */
static struct tag *
maketagent(char *file, LINENUM linenum, char *pattern, int endline)
{
	struct tag *tp;

	tp = emalloc(sizeof(struct tag));
	tp->tag_file = emalloc(strlen(file) + 1);
	strcpy(tp->tag_file, file);
	tp->tag_linenum = linenum;
	tp->tag_endline = endline;
	if (pattern == NULL)
		tp->tag_pattern = NULL;
	else {
		tp->tag_pattern = emalloc(strlen(pattern) + 1);
		strcpy(tp->tag_pattern, pattern);
	}
	return tp;
}

/* Search for a tag.
 * This is a stripped-down version of search().
 * We don't use search() for several reasons:
 *   -  We don't want to blow away any search string we may have saved.
 *   -  The various regular-expression functions (from different systems:
 *      regcmp vs. re_comp) behave differently in the presence of
 *      parentheses (which are almost always found in a tag). */
static POSITION
ctagsearch(void)
{
	POSITION pos, linepos;
	LINENUM linenum;
	int len;
	char *line;

	pos = ch_zero();
	linenum = find_linenum(pos);

	for (;;) {
		/* Get lines until we find a matching one or
		 * until we hit end-of-file. */
		if (ABORT_SIGS())
			return NULL_POSITION;

		/* Read the next line, and save the
		 * starting position of that line in linepos. */
		linepos = pos;
		pos = forw_raw_line(pos, &line, (int *)NULL);
		if (linenum != 0)
			linenum++;

		if (pos == NULL_POSITION) {
			/* We hit EOF without a match. */
			error("Tag not found", NULL_PARG);
			return NULL_POSITION;
		}

		/* If we're using line numbers, we might as well
		 * remember the information we have now (the position
		 * and line number of the current line). */
		if (linenums)
			add_lnum(linenum, pos);

		/* Test the line to see if we have a match.
		 * Use strncmp because the pattern may be
		 * truncated (in the tags file) if it is too long.
		 * If tagendline is set, make sure we match all
		 * the way to end of line (no extra chars after the match). */
		len = strlen(curtag->tag_pattern);
		if (strncmp(curtag->tag_pattern, line, len) == 0 &&
		    (!curtag->tag_endline || line[len] == '\0' || line[len] == '\r')) {
			curtag->tag_linenum = find_linenum(linepos);
			break;
		}
	}

	return linepos;
}

/* Position the current file at at what is hopefully the tag that was chosen
 * using either findtag() or one of nextgtag() and prevgtag().  Returns -1
 * if it was unable to position at the tag, 0 if successful. */
static POSITION
gtagsearch(void)
{
	if (curtag == NULL)
		return NULL_POSITION;  /* No gtags loaded! */
	return find_pos(curtag->tag_linenum);
}

/* Search for a tag. */
static POSITION
tagsearch(void)
{
	if (curtag == NULL)
		return NULL_POSITION;  /* No gtags loaded! */
	if (curtag->tag_linenum != 0)
		return gtagsearch();
	else
		return ctagsearch();
}

static int circular;        /* 1: circular tag structure */

/* Return the filename required for the next gtag in the queue that was setup
 * by findgtag().  The next call to gtagsearch() will try to position at the
 * appropriate tag. */
static char *
nextgtag(void)
{
	struct tag *tp;

	if (curtag == NULL)
		/* No tag loaded */
		return NULL;

	tp = curtag->next;
	if (tp == TAG_END) {
		if (!circular)
			return NULL;
		/* Wrapped around to the head of the queue */
		curtag = taglist.tl_first;
		curseq = 1;
	} else {
		curtag = tp;
		curseq++;
	}
	return curtag->tag_file;
}

/* Go to the next tag. */
static char *
nexttag(int n)
{
	char *tagfile = (char *) NULL;

	while (n-- > 0)
		tagfile = nextgtag();
	return tagfile;
}

/* Return the filename required for the previous gtag in the queue that was
 * setup by findgtat().  The next call to gtagsearch() will try to position
 * at the appropriate tag. */
static char *
prevgtag(void)
{
	struct tag *tp;

	if (curtag == NULL)
		/* No tag loaded */
		return NULL;

	tp = curtag->prev;
	if (tp == TAG_END) {
		if (!circular)
			return NULL;
		/* Wrapped around to the tail of the queue */
		curtag = taglist.tl_last;
		curseq = total;
	} else {
		curtag = tp;
		curseq--;
	}
	return curtag->tag_file;
}

/* Go to the previous tag. */
static char *
prevtag(int n)
{
	char *tagfile = (char *) NULL;

	while (n-- > 0)
		tagfile = prevgtag();
	return tagfile;
}
#endif /* TAGS */

static int
num_error(const char *printopt, int *errp)
{
	PARG parg;

	if (errp != NULL) {
		*errp = TRUE;
		return -1;
	}
	if (printopt != NULL) {
		parg.p_string = printopt;
		error("Number is required after %s", &parg);
	}
	quit(QUIT_ERROR);
	return -1;
}

/* Translate a string into a number.
 * Like atoi(), but takes a pointer to a char *, and updates
 * the char * to point after the translated number. */
static int
getnum(const char **sp, const char *printopt, int *errp)
{
	const char *s;
	int n;
	int neg;

	s = skipsp(*sp);
	neg = FALSE;
	if (*s == '-') {
		neg = TRUE;
		s++;
	}
	if (*s < '0' || *s > '9')
		return num_error(printopt, errp);

	n = 0;
	while (*s >= '0' && *s <= '9')
		n = 10 * n + *s++ - '0';
	*sp = s;
	if (errp != NULL)
		*errp = FALSE;
	if (neg)
		n = -n;
	return n;
}

#if TAGS

/*****************************************************************************
 * ctags
 */

/* Find tags in the "tags" file.
 * Sets curtag to the first tag entry. */
static enum tag_result
findctag(char *tag)
{
	char *p;
	FILE *f;
	int taglen;
	LINENUM taglinenum;
	char *tagfile;
	char *tagpattern;
	int tagendline;
	int search_char;
	int err;
	char tline[TAGLINE_SIZE];
	struct tag *tp;

	p = shell_unquote(tags);
	f = fopen(p, "r");
	free(p);
	if (f == NULL)
		return TAG_NOFILE;

	cleantags();
	total = 0;
	taglen = strlen(tag);

	/* Search the tags file for the desired tag. */
	while (fgets(tline, sizeof(tline), f) != NULL) {
		if (tline[0] == '!')
			/* Skip header of extended format. */
			continue;
		if (strncmp(tag, tline, taglen) != 0 || !WHITESP(tline[taglen]))
			continue;

		/* Found it.
		 * The line contains the tag, the filename and the
		 * location in the file, separated by white space.
		 * The location is either a decimal line number,
		 * or a search pattern surrounded by a pair of delimiters.
		 * Parse the line and extract these parts. */
		tagpattern = NULL;

		/* Skip over the whitespace after the tag name. */
		p = skipsp(tline+taglen);
		if (*p == '\0')
			/* File name is missing! */
			continue;

		/* Save the file name.
		 * Skip over the whitespace after the file name. */
		tagfile = p;
		while (!WHITESP(*p) && *p != '\0')
			p++;
		*p++ = '\0';
		p = skipsp(p);
		if (*p == '\0')
			/* Pattern is missing! */
			continue;

		/* First see if it is a line number. */
		tagendline = 0;
		taglinenum = getnum(&p, 0, &err);
		if (err) {
			/* No, it must be a pattern.
			 * Delete the initial "^" (if present) and
			 * the final "$" from the pattern.
			 * Delete any backslash in the pattern. */
			taglinenum = 0;
			search_char = *p++;
			if (*p == '^')
				p++;
			tagpattern = p;
			while (*p != search_char && *p != '\0') {
				if (*p == '\\')
					p++;
				p++;
			}
			tagendline = (p[-1] == '$');
			if (tagendline)
				p--;
			*p = '\0';
		}
		tp = maketagent(tagfile, taglinenum, tagpattern, tagendline);
		TAG_INS(tp);
		total++;
	}
	fclose(f);
	if (total == 0)
		return TAG_NOTAG;
	curtag = taglist.tl_first;
	curseq = 1;
	return TAG_FOUND;
}

/* Edit current tagged file. */
static int
edit_tagfile(void)
{
	if (curtag == NULL)
		return 1;
	return edit(curtag->tag_file);
}


/*******************************************************************************
 * gtags
 */

/* The getentry() parses both standard and extended ctags -x format.
 *
 * [standard format]
 * <tag>   <lineno>  <file>         <image>
 * +------------------------------------------------
 * |main     30      main.c         main(argc, argv)
 * |func     21      subr.c         func(arg)
 *
 * The following commands write this format.
 *      o Traditinal Ctags with -x option
 *      o Global with -x option
 *              See <http://www.gnu.org/software/global/global.html>
 *
 * [extended format]
 * <tag>   <type>  <lineno>   <file>        <image>
 * +----------------------------------------------------------
 * |main     function 30      main.c         main(argc, argv)
 * |func     function 21      subr.c         func(arg)
 *
 * The following commands write this format.
 *      o Exuberant Ctags with -x option
 *              See <http://ctags.sourceforge.net>
 *
 * Returns 0 on success, -1 on error.
 * The tag, file, and line will each be NUL-terminated pointers
 * into buf.
 * buf - standard or extended ctags -x format data
 * tag - name of the tag we actually found
 * file - file in which to find this tag
 * line - line number of file where this tag is found */
static int
getentry(char *buf, char **tag, char **file, char **line)
{
	char *p = buf;

	for (*tag = p;  *p && !IS_SPACE(*p);  p++)      /* tag name */
		;
	if (*p == 0)
		return -1;
	*p++ = 0;
	for ( ;  *p && IS_SPACE(*p);  p++)              /* (skip blanks) */
		;
	if (*p == 0)
		return -1;
	/* If the second part begin with other than digit,
	 * it is assumed tag type. Skip it. */
	if (!IS_DIGIT(*p)) {
		for ( ;  *p && !IS_SPACE(*p);  p++)     /* (skip tag type) */
			;
		for (;  *p && IS_SPACE(*p);  p++)       /* (skip blanks) */
			;
	}
	if (!IS_DIGIT(*p))
		return -1;
	*line = p;                                      /* line number */
	for (*line = p;  *p && !IS_SPACE(*p);  p++)
		;
	if (*p == 0)
		return -1;
	*p++ = 0;
	for ( ; *p && IS_SPACE(*p);  p++)               /* (skip blanks) */
		;
	if (*p == 0)
		return -1;
	*file = p;                                      /* file name */
	for (*file = p;  *p && !IS_SPACE(*p);  p++)
		;
	if (*p == 0)
		return -1;
	*p = 0;

	/* value check */
	if (strlen(*tag) && strlen(*line) && strlen(*file) && atoi(*line) > 0)
		return 0;
	return -1;
}


/* Find tags in the GLOBAL's tag file.
 * The findgtag() will try and load information about the requested tag.
 * It does this by calling "global -x tag" and storing the parsed output
 * for future use by gtagsearch().
 * Sets curtag to the first tag entry. */
static enum tag_result
findgtag(char *tag, int type)
{
	char buf[256];
	FILE *fp;
	struct tag *tp;

	if (type != T_CTAGS_X && tag == NULL)
		return TAG_NOFILE;

	cleantags();
	total = 0;

	/* If type == T_CTAGS_X then read ctags's -x format from stdin
	 * else execute global(1) and read from it. */
	if (type == T_CTAGS_X) {
		fp = stdin;
		/* Set tag default because we cannot read stdin again. */
		tags = "tags";
	} else {
		char *command;
		char *flag;
		char *qtag;
		char *cmd = lgetenv("LESSGLOBALTAGS");

		if (cmd == NULL || *cmd == '\0')
			return TAG_NOFILE;
		/* Get suitable flag value for global(1). */
		switch (type) {
		case T_GTAGS:
			flag = "" ;
			break;
		case T_GRTAGS:
			flag = "r";
			break;
		case T_GSYMS:
			flag = "s";
			break;
		case T_GPATH:
			flag = "P";
			break;
		default:
			return TAG_NOTYPE;
		}

		/* Get our data from global(1). */
		qtag = shell_quote(tag);
		if (qtag == NULL)
			qtag = tag;
		command = emalloc(strlen(cmd) + strlen(flag) +
				strlen(qtag) + 5);
		sprintf(command, "%s -x%s %s", cmd, flag, qtag);
		if (qtag != tag)
			free(qtag);
		fp = popen(command, "r");
		free(command);
	}
	if (fp != NULL) {
		while (fgets(buf, sizeof(buf), fp)) {
			char *name, *file, *line;
			int len;

			if (sigs) {
				if (fp != stdin)
					pclose(fp);
				return TAG_INTR;
			}
			len = strlen(buf);
			if (len > 0 && buf[len-1] == '\n')
				buf[len-1] = '\0';
			else {
				int c;
				do {
					c = fgetc(fp);
				} while (c != '\n' && c != EOF);
			}

			if (getentry(buf, &name, &file, &line)) {
				/* Couldn't parse this line for some reason.
				 * We'll just pretend it never happened. */
				break;
			}

			/* Make new entry and add to list. */
			tp = maketagent(file, (LINENUM) atoi(line), NULL, 0);
			TAG_INS(tp);
			total++;
		}
		if (fp != stdin) {
			if (pclose(fp)) {
				curtag = NULL;
				total = curseq = 0;
				return TAG_NOFILE;
			}
		}
	}

	/* Check to see if we found anything. */
	tp = taglist.tl_first;
	if (tp == TAG_END)
		return TAG_NOTAG;
	curtag = tp;
	curseq = 1;
	return TAG_FOUND;
}



/* Get tag mode. */
static int
gettagtype(void)
{
	int f;

	if (strcmp(tags, "GTAGS") == 0)
		return T_GTAGS;
	if (strcmp(tags, "GRTAGS") == 0)
		return T_GRTAGS;
	if (strcmp(tags, "GSYMS") == 0)
		return T_GSYMS;
	if (strcmp(tags, "GPATH") == 0)
		return T_GPATH;
	if (strcmp(tags, "-") == 0)
		return T_CTAGS_X;
	f = open(tags, O_RDONLY);
	if (f >= 0) {
		close(f);
		return T_CTAGS;
	}
	return T_GTAGS;
}

/* Find tags in tag file.
 * Find a tag in the "tags" file.
 * Sets "tag_file" to the name of the file containing the tag,
 * and "tagpattern" to the search pattern which should be used
 * to find the tag. */
static void
findtag(char *tag)
{
	int type = gettagtype();
	enum tag_result result;

	if (type == T_CTAGS)
		result = findctag(tag);
	else
		result = findgtag(tag, type);
	switch (result) {
	case TAG_FOUND:
	case TAG_INTR:
		break;
	case TAG_NOFILE:
		error("No tags file", NULL_PARG);
		break;
	case TAG_NOTAG:
		error("No such tag in tags file", NULL_PARG);
		break;
	case TAG_NOTYPE:
		error("unknown tag type", NULL_PARG);
		break;
	}
}

/* Handler for -t option. */
static void
opt_t(int type, char *s)
{
	IFILE save_ifile;
	POSITION pos;

	switch (type) {
	case INIT:
		tagoption = s;
		/* Do the rest in main() */
		break;
	case TOGGLE:
		if (secure) {
			error("tags support is not available", NULL_PARG);
			break;
		}
		findtag(skipsp(s));
		save_ifile = save_curr_ifile();
		/* Try to open the file containing the tag
		 * and search for the tag in that file. */
		if (edit_tagfile() || (pos = tagsearch()) == NULL_POSITION) {
			/* Failed: reopen the old file. */
			reedit_ifile(save_ifile);
			break;
		}
		unsave_ifile(save_ifile);
		jump_loc(pos, jump_sline);
		break;
	}
}
#endif

#if TAGS || TAB_COMPLETE_FILENAME || EXAMINE || LOGFILE || SHELL_ESCAPE
/* Expand a string, substituting any "%" with the current filename,
 * and any "#" with the previous filename.
 * But a string of N "%"s is just replaced with N-1 "%"s.
 * Likewise for a string of N "#"s.
 * {{ This is a lot of work just to support % and #. }} */
static char *
fexpand(char *s)
{
	char *fr, *to;
	int n;
	char *e;
	IFILE ifile;

#define fchar_ifile(c) \
	((c) == '%' ? curr_ifile : \
	 (c) == '#' ? old_ifile : NULL_IFILE)

	/* Make one pass to see how big a buffer we
	 * need to allocate for the expanded string. */
	n = 0;
	for (fr = s;  *fr != '\0';  fr++) {
		switch (*fr) {
		case '%':
		case '#':
			if (fr > s && fr[-1] == *fr) {
				/* Second (or later) char in a string
				 * of identical chars.  Treat as normal. */
				n++;
			} else if (fr[1] != *fr) {
				/* Single char (not repeated).  Treat specially. */
				ifile = fchar_ifile(*fr);
				if (ifile == NULL_IFILE)
					n++;
				else
					n += strlen(get_filename(ifile));
			}
			/* Else it is the first char in a string of
			 * identical chars.  Just discard it. */
			break;
		default:
			n++;
			break;
		}
	}

	e = emalloc(n+1);

	/* Now copy the string, expanding any "%" or "#". */
	to = e;
	for (fr = s;  *fr != '\0';  fr++) {
		switch (*fr) {
		case '%':
		case '#':
			if (fr > s && fr[-1] == *fr) {
				*to++ = *fr;
			} else if (fr[1] != *fr) {
				ifile = fchar_ifile(*fr);
				if (ifile == NULL_IFILE)
					*to++ = *fr;
				else {
					strcpy(to, get_filename(ifile));
					to += strlen(to);
				}
			}
			break;
		default:
			*to++ = *fr;
			break;
		}
	}
	*to = '\0';
	return e;
}
#endif

/* Execute a shell command.
 * Return a pointer to a pipe connected to the shell command's standard output. */
static FILE *
shell_cmd(char *cmd)
{
	FILE *fd;

#if HAVE_SHELL
	const char *shell;

	shell = lgetenv("SHELL");
	if (shell != NULL && *shell != '\0') {
		char *scmd;
		char *esccmd;

		/* Read the output of <$SHELL -c cmd>.
		 * Escape any metacharacters in the command. */
		esccmd = shell_quote(cmd);
		if (esccmd == NULL) {
			fd = popen(cmd, "r");
		} else {
			int len = strlen(shell) + strlen(esccmd) + 5;
			scmd = emalloc(len);
			snprintf(scmd, len, "%s -c %s", shell, esccmd);
			free(esccmd);
			fd = popen(scmd, "r");
			free(scmd);
		}
	} else
#endif
	fd = popen(cmd, "r");
	/* Redirection in `popen' might have messed with the
	 * standard devices.  Restore binary input mode. */
	return fd;
}

/* Read a string from a file.
 * Return a pointer to the string in memory. */
static char *
readfd(FILE *fd)
{
	int len;
	int ch;
	char *buf;
	char *p;

	/* Make a guess about how many chars in the string
	 * and allocate a buffer to hold it. */
	len = 100;
	buf = emalloc(len);
	for (p = buf;  ;  p++) {
		if ((ch = getc(fd)) == '\n' || ch == EOF)
			break;
		if (p - buf >= len-1) {
			/* The string is too big to fit in the buffer we have.
			 * Allocate a new buffer, twice as big. */
			len *= 2;
			*p = '\0';
			p = emalloc(len);
			strcpy(p, buf);
			free(buf);
			buf = p;
			p = buf + strlen(buf);
		}
		*p = ch;
	}
	*p = '\0';
	return buf;
}

#if TAGS || TAB_COMPLETE_FILENAME || EXAMINE || LOGFILE
/* Expand a filename, doing any system-specific metacharacter substitutions */
static char *
lglob(char *filename)
{
	char *gfilename;
	char *ofilename;

	ofilename = fexpand(filename);
	if (secure)
		return ofilename;
	filename = shell_unquote(ofilename);

#ifdef DECL_GLOB_LIST
{
	/* The globbing function returns a list of names. */
	int length;
	char *p;
	char *qfilename;
	DECL_GLOB_LIST(list)

	GLOB_LIST(filename, list);
	if (GLOB_LIST_FAILED(list)) {
		free(filename);
		return ofilename;
	}
	length = 1; /* Room for trailing null byte */
	for (SCAN_GLOB_LIST(list, p)) {
		INIT_GLOB_LIST(list, p);
		qfilename = shell_quote(p);
		if (qfilename != NULL) {
			length += strlen(qfilename) + 1;
			free(qfilename);
		}
	}
	gfilename = emalloc(length);
	for (SCAN_GLOB_LIST(list, p)) {
		INIT_GLOB_LIST(list, p);
		qfilename = shell_quote(p);
		if (qfilename != NULL) {
			sprintf(gfilename + strlen(gfilename), "%s ", qfilename);
			free(qfilename);
		}
	}
	/* Overwrite the final trailing space with a null terminator. */
	*--p = '\0';
	GLOB_LIST_DONE(list);
}
#else
#ifdef DECL_GLOB_NAME
{
	/* The globbing function returns a single name, and
	 * is called multiple times to walk thru all names. */
	char *p;
	int len;
	int n;
	char *pathname;
	char *qpathname;
	DECL_GLOB_NAME(fnd,drive,dir,fname,ext,handle)

	GLOB_FIRST_NAME(filename, &fnd, handle);
	if (GLOB_FIRST_FAILED(handle)) {
		free(filename);
		return ofilename;
	}

	_splitpath(filename, drive, dir, fname, ext);
	len = 100;
	gfilename = emalloc(len);
	p = gfilename;
	do {
		n = strlen(drive) + strlen(dir) + strlen(fnd.GLOB_NAME) + 1;
		pathname = emalloc(n);
		snprintf(pathname, n, "%s%s%s", drive, dir, fnd.GLOB_NAME);
		qpathname = shell_quote(pathname);
		free(pathname);
		if (qpathname != NULL) {
			n = strlen(qpathname);
			while (p - gfilename + n + 2 >= len) {
				/* No room in current buffer.
				 * Allocate a bigger one. */
				len *= 2;
				*p = '\0';
				p = emalloc(len);
				strcpy(p, gfilename);
				free(gfilename);
				gfilename = p;
				p = gfilename + strlen(gfilename);
			}
			strcpy(p, qpathname);
			free(qpathname);
			p += n;
			*p++ = ' ';
		}
	} while (GLOB_NEXT_NAME(handle, &fnd) == 0);

	/* Overwrite the final trailing space with a null terminator. */
	*--p = '\0';
	GLOB_NAME_DONE(handle);
}
#else
{
	/* We get the shell to glob the filename for us by passing
	 * an "echo" command to the shell and reading its output. */
	FILE *fd;
	char *s;
	char *lessecho;
	char *cmd;
	char *esc;
	int len;

	esc = get_meta_escape();
	if (strlen(esc) == 0)
		esc = "-";
	esc = shell_quote(esc);
	if (esc == NULL) {
		free(filename);
		return ofilename;
	}
	lessecho = lgetenv("LESSECHO");
	if (lessecho == NULL || *lessecho == '\0')
		lessecho = "lessecho";
	/* Invoke lessecho, and read its output (a globbed list of filenames). */
	len = strlen(lessecho) + strlen(ofilename) + (7*strlen(metachars())) + 24;
	cmd = emalloc(len);
	snprintf(cmd, len, "%s -p0x%x -d0x%x -e%s ", lessecho, openquote, closequote, esc);
	free(esc);
	for (s = metachars();  *s != '\0';  s++)
		sprintf(cmd + strlen(cmd), "-n0x%x ", *s);
	sprintf(cmd + strlen(cmd), "-- %s", ofilename);
	fd = shell_cmd(cmd);
	free(cmd);
	if (fd == NULL) {
		/* Cannot create the pipe.
		 * Just return the original (fexpanded) filename. */
		free(filename);
		return ofilename;
	}
	gfilename = readfd(fd);
	pclose(fd);
	if (*gfilename == '\0') {
		free(gfilename);
		free(filename);
		return ofilename;
	}
}
#endif
#endif
	free(filename);
	free(ofilename);
	return gfilename;
}
#endif /* TAGS || TAB_COMPLETE_FILENAME || EXAMINE || LOGFILE */

#if TAGS
/* Handler for -T option. */
static void
opt__T(int type, char *s)
{
	PARG parg;

	switch (type) {
	case INIT:
		tags = s;
		break;
	case TOGGLE:
		s = skipsp(s);
		tags = lglob(s);
		break;
	case QUERY:
		parg.p_string = tags;
		error("Tags file \"%s\"", &parg);
		break;
	}
}
#endif

/* Handler for -p option. */
static void
opt_p(int type, const char *s)
{
	switch (type) {
	case INIT:
		/* Unget a search command for the specified string.
		 * {{ This won't work if the "/" command is
		 *    changed or invalidated by a .lesskey file. }} */
		plusoption = TRUE;
		ungetsc(s);
		/* In "more" mode, the -p argument is a command,
		 * not a search string, so we don't need a slash. */
		if (!less_is_more)
			ungetsc("/");
		break;
	}
}



/* Handler for -P option. */
static void
opt__P(int type, const char *s)
{
	char **proto;
	PARG parg;

	switch (type) {
	case INIT:
	case TOGGLE:
		/* Figure out which prototype string should be changed. */
		switch (*s) {
		case 's':  proto = &prproto[PR_SHORT];  s++;    break;
		case 'm':  proto = &prproto[PR_MEDIUM]; s++;    break;
		case 'M':  proto = &prproto[PR_LONG];   s++;    break;
		case '=':  proto = &eqproto;            s++;    break;
		case 'h':  proto = &hproto;             s++;    break;
		case 'w':  proto = &wproto;             s++;    break;
		default:   proto = &prproto[PR_SHORT];          break;
		}
		free(*proto);
		*proto = save(s);
		break;
	case QUERY:
		parg.p_string = prproto[pr_type];
		error("%s", &parg);
		break;
	}
}


/* Set max amount of buffer space.
 * bufspace is in units of 1024 bytes.  -1 mean no limit. */
static void
ch_setbufspace(int set_bufspace)
{
	if (set_bufspace < 0)
		maxbufs = -1;
	else {
		maxbufs = ((set_bufspace * 1024) + LBUFSIZE-1) / LBUFSIZE;
		if (maxbufs < 1)
			maxbufs = 1;
	}
}

/* Forget that we have a compiled pattern. */
static void
uncompile_pattern(void)
{
#ifndef NO_REGEX
	if (regpattern != NULL)
		regfree(regpattern);
	regpattern = NULL;
#endif
	last_pattern = NULL;
}

/* Change the caseless-ness of searches.
 * Updates the internal search state to reflect a change in the -i flag. */
static void
chg_caseless(void)
{
	if (!is_ucase_pattern)
		/* Pattern did not have uppercase.
		 * Just set the search caselessness to the global caselessness. */
		is_caseless = caseless;
	else
		/* Pattern did have uppercase.
		 * Discard the pattern; we can't change search caselessness now. */
		uncompile_pattern();
}


/* Handler for the -i option. */
static void
opt_i(int type, const char *s)
{
	(void)&s;
	switch (type) {
	case TOGGLE:
		chg_caseless();
		break;
	case QUERY:
	case INIT:
		break;
	}
}

/* Display the less version message. */
static void
dispversion(void)
{
	PARG parg;

	parg.p_string = version;
	error("less %s", &parg);
}

/* Handler for the -V option. */
static void
opt__V(int type, const char *s)
{
	(void)&s;
	switch (type) {
	case TOGGLE:
	case QUERY:
		dispversion();
		break;
	case INIT:
		/* Force output to stdout per GNU standard for --version output. */
		any_display = 1;
		putstr("less ");
		putstr(version);
		putstr("\nCopyright (C) 1984-2007 Mark Nudelman\n\n");
		putstr("less comes with NO WARRANTY, to the extent permitted by law.\n");
		putstr("For information about the terms of redistribution,\n");
		putstr("see the file named README in the less distribution.\n");
		putstr("Homepage: http://www.greenwoodsoftware.com/less\n");
		quit(QUIT_OK);
	}
}

/* Handler for the -x option. */
static void
opt_x(int type, const char *s)
{
	char msg[60+(4*TABSTOP_MAX)];
	int i;
	PARG p;

	switch (type) {
	case INIT:
	case TOGGLE:
		/* Start at 1 because tabstops[0] is always zero. */
		for (i = 1;  i < TABSTOP_MAX;  ) {
			int n = 0;
			s = skipsp(s);
			while (*s >= '0' && *s <= '9')
				n = (10 * n) + (*s++ - '0');
			if (n > tabstops[i-1])
				tabstops[i++] = n;
			s = skipsp(s);
			if (*s++ != ',')
				break;
		}
		if (i < 2)
			return;
		ntabstops = i;
		tabdefault = tabstops[ntabstops-1] - tabstops[ntabstops-2];
		break;
	case QUERY:
		strcpy(msg, "Tab stops ");
		if (ntabstops > 2) {
			for (i = 1;  i < ntabstops;  i++) {
				if (i > 1)
					strcat(msg, ",");
				sprintf(msg+strlen(msg), "%d", tabstops[i]);
			}
			sprintf(msg+strlen(msg), " and then ");
		}
		sprintf(msg+strlen(msg), "every %d spaces",
			tabdefault);
		p.p_string = msg;
		error("%s", &p);
		break;
	}
}


/* Handler for the -" option. */
static void
opt_quote(int type, const char *s)
{
	char buf[3];
	PARG parg;

	switch (type) {
	case INIT:
	case TOGGLE:
		if (s[0] == '\0') {
			openquote = closequote = '\0';
			break;
		}
		if (s[1] != '\0' && s[2] != '\0') {
			error("-\" must be followed by 1 or 2 chars", NULL_PARG);
			return;
		}
		openquote = s[0];
		if (s[1] == '\0')
			closequote = openquote;
		else
			closequote = s[1];
		break;
	case QUERY:
		buf[0] = openquote;
		buf[1] = closequote;
		buf[2] = '\0';
		parg.p_string = buf;
		error("quotes %s", &parg);
		break;
	}
}

/* "-?" means display a help message.
 * If from the command line, exit immediately. */
static void
opt_query(int type, const char *s)
{
	(void)&s;
	switch (type) {
	case QUERY:
	case TOGGLE:
		error("Use \"h\" for help", NULL_PARG);
		break;
	case INIT:
		dohelp = 1;
	}
}

/* Get the "screen window" size. */
static int
get_swindow(void)
{
	if (swindow > 0)
		return swindow;
	return sc_height + swindow;
}

/* "Toggle" a triple-valued option. */
static int
flip_triple(int val, int lc)
{
	if (lc)
		return ((val == OPT_ON) ? OPT_OFF : OPT_ON);
	else
		return ((val == OPT_ONPLUS) ? OPT_OFF : OPT_ONPLUS);
}

/* Return a string suitable for printing as the "name" of an option.
 * For example, if the option letter is 'x', just return "-x". */
static char *
propt(int c)
{
	char buf[32];

	sprintf(buf2, "-%s", prchar(c, buf));
	return buf2;
}

static const struct loption *findopt(int c);

/* Determine if an option is a single character option (BOOL or TRIPLE),
 * or if it a multi-character option (NUMBER). */
static int
single_char_option(int c)
{
	const struct loption *o;

	o = findopt(c);
	if (o == NULL)
		return TRUE;
	return ((o->otype & (BOOL|TRIPLE|NOVAR|NO_TOGGLE)) != 0);
}

/* Return the prompt to be used for a given option letter.
 * Only string and number valued options have prompts. */
static char *
opt_prompt(int c)
{
	const struct loption *o;

	o = findopt(c);
	if (o == NULL || (o->otype & (STRING|NUMBER)) == 0)
		return NULL;
	return o->odesc[0];
}

/* Return whether or not there is a string option pending;
 * that is, if the previous option was a string-valued option letter
 * (like -P) without a following string.
 * In that case, the current option is taken to be the string for
 * the previous option. */
#define isoptpending() (pendopt != NULL)

/* Print error message about missing string. */
static void
nostring(const char *printopt)
{
	PARG parg;
	parg.p_string = printopt;
	error("Value is required after %s", &parg);
}

/* Print error message if a STRING type option is not followed by a string. */
static void
nopendopt(void)
{
	nostring(propt(pendopt->oletter));
}

/* Scan to end of string or to an END_OPTION_STRING character.
 * In the latter case, replace the char with a null char.
 * Return a pointer to the remainder of the string, if any. */
static char *
optstring(char *s, char **p_str, const char *printopt, const char *validchars)
{
	char *p;

	if (*s == '\0') {
		nostring(printopt);
		quit(QUIT_ERROR);
	}
	*p_str = s;
	for (p = s;  *p != '\0';  p++) {
		if (*p == END_OPTION_STRING ||
		    (validchars != NULL && strchr(validchars, *p) == NULL)) {
			switch (*p) {
			case END_OPTION_STRING:
			case ' ':  case '\t':  case '-':
				/* Replace the char with a null to terminate string. */
				*p++ = '\0';
				break;
			default:
				/* Cannot replace char; make a copy of the string. */
				*p_str = emalloc(p-s+1);
				strncpy(*p_str, s, p-s);
				(*p_str)[p-s] = '\0';
				break;
			}
			break;
		}
	}
	return p;
}

/* Translate a string into a fraction, represented by the part of a
 * number which would follow a decimal point.
 * The value of the fraction is returned as parts per NUM_FRAC_DENOM.
 * That is, if "n" is returned, the fraction intended is n/NUM_FRAC_DENOM. */
static long
getfraction(const char **sp, char *printopt, int *errp)
{
	const char *s;
	long frac = 0;
	int fraclen = 0;

	s = skipsp(*sp);
	if (*s < '0' || *s > '9')
		return num_error(printopt, errp);

	for ( ;  *s >= '0' && *s <= '9';  s++) {
		frac = (frac * 10) + (*s - '0');
		fraclen++;
	}
	if (fraclen > NUM_LOG_FRAC_DENOM)
		while (fraclen-- > NUM_LOG_FRAC_DENOM)
			frac /= 10;
	else
		while (fraclen++ < NUM_LOG_FRAC_DENOM)
			frac *= 10;
	*sp = s;
	if (errp != NULL)
		*errp = FALSE;
	return frac;
}

/* See how many characters of two strings are identical.
 * If uppercase is true, the first string must begin with an uppercase
 * character; the remainder of the first string may be either case. */
static int
sprefix(const char *ps, const char *s, int uppercase)
{
	int c;
	int sc;
	int len = 0;

	for ( ;  *s != '\0';  s++, ps++) {
		c = *ps;
		if (uppercase) {
			if (len == 0 && ASCII_IS_LOWER(c))
				return -1;
			if (ASCII_IS_UPPER(c))
				c = ASCII_TO_LOWER(c);
		}
		sc = *s;
		if (len > 0 && ASCII_IS_UPPER(sc))
			sc = ASCII_TO_LOWER(sc);
		if (c != sc)
			break;
		len++;
	}
	return len;
}


/* Like read() system call, but is deliberately interruptible.
 * A call to intread() from a signal handler will interrupt
 * any pending iread(). */
static int
iread(int fd, char *buf, unsigned int len)
{
	int n;

start:
	if (SET_JUMP(read_label)) {
		/* We jumped here from intread. */
		reading = 0;
		{
		  sigset_t mask;
		  sigemptyset(&mask);
		  sigprocmask(SIG_SETMASK, &mask, NULL);
		}
		return READ_INTR;
	}

	flush();
	reading = 1;
	n = read(fd, buf, len);

	/* This is a kludge to workaround a problem on some systems
	 * where terminating a remote tty connection causes read() to
	 * start returning 0 forever, instead of -1. */
	{
		if (!ignore_eoi) {
			if (n == 0)
				consecutive_nulls++;
			else
				consecutive_nulls = 0;
			if (consecutive_nulls > 20)
				quit(QUIT_ERROR);
		}
	}

	reading = 0;
	if (n < 0) {
		/* Certain values of errno indicate we should just retry the read. */
#ifdef EINTR
		if (errno == EINTR)
			goto start;
#endif
#ifdef EAGAIN
		if (errno == EAGAIN)
			goto start;
#endif
		return -1;
	}
	return n;
}


/* errno_message: Return an error message based on the value of "errno". */
static char *
errno_message(char *filename)
{
	char *p;
	char *m;
	int len;
	p = strerror(errno);
	len = strlen(filename) + strlen(p) + 3;
	m = emalloc(len);
	snprintf(m, len, "%s: %s", filename, p);
	return m;
}

/* Return the specified percentage of a POSITION. */
static POSITION
percent_pos(POSITION pos, int percent)
{
	/* Change percent (parts per 100) to perden (parts per NUM_FRAC_DENOM). */
	long perden = (percent * (NUM_FRAC_DENOM / 100)) + (fraction / 100);
	POSITION temp;

	if (perden == 0)
		return 0;
	temp = pos * perden;  /* This might overflow. */
	if (temp / perden == pos)
		/* No overflow */
		return temp / NUM_FRAC_DENOM;
	else
		/* Above calculation overflows;
		 * use a method that is less precise but won't overflow. */
		return perden * (pos / NUM_FRAC_DENOM);
}

/* Routines dealing with the "position" table.
 * This is a table which tells the position (in the input file) of the
 * first char on each currently displayed line.
 *
 * {{ The position table is scrolled by moving all the entries.
 *    Would be better to have a circular table
 *    and just change a couple of pointers. }} */


/* Allocate or reallocate the position table. */
static void
pos_init(void)
{
	struct scrpos scrpos;

	if (sc_height <= table_size)
		return;
	/* If we already have a table, remember the first line in it
	 * before we free it, so we can copy that line to the new table. */
	if (table != NULL) {
		get_scrpos(&scrpos);
		free((char*)table);
	} else
		scrpos.pos = NULL_POSITION;
	table = emalloc(sc_height * sizeof(POSITION));
	table_size = sc_height;
	pos_clear();
	if (scrpos.pos != NULL_POSITION)
		table[scrpos.ln-1] = scrpos.pos;
}

/* Prompting and other messages.
 * There are three flavors of prompts, SHORT, MEDIUM and LONG,
 * selected by the -m/-M options.
 * There is also the "equals message", printed by the = command.
 * A prompt is a message composed of various pieces, such as the
 * name of the file being viewed, the percentage into the file, etc. */


/* Return the value of a prototype conditional.
 * A prototype string may include conditionals which consist of a
 * question mark followed by a single letter.
 * Here we decode that letter and return the appropriate boolean value. */
static int
cond(char c, int where)
{
	POSITION len;

	switch (c) {
	case 'a':       /* Anything in the message yet? */
		return mp > message;
	case 'b':       /* Current byte offset known? */
		return curr_byte(where) != NULL_POSITION;
	case 'c':
		return (hshift != 0);
	case 'e':       /* At end of file? */
		return hit_eof;
	case 'f':       /* Filename known? */
		return (strcmp(get_filename(curr_ifile), "-") != 0);
	case 'l':       /* Line number known? */
	case 'd':       /* Same as l */
		return linenums;
	case 'L':       /* Final line number known? */
	case 'D':       /* Final page number known? */
		return (linenums && ch_length() != NULL_POSITION);
	case 'm':       /* More than one file? */
#if TAGS
		return (ntags() ? (ntags() > 1) : (nifile() > 1));
#else
		return (nifile() > 1);
#endif
	case 'n':       /* First prompt in a new file? */
#if TAGS
		return (ntags() ? 1 : new_file);
#else
		return new_file;
#endif
	case 'p':       /* Percent into file (bytes) known? */
		return (curr_byte(where) != NULL_POSITION &&
				ch_length() > 0);
	case 'P':       /* Percent into file (lines) known? */
		return (currline(where) != 0 &&
				(len = ch_length()) > 0 &&
				find_linenum(len) != 0);
	case 's':       /* Size of file known? */
	case 'B':
		return (ch_length() != NULL_POSITION);
	case 'x':       /* Is there a "next" file? */
#if TAGS
		if (ntags())
			return 0;
#endif
		return (next_ifile(curr_ifile) != NULL_IFILE);
	}
	return 0;
}


/* Return a prompt.
 * This depends on the prompt type (SHORT, MEDIUM, LONG), etc.
 * If we can't come up with an appropriate prompt, return NULL
 * and the caller will prompt with a colon. */
static char *
pr_string(void)
{
	char *prompt;
	int type;

	type = (!less_is_more) ? pr_type : pr_type ? 0 : 1;
	prompt = pr_expand((ch_getflags() & CH_HELPFILE) ?
				hproto : prproto[type],
			sc_width-2);
	new_file = 0;
	return prompt;
}


/* Get size of the output screen. */
static void
scrsize(void)
{
	const char *s;
	int sys_height;
	int sys_width;

#define DEF_SC_WIDTH    80
#define DEF_SC_HEIGHT   24


	get_terminal_width_height(&sys_width, &sys_height);

	if (sys_height > 0)
		sc_height = sys_height;
	else if ((s = lgetenv("LINES")) != NULL)
		sc_height = atoi(s);
	else
		sc_height = DEF_SC_HEIGHT;

	if (sys_width > 0)
		sc_width = sys_width;
	else if ((s = lgetenv("COLUMNS")) != NULL)
		sc_width = atoi(s);
	else
		sc_width = DEF_SC_WIDTH;
}

/* Return the characters actually input by a "special" key. */
static const char *
special_key_str(int key)
{
	char *s;

	switch (key) {
	case SK_RIGHT_ARROW:
		s = "\033[C";
		break;
	case SK_LEFT_ARROW:
		s = "\033[D";
		break;
	case SK_UP_ARROW:
		s = "\033[A";
		break;
	case SK_DOWN_ARROW:
		s = "\033[B";
		break;
	case SK_PAGE_UP:
		s = "\033[5~";
		break;
	case SK_PAGE_DOWN:
		s = "\033[6~";
		break;
	case SK_HOME:
		s = "\033[1~";
		break;
	case SK_END:
		s = "\033[4~";
		break;
	case SK_DELETE:
		s = erase_char == '\010' ? "\177" : "\033[3~";
		break;
	case SK_CONTROL_K:
		s = "\013";     /* CONTROL('K') */
		break;
	default:
		return NULL;
	}
	return s;
}

static void
get_term(void)
{
	int term_is_ansi = 0;

	/* Find out what kind of terminal this is. */
	if (lgetenv("TERM") == NULL)
		term_is_ansi = 1;

	/* Get size of the screen. */
	scrsize();
	pos_init();

	/* make minimal harcoded vt100-like TERM by dzo@ */

	ignaw = !term_is_ansi;

	/* Get various string-valued capabilities. */
	sc_u_in = term_is_ansi ? "\033[36m" : "\033[4m";
	sc_bl_in = term_is_ansi ? "\033[32m" : "\033[5m";
	sc_visual_bell = term_is_ansi ? "" : "\033[?5h\033[?5l";
	sc_backspace = term_is_ansi ? "\033[D" : "\b";
}

static void multi_search(char *pattern, int n);


#if TAB_COMPLETE_FILENAME
static int cmd_complete(int action);
/* These variables are statics used by cmd_complete. */
static int in_completion;
static char *tk_text;
static char *tk_original;
static char *tk_ipoint;
static char *tk_trial;
static struct textlist tk_tlist;
#endif


static void cmd_repaint(const char *old_cp);


/* Reset command buffer (to empty). */
static void
cmd_reset(void)
{
	cmdbuf[0] = '\0';
	cp = cmdbuf;
	cmd_col = 0;
	cmd_offset = 0;
	literal = 0;
#if SUPPORT_UTF8
	cmd_mbc_buf_len = 0;
#endif
}

/* Clear command line. */
static void
clear_cmd(void)
{
	cmd_col = prompt_col = 0;
#if SUPPORT_UTF8
	cmd_mbc_buf_len = 0;
#endif
}

/* Display a string, usually as a prompt for input into the command buffer. */
static void
cmd_putstr(const char *s)
{
#if SUPPORT_UTF8
	LWCHAR prev_ch = 0;
	LWCHAR ch;
#endif
	const char *endline = s + strlen(s);
	while (*s != '\0') {
		const char *ns = s;
#if SUPPORT_UTF8
		ch = step_char(&ns, +1, endline);
#else
		step_char(&ns, +1, endline);
#endif
		while (s < ns)
			putchr(*s++);
		if (!utf_mode) {
			cmd_col++;
			prompt_col++;
		}
#if SUPPORT_UTF8
		 else if (!is_composing_char(ch) &&
			   !is_combining_char(prev_ch, ch)) {
			int width = is_wide_char(ch) ? 2 : 1;
			cmd_col += width;
			prompt_col += width;
		}
		prev_ch = ch;
#endif
	}
}

/* How many characters are in the command buffer? */
static int
len_cmdbuf(void)
{
	const char *s = cmdbuf;
	const char *endline = s + strlen(s);
	int len = 0;

	while (*s != '\0') {
		step_char(&s, +1, endline);
		len++;
	}
	return len;
}

/* Common part of cmd_step_right() and cmd_step_left(). */
static char *
#if SUPPORT_UTF8
cmd_step_common(char *p, LWCHAR ch, int len, int *pwidth, int *bswidth)
#else
cmd_step_common(LWCHAR ch, int *p_width, int *bswidth)
#endif
{
	char *pr;
	char buf[32];

#if SUPPORT_UTF8
	if (len == 1) {
#endif
		pr = prchar((int) ch, buf);
		if (p_width != NULL || bswidth != NULL) {
			int len = strlen(pr);
			if (p_width != NULL)
				*p_width = len;
			if (bswidth != NULL)
				*bswidth = len;
		}
#if SUPPORT_UTF8
	} else {
		char buf[32];

		pr = prutfchar(ch, buf);
		if (p_width != NULL || bswidth != NULL) {
			if (is_composing_char(ch)) {
				if (p_width != NULL)
					*p_width = 0;
				if (bswidth != NULL)
					*bswidth = 0;
			} else if (is_ubin_char(ch)) {
				int len = strlen(pr);
				if (p_width != NULL)
					*p_width = len;
				if (bswidth != NULL)
					*bswidth = len;
			} else {
				LWCHAR prev_ch = step_char(&p, -1, cmdbuf);
				if (is_combining_char(prev_ch, ch)) {
					if (p_width != NULL)
						*p_width = 0;
					if (bswidth != NULL)
						*bswidth = 0;
				} else {
					if (p_width != NULL)
						*p_width = is_wide_char(ch)
							?       2
							:       1;
					if (bswidth != NULL)
						*bswidth = 1;
				}
			}
		}
	}
#endif
	return pr;
}

/* Step a pointer one character right in the command buffer. */
static char *
cmd_step_right(const char **pp, int *p_width, int *bswidth)
{
	const char *p = *pp;
	LWCHAR ch = step_char(pp, +1, p + strlen(p));

#if SUPPORT_UTF8
	return cmd_step_common(p, ch, *pp - p, p_width, bswidth);
#else
	return cmd_step_common(ch, p_width, bswidth);
#endif
}

/* Step a pointer one character left in the command buffer. */
static char *
cmd_step_left(const char **pp, int *p_width, int *bswidth)
{
#if SUPPORT_UTF8
	const char *p = *pp;
#endif
	LWCHAR ch = step_char(pp, -1, cmdbuf);

#if SUPPORT_UTF8
	return cmd_step_common(*pp, ch, p - *pp, p_width, bswidth);
#else
	return cmd_step_common(ch, p_width, bswidth);
#endif
}

/* Put the cursor at "home" (just after the prompt),
 * and set cp to the corresponding char in cmdbuf. */
static void
cmd_home(void)
{
	while (cmd_col > prompt_col) {
		int width, bswidth;

		cmd_step_left(&cp, &width, &bswidth);
		while (bswidth-- > 0)
			putbs();
		cmd_col -= width;
	}

	cp = &cmdbuf[cmd_offset];
}

/* Shift the cmdbuf display right a half-screen. */
static void
cmd_rshift(void)
{
	const char *s;
	const char *save_cp;
	int cols;

	/* Start at the first displayed char, count how far to the
	 * left we'd have to move to traverse a half-screen width
	 * of displayed characters. */
	s = cmdbuf + cmd_offset;
	cols = 0;
	while (cols < (sc_width - prompt_col) / 2 && s > cmdbuf) {
		int width;
		cmd_step_left(&s, &width, NULL);
		cols += width;
	}

	cmd_offset = s - cmdbuf;
	save_cp = cp;
	cmd_home();
	cmd_repaint(save_cp);
}

/* Move cursor left one character. */
static int
cmd_left(void)
{
	const char *ncp;
	int width, bswidth;

	if (cp <= cmdbuf) {
		/* Already at the beginning of the line */
		return CC_OK;
	}
	ncp = cp;
	while (ncp > cmdbuf) {
		cmd_step_left(&ncp, &width, &bswidth);
		if (width > 0)
			break;
	}
	if (cmd_col < prompt_col + width)
		cmd_rshift();
	cp = ncp;
	cmd_col -= width;
	while (bswidth-- > 0)
		putbs();
	return CC_OK;
}

/* Repaint the line from cp onwards.
 * Then position the cursor just after the char old_cp (a pointer into cmdbuf). */
static void
cmd_repaint(const char *old_cp)
{
	/* Repaint the line from the current position. */
	clear_eol();
	while (*cp != '\0') {
		const char *np = cp;
		int width;
		char *pr = cmd_step_right(&np, &width, NULL);
		if (cmd_col + width >= sc_width)
			break;
		cp = np;
		putstr(pr);
		cmd_col += width;
	}
	while (*cp != '\0') {
		const char *np = cp;
		int width;
		char *pr = cmd_step_right(&np, &width, NULL);
		if (width > 0)
			break;
		cp = np;
		putstr(pr);
	}

	/* Back up the cursor to the correct position. */
	while (cp > old_cp)
		cmd_left();
}


/* Shift the cmdbuf display left a half-screen. */
static void
cmd_lshift(void)
{
	const char *s;
	const char *save_cp;
	int cols;

	/* Start at the first displayed char, count how far to the
	 * right we'd have to move to reach the center of the screen. */
	s = cmdbuf + cmd_offset;
	cols = 0;
	while (cols < (sc_width - prompt_col) / 2 && *s != '\0') {
		int width;
		cmd_step_right(&s, &width, NULL);
		cols += width;
	}
	while (*s != '\0') {
		int width;
		const char *ns = s;
		cmd_step_right(&ns, &width, NULL);
		if (width > 0)
			break;
		s = ns;
	}

	cmd_offset = s - cmdbuf;
	save_cp = cp;
	cmd_home();
	cmd_repaint(save_cp);
}

/* Move cursor right one character. */
static int
cmd_right(void)
{
	const char *pr;
	const char *ncp;
	int width;

	if (*cp == '\0') {
		/* Already at the end of the line. */
		return CC_OK;
	}
	ncp = cp;
	pr = cmd_step_right(&ncp, &width, NULL);
	if (cmd_col + width >= sc_width)
		cmd_lshift();
	else if (cmd_col + width == sc_width - 1 && cp[1] != '\0')
		cmd_lshift();
	cp = ncp;
	cmd_col += width;
	putstr(pr);
	while (*cp != '\0') {
		pr = cmd_step_right(&ncp, &width, NULL);
		if (width > 0)
			break;
		putstr(pr);
		cp = ncp;
	}
	return CC_OK;
}


/* Insert a char into the command buffer, at the current position. */
static int
cmd_ichar(char *cs, int clen)
{
	char *s;

	if (strlen(cmdbuf) + clen >= sizeof(cmdbuf)-1) {
		/* No room in the command buffer for another char. */
		bell();
		return CC_ERROR;
	}

	/* Make room for the new character (shift the tail of the buffer right). */
	for (s = &cmdbuf[strlen(cmdbuf)];  s >= cp;  s--)
		s[clen] = s[0];
	/* Insert the character into the buffer. */
	for (s = (char *)cp;  s < cp + clen;  s++)
		*s = *cs++;
	/* Reprint the tail of the line from the inserted char. */
	cmd_repaint(cp);
	cmd_right();
	return CC_OK;
}

/* Backspace in the command buffer.
 * Delete the char to the left of the cursor. */
static int
cmd_erase(void)
{
	char *s;
	int clen;

	if (cp == cmdbuf) {
		/* Backspace past beginning of the buffer:
		 * this usually means abort the command. */
		return CC_QUIT;
	}
	/* Move cursor left (to the char being erased). */
	s = (char *)cp;
	cmd_left();
	clen = s - cp;

	/* Remove the char from the buffer (shift the buffer left). */
	for (s = (char *)cp;  ;  s++) {
		s[0] = s[clen];
		if (s[0] == '\0')
			break;
	}

	/* Repaint the buffer after the erased char. */
	cmd_repaint(cp);

	/* We say that erasing the entire command string causes us
	 * to abort the current command, if CF_QUIT_ON_ERASE is set. */
	if ((curr_cmdflags & CF_QUIT_ON_ERASE) && cp == cmdbuf && *cp == '\0')
		return CC_QUIT;
	return CC_OK;
}

/* Delete the char under the cursor. */
static int
cmd_delete(void)
{
	if (*cp == '\0') {
		/* At end of string; there is no char under the cursor. */
		return CC_OK;
	}
	/* Move right, then use cmd_erase. */
	cmd_right();
	cmd_erase();
	return CC_OK;
}

/* Delete the "word" to the left of the cursor. */
static int
cmd_werase(void)
{
	if (cp > cmdbuf && cp[-1] == ' ') {
		/* If the char left of cursor is a space,
		 * erase all the spaces left of cursor (to the first non-space). */
		while (cp > cmdbuf && cp[-1] == ' ')
			(void) cmd_erase();
	} else {
		/* If the char left of cursor is not a space,
		 * erase all the nonspaces left of cursor (the whole "word"). */
		while (cp > cmdbuf && cp[-1] != ' ')
			(void) cmd_erase();
	}
	return CC_OK;
}

/* Delete the "word" under the cursor. */
static int
cmd_wdelete(void)
{
	if (*cp == ' ') {
		/* If the char under the cursor is a space,
		 * delete it and all the spaces right of cursor. */
		while (*cp == ' ')
			(void) cmd_delete();
	} else {
		/* If the char under the cursor is not a space,
		 * delete it and all nonspaces right of cursor (the whole word). */
		while (*cp != ' ' && *cp != '\0')
			(void) cmd_delete();
	}
	return CC_OK;
}

/* Delete all chars in the command buffer. */
static int
cmd_kill(void)
{
	if (cmdbuf[0] == '\0') {
		/* Buffer is already empty; abort the current command. */
		return CC_QUIT;
	}
	cmd_offset = 0;
	cmd_home();
	cmdbuf[0] = '\0';
	cmd_repaint(cp);

	/* We say that erasing the entire command string causes us
	 * to abort the current command, if CF_QUIT_ON_ERASE is set. */
	if (curr_cmdflags & CF_QUIT_ON_ERASE)
		return CC_QUIT;
	return CC_OK;
}

/* Select an mlist structure to be the current command history. */
static void
set_mlist(void *mlist, int cmdflags)
{
	curr_mlist = (struct mlist *) mlist;
	curr_cmdflags = cmdflags;

	/* Make sure the next up-arrow moves to the last string in the mlist. */
	if (curr_mlist != NULL)
		curr_mlist->curr_mp = curr_mlist;
}

#if CMD_HISTORY
/* Move up or down in the currently selected command history list. */
static int
cmd_updown(int action)
{
	char *s;

	if (curr_mlist == NULL) {
		/* The current command has no history list. */
		bell();
		return CC_OK;
	}
	cmd_home();
	clear_eol();
	/* Move curr_mp to the next/prev entry. */
	if (action == EC_UP)
		curr_mlist->curr_mp = curr_mlist->curr_mp->prev;
	else
		curr_mlist->curr_mp = curr_mlist->curr_mp->next;
	/* Copy the entry into cmdbuf and echo it on the screen. */
	s = curr_mlist->curr_mp->string;
	if (s == NULL)
		s = "";
	strcpy(cmdbuf, s);
	for (cp = cmdbuf;  *cp != '\0';  )
		cmd_right();
	return CC_OK;
}
#endif

/* Add a string to a history list. */
#if CMD_HISTORY
static void
cmd_addhist(struct mlist *mlist, char *cmd)
{
	struct mlist *ml;

	/* Don't save a trivial command. */
	if (strlen(cmd) == 0)
		return;

	/* Save the command unless it's a duplicate of the
	 * last command in the history. */
	ml = mlist->prev;
	if (ml == mlist || strcmp(ml->string, cmd) != 0) {
		/* Did not find command in history.
		 * Save the command and put it at the end of the history list. */
		ml = emalloc(sizeof(struct mlist));
		ml->string = save(cmd);
		ml->next = mlist;
		ml->prev = mlist->prev;
		mlist->prev->next = ml;
		mlist->prev = ml;
	}
	/* Point to the cmd just after the just-accepted command.
	 * Thus, an UPARROW will always retrieve the previous command. */
	mlist->curr_mp = ml->next;
}
#endif

/* Accept the command in the command buffer.
 * Add it to the currently selected history list. */
static void
cmd_accept(void)
{
#if CMD_HISTORY
	/* Nothing to do if there is no currently selected history list. */
	if (curr_mlist == NULL)
		return;
	cmd_addhist(curr_mlist, cmdbuf);
	curr_mlist->modified = 1;
#endif
}

/* Get command character.
 * The character normally comes from the keyboard,
 * but may come from ungotten characters
 * (characters previously given to ungetcc or ungetsc). */
static int
getcc(void)
{
	if (ungotp == NULL)
		/* Normal case: no ungotten chars, so get one from the user. */
		return getchr();

	if (ungotp > ungot)
		/* Return the next ungotten char. */
		return (*--ungotp);

	/* We have just run out of ungotten chars. */
	ungotp = NULL;
	if (len_cmdbuf() == 0 || !empty_screen())
		return getchr();
	/* Command is incomplete, so try to complete it. */
	switch (mca) {
	case A_DIGIT:
		/* We have a number but no command.  Treat as #g. */
		return 'g';

	case A_F_SEARCH:
	case A_B_SEARCH:
		/* We have "/string" but no newline.  Add the \n. */
		return '\n';

	default:
		/* Some other incomplete command.  Let user complete it. */
		return getchr();
	}
}

/* Decode a command from the edittables list. */
static int
ecmd_decode(const char *cmd, const char **sp)
{
	return cmd_decode(list_ecmd_tables, cmd, sp);
}

/* See if a char is a special line-editing command. */
static int
editchar(int c, int flags)
{
	int action;
	int nch;
	const char *s;
	char usercmd[MAX_CMDLEN+1];

	/* An editing character could actually be a sequence of characters;
	 * for example, an escape sequence sent by pressing the uparrow key.
	 * To match the editing string, we use the command decoder
	 * but give it the edit-commands command table
	 * This table is constructed to match the user's keyboard. */
	if (c == erase_char || c == erase2_char)
		return EC_BACKSPACE;
	if (c == kill_char)
		return EC_LINEKILL;

	/* Collect characters in a buffer.
	 * Start with the one we have, and get more if we need them. */
	nch = 0;
	do {
		if (nch > 0)
			c = getcc();
		usercmd[nch] = c;
		usercmd[nch+1] = '\0';
		nch++;
		action = ecmd_decode(usercmd, &s);
	} while (action == A_PREFIX);

	if (flags & EC_NORIGHTLEFT) {
		switch (action) {
		case EC_RIGHT:
		case EC_LEFT:
			action = A_INVALID;
			break;
		}
	}
#if CMD_HISTORY
	if (flags & EC_NOHISTORY) {
		/* The caller says there is no history list.
		 * Reject any history-manipulation action. */
		switch (action) {
		case EC_UP:
		case EC_DOWN:
			action = A_INVALID;
			break;
		}
	}
#endif
#if TAB_COMPLETE_FILENAME
	if (flags & EC_NOCOMPLETE) {
		/* The caller says we don't want any filename completion cmds.
		 * Reject them. */
		switch (action) {
		case EC_F_COMPLETE:
		case EC_B_COMPLETE:
		case EC_EXPAND:
			action = A_INVALID;
			break;
		}
	}
#endif
	if ((flags & EC_PEEK) || action == A_INVALID) {
		/* We're just peeking, or we didn't understand the command.
		 * Unget all the characters we read in the loop above.
		 * This does NOT include the original character that was
		 * passed in as a parameter. */
		while (nch > 1) {
			ungetcc(usercmd[--nch]);
		}
	} else {
		if (s != NULL)
			ungetsc(s);
	}
	return action;
}

/* Try to perform a line-edit function on the command buffer,
 * using a specified char as a line-editing command.
 * Returns:
 *      CC_PASS The char does not invoke a line edit function.
 *      CC_OK   Line edit function done.
 *      CC_QUIT The char requests the current command to be aborted. */
static int
cmd_edit(int c)
{
	int action;
	int flags;

#if TAB_COMPLETE_FILENAME
#define not_in_completion()     in_completion = 0
#else
#define not_in_completion()
#endif

	/* See if the char is indeed a line-editing command. */
	flags = 0;
#if CMD_HISTORY
	if (curr_mlist == NULL)
		/* No current history; don't accept history manipulation cmds. */
		flags |= EC_NOHISTORY;
#endif
#if TAB_COMPLETE_FILENAME
	if (curr_mlist == ml_search)
		/* In a search command; don't accept file-completion cmds. */
		flags |= EC_NOCOMPLETE;
#endif

	action = editchar(c, flags);

	switch (action) {
	case EC_RIGHT:
		not_in_completion();
		return cmd_right();
	case EC_LEFT:
		not_in_completion();
		return cmd_left();
	case EC_W_RIGHT:
		not_in_completion();
		while (*cp != '\0' && *cp != ' ')
			cmd_right();
		while (*cp == ' ')
			cmd_right();
		return CC_OK;
	case EC_W_LEFT:
		not_in_completion();
		while (cp > cmdbuf && cp[-1] == ' ')
			cmd_left();
		while (cp > cmdbuf && cp[-1] != ' ')
			cmd_left();
		return CC_OK;
	case EC_HOME:
		not_in_completion();
		cmd_offset = 0;
		cmd_home();
		cmd_repaint(cp);
		return CC_OK;
	case EC_END:
		not_in_completion();
		while (*cp != '\0')
			cmd_right();
		return CC_OK;
	case EC_INSERT:
		not_in_completion();
		return CC_OK;
	case EC_BACKSPACE:
		not_in_completion();
		return cmd_erase();
	case EC_LINEKILL:
		not_in_completion();
		return cmd_kill();
	case EC_W_BACKSPACE:
		not_in_completion();
		return cmd_werase();
	case EC_DELETE:
		not_in_completion();
		return cmd_delete();
	case EC_W_DELETE:
		not_in_completion();
		return cmd_wdelete();
	case EC_LITERAL:
		literal = 1;
		return CC_OK;
#if CMD_HISTORY
	case EC_UP:
	case EC_DOWN:
		not_in_completion();
		return cmd_updown(action);
#endif
#if TAB_COMPLETE_FILENAME
	case EC_F_COMPLETE:
	case EC_B_COMPLETE:
	case EC_EXPAND:
		return cmd_complete(action);
#endif
	case EC_NOACTION:
		return CC_OK;
	default:
		not_in_completion();
		return CC_PASS;
	}
}

#if EXAMINE || TAB_COMPLETE_FILENAME
static char *
forw_textlist(struct textlist *tlist, char *prev)
{
	char *s;

	/* prev == NULL means return the first word in the list.
	 * Otherwise, return the word after "prev". */
	if (prev == NULL)
		s = tlist->string;
	else
		s = prev + strlen(prev);
	if (s >= tlist->endstring)
		return NULL;
	while (*s == '\0')
		s++;
	if (s >= tlist->endstring)
		return NULL;
	return s;
}
#endif /* EXAMINE || TAB_COMPLETE_FILENAME */

#if TAB_COMPLETE_FILENAME
/* Insert a string into the command buffer, at the current position. */
static int
cmd_istr(char *str)
{
	char *s;
	int action;
	char *endline = str + strlen(str);

	for (s = str;  *s != '\0';  ) {
		char *os = s;
		step_char(&s, +1, endline);
		action = cmd_ichar(os, s - os);
		if (action != CC_OK) {
			bell();
			return action;
		}
	}
	return CC_OK;
}

/* Find the beginning and end of the "current" word.
 * This is the word which the cursor (cp) is inside or at the end of.
 * Return pointer to the beginning of the word and put the
 * cursor at the end of the word. */
static char *
delimit_word(void)
{
	char *word;
	char *p;
	int delim_quoted = 0;
	int meta_quoted = 0;
	char *esc = get_meta_escape();
	int esclen = strlen(esc);

	/* Move cursor to end of word. */
	if (*cp != ' ' && *cp != '\0') {
		/* Cursor is on a nonspace.
		 * Move cursor right to the next space. */
		while (*cp != ' ' && *cp != '\0')
			cmd_right();
	} else if (cp > cmdbuf && cp[-1] != ' ') {
		/* Cursor is on a space, and char to the left is a nonspace.
		 * We're already at the end of the word. */
		;
	}
	/* Find the beginning of the word which the cursor is in. */
	if (cp == cmdbuf)
		return NULL;
	/* If we have an unbalanced quote (that is, an open quote
	 * without a corresponding close quote), we return everything
	 * from the open quote, including spaces. */
	for (word = cmdbuf;  word < cp;  word++)
		if (*word != ' ')
			break;
	if (word >= cp)
		return cp;
	for (p = cmdbuf;  p < cp;  p++) {
		if (meta_quoted) {
			meta_quoted = 0;
		} else if (esclen > 0 && p + esclen < cp &&
			   strncmp(p, esc, esclen) == 0) {
			meta_quoted = 1;
			p += esclen - 1;
		} else if (delim_quoted) {
			if (*p == closequote)
				delim_quoted = 0;
		} else /* (!delim_quoted) */ {
			if (*p == openquote)
				delim_quoted = 1;
			else if (*p == ' ')
				word = p+1;
		}
	}
	return word;
}


/* Return a blank-separated list of filenames which "complete"
 * the given string. */
static char *
fcomplete(char *s)
{
	char *fpat;
	char *qs;

	if (secure)
		return NULL;
	/* Complete the filename "s" by globbing "s*". */
	fpat = emalloc(strlen(s) + 2);
	sprintf(fpat, "%s*", s);
	qs = lglob(fpat);
	s = shell_unquote(qs);
	if (strcmp(s, fpat) == 0) {
		/* The filename didn't expand. */
		free(qs);
		qs = NULL;
	}
	free(s);
	free(fpat);
	return qs;
}

/* Set things up to enter completion mode.
 * Expand the word under the cursor into a list of filenames
 * which start with that word, and set tk_text to that list. */
static void
init_compl(void)
{
	char *word;
	char c;

	/* Get rid of any previous tk_text. */
	if (tk_text != NULL) {
		free(tk_text);
		tk_text = NULL;
	}
	/* Find the original (uncompleted) word in the command buffer. */
	word = delimit_word();
	if (word == NULL)
		return;
	/* Set the insertion point to the point in the command buffer
	 * where the original (uncompleted) word now sits. */
	tk_ipoint = word;
	/* Save the original (uncompleted) word */
	if (tk_original != NULL)
		free(tk_original);
	tk_original = emalloc(cp-word+1);
	strncpy(tk_original, word, cp-word);
	/* Get the expanded filename.
	 * This may result in a single filename, or
	 * a blank-separated list of filenames. */
	c = *cp;
	*cp = '\0';
	if (*word != openquote) {
		tk_text = fcomplete(word);
	} else {
		char *qword = shell_quote(word+1);
		if (qword == NULL)
			tk_text = fcomplete(word+1);
		else {
			tk_text = fcomplete(qword);
			free(qword);
		}
	}
	*cp = c;
}

static char *
back_textlist(struct textlist *tlist, char *prev)
{
	char *s;

	/* prev == NULL means return the last word in the list.
	 * Otherwise, return the word before "prev". */
	if (prev == NULL)
		s = tlist->endstring;
	else if (prev <= tlist->string)
		return NULL;
	else
		s = prev - 1;
	while (*s == '\0')
		s--;
	if (s <= tlist->string)
		return NULL;
	while (s[-1] != '\0' && s > tlist->string)
		s--;
	return s;
}

/* Return the next word in the current completion list. */
static char *
next_compl(int action, char *prev)
{
	switch (action) {
	case EC_F_COMPLETE:
		return forw_textlist(&tk_tlist, prev);
	case EC_B_COMPLETE:
		return back_textlist(&tk_tlist, prev);
	}
	/* Cannot happen */
	return "?";
}
#endif

/* Is it possible to seek on a file descriptor? */
static int
seekable(int f)
{
	return (lseek(f, (off_t)1, SEEK_SET) != BAD_LSEEK);
}

/* Predefined character sets,
 * selected by the LESSCHARSET environment variable. */
struct charset {
	const char *name;
	const char *desc;
};

static const struct charset charsets[] = {
	{ "ascii",                "8bcccbcc18b95.b" },
	{ "utf-8",                "8bcccbcc18b95.b126.bb" },
#define UTF8_charset 1
	{ "iso8859",              "8bcccbcc18b95.33b." },
	{ "latin3",               "8bcccbcc18b95.33b5.b8.b15.b4.b12.b18.b12.b." },
	{ "arabic",               "8bcccbcc18b95.33b.3b.7b2.13b.3b.b26.5b19.b" },
	{ "greek",                "8bcccbcc18b95.33b4.2b4.b3.b35.b44.b" },
	{ "greek2005",            "8bcccbcc18b95.33b14.b35.b44.b" },
	{ "hebrew",               "8bcccbcc18b95.33b.b29.32b28.2b2.b" },
	{ "koi8-r",               "8bcccbcc18b95.b." },
	{ "KOI8-T",               "8bcccbcc18b95.b8.b6.b8.b.b.5b7.3b4.b4.b3.b.b.3b." },
	{ "georgianps",           "8bcccbcc18b95.3b11.4b12.2b." },
	{ "tcvn",                 "b..b...bcccbccbbb7.8b95.b48.5b." },
	{ "TIS-620",              "8bcccbcc18b95.b.4b.11b7.8b." },
	{ "next",                 "8bcccbcc18b95.bb125.bb" },
	{ "dos",                  "8bcccbcc12bc5b95.b." },
	{ "windows-1251",         "8bcccbcc12bc5b95.b24.b." },
	{ "windows-1252",         "8bcccbcc12bc5b95.b.b11.b.2b12.b." },
	{ "windows-1255",         "8bcccbcc12bc5b95.b.b8.b.5b9.b.4b." },
	{ "ebcdic",               "5bc6bcc7bcc41b.9b7.9b5.b..8b6.10b6.b9.7b9.8b8.17b3.3b9.7b9.8b8.6b10.b.b.b." },
	{ "IBM-1047",             "4cbcbc3b9cbccbccbb4c6bcc5b3cbbc4bc4bccbc191.b" },
	{ NULL, NULL }
};


/* Define a charset, given a description string.
 * The string consists of 256 letters,
 * one for each character in the charset.
 * If the string is shorter than 256 letters, missing letters
 * are taken to be identical to the last one.
 * A decimal number followed by a letter is taken to be a
 * repetition of the letter.
 *
 * Each letter is one of:
 *      . normal character
 *      b binary character
 *      c control character
 */
static void
ichardef(const char *s)
{
	char *chdfp;
	int n;
	char v;

	n = 0;
	v = 0;
	chdfp = chardef;
	while (*s != '\0') {
		switch (*s++) {
		case '.':
			v = 0;
			break;
		case 'c':
			v = IS_CONTROL_CHAR;
			break;
		case 'b':
			v = IS_BINARY_CHAR|IS_CONTROL_CHAR;
			break;

		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			n = (10 * n) + (s[-1] - '0');
			continue;

		default:
			error("invalid chardef", NULL_PARG);
			quit(QUIT_ERROR);
		}

		do {
			if (chdfp >= chardef + sizeof(chardef)) {
				error("chardef longer than 256", NULL_PARG);
				quit(QUIT_ERROR);
			}
			*chdfp++ = v;
		} while (--n > 0);
		n = 0;
	}

	while (chdfp < chardef + sizeof(chardef))
		*chdfp++ = v;
}

/* Define a charset, given a charset name.
 * The valid charset names are listed in the "charsets" array. */
static int
icharset(const char *name, int no_error)
{
	const struct charset *p;
	const struct cs_alias *a;

	if (name == NULL || *name == '\0')
		return 0;

	/* First see if the name is an alias. */
	a = find_string_from_struct(cs_aliases, sizeof(struct cs_alias), name);
	if(a->name != NULL)
		name = a->oname;

	p = find_string_from_struct(charsets, sizeof(struct charset), name);
	if(p->name != NULL) {
		ichardef(p->desc);
#if SUPPORT_UTF8
		if (p == &charsets[UTF8_charset])
			utf_mode = 1;
#endif
		return 1;
	}

	if (!no_error) {
		error("invalid charset name", NULL_PARG);
		quit(QUIT_ERROR);
	}
	return 0;
}

#if HAVE_LOCALE
/* Define a charset, given a locale name. */
static void
ilocale(void)
{
	int c;

	for (c = 0;  c < (int) sizeof(chardef);  c++) {
		if (isprint(c))
			chardef[c] = 0;
		else if (iscntrl(c))
			chardef[c] = IS_CONTROL_CHAR;
		else
			chardef[c] = IS_BINARY_CHAR|IS_CONTROL_CHAR;
	}
}
#endif

/* Is a given character a "binary" character? */
static inline int
binary_char(unsigned char c)
{
	return (chardef[c] & IS_BINARY_CHAR);
}

/* Try to determine if a file is "binary".
 * This is just a guess, and we need not try too hard to make it accurate. */
static int
bin_file(int f)
{
	int i;
	int n;
	int bin_count = 0;
	unsigned char data[256];

	if (!seekable(f))
		return 0;
	if (lseek(f, (off_t)0, SEEK_SET) == BAD_LSEEK)
		return 0;
	n = read(f, data, sizeof(data));
	for (i = 0;  i < n;  i++) {
		char c = data[i];
		if (ctldisp == OPT_ONPLUS && IS_CSI_START(c)) {
			while (++i < n && is_ansi_middle(data[i]))
				continue;
		} else if (binary_char(c))
			bin_count++;
	}
	/* Call it a binary file if there are more than 5 binary characters
	 * in the first 256 bytes of the file. */
	return (bin_count > 5);
}

/* See if we should open a "replacement file"
 * instead of the file we're about to open. */
static char *
open_altfile(const char *filename, int *pf, void **pfd)
{
	const char *lessopen;
	char *cmd;
	int len;
	FILE *fd;
	int returnfd = 0;

	if (!use_lessopen || secure)
		return NULL;
	ch_ungetchar(-1);
	if ((lessopen = lgetenv("LESSOPEN")) == NULL)
		return NULL;
	if (strcmp(filename, "-") == 0)
		return NULL;
	if (*lessopen == '|') {
		/* If LESSOPEN starts with a |, it indicates
		 * a "pipe preprocessor". */
		lessopen++;
		returnfd = 1;
	}

	len = strlen(lessopen) + strlen(filename) + 2;
	cmd = emalloc(len);
	snprintf(cmd, len, lessopen, filename);
	fd = shell_cmd(cmd);
	free(cmd);
	if (fd == NULL) {
		/* Cannot create the pipe. */
		return NULL;
	}
	if (returnfd) {
		int f;
		char c;

		/* Read one char to see if the pipe will produce any data.
		 * If it does, push the char back on the pipe. */
		f = fileno(fd);
		if (read(f, &c, 1) != 1) {
			/* Pipe is empty.  This means there is no alt file. */
			pclose(fd);
			return NULL;
		}
		ch_ungetchar(c);
		*pfd = (void *) fd;
		*pf = f;
		return save("-");
	}
	cmd = readfd(fd);
	pclose(fd);
	if (*cmd == '\0')
		/* Pipe is empty.  This means there is no alt file. */
		return NULL;
	return cmd;
}

/* Close a replacement file. */
static void
close_altfile(const char *altfilename, char *filename, char *pipefd)
{
	const char *lessclose;
	FILE *fd;
	char *cmd;
	int len;

	if (secure)
		return;
	if (pipefd != NULL) {
		pclose((FILE*) pipefd);
	}
	if ((lessclose = lgetenv("LESSCLOSE")) == NULL)
		return;
	len = strlen(lessclose) + strlen(filename) + strlen(altfilename) + 2;
	cmd = emalloc(len);
	snprintf(cmd, len, lessclose, filename, altfilename);
	fd = shell_cmd(cmd);
	free(cmd);
	if (fd != NULL)
		pclose(fd);
}

/* Is the specified file a directory? */
static int is_dir(char *filename)
{
	int isdir;

	filename = shell_unquote(filename);
	isdir = is_directory(filename, 1, NULL) > 0;
	free(filename);
	return isdir;
}

/* Returns NULL if the file can be opened and
 * is an ordinary file, otherwise an error message
 * (if it cannot be opened or is a directory, etc.) */
static char *bad_file(char *filename)
{
	char *m = NULL;

	filename = shell_unquote(filename);
	if (!force_open && is_dir(filename)) {
		static const char is_a_dir[] = " is a directory";

		m = emalloc(strlen(filename) + sizeof(is_a_dir));
		strcpy(m, filename);
		strcat(m, is_a_dir);
	} else {
		int r;
		struct stat statbuf;

		r = stat(filename, &statbuf);
		if (r < 0) {
			m = errno_message(filename);
		} else if (force_open) {
			m = NULL;
		} else if (!S_ISREG(statbuf.st_mode)) {
			static const char not_reg[] = " is not a regular file (use -f to see it)";
			m = emalloc(strlen(filename) + sizeof(not_reg));
			strcpy(m, filename);
			strcat(m, not_reg);
		}
	}
	free(filename);
	return m;
}

/* Unlink an ifile from the ifile list. */
static void
unlink_ifile(struct ifile *p)
{
	p->h_next->h_prev = p->h_prev;
	p->h_prev->h_next = p->h_next;
	incr_index(p->h_next, -1);
	ifiles--;
}

/* Return a different ifile from the given one. */
static IFILE
getoff_ifile(IFILE ifile)
{
	IFILE newifile;

	if ((newifile = prev_ifile(ifile)) != NULL_IFILE)
		return newifile;
	if ((newifile = next_ifile(ifile)) != NULL_IFILE)
		return newifile;
	return NULL_IFILE;
}

/* Delete an existing ifile structure. */
static void
del_ifile(IFILE h)
{
	struct ifile *p;

	if (h == NULL_IFILE)
		return;
	/* If the ifile we're deleting is the currently open ifile,
	 * move off it. */
	unmark(h);
	if (h == curr_ifile)
		curr_ifile = getoff_ifile(curr_ifile);
	p = int_ifile(h);
	unlink_ifile(p);
	free(p->h_filename);
	free(p);
}

/* Save the file position to be associated with a given file. */
static void
store_pos(IFILE ifile, struct scrpos *scrpos)
{
	int_ifile(ifile)->h_scrpos = *scrpos;
}

/* Recall the file position associated with a file.
 * If no position has been associated with the file, return NULL_POSITION. */
static void
get_pos(IFILE ifile, struct scrpos *scrpos)
{
	*scrpos = int_ifile(ifile)->h_scrpos;
}

/* Mark the ifile as "opened". */
static void
set_open(IFILE ifile)
{
	int_ifile(ifile)->h_opened = 1;
}

/* Return whether the ifile has been opened previously. */
static int
opened(IFILE ifile)
{
	return int_ifile(ifile)->h_opened;
}

static int
held_ifile(IFILE ifile)
{
	return int_ifile(ifile)->h_hold;
}

static void *
get_filestate(IFILE ifile)
{
	return int_ifile(ifile)->h_filestate;
}

static void
set_filestate(IFILE ifile, void *filestate)
{
	int_ifile(ifile)->h_filestate = filestate;
}

/* Textlist functions deal with a list of words separated by spaces.
 * init_textlist sets up a textlist structure.
 * forw_textlist uses that structure to iterate thru the list of
 * words, returning each one as a standard null-terminated string.
 * back_textlist does the same, but runs thru the list backwards. */
#if EXAMINE || TAB_COMPLETE_FILENAME
static void
init_textlist(struct textlist *tlist, char *str)
{
	char *s;
	int meta_quoted = 0;
	int delim_quoted = 0;
	char *esc = get_meta_escape();
	int esclen = strlen(esc);

	tlist->string = skipsp(str);
	tlist->endstring = tlist->string + strlen(tlist->string);
	for (s = str;  s < tlist->endstring;  s++) {
		if (meta_quoted) {
			meta_quoted = 0;
		} else if (esclen > 0 && s + esclen < tlist->endstring &&
			   strncmp(s, esc, esclen) == 0) {
			meta_quoted = 1;
			s += esclen - 1;
		} else if (delim_quoted) {
			if (*s == closequote)
				delim_quoted = 0;
		} else /* (!delim_quoted) */ {
			if (*s == openquote)
				delim_quoted = 1;
			else if (*s == ' ')
				*s = '\0';
		}
	}
}
#endif /* EXAMINE ||  TAB_COMPLETE_FILENAME */

static void
init_hashtbl(void)
{
	int h;

	for (h = 0;  h < BUFHASH_SIZE;  h++) {
		thisfile->hashtbl[h].buf_hnext = END_OF_HCHAIN(h);
		thisfile->hashtbl[h].buf_hprev = END_OF_HCHAIN(h);
	}
}

/* Delete all buffers for this file. */
static void
ch_delbufs(void)
{
	struct buf *bp;

	while (ch_bufhead != END_OF_CHAIN) {
		bp = ch_bufhead;
		bp->next->prev = bp->prev;
		bp->prev->next = bp->next;
		free(bp);
	}
	ch_nbufs = 0;
	init_hashtbl();
}

/* Close a filestate. */
static void
ch_close(void)
{
	int keepstate = FALSE;

	if (thisfile == NULL)
		return;

	if (ch_flags & (CH_CANSEEK|CH_POPENED|CH_HELPFILE)) {
		/* We can seek or re-open, so we don't need to keep buffers. */
		ch_delbufs();
	} else
		keepstate = TRUE;
	if (!(ch_flags & CH_KEEPOPEN)) {
		/* We don't need to keep the file descriptor open
		 * (because we can re-open it.)
		 * But don't really close it if it was opened via popen(),
		 * because pclose() wants to close it. */
		if (!(ch_flags & (CH_POPENED|CH_HELPFILE)))
			close(ch_file);
		ch_file = -1;
	} else
		keepstate = TRUE;
	if (!keepstate) {
		/* We don't even need to keep the filestate structure. */
		free(thisfile);
		thisfile = NULL;
		set_filestate(curr_ifile, (void *) NULL);
	}
}

/* Close the current input file. */
static void
close_file(void)
{
	struct scrpos scrpos;

	if (curr_ifile == NULL_IFILE)
		return;

	/* Save the current position so that we can return to
	 * the same position if we edit this file again. */
	get_scrpos(&scrpos);
	if (scrpos.pos != NULL_POSITION) {
		store_pos(curr_ifile, &scrpos);
		lastmark();
	}
	/* Close the file descriptor, unless it is a pipe. */
	ch_close();
	/* If we opened a file using an alternate name,
	 * do special stuff to close it. */
	if (curr_altfilename != NULL) {
		close_altfile(curr_altfilename, get_filename(curr_ifile),
				curr_altpipe);
		free(curr_altfilename);
		curr_altfilename = NULL;
	}
	curr_ifile = NULL_IFILE;
	curr_ino = curr_dev = 0;
}


#if LOGFILE
/* Close the logfile.
 * If we haven't read all of standard input into it, do that now. */
static void
end_logfile(void)
{
	static int tried = FALSE;

	if (logfile < 0)
		return;
	if (!tried && ch_fsize == NULL_POSITION) {
		tried = TRUE;
		ierror("Finishing logfile", NULL_PARG);
		while (ch_forw_get() != EOI)
			if (ABORT_SIGS())
				break;
	}
	close(logfile);
	logfile = -1;
	namelogfile = NULL;
}

/* Start a log file AFTER less has already been running.
 * Invoked from the - command; see toggle_option().
 * Write all the existing buffered data to the log file. */
static void
sync_logfile(void)
{
	struct buf *bp;
	int warned = FALSE;
	BLOCKNUM block;
	BLOCKNUM nblocks;

	nblocks = (ch_fpos + LBUFSIZE - 1) / LBUFSIZE;
	for (block = 0;  block < nblocks;  block++) {
		for (bp = ch_bufhead;  ;  bp = bp->next) {
			if (bp == END_OF_CHAIN) {
				if (!warned) {
					error("Warning: log file is incomplete",
						NULL_PARG);
					warned = TRUE;
				}
				break;
			}
			if (bp->block == block) {
				write(logfile, (char *) bp->data, bp->datasize);
				break;
			}
		}
	}
}
#endif

/* Flush (discard) any saved file state, including buffer contents. */
static void
ch_flush(void)
{
	struct buf *bp;

	if (thisfile == NULL)
		return;

	if (!(ch_flags & CH_CANSEEK)) {
		/* If input is a pipe, we don't flush buffer contents,
		 * since the contents can't be recovered. */
		ch_fsize = NULL_POSITION;
		return;
	}

	/* Initialize all the buffers. */
	for (bp = ch_bufhead;  bp != END_OF_CHAIN;  bp = bp->next)
		bp->block = -1;

	/* Figure out the size of the file, if we can. */
	ch_fsize = filesize(ch_file);

	/* Seek to a known position: the beginning of the file. */
	ch_fpos = 0;
	ch_block = 0; /* ch_fpos / LBUFSIZE; */
	ch_offset = 0; /* ch_fpos % LBUFSIZE; */

	/* This is a kludge to workaround a Linux kernel bug: files in
	 * /proc have a size of 0 according to fstat() but have readable
	 * data.  They are sometimes, but not always, seekable.
	 * Force them to be non-seekable here. */
	if (ch_fsize == 0) {
		ch_fsize = NULL_POSITION;
		ch_flags &= ~CH_CANSEEK;
	}

	if (lseek(ch_file, (off_t)0, SEEK_SET) == BAD_LSEEK) {
		/* Warning only; even if the seek fails for some reason,
		 * there's a good chance we're at the beginning anyway.
		 * {{ I think this is bogus reasoning. }} */
		error("seek error to 0", NULL_PARG);
	}
}

/* Initialize file state for a new file. */
static void
ch_init(int f, int flags)
{
	/* See if we already have a filestate for this file. */
	thisfile = (struct filestate *) get_filestate(curr_ifile);
	if (thisfile == NULL) {
		/* Allocate and initialize a new filestate. */
		thisfile = (struct filestate *)
				calloc(1, sizeof(struct filestate));
		thisfile->buf_next = thisfile->buf_prev = END_OF_CHAIN;
		thisfile->nbufs = 0;
		thisfile->flags = 0;
		thisfile->fpos = 0;
		thisfile->block = 0;
		thisfile->offset = 0;
		thisfile->file = -1;
		thisfile->fsize = NULL_POSITION;
		ch_flags = flags;
		init_hashtbl();
		/* Try to seek; set CH_CANSEEK if it works. */
		if ((flags & CH_CANSEEK) && !seekable(f))
			ch_flags &= ~CH_CANSEEK;
		set_filestate(curr_ifile, (void *) thisfile);
	}
	if (thisfile->file == -1)
		thisfile->file = f;
	ch_flush();
}

/* Edit a new file (given its IFILE).
 * ifile == NULL means just close the current file. */
static int
edit_ifile(IFILE ifile)
{
	int f = -1;
	int answer;
	int no_display;
	int chflags;
	char *filename;
	char *open_filename;
	char *qopen_filename;
	char *alt_filename;
	char *bad_filename;
	void *alt_pipe;
	IFILE was_curr_ifile;
	PARG parg;

	if (ifile == curr_ifile) {
		/* Already have the correct file open. */
		return 0;
	}

	/* We must close the currently open file now.
	 * This is necessary to make the open_altfile/close_altfile pairs
	 * nest properly (or rather to avoid nesting at all).
	 * {{ Some stupid implementations of popen() mess up if you do:
	 *    fA = popen("A"); fB = popen("B"); pclose(fA); pclose(fB); }} */
#if LOGFILE
	end_logfile();
#endif
	was_curr_ifile = save_curr_ifile();
	if (curr_ifile != NULL_IFILE) {
		chflags = ch_getflags();
		close_file();
		if ((chflags & CH_HELPFILE) && held_ifile(was_curr_ifile) <= 1) {
			/* Don't keep the help file in the ifile list. */
			del_ifile(was_curr_ifile);
			was_curr_ifile = old_ifile;
		}
	}

	if (ifile == NULL_IFILE) {
		/* No new file to open.
		 * (Don't set old_ifile, because if you call edit_ifile(NULL),
		 *  you're supposed to have saved curr_ifile yourself,
		 *  and you'll restore it if necessary.) */
		unsave_ifile(was_curr_ifile);
		return 0;
	}

	filename = save(get_filename(ifile));
	/* See if LESSOPEN specifies an "alternate" file to open. */
	alt_pipe = NULL;
	alt_filename = open_altfile(filename, &f, &alt_pipe);
	open_filename = (alt_filename != NULL) ? alt_filename : filename;
	qopen_filename = shell_unquote(open_filename);

	chflags = 0;
	if (alt_pipe != NULL) {
		/* The alternate "file" is actually a pipe.
		 * f has already been set to the file descriptor of the pipe
		 * in the call to open_altfile above.
		 * Keep the file descriptor open because it was opened
		 * via popen(), and pclose() wants to close it. */
		chflags |= CH_POPENED;
	} else if (strcmp(open_filename, "-") == 0) {
		/* Use standard input.
		 * Keep the file descriptor open because we can't reopen it. */
		f = fd0;
		chflags |= CH_KEEPOPEN;
	} else if (strcmp(open_filename, FAKE_HELPFILE) == 0) {
		f = -1;
		chflags |= CH_HELPFILE;
	} else if ((bad_filename = bad_file(open_filename)) != NULL) {
		/* It looks like a bad file.  Don't try to open it. */
		parg.p_string = bad_filename;
		error("%s", &parg);
		free(bad_filename);
	    err1:
		if (alt_filename != NULL) {
			close_altfile(alt_filename, filename, alt_pipe);
			free(alt_filename);
		}
		del_ifile(ifile);
		free(qopen_filename);
		free(filename);
		/* Re-open the current file. */
		if (was_curr_ifile == ifile) {
			/* Whoops.  The "current" ifile is the one we just deleted.
			 * Just give up. */
			quit(QUIT_ERROR);
		}
		reedit_ifile(was_curr_ifile);
		return 1;
	} else if ((f = open(qopen_filename, O_RDONLY)) < 0) {
		/* Got an error trying to open it. */
		char *e = errno_message(filename);
		parg.p_string = e;
		error("%s", &parg);
		free(e);
		goto err1;
	} else {
		chflags |= CH_CANSEEK;
		if (!force_open && !opened(ifile) && bin_file(f)) {
			/* Looks like a binary file.
			 * Ask user if we should proceed. */
			parg.p_string = filename;
			answer = query("\"%s\" may be a binary file.  See it anyway? ",
				&parg);
			if (answer != 'y' && answer != 'Y') {
				close(f);
				goto err1;
			}
		}
	}

	/* Get the new ifile.
	 * Get the saved position for the file. */
	if (was_curr_ifile != NULL_IFILE) {
		old_ifile = was_curr_ifile;
		unsave_ifile(was_curr_ifile);
	}
	curr_ifile = ifile;
	curr_altfilename = alt_filename;
	curr_altpipe = alt_pipe;
	set_open(curr_ifile); /* File has been opened */
	get_pos(curr_ifile, &initial_scrpos);
	new_file = TRUE;
	ch_init(f, chflags);

	if (!(chflags & CH_HELPFILE)) {
#if LOGFILE
		if (namelogfile != NULL && is_tty)
			use_logfile(namelogfile);
#endif
		/* Remember the i-number and device of the opened file. */
		{
			struct stat statbuf;
			int r = stat(qopen_filename, &statbuf);
			if (r == 0) {
				curr_ino = statbuf.st_ino;
				curr_dev = statbuf.st_dev;
			}
		}
		if (every_first_cmd != NULL)
			ungetsc(every_first_cmd);
	}

	free(qopen_filename);
	no_display = !any_display;
	flush();
	any_display = TRUE;

	if (is_tty) {
		/* Output is to a real tty. */

		/* Indicate there is nothing displayed yet. */
		pos_clear();
		clr_linenum();
#if HILITE_SEARCH
		clr_hilite();
#endif
#if CMD_HISTORY
		cmd_addhist(ml_examine, filename);
#endif
		if (no_display && errmsgs > 0) {
			/* We displayed some messages on error output
			 * (file descriptor 2; see error() function).
			 * Before erasing the screen contents,
			 * display the file name and wait for a keystroke. */
			parg.p_string = filename;
			error("%s", &parg);
		}
	}
	free(filename);
	return 0;
}

#if EXAMINE
/* Edit a space-separated list of files.
 * For each filename in the list, enter it into the ifile list.
 * Then edit the first one. */
static int
edit_list(char *filelist)
{
	IFILE save_ifile;
	char *good_filename;
	char *filename;
	char *gfilelist;
	char *gfilename;
	struct textlist tl_files;
	struct textlist tl_gfiles;

	save_ifile = save_curr_ifile();
	good_filename = NULL;

	/* Run thru each filename in the list.
	 * Try to glob the filename.
	 * If it doesn't expand, just try to open the filename.
	 * If it does expand, try to open each name in that list. */
	init_textlist(&tl_files, filelist);
	filename = NULL;
	while ((filename = forw_textlist(&tl_files, filename)) != NULL) {
		gfilelist = lglob(filename);
		init_textlist(&tl_gfiles, gfilelist);
		gfilename = NULL;
		while ((gfilename = forw_textlist(&tl_gfiles, gfilename)) != NULL) {
			if (edit(gfilename) == 0 && good_filename == NULL)
				good_filename = get_filename(curr_ifile);
		}
		free(gfilelist);
	}
	/* Edit the first valid filename in the list. */
	if (good_filename == NULL) {
		unsave_ifile(save_ifile);
		return 1;
	}
	if (get_ifile(good_filename, curr_ifile) == curr_ifile) {
		/* Trying to edit the current file; don't reopen it. */
		unsave_ifile(save_ifile);
		return 0;
	}
	reedit_ifile(save_ifile);
	return edit(good_filename);
}
#endif

/* Edit the first file in the command line (ifile) list. */
static int
edit_first(void)
{
	curr_ifile = NULL_IFILE;
	return edit_next(1);
}

static int
edit_prev(int n)
{
	return edit_istep(curr_ifile, n, -1);
}

/* Edit the last file in the command line (ifile) list. */
static int
edit_last(void)
{
	curr_ifile = NULL_IFILE;
	return edit_prev(1);
}

/* Edit a specific file in the command line (ifile) list. */
static int
edit_index(int n)
{
	IFILE h;

	h = NULL_IFILE;
	do {
		if ((h = next_ifile(h)) == NULL_IFILE) {
			/* Reached end of the list without finding it. */
			return 1;
		}
	} while (get_index(h) != n);

	return edit_ifile(h);
}

static void
reopen_curr_ifile(void)
{
	IFILE save_ifile = save_curr_ifile();
	close_file();
	reedit_ifile(save_ifile);
}

/* Edit standard input. */
static int
edit_stdin(void)
{
	if (isatty(fd0)) {
		error("Missing filename (\"less --help\" for help)", NULL_PARG);
		quit(QUIT_OK);
	}
	return edit("-");
}

/* Copy a file directly to standard output.
 * Used if standard output is not a tty. */
static void
cat_file(void)
{
	int c;

	while ((c = ch_forw_get()) != EOI)
		putchr(c);
	flush();
}

#if TAB_COMPLETE_FILENAME
/* Complete the filename before (or under) the cursor.
 * cmd_complete may be called multiple times.  The global in_completion
 * remembers whether this call is the first time (create the list),
 * or a subsequent time (step thru the list). */
static int
cmd_complete(int action)
{
	char *s;

	if (!in_completion || action == EC_EXPAND) {
		/* Expand the word under the cursor and
		 * use the first word in the expansion
		 * (or the entire expansion if we're doing EC_EXPAND). */
		init_compl();
		if (tk_text == NULL) {
			bell();
			return CC_OK;
		}
		if (action == EC_EXPAND) {
			/* Use the whole list. */
			tk_trial = tk_text;
		} else {
			/* Use the first filename in the list. */
			in_completion = 1;
			init_textlist(&tk_tlist, tk_text);
			tk_trial = next_compl(action, (char*)NULL);
		}
	} else {
		/* We already have a completion list.
		 * Use the next/previous filename from the list. */
		tk_trial = next_compl(action, tk_trial);
	}

	/* Remove the original word, or the previous trial completion. */
	while (cp > tk_ipoint)
		(void) cmd_erase();

	if (tk_trial == NULL) {
		/* There are no more trial completions.
		 * Insert the original (uncompleted) filename. */
		in_completion = 0;
		if (cmd_istr(tk_original) != CC_OK)
			goto fail;
	} else {
		/* Insert trial completion. */
		if (cmd_istr(tk_trial) != CC_OK)
			goto fail;
		/* If it is a directory, append a slash. */
		if (is_dir(tk_trial)) {
			if (cp > cmdbuf && cp[-1] == closequote)
				(void) cmd_erase();
			s = lgetenv("LESSSEPARATOR");
			if (s == NULL)
				s = "/";
			if (cmd_istr(s) != CC_OK)
				goto fail;
		}
	}

	return CC_OK;

fail:
	in_completion = 0;
	bell();
	return CC_OK;
}
#endif /* TAB_COMPLETE_FILENAME */

static int
in_mca(void)
{
	return (mca != 0 && mca != A_PREFIX);
}

/* Process a single character of a multi-character command, such as
 * a number, or the pattern of a search command.
 * Returns:
 *      CC_OK           The char was accepted.
 *      CC_QUIT         The char requests the command to be aborted.
 *      CC_ERROR        The char could not be accepted due to an error. */
static int
cmd_char(int c)
{
	int action;
	int len;

	if (!utf_mode) {
		cmd_mbc_buf[0] = c;
		len = 1;
	}
#if SUPPORT_UTF8
	else {
		static int cmd_mbc_buf_index;
		/* Perform strict validation in all possible cases. */
		if (cmd_mbc_buf_len == 0) {
		 retry:
			cmd_mbc_buf_index = 1;
			*cmd_mbc_buf = c;
			if (IS_ASCII_OCTET(c))
				cmd_mbc_buf_len = 1;
			else if (IS_UTF8_LEAD(c)) {
				cmd_mbc_buf_len = utf_len(c);
				return CC_OK;
			} else {
				/* UTF8_INVALID or stray UTF8_TRAIL */
				bell();
				return CC_ERROR;
			}
		} else if (IS_UTF8_TRAIL(c)) {
			cmd_mbc_buf[cmd_mbc_buf_index++] = c;
			if (cmd_mbc_buf_index < cmd_mbc_buf_len)
				return CC_OK;
			if (!is_utf8_well_formed(cmd_mbc_buf)) {
				/* complete, but not well formed (non-shortest form), sequence */
				cmd_mbc_buf_len = 0;
				bell();
				return CC_ERROR;
			}
		} else {
			/* Flush incomplete (truncated) sequence. */
			cmd_mbc_buf_len = 0;
			bell();
			/* Handle new char. */
			goto retry;
		}

		len = cmd_mbc_buf_len;
		cmd_mbc_buf_len = 0;
	}
#endif
	if (literal) {
		/* Insert the char, even if it is a line-editing char. */
		literal = 0;
		return cmd_ichar(cmd_mbc_buf, len);
	}

	/* See if it is a line-editing character. */
	if (in_mca() && len == 1) {
		action = cmd_edit(c);
		switch (action) {
		case CC_OK:
		case CC_QUIT:
			return action;
		case CC_PASS:
			break;
		}
	}

	/* Insert the char into the command buffer. */
	return cmd_ichar(cmd_mbc_buf, len);
}

/* Return the number currently in the command buffer. */
static LINENUM
cmd_int(long *frac)
{
	const char *p;
	LINENUM n = 0;
	int err;

	for (p = cmdbuf;  *p >= '0' && *p <= '9';  p++)
		n = (n * 10) + (*p - '0');
	*frac = 0;
	if (*p++ == '.') {
		*frac = getfraction(&p, NULL, &err);
		/* {{ do something if err is set? }} */
	}
	return n;
}

/* Return a pointer to the command buffer. */
static inline char *
get_cmdbuf(void)
{
	return cmdbuf;
}

#if CMD_HISTORY
/* Return the last (most recent) string in the current command history. */
static char *
cmd_lastpattern(void)
{
	if (curr_mlist == NULL)
		return NULL;
	return curr_mlist->curr_mp->prev->string;
}

/* Get the name of the history file. */
static char *
histfile_name(void)
{
	const char *home_env;
	const char *name;
	char *o;
	int len;

	/* See if filename is explicitly specified by $LESSHISTFILE. */
	name = lgetenv("LESSHISTFILE");
	if (name != NULL && *name != '\0') {
		if (strcmp(name, "-") == 0 || strcmp(name, bb_dev_null) == 0)
			/* $LESSHISTFILE == "-" means don't use a history file. */
			return NULL;
		return save(name);
	}

	/* Otherwise, file is in $HOME. */
	home_env = lgetenv("HOME");
	if (home_env == NULL || *home_env == '\0')
		return NULL;

	len = strlen(home_env) + strlen(LESSHISTFILE) + 2;
	o = emalloc(len);
	snprintf(o, len, "%s/%s", home_env, LESSHISTFILE);
	return o;
}
#endif /* CMD_HISTORY */

/* Initialize history from a .lesshist file. */
static void
init_cmdhist(void)
{
#if CMD_HISTORY
	struct mlist *ml = NULL;
	char line[CMDBUF_SIZE];
	char *filename;
	FILE *f;
	char *p;

	filename = histfile_name();
	if (filename == NULL)
		return;
	f = fopen(filename, "r");
	free(filename);
	if (f == NULL)
		return;
	if (fgets(line, sizeof(line), f) == NULL ||
	    strncmp(line, HISTFILE_FIRST_LINE, strlen(HISTFILE_FIRST_LINE)) != 0) {
		fclose(f);
		return;
	}
	while (fgets(line, sizeof(line), f) != NULL) {
		for (p = line;  *p != '\0';  p++) {
			if (*p == '\n' || *p == '\r') {
				*p = '\0';
				break;
			}
		}
		if (strcmp(line, HISTFILE_SEARCH_SECTION) == 0)
			ml = &mlist_search;
		else if (strcmp(line, HISTFILE_SHELL_SECTION) == 0) {
#if SHELL_ESCAPE || PIPEC
			ml = &mlist_shell;
#else
			ml = NULL;
#endif
		} else if (*line == '"') {
			if (ml != NULL)
				cmd_addhist(ml, line+1);
		}
	}
	fclose(f);
#endif /* CMD_HISTORY */
}

#if CMD_HISTORY
static void
save_mlist(struct mlist *ml, FILE *f)
{
	int histsize = 0;
	int n;
	const char *s;

	s = lgetenv("LESSHISTSIZE");
	if (s != NULL)
		histsize = atoi(s);
	if (histsize == 0)
		histsize = 100;

	ml = ml->prev;
	for (n = 0;  n < histsize;  n++) {
		if (ml->string == NULL)
			break;
		ml = ml->prev;
	}
	for (ml = ml->next;  ml->string != NULL;  ml = ml->next)
		fprintf(f, "\"%s\n", ml->string);
}
#endif /* CMD_HISTORY */

static void
save_cmdhist(void)
{
#if CMD_HISTORY
	char *filename;
	FILE *f;
	int modified = 0;

	filename = histfile_name();
	if (filename == NULL)
		return;
	if (mlist_search.modified)
		modified = 1;
#if SHELL_ESCAPE || PIPEC
	if (mlist_shell.modified)
		modified = 1;
#endif
	if (!modified)
		return;
	f = fopen(filename, "w");
	free(filename);
	if (f == NULL)
		return;
	/* Make history file readable only by owner. */
	fchmod(fileno(f), 0600);

	fprintf(f, "%s\n", HISTFILE_FIRST_LINE);

	fprintf(f, "%s\n", HISTFILE_SEARCH_SECTION);
	save_mlist(&mlist_search, f);

#if SHELL_ESCAPE || PIPEC
	fprintf(f, "%s\n", HISTFILE_SHELL_SECTION);
	save_mlist(&mlist_shell, f);
#endif

	fclose(f);
#endif /* CMD_HISTORY */
}

/* Clear the attn hilite. */
static void
clear_attn(void)
{
	int slinenum;
	POSITION old_start_attnpos;
	POSITION old_end_attnpos;
	POSITION pos;
	POSITION epos;
	int moved = 0;

	if (start_attnpos == NULL_POSITION)
		return;
	old_start_attnpos = start_attnpos;
	old_end_attnpos = end_attnpos;
	start_attnpos = end_attnpos = NULL_POSITION;

	if (squished)
		repaint();

	for (slinenum = TOP;  slinenum < TOP + sc_height-1;  slinenum++) {
		pos = position(slinenum);
		if (pos == NULL_POSITION)
			continue;
		epos = position(slinenum+1);
		if (pos < old_end_attnpos &&
		     (epos == NULL_POSITION || epos > old_start_attnpos)) {
			(void) forw_line(pos);
			goto_line(slinenum);
			put_line();
			moved = 1;
		}
	}
	if (moved)
		lower_left();
}

/* Move the cursor to start of prompt line before executing a command.
 * This looks nicer if the command takes a long time before
 * updating the screen. */
static void
cmd_exec(void)
{
	clear_attn();
	clear_bot();
	flush();
}

/* Set up the display to start a new multi-character command. */
static void
start_mca(int action, const char *prompt, void *mlist, int cmdflags)
{
	mca = action;
	clear_bot();
	clear_cmd();
	cmd_putstr(prompt);
	set_mlist(mlist, cmdflags);
}


/* Set up the display to start a new search command. */
static void
mca_search(void)
{
	if (search_type & SRCH_FORW)
		mca = A_F_SEARCH;
	else
		mca = A_B_SEARCH;

	clear_bot();
	clear_cmd();

	if (search_type & SRCH_NO_MATCH)
		cmd_putstr("Non-match ");
	if (search_type & SRCH_FIRST_FILE)
		cmd_putstr("First-file ");
	if (search_type & SRCH_PAST_EOF)
		cmd_putstr("EOF-ignore ");
	if (search_type & SRCH_NO_MOVE)
		cmd_putstr("Keep-pos ");
	if (search_type & SRCH_NO_REGEX)
		cmd_putstr("Regex-off ");

	if (search_type & SRCH_FORW)
		cmd_putstr("/");
	else
		cmd_putstr("?");
	set_mlist(ml_search, 0);
}

/* Set up the display to start a new toggle-option command. */
static void
mca_opt_toggle(void)
{
	int no_prompt;
	int flag;
	char *dash;

	no_prompt = (optflag & OPT_NO_PROMPT);
	flag = (optflag & ~OPT_NO_PROMPT);
	dash = (flag == OPT_NO_TOGGLE) ? "_" : "-";

	mca = A_OPT_TOGGLE;
	clear_bot();
	clear_cmd();
	cmd_putstr(dash);
	if (optgetname)
		cmd_putstr(dash);
	if (no_prompt)
		cmd_putstr("(P)");
	switch (flag) {
	case OPT_UNSET:
		cmd_putstr("+");
		break;
	case OPT_SET:
		cmd_putstr("!");
		break;
	}
	set_mlist(NULL, 0);
}

/* Set attnpos. */
static void
set_attnpos(POSITION pos)
{
	int c;

	if (pos != NULL_POSITION) {
		if (ch_seek(pos))
			return;
		for (;;) {
			c = ch_forw_get();
			if (c == EOI)
				return;
			if (c != '\n' && c != '\r')
				break;
			pos++;
		}
	}
	start_attnpos = pos;
	for (;;) {
		c = ch_forw_get();
		pos++;
		if (c == EOI || c == '\n' || c == '\r')
			break;
	}
	end_attnpos = pos;
}


/* Jump to a specified position in the file.
 * Like jump_loc, but the position need not be
 * the first character in a line. */
static void
jump_line_loc(POSITION pos, int sline)
{
	int c;

	if (ch_seek(pos) == 0) {
		/* Back up to the beginning of the line. */
		while ((c = ch_back_get()) != '\n' && c != EOI)
			;
		if (c == '\n')
			(void) ch_forw_get();
		pos = ch_tell();
	}
	if (show_attn)
		set_attnpos(pos);
	jump_loc(pos, sline);
}

/* Routines to perform bracket matching functions. */

/* Try to match the n-th open bracket
 *  which appears in the top displayed line (forwdir),
 * or the n-th close bracket
 *  which appears in the bottom displayed line (!forwdir).
 * The characters which serve as "open bracket" and
 * "close bracket" are given. */
static void
match_brac(int obrac, int cbrac, int forwdir, int n)
{
	int c;
	int nest;
	POSITION pos;
	int (*chget)(void);

	/* Seek to the line containing the open bracket.
	 * This is either the top or bottom line on the screen,
	 * depending on the type of bracket. */
	pos = position((forwdir) ? TOP : BOTTOM);
	if (pos == NULL_POSITION || ch_seek(pos)) {
		if (forwdir)
			error("Nothing in top line", NULL_PARG);
		else
			error("Nothing in bottom line", NULL_PARG);
		return;
	}

	/* Look thru the line to find the open bracket to match. */
	do {
		if ((c = ch_forw_get()) == '\n' || c == EOI) {
			if (forwdir)
				error("No bracket in top line", NULL_PARG);
			else
				error("No bracket in bottom line", NULL_PARG);
			return;
		}
	} while (c != obrac || --n > 0);

	/* Position the file just "after" the open bracket
	 * (in the direction in which we will be searching).
	 * If searching forward, we are already after the bracket.
	 * If searching backward, skip back over the open bracket. */
	if (!forwdir)
		(void) ch_back_get();

	/* Search the file for the matching bracket. */
	chget = (forwdir) ? ch_forw_get : ch_back_get;
	nest = 0;
	while ((c = (*chget)()) != EOI) {
		if (c == obrac)
			nest++;
		else if (c == cbrac && --nest < 0) {
			/* Found the matching bracket.
			 * If searching backward, put it on the top line.
			 * If searching forward, put it on the bottom line. */
			jump_line_loc(ch_tell(), forwdir ? -1 : 1);
			return;
		}
	}
	error("No matching bracket", NULL_PARG);
}

static void toggle_option(int c, char *s, int how_toggle);

/* Execute a multicharacter command. */
static void
exec_mca(void)
{
	char *cbuf;

	cmd_exec();
	cbuf = get_cmdbuf();

	switch (mca) {
	case A_F_SEARCH:
	case A_B_SEARCH:
		multi_search(cbuf, (int) number);
		break;
	case A_FIRSTCMD:
		/* Skip leading spaces or + signs in the string. */
		while (*cbuf == '+' || *cbuf == ' ')
			cbuf++;
		if (every_first_cmd != NULL)
			free(every_first_cmd);
		if (*cbuf == '\0')
			every_first_cmd = NULL;
		else
			every_first_cmd = save(cbuf);
		break;
	case A_OPT_TOGGLE:
		toggle_option(optchar, cbuf, optflag);
		optchar = '\0';
		break;
	case A_F_BRACKET:
		match_brac(cbuf[0], cbuf[1], 1, (int) number);
		break;
	case A_B_BRACKET:
		match_brac(cbuf[1], cbuf[0], 0, (int) number);
		break;
#if EXAMINE
	case A_EXAMINE:
		if (secure)
			break;
		edit_list(cbuf);
#if TAGS
		/* If tag structure is loaded then clean it up. */
		cleantags();
#endif
		break;
#endif
#if SHELL_ESCAPE
	case A_SHELL:
		/* !! just uses whatever is in shellcmd.
		 * Otherwise, copy cmdbuf to shellcmd,
		 * expanding any special characters ("%" or "#"). */
		if (*cbuf != '!') {
			if (shellcmd != NULL)
				free(shellcmd);
			shellcmd = fexpand(cbuf);
		}

		if (secure)
			break;
		if (shellcmd == NULL)
			lsystem("", "!done");
		else
			lsystem(shellcmd, "!done");
		break;
#endif
#if PIPEC
	case A_PIPE:
		if (secure)
			break;
		(void) pipe_mark(pipec, cbuf);
		error("|done", NULL_PARG);
		break;
#endif
	}
}

/* Add a character to a multi-character command. */
static int
mca_char(int c)
{
	const char *p;
	int flag;
	char buf[3];
	PARG parg;

	switch (mca) {
	case 0:
		/* Not in a multicharacter command. */
		return NO_MCA;

	case A_PREFIX:
		/* In the prefix of a command.
		 * This not considered a multichar command
		 * (even tho it uses cmdbuf, etc.).
		 * It is handled in the commands() switch. */
		return NO_MCA;

	case A_DIGIT:
		/* Entering digits of a number.
		 * Terminated by a non-digit. */
		if (!((c >= '0' && c <= '9') || c == '.') &&
		  editchar(c, EC_PEEK|EC_NOHISTORY|EC_NOCOMPLETE|EC_NORIGHTLEFT) == A_INVALID) {
			/* Not part of the number.
			 * Treat as a normal command character. */
			number = cmd_int(&fraction);
			mca = 0;
			cmd_accept();
			return NO_MCA;
		}
		break;

	case A_OPT_TOGGLE:
		/* Special case for the TOGGLE_OPTION command.
		 * If the option letter which was entered is a
		 * single-char option, execute the command immediately,
		 * so user doesn't have to hit RETURN.
		 * If the first char is + or -, this indicates
		 * OPT_UNSET or OPT_SET respectively, instead of OPT_TOGGLE.
		 * "--" begins inputting a long option name. */
		if (optchar == '\0' && len_cmdbuf() == 0) {
			flag = (optflag & ~OPT_NO_PROMPT);
			if (flag == OPT_NO_TOGGLE) {
				switch (c) {
				case '_':
					/* "__" = long option name. */
					optgetname = TRUE;
					mca_opt_toggle();
					return MCA_MORE;
				}
			} else {
				switch (c) {
				case '+':
					/* "-+" = UNSET. */
					optflag = (flag == OPT_UNSET) ?
						OPT_TOGGLE : OPT_UNSET;
					mca_opt_toggle();
					return MCA_MORE;
				case '!':
					/* "-!" = SET */
					optflag = (flag == OPT_SET) ?
						OPT_TOGGLE : OPT_SET;
					mca_opt_toggle();
					return MCA_MORE;
				case CONTROL('P'):
					optflag ^= OPT_NO_PROMPT;
					mca_opt_toggle();
					return MCA_MORE;
				case '-':
					/* "--" = long option name. */
					optgetname = TRUE;
					mca_opt_toggle();
					return MCA_MORE;
				}
			}
		}
		if (optgetname) {
			/* We're getting a long option name.
			 * See if we've matched an option name yet.
			 * If so, display the complete name and stop
			 * accepting chars until user hits RETURN. */
			const struct loption *o;
			const char *oname;
			int lc;

			if (c == '\n' || c == '\r') {
				/* When the user hits RETURN, make sure
				 * we've matched an option name, then
				 * pretend he just entered the equivalent
				 * option letter. */
				if (optchar == '\0') {
					parg.p_string = get_cmdbuf();
					error("There is no --%s option", &parg);
					return MCA_DONE;
				}
				optgetname = FALSE;
				cmd_reset();
				c = optchar;
			} else {
				if (optchar != '\0') {
					/* Already have a match for the name.
					 * Don't accept anything but erase/kill. */
					if (c == erase_char ||
					    c == erase2_char ||
					    c == kill_char)
						return MCA_DONE;
					return MCA_MORE;
				}
				/* Add char to cmd buffer and try to match
				 * the option name. */
				if (cmd_char(c) == CC_QUIT)
					return MCA_DONE;
				p = get_cmdbuf();
				lc = ASCII_IS_LOWER(p[0]);
				o = findopt_name(&p, &oname, NULL);
				if (o != NULL) {
					/* Got a match.
					 * Remember the option letter and
					 * display the full option name. */
					optchar = o->oletter;
					if (!lc && ASCII_IS_LOWER(optchar))
						optchar = ASCII_TO_UPPER(optchar);
					cmd_reset();
					mca_opt_toggle();
					for (;  *oname != '\0';  oname++) {
						c = *oname;
						if (!lc && ASCII_IS_LOWER(c))
							c = ASCII_TO_UPPER(c);
						if (cmd_char(c) != CC_OK)
							return MCA_DONE;
					}
				}
				return MCA_MORE;
			}
		} else {
			if (c == erase_char || c == erase2_char || c == kill_char)
				break;
			if (optchar != '\0')
				/* We already have the option letter. */
				break;
		}

		optchar = c;
		if ((optflag & ~OPT_NO_PROMPT) != OPT_TOGGLE ||
		    single_char_option(c)) {
			toggle_option(c, "", optflag);
			return MCA_DONE;
		}
		/* Display a prompt appropriate for the option letter. */
		if ((p = opt_prompt(c)) == NULL) {
			buf[0] = '-';
			buf[1] = c;
			buf[2] = '\0';
			p = buf;
		}
		start_mca(A_OPT_TOGGLE, p, (void*)NULL, 0);
		return MCA_MORE;

	case A_F_SEARCH:
	case A_B_SEARCH:
		/* Special case for search commands.
		 * Certain characters as the first char of
		 * the pattern have special meaning:
		 *      !  Toggle the NO_MATCH flag
		 *      *  Toggle the PAST_EOF flag
		 *      @  Toggle the FIRST_FILE flag */
		if (len_cmdbuf() > 0)
			/* Only works for the first char of the pattern. */
			break;

		flag = 0;
		switch (c) {
		case CONTROL('E'): /* ignore END of file */
		case '*':
			flag = SRCH_PAST_EOF;
			break;
		case CONTROL('F'): /* FIRST file */
		case '@':
			flag = SRCH_FIRST_FILE;
			break;
		case CONTROL('K'): /* KEEP position */
			flag = SRCH_NO_MOVE;
			break;
		case CONTROL('R'): /* Don't use REGULAR EXPRESSIONS */
			flag = SRCH_NO_REGEX;
			break;
		case CONTROL('N'): /* NOT match */
		case '!':
			flag = SRCH_NO_MATCH;
			break;
		}
		if (flag != 0) {
			search_type ^= flag;
			mca_search();
			return MCA_MORE;
		}
		break;
	}

	/* Any other multicharacter command
	 * is terminated by a newline. */
	if (c == '\n' || c == '\r') {
		/* Execute the command. */
		exec_mca();
		return MCA_DONE;
	}

	/* Append the char to the command buffer. */
	if (cmd_char(c) == CC_QUIT)
		/* Abort the multi-char command. */
		return MCA_DONE;

	if ((mca == A_F_BRACKET || mca == A_B_BRACKET) && len_cmdbuf() >= 2) {
		/* Special case for the bracket-matching commands.
		 * Execute the command after getting exactly two
		 * characters from the user. */
		exec_mca();
		return MCA_DONE;
	}

	/* Need another character. */
	return MCA_MORE;
}


/* Discard any buffered file data. */
static void
clear_buffers(void)
{
	if (!(ch_getflags() & CH_CANSEEK))
		return;
	ch_flush();
	clr_linenum();
#if HILITE_SEARCH
	clr_hilite();
#endif
}

static void jump_forw(void);

/* Make sure the screen is displayed. */
static void
make_display(void)
{
	/* If nothing is displayed yet, display starting from initial_scrpos. */
	if (empty_screen()) {
		if (initial_scrpos.pos == NULL_POSITION)
			/* {{ Maybe this should be:
			 *    jump_loc(ch_zero(), jump_sline);
			 *    but this behavior seems rather unexpected
			 *    on the first screen. }} */
			jump_loc(ch_zero(), 1);
		else
			jump_loc(initial_scrpos.pos, initial_scrpos.ln);
	} else if (screen_trashed) {
		int save_top_scroll = top_scroll;
		int save_ignore_eoi = ignore_eoi;
		top_scroll = 1;
		ignore_eoi = 0;
		if (screen_trashed == 2) {
			/* Special case used by ignore_eoi: re-open the input file
			 * and jump to the end of the file. */
			reopen_curr_ifile();
			jump_forw();
		}
		repaint();
		top_scroll = save_top_scroll;
		ignore_eoi = save_ignore_eoi;
	}
}

/* Display the appropriate prompt. */
static void
prompt(void)
{
	char *p;

	if (ungotp != NULL && ungotp > ungot) {
		/* No prompt necessary if commands are from
		 * ungotten chars rather than from the user. */
		return;
	}

	/* Make sure the screen is displayed. */
	make_display();
	bottompos = position(BOTTOM_PLUS_ONE);

	/* If we've hit EOF on the last file, and the -E flag is set
	 * (or -F is set and this is the first prompt), then quit.
	 * {{ Relying on "first prompt" to detect a single-screen file
	 * fails if +G is used, for example. }} */
	if ((get_quit_at_eof() == OPT_ONPLUS || quit_if_one_screen) &&
	    hit_eof && !(ch_getflags() & CH_HELPFILE) &&
	    next_ifile(curr_ifile) == NULL_IFILE)
		quit(QUIT_OK);
	quit_if_one_screen = FALSE;

	/* Select the proper prompt and display it. */
	/* If the previous action was a forward movement,
	 * don't clear the bottom line of the display;
	 * just print the prompt since the forward movement guarantees
	 * that we're in the right position to display the prompt.
	 * Clearing the line could cause a problem: for example, if the last
	 * line displayed ended at the right screen edge without a newline,
	 * then clearing would clear the last displayed line rather than
	 * the prompt line. */
	if (!forw_prompt)
		clear_bot();
	clear_cmd();
	forw_prompt = 0;
	p = pr_string();
	if (p == NULL || *p == '\0')
		putchr(':');
	else {
		at_enter(AT_STANDOUT);
		putstr(p);
		at_exit();
	}
	clear_eol();
}


/* Search for a pattern, possibly in multiple files.
 * If SRCH_FIRST_FILE is set, begin searching at the first file.
 * If SRCH_PAST_EOF is set, continue the search thru multiple files. */
static void
multi_search(char *pattern, int n)
{
	int nomore;
	IFILE save_ifile;
	int changed_file;

	changed_file = 0;
	save_ifile = save_curr_ifile();

	if (search_type & SRCH_FIRST_FILE) {
		/* Start at the first (or last) file
		 * in the command line list. */
		if (search_type & SRCH_FORW)
			nomore = edit_first();
		else
			nomore = edit_last();
		if (nomore) {
			unsave_ifile(save_ifile);
			return;
		}
		changed_file = 1;
		search_type &= ~SRCH_FIRST_FILE;
	}

	for (;;) {
		n = search(pattern, n);
		/* The SRCH_NO_MOVE flag doesn't "stick": it gets cleared
		 * after being used once.  This allows "n" to work after
		 * using a /@@ search. */
		search_type &= ~SRCH_NO_MOVE;
		if (n == 0) {
			/* Found it. */
			unsave_ifile(save_ifile);
			return;
		}

		if (n < 0)
			/* Some kind of error in the search.
			 * Error message has been printed by search(). */
			break;

		if ((search_type & SRCH_PAST_EOF) == 0)
			/* We didn't find a match, but we're
			 * supposed to search only one file. */
			break;
		/* Move on to the next file. */
		if (search_type & SRCH_FORW)
			nomore = edit_next(1);
		else
			nomore = edit_prev(1);
		if (nomore)
			break;
		changed_file = 1;
	}

	/* Didn't find it.
	 * Print an error message if we haven't already. */
	if (n > 0)
		error("Pattern not found", NULL_PARG);

	if (changed_file) {
		/* Restore the file we were originally viewing. */
		reedit_ifile(save_ifile);
	} else {
		unsave_ifile(save_ifile);
	}
}


/* Hide search string highlighting. */
static void
undo_search(void)
{
	if (!prev_pattern()) {
		error("No previous regular expression", NULL_PARG);
		return;
	}
#if HILITE_SEARCH
	hide_hilite = !hide_hilite;
	repaint_hilite(1);
#endif
}

/*  Routines to search a file for a pattern. */

#ifdef REG_EXTENDED
#define REGCOMP_FLAG    REG_EXTENDED
#else
#define REGCOMP_FLAG    0
#endif

/* Get the length of a buffer needed to convert a string. */
static int
cvt_length(int len)
{
	if (utf_mode)
		/* Just copying a string in UTF-8 mode can cause it to grow
		 * in length.
		 * Six output bytes for one input byte is the worst case
		 * (and unfortunately is far more than is needed in any
		 * non-pathological situation, so this is very wasteful). */
		len *= 6;
	return len + 1;
}


/* Convert text.  Perform one or more of these transformations: */
static void
cvt_text(char *odst, const char *osrc, int *lenp, int ops)
{
	char *dst;
	const char *src;
	const char *src_end;
	LWCHAR ch;

	if (lenp != NULL)
		src_end = osrc + *lenp;
	else
		src_end = osrc + strlen(osrc);

	for (src = osrc, dst = odst;  src < src_end;  ) {
		ch = step_char(&src, +1, src_end);
		if ((ops & CVT_TO_LC) && IS_UPPER(ch)) {
			/* Convert uppercase to lowercase. */
			put_wchar(&dst, TO_LOWER(ch));
		} else if ((ops & CVT_BS) && ch == '\b' && dst > odst) {
			/* Delete backspace and preceding char. */
#if SUPPORT_UTF8
			do {
				dst--;
			} while (dst > odst
					&& !IS_ASCII_OCTET(*dst)
					&& !IS_UTF8_LEAD(*dst)
				);
#else
			dst--;
#endif
		} else if ((ops & CVT_ANSI) && IS_CSI_START(ch)) {
			/* Skip to end of ANSI escape sequence. */
			src++;  /* skip the CSI start char */
			while (src < src_end)
				if (!is_ansi_middle(*src++))
					break;
		} else
			/* Just copy. */
			put_wchar(&dst, ch);
	}
	if ((ops & CVT_CRLF) && dst > odst && dst[-1] == '\r')
		dst--;
	*dst = '\0';
	if (lenp != NULL)
		*lenp = dst - odst;
}

/* Determine which conversions to perform. */
static int
get_cvt_ops(void)
{
	int ops = 0;
	if (is_caseless || bs_mode == BS_SPECIAL) {
		if (is_caseless)
			ops |= CVT_TO_LC;
		if (bs_mode == BS_SPECIAL)
			ops |= CVT_BS;
		if (bs_mode != BS_CONTROL)
			ops |= CVT_CRLF;
	} else if (bs_mode != BS_CONTROL) {
		ops |= CVT_CRLF;
	}
	if (ctldisp == OPT_ONPLUS)
		ops |= CVT_ANSI;
	return ops;
}

/* Are there any uppercase letters in this string? */
static int
is_ucase(const char *str)
{
	const char *str_end = str + strlen(str);
	LWCHAR ch;

	while (str < str_end) {
		ch = step_char(&str, +1, str_end);
		if (IS_UPPER(ch))
			return 1;
	}
	return 0;
}

/* Compile a search pattern, for future use by match_pattern. */
static int
compile_pattern2(char *pattern)
{
#ifndef NO_REGEX
	if ((search_type & SRCH_NO_REGEX) == 0) {
		regex_t *s = emalloc(sizeof(regex_t));
		if (regcomp(s, pattern, REGCOMP_FLAG)) {
			free(s);
			error("Invalid pattern", NULL_PARG);
			return -1;
		}
		if (regpattern != NULL)
			regfree(regpattern);
		regpattern = s;
	}
#endif

	if (last_pattern != NULL)
		free(last_pattern);
	last_pattern = (char *) calloc(1, strlen(pattern)+1);
	if (last_pattern != NULL)
		strcpy(last_pattern, pattern);

	last_search_type = search_type;
	return 0;
}

/* Like compile_pattern, but convert the pattern to lowercase if necessary. */
static int
compile_pattern(char *pattern)
{
	char *cvt_pattern;
	int result;

	if (caseless != OPT_ONPLUS)
		cvt_pattern = pattern;
	else {
		cvt_pattern = emalloc(cvt_length(strlen(pattern)));
		cvt_text(cvt_pattern, pattern, (int *)NULL, CVT_TO_LC);
	}
	result = compile_pattern2(cvt_pattern);
	if (cvt_pattern != pattern)
		free(cvt_pattern);
	return result;
}

static int
match(char *pattern, int pattern_len, char *buf, int buf_len,
		char **pfound, char **pend)
{
	char *pp, *lp;
	char *pattern_end = pattern + pattern_len;
	char *buf_end = buf + buf_len;

	for ( ;  buf < buf_end;  buf++) {
		for (pp = pattern, lp = buf;  *pp == *lp;  pp++, lp++)
			if (pp == pattern_end || lp == buf_end)
				break;
		if (pp == pattern_end) {
			if (pfound != NULL)
				*pfound = buf;
			if (pend != NULL)
				*pend = lp;
			return 1;
		}
	}
	return 0;
}


/* Perform a pattern match with the previously compiled pattern.
 * Set sp and ep to the start and end of the matched string. */
static int
#if NO_REGEX
match_pattern(char *line, int line_len, char **sp, char **ep)
#else
match_pattern(char *line, int line_len, char **sp, char **ep, int notbol)
#endif
{
	int matched;

	if (last_search_type & SRCH_NO_REGEX)
	       return match(last_pattern, strlen(last_pattern), line, line_len, sp, ep);

#if NO_REGEX
	matched = match(last_pattern, strlen(last_pattern), line, line_len, sp, ep);
#else
	{
		regmatch_t rm;
		int flags = (notbol) ? REG_NOTBOL : 0;
		matched = !regexec(regpattern, line, 1, &rm, flags);
		if (!matched)
			return 0;
		*sp = line + rm.rm_so;
		*ep = line + rm.rm_eo;
	}
#endif
	return matched;
}


/* Figure out where to start a search. */
static POSITION
search_pos(void)
{
	POSITION pos;
	int linenum;

	if (empty_screen()) {
		/* Start at the beginning (or end) of the file.
		 * The empty_screen() case is mainly for
		 * command line initiated searches;
		 * for example, "+/xyz" on the command line.
		 * Also for multi-file (SRCH_PAST_EOF) searches. */
		if (search_type & SRCH_FORW) {
			return ch_zero();
		} else {
			pos = ch_length();
			if (pos == NULL_POSITION) {
				(void) ch_end_seek();
				pos = ch_length();
			}
			return pos;
		}
	}
	if (how_search) {
		/* Search does not include current screen. */
		if (search_type & SRCH_FORW)
			linenum = BOTTOM_PLUS_ONE;
		else
			linenum = TOP;
		pos = position(linenum);
	} else {
		/* Search includes current screen.
		 * It starts at the jump target (if searching backwards),
		 * or at the jump target plus one (if forwards). */
		linenum = adjsline(jump_sline);
		pos = position(linenum);
		if (search_type & SRCH_FORW) {
			pos = forw_raw_line(pos, (const char **)NULL, (int *)NULL);
			while (pos == NULL_POSITION) {
				if (++linenum >= sc_height)
					break;
				pos = position(linenum);
			}
		} else {
			while (pos == NULL_POSITION) {
				if (--linenum < 0)
					break;
				pos = position(linenum);
			}
		}
	}
	return pos;
}

/* Search a subset of the file, specified by start/end position. */
static int
search_range(POSITION pos, POSITION endpos, int search_type_r, int matches,
		int maxlines, POSITION *plinepos, POSITION *pendpos_r)
{
	const char *line;
	char *cline;
	int line_len;
	LINENUM linenum;
	char *sp, *ep;
	int line_match;
	int cvt_ops;
	POSITION linepos, oldpos;

	linenum = find_linenum(pos);
	oldpos = pos;
	for (;;) {
		/* Get lines until we find a matching one or until
		 * we hit end-of-file (or beginning-of-file if we're
		 * going backwards), or until we hit the end position. */
		if (ABORT_SIGS()) {
			/* A signal aborts the search. */
			return -1;
		}

		if ((endpos != NULL_POSITION && pos >= endpos) || maxlines == 0) {
			/* Reached end position without a match. */
			if (pendpos_r != NULL)
				*pendpos_r = pos;
			return matches;
		}
		if (maxlines > 0)
			maxlines--;

		if (search_type_r & SRCH_FORW) {
			/* Read the next line, and save the
			 * starting position of that line in linepos. */
			linepos = pos;
			pos = forw_raw_line(pos, &line, &line_len);
			if (linenum != 0)
				linenum++;
		} else {
			/* Read the previous line and save the
			 * starting position of that line in linepos. */
			pos = back_raw_line(pos, &line, &line_len);
			linepos = pos;
			if (linenum != 0)
				linenum--;
		}

		if (pos == NULL_POSITION) {
			/* Reached EOF/BOF without a match. */
			if (pendpos_r != NULL)
				*pendpos_r = oldpos;
			return matches;
		}

		/* If we're using line numbers, we might as well
		 * remember the information we have now (the position
		 * and line number of the current line).
		 * Don't do it for every line because it slows down
		 * the search.  Remember the line number only if
		 * we're "far" from the last place we remembered it. */
		if (linenums && abs((int)(pos - oldpos)) > 1024)
			add_lnum(linenum, pos);
		oldpos = pos;

		/* If it's a caseless search, convert the line to lowercase.
		 * If we're doing backspace processing, delete backspaces. */
		cvt_ops = get_cvt_ops();
		cline = calloc(1, cvt_length(line_len));
		cvt_text(cline, line, &line_len, cvt_ops);

		/* Test the next line to see if we have a match.
		 * We are successful if we either want a match and got one,
		 * or if we want a non-match and got one. */
		line_match = MATCH_PATTERN(cline, line_len, &sp, &ep, 0);
		line_match = (!(search_type_r & SRCH_NO_MATCH) && line_match) ||
				((search_type_r & SRCH_NO_MATCH) && !line_match);
		if (!line_match) {
			free(cline);
			continue;
		}
		/* Got a match. */
		if (search_type_r & SRCH_FIND_ALL) {
#if HILITE_SEARCH
			/* We are supposed to find all matches in the range.
			 * Just add the matches in this line to the
			 * hilite list and keep searching. */
			if (line_match)
				hilite_line(linepos, cline, line_len, sp, ep, cvt_ops);
#endif
			free(cline);
		} else if (--matches <= 0) {
			/* Found the one match we're looking for.
			 * Return it. */
#if HILITE_SEARCH
			if (hilite_search == OPT_ON) {
				/* Clear the hilite list and add only
				 * the matches in this one line. */
				clr_hilite();
				if (line_match)
					hilite_line(linepos, cline, line_len, sp, ep, cvt_ops);
			}
#endif
			free(cline);
			if (plinepos != NULL)
				*plinepos = linepos;
			return 0;
		}
	}
}

#if CMD_HISTORY
/* search for a pattern in history. If found, compile that pattern. */
static int
hist_pattern(void)
{
	char *pattern;

	set_mlist(ml_search, 0);
	pattern = cmd_lastpattern();
	if (pattern == NULL)
		return 0;

	if (compile_pattern(pattern) < 0)
		return 0;

	is_ucase_pattern = is_ucase(pattern);
	if (is_ucase_pattern && caseless != OPT_ONPLUS)
		is_caseless = 0;
	else
		is_caseless = caseless;

#if HILITE_SEARCH
	if (hilite_search == OPT_ONPLUS && !hide_hilite)
		hilite_screen();
#endif

	return 1;
}
#else /* CMD_HISTORY */
#define hist_pattern()       0
#endif /* CMD_HISTORY */

/* Search for the n-th occurrence of a specified pattern,
 * either forward or backward.
 * Return the number of matches not yet found in this file
 * (that is, n minus the number of matches found).
 * Return -1 if the search should be aborted.
 * Caller may continue the search in another file
 * if less than n matches are found in this file. */
static int
search(char *pattern, int n)
{
	POSITION pos;

	if (pattern == NULL || *pattern == '\0') {
		/* A null pattern means use the previously compiled pattern. */
		if (!prev_pattern() && !hist_pattern()) {
			error("No previous regular expression", NULL_PARG);
			return -1;
		}
		if ((search_type & SRCH_NO_REGEX) !=
		    (last_search_type & SRCH_NO_REGEX)) {
			error("Please re-enter search pattern", NULL_PARG);
			return -1;
		}
#if HILITE_SEARCH
		if (hilite_search == OPT_ON) {
			/* Erase the highlights currently on screen.
			 * If the search fails, we'll redisplay them later. */
			repaint_hilite(0);
		}
		if (hilite_search == OPT_ONPLUS && hide_hilite) {
			/* Highlight any matches currently on screen,
			 * before we actually start the search. */
			hide_hilite = 0;
			hilite_screen();
		}
		hide_hilite = 0;
#endif
	} else {
		/* Compile the pattern. */
		if (compile_pattern(pattern) < 0)
			return -1;
		/* Ignore case if -I is set OR
		 * -i is set AND the pattern is all lowercase. */
		is_ucase_pattern = is_ucase(pattern);
		if (is_ucase_pattern && caseless != OPT_ONPLUS)
			is_caseless = 0;
		else
			is_caseless = caseless;
#if HILITE_SEARCH
		if (hilite_search) {
			/* Erase the highlights currently on screen.
			 * Also permanently delete them from the hilite list. */
			repaint_hilite(0);
			hide_hilite = 0;
			clr_hilite();
		}
		if (hilite_search == OPT_ONPLUS) {
			/* Highlight any matches currently on screen,
			 * before we actually start the search. */
			hilite_screen();
		}
#endif
	}

	/* Figure out where to start the search. */
	pos = search_pos();
	if (pos == NULL_POSITION) {
		/* Can't find anyplace to start searching from. */
		if (search_type & SRCH_PAST_EOF)
			return n;
		/* repaint(); -- why was this here? */
		error("Nothing to search", NULL_PARG);
		return -1;
	}

	n = search_range(pos, NULL_POSITION, search_type, n, -1,
			&pos, (POSITION*)NULL);
	if (n != 0) {
		/* Search was unsuccessful. */
#if HILITE_SEARCH
		if (hilite_search == OPT_ON && n > 0)
			/* Redisplay old hilites. */
			repaint_hilite(1);
#endif
		return n;
	}

	if (!(search_type & SRCH_NO_MOVE)) {
		/* Go to the matching line. */
		jump_loc(pos, jump_sline);
	}

#if HILITE_SEARCH
	if (hilite_search == OPT_ON)
		/* Display new hilites in the matching line. */
		repaint_hilite(1);
#endif
	return 0;
}

/* Simple pattern matching function.
 * It supports no metacharacters like *, etc. */


/* Seek to the beginning of the file, or as close to it as we can get.
 * We may not be able to seek there if input is a pipe and the
 * beginning of the pipe is no longer buffered. */
static int
ch_beg_seek(void)
{
	struct buf *bp, *firstbp;

	/* Try a plain ch_seek first. */
	if (ch_seek(ch_zero()) == 0)
		return 0;

	/* Can't get to position 0.
	 * Look thru the buffers for the one closest to position 0. */
	firstbp = bp = ch_bufhead;
	if (bp == END_OF_CHAIN)
		return 1;
	while ((bp = bp->next) != END_OF_CHAIN)
		if (bp->block < firstbp->block)
			firstbp = bp;
	ch_block = firstbp->block;
	ch_offset = 0;
	return 0;
}


/* Routines which jump to a new location in the file. */

/* Jump to the end of the file. */
static void
jump_forw(void)
{
	POSITION pos;
	POSITION end_pos;

	if (ch_end_seek()) {
		error("Cannot seek to end of file", NULL_PARG);
		return;
	}
	/* Position the last line in the file at the last screen line.
	 * Go back one line from the end of the file
	 * to get to the beginning of the last line. */
	pos_clear();
	end_pos = ch_tell();
	pos = back_line(end_pos);
	if (pos == NULL_POSITION)
		jump_loc((POSITION)0, sc_height-1);
	else {
		jump_loc(pos, sc_height-1);
		if (position(sc_height-1) != end_pos)
			repaint();
	}
}

/* Jump to line n in the file. */
static void
jump_back(LINENUM linenum)
{
	POSITION pos;
	PARG parg;

	/* Find the position of the specified line.
	 * If we can seek there, just jump to it.
	 * If we can't seek, but we're trying to go to line number 1,
	 * use ch_beg_seek() to get as close as we can. */
	pos = find_pos(linenum);
	if (pos != NULL_POSITION && ch_seek(pos) == 0) {
		if (show_attn)
			set_attnpos(pos);
		jump_loc(pos, jump_sline);
	} else if (linenum <= 1 && ch_beg_seek() == 0) {
		jump_loc(ch_tell(), jump_sline);
		error("Cannot seek to beginning of file", NULL_PARG);
	} else {
		parg.p_linenum = linenum;
		error("Cannot seek to line number %n", &parg);
	}
}


/* Jump to a specified percentage into the file. */
static void
jump_percent(int percent)
{
	POSITION pos, len;

	/* Determine the position in the file
	 * (the specified percentage of the file's length). */
	if ((len = ch_length()) == NULL_POSITION) {
		ierror("Determining length of file", NULL_PARG);
		ch_end_seek();
	}
	if ((len = ch_length()) == NULL_POSITION) {
		error("Don't know length of file", NULL_PARG);
		return;
	}
	pos = percent_pos(len, percent);
	if (pos >= len)
		pos = len-1;

	jump_line_loc(pos, jump_sline);
}


/* Define the printing format for control (or binary utf) chars. */
static void
setbinfmt(const char *s, const char **fmtvarptr, const char *default_fmt)
{
	if (s && utf_mode) {
		/* It would be too hard to account for width otherwise. */
		const char *t = s;
		while (*t) {
			if (*t < ' ' || *t > '~') {
				s = default_fmt;
				goto l_attr;
			}
			t++;
		}
	}

	/* %n is evil */
	if (s == NULL || *s == '\0' ||
	    (*s == '*' && (s[1] == '\0' || s[2] == '\0' || strchr(s + 2, 'n'))) ||
	    (*s != '*' && strchr(s, 'n')))
		s = default_fmt;

	/* Select the attributes if it starts with "*". */
 l_attr:
	if (*s == '*') {
		switch (s[1]) {
		case 'd':  binattr = AT_BOLD;      break;
		case 'k':  binattr = AT_BLINK;     break;
		case 's':  binattr = AT_STANDOUT;  break;
		case 'u':  binattr = AT_UNDERLINE; break;
		default:   binattr = AT_NORMAL;    break;
		}
		s += 2;
	}
	*fmtvarptr = s;
}

/* */
static void
set_charset(void)
{
	const char *s;

	/* See if environment variable LESSCHARSET is defined. */
	s = lgetenv("LESSCHARSET");
	if (icharset(s, 0))
		return;

	/* LESSCHARSET is not defined: try LESSCHARDEF. */
	s = lgetenv("LESSCHARDEF");
	if (s != NULL && *s != '\0') {
		ichardef(s);
		return;
	}

#if HAVE_LOCALE
#ifdef CODESET
	/* Try using the codeset name as the charset name. */
	s = nl_langinfo(CODESET);
	if (icharset(s, 1))
		return;
#endif
#endif

#if SUPPORT_UTF8
	/* Check whether LC_ALL, LC_CTYPE or LANG look like UTF-8 is used. */
	if ((s = lgetenv("LC_ALL")) != NULL ||
	    (s = lgetenv("LC_CTYPE")) != NULL ||
	    (s = lgetenv("LANG")) != NULL) {
		if (   strstr(s, "UTF-8") != NULL || strstr(s, "utf-8") != NULL
		    || strstr(s, "UTF8")  != NULL || strstr(s, "utf8")  != NULL)
			if (icharset("utf-8", 1))
				return;
	}
#endif

#if HAVE_LOCALE
	/* Get character definitions from locale functions,
	 * rather than from predefined charset entry. */
	ilocale();
	/* Default to "latin1". */
	(void) icharset("latin1", 1);
#endif
}

/* Routines to decode user commands.
 *
 * This is all table driven.
 * A command table is a sequence of command descriptors.
 * Each command descriptor is a sequence of bytes with the following format:
 *      <c1><c2>...<cN><0><action>
 * The characters c1,c2,...,cN are the command string; that is,
 * the characters which the user must type.
 * It is terminated by a null <0> byte.
 * The byte after the null byte is the action code associated
 * with the command string.
 * If an action byte is OR-ed with A_EXTRA, this indicates
 * that the option byte is followed by an extra string.
 *
 * There may be many command tables.
 * The first (default) table is built-in.
 * Other tables are read in from "lesskey" files.
 * All the tables are linked together and are searched in order. */

/* Format of a lesskey file:
 *
 *      LESSKEY_MAGIC (4 bytes)
 *       sections...
 *      END_LESSKEY_MAGIC (4 bytes)
 *
 * Each section is:
 *
 *      section_MAGIC (1 byte)
 *      section_length (2 bytes)
 *      key table (section_length bytes) */
#define C0_LESSKEY_MAGIC        '\0'
#define C1_LESSKEY_MAGIC        'M'
#define C2_LESSKEY_MAGIC        '+'
#define C3_LESSKEY_MAGIC        'G'

#define CMD_SECTION             'c'
#define EDIT_SECTION            'e'
#define VAR_SECTION             'v'
#define END_SECTION             'x'

#define C0_END_LESSKEY_MAGIC    'E'
#define C1_END_LESSKEY_MAGIC    'n'
#define C2_END_LESSKEY_MAGIC    'd'

/* */
#define KRADIX          64

#define SK(k) \
	SK_SPECIAL_KEY, (k), 6, 1, 1, 1
/* Command table is ordered roughly according to expected
 * frequency of use, so the common commands are near the beginning. */

static const unsigned char cmdtable[] =
{
	'\r',0,                         A_F_LINE,
	'\n',0,                         A_F_LINE,
	'e',0,                          A_F_LINE,
	'j',0,                          A_F_LINE,
	SK(SK_DOWN_ARROW),0,            A_F_LINE,
	CONTROL('E'),0,                 A_F_LINE,
	CONTROL('N'),0,                 A_F_LINE,
	'k',0,                          A_B_LINE,
	'y',0,                          A_B_LINE,
	CONTROL('Y'),0,                 A_B_LINE,
	SK(SK_CONTROL_K),0,             A_B_LINE,
	CONTROL('P'),0,                 A_B_LINE,
	SK(SK_UP_ARROW),0,              A_B_LINE,
	'J',0,                          A_FF_LINE,
	'K',0,                          A_BF_LINE,
	'Y',0,                          A_BF_LINE,
	'd',0,                          A_F_SCROLL,
	CONTROL('D'),0,                 A_F_SCROLL,
	'u',0,                          A_B_SCROLL,
	CONTROL('U'),0,                 A_B_SCROLL,
	' ',0,                          A_F_SCREEN,
	'f',0,                          A_F_SCREEN,
	CONTROL('F'),0,                 A_F_SCREEN,
	CONTROL('V'),0,                 A_F_SCREEN,
	SK(SK_PAGE_DOWN),0,             A_F_SCREEN,
	'b',0,                          A_B_SCREEN,
	CONTROL('B'),0,                 A_B_SCREEN,
	ESC,'v',0,                      A_B_SCREEN,
	SK(SK_PAGE_UP),0,               A_B_SCREEN,
	'z',0,                          A_F_WINDOW,
	'w',0,                          A_B_WINDOW,
	ESC,' ',0,                      A_FF_SCREEN,
	'F',0,                          A_F_FOREVER,
	'R',0,                          A_FREPAINT,
	'r',0,                          A_REPAINT,
	CONTROL('R'),0,                 A_REPAINT,
	CONTROL('L'),0,                 A_REPAINT,
	ESC,'u',0,                      A_UNDO_SEARCH,
	'g',0,                          A_GOLINE,
	ESC,'[','H',0,                  A_GOLINE,
	SK(SK_HOME),0,                  A_GOLINE,
	'<',0,                          A_GOLINE,
	ESC,'<',0,                      A_GOLINE,
	'p',0,                          A_PERCENT,
	'%',0,                          A_PERCENT,
	ESC,'[',0,                      A_LSHIFT,
	ESC,']',0,                      A_RSHIFT,
	ESC,'(',0,                      A_LSHIFT,
	ESC,')',0,                      A_RSHIFT,
	SK(SK_RIGHT_ARROW),0,           A_RSHIFT,
	SK(SK_LEFT_ARROW),0,            A_LSHIFT,
	'{',0,                          A_F_BRACKET|A_EXTRA,    '{','}',0,
	'}',0,                          A_B_BRACKET|A_EXTRA,    '{','}',0,
	'(',0,                          A_F_BRACKET|A_EXTRA,    '(',')',0,
	')',0,                          A_B_BRACKET|A_EXTRA,    '(',')',0,
	'[',0,                          A_F_BRACKET|A_EXTRA,    '[',']',0,
	']',0,                          A_B_BRACKET|A_EXTRA,    '[',']',0,
	ESC,CONTROL('F'),0,             A_F_BRACKET,
	ESC,CONTROL('B'),0,             A_B_BRACKET,
	'G',0,                          A_GOEND,
	ESC,'>',0,                      A_GOEND,
	'>',0,                          A_GOEND,
	ESC,'[','F',0,                  A_GOEND,
	SK(SK_END),0,                   A_GOEND,
	'P',0,                          A_GOPOS,

	'0',0,                          A_DIGIT,
	'1',0,                          A_DIGIT,
	'2',0,                          A_DIGIT,
	'3',0,                          A_DIGIT,
	'4',0,                          A_DIGIT,
	'5',0,                          A_DIGIT,
	'6',0,                          A_DIGIT,
	'7',0,                          A_DIGIT,
	'8',0,                          A_DIGIT,
	'9',0,                          A_DIGIT,
	'.',0,                          A_DIGIT,

	'=',0,                          A_STAT,
	CONTROL('G'),0,                 A_STAT,
	':','f',0,                      A_STAT,
	'/',0,                          A_F_SEARCH,
	'?',0,                          A_B_SEARCH,
	ESC,'/',0,                      A_F_SEARCH|A_EXTRA,     '*',0,
	ESC,'?',0,                      A_B_SEARCH|A_EXTRA,     '*',0,
	'n',0,                          A_AGAIN_SEARCH,
	ESC,'n',0,                      A_T_AGAIN_SEARCH,
	'N',0,                          A_REVERSE_SEARCH,
	ESC,'N',0,                      A_T_REVERSE_SEARCH,
	'm',0,                          A_SETMARK,
	'\'',0,                         A_GOMARK,
	CONTROL('X'),CONTROL('X'),0,    A_GOMARK,
	'E',0,                          A_EXAMINE,
	':','e',0,                      A_EXAMINE,
	CONTROL('X'),CONTROL('V'),0,    A_EXAMINE,
	':','n',0,                      A_NEXT_FILE,
	':','p',0,                      A_PREV_FILE,
	't',0,                          A_NEXT_TAG,
	'T',0,                          A_PREV_TAG,
	':','x',0,                      A_INDEX_FILE,
	':','d',0,                      A_REMOVE_FILE,
	'-',0,                          A_OPT_TOGGLE,
	':','t',0,                      A_OPT_TOGGLE|A_EXTRA,   't',0,
	's',0,                          A_OPT_TOGGLE|A_EXTRA,   'o',0,
	'_',0,                          A_DISP_OPTION,
	'|',0,                          A_PIPE,
	'v',0,                          A_VISUAL,
	'!',0,                          A_SHELL,
	'+',0,                          A_FIRSTCMD,

	'H',0,                          A_HELP,
	'h',0,                          A_HELP,
	SK(SK_F1),0,                    A_HELP,
	'V',0,                          A_VERSION,
	'q',0,                          A_QUIT,
	'Q',0,                          A_QUIT,
	':','q',0,                      A_QUIT,
	':','Q',0,                      A_QUIT,
	'Z','Z',0,                      A_QUIT
};

static const unsigned char edittable[] =
{
	'\t',0,                         EC_F_COMPLETE,  /* TAB */
	'\17',0,                        EC_B_COMPLETE,  /* BACKTAB */
	SK(SK_BACKTAB),0,               EC_B_COMPLETE,  /* BACKTAB */
	ESC,'\t',0,                     EC_B_COMPLETE,  /* ESC TAB */
	CONTROL('L'),0,                 EC_EXPAND,      /* CTRL-L */
	CONTROL('V'),0,                 EC_LITERAL,     /* BACKSLASH */
	CONTROL('A'),0,                 EC_LITERAL,     /* BACKSLASH */
	ESC,'l',0,                      EC_RIGHT,       /* ESC l */
	SK(SK_RIGHT_ARROW),0,           EC_RIGHT,       /* RIGHTARROW */
	ESC,'h',0,                      EC_LEFT,        /* ESC h */
	SK(SK_LEFT_ARROW),0,            EC_LEFT,        /* LEFTARROW */
	ESC,'b',0,                      EC_W_LEFT,      /* ESC b */
	ESC,SK(SK_LEFT_ARROW),0,        EC_W_LEFT,      /* ESC LEFTARROW */
	SK(SK_CTL_LEFT_ARROW),0,        EC_W_LEFT,      /* CTRL-LEFTARROW */
	ESC,'w',0,                      EC_W_RIGHT,     /* ESC w */
	ESC,SK(SK_RIGHT_ARROW),0,       EC_W_RIGHT,     /* ESC RIGHTARROW */
	SK(SK_CTL_RIGHT_ARROW),0,       EC_W_RIGHT,     /* CTRL-RIGHTARROW */
	ESC,'i',0,                      EC_INSERT,      /* ESC i */
	SK(SK_INSERT),0,                EC_INSERT,      /* INSERT */
	ESC,'x',0,                      EC_DELETE,      /* ESC x */
	ESC,'[','3','~',0,              EC_DELETE,      /* DELETE */
	SK(SK_DELETE),0,                EC_DELETE,      /* DELETE */
	ESC,'X',0,                      EC_W_DELETE,    /* ESC X */
	ESC,SK(SK_DELETE),0,            EC_W_DELETE,    /* ESC DELETE */
	SK(SK_CTL_DELETE),0,            EC_W_DELETE,    /* CTRL-DELETE */
	SK(SK_CTL_BACKSPACE),0,         EC_W_BACKSPACE, /* CTRL-BACKSPACE */
	ESC,'\b',0,                     EC_W_BACKSPACE, /* ESC BACKSPACE */
	ESC,'0',0,                      EC_HOME,        /* ESC 0 */
	ESC,'[','H',0,                  EC_HOME,        /* HOME */
	SK(SK_HOME),0,                  EC_HOME,        /* HOME */
	ESC,'$',0,                      EC_END,         /* ESC $ */
	ESC,'[','F',0,                  EC_END,         /* END */
	SK(SK_END),0,                   EC_END,         /* END */
	ESC,'k',0,                      EC_UP,          /* ESC k */
	SK(SK_UP_ARROW),0,              EC_UP,          /* UPARROW */
	ESC,'j',0,                      EC_DOWN,        /* ESC j */
	SK(SK_DOWN_ARROW),0,            EC_DOWN,        /* DOWNARROW */
};



/* Expand special key abbreviations in a command table. */
static void
expand_special_keys(char *keys_table, int len)
{
	char *fm;
	char *to;
	int a;
	const char *repl;
	int klen;

	for (fm = keys_table;  fm < keys_table + len; ) {
		/* Rewrite each command in the table with any
		 * special key abbreviations expanded. */
		for (to = fm;  *fm != '\0'; ) {
			if (*fm != SK_SPECIAL_KEY) {
				*to++ = *fm++;
				continue;
			}
			/* After SK_SPECIAL_KEY, next byte is the type
			 * of special key (one of the SK_* contants),
			 * and the byte after that is the number of bytes,
			 * N, reserved by the abbreviation (including the
			 * SK_SPECIAL_KEY and key type bytes).
			 * Replace all N bytes with the actual bytes
			 * output by the special key on this terminal. */
			repl = special_key_str(fm[1]);
			klen = fm[2] & 0377;
			fm += klen;
			if (repl == NULL || (int) strlen(repl) > klen)
				repl = "\377";
			while (*repl != '\0')
				*to++ = *repl++;
		}
		*to++ = '\0';
		/* Fill any unused bytes between end of command and
		 * the action byte with A_SKIP. */
		while (to <= fm)
			*to++ = A_SKIP;
		fm++;
		a = *fm++ & 0377;
		if (a & A_EXTRA) {
			while (*fm++ != '\0')
				continue;
		}
	}
}

/* Add a command table. */
static int
add_cmd_table(struct tablelist **tlist, char *buf, int len)
{
	struct tablelist *t;

	if (len == 0)
		return 0;
	/* Allocate a tablelist structure, initialize it,
	 * and link it into the list of tables. */
	if ((t = (struct tablelist *)
			calloc(1, sizeof(struct tablelist))) == NULL) {
		return -1;
	}
	expand_special_keys(buf, len);
	t->t_start = buf;
	t->t_end = buf + len;
	t->t_next = *tlist;
	*tlist = t;
	return 0;
}

/* Add a command table. */
static void
add_fcmd_table(char *buf, int len)
{
	if (add_cmd_table(&list_fcmd_tables, buf, len) < 0)
		error("Warning: some commands disabled", NULL_PARG);
}

/* Add an editing command table. */
static void
add_ecmd_table(char *buf, int len)
{
	if (add_cmd_table(&list_ecmd_tables, buf, len) < 0)
		error("Warning: some edit commands disabled", NULL_PARG);
}


#if USERFILE
/* Add an environment variable table. */
static void
add_var_table(struct tablelist **tlist, char *buf, int len)
{
	if (add_cmd_table(tlist, buf, len) < 0)
		error("Warning: environment variables from lesskey file unavailable", NULL_PARG);
}

/* Get an "integer" from a lesskey file.
 * Integers are stored in a funny format:
 * two bytes, low order first, in radix KRADIX. */
static int
gint(char **sp)
{
	int n;

	n = *(*sp)++;
	n += *(*sp)++ * KRADIX;
	return n;
}

/* Process a new (post-v241) lesskey file. */
static int
new_lesskey(char *buf, int len, int sysvar)
{
	char *p;
	int c;
	int n;

	/* New-style lesskey file. Extract the pieces. */
	if (buf[len-3] != C0_END_LESSKEY_MAGIC ||
	    buf[len-2] != C1_END_LESSKEY_MAGIC ||
	    buf[len-1] != C2_END_LESSKEY_MAGIC)
		return -1;
	p = buf + 4;
	for (;;) {
		c = *p++;
		switch (c) {
		case CMD_SECTION:
			n = gint(&p);
			add_fcmd_table(p, n);
			p += n;
			break;
		case EDIT_SECTION:
			n = gint(&p);
			add_ecmd_table(p, n);
			p += n;
			break;
		case VAR_SECTION:
			n = gint(&p);
			add_var_table((sysvar) ?
				&list_sysvar_tables : &list_var_tables, p, n);
			p += n;
			break;
		case END_SECTION:
			return 0;
		default:
			/* Unrecognized section type. */
			return -1;
		}
	}
}

/* Process an old (pre-v241) lesskey file. */
static int
old_lesskey(char *buf, int len)
{
	/* Old-style lesskey file.
	 * The file must end with either
	 *     ...,cmd,0,action
	 * or  ...,cmd,0,action|A_EXTRA,string,0
	 * So the last byte or the second to last byte must be zero. */
	if (buf[len-1] != '\0' && buf[len-2] != '\0')
		return -1;
	add_fcmd_table(buf, len);
	return 0;
}

/* Set up a user command table, based on a "lesskey" file. */
static int
lesskey(char *filename, int sysvar)
{
	char *buf;
	POSITION len;
	long n;
	int f;

	if (secure)
		return 1;
	/* Try to open the lesskey file. */
	filename = shell_unquote(filename);
	f = open(filename, O_RDONLY);
	free(filename);
	if (f < 0)
		return 1;

	/* Read the file into a buffer.
	 * We first figure out the size of the file and allocate space for it.
	 * {{ Minimal error checking is done here.
	 *    A garbage .less file will produce strange results.
	 *    To avoid a large amount of error checking code here, we
	 *    rely on the lesskey program to generate a good .less file. }} */
	len = filesize(f);
	if (len == NULL_POSITION || len < 3) {
		/* Bad file (valid file must have at least 3 chars). */
		close(f);
		return -1;
	}
	if ((buf = (char *) calloc((int)len, sizeof(char))) == NULL) {
		close(f);
		return -1;
	}
	if (lseek(f, (off_t)0, SEEK_SET) == BAD_LSEEK) {
		free(buf);
		close(f);
		return -1;
	}
	n = read(f, buf, (unsigned int) len);
	close(f);
	if (n != len) {
		free(buf);
		return -1;
	}

	/* Figure out if this is an old-style (before version 241)
	 * or new-style lesskey file format. */
	if (buf[0] != C0_LESSKEY_MAGIC || buf[1] != C1_LESSKEY_MAGIC ||
	    buf[2] != C2_LESSKEY_MAGIC || buf[3] != C3_LESSKEY_MAGIC)
		return old_lesskey(buf, (int)len);
	return new_lesskey(buf, (int)len, sysvar);
}

/* Return a pathname that points to a specified file in a specified directory.
 * Return NULL if the file does not exist in the directory. */
static char *
dirfile(char *dirname, char *filename)
{
	char *pathname;
	char *qpathname;
	int len;
	int f;

	if (dirname == NULL || *dirname == '\0')
		return NULL;
	/* Construct the full pathname. */
	len= strlen(dirname) + strlen(filename) + 2;
	pathname = (char *) calloc(len, sizeof(char));
	if (pathname == NULL)
		return NULL;
	snprintf(pathname, len, "%s/%s", dirname, filename);
	/* Make sure the file exists. */
	qpathname = shell_unquote(pathname);
	f = open(qpathname, O_RDONLY);
	if (f < 0) {
		free(pathname);
		pathname = NULL;
	} else {
		close(f);
	}
	free(qpathname);
	return pathname;
}

/* Return the full pathname of the given file in the "home directory". */
static char *
homefile(char *filename)
{
	char *pathname;

	/* Try $HOME/filename. */
	pathname = dirfile(lgetenv("HOME"), filename);
	if (pathname != NULL)
		return pathname;
	return NULL;
}

/* Add the standard lesskey file "$HOME/.less" */
static void
add_hometable(char *envname, char *def_filename, int sysvar)
{
	char *filename;
	PARG parg;

	if (envname != NULL && (filename = lgetenv(envname)) != NULL)
		filename = save(filename);
	else if (sysvar)
		filename = save(def_filename);
	else
		filename = homefile(def_filename);
	if (filename == NULL)
		return;
	if (lesskey(filename, sysvar) < 0) {
		parg.p_string = filename;
		error("Cannot use lesskey file \"%s\"", &parg);
	}
	free(filename);
}
#endif /* USERFILE */

/* Initialize the command lists. */
static void
init_cmds(void)
{
	/* Add the default command tables. */
	char *p = malloc(sizeof(cmdtable));
	memcpy(p, cmdtable, sizeof(cmdtable));
	add_fcmd_table(p, sizeof(cmdtable));
	p = malloc(sizeof(edittable));
	memcpy(p, edittable, sizeof(edittable));
	add_ecmd_table(p, sizeof(edittable));
#if USERFILE
	/* For backwards compatibility,
	 * try to add tables in the OLD system lesskey file. */
#ifdef BINDIR
	add_hometable(NULL, BINDIR "/.sysless", 1);
#endif
	/* Try to add the tables in the system lesskey file. */
	add_hometable("LESSKEY_SYSTEM", LESSKEYFILE_SYS, 1);
	/* Try to add the tables in the standard lesskey file "$HOME/.less". */
	add_hometable("LESSKEY", LESSKEYFILE, 0);
#endif
}

/* Decode a command from the cmdtables list. */
static int
fcmd_decode(const char *cmd, const char **sp)
{
	return cmd_decode(list_fcmd_tables, cmd, sp);
}


/* Main command processor.
 * Accept and execute commands until a quit command. */
static void
commands(void)
{
	int c;
	int action;
	char *cbuf;
	int newaction;
	int save_search_type;
	const char *extra;
	char tbuf[2];
	PARG parg;
	IFILE c_old_ifile;
	IFILE c_new_ifile;
#if TAGS
	char *tagfile;
#endif

	search_type = SRCH_FORW;
	wscroll = (sc_height + 1) / 2;
	newaction = A_NOACTION;

	for (;;) {
		mca = 0;
		cmd_accept();
		number = 0;
		optchar = '\0';

		/* See if any signals need processing. */
		if (sigs) {
			psignals();
			if (quitting)
				quit(QUIT_SAVED_STATUS);
		}

		/* Display prompt and accept a character. */
		cmd_reset();
		prompt();
		if (sigs)
			continue;
		if (newaction == A_NOACTION)
			c = getcc();

	again:
		if (sigs)
			continue;

		if (newaction != A_NOACTION) {
			action = newaction;
			newaction = A_NOACTION;
		} else {
			/* If we are in a multicharacter command, call mca_char.
			 * Otherwise we call fcmd_decode to determine the
			 * action to be performed. */
			if (mca)
				switch (mca_char(c)) {
				case MCA_MORE:
					/* Need another character. */
					c = getcc();
					goto again;
				case MCA_DONE:
					/* Command has been handled by mca_char.
					 * Start clean with a prompt. */
					continue;
				case NO_MCA:
					/* Not a multi-char command
					 * (at least, not anymore). */
					break;
				}

			/* Decode the command character and decide what to do. */
			if (mca) {
				/* We're in a multichar command.
				 * Add the character to the command buffer
				 * and display it on the screen.
				 * If the user backspaces past the start
				 * of the line, abort the command. */
				if (cmd_char(c) == CC_QUIT || len_cmdbuf() == 0)
					continue;
				cbuf = get_cmdbuf();
			} else {
				/* Don't use cmd_char if we're starting fresh
				 * at the beginning of a command, because we
				 * don't want to echo the command until we know
				 * it is a multichar command.  We also don't
				 * want erase_char/kill_char to be treated
				 * as line editing characters. */
				tbuf[0] = c;
				tbuf[1] = '\0';
				cbuf = tbuf;
			}
			extra = NULL;
			action = fcmd_decode(cbuf, &extra);
			/* If an "extra" string was returned,
			 * process it as a string of command characters. */
			if (extra != NULL)
				ungetsc(extra);
		}
		/* Clear the cmdbuf string.
		 * (But not if we're in the prefix of a command,
		 * because the partial command string is kept there.) */
		if (action != A_PREFIX)
			cmd_reset();

		switch (action) {
		case A_DIGIT:
			/* First digit of a number. */
			start_mca(A_DIGIT, ":", (void*)NULL, CF_QUIT_ON_ERASE);
			goto again;

		case A_F_WINDOW:
			/* Forward one window (and set the window size). */
			if (number > 0)
				swindow = (int) number;
			/* FALLTHRU */
		case A_F_SCREEN:
			/* Forward one screen. */
			if (number <= 0)
				number = get_swindow();
			cmd_exec();
			if (show_attn)
				set_attnpos(bottompos);
			forward((int) number, 0, 1);
			break;

		case A_B_WINDOW:
			/* Backward one window (and set the window size). */
			if (number > 0)
				swindow = (int) number;
			/* FALLTHRU */
		case A_B_SCREEN:
			/* Backward one screen. */
			if (number <= 0)
				number = get_swindow();
			cmd_exec();
			backward((int) number, 0, 1);
			break;

		case A_F_LINE:
			/* Forward N (default 1) line. */
			if (number <= 0)
				number = 1;
			cmd_exec();
			if (show_attn == OPT_ONPLUS && number > 1)
				set_attnpos(bottompos);
			forward((int) number, 0, 0);
			break;

		case A_B_LINE:
			/* Backward N (default 1) line. */
			if (number <= 0)
				number = 1;
			cmd_exec();
			backward((int) number, 0, 0);
			break;

		case A_FF_LINE:
			/* Force forward N (default 1) line. */
			if (number <= 0)
				number = 1;
			cmd_exec();
			if (show_attn == OPT_ONPLUS && number > 1)
				set_attnpos(bottompos);
			forward((int) number, 1, 0);
			break;

		case A_BF_LINE:
			/* Force backward N (default 1) line. */
			if (number <= 0)
				number = 1;
			cmd_exec();
			backward((int) number, 1, 0);
			break;

		case A_FF_SCREEN:
			/* Force forward one screen. */
			if (number <= 0)
				number = get_swindow();
			cmd_exec();
			if (show_attn == OPT_ONPLUS)
				set_attnpos(bottompos);
			forward((int) number, 1, 0);
			break;

		case A_F_FOREVER:
			/* Forward forever, ignoring EOF. */
			if (ch_getflags() & CH_HELPFILE)
				break;
			cmd_exec();
			jump_forw();
			ignore_eoi = 1;
			hit_eof = 0;
			while (!sigs) {
				make_display();
				forward(1, 0, 0);
			}
			ignore_eoi = 0;
			/* This gets us back in "F mode" after processing
			 * a non-abort signal (e.g. window-change). */
			if (sigs && !ABORT_SIGS())
				newaction = A_F_FOREVER;
			break;

		case A_F_SCROLL:
			/* Forward N lines
			 * (default same as last 'd' or 'u' command). */
			if (number > 0)
				wscroll = (int) number;
			cmd_exec();
			if (show_attn == OPT_ONPLUS)
				set_attnpos(bottompos);
			forward(wscroll, 0, 0);
			break;

		case A_B_SCROLL:
			/* Forward N lines
			 * (default same as last 'd' or 'u' command). */
			if (number > 0)
				wscroll = (int) number;
			cmd_exec();
			backward(wscroll, 0, 0);
			break;

		case A_FREPAINT:
			/* Flush buffers, then repaint screen.
			 * Don't flush the buffers on a pipe! */
			clear_buffers();
			/* FALLTHRU */
		case A_REPAINT:
			/* Repaint screen. */
			cmd_exec();
			repaint();
			break;

		case A_GOLINE:
			/* Go to line N, default beginning of file. */
			if (number <= 0)
				number = 1;
			cmd_exec();
			jump_back(number);
			break;

		case A_PERCENT:
			/* Go to a specified percentage into the file. */
			if (number < 0) {
				number = 0;
				fraction = 0;
			}
			if (number > 100) {
				number = 100;
				fraction = 0;
			}
			cmd_exec();
			jump_percent((int) number);
			break;

		case A_GOEND:
			/* Go to line N, default end of file. */
			cmd_exec();
			if (number <= 0)
				jump_forw();
			else
				jump_back(number);
			break;

		case A_GOPOS:
			/* Go to a specified byte position in the file. */
			cmd_exec();
			if (number < 0)
				number = 0;
			jump_line_loc((POSITION) number, jump_sline);
			break;

		case A_STAT:
			/* Print file name, etc. */
			if (ch_getflags() & CH_HELPFILE)
				break;
			cmd_exec();
			parg.p_string = pr_expand(eqproto, 0);
			error("%s", &parg);
			break;

		case A_VERSION:
			/* Print version number, without the "@(#)". */
			cmd_exec();
			dispversion();
			break;

		case A_QUIT:
			/* Exit. */
			if (curr_ifile != NULL_IFILE &&
			    ch_getflags() & CH_HELPFILE) {
				/* Quit while viewing the help file
				 * just means return to viewing the
				 * previous file. */
				hshift = save_hshift;
				if (edit_prev(1) == 0)
					break;
			}
			if (extra != NULL)
				quit(*extra);
			quit(QUIT_OK);

/* Define abbreviation for a commonly used sequence below. */
#define DO_SEARCH() \
			if (number <= 0) number = 1;    \
			mca_search();                   \
			cmd_exec();                     \
			multi_search((char *)NULL, (int) number);


		case A_F_SEARCH:
			/* Search forward for a pattern.
			 * Get the first char of the pattern. */
			search_type = SRCH_FORW;
			if (number <= 0)
				number = 1;
			mca_search();
			c = getcc();
			goto again;

		case A_B_SEARCH:
			/* Search backward for a pattern.
			 * Get the first char of the pattern. */
			search_type = SRCH_BACK;
			if (number <= 0)
				number = 1;
			mca_search();
			c = getcc();
			goto again;

		case A_AGAIN_SEARCH:
			/* Repeat previous search. */
			DO_SEARCH();
			break;

		case A_T_AGAIN_SEARCH:
			/* Repeat previous search, multiple files. */
			search_type |= SRCH_PAST_EOF;
			DO_SEARCH();
			break;

		case A_REVERSE_SEARCH:
			/* Repeat previous search, in reverse direction. */
			save_search_type = search_type;
			search_type = SRCH_REVERSE(search_type);
			DO_SEARCH();
			search_type = save_search_type;
			break;

		case A_T_REVERSE_SEARCH:
			/* Repeat previous search,
			 * multiple files in reverse direction. */
			save_search_type = search_type;
			search_type = SRCH_REVERSE(search_type);
			search_type |= SRCH_PAST_EOF;
			DO_SEARCH();
			search_type = save_search_type;
			break;

		case A_UNDO_SEARCH:
			undo_search();
			break;

		case A_HELP:
			/* Help. */
			if (ch_getflags() & CH_HELPFILE)
				break;
			cmd_exec();
			save_hshift = hshift;
			hshift = 0;
			(void) edit(FAKE_HELPFILE);
			break;

		case A_EXAMINE:
#if EXAMINE
			/* Edit a new file.  Get the filename. */
			if (secure) {
				error("Command not available", NULL_PARG);
				break;
			}
			start_mca(A_EXAMINE, "Examine: ", ml_examine, 0);
			c = getcc();
			goto again;
#else
			error("Command not available", NULL_PARG);
			break;
#endif

		case A_VISUAL:
			/* Invoke an editor on the input file. */
#if EDITOR
			if (secure) {
				error("Command not available", NULL_PARG);
				break;
			}
			if (ch_getflags() & CH_HELPFILE)
				break;
			if (strcmp(get_filename(curr_ifile), "-") == 0) {
				error("Cannot edit standard input", NULL_PARG);
				break;
			}
			if (curr_altfilename != NULL) {
				error("WARNING: This file was viewed via LESSOPEN",
					NULL_PARG);
			}
			start_mca(A_SHELL, "!", ml_shell, 0);
			/* Expand the editor prototype string
			 * and pass it to the system to execute.
			 * (Make sure the screen is displayed so the
			 * expansion of "+%lm" works.) */
			make_display();
			cmd_exec();
			lsystem(pr_expand(editproto, 0), (char*)NULL);
			break;
#else
			error("Command not available", NULL_PARG);
			break;
#endif

		case A_NEXT_FILE:
			/* Examine next file. */
#if TAGS
			if (ntags()) {
				error("No next file", NULL_PARG);
				break;
			}
#endif
			if (number <= 0)
				number = 1;
			if (edit_next((int) number)) {
				if (get_quit_at_eof() && hit_eof &&
				    !(ch_getflags() & CH_HELPFILE))
					quit(QUIT_OK);
				parg.p_string = (number > 1) ? "(N-th) " : "";
				error("No %snext file", &parg);
			}
			break;

		case A_PREV_FILE:
			/* Examine previous file. */
#if TAGS
			if (ntags()) {
				error("No previous file", NULL_PARG);
				break;
			}
#endif
			if (number <= 0)
				number = 1;
			if (edit_prev((int) number)) {
				parg.p_string = (number > 1) ? "(N-th) " : "";
				error("No %sprevious file", &parg);
			}
			break;

		case A_NEXT_TAG:
#if TAGS
			if (number <= 0)
				number = 1;
			tagfile = nexttag((int) number);
			if (tagfile == NULL) {
				error("No next tag", NULL_PARG);
				break;
			}
			if (edit(tagfile) == 0) {
				POSITION pos = tagsearch();
				if (pos != NULL_POSITION)
					jump_loc(pos, jump_sline);
			}
#else
			error("Command not available", NULL_PARG);
#endif
			break;

		case A_PREV_TAG:
#if TAGS
			if (number <= 0)
				number = 1;
			tagfile = prevtag((int) number);
			if (tagfile == NULL) {
				error("No previous tag", NULL_PARG);
				break;
			}
			if (edit(tagfile) == 0) {
				POSITION pos = tagsearch();
				if (pos != NULL_POSITION)
					jump_loc(pos, jump_sline);
			}
#else
			error("Command not available", NULL_PARG);
#endif
			break;

		case A_INDEX_FILE:
			/* Examine a particular file. */
			if (number <= 0)
				number = 1;
			if (edit_index((int) number))
				error("No such file", NULL_PARG);
			break;

		case A_REMOVE_FILE:
			if (ch_getflags() & CH_HELPFILE)
				break;
			c_old_ifile = curr_ifile;
			c_new_ifile = getoff_ifile(curr_ifile);
			if (c_new_ifile == NULL_IFILE) {
				bell();
				break;
			}
			if (edit_ifile(c_new_ifile) != 0) {
				reedit_ifile(c_old_ifile);
				break;
			}
			del_ifile(c_old_ifile);
			break;

		case A_OPT_TOGGLE:
			optflag = OPT_TOGGLE;
			optgetname = FALSE;
			mca_opt_toggle();
			c = getcc();
			goto again;

		case A_DISP_OPTION:
			/* Report a flag setting. */
			optflag = OPT_NO_TOGGLE;
			optgetname = FALSE;
			mca_opt_toggle();
			c = getcc();
			goto again;

		case A_FIRSTCMD:
			/* Set an initial command for new files. */
			start_mca(A_FIRSTCMD, "+", (void*)NULL, 0);
			c = getcc();
			goto again;

		case A_SHELL:
			/* Shell escape. */
#if SHELL_ESCAPE
			if (secure) {
				error("Command not available", NULL_PARG);
				break;
			}
			start_mca(A_SHELL, "!", ml_shell, 0);
			c = getcc();
			goto again;
#else
			error("Command not available", NULL_PARG);
			break;
#endif

		case A_SETMARK:
			/* Set a mark. */
			if (ch_getflags() & CH_HELPFILE)
				break;
			start_mca(A_SETMARK, "mark: ", (void*)NULL, 0);
			c = getcc();
			if (c == erase_char || c == erase2_char ||
			    c == kill_char || c == '\n' || c == '\r')
				break;
			setmark(c);
			break;

		case A_GOMARK:
			/* Go to a mark. */
			start_mca(A_GOMARK, "goto mark: ", (void*)NULL, 0);
			c = getcc();
			if (c == erase_char || c == erase2_char ||
			    c == kill_char || c == '\n' || c == '\r')
				break;
			cmd_exec();
			gomark(c);
			break;

		case A_PIPE:
#if PIPEC
			if (secure) {
				error("Command not available", NULL_PARG);
				break;
			}
			start_mca(A_PIPE, "|mark: ", (void*)NULL, 0);
			c = getcc();
			if (c == erase_char || c == erase2_char || c == kill_char)
				break;
			if (c == '\n' || c == '\r')
				c = '.';
			if (badmark(c))
				break;
			pipec = c;
			start_mca(A_PIPE, "!", ml_shell, 0);
			c = getcc();
			goto again;
#else
			error("Command not available", NULL_PARG);
			break;
#endif

		case A_B_BRACKET:
		case A_F_BRACKET:
			start_mca(action, "Brackets: ", (void*)NULL, 0);
			c = getcc();
			goto again;

		case A_LSHIFT:
			if (number > 0)
				shift_count = number;
			else
				number = (shift_count > 0) ?
					shift_count : sc_width / 2;
			if (number > hshift)
				number = hshift;
			hshift -= number;
			screen_trashed = 1;
			break;

		case A_RSHIFT:
			if (number > 0)
				shift_count = number;
			else
				number = (shift_count > 0) ?
					shift_count : sc_width / 2;
			hshift += number;
			screen_trashed = 1;
			break;

		case A_PREFIX:
			/* The command is incomplete (more chars are needed).
			 * Display the current char, so the user knows
			 * what's going on, and get another character. */
			if (mca != A_PREFIX) {
				cmd_reset();
				start_mca(A_PREFIX, " ", (void*)NULL,
					CF_QUIT_ON_ERASE);
				(void) cmd_char(c);
			}
			c = getcc();
			goto again;

		case A_NOACTION:
			break;

		default:
			bell();
			break;
		}
	}
}

/* Handler for the -b option. */
static void
opt_b(int type, const char *s)
{
	(void)&s;
	switch (type) {
	case INIT:
	case TOGGLE:
		/* Set the new number of buffers. */
		ch_setbufspace(bufspace);
		break;
	case QUERY:
		break;
	}
}

/* Handlers for -j option. */
static void
opt_j(int type, const char *s)
{
	PARG parg;
	char buf[16];
	int len;
	int err;

	switch (type) {
	case INIT:
	case TOGGLE:
		if (*s == '.') {
			s++;
			jump_sline_fraction = getfraction(&s, "j", &err);
			if (err)
				error("Invalid line fraction", NULL_PARG);
			else
				calc_jump_sline();
		} else {
			int sline = getnum(&s, "j", &err);
			if (err)
				error("Invalid line number", NULL_PARG);
			else {
				jump_sline = sline;
				jump_sline_fraction = -1;
			}
		}
		break;
	case QUERY:
		if (jump_sline_fraction < 0) {
			parg.p_int =  jump_sline;
			error("Position target at screen line %d", &parg);
		} else {

			sprintf(buf, ".%06ld", jump_sline_fraction);
			len = strlen(buf);
			while (len > 2 && buf[len-1] == '0')
				len--;
			buf[len] = '\0';
			parg.p_string = buf;
			error("Position target at screen position %s", &parg);
		}
		break;
	}
}

#if USERFILE
static void
opt_k(int type, const char *s)
{
	PARG parg;

	switch (type) {
	case INIT:
		if (lesskey(s, 0)) {
			parg.p_string = s;
			error("Cannot use lesskey file \"%s\"", &parg);
		}
		break;
	}
}
#endif

/* Handlers for -l option. */
static void
opt_l(int type, const char *s)
{
	int err;
	int n;
	const char *t;

	switch (type) {
	case INIT:
		t = s;
		n = getnum(&t, "l", &err);
		if (err || n <= 0) {
			error("Line number is required after -l", NULL_PARG);
			return;
		}
		plusoption = TRUE;
		ungetsc(s);
		break;
	}
}

#if LOGFILE
/* Handler for -o option. */
static void
opt_o(int type, const char *s)
{
	PARG parg;

	if (secure) {
		error("log file support is not available", NULL_PARG);
		return;
	}
	switch (type) {
	case INIT:
		namelogfile = s;
		break;
	case TOGGLE:
		if (ch_getflags() & CH_CANSEEK) {
			error("Input is not a pipe", NULL_PARG);
			return;
		}
		if (logfile >= 0) {
			error("Log file is already in use", NULL_PARG);
			return;
		}
		s = skipsp(s);
		namelogfile = lglob(s);
		use_logfile(namelogfile);
		sync_logfile();
		break;
	case QUERY:
		if (logfile < 0)
			error("No log file", NULL_PARG);
		else {
			parg.p_string = namelogfile;
			error("Log file \"%s\"", &parg);
		}
		break;
	}
}

/* Handler for -O option. */
static void
opt__O(int type, const char *s)
{
	force_logfile = TRUE;
	opt_o(type, s);
}
#endif

/* Table of all options and their semantics.
 *
 * For BOOL and TRIPLE options, odesc[0], odesc[1], odesc[2] are
 * the description of the option when set to 0, 1 or 2, respectively.
 * For NUMBER options, odesc[0] is the prompt to use when entering
 * a new value, and odesc[1] is the description, which should contain
 * one %d which is replaced by the value of the number.
 * For STRING options, odesc[0] is the prompt to use when entering
 * a new value, and odesc[1], if not NULL, is the set of characters
 * that are valid in the string. */
static const struct loption option[] =
{
	{ 'a', "search-skip-screen", BOOL, OPT_OFF, &how_search, NULL,
		{
			"Search includes displayed screen",
			"Search skips displayed screen",
			NULL
		}
	},

	{ 'b', "buffers", NUMBER|INIT_HANDLER, 64, &bufspace, opt_b,
		{
			"Max buffer space per file (K): ",
			"Max buffer space per file: %dK",
			NULL
		}
	},
	{ 'B', "auto-buffers", BOOL, OPT_ON, &autobuf, NULL,
		{
			"Don't automatically allocate buffers",
			"Automatically allocate buffers when needed",
			NULL
		}
	},
	{ 'c', "clear-screen", TRIPLE, OPT_OFF, &top_scroll, NULL,
		{
			"Repaint by scrolling from bottom of screen",
			"Repaint by painting from top of screen",
			"Repaint by painting from top of screen"
		}
	},
	{ 'e', "quit-at-eof", TRIPLE, OPT_OFF, &quit_at_eof, NULL,
		{
			"Don't quit at end-of-file",
			"Quit at end-of-file",
			"Quit immediately at end-of-file"
		}
	},
	{ 'f', "force", BOOL, OPT_OFF, &force_open, NULL,
		{
			"Open only regular files",
			"Open even non-regular files",
			NULL
		}
	},
	{ 'F', "quit-if-one-screen", BOOL, OPT_OFF, &quit_if_one_screen, NULL,
		{
			"Don't quit if end-of-file on first screen",
			"Quit if end-of-file on first screen",
			NULL
		}
	},
#if HILITE_SEARCH
	{ 'g', "hilite-search", TRIPLE|HL_REPAINT, OPT_ONPLUS, &hilite_search, NULL,
		{
			"Don't highlight search matches",
			"Highlight matches for previous search only",
			"Highlight all matches for previous search pattern",
		}
	},
#endif
	{ 'h', "max-back-scroll", NUMBER, -1, &back_scroll, NULL,
		{
			"Backwards scroll limit: ",
			"Backwards scroll limit is %d lines",
			NULL
		}
	},
	{ 'i', "ignore-case", TRIPLE|HL_REPAINT, OPT_OFF, &caseless, opt_i,
		{
			"Case is significant in searches",
			"Ignore case in searches",
			"Ignore case in searches and in patterns"
		}
	},
	{ 'j', "jump-target", STRING, 0, NULL, opt_j,
		{
			"Target line: ",
			"0123456789.",
			NULL
		}
	},
	{ 'J', "status-column", BOOL|REPAINT, OPT_OFF, &status_col, NULL,
		{
			"Don't display a status column",
			"Display a status column",
			NULL
		}
	},
#if USERFILE
	{ 'k', "lesskey-file",  STRING|NO_TOGGLE|NO_QUERY, 0, NULL, opt_k,
		{ NULL, NULL, NULL }
	},
#endif
	{ 'K', "quit-on-intr", BOOL, OPT_OFF, &quit_on_intr, NULL,
		{
			"Interrupt (ctrl-C) returns to prompt",
			"Interrupt (ctrl-C) exits less",
			NULL
		}
	},
	{ 'l', NULL, STRING|NO_TOGGLE|NO_QUERY, 0, NULL, opt_l,
		{ NULL, NULL, NULL }
	},
	{ 'L', "no-lessopen", BOOL, OPT_ON, &use_lessopen, NULL,
		{
			"Don't use the LESSOPEN filter",
			"Use the LESSOPEN filter",
			NULL
		}
	},
	{ 'm', "long-prompt", TRIPLE, OPT_OFF, &pr_type, NULL,
		{
			"Short prompt",
			"Medium prompt",
			"Long prompt"
		}
	},
	{ 'n', "line-numbers", TRIPLE|REPAINT, OPT_ON, &linenums, NULL,
		{
			"Don't use line numbers",
			"Use line numbers",
			"Constantly display line numbers"
		}
	},
#if LOGFILE
	{ 'o', "log-file", STRING, 0, NULL, opt_o,
		{ "log file: ", NULL, NULL }
	},
	{ 'O', "LOG-FILE", STRING, 0, NULL, opt__O,
		{ "Log file: ", NULL, NULL }
	},
#endif
	{ 'p', "pattern", STRING|NO_TOGGLE|NO_QUERY, 0, NULL, opt_p,
		{ NULL, NULL, NULL }
	},
	{ 'P', "prompt", STRING, 0, NULL, opt__P,
		{ "prompt: ", NULL, NULL }
	},
	{ 'q', "quiet", TRIPLE, OPT_OFF, &quiet, NULL,
		{
			"Ring the bell for errors AND at eof/bof",
			"Ring the bell for errors but not at eof/bof",
			"Never ring the bell"
		}
	},
	{ 'r', "raw-control-chars", TRIPLE|REPAINT, OPT_OFF, &ctldisp, NULL,
		{
			"Display control characters as ^X",
			"Display control characters directly",
			"Display control characters directly, processing ANSI sequences"
		}
	},
	{ 's', "squeeze-blank-lines", BOOL|REPAINT, OPT_OFF, &squeeze, NULL,
		{
			"Display all blank lines",
			"Squeeze multiple blank lines",
			NULL
		}
	},
	{ 'S', "chop-long-lines", BOOL|REPAINT, OPT_OFF, &chopline, NULL,
		{
			"Fold long lines",
			"Chop long lines",
			NULL
		}
	},
#if TAGS
	{ 't', "tag", STRING|NO_QUERY, 0, NULL, opt_t,
		{ "tag: ", NULL, NULL }
	},
	{ 'T', "tag-file", STRING, 0, NULL, opt__T,
		{ "tags file: ", NULL, NULL }
	},
#endif
	{ 'u', "underline-special", TRIPLE|REPAINT, OPT_OFF, &bs_mode, NULL,
		{
			"Display underlined text in underline mode",
			"Backspaces cause overstrike",
			"Print backspace as ^H"
		}
	},
	{ 'V', "version", NOVAR, 0, NULL, opt__V,
		{ NULL, NULL, NULL }
	},
	{ 'w', "hilite-unread", TRIPLE|REPAINT, OPT_OFF, &show_attn, NULL,
		{
			"Don't highlight first unread line",
			"Highlight first unread line after forward-screen",
			"Highlight first unread line after any forward movement",
		}
	},
	{ 'x', "tabs", STRING|REPAINT, 0, NULL, opt_x,
		{
			"Tab stops: ",
			"0123456789,",
			NULL
		}
	},
	{ 'y', "max-forw-scroll", NUMBER, -1, &forw_scroll, NULL,
		{
			"Forward scroll limit: ",
			"Forward scroll limit is %d lines",
			NULL
		}
	},
	{ 'z', "window", NUMBER, -1, &swindow, NULL,
		{
			"Scroll window size: ",
			"Scroll window size is %d lines",
			NULL
		}
	},
	{ '"', "quotes", STRING, 0, NULL, opt_quote,
		{ "quotes: ", NULL, NULL }
	},
	{ '~', "tilde", BOOL|REPAINT, OPT_ON, &twiddle, NULL,
		{
			"Don't show tildes after end of file",
			"Show tildes after end of file",
			NULL
		}
	},
	{ '?', "help", NOVAR, 0, NULL, opt_query,
		{ NULL, NULL, NULL }
	},
	{ '#', "shift", NUMBER, 0, &shift_count, NULL,
		{
			"Horizontal shift: ",
			"Horizontal shift %d positions",
			NULL
		}
	},
	{ '.', "follow-name", BOOL, FOLLOW_DESC, &follow_mode, NULL,
		{
			"F command Follows file descriptor",
			"F command Follows file name",
			NULL
		}
	},
	{ '\0', NULL, NOVAR, 0, NULL, NULL, { NULL, NULL, NULL } }
};

/* Process command line options.
 *
 * Each option is a single letter which controls a program variable.
 * The options have defaults which may be changed via
 * the command line option, toggled via the "-" command,
 * or queried via the "_" command. */

/* Scan an argument (either from the command line or from the
 * LESS environment variable) and process it. */
static void
scan_option(char *s)
{
	const struct loption *o;
	int optc;
	const char *optname;
	const char *printopt;
	const char *printopt_plus;
	char *str;
	int set_default;
	int lc;
	int err;
	PARG parg;

	if (s == NULL)
		return;

	/* If we have a pending option which requires an argument,
	 * handle it now.
	 * This happens if the previous option was, for example, "-P"
	 * without a following string.  In that case, the current
	 * option is simply the argument for the previous option. */
	if (pendopt != NULL) {
		switch (pendopt->otype & OTYPE) {
		case STRING:
			(*pendopt->ofunc)(INIT, s);
			break;
		case NUMBER:
			printopt = propt(pendopt->oletter);
			*(pendopt->ovar) = getnum((const char **)&s, printopt, (int*)NULL);
			break;
		}
		pendopt = NULL;
		return;
	}

	set_default = FALSE;
	optname = NULL;

	while (*s != '\0') {
		/* Check some special cases first. */
		switch (optc = *s++) {
		case ' ':
		case '\t':
		case END_OPTION_STRING:
			continue;
		case '-':
			/* "--" indicates an option name instead of a letter. */
			if (*s == '-') {
				optname = ++s;
				break;
			}
			/* "-+" means set these options back to their defaults.
			 * (They may have been set otherwise by previous
			 * options.) */
			set_default = (*s == '+');
			if (set_default)
				s++;
			continue;
		case '+':
			/* An option prefixed by a "+" is ungotten, so
			 * that it is interpreted as less commands
			 * processed at the start of the first input file.
			 * "++" means process the commands at the start of
			 * EVERY input file. */
			plusoption = TRUE;
			printopt_plus = propt('+');
			s = optstring(s, &str, printopt_plus, NULL);
			if (*str == '+')
				every_first_cmd = save(++str);
			else
				ungetsc(str);
			continue;
		case '0':  case '1':  case '2':  case '3':  case '4':
		case '5':  case '6':  case '7':  case '8':  case '9':
			/* Special "more" compatibility form "-<number>"
			 * instead of -z<number> to set the scrolling
			 * window size. */
			s--;
			optc = 'z';
			break;
		case 'n':
			if (less_is_more)
				optc = 'z';
			break;
		}

		/* Not a special case.
		 * Look up the option letter in the option table. */
		err = 0;
		if (optname == NULL) {
			printopt = propt(optc);
			lc = ASCII_IS_LOWER(optc);
			o = findopt(optc);
		} else {
			printopt = optname;
			lc = ASCII_IS_LOWER(optname[0]);
			o = findopt_name(&optname, NULL, &err);
			s = (char *)optname;
			optname = NULL;
			if (*s == '\0' || *s == ' ') {
				/* The option name matches exactly. */
				;
			} else if (*s == '=') {
				/* The option name is followed by "=value". */
				if (o != NULL &&
				    (o->otype & OTYPE) != STRING &&
				    (o->otype & OTYPE) != NUMBER) {
					parg.p_string = printopt;
					error("The %s option should not be followed by =",
						&parg);
					quit(QUIT_ERROR);
				}
				s++;
			} else {
				/* The specified name is longer than the
				 * real option name. */
				o = NULL;
			}
		}
		if (o == NULL) {
			parg.p_string = printopt;
			if (err == OPT_AMBIG)
				error("%s is an ambiguous abbreviation (\"less --help\" for help)",
					&parg);
			else
				error("There is no %s option (\"less --help\" for help)",
					&parg);
			quit(QUIT_ERROR);
		}

		str = NULL;
		switch (o->otype & OTYPE) {
		case BOOL:
			if (set_default)
				*(o->ovar) = o->odefault;
			else
				*(o->ovar) = ! o->odefault;
			break;
		case TRIPLE:
			if (set_default)
				*(o->ovar) = o->odefault;
			else
				*(o->ovar) = flip_triple(o->odefault, lc);
			break;
		case STRING:
			if (*s == '\0') {
				/* Set pendopt and return.
				 * We will get the string next time
				 * scan_option is called. */
				pendopt = o;
				return;
			}
			/* Don't do anything here.
			 * All processing of STRING options is done by
			 * the handling function. */
			while (*s == ' ')
				s++;
			s = optstring(s, &str, printopt, o->odesc[1]);
			break;
		case NUMBER:
			if (*s == '\0') {
				pendopt = o;
				return;
			}
			*(o->ovar) = getnum((const char **)&s, printopt, (int*)NULL);
			break;
		}
		/* If the option has a handling function, call it. */
		if (o->ofunc != NULL)
			(*o->ofunc)(INIT, str);
	}
}

static int
is_optchar(char c)
{
	if (ASCII_IS_UPPER(c))
		return 1;
	if (ASCII_IS_LOWER(c))
		return 1;
	if (c == '-')
		return 1;
	return 0;
}

/* Find an option in the option table, given its option name.
 * p_optname is the (possibly partial) name to look for, and
 * is updated to point after the matched name.
 * p_oname if non-NULL is set to point to the full option name. */
static const struct loption *
findopt_name(const char **ptr_optname, const char **p_oname, int *p_err)
{
	const char *optname = *ptr_optname;
	const struct loption *o;
	const char *oname;
	int len;
	int uppercase;
	const struct loption *maxo = NULL;
	const char *maxoname = NULL;
	int maxlen = 0;
	int ambig = 0;
	int exact = 0;

	/* Check all options. */
	for (o = option;  o->oletter != '\0';  o++) {
		/* Check all names for this option. */
		oname = o->onames;
	next_quiet:
		/* Try normal match first (uppercase == 0),
		 * then, then if it's a TRIPLE option,
		 * try uppercase match (uppercase == 1). */
		for (uppercase = 0;  uppercase <= 1;  uppercase++) {
			len = sprefix(optname, oname, uppercase);
			if (len <= 0 || is_optchar(optname[len])) {
				/* We didn't use all of the option name. */
				continue;
			}
			if (!exact && len == maxlen)
				/* Already had a partial match,
				 * and now there's another one that
				 * matches the same length. */
				ambig = 1;
			else if (len > maxlen) {
				/* Found a better match than the one we had. */
				maxo = o;
				maxoname = oname;
				maxlen = len;
				ambig = 0;
				exact = (len == (int)strlen(oname));
			}
			if (!(o->otype & TRIPLE))
				break;
		}
		if(oname != NULL && strcmp(oname, "quiet") == 0) {
			oname = "silent";
			goto next_quiet;
		}
	}
	if (ambig) {
		/* Name matched more than one option. */
		if (p_err != NULL)
			*p_err = OPT_AMBIG;
		return NULL;
	}
	*ptr_optname = optname + maxlen;
	if (p_oname != NULL)
		*p_oname = maxoname == NULL ? NULL : maxoname;
	return maxo;
}

/* Initialize each option to its default value. */
static void
init_option(void)
{
	const struct loption *o;
	const char *p;

	p = lgetenv("LESS_IS_MORE");
	if (p != NULL && *p != '\0')
		less_is_more = 1;

	for (o = option;  o->oletter != '\0';  o++) {
		/* Set each variable to its default. */
		if (o->ovar != NULL)
			*(o->ovar) = o->odefault;
		if (o->otype & INIT_HANDLER)
			(*(o->ofunc))(INIT, (char *) NULL);
	}
}

/* Find an option in the option table, given its option letter. */
static const struct loption *
findopt(int c)
{
	const struct loption *o;

	for (o = option;  o->oletter != '\0';  o++) {
		if (o->oletter == c)
			return o;
		if ((o->otype & TRIPLE) && ASCII_TO_UPPER(o->oletter) == c)
			return o;
	}
	return NULL;
}

/* Toggle command line flags from within the program.
 * Used by the "-" and "_" commands.
 * how_toggle may be:
 *      OPT_NO_TOGGLE   just report the current setting, without changing it.
 *      OPT_TOGGLE      invert the current setting
 *      OPT_UNSET       set to the default value
 *      OPT_SET         set to the inverse of the default value */
static void
toggle_option(int c, char *s, int how_toggle)
{
	const struct loption *o;
	int num;
	int no_prompt;
	int err;
	PARG parg;

	no_prompt = (how_toggle & OPT_NO_PROMPT);
	how_toggle &= ~OPT_NO_PROMPT;

	/* Look up the option letter in the option table. */
	o = findopt(c);
	if (o == NULL) {
		parg.p_string = propt(c);
		error("There is no %s option", &parg);
		return;
	}

	if (how_toggle == OPT_TOGGLE && (o->otype & NO_TOGGLE)) {
		parg.p_string = propt(c);
		error("Cannot change the %s option", &parg);
		return;
	}

	if (how_toggle == OPT_NO_TOGGLE && (o->otype & NO_QUERY)) {
		parg.p_string = propt(c);
		error("Cannot query the %s option", &parg);
		return;
	}

	/* Check for something which appears to be a do_toggle
	 * (because the "-" command was used), but really is not.
	 * This could be a string option with no string, or
	 * a number option with no number. */
	switch (o->otype & OTYPE) {
	case STRING:
	case NUMBER:
		if (how_toggle == OPT_TOGGLE && *s == '\0')
			how_toggle = OPT_NO_TOGGLE;
		break;
	}

#if HILITE_SEARCH
	if (how_toggle != OPT_NO_TOGGLE && (o->otype & HL_REPAINT))
		repaint_hilite(0);
#endif

	/* Now actually toggle (change) the variable. */
	if (how_toggle != OPT_NO_TOGGLE) {
		switch (o->otype & OTYPE) {
		case BOOL:
			/* Boolean. */
			switch (how_toggle) {
			case OPT_TOGGLE:
				*(o->ovar) = ! *(o->ovar);
				break;
			case OPT_UNSET:
				*(o->ovar) = o->odefault;
				break;
			case OPT_SET:
				*(o->ovar) = ! o->odefault;
				break;
			}
			break;
		case TRIPLE:
			/* Triple:
			 *      If user gave the lower case letter, then switch
			 *      to 1 unless already 1, in which case make it 0.
			 *      If user gave the upper case letter, then switch
			 *      to 2 unless already 2, in which case make it 0. */
			switch (how_toggle) {
			case OPT_TOGGLE:
				*(o->ovar) = flip_triple(*(o->ovar),
						ASCII_IS_LOWER(c));
				break;
			case OPT_UNSET:
				*(o->ovar) = o->odefault;
				break;
			case OPT_SET:
				*(o->ovar) = flip_triple(o->odefault,
						ASCII_IS_LOWER(c));
				break;
			}
			break;
		case STRING:
			/* String: don't do anything here.
			 *      The handling function will do everything. */
			switch (how_toggle) {
			case OPT_SET:
			case OPT_UNSET:
				error("Cannot use \"-+\" or \"--\" for a string option",
					NULL_PARG);
				return;
			}
			break;
		case NUMBER:
			/* Number: set the variable to the given number. */
			switch (how_toggle) {
			case OPT_TOGGLE:
				num = getnum((const char **)&s, NULL, &err);
				if (!err)
					*(o->ovar) = num;
				break;
			case OPT_UNSET:
				*(o->ovar) = o->odefault;
				break;
			case OPT_SET:
				error("Can't use \"-!\" for a numeric option",
					NULL_PARG);
				return;
			}
			break;
		}
	}

	/* Call the handling function for any special action
	 * specific to this option. */
	if (o->ofunc != NULL)
		(*o->ofunc)((how_toggle==OPT_NO_TOGGLE) ? QUERY : TOGGLE, s);

#if HILITE_SEARCH
	if (how_toggle != OPT_NO_TOGGLE && (o->otype & HL_REPAINT))
		chg_hilite();
#endif

	if (!no_prompt) {
		/* Print a message describing the new setting. */
		switch (o->otype & OTYPE) {
		case BOOL:
		case TRIPLE:
			/* Print the odesc message. */
			error(o->odesc[*(o->ovar)], NULL_PARG);
			break;
		case NUMBER:
			/* The message is in odesc[1] and has a %d for
			 * the value of the variable. */
			parg.p_int = *(o->ovar);
			error(o->odesc[1], &parg);
			break;
		case STRING:
			/* Message was already printed by the handling function. */
			break;
		}
	}

	if (how_toggle != OPT_NO_TOGGLE && (o->otype & REPAINT))
		screen_trashed = TRUE;
}


/* Entry point. */
int less_main(void)
{
	IFILE ifile;
	const char *s;
	char **argv = Gargv;
	int argc = Gargc;

	argv++;
	argc--;

	ob = obuf;
	ch_ungotchar = -1;
	binattr = AT_STANDOUT;
	start_attnpos = NULL_POSITION;
	end_attnpos = NULL_POSITION;
	jump_sline_fraction = -1;
	openquote = '"';
	closequote = '"';
	first_time = 1;
	hilite_anchor.hl_startpos = hilite_anchor.hl_endpos = NULL_POSITION;
	ifile_anchor.h_next = ifile_anchor.h_prev = &ifile_anchor;
	ifile_anchor.h_scrpos.pos = NULL_POSITION;
	maxbufs = -1;
	ntabstops = 1;
	tabdefault = 8;
#if CMD_HISTORY
	mlist_search.next = mlist_search.prev = mlist_search.curr_mp = &mlist_search;
	mlist_examine.next = mlist_examine.prev = mlist_examine.curr_mp = &mlist_examine;
#if SHELL_ESCAPE || PIPEC
	mlist_shell.next = mlist_shell.prev = mlist_shell.curr_mp = &mlist_shell;
#endif
#endif /* CMD_HISTORY */

	secure = 0;
	s = lgetenv("LESSSECURE");
	if (s != NULL && *s != '\0')
		secure = 1;

	/* Process command line arguments and LESS environment arguments.
	 * Command line arguments override environment arguments. */
	is_tty = isatty(1);
	/* standart input may be not terminal */
	tty = open(bb_dev_tty, O_RDONLY);
	if(tty < 0)
		tty = 2;
	get_term();
	init_cmds();

	/* Initialize charset data structures. */
#if HAVE_LOCALE
	setlocale(LC_ALL, "");
#endif
	set_charset();

	s = lgetenv("LESSBINFMT");
	setbinfmt(s, &binfmt, "*s<%02X>");

	s = lgetenv("LESSUTFBINFMT");
	setbinfmt(s, &utfbinfmt, "<U+%04lX>");

	/* Initialize from environment variables. */
	end_ansi_chars = lgetenv("LESSANSIENDCHARS");
	if (end_ansi_chars == NULL || *end_ansi_chars == '\0')
		end_ansi_chars = "m";

	mid_ansi_chars = lgetenv("LESSANSIMIDCHARS");
	if (mid_ansi_chars == NULL || *mid_ansi_chars == '\0')
		mid_ansi_chars = "0123456789;[?!\"'#%()*+ ";

	linebuf = emalloc(LINEBUF_SIZE);
	attr = emalloc(LINEBUF_SIZE);
	size_linebuf = LINEBUF_SIZE;

	init_cmdhist();
	init_option();

	/* If the name of the executable program is "more",
	 * act like LESS_IS_MORE is set. */
	if (strcmp(bb_applet_name, "more") == 0)
		less_is_more = 1;

	/* Initialize the prompt prototype strings. */
	prproto[0] = save(s_proto);
	prproto[1] = save(less_is_more ? more_proto : m_proto);
	prproto[2] = save(M_proto);
	eqproto = save(e_proto);
	hproto = save(h_proto);
	wproto = save(w_proto);

	s = lgetenv(less_is_more ? "MORE" : "LESS");
	if (s != NULL)
		scan_option(save(s));

#define isoptstring(s)  (((s)[0] == '-' || (s)[0] == '+') && (s)[1] != '\0')
	while (argc > 0 && (isoptstring(*argv) || isoptpending())) {
		s = *argv++;
		argc--;
		if (strcmp(s, "--") == 0)
			break;
		scan_option(save(s));
	}
#undef isoptstring

	if (isoptpending()) {
		/* Last command line option was a flag requiring a
		 * following string, but there was no following string. */
		nopendopt();
		quit(QUIT_OK);
	}

	if (less_is_more && get_quit_at_eof())
		quit_if_one_screen = TRUE;

#if EDITOR
	editor = lgetenv("VISUAL");
	if (editor == NULL || *editor == '\0') {
		editor = lgetenv("EDITOR");
		if (editor == NULL || *editor == '\0')
			editor = EDIT_PGM;
	}
	editproto = lgetenv("LESSEDIT");
	if (editproto == NULL || *editproto == '\0')
		editproto = "%E ?lm+%lm. %f";
#endif

	/* Call get_ifile with all the command line filenames
	 * to "register" them with the ifile system. */
	ifile = NULL_IFILE;
	if (dohelp)
		ifile = get_ifile(FAKE_HELPFILE, ifile);
	while (argc-- > 0) {
		char *filename;
		filename = shell_quote(*argv);
		if (filename == NULL)
			filename = *argv;
		argv++;
		(void) get_ifile(filename, ifile);
		ifile = prev_ifile(NULL_IFILE);
	}
	/* Set up terminal, etc. */
	if (!is_tty) {
		/* Output is not a tty.
		 * Just copy the input file(s) to output. */
		if (nifile() == 0) {
			if (edit_stdin() == 0)
				cat_file();
		} else if (edit_first() == 0) {
			do {
				cat_file();
			} while (edit_next(1) == 0);
		}
		quit(QUIT_OK);
	}

	/* Initialize the mark table to show no marks are set. */
	{
		int i;

		for (i = 0;  i < NMARKS;  i++)
			marks[i].m_scrpos.pos = NULL_POSITION;
	}
	raw_mode(1);
	init_signals(1);

	/* Select the first file to examine. */
#if TAGS
	if (tagoption != NULL || strcmp(tags, "-") == 0) {
		/* A -t option was given.
		 * Verify that no filenames were also given.
		 * Edit the file selected by the "tags" search,
		 * and search for the proper line in the file. */
		if (nifile() > 0) {
			error("No filenames allowed with -t option", NULL_PARG);
			quit(QUIT_ERROR);
		}
		findtag(tagoption);
		if (edit_tagfile())  /* Edit file which contains the tag */
			quit(QUIT_ERROR);
		/* Search for the line which contains the tag.
		 * Set up initial_scrpos so we display that line. */
		initial_scrpos.pos = tagsearch();
		if (initial_scrpos.pos == NULL_POSITION)
			quit(QUIT_ERROR);
		initial_scrpos.ln = jump_sline;
	} else
#endif
	if (nifile() == 0) {
		if (edit_stdin())  /* Edit standard input */
			quit(QUIT_ERROR);
	} else {
		if (edit_first())  /* Edit first valid file in cmd line */
			quit(QUIT_ERROR);
	}

	init();
	commands();
	quit(QUIT_OK);
	return 0;
}
#endif /* APPLET_USAGE */
