/* MiniUPnP Project
 * http://miniupnp.free.fr/ or http://miniupnp.tuxfamily.org/
 * (c) 2006-2016 Thomas Bernard
 *
 * PCP code Copyright (c) 2013 by Cisco Systems, Inc.
 * All rights reserved.
 *
 * busybox port by Vladimir Oleynik (c) 2016 <dzo@simtreas.ru>
 *

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/
#ifdef APPLET_USAGE
#define miniupnpd_trivial_usage \
	"[-f config_file] [-i ext_ifname] [-o ext_ip] [-a listening_ip] [-p port] [-d] [-U] [-S] [-N]\n" \
	"[-u uuid] [-s serial] [-m model_number] [-t notify_interval] [-P pid_filename] [-W down -B up]\n" \
	"[-w url] [-r clean_ruleset_interval] [-A \"permission rule\"] [-b BOOTID]"
#define miniupnpd_full_usage \
	"The miniUPnP daemon is an UPnP IGD (internet gateway device).\n" \
	"\tNotify interval is in seconds. Default is 30 seconds.\n" \
	"\t-S sets \"secure\" mode : clients can only add mappings to their own ip\n" \
	"\t-U causes miniupnpd to report system uptime instead of daemon uptime.\n" \
	"\t-N enables NAT-PMP functionality.\n" \
	"\t-W with -B sets bitrates reported by daemon in bits per second.\n" \
	"\t-w sets the presentation url. Default is http address on port 80\n" \
	"\t-A use following syntax for permission rules :\n" \
	"\t  (allow|deny) (external port range) ip/mask (internal port range)\n" \
	"\texamples :\n" \
	"\t  \"allow 1024-65535 192.168.1.0/24 1024-65535\"\n" \
	"\t  \"deny 0-65535 0.0.0.0/0 0-65535\"\n" \
	"\t-b sets the value of BOOTID.UPNP.ORG SSDP header\n" \
	"\t-h prints this help and quits."
#else

#define USE_IFACEWATCHER 1
#define USE_NETFILTER 1
/* Enable IGD2 "Port Triggering" as defined in Section 2.5.16
 * figure 2.2 in UPnP-gw-WANIPConnection-v2-Service.pdf */
#define ENABLE_PORT_TRIGGERING

#define OS_NAME         "Linux"
#define OS_VERSION      "Linux/3.10.17"
#define OS_URL          "http://www.kernel.org/"

#undef DEBUG

/* The NAT-PMP specification which can be found at the url :
 * http://files.dns-sd.org/draft-cheshire-nat-pmp.txt
 * draft version 3 of April 2008
 * define 5351 as listening port for the gateway,
 * and the 224.0.0.1 port 5350 as the local link
 * multicast address for address change announces.
 * Previous versions of the specification defined 5351
 * as the port for address change announces. */
#define NATPMP_PORT (5351)
#define NATPMP_NOTIF_PORT       (5350)
#define NATPMP_NOTIF_ADDR       ("224.0.0.1")

/* syslog facility to be used by miniupnpd */
#define LOG_MINIUPNPD            LOG_DAEMON

/* Comment the following line to disable PCP operations */
#define ENABLE_PCP

/* Define one or none of the two following macros in order to make some
 * clients happy. It will change the XML Root Description of the IGD.
 * Enabling the Layer3Forwarding Service seems to be the more compatible
 * option. */
#define ENABLE_L3F_SERVICE

/* define ADVERTISE_WANPPPCONN to allow buggy Control Point to use
 * WANPPPConnection instead of WANIPConnection. */
#define ADVERTISE_WANPPPCONN

#define HAVE_IP_MREQN

/* UPnP Events support. Working well enough to be enabled by default.
 * It can be disabled to save a few bytes. */
#define ENABLE_EVENTS

/* If SSDP_RESPOND_SAME_VERSION is defined, the M-SEARCH response
 * include the same device version as was contained in the search
 * request. It conforms to UPnP DA v1.1 */
#define SSDP_RESPOND_SAME_VERSION

/* disable reading and parsing of config file (miniupnpd.conf) */
/*#define DISABLE_CONFIG_FILE*/

#if defined(ENABLE_PCP) && ENABLE_FEATURE_IPV6
#define ENABLE_UPNPPINHOLE
#endif

/* maximum lenght of SSDP packets we are generating
 * (reception is done in a 1500byte buffer) */
#define SSDP_PACKET_MAX_LEN 1024

#ifndef __USE_GNU
#define __USE_GNU 1
#endif

#include "busybox.h"
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <inttypes.h>
#include <netinet/in.h>
#include <time.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <ctype.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <linux/un.h>
#include <assert.h>

#ifndef HAVE_GET_FREE_IFADDRS
#undef ENABLE_PCP
#endif

#if ENABLE_FEATURE_IPV6 || defined(ENABLE_PCP)
#include <ifaddrs.h>
#endif

#include "uuid.h"
#include "bb_xopen.h"
#include "bb_inet.h"
#include "bb_sig.h"
#define FOR_MINIUPNPD extern
#include "bb_libiptc.h"
#include "config_read.h"

#if defined(__UCLIBC__) && !ENABLE_FEATURE_IPV6
#if !defined __UCLIBC_HAS_IPV6__ && defined __UCLIBC_STRICT_HEADERS__
/* IPv6 address special for PCP */
#define s6_addr                  __in6_u.__u6_addr8
#define INET6_ADDRSTRLEN 46
#define IN6_IS_ADDR_V4MAPPED(a) \
	((((__const uint32_t *) (a))[0] == 0)                                 \
	&& (((__const uint32_t *) (a))[1] == 0)                              \
	&& (((__const uint32_t *) (a))[2] == htonl (0xffff)))
#define IN6_IS_ADDR_UNSPECIFIED(a) \
	(((__const uint32_t *) (a))[0] == 0                                   \
	&& ((__const uint32_t *) (a))[1] == 0                                \
	&& ((__const uint32_t *) (a))[2] == 0                                \
	&& ((__const uint32_t *) (a))[3] == 0)
#define IN6_ARE_ADDR_EQUAL(a,b) \
	((((__const uint32_t *) (a))[0] == ((__const uint32_t *) (b))[0])     \
	&& (((__const uint32_t *) (a))[1] == ((__const uint32_t *) (b))[1])  \
	&& (((__const uint32_t *) (a))[2] == ((__const uint32_t *) (b))[2])  \
	&& (((__const uint32_t *) (a))[3] == ((__const uint32_t *) (b))[3]))

struct in6_addr
  {
      union
	    {
		    uint8_t __u6_addr8[16];
		    uint16_t __u6_addr16[8];
		    uint32_t __u6_addr32[4];
	} __in6_u;
};
#endif
#endif


#define MINIUPNPD_VERSION "2.0"
#define MINIUPNPD_DATE  "20160512"

#ifndef XSTR
#define XSTR(s) STR(s)
#define STR(s) #s
#endif /* XSTR */

/* UPnP version reported in XML descriptions
 * 1.0 / 1.1 / 2.0 depending on which UDA (UPnP Device Architecture) Version */
#define UPNP_VERSION_MAJOR      1
#define UPNP_VERSION_MINOR      1
#define UPNP_VERSION_MAJOR_STR  XSTR(UPNP_VERSION_MAJOR)
#define UPNP_VERSION_MINOR_STR  XSTR(UPNP_VERSION_MINOR)

#define UNUSED(arg)     (void)(arg)

#ifdef DEBUG
#define DEBUG_LOG(l)    l
#else
#define DEBUG_LOG(l)
#endif

#if !defined(__MUSL__)
#include <sys/queue.h>
#else
/* #include <sys/queue.h> - musl have not */
/*
 * List definitions.
 */
#define LIST_HEAD(name, type)                                           \
struct name {                                                           \
	struct type *lh_first;  /* first element */                     \
}

#define LIST_ENTRY(type)                                                \
struct {                                                                \
	struct type *le_next;   /* next element */                      \
	struct type **le_prev;  /* address of previous next element */  \
}

/*
 * List functions.
 */
#define LIST_INIT(head) do {                                            \
	(head)->lh_first = NULL;                                        \
} while (/*CONSTCOND*/0)

#define LIST_INSERT_HEAD(head, elm, field) do {                         \
	if (((elm)->field.le_next = (head)->lh_first) != NULL)          \
		(head)->lh_first->field.le_prev = &(elm)->field.le_next;\
	(head)->lh_first = (elm);                                       \
	(elm)->field.le_prev = &(head)->lh_first;                       \
} while (/*CONSTCOND*/0)

#define LIST_REMOVE(elm, field) do {                                    \
	if ((elm)->field.le_next != NULL)                               \
		(elm)->field.le_next->field.le_prev =                   \
		    (elm)->field.le_prev;                               \
	*(elm)->field.le_prev = (elm)->field.le_next;                   \
} while (/*CONSTCOND*/0)
/* end #include <sys/queue.h> */
#endif /* __MUSL__ */

/* theses macros are designed to read/write unsigned short/long int
 * from an unsigned char array in network order (big endian).
 * Avoid pointer casting, so avoid accessing unaligned memory, which
 * can crash with some cpu's */
static inline uint32_t readnu32(const uint8_t * p)
{
	return (p[0] << 24 | p[1] << 16 | p[2] << 8 | p[3]);
}
#define READNU32(p) readnu32(p)
static inline uint16_t readnu16(const uint8_t * p)
{
	return (p[0] << 8 | p[1]);
}
#define READNU16(p) readnu16(p)
static inline void writenu32(uint8_t * p, uint32_t n)
{
	p[0] = (n & 0xff000000) >> 24;
	p[1] = (n & 0xff0000) >> 16;
	p[2] = (n & 0xff00) >> 8;
	p[3] = n & 0xff;
}
#define WRITENU32(p, n) writenu32(p, n)
static inline void writenu16(uint8_t * p, uint16_t n)
{
	p[0] = (n & 0xff00) >> 8;
	p[1] = n & 0xff;
}
#define WRITENU16(p, n) writenu16(p, n)


/* Paths and other URLs in the miniupnpd http server */

#define ROOTDESC_PATH           "/rootDesc.xml"

#define WANCFG_PATH             "/WANCfg.xml"
#define WANCFG_CONTROLURL       "/ctl/CmnIfCfg"
#define WANCFG_EVENTURL         "/evt/CmnIfCfg"

#define WANIPC_PATH             "/WANIPCn.xml"
#define WANIPC_CONTROLURL       "/ctl/IPConn"
#define WANIPC_EVENTURL         "/evt/IPConn"

#ifdef ENABLE_L3F_SERVICE
#define L3F_PATH                "/L3F.xml"
#define L3F_CONTROLURL          "/ctl/L3F"
#define L3F_EVENTURL            "/evt/L3F"
#endif

/* structure and list for storing lan addresses
 * with ascii representation and mask */
struct lan_addr_s {
	char ifname[IFNAMSIZ];  /* example: eth0 */
#if ENABLE_FEATURE_IPV6
	unsigned int index;             /* use if_nametoindex() */
#endif
	char str[16];   /* example: 192.168.0.1 */
	struct in_addr addr, mask;      /* ip/mask */
	LIST_ENTRY(lan_addr_s) list;
};
LIST_HEAD(lan_addr_list, lan_addr_s);

/* strings used in the root device xml description */
#define ROOTDEV_FRIENDLYNAME            OS_NAME " router"
#define ROOTDEV_MANUFACTURER            OS_NAME
#define ROOTDEV_MANUFACTURERURL         OS_URL
#define ROOTDEV_MODELNAME               OS_NAME " router"
#define ROOTDEV_MODELDESCRIPTION        OS_NAME " router"
#define ROOTDEV_MODELURL                OS_URL

#define WANDEV_FRIENDLYNAME             "WANDevice"
#define WANDEV_MANUFACTURER             "MiniUPnP"
#define WANDEV_MANUFACTURERURL          "http://miniupnp.free.fr/"
#define WANDEV_MODELNAME                "WAN Device"
#define WANDEV_MODELDESCRIPTION         "WAN Device"
#define WANDEV_MODELNUMBER              MINIUPNPD_DATE
#define WANDEV_MODELURL                 "http://miniupnp.free.fr/"
#define WANDEV_UPC                      "000000000000"
/* UPC is 12 digit (barcode) */

#define WANCDEV_FRIENDLYNAME            "WANConnectionDevice"
#define WANCDEV_MANUFACTURER            WANDEV_MANUFACTURER
#define WANCDEV_MANUFACTURERURL         WANDEV_MANUFACTURERURL
#define WANCDEV_MODELNAME               "MiniUPnPd"
#define WANCDEV_MODELDESCRIPTION        "MiniUPnP daemon"
#define WANCDEV_MODELNUMBER             MINIUPNPD_DATE
#define WANCDEV_MODELURL                "http://miniupnp.free.fr/"
#define WANCDEV_UPC                     "000000000000"
/* UPC is 12 digit (barcode) */

/* IGD v1 */
#define SERVICE_TYPE_WANIPC "urn:schemas-upnp-org:service:WANIPConnection:1"
#define SERVICE_ID_WANIPC   "urn:upnp-org:serviceId:WANIPConn1"


/* runtime boolean flags */
#define LOGPACKETSMASK          0x0001
#define SYSUPTIMEMASK           0x0002
#if ENABLE_FEATURE_MINIUPNPD_NATPMP
#define ENABLENATPMPMASK        0x0004
#endif
#define CHECKCLIENTIPMASK       0x0008
#define SECUREMODEMASK          0x0010

#define ENABLEUPNPMASK          0x0020

#if ENABLE_FEATURE_IPV6
#define IPV6DISABLEDMASK        0x0080
#endif
#ifdef ENABLE_PCP
#define PCP_ALLOWTHIRDPARTYMASK 0x0400
#endif

#define SETFLAG(mask)   bb_common_bss.runtime_flags |= mask
#define GETFLAG(mask)   (bb_common_bss.runtime_flags & mask)
#define CLEARFLAG(mask) bb_common_bss.runtime_flags &= ~mask


/* for the root description
 * The child list reference is stored in "data" member using the
 * INITHELPER macro with index/nchild always in the
 * same order, whatever the endianness */
struct XMLElt {
	const char *eltname;    /* begin with '/' if no child */
	const char *data;       /* Value */
};

/* for service description */
struct serviceDesc {
	const struct action *actionList;
	const struct stateVar *serviceStateTable;
};

struct action {
	const char *name;
	const struct argument *args;
};

struct argument {       /* the name of the arg is obtained from the variable */
	unsigned char dir;      /* MSB : don't append "New" Flag,
				 * 5 Medium bits : magic argument name index
				 * 2 LSB : 1 = in, 2 = out */
	unsigned char relatedVar;       /* index of the related variable */
};

struct stateVar {
	const char * name;
	unsigned char itype;    /* MSB: sendEvent flag, 7 LSB: index in upnptypes */
	unsigned char idefault; /* default value */
	unsigned char iallowedlist;     /* index in allowed values list
					 * or in allowed range list */
	unsigned char ieventvalue;      /* fixed value returned or magical values */
};

/* little endian
 * The code has now be tested on big endian architecture */
#define INITHELPER(i, n) ((char *)(((n)<<16)|(i)))

#if !defined(SA_LEN)
# define SA_LEN(sa) ((sa)->sa_len)
#endif

#ifndef MAX
# define MAX(a,b) (((a)>(b))?(a):(b))
#endif

/* netfilter cannot store redirection descriptions, so we use our
 * own structure to store them */
struct rdr_desc {
	struct rdr_desc * next;
	unsigned int timestamp;
	unsigned short eport;
	short proto;
	char str[];
};


CALC_BSS struct BB_COMMON_BSS
{
	/* network interface for internet */
	const char *ext_if_name;

	/* forced ip address to use for this interface
	 * when NULL, getifaddr() is used */
	const char *use_ext_ip_addr;

	/* parameters to return to upnp client when asked */
	unsigned long downstream_bitrate;
	unsigned long upstream_bitrate;

	/* startup time */
	time_t startup_time;

#ifdef ENABLE_PCP
	unsigned long int min_lifetime;
	unsigned long int max_lifetime;
#endif

	int runtime_flags;

	const char *pidfilename;

	char uuidvalue_igd[42]; /* "uuid:00000000-0000-0000-0000-000000000000" uuid of root device (IGD) */
	char uuidvalue_wan[42]; /* "uuid:00000000-0000-0000-0000-000000000000" uuid of WAN Device */
	char uuidvalue_wcd[42]; /* "uuid:00000000-0000-0000-0000-000000000000" uuid of WAN Connection Device */
#define SERIALNUMBER_MAX_LEN (10)
	char serialnumber[SERIALNUMBER_MAX_LEN];

#define MODELNUMBER_MAX_LEN (48)
	char modelnumber[MODELNUMBER_MAX_LEN];

	/* presentation url :
	 * http://nnn.nnn.nnn.nnn:ppppp/  => max 30 bytes including terminating 0 */
#define PRESENTATIONURL_MAX_LEN (64)
	char presentationurl[PRESENTATIONURL_MAX_LEN];

	/* UPnP permission rules : */
	llist_t *upnppermlist_list_t;

	/* For automatic removal of expired rules (with LeaseDuration) */
	unsigned int nextruletoclean_timestamp;

#ifdef USE_NETFILTER
	/* chain names to use in the nat and filter tables. */

	/* iptables -t nat -N MINIUPNPD
	 * iptables -t nat -A PREROUTING -i <ext_if_name> -j MINIUPNPD */
	const char *miniupnpd_nat_chain;

	/* iptables -t nat -N MINIUPNPD-POSTROUTING
	 * iptables -t nat -A POSTROUTING -o <ext_if_name> -j MINIUPNPD-POSTROUTING */
	const char *miniupnpd_nat_postrouting_chain;

	/* iptables -t filter -N MINIUPNPD
	 * iptables -t filter -A FORWARD -i <ext_if_name> ! -o <ext_if_name> -j MINIUPNPD */
	const char *miniupnpd_forward_chain;

#ifdef ENABLE_UPNPPINHOLE
	/* ip6tables -t filter -N MINIUPNPD
	 * ip6tables -t filter -A FORWARD -i <ext_if_name> ! -o <ext_if_name> -j MINIUPNPD */
	const char *miniupnpd_v6_filter_chain;
#endif /* ENABLE_UPNPPINHOLE */

#endif /* USE_NETFILTER */

	struct lan_addr_list lan_addrs; /* lan addresses to listen to SSDP traffic */

#if ENABLE_FEATURE_IPV6
	/* ipv6 address used for HTTP */
	char ipv6_addr_for_http_with_brackets[64];
	/* address used to bind local services */
	struct in6_addr ipv6_bind_addr;
#endif
	/* can contain a IPv4-mapped IPv6 address */
	struct in6_addr external_addr;

	/* Path of the Unix socket used to communicate with MiniSSDPd */
	const char *minissdpdsocketpath;

/* BOOTID.UPNP.ORG and CONFIGID.UPNP.ORG */
/* See UPnP Device Architecture v1.1 section 1.2 Advertisement :
 * The field value of the BOOTID.UPNP.ORG header field MUST be increased
 * each time a device (re)joins the network and sends an initial announce
 * (a "reboot" in UPnP terms), or adds a UPnP-enabled interface.
 * Unless the device explicitly announces a change in the BOOTID.UPNP.ORG
 * field value using an SSDP message, as long as the device remains
 * continuously available in the network, the same BOOTID.UPNP.ORG field
 * value MUST be used in all repeat announcements, search responses,
 * update messages and eventually bye-bye messages. */
	unsigned int upnp_bootid;      /* 1 - BOOTID.UPNP.ORG */
/* The field value of the CONFIGID.UPNP.ORG header field identifies the
 * current set of device and service descriptions; control points can
 * parse this header field to detect whether they need to send new
 * description query messages. */
/* UPnP 1.1 devices MAY freely assign configid numbers from 0 to
 * 16777215 (2^24-1). Higher numbers are reserved for future use, and
 * can be assigned by the Technical Committee. The configuration of a
 * root device consists of the following information: the DDD of the
 * root device and all its embedded devices, and the SCPDs of all the
 * contained services. If any part of the configuration changes, the
 * CONFIGID.UPNP.ORG field value MUST be changed.
 * DDD = Device Description Document
 * SCPD = Service Control Protocol Description */
	unsigned int upnp_configid; /* 1337 - CONFIGID.UPNP.ORG */

	/* Subscriber list */
	LIST_HEAD(listhead, subscriber) subscriberlist;

	/* notify list */
	LIST_HEAD(listheadnotif, upnp_event_notify) notifylist;

	/* variables used by signals */
	volatile sig_atomic_t quitting;

	/* pointer to the chained list where descriptions are stored */
	struct rdr_desc *rdr_desc_list;

	LIST_HEAD(send_listhead, scheduled_send) send_list;

	volatile sig_atomic_t should_send_public_address_change_notif;

} bb_common_bss;
#define external_addr bb_common_bss.external_addr
#define subscriberlist bb_common_bss.subscriberlist
#define notifylist bb_common_bss.notifylist

/**
 * convert a struct sockaddr to a human readable string.
 * [ipv6]:port or ipv4:port
 * return the number of characters used (as snprintf)
 */
static int
sockaddr_to_string(const struct sockaddr * addr, char * str, size_t size)
{
	char buffer[64];
	unsigned short port = 0;
	int n = -1;

	switch(addr->sa_family)
	{
#if ENABLE_FEATURE_IPV6
	case AF_INET6:
		if(inet_ntop(addr->sa_family,
			     &((struct sockaddr_in6 *)addr)->sin6_addr,
			     buffer, sizeof(buffer)) == NULL) {
			snprintf(buffer, sizeof(buffer), "inet_ntop: %s", strerror(errno));
		}
		port = ntohs(((struct sockaddr_in6 *)addr)->sin6_port);
		if(((struct sockaddr_in6 *)addr)->sin6_scope_id > 0) {
			char ifname[IF_NAMESIZE];
			if(if_indextoname(((struct sockaddr_in6 *)addr)->sin6_scope_id, ifname) == NULL)
				strcpy(ifname, "ERROR");
			n = snprintf(str, size, "[%s%%%s]:%hu", buffer, ifname, port);
		} else {
			n = snprintf(str, size, "[%s]:%hu", buffer, port);
		}
		break;
#endif /* ENABLE_FEATURE_IPV6 */
	case AF_INET:
		if(inet_ntop(addr->sa_family,
			     &((struct sockaddr_in *)addr)->sin_addr,
			     buffer, sizeof(buffer)) == NULL) {
			snprintf(buffer, sizeof(buffer), "inet_ntop: %s", strerror(errno));
		}
		port = ntohs(((struct sockaddr_in *)addr)->sin_port);
		n = snprintf(str, size, "%s:%hu", buffer, port);
		break;
#ifdef AF_LINK
#if defined(__sun)
		/* solaris does not seem to have link_ntoa */
		/* #define link_ntoa _link_ntoa */
#define link_ntoa(x) "dummy-link_ntoa"
#endif
	case AF_LINK:
		{
			struct sockaddr_dl * sdl = (struct sockaddr_dl *)addr;
			n = snprintf(str, size, "index=%hu type=%d %s",
				     sdl->sdl_index, sdl->sdl_type,
				     link_ntoa(sdl));
		}
		break;
#endif  /* AF_LINK */
	default:
		n = snprintf(str, size, "unknown address family %d", addr->sa_family);
	}
	return n;
}


/**
 * set the file description as non blocking
 * return 0 in case of failure, 1 in case of success
 */
static int set_non_blocking(int fd)
{
	int flags = fcntl(fd, F_GETFL);
	if(flags < 0)
		return 0;
	if(fcntl(fd, F_SETFL, flags | O_NONBLOCK) < 0)
		return 0;
	return 1;
}

#if ENABLE_FEATURE_IPV6
/**
 * nfnl_addattr_l - Add variable length attribute to nlmsghdr
 * @n: netlink message header to which attribute is to be added
 * @maxlen: maximum length of netlink message header
 * @type: type of new attribute
 * @data: content of new attribute
 * @len: attribute length
 */
#define NFA_ALIGNTO     4
#define NFA_ALIGN(len)  (((len) + NFA_ALIGNTO - 1) & ~(NFA_ALIGNTO - 1))
#define NFA_LENGTH(len) (NFA_ALIGN(sizeof(struct nfattr)) + (len))
#define NFA_DATA(nfa)   ((void *)(((char *)(nfa)) + NFA_LENGTH(0)))
#define NLMSG_TAIL(nlh) \
	(((void *) (nlh)) + NLMSG_ALIGN((nlh)->nlmsg_len))

/* Generic structure for encapsulation optional netfilter information.
 * It is reminiscent of sockaddr, but with sa_family replaced
 * with attribute type.
 * ! This should someday be put somewhere generic as now rtnetlink and
 * ! nfnetlink use the same attributes methods. - J. Schulist.
 */

struct nfattr
{
	u_int16_t nfa_len;
	u_int16_t nfa_type;     /* we use 15 bits for the type, and the highest
				 * bit to indicate whether the payload is nested */
};

static int nfnl_addattr_l(struct nlmsghdr *n, int maxlen, int type, const void *data,
		   int alen)
{
	int len = NFA_LENGTH(alen);
	struct nfattr *nfa;

	assert(n);
	assert(maxlen > 0);
	assert(type >= 0);

	if ((len + (int)NLMSG_ALIGN(n->nlmsg_len)) > maxlen) {
		errno = ENOSPC;
		return -1;
	}

	nfa = NLMSG_TAIL(n);
	nfa->nfa_type = type;
	nfa->nfa_len = len;
	memcpy(NFA_DATA(nfa), data, alen);
	n->nlmsg_len = (NLMSG_ALIGN(n->nlmsg_len) + NFA_ALIGN(len));
	return 0;
}


static int get_src_for_route_to(const struct sockaddr * dst,
		     void * src, size_t * src_len,
		     int * index)
{
	int fd = -1;
	struct nlmsghdr *h;
	int status;
	struct {
		struct nlmsghdr n;
		struct rtmsg r;
		char buf[1024];
	} req;
	struct sockaddr_nl nladdr;
	struct iovec iov = {
		.iov_base = (void*) &req.n,
	};
	struct msghdr msg = {
		.msg_name = &nladdr,
		.msg_namelen = sizeof(nladdr),
		.msg_iov = &iov,
		.msg_iovlen = 1,
	};
	const struct sockaddr_in * dst4;
	const struct sockaddr_in6 * dst6;
	memset(&req, 0, sizeof(req));
	req.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
	req.n.nlmsg_flags = NLM_F_REQUEST;
	req.n.nlmsg_type = RTM_GETROUTE;
	req.r.rtm_family = dst->sa_family;
	req.r.rtm_table = 0;
	req.r.rtm_protocol = 0;
	req.r.rtm_scope = 0;
	req.r.rtm_type = 0;
	req.r.rtm_src_len = 0;
	req.r.rtm_dst_len = 0;
	req.r.rtm_tos = 0;

	{
		char dst_str[128];
		sockaddr_to_string(dst, dst_str, sizeof(dst_str));
		DEBUG_LOG(syslog(LOG_DEBUG, "get_src_for_route_to (%s)", dst_str);)
	}
	/* add address */
	if(dst->sa_family == AF_INET) {
		dst4 = (const struct sockaddr_in *)dst;
		nfnl_addattr_l(&req.n, sizeof(req), RTA_DST, &dst4->sin_addr, 4);
		req.r.rtm_dst_len = 32;
	} else {
		dst6 = (const struct sockaddr_in6 *)dst;
		nfnl_addattr_l(&req.n, sizeof(req), RTA_DST, &dst6->sin6_addr, 16);
		req.r.rtm_dst_len = 128;
	}
	fd = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
	if (fd < 0) {
		bb_perror_msg("socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE)");
		return -1;
	}

	memset(&nladdr, 0, sizeof(nladdr));
	nladdr.nl_family = AF_NETLINK;

	req.n.nlmsg_seq = 1;
	iov.iov_len = req.n.nlmsg_len;

	status = sendmsg(fd, &msg, 0);

	if (status < 0) {
		bb_perror_msg("sendmsg(rtnetlink)");
		goto error;
	}

	memset(&req, 0, sizeof(req));

	for(;;) {
		iov.iov_len = sizeof(req);
		status = recvmsg(fd, &msg, 0);
		if(status < 0) {
			if (errno == EINTR || errno == EAGAIN)
				continue;
			bb_perror_msg("recvmsg(rtnetlink)");
			goto error;
		}
		if(status == 0) {
			bb_error_msg("recvmsg(rtnetlink) EOF");
			goto error;
		}
		for (h = (struct nlmsghdr*)&req.n; status >= (int)sizeof(*h); ) {
			int len = h->nlmsg_len;
			int l = len - sizeof(*h);

			if (l<0 || len>status) {
				if (msg.msg_flags & MSG_TRUNC) {
					bb_error_msg("Truncated message");
				}
				bb_error_msg("malformed message: len=%d", len);
				goto error;
			}

			if(nladdr.nl_pid != 0 || h->nlmsg_seq != 1/*seq*/) {
				bb_error_msg("wrong seq = %d", h->nlmsg_seq);
				/* Don't forget to skip that message. */
				status -= NLMSG_ALIGN(len);
				h = (struct nlmsghdr*)((char*)h + NLMSG_ALIGN(len));
				continue;
			}

			if(h->nlmsg_type == NLMSG_ERROR) {
				struct nlmsgerr *err = (struct nlmsgerr*)NLMSG_DATA(h);
				bb_error_msg("NLMSG_ERROR %d : %s", err->error, strerror(-err->error));
				goto error;
			}
			if(h->nlmsg_type == RTM_NEWROUTE) {
				struct rtattr * rta;

				len = h->nlmsg_len - NLMSG_LENGTH(sizeof(struct rtmsg));
				for(rta = RTM_RTA(NLMSG_DATA((h))); RTA_OK(rta, len); rta = RTA_NEXT(rta,len)) {
					unsigned char * data = RTA_DATA(rta);
					if(rta->rta_type == RTA_PREFSRC) {
						if(src_len && src) {
							if(*src_len < RTA_PAYLOAD(rta)) {
								xlog_warn("cannot copy src: %u<%lu",
								       (unsigned)*src_len, (unsigned long)RTA_PAYLOAD(rta));
								goto error;
							}
							*src_len = RTA_PAYLOAD(rta);
							memcpy(src, data, RTA_PAYLOAD(rta));
						}
					} else if(rta->rta_type == RTA_OIF) {
						if(index)
							memcpy(index, data, sizeof(int));
					}
				}
				close(fd);
				return 0;
			}
			status -= NLMSG_ALIGN(len);
			h = (struct nlmsghdr*)((char*)h + NLMSG_ALIGN(len));
		}
	}
	xlog_warn("get_src_for_route_to() : src not found");
error:
	if(fd >= 0)
		close(fd);
	return -1;
}
#endif /* ENABLE_FEATURE_IPV6 */

/* get the LAN which the peer belongs to */
static struct lan_addr_s *
get_lan_for_peer(const struct sockaddr * peer)
{
	struct lan_addr_s * lan_addr = NULL;

#if ENABLE_FEATURE_IPV6
	if(peer->sa_family == AF_INET6)
	{
		struct sockaddr_in6 * peer6 = (struct sockaddr_in6 *)peer;
		if(IN6_IS_ADDR_V4MAPPED(&peer6->sin6_addr))
		{
			struct in_addr peer_addr;
			memcpy(&peer_addr, &peer6->sin6_addr.s6_addr[12], 4);
			for(lan_addr = bb_common_bss.lan_addrs.lh_first;
			    lan_addr != NULL;
			    lan_addr = lan_addr->list.le_next)
			{
				if( (peer_addr.s_addr & lan_addr->mask.s_addr)
				   == (lan_addr->addr.s_addr & lan_addr->mask.s_addr))
					break;
			}
		}
		else
		{
			int index = -1;
			if(peer6->sin6_scope_id > 0)
				index = (int)peer6->sin6_scope_id;
			else
			{
				if(get_src_for_route_to(peer, NULL, NULL, &index) < 0)
					return NULL;
			}
			DEBUG_LOG(syslog(LOG_DEBUG, "%s looking for LAN interface index=%d",
			       "get_lan_for_peer()", index);)
			for(lan_addr = bb_common_bss.lan_addrs.lh_first;
			    lan_addr != NULL;
			    lan_addr = lan_addr->list.le_next)
			{
				DEBUG_LOG(syslog(LOG_DEBUG,
				       "ifname=%s index=%u str=%s addr=%08x mask=%08x",
				       lan_addr->ifname, lan_addr->index,
				       lan_addr->str,
				       ntohl(lan_addr->addr.s_addr),
				       ntohl(lan_addr->mask.s_addr));)
				if(index == (int)lan_addr->index)
					break;
			}
		}
	}
	else if(peer->sa_family == AF_INET)
	{
#endif /* ENABLE_FEATURE_IPV6 */
		for(lan_addr = bb_common_bss.lan_addrs.lh_first;
		    lan_addr != NULL;
		    lan_addr = lan_addr->list.le_next)
		{
			if( (((const struct sockaddr_in *)peer)->sin_addr.s_addr & lan_addr->mask.s_addr)
			   == (lan_addr->addr.s_addr & lan_addr->mask.s_addr))
				break;
		}
#if ENABLE_FEATURE_IPV6
	}
#endif

	return lan_addr;
}


#if ENABLE_FEATURE_IPV6
#define sendto_schedule(sockfd, buf, len, flags, dest_addr, addrlen, delay) \
	sendto_schedule2(sockfd, buf, len, flags, dest_addr, addrlen, NULL, delay)
#else
#define sendto_schedule(sockfd, buf, len, flags, dest_addr, addrlen, delay) \
	sendto_schedule2(sockfd, buf, len, flags, dest_addr, addrlen, delay)
#endif

enum E_STATE {ESCHEDULED=1, EWAITREADY=2, ESENDNOW=3};

/* state diagram for a packet :
 *
 *                     |
 *                     V
 * -> ESCHEDULED -> ESENDNOW -> sent
 *                    ^  |
 *                    |  V
 *                EWAITREADY -> sent
 */
struct scheduled_send {
	LIST_ENTRY(scheduled_send) entries;
	struct timeval ts;
	enum E_STATE state;
	int sockfd;
	const void * buf;
	size_t len;
	int flags;
	const struct sockaddr *dest_addr;
	socklen_t addrlen;
	const struct sockaddr_in6 *src_addr;
	char data[];
};

/*
 * ssize_t sendto(int sockfd, const void *buf, size_t len, int flags,
 *                const struct sockaddr *dest_addr, socklen_t addrlen);
 */
static ssize_t
send_from_to(int sockfd, const void *buf, size_t len, int flags,
	     const struct sockaddr *dest_addr, socklen_t addrlen)
{
	return sendto(sockfd, buf, len, flags, dest_addr, addrlen);
}

/* sendto_schedule() : see sendto(2)
 * schedule sendto() call after delay (milliseconds) */
static ssize_t
sendto_schedule2(int sockfd, const void *buf, size_t len, int flags,
		 const struct sockaddr *dest_addr, socklen_t addrlen,
#if ENABLE_FEATURE_IPV6
		 const struct sockaddr_in6 *src_addr,
#endif
		 unsigned int delay)
{
	ssize_t n;
	size_t alloc_len;
	struct timeval tv;
	struct scheduled_send * elt;
	enum E_STATE state;

	if(delay == 0) {
		/* first try to send at once */
		n = send_from_to(sockfd, buf, len, flags, dest_addr, addrlen);
		if(n >= 0)
			return n;
		else if(errno == EAGAIN || errno == EWOULDBLOCK) {
			/* use select() on this socket */
			state = EWAITREADY;
		} else if(errno == EINTR) {
			state = ESENDNOW;
		} else {
			/* uncatched error */
			return n;
		}
	} else {
		state = ESCHEDULED;
	}

	/* schedule */
	if(gettimeofday(&tv, 0) < 0) {
		return -1;
	}
	/* allocate enough space for structure + buffers */
	alloc_len = sizeof(struct scheduled_send) + len + addrlen;
#if ENABLE_FEATURE_IPV6
	if(src_addr)
		alloc_len += sizeof(struct sockaddr_in6);
#endif
	elt = malloc(alloc_len);
	if(elt == NULL) {
		bb_error_msg("malloc failed to allocate %u bytes", (unsigned)alloc_len);
		return -1;
	}
	elt->state = state;
	/* time the packet should be sent */
	elt->ts.tv_sec = tv.tv_sec + (delay / 1000);
	elt->ts.tv_usec = tv.tv_usec + (delay % 1000) * 1000;
	if(elt->ts.tv_usec > 1000000) {
		elt->ts.tv_sec++;
		elt->ts.tv_usec -= 1000000;
	}
	elt->sockfd = sockfd;
	elt->flags = flags;
	memcpy(elt->data, dest_addr, addrlen);
	elt->dest_addr = (struct sockaddr *)elt->data;
	elt->addrlen = addrlen;
#if ENABLE_FEATURE_IPV6
	if(src_addr) {
		elt->src_addr = (struct sockaddr_in6 *)(elt->data + addrlen);
		memcpy((void *)elt->src_addr, src_addr, sizeof(struct sockaddr_in6));
		elt->buf = (void *)(elt->data + addrlen + sizeof(struct sockaddr_in6));
	} else
#endif
		{
			elt->src_addr = NULL;
			elt->buf = (void *)(elt->data + addrlen);
		}
	elt->len = len;
	memcpy((void *)elt->buf, buf, len);
	/* insert */
	LIST_INSERT_HEAD( &bb_common_bss.send_list, elt, entries);
	return 0;
}

/* try to send at once, and queue the packet if needed */
static ssize_t
sendto_or_schedule(int sockfd, const void *buf, size_t len, int flags,
		   const struct sockaddr *dest_addr, socklen_t addrlen)
{
	return sendto_schedule2(sockfd, buf, len, flags, dest_addr, addrlen,
#if ENABLE_FEATURE_IPV6
		NULL,
#endif
		0);
}

/* return number of scheduled sendto
 * set next_send to timestamp to send next packet
 * return number of scheduled send in list */
static int get_next_scheduled_send(struct timeval * next_send)
{
	int n = 0;
	struct scheduled_send * elt;
	if(next_send == NULL)
		return -1;
	for(elt = bb_common_bss.send_list.lh_first; elt != NULL; elt = elt->entries.le_next) {
		if(n == 0 || (elt->ts.tv_sec < next_send->tv_sec) ||
		   (elt->ts.tv_sec == next_send->tv_sec && elt->ts.tv_usec < next_send->tv_usec)) {
			next_send->tv_sec = elt->ts.tv_sec;
			next_send->tv_usec = elt->ts.tv_usec;
		}
		n++;
	}
	return n;
}

/* update writefds for select() call
 * return the number of packets to try to send at once */
static int get_sendto_fds(fd_set * writefds, int * max_fd, const struct timeval * now)
{
	int n = 0;
	struct scheduled_send * elt;
	for(elt = bb_common_bss.send_list.lh_first; elt != NULL; elt = elt->entries.le_next) {
		if(elt->state == EWAITREADY) {
			/* last sendto() call returned EAGAIN/EWOULDBLOCK */
			FD_SET(elt->sockfd, writefds);
			if(elt->sockfd > *max_fd)
				*max_fd = elt->sockfd;
			n++;
		} else if((elt->ts.tv_sec < now->tv_sec) ||
			  (elt->ts.tv_sec == now->tv_sec && elt->ts.tv_usec <= now->tv_usec)) {
			/* we waited long enough, now send ! */
			elt->state = ESENDNOW;
			n++;
		}
	}
	return n;
}

/* executed sendto() when needed */
static int try_sendto(fd_set * writefds)
{
	int ret = 0;
	ssize_t n;
	struct scheduled_send * elt;
	struct scheduled_send * next;
	for(elt = bb_common_bss.send_list.lh_first; elt != NULL; elt = next) {
		next = elt->entries.le_next;
		if((elt->state == ESENDNOW) ||
		   (elt->state == EWAITREADY && FD_ISSET(elt->sockfd, writefds))) {
			n = send_from_to(elt->sockfd, elt->buf, elt->len, elt->flags,
						elt->dest_addr, elt->addrlen);
			/*n = sendto(elt->sockfd, elt->buf, elt->len, elt->flags,
				   elt->dest_addr, elt->addrlen);*/
			if(n < 0) {
				if(errno == EINTR) {
					/* retry at once */
					elt->state = ESENDNOW;
					continue;
				} else if(errno == EAGAIN || errno == EWOULDBLOCK) {
					/* retry once the socket is ready for writing */
					elt->state = EWAITREADY;
					continue;
				} else {
					char addr_str[64];
					/* uncatched error */
					if(sockaddr_to_string(elt->dest_addr, addr_str, sizeof(addr_str)) <= 0)
						addr_str[0] = '\0';
					bb_perror_msg("%s(sock=%d, len=%u, dest=%s): sendto",
					       "try_sendto", elt->sockfd, (unsigned)elt->len,
					       addr_str);
					ret--;
				}
			} else if((int)n != (int)elt->len) {
				xlog_warn("%s: %d bytes sent out of %d",
				       "try_sendto", (int)n, (int)elt->len);
			}
			/* remove from the list */
			LIST_REMOVE(elt, entries);
			free(elt);
		}
	}
	return ret;
}

/* maximum execution time for finalize_sendto() in milliseconds */
#define FINALIZE_SENDTO_DELAY   (500)

/* empty the list */
static void finalize_sendto(void)
{
	ssize_t n;
	struct scheduled_send * elt;
	struct scheduled_send * next;
	fd_set writefds;
	struct timeval deadline;
	struct timeval now;
	struct timeval timeout;
	int max_fd;

	if(gettimeofday(&deadline, NULL) < 0) {
		bb_perror_msg("gettimeofday");
		return;
	}
	deadline.tv_usec += FINALIZE_SENDTO_DELAY*1000;
	if(deadline.tv_usec > 1000000) {
		deadline.tv_sec++;
		deadline.tv_usec -= 1000000;
	}
	while(bb_common_bss.send_list.lh_first) {
		FD_ZERO(&writefds);
		max_fd = -1;
		for(elt = bb_common_bss.send_list.lh_first; elt != NULL; elt = next) {
			next = elt->entries.le_next;
			DEBUG_LOG(syslog(LOG_DEBUG, "finalize_sendto(): %d bytes on socket %d",
			       (int)elt->len, elt->sockfd);)
			n = send_from_to(elt->sockfd, elt->buf, elt->len, elt->flags,
						elt->dest_addr, elt->addrlen);
			/*n = sendto(elt->sockfd, elt->buf, elt->len, elt->flags,
				   elt->dest_addr, elt->addrlen);*/
			if(n < 0) {
				if(errno==EAGAIN || errno==EWOULDBLOCK) {
					FD_SET(elt->sockfd, &writefds);
					if(elt->sockfd > max_fd)
						max_fd = elt->sockfd;
					continue;
				}
				xlog_warn("finalize_sendto(): socket=%d sendto: %m", elt->sockfd);
			}
			/* remove from the list */
			LIST_REMOVE(elt, entries);
			free(elt);
		}
		/* check deadline */
		if(gettimeofday(&now, NULL) < 0) {
			bb_perror_msg("gettimeofday");
			return;
		}
		if(now.tv_sec > deadline.tv_sec ||
		   (now.tv_sec == deadline.tv_sec && now.tv_usec > deadline.tv_usec)) {
			/* deadline ! */
			while((elt = bb_common_bss.send_list.lh_first) != NULL) {
				LIST_REMOVE(elt, entries);
				free(elt);
			}
			return;
		}
		/* compute timeout value */
		timeout.tv_sec = deadline.tv_sec - now.tv_sec;
		timeout.tv_usec = deadline.tv_usec - now.tv_usec;
		if(timeout.tv_usec < 0) {
			timeout.tv_sec--;
			timeout.tv_usec += 1000000;
		}
		if(max_fd >= 0) {
			if(select(max_fd + 1, NULL, &writefds, NULL, &timeout) < 0) {
				bb_perror_msg("select");
				return;
			}
		}
	}
}

/* getifaddr()
 * take a network interface name and write the
 * ip v4 address as text in the buffer
 * returns: 0 success, -1 failure */
static int
getifaddr(const char * ifname, char * buf, int len,
	  struct in_addr * addr, struct in_addr * mask)
{
	/* use ioctl SIOCGIFADDR. Works only for ip v4 */
	/* SIOCGIFADDR struct ifreq *  */
	int s;
	struct ifreq ifr;
	int ifrlen;
	struct sockaddr_in * ifaddr;
	ifrlen = sizeof(ifr);

	if(!ifname || ifname[0]=='\0')
		return -1;
	s = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM);
	if(s < 0)
		return -1;
	safe_ifname_cp(&ifr, ifname);
	if(ioctl(s, SIOCGIFFLAGS, &ifr, &ifrlen) < 0) {
		DEBUG_LOG(syslog(LOG_DEBUG, "ioctl(s, SIOCGIFFLAGS, ...): %m");)
		close(s);
		return -1;
	}
	if ((ifr.ifr_flags & IFF_UP) == 0) {
		DEBUG_LOG(syslog(LOG_DEBUG, "network interface %s is down", ifname);)
		close(s);
		return -1;
	}
	safe_ifname_cp(&ifr, ifname);
	if(xioctl(s, SIOCGIFADDR, &ifr, "ioctl(s, SIOCGIFADDR, ...)") < 0) {
		close(s);
		return -1;
	}
	ifaddr = (struct sockaddr_in *)&ifr.ifr_addr;
	if(addr) *addr = ifaddr->sin_addr;
	if(buf) {
		if(!inet_ntop(AF_INET, &ifaddr->sin_addr, buf, len)) {
			bb_perror_msg("inet_ntop()");
			close(s);
			return -1;
		}
	}
	if(mask) {
		safe_ifname_cp(&ifr, ifname);
		if(xioctl(s, SIOCGIFNETMASK, &ifr, "ioctl(s, SIOCGIFNETMASK, ...)") < 0) {
			close(s);
			return -1;
		}
#ifdef ifr_netmask
		*mask = ((struct sockaddr_in *)&ifr.ifr_netmask)->sin_addr;
#else
		*mask = ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr;
#endif
	}
	close(s);
	return 0;
}


#if ENABLE_FEATURE_IPV6
/* find a non link local IP v6 address for the interface.
 * if ifname is NULL, look for all interfaces */
static int
find_ipv6_addr(const char * ifname,
	       char * dst, int n)
{
	struct ifaddrs * ifap;
	struct ifaddrs * ife;
	const struct sockaddr_in6 * addr;
	char buf[64];
	int r = 0;

	if(!dst)
		return -1;

	if(getifaddrs(&ifap)<0)
	{
		bb_perror_msg("getifaddrs");
		return -1;
	}
	for(ife = ifap; ife; ife = ife->ifa_next)
	{
		/* skip other interfaces if one was specified */
		if(ifname && (0 != strcmp(ifname, ife->ifa_name)))
			continue;
		if(ife->ifa_addr == NULL)
			continue;
		if(ife->ifa_addr->sa_family == AF_INET6)
		{
			addr = (const struct sockaddr_in6 *)ife->ifa_addr;
			if(!IN6_IS_ADDR_LOOPBACK(&addr->sin6_addr)
			   && !IN6_IS_ADDR_LINKLOCAL(&addr->sin6_addr))
			{
				inet_ntop(ife->ifa_addr->sa_family,
					  &addr->sin6_addr,
					  buf, sizeof(buf));
				/* add brackets */
				snprintf(dst, n, "[%s]", buf);
				r = 1;
			}
		}
	}
	freeifaddrs(ifap);
	return r;
}
#endif


#define STATUS_UNCONFIGURED (0)
#define STATUS_CONNECTING (1)
#define STATUS_CONNECTED (2)
#define STATUS_PENDINGDISCONNECT (3)
#define STATUS_DISCONNECTING (4)
#define STATUS_DISCONNECTED (5)

/**
 * get the connection status
 * return values :
 *  0 - Unconfigured
 *  1 - Connecting
 *  2 - Connected
 *  3 - PendingDisconnect
 *  4 - Disconnecting
 *  5 - Disconnected */
static int get_wan_connection_status(const char * ifname)
{
	char addr[INET_ADDRSTRLEN];
	int r;

	/* we need a better implementation here.
	 * I'm afraid it should be device specific */
	r = getifaddr(ifname, addr, INET_ADDRSTRLEN, NULL, NULL);
	return (r < 0) ? STATUS_DISCONNECTED : STATUS_CONNECTED;
}

/**
 * return the same value as get_wan_connection_status()
 * as a C string */
static const char *get_wan_connection_status_str(const char * ifname)
{
	int status;
	const char * str = NULL;

	status = get_wan_connection_status(ifname);
	switch(status) {
	case 0:
		str = "Unconfigured";
		break;
	case 1:
		str = "Connecting";
		break;
	case 2:
		str = "Connected";
		break;
	case 3:
		str = "PendingDisconnect";
		break;
	case 4:
		str = "Disconnecting";
		break;
	case 5:
		str = "Disconnected";
		break;
	}
	return str;
}


#include <linux/types.h>

#define IP_NAT_RANGE_MAP_IPS 1
#define IP_NAT_RANGE_PROTO_SPECIFIED 2
#define IP_NAT_RANGE_PROTO_RANDOM 4
#define IP_NAT_RANGE_PERSISTENT 8

union nf_conntrack_man_proto {
	__be16 all;
	struct { __be16 port; } tcp;
	struct { __be16 port; } udp;
	struct { __be16 id;   } icmp;
	struct { __be16 port; } dccp;
	struct { __be16 port; } sctp;
	struct { __be16 key;  } gre;
};

struct nf_nat_range {
	unsigned int flags;
	__be32 min_ip, max_ip;
	union nf_conntrack_man_proto min, max;
};

struct nf_nat_multi_range_compat {
	unsigned int rangesize;
	struct nf_nat_range range[1];
};

#define nf_nat_multi_range nf_nat_multi_range_compat

#define ip_nat_multi_range      nf_nat_multi_range
#define ip_nat_range            nf_nat_range
#define IPTC_HANDLE             struct xtc_handle *

/* IPT_ALIGN was renamed XT_ALIGN in iptables-1.4.11 */
#ifndef IPT_ALIGN
#define IPT_ALIGN XT_ALIGN
#endif

/* dummy init and shutdown functions
 * Only test iptc_init() */
static void init_redirect(void)
{
	IPTC_HANDLE h = iptc_init("nat");
	if(!h)
		bb_error_msg_and_die("Failed to init redirection engine: %s. EXITING", iptc_strerror(errno));
	 else
		iptc_free(h);
}

/* convert an ip address to string */
static int snprintip(char * dst, size_t size, uint32_t ip)
{
	return snprintf(dst, size,
	       "%u.%u.%u.%u", ip >> 24, (ip >> 16) & 0xff,
	       (ip >> 8) & 0xff, ip & 0xff);
}

/* add a description to the list of redirection descriptions */
static void
add_redirect_desc(unsigned short eport, int proto,
		  const char * desc, unsigned int timestamp)
{
	struct rdr_desc * p;
	size_t l;
	/* set a default description if none given */
	if(!desc)
		desc = "miniupnpd";
	l = strlen(desc) + 1;
	p = malloc(sizeof(struct rdr_desc) + l);
	if(p) {
		p->next = bb_common_bss.rdr_desc_list;
		p->timestamp = timestamp;
		p->eport = eport;
		p->proto = (short)proto;
		memcpy(p->str, desc, l);
		bb_common_bss.rdr_desc_list = p;
	}
}

/* delete a description from the list */
static void
del_redirect_desc(unsigned short eport, int proto)
{
	struct rdr_desc * p, * last;
	p = bb_common_bss.rdr_desc_list;
	last = 0;
	while(p)
	{
		if(p->eport == eport && p->proto == proto)
		{
			if(!last)
				bb_common_bss.rdr_desc_list = p->next;
			else
				last->next = p->next;
			free(p);
			return;
		}
		last = p;
		p = p->next;
	}
}

/* go through the list to find the description */
static void
get_redirect_desc(unsigned short eport, int proto,
		  char * desc, int desclen,
		  unsigned int * timestamp)
{
	struct rdr_desc * p;
	for(p = bb_common_bss.rdr_desc_list; p; p = p->next)
	{
		if(p->eport == eport && p->proto == (short)proto)
		{
			if(desc)
				safe_strncpy(desc, p->str, desclen);
			if(timestamp)
				*timestamp = p->timestamp;
			return;
		}
	}
	/* if no description was found, return miniupnpd as default */
	if(desc)
		safe_strncpy(desc, "miniupnpd", desclen);
	if(timestamp)
		*timestamp = 0;
}

static int
get_nat_redirect_rule(const char * nat_chain_name, unsigned short eport, int proto,
		  char * iaddr, int iaddrlen, unsigned short * iport,
		  char * desc, int desclen,
		  char * rhost, int rhostlen,
		  unsigned int * timestamp,
		  u_int64_t * packets, u_int64_t * bytes)
{
	int r = -1;
	IPTC_HANDLE h;
	const struct ipt_entry * e;
	const struct ipt_entry_target * target;
	const struct ip_nat_multi_range * mr;
	const struct ipt_entry_match *match;

	h = iptc_init("nat");
	if(!h) {
		bb_error_msg("%s() : iptc_init() failed : %s", "get_nat_redirect_rule", iptc_strerror(errno));
		return -1;
	}
	if(!iptc_is_chain(nat_chain_name, h)) {
		bb_error_msg("chain %s not found", nat_chain_name);
	} else {
		for(e = iptc_first_rule(nat_chain_name, h); e; e = iptc_next_rule(e, h)) {
			if(proto==e->ip.proto) {
				match = (const struct ipt_entry_match *)&e->elems;
				if(0 == strncmp(match->u.user.name, "tcp", IPT_FUNCTION_MAXNAMELEN)) {
					const struct ipt_tcp * info;
					info = (const struct ipt_tcp *)match->data;
					if(eport != info->dpts[0])
						continue;
				} else {
					const struct ipt_udp * info;
					info = (const struct ipt_udp *)match->data;
					if(eport != info->dpts[0])
						continue;
				}
				target = (void *)e + e->target_offset;
				/* target = ipt_get_target(e); */
				mr = (const struct ip_nat_multi_range *)&target->data[0];
				snprintip(iaddr, iaddrlen, ntohl(mr->range[0].min_ip));
				*iport = ntohs(mr->range[0].min.all);
				get_redirect_desc(eport, proto, desc, desclen, timestamp);
				if(packets)
					*packets = e->counters.pcnt;
				if(bytes)
					*bytes = e->counters.bcnt;
				/* rhost */
				if(e->ip.src.s_addr && rhost) {
					snprintip(rhost, rhostlen, ntohl(e->ip.src.s_addr));
				}
				r = 0;
				break;
			}
		}
	}
	if(h)
		iptc_free(h);
	return r;
}

/* get_redirect_rule() gets internal IP and port from
 * interface, external port and protocol
 * return value :
 *  0 success (rule found)
 * -1 error or rule not found
 */
static int
get_redirect_rule(unsigned short eport, int proto,
		  char * iaddr, int iaddrlen, unsigned short * iport,
		  char * desc, int desclen,
		  char * rhost, int rhostlen,
		  unsigned int * timestamp,
		  u_int64_t * packets, u_int64_t * bytes)
{
	return get_nat_redirect_rule(bb_common_bss.miniupnpd_nat_chain,
				     eport, proto,
				     iaddr, iaddrlen, iport,
				     desc, desclen,
				     rhost, rhostlen,
				     timestamp, packets, bytes);
}

/* get_redirect_rule_by_index()
 * return values :
 *  0 success (rule found)
 * -1 error or rule not found */
static int
get_redirect_rule_by_index(int index,
			   char * ifname, unsigned short * eport,
			   char * iaddr, int iaddrlen, unsigned short * iport,
			   int * proto, char * desc, int desclen,
			   char * rhost, int rhostlen,
			   unsigned int * timestamp,
			   u_int64_t * packets, u_int64_t * bytes)
{
	int r = -1;
	int i = 0;
	IPTC_HANDLE h;
	const struct ipt_entry * e;
	const struct ipt_entry_target * target;
	const struct ip_nat_multi_range * mr;
	const struct ipt_entry_match *match;
	UNUSED(ifname);

	h = iptc_init("nat");
	if(!h)
	{
		bb_error_msg("%s() : iptc_init() failed : %s", "get_redirect_rule_by_index", iptc_strerror(errno));
		return -1;
	}
	if(!iptc_is_chain(bb_common_bss.miniupnpd_nat_chain, h))
	{
		bb_error_msg("chain %s not found", bb_common_bss.miniupnpd_nat_chain);
	}
	else
	{
		for(e = iptc_first_rule(bb_common_bss.miniupnpd_nat_chain, h); e; e = iptc_next_rule(e, h))
		{
			if(i==index)
			{
				*proto = e->ip.proto;
				match = (const struct ipt_entry_match *)&e->elems;
				if(0 == strncmp(match->u.user.name, "tcp", IPT_FUNCTION_MAXNAMELEN))
				{
					const struct ipt_tcp * info;
					info = (const struct ipt_tcp *)match->data;
					*eport = info->dpts[0];
				}
				else
				{
					const struct ipt_udp * info;
					info = (const struct ipt_udp *)match->data;
					*eport = info->dpts[0];
				}
				target = (void *)e + e->target_offset;
				mr = (const struct ip_nat_multi_range *)&target->data[0];
				snprintip(iaddr, iaddrlen, ntohl(mr->range[0].min_ip));
				*iport = ntohs(mr->range[0].min.all);
				get_redirect_desc(*eport, *proto, desc, desclen, timestamp);
				if(packets)
					*packets = e->counters.pcnt;
				if(bytes)
					*bytes = e->counters.bcnt;
				/* rhost */
				if(rhost && rhostlen > 0) {
					if(e->ip.src.s_addr) {
						snprintip(rhost, rhostlen, ntohl(e->ip.src.s_addr));
					} else {
						rhost[0] = '\0';
					}
				}
				r = 0;
				break;
			}
			i++;
		}
	}
	if(h)
		iptc_free(h);
	return r;
}

/* delete_rule_and_commit() :
 * subfunction used in delete_redirect_and_filter_rules() */
static int
delete_rule_and_commit(unsigned int index, IPTC_HANDLE h,
		       const char * miniupnpd_chain,
		       const char * logcaller)
{
	int r = 0;
	if(!iptc_delete_num_entry(miniupnpd_chain, index, h))
	{
		bb_error_msg("%s() : iptc_delete_num_entry(): %s", logcaller, iptc_strerror(errno));
		r = -1;
	}
	else if(!iptc_commit(h))
	{
		bb_error_msg("%s() : iptc_commit(): %s", logcaller, iptc_strerror(errno));
		r = -1;
	}
	if(h)
		iptc_free(h);
	return r;
}

static int
delete_redirect_and_filter_rules(unsigned short eport, int proto)
{
	int r = -1, r2 = -1;
	unsigned index = 0;
	unsigned i = 0;
	IPTC_HANDLE h;
	const struct ipt_entry * e;
	const struct ipt_entry_target * target;
	const struct ip_nat_multi_range * mr;
	const struct ipt_entry_match *match;
	unsigned short iport = 0;
	uint32_t iaddr = 0;

	h = iptc_init("nat");
	if(!h)
	{
		bb_error_msg("%s() : iptc_init() failed : %s", "delete_redirect_and_filter_rules", iptc_strerror(errno));
		return -1;
	}
	/* First step : find the right nat rule */
	if(!iptc_is_chain(bb_common_bss.miniupnpd_nat_chain, h))
	{
		bb_error_msg("chain %s not found", bb_common_bss.miniupnpd_nat_chain);
	}
	else
	{
		for(e = iptc_first_rule(bb_common_bss.miniupnpd_nat_chain, h); e; e = iptc_next_rule(e, h), i++)
		{
			if(proto==e->ip.proto)
			{
				match = (const struct ipt_entry_match *)&e->elems;
				if(0 == strncmp(match->u.user.name, "tcp", IPT_FUNCTION_MAXNAMELEN))
				{
					const struct ipt_tcp * info;
					info = (const struct ipt_tcp *)match->data;
					if(eport != info->dpts[0])
						continue;
				}
				else
				{
					const struct ipt_udp * info;
					info = (const struct ipt_udp *)match->data;
					if(eport != info->dpts[0])
						continue;
				}
				/* get the index, the internal address and the internal port
				 * of the rule */
				index = i;
				target = (void *)e + e->target_offset;
				mr = (const struct ip_nat_multi_range *)&target->data[0];
				iaddr = mr->range[0].min_ip;
				iport = ntohs(mr->range[0].min.all);
				r = 0;
				break;
			}
		}
	}
	if(h)
		iptc_free(h);
	if(r == 0) {
		DEBUG_LOG(syslog(LOG_INFO, "Trying to delete nat rule at index %u", index);)
		/* Now delete both rules */
		/* first delete the nat rule */
		h = iptc_init("nat");
		if(h) {
			r = delete_rule_and_commit(index, h, bb_common_bss.miniupnpd_nat_chain, "delete_redirect_rule");
		}
		if((r == 0) && (h = iptc_init("filter"))) {
			i = 0;
			/* we must find the right index for the filter rule */
			for(e = iptc_first_rule(bb_common_bss.miniupnpd_forward_chain, h); e; e = iptc_next_rule(e, h), i++) {
				if(proto==e->ip.proto) {
					match = (const struct ipt_entry_match *)&e->elems;
					/*syslog(LOG_DEBUG, "filter rule #%u: %s %s",
					       i, match->u.user.name, inet_ntoa(e->ip.dst));*/
					if(0 == strncmp(match->u.user.name, "tcp", IPT_FUNCTION_MAXNAMELEN)) {
						const struct ipt_tcp * info;
						info = (const struct ipt_tcp *)match->data;
						if(iport != info->dpts[0])
							continue;
					} else {
						const struct ipt_udp * info;
						info = (const struct ipt_udp *)match->data;
						if(iport != info->dpts[0])
							continue;
					}
					if(iaddr != e->ip.dst.s_addr)
						continue;
					index = i;
					DEBUG_LOG(syslog(LOG_INFO, "Trying to delete filter rule at index %u", index);)
					r = delete_rule_and_commit(index, h, bb_common_bss.miniupnpd_forward_chain, "delete_filter_rule");
					h = NULL;
					break;
				}
			}
		}
		if(h)
			iptc_free(h);
	}

	/*delete PEER rule*/
	if((h = iptc_init("nat"))) {
		i = 0;
		/* we must find the right index for the filter rule */
		for(e = iptc_first_rule(bb_common_bss.miniupnpd_nat_postrouting_chain, h); e; e = iptc_next_rule(e, h), i++) {
			if(proto==e->ip.proto) {
				target = (void *)e + e->target_offset;
				mr = (const struct ip_nat_multi_range *)&target->data[0];
				DEBUG_LOG(syslog(LOG_DEBUG, "postrouting rule #%u: %s %s %hu",
				       i, target->u.user.name, inet_ntoa(e->ip.src), ntohs(mr->range[0].min.all));)
				/* target->u.user.name SNAT / MASQUERADE */
				if (eport != ntohs(mr->range[0].min.all)) {
					continue;
				}
				iaddr = e->ip.src.s_addr;
				match = (const struct ipt_entry_match *)&e->elems;
				if(0 == strncmp(match->u.user.name, "tcp", IPT_FUNCTION_MAXNAMELEN)) {
					const struct ipt_tcp * info;
					info = (const struct ipt_tcp *)match->data;
					iport = info->spts[0];
				} else {
					const struct ipt_udp * info;
					info = (const struct ipt_udp *)match->data;
					iport = info->dpts[0];
				}

				index = i;
				DEBUG_LOG(syslog(LOG_INFO, "Trying to delete peer rule at index %u", index);)
				r2 = delete_rule_and_commit(index, h, bb_common_bss.miniupnpd_nat_postrouting_chain, "delete_peer_rule");
				h = NULL;
				break;
			}
		}
	}

	if(h)
		iptc_free(h);
	/*delete DSCP rule*/
	if((r2==0)&&(h = iptc_init("mangle"))) {
		i = 0;
		index = -1;
		/* we must find the right index for the filter rule */
		for(e = iptc_first_rule(bb_common_bss.miniupnpd_nat_chain, h); e; e = iptc_next_rule(e, h), i++) {
			if(proto==e->ip.proto) {
				match = (const struct ipt_entry_match *)&e->elems;
				/*syslog(LOG_DEBUG, "filter rule #%u: %s %s",
				       i, match->u.user.name, inet_ntoa(e->ip.dst));*/
				if(0 == strncmp(match->u.user.name, "tcp", IPT_FUNCTION_MAXNAMELEN)) {
					const struct ipt_tcp * info;
					info = (const struct ipt_tcp *)match->data;
					if(iport != info->spts[0])
						continue;
				} else {
					const struct ipt_udp * info;
					info = (const struct ipt_udp *)match->data;
					if(iport != info->spts[0])
						continue;
				}
				if(iaddr != e->ip.src.s_addr)
					continue;
				index = i;
				DEBUG_LOG(syslog(LOG_INFO, "Trying to delete dscp rule at index %u", index);)
				r2 = delete_rule_and_commit(index, h, bb_common_bss.miniupnpd_nat_chain, "delete_dscp_rule");
				h = NULL;
				break;
			}
		}
	if (h)
		iptc_free(h);
	}

	del_redirect_desc(eport, proto);
	return r*r2;
}

/* ==================================== */
/* TODO : add the -m state --state NEW,ESTABLISHED,RELATED
 * only for the filter rule */
static struct ipt_entry_match *
get_tcp_match(unsigned short dport, unsigned short sport)
{
	struct ipt_entry_match *match;
	struct ipt_tcp * tcpinfo;
	size_t size;
	size =   IPT_ALIGN(sizeof(struct ipt_entry_match))
	       + IPT_ALIGN(sizeof(struct ipt_tcp));
	match = calloc(1, size);
	match->u.match_size = size;
	safe_strncpy(match->u.user.name, "tcp", sizeof(match->u.user.name));
	tcpinfo = (struct ipt_tcp *)match->data;
	if (sport == 0) {
		tcpinfo->spts[0] = 0;           /* all source ports */
		tcpinfo->spts[1] = 0xFFFF;
	} else {
		tcpinfo->spts[0] = sport;          /* specified source port */
		tcpinfo->spts[1] = sport;
	}
	if (dport == 0) {
		tcpinfo->dpts[0] = 0;      /* all destination ports */
		tcpinfo->dpts[1] = 0xFFFF;
	} else {
		tcpinfo->dpts[0] = dport;       /* specified destination port */
		tcpinfo->dpts[1] = dport;
	}
	return match;
}

static struct ipt_entry_match *
get_udp_match(unsigned short dport, unsigned short sport)
{
	struct ipt_entry_match *match;
	struct ipt_udp * udpinfo;
	size_t size;
	size =   IPT_ALIGN(sizeof(struct ipt_entry_match))
	       + IPT_ALIGN(sizeof(struct ipt_udp));
	match = calloc(1, size);
	match->u.match_size = size;
	safe_strncpy(match->u.user.name, "udp", sizeof(match->u.user.name));
	udpinfo = (struct ipt_udp *)match->data;
	if (sport == 0) {
		udpinfo->spts[0] = 0;      /* all source ports */
		udpinfo->spts[1] = 0xFFFF;
	} else {
		udpinfo->spts[0] = sport;          /* specified source port */
		udpinfo->spts[1] = sport;
	}
	if (dport == 0) {
		udpinfo->dpts[0] = 0;      /* all destination ports */
		udpinfo->dpts[1] = 0xFFFF;
	} else {
		udpinfo->dpts[0] = dport;   /* specified destination port */
		udpinfo->dpts[1] = dport;
	}
	return match;
}

static struct ipt_entry_target *
get_dnat_target(const char * daddr, unsigned short dport)
{
	struct ipt_entry_target * target;
	struct ip_nat_multi_range * mr;
	struct ip_nat_range * range;
	size_t size;

	size =   IPT_ALIGN(sizeof(struct ipt_entry_target))
	       + IPT_ALIGN(sizeof(struct ip_nat_multi_range));
	target = calloc(1, size);
	target->u.target_size = size;
	safe_strncpy(target->u.user.name, "DNAT", sizeof(target->u.user.name));
	/* one ip_nat_range already included in ip_nat_multi_range */
	mr = (struct ip_nat_multi_range *)&target->data[0];
	mr->rangesize = 1;
	range = &mr->range[0];
	range->min_ip = range->max_ip = inet_addr(daddr);
	range->flags |= IP_NAT_RANGE_MAP_IPS;
	range->min.all = range->max.all = htons(dport);
	range->flags |= IP_NAT_RANGE_PROTO_SPECIFIED;
	return target;
}

#ifdef ENABLE_PORT_TRIGGERING
static struct ipt_entry_target *
get_masquerade_target(unsigned short port)
{
	struct ipt_entry_target * target;
	struct ip_nat_multi_range * mr;
	struct ip_nat_range * range;
	size_t size;

	size =   IPT_ALIGN(sizeof(struct ipt_entry_target))
	       + IPT_ALIGN(sizeof(struct ip_nat_multi_range));
	target = calloc(1, size);
	target->u.target_size = size;
	safe_strncpy(target->u.user.name, "MASQUERADE", sizeof(target->u.user.name));
	/* one ip_nat_range already included in ip_nat_multi_range */
	mr = (struct ip_nat_multi_range *)&target->data[0];
	mr->rangesize = 1;
	range = &mr->range[0];
	range->min.tcp.port = range->max.tcp.port = htons(port);
	/*range->min.all = range->max.all = htons(port);*/
	range->flags |= IP_NAT_RANGE_PROTO_SPECIFIED;
	return target;
}
#endif /* ENABLE_PORT_TRIGGERING */

/* iptc_init_verify_and_append()
 * return 0 on success, -1 on failure */
static int
iptc_init_verify_and_append(const char * table,
			    const char * miniupnpd_chain,
			    struct ipt_entry * e,
			    const char * logcaller)
{
	IPTC_HANDLE h;
	h = iptc_init(table);
	if(!h)
	{
		bb_error_msg("%s() : iptc_init() error : %s", logcaller, iptc_strerror(errno));
		return -1;
	}
	if(!iptc_is_chain(miniupnpd_chain, h))
	{
		bb_error_msg("%s() : chain %s not found", logcaller, miniupnpd_chain);
		if(h)
			iptc_free(h);
		return -1;
	}
	/* iptc_insert_entry(miniupnpd_chain, e, n, h/&h) could also be used */
	if(!iptc_append_entry(miniupnpd_chain, e, h))
	{
		bb_error_msg("%s() : iptc_append_entry() error : %s", logcaller, iptc_strerror(errno));
		if(h)
			iptc_free(h);
		return -1;
	}
	if(!iptc_commit(h))
	{
		bb_error_msg("%s() : iptc_commit() error : %s", logcaller, iptc_strerror(errno));
		if(h)
			iptc_free(h);
		return -1;
	}
	if(h)
		iptc_free(h);
	return 0;
}

/* add nat rule
 * iptables -t nat -A MINIUPNPD -p <proto> [-s <rhost>] --dport <eport> -j DNAT --to <iaddr>:<iport>
 * */
static int
addnatrule(int proto, unsigned short eport,
	   const char * iaddr, unsigned short iport,
	   const char * rhost)
{
	int r = 0;
	struct ipt_entry * e;
	struct ipt_entry * tmp;
	struct ipt_entry_match *match = NULL;
	struct ipt_entry_target *target = NULL;

	e = calloc(1, sizeof(struct ipt_entry));
	if(!e) {
		bb_error_msg("%s: calloc(%d) error", "addnatrule", (int)sizeof(struct ipt_entry));
		return -1;
	}
	e->ip.proto = proto;
	if(proto == IPPROTO_TCP) {
		match = get_tcp_match(eport, 0);
	} else {
		match = get_udp_match(eport, 0);
	}
	e->nfcache = NFC_IP_DST_PT;
	target = get_dnat_target(iaddr, iport);
	e->nfcache |= NFC_UNKNOWN;
	tmp = realloc(e, sizeof(struct ipt_entry)
		       + match->u.match_size
				   + target->u.target_size);
	if(!tmp) {
		bb_error_msg("%s: realloc(%d) error", "addnatrule",
		       (int)(sizeof(struct ipt_entry) + match->u.match_size + target->u.target_size));
		free(e);
		free(match);
		free(target);
		return -1;
	}
	e = tmp;
	memcpy(e->elems, match, match->u.match_size);
	memcpy(e->elems + match->u.match_size, target, target->u.target_size);
	e->target_offset = sizeof(struct ipt_entry)
			   + match->u.match_size;
	e->next_offset = sizeof(struct ipt_entry)
			 + match->u.match_size
					 + target->u.target_size;
	/* remote host */
	if(rhost && (rhost[0] != '\0') && (0 != strcmp(rhost, "*"))) {
		e->ip.src.s_addr = inet_addr(rhost);
		e->ip.smsk.s_addr = INADDR_NONE;
	}

	r = iptc_init_verify_and_append("nat", bb_common_bss.miniupnpd_nat_chain, e, "addnatrule");
	free(target);
	free(match);
	free(e);
	return r;
}

/* for "Port Triggering"
 * Section 2.5.16 figure 2.2 in UPnP-gw-WANIPConnection-v2-Service.pdf
 *
 *   When a control point creates a port forwarding rule with AddPortMapping()
 *   action for inbound traffic , this rule MUST also be applied when NAT port
 *   triggering occurs for outbound traffic.
 *
 * iptables -t nat -A MINIUPNPD-POSTROUTING {-o <extif>} -s <iaddr>
 *          -p <proto> [-d <rhost>] --sport <iport> -j MASQUERADE --to-ports <eport>
 */
#ifdef ENABLE_PORT_TRIGGERING
static int
addmasqueraderule(int proto,
	   unsigned short eport,
	   const char * iaddr, unsigned short iport,
	   const char * rhost/*, const char * extif*/)
{
	int r = 0;
	struct ipt_entry * e;
	struct ipt_entry * tmp;
	struct ipt_entry_match *match = NULL;
	struct ipt_entry_target *target = NULL;

	e = calloc(1, sizeof(struct ipt_entry));
	if(!e) {
		bb_error_msg("%s: calloc(%d) error", "addmasqueraderule", (int)sizeof(struct ipt_entry));
		return -1;
	}
	e->ip.proto = proto;
	if(proto == IPPROTO_TCP) {
		match = get_tcp_match(0, iport);
	} else {
		match = get_udp_match(0, iport);
	}
	e->nfcache = NFC_IP_DST_PT;
	target = get_masquerade_target(eport);
	e->nfcache |= NFC_UNKNOWN;
	tmp = realloc(e, sizeof(struct ipt_entry)
		       + match->u.match_size
				   + target->u.target_size);
	if(!tmp) {
		bb_error_msg("%s: realloc(%d) error", "addmasqueraderule",
		       (int)(sizeof(struct ipt_entry) + match->u.match_size + target->u.target_size));
		free(e);
		free(match);
		free(target);
		return -1;
	}
	e = tmp;
	memcpy(e->elems, match, match->u.match_size);
	memcpy(e->elems + match->u.match_size, target, target->u.target_size);
	e->target_offset = sizeof(struct ipt_entry)
			   + match->u.match_size;
	e->next_offset = sizeof(struct ipt_entry)
			 + match->u.match_size
					 + target->u.target_size;
	/* internal host */
	if(iaddr && (iaddr[0] != '\0') && (0 != strcmp(iaddr, "*")))
	{
		e->ip.src.s_addr = inet_addr(iaddr);
		e->ip.smsk.s_addr = INADDR_NONE;
	}
	/* remote host */
	if(rhost && (rhost[0] != '\0') && (0 != strcmp(rhost, "*"))) {
		e->ip.dst.s_addr = inet_addr(rhost);
		e->ip.dmsk.s_addr = INADDR_NONE;
	}

	r = iptc_init_verify_and_append("nat", bb_common_bss.miniupnpd_nat_postrouting_chain, e, "addmasqueraderule");
	free(target);
	free(match);
	free(e);
	return r;
}
#endif /* ENABLE_PORT_TRIGGERING */

/* ================================= */
static struct ipt_entry_target *
get_accept_target(void)
{
	struct ipt_entry_target * target = NULL;
	size_t size;
	size =   IPT_ALIGN(sizeof(struct ipt_entry_target))
	       + IPT_ALIGN(sizeof(int));
	target = calloc(1, size);
	target->u.user.target_size = size;
	safe_strncpy(target->u.user.name, "ACCEPT", sizeof(target->u.user.name));
	return target;
}

/* add_filter_rule()
 * iptables -t filter -A MINIUPNPD [-s <rhost>] -p <proto> -d <iaddr> --dport <iport> -j ACCEPT */
static int
add_filter_rule(int proto, const char * rhost,
		const char * iaddr, unsigned short iport)
{
	int r = 0;
	struct ipt_entry * e;
	struct ipt_entry * tmp;
	struct ipt_entry_match *match = NULL;
	struct ipt_entry_target *target = NULL;

	e = calloc(1, sizeof(struct ipt_entry));
	if(!e) {
		bb_error_msg("%s: calloc(%d) error", "add_filter_rule", (int)sizeof(struct ipt_entry));
		return -1;
	}
	e->ip.proto = proto;
	if(proto == IPPROTO_TCP) {
		match = get_tcp_match(iport,0);
	} else {
		match = get_udp_match(iport,0);
	}
	e->nfcache = NFC_IP_DST_PT;
	e->ip.dst.s_addr = inet_addr(iaddr);
	e->ip.dmsk.s_addr = INADDR_NONE;
	target = get_accept_target();
	e->nfcache |= NFC_UNKNOWN;
	tmp = realloc(e, sizeof(struct ipt_entry)
		       + match->u.match_size
				   + target->u.target_size);
	if(!tmp) {
		bb_error_msg("%s: realloc(%d) error", "add_filter_rule",
		       (int)(sizeof(struct ipt_entry) + match->u.match_size + target->u.target_size));
		free(e);
		free(match);
		free(target);
		return -1;
	}
	e = tmp;
	memcpy(e->elems, match, match->u.match_size);
	memcpy(e->elems + match->u.match_size, target, target->u.target_size);
	e->target_offset = sizeof(struct ipt_entry)
			   + match->u.match_size;
	e->next_offset = sizeof(struct ipt_entry)
			 + match->u.match_size
					 + target->u.target_size;
	/* remote host */
	if(rhost && (rhost[0] != '\0') && (0 != strcmp(rhost, "*")))
	{
		e->ip.src.s_addr = inet_addr(rhost);
		e->ip.smsk.s_addr = INADDR_NONE;
	}

	r = iptc_init_verify_and_append("filter", bb_common_bss.miniupnpd_forward_chain, e, "add_filter_rule");
	free(target);
	free(match);
	free(e);
	return r;
}

/* return an (malloc'ed) array of "external" port for which there is
 * a port mapping. number is the size of the array */
static unsigned short *
get_portmappings_in_range(unsigned short startport, unsigned short endport,
			  int proto, unsigned int * number)
{
	unsigned short * array;
	unsigned int capacity;
	unsigned short eport;
	IPTC_HANDLE h;
	const struct ipt_entry * e;
	const struct ipt_entry_match *match;

	*number = 0;
	capacity = 128;
	array = calloc(capacity, sizeof(unsigned short));
	if(!array)
	{
		bb_error_msg("%s() : calloc error", "get_portmappings_in_range");
		return NULL;
	}

	h = iptc_init("nat");
	if(!h)
	{
		bb_error_msg("%s() : iptc_init() failed : %s", "get_portmappings_in_range", iptc_strerror(errno));
		free(array);
		return NULL;
	}
	if(!iptc_is_chain(bb_common_bss.miniupnpd_nat_chain, h))
	{
		bb_error_msg("chain %s not found", bb_common_bss.miniupnpd_nat_chain);
		free(array);
		array = NULL;
	}
	else
	{
		for(e = iptc_first_rule(bb_common_bss.miniupnpd_nat_chain, h); e; e = iptc_next_rule(e, h))
		{
			if(proto == e->ip.proto)
			{
				match = (const struct ipt_entry_match *)&e->elems;
				if(0 == strncmp(match->u.user.name, "tcp", IPT_FUNCTION_MAXNAMELEN))
				{
					const struct ipt_tcp * info;
					info = (const struct ipt_tcp *)match->data;
					eport = info->dpts[0];
				}
				else
				{
					const struct ipt_udp * info;
					info = (const struct ipt_udp *)match->data;
					eport = info->dpts[0];
				}
				if(startport <= eport && eport <= endport)
				{
					if(*number >= capacity)
					{
						unsigned short * tmp;
						/* need to increase the capacity of the array */
						tmp = realloc(array, sizeof(unsigned short)*capacity);
						if(!tmp)
						{
							bb_error_msg("get_portmappings_in_range() : realloc(%u) error",
							       (unsigned)sizeof(unsigned short)*capacity);
							*number = 0;
							free(array);
							array = NULL;
							break;
						}
						array = tmp;
					}
					array[*number] = eport;
					(*number)++;
				}
			}
		}
	}
	if(h)
		iptc_free(h);
	return array;
}

/* update the port mapping decription and timestamp */
static int
update_portmapping_desc_timestamp(const char * ifname,
		   unsigned short eport, int proto,
		   const char * desc, unsigned int timestamp)
{
	UNUSED(ifname);
	del_redirect_desc(eport, proto);
	add_redirect_desc(eport, proto, desc, timestamp);
	return 0;
}

static int
update_rule_and_commit(const char * table, const char * chain,
		       unsigned index, const struct ipt_entry *e)
{
	IPTC_HANDLE h;
	int r = 0;

	h = iptc_init(table);
	if(!h) {
		bb_error_msg("%s() : iptc_init() failed : %s", "update_rule_and_commit", iptc_strerror(errno));
		return -1;
	}
	if(!iptc_replace_entry(chain, e, index, h)) {
		bb_error_msg("%s(): iptc_replace_entry: %s", "update_rule_and_commit", iptc_strerror(errno));
		r = -1;
	} else if(!iptc_commit(h)) {
		bb_error_msg("%s(): iptc_commit: %s", "update_rule_and_commit", iptc_strerror(errno));
		r = -1;
	}
	iptc_free(h);
	return r;
}

/* update the port mapping internal port, decription and timestamp */
static int
update_portmapping(const char * ifname, unsigned short eport, int proto,
		   unsigned short iport, const char * desc,
		   unsigned int timestamp)
{
	int r = 0;
	int found = 0;
	unsigned index = 0;
	unsigned i = 0;
	IPTC_HANDLE h;
	const struct ipt_entry * e;
	struct ipt_entry * new_e = NULL;
	size_t entry_len;
	const struct ipt_entry_target * target;
	struct ip_nat_multi_range * mr;
	const struct ipt_entry_match *match;
	uint32_t iaddr = 0;
	unsigned short old_iport = 0;

	h = iptc_init("nat");
	if(!h) {
		bb_error_msg("%s() : iptc_init() failed : %s", "update_portmapping", iptc_strerror(errno));
		return -1;
	}
	/* First step : find the right nat rule */
	if(!iptc_is_chain(bb_common_bss.miniupnpd_nat_chain, h)) {
		bb_error_msg("chain %s not found", bb_common_bss.miniupnpd_nat_chain);
		r = -1;
	} else {
		for(e = iptc_first_rule(bb_common_bss.miniupnpd_nat_chain, h); e; e = iptc_next_rule(e, h), i++) {
			if(proto==e->ip.proto) {
				match = (const struct ipt_entry_match *)&e->elems;
				if(0 == strncmp(match->u.user.name, "tcp", IPT_FUNCTION_MAXNAMELEN)) {
					const struct ipt_tcp * info;
					info = (const struct ipt_tcp *)match->data;
					if(eport != info->dpts[0])
						continue;
				} else {
					const struct ipt_udp * info;
					info = (const struct ipt_udp *)match->data;
					if(eport != info->dpts[0])
						continue;
				}
				/* we found the right rule */
				found = 1;
				index = i;
				target = (void *)e + e->target_offset;
				mr = (struct ip_nat_multi_range *)&target->data[0];
				iaddr = mr->range[0].min_ip;
				old_iport = ntohs(mr->range[0].min.all);
				entry_len = sizeof(struct ipt_entry) + match->u.match_size + target->u.target_size;
				new_e = malloc(entry_len);
				if(new_e == NULL) {
					bb_error_msg("%s: malloc(%u) error", "update_portmapping", (unsigned)entry_len);
					r = -1;
				} else {
					memcpy(new_e, e, entry_len);
				}
				break;
			}
		}
	}
	iptc_free(h);
	if(!found || r < 0)
		return -1;
	DEBUG_LOG(syslog(LOG_INFO, "Trying to update nat rule at index %u", index);)
	target = (void *)new_e + new_e->target_offset;
	mr = (struct ip_nat_multi_range *)&target->data[0];
	mr->range[0].min.all = mr->range[0].max.all = htons(iport);
	/* first update the nat rule */
	r = update_rule_and_commit("nat", bb_common_bss.miniupnpd_nat_chain, index, new_e);
	free(new_e); new_e = NULL;
	if(r < 0)
		return r;

	/* update filter rule */
	h = iptc_init("filter");
	if(!h) {
		bb_error_msg("%s() : iptc_init() failed : %s", "update_portmapping", iptc_strerror(errno));
		return -1;
	}
	i = 0; found = 0;
	if(!iptc_is_chain(bb_common_bss.miniupnpd_forward_chain, h)) {
		bb_error_msg("chain %s not found", bb_common_bss.miniupnpd_forward_chain);
	} else {
		/* we must find the right index for the filter rule */
		for(e = iptc_first_rule(bb_common_bss.miniupnpd_forward_chain, h); e; e = iptc_next_rule(e, h), i++) {
			if(proto!=e->ip.proto)
				continue;
			target = (void *)e + e->target_offset;
			match = (const struct ipt_entry_match *)&e->elems;
			if(0 == strncmp(match->u.user.name, "tcp", IPT_FUNCTION_MAXNAMELEN)) {
				const struct ipt_tcp * info;
				info = (const struct ipt_tcp *)match->data;
				if(old_iport != info->dpts[0])
					continue;
			} else {
				const struct ipt_udp * info;
				info = (const struct ipt_udp *)match->data;
				if(old_iport != info->dpts[0])
					continue;
			}
			if(iaddr != e->ip.dst.s_addr)
				continue;
			index = i;
			found = 1;
			entry_len = sizeof(struct ipt_entry) + match->u.match_size + target->u.target_size;
			new_e = malloc(entry_len);
			if(new_e == NULL) {
				bb_error_msg("%s: malloc(%u) error", "update_portmapping", (unsigned)entry_len);
				r = -1;
			} else {
				memcpy(new_e, e, entry_len);
			}
			break;
		}
	}
	iptc_free(h);
	if(!found || r < 0)
		return -1;
	DEBUG_LOG(syslog(LOG_INFO, "Trying to update filter rule at index %u", index);)
	match = (struct ipt_entry_match *)&new_e->elems;
	if(0 == strncmp(match->u.user.name, "tcp", IPT_FUNCTION_MAXNAMELEN)) {
		struct ipt_tcp * info;
		info = (struct ipt_tcp *)match->data;
		info->dpts[0] = info->dpts[1] = iport;
	} else {
		struct ipt_udp * info;
		info = (struct ipt_udp *)match->data;
		info->dpts[0] = info->dpts[1] = iport;
	}
	r = update_rule_and_commit("filter", bb_common_bss.miniupnpd_forward_chain, index, new_e);
	free(new_e); new_e = NULL;
	if(r < 0)
		return r;

	return update_portmapping_desc_timestamp(ifname, eport, proto, desc, timestamp);
}

struct rule_state
{
	u_int64_t packets;
	u_int64_t bytes;
	struct rule_state * next;
	unsigned short eport;
	unsigned char proto;
	unsigned char to_remove;
};


/* from <inttypes.h> */
#ifndef PRIu64
#define PRIu64 "llu"
#endif

/* proto_atoi()
 * convert the string "UDP" or "TCP" to IPPROTO_UDP and IPPROTO_UDP */
static int proto_atoi(const char * protocol)
{
	int proto = IPPROTO_TCP;
	if(strcasecmp(protocol, "UDP") == 0)
		proto = IPPROTO_UDP;
#ifdef IPPROTO_UDPLITE
	else if(strcasecmp(protocol, "UDPLITE") == 0)
		proto = IPPROTO_UDPLITE;
#endif /* IPPROTO_UDPLITE */
	return proto;
}

#ifdef DEBUG
/* proto_itoa()
 * convert IPPROTO_UDP, IPPROTO_UDP, etc. to "UDP", "TCP" */
static const char *proto_itoa(int proto)
{
	const char * protocol;
	switch(proto) {
	case IPPROTO_UDP:
		protocol = "UDP";
		break;
	case IPPROTO_TCP:
		protocol = "TCP";
		break;
#ifdef IPPROTO_UDPLITE
	case IPPROTO_UDPLITE:
		protocol = "UDPLITE";
		break;
#endif /* IPPROTO_UDPLITE */
	default:
		protocol = "*UNKNOWN*";
	}
	return protocol;
}
#endif


/* UPnP permission rule samples:
 * allow 1024-65535 192.168.3.0/24 1024-65535
 * deny 0-65535 192.168.1.125/32 0-65535 */
struct upnpperm {
	enum {UPNPPERM_ALLOW=1, UPNPPERM_DENY=2 } type;
				/* is it an allow or deny permission rule ? */
	u_short eport_min, eport_max;   /* external port range */
	struct in_addr address, mask;   /* ip/mask */
	u_short iport_min, iport_max;   /* internal port range */
};

static const char *strtol_port(const char *s, u_short *p)
{
	long i;
	char *e;

	errno = 0;
	i = strtol(s, &e, 0);
	if(errno != 0 || s == e || i < 0 || i > 65535)
		return NULL;
	*p = (u_short)i;
	return e;
}

/* " [0-9]+(-[0-9]+) " */
static const char *read_permission_ports(const char *p, u_short *pmin, u_short *pmax)
{
	p = strtol_port(p, pmin);
	if(p == NULL)
		return NULL;
	if(*p == '-') {
		p = strtol_port(p + 1, pmax);
		if(p == NULL || *pmin > *pmax)
			return NULL;
	} else if(*p == '\0') {
		*pmax = *pmin;
	} else
		return NULL;
	return *p == '\0' ? p : NULL;
}

/* read_permission_line()
 * parse the a permission line which format is :
 * (deny|allow) [0-9]+(-[0-9]+) ip/mask [0-9]+(-[0-9]+)
 * ip/mask is either 192.168.1.1/24 or 192.168.1.1/255.255.255.0
 */
static struct upnpperm *read_permission_line(char *tokens[])
{
	char *q;
	struct upnpperm perm;

	if(0 == strcasecmp(tokens[0], "allow"))
		perm.type = UPNPPERM_ALLOW;
	else if(0 == strcasecmp(tokens[0], "deny"))
		perm.type = UPNPPERM_DENY;
	else
		return NULL;
	/* second token: eport or eport_min-eport_max */
	if(read_permission_ports(tokens[1], &perm.eport_min, &perm.eport_max) == NULL)
		return NULL;
	/* third token:  ip/mask */
	q = tokens[2];
	if(!isdigit(*q))
		return NULL;
	for(; isdigit(*q) || (*q == '.'); q++)
		;
	if(*q=='/') {
		char *p;

		*q = '\0';
		if(!inet_aton(tokens[2], &perm.address))
			return NULL;
		q++;
		p = q;
		while(isdigit(*q))
			q++;
		if(*q == '.') {
			while(*q == '.' || isdigit(*q))
				q++;
			if(*q != '\0')
				return NULL;
			if(!inet_aton(p, &perm.mask))
				return NULL;
		} else if(*q != '\0')
			return NULL;
		else {
			int n_bits = atoi(p);
			if(n_bits > 32)
				return NULL;
			perm.mask.s_addr = htonl(n_bits ? (0xffffffffu << (32 - n_bits)) : 0);
		}
	} else if(*q == '\0') {
		if(!inet_aton(tokens[2], &perm.address))
			return NULL;
		perm.mask.s_addr = 0xffffffffu;
	} else {
		return NULL;
	}
	/* fourth token: iport or iport_min-iport_max */
	if(read_permission_ports(tokens[3], &perm.iport_min, &perm.iport_max) == NULL)
		return NULL;
	return memcpy(xmalloc(sizeof(struct upnpperm)), &perm, sizeof(struct upnpperm));
}

/* match_permission()
 * returns: 1 if eport, address, iport matches the permission rule
 *          0 if no match */
static int
match_permission(const struct upnpperm * perm,
		 u_short eport, struct in_addr address, u_short iport)
{
	if( (eport < perm->eport_min) || (perm->eport_max < eport))
		return 0;
	if( (iport < perm->iport_min) || (perm->iport_max < iport))
		return 0;
	if( (address.s_addr & perm->mask.s_addr)
	   != (perm->address.s_addr & perm->mask.s_addr) )
		return 0;
	return 1;
}

/* returns: 0 if the upnp rule should be rejected,
 *          1 if it could be accepted */
static int
check_upnp_rule_against_permissions(u_short eport, struct in_addr address, u_short iport)
{
	int i;
	llist_t *pp;

	for(i = 0, pp = bb_common_bss.upnppermlist_list_t; pp; pp = pp->link, i++) {
		const struct upnpperm *permary = pp->data;
		if(match_permission(permary, eport, address, iport)) {
			DEBUG_LOG(syslog(LOG_DEBUG,
			       "UPnP permission rule %d matched : port mapping %s",
			       i, (permary->type == UPNPPERM_ALLOW)?"accepted":"rejected");)
			return (permary->type == UPNPPERM_ALLOW);
		}
	}
	DEBUG_LOG(syslog(LOG_DEBUG, "no permission rule matched : accept by default");)
	return 1;       /* Default : accept */
}

#ifdef ENABLE_EVENTS

enum subscriber_service_enum {
 EWanCFG = 1,
 EWanIPC,
#ifdef ENABLE_L3F_SERVICE
 EL3F,
#endif
};

/* stuctures definitions */
struct subscriber {
	LIST_ENTRY(subscriber) entries;
	struct upnp_event_notify * notify;
	time_t timeout;
	uint32_t seq;
	enum subscriber_service_enum service;
	char uuid[42];
	char callback[];
};

struct upnp_event_notify
{
	LIST_ENTRY(upnp_event_notify) entries;
	int s;  /* socket */
	enum { ECreated=1,
	       EConnecting,
	       ESending,
	       EWaitingForResponse,
	       EFinished,
	       EError } state;
	struct subscriber * sub;
	char * buffer;
	int buffersize;
	int tosend;
	int sent;
	const char * path;
#if ENABLE_FEATURE_IPV6
	int ipv6;
	char addrstr[48];
#else
	char addrstr[16];
#endif
	char portstr[8];
};

/* create and add the notify object to the list */
static void
upnp_event_create_notify(struct subscriber * sub)
{
	struct upnp_event_notify *obj;
	/*struct timeval sock_timeout;*/

	obj = calloc(1, sizeof(struct upnp_event_notify));
	if(!obj) {
		bb_perror_msg("%s: calloc()", "upnp_event_create_notify");
		return;
	}
	obj->sub = sub;
	obj->state = ECreated;
#if ENABLE_FEATURE_IPV6
	obj->s = bb_xsocket((obj->sub->callback[7] == '[') ? BB_XSOCKET_FLG_AF_INET6 : 0);
#else
	obj->s = bb_xsocket(0);
#endif
	if(obj->s < 0)
		goto error;
	/* set socket non blocking */
	if(!set_non_blocking(obj->s)) {
		bb_perror_msg("%s: set_non_blocking()", "upnp_event_create_notify");
		goto error;
	}
	if(sub)
		sub->notify = obj;
	LIST_INSERT_HEAD(&notifylist, obj, entries);
	return;
error:
	if(obj->s >= 0)
		close(obj->s);
	free(obj);
}

/* notifies all subscriber of a number of port mapping change
 * or external ip address change */
static void
upnp_event_var_change_notify(enum subscriber_service_enum service)
{
	struct subscriber * sub;
	for(sub = subscriberlist.lh_first; sub != NULL; sub = sub->entries.le_next) {
		if(sub->service == service && sub->notify == NULL)
			upnp_event_create_notify(sub);
	}
}

static int
add_redirect_rule2(const char * rhost, unsigned short eport,
		   const char * iaddr, unsigned short iport, int proto,
				   const char * desc, unsigned int timestamp)
{
	int r;

	r = addnatrule(proto, eport, iaddr, iport, rhost);
	if(r >= 0) {
		add_redirect_desc(eport, proto, desc, timestamp);
#ifdef ENABLE_PORT_TRIGGERING
		/* http://www.netfilter.org/documentation/HOWTO/NAT-HOWTO-6.html#ss6.3
		 * The default behavior is to alter the connection as little
		 * as possible, within the constraints of the rule given by
		 * the user.
		 * This means we won't remap ports unless we have to. */
		if(iport != eport) {
			/* TODO : check if this should be done only with UDP */
			r = addmasqueraderule(proto, eport, iaddr, iport, rhost/*, ifname*/);
#ifdef DEBUG
			if(r < 0) {
				syslog(LOG_NOTICE, "add_redirect_rule2(): addmasqueraderule returned %d", r);
			}
#endif
		}
#endif /* ENABLE_PORT_TRIGGERING */
	}
	return r;
}

/* same as upnp_redirect() without any check */
static int
upnp_redirect_internal(const char * rhost, unsigned short eport,
		       const char * iaddr, unsigned short iport,
		       int proto, const char * desc,
		       unsigned int timestamp)
{
	if(add_redirect_rule2(rhost, eport, iaddr, iport, proto, desc, timestamp) < 0)
		return -1;

	if(add_filter_rule(proto, rhost, iaddr, iport) < 0) {
		/* clean up the redirect rule */
		return -1;
	}
	if(timestamp > 0) {
		if(!bb_common_bss.nextruletoclean_timestamp || (timestamp < bb_common_bss.nextruletoclean_timestamp))
			bb_common_bss.nextruletoclean_timestamp = timestamp;
	}
#ifdef ENABLE_EVENTS
	/* the number of port mappings changed, we must
	 * inform the subscribers */
	upnp_event_var_change_notify(EWanIPC);
#endif
	return 0;
}

/* upnp_redirect()
 * calls OS/fw dependant implementation of the redirection.
 * protocol should be the string "TCP" or "UDP"
 * returns: 0 on success
 *          -1 failed to redirect
 *          -2 already redirected
 *          -3 permission check failed
 *          -4 already redirected (other mechanism)
 */
static int
upnp_redirect(const char * rhost, unsigned short eport,
	      const char * iaddr, unsigned short iport,
	      const char * protocol, const char * desc,
	      unsigned int leaseduration)
{
	int proto, r;
	char iaddr_old[32];
	char rhost_old[32];
	unsigned short iport_old;
	struct in_addr address;
	unsigned int timestamp;

	proto = proto_atoi(protocol);
	if(inet_aton(iaddr, &address) <= 0) {
		bb_error_msg("inet_aton(%s) FAILED", iaddr);
		return -1;
	}

	if(!check_upnp_rule_against_permissions(eport, address, iport)) {
		DEBUG_LOG(syslog(LOG_INFO, "redirection permission check failed for "
				 "%hu->%s:%hu %s", eport, iaddr, iport, protocol);)
		return -3;
	}
	/* IGDv1 (WANIPConnection:1 Service Template Version 1.01 / Nov 12, 2001)
	 * - 2.2.20.PortMappingDescription :
	 *  Overwriting Previous / Existing Port Mappings:
	 * If the RemoteHost, ExternalPort, PortMappingProtocol and InternalClient
	 * are exactly the same as an existing mapping, the existing mapping values
	 * for InternalPort, PortMappingDescription, PortMappingEnabled and
	 * PortMappingLeaseDuration are overwritten.
	 *  Rejecting a New Port Mapping:
	 * In cases where the RemoteHost, ExternalPort and PortMappingProtocol
	 * are the same as an existing mapping, but the InternalClient is
	 * different, the action is rejected with an appropriate error.
	 *  Add or Reject New Port Mapping behavior based on vendor implementation:
	 * In cases where the ExternalPort, PortMappingProtocol and InternalClient
	 * are the same, but RemoteHost is different, the vendor can choose to
	 * support both mappings simultaneously, or reject the second mapping
	 * with an appropriate error.
	 *
	 * - 2.4.16.AddPortMapping
	 * This action creates a new port mapping or overwrites an existing
	 * mapping with the same internal client. If the ExternalPort and
	 * PortMappingProtocol pair is already mapped to another internal client,
	 * an error is returned.
	 *
	 * IGDv2 (WANIPConnection:2 Service Standardized DCP (SDCP) Sep 10, 2010)
	 * Protocol ExternalPort RemoteHost InternalClient Result
	 *     =         =           ≠           ≠         Failure
	 *     =         =           ≠           =         Failure or success
	 *                                                 (vendor specific)
	 *     =         =           =           ≠         Failure
	 *     =         =           =           =         Success (overwrite)
	 */
	rhost_old[0] = '\0';
	r = get_redirect_rule(eport, proto, iaddr_old, sizeof(iaddr_old), &iport_old, 0, 0,
			      rhost_old, sizeof(rhost_old),
			      &timestamp, 0, 0);
	if(r == 0) {
		if(strcmp(iaddr, iaddr_old)==0 &&
		   ((rhost == NULL && rhost_old[0]=='\0') ||
		    (rhost && (strcmp(rhost, "*") == 0) && rhost_old[0]=='\0') ||
		    (rhost && (strcmp(rhost, rhost_old) == 0)))) {
			DEBUG_LOG(syslog(LOG_INFO, "updating existing port mapping %hu %s (rhost '%s') => %s:%hu",
				eport, protocol, rhost_old, iaddr_old, iport_old);)
			timestamp = (leaseduration > 0) ? time(NULL) + leaseduration : 0;
			if(iport != iport_old) {
				r = update_portmapping(bb_common_bss.ext_if_name, eport, proto, iport, desc, timestamp);
			} else {
				r = update_portmapping_desc_timestamp(bb_common_bss.ext_if_name, eport, proto, desc, timestamp);
			}
			return r;
		} else {
			DEBUG_LOG(syslog(LOG_INFO, "port %hu %s (rhost '%s') already redirected to %s:%hu",
				eport, protocol, rhost_old, iaddr_old, iport_old);)
			return -2;
		}
	} else {
		timestamp = (leaseduration > 0) ? time(NULL) + leaseduration : 0;
		DEBUG_LOG(syslog(LOG_INFO, "redirecting port %hu to %s:%hu protocol %s for: %s",
			eport, iaddr, iport, protocol, desc);)
		return upnp_redirect_internal(rhost, eport, iaddr, iport, proto,
					      desc, timestamp);
	}
}

/* Firewall independant code which call the FW dependant code.
 * returns : 0 on success
 *           -1 failed to get the port mapping entry or no entry exists */
static int
upnp_get_redirection_infos(unsigned short eport, const char * protocol,
			   unsigned short * iport,
			   char * iaddr, int iaddrlen,
			   char * desc, int desclen,
			   char * rhost, int rhostlen,
			   unsigned int * leaseduration)
{
	int r;
	unsigned int timestamp;
	time_t current_time;

	if(desc && (desclen > 0))
		desc[0] = '\0';
	if(rhost && (rhostlen > 0))
		rhost[0] = '\0';
	r = get_redirect_rule(eport, proto_atoi(protocol), iaddr, iaddrlen, iport, desc, desclen,
			      rhost, rhostlen, &timestamp,
			      0, 0);
	if(r == 0 &&
	   timestamp > 0 &&
	   timestamp > (unsigned int)(current_time = time(NULL))) {
		*leaseduration = timestamp - current_time;
	} else {
		*leaseduration = 0;
	}
	return r;
}

/* returns : 0 on success
 *           -1 failed to get the port mapping or index out of range */
static int
upnp_get_redirection_infos_by_index(int index,
				    unsigned short * eport, char * protocol,
				    unsigned short * iport,
				    char * iaddr, int iaddrlen,
				    char * desc, int desclen,
				    char * rhost, int rhostlen,
				    unsigned int * leaseduration)
{
	/*char ifname[IFNAMSIZ];*/
	int proto = 0;
	unsigned int timestamp;
	time_t current_time;

	if(desc && (desclen > 0))
		desc[0] = '\0';
	if(rhost && (rhostlen > 0))
		rhost[0] = '\0';
	if(get_redirect_rule_by_index(index, 0/*ifname*/, eport, iaddr, iaddrlen,
				      iport, &proto, desc, desclen,
				      rhost, rhostlen, &timestamp,
				      0, 0) < 0)
		return -1;
	else {
		current_time = time(NULL);
		*leaseduration = (timestamp > (unsigned int)current_time)
				 ? (timestamp - current_time)
				 : 0;
		if(proto == IPPROTO_TCP)
			memcpy(protocol, "TCP", 4);
#ifdef IPPROTO_UDPLITE
		else if(proto == IPPROTO_UDPLITE)
			memcpy(protocol, "UDPLITE", 8);
#endif /* IPPROTO_UDPLITE */
		else
			memcpy(protocol, "UDP", 4);
		return 0;
	}
}

/* called from natpmp.c too */
static int
_upnp_delete_redir(unsigned short eport, int proto)
{
	int r;
	r = delete_redirect_and_filter_rules(eport, proto);

#ifdef ENABLE_EVENTS
	upnp_event_var_change_notify(EWanIPC);
#endif
	return r;
}

static int
upnp_delete_redirection(unsigned short eport, const char * protocol)
{
	DEBUG_LOG(syslog(LOG_INFO, "removing redirect rule port %hu %s", eport, protocol);)
	return _upnp_delete_redir(eport, proto_atoi(protocol));
}

/* return the number of port mapping entries
 * TODO: improve this code. */
static int
upnp_get_portmapping_number_of_entries(void)
{
	int n = 0, r = 0;
	unsigned short eport, iport;
	char protocol[4], iaddr[32], desc[64], rhost[32];
	unsigned int leaseduration;
	do {
		protocol[0] = '\0'; iaddr[0] = '\0'; desc[0] = '\0';
		r = upnp_get_redirection_infos_by_index(n, &eport, protocol, &iport,
							iaddr, sizeof(iaddr),
							desc, sizeof(desc),
							rhost, sizeof(rhost),
							&leaseduration);
		n++;
	} while(r==0);
	return (n-1);
}

/* functions used to remove unused rules
 * As a side effect, delete expired rules (based on LeaseDuration)
 * return a linked list of all rules
 * or an empty list if there are not enough
 * As a "side effect", delete rules which are expired */
static struct rule_state *
get_upnp_rules_state_list(int max_rules_number_target)
{
	int proto;
	unsigned short iport;
	unsigned int timestamp;
	struct rule_state * tmp;
	struct rule_state * list = 0;
	struct rule_state * * p;
	int i = 0;
	time_t current_time;

	/*ifname[0] = '\0';*/
	tmp = malloc(sizeof(struct rule_state));
	if(!tmp)
		return 0;
	current_time = time(NULL);
	bb_common_bss.nextruletoclean_timestamp = 0;
	while(get_redirect_rule_by_index(i, /*ifname*/0, &tmp->eport, 0, 0,
				      &iport, &proto, 0, 0, 0,0, &timestamp,
								  &tmp->packets, &tmp->bytes) >= 0)
	{
		tmp->to_remove = 0;
		if(timestamp > 0) {
			/* need to remove this port mapping ? */
			if(timestamp <= (unsigned int)current_time)
				tmp->to_remove = 1;
			else if((bb_common_bss.nextruletoclean_timestamp <= (unsigned int)current_time)
			       || (timestamp < bb_common_bss.nextruletoclean_timestamp))
				bb_common_bss.nextruletoclean_timestamp = timestamp;
		}
		tmp->proto = (short)proto;
		/* add tmp to list */
		tmp->next = list;
		list = tmp;
		/* prepare next iteration */
		i++;
		tmp = malloc(sizeof(struct rule_state));
		if(!tmp)
			break;
	}
	free(tmp);
	/* remove the redirections that need to be removed */
	for(p = &list, tmp = list; tmp; tmp = *p) {
		if(tmp->to_remove) {
			DEBUG_LOG(syslog(LOG_NOTICE, "remove port mapping %hu %s because it has expired",
			       tmp->eport, proto_itoa(tmp->proto));)
			_upnp_delete_redir(tmp->eport, tmp->proto);
			*p = tmp->next;
			free(tmp);
			i--;
		} else {
			p = &(tmp->next);
		}
	}
	/* return empty list if not enough redirections */
	if(i<=max_rules_number_target)
		while(list) {
			tmp = list;
			list = tmp->next;
			free(tmp);
		}
	/* return list */
	return list;
}

static void remove_unused_rules(struct rule_state * list)
{
	unsigned short iport;
	struct rule_state * tmp;
	u_int64_t packets;
	u_int64_t bytes;
	unsigned int timestamp;
	int n = 0;

	while(list) {
		/* remove the rule if no traffic has used it */
		if(get_redirect_rule(list->eport, list->proto, 0, 0, &iport, 0, 0, 0, 0, &timestamp,
				     &packets, &bytes) >= 0) {
			if(packets == list->packets && bytes == list->bytes) {
				DEBUG_LOG(syslog(LOG_DEBUG, "removing unused mapping %hu %s : still "
				       "%" PRIu64 "packets %" PRIu64 "bytes",
				       list->eport, proto_itoa(list->proto),
				       packets, bytes);)
				_upnp_delete_redir(list->eport, list->proto);
				n++;
			}
		}
		tmp = list;
		list = tmp->next;
		free(tmp);
	}
#ifdef DEBUG
	if(n>0)
		syslog(LOG_NOTICE, "removed %d unused rules", n);
#endif
}

/* return a list of all "external" ports for which a port
 * mapping exists */
static unsigned short *
upnp_get_portmappings_in_range(unsigned short startport,
			       unsigned short endport,
			       const char * protocol,
			       unsigned int * number)
{
	int proto;
	proto = proto_atoi(protocol);
	if(!number)
		return NULL;
	return get_portmappings_in_range(startport, endport, proto, number);
}


/* renew a subscription (update the timeout) */
static const char *
upnpevents_renewSubscription(const char * sid, int sidlen, int timeout)
{
	struct subscriber * sub;
	for(sub = subscriberlist.lh_first; sub != NULL; sub = sub->entries.le_next) {
		if((sidlen == 41) && (memcmp(sid, sub->uuid, 41) == 0)) {
			sub->timeout = (timeout ? time(NULL) + timeout : 0);
			return sub->uuid;
		}
	}
	return NULL;
}

static int upnpevents_removeSubscriber(const char * sid, int sidlen)
{
	struct subscriber * sub;
	if(!sid)
		return -1;
	for(sub = subscriberlist.lh_first; sub != NULL; sub = sub->entries.le_next) {
		if((sidlen == 41) && (memcmp(sid, sub->uuid, 41) == 0)) {
			if(sub->notify) {
				sub->notify->sub = NULL;
			}
			LIST_REMOVE(sub, entries);
			free(sub);
			return 0;
		}
	}
	return -1;
}


static void
upnp_event_notify_connect(struct upnp_event_notify * obj)
{
	unsigned int i;
	const char * p;
	unsigned short port;
#if ENABLE_FEATURE_IPV6
	struct sockaddr_storage addr;
	socklen_t addrlen;
#else
	struct sockaddr_in addr;
	socklen_t addrlen;
#endif

	if(!obj)
		return;
	memset(&addr, 0, sizeof(addr));
	i = 0;
	if(obj->sub == NULL) {
		obj->state = EError;
		return;
	}
	p = obj->sub->callback;
	p += 7; /* http:// */
#if ENABLE_FEATURE_IPV6
	if(*p == '[') { /* ip v6 */
		obj->addrstr[i++] = '[';
		p++;
		obj->ipv6 = 1;
		while(*p != ']' && i < (sizeof(obj->addrstr)-1))
			obj->addrstr[i++] = *(p++);
		if(*p == ']')
			p++;
		if(i < (sizeof(obj->addrstr)-1))
			obj->addrstr[i++] = ']';
	} else {
#endif
		while(*p != '/' && *p != ':' && i < (sizeof(obj->addrstr)-1))
			obj->addrstr[i++] = *(p++);
#if ENABLE_FEATURE_IPV6
	}
#endif
	obj->addrstr[i] = '\0';
	if(*p == ':') {
		obj->portstr[0] = *p;
		i = 1;
		p++;
		port = (unsigned short)atoi(p);
		while(*p != '/') {
			if(i<7) obj->portstr[i++] = *p;
			p++;
		}
		obj->portstr[i] = 0;
	} else {
		port = 80;
		obj->portstr[0] = '\0';
	}
	obj->path = p;
#if ENABLE_FEATURE_IPV6
	if(obj->ipv6) {
		char addrstr_tmp[48];
		struct sockaddr_in6 * sa = (struct sockaddr_in6 *)&addr;
		sa->sin6_family = AF_INET6;
		i = (int)strlen(obj->addrstr);
		if(i > 2) {
			i -= 2;
			memcpy(addrstr_tmp, obj->addrstr + 1, i);
			addrstr_tmp[i] = '\0';
			inet_pton(AF_INET6, addrstr_tmp, &(sa->sin6_addr));
		}
		sa->sin6_port = htons(port);
		addrlen = sizeof(struct sockaddr_in6);
	} else {
		struct sockaddr_in * sa = (struct sockaddr_in *)&addr;
		sa->sin_family = AF_INET;
		inet_pton(AF_INET, obj->addrstr, &(sa->sin_addr));
		sa->sin_port = htons(port);
		addrlen = sizeof(struct sockaddr_in);
	}
#else
	addr.sin_family = AF_INET;
	inet_aton(obj->addrstr, &addr.sin_addr);
	addr.sin_port = htons(port);
	addrlen = sizeof(struct sockaddr_in);
#endif
	DEBUG_LOG(syslog(LOG_DEBUG, "%s: '%s' %hu '%s'", "upnp_event_notify_connect",
	       obj->addrstr, port, obj->path);)
	obj->state = EConnecting;
	if(connect(obj->s, (struct sockaddr *)&addr, addrlen) < 0) {
		if(errno != EINPROGRESS && errno != EWOULDBLOCK) {
			bb_perror_msg("%s: connect(%d, %s, %u)", "upnp_event_notify_connect", obj->s,
			       obj->addrstr, addrlen);
			obj->state = EError;
		}
	}
}


static void upnp_event_send(struct upnp_event_notify * obj)
{
	int i;
	DEBUG_LOG(syslog(LOG_DEBUG, "%s: sending event notify message to %s%s",
	       "upnp_event_send", obj->addrstr, obj->portstr);)
	DEBUG_LOG(syslog(LOG_DEBUG, "%s: msg: %s",
	       "upnp_event_send", obj->buffer + obj->sent);)
	i = send(obj->s, obj->buffer + obj->sent, obj->tosend - obj->sent, 0);
	if(i<0) {
		if(errno != EAGAIN && errno != EWOULDBLOCK && errno != EINTR) {
			DEBUG_LOG(syslog(LOG_NOTICE, "%s: send(): %m", "upnp_event_send");)
			obj->state = EError;
			return;
		} else {
			/* EAGAIN or EWOULDBLOCK or EINTR : no data sent */
			i = 0;
		}
	}
#ifdef DEBUG
	if(i != (obj->tosend - obj->sent))
		syslog(LOG_NOTICE, "%s: %d bytes send out of %d",
		       "upnp_event_send", i, obj->tosend - obj->sent);
#endif
	obj->sent += i;
	if(obj->sent == obj->tosend)
		obj->state = EWaitingForResponse;
}

static void upnp_event_recv(struct upnp_event_notify * obj)
{
	int n;
	n = recv(obj->s, obj->buffer, obj->buffersize, 0);
	if(n<0) {
		if(errno != EAGAIN &&
		   errno != EWOULDBLOCK &&
		   errno != EINTR) {
			bb_perror_msg("%s: recv()", "upnp_event_recv");
			obj->state = EError;
		}
		return;
	}
	DEBUG_LOG(syslog(LOG_DEBUG, "%s: (%dbytes) %.*s", "upnp_event_recv", n, n, obj->buffer);)
	/* TODO : do something with the data recevied ?
	 * right now, n (number of bytes received) is ignored
	 * We may need to recv() more bytes. */
	obj->state = EFinished;
	if(obj->sub)
		obj->sub->seq++;
}

static void upnpevents_selectfds(fd_set *readset, fd_set *writeset, int * max_fd)
{
	struct upnp_event_notify * obj;
	for(obj = notifylist.lh_first; obj != NULL; obj = obj->entries.le_next) {
		DEBUG_LOG(syslog(LOG_DEBUG, "upnpevents_selectfds: %p %d %d", obj, obj->state, obj->s);)
		if(obj->s >= 0) {
			switch(obj->state) {
			case ECreated:
				upnp_event_notify_connect(obj);
				if(obj->state != EConnecting)
					break;
			case EConnecting:
			case ESending:
				FD_SET(obj->s, writeset);
				if(obj->s > *max_fd)
					*max_fd = obj->s;
				break;
			case EWaitingForResponse:
				FD_SET(obj->s, readset);
				if(obj->s > *max_fd)
					*max_fd = obj->s;
				break;
			default:
				;
			}
		}
	}
}


#endif


/* Event magical values codes */
#define SETUPREADY_MAGICALVALUE (248)
#define CONNECTIONSTATUS_MAGICALVALUE (249)
#define FIREWALLENABLED_MAGICALVALUE (250)
#define INBOUNDPINHOLEALLOWED_MAGICALVALUE (251)
#define SYSTEMUPDATEID_MAGICALVALUE (252)
#define PORTMAPPINGNUMBEROFENTRIES_MAGICALVALUE (253)
#define EXTERNALIPADDRESS_MAGICALVALUE (254)
#define DEFAULTCONNECTIONSERVICE_MAGICALVALUE (255)


static const char * const upnptypes[] =
{
	"string",
	"boolean",
	"ui2",
	"ui4",
	"bin.base64"
};

static const char * const upnpdefaultvalues[] =
{
	0,
	"IP_Routed"/*"Unconfigured"*/, /* 1 default value for ConnectionType */
	"3600", /* 2 default value for PortMappingLeaseDuration */
	"Unconfigured", /* 3 default value for ConnectionStatus */
	"0", /* 4 default value for RSIPAvailable */
	"1", /* 5 default value for NATEnabled */
	"ERROR_NONE", /* 6 default value for LastConnectionError */
};

static const char * const upnpallowedvalues[] =
{
	0,      /* 0 */
	"DSL",  /* 1 */
	"POTS",
	"Cable",
	"Ethernet",
	0,
	"Up",   /* 6 */
	"Down",
	"Initializing",
	"Unavailable",
	0,
	"TCP",  /* 11 */
	"UDP",
	0,
	"Unconfigured", /* 14 */
	"IP_Routed",
	"IP_Bridged",
	0,
	"Unconfigured", /* 18 */
	"Connecting",
	"Connected",
	"PendingDisconnect",
	"Disconnecting",
	"Disconnected",
	0,
	"ERROR_NONE",   /* 25 */
/* Optionals values :
 * ERROR_COMMAND_ABORTED
 * ERROR_NOT_ENABLED_FOR_INTERNET
 * ERROR_USER_DISCONNECT
 * ERROR_ISP_DISCONNECT
 * ERROR_IDLE_DISCONNECT
 * ERROR_FORCED_DISCONNECT
 * ERROR_NO_CARRIER
 * ERROR_IP_CONFIGURATION
 * ERROR_UNKNOWN */
	0,
	"",             /* 27 */
	0
};

static const int upnpallowedranges[] = {
	0,
	/* 1 PortMappingLeaseDuration */
	0,
	604800,
	/* 3 InternalPort */
	1,
	65535,
	/* 5 LeaseTime */
	1,
	86400,
	/* 7 OutboundPinholeTimeout */
	100,
	200,
};

static const char * const magicargname[] = {
	0,
	"StartPort",            /* 1 */
	"EndPort",              /* 2 */
	"RemoteHost",           /* 3 */
	"RemotePort",           /* 4 */
	"InternalClient",       /* 5 */
	"InternalPort",         /* 6 */
	"IsWorking",            /* 7 */
};

static const char xmlver[] = "<?xml version=\"1.0\"?>\r\n";
#define root_service "scpd xmlns=\"urn:schemas-upnp-org:service-1-0\""
static const char root_device[] = "root xmlns=\"urn:schemas-upnp-org:device-1-0\"";

/* root Description of the UPnP Device
 * fixed to match UPnP_IGD_InternetGatewayDevice 1.0.pdf
 * Needs to be checked with UPnP-gw-InternetGatewayDevice-v2-Device.pdf
 * presentationURL is only "recommended" but the router doesn't appears
 * in "Network connections" in Windows XP if it is not present. */
static const struct XMLElt rootDesc[] =
{
/* 0 */
	{root_device, INITHELPER(1,2)},
	{"specVersion", INITHELPER(3,2)},
#if defined(ENABLE_L3F_SERVICE)
	{"device", INITHELPER(5,13)},
#else
	{"device", INITHELPER(5,12)},
#endif
	{"/major", UPNP_VERSION_MAJOR_STR},
	{"/minor", UPNP_VERSION_MINOR_STR},
/* 5 */
	{"/deviceType", "urn:schemas-upnp-org:device:InternetGatewayDevice:1"},
	{"/friendlyName", ROOTDEV_FRIENDLYNAME},        /* required */
	{"/manufacturer", ROOTDEV_MANUFACTURER},        /* required */
/* 8 */
	{"/manufacturerURL", ROOTDEV_MANUFACTURERURL},  /* optional */
	{"/modelDescription", ROOTDEV_MODELDESCRIPTION}, /* recommended */
	{"/modelName", ROOTDEV_MODELNAME},      /* required */
	{"/modelNumber", bb_common_bss.modelnumber},
	{"/modelURL", ROOTDEV_MODELURL},
	{"/serialNumber", bb_common_bss.serialnumber},
	{"/UDN", bb_common_bss.uuidvalue_igd},        /* required */
	/* see if /UPC is needed. */
#define SERVICES_OFFSET 58
#if defined(ENABLE_L3F_SERVICE)
	/* here we dening Services for the root device :
	 * L3F and DUMMY and DeviceProtection */
#ifdef ENABLE_L3F_SERVICE
#define NSERVICES1 1
#else
#define NSERVICES1 0
#endif
#define NSERVICES2 0
#define NSERVICES3 0
#define NSERVICES (NSERVICES1+NSERVICES2+NSERVICES3)
	{"serviceList", INITHELPER(SERVICES_OFFSET,NSERVICES)},
	{"deviceList", INITHELPER(18,1)},
	{"/presentationURL", bb_common_bss.presentationurl},  /* recommended */
#else
	{"deviceList", INITHELPER(18,1)},
	{"/presentationURL", bb_common_bss.presentationurl},  /* recommended */
	{0,0},
#endif
/* 18 */
	{"device", INITHELPER(19,13)},
/* 19 */
	{"/deviceType", "urn:schemas-upnp-org:device:WANDevice:1"}, /* required */
	{"/friendlyName", WANDEV_FRIENDLYNAME},
	{"/manufacturer", WANDEV_MANUFACTURER},
	{"/manufacturerURL", WANDEV_MANUFACTURERURL},
	{"/modelDescription" , WANDEV_MODELDESCRIPTION},
	{"/modelName", WANDEV_MODELNAME},
	{"/modelNumber", WANDEV_MODELNUMBER},
	{"/modelURL", WANDEV_MODELURL},
	{"/serialNumber", bb_common_bss.serialnumber},
	{"/UDN", bb_common_bss.uuidvalue_wan},
	{"/UPC", WANDEV_UPC},   /* UPC (=12 digit barcode) is optional */
/* 30 */
	{"serviceList", INITHELPER(32,1)},
	{"deviceList", INITHELPER(38,1)},
/* 32 */
	{"service", INITHELPER(33,5)},
/* 33 */
	{"/serviceType",
			"urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1"},
	/*{"/serviceId", "urn:upnp-org:serviceId:WANCommonInterfaceConfig"}, */
	{"/serviceId", "urn:upnp-org:serviceId:WANCommonIFC1"}, /* required */
	{"/SCPDURL", WANCFG_PATH},
	{"/controlURL", WANCFG_CONTROLURL},
	{"/eventSubURL", WANCFG_EVENTURL},
/* 38 */
	{"device", INITHELPER(39,12)},
/* 39 */
	{"/deviceType", "urn:schemas-upnp-org:device:WANConnectionDevice:1"},
	{"/friendlyName", WANCDEV_FRIENDLYNAME},
	{"/manufacturer", WANCDEV_MANUFACTURER},
	{"/manufacturerURL", WANCDEV_MANUFACTURERURL},
	{"/modelDescription", WANCDEV_MODELDESCRIPTION},
	{"/modelName", WANCDEV_MODELNAME},
	{"/modelNumber", WANCDEV_MODELNUMBER},
	{"/modelURL", WANCDEV_MODELURL},
	{"/serialNumber", bb_common_bss.serialnumber},
	{"/UDN", bb_common_bss.uuidvalue_wcd},
	{"/UPC", WANCDEV_UPC},  /* UPC (=12 digit Barcode) is optional */
	{"serviceList", INITHELPER(51,1)},
/* 51 */
	{"service", INITHELPER(53,5)},
	{"service", INITHELPER(58,5)},
/* 53 */
	{"/serviceType", SERVICE_TYPE_WANIPC},
		/* urn:schemas-upnp-org:service:WANIPConnection:2 for v2 */
	{"/serviceId", SERVICE_ID_WANIPC},
		/* urn:upnp-org:serviceId:WANIPConn1 or 2 */
	{"/SCPDURL", WANIPC_PATH},
	{"/controlURL", WANIPC_CONTROLURL},
	{"/eventSubURL", WANIPC_EVENTURL},
/* 58 / 63 = SERVICES_OFFSET*/
#if defined(ENABLE_L3F_SERVICE)
	{"service", INITHELPER(SERVICES_OFFSET+2,5)},
	{"service", INITHELPER(SERVICES_OFFSET+7,5)},
#endif
#ifdef ENABLE_L3F_SERVICE
/* 60 / 65 = SERVICES_OFFSET+2 */
	{"/serviceType", "urn:schemas-upnp-org:service:Layer3Forwarding:1"},
	{"/serviceId", "urn:upnp-org:serviceId:L3Forwarding1"},
	{"/SCPDURL", L3F_PATH},
	{"/controlURL", L3F_CONTROLURL}, /* The Layer3Forwarding service is only */
	{"/eventSubURL", L3F_EVENTURL}, /* recommended, not mandatory */
#endif
	{0, 0}
};

/* WANIPCn.xml */
/* see UPnP_IGD_WANIPConnection 1.0.pdf
static struct XMLElt scpdWANIPCn[] =
{
	{root_service, {INITHELPER(1,2)}},
	{0, {0}}
};
*/
static const struct argument AddPortMappingArgs[] =
{
	{1, 11},        /* RemoteHost */
	{1, 12},        /* ExternalPort */
	{1, 14},        /* PortMappingProtocol */
	{1, 13},        /* InternalPort */
	{1, 15},        /* InternalClient */
	{1, 9},         /* PortMappingEnabled */
	{1, 16},        /* PortMappingDescription */
	{1, 10},        /* PortMappingLeaseDuration */
	{0, 0}
};

static const struct argument GetExternalIPAddressArgs[] =
{
	{2, 7},
	{0, 0}
};

static const struct argument DeletePortMappingArgs[] =
{
	{1, 11},        /* RemoteHost */
	{1, 12},        /* ExternalPort */
	{1, 14},        /* PortMappingProtocol */
	{0, 0}
};

static const struct argument SetConnectionTypeArgs[] =
{
	{1, 0},
	{0, 0}
};

static const struct argument GetConnectionTypeInfoArgs[] =
{
	{2, 0},
	{2, 1},
	{0, 0}
};

static const struct argument GetStatusInfoArgs[] =
{
	{2, 2},
	{2, 4},
	{2, 3},
	{0, 0}
};

static const struct argument GetNATRSIPStatusArgs[] =
{
	{2, 5},
	{2, 6},
	{0, 0}
};

static const struct argument GetGenericPortMappingEntryArgs[] =
{
	{1, 8},
	{2, 11},
	{2, 12},
	{2, 14},
	{2, 13},
	{2, 15},
	{2, 9},
	{2, 16},
	{2, 10},
	{0, 0}
};

static const struct argument GetSpecificPortMappingEntryArgs[] =
{
	{1, 11},
	{1, 12},
	{1, 14},
	{2, 13},
	{2, 15},
	{2, 9},
	{2, 16},
	{2, 10},
	{0, 0}
};

static const struct action WANIPCnActions[] =
{
	{"SetConnectionType", SetConnectionTypeArgs}, /* R */
	{"GetConnectionTypeInfo", GetConnectionTypeInfoArgs}, /* R */
	{"RequestConnection", 0}, /* R */
	/*{"RequestTermination", 0},*/ /* O */
	{"ForceTermination", 0}, /* R */
	/*{"SetAutoDisconnectTime", 0},*/ /* O */
	/*{"SetIdleDisconnectTime", 0},*/ /* O */
	/*{"SetWarnDisconnectDelay", 0}, */ /* O */
	{"GetStatusInfo", GetStatusInfoArgs}, /* R */
	/*GetAutoDisconnectTime*/ /* O */
	/*GetIdleDisconnectTime*/ /* O */
	/*GetWarnDisconnectDelay*/ /* O */
	{"GetNATRSIPStatus", GetNATRSIPStatusArgs}, /* R */
	{"GetGenericPortMappingEntry", GetGenericPortMappingEntryArgs}, /* R */
	{"GetSpecificPortMappingEntry", GetSpecificPortMappingEntryArgs}, /* R */
	{"AddPortMapping", AddPortMappingArgs}, /* R */
	{"DeletePortMapping", DeletePortMappingArgs}, /* R */
	{"GetExternalIPAddress", GetExternalIPAddressArgs}, /* R */
	{0, 0}
};
/* R=Required, O=Optional */

/* ignore "warning: missing initializer" */
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"

/* struct stateVar :
 * {name, itype(&event), idefault, iallowedlist, ieventvalue} */
static const struct stateVar WANIPCnVars[] =
{
/* 0 */
	{"ConnectionType", 0, 1, 0, 15}, /* required */
	{"PossibleConnectionTypes", 0|0x80, 0, 14, 15},
	 /* Required
	  * Allowed values : Unconfigured / IP_Routed / IP_Bridged */
	{"ConnectionStatus", 0|0x80, 3, 18,
	 CONNECTIONSTATUS_MAGICALVALUE }, /* required */
	 /* Allowed Values : Unconfigured / Connecting(opt) / Connected
	  *                  PendingDisconnect(opt) / Disconnecting (opt)
	  *                  Disconnected */
	{"Uptime", 3, 0},       /* Required */
	{"LastConnectionError", 0, 6, 25},      /* required : */
	 /* Allowed Values : ERROR_NONE(req) / ERROR_COMMAND_ABORTED(opt)
	  *                  ERROR_NOT_ENABLED_FOR_INTERNET(opt)
	  *                  ERROR_USER_DISCONNECT(opt)
	  *                  ERROR_ISP_DISCONNECT(opt)
	  *                  ERROR_IDLE_DISCONNECT(opt)
	  *                  ERROR_FORCED_DISCONNECT(opt)
	  *                  ERROR_NO_CARRIER(opt)
	  *                  ERROR_IP_CONFIGURATION(opt)
	  *                  ERROR_UNKNOWN(opt) */
	{"RSIPAvailable", 1, 4}, /* required */
	{"NATEnabled", 1, 5},    /* required */
	{"ExternalIPAddress", 0|0x80, 0, 0,
	 EXTERNALIPADDRESS_MAGICALVALUE}, /* required. Default : empty string */
	{"PortMappingNumberOfEntries", 2|0x80, 0, 0,
	 PORTMAPPINGNUMBEROFENTRIES_MAGICALVALUE}, /* required >= 0 */
	{"PortMappingEnabled", 1, 0}, /* Required */
/* 10 */
	{"PortMappingLeaseDuration", 3, 2, 1}, /* required */
	/* TODO : for IGD v2 :
	 * <stateVariable sendEvents="no">
	 *   <name>PortMappingLeaseDuration</name>
	 *   <dataType>ui4</dataType>
	 *   <defaultValue>Vendor-defined</defaultValue>
	 *   <allowedValueRange>
	 *      <minimum>0</minimum>
	 *      <maximum>604800</maximum>
	 *   </allowedValueRange>
	 * </stateVariable> */
	{"RemoteHost", 0, 0},   /* required. Default : empty string */
	{"ExternalPort", 2, 0}, /* required */
	{"InternalPort", 2, 0, 3}, /* required */
	{"PortMappingProtocol", 0, 0, 11}, /* required allowedValues: TCP/UDP */
	{"InternalClient", 0, 0}, /* required */
	{"PortMappingDescription", 0, 0}, /* required default: empty string */
	{0, 0}
};

static const struct serviceDesc scpdWANIPCn =
{ WANIPCnActions, WANIPCnVars };

/* WANCfg.xml */
/* See UPnP_IGD_WANCommonInterfaceConfig 1.0.pdf */

static const struct argument GetCommonLinkPropertiesArgs[] =
{
	{2, 0},
	{2, 1},
	{2, 2},
	{2, 3},
	{0, 0}
};

static const struct argument GetTotalBytesSentArgs[] =
{
	{2, 4},
	{0, 0}
};

static const struct argument GetTotalBytesReceivedArgs[] =
{
	{2, 5},
	{0, 0}
};

static const struct argument GetTotalPacketsSentArgs[] =
{
	{2, 6},
	{0, 0}
};

static const struct argument GetTotalPacketsReceivedArgs[] =
{
	{2, 7},
	{0, 0}
};

static const struct action WANCfgActions[] =
{
	{"GetCommonLinkProperties", GetCommonLinkPropertiesArgs}, /* Required */
	{"GetTotalBytesSent", GetTotalBytesSentArgs},             /* optional */
	{"GetTotalBytesReceived", GetTotalBytesReceivedArgs},     /* optional */
	{"GetTotalPacketsSent", GetTotalPacketsSentArgs},         /* optional */
	{"GetTotalPacketsReceived", GetTotalPacketsReceivedArgs}, /* optional */
	{0, 0}
};

/* See UPnP_IGD_WANCommonInterfaceConfig 1.0.pdf */
static const struct stateVar WANCfgVars[] =
{
	{"WANAccessType", 0, 0, 1},
	/* Allowed Values : DSL / POTS / Cable / Ethernet
	 * Default value : empty string */
	{"Layer1UpstreamMaxBitRate", 3, 0},
	{"Layer1DownstreamMaxBitRate", 3, 0},
	{"PhysicalLinkStatus", 0|0x80, 0, 6, 6},
	/*  allowed values :
	 *      Up / Down / Initializing (optional) / Unavailable (optionnal)
	 *  no Default value
	 *  Evented */
	{"TotalBytesSent", 3, 0},          /* Optional */
	{"TotalBytesReceived", 3, 0},  /* Optional */
	{"TotalPacketsSent", 3, 0},    /* Optional */
	{"TotalPacketsReceived", 3, 0},/* Optional */
	/*{"MaximumActiveConnections", 2, 0},   // allowed Range value // OPTIONAL */
	/*{"WANAccessProvider", 0, 0},*/   /* Optional */
	{0, 0}
};

static const struct serviceDesc scpdWANCfg =
{ WANCfgActions, WANCfgVars };

#ifdef ENABLE_L3F_SERVICE
/* Read UPnP_IGD_Layer3Forwarding_1.0.pdf */
static const struct argument SetDefaultConnectionServiceArgs[] =
{
	{1, 0}, /* in */
	{0, 0}
};

static const struct argument GetDefaultConnectionServiceArgs[] =
{
	{2, 0}, /* out */
	{0, 0}
};

static const struct action L3FActions[] =
{
	{"SetDefaultConnectionService", SetDefaultConnectionServiceArgs}, /* Req */
	{"GetDefaultConnectionService", GetDefaultConnectionServiceArgs}, /* Req */
	{0, 0}
};

static const struct stateVar L3FVars[] =
{
	{"DefaultConnectionService", 0|0x80, 0, 0,
	 DEFAULTCONNECTIONSERVICE_MAGICALVALUE}, /* Required */
	{0, 0}
};

static const struct serviceDesc scpdL3F =
{ L3FActions, L3FVars };
#endif

/* strcat_str()
 * concatenate the string and use realloc to increase the
 * memory buffer if needed. */
typedef struct CONCATENATE_STR {
	char *str;
	size_t len;
	size_t allocated;
} concatenate_str_t;

static void strcat_str(concatenate_str_t *s, int nstr, ...)
{
	va_list slist1;
	size_t sz = 0;
	size_t *psz;
	const char **pstrs;
	int i;

	va_start(slist1, nstr);
	pstrs = alloca((nstr * sizeof(char *))+(nstr * sizeof(size_t)));
	psz = (size_t *)(pstrs + nstr);
	for(i = 0; i < nstr; i++) {
		psz[i] = strlen(pstrs[i] = va_arg(slist1, const char *));
		sz += psz[i];
	}
	va_end(slist1);

	if(s->allocated <= (s->len + sz)) {
		char *p;
		size_t newlen = (s->allocated + sz + 256) & ~255;

		p = realloc(s->str, newlen);
		if(p == NULL) /* handle a failure of realloc() */
			return;
		s->str = p;
		s->allocated = newlen;
	}
	sz = s->len;
	for(i = 0; i < nstr; i++) {
		memcpy(s->str + sz, pstrs[i], psz[i] + 1);
		sz += psz[i];
	}
	s->len = sz;
}

/* iterative subroutine using a small stack
 * This way, the progam stack usage is kept low */
static concatenate_str_t *genXML(concatenate_str_t *cs, const struct XMLElt *p)
{
	unsigned short i, j;
	unsigned long k;
	int top;
	const char * eltname;
	struct {
		unsigned short i;
		unsigned short j;
		const char * eltname;
	} pile[16]; /* stack */
	top = -1;
	i = 0;  /* current node */
	j = 1;  /* i + number of nodes*/
	for(;;) {
		eltname = p[i].eltname;
		if(!eltname)
			return cs;
		if(eltname[0] == '/') {
			if(p[i].data && p[i].data[0]) {
				/*printf("<%s>%s<%s>\n", eltname+1, p[i].data, eltname); */
				strcat_str(cs, 7, "<", eltname+1, ">", p[i].data, "<", eltname, ">");
			}
			for(;;) {
				if(top < 0)
					return cs;
				i = ++(pile[top].i);
				j = pile[top].j;
				/*printf("  pile[%d]\t%d %d\n", top, i, j); */
				if(i==j) {
					/*printf("</%s>\n", pile[top].eltname); */
					strcat_str(cs, 3, "</", pile[top].eltname, ">");
					top--;
				} else
					break;
			}
		} else {
			/*printf("<%s>\n", eltname); */
			strcat_str(cs, 2, "<", eltname);
			if(memcmp(eltname, "root ", 5) == 0) {
				char configid_str[32];
				/* add configId attribute, required by UDA 1.1 */
				snprintf(configid_str, sizeof(configid_str), " configId=\"%u\"", bb_common_bss.upnp_configid);
				strcat_str(cs, 1, configid_str);
			}
			strcat_str(cs, 1, ">");
			k = (unsigned long)p[i].data;
			i = k & 0xffff;
			j = i + (k >> 16);
			top++;
			/*printf(" +pile[%d]\t%d %d\n", top, i, j); */
			pile[top].i = i;
			pile[top].j = j;
			pile[top].eltname = eltname;
		}
	}
}

/* genRootDesc() :
 * - Generate the root description of the UPnP device.
 * - the len argument is used to return the length of
 *   the returned string.
 * - tmp_uuid argument is used to build the uuid string */
static char *genRootDesc(int *len)
{
	concatenate_str_t concatenate_str;

	strcat_str(memset(&concatenate_str, 0, sizeof(concatenate_str_t)), 1, xmlver);
	genXML(&concatenate_str, rootDesc);
	return concatenate_str.str;
}

/* genServiceDesc() :
 * Generate service description with allowed methods and
 * related variables. */
static char *
genServiceDesc(int *len, const struct serviceDesc *s)
{
	int i, j;
	const struct action *acts = s->actionList;
	const struct stateVar *vars = s->serviceStateTable;
	const struct argument *args;
	const char *p;
	concatenate_str_t concatenate_str, *cs;

	cs = &concatenate_str;
	strcat_str(memset(cs, 0, sizeof(concatenate_str_t)), 2, xmlver, "<" root_service ">"
		"<specVersion><major>" UPNP_VERSION_MAJOR_STR "</major>"
		"<minor>" UPNP_VERSION_MINOR_STR "</minor></specVersion>" "<actionList>");

	i = 0;
	while(acts[i].name) {
		strcat_str(cs, 3, "<action><name>", acts[i].name, "</name>");
		/* argument List */
		args = acts[i].args;
		if(args) {
			strcat_str(cs, 1, "<argumentList>");
			j = 0;
			while(args[j].dir) {
				strcat_str(cs, 1, "<argument><name>");
				if((args[j].dir & 0x80) == 0)
					strcat_str(cs, 1, "New");
				p = vars[args[j].relatedVar].name;
				if(args[j].dir & 0x7c) {
					/* use magic values ... */
					strcat_str(cs, 1, magicargname[(args[j].dir & 0x7c) >> 2]);
				} else if(0 == memcmp(p, "PortMapping", 11)
				   && 0 != memcmp(p + 11, "Description", 11)) {
					if(0 == memcmp(p + 11, "NumberOfEntries", 15)) {
						/* PortMappingNumberOfEntries */
						strcat_str(cs, 1, "PortMappingIndex");
					} else {
						/* PortMappingEnabled
						 * PortMappingLeaseDuration
						 * PortMappingProtocol */
						strcat_str(cs, 1, p + 11);
					}
				} else {
					strcat_str(cs, 1, p);
				}
				strcat_str(cs, 5, "</name><direction>", ((args[j].dir&0x03)==1?"in":"out"),
						"</direction><relatedStateVariable>", p,
						"</relatedStateVariable></argument>");
				j++;
			}
			strcat_str(cs, 1, "</argumentList>");
		}
		strcat_str(cs, 1,  "</action>");
		/* strcat_str(cs, 1, "\n"); // TEMP ! */
		i++;
	}
	strcat_str(cs, 1, "</actionList><serviceStateTable>");
	i = 0;
	while(vars[i].name) {
		strcat_str(cs, 1, "<stateVariable sendEvents=\"");
#ifdef ENABLE_EVENTS
		strcat_str(cs, 1, ((vars[i].itype & 0x80)?"yes":"no"));
#else
		/* for the moment always send no. Wait for SUBSCRIBE implementation
		 * before setting it to yes */
		strcat_str(cs, 1, "no");
#endif
		strcat_str(cs, 5, "\"><name>", vars[i].name, "</name><dataType>",
			upnptypes[vars[i].itype & 0x0f], "</dataType>");
		/*if(vars[i].defaultValue) */
		if(vars[i].idefault)
		  strcat_str(cs, 3, "<defaultValue>", upnpdefaultvalues[vars[i].idefault], "</defaultValue>");
		if(vars[i].iallowedlist) {
		  if((vars[i].itype & 0x0f) == 0) {
		    /* string */
		    strcat_str(cs, 1, "<allowedValueList>");
		    for(j=vars[i].iallowedlist; upnpallowedvalues[j]; j++)
		      strcat_str(cs, 3, "<allowedValue>", upnpallowedvalues[j], "</allowedValue>");
		    strcat_str(cs, 1, "</allowedValueList>");
		  } else {
		    /* ui2 and ui4 */
		    char ibuf[200];
		    snprintf(ibuf, sizeof(ibuf), "<allowedValueRange><minimum>%d</minimum><maximum>%d</maximum></allowedValueRange>",
				upnpallowedranges[vars[i].iallowedlist], upnpallowedranges[vars[i].iallowedlist+1]);
		    strcat_str(cs, 1, ibuf);
		  }
		}
		strcat_str(cs, 1, "</stateVariable>");
		/* strcat_str(cs, 1, "\n"); // TEMP ! */
		i++;
	}
	strcat_str(cs, 1, "</serviceStateTable></scpd>");
	*len = cs->len;
	return cs->str;
}

/* Generate the WANIPConnection xml description */
static char *genWANIPCn(int * len)
{
	return genServiceDesc(len, &scpdWANIPCn);
}

/* Generate the WANInterfaceConfig xml description. */
static char *genWANCfg(int * len)
{
	return genServiceDesc(len, &scpdWANCfg);
}

#ifdef ENABLE_L3F_SERVICE
static char *genL3F(int * len)
{
	return genServiceDesc(len, &scpdL3F);
}
#endif

#ifdef ENABLE_EVENTS
static char *genEventVars(int * len, const struct serviceDesc * s)
{
	char tmp[16];
	const struct stateVar * v = s->serviceStateTable;
	concatenate_str_t concatenate_str, *cs;

	cs = &concatenate_str;
	strcat_str(memset(cs, 0, sizeof(concatenate_str_t)), 1,
			"<e:propertyset xmlns:e=\"urn:schemas-upnp-org:event-1-0\">");
	while(v->name) {
		if(v->itype & 0x80) {
			strcat_str(cs, 3, "<e:property><", v->name, ">");
			switch(v->ieventvalue) {
			case 0:
				break;
			case CONNECTIONSTATUS_MAGICALVALUE:
				/* or get_wan_connection_status_str(ext_if_name) */
				strcat_str(cs, 1, upnpallowedvalues[18 + get_wan_connection_status(bb_common_bss.ext_if_name)]);
				break;
			case PORTMAPPINGNUMBEROFENTRIES_MAGICALVALUE:
				/* Port mapping number of entries magical value */
				snprintf(tmp, sizeof(tmp), "%d", upnp_get_portmapping_number_of_entries());
				strcat_str(cs, 1, tmp);
				break;
			case EXTERNALIPADDRESS_MAGICALVALUE:
				/* External ip address magical value */
				if(bb_common_bss.use_ext_ip_addr)
					strcat_str(cs, 1, bb_common_bss.use_ext_ip_addr);
				else {
					char ext_ip_addr[INET_ADDRSTRLEN];
					if(getifaddr(bb_common_bss.ext_if_name, ext_ip_addr, INET_ADDRSTRLEN, NULL, NULL) < 0) {
						strcat_str(cs, 1, "0.0.0.0");
					} else {
						strcat_str(cs, 1, ext_ip_addr);
					}
				}
				break;
			case DEFAULTCONNECTIONSERVICE_MAGICALVALUE:
				/* DefaultConnectionService magical value */
				strcat_str(cs, 2, bb_common_bss.uuidvalue_wcd,
					":WANConnectionDevice:1,urn:upnp-org:serviceId:WANIPConn1");
				break;
			default:
				strcat_str(cs, 1, upnpallowedvalues[v->ieventvalue]);
			}
			strcat_str(cs, 3, "</", v->name, "></e:property>");
		}
		v++;
	}
	strcat_str(cs, 1, "</e:propertyset>");
	*len = cs->len;
	return cs->str;
}

static char *getVarsWANIPCn(int * l)
{
	return genEventVars(l, &scpdWANIPCn);
}

static char *getVarsWANCfg(int * l)
{
	return genEventVars(l, &scpdWANCfg);
}

#ifdef ENABLE_L3F_SERVICE
static char *getVarsL3F(int * l)
{
	return genEventVars(l, &scpdL3F);
}
#endif

#endif /* ENABLE_EVENTS */

/* $Id: upnphttp.h,v 1.42 2015/12/16 10:21:49 nanard Exp $ */

#ifndef UPNPHTTP_H_INCLUDED
#define UPNPHTTP_H_INCLUDED

#define UPNP_VERSION_STRING "UPnP/" UPNP_VERSION_MAJOR_STR "." UPNP_VERSION_MINOR_STR

/* server: HTTP header returned in all HTTP responses : */
#define MINIUPNPD_SERVER_STRING OS_VERSION " " UPNP_VERSION_STRING " MiniUPnPd/" MINIUPNPD_VERSION

/*
 states :
  0 - waiting for data to read
  1 - waiting for HTTP Post Content.
  ...
  >= 100 - to be deleted
*/
enum httpStates {
	EWaitingForHttpRequest = 0,
	EWaitingForHttpContent,
	ESendingContinue,
	ESendingAndClosing,
	EToDelete = 100
};

enum httpCommands {
	EUnknown = 0,
	EGet,
	EPost,
	ESubscribe,
	EUnSubscribe
};

struct upnphttp {
	int socket;
	struct in_addr clientaddr;      /* client address */
#if ENABLE_FEATURE_IPV6
	int ipv6;
	struct in6_addr clientaddr_v6;
#endif
#ifdef DEBUG
	char clientaddr_str[64];        /* used for syslog() output */
#endif
	enum httpStates state;
	char HttpVer[16];
	/* request */
	char * req_buf;
	char accept_language[8];
	int req_buflen;
	int req_contentlen;
	int req_contentoff;     /* header length */
	enum httpCommands req_command;
	int req_soapActionOff;
	int req_soapActionLen;
	int req_HostOff;        /* Host: header */
	int req_HostLen;
#ifdef ENABLE_EVENTS
	int req_CallbackOff;    /* For SUBSCRIBE */
	int req_CallbackLen;
	int req_Timeout;
	int req_SIDOff;         /* For UNSUBSCRIBE */
	int req_SIDLen;
	const char * res_SID;
#endif
	int respflags;                          /* see FLAG_* constants below */
	/* response */
	char * res_buf;
	int res_buflen;
	int res_sent;
	int res_buf_alloclen;
	LIST_ENTRY(upnphttp) entries;
};

/* Include the "Timeout:" header in response */
#define FLAG_TIMEOUT    0x01
/* Include the "SID:" header in response */
#define FLAG_SID                0x02

/* If set, the POST request included a "Expect: 100-continue" header */
#define FLAG_CONTINUE   0x40

/* If set, the Content-Type is set to text/xml, otherwise it is text/xml */
#define FLAG_HTML               0x80

/* If set, the corresponding Allow: header is set */
#define FLAG_ALLOW_POST                 0x100
#define FLAG_ALLOW_SUB_UNSUB    0x200


#endif

/* $Id: upnphttp.c,v 1.105 2016/02/16 12:15:02 nanard Exp $ */
/* vim: tabstop=4 shiftwidth=4 noexpandtab */
/* Project :  miniupnp
 * Website :  http://miniupnp.free.fr/ or http://miniupnp.tuxfamily.org/
 * Author :   Thomas Bernard
 * Copyright (c) 2005-2015 Thomas Bernard
 * This software is subject to the conditions detailed in the
 * LICENCE file included in this distribution.
 * */

static struct upnphttp *
New_upnphttp(int s)
{
	struct upnphttp * ret;
	if(s<0)
		return NULL;
	ret = malloc(sizeof(struct upnphttp));
	if(ret == NULL)
		return NULL;
	memset(ret, 0, sizeof(struct upnphttp));
	ret->socket = s;
	if(!set_non_blocking(s))
		xlog_warn("New_upnphttp::set_non_blocking(): %m");
	return ret;
}

static void
CloseSocket_upnphttp(struct upnphttp * h)
{
	/* SSL_shutdown() ? */
	if(close(h->socket) < 0)
		bb_perror_msg("CloseSocket_upnphttp: close(%d)", h->socket);
	h->socket = -1;
	h->state = EToDelete;
}

static void
Delete_upnphttp(struct upnphttp * h)
{
	if(h) {
		if(h->socket >= 0)
			CloseSocket_upnphttp(h);
		if(h->req_buf)
			free(h->req_buf);
		if(h->res_buf)
			free(h->res_buf);
		free(h);
	}
}

/* parse HttpHeaders of the REQUEST
 * This function is called after the \r\n\r\n character
 * sequence has been found in h->req_buf */
static void
ParseHttpHeaders(struct upnphttp * h)
{
	char * line;
	char * colon;
	char * p;
	int n;
	if((h->req_buf == NULL) || (h->req_contentoff <= 0))
		return;
	line = h->req_buf;
	while(line < (h->req_buf + h->req_contentoff)) {
		colon = line;
		while(*colon != ':') {
			if(*colon == '\r' || *colon == '\n') {
				colon = NULL;   /* no ':' character found on the line */
				break;
			}
			colon++;
		}
		if(colon) {
			if(strncasecmp(line, "Content-Length:", 15)==0) {
				p = colon;
				while((*p < '0' || *p > '9') && (*p != '\r') && (*p != '\n'))
					p++;
				h->req_contentlen = atoi(p);
				if(h->req_contentlen < 0) {
					xlog_warn("ParseHttpHeaders() invalid Content-Length %d", h->req_contentlen);
					h->req_contentlen = 0;
				}
				/*printf("*** Content-Lenght = %d ***\n", h->req_contentlen);
				printf("    readbufflen=%d contentoff = %d\n",
					h->req_buflen, h->req_contentoff);*/
			} else if(strncasecmp(line, "Host:", 5)==0) {
				p = colon;
				n = 0;
				while(*p == ':' || *p == ' ' || *p == '\t')
					p++;
				while(p[n]>' ')
					n++;
				h->req_HostOff = p - h->req_buf;
				h->req_HostLen = n;
			} else if(strncasecmp(line, "SOAPAction:", 11)==0) {
				p = colon;
				n = 0;
				while(*p == ':' || *p == ' ' || *p == '\t')
					p++;
				while(p[n]>=' ')
					n++;
				if((p[0] == '"' && p[n-1] == '"')
				  || (p[0] == '\'' && p[n-1] == '\''))
				{
					p++; n -= 2;
				}
				h->req_soapActionOff = p - h->req_buf;
				h->req_soapActionLen = n;
			} else if(strncasecmp(line, "accept-language:", 16) == 0) {
				p = colon;
				n = 0;
				while(*p == ':' || *p == ' ' || *p == '\t')
					p++;
				while(p[n]>=' ')
					n++;
				DEBUG_LOG(syslog(LOG_DEBUG, "accept-language HTTP header : '%.*s'", n, p);)
				/* keep only the 1st accepted language */
				n = 0;
				while(p[n]>' ' && p[n] != ',')
					n++;
				if(n >= (int)sizeof(h->accept_language))
					n = (int)sizeof(h->accept_language) - 1;
				memcpy(h->accept_language, p, n);
				h->accept_language[n] = '\0';
			}
			else if(strncasecmp(line, "expect:", 7) == 0) {
				p = colon;
				n = 0;
				while(*p == ':' || *p == ' ' || *p == '\t')
					p++;
				while(p[n]>=' ')
					n++;
				if(strncasecmp(p, "100-continue", 12) == 0) {
					h->respflags |= FLAG_CONTINUE;
					DEBUG_LOG(syslog(LOG_DEBUG, "\"Expect: 100-Continue\" header detected");)
				}
			}
#ifdef ENABLE_EVENTS
			else if(strncasecmp(line, "Callback:", 9)==0) {
				/* The Callback can contain several urls :
				 * If there is more than one URL, when the service sends
				 * events, it will try these URLs in order until one
				 * succeeds. One or more URLs each enclosed by angle
				 * brackets ("<" and ">") */
				p = colon;
				while(*p != '<' && *p != '\r' )
					p++;
				n = 0;
				while(p[n] != '\r')
					n++;
				while(n > 0 && p[n] != '>')
					n--;
				/* found last > character */
				h->req_CallbackOff = p - h->req_buf;
				h->req_CallbackLen = MAX(0, n + 1);
			} else if(strncasecmp(line, "SID:", 4)==0) {
				p = bb_skip_whitespace(colon + 1);
				n = 0;
				while(!isspace(p[n]))
					n++;
				h->req_SIDOff = p - h->req_buf;
				h->req_SIDLen = n;
			}
			/* Timeout: Seconds-nnnn */
/* TIMEOUT
Recommended. Requested duration until subscription expires,
either number of seconds or infinite. Recommendation
by a UPnP Forum working committee. Defined by UPnP vendor.
 Consists of the keyword "Second-" followed (without an
intervening space) by either an integer or the keyword "infinite". */
			else if(strncasecmp(line, "Timeout:", 8)==0) {
				p = colon + 1;
				while((*p == ' ') || (*p == '\t'))
					p++;
				if(strncasecmp(p, "Second-", 7)==0) {
					h->req_Timeout = atoi(p+7);
				}
			}
#endif /* ENABLE_EVENTS */
		}
		/* the loop below won't run off the end of the buffer
		 * because the buffer is guaranteed to contain the \r\n\r\n
		 * character sequence */
		while(!(line[0] == '\r' && line[1] == '\n'))
			line++;
		line += 2;
	}
}

static const char httpresphead[] =
	"%s %d %s\r\n"
	"Content-Type: %s\r\n"
	"Connection: close\r\n"
	"Content-Length: %d\r\n"
	"Server: " MINIUPNPD_SERVER_STRING "\r\n"
	"Ext:\r\n"
	;       /*"\r\n";*/
/*
		"<?xml version=\"1.0\"?>\n"
		"<s:Envelope xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\" "
		"s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\">"
		"<s:Body>"

		"</s:Body>"
		"</s:Envelope>";
*/

/* with response code and response message
 * also allocate enough memory */
static int
BuildHeader_upnphttp(struct upnphttp * h, int respcode,
		     const char * respmsg,
		     int bodylen)
{
	int templen;
	if(!h->res_buf ||
	   h->res_buf_alloclen < ((int)sizeof(httpresphead) + 256 + bodylen)) {
		if(h->res_buf)
			free(h->res_buf);
		templen = sizeof(httpresphead) + 256 + bodylen;
		h->res_buf = (char *)malloc(templen);
		if(!h->res_buf) {
			bb_error_msg("malloc error in BuildHeader_upnphttp()");
			return -1;
		}
		h->res_buf_alloclen = templen;
	}
	h->res_sent = 0;
	h->res_buflen = snprintf(h->res_buf, h->res_buf_alloclen,
				 httpresphead, h->HttpVer,
				 respcode, respmsg,
				 (h->respflags&FLAG_HTML)?"text/html":"text/xml; charset=\"utf-8\"",
							 bodylen);
	/* Content-Type MUST be 'text/xml; charset="utf-8"' according to UDA v1.1 */
	/* Content-Type MUST be 'text/xml' according to UDA v1.0 */
	/* Additional headers */
#ifdef ENABLE_EVENTS
	if(h->respflags & FLAG_TIMEOUT) {
		h->res_buflen += snprintf(h->res_buf + h->res_buflen,
					  h->res_buf_alloclen - h->res_buflen,
					  "Timeout: Second-");
		if(h->req_Timeout) {
			h->res_buflen += snprintf(h->res_buf + h->res_buflen,
						  h->res_buf_alloclen - h->res_buflen,
						  "%d\r\n", h->req_Timeout);
		} else {
			h->res_buflen += snprintf(h->res_buf + h->res_buflen,
						  h->res_buf_alloclen - h->res_buflen,
						  "infinite\r\n");
		}
	}
	if(h->respflags & FLAG_SID) {
		h->res_buflen += snprintf(h->res_buf + h->res_buflen,
					  h->res_buf_alloclen - h->res_buflen,
					  "SID: %s\r\n", h->res_SID);
	}
#endif
	if(h->respflags & FLAG_ALLOW_POST) {
		h->res_buflen += snprintf(h->res_buf + h->res_buflen,
					  h->res_buf_alloclen - h->res_buflen,
					  "Allow: %s\r\n", "POST");
	} else if(h->respflags & FLAG_ALLOW_SUB_UNSUB) {
		h->res_buflen += snprintf(h->res_buf + h->res_buflen,
					  h->res_buf_alloclen - h->res_buflen,
					  "Allow: %s\r\n", "SUBSCRIBE, UNSUBSCRIBE");
	}
	if(h->accept_language[0] != '\0') {
		/* defaulting to "en" */
		h->res_buflen += snprintf(h->res_buf + h->res_buflen,
					  h->res_buf_alloclen - h->res_buflen,
					  "Content-Language: %s\r\n",
					  h->accept_language[0] == '*' ? "en" : h->accept_language);
	}
	h->res_buf[h->res_buflen++] = '\r';
	h->res_buf[h->res_buflen++] = '\n';
	if(h->res_buf_alloclen < (h->res_buflen + bodylen)) {
		char *tmp = realloc(h->res_buf, (h->res_buflen + bodylen));

		if(tmp) {
			h->res_buf = tmp;
			h->res_buf_alloclen = h->res_buflen + bodylen;
		} else {
			bb_error_msg("realloc error in BuildHeader_upnphttp()");
			return -1;
		}
	}
	return 0;
}


static void
BuildResp2_upnphttp(struct upnphttp * h, int respcode,
		    const char * respmsg,
		    const char * body, int bodylen)
{
	int r = BuildHeader_upnphttp(h, respcode, respmsg, bodylen);
	if(body && (r >= 0)) {
		memcpy(h->res_buf + h->res_buflen, body, bodylen);
		h->res_buflen += bodylen;
	}
}



/* findendheaders() find the \r\n\r\n character sequence and
 * return a pointer to it.
 * It returns NULL if not found */
static const char *
findendheaders(const char * s, int len)
{
	while(len-->3) {
		if(s[0]=='\r' && s[1]=='\n' && s[2]=='\r' && s[3]=='\n')
			return s;
		s++;
	}
	return NULL;
}

#ifdef ENABLE_EVENTS
/**
 * checkCallbackURL()
 * check that url is on originating IP
 * extract first correct URL
 * returns 0 if the callback header value is not valid
 * 1 if it is valid.
 */
static int
checkCallbackURL(struct upnphttp * h)
{
	char addrstr[48];
	int ipv6;
	const char * p;
	unsigned int i;

start_again:
	if(h->req_CallbackOff <= 0 || h->req_CallbackLen < 10)
		return 0;
	if(memcmp(h->req_buf + h->req_CallbackOff, "<http://", 8) != 0) {
		p = h->req_buf + h->req_CallbackOff + 1;
		goto invalid;
	}
	/* extract host from url to addrstr[] */
	i = 0;
	p = h->req_buf + h->req_CallbackOff + 8;
	if(*p == '[') {
		p++;
		ipv6 = 1;
		while(*p != ']' && *p != '>' && i < (sizeof(addrstr)-1)
		      && p < (h->req_buf + h->req_CallbackOff + h->req_CallbackLen))
			addrstr[i++] = *(p++);
	} else {
		ipv6 = 0;
		while(*p != '/' && *p != ':' && *p != '>' && i < (sizeof(addrstr)-1)
		      && p < (h->req_buf + h->req_CallbackOff + h->req_CallbackLen))
			addrstr[i++] = *(p++);
	}
	addrstr[i] = '\0';
	/* check addrstr */
	if(ipv6) {
#if ENABLE_FEATURE_IPV6
		struct in6_addr addr;
		if(inet_pton(AF_INET6, addrstr, &addr) <= 0)
			goto invalid;
		if(!h->ipv6
		  || (0!=memcmp(&addr, &(h->clientaddr_v6), sizeof(struct in6_addr))))
			goto invalid;
#else
		goto invalid;
#endif
	} else {
		struct in_addr addr;
		if(inet_pton(AF_INET, addrstr, &addr) <= 0)
			goto invalid;
#if ENABLE_FEATURE_IPV6
		if(h->ipv6) {
			if(!IN6_IS_ADDR_V4MAPPED(&(h->clientaddr_v6)))
				goto invalid;
			if(0!=memcmp(&addr, ((const char *)&(h->clientaddr_v6) + 12), 4))
				goto invalid;
		} else {
			if(0!=memcmp(&addr, &(h->clientaddr), sizeof(struct in_addr)))
				goto invalid;
		}
#else
		if(0!=memcmp(&addr, &(h->clientaddr), sizeof(struct in_addr)))
			goto invalid;
#endif
	}
	/* select only the good callback url */
	while(p < h->req_buf + h->req_CallbackOff + h->req_CallbackLen && *p != '>')
		p++;
	h->req_CallbackOff++;   /* skip initial '<' */
	h->req_CallbackLen = (int)(p - h->req_buf - h->req_CallbackOff);
	return 1;
invalid:
	while(p < h->req_buf + h->req_CallbackOff + h->req_CallbackLen && *p != '>')
		p++;
	if(*p != '>') return 0;
	while(p < h->req_buf + h->req_CallbackOff + h->req_CallbackLen && *p != '<')
		p++;
	if(*p != '<') return 0;
	h->req_CallbackLen -= (int)(p - h->req_buf - h->req_CallbackOff);
	h->req_CallbackOff = (int)(p - h->req_buf);
	goto start_again;
}
#endif /* ENABLE_EVENTS */

static void SendRespAndClose_upnphttp(struct upnphttp * h)
{
	ssize_t n;

	while (h->res_sent < h->res_buflen) {
		n = send(h->socket, h->res_buf + h->res_sent,
			 h->res_buflen - h->res_sent, 0);
		if(n<0) {
			if(errno == EINTR)
				continue;       /* try again immediatly */
			if(errno == EAGAIN || errno == EWOULDBLOCK) {
				/* try again later */
				h->state = ESendingAndClosing;
				return;
			}
			bb_perror_msg("send(res_buf)");
			break; /* avoid infinite loop */
		} else if(n == 0) {
			bb_error_msg("send(res_buf): %d bytes sent (out of %d)",
							h->res_sent, h->res_buflen);
			break;
		} else {
			h->res_sent += n;
		}
	}
	CloseSocket_upnphttp(h);
}

/* very minimalistic 404 error message */
static void Send404(struct upnphttp *h)
{
	static const char body404[] =
		"<HTML><HEAD><TITLE>404 Not Found</TITLE></HEAD>"
		"<BODY><H1>Not Found</H1>The requested URL was not found"
		" on this server.</BODY></HTML>\r\n";

	h->respflags = FLAG_HTML;
	BuildResp2_upnphttp(h, 404, "Not Found", body404, sizeof(body404) - 1);
	SendRespAndClose_upnphttp(h);
}


/* fill the res_buf buffer with the complete
 * HTTP 200 OK response from the body passed as argument */
static void BuildResp_upnphttp(struct upnphttp *h, const char * body, int bodylen)
{
	BuildResp2_upnphttp(h, 200, "OK", body, bodylen);
}

static int SendResp_upnphttp(struct upnphttp * h)
{
	ssize_t n;

	while (h->res_sent < h->res_buflen) {
		n = send(h->socket, h->res_buf + h->res_sent,
			 h->res_buflen - h->res_sent, 0);
		if(n<0) {
			if(errno == EINTR)
				continue;       /* try again immediatly */
			if(errno == EAGAIN || errno == EWOULDBLOCK) {
				/* try again later */
				return 0;
			}
			bb_perror_msg("send(res_buf)");
			break; /* avoid infinite loop */
		} else if(n == 0) {
			bb_error_msg("send(res_buf): %d bytes sent (out of %d)",
							h->res_sent, h->res_buflen);
			break;
		} else {
			h->res_sent += n;
		}
	}
	return 1;       /* finished */
}


struct NameValue {
	struct NameValue *l_next;
	char name[64];
	char value[128];
};

struct NameValueParserData {
	struct NameValue *l_head;
	char curelt[64];
	char *portListing;
	int portListingLength;
	int topelt;
	const char *cdata;
	int cdatalen;
};


static void NameValueParserStartElt(void *d, const char *name, int l)
{
	struct NameValueParserData *data = (struct NameValueParserData *)d;

	data->topelt = 1;
	if(l > 63)
		l = 63;
	memcpy(data->curelt, name, l);
	data->curelt[l] = '\0';
	data->cdata = NULL;
	data->cdatalen = 0;
}

static void NameValueParserEndElt(void * d, const char *name, int l_unuse)
{
	struct NameValueParserData * data = (struct NameValueParserData *)d;
	struct NameValue * nv;
	(void)name;
	(void)l_unuse;

	if(!data->topelt)
		return;
	if(strcmp(data->curelt, "NewPortListing") != 0) {
		int l;
		/* standard case. Limited to n chars strings */
		l = data->cdatalen;
		nv = malloc(sizeof(struct NameValue));
		if(nv == NULL)
		{
			/* malloc error */
			return;
		}
		if(l>=(int)sizeof(nv->value))
			l = sizeof(nv->value) - 1;
		safe_strncpy(nv->name, data->curelt, 64);
		if(data->cdata != NULL) {
			memcpy(nv->value, data->cdata, l);
			nv->value[l] = '\0';
		} else {
			nv->value[0] = '\0';
		}
		nv->l_next = data->l_head;      /* insert in list */
		data->l_head = nv;
	}
	data->cdata = NULL;
	data->cdatalen = 0;
	data->topelt = 0;
}

static void NameValueParserGetData(void * d, const char * datas, int l)
{
    struct NameValueParserData * data = (struct NameValueParserData *)d;
	if(strcmp(data->curelt, "NewPortListing") == 0) {
		/* specific case for NewPortListing which is a XML Document */
		data->portListing = malloc(l + 1);
		if(!data->portListing) {
			/* malloc error */
			return;
		}
		memcpy(data->portListing, datas, l);
		data->portListing[l] = '\0';
		data->portListingLength = l;
	} else {
		/* standard case. */
		data->cdata = datas;
		data->cdatalen = l;
	}
}

/* minimal xml parser */
#ifndef MINIXML_H_INCLUDED
#define MINIXML_H_INCLUDED
#define IS_WHITE_SPACE(c) ((c==' ') || (c=='\t') || (c=='\r') || (c=='\n'))

/* if a callback function pointer is set to NULL,
 * the function is not called */
struct xmlparser {
	const char *xmlstart;
	const char *xmlend;
	const char *xml;        /* pointer to current character */
	int xmlsize;
	void * data;
	void (*starteltfunc) (void *, const char *, int);
	void (*endeltfunc) (void *, const char *, int);
	void (*datafunc) (void *, const char *, int);
	void (*attfunc) (void *, const char *, int, const char *, int);
};


#endif

/* parseatt : used to parse the argument list
 * return 0 (false) in case of success and -1 (true) if the end
 * of the xmlbuffer is reached. */
static int parseatt(struct xmlparser * p)
{
	const char * attname;
	int attnamelen;
	const char * attvalue;
	int attvaluelen;
	while(p->xml < p->xmlend)
	{
		if(*p->xml=='/' || *p->xml=='>')
			return 0;
		if( !IS_WHITE_SPACE(*p->xml) )
		{
			char sep;
			attname = p->xml;
			attnamelen = 0;
			while(*p->xml!='=' && !IS_WHITE_SPACE(*p->xml) )
			{
				attnamelen++; p->xml++;
				if(p->xml >= p->xmlend)
					return -1;
			}
			while(*(p->xml++) != '=')
			{
				if(p->xml >= p->xmlend)
					return -1;
			}
			while(IS_WHITE_SPACE(*p->xml))
			{
				p->xml++;
				if(p->xml >= p->xmlend)
					return -1;
			}
			sep = *p->xml;
			if(sep=='\'' || sep=='\"')
			{
				p->xml++;
				if(p->xml >= p->xmlend)
					return -1;
				attvalue = p->xml;
				attvaluelen = 0;
				while(*p->xml != sep)
				{
					attvaluelen++; p->xml++;
					if(p->xml >= p->xmlend)
						return -1;
				}
			} else {
				attvalue = p->xml;
				attvaluelen = 0;
				while(   !IS_WHITE_SPACE(*p->xml)
					  && *p->xml != '>' && *p->xml != '/')
				{
					attvaluelen++; p->xml++;
					if(p->xml >= p->xmlend)
						return -1;
				}
			}
			/*printf("%.*s='%.*s'\n",
			       attnamelen, attname, attvaluelen, attvalue);*/
			if(p->attfunc)
				p->attfunc(p->data, attname, attnamelen, attvalue, attvaluelen);
		}
		p->xml++;
	}
	return -1;
}

/* parseelt parse the xml stream and
 * call the callback functions when needed... */
static void parseelt(struct xmlparser * p)
{
	int i;
	const char * elementname;
	while(p->xml < (p->xmlend - 1)) {
		if((p->xml + 4) <= p->xmlend && (0 == memcmp(p->xml, "<!--", 4))) {
			p->xml += 3;
			/* ignore comments */
			do {
				p->xml++;
				if ((p->xml + 3) >= p->xmlend)
					return;
			}
			while(memcmp(p->xml, "-->", 3) != 0);
			p->xml += 3;
		} else if((p->xml)[0]=='<' && (p->xml)[1]!='?') {
			i = 0; elementname = ++p->xml;
			while( !IS_WHITE_SPACE(*p->xml) && (*p->xml!='>') && (*p->xml!='/')) {
				i++; p->xml++;
				if (p->xml >= p->xmlend)
					return;
				/* to ignore namespace : */
				if(*p->xml==':') {
					i = 0;
					elementname = ++p->xml;
				}
			}
			if(i>0) {
				if(p->starteltfunc)
					p->starteltfunc(p->data, elementname, i);
				if(parseatt(p))
					return;
				if(*p->xml!='/') {
					const char * data;
					i = 0; data = ++p->xml;
					if (p->xml >= p->xmlend)
						return;
					while( IS_WHITE_SPACE(*p->xml) ) {
						i++; p->xml++;
						if (p->xml >= p->xmlend)
							return;
					}
					if(memcmp(p->xml, "<![CDATA[", 9) == 0) {
						/* CDATA handling */
						p->xml += 9;
						data = p->xml;
						i = 0;
						while(memcmp(p->xml, "]]>", 3) != 0) {
							i++; p->xml++;
							if ((p->xml + 3) >= p->xmlend)
								return;
						}
						if(i>0 && p->datafunc)
							p->datafunc(p->data, data, i);
						while(*p->xml!='<') {
							p->xml++;
							if (p->xml >= p->xmlend)
								return;
						}
					} else {
						while(*p->xml!='<') {
							i++; p->xml++;
							if ((p->xml + 1) >= p->xmlend)
								return;
						}
						if(i>0 && p->datafunc && *(p->xml + 1) == '/')
							p->datafunc(p->data, data, i);
					}
				}
			} else if(*p->xml == '/') {
				i = 0; elementname = ++p->xml;
				if (p->xml >= p->xmlend)
					return;
				while((*p->xml != '>')) {
					i++; p->xml++;
					if (p->xml >= p->xmlend)
						return;
				}
				if(p->endeltfunc)
					p->endeltfunc(p->data, elementname, i);
				p->xml++;
			}
		} else {
			p->xml++;
		}
	}
}

/* parsexml()
 * the xmlparser structure must be initialized before the call
 * the following structure members have to be initialized :
 * xmlstart, xmlsize, data, *func
 * xml is for internal usage, xmlend is computed automatically */
/* the parser must be initialized before calling this function */
static void parsexml(struct xmlparser * parser)
{
	parser->xml = parser->xmlstart;
	parser->xmlend = parser->xmlstart + parser->xmlsize;
	parseelt(parser);
}

static void
ParseNameValue(const char * buffer, int bufsize,
	       struct NameValueParserData * data)
{
	struct xmlparser parser;
	data->l_head = NULL;
	data->portListing = NULL;
	data->portListingLength = 0;
	/* init xmlparser object */
	parser.xmlstart = buffer;
	parser.xmlsize = bufsize;
	parser.data = data;
	parser.starteltfunc = NameValueParserStartElt;
	parser.endeltfunc = NameValueParserEndElt;
	parser.datafunc = NameValueParserGetData;
	parser.attfunc = 0;
	parsexml(&parser);
}

static void
ClearNameValueList(struct NameValueParserData * pdata)
{
    struct NameValue * nv;
	if(pdata->portListing)
	{
		free(pdata->portListing);
		pdata->portListing = NULL;
		pdata->portListingLength = 0;
	}
    while((nv = pdata->l_head) != NULL)
    {
		pdata->l_head = nv->l_next;
	free(nv);
    }
}

static char *
GetValueFromNameValueList(struct NameValueParserData * pdata,
			  const char * Name)
{
    struct NameValue * nv;
    char * p = NULL;
    for(nv = pdata->l_head;
	(nv != NULL) && (p == NULL);
	nv = nv->l_next)
    {
	if(strcmp(nv->name, Name) == 0)
	    p = nv->value;
    }
    return p;
}

struct ifdata {
	unsigned long opackets;
	unsigned long ipackets;
	unsigned long obytes;
	unsigned long ibytes;
	unsigned long baudrate;
};

/* utility function */
static int is_numeric(const char * s)
{
	while(*s) {
		if(*s < '0' || *s > '9') return 0;
		s++;
	}
	return 1;
}

static void
BuildSendAndCloseSoapResp(struct upnphttp * h,
			  const char * body, int bodylen)
{
	static const char beforebody[] =
		"<?xml version=\"1.0\"?>\r\n"
		"<s:Envelope xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\" "
		"s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\">"
		"<s:Body>";

	static const char afterbody[] =
		"</s:Body>"
		"</s:Envelope>\r\n";

	int r = BuildHeader_upnphttp(h, 200, "OK",  sizeof(beforebody) - 1
				     + sizeof(afterbody) - 1 + bodylen );

	if(r >= 0) {
		memcpy(h->res_buf + h->res_buflen, beforebody, sizeof(beforebody) - 1);
		h->res_buflen += sizeof(beforebody) - 1;

		memcpy(h->res_buf + h->res_buflen, body, bodylen);
		h->res_buflen += bodylen;

		memcpy(h->res_buf + h->res_buflen, afterbody, sizeof(afterbody) - 1);
		h->res_buflen += sizeof(afterbody) - 1;
	} else {
		BuildResp2_upnphttp(h, 500, "Internal Server Error", NULL, 0);
	}

	SendRespAndClose_upnphttp(h);
}

static void
GetConnectionTypeInfo(struct upnphttp * h, const char * action, const char * ns)
{
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewConnectionType>IP_Routed</NewConnectionType>"
		"<NewPossibleConnectionTypes>IP_Routed</NewPossibleConnectionTypes>"
		"</u:%sResponse>";
	char body[512];
	int bodylen;

	bodylen = snprintf(body, sizeof(body), resp,
		 action, ns, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

/* that is the answer */
#define BAUDRATE_DEFAULT 4200000

static int getifstats(const char *ifname, struct ifdata *data)
{
	FILE *f;
	char line[512];
	int r = -1;
	int i;
	config_read_t *crt;

	data->baudrate = BAUDRATE_DEFAULT;
	data->opackets = 0;
	data->ipackets = 0;
	data->obytes = 0;
	data->ibytes = 0;
	if(!ifname || ifname[0]=='\0')
		return -1;
	crt = config_read_prepare("/proc/net/dev", 'w');
	i = 0;
	while( (crt = config_read(crt, "\0~\0\0:")) != NULL) {
		if(i != 2) {
			/* discard the two header lines */
			i++;
			continue;
		}
		/* eth0: ibytes ipackets x1 x2 x3 x4 x5 x6 obytes opackets ... */
		if(crt->ntokens < 11)
			continue;
		/* TODO : how to handle aliases ? */
		if(strcmp(ifname, crt->tokens[0]))
			continue;
		data->ibytes = strtoul(crt->tokens[1], 0, 0);
		data->ipackets = strtoul(crt->tokens[2], 0, 0);
		/* skip 6 columns */
		data->obytes = strtoul(crt->tokens[9], 0, 0);
		data->opackets = strtoul(crt->tokens[10], 0, 0);
		r = 0;
		config_close(crt);
		break;
	}
	/* get interface speed */
	snprintf(line, sizeof(line), "/sys/class/net/%s/speed", ifname);
	f = fopen(line, "r");
	if(f) {
		if(fgets(line, sizeof(line), f)) {
			i = atoi(line); /* 65535 means unknown */
			if(i > 0 && i < 65535)
				data->baudrate = 1000000*i;
		}
		fclose(f);
	}
#ifdef DEBUG
	  else {
		syslog(LOG_INFO, "cannot read %s file : %m", line);
	}
#endif
	return r;
}

/* maximum value for a UPNP ui4 type variable */
#define UPNP_UI4_MAX (4294967295ul)

static void
GetTotalBytesSent(struct upnphttp * h, const char * action, const char * ns)
{
	int r;

	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewTotalBytesSent>%lu</NewTotalBytesSent>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;
	struct ifdata data;

	r = getifstats(bb_common_bss.ext_if_name, &data);
	bodylen = snprintf(body, sizeof(body), resp,
		action, ns, /* was "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1" */
		r<0?0:data.obytes, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetTotalBytesReceived(struct upnphttp * h, const char * action, const char * ns)
{
	int r;

	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewTotalBytesReceived>%lu</NewTotalBytesReceived>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;
	struct ifdata data;

	r = getifstats(bb_common_bss.ext_if_name, &data);
	/* TotalBytesReceived
	 * This variable represents the cumulative counter for total number of
	 * bytes received downstream across all connection service instances on
	 * WANDevice. The count rolls over to 0 after it reaching the maximum
	 * value (2^32)-1. */
	bodylen = snprintf(body, sizeof(body), resp,
		 action, ns, /* was "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1" */
		 r<0?0:data.ibytes, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetTotalPacketsSent(struct upnphttp * h, const char * action, const char * ns)
{
	int r;

	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewTotalPacketsSent>%lu</NewTotalPacketsSent>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;
	struct ifdata data;

	r = getifstats(bb_common_bss.ext_if_name, &data);
	bodylen = snprintf(body, sizeof(body), resp,
		 action, ns,/*"urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",*/
		 r<0?0:data.opackets, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetTotalPacketsReceived(struct upnphttp * h, const char * action, const char * ns)
{
	int r;

	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewTotalPacketsReceived>%lu</NewTotalPacketsReceived>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;
	struct ifdata data;

	r = getifstats(bb_common_bss.ext_if_name, &data);
	bodylen = snprintf(body, sizeof(body), resp,
		 action, ns, /* was "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1" */
		 r<0?0:data.ipackets, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetCommonLinkProperties(struct upnphttp * h, const char * action, const char * ns)
{
	/* WANAccessType : set depending on the hardware :
	 * DSL, POTS (plain old Telephone service), Cable, Ethernet */
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewWANAccessType>%s</NewWANAccessType>"
		"<NewLayer1UpstreamMaxBitRate>%lu</NewLayer1UpstreamMaxBitRate>"
		"<NewLayer1DownstreamMaxBitRate>%lu</NewLayer1DownstreamMaxBitRate>"
		"<NewPhysicalLinkStatus>%s</NewPhysicalLinkStatus>"
		"</u:%sResponse>";

	char body[2048];
	int bodylen;
	struct ifdata data;
	const char * status = "Up";     /* Up, Down (Required),
				     * Initializing, Unavailable (Optional) */
	const char * wan_access_type = "Cable"; /* DSL, POTS, Cable, Ethernet */
	char ext_ip_addr[INET_ADDRSTRLEN];

	if((bb_common_bss.downstream_bitrate == 0) || (bb_common_bss.upstream_bitrate == 0))
	{
		if(getifstats(bb_common_bss.ext_if_name, &data) >= 0)
		{
			if(bb_common_bss.downstream_bitrate == 0)
				bb_common_bss.downstream_bitrate = data.baudrate;
			if(bb_common_bss.upstream_bitrate == 0)
				bb_common_bss.upstream_bitrate = data.baudrate;
		}
	}
	if(getifaddr(bb_common_bss.ext_if_name, ext_ip_addr, INET_ADDRSTRLEN, NULL, NULL) < 0) {
		status = "Down";
	}
	bodylen = snprintf(body, sizeof(body), resp,
	    action, ns, /* was "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1" */
	    wan_access_type,
	    bb_common_bss.upstream_bitrate, bb_common_bss.downstream_bitrate,
	    status, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetStatusInfo(struct upnphttp * h, const char * action, const char * ns)
{
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewConnectionStatus>%s</NewConnectionStatus>"
		"<NewLastConnectionError>ERROR_NONE</NewLastConnectionError>"
		"<NewUptime>%ld</NewUptime>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;
	time_t uptime;
	const char * status;
	/* ConnectionStatus possible values :
	 * Unconfigured, Connecting, Connected, PendingDisconnect,
	 * Disconnecting, Disconnected */

	status = get_wan_connection_status_str(bb_common_bss.ext_if_name);
	uptime = (time(NULL) - bb_common_bss.startup_time);
	bodylen = snprintf(body, sizeof(body), resp,
		action, ns, /*SERVICE_TYPE_WANIPC,*/
		status, (long)uptime, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetNATRSIPStatus(struct upnphttp * h, const char * action, const char * ns)
{
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewRSIPAvailable>0</NewRSIPAvailable>"
		"<NewNATEnabled>1</NewNATEnabled>"
		"</u:%sResponse>";
	char body[512];
	int bodylen;
	/* 2.2.9. RSIPAvailable
	 * This variable indicates if Realm-specific IP (RSIP) is available
	 * as a feature on the InternetGatewayDevice. RSIP is being defined
	 * in the NAT working group in the IETF to allow host-NATing using
	 * a standard set of message exchanges. It also allows end-to-end
	 * applications that otherwise break if NAT is introduced
	 * (e.g. IPsec-based VPNs).
	 * A gateway that does not support RSIP should set this variable to 0. */
	bodylen = snprintf(body, sizeof(body), resp,
		action, ns, /*SERVICE_TYPE_WANIPC,*/
		action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetExternalIPAddress(struct upnphttp * h, const char * action, const char * ns)
{
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewExternalIPAddress>%s</NewExternalIPAddress>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;
	char ext_ip_addr[INET_ADDRSTRLEN];
	/* Does that method need to work with IPv6 ?
	 * There is usually no NAT with IPv6 */

	if(bb_common_bss.use_ext_ip_addr) {
		safe_strncpy(ext_ip_addr, bb_common_bss.use_ext_ip_addr, INET_ADDRSTRLEN);
	} else if(getifaddr(bb_common_bss.ext_if_name, ext_ip_addr, INET_ADDRSTRLEN, NULL, NULL) < 0) {
		bb_error_msg("Failed to get ip address for interface %s",
			bb_common_bss.ext_if_name);
		safe_strncpy(ext_ip_addr, "0.0.0.0", INET_ADDRSTRLEN);
	}
	bodylen = snprintf(body, sizeof(body), resp,
		      action, ns, /*SERVICE_TYPE_WANIPC,*/
				  ext_ip_addr, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

/* Standard Errors:
 *
 * errorCode errorDescription Description
 * --------     ---------------- -----------
 * 401          Invalid Action  No action by that name at this service.
 * 402          Invalid Args    Could be any of the following: not enough in args,
 *                                                      too many in args, no in arg by that name,
 *                                                      one or more in args are of the wrong data type.
 * 403          Out of Sync     Out of synchronization.
 * 501          Action Failed   May be returned in current state of service
 *                                                      prevents invoking that action.
 * 600-699      TBD                     Common action errors. Defined by UPnP Forum
 *                                                      Technical Committee.
 * 700-799      TBD                     Action-specific errors for standard actions.
 *                                                      Defined by UPnP Forum working committee.
 * 800-899      TBD                     Action-specific errors for non-standard actions.
 *                                                      Defined by UPnP vendor.
*/
static void SoapError(struct upnphttp * h, int errCode, const char * errDesc)
{
	static const char resp[] =
		"<s:Envelope "
		"xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\" "
		"s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\">"
		"<s:Body>"
		"<s:Fault>"
		"<faultcode>s:Client</faultcode>"
		"<faultstring>UPnPError</faultstring>"
		"<detail>"
		"<UPnPError xmlns=\"urn:schemas-upnp-org:control-1-0\">"
		"<errorCode>%d</errorCode>"
		"<errorDescription>%s</errorDescription>"
		"</UPnPError>"
		"</detail>"
		"</s:Fault>"
		"</s:Body>"
		"</s:Envelope>";

	char body[2048];
	int bodylen;

	DEBUG_LOG(syslog(LOG_INFO, "Returning UPnPError %d: %s", errCode, errDesc);)
	bodylen = snprintf(body, sizeof(body), resp, errCode, errDesc);
	BuildResp2_upnphttp(h, 500, "Internal Server Error", body, bodylen);
	SendRespAndClose_upnphttp(h);
}

static const char *miniupnpd_gethostbyname4(const char *int_ip, struct in_addr *presult_ip)
{
	if (inet_pton(AF_INET, int_ip, presult_ip) <= 0) {
		/* TODO : deal with more than one ip per hostname */
		if(bb_gethostbyname(BB_GHBN_IN_ADDR | BB_GHBN_ADD_ERR_MSG, int_ip, presult_ip,
					"Failed to convert hostname to ip address") != NULL) {
			return inet_ntoa(*presult_ip);
		} else {
			return NULL;
		}
	}
	return int_ip;
}

/* AddPortMapping method of WANIPConnection Service
 * Ignored argument : NewEnabled */
static void
AddPortMapping(struct upnphttp * h, const char * action, const char * ns)
{
	int r;

	/*static const char resp[] =
		"<u:AddPortMappingResponse "
		"xmlns:u=\"" SERVICE_TYPE_WANIPC "\"/>";*/
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\"/>";

	char body[512];
	int bodylen;
	struct NameValueParserData data;
	const char *int_ip;
	char *int_port, *ext_port, *protocol, *desc;
	char * leaseduration_str;
	unsigned int leaseduration;
	char *r_host;
	unsigned short iport, eport;

	struct in_addr result_ip;/*unsigned char result_ip[16];*/ /* inet_pton() */

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
	int_ip = GetValueFromNameValueList(&data, "NewInternalClient");
	if (!int_ip) {
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}

	/* IGD 2 MUST support both wildcard and specific IP address values
	 * for RemoteHost (only the wildcard value was REQUIRED in release 1.0) */
	r_host = GetValueFromNameValueList(&data, "NewRemoteHost");

	/* if ip not valid assume hostname and convert */
	int_ip = miniupnpd_gethostbyname4(int_ip, &result_ip);
	if(int_ip == NULL) {
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}

	/* check if NewInternalAddress is the client address */
	if(GETFLAG(SECUREMODEMASK)) {
		if(h->clientaddr.s_addr != result_ip.s_addr) {
			DEBUG_LOG(syslog(LOG_INFO, "Client %s tried to redirect port to %s",
			       inet_ntoa(h->clientaddr), int_ip);)
			ClearNameValueList(&data);
			SoapError(h, 718, "ConflictInMappingEntry");
			return;
		}
	}

	int_port = GetValueFromNameValueList(&data, "NewInternalPort");
	ext_port = GetValueFromNameValueList(&data, "NewExternalPort");
	protocol = GetValueFromNameValueList(&data, "NewProtocol");
	desc = GetValueFromNameValueList(&data, "NewPortMappingDescription");
	leaseduration_str = GetValueFromNameValueList(&data, "NewLeaseDuration");

	if (!int_port || !ext_port || !protocol) {
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}

	eport = (unsigned short)atoi(ext_port);
	iport = (unsigned short)atoi(int_port);

	if (strcmp(ext_port, "*") == 0 || eport == 0) {
		ClearNameValueList(&data);
		SoapError(h, 716, "Wildcard not permited in ExtPort");
		return;
	}

	leaseduration = leaseduration_str ? atoi(leaseduration_str) : 0;
	DEBUG_LOG(syslog(LOG_INFO, "%s: ext port %hu to %s:%hu protocol %s for: %s leaseduration=%u rhost=%s",
	       action, eport, int_ip, iport, protocol, desc, leaseduration,
	       r_host ? r_host : "NULL");)

	r = upnp_redirect(r_host, eport, int_ip, iport, protocol, desc, leaseduration);

	ClearNameValueList(&data);

	/* possible error codes for AddPortMapping :
	 * 402 - Invalid Args
	 * 501 - Action Failed
	 * 715 - Wildcard not permited in SrcAddr
	 * 716 - Wildcard not permited in ExtPort
	 * 718 - ConflictInMappingEntry
	 * 724 - SamePortValuesRequired (deprecated in IGD v2)
     * 725 - OnlyPermanentLeasesSupported
	     The NAT implementation only supports permanent lease times on
	     port mappings (deprecated in IGD v2)
     * 726 - RemoteHostOnlySupportsWildcard
	     RemoteHost must be a wildcard and cannot be a specific IP
	     address or DNS name (deprecated in IGD v2)
     * 727 - ExternalPortOnlySupportsWildcard
	     ExternalPort must be a wildcard and cannot be a specific port
	     value (deprecated in IGD v2)
     * 728 - NoPortMapsAvailable
	     There are not enough free ports available to complete the mapping
	     (added in IGD v2)
	 * 729 - ConflictWithOtherMechanisms (added in IGD v2) */
	switch(r)
	{
	case 0: /* success */
		bodylen = snprintf(body, sizeof(body), resp,
				   action, ns/*SERVICE_TYPE_WANIPC*/);
		BuildSendAndCloseSoapResp(h, body, bodylen);
		break;
	case -4:
	case -2:        /* already redirected */
	case -3:        /* not permitted */
		SoapError(h, 718, "ConflictInMappingEntry");
		break;
	default:
		SoapError(h, 501, "ActionFailed");
	}
}

/* AddAnyPortMapping was added in WANIPConnection v2 */
static void
AddAnyPortMapping(struct upnphttp * h, const char * action, const char * ns)
{
	int r;
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewReservedPort>%hu</NewReservedPort>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;

	struct NameValueParserData data;
	const char * int_ip, * int_port, * ext_port, * protocol, * desc;
	const char * r_host;
	unsigned short iport, eport;
	const char * leaseduration_str;
	unsigned int leaseduration;

	struct in_addr result_ip;/*unsigned char result_ip[16];*/ /* inet_pton() */

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
	r_host = GetValueFromNameValueList(&data, "NewRemoteHost");
	ext_port = GetValueFromNameValueList(&data, "NewExternalPort");
	protocol = GetValueFromNameValueList(&data, "NewProtocol");
	int_port = GetValueFromNameValueList(&data, "NewInternalPort");
	int_ip = GetValueFromNameValueList(&data, "NewInternalClient");
	/* NewEnabled */
	desc = GetValueFromNameValueList(&data, "NewPortMappingDescription");
	leaseduration_str = GetValueFromNameValueList(&data, "NewLeaseDuration");

	leaseduration = leaseduration_str ? atoi(leaseduration_str) : 0;
	if(leaseduration == 0)
		leaseduration = 604800;

	if (!int_ip || !ext_port || !int_port) {
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}

	eport = (unsigned short)atoi(ext_port);
	iport = (unsigned short)atoi(int_port);
	if(iport == 0 || !is_numeric(ext_port)) {
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}

	/* if ip not valid assume hostname and convert */
	int_ip = miniupnpd_gethostbyname4(int_ip, &result_ip);
	if(int_ip == NULL) {
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}

	/* check if NewInternalAddress is the client address */
	if(GETFLAG(SECUREMODEMASK)) {
		if(h->clientaddr.s_addr != result_ip.s_addr) {
			DEBUG_LOG(syslog(LOG_INFO, "Client %s tried to redirect port to %s",
			       inet_ntoa(h->clientaddr), int_ip);)
			ClearNameValueList(&data);
			SoapError(h, 606, "Action not authorized");
			return;
		}
	}

	/* TODO : accept a different external port
	 * have some smart strategy to choose the port */
	for(;;) {
		r = upnp_redirect(r_host, eport, int_ip, iport, protocol, desc, leaseduration);
		if(r==-2 && eport < 65535) {
			eport++;
		} else {
			break;
		}
	}

	ClearNameValueList(&data);

	switch(r)
	{
	case 0: /* success */
		bodylen = snprintf(body, sizeof(body), resp,
			      action, ns, /*SERVICE_TYPE_WANIPC,*/
					  eport, action);
		BuildSendAndCloseSoapResp(h, body, bodylen);
		break;
	case -2:        /* already redirected */
		SoapError(h, 718, "ConflictInMappingEntry");
		break;
	case -3:        /* not permitted */
		SoapError(h, 606, "Action not authorized");
		break;
	default:
		SoapError(h, 501, "ActionFailed");
	}
}

static void
GetSpecificPortMappingEntry(struct upnphttp * h, const char * action, const char * ns)
{
	int r;

	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewInternalPort>%u</NewInternalPort>"
		"<NewInternalClient>%s</NewInternalClient>"
		"<NewEnabled>1</NewEnabled>"
		"<NewPortMappingDescription>%s</NewPortMappingDescription>"
		"<NewLeaseDuration>%u</NewLeaseDuration>"
		"</u:%sResponse>";

	char body[1024];
	int bodylen;
	struct NameValueParserData data;
	const char *ext_port, *protocol;
#ifdef DEBUG
	const char *r_host;
#endif
	unsigned short eport, iport;
	char int_ip[32];
	char desc[64];
	unsigned int leaseduration = 0;

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
#ifdef DEBUG
	r_host = GetValueFromNameValueList(&data, "NewRemoteHost");
#endif
	ext_port = GetValueFromNameValueList(&data, "NewExternalPort");
	protocol = GetValueFromNameValueList(&data, "NewProtocol");

	if(!ext_port || !protocol)
	{
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}

	eport = (unsigned short)atoi(ext_port);
	if(eport == 0)
	{
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}

	/* TODO : add r_host as an input parameter ...
	 * We prevent several Port Mapping with same external port
	 * but different remoteHost to be set up, so that is not
	 * a priority. */
	r = upnp_get_redirection_infos(eport, protocol, &iport,
				       int_ip, sizeof(int_ip),
				       desc, sizeof(desc),
				       NULL, 0,
				       &leaseduration);

	if(r < 0) {
		SoapError(h, 714, "NoSuchEntryInArray");
	} else {
		DEBUG_LOG(syslog(LOG_INFO, "%s: rhost='%s' %s %s found => %s:%u desc='%s'",
					action,
					r_host ? r_host : "NULL", ext_port, protocol, int_ip,
					(unsigned int)iport, desc);)
		bodylen = snprintf(body, sizeof(body), resp,
				action, ns/*SERVICE_TYPE_WANIPC*/,
				(unsigned int)iport, int_ip, desc, leaseduration,
				action);
		BuildSendAndCloseSoapResp(h, body, bodylen);
	}

	ClearNameValueList(&data);
}

static void
DeletePortMapping(struct upnphttp * h, const char * action, const char * ns)
{
	int r;

	/*static const char resp[] =
		"<u:DeletePortMappingResponse "
		"xmlns:u=\"" SERVICE_TYPE_WANIPC "\">"
		"</u:DeletePortMappingResponse>";*/
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"</u:%sResponse>";

	char body[512];
	int bodylen;
	struct NameValueParserData data;
	const char * ext_port, * protocol;
	unsigned short eport;

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
	ext_port = GetValueFromNameValueList(&data, "NewExternalPort");
	protocol = GetValueFromNameValueList(&data, "NewProtocol");

	if(!ext_port || !protocol)
	{
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}

	eport = (unsigned short)atoi(ext_port);
	if(eport == 0) {
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}
	DEBUG_LOG(syslog(LOG_INFO, "%s: external port: %hu, protocol: %s", action, eport, protocol);)

	/* if in secure mode, check the IP
	 * Removing a redirection is not a security threat,
	 * just an annoyance for the user using it. So this is not
	 * a priority. */
	if(GETFLAG(SECUREMODEMASK)) {
		char int_ip[32];
		struct in_addr int_ip_addr;
		unsigned short iport;
		unsigned int leaseduration = 0;
		r = upnp_get_redirection_infos(eport, protocol, &iport,
					       int_ip, sizeof(int_ip),
					       NULL, 0, NULL, 0,
					       &leaseduration);
		if(r >= 0) {
			if(inet_pton(AF_INET, int_ip, &int_ip_addr) > 0) {
				if(h->clientaddr.s_addr != int_ip_addr.s_addr) {
					SoapError(h, 606, "Action not authorized");
					/*SoapError(h, 714, "NoSuchEntryInArray");*/
					ClearNameValueList(&data);
					return;
				}
			}
		}
	}

	r = upnp_delete_redirection(eport, protocol);

	if(r < 0)
	{
		SoapError(h, 714, "NoSuchEntryInArray");
	}
	else
	{
		bodylen = snprintf(body, sizeof(body), resp,
				   action, ns, action);
		BuildSendAndCloseSoapResp(h, body, bodylen);
	}

	ClearNameValueList(&data);
}

/* DeletePortMappingRange was added in IGD spec v2 */
static void
DeletePortMappingRange(struct upnphttp *h, const char *action, const char *ns)
{
	/*static const char resp[] =
		"<u:DeletePortMappingRangeResponse "
		"xmlns:u=\"" SERVICE_TYPE_WANIPC "\">"
		"</u:DeletePortMappingRangeResponse>";*/
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"</u:%sResponse>";
	char body[512];
	int bodylen;
	struct NameValueParserData data;
	const char * protocol;
	const char * startport_s, * endport_s;
	unsigned short startport, endport;
	/*int manage;*/
	unsigned short * port_list;
	unsigned int i, number = 0;

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
	startport_s = GetValueFromNameValueList(&data, "NewStartPort");
	endport_s = GetValueFromNameValueList(&data, "NewEndPort");
	protocol = GetValueFromNameValueList(&data, "NewProtocol");
	/*manage = atoi(GetValueFromNameValueList(&data, "NewManage"));*/
	if(startport_s == NULL || endport_s == NULL || protocol == NULL ||
	   !is_numeric(startport_s) || !is_numeric(endport_s)) {
		SoapError(h, 402, "Invalid Args");
		ClearNameValueList(&data);
		return;
	}
	startport = (unsigned short)atoi(startport_s);
	endport = (unsigned short)atoi(endport_s);

	/* possible errors :
	   606 - Action not authorized
	   730 - PortMappingNotFound
	   733 - InconsistentParameter
	 */
	if(startport > endport) {
		SoapError(h, 733, "InconsistentParameter");
		ClearNameValueList(&data);
		return;
	}
	DEBUG_LOG(syslog(LOG_INFO, "%s: deleting external ports: %hu-%hu, protocol: %s", action, startport, endport, protocol);)

	port_list = upnp_get_portmappings_in_range(startport, endport,
						   protocol, &number);
	if(number == 0) {
		SoapError(h, 730, "PortMappingNotFound");
		ClearNameValueList(&data);
		free(port_list);
		return;
	}

	for(i = 0; i < number; i++) {
#ifdef DEBUG
		int r =
#endif
			upnp_delete_redirection(port_list[i], protocol);
		DEBUG_LOG(syslog(LOG_INFO, "%s: deleting external port: %hu, protocol: %s: %s",
		       action, port_list[i], protocol, r < 0 ? "failed" : "ok");)
	}
	free(port_list);
	bodylen = snprintf(body, sizeof(body), resp,
			   action, ns, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);

	ClearNameValueList(&data);
}

static void
GetGenericPortMappingEntry(struct upnphttp * h, const char * action, const char * ns)
{
	int r;

	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewRemoteHost>%s</NewRemoteHost>"
		"<NewExternalPort>%u</NewExternalPort>"
		"<NewProtocol>%s</NewProtocol>"
		"<NewInternalPort>%u</NewInternalPort>"
		"<NewInternalClient>%s</NewInternalClient>"
		"<NewEnabled>1</NewEnabled>"
		"<NewPortMappingDescription>%s</NewPortMappingDescription>"
		"<NewLeaseDuration>%u</NewLeaseDuration>"
		"</u:%sResponse>";

	long int index = 0;
	unsigned short eport, iport;
	const char * m_index;
	char * endptr;
	char protocol[8], iaddr[32];
	char desc[64];
	char rhost[40];
	unsigned int leaseduration = 0;
	struct NameValueParserData data;

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
	m_index = GetValueFromNameValueList(&data, "NewPortMappingIndex");

	if(!m_index) {
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}
	errno = 0;      /* To distinguish success/failure after call */
	index = strtol(m_index, &endptr, 10);
	if((errno == ERANGE && (index == LONG_MAX || index == LONG_MIN))
	   || (errno != 0 && index == 0) || (m_index == endptr))
	{
		/* should condition (*endptr != '\0') be also an error ? */
		if(m_index == endptr)
			xlog_warn("%s: no digits were found in <%s>",
			       "GetGenericPortMappingEntry", "NewPortMappingIndex");
		else
			xlog_warn("%s: strtol('%s'): %m",
			       "GetGenericPortMappingEntry", m_index);
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}
	DEBUG_LOG(syslog(LOG_INFO, "%s: index=%d", action, (int)index);)

	rhost[0] = '\0';
	r = upnp_get_redirection_infos_by_index((int)index, &eport, protocol, &iport,
					    iaddr, sizeof(iaddr),
						desc, sizeof(desc),
						rhost, sizeof(rhost),
						&leaseduration);

	if(r < 0) {
		SoapError(h, 713, "SpecifiedArrayIndexInvalid");
	} else {
		int bodylen;
		char body[2048];
		bodylen = snprintf(body, sizeof(body), resp,
			action, ns, /*SERVICE_TYPE_WANIPC,*/ rhost,
			(unsigned int)eport, protocol, (unsigned int)iport, iaddr, desc,
		    leaseduration, action);
		BuildSendAndCloseSoapResp(h, body, bodylen);
	}

	ClearNameValueList(&data);
}

/* GetListOfPortMappings was added in the IGD v2 specification */
static void
GetListOfPortMappings(struct upnphttp * h, const char * action, const char * ns)
{
	static const char resp_start[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewPortListing><![CDATA[";
	static const char resp_end[] =
		"]]></NewPortListing>"
		"</u:%sResponse>";

	static const char list_start[] =
		"<p:PortMappingList xmlns:p=\"urn:schemas-upnp-org:gw:WANIPConnection\""
		" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\""
		" xsi:schemaLocation=\"urn:schemas-upnp-org:gw:WANIPConnection"
		" http://www.upnp.org/schemas/gw/WANIPConnection-v2.xsd\">";
	static const char list_end[] =
		"</p:PortMappingList>";

	static const char entry[] =
		"<p:PortMappingEntry>"
		"<p:NewRemoteHost>%s</p:NewRemoteHost>"
		"<p:NewExternalPort>%hu</p:NewExternalPort>"
		"<p:NewProtocol>%s</p:NewProtocol>"
		"<p:NewInternalPort>%hu</p:NewInternalPort>"
		"<p:NewInternalClient>%s</p:NewInternalClient>"
		"<p:NewEnabled>1</p:NewEnabled>"
		"<p:NewDescription>%s</p:NewDescription>"
		"<p:NewLeaseTime>%u</p:NewLeaseTime>"
		"</p:PortMappingEntry>";

	char * body;
	size_t bodyalloc;
	int bodylen;

	int r = -1;
	unsigned short iport;
	char int_ip[32];
	char desc[64];
	char rhost[64];
	unsigned int leaseduration = 0;

	struct NameValueParserData data;
	const char * startport_s, * endport_s;
	unsigned short startport, endport;
	const char * protocol;
	/*int manage;*/
	const char * number_s;
	int number;
	unsigned short * port_list;
	unsigned int i, list_size = 0;

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
	startport_s = GetValueFromNameValueList(&data, "NewStartPort");
	endport_s = GetValueFromNameValueList(&data, "NewEndPort");
	protocol = GetValueFromNameValueList(&data, "NewProtocol");
	/*manage_s = GetValueFromNameValueList(&data, "NewManage");*/
	number_s = GetValueFromNameValueList(&data, "NewNumberOfPorts");
	if(startport_s == NULL || endport_s == NULL || protocol == NULL ||
	   number_s == NULL || !is_numeric(number_s) ||
	   !is_numeric(startport_s) || !is_numeric(endport_s)) {
		SoapError(h, 402, "Invalid Args");
		ClearNameValueList(&data);
		return;
	}

	startport = (unsigned short)atoi(startport_s);
	endport = (unsigned short)atoi(endport_s);
	/*manage = atoi(manage_s);*/
	number = atoi(number_s);
	if(number == 0) number = 1000;  /* return up to 1000 mappings by default */

	if(startport > endport) {
		SoapError(h, 733, "InconsistentParameter");
		ClearNameValueList(&data);
		return;
	}
/*
build the PortMappingList xml document :

<p:PortMappingList xmlns:p="urn:schemas-upnp-org:gw:WANIPConnection"
xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
xsi:schemaLocation="urn:schemas-upnp-org:gw:WANIPConnection
http://www.upnp.org/schemas/gw/WANIPConnection-v2.xsd">
<p:PortMappingEntry>
<p:NewRemoteHost>202.233.2.1</p:NewRemoteHost>
<p:NewExternalPort>2345</p:NewExternalPort>
<p:NewProtocol>TCP</p:NewProtocol>
<p:NewInternalPort>2345</p:NewInternalPort>
<p:NewInternalClient>192.168.1.137</p:NewInternalClient>
<p:NewEnabled>1</p:NewEnabled>
<p:NewDescription>dooom</p:NewDescription>
<p:NewLeaseTime>345</p:NewLeaseTime>
</p:PortMappingEntry>
</p:PortMappingList>
*/
	bodyalloc = 4096;
	body = malloc(bodyalloc);
	if(!body) {
		ClearNameValueList(&data);
		SoapError(h, 501, "ActionFailed");
		return;
	}
	bodylen = snprintf(body, bodyalloc, resp_start,
		      action, ns/*SERVICE_TYPE_WANIPC*/);
	if(bodylen < 0) {
		SoapError(h, 501, "ActionFailed");
		free(body);
		return;
	}
	memcpy(body+bodylen, list_start, sizeof(list_start));
	bodylen += (sizeof(list_start) - 1);

	port_list = upnp_get_portmappings_in_range(startport, endport,
						   protocol, &list_size);
	/* loop through port mappings */
	for(i = 0; number > 0 && i < list_size; i++) {
		/* have a margin of 1024 bytes to store the new entry */
		if((unsigned int)bodylen + 1024 > bodyalloc) {
			char * body_sav = body;
			bodyalloc += 4096;
			body = realloc(body, bodyalloc);
			if(!body) {
				syslog(LOG_CRIT, "realloc(%p, %u) FAILED", body_sav, (unsigned)bodyalloc);
				ClearNameValueList(&data);
				SoapError(h, 501, "ActionFailed");
				free(body_sav);
				free(port_list);
				return;
			}
		}
		rhost[0] = '\0';
		r = upnp_get_redirection_infos(port_list[i], protocol, &iport,
					       int_ip, sizeof(int_ip),
					       desc, sizeof(desc),
					       rhost, sizeof(rhost),
					       &leaseduration);
		if(r == 0) {
			bodylen += snprintf(body+bodylen, bodyalloc-bodylen, entry,
					    rhost, port_list[i], protocol,
					    iport, int_ip, desc, leaseduration);
			number--;
		}
	}
	free(port_list);
	port_list = NULL;

	if((bodylen + sizeof(list_end) + 1024) > bodyalloc) {
		char * body_sav = body;
		bodyalloc += (sizeof(list_end) + 1024);
		body = realloc(body, bodyalloc);
		if(!body) {
			syslog(LOG_CRIT, "realloc(%p, %u) FAILED", body_sav, (unsigned)bodyalloc);
			ClearNameValueList(&data);
			SoapError(h, 501, "ActionFailed");
			free(body_sav);
			return;
		}
	}
	memcpy(body+bodylen, list_end, sizeof(list_end));
	bodylen += (sizeof(list_end) - 1);
	bodylen += snprintf(body+bodylen, bodyalloc-bodylen, resp_end,
			    action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
	free(body);

	ClearNameValueList(&data);
}

#ifdef ENABLE_L3F_SERVICE
static void
SetDefaultConnectionService(struct upnphttp * h, const char * action, const char * ns)
{
	/*static const char resp[] =
		"<u:SetDefaultConnectionServiceResponse "
		"xmlns:u=\"urn:schemas-upnp-org:service:Layer3Forwarding:1\">"
		"</u:SetDefaultConnectionServiceResponse>";*/
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"</u:%sResponse>";
	char body[512];
	int bodylen;
	struct NameValueParserData data;
	char * p;
	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
	p = GetValueFromNameValueList(&data, "NewDefaultConnectionService");
	if(p) {
		/* 720 InvalidDeviceUUID
		 * 721 InvalidServiceID
		 * 723 InvalidConnServiceSelection */
		DEBUG_LOG(syslog(LOG_INFO, "%s(%s) : Ignored", action, p);)
		bodylen = snprintf(body, sizeof(body), resp,
					   action, ns, action);
		BuildSendAndCloseSoapResp(h, body, bodylen);
	} else {
		/* missing argument */
		SoapError(h, 402, "Invalid Args");
	}
	ClearNameValueList(&data);
}

static void
GetDefaultConnectionService(struct upnphttp * h, const char * action, const char * ns)
{
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewDefaultConnectionService>%s:WANConnectionDevice:1,"
		SERVICE_ID_WANIPC "</NewDefaultConnectionService>"
		"</u:%sResponse>";
	/* example from UPnP_IGD_Layer3Forwarding 1.0.pdf :
	 * uuid:44f5824f-c57d-418c-a131-f22b34e14111:WANConnectionDevice:1,
	 * urn:upnp-org:serviceId:WANPPPConn1 */
	char body[1024];
	int bodylen;

	/* namespace : urn:schemas-upnp-org:service:Layer3Forwarding:1 */
	bodylen = snprintf(body, sizeof(body), resp,
			   action, ns, bb_common_bss.uuidvalue_wcd, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}
#endif

/* Added for compliance with WANIPConnection v2 */
static void
SetConnectionType(struct upnphttp * h, const char * action, const char * ns)
{
	struct NameValueParserData data;
	UNUSED(action);
	UNUSED(ns);

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
	/* Unconfigured, IP_Routed, IP_Bridged */
	ClearNameValueList(&data);
	/* always return a ReadOnly error */
	SoapError(h, 731, "ReadOnly");
}

/* Added for compliance with WANIPConnection v2 */
static void
RequestConnection(struct upnphttp * h, const char * action, const char * ns)
{
	UNUSED(action);
	UNUSED(ns);
	SoapError(h, 606, "Action not authorized");
}

/* Added for compliance with WANIPConnection v2 */
static void
ForceTermination(struct upnphttp * h, const char * action, const char * ns)
{
	UNUSED(action);
	UNUSED(ns);
	SoapError(h, 606, "Action not authorized");
}

/*
If a control point calls QueryStateVariable on a state variable that is not
buffered in memory within (or otherwise available from) the service,
the service must return a SOAP fault with an errorCode of 404 Invalid Var.

QueryStateVariable remains useful as a limited test tool but may not be
part of some future versions of UPnP.
*/
static void
QueryStateVariable(struct upnphttp * h, const char * action, const char * ns)
{
	static const char resp[] =
	"<u:%sResponse "
	"xmlns:u=\"%s\">"
		"<return>%s</return>"
	"</u:%sResponse>";

	char body[512];
	int bodylen;
	struct NameValueParserData data;
	const char * var_name;

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
	/*var_name = GetValueFromNameValueList(&data, "QueryStateVariable"); */
	var_name = GetValueFromNameValueList(&data, "varName");

	/*syslog(LOG_INFO, "QueryStateVariable(%.40s)", var_name); */

	if(!var_name) {
		SoapError(h, 402, "Invalid Args");
	} else if(strcmp(var_name, "ConnectionStatus") == 0) {
		const char * status;

		status = get_wan_connection_status_str(bb_common_bss.ext_if_name);
		bodylen = snprintf(body, sizeof(body), resp,
			   action, ns,/*"urn:schemas-upnp-org:control-1-0",*/
				   status, action);
		BuildSendAndCloseSoapResp(h, body, bodylen);
	} else if(strcmp(var_name, "PortMappingNumberOfEntries") == 0) {
		char strn[10];
		snprintf(strn, sizeof(strn), "%i",
			 upnp_get_portmapping_number_of_entries());
		bodylen = snprintf(body, sizeof(body), resp,
			   action, ns,/*"urn:schemas-upnp-org:control-1-0",*/
				   strn, action);
		BuildSendAndCloseSoapResp(h, body, bodylen);
	} else {
		DEBUG_LOG(syslog(LOG_NOTICE, "%s: Unknown: %s", action, var_name?var_name:"");)
		SoapError(h, 404, "Invalid Var");
	}

	ClearNameValueList(&data);
}

#ifdef ENABLE_UPNPPINHOLE

#include <libiptc/libip6tc.h>

#define IP6TC_HANDLE struct xtc_handle6 *

static int next_uid = 1;

static LIST_HEAD(pinhole_list_t, pinhole_t) pinhole_list;

static struct pinhole_t *
get_pinhole(unsigned short uid);

struct pinhole_t {
	struct in6_addr saddr;
	struct in6_addr daddr;
	LIST_ENTRY(pinhole_t) entries;
	unsigned int timestamp;
	unsigned short sport;
	unsigned short dport;
	unsigned short uid;
	unsigned char proto;
	char desc[];
};

static void init_iptpinhole(void)
{
	LIST_INIT(&pinhole_list);
}

/* return uid */
static int
add_to_pinhole_list(struct in6_addr * saddr, unsigned short sport,
		    struct in6_addr * daddr, unsigned short dport,
		    int proto, const char *desc, unsigned int timestamp)
{
	struct pinhole_t * p;

	p = calloc(1, sizeof(struct pinhole_t) + strlen(desc) + 1);
	if(!p) {
		bb_error_msg("add_to_pinhole_list calloc() error");
		return -1;
	}
	strcpy(p->desc, desc);
	memcpy(&p->saddr, saddr, sizeof(struct in6_addr));
	p->sport = sport;
	memcpy(&p->daddr, daddr, sizeof(struct in6_addr));
	p->dport = dport;
	p->timestamp = timestamp;
	p->proto = (unsigned char)proto;
	LIST_INSERT_HEAD(&pinhole_list, p, entries);
	while(get_pinhole(next_uid) != NULL) {
		next_uid++;
		if(next_uid > 65535)
			next_uid = 1;
	}
	p->uid = next_uid;
	next_uid++;
	if(next_uid > 65535)
		next_uid = 1;
	return p->uid;
}

static struct pinhole_t *
get_pinhole(unsigned short uid)
{
	struct pinhole_t * p;

	for(p = pinhole_list.lh_first; p != NULL; p = p->entries.le_next) {
		if(p->uid == uid)
			return p;
	}
	return NULL;    /* not found */
}

/* new_match()
 * Allocate and set a new ip6t_entry_match structure
 * The caller must free() it after usage */
static struct ip6t_entry_match *
new_match(int proto, unsigned short sport, unsigned short dport)
{
	struct ip6t_entry_match *match;
	struct ip6t_tcp *info;  /* TODO : use ip6t_udp if needed */
	size_t size;
	const char * name;
	size =   XT_ALIGN(sizeof(struct ip6t_entry_match))
	       + XT_ALIGN(sizeof(struct ip6t_tcp));
	match = calloc(1, size);
	match->u.user.match_size = size;
	switch(proto) {
	case IPPROTO_TCP:
		name = "tcp";
		break;
	case IPPROTO_UDP:
		name = "udp";
		break;
	case IPPROTO_UDPLITE:
		name = "udplite";
		break;
	default:
		name = NULL;
	}
	if(name)
		safe_strncpy(match->u.user.name, name, sizeof(match->u.user.name));
	else
		xlog_warn("no name for protocol %d", proto);
	info = (struct ip6t_tcp *)match->data;
	if(sport) {
		info->spts[0] = sport;  /* specified source port */
		info->spts[1] = sport;
	} else {
		info->spts[0] = 0;              /* all source ports */
		info->spts[1] = 0xFFFF;
	}
	info->dpts[0] = dport;  /* specified destination port */
	info->dpts[1] = dport;
	return match;
}

static struct ip6t_entry_target *
get_accept_target6(void)
{
	struct ip6t_entry_target * target = NULL;
	size_t size;
	size =   XT_ALIGN(sizeof(struct ip6t_entry_target))
	       + XT_ALIGN(sizeof(int));
	target = calloc(1, size);
	target->u.user.target_size = size;
	safe_strncpy(target->u.user.name, "ACCEPT", sizeof(target->u.user.name));
	return target;
}

static int
ip6tc_init_verify_append(const char * table,
			 const char * chain,
			 struct ip6t_entry * e)
{
	IP6TC_HANDLE h;

	h = ip6tc_init(table);
	if(!h) {
		bb_error_msg("ip6tc_init error : %s", ip6tc_strerror(errno));
		return -1;
	}
	if(!ip6tc_is_chain(chain, h)) {
		bb_error_msg("chain %s not found", chain);
		goto error;
	}
	if(!ip6tc_append_entry(chain, e, h)) {
		bb_error_msg("ip6tc_append_entry() error : %s", ip6tc_strerror(errno));
		goto error;
	}
	if(!ip6tc_commit(h)) {
		bb_error_msg("ip6tc_commit() error : %s", ip6tc_strerror(errno));
		goto error;
	}
	return 0;       /* ok */
error:
	ip6tc_free(h);
	return -1;
}

/*
ip6tables -I %s %d -p %s -i %s -s %s --sport %hu -d %s --dport %hu -j ACCEPT
ip6tables -I %s %d -p %s -i %s --sport %hu -d %s --dport %hu -j ACCEPT

miniupnpd_forward_chain, line_number, proto, ext_if_name, raddr, rport, iaddr, iport

ip6tables -t raw -I PREROUTING %d -p %s -i %s -s %s --sport %hu -d %s --dport %hu -j TRACE
ip6tables -t raw -I PREROUTING %d -p %s -i %s --sport %hu -d %s --dport %hu -j TRACE
*/
static int add_pinhole(const char * ifname,
		const char * rem_host, unsigned short rem_port,
		const char * int_client, unsigned short int_port,
		int proto, const char * desc, unsigned int timestamp)
{
	int uid;
	struct ip6t_entry * e;
	struct ip6t_entry * tmp;
	struct ip6t_entry_match *match = NULL;
	struct ip6t_entry_target *target = NULL;

	e = calloc(1, sizeof(struct ip6t_entry));
	if(!e) {
		bb_error_msg("%s: calloc(%d) failed",
		       "add_pinhole", (int)sizeof(struct ip6t_entry));
		return -1;
	}
	e->ipv6.proto = proto;
	if (proto)
		e->ipv6.flags |= IP6T_F_PROTO;

	if(ifname)
		safe_strncpy(e->ipv6.iniface, ifname, IFNAMSIZ);
	if(rem_host && (rem_host[0] != '\0')) {
		inet_pton(AF_INET6, rem_host, &e->ipv6.src);
		memset(&e->ipv6.smsk, 0xff, sizeof(e->ipv6.smsk));
	}
	inet_pton(AF_INET6, int_client, &e->ipv6.dst);
	memset(&e->ipv6.dmsk, 0xff, sizeof(e->ipv6.dmsk));

	/*e->nfcache = NFC_IP_DST_PT;*/
	/*e->nfcache |= NFC_UNKNOWN;*/

	match = new_match(proto, rem_port, int_port);
	target = get_accept_target6();
	tmp = realloc(e, sizeof(struct ip6t_entry)
		       + match->u.match_size
		       + target->u.target_size);
	if(!tmp) {
		bb_error_msg("%s: realloc(%d) failed",
		       "add_pinhole", (int)(sizeof(struct ip6t_entry) + match->u.match_size + target->u.target_size));
		free(e);
		free(match);
		free(target);
		return -1;
	}
	e = tmp;
	memcpy(e->elems, match, match->u.match_size);
	memcpy(e->elems + match->u.match_size, target, target->u.target_size);
	e->target_offset = sizeof(struct ip6t_entry)
			   + match->u.match_size;
	e->next_offset = sizeof(struct ip6t_entry)
			 + match->u.match_size
					 + target->u.target_size;
	free(match);
	free(target);

	if(ip6tc_init_verify_append("filter", bb_common_bss.miniupnpd_v6_filter_chain, e) < 0) {
		free(e);
		return -1;
	}
	uid = add_to_pinhole_list(&e->ipv6.src, rem_port,
				  &e->ipv6.dst, int_port,
				  proto, desc, timestamp);
	free(e);
	return uid;
}

static int
find_pinhole(const char * ifname,
	     const char * rem_host, unsigned short rem_port,
	     const char * int_client, unsigned short int_port,
	     int proto,
	     char *desc, int desc_len, unsigned int * timestamp)
{
	struct pinhole_t * p;
	struct in6_addr saddr;
	struct in6_addr daddr;
	UNUSED(ifname);

	if(rem_host && (rem_host[0] != '\0')) {
		inet_pton(AF_INET6, rem_host, &saddr);
	} else {
		memset(&saddr, 0, sizeof(struct in6_addr));
	}
	inet_pton(AF_INET6, int_client, &daddr);
	for(p = pinhole_list.lh_first; p != NULL; p = p->entries.le_next) {
		if((proto == p->proto) && (rem_port == p->sport) &&
		   (0 == memcmp(&saddr, &p->saddr, sizeof(struct in6_addr))) &&
		   (int_port == p->dport) &&
		   (0 == memcmp(&daddr, &p->daddr, sizeof(struct in6_addr)))) {
			if(desc) safe_strncpy(desc, p->desc, desc_len);
			if(timestamp) *timestamp = p->timestamp;
			return (int)p->uid;
		}
	}
	return -2;      /* not found */
}

static int delete_pinhole(unsigned short uid)
{
	struct pinhole_t * p;
	IP6TC_HANDLE h;
	const struct ip6t_entry * e;
	const struct ip6t_entry_match *match = NULL;
	/*const struct ip6t_entry_target *target = NULL;*/
	unsigned int index;

	p = get_pinhole(uid);
	if(!p)
		return -2;      /* not found */

	h = ip6tc_init("filter");
	if(!h) {
		bb_error_msg("ip6tc_init error : %s", ip6tc_strerror(errno));
		return -1;
	}
	if(!ip6tc_is_chain(bb_common_bss.miniupnpd_v6_filter_chain, h)) {
		bb_error_msg("chain %s not found", bb_common_bss.miniupnpd_v6_filter_chain);
		goto error;
	}
	index = 0;
	for(e = ip6tc_first_rule(bb_common_bss.miniupnpd_v6_filter_chain, h);
	    e;
	    e = ip6tc_next_rule(e, h)) {
		if((e->ipv6.proto == p->proto) &&
		   (0 == memcmp(&e->ipv6.src, &p->saddr, sizeof(e->ipv6.src))) &&
		   (0 == memcmp(&e->ipv6.dst, &p->daddr, sizeof(e->ipv6.dst)))) {
			const struct ip6t_tcp * info;
			match = (const struct ip6t_entry_match *)&e->elems;
			info = (const struct ip6t_tcp *)&match->data;
			if((info->spts[0] == p->sport) && (info->dpts[0] == p->dport)) {
				if(!ip6tc_delete_num_entry(bb_common_bss.miniupnpd_v6_filter_chain, index, h)) {
					bb_error_msg("ip6tc_delete_num_entry(%s,%u,...): %s",
					       bb_common_bss.miniupnpd_v6_filter_chain, index, ip6tc_strerror(errno));
					goto error;
				}
				if(!ip6tc_commit(h)) {
					bb_error_msg("ip6tc_commit(): %s",
					       ip6tc_strerror(errno));
					goto error;
				}
				ip6tc_free(h);
				LIST_REMOVE(p, entries);
				return 0;       /* ok */
			}
		}
		index++;
	}
	ip6tc_free(h);
	xlog_warn("delete_pinhole() rule with PID=%hu not found", uid);
	LIST_REMOVE(p, entries);
	return -2;      /* not found */
error:
	ip6tc_free(h);
	return -1;
}

static int update_pinhole(unsigned short uid, unsigned int timestamp)
{
	struct pinhole_t * p;

	p = get_pinhole(uid);
	if(p) {
		p->timestamp = timestamp;
		return 0;
	} else {
		return -2;      /* Not found */
	}
}

static int
get_pinhole_info(unsigned short uid,
		 char * rem_host, int rem_hostlen,
		 unsigned short * rem_port,
		 char * int_client, int int_clientlen,
		 unsigned short * int_port,
		 int * proto, char * desc, int desclen,
		 unsigned int * timestamp,
		 u_int64_t * packets, u_int64_t * bytes)
{
	struct pinhole_t * p;

	p = get_pinhole(uid);
	if(!p)
		return -2;      /* Not found */
	if(rem_host && (rem_host[0] != '\0')) {
		if(inet_ntop(AF_INET6, &p->saddr, rem_host, rem_hostlen) == NULL)
			return -1;
	}
	if(rem_port)
		*rem_port = p->sport;
	if(int_client) {
		if(inet_ntop(AF_INET6, &p->daddr, int_client, int_clientlen) == NULL)
			return -1;
	}
	if(int_port)
		*int_port = p->dport;
	if(proto)
		*proto = p->proto;
	if(timestamp)
		*timestamp = p->timestamp;
	if (desc)
		safe_strncpy(desc, p->desc, desclen);
	if(packets || bytes) {
		/* theses informations need to be read from netfilter */
		IP6TC_HANDLE h;
		const struct ip6t_entry * e;
		const struct ip6t_entry_match * match;
		h = ip6tc_init("filter");
		if(!h) {
			bb_error_msg("ip6tc_init error : %s", ip6tc_strerror(errno));
			return -1;
		}
		for(e = ip6tc_first_rule(bb_common_bss.miniupnpd_v6_filter_chain, h);
		    e;
		    e = ip6tc_next_rule(e, h)) {
			if((e->ipv6.proto == p->proto) &&
			   (0 == memcmp(&e->ipv6.src, &p->saddr, sizeof(e->ipv6.src))) &&
			   (0 == memcmp(&e->ipv6.dst, &p->daddr, sizeof(e->ipv6.dst)))) {
				const struct ip6t_tcp * info;
				match = (const struct ip6t_entry_match *)&e->elems;
				info = (const struct ip6t_tcp *)&match->data;
				if((info->spts[0] == p->sport) && (info->dpts[0] == p->dport)) {
					if(packets)
						*packets = e->counters.pcnt;
					if(bytes)
						*bytes = e->counters.bcnt;
					break;
				}
			}
		}
		ip6tc_free(h);
	}
	return 0;
}

static int get_pinhole_uid_by_index(int index)
{
	struct pinhole_t * p;

	for(p = pinhole_list.lh_first; p != NULL; p = p->entries.le_next)
		if (!index--)
			return p->uid;
	return -1;
}

static int clean_pinhole_list(unsigned int * next_timestamp)
{
	unsigned int min_ts = UINT_MAX;
	struct pinhole_t * p;
	time_t current_time;
	int n = 0;

	current_time = time(NULL);
	p = pinhole_list.lh_first;
	while(p != NULL) {
		if(p->timestamp <= (unsigned int)current_time) {
			unsigned short uid = p->uid;
			DEBUG_LOG(syslog(LOG_INFO, "removing expired pinhole with uid=%hu", uid);)
			p = p->entries.le_next;
			if(delete_pinhole(uid) == 0)
				n++;
			else
				break;
		} else {
			if(p->timestamp < min_ts)
				min_ts = p->timestamp;
			p = p->entries.le_next;
		}
	}
	if(next_timestamp && (min_ts != UINT_MAX))
		*next_timestamp = min_ts;
	return n;
}

#endif /* ENABLE_UPNPPINHOLE */


/* Encode length by using 7bit per Byte :
 * Most significant bit of each byte specifies that the
 * following byte is part of the code */

/* n : unsigned
 * p : unsigned char *
 */
#define DECODELENGTH(n, p) n = 0; \
			   do { n = (n << 7) | (*p & 0x7f); } \
			   while((*(p++)&0x80) && (n<(1<<25)));

/* n :    unsigned
 * READ : function/macro to read one byte (unsigned char)
 */
#define DECODELENGTH_READ(n, READ) \
	n = 0; \
	do { \
		unsigned char c; \
		READ(c); \
		n = (n << 7) | (c & 0x07f); \
		if(!(c&0x80)) break; \
	} while(n<(1<<25));

/* n :       unsigned
 * p :       unsigned char *
 * p_limit : unsigned char *
 */
#define DECODELENGTH_CHECKLIMIT(n, p, p_limit) \
	n = 0; \
	do { \
		if((p) >= (p_limit)) break; \
		n = (n << 7) | (*(p) & 0x7f); \
	} while((*((p)++)&0x80) && (n<(1<<25)));


/* n : unsigned
 * p : unsigned char *
 */
#define CODELENGTH(n, p) if(n>=268435456) *(p++) = (n >> 28) | 0x80; \
			 if(n>=2097152) *(p++) = (n >> 21) | 0x80; \
			 if(n>=16384) *(p++) = (n >> 14) | 0x80; \
			 if(n>=128) *(p++) = (n >> 7) | 0x80; \
			 *(p++) = n & 0x7f;

/* SSDP ip/port */
#define SSDP_PORT (1900)
#define SSDP_MCAST_ADDR ("239.255.255.250")
#define LL_SSDP_MCAST_ADDR "FF02::C"
#define SL_SSDP_MCAST_ADDR "FF05::C"
#define GL_SSDP_MCAST_ADDR "FF0E::C"

/* AddMulticastMembership()
 * param s                      socket
 * param lan_addr       lan address
 */
static int
AddMulticastMembership(int s, struct lan_addr_s * lan_addr)
{
#ifndef HAVE_IP_MREQN
	/* The ip_mreqn structure appeared in Linux 2.4. */
	struct ip_mreq imr;     /* Ip multicast membership */
#else   /* HAVE_IP_MREQN */
	struct ip_mreqn imr;    /* Ip multicast membership */
#endif  /* HAVE_IP_MREQN */

    /* setting up imr structure */
    imr.imr_multiaddr.s_addr = inet_addr(SSDP_MCAST_ADDR);
    /*imr.imr_interface.s_addr = htonl(INADDR_ANY);*/
#ifndef HAVE_IP_MREQN
	imr.imr_interface.s_addr = lan_addr->addr.s_addr;
#else   /* HAVE_IP_MREQN */
    imr.imr_address.s_addr = lan_addr->addr.s_addr;
#if ENABLE_FEATURE_IPV6
	imr.imr_ifindex = lan_addr->index;
#else   /* ENABLE_FEATURE_IPV6 */
    imr.imr_ifindex = if_nametoindex(lan_addr->ifname);
#endif  /* ENABLE_FEATURE_IPV6 */
#endif  /* HAVE_IP_MREQN */

#ifndef HAVE_IP_MREQN
	if (setsockopt(s, IPPROTO_IP, IP_ADD_MEMBERSHIP, (void *)&imr, sizeof(struct ip_mreq)) < 0)
#else   /* HAVE_IP_MREQN */
	if (setsockopt(s, IPPROTO_IP, IP_ADD_MEMBERSHIP, (void *)&imr, sizeof(struct ip_mreqn)) < 0)
#endif  /* HAVE_IP_MREQN */
	{
		bb_perror_msg("setsockopt(udp, IP_ADD_MEMBERSHIP)");
		return -1;
	}

	return 0;
}

/* AddMulticastMembershipIPv6()
 * param s      socket (IPv6)
 * param ifindex : interface index (0 : All interfaces) */
#if ENABLE_FEATURE_IPV6
static int
AddMulticastMembershipIPv6(int s, unsigned int ifindex)
{
	struct ipv6_mreq mr;

	memset(&mr, 0, sizeof(mr));
	mr.ipv6mr_interface = ifindex;  /* 0 : all interfaces */
#ifndef IPV6_ADD_MEMBERSHIP
#define IPV6_ADD_MEMBERSHIP IPV6_JOIN_GROUP
#endif
	inet_pton(AF_INET6, LL_SSDP_MCAST_ADDR, &mr.ipv6mr_multiaddr);
	if(setsockopt(s, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP, &mr, sizeof(struct ipv6_mreq)) < 0) {
		bb_perror_msg("setsockopt(udp, IPV6_ADD_MEMBERSHIP)");
		return -1;
	}
	inet_pton(AF_INET6, SL_SSDP_MCAST_ADDR, &mr.ipv6mr_multiaddr);
	if(setsockopt(s, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP, &mr, sizeof(struct ipv6_mreq)) < 0) {
		bb_perror_msg("setsockopt(udp, IPV6_ADD_MEMBERSHIP)");
		return -1;
	}
	inet_pton(AF_INET6, GL_SSDP_MCAST_ADDR, &mr.ipv6mr_multiaddr);
	if(setsockopt(s, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP, &mr, sizeof(struct ipv6_mreq)) < 0) {
		bb_perror_msg("setsockopt(udp, IPV6_ADD_MEMBERSHIP)");
		return -1;
	}
	return 0;
}
#endif

/* Open and configure the socket listening for
 * SSDP udp packets sent on 239.255.255.250 port 1900
 * SSDP v6 udp packets sent on FF02::C, or FF05::C, port 1900 */
static int
OpenAndConfSSDPReceiveSocket(int ipv6)
{
	int s;
	struct sockaddr_storage sockname;
	socklen_t sockname_len;
	struct lan_addr_s * lan_addr;

	if( (s = bb_xsocket((ipv6 ? BB_XSOCKET_FLG_AF_INET6 : 0) | BB_XSOCKET_FLG_SOCK_DGRAM)) < 0)
		return -1;

	memset(&sockname, 0, sizeof(struct sockaddr_storage));
#if ENABLE_FEATURE_IPV6
	if(ipv6) {
		struct sockaddr_in6 * saddr = (struct sockaddr_in6 *)&sockname;
		saddr->sin6_family = AF_INET6;
		saddr->sin6_port = htons(SSDP_PORT);
		saddr->sin6_addr = bb_common_bss.ipv6_bind_addr;
		sockname_len = sizeof(struct sockaddr_in6);
	} else
#endif
	{
		struct sockaddr_in * saddr = (struct sockaddr_in *)&sockname;
		saddr->sin_family = AF_INET;
		saddr->sin_port = htons(SSDP_PORT);
		/* NOTE : it seems it doesnt work when binding on the specific address */
		/*saddr->sin_addr.s_addr = inet_addr(UPNP_MCAST_ADDR);*/
		saddr->sin_addr.s_addr = htonl(INADDR_ANY);
		/*saddr->sin_addr.s_addr = inet_addr(ifaddr);*/
		sockname_len = sizeof(struct sockaddr_in);
	}

	if(setsockopt_1i(s, SOL_SOCKET, SO_REUSEADDR, 1) < 0)
		xlog_warn("setsockopt(udp, SO_REUSEADDR): %m");

	if(!set_non_blocking(s))
		xlog_warn("%s: set_non_blocking(): %m", "OpenAndConfSSDPReceiveSocket");

#if defined(SO_BINDTODEVICE)
	/* One and only one LAN interface */
	if(bb_common_bss.lan_addrs.lh_first != NULL && bb_common_bss.lan_addrs.lh_first->list.le_next == NULL
	   && strlen(bb_common_bss.lan_addrs.lh_first->ifname) > 0)
	{
		if(setsockopt(s, SOL_SOCKET, SO_BINDTODEVICE,
			      bb_common_bss.lan_addrs.lh_first->ifname,
			      strlen(bb_common_bss.lan_addrs.lh_first->ifname)) < 0)
		    xlog_warn("%s: setsockopt(udp%s, SO_BINDTODEVICE, %s): %m",
			       "OpenAndConfSSDPReceiveSocket", ipv6 ? "6" : "",
			       bb_common_bss.lan_addrs.lh_first->ifname);
	}
#endif /* defined(SO_BINDTODEVICE) */

	if(bind(s, (struct sockaddr *)&sockname, sockname_len) < 0) {
		bb_perror_msg("%s: bind(udp%s)", "OpenAndConfSSDPReceiveSocket", ipv6 ? "6" : "");
		close(s);
		return -1;
	}

#if ENABLE_FEATURE_IPV6
	if(ipv6) {
		for(lan_addr = bb_common_bss.lan_addrs.lh_first; lan_addr != NULL; lan_addr = lan_addr->list.le_next) {
			if(AddMulticastMembershipIPv6(s, lan_addr->index) < 0) {
				xlog_warn("Failed to add IPv6 multicast membership for interface %s",
				       lan_addr->str ? lan_addr->str : "NULL");
			}
		}
	} else
#endif
	{
		for(lan_addr = bb_common_bss.lan_addrs.lh_first; lan_addr != NULL; lan_addr = lan_addr->list.le_next) {
			if(AddMulticastMembership(s, lan_addr) < 0) {
				xlog_warn("Failed to add multicast membership for interface %s",
				       strlen(lan_addr->str) ? lan_addr->str : "NULL");
			}
		}
	}

	return s;
}

/* open the UDP socket used to send SSDP notifications to
 * the multicast group reserved for them */
static int
OpenAndConfSSDPNotifySocket(in_addr_t addr)
{
	int s;

	struct in_addr mc_if;
	struct sockaddr_in sockname;

	if( (s = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM)) < 0)
		return -1;

	mc_if.s_addr = addr;    /*inet_addr(addr);*/

	if(setsockopt_1i(s, IPPROTO_IP, IP_MULTICAST_LOOP, 0) < 0) {
		bb_perror_msg("setsockopt(udp_notify, IP_MULTICAST_LOOP)");
		close(s);
		return -1;
	}

	if(setsockopt(s, IPPROTO_IP, IP_MULTICAST_IF, (char *)&mc_if, sizeof(mc_if)) < 0) {
		bb_perror_msg("setsockopt(udp_notify, IP_MULTICAST_IF)");
		close(s);
		return -1;
	}

	/* UDA v1.1 says :
		The TTL for the IP packet SHOULD default to 2 and SHOULD be configurable. */
	/* TODO: Make TTL be configurable */
	if(setsockopt_1i(s, IPPROTO_IP, IP_MULTICAST_TTL, 2) < 0)
		xlog_warn("setsockopt(udp_notify, IP_MULTICAST_TTL,): %m");

	if(setsockopt_1i(s, SOL_SOCKET, SO_BROADCAST, 1) < 0) {
		bb_perror_msg("setsockopt(udp_notify, SO_BROADCAST)");
		close(s);
		return -1;
	}

	/* bind() socket before using sendto() is not mandatory
	 * (sendto() will implicitly bind the socket when called on
	 * an unbound socket)
	 * here it is used to se a specific sending address */
	memset(&sockname, 0, sizeof(struct sockaddr_in));
	sockname.sin_family = AF_INET;
	sockname.sin_addr.s_addr = addr;    /*inet_addr(addr);*/

	if (bind(s, (struct sockaddr *)&sockname, sizeof(struct sockaddr_in)) < 0) {
		bb_perror_msg("bind(udp_notify)");
		close(s);
		return -1;
	}

	return s;
}

#if ENABLE_FEATURE_IPV6
/* open the UDP socket used to send SSDP notifications to
 * the multicast group reserved for them. IPv6 */
static int
OpenAndConfSSDPNotifySocketIPv6(unsigned int if_index)
{
	int s;
	struct sockaddr_in6 sockname;

	s = bb_xsocket(BB_XSOCKET_FLG_AF_INET6 | BB_XSOCKET_FLG_SOCK_DGRAM);
	if(s < 0)
		return -1;
	if(setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_IF, &if_index, sizeof(if_index)) < 0) {
		bb_perror_msg("setsockopt(udp_notify IPv6, IPV6_MULTICAST_IF, %u)", if_index);
		close(s);
		return -1;
	}
	if(setsockopt_1i(s, IPPROTO_IPV6, IPV6_MULTICAST_LOOP, 0) < 0) {
		bb_perror_msg("setsockopt(udp_notify, IPV6_MULTICAST_LOOP)");
		close(s);
		return -1;
	}
	/* UDA 2.0 : The hop limit of each IP packet for a Site-Local scope
	 * multicast message SHALL be configurable and SHOULD default to 10 */
	if(setsockopt_1i(s, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, 10) < 0) {
		bb_perror_msg("setsockopt(udp_notify, IPV6_MULTICAST_HOPS)");
		close(s);
		return -1;
	}

	/* bind() socket before using sendto() is not mandatory
	 * (sendto() will implicitly bind the socket when called on
	 * an unbound socket)
	 * but explicit bind permits to set port/scope_id/etc. */
	memset(&sockname, 0, sizeof(sockname));
	sockname.sin6_family = AF_INET6;
	sockname.sin6_addr = in6addr_any;
	/*sockname.sin6_port = htons(port);*/
	/*sockname.sin6_scope_id = if_index;*/
	if(bind(s, (struct sockaddr *)&sockname, sizeof(sockname)) < 0) {
		bb_perror_msg("bind(udp_notify IPv6)");
		close(s);
		return -1;
	}

	return s;
}
#endif

static int
OpenAndConfSSDPNotifySockets(int * sockets)
/*OpenAndConfSSDPNotifySockets(int * sockets,
			     struct lan_addr_s * lan_addr, int n_lan_addr)*/
{
	int i;
	struct lan_addr_s * lan_addr;

	for(i=0, lan_addr = bb_common_bss.lan_addrs.lh_first; lan_addr != NULL;
	    lan_addr = lan_addr->list.le_next)
	{
		sockets[i] = OpenAndConfSSDPNotifySocket(lan_addr->addr.s_addr);
		if(sockets[i] < 0)
			goto error;
		i++;
#if ENABLE_FEATURE_IPV6
		if(GETFLAG(IPV6DISABLEDMASK))
		{
			sockets[i] = -1;
		}
		else
		{
			sockets[i] = OpenAndConfSSDPNotifySocketIPv6(lan_addr->index);
			if(sockets[i] < 0)
				goto error;
		}
		i++;
#endif
	}
	return 0;
error:
	while(--i >= 0)
	{
		close(sockets[i]);
		sockets[i] = -1;
	}
	return -1;
}

/*
 * response from a LiveBox (Wanadoo)
HTTP/1.1 200 OK
CACHE-CONTROL: max-age=1800
DATE: Thu, 01 Jan 1970 04:03:23 GMT
EXT:
LOCATION: http://192.168.0.1:49152/gatedesc.xml
SERVER: Linux/2.4.17, UPnP/1.0, Intel SDK for UPnP devices /1.2
ST: upnp:rootdevice
USN: uuid:75802409-bccb-40e7-8e6c-fa095ecce13e::upnp:rootdevice

 * response from a Linksys 802.11b :
HTTP/1.1 200 OK
Cache-Control:max-age=120
Location:http://192.168.5.1:5678/rootDesc.xml
Server:NT/5.0 UPnP/1.0
ST:upnp:rootdevice
USN:uuid:upnp-InternetGatewayDevice-1_0-0090a2777777::upnp:rootdevice
EXT:
 */

/* Responds to a SSDP "M-SEARCH"
 * s :          socket to use
 * addr :       peer
 * st, st_len : ST: header
 * suffix :     suffix for USN: header
 * host, port : our HTTP host, port
 * delay :      in milli-seconds
 */
static void
SendSSDPResponse(int s, const struct sockaddr * addr,
		 const char * st, int st_len, const char * suffix,
		 const char * host, unsigned short http_port,
		 const char * uuidvalue, unsigned int delay)
{
	int l, n;
	char buf[SSDP_PACKET_MAX_LEN];
	char addr_str[64];
	socklen_t addrlen;
	int st_is_uuid;

	st_is_uuid = (st_len == (int)strlen(uuidvalue)) &&
		      (memcmp(uuidvalue, st, st_len) == 0);
	/*
	 * follow guideline from document "UPnP Device Architecture 1.0"
	 * uppercase is recommended.
	 * DATE: is recommended
	 * SERVER: OS/ver UPnP/1.0 miniupnpd/1.0
	 * - check what to put in the 'Cache-Control' header
	 *
	 * have a look at the document "UPnP Device Architecture v1.1 */
	l = snprintf(buf, sizeof(buf), "HTTP/1.1 200 OK\r\n"
		"CACHE-CONTROL: max-age=120\r\n"
		"ST: %.*s%s\r\n"
		"USN: %s%s%.*s%s\r\n"
		"EXT:\r\n"
		"SERVER: " MINIUPNPD_SERVER_STRING "\r\n"
		"LOCATION: http://%s:%u" ROOTDESC_PATH "\r\n"
		"OPT: \"http://schemas.upnp.org/upnp/1/0/\"; ns=01\r\n" /* UDA v1.1 */
		"01-NLS: %u\r\n" /* same as BOOTID. UDA v1.1 */
		"BOOTID.UPNP.ORG: %u\r\n" /* UDA v1.1 */
		"CONFIGID.UPNP.ORG: %u\r\n" /* UDA v1.1 */
		"\r\n",
		st_len, st, suffix,
		uuidvalue, st_is_uuid ? "" : "::",
		st_is_uuid ? 0 : st_len, st, suffix,
		host, (unsigned int)http_port,
		bb_common_bss.upnp_bootid, bb_common_bss.upnp_bootid, bb_common_bss.upnp_configid);
	if(l<0) {
		bb_perror_msg("%s: snprintf failed", "SendSSDPResponse()");
		return;
	} else if((unsigned)l>=sizeof(buf)) {
		xlog_warn("%s: truncated output (%u>=%u)",
		       "SendSSDPResponse()", (unsigned)l, (unsigned)sizeof(buf));
		l = sizeof(buf) - 1;
	}
	addrlen =
#if ENABLE_FEATURE_IPV6
		(addr->sa_family == AF_INET6)
		  ? sizeof(struct sockaddr_in6) : sizeof(struct sockaddr_in);
#else
		sizeof(struct sockaddr_in);
#endif
	n = sendto_schedule(s, buf, l, 0,
			    addr, addrlen, delay);
	sockaddr_to_string(addr, addr_str, sizeof(addr_str));
	DEBUG_LOG(syslog(LOG_DEBUG, "%s: %d bytes to %s ST: %.*s", "SendSSDPResponse()", n, addr_str, l, buf);)
	if(n < 0)
		bb_perror_msg("%s: sendto(udp)", "SendSSDPResponse()");
}

static struct {
	const char *s;
	const int version;
	const char *uuid;
} const known_service_types[] =
{
	{"upnp:rootdevice", 0, bb_common_bss.uuidvalue_igd},
	/* IGD v1 */
	{"urn:schemas-upnp-org:device:InternetGatewayDevice:", 1, bb_common_bss.uuidvalue_igd},
	{"urn:schemas-upnp-org:device:WANConnectionDevice:", 1, bb_common_bss.uuidvalue_wcd},
	{"urn:schemas-upnp-org:device:WANDevice:", 1, bb_common_bss.uuidvalue_wan},
	{"urn:schemas-upnp-org:service:WANIPConnection:", 1, bb_common_bss.uuidvalue_wcd},
	{"urn:schemas-upnp-org:service:WANCommonInterfaceConfig:", 1, bb_common_bss.uuidvalue_wan},
#ifdef ADVERTISE_WANPPPCONN
	/* We use WAN IP Connection, not PPP connection,
	 * but buggy control points may try to use WanPPPConnection
	 * anyway */
	{"urn:schemas-upnp-org:service:WANPPPConnection:", 1, bb_common_bss.uuidvalue_wcd},
#endif /* ADVERTISE_WANPPPCONN */
#ifdef ENABLE_L3F_SERVICE
	{"urn:schemas-upnp-org:service:Layer3Forwarding:", 1, bb_common_bss.uuidvalue_igd},
#endif /* ENABLE_L3F_SERVICE */
/* we might want to support urn:schemas-wifialliance-org:device:WFADevice:1
 * urn:schemas-wifialliance-org:device:WFADevice:1
 * in the future */
	{0, 0, 0}
};

/* SendSSDPNotify() sends the SSDP NOTIFY to a specific
 * destination, for a specific UPnP service or device */
static void
SendSSDPNotify(int s, const struct sockaddr * dest, socklen_t dest_len,
	       const char * dest_str,
	       const char * host, unsigned short http_port,
	       const char * nt, const char * suffix,
	       const char * usn1, const char * usn2, const char * usn3,
	       unsigned int lifetime)
{
	char bufr[SSDP_PACKET_MAX_LEN];
	int n, l;

	l = snprintf(bufr, sizeof(bufr),
		"NOTIFY * HTTP/1.1\r\n"
		"HOST: %s:%d\r\n"
		"CACHE-CONTROL: max-age=%u\r\n"
		"LOCATION: http://%s:%u" ROOTDESC_PATH "\r\n"
		"SERVER: " MINIUPNPD_SERVER_STRING "\r\n"
		"NT: %s%s\r\n"
		"USN: %s%s%s%s\r\n"
		"NTS: ssdp:alive\r\n"
		"OPT: \"http://schemas.upnp.org/upnp/1/0/\"; ns=01\r\n" /* UDA v1.1 */
		"01-NLS: %u\r\n" /* same as BOOTID field. UDA v1.1 */
		"BOOTID.UPNP.ORG: %u\r\n" /* UDA v1.1 */
		"CONFIGID.UPNP.ORG: %u\r\n" /* UDA v1.1 */
		"\r\n",
		dest_str, SSDP_PORT,            /* HOST: */
		lifetime,                       /* CACHE-CONTROL: */
		host, (unsigned int)http_port,  /* LOCATION: */
		nt, suffix,                     /* NT: */
		usn1, usn2, usn3, suffix,       /* USN: */
		bb_common_bss.upnp_bootid,      /* 01-NLS: */
		bb_common_bss.upnp_bootid,      /* BOOTID.UPNP.ORG: */
		bb_common_bss.upnp_configid);   /* CONFIGID.UPNP.ORG: */
	if(l<0) {
		bb_error_msg("%s: snprintf error", "SendSSDPNotify()");
		return;
	} else if((unsigned int)l >= sizeof(bufr)) {
		xlog_warn("%s: truncated output (%u>=%u)",
		       "SendSSDPNotify()", (unsigned)l, (unsigned)sizeof(bufr));
		l = sizeof(bufr) - 1;
	}
	n = sendto_or_schedule(s, bufr, l, 0, dest, dest_len);
	if(n < 0) {
		bb_perror_msg("sendto(udp_notify=%d, %s)", s, host ? host : "NULL");
	}
#ifdef DEBUG
	  else if(n != l) {
		syslog(LOG_NOTICE, "sendto() sent %d out of %d bytes", n, l);
	}
#endif
	/* Due to the unreliable nature of UDP, devices SHOULD send the entire
	 * set of discovery messages more than once with some delay between
	 * sets e.g. a few hundred milliseconds. To avoid network congestion
	 * discovery messages SHOULD NOT be sent more than three times. */
	n = sendto_schedule(s, bufr, l, 0, dest, dest_len, 250);
	if(n < 0)
		bb_perror_msg("sendto(udp_notify=%d, %s)", s, host ? host : "NULL");
}

/* SendSSDPNotifies() send SSPD NOTIFY for a specific
 * LAN (network interface) for all devices / services */
static void
SendSSDPNotifies(int s, const char * host, unsigned short http_port,
		 unsigned int lifetime, int ipv6)
{
#if ENABLE_FEATURE_IPV6
	struct sockaddr_storage sockname;
	/* UDA 1.1 AnnexA and UDA 2.0 only allow/define the use of
	 * Link-Local and Site-Local multicast scopes */
	static struct { const char * p1, * p2; } const mcast_addrs[] =
		{ { LL_SSDP_MCAST_ADDR, "[" LL_SSDP_MCAST_ADDR "]" },   /* Link Local */
		  { SL_SSDP_MCAST_ADDR, "[" SL_SSDP_MCAST_ADDR "]" },   /* Site Local */
		  { GL_SSDP_MCAST_ADDR, "[" GL_SSDP_MCAST_ADDR "]" },   /* Global */
		  { NULL, NULL } };
	int j;
#else /* ENABLE_FEATURE_IPV6 */
	struct sockaddr_in sockname;
#endif /* ENABLE_FEATURE_IPV6 */
	socklen_t sockname_len;
	const char * dest_str;
	int i;
	char ver_str[4];
#if !ENABLE_FEATURE_IPV6
	UNUSED(ipv6);
#endif /* ENABLE_FEATURE_IPV6 */

	memset(&sockname, 0, sizeof(sockname));
#if ENABLE_FEATURE_IPV6
	/* first iterate destinations for this LAN interface (only 1 for IPv4) */
	for(j = 0; (mcast_addrs[j].p1 != 0 && ipv6) || j < 1; j++) {
		if(ipv6) {
			struct sockaddr_in6 * p = (struct sockaddr_in6 *)&sockname;
			sockname_len = sizeof(struct sockaddr_in6);
			p->sin6_family = AF_INET6;
			p->sin6_port = htons(SSDP_PORT);
			inet_pton(AF_INET6, mcast_addrs[j].p1, &(p->sin6_addr));
			dest_str = mcast_addrs[j].p2;
			/* UPnP Device Architecture 1.1 :
			 * Devices MUST multicast SSDP messages for each of the UPnP-enabled
			 * interfaces. The scope of multicast SSDP messages MUST be
			 * link local FF02::C if the message is sent from a link local address.
			 * If the message is sent from a global address it MUST be multicast
			 * using either global scope FF0E::C or site local scope FF05::C.
			 * In networks with complex topologies and overlapping sites, use of
			 * global scope is RECOMMENDED. */
		} else {
#else /* ENABLE_FEATURE_IPV6 */
		{
#endif /* ENABLE_FEATURE_IPV6 */
			/* IPv4 */
			struct sockaddr_in *p = (struct sockaddr_in *)&sockname;
			sockname_len = sizeof(struct sockaddr_in);
			p->sin_family = AF_INET;
			p->sin_port = htons(SSDP_PORT);
			p->sin_addr.s_addr = inet_addr(SSDP_MCAST_ADDR);
			dest_str = SSDP_MCAST_ADDR;
		}

		/* iterate all services / devices */
		for(i = 0; known_service_types[i].s; i++) {
			if(i==0)
				ver_str[0] = '\0';
			else
				snprintf(ver_str, sizeof(ver_str), "%d", known_service_types[i].version);
				SendSSDPNotify(s, (struct sockaddr *)&sockname, sockname_len, dest_str,
				       host, http_port,
				       known_service_types[i].s, ver_str,       /* NT: */
				       known_service_types[i].uuid, "::",
				       known_service_types[i].s, /* ver_str,    USN: */
				       lifetime);
			/* for devices, also send NOTIFY on the uuid */
			if(0==memcmp(known_service_types[i].s,
				     "urn:schemas-upnp-org:device", sizeof("urn:schemas-upnp-org:device")-1)) {
				SendSSDPNotify(s, (struct sockaddr *)&sockname, sockname_len, dest_str,
					       host, http_port,
					       known_service_types[i].uuid, "", /* NT: */
					       known_service_types[i].uuid, "", "", /* ver_str, USN: */
					       lifetime);
			}
		} /* for(i = 0; known_service_types[i].s; i++) */
#if ENABLE_FEATURE_IPV6
	} /* for(j = 0; (mcast_addrs[j].p1 != 0 && ipv6) || j < 1; j++) */
#endif /* ENABLE_FEATURE_IPV6 */
}

/* SendSSDPNotifies2() sends SSDP NOTIFY packets on all interfaces
 * for all destinations, all devices / services */
static void
SendSSDPNotifies2(int * sockets,
		  unsigned short http_port,
		  unsigned int lifetime)
{
	int i;
	struct lan_addr_s * lan_addr;
	for(i = 0, lan_addr = bb_common_bss.lan_addrs.lh_first; lan_addr != NULL;
	    lan_addr = lan_addr->list.le_next) {
		SendSSDPNotifies(sockets[i], lan_addr->str, http_port,
				 lifetime, 0);
		i++;
#if ENABLE_FEATURE_IPV6
		if(sockets[i] >= 0) {
			SendSSDPNotifies(sockets[i], bb_common_bss.ipv6_addr_for_http_with_brackets, http_port, lifetime, 1);
		}
		i++;
#endif  /* ENABLE_FEATURE_IPV6 */
	}
}

static void
ProcessSSDPData(int s, const char *bufr, int n,
		const struct sockaddr * sender,
		unsigned short http_port)
{
	int i, l;
	struct lan_addr_s * lan_addr = NULL;
	const char * st = NULL;
	int st_len = 0;
	char sender_str[64];
	char ver_str[4];
	const char * announced_host = NULL;
	unsigned int delay = 50; /* Non-zero default delay to prevent flooding */
	/* UPnP Device Architecture v1.1.  1.3.3 Search response :
	 * Devices responding to a multicast M-SEARCH SHOULD wait a random period
	 * of time between 0 seconds and the number of seconds specified in the
	 * MX field value of the search request before responding, in order to
	 * avoid flooding the requesting control point with search responses
	 * from multiple devices. If the search request results in the need for
	 * a multiple part response from the device, those multiple part
	 * responses SHOULD be spread at random intervals through the time period
	 * from 0 to the number of seconds specified in the MX header field. */

	/* get the string representation of the sender address */
	sockaddr_to_string(sender, sender_str, sizeof(sender_str));
	lan_addr = get_lan_for_peer(sender);
	if(lan_addr == NULL)
	{
		xlog_warn("SSDP packet sender %s not from a LAN, ignoring",
		       sender_str);
		return;
	}

	if(memcmp(bufr, "NOTIFY", 6) == 0)
	{
		/* ignore NOTIFY packets. We could log the sender and device type */
		return;
	}
	else if(memcmp(bufr, "M-SEARCH", 8) == 0)
	{
		i = 0;
		while(i < n)
		{
			while((i < n - 1) && (bufr[i] != '\r' || bufr[i+1] != '\n'))
				i++;
			i += 2;
			if((i < n - 3) && (strncasecmp(bufr+i, "st:", 3) == 0))
			{
				st = bufr+i+3;
				st_len = 0;
				while((*st == ' ' || *st == '\t') && (st < bufr + n))
					st++;
				while(st[st_len]!='\r' && st[st_len]!='\n'
				     && (st + st_len < bufr + n))
					st_len++;
				l = st_len;
				while(l > 0 && st[l-1] != ':')
					l--;
				DEBUG_LOG(syslog(LOG_DEBUG, "ST: %.*s (ver=%d)", st_len, st, atoi(st+l));)
				/*j = 0;*/
				/*while(bufr[i+j]!='\r') j++;*/
				/*syslog(LOG_INFO, "%.*s", j, bufr+i);*/
			}
		}
		if(st && (st_len > 0)) {
			DEBUG_LOG(syslog(LOG_INFO, "SSDP M-SEARCH from %s ST: %.*s", sender_str, st_len, st);)
			/* find in which sub network the client is */
			if(sender->sa_family == AF_INET) {
				if (lan_addr == NULL) {
					bb_perror_msg("Can't find in which sub network the client %s is",
					       sender_str);
					return;
				}
				announced_host = lan_addr->str;
			}
#if ENABLE_FEATURE_IPV6
			else
			{
				/* IPv6 address with brackets */
				announced_host = bb_common_bss.ipv6_addr_for_http_with_brackets;
			}
#endif
			/* Responds to request with a device as ST header */
			for(i = 0; known_service_types[i].s; i++)
			{
				l = (int)strlen(known_service_types[i].s);
				if(l<=st_len && (0 == memcmp(st, known_service_types[i].s, l)))
				{
					/* SSDP_RESPOND_SAME_VERSION :
					 * response is urn:schemas-upnp-org:service:WANIPConnection:1 when
					 * M-SEARCH included urn:schemas-upnp-org:service:WANIPConnection:1
					 * else the implemented versions is included in the response
					 *
					 * From UPnP Device Architecture v1.1 :
					 * 1.3.2 [...] Updated versions of device and service types
					 * are REQUIRED to be fully backward compatible with
					 * previous versions. Devices MUST respond to M-SEARCH
					 * requests for any supported version. For example, if a
					 * device implements “urn:schemas-upnporg:service:xyz:2”,
					 * it MUST respond to search requests for both that type
					 * and “urn:schemas-upnp-org:service:xyz:1”. The response
					 * MUST specify the same version as was contained in the
					 * search request. [...] */
#ifndef SSDP_RESPOND_SAME_VERSION
					if(i==0)
						ver_str[0] = '\0';
					else
						snprintf(ver_str, sizeof(ver_str), "%d", known_service_types[i].version);
#endif
					DEBUG_LOG(syslog(LOG_INFO, "Single search found");)
					SendSSDPResponse(s, sender,
#ifdef SSDP_RESPOND_SAME_VERSION
							 st, st_len, "",
#else
							 known_service_types[i].s, l, ver_str,
#endif
							 announced_host, http_port,
							 known_service_types[i].uuid,
							 delay);
					break;
				}
			}
			/* Responds to request with ST: ssdp:all */
			/* strlen("ssdp:all") == 8 */
			if(st_len==8 && (0 == memcmp(st, "ssdp:all", 8))) {
				DEBUG_LOG(syslog(LOG_INFO, "ssdp:all found");)
				for(i=0; known_service_types[i].s; i++) {
					if(i==0)
						ver_str[0] = '\0';
					else
						snprintf(ver_str, sizeof(ver_str), "%d", known_service_types[i].version);
					l = (int)strlen(known_service_types[i].s);
					SendSSDPResponse(s, sender,
							 known_service_types[i].s, l, ver_str,
							 announced_host, http_port,
							 known_service_types[i].uuid,
							 delay);
				}
				/* also answer for uuid */
				SendSSDPResponse(s, sender, bb_common_bss.uuidvalue_igd, strlen(bb_common_bss.uuidvalue_igd), "",
						 announced_host, http_port,
						 bb_common_bss.uuidvalue_igd, delay);
				SendSSDPResponse(s, sender, bb_common_bss.uuidvalue_wan, strlen(bb_common_bss.uuidvalue_wan), "",
						 announced_host, http_port,
						 bb_common_bss.uuidvalue_wan, delay);
				SendSSDPResponse(s, sender, bb_common_bss.uuidvalue_wcd, strlen(bb_common_bss.uuidvalue_wcd), "",
						 announced_host, http_port,
						 bb_common_bss.uuidvalue_wcd, delay);
			}
			/* responds to request by UUID value */
			l = (int)strlen(bb_common_bss.uuidvalue_igd);
			if(l==st_len) {
				if(0 == memcmp(st, bb_common_bss.uuidvalue_igd, l)) {
					DEBUG_LOG(syslog(LOG_INFO, "ssdp:uuid (IGD) found");)
					SendSSDPResponse(s, sender, st, st_len, "",
							 announced_host, http_port,
							 bb_common_bss.uuidvalue_igd, delay);
				} else if(0 == memcmp(st, bb_common_bss.uuidvalue_wan, l)) {
					DEBUG_LOG(syslog(LOG_INFO, "ssdp:uuid (WAN) found");)
					SendSSDPResponse(s, sender, st, st_len, "",
							 announced_host, http_port,
							 bb_common_bss.uuidvalue_wan, delay);
				} else if(0 == memcmp(st, bb_common_bss.uuidvalue_wcd, l)) {
					DEBUG_LOG(syslog(LOG_INFO, "ssdp:uuid (WCD) found");)
					SendSSDPResponse(s, sender, st, st_len, "",
							 announced_host, http_port,
							 bb_common_bss.uuidvalue_wcd, delay);
				}
			}
		} else {
			DEBUG_LOG(syslog(LOG_INFO, "Invalid SSDP M-SEARCH from %s", sender_str));
		}
	}
#ifdef DEBUG
	  else {
		syslog(LOG_NOTICE, "Unknown udp packet received from %s", sender_str);
	}
#endif
}

/* ProcessSSDPRequest()
 * process SSDP M-SEARCH requests and responds to them */
static void
ProcessSSDPRequest(int s, unsigned short http_port)
{
	int n;
	char bufr[1500];
	socklen_t len_r;
#if ENABLE_FEATURE_IPV6
	struct sockaddr_storage sendername;
	len_r = sizeof(struct sockaddr_storage);
#else
	struct sockaddr_in sendername;
	len_r = sizeof(struct sockaddr_in);
#endif

	n = recvfrom(s, bufr, sizeof(bufr), 0,
		     (struct sockaddr *)&sendername, &len_r);
	if(n < 0) {
		/* EAGAIN, EWOULDBLOCK, EINTR : silently ignore (try again next time)
		 * other errors : log to LOG_ERR */
		if(errno != EAGAIN && errno != EWOULDBLOCK && errno != EINTR)
			bb_perror_msg("recvfrom(udp)");
		return;
	}
	ProcessSSDPData(s, bufr, n, (struct sockaddr *)&sendername, http_port);
}


static int
SendSSDPbyebye(int s, const struct sockaddr * dest, socklen_t destlen,
	       const char * dest_str,
	       const char * nt, const char * suffix,
	       const char * usn1, const char * usn2, const char * usn3)
{
	int n, l;
	char bufr[SSDP_PACKET_MAX_LEN];

	l = snprintf(bufr, sizeof(bufr),
		     "NOTIFY * HTTP/1.1\r\n"
		     "HOST: %s:%d\r\n"
		     "NT: %s%s\r\n"
		     "USN: %s%s%s%s\r\n"
		     "NTS: ssdp:byebye\r\n"
		     "OPT: \"http://schemas.upnp.org/upnp/1/0/\"; ns=01\r\n" /* UDA v1.1 */
		     "01-NLS: %u\r\n" /* same as BOOTID field. UDA v1.1 */
		     "BOOTID.UPNP.ORG: %u\r\n" /* UDA v1.1 */
		     "CONFIGID.UPNP.ORG: %u\r\n" /* UDA v1.1 */
		     "\r\n",
		     dest_str, SSDP_PORT,       /* HOST : */
		     nt, suffix,                /* NT: */
		     usn1, usn2, usn3, suffix,  /* USN: */
		     bb_common_bss.upnp_bootid, bb_common_bss.upnp_bootid, bb_common_bss.upnp_configid);
	if(l<0) {
		bb_error_msg("%s: snprintf error", "SendSSDPbyebye()");
		return -1;
	} else if((unsigned int)l >= sizeof(bufr)) {
		xlog_warn("%s: truncated output (%u>=%u)",
		       "SendSSDPbyebye()", (unsigned)l, (unsigned)sizeof(bufr));
		l = sizeof(bufr) - 1;
	}
	n = sendto_or_schedule(s, bufr, l, 0, dest, destlen);
	if(n < 0) {
		bb_perror_msg("sendto(udp_shutdown=%d)", s);
		return -1;
	}
#ifdef DEBUG
	  else if(n != l) {
		syslog(LOG_NOTICE, "sendto() sent %d out of %d bytes", n, l);
		return -1;
	}
#endif
	return 0;
}

/* This will broadcast ssdp:byebye notifications to inform
 * the network that UPnP is going down. */
static int
SendSSDPGoodbye(int * sockets, int n_sockets)
{
	struct sockaddr_in sockname4;
#if ENABLE_FEATURE_IPV6
	struct sockaddr_in6 sockname6;
	struct sockaddr * sockname;
	socklen_t socknamelen;
	int ipv6 = 0;
#endif
	int i, j;
	char ver_str[4];
	int ret = 0;
	const char * dest_str;

    memset(&sockname4, 0, sizeof(struct sockaddr_in));
    sockname4.sin_family = AF_INET;
    sockname4.sin_port = htons(SSDP_PORT);
    sockname4.sin_addr.s_addr = inet_addr(SSDP_MCAST_ADDR);
#if ENABLE_FEATURE_IPV6
	memset(&sockname6, 0, sizeof(struct sockaddr_in6));
	sockname6.sin6_family = AF_INET6;
	sockname6.sin6_port = htons(SSDP_PORT);
	inet_pton(AF_INET6, LL_SSDP_MCAST_ADDR, &(sockname6.sin6_addr));
#else
	dest_str = SSDP_MCAST_ADDR;
#endif

	for(j=0; j<n_sockets; j++)
	{
		if(sockets[j] < 0)
			continue;
#if ENABLE_FEATURE_IPV6
		ipv6 = j & 1;
		if(ipv6) {
			dest_str = "[" LL_SSDP_MCAST_ADDR "]";
			sockname = (struct sockaddr *)&sockname6;
			socknamelen = sizeof(struct sockaddr_in6);
		} else {
			dest_str = SSDP_MCAST_ADDR;
			sockname = (struct sockaddr *)&sockname4;
			socknamelen = sizeof(struct sockaddr_in);
		}
#endif
	    for(i=0; known_service_types[i].s; i++)
	    {
			if(i==0)
				ver_str[0] = '\0';
			else
				snprintf(ver_str, sizeof(ver_str), "%d", known_service_types[i].version);
			ret += SendSSDPbyebye(sockets[j],
#if ENABLE_FEATURE_IPV6
					      sockname, socknamelen,
#else
					      (struct sockaddr *)&sockname4, sizeof(struct sockaddr_in),
#endif
					      dest_str,
					      known_service_types[i].s, ver_str,        /* NT: */
					      known_service_types[i].uuid, "::",
					      known_service_types[i].s); /* ver_str, USN: */
			if(0==memcmp(known_service_types[i].s,
				     "urn:schemas-upnp-org:device", sizeof("urn:schemas-upnp-org:device")-1))
			{
				ret += SendSSDPbyebye(sockets[j],
#if ENABLE_FEATURE_IPV6
						      sockname, socknamelen,
#else
						      (struct sockaddr *)&sockname4, sizeof(struct sockaddr_in),
#endif
						      dest_str,
						      known_service_types[i].uuid, "",  /* NT: */
						      known_service_types[i].uuid, "", ""); /* ver_str, USN: */
			}
	}
	}
	return ret;
}

/* SubmitServicesToMiniSSDPD() :
 * register services offered by MiniUPnPd to a running instance of
 * MiniSSDPd */
static int
SubmitServicesToMiniSSDPD(const char * host, unsigned short port)
{
	struct sockaddr_un addr;
	int s;
	unsigned char buffer[2048];
	char strbuf[256];
	unsigned char * p;
	int i, l, n;
	char ver_str[4];

	s = bb_xsocket(BB_XSOCKET_FLG_AF_UNIX);
	if(s < 0)
		return -1;
	addr.sun_family = AF_UNIX;
	safe_strncpy(addr.sun_path, bb_common_bss.minissdpdsocketpath, sizeof(addr.sun_path));
	if(connect(s, (struct sockaddr *)&addr, sizeof(struct sockaddr_un)) < 0) {
		bb_perror_msg("connect(\"%s\")", bb_common_bss.minissdpdsocketpath);
		close(s);
		return -1;
	}
	for(i = 0; known_service_types[i].s; i++) {
		buffer[0] = 4;  /* request type 4 : submit service */
		/* 4 strings following : ST (service type), USN, Server, Location */
		p = buffer + 1;
		l = (int)strlen(known_service_types[i].s);
		if(i > 0)
			l++;
		CODELENGTH(l, p);
		memcpy(p, known_service_types[i].s, l);
		if(i > 0)
			p[l-1] = '1';
		p += l;
		if(i==0)
			ver_str[0] = '\0';
		else
			snprintf(ver_str, sizeof(ver_str), "%d", known_service_types[i].version);
		l = snprintf(strbuf, sizeof(strbuf), "%s::%s%s",
			     known_service_types[i].uuid, known_service_types[i].s, ver_str);
		if(l<0) {
			xlog_warn("SubmitServicesToMiniSSDPD: snprintf %m");
			continue;
		} else if((unsigned)l>=sizeof(strbuf)) {
			l = sizeof(strbuf) - 1;
		}
		CODELENGTH(l, p);
		memcpy(p, strbuf, l);
		p += l;
		l = (int)strlen(MINIUPNPD_SERVER_STRING);
		CODELENGTH(l, p);
		memcpy(p, MINIUPNPD_SERVER_STRING, l);
		p += l;
		l = snprintf(strbuf, sizeof(strbuf), "http://%s:%u" ROOTDESC_PATH,
			     host, (unsigned int)port);
		if(l<0) {
			xlog_warn("SubmitServicesToMiniSSDPD: snprintf %m");
			continue;
		} else if((unsigned)l>=sizeof(strbuf)) {
			l = sizeof(strbuf) - 1;
		}
		CODELENGTH(l, p);
		memcpy(p, strbuf, l);
		p += l;
		/* now write the encoded data */
		n = p - buffer; /* bytes to send */
		p = buffer;     /* start */
		while(n > 0) {
			l = write(s, p, n);
			if (l < 0) {
				if(errno == EINTR)
					continue;
				bb_perror_msg("write()");
				close(s);
				return -1;
			} else if (l == 0) {
				bb_error_msg("write() returned 0");
				close(s);
				return -1;
			}
			p += l;
			n -= l;
		}
	}
	close(s);
	return 0;
}

#include <sys/file.h>

#if ENABLE_FEATURE_MINIUPNPD_NATPMP

# ifndef ENABLE_PCP
# define PCP_MAX_LEN 32
# endif

#endif

#ifdef USE_IFACEWATCHER
/* $Id: ifacewatcher.c,v 1.8 2015/03/07 15:52:33 nanard Exp $ */
/* ifacewatcher.c
 *
 * This file implements dynamic serving of new network interfaces
 * which weren't available during daemon start. It also takes care
 * of interfaces which become unavailable.
 *
 * Copyright (c) 2011, Alexey Osipov <simba@lerlan.ru>
*/


static int OpenAndConfInterfaceWatchSocket(void)
{
	int s;
	struct sockaddr_nl addr;

	s = socket(PF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
	if (s == -1) {
		bb_perror_msg("socket(PF_NETLINK, SOCK_RAW, NETLINK_ROUTE)");
		return -1;
	}

	memset(&addr, 0, sizeof(addr));
	addr.nl_family = AF_NETLINK;
	addr.nl_groups = RTMGRP_LINK | RTMGRP_IPV4_IFADDR;
	/*addr.nl_groups = RTMGRP_LINK | RTMGRP_IPV4_IFADDR | RTMGRP_IPV6_IFADDR;*/

	if (bind(s, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		bb_perror_msg("bind(netlink)");
		close(s);
		return -1;
	}

	return s;
}

static void ProcessInterfaceWatchNotify(int s)
{
	char buffer[4096];
	struct iovec iov;
	struct msghdr hdr;
	struct nlmsghdr *nlhdr;
	struct ifaddrmsg *ifa;
	int len;

	struct rtattr *rth;
	int rtl;

	unsigned int ext_if_name_index = 0;

	iov.iov_base = buffer;
	iov.iov_len = sizeof(buffer);

	memset(&hdr, 0, sizeof(hdr));
	hdr.msg_iov = &iov;
	hdr.msg_iovlen = 1;

	len = recvmsg(s, &hdr, 0);
	if (len < 0) {
		bb_perror_msg("recvmsg(s, &hdr, 0)");
		return;
	}

	if(bb_common_bss.ext_if_name)
		ext_if_name_index = if_nametoindex(bb_common_bss.ext_if_name);

	for (nlhdr = (struct nlmsghdr *) buffer; NLMSG_OK (nlhdr, (unsigned int)len); nlhdr = NLMSG_NEXT (nlhdr, len)) {
#ifdef DEBUG
		int is_del = 0;
#endif
		char address[48];
		char ifname[IFNAMSIZ];
		address[0] = '\0';
		ifname[0] = '\0';
		if (nlhdr->nlmsg_type == NLMSG_DONE)
			break;
		switch(nlhdr->nlmsg_type) {
		case RTM_DELLINK:
#ifdef DEBUG
			is_del = 1;
#endif
		case RTM_NEWLINK:
			break;
		case RTM_DELADDR:
#ifdef DEBUG
			is_del = 1;
#endif
		case RTM_NEWADDR:
			/* see /usr/include/linux/netlink.h
			 * and /usr/include/linux/rtnetlink.h */
			ifa = (struct ifaddrmsg *) NLMSG_DATA(nlhdr);
			DEBUG_LOG(syslog(LOG_DEBUG, "%s %s index=%d fam=%d", "ProcessInterfaceWatchNotify",
			       is_del ? "RTM_DELADDR" : "RTM_NEWADDR",
			       ifa->ifa_index, ifa->ifa_family);)
			for(rth = IFA_RTA(ifa), rtl = IFA_PAYLOAD(nlhdr);
			    rtl && RTA_OK(rth, rtl);
			    rth = RTA_NEXT(rth, rtl)) {
				char tmp[128];
				memset(tmp, 0, sizeof(tmp));
				switch(rth->rta_type) {
				case IFA_ADDRESS:
				case IFA_LOCAL:
				case IFA_BROADCAST:
				case IFA_ANYCAST:
					inet_ntop(ifa->ifa_family, RTA_DATA(rth), tmp, sizeof(tmp));
					if(rth->rta_type == IFA_ADDRESS)
						safe_strncpy(address, tmp, sizeof(address));
					break;
				case IFA_LABEL:
					safe_strncpy(tmp, RTA_DATA(rth), sizeof(tmp));
					safe_strncpy(ifname, tmp, sizeof(ifname));
					break;
				case IFA_CACHEINFO:
					{
						struct ifa_cacheinfo *cache_info;
						cache_info = RTA_DATA(rth);
						snprintf(tmp, sizeof(tmp), "valid=%u prefered=%u",
							 cache_info->ifa_valid, cache_info->ifa_prefered);
					}
					break;
				default:
					safe_strncpy(tmp, "*unknown*", sizeof(tmp));
				}
				DEBUG_LOG(syslog(LOG_DEBUG, " - %u - %s type=%d",
							ifa->ifa_index, tmp, rth->rta_type);)
			}
			if(ifa->ifa_index == ext_if_name_index) {
				bb_common_bss.should_send_public_address_change_notif = 1;
			}
			break;
		default:
			DEBUG_LOG(syslog(LOG_DEBUG, "%s type %d ignored", "ProcessInterfaceWatchNotify", nlhdr->nlmsg_type));
		}
	}
}
#endif

#ifndef DEFAULT_CONFIG
#define DEFAULT_CONFIG "/etc/miniupnpd.conf"
#endif

/* OpenAndConfHTTPSocket() :
 * setup the socket used to handle incoming HTTP connections. */
static int
#if ENABLE_FEATURE_IPV6
OpenAndConfHTTPSocket(unsigned short * port, int ipv6)
#else
OpenAndConfHTTPSocket(unsigned short * port)
#endif
{
	int s;
#if ENABLE_FEATURE_IPV6
	struct sockaddr_in6 listenname6;
	struct sockaddr_in listenname4;
#else
	struct sockaddr_in listenname;
#endif
	socklen_t listenname_len;

	s = bb_xsocket(BB_XSOCKET_FLG_NO_WARN
#if ENABLE_FEATURE_IPV6
		   | (ipv6 ? BB_XSOCKET_FLG_AF_INET6 : 0)
#endif
		   );
#if ENABLE_FEATURE_IPV6
	if(s < 0 && ipv6 && errno == EAFNOSUPPORT)
	{
		/* the system doesn't support IPV6 */
		xlog_warn("socket(PF_INET6, ...) failed with EAFNOSUPPORT, disabling IPv6");
		SETFLAG(IPV6DISABLEDMASK);
		ipv6 = 0;
		s = bb_xsocket(BB_XSOCKET_FLG_NO_WARN);
	}
#endif
	if(s < 0) {
		bb_perror_msg("socket(http)");
		return -1;
	}

	if(setsockopt_1i(s, SOL_SOCKET, SO_REUSEADDR, 1) < 0)
		xlog_warn("setsockopt(http, SO_REUSEADDR): %m");

	if(!set_non_blocking(s))
		xlog_warn("set_non_blocking(http): %m");

#if ENABLE_FEATURE_IPV6
	if(ipv6) {
		memset(&listenname6, 0, sizeof(struct sockaddr_in6));
		listenname6.sin6_family = AF_INET6;
		listenname6.sin6_port = htons(*port);
		listenname6.sin6_addr = bb_common_bss.ipv6_bind_addr;
		listenname_len =  sizeof(struct sockaddr_in6);
	} else {
		memset(&listenname4, 0, sizeof(struct sockaddr_in));
		listenname4.sin_family = AF_INET;
		listenname4.sin_port = htons(*port);
		listenname4.sin_addr.s_addr = htonl(INADDR_ANY);
		listenname_len =  sizeof(struct sockaddr_in);
	}
#else
	memset(&listenname, 0, sizeof(struct sockaddr_in));
	listenname.sin_family = AF_INET;
	listenname.sin_port = htons(*port);
	listenname.sin_addr.s_addr = htonl(INADDR_ANY);
	listenname_len =  sizeof(struct sockaddr_in);
#endif

#if defined(SO_BINDTODEVICE)
	/* One and only one LAN interface */
	if(bb_common_bss.lan_addrs.lh_first != NULL && bb_common_bss.lan_addrs.lh_first->list.le_next == NULL
	   && strlen(bb_common_bss.lan_addrs.lh_first->ifname) > 0)
	{
		if(setsockopt(s, SOL_SOCKET, SO_BINDTODEVICE,
			      bb_common_bss.lan_addrs.lh_first->ifname,
			      strlen(bb_common_bss.lan_addrs.lh_first->ifname)) < 0)
			xlog_warn("setsockopt(http, SO_BINDTODEVICE, %s): %m",
			       bb_common_bss.lan_addrs.lh_first->ifname);
	}
#endif /* defined(SO_BINDTODEVICE) */

#if ENABLE_FEATURE_IPV6
	if(bind(s, ipv6 ? (struct sockaddr *)&listenname6 : (struct sockaddr *)&listenname4,
		listenname_len) < 0)
#else
	if(bind(s, (struct sockaddr *)&listenname, listenname_len) < 0)
#endif
	{
		bb_perror_msg("bind(http)");
		close(s);
		return -1;
	}

	if(listen(s, 5) < 0) {
		bb_perror_msg("listen(http)");
		close(s);
		return -1;
	}

	if(*port == 0) {
#if ENABLE_FEATURE_IPV6
		if(ipv6) {
			struct sockaddr_in6 sockinfo;
			socklen_t len = sizeof(struct sockaddr_in6);
			if (getsockname(s, (struct sockaddr *)&sockinfo, &len) < 0) {
				bb_perror_msg("getsockname()");
			} else {
				*port = ntohs(sockinfo.sin6_port);
			}
		} else {
#endif /* ENABLE_FEATURE_IPV6 */
			struct sockaddr_in sockinfo;
			socklen_t len = sizeof(struct sockaddr_in);
			if (getsockname(s, (struct sockaddr *)&sockinfo, &len) < 0) {
				bb_perror_msg("getsockname()");
			} else {
				*port = ntohs(sockinfo.sin_port);
			}
#if ENABLE_FEATURE_IPV6
		}
#endif /* ENABLE_FEATURE_IPV6 */
	}
	return s;
}

static struct upnphttp *
ProcessIncomingHTTP(int shttpl, const char * protocol)
{
	int shttp;
	socklen_t clientnamelen;
#if ENABLE_FEATURE_IPV6
	struct sockaddr_storage clientname;
	clientnamelen = sizeof(struct sockaddr_storage);
#else
	struct sockaddr_in clientname;
	clientnamelen = sizeof(struct sockaddr_in);
#endif
	shttp = accept(shttpl, (struct sockaddr *)&clientname, &clientnamelen);
	if(shttp<0) {
		/* ignore EAGAIN, EWOULDBLOCK, EINTR, we just try again later */
		if(errno != EAGAIN && errno != EWOULDBLOCK && errno != EINTR)
			bb_perror_msg("accept(http)");
	} else {
		struct upnphttp *tmp = 0;
		char addr_str[64];

		sockaddr_to_string((struct sockaddr *)&clientname, addr_str, sizeof(addr_str));
		if(get_lan_for_peer((struct sockaddr *)&clientname) == NULL) {
			/* The peer is not a LAN ! */
			xlog_warn("%s peer %s is not from a LAN, closing the connection",
			       protocol, addr_str);
			close(shttp);
		} else {
			/* Create a new upnphttp object and add it to
			 * the active upnphttp object list */
			tmp = New_upnphttp(shttp);
			if(tmp) {
#if ENABLE_FEATURE_IPV6
				if(clientname.ss_family == AF_INET) {
					tmp->clientaddr = ((struct sockaddr_in *)&clientname)->sin_addr;
				} else if(clientname.ss_family == AF_INET6) {
					struct sockaddr_in6 * addr = (struct sockaddr_in6 *)&clientname;
					if(IN6_IS_ADDR_V4MAPPED(&addr->sin6_addr)) {
						memcpy(&tmp->clientaddr,
						       &addr->sin6_addr.s6_addr[12],
						       4);
					} else {
						tmp->ipv6 = 1;
						memcpy(&tmp->clientaddr_v6,
						       &addr->sin6_addr,
						       sizeof(struct in6_addr));
					}
				}
#else
				tmp->clientaddr = clientname.sin_addr;
#endif
#ifdef DEBUG
				memcpy(tmp->clientaddr_str, addr_str, sizeof(tmp->clientaddr_str));
#endif
				return tmp;
			} else {
				bb_error_msg("New_upnphttp() failed");
				close(shttp);
			}
		}
	}
	return NULL;
}

/* Handler for the SIGTERM signal (kill)
 * SIGINT is also handled */
static void
sigterm(int sig)
{
	UNUSED(sig);
	/*int save_errno = errno; */
	/*signal(sig, SIG_IGN);*/       /* Ignore this signal while we are quitting */
	/* Note : isn't it useless ? */

	bb_common_bss.quitting = 1;
	/*errno = save_errno;*/
}

/* Handler for the SIGUSR1 signal indicating public IP address change. */
static void
sigusr1(int sig)
{
	UNUSED(sig);
	bb_common_bss.should_send_public_address_change_notif = 1;
}

/* record the startup time, for returning uptime */
static void
set_startup_time(int sysuptime)
{
	bb_common_bss.startup_time = time(NULL);
	if(bb_common_bss.startup_time > 60*60*24 && bb_common_bss.upnp_bootid == 1) {
		/* We know we are not January the 1st 1970 */
		bb_common_bss.upnp_bootid = (unsigned int)bb_common_bss.startup_time;
		/* from UDA v1.1 :
		 * A convenient mechanism is to set this field value to the time
		 * that the device sends its initial announcement, expressed as
		 * seconds elapsed since midnight January 1, 1970; */
	}
	if(sysuptime) {
		/* use system uptime instead of daemon uptime */
		char buff[64];
		int uptime = 0, fd;
		fd = bb_xopen("/proc/uptime", O_RDONLY|BB_OPEN_WARN);
		if(fd >= 0) {
			memset(buff, 0, sizeof(buff));
			if(read(fd, buff, sizeof(buff) - 1) < 0) {
				bb_perror_msg("read(\"/proc/uptime\"");
			} else {
				uptime = atoi(buff);
				DEBUG_LOG(syslog(LOG_INFO, "system uptime is %d seconds", uptime);)
			}
			close(fd);
			bb_common_bss.startup_time -= uptime;
		}
	}
}

/* structure containing variables used during "main loop"
 * that are filled during the init */
struct runtime_vars {
	/* LAN IP addresses for SSDP traffic and HTTP */
	/* moved to global vars */
	int port;       /* HTTP Port */
	int notify_interval;    /* seconds between SSDP announces */
	/* unused rules cleaning related variables : */
	int clean_ruleset_threshold;    /* threshold for removing unused rules */
	int clean_ruleset_interval;             /* (minimum) interval between checks */
};

/* parselanaddr()
 * parse address with mask
 * ex: 192.168.1.1/24 or 192.168.1.1/255.255.255.0
 *
 * Can also use the interface name (ie eth0)
 *
 * return value :
 *    0 : ok
 *   -1 : error */
static int
parselanaddr(struct lan_addr_s *lan_addr, const char *str)
{
	const char * p;
	int n;
	char tmp[16];

	memset(lan_addr, 0, sizeof(struct lan_addr_s));
	p = str;
	while(*p && *p != '/')
		p++;
	n = p - str;
	if(!isdigit(str[0]) && n < (int)sizeof(lan_addr->ifname)) {
		/* not starting with a digit : suppose it is an interface name */
		safe_strncpy(lan_addr->ifname, str, n + 1);
		if(getifaddr(lan_addr->ifname, lan_addr->str, sizeof(lan_addr->str),
			     &lan_addr->addr, &lan_addr->mask) < 0) {
#if ENABLE_FEATURE_IPV6
			bb_error_msg("interface \"%s\" has no IPv4 address", str);
			lan_addr->str[0] = '\0';
			lan_addr->addr.s_addr = htonl(0x00000000u);
			lan_addr->mask.s_addr = htonl(0xffffffffu);
#else
			goto parselan_error;
#endif /* ENABLE_FEATURE_IPV6 */
		}
		/*printf("%s => %s\n", lan_addr->ifname, lan_addr->str);*/
	} else {
		if(n>15)
			goto parselan_error;
		safe_strncpy(lan_addr->ifname, str, n + 1);
		if(!inet_aton(lan_addr->str, &lan_addr->addr))
			goto parselan_error;
	}
	if(*p == '/') {
		const char *q = ++p;

		while(*p && isdigit(*p))
			p++;
		if(*p=='.') {
			/* parse mask in /255.255.255.0 format */
			while(*p && (*p=='.' || isdigit(*p)))
				p++;
			n = p - q;
			if(n>15)
				goto parselan_error;
			safe_strncpy(tmp, q, n + 1);
			if(!inet_aton(tmp, &lan_addr->mask))
				goto parselan_error;
		} else {
			/* it is a /24 format */
			int nbits = atoi(q);
			if(nbits > 32 || nbits < 0)
				goto parselan_error;
			lan_addr->mask.s_addr = htonl(nbits ? (0xffffffffu << (32 - nbits)) : 0);
		}
	} else if(lan_addr->mask.s_addr == 0) {
		/* by default, networks are /24 */
		lan_addr->mask.s_addr = htonl(0xffffff00u);
	}
#if ENABLE_FEATURE_IPV6
	if(lan_addr->ifname[0] != '\0') {
		lan_addr->index = if_nametoindex(lan_addr->ifname);
		if(lan_addr->index == 0)
			bb_error_msg("cannot get index for network interface %s", lan_addr->ifname);
	} else {
		bb_error_msg("Error: please specify LAN network interface by name instead of IPv4 address : %s", str);
		return -1;
	}
#endif
	return 0;
parselan_error:
	bb_error_msg("error parsing address/mask (or interface name) : %s", str);
	return -1;
}

/* fill uuidvalue_wan and uuidvalue_wcd based on uuidvalue_igd */
static void complete_uuidvalues(void)
{
	size_t len;
	len = strlen(bb_common_bss.uuidvalue_igd);
	memcpy(bb_common_bss.uuidvalue_wan, bb_common_bss.uuidvalue_igd, len+1);
	switch(bb_common_bss.uuidvalue_wan[len-1]) {
	case '9':
		bb_common_bss.uuidvalue_wan[len-1] = 'a';
		break;
	case 'f':
		bb_common_bss.uuidvalue_wan[len-1] = '0';
		break;
	default:
		bb_common_bss.uuidvalue_wan[len-1]++;
	}
	memcpy(bb_common_bss.uuidvalue_wcd, bb_common_bss.uuidvalue_wan, len+1);
	switch(bb_common_bss.uuidvalue_wcd[len-1]) {
	case '9':
		bb_common_bss.uuidvalue_wcd[len-1] = 'a';
		break;
	case 'f':
		bb_common_bss.uuidvalue_wcd[len-1] = '0';
		break;
	default:
		bb_common_bss.uuidvalue_wcd[len-1]++;
	}
}

/* check for another instance running
 * returns: 0 only instance
 *          -1 invalid filename
 *          -2 another instance running  */
static int checkforrunning(const char * fname)
{
	char buffer[64];
	int pidfile;
	pid_t pid;

	if(!fname || (strlen(fname) == 0))
		return -1;

	if( (pidfile = open(fname, O_RDONLY)) < 0)
		return 0;

	memset(buffer, 0, 64);

	if(read(pidfile, buffer, 63)) {
		if( (pid = atol(buffer)) > 0) {
			if(!kill(pid, 0)) {
				close(pidfile);
				return -2;
			}
		}
	}

	close(pidfile);

	return 0;
}

#define PCP_MIN_LEN           24


/* Current assumptions:
   - IPv4 is always NATted (internal -> external)
   - IPv6 is always firewalled (this may need some work, NAT6* do exist)

   - we make the judgement based on (in order, picking first one available):
     - third party adress
     - internal client address

   TODO : handle NAT46, NAT64, NPT66. In addition, beyond FW/NAT
   choice, should also add for proxy (=as much as possible transparent
   pass-through to one or more servers).

   TODO: IPv6 permission handling (for the time being, we just assume
   anyone on IPv6 is a good guy, but fixing that would include
   upnppermissions rewrite to be AF neutral).
*/


#ifdef ENABLE_PCP

#define PCP_MAX_LEN           1100
#define PCP_OPCODE_ANNOUNCE   0
#define PCP_OPCODE_MAP        1
#define PCP_OPCODE_PEER       2

/* Possible response codes sent by server, as a result of client request*/
#define PCP_SUCCESS                     0

#define PCP_ERR_UNSUPP_VERSION          1
/** The version number at the start of the PCP Request
 * header is not recognized by this PCP server.  This is a long
 * lifetime error.  This document describes PCP version 2.
 */

#define PCP_ERR_NOT_AUTHORIZED          2
/**The requested operation is disabled for this PCP
 * client, or the PCP client requested an operation that cannot be
 * fulfilled by the PCP server's security policy.  This is a long
 * lifetime error.
 */

#define PCP_ERR_MALFORMED_REQUEST       3
/**The request could not be successfully parsed.
 * This is a long lifetime error.
 */

#define PCP_ERR_UNSUPP_OPCODE           4
/** Unsupported Opcode.  This is a long lifetime error.
 */

#define PCP_ERR_UNSUPP_OPTION           5
/**Unsupported Option.  This error only occurs if the
 * Option is in the mandatory-to-process range.  This is a long
 * lifetime error.
 */

#define PCP_ERR_MALFORMED_OPTION        6
/**Malformed Option (e.g., appears too many times,
 * invalid length).  This is a long lifetime error.
 */

#define PCP_ERR_NETWORK_FAILURE         7
/**The PCP server or the device it controls are
 * experiencing a network failure of some sort (e.g., has not
 * obtained an External IP address).  This is a short lifetime error.
 */

#define PCP_ERR_NO_RESOURCES            8
/**Request is well-formed and valid, but the server has
 * insufficient resources to complete the requested operation at this
 * time.  For example, the NAT device cannot create more mappings at
 * this time, is short of CPU cycles or memory, or is unable to
 * handle the request due to some other temporary condition.  The
 * same request may succeed in the future.  This is a system-wide
 * error, different from USER_EX_QUOTA.  This can be used as a catch-
 * all error, should no other error message be suitable.  This is a
 * short lifetime error.
 */

#define PCP_ERR_UNSUPP_PROTOCOL         9
/**Unsupported transport protocol, e.g.  SCTP in a
 * NAT that handles only UDP and TCP.  This is a long lifetime error.
 */

#define PCP_ERR_USER_EX_QUOTA           10
/** This attempt to create a new mapping would exceed
 * this subscriber's port quota.  This is a short lifetime error.
 */

#define PCP_ERR_CANNOT_PROVIDE_EXTERNAL 11
/** The suggested external port and/or
 * external address cannot be provided.  This error MUST only be
 * returned for:
 *      *  MAP requests that included the PREFER_FAILURE Option
 *          (normal MAP requests will return an available external port)
 *      *  MAP requests for the SCTP protocol (PREFER_FAILURE is implied)
 *      *  PEER requests
 */

#define PCP_ERR_ADDRESS_MISMATCH        12
/** The source IP address of the request packet does
 * not match the contents of the PCP Client's IP Address field, due
 * to an unexpected NAT on the path between the PCP client and the
 * PCP-controlled NAT or firewall.  This is a long lifetime error.
 */

#define PCP_ERR_EXCESSIVE_REMOTE_PEERS  13
/** The PCP server was not able to create the
 * filters in this request.  This result code MUST only be returned
 * if the MAP request contained the FILTER Option.  See Section 13.3
 * for processing information.  This is a long lifetime error.
 */

typedef enum pcp_options  {
    PCP_OPTION_3RD_PARTY = 1,
    PCP_OPTION_PREF_FAIL = 2,
    PCP_OPTION_FILTER = 3,
} pcp_options_t;


#define PCP_COMMON_REQUEST_SIZE (24)

#define PCP_COMMON_RESPONSE_SIZE (24)

#define PCP_OPTION_HDR_SIZE (4)

#define PCP_MAP_V2_SIZE (36)

#define PCP_MAP_V1_SIZE (24)

#define PCP_PEER_V1_SIZE (44)

#define PCP_PEER_V2_SIZE (56)

#define PCP_PREFER_FAIL_OPTION_SIZE (4)

#define PCP_3RD_PARTY_OPTION_SIZE (20)

#define PCP_FILTER_OPTION_SIZE (24)

/* default server settings, highest version supported is the default */
#define PCP_SERVER_INFO_THIS_SERVER_INFO 2


/* structure holding information from PCP msg*/
/* all variables are in host byte order except IP addresses */
typedef struct pcp_info {
	uint8_t     version;
	uint8_t     opcode;
	uint8_t     result_code;
	uint32_t    lifetime;             /* lifetime of the mapping */
	uint32_t    epochtime;
	/* both MAP and PEER opcode specific information */
	uint32_t        nonce[3];       /* random value generated by client */
	uint8_t     protocol;
	uint16_t    int_port;
	const struct in6_addr    *int_ip; /* in network order */
	uint16_t    ext_port;
	const struct in6_addr    *ext_ip; /* Suggested external IP in network order*/
	/* PEER specific information */

	uint8_t is_map_op;
	uint8_t is_peer_op;
	const struct in6_addr *thirdp_ip;
	const struct in6_addr *mapped_ip;
	char mapped_str[INET6_ADDRSTRLEN];
	int pfailure_present;
	struct in6_addr sender_ip;
	int is_fw; /* is this firewall operation? if not, nat. */
	char desc[64];
} pcp_info_t;

#if ENABLE_FEATURE_MINIUPNPD_NATPMP || ENABLE_FEATURE_IPV6
/* getPCPOpCodeStr()
 * return a string representation of the PCP OpCode
 * can be used for debug output */
static const char * getPCPOpCodeStr(uint8_t opcode)
{
	switch(opcode) {
	case PCP_OPCODE_ANNOUNCE:
		return "ANNOUNCE";
	case PCP_OPCODE_MAP:
		return "MAP";
	case PCP_OPCODE_PEER:
		return "PEER";
	default:
		return "UNKNOWN";
	}
}

/* useful to copy ext_ip only if needed, as request and response
 * buffers are same */
static void copyIPv6IfDifferent(void * dest, const void * src)
{
	if(dest != src) {
		memcpy(dest, src, sizeof(struct in6_addr));
	}
}

/*
 * Function extracting information from common_req (common request header)
 * into pcp_msg_info.
 * @return : when no problem occurred 0 is returned, 1 otherwise and appropriate
 *          result code is assigned to pcp_msg_info->result_code to indicate
 *          what kind of error occurred
 */
static int parseCommonRequestHeader(const uint8_t *common_req, pcp_info_t *pcp_msg_info)
{
	pcp_msg_info->version = common_req[0] ;
	pcp_msg_info->opcode = common_req[1] & 0x7f;
	pcp_msg_info->lifetime = READNU32(common_req + 4);
	pcp_msg_info->int_ip = (struct in6_addr *)(common_req + 8);
	pcp_msg_info->mapped_ip = (struct in6_addr *)(common_req + 8);


	if ( (pcp_msg_info->version > PCP_SERVER_INFO_THIS_SERVER_INFO) ) {
		pcp_msg_info->result_code = PCP_ERR_UNSUPP_VERSION;
		return 1;
	}

	if (pcp_msg_info->lifetime > bb_common_bss.max_lifetime ) {
		pcp_msg_info->lifetime = bb_common_bss.max_lifetime;
	}

	if ( (pcp_msg_info->lifetime < bb_common_bss.min_lifetime) && (pcp_msg_info->lifetime != 0) ) {
		pcp_msg_info->lifetime = bb_common_bss.min_lifetime;
	}

	return 0;
}

static void parsePCPMAP_version1(const uint8_t *map_v1,
		pcp_info_t *pcp_msg_info)
{
	pcp_msg_info->is_map_op = 1;
	pcp_msg_info->protocol = map_v1[0];
	pcp_msg_info->int_port = READNU16(map_v1 + 4);
	pcp_msg_info->ext_port = READNU16(map_v1 + 6);

	pcp_msg_info->ext_ip = (struct in6_addr *)(map_v1 + 8);
}

static void parsePCPMAP_version2(const uint8_t *map_v2,
		pcp_info_t *pcp_msg_info)
{
	pcp_msg_info->is_map_op = 1;
	memcpy(pcp_msg_info->nonce, map_v2, 12);
	pcp_msg_info->protocol = map_v2[12];
	pcp_msg_info->int_port = READNU16(map_v2 + 16);
	pcp_msg_info->ext_port = READNU16(map_v2 + 18);

	pcp_msg_info->ext_ip = (struct in6_addr *)(map_v2 + 20);
}

static int parsePCPOption(uint8_t* pcp_buf, int remain, pcp_info_t *pcp_msg_info)
{
	unsigned short option_length;

	/* Do centralized option sanity checks here. */

	if (remain < (int)PCP_OPTION_HDR_SIZE) {
		pcp_msg_info->result_code = PCP_ERR_MALFORMED_OPTION;
		return 0;
	}

	option_length = READNU16(pcp_buf + 2) + 4;      /* len */

	if (remain < option_length) {
		pcp_msg_info->result_code = PCP_ERR_MALFORMED_OPTION;
		return 0;
	}

	switch (pcp_buf[0]) { /* code */

	case PCP_OPTION_3RD_PARTY:

		if (option_length != PCP_3RD_PARTY_OPTION_SIZE) {
			pcp_msg_info->result_code = PCP_ERR_MALFORMED_OPTION;
			return 0;
		}
		if (pcp_msg_info->thirdp_ip ) {
			bb_error_msg("PCP: THIRD PARTY OPTION was already present. \n");
			pcp_msg_info->result_code = PCP_ERR_MALFORMED_OPTION;
			return 0;
		} else {
			pcp_msg_info->thirdp_ip = (struct in6_addr *)(pcp_buf + 4);
			pcp_msg_info->mapped_ip = (struct in6_addr *)(pcp_buf + 4);
		}
		break;

	case PCP_OPTION_PREF_FAIL:

		if (option_length != PCP_PREFER_FAIL_OPTION_SIZE) {
			pcp_msg_info->result_code = PCP_ERR_MALFORMED_OPTION;
			return 0;
		}
		if (pcp_msg_info->opcode != PCP_OPCODE_MAP) {
			DEBUG_LOG(syslog(LOG_DEBUG, "PCP: Unsupported OPTION for given OPCODE.\n");)
			pcp_msg_info->result_code = PCP_ERR_MALFORMED_REQUEST;
		}
		if (pcp_msg_info->pfailure_present != 0 ) {
			DEBUG_LOG(syslog(LOG_DEBUG, "PCP: PREFER FAILURE OPTION was already present.\n");)
			pcp_msg_info->result_code = PCP_ERR_MALFORMED_OPTION;
		} else {
			pcp_msg_info->pfailure_present = 1;
		}
		break;

	case PCP_OPTION_FILTER:
		/* TODO fully implement filter */
		if (option_length != PCP_FILTER_OPTION_SIZE) {
			pcp_msg_info->result_code = PCP_ERR_MALFORMED_OPTION;
			return 0;
		}
		if (pcp_msg_info->opcode != PCP_OPCODE_MAP) {
			bb_error_msg("PCP: Unsupported OPTION for given OPCODE.");
			pcp_msg_info->result_code = PCP_ERR_MALFORMED_REQUEST;
			return 0;
		}
		break;

	default:
		if (pcp_buf[0] < 128) {
			bb_error_msg("PCP: Unrecognized mandatory PCP OPTION: %d", (int)pcp_buf[0]);
			/* Mandatory to understand */
			pcp_msg_info->result_code = PCP_ERR_UNSUPP_OPTION;
			remain = 0;
			break;
		}
		/* TODO - log optional not understood options? */
		break;
	}
	return option_length;
}


static void parsePCPOptions(void* pcp_buf, int remain, pcp_info_t *pcp_msg_info)
{
	int option_length;

	while (remain > 0) {
		option_length = parsePCPOption(pcp_buf, remain, pcp_msg_info);
		if (!option_length)
			break;
		remain -= option_length;
		pcp_buf += option_length;
	}
	if (remain > 0) {
		xlog_warn("%s: remain=%d", "parsePCPOptions", remain);
	}
}


static int getifaddr_in6(const char * ifname, int af, struct in6_addr *addr)
{
#if ENABLE_FEATURE_IPV6
	struct ifaddrs * ifap;
	struct ifaddrs * ife;
	const struct sockaddr_in6 * tmpaddr;
	int found = 0;

	if(!ifname || ifname[0]=='\0')
		return -1;
	if(getifaddrs(&ifap)<0) {
		bb_perror_msg("getifaddrs");
		return -1;
	}
	for(ife = ifap; ife && !found; ife = ife->ifa_next) {
		/* skip other interfaces if one was specified */
		if(ifname && (0 != strcmp(ifname, ife->ifa_name)))
			continue;
		if(ife->ifa_addr == NULL)
			continue;
		if (ife->ifa_addr->sa_family != af)
			continue;
		switch(ife->ifa_addr->sa_family)
		{
		case AF_INET:
			/* IPv4-mapped IPv6 address ::ffff:1.2.3.4 */
			memset(addr->s6_addr, 0, 10);
			addr->s6_addr[10] = 0xff;
			addr->s6_addr[11] = 0xff;
			memcpy(addr->s6_addr + 12,
			       &(((struct sockaddr_in *)ife->ifa_addr)->sin_addr.s_addr),
			       4);
			found = 1;
			break;

#if ENABLE_FEATURE_IPV6
		case AF_INET6:
			tmpaddr = (const struct sockaddr_in6 *)ife->ifa_addr;
			if(!IN6_IS_ADDR_LOOPBACK(&tmpaddr->sin6_addr)
			   && !IN6_IS_ADDR_LINKLOCAL(&tmpaddr->sin6_addr))
			{
				memcpy(addr->s6_addr,
				       &tmpaddr->sin6_addr,
				       16);
				found = 1;
			}
			break;
#endif
		}
	}
	freeifaddrs(ifap);
	return (found ? 0 : -1);
#else /* ENABLE_FEATURE_IPV6 */
	/* IPv4 only */
	struct in_addr addr4;
	if(af != AF_INET)
		return -1;
	if(getifaddr(ifname, NULL, 0, &addr4, NULL) < 0)
		return -1;
	/* IPv4-mapped IPv6 address ::ffff:1.2.3.4 */
	memset(addr->s6_addr, 0, 10);
	addr->s6_addr[10] = 0xff;
	addr->s6_addr[11] = 0xff;
	memcpy(addr->s6_addr + 12, &addr4.s_addr, 4);
	return 0;
#endif
}

/* CheckExternalAddress()
 * Check that suggested external address in request match a real external
 * IP address.
 * Suggested address can also be 0 IPv4 or IPv6 address.
 *  (see http://tools.ietf.org/html/rfc6887#section-10 )
 * return values :
 *   0 : check is OK
 *  -1 : check failed */
static int CheckExternalAddress(pcp_info_t* pcp_msg_info)
{
	int af;

	af = IN6_IS_ADDR_V4MAPPED(pcp_msg_info->mapped_ip)
		? AF_INET : AF_INET6;

	pcp_msg_info->is_fw = af == AF_INET6;

	if (pcp_msg_info->is_fw) {
		external_addr = *pcp_msg_info->mapped_ip;
	} else {
		/* TODO : be able to handle case with multiple
		 * external addresses */
		if(bb_common_bss.use_ext_ip_addr) {
			if (inet_pton(AF_INET, bb_common_bss.use_ext_ip_addr,
				      ((uint32_t*)external_addr.s6_addr)+3) == 1) {
				((uint32_t*)external_addr.s6_addr)[0] = 0;
				((uint32_t*)external_addr.s6_addr)[1] = 0;
				((uint32_t*)external_addr.s6_addr)[2] = htonl(0xFFFF);
			} else if (inet_pton(AF_INET6, bb_common_bss.use_ext_ip_addr, external_addr.s6_addr)
				   != 1) {
				pcp_msg_info->result_code = PCP_ERR_NETWORK_FAILURE;
				return -1;
			}
		} else {
			if(!bb_common_bss.ext_if_name || bb_common_bss.ext_if_name[0]=='\0') {
				pcp_msg_info->result_code = PCP_ERR_NETWORK_FAILURE;
				return -1;
			}
			if(getifaddr_in6(bb_common_bss.ext_if_name, af, &external_addr) < 0) {
				pcp_msg_info->result_code = PCP_ERR_NETWORK_FAILURE;
				return -1;
			}
		}
	}
	if (pcp_msg_info->ext_ip == NULL ||
	    IN6_IS_ADDR_UNSPECIFIED(pcp_msg_info->ext_ip) ||
	    (IN6_IS_ADDR_V4MAPPED(pcp_msg_info->ext_ip)
	      && ((uint32_t *)pcp_msg_info->ext_ip->s6_addr)[3] == INADDR_ANY)) {
		/* no suggested external address : use real external address */
		pcp_msg_info->ext_ip = &external_addr;
		return 0;
	}

	if (!IN6_ARE_ADDR_EQUAL(pcp_msg_info->ext_ip, &external_addr)) {
		bb_perror_msg("PCP: External IP in request didn't match interface IP");
		if (pcp_msg_info->pfailure_present) {
			pcp_msg_info->result_code = PCP_ERR_CANNOT_PROVIDE_EXTERNAL;
			return -1;
		} else {
			pcp_msg_info->ext_ip = &external_addr;
		}

	}

	return 0;
}


static const char* inet_n46top(const struct in6_addr* in,
			       char* buf, size_t buf_len)
{
	if (IN6_IS_ADDR_V4MAPPED(in)) {
		return inet_ntop(AF_INET, ((uint32_t*)(in->s6_addr))+3, buf, buf_len);
	} else {
		return inet_ntop(AF_INET6, in, buf, buf_len);
	}
}

static int CreatePCPMap_NAT(pcp_info_t *pcp_msg_info)
{
	int r = 0;
	char iaddr_old[INET6_ADDRSTRLEN];
	uint16_t iport_old, eport_first = 0;
	int any_eport_allowed = 0;
	unsigned int timestamp = time(NULL) + pcp_msg_info->lifetime;

	if (pcp_msg_info->ext_port == 0) {
		pcp_msg_info->ext_port = pcp_msg_info->int_port;
	}

	/* TODO: Support non-TCP/UDP */
	if (pcp_msg_info->ext_port == 0) {
		return PCP_ERR_MALFORMED_REQUEST;
	}

	do {
		if (eport_first == 0) { /* first time in loop */
			eport_first = pcp_msg_info->ext_port;
		} else if (pcp_msg_info->ext_port == eport_first) { /* no eport available */
			/* all eports rejected by permissions? */
			if (any_eport_allowed == 0)
				return PCP_ERR_NOT_AUTHORIZED;
			/* at least one eport allowed (but none available) */
			return PCP_ERR_NO_RESOURCES;
		}
		if ((IN6_IS_ADDR_V4MAPPED(pcp_msg_info->mapped_ip) &&
		     (!check_upnp_rule_against_permissions(pcp_msg_info->ext_port,
							   ((struct in_addr*)pcp_msg_info->mapped_ip->s6_addr)[3],
							   pcp_msg_info->int_port)))) {
			if (pcp_msg_info->pfailure_present) {
				return PCP_ERR_CANNOT_PROVIDE_EXTERNAL;
			}
			pcp_msg_info->ext_port++;
			if (pcp_msg_info->ext_port == 0) { /* skip port zero */
				pcp_msg_info->ext_port++;
			}
			continue;
		}
		any_eport_allowed = 1;
		r = get_redirect_rule(pcp_msg_info->ext_port,
				      pcp_msg_info->protocol,
				      iaddr_old, sizeof(iaddr_old),
				      &iport_old, 0, 0, 0, 0,
				      NULL/*&timestamp*/, 0, 0);

		if(r==0) {
			if((strcmp(pcp_msg_info->mapped_str, iaddr_old)!=0)
			   || (pcp_msg_info->int_port != iport_old)) {
				/* redirection already existing */
				if (pcp_msg_info->pfailure_present) {
					return PCP_ERR_CANNOT_PROVIDE_EXTERNAL;
				}
			} else {
				DEBUG_LOG(syslog(LOG_INFO, "port %hu %s already redirected to %s:%hu, replacing",
				       pcp_msg_info->ext_port, (pcp_msg_info->protocol==IPPROTO_TCP)?"tcp":"udp",
				       iaddr_old, iport_old);)
				/* remove and then add again */
				if (_upnp_delete_redir(pcp_msg_info->ext_port,
						       pcp_msg_info->protocol)==0) {
					break;
				} else if (pcp_msg_info->pfailure_present) {
					return PCP_ERR_CANNOT_PROVIDE_EXTERNAL;
				}
			}
			pcp_msg_info->ext_port++;
			if (pcp_msg_info->ext_port == 0) { /* skip port zero */
				pcp_msg_info->ext_port++;
			}
		}
	} while (r==0);

	r = upnp_redirect_internal(NULL,
				   pcp_msg_info->ext_port,
				   pcp_msg_info->mapped_str,
				   pcp_msg_info->int_port,
				   pcp_msg_info->protocol,
				   pcp_msg_info->desc,
				   timestamp);
	if (r < 0)
		return PCP_ERR_NO_RESOURCES;
	return PCP_SUCCESS;
}


#ifdef ENABLE_UPNPPINHOLE
/* upnp_add_inboundpinhole()
 * returns:  1 on success
 *          -1 Pinhole space exhausted
 *          -4 invalid arguments
 *         -42 not implemented
 * TODO : return uid on success (positive) or error value (negative)
 */
static int
upnp_add_inboundpinhole(const char * raddr,
			unsigned short rport, const char * iaddr,
			unsigned short iport, int proto,
			char *desc, unsigned int leasetime, int * uid)
{
	int r;
	time_t current;
	unsigned int timestamp;
	struct in6_addr address;

	r = inet_pton(AF_INET6, iaddr, &address);
	if(r <= 0) {
		bb_error_msg("inet_pton(%d, %s, %p) FAILED", AF_INET6, iaddr, &address);
		return -4;
	}
	current = time(NULL);
	timestamp = current + leasetime;
	r = 0;

#if defined(USE_NETFILTER)
	*uid = add_pinhole (0/*ext_if_name*/, raddr, rport,
			    iaddr, iport, proto, desc, timestamp);
	return *uid >= 0 ? 1 : -1;
#else
	return -42;     /* not implemented */
#endif
}

/* update the lease time */
static int
upnp_update_inboundpinhole(unsigned short uid, unsigned int leasetime)
{
#if defined(USE_NETFILTER)
	unsigned int timestamp;

	timestamp = time(NULL) + leasetime;
	return update_pinhole(uid, timestamp);
#else
	UNUSED(uid); UNUSED(leasetime);

	return -42; /* not implemented */
#endif
}

/* find an inbound pinhole base on remove host:port / local host:port
 * return the (positive) uid or a negative value if not found */
static int
upnp_find_inboundpinhole(const char * raddr, unsigned short rport,
			 const char * iaddr, unsigned short iport,
			 int proto, char * desc, int desc_len, unsigned int * leasetime)
{
#if defined(USE_NETFILTER)
	int uid;
	uid = find_pinhole(NULL, raddr, rport, iaddr, iport, proto,
			   desc, desc_len, leasetime);
	return uid;
#else
	return -42;
#endif
}
#endif

static int CreatePCPMap_FW(pcp_info_t *pcp_msg_info)
{
#ifdef ENABLE_UPNPPINHOLE
	int uid;
	int r;
	/* first check if pinhole already exists */
	uid = upnp_find_inboundpinhole(NULL, 0,
					pcp_msg_info->mapped_str,
					pcp_msg_info->int_port,
					pcp_msg_info->protocol,
					NULL, 0, /* desc */
					NULL /* lifetime */);
	if(uid >= 0) {
		/* pinhole already exists, updating */
		DEBUG_LOG(syslog(LOG_INFO, "updating pinhole to %s:%hu %s",
		       pcp_msg_info->mapped_str, pcp_msg_info->int_port,
		       (pcp_msg_info->protocol == IPPROTO_TCP)?"TCP":"UDP");)
		r = upnp_update_inboundpinhole((unsigned short)uid, pcp_msg_info->lifetime);
		return r >= 0 ? PCP_SUCCESS : PCP_ERR_NO_RESOURCES;
	} else {
		r = upnp_add_inboundpinhole(NULL, 0,
						pcp_msg_info->mapped_str,
						pcp_msg_info->int_port,
						pcp_msg_info->protocol,
						pcp_msg_info->desc,
						pcp_msg_info->lifetime,
						&uid);
		if (r < 0)
			return PCP_ERR_NO_RESOURCES;
		pcp_msg_info->ext_port = pcp_msg_info->int_port;
		return PCP_SUCCESS;
	}
#else
	UNUSED(pcp_msg_info);
	return PCP_ERR_NO_RESOURCES;
#endif /* ENABLE_UPNPPINHOLE */
}

#ifdef ENABLE_UPNPPINHOLE
/* remove the inbound pinhole */
static int
upnp_delete_inboundpinhole(unsigned short uid)
{
#if defined(USE_NETFILTER)
	return delete_pinhole(uid);
#else
	UNUSED(uid);

	return -1;
#endif
}

/* get from uid
 * return values :
 *   0   OK
 *  -1   Internal error
 *  -2   NOT FOUND (no such entry)
 *  ..
 *  -42  Not implemented
 */
static int
upnp_get_pinhole_info(unsigned short uid,
		      char * raddr, int raddrlen,
		      unsigned short * rport,
		      char * iaddr, int iaddrlen,
		      unsigned short * iport,
		      int * proto, char * desc, int desclen,
		      unsigned int * leasetime,
		      unsigned int * packets)
{
	/* Call Firewall specific code to get IPv6 pinhole infos */
#if defined(USE_NETFILTER)
	int r;
	unsigned int timestamp;
	u_int64_t packets_tmp;
	/*u_int64_t bytes_tmp;*/

	r = get_pinhole_info(uid, raddr, raddrlen, rport,
			     iaddr, iaddrlen, iport,
			     proto, desc, desclen,
			     leasetime ? &timestamp : NULL,
			     packets ? &packets_tmp : NULL,
			     NULL/*&bytes_tmp*/);
	if(r >= 0) {
		if(leasetime) {
			time_t current_time;
			current_time = time(NULL);
			if(timestamp > (unsigned int)current_time)
				*leasetime = timestamp - current_time;
			else
				*leasetime = 0;
		}
		if(packets)
			*packets = (unsigned int)packets_tmp;
	}
	return r;
#else
	UNUSED(uid);
	UNUSED(raddr); UNUSED(raddrlen); UNUSED(rport);
	UNUSED(iaddr); UNUSED(iaddrlen); UNUSED(iport);
	UNUSED(proto); UNUSED(desc); UNUSED(desclen);
	UNUSED(leasetime); UNUSED(packets);
	return -42;     /* not implemented */
#endif
}

/*
 * return values:
 * -1 = not found
 * 0 .. 65535 = uid of the rule for the index
 */
static int
upnp_get_pinhole_uid_by_index(int index)
{
#if defined (USE_NETFILTER)
	return get_pinhole_uid_by_index(index);
#else
	UNUSED(index);
	return -42;
#endif /* defined (USE_NETFILTER) */
}

#endif

/*                internal  external  PCP remote peer  actual remote peer
 *                --------  -------   ---------------  ------------------
 * IPv4 firewall   IPv4      IPv4         IPv4              IPv4
 * IPv6 firewall   IPv6      IPv6         IPv6              IPv6
 *         NAT44   IPv4      IPv4         IPv4              IPv4
 *         NAT46   IPv4      IPv6         IPv4              IPv6
 *         NAT64   IPv6      IPv4         IPv6              IPv4
 *         NPTv6   IPv6      IPv6         IPv6              IPv6
 *
 *             Address Families with MAP and PEER
 *
 * The 'internal' address is implicitly the same as the source IP
 * address of the PCP request, except when the THIRD_PARTY option is
 * used.
 *
 * The 'external' address is the Suggested External Address field of the
 * MAP or PEER request, and its address family is usually the same as
 * the 'internal' address family, except when technologies like NAT64
 * are used.
 *
 * The 'remote peer' address is the remote peer IP address of the PEER
 * request or the FILTER option of the MAP request, and is always the
 * same address family as the 'internal' address, even when NAT64 is
 * used.  In NAT64, the IPv6 PCP client is not necessarily aware of the
 * NAT64 or aware of the actual IPv4 address of the remote peer, so it
 * expresses the IPv6 address from its perspective.     */

/* TODO: Support more than basic NAT44 / IPv6 firewall cases. */
static void CreatePCPMap(pcp_info_t *pcp_msg_info)
{
	int r;

	if (pcp_msg_info->is_fw)
		r = CreatePCPMap_FW(pcp_msg_info);
	else
		r = CreatePCPMap_NAT(pcp_msg_info);
	pcp_msg_info->result_code = r;
#ifdef DEBUG
	syslog(r == PCP_SUCCESS ? LOG_INFO : LOG_ERR,
		"PCP MAP: %s mapping %s %hu->%s:%hu '%s'",
		r == PCP_SUCCESS ? "added" : "failed to add",
#else
	if(r != PCP_SUCCESS)
		bb_error_msg("PCP MAP: failed to add mapping %s %hu->%s:%hu '%s'",
#endif
	       (pcp_msg_info->protocol==IPPROTO_TCP)?"TCP":"UDP",
	       pcp_msg_info->ext_port,
	       pcp_msg_info->mapped_str,
	       pcp_msg_info->int_port,
	       pcp_msg_info->desc);
}

static void DeletePCPMap(pcp_info_t *pcp_msg_info)
{
	uint16_t iport = pcp_msg_info->int_port;  /* private port */
	uint8_t  proto = pcp_msg_info->protocol;
	int r=-1;
	/* remove the mapping */
	/* remove all the mappings for this client */
	int index;
	unsigned short eport2, iport2;
	char iaddr2[INET6_ADDRSTRLEN];
	int proto2;
	char desc[64];
	unsigned int timestamp;
#ifdef ENABLE_UPNPPINHOLE
	int uid = -1;
#endif /* ENABLE_UPNPPINHOLE */

	/* iterate through all rules and delete the requested ones */
	for (index = 0 ;
	     (!pcp_msg_info->is_fw &&
	      get_redirect_rule_by_index(index, 0,
					 &eport2, iaddr2, sizeof(iaddr2),
					 &iport2, &proto2,
					 desc, sizeof(desc),
					 0, 0, &timestamp, 0, 0) >= 0)
#ifdef ENABLE_UPNPPINHOLE
	       ||
	     (pcp_msg_info->is_fw &&
	      (uid=upnp_get_pinhole_uid_by_index(index))>=0 &&
	      upnp_get_pinhole_info((unsigned short)uid,
				    NULL, 0, NULL,
				    iaddr2, sizeof(iaddr2), &iport2,
				    &proto2, desc, sizeof(desc),
				    &timestamp, NULL) >= 0)
#endif /* ENABLE_UPNPPINHOLE */
		     ;
	     index++)
		if(0 == strcmp(iaddr2, pcp_msg_info->mapped_str)
		   && (proto2==proto)
		   && ((iport2==iport) || (iport==0))) {
			if(0 != strcmp(desc, pcp_msg_info->desc)) {
				/* nonce does not match */
				pcp_msg_info->result_code = PCP_ERR_NOT_AUTHORIZED;
				bb_error_msg("Unauthorized to remove PCP mapping internal port %hu, protocol %s",
				       iport, (pcp_msg_info->protocol == IPPROTO_TCP)?"TCP":"UDP");
				return;
			} else if (!pcp_msg_info->is_fw) {
				r = _upnp_delete_redir(eport2, proto2);
			} else {
#ifdef ENABLE_UPNPPINHOLE
				r = upnp_delete_inboundpinhole(uid);
#endif /* ENABLE_UPNPPINHOLE */
			}
			break;
		}
	if (r >= 0) {
		DEBUG_LOG(syslog(LOG_INFO, "PCP: %s port %hu mapping removed",
		       proto2==IPPROTO_TCP?"TCP":"UDP", eport2));
	} else {
		bb_error_msg("Failed to remove PCP mapping internal port %hu, protocol %s",
		       iport, (pcp_msg_info->protocol == IPPROTO_TCP)?"TCP":"UDP");
		pcp_msg_info->result_code = PCP_ERR_NO_RESOURCES;
	}
}

static int ValidatePCPMsg(pcp_info_t *pcp_msg_info)
{
	if (pcp_msg_info->result_code)
		return 0;

	/* RFC 6887, section 8.2: MUST return address mismatch if NAT
	 * in middle. */
	if (memcmp(pcp_msg_info->int_ip,
		   &pcp_msg_info->sender_ip,
		   sizeof(pcp_msg_info->sender_ip)) != 0) {
		pcp_msg_info->result_code = PCP_ERR_ADDRESS_MISMATCH;
		return 0;
	}

	if (pcp_msg_info->thirdp_ip) {
		if (!GETFLAG(PCP_ALLOWTHIRDPARTYMASK)) {
			pcp_msg_info->result_code = PCP_ERR_UNSUPP_OPTION;
			return 0;
		}

		/* RFC687, section 13.1 - if sender ip == THIRD_PARTY,
		 * it's an error. */
		if (memcmp(pcp_msg_info->thirdp_ip,
			   &pcp_msg_info->sender_ip,
			   sizeof(pcp_msg_info->sender_ip)) == 0) {
			pcp_msg_info->result_code = PCP_ERR_MALFORMED_REQUEST;
			return 0;
		}
	}

	/* Produce mapped_str for future use. */
	if (!inet_n46top(pcp_msg_info->mapped_ip, pcp_msg_info->mapped_str,
			 sizeof(pcp_msg_info->mapped_str))) {
		bb_perror_msg("inet_ntop(pcpserver)");
		return 0;
	}

	/* protocol zero means 'all protocols' : internal port MUST be zero */
	if (pcp_msg_info->protocol == 0 && pcp_msg_info->int_port != 0) {
		bb_error_msg("PCP %s: Protocol was ZERO, but internal port "
		       "has non-ZERO value.", getPCPOpCodeStr(pcp_msg_info->opcode));
		pcp_msg_info->result_code = PCP_ERR_MALFORMED_REQUEST;
		return 0;
	}

	if (pcp_msg_info->pfailure_present) {
		if ( (IN6_IS_ADDR_UNSPECIFIED(pcp_msg_info->ext_ip) ||
		      ((IN6_IS_ADDR_V4MAPPED(pcp_msg_info->ext_ip)) &&
		       (((uint32_t*)pcp_msg_info->ext_ip->s6_addr)[3] == 0))) &&
		     (pcp_msg_info->ext_port == 0)
		   )
		{
			pcp_msg_info->result_code = PCP_ERR_MALFORMED_OPTION;
			return 0;
		}
	}

	if (CheckExternalAddress(pcp_msg_info))
		return 0;

	/* Fill in the desc that describes uniquely what flow we're
	 * dealing with (same code used in both create + delete of
	 * MAP/PEER) */
	switch (pcp_msg_info->opcode) {
	case PCP_OPCODE_MAP:
	case PCP_OPCODE_PEER:
		snprintf(pcp_msg_info->desc, sizeof(pcp_msg_info->desc),
			 "PCP %s %08x%08x%08x",
			 getPCPOpCodeStr(pcp_msg_info->opcode),
			 pcp_msg_info->nonce[0],
			 pcp_msg_info->nonce[1], pcp_msg_info->nonce[2]);
		break;
	}
	return 1;
}

/*
 * return value indicates whether the request is valid or not.
 * Based on the return value simple response can be formed.
 */
static int processPCPRequest(void * req, int req_size, pcp_info_t *pcp_msg_info)
{
	int remainingSize;

	/* start with PCP_SUCCESS as result code,
	 * if everything is OK value will be unchanged */
	pcp_msg_info->result_code = PCP_SUCCESS;

	remainingSize = req_size;

	/* discard request that exceeds maximal length,
	   or that is shorter than PCP_MIN_LEN (=24)
	   or that is not the multiple of 4 */
	if (req_size < 3)
		return 0; /* ignore msg */

	if (req_size < PCP_MIN_LEN) {
		pcp_msg_info->result_code = PCP_ERR_MALFORMED_REQUEST;
		return 1; /* send response */
	}

	if ( (req_size > PCP_MAX_LEN) || ( (req_size & 3) != 0)) {
		bb_error_msg("PCP: Size of PCP packet(%d) is larger than %d bytes or "
		       "the size is not multiple of 4.\n", req_size, PCP_MAX_LEN);
		pcp_msg_info->result_code = PCP_ERR_MALFORMED_REQUEST;
		return 1; /* send response */
	}

	/* first parse request header */
	if (parseCommonRequestHeader(req, pcp_msg_info) ) {
		return 1;
	}

	remainingSize -= PCP_COMMON_REQUEST_SIZE;
	req += PCP_COMMON_REQUEST_SIZE;

	if (pcp_msg_info->version == 1) {
		/* legacy PCP version 1 support */
		switch (pcp_msg_info->opcode) {
		case PCP_OPCODE_MAP:

			remainingSize -= PCP_MAP_V1_SIZE;
			if (remainingSize < 0) {
				pcp_msg_info->result_code = PCP_ERR_MALFORMED_REQUEST;
				return pcp_msg_info->result_code;
			}

			parsePCPMAP_version1(req, pcp_msg_info);

			req += PCP_MAP_V1_SIZE;

			parsePCPOptions(req, remainingSize, pcp_msg_info);
			if (ValidatePCPMsg(pcp_msg_info)) {
				if (pcp_msg_info->lifetime == 0) {
					DeletePCPMap(pcp_msg_info);
				} else {
					CreatePCPMap(pcp_msg_info);
				}
			} else {
				bb_error_msg("PCP: Invalid PCP v1 MAP message.");
				return pcp_msg_info->result_code;
			}
			break;

		default:
			pcp_msg_info->result_code = PCP_ERR_UNSUPP_OPCODE;
			break;
		}

	} else if (pcp_msg_info->version == 2) {
		/* RFC 6887 PCP support
		 * http://tools.ietf.org/html/rfc6887 */
		switch (pcp_msg_info->opcode) {
		case PCP_OPCODE_ANNOUNCE:
			/* should check PCP Client's IP Address in request */
			/* see http://tools.ietf.org/html/rfc6887#section-14.1 */
			break;
		case PCP_OPCODE_MAP:

			remainingSize -= PCP_MAP_V2_SIZE;
			if (remainingSize < 0) {
				pcp_msg_info->result_code = PCP_ERR_MALFORMED_REQUEST;
				return pcp_msg_info->result_code;
			}

			parsePCPMAP_version2(req, pcp_msg_info);
			req += PCP_MAP_V2_SIZE;

			parsePCPOptions(req, remainingSize, pcp_msg_info);

			if (ValidatePCPMsg(pcp_msg_info)) {
				if (pcp_msg_info->lifetime == 0) {
					DeletePCPMap(pcp_msg_info);
				} else {
					CreatePCPMap(pcp_msg_info);
				}
			} else {
				bb_error_msg("PCP: Invalid PCP v2 MAP message.");
				return pcp_msg_info->result_code;
			}


			break;

		default:
			pcp_msg_info->result_code = PCP_ERR_UNSUPP_OPCODE;
			break;
		}
	} else {
		pcp_msg_info->result_code = PCP_ERR_UNSUPP_VERSION;
		return pcp_msg_info->result_code;
	}
	return 1;
}


static void createPCPResponse(unsigned char *response, pcp_info_t *pcp_msg_info)
{
	response[2] = 0;        /* reserved */
	memset(response + 12, 0, 12);   /* reserved */
	if (pcp_msg_info->result_code == PCP_ERR_UNSUPP_VERSION ) {
		/* highest supported version */
		response[0] = PCP_SERVER_INFO_THIS_SERVER_INFO;
	} else {
		response[0] = pcp_msg_info->version;
	}

	response[1] |= 0x80;    /* r_opcode */
	response[3] = pcp_msg_info->result_code;
	WRITENU32(response + 8, time(NULL) - bb_common_bss.startup_time); /* epochtime */
	switch (pcp_msg_info->result_code) {
	/*long lifetime errors*/
	case PCP_ERR_UNSUPP_VERSION:
	case PCP_ERR_NOT_AUTHORIZED:
	case PCP_ERR_MALFORMED_REQUEST:
	case PCP_ERR_UNSUPP_OPCODE:
	case PCP_ERR_UNSUPP_OPTION:
	case PCP_ERR_MALFORMED_OPTION:
	case PCP_ERR_UNSUPP_PROTOCOL:
	case PCP_ERR_ADDRESS_MISMATCH:
	case PCP_ERR_CANNOT_PROVIDE_EXTERNAL:
	case PCP_ERR_EXCESSIVE_REMOTE_PEERS:
		WRITENU32(response + 4, 0);     /* lifetime */
		break;

	case PCP_ERR_NETWORK_FAILURE:
	case PCP_ERR_NO_RESOURCES:
	case PCP_ERR_USER_EX_QUOTA:
		WRITENU32(response + 4, 30);    /* lifetime */
		break;

	case PCP_SUCCESS:
	default:
		WRITENU32(response + 4, pcp_msg_info->lifetime);        /* lifetime */
		break;
	}

	if (response[1] == 0x81) { /* MAP response */
		if (response[0] == 1) { /* version */
			WRITENU16(response + PCP_COMMON_RESPONSE_SIZE + 4, pcp_msg_info->int_port);
			WRITENU16(response + PCP_COMMON_RESPONSE_SIZE + 6, pcp_msg_info->ext_port);
			copyIPv6IfDifferent(response + PCP_COMMON_RESPONSE_SIZE + 8,
					    pcp_msg_info->ext_ip);
		}
		else if (response[0] == 2) {
			WRITENU16(response + PCP_COMMON_RESPONSE_SIZE + 16, pcp_msg_info->int_port);
			WRITENU16(response + PCP_COMMON_RESPONSE_SIZE + 18, pcp_msg_info->ext_port);
			copyIPv6IfDifferent(response + PCP_COMMON_RESPONSE_SIZE + 20,
					    pcp_msg_info->ext_ip);
		}
	}
}

/*
 * receiveraddr is only used for IPV6
 *
 * returns 0 upon success 1 otherwise
 */

#if ENABLE_FEATURE_IPV6
static int ProcessIncomingPCPPacket(int s, unsigned char *buff, int len,
			     const struct sockaddr *senderaddr,
			     const struct sockaddr_in6 *receiveraddr)
#else
static int ProcessIncomingPCPPacket(int s, unsigned char *buff, int len,
			     const struct sockaddr *senderaddr,
			     void *unuse)
#endif
{
	pcp_info_t pcp_msg_info;
	struct lan_addr_s * lan_addr;
	char addr_str[64];

#if !ENABLE_FEATURE_IPV6
	UNUSED(unuse);
#endif
	memset(&pcp_msg_info, 0, sizeof(pcp_info_t));

	if(senderaddr->sa_family == AF_INET) {
		const struct sockaddr_in * senderaddr_v4 =
			(const struct sockaddr_in *)senderaddr;
		pcp_msg_info.sender_ip.s6_addr[11] = 0xff;
		pcp_msg_info.sender_ip.s6_addr[10] = 0xff;
		memcpy(pcp_msg_info.sender_ip.s6_addr+12,
		       &senderaddr_v4->sin_addr, 4);
	}
#if ENABLE_FEATURE_IPV6
	  else if(senderaddr->sa_family == AF_INET6) {
		const struct sockaddr_in6 * senderaddr_v6 =
			(const struct sockaddr_in6 *)senderaddr;
		pcp_msg_info.sender_ip = senderaddr_v6->sin6_addr;
	}
#endif
	  else {
		xlog_warn("unknown PCP packet sender address family %d",
		       senderaddr->sa_family);
		return 0;
	}

	if(sockaddr_to_string(senderaddr, addr_str, sizeof(addr_str)))
		DEBUG_LOG(syslog(LOG_DEBUG, "PCP request received from %s %dbytes", addr_str, len));

	if(buff[1] & 128) {
		/* discarding PCP responses silently */
		return 0;
	}

	/* If we're in allow third party-mode, we probably don't care
	 * about locality either. Let's hope firewall is ok. */
	if (!GETFLAG(PCP_ALLOWTHIRDPARTYMASK)) {
		lan_addr = get_lan_for_peer(senderaddr);
		if(lan_addr == NULL) {
			xlog_warn("PCP packet sender %s not from a LAN, ignoring",
			       addr_str);
			return 0;
		}
	}

	if (processPCPRequest(buff, len, &pcp_msg_info) ) {

		createPCPResponse(buff, &pcp_msg_info);

		if(len < PCP_MIN_LEN)
			len = PCP_MIN_LEN;
		else
			len = (len + 3) & ~3;   /* round up resp. length to multiple of 4 */
		len = sendto_schedule2(s, buff, len, 0, senderaddr,
#if ENABLE_FEATURE_IPV6
			   (senderaddr->sa_family == AF_INET) ?
				  sizeof(struct sockaddr_in) :
				  sizeof(struct sockaddr_in6), receiveraddr, 0);
#else
				  sizeof(struct sockaddr_in), 0);
#endif
		if( len < 0 )
			bb_perror_msg("sendto(pcpserver)");
	}

	return 0;
}
#endif /* ENABLE_FEATURE_MINIUPNPD_NATPMP || ENABLE_FEATURE_IPV6 */

#if ENABLE_FEATURE_IPV6
static int OpenAndConfPCPv6Socket(void)
{
	int s;
	struct sockaddr_in6 addr;
	s = bb_xsocket(BB_XSOCKET_FLG_AF_INET6 | BB_XSOCKET_FLG_SOCK_DGRAM);
	if(s < 0)
		return -1;
	if(setsockopt_1i(s, SOL_SOCKET, SO_REUSEADDR, 1) < 0) {
		xlog_warn("%s: setsockopt(SO_REUSEADDR): %m",
		       "OpenAndConfPCPv6Socket");
	}
#ifdef IPV6_RECVPKTINFO
	/* see RFC3542 */
	if(setsockopt_1i(s, IPPROTO_IPV6, IPV6_RECVPKTINFO, 1) < 0) {
		xlog_warn("%s: setsockopt(IPV6_RECVPKTINFO): %m",
		       "OpenAndConfPCPv6Socket");
	}
#endif
	if(!set_non_blocking(s)) {
		xlog_warn("%s: set_non_blocking(): %m",
		       "OpenAndConfPCPv6Socket");
	}
	memset(&addr, 0, sizeof(addr));
	addr.sin6_family = AF_INET6;
	addr.sin6_port = htons(NATPMP_PORT);
	addr.sin6_addr = bb_common_bss.ipv6_bind_addr;
	if(bind(s, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		bb_perror_msg("%s: bind()", "OpenAndConfPCPv6Socket");
		close(s);
		return -1;
	}
	return s;
}
#endif /*ENABLE_FEATURE_IPV6*/
#endif /*ENABLE_PCP*/

#if ENABLE_FEATURE_MINIUPNPD_NATPMP

static int OpenAndConfNATPMPSocket(in_addr_t addr)
{
	int snatpmp;
	snatpmp = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM);
	if(snatpmp < 0)
		return -1;
	if(setsockopt_1i(snatpmp, SOL_SOCKET, SO_REUSEADDR, 1) < 0) {
		xlog_warn("%s: setsockopt(SO_REUSEADDR): %m",
		       "OpenAndConfNATPMPSocket");
	}
	if(!set_non_blocking(snatpmp)) {
		xlog_warn("%s: set_non_blocking(): %m",
		       "OpenAndConfNATPMPSocket");
	}
	{
		struct sockaddr_in natpmp_addr;
		memset(&natpmp_addr, 0, sizeof(natpmp_addr));
		natpmp_addr.sin_family = AF_INET;
		natpmp_addr.sin_port = htons(NATPMP_PORT);
		/*natpmp_addr.sin_addr.s_addr = INADDR_ANY; */
		natpmp_addr.sin_addr.s_addr = addr;
		if(bind(snatpmp, (struct sockaddr *)&natpmp_addr, sizeof(natpmp_addr)) < 0) {
			bb_perror_msg("%s: bind()", "OpenAndConfNATPMPSocket");
			close(snatpmp);
			return -1;
		}
	}
	return snatpmp;
}

static int OpenAndConfNATPMPSockets(int * sockets)
{
	int i;
	struct lan_addr_s * lan_addr;
	for(i = 0, lan_addr = bb_common_bss.lan_addrs.lh_first; lan_addr != NULL;
	    lan_addr = lan_addr->list.le_next)
	{
		sockets[i] = OpenAndConfNATPMPSocket(lan_addr->addr.s_addr);
		if(sockets[i] < 0)
			goto error;
		i++;
	}
	return 0;
error:
	while(--i >= 0)
	{
		close(sockets[i]);
		sockets[i] = -1;
	}
	return -1;
}

static void FillPublicAddressResponse(unsigned char * resp, in_addr_t senderaddr)
{
	char tmp[16];
	UNUSED(senderaddr);

	if(bb_common_bss.use_ext_ip_addr) {
		inet_pton(AF_INET, bb_common_bss.use_ext_ip_addr, resp+8);
	} else {
		if(!bb_common_bss.ext_if_name || bb_common_bss.ext_if_name[0]=='\0') {
			resp[3] = 3;    /* Network Failure (e.g. NAT box itself
					 * has not obtained a DHCP lease) */
		} else if(getifaddr(bb_common_bss.ext_if_name, tmp, INET_ADDRSTRLEN, NULL, NULL) < 0) {
			bb_error_msg("Failed to get IP for interface %s", bb_common_bss.ext_if_name);
			resp[3] = 3;    /* Network Failure (e.g. NAT box itself
					 * has not obtained a DHCP lease) */
		} else {
			inet_pton(AF_INET, tmp, resp+8); /* ok */
		}
	}
}


/** read the request from the socket, process it and then send the
 * response back.
 */
static void ProcessIncomingNATPMPPacket(int s, unsigned char *msg_buff, int len,
		struct sockaddr_in *senderaddr)
{
	unsigned char *req=msg_buff;    /* request udp packet */
	unsigned char resp[32]; /* response udp packet */
	int resplen;
	int n = len;
	char senderaddrstr[16];

	if(!inet_ntop(AF_INET, &senderaddr->sin_addr, senderaddrstr, sizeof(senderaddrstr)))
		bb_perror_msg("inet_ntop(natpmp)");

	DEBUG_LOG(syslog(LOG_INFO, "NAT-PMP request received from %s:%hu %dbytes",
	       senderaddrstr, ntohs(senderaddr->sin_port), n);)

	if(n<2 || ((((req[1]-1)&~1)==0) && n<12)) {
		xlog_warn("discarding NAT-PMP request (too short) %dBytes", n);
		return;
	}
	if(req[1] & 128) {
		/* discarding NAT-PMP responses silently */
		return;
	}
	memset(resp, 0, sizeof(resp));
	resplen = 8;
	resp[1] = 128 + req[1]; /* response OPCODE is request OPCODE + 128 */
	/* setting response TIME STAMP :
	 * time elapsed since its port mapping table was initialized on
	 * startup or reset for any other reason */
	WRITENU32(resp+4, time(NULL) - bb_common_bss.startup_time);
	if(req[0] > 0) {
		/* invalid version */
		xlog_warn("unsupported NAT-PMP version : %u",
		       (unsigned)req[0]);
		resp[3] = 1;    /* unsupported version */
	} else switch(req[1]) {
	case 0: /* Public address request */
		DEBUG_LOG(syslog(LOG_INFO, "NAT-PMP public address request");)
		FillPublicAddressResponse(resp, senderaddr->sin_addr.s_addr);
		resplen = 12;
		break;
	case 1: /* UDP port mapping request */
	case 2: /* TCP port mapping request */
		{
			unsigned short iport;   /* private port */
			unsigned short eport;   /* public port */
			uint32_t lifetime;              /* lifetime=0 => remove port mapping */
			int r;
			int proto;
			char iaddr_old[16];
			unsigned short iport_old;
			unsigned int timestamp;

			iport = READNU16(req+4);
			eport = READNU16(req+6);
			lifetime = READNU32(req+8);
			proto = (req[1]==1)?IPPROTO_UDP:IPPROTO_TCP;
			DEBUG_LOG(syslog(LOG_INFO, "NAT-PMP port mapping request : "
					 "%hu->%s:%hu %s lifetime=%us",
					 eport, senderaddrstr, iport,
					 (req[1]==1)?"udp":"tcp", lifetime);)
			/* TODO: accept port mapping if iport ok but eport not ok
			 * (and set eport correctly) */
			if(lifetime == 0) {
				/* remove the mapping */
				/* RFC6886 :
				 * A client MAY also send an explicit packet to request deletion of a
				 * mapping that is no longer needed. A client requests explicit
				 * deletion of a mapping by sending a message to the NAT gateway
				 * requesting the mapping, with the Requested Lifetime in Seconds set to
				 * zero. The Suggested External Port MUST be set to zero by the client
				 * on sending, and MUST be ignored by the gateway on reception. */
				int index = 0;
				unsigned short eport2, iport2;
				char iaddr2[16];
				int proto2;
				char desc[64];
				eport = 0; /* to indicate correct removing of port mapping */
				while(get_redirect_rule_by_index(index, 0,
					  &eport2, iaddr2, sizeof(iaddr2),
						  &iport2, &proto2,
						  desc, sizeof(desc),
					  0, 0, &timestamp, 0, 0) >= 0) {
					DEBUG_LOG(syslog(LOG_DEBUG, "%d %d %hu->'%s':%hu '%s'",
					       index, proto2, eport2, iaddr2, iport2, desc);)
					if(0 == strcmp(iaddr2, senderaddrstr)
					  && 0 == memcmp(desc, "NAT-PMP", 7)) {
						/* (iport == 0) => remove all the mappings for this client */
						if((iport == 0) || ((iport == iport2) && (proto == proto2))) {
							r = _upnp_delete_redir(eport2, proto2);
							if(r < 0) {
								bb_error_msg("Failed to remove NAT-PMP mapping eport %hu, protocol %s",
								       eport2, (proto2==IPPROTO_TCP)?"TCP":"UDP");
								resp[3] = 2;    /* Not Authorized/Refused */
								break;
							} else {
								DEBUG_LOG(syslog(LOG_INFO, "NAT-PMP %s port %hu mapping removed",
								       proto2==IPPROTO_TCP?"TCP":"UDP", eport2);)
								index--;
							}
						}
					}
					index++;
				}
			} else if(iport==0) {
				resp[3] = 2;    /* Not Authorized/Refused */
			} else { /* iport > 0 && lifetime > 0 */
				unsigned short eport_first = 0;
				int any_eport_allowed = 0;
				char desc[64];
				if(eport==0)    /* if no suggested external port, use same a internal port */
					eport = iport;
				while(resp[3] == 0) {
					if(eport_first == 0) { /* first time in loop */
						eport_first = eport;
					} else if(eport == eport_first) { /* no eport available */
						if(any_eport_allowed == 0) { /* all eports rejected by permissions */
							bb_error_msg("No allowed eport for NAT-PMP %hu %s->%s:%hu",
							       eport, (proto==IPPROTO_TCP)?"tcp":"udp", senderaddrstr, iport);
							resp[3] = 2;    /* Not Authorized/Refused */
						} else { /* at least one eport allowed (but none available) */
							bb_error_msg("Failed to find available eport for NAT-PMP %hu %s->%s:%hu",
							       eport, (proto==IPPROTO_TCP)?"tcp":"udp", senderaddrstr, iport);
							resp[3] = 4;    /* Out of resources */
						}
						break;
					}
					if(!check_upnp_rule_against_permissions(eport, senderaddr->sin_addr, iport)) {
						eport++;
						if(eport == 0) eport++; /* skip port zero */
						continue;
					}
					any_eport_allowed = 1;  /* at lease one eport is allowed */
					r = get_redirect_rule(eport, proto, iaddr_old, sizeof(iaddr_old),
							      &iport_old, 0, 0, 0, 0,
							      &timestamp, 0, 0);
					if(r==0) {
						if(strcmp(senderaddrstr, iaddr_old)==0
						    && iport==iport_old) {
							/* redirection already existing */
							DEBUG_LOG(syslog(LOG_INFO, "port %hu %s already redirected to %s:%hu, replacing",
							       eport, (proto==IPPROTO_TCP)?"tcp":"udp", iaddr_old, iport_old);)
							/* remove and then add again */
							if(_upnp_delete_redir(eport, proto) < 0) {
								bb_error_msg("failed to remove port mapping");
								break;
							}
						} else {
							eport++;
							if(eport == 0) eport++; /* skip port zero */
							continue;
						}
					}
					/* do the redirection */
					timestamp = time(NULL) + lifetime;
					snprintf(desc, sizeof(desc), "NAT-PMP %hu %s",
						 eport, (proto==IPPROTO_TCP)?"tcp":"udp");
					/* TODO : check return code */
					if(upnp_redirect_internal(NULL, eport, senderaddrstr,
								  iport, proto, desc,
								  timestamp) < 0) {
						bb_error_msg("Failed to add NAT-PMP %hu %s->%s:%hu '%s'",
						       eport, (proto==IPPROTO_TCP)?"tcp":"udp", senderaddrstr, iport, desc);
						resp[3] = 3;  /* Failure */
					}
					break;
				}
			}
			WRITENU16(resp+8, iport);       /* private port */
			WRITENU16(resp+10, eport);      /* public port */
			WRITENU32(resp+12, lifetime);   /* Port Mapping lifetime */
		}
		resplen = 16;
		break;
	default:
		resp[3] = 5;    /* Unsupported OPCODE */
	}
	n = sendto_or_schedule(s, resp, resplen, 0,
		   (struct sockaddr *)senderaddr, sizeof(*senderaddr));
	if(n<0) {
		bb_perror_msg("sendto(natpmp)");
	} else if(n<resplen) {
		bb_error_msg("sendto(natpmp): sent only %d bytes out of %d", n, resplen);
	}
}

/* should be called when the public IP address changed */
static void SendNATPMPPublicAddressChangeNotification(int * sockets, int n_sockets)
{
	struct sockaddr_in sockname;
	unsigned char notif[12];
	int j, n;

	notif[0] = 0;   /* vers */
	notif[1] = 128; /* OP code */
	notif[2] = 0;   /* result code */
	notif[3] = 0;   /* result code */
	/* seconds since "start of epoch" :
	 * time elapsed since the port mapping table was initialized on
	 * startup or reset for any other reason */
	WRITENU32(notif+4, time(NULL) - bb_common_bss.startup_time);
	FillPublicAddressResponse(notif, 0);
	if(notif[3])
	{
		xlog_warn("%s: cannot get public IP address, stopping",
		       "SendNATPMPPublicAddressChangeNotification");
		return;
	}
	memset(&sockname, 0, sizeof(struct sockaddr_in));
	sockname.sin_family = AF_INET;
	sockname.sin_addr.s_addr = inet_addr(NATPMP_NOTIF_ADDR);

	for(j=0; j<n_sockets; j++) {
		if(sockets[j] < 0)
			continue;
		/* Port to use in 2006 version of the NAT-PMP specification */
		sockname.sin_port = htons(NATPMP_PORT);
		n = sendto_or_schedule(sockets[j], notif, 12, 0,
			   (struct sockaddr *)&sockname, sizeof(struct sockaddr_in));
		if(n < 0) {
			bb_perror_msg("%s: sendto(s_udp=%d)",
			       "SendNATPMPPublicAddressChangeNotification", sockets[j]);
			return;
		}
		/* Port to use in 2008 version of the NAT-PMP specification */
		sockname.sin_port = htons(NATPMP_NOTIF_PORT);
		n = sendto_or_schedule(sockets[j], notif, 12, 0,
			   (struct sockaddr *)&sockname, sizeof(struct sockaddr_in));
		if(n < 0) {
			bb_perror_msg("%s: sendto(s_udp=%d)",
			       "SendNATPMPPublicAddressChangeNotification", sockets[j]);
			return;
		}
	}
}

#endif /* ENABLE_FEATURE_MINIUPNPD_NATPMP */


struct STR_SAVE_OPT {
	const char *str_opt;
	const char **ptr_value;
};
/* init phase :
 * 1) read configuration file
 * 2) read command line arguments
 * 3) daemonize
 * 4) open syslog
 * 5) check and write pid file
 * 6) set startup time stamp
 * 7) compute presentation URL
 * 8) set signal handlers
 * 9) init random generator (srandom())
 * 10) init redirection engine
 * 11) reload mapping from leasefile */
static void init(struct runtime_vars *v)
{
	int i, opt_bits;
#ifdef DEBUG
	int debug_flag = 0;
#endif
	/*const char * logfilename = 0;*/
	const char *presurl = NULL;
	struct lan_addr_s *lan_addr;
	struct lan_addr_s *lan_addr2;
	char *s_upnp_bootid = NULL;
	const char *s_uuidvalue_igd = NULL;
	const char *s_serialnumber = NULL;
	const char *s_modelnumber = NULL;
	char *s_downstream_bitrate = NULL;
	char *s_upstream_bitrate = NULL;
	llist_t *lan_addr_llist_t = NULL;
#ifndef DISABLE_CONFIG_FILE
	const char *optionsfile = NULL;
	config_read_t *crt;
	llist_t *lan_addr_llist_cfg = NULL;
	llist_t *upnppermlist_list_cfg = NULL;
	struct STR_SAVE_OPT str_save_opt[] = {
		{ "ext_ifname", &bb_common_bss.ext_if_name },
		{ "ext_ip", &bb_common_bss.use_ext_ip_addr },
		{ "presentation_url", &presurl },
#ifdef USE_NETFILTER
		{ "upnp_forward_chain", &bb_common_bss.miniupnpd_forward_chain },
		{ "upnp_nat_chain", &bb_common_bss.miniupnpd_nat_chain },
		{ "upnp_nat_postrouting_chain", &bb_common_bss.miniupnpd_nat_postrouting_chain },
#endif
		{ "uuid", &s_uuidvalue_igd },
		{ "serial", &s_serialnumber },
		{ "model_number", &s_modelnumber },
		{ "minissdpdsocket", &bb_common_bss.minissdpdsocketpath },
		{ NULL, NULL }
		};
#endif /* DISABLE_CONFIG_FILE */

	/* set initial values */
	SETFLAG(ENABLEUPNPMASK);        /* UPnP is enabled by default */
#if ENABLE_FEATURE_IPV6
	bb_common_bss.ipv6_bind_addr = in6addr_any;
#endif

	LIST_INIT(&bb_common_bss.lan_addrs);
	v->port = -1;

	/* set data */
#ifdef ENABLE_PCP
	bb_common_bss.max_lifetime = 86400;
	bb_common_bss.min_lifetime = 120;
#endif

#ifdef USE_NETFILTER
#ifdef ENABLE_UPNPPINHOLE
	bb_common_bss.miniupnpd_v6_filter_chain = "MINIUPNPD";
#endif /* ENABLE_UPNPPINHOLE */
#endif
	bb_common_bss.modelnumber[0] = '1';
	bb_common_bss.modelnumber[1] = '\0';
	bb_common_bss.pidfilename = "/var/run/miniupnpd.pid";
	strcpy(bb_common_bss.serialnumber, "00000000");
	bb_common_bss.upnp_bootid = 1;
	bb_common_bss.upnp_configid = 1337;
	strcpy(bb_common_bss.uuidvalue_igd, "uuid:00000000-0000-0000-0000-000000000000");
	strcpy(bb_common_bss.uuidvalue_wan, "uuid:00000000-0000-0000-0000-000000000000");
	strcpy(bb_common_bss.uuidvalue_wcd, "uuid:00000000-0000-0000-0000-000000000000");

#define DEFAULT_NOTIFY_INTERVAL 30
#define DEFAULT_CLEAN_RULESET_INTERVAL 0
#ifndef DISABLE_CONFIG_FILE
	v->notify_interval = -1;        /* seconds between SSDP announces */
	v->clean_ruleset_interval = -1;  /* interval between ruleset check. 0=disabled */
#else
	v->notify_interval = DEFAULT_NOTIFY_INTERVAL;
	v->clean_ruleset_interval = DEFAULT_CLEAN_RULESET_INTERVAL;
#endif

	v->clean_ruleset_threshold = 20;


	opt_bits = bb_xgetopt_ulflags("dUS" USE_FEATURE_MINIUPNPD_NATPMP("N")
#ifndef DISABLE_CONFIG_FILE
			"f:"
#endif
			"hb:o:"
			"t:r:"
			"u:s:m:"
			"i:p:"
			"P:w:"
			"W:B:"
			"a:A:\0:-trp:?0:a::A::h-",
#ifndef DISABLE_CONFIG_FILE
			&optionsfile,
#endif
			&s_upnp_bootid, &bb_common_bss.use_ext_ip_addr,
			&v->notify_interval, &v->clean_ruleset_interval,
			&s_uuidvalue_igd, &s_serialnumber, &s_modelnumber,
			&bb_common_bss.ext_if_name, &v->port,
			&bb_common_bss.pidfilename, &presurl,
			&s_downstream_bitrate, &s_upstream_bitrate,
			&lan_addr_llist_t, &bb_common_bss.upnppermlist_list_t);

#ifdef DEBUG
	debug_flag = opt_bits & 1;
#endif
	i = LOG_PID|LOG_CONS;
#ifdef DEBUG
	if(debug_flag)
		i |= LOG_PERROR;   /* also log on stderr */
#endif
	bb_openlog(i, LOG_MINIUPNPD);

#ifndef DISABLE_CONFIG_FILE
	/* read options file first since
	 * command line arguments have final say */
	if(optionsfile == NULL)
		optionsfile = DEFAULT_CONFIG;
	crt = config_read_prepare(optionsfile, 'w');
	while( (crt = config_read(crt, "#~\0\0=")) != NULL) {
		if(!strcmp(crt->tokens[0], "allow") || !strcmp(crt->tokens[0], "deny")) {
			if(crt->ntokens == 4) {
				/* restore spaces */
				struct upnpperm *perm = read_permission_line(crt->tokens);
				if(perm == NULL) {
					bb_error_msg("%s:%d: permission rule parsing error", optionsfile, crt->num_line);
				} else {
					upnppermlist_list_cfg = llist_add_to_end(upnppermlist_list_cfg, perm);
				}
			} else
				bb_error_msg("%s:%d: permission rule parsing error", optionsfile, crt->num_line);
			continue;
		}
		if(crt->ntokens != 2 || (crt->flags[0] & ~CR_FLAGS_FOUND_SPACE_DELIM) != CR_FLAGS_FOUND_SPEC_DELIM) {
			bb_error_msg("%s:%d: parsing error", optionsfile, crt->num_line);
			continue;
		}

		if(!strcmp(crt->tokens[0], "listening_ip")) {
			lan_addr_llist_cfg = llist_add_to_end(lan_addr_llist_cfg, xstrdup(crt->tokens[1]));
#if ENABLE_FEATURE_IPV6
		} else if(!strcmp(crt->tokens[0], "ipv6_listening_ip")) {
			if (inet_pton(AF_INET6, crt->tokens[1], &bb_common_bss.ipv6_bind_addr) < 1)
				bb_error_msg("%s:%d: can't parse \"%s\" as valid IPv6 listening address", optionsfile, crt->num_line, crt->tokens[1]);
#endif
		} else if(!strcmp(crt->tokens[0], "port")) {
			if(v->port == -1) /* use from command option priority */
				v->port = atoi(crt->tokens[1]);
		} else if(!strcmp(crt->tokens[0], "bitrate_up")) {
			if(s_upstream_bitrate == NULL) /* use from command option priority */
				bb_common_bss.upstream_bitrate = strtoul(crt->tokens[1], 0, 0);
		} else if(!strcmp(crt->tokens[0], "bitrate_down")) {
			if(s_downstream_bitrate == NULL) /* use from command option priority */
				bb_common_bss.downstream_bitrate = strtoul(crt->tokens[1], 0, 0);
		} else if(!strcmp(crt->tokens[0], "notify_interval")) {
			if(v->notify_interval == -1)    /* use from command option priority */
				v->notify_interval = atoi(crt->tokens[1]);
		} else if(!strcmp(crt->tokens[0], "system_uptime")) {
			if(strcmp(crt->tokens[1], "yes") == 0)
				SETFLAG(SYSUPTIMEMASK); /*sysuptime = 1;*/
		} else if(!strcmp(crt->tokens[0], "clean_ruleset_threshold")) {
			v->clean_ruleset_threshold = atoi(crt->tokens[1]);
		} else if(!strcmp(crt->tokens[0], "clean_ruleset_interval")) {
			if(v->clean_ruleset_interval == -1) /* use from command option priority */
				v->clean_ruleset_interval = atoi(crt->tokens[1]);
#if ENABLE_FEATURE_MINIUPNPD_NATPMP
		} else if(!strcmp(crt->tokens[0], "enable_natpmp")) {
			if(strcmp(crt->tokens[1], "yes") == 0)
				opt_bits |= 8;      /*enablenatpmp = 1;*/
			else
				if(atoi(crt->tokens[1]))
					opt_bits |= 8;
#endif
#ifdef ENABLE_PCP
		} else if(!strcmp(crt->tokens[0], "min_lifetime")) {
			bb_common_bss.min_lifetime = atoi(crt->tokens[1]);
			if (bb_common_bss.min_lifetime > 120 )
				bb_common_bss.min_lifetime = 120;
		} else if(!strcmp(crt->tokens[0], "max_lifetime")) {
			bb_common_bss.max_lifetime = atoi(crt->tokens[1]);
			if (bb_common_bss.max_lifetime > 86400 )
				bb_common_bss.max_lifetime = 86400;
		} else if(!strcmp(crt->tokens[0], "pcp_allow_thirdparty")) {
			if(strcmp(crt->tokens[1], "yes") == 0)
				SETFLAG(PCP_ALLOWTHIRDPARTYMASK);
#endif  /* ENABLE_PCP */
		} else if(!strcmp(crt->tokens[0], "enable_upnp")) {
			if(strcmp(crt->tokens[1], "yes") != 0)
				CLEARFLAG(ENABLEUPNPMASK);
		} else if(!strcmp(crt->tokens[0], "secure_mode")) {
			if(strcmp(crt->tokens[1], "yes") == 0)
				SETFLAG(SECUREMODEMASK);
		} else {
			struct STR_SAVE_OPT *p_save_opt = find_string_from_struct(str_save_opt, sizeof(struct STR_SAVE_OPT), crt->tokens[0]);
			if(p_save_opt->str_opt == NULL)
				bb_error_msg("%s:%d: unknown option \"%s\"", optionsfile, crt->num_line, crt->tokens[0]);
			else if(*p_save_opt->ptr_value != NULL)
				*p_save_opt->ptr_value = xstrdup(crt->tokens[1]);
		}
	}
#ifdef ENABLE_PCP
	/* if lifetimes are inverse */
	if (bb_common_bss.min_lifetime >= bb_common_bss.max_lifetime) {
		bb_error_msg_and_die("minimum lifetime (%lu) is greater than or equal to maximum lifetime (%lu).\n"
				"Check your configuration file %s",
				bb_common_bss.min_lifetime, bb_common_bss.max_lifetime, optionsfile);
	}
#endif  /* ENABLE_PCP */

	/* set default */
	if(v->notify_interval == -1)
		v->notify_interval = DEFAULT_NOTIFY_INTERVAL;
	if(v->clean_ruleset_interval == -1)
		v->clean_ruleset_interval = DEFAULT_CLEAN_RULESET_INTERVAL;
#ifdef USE_NETFILTER
	if(bb_common_bss.miniupnpd_forward_chain == NULL)
		bb_common_bss.miniupnpd_forward_chain = "MINIUPNPD";
	if(bb_common_bss.miniupnpd_nat_chain == NULL)
		bb_common_bss.miniupnpd_nat_chain = "MINIUPNPD";
	if(bb_common_bss.miniupnpd_nat_postrouting_chain == NULL)
		bb_common_bss.miniupnpd_nat_postrouting_chain = "MINIUPNPD-POSTROUTING";
#endif
	if(bb_common_bss.minissdpdsocketpath == NULL)
		bb_common_bss.minissdpdsocketpath = "/var/run/minissdpd.sock";

	/* set lan_addr lists elements before command line specified */
	while(lan_addr_llist_t != NULL) {
		lan_addr_llist_cfg = llist_add_to_end(lan_addr_llist_cfg, lan_addr_llist_t->data);
		lan_addr_llist_t = llist_free_one(lan_addr_llist_t);
	}
	lan_addr_llist_t = lan_addr_llist_cfg;
#endif /* DISABLE_CONFIG_FILE */

	if((opt_bits & 2))
		SETFLAG(SYSUPTIMEMASK); /*sysuptime = 1;*/
	if((opt_bits & 4))
		SETFLAG(SECUREMODEMASK);
#if ENABLE_FEATURE_MINIUPNPD_NATPMP
	if((opt_bits & 8))
		SETFLAG(ENABLENATPMPMASK); /*enablenatpmp = 1;*/
#endif
	if(s_upnp_bootid != NULL)
		bb_common_bss.upnp_bootid = (unsigned int)strtoul(s_upnp_bootid, NULL, 10);
	if(s_uuidvalue_igd != NULL) {
		safe_strncpy(bb_common_bss.uuidvalue_igd + 5, s_uuidvalue_igd, sizeof(bb_common_bss.uuidvalue_igd) - 5);
		complete_uuidvalues();
	}
	if(s_serialnumber != NULL)
		safe_strncpy(bb_common_bss.serialnumber, s_serialnumber, SERIALNUMBER_MAX_LEN);
	if(s_modelnumber != NULL)
		safe_strncpy(bb_common_bss.modelnumber, s_modelnumber, MODELNUMBER_MAX_LEN);
	if(s_downstream_bitrate != NULL)
		bb_common_bss.downstream_bitrate = strtoul(s_downstream_bitrate, 0, 0);
	if(s_upstream_bitrate != NULL)
		bb_common_bss.upstream_bitrate = strtoul(s_upstream_bitrate, 0, 0);
	while(lan_addr_llist_t != NULL) {
		lan_addr = xmalloc(sizeof(struct lan_addr_s));
		if(parselanaddr(lan_addr, lan_addr_llist_t->data) != 0) {
			bb_error_msg("can't parse \"%s\" as a valid "
#if !ENABLE_FEATURE_IPV6
						"LAN address or "
#endif
						"interface name", (char *)lan_addr_llist_t->data);
		} else {
			/* check if we already have this address */
			for(lan_addr2 = bb_common_bss.lan_addrs.lh_first; lan_addr2 != NULL; lan_addr2 = lan_addr2->list.le_next) {
				if (!strcmp(lan_addr2->str, lan_addr->str))
					break;
			}
			if (lan_addr2 == NULL) {
				LIST_INSERT_HEAD(&bb_common_bss.lan_addrs, lan_addr, list);
				lan_addr = NULL;
			}
		}
		free(lan_addr);
		lan_addr_llist_t = llist_free_one(lan_addr_llist_t);
	}
	if(bb_common_bss.upnppermlist_list_t) {
		llist_t *pp;

		for(pp = bb_common_bss.upnppermlist_list_t; pp; ) {
			config_read_t *ct = config_read(config_read_prepare(pp->data, 'S'), "\0\1\2\3\4\0\4");
			struct upnpperm *perm = NULL;

			if(ct != NULL && ct->ntokens == 4) {
				perm = read_permission_line(ct->tokens);
				config_close(ct);
			}

			if(perm == NULL) {
				bb_error_msg("-A \"%s\" permission rule parsing error", (char *)pp->data);
			} else {
#ifndef DISABLE_CONFIG_FILE
				/* set command line specified after config permission lines */
				upnppermlist_list_cfg = llist_add_to_end(upnppermlist_list_cfg, perm);
				pp = llist_free_one(pp);
#else
				pp->data = perm;
				pp = pp->link;
#endif
			}
		}
	}
#ifndef DISABLE_CONFIG_FILE
	bb_common_bss.upnppermlist_list_t = upnppermlist_list_cfg;
#endif
	if(!bb_common_bss.ext_if_name || !bb_common_bss.lan_addrs.lh_first)
		bb_error_msg_and_die("Need set interface(s)");

	if(checkforrunning(bb_common_bss.pidfilename) < 0)
		bb_error_msg_and_die("is already running. EXITING");

#ifdef DEBUG
	if(!debug_flag) {
		/* speed things up and ignore LOG_INFO and LOG_DEBUG */
		setlogmask(LOG_UPTO(LOG_NOTICE));
#endif
		xdaemon(XDAEMON_PF_ARG, bb_common_bss.pidfilename);
#ifdef DEBUG
	}
#endif
	set_startup_time(GETFLAG(SYSUPTIMEMASK));

	/* presentation url */
	if(presurl) {
		safe_strncpy(bb_common_bss.presentationurl, presurl, PRESENTATIONURL_MAX_LEN);
	} else {
		snprintf(bb_common_bss.presentationurl, PRESENTATIONURL_MAX_LEN,
			 "http://%s/", bb_common_bss.lan_addrs.lh_first->str);
			 /*"http://%s:%d/", bb_common_bss.lan_addrs.lh_first->str, 80);*/
	}

	/* set signal handler */
	bb_signal((1<<SIGTERM)|(1<<SIGINT), sigterm);
	bb_signal((1<<SIGPIPE), SIG_IGN);
	bb_signal((1<<SIGUSR1), sigusr1);

	/* initialize random number generator */
	srandom((unsigned int)time(NULL));

	/* initialize redirection engine (and pinholes) */
	init_redirect();
#ifdef ENABLE_UPNPPINHOLE
#ifdef USE_NETFILTER
	init_iptpinhole();
#endif
#endif
}

/* Sends the description generated by the parameter */
static void sendXMLdesc(struct upnphttp * h, char * (f)(int *))
{
	char *desc;
	int len;
	desc = f(&len);
	if(!desc) {
		static const char error500[] = "<HTML><HEAD><TITLE>Error 500</TITLE>"
		   "</HEAD><BODY>Internal Server Error</BODY></HTML>\r\n";

		bb_error_msg("Failed to generate XML description");
		h->respflags = FLAG_HTML;
		BuildResp2_upnphttp(h, 500, "Internal Server Error",
				    error500, sizeof(error500)-1);
	} else {
		BuildResp_upnphttp(h, desc, len);
	}
	SendRespAndClose_upnphttp(h);
	free(desc);
}

/* create a new subscriber */
static struct subscriber *newSubscriber(const char *eventurl, const char *callback, int callbacklen)
{
	struct subscriber * tmp;
	if(!eventurl || !callback || !callbacklen)
		return NULL;
	tmp = calloc(1, sizeof(struct subscriber)+callbacklen+1);
	if(!tmp)
		return NULL;
	if(strcmp(eventurl, WANCFG_EVENTURL)==0)
		tmp->service = EWanCFG;
	else if(strcmp(eventurl, WANIPC_EVENTURL)==0)
		tmp->service = EWanIPC;
#ifdef ENABLE_L3F_SERVICE
	else if(strcmp(eventurl, L3F_EVENTURL)==0)
		tmp->service = EL3F;
#endif
	else {
		free(tmp);
		return NULL;
	}
	memcpy(tmp->callback, callback, callbacklen);
	tmp->callback[callbacklen] = '\0';
	{
		uuid_t uuid;
		uuid_generate(uuid);
		memcpy(tmp->uuid, "uuid:", 5);
		uuid_unparse(uuid, tmp->uuid + 5);
	}
	return tmp;
}

/* creates a new subscriber and adds it to the subscriber list
 * also initiate 1st notify
 * TODO : add a check on the number of subscriber in order to
 * prevent memory overflow... */
static const char *
upnpevents_addSubscriber(const char * eventurl,
			 const char * callback, int callbacklen,
			 int timeout)
{
	struct subscriber * tmp;
	/* static char uuid[42];*/
	/* "uuid:00000000-0000-0000-0000-000000000000"; 5+36+1=42bytes */
	DEBUG_LOG(syslog(LOG_DEBUG, "addSubscriber(%s, %.*s, %d)", eventurl, callbacklen, callback, timeout);)
	/* safe_strncpy(uuid, uuidvalue, sizeof(uuid)); */
	tmp = newSubscriber(eventurl, callback, callbacklen);
	if(!tmp)
		return NULL;
	if(timeout)
		tmp->timeout = time(NULL) + timeout;
	LIST_INSERT_HEAD(&subscriberlist, tmp, entries);
	upnp_event_create_notify(tmp);
	return tmp->uuid;
}

static void
ProcessHTTPSubscribe_upnphttp(struct upnphttp * h, const char * path)
{
	const char * sid;
	DEBUG_LOG(syslog(LOG_DEBUG, "ProcessHTTPSubscribe %s", path);
	syslog(LOG_DEBUG, "Callback '%.*s' Timeout=%d",
	       h->req_CallbackLen, h->req_buf + h->req_CallbackOff,
	       h->req_Timeout);
	syslog(LOG_DEBUG, "SID '%.*s'", h->req_SIDLen, h->req_buf + h->req_SIDOff);)
	if(h->req_Timeout == 0) {
		/* Second-infinite is forbidden with UDA v1.1 and later :
		 * (UDA 1.1 : 4.1.1 Subscription)
		 * UPnP 1.1 control points MUST NOT subscribe using keyword infinite,
		 * UPnP 1.1 devices MUST NOT set actual subscription durations to
		 * "infinite". The presence of infinite in a request MUST be silently
		 * ignored by a UPnP 1.1 device (the presence of infinite is handled
		 * by the device as if the TIMEOUT header field in a request was not
		 * present) . The keyword infinite MUST NOT be returned by a UPnP 1.1
		 * device. */
		h->req_Timeout = 1800;  /* default to 30 minutes */
	}
	if((h->req_CallbackOff <= 0) && (h->req_SIDOff <= 0)) {
		/* Missing or invalid CALLBACK : 412 Precondition Failed.
		 * If CALLBACK header is missing or does not contain a valid HTTP URL,
		 * the publisher must respond with HTTP error 412 Precondition Failed*/
		BuildResp2_upnphttp(h, 412, "Precondition Failed", 0, 0);
		SendRespAndClose_upnphttp(h);
	} else {
	/* - add to the subscriber list
	 * - respond HTTP/x.x 200 OK
	 * - Send the initial event message */
	/* Server:, SID:; Timeout: Second-(xx|infinite) */
	/* Check that the callback URL is on the same IP as
	 * the request, and not on the internet, nor on ourself (DOS attack ?) */
		if(h->req_CallbackOff > 0) {
			if(checkCallbackURL(h)) {
				sid = upnpevents_addSubscriber(path, h->req_buf + h->req_CallbackOff,
							       h->req_CallbackLen, h->req_Timeout);
				h->respflags = FLAG_TIMEOUT;
				if(sid) {
					DEBUG_LOG(syslog(LOG_DEBUG, "generated sid=%s", sid);)
					h->respflags |= FLAG_SID;
					h->res_SID = sid;
				}
				BuildResp_upnphttp(h, 0, 0);
			} else {
				xlog_warn("Invalid Callback in SUBSCRIBE %.*s",
				       h->req_CallbackLen, h->req_buf + h->req_CallbackOff);
				BuildResp2_upnphttp(h, 412, "Precondition Failed", 0, 0);
			}
		} else {
			/* subscription renew */
			/* Invalid SID
412 Precondition Failed. If a SID does not correspond to a known,
un-expired subscription, the publisher must respond
with HTTP error 412 Precondition Failed. */
				sid = upnpevents_renewSubscription(h->req_buf + h->req_SIDOff,
								   h->req_SIDLen, h->req_Timeout);
				if(!sid) {
					BuildResp2_upnphttp(h, 412, "Precondition Failed", 0, 0);
				} else {
					h->respflags = FLAG_TIMEOUT | FLAG_SID;
					h->res_SID = sid;
					BuildResp_upnphttp(h, 0, 0);
				}
		}
		SendRespAndClose_upnphttp(h);
	}
}

static void
Send405(struct upnphttp * h)
{
	static const char body405[] =
		"<HTML><HEAD><TITLE>405 Method Not Allowed</TITLE></HEAD>"
		"<BODY><H1>Method Not Allowed</H1>The HTTP Method "
		"is not allowed on this resource.</BODY></HTML>\r\n";

	h->respflags |= FLAG_HTML;
	BuildResp2_upnphttp(h, 405, "Method Not Allowed",
			    body405, sizeof(body405) - 1);
	SendRespAndClose_upnphttp(h);
}

/* very minimalistic 501 error message */
static void
Send501(struct upnphttp * h)
{
	static const char body501[] =
		"<HTML><HEAD><TITLE>501 Not Implemented</TITLE></HEAD>"
		"<BODY><H1>Not Implemented</H1>The HTTP Method "
		"is not implemented by this server.</BODY></HTML>\r\n";

	h->respflags = FLAG_HTML;
	BuildResp2_upnphttp(h, 501, "Not Implemented",
			    body501, sizeof(body501) - 1);
	SendRespAndClose_upnphttp(h);
}

/* Windows XP as client send the following requests :
 * GetConnectionTypeInfo
 * GetNATRSIPStatus
 * ? GetTotalBytesSent - WANCommonInterfaceConfig
 * ? GetTotalBytesReceived - idem
 * ? GetTotalPacketsSent - idem
 * ? GetTotalPacketsReceived - idem
 * GetCommonLinkProperties - idem
 * GetStatusInfo - WANIPConnection
 * GetExternalIPAddress
 * QueryStateVariable / ConnectionStatus!
 */
static const struct
{
	const char * methodName;
	void (*methodImpl)(struct upnphttp *, const char *, const char *);
}
soapMethods[] =
{
	/* WANCommonInterfaceConfig */
	{ "QueryStateVariable", QueryStateVariable},
	{ "GetTotalBytesSent", GetTotalBytesSent},
	{ "GetTotalBytesReceived", GetTotalBytesReceived},
	{ "GetTotalPacketsSent", GetTotalPacketsSent},
	{ "GetTotalPacketsReceived", GetTotalPacketsReceived},
	{ "GetCommonLinkProperties", GetCommonLinkProperties},
	{ "GetStatusInfo", GetStatusInfo},
	/* WANIPConnection */
	{ "GetConnectionTypeInfo", GetConnectionTypeInfo },
	{ "GetNATRSIPStatus", GetNATRSIPStatus},
	{ "GetExternalIPAddress", GetExternalIPAddress},
	{ "AddPortMapping", AddPortMapping},
	{ "DeletePortMapping", DeletePortMapping},
	{ "GetGenericPortMappingEntry", GetGenericPortMappingEntry},
	{ "GetSpecificPortMappingEntry", GetSpecificPortMappingEntry},
	/* Required in WANIPConnection:2 */
	{ "SetConnectionType", SetConnectionType},
	{ "RequestConnection", RequestConnection},
	{ "ForceTermination", ForceTermination},
	{ "AddAnyPortMapping", AddAnyPortMapping},
	{ "DeletePortMappingRange", DeletePortMappingRange},
	{ "GetListOfPortMappings", GetListOfPortMappings},
#ifdef ENABLE_L3F_SERVICE
	/* Layer3Forwarding */
	{ "SetDefaultConnectionService", SetDefaultConnectionService},
	{ "GetDefaultConnectionService", GetDefaultConnectionService},
#endif
	{ 0, 0 }
};

/* this method executes the requested Soap Action */
static void ExecuteSoapAction(struct upnphttp * h, const char * action, int n)
{
	char * p;
	char * p2;
	int i, len, methodlen;
	char namespace[256];

	/* SoapAction example :
	 * urn:schemas-upnp-org:service:WANIPConnection:1#GetStatusInfo */
	p = strchr(action, '#');
	if(p && (p - action) < n) {
		for(i = 0; i < ((int)sizeof(namespace) - 1) && (action + i) < p; i++)
			namespace[i] = action[i];
		namespace[i] = '\0';
		p++;
		p2 = strchr(p, '"');
		if(p2 && (p2 - action) <= n)
			methodlen = p2 - p;
		else
			methodlen = n - (p - action);
		/*syslog(LOG_DEBUG, "SoapMethod: %.*s %d %d %p %p %d",
		       methodlen, p, methodlen, n, action, p, (int)(p - action));*/
		for(i = 0; soapMethods[i].methodName; i++) {
			len = strlen(soapMethods[i].methodName);
			if((len == methodlen) && memcmp(p, soapMethods[i].methodName, len) == 0) {
				soapMethods[i].methodImpl(h, soapMethods[i].methodName, namespace);
				return;
			}
		}
		DEBUG_LOG(syslog(LOG_NOTICE, "SoapMethod: Unknown: %.*s %s", methodlen, p, namespace);)
	}
#ifdef DEBUG
	  else {
		syslog(LOG_NOTICE, "cannot parse SoapAction");
	}
#endif

	SoapError(h, 401, "Invalid Action");
}

#ifdef ENABLE_EVENTS
static void
ProcessHTTPUnSubscribe_upnphttp(struct upnphttp * h, const char * path)
{
	DEBUG_LOG(syslog(LOG_DEBUG, "ProcessHTTPUnSubscribe %s", path);
	syslog(LOG_DEBUG, "SID '%.*s'", h->req_SIDLen, h->req_buf + h->req_SIDOff);)
	/* Remove from the list */
	if(upnpevents_removeSubscriber(h->req_buf + h->req_SIDOff, h->req_SIDLen) < 0) {
		BuildResp2_upnphttp(h, 412, "Precondition Failed", 0, 0);
	} else {
		BuildResp_upnphttp(h, 0, 0);
	}
	SendRespAndClose_upnphttp(h);
}
#endif

/* ProcessHTTPPOST_upnphttp()
 * executes the SOAP query if it is possible */
static void
ProcessHTTPPOST_upnphttp(struct upnphttp * h)
{
	if((h->req_buflen - h->req_contentoff) >= h->req_contentlen) {
		/* the request body is received */
		if(h->req_soapActionOff > 0) {
			/* we can process the request */
			DEBUG_LOG(syslog(LOG_INFO, "SOAPAction: %.*s",
			       h->req_soapActionLen, h->req_buf + h->req_soapActionOff);)
			ExecuteSoapAction(h,
				h->req_buf + h->req_soapActionOff,
				h->req_soapActionLen);
		} else {
			static const char err400str[] =
				"<html><body>Bad request</body></html>";
			DEBUG_LOG(syslog(LOG_INFO, "No SOAPAction in HTTP headers");)
			h->respflags = FLAG_HTML;
			BuildResp2_upnphttp(h, 400, "Bad Request",
					    err400str, sizeof(err400str) - 1);
			SendRespAndClose_upnphttp(h);
		}
	} else if(h->respflags & FLAG_CONTINUE) {
		/* Sending the 100 Continue response */
		if(!h->res_buf) {
			h->res_buf = malloc(256);
			h->res_buf_alloclen = 256;
		}
		h->res_buflen = snprintf(h->res_buf, h->res_buf_alloclen,
					 "%s 100 Continue\r\n\r\n", h->HttpVer);
		h->res_sent = 0;
		h->state = ESendingContinue;
		if(SendResp_upnphttp(h))
			h->state = EWaitingForHttpContent;
	} else {
		/* waiting for remaining data */
		h->state = EWaitingForHttpContent;
	}
}

/* Parse and process Http Query
 * called once all the HTTP headers have been received,
 * so it is guaranteed that h->req_buf contains the \r\n\r\n
 * character sequence */
static void
ProcessHttpQuery_upnphttp(struct upnphttp * h)
{
	static const struct {
		const char *path;
		char *(*f)(int *);
	} path_desc[] = {
		{ ROOTDESC_PATH, genRootDesc},
		{ WANIPC_PATH, genWANIPCn},
		{ WANCFG_PATH, genWANCfg},
#ifdef ENABLE_L3F_SERVICE
		{ L3F_PATH, genL3F},
#endif
		{ NULL, NULL}
	};
	char HttpCommand[16];
	char HttpUrl[128];
	char * HttpVer;
	char * p;
	int i;

	p = h->req_buf;
	if(!p)
		return;
	/* note : checking (*p != '\r') is enough to avoid runing off the
	 * end of the buffer, because h->req_buf is guaranteed to contain
	 * the \r\n\r\n character sequence */
	for(i = 0; i<15 && *p != ' ' && *p != '\r'; i++)
		HttpCommand[i] = *(p++);
	HttpCommand[i] = '\0';
	while(*p==' ')
		p++;
	for(i = 0; i<127 && *p != ' ' && *p != '\r'; i++)
		HttpUrl[i] = *(p++);
	HttpUrl[i] = '\0';
	while(*p==' ')
		p++;
	HttpVer = h->HttpVer;
	for(i = 0; i<15 && *p != '\r'; i++)
		HttpVer[i] = *(p++);
	HttpVer[i] = '\0';
	DEBUG_LOG(syslog(LOG_INFO, "HTTP REQUEST from %s : %s %s (%s)",
	       h->clientaddr_str, HttpCommand, HttpUrl, HttpVer);)
	ParseHttpHeaders(h);
	if(h->req_HostOff > 0 && h->req_HostLen > 0) {
		DEBUG_LOG(syslog(LOG_DEBUG, "Host: %.*s", h->req_HostLen, h->req_buf + h->req_HostOff);)
		p = h->req_buf + h->req_HostOff;
		if(*p == '[') {
			/* IPv6 */
			p++;
			while(p < h->req_buf + h->req_HostOff + h->req_HostLen) {
				if(*p == ']') break;
				/* TODO check *p in [0-9a-f:.] */
				p++;
			}
			if(*p != ']') {
				DEBUG_LOG(syslog(LOG_NOTICE, "DNS rebinding attack suspected (Host: %.*s)", h->req_HostLen, h->req_buf + h->req_HostOff);)
				Send404(h);/* 403 */
				return;
			}
			p++;
			/* TODO : Check port */
		} else {
			for(i = 0; i < h->req_HostLen; i++) {
				if(*p != ':' && *p != '.' && (*p > '9' || *p < '0')) {
					DEBUG_LOG(syslog(LOG_NOTICE, "DNS rebinding attack suspected (Host: %.*s)", h->req_HostLen, h->req_buf + h->req_HostOff);)
					Send404(h);/* 403 */
					return;
				}
				p++;
			}
		}
	}
	if(strcmp("POST", HttpCommand) == 0) {
		h->req_command = EPost;
		ProcessHTTPPOST_upnphttp(h);
	} else if(strcmp("GET", HttpCommand) == 0) {
		h->req_command = EGet;
		for(i=0; path_desc[i].path; i++) {
			if(strcasecmp(path_desc[i].path, HttpUrl) == 0) {
				if(path_desc[i].f)
					sendXMLdesc(h, path_desc[i].f);
				else
					continue;
				return;
			}
		}
		if(0 == memcmp(HttpUrl, "/ctl/", 5)) {
			/* 405 Method Not Allowed
			 * Allow: POST */
			h->respflags = FLAG_ALLOW_POST;
			Send405(h);
			return;
		}
#ifdef ENABLE_EVENTS
		if(0 == memcmp(HttpUrl, "/evt/", 5)) {
			/* 405 Method Not Allowed
			 * Allow: SUBSCRIBE, UNSUBSCRIBE */
			h->respflags = FLAG_ALLOW_SUB_UNSUB;
			Send405(h);
			return;
		}
#endif
		DEBUG_LOG(syslog(LOG_NOTICE, "%s not found, responding ERROR 404", HttpUrl);)
		Send404(h);
	}
#ifdef ENABLE_EVENTS
	else if(strcmp("SUBSCRIBE", HttpCommand) == 0) {
		h->req_command = ESubscribe;
		ProcessHTTPSubscribe_upnphttp(h, HttpUrl);
	} else if(strcmp("UNSUBSCRIBE", HttpCommand) == 0) {
		h->req_command = EUnSubscribe;
		ProcessHTTPUnSubscribe_upnphttp(h, HttpUrl);
	}
#else
	else if(strcmp("SUBSCRIBE", HttpCommand) == 0) {
		DEBUG_LOG(syslog(LOG_NOTICE, "SUBSCRIBE not implemented. ENABLE_EVENTS compile option disabled");)
		Send501(h);
	}
#endif
	else {
		DEBUG_LOG(syslog(LOG_NOTICE, "Unsupported HTTP Command %s", HttpCommand);)
		Send501(h);
	}
}

static void
Process_upnphttp(struct upnphttp * h)
{
	char * h_tmp;
	char buf[2048];
	int n;

	if(!h)
		return;
	switch(h->state)
	{
	case EWaitingForHttpRequest:
		n = recv(h->socket, buf, sizeof(buf), 0);
		if(n<0) {
			if(errno != EAGAIN && errno != EWOULDBLOCK && errno != EINTR) {
				bb_perror_msg("recv (state0)");
				h->state = EToDelete;
			}
			/* if errno is EAGAIN, EWOULDBLOCK or EINTR, try again later */
		} else if(n==0) {
			xlog_warn("HTTP Connection from %s closed unexpectedly", inet_ntoa(h->clientaddr));
			h->state = EToDelete;
		} else {
			const char * endheaders;
			/* if 1st arg of realloc() is null,
			 * realloc behaves the same as malloc() */
			h_tmp = (char *)realloc(h->req_buf, n + h->req_buflen + 1);
			if (h_tmp == NULL) {
				xlog_warn("Unable to allocate new memory for h->req_buf)");
				h->state = EToDelete;
			} else {
				h->req_buf = h_tmp;
				memcpy(h->req_buf + h->req_buflen, buf, n);
				h->req_buflen += n;
				h->req_buf[h->req_buflen] = '\0';
			}
			/* search for the string "\r\n\r\n" */
			endheaders = findendheaders(h->req_buf, h->req_buflen);
			if(endheaders) {
				/* at this point, the request buffer (h->req_buf)
				 * is guaranteed to contain the \r\n\r\n character sequence */
				h->req_contentoff = endheaders - h->req_buf + 4;
				ProcessHttpQuery_upnphttp(h);
			}
		}
		break;
	case EWaitingForHttpContent:
		n = recv(h->socket, buf, sizeof(buf), 0);
		if(n<0) {
			if(errno != EAGAIN && errno != EWOULDBLOCK && errno != EINTR) {
				bb_perror_msg("recv (state1)");
				h->state = EToDelete;
			}
			/* if errno is EAGAIN, EWOULDBLOCK or EINTR, try again later */
		} else if(n==0) {
			xlog_warn("HTTP Connection from %s closed unexpectedly", inet_ntoa(h->clientaddr));
			h->state = EToDelete;
		} else {
			void * tmp = realloc(h->req_buf, n + h->req_buflen);
			if(!tmp) {
				bb_perror_msg("memory allocation error");
				h->state = EToDelete;
			} else {
				h->req_buf = tmp;
				memcpy(h->req_buf + h->req_buflen, buf, n);
				h->req_buflen += n;
				if((h->req_buflen - h->req_contentoff) >= h->req_contentlen)
					ProcessHTTPPOST_upnphttp(h);
			}
		}
		break;
	case ESendingContinue:
		if(SendResp_upnphttp(h))
			h->state = EWaitingForHttpContent;
		break;
	case ESendingAndClosing:
		SendRespAndClose_upnphttp(h);
		break;
	default:
		xlog_warn("Unexpected state: %d", h->state);
	}
}

#ifdef ENABLE_EVENTS
static void upnp_event_prepare(struct upnp_event_notify *obj)
{
	static const char notifymsg[] =
		"NOTIFY %s HTTP/1.1\r\n"
		"Host: %s%s\r\n"
#if (UPNP_VERSION_MAJOR == 1) && (UPNP_VERSION_MINOR == 0)
		"Content-Type: text/xml\r\n"    /* UDA v1.0 */
#else
		"Content-Type: text/xml; charset=\"utf-8\"\r\n" /* UDA v1.1 or later */
#endif
		"Content-Length: %d\r\n"
		"NT: upnp:event\r\n"
		"NTS: upnp:propchange\r\n"
		"SID: %s\r\n"
		"SEQ: %u\r\n"
		"Connection: close\r\n"
		"Cache-Control: no-cache\r\n"
		"\r\n"
		"%.*s\r\n";
	char * xml;
	int l;
	if(obj->sub == NULL) {
		obj->state = EError;
		return;
	}
	switch(obj->sub->service) {
	case EWanCFG:
		xml = getVarsWANCfg(&l);
		break;
	case EWanIPC:
		xml = getVarsWANIPCn(&l);
		break;
#ifdef ENABLE_L3F_SERVICE
	case EL3F:
		xml = getVarsL3F(&l);
		break;
#endif
	default:
		xml = NULL;
		l = 0;
	}
	obj->buffersize = 1024;
	obj->buffer = malloc(obj->buffersize);
	if(!obj->buffer) {
		bb_error_msg("%s: malloc returned NULL", "upnp_event_prepare");
		if(xml)
			free(xml);
		obj->state = EError;
		return;
	}
	obj->tosend = snprintf(obj->buffer, obj->buffersize, notifymsg,
			       obj->path, obj->addrstr, obj->portstr, l+2,
			       obj->sub->uuid, obj->sub->seq,
			       l, xml);
	if(xml) {
		free(xml);
		xml = NULL;
	}
	obj->state = ESending;
}

static void upnp_event_process_notify(struct upnp_event_notify *obj)
{
	int err;
	socklen_t len;
	switch(obj->state) {
	case EConnecting:
		/* now connected or failed to connect */
		len = sizeof(err);
		if(getsockopt(obj->s, SOL_SOCKET, SO_ERROR, &err, &len) < 0) {
			bb_perror_msg("%s: getsockopt", "upnp_event_process_notify");
			obj->state = EError;
			break;
		}
		if(err != 0) {
			errno = err;
			xlog_warn("%s: connect(%s%s): %m", "upnp_event_process_notify",
			       obj->addrstr, obj->portstr);
			obj->state = EError;
			break;
		}
		upnp_event_prepare(obj);
		if(obj->state == ESending)
			upnp_event_send(obj);
		break;
	case ESending:
		upnp_event_send(obj);
		break;
	case EWaitingForResponse:
		upnp_event_recv(obj);
		break;
	case EFinished:
		close(obj->s);
		obj->s = -1;
		break;
	default:
		bb_error_msg("%s: unknown state", "upnp_event_process_notify");
	}
}

static void upnpevents_processfds(fd_set *readset, fd_set *writeset)
{
	struct upnp_event_notify * obj;
	struct upnp_event_notify * next;
	struct subscriber * sub;
	struct subscriber * subnext;
	time_t curtime;
	for(obj = notifylist.lh_first; obj != NULL; obj = obj->entries.le_next) {
		DEBUG_LOG(syslog(LOG_DEBUG, "%s: %p %d %d %d %d",
		       "upnpevents_processfds", obj, obj->state, obj->s,
		       FD_ISSET(obj->s, readset), FD_ISSET(obj->s, writeset));)
		if(obj->s >= 0) {
			if(FD_ISSET(obj->s, readset) || FD_ISSET(obj->s, writeset))
				upnp_event_process_notify(obj);
		}
	}
	obj = notifylist.lh_first;
	while(obj != NULL) {
		next = obj->entries.le_next;
		if(obj->state == EError || obj->state == EFinished) {
			if(obj->s >= 0) {
				close(obj->s);
			}
			if(obj->sub)
				obj->sub->notify = NULL;
			/* remove also the subscriber from the list if there was an error */
			if(obj->state == EError && obj->sub) {
				LIST_REMOVE(obj->sub, entries);
				free(obj->sub);
			}
			if(obj->buffer) {
				free(obj->buffer);
			}
			LIST_REMOVE(obj, entries);
			free(obj);
		}
		obj = next;
	}
	/* remove timeouted subscribers */
	curtime = time(NULL);
	for(sub = subscriberlist.lh_first; sub != NULL; ) {
		subnext = sub->entries.le_next;
		if(sub->timeout && curtime > sub->timeout && sub->notify == NULL) {
			DEBUG_LOG(syslog(LOG_INFO, "subscriber timeouted : %u > %u SID=%s",
			       (unsigned)curtime, (unsigned)sub->timeout, sub->uuid);)
			LIST_REMOVE(sub, entries);
			free(sub);
		}
		sub = subnext;
	}
}
#endif

#if (ENABLE_FEATURE_IPV6 && defined(ENABLE_PCP)) || ENABLE_FEATURE_MINIUPNPD_NATPMP
/*
 * Receives NATPMP and PCP packets and stores them in msg_buff.
 * The sender information is stored in senderaddr.
 * Returns number of bytes recevied, even if number is negative.
 */
static int ReceiveNATPMPOrPCPPacket(int s, struct sockaddr * senderaddr,
			     socklen_t * senderaddrlen,
			     unsigned char * msg_buff, size_t msg_buff_size)
{
	int n;

	n = recvfrom(s, msg_buff, msg_buff_size, 0,
		     senderaddr, senderaddrlen);

	if(n<0) {
		/* EAGAIN, EWOULDBLOCK and EINTR : silently ignore (retry next time)
		 * other errors : log to LOG_ERR */
		if(errno != EAGAIN &&
		   errno != EWOULDBLOCK &&
		   errno != EINTR) {
			bb_perror_msg("recvfrom(natpmp)");
		}
		return n;
	}
	return n;
}
#endif

/* === main === */
/* process HTTP or SSDP requests */
int miniupnpd_main(void)
{
	int i;
	int shttpl = -1;        /* socket for HTTP */
	int sudp = -1;          /* IP v4 socket for receiving SSDP */
#if ENABLE_FEATURE_IPV6
	int sudpv6 = -1;        /* IP v6 socket for receiving SSDP */
#endif
#if ENABLE_FEATURE_MINIUPNPD_NATPMP
	int * snatpmp = NULL;   /* also used for PCP */
#endif
#if ENABLE_FEATURE_IPV6 && defined(ENABLE_PCP)
	int spcp_v6 = -1;
#endif
#ifdef USE_IFACEWATCHER
	int sifacewatcher = -1;
#endif
	int * snotify = NULL;
	int addr_count;
	LIST_HEAD(httplisthead, upnphttp) upnphttphead;
	struct upnphttp *e = 0;
	struct upnphttp *next;
	fd_set readset; /* for select() */
	fd_set writeset;
	struct timeval timeout, timeofday, lasttimeofday = {0, 0};
	int max_fd = -1;
	struct runtime_vars v;
	/* variables used for the unused-rule cleanup process */
	struct rule_state * rule_list = 0;
	struct timeval checktime = {0, 0};
	struct lan_addr_s *lan_addr;
#ifdef ENABLE_UPNPPINHOLE
	unsigned int next_pinhole_ts;
#endif

	init(&v);

	/* count lan addrs */
	addr_count = 0;
	for(lan_addr = bb_common_bss.lan_addrs.lh_first; lan_addr != NULL; lan_addr = lan_addr->list.le_next)
		addr_count++;
	if(addr_count > 0) {
#if !ENABLE_FEATURE_IPV6
		snotify = calloc(addr_count, sizeof(int));
#else
		/* one for IPv4, one for IPv6 */
		snotify = calloc(addr_count * 2, sizeof(int));
#endif
	}
#if ENABLE_FEATURE_MINIUPNPD_NATPMP
	if(addr_count > 0) {
		snatpmp = malloc(addr_count * sizeof(int));
		for(i = 0; i < addr_count; i++)
			snatpmp[i] = -1;
	}
#endif

	LIST_INIT(&upnphttphead);

	if(
#if ENABLE_FEATURE_MINIUPNPD_NATPMP
	!GETFLAG(ENABLENATPMPMASK) &&
#endif
	!GETFLAG(ENABLEUPNPMASK) ) {
		bb_error_msg("Why did you run me anyway?");
		return 0;
	}
	DEBUG_LOG(syslog(LOG_INFO, "version " MINIUPNPD_VERSION " starting%s%sext if %s BOOTID=%u",
#if ENABLE_FEATURE_MINIUPNPD_NATPMP
#ifdef ENABLE_PCP
	       GETFLAG(ENABLENATPMPMASK) ? " NAT-PMP/PCP " : " ",
#else
	       GETFLAG(ENABLENATPMPMASK) ? " NAT-PMP " : " ",
#endif
#else
	       " ",
#endif
	       GETFLAG(ENABLEUPNPMASK) ? "UPnP-IGD " : "",
	       bb_common_bss.ext_if_name, bb_common_bss.upnp_bootid);)
	if(GETFLAG(ENABLEUPNPMASK)) {
		unsigned short listen_port = (v.port > 0) ? v.port : 0;

		/* open socket for HTTP connections. Listen on the 1st LAN address */
#if ENABLE_FEATURE_IPV6
		shttpl = OpenAndConfHTTPSocket(&listen_port, 1);
#else
		shttpl = OpenAndConfHTTPSocket(&listen_port);
#endif
		if(shttpl < 0)
			bb_error_msg_and_die("Failed to open socket for HTTP. EXITING");
		v.port = listen_port;
		DEBUG_LOG(syslog(LOG_NOTICE, "HTTP listening on port %d", v.port);)
#if ENABLE_FEATURE_IPV6
		if(find_ipv6_addr(bb_common_bss.lan_addrs.lh_first ? bb_common_bss.lan_addrs.lh_first->ifname : NULL,
				  bb_common_bss.ipv6_addr_for_http_with_brackets, sizeof(bb_common_bss.ipv6_addr_for_http_with_brackets)) > 0) {
			DEBUG_LOG(syslog(LOG_NOTICE, "HTTP IPv6 address given to control points : %s",
			       bb_common_bss.ipv6_addr_for_http_with_brackets));
		} else {
			memcpy(bb_common_bss.ipv6_addr_for_http_with_brackets, "[::1]", 6);
			xlog_warn("no HTTP IPv6 address, disabling IPv6");
			SETFLAG(IPV6DISABLEDMASK);
		}
#endif

		/* open socket for SSDP connections */
		sudp = OpenAndConfSSDPReceiveSocket(0);
		if(sudp < 0) {
			DEBUG_LOG(syslog(LOG_NOTICE, "Failed to open socket for receiving SSDP. Trying to use MiniSSDPd");)
			if(SubmitServicesToMiniSSDPD(bb_common_bss.lan_addrs.lh_first->str, v.port) < 0)
				bb_error_msg_and_die("Failed to connect to MiniSSDPd. EXITING");
		}
#if ENABLE_FEATURE_IPV6
		if(!GETFLAG(IPV6DISABLEDMASK)) {
			sudpv6 = OpenAndConfSSDPReceiveSocket(1);
			if(sudpv6 < 0)
				xlog_warn("Failed to open socket for receiving SSDP (IP v6).");
		}
#endif

		/* open socket for sending notifications */
		if(OpenAndConfSSDPNotifySockets(snotify) < 0)
			bb_error_msg_and_die("Failed to open sockets for sending SSDP notify messages. EXITING");

#ifdef USE_IFACEWATCHER
		/* open socket for kernel notifications about new network interfaces */
		if (sudp >= 0) {
			sifacewatcher = OpenAndConfInterfaceWatchSocket();
			if (sifacewatcher < 0)
				bb_error_msg("Failed to open socket for receiving network interface notifications");
		}
#endif
	}

#if ENABLE_FEATURE_MINIUPNPD_NATPMP
	/* open socket for NAT PMP traffic */
	if(GETFLAG(ENABLENATPMPMASK)) {
		if(OpenAndConfNATPMPSockets(snatpmp) < 0)
#ifdef ENABLE_PCP
		{
			bb_error_msg("Failed to open sockets for NAT-PMP/PCP.");
		}
#ifdef DEBUG
		  else {
			syslog(LOG_NOTICE, "Listening for NAT-PMP/PCP traffic on port %u", NATPMP_PORT);
		}
#endif
#else
		{
			bb_error_msg("Failed to open sockets for NAT PMP.");
		}
#ifdef DEBUG
		  else {
			syslog(LOG_NOTICE, "Listening for NAT-PMP traffic on port %u", NATPMP_PORT);
		}

#endif
#endif /* ENABLE_PCP */
	}
#endif /* ENABLE_FEATURE_MINIUPNPD_NATPMP */

#if ENABLE_FEATURE_IPV6 && defined(ENABLE_PCP)
	spcp_v6 = OpenAndConfPCPv6Socket();
#endif

	/* main loop */
	while(!bb_common_bss.quitting) {
		/* Correct startup_time if it was set with a RTC close to 0 */
		if((bb_common_bss.startup_time<60*60*24) && (time(NULL)>60*60*24))
			set_startup_time(GETFLAG(SYSUPTIMEMASK));
		/* send public address change notifications if needed */
		if(bb_common_bss.should_send_public_address_change_notif) {
			DEBUG_LOG(syslog(LOG_INFO, "should send external iface address change notification(s)");)
#if ENABLE_FEATURE_MINIUPNPD_NATPMP
			if(GETFLAG(ENABLENATPMPMASK))
				SendNATPMPPublicAddressChangeNotification(snatpmp, addr_count);
#endif
#ifdef ENABLE_EVENTS
			if(GETFLAG(ENABLEUPNPMASK))
				upnp_event_var_change_notify(EWanIPC);
#endif
			bb_common_bss.should_send_public_address_change_notif = 0;
		}
		/* Check if we need to send SSDP NOTIFY messages and do it if
		 * needed */
		if(gettimeofday(&timeofday, 0) < 0) {
			bb_perror_msg("gettimeofday()");
			timeout.tv_sec = v.notify_interval;
			timeout.tv_usec = 0;
		} else {
			/* the comparaison is not very precise but who cares ? */
			if(timeofday.tv_sec >= (lasttimeofday.tv_sec + v.notify_interval)) {
				if (GETFLAG(ENABLEUPNPMASK))
					SendSSDPNotifies2(snotify,
						  (unsigned short)v.port,
						  v.notify_interval << 1);
				memcpy(&lasttimeofday, &timeofday, sizeof(struct timeval));
				timeout.tv_sec = v.notify_interval;
				timeout.tv_usec = 0;
			} else {
				timeout.tv_sec = lasttimeofday.tv_sec + v.notify_interval
						 - timeofday.tv_sec;
				if(timeofday.tv_usec > lasttimeofday.tv_usec) {
					timeout.tv_usec = 1000000 + lasttimeofday.tv_usec
							  - timeofday.tv_usec;
					timeout.tv_sec--;
				} else {
					timeout.tv_usec = lasttimeofday.tv_usec - timeofday.tv_usec;
				}
			}
		}
		/* remove unused rules */
		if( v.clean_ruleset_interval
		  && (timeofday.tv_sec >= checktime.tv_sec + v.clean_ruleset_interval))
		{
			if(rule_list) {
				remove_unused_rules(rule_list);
				rule_list = NULL;
			} else {
				rule_list = get_upnp_rules_state_list(v.clean_ruleset_threshold);
			}
			memcpy(&checktime, &timeofday, sizeof(struct timeval));
		}
		/* Remove expired port mappings, based on UPnP IGD LeaseDuration
		 * or NAT-PMP lifetime) */
		if(bb_common_bss.nextruletoclean_timestamp
		  && ((unsigned int)timeofday.tv_sec >= bb_common_bss.nextruletoclean_timestamp)) {
			DEBUG_LOG(syslog(LOG_DEBUG, "cleaning expired Port Mappings");)
			get_upnp_rules_state_list(0);
		}
		if(bb_common_bss.nextruletoclean_timestamp
		  && ((unsigned int)timeout.tv_sec >= (bb_common_bss.nextruletoclean_timestamp - timeofday.tv_sec)))
		{
			timeout.tv_sec = bb_common_bss.nextruletoclean_timestamp - timeofday.tv_sec;
			timeout.tv_usec = 0;
			DEBUG_LOG(syslog(LOG_DEBUG, "setting timeout to %u sec", (unsigned)timeout.tv_sec);)
		}
#ifdef ENABLE_UPNPPINHOLE
		/* Clean up expired IPv6 PinHoles */
		next_pinhole_ts = 0;
#if defined(USE_NETFILTER)
		clean_pinhole_list(&next_pinhole_ts);
#endif
		if(next_pinhole_ts &&
		   timeout.tv_sec >= (int)(next_pinhole_ts - timeofday.tv_sec)) {
			timeout.tv_sec = next_pinhole_ts - timeofday.tv_sec;
			timeout.tv_usec = 0;
		}
#endif /* ENABLE_UPNPPINHOLE */

		/* select open sockets (SSDP, HTTP listen, and all HTTP soap sockets) */
		FD_ZERO(&readset);
		FD_ZERO(&writeset);

		if (sudp >= 0) {
			FD_SET(sudp, &readset);
			max_fd = MAX( max_fd, sudp);
#ifdef USE_IFACEWATCHER
			if (sifacewatcher >= 0) {
				FD_SET(sifacewatcher, &readset);
				max_fd = MAX(max_fd, sifacewatcher);
			}
#endif
		}
		if (shttpl >= 0) {
			FD_SET(shttpl, &readset);
			max_fd = MAX( max_fd, shttpl);
		}
#if ENABLE_FEATURE_IPV6
		if (sudpv6 >= 0) {
			FD_SET(sudpv6, &readset);
			max_fd = MAX( max_fd, sudpv6);
		}
#endif

		i = 0;  /* active HTTP connections count */
		for(e = upnphttphead.lh_first; e != NULL; e = e->entries.le_next) {
			if(e->socket >= 0) {
				if(e->state <= EWaitingForHttpContent)
					FD_SET(e->socket, &readset);
				else if(e->state == ESendingAndClosing)
					FD_SET(e->socket, &writeset);
				else
					continue;
				max_fd = MAX(max_fd, e->socket);
				i++;
			}
		}
		/* for debug */
#if ENABLE_FEATURE_MINIUPNPD_NATPMP
		for(i=0; i<addr_count; i++) {
			if(snatpmp[i] >= 0) {
				FD_SET(snatpmp[i], &readset);
				max_fd = MAX( max_fd, snatpmp[i]);
			}
		}
#endif
#if ENABLE_FEATURE_IPV6 && defined(ENABLE_PCP)
		if(spcp_v6 >= 0) {
			FD_SET(spcp_v6, &readset);
			max_fd = MAX(max_fd, spcp_v6);
		}
#endif

#ifdef ENABLE_EVENTS
		upnpevents_selectfds(&readset, &writeset, &max_fd);
#endif

		/* queued "sendto" */
		{
			struct timeval next_send;
			i = get_next_scheduled_send(&next_send);
			if(i > 0) {
				i = get_sendto_fds(&writeset, &max_fd, &timeofday);
				if(timeofday.tv_sec > next_send.tv_sec ||
				   (timeofday.tv_sec == next_send.tv_sec && timeofday.tv_usec >= next_send.tv_usec)) {
					if(i > 0) {
						timeout.tv_sec = 0;
						timeout.tv_usec = 0;
					}
				} else {
					struct timeval tmp_timeout;
					tmp_timeout.tv_sec = (next_send.tv_sec - timeofday.tv_sec);
					tmp_timeout.tv_usec = (next_send.tv_usec - timeofday.tv_usec);
					if(tmp_timeout.tv_usec < 0) {
						tmp_timeout.tv_usec += 1000000;
						tmp_timeout.tv_sec--;
					}
					if(timeout.tv_sec > tmp_timeout.tv_sec
					   || (timeout.tv_sec == tmp_timeout.tv_sec && timeout.tv_usec > tmp_timeout.tv_usec)) {
						timeout.tv_sec = tmp_timeout.tv_sec;
						timeout.tv_usec = tmp_timeout.tv_usec;
					}
				}
			}
		}

		if(select(max_fd+1, &readset, &writeset, 0, &timeout) < 0) {
			if(bb_common_bss.quitting) goto shutdown;
			if(errno == EINTR) continue; /* interrupted by a signal, start again */
			bb_perror_msg_and_die("select(all). Failed to select open sockets. EXITING");
		}
		i = try_sendto(&writeset);
		if(i < 0)
			bb_error_msg("try_sendto failed to send %d packets", -i);
#ifdef ENABLE_EVENTS
		upnpevents_processfds(&readset, &writeset);
#endif
#if ENABLE_FEATURE_MINIUPNPD_NATPMP
		/* process NAT-PMP packets */
		for(i=0; i<addr_count; i++) {
			if((snatpmp[i] >= 0) && FD_ISSET(snatpmp[i], &readset)) {
				unsigned char msg_buff[PCP_MAX_LEN];
				struct sockaddr_in senderaddr;
				socklen_t senderaddrlen;
				int len;
				memset(msg_buff, 0, PCP_MAX_LEN);
				senderaddrlen = sizeof(senderaddr);
				len = ReceiveNATPMPOrPCPPacket(snatpmp[i],
							       (struct sockaddr *)&senderaddr,
							       &senderaddrlen,
							       msg_buff, sizeof(msg_buff));
				if (len < 1)
					continue;
#ifdef ENABLE_PCP
				if (msg_buff[0]==0) {  /* version equals to 0 -> means NAT-PMP */
					/* Check if the packet is coming from a LAN to enforce RFC6886 :
					 * The NAT gateway MUST NOT accept mapping requests destined to the NAT
					 * gateway's external IP address or received on its external network
					 * interface.  Only packets received on the internal interface(s) with a
					 * destination address matching the internal address(es) of the NAT
					 * gateway should be allowed. */
					/* TODO : move to ProcessIncomingNATPMPPacket() ? */
					lan_addr = get_lan_for_peer((struct sockaddr *)&senderaddr);
					if(lan_addr == NULL) {
						char sender_str[64];
						sockaddr_to_string((struct sockaddr *)&senderaddr, sender_str, sizeof(sender_str));
						xlog_warn("NAT-PMP packet sender %s not from a LAN, ignoring",
						       sender_str);
						continue;
					}
					ProcessIncomingNATPMPPacket(snatpmp[i], msg_buff, len,
								    &senderaddr);
				} else { /* everything else can be PCP */
					ProcessIncomingPCPPacket(snatpmp[i], msg_buff, len,
								 (struct sockaddr *)&senderaddr, NULL);
				}

#else
				/* Check if the packet is coming from a LAN to enforce RFC6886 :
				 * The NAT gateway MUST NOT accept mapping requests destined to the NAT
				 * gateway's external IP address or received on its external network
				 * interface.  Only packets received on the internal interface(s) with a
				 * destination address matching the internal address(es) of the NAT
				 * gateway should be allowed. */
				/* TODO : move to ProcessIncomingNATPMPPacket() ? */
				lan_addr = get_lan_for_peer((struct sockaddr *)&senderaddr);
				if(lan_addr == NULL) {
					char sender_str[64];
					sockaddr_to_string((struct sockaddr *)&senderaddr, sender_str, sizeof(sender_str));
					xlog_warn("NAT-PMP packet sender %s not from a LAN, ignoring", sender_str);
					continue;
				}
				ProcessIncomingNATPMPPacket(snatpmp[i], msg_buff, len, &senderaddr);
#endif
			}
		}
#endif
#if ENABLE_FEATURE_IPV6 && defined(ENABLE_PCP)
		/* in IPv6, only PCP is supported, not NAT-PMP */
		if(spcp_v6 >= 0 && FD_ISSET(spcp_v6, &readset))
		{
			unsigned char msg_buff[PCP_MAX_LEN];
			struct sockaddr_in6 senderaddr;
			socklen_t senderaddrlen;
			struct sockaddr_in6 receiveraddr;
			int len;
			memset(msg_buff, 0, PCP_MAX_LEN);
			senderaddrlen = sizeof(senderaddr);
			len = ReceiveNATPMPOrPCPPacket(spcp_v6,
						       (struct sockaddr *)&senderaddr,
						       &senderaddrlen,
						       msg_buff, sizeof(msg_buff));
			if(len >= 1)
				ProcessIncomingPCPPacket(spcp_v6, msg_buff, len,
							 (struct sockaddr *)&senderaddr,
							 &receiveraddr);
		}
#endif
		/* process SSDP packets */
		if(sudp >= 0 && FD_ISSET(sudp, &readset)) {
			/*syslog(LOG_INFO, "Received UDP Packet");*/
			ProcessSSDPRequest(sudp, (unsigned short)v.port);
		}
#if ENABLE_FEATURE_IPV6
		if(sudpv6 >= 0 && FD_ISSET(sudpv6, &readset)) {
			DEBUG_LOG(syslog(LOG_INFO, "Received UDP Packet (IPv6)");)
			ProcessSSDPRequest(sudpv6, (unsigned short)v.port);
		}
#endif
#ifdef USE_IFACEWATCHER
		/* process kernel notifications */
		if (sifacewatcher >= 0 && FD_ISSET(sifacewatcher, &readset))
			ProcessInterfaceWatchNotify(sifacewatcher);
#endif

		/* process active HTTP connections */
		/* LIST_FOREACH macro is not available under linux */
		for(e = upnphttphead.lh_first; e != NULL; e = e->entries.le_next) {
			if(e->socket >= 0) {
				if(FD_ISSET(e->socket, &readset) || FD_ISSET(e->socket, &writeset))
					Process_upnphttp(e);
			}
		}
		/* process incoming HTTP connections */
		if(shttpl >= 0 && FD_ISSET(shttpl, &readset)) {
			struct upnphttp * tmp = ProcessIncomingHTTP(shttpl, "HTTP");

			if(tmp) {
				LIST_INSERT_HEAD(&upnphttphead, tmp, entries);
			}
		}
		/* delete finished HTTP connections */
		for(e = upnphttphead.lh_first; e != NULL; ) {
			next = e->entries.le_next;
			if(e->state >= EToDelete) {
				LIST_REMOVE(e, entries);
				Delete_upnphttp(e);
			}
			e = next;
		}

	}       /* end of main loop */

shutdown:
	DEBUG_LOG(syslog(LOG_NOTICE, "shutting down MiniUPnPd");)
	/* send good-bye */
	if (GETFLAG(ENABLEUPNPMASK)) {
#if !ENABLE_FEATURE_IPV6
		if(SendSSDPGoodbye(snotify, addr_count) < 0)
#else
		if(SendSSDPGoodbye(snotify, addr_count * 2) < 0)
#endif
			bb_error_msg("Failed to broadcast good-bye notifications");
	}
	/* try to send pending packets */
	finalize_sendto();

	/* close out open sockets */
	while(upnphttphead.lh_first != NULL) {
		e = upnphttphead.lh_first;
		LIST_REMOVE(e, entries);
		Delete_upnphttp(e);
	}

	if (sudp >= 0) close(sudp);
	if (shttpl >= 0) close(shttpl);
#if ENABLE_FEATURE_IPV6
	if (sudpv6 >= 0) close(sudpv6);
#endif
#ifdef USE_IFACEWATCHER
	if(sifacewatcher >= 0) close(sifacewatcher);
#endif
#if ENABLE_FEATURE_MINIUPNPD_NATPMP
	for(i=0; i<addr_count; i++) {
		if(snatpmp[i]>=0) {
			close(snatpmp[i]);
			snatpmp[i] = -1;
		}
	}
#endif
#if ENABLE_FEATURE_IPV6 && defined(ENABLE_PCP)
	if(spcp_v6 >= 0) {
		close(spcp_v6);
		spcp_v6 = -1;
	}
#endif
	if (GETFLAG(ENABLEUPNPMASK)) {
#if !ENABLE_FEATURE_IPV6
		for(i = 0; i < addr_count; i++)
#else
		for(i = 0; i < addr_count * 2; i++)
#endif
			close(snotify[i]);
	}

	/* remove pidfile */
	if(bb_common_bss.pidfilename && (unlink(bb_common_bss.pidfilename) < 0))
		bb_perror_msg_and_die("Failed to remove pidfile %s", bb_common_bss.pidfilename);

	return 0;
}
#endif /* APPLET_USAGE */
