/*
 * Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 * Released under the terms of the GNU GPL v2.0.
 *
 * Introduced single menu mode (show all sub-menus in one large tree).
 * 2002-11-06 Petr Baudis <pasky@ucw.cz>
 *
 * Directly use liblxdialog library routines.
 * 2002-11-14 Petr Baudis <pasky@ucw.cz>
 *
 * Internal lxdialog as one file with all static, removed unused code
 * 2007-08-25 (c) Vladimir Oleynik <dzo@simtreas.ru>
 *
 * Directly use libncurses-5.2 as one file with all static, removed unused code
 * 2016-10-05 (c) Vladimir Oleynik <dzo@simtreas.ru>
 */

/*
 *  dialog.h -- common declarations for all dialog modules
 *
 *  AUTHOR: Savio Lam (lam836@cs.cuhk.hk)
 *  MODIFIED FOR LINUX KERNEL CONFIG BY: William Roadcap (roadcap@cfw.com)
 *     Stuart Herbert - S.Herbert@sheffield.ac.uk: radiolist extension
 *     Alessandro Rubini - rubini@ipvvis.unipv.it: merged the two
 *
 *  Changes by Clifford Wolf (god@clifford.at)
 *
 *  [ 1998-06-13 ]
 *
 *    *)  A bugfix for the Page-Down problem
 *
 *    *)  Formerly when I used Page Down and Page Up, the cursor would be set
 *        to the first position in the menu box.  Now lxdialog is a bit
 *        smarter and works more like other menu systems (just have a look at
 *        it).
 *
 *    *)  Formerly if I selected something my scrolling would be broken because
 *        lxdialog is re-invoked by the Menuconfig shell script, can't
 *        remember the last scrolling position, and just sets it so that the
 *        cursor is at the bottom of the box.  Now it writes the temporary file
 *        lxdialog.scrltmp which contains this information. The file is
 *        deleted by lxdialog if the user leaves a submenu or enters a new
 *        one, but it would be nice if Menuconfig could make another "rm -f"
 *        just to be sure.  Just try it out - you will recognise a difference!
 *
 *  [ 1998-06-14 ]
 *
 *    *)  Now lxdialog is crash-safe against broken "lxdialog.scrltmp" files
 *        and menus change their size on the fly.
 *
 *    *)  If for some reason the last scrolling position is not saved by
 *        lxdialog, it sets the scrolling so that the selected item is in the
 *        middle of the menu box, not at the bottom.
 *
 * 02 January 1999, Michael Elizabeth Chastain (mec@shout.net)
 * Reset 'scroll' to 0 if the value from lxdialog.scrltmp is bogus.
 * This fixes a bug in Menuconfig where using ' ' to descend into menus
 * would leave mis-synchronized lxdialog.scrltmp files lying around,
 * fscanf would read in 'scroll', and eventually that value would get used.
 *
 *
 * This is NOT the official version of dialog.  This version has been
 * significantly modified from the original.  It is for use by the Linux
 * kernel configuration script.  Please do not bother Savio Lam with
 * questions about this program.
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/termios.h>
#include <stdio.h>
#include <assert.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <stdarg.h>
#include <termios.h>
#include <limits.h>

/* mini curses 5.2 */
/*
 * Copyright (c) 1998,1999,2000 Free Software Foundation, Inc.              *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

/****************************************************************************
 *  Author: Zeyd M. Ben-Halim <zmbenhal@netcom.com> 1992,1995               *
 *     and: Eric S. Raymond <esr@snark.thyrsus.com>                         *
 ****************************************************************************/

#undef bool
typedef int bool;

#undef TRUE
#define TRUE    1

#undef FALSE
#define FALSE   0

static int      LINES;
static int      COLS;

typedef unsigned long chtype;

/* line graphics */

typedef struct _win_st WINDOW;


#ifndef CURSES_PRIV_H
#define CURSES_PRIV_H 1

/* #include <ncurses_cfg.h> */

#define GCC_NORETURN __attribute__((noreturn))
#define GCC_UNUSED __attribute__((unused))
#define HAVE_LIBGPM 1
#define HAVE_NANOSLEEP 1
#define HAVE_POLL_H 1
#define HAVE_SELECT 1
#define HAVE_SETBUFFER 1
#define HAVE_SETVBUF 1
#define HAVE_SIGACTION 1
#define HAVE_SIZECHANGE 1
#define HAVE_SYS_TIME_H 1
#define HAVE_SYS_TIME_SELECT 1
#define HAVE_TERMIOS_H 1
#define HAVE_TERMIO_H 1
#define HAVE_VSNPRINTF 1
#define HAVE_WORKING_POLL 1
#define NCURSES_EXT_FUNCS 1
#define RETSIGTYPE void
#define TERMINFO "/usr/share/terminfo"
#define TERMINFO_DIRS "/usr/share/terminfo"
#define USE_HASHMAP 1
#define USE_SIGWINCH 1

/*
 * This file is generated by ./MKncurses_def.sh
 */

#ifndef NC_DEFINE_H
#define NC_DEFINE_H 1

#ifndef GCC_NORETURN
#define GCC_NORETURN /* nothing */
#endif

#ifndef GCC_UNUSED
#define GCC_UNUSED /* nothing */
#endif

#ifndef HAVE_LIBGPM
#define HAVE_LIBGPM 0
#endif

#ifndef HAVE_NANOSLEEP
#define HAVE_NANOSLEEP 0
#endif

#ifndef HAVE_SELECT
#define HAVE_SELECT 0
#endif

#ifndef HAVE_SETBUFFER
#define HAVE_SETBUFFER 0
#endif

#ifndef HAVE_SETVBUF
#define HAVE_SETVBUF 0
#endif

#ifndef HAVE_SIGACTION
#define HAVE_SIGACTION 0
#endif

#ifndef HAVE_SIGVEC
#define HAVE_SIGVEC 0
#endif

#ifndef HAVE_SIZECHANGE
#define HAVE_SIZECHANGE 0
#endif

#ifndef HAVE_SYS_TIME_H
#define HAVE_SYS_TIME_H 0
#endif

#ifndef HAVE_SYS_TIME_SELECT
#define HAVE_SYS_TIME_SELECT 0
#endif

#ifndef HAVE_TERMIOS_H
#define HAVE_TERMIOS_H 0
#endif

#ifndef HAVE_TERMIO_H
#define HAVE_TERMIO_H 0
#endif

#ifndef HAVE_TYPE_SIGACTION
#define HAVE_TYPE_SIGACTION 0
#endif

#ifndef HAVE_VSNPRINTF
#define HAVE_VSNPRINTF 0
#endif

#ifndef HAVE_WORKING_POLL
#define HAVE_WORKING_POLL 0
#endif

#ifndef NCURSES_EXT_FUNCS
#define NCURSES_EXT_FUNCS 0
#endif

#ifndef SVR4_TERMIO
#define SVR4_TERMIO 0
#endif

#ifndef USE_SCROLL_HINTS
#define USE_SCROLL_HINTS 0
#endif

#ifndef USE_SIGWINCH
#define USE_SIGWINCH 0
#endif

#endif /* NC_DEFINE_H */

/* end #include <ncurses_cfg.h> */

#ifndef PATH_MAX
# if defined(_POSIX_PATH_MAX)
#  define PATH_MAX _POSIX_PATH_MAX
# elif defined(MAXPATHLEN)
#  define PATH_MAX MAXPATHLEN
# else
#  define PATH_MAX 255  /* the Posix minimum path-size */
# endif
#endif

/* Some systems have a broken 'select()', but workable 'poll()'.  Use that */
#if HAVE_WORKING_POLL
#define USE_FUNC_POLL 1
#include <poll.h>
#else
#define USE_FUNC_POLL 0
#endif

/* Alessandro Rubini's GPM (general-purpose mouse) */
#if HAVE_LIBGPM && HAVE_GPM_H
#define USE_GPM_SUPPORT 1
#else
#define USE_GPM_SUPPORT 0
#endif

/* QNX mouse support */
#if defined(__QNX__) && !defined(__QNXNTO__)
#define USE_QNX_MOUSE 1
#else
#define USE_QNX_MOUSE 0
#endif

/* EMX mouse support */
#ifdef __EMX__
#define USE_EMX_MOUSE
#endif

#define DEFAULT_MAXCLICK 166
#define EV_MAX          8       /* size of mouse circular event queue */

/*
 * If we don't have signals to support it, don't add a sigwinch handler.
 * In any case, resizing is an extended feature.  Use it if we've got it.
 */
#if !NCURSES_EXT_FUNCS
#undef HAVE_SIZECHANGE
#endif

#if HAVE_SIZECHANGE
#define USE_SIZECHANGE 1
#else
#undef USE_SIGWINCH
#endif

/*
 * If desired, one can configure this, disabling environment variables that
 * point to custom terminfo/termcap locations.
 */
#define use_terminfo_vars() 1

#if USE_SCROLL_HINTS
#define if_USE_SCROLL_HINTS(stmt) stmt
#else
#define if_USE_SCROLL_HINTS(stmt) /*nothing*/
#endif

/* #include <curses.h> */
#ifndef __NCURSES_H
#define __NCURSES_H

#define CURSES 1
#define CURSES_H 1

/* This should be defined for the enhanced functionality to be visible.
 * However, none of the wide-character (enhanced) functionality is implemented.
 * So we do not define it (yet).
#define _XOPEN_CURSES 1
 */

/* These are defined only in curses.h, and are used for conditional compiles */
#define NCURSES_VERSION_MAJOR 5
#define NCURSES_VERSION_MINOR 2
#define NCURSES_VERSION_PATCH 20001021

/* This is defined in more than one ncurses header, for identification */
#undef  NCURSES_VERSION
#define NCURSES_VERSION "5.2"

#undef  NCURSES_COLOR_T
#define NCURSES_COLOR_T short

#undef  NCURSES_SIZE_T
#define NCURSES_SIZE_T short

/* XSI and SVr4 specify that curses implements 'bool'.  However, C++ may also
 * implement it.  If so, we must use the C++ compiler's type to avoid conflict
 * with other interfaces.
 */

#undef TRUE
#define TRUE    1

#undef FALSE
#define FALSE   0

#undef unctrl
static const char *unctrl(chtype);

#if !HAVE_TYPE_SIGACTION
typedef struct sigaction sigaction_t;
#endif

/*
 * XSI attributes.  In the ncurses implementation, they are identical to the
 * A_ attributes.
 */
#define WA_ATTRIBUTES   A_ATTRIBUTES
#define WA_NORMAL       A_NORMAL
#define WA_STANDOUT     A_STANDOUT
#define WA_UNDERLINE    A_UNDERLINE
#define WA_REVERSE      A_REVERSE
#define WA_BLINK        A_BLINK
#define WA_DIM          A_DIM
#define WA_BOLD         A_BOLD
#define WA_ALTCHARSET   A_ALTCHARSET
#define WA_INVIS        A_INVIS
#define WA_PROTECT      A_PROTECT
#define WA_HORIZONTAL   A_HORIZONTAL
#define WA_LEFT         A_LEFT
#define WA_LOW          A_LOW
#define WA_RIGHT        A_RIGHT
#define WA_TOP          A_TOP
#define WA_VERTICAL     A_VERTICAL

/* colors */
static int COLORS;
static int COLOR_PAIRS;

#define COLOR_BLACK     0
#define COLOR_RED       1
#define COLOR_GREEN     2
#define COLOR_YELLOW    3
#define COLOR_BLUE      4
#define COLOR_MAGENTA   5
#define COLOR_CYAN      6
#define COLOR_WHITE     7

/* VT100 symbols begin here */
#define ACS_ULCORNER    (acs_map['l'])  /* upper left corner */
#define ACS_LLCORNER    (acs_map['m'])  /* lower left corner */
#define ACS_URCORNER    (acs_map['k'])  /* upper right corner */
#define ACS_LRCORNER    (acs_map['j'])  /* lower right corner */
#define ACS_LTEE        (acs_map['t'])  /* tee pointing right */
#define ACS_RTEE        (acs_map['u'])  /* tee pointing left */
#define ACS_BTEE        (acs_map['v'])  /* tee pointing up */
#define ACS_TTEE        (acs_map['w'])  /* tee pointing down */
#define ACS_HLINE       (acs_map['q'])  /* horizontal line */
#define ACS_VLINE       (acs_map['x'])  /* vertical line */
#define ACS_PLUS        (acs_map['n'])  /* large plus or crossover */
#define ACS_S1          (acs_map['o'])  /* scan line 1 */
#define ACS_S9          (acs_map['s'])  /* scan line 9 */
#define ACS_DIAMOND     (acs_map['`'])  /* diamond */
#define ACS_CKBOARD     (acs_map['a'])  /* checker board (stipple) */
#define ACS_DEGREE      (acs_map['f'])  /* degree symbol */
#define ACS_PLMINUS     (acs_map['g'])  /* plus/minus */
#define ACS_BULLET      (acs_map['~'])  /* bullet */
/* Teletype 5410v1 symbols begin here */
#define ACS_LARROW      (acs_map[','])  /* arrow pointing left */
#define ACS_RARROW      (acs_map['+'])  /* arrow pointing right */
#define ACS_DARROW      (acs_map['.'])  /* arrow pointing down */
#define ACS_UARROW      (acs_map['-'])  /* arrow pointing up */
#define ACS_BOARD       (acs_map['h'])  /* board of squares */
#define ACS_LANTERN     (acs_map['i'])  /* lantern symbol */
#define ACS_BLOCK       (acs_map['0'])  /* solid square block */
/*
 * These aren't documented, but a lot of System Vs have them anyway
 * (you can spot pprryyzz{{||}} in a lot of AT&T terminfo strings).
 * The ACS_names may not match AT&T's, our source didn't know them.
 */
#define ACS_S3          (acs_map['p'])  /* scan line 3 */
#define ACS_S7          (acs_map['r'])  /* scan line 7 */
#define ACS_LEQUAL      (acs_map['y'])  /* less/equal */
#define ACS_GEQUAL      (acs_map['z'])  /* greater/equal */
#define ACS_PI          (acs_map['{'])  /* Pi */
#define ACS_NEQUAL      (acs_map['|'])  /* not equal */
#define ACS_STERLING    (acs_map['}'])  /* UK pound sign */

/*
 * Line drawing ACS names are of the form ACS_trbl, where t is the top, r
 * is the right, b is the bottom, and l is the left.  t, r, b, and l might
 * be B (blank), S (single), D (double), or T (thick).  The subset defined
 * here only uses B and S.
 */
#define ACS_BSSB        ACS_ULCORNER
#define ACS_SSBB        ACS_LLCORNER
#define ACS_BBSS        ACS_URCORNER
#define ACS_SBBS        ACS_LRCORNER
#define ACS_SBSS        ACS_RTEE
#define ACS_SSSB        ACS_LTEE
#define ACS_SSBS        ACS_BTEE
#define ACS_BSSS        ACS_TTEE
#define ACS_BSBS        ACS_HLINE
#define ACS_SBSB        ACS_VLINE
#define ACS_SSSS        ACS_PLUS

#if     !defined(ERR) || ((ERR) != -1)
#define ERR     (-1)
#endif

#if     !defined(OK) || ((OK) != 0)
#define OK      (0)
#endif

/* values for the _flags member */
#define _SUBWIN         0x01    /* is this a sub-window? */
#define _ENDLINE        0x02    /* is the window flush right? */
#define _FULLWIN        0x04    /* is the window full-screen? */
#define _SCROLLWIN      0x08    /* bottom edge is at screen bottom? */
#define _ISPAD          0x10    /* is this window a pad? */
#define _HASMOVED       0x20    /* has cursor moved since last refresh? */
#define _WRAPPED        0x40    /* cursor was just wrappped */

/*
 * this value is used in the firstchar and lastchar fields to mark
 * unchanged lines
 */
#define _NOCHANGE       -1

/*
 * this value is used in the oldindex field to mark lines created by insertions
 * and scrolls.
 */
#define _NEWINDEX       -1

typedef struct screen  SCREEN;

typedef chtype  attr_t;         /* ...must be at least as wide as chtype */

struct ldat
{
	chtype  *text;          /* text of the line */
	NCURSES_SIZE_T firstchar; /* first changed character in the line */
	NCURSES_SIZE_T lastchar; /* last changed character in the line */
	NCURSES_SIZE_T oldindex; /* index of the line at last update */
};

struct _win_st
{
	NCURSES_SIZE_T _cury, _curx; /* current cursor position */

	/* window location and size */
	NCURSES_SIZE_T _maxy, _maxx; /* maximums of x and y, NOT window size */
	NCURSES_SIZE_T _begy, _begx; /* screen coords of upper-left-hand corner */

	short   _flags;         /* window state flags */

	/* attribute tracking */
	attr_t  _attrs;         /* current attribute for non-space character */
	chtype  _bkgd;          /* current background char/attribute pair */

	/* option values set by user */
	bool    _notimeout;     /* no time out on function-key entry? */
	bool    _clear;         /* consider all data in the window invalid? */
	bool    _leaveok;       /* OK to not reset cursor on exit? */
	bool    _scroll;        /* OK to scroll this window? */
	bool    _idlok;         /* OK to use insert/delete line? */
	bool    _idcok;         /* OK to use insert/delete char? */
	bool    _immed;         /* window in immed mode? (not yet used) */
	bool    _sync;          /* window in sync mode? */
	bool    _use_keypad;    /* process function keys into KEY_ symbols? */
	int     _delay;         /* 0 = nodelay, <0 = blocking, >0 = delay */

	struct ldat *_line;     /* the actual line data */

	/* global screen state */
	NCURSES_SIZE_T _regtop; /* top line of scrolling region */
	NCURSES_SIZE_T _regbottom; /* bottom line of scrolling region */

	/* these are used only if this is a sub-window */
	int     _parx;          /* x coordinate of this window in parent */
	int     _pary;          /* y coordinate of this window in parent */
	WINDOW  *_parent;       /* pointer to parent if a sub-window */

	/* these are used only if this is a pad */
	struct pdat
	{
	    NCURSES_SIZE_T _pad_y,      _pad_x;
	    NCURSES_SIZE_T _pad_top,    _pad_left;
	    NCURSES_SIZE_T _pad_bottom, _pad_right;
	} _pad;

	NCURSES_SIZE_T _yoffset; /* real begy is _begy + _yoffset */
};


/*
 * This global was an undocumented feature under AIX curses.
 */
static int ESCDELAY;    /* ESC expire time in milliseconds */

/*
 * These functions are extensions - not in XSI Curses.
 */
static int keyok (int, bool);
static int resizeterm (int, int);
static int wresize (WINDOW *, int, int);

/*
 * GCC (and some other compilers) define '__attribute__'; we're using this
 * macro to alert the compiler to flag inconsistencies in printf/scanf-like
 * function calls.  Just in case '__attribute__' isn't defined, make a dummy.
 * G++ doesn't accept it anyway.
 */
#if !defined(__GNUC__) && !defined(__attribute__)
#define __attribute__(p) /* nothing */
#endif

/*
 * Function prototypes.  This is the complete XSI Curses list of required
 * functions.  Those marked `generated' will have sources generated from the
 * macro definitions later in this file, in order to satisfy XPG4.2
 * requirements.
 */

/*
 * OS/2's native linker complains if we don't initialize public data when
 * constructing a dll (reported by J.J.G.Ripoll).
 */
static WINDOW *stdscr;
static WINDOW *curscr;
static WINDOW *newscr;

static SCREEN *_nc_screen_chain;

static int cbreak(void);                                /* implemented */
static int delwin(WINDOW *);                            /* implemented */
static int doupdate(void);                              /* implemented */
static int endwin(void);                                /* implemented */
static int flash(void);                                 /* implemented */
static bool has_colors(void);                           /* implemented */
static WINDOW *initscr(void);                           /* implemented */
static int init_pair(short,short,short);                /* implemented */
static int keypad(WINDOW *,bool);                       /* implemented */
static WINDOW *newwin(int,int,int,int);                 /* implemented */
static int noecho(void);                                /* implemented */
static int scrollok(WINDOW *,bool);                     /* implemented */
static int start_color(void);                           /* implemented */
static WINDOW *subwin(WINDOW *,int,int,int,int);        /* implemented */
static int waddch(WINDOW *, const chtype);              /* implemented */
static int waddnstr(WINDOW *,const char *const,int);    /* implemented */
static void wbkgdset(WINDOW *,chtype);                  /* implemented */
static int wclrtoeol(WINDOW *);                         /* implemented */
static int wgetch(WINDOW *);                            /* implemented */
static chtype winch(WINDOW *);                          /* implemented */
static int wmove(WINDOW *,int,int);                     /* implemented */
static int wnoutrefresh(WINDOW *);                      /* implemented */
#define GCC_PRINTFLIKE(fmt,var) __attribute__((format(printf,fmt,var)))
static int wprintw(WINDOW *, const char *,...)  /* implemented */
		GCC_PRINTFLIKE(2,3);
static int wrefresh(WINDOW *);                          /* implemented */
static int wscrl(WINDOW *,int);                         /* implemented */
static int wtouchln(WINDOW *,int,int,int);              /* implemented */

static int baudrate(void);                              /* implemented */
static int beep(void);                                  /* implemented */
static int clearok(WINDOW *,bool);                      /* implemented */
static int COLOR_PAIR(int);                             /* generated */
static int curs_set(int);                               /* implemented */
static int def_prog_mode(void);                         /* implemented */
static int def_shell_mode(void);                        /* implemented */
static int delay_output(int);                           /* implemented */
static WINDOW *derwin(WINDOW *,int,int,int,int);        /* implemented */
static char erasechar(void);                            /* implemented */
static int flushinp(void);                              /* implemented */
static bool has_ic(void);                               /* implemented */
static bool is_wintouched(WINDOW *);                    /* implemented */
static char killchar(void);                             /* implemented */
static int mvcur(int,int,int,int);                      /* implemented */
static int mvwaddch(WINDOW *, int, int, const chtype);  /* generated */
static int mvwhline(WINDOW *, int, int, chtype, int);   /* generated */
static int napms(int);                                  /* implemented */
static SCREEN *newterm(const char *,FILE *,FILE *);     /* implemented */
static int nl(void);                                    /* implemented */
static int noraw(void);                                 /* implemented */
static int pair_content(short,short*,short*);           /* implemented */
static int PAIR_NUMBER(int);                            /* generated */
static int putp(const char *);                          /* implemented */
static int reset_prog_mode(void);                       /* implemented */
static int reset_shell_mode(void);                      /* implemented */
static SCREEN *set_term(SCREEN *);                      /* implemented */
static chtype termattrs(void);                          /* implemented */
static int touchline(WINDOW *, int, int);               /* generated */
static char *tparm(const char *, ...);          /* implemented */
static int typeahead(int);                              /* implemented */
static int ungetch(int);                                /* implemented */
static int vidattr(chtype);                             /* implemented */
static int vidputs(chtype, int (*)(int));               /* implemented */
static int vwprintw(WINDOW *, const char *,va_list);    /* implemented */
static int waddstr(WINDOW *,const char *);              /* generated */
static int wechochar(WINDOW *, const chtype);           /* implemented */
static int wgetnstr(WINDOW *,char *,int);               /* implemented */
static int whline(WINDOW *, chtype, int);               /* implemented */
static void wsyncdown(WINDOW *);                        /* implemented */
static void wsyncup(WINDOW *);                          /* implemented */

/* attributes */

#define NCURSES_BITS(mask,shift) ((mask) << ((shift) + 8))

#define A_NORMAL        0L
#define A_ATTRIBUTES    NCURSES_BITS(~(1UL - 1UL),0)
#define A_CHARTEXT      (NCURSES_BITS(1UL,0) - 1UL)
#define A_COLOR         NCURSES_BITS(((1UL) << 8) - 1UL,0)
#define A_STANDOUT      NCURSES_BITS(1UL,8)
#define A_UNDERLINE     NCURSES_BITS(1UL,9)
#define A_REVERSE       NCURSES_BITS(1UL,10)
#define A_BLINK         NCURSES_BITS(1UL,11)
#define A_DIM           NCURSES_BITS(1UL,12)
#define A_BOLD          NCURSES_BITS(1UL,13)
#define A_ALTCHARSET    NCURSES_BITS(1UL,14)
#define A_INVIS         NCURSES_BITS(1UL,15)

/* Tradeoff on 32-bit machines ('protect' vs widec).  The others (e.g., left
 * highlight are not implemented in any terminal descriptions, anyway.
 */
#if ((16 + 8) < 64)
#define A_PROTECT       NCURSES_BITS(1UL,16)
#define A_HORIZONTAL    NCURSES_BITS(1UL,17)
#define A_LEFT          NCURSES_BITS(1UL,18)
#define A_LOW           NCURSES_BITS(1UL,19)
#define A_RIGHT         NCURSES_BITS(1UL,20)
#define A_TOP           NCURSES_BITS(1UL,21)
#define A_VERTICAL      NCURSES_BITS(1UL,22)
#else
#define A_PROTECT       0L
#define A_HORIZONTAL    0L
#define A_LEFT          0L
#define A_LOW           0L
#define A_RIGHT         0L
#define A_TOP           0L
#define A_VERTICAL      0L
#endif

#define COLOR_PAIR(n)   NCURSES_BITS(n, 0)
#define PAIR_NUMBER(a)  (((a) & A_COLOR) >> 8)

/*
 * pseudo functions
 */
#define setterm(term)           setupterm(term, 1, (int *)0)

#define fixterm()               reset_prog_mode()
#define resetterm()             reset_shell_mode()
#define saveterm()              def_prog_mode()
#define crmode()                cbreak()
#define nocrmode()              nocbreak()
#define gettmode()

#define getyx(win,y,x)          (y = (win)?(win)->_cury:ERR, x = (win)?(win)->_curx:ERR)
#define getbegyx(win,y,x)       (y = (win)?(win)->_begy:ERR, x = (win)?(win)->_begx:ERR)
#define getmaxyx(win,y,x)       (y = (win)?((win)->_maxy + 1):ERR, x = (win)?((win)->_maxx + 1):ERR)
#define getparyx(win,y,x)       (y = (win)?(win)->_pary:ERR, x = (win)?(win)->_parx:ERR)
#define getsyx(y,x) do { if(newscr->_leaveok) (y)=(x)=-1; \
			 else getyx(newscr,(y),(x)); \
		    } while(0)
#define setsyx(y,x) do { if((y)==-1 && (x)==-1) newscr->_leaveok=TRUE; \
			 else {newscr->_leaveok=FALSE;wmove(newscr,(y),(x));} \
		    } while(0)

/* It seems older SYSV curses versions define these */
#define getattrs(win)           ((win)?(win)->_attrs:A_NORMAL)
#define getcurx(win)            ((win)?(win)->_curx:ERR)
#define getcury(win)            ((win)?(win)->_cury:ERR)
#define getbegx(win)            ((win)?(win)->_begx:ERR)
#define getbegy(win)            ((win)?(win)->_begy:ERR)
#define getmaxx(win)            ((win)?((win)->_maxx + 1):ERR)
#define getmaxy(win)            ((win)?((win)->_maxy + 1):ERR)
#define getparx(win)            ((win)?(win)->_parx:ERR)
#define getpary(win)            ((win)?(win)->_pary:ERR)

#define scroll(win)             wscrl(win,1)

#define touchline(win, s, c)    wtouchln((win), s, c, 1)

#define waddstr(win,str)        waddnstr(win,str,-1)

/*
 * mv functions
 */

#define mvwaddch(win,y,x,ch)            (wmove(win,y,x) == ERR ? ERR : waddch(win,ch))
#define mvwhline(win,y,x,c,n)           (wmove(win,y,x) == ERR ? ERR : whline(win,c,n))

/*
 * XSI curses macros for XPG4 conformance.
 * The underlying functions needed to make these work are:
 * waddnwstr(), waddchnwstr(), wadd_wch(), wborder_set(), wchgat(),
 * wecho_wchar(), wgetn_wstr(), wget_wch(), whline_set(), vhline_set(),
 * winnwstr(), wins_nwstr(), wins_wch(), win_wch(), win_wchnstr().
 * Except for wchgat(), these are not yet implemented.  They will be someday.
 */
#define add_wch(c)                      wadd_wch(stdscr,c)
#define addnwstr(wstr,n)                waddnwstr(stdscr,wstr,n)
#define addwstr(wstr,n)                 waddnwstr(stdscr,wstr,-1)
#define box_set(w,v,h)                  wborder_set(w,v,v,h,h,0,0,0,0)
#define echo_wchar(c)                   wecho_wchar(stdscr,c)
#define get_wch(c)                      wget_wch(stdscr,c)
#define get_wstr(t)                     wgetn_wstr(stdscr,t,-1)
#define getn_wstr(t,n)                  wgetn_wstr(stdscr,t,n)
#define hline_set(c,n)                  whline_set(stdscr,c,n)
#define in_wch(c)                       win_wch(stdscr,c)
#define in_wchnstr(c,n)                 win_wchnstr(stdscr,c,n)
#define in_wchstr(c)                    win_wchnstr(stdscr,c,-1)
#define innwstr(c,n)                    winnwstr(stdscr,c,n)
#define ins_nwstr(t,n)                  wins_nwstr(stdscr,t,n)
#define ins_wch(c)                      wins_wch(stdscr,c)
#define ins_wstr(t)                     wins_nwstr(stdscr,t,-1)
#define inwstr(c)                       winnwstr(stdscr,c,-1)

#define mvadd_wch(y,x,c)                mvwadd_wch(stdscr,y,x,c)
#define mvaddnwstr(y,x,wstr,n)          mvwaddnwstr(stdscr,y,x,wstr,n)
#define mvaddwstr(y,x,wstr,n)           mvwaddnwstr(stdscr,y,x,wstr,-1)
#define mvget_wch(y,x,c)                mvwget_wch(stdscr,y,x,c)
#define mvget_wstr(y,x,t)               mvwgetn_wstr(stdscr,y,x,t,-1)
#define mvgetn_wstr(y,x,t,n)            mvwgetn_wstr(stdscr,y,x,t,n)
#define mvhline_set(y,x,c,n)            mvwhline_set(stdscr,y,x,c,n)
#define mvin_wch(y,x,c)                 mvwin_wch(stdscr,y,x,c)
#define mvin_wchnstr(y,x,c,n)           mvwin_wchnstr(stdscr,y,x,c,n)
#define mvin_wchstr(y,x,c)              mvwin_wchnstr(stdscr,y,x,c,-1)
#define mvinnwstr(y,x,c,n)              mvwinnwstr(stdscr,y,x,c,n)
#define mvins_nwstr(y,x,t,n)            mvwins_nwstr(stdscr,y,x,t,n)
#define mvins_wch(y,x,c)                mvwins_wch(stdscr,y,x,c)
#define mvins_wstr(y,x,t)               mvwins_nwstr(stdscr,y,x,t,-1)
#define mvinwstr(y,x,c)                 mvwinnwstr(stdscr,y,x,c,-1)
#define mvvline_set(y,x,c,n)            mvwvline_set(stdscr,y,x,c,n)

#define mvwadd_wch(win,y,x,c)           (wmove(win,y,x) == ERR ? ERR : wadd_wch(stdscr,c))
#define mvwaddnwstr(win,y,x,wstr,n)     (wmove(win,y,x) == ERR ? ERR : waddnwstr(stdscr,wstr,n))
#define mvwaddwstr(win,y,x,wstr,n)      (wmove(win,y,x) == ERR ? ERR : waddnwstr(stdscr,wstr,-1))
#define mvwget_wch(win,y,x,c)           (wmove(win,y,x) == ERR ? ERR : wget_wch(win,c))
#define mvwget_wstr(win,y,x,t)          (wmove(win,y,x) == ERR ? ERR : wgetn_wstr(win,t,-1))
#define mvwgetn_wstr(win,y,x,t,n)       (wmove(win,y,x) == ERR ? ERR : wgetn_wstr(win,t,n))
#define mvwhline_set(win,y,x,c,n)       (wmove(win,y,x) == ERR ? ERR : whline_set(win,c,n))
#define mvwin_wch(win,y,x,c)            (wmove(win,y,x) == ERR ? ERR : win_wch(win,c))
#define mvwin_wchnstr(win,y,x,c,n)      (wmove(win,y,x) == ERR ? ERR : win_wchnstr(stdscr,c,n))
#define mvwin_wchstr(win,y,x,c)         (wmove(win,y,x) == ERR ? ERR : win_wchnstr(stdscr,c,-1))
#define mvwinnwstr(win,y,x,c,n)         (wmove(win,y,x) == ERR ? ERR : winnwstr(stdscr,c,n))
#define mvwins_nwstr(win,y,x,t,n)       (wmove(win,y,x) == ERR ? ERR : wins_nwstr(stdscr,t,n))
#define mvwins_wch(win,y,x,c)           (wmove(win,y,x) == ERR ? ERR : wins_wch(stdscr,c))
#define mvwins_wstr(win,y,x,t)          (wmove(win,y,x) == ERR ? ERR : wins_nwstr(stdscr,t,-1))
#define mvwinwstr(win,y,x,c)            (wmove(win,y,x) == ERR ? ERR : winnwstr(stdscr,c,-1))
#define mvwvline_set(win,y,x,c,n)       (wmove(win,y,x) == ERR ? ERR : wvline_set(win,c,n))

#define vline_set(c,n)                  wvline_set(stdscr,c,n)
#define waddwstr(win,wstr,n)            waddnwstr(win,wstr,-1)
#define wget_wstr(w,t)                  wgetn_wstr(w,t,-1)
#define win_wchstr(w,c)                 win_wchnstr(w,c,-1)
#define wins_wstr(w,t)                  wins_nwstr(w,t,-1)
#define winwstr(w,c)                    winnwstr(w,c,-1)


/*
 * Pseudo-character tokens outside ASCII range.  The curses wgetch() function
 * will return any given one of these only if the corresponding k- capability
 * is defined in your terminal's terminfo entry.
 */
#define KEY_CODE_YES    0400            /* A wchar_t contains a key code */
#define KEY_MIN         0401            /* Minimum curses key */
#define KEY_BREAK       0401            /* Break key (unreliable) */
#define KEY_DOWN        0402            /* Down-arrow */
#define KEY_UP          0403            /* Up-arrow */
#define KEY_LEFT        0404            /* Left-arrow */
#define KEY_RIGHT       0405            /* Right-arrow */
#define KEY_HOME        0406            /* Home key (upward+left arrow) */
#define KEY_BACKSPACE   0407            /* Backspace (unreliable) */
#define KEY_F0          0410            /* Function keys.  Space for 64 */
#define KEY_F(n)        (KEY_F0+(n))    /* Value of function key n */
#define KEY_DL          0510            /* Delete line */
#define KEY_IL          0511            /* Insert line */
#define KEY_DC          0512            /* Delete character */
#define KEY_IC          0513            /* Insert char or enter insert mode */
#define KEY_EIC         0514            /* Exit insert char mode */
#define KEY_CLEAR       0515            /* Clear screen */
#define KEY_EOS         0516            /* Clear to end of screen */
#define KEY_EOL         0517            /* Clear to end of line */
#define KEY_SF          0520            /* Scroll 1 line forward */
#define KEY_SR          0521            /* Scroll 1 line backward (reverse) */
#define KEY_NPAGE       0522            /* Next page */
#define KEY_PPAGE       0523            /* Previous page */
#define KEY_STAB        0524            /* Set tab */
#define KEY_CTAB        0525            /* Clear tab */
#define KEY_CATAB       0526            /* Clear all tabs */
#define KEY_ENTER       0527            /* Enter or send (unreliable) */
#define KEY_SRESET      0530            /* Soft (partial) reset (unreliable) */
#define KEY_RESET       0531            /* Reset or hard reset (unreliable) */
#define KEY_PRINT       0532            /* Print */
#define KEY_LL          0533            /* Home down or bottom (lower left) */

/* The keypad is arranged like this: */
/* a1    up    a3   */
/* left   b2  right  */
/* c1   down   c3   */

#define KEY_A1          0534            /* Upper left of keypad */
#define KEY_A3          0535            /* Upper right of keypad */
#define KEY_B2          0536            /* Center of keypad */
#define KEY_C1          0537            /* Lower left of keypad */
#define KEY_C3          0540            /* Lower right of keypad */
#define KEY_BTAB        0541            /* Back tab */
#define KEY_BEG         0542            /* Beg (beginning) */
#define KEY_CANCEL      0543            /* Cancel */
#define KEY_CLOSE       0544            /* Close */
#define KEY_COMMAND     0545            /* Cmd (command) */
#define KEY_COPY        0546            /* Copy */
#define KEY_CREATE      0547            /* Create */
#define KEY_END         0550            /* End */
#define KEY_EXIT        0551            /* Exit */
#define KEY_FIND        0552            /* Find */
#define KEY_HELP        0553            /* Help */
#define KEY_MARK        0554            /* Mark */
#define KEY_MESSAGE     0555            /* Message */
#define KEY_MOVE        0556            /* Move */
#define KEY_NEXT        0557            /* Next */
#define KEY_OPEN        0560            /* Open */
#define KEY_OPTIONS     0561            /* Options */
#define KEY_PREVIOUS    0562            /* Prev (previous) */
#define KEY_REDO        0563            /* Redo */
#define KEY_REFERENCE   0564            /* Ref (reference) */
#define KEY_REFRESH     0565            /* Refresh */
#define KEY_REPLACE     0566            /* Replace */
#define KEY_RESTART     0567            /* Restart */
#define KEY_RESUME      0570            /* Resume */
#define KEY_SAVE        0571            /* Save */
#define KEY_SBEG        0572            /* Shifted Beg (beginning) */
#define KEY_SCANCEL     0573            /* Shifted Cancel */
#define KEY_SCOMMAND    0574            /* Shifted Command */
#define KEY_SCOPY       0575            /* Shifted Copy */
#define KEY_SCREATE     0576            /* Shifted Create */
#define KEY_SDC         0577            /* Shifted Delete char */
#define KEY_SDL         0600            /* Shifted Delete line */
#define KEY_SELECT      0601            /* Select */
#define KEY_SEND        0602            /* Shifted End */
#define KEY_SEOL        0603            /* Shifted Clear line */
#define KEY_SEXIT       0604            /* Shifted Dxit */
#define KEY_SFIND       0605            /* Shifted Find */
#define KEY_SHELP       0606            /* Shifted Help */
#define KEY_SHOME       0607            /* Shifted Home */
#define KEY_SIC         0610            /* Shifted Input */
#define KEY_SLEFT       0611            /* Shifted Left arrow */
#define KEY_SMESSAGE    0612            /* Shifted Message */
#define KEY_SMOVE       0613            /* Shifted Move */
#define KEY_SNEXT       0614            /* Shifted Next */
#define KEY_SOPTIONS    0615            /* Shifted Options */
#define KEY_SPREVIOUS   0616            /* Shifted Prev */
#define KEY_SPRINT      0617            /* Shifted Print */
#define KEY_SREDO       0620            /* Shifted Redo */
#define KEY_SREPLACE    0621            /* Shifted Replace */
#define KEY_SRIGHT      0622            /* Shifted Right arrow */
#define KEY_SRSUME      0623            /* Shifted Resume */
#define KEY_SSAVE       0624            /* Shifted Save */
#define KEY_SSUSPEND    0625            /* Shifted Suspend */
#define KEY_SUNDO       0626            /* Shifted Undo */
#define KEY_SUSPEND     0627            /* Suspend */
#define KEY_UNDO        0630            /* Undo */
#define KEY_MOUSE       0631            /* Mouse event has occurred */
#define KEY_RESIZE      0632            /* Terminal resize event */
#define KEY_MAX         0777            /* Maximum key value */

/* mouse interface */
#define NCURSES_MOUSE_VERSION   1

/* event masks */
#define BUTTON1_RELEASED        000000000001L
#define BUTTON1_PRESSED         000000000002L
#define BUTTON1_CLICKED         000000000004L
#define BUTTON1_DOUBLE_CLICKED  000000000010L
#define BUTTON1_TRIPLE_CLICKED  000000000020L
#define BUTTON1_RESERVED_EVENT  000000000040L
#define BUTTON2_RELEASED        000000000100L
#define BUTTON2_PRESSED         000000000200L
#define BUTTON2_CLICKED         000000000400L
#define BUTTON2_DOUBLE_CLICKED  000000001000L
#define BUTTON2_TRIPLE_CLICKED  000000002000L
#define BUTTON2_RESERVED_EVENT  000000004000L
#define BUTTON3_RELEASED        000000010000L
#define BUTTON3_PRESSED         000000020000L
#define BUTTON3_CLICKED         000000040000L
#define BUTTON3_DOUBLE_CLICKED  000000100000L
#define BUTTON3_TRIPLE_CLICKED  000000200000L
#define BUTTON3_RESERVED_EVENT  000000400000L
#define BUTTON4_RELEASED        000001000000L
#define BUTTON4_PRESSED         000002000000L
#define BUTTON4_CLICKED         000004000000L
#define BUTTON4_DOUBLE_CLICKED  000010000000L
#define BUTTON4_TRIPLE_CLICKED  000020000000L
#define BUTTON4_RESERVED_EVENT  000040000000L
#define BUTTON_CTRL             000100000000L
#define BUTTON_SHIFT            000200000000L
#define BUTTON_ALT              000400000000L
#define ALL_MOUSE_EVENTS        000777777777L
#define REPORT_MOUSE_POSITION   001000000000L

/* macros to extract single event-bits from masks */
#define BUTTON_RELEASE(e, x)            ((e) & (001 << (6 * ((x) - 1))))
#define BUTTON_PRESS(e, x)              ((e) & (002 << (6 * ((x) - 1))))
#define BUTTON_CLICK(e, x)              ((e) & (004 << (6 * ((x) - 1))))
#define BUTTON_DOUBLE_CLICK(e, x)       ((e) & (010 << (6 * ((x) - 1))))
#define BUTTON_TRIPLE_CLICK(e, x)       ((e) & (020 << (6 * ((x) - 1))))
#define BUTTON_RESERVED_EVENT(e, x)     ((e) & (040 << (6 * ((x) - 1))))

typedef unsigned long mmask_t;

typedef struct
{
    short id;           /* ID to distinguish multiple devices */
    int x, y, z;        /* event coordinates (character-cell) */
    mmask_t bstate;     /* button state bits */
}
MEVENT;

/* Debugging : use with libncurses_g.a */

/* trace masks */
#define TRACE_DISABLE   0x0000  /* turn off tracing */
#define TRACE_TIMES     0x0001  /* trace user and system times of updates */
#define TRACE_TPUTS     0x0002  /* trace tputs calls */
#define TRACE_UPDATE    0x0004  /* trace update actions, old & new screens */
#define TRACE_MOVE      0x0008  /* trace cursor moves and scrolls */
#define TRACE_CHARPUT   0x0010  /* trace all character outputs */
#define TRACE_ORDINARY  0x001F  /* trace all update actions */
#define TRACE_CALLS     0x0020  /* trace all curses calls */
#define TRACE_VIRTPUT   0x0040  /* trace virtual character puts */
#define TRACE_IEVENT    0x0080  /* trace low-level input processing */
#define TRACE_BITS      0x0100  /* trace state of TTY control bits */
#define TRACE_ICALLS    0x0200  /* trace internal/nested calls */
#define TRACE_CCALLS    0x0400  /* trace per-character calls */
#define TRACE_DATABASE  0x0800  /* trace read/write of terminfo/termcap data */
#define TRACE_ATTRS     0x1000  /* trace attribute updates */
#define TRACE_MAXIMUM   0xffff  /* maximum trace level */

#endif /* __NCURSES_H */
/* end #include <curses.h> */


/*
 * Note:  ht/cbt expansion flakes out randomly under Linux 1.1.47, but only
 * when we're throwing control codes at the screen at high volume.  To see
 * this, re-enable USE_HARD_TABS and run worm for a while.  Other systems
 * probably don't want to define this either due to uncertainties about tab
 * delays and expansion in raw mode.
 */

struct tries {
	struct tries    *child;     /* ptr to child.  NULL if none          */
	struct tries    *sibling;   /* ptr to sibling.  NULL if none        */
	unsigned char    ch;        /* character at this node               */
	unsigned short   value;     /* code of string so far.  0 if none.   */
};

/*
 * Definitions for color pairs
 */
#define C_SHIFT 8               /* we need more bits than there are colors */
#define C_MASK  ((1 << C_SHIFT) - 1)

#define PAIR_OF(fg, bg) ((((fg) & C_MASK) << C_SHIFT) | ((bg) & C_MASK))

/*
 * Common/troublesome character definitions
 */
#define L_BRACE '{'
#define R_BRACE '}'
#define S_QUOTE '\''

/*
 * Structure for palette tables
 */

typedef struct
{
    short red, green, blue;
}
color_t;

#define MAXCOLUMNS    135
#define MAXLINES      66
#define FIFO_SIZE     MAXCOLUMNS+2  /* for nocbreak mode input */

#define ACS_LEN       128

#define WINDOWLIST struct _win_list

/*
 * Structure for soft labels.
 */

typedef struct
{
	char *text;             /* text for the label */
	char *form_text;        /* formatted text (left/center/...) */
	int x;                  /* x coordinate of this field */
	char dirty;             /* this label has changed */
	char visible;           /* field is visible */
} slk_ent;

typedef struct {
	char dirty;             /* all labels have changed */
	char hidden;            /* soft labels are hidden */
	struct _win_st *win;
	slk_ent *ent;
	char*  buffer;           /* buffer for labels */
	short  maxlab;           /* number of available labels */
	short  labcnt;           /* number of allocated labels */
	short  maxlen;           /* length of labels */
	chtype attr;             /* soft label attribute */
} SLK;

struct screen {
	int             _ifd;           /* input file ptr for screen        */
	FILE            *_ofp;          /* output file ptr for screen       */
	char            *_setbuf;       /* buffered I/O for output          */
	int             _buffered;      /* setvbuf uses _setbuf data        */
	int             _checkfd;       /* filedesc for typeahead check     */
	struct term     *_term;         /* terminal type information        */
	short           _lines;         /* screen lines                     */
	short           _columns;       /* screen columns                   */
	short           _lines_avail;   /* lines available for stdscr       */
	short           _topstolen;     /* lines stolen from top            */

	WINDOW          *_curscr;       /* current screen                   */
	WINDOW          *_newscr;       /* virtual screen to be updated to  */
	WINDOW          *_stdscr;       /* screen's full-window context     */

	struct tries    *_keytry;       /* "Try" for use with keypad mode   */
	struct tries    *_key_ok;       /* Disabled keys via keyok(,FALSE)  */
	int             _tried;         /* keypad mode was initialized      */

	unsigned int    _fifo[FIFO_SIZE];       /* input push-back buffer   */
	short           _fifohead,      /* head of fifo queue               */
			_fifotail,      /* tail of fifo queue               */
			_fifopeek,      /* where to peek for next char      */
			_fifohold;      /* set if breakout marked           */

	int             _endwin;        /* are we out of window mode?       */
	unsigned long   _current_attr;  /* terminal attribute current set   */
	int             _coloron;       /* is color enabled?                */
	int             _cursor;        /* visibility of the cursor         */
	int             _cursrow;       /* physical cursor row              */
	int             _curscol;       /* physical cursor column           */
	int             _nl;            /* True if NL -> CR/NL is on        */
	int             _raw;           /* True if in raw mode              */
	int             _cbreak;        /* 1 if in cbreak mode              */
					/* > 1 if in halfdelay mode         */
	int             _echo;          /* True if echo on                  */
	int             _use_meta;      /* use the meta key?                */
	SLK             *_slk;          /* ptr to soft key struct / NULL    */
	int             slk_format;     /* selected format for this screen  */
	/* cursor movement costs; units are 10ths of milliseconds */
	int             _no_padding;    /* flag to set if padding disabled  */
	int             _char_padding;  /* cost of character put            */
	int             _cr_cost;       /* cost of (carriage_return)        */
	int             _cup_cost;      /* cost of (cursor_address)         */
	int             _home_cost;     /* cost of (cursor_home)            */
	int             _ll_cost;       /* cost of (cursor_to_ll)           */
	int             _cub1_cost;     /* cost of (cursor_left)            */
	int             _cuf1_cost;     /* cost of (cursor_right)           */
	int             _cud1_cost;     /* cost of (cursor_down)            */
	int             _cuu1_cost;     /* cost of (cursor_up)              */
	int             _cub_cost;      /* cost of (parm_cursor_left)       */
	int             _cuf_cost;      /* cost of (parm_cursor_right)      */
	int             _cud_cost;      /* cost of (parm_cursor_down)       */
	int             _cuu_cost;      /* cost of (parm_cursor_up)         */
	int             _hpa_cost;      /* cost of (column_address)         */
	int             _vpa_cost;      /* cost of (row_address)            */
	/* used in tty_update.c, must be chars */
	int             _ed_cost;       /* cost of (clr_eos)                */
	int             _el_cost;       /* cost of (clr_eol)                */
	int             _el1_cost;      /* cost of (clr_bol)                */
	int             _dch1_cost;     /* cost of (delete_character)       */
	int             _ich1_cost;     /* cost of (insert_character)       */
	int             _dch_cost;      /* cost of (parm_dch)               */
	int             _ich_cost;      /* cost of (parm_ich)               */
	int             _ech_cost;      /* cost of (erase_chars)            */
	int             _rep_cost;      /* cost of (repeat_char)            */
	int             _hpa_ch_cost;   /* cost of (column_address)         */
	int             _cup_ch_cost;   /* cost of (cursor_address)         */
	int             _cuf_ch_cost;   /* cost of (parm_cursor_right)      */
	int             _inline_cost;   /* cost of inline-move              */
	int             _smir_cost;     /* cost of (enter_insert_mode)      */
	int             _rmir_cost;     /* cost of (exit_insert_mode)       */
	int             _ip_cost;       /* cost of (insert_padding)         */
	/* used in lib_mvcur.c */
	char *          _address_cursor;
	/* used in tty_update.c */
	int             _scrolling;     /* 1 if terminal's smart enough to  */

	/* used in lib_color.c */
	color_t         *_color_table;  /* screen's color palette            */
	int             _color_count;   /* count of colors in palette        */
	unsigned short  *_color_pairs;  /* screen's color pair list          */
	int             _pair_count;    /* count of color pairs              */
#if NCURSES_EXT_FUNCS
	bool            _default_color; /* use default colors                */
	bool            _has_sgr_39_49; /* has ECMA default color support    */
	int             _default_fg;    /* assumed default foreground        */
	int             _default_bg;    /* assumed default background        */
#endif
	chtype          _xmc_suppress;  /* attributes to suppress if xmc     */
	chtype          _xmc_triggers;  /* attributes to process if xmc      */
	chtype          _acs_map[ACS_LEN];

	/* used in lib_vidattr.c */
	bool            _use_rmso;      /* true if we may use 'rmso'         */
	bool            _use_rmul;      /* true if we may use 'rmul'         */

	/*
	 * These data correspond to the state of the idcok() and idlok()
	 * functions.  A caveat is in order here:  the XSI and SVr4
	 * documentation specify that these functions apply to the window which
	 * is given as an argument.  However, ncurses implements this logic
	 * only for the newscr/curscr update process, _not_ per-window.
	 */
	bool            _nc_sp_idlok;
	bool            _nc_sp_idcok;
#define _nc_idlok SP->_nc_sp_idlok
#define _nc_idcok SP->_nc_sp_idcok

	/*
	 * These are the data that support the mouse interface.
	 */
	int             _maxclick;
	bool            (*_mouse_event) (SCREEN *);
	bool            (*_mouse_inline)(SCREEN *);
	bool            (*_mouse_parse) (int);
	void            (*_mouse_resume)(SCREEN *);
	void            (*_mouse_wrap)  (SCREEN *);
	int             _mouse_fd;      /* file-descriptor, if any */

	/*
	 * This supports automatic resizing
	 */
#if USE_SIZECHANGE
	int             (*_resize)(int,int);
#endif

	/*
	 * Linked-list of all windows, to support '_nc_resizeall()' and
	 * '_nc_freeall()'
	 */
	WINDOWLIST      *_nc_sp_windows;
#define _nc_windows SP->_nc_sp_windows

	bool            _sig_winch;
	SCREEN          *_next_screen;

	/* hashes for old and new lines */
	unsigned long   *oldhash, *newhash;

	bool            _cleanup;       /* cleanup after int/quit signal */
	int             (*_outch)(int); /* output handler if not putc */
};

static SCREEN *_nc_screen_chain;

WINDOWLIST {
	WINDOWLIST *next;
	WINDOW  *win;
};

typedef struct {
	int     line;                   /* lines to take, < 0 => from bottom*/
	int     (*hook)(struct _win_st *, int); /* callback for user        */
	struct _win_st *w;              /* maybe we need this for cleanup   */
} ripoff_t;

/* The terminfo source is assumed to be 7-bit ASCII */
#define is7bits(c)      ((unsigned)(c) < 128)

#ifndef min
#define min(a,b)        ((a) > (b)  ?  (b)  :  (a))
#endif

#ifndef max
#define max(a,b)        ((a) < (b)  ?  (b)  :  (a))
#endif

/* usually in <unistd.h> */
#ifndef STDIN_FILENO
#define STDIN_FILENO 0
#endif

#ifndef STDOUT_FILENO
#define STDOUT_FILENO 1
#endif

#ifndef STDERR_FILENO
#define STDERR_FILENO 2
#endif

#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 0
#endif

#ifndef EXIT_FAILURE
#define EXIT_FAILURE 1
#endif

#ifndef R_OK
#define R_OK    4               /* Test for read permission.  */
#endif
#ifndef W_OK
#define W_OK    2               /* Test for write permission.  */
#endif
#ifndef X_OK
#define X_OK    1               /* Test for execute permission.  */
#endif
#ifndef F_OK
#define F_OK    0               /* Test for existence.  */
#endif

#ifndef O_BINARY
#define O_BINARY 0
#endif

#define TextOf(c)    ((c) & (chtype)A_CHARTEXT)
#define AttrOf(c)    ((c) & (chtype)A_ATTRIBUTES)

#define BLANK        (' '|A_NORMAL)

#define CHANGED     -1

#define CHANGED_CELL(line,col) \
	if (line->firstchar == _NOCHANGE) \
		line->firstchar = line->lastchar = col; \
	else if ((col) < line->firstchar) \
		line->firstchar = col; \
	else if ((col) > line->lastchar) \
		line->lastchar = col

#define CHANGED_RANGE(line,start,end) \
	if (line->firstchar == _NOCHANGE \
	 || line->firstchar > (start)) \
		line->firstchar = start; \
	if (line->lastchar == _NOCHANGE \
	 || line->lastchar < (end)) \
		line->lastchar = end

#define CHANGED_TO_EOL(line,start,end) \
	if (line->firstchar == _NOCHANGE \
	 || line->firstchar > (start)) \
		line->firstchar = start; \
	line->lastchar = end

#define SIZEOF(v) (sizeof(v)/sizeof(v[0]))

#define FreeIfNeeded(p)  if ((p) != 0) free(p)

/* FreeAndNull() is not a comma-separated expression because some compilers
 * do not accept a mixture of void with values.
 */
#define FreeAndNull(p)   free(p); p = 0

/* #include <nc_alloc.h> */

#ifndef ExitProgram
#define ExitProgram(code) return code
#endif

/* doalloc.c */
static void *_nc_doalloc(void *, size_t);

#define typeMalloc(type,elts) (type *)malloc((elts)*sizeof(type))
#define typeCalloc(type,elts) (type *)calloc((elts),sizeof(type))
#define typeRealloc(type,elts,ptr) (type *)_nc_doalloc(ptr, (elts)*sizeof(type))
/* end #include <nc_alloc.h> */


/*
 * Prefixes for call/return points of library function traces.  We use these to
 * instrument the public functions so that the traces can be easily transformed
 * into regression scripts.
 */
#define T_CALLED(fmt) "called " fmt
#define T_CREATE(fmt) "create " fmt
#define T_RETURN(fmt) "return " fmt

#define T(a)
#define TR(n, a)
#define returnAttr(code) return code
#define returnChar(code) return code
#define returnCode(code) return code
#define returnPtr(code)  return code
#define returnVoid       return
#define returnWin(code)  return code

#define _trace_key(ch) ((ch > KEY_MIN) ? keyname(ch) : _tracechar((unsigned char)ch))

#define ALL_BUT_COLOR ((chtype)~(A_COLOR))
#define IGNORE_COLOR_OFF FALSE
#define NONBLANK_ATTR (A_BOLD|A_DIM|A_BLINK)
#define XMC_CHANGES(c) ((c) & SP->_xmc_suppress)


#define toggle_attr_on(S,at) {\
   if (PAIR_NUMBER(at) > 0)\
      (S) = ((S) & ALL_BUT_COLOR) | (at);\
   else\
      (S) |= (at);\
   TR(TRACE_ATTRS, ("new attribute is %s", _traceattr((S))));}


#define toggle_attr_off(S,at) {\
   if (IGNORE_COLOR_OFF == TRUE) {\
      if (PAIR_NUMBER(at) == 0xff) /* turn off color */\
	 (S) &= ~(at);\
      else /* leave color alone */\
	 (S) &= ~((at)&ALL_BUT_COLOR);\
   } else {\
      if (PAIR_NUMBER(at) > 0x00) /* turn off color */\
	 (S) &= ~(at|A_COLOR);\
      else /* leave color alone */\
	 (S) &= ~(at);\
   }\
   TR(TRACE_ATTRS, ("new attribute is %s", _traceattr((S))));}

#define DelCharCost(count) \
		((parm_dch != 0) \
		? SP->_dch_cost \
		: ((delete_character != 0) \
			? (SP->_dch1_cost * count) \
			: INFINITY))

#define InsCharCost(count) \
		((parm_ich != 0) \
		? SP->_ich_cost \
		: ((enter_insert_mode && exit_insert_mode) \
		  ? SP->_smir_cost + SP->_rmir_cost + (SP->_ip_cost * count) \
		  : ((insert_character != 0) \
		    ? (SP->_ich1_cost * count) \
		    : INFINITY)))

#define UpdateAttrs(c)  if (SP->_current_attr != AttrOf(c)) \
				vidattr(AttrOf(c))

/* lib_acs.c */
static void _nc_init_acs(void); /* corresponds to traditional 'init_acs()' */
static int _nc_msec_cost(const char *const, int);  /* used by 'tack' program */

/* lib_mvcur.c */
#define INFINITY        1000000 /* cost: too high to use */

static void _nc_mvcur_init(void);
static void _nc_mvcur_resume(void);
static void _nc_mvcur_wrap(void);

static int _nc_scrolln(int, int, int, int);

static void _nc_screen_init(void);
static void _nc_screen_resume(void);
static void _nc_screen_wrap(void);

/* lib_mvcur.c */
#define INFINITY        1000000 /* cost: too high to use */

typedef struct {
    char *s_head;
    char *s_tail;
    size_t s_size;
} string_desc;

/* strings.c */
static string_desc *_nc_str_init(string_desc * dst, char *src, size_t len);
static string_desc *_nc_str_null(string_desc * dst, size_t len);
static string_desc *_nc_str_copy(string_desc * dst, string_desc * src);
static bool _nc_safe_strcat(string_desc * dst, const char *src);
static bool _nc_safe_strcpy(string_desc * dst, const char *src);

static void _nc_mvcur_init(void);
static void _nc_mvcur_resume(void);
static void _nc_mvcur_wrap(void);

static int _nc_scrolln(int, int, int, int);

static void _nc_screen_init(void);
static void _nc_screen_resume(void);
static void _nc_screen_wrap(void);

/* safe_sprintf.c */
static char * _nc_printf_string(const char *fmt, va_list ap);

/* tries.c */
static void _nc_add_to_try(struct tries **tree, const char *str, unsigned short code);
static char *_nc_expand_try(struct tries *tree, unsigned short code, int *count, size_t len);
static int _nc_remove_key(struct tries **tree, unsigned short code);

/* elsewhere ... */
static WINDOW *_nc_makenew(int, int, int, int, int);
static char *_nc_home_terminfo(void);
static chtype _nc_background(WINDOW *);
static chtype _nc_render(WINDOW *, chtype);
static int _nc_access(const char *, int);
static int _nc_baudrate(int);
static int _nc_getenv_num(const char *);
static int _nc_keypad(bool);
static int _nc_ospeed(int);
static int _nc_outch(int);
static int _nc_setupscreen(short, short const, FILE *);
static int _nc_timed_wait(int, int, int *);
static void _nc_do_color(int, int, bool, int (*)(int));
static void _nc_freewin(WINDOW *win);
static void _nc_hash_map(void);
static void _nc_init_keytry(void);
static void _nc_make_oldhash(int i);
static void _nc_flush(void);
static void _nc_scroll_oldhash(int n, int top, int bot);
static void _nc_scroll_optimize(void);
static void _nc_scroll_window(WINDOW *, int const, short const, short const, chtype);
static void _nc_set_buffer(FILE *, bool);
static void _nc_signal_handler(bool);
static void _nc_synchook(WINDOW *win);

#if USE_SIZECHANGE
static void _nc_update_screensize(void);
#endif

/* scroll indices */
static int *_nc_oldnums;

#define USE_SETBUF_0 0

#define NC_BUFFERED(flag) \
	if ((SP->_buffered != 0) != flag) \
		_nc_set_buffer(SP->_ofp, flag)

#define NC_OUTPUT ((SP != 0) ? SP->_ofp : stdout)

/* current screen is private data; avoid possible linking conflicts too */
static SCREEN *SP;
#define _nc_alloc_screen() ((SP = typeCalloc(SCREEN, 1)) != 0)
#define _nc_set_screen(sp) SP = sp

/*
 * We don't want to use the lines or columns capabilities internally,
 * because if the application is running multiple screens under
 * X windows, it's quite possible they could all have type xterm
 * but have different sizes!  So...
 */
#define screen_lines    SP->_lines
#define screen_columns  SP->_columns

static int _nc_slk_format;  /* != 0 if slk_init() called */
static int _nc_slk_initialize(WINDOW *, int);

/*
 * Some constants related to SLK's
 */
#define MAX_SKEY_OLD       8    /* count of soft keys */
#define MAX_SKEY_LEN_OLD   8    /* max length of soft key text */
#define MAX_SKEY_PC       12    /* This is what most PC's have */
#define MAX_SKEY_LEN_PC    5

/* Macro to check whether or not we use a standard format */
#define SLK_STDFMT(fmt) (fmt < 3)
/* Macro to determine height of label window */
#define SLK_LINES(fmt)  (SLK_STDFMT(fmt) ? 1 : ((fmt) - 2))

#define MAX_SKEY(fmt)     (SLK_STDFMT(fmt)? MAX_SKEY_OLD : MAX_SKEY_PC)
#define MAX_SKEY_LEN(fmt) (SLK_STDFMT(fmt)? MAX_SKEY_LEN_OLD : MAX_SKEY_LEN_PC)

static int _nc_ripoffline(int line, int (*init)(WINDOW *,int));

#endif /* CURSES_PRIV_H */

/****************************************************************************
 * Copyright (c) 1998-2000 Free Software Foundation, Inc.                   *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

/****************************************************************************
 *  Author: Zeyd M. Ben-Halim <zmbenhal@netcom.com> 1992,1995               *
 *     and: Eric S. Raymond <esr@snark.thyrsus.com>                         *
 ****************************************************************************/

/*
 * $Id: tic.h,v 1.35 2000/09/09 19:47:16 tom Exp $
 *      tic.h - Global variables and structures for the terminfo
 *                      compiler.
 */

/* tic.h */



/*
** The format of compiled terminfo files is as follows:
**
**              Header (12 bytes), containing information given below
**              Names Section, containing the names of the terminal
**              Boolean Section, containing the values of all of the
**                              boolean capabilities
**                              A null byte may be inserted here to make
**                              sure that the Number Section begins on an
**                              even word boundary.
**              Number Section, containing the values of all of the numeric
**                              capabilities, each as a short integer
**              String Section, containing short integer offsets into the
**                              String Table, one per string capability
**              String Table, containing the actual characters of the string
**                              capabilities.
**
**      NOTE that all short integers in the file are stored using VAX/PDP-style
**      byte-order, i.e., least-significant byte first.
**
**      There is no structure definition here because it would only confuse
**      matters.  Terminfo format is a raw byte layout, not a structure
**      dump.  If you happen to be on a little-endian machine with 16-bit
**      shorts that requires no padding between short members in a struct,
**      then there is a natural C structure that captures the header, but
**      not very helpfully.
*/

#define MAGIC           0432    /* first two bytes of a compiled entry */
#define MAGIC2          01036   /* first two bytes of a compiled 32-bit entry */
#define IS_TIC_MAGIC(p) (LOW_MSB(p) == MAGIC || LOW_MSB(p) == MAGIC2)

/*
 * The "maximum" here is misleading; XSI guarantees minimum values, which a
 * given implementation may exceed.
 */
#define MAX_NAME_SIZE   512     /* maximum legal name field size (XSI:127) */
#define MAX_ENTRY_SIZE  4096    /* maximum legal entry size */

#define MAX_ALIAS       32      /* POSIX minimum for PATH_MAX */

/* location of user's personal info directory */
#define PRIVATE_INFO    "%s/.terminfo"  /* plug getenv("HOME") into %s */

/*
 * Some traces are designed to be used via tic's verbose option (and similar in
 * infocmp and toe) rather than the 'trace()' function.  So we use the bits
 * above the normal trace() parameter as a debug-level.
 */

#define MAX_DEBUG_LEVEL 15
#define DEBUG_LEVEL(n)  ((n) << 12)     /* see TRACE_MAXIMUM */

#define DEBUG(n, a)     /*nothing*/

/*
 * These are the types of tokens returned by the scanner.  The first
 * three are also used in the hash table of capability names.  The scanner
 * returns one of these values after loading the specifics into the global
 * structure curr_token.
 */

#define BOOLEAN 0               /* Boolean capability */
#define NUMBER 1                /* Numeric capability */
#define STRING 2                /* String-valued capability */
#define CANCEL 3                /* Capability to be cancelled in following tc's */
#define NAMES  4                /* The names for a terminal type */
#define UNDEF  5                /* Undefined */

#define NO_PUSHBACK     -1      /* used in pushtype to indicate no pushback */

	/*
	 *      The global structure in which the specific parts of a
	 *      scanned token are returned.
	 *
	 */

struct token
{
	char    *tk_name;               /* name of capability */
	int     tk_valnumber;   /* value of capability (if a number) */
	char    *tk_valstring;  /* value of capability (if a string) */
};

	/*
	 * List of keynames with their corresponding code.
	 */
struct kn {
	const char *name;
	int code;
};

	/*
	 * Offsets to string capabilities, with the corresponding functionkey
	 * codes.
	 */
struct tinfo_fkeys {
	unsigned offset;
	chtype code;
};

static struct tinfo_fkeys _nc_tinfo_fkeys[];

	/*
	 * The file comp_captab.c contains an array of these structures, one
	 * per possible capability.  These are indexed by a hash table array of
	 * pointers to the same structures for use by the parser.
	 */

struct name_table_entry
{
	const char *nte_name;   /* name to hash on */
	int     nte_type;       /* BOOLEAN, NUMBER or STRING */
	short   nte_index;      /* index of associated variable in its array */
	short   nte_link;       /* index in table of next hash, or -1 */
};

struct alias
{
	const char      *from;
	const char      *to;
	const char      *source;
};

#define NOTFOUND        ((struct name_table_entry *) 0)

/* out-of-band values for representing absent capabilities */
#define ABSENT_BOOLEAN          -1
#define ABSENT_NUMERIC          -1
#define ABSENT_STRING           (char *)0

/* out-of-band values for representing cancels */
#define CANCELLED_BOOLEAN       (char)(-2)
#define CANCELLED_NUMERIC       -2
#define CANCELLED_STRING        (char *)-1

#define VALID_BOOLEAN(s) ((s) >= 0)
#define VALID_NUMERIC(s) ((s) >= 0)
#define VALID_STRING(s) ((s) != CANCELLED_STRING && (s) != ABSENT_STRING)

/* termcap entries longer than this may break old binaries */
#define MAX_TERMCAP_LENGTH      1023

/* this is a documented limitation of terminfo */
#define MAX_TERMINFO_LENGTH     4096

/* access.c */
static char *_nc_basename(char *);

/* comp_scan.c: lexical analysis */
static int _nc_curr_line;
#define SYN_TERMINFO    0
#define SYN_TERMCAP     1

/* comp_error.c: warning & abort messages */
static void _nc_err_abort(const char *const,...) GCC_PRINTFLIKE(1,2) GCC_NORETURN;

/* lib_tputs.c */
static int _nc_nulls_sent;              /* Add one for every null sent */

/* read_entry.c */
static const char *_nc_tic_dir(const char *);

/* end tic.h */

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

/****************************************************************************
 *  Author: Zeyd M. Ben-Halim <zmbenhal@netcom.com> 1992,1995               *
 *     and: Eric S. Raymond <esr@snark.thyrsus.com>                         *
 ****************************************************************************/

/*
 * Common macros for lib_getch.c, lib_ungetch.c
 *
 * $Id: fifo_defs.h,v 1.2 1998/02/11 12:13:56 tom Exp $
 */

#ifndef FIFO_DEFS_H
#define FIFO_DEFS_H 1

#define HEAD    SP->_fifohead
#define tail    SP->_fifotail
/* peek points to next uninterpreted character */
#define peek    SP->_fifopeek

#define h_inc() { HEAD == FIFO_SIZE-1 ? HEAD = 0 : HEAD++; if (HEAD == tail) HEAD = -1, tail = 0;}
#define h_dec() { HEAD == 0 ?  HEAD = FIFO_SIZE-1 : HEAD--; if (HEAD == tail) tail = -1;}
#define t_inc() { tail == FIFO_SIZE-1 ? tail = 0 : tail++; if (tail == HEAD) tail = -1;}
#define t_dec() { tail == 0 ?  tail = FIFO_SIZE-1 : tail--; if (HEAD == tail) fifo_clear();}
#define p_inc() { peek == FIFO_SIZE-1 ? peek = 0 : peek++;}

#define cooked_key_in_fifo()    (HEAD!=-1 && peek!=HEAD)
#define raw_key_in_fifo()       (HEAD!=-1 && peek!=tail)

#undef HIDE_EINTR

#endif /* FIFO_DEFS_H */
/****************************************************************************
 * Copyright (c) 1998,1999,2000 Free Software Foundation, Inc.              *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

/****************************************************************************
 *  Author: Zeyd M. Ben-Halim <zmbenhal@netcom.com> 1992,1995               *
 *     and: Eric S. Raymond <esr@snark.thyrsus.com>                         *
 ****************************************************************************/

/* $Id: term_entry.h,v 1.29 2000/03/19 02:04:15 tom Exp $ */

/*
 *      term_entry.h -- interface to entry-manipulation code
 */

#ifndef _TERM_ENTRY_H
#define _TERM_ENTRY_H

/* #include <term.h> */

#undef  NCURSES_XNAMES
#define NCURSES_XNAMES 1

/* We will use these symbols to hide differences between
 * termios/termio/sgttyb interfaces.
 */
#undef  TTY
#undef  SET_TTY
#undef  GET_TTY

/* Assume Posix termio if we have the header and function */
#if HAVE_TERMIOS_H

#undef  TERMIOS
#define TERMIOS 1

#define TTY struct termios

#else /* !HAVE_TERMIOS_H */

#if HAVE_TERMIO_H

#undef  TERMIOS
#define TERMIOS 1

#include <termio.h>
#define TTY struct termio

/* Add definitions to make termio look like termios.
 * But ifdef it, since there are some implementations
 * that try to do this for us in a fake <termio.h>.
 */
#ifndef TCSANOW
#define TCSANOW TCSETA
#endif
#ifndef TCSADRAIN
#define TCSADRAIN TCSETAW
#endif
#ifndef TCSAFLUSH
#define TCSAFLUSH TCSETAF
#endif
#ifndef tcsetattr
#define tcsetattr(fd, cmd, arg) ioctl(fd, cmd, arg)
#endif
#ifndef tcgetattr
#define tcgetattr(fd, arg) ioctl(fd, TCGETA, arg)
#endif
#ifndef cfgetospeed
#define cfgetospeed(t) ((t)->c_cflag & CBAUD)
#endif
#ifndef TCIFLUSH
#define TCIFLUSH 0
#endif
#ifndef TCOFLUSH
#define TCOFLUSH 1
#endif
#ifndef TCIOFLUSH
#define TCIOFLUSH 2
#endif
#ifndef tcflush
#define tcflush(fd, arg) ioctl(fd, TCFLSH, arg)
#endif

#else /* !HAVE_TERMIO_H */

#undef TERMIOS
#include <sgtty.h>
#define TTY struct sgttyb

#endif /* HAVE_TERMIO_H */

#endif /* HAVE_TERMIOS_H */

#ifdef TERMIOS
#define GET_TTY(fd, buf) tcgetattr(fd, buf)
#define SET_TTY(fd, buf) tcsetattr(fd, TCSADRAIN, buf)
#else
#define GET_TTY(fd, buf) gtty(fd, buf)
#define SET_TTY(fd, buf) stty(fd, buf)
#endif

#define NAMESIZE 256

#define CUR cur_term->type.

#define auto_left_margin               CUR Booleans[0]
#define auto_right_margin              CUR Booleans[1]
#define no_esc_ctlc                    CUR Booleans[2]
#define ceol_standout_glitch           CUR Booleans[3]
#define eat_newline_glitch             CUR Booleans[4]
#define erase_overstrike               CUR Booleans[5]
#define generic_type                   CUR Booleans[6]
#define hard_copy                      CUR Booleans[7]
#define has_meta_key                   CUR Booleans[8]
#define has_status_line                CUR Booleans[9]
#define insert_null_glitch             CUR Booleans[10]
#define memory_above                   CUR Booleans[11]
#define memory_below                   CUR Booleans[12]
#define move_insert_mode               CUR Booleans[13]
#define move_standout_mode             CUR Booleans[14]
#define over_strike                    CUR Booleans[15]
#define status_line_esc_ok             CUR Booleans[16]
#define dest_tabs_magic_smso           CUR Booleans[17]
#define tilde_glitch                   CUR Booleans[18]
#define transparent_underline          CUR Booleans[19]
#define xon_xoff                       CUR Booleans[20]
#define needs_xon_xoff                 CUR Booleans[21]
#define prtr_silent                    CUR Booleans[22]
#define hard_cursor                    CUR Booleans[23]
#define non_rev_rmcup                  CUR Booleans[24]
#define no_pad_char                    CUR Booleans[25]
#define non_dest_scroll_region         CUR Booleans[26]
#define can_change                     CUR Booleans[27]
#define back_color_erase               CUR Booleans[28]
#define hue_lightness_saturation       CUR Booleans[29]
#define col_addr_glitch                CUR Booleans[30]
#define cr_cancels_micro_mode          CUR Booleans[31]
#define has_print_wheel                CUR Booleans[32]
#define row_addr_glitch                CUR Booleans[33]
#define semi_auto_right_margin         CUR Booleans[34]
#define cpi_changes_res                CUR Booleans[35]
#define lpi_changes_res                CUR Booleans[36]
#define columns                        CUR Numbers[0]
#define init_tabs                      CUR Numbers[1]
#define lines                          CUR Numbers[2]
#define lines_of_memory                CUR Numbers[3]
#define magic_cookie_glitch            CUR Numbers[4]
#define padding_baud_rate              CUR Numbers[5]
#define virtual_terminal               CUR Numbers[6]
#define width_status_line              CUR Numbers[7]
#define num_labels                     CUR Numbers[8]
#define label_height                   CUR Numbers[9]
#define label_width                    CUR Numbers[10]
#define max_attributes                 CUR Numbers[11]
#define maximum_windows                CUR Numbers[12]
#define max_colors                     CUR Numbers[13]
#define max_pairs                      CUR Numbers[14]
#define no_color_video                 CUR Numbers[15]
#define buffer_capacity                CUR Numbers[16]
#define dot_vert_spacing               CUR Numbers[17]
#define dot_horz_spacing               CUR Numbers[18]
#define max_micro_address              CUR Numbers[19]
#define max_micro_jump                 CUR Numbers[20]
#define micro_col_size                 CUR Numbers[21]
#define micro_line_size                CUR Numbers[22]
#define number_of_pins                 CUR Numbers[23]
#define output_res_char                CUR Numbers[24]
#define output_res_line                CUR Numbers[25]
#define output_res_horz_inch           CUR Numbers[26]
#define output_res_vert_inch           CUR Numbers[27]
#define print_rate                     CUR Numbers[28]
#define wide_char_size                 CUR Numbers[29]
#define buttons                        CUR Numbers[30]
#define bit_image_entwining            CUR Numbers[31]
#define bit_image_type                 CUR Numbers[32]
#define back_tab                       CUR Strings[0]
#define bell                           CUR Strings[1]
#define carriage_return                CUR Strings[2]
#define change_scroll_region           CUR Strings[3]
#define clear_all_tabs                 CUR Strings[4]
#define clear_screen                   CUR Strings[5]
#define clr_eol                        CUR Strings[6]
#define clr_eos                        CUR Strings[7]
#define column_address                 CUR Strings[8]
#define command_character              CUR Strings[9]
#define cursor_address                 CUR Strings[10]
#define cursor_down                    CUR Strings[11]
#define cursor_home                    CUR Strings[12]
#define cursor_invisible               CUR Strings[13]
#define cursor_left                    CUR Strings[14]
#define cursor_mem_address             CUR Strings[15]
#define cursor_normal                  CUR Strings[16]
#define cursor_right                   CUR Strings[17]
#define cursor_to_ll                   CUR Strings[18]
#define cursor_up                      CUR Strings[19]
#define cursor_visible                 CUR Strings[20]
#define delete_character               CUR Strings[21]
#define delete_line                    CUR Strings[22]
#define dis_status_line                CUR Strings[23]
#define down_half_line                 CUR Strings[24]
#define enter_alt_charset_mode         CUR Strings[25]
#define enter_blink_mode               CUR Strings[26]
#define enter_bold_mode                CUR Strings[27]
#define enter_ca_mode                  CUR Strings[28]
#define enter_delete_mode              CUR Strings[29]
#define enter_dim_mode                 CUR Strings[30]
#define enter_insert_mode              CUR Strings[31]
#define enter_secure_mode              CUR Strings[32]
#define enter_protected_mode           CUR Strings[33]
#define enter_reverse_mode             CUR Strings[34]
#define enter_standout_mode            CUR Strings[35]
#define enter_underline_mode           CUR Strings[36]
#define erase_chars                    CUR Strings[37]
#define exit_alt_charset_mode          CUR Strings[38]
#define exit_attribute_mode            CUR Strings[39]
#define exit_ca_mode                   CUR Strings[40]
#define exit_delete_mode               CUR Strings[41]
#define exit_insert_mode               CUR Strings[42]
#define exit_standout_mode             CUR Strings[43]
#define exit_underline_mode            CUR Strings[44]
#define flash_screen                   CUR Strings[45]
#define form_feed                      CUR Strings[46]
#define from_status_line               CUR Strings[47]
#define init_1string                   CUR Strings[48]
#define init_2string                   CUR Strings[49]
#define init_3string                   CUR Strings[50]
#define init_file                      CUR Strings[51]
#define insert_character               CUR Strings[52]
#define insert_line                    CUR Strings[53]
#define insert_padding                 CUR Strings[54]
#define key_backspace                  CUR Strings[55]
#define key_catab                      CUR Strings[56]
#define key_clear                      CUR Strings[57]
#define key_ctab                       CUR Strings[58]
#define key_dc                         CUR Strings[59]
#define key_dl                         CUR Strings[60]
#define key_down                       CUR Strings[61]
#define key_eic                        CUR Strings[62]
#define key_eol                        CUR Strings[63]
#define key_eos                        CUR Strings[64]
#define key_f0                         CUR Strings[65]
#define key_f1                         CUR Strings[66]
#define key_f10                        CUR Strings[67]
#define key_f2                         CUR Strings[68]
#define key_f3                         CUR Strings[69]
#define key_f4                         CUR Strings[70]
#define key_f5                         CUR Strings[71]
#define key_f6                         CUR Strings[72]
#define key_f7                         CUR Strings[73]
#define key_f8                         CUR Strings[74]
#define key_f9                         CUR Strings[75]
#define key_home                       CUR Strings[76]
#define key_ic                         CUR Strings[77]
#define key_il                         CUR Strings[78]
#define key_left                       CUR Strings[79]
#define key_ll                         CUR Strings[80]
#define key_npage                      CUR Strings[81]
#define key_ppage                      CUR Strings[82]
#define key_right                      CUR Strings[83]
#define key_sf                         CUR Strings[84]
#define key_sr                         CUR Strings[85]
#define key_stab                       CUR Strings[86]
#define key_up                         CUR Strings[87]
#define keypad_local                   CUR Strings[88]
#define keypad_xmit                    CUR Strings[89]
#define lab_f0                         CUR Strings[90]
#define lab_f1                         CUR Strings[91]
#define lab_f10                        CUR Strings[92]
#define lab_f2                         CUR Strings[93]
#define lab_f3                         CUR Strings[94]
#define lab_f4                         CUR Strings[95]
#define lab_f5                         CUR Strings[96]
#define lab_f6                         CUR Strings[97]
#define lab_f7                         CUR Strings[98]
#define lab_f8                         CUR Strings[99]
#define lab_f9                         CUR Strings[100]
#define meta_off                       CUR Strings[101]
#define meta_on                        CUR Strings[102]
#define newline                        CUR Strings[103]
#define pad_char                       CUR Strings[104]
#define parm_dch                       CUR Strings[105]
#define parm_delete_line               CUR Strings[106]
#define parm_down_cursor               CUR Strings[107]
#define parm_ich                       CUR Strings[108]
#define parm_index                     CUR Strings[109]
#define parm_insert_line               CUR Strings[110]
#define parm_left_cursor               CUR Strings[111]
#define parm_right_cursor              CUR Strings[112]
#define parm_rindex                    CUR Strings[113]
#define parm_up_cursor                 CUR Strings[114]
#define pkey_key                       CUR Strings[115]
#define pkey_local                     CUR Strings[116]
#define pkey_xmit                      CUR Strings[117]
#define print_screen                   CUR Strings[118]
#define prtr_off                       CUR Strings[119]
#define prtr_on                        CUR Strings[120]
#define repeat_char                    CUR Strings[121]
#define reset_1string                  CUR Strings[122]
#define reset_2string                  CUR Strings[123]
#define reset_3string                  CUR Strings[124]
#define reset_file                     CUR Strings[125]
#define restore_cursor                 CUR Strings[126]
#define row_address                    CUR Strings[127]
#define save_cursor                    CUR Strings[128]
#define scroll_forward                 CUR Strings[129]
#define scroll_reverse                 CUR Strings[130]
#define set_attributes                 CUR Strings[131]
#define set_tab                        CUR Strings[132]
#define set_window                     CUR Strings[133]
#define tab                            CUR Strings[134]
#define to_status_line                 CUR Strings[135]
#define underline_char                 CUR Strings[136]
#define up_half_line                   CUR Strings[137]
#define init_prog                      CUR Strings[138]
#define key_a1                         CUR Strings[139]
#define key_a3                         CUR Strings[140]
#define key_b2                         CUR Strings[141]
#define key_c1                         CUR Strings[142]
#define key_c3                         CUR Strings[143]
#define prtr_non                       CUR Strings[144]
#define char_padding                   CUR Strings[145]
#define acs_chars                      CUR Strings[146]
#define plab_norm                      CUR Strings[147]
#define key_btab                       CUR Strings[148]
#define enter_xon_mode                 CUR Strings[149]
#define exit_xon_mode                  CUR Strings[150]
#define enter_am_mode                  CUR Strings[151]
#define exit_am_mode                   CUR Strings[152]
#define xon_character                  CUR Strings[153]
#define xoff_character                 CUR Strings[154]
#define ena_acs                        CUR Strings[155]
#define label_on                       CUR Strings[156]
#define label_off                      CUR Strings[157]
#define key_beg                        CUR Strings[158]
#define key_cancel                     CUR Strings[159]
#define key_close                      CUR Strings[160]
#define key_command                    CUR Strings[161]
#define key_copy                       CUR Strings[162]
#define key_create                     CUR Strings[163]
#define key_end                        CUR Strings[164]
#define key_enter                      CUR Strings[165]
#define key_exit                       CUR Strings[166]
#define key_find                       CUR Strings[167]
#define key_help                       CUR Strings[168]
#define key_mark                       CUR Strings[169]
#define key_message                    CUR Strings[170]
#define key_move                       CUR Strings[171]
#define key_next                       CUR Strings[172]
#define key_open                       CUR Strings[173]
#define key_options                    CUR Strings[174]
#define key_previous                   CUR Strings[175]
#define key_print                      CUR Strings[176]
#define key_redo                       CUR Strings[177]
#define key_reference                  CUR Strings[178]
#define key_refresh                    CUR Strings[179]
#define key_replace                    CUR Strings[180]
#define key_restart                    CUR Strings[181]
#define key_resume                     CUR Strings[182]
#define key_save                       CUR Strings[183]
#define key_suspend                    CUR Strings[184]
#define key_undo                       CUR Strings[185]
#define key_sbeg                       CUR Strings[186]
#define key_scancel                    CUR Strings[187]
#define key_scommand                   CUR Strings[188]
#define key_scopy                      CUR Strings[189]
#define key_screate                    CUR Strings[190]
#define key_sdc                        CUR Strings[191]
#define key_sdl                        CUR Strings[192]
#define key_select                     CUR Strings[193]
#define key_send                       CUR Strings[194]
#define key_seol                       CUR Strings[195]
#define key_sexit                      CUR Strings[196]
#define key_sfind                      CUR Strings[197]
#define key_shelp                      CUR Strings[198]
#define key_shome                      CUR Strings[199]
#define key_sic                        CUR Strings[200]
#define key_sleft                      CUR Strings[201]
#define key_smessage                   CUR Strings[202]
#define key_smove                      CUR Strings[203]
#define key_snext                      CUR Strings[204]
#define key_soptions                   CUR Strings[205]
#define key_sprevious                  CUR Strings[206]
#define key_sprint                     CUR Strings[207]
#define key_sredo                      CUR Strings[208]
#define key_sreplace                   CUR Strings[209]
#define key_sright                     CUR Strings[210]
#define key_srsume                     CUR Strings[211]
#define key_ssave                      CUR Strings[212]
#define key_ssuspend                   CUR Strings[213]
#define key_sundo                      CUR Strings[214]
#define req_for_input                  CUR Strings[215]
#define key_f11                        CUR Strings[216]
#define key_f12                        CUR Strings[217]
#define key_f13                        CUR Strings[218]
#define key_f14                        CUR Strings[219]
#define key_f15                        CUR Strings[220]
#define key_f16                        CUR Strings[221]
#define key_f17                        CUR Strings[222]
#define key_f18                        CUR Strings[223]
#define key_f19                        CUR Strings[224]
#define key_f20                        CUR Strings[225]
#define key_f21                        CUR Strings[226]
#define key_f22                        CUR Strings[227]
#define key_f23                        CUR Strings[228]
#define key_f24                        CUR Strings[229]
#define key_f25                        CUR Strings[230]
#define key_f26                        CUR Strings[231]
#define key_f27                        CUR Strings[232]
#define key_f28                        CUR Strings[233]
#define key_f29                        CUR Strings[234]
#define key_f30                        CUR Strings[235]
#define key_f31                        CUR Strings[236]
#define key_f32                        CUR Strings[237]
#define key_f33                        CUR Strings[238]
#define key_f34                        CUR Strings[239]
#define key_f35                        CUR Strings[240]
#define key_f36                        CUR Strings[241]
#define key_f37                        CUR Strings[242]
#define key_f38                        CUR Strings[243]
#define key_f39                        CUR Strings[244]
#define key_f40                        CUR Strings[245]
#define key_f41                        CUR Strings[246]
#define key_f42                        CUR Strings[247]
#define key_f43                        CUR Strings[248]
#define key_f44                        CUR Strings[249]
#define key_f45                        CUR Strings[250]
#define key_f46                        CUR Strings[251]
#define key_f47                        CUR Strings[252]
#define key_f48                        CUR Strings[253]
#define key_f49                        CUR Strings[254]
#define key_f50                        CUR Strings[255]
#define key_f51                        CUR Strings[256]
#define key_f52                        CUR Strings[257]
#define key_f53                        CUR Strings[258]
#define key_f54                        CUR Strings[259]
#define key_f55                        CUR Strings[260]
#define key_f56                        CUR Strings[261]
#define key_f57                        CUR Strings[262]
#define key_f58                        CUR Strings[263]
#define key_f59                        CUR Strings[264]
#define key_f60                        CUR Strings[265]
#define key_f61                        CUR Strings[266]
#define key_f62                        CUR Strings[267]
#define key_f63                        CUR Strings[268]
#define clr_bol                        CUR Strings[269]
#define clear_margins                  CUR Strings[270]
#define set_left_margin                CUR Strings[271]
#define set_right_margin               CUR Strings[272]
#define label_format                   CUR Strings[273]
#define set_clock                      CUR Strings[274]
#define display_clock                  CUR Strings[275]
#define remove_clock                   CUR Strings[276]
#define create_window                  CUR Strings[277]
#define goto_window                    CUR Strings[278]
#define hangup                         CUR Strings[279]
#define dial_phone                     CUR Strings[280]
#define quick_dial                     CUR Strings[281]
#define tone                           CUR Strings[282]
#define pulse                          CUR Strings[283]
#define flash_hook                     CUR Strings[284]
#define fixed_pause                    CUR Strings[285]
#define wait_tone                      CUR Strings[286]
#define user0                          CUR Strings[287]
#define user1                          CUR Strings[288]
#define user2                          CUR Strings[289]
#define user3                          CUR Strings[290]
#define user4                          CUR Strings[291]
#define user5                          CUR Strings[292]
#define user6                          CUR Strings[293]
#define user7                          CUR Strings[294]
#define user8                          CUR Strings[295]
#define user9                          CUR Strings[296]
#define orig_pair                      CUR Strings[297]
#define orig_colors                    CUR Strings[298]
#define initialize_color               CUR Strings[299]
#define initialize_pair                CUR Strings[300]
#define set_color_pair                 CUR Strings[301]
#define set_foreground                 CUR Strings[302]
#define set_background                 CUR Strings[303]
#define change_char_pitch              CUR Strings[304]
#define change_line_pitch              CUR Strings[305]
#define change_res_horz                CUR Strings[306]
#define change_res_vert                CUR Strings[307]
#define define_char                    CUR Strings[308]
#define enter_doublewide_mode          CUR Strings[309]
#define enter_draft_quality            CUR Strings[310]
#define enter_italics_mode             CUR Strings[311]
#define enter_leftward_mode            CUR Strings[312]
#define enter_micro_mode               CUR Strings[313]
#define enter_near_letter_quality      CUR Strings[314]
#define enter_normal_quality           CUR Strings[315]
#define enter_shadow_mode              CUR Strings[316]
#define enter_subscript_mode           CUR Strings[317]
#define enter_superscript_mode         CUR Strings[318]
#define enter_upward_mode              CUR Strings[319]
#define exit_doublewide_mode           CUR Strings[320]
#define exit_italics_mode              CUR Strings[321]
#define exit_leftward_mode             CUR Strings[322]
#define exit_micro_mode                CUR Strings[323]
#define exit_shadow_mode               CUR Strings[324]
#define exit_subscript_mode            CUR Strings[325]
#define exit_superscript_mode          CUR Strings[326]
#define exit_upward_mode               CUR Strings[327]
#define micro_column_address           CUR Strings[328]
#define micro_down                     CUR Strings[329]
#define micro_left                     CUR Strings[330]
#define micro_right                    CUR Strings[331]
#define micro_row_address              CUR Strings[332]
#define micro_up                       CUR Strings[333]
#define order_of_pins                  CUR Strings[334]
#define parm_down_micro                CUR Strings[335]
#define parm_left_micro                CUR Strings[336]
#define parm_right_micro               CUR Strings[337]
#define parm_up_micro                  CUR Strings[338]
#define select_char_set                CUR Strings[339]
#define set_bottom_margin              CUR Strings[340]
#define set_bottom_margin_parm         CUR Strings[341]
#define set_left_margin_parm           CUR Strings[342]
#define set_right_margin_parm          CUR Strings[343]
#define set_top_margin                 CUR Strings[344]
#define set_top_margin_parm            CUR Strings[345]
#define start_bit_image                CUR Strings[346]
#define start_char_set_def             CUR Strings[347]
#define stop_bit_image                 CUR Strings[348]
#define stop_char_set_def              CUR Strings[349]
#define subscript_characters           CUR Strings[350]
#define superscript_characters         CUR Strings[351]
#define these_cause_cr                 CUR Strings[352]
#define zero_motion                    CUR Strings[353]
#define char_set_names                 CUR Strings[354]
#define key_mouse                      CUR Strings[355]
#define mouse_info                     CUR Strings[356]
#define req_mouse_pos                  CUR Strings[357]
#define get_mouse                      CUR Strings[358]
#define set_a_foreground               CUR Strings[359]
#define set_a_background               CUR Strings[360]
#define pkey_plab                      CUR Strings[361]
#define device_type                    CUR Strings[362]
#define code_set_init                  CUR Strings[363]
#define set0_des_seq                   CUR Strings[364]
#define set1_des_seq                   CUR Strings[365]
#define set2_des_seq                   CUR Strings[366]
#define set3_des_seq                   CUR Strings[367]
#define set_lr_margin                  CUR Strings[368]
#define set_tb_margin                  CUR Strings[369]
#define bit_image_repeat               CUR Strings[370]
#define bit_image_newline              CUR Strings[371]
#define bit_image_carriage_return      CUR Strings[372]
#define color_names                    CUR Strings[373]
#define define_bit_image_region        CUR Strings[374]
#define end_bit_image_region           CUR Strings[375]
#define set_color_band                 CUR Strings[376]
#define set_page_length                CUR Strings[377]
#define display_pc_char                CUR Strings[378]
#define enter_pc_charset_mode          CUR Strings[379]
#define exit_pc_charset_mode           CUR Strings[380]
#define enter_scancode_mode            CUR Strings[381]
#define exit_scancode_mode             CUR Strings[382]
#define pc_term_options                CUR Strings[383]
#define scancode_escape                CUR Strings[384]
#define alt_scancode_esc               CUR Strings[385]
#define enter_horizontal_hl_mode       CUR Strings[386]
#define enter_left_hl_mode             CUR Strings[387]
#define enter_low_hl_mode              CUR Strings[388]
#define enter_right_hl_mode            CUR Strings[389]
#define enter_top_hl_mode              CUR Strings[390]
#define enter_vertical_hl_mode         CUR Strings[391]
#define set_a_attributes               CUR Strings[392]
#define set_pglen_inch                 CUR Strings[393]

#define BOOLWRITE 37
#define NUMWRITE  33
#define STRWRITE  394

/* older synonyms for some capabilities */
#define beehive_glitch  no_esc_ctlc
#define teleray_glitch  dest_tabs_magic_smso
#define micro_char_size micro_col_size

#define termcap_init2                  CUR Strings[394]
#define termcap_reset                  CUR Strings[395]
#define magic_cookie_glitch_ul         CUR Numbers[33]
#define backspaces_with_bs             CUR Booleans[37]
#define crt_no_scrolling               CUR Booleans[38]
#define no_correctly_working_cr        CUR Booleans[39]
#define carriage_return_delay          CUR Numbers[34]
#define new_line_delay                 CUR Numbers[35]
#define linefeed_if_not_lf             CUR Strings[396]
#define backspace_if_not_bs            CUR Strings[397]
#define gnu_has_meta_key               CUR Booleans[40]
#define linefeed_is_newline            CUR Booleans[41]
#define backspace_delay                CUR Numbers[36]
#define horizontal_tab_delay           CUR Numbers[37]
#define number_of_function_keys        CUR Numbers[38]
#define other_non_function_keys        CUR Strings[398]
#define arrow_key_map                  CUR Strings[399]
#define has_hardware_tabs              CUR Booleans[42]
#define return_does_clr_eol            CUR Booleans[43]
#define acs_ulcorner                   CUR Strings[400]
#define acs_llcorner                   CUR Strings[401]
#define acs_urcorner                   CUR Strings[402]
#define acs_lrcorner                   CUR Strings[403]
#define acs_ltee                       CUR Strings[404]
#define acs_rtee                       CUR Strings[405]
#define acs_btee                       CUR Strings[406]
#define acs_ttee                       CUR Strings[407]
#define acs_hline                      CUR Strings[408]
#define acs_vline                      CUR Strings[409]
#define acs_plus                       CUR Strings[410]
#define memory_lock                    CUR Strings[411]
#define memory_unlock                  CUR Strings[412]
#define box_chars_1                    CUR Strings[413]


/*
 * Predefined terminfo array sizes
 */
#define BOOLCOUNT 44
#define NUMCOUNT  39
#define STRCOUNT  414

/* used by code for comparing entries */
#define acs_chars_index  146

typedef struct termtype {       /* in-core form of terminfo data */
    char  *term_names;          /* str_table offset of term names */
    char  *str_table;           /* pointer to string table */
    char  *Booleans;            /* array of boolean values */
    short *Numbers;             /* array of integer values */
    char  **Strings;            /* array of string offsets */

#if NCURSES_XNAMES
    char  *ext_str_table;       /* pointer to extended string table */
    char  **ext_Names;          /* corresponding names */

    unsigned short num_Booleans;/* count total Booleans */
    unsigned short num_Numbers; /* count total Numbers */
    unsigned short num_Strings; /* count total Strings */

    unsigned short ext_Booleans;/* count extensions to Booleans */
    unsigned short ext_Numbers; /* count extensions to Numbers */
    unsigned short ext_Strings; /* count extensions to Strings */
#endif /* NCURSES_XNAMES */

} TERMTYPE;

typedef struct term {           /* describe an actual terminal */
    TERMTYPE    type;           /* terminal type description */
    short       Filedes;        /* file description being written to */
    TTY         Ottyb,          /* original state of the terminal */
		Nttyb;          /* current state of the terminal */
    int         _baudrate;      /* used to compute padding */
} TERMINAL;

static TERMINAL *cur_term;

/* internals */
static int _nc_set_tty_mode(TTY *buf);
static int _nc_get_tty_mode(TTY *buf);
static int _nc_read_entry(const char * const, char * const, TERMTYPE *const);
static int _nc_read_file_entry(const char *const, TERMTYPE *);
static const TERMTYPE *_nc_fallback(const char *);

/* entry points */
static TERMINAL *set_curterm(TERMINAL *);

/* miscellaneous entry points */
static int setupterm(const char *,int,int *);

/* terminfo entry points, also declared in curses.h */
#if !defined(__NCURSES_H)
static char *tparm(const char *, ...);
static int putp(const char *);
#endif /* __NCURSES_H */

/* termcap database emulation (XPG4 uses const only for 2nd param of tgetent) */
#if !defined(_NCU_TERMCAP_H)
static int tputs(const char *, int, int (*)(int));
#endif /* _NCU_TERMCAP_H */
/* end #include <term.h> */

#define MAX_USES        32
#define MAX_CROSSLINKS  16

typedef struct entry {
	TERMTYPE        tterm;
	int             nuses;
	struct
	{
	    char                *name;
	    struct entry        *link;
	    long        line;
	}
	uses[MAX_USES];
	int             ncrosslinks;
	struct entry    *crosslinks[MAX_CROSSLINKS];
	long            cstart, cend;
	long            startline;
	struct entry    *next;
	struct entry    *last;
}
ENTRY;

#if NCURSES_XNAMES
#define NUM_BOOLEANS(tp) (tp)->num_Booleans
#define NUM_NUMBERS(tp)  (tp)->num_Numbers
#define NUM_STRINGS(tp)  (tp)->num_Strings
#define EXT_NAMES(tp,i,limit,index,table) (i >= limit) ? tp->ext_Names[index] : table[i]
#else
#define NUM_BOOLEANS(tp) BOOLCOUNT
#define NUM_NUMBERS(tp)  NUMCOUNT
#define NUM_STRINGS(tp)  STRCOUNT
#define EXT_NAMES(tp,i,limit,index,table) table[i]
#endif

#define NUM_EXT_NAMES(tp) ((tp)->ext_Booleans + (tp)->ext_Numbers + (tp)->ext_Strings)

#define for_each_boolean(n,tp) for(n = 0; n < NUM_BOOLEANS(tp); n++)
#define for_each_number(n,tp)  for(n = 0; n < NUM_NUMBERS(tp);  n++)
#define for_each_string(n,tp)  for(n = 0; n < NUM_STRINGS(tp);  n++)

#define ExtBoolname(tp,i,names) EXT_NAMES(tp, i, BOOLCOUNT, (i - (tp->num_Booleans - tp->ext_Booleans)), names)
#define ExtNumname(tp,i,names)  EXT_NAMES(tp, i, NUMCOUNT, (i - (tp->num_Numbers - tp->ext_Numbers)) + tp->ext_Booleans, names)
#define ExtStrname(tp,i,names)  EXT_NAMES(tp, i, STRCOUNT, (i - (tp->num_Strings - tp->ext_Strings)) + (tp->ext_Numbers + tp->ext_Booleans), names)

#define for_entry_list(qp)      for (qp = _nc_head; qp; qp = qp->next)

#define MAX_LINE        132

#define NULLHOOK        (bool(*)(ENTRY *))0

/*
 * Note that WANTED and PRESENT are not simple inverses!  If a capability
 * has been explicitly cancelled, it's not considered WANTED.
 */
#define WANTED(s)       ((s) == ABSENT_STRING)
#define PRESENT(s)      (((s) != ABSENT_STRING) && ((s) != CANCELLED_STRING))

#define ANDMISSING(p,q) \
		{if (PRESENT(p) && !PRESENT(q)) _nc_warning(#p " but no " #q);}

#define PAIRED(p,q) \
		{ \
		if (PRESENT(q) && !PRESENT(p)) \
			_nc_warning(#q " but no " #p); \
		if (PRESENT(p) && !PRESENT(q)) \
			_nc_warning(#p " but no " #q); \
		}

/* free_ttype.c: elementary allocation code */
static void _nc_free_termtype(TERMTYPE *);

/* lib_acs.c */
static void _nc_init_acs(void); /* corresponds to traditional 'init_acs()' */

/* parse_entry.c: entry-parsing code */
#if NCURSES_XNAMES
static bool _nc_user_definable;
#endif

#endif /* _TERM_ENTRY_H */
/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

/****************************************************************************
 *  Author: Zeyd M. Ben-Halim <zmbenhal@netcom.com> 1992,1995               *
 *     and: Eric S. Raymond <esr@snark.thyrsus.com>                         *
 ****************************************************************************/

/* $Id: termcap.h.in,v 1.12 2000/10/08 01:06:37 tom Exp $ */

#ifndef _NCU_TERMCAP_H
#define _NCU_TERMCAP_H  1

static char PC;
static short ospeed;

#if !defined(_NCU_TERM_H)
static int tputs(const char *, int, int (*)(int));
#endif


#endif /* _NCU_TERMCAP_H */


static int LINES, COLS;
static int TABSIZE;


static char *
_nc_basename(char *path)
{
    char *result = strrchr(path, '/');
#ifdef __EMX__
    if (result == 0)
	result = strrchr(path, '\\');
#endif
    if (result == 0)
	result = path;
    else
	result++;
    return result;
}

static int
_nc_access(const char *path, int mode)
{
    if (access(path, mode) < 0) {
	if ((mode & W_OK) != 0
	    && errno == ENOENT
	    && strlen(path) < PATH_MAX) {
	    char head[PATH_MAX];
	    char *leaf = _nc_basename(strcpy(head, path));

	    if (leaf == 0)
		leaf = head;
	    *leaf = '\0';
	    if (head == leaf)
		(void) strcpy(head, ".");

	    return access(head, R_OK | W_OK | X_OK);
	}
	return -1;
    }
    return 0;
}

/*
**      add_tries.c
**
**      Add keycode/string to tries-tree.
**
*/

#define SET_TRY(dst,src) if ((dst->ch = *src++) == 128) dst->ch = '\0'
#define CMP_TRY(a,b) ((a)? (a == b) : (b == 128))

static void
_nc_add_to_try(struct tries **tree, const char *str, unsigned short code)
{
    static bool out_of_memory = FALSE;
    struct tries *ptr, *savedptr;
    unsigned const char *txt = (unsigned const char *) str;

    if (txt == 0 || *txt == '\0' || out_of_memory || code == 0)
	return;

    if ((*tree) != 0) {
	ptr = savedptr = (*tree);

	for (;;) {
	    unsigned char cmp = *txt;

	    while (!CMP_TRY(ptr->ch, cmp)
		&& ptr->sibling != 0)
		ptr = ptr->sibling;

	    if (CMP_TRY(ptr->ch, cmp)) {
		if (*(++txt) == '\0') {
		    ptr->value = code;
		    return;
		}
		if (ptr->child != 0)
		    ptr = ptr->child;
		else
		    break;
	    } else {
		if ((ptr->sibling = typeCalloc(struct tries, 1)) == 0) {
		    out_of_memory = TRUE;
		    return;
		}

		savedptr = ptr = ptr->sibling;
		SET_TRY(ptr, txt);
		ptr->value = 0;

		break;
	    }
	}                       /* end for (;;) */
    } else {                    /* (*tree) == 0 :: First sequence to be added */
	savedptr = ptr = (*tree) = typeCalloc(struct tries, 1);

	if (ptr == 0) {
	    out_of_memory = TRUE;
	    return;
	}

	SET_TRY(ptr, txt);
	ptr->value = 0;
    }

    /* at this point, we are adding to the try.  ptr->child == 0 */

    while (*txt) {
	ptr->child = typeCalloc(struct tries, 1);

	ptr = ptr->child;

	if (ptr == 0) {
	    out_of_memory = TRUE;

	    while ((ptr = savedptr) != 0) {
		savedptr = ptr->child;
		free(ptr);
	    }

	    return;
	}

	SET_TRY(ptr, txt);
	ptr->value = 0;
    }

    ptr->value = code;
    return;
}

/*
 *      comp_captab.c -- The names of the capabilities indexed via a hash
 *                       table for the compiler.
 *
 */

#if (BOOLCOUNT!=44)||(NUMCOUNT!=39)||(STRCOUNT!=414)
#error  --> term.h and comp_captab.c disagree about the <--
#error  --> numbers of booleans, numbers and/or strings <--
#endif

/*
 *      comp_error.c -- Error message routines
 *
 */

static char termtype[MAX_NAME_SIZE+1];

static inline void where_is_problem(void)
{
	fprintf (stderr, "\"NULL\"");
	if (_nc_curr_line >= 0)
		fprintf (stderr, ", line %d", _nc_curr_line);
	fprintf (stderr, ", col 0");
	if (termtype[0])
		fprintf (stderr, ", terminal '%s'", termtype);
	fputc(':', stderr);
	fputc(' ', stderr);
}

static void _nc_err_abort(const char *const fmt, ...)
{
va_list argp;

	where_is_problem();
	va_start(argp,fmt);
	vfprintf (stderr, fmt, argp);
	fprintf (stderr, "\n");
	va_end(argp);
	exit(EXIT_FAILURE);
}


/*
 * Wrapper for malloc/realloc.  Standard implementations allow realloc with
 * a null pointer, but older libraries may not (e.g., SunOS).
 *
 * Also if realloc fails, we discard the old memory to avoid leaks.
 */


static void *_nc_doalloc(void *oldp, size_t amount)
{
	void *newp;

	if (oldp != 0) {
		if ((newp = realloc(oldp, amount)) == 0) {
			free(oldp);
			errno = ENOMEM;         /* just in case 'free' reset */
		}
	} else {
		newp = typeMalloc(char, amount);
	}
	return newp;
}

static const TERMTYPE *_nc_fallback(const char *name GCC_UNUSED)
{
	/* the fallback list is empty */
	return((TERMTYPE *)0);
}

/*
 * free_ttype.c -- allocation functions for TERMTYPE
 *
 *      _nc_free_termtype()
 *      use_extended_names()
 *
 */

static void _nc_free_termtype(TERMTYPE *ptr)
{
    FreeIfNeeded(ptr->str_table);
    FreeIfNeeded(ptr->term_names);
    FreeIfNeeded(ptr->Booleans);
    FreeIfNeeded(ptr->Numbers);
    FreeIfNeeded(ptr->Strings);
#if NCURSES_XNAMES
    FreeIfNeeded(ptr->ext_str_table);
    FreeIfNeeded(ptr->ext_Names);
#endif
    memset(ptr, 0, sizeof(TERMTYPE));
}

#if NCURSES_XNAMES
static bool _nc_user_definable = TRUE;
#endif

/* getenv_num.c -- obtain a number from the environment */
static int
_nc_getenv_num(const char *name)
{
	char *dst = 0;
	char *src = getenv(name);
	long value;

	if ((src == 0)
	 || (value = strtol(src, &dst, 0)) < 0
	 || (dst == src)
	 || (*dst != '\0')
	 || (int)value < value)
		value = -1;

	return (int) value;
}

/******************************************************************************

NAME
   hardscroll.c -- hardware-scrolling optimization for ncurses

SYNOPSIS
   void _nc_scroll_optimize(void)

DESCRIPTION
			OVERVIEW

This algorithm for computes optimum hardware scrolling to transform an
old screen (curscr) into a new screen (newscr) via vertical line moves.

Because the screen has a `grain' (there are insert/delete/scroll line
operations but no insert/delete/scroll column operations), it is efficient
break the update algorithm into two pieces: a first stage that does only line
moves, optimizing the end product of user-invoked insertions, deletions, and
scrolls; and a second phase (corresponding to the present doupdate code in
ncurses) that does only line transformations.

The common case we want hardware scrolling for is to handle line insertions
and deletions in screen-oriented text-editors.  This two-stage approach will
accomplish that at a low computation and code-size cost.

			LINE-MOVE COMPUTATION

Now, to a discussion of the line-move computation.

For expository purposes, consider the screen lines to be represented by
integers 0..23 (with the understanding that the value of 23 may vary).
Let a new line introduced by insertion, scrolling, or at the bottom of
the screen following a line delete be given the index -1.

Assume that the real screen starts with lines 0..23.  Now, we have
the following possible line-oriented operations on the screen:

Insertion: inserts a line at a given screen row, forcing all lines below
to scroll forward.  The last screen line is lost.  For example, an insertion
at line 5 would produce: 0..4 -1 5..23.

Deletion: deletes a line at a given screen row, forcing all lines below
to scroll forward.  The last screen line is made new.  For example, a deletion
at line 7 would produce: 0..6 8..23 -1.

Scroll up: move a range of lines up 1.  The bottom line of the range
becomes new.  For example, scrolling up the region from 9 to 14 will
produce 0..8 10..14 -1 15..23.

Scroll down: move a range of lines down 1.  The top line of the range
becomes new.  For example, scrolling down the region from 12 to 16 will produce
0..11 -1 12..15 17..23.

Now, an obvious property of all these operations is that they preserve the
order of old lines, though not their position in the sequence.

The key trick of this algorithm is that the original line indices described
above are actually maintained as _line[].oldindex fields in the window
structure, and stick to each line through scroll and insert/delete operations.

Thus, it is possible at update time to look at the oldnum fields and compute
an optimal set of il/dl/scroll operations that will take the real screen
lines to the virtual screen lines.  Once these vertical moves have been done,
we can hand off to the second stage of the update algorithm, which does line
transformations.

Note that the move computation does not need to have the full generality
of a diff algorithm (which it superficially resembles) because lines cannot
be moved out of order.

			THE ALGORITHM

The scrolling is done in two passes. The first pass is from top to bottom
scroling hunks UP. The second one is from bottom to top scrolling hunks DOWN.
Obviously enough, no lines to be scrolled will be destroyed. (lav)

HOW TO TEST THIS:

Use the following production:

hardscroll: hardscroll.c
	$(CC) -g -DSCROLLDEBUG hardscroll.c -o hardscroll

Then just type scramble vectors and watch.  The following test loads are
a representative sample of cases:

-----------------------------  CUT HERE ------------------------------------
# No lines moved
 0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20 21 22 23
#
# A scroll up
 1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 -1
#
# A scroll down
-1  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20 21 22
#
# An insertion (after line 12)
 0  1  2  3  4  5  6  7  8  9 10 11 12 -1 13 14 15 16 17 18 19 20 21 22
#
# A simple deletion (line 10)
 0  1  2  3  4  5  6  7  8  9  11 12 13 14 15 16 17 18 19 20 21 22 23 -1
#
# A more complex case
-1 -1 -1 -1 -1  3  4  5  6  7  -1 -1  8  9 10 11 12 13 14 15 16 17 -1 -1
-----------------------------  CUT HERE ------------------------------------

AUTHOR
    Eric S. Raymond <esr@snark.thyrsus.com>, November 1994
    New algorithm by Alexander V. Lukyanov <lav@yars.free.net>, Aug 1997

*****************************************************************************/


/* OLDNUM(n) indicates which line will be shifted to the position n.
   if OLDNUM(n) == _NEWINDEX, then the line n in new, not shifted from
   somewhere. */
# if USE_HASHMAP
static int *_nc_oldnums = 0;
static int oldnums_allocated = 0;
#  define oldnums       _nc_oldnums
#  define OLDNUM(n)     _nc_oldnums[n]
# else /* !USE_HASHMAP */
#  define OLDNUM(n)     newscr->_line[n].oldindex
# endif /* !USE_HASHMAP */

static void _nc_scroll_optimize(void)
/* scroll optimization to transform curscr to newscr */
{
    int i;
    int start, end, shift;

    TR(TRACE_ICALLS, ("_nc_scroll_optimize() begins"));

#if !defined(SCROLLDEBUG) && !defined(HASHDEBUG)
#if USE_HASHMAP
    /* get enough storage */
    if (oldnums_allocated < screen_lines)
    {
	int *new_oldnums = typeRealloc(int, screen_lines, oldnums);
	if (!new_oldnums)
	    return;
	oldnums = new_oldnums;
	oldnums_allocated = screen_lines;
    }
    /* calculate the indices */
    _nc_hash_map();
#endif
#endif /* !defined(SCROLLDEBUG) && !defined(HASHDEBUG) */

    /* pass 1 - from top to bottom scrolling up */
    for (i = 0; i < screen_lines; )
    {
	while (i < screen_lines && (OLDNUM(i) == _NEWINDEX || OLDNUM(i) <= i))
	    i++;
	if (i >= screen_lines)
	    break;

	shift = OLDNUM(i) - i; /* shift > 0 */
	start = i;

	i++;
	while (i < screen_lines && OLDNUM(i) != _NEWINDEX && OLDNUM(i) - i == shift)
	    i++;
	end = i-1 + shift;

	TR(TRACE_UPDATE | TRACE_MOVE, ("scroll [%d, %d] by %d", start, end, shift));
#if !defined(SCROLLDEBUG) && !defined(HASHDEBUG)
	if (_nc_scrolln(shift, start, end, screen_lines - 1) == ERR)
	{
	    TR(TRACE_UPDATE | TRACE_MOVE, ("unable to scroll"));
	    continue;
	}
#endif /* !defined(SCROLLDEBUG) && !defined(HASHDEBUG) */
    }

    /* pass 2 - from bottom to top scrolling down */
    for (i = screen_lines-1; i >= 0; )
    {
	while (i >= 0 && (OLDNUM(i) == _NEWINDEX || OLDNUM(i) >= i))
	    i--;
	if (i < 0)
	    break;

	shift = OLDNUM(i) - i; /* shift < 0 */
	end = i;

	i--;
	while (i >= 0 && OLDNUM(i) != _NEWINDEX && OLDNUM(i) - i == shift)
	    i--;
	start = i+1 - (-shift);

	TR(TRACE_UPDATE | TRACE_MOVE, ("scroll [%d, %d] by %d", start, end, shift));
#if !defined(SCROLLDEBUG) && !defined(HASHDEBUG)
	if (_nc_scrolln(shift, start, end, screen_lines - 1) == ERR)
	{
	    TR(TRACE_UPDATE | TRACE_MOVE, ("unable to scroll"));
	    continue;
	}
#endif /* !defined(SCROLLDEBUG) && !defined(HASHDEBUG) */
    }
}

/* hardscroll.c ends here */

/******************************************************************************

NAME
   hashmap.c -- fill in scramble vector based on text hashes

SYNOPSIS
   void _nc_hash_map(void)

DESCRIPTION:
   This code attempts to recognize pairs of old and new lines in the physical
and virtual screens.  When a line pair is recognized, the old line index is
placed in the oldindex member of the virtual screen line, to be used by the
vertical-motion optimizer portion of the update logic (see hardscroll.c).

   Line pairs are recognized by applying a modified Heckel's algorithm,
sped up by hashing.  If a line hash is unique in both screens, those
lines must be a pair. Then if the lines just before or after the pair
are the same or similar, they are a pair too.

   We don't worry about false pairs produced by hash collisions, on the
assumption that such cases are rare and will only make the latter stages
of update less efficient, not introduce errors.

HOW TO TEST THIS:

Use the following production:

hashmap: hashmap.c
	$(CC) -g -DHASHDEBUG hashmap.c hardscroll.c ../objects/lib_trace.o -o hashmap

AUTHOR
    Eric S. Raymond <esr@snark.thyrsus.com>, May 1996
    Bug fixes and improvements by Alexander V. Lukyanov <lav@yars.free.net>, 1997

*****************************************************************************/

#define OLDTEXT(n)     curscr->_line[n].text
#define NEWTEXT(m)     newscr->_line[m].text
#define TEXTWIDTH      (curscr->_maxx+1)
#define PENDING(n)     (newscr->_line[n].firstchar != _NOCHANGE)

static inline unsigned long hash(chtype *text)
{
    int i;
    chtype ch;
    unsigned long result = 0;
    for (i = TEXTWIDTH; i>0; i--)
    {
	ch = *text++;
	result += (result<<5) + ch;
    }
    return result;
}

/* approximate update cost */
static int update_cost(chtype *from,chtype *to)
{
    int cost=0;
    int i;

    for (i=TEXTWIDTH; i>0; i--)
	if (*from++ != *to++)
	    cost++;

    return cost;
}
static int update_cost_from_blank(chtype *to)
{
    int cost=0;
    int i;
    chtype blank = BLANK;

    if (back_color_erase)
	blank |= (stdscr->_bkgd & A_COLOR);

    for (i=TEXTWIDTH; i>0; i--)
	if (blank != *to++)
	    cost++;

    return cost;
}

/*
 * Returns true when moving line 'from' to line 'to' seems to be cost
 * effective. 'blank' indicates whether the line 'to' would become blank.
 */
static inline bool cost_effective(const int from, const int to, const bool blank)
{
    int new_from;

    if (from == to)
	return FALSE;

    new_from = OLDNUM(from);
    if (new_from == _NEWINDEX)
	new_from = from;

    /*
     * On the left side of >= is the cost before moving;
     * on the right side -- cost after moving.
     */
    return (((blank ? update_cost_from_blank(NEWTEXT(to))
		    : update_cost(OLDTEXT(to),NEWTEXT(to)))
	     + update_cost(OLDTEXT(new_from),NEWTEXT(from)))
	 >= ((new_from==from ? update_cost_from_blank(NEWTEXT(from))
			     : update_cost(OLDTEXT(new_from),NEWTEXT(from)))
	     + update_cost(OLDTEXT(from),NEWTEXT(to)))) ? TRUE : FALSE;
}


typedef struct
{
    unsigned long       hashval;
    int         oldcount, newcount;
    int         oldindex, newindex;
}
    sym;

static sym *hashtab=0;
static int lines_alloc=0;

static void grow_hunks(void)
{
    int start, end, shift;
    int back_limit, forward_limit;          /* limits for cells to fill */
    int back_ref_limit, forward_ref_limit;  /* limits for refrences */
    int i;
    int next_hunk;

    /*
     * This is tricky part.  We have unique pairs to use as anchors.
     * Use these to deduce the presence of spans of identical lines.
     */
    back_limit = 0;
    back_ref_limit = 0;

    i = 0;
    while (i < screen_lines && OLDNUM(i) == _NEWINDEX)
	i++;
    for ( ; i < screen_lines; i=next_hunk)
    {
	start = i;
	shift = OLDNUM(i) - i;

	/* get forward limit */
	i = start+1;
	while (i < screen_lines && OLDNUM(i) != _NEWINDEX && OLDNUM(i) - i == shift)
	    i++;
	end = i;
	while (i < screen_lines && OLDNUM(i) == _NEWINDEX)
	    i++;
	next_hunk = i;
	forward_limit = i;
	if (i >= screen_lines || OLDNUM(i) >= i)
	    forward_ref_limit = i;
	else
	    forward_ref_limit = OLDNUM(i);

	i = start-1;
	/* grow back */
	if (shift < 0)
	    back_limit = back_ref_limit + (-shift);
	while (i >= back_limit)
	{
	    if(SP->newhash[i] == SP->oldhash[i+shift]
	    || cost_effective(i+shift, i, shift<0))
	    {
		OLDNUM(i) = i+shift;
		TR(TRACE_UPDATE | TRACE_MOVE,
		   ("connected new line %d to old line %d (backward continuation)",
		    i, i+shift));
	    }
	    else
	    {
		TR(TRACE_UPDATE | TRACE_MOVE,
		   ("not connecting new line %d to old line %d (backward continuation)",
		    i, i+shift));
		break;
	    }
	    i--;
	}

	i = end;
	/* grow forward */
	if (shift > 0)
	    forward_limit = forward_ref_limit - shift;
	while (i < forward_limit)
	{
	    if(SP->newhash[i] == SP->oldhash[i+shift]
	    || cost_effective(i+shift, i, shift>0))
	    {
		OLDNUM(i) = i+shift;
		TR(TRACE_UPDATE | TRACE_MOVE,
		   ("connected new line %d to old line %d (forward continuation)",
		    i, i+shift));
	    }
	    else
	    {
		TR(TRACE_UPDATE | TRACE_MOVE,
		   ("not connecting new line %d to old line %d (forward continuation)",
		    i, i+shift));
		break;
	    }
	    i++;
	}

	back_ref_limit = back_limit = i;
	if (shift > 0)
	    back_ref_limit += shift;
    }
}

static void _nc_hash_map(void)
{
    sym *sp;
    int i;
    int start, shift, size;


    if (screen_lines > lines_alloc)
    {
	if (hashtab)
	    free (hashtab);
	hashtab = typeMalloc(sym, (screen_lines+1)*2);
	if (!hashtab)
	{
	    if (SP->oldhash)
	    {
		FreeAndNull(SP->oldhash);
	    }
	    lines_alloc = 0;
	    return;
	}
	lines_alloc = screen_lines;
    }

    if (SP->oldhash && SP->newhash)
    {
	/* re-hash only changed lines */
	for (i = 0; i < screen_lines; i++)
	{
	    if (PENDING(i))
		SP->newhash[i] = hash(NEWTEXT(i));
	}
    }
    else
    {
	/* re-hash all */
	if (SP->oldhash == 0)
	    SP->oldhash = typeCalloc (unsigned long, screen_lines);
	if (SP->newhash == 0)
	    SP->newhash = typeCalloc (unsigned long, screen_lines);
	if (!SP->oldhash || !SP->newhash)
	    return; /* malloc failure */
	for (i = 0; i < screen_lines; i++)
	{
	    SP->newhash[i] = hash(NEWTEXT(i));
	    SP->oldhash[i] = hash(OLDTEXT(i));
	}
    }

#ifdef HASH_VERIFY
    for (i = 0; i < screen_lines; i++)
    {
	if(SP->newhash[i] != hash(NEWTEXT(i)))
	    fprintf(stderr,"error in SP->newhash[%d]\n",i);
	if(SP->oldhash[i] != hash(OLDTEXT(i)))
	    fprintf(stderr,"error in SP->oldhash[%d]\n",i);
    }
#endif

    /*
     * Set up and count line-hash values.
     */
    memset(hashtab, '\0', sizeof(*hashtab)*(screen_lines+1)*2);
    for (i = 0; i < screen_lines; i++)
    {
	unsigned long hashval = SP->oldhash[i];

	for (sp = hashtab; sp->hashval; sp++)
	    if (sp->hashval == hashval)
		break;
	sp->hashval = hashval;  /* in case this is a new entry */
	sp->oldcount++;
	sp->oldindex = i;
    }
    for (i = 0; i < screen_lines; i++)
    {
	unsigned long hashval = SP->newhash[i];

	for (sp = hashtab; sp->hashval; sp++)
	    if (sp->hashval == hashval)
		break;
	sp->hashval = hashval;  /* in case this is a new entry */
	sp->newcount++;
	sp->newindex = i;

	OLDNUM(i) = _NEWINDEX;  /* initialize old indices array */
    }

    /*
     * Mark line pairs corresponding to unique hash pairs.
     *
     * We don't mark lines with offset 0, because it can make fail
     * extending hunks by cost_effective. Otherwise, it does not
     * have any side effects.
     */
    for (sp = hashtab; sp->hashval; sp++)
	if (sp->oldcount == 1 && sp->newcount == 1
	    && sp->oldindex != sp->newindex)
	{
	    TR(TRACE_UPDATE | TRACE_MOVE,
	       ("new line %d is hash-identical to old line %d (unique)",
		   sp->newindex, sp->oldindex));
	    OLDNUM(sp->newindex) = sp->oldindex;
	}

    grow_hunks();

    /*
     * Eliminate bad or impossible shifts -- this includes removing
     * those hunks which could not grow because of conflicts, as well
     * those which are to be moved too far, they are likely to destroy
     * more than carry.
     */
    for (i = 0; i < screen_lines; )
    {
	while (i < screen_lines && OLDNUM(i) == _NEWINDEX)
	    i++;
	if (i >= screen_lines)
	    break;
	start = i;
	shift = OLDNUM(i) - i;
	i++;
	while (i < screen_lines && OLDNUM(i) != _NEWINDEX && OLDNUM(i) - i == shift)
	    i++;
	size = i - start;
	if (size < 3 || size+min(size/8,2) < abs(shift))
	{
	    while (start < i)
	    {
		OLDNUM(start) = _NEWINDEX;
		start++;
	    }
	}
    }

    /* After clearing invalid hunks, try grow the rest. */
    grow_hunks();

}

static void _nc_make_oldhash(int i)
{
    if (SP->oldhash)
	SP->oldhash[i] = hash(OLDTEXT(i));
}

static void _nc_scroll_oldhash(int n, int top, int bot)
{
    int size;
    int i;

    if (!SP->oldhash)
	return;

    size = sizeof(*SP->oldhash) * (bot-top+1-abs(n));
    if (n > 0)
    {
	memmove (SP->oldhash+top, SP->oldhash+top+n, size);
	for (i = bot; i > bot-n; i--)
	    SP->oldhash[i] = hash(OLDTEXT(i));
    }
    else
    {
	memmove (SP->oldhash+top-n, SP->oldhash+top, size);
	for (i = top; i < top-n; i++)
	    SP->oldhash[i] = hash(OLDTEXT(i));
    }
}


/* hashmap.c ends here */

/*
 *      home_terminfo.c -- return the $HOME/.terminfo string, expanded
 */


#define my_length (strlen(home) + sizeof(PRIVATE_INFO))

/* ncurses extension...fall back on user's private directory */

static char *
_nc_home_terminfo(void)
{
    char *home;
    static char *temp = 0;

    if (use_terminfo_vars()) {
	if (temp == 0) {
	    if ((home = getenv("HOME")) != 0
		&& my_length <= PATH_MAX) {
		temp = typeMalloc(char, my_length);
		if (temp == 0)
		    _nc_err_abort("Out of memory");
		(void) sprintf(temp, PRIVATE_INFO, home);
	    }
	}
	return temp;
    }
    return 0;
}

/*
**      _nc_init_keytry()
**
**      Construct the try for the current terminal's keypad keys.
**
*/

/* LINT_PREPRO
#if 0*/
/* #include <init_keytry.h> */
#ifndef NCU_KEYS_H
#define NCU_KEYS_H 1

/* This file was generated by MAKE_KEYS */

static struct tinfo_fkeys _nc_tinfo_fkeys[] = {
	{  139, KEY_A1           },     /* key_a1 */
	{  140, KEY_A3           },     /* key_a3 */
	{  141, KEY_B2           },     /* key_b2 */
	{   55, KEY_BACKSPACE    },     /* key_backspace */
	{  158, KEY_BEG          },     /* key_beg */
	{  148, KEY_BTAB         },     /* key_btab */
	{  142, KEY_C1           },     /* key_c1 */
	{  143, KEY_C3           },     /* key_c3 */
	{  159, KEY_CANCEL       },     /* key_cancel */
	{   56, KEY_CATAB        },     /* key_catab */
	{   57, KEY_CLEAR        },     /* key_clear */
	{  160, KEY_CLOSE        },     /* key_close */
	{  161, KEY_COMMAND      },     /* key_command */
	{  162, KEY_COPY         },     /* key_copy */
	{  163, KEY_CREATE       },     /* key_create */
	{   58, KEY_CTAB         },     /* key_ctab */
	{   59, KEY_DC           },     /* key_dc */
	{   60, KEY_DL           },     /* key_dl */
	{   61, KEY_DOWN         },     /* key_down */
	{   62, KEY_EIC          },     /* key_eic */
	{  164, KEY_END          },     /* key_end */
	{  165, KEY_ENTER        },     /* key_enter */
	{   63, KEY_EOL          },     /* key_eol */
	{   64, KEY_EOS          },     /* key_eos */
	{  166, KEY_EXIT         },     /* key_exit */
	{   65, KEY_F(0)         },     /* key_f0 */
	{   66, KEY_F(1)         },     /* key_f1 */
	{   67, KEY_F(10)        },     /* key_f10 */
	{  216, KEY_F(11)        },     /* key_f11 */
	{  217, KEY_F(12)        },     /* key_f12 */
	{  218, KEY_F(13)        },     /* key_f13 */
	{  219, KEY_F(14)        },     /* key_f14 */
	{  220, KEY_F(15)        },     /* key_f15 */
	{  221, KEY_F(16)        },     /* key_f16 */
	{  222, KEY_F(17)        },     /* key_f17 */
	{  223, KEY_F(18)        },     /* key_f18 */
	{  224, KEY_F(19)        },     /* key_f19 */
	{   68, KEY_F(2)         },     /* key_f2 */
	{  225, KEY_F(20)        },     /* key_f20 */
	{  226, KEY_F(21)        },     /* key_f21 */
	{  227, KEY_F(22)        },     /* key_f22 */
	{  228, KEY_F(23)        },     /* key_f23 */
	{  229, KEY_F(24)        },     /* key_f24 */
	{  230, KEY_F(25)        },     /* key_f25 */
	{  231, KEY_F(26)        },     /* key_f26 */
	{  232, KEY_F(27)        },     /* key_f27 */
	{  233, KEY_F(28)        },     /* key_f28 */
	{  234, KEY_F(29)        },     /* key_f29 */
	{   69, KEY_F(3)         },     /* key_f3 */
	{  235, KEY_F(30)        },     /* key_f30 */
	{  236, KEY_F(31)        },     /* key_f31 */
	{  237, KEY_F(32)        },     /* key_f32 */
	{  238, KEY_F(33)        },     /* key_f33 */
	{  239, KEY_F(34)        },     /* key_f34 */
	{  240, KEY_F(35)        },     /* key_f35 */
	{  241, KEY_F(36)        },     /* key_f36 */
	{  242, KEY_F(37)        },     /* key_f37 */
	{  243, KEY_F(38)        },     /* key_f38 */
	{  244, KEY_F(39)        },     /* key_f39 */
	{   70, KEY_F(4)         },     /* key_f4 */
	{  245, KEY_F(40)        },     /* key_f40 */
	{  246, KEY_F(41)        },     /* key_f41 */
	{  247, KEY_F(42)        },     /* key_f42 */
	{  248, KEY_F(43)        },     /* key_f43 */
	{  249, KEY_F(44)        },     /* key_f44 */
	{  250, KEY_F(45)        },     /* key_f45 */
	{  251, KEY_F(46)        },     /* key_f46 */
	{  252, KEY_F(47)        },     /* key_f47 */
	{  253, KEY_F(48)        },     /* key_f48 */
	{  254, KEY_F(49)        },     /* key_f49 */
	{   71, KEY_F(5)         },     /* key_f5 */
	{  255, KEY_F(50)        },     /* key_f50 */
	{  256, KEY_F(51)        },     /* key_f51 */
	{  257, KEY_F(52)        },     /* key_f52 */
	{  258, KEY_F(53)        },     /* key_f53 */
	{  259, KEY_F(54)        },     /* key_f54 */
	{  260, KEY_F(55)        },     /* key_f55 */
	{  261, KEY_F(56)        },     /* key_f56 */
	{  262, KEY_F(57)        },     /* key_f57 */
	{  263, KEY_F(58)        },     /* key_f58 */
	{  264, KEY_F(59)        },     /* key_f59 */
	{   72, KEY_F(6)         },     /* key_f6 */
	{  265, KEY_F(60)        },     /* key_f60 */
	{  266, KEY_F(61)        },     /* key_f61 */
	{  267, KEY_F(62)        },     /* key_f62 */
	{  268, KEY_F(63)        },     /* key_f63 */
	{   73, KEY_F(7)         },     /* key_f7 */
	{   74, KEY_F(8)         },     /* key_f8 */
	{   75, KEY_F(9)         },     /* key_f9 */
	{  167, KEY_FIND         },     /* key_find */
	{  168, KEY_HELP         },     /* key_help */
	{   76, KEY_HOME         },     /* key_home */
	{   77, KEY_IC           },     /* key_ic */
	{   78, KEY_IL           },     /* key_il */
	{   79, KEY_LEFT         },     /* key_left */
	{   80, KEY_LL           },     /* key_ll */
	{  169, KEY_MARK         },     /* key_mark */
	{  170, KEY_MESSAGE      },     /* key_message */
	{  355, KEY_MOUSE        },     /* key_mouse */
	{  171, KEY_MOVE         },     /* key_move */
	{  172, KEY_NEXT         },     /* key_next */
	{   81, KEY_NPAGE        },     /* key_npage */
	{  173, KEY_OPEN         },     /* key_open */
	{  174, KEY_OPTIONS      },     /* key_options */
	{   82, KEY_PPAGE        },     /* key_ppage */
	{  175, KEY_PREVIOUS     },     /* key_previous */
	{  176, KEY_PRINT        },     /* key_print */
	{  177, KEY_REDO         },     /* key_redo */
	{  178, KEY_REFERENCE    },     /* key_reference */
	{  179, KEY_REFRESH      },     /* key_refresh */
	{  180, KEY_REPLACE      },     /* key_replace */
	{  181, KEY_RESTART      },     /* key_restart */
	{  182, KEY_RESUME       },     /* key_resume */
	{   83, KEY_RIGHT        },     /* key_right */
	{  183, KEY_SAVE         },     /* key_save */
	{  186, KEY_SBEG         },     /* key_sbeg */
	{  187, KEY_SCANCEL      },     /* key_scancel */
	{  188, KEY_SCOMMAND     },     /* key_scommand */
	{  189, KEY_SCOPY        },     /* key_scopy */
	{  190, KEY_SCREATE      },     /* key_screate */
	{  191, KEY_SDC          },     /* key_sdc */
	{  192, KEY_SDL          },     /* key_sdl */
	{  193, KEY_SELECT       },     /* key_select */
	{  194, KEY_SEND         },     /* key_send */
	{  195, KEY_SEOL         },     /* key_seol */
	{  196, KEY_SEXIT        },     /* key_sexit */
	{   84, KEY_SF           },     /* key_sf */
	{  197, KEY_SFIND        },     /* key_sfind */
	{  198, KEY_SHELP        },     /* key_shelp */
	{  199, KEY_SHOME        },     /* key_shome */
	{  200, KEY_SIC          },     /* key_sic */
	{  201, KEY_SLEFT        },     /* key_sleft */
	{  202, KEY_SMESSAGE     },     /* key_smessage */
	{  203, KEY_SMOVE        },     /* key_smove */
	{  204, KEY_SNEXT        },     /* key_snext */
	{  205, KEY_SOPTIONS     },     /* key_soptions */
	{  206, KEY_SPREVIOUS    },     /* key_sprevious */
	{  207, KEY_SPRINT       },     /* key_sprint */
	{   85, KEY_SR           },     /* key_sr */
	{  208, KEY_SREDO        },     /* key_sredo */
	{  209, KEY_SREPLACE     },     /* key_sreplace */
	{  210, KEY_SRIGHT       },     /* key_sright */
	{  211, KEY_SRSUME       },     /* key_srsume */
	{  212, KEY_SSAVE        },     /* key_ssave */
	{  213, KEY_SSUSPEND     },     /* key_ssuspend */
	{   86, KEY_STAB         },     /* key_stab */
	{  214, KEY_SUNDO        },     /* key_sundo */
	{  184, KEY_SUSPEND      },     /* key_suspend */
	{  185, KEY_UNDO         },     /* key_undo */
	{   87, KEY_UP           },     /* key_up */
	{ 0, 0} };

#endif /* NCU_KEYS_H */

/* LINT_PREPRO
#endif*/

static void _nc_init_keytry(void)
{
	size_t n;

	/* The SP->_keytry value is initialized in newterm(), where the SP
	 * structure is created, because we can not tell where keypad() or
	 * mouse_activate() (which will call keyok()) are first called.
	 */

	for (n = 0; _nc_tinfo_fkeys[n].code; n++)
		if (_nc_tinfo_fkeys[n].offset < STRCOUNT)
		_nc_add_to_try(&(SP->_keytry),
			CUR Strings[_nc_tinfo_fkeys[n].offset],
			_nc_tinfo_fkeys[n].code);
}

/*
 * Enable (or disable) ncurses' interpretation of a keycode by adding (or
 * removing) the corresponding 'tries' entry.
 *
 * Do this by storing a second tree of tries, which records the disabled keys.
 * The simplest way to copy is to make a function that returns the string (with
 * nulls set to 0200), then use that to reinsert the string into the
 * corresponding tree.
 */

static int keyok(int c, bool flag)
{
	int code = ERR;
	int count = 0;
	char *s;

	T((T_CALLED("keyok(%d,%d)"), c, flag));
	if (flag) {
		while ((s = _nc_expand_try(SP->_key_ok, c, &count, 0)) != 0
		 && _nc_remove_key(&(SP->_key_ok), c)) {
			_nc_add_to_try(&(SP->_keytry), s, c);
			free(s);
			code = OK;
			count = 0;
		}
	} else {
		while ((s = _nc_expand_try(SP->_keytry, c, &count, 0)) != 0
		 && _nc_remove_key(&(SP->_keytry), c)) {
			_nc_add_to_try(&(SP->_key_ok), s, c);
			free(s);
			code = OK;
			count = 0;
		}
	}
	returnCode(code);
}


static chtype acs_map[ACS_LEN];

static void _nc_init_acs(void)
{
	T(("initializing ACS map"));

	/*
	 * Initializations for a UNIX-like multi-terminal environment.  Use
	 * ASCII chars and count on the terminfo description to do better.
	 */
	ACS_ULCORNER = '+';     /* should be upper left corner */
	ACS_LLCORNER = '+';     /* should be lower left corner */
	ACS_URCORNER = '+';     /* should be upper right corner */
	ACS_LRCORNER = '+';     /* should be lower right corner */
	ACS_RTEE     = '+';     /* should be tee pointing left */
	ACS_LTEE     = '+';     /* should be tee pointing right */
	ACS_BTEE     = '+';     /* should be tee pointing up */
	ACS_TTEE     = '+';     /* should be tee pointing down */
	ACS_HLINE    = '-';     /* should be horizontal line */
	ACS_VLINE    = '|';     /* should be vertical line */
	ACS_PLUS     = '+';     /* should be large plus or crossover */
	ACS_S1       = '~';     /* should be scan line 1 */
	ACS_S9       = '_';     /* should be scan line 9 */
	ACS_DIAMOND  = '+';     /* should be diamond */
	ACS_CKBOARD  = ':';     /* should be checker board (stipple) */
	ACS_DEGREE   = '\'';    /* should be degree symbol */
	ACS_PLMINUS  = '#';     /* should be plus/minus */
	ACS_BULLET   = 'o';     /* should be bullet */
	ACS_LARROW   = '<';     /* should be arrow pointing left */
	ACS_RARROW   = '>';     /* should be arrow pointing right */
	ACS_DARROW   = 'v';     /* should be arrow pointing down */
	ACS_UARROW   = '^';     /* should be arrow pointing up */
	ACS_BOARD    = '#';     /* should be board of squares */
	ACS_LANTERN  = '#';     /* should be lantern symbol */
	ACS_BLOCK    = '#';     /* should be solid square block */
	/* these defaults were invented for ncurses */
	ACS_S3       = '-';     /* should be scan line 3 */
	ACS_S7       = '-';     /* should be scan line 7 */
	ACS_LEQUAL   = '<';     /* should be less-than-or-equal-to */
	ACS_GEQUAL   = '>';     /* should be greater-than-or-equal-to */
	ACS_PI       = '*';     /* should be greek pi */
	ACS_NEQUAL   = '!';     /* should be not-equal */
	ACS_STERLING = 'f';     /* should be pound-sterling symbol */

	if (ena_acs != NULL)
	{
		putp(ena_acs);
	}

#define ALTCHAR(c)      ((chtype)(((unsigned char)(c)) | A_ALTCHARSET))

	if (acs_chars != NULL) {
	    size_t i = 0;
	    size_t length = strlen(acs_chars);

		while (i < length)
			switch (acs_chars[i]) {
			case 'l':case 'm':case 'k':case 'j':
			case 'u':case 't':case 'v':case 'w':
			case 'q':case 'x':case 'n':case 'o':
			case 's':case '`':case 'a':case 'f':
			case 'g':case '~':case ',':case '+':
			case '.':case '-':case 'h':case 'i':
			case '0':case 'p':case 'r':case 'y':
			case 'z':case '{':case '|':case '}':
				acs_map[(unsigned int)acs_chars[i]] =
					ALTCHAR(acs_chars[i+1]);
				i++;
				/* FALLTHRU */
			default:
				i++;
				break;
			}
	}
}

/*
**      lib_addch.c
**
**      The routine waddch().
**
*/

/*
 * Ugly microtweaking alert.  Everything from here to end of module is
 * likely to be speed-critical -- profiling data sure says it is!
 * Most of the important screen-painting functions are shells around
 * waddch().  So we make every effort to reduce function-call overhead
 * by inlining stuff, even at the cost of making wrapped copies for
 * export.  Also we supply some internal versions that don't call the
 * window sync hook, for use by string-put functions.
 */

/* Return bit mask for clearing color pair number if given ch has color */
#define COLOR_MASK(ch) (~(chtype)((ch)&A_COLOR?A_COLOR:0))

static inline chtype
render_char(WINDOW *win, chtype ch)
/* compute a rendition of the given char correct for the current context */
{
    chtype a = win->_attrs;

    if (ch == ' ') {
	/* color in attrs has precedence over bkgd */
	ch = a | (win->_bkgd & COLOR_MASK(a));
    } else {
	/* color in attrs has precedence over bkgd */
	a |= (win->_bkgd & A_ATTRIBUTES) & COLOR_MASK(a);
	/* color in ch has precedence */
	ch |= (a & COLOR_MASK(ch));
    }

    TR(TRACE_VIRTPUT, ("bkg = %lx, attrs = %lx -> ch = %lx", win->_bkgd,
	    win->_attrs, ch));

    return (ch);
}

static chtype
_nc_background(WINDOW *win)
/* make render_char() visible while still allowing us to inline it below */
{
    return (win->_bkgd);
}

static chtype
_nc_render(WINDOW *win, chtype ch)
/* make render_char() visible while still allowing us to inline it below */
{
    return render_char(win, ch);
}

#define CHECK_POSITION(win, x, y)       /* nothing */

static inline int
waddch_literal(WINDOW *win, chtype ch)
{
    int x;
    struct ldat *line;

    x = win->_curx;

    CHECK_POSITION(win, x, win->_cury);

    ch = render_char(win, ch);
    TR(TRACE_VIRTPUT, ("win attr = %s", _traceattr(win->_attrs)));

    line = win->_line + win->_cury;

    CHANGED_CELL(line, x);

    line->text[x++] = ch;

    TR(TRACE_VIRTPUT, ("(%d, %d) = %s", win->_cury, x, _tracechtype(ch)));
    if (x > win->_maxx) {
	/*
	 * The _WRAPPED flag is useful only for telling an application that
	 * we've just wrapped the cursor.  We don't do anything with this flag
	 * except set it when wrapping, and clear it whenever we move the
	 * cursor.  If we try to wrap at the lower-right corner of a window, we
	 * cannot move the cursor (since that wouldn't be legal).  So we return
	 * an error (which is what SVr4 does).  Unlike SVr4, we can
	 * successfully add a character to the lower-right corner (Solaris 2.6
	 * does this also, however).
	 */
	win->_flags |= _WRAPPED;
	if (++win->_cury > win->_regbottom) {
	    win->_cury = win->_regbottom;
	    win->_curx = win->_maxx;
	    if (!win->_scroll)
		return (ERR);
	    scroll(win);
	}
	win->_curx = 0;
	return (OK);
    }
    win->_curx = x;
    return OK;
}

static inline int
waddch_nosync(WINDOW *win, const chtype ch)
/* the workhorse function -- add a character to the given window */
{
    int x, y;
    int t = 0;
    const char *s = 0;

    if ((ch & A_ALTCHARSET)
	|| ((t = TextOf(ch)) > 127)
	|| ((s = unctrl(t))[1] == 0))
	return waddch_literal(win, ch);

    x = win->_curx;
    y = win->_cury;

    switch (t) {
    case '\t':
	x += (TABSIZE - (x % TABSIZE));

	/*
	 * Space-fill the tab on the bottom line so that we'll get the
	 * "correct" cursor position.
	 */
	if ((!win->_scroll && (y == win->_regbottom))
	    || (x <= win->_maxx)) {
	    chtype blank = (' ' | AttrOf(ch));
	    while (win->_curx < x) {
		if (waddch_literal(win, blank) == ERR)
		    return (ERR);
	    }
	    break;
	} else {
	    wclrtoeol(win);
	    win->_flags |= _WRAPPED;
	    if (++y > win->_regbottom) {
		x = win->_maxx;
		y--;
		if (win->_scroll) {
		    scroll(win);
		    x = 0;
		}
	    } else {
		x = 0;
	    }
	}
	break;
    case '\n':
	wclrtoeol(win);
	if (++y > win->_regbottom) {
	    y--;
	    if (win->_scroll)
		scroll(win);
	    else
		return (ERR);
	}
	/* FALLTHRU */
    case '\r':
	x = 0;
	win->_flags &= ~_WRAPPED;
	break;
    case '\b':
	if (x == 0)
	    return (OK);
	x--;
	win->_flags &= ~_WRAPPED;
	break;
    default:
	while (*s)
	    if (waddch_literal(win, (*s++) | AttrOf(ch)) == ERR)
		return ERR;
	return (OK);
    }

    win->_curx = x;
    win->_cury = y;

    return (OK);
}

static int
_nc_waddch_nosync(WINDOW *win, const chtype c)
/* export copy of waddch_nosync() so the string-put functions can use it */
{
    return (waddch_nosync(win, c));
}

/*
 * The versions below call _nc_synhook().  We wanted to avoid this in the
 * version exported for string puts; they'll call _nc_synchook once at end
 * of run.
 */

/* These are actual entry points */

int
waddch(WINDOW *win, const chtype ch)
{
    int code = ERR;

    TR(TRACE_VIRTPUT | TRACE_CCALLS, (T_CALLED("waddch(%p, %s)"), win,
	    _tracechtype(ch)));

    if (win && (waddch_nosync(win, ch) != ERR)) {
	_nc_synchook(win);
	code = OK;
    }

    TR(TRACE_VIRTPUT | TRACE_CCALLS, (T_RETURN("%d"), code));
    return (code);
}

static int
wechochar(WINDOW *win, const chtype ch)
{
    int code = ERR;

    TR(TRACE_VIRTPUT | TRACE_CCALLS, (T_CALLED("wechochar(%p, %s)"), win,
	    _tracechtype(ch)));

    if (win && (waddch_nosync(win, ch) != ERR)) {
	bool save_immed = win->_immed;
	win->_immed = TRUE;
	_nc_synchook(win);
	win->_immed = save_immed;
	code = OK;
    }
    TR(TRACE_VIRTPUT | TRACE_CCALLS, (T_RETURN("%d"), code));
    return (code);
}

/*
**      lib_addstr.c
*
**      The routines waddnstr(), waddchnstr().
**
*/


static int
waddnstr(WINDOW *win, const char *const astr, int n)
{
    unsigned const char *str = (unsigned const char *) astr;
    int code = ERR;

    T((T_CALLED("waddnstr(%p,%s,%d)"), win, _nc_visbuf(astr), n));

    if (win && (str != 0)) {
	TR(TRACE_VIRTPUT | TRACE_ATTRS, ("... current %s", _traceattr(win->_attrs)));
	TR(TRACE_VIRTPUT, ("str is not null"));
	code = OK;
	if (n < 0)
	    n = (int) strlen(astr);

	while ((n-- > 0) && (*str != '\0')) {
	    TR(TRACE_VIRTPUT, ("*str = %#x", *str));
	    if (_nc_waddch_nosync(win, (chtype) * str++) == ERR) {
		code = ERR;
		break;
	    }
	}
	_nc_synchook(win);
    }
    TR(TRACE_VIRTPUT, ("waddnstr returns %d", code));
    returnCode(code);
}

/*
 *      lib_baudrate.c
 *
 */

/*
 *      int
 *      baudrate()
 *
 *      Returns the current terminal's baud rate.
 *
 */

struct speed {
    int s;      /* value for 'ospeed' is an index */
    int sp;     /* the actual speed */
};

static struct speed const speeds[] =
{
    {B0, 0},
    {B50, 50},
    {B75, 75},
    {B110, 110},
    {B134, 134},
    {B150, 150},
    {B200, 200},
    {B300, 300},
    {B600, 600},
    {B1200, 1200},
    {B1800, 1800},
    {B2400, 2400},
    {B4800, 4800},
    {B9600, 9600},
#ifdef B19200
    {B19200, 19200},
#else
#ifdef EXTA
    {EXTA, 19200},
#endif
#endif
#ifdef B38400
    {B38400, 38400},
#else
#ifdef EXTB
    {EXTB, 38400},
#endif
#endif
#ifdef B57600
    {B57600, 57600},
#endif
#ifdef B115200
    {B115200, 115200},
#endif
#ifdef B230400
    {B230400, 230400},
#endif
#ifdef B460800
    {B460800, 460800},
#endif
};

static int
_nc_baudrate(int OSpeed)
{
    static int last_OSpeed;
    static int last_baudrate;

    int result;
    unsigned i;

    if (OSpeed == last_OSpeed) {
	result = last_baudrate;
    } else {
	result = ERR;
	if (OSpeed >= 0) {
	    for (i = 0; i < SIZEOF(speeds); i++) {
		if (speeds[i].s == OSpeed) {
		    result = speeds[i].sp;
		    break;
		}
	    }
	}
	last_baudrate = result;
    }
    return (result);
}

static int
_nc_ospeed(int BaudRate)
{
    int result = 1;
    unsigned i;

    if (BaudRate >= 0) {
	for (i = 0; i < SIZEOF(speeds); i++) {
	    if (speeds[i].sp == BaudRate) {
		result = speeds[i].s;
		break;
	    }
	}
    }
    return (result);
}

static int
baudrate(void)
{
    int result;

    T((T_CALLED("baudrate()")));

#ifdef TERMIOS
	ospeed = cfgetospeed(&cur_term->Nttyb);
#else
	ospeed = cur_term->Nttyb.sg_ospeed;
#endif
    result = _nc_baudrate(ospeed);
    if (cur_term != 0)
	cur_term->_baudrate = result;

    returnCode(result);
}

/*
 *      beep()
 *
 *      Sound the current terminal's audible bell if it has one.   If not,
 *      flash the screen if possible.
 *
 */

static int beep(void)
{
	int res = ERR;

	T((T_CALLED("beep()")));

	/* FIXME: should make sure that we are not in altchar mode */
	if (bell) {
		res = putp(bell);
		_nc_flush();
	} else if (flash_screen) {
		res = putp(flash_screen);
		_nc_flush();
	}

	returnCode(res);
}


void wbkgdset(WINDOW *win, chtype ch)
{
  T((T_CALLED("wbkgdset(%p,%s)"), win, _tracechtype(ch)));

  if (win) {
    chtype off = AttrOf(win->_bkgd);
    chtype on  = AttrOf(ch);

    toggle_attr_off(win->_attrs,off);
    toggle_attr_on (win->_attrs,on);

    if (TextOf(ch)==0)
      ch |= BLANK;
    win->_bkgd = ch;
  }
  returnVoid;
}

/*
**      lib_clearok.c
**
**      The routine clearok.
**
*/


int clearok(WINDOW *win, bool flag)
{
	T((T_CALLED("clearok(%p,%d)"), win, flag));

	if (win) {
	  win->_clear = flag;
	  returnCode(OK);
	}
	else
	  returnCode(ERR);
}

/*
**      lib_clreol.c
**
**      The routine wclrtoeol().
**
*/


int
wclrtoeol(WINDOW *win)
{
    int code = ERR;

    T((T_CALLED("wclrtoeol(%p)"), win));

    if (win) {
	chtype blank;
	chtype *ptr, *end;
	struct ldat *line;
	NCURSES_SIZE_T y = win->_cury;
	NCURSES_SIZE_T x = win->_curx;

	/*
	 * If we have just wrapped the cursor, the clear applies to the
	 * new line, unless we are at the lower right corner.
	 */
	if (win->_flags & _WRAPPED
	    && y < win->_maxy) {
	    win->_flags &= ~_WRAPPED;
	}

	/*
	 * There's no point in clearing if we're not on a legal
	 * position, either.
	 */
	if (win->_flags & _WRAPPED
	    || y > win->_maxy
	    || x > win->_maxx)
	    returnCode(ERR);

	blank = _nc_background(win);
	line = &win->_line[y];
	CHANGED_TO_EOL(line, x, win->_maxx);

	ptr = &(line->text[x]);
	end = &(line->text[win->_maxx]);

	while (ptr <= end)
	    *ptr++ = blank;

	_nc_synchook(win);
	code = OK;
    }
    returnCode(code);
}

/* lib_color.c
 *
 * Handles color emulation of SYS V curses
 */



/*
 * These should be screen structure members.  They need to be globals for
 * historical reasons.  So we assign them in start_color() and also in
 * set_term()'s screen-switching logic.
 */
static int COLOR_PAIRS = 0;
static int COLORS = 0;

/*
 * Given a RGB range of 0..1000, we'll normally set the individual values
 * to about 2/3 of the maximum, leaving full-range for bold/bright colors.
 */
#define RGB_ON  680
#define RGB_OFF 0
/* *INDENT-OFF* */
static const color_t cga_palette[] =
{
    /*  R               G               B */
    {RGB_OFF,           RGB_OFF,        RGB_OFF},       /* COLOR_BLACK */
    {RGB_ON,            RGB_OFF,        RGB_OFF},       /* COLOR_RED */
    {RGB_OFF,           RGB_ON,         RGB_OFF},       /* COLOR_GREEN */
    {RGB_ON,            RGB_ON,         RGB_OFF},       /* COLOR_YELLOW */
    {RGB_OFF,           RGB_OFF,        RGB_ON},        /* COLOR_BLUE */
    {RGB_ON,            RGB_OFF,        RGB_ON},        /* COLOR_MAGENTA */
    {RGB_OFF,           RGB_ON,         RGB_ON},        /* COLOR_CYAN */
    {RGB_ON,            RGB_ON,         RGB_ON},        /* COLOR_WHITE */
};

static const color_t hls_palette[] =
{
    /*  H       L       S */
    {   0,      0,      0},             /* COLOR_BLACK */
    {   120,    50,     100},           /* COLOR_RED */
    {   240,    50,     100},           /* COLOR_GREEN */
    {   180,    50,     100},           /* COLOR_YELLOW */
    {   330,    50,     100},           /* COLOR_BLUE */
    {   60,     50,     100},           /* COLOR_MAGENTA */
    {   300,    50,     100},           /* COLOR_CYAN */
    {   0,      50,     100},           /* COLOR_WHITE */
};
/* *INDENT-ON* */

#if NCURSES_EXT_FUNCS
/*
 * These are called from _nc_do_color(), which in turn is called from
 * vidattr - so we have to assume that SP may be null.
 */
static int
default_fg(void)
{
    return (SP != 0) ? SP->_default_fg : COLOR_WHITE;
}

static int
default_bg(void)
{
    return SP != 0 ? SP->_default_bg : COLOR_BLACK;
}
#else
#define default_fg() COLOR_WHITE
#define default_bg() COLOR_BLACK
#endif

/*
 * SVr4 curses is known to interchange color codes (1,4) and (3,6), possibly
 * to maintain compatibility with a pre-ANSI scheme.  The same scheme is
 * also used in the FreeBSD syscons.
 */
static int
toggled_colors(int c)
{
    if (c < 16) {
	static const int table[] =
	{0, 4, 2, 6, 1, 5, 3, 7,
	    8, 12, 10, 14, 9, 13, 11, 15};
	c = table[c];
    }
    return c;
}

static void
set_background_color(int bg, int (*outc) (int))
{
    if (set_a_background) {
	tputs(tparm(set_a_background, bg), 1, outc);
    } else {
	tputs(tparm(set_background, toggled_colors(bg)), 1, outc);
    }
}

static void
set_foreground_color(int fg, int (*outc) (int))
{
    if (set_a_foreground) {
	tputs(tparm(set_a_foreground, fg), 1, outc);
    } else {
	tputs(tparm(set_foreground, toggled_colors(fg)), 1, outc);
    }
}

static bool
set_original_colors(void)
{
    if (orig_pair != 0) {
	putp(orig_pair);
	return TRUE;
    } else if (orig_colors != NULL) {
	putp(orig_colors);
	return TRUE;
    }
    return FALSE;
}

int
start_color(void)
{
    int n;
    const color_t *tp;

    T((T_CALLED("start_color()")));

    if (set_original_colors() != TRUE) {
	set_foreground_color(default_fg(), _nc_outch);
	set_background_color(default_bg(), _nc_outch);
    }

    if (VALID_NUMERIC(max_pairs))
	COLOR_PAIRS = SP->_pair_count = max_pairs;
    else
	returnCode(ERR);
    if ((SP->_color_pairs = typeCalloc(unsigned short, max_pairs)) == 0)
	  returnCode(ERR);
    SP->_color_pairs[0] = PAIR_OF(default_fg(), default_bg());
    if (VALID_NUMERIC(max_colors))
	COLORS = SP->_color_count = max_colors;
    else
	returnCode(ERR);
    SP->_coloron = 1;

    if ((SP->_color_table = typeMalloc(color_t, COLORS)) == 0)
	returnCode(ERR);
    tp = (hue_lightness_saturation) ? hls_palette : cga_palette;
    for (n = 0; n < COLORS; n++) {
	if (n < 8) {
	    SP->_color_table[n] = tp[n];
	} else {
	    SP->_color_table[n] = tp[n % 8];
	    if (hue_lightness_saturation) {
		SP->_color_table[n].green = 100;
	    } else {
		if (SP->_color_table[n].red)
		    SP->_color_table[n].red = 1000;
		if (SP->_color_table[n].green)
		    SP->_color_table[n].green = 1000;
		if (SP->_color_table[n].blue)
		    SP->_color_table[n].blue = 1000;
	    }
	}
    }

    T(("started color: COLORS = %d, COLOR_PAIRS = %d", COLORS, COLOR_PAIRS));

    returnCode(OK);
}

/*
 * Extension (1997/1/18) - Allow negative f/b values to set default color
 * values.
 */
int
init_pair(short pair, short f, short b)
{
    unsigned result;

    T((T_CALLED("init_pair(%d,%d,%d)"), pair, f, b));

    if ((pair < 0) || (pair >= COLOR_PAIRS))
	returnCode(ERR);
#if NCURSES_EXT_FUNCS
    if (SP->_default_color) {
	if (f < 0)
	    f = C_MASK;
	if (b < 0)
	    b = C_MASK;
	if (f >= COLORS && f != C_MASK)
	    returnCode(ERR);
	if (b >= COLORS && b != C_MASK)
	    returnCode(ERR);
    } else
#endif
    {
	if ((f < 0) || (f >= COLORS)
	    || (b < 0) || (b >= COLORS)
	    || (pair < 1))
	    returnCode(ERR);
    }

    /*
     * When a pair's content is changed, replace its colors (if pair was
     * initialized before a screen update is performed replacing original
     * pair colors with the new ones).
     */
    result = PAIR_OF(f, b);
    if (SP->_color_pairs[pair] != 0
	&& SP->_color_pairs[pair] != result) {
	int y, x;
	attr_t z = COLOR_PAIR(pair);

	for (y = 0; y <= curscr->_maxy; y++) {
	    struct ldat *ptr = &(curscr->_line[y]);
	    bool changed = FALSE;
	    for (x = 0; x <= curscr->_maxx; x++) {
		if ((ptr->text[x] & A_COLOR) == z) {
		    /* Set the old cell to zero to ensure it will be
		       updated on the next doupdate() */
		    ptr->text[x] = 0;
		    CHANGED_CELL(ptr, x);
		    changed = TRUE;
		}
	    }
	    if (changed)
		_nc_make_oldhash(y);
	}
    }
    SP->_color_pairs[pair] = result;
    if ((int) (SP->_current_attr & A_COLOR) == COLOR_PAIR(pair))
	SP->_current_attr |= A_COLOR;   /* force attribute update */

    if (initialize_pair) {
	const color_t *tp = hue_lightness_saturation ? hls_palette : cga_palette;

	T(("initializing pair: pair = %d, fg=(%d,%d,%d), bg=(%d,%d,%d)",
		pair,
		tp[f].red, tp[f].green, tp[f].blue,
		tp[b].red, tp[b].green, tp[b].blue));

	if (initialize_pair) {
	    putp(tparm(initialize_pair,
		    pair,
		    tp[f].red, tp[f].green, tp[f].blue,
		    tp[b].red, tp[b].green, tp[b].blue));
	}
    }

    returnCode(OK);
}

bool
has_colors(void)
{
    T((T_CALLED("has_colors()")));
    returnCode((VALID_NUMERIC(max_colors) && VALID_NUMERIC(max_pairs)
	    && (((set_foreground != NULL)
		    && (set_background != NULL))
		|| ((set_a_foreground != NULL)
		    && (set_a_background != NULL))
		|| set_color_pair)) ? TRUE : FALSE);
}

static int
pair_content(short pair, short *f, short *b)
{
    T((T_CALLED("pair_content(%d,%p,%p)"), pair, f, b));

    if ((pair < 0) || (pair >= COLOR_PAIRS))
	returnCode(ERR);
    if (f)
	*f = ((SP->_color_pairs[pair] >> C_SHIFT) & C_MASK);
    if (b)
	*b = (SP->_color_pairs[pair] & C_MASK);

    returnCode(OK);
}

static void
_nc_do_color(int old_pair, int pair, bool reverse, int (*outc) (int))
{
    NCURSES_COLOR_T fg = C_MASK, bg = C_MASK;
    NCURSES_COLOR_T old_fg, old_bg;

    if (pair < 0 || pair >= COLOR_PAIRS) {
	return;
    } else if (pair != 0) {
	if (set_color_pair) {
	    tputs(tparm(set_color_pair, pair), 1, outc);
	    return;
	} else if (SP != 0) {
	    pair_content(pair, &fg, &bg);
	}
    }

    if (old_pair >= 0 && SP != 0) {
	old_fg = old_bg = 0;
	pair_content(old_pair, &old_fg, &old_bg);
	if ((fg == C_MASK && old_fg != C_MASK)
	    || (bg == C_MASK && old_bg != C_MASK)) {
#if NCURSES_EXT_FUNCS
	    /*
	     * A minor optimization - but extension.  If "AX" is specified in
	     * the terminal description, treat it as screen's indicator of ECMA
	     * SGR 39 and SGR 49, and assume the two sequences are independent.
	     */
	    if (SP->_has_sgr_39_49 && old_bg == C_MASK && old_fg != C_MASK) {
		tputs("\033[39m", 1, outc);
	    } else if (SP->_has_sgr_39_49 && old_fg == C_MASK && old_bg != C_MASK) {
		tputs("\033[49m", 1, outc);
	    } else
#endif
		set_original_colors();
	}
    } else {
	set_original_colors();
	if (old_pair < 0)
	    return;
    }

#if NCURSES_EXT_FUNCS
    if (fg == C_MASK)
	fg = default_fg();
    if (bg == C_MASK)
	bg = default_bg();
#endif

    if (reverse) {
	NCURSES_COLOR_T xx = fg;
	fg = bg;
	bg = xx;
    }

    TR(TRACE_ATTRS, ("setting colors: pair = %d, fg = %d, bg = %d", pair, fg, bg));

    if (fg != C_MASK) {
	set_foreground_color(fg, outc);
    }
    if (bg != C_MASK) {
	set_background_color(bg, outc);
    }
}
/*
 * Module that "owns" the 'cur_term' variable:
 *
 *      TERMINAL *set_curterm(TERMINAL *)
 */

static TERMINAL *cur_term = 0;

static TERMINAL *set_curterm(TERMINAL *termp)
{
	TERMINAL *oldterm = cur_term;

	if ((cur_term = termp) != 0) {
		ospeed = _nc_ospeed(cur_term->_baudrate);
		PC = (pad_char != NULL) ? pad_char[0] : 0;
	}
	return oldterm;
}


/*
**      lib_data.c
**
**      Common data that may/may not be allocated, but is referenced globally
**
*/


/*
 * The variable 'SP' will be defined as a function on systems that cannot link
 * data-only modules, since it is used in a lot of places within ncurses and we
 * cannot guarantee that any application will use any particular function.  We
 * put the WINDOW variables in this module, because it appears that any
 * application that uses them will also use 'SP'.
 *
 * This module intentionally does not reference other ncurses modules, to avoid
 * module coupling that increases the size of the executable.
 */
static SCREEN *SP;              /* Some linkers require initialized data... */

/*
**      lib_delwin.c
**
**      The routine delwin().
**
*/
static bool have_children(WINDOW *win)
{
	WINDOWLIST *p;
	for (p = _nc_windows; p != 0; p = p->next) {
		if (p->win->_flags & _SUBWIN
		 && p->win->_parent == win)
			return TRUE;
	}
	return FALSE;
}

#define touchwin(win)           wtouchln((win), 0, getmaxy(win), 1)
int delwin(WINDOW *win)
{
	T((T_CALLED("delwin(%p)"), win));

	if (win == 0
	 || have_children(win))
		returnCode(ERR);

	if (win->_flags & _SUBWIN)
		touchwin(win->_parent);
	else if (curscr != 0)
		touchwin(curscr);

	_nc_freewin(win);

	returnCode(OK);
}

/*
 *      echo.c
 *
 *      Routines:
 *              noecho()
 *
 */


int noecho(void)
{
	T((T_CALLED("noecho()")));
	SP->_echo = FALSE;
	returnCode(OK);
}

/*
**      lib_endwin.c
**
**      The routine endwin().
**
*/


int
endwin(void)
{
	T((T_CALLED("endwin()")));

	if (SP) {
		SP->_endwin = TRUE;
		SP->_mouse_wrap(SP);
		_nc_screen_wrap();
		_nc_mvcur_wrap();       /* wrap up cursor addressing */
		returnCode(reset_shell_mode());
	}

	returnCode(ERR);
}

/*
 *      flash.c
 *
 *      The routine flash().
 *
 */

/*
 *      flash()
 *
 *      Flash the current terminal's screen if possible.   If not,
 *      sound the audible bell if one exists.
 *
 */

int flash(void)
{
	int res = ERR;

	T((T_CALLED("flash()")));

	/* FIXME: should make sure that we are not in altchar mode */
	if (flash_screen) {
		res = putp(flash_screen);
		_nc_flush();
	} else if (bell) {
		res = putp(bell);
		_nc_flush();
	}

	returnCode(res);
}

/*
 * DO NOT EDIT THIS FILE BY HAND!
 * It is generated by MKlib_gen.sh.
 *
 * This is a file of trivial functions generated from macro
 * definitions in curses.h to satisfy the XSI Curses requirement
 * that every macro also exist as a callable function.
 *
 * It will never be linked unless you call one of the entry
 * points with its normal macro definition disabled. In that
 * case, if you have no shared libraries, it will indirectly
 * pull most of the rest of the library into your link image.
 */


#undef COLOR_PAIR
static int COLOR_PAIR ( int z)
{
	T((T_CALLED("COLOR_PAIR(%d)"), z)); returnCode( ((z) << ((0) + 8)));
}


#undef mvwaddch
static int mvwaddch ( WINDOW * a1, int a2, int a3, const chtype z)
{
	T((T_CALLED("mvwaddch(%p,%d,%d,%s)"), a1, a2, a3, _tracechtype2(3,z))); returnCode( (wmove(a1,a2,a3) == (-1) ? (-1) : waddch(a1,z)));
}


#undef mvwhline
static int mvwhline ( WINDOW * a1, int a2, int a3, chtype a4, int z)
{
	T((T_CALLED("mvwhline(%p,%d,%d,%s,%d)"), a1, a2, a3, _tracechtype2(3,a4), z)); returnCode( (wmove(a1,a2,a3) == (-1) ? (-1) : whline(a1,a4,z)));
}


#undef PAIR_NUMBER
static int PAIR_NUMBER ( int z)
{
	T((T_CALLED("PAIR_NUMBER(%d)"), z)); returnCode( (((z) & ((((1UL) << 8) - 1UL) << ((0) + 8))) >> 8));
}



#undef touchline
static int touchline ( WINDOW * a1, int a2, int z)
{
	T((T_CALLED("touchline(%p,%d,%d)"), a1, a2, z)); returnCode( wtouchln((a1), a2, z, 1));
}


#undef waddstr
static int waddstr ( WINDOW * a1, const char * z)
{
	T((T_CALLED("waddstr(%p,%s)"), a1, _nc_visbuf2(1,z))); returnCode( waddnstr(a1,z,-1));
}


#undef wattrset
static int wattrset ( WINDOW * a1, int z)
{
	T((T_CALLED("wattrset(%p,%d)"), a1, z)); returnCode( ((a1)->_attrs = (z)));
}

#undef mvwaddstr
static int mvwaddstr ( WINDOW * a1, int a2, int a3, const char * z)
{
	T((T_CALLED("mvwaddstr(%p,%d,%d,%s)"), a1, a2, a3, _nc_visbuf2(3,z))); returnCode( (wmove(a1,a2,a3) == (-1) ? (-1) : waddnstr(a1,z,-1)));
}

/*
**      lib_getch.c
**
**      The routine getch().
**
*/


static int ESCDELAY = 1000;            /* max interval betw. chars in funkeys, in millisecs */

static inline int
fifo_peek(void)
{
    int ch = SP->_fifo[peek];
    TR(TRACE_IEVENT, ("peeking at %d", peek));

    p_inc();
    return ch;
}

static inline int
fifo_pull(void)
{
    int ch;
    ch = SP->_fifo[HEAD];
    TR(TRACE_IEVENT, ("pulling %d from %d", ch, HEAD));

    if (peek == HEAD) {
	h_inc();
	peek = HEAD;
    } else
	h_inc();

    return ch;
}

static inline int
fifo_push(void)
{
    int n;
    unsigned int ch;

    if (tail == -1)
	return ERR;

#ifdef HIDE_EINTR
  again:
    errno = 0;
#endif

#if USE_GPM_SUPPORT || defined(USE_EMX_MOUSE)
    if ((SP->_mouse_fd >= 0)
	&& (_nc_timed_wait(3, -1, (int *) 0) & 2)) {
	SP->_mouse_event(SP);
	ch = KEY_MOUSE;
	n = 1;
    } else
#endif
    {
	unsigned char c2 = 0;
	n = read(SP->_ifd, &c2, 1);
	ch = c2 & 0xff;
    }

#ifdef HIDE_EINTR
    /*
     * Under System V curses with non-restarting signals, getch() returns
     * with value ERR when a handled signal keeps it from completing.
     * If signals restart system calls, OTOH, the signal is invisible
     * except to its handler.
     *
     * We don't want this difference to show.  This piece of code
     * tries to make it look like we always have restarting signals.
     */
    if (n <= 0 && errno == EINTR)
	goto again;
#endif

    if ((n == -1) || (n == 0)) {
	TR(TRACE_IEVENT, ("read(%d,&ch,1)=%d, errno=%d", SP->_ifd, n, errno));
	ch = ERR;
    }
    TR(TRACE_IEVENT, ("read %d characters", n));

    SP->_fifo[tail] = ch;
    SP->_fifohold = 0;
    if (HEAD == -1)
	HEAD = peek = tail;
    t_inc();
    TR(TRACE_IEVENT, ("pushed %#x at %d", ch, tail));
    return ch;
}

static inline void
fifo_clear(void)
{
    int i;
    for (i = 0; i < FIFO_SIZE; i++)
	SP->_fifo[i] = 0;
    HEAD = -1;
    tail = peek = 0;
}

static int kgetch(WINDOW *);

#define wgetch_should_refresh(win) (\
	(is_wintouched(win) || (win->_flags & _HASMOVED)) \
	&& !(win->_flags & _ISPAD))

int
wgetch(WINDOW *win)
{
    int ch;

    T((T_CALLED("wgetch(%p)"), win));

    if (!win)
	returnCode(ERR);

    if (cooked_key_in_fifo()) {
	if (wgetch_should_refresh(win))
	    wrefresh(win);

	ch = fifo_pull();
	T(("wgetch returning (pre-cooked): %#x = %s", ch, _trace_key(ch)));
	returnCode(ch);
    }

    /*
     * Handle cooked mode.  Grab a string from the screen,
     * stuff its contents in the FIFO queue, and pop off
     * the first character to return it.
     */
    if (HEAD == -1 && !SP->_raw && !SP->_cbreak) {
	char buf[MAXCOLUMNS], *sp;

	TR(TRACE_IEVENT, ("filling queue in cooked mode"));

	wgetnstr(win, buf, MAXCOLUMNS);

	/* ungetch in reverse order */
	ungetch('\n');
	for (sp = buf + strlen(buf); sp > buf; sp--)
	    ungetch(sp[-1]);

	returnCode(fifo_pull());
    }

    if (wgetch_should_refresh(win))
	wrefresh(win);

    if (!win->_notimeout && (win->_delay >= 0 || SP->_cbreak > 1)) {
	int delay;

	TR(TRACE_IEVENT, ("timed delay in wgetch()"));
	if (SP->_cbreak > 1)
	    delay = (SP->_cbreak - 1) * 100;
	else
	    delay = win->_delay;

	TR(TRACE_IEVENT, ("delay is %d milliseconds", delay));

	if (HEAD == -1)         /* fifo is empty */
	    if (!_nc_timed_wait(3, delay, (int *) 0))
		returnCode(ERR);
	/* else go on to read data available */
    }

    if (win->_use_keypad) {
	/*
	 * This is tricky.  We only want to get special-key
	 * events one at a time.  But we want to accumulate
	 * mouse events until either (a) the mouse logic tells
	 * us it's picked up a complete gesture, or (b)
	 * there's a detectable time lapse after one.
	 *
	 * Note: if the mouse code starts failing to compose
	 * press/release events into clicks, you should probably
	 * increase the wait with mouseinterval().
	 */
	int runcount = 0;

	do {
	    ch = kgetch(win);
	    if (ch == KEY_MOUSE) {
		++runcount;
		if (SP->_mouse_inline(SP))
		    break;
	    }
	} while
	    (ch == KEY_MOUSE
	    && (_nc_timed_wait(3, SP->_maxclick, (int *) 0)
		|| !SP->_mouse_parse(runcount)));
	if (runcount > 0 && ch != KEY_MOUSE) {
	    /* mouse event sequence ended by keystroke, push it */
	    ungetch(ch);
	    ch = KEY_MOUSE;
	}
    } else {
	if (HEAD == -1)
	    fifo_push();
	ch = fifo_pull();
    }

    if (ch == ERR) {
#if USE_SIZECHANGE
	if (SP->_sig_winch) {
	    _nc_update_screensize();
	    /* resizeterm can push KEY_RESIZE */
	    if (cooked_key_in_fifo()) {
		ch = fifo_pull();
		T(("wgetch returning (pre-cooked): %#x = %s", ch, _trace_key(ch)));
		returnCode(ch);
	    }
	}
#endif
	T(("wgetch returning ERR"));
	returnCode(ERR);
    }

    /*
     * If echo() is in effect, display the printable version of the
     * key on the screen.  Carriage return and backspace are treated
     * specially by Solaris curses:
     *
     * If carriage return is defined as a function key in the
     * terminfo, e.g., kent, then Solaris may return either ^J (or ^M
     * if nonl() is set) or KEY_ENTER depending on the echo() mode.
     * We echo before translating carriage return based on nonl(),
     * since the visual result simply moves the cursor to column 0.
     *
     * Backspace is a different matter.  Solaris curses does not
     * translate it to KEY_BACKSPACE if kbs=^H.  This does not depend
     * on the stty modes, but appears to be a hardcoded special case.
     * This is a difference from ncurses, which uses the terminfo entry.
     * However, we provide the same visual result as Solaris, moving the
     * cursor to the left.
     */
    if (SP->_echo && !(win->_flags & _ISPAD)) {
	chtype backup = (ch == KEY_BACKSPACE) ? '\b' : ch;
	if (backup < KEY_MIN)
	    wechochar(win, backup);
    }

    /*
     * Simulate ICRNL mode
     */
    if ((ch == '\r') && SP->_nl)
	ch = '\n';

    /* Strip 8th-bit if so desired.  We do this only for characters that
     * are in the range 128-255, to provide compatibility with terminals
     * that display only 7-bit characters.  Note that 'ch' may be a
     * function key at this point, so we mustn't strip _those_.
     */
    if ((ch < KEY_MIN) && (ch & 0x80))
	if (!SP->_use_meta)
	    ch &= 0x7f;

    T(("wgetch returning : %#x = %s", ch, _trace_key(ch)));

    returnCode(ch);
}

/*
**      int
**      kgetch()
**
**      Get an input character, but take care of keypad sequences, returning
**      an appropriate code when one matches the input.  After each character
**      is received, set an alarm call based on ESCDELAY.  If no more of the
**      sequence is received by the time the alarm goes off, pass through
**      the sequence gotten so far.
**
**      This function must be called when there is no cooked keys in queue.
**      (that is HEAD==-1 || peek==HEAD)
**
*/

static int
kgetch(WINDOW *win GCC_UNUSED)
{
    struct tries *ptr;
    int ch = 0;
    int timeleft = ESCDELAY;

    TR(TRACE_IEVENT, ("kgetch(%p) called", win));

    ptr = SP->_keytry;

    for (;;) {
	if (!raw_key_in_fifo()) {
	    if (fifo_push() == ERR) {
		peek = HEAD;    /* the keys stay uninterpreted */
		return ERR;
	    }
	}
	ch = fifo_peek();
	if (ch >= KEY_MIN) {
	    peek = HEAD;
	    /* assume the key is the last in fifo */
	    t_dec();            /* remove the key */
	    return ch;
	}

	TR(TRACE_IEVENT, ("ch: %s", _trace_key((unsigned char) ch)));
	while ((ptr != NULL) && (ptr->ch != (unsigned char) ch))
	    ptr = ptr->sibling;

	if (ptr == NULL)
	    break;

	if (ptr->value != 0) {  /* sequence terminated */
	    TR(TRACE_IEVENT, ("end of sequence"));
	    if (peek == tail)
		fifo_clear();
	    else
		HEAD = peek;
	    return (ptr->value);
	}

	ptr = ptr->child;

	if (!raw_key_in_fifo()) {
	    TR(TRACE_IEVENT, ("waiting for rest of sequence"));
	    if (!_nc_timed_wait(3, timeleft, &timeleft)) {
		TR(TRACE_IEVENT, ("ran out of time"));
		break;
	    }
	}
    }
    ch = fifo_pull();
    peek = HEAD;
    return ch;
}

/*
**      lib_getstr.c
**
**      The routine wgetstr().
**
*/


/*
 * This wipes out the last character, no matter whether it was a tab, control
 * or other character, and handles reverse wraparound.
 */
static char *WipeOut(WINDOW *win, int y, int x, char *first, char *last, bool echoed)
{
	if (last > first) {
		*--last = '\0';
		if (echoed) {
			int y1 = win->_cury;
			int x1 = win->_curx;

			wmove(win, y, x);
			waddstr(win, first);
			getyx(win, y, x);
			while (win->_cury < y1
			   || (win->_cury == y1 && win->_curx < x1))
				waddch(win, ' ');

			wmove(win, y, x);
		}
	}
	return last;
}

static int wgetnstr(WINDOW *win, char *str, int maxlen)
{
TTY     buf;
bool    oldnl, oldecho, oldraw, oldcbreak;
char    erasec;
char    killc;
char    *oldstr;
int ch;
int     y, x;

	T((T_CALLED("wgetnstr(%p,%p, %d)"), win, str, maxlen));

	if (!win)
	  returnCode(ERR);

	_nc_get_tty_mode(&buf);

	oldnl = SP->_nl;
	oldecho = SP->_echo;
	oldraw = SP->_raw;
	oldcbreak = SP->_cbreak;
	nl();
	noecho();
	noraw();
	cbreak();

	erasec = erasechar();
	killc = killchar();

	oldstr = str;
	getyx(win, y, x);

	if (is_wintouched(win) || (win->_flags & _HASMOVED))
		wrefresh(win);

	while ((ch = wgetch(win)) != ERR) {
		/*
		 * Some terminals (the Wyse-50 is the most common) generate
		 * a \n from the down-arrow key.  With this logic, it's the
		 * user's choice whether to set kcud=\n for wgetch();
		 * terminating *getstr() with \n should work either way.
		 */
		if (ch == '\n'
		 || ch == '\r'
		 || ch == KEY_DOWN
		 || ch == KEY_ENTER) {
			if (oldecho == TRUE
			 && win->_cury == win->_maxy
			 && win->_scroll)
				wechochar(win, '\n');
			break;
		}
		if (ch == erasec || ch == KEY_LEFT || ch == KEY_BACKSPACE) {
			if (str > oldstr) {
				str = WipeOut(win, y, x, oldstr, str, oldecho);
			}
		} else if (ch == killc) {
			while (str > oldstr) {
				str = WipeOut(win, y, x, oldstr, str, oldecho);
			}
		} else if (ch >= KEY_MIN
			   || (maxlen >= 0 && str - oldstr >= maxlen)) {
			beep();
		} else {
			*str++ = ch;
			if (oldecho == TRUE) {
				int oldy = win->_cury;
				if (waddch(win, ch) == ERR) {
					/*
					 * We can't really use the lower-right
					 * corner for input, since it'll mess
					 * up bookkeeping for erases.
					 */
					win->_flags &= ~_WRAPPED;
					waddch(win, ' ');
					str = WipeOut(win, y, x, oldstr, str, oldecho);
					continue;
				} else if (win->_flags & _WRAPPED) {
					/*
					 * If the last waddch forced a wrap &
					 * scroll, adjust our reference point
					 * for erasures.
					 */
					if (win->_scroll
					 && oldy == win->_maxy
					 && win->_cury == win->_maxy) {
						if (--y <= 0) {
							y = 0;
						}
					}
					win->_flags &= ~_WRAPPED;
				}
				wrefresh(win);
			}
		}
	}

	win->_curx = 0;
	win->_flags &= ~_WRAPPED;
	if (win->_cury < win->_maxy)
		win->_cury++;
	wrefresh(win);

	/* Restore with a single I/O call, to fix minor asymmetry between
	 * raw/noraw, etc.
	 */
	SP->_nl = oldnl;
	SP->_echo = oldecho;
	SP->_raw = oldraw;
	SP->_cbreak = oldcbreak;

	_nc_set_tty_mode(&buf);

	*str = '\0';
	if (ch == ERR)
		returnCode(ERR);

	T(("wgetnstr returns %s", _nc_visbuf(oldstr)));

	returnCode(OK);
}

/*
**      lib_has_cap.c
**
**      The routines to query terminal capabilities
**
*/



static bool has_ic(void)
{
	T((T_CALLED("has_ic()")));
	returnCode(cur_term &&
	  (insert_character || parm_ich
	   ||  (enter_insert_mode && exit_insert_mode))
	  &&  (delete_character || parm_dch));
}


/*
**      lib_hline.c
**
**      The routine whline().
**
*/


static int
whline(WINDOW *win, chtype ch, int n)
{
    int code = ERR;
    NCURSES_SIZE_T start;
    NCURSES_SIZE_T end;

    T((T_CALLED("whline(%p,%s,%d)"), win, _tracechtype(ch), n));

    if (win) {
	struct ldat *line = &(win->_line[win->_cury]);

	start = win->_curx;
	end = start + n - 1;
	if (end > win->_maxx)
	    end = win->_maxx;

	CHANGED_RANGE(line, start, end);

	if (ch == 0)
	    ch = ACS_HLINE;
	ch = _nc_render(win, ch);

	while (end >= start) {
	    line->text[end] = ch;
	    end--;
	}

	_nc_synchook(win);
	code = OK;
    }
    returnCode(code);
}

/*
**      lib_initscr.c
**
**      The routines initscr(), and termname().
**
*/

WINDOW *initscr(void)
{
static  bool initialized = FALSE;
const char *name;
int value;

	T((T_CALLED("initscr()")));
	/* Portable applications must not call initscr() more than once */
	if (!initialized) {
		initialized = TRUE;

		if ((name = getenv("TERM")) == 0
		 || *name == '\0')
			name = "unknown";
		if (newterm(name, stdout, stdin) == 0) {
			fprintf(stderr, "Error opening terminal: %s.\n", name);
			exit(EXIT_FAILURE);
		}

		/* allow user to set maximum escape delay from the environment */
		if ((value = _nc_getenv_num("ESCDELAY")) >= 0) {
			ESCDELAY = value;
		}

		/* def_shell_mode - done in newterm/_nc_setupscreen */
		def_prog_mode();
	}
	returnWin(stdscr);
}

/*
 *      lib_kernel.c
 *
 *      Misc. low-level routines:
 *              erasechar()
 *              killchar()
 *              flushinp()
 *
 * The baudrate() and delay_output() functions could logically live here,
 * but are in other modules to reduce the static-link size of programs
 * that use only these facilities.
 */

/*
 *      erasechar()
 *
 *      Return erase character as given in cur_term->Ottyb.
 *
 */

static char
erasechar(void)
{
	T((T_CALLED("erasechar()")));

	if (cur_term != 0) {
#ifdef TERMIOS
		returnCode(cur_term->Ottyb.c_cc[VERASE]);
#else
		returnCode(cur_term->Ottyb.sg_erase);
#endif
	}
	returnCode(ERR);
}



/*
 *      killchar()
 *
 *      Return kill character as given in cur_term->Ottyb.
 *
 */

static char
killchar(void)
{
	T((T_CALLED("killchar()")));

	if (cur_term != 0) {
#ifdef TERMIOS
		returnCode(cur_term->Ottyb.c_cc[VKILL]);
#else
		returnCode(cur_term->Ottyb.sg_kill);
#endif
	}
	returnCode(ERR);
}



/*
 *      flushinp()
 *
 *      Flush any input on cur_term->Filedes
 *
 */

static int flushinp(void)
{
	T((T_CALLED("flushinp()")));

	if (cur_term != 0) {
#ifdef TERMIOS
		tcflush(cur_term->Filedes, TCIFLUSH);
#else
		errno = 0;
		do {
		    ioctl(cur_term->Filedes, TIOCFLUSH, 0);
		} while
		    (errno == EINTR);
#endif
		if (SP) {
			SP->_fifohead = -1;
			SP->_fifotail = 0;
			SP->_fifopeek = 0;
		}
		returnCode(OK);
	}
	returnCode(ERR);
}
/* generated by MKkeyname.awk */


/*
 * This module is intended to encapsulate ncurses's interface to pointing
 * devices.
 *
 * The first method used is xterm's internal mouse-tracking facility.
 * The second is Alessandro Rubini's GPM server.
 *
 * Notes for implementors of new mouse-interface methods:
 *
 * The code is logically split into a lower level that accepts event reports
 * in a device-dependent format and an upper level that parses mouse gestures
 * and filters events.  The mediating data structure is a circular queue of
 * MEVENT structures.
 *
 * Functionally, the lower level's job is to pick up primitive events and
 * put them on the circular queue.  This can happen in one of two ways:
 * either (a) _nc_mouse_event() detects a series of incoming mouse reports
 * and queues them, or (b) code in lib_getch.c detects the kmous prefix in
 * the keyboard input stream and calls _nc_mouse_inline to queue up a series
 * of adjacent mouse reports.
 *
 * In either case, _nc_mouse_parse() should be called after the series is
 * accepted to parse the digested mouse reports (low-level MEVENTs) into
 * a gesture (a high-level or composite MEVENT).
 *
 * Don't be too shy about adding new event types or modifiers, if you can find
 * room for them in the 32-bit mask.  The API is written so that users get
 * feedback on which theoretical event types they won't see when they call
 * mousemask. There's one bit per button (the RESERVED_EVENT bit) not being
 * used yet, and a couple of bits open at the high end.
 */

#ifdef __EMX__
#  include <io.h>
#  define  INCL_DOS
#  define  INCL_VIO
#  define  INCL_KBD
#  define  INCL_MOU
#  define  INCL_DOSPROCESS
#  include <os2.h>              /* Need to include before the others */
#endif


#if USE_GPM_SUPPORT
#ifndef LINT                    /* don't need this for llib-lncurses */
#undef buttons                  /* term.h defines this, and gpm uses it! */
#include <gpm.h>
#include <linux/keyboard.h>     /* defines KG_* macros */
#endif
#endif

#define MY_TRACE TRACE_ICALLS|TRACE_IEVENT

#define INVALID_EVENT   -1

static int mousetype;
#define M_XTERM         -1      /* use xterm's mouse tracking? */
#define M_NONE          0       /* no mouse device */
#define M_GPM           1       /* use GPM */
#define M_QNX           2       /* QNX mouse on console */
#define M_QNX_TERM      3       /* QNX mouse on pterm/xterm (using qansi-m) */

#if USE_GPM_SUPPORT
#ifndef LINT
static Gpm_Connect gpm_connect;
#endif
#endif

static mmask_t eventmask;       /* current event mask */

static bool _nc_mouse_parse(int);
static void _nc_mouse_resume(SCREEN *);
static void _nc_mouse_wrap(SCREEN *);

/* maintain a circular list of mouse events */

/* The definition of the circular list size (EV_MAX), is in curses.priv.h, so
 * wgetch() may refer to the size and call _nc_mouse_parse() before circular
 * list overflow.
 */
static MEVENT events[EV_MAX];   /* hold the last mouse event seen */
static MEVENT *eventp = events; /* next free slot in event queue */
#define NEXT(ep)        ((ep == events + EV_MAX - 1) ? events : ep + 1)
#define PREV(ep)        ((ep == events) ? events + EV_MAX - 1 : ep - 1)

#ifdef USE_EMX_MOUSE

#  define TOP_ROW          0
#  define LEFT_COL         0

static int mouse_wfd;
static int mouse_thread;
static int mouse_activated;
static char mouse_buttons[] =
{0, 1, 3, 2};

#  define M_FD(sp) sp->_mouse_fd

static void
write_event(int down, int button, int x, int y)
{
    char buf[6];
    unsigned long ignore;

    strncpy(buf, key_mouse, 3); /* should be "\033[M" */
    buf[3] = ' ' + (button - 1) + (down ? 0 : 0x40);
    buf[4] = ' ' + x - LEFT_COL + 1;
    buf[5] = ' ' + y - TOP_ROW + 1;
    DosWrite(mouse_wfd, buf, 6, &ignore);
}

static void
mouse_server(unsigned long ignored GCC_UNUSED)
{
    unsigned short fWait = MOU_WAIT;
    /* NOPTRRECT mourt = { 0,0,24,79 }; */
    MOUEVENTINFO mouev;
    HMOU hmou;
    unsigned short mask = MOUSE_BN1_DOWN | MOUSE_BN2_DOWN | MOUSE_BN3_DOWN;
    int nbuttons = 3;
    int oldstate = 0;
    char err[80];
    unsigned long rc;

    /* open the handle for the mouse */
    if (MouOpen(NULL, &hmou) == 0) {
	rc = MouSetEventMask(&mask, hmou);
	if (rc) {               /* retry with 2 buttons */
	    mask = MOUSE_BN1_DOWN | MOUSE_BN2_DOWN;
	    rc = MouSetEventMask(&mask, hmou);
	    nbuttons = 2;
	}
	if (rc == 0 && MouDrawPtr(hmou) == 0) {
	    for (;;) {
		/* sit and wait on the event queue */
		rc = MouReadEventQue(&mouev, &fWait, hmou);
		if (rc) {
		    sprintf(err, "Error reading mouse queue, rc=%lu.\r\n", rc);
		    break;
		}
		if (!mouse_activated)
		    goto finish;

		/*
		 * OS/2 numbers a 3-button mouse inconsistently from other
		 * platforms:
		 *      1 = left
		 *      2 = right
		 *      3 = middle.
		 */
		if ((mouev.fs ^ oldstate) & MOUSE_BN1_DOWN)
		    write_event(mouev.fs & MOUSE_BN1_DOWN,
				mouse_buttons[1], mouev.col, mouev.row);
		if ((mouev.fs ^ oldstate) & MOUSE_BN2_DOWN)
		    write_event(mouev.fs & MOUSE_BN2_DOWN,
				mouse_buttons[3], mouev.col, mouev.row);
		if ((mouev.fs ^ oldstate) & MOUSE_BN3_DOWN)
		    write_event(mouev.fs & MOUSE_BN3_DOWN,
				mouse_buttons[2], mouev.col, mouev.row);

	      finish:
		oldstate = mouev.fs;
	    }
	} else
	    sprintf(err, "Error setting event mask, buttons=%d, rc=%lu.\r\n",
		    nbuttons, rc);

	DosWrite(2, err, strlen(err), &rc);
	MouClose(hmou);
    }
    DosExit(EXIT_THREAD, 0L);
}

static void
server_state(const int state)
{                               /* It would be nice to implement pointer-off and stop looping... */
    mouse_activated = state;
}

#endif

static int initialized;

static void
initialize_mousetype(void)
{
    static const char *xterm_kmous = "\033[M";

    /* Try gpm first, because gpm may be configured to run in xterm */
#if USE_GPM_SUPPORT
    /* GPM: initialize connection to gpm server */
    gpm_connect.eventMask = GPM_DOWN | GPM_UP;
    gpm_connect.defaultMask = ~(gpm_connect.eventMask | GPM_HARD);
    gpm_connect.minMod = 0;
    gpm_connect.maxMod = ~((1 << KG_SHIFT) | (1 << KG_SHIFTL) | (1 << KG_SHIFTR));
    if (Gpm_Open(&gpm_connect, 0) >= 0) {       /* returns the file-descriptor */
	mousetype = M_GPM;
	SP->_mouse_fd = gpm_fd;
	return;
    }
#endif

    /* OS/2 VIO */
#ifdef USE_EMX_MOUSE
    if (!mouse_thread
	&& strstr(cur_term->type.term_names, "xterm") == 0
	&& key_mouse) {
	int handles[2];

	if (pipe(handles) < 0) {
	    perror("mouse pipe error");
	    return;
	} else {
	    int rc;

	    if (!mouse_buttons[0]) {
		char *s = getenv("MOUSE_BUTTONS_123");

		mouse_buttons[0] = 1;
		if (s && strlen(s) >= 3) {
		    mouse_buttons[1] = s[0] - '0';
		    mouse_buttons[2] = s[1] - '0';
		    mouse_buttons[3] = s[2] - '0';
		}
	    }
	    mouse_wfd = handles[1];
	    M_FD(SP) = handles[0];
	    /* Needed? */
	    setmode(handles[0], O_BINARY);
	    setmode(handles[1], O_BINARY);
	    /* Do not use CRT functions, we may single-threaded. */
	    rc = DosCreateThread((unsigned long *) &mouse_thread,
				 mouse_server, 0, 0, 8192);
	    if (rc) {
		printf("mouse thread error %d=%#x", rc, rc);
		return;
	    } else {
		mousetype = M_XTERM;
		return;
	    }
	}
    }
#endif

    /* we know how to recognize mouse events under "xterm" */
    if (key_mouse != 0) {
	if (!strcmp(key_mouse, xterm_kmous)) {
	    mousetype = M_XTERM;
	    return;
	}
    } else if (strstr(cur_term->type.term_names, "xterm") != 0) {
	(void) _nc_add_to_try(&(SP->_keytry), xterm_kmous, KEY_MOUSE);
	mousetype = M_XTERM;
	return;
    }
}

static void
_nc_mouse_init(void)
/* initialize the mouse */
{
    int i;

    if (!initialized) {
	initialized = TRUE;

	TR(MY_TRACE, ("_nc_mouse_init() called"));

	for (i = 0; i < EV_MAX; i++)
	    events[i].id = INVALID_EVENT;

	initialize_mousetype();

	T(("_nc_mouse_init() set mousetype to %d", mousetype));
    }
}

static bool
_nc_mouse_event(SCREEN * sp GCC_UNUSED)
/* query to see if there is a pending mouse event */
{
#if USE_GPM_SUPPORT
    /* GPM: query server for event, return TRUE if we find one */
    Gpm_Event ev;

    if (gpm_fd >= 0
	&& (_nc_timed_wait(3, 0, (int *) 0) & 2) != 0
	&& Gpm_GetEvent(&ev) == 1) {
	eventp->id = 0;         /* there's only one mouse... */

	eventp->bstate = 0;
	switch (ev.type & 0x0f) {
	case (GPM_DOWN):
	    if (ev.buttons & GPM_B_LEFT)
		eventp->bstate |= BUTTON1_PRESSED;
	    if (ev.buttons & GPM_B_MIDDLE)
		eventp->bstate |= BUTTON2_PRESSED;
	    if (ev.buttons & GPM_B_RIGHT)
		eventp->bstate |= BUTTON3_PRESSED;
	    break;
	case (GPM_UP):
	    if (ev.buttons & GPM_B_LEFT)
		eventp->bstate |= BUTTON1_RELEASED;
	    if (ev.buttons & GPM_B_MIDDLE)
		eventp->bstate |= BUTTON2_RELEASED;
	    if (ev.buttons & GPM_B_RIGHT)
		eventp->bstate |= BUTTON3_RELEASED;
	    break;
	default:
	    break;
	}

	eventp->x = ev.x - 1;
	eventp->y = ev.y - 1;
	eventp->z = 0;

	/* bump the next-free pointer into the circular list */
	eventp = NEXT(eventp);
	return (TRUE);
    }
#endif

#ifdef USE_EMX_MOUSE
    if (SP->_mouse_fd >= 0
	&& (_nc_timed_wait(3, 0, (int *) 0) & 2) != 0) {
	char kbuf[3];

	int i, res = read(M_FD(sp), &kbuf, 3);  /* Eat the prefix */
	if (res != 3)
	    printf("Got %d chars instead of 3 for prefix.\n", res);
	for (i = 0; i < res; i++) {
	    if (kbuf[i] != key_mouse[i])
		printf("Got char %d instead of %d for prefix.\n",
		       (int) kbuf[i], (int) key_mouse[i]);
	}
	return TRUE;
    }
#endif /* USE_EMX_MOUSE */

    /* xterm: never have to query, mouse events are in the keyboard stream */
    return (FALSE);             /* no event waiting */
}

static bool
_nc_mouse_inline(SCREEN * sp)
/* mouse report received in the keyboard stream -- parse its info */
{
    TR(MY_TRACE, ("_nc_mouse_inline() called"));

    if (mousetype == M_XTERM) {
	unsigned char kbuf[4];
	MEVENT *prev;
	size_t grabbed;
	int res;

	/* This code requires that your xterm entry contain the kmous
	 * capability and that it be set to the \E[M documented in the
	 * Xterm Control Sequences reference.  This is how we
	 * arrange for mouse events to be reported via a KEY_MOUSE
	 * return value from wgetch().  After this value is received,
	 * _nc_mouse_inline() gets called and is immediately
	 * responsible for parsing the mouse status information
	 * following the prefix.
	 *
	 * The following quotes from the ctrlseqs.ms document in the
	 * X distribution, describing the X mouse tracking feature:
	 *
	 * Parameters for all mouse tracking escape sequences
	 * generated by xterm encode numeric parameters in a single
	 * character as value+040.  For example, !  is 1.
	 *
	 * On button press or release, xterm sends ESC [ M CbCxCy.
	 * The low two bits of Cb encode button information: 0=MB1
	 * pressed, 1=MB2 pressed, 2=MB3 pressed, 3=release.  The
	 * upper bits encode what modifiers were down when the
	 * button was pressed and are added together.  4=Shift,
	 * 8=Meta, 16=Control.  Cx and Cy are the x and y coordinates
	 * of the mouse event.  The upper left corner is (1,1).
	 *
	 * (End quote)  By the time we get here, we've eaten the
	 * key prefix.  FYI, the loop below is necessary because
	 * mouse click info isn't guaranteed to present as a
	 * single clist item.  It always does under Linux but often
	 * fails to under Solaris.
	 */
	for (grabbed = 0; grabbed < 3; grabbed += res) {

	    /* For VIO mouse we add extra bit 64 to disambiguate button-up. */
#ifdef USE_EMX_MOUSE
	    res = read(M_FD(sp) >= 0 ? M_FD(sp) : sp->_ifd, &kbuf, 3);
#else
	    res = read(sp->_ifd, kbuf + grabbed, 3 - grabbed);
#endif
	    if (res == -1)
		break;
	}
	kbuf[3] = '\0';

	TR(TRACE_IEVENT,
	   ("_nc_mouse_inline sees the following xterm data: '%s'", kbuf));

	eventp->id = 0;         /* there's only one mouse... */

	/* processing code goes here */
	eventp->bstate = 0;
	switch (kbuf[0] & 0x3) {
	case 0x0:
	    eventp->bstate = BUTTON1_PRESSED;
#ifdef USE_EMX_MOUSE
	    if (kbuf[0] & 0x40)
		eventp->bstate = BUTTON1_RELEASED;
#endif
	    break;

	case 0x1:
	    eventp->bstate = BUTTON2_PRESSED;
#ifdef USE_EMX_MOUSE
	    if (kbuf[0] & 0x40)
		eventp->bstate = BUTTON2_RELEASED;
#endif
	    break;

	case 0x2:
	    eventp->bstate = BUTTON3_PRESSED;
#ifdef USE_EMX_MOUSE
	    if (kbuf[0] & 0x40)
		eventp->bstate = BUTTON3_RELEASED;
#endif
	    break;

	case 0x3:
	    /*
	     * Release events aren't reported for individual buttons,
	     * just for the button set as a whole...
	     */
	    eventp->bstate =
		(BUTTON1_RELEASED |
		 BUTTON2_RELEASED |
		 BUTTON3_RELEASED);
	    /*
	     * ...however, because there are no kinds of mouse events under
	     * xterm that can intervene between press and release, we can
	     * deduce which buttons were actually released by looking at the
	     * previous event.
	     */
	    prev = PREV(eventp);
	    if (!(prev->bstate & BUTTON1_PRESSED))
		eventp->bstate &= ~BUTTON1_RELEASED;
	    if (!(prev->bstate & BUTTON2_PRESSED))
		eventp->bstate &= ~BUTTON2_RELEASED;
	    if (!(prev->bstate & BUTTON3_PRESSED))
		eventp->bstate &= ~BUTTON3_RELEASED;
	    break;
	}

	if (kbuf[0] & 4) {
	    eventp->bstate |= BUTTON_SHIFT;
	}
	if (kbuf[0] & 8) {
	    eventp->bstate |= BUTTON_ALT;
	}
	if (kbuf[0] & 16) {
	    eventp->bstate |= BUTTON_CTRL;
	}

	eventp->x = (kbuf[1] - ' ') - 1;
	eventp->y = (kbuf[2] - ' ') - 1;
	TR(MY_TRACE,
	   ("_nc_mouse_inline: primitive mouse-event %s has slot %ld",
	    _tracemouse(eventp),
	    (long) (eventp - events)));

	/* bump the next-free pointer into the circular list */
	eventp = NEXT(eventp);
#if 0                           /* this return would be needed for QNX's mods to lib_getch.c */
	return (TRUE);
#endif
    }

    return (FALSE);
}

static void
mouse_activate(bool on)
{
    if (!on && !initialized)
	return;

    _nc_mouse_init();

    if (on) {

	switch (mousetype) {
	case M_XTERM:
#if NCURSES_EXT_FUNCS
	    keyok(KEY_MOUSE, on);
#endif
#ifdef USE_EMX_MOUSE
	    server_state(1);
#else
	    putp("\033[?1000h");
#endif
	    break;
#if USE_GPM_SUPPORT
	case M_GPM:
	    SP->_mouse_fd = gpm_fd;
	    break;
#endif
	}
	/* Make runtime binding to cut down on object size of applications that
	 * do not use the mouse (e.g., 'clear').
	 */
	SP->_mouse_event = _nc_mouse_event;
	SP->_mouse_inline = _nc_mouse_inline;
	SP->_mouse_parse = _nc_mouse_parse;
	SP->_mouse_resume = _nc_mouse_resume;
	SP->_mouse_wrap = _nc_mouse_wrap;

    } else {

	switch (mousetype) {
	case M_XTERM:
#ifdef USE_EMX_MOUSE
	    server_state(0);
#else
	    putp("\033[?1000l");
#endif
	    break;
#if USE_GPM_SUPPORT
	case M_GPM:
	    break;
#endif
	}
    }
    _nc_flush();
}

/**************************************************************************
 *
 * Device-independent code
 *
 **************************************************************************/

static bool
_nc_mouse_parse(int runcount)
/* parse a run of atomic mouse events into a gesture */
{
    MEVENT *ep, *runp, *next, *prev = PREV(eventp);
    int n;
    bool merge;

    TR(MY_TRACE, ("_nc_mouse_parse(%d) called", runcount));

    /*
     * When we enter this routine, the event list next-free pointer
     * points just past a run of mouse events that we know were separated
     * in time by less than the critical click interval. The job of this
     * routine is to collaps this run into a single higher-level event
     * or gesture.
     *
     * We accomplish this in two passes.  The first pass merges press/release
     * pairs into click events.  The second merges runs of click events into
     * double or triple-click events.
     *
     * It's possible that the run may not resolve to a single event (for
     * example, if the user quadruple-clicks).  If so, leading events
     * in the run are ignored.
     *
     * Note that this routine is independent of the format of the specific
     * format of the pointing-device's reports.  We can use it to parse
     * gestures on anything that reports press/release events on a per-
     * button basis, as long as the device-dependent mouse code puts stuff
     * on the queue in MEVENT format.
     */
    if (runcount == 1) {
	TR(MY_TRACE,
	   ("_nc_mouse_parse: returning simple mouse event %s at slot %ld",
	    _tracemouse(prev),
	    (long) (prev - events)));
	return (prev->id >= 0)
	    ? ((prev->bstate & eventmask) ? TRUE : FALSE)
	    : FALSE;
    }

    /* find the start of the run */
    runp = eventp;
    for (n = runcount; n > 0; n--) {
	runp = PREV(runp);
    }

    /* first pass; merge press/release pairs */
    do {
	merge = FALSE;
	for (ep = runp; next = NEXT(ep), next != eventp; ep = next) {
	    if (ep->x == next->x && ep->y == next->y
		&& (ep->bstate & (BUTTON1_PRESSED | BUTTON2_PRESSED | BUTTON3_PRESSED))
		&& (!(ep->bstate & BUTTON1_PRESSED)
		    == !(next->bstate & BUTTON1_RELEASED))
		&& (!(ep->bstate & BUTTON2_PRESSED)
		    == !(next->bstate & BUTTON2_RELEASED))
		&& (!(ep->bstate & BUTTON3_PRESSED)
		    == !(next->bstate & BUTTON3_RELEASED))
		) {
		if ((eventmask & BUTTON1_CLICKED)
		    && (ep->bstate & BUTTON1_PRESSED)) {
		    ep->bstate &= ~BUTTON1_PRESSED;
		    ep->bstate |= BUTTON1_CLICKED;
		    merge = TRUE;
		}
		if ((eventmask & BUTTON2_CLICKED)
		    && (ep->bstate & BUTTON2_PRESSED)) {
		    ep->bstate &= ~BUTTON2_PRESSED;
		    ep->bstate |= BUTTON2_CLICKED;
		    merge = TRUE;
		}
		if ((eventmask & BUTTON3_CLICKED)
		    && (ep->bstate & BUTTON3_PRESSED)) {
		    ep->bstate &= ~BUTTON3_PRESSED;
		    ep->bstate |= BUTTON3_CLICKED;
		    merge = TRUE;
		}
		if (merge)
		    next->id = INVALID_EVENT;
	    }
	}
    } while
	(merge);

    /*
     * Second pass; merge click runs.  At this point, click events are
     * each followed by one invalid event. We merge click events
     * forward in the queue.
     *
     * NOTE: There is a problem with this design!  If the application
     * allows enough click events to pile up in the circular queue so
     * they wrap around, it will cheerfully merge the newest forward
     * into the oldest, creating a bogus doubleclick and confusing
     * the queue-traversal logic rather badly.  Generally this won't
     * happen, because calling getmouse() marks old events invalid and
     * ineligible for merges.  The true solution to this problem would
     * be to timestamp each MEVENT and perform the obvious sanity check,
     * but the timer element would have to have sub-second resolution,
     * which would get us into portability trouble.
     */
    do {
	MEVENT *follower;

	merge = FALSE;
	for (ep = runp; next = NEXT(ep), next != eventp; ep = next)
	    if (ep->id != INVALID_EVENT) {
		if (next->id != INVALID_EVENT)
		    continue;
		follower = NEXT(next);
		if (follower->id == INVALID_EVENT)
		    continue;

		/* merge click events forward */
		if ((ep->bstate &
		     (BUTTON1_CLICKED | BUTTON2_CLICKED | BUTTON3_CLICKED))
		    && (follower->bstate &
			(BUTTON1_CLICKED | BUTTON2_CLICKED | BUTTON3_CLICKED))) {
		    if ((eventmask & BUTTON1_DOUBLE_CLICKED)
			&& (follower->bstate & BUTTON1_CLICKED)) {
			follower->bstate &= ~BUTTON1_CLICKED;
			follower->bstate |= BUTTON1_DOUBLE_CLICKED;
			merge = TRUE;
		    }
		    if ((eventmask & BUTTON2_DOUBLE_CLICKED)
			&& (follower->bstate & BUTTON2_CLICKED)) {
			follower->bstate &= ~BUTTON2_CLICKED;
			follower->bstate |= BUTTON2_DOUBLE_CLICKED;
			merge = TRUE;
		    }
		    if ((eventmask & BUTTON3_DOUBLE_CLICKED)
			&& (follower->bstate & BUTTON3_CLICKED)) {
			follower->bstate &= ~BUTTON3_CLICKED;
			follower->bstate |= BUTTON3_DOUBLE_CLICKED;
			merge = TRUE;
		    }
		    if (merge)
			ep->id = INVALID_EVENT;
		}

		/* merge double-click events forward */
		if ((ep->bstate &
		     (BUTTON1_DOUBLE_CLICKED
		      | BUTTON2_DOUBLE_CLICKED
		      | BUTTON3_DOUBLE_CLICKED))
		    && (follower->bstate &
			(BUTTON1_CLICKED | BUTTON2_CLICKED | BUTTON3_CLICKED))) {
		    if ((eventmask & BUTTON1_TRIPLE_CLICKED)
			&& (follower->bstate & BUTTON1_CLICKED)) {
			follower->bstate &= ~BUTTON1_CLICKED;
			follower->bstate |= BUTTON1_TRIPLE_CLICKED;
			merge = TRUE;
		    }
		    if ((eventmask & BUTTON2_TRIPLE_CLICKED)
			&& (follower->bstate & BUTTON2_CLICKED)) {
			follower->bstate &= ~BUTTON2_CLICKED;
			follower->bstate |= BUTTON2_TRIPLE_CLICKED;
			merge = TRUE;
		    }
		    if ((eventmask & BUTTON3_TRIPLE_CLICKED)
			&& (follower->bstate & BUTTON3_CLICKED)) {
			follower->bstate &= ~BUTTON3_CLICKED;
			follower->bstate |= BUTTON3_TRIPLE_CLICKED;
			merge = TRUE;
		    }
		    if (merge)
			ep->id = INVALID_EVENT;
		}
	    }
    } while
	(merge);

    /*
     * Now try to throw away trailing events flagged invalid, or that
     * don't match the current event mask.
     */
    for (; runcount; prev = PREV(eventp), runcount--)
	if (prev->id == INVALID_EVENT || !(prev->bstate & eventmask)) {
	    eventp = prev;
	}

    /* after all this, do we have a valid event? */
    return (PREV(eventp)->id != INVALID_EVENT);
}

static void
_nc_mouse_wrap(SCREEN * sp GCC_UNUSED)
/* release mouse -- called by endwin() before shellout/exit */
{
    TR(MY_TRACE, ("_nc_mouse_wrap() called"));

    switch (mousetype) {
    case M_XTERM:
	if (eventmask)
	    mouse_activate(FALSE);
	break;
#if USE_GPM_SUPPORT
	/* GPM: pass all mouse events to next client */
    case M_GPM:
	break;
#endif
    }
}

static void
_nc_mouse_resume(SCREEN * sp GCC_UNUSED)
/* re-connect to mouse -- called by doupdate() after shellout */
{
    TR(MY_TRACE, ("_nc_mouse_resume() called"));

    /* xterm: re-enable reporting */
    if (mousetype == M_XTERM && eventmask)
	mouse_activate(TRUE);

    /* GPM: reclaim our event set */
}

/* lib_mouse.c ends here */

/*
**      lib_move.c
**
**      The routine wmove().
**
*/


int
wmove(WINDOW *win, int y, int x)
{
    T((T_CALLED("wmove(%p,%d,%d)"), win, y, x));

    if (win &&
	x >= 0 && x <= win->_maxx &&
	y >= 0 && y <= win->_maxy) {
	win->_curx = (NCURSES_SIZE_T) x;
	win->_cury = (NCURSES_SIZE_T) y;

	win->_flags &= ~_WRAPPED;
	win->_flags |= _HASMOVED;
	returnCode(OK);
    } else
	returnCode(ERR);
}

/*
**      lib_mvcur.c
**
**      The routines for moving the physical cursor and scrolling:
**
**              void _nc_mvcur_init(void)
**
**              void _nc_mvcur_resume(void)
**
**              int mvcur(int old_y, int old_x, int new_y, int new_x)
**
**              void _nc_mvcur_wrap(void)
**
** Comparisons with older movement optimizers:
**    SVr3 curses mvcur() can't use cursor_to_ll or auto_left_margin.
**    4.4BSD curses can't use cuu/cud/cuf/cub/hpa/vpa/tab/cbt for local
** motions.  It doesn't use tactics based on auto_left_margin.  Weirdly
** enough, it doesn't use its own hardware-scrolling routine to scroll up
** destination lines for out-of-bounds addresses!
**    old ncurses optimizer: less accurate cost computations (in fact,
** it was broken and had to be commented out!).
**
** Compile with -DMAIN to build an interactive tester/timer for the movement
** optimizer.  You can use it to investigate the optimizer's behavior.
** You can also use it for tuning the formulas used to determine whether
** or not full optimization is attempted.
**
** This code has a nasty tendency to find bugs in terminfo entries, because it
** exercises the non-cup movement capabilities heavily.  If you think you've
** found a bug, try deleting subsets of the following capabilities (arranged
** in decreasing order of suspiciousness): it, tab, cbt, hpa, vpa, cuu, cud,
** cuf, cub, cuu1, cud1, cuf1, cub1.  It may be that one or more are wrong.
**
** Note: you should expect this code to look like a resource hog in a profile.
** That's because it does a lot of I/O, through the tputs() calls.  The I/O
** cost swamps the computation overhead (and as machines get faster, this
** will become even more true).  Comments in the test exerciser at the end
** go into detail about tuning and how you can gauge the optimizer's
** effectiveness.
**/

/****************************************************************************
 *
 * Constants and macros for optimizer tuning.
 *
 ****************************************************************************/

/*
 * The average overhead of a full optimization computation in character
 * transmission times.  If it's too high, the algorithm will be a bit
 * over-biased toward using cup rather than local motions; if it's too
 * low, the algorithm may spend more time than is strictly optimal
 * looking for non-cup motions.  Profile the optimizer using the `t'
 * command of the exerciser (see below), and round to the nearest integer.
 *
 * Yes, I (esr) thought about computing expected overhead dynamically, say
 * by derivation from a running average of optimizer times.  But the
 * whole point of this optimization is to *decrease* the frequency of
 * system calls. :-)
 */
#define COMPUTE_OVERHEAD        1       /* I use a 90MHz Pentium @ 9.6Kbps */

/*
 * LONG_DIST is the distance we consider to be just as costly to move over as a
 * cup sequence is to emit.  In other words, it's the length of a cup sequence
 * adjusted for average computation overhead.  The magic number is the length
 * of "\033[yy;xxH", the typical cup sequence these days.
 */
#define LONG_DIST               (8 - COMPUTE_OVERHEAD)

/*
 * Tell whether a motion is optimizable by local motions.  Needs to be cheap to
 * compute. In general, all the fast moves go to either the right or left edge
 * of the screen.  So any motion to a location that is (a) further away than
 * LONG_DIST and (b) further inward from the right or left edge than LONG_DIST,
 * we'll consider nonlocal.
 */
#define NOT_LOCAL(fy, fx, ty, tx)       ((tx > LONG_DIST) && (tx < screen_lines - 1 - LONG_DIST) && (abs(ty-fy) + abs(tx-fx) > LONG_DIST))

/****************************************************************************
 *
 * External interfaces
 *
 ****************************************************************************/

/*
 * For this code to work OK, the following components must live in the
 * screen structure:
 *
 *      int             _char_padding;  // cost of character put
 *      int             _cr_cost;       // cost of (carriage_return)
 *      int             _cup_cost;      // cost of (cursor_address)
 *      int             _home_cost;     // cost of (cursor_home)
 *      int             _ll_cost;       // cost of (cursor_to_ll)
 *      int             _cub1_cost;     // cost of (cursor_left)
 *      int             _cuf1_cost;     // cost of (cursor_right)
 *      int             _cud1_cost;     // cost of (cursor_down)
 *      int             _cuu1_cost;     // cost of (cursor_up)
 *      int             _cub_cost;      // cost of (parm_cursor_left)
 *      int             _cuf_cost;      // cost of (parm_cursor_right)
 *      int             _cud_cost;      // cost of (parm_cursor_down)
 *      int             _cuu_cost;      // cost of (parm_cursor_up)
 *      int             _hpa_cost;      // cost of (column_address)
 *      int             _vpa_cost;      // cost of (row_address)
 *      int             _ech_cost;      // cost of (erase_chars)
 *      int             _rep_cost;      // cost of (repeat_char)
 *
 */

#define CURRENT_ROW     SP->_cursrow    /* phys cursor row */
#define CURRENT_COLUMN  SP->_curscol    /* phys cursor column */
#define CURRENT_ATTR    SP->_current_attr       /* current phys attribute */
#define REAL_ATTR       SP->_current_attr       /* phys current attribute */
#define WANT_CHAR(y, x) SP->_newscr->_line[y].text[x]   /* desired state */
#define BAUDRATE        cur_term->_baudrate     /* bits per second */

#define OPT_SIZE 512

static int normalized_cost(const char *const cap, int affcnt);

/****************************************************************************
 *
 * Initialization/wrapup (including cost pre-computation)
 *
 ****************************************************************************/

#define CostOf(cap,affcnt) _nc_msec_cost(cap,affcnt);
#define NormalizedCost(cap,affcnt) normalized_cost(cap,affcnt);

static int
_nc_msec_cost(const char *const cap, int affcnt)
/* compute the cost of a given operation */
{
    if (cap == 0)
	return (INFINITY);
    else {
	const char *cp;
	float cum_cost = 0.0;

	for (cp = cap; *cp; cp++) {
	    /* extract padding, either mandatory or required */
	    if (cp[0] == '$' && cp[1] == '<' && strchr(cp, '>')) {
		float number = 0.0;

		for (cp += 2; *cp != '>'; cp++) {
		    if (isdigit(*cp))
			number = number * 10 + (*cp - '0');
		    else if (*cp == '*')
			number *= affcnt;
		    else if (*cp == '.' && (*++cp != '>') && isdigit(*cp))
			number += (*cp - '0') / 10.0;
		}

		if (!(SP->_no_padding))
		    cum_cost += number * 10;
	    } else
		cum_cost += SP->_char_padding;
	}

	return ((int) cum_cost);
    }
}

static int
normalized_cost(const char *const cap, int affcnt)
/* compute the effective character-count for an operation (round up) */
{
    int cost = _nc_msec_cost(cap, affcnt);
    if (cost != INFINITY)
	cost = (cost + SP->_char_padding - 1) / SP->_char_padding;
    return cost;
}

static void
reset_scroll_region(void)
/* Set the scroll-region to a known state (the default) */
{
    if (change_scroll_region) {
	putp(tparm(change_scroll_region, 0, screen_lines - 1));
    }
}

static void
_nc_mvcur_resume(void)
/* what to do at initialization time and after each shellout */
{
    /* initialize screen for cursor access */
    if (enter_ca_mode) {
	putp(enter_ca_mode);
    }

    /*
     * Doing this here rather than in _nc_mvcur_wrap() ensures that
     * ncurses programs will see a reset scroll region even if a
     * program that messed with it died ungracefully.
     *
     * This also undoes the effects of terminal init strings that assume
     * they know the screen size.  This is useful when you're running
     * a vt100 emulation through xterm.
     */
    reset_scroll_region();
    SP->_cursrow = SP->_curscol = -1;

    /* restore cursor shape */
    if (SP->_cursor != -1) {
	int cursor = SP->_cursor;
	SP->_cursor = -1;
	curs_set(cursor);
    }
}

static void
_nc_mvcur_init(void)
/* initialize the cost structure */
{
    /*
     * 9 = 7 bits + 1 parity + 1 stop.
     */
    SP->_char_padding = (9 * 1000 * 10) / (BAUDRATE > 0 ? BAUDRATE : 9600);
    if (SP->_char_padding <= 0)
	SP->_char_padding = 1;  /* must be nonzero */
    TR(TRACE_CHARPUT | TRACE_MOVE, ("char_padding %d msecs", SP->_char_padding));

    /* non-parameterized local-motion strings */
    SP->_cr_cost = CostOf(carriage_return, 0);
    SP->_home_cost = CostOf(cursor_home, 0);
    SP->_ll_cost = CostOf(cursor_to_ll, 0);
    SP->_cub1_cost = CostOf(cursor_left, 0);
    SP->_cuf1_cost = CostOf(cursor_right, 0);
    SP->_cud1_cost = CostOf(cursor_down, 0);
    SP->_cuu1_cost = CostOf(cursor_up, 0);

    SP->_smir_cost = CostOf(enter_insert_mode, 0);
    SP->_rmir_cost = CostOf(exit_insert_mode, 0);
    SP->_ip_cost = 0;
    if (insert_padding) {
	SP->_ip_cost = CostOf(insert_padding, 0);
    }

    /*
     * Assumption: if the terminal has memory_relative addressing, the
     * initialization strings or smcup will set single-page mode so we
     * can treat it like absolute screen addressing.  This seems to be true
     * for all cursor_mem_address terminal types in the terminfo database.
     */
    SP->_address_cursor = cursor_address ? cursor_address : cursor_mem_address;

    /*
     * Parametrized local-motion strings.  This static cost computation
     * depends on the following assumptions:
     *
     * (1) They never have * padding.  In the entire master terminfo database
     *     as of March 1995, only the obsolete Zenith Z-100 pc violates this.
     *     (Proportional padding is found mainly in insert, delete and scroll
     *     capabilities).
     *
     * (2) The average case of cup has two two-digit parameters.  Strictly,
     *     the average case for a 24 * 80 screen has ((10*10*(1 + 1)) +
     *     (14*10*(1 + 2)) + (10*70*(2 + 1)) + (14*70*4)) / (24*80) = 3.458
     *     digits of parameters.  On a 25x80 screen the average is 3.6197.
     *     On larger screens the value gets much closer to 4.
     *
     * (3) The average case of cub/cuf/hpa/ech/rep has 2 digits of parameters
     *     (strictly, (((10 * 1) + (70 * 2)) / 80) = 1.8750).
     *
     * (4) The average case of cud/cuu/vpa has 2 digits of parameters
     *     (strictly, (((10 * 1) + (14 * 2)) / 24) = 1.5833).
     *
     * All these averages depend on the assumption that all parameter values
     * are equally probable.
     */
    SP->_cup_cost = CostOf(tparm(SP->_address_cursor, 23, 23), 1);
    SP->_cub_cost = CostOf(tparm(parm_left_cursor, 23), 1);
    SP->_cuf_cost = CostOf(tparm(parm_right_cursor, 23), 1);
    SP->_cud_cost = CostOf(tparm(parm_down_cursor, 23), 1);
    SP->_cuu_cost = CostOf(tparm(parm_up_cursor, 23), 1);
    SP->_hpa_cost = CostOf(tparm(column_address, 23), 1);
    SP->_vpa_cost = CostOf(tparm(row_address, 23), 1);

    /* non-parameterized screen-update strings */
    SP->_ed_cost = NormalizedCost(clr_eos, 1);
    SP->_el_cost = NormalizedCost(clr_eol, 1);
    SP->_el1_cost = NormalizedCost(clr_bol, 1);
    SP->_dch1_cost = NormalizedCost(delete_character, 1);
    SP->_ich1_cost = NormalizedCost(insert_character, 1);

    /* parameterized screen-update strings */
    SP->_dch_cost = NormalizedCost(tparm(parm_dch, 23), 1);
    SP->_ich_cost = NormalizedCost(tparm(parm_ich, 23), 1);
    SP->_ech_cost = NormalizedCost(tparm(erase_chars, 23), 1);
    SP->_rep_cost = NormalizedCost(tparm(repeat_char, ' ', 23), 1);

    SP->_cup_ch_cost = NormalizedCost(tparm(SP->_address_cursor, 23, 23), 1);
    SP->_hpa_ch_cost = NormalizedCost(tparm(column_address, 23), 1);
    SP->_cuf_ch_cost = NormalizedCost(tparm(parm_right_cursor, 23), 1);
    SP->_inline_cost = min(SP->_cup_ch_cost,
			   min(SP->_hpa_ch_cost,
			       SP->_cuf_ch_cost));

    /*
     * If save_cursor is used within enter_ca_mode, we should not use it for
     * scrolling optimization, since the corresponding restore_cursor is not
     * nested on the various terminals (vt100, xterm, etc.) which use this
     * feature.
     */
    if (save_cursor != 0
	&& enter_ca_mode != 0
	&& strstr(enter_ca_mode, save_cursor) != 0) {
	T(("...suppressed sc/rc capability due to conflict with smcup/rmcup"));
	save_cursor = 0;
	restore_cursor = 0;
    }

    /*
     * A different, possibly better way to arrange this would be to set
     * SP->_endwin = TRUE at window initialization time and let this be
     * called by doupdate's return-from-shellout code.
     */
    _nc_mvcur_resume();
}

static void
_nc_mvcur_wrap(void)
/* wrap up cursor-addressing mode */
{
    /* leave cursor at screen bottom */
    mvcur(-1, -1, screen_lines - 1, 0);

    /* set cursor to normal mode */
    if (SP->_cursor != -1)
	curs_set(1);

    if (exit_ca_mode) {
	putp(exit_ca_mode);
    }
    /*
     * Reset terminal's tab counter.  There's a long-time bug that
     * if you exit a "curses" program such as vi or more, tab
     * forward, and then backspace, the cursor doesn't go to the
     * right place.  The problem is that the kernel counts the
     * escape sequences that reset things as column positions.
     * Utter a \r to reset this invisibly.
     */
    _nc_outch('\r');
}

/****************************************************************************
 *
 * Optimized cursor movement
 *
 ****************************************************************************/

/*
 * Perform repeated-append, returning cost
 */
static inline int
repeated_append(string_desc * target, int total, int num, int repeat, const char *src)
{
    size_t need = repeat * strlen(src);

    if (need < target->s_size) {
	while (repeat-- > 0) {
	    if (_nc_safe_strcat(target, src)) {
		total += num;
	    } else {
		total = INFINITY;
		break;
	    }
	}
    } else {
	total = INFINITY;
    }
    return total;
}

#ifndef NO_OPTIMIZE
#define NEXTTAB(fr)     (fr + init_tabs - (fr % init_tabs))

/*
 * Assume back_tab (CBT) does not wrap backwards at the left margin, return
 * a negative value at that point to simplify the loop.
 */
#define LASTTAB(fr)     ((fr > 0) ? ((fr - 1) / init_tabs) * init_tabs : -1)

static int
relative_move(string_desc * target, int from_y, int from_x, int to_y, int
	      to_x, bool ovw)
/* move via local motions (cuu/cuu1/cud/cud1/cub1/cub/cuf1/cuf/vpa/hpa) */
{
    string_desc save;
    int n, vcost = 0, hcost = 0;

    (void) _nc_str_copy(&save, target);

    if (to_y != from_y) {
	vcost = INFINITY;

	if (row_address != 0
	    && _nc_safe_strcat(target, tparm(row_address, to_y))) {
	    vcost = SP->_vpa_cost;
	}

	if (to_y > from_y) {
	    n = (to_y - from_y);

	    if (parm_down_cursor
		&& SP->_cud_cost < vcost
		&& _nc_safe_strcat(_nc_str_copy(target, &save),
				   tparm(parm_down_cursor, n))) {
		vcost = SP->_cud_cost;
	    }

	    if (cursor_down && (n * SP->_cud1_cost < vcost)) {
		vcost = repeated_append(_nc_str_copy(target, &save), 0,
					SP->_cud1_cost, n, cursor_down);
	    }
	} else {                /* (to_y < from_y) */
	    n = (from_y - to_y);

	    if (parm_up_cursor
		&& SP->_cup_cost < vcost
		&& _nc_safe_strcat(_nc_str_copy(target, &save),
				   tparm(parm_up_cursor, n))) {
		vcost = SP->_cup_cost;
	    }

	    if (cursor_up && (n * SP->_cuu1_cost < vcost)) {
		vcost = repeated_append(_nc_str_copy(target, &save), 0,
					SP->_cuu1_cost, n, cursor_up);
	    }
	}

	if (vcost == INFINITY)
	    return (INFINITY);
    }

    save = *target;

    if (to_x != from_x) {
	char str[OPT_SIZE];
	string_desc check;

	hcost = INFINITY;

	if (column_address
	    && _nc_safe_strcat(_nc_str_copy(target, &save),
			       tparm(column_address, to_x))) {
	    hcost = SP->_hpa_cost;
	}

	if (to_x > from_x) {
	    n = to_x - from_x;

	    if (parm_right_cursor
		&& SP->_cuf_cost < hcost
		&& _nc_safe_strcat(_nc_str_copy(target, &save),
				   tparm(parm_right_cursor, n))) {
		hcost = SP->_cuf_cost;
	    }

	    if (cursor_right) {
		int lhcost = 0;

		(void) _nc_str_init(&check, str, sizeof(str));

#if defined(REAL_ATTR) && defined(WANT_CHAR)
		/*
		 * If we have no attribute changes, overwrite is cheaper.
		 * Note: must suppress this by passing in ovw = FALSE whenever
		 * WANT_CHAR would return invalid data.  In particular, this
		 * is true between the time a hardware scroll has been done
		 * and the time the structure WANT_CHAR would access has been
		 * updated.
		 */
		if (ovw) {
		    int i;

		    for (i = 0; i < n; i++)
			if ((WANT_CHAR(to_y, from_x + i) & A_ATTRIBUTES) != CURRENT_ATTR) {
			    ovw = FALSE;
			    break;
			}
		}
		if (ovw) {
		    int i;

		    for (i = 0; i < n; i++)
			*check.s_tail++ = WANT_CHAR(to_y, from_x + i);
		    *check.s_tail = '\0';
		    check.s_size -= n;
		    lhcost += n * SP->_char_padding;
		} else
#endif /* defined(REAL_ATTR) && defined(WANT_CHAR) */
		{
		    lhcost = repeated_append(&check, lhcost, SP->_cuf1_cost,
					     n, cursor_right);
		}

		if (lhcost < hcost
		    && _nc_safe_strcat(_nc_str_copy(target, &save), str)) {
		    hcost = lhcost;
		}
	    }
	} else {                /* (to_x < from_x) */
	    n = from_x - to_x;

	    if (parm_left_cursor
		&& SP->_cub_cost < hcost
		&& _nc_safe_strcat(_nc_str_copy(target, &save),
				   tparm(parm_left_cursor, n))) {
		hcost = SP->_cub_cost;
	    }

	    if (cursor_left) {
		int lhcost = 0;

		(void) _nc_str_init(&check, str, sizeof(str));

		lhcost = repeated_append(&check, lhcost, SP->_cub1_cost, n, cursor_left);

		if (lhcost < hcost
		    && _nc_safe_strcat(_nc_str_copy(target, &save), str)) {
		    hcost = lhcost;
		}
	    }
	}

	if (hcost == INFINITY)
	    return (INFINITY);
    }

    return (vcost + hcost);
}
#endif /* !NO_OPTIMIZE */

/*
 * With the machinery set up above, it's conceivable that
 * onscreen_mvcur could be modified into a recursive function that does
 * an alpha-beta search of motion space, as though it were a chess
 * move tree, with the weight function being boolean and the search
 * depth equated to length of string.  However, this would jack up the
 * computation cost a lot, especially on terminals without a cup
 * capability constraining the search tree depth.  So we settle for
 * the simpler method below.
 */

static inline int
onscreen_mvcur(int yold, int xold, int ynew, int xnew, bool ovw)
/* onscreen move from (yold, xold) to (ynew, xnew) */
{
    string_desc result;
    char buffer[OPT_SIZE];
    int tactic = 0, newcost, usecost = INFINITY;
    int t5_cr_cost;

#define NullResult _nc_str_null(&result, sizeof(buffer))
#define InitResult _nc_str_init(&result, buffer, sizeof(buffer))

    /* tactic #0: use direct cursor addressing */
    if (_nc_safe_strcpy(InitResult, tparm(SP->_address_cursor, ynew, xnew))) {
	tactic = 0;
	usecost = SP->_cup_cost;

	/*
	 * We may be able to tell in advance that the full optimization
	 * will probably not be worth its overhead.  Also, don't try to
	 * use local movement if the current attribute is anything but
	 * A_NORMAL...there are just too many ways this can screw up
	 * (like, say, local-movement \n getting mapped to some obscure
	 * character because A_ALTCHARSET is on).
	 */
	if (yold == -1 || xold == -1 || NOT_LOCAL(yold, xold, ynew, xnew)) {
	    goto nonlocal;
	}
    }
#ifndef NO_OPTIMIZE
    /* tactic #1: use local movement */
    if (yold != -1 && xold != -1
	&& ((newcost = relative_move(NullResult, yold, xold, ynew, xnew,
				     ovw)) != INFINITY)
	&& newcost < usecost) {
	tactic = 1;
	usecost = newcost;
    }

    /* tactic #2: use carriage-return + local movement */
    if (yold != -1 && carriage_return
	&& ((newcost = relative_move(NullResult, yold, 0, ynew, xnew, ovw))
	    != INFINITY)
	&& SP->_cr_cost + newcost < usecost) {
	tactic = 2;
	usecost = SP->_cr_cost + newcost;
    }

    /* tactic #3: use home-cursor + local movement */
    if (cursor_home
	&& ((newcost = relative_move(NullResult, 0, 0, ynew, xnew, ovw)) != INFINITY)
	&& SP->_home_cost + newcost < usecost) {
	tactic = 3;
	usecost = SP->_home_cost + newcost;
    }

    /* tactic #4: use home-down + local movement */
    if (cursor_to_ll
	&& ((newcost = relative_move(NullResult, screen_lines - 1, 0, ynew,
				     xnew, ovw)) != INFINITY)
	&& SP->_ll_cost + newcost < usecost) {
	tactic = 4;
	usecost = SP->_ll_cost + newcost;
    }

    /*
     * tactic #5: use left margin for wrap to right-hand side,
     * unless strange wrap behavior indicated by xenl might hose us.
     */
    t5_cr_cost = (xold > 0 ? SP->_cr_cost : 0);
    if (auto_left_margin && !eat_newline_glitch
	&& yold > 0 && cursor_left
	&& ((newcost = relative_move(NullResult, yold - 1, screen_columns -
				     1, ynew, xnew, ovw)) != INFINITY)
	&& t5_cr_cost + SP->_cub1_cost + newcost < usecost) {
	tactic = 5;
	usecost = t5_cr_cost + SP->_cub1_cost + newcost;
    }

    /*
     * These cases are ordered by estimated relative frequency.
     */
    if (tactic)
	InitResult;
    switch (tactic) {
    case 1:
	(void) relative_move(&result, yold, xold, ynew, xnew, ovw);
	break;
    case 2:
	(void) _nc_safe_strcpy(&result, carriage_return);
	(void) relative_move(&result, yold, 0, ynew, xnew, ovw);
	break;
    case 3:
	(void) _nc_safe_strcpy(&result, cursor_home);
	(void) relative_move(&result, 0, 0, ynew, xnew, ovw);
	break;
    case 4:
	(void) _nc_safe_strcpy(&result, cursor_to_ll);
	(void) relative_move(&result, screen_lines - 1, 0, ynew, xnew, ovw);
	break;
    case 5:
	if (xold > 0)
	    (void) _nc_safe_strcat(&result, carriage_return);
	(void) _nc_safe_strcat(&result, cursor_left);
	(void) relative_move(&result, yold - 1, screen_columns - 1, ynew,
			     xnew, ovw);
	break;
    }
#endif /* !NO_OPTIMIZE */

  nonlocal:
    if (usecost != INFINITY) {
	tputs(buffer, 1, _nc_outch);
	return (OK);
    } else
	return (ERR);
}

static int
mvcur(int yold, int xold, int ynew, int xnew)
/* optimized cursor move from (yold, xold) to (ynew, xnew) */
{
    TR(TRACE_MOVE, ("mvcur(%d,%d,%d,%d) called", yold, xold, ynew, xnew));

    if (yold == ynew && xold == xnew)
	return (OK);

    /*
     * Most work here is rounding for terminal boundaries getting the
     * column position implied by wraparound or the lack thereof and
     * rolling up the screen to get ynew on the screen.
     */

    if (xnew >= screen_columns) {
	ynew += xnew / screen_columns;
	xnew %= screen_columns;
    }
    if (xold >= screen_columns) {
	int l;

	l = (xold + 1) / screen_columns;
	yold += l;
	if (yold >= screen_lines)
	    l -= (yold - screen_lines - 1);

	while (l > 0) {
	    if (newline) {
		tputs(newline, 0, _nc_outch);
	    } else
		putchar('\n');
	    l--;
	    if (xold > 0) {
		if (carriage_return) {
		    tputs(carriage_return, 0, _nc_outch);
		} else
		    putchar('\r');
		xold = 0;
	    }
	}
    }

    if (yold > screen_lines - 1)
	yold = screen_lines - 1;
    if (ynew > screen_lines - 1)
	ynew = screen_lines - 1;

    /* destination location is on screen now */
    return (onscreen_mvcur(yold, xold, ynew, xnew, TRUE));
}

/* lib_mvcur.c ends here */

/*
 *      lib_napms.c
 *
 *      The routine napms.
 *
 */


#if HAVE_NANOSLEEP
#if HAVE_SYS_TIME_H
#include <sys/time.h>           /* needed for MacOS X DP3 */
#endif
#elif USE_FUNC_POLL
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#elif HAVE_SELECT
#if HAVE_SYS_TIME_H && HAVE_SYS_TIME_SELECT
#include <sys/time.h>
#endif
#include <sys/select.h>
#endif

static int
napms(int ms)
{
    T((T_CALLED("napms(%d)"), ms));

#if HAVE_NANOSLEEP
    {
	struct timespec ts;
	ts.tv_sec = ms / 1000;
	ts.tv_nsec = (ms % 1000) * 1000000;
	nanosleep(&ts, NULL);
    }
#elif USE_FUNC_POLL
    {
	struct pollfd fds[1];
	poll(fds, 0, ms);
    }
#elif HAVE_SELECT
    {
	struct timeval tval;
	tval.tv_sec = ms / 1000;
	tval.tv_usec = (ms % 1000) * 1000;
	select(0, NULL, NULL, NULL, &tval);
    }
#endif
    returnCode(OK);
}

/*
**      lib_newterm.c
**
**      The newterm() function.
**
*/


#if SVR4_TERMIO && !defined(_POSIX_SOURCE)
#define _POSIX_SOURCE
#endif

#ifndef ONLCR                   /* Allows compilation under the QNX 4.2 OS */
#define ONLCR 0
#endif

/*
 * SVr4/XSI Curses specify that hardware echo is turned off in initscr, and not
 * restored during the curses session.  The library simulates echo in software.
 * (The behavior is unspecified if the application enables hardware echo).
 *
 * The newterm function also initializes terminal settings, and since initscr
 * is supposed to behave as if it calls newterm, we do it here.
 */
static inline int
_nc_initscr(void)
{
    /* for extended XPG4 conformance requires cbreak() at this point */
    /* (SVr4 curses does this anyway) */
    cbreak();

#ifdef TERMIOS
    cur_term->Nttyb.c_lflag &= ~(ECHO | ECHONL);
    cur_term->Nttyb.c_iflag &= ~(ICRNL | INLCR | IGNCR);
    cur_term->Nttyb.c_oflag &= ~(ONLCR);
#else
    cur_term->Nttyb.sg_flags &= ~(ECHO | CRMOD);
#endif
    return _nc_set_tty_mode(&cur_term->Nttyb);
}

/*
 * filter() has to be called before either initscr() or newterm(), so there is
 * apparently no way to make this flag apply to some terminals and not others,
 * aside from possibly delaying a filter() call until some terminals have been
 * initialized.
 */
static int filter_mode = FALSE;

static SCREEN *
newterm(const char *name, FILE * ofp, FILE * ifp)
{
    int errret;
    int slk_format = _nc_slk_format;
    SCREEN *current;

    T((T_CALLED("newterm(\"%s\",%p,%p)"), name, ofp, ifp));

    /* this loads the capability entry, then sets LINES and COLS */
    if (setupterm(name, fileno(ofp), &errret) == ERR)
	return 0;

    /* implement filter mode */
    if (filter_mode) {
	LINES = 1;

	if (VALID_NUMERIC(init_tabs))
	    TABSIZE = init_tabs;
	else
	    TABSIZE = 8;

	T(("TABSIZE = %d", TABSIZE));

	clear_screen = 0;
	cursor_down = parm_down_cursor = 0;
	cursor_address = 0;
	cursor_up = parm_up_cursor = 0;
	row_address = 0;

	cursor_home = carriage_return;
    }

    /* If we must simulate soft labels, grab off the line to be used.
       We assume that we must simulate, if it is none of the standard
       formats (4-4  or 3-2-3) for which there may be some hardware
       support. */
    if (num_labels <= 0 || !SLK_STDFMT(slk_format))
	if (slk_format) {
	    if (ERR == _nc_ripoffline(-SLK_LINES(slk_format),
		    _nc_slk_initialize))
		return 0;
	}
    /* this actually allocates the screen structure, and saves the
     * original terminal settings.
     */
    current = SP;
    _nc_set_screen(0);
    if (_nc_setupscreen(LINES, COLS, ofp) == ERR) {
	_nc_set_screen(current);
	return 0;
    }

    /* if the terminal type has real soft labels, set those up */
    if (slk_format && num_labels > 0 && SLK_STDFMT(slk_format))
	_nc_slk_initialize(stdscr, COLS);

    SP->_ifd = fileno(ifp);
    SP->_checkfd = fileno(ifp);
    typeahead(fileno(ifp));
#ifdef TERMIOS
    SP->_use_meta = ((cur_term->Ottyb.c_cflag & CSIZE) == CS8 &&
	!(cur_term->Ottyb.c_iflag & ISTRIP));
#else
    SP->_use_meta = FALSE;
#endif
    SP->_endwin = FALSE;

    /* Check whether we can optimize scrolling under dumb terminals in case
     * we do not have any of these capabilities, scrolling optimization
     * will be useless.
     */
    SP->_scrolling = ((scroll_forward && scroll_reverse) ||
	((parm_rindex || parm_insert_line || insert_line) &&
	    (parm_index || parm_delete_line || delete_line)));

    baudrate();                 /* sets a field in the SP structure */

    SP->_keytry = 0;

    /*
     * Check for mismatched graphic-rendition capabilities.  Most SVr4
     * terminfo trees contain entries that have rmul or rmso equated to
     * sgr0 (Solaris curses copes with those entries).  We do this only for
     * curses, since many termcap applications assume that smso/rmso and
     * smul/rmul are paired, and will not function properly if we remove
     * rmso or rmul.  Curses applications shouldn't be looking at this
     * detail.
     */
#define SGR0_TEST(mode) (mode != 0) && (exit_attribute_mode == 0 || strcmp(mode, exit_attribute_mode))
    SP->_use_rmso = SGR0_TEST(exit_standout_mode);
    SP->_use_rmul = SGR0_TEST(exit_underline_mode);

    /* compute movement costs so we can do better move optimization */
    _nc_mvcur_init();

    /* initialize terminal to a sane state */
    _nc_screen_init();

    /* Initialize the terminal line settings. */
    _nc_initscr();

    _nc_signal_handler(TRUE);

    T((T_RETURN("%p"), SP));
    return (SP);
}

/*
**      lib_newwin.c
**
**      The routines newwin(), subwin() and their dependent
**
*/


static void
_nc_freewin(WINDOW *win)
{
    WINDOWLIST *p, *q;
    int i;

    if (win != 0) {
	for (p = _nc_windows, q = 0; p != 0; q = p, p = p->next) {
	    if (p->win == win) {
		if (q == 0)
		    _nc_windows = p->next;
		else
		    q->next = p->next;
		free(p);

		if (!(win->_flags & _SUBWIN)) {
		    for (i = 0; i <= win->_maxy; i++)
			FreeIfNeeded(win->_line[i].text);
		}
		free(win->_line);
		free(win);

		if (win == curscr)
		    curscr = 0;
		if (win == stdscr)
		    stdscr = 0;
		if (win == newscr)
		    newscr = 0;

		T(("...deleted win=%p", win));
		break;
	    }
	}
    }
}

WINDOW *
newwin(int num_lines, int num_columns, int begy, int begx)
{
    WINDOW *win;
    chtype *ptr;
    int i;

    T((T_CALLED("newwin(%d,%d,%d,%d)"), num_lines, num_columns, begy, begx));

    if (begy < 0 || begx < 0 || num_lines < 0 || num_columns < 0)
	returnWin(0);

    if (num_lines == 0)
	num_lines = SP->_lines_avail - begy;
    if (num_columns == 0)
	num_columns = screen_columns - begx;

    if (num_columns + begx > SP->_columns || num_lines + begy > SP->_lines_avail)
	returnWin(0);

    if ((win = _nc_makenew(num_lines, num_columns, begy, begx, 0)) == 0)
	returnWin(0);

    for (i = 0; i < num_lines; i++) {
	win->_line[i].text = typeCalloc(chtype, (unsigned) num_columns);
	if (win->_line[i].text == 0) {
	    _nc_freewin(win);
	    returnWin(0);
	}
	for (ptr = win->_line[i].text; ptr < win->_line[i].text +
	    num_columns;)
	    *ptr++ = ' ';
    }

    T(("newwin: returned window is %p", win));

    returnWin(win);
}

static WINDOW *
derwin(WINDOW *orig, int num_lines, int num_columns, int begy, int begx)
{
    WINDOW *win;
    int i;
    int flags = _SUBWIN;

    T((T_CALLED("derwin(%p,%d,%d,%d,%d)"), orig, num_lines, num_columns,
	    begy, begx));

    /*
       ** make sure window fits inside the original one
     */
    if (begy < 0 || begx < 0 || orig == 0 || num_lines < 0 || num_columns < 0)
	returnWin(0);
    if (begy + num_lines > orig->_maxy + 1
	|| begx + num_columns > orig->_maxx + 1)
	returnWin(0);

    if (num_lines == 0)
	num_lines = orig->_maxy + 1 - begy;

    if (num_columns == 0)
	num_columns = orig->_maxx + 1 - begx;

    if (orig->_flags & _ISPAD)
	flags |= _ISPAD;

    if ((win = _nc_makenew(num_lines, num_columns, orig->_begy + begy,
		orig->_begx + begx, flags)) == 0)
	returnWin(0);

    win->_pary = begy;
    win->_parx = begx;
    win->_attrs = orig->_attrs;
    win->_bkgd = orig->_bkgd;

    for (i = 0; i < num_lines; i++)
	win->_line[i].text = &orig->_line[begy++].text[begx];

    win->_parent = orig;

    T(("derwin: returned window is %p", win));

    returnWin(win);
}

WINDOW *
subwin(WINDOW *w, int l, int c, int y, int x)
{
    T((T_CALLED("subwin(%p, %d, %d, %d, %d)"), w, l, c, y, x));
    T(("parent has begy = %d, begx = %d", w->_begy, w->_begx));

    returnWin(derwin(w, l, c, y - w->_begy, x - w->_begx));
}

static bool
dimension_limit(int value)
{
    NCURSES_SIZE_T test = value;
    return (test == value && value > 0);
}

static WINDOW *
_nc_makenew(int num_lines, int num_columns, int begy, int begx, int flags)
{
    int i;
    WINDOWLIST *wp;
    WINDOW *win;
    bool is_pad = (flags & _ISPAD);

    T(("_nc_makenew(%d,%d,%d,%d)", num_lines, num_columns, begy, begx));

    if (!dimension_limit(num_lines) || !dimension_limit(num_columns))
	return 0;

    if ((wp = typeCalloc(WINDOWLIST, 1)) == 0)
	return 0;

    if ((win = typeCalloc(WINDOW, 1)) == 0)
	  return 0;

    if ((win->_line = typeCalloc(struct ldat, ((unsigned) num_lines))) == 0) {
	free(win);
	return 0;
    }

    win->_curx = 0;
    win->_cury = 0;
    win->_maxy = num_lines - 1;
    win->_maxx = num_columns - 1;
    win->_begy = begy;
    win->_begx = begx;
    win->_yoffset = SP->_topstolen;

    win->_flags = flags;
    win->_attrs = A_NORMAL;
    win->_bkgd = BLANK;

    win->_clear = is_pad ? FALSE : (num_lines == screen_lines && num_columns
	== screen_columns);
    win->_idlok = FALSE;
    win->_idcok = TRUE;
    win->_scroll = FALSE;
    win->_leaveok = FALSE;
    win->_use_keypad = FALSE;
    win->_delay = -1;
    win->_immed = FALSE;
    win->_sync = 0;
    win->_parx = -1;
    win->_pary = -1;
    win->_parent = 0;

    win->_regtop = 0;
    win->_regbottom = num_lines - 1;

    win->_pad._pad_y = -1;
    win->_pad._pad_x = -1;
    win->_pad._pad_top = -1;
    win->_pad._pad_bottom = -1;
    win->_pad._pad_left = -1;
    win->_pad._pad_right = -1;

    for (i = 0; i < num_lines; i++) {
	/*
	 * This used to do
	 *
	 * win->_line[i].firstchar = win->_line[i].lastchar = _NOCHANGE;
	 *
	 * which marks the whole window unchanged.  That's how
	 * SVr1 curses did it, but SVr4 curses marks the whole new
	 * window changed.
	 *
	 * With the old SVr1-like code, say you have stdscr full of
	 * characters, then create a new window with newwin(),
	 * then do a printw(win, "foo        ");, the trailing spaces are
	 * completely ignored by the following refreshes.  So, you
	 * get "foojunkjunk" on the screen instead of "foo        " as
	 * you actually intended.
	 *
	 * SVr4 doesn't do this.  Instead the spaces are actually written.
	 * So that's how we want ncurses to behave.
	 */
	win->_line[i].firstchar = 0;
	win->_line[i].lastchar = num_columns - 1;

	if_USE_SCROLL_HINTS(win->_line[i].oldindex = i);
    }

    if (!is_pad && (begx + num_columns == screen_columns)) {
	win->_flags |= _ENDLINE;

	if (begx == 0 && num_lines == screen_lines && begy == 0)
	    win->_flags |= _FULLWIN;

	if (begy + num_lines == screen_lines)
	    win->_flags |= _SCROLLWIN;
    }

    wp->next = _nc_windows;
    wp->win = win;
    _nc_windows = wp;

    T((T_CREATE("window %p"), win));

    return (win);
}

/*
 *      nl.c
 *
 *      Routines:
 *              nl()
 *
 */


static int
nl(void)
{
    T((T_CALLED("nl()")));

    SP->_nl = TRUE;

#ifdef __EMX__
    _nc_flush();
    _fsetmode(NC_OUTPUT, "t");
#endif

    returnCode(OK);
}

/*
**      lib_options.c
**
**      The routines to handle option setting.
**
*/


int
keypad(WINDOW *win, bool flag)
{
    T((T_CALLED("keypad(%p,%d)"), win, flag));

    if (win) {
	win->_use_keypad = flag;
	returnCode(_nc_keypad(flag));
    } else
	returnCode(ERR);
}

/* curs_set() moved here to narrow the kernel interface */

static int
curs_set(int vis)
{
    int cursor = SP->_cursor;

    T((T_CALLED("curs_set(%d)"), vis));

    if (vis < 0 || vis > 2)
	returnCode(ERR);

    if (vis == cursor)
	returnCode(cursor);

    switch (vis) {
    case 2:
	if (cursor_visible) {
	    putp(cursor_visible);
	} else
	    returnCode(ERR);
	break;
    case 1:
	if (cursor_normal) {
	    putp(cursor_normal);
	} else
	    returnCode(ERR);
	break;
    case 0:
	if (cursor_invisible) {
	    putp(cursor_invisible);
	} else
	    returnCode(ERR);
	break;
    }
    SP->_cursor = vis;
    _nc_flush();

    returnCode(cursor == -1 ? 1 : cursor);
}

static int
typeahead(int fd)
{
    T((T_CALLED("typeahead(%d)"), fd));
    SP->_checkfd = fd;
    returnCode(OK);
}

/* Turn the keypad on/off
 *
 * Note:  we flush the output because changing this mode causes some terminals
 * to emit different escape sequences for cursor and keypad keys.  If we don't
 * flush, then the next wgetch may get the escape sequence that corresponds to
 * the terminal state _before_ switching modes.
 */
static int
_nc_keypad(bool flag)
{
    if (flag && keypad_xmit) {
	putp(keypad_xmit);
	_nc_flush();
    } else if (!flag && keypad_local) {
	putp(keypad_local);
	_nc_flush();
    }

    if (flag && !SP->_tried) {
	_nc_init_keytry();
	SP->_tried = TRUE;
    }
    return (OK);
}

/*
**      lib_printw.c
**
**      The routines printw(), wprintw() and friends.
**
*/


int wprintw(WINDOW *win, const char *fmt, ...)
{
	va_list argp;
	int code;

	T(("wprintw(%p,%s,...) called", win, _nc_visbuf(fmt)));

	va_start(argp, fmt);
	code = vwprintw(win, fmt, argp);
	va_end(argp);

	return code;
}

static int vwprintw(WINDOW *win, const char *fmt, va_list argp)
{
	char *buf = _nc_printf_string(fmt, argp);
	int code = ERR;

	if (buf != 0) {
		code = waddstr(win, buf);
	}
	return code;
}

/*
 *      raw.c
 *
 *      Routines:
 *              cbreak()
 *              noraw()
 *              qiflush()
 *              noqiflush()
 *              intrflush()
 *
 */

#define COOKED_INPUT    (IXON|BRKINT|PARMRK)

#define BEFORE(s)
#define AFTER(s)

int
cbreak(void)
{
    T((T_CALLED("cbreak()")));

    SP->_cbreak = 1;

#ifdef __EMX__
    setmode(SP->_ifd, O_BINARY);
#endif

#ifdef TERMIOS
    BEFORE("cbreak");
    cur_term->Nttyb.c_lflag &= ~ICANON;
    cur_term->Nttyb.c_iflag &= ~ICRNL;
    cur_term->Nttyb.c_lflag |= ISIG;
    cur_term->Nttyb.c_cc[VMIN] = 1;
    cur_term->Nttyb.c_cc[VTIME] = 0;
    AFTER("cbreak");
#else
    cur_term->Nttyb.sg_flags |= CBREAK;
#endif
    returnCode(_nc_set_tty_mode(&cur_term->Nttyb));
}

static int
noraw(void)
{
    T((T_CALLED("noraw()")));

    SP->_raw = FALSE;
    SP->_cbreak = 0;

#ifdef __EMX__
    setmode(SP->_ifd, O_TEXT);
#endif

#ifdef TERMIOS
    BEFORE("noraw");
    cur_term->Nttyb.c_lflag |= ISIG | ICANON |
	(cur_term->Ottyb.c_lflag & IEXTEN);
    cur_term->Nttyb.c_iflag |= COOKED_INPUT;
    AFTER("noraw");
#else
    cur_term->Nttyb.sg_flags &= ~(RAW | CBREAK);
#endif
    returnCode(_nc_set_tty_mode(&cur_term->Nttyb));
}

/*
 *      lib_refresh.c
 *
 *      The routines wrefresh() and wnoutrefresh().
 *
 */


int
wrefresh(WINDOW *win)
{
    int code;

    T((T_CALLED("wrefresh(%p)"), win));

    if (win == curscr) {
	curscr->_clear = TRUE;
	code = doupdate();
    } else if ((code = wnoutrefresh(win)) == OK) {
	if (win->_clear)
	    newscr->_clear = TRUE;
	code = doupdate();
	/*
	 * Reset the clearok() flag in case it was set for the special
	 * case in hardscroll.c (if we don't reset it here, we'll get 2
	 * refreshes because the flag is copied from stdscr to newscr).
	 * Resetting the flag shouldn't do any harm, anyway.
	 */
	win->_clear = FALSE;
    }
    returnCode(code);
}

int
wnoutrefresh(WINDOW *win)
{
    NCURSES_SIZE_T limit_x;
    NCURSES_SIZE_T i, j;
    NCURSES_SIZE_T begx;
    NCURSES_SIZE_T begy;
    NCURSES_SIZE_T m, n;
#if USE_SCROLL_HINTS
    bool wide;
#endif
    T((T_CALLED("wnoutrefresh(%p)"), win));

    /*
     * This function will break badly if we try to refresh a pad.
     */
    if ((win == 0)
	|| (win->_flags & _ISPAD))
	returnCode(ERR);

    /* put them here so "win == 0" won't break our code */
    begx = win->_begx;
    begy = win->_begy;

    newscr->_bkgd = win->_bkgd;
    newscr->_attrs = win->_attrs;

    /* merge in change information from all subwindows of this window */
    wsyncdown(win);

#if USE_SCROLL_HINTS
    /*
     * For pure efficiency, we'd want to transfer scrolling information
     * from the window to newscr whenever the window is wide enough that
     * its update will dominate the cost of the update for the horizontal
     * band of newscr that it occupies.  Unfortunately, this threshold
     * tends to be complex to estimate, and in any case scrolling the
     * whole band and rewriting the parts outside win's image would look
     * really ugly.  So.  What we do is consider the window "wide" if it
     * either (a) occupies the whole width of newscr, or (b) occupies
     * all but at most one column on either vertical edge of the screen
     * (this caters to fussy people who put boxes around full-screen
     * windows).  Note that changing this formula will not break any code,
     * merely change the costs of various update cases.
     */
    wide = (begx <= 1 && win->_maxx >= (newscr->_maxx - 1));
#endif
    win->_flags &= ~_HASMOVED;

    /*
     * Microtweaking alert!  This double loop is one of the genuine
     * hot spots in the code.  Even gcc doesn't seem to do enough
     * common-subexpression chunking to make it really tense,
     * so we'll force the issue.
     */

    /* limit(n) */
    limit_x = win->_maxx;
    /* limit(j) */
    if (limit_x > win->_maxx)
	limit_x = win->_maxx;

    for (i = 0, m = begy + win->_yoffset;
	i <= win->_maxy && m <= newscr->_maxy;
	i++, m++) {
	struct ldat *nline = &newscr->_line[m];
	struct ldat *oline = &win->_line[i];

	if (oline->firstchar != _NOCHANGE) {
	    int last = oline->lastchar;

	    if (last > limit_x)
		last = limit_x;

	    for (j = oline->firstchar, n = j + begx; j <= last; j++, n++) {
		if (oline->text[j] != nline->text[n]) {
		    nline->text[n] = oline->text[j];
		    CHANGED_CELL(nline, n);
		}
	    }

	}
#if USE_SCROLL_HINTS
	if (wide) {
	    int oind = oline->oldindex;

	    nline->oldindex = (oind == _NEWINDEX) ? _NEWINDEX : begy + oind
		+ win->_yoffset;
	}
#endif /* USE_SCROLL_HINTS */

	oline->firstchar = oline->lastchar = _NOCHANGE;
	if_USE_SCROLL_HINTS(oline->oldindex = i);
    }

    if (win->_clear) {
	win->_clear = FALSE;
	newscr->_clear = TRUE;
    }

    if (!win->_leaveok) {
	newscr->_cury = win->_cury + win->_begy + win->_yoffset;
	newscr->_curx = win->_curx + win->_begx;
    }
    newscr->_leaveok = win->_leaveok;

    returnCode(OK);
}

/*
**      lib_scroll.c
**
**      The routine wscrl(win, n).
**  positive n scroll the window up (ie. move lines down)
**  negative n scroll the window down (ie. move lines up)
**
*/


static void
_nc_scroll_window(WINDOW *win, int const n, NCURSES_SIZE_T const top,
    NCURSES_SIZE_T const bottom, chtype blank)
{
    int line, j;
    size_t to_copy = (size_t) (sizeof(chtype) * (win->_maxx + 1));

    TR(TRACE_MOVE, ("_nc_scroll_window(%p, %d, %d, %d)", win, n, top, bottom));

    /*
     * This used to do a line-text pointer-shuffle instead of text copies.
     * That (a) doesn't work when the window is derived and doesn't have
     * its own storage, (b) doesn't save you a lot on modern machines
     * anyway.  Your typical memcpy implementations are coded in
     * assembler using a tight BLT loop; for the size of copies we're
     * talking here, the total execution time is dominated by the one-time
     * setup cost.  So there is no point in trying to be excessively
     * clever -- esr.
     */

    /* shift n lines downwards */
    if (n < 0) {
	for (line = bottom; line >= top - n; line--) {
	    memcpy(win->_line[line].text,
		win->_line[line + n].text,
		to_copy);
	    if_USE_SCROLL_HINTS(win->_line[line].oldindex = win->_line[line
		    + n].oldindex);
	}
	for (line = top; line < top - n; line++) {
	    for (j = 0; j <= win->_maxx; j++)
		win->_line[line].text[j] = blank;
	    if_USE_SCROLL_HINTS(win->_line[line].oldindex = _NEWINDEX);
	}
    }

    /* shift n lines upwards */
    if (n > 0) {
	for (line = top; line <= bottom - n; line++) {
	    memcpy(win->_line[line].text,
		win->_line[line + n].text,
		to_copy);
	    if_USE_SCROLL_HINTS(win->_line[line].oldindex = win->_line[line
		    + n].oldindex);
	}
	for (line = bottom; line > bottom - n; line--) {
	    for (j = 0; j <= win->_maxx; j++)
		win->_line[line].text[j] = blank;
	    if_USE_SCROLL_HINTS(win->_line[line].oldindex = _NEWINDEX);
	}
    }
    touchline(win, top, bottom - top + 1);
}

int
wscrl(WINDOW *win, int n)
{
    T((T_CALLED("wscrl(%p,%d)"), win, n));

    if (!win || !win->_scroll)
	returnCode(ERR);

    if (n == 0)
	returnCode(OK);

    if ((n > (win->_regbottom - win->_regtop)) ||
	(-n > (win->_regbottom - win->_regtop)))
	returnCode(ERR);

    _nc_scroll_window(win, n, win->_regtop, win->_regbottom, _nc_background(win));

    _nc_synchook(win);
    returnCode(OK);
}

/*
**      lib_scrollok.c
**
**      The routine scrollok.
**
*/


int scrollok(WINDOW *win, bool flag)
{
	T((T_CALLED("scrollok(%p,%d)"), win, flag));

	if (win) {
	  win->_scroll = flag;
	  returnCode(OK);
	}
	else
	  returnCode(ERR);
}

/*
**      lib_set_term.c
**
**      The routine set_term().
**
*/

static SCREEN *
set_term(SCREEN * screenp)
{
    SCREEN *oldSP;

    T((T_CALLED("set_term(%p)"), screenp));

    oldSP = SP;
    _nc_set_screen(screenp);

    set_curterm(SP->_term);
    curscr = SP->_curscr;
    newscr = SP->_newscr;
    stdscr = SP->_stdscr;
    COLORS = SP->_color_count;
    COLOR_PAIRS = SP->_pair_count;
    memcpy(acs_map, SP->_acs_map, sizeof(chtype) * ACS_LEN);

    T((T_RETURN("%p"), oldSP));
    return (oldSP);
}

static ripoff_t rippedoff[5];
static ripoff_t *rsp = rippedoff;
#define N_RIPS SIZEOF(rippedoff)

static bool
no_mouse_event(SCREEN * sp GCC_UNUSED)
{
    return FALSE;
}

static bool
no_mouse_inline(SCREEN * sp GCC_UNUSED)
{
    return FALSE;
}

static bool
no_mouse_parse(int code GCC_UNUSED)
{
    return TRUE;
}

static void
no_mouse_resume(SCREEN * sp GCC_UNUSED)
{
}

static void
no_mouse_wrap(SCREEN * sp GCC_UNUSED)
{
}

static int
_nc_setupscreen(short slines, short const scolumns, FILE * output)
/* OS-independent screen initializations */
{
    int bottom_stolen = 0;
    size_t i;

    assert(SP == 0);            /* has been reset in newterm() ! */
    if (!_nc_alloc_screen())
	return ERR;

    SP->_next_screen = _nc_screen_chain;
    _nc_screen_chain = SP;

    _nc_set_buffer(output, TRUE);
    SP->_term = cur_term;
    SP->_lines = slines;
    SP->_lines_avail = slines;
    SP->_columns = scolumns;
    SP->_cursrow = -1;
    SP->_curscol = -1;
    SP->_nl = TRUE;
    SP->_raw = FALSE;
    SP->_cbreak = 0;
    SP->_echo = TRUE;
    SP->_fifohead = -1;
    SP->_endwin = TRUE;
    SP->_ofp = output;
    SP->_cursor = -1;           /* cannot know real cursor shape */

    SP->_no_padding = getenv("NCURSES_NO_PADDING") != 0;
    TR(TRACE_CHARPUT | TRACE_MOVE, ("padding will%s be used",
	    SP->_no_padding ? " not" : ""));

#if NCURSES_EXT_FUNCS
    SP->_default_color = FALSE;
    SP->_has_sgr_39_49 = FALSE;

    /*
     * Set our assumption of the terminal's default foreground and background
     * colors.  The curs_color man-page states that we can assume that the
     * background is black.  The origin of this assumption appears to be
     * terminals that displayed colored text, but no colored backgrounds, e.g.,
     * the first colored terminals around 1980.  More recent ones with better
     * technology can display not only colored backgrounds, but all
     * combinations.  So a terminal might be something other than "white" on
     * black (green/black looks monochrome too), but black on white or even
     * on ivory.
     *
     * White-on-black is the simplest thing to use for monochrome.  Almost
     * all applications that use color paint both text and background, so
     * the distinction is moot.  But a few do not - which is why we leave this
     * configurable (a better solution is to use assume_default_colors() for
     * the rare applications that do require that sort of appearance, since
     * is appears that more users expect to be able to make a white-on-black
     * or black-on-white display under control of the application than not).
     */
    SP->_default_fg = COLOR_WHITE;
    SP->_default_bg = COLOR_BLACK;

#endif /* NCURSES_EXT_FUNCS */

    SP->_maxclick = DEFAULT_MAXCLICK;
    SP->_mouse_event = no_mouse_event;
    SP->_mouse_inline = no_mouse_inline;
    SP->_mouse_parse = no_mouse_parse;
    SP->_mouse_resume = no_mouse_resume;
    SP->_mouse_wrap = no_mouse_wrap;
    SP->_mouse_fd = -1;

    /*
     * If we've no magic cookie support, we suppress attributes that xmc
     * would affect, i.e., the attributes that affect the rendition of a
     * space.  Note that this impacts the alternate character set mapping
     * as well.
     */
    if (magic_cookie_glitch > 0) {

	SP->_xmc_triggers = termattrs() & (
	    A_ALTCHARSET |
	    A_BLINK |
	    A_BOLD |
	    A_REVERSE |
	    A_STANDOUT |
	    A_UNDERLINE
	    );
	SP->_xmc_suppress = SP->_xmc_triggers & (chtype) ~ (A_BOLD);

	T(("magic cookie attributes %s", _traceattr(SP->_xmc_suppress)));
	magic_cookie_glitch = ABSENT_NUMERIC;
	acs_chars = 0;
    }
    _nc_init_acs();
    memcpy(SP->_acs_map, acs_map, sizeof(chtype) * ACS_LEN);

    _nc_idcok = TRUE;
    _nc_idlok = FALSE;

    _nc_windows = 0;            /* no windows yet */

    SP->oldhash = 0;
    SP->newhash = 0;

    T(("creating newscr"));
    if ((newscr = newwin(slines, scolumns, 0, 0)) == 0)
	return ERR;

    T(("creating curscr"));
    if ((curscr = newwin(slines, scolumns, 0, 0)) == 0)
	return ERR;

    SP->_newscr = newscr;
    SP->_curscr = curscr;
#if USE_SIZECHANGE
    SP->_resize = resizeterm;
#endif

    newscr->_clear = TRUE;
    curscr->_clear = FALSE;

    for (i = 0, rsp = rippedoff; rsp->line && (i < N_RIPS); rsp++, i++) {
	if (rsp->hook) {
	    WINDOW *w;
	    int count = (rsp->line < 0) ? -rsp->line : rsp->line;

	    if (rsp->line < 0) {
		w = newwin(count, scolumns, SP->_lines_avail - count, 0);
		if (w) {
		    rsp->w = w;
		    rsp->hook(w, scolumns);
		    bottom_stolen += count;
		} else
		    return ERR;
	    } else {
		w = newwin(count, scolumns, 0, 0);
		if (w) {
		    rsp->w = w;
		    rsp->hook(w, scolumns);
		    SP->_topstolen += count;
		} else
		    return ERR;
	    }
	    SP->_lines_avail -= count;
	}
	rsp->line = 0;
    }
    /* reset the stack */
    rsp = rippedoff;

    T(("creating stdscr"));
    assert((SP->_lines_avail + SP->_topstolen + bottom_stolen) == slines);
    if ((stdscr = newwin(LINES = SP->_lines_avail, scolumns, 0, 0)) == 0)
	return ERR;
    SP->_stdscr = stdscr;

    def_shell_mode();
    def_prog_mode();

    return OK;
}

/* The internal implementation interprets line as the number of
   lines to rip off from the top or bottom.
   */
static int
_nc_ripoffline(int line, int (*init) (WINDOW *, int))
{
    if (line == 0)
	return (OK);

    if (rsp >= rippedoff + N_RIPS)
	return (ERR);

    rsp->line = line;
    rsp->hook = init;
    rsp->w = 0;
    rsp++;

    return (OK);
}

/*
 * Terminal setup routines common to termcap and terminfo:
 *
 *              use_env(bool)
 *              setupterm(char *, int, int *)
 */

/****************************************************************************
 *
 * Terminal size computation
 *
 ****************************************************************************/

/*
 * SCO defines TIOCGSIZE and the corresponding struct.  Other systems (SunOS,
 * Solaris, IRIX) define TIOCGWINSZ and struct winsize.
 */
#ifdef TIOCGSIZE
# define IOCTL_WINSIZE TIOCGSIZE
# define STRUCT_WINSIZE struct ttysize
# define WINSIZE_ROWS(n) (int)n.ts_lines
# define WINSIZE_COLS(n) (int)n.ts_cols
#else
# ifdef TIOCGWINSZ
#  define IOCTL_WINSIZE TIOCGWINSZ
#  define STRUCT_WINSIZE struct winsize
#  define WINSIZE_ROWS(n) (int)n.ws_row
#  define WINSIZE_COLS(n) (int)n.ws_col
# endif
#endif

static int _use_env = TRUE;

static void do_prototype(void);


static void
_nc_get_screensize(int *linep, int *colp)
/* Obtain lines/columns values from the environment and/or terminfo entry */
{
    /* figure out the size of the screen */
    T(("screen size: terminfo lines = %d columns = %d", lines, columns));

    if (!_use_env) {
	*linep = (int) lines;
	*colp = (int) columns;
    } else {                    /* usually want to query LINES and COLUMNS from environment */
	int value;

	*linep = *colp = 0;

	/* first, look for environment variables */
	if ((value = _nc_getenv_num("LINES")) > 0) {
	    *linep = value;
	}
	if ((value = _nc_getenv_num("COLUMNS")) > 0) {
	    *colp = value;
	}
	T(("screen size: environment LINES = %d COLUMNS = %d", *linep, *colp));

#ifdef __EMX__
	if (*linep <= 0 || *colp <= 0) {
	    int screendata[2];
	    _scrsize(screendata);
	    *colp = screendata[0];
	    *linep = screendata[1];
	    T(("EMX screen size: environment LINES = %d COLUMNS = %d",
		    *linep, *colp));
	}
#endif
#if HAVE_SIZECHANGE
	/* if that didn't work, maybe we can try asking the OS */
	if (*linep <= 0 || *colp <= 0) {
	    if (isatty(cur_term->Filedes)) {
		STRUCT_WINSIZE size;

		errno = 0;
		do {
		    if (ioctl(cur_term->Filedes, IOCTL_WINSIZE, &size) < 0
			&& errno != EINTR)
			goto failure;
		} while
		    (errno == EINTR);

		/*
		 * Solaris lets users override either dimension with an
		 * environment variable.
		 */
		if (*linep <= 0)
		    *linep = WINSIZE_ROWS(size);
		if (*colp <= 0)
		    *colp = WINSIZE_COLS(size);
	    }
	    /* FALLTHRU */
	  failure:;
	}
#endif /* HAVE_SIZECHANGE */

	/* if we can't get dynamic info about the size, use static */
	if (*linep <= 0 || *colp <= 0)
	    if (lines > 0 && columns > 0) {
		*linep = (int) lines;
		*colp = (int) columns;
	    }

	/* the ultimate fallback, assume fixed 24x80 size */
	if (*linep <= 0 || *colp <= 0) {
	    *linep = 24;
	    *colp = 80;
	}

	/*
	 * Put the derived values back in the screen-size caps, so
	 * tigetnum() and tgetnum() will do the right thing.
	 */
	lines = (short) (*linep);
	columns = (short) (*colp);
    }

    T(("screen size is %dx%d", *linep, *colp));

    if (VALID_NUMERIC(init_tabs))
	TABSIZE = (int) init_tabs;
    else
	TABSIZE = 8;
    T(("TABSIZE = %d", TABSIZE));

}

#if USE_SIZECHANGE
static void
_nc_update_screensize(void)
{
    int my_lines, my_cols;

    _nc_get_screensize(&my_lines, &my_cols);
    if (SP != 0 && SP->_resize != 0)
	SP->_resize(my_lines, my_cols);
}
#endif

/****************************************************************************
 *
 * Terminal setup
 *
 ****************************************************************************/

#define ret_error(code, fmt, arg)       if (errret) {\
					    *errret = code;\
					    returnCode(ERR);\
					} else {\
					    fprintf(stderr, fmt, arg);\
					    exit(EXIT_FAILURE);\
					}

#define ret_error0(code, msg)           if (errret) {\
					    *errret = code;\
					    returnCode(ERR);\
					} else {\
					    fprintf(stderr, msg);\
					    exit(EXIT_FAILURE);\
					}

static int
grab_entry(const char *const tn, TERMTYPE * const tp)
/* return 1 if entry found, 0 if not found, -1 if database not accessible */
{
    char filename[PATH_MAX];
    int status;

    /*
     * $TERM shouldn't contain pathname delimiters.
     */
    if (strchr(tn, '/'))
	return 0;

    if ((status = _nc_read_entry(tn, filename, tp)) != 1) {
	;
    }

    /*
     * If we have an entry, force all of the cancelled strings to null
     * pointers so we don't have to test them in the rest of the library.
     * (The terminfo compiler bypasses this logic, since it must know if
     * a string is cancelled, for merging entries).
     */
    if (status == 1) {
	int n;
	for_each_boolean(n, tp)
	    if (!VALID_BOOLEAN(tp->Booleans[n]))
	    tp->Booleans[n] = FALSE;
	for_each_string(n, tp)
	    if (tp->Strings[n] == CANCELLED_STRING)
	    tp->Strings[n] = ABSENT_STRING;
    }
    return (status);
}

/*
 *      setupterm(termname, Filedes, errret)
 *
 *      Find and read the appropriate object file for the terminal
 *      Make cur_term point to the structure.
 *
 */

static int
setupterm(const char *tname, int Filedes, int *errret)
{
    struct term *term_ptr;
    int status;

    T((T_CALLED("setupterm(%s,%d,%p)"), _nc_visbuf(tname), Filedes, errret));

    if (tname == 0) {
	tname = getenv("TERM");
	if (tname == 0 || *tname == '\0') {
	    ret_error0(-1, "TERM environment variable not set.\n");
	}
    }
    if (strlen(tname) > MAX_NAME_SIZE) {
	ret_error(-1, "TERM environment must be <= %d characters.\n",
	    MAX_NAME_SIZE);
    }

    T(("your terminal name is %s", tname));

    term_ptr = typeCalloc(TERMINAL, 1);

    if (term_ptr == 0) {
	ret_error0(-1, "Not enough memory to create terminal structure.\n");
    }
    status = grab_entry(tname, &term_ptr->type);

    /* try fallback list if entry on disk */
    if (status != 1) {
	const TERMTYPE *fallback = _nc_fallback(tname);

	if (fallback) {
	    term_ptr->type = *fallback;
	    status = 1;
	}
    }

    if (status == -1) {
	ret_error0(-1, "terminals database is inaccessible\n");
    } else if (status == 0) {
	ret_error(0, "'%s': unknown terminal type.\n", tname);
    }

    /*
     * Improve on SVr4 curses.  If an application mixes curses and termcap
     * calls, it may call both initscr and tgetent.  This is not really a
     * good thing to do, but can happen if someone tries using ncurses with
     * the readline library.  The problem we are fixing is that when
     * tgetent calls setupterm, the resulting Ottyb struct in cur_term is
     * zeroed.  A subsequent call to endwin uses the zeroed terminal
     * settings rather than the ones saved in initscr.  So we check if
     * cur_term appears to contain terminal settings for the same output
     * file as our current call - and copy those terminal settings.  (SVr4
     * curses does not do this, however applications that are working
     * around the problem will still work properly with this feature).
     */
    if (cur_term != 0) {
	if (cur_term->Filedes == Filedes)
	    term_ptr->Ottyb = cur_term->Ottyb;
    }

    set_curterm(term_ptr);

    if (command_character && getenv("CC"))
	do_prototype();

    /*
     * Allow output redirection.  This is what SVr3 does.
     * If stdout is directed to a file, screen updates go
     * to standard error.
     */
    if (Filedes == STDOUT_FILENO && !isatty(Filedes))
	Filedes = STDERR_FILENO;
    cur_term->Filedes = Filedes;

    _nc_get_screensize(&LINES, &COLS);

    if (errret)
	*errret = 1;

    T((T_CREATE("screen %s %dx%d"), tname, LINES, COLS));

    if (generic_type) {
	ret_error(0, "'%s': I need something more specific.\n", tname);
    }
    if (hard_copy) {
	ret_error(1, "'%s': I can't handle hardcopy terminals.\n", tname);
    }
    returnCode(OK);
}

/*
**      do_prototype()
**
**      Take the real command character out of the CC environment variable
**      and substitute it in for the prototype given in 'command_character'.
**
*/

static void
do_prototype(void)
{
    int i;
    char CC;
    char proto;
    char *tmp;

    tmp = getenv("CC");
    CC = *tmp;
    proto = *command_character;

    for_each_string(i, &(cur_term->type)) {
	for (tmp = cur_term->type.Strings[i]; *tmp; tmp++) {
	    if (*tmp == proto)
		*tmp = CC;
	}
    }
}

/*
 *      lib_slk.c
 *      Soft key routines.
 */

/*
 * We'd like to move these into the screen context structure, but cannot,
 * because slk_init() is called before initscr()/newterm().
 */

/*
 * Paint the info line for the PC style SLK emulation.
 *
 */
static void
slk_paint_info(WINDOW *win)
{
  if (win && SP->slk_format==4)
    {
      int i;

      mvwhline (win,0,0,0,getmaxx(win));
      wmove (win,0,0);

      for (i = 0; i < SP->_slk->maxlab; i++) {
	if (win && SP->slk_format==4)
	  {
	    mvwaddch(win,0,SP->_slk->ent[i].x,'F');
	    if (i<9)
	      waddch(win,'1'+i);
	    else
	      {
		waddch(win,'1');
		waddch(win,'0' + (i-9));
	      }
	  }
      }
    }
}

/*
 * Initialize soft labels.
 * Called from newterm()
 */
static int
_nc_slk_initialize(WINDOW *stwin, int cols)
{
int i, x;
int res = OK;
char *p;

	T(("slk_initialize()"));

	if (SP->_slk)
	  { /* we did this already, so simply return */
	    return(OK);
	  }
	else
	  if ((SP->_slk = typeCalloc(SLK, 1)) == 0)
	    return(ERR);

	SP->_slk->ent    = NULL;
	SP->_slk->buffer = NULL;
	SP->_slk->attr   = A_STANDOUT;

	SP->_slk->maxlab = (num_labels > 0) ?
	  num_labels : MAX_SKEY(_nc_slk_format);
	SP->_slk->maxlen = (num_labels > 0) ?
	  label_width * label_height : MAX_SKEY_LEN(_nc_slk_format);
	SP->_slk->labcnt = (SP->_slk->maxlab < MAX_SKEY(_nc_slk_format)) ?
	  MAX_SKEY(_nc_slk_format) : SP->_slk->maxlab;

	SP->_slk->ent = typeCalloc(slk_ent, SP->_slk->labcnt);
	if (SP->_slk->ent == NULL)
	  goto exception;

	p = SP->_slk->buffer = (char*) calloc(2*SP->_slk->labcnt,(1+SP->_slk->maxlen));
	if (SP->_slk->buffer == NULL)
	  goto exception;

	for (i = 0; i < SP->_slk->labcnt; i++) {
		SP->_slk->ent[i].text = p;
		p += (1 + SP->_slk->maxlen);
		SP->_slk->ent[i].form_text = p;
		p += (1 + SP->_slk->maxlen);
		memset(SP->_slk->ent[i].form_text, ' ', (unsigned)(SP->_slk->maxlen));
		SP->_slk->ent[i].visible = (i < SP->_slk->maxlab);
	}
	if (_nc_slk_format >= 3) /* PC style */
	  {
	    int gap = (cols - 3 * (3 + 4*SP->_slk->maxlen))/2;

	    if (gap < 1)
	      gap = 1;

	    for (i = x = 0; i < SP->_slk->maxlab; i++) {
	      SP->_slk->ent[i].x = x;
	      x += SP->_slk->maxlen;
	      x += (i==3 || i==7) ? gap : 1;
	    }
	    if (_nc_slk_format == 4)
	      slk_paint_info (stwin);
	  }
	else {
	  if (_nc_slk_format == 2) {    /* 4-4 */
	    int gap = cols - (SP->_slk->maxlab * SP->_slk->maxlen) - 6;

	    if (gap < 1)
			gap = 1;
	    for (i = x = 0; i < SP->_slk->maxlab; i++) {
	      SP->_slk->ent[i].x = x;
	      x += SP->_slk->maxlen;
	      x += (i == 3) ? gap : 1;
	    }
	  }
	  else
	    {
	      if (_nc_slk_format == 1) { /* 1 -> 3-2-3 */
		int gap = (cols - (SP->_slk->maxlab * SP->_slk->maxlen) - 5) / 2;

		if (gap < 1)
		  gap = 1;
		for (i = x = 0; i < SP->_slk->maxlab; i++) {
		  SP->_slk->ent[i].x = x;
		  x += SP->_slk->maxlen;
		  x += (i == 2 || i == 4) ? gap : 1;
		}
	      }
	      else
		goto exception;
	    }
	}
	SP->_slk->dirty = TRUE;
	if ((SP->_slk->win = stwin) == NULL)
	{
	exception:
		if (SP->_slk)
		{
		   FreeIfNeeded(SP->_slk->buffer);
		   FreeIfNeeded(SP->_slk->ent);
		   free(SP->_slk);
		   SP->_slk = (SLK*)0;
		   res = (ERR);
		}
	}

	/* We now reset the format so that the next newterm has again
	 * per default no SLK keys and may call slk_init again to
	 * define a new layout. (juergen 03-Mar-1999)
	 */
	SP->slk_format = _nc_slk_format;
	_nc_slk_format = 0;
	return(res);
}


/*
**      lib_touch.c
**
**         The routines wtouchln(),
**                      is_wintouched().
**
*/


static bool is_wintouched(WINDOW *win)
{
int i;

	T((T_CALLED("is_wintouched(%p)"), win));

	if (win)
		for (i = 0; i <= win->_maxy; i++)
			if (win->_line[i].firstchar != _NOCHANGE)
				returnCode(TRUE);
	returnCode(FALSE);
}

int wtouchln(WINDOW *win, int y, int n, int changed)
{
int i;

	T((T_CALLED("wtouchln(%p,%d,%d,%d)"), win, y, n, changed));

	if (!win || (n<0) || (y<0) || (y>win->_maxy))
	  returnCode(ERR);

	for (i = y; i < y+n; i++) {
		if (i>win->_maxy) break;
		win->_line[i].firstchar = changed ? 0 : _NOCHANGE;
		win->_line[i].lastchar = changed ? win->_maxx : _NOCHANGE;
	}
	returnCode(OK);
}

/*
 *      tparm.c
 *
 */


/*
 *      char *
 *      tparm(string, ...)
 *
 *      Substitute the given parameters into the given string by the following
 *      rules (taken from terminfo(5)):
 *
 *           Cursor addressing and other strings  requiring  parame-
 *      ters in the terminal are described by a parameterized string
 *      capability, with like escapes %x in  it.   For  example,  to
 *      address  the  cursor, the cup capability is given, using two
 *      parameters: the row and column to  address  to.   (Rows  and
 *      columns  are  numbered  from  zero and refer to the physical
 *      screen visible to the user, not to any  unseen  memory.)  If
 *      the terminal has memory relative cursor addressing, that can
 *      be indicated by
 *
 *           The parameter mechanism uses  a  stack  and  special  %
 *      codes  to manipulate it.  Typically a sequence will push one
 *      of the parameters onto the stack and then print it  in  some
 *      format.  Often more complex operations are necessary.
 *
 *           The % encodings have the following meanings:
 *
 *           %%        outputs `%'
 *           %c        print pop() like %c in printf()
 *           %s        print pop() like %s in printf()
 *           %[[:]flags][width[.precision]][doxXs]
 *                     as in printf, flags are [-+#] and space
 *                     The ':' is used to avoid making %+ or %-
 *                     patterns (see below).
 *
 *           %p[1-9]   push ith parm
 *           %P[a-z]   set dynamic variable [a-z] to pop()
 *           %g[a-z]   get dynamic variable [a-z] and push it
 *           %P[A-Z]   set static variable [A-Z] to pop()
 *           %g[A-Z]   get static variable [A-Z] and push it
 *           %l        push strlen(pop)
 *           %'c'      push char constant c
 *           %{nn}     push integer constant nn
 *
 *           %+ %- %* %/ %m
 *                     arithmetic (%m is mod): push(pop() op pop())
 *           %& %| %^  bit operations: push(pop() op pop())
 *           %= %> %<  logical operations: push(pop() op pop())
 *           %A %O     logical and & or operations for conditionals
 *           %! %~     unary operations push(op pop())
 *           %i        add 1 to first two parms (for ANSI terminals)
 *
 *           %? expr %t thenpart %e elsepart %;
 *                     if-then-else, %e elsepart is optional.
 *                     else-if's are possible ala Algol 68:
 *                     %? c1 %t b1 %e c2 %t b2 %e c3 %t b3 %e c4 %t b4 %e b5 %;
 *
 *      For those of the above operators which are binary and not commutative,
 *      the stack works in the usual way, with
 *                      %gx %gy %m
 *      resulting in x mod y, not the reverse.
 */

#define STACKSIZE       20

typedef struct {
    union {
	unsigned int num;
	char *str;
    } data;
    bool num_type;
} stack_frame;

static stack_frame stack[STACKSIZE];
static int stack_ptr;

static char *out_buff;
static size_t out_size;
static size_t out_used;

static void
really_get_space(size_t need)
{
    out_size = need * 2;
    out_buff = typeRealloc(char, out_size, out_buff);
    if (out_buff == 0)
	_nc_err_abort("Out of memory");
}

static inline void
get_space(size_t need)
{
    need += out_used;
    if (need > out_size)
	really_get_space(need);
}

static inline void
save_text(const char *fmt, const char *s, int len)
{
    size_t s_len = strlen(s);
    if (len > (int) s_len)
	s_len = len;

    get_space(s_len + 1);

    (void) sprintf(out_buff + out_used, fmt, s);
    out_used += strlen(out_buff + out_used);
}

static inline void
save_number(const char *fmt, int number, int len)
{
    if (len < 30)
	len = 30;               /* actually log10(MAX_INT)+1 */

    get_space(len + 1);

    (void) sprintf(out_buff + out_used, fmt, number);
    out_used += strlen(out_buff + out_used);
}

static inline void
tp_save_char(int c)
{
    if (c == 0)
	c = 0200;
    get_space(1);
    out_buff[out_used++] = c;
}

static inline void
npush(int x)
{
    if (stack_ptr < STACKSIZE) {
	stack[stack_ptr].num_type = TRUE;
	stack[stack_ptr].data.num = x;
	stack_ptr++;
    }
}

static inline int
npop(void)
{
    int result = 0;
    if (stack_ptr > 0) {
	stack_ptr--;
	if (stack[stack_ptr].num_type)
	    result = stack[stack_ptr].data.num;
    }
    return result;
}

static inline void
spush(char *x)
{
    if (stack_ptr < STACKSIZE) {
	stack[stack_ptr].num_type = FALSE;
	stack[stack_ptr].data.str = x;
	stack_ptr++;
    }
}

static inline char *
spop(void)
{
    static char dummy[] = "";   /* avoid const-cast */
    char *result = dummy;
    if (stack_ptr > 0) {
	stack_ptr--;
	if (!stack[stack_ptr].num_type && stack[stack_ptr].data.str != 0)
	    result = stack[stack_ptr].data.str;
    }
    return result;
}

static inline const char *
parse_format(const char *s, char *format, int *len)
{
    bool done = FALSE;
    bool allowminus = FALSE;
    bool dot = FALSE;
    bool err = FALSE;
    char *fmt = format;
    int prec = 0;
    int width = 0;
    int value = 0;

    *len = 0;
    *format++ = '%';
    while (*s != '\0' && !done) {
	switch (*s) {
	case 'c':               /* FALLTHRU */
	case 'd':               /* FALLTHRU */
	case 'o':               /* FALLTHRU */
	case 'x':               /* FALLTHRU */
	case 'X':               /* FALLTHRU */
	case 's':
	    *format++ = *s;
	    done = TRUE;
	    break;
	case '.':
	    *format++ = *s++;
	    if (dot) {
		err = TRUE;
	    } else {
		dot = TRUE;
		prec = value;
	    }
	    value = 0;
	    break;
	case '#':
	    *format++ = *s++;
	    break;
	case ' ':
	    *format++ = *s++;
	    break;
	case ':':
	    s++;
	    allowminus = TRUE;
	    break;
	case '-':
	    if (allowminus) {
		*format++ = *s++;
	    } else {
		done = TRUE;
	    }
	    break;
	default:
	    if (isdigit(*s)) {
		value = (value * 10) + (*s - '0');
		if (value > 10000)
		    err = TRUE;
		*format++ = *s++;
	    } else {
		done = TRUE;
	    }
	}
    }

    /*
     * If we found an error, ignore (and remove) the flags.
     */
    if (err) {
	prec = width = value = 0;
	format = fmt;
	*format++ = '%';
	*format++ = *s;
    }

    if (dot)
	width = value;
    else
	prec = value;

    *format = '\0';
    /* return maximum string length in print */
    *len = (prec > width) ? prec : width;
    return s;
}

#define isUPPER(c) ((c) >= 'A' && (c) <= 'Z')
#define isLOWER(c) ((c) >= 'a' && (c) <= 'z')

static inline char *
tparam_internal(const char *string, va_list ap)
{
#define NUM_VARS 26
    char *p_is_s[9];
    int param[9];
    int lastpop;
    int popcount;
    int number;
    int len;
    int level;
    int x, y;
    int i;
    size_t len2;
    const char *cp;
    static size_t len_fmt;
    static char dummy[] = "";
    static char *format;
    static int dynamic_var[NUM_VARS];
    static int static_vars[NUM_VARS];

    out_used = 0;
    if (string == NULL)
	return NULL;

    if ((len2 = strlen(string)) > len_fmt) {
	len_fmt = len2 + len_fmt + 2;
	if ((format = typeRealloc(char, len_fmt, format)) == 0)
	      return 0;
    }

    /*
     * Find the highest parameter-number referred to in the format string.
     * Use this value to limit the number of arguments copied from the
     * variable-length argument list.
     */

    number = 0;
    lastpop = -1;
    popcount = 0;
    memset(p_is_s, 0, sizeof(p_is_s));

    /*
     * Analyze the string to see how many parameters we need from the varargs
     * list, and what their types are.  We will only accept string parameters
     * if they appear as a %l or %s format following an explicit parameter
     * reference (e.g., %p2%s).  All other parameters are numbers.
     *
     * 'number' counts coarsely the number of pop's we see in the string, and
     * 'popcount' shows the highest parameter number in the string.  We would
     * like to simply use the latter count, but if we are reading termcap
     * strings, there may be cases that we cannot see the explicit parameter
     * numbers.
     */
    for (cp = string; (cp - string) < (int) len2;) {
	if (*cp == '%') {
	    cp++;
	    cp = parse_format(cp, format, &len);
	    switch (*cp) {
	    default:
		break;

	    case 'd':           /* FALLTHRU */
	    case 'o':           /* FALLTHRU */
	    case 'x':           /* FALLTHRU */
	    case 'X':           /* FALLTHRU */
	    case 'c':           /* FALLTHRU */
		number++;
		lastpop = -1;
		break;

	    case 'l':
	    case 's':
		if (lastpop > 0)
		    p_is_s[lastpop - 1] = dummy;
		++number;
		break;

	    case 'p':
		cp++;
		i = (*cp - '0');
		if (i >= 0 && i <= 9) {
		    lastpop = i;
		    if (lastpop > popcount)
			popcount = lastpop;
		}
		break;

	    case 'P':
	    case 'g':
		cp++;
		break;

	    case S_QUOTE:
		cp += 2;
		lastpop = -1;
		break;

	    case L_BRACE:
		cp++;
		while (*cp >= '0' && *cp <= '9') {
		    cp++;
		}
		break;

	    case '+':
	    case '-':
	    case '*':
	    case '/':
	    case 'm':
	    case 'A':
	    case 'O':
	    case '&':
	    case '|':
	    case '^':
	    case '=':
	    case '<':
	    case '>':
	    case '!':
	    case '~':
		lastpop = -1;
		number += 2;
		break;

	    case 'i':
		lastpop = -1;
		if (popcount < 2)
		    popcount = 2;
		break;
	    }
	}
	if (*cp != '\0')
	    cp++;
    }

    if (number > 9)
	number = 9;
    for (i = 0; i < max(popcount, number); i++) {
	/*
	 * A few caps (such as plab_norm) have string-valued parms.
	 * We'll have to assume that the caller knows the difference, since
	 * a char* and an int may not be the same size on the stack.
	 */
	if (p_is_s[i] != 0) {
	    p_is_s[i] = va_arg(ap, char *);
	} else {
	    param[i] = va_arg(ap, int);
	}
    }

    /*
     * This is a termcap compatibility hack.  If there are no explicit pop
     * operations in the string, load the stack in such a way that
     * successive pops will grab successive parameters.  That will make
     * the expansion of (for example) \E[%d;%dH work correctly in termcap
     * style, which means tparam() will expand termcap strings OK.
     */
    stack_ptr = 0;
    if (popcount == 0) {
	popcount = number;
	for (i = number - 1; i >= 0; i--)
	    npush(param[i]);
    }

    while (*string) {
	if (*string != '%') {
	    tp_save_char(*string);
	} else {
	    string++;
	    string = parse_format(string, format, &len);
	    switch (*string) {
	    default:
		break;
	    case '%':
		tp_save_char('%');
		break;

	    case 'd':           /* FALLTHRU */
	    case 'o':           /* FALLTHRU */
	    case 'x':           /* FALLTHRU */
	    case 'X':           /* FALLTHRU */
	    case 'c':           /* FALLTHRU */
		save_number(format, npop(), len);
		break;

	    case 'l':
		save_number("%d", strlen(spop()), 0);
		break;

	    case 's':
		save_text(format, spop(), len);
		break;

	    case 'p':
		string++;
		i = (*string - '1');
		if (i >= 0 && i < 9) {
		    if (p_is_s[i])
			spush(p_is_s[i]);
		    else
			npush(param[i]);
		}
		break;

	    case 'P':
		string++;
		if (isUPPER(*string)) {
		    i = (*string - 'A');
		    static_vars[i] = npop();
		} else if (isLOWER(*string)) {
		    i = (*string - 'a');
		    dynamic_var[i] = npop();
		}
		break;

	    case 'g':
		string++;
		if (isUPPER(*string)) {
		    i = (*string - 'A');
		    npush(static_vars[i]);
		} else if (isLOWER(*string)) {
		    i = (*string - 'a');
		    npush(dynamic_var[i]);
		}
		break;

	    case S_QUOTE:
		string++;
		npush(*string);
		string++;
		break;

	    case L_BRACE:
		number = 0;
		string++;
		while (*string >= '0' && *string <= '9') {
		    number = number * 10 + *string - '0';
		    string++;
		}
		npush(number);
		break;

	    case '+':
		npush(npop() + npop());
		break;

	    case '-':
		y = npop();
		x = npop();
		npush(x - y);
		break;

	    case '*':
		npush(npop() * npop());
		break;

	    case '/':
		y = npop();
		x = npop();
		npush(y ? (x / y) : 0);
		break;

	    case 'm':
		y = npop();
		x = npop();
		npush(y ? (x % y) : 0);
		break;

	    case 'A':
		npush(npop() && npop());
		break;

	    case 'O':
		npush(npop() || npop());
		break;

	    case '&':
		npush(npop() & npop());
		break;

	    case '|':
		npush(npop() | npop());
		break;

	    case '^':
		npush(npop() ^ npop());
		break;

	    case '=':
		y = npop();
		x = npop();
		npush(x == y);
		break;

	    case '<':
		y = npop();
		x = npop();
		npush(x < y);
		break;

	    case '>':
		y = npop();
		x = npop();
		npush(x > y);
		break;

	    case '!':
		npush(!npop());
		break;

	    case '~':
		npush(~npop());
		break;

	    case 'i':
		if (p_is_s[0] == 0)
		    param[0]++;
		if (p_is_s[1] == 0)
		    param[1]++;
		break;

	    case '?':
		break;

	    case 't':
		x = npop();
		if (!x) {
		    /* scan forward for %e or %; at level zero */
		    string++;
		    level = 0;
		    while (*string) {
			if (*string == '%') {
			    string++;
			    if (*string == '?')
				level++;
			    else if (*string == ';') {
				if (level > 0)
				    level--;
				else
				    break;
			    } else if (*string == 'e' && level == 0)
				break;
			}

			if (*string)
			    string++;
		    }
		}
		break;

	    case 'e':
		/* scan forward for a %; at level zero */
		string++;
		level = 0;
		while (*string) {
		    if (*string == '%') {
			string++;
			if (*string == '?')
			    level++;
			else if (*string == ';') {
			    if (level > 0)
				level--;
			    else
				break;
			}
		    }

		    if (*string)
			string++;
		}
		break;

	    case ';':
		break;

	    }                   /* endswitch (*string) */
	}                       /* endelse (*string == '%') */

	if (*string == '\0')
	    break;

	string++;
    }                           /* endwhile (*string) */

    get_space(1);
    out_buff[out_used] = '\0';

    T((T_RETURN("%s"), _nc_visbuf(out_buff)));
    return (out_buff);
}

static char *
tparm(const char *string,...)
{
    va_list ap;
    char *result;

    va_start(ap, string);
    result = tparam_internal(string, ap);
    va_end(ap);
    return result;
}

/*
 *      tputs.c
 *              delay_output()
 *              _nc_outch()
 *              tputs()
 *
 */

static char PC = 0;                    /* used by termcap library */
static short ospeed = 0;               /* used by termcap library */

static int _nc_nulls_sent = 0;         /* used by 'tack' program */

static int (*my_outch) (int c) = _nc_outch;

static int
delay_output(int ms)
{
    T((T_CALLED("delay_output(%d)"), ms));

    if (no_pad_char) {
	_nc_flush();
	napms(ms);
    } else {
	int nullcount;

	nullcount = (ms * _nc_baudrate(ospeed)) / 10000;
	for (_nc_nulls_sent += nullcount; nullcount > 0; nullcount--)
	    my_outch(PC);
	if (my_outch == _nc_outch)
	    _nc_flush();
    }

    returnCode(OK);
}

static void
_nc_flush(void)
{
    (void) fflush(NC_OUTPUT);
}

static int
_nc_outch(int ch)
{
    if (SP != 0
	&& SP->_cleanup) {
	char tmp = ch;
	/*
	 * POSIX says write() is safe in a signal handler, but the
	 * buffered I/O is not.
	 */
	write(fileno(NC_OUTPUT), &tmp, 1);
    } else {
	putc(ch, NC_OUTPUT);
    }
    return OK;
}

static int
putp(const char *string)
{
    return tputs(string, 1, _nc_outch);
}

static int
tputs(const char *string, int affcnt, int (*outc) (int))
{
    bool always_delay;
    bool normal_delay;
    int number;

    if (!VALID_STRING(string))
	return ERR;

    if (cur_term == 0) {
	always_delay = FALSE;
	normal_delay = TRUE;
    } else {
	always_delay = (string == bell) || (string == flash_screen);
	normal_delay =
	    !xon_xoff
	    && padding_baud_rate
	    && (SP == 0 || !(SP->_no_padding))
	    && (_nc_baudrate(ospeed) >= padding_baud_rate);
    }

    my_outch = outc;            /* redirect delay_output() */
    while (*string) {
	if (*string != '$')
	    (*outc) (*string);
	else {
	    string++;
	    if (*string != '<') {
		(*outc) ('$');
		if (*string)
		    (*outc) (*string);
	    } else {
		bool mandatory;

		string++;
		if ((!isdigit(*string) && *string != '.') || !strchr(string, '>')) {
		    (*outc) ('$');
		    (*outc) ('<');
		    continue;
		}

		number = 0;
		while (isdigit(*string)) {
		    number = number * 10 + (*string - '0');
		    string++;
		}
		number *= 10;
		if (*string == '.') {
		    string++;
		    if (isdigit(*string)) {
			number += (*string - '0');
			string++;
		    }
		    while (isdigit(*string))
			string++;
		}

		mandatory = FALSE;
		while (*string == '*' || *string == '/') {
		    if (*string == '*') {
			number *= affcnt;
			string++;
		    } else {    /* if (*string == '/') */
			mandatory = TRUE;
			string++;
		    }
		}

		if (number > 0
		    && (always_delay
			|| normal_delay
			|| mandatory))
		    delay_output(number / 10);

	    }                   /* endelse (*string == '<') */
	}                       /* endelse (*string == '$') */

	if (*string == '\0')
	    break;

	string++;
    }

    my_outch = _nc_outch;
    return OK;
}

/*
**      lib_tstp.c
**
**      The routine _nc_signal_handler().
**
*/


#if defined(SIGTSTP) && (HAVE_SIGACTION || HAVE_SIGVEC)
#define USE_SIGTSTP 1
#else
#define USE_SIGTSTP 0
#endif

/*
 * Note: This code is fragile!  Its problem is that different OSs
 * handle restart of system calls interrupted by signals differently.
 * The ncurses code needs signal-call restart to happen -- otherwise,
 * interrupted wgetch() calls will return FAIL, probably making the
 * application think the input stream has ended and it should
 * terminate.  In particular, you know you have this problem if, when
 * you suspend an ncurses-using lynx with ^Z and resume, it dies
 * immediately.
 *
 * Default behavior of POSIX sigaction(2) is not to restart
 * interrupted system calls, but Linux's sigaction does it anyway (at
 * least, on and after the 1.1.47 I (esr) use).  Thus this code works
 * OK under Linux.  The 4.4BSD sigaction(2) supports a (non-portable)
 * SA_RESTART flag that forces the right behavior.  Thus, this code
 * should work OK under BSD/OS, NetBSD, and FreeBSD (let us know if it
 * does not).
 *
 * Stock System Vs (and anything else using a strict-POSIX
 * sigaction(2) without SA_RESTART) may have a problem.  Possible
 * solutions:
 *
 *    sigvec      restarts by default (SV_INTERRUPT flag to not restart)
 *    signal      restarts by default in SVr4 (assuming you link with -lucb)
 *                and BSD, but not SVr3.
 *    sigset      restarts, but is only available under SVr4/Solaris.
 *
 * The signal(3) call is mandated by the ANSI standard, and its
 * interaction with sigaction(2) is described in the POSIX standard
 * (3.3.4.2, page 72,line 934).  According to section 8.1, page 191,
 * however, signal(3) itself is not required by POSIX.1.  And POSIX is
 * silent on whether it is required to restart signals.
 *
 * So.  The present situation is, we use sigaction(2) with no
 * guarantee of restart anywhere but on Linux and BSD.  We could
 * switch to signal(3) and collar Linux, BSD, and SVr4.  Any way
 * we slice it, System V UNIXes older than SVr4 will probably lose
 * (this may include XENIX).
 *
 * This implementation will probably be changed to use signal(3) in
 * the future.  If nothing else, it's simpler...
 */

#if USE_SIGTSTP
static void
tstp(int dummy GCC_UNUSED)
{
    sigset_t mask, omask;
    sigaction_t act, oact;

#ifdef SIGTTOU
    int sigttou_blocked;
#endif

    T(("tstp() called"));

    /*
     * The user may have changed the prog_mode tty bits, so save them.
     *
     * But first try to detect whether we still are in the foreground
     * process group - if not, an interactive shell may already have
     * taken ownership of the tty and modified the settings when our
     * parent was stopped before us, and we would likely pick up the
     * settings already modified by the shell.
     */
    if (SP != 0 && !SP->_endwin)        /* don't do this if we're not in curses */
	if (tcgetpgrp(STDIN_FILENO) == getpgrp())
	    def_prog_mode();

    /*
     * Block window change and timer signals.  The latter
     * is because applications use timers to decide when
     * to repaint the screen.
     */
    (void) sigemptyset(&mask);
    (void) sigaddset(&mask, SIGALRM);
#if USE_SIGWINCH
    (void) sigaddset(&mask, SIGWINCH);
#endif
    (void) sigprocmask(SIG_BLOCK, &mask, &omask);

#ifdef SIGTTOU
    sigttou_blocked = sigismember(&omask, SIGTTOU);
    if (!sigttou_blocked) {
	(void) sigemptyset(&mask);
	(void) sigaddset(&mask, SIGTTOU);
	(void) sigprocmask(SIG_BLOCK, &mask, NULL);
    }
#endif

    /*
     * End window mode, which also resets the terminal state to the
     * original (pre-curses) modes.
     */
    endwin();

    /* Unblock SIGTSTP. */
    (void) sigemptyset(&mask);
    (void) sigaddset(&mask, SIGTSTP);
#ifdef SIGTTOU
    if (!sigttou_blocked) {
	/* Unblock this too if it wasn't blocked on entry */
	(void) sigaddset(&mask, SIGTTOU);
    }
#endif
    (void) sigprocmask(SIG_UNBLOCK, &mask, NULL);

    /* Now we want to resend SIGSTP to this process and suspend it */
    act.sa_handler = SIG_DFL;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
#ifdef SA_RESTART
    act.sa_flags |= SA_RESTART;
#endif /* SA_RESTART */
    sigaction(SIGTSTP, &act, &oact);
    kill(getpid(), SIGTSTP);

    /* Process gets suspended...time passes...process resumes */

    T(("SIGCONT received"));
    sigaction(SIGTSTP, &oact, NULL);
    flushinp();

    /*
     * If the user modified the tty state while suspended, he wants
     * those changes to stick.  So save the new "default" terminal state.
     */
    def_shell_mode();

    /*
     * This relies on the fact that doupdate() will restore the
     * program-mode tty state, and issue enter_ca_mode if need be.
     */
    doupdate();

    /* Reset the signals. */
    (void) sigprocmask(SIG_SETMASK, &omask, NULL);
}
#endif /* USE_SIGTSTP */

static void
cleanup(int sig)
{
    static int nested;

    /*
     * Actually, doing any sort of I/O from within an signal handler is
     * "unsafe".  But we'll _try_ to clean up the screen and terminal
     * settings on the way out.
     */
    if (!nested++
	&& (sig == SIGINT
	    || sig == SIGQUIT)) {
#if HAVE_SIGACTION || HAVE_SIGVEC
	sigaction_t act;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	act.sa_handler = SIG_IGN;
	if (sigaction(sig, &act, (sigaction_t *) 0) == 0)
#else
	if (signal(sig, SIG_IGN) != SIG_ERR)
#endif
	{
	    SCREEN *scan = _nc_screen_chain;
	    while (scan) {
		if (SP != 0
		    && SP->_ofp != 0
		    && isatty(fileno(SP->_ofp))) {
		    SP->_cleanup = TRUE;
		    SP->_outch = _nc_outch;
		}
		set_term(scan);
		endwin();
		if (SP)
		    SP->_endwin = FALSE;        /* in case we have an atexit! */
		scan = scan->_next_screen;
	    }
	}
    }
    exit(EXIT_FAILURE);
}

#if USE_SIGWINCH
static void
sigwinch(int sig GCC_UNUSED)
{
    SCREEN *scan = _nc_screen_chain;
    while (scan) {
	scan->_sig_winch = TRUE;
	scan = scan->_next_screen;
    }
}
#endif /* USE_SIGWINCH */

/*
 * If the given signal is still in its default state, set it to the given
 * handler.
 */
#if HAVE_SIGACTION || HAVE_SIGVEC
static int
CatchIfDefault(int sig, sigaction_t * act)
{
    sigaction_t old_act;

    if (sigaction(sig, (sigaction_t *) 0, &old_act) == 0
	&& (old_act.sa_handler == SIG_DFL
#if USE_SIGWINCH
	    || (sig == SIGWINCH && old_act.sa_handler == SIG_IGN)
#endif
	)) {
	(void) sigaction(sig, act, (sigaction_t *) 0);
	return TRUE;
    }
    return FALSE;
}
#else
static int
CatchIfDefault(int sig, RETSIGTYPE(*handler) (int))
{
    void (*ohandler) (int);

    ohandler = signal(sig, SIG_IGN);
    if (ohandler == SIG_DFL
#if USE_SIGWINCH
	|| (sig == SIGWINCH && ohandler == SIG_IGN)
#endif
	) {
	signal(sig, handler);
	return TRUE;
    } else {
	signal(sig, ohandler);
	return FALSE;
    }
}
#endif

/*
 * This is invoked once at the beginning (e.g., from 'initscr()'), to
 * initialize the signal catchers, and thereafter when spawning a shell (and
 * returning) to disable/enable the SIGTSTP (i.e., ^Z) catcher.
 *
 * If the application has already set one of the signals, we'll not modify it
 * (during initialization).
 *
 * The XSI document implies that we shouldn't keep the SIGTSTP handler if
 * the caller later changes its mind, but that doesn't seem correct.
 */
static void
_nc_signal_handler(bool enable)
{
#if USE_SIGTSTP                 /* Xenix 2.x doesn't have SIGTSTP, for example */
    static sigaction_t act, oact;
    static int ignore;

    if (!ignore) {
	if (!enable) {
	    act.sa_handler = SIG_IGN;
	    sigaction(SIGTSTP, &act, &oact);
	} else if (act.sa_handler) {
	    sigaction(SIGTSTP, &oact, NULL);
	} else {                /*initialize */
	    sigemptyset(&act.sa_mask);
	    act.sa_flags = 0;
#if USE_SIGWINCH
	    act.sa_handler = sigwinch;
	    CatchIfDefault(SIGWINCH, &act);
#endif

#ifdef SA_RESTART
	    act.sa_flags |= SA_RESTART;
#endif /* SA_RESTART */
	    act.sa_handler = cleanup;
	    CatchIfDefault(SIGINT, &act);
	    CatchIfDefault(SIGTERM, &act);

	    act.sa_handler = tstp;
	    if (!CatchIfDefault(SIGTSTP, &act))
		ignore = TRUE;
	}
    }
#else /* !USE_SIGTSTP */
    if (enable) {
#if HAVE_SIGACTION || HAVE_SIGVEC
	static sigaction_t act;
	sigemptyset(&act.sa_mask);
#if USE_SIGWINCH
	act.sa_handler = sigwinch;
	CatchIfDefault(SIGWINCH, &act);
#endif
#ifdef SA_RESTART
	act.sa_flags |= SA_RESTART;
#endif /* SA_RESTART */
	act.sa_handler = cleanup;
	CatchIfDefault(SIGINT, &act);
	CatchIfDefault(SIGTERM, &act);

#else /* !(HAVE_SIGACTION || HAVE_SIGVEC) */

	CatchIfDefault(SIGINT, cleanup);
	CatchIfDefault(SIGTERM, cleanup);
#if USE_SIGWINCH
	CatchIfDefault(SIGWINCH, sigwinch);
#endif
#endif /* !(HAVE_SIGACTION || HAVE_SIGVEC) */
    }
#endif /* !USE_SIGTSTP */
}

/*
 *              def_prog_mode()
 *              def_shell_mode()
 *              reset_prog_mode()
 *              reset_shell_mode()
 *              savetty()
 *              resetty()
 */

#undef tabs

#ifdef TAB3
# define tabs TAB3
#else
# ifdef XTABS
#  define tabs XTABS
# else
#  ifdef OXTABS
#   define tabs OXTABS
#  else
#   define tabs 0
#  endif
# endif
#endif

static int _nc_get_tty_mode(TTY *buf)
{
	if (cur_term == 0
	 || GET_TTY(cur_term->Filedes, buf) != 0)
		return(ERR);
	TR(TRACE_BITS,("_nc_get_tty_mode: %s", _nc_tracebits()));
	return (OK);
}

static int _nc_set_tty_mode(TTY *buf)
{
	if (cur_term == 0
	 || SET_TTY(cur_term->Filedes, buf) != 0)
		return(ERR);
	TR(TRACE_BITS,("_nc_set_tty_mode: %s", _nc_tracebits()));
	return (OK);
}

static int def_shell_mode(void)
{
	T((T_CALLED("def_shell_mode()")));

	/*
	 * Turn off the XTABS bit in the tty structure if it was on.  If XTABS
	 * was on, remove the tab and backtab capabilities.
	 */

	if (_nc_get_tty_mode(&cur_term->Ottyb) != OK)
		returnCode(ERR);
#ifdef TERMIOS
	if (cur_term->Ottyb.c_oflag & tabs)
		tab = back_tab = NULL;
#else
	if (cur_term->Ottyb.sg_flags & XTABS)
		tab = back_tab = NULL;
#endif
	returnCode(OK);
}

static int def_prog_mode(void)
{
	T((T_CALLED("def_prog_mode()")));

	if (_nc_get_tty_mode(&cur_term->Nttyb) != OK)
		returnCode(ERR);
#ifdef TERMIOS
	cur_term->Nttyb.c_oflag &= ~tabs;
#else
	cur_term->Nttyb.sg_flags &= ~XTABS;
#endif
	returnCode(OK);
}

static int reset_prog_mode(void)
{
	T((T_CALLED("reset_prog_mode()")));

	if (cur_term != 0) {
		_nc_set_tty_mode(&cur_term->Nttyb);
		if (SP) {
			if (stdscr && stdscr->_use_keypad)
				_nc_keypad(TRUE);
			NC_BUFFERED(TRUE);
		}
		returnCode(OK);
	}
	returnCode(ERR);
}

static int reset_shell_mode(void)
{
	T((T_CALLED("reset_shell_mode()")));

	if (cur_term != 0) {
		if (SP)
		{
			_nc_keypad(FALSE);
			_nc_flush();
			NC_BUFFERED(FALSE);
		}
		returnCode(_nc_set_tty_mode(&cur_term->Ottyb));
	}
	returnCode(ERR);
}

/*
**      lib_twait.c
**
**      The routine _nc_timed_wait().
**
**      (This file was originally written by Eric Raymond; however except for
**      comments, none of the original code remains - T.Dickey).
*/

#ifdef __BEOS__
#include <OS.h>
#endif


#if USE_FUNC_POLL
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# endif
#elif HAVE_SELECT
# if HAVE_SYS_TIME_H && HAVE_SYS_TIME_SELECT
#  include <sys/time.h>
# endif
# include <sys/select.h>
#endif

static long
_nc_gettime(bool first)
{
    long res;

    static struct timeval t0;
    struct timeval t1;
    gettimeofday(&t1, (struct timezone *) 0);
    if (first) {
	t0 = t1;
    }
    res = (t1.tv_sec - t0.tv_sec) * 1000
	+ (t1.tv_usec - t0.tv_usec) / 1000;
    T(("%s time: %ld msec", first ? "get" : "elapsed", res));
    return res;
}

/*
 * Wait a specified number of milliseconds, returning nonzero if the timer
 * didn't expire before there is activity on the specified file descriptors.
 * The file-descriptors are specified by the mode:
 *      0 - none (absolute time)
 *      1 - ncurses' normal input-descriptor
 *      2 - mouse descriptor, if any
 *      3 - either input or mouse.
 * We return a mask that corresponds to the mode (e.g., 2 for mouse activity).
 *
 * If the milliseconds given are -1, the wait blocks until activity on the
 * descriptors.
 */
static int
_nc_timed_wait(int mode, int milliseconds, int *timeleft)
{
    int fd;
    int count;

    int result;

#if USE_FUNC_POLL
    struct pollfd fds[2];
#elif defined(__BEOS__)
#elif HAVE_SELECT
    static fd_set set;
#endif

    long starttime, returntime;

    T(("start twait: %d milliseconds, mode: %d", milliseconds, mode));

  retry:
    starttime = _nc_gettime(TRUE);

    count = 0;

#if USE_FUNC_POLL
    memset(fds, 0, sizeof(fds));
    if (mode & 1) {
	fds[count].fd = SP->_ifd;
	fds[count].events = POLLIN;
	count++;
    }
    if ((mode & 2)
	&& (fd = SP->_mouse_fd) >= 0) {
	fds[count].fd = fd;
	fds[count].events = POLLIN;
	count++;
    }
    result = poll(fds, count, milliseconds);

#elif defined(__BEOS__)
    /*
     * BeOS's select() is declared in socket.h, so the configure script does
     * not see it.  That's just as well, since that function works only for
     * sockets.  This (using snooze and ioctl) was distilled from Be's patch
     * for ncurses which uses a separate thread to simulate select().
     *
     * FIXME: the return values from the ioctl aren't very clear if we get
     * interrupted.
     */
    result = 0;
    if (mode & 1) {
	bigtime_t d;
	bigtime_t useconds = milliseconds * 1000;
	int n, howmany;

	if (useconds == 0)      /* we're here to go _through_ the loop */
	    useconds = 1;

	for (d = 0; d < useconds; d += 5000) {
	    n = 0;
	    howmany = ioctl(0, 'ichr', &n);
	    if (howmany >= 0 && n > 0) {
		result = 1;
		break;
	    }
	    if (useconds > 1)
		snooze(5000);
	    milliseconds -= 5;
	}
    } else if (milliseconds > 0) {
	snooze(milliseconds * 1000);
	milliseconds = 0;
    }
#elif HAVE_SELECT
    /*
     * select() modifies the fd_set arguments; do this in the
     * loop.
     */
    FD_ZERO(&set);

    if (mode & 1) {
	FD_SET(SP->_ifd, &set);
	count = SP->_ifd + 1;
    }
    if ((mode & 2)
	&& (fd = SP->_mouse_fd) >= 0) {
	FD_SET(fd, &set);
	count = max(fd, count) + 1;
    }

    if (milliseconds >= 0) {
	struct timeval ntimeout;
	ntimeout.tv_sec = milliseconds / 1000;
	ntimeout.tv_usec = (milliseconds % 1000) * 1000;
	result = select(count, &set, NULL, NULL, &ntimeout);
    } else {
	result = select(count, &set, NULL, NULL, NULL);
    }
#endif

    returntime = _nc_gettime(FALSE);

    if (milliseconds >= 0)
	milliseconds -= (returntime - starttime);

    /*
     * If the timeout hasn't expired, and we've gotten no data,
     * this is probably a system where 'select()' needs to be left
     * alone so that it can complete.  Make this process sleep,
     * then come back for more.
     */
    if (result == 0 && milliseconds > 100) {
	napms(100);
	milliseconds -= 100;
	goto retry;
    }

    /* return approximate time left in milliseconds */
    if (timeleft)
	*timeleft = milliseconds;

    T(("end twait: returned %d (%d), remaining time %d msec",
       result, errno, milliseconds));

    /*
     * Both 'poll()' and 'select()' return the number of file descriptors
     * that are active.  Translate this back to the mask that denotes which
     * file-descriptors, so that we don't need all of this system-specific
     * code everywhere.
     */
    if (result != 0) {
	if (result > 0) {
	    result = 0;
#if USE_FUNC_POLL
	    for (count = 0; count < 2; count++) {
		if ((mode & (1 << count))
		    && (fds[count].revents & POLLIN)) {
		    result |= (1 << count);
		}
	    }
#elif defined(__BEOS__)
	    result = 1;         /* redundant, but simple */
#elif HAVE_SELECT
	    if ((mode & 2)
		&& (fd = SP->_mouse_fd) >= 0
		&& FD_ISSET(fd, &set))
		result |= 2;
	    if ((mode & 1)
		&& FD_ISSET(SP->_ifd, &set))
		result |= 1;
#endif
	} else
	    result = 0;
    }

    return (result);
}

/*
**      lib_ungetch.c
**
**      The routine ungetch().
**
*/


static int ungetch(int ch)
{
	if (tail == -1)
		return ERR;
	if (HEAD == -1) {
		HEAD = 0;
		t_inc()
		peek = tail; /* no raw keys */
	} else
		h_dec();

	SP->_fifo[HEAD] = ch;
	T(("ungetch %#x ok", ch));
	return OK;
}

/*
 *      vidputs(newmode, outc)
 *
 *      newmode is taken to be the logical 'or' of the symbols in curses.h
 *      representing graphic renditions.  The terminal is set to be in all of
 *      the given modes, if possible.
 *
 *      if the new attribute is normal
 *              if exit-alt-char-set exists
 *                      emit it
 *              emit exit-attribute-mode
 *      else if set-attributes exists
 *              use it to set exactly what you want
 *      else
 *              if exit-attribute-mode exists
 *                      turn off everything
 *              else
 *                      turn off those which can be turned off and aren't in
 *                      newmode.
 *              turn on each mode which should be on and isn't, one by one
 *
 *      NOTE that this algorithm won't achieve the desired mix of attributes
 *      in some cases, but those are probably just those cases in which it is
 *      actually impossible, anyway, so...
 *
 *      NOTE that we cannot assume that there's no interaction between color
 *      and other attribute resets.  So each time we reset color (or other
 *      attributes) we'll have to be prepared to restore the other.
 */


#define doPut(mode) tputs(mode, 1, outc)

#define TurnOn(mask,mode) \
	if ((turn_on & mask) && mode) { doPut(mode); }

#define TurnOff(mask,mode) \
	if ((turn_off & mask) && mode) { doPut(mode); turn_off &= ~mask; }

	/* if there is no current screen, assume we *can* do color */
#define SetColorsIf(why,old_attr) \
	if (can_color && (why)) { \
		int old_pair = PAIR_NUMBER(old_attr); \
		TR(TRACE_ATTRS, ("old pair = %d -- new pair = %d", old_pair, pair)); \
		if ((pair != old_pair) \
		 || (fix_pair0 && (pair == 0)) \
		 || (reverse ^ ((old_attr & A_REVERSE) != 0))) { \
			_nc_do_color(old_pair, pair, reverse, outc); \
		} \
	}

static int
vidputs(attr_t newmode, int (*outc) (int))
{
    static attr_t previous_attr = A_NORMAL;
    attr_t turn_on, turn_off;
    int pair;
    bool reverse = FALSE;
    bool used_ncv = FALSE;
    bool can_color = (SP == 0 || SP->_coloron);
#if NCURSES_EXT_FUNCS
    bool fix_pair0 = (SP != 0 && SP->_coloron && !SP->_default_color);
#else
#define fix_pair0 FALSE
#endif

    T((T_CALLED("vidputs(%s)"), _traceattr(newmode)));

    /* this allows us to go on whether or not newterm() has been called */
    if (SP)
	previous_attr = SP->_current_attr;

    TR(TRACE_ATTRS, ("previous attribute was %s", _traceattr(previous_attr)));

    if ((SP != 0)
	&& (magic_cookie_glitch > 0))
	newmode &= ~(SP->_xmc_suppress);

    /*
     * If we have a terminal that cannot combine color with video
     * attributes, use the colors in preference.
     */
    if (((newmode & A_COLOR) != 0
	 || fix_pair0)
	&& (no_color_video > 0)) {
	/*
	 * If we had chosen the A_xxx definitions to correspond to the
	 * no_color_video mask, we could simply shift it up and mask off the
	 * attributes.  But we did not (actually copied Solaris' definitions).
	 * However, this is still simpler/faster than a lookup table.
	 *
	 * The 63 corresponds to A_STANDOUT, A_UNDERLINE, A_REVERSE, A_BLINK,
	 * A_DIM, A_BOLD which are 1:1 with no_color_video.  The bits that
	 * correspond to A_INVIS, A_PROTECT (192) must be shifted up 1 and
	 * A_ALTCHARSET (256) down 2 to line up.  We use the NCURSES_BITS
	 * macro so this will work properly for the wide-character layout.
	 */
	attr_t mask = NCURSES_BITS((no_color_video & 63)
				   | ((no_color_video & 192) << 1)
				   | ((no_color_video & 256) >> 2), 8);

	if (mask & A_REVERSE && newmode & A_REVERSE) {
	    reverse = TRUE;
	    mask &= ~A_REVERSE;
	}
	newmode &= ~mask;
    }

    if (newmode == previous_attr)
	returnCode(OK);

    pair = PAIR_NUMBER(newmode);

    if (reverse) {
	newmode &= ~A_REVERSE;
    }

    turn_off = (~newmode & previous_attr) & ALL_BUT_COLOR;
    turn_on = (newmode & ~previous_attr) & ALL_BUT_COLOR;

    SetColorsIf(((pair == 0) && !fix_pair0), previous_attr);

    if (newmode == A_NORMAL) {
	if ((previous_attr & A_ALTCHARSET) && exit_alt_charset_mode) {
	    doPut(exit_alt_charset_mode);
	    previous_attr &= ~A_ALTCHARSET;
	}
	if (previous_attr) {
	    if (exit_attribute_mode) {
		doPut(exit_attribute_mode);
	    } else {
		if (!SP || SP->_use_rmul) {
		    TurnOff(A_UNDERLINE, exit_underline_mode);
		}
		if (!SP || SP->_use_rmso) {
		    TurnOff(A_STANDOUT, exit_standout_mode);
		}
	    }
	    previous_attr &= ~A_COLOR;
	}

	SetColorsIf((pair != 0) || fix_pair0, previous_attr);
    } else if (set_attributes && !used_ncv) {
	if (turn_on || turn_off) {
	    tputs(tparm(set_attributes,
			(newmode & A_STANDOUT) != 0,
			(newmode & A_UNDERLINE) != 0,
			(newmode & A_REVERSE) != 0,
			(newmode & A_BLINK) != 0,
			(newmode & A_DIM) != 0,
			(newmode & A_BOLD) != 0,
			(newmode & A_INVIS) != 0,
			(newmode & A_PROTECT) != 0,
			(newmode & A_ALTCHARSET) != 0), 1, outc);
	    previous_attr &= ~A_COLOR;
	}
	SetColorsIf((pair != 0) || fix_pair0, previous_attr);
    } else {

	TR(TRACE_ATTRS, ("turning %s off", _traceattr(turn_off)));

	TurnOff(A_ALTCHARSET, exit_alt_charset_mode);

	if (!SP || SP->_use_rmul) {
	    TurnOff(A_UNDERLINE, exit_underline_mode);
	}

	if (!SP || SP->_use_rmso) {
	    TurnOff(A_STANDOUT, exit_standout_mode);
	}

	if (turn_off && exit_attribute_mode) {
	    doPut(exit_attribute_mode);
	    turn_on |= (newmode & (chtype) (~A_COLOR));
	    previous_attr &= ~A_COLOR;
	}
	SetColorsIf((pair != 0) || fix_pair0, previous_attr);

	TR(TRACE_ATTRS, ("turning %s on", _traceattr(turn_on)));
	/* *INDENT-OFF* */
	TurnOn(A_ALTCHARSET,    enter_alt_charset_mode);
	TurnOn(A_BLINK,         enter_blink_mode);
	TurnOn(A_BOLD,          enter_bold_mode);
	TurnOn(A_DIM,           enter_dim_mode);
	TurnOn(A_REVERSE,       enter_reverse_mode);
	TurnOn(A_STANDOUT,      enter_standout_mode);
	TurnOn(A_PROTECT,       enter_protected_mode);
	TurnOn(A_INVIS,         enter_secure_mode);
	TurnOn(A_UNDERLINE,     enter_underline_mode);
	TurnOn(A_HORIZONTAL,    enter_horizontal_hl_mode);
	TurnOn(A_LEFT,          enter_left_hl_mode);
	TurnOn(A_LOW,           enter_low_hl_mode);
	TurnOn(A_RIGHT,         enter_right_hl_mode);
	TurnOn(A_TOP,           enter_top_hl_mode);
	TurnOn(A_VERTICAL,      enter_vertical_hl_mode);
	/* *INDENT-ON* */

    }

    if (reverse)
	newmode |= A_REVERSE;

    if (SP)
	SP->_current_attr = newmode;
    else
	previous_attr = newmode;

    returnCode(OK);
}

static int
vidattr(attr_t newmode)
{
    T((T_CALLED("vidattr(%s)"), _traceattr(newmode)));

    returnCode(vidputs(newmode, _nc_outch));
}

static chtype
termattrs(void)
{
    chtype attrs = A_NORMAL;

    T((T_CALLED("termattrs()")));
    if (enter_alt_charset_mode)
	attrs |= A_ALTCHARSET;

    if (enter_blink_mode)
	attrs |= A_BLINK;

    if (enter_bold_mode)
	attrs |= A_BOLD;

    if (enter_dim_mode)
	attrs |= A_DIM;

    if (enter_reverse_mode)
	attrs |= A_REVERSE;

    if (enter_standout_mode)
	attrs |= A_STANDOUT;

    if (enter_protected_mode)
	attrs |= A_PROTECT;

    if (enter_secure_mode)
	attrs |= A_INVIS;

    if (enter_underline_mode)
	attrs |= A_UNDERLINE;

    if (SP->_coloron)
	attrs |= A_COLOR;

    returnChar(attrs);
}

/*
**      lib_winch.c
**
**      The routine winch().
**
*/


chtype
winch(WINDOW *win)
{
    T((T_CALLED("winch(%p)"), win));
    if (win != 0) {
	returnChar(win->_line[win->_cury].text[win->_curx]);
    } else {
	returnChar(0);
    }
}

/*
**      lib_window.c
**
**
*/


static void _nc_synchook(WINDOW *win)
/* hook to be called after each window change */
{
	if (win->_immed) wrefresh(win);
	if (win->_sync) wsyncup(win);
}

static void wsyncup(WINDOW *win)
/* mark changed every cell in win's ancestors that is changed in win */
/* Rewritten by J. Pfeifer, 1-Apr-96 (don't even think that...)      */
{
  WINDOW        *wp;

  if (win && win->_parent)
    for (wp = win; wp->_parent; wp = wp->_parent)
      {
	int y;
	WINDOW *pp = wp->_parent;

	assert((wp->_pary <= pp->_maxy) &&
	       ((wp->_pary+wp->_maxy) <= pp->_maxy));

	for (y = 0; y <= wp->_maxy; y++)
	  {
	    int left = wp->_line[y].firstchar;
	    if (left >= 0) /* line is touched */
	      {
		struct ldat *line = &(pp->_line[wp->_pary + y]);
		/* left & right character in parent window coordinates */
		int right = wp->_line[y].lastchar + wp->_parx;
		left += wp->_parx;

		CHANGED_RANGE(line, left, right);
	      }
	  }
      }
}

static void wsyncdown(WINDOW *win)
/* mark changed every cell in win that is changed in any of its ancestors */
/* Rewritten by J. Pfeifer, 1-Apr-96 (don't even think that...)           */
{
  if (win && win->_parent)
    {
      WINDOW *pp = win->_parent;
      int y;

      /* This recursion guarantees, that the changes are propagated down-
	 wards from the root to our direct parent. */
      wsyncdown(pp);

      /* and now we only have to propagate the changes from our direct
	 parent, if there are any. */
      assert((win->_pary <= pp->_maxy) &&
	     ((win->_pary + win->_maxy) <= pp->_maxy));

      for (y = 0; y <= win->_maxy; y++)
	{
	  if (pp->_line[win->_pary + y].firstchar >= 0) /* parent changed */
	    {
	      struct ldat *line = &(win->_line[y]);
	      /* left and right character in child coordinates */
	      int left  = pp->_line[win->_pary + y].firstchar - win->_parx;
	      int right = pp->_line[win->_pary + y].lastchar  - win->_parx;
	      /* The change maybe outside the childs range */
	      if (left<0)
		left = 0;
	      if (right > win->_maxx)
		right = win->_maxx;
	      CHANGED_RANGE(line, left, right);
	    }
	}
    }
}

/*
 *      read_entry.c -- Routine for reading in a compiled terminfo file
 *
 */

/*
 *      int
 *      _nc_read_file_entry(filename, ptr)
 *
 *      Read the compiled terminfo entry in the given file into the
 *      structure pointed to by ptr, allocating space for the string
 *      table.
 */

#undef  BYTE
#define BYTE(p,n)       (unsigned char)((p)[n])

#define IS_NEG1(p)      ((BYTE(p,0) == 0377) && (BYTE(p,1) == 0377))
#define IS_NEG2(p)      ((BYTE(p,0) == 0376) && (BYTE(p,1) == 0377))
#define LOW_MSB(p)      (BYTE(p,0) + 256*BYTE(p,1))

static bool have_tic_directory = FALSE;
static bool keep_tic_directory = FALSE;

/*
 * Record the "official" location of the terminfo directory, according to
 * the place where we're writing to, or the normal default, if not.
 */
static const char *
_nc_tic_dir(const char *path)
{
    static const char *result = TERMINFO;

    if (!keep_tic_directory) {
	if (path != 0) {
	    result = path;
	    have_tic_directory = TRUE;
	} else if (!have_tic_directory && use_terminfo_vars()) {
	    char *envp;
	    if ((envp = getenv("TERMINFO")) != 0)
		return _nc_tic_dir(envp);
	}
    }
    return result;
}

static void
convert_shorts(char *buf, short *Numbers, int count)
{
    int i;
    for (i = 0; i < count; i++) {
	if (IS_NEG1(buf + 2 * i))
	    Numbers[i] = ABSENT_NUMERIC;
	else if (IS_NEG2(buf + 2 * i))
	    Numbers[i] = CANCELLED_NUMERIC;
	else
	    Numbers[i] = LOW_MSB(buf + 2 * i);
	TR(TRACE_DATABASE, ("get Numbers[%d]=%d", i, Numbers[i]));
    }
}

#define MAX_OF_TYPE(t)   (int)(((unsigned t)(~0))>>1)
static void
convert_32bits(char *buf, short *Numbers, int count)
{
    int i, j;
    unsigned char ch;
    for (i = 0; i < count; i++) {
	int value = 0;
	for (j = 0; j < 4; ++j) {
	    ch = (unsigned char)(*buf++);
	    value |= (ch << (8 * j));
	}
	if (value == -1)
	    Numbers[i] = ABSENT_NUMERIC;
	else if (value == -2)
	    Numbers[i] = CANCELLED_NUMERIC;
	else if (value > MAX_OF_TYPE(short))
	    Numbers[i] = MAX_OF_TYPE(short);
	else
	    Numbers[i] = (short) value;
	TR(TRACE_DATABASE, ("get Numbers[%d]=%d", i, Numbers[i]));
    }
}

static void
convert_strings(char *buf, char **Strings, int count, int size, char *table)
{
    int i;
    char *p;

    for (i = 0; i < count; i++) {
	if (IS_NEG1(buf + 2 * i)) {
	    Strings[i] = ABSENT_STRING;
	} else if (IS_NEG2(buf + 2 * i)) {
	    Strings[i] = CANCELLED_STRING;
	} else if (LOW_MSB(buf + 2 * i) > size) {
	    Strings[i] = ABSENT_STRING;
	} else {
	    Strings[i] = (LOW_MSB(buf + 2 * i) + table);
	    TR(TRACE_DATABASE, ("Strings[%d] = %s", i, _nc_visbuf(Strings[i])));
	}

	/* make sure all strings are NUL terminated */
	if (VALID_STRING(Strings[i])) {
	    for (p = Strings[i]; p <= table + size; p++)
		if (*p == '\0')
		    break;
	    /* if there is no NUL, ignore the string */
	    if (p > table + size)
		Strings[i] = ABSENT_STRING;
	}
    }
}

#define read_shorts(fd, buf, count) (read(fd, buf, (count)*2) == (count)*2)
#define read_numbers(fd, buf, count) \
	(read(fd, buf, (count)*(unsigned)size_of_numbers) == (int) (count)*size_of_numbers)

#define ead_(fd, buf, count) (read(fd, buf, (count)*2) == (count)*2)

#define even_boundary(value) \
    if ((value) % 2 != 0) read(fd, buf, 1)

static int
read_termtype(int fd, TERMTYPE * ptr)
/* return 1 if read, 0 if not found or garbled */
{
    int name_size, bool_count, num_count, str_count, str_size;
    int i, magic2 = 0, size_of_numbers = 2;
    char buf[MAX_ENTRY_SIZE];

    TR(TRACE_DATABASE, ("READ termtype header @%d", tell(fd)));

    memset(ptr, 0, sizeof(*ptr));

    /* grab the header */
    if (!read_shorts(fd, buf, 6)
	|| !IS_TIC_MAGIC(buf)) {
	return 0;
    }
    magic2 = LOW_MSB(buf) == MAGIC2;
    if (magic2)
	size_of_numbers = 4;


    _nc_free_termtype(ptr);
    name_size = LOW_MSB(buf + 2);
    bool_count = LOW_MSB(buf + 4);
    num_count = LOW_MSB(buf + 6);
    str_count = LOW_MSB(buf + 8);
    str_size = LOW_MSB(buf + 10);

    TR(TRACE_DATABASE,
       ("TERMTYPE name_size=%d, bool=%d/%d, num=%d/%d str=%d/%d(%d)",
	name_size, bool_count, BOOLCOUNT, num_count, NUMCOUNT,
	str_count, STRCOUNT, str_size));
    if (name_size < 0
	|| bool_count < 0
	|| num_count < 0
	|| str_count < 0
	|| str_size < 0) {
	return 0;
    }

    if (str_size) {
	/* try to allocate space for the string table */
	if (str_count * 2 >= (int) sizeof(buf)
	    || (ptr->str_table = typeMalloc(char, (unsigned) str_size)) == 0) {
	    return 0;
	}
    } else {
	str_count = 0;
    }

    /* grab the name (a null-terminate string) */
    read(fd, buf, min(MAX_NAME_SIZE, (unsigned) name_size));
    buf[MAX_NAME_SIZE] = '\0';
    ptr->term_names = typeCalloc(char, strlen(buf) + 1);
    if (ptr->term_names == NULL) {
	return 0;
    }
    (void) strcpy(ptr->term_names, buf);
    if (name_size > MAX_NAME_SIZE)
	lseek(fd, (off_t) (name_size - MAX_NAME_SIZE), 1);

    /* grab the booleans */
    if ((ptr->Booleans = typeCalloc(char, max(BOOLCOUNT, bool_count))) == 0
	|| read(fd, ptr->Booleans, (unsigned) bool_count) < bool_count) {
	return 0;
    }

    /*
     * If booleans end on an odd byte, skip it.  The machine they
     * originally wrote terminfo on must have been a 16-bit
     * word-oriented machine that would trap out if you tried a
     * word access off a 2-byte boundary.
     */
    even_boundary(name_size + bool_count);

    /* grab the numbers */
    if ((ptr->Numbers = typeCalloc(short, max(NUMCOUNT, num_count))) == 0
	|| !read_numbers(fd, buf, num_count)) {
	return 0;
    }
    if (!magic2) {
	convert_shorts(buf, ptr->Numbers, num_count);
    } else {
	convert_32bits(buf, ptr->Numbers, num_count);
    }

    if ((ptr->Strings = typeCalloc(char *, max(STRCOUNT, str_count))) == 0)
	  return 0;

    if (str_count) {
	/* grab the string offsets */
	if (!read_shorts(fd, buf, str_count)) {
	    return 0;
	}
	/* finally, grab the string table itself */
	if (read(fd, ptr->str_table, (unsigned) str_size) != str_size)
	    return 0;
	convert_strings(buf, ptr->Strings, str_count, str_size, ptr->str_table);
    }
#if NCURSES_XNAMES

    ptr->num_Booleans = BOOLCOUNT;
    ptr->num_Numbers = NUMCOUNT;
    ptr->num_Strings = STRCOUNT;

    /*
     * Read extended entries, if any, after the normal end of terminfo data.
     */
    even_boundary(str_size);
    TR(TRACE_DATABASE, ("READ extended_header @%d", tell(fd)));
    if (_nc_user_definable && read_shorts(fd, buf, 5)) {
	int ext_bool_count = LOW_MSB(buf + 0);
	int ext_num_count = LOW_MSB(buf + 2);
	int ext_str_count = LOW_MSB(buf + 4);
	int ext_str_size = LOW_MSB(buf + 6);
	int ext_str_limit = LOW_MSB(buf + 8);
	int need = (ext_bool_count + ext_num_count + ext_str_count);
	int base = 0;

	if (need >= (int) sizeof(buf)
	    || ext_str_size >= (int) sizeof(buf)
	    || ext_str_limit >= (int) sizeof(buf)
	    || ext_bool_count < 0
	    || ext_num_count < 0
	    || ext_str_count < 0
	    || ext_str_size < 0
	    || ext_str_limit < 0)
	    return 0;

	ptr->num_Booleans = BOOLCOUNT + ext_bool_count;
	ptr->num_Numbers = NUMCOUNT + ext_num_count;
	ptr->num_Strings = STRCOUNT + ext_str_count;

	ptr->Booleans = typeRealloc(char, ptr->num_Booleans, ptr->Booleans);
	ptr->Numbers = typeRealloc(short, ptr->num_Numbers, ptr->Numbers);
	ptr->Strings = typeRealloc(char *, ptr->num_Strings, ptr->Strings);

	TR(TRACE_DATABASE, ("extended header is %d/%d/%d(%d:%d)",
			    ext_bool_count, ext_num_count, ext_str_count,
			    ext_str_size, ext_str_limit));

	TR(TRACE_DATABASE, ("READ %d extended-booleans @%d",
			    ext_bool_count, tell(fd)));
	if ((ptr->ext_Booleans = ext_bool_count) != 0) {
	    if (read(fd, ptr->Booleans + BOOLCOUNT, (unsigned)
		     ext_bool_count) != ext_bool_count)
		return 0;
	}
	even_boundary(ext_bool_count);

	TR(TRACE_DATABASE, ("READ %d extended-numbers @%d",
			    ext_num_count, tell(fd)));
	if ((ptr->ext_Numbers = ext_num_count) != 0) {
	    if (!read_numbers(fd, buf, ext_num_count))
		return 0;
	    TR(TRACE_DATABASE, ("Before converting extended-numbers"));
	    if (!magic2) {
		convert_shorts(buf, ptr->Numbers + NUMCOUNT, ext_num_count);
	    } else {
		convert_32bits(buf, ptr->Numbers + NUMCOUNT, ext_num_count);
	    }
	}

	TR(TRACE_DATABASE, ("READ extended-offsets @%d", tell(fd)));
	if ((ext_str_count || need)
	    && !read_shorts(fd, buf, ext_str_count + need))
	    return 0;

	TR(TRACE_DATABASE, ("READ %d bytes of extended-strings @%d",
			    ext_str_limit, tell(fd)));

	if (ext_str_limit) {
	    if ((ptr->ext_str_table = typeMalloc(char, ext_str_limit)) == 0)
		  return 0;
	    if (read(fd, ptr->ext_str_table, ext_str_limit) != ext_str_limit)
		return 0;
	    TR(TRACE_DATABASE, ("first extended-string is %s", _nc_visbuf(ptr->ext_str_table)));
	}

	if ((ptr->ext_Strings = ext_str_count) != 0) {
	    TR(TRACE_DATABASE,
	       ("Before computing extended-string capabilities str_count=%d, ext_str_count=%d",
		str_count, ext_str_count));
	    convert_strings(buf, ptr->Strings + str_count, ext_str_count,
			    ext_str_limit, ptr->ext_str_table);
	    for (i = ext_str_count - 1; i >= 0; i--) {
		TR(TRACE_DATABASE, ("MOVE from [%d:%d] %s",
				    i, i + str_count,
				    _nc_visbuf(ptr->Strings[i + str_count])));
		ptr->Strings[i + STRCOUNT] = ptr->Strings[i + str_count];
		if (VALID_STRING(ptr->Strings[i + STRCOUNT]))
		    base += (strlen(ptr->Strings[i + STRCOUNT]) + 1);
		TR(TRACE_DATABASE, ("... to    [%d] %s",
				    i + STRCOUNT,
				    _nc_visbuf(ptr->Strings[i + STRCOUNT])));
	    }
	}

	if (need) {
	    if ((ptr->ext_Names = typeCalloc(char *, need)) == 0)
		  return 0;
	    TR(TRACE_DATABASE,
	       ("ext_NAMES starting @%d in extended_strings, first = %s",
		base, _nc_visbuf(ptr->ext_str_table + base)));
	    convert_strings(buf + (2 * ext_str_count), ptr->ext_Names, need,
			    ext_str_limit, ptr->ext_str_table + base);
	}

	T(("...done reading terminfo bool %d(%d) num %d(%d) str %d(%d)",
	   ptr->num_Booleans, ptr->ext_Booleans,
	   ptr->num_Numbers, ptr->ext_Numbers,
	   ptr->num_Strings, ptr->ext_Strings));

	TR(TRACE_DATABASE, ("extend: num_Booleans:%d", ptr->num_Booleans));
    } else
#endif /* NCURSES_XNAMES */
    {
	T(("...done reading terminfo bool %d num %d str %d",
	   bool_count, num_count, str_count));
#if NCURSES_XNAMES
	TR(TRACE_DATABASE, ("normal: num_Booleans:%d", ptr->num_Booleans));
#endif
    }

    for (i = bool_count; i < BOOLCOUNT; i++)
	ptr->Booleans[i] = FALSE;
    for (i = num_count; i < NUMCOUNT; i++)
	ptr->Numbers[i] = ABSENT_NUMERIC;
    for (i = str_count; i < STRCOUNT; i++)
	ptr->Strings[i] = ABSENT_STRING;

    return 1;
}

static int
_nc_read_file_entry(const char *const filename, TERMTYPE * ptr)
/* return 1 if read, 0 if not found or garbled */
{
    int code, fd = -1;

    if (_nc_access(filename, R_OK) < 0
	|| (fd = open(filename, O_RDONLY | O_BINARY)) < 0) {
	T(("cannot open terminfo %s (errno=%d)", filename, errno));
	return 0;
    }

    T(("read terminfo %s", filename));
    if ((code = read_termtype(fd, ptr)) == 0)
	_nc_free_termtype(ptr);
    close(fd);

    return (code);
}

/*
 * Build a terminfo pathname and try to read the data.  Returns 1 on success,
 * 0 on failure.
 */
static int
_nc_read_tic_entry(char *const filename,
		   const char *const dir, const char *ttn, TERMTYPE * const tp)
{
/* maximum safe length of terminfo root directory name */
#define MAX_TPATH       (PATH_MAX - MAX_ALIAS - 6)

    if (strlen(dir) > MAX_TPATH)
	return 0;
    (void) sprintf(filename, "%s/%s", dir, ttn);
    return _nc_read_file_entry(filename, tp);
}

/*
 * Process the list of :-separated directories, looking for the terminal type.
 * We don't use strtok because it does not show us empty tokens.
 */
static int
_nc_read_terminfo_dirs(const char *dirs, char *const filename, const char *const
		       ttn, TERMTYPE * const tp)
{
    char *list, *a;
    const char *b;
    int code = 0;

    /* we'll modify the argument, so we must copy */
    if ((b = a = list = strdup(dirs)) == NULL)
	return 0;

    for (;;) {
	int c = *a;
	if (c == 0 || c == ':') {
	    *a = 0;
	    if ((b + 1) >= a)
		b = TERMINFO;
	    if (_nc_read_tic_entry(filename, b, ttn, tp) == 1) {
		code = 1;
		break;
	    }
	    b = a + 1;
	    if (c == 0)
		break;
	}
	a++;
    }

    free(list);
    return (code);
}

/*
 *      _nc_read_entry(char *tn, char *filename, TERMTYPE *tp)
 *
 *      Find and read the compiled entry for a given terminal type,
 *      if it exists.  We take pains here to make sure no combination
 *      of environment variables and terminal type name can be used to
 *      overrun the file buffer.
 */

static int
_nc_read_entry(const char *const tn, char *const filename, TERMTYPE * const tp)
{
    char *envp;
    char ttn[MAX_ALIAS + 3];

    /* truncate the terminal name to prevent dangerous buffer airline */
    (void) sprintf(ttn, "%c/%.*s", *tn, MAX_ALIAS, tn);

    /* This is System V behavior, in conjunction with our requirements for
     * writing terminfo entries.
     */
    if (have_tic_directory
	&& _nc_read_tic_entry(filename, _nc_tic_dir(0), ttn, tp) == 1)
	return 1;

    if (use_terminfo_vars()) {
	if ((envp = getenv("TERMINFO")) != 0
	    && _nc_read_tic_entry(filename, _nc_tic_dir(envp), ttn, tp) == 1)
	    return 1;

	/* this is an ncurses extension */
	if ((envp = _nc_home_terminfo()) != 0) {
	    if (_nc_read_tic_entry(filename, envp, ttn, tp) == 1) {
		return 1;
	    }
	}

	/* this is an ncurses extension */
	if ((envp = getenv("TERMINFO_DIRS")) != 0)
	    return _nc_read_terminfo_dirs(envp, filename, ttn, tp);
    }

    /* Try the system directory.  Note that the TERMINFO_DIRS value, if
     * defined by the configure script, begins with a ":", which will be
     * interpreted as TERMINFO.
     */
#ifdef TERMINFO_DIRS
    return _nc_read_terminfo_dirs(TERMINFO_DIRS, filename, ttn, tp);
#else
    return _nc_read_tic_entry(filename, TERMINFO, ttn, tp);
#endif
}

/*
 * This is an extension to the curses library.  It provides callers with a hook
 * into the NCURSES data to resize windows, primarily for use by programs
 * running in an X Window terminal (e.g., xterm).  I abstracted this module
 * from my application library for NCURSES because it must be compiled with
 * the private data structures -- T.Dickey 1995/7/4.
 */


/*
 * This function reallocates NCURSES window structures.  It is invoked in
 * response to a SIGWINCH interrupt.  Other user-defined windows may also need
 * to be reallocated.
 *
 * Because this performs memory allocation, it should not (in general) be
 * invoked directly from the signal handler.
 */
static int
resizeterm(int ToLines, int ToCols)
{
	int stolen = screen_lines - SP->_lines_avail;
	int bottom = screen_lines + SP->_topstolen - stolen;

	T((T_CALLED("resizeterm(%d,%d) old(%d,%d)"),
		ToLines, ToCols,
		screen_lines, screen_columns));

	SP->_sig_winch = FALSE;

	if (ToLines != screen_lines
	 || ToCols  != screen_columns) {
		WINDOWLIST *wp;

#if USE_SIGWINCH
		ungetch(KEY_RESIZE);    /* so application can know this */
		clearok(curscr, TRUE);  /* screen contents are unknown */
#endif

		for (wp = _nc_windows; wp != 0; wp = wp->next) {
			WINDOW *win = wp->win;
			int myLines = win->_maxy + 1;
			int myCols  = win->_maxx + 1;

			/* pads aren't treated this way */
			if (win->_flags & _ISPAD)
				continue;

			if (win->_begy >= bottom) {
				win->_begy += (ToLines - screen_lines);
			} else {
				if (myLines == screen_lines - stolen
				 && ToLines != screen_lines)
					myLines = ToLines - stolen;
				else
				if (myLines == screen_lines
				 && ToLines != screen_lines)
					myLines = ToLines;
			}

			if (myCols  == screen_columns
			 && ToCols  != screen_columns)
				myCols = ToCols;

			if (wresize(win, myLines, myCols) != OK)
				returnCode(ERR);
		}

		screen_lines   = lines    = ToLines;
		screen_columns = columns  = ToCols;

		SP->_lines_avail = lines - stolen;

		if (SP->oldhash) { FreeAndNull(SP->oldhash); }
		if (SP->newhash) { FreeAndNull(SP->newhash); }
	}

	/*
	 * Always update LINES, to allow for call from lib_doupdate.c which
	 * needs to have the count adjusted by the stolen (ripped off) lines.
	 */
	LINES = ToLines - stolen;
	COLS  = ToCols;

	returnCode(OK);
}

/*
 * Wrapper for vsprintf that allocates a buffer big enough to hold the result.
 */
static char *
_nc_printf_string(const char *fmt, va_list ap)
{
	static int rows, cols;
	static char *buf;
	static size_t len;

	if (screen_lines > rows || screen_columns > cols) {
		if (screen_lines   > rows) rows = screen_lines;
		if (screen_columns > cols) cols = screen_columns;
		len = (rows * (cols + 1)) + 1;
		buf = typeRealloc(char, len, buf);
		if (buf == 0) {
			return(0);
		}
	}

	if (buf != 0) {
# if HAVE_VSNPRINTF
		vsnprintf(buf, len, fmt, ap);   /* GNU extension */
# else
		vsprintf(buf, fmt, ap);         /* ANSI */
# endif
	}
	return buf;
}

/*
**      setbuf.c
**
**      Support for set_term(), reset_shell_mode(), reset_prog_mode().
**
*/


/*
 * If the output file descriptor is connected to a tty (the typical case) it
 * will probably be line-buffered.  Keith Bostic pointed out that we don't want
 * this; it hoses people running over networks by forcing out a bunch of small
 * packets instead of one big one, so screen updates on ptys look jerky.
 * Restore block buffering to prevent this minor lossage.
 *
 * The buffer size is a compromise.  Ideally we'd like a buffer that can hold
 * the maximum possible update size (the whole screen plus cup commands to
 * change lines as it's painted).  On a 66-line xterm this can become
 * excessive.  So we min it with the amount of data we think we can get through
 * two Ethernet packets (maximum packet size - 100 for TCP/IP overhead).
 *
 * Why two ethernet packets?  It used to be one, on the theory that said
 * packets define the maximum size of atomic update.  But that's less than the
 * 2000 chars on a 25 x 80 screen, and we don't want local updates to flicker
 * either.  Two packet lengths will handle up to a 35 x 80 screen.
 *
 * The magic '6' is the estimated length of the end-of-line cup sequence to go
 * to the next line.  It's generous.  We used to mess with the buffering in
 * init_mvcur() after cost computation, but that lost the sequences emitted by
 * init_acs() in setupscreen().
 *
 * "The setvbuf function may be used only after the stream pointed to by stream
 * has been associated with an open file and before any other operation is
 * performed on the stream." (ISO 7.9.5.6.)
 *
 * Grrrr...
 *
 * On a lighter note, many implementations do in fact allow an application to
 * reset the buffering after it has been written to.  We try to do this because
 * otherwise we leave stdout in buffered mode after endwin() is called.  (This
 * also happens with SVr4 curses).
 *
 * There are pros/cons:
 *
 * con:
 *      There is no guarantee that we can reestablish buffering once we've
 *      dropped it.
 *
 *      We _may_ lose data if the implementation does not coordinate this with
 *      fflush.
 *
 * pro:
 *      An implementation is more likely to refuse to change the buffering than
 *      to do it in one of the ways mentioned above.
 *
 *      The alternative is to have the application try to change buffering
 *      itself, which is certainly no improvement.
 *
 * Just in case it does not work well on a particular system, the calls to
 * change buffering are all via the macro NC_BUFFERED.  Some implementations
 * do indeed get confused by changing setbuf on/off, and will overrun the
 * buffer.  So we disable this by default (there may yet be a workaround).
 */
static void _nc_set_buffer(FILE *ofp, bool buffered)
{
	/* optional optimization hack -- do before any output to ofp */
#if HAVE_SETVBUF || HAVE_SETBUFFER
	unsigned buf_len;
	char *buf_ptr;

	if (getenv("NCURSES_NO_SETBUF") != 0)
		return;

	fflush(ofp);
	if ((SP->_buffered = buffered) != 0) {
		buf_len = min(LINES * (COLS + 6), 2800);
		if ((buf_ptr = SP->_setbuf) == 0) {
			if ((buf_ptr = typeMalloc(char, buf_len)) == NULL)
				return;
			SP->_setbuf = buf_ptr;
			/* Don't try to free this! */
		}
#if !USE_SETBUF_0
		else return;
#endif
	} else {
#if !USE_SETBUF_0
		return;
#else
		buf_len = 0;
		buf_ptr = 0;
#endif
	}

#if HAVE_SETVBUF
#ifdef SETVBUF_REVERSED /* pre-svr3? */
	(void) setvbuf(ofp, buf_ptr, buf_len, buf_len ? _IOFBF : _IOLBF);
#else
	(void) setvbuf(ofp, buf_ptr, buf_len ? _IOFBF : _IOLBF, buf_len);
#endif
#elif HAVE_SETBUFFER
	(void) setbuffer(ofp, buf_ptr, (int)buf_len);
#endif

#endif /* HAVE_SETVBUF || HAVE_SETBUFFER */
}


/* This file provides sigaction() emulation using sigvec() */
/* Use only if this is non POSIX system */

#if !HAVE_SIGACTION && HAVE_SIGVEC

static int
sigaction (int sig, sigaction_t * sigact, sigaction_t * osigact)
{
  return sigvec(sig, sigact, osigact);
}

static int
sigemptyset (sigset_t * mask)
{
  *mask = 0;
  return 0;
}

static int
sigprocmask (int mode, sigset_t * mask, sigset_t * omask)
{
   sigset_t current = sigsetmask(0);

   if (omask) *omask = current;

   if (mode==SIG_BLOCK)
      current |= *mask;
   else if (mode==SIG_UNBLOCK)
      current &= ~*mask;
   else if (mode==SIG_SETMASK)
      current = *mask;

   sigsetmask(current);
   return 0;
}

static int
sigsuspend (sigset_t * mask)
{
  return sigpause (*mask);
}

static int
sigdelset (sigset_t * mask, int sig)
{
  *mask &= ~sigmask (sig);
  return 0;
}

static int
sigaddset (sigset_t * mask, int sig)
{
  *mask |= sigmask (sig);
  return 0;
}

static int
sigismember (sigset_t * mask, int sig)
{
  return (*mask & sigmask (sig)) != 0;
}

#endif

/*
**      lib_mvcur.c
**/


/****************************************************************************
 * Useful string functions (especially for mvcur)
 ****************************************************************************/

/*
 * Initialize the descriptor so we can append to it.
 */
static string_desc *
_nc_str_init(string_desc * dst, char *src, size_t len)
{
    if (dst != 0) {
	dst->s_head = src;
	dst->s_tail = src;
	dst->s_size = len - 1;
	if (src != 0)
	    *src = 0;
    }
    return dst;
}

/*
 * Initialize the descriptor for only tracking the amount of memory used.
 */
static string_desc *
_nc_str_null(string_desc * dst, size_t len)
{
    return _nc_str_init(dst, 0, len);
}

/*
 * Copy a descriptor
 */
static string_desc *
_nc_str_copy(string_desc * dst, string_desc * src)
{
    *dst = *src;
    return dst;
}

/*
 * Replaces strcat into a fixed buffer, returning false on failure.
 */
static bool
_nc_safe_strcat(string_desc * dst, const char *src)
{
    if (src != 0) {
	size_t len = strlen(src);

	if (len < dst->s_size) {
	    if (dst->s_tail != 0) {
		strcpy(dst->s_tail, src);
		dst->s_tail += len;
	    }
	    dst->s_size -= len;
	    return TRUE;
	}
    }
    return FALSE;
}

/*
 * Replaces strcpy into a fixed buffer, returning false on failure.
 */
static bool
_nc_safe_strcpy(string_desc * dst, const char *src)
{
    if (src != 0) {
	size_t len = strlen(src);

	if (len < dst->s_size) {
	    if (dst->s_head != 0) {
		strcpy(dst->s_head, src);
		dst->s_tail = dst->s_head + len;
	    }
	    dst->s_size -= len;
	    return TRUE;
	}
    }
    return FALSE;
}

/*
**      tries.c
**
**      Functions to manage the tree of partial-completions for keycodes.
**
*/


/*
 * Expand a keycode into the string that it corresponds to, returning null if
 * no match was found, otherwise allocating a string of the result.
 */
static char *_nc_expand_try(struct tries *tree, unsigned short code, int *count, size_t len)
{
	struct tries *ptr = tree;
	char *result = 0;

	if (code != 0) {
		while (ptr != 0) {
			if ((result = _nc_expand_try(ptr->child, code, count, len + 1)) != 0) {
				break;
			}
			if (ptr->value == code) {
				*count -= 1;
				if (*count == -1) {
					result = typeCalloc(char, len+2);
					break;
				}
			}
			ptr = ptr->sibling;
		}
	}
	if (result != 0) {
		if ((result[len] = ptr->ch) == 0)
			*((unsigned char *)(result+len)) = 128;
	}
	return result;
}

/*
 * Remove a code from the specified tree, freeing the unused nodes.  Returns
 * true if the code was found/removed.
 */
static int _nc_remove_key(struct tries **tree, unsigned short code)
{
	T((T_CALLED("_nc_remove_key(%p,%d)"), tree, code));

	if (code == 0)
		returnCode(FALSE);

	while (*tree != 0) {
		if (_nc_remove_key(&(*tree)->child, code)) {
			returnCode(TRUE);
		}
		if ((*tree)->value == code) {
			if((*tree)->child) {
				/* don't cut the whole sub-tree */
				(*tree)->value = 0;
			} else {
				struct tries *to_free = *tree;
				*tree = (*tree)->sibling;
				free(to_free);
			}
			returnCode(TRUE);
		}
		tree = &(*tree)->sibling;
	}
	returnCode(FALSE);
}


/*-----------------------------------------------------------------
 *
 *      lib_doupdate.c
 *
 *      The routine doupdate() and its dependents.  Also _nc_outstr(),
 *      so all physical output is concentrated here (except _nc_outch()
 *      in lib_tputs.c).
 *
 *-----------------------------------------------------------------*/

#if HAVE_SYS_TIME_H && HAVE_SYS_TIME_SELECT
#include <sys/time.h>
#endif

#if USE_FUNC_POLL
#elif HAVE_SELECT
#include <sys/select.h>
#endif


/*
 * This define controls the line-breakout optimization.  Every once in a
 * while during screen refresh, we want to check for input and abort the
 * update if there's some waiting.  CHECK_INTERVAL controls the number of
 * changed lines to be emitted between input checks.
 *
 * Note: Input-check-and-abort is no longer done if the screen is being
 * updated from scratch.  This is a feature, not a bug.
 */
#define CHECK_INTERVAL  5

#define FILL_BCE() (SP->_coloron && !SP->_default_color && !back_color_erase)

/*
 * Enable checking to see if doupdate and friends are tracking the true
 * cursor position correctly.  NOTE: this is a debugging hack which will
 * work ONLY on ANSI-compatible terminals!
 */
/* #define POSITION_DEBUG */

static int ClrBottom(int total);
static void ClearScreen(chtype blank);
static void ClrUpdate(void);
static void DelChar(int count);
static void InsStr(chtype * line, int count);
static void TransformLine(int const lineno);

#define position_check(expected_y, expected_x, legend)  /* nothing */

/****************************************************************************
 *
 * Optimized update code
 *
 ****************************************************************************/

static inline void
GoTo(int const row, int const col)
{
    chtype oldattr = SP->_current_attr;

    TR(TRACE_MOVE, ("GoTo(%d, %d) from (%d, %d)",
		    row, col, SP->_cursrow, SP->_curscol));

    position_check(SP->_cursrow, SP->_curscol, "GoTo");

    /*
     * Force restore even if msgr is on when we're in an alternate
     * character set -- these have a strong tendency to screw up the
     * CR & LF used for local character motions!
     */
    if ((oldattr & A_ALTCHARSET)
	|| (oldattr && !move_standout_mode)) {
	TR(TRACE_CHARPUT, ("turning off (%#lx) %s before move",
			   oldattr, _traceattr(oldattr)));
	vidattr(A_NORMAL);
    }

    mvcur(SP->_cursrow, SP->_curscol, row, col);
    SP->_cursrow = row;
    SP->_curscol = col;
    position_check(SP->_cursrow, SP->_curscol, "GoTo2");
}

static inline void
PutAttrChar(chtype ch)
{
    int data;

    if (tilde_glitch && (TextOf(ch) == '~'))
	ch = ('`' | AttrOf(ch));

    TR(TRACE_CHARPUT, ("PutAttrChar(%s) at (%d, %d)",
		       _tracechtype(ch),
		       SP->_cursrow, SP->_curscol));
    UpdateAttrs(ch);
    data = TextOf(ch);
    if (SP->_outch != 0) {
	SP->_outch(data);
    } else {
	putc(data, SP->_ofp);   /* macro's fastest... */
    }
    SP->_curscol++;
    if (char_padding) {
	putp(char_padding);
    }
}

static bool
check_pending(void)
/* check for pending input */
{
    bool have_pending = FALSE;

    /*
     * Only carry out this check when the flag is zero, otherwise we'll
     * have the refreshing slow down drastically (or stop) if there's an
     * unread character available.
     */
    if (SP->_fifohold != 0)
	return FALSE;

    if (SP->_checkfd >= 0) {
#if USE_FUNC_POLL
	struct pollfd fds[1];
	fds[0].fd = SP->_checkfd;
	fds[0].events = POLLIN;
	if (poll(fds, 1, 0) > 0) {
	    have_pending = TRUE;
	}
#elif defined(__BEOS__)
	/*
	 * BeOS's select() is declared in socket.h, so the configure script does
	 * not see it.  That's just as well, since that function works only for
	 * sockets.  This (using snooze and ioctl) was distilled from Be's patch
	 * for ncurses which uses a separate thread to simulate select().
	 *
	 * FIXME: the return values from the ioctl aren't very clear if we get
	 * interrupted.
	 */
	int n = 0;
	int howmany = ioctl(0, 'ichr', &n);
	if (howmany >= 0 && n > 0) {
	    have_pending = TRUE;
	}
#elif HAVE_SELECT
	fd_set fdset;
	struct timeval ktimeout;

	ktimeout.tv_sec =
	    ktimeout.tv_usec = 0;

	FD_ZERO(&fdset);
	FD_SET(SP->_checkfd, &fdset);
	if (select(SP->_checkfd + 1, &fdset, NULL, NULL, &ktimeout) != 0) {
	    have_pending = TRUE;
	}
#endif
    }
    if (have_pending) {
	SP->_fifohold = 5;
	_nc_flush();
    }
    return FALSE;
}

/*
 * No one supports recursive inline functions.  However, gcc is quieter if we
 * instantiate the recursive part separately.
 */
static void callPutChar(chtype const);

/* put char at lower right corner */
static void
PutCharLR(chtype const ch)
{
    if (!auto_right_margin) {
	/* we can put the char directly */
	PutAttrChar(ch);
    } else if (enter_am_mode && exit_am_mode) {
	/* we can suppress automargin */
	putp(exit_am_mode);

	PutAttrChar(ch);
	SP->_curscol--;
	position_check(SP->_cursrow, SP->_curscol, "exit_am_mode");

	putp(enter_am_mode);
    } else if ((enter_insert_mode && exit_insert_mode)
	       || insert_character || parm_ich) {
	GoTo(screen_lines - 1, screen_columns - 2);
	callPutChar(ch);
	GoTo(screen_lines - 1, screen_columns - 2);
	InsStr(newscr->_line[screen_lines - 1].text + screen_columns - 2, 1);
    }
}

static void
wrap_cursor(void)
{
    if (eat_newline_glitch) {
	/*
	 * xenl can manifest two different ways.  The vt100
	 * way is that, when you'd expect the cursor to wrap,
	 * it stays hung at the right margin (on top of the
	 * character just emitted) and doesn't wrap until the
	 * *next* graphic char is emitted.  The c100 way is
	 * to ignore LF received just after an am wrap.
	 *
	 * An aggressive way to handle this would be to
	 * emit CR/LF after the char and then assume the wrap
	 * is done, you're on the first position of the next
	 * line, and the terminal out of its weird state.
	 * Here it's safe to just tell the code that the
	 * cursor is in hyperspace and let the next mvcur()
	 * call straighten things out.
	 */
	SP->_curscol = -1;
	SP->_cursrow = -1;
    } else if (auto_right_margin) {
	SP->_curscol = 0;
	SP->_cursrow++;
    } else {
	SP->_curscol--;
    }
    position_check(SP->_cursrow, SP->_curscol, "wrap_cursor");
}

static inline void
PutChar(chtype const ch)
/* insert character, handling automargin stuff */
{
    if (SP->_cursrow == screen_lines - 1 && SP->_curscol == screen_columns - 1)
	PutCharLR(ch);
    else
	PutAttrChar(ch);

    if (SP->_curscol >= screen_columns)
	wrap_cursor();

    position_check(SP->_cursrow, SP->_curscol, "PutChar");
}

/*
 * Check whether the given character can be output by clearing commands.  This
 * includes test for being a space and not including any 'bad' attributes, such
 * as A_REVERSE.  All attribute flags which don't affect appearance of a space
 * or can be output by clearing (A_COLOR in case of bce-terminal) are excluded.
 */
static inline bool
can_clear_with(chtype ch)
{
    if (!back_color_erase && SP->_coloron) {
	if (ch & A_COLOR)
	    return FALSE;
#if NCURSES_EXT_FUNCS
	if (!SP->_default_color)
	    return FALSE;
	if (SP->_default_fg != C_MASK || SP->_default_bg != C_MASK)
	    return FALSE;
#endif
    }
    return ((ch & ~(NONBLANK_ATTR | A_COLOR)) == BLANK);
}

/*
 * Issue a given span of characters from an array.
 * Must be functionally equivalent to:
 *      for (i = 0; i < num; i++)
 *          PutChar(ntext[i]);
 * but can leave the cursor positioned at the middle of the interval.
 *
 * Returns: 0 - cursor is at the end of interval
 *          1 - cursor is somewhere in the middle
 *
 * This code is optimized using ech and rep.
 */
static int
EmitRange(const chtype * ntext, int num)
{
    int i;

    if (erase_chars || repeat_char) {
	while (num > 0) {
	    int runcount;
	    chtype ntext0;

	    while (num > 1 && ntext[0] != ntext[1]) {
		PutChar(ntext[0]);
		ntext++;
		num--;
	    }
	    ntext0 = ntext[0];
	    if (num == 1) {
		PutChar(ntext0);
		return 0;
	    }
	    runcount = 2;

	    while (runcount < num && ntext[runcount] == ntext0)
		runcount++;

	    /*
	     * The cost expression in the middle isn't exactly right.
	     * _cup_ch_cost is an upper bound on the cost for moving to the
	     * end of the erased area, but not the cost itself (which we
	     * can't compute without emitting the move).  This may result
	     * in erase_chars not getting used in some situations for
	     * which it would be marginally advantageous.
	     */
	    if (erase_chars
		&& runcount > SP->_ech_cost + SP->_cup_ch_cost
		&& can_clear_with(ntext0)) {
		UpdateAttrs(ntext0);
		putp(tparm(erase_chars, runcount));

		/*
		 * If this is the last part of the given interval,
		 * don't bother moving cursor, since it can be the
		 * last update on the line.
		 */
		if (runcount < num) {
		    GoTo(SP->_cursrow, SP->_curscol + runcount);
		} else {
		    return 1;   /* cursor stays in the middle */
		}
	    } else if (repeat_char && runcount > SP->_rep_cost) {
		bool wrap_possible = (SP->_curscol + runcount >= screen_columns);
		int rep_count = runcount;

		if (wrap_possible)
		    rep_count--;

		UpdateAttrs(ntext0);
		putp(tparm(repeat_char, TextOf(ntext0), rep_count));
		SP->_curscol += rep_count;

		if (wrap_possible)
		    PutChar(ntext0);
	    } else {
		for (i = 0; i < runcount; i++)
		    PutChar(ntext[i]);
	    }
	    ntext += runcount;
	    num -= runcount;
	}
	return 0;
    }

    for (i = 0; i < num; i++)
	PutChar(ntext[i]);
    return 0;
}

/*
 * Output the line in the given range [first .. last]
 *
 * If there's a run of identical characters that's long enough to justify
 * cursor movement, use that also.
 *
 * Returns: same as EmitRange
 */
static int
PutRange(
	    const chtype * otext,
	    const chtype * ntext,
	    int row,
	    int first, int last)
{
    int j, run;

    TR(TRACE_CHARPUT, ("PutRange(%p, %p, %d, %d, %d)",
		       otext, ntext, row, first, last));

    if (otext != ntext
	&& (last - first + 1) > SP->_inline_cost) {
	for (j = first, run = 0; j <= last; j++) {
	    if (otext[j] == ntext[j]) {
		run++;
	    } else {
		if (run > SP->_inline_cost) {
		    int before_run = (j - run);
		    EmitRange(ntext + first, before_run - first);
		    GoTo(row, first = j);
		}
		run = 0;
	    }
	}
    }
    return EmitRange(ntext + first, last - first + 1);
}

static void
callPutChar(chtype const ch)
{
    PutChar(ch);
}

/* leave unbracketed here so 'indent' works */
#define MARK_NOCHANGE(win,row) \
		win->_line[row].firstchar = _NOCHANGE; \
		win->_line[row].lastchar = _NOCHANGE; \
		if_USE_SCROLL_HINTS(win->_line[row].oldindex = row)

int
doupdate(void)
{
    int i;
    int nonempty;

    T((T_CALLED("doupdate()")));

    _nc_signal_handler(FALSE);

    if (SP->_fifohold)
	SP->_fifohold--;

#if USE_SIZECHANGE
    if (SP->_endwin || SP->_sig_winch) {
	/*
	 * This is a transparent extension:  XSI does not address it,
	 * and applications need not know that ncurses can do it.
	 *
	 * Check if the terminal size has changed while curses was off
	 * (this can happen in an xterm, for example), and resize the
	 * ncurses data structures accordingly.
	 */
	_nc_update_screensize();
    }
#endif

    if (SP->_endwin) {

	T(("coming back from shell mode"));
	reset_prog_mode();

	_nc_mvcur_resume();
	_nc_screen_resume();
	SP->_mouse_resume(SP);

	SP->_endwin = FALSE;
    }

    nonempty = 0;
    if (curscr->_clear || newscr->_clear) {     /* force refresh ? */
	TR(TRACE_UPDATE, ("clearing and updating from scratch"));
	ClrUpdate();
	curscr->_clear = FALSE; /* reset flag */
	newscr->_clear = FALSE; /* reset flag */
    } else {
	int changedlines = CHECK_INTERVAL;

	if (check_pending())
	    goto cleanup;

	nonempty = min(screen_lines, newscr->_maxy + 1);

	if (SP->_scrolling) {
	    _nc_scroll_optimize();
	}

	nonempty = ClrBottom(nonempty);

	TR(TRACE_UPDATE, ("Transforming lines, nonempty %d", nonempty));
	for (i = 0; i < nonempty; i++) {
	    /*
	     * Here is our line-breakout optimization.
	     */
	    if (changedlines == CHECK_INTERVAL) {
		if (check_pending())
		    goto cleanup;
		changedlines = 0;
	    }

	    /*
	     * newscr->line[i].firstchar is normally set
	     * by wnoutrefresh.  curscr->line[i].firstchar
	     * is normally set by _nc_scroll_window in the
	     * vertical-movement optimization code,
	     */
	    if (newscr->_line[i].firstchar != _NOCHANGE
		|| curscr->_line[i].firstchar != _NOCHANGE) {
		TransformLine(i);
		changedlines++;
	    }

	    /* mark line changed successfully */
	    if (i <= newscr->_maxy) {
		MARK_NOCHANGE(newscr, i)
	    }
	    if (i <= curscr->_maxy) {
		MARK_NOCHANGE(curscr, i)
	    }
	}
    }

    /* put everything back in sync */
    for (i = nonempty; i <= newscr->_maxy; i++) {
	MARK_NOCHANGE(newscr, i)
    }
    for (i = nonempty; i <= curscr->_maxy; i++) {
	MARK_NOCHANGE(curscr, i)
    }

    if (!newscr->_leaveok) {
	curscr->_curx = newscr->_curx;
	curscr->_cury = newscr->_cury;

	GoTo(curscr->_cury, curscr->_curx);
    }

  cleanup:
    /*
     * Keep the physical screen in normal mode in case we get other
     * processes writing to the screen.
     */
    UpdateAttrs(A_NORMAL);

    _nc_flush();
    curscr->_attrs = newscr->_attrs;

    _nc_signal_handler(TRUE);

    returnCode(OK);
}

/*
 *      ClrBlank(win)
 *
 *      Returns the attributed character that corresponds to the "cleared"
 *      screen.  If the terminal has the back-color-erase feature, this will be
 *      colored according to the wbkgd() call.
 *
 *      We treat 'curscr' specially because it isn't supposed to be set directly
 *      in the wbkgd() call.  Assume 'stdscr' for this case.
 */
#define BCE_ATTRS (A_NORMAL|A_COLOR)
#define BCE_BKGD(win) (((win) == curscr ? stdscr : (win))->_bkgd)

static inline chtype
ClrBlank(WINDOW *win)
{
    chtype blank = BLANK;
    if (back_color_erase)
	blank |= (BCE_BKGD(win) & BCE_ATTRS);
    return blank;
}

/*
**      ClrUpdate()
**
**      Update by clearing and redrawing the entire screen.
**
*/

static void
ClrUpdate(void)
{
    int i;
    chtype blank = ClrBlank(stdscr);
    int nonempty = min(screen_lines, newscr->_maxy + 1);

    TR(TRACE_UPDATE, ("ClrUpdate() called"));

    ClearScreen(blank);

    TR(TRACE_UPDATE, ("updating screen from scratch"));

    nonempty = ClrBottom(nonempty);

    for (i = 0; i < nonempty; i++)
	TransformLine(i);
}

/*
**      ClrToEOL(blank)
**
**      Clear to end of current line, starting at the cursor position
*/

static void
ClrToEOL(chtype blank, bool needclear)
{
    int j;

    if (curscr != 0
	&& SP->_cursrow >= 0) {
	for (j = SP->_curscol; j < screen_columns; j++) {
	    if (j >= 0) {
		chtype *cp = &(curscr->_line[SP->_cursrow].text[j]);

		if (*cp != blank) {
		    *cp = blank;
		    needclear = TRUE;
		}
	    }
	}
    } else {
	needclear = TRUE;
    }

    if (needclear) {
	UpdateAttrs(blank);
	if (SP->_el_cost > (screen_columns - SP->_curscol)) {
	    int count = (screen_columns - SP->_curscol);
	    while (count-- > 0)
		PutChar(blank);
	} else {
	    putp(clr_eol);
	}
    }
}

/*
**      ClrToEOS(blank)
**
**      Clear to end of screen, starting at the cursor position
*/

static void
ClrToEOS(chtype blank)
{
    int row, col;

    row = SP->_cursrow;
    col = SP->_curscol;

    UpdateAttrs(blank);
    tputs(clr_eos, screen_lines - row, _nc_outch);

    while (col < screen_columns)
	curscr->_line[row].text[col++] = blank;

    for (row++; row < screen_lines; row++) {
	for (col = 0; col < screen_columns; col++)
	    curscr->_line[row].text[col] = blank;
    }
}

/*
 *      ClrBottom(total)
 *
 *      Test if clearing the end of the screen would satisfy part of the
 *      screen-update.  Do this by scanning backwards through the lines in the
 *      screen, checking if each is blank, and one or more are changed.
 */
static int
ClrBottom(int total)
{
    int row;
    int col;
    int top = total;
    int last = min(screen_columns, newscr->_maxx + 1);
    chtype blank = ClrBlank(stdscr);
    bool ok;

    if (clr_eos && can_clear_with(blank)) {

	for (row = total - 1; row >= 0; row--) {
	    for (col = 0, ok = TRUE; ok && col < last; col++) {
		ok = (newscr->_line[row].text[col] == blank);
	    }
	    if (!ok)
		break;

	    for (col = 0; ok && col < last; col++) {
		ok = (curscr->_line[row].text[col] == blank);
	    }
	    if (!ok)
		top = row;
	}

	/* don't use clr_eos for just one line if clr_eol available */
	if (top < total - 1 || (top < total && !clr_eol && !clr_bol)) {
	    GoTo(top, 0);
	    ClrToEOS(blank);
	    total = top;
	    if (SP->oldhash && SP->newhash) {
		for (row = top; row < screen_lines; row++)
		    SP->oldhash[row] = SP->newhash[row];
	    }
	}
    }
    return total;
}

/*
**      TransformLine(lineno)
**
**      Transform the given line in curscr to the one in newscr, using
**      Insert/Delete Character if _nc_idcok && has_ic().
**
**              firstChar = position of first different character in line
**              oLastChar = position of last different character in old line
**              nLastChar = position of last different character in new line
**
**              move to firstChar
**              overwrite chars up to min(oLastChar, nLastChar)
**              if oLastChar < nLastChar
**                      insert newLine[oLastChar+1..nLastChar]
**              else
**                      delete oLastChar - nLastChar spaces
*/

static void
TransformLine(int const lineno)
{
    int firstChar, oLastChar, nLastChar;
    chtype *newLine = newscr->_line[lineno].text;
    chtype *oldLine = curscr->_line[lineno].text;
    int n;
    bool attrchanged = FALSE;

    TR(TRACE_UPDATE, ("TransformLine(%d) called", lineno));

    /* copy new hash value to old one */
    if (SP->oldhash && SP->newhash)
	SP->oldhash[lineno] = SP->newhash[lineno];

#define ColorOf(n) ((n) & A_COLOR)
#define unColor(n) ((n) & ALL_BUT_COLOR)
    /*
     * If we have colors, there is the possibility of having two color pairs
     * that display as the same colors.  For instance, Lynx does this.  Check
     * for this case, and update the old line with the new line's colors when
     * they are equivalent.
     */
    if (SP->_coloron) {
	chtype oldColor;
	chtype newColor;
	int oldPair;
	int newPair;

	for (n = 0; n < screen_columns; n++) {
	    if (newLine[n] != oldLine[n]) {
		oldColor = ColorOf(oldLine[n]);
		newColor = ColorOf(newLine[n]);
		if (oldColor != newColor
		    && unColor(oldLine[n]) == unColor(newLine[n])) {
		    oldPair = PAIR_NUMBER(oldColor);
		    newPair = PAIR_NUMBER(newColor);
		    if (oldPair < COLOR_PAIRS
			&& newPair < COLOR_PAIRS
			&& SP->_color_pairs[oldPair] == SP->_color_pairs[newPair]) {
			oldLine[n] &= ~A_COLOR;
			oldLine[n] |= ColorOf(newLine[n]);
		    }
		}
	    }
	}
    }

    if (ceol_standout_glitch && clr_eol) {
	firstChar = 0;
	while (firstChar < screen_columns) {
	    if (AttrOf(newLine[firstChar]) != AttrOf(oldLine[firstChar]))
		attrchanged = TRUE;
	    firstChar++;
	}
    }

    firstChar = 0;

    if (attrchanged) {          /* we may have to disregard the whole line */
	GoTo(lineno, firstChar);
	ClrToEOL(ClrBlank(curscr), FALSE);
	PutRange(oldLine, newLine, lineno, 0, (screen_columns - 1));
    } else {
	chtype blank;

	/* find the first differing character */
	while (firstChar < screen_columns &&
	       newLine[firstChar] == oldLine[firstChar])
	    firstChar++;

	/* if there wasn't one, we're done */
	if (firstChar >= screen_columns)
	    return;

	/* it may be cheap to clear leading whitespace with clr_bol */
	if (clr_bol && can_clear_with(blank = newLine[0])) {
	    int oFirstChar, nFirstChar;

	    for (oFirstChar = 0; oFirstChar < screen_columns; oFirstChar++)
		if (oldLine[oFirstChar] != blank)
		    break;
	    for (nFirstChar = 0; nFirstChar < screen_columns; nFirstChar++)
		if (newLine[nFirstChar] != blank)
		    break;

	    if (nFirstChar > oFirstChar + SP->_el1_cost) {
		if (nFirstChar >= screen_columns && SP->_el_cost <= SP->_el1_cost) {
		    GoTo(lineno, 0);
		    UpdateAttrs(blank);
		    putp(clr_eol);
		} else {
		    GoTo(lineno, nFirstChar - 1);
		    UpdateAttrs(blank);
		    putp(clr_bol);
		}

		while (firstChar < nFirstChar)
		    oldLine[firstChar++] = blank;

		if (firstChar >= screen_columns)
		    return;
	    }
	}

	blank = newLine[screen_columns - 1];

	if (!can_clear_with(blank)) {
	    /* find the last differing character */
	    nLastChar = screen_columns - 1;

	    while (nLastChar > firstChar
		   && newLine[nLastChar] == oldLine[nLastChar])
		nLastChar--;

	    if (nLastChar >= firstChar) {
		GoTo(lineno, firstChar);
		PutRange(oldLine, newLine, lineno, firstChar, nLastChar);
		memcpy(oldLine + firstChar,
		       newLine + firstChar,
		       (nLastChar - firstChar + 1) * sizeof(chtype));
	    }
	    return;
	}

	/* find last non-blank character on old line */
	oLastChar = screen_columns - 1;
	while (oLastChar > firstChar && oldLine[oLastChar] == blank)
	    oLastChar--;

	/* find last non-blank character on new line */
	nLastChar = screen_columns - 1;
	while (nLastChar > firstChar && newLine[nLastChar] == blank)
	    nLastChar--;

	if ((nLastChar == firstChar)
	    && (SP->_el_cost < (oLastChar - nLastChar))) {
	    GoTo(lineno, firstChar);
	    if (newLine[firstChar] != blank)
		PutChar(newLine[firstChar]);
	    ClrToEOL(blank, FALSE);
	} else if ((nLastChar != oLastChar)
		   && (newLine[nLastChar] != oldLine[oLastChar]
		       || !(_nc_idcok && has_ic()))) {
	    GoTo(lineno, firstChar);
	    if ((oLastChar - nLastChar) > SP->_el_cost) {
		if (PutRange(oldLine, newLine, lineno, firstChar, nLastChar))
		    GoTo(lineno, nLastChar + 1);
		ClrToEOL(blank, FALSE);
	    } else {
		n = max(nLastChar, oLastChar);
		PutRange(oldLine, newLine, lineno, firstChar, n);
	    }
	} else {
	    int nLastNonblank = nLastChar;
	    int oLastNonblank = oLastChar;

	    /* find the last characters that really differ */
	    while (newLine[nLastChar] == oldLine[oLastChar]) {
		if (nLastChar != 0
		    && oLastChar != 0) {
		    nLastChar--;
		    oLastChar--;
		} else {
		    break;
		}
	    }

	    n = min(oLastChar, nLastChar);
	    if (n >= firstChar) {
		GoTo(lineno, firstChar);
		PutRange(oldLine, newLine, lineno, firstChar, n);
	    }

	    if (oLastChar < nLastChar) {
		int m = max(nLastNonblank, oLastNonblank);
		GoTo(lineno, n + 1);
		if (InsCharCost(nLastChar - oLastChar)
		    > (m - n)) {
		    PutRange(oldLine, newLine, lineno, n + 1, m);
		} else {
		    InsStr(&newLine[n + 1], nLastChar - oLastChar);
		}
	    } else if (oLastChar > nLastChar) {
		GoTo(lineno, n + 1);
		if (DelCharCost(oLastChar - nLastChar)
		    > SP->_el_cost + nLastNonblank - (n + 1)) {
		    if (PutRange(oldLine, newLine, lineno,
				 n + 1, nLastNonblank))
			GoTo(lineno, nLastNonblank + 1);
		    ClrToEOL(blank, FALSE);
		} else {
		    /*
		     * The delete-char sequence will
		     * effectively shift in blanks from the
		     * right margin of the screen.  Ensure
		     * that they are the right color by
		     * setting the video attributes from
		     * the last character on the row.
		     */
		    UpdateAttrs(blank);
		    DelChar(oLastChar - nLastChar);
		}
	    }
	}
    }

    /* update the code's internal representation */
    if (screen_columns > firstChar)
	memcpy(oldLine + firstChar,
	       newLine + firstChar,
	       (screen_columns - firstChar) * sizeof(chtype));
}

/*
**      ClearScreen(blank)
**
**      Clear the physical screen and put cursor at home
**
*/

static void
ClearScreen(chtype blank)
{
    int i, j;
    bool fast_clear = (clear_screen || clr_eos || clr_eol);

    TR(TRACE_UPDATE, ("ClearScreen() called"));

#if NCURSES_EXT_FUNCS
    if (SP->_coloron
	&& !SP->_default_color) {
	_nc_do_color(COLOR_PAIR(SP->_current_attr), 0, FALSE, _nc_outch);
	if (!back_color_erase) {
	    fast_clear = FALSE;
	}
    }
#endif

    if (fast_clear) {
	if (clear_screen) {
	    UpdateAttrs(blank);
	    putp(clear_screen);
	    SP->_cursrow = SP->_curscol = 0;
	    position_check(SP->_cursrow, SP->_curscol, "ClearScreen");
	} else if (clr_eos) {
	    SP->_cursrow = SP->_curscol = -1;
	    GoTo(0, 0);

	    UpdateAttrs(blank);
	    putp(clr_eos);
	} else if (clr_eol) {
	    SP->_cursrow = SP->_curscol = -1;

	    UpdateAttrs(blank);
	    for (i = 0; i < screen_lines; i++) {
		GoTo(i, 0);
		putp(clr_eol);
	    }
	    GoTo(0, 0);
	}
    } else {
	UpdateAttrs(blank);
	for (i = 0; i < screen_lines; i++) {
	    GoTo(i, 0);
	    for (j = 0; j < screen_columns; j++)
		PutChar(blank);
	}
	GoTo(0, 0);
    }

    for (i = 0; i < screen_lines; i++) {
	for (j = 0; j < screen_columns; j++)
	    curscr->_line[i].text[j] = blank;
    }

    TR(TRACE_UPDATE, ("screen cleared"));
}

/*
**      InsStr(line, count)
**
**      Insert the count characters pointed to by line.
**
*/

static void
InsStr(chtype * line, int count)
{
    TR(TRACE_UPDATE, ("InsStr(%p,%d) called", line, count));

    /* Prefer parm_ich as it has the smallest cost - no need to shift
     * the whole line on each character. */
    /* The order must match that of InsCharCost. */
    if (parm_ich) {
	tputs(tparm(parm_ich, count), count, _nc_outch);
	while (count) {
	    PutAttrChar(*line);
	    line++;
	    count--;
	}
    } else if (enter_insert_mode && exit_insert_mode) {
	putp(enter_insert_mode);
	while (count) {
	    PutAttrChar(*line);
	    if (insert_padding) {
		putp(insert_padding);
	    }
	    line++;
	    count--;
	}
	putp(exit_insert_mode);
    } else {
	while (count) {
	    putp(insert_character);
	    PutAttrChar(*line);
	    if (insert_padding) {
		putp(insert_padding);
	    }
	    line++;
	    count--;
	}
    }
    position_check(SP->_cursrow, SP->_curscol, "InsStr");
}

/*
**      DelChar(count)
**
**      Delete count characters at current position
**
*/

static void
DelChar(int count)
{
    int n;

    TR(TRACE_UPDATE, ("DelChar(%d) called, position = (%d,%d)", count,
		      newscr->_cury, newscr->_curx));

    if (parm_dch) {
	tputs(tparm(parm_dch, count), count, _nc_outch);
    } else {
	for (n = 0; n < count; n++) {
	    putp(delete_character);
	}
    }
}

/*
 * Physical-scrolling support
 *
 * This code was adapted from Keith Bostic's hardware scrolling
 * support for 4.4BSD curses.  I (esr) translated it to use terminfo
 * capabilities, narrowed the call interface slightly, and cleaned
 * up some convoluted tests.  I also added support for the memory_above
 * memory_below, and non_dest_scroll_region capabilities.
 *
 * For this code to work, we must have either
 * change_scroll_region and scroll forward/reverse commands, or
 * insert and delete line capabilities.
 * When the scrolling region has been set, the cursor has to
 * be at the last line of the region to make the scroll up
 * happen, or on the first line of region to scroll down.
 *
 * This code makes one aesthetic decision in the opposite way from
 * BSD curses.  BSD curses preferred pairs of il/dl operations
 * over scrolls, allegedly because il/dl looked faster.  We, on
 * the other hand, prefer scrolls because (a) they're just as fast
 * on many terminals and (b) using them avoids bouncing an
 * unchanged bottom section of the screen up and down, which is
 * visually nasty.
 *
 * (lav): added more cases, used dl/il when bot==maxy and in csr case.
 *
 * I used assumption that capabilities il/il1/dl/dl1 work inside
 * changed scroll region not shifting screen contents outside of it.
 * If there are any terminals behaving different way, it would be
 * necessary to add some conditions to scroll_csr_forward/backward.
 */

/* Try to scroll up assuming given csr (miny, maxy). Returns ERR on failure */
static int
scroll_csr_forward(int n, int top, int bot, int miny, int maxy, chtype blank)
{
    int i, j;

    if (n == 1 && scroll_forward && top == miny && bot == maxy) {
	GoTo(bot, 0);
	UpdateAttrs(blank);
	tputs(scroll_forward, 0, _nc_outch);
    } else if (n == 1 && delete_line && bot == maxy) {
	GoTo(top, 0);
	UpdateAttrs(blank);
	tputs(delete_line, 0, _nc_outch);
    } else if (parm_index && top == miny && bot == maxy) {
	GoTo(bot, 0);
	UpdateAttrs(blank);
	tputs(tparm(parm_index, n, 0), n, _nc_outch);
    } else if (parm_delete_line && bot == maxy) {
	GoTo(top, 0);
	UpdateAttrs(blank);
	tputs(tparm(parm_delete_line, n, 0), n, _nc_outch);
    } else if (scroll_forward && top == miny && bot == maxy) {
	GoTo(bot, 0);
	UpdateAttrs(blank);
	for (i = 0; i < n; i++) {
	    tputs(scroll_forward, 0, _nc_outch);
	}
    } else if (delete_line && bot == maxy) {
	GoTo(top, 0);
	UpdateAttrs(blank);
	for (i = 0; i < n; i++) {
	    tputs(delete_line, 0, _nc_outch);
	}
    } else
	return ERR;

#if NCURSES_EXT_FUNCS
    if (FILL_BCE()) {
	for (i = 0; i < n; i++) {
	    GoTo(bot - i, 0);
	    for (j = 0; j < screen_columns; j++)
		PutChar(blank);
	}
    }
#endif
    return OK;
}

/* Try to scroll down assuming given csr (miny, maxy). Returns ERR on failure */
/* n > 0 */
static int
scroll_csr_backward(int n, int top, int bot, int miny, int maxy, chtype blank)
{
    int i, j;

    if (n == 1 && scroll_reverse && top == miny && bot == maxy) {
	GoTo(top, 0);
	UpdateAttrs(blank);
	tputs(scroll_reverse, 0, _nc_outch);
    } else if (n == 1 && insert_line && bot == maxy) {
	GoTo(top, 0);
	UpdateAttrs(blank);
	tputs(insert_line, 0, _nc_outch);
    } else if (parm_rindex && top == miny && bot == maxy) {
	GoTo(top, 0);
	UpdateAttrs(blank);
	tputs(tparm(parm_rindex, n, 0), n, _nc_outch);
    } else if (parm_insert_line && bot == maxy) {
	GoTo(top, 0);
	UpdateAttrs(blank);
	tputs(tparm(parm_insert_line, n, 0), n, _nc_outch);
    } else if (scroll_reverse && top == miny && bot == maxy) {
	GoTo(top, 0);
	UpdateAttrs(blank);
	for (i = 0; i < n; i++) {
	    tputs(scroll_reverse, 0, _nc_outch);
	}
    } else if (insert_line && bot == maxy) {
	GoTo(top, 0);
	UpdateAttrs(blank);
	for (i = 0; i < n; i++) {
	    tputs(insert_line, 0, _nc_outch);
	}
    } else
	return ERR;

#if NCURSES_EXT_FUNCS
    if (FILL_BCE()) {
	for (i = 0; i < n; i++) {
	    GoTo(top + i, 0);
	    for (j = 0; j < screen_columns; j++)
		PutChar(blank);
	}
    }
#endif
    return OK;
}

/* scroll by using delete_line at del and insert_line at ins */
/* n > 0 */
static int
scroll_idl(int n, int del, int ins, chtype blank)
{
    int i;

    if (!((parm_delete_line || delete_line) && (parm_insert_line || insert_line)))
	return ERR;

    GoTo(del, 0);
    UpdateAttrs(blank);
    if (n == 1 && delete_line) {
	tputs(delete_line, 0, _nc_outch);
    } else if (parm_delete_line) {
	tputs(tparm(parm_delete_line, n, 0), n, _nc_outch);
    } else {                    /* if (delete_line) */
	for (i = 0; i < n; i++) {
	    tputs(delete_line, 0, _nc_outch);
	}
    }

    GoTo(ins, 0);
    UpdateAttrs(blank);
    if (n == 1 && insert_line) {
	tputs(insert_line, 0, _nc_outch);
    } else if (parm_insert_line) {
	tputs(tparm(parm_insert_line, n, 0), n, _nc_outch);
    } else {                    /* if (insert_line) */
	for (i = 0; i < n; i++) {
	    tputs(insert_line, 0, _nc_outch);
	}
    }

    return OK;
}

static int
_nc_scrolln(int n, int top, int bot, int maxy)
/* scroll region from top to bot by n lines */
{
    chtype blank = ClrBlank(stdscr);
    int i;
    bool cursor_saved = FALSE;
    int res;

    TR(TRACE_MOVE, ("mvcur_scrolln(%d, %d, %d, %d)", n, top, bot, maxy));

    if (n > 0) {                /* scroll up (forward) */
	/*
	 * Explicitly clear if stuff pushed off top of region might
	 * be saved by the terminal.
	 */
	res = scroll_csr_forward(n, top, bot, 0, maxy, blank);

	if (res == ERR && change_scroll_region) {
	    if ((((n == 1 && scroll_forward) || parm_index)
		 && (SP->_cursrow == bot || SP->_cursrow == bot - 1))
		&& save_cursor && restore_cursor) {
		cursor_saved = TRUE;
		tputs(save_cursor, 0, _nc_outch);
	    }
	    tputs(tparm(change_scroll_region, top, bot), 0, _nc_outch);
	    if (cursor_saved) {
		tputs(restore_cursor, 0, _nc_outch);
	    } else {
		SP->_cursrow = SP->_curscol = -1;
	    }

	    res = scroll_csr_forward(n, top, bot, top, bot, blank);

	    tputs(tparm(change_scroll_region, 0, maxy), 0, _nc_outch);
	    SP->_cursrow = SP->_curscol = -1;
	}

	if (res == ERR && _nc_idlok)
	    res = scroll_idl(n, top, bot - n + 1, blank);

	/*
	 * Clear the newly shifted-in text.
	 */
	if (res != ERR
	    && (non_dest_scroll_region || (memory_below && bot == maxy))) {
	    if (bot == maxy && clr_eos) {
		GoTo(bot - n, 0);
		ClrToEOS(BLANK);
	    } else {
		for (i = 0; i < n; i++) {
		    GoTo(bot - i, 0);
		    ClrToEOL(BLANK, FALSE);
		}
	    }
	}

    } else {                    /* (n < 0) - scroll down (backward) */
	res = scroll_csr_backward(-n, top, bot, 0, maxy, blank);

	if (res == ERR && change_scroll_region) {
	    if (top != 0 && (SP->_cursrow == top || SP->_cursrow == top - 1)
		&& save_cursor && restore_cursor) {
		cursor_saved = TRUE;
		tputs(save_cursor, 0, _nc_outch);
	    }
	    tputs(tparm(change_scroll_region, top, bot), 0, _nc_outch);
	    if (cursor_saved) {
		tputs(restore_cursor, 0, _nc_outch);
	    } else {
		SP->_cursrow = SP->_curscol = -1;
	    }

	    res = scroll_csr_backward(-n, top, bot, top, bot, blank);

	    tputs(tparm(change_scroll_region, 0, maxy), 0, _nc_outch);
	    SP->_cursrow = SP->_curscol = -1;
	}

	if (res == ERR && _nc_idlok)
	    res = scroll_idl(-n, bot + n + 1, top, blank);

	/*
	 * Clear the newly shifted-in text.
	 */
	if (res != ERR
	    && (non_dest_scroll_region || (memory_above && top == 0))) {
	    for (i = 0; i < -n; i++) {
		GoTo(i + top, 0);
		ClrToEOL(BLANK, FALSE);
	    }
	}
    }

    if (res == ERR)
	return (ERR);

    _nc_scroll_window(curscr, n, top, bot, blank);

    /* shift hash values too - they can be reused */
    _nc_scroll_oldhash(n, top, bot);

    return (OK);
}

static void
_nc_screen_resume(void)
{
    /* make sure terminal is in a sane known state */
    SP->_current_attr = A_NORMAL;
    newscr->_clear = TRUE;

    if (SP->_coloron == TRUE && orig_pair)
	putp(orig_pair);
    if (exit_attribute_mode)
	putp(exit_attribute_mode);
    else {
	/* turn off attributes */
	if (exit_alt_charset_mode)
	    putp(exit_alt_charset_mode);
	if (exit_standout_mode)
	    putp(exit_standout_mode);
	if (exit_underline_mode)
	    putp(exit_underline_mode);
    }
    if (exit_insert_mode)
	putp(exit_insert_mode);
    if (enter_am_mode && exit_am_mode)
	putp(auto_right_margin ? enter_am_mode : exit_am_mode);
}

static void
_nc_screen_init(void)
{
    _nc_screen_resume();
}

/* wrap up screen handling */
static void
_nc_screen_wrap(void)
{
    UpdateAttrs(A_NORMAL);
#if NCURSES_EXT_FUNCS
    if (SP->_coloron
	&& !SP->_default_color) {
	SP->_default_color = TRUE;
	_nc_do_color(-1, 0, FALSE, _nc_outch);
	SP->_default_color = FALSE;

	mvcur(SP->_cursrow, SP->_curscol, screen_lines - 1, 0);
	SP->_cursrow = screen_lines - 1;
	SP->_curscol = 0;

	ClrToEOL(BLANK, TRUE);
    }
#endif
}

/* generated by MKunctrl.awk */


#undef unctrl

const char *unctrl(chtype ch)
{
static const char* const table[] = {
    "^\100", "^\101", "^\102", "^\103", "^\104", "^\105", "^\106", "^\107",
    "^\110", "^\111", "^\112", "^\113", "^\114", "^\115", "^\116", "^\117",
    "^\120", "^\121", "^\122", "^\123", "^\124", "^\125", "^\126", "^\127",
    "^\130", "^\131", "^\132", "^\133", "^\134", "^\135", "^\136", "^\137",
    "\040",  "\041",  "\042",  "\043",  "\044",  "\045",  "\046",  "\047",
    "\050",  "\051",  "\052",  "\053",  "\054",  "\055",  "\056",  "\057",
    "\060",  "\061",  "\062",  "\063",  "\064",  "\065",  "\066",  "\067",
    "\070",  "\071",  "\072",  "\073",  "\074",  "\075",  "\076",  "\077",
    "\100",  "\101",  "\102",  "\103",  "\104",  "\105",  "\106",  "\107",
    "\110",  "\111",  "\112",  "\113",  "\114",  "\115",  "\116",  "\117",
    "\120",  "\121",  "\122",  "\123",  "\124",  "\125",  "\126",  "\127",
    "\130",  "\131",  "\132",  "\133",  "\134",  "\135",  "\136",  "\137",
    "\140",  "\141",  "\142",  "\143",  "\144",  "\145",  "\146",  "\147",
    "\150",  "\151",  "\152",  "\153",  "\154",  "\155",  "\156",  "\157",
    "\160",  "\161",  "\162",  "\163",  "\164",  "\165",  "\166",  "\167",
    "\170",  "\171",  "\172",  "\173",  "\174",  "\175",  "\176",  "^?",
    "~\100", "~\101", "~\102", "~\103", "~\104", "~\105", "~\106", "~\107",
    "~\110", "~\111", "~\112", "~\113", "~\114", "~\115", "~\116", "~\117",
    "~\120", "~\121", "~\122", "~\123", "~\124", "~\125", "~\126", "~\127",
    "~\130", "~\131", "~\132", "~\133", "~\134", "~\135", "~\136", "~\137",
    "\240",  "\241",  "\242",  "\243",  "\244",  "\245",  "\246",  "\247",
    "\250",  "\251",  "\252",  "\253",  "\254",  "\255",  "\256",  "\257",
    "\260",  "\261",  "\262",  "\263",  "\264",  "\265",  "\266",  "\267",
    "\270",  "\271",  "\272",  "\273",  "\274",  "\275",  "\276",  "\277",
    "\300",  "\301",  "\302",  "\303",  "\304",  "\305",  "\306",  "\307",
    "\310",  "\311",  "\312",  "\313",  "\314",  "\315",  "\316",  "\317",
    "\320",  "\321",  "\322",  "\323",  "\324",  "\325",  "\326",  "\327",
    "\330",  "\331",  "\332",  "\333",  "\334",  "\335",  "\336",  "\337",
    "\340",  "\341",  "\342",  "\343",  "\344",  "\345",  "\346",  "\347",
    "\350",  "\351",  "\352",  "\353",  "\354",  "\355",  "\356",  "\357",
    "\360",  "\361",  "\362",  "\363",  "\364",  "\365",  "\366",  "\367",
    "\370",  "\371",  "\372",  "\373",  "\374",  "\375",  "\376",  "\377"
};

	return (const char *)table[TextOf(ch)];
}

/*
 * Reallocate a curses WINDOW struct to either shrink or grow to the specified
 * new lines/columns.  If it grows, the new character cells are filled with
 * blanks.  The application is responsible for repainting the blank area.
 */

#define DOALLOC(p,t,n)  typeRealloc(t, n, p)
#define ld_ALLOC(p,n)   DOALLOC(p,struct ldat,n)
#define c_ALLOC(p,n)    DOALLOC(p,chtype,n)

static int
wresize(WINDOW *win, int ToLines, int ToCols)
{
    int row;
    int size_x, size_y;
    struct ldat *pline;
    chtype blank;

    if (!win || --ToLines < 0 || --ToCols < 0)
	returnCode(ERR);

    size_x = win->_maxx;
    size_y = win->_maxy;

    if (ToLines == size_y
	&& ToCols == size_x)
	returnCode(OK);

    if ((win->_flags & _SUBWIN)) {
	/*
	 * Check if the new limits will fit into the parent window's size.  If
	 * not, do not resize.  We could adjust the location of the subwindow,
	 * but the application may not like that.
	 */
	if (win->_pary + ToLines > win->_parent->_maxy
	    || win->_parx + ToCols > win->_parent->_maxx) {
	    returnCode(ERR);
	}
	pline = win->_parent->_line;
    } else {
	pline = 0;
    }

    /*
     * If the number of lines has changed, adjust the size of the overall
     * vector:
     */
    if (ToLines != size_y) {
	if (!(win->_flags & _SUBWIN)) {
	    for (row = ToLines + 1; row <= size_y; row++)
		free((char *) (win->_line[row].text));
	}

	win->_line = ld_ALLOC(win->_line, ToLines + 1);
	if (win->_line == 0)
	    returnCode(ERR);

	for (row = size_y + 1; row <= ToLines; row++) {
	    win->_line[row].text = 0;
	    win->_line[row].firstchar = 0;
	    win->_line[row].lastchar = ToCols;
	    if ((win->_flags & _SUBWIN)) {
		win->_line[row].text =
		    &pline[win->_pary + row].text[win->_parx];
	    }
	}
    }

    /*
     * Adjust the width of the columns:
     */
    blank = _nc_background(win);
    for (row = 0; row <= ToLines; row++) {
	chtype *s = win->_line[row].text;
	int begin = (s == 0) ? 0 : size_x + 1;
	int end = ToCols;

	if_USE_SCROLL_HINTS(win->_line[row].oldindex = row);

	if (ToCols != size_x || s == 0) {
	    if (!(win->_flags & _SUBWIN)) {
		win->_line[row].text = s = c_ALLOC(s, ToCols + 1);
		if (win->_line[row].text == 0)
		    returnCode(ERR);
	    } else if (s == 0) {
		win->_line[row].text = s =
		    &pline[win->_pary + row].text[win->_parx];
	    }

	    if (end >= begin) { /* growing */
		if (win->_line[row].firstchar < begin)
		    win->_line[row].firstchar = begin;
		win->_line[row].lastchar = ToCols;
		do {
		    s[end] = blank;
		} while (--end >= begin);
	    } else {            /* shrinking */
		win->_line[row].firstchar = 0;
		win->_line[row].lastchar = ToCols;
	    }
	}
    }

    /*
     * Finally, adjust the parameters showing screen size and cursor
     * position:
     */
    win->_maxx = ToCols;
    win->_maxy = ToLines;

    if (win->_regtop > win->_maxy)
	win->_regtop = win->_maxy;
    if (win->_regbottom > win->_maxy
	|| win->_regbottom == size_y)
	win->_regbottom = win->_maxy;

    if (win->_curx > win->_maxx)
	win->_curx = win->_maxx;
    if (win->_cury > win->_maxy)
	win->_cury = win->_maxy;

    returnCode(OK);
}

#define LKC_DIRECT_LINK
#include "lkc.h"


#define ESC 27
#define TAB 9
#define MAX_LEN 2048
#define BUF_SIZE (10*1024)
#define MIN(x,y) (x < y ? x : y)
#define MAX(x,y) (x > y ? x : y)


#ifndef ACS_ULCORNER
#define ACS_ULCORNER '+'
#endif
#ifndef ACS_LLCORNER
#define ACS_LLCORNER '+'
#endif
#ifndef ACS_URCORNER
#define ACS_URCORNER '+'
#endif
#ifndef ACS_LRCORNER
#define ACS_LRCORNER '+'
#endif
#ifndef ACS_HLINE
#define ACS_HLINE '-'
#endif
#ifndef ACS_VLINE
#define ACS_VLINE '|'
#endif
#ifndef ACS_LTEE
#define ACS_LTEE '+'
#endif
#ifndef ACS_RTEE
#define ACS_RTEE '+'
#endif
#ifndef ACS_UARROW
#define ACS_UARROW '^'
#endif
#ifndef ACS_DARROW
#define ACS_DARROW 'v'
#endif

/*
 * Attribute names
 */
#define screen_attr                   attributes[0]
#define shadow_attr                   attributes[1]
#define dialog_attr                   attributes[2]
#define title_attr                    attributes[3]
#define border_attr                   attributes[4]
#define button_active_attr            attributes[5]
#define button_inactive_attr          attributes[6]
#define button_key_active_attr        attributes[7]
#define button_key_inactive_attr      attributes[8]
#define button_label_active_attr      attributes[9]
#define button_label_inactive_attr    attributes[10]
#define inputbox_attr                 attributes[11]
#define inputbox_border_attr          attributes[12]
#define searchbox_attr                attributes[13]
#define searchbox_title_attr          attributes[14]
#define searchbox_border_attr         attributes[15]
#define position_indicator_attr       attributes[16]
#define menubox_attr                  attributes[17]
#define menubox_border_attr           attributes[18]
#define item_attr                     attributes[19]
#define item_selected_attr            attributes[20]
#define tag_attr                      attributes[21]
#define tag_selected_attr             attributes[22]
#define tag_key_attr                  attributes[23]
#define tag_key_selected_attr         attributes[24]
#define check_attr                    attributes[25]
#define check_selected_attr           attributes[26]
#define uarrow_attr                   attributes[27]
#define darrow_attr                   attributes[28]

/* number of attributes */
#define ATTRIBUTE_COUNT               29


struct dialog_list_item {
	char *name;
	int namelen;
	char *tag;
	int selected; /* Set to 1 by dialog_*() function. */
};

/*
 * This is the base for fictitious keys, which activate
 * the buttons.
 *
 * Mouse-generated keys are the following:
 *   -- the first 32 are used as numbers, in addition to '0'-'9'
 *   -- the lowercase are used to signal mouse-enter events (M_EVENT + 'o')
 *   -- uppercase chars are used to invoke the button (M_EVENT + 'O')
 */
#define M_EVENT (KEY_MAX+1)


/*
 * The `flag' parameter in checklist is used to select between
 * radiolist and checklist
 */
#define FLAG_CHECK 1
#define FLAG_RADIO 0

/*
 *  colors.h -- color attribute definitions
 */


/*
 *   Default color definitions
 *
 *   *_FG = foreground
 *   *_BG = background
 *   *_HL = highlight?
 */
#define SCREEN_FG                    COLOR_CYAN
#define SCREEN_BG                    COLOR_BLUE
#define SCREEN_HL                    TRUE

#define SHADOW_FG                    COLOR_BLACK
#define SHADOW_BG                    COLOR_BLACK
#define SHADOW_HL                    TRUE

#define DIALOG_FG                    COLOR_BLACK
#define DIALOG_BG                    COLOR_WHITE
#define DIALOG_HL                    FALSE

#define TITLE_FG                     COLOR_YELLOW
#define TITLE_BG                     COLOR_WHITE
#define TITLE_HL                     TRUE

#define BORDER_FG                    COLOR_WHITE
#define BORDER_BG                    COLOR_WHITE
#define BORDER_HL                    TRUE

#define BUTTON_ACTIVE_FG             COLOR_WHITE
#define BUTTON_ACTIVE_BG             COLOR_BLUE
#define BUTTON_ACTIVE_HL             TRUE

#define BUTTON_INACTIVE_FG           COLOR_BLACK
#define BUTTON_INACTIVE_BG           COLOR_WHITE
#define BUTTON_INACTIVE_HL           FALSE

#define BUTTON_KEY_ACTIVE_FG         COLOR_WHITE
#define BUTTON_KEY_ACTIVE_BG         COLOR_BLUE
#define BUTTON_KEY_ACTIVE_HL         TRUE

#define BUTTON_KEY_INACTIVE_FG       COLOR_RED
#define BUTTON_KEY_INACTIVE_BG       COLOR_WHITE
#define BUTTON_KEY_INACTIVE_HL       FALSE

#define BUTTON_LABEL_ACTIVE_FG       COLOR_YELLOW
#define BUTTON_LABEL_ACTIVE_BG       COLOR_BLUE
#define BUTTON_LABEL_ACTIVE_HL       TRUE

#define BUTTON_LABEL_INACTIVE_FG     COLOR_BLACK
#define BUTTON_LABEL_INACTIVE_BG     COLOR_WHITE
#define BUTTON_LABEL_INACTIVE_HL     TRUE

#define INPUTBOX_FG                  COLOR_BLACK
#define INPUTBOX_BG                  COLOR_WHITE
#define INPUTBOX_HL                  FALSE

#define INPUTBOX_BORDER_FG           COLOR_BLACK
#define INPUTBOX_BORDER_BG           COLOR_WHITE
#define INPUTBOX_BORDER_HL           FALSE

#define SEARCHBOX_FG                 COLOR_BLACK
#define SEARCHBOX_BG                 COLOR_WHITE
#define SEARCHBOX_HL                 FALSE

#define SEARCHBOX_TITLE_FG           COLOR_YELLOW
#define SEARCHBOX_TITLE_BG           COLOR_WHITE
#define SEARCHBOX_TITLE_HL           TRUE

#define SEARCHBOX_BORDER_FG          COLOR_WHITE
#define SEARCHBOX_BORDER_BG          COLOR_WHITE
#define SEARCHBOX_BORDER_HL          TRUE

#define POSITION_INDICATOR_FG        COLOR_YELLOW
#define POSITION_INDICATOR_BG        COLOR_WHITE
#define POSITION_INDICATOR_HL        TRUE

#define MENUBOX_FG                   COLOR_BLACK
#define MENUBOX_BG                   COLOR_WHITE
#define MENUBOX_HL                   FALSE

#define MENUBOX_BORDER_FG            COLOR_WHITE
#define MENUBOX_BORDER_BG            COLOR_WHITE
#define MENUBOX_BORDER_HL            TRUE

#define ITEM_FG                      COLOR_BLACK
#define ITEM_BG                      COLOR_WHITE
#define ITEM_HL                      FALSE

#define ITEM_SELECTED_FG             COLOR_WHITE
#define ITEM_SELECTED_BG             COLOR_BLUE
#define ITEM_SELECTED_HL             TRUE

#define TAG_FG                       COLOR_YELLOW
#define TAG_BG                       COLOR_WHITE
#define TAG_HL                       TRUE

#define TAG_SELECTED_FG              COLOR_YELLOW
#define TAG_SELECTED_BG              COLOR_BLUE
#define TAG_SELECTED_HL              TRUE

#define TAG_KEY_FG                   COLOR_YELLOW
#define TAG_KEY_BG                   COLOR_WHITE
#define TAG_KEY_HL                   TRUE

#define TAG_KEY_SELECTED_FG          COLOR_YELLOW
#define TAG_KEY_SELECTED_BG          COLOR_BLUE
#define TAG_KEY_SELECTED_HL          TRUE

#define CHECK_FG                     COLOR_BLACK
#define CHECK_BG                     COLOR_WHITE
#define CHECK_HL                     FALSE

#define CHECK_SELECTED_FG            COLOR_WHITE
#define CHECK_SELECTED_BG            COLOR_BLUE
#define CHECK_SELECTED_HL            TRUE

#define UARROW_FG                    COLOR_GREEN
#define UARROW_BG                    COLOR_WHITE
#define UARROW_HL                    TRUE

#define DARROW_FG                    COLOR_GREEN
#define DARROW_BG                    COLOR_WHITE
#define DARROW_HL                    TRUE

/* End of default color definitions */

#define C_ATTR(x,y)                  ((x ? A_BOLD : 0) | COLOR_PAIR((y)))
#define COLOR_NAME_LEN               10
#define COLOR_COUNT                  8

/*
 *  util.c
 */

/*
 * Attribute values, default is for mono display
 */
static chtype attributes[] =
{
    A_NORMAL,                   /* screen_attr */
    A_NORMAL,                   /* shadow_attr */
    A_NORMAL,                   /* dialog_attr */
    A_BOLD,                     /* title_attr */
    A_NORMAL,                   /* border_attr */
    A_REVERSE,                  /* button_active_attr */
    A_DIM,                      /* button_inactive_attr */
    A_REVERSE,                  /* button_key_active_attr */
    A_BOLD,                     /* button_key_inactive_attr */
    A_REVERSE,                  /* button_label_active_attr */
    A_NORMAL,                   /* button_label_inactive_attr */
    A_NORMAL,                   /* inputbox_attr */
    A_NORMAL,                   /* inputbox_border_attr */
    A_NORMAL,                   /* searchbox_attr */
    A_BOLD,                     /* searchbox_title_attr */
    A_NORMAL,                   /* searchbox_border_attr */
    A_BOLD,                     /* position_indicator_attr */
    A_NORMAL,                   /* menubox_attr */
    A_NORMAL,                   /* menubox_border_attr */
    A_NORMAL,                   /* item_attr */
    A_REVERSE,                  /* item_selected_attr */
    A_BOLD,                     /* tag_attr */
    A_REVERSE,                  /* tag_selected_attr */
    A_BOLD,                     /* tag_key_attr */
    A_REVERSE,                  /* tag_key_selected_attr */
    A_BOLD,                     /* check_attr */
    A_REVERSE,                  /* check_selected_attr */
    A_BOLD,                     /* uarrow_attr */
    A_BOLD                      /* darrow_attr */
};


/*
 * Table of color values
 */
static int color_table[][3] =
{
    {SCREEN_FG, SCREEN_BG, SCREEN_HL},
    {SHADOW_FG, SHADOW_BG, SHADOW_HL},
    {DIALOG_FG, DIALOG_BG, DIALOG_HL},
    {TITLE_FG, TITLE_BG, TITLE_HL},
    {BORDER_FG, BORDER_BG, BORDER_HL},
    {BUTTON_ACTIVE_FG, BUTTON_ACTIVE_BG, BUTTON_ACTIVE_HL},
    {BUTTON_INACTIVE_FG, BUTTON_INACTIVE_BG, BUTTON_INACTIVE_HL},
    {BUTTON_KEY_ACTIVE_FG, BUTTON_KEY_ACTIVE_BG, BUTTON_KEY_ACTIVE_HL},
    {BUTTON_KEY_INACTIVE_FG, BUTTON_KEY_INACTIVE_BG, BUTTON_KEY_INACTIVE_HL},
    {BUTTON_LABEL_ACTIVE_FG, BUTTON_LABEL_ACTIVE_BG, BUTTON_LABEL_ACTIVE_HL},
    {BUTTON_LABEL_INACTIVE_FG, BUTTON_LABEL_INACTIVE_BG,
     BUTTON_LABEL_INACTIVE_HL},
    {INPUTBOX_FG, INPUTBOX_BG, INPUTBOX_HL},
    {INPUTBOX_BORDER_FG, INPUTBOX_BORDER_BG, INPUTBOX_BORDER_HL},
    {SEARCHBOX_FG, SEARCHBOX_BG, SEARCHBOX_HL},
    {SEARCHBOX_TITLE_FG, SEARCHBOX_TITLE_BG, SEARCHBOX_TITLE_HL},
    {SEARCHBOX_BORDER_FG, SEARCHBOX_BORDER_BG, SEARCHBOX_BORDER_HL},
    {POSITION_INDICATOR_FG, POSITION_INDICATOR_BG, POSITION_INDICATOR_HL},
    {MENUBOX_FG, MENUBOX_BG, MENUBOX_HL},
    {MENUBOX_BORDER_FG, MENUBOX_BORDER_BG, MENUBOX_BORDER_HL},
    {ITEM_FG, ITEM_BG, ITEM_HL},
    {ITEM_SELECTED_FG, ITEM_SELECTED_BG, ITEM_SELECTED_HL},
    {TAG_FG, TAG_BG, TAG_HL},
    {TAG_SELECTED_FG, TAG_SELECTED_BG, TAG_SELECTED_HL},
    {TAG_KEY_FG, TAG_KEY_BG, TAG_KEY_HL},
    {TAG_KEY_SELECTED_FG, TAG_KEY_SELECTED_BG, TAG_KEY_SELECTED_HL},
    {CHECK_FG, CHECK_BG, CHECK_HL},
    {CHECK_SELECTED_FG, CHECK_SELECTED_BG, CHECK_SELECTED_HL},
    {UARROW_FG, UARROW_BG, UARROW_HL},
    {DARROW_FG, DARROW_BG, DARROW_HL},
};                              /* color_table */

/*
 * Set window to attribute 'attr'
 */
static void
attr_clear (WINDOW * win, int height, int width, chtype attr)
{
    int i, j;

    wattrset (win, attr);
    for (i = 0; i < height; i++) {
	wmove (win, i, 0);
	for (j = 0; j < width; j++)
	    waddch (win, ' ');
    }
    touchwin (win);
}

static void dialog_clear (void)
{
    attr_clear (stdscr, LINES, COLS, screen_attr);
    wnoutrefresh (stdscr);
}

/*
 * Do some initialization for dialog
 */
static void
init_dialog (void)
{
    int i;

    initscr ();                 /* Init curses */
    keypad (stdscr, TRUE);
    cbreak ();
    noecho ();

   /*
    * Setup for color display
    */
    if (has_colors ()) {        /* Terminal supports color? */
	start_color ();

	/* Initialize color pairs */
	for (i = 0; i < ATTRIBUTE_COUNT; i++)
	    init_pair (i + 1, color_table[i][0], color_table[i][1]);

	/* Setup color attributes */
	for (i = 0; i < ATTRIBUTE_COUNT; i++)
	    attributes[i] = C_ATTR (color_table[i][2], i + 1);
    }

    dialog_clear ();
}

/*
 * End using dialog functions.
 */
static inline void
end_dialog (void)
{
    endwin ();
}


/*
 * Print a string of text in a window, automatically wrap around to the
 * next line if the string is too long to fit on one line. Newline
 * characters '\n' are replaced by spaces.  We start on a new line
 * if there is no room for at least 4 nonblanks following a double-space.
 */
static void
print_autowrap (WINDOW * win, const char *prompt, int width, int y, int x)
{
    int newl, cur_x, cur_y;
    int i, prompt_len, room, wlen;
    char tempstr[MAX_LEN + 1], *word, *sp, *sp2;

    strcpy (tempstr, prompt);

    prompt_len = strlen(tempstr);

    /*
     * Remove newlines
     */
    for(i=0; i<prompt_len; i++) {
	if(tempstr[i] == '\n') tempstr[i] = ' ';
    }

    if (prompt_len <= width - x * 2) {  /* If prompt is short */
	wmove (win, y, (width - prompt_len) / 2);
	waddstr (win, tempstr);
    } else {
	cur_x = x;
	cur_y = y;
	newl = 1;
	word = tempstr;
	while (word && *word) {
	    sp = index(word, ' ');
	    if (sp)
		*sp++ = 0;

	    /* Wrap to next line if either the word does not fit,
	       or it is the first word of a new sentence, and it is
	       short, and the next word does not fit. */
	    room = width - cur_x;
	    wlen = strlen(word);
	    if (wlen > room ||
	       (newl && wlen < 4 && sp && wlen+1+(int)strlen(sp) > room
		     && (!(sp2 = index(sp, ' ')) || wlen+1+(sp2-sp) > room))) {
		cur_y++;
		cur_x = x;
	    }
	    wmove (win, cur_y, cur_x);
	    waddstr (win, word);
	    getyx (win, cur_y, cur_x);
	    cur_x++;
	    if (sp && *sp == ' ') {
		cur_x++;        /* double space */
		while (*++sp == ' ');
		newl = 1;
	    } else
		newl = 0;
	    word = sp;
	}
    }
}

/*
 * Print a button
 */
static void
print_button (WINDOW * win, const char *label, int y, int x, int selected)
{
    int i, temp;

    wmove (win, y, x);
    wattrset (win, selected ? button_active_attr : button_inactive_attr);
    waddstr (win, "<");
    temp = strspn (label, " ");
    label += temp;
    wattrset (win, selected ? button_label_active_attr
	      : button_label_inactive_attr);
    for (i = 0; i < temp; i++)
	waddch (win, ' ');
    wattrset (win, selected ? button_key_active_attr
	      : button_key_inactive_attr);
    waddch (win, label[0]);
    wattrset (win, selected ? button_label_active_attr
	      : button_label_inactive_attr);
    waddstr (win, (char *)label + 1);
    wattrset (win, selected ? button_active_attr : button_inactive_attr);
    waddstr (win, ">");
    wmove (win, y, x + temp + 1);
}

/*
 * Draw a rectangular box with line drawing characters
 */
static void
draw_box (WINDOW * win, int y, int x, int height, int width,
	  chtype box, chtype border)
{
    int i, j;

    wattrset (win, 0);
    for (i = 0; i < height; i++) {
	wmove (win, y + i, x);
	for (j = 0; j < width; j++)
	    if (!i && !j)
		waddch (win, border | ACS_ULCORNER);
	    else if (i == height - 1 && !j)
		waddch (win, border | ACS_LLCORNER);
	    else if (!i && j == width - 1)
		waddch (win, box | ACS_URCORNER);
	    else if (i == height - 1 && j == width - 1)
		waddch (win, box | ACS_LRCORNER);
	    else if (!i)
		waddch (win, border | ACS_HLINE);
	    else if (i == height - 1)
		waddch (win, box | ACS_HLINE);
	    else if (!j)
		waddch (win, border | ACS_VLINE);
	    else if (j == width - 1)
		waddch (win, box | ACS_VLINE);
	    else
		waddch (win, box | ' ');
    }
}

/*
 * Draw shadows along the right and bottom edge to give a more 3D look
 * to the boxes
 */
static void
draw_shadow (WINDOW * win, int y, int x, int height, int width)
{
    int i;

    if (has_colors ()) {        /* Whether terminal supports color? */
	wattrset (win, shadow_attr);
	wmove (win, y + height, x + 2);
	for (i = 0; i < width; i++)
	    waddch (win, winch (win) & A_CHARTEXT);
	for (i = y + 1; i < y + height + 1; i++) {
	    wmove (win, i, x + width);
	    waddch (win, winch (win) & A_CHARTEXT);
	    waddch (win, winch (win) & A_CHARTEXT);
	}
	wnoutrefresh (win);
    }
}

/*
 *  Return the position of the first alphabetic character in a string.
 */
static int
first_alpha(const char *string, const char *exempt)
{
	int i, in_paren=0, c;

	for (i = 0; i < (int)strlen(string); i++) {
		c = tolower(string[i]);

		if (strchr("<[(", c)) ++in_paren;
		if (strchr(">])", c) && in_paren > 0) --in_paren;

		if ((! in_paren) && isalpha(c) &&
		     strchr(exempt, c) == 0)
			return i;
	}

	return 0;
}

/*
 * Get the first selected item in the dialog_list_item list.
 */
static struct dialog_list_item *
first_sel_item(int item_no, struct dialog_list_item ** items)
{
	int i;

	for (i = 0; i < item_no; i++) {
		if (items[i]->selected)
			return items[i];
	}

	return NULL;
}

/*
 *  yesno.c -- implements the yes/no box
 */

/*
 * Display termination buttons
 */
static void
print_buttons_yes_no(WINDOW *dialog, int height, int width, int selected)
{
    int x = width / 2 - 10;
    int y = height - 2;

    print_button (dialog, " Yes ", y, x, selected == 0);
    print_button (dialog, "  No  ", y, x + 13, selected == 1);

    wmove(dialog, y, x+1 + 13*selected );
    wrefresh (dialog);
}

/*
 * Display a dialog box with two buttons - Yes and No
 */
static int
dialog_yesno (const char *title, const char *prompt, int height, int width)
{
    int i, x, y, key = 0, button = 0;
    WINDOW *dialog;

    /* center dialog box on screen */
    x = (COLS - width) / 2;
    y = (LINES - height) / 2;

    draw_shadow (stdscr, y, x, height, width);

    dialog = newwin (height, width, y, x);
    keypad (dialog, TRUE);

    draw_box (dialog, 0, 0, height, width, dialog_attr, border_attr);
    wattrset (dialog, border_attr);
    mvwaddch (dialog, height-3, 0, ACS_LTEE);
    for (i = 0; i < width - 2; i++)
	waddch (dialog, ACS_HLINE);
    wattrset (dialog, dialog_attr);
    waddch (dialog, ACS_RTEE);

    if (title != NULL && (int)strlen(title) >= width-2 ) {
	/* truncate long title -- mec */
	char * title2 = malloc(width-2+1);
	memcpy( title2, title, width-2 );
	title2[width-2] = '\0';
	title = title2;
    }

    if (title != NULL) {
	wattrset (dialog, title_attr);
	mvwaddch (dialog, 0, (width - strlen(title))/2 - 1, ' ');
	waddstr (dialog, (char *)title);
	waddch (dialog, ' ');
    }

    wattrset (dialog, dialog_attr);
    print_autowrap (dialog, prompt, width - 2, 1, 3);

    print_buttons_yes_no(dialog, height, width, 0);

    while (key != ESC) {
	key = wgetch (dialog);
	switch (key) {
	case 'Y':
	case 'y':
	    delwin (dialog);
	    return 0;
	case 'N':
	case 'n':
	    delwin (dialog);
	    return 1;

	case TAB:
	case KEY_LEFT:
	case KEY_RIGHT:
	    button = ((key == KEY_LEFT ? --button : ++button) < 0)
			? 1 : (button > 1 ? 0 : button);

	    print_buttons_yes_no(dialog, height, width, button);
	    wrefresh (dialog);
	    break;
	case ' ':
	case '\n':
	    delwin (dialog);
	    return button;
	case ESC:
	    break;
	}
    }

    delwin (dialog);
    return -1;                  /* ESC pressed */
}

/*
 *  checklist.c -- implements the checklist box
 */

static int list_width, check_x, item_x_chl, checkflag;

/*
 * Print list item
 */
static void
print_item_checklist (WINDOW * win, const char *item, int status,
	    int choice, int selected)
{
    int i;

    /* Clear 'residue' of last item */
    wattrset (win, menubox_attr);
    wmove (win, choice, 0);
    for (i = 0; i < list_width; i++)
	waddch (win, ' ');

    wmove (win, choice, check_x);
    wattrset (win, selected ? check_selected_attr : check_attr);
    if (checkflag == FLAG_CHECK)
	wprintw (win, "[%c]", status ? 'X' : ' ');
    else
	wprintw (win, "(%c)", status ? 'X' : ' ');

    wattrset (win, selected ? tag_selected_attr : tag_attr);
    mvwaddch(win, choice, item_x_chl, item[0]);
    wattrset (win, selected ? item_selected_attr : item_attr);
    waddstr (win, (char *)item+1);
    if (selected) {
	wmove (win, choice, check_x+1);
	wrefresh (win);
    }
}

/*
 * Print the scroll indicators.
 */
static void
print_arrows_checklist (WINDOW * win, int choice, int item_no, int scroll,
		int y, int x, int height)
{
    wmove(win, y, x);

    if (scroll > 0) {
	wattrset (win, uarrow_attr);
	waddch (win, ACS_UARROW);
	waddstr (win, "(-)");
    }
    else {
	wattrset (win, menubox_attr);
	waddch (win, ACS_HLINE);
	waddch (win, ACS_HLINE);
	waddch (win, ACS_HLINE);
	waddch (win, ACS_HLINE);
    }

   y = y + height + 1;
   wmove(win, y, x);

   if ((height < item_no) && (scroll + choice < item_no - 1)) {
	wattrset (win, darrow_attr);
	waddch (win, ACS_DARROW);
	waddstr (win, "(+)");
    }
    else {
	wattrset (win, menubox_border_attr);
	waddch (win, ACS_HLINE);
	waddch (win, ACS_HLINE);
	waddch (win, ACS_HLINE);
	waddch (win, ACS_HLINE);
   }
}

/*
 *  Display the termination buttons
 */
static void
print_buttons_checklist( WINDOW *dialog, int height, int width, int selected)
{
    int x = width / 2 - 11;
    int y = height - 2;

    print_button (dialog, "Select", y, x, selected == 0);
    print_button (dialog, " Help ", y, x + 14, selected == 1);

    wmove(dialog, y, x+1 + 14*selected);
    wrefresh (dialog);
}

/*
 * Display a dialog box with a list of options that can be turned on or off
 * The `flag' parameter is used to select between radiolist and checklist.
 */
static int
dialog_checklist (const char *title, const char *prompt, int height, int width,
	int list_height, int item_no, struct dialog_list_item ** items,
	int flag)

{
    int i, x, y, box_x, box_y;
    int key = 0, button = 0, choice = 0, scroll = 0, max_choice, *status;
    WINDOW *dialog, *list;

    checkflag = flag;

    /* Allocate space for storing item on/off status */
    if ((status = malloc (sizeof (int) * item_no)) == NULL) {
	endwin ();
	fprintf (stderr,
		 "\nCan't allocate memory in dialog_checklist().\n");
	exit (-1);
    }

    /* Initializes status */
    for (i = 0; i < item_no; i++) {
	status[i] = (items[i]->selected == 1); /* ON */
	if ((!choice && status[i]) || items[i]->selected == 2) /* SELECTED */
	    choice = i + 1;
    }
    if (choice)
	    choice--;

    max_choice = MIN (list_height, item_no);

    /* center dialog box on screen */
    x = (COLS - width) / 2;
    y = (LINES - height) / 2;

    draw_shadow (stdscr, y, x, height, width);

    dialog = newwin (height, width, y, x);
    keypad (dialog, TRUE);

    draw_box (dialog, 0, 0, height, width, dialog_attr, border_attr);
    wattrset (dialog, border_attr);
    mvwaddch (dialog, height-3, 0, ACS_LTEE);
    for (i = 0; i < width - 2; i++)
	waddch (dialog, ACS_HLINE);
    wattrset (dialog, dialog_attr);
    waddch (dialog, ACS_RTEE);

    if (title != NULL && (int)strlen(title) >= width-2 ) {
	/* truncate long title -- mec */
	char * title2 = malloc(width-2+1);
	memcpy( title2, title, width-2 );
	title2[width-2] = '\0';
	title = title2;
    }

    if (title != NULL) {
	wattrset (dialog, title_attr);
	mvwaddch (dialog, 0, (width - strlen(title))/2 - 1, ' ');
	waddstr (dialog, (char *)title);
	waddch (dialog, ' ');
    }

    wattrset (dialog, dialog_attr);
    print_autowrap (dialog, prompt, width - 2, 1, 3);

    list_width = width - 6;
    box_y = height - list_height - 5;
    box_x = (width - list_width) / 2 - 1;

    /* create new window for the list */
    list = subwin (dialog, list_height, list_width, y+box_y+1, x+box_x+1);

    keypad (list, TRUE);

    /* draw a box around the list items */
    draw_box (dialog, box_y, box_x, list_height + 2, list_width + 2,
	      menubox_border_attr, menubox_attr);

    /* Find length of longest item in order to center checklist */
    check_x = 0;
    for (i = 0; i < item_no; i++)
	check_x = MAX (check_x, (int)strlen (items[i]->name) + 4);

    check_x = (list_width - check_x) / 2;
    item_x_chl = check_x + 4;

    if (choice >= list_height) {
	scroll = choice - list_height + 1;
	choice -= scroll;
    }

    /* Print the list */
    for (i = 0; i < max_choice; i++) {
	print_item_checklist (list, items[scroll + i]->name,
		    status[i+scroll], i, i == choice);
    }

    print_arrows_checklist(dialog, choice, item_no, scroll,
			box_y, box_x + check_x + 5, list_height);

    print_buttons_checklist(dialog, height, width, 0);

    wnoutrefresh (list);
    wnoutrefresh (dialog);
    doupdate ();

    while (key != ESC) {
	key = wgetch (dialog);

	for (i = 0; i < max_choice; i++)
	    if (toupper(key) == toupper(items[scroll + i]->name[0]))
		break;


	if ( i < max_choice || key == KEY_UP || key == KEY_DOWN ||
	    key == '+' || key == '-' ) {
	    if (key == KEY_UP || key == '-') {
		if (!choice) {
		    if (!scroll)
			continue;
		    /* Scroll list down */
		    if (list_height > 1) {
			/* De-highlight current first item */
			print_item_checklist (list, items[scroll]->name,
					status[scroll], 0, FALSE);
			scrollok (list, TRUE);
			wscrl (list, -1);
			scrollok (list, FALSE);
		    }
		    scroll--;
		    print_item_checklist (list, items[scroll]->name,
				status[scroll], 0, TRUE);
		    wnoutrefresh (list);

		    print_arrows_checklist(dialog, choice, item_no, scroll,
				box_y, box_x + check_x + 5, list_height);

		    wrefresh (dialog);

		    continue;   /* wait for another key press */
		} else
		    i = choice - 1;
	    } else if (key == KEY_DOWN || key == '+') {
		if (choice == max_choice - 1) {
		    if (scroll + choice >= item_no - 1)
			continue;
		    /* Scroll list up */
		    if (list_height > 1) {
			/* De-highlight current last item before scrolling up */
			print_item_checklist (list,
				    items[scroll + max_choice - 1]->name,
				    status[scroll + max_choice - 1],
				    max_choice - 1, FALSE);
			scrollok (list, TRUE);
			scroll (list);
			scrollok (list, FALSE);
		    }
		    scroll++;
		    print_item_checklist (list,
				items[scroll + max_choice - 1]->name,
				status[scroll + max_choice - 1],
				max_choice - 1, TRUE);
		    wnoutrefresh (list);

		    print_arrows_checklist(dialog, choice, item_no, scroll,
				box_y, box_x + check_x + 5, list_height);

		    wrefresh (dialog);

		    continue;   /* wait for another key press */
		} else
		    i = choice + 1;
	    }
	    if (i != choice) {
		/* De-highlight current item */
		print_item_checklist (list, items[scroll + choice]->name,
			    status[scroll + choice], choice, FALSE);
		/* Highlight new item */
		choice = i;
		print_item_checklist (list, items[scroll + choice]->name,
			    status[scroll + choice], choice, TRUE);
		wnoutrefresh (list);
		wrefresh (dialog);
	    }
	    continue;           /* wait for another key press */
	}
	switch (key) {
	case 'H':
	case 'h':
	case '?':
	    for (i = 0; i < item_no; i++)
		items[i]->selected = 0;
	    items[scroll + choice]->selected = 1;
	    delwin (dialog);
	    free (status);
	    return 1;
	case TAB:
	case KEY_LEFT:
	case KEY_RIGHT:
	    button = ((key == KEY_LEFT ? --button : ++button) < 0)
			? 1 : (button > 1 ? 0 : button);

	    print_buttons_checklist(dialog, height, width, button);
	    wrefresh (dialog);
	    break;
	case 'S':
	case 's':
	case ' ':
	case '\n':
	    if (!button) {
		if (flag == FLAG_CHECK) {
		    status[scroll + choice] = !status[scroll + choice];
		    wmove (list, choice, check_x);
		    wattrset (list, check_selected_attr);
		    wprintw (list, "[%c]", status[scroll + choice] ? 'X' : ' ');
		} else {
		    if (!status[scroll + choice]) {
			for (i = 0; i < item_no; i++)
			    status[i] = 0;
			status[scroll + choice] = 1;
			for (i = 0; i < max_choice; i++)
			    print_item_checklist (list,
					items[scroll + i]->name,
					status[scroll + i], i, i == choice);
		    }
		}
		wnoutrefresh (list);
		wrefresh (dialog);

		for (i = 0; i < item_no; i++) {
			items[i]->selected = status[i];
		}
	    } else {
		    for (i = 0; i < item_no; i++)
			    items[i]->selected = 0;
		    items[scroll + choice]->selected = 1;
	    }
	    delwin (dialog);
	    free (status);
	    return button;
	case 'X':
	case 'x':
	    key = ESC;
	case ESC:
	    break;
	}

	/* Now, update everything... */
	doupdate ();
    }


    delwin (dialog);
    free (status);
    return -1;                  /* ESC pressed */
}

/*
 *  inputbox.c -- implements the input box
 */

static unsigned char dialog_input_result[MAX_LEN + 1];

/*
 *  Print the termination buttons
 */
static void
print_buttons_inputbox(WINDOW *dialog, int height, int width, int selected)
{
    int x = width / 2 - 11;
    int y = height - 2;

    print_button (dialog, "  Ok  ", y, x, selected==0);
    print_button (dialog, " Help ", y, x + 14, selected==1);

    wmove(dialog, y, x+1+14*selected);
    wrefresh(dialog);
}

/*
 * Display a dialog box for inputing a string
 */
static int
dialog_inputbox (const char *title, const char *prompt, int height, int width,
		 const char *init)
{
    int i, x, y, box_y, box_x, box_width;
    int input_x = 0, scroll = 0, key = 0, button = -1;
    unsigned char *instr = dialog_input_result;
    WINDOW *dialog;

    /* center dialog box on screen */
    x = (COLS - width) / 2;
    y = (LINES - height) / 2;


    draw_shadow (stdscr, y, x, height, width);

    dialog = newwin (height, width, y, x);
    keypad (dialog, TRUE);

    draw_box (dialog, 0, 0, height, width, dialog_attr, border_attr);
    wattrset (dialog, border_attr);
    mvwaddch (dialog, height-3, 0, ACS_LTEE);
    for (i = 0; i < width - 2; i++)
	waddch (dialog, ACS_HLINE);
    wattrset (dialog, dialog_attr);
    waddch (dialog, ACS_RTEE);

    if (title != NULL && (int)strlen(title) >= width-2 ) {
	/* truncate long title -- mec */
	char * title2 = malloc(width-2+1);
	memcpy( title2, title, width-2 );
	title2[width-2] = '\0';
	title = title2;
    }

    if (title != NULL) {
	wattrset (dialog, title_attr);
	mvwaddch (dialog, 0, (width - strlen(title))/2 - 1, ' ');
	waddstr (dialog, (char *)title);
	waddch (dialog, ' ');
    }

    wattrset (dialog, dialog_attr);
    print_autowrap (dialog, prompt, width - 2, 1, 3);

    /* Draw the input field box */
    box_width = width - 6;
    getyx (dialog, y, x);
    box_y = y + 2;
    box_x = (width - box_width) / 2;
    draw_box (dialog, y + 1, box_x - 1, 3, box_width + 2,
	      border_attr, dialog_attr);

    print_buttons_inputbox(dialog, height, width, 0);

    /* Set up the initial value */
    wmove (dialog, box_y, box_x);
    wattrset (dialog, inputbox_attr);

    if (!init)
	instr[0] = '\0';
    else
	strcpy ((char *)instr, init);

    input_x = strlen ((char *)instr);

    if (input_x >= box_width) {
	scroll = input_x - box_width + 1;
	input_x = box_width - 1;
	for (i = 0; i < box_width - 1; i++)
	    waddch (dialog, instr[scroll + i]);
    } else
	waddstr (dialog, (const char *)instr);

    wmove (dialog, box_y, box_x + input_x);

    wrefresh (dialog);

    while (key != ESC) {
	key = wgetch (dialog);

	if (button == -1) {     /* Input box selected */
	    switch (key) {
	    case TAB:
	    case KEY_UP:
	    case KEY_DOWN:
		break;
	    case KEY_LEFT:
		continue;
	    case KEY_RIGHT:
		continue;
	    case KEY_BACKSPACE:
	    case 127:
		if (input_x || scroll) {
		    wattrset (dialog, inputbox_attr);
		    if (!input_x) {
			scroll = scroll < box_width - 1 ?
			    0 : scroll - (box_width - 1);
			wmove (dialog, box_y, box_x);
			for (i = 0; i < box_width; i++)
			    waddch (dialog, instr[scroll + input_x + i] ?
				    instr[scroll + input_x + i] : ' ');
			input_x = strlen ((char *)instr) - scroll;
		    } else
			input_x--;
		    instr[scroll + input_x] = '\0';
		    mvwaddch (dialog, box_y, input_x + box_x, ' ');
		    wmove (dialog, box_y, input_x + box_x);
		    wrefresh (dialog);
		}
		continue;
	    default:
		if (key < 0x100 && isprint (key)) {
		    if (scroll + input_x < MAX_LEN) {
			wattrset (dialog, inputbox_attr);
			instr[scroll + input_x] = key;
			instr[scroll + input_x + 1] = '\0';
			if (input_x == box_width - 1) {
			    scroll++;
			    wmove (dialog, box_y, box_x);
			    for (i = 0; i < box_width - 1; i++)
				waddch (dialog, instr[scroll + i]);
			} else {
			    wmove (dialog, box_y, input_x++ + box_x);
			    waddch (dialog, key);
			}
			wrefresh (dialog);
		    } else
			flash ();       /* Alarm user about overflow */
		    continue;
		}
	    }
	}
	switch (key) {
	case 'O':
	case 'o':
	    delwin (dialog);
	    return 0;
	case 'H':
	case 'h':
	    delwin (dialog);
	    return 1;
	case KEY_UP:
	case KEY_LEFT:
	    switch (button) {
	    case -1:
		button = 1;     /* Indicates "Cancel" button is selected */
		print_buttons_inputbox(dialog, height, width, 1);
		break;
	    case 0:
		button = -1;    /* Indicates input box is selected */
		print_buttons_inputbox(dialog, height, width, 0);
		wmove (dialog, box_y, box_x + input_x);
		wrefresh (dialog);
		break;
	    case 1:
		button = 0;     /* Indicates "OK" button is selected */
		print_buttons_inputbox(dialog, height, width, 0);
		break;
	    }
	    break;
	case TAB:
	case KEY_DOWN:
	case KEY_RIGHT:
	    switch (button) {
	    case -1:
		button = 0;     /* Indicates "OK" button is selected */
		print_buttons_inputbox(dialog, height, width, 0);
		break;
	    case 0:
		button = 1;     /* Indicates "Cancel" button is selected */
		print_buttons_inputbox(dialog, height, width, 1);
		break;
	    case 1:
		button = -1;    /* Indicates input box is selected */
		print_buttons_inputbox(dialog, height, width, 0);
		wmove (dialog, box_y, box_x + input_x);
		wrefresh (dialog);
		break;
	    }
	    break;
	case ' ':
	case '\n':
	    delwin (dialog);
	    return (button == -1 ? 0 : button);
	case 'X':
	case 'x':
	    key = ESC;
	case ESC:
	    break;
	}
    }

    delwin (dialog);
    return -1;                  /* ESC pressed */
}

/*
 *  menubox.c -- implements the menu box
 */

static int menu_width, item_x;

/*
 * Print menu item
 */
static void
print_item_menubox (WINDOW * win, const char *item, int choice, int selected, int hotkey)
{
    int j;
    char menu_item[menu_width+1];

    strncpy(menu_item, item, menu_width);
    menu_item[menu_width] = 0;
    j = first_alpha(menu_item, "YyNnMmHh");

    /* Clear 'residue' of last item */
    wattrset (win, menubox_attr);
    wmove (win, choice, 0);
    wclrtoeol(win);
    wattrset (win, selected ? item_selected_attr : item_attr);
    mvwaddstr (win, choice, item_x, menu_item);
    if (hotkey) {
	wattrset (win, selected ? tag_key_selected_attr : tag_key_attr);
	mvwaddch(win, choice, item_x+j, menu_item[j]);
    }
    if (selected) {
	wmove (win, choice, item_x+1);
	wrefresh (win);
    }
}

/*
 * Print the scroll indicators.
 */
static void
print_arrows_menubox (WINDOW * win, int item_no, int scroll,
		int y, int x, int height)
{
    int cur_y, cur_x;

    getyx(win, cur_y, cur_x);

    wmove(win, y, x);

    if (scroll > 0) {
	wattrset (win, uarrow_attr);
	waddch (win, ACS_UARROW);
	waddstr (win, "(-)");
    }
    else {
	wattrset (win, menubox_attr);
	waddch (win, ACS_HLINE);
	waddch (win, ACS_HLINE);
	waddch (win, ACS_HLINE);
	waddch (win, ACS_HLINE);
    }

   y = y + height + 1;
   wmove(win, y, x);

   if ((height < item_no) && (scroll + height < item_no)) {
	wattrset (win, darrow_attr);
	waddch (win, ACS_DARROW);
	waddstr (win, "(+)");
    }
    else {
	wattrset (win, menubox_border_attr);
	waddch (win, ACS_HLINE);
	waddch (win, ACS_HLINE);
	waddch (win, ACS_HLINE);
	waddch (win, ACS_HLINE);
   }

   wmove(win, cur_y, cur_x);
}

/*
 * Display the termination buttons.
 */
static void
print_buttons_menubox (WINDOW *win, int height, int width, int selected)
{
    int x = width / 2 - 16;
    int y = height - 2;

    print_button (win, "Select", y, x, selected == 0);
    print_button (win, " Exit ", y, x + 12, selected == 1);
    print_button (win, " Help ", y, x + 24, selected == 2);

    wmove(win, y, x+1+12*selected);
    wrefresh (win);
}

/*
 * Display a menu for choosing among a number of options
 */
static int
dialog_menu (const char *title, const char *prompt, int height, int width,
		int menu_height, const char *current, int item_no,
		struct dialog_list_item ** items)
{
    int i, j, x, y, box_x, box_y;
    int key = 0, button = 0, scroll = 0, choice = 0, first_item = 0, max_choice;
    WINDOW *dialog, *menu;
    FILE *f;

    max_choice = MIN (menu_height, item_no);

    /* center dialog box on screen */
    x = (COLS - width) / 2;
    y = (LINES - height) / 2;

    draw_shadow (stdscr, y, x, height, width);

    dialog = newwin (height, width, y, x);
    keypad (dialog, TRUE);

    draw_box (dialog, 0, 0, height, width, dialog_attr, border_attr);
    wattrset (dialog, border_attr);
    mvwaddch (dialog, height - 3, 0, ACS_LTEE);
    for (i = 0; i < width - 2; i++)
	waddch (dialog, ACS_HLINE);
    wattrset (dialog, dialog_attr);
    wbkgdset (dialog, dialog_attr & A_COLOR);
    waddch (dialog, ACS_RTEE);

    if (title != NULL && (int)strlen(title) >= width-2 ) {
	/* truncate long title -- mec */
	char * title2 = malloc(width-2+1);
	memcpy( title2, title, width-2 );
	title2[width-2] = '\0';
	title = title2;
    }

    if (title != NULL) {
	wattrset (dialog, title_attr);
	mvwaddch (dialog, 0, (width - strlen(title))/2 - 1, ' ');
	waddstr (dialog, (char *)title);
	waddch (dialog, ' ');
    }

    wattrset (dialog, dialog_attr);
    print_autowrap (dialog, prompt, width - 2, 1, 3);

    menu_width = width - 6;
    box_y = height - menu_height - 5;
    box_x = (width - menu_width) / 2 - 1;

    /* create new window for the menu */
    menu = subwin (dialog, menu_height, menu_width,
		y + box_y + 1, x + box_x + 1);
    keypad (menu, TRUE);

    /* draw a box around the menu items */
    draw_box (dialog, box_y, box_x, menu_height + 2, menu_width + 2,
	      menubox_border_attr, menubox_attr);

    /*
     * Find length of longest item in order to center menu.
     * Set 'choice' to default item.
     */
    item_x = 0;
    for (i = 0; i < item_no; i++) {
	item_x = MAX (item_x, MIN(menu_width, (int)strlen (items[i]->name) + 2));
	if (strcmp(current, items[i]->tag) == 0) choice = i;
    }

    item_x = (menu_width - item_x) / 2;

    /* get the scroll info from the temp file */
    if ( (f=fopen("lxdialog.scrltmp","r")) != NULL ) {
	if ( (fscanf(f,"%d\n",&scroll) == 1) && (scroll <= choice) &&
	     (scroll+max_choice > choice) && (scroll >= 0) &&
	     (scroll+max_choice <= item_no) ) {
	    first_item = scroll;
	    choice = choice - scroll;
	    fclose(f);
	} else {
	    scroll=0;
	    remove("lxdialog.scrltmp");
	    fclose(f);
	    f=NULL;
	}
    }
    if ( (choice >= max_choice) || (f==NULL && choice >= max_choice/2) ) {
	if (choice >= item_no-max_choice/2)
	    scroll = first_item = item_no-max_choice;
	else
	    scroll = first_item = choice - max_choice/2;
	choice = choice - scroll;
    }

    /* Print the menu */
    for (i=0; i < max_choice; i++) {
	print_item_menubox (menu, items[first_item + i]->name, i, i == choice,
		    (items[first_item + i]->tag[0] != ':'));
    }

    wnoutrefresh (menu);

    print_arrows_menubox(dialog, item_no, scroll,
		 box_y, box_x+item_x+1, menu_height);

    print_buttons_menubox (dialog, height, width, 0);
    wmove (menu, choice, item_x+1);
    wrefresh (menu);

    while (key != ESC) {
	key = wgetch(menu);

	if (key < 256 && isalpha(key)) key = tolower(key);

	if (strchr("ynmh", key))
		i = max_choice;
	else {
	for (i = choice+1; i < max_choice; i++) {
		j = first_alpha(items[scroll + i]->name, "YyNnMmHh");
		if (key == tolower(items[scroll + i]->name[j]))
			break;
	}
	if (i == max_choice)
		for (i = 0; i < max_choice; i++) {
			j = first_alpha(items[scroll + i]->name, "YyNnMmHh");
			if (key == tolower(items[scroll + i]->name[j]))
				break;
		}
	}

	if (i < max_choice ||
	    key == KEY_UP || key == KEY_DOWN ||
	    key == '-' || key == '+' ||
	    key == KEY_PPAGE || key == KEY_NPAGE) {

	    print_item_menubox (menu, items[scroll + choice]->name, choice, FALSE,
		       (items[scroll + choice]->tag[0] != ':'));

	    if (key == KEY_UP || key == '-') {
		if (choice < 2 && scroll) {
		    /* Scroll menu down */
		    scrollok (menu, TRUE);
		    wscrl (menu, -1);
		    scrollok (menu, FALSE);

		    scroll--;

		    print_item_menubox (menu, items[scroll]->name, 0, FALSE,
			       (items[scroll]->tag[0] != ':'));
		} else
		    choice = MAX(choice - 1, 0);

	    } else if (key == KEY_DOWN || key == '+')  {

		print_item_menubox (menu, items[scroll + choice]->name, choice, FALSE,
				(items[scroll + choice]->tag[0] != ':'));

		if ((choice > max_choice-3) &&
		    (scroll + max_choice < item_no)
		   ) {
		    /* Scroll menu up */
		    scrollok (menu, TRUE);
		    scroll (menu);
		    scrollok (menu, FALSE);

		    scroll++;

		    print_item_menubox (menu, items[scroll + max_choice - 1]->name,
			       max_choice-1, FALSE,
			       (items[scroll + max_choice - 1]->tag[0] != ':'));
		} else
		    choice = MIN(choice+1, max_choice-1);

	    } else if (key == KEY_PPAGE) {
		scrollok (menu, TRUE);
		for (i=0; (i < max_choice); i++) {
		    if (scroll > 0) {
			wscrl (menu, -1);
			scroll--;
			print_item_menubox (menu, items[scroll]->name, 0, FALSE,
			(items[scroll]->tag[0] != ':'));
		    } else {
			if (choice > 0)
			    choice--;
		    }
		}
		scrollok (menu, FALSE);

	    } else if (key == KEY_NPAGE) {
		for (i=0; (i < max_choice); i++) {
		    if (scroll+max_choice < item_no) {
			scrollok (menu, TRUE);
			scroll(menu);
			scrollok (menu, FALSE);
			scroll++;
			print_item_menubox (menu, items[scroll + max_choice - 1]->name,
				    max_choice-1, FALSE,
				    (items[scroll + max_choice - 1]->tag[0] != ':'));
		    } else {
			if (choice+1 < max_choice)
			    choice++;
		    }
		}

	    } else
		choice = i;

	    print_item_menubox (menu, items[scroll + choice]->name, choice, TRUE,
		       (items[scroll + choice]->tag[0] != ':'));

	    print_arrows_menubox(dialog, item_no, scroll,
			 box_y, box_x+item_x+1, menu_height);

	    wnoutrefresh (dialog);
	    wrefresh (menu);

	    continue;           /* wait for another key press */
	}

	switch (key) {
	case KEY_LEFT:
	case TAB:
	case KEY_RIGHT:
	    button = ((key == KEY_LEFT ? --button : ++button) < 0)
			? 2 : (button > 2 ? 0 : button);

	    print_buttons_menubox(dialog, height, width, button);
	    wrefresh (menu);
	    break;
	case ' ':
	case 's':
	case 'y':
	case 'n':
	case 'm':
	case '/':
	    /* save scroll info */
	    if ( (f=fopen("lxdialog.scrltmp","w")) != NULL ) {
		fprintf(f,"%d\n",scroll);
		fclose(f);
	    }
	    delwin (dialog);
	    items[scroll + choice]->selected = 1;
	    switch (key) {
	    case 's': return 3;
	    case 'y': return 3;
	    case 'n': return 4;
	    case 'm': return 5;
	    case ' ': return 6;
	    case '/': return 7;
	    }
	    return 0;
	case 'h':
	case '?':
	    button = 2;
	case '\n':
	    delwin (dialog);
	    items[scroll + choice]->selected = 1;

	    remove("lxdialog.scrltmp");
	    return button;
	case 'e':
	case 'x':
	    key = ESC;
	case ESC:
	    break;
	}
    }

    delwin (dialog);
    remove("lxdialog.scrltmp");
    return -1;                  /* ESC pressed */
}

/*
 *  textbox.c -- implements the text box
 */

static int hscroll, fd, file_size, bytes_read;
static int begin_reached = 1, end_reached, page_length;
static char *buf, *page;


/*
 * Go back 'n' lines in text file. Called by dialog_textbox().
 * 'page' will be updated to point to the desired line in 'buf'.
 */
static void
back_lines (int n)
{
    int i, fpos;

    begin_reached = 0;
    /* We have to distinguish between end_reached and !end_reached
       since at end of file, the line is not ended by a '\n'.
       The code inside 'if' basically does a '--page' to move one
       character backward so as to skip '\n' of the previous line */
    if (!end_reached) {
	/* Either beginning of buffer or beginning of file reached? */
	if (page == buf) {
	    if ((fpos = lseek (fd, 0, SEEK_CUR)) == -1) {
		endwin ();
		fprintf (stderr, "\nError moving file pointer in "
			 "back_lines().\n");
		exit (-1);
	    }
	    if (fpos > bytes_read) {    /* Not beginning of file yet */
		/* We've reached beginning of buffer, but not beginning of
		   file yet, so read previous part of file into buffer.
		   Note that we only move backward for BUF_SIZE/2 bytes,
		   but not BUF_SIZE bytes to avoid re-reading again in
		   print_page() later */
		/* Really possible to move backward BUF_SIZE/2 bytes? */
		if (fpos < BUF_SIZE / 2 + bytes_read) {
		    /* No, move less then */
		    if (lseek (fd, 0, SEEK_SET) == -1) {
			endwin ();
			fprintf (stderr, "\nError moving file pointer in "
				 "back_lines().\n");
			exit (-1);
		    }
		    page = buf + fpos - bytes_read;
		} else {        /* Move backward BUF_SIZE/2 bytes */
		    if (lseek (fd, -(BUF_SIZE / 2 + bytes_read), SEEK_CUR)
			== -1) {
			endwin ();
			fprintf (stderr, "\nError moving file pointer "
				 "in back_lines().\n");
			exit (-1);
		    }
		    page = buf + BUF_SIZE / 2;
		}
		if ((bytes_read = read (fd, buf, BUF_SIZE)) == -1) {
		    endwin ();
		    fprintf (stderr, "\nError reading file in back_lines().\n");
		    exit (-1);
		}
		buf[bytes_read] = '\0';
	    } else {            /* Beginning of file reached */
		begin_reached = 1;
		return;
	    }
	}
	if (*(--page) != '\n') {        /* '--page' here */
	    /* Something's wrong... */
	    endwin ();
	    fprintf (stderr, "\nInternal error in back_lines().\n");
	    exit (-1);
	}
    }
    /* Go back 'n' lines */
    for (i = 0; i < n; i++)
	do {
	    if (page == buf) {
		if ((fpos = lseek (fd, 0, SEEK_CUR)) == -1) {
		    endwin ();
		    fprintf (stderr,
			  "\nError moving file pointer in back_lines().\n");
		    exit (-1);
		}
		if (fpos > bytes_read) {
		    /* Really possible to move backward BUF_SIZE/2 bytes? */
		    if (fpos < BUF_SIZE / 2 + bytes_read) {
			/* No, move less then */
			if (lseek (fd, 0, SEEK_SET) == -1) {
			    endwin ();
			    fprintf (stderr, "\nError moving file pointer "
				     "in back_lines().\n");
			    exit (-1);
			}
			page = buf + fpos - bytes_read;
		    } else {    /* Move backward BUF_SIZE/2 bytes */
			if (lseek (fd, -(BUF_SIZE / 2 + bytes_read),
				   SEEK_CUR) == -1) {
			    endwin ();
			    fprintf (stderr, "\nError moving file pointer"
				     " in back_lines().\n");
			    exit (-1);
			}
			page = buf + BUF_SIZE / 2;
		    }
		    if ((bytes_read = read (fd, buf, BUF_SIZE)) == -1) {
			endwin ();
			fprintf (stderr, "\nError reading file in "
				 "back_lines().\n");
			exit (-1);
		    }
		    buf[bytes_read] = '\0';
		} else {        /* Beginning of file reached */
		    begin_reached = 1;
		    return;
		}
	    }
	} while (*(--page) != '\n');
    page++;
}

/*
 * Return current line of text. Called by dialog_textbox() and print_line().
 * 'page' should point to start of current line before calling, and will be
 * updated to point to start of next line.
 */
static char *
get_line (void)
{
    int i = 0, fpos;
    static char line[MAX_LEN + 1];

    end_reached = 0;
    while (*page != '\n') {
	if (*page == '\0') {
	    /* Either end of file or end of buffer reached */
	    if ((fpos = lseek (fd, 0, SEEK_CUR)) == -1) {
		endwin ();
		fprintf (stderr, "\nError moving file pointer in "
			 "get_line().\n");
		exit (-1);
	    }
	    if (fpos < file_size) {     /* Not end of file yet */
		/* We've reached end of buffer, but not end of file yet,
		   so read next part of file into buffer */
		if ((bytes_read = read (fd, buf, BUF_SIZE)) == -1) {
		    endwin ();
		    fprintf (stderr, "\nError reading file in get_line().\n");
		    exit (-1);
		}
		buf[bytes_read] = '\0';
		page = buf;
	    } else {
		if (!end_reached)
		    end_reached = 1;
		break;
	    }
	} else if (i < MAX_LEN)
	    line[i++] = *(page++);
	else {
	    /* Truncate lines longer than MAX_LEN characters */
	    if (i == MAX_LEN)
		line[i++] = '\0';
	    page++;
	}
    }
    if (i <= MAX_LEN)
	line[i] = '\0';
    if (!end_reached)
	page++;                 /* move pass '\n' */

    return line;
}

/*
 * Print a new line of text. Called by dialog_textbox() and print_page().
 */
static void
print_line (WINDOW * win, int row, int width)
{
    char *line;

    line = get_line ();
    line += MIN ((int)strlen (line), hscroll);       /* Scroll horizontally */
    wmove (win, row, 0);        /* move cursor to correct line */
    waddch (win, ' ');
    waddnstr (win, line, MIN ((int)strlen (line), width - 2));

    /* Clear 'residue' of previous line */
    wclrtoeol(win);
}

/*
 * Print a new page of text. Called by dialog_textbox().
 */
static void
print_page (WINDOW * win, int height, int width)
{
    int i, passed_end = 0;

    page_length = 0;
    for (i = 0; i < height; i++) {
	print_line (win, i, width);
	if (!passed_end)
	    page_length++;
	if (end_reached && !passed_end)
	    passed_end = 1;
    }
    wnoutrefresh (win);
}


/*
 * Print current position
 */
static void
print_position (WINDOW * win, int height, int width)
{
    int fpos, percent;

    if ((fpos = lseek (fd, 0, SEEK_CUR)) == -1) {
	endwin ();
	fprintf (stderr, "\nError moving file pointer in print_position().\n");
	exit (-1);
    }
    wattrset (win, position_indicator_attr);
    wbkgdset (win, position_indicator_attr & A_COLOR);
    percent = !file_size ?
	100 : ((fpos - bytes_read + page - buf) * 100) / file_size;
    wmove (win, height - 3, width - 9);
    wprintw (win, "(%3d%%)", percent);
}

/*
 * Display text from a file in a dialog box.
 */
static int
dialog_textbox (const char *title, const char *file, int height, int width)
{
    int i, x, y, cur_x, cur_y, fpos, key = 0;
    int passed_end;
    WINDOW *dialog, *text;

    /* Open input file for reading */
    if ((fd = open (file, O_RDONLY)) == -1) {
	endwin ();
	fprintf (stderr, "\nCan't open input file in dialog_textbox().\n");
	exit (-1);
    }
    /* Get file size. Actually, 'file_size' is the real file size - 1,
       since it's only the last byte offset from the beginning */
    if ((file_size = lseek (fd, 0, SEEK_END)) == -1) {
	endwin ();
	fprintf (stderr, "\nError getting file size in dialog_textbox().\n");
	exit (-1);
    }
    /* Restore file pointer to beginning of file after getting file size */
    if (lseek (fd, 0, SEEK_SET) == -1) {
	endwin ();
	fprintf (stderr, "\nError moving file pointer in dialog_textbox().\n");
	exit (-1);
    }
    /* Allocate space for read buffer */
    if ((buf = malloc (BUF_SIZE + 1)) == NULL) {
	endwin ();
	fprintf (stderr, "\nCan't allocate memory in dialog_textbox().\n");
	exit (-1);
    }
    if ((bytes_read = read (fd, buf, BUF_SIZE)) == -1) {
	endwin ();
	fprintf (stderr, "\nError reading file in dialog_textbox().\n");
	exit (-1);
    }
    buf[bytes_read] = '\0';     /* mark end of valid data */
    page = buf;                 /* page is pointer to start of page to be displayed */

    /* center dialog box on screen */
    x = (COLS - width) / 2;
    y = (LINES - height) / 2;


    draw_shadow (stdscr, y, x, height, width);

    dialog = newwin (height, width, y, x);
    keypad (dialog, TRUE);

    /* Create window for text region, used for scrolling text */
    text = subwin (dialog, height - 4, width - 2, y + 1, x + 1);
    wattrset (text, dialog_attr);
    wbkgdset (text, dialog_attr & A_COLOR);

    keypad (text, TRUE);

    /* register the new window, along with its borders */
    draw_box (dialog, 0, 0, height, width, dialog_attr, border_attr);

    wattrset (dialog, border_attr);
    mvwaddch (dialog, height-3, 0, ACS_LTEE);
    for (i = 0; i < width - 2; i++)
	waddch (dialog, ACS_HLINE);
    wattrset (dialog, dialog_attr);
    wbkgdset (dialog, dialog_attr & A_COLOR);
    waddch (dialog, ACS_RTEE);

    if (title != NULL && (int)strlen(title) >= width-2 ) {
	/* truncate long title -- mec */
	char * title2 = malloc(width-2+1);
	memcpy( title2, title, width-2 );
	title2[width-2] = '\0';
	title = title2;
    }

    if (title != NULL) {
	wattrset (dialog, title_attr);
	mvwaddch (dialog, 0, (width - strlen(title))/2 - 1, ' ');
	waddstr (dialog, (char *)title);
	waddch (dialog, ' ');
    }
    print_button (dialog, " Exit ", height - 2, width / 2 - 4, TRUE);
    wnoutrefresh (dialog);
    getyx (dialog, cur_y, cur_x);       /* Save cursor position */

    /* Print first page of text */
    attr_clear (text, height - 4, width - 2, dialog_attr);
    print_page (text, height - 4, width - 2);
    print_position (dialog, height, width);
    wmove (dialog, cur_y, cur_x);       /* Restore cursor position */
    wrefresh (dialog);

    while ((key != ESC) && (key != '\n')) {
	key = wgetch (dialog);
	switch (key) {
	case 'E':               /* Exit */
	case 'e':
	case 'X':
	case 'x':
	    delwin (dialog);
	    free (buf);
	    close (fd);
	    return 0;
	case 'g':               /* First page */
	case KEY_HOME:
	    if (!begin_reached) {
		begin_reached = 1;
		/* First page not in buffer? */
		if ((fpos = lseek (fd, 0, SEEK_CUR)) == -1) {
		    endwin ();
		    fprintf (stderr,
		      "\nError moving file pointer in dialog_textbox().\n");
		    exit (-1);
		}
		if (fpos > bytes_read) {        /* Yes, we have to read it in */
		    if (lseek (fd, 0, SEEK_SET) == -1) {
			endwin ();
			fprintf (stderr, "\nError moving file pointer in "
				 "dialog_textbox().\n");
			exit (-1);
		    }
		    if ((bytes_read = read (fd, buf, BUF_SIZE)) == -1) {
			endwin ();
			fprintf (stderr,
			     "\nError reading file in dialog_textbox().\n");
			exit (-1);
		    }
		    buf[bytes_read] = '\0';
		}
		page = buf;
		print_page (text, height - 4, width - 2);
		print_position (dialog, height, width);
		wmove (dialog, cur_y, cur_x);   /* Restore cursor position */
		wrefresh (dialog);
	    }
	    break;
	case 'G':               /* Last page */
	case KEY_END:

	    end_reached = 1;
	    /* Last page not in buffer? */
	    if ((fpos = lseek (fd, 0, SEEK_CUR)) == -1) {
		endwin ();
		fprintf (stderr,
		      "\nError moving file pointer in dialog_textbox().\n");
		exit (-1);
	    }
	    if (fpos < file_size) {     /* Yes, we have to read it in */
		if (lseek (fd, -BUF_SIZE, SEEK_END) == -1) {
		    endwin ();
		    fprintf (stderr,
		      "\nError moving file pointer in dialog_textbox().\n");
		    exit (-1);
		}
		if ((bytes_read = read (fd, buf, BUF_SIZE)) == -1) {
		    endwin ();
		    fprintf (stderr,
			     "\nError reading file in dialog_textbox().\n");
		    exit (-1);
		}
		buf[bytes_read] = '\0';
	    }
	    page = buf + bytes_read;
	    back_lines (height - 4);
	    print_page (text, height - 4, width - 2);
	    print_position (dialog, height, width);
	    wmove (dialog, cur_y, cur_x);       /* Restore cursor position */
	    wrefresh (dialog);
	    break;
	case 'K':               /* Previous line */
	case 'k':
	case KEY_UP:
	    if (!begin_reached) {
		back_lines (page_length + 1);

		/* We don't call print_page() here but use scrolling to ensure
		   faster screen update. However, 'end_reached' and
		   'page_length' should still be updated, and 'page' should
		   point to start of next page. This is done by calling
		   get_line() in the following 'for' loop. */
		scrollok (text, TRUE);
		wscrl (text, -1);       /* Scroll text region down one line */
		scrollok (text, FALSE);
		page_length = 0;
		passed_end = 0;
		for (i = 0; i < height - 4; i++) {
		    if (!i) {
			/* print first line of page */
			print_line (text, 0, width - 2);
			wnoutrefresh (text);
		    } else
			/* Called to update 'end_reached' and 'page' */
			get_line ();
		    if (!passed_end)
			page_length++;
		    if (end_reached && !passed_end)
			passed_end = 1;
		}

		print_position (dialog, height, width);
		wmove (dialog, cur_y, cur_x);   /* Restore cursor position */
		wrefresh (dialog);
	    }
	    break;
	case 'B':               /* Previous page */
	case 'b':
	case KEY_PPAGE:
	    if (begin_reached)
		break;
	    back_lines (page_length + height - 4);
	    print_page (text, height - 4, width - 2);
	    print_position (dialog, height, width);
	    wmove (dialog, cur_y, cur_x);
	    wrefresh (dialog);
	    break;
	case 'J':               /* Next line */
	case 'j':
	case KEY_DOWN:
	    if (!end_reached) {
		begin_reached = 0;
		scrollok (text, TRUE);
		scroll (text);  /* Scroll text region up one line */
		scrollok (text, FALSE);
		print_line (text, height - 5, width - 2);
		wnoutrefresh (text);
		print_position (dialog, height, width);
		wmove (dialog, cur_y, cur_x);   /* Restore cursor position */
		wrefresh (dialog);
	    }
	    break;
	case KEY_NPAGE:         /* Next page */
	case ' ':
	    if (end_reached)
		break;

	    begin_reached = 0;
	    print_page (text, height - 4, width - 2);
	    print_position (dialog, height, width);
	    wmove (dialog, cur_y, cur_x);
	    wrefresh (dialog);
	    break;
	case '0':               /* Beginning of line */
	case 'H':               /* Scroll left */
	case 'h':
	case KEY_LEFT:
	    if (hscroll <= 0)
		break;

	    if (key == '0')
		hscroll = 0;
	    else
		hscroll--;
	    /* Reprint current page to scroll horizontally */
	    back_lines (page_length);
	    print_page (text, height - 4, width - 2);
	    wmove (dialog, cur_y, cur_x);
	    wrefresh (dialog);
	    break;
	case 'L':               /* Scroll right */
	case 'l':
	case KEY_RIGHT:
	    if (hscroll >= MAX_LEN)
		break;
	    hscroll++;
	    /* Reprint current page to scroll horizontally */
	    back_lines (page_length);
	    print_page (text, height - 4, width - 2);
	    wmove (dialog, cur_y, cur_x);
	    wrefresh (dialog);
	    break;
	case ESC:
	    break;
	}
    }

    delwin (dialog);
    free (buf);
    close (fd);
    return 1;                   /* ESC pressed */
}

static char menu_backtitle[128];
static const char mconf_readme[] =
"Overview\n"
"--------\n"
"Some features may be built directly into BusyBox.  Some features\n"
"may be completely removed altogether.  There are also certain\n"
"parameters which are not really features, but must be\n"
"entered in as decimal or hexadecimal numbers or possibly text.\n"
"\n"
"Menu items beginning with [*] or [ ] represent features\n"
"configured to be built in or removed respectively.\n"
"\n"
"To change any of these features, highlight it with the cursor\n"
"keys and press <Y> to build it in or <N> to removed it.\n"
"You may also press the <Space Bar> to cycle\n"
"through the available options (ie. Y->N->Y).\n"
"\n"
"Some additional keyboard hints:\n"
"\n"
"Menus\n"
"----------\n"
"o  Use the Up/Down arrow keys (cursor keys) to highlight the item\n"
"   you wish to change or submenu wish to select and press <Enter>.\n"
"   Submenus are designated by \"--->\".\n"
"\n"
"   Shortcut: Press the option's highlighted letter (hotkey).\n"
"             Pressing a hotkey more than once will sequence\n"
"             through all visible items which use that hotkey.\n"
"\n"
"   You may also use the <PAGE UP> and <PAGE DOWN> keys to scroll\n"
"   unseen options into view.\n"
"\n"
"o  To exit a menu use the cursor keys to highlight the <Exit> button\n"
"   and press <ENTER>.\n"
"\n"
"   Shortcut: Press <ESC><ESC> or <E> or <X> if there is no hotkey\n"
"             using those letters.  You may press a single <ESC>, but\n"
"             there is a delayed response which you may find annoying.\n"
"\n"
"   Also, the <TAB> and cursor keys will cycle between <Select>,\n"
"   <Exit> and <Help>\n"
"\n"
"o  To get help with an item, use the cursor keys to highlight <Help>\n"
"   and Press <ENTER>.\n"
"\n"
"   Shortcut: Press <H> or <?>.\n"
"\n"
"\n"
"Radiolists  (Choice lists)\n"
"-----------\n"
"o  Use the cursor keys to select the option you wish to set and press\n"
"   <S> or the <SPACE BAR>.\n"
"\n"
"   Shortcut: Press the first letter of the option you wish to set then\n"
"             press <S> or <SPACE BAR>.\n"
"\n"
"o  To see available help for the item, use the cursor keys to highlight\n"
"   <Help> and Press <ENTER>.\n"
"\n"
"   Shortcut: Press <H> or <?>.\n"
"\n"
"   Also, the <TAB> and cursor keys will cycle between <Select> and\n"
"   <Help>\n"
"\n"
"\n"
"Data Entry\n"
"-----------\n"
"o  Enter the requested information and press <ENTER>\n"
"   If you are entering hexadecimal values, it is not necessary to\n"
"   add the '0x' prefix to the entry.\n"
"\n"
"o  For help, use the <TAB> or cursor keys to highlight the help option\n"
"   and press <ENTER>.  You can try <TAB><H> as well.\n"
"\n"
"\n"
"Text Box    (Help Window)\n"
"--------\n"
"o  Use the cursor keys to scroll up/down/left/right.  The VI editor\n"
"   keys h,j,k,l function here as do <SPACE BAR> and <B> for those\n"
"   who are familiar with less and lynx.\n"
"\n"
"o  Press <E>, <X>, <Enter> or <Esc><Esc> to exit.\n"
"\n"
"\n"
"Alternate Configuration Files\n"
"-----------------------------\n"
"Menuconfig supports the use of alternate configuration files for\n"
"those who, for various reasons, find it necessary to switch\n"
"between different configurations.\n"
"\n"
"At the end of the main menu you will find two options.  One is\n"
"for saving the current configuration to a file of your choosing.\n"
"The other option is for loading a previously saved alternate\n"
"configuration.\n"
"\n"
"Even if you don't use alternate configuration files, but you\n"
"find during a Menuconfig session that you have completely messed\n"
"up your settings, you may use the \"Load Alternate...\" option to\n"
"restore your previously saved settings from \".config\" without\n"
"restarting Menuconfig.\n"
"\n"
"Other information\n"
"-----------------\n"
"If you use Menuconfig in an XTERM window make sure you have your\n"
"$TERM variable set to point to a xterm definition which supports color.\n"
"Otherwise, Menuconfig will look rather bad.  Menuconfig will not\n"
"display correctly in a RXVT window because rxvt displays only one\n"
"intensity of color, bright.\n"
"\n"
"Menuconfig will display larger menus on screens or xterms which are\n"
"set to display more than the standard 25 row by 80 column geometry.\n"
"In order for this to work, the \"stty size\" command must be able to\n"
"display the screen's current row and column geometry.  I STRONGLY\n"
"RECOMMEND that you make sure you do NOT have the shell variables\n"
"LINES and COLUMNS exported into your environment.  Some distributions\n"
"export those variables via /etc/profile.  Some ncurses programs can\n"
"become confused when those variables (LINES & COLUMNS) don't reflect\n"
"the true screen size.\n"
"\n"
"Optional personality available\n"
"------------------------------\n"
"If you prefer to have all of the options listed in a single\n"
"menu, rather than the default multimenu hierarchy, run the menuconfig\n"
"with MENUCONFIG_MODE environment variable set to single_menu. Example:\n"
"\n"
"make MENUCONFIG_MODE=single_menu menuconfig\n"
"\n"
"<Enter> will then unroll the appropriate category, or enfold it if it\n"
"is already unrolled.\n"
"\n"
"Note that this mode can eventually be a little more CPU expensive\n"
"(especially with a larger number of unrolled categories) than the\n"
"default mode.\n",
menu_instructions[] =
	"Arrow keys navigate the menu.  "
	"<Enter> selects submenus --->.  "
	"Highlighted letters are hotkeys.  "
	"Pressing <Y> selectes a feature, while <N> will exclude a feature.  "
	"Press <Esc><Esc> to exit, <?> for Help, </> for Search.  "
	"Legend: [*] feature is selected  [ ] feature is excluded",
radiolist_instructions[] =
	"Use the arrow keys to navigate this window or "
	"press the hotkey of the item you wish to select "
	"followed by the <SPACE BAR>. "
	"Press <?> for additional information about this option.",
inputbox_instructions_int[] =
	"Please enter a decimal value. "
	"Fractions will not be accepted.  "
	"Use the <TAB> key to move from the input field to the buttons below it.",
inputbox_instructions_hex[] =
	"Please enter a hexadecimal value. "
	"Use the <TAB> key to move from the input field to the buttons below it.",
inputbox_instructions_string[] =
	"Please enter a string value. "
	"Use the <TAB> key to move from the input field to the buttons below it.",
setmod_text[] =
	"This feature depends on another which has been configured as a module.\n"
	"As a result, this feature will be built as a module.",
nohelp_text[] =
	"There is no help available for this option.\n",
load_config_text[] =
	"Enter the name of the configuration file you wish to load.  "
	"Accept the name shown to restore the configuration you "
	"last retrieved.  Leave blank to abort.",
load_config_help[] =
	"\n"
	"For various reasons, one may wish to keep several different BusyBox\n"
	"configurations available on a single machine.\n"
	"\n"
	"If you have saved a previous configuration in a file other than the\n"
	"BusyBox's default, entering the name of the file here will allow you\n"
	"to modify that configuration.\n"
	"\n"
	"If you are uncertain, then you have probably never used alternate\n"
	"configuration files.  You should therefor leave this blank to abort.\n",
save_config_text[] =
	"Enter a filename to which this configuration should be saved "
	"as an alternate.  Leave blank to abort.",
save_config_help[] =
	"\n"
	"For various reasons, one may wish to keep different BusyBox\n"
	"configurations available on a single machine.\n"
	"\n"
	"Entering a file name here will allow you to later retrieve, modify\n"
	"and use the current configuration as an alternate to whatever\n"
	"configuration options you have selected at that time.\n"
	"\n"
	"If you are uncertain what all this means then you should probably\n"
	"leave this blank.\n",
search_help[] =
	"\n"
	"Search for CONFIG_ symbols and display their relations.\n"
	"Example: search for \"^FOO\"\n"
	"Result:\n"
	"-----------------------------------------------------------------\n"
	"Symbol: FOO [=m]\n"
	"Prompt: Foo bus is used to drive the bar HW\n"
	"Defined at drivers/pci/Kconfig:47\n"
	"Depends on: X86_LOCAL_APIC && X86_IO_APIC || IA64\n"
	"Location:\n"
	"  -> Bus options (PCI, PCMCIA, EISA, MCA, ISA)\n"
	"    -> PCI support (PCI [=y])\n"
	"      -> PCI access mode (<choice> [=y])\n"
	"Selects: LIBCRC32\n"
	"Selected by: BAR\n"
	"-----------------------------------------------------------------\n"
	"o The line 'Prompt:' shows the text used in the menu structure for\n"
	"  this CONFIG_ symbol\n"
	"o The 'Defined at' line tell at what file / line number the symbol\n"
	"  is defined\n"
	"o The 'Depends on:' line tell what symbols needs to be defined for\n"
	"  this symbol to be visible in the menu (selectable)\n"
	"o The 'Location:' lines tell where in the menu structure this symbol\n"
	"  is located\n"
	"    A location followed by a [=y] indicate that this is a selectable\n"
	"    menu item - and current value is displayed inside brackets.\n"
	"o The 'Selects:' line tell what symbol will be automatically\n"
	"  selected if this symbol is selected (y or m)\n"
	"o The 'Selected by' line tell what symbol has selected this symbol\n"
	"\n"
	"Only relevant lines are shown.\n"
	"\n\n"
	"Search examples:\n"
	"Examples: USB  => find all CONFIG_ symbols containing USB\n"
	"          ^USB => find all CONFIG_ symbols starting with USB\n"
	"          USB$ => find all CONFIG_ symbols ending with USB\n"
	"\n";

static char filename[256+1] = ".config";
static int indent;
static struct termios ios_org;
static int rows, cols;
static struct menu *current_menu;
static int child_count;
static int single_menu_mode;

static struct dialog_list_item *items[16384]; /* FIXME: This ought to be dynamic. */
static int item_no;

static void conf(struct menu *menu);
static void conf_choice(struct menu *menu);
static void conf_string(struct menu *menu);
static void conf_load(void);
static void conf_save(void);
static void show_textbox(const char *title, const char *text, int r, int c);
static void show_helptext(const char *title, const char *text);
static void show_help(struct menu *menu);
static void show_file(const char *filename, const char *title, int r, int c);

static void init_wsize(void)
{
	struct winsize ws;
	char *env;

	if (!ioctl(STDIN_FILENO, TIOCGWINSZ, &ws)) {
		rows = ws.ws_row;
		cols = ws.ws_col;
	}

	if (!rows) {
		env = getenv("LINES");
		if (env)
			rows = atoi(env);
		if (!rows)
			rows = 24;
	}
	if (!cols) {
		env = getenv("COLUMNS");
		if (env)
			cols = atoi(env);
		if (!cols)
			cols = 80;
	}

	if (rows < 19 || cols < 80) {
		fprintf(stderr, "Your display is too small to run Menuconfig!\n");
		fprintf(stderr, "It must be at least 19 lines by 80 columns.\n");
		exit(1);
	}

	rows -= 4;
	cols -= 5;
}

static void cinit(void)
{
	item_no = 0;
}

static void cmake(void)
{
	items[item_no] = malloc(sizeof(struct dialog_list_item));
	memset(items[item_no], 0, sizeof(struct dialog_list_item));
	items[item_no]->tag = malloc(32); items[item_no]->tag[0] = 0;
	items[item_no]->name = malloc(512); items[item_no]->name[0] = 0;
	items[item_no]->namelen = 0;
	item_no++;
}

static int cprint_name(const char *fmt, ...)
{
	va_list ap;
	int res;

	if (!item_no)
		cmake();
	va_start(ap, fmt);
	res = vsnprintf(items[item_no - 1]->name + items[item_no - 1]->namelen,
			512 - items[item_no - 1]->namelen, fmt, ap);
	if (res > 0)
		items[item_no - 1]->namelen += res;
	va_end(ap);

	return res;
}

static int cprint_tag(const char *fmt, ...)
{
	va_list ap;
	int res;

	if (!item_no)
		cmake();
	va_start(ap, fmt);
	res = vsnprintf(items[item_no - 1]->tag, 32, fmt, ap);
	va_end(ap);

	return res;
}

static void cdone(void)
{
	int i;

	for (i = 0; i < item_no; i++) {
		free(items[i]->tag);
		free(items[i]->name);
		free(items[i]);
	}

	item_no = 0;
}

static void get_prompt_str(struct gstr *r, struct property *prop)
{
	int i, j;
	struct menu *submenu[8], *menu;

	str_printf(r, "Prompt: %s\n", prop->text);
	str_printf(r, "  Defined at %s:%d\n", prop->menu->file->name,
		prop->menu->lineno);
	if (!expr_is_yes(prop->visible.expr)) {
		str_append(r, "  Depends on: ");
		expr_gstr_print(prop->visible.expr, r);
		str_append(r, "\n");
	}
	menu = prop->menu->parent;
	for (i = 0; menu != &rootmenu && i < 8; menu = menu->parent)
		submenu[i++] = menu;
	if (i > 0) {
		str_printf(r, "  Location:\n");
		for (j = 4; --i >= 0; j += 2) {
			menu = submenu[i];
			str_printf(r, "%*c-> %s", j, ' ', menu_get_prompt(menu));
			if (menu->sym) {
				str_printf(r, " (%s [=%s])", menu->sym->name ?
					menu->sym->name : "<choice>",
					sym_get_string_value(menu->sym));
			}
			str_append(r, "\n");
		}
	}
}

static void get_symbol_str(struct gstr *r, struct symbol *sym)
{
	bool hit;
	struct property *prop;

	str_printf(r, "Symbol: %s [=%s]\n", sym->name,
				       sym_get_string_value(sym));
	for_all_prompts(sym, prop)
		get_prompt_str(r, prop);
	hit = false;
	for_all_properties(sym, prop, P_SELECT) {
		if (!hit) {
			str_append(r, "  Selects: ");
			hit = true;
		} else
			str_printf(r, " && ");
		expr_gstr_print(prop->expr, r);
	}
	if (hit)
		str_append(r, "\n");
	if (sym->rev_dep.expr) {
		str_append(r, "  Selected by: ");
		expr_gstr_print(sym->rev_dep.expr, r);
		str_append(r, "\n");
	}
	str_append(r, "\n\n");
}

static struct gstr get_relations_str(struct symbol **sym_arr)
{
	struct symbol *sym;
	struct gstr res = str_new();
	int i;

	for (i = 0; sym_arr && (sym = sym_arr[i]); i++)
		get_symbol_str(&res, sym);
	if (!i)
		str_append(&res, "No matches found.\n");
	return res;
}

static void search_conf(void)
{
	struct symbol **sym_arr;
	struct gstr res;

again:
	switch (dialog_inputbox("Search Configuration Parameter",
				"Enter Keyword", 10, 75,
				NULL)) {
	case 0:
		break;
	case 1:
		show_helptext("Search Configuration", search_help);
		goto again;
	default:
		return;
	}

	sym_arr = sym_re_search((char *)dialog_input_result);
	res = get_relations_str(sym_arr);
	free(sym_arr);
	show_textbox("Search Results", str_get(&res), 0, 0);
	str_free(&res);
}

static void build_conf(struct menu *menu)
{
	struct symbol *sym;
	struct property *prop;
	struct menu *child;
	int type, tmp, doint = 2;
	tristate val;
	char ch;

	if (!menu_is_visible(menu))
		return;

	sym = menu->sym;
	prop = menu->prompt;
	if (!sym) {
		if (prop && menu != current_menu) {
			const char *prompt = menu_get_prompt(menu);
			switch (prop->type) {
			case P_MENU:
				child_count++;
				cmake();
				cprint_tag("m%p", menu);

				if (single_menu_mode) {
					cprint_name("%s%*c%s",
						menu->data ? "-->" : "++>",
						indent + 1, ' ', prompt);
				} else {
					cprint_name("   %*c%s  --->", indent + 1, ' ', prompt);
				}

				if (single_menu_mode && menu->data)
					goto conf_childs;
				return;
			default:
				if (prompt) {
					child_count++;
					cmake();
					cprint_tag(":%p", menu);
					cprint_name("---%*c%s", indent + 1, ' ', prompt);
				}
			}
		} else
			doint = 0;
		goto conf_childs;
	}

	cmake();
	type = sym_get_type(sym);
	if (sym_is_choice(sym)) {
		struct symbol *def_sym = sym_get_choice_value(sym);
		struct menu *def_menu = NULL;

		child_count++;
		for (child = menu->list; child; child = child->next) {
			if (menu_is_visible(child) && child->sym == def_sym)
				def_menu = child;
		}

		val = sym_get_tristate_value(sym);
		if (sym_is_changable(sym)) {
			cprint_tag("t%p", menu);
			switch (type) {
			case S_BOOLEAN:
				cprint_name("[%c]", val == no ? ' ' : '*');
				break;
			case S_TRISTATE:
				switch (val) {
				case yes: ch = '*'; break;
				case mod: ch = 'M'; break;
				default:  ch = ' '; break;
				}
				cprint_name("<%c>", ch);
				break;
			}
		} else {
			cprint_tag("%c%p", def_menu ? 't' : ':', menu);
			cprint_name("   ");
		}

		cprint_name("%*c%s", indent + 1, ' ', menu_get_prompt(menu));
		if (val == yes) {
			if (def_menu) {
				cprint_name(" (%s)", menu_get_prompt(def_menu));
				cprint_name("  --->");
				if (def_menu->list) {
					indent += 2;
					build_conf(def_menu);
					indent -= 2;
				}
			}
			return;
		}
	} else {
		if (menu == current_menu) {
			cprint_tag(":%p", menu);
			cprint_name("---%*c%s", indent + 1, ' ', menu_get_prompt(menu));
			goto conf_childs;
		}
		child_count++;
		val = sym_get_tristate_value(sym);
		if (sym_is_choice_value(sym) && val == yes) {
			cprint_tag(":%p", menu);
			cprint_name("   ");
		} else {
			switch (type) {
			case S_BOOLEAN:
				cprint_tag("t%p", menu);
				if (sym_is_changable(sym))
					cprint_name("[%c]", val == no ? ' ' : '*');
				else
					cprint_name("---");
				break;
			case S_TRISTATE:
				cprint_tag("t%p", menu);
				switch (val) {
				case yes: ch = '*'; break;
				case mod: ch = 'M'; break;
				default:  ch = ' '; break;
				}
				if (sym_is_changable(sym))
					cprint_name("<%c>", ch);
				else
					cprint_name("---");
				break;
			default:
				cprint_tag("s%p", menu);
				tmp = cprint_name("(%s)", sym_get_string_value(sym));
				tmp = indent - tmp + 4;
				if (tmp < 0)
					tmp = 0;
				cprint_name("%*c%s%s", tmp, ' ', menu_get_prompt(menu),
					(sym_has_value(sym) || !sym_is_changable(sym)) ?
					"" : " (NEW)");
				goto conf_childs;
			}
		}
		cprint_name("%*c%s%s", indent + 1, ' ', menu_get_prompt(menu),
			(sym_has_value(sym) || !sym_is_changable(sym)) ?
			"" : " (NEW)");
		if (menu->prompt->type == P_MENU) {
			cprint_name("  --->");
			return;
		}
	}

conf_childs:
	indent += doint;
	for (child = menu->list; child; child = child->next)
		build_conf(child);
	indent -= doint;
}

static void conf(struct menu *menu)
{
	struct dialog_list_item *active_item = NULL;
	struct menu *submenu;
	const char *prompt = menu_get_prompt(menu);
	struct symbol *sym;
	char active_entry[40];
	int stat, type;

	unlink("lxdialog.scrltmp");
	active_entry[0] = 0;
	while (1) {
		indent = 0;
		child_count = 0;
		current_menu = menu;
		cdone(); cinit();
		build_conf(menu);
		if (!child_count)
			break;
		if (menu == &rootmenu) {
			cmake(); cprint_tag(":"); cprint_name("--- ");
			cmake(); cprint_tag("L"); cprint_name("Load an Alternate Configuration File");
			cmake(); cprint_tag("S"); cprint_name("Save Configuration to an Alternate File");
		}
		dialog_clear();
		stat = dialog_menu(prompt ? prompt : "Main Menu",
				menu_instructions, rows, cols, rows - 10,
				active_entry, item_no, items);
		if (stat < 0)
			return;

		if (stat == 1 || stat == 255)
			break;

		active_item = first_sel_item(item_no, items);
		if (!active_item)
			continue;
		active_item->selected = 0;
		strncpy(active_entry, active_item->tag, sizeof(active_entry));
		active_entry[sizeof(active_entry)-1] = 0;
		type = active_entry[0];
		if (!type)
			continue;

		sym = NULL;
		submenu = NULL;
		if (sscanf(active_entry + 1, "%p", &submenu) == 1)
			sym = submenu->sym;

		switch (stat) {
		case 0:
			switch (type) {
			case 'm':
				if (single_menu_mode)
					submenu->data = (void *) (long) !submenu->data;
				else
					conf(submenu);
				break;
			case 't':
				if (sym_is_choice(sym) && sym_get_tristate_value(sym) == yes)
					conf_choice(submenu);
				else if (submenu->prompt->type == P_MENU)
					conf(submenu);
				break;
			case 's':
				conf_string(submenu);
				break;
			case 'L':
				conf_load();
				break;
			case 'S':
				conf_save();
				break;
			}
			break;
		case 2:
			if (sym)
				show_help(submenu);
			else
				show_helptext("README", mconf_readme);
			break;
		case 3:
			if (type == 't') {
				if (sym_set_tristate_value(sym, yes))
					break;
				if (sym_set_tristate_value(sym, mod))
					show_textbox(NULL, setmod_text, 6, 74);
			}
			break;
		case 4:
			if (type == 't')
				sym_set_tristate_value(sym, no);
			break;
		case 5:
			if (type == 't')
				sym_set_tristate_value(sym, mod);
			break;
		case 6:
			if (type == 't')
				sym_toggle_tristate_value(sym);
			else if (type == 'm')
				conf(submenu);
			break;
		case 7:
			search_conf();
			break;
		}
	}
}

static void show_textbox(const char *title, const char *text, int r, int c)
{
	int fd;

	fd = creat(".help.tmp", 0777);
	write(fd, text, strlen(text));
	close(fd);
	show_file(".help.tmp", title, r, c);
	unlink(".help.tmp");
}

static void show_helptext(const char *title, const char *text)
{
	show_textbox(title, text, 0, 0);
}

static void show_help(struct menu *menu)
{
	struct gstr help = str_new();
	struct symbol *sym = menu->sym;

	if (sym->help)
	{
		if (sym->name) {
			str_printf(&help, "%s:\n\n", sym->name);
			str_append(&help, sym->help);
			str_append(&help, "\n");
		}
	} else {
		str_append(&help, nohelp_text);
	}
	get_symbol_str(&help, sym);
	show_helptext(menu_get_prompt(menu), str_get(&help));
	str_free(&help);
}

static void show_file(const char *filename, const char *title, int r, int c)
{
	while (dialog_textbox(title, filename, r ? r : rows, c ? c : cols) < 0)
		;
}

static void conf_choice(struct menu *menu)
{
	const char *prompt = menu_get_prompt(menu);
	struct menu *child;
	struct symbol *active;

	active = sym_get_choice_value(menu->sym);
	while (1) {
		current_menu = menu;
		cdone(); cinit();
		for (child = menu->list; child; child = child->next) {
			if (!menu_is_visible(child))
				continue;
			cmake();
			cprint_tag("%p", child);
			cprint_name("%s", menu_get_prompt(child));
			if (child->sym == sym_get_choice_value(menu->sym))
				items[item_no - 1]->selected = 1; /* ON */
			else if (child->sym == active)
				items[item_no - 1]->selected = 2; /* SELECTED */
			else
				items[item_no - 1]->selected = 0; /* OFF */
		}

		switch (dialog_checklist(prompt ? prompt : "Main Menu",
					radiolist_instructions, 15, 70, 6,
					item_no, items, FLAG_RADIO)) {
		case 0:
			if (sscanf(first_sel_item(item_no, items)->tag, "%p", &child) != 1)
				break;
			sym_set_tristate_value(child->sym, yes);
			return;
		case 1:
			if (sscanf(first_sel_item(item_no, items)->tag, "%p", &child) == 1) {
				show_help(child);
				active = child->sym;
			} else
				show_help(menu);
			break;
		case 255:
			return;
		}
	}
}

static void conf_string(struct menu *menu)
{
	const char *prompt = menu_get_prompt(menu);

	while (1) {
		char *heading;

		switch (sym_get_type(menu->sym)) {
		case S_INT:
			heading = (char *) inputbox_instructions_int;
			break;
		case S_HEX:
			heading = (char *) inputbox_instructions_hex;
			break;
		case S_STRING:
			heading = (char *) inputbox_instructions_string;
			break;
		default:
			heading = "Internal mconf error!";
			/* panic? */;
		}

		switch (dialog_inputbox(prompt ? prompt : "Main Menu",
					heading, 10, 75,
					sym_get_string_value(menu->sym))) {
		case 0:
			if (sym_set_string_value(menu->sym, (char *)dialog_input_result))
				return;
			show_textbox(NULL, "You have made an invalid entry.", 5, 43);
			break;
		case 1:
			show_help(menu);
			break;
		case 255:
			return;
		}
	}
}

static void conf_load(void)
{
	while (1) {
		switch (dialog_inputbox(NULL, load_config_text, 11, 55,
					filename)) {
		case 0:
			if (!dialog_input_result[0])
				return;
			if (!conf_read((char *)dialog_input_result))
				return;
			show_textbox(NULL, "File does not exist!", 5, 38);
			break;
		case 1:
			show_helptext("Load Alternate Configuration", load_config_help);
			break;
		case 255:
			return;
		}
	}
}

static void conf_save(void)
{
	while (1) {
		switch (dialog_inputbox(NULL, save_config_text, 11, 55,
					filename)) {
		case 0:
			if (!dialog_input_result[0])
				return;
			if (!conf_write((char *)dialog_input_result))
				return;
			show_textbox(NULL, "Can't create file!  Probably a nonexistent directory.", 5, 60);
			break;
		case 1:
			show_helptext("Save Alternate Configuration", save_config_help);
			break;
		case 255:
			return;
		}
	}
}

static void conf_cleanup(void)
{
	tcsetattr(1, TCSAFLUSH, &ios_org);
	unlink(".help.tmp");
}

static void winch_handler(int sig GCC_UNUSED)
{
	struct winsize ws;

	if (ioctl(1, TIOCGWINSZ, &ws) == -1) {
		rows = 24;
		cols = 80;
	} else {
		rows = ws.ws_row;
		cols = ws.ws_col;
	}

	if (rows < 19 || cols < 80) {
		end_dialog();
		fprintf(stderr, "Your display is too small to run Menuconfig!\n");
		fprintf(stderr, "It must be at least 19 lines by 80 columns.\n");
		exit(1);
	}

	rows -= 4;
	cols -= 5;

}

int main(int ac GCC_UNUSED, char **av)
{
	struct symbol *sym;
	char *mode;
	int stat;

	conf_parse(av[1]);
	conf_read(NULL);

	sym = sym_lookup("VERSION", 0);
	sym_calc_value(sym);
	snprintf(menu_backtitle, 128, "BusyBox v%s Configuration",
		sym_get_string_value(sym));

	mode = getenv("MENUCONFIG_MODE");
	if (mode) {
		if (!strcasecmp(mode, "single_menu"))
			single_menu_mode = 1;
	}

	tcgetattr(1, &ios_org);
	atexit(conf_cleanup);
	init_wsize();
	init_dialog();
	signal(SIGWINCH, winch_handler);
	conf(&rootmenu);
	end_dialog();

	/* Restart dialog to act more like when lxdialog was still separate */
	init_dialog();
	do {
		stat = dialog_yesno(NULL,
			"Do you wish to save your new BusyBox configuration?",
			5, 60);
	} while (stat < 0);
	end_dialog();

	if (stat == 0) {
		conf_write(NULL);
		printf("\n"
			"*** End of BusyBox configuration.\n");
	} else
		printf("\n\nYour BusyBox configuration changes were NOT saved.\n\n");

	return 0;
}
