/*
 * mdassemble - assemble Linux "md" devices aka RAID arrays.
 *
 * Copyright (C) 2001-2009 Neil Brown <neilb@suse.de>
 * Copyright (C) 2003 Luca Berra <bluca@vodka.it>
 *
 *  Retrieve drive serial numbers for scsi disks
 *   Copyright (C) 2007-2008 Intel Corporation
 *  Intel(R) Matrix Storage Manager hardware and firmware support routines
 *   Copyright (C) 2008 Intel Corporation
 *
 *  crc32.c -- compute the CRC-32 of a data stream
 *   Copyright (C) 1995-2003 Mark Adler
 *
 * Busybox port by Vladimir Oleynik (c) 2013 <dzo@simtreas.ru>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

#ifndef CONFFILE
#define CONFFILE "/etc/mdadm.conf"
#endif
#ifndef CONFFILE2
/* for Debian compatibility .... */
#define CONFFILE2 "/etc/mdadm/mdadm.conf"
#endif

#include "busybox.h"

#include <asm/types.h>
#include <sys/mman.h>
#include <sys/select.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <sys/sysmacros.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <dirent.h>
#include <glob.h>
#include <fnmatch.h>
#include <signal.h>
#include <strings.h>
#include <time.h>
#include <sys/time.h>
#include <syslog.h>
#include <linux/kdev_t.h>
#include <sys/mount.h>
#include <endian.h>
#include <assert.h>

#include <inttypes.h>
#include "sha1.h"
#include "bb_xopen.h"

#if defined(__BIONIC__)
#include "sg.h"
#include "scsi.h"
#else
#include <scsi/sg.h>
#include <scsi/scsi.h>
#endif

#ifndef BLKSSZGET
#include <linux/fs.h>
#endif

#ifdef __dietlibc__
/* dietlibc has deprecated random!! */
#define random rand
#endif

#ifdef HAVE_FTW
# include <ftw.h>
#endif

#ifndef HAVE_NFTW
# ifndef FTW_PHYS
#  define FTW_PHYS 1
# endif
# ifndef HAVE_FTW
  struct FTW {};
# endif
#endif



/*
 * RAID superblock.
 *
 * The RAID superblock maintains some statistics on each RAID configuration.
 * Each real device in the RAID set contains it near the end of the device.
 * Some of the ideas are copied from the ext2fs implementation.
 *
 * We currently use 4096 bytes as follows:
 *
 *      word offset     function
 *
 *         0  -    31   Constant generic RAID device information.
 *        32  -    63   Generic state information.
 *        64  -   127   Personality specific information.
 *       128  -   511   12 32-words descriptors of the disks in the raid set.
 *       512  -   911   Reserved.
 *       912  -  1023   Disk specific descriptor.
 */

/*
 * If x is the real device size in bytes, we return an apparent size of:
 *
 *      y = (x & ~(MD_RESERVED_BYTES - 1)) - MD_RESERVED_BYTES
 *
 * and place the 4kB superblock at offset y.
 */
#define MD_RESERVED_BYTES               (64 * 1024)
#define MD_RESERVED_SECTORS             (MD_RESERVED_BYTES / 512)
#define MD_RESERVED_BLOCKS              (MD_RESERVED_BYTES / BLOCK_SIZE)

#define MD_NEW_SIZE_SECTORS(x)          ((x & ~(MD_RESERVED_SECTORS - 1)) - MD_RESERVED_SECTORS)
#define MD_NEW_SIZE_BLOCKS(x)           ((x & ~(MD_RESERVED_BLOCKS - 1)) - MD_RESERVED_BLOCKS)

#define MD_SB_BYTES                     4096
#define MD_SB_WORDS                     (MD_SB_BYTES / 4)
#define MD_SB_BLOCKS                    (MD_SB_BYTES / BLOCK_SIZE)
#define MD_SB_SECTORS                   (MD_SB_BYTES / 512)

/* The following are counted in 32-bit words */
#define MD_SB_GENERIC_CONSTANT_WORDS    32
#define MD_SB_GENERIC_STATE_WORDS       32
#define MD_SB_GENERIC_WORDS             (MD_SB_GENERIC_CONSTANT_WORDS + MD_SB_GENERIC_STATE_WORDS)
#define MD_SB_PERSONALITY_WORDS         64
#define MD_SB_DESCRIPTOR_WORDS          32
#define MD_SB_DISKS                     27
#define MD_SB_DISKS_WORDS               (MD_SB_DISKS*MD_SB_DESCRIPTOR_WORDS)
#define MD_SB_RESERVED_WORDS            (1024 - MD_SB_GENERIC_WORDS - MD_SB_PERSONALITY_WORDS - MD_SB_DISKS_WORDS - MD_SB_DESCRIPTOR_WORDS)
#define MD_SB_EQUAL_WORDS               (MD_SB_GENERIC_WORDS + MD_SB_PERSONALITY_WORDS + MD_SB_DISKS_WORDS)

/* Device "operational" state bits */
#define MD_DISK_FAULTY          0 /* disk is faulty / operational */
#define MD_DISK_ACTIVE          1 /* disk is running but may not be in sync */
#define MD_DISK_SYNC            2 /* disk is in sync with the raid set */
#define MD_DISK_REMOVED         3 /* disk is in sync with the raid set */

#define MD_DISK_WRITEMOSTLY     9 /* disk is "write-mostly" is RAID1 config.
				   * read requests will only be sent here in
				   * dire need */

typedef struct mdp_device_descriptor_s {
	__u32 number;           /* 0 Device number in the entire set          */
	__u32 major;            /* 1 Device major number                      */
	__u32 minor;            /* 2 Device minor number                      */
	__u32 raid_disk;        /* 3 The role of the device in the raid set   */
	__u32 state;            /* 4 Operational state                        */
	__u32 reserved[MD_SB_DESCRIPTOR_WORDS - 5];
} mdp_disk_t;

#define MD_SB_MAGIC             0xa92b4efc

/* Superblock state bits */
#define MD_SB_CLEAN             0
#define MD_SB_ERRORS            1
#define MD_SB_BBM_ERRORS        2
#define MD_SB_BLOCK_CONTAINER_RESHAPE 3 /* block container wide reshapes */
#define MD_SB_BLOCK_VOLUME      4 /* block activation of array, other arrays
				   * in container can be activated */
#define MD_SB_BITMAP_PRESENT    8 /* bitmap may be present nearby */

typedef struct mdp_superblock_s {
	/* Constant generic information */
	__u32 md_magic;         /*  0 MD identifier                           */
	__u32 major_version;    /*  1 major version to which the set conforms */
	__u32 minor_version;    /*  2 minor version ...                       */
	__u32 patch_version;    /*  3 patchlevel version ...                  */
	__u32 gvalid_words;     /*  4 Number of used words in this section    */
	__u32 set_uuid0;        /*  5 Raid set identifier                     */
	__u32 ctime;            /*  6 Creation time                           */
	__u32 level;            /*  7 Raid personality                        */
	__u32 size;             /*  8 Apparent size of each individual disk   */
	__u32 nr_disks;         /*  9 total disks in the raid set             */
	__u32 raid_disks;       /* 10 disks in a fully functional raid set    */
	__u32 md_minor;         /* 11 preferred MD minor device number        */
	__u32 not_persistent;   /* 12 does it have a persistent superblock    */
	__u32 set_uuid1;        /* 13 Raid set identifier #2                  */
	__u32 set_uuid2;        /* 14 Raid set identifier #3                  */
	__u32 set_uuid3;        /* 15 Raid set identifier #4                  */
	__u32 gstate_creserved[MD_SB_GENERIC_CONSTANT_WORDS - 16];

	/* Generic state information */
	__u32 utime;            /*  0 Superblock update time                  */
	__u32 state;            /*  1 State bits (clean, ...)                 */
	__u32 active_disks;     /*  2 Number of currently active disks        */
	__u32 working_disks;    /*  3 Number of working disks                 */
	__u32 failed_disks;     /*  4 Number of failed disks                  */
	__u32 spare_disks;      /*  5 Number of spare disks                   */
	__u32 sb_csum;          /*  6 checksum of the whole superblock        */
#if  __BYTE_ORDER ==  __BIG_ENDIAN
	__u32 events_hi;        /*  7 high-order of superblock update count   */
	__u32 events_lo;        /*  8 low-order of superblock update count    */
	__u32 cp_events_hi;     /*  9 high-order of checkpoint update count   */
	__u32 cp_events_lo;     /* 10 low-order of checkpoint update count    */
#else
	__u32 events_lo;        /*  7 low-order of superblock update count    */
	__u32 events_hi;        /*  8 high-order of superblock update count   */
	__u32 cp_events_lo;     /*  9 low-order of checkpoint update count    */
	__u32 cp_events_hi;     /* 10 high-order of checkpoint update count   */
#endif
	__u32 recovery_cp;      /* 11 recovery checkpoint sector count        */
	/* There are only valid for minor_version > 90 */
	__u64 reshape_position; /* 12,13 next address in array-space for reshape */
	__u32 new_level;        /* 14 new level we are reshaping to           */
	__u32 delta_disks;      /* 15 change in number of raid_disks          */
	__u32 new_layout;       /* 16 new layout                              */
	__u32 new_chunk;        /* 17 new chunk size (bytes)                  */
	__u32 gstate_sreserved[MD_SB_GENERIC_STATE_WORDS - 18];

	/* Personality information */
	__u32 layout;           /*  0 the array's physical layout             */
	__u32 chunk_size;       /*  1 chunk size in bytes                     */
	__u32 root_pv;          /*  2 LV root PV */
	__u32 root_block;       /*  3 LV root block */
	__u32 pstate_reserved[MD_SB_PERSONALITY_WORDS - 4];

	/* Disks information */
	mdp_disk_t disks[MD_SB_DISKS];

	/* Reserved */
	__u32 reserved[MD_SB_RESERVED_WORDS];

	/* Active descriptor */
	mdp_disk_t this_disk;

} mdp_super_t;

#ifdef __TINYC__
typedef unsigned long long __u64;
#endif

static inline __u64 md_event(mdp_super_t *sb) {
	__u64 ev = sb->events_hi;
	return (ev<<32)| sb->events_lo;
}


/* ioctls */

/* status */
#define RAID_VERSION            _IOR (MD_MAJOR, 0x10, mdu_version_t)
#define GET_ARRAY_INFO          _IOR (MD_MAJOR, 0x11, mdu_array_info_t)
#define GET_DISK_INFO           _IOR (MD_MAJOR, 0x12, mdu_disk_info_t)
#define PRINT_RAID_DEBUG        _IO (MD_MAJOR, 0x13)
#define RAID_AUTORUN            _IO (MD_MAJOR, 0x14)

/* configuration */
#define ADD_NEW_DISK            _IOW (MD_MAJOR, 0x21, mdu_disk_info_t)
#define SET_ARRAY_INFO          _IOW (MD_MAJOR, 0x23, mdu_array_info_t)
#define SET_BITMAP_FILE         _IOW (MD_MAJOR, 0x2b, int)

/* usage */
#define RUN_ARRAY               _IOW (MD_MAJOR, 0x30, mdu_param_t)
#define START_ARRAY             _IO (MD_MAJOR, 0x31)
#define STOP_ARRAY              _IO (MD_MAJOR, 0x32)
#define STOP_ARRAY_RO           _IO (MD_MAJOR, 0x33)
#define RESTART_ARRAY_RW        _IO (MD_MAJOR, 0x34)

typedef struct mdu_version_s {
	int major;
	int minor;
	int patchlevel;
} mdu_version_t;

typedef struct mdu_array_info_s {
	/* Generic constant information */
	int major_version;
	int minor_version;
	int patch_version;
	int ctime;
	int level;
	int size;
	int nr_disks;
	int raid_disks;
	int md_minor;
	int not_persistent;

	/* Generic state information */
	int utime;              /*  0 Superblock update time                  */
	int state;              /*  1 State bits (clean, ...)                 */
	int active_disks;       /*  2 Number of currently active disks        */
	int working_disks;      /*  3 Number of working disks                 */
	int failed_disks;       /*  4 Number of failed disks                  */
	int spare_disks;        /*  5 Number of spare disks                   */

	/* Personality information */
	int layout;             /* 0 the array's physical layout             */
	int chunk_size;         /* 1 chunk size in bytes */
} mdu_array_info_t;

typedef struct mdu_disk_info_s {
	/* configuration/status of one particular disk */
	int number;
	int major;
	int minor;
	int raid_disk;
	int state;
} mdu_disk_info_t;

typedef struct mdu_start_info_s {
	/* configuration/status of one particular disk */
	int major;
	int minor;
	int raid_disk;
	int state;

} mdu_start_info_t;

typedef struct mdu_param_s
{
	int                     personality;    /* 1,2,3,4 */
	int                     chunk_size;     /* in bytes */
	int                     max_fault;      /* unused for now */
} mdu_param_t;

/* The IMSM Capability (IMSM AHCI and ISCU OROM/EFI variable) Version Table definition */
struct imsm_orom {
	__u8 signature[4];
	__u8 table_ver_major; /* Currently 2 (can change with future revs) */
	__u8 table_ver_minor; /* Currently 2 (can change with future revs) */
	__u16 major_ver; /* Example: 8 as in 8.6.0.1020 */
	__u16 minor_ver; /* Example: 6 as in 8.6.0.1020 */
	__u16 hotfix_ver; /* Example: 0 as in 8.6.0.1020 */
	__u16 build; /* Example: 1020 as in 8.6.0.1020 */
	__u8 len; /* number of bytes in this entire table */
	__u8 checksum; /* checksum of all the bytes in this table */
	__u16 rlc; /* RAID Level Capability */
	/* we assume the cpu is x86 as the orom should not be found
		anywhere else */
	#define IMSM_OROM_RLC_RAID0 (1 << 0)
	#define IMSM_OROM_RLC_RAID1 (1 << 1)
	#define IMSM_OROM_RLC_RAID10 (1 << 2)
	#define IMSM_OROM_RLC_RAID1E (1 << 3)
	#define IMSM_OROM_RLC_RAID5 (1 << 4)
	#define IMSM_OROM_RLC_RAID_CNG (1 << 5)
	__u16 sss; /* Strip Size Supported */
	#define IMSM_OROM_SSS_2kB (1 << 0)
	#define IMSM_OROM_SSS_4kB (1 << 1)
	#define IMSM_OROM_SSS_8kB (1 << 2)
	#define IMSM_OROM_SSS_16kB (1 << 3)
	#define IMSM_OROM_SSS_32kB (1 << 4)
	#define IMSM_OROM_SSS_64kB (1 << 5)
	#define IMSM_OROM_SSS_128kB (1 << 6)
	#define IMSM_OROM_SSS_256kB (1 << 7)
	#define IMSM_OROM_SSS_512kB (1 << 8)
	#define IMSM_OROM_SSS_1MB (1 << 9)
	#define IMSM_OROM_SSS_2MB (1 << 10)
	#define IMSM_OROM_SSS_4MB (1 << 11)
	#define IMSM_OROM_SSS_8MB (1 << 12)
	#define IMSM_OROM_SSS_16MB (1 << 13)
	#define IMSM_OROM_SSS_32MB (1 << 14)
	#define IMSM_OROM_SSS_64MB (1 << 15)
	__u16 dpa; /* Disks Per Array supported */
	#define IMSM_OROM_DISKS_PER_ARRAY 6
	__u16 tds; /* Total Disks Supported */
	#define IMSM_OROM_TOTAL_DISKS 6
	__u8 vpa; /* # Volumes Per Array supported */
	#define IMSM_OROM_VOLUMES_PER_ARRAY 2
	__u8 vphba; /* # Volumes Per Host Bus Adapter supported */
	#define IMSM_OROM_VOLUMES_PER_HBA 4
	/* Attributes supported. This should map to the
	 * attributes in the MPB. Also, lower 16 bits
	 * should match/duplicate RLC bits above.
	 */
	__u32 attr;
	#define IMSM_OROM_ATTR_RAID0 IMSM_OROM_RLC_RAID0
	#define IMSM_OROM_ATTR_RAID1 IMSM_OROM_RLC_RAID1
	#define IMSM_OROM_ATTR_RAID10 IMSM_OROM_RLC_RAID10
	#define IMSM_OROM_ATTR_RAID1E IMSM_OROM_RLC_RAID1E
	#define IMSM_OROM_ATTR_RAID5 IMSM_OROM_RLC_RAID5
	#define IMSM_OROM_ATTR_RAID_CNG IMSM_OROM_RLC_RAID_CNG
	#define IMSM_OROM_ATTR_2TB_DISK (1 << 26)
	#define IMSM_OROM_ATTR_2TB (1 << 29)
	#define IMSM_OROM_ATTR_PM (1 << 30)
	#define IMSM_OROM_ATTR_ChecksumVerify (1 << 31)
	__u32 capabilities;
	#define IMSM_OROM_CAPABILITIES_Ext_SATA (1 << 0)
	#define IMSM_OROM_CAPABILITIES_TurboMemory (1 << 1)
	#define IMSM_OROM_CAPABILITIES_HddPassword (1 << 2)
	#define IMSM_OROM_CAPABILITIES_DiskCoercion (1 << 3)
	__u32 driver_features;
	#define IMSM_OROM_CAPABILITIES_HDDUnlock (1 << 0)
	#define IMSM_OROM_CAPABILITIES_LEDLoc (1 << 1)
	#define IMSM_OROM_CAPABILITIES_EnterpriseSystem (1 << 2)
	#define IMSM_OROM_CAPABILITIES_Zpodd (1 << 3)
	#define IMSM_OROM_CAPABILITIES_LargeDramCache (1 << 4)
	#define IMSM_OROM_CAPABILITIES_Rohi (1 << 5)
	#define IMSM_OROM_CAPABILITIES_ReadPatrol (1 << 6)
	#define IMSM_OROM_CAPABILITIES_XorHw (1 << 7)
} __attribute__((packed));

enum sys_dev_type {
	SYS_DEV_UNKNOWN = 0,
	SYS_DEV_SAS,
	SYS_DEV_SATA,
	SYS_DEV_MAX
};

struct sys_dev {
	enum sys_dev_type type;
	char *path;
	char *pci_id;
	__u16  dev_id;
	struct sys_dev *next;
};

struct efi_guid {
	__u8 b[16];
};

static inline char *guid_str(char *buf, struct efi_guid guid)
{
	sprintf(buf, "%02x%02x%02x%02x-%02x%02x-%02x%02x-"
		     "%02x%02x-%02x%02x%02x%02x%02x%02x",
		 guid.b[3], guid.b[2], guid.b[1], guid.b[0],
		 guid.b[5], guid.b[4], guid.b[7], guid.b[6],
		 guid.b[8], guid.b[9], guid.b[10], guid.b[11],
		 guid.b[12], guid.b[13], guid.b[14], guid.b[15]);
	return buf;
}

static char *diskfd_to_devpath(int fd);
static struct sys_dev *find_intel_devices(void);
static const struct imsm_orom *find_imsm_capability(enum sys_dev_type hba_id);
static int path_attached_to_hba(const char *disk_dev_path, const char *hba_path);


/* doubley linked lists */
/* This is free software. No strings attached. No copyright claimed */

struct __dl_head
{
    void * dh_prev;
    void * dh_next;
};

#define dl_alloc(size)  ((void*)(((char*)calloc(1,(size)+sizeof(struct __dl_head)))+sizeof(struct __dl_head)))
#define dl_new(t)       ((t*)dl_alloc(sizeof(t)))
#define dl_newv(t,n)    ((t*)dl_alloc(sizeof(t)*n))

#define dl_next(p) *(&(((struct __dl_head*)(p))[-1].dh_next))
#define dl_prev(p) *(&(((struct __dl_head*)(p))[-1].dh_prev))

static void *dl_head(void);
static char *dl_strdup(char *);
static char *dl_strndup(char *, int);
static void dl_add(void*, void*);
static void dl_del(void*);
static void dl_free(void*);
static void dl_init(void*);

/* Structure definitions ext  for MBR and GPT partition tables */

#define MBR_SIGNATURE_MAGIC     __cpu_to_le16(0xAA55)
#define MBR_PARTITIONS               4

struct MBR_part_record {
  __u8 bootable;
  __u8 first_head;
  __u8 first_sector;
  __u8 first_cyl;
  __u8 part_type;
  __u8 last_head;
  __u8 last_sector;
  __u8 last_cyl;
  __u32 first_sect_lba;
  __u32 blocks_num;
};

struct MBR {
	__u8 pad[446];
	struct MBR_part_record parts[MBR_PARTITIONS];
	__u16 magic;
} __attribute__((packed));



#define GPT_SIGNATURE_MAGIC     __cpu_to_le64(0x5452415020494645ULL)
#define MBR_GPT_PARTITION_TYPE       0xEE

struct GPT_part_entry {
	unsigned char type_guid[16];
	unsigned char partition_guid[16];
	__u64 starting_lba;
	__u64 ending_lba;
	unsigned char attr_bits[8];
	unsigned char name[72];
} __attribute__((packed));

struct GPT {
	__u64 magic;
	__u32 revision;
	__u32 header_size;
	__u32 crc;
	__u32 pad1;
	__u64 current_lba;
	__u64 backup_lba;
	__u64 first_lba;
	__u64 last_lba;
	__u8 guid[16];
	__u64 part_start;
	__u32 part_cnt;
	__u32 part_size;
	__u32 part_crc;
	__u8 pad2[420];
} __attribute__((packed));

/* general information that might be extracted from a superblock */
struct mdinfo {
	mdu_array_info_t        array;
	mdu_disk_info_t         disk;
	__u64                   events;
	int                     uuid[4];
	char                    name[33];
	unsigned long long      data_offset;
	unsigned long long      component_size; /* same as array.size, except in
						 * sectors and up to 64bits.
						 */
	unsigned long long      custom_array_size; /* size for non-default sized
						    * arrays (in sectors)
						    */
#define NO_RESHAPE              0
#define VOLUME_RESHAPE          1
#define CONTAINER_RESHAPE       2
	int                     reshape_active;
	unsigned long long      reshape_progress;
	int                     recovery_blocked; /* for external metadata it
						   * indicates that there is
						   * reshape in progress in
						   * container,
						   * for native metadata it is
						   * reshape_active field mirror
						   */
	union {
		unsigned long long resync_start; /* per-array resync position */
		unsigned long long recovery_start; /* per-device rebuild position */
		#define MaxSector  (~0ULL) /* resync/recovery complete position */
	};
	long                    bitmap_offset;  /* 0 == none, 1 == a file */
	unsigned long           safe_mode_delay; /* ms delay to mark clean */
	int                     new_level, delta_disks, new_layout, new_chunk;
	int                     errors;
	unsigned long           cache_size; /* size of raid456 stripe cache*/
	int                     mismatch_cnt;
	char                    text_version[50];

	int container_member; /* for assembling external-metatdata arrays
			       * This is to be used internally by metadata
			       * handler only */
	int container_enough; /* flag external handlers can set to
			       * indicate that subarrays have not enough (-1),
			       * enough to start (0), or all expected disks (1) */
	char            sys_name[20];
	struct mdinfo *devs;
	struct mdinfo *next;

	/* Device info for mdmon: */
	int recovery_fd;
	int state_fd;
	#define DS_FAULTY       1
	#define DS_INSYNC       2
	#define DS_WRITE_MOSTLY 4
	#define DS_SPARE        8
	#define DS_BLOCKED      16
	#define DS_REMOVE       1024
	#define DS_UNBLOCK      2048
	int prev_state, curr_state, next_state;

};

struct active_array {
	struct mdinfo info;
	struct supertype *container;
	struct active_array *next, *replaces;
	int to_remove;

	int action_fd;
	int resync_start_fd;
	int metadata_fd; /* for monitoring rw/ro status */
	int sync_completed_fd; /* for checkpoint notification events */
	unsigned long long last_checkpoint; /* sync_completed fires for many
					     * reasons this field makes sure the
					     * kernel has made progress before
					     * moving the checkpoint.  It is
					     * cleared by the metadata handler
					     * when it determines recovery is
					     * terminated.
					     */

	int check_degraded; /* flag set by mon, read by manage */
	int check_reshape; /* flag set by mon, read by manage */

	int devnum;
};

/*
 * Metadata updates are handled by the monitor thread,
 * as it has exclusive access to the metadata.
 * When the manager want to updates metadata, either
 * for it's own reason (e.g. committing a spare) or
 * on behalf of mdadm, it creates a metadata_update
 * structure and queues it to the monitor.
 * Updates are created and processed by code under the
 * superswitch.  All common code sees them as opaque
 * blobs.
 */

struct metadata_update {
	int     len;
	char    *buf;
	void    *space; /* allocated space that monitor will use */
	void    **space_list; /* list of allocated spaces that monitor can
			       * use or that it returned.
			       */
	struct metadata_update *next;
};

#define MD_MAJOR 9

/* helper routine to determine resync completion since MaxSector is a
 * moving target */
static inline int is_resync_complete(struct mdinfo *array)
{
	if (array->resync_start >= array->component_size)
		return 1;
	return 0;
}

#define MD_MAJOR 9
#define MdpMinorShift 6

#ifndef BLKGETSIZE64
#define BLKGETSIZE64 _IOR(0x12,114,size_t) /* return device size in bytes (u64 *arg) */
#endif

#define DEFAULT_CHUNK 512
#define DEFAULT_BITMAP_CHUNK 4096
#define DEFAULT_BITMAP_DELAY 5
#define DEFAULT_MAX_WRITE_BEHIND 256

/*
 * in-memory bitmap:
 *
 * Use 16 bit block counters to track pending writes to each "chunk".
 * The 2 high order bits are special-purpose, the first is a flag indicating
 * whether a resync is needed.  The second is a flag indicating whether a
 * resync is active.
 * This means that the counter is actually 14 bits:
 *
 * +--------+--------+------------------------------------------------+
 * | resync | resync |               counter                          |
 * | needed | active |                                                |
 * |  (0-1) |  (0-1) |              (0-16383)                         |
 * +--------+--------+------------------------------------------------+
 *
 * The "resync needed" bit is set when:
 *    a '1' bit is read from storage at startup.
 *    a write request fails on some drives
 *    a resync is aborted on a chunk with 'resync active' set
 * It is cleared (and resync-active set) when a resync starts across all drives
 * of the chunk.
 *
 *
 * The "resync active" bit is set when:
 *    a resync is started on all drives, and resync_needed is set.
 *       resync_needed will be cleared (as long as resync_active wasn't already set).
 * It is cleared when a resync completes.
 *
 * The counter counts pending write requests, plus the on-disk bit.
 * When the counter is '1' and the resync bits are clear, the on-disk
 * bit can be cleared aswell, thus setting the counter to 0.
 * When we set a bit, or in the counter (to start a write), if the fields is
 * 0, we first set the disk bit and set the counter to 1.
 *
 * If the counter is 0, the on-disk bit is clear and the stipe is clean
 * Anything that dirties the stipe pushes the counter to 2 (at least)
 * and sets the on-disk bit (lazily).
 * If a periodic sweep find the counter at 2, it is decremented to 1.
 * If the sweep find the counter at 1, the on-disk bit is cleared and the
 * counter goes to zero.
 *
 * Also, we'll hijack the "map" pointer itself and use it as two 16 bit block
 * counters as a fallback when "page" memory cannot be allocated:
 *
 * Normal case (page memory allocated):
 *
 *     page pointer (32-bit)
 *
 *     [ ] ------+
 *               |
 *               +-------> [   ][   ]..[   ] (4096 byte page == 2048 counters)
 *                          c1   c2    c2048
 *
 * Hijacked case (page memory allocation failed):
 *
 *     hijacked page pointer (32-bit)
 *
 *     [                  ][              ] (no page memory allocated)
 *      counter #1 (16-bit) counter #2 (16-bit)
 *
 */

/* bitmap structures: */

#define BITMAP_MAGIC 0x6d746962

/* use these for bitmap->flags and bitmap->sb->state bit-fields */
enum bitmap_state {
	BITMAP_ACTIVE = 0x001, /* the bitmap is in use */
	BITMAP_STALE  = 0x002  /* the bitmap file is out of date or had -EIO */
};

/* the superblock at the front of the bitmap file -- little endian */
typedef struct bitmap_super_s {
	__u32 magic;        /*  0  BITMAP_MAGIC */
	__u32 version;      /*  4  the bitmap major for now, could change... */
	__u8  uuid[16];     /*  8  128 bit uuid - must match md device uuid */
	__u64 events;       /* 24  event counter for the bitmap (1)*/
	__u64 events_cleared;/*32  event counter when last bit cleared (2) */
	__u64 sync_size;    /* 40  the size of the md device's sync range(3) */
	__u32 state;        /* 48  bitmap state information */
	__u32 chunksize;    /* 52  the bitmap chunk size in bytes */
	__u32 daemon_sleep; /* 56  seconds between disk flushes */
	__u32 write_behind; /* 60  number of outstanding write-behind writes */

	__u8  pad[256 - 64]; /* set to zero */
} bitmap_super_t;

/* Redhat don't like to #include <asm/byteorder.h>, and
 * some time include <linux/byteorder/xxx_endian.h> isn't enough,
 * and there is no standard conversion function so... */
/* And dietlibc doesn't think byteswap is ok, so.. */
/*  #include <byteswap.h> */
#ifndef bswap_16
#define bswap_16(x) (((x) & 0x00ffU) << 8 | \
		     ((x) & 0xff00U) >> 8)
#endif
#ifndef bswap_32
#define bswap_32(x) (((x) & 0x000000ffU) << 24 | \
		     ((x) & 0xff000000U) >> 24 | \
		     ((x) & 0x0000ff00U) << 8  | \
		     ((x) & 0x00ff0000U) >> 8)
#endif
#ifndef bswap_64
#define bswap_64(x) (((x) & 0x00000000000000ffULL) << 56 | \
		     ((x) & 0xff00000000000000ULL) >> 56 | \
		     ((x) & 0x000000000000ff00ULL) << 40 | \
		     ((x) & 0x00ff000000000000ULL) >> 40 | \
		     ((x) & 0x0000000000ff0000ULL) << 24 | \
		     ((x) & 0x0000ff0000000000ULL) >> 24 | \
		     ((x) & 0x00000000ff000000ULL) << 8 | \
		     ((x) & 0x000000ff00000000ULL) >> 8)
#endif

#if !defined(__KLIBC__)
#if BYTE_ORDER == LITTLE_ENDIAN
#define __cpu_to_le16(_x) (_x)
#define __cpu_to_le32(_x) (_x)
#define __cpu_to_le64(_x) (_x)
#define __le16_to_cpu(_x) (_x)
#define __le32_to_cpu(_x) (_x)
#define __le64_to_cpu(_x) (_x)

#define __cpu_to_be16(_x) bswap_16(_x)
#define __cpu_to_be32(_x) bswap_32(_x)
#define __cpu_to_be64(_x) bswap_64(_x)
#define __be16_to_cpu(_x) bswap_16(_x)
#define __be32_to_cpu(_x) bswap_32(_x)
#define __be64_to_cpu(_x) bswap_64(_x)
#elif BYTE_ORDER == BIG_ENDIAN
#define __cpu_to_le16(_x) bswap_16(_x)
#define __cpu_to_le32(_x) bswap_32(_x)
#define __cpu_to_le64(_x) bswap_64(_x)
#define __le16_to_cpu(_x) bswap_16(_x)
#define __le32_to_cpu(_x) bswap_32(_x)
#define __le64_to_cpu(_x) bswap_64(_x)

#define __cpu_to_be16(_x) (_x)
#define __cpu_to_be32(_x) (_x)
#define __cpu_to_be64(_x) (_x)
#define __be16_to_cpu(_x) (_x)
#define __be32_to_cpu(_x) (_x)
#define __be64_to_cpu(_x) (_x)
#else
#  error "unknown endianness."
#endif
#endif /* __KLIBC__ */


/*
 * min()/max()/clamp() macros that also do
 * strict type-checking.. See the
 * "unnecessary" pointer comparison.
 */
#define min(x, y) ({                            \
	typeof(x) _min1 = (x);                  \
	typeof(y) _min2 = (y);                  \
	(void) (&_min1 == &_min2);              \
	_min1 < _min2 ? _min1 : _min2; })

#define max(x, y) ({                            \
	typeof(x) _max1 = (x);                  \
	typeof(y) _max2 = (y);                  \
	(void) (&_max1 == &_max2);              \
	_max1 > _max2 ? _max1 : _max2; })

struct createinfo {
	int     uid;
	int     gid;
	int     autof;
	int     mode;
	int     symlinks;
	struct supertype *supertype;
};

#define Name "mdadm"

/* structures read from config file */
/* List of mddevice names and identifiers
 * Identifiers can be:
 *    uuid=128-hex-uuid
 *    super-minor=decimal-minor-number-from-superblock
 *    devices=comma,separated,list,of,device,names,with,wildcards
 *
 * If multiple fields are present, the intersection of all matching
 * devices is considered
 */
#define UnSet (0xfffe)
struct mddev_ident {
	char    *devname;

	int     uuid_set;
	int     uuid[4];
	char    name[33];

	int super_minor;

	char    *devices;       /* comma separated list of device
				 * names with wild cards
				 */
	int     level;
	int raid_disks;
	int spare_disks;
	struct supertype *st;
	int     autof;          /* 1 for normal, 2 for partitioned */
	char    *spare_group;
	char    *bitmap_file;
	int     bitmap_fd;

	char    *container;     /* /dev/whatever name of container, or
				 * uuid of container.  You would expect
				 * this to be the 'devname' or UUID
				 * of some other entry.
				 */
	char    *member;        /* subarray within a container */

	struct mddev_ident *next;
	union {
		/* fields needed by different users of this structure */
		int assembled;  /* set when assembly succeeds */
	};
};

/* List of device names - wildcards expanded */
struct mddev_dev {
	char *devname;
	int disposition;        /* 'a' for add, 'r' for remove, 'f' for fail.
				 * Not set for names read from .config
				 */
	char writemostly;       /* 1 for 'set writemostly', 2 for 'clear writemostly' */
	char re_add;
	char used;              /* set when used */
	struct mddev_dev *next;
};

typedef struct mapping {
	const char *name;
	int num;
} mapping_t;


struct mdstat_ent {
	char            *dev;
	int             devnum;
	int             active;
	char            *level;
	char            *pattern; /* U or up, _ for down */
	int             percent; /* -1 if no resync */
	int             resync; /* 3 if check, 2 if reshape, 1 if resync, 0 if recovery */
	int             devcnt;
	int             raid_disks;
	char *          metadata_version;
	struct dev_member {
		char                    *name;
		struct dev_member       *next;
	}               *members;
	struct mdstat_ent *next;
};

static struct mdstat_ent *mdstat_read(int hold, int start);
static void free_mdstat(struct mdstat_ent *ms);
static int mddev_busy(int devnum);

struct map_ent {
	struct map_ent *next;
	int     devnum;
	char    metadata[20];
	int     uuid[4];
	int     bad;
	char    *path;
};

/* various details can be requested */
enum sysfs_read_flags {
	GET_LEVEL       = (1 << 0),
	GET_LAYOUT      = (1 << 1),
	GET_COMPONENT   = (1 << 2),
	GET_CHUNK       = (1 << 3),
	GET_CACHE       = (1 << 4),
	GET_MISMATCH    = (1 << 5),
	GET_VERSION     = (1 << 6),
	GET_DISKS       = (1 << 7),
	GET_DEGRADED    = (1 << 8),
	GET_SAFEMODE    = (1 << 9),
	GET_BITMAP_LOCATION = (1 << 10),

	GET_DEVS        = (1 << 20), /* gets role, major, minor */
	GET_OFFSET      = (1 << 21),
	GET_SIZE        = (1 << 22),
	GET_STATE       = (1 << 23),
	GET_ERROR       = (1 << 24),
};

static void sysfs_free(struct mdinfo *sra);
static struct mdinfo *sysfs_read(int fd, int devnum, unsigned long options);
static int sysfs_set_num(struct mdinfo *sra, struct mdinfo *dev,
			 char *name, unsigned long long val);
static int sysfs_uevent(struct mdinfo *sra, char *event);
static int sysfs_fd_get_ll(int fd, unsigned long long *val);

static char *map_dev(int major, int minor, int create);

/* 'struct reshape' records the intermediate states
 * a general reshape.
 * The starting geometry is converted to the 'before' geometry
 * by at most an atomic level change. They could be the same.
 * Similarly the 'after' geometry is converted to the final
 * geometry by at most a level change.
 * Note that 'before' and 'after' must have the same level.
 * 'blocks' is the minimum number of sectors for a reshape unit.
 * This will be a multiple of the stripe size in each of the
 * 'before' and 'after' geometries.
 * If 'blocks' is 0, no restriping is necessary.
 */
struct reshape {
	int level;
	int parity; /* number of parity blocks/devices */
	struct {
		int layout;
		int data_disks;
	} before, after;
	unsigned long long backup_blocks;
	unsigned long long stripes; /* number of old stripes that comprise 'blocks'*/
	unsigned long long new_size; /* New size of array in sectors */
};

/* A superswitch provides entry point the a metadata handler.
 *
 * The super_switch primarily operates on some "metadata" that
 * is accessed via the 'supertype'.
 * This metadata has one of three possible sources.
 * 1/ It is read from a single device.  In this case it may not completely
 *    describe the array or arrays as some information might be on other
 *    devices.
 * 2/ It is read from all devices in a container.  In this case all
 *    information is present.
 *
 * When creating an array inside a container, the metadata will be
 * formed by a combination of 2 and 3.  The metadata or the array is read,
 * then new information is added.
 *
 * The metadata must sometimes have a concept of a 'current' array
 * and a 'current' device.
 *
 * The 'current' device is either the device that the metadata was read from
 * in case 1, or the last device added by add_to_super in case 3.
 * Case 2 does not identify a 'current' device.
 */
struct superswitch {

	/* Used:
	 *   to get uuid to storing in bitmap metadata
	 *   and 'reshape' backup-data metadata
	 *   To see if a device is being re-added to an array it was part of.
	 */
	void (*uuid_from_super)(struct supertype *st, int uuid[4]);

	/* Extract generic details from metadata.  This could be details about
	 * the container, or about an individual array within the container.
	 * The determination is made either by:
	 *   load_super being given a 'component' string.
	 *   validate_geometry determining what to create.
	 * The info includes both array information and device information.
	 * The particular device should be:
	 *   The last device added by add_to_super
	 *   The device the metadata was loaded from by load_super
	 * If 'map' is present, then it is an array raid_disks long
	 * (raid_disk must already be set and correct) and it is filled
	 * with 1 for slots that are thought to be active and 0 for slots which
	 * appear to be failed/missing.
	 * *info is zeroed out before data is added.
	 */
	void (*getinfo_super)(struct supertype *st, struct mdinfo *info, char *map);
	struct mdinfo *(*getinfo_super_disks)(struct supertype *st);
	/* Check if the given metadata is flagged as belonging to "this"
	 * host.  0 for 'no', 1 for 'yes', -1 for "Don't record homehost"
	 */
	int (*match_home)(struct supertype *st, char *homehost);

	/* Make one of several generic modifications to metadata
	 * prior to assembly (or other times).
	 *   sparc2.2  - first bug in early 0.90 metadata
	 *   super-minor - change name of 0.90 metadata
	 *   summaries - 'correct' any redundant data
	 *   resync - mark array as dirty to trigger a resync.
	 *   uuid - set new uuid - only 0.90 or 1.x
	 *   name - change the name of the array (where supported)
	 *   homehost - change which host this array is tied to.
	 *   devicesize - If metadata is at start of device, change recorded
	 *               device size to match actual device size
	 *   byteorder - swap bytes for 0.90 metadata
	 *
	 *   force-one  - mark that device as uptodate, not old or failed.
	 *   force-array - mark array as clean if it would not otherwise
	 *               assemble
	 *   assemble   - not sure how this is different from force-one...
	 *   linear-grow-new - add a new device to a linear array, but don't
	 *                   change the size: so superblock still matches
	 *   linear-grow-update - now change the size of the array.
	 *   writemostly - set the WriteMostly1 bit in the superblock devflags
	 *   readwrite - clear the WriteMostly1 bit in the superblock devflags
	 */
	int (*update_super)(struct supertype *st, struct mdinfo *info,
			    char *update,
			    char *devname, int verbose,
			    int uuid_set, char *homehost);

	/* update the metadata to include new device, either at create or
	 * when hot-adding a spare.
	 */
	int (*add_to_super)(struct supertype *st, mdu_disk_info_t *dinfo,
			    int fd, char *devname);
	/* update the metadata to delete a device,
	 * when hot-removing.
	 */
	int (*remove_from_super)(struct supertype *st, mdu_disk_info_t *dinfo);

	/* Write metadata to one device when fixing problems or adding
	 * a new device.
	 */
	int (*store_super)(struct supertype *st, int fd);

	/*  Write all metadata for this array.
	 */
	int (*write_init_super)(struct supertype *st);
	int (*compare_super)(struct supertype *st, struct supertype *tst);
	int (*load_super)(struct supertype *st, int fd, char *devname);
	int (*load_container)(struct supertype *st, int fd, char *devname);
	struct supertype * (*match_metadata_desc)(char *arg);
	unsigned long long (*min_acceptable_spare_size)(struct supertype *st);
	void (*free_super)(struct supertype *st);

	struct mdinfo *(*container_content)(struct supertype *st, char *subarray);
	/* query the supertype for default geometry */
	void (*default_geometry)(struct supertype *st, int *level, int *layout, int *chunk); /* optional */
	/* Permit subarray's to be deleted from inactive containers */
	int (*kill_subarray)(struct supertype *st); /* optional */
	/* Permit subarray's to be modified */
	int (*update_subarray)(struct supertype *st, char *subarray,
			       char *update, struct mddev_ident *ident); /* optional */
	/*
	 * Return statically allocated string that represents metadata specific
	 * controller domain of the disk. The domain is used in disk domain
	 * matching functions. Disks belong to the same domain if the they have
	 * the same domain from mdadm.conf and belong the same metadata domain.
	 * Returning NULL or not providing this handler means that metadata
	 * does not distinguish the differences between disks that belong to
	 * different controllers. They are in the domain specified by
	 * configuration file (mdadm.conf).
	 * In case when the metadata has the notion of domains based on disk
	 * it shall return NULL for disks that do not belong to the controller
	 * the supported domains. Such disks will form another domain and won't
	 * be mixed with supported ones.
	 */
	const char *(*get_disk_controller_domain)(const char *path);

	int swapuuid; /* true if uuid is bigending rather than hostendian */
	int external;
	const char *name; /* canonical metadata name */
};

static const struct superswitch super0, super1, super_imsm, super_ddf, mbr, gpt;

static const struct superswitch * const superlist[] =
{
	&super0, &super1,
	&super_ddf, &super_imsm,
	&mbr, &gpt,
	NULL
};


/* A supertype holds a particular collection of metadata.
 * It identifies the metadata type by the superswitch, and the particular
 * sub-version of that metadata type.
 * metadata read in or created is stored in 'sb' and 'info'.
 * There are also fields used by mdmon to track containers.
 *
 * A supertype may refer to:
 *   Just an array, possibly in a container
 *   A container, not identifying any particular array
 *   Info read from just one device, not yet fully describing the array/container.
 *
 *
 * A supertype is created by:
 *   super_by_fd
 *   guess_super
 *   dup_super
 */
struct supertype {
	const struct superswitch *ss;
	int minor_version;
	int max_devs;
	int container_dev;    /* devnum of container */
	void *sb;
	void *info;
	int ignore_hw_compat; /* used to inform metadata handlers that it should ignore
				 HW/firmware related incompatability to load metadata.
				 Used when examining metadata to display content of disk
				 when user has no hw/firmare compatible system.
			      */
	struct metadata_update *updates;
	struct metadata_update **update_tail;

	/* extra stuff used by mdmon */
	struct active_array *arrays;
	int sock; /* listen to external programs */
	int devnum;
	char *devname; /* e.g. md0.  This appears in metadata_verison:
			*  external:/md0/12
			*/
	int devcnt;
	int retry_soon;

	struct mdinfo *devs;

};

enum guess_types { guess_any, guess_array, guess_partitions };
static struct supertype *guess_super_type(int fd, enum guess_types guess_type);
static inline struct supertype *guess_super(int fd) {
	return guess_super_type(fd, guess_any);
}
static struct supertype *dup_super(struct supertype *st);
static int get_dev_size(int fd, char *dname, unsigned long long *sizep);
static int must_be_container(int fd);
static void wait_for(char *dev, int fd);

/*
 * Data structures for policy management.
 * Each device can have a policy structure that lists
 * various name/value pairs each possibly with a metadata associated.
 * The policy list is sorted by name/value/metadata
 */
struct dev_policy {
	struct dev_policy *next;
	const char *name; /* None of these strings are allocated.  They are
			 * all just references to strings which are known
			 * to exist elsewhere.
			 * name and metadata can be compared by address equality.
			 */
	const char *metadata;
	const char *value;
};

static const char pol_act[] = "action";
static const char rule_path[] = "path";
static const char rule_type[] = "type";

static const char rule_policy[] = "policy";
static const char rule_part[] = "part-policy";

static const char pol_metadata[] = "metadata";
static const char pol_domain[] = "domain";
static const char pol_auto[] = "auto";

/* iterate over the sublist starting at list, having the same
 * 'name' as 'list', and matching the given metadata (Where
 * NULL matches anything
 */
#define pol_for_each(item, list, _metadata)                             \
	for (item = list;                                               \
	     item && item->name == list->name;                          \
	     item = item->next)                                         \
		if (!(!_metadata || !item->metadata || _metadata == item->metadata)) \
			; else

/* policy records read from mdadm are largely just name-value pairs.
 * The names are constants, not strdupped */
struct pol_rule {
	struct pol_rule *next;
	const char *type;     /* rule_policy or rule_part */
	struct rule {
		struct rule *next;
		const char *name;
		char *value;
		char *dups; /* duplicates of 'value' with a partNN appended */
	} *rule;
};

static void policyline(char *line, const char *type);
static void policy_add(const char *type, ...);

static struct dev_policy *disk_policy(struct mdinfo *disk);
static struct dev_policy *devnum_policy(int dev);
static void dev_policy_free(struct dev_policy *p);

static struct dev_policy *pol_find(struct dev_policy *pol, const char *name);

enum policy_action {
	act_default,
	act_include,
	act_re_add,
	act_spare,      /* This only applies to bare devices */
	act_spare_same_slot, /* this allows non-bare devices,
			      * but only if recent removal */
	act_force_spare, /* this allow non-bare devices in any case */
	act_err
};

static int policy_action_allows(struct dev_policy *plist, const char *metadata,
				enum policy_action want);
static int disk_action_allows(struct mdinfo *disk, const char *metadata,
			      enum policy_action want);

struct domainlist {
	struct domainlist *next;
	const char *dom;
};

static int domain_test(struct domainlist *dom, struct dev_policy *pol,
		       const char *metadata);
static void domain_free(struct domainlist *dl);
static void domain_merge(struct domainlist **domp, struct dev_policy *pol,
			 const char *metadata);

static int md_get_version(int fd);
static int parse_uuid(char *str, int uuid[4]);

static int get_mdp_major(void);
static int dev_open(char *dev, int flags);
static int same_dev(char *one, char *two);

static int parse_auto(char *str, char *msg, int config);
static struct mddev_dev *conf_get_devs(void);
static int conf_test_metadata(const char *version, struct dev_policy *pol, int is_homehost);
static char *conf_line(FILE *file);
static char *conf_word(FILE *file, int allow_key);
static int conf_name_is_free(char *name);
static int devname_matches(char *name, char *match);

static void free_line(char *line);
static int match_oneof(char *devices, char *devname);

static const int uuid_zero[4] = { 0, 0, 0, 0 };

static int same_uuid(int a[4], int b[4], int swapuuid);
static void copy_uuid(void *a, int b[4], int swapuuid);
static unsigned long calc_csum(void *super, int bytes);
static int enough(int level, int raid_disks, int layout, int clean,
		   char *avail);
static void remove_partitions(int fd);
static int test_partition(int fd);
static int add_disk(int mdfd, struct mdinfo *info);
static int set_array_info(int mdfd, struct mdinfo *info);

#define NoMdDev (1<<23)

/* values for 'trustworthy' */
#define LOCAL   1
#define LOCAL_ANY 10
#define FOREIGN 2
#define METADATA 3

static int check_env(char *name);

static char *devnum2devname(int num);
static void fmt_devname(char *name, int num);
static int devname2devnum(char *name);
static int stat2devnum(struct stat *st);
static int fd2devnum(int fd);

static inline int dev2major(int d)
{
	if (d >= 0)
		return MD_MAJOR;
	else
		return get_mdp_major();
}

static inline int dev2minor(int d)
{
	if (d >= 0)
		return d;
	return (-1-d) << MdpMinorShift;
}

#define _ROUND_UP(val, base)    (((val) + (base) - 1) & ~(base - 1))
#define ROUND_UP(val, base)     _ROUND_UP(val, (typeof(val))(base))
#define ROUND_UP_PTR(ptr, base) ((typeof(ptr)) \
				 (ROUND_UP((unsigned long)(ptr), base)))

static inline int is_subarray(char *vers)
{
	/* The version string for a 'subarray' (an array in a container)
	 * is
	 *    /containername/componentname    for normal read-write arrays
	 *    -containername/componentname    for arrays which mdmon must not
	 *                                    reconfigure.  They might be read-only
	 *                                    or might be undergoing reshape etc.
	 * containername is e.g. md0, md_d1
	 * componentname is dependant on the metadata. e.g. '1' 'S1' ...
	 */
	return (*vers == '/' || *vers == '-');
}

#ifdef DEBUG
#define dprintf(fmt, arg...) \
	fprintf(stderr, fmt, ##arg)
#else
#define dprintf(fmt, arg...) \
	({ if (0) fprintf(stderr, fmt, ##arg); 0; })
#endif

#define LEVEL_MULTIPATH         (-4)
#define LEVEL_LINEAR            (-1)
#define LEVEL_FAULTY            (-5)

/* kernel module doesn't know about these */
#define LEVEL_CONTAINER         (-100)
#define LEVEL_UNSUPPORTED       (-200)


/* faulty stuff */

#define WriteTransient  0
#define ReadTransient   1
#define WritePersistent 2
#define ReadPersistent  3
#define WriteAll        4 /* doesn't go to device */
#define ReadFixable     5
#define Modes   6

#define ClearErrors     31
#define ClearFaults     30

#define AllPersist      100 /* internal use only */
#define NoPersist       101

#define ModeMask        0x1f
#define ModeShift       5


#ifdef __TINYC__
#undef minor
#undef major
#undef makedev
#define minor(x) ((x)&0xff)
#define major(x) (((x)>>8)&0xff)
#define makedev(M,m) (((M)<<8) | (m))
#endif

/* for raid4/5/6 */
#define ALGORITHM_LEFT_ASYMMETRIC       0
#define ALGORITHM_RIGHT_ASYMMETRIC      1
#define ALGORITHM_LEFT_SYMMETRIC        2
#define ALGORITHM_RIGHT_SYMMETRIC       3

/* Define non-rotating (raid4) algorithms.  These allow
 * conversion of raid4 to raid5.
 */
#define ALGORITHM_PARITY_0              4 /* P or P,Q are initial devices */
#define ALGORITHM_PARITY_N              5 /* P or P,Q are final devices. */

/* DDF RAID6 layouts differ from md/raid6 layouts in two ways.
 * Firstly, the exact positioning of the parity block is slightly
 * different between the 'LEFT_*' modes of md and the "_N_*" modes
 * of DDF.
 * Secondly, or order of datablocks over which the Q syndrome is computed
 * is different.
 * Consequently we have different layouts for DDF/raid6 than md/raid6.
 * These layouts are from the DDFv1.2 spec.
 * Interestingly DDFv1.2-Errata-A does not specify N_CONTINUE but
 * leaves RLQ=3 as 'Vendor Specific'
 */

#define ALGORITHM_ROTATING_ZERO_RESTART 8 /* DDF PRL=6 RLQ=1 */
#define ALGORITHM_ROTATING_N_RESTART    9 /* DDF PRL=6 RLQ=2 */
#define ALGORITHM_ROTATING_N_CONTINUE   10 /*DDF PRL=6 RLQ=3 */


/* For every RAID5 algorithm we define a RAID6 algorithm
 * with exactly the same layout for data and parity, and
 * with the Q block always on the last device (N-1).
 * This allows trivial conversion from RAID5 to RAID6
 */
#define ALGORITHM_LEFT_ASYMMETRIC_6     16
#define ALGORITHM_RIGHT_ASYMMETRIC_6    17
#define ALGORITHM_LEFT_SYMMETRIC_6      18
#define ALGORITHM_RIGHT_SYMMETRIC_6     19
#define ALGORITHM_PARITY_0_6            20
#define ALGORITHM_PARITY_N_6            ALGORITHM_PARITY_N

/* Define PATH_MAX in case we don't use glibc or standard library does
 * not have PATH_MAX defined. Assume max path length is 4K characters.
 */
#ifndef PATH_MAX
#define PATH_MAX        4096
#endif

#define RESYNC_NONE -1
#define RESYNC_DELAYED -2
#define RESYNC_PENDING -3
#define RESYNC_UNKNOWN -4

/* When using "GET_DISK_INFO" it isn't certain how high
 * we need to check.  So we impose an absolute limit of
 * MAX_DISKS.  This needs to be much more than the largest
 * number of devices any metadata can support.  Currently
 * v1.x can support 1920
 */
#define MAX_DISKS       4096

static void probe_roms_exit(void);
static int probe_roms_init(unsigned long align);
typedef int (*scan_fn)(const void *start, const void *end, const void *data);
static int scan_adapter_roms(scan_fn fn);
static void probe_roms(void);

struct conf_dev {
    struct conf_dev *next;
    char *name;
};

struct devmap {
    int major, minor;
    char *name;
    struct devmap *next;
};

struct resource {
	unsigned long start;
	unsigned long end;
	unsigned long data;
	const char *name;
};

#define ARR_SIZE 6
static const struct resource adapter_rom_resources_st[ARR_SIZE] = { {
	.name   = "Adapter ROM",
	.start  = 0xc8000,
	.data   = 0,
	.end    = 0,
}, {
	.name   = "Adapter ROM",
	.start  = 0,
	.data   = 0,
	.end    = 0,
}, {
	.name   = "Adapter ROM",
	.start  = 0,
	.data   = 0,
	.end    = 0,
}, {
	.name   = "Adapter ROM",
	.start  = 0,
	.data   = 0,
	.end    = 0,
}, {
	.name   = "Adapter ROM",
	.start  = 0,
	.data   = 0,
	.end    = 0,
}, {
	.name   = "Adapter ROM",
	.start  = 0,
	.data   = 0,
	.end    = 0,
} };


CALC_BSS struct BB_COMMON_BSS
{
	int _sigbus;
	unsigned long rom_align;
	char abuf[4096+4096];
	char major_minor_buf[30];
	struct conf_dev *cdevlist;
	struct pol_rule *config_rules;
	int config_rules_has_path;
	struct devmap *devmap_devlist;
	int devlist_ready;
	struct mddev_dev *mddev_dev_dlist;
	struct imsm_orom g_imsm_orom[SYS_DEV_MAX];
	int populated_orom[SYS_DEV_MAX];
	int populated_efi[SYS_DEV_MAX];
	struct imsm_orom imsm_efi[SYS_DEV_MAX];
	struct sys_dev *intel_devices;
	int loaded;
	struct mddev_ident *mddevlist;
	struct mddev_ident **mddevlp; /* = &mddevlist */
	const char *prev_metadata;
	const char *conffile;
	int mdp_major;
	int mdstat_fd;
	int rom_fd;
	void *rom_mem;
	struct resource adapter_rom_resources[ARR_SIZE];
	struct resource video_rom_resource;
} bb_common_bss;
#define _sigbus bb_common_bss._sigbus
#define rom_align bb_common_bss.rom_align
#define abuf bb_common_bss.abuf
#define major_minor_buf bb_common_bss.major_minor_buf
#define cdevlist bb_common_bss.cdevlist
#define config_rules bb_common_bss.config_rules
#define config_rules_has_path bb_common_bss.config_rules_has_path
#define devmap_devlist bb_common_bss.devmap_devlist
#define devlist_ready bb_common_bss.devlist_ready
#define mddev_dev_dlist bb_common_bss.mddev_dev_dlist
#define g_imsm_orom bb_common_bss.g_imsm_orom
#define populated_orom bb_common_bss.populated_orom
#define populated_efi bb_common_bss.populated_efi
#define imsm_efi bb_common_bss.imsm_efi
#define intel_devices bb_common_bss.intel_devices
#define loaded bb_common_bss.loaded
#define mddevlist bb_common_bss.mddevlist
#define mddevlp bb_common_bss.mddevlp
#define prev_metadata bb_common_bss.prev_metadata
#define conffile bb_common_bss.conffile
#define mdp_major bb_common_bss.mdp_major
#define mdstat_fd bb_common_bss.mdstat_fd
#define rom_fd bb_common_bss.rom_fd
#define rom_mem bb_common_bss.rom_mem
#define adapter_rom_resources bb_common_bss.adapter_rom_resources
#define video_rom_resource bb_common_bss.video_rom_resource


#ifndef HAVE_POSIX_MEMALIGN
int posix_memalign(void **res, size_t align, size_t len)
{
	if (align < sizeof(void *)) return EINVAL;
	void *mem = memalign(align, len);
	if (!mem) return errno;
	*res = mem;
	return 0;
}
#endif

/* Compute SHA1 message digest for LEN bytes beginning at BUFFER.  The
   result is always in little endian byte order, so that a byte-wise
   output yields to the wanted ASCII representation of the message
   digest.  */
static void *sha1_buffer (const char *buffer, size_t len, void *resblock)
{
  sha1_ctx_t ctx;

  /* Initialize the computation context.  */
  sha1_begin(&ctx);

  /* Process whole buffer but last len % 64 bytes.  */
  sha1_hash (buffer, len, &ctx);

  /* Put result in desired memory area.  */
  return sha1_end (resblock, &ctx);
}


/* from mdopen.c */
static int open_mddev(const char *dev)
{
	int mdfd = bb_xopen(dev, O_RDWR | BB_OPEN_WARN);
	if (mdfd >= 0 && md_get_version(mdfd) <= 0) {
		bb_error_msg("%s does not appear to be an md device", dev);
		close(mdfd);
		mdfd = -1;
	}
	return mdfd;
}

static inline int create_mddev(const char *dev)
{
	return open_mddev(dev);
}

static const char DefaultConfFile[] = CONFFILE;
static const char DefaultAltConfFile[] = CONFFILE2;



static int name_matches(char *found, char *required, char *homehost)
{
	/* See if the name found matches the required name, possibly
	 * prefixed with 'homehost'
	 */
	char fnd[33];

	strncpy(fnd, found, 32);
	fnd[32] = 0;
	if (strcmp(found, required)==0)
		return 1;
	if (homehost) {
		int l = strlen(homehost);
		if (l < 32 && fnd[l] == ':' &&
		    strcmp(fnd+l+1, required)==0)
			return 1;
	}
	return 0;
}

static int is_member_busy(char *metadata_version)
{
	/* check if the given member array is active */
	struct mdstat_ent *mdstat = mdstat_read(1, 0);
	struct mdstat_ent *ent;
	int busy = 0;

	for (ent = mdstat; ent; ent = ent->next) {
		if (ent->metadata_version == NULL)
			continue;
		if (strncmp(ent->metadata_version, "external:", 9) != 0)
			continue;
		if (!is_subarray(&ent->metadata_version[9]))
			continue;
		/* Skip first char - it can be '/' or '-' */
		if (strcmp(&ent->metadata_version[10], metadata_version+1) == 0) {
			busy = 1;
			break;
		}
	}
	free_mdstat(mdstat);

	return busy;
}

static int ident_matches(struct mddev_ident *ident,
			 struct mdinfo *content,
			 struct supertype *tst,
			 char *homehost,
			 char *update, char *devname)
{

	if (ident->uuid_set && (!update || strcmp(update, "uuid")!= 0) &&
	    same_uuid(content->uuid, ident->uuid, tst->ss->swapuuid)==0 &&
	    memcmp(content->uuid, uuid_zero, sizeof(int[4])) != 0) {
		if (devname)
			bb_error_msg("%s has wrong uuid.", devname);
		return 0;
	}
	if (ident->name[0] && (!update || strcmp(update, "name")!= 0) &&
	    name_matches(content->name, ident->name, homehost)==0) {
		if (devname)
			bb_error_msg("%s has wrong name.", devname);
		return 0;
	}
	if (ident->super_minor != UnSet &&
	    ident->super_minor != content->array.md_minor) {
		if (devname)
			bb_error_msg("%s has wrong super-minor.", devname);
		return 0;
	}
	if (ident->level != UnSet &&
	    ident->level != content->array.level) {
		if (devname)
			bb_error_msg("%s has wrong raid level.", devname);
		return 0;
	}
	if (ident->raid_disks != UnSet &&
	    ident->raid_disks!= content->array.raid_disks) {
		if (devname)
			bb_error_msg("%s requires wrong number of drives.", devname);
		return 0;
	}
	if (ident->member && ident->member[0]) {
		/* content->text_version must match */
		char *s = strchr(content->text_version+1, '/');
		if (s == NULL) {
			if (devname)
				bb_error_msg("%s is not a container and one is required.",
					devname);
			return 0;
		} else if (strcmp(ident->member, s+1) != 0) {
			if (devname)
				bb_error_msg("skipping wrong member %s is %s",
					content->text_version, devname);
			return 0;
		}
	}
	return 1;
}


static int Assemble(struct supertype *st, char *mddev,
	     struct mddev_ident *ident,
	     struct mddev_dev *devlist,
	     int runstop, char *update, char *homehost, int require_homehost,
	     int verbose, int force)
{
	/*
	 * The task of Assemble is to find a collection of
	 * devices that should (according to their superblocks)
	 * form an array, and to give this collection to the MD driver.
	 * In Linux-2.4 and later, this involves submitting a
	 * SET_ARRAY_INFO ioctl with no arg - to prepare
	 * the array - and then submit a number of
	 * ADD_NEW_DISK ioctls to add disks into
	 * the array.  Finally RUN_ARRAY might
	 * be submitted to start the array.
	 *
	 * Much of the work of Assemble is in finding and/or
	 * checking the disks to make sure they look right.
	 *
	 * If mddev is not set, then scan must be set and we
	 *  read through the config file for dev+uuid mapping
	 *  We recurse, setting mddev, for each device that
	 *    - isn't running
	 *    - has a valid uuid (or any uuid if !uuidset)
	 *
	 * If mddev is set, we try to determine state of md.
	 *   check version - must be at least 0.90.0
	 *   check kernel version.  must be at least 2.4.
	 *    If not, we can possibly fall back on START_ARRAY
	 *   Try to GET_ARRAY_INFO.
	 *     If possible, give up
	 *     If not, try to STOP_ARRAY just to make sure
	 *
	 * If !uuidset and scan, look in conf-file for uuid
	 *       If not found, give up
	 * If !devlist and scan and uuidset, get list of devs from conf-file
	 *
	 * For each device:
	 *   Check superblock - discard if bad
	 *   Check uuid (set if we don't have one) - discard if no match
	 *   Check superblock similarity if we have a superblock - discard if different
	 *   Record events, devicenum
	 * This should give us a list of devices for the array
	 * We should collect the most recent event number
	 *
	 * Count disks with recent enough event count
	 * While force && !enough disks
	 *    Choose newest rejected disks, update event count
	 *     mark clean and rewrite superblock
	 * If recent kernel:
	 *    SET_ARRAY_INFO
	 *    foreach device with recent events : ADD_NEW_DISK
	 *    if runstop == 1 || "enough" disks and runstop==0 -> RUN_ARRAY
	 * If old kernel:
	 *    Check the device numbers in superblock are right
	 *    update superblock if any changes
	 *    START_ARRAY
	 *
	 */
	int mdfd;
	int clean;
	int auto_assem = (mddev == NULL && !ident->uuid_set &&
			  ident->super_minor == UnSet && ident->name[0] == 0
			  && (ident->container == NULL || ident->member == NULL));
	int vers = vers; /* Keep gcc quite - it really is initialised */
	struct {
		char *devname;
		int uptodate; /* set once we decide that this device is as
			       * recent as everything else in the array.
			       */
		struct mdinfo i;
	} *devices;
	char *devmap;
	int *best = NULL; /* indexed by raid_disk */
	int bestcnt = 0;
	int devcnt = 0;
	unsigned int okcnt, sparecnt, rebuilding_cnt;
	unsigned int req_cnt;
	int i;
	int most_recent = 0;
	int chosen_drive;
	int change = 0;
	int inargv = 0;
	int report_missmatch;
	int start_partial_ok = (runstop >= 0) &&
		(force || devlist==NULL || auto_assem);
	unsigned int num_devs;
	struct mddev_dev *tmpdev;
	struct mdinfo info;
	struct mdinfo *content = NULL;
	char *avail;
	int nextspare = 0;
	char *name = NULL;
	int trustworthy;
	char chosen_name[1024];
	struct domainlist *domains = NULL;
	int rv;

	/*
	 * If any subdevs are listed, then any that don't
	 * match ident are discarded.  Remainder must all match and
	 * become the array.
	 * If no subdevs, then we scan all devices in the config file, but
	 * there must be something in the identity
	 */

	if (!devlist &&
	    ident->uuid_set == 0 &&
	    (ident->super_minor < 0 || ident->super_minor == UnSet) &&
	    ident->name[0] == 0 &&
	    (ident->container == NULL || ident->member == NULL) &&
	    ident->devices == NULL) {
		bb_error_msg("No identity information available for %s - cannot assemble.",
			mddev ? mddev : "further assembly");
		return 1;
	}

	if (devlist == NULL)
		devlist = conf_get_devs();
	else if (mddev)
		inargv = 1;

	report_missmatch = ((inargv && verbose >= 0) || verbose > 0);
 try_again:
	/* We come back here when doing auto-assembly and attempting some
	 * set of devices failed.  Those are now marked as ->used==2 and
	 * we ignore them and try again
	 */

	tmpdev = devlist; num_devs = 0;
	while (tmpdev) {
		if (tmpdev->used)
			tmpdev->used = 2;
		else
			num_devs++;
		tmpdev = tmpdev->next;
	}

	if (!st && ident->st) st = ident->st;

	if (verbose>0)
	    bb_error_msg("looking for devices for %s",
		    mddev ? mddev : "further assembly");

	/* first walk the list of devices to find a consistent set
	 * that match the criterea, if that is possible.
	 * We flag the ones we like with 'used'.
	 */
	for (tmpdev = devlist;
	     tmpdev;
	     tmpdev = tmpdev ? tmpdev->next : NULL) {
		char *devname = tmpdev->devname;
		int dfd;
		struct stat stb;
		struct supertype *tst;
		struct dev_policy *pol = NULL;
		int found_container = 0;

		if (tmpdev->used > 1) continue;

		if (ident->devices &&
		    !match_oneof(ident->devices, devname)) {
			if (report_missmatch)
				bb_error_msg("%s is not one of %s", devname, ident->devices);
			continue;
		}

		tst = dup_super(st);

		dfd = dev_open(devname, O_RDONLY);
		if (dfd < 0) {
			if (report_missmatch)
				bb_perror_msg("cannot open device %s", devname);
			tmpdev->used = 2;
		} else if (fstat(dfd, &stb)< 0) {
			/* Impossible! */
			bb_perror_msg("fstat failed for %s", devname);
			tmpdev->used = 2;
		} else if ((stb.st_mode & S_IFMT) != S_IFBLK) {
			bb_error_msg("%s is not a block device.", devname);
			tmpdev->used = 2;
		} else if (must_be_container(dfd)) {
			if (st) {
				/* already found some components, this cannot
				 * be another one.
				 */
				if (report_missmatch)
					bb_error_msg("%s is a container, but we are looking for components",
						devname);
				tmpdev->used = 2;
			} else if (!tst->ss->load_container
				   || tst->ss->load_container(tst, dfd, NULL)) {
				if (report_missmatch)
					bb_error_msg("no correct container type: %s", devname);
				tmpdev->used = 2;
			} else if (auto_assem &&
				   !conf_test_metadata(tst->ss->name, (pol = devnum_policy(stb.st_rdev)),
						       tst->ss->match_home(tst, homehost) == 1)) {
				if (report_missmatch)
					bb_error_msg("%s has metadata type %s for which "
						"auto-assembly is disabled",
						devname, tst->ss->name);
				tmpdev->used = 2;
			} else
				found_container = 1;
		} else {
			if (!tst && (tst = guess_super(dfd)) == NULL) {
				if (report_missmatch)
					bb_error_msg("no recogniseable superblock on %s",
						devname);
				tmpdev->used = 2;
			} else if (tst->ss->load_super(tst,dfd, NULL)) {
				if (report_missmatch)
					bb_error_msg("no RAID superblock on %s",
						devname);
				tmpdev->used = 2;
			} else if (tst->ss->compare_super == NULL) {
				if (report_missmatch)
					bb_error_msg("Cannot assemble %s metadata on %s",
						tst->ss->name, devname);
				tmpdev->used = 2;
			} else if (auto_assem && st == NULL &&
				   !conf_test_metadata(tst->ss->name, (pol = devnum_policy(stb.st_rdev)),
						       tst->ss->match_home(tst, homehost) == 1)) {
				if (report_missmatch)
					bb_error_msg("%s has metadata type %s for which "
						"auto-assembly is disabled",
						devname, tst->ss->name);
				tmpdev->used = 2;
			}
		}
		if (dfd >= 0) close(dfd);
		if (tmpdev->used == 2) {
			if (auto_assem || !inargv)
				/* Ignore unrecognised devices during auto-assembly */
				goto loop;
			if (ident->uuid_set || ident->name[0] ||
			    ident->super_minor != UnSet)
				/* Ignore unrecognised device if looking for
				 * specific array */
				goto loop;


			bb_error_msg("%s has no superblock - assembly aborted",
				devname);
			if (st)
				st->ss->free_super(st);
			dev_policy_free(pol);
			domain_free(domains);
			return 1;
		}

		if (found_container) {
			/* tmpdev is a container.  We need to be either
			 * looking for a member, or auto-assembling
			 */
			/* should be safe to try an exclusive open now, we
			 * have rejected anything that some other mdadm might
			 * be looking at
			 */
			dfd = dev_open(devname, O_RDONLY | O_EXCL);
			if (dfd < 0) {
				if (report_missmatch)
					bb_error_msg("%s is busy - skipping", devname);
				goto loop;
			}
			close(dfd);

			if (ident->container) {
				if (ident->container[0] == '/' &&
				    !same_dev(ident->container, devname)) {
					if (report_missmatch)
						bb_error_msg("%s is not the container required (%s)",
							devname, ident->container);
					goto loop;
				}
				if (ident->container[0] != '/') {
					/* we have a uuid */
					int uuid[4];

					content = &info;
					tst->ss->getinfo_super(tst, content, NULL);

					if (!parse_uuid(ident->container, uuid) ||
					    !same_uuid(content->uuid, uuid, tst->ss->swapuuid)) {
						if (report_missmatch)
							bb_error_msg("%s has wrong UUID to be required container",
								devname);
						goto loop;
					}
				}
			}
			/* It is worth looking inside this container.
			 */
			if (verbose > 0)
				bb_error_msg("looking in container %s",
					devname);

			for (content = tst->ss->container_content(tst, NULL);
			     content;
			     content = content->next) {

				if (!ident_matches(ident, content, tst,
						   homehost, update,
						   report_missmatch ? devname : NULL))
					/* message already printed */;
				else if (is_member_busy(content->text_version)) {
					if (report_missmatch)
						bb_error_msg("member %s in %s is already assembled",
							content->text_version,
							devname);
				} else if (content->array.state & (1<<MD_SB_BLOCK_VOLUME)) {
					/* do not assemble arrays with unsupported configurations */
					bb_error_msg("Cannot activate member %s in %s.",
						content->text_version, devname);
				} else
					break;
			}
			if (!content) {
				tmpdev->used = 2;
				goto loop; /* empty container */
			}

			st = tst; tst = NULL;
			if (!auto_assem && inargv && tmpdev->next != NULL) {
				bb_error_msg("%s is a container, but is not "
					"only device given: confused and aborting",
					devname);
				st->ss->free_super(st);
				dev_policy_free(pol);
				domain_free(domains);
				return 1;
			}
			if (verbose > 0)
				bb_error_msg("found match on member %s in %s",
					content->text_version, devname);

			/* make sure we finished the loop */
			tmpdev = NULL;
			goto loop;
		} else {

			content = &info;
			tst->ss->getinfo_super(tst, content, NULL);

			if (!ident_matches(ident, content, tst,
					   homehost, update,
					   report_missmatch ? devname : NULL))
				goto loop;

			/* should be safe to try an exclusive open now, we
			 * have rejected anything that some other mdadm might
			 * be looking at
			 */
			dfd = dev_open(devname, O_RDONLY | O_EXCL);
			if (dfd < 0) {
				if (report_missmatch)
					bb_error_msg("%s is busy - skipping", devname);
				goto loop;
			}
			close(dfd);

			if (st == NULL)
				st = dup_super(tst);
			if (st->minor_version == -1)
				st->minor_version = tst->minor_version;

			if (memcmp(content->uuid, uuid_zero,
				   sizeof(int[4])) == 0) {
				/* this is a floating spare.  It cannot define
				 * an array unless there are no more arrays of
				 * this type to be found.  It can be included
				 * in an array of this type though.
				 */
				tmpdev->used = 3;
				goto loop;
			}

			if (st->ss != tst->ss ||
			    st->minor_version != tst->minor_version ||
			    st->ss->compare_super(st, tst) != 0) {
				/* Some mismatch. If exactly one array matches this host,
				 * we can resolve on that one.
				 * Or, if we are auto assembling, we just ignore the second
				 * for now.
				 */
				if (auto_assem)
					goto loop;
				if (homehost) {
					int first = st->ss->match_home(st, homehost);
					int last = tst->ss->match_home(tst, homehost);
					if (first != last &&
					    (first == 1 || last == 1)) {
						/* We can do something */
						if (first) {/* just ignore this one */
							if (report_missmatch)
								bb_error_msg("%s misses out due to wrong homehost",
									devname);
							goto loop;
						} else { /* reject all those sofar */
							struct mddev_dev *td;
							if (report_missmatch)
								bb_error_msg("%s overrides previous devices due to good homehost",
									devname);
							for (td=devlist; td != tmpdev; td=td->next)
								if (td->used == 1)
									td->used = 0;
							tmpdev->used = 1;
							goto loop;
						}
					}
				}
				bb_error_msg("superblock on %s doesn't match others - assembly aborted",
					devname);
				tst->ss->free_super(tst);
				st->ss->free_super(st);
				dev_policy_free(pol);
				domain_free(domains);
				return 1;
			}
			tmpdev->used = 1;
		}
	loop:
		/* Collect domain information from members only */
		if (tmpdev && tmpdev->used == 1) {
			if (!pol)
				pol = devnum_policy(stb.st_rdev);
			domain_merge(&domains, pol, tst?tst->ss->name:NULL);
		}
		dev_policy_free(pol);
		pol = NULL;
		if (tst)
			tst->ss->free_super(tst);
	}

	/* Check if we found some imsm spares but no members */
	if ((auto_assem ||
	     (ident->uuid_set &&
	      memcmp(uuid_zero, ident->uuid,sizeof(uuid_zero)) == 0)) &&
	    (!st || !st->sb))
		for (tmpdev = devlist; tmpdev; tmpdev = tmpdev->next) {
			if (tmpdev->used != 3)
				continue;
			tmpdev->used = 1;
			content = &info;

			if (!st->sb) {
				/* we need sb from one of the spares */
				int dfd = dev_open(tmpdev->devname, O_RDONLY);
				if (dfd < 0 ||
				    st->ss->load_super(st, dfd, NULL))
					tmpdev->used = 2;
				if (dfd > 0)
					close(dfd);
			}
		}

	/* Now reject spares that don't match domains of identified members */
	for (tmpdev = devlist; tmpdev; tmpdev = tmpdev->next) {
		struct stat stb;
		if (tmpdev->used != 3)
			continue;
		if (stat(tmpdev->devname, &stb)< 0) {
			bb_perror_msg("fstat failed for %s", tmpdev->devname);
			tmpdev->used = 2;
		} else {
			struct dev_policy *pol = devnum_policy(stb.st_rdev);
			int dt = domain_test(domains, pol, NULL);
			if (inargv && dt != 0)
				/* take this spare as domains match
				 * if there are any */
				tmpdev->used = 1;
			else if (!inargv && dt == 1)
				/* device wasn't explicitly listed, so need
				 * explicit domain match - which we have */
				tmpdev->used = 1;
			else
				/* if domains don't match mark as unused */
				tmpdev->used = 0;
			dev_policy_free(pol);
		}
	}
	domain_free(domains);

	if (!st || !st->sb || !content)
		return 2;

	/* Now need to open the array device.  Use create_mddev */
	if (content == &info)
		st->ss->getinfo_super(st, content, NULL);

	trustworthy = FOREIGN;
	name = content->name;
	switch (st->ss->match_home(st, homehost)
		?: st->ss->match_home(st, "any")) {
	case 1:
		trustworthy = LOCAL;
		name = strchr(content->name, ':');
		if (name)
			name++;
		else
			name = content->name;
		break;
	}
	if (!auto_assem)
		/* If the array is listed in mdadm.conf or on
		 * command line, then we trust the name
		 * even if the array doesn't look local
		 */
		trustworthy = LOCAL;

	if (name[0] == 0 &&
	    content->array.level == LEVEL_CONTAINER) {
		name = content->text_version;
		trustworthy = METADATA;
	}

	if (name[0] && trustworthy != LOCAL &&
	    ! require_homehost &&
	    conf_name_is_free(name))
		trustworthy = LOCAL;

	if (trustworthy == LOCAL &&
	    strchr(name, ':'))
		/* Ignore 'host:' prefix of name */
		name = strchr(name, ':')+1;

	mdfd = create_mddev(mddev);
	if (mdfd < 0) {
		st->ss->free_super(st);
		if (auto_assem)
			goto try_again;
		return 1;
	}
	mddev = chosen_name;
	vers = md_get_version(mdfd);
	if (vers < 9000) {
		bb_error_msg("Assemble requires driver version 0.90.0 or later.\n"
			"    Upgrade your kernel or try --build");
		close(mdfd);
		return 1;
	}
	if (mddev_busy(fd2devnum(mdfd))) {
		bb_error_msg("%s already active, cannot restart it!", mddev);
		for (tmpdev = devlist ;
		     tmpdev && tmpdev->used != 1;
		     tmpdev = tmpdev->next)
			;
		if (tmpdev && auto_assem)
			bb_error_msg("  %s needed for %s...",
				mddev, tmpdev->devname);
		close(mdfd);
		mdfd = -3;
		st->ss->free_super(st);
		if (auto_assem)
			goto try_again;
		return 1;
	}
	ioctl(mdfd, STOP_ARRAY, NULL); /* just incase it was started but has no content */

	/* Ok, no bad inconsistancy, we can try updating etc */
	devices = xmalloc(num_devs * sizeof(*devices));
	devmap = zmalloc(num_devs * content->array.raid_disks);
	for (tmpdev = devlist; tmpdev; tmpdev=tmpdev->next)
	  if (tmpdev->used == 1) {
		char *devname = tmpdev->devname;
		struct stat stb;
		/* looks like a good enough match to update the super block if needed */
		{
			struct supertype *tst = dup_super(st);
			int dfd;
			dfd = dev_open(devname, O_RDWR|O_EXCL);

			if (dfd < 0 || tst->ss->load_super(tst, dfd, NULL) != 0) {
				bb_error_msg("cannot re-read metadata from %s - aborting",
					devname);
				if (dfd >= 0)
					close(dfd);
				close(mdfd);
				free(devices);
				free(devmap);
				return 1;
			}
			tst->ss->getinfo_super(tst, content, devmap + devcnt * content->array.raid_disks);
			tst->ss->free_super(tst);
			close(dfd);
		}

		stat(devname, &stb);

		if (verbose > 0)
			bb_error_msg("%s is identified as a member of %s, slot %d.",
				devname, mddev, content->disk.raid_disk);
		devices[devcnt].devname = devname;
		devices[devcnt].uptodate = 0;
		devices[devcnt].i = *content;
		devices[devcnt].i.disk.major = major(stb.st_rdev);
		devices[devcnt].i.disk.minor = minor(stb.st_rdev);
		if (most_recent < devcnt) {
			if (devices[devcnt].i.events
			    > devices[most_recent].i.events)
				most_recent = devcnt;
		}
		if (content->array.level == LEVEL_MULTIPATH)
			/* with multipath, the raid_disk from the superblock is meaningless */
			i = devcnt;
		else
			i = devices[devcnt].i.disk.raid_disk;
		if (i+1 == 0) {
			if (nextspare < content->array.raid_disks)
				nextspare = content->array.raid_disks;
			i = nextspare++;
		} else {
			if (i >= content->array.raid_disks &&
			    i >= nextspare)
				nextspare = i+1;
		}
		if (i < 10000) {
			if (i >= bestcnt) {
				int newbestcnt = i+10;
				int *newbest = xmalloc(sizeof(int)*newbestcnt);
				int c;
				for (c=0; c < newbestcnt; c++)
					if (c < bestcnt)
						newbest[c] = best[c];
					else
						newbest[c] = -1;
				if (best)
					free(best);
				best = newbest;
				bestcnt = newbestcnt;
			}
			if (best[i] >=0 &&
			    devices[best[i]].i.events
			    == devices[devcnt].i.events
			    && (devices[best[i]].i.disk.minor
				!= devices[devcnt].i.disk.minor)
			    && st->ss == &super0
			    && content->array.level != LEVEL_MULTIPATH) {
				/* two different devices with identical superblock.
				 * Could be a mis-detection caused by overlapping
				 * partitions.  fail-safe.
				 */
				bb_error_msg("WARNING %s and %s appear"
					" to have very similar superblocks.\n"
					"      If they are really different, "
					"please --zero the superblock on one\n"
					"      If they are the same or overlap,"
					" please remove one from %s.",
					devices[best[i]].devname, devname,
					inargv ? "the list" :
					   "the\n      DEVICE list in mdadm.conf"
					);
				close(mdfd);
				free(devices);
				free(devmap);
				return 1;
			}
			if (best[i] == -1
			    || (devices[best[i]].i.events
				< devices[devcnt].i.events))
				best[i] = devcnt;
		}
		devcnt++;
	}

	if (devcnt == 0) {
		bb_error_msg("no devices found for %s", mddev);
		if (st)
			st->ss->free_super(st);
		close(mdfd);
		free(devices);
		free(devmap);
		return 1;
	}

	if (update && strcmp(update, "byteorder")==0)
		st->minor_version = 90;

	st->ss->getinfo_super(st, content, NULL);
	clean = content->array.state & 1;

	/* now we have some devices that might be suitable.
	 * I wonder how many
	 */
	avail = zmalloc(content->array.raid_disks);
	okcnt = 0;
	sparecnt=0;
	rebuilding_cnt=0;
	for (i=0; i< bestcnt; i++) {
		int j = best[i];
		int event_margin = 1; /* always allow a difference of '1'
				       * like the kernel does
				       */
		if (j < 0) continue;
		/* note: we ignore error flags in multipath arrays
		 * as they don't make sense
		 */
		if (content->array.level != LEVEL_MULTIPATH)
			if (!(devices[j].i.disk.state & (1<<MD_DISK_ACTIVE))) {
				if (!(devices[j].i.disk.state
				      & (1<<MD_DISK_FAULTY))) {
					devices[j].uptodate = 1;
					sparecnt++;
				}
				continue;
			}
		/* If this device thinks that 'most_recent' has failed, then
		 * we must reject this device.
		 */
		if (j != most_recent &&
		    content->array.raid_disks > 0 &&
		    devices[most_recent].i.disk.raid_disk >= 0 &&
		    devmap[j * content->array.raid_disks + devices[most_recent].i.disk.raid_disk] == 0) {
			if (verbose > -1)
				bb_error_msg("ignoring %s as it reports %s as failed",
					devices[j].devname, devices[most_recent].devname);
			best[i] = -1;
			continue;
		}
		if (devices[j].i.events+event_margin >=
		    devices[most_recent].i.events) {
			devices[j].uptodate = 1;
			if (i < content->array.raid_disks) {
				if (devices[j].i.recovery_start == MaxSector ||
				    (content->reshape_active &&
				     ((i >= content->array.raid_disks - content->delta_disks) ||
				      (i >= content->array.raid_disks - content->delta_disks - 1
				       && content->array.level == 4)))) {
					okcnt++;
					avail[i]=1;
				} else
					rebuilding_cnt++;
			} else
				sparecnt++;
		}
	}
	free(devmap);
	while (force &&
	       (!enough(content->array.level, content->array.raid_disks,
			content->array.layout, 1,
			avail)
		||
		(content->reshape_active && content->delta_disks > 0 &&
		 !enough(content->array.level, (content->array.raid_disks
						- content->delta_disks),
			 content->new_layout, 1,
			 avail)
			))) {
		/* Choose the newest best drive which is
		 * not up-to-date, update the superblock
		 * and add it.
		 */
		int fd;
		struct supertype *tst;
		unsigned long long current_events;
		chosen_drive = -1;
		for (i = 0; i < content->array.raid_disks && i < bestcnt; i++) {
			int j = best[i];
			if (j>=0 &&
			    !devices[j].uptodate &&
			    devices[j].i.recovery_start == MaxSector &&
			    (chosen_drive < 0 ||
			     devices[j].i.events
			     > devices[chosen_drive].i.events))
				chosen_drive = j;
		}
		if (chosen_drive < 0)
			break;
		current_events = devices[chosen_drive].i.events;
	add_another:
		if (verbose >= 0)
			bb_error_msg("forcing event count in %s(%d) from %d upto %d",
				devices[chosen_drive].devname,
				devices[chosen_drive].i.disk.raid_disk,
				(int)(devices[chosen_drive].i.events),
				(int)(devices[most_recent].i.events));
		fd = dev_open(devices[chosen_drive].devname, O_RDWR|O_EXCL);
		if (fd < 0) {
			bb_error_msg("Couldn't open %s for write - not updating",
				devices[chosen_drive].devname);
			devices[chosen_drive].i.events = 0;
			continue;
		}
		tst = dup_super(st);
		if (tst->ss->load_super(tst,fd, NULL)) {
			close(fd);
			bb_error_msg("RAID superblock disappeared from %s - not updating.",
				devices[chosen_drive].devname);
			devices[chosen_drive].i.events = 0;
			continue;
		}
		content->events = devices[most_recent].i.events;
		tst->ss->update_super(tst, content, "force-one",
				     devices[chosen_drive].devname, verbose,
				     0, NULL);

		if (tst->ss->store_super(tst, fd)) {
			close(fd);
			bb_error_msg("Could not re-write superblock on %s",
				devices[chosen_drive].devname);
			devices[chosen_drive].i.events = 0;
			tst->ss->free_super(tst);
			continue;
		}
		close(fd);
		devices[chosen_drive].i.events = devices[most_recent].i.events;
		devices[chosen_drive].uptodate = 1;
		avail[chosen_drive] = 1;
		okcnt++;
		tst->ss->free_super(tst);

		/* If there are any other drives of the same vintage,
		 * add them in as well.  We can't lose and we might gain
		 */
		for (i = 0; i < content->array.raid_disks && i < bestcnt ; i++) {
			int j = best[i];
			if (j >= 0 &&
			    !devices[j].uptodate &&
			    devices[j].i.recovery_start == MaxSector &&
			    devices[j].i.events == current_events) {
				chosen_drive = j;
				goto add_another;
			}
		}
	}

	/* Now we want to look at the superblock which the kernel will base things on
	 * and compare the devices that we think are working with the devices that the
	 * superblock thinks are working.
	 * If there are differences and --force is given, then update this chosen
	 * superblock.
	 */
	chosen_drive = -1;
	st->ss->free_super(st);
	for (i=0; chosen_drive < 0 && i<bestcnt; i++) {
		int j = best[i];
		int fd;

		if (j<0)
			continue;
		if (!devices[j].uptodate)
			continue;
		if (devices[j].i.events < devices[most_recent].i.events)
			continue;
		chosen_drive = j;
		if ((fd=dev_open(devices[j].devname, O_RDONLY|O_EXCL))< 0) {
			bb_perror_msg("Cannot open %s", devices[j].devname);
			close(mdfd);
			free(devices);
			return 1;
		}
		if (st->ss->load_super(st,fd, NULL)) {
			close(fd);
			bb_error_msg("RAID superblock has disappeared from %s",
				devices[j].devname);
			close(mdfd);
			free(devices);
			return 1;
		}
		close(fd);
	}
	if (st->sb == NULL) {
		bb_error_msg("No suitable drives found for %s", mddev);
		close(mdfd);
		free(devices);
		return 1;
	}
	st->ss->getinfo_super(st, content, NULL);
	for (i=0; i<bestcnt; i++) {
		int j = best[i];
		unsigned int desired_state;

		if (i < content->array.raid_disks)
			desired_state = (1<<MD_DISK_ACTIVE) | (1<<MD_DISK_SYNC);
		else
			desired_state = 0;

		if (j<0)
			continue;
		if (!devices[j].uptodate)
			continue;

		devices[j].i.disk.state = desired_state;
		if (!(devices[j].i.array.state & 1))
			clean = 0;

		if (st->ss->update_super(st, &devices[j].i, "assemble", NULL,
					 verbose, 0, NULL)) {
			if (force) {
				if (verbose >= 0)
					bb_error_msg("clearing FAULTY flag for device %d in %s for %s",
						j, mddev, devices[j].devname);
				change = 1;
			} else {
				if (verbose >= -1)
					bb_error_msg("device %d in %s has wrong state in superblock, but %s seems ok",
						i, mddev, devices[j].devname);
			}
		}
	}
	if (force && !clean &&
	    !enough(content->array.level, content->array.raid_disks,
		    content->array.layout, clean,
		    avail)) {
		change += st->ss->update_super(st, content, "force-array",
					devices[chosen_drive].devname, verbose,
					       0, NULL);
		clean = 1;
	}

	if (change) {
		int fd;
		fd = dev_open(devices[chosen_drive].devname, O_RDWR|O_EXCL);
		if (fd < 0) {
			bb_error_msg("Could not open %s for write - cannot Assemble array.",
				devices[chosen_drive].devname);
			close(mdfd);
			free(devices);
			return 1;
		}
		if (st->ss->store_super(st, fd)) {
			close(fd);
			bb_error_msg("Could not re-write superblock on %s",
				devices[chosen_drive].devname);
			close(mdfd);
			free(devices);
			return 1;
		}
		if (verbose >= 0)
			bb_error_msg("Marking array %s as 'clean'", mddev);
		close(fd);
	}

	/* count number of in-sync devices according to the superblock.
	 * We must have this number to start the array without -s or -R
	 */
	req_cnt = content->array.working_disks;

	/* Almost ready to actually *do* something */

	/* First, fill in the map, so that udev can find our name
	 * as soon as we become active.
	 */
	rv = set_array_info(mdfd, content);
	if (rv) {
		bb_perror_msg("failed to set array info for %s", mddev);
		ioctl(mdfd, STOP_ARRAY, NULL);
		close(mdfd);
		free(devices);
		return 1;
	}
	if (ident->bitmap_fd >= 0) {
		if (xioctl(mdfd, SET_BITMAP_FILE, (void *)(ptrdiff_t)ident->bitmap_fd, "SET_BITMAP_FILE failed") != 0) {
			ioctl(mdfd, STOP_ARRAY, NULL);
			close(mdfd);
			free(devices);
			return 1;
		}
	} else if (ident->bitmap_file) {
		/* From config file */
		int bmfd = bb_xopen(ident->bitmap_file, O_RDWR | BB_OPEN_WARN);
		if (bmfd < 0) {
			ioctl(mdfd, STOP_ARRAY, NULL);
			close(mdfd);
			free(devices);
			return 1;
		}
		if (xioctl(mdfd, SET_BITMAP_FILE, (void *)(ptrdiff_t)bmfd, "Failed to set bitmapfile for %s", mddev) != 0) {
			close(bmfd);
			ioctl(mdfd, STOP_ARRAY, NULL);
			close(mdfd);
			free(devices);
			return 1;
		}
		close(bmfd);
	}

	/* First, add the raid disks, but add the chosen one last */
	for (i=0; i<= bestcnt; i++) {
		int j;
		if (i < bestcnt) {
			j = best[i];
			if (j == chosen_drive)
				continue;
		} else
			j = chosen_drive;

		if (j >= 0 /* && devices[j].uptodate */) {
			int dfd = dev_open(devices[j].devname,
					   O_RDWR|O_EXCL);
			if (dfd >= 0) {
				remove_partitions(dfd);
				close(dfd);
			}
			rv = add_disk(mdfd, &devices[j].i);

			if (rv) {
				bb_perror_msg("failed to add "
						"%s to %s",
					devices[j].devname, mddev);
				if (i < content->array.raid_disks
				    || i == bestcnt)
					okcnt--;
				else
					sparecnt--;
			} else if (verbose > 0)
				bb_error_msg("added %s "
						"to %s as %d%s",
					devices[j].devname, mddev,
					devices[j].i.disk.raid_disk,
					devices[j].uptodate?"":
					" (possibly out of date)");
		} else if (verbose > 0 && i < content->array.raid_disks)
			bb_error_msg("no uptodate device for "
					"slot %d of %s", i, mddev);
	}

	if (content->array.level == LEVEL_CONTAINER) {
		if (verbose >= 0) {
			fprintf(stderr, "%s: Container %s has been "
				"assembled with %d drive%s",
				bb_applet_name,
				mddev, okcnt+sparecnt, okcnt+sparecnt==1?"":"s");
			if (okcnt < (unsigned)content->array.raid_disks)
				fprintf(stderr, " (out of %d)",
					content->array.raid_disks);
			fprintf(stderr, "\n");
		}
		st->ss->free_super(st);
		sysfs_uevent(content, "change");
		wait_for(chosen_name, mdfd);
		close(mdfd);
		free(devices);
		return 0;
	}

	if (runstop == 1 ||
	    (runstop <= 0 &&
	     ( enough(content->array.level, content->array.raid_disks,
		      content->array.layout, clean, avail) &&
	       (okcnt + rebuilding_cnt >= req_cnt || start_partial_ok)
		     ))) {
		/* This array is good-to-go.
		 * If a reshape is in progress then we might need to
		 * continue monitoring it.  In that case we start
		 * it read-only and let the grow code make it writable.
		 */
		rv = ioctl(mdfd, RUN_ARRAY, NULL);

		if (rv == 0) {
			if (verbose >= 0) {
				fprintf(stderr, "%s: %s has been started with %d drive%s",
					bb_applet_name,
					mddev, okcnt, okcnt==1?"":"s");
				if (okcnt < (unsigned)content->array.raid_disks)
					fprintf(stderr, " (out of %d)", content->array.raid_disks);
				if (rebuilding_cnt)
					fprintf(stderr, "%s %d rebuilding", sparecnt?",":" and", rebuilding_cnt);
				if (sparecnt)
					fprintf(stderr, " and %d spare%s", sparecnt, sparecnt==1?"":"s");
				fprintf(stderr, ".\n");
			}
			if (content->reshape_active &&
			    content->array.level >= 4 &&
			    content->array.level <= 6) {
				/* might need to increase the size
				 * of the stripe cache - default is 256
				 */
				if (256 < 4 * (content->array.chunk_size/4096)) {
					struct mdinfo *sra = sysfs_read(mdfd, 0, 0);
					if (sra)
						sysfs_set_num(sra, NULL,
							      "stripe_cache_size",
							      (4 * content->array.chunk_size / 4096) + 1);
					sysfs_free(sra);
				}
			}
			if (okcnt < (unsigned)content->array.raid_disks) {
				/* If any devices did not get added
				 * because the kernel rejected them based
				 * on event count, try adding them
				 * again providing the action policy is
				 * 're-add' or greater.  The bitmap
				 * might allow them to be included, or
				 * they will become spares.
				 */
				for (i = 0; i < bestcnt; i++) {
					int j = best[i];
					if (j >= 0 && !devices[j].uptodate) {
						if (!disk_action_allows(&devices[j].i, st->ss->name, act_re_add))
							continue;
						rv = add_disk(mdfd, &devices[j].i);
						if (rv == 0 && verbose >= 0)
							bb_error_msg("%s has been re-added.",
								devices[j].devname);
					}
				}
			}
			wait_for(mddev, mdfd);
			close(mdfd);
			if (auto_assem) {
				int usecs = 1;
				/* There is a nasty race with 'mdadm --monitor'.
				 * If it opens this device before we close it,
				 * it gets an incomplete open on which IO
				 * doesn't work and the capacity is
				 * wrong.
				 * If we reopen (to check for layered devices)
				 * before --monitor closes, we loose.
				 *
				 * So: wait upto 1 second for there to be
				 * a non-zero capacity.
				 */
				while (usecs < 1000) {
					mdfd = open(mddev, O_RDONLY);
					if (mdfd >= 0) {
						unsigned long long size;
						if (get_dev_size(mdfd, NULL, &size) && size > 0)
							break;
						close(mdfd);
					}
					xusleep(usecs);
					usecs <<= 1;
				}
			}
			free(devices);
			return 0;
		}
		bb_perror_msg("failed to RUN_ARRAY %s", mddev);

		if (!enough(content->array.level, content->array.raid_disks,
			    content->array.layout, 1, avail))
			bb_error_msg("Not enough devices to "
				"start the array.");
		else if (!enough(content->array.level,
				 content->array.raid_disks,
				 content->array.layout, clean,
				 avail))
			bb_error_msg("Not enough devices to "
				"start the array while not clean "
				"- consider --force.");

		if (auto_assem)
			ioctl(mdfd, STOP_ARRAY, NULL);
		close(mdfd);
		free(devices);
		return 1;
	}
	if (runstop == -1) {
		fprintf(stderr, "%s: %s assembled from %d drive%s",
			bb_applet_name,
			mddev, okcnt, okcnt==1?"":"s");
		if (okcnt != (unsigned)content->array.raid_disks)
			fprintf(stderr, " (out of %d)", content->array.raid_disks);
		fprintf(stderr, ", but not started.\n");
		close(mdfd);
		free(devices);
		return 0;
	}
	if (verbose >= -1) {
		fprintf(stderr, "%s: %s assembled from %d drive%s",
			bb_applet_name,
			mddev, okcnt, okcnt==1?"":"s");
		if (rebuilding_cnt)
			fprintf(stderr, "%s %d rebuilding", sparecnt?", ":" and ", rebuilding_cnt);
		if (sparecnt)
			fprintf(stderr, " and %d spare%s", sparecnt, sparecnt==1?"":"s");
		if (!enough(content->array.level, content->array.raid_disks,
			    content->array.layout, 1, avail))
			fprintf(stderr, " - not enough to start the array.\n");
		else if (!enough(content->array.level,
				 content->array.raid_disks,
				 content->array.layout, clean,
				 avail))
			fprintf(stderr, " - not enough to start the "
				"array while not clean - consider "
				"--force.\n");
		else {
			if (req_cnt == (unsigned)content->array.raid_disks)
				fprintf(stderr, " - need all %d to start it", req_cnt);
			else
				fprintf(stderr, " - need %d of %d to start", req_cnt, content->array.raid_disks);
			fprintf(stderr, " (use --run to insist).\n");
		}
	}
	if (auto_assem)
		ioctl(mdfd, STOP_ARRAY, NULL);
	close(mdfd);
	free(devices);
	return 1;
}

#define REGISTER_DEV            _IO (MD_MAJOR, 1)
#define START_MD                _IO (MD_MAJOR, 2)
#define STOP_MD                 _IO (MD_MAJOR, 3)

static int Manage_ro(char *devname, int fd, int readonly)
{
	/* switch to readonly or rw
	 *
	 * requires >= 0.90.0
	 * first check that array is runing
	 * use RESTART_ARRAY_RW or STOP_ARRAY_RO
	 *
	 */
	mdu_array_info_t array;
	int rv = 0;

	if (md_get_version(fd) < 9000) {
		bb_error_msg("need md driver version 0.90.0 or later");
		return 1;
	}
	if (xioctl(fd, GET_ARRAY_INFO, &array, "%s does not appear to be active", devname)) {
		rv = 1;
		goto out;
	}

	if (readonly>0) {
		if (xioctl(fd, STOP_ARRAY_RO, NULL, "failed to set readonly for %s", devname)) {
			rv = 1;
			goto out;
		}
	} else if (readonly < 0) {
		if (xioctl(fd, RESTART_ARRAY_RW, NULL, "failed to set writable for %s", devname)) {
			rv = 1;
			goto out;
		}
	}
out:
	return rv;
}

/*
 * Read the config file
 *
 * conf_get_uuids gets a list of devicename+uuid pairs
 * conf_get_devs gets device names after expanding wildcards
 *
 * Each keeps the returned list and frees it when asked to make
 * a new list.
 *
 * The format of the config file needs to be fairly extensible.
 * Now, arrays only have names and uuids and devices merely are.
 * But later arrays might want names, and devices might want superblock
 * versions, and who knows what else.
 * I like free format, abhore backslash line continuation, adore
 *   indentation for structure and am ok about # comments.
 *
 * So, each line that isn't blank or a #comment must either start
 *  with a key word, and not be indented, or must start with a
 *  non-key-word and must be indented.
 *
 * Keywords are DEVICE and ARRAY ... and several others.
 * DEV{ICE} introduces some devices that might contain raid components.
 * e.g.
 *   DEV style=0 /dev/sda* /dev/hd*
 *   DEV style=1 /dev/sd[b-f]*
 * ARR{AY} describes an array giving md device and attributes like uuid=whatever
 * e.g.
 *   ARRAY /dev/md0 uuid=whatever name=something
 * Spaces separate words on each line.  Quoting, with "" or '' protects them,
 * but may not wrap over lines
 *
 */


enum linetype { Devices, Array, Mailaddr, Mailfrom, Program, CreateDev,
		Homehost, AutoMode, Policy, PartPolicy, LTEnd };
static const char * const keywords[] = {
	[Devices]  = "devices",
	[Array]    = "array",
	[Mailaddr] = "mailaddr",
	[Mailfrom] = "mailfrom",
	[Program]  = "program",
	[CreateDev]= "create",
	[Homehost] = "homehost",
	[AutoMode] = "auto",
	[Policy]   = "policy",
	[PartPolicy]="part-policy",
	[LTEnd]    = NULL
};

/*
 * match_keyword returns an index into the keywords array, or -1 for no match
 * case is ignored, and at least three characters must be given
 */

static int match_keyword(char *word)
{
	int len = strlen(word);
	int n;

	if (len < 3) return -1;
	for (n=0; keywords[n]; n++) {
		if (strncasecmp(word, keywords[n], len)==0)
			return n;
	}
	return -1;
}

/*
 * conf_line reads one logical line from the conffile.
 * It skips comments and continues until it finds a line that starts
 * with a non blank/comment.  This character is pushed back for the next call
 * A doubly linked list of words is returned.
 * the first word will be a keyword.  Other words will have had quotes removed.
 */

char *conf_line(FILE *file)
{
	char *w;
	char *list;

	w = conf_word(file, 1);
	if (w == NULL) return NULL;

	list = dl_strdup(w);
	free(w);
	dl_init(list);

	while ((w = conf_word(file,0))){
		char *w2 = dl_strdup(w);
		free(w);
		dl_add(list, w2);
	}
/*    printf("got a line\n");*/
	return list;
}

void free_line(char *line)
{
	char *w;
	for (w=dl_next(line); w != line; w=dl_next(line)) {
		dl_del(w);
		dl_free(w);
	}
	dl_free(line);
}

static struct mddev_dev *load_partitions(void)
{
	FILE *f = fopen("/proc/partitions", "r");
	char buf[1024];
	struct mddev_dev *rv = NULL;

	if (f == NULL) {
		bb_perror_msg("cannot open /proc/partitions");
		return NULL;
	}
	while (fgets(buf, 1024, f)) {
		int major, minor;
		char *name, *mp;
		struct mddev_dev *d;

		buf[1023] = '\0';
		if (buf[0] != ' ')
			continue;
		major = strtoul(buf, &mp, 10);
		if (mp == buf || *mp != ' ')
			continue;
		minor = strtoul(mp, NULL, 10);

		name = map_dev(major, minor, 1);
		if (!name)
			continue;
		d = xmalloc(sizeof(*d));
		d->devname = strdup(name);
		d->next = rv;
		d->used = 0;
		rv = d;
	}
	fclose(f);
	return rv;
}

static struct mddev_dev *load_containers(void)
{
	struct mdstat_ent *mdstat = mdstat_read(1, 0);
	struct mdstat_ent *ent;
	struct mddev_dev *d;
	struct mddev_dev *rv = NULL;

	if (!mdstat)
		return NULL;

	for (ent = mdstat; ent; ent = ent->next)
		if (ent->metadata_version &&
		    strncmp(ent->metadata_version, "external:", 9) == 0 &&
		    !is_subarray(&ent->metadata_version[9])) {
			d = xmalloc(sizeof(*d));
			d->devname = xasprintf("/dev/%s", ent->dev);
			d->next = rv;
			d->used = 0;
			rv = d;
		}
	free_mdstat(mdstat);

	return rv;
}


int parse_auto(char *str, char *msg, int config)
{
	int autof;
	if (str == NULL || *str == 0)
		autof = 2;
	else if (strcasecmp(str,"no")==0)
		autof = 1;
	else if (strcasecmp(str,"yes")==0)
		autof = 2;
	else if (strcasecmp(str,"md")==0)
		autof = config?5:3;
	else {
		/* There might be digits, and maybe a hypen, at the end */
		char *e = str + strlen(str);
		int num = 4;
		int len;
		while (e > str && isdigit(e[-1]))
			e--;
		if (*e) {
			num = atoi(e);
			if (num <= 0) num = 1;
		}
		if (e > str && e[-1] == '-')
			e--;
		len = e - str;
		if ((len == 2 && strncasecmp(str,"md",2)==0)) {
			autof = config ? 5 : 3;
		} else if ((len == 3 && strncasecmp(str,"yes",3)==0)) {
			autof = 2;
		} else if ((len == 3 && strncasecmp(str,"mdp",3)==0)) {
			autof = config ? 6 : 4;
		} else if ((len == 1 && strncasecmp(str,"p",1)==0) ||
			   (len >= 4 && strncasecmp(str,"part",4)==0)) {
			autof = 6;
		} else {
			bb_error_msg("%s arg of \"%s\" unrecognised: use no,yes,md,mdp,part\n"
				"        optionally followed by a number.",
				msg, str);
			exit(2);
		}
		autof |= num << 3;
	}
	return autof;
}

#if 0   /* set but unused */
static struct createinfo createinfo = {
	.autof = 2, /* by default, create devices with standard names */
	.symlinks = 1,
#ifdef DEBIAN
	.gid = 6, /* disk */
	.mode = 0660,
#else
	.mode = 0600,
#endif
};
#endif

static void createline(char *line)
{
	char *w;
	char *ep;

#if 1
	/* set but unused */
	struct createinfo createinfo;
	createinfo.autof = 2;
	createinfo.symlinks = 1;
#ifdef DEBIAN
	createinfo.gid = 6; /* disk */
	createinfo.mode = 0660;
#else
	createinfo.mode = 0600;
#endif
#endif
	for (w=dl_next(line); w!=line; w=dl_next(w)) {
		if (strncasecmp(w, "auto=", 5) == 0)
			createinfo.autof = parse_auto(w+5, "auto=", 1);
		else if (strncasecmp(w, "owner=", 6) == 0) {
			if (w[6] == 0) {
				bb_error_msg("missing owner name");
				continue;
			}
			createinfo.uid = strtoul(w+6, &ep, 10);
			if (*ep != 0) {
				struct passwd *pw;
				/* must be a name */
				pw = getpwnam(w+6);
				if (pw)
					createinfo.uid = pw->pw_uid;
				else
					bb_error_msg("CREATE user %s not found", w+6);
			}
		} else if (strncasecmp(w, "group=", 6) == 0) {
			if (w[6] == 0) {
				bb_error_msg("missing group name");
				continue;
			}
			createinfo.gid = strtoul(w+6, &ep, 10);
			if (*ep != 0) {
				struct group *gr;
				/* must be a name */
				gr = getgrnam(w+6);
				if (gr)
					createinfo.gid = gr->gr_gid;
				else
					bb_error_msg("CREATE group %s not found", w+6);
			}
		} else if (strncasecmp(w, "mode=", 5) == 0) {
			if (w[5] == 0) {
				bb_error_msg("missing CREATE mode");
				continue;
			}
			createinfo.mode = strtoul(w+5, &ep, 8);
			if (*ep != 0) {
				createinfo.mode = 0600;
				bb_error_msg("unrecognised CREATE mode %s",
					w+5);
			}
		} else if (strncasecmp(w, "metadata=", 9) == 0) {
			/* style of metadata to use by default */
			int i;
			for (i=0; superlist[i] && !createinfo.supertype; i++)
				createinfo.supertype =
					superlist[i]->match_metadata_desc(w+9);
			if (!createinfo.supertype)
				bb_error_msg("metadata format %s unknown, ignoring",
					w+9);
		} else if (strncasecmp(w, "symlinks=yes", 12) == 0)
			createinfo.symlinks = 1;
		else if  (strncasecmp(w, "symlinks=no", 11) == 0)
			createinfo.symlinks = 0;
		else {
			bb_error_msg("unrecognised word on CREATE line: %s", w);
		}
	}
}

static void devline(char *line)
{
	char *w;
	struct conf_dev *cd;

	for (w=dl_next(line); w != line; w=dl_next(w)) {
		if (w[0] == '/' || strcasecmp(w, "partitions") == 0 ||
		    strcasecmp(w, "containers") == 0) {
			cd = xmalloc(sizeof(*cd));
			cd->name = strdup(w);
			cd->next = cdevlist;
			cdevlist = cd;
		} else {
			bb_error_msg("unreconised word on DEVICE line: %s", w);
		}
	}
}


static int is_number(char *w)
{
	/* check if there are 1 or more digits and nothing else */
	int digits = 0;
	while (*w && isdigit(*w)) {
		digits++;
		w++;
	}
	return (digits && ! *w);
}


static int map_name(const char *name)
{
	static const mapping_t pers[] = {
		{ "linear", LEVEL_LINEAR},
		{ "raid0", 0},
		{ "0", 0},
		{ "stripe", 0},
		{ "raid1", 1},
		{ "1", 1},
		{ "mirror", 1},
		{ "raid4", 4},
		{ "4", 4},
		{ "raid5", 5},
		{ "5", 5},
		{ "multipath", LEVEL_MULTIPATH},
		{ "mp", LEVEL_MULTIPATH},
		{ "raid6", 6},
		{ "6", 6},
		{ "raid10", 10},
		{ "10", 10},
		{ "faulty", LEVEL_FAULTY},
		{ "container", LEVEL_CONTAINER},
		{ NULL, UnSet}
	};
	return ((const mapping_t *)find_string_from_struct(pers, sizeof(mapping_t), name))->num;
}

static void arrayline(char *line)
{
	char *w;

	struct mddev_ident mis;
	struct mddev_ident *mi;

	mis.uuid_set = 0;
	mis.super_minor = UnSet;
	mis.level = UnSet;
	mis.raid_disks = UnSet;
	mis.spare_disks = 0;
	mis.devices = NULL;
	mis.devname = NULL;
	mis.spare_group = NULL;
	mis.autof = 0;
	mis.next = NULL;
	mis.st = NULL;
	mis.bitmap_fd = -1;
	mis.bitmap_file = NULL;
	mis.name[0] = 0;
	mis.container = NULL;
	mis.member = NULL;

	for (w=dl_next(line); w!=line; w=dl_next(w)) {
		if (w[0] == '/' || strchr(w, '=') == NULL) {
			/* This names the device, or is '<ignore>'.
			 * The rules match those in create_mddev.
			 * 'w' must be:
			 *  /dev/md/{anything}
			 *  /dev/mdNN
			 *  /dev/md_dNN
			 *  <ignore>
			 *  or anything that doesn't start '/' or '<'
			 */
			if (strcasecmp(w, "<ignore>") == 0 ||
			    strncmp(w, "/dev/md/", 8) == 0 ||
			    (w[0] != '/' && w[0] != '<') ||
			    (strncmp(w, "/dev/md", 7) == 0 &&
			     is_number(w+7)) ||
			    (strncmp(w, "/dev/md_d", 9) == 0 &&
			     is_number(w+9))
				) {
				/* This is acceptable */;
				if (mis.devname)
					bb_error_msg("only give one "
						"device per ARRAY line: %s and %s",
						mis.devname, w);
				else
					mis.devname = w;
			}else {
				bb_error_msg("%s is an invalid name for "
					"an md device - ignored.", w);
			}
		} else if (strncasecmp(w, "uuid=", 5)==0 ) {
			if (mis.uuid_set)
				bb_error_msg("only specify uuid once, %s ignored.",
					w);
			else {
				if (parse_uuid(w+5, mis.uuid))
					mis.uuid_set = 1;
				else
					bb_error_msg("bad uuid: %s", w);
			}
		} else if (strncasecmp(w, "super-minor=", 12)==0 ) {
			if (mis.super_minor != UnSet)
				bb_error_msg("only specify super-minor once, %s ignored.",
					w);
			else {
				int minor = bb_xparse_i_number(10 | BB_PN_SET_LOWER1 | BB_PN_ERR_STRV | BB_PN_RETURN_ERR,
									w+12, "invalid super-minor number: %s", w);
				if (minor)
					mis.super_minor = minor;
			}
		} else if (strncasecmp(w, "name=", 5)==0) {
			if (mis.name[0])
				bb_error_msg("only specify name once, %s ignored.",
					w);
			else if (strlen(w+5) > 32)
				bb_error_msg("name too long, ignoring %s", w);
			else
				strcpy(mis.name, w+5);

		} else if (strncasecmp(w, "bitmap=", 7) == 0) {
			if (mis.bitmap_file)
				bb_error_msg("only specify bitmap file once. %s ignored",
					w);
			else
				mis.bitmap_file = strdup(w+7);

		} else if (strncasecmp(w, "devices=", 8 ) == 0 ) {
			if (mis.devices)
				bb_error_msg("only specify devices once (use a comma separated list). %s ignored",
					w);
			else
				mis.devices = strdup(w+8);
		} else if (strncasecmp(w, "spare-group=", 12) == 0 ) {
			if (mis.spare_group)
				bb_error_msg("only specify one spare group per array. %s ignored.",
					w);
			else
				mis.spare_group = strdup(w+12);
		} else if (strncasecmp(w, "level=", 6) == 0 ) {
			/* this is mainly for compatability with --brief output */
			mis.level = map_name(w+6);
		} else if (strncasecmp(w, "disks=", 6) == 0 ) {
			/* again, for compat */
			mis.raid_disks = atoi(w+6);
		} else if (strncasecmp(w, "num-devices=", 12) == 0 ) {
			/* again, for compat */
			mis.raid_disks = atoi(w+12);
		} else if (strncasecmp(w, "spares=", 7) == 0 ) {
			/* for warning if not all spares present */
			mis.spare_disks = atoi(w+7);
		} else if (strncasecmp(w, "metadata=", 9) == 0) {
			/* style of metadata on the devices. */
			int i;

			for(i=0; superlist[i] && !mis.st; i++)
				mis.st = superlist[i]->match_metadata_desc(w+9);

			if (!mis.st)
				bb_error_msg("metadata format %s unknown, ignored.", w+9);
		} else if (strncasecmp(w, "auto=", 5) == 0 ) {
			/* whether to create device special files as needed */
			mis.autof = parse_auto(w+5, "auto type", 0);
		} else if (strncasecmp(w, "member=", 7) == 0) {
			/* subarray within a container */
			mis.member = strdup(w+7);
		} else if (strncasecmp(w, "container=", 10) == 0) {
			/* the container holding this subarray.  Either a device name
			 * or a uuid */
			mis.container = strdup(w+10);
		} else {
			bb_error_msg("unrecognised word on ARRAY line: %s",
				w);
		}
	}
	if (mis.uuid_set == 0 && mis.devices == NULL &&
	    mis.super_minor == UnSet && mis.name[0] == 0 &&
	    (mis.container == NULL || mis.member == NULL))
		bb_error_msg("ARRAY line %s has no identity information.", mis.devname);
	else {
		mi = xmalloc(sizeof(*mi));
		*mi = mis;
		mi->devname = mis.devname ? strdup(mis.devname) : NULL;
		mi->next = NULL;
		*mddevlp = mi;
		mddevlp = &mi->next;
	}
}

static const char auto_yes[] = "yes";
static const char auto_no[] = "no";
static const char auto_homehost[] = "homehost";

static void autoline(char *line)
{
	char *w;
	char *seen;
	int super_cnt;
	const char *dflt = auto_yes;
	int homehost = 0;
	int i;

	/* Parse the 'auto' line creating policy statements for the 'auto' policy.
	 *
	 * The default is 'yes' but the 'auto' line might over-ride that.
	 * Words in the line are processed in order with the first
	 * match winning.
	 * word can be:
	 *   +version   - that version can be assembled
	 *   -version   - that version cannot be auto-assembled
	 *   yes or +all - any other version can be assembled
	 *   no or -all  - no other version can be assembled.
	 *   homehost   - any array associated by 'homehost' to this
	 *                host can be assembled.
	 *
	 * Thus:
	 *   +ddf -0.90 homehost -all
	 * will auto-assemble any ddf array, no 0.90 array, and
	 * any other array (imsm, 1.x) if and only if it is identified
	 * as belonging to this host.
	 *
	 * We translate that to policy by creating 'auto=yes' when we see
	 * a '+version' line, 'auto=no' if we see '-version' before 'homehost',
	 * or 'auto=homehost' if we see '-version' after 'homehost'.
	 * When we see yes, no, +all or -all we stop and any version that hasn't
	 * been seen gets an appropriate auto= entry.
	 */

	for (super_cnt = 0; superlist[super_cnt]; super_cnt++)
		;
	seen = calloc(super_cnt, 1);

	for (w = dl_next(line); w != line ; w = dl_next(w)) {
		const char *val;

		if (strcasecmp(w, "yes") == 0) {
			dflt = auto_yes;
			break;
		}
		if (strcasecmp(w, "no") == 0) {
			if (homehost)
				dflt = auto_homehost;
			else
				dflt = auto_no;
			break;
		}
		if (strcasecmp(w, "homehost") == 0) {
			homehost = 1;
			continue;
		}
		if (w[0] == '+')
			val = auto_yes;
		else if (w[0] == '-') {
			if (homehost)
				val = auto_homehost;
			else
				val = auto_no;
		} else
			continue;

		if (strcasecmp(w+1, "all") == 0) {
			dflt = val;
			break;
		}
		for (i = 0; superlist[i]; i++) {
			const char *version = superlist[i]->name;
			if (strcasecmp(w+1, version) == 0)
				break;
			/* 1 matches 1.x, 0 matches 0.90 */
			if (version[1] == '.' &&
			    strlen(w+1) == 1 &&
			    w[1] == version[0])
				break;
			/* 1.anything matches 1.x */
			if (strcmp(version, "1.x") == 0 &&
			    strncmp(w+1, "1.", 2) == 0)
				break;
		}
		if (superlist[i] == NULL)
			/* ignore this word */
			continue;
		if (seen[i])
			/* already know about this metadata */
			continue;
		policy_add(rule_policy, pol_auto, val, pol_metadata, superlist[i]->name, NULL);
		seen[i] = 1;
	}
	for (i = 0; i < super_cnt; i++)
		if (!seen[i])
			policy_add(rule_policy, pol_auto, dflt, pol_metadata, superlist[i]->name, NULL);

	free(seen);
}

static void load_conffile(void)
{
	FILE *f;
	char *line;

	if (loaded) return;

	f = fopen(conffile, "r");
	/* Debian chose to relocate mdadm.conf into /etc/mdadm/.
	 * To allow Debian users to compile from clean source and still
	 * have a working mdadm, we read /etc/mdadm/mdadm.conf
	 * if /etc/mdadm.conf doesn't exist
	 */
	if (f == NULL && conffile == DefaultConfFile) {
		f = fopen(DefaultAltConfFile, "r");
		if (f)
			conffile = DefaultAltConfFile;
	}
	if (f == NULL)
		return;

	loaded = 1;
	while ((line=conf_line(f))) {
		switch(match_keyword(line)) {
		case Devices:
			devline(line);
			break;
		case Array:
			arrayline(line);
			break;
		case Mailaddr:
			/* unuse */
			break;
		case Mailfrom:
			/* unuse */
			break;
		case Program:
			/* unuse */
			break;
		case CreateDev:
			createline(line);
			break;
		case Homehost:
			/* unuse */
			break;
		case AutoMode:
			autoline(line);
			break;
		case Policy:
			policyline(line, rule_policy);
			break;
		case PartPolicy:
			policyline(line, rule_part);
			break;
		default:
			bb_error_msg("Unknown keyword %s", line);
		}
		free_line(line);
	}

	fclose(f);
}

static struct mddev_ident *conf_get_ident(char *dev)
{
	struct mddev_ident *rv;
	load_conffile();
	rv = mddevlist;
	while (dev && rv && (rv->devname == NULL
			     || !devname_matches(dev, rv->devname)))
		rv = rv->next;
	return rv;
}

static void append_dlist(struct mddev_dev **dlp, struct mddev_dev *list)
{
	while (*dlp)
		dlp = &(*dlp)->next;
	*dlp = list;
}

struct mddev_dev *conf_get_devs()
{
	glob_t globbuf;
	struct conf_dev *cd;
	int flags = 0;
	unsigned int i;

	while (mddev_dev_dlist) {
		struct mddev_dev *t = mddev_dev_dlist;
		mddev_dev_dlist = mddev_dev_dlist->next;
		free(t->devname);
		free(t);
	}

	load_conffile();

	if (cdevlist == NULL) {
		/* default to 'partitions' and 'containers' */
		mddev_dev_dlist = load_partitions();
		append_dlist(&mddev_dev_dlist, load_containers());
	}

	for (cd=cdevlist; cd; cd=cd->next) {
		if (strcasecmp(cd->name, "partitions")==0)
			append_dlist(&mddev_dev_dlist, load_partitions());
		else if (strcasecmp(cd->name, "containers")==0)
			append_dlist(&mddev_dev_dlist, load_containers());
		else {
			glob(cd->name, flags, NULL, &globbuf);
			flags |= GLOB_APPEND;
		}
	}
	if (flags & GLOB_APPEND) {
		for (i=0; i<globbuf.gl_pathc; i++) {
			struct mddev_dev *t = xmalloc(sizeof(*t));
			t->devname = strdup(globbuf.gl_pathv[i]);
			t->next = mddev_dev_dlist;
			t->used = 0;
			mddev_dev_dlist = t;
		}
		globfree(&globbuf);
	}

	return mddev_dev_dlist;
}

int conf_test_metadata(const char *version, struct dev_policy *pol, int is_homehost)
{
	/* If anyone said 'yes', that sticks.
	 * else if homehost applies, use that
	 * else if there is a 'no', say 'no'.
	 * else 'yes'.
	 */
	struct dev_policy *p;
	int no=0, found_homehost=0;
	load_conffile();

	pol = pol_find(pol, pol_auto);
	pol_for_each(p, pol, version) {
		if (strcmp(p->value, "yes") == 0)
			return 1;
		if (strcmp(p->value, "homehost") == 0)
			found_homehost = 1;
		if (strcmp(p->value, "no") == 0)
			no = 1;
	}
	if (is_homehost && found_homehost)
		return 1;
	if (no)
		return 0;
	return 1;
}

int match_oneof(char *devices, char *devname)
{
    /* check if one of the comma separated patterns in devices
     * matches devname
     */

    while (devices && *devices) {
	char patn[1024];
	char *p = devices;
	devices = strchr(devices, ',');
	if (!devices)
	    devices = p + strlen(p);
	if (devices-p < 1024) {
		strncpy(patn, p, devices-p);
		patn[devices-p] = 0;
		if (fnmatch(patn, devname, FNM_PATHNAME)==0)
			return 1;
	}
	if (*devices == ',')
		devices++;
    }
    return 0;
}

int devname_matches(char *name, char *match)
{
	/* See if the given array name matches the
	 * given match from config file.
	 *
	 * First strip and /dev/md/ or /dev/, then
	 * see if there might be a numeric match of
	 *  mdNN with NN
	 * then just strcmp
	 */
	if (strncmp(name, "/dev/md/", 8) == 0)
		name += 8;
	else if (strncmp(name, "/dev/", 5) == 0)
		name += 5;

	if (strncmp(match, "/dev/md/", 8) == 0)
		match += 8;
	else if (strncmp(match, "/dev/", 5) == 0)
		match += 5;


	if (strncmp(name, "md", 2) == 0 &&
	    isdigit(name[2]))
		name += 2;
	if (strncmp(match, "md", 2) == 0 &&
	    isdigit(match[2]))
		match += 2;

	return (strcmp(name, match) == 0);
}

int conf_name_is_free(char *name)
{
	/* Check if this name is already take by an ARRAY entry in
	 * the config file.
	 * It can be taken either by a match on devname, name, or
	 * even super-minor.
	 */
	struct mddev_ident *dev;

	load_conffile();
	for (dev = mddevlist; dev; dev = dev->next) {
		char nbuf[100];
		if (dev->devname && devname_matches(name, dev->devname))
			return 0;
		if (dev->name[0] && devname_matches(name, dev->name))
			return 0;
		sprintf(nbuf, "%d", dev->super_minor);
		if (dev->super_minor != UnSet &&
		    devname_matches(name, nbuf))
			return 0;
	}
	return 1;
}

/*
 * Policy module for mdadm.
 * A policy statement about a device lists a set of values for each
 * of a set of names.  Each value can have a metadata type as context.
 *
 * names include:
 *   action - the actions that can be taken on hot-plug
 *   domain - the domain(s) that the device is part of
 *
 * Policy information is extracted from various sources, but
 * particularly from a set of policy rules in mdadm.conf
 */

static void pol_new(struct dev_policy **pol, const char *name, const char *val,
		    const char *metadata)
{
	struct dev_policy *n = xmalloc(sizeof(*n));
	const char *real_metadata = NULL;
	int i;

	n->name = name;
	n->value = val;

	/* We need to normalise the metadata name */
	if (metadata) {
		for (i = 0; superlist[i] ; i++)
			if (strcmp(metadata, superlist[i]->name) == 0) {
				real_metadata = superlist[i]->name;
				break;
			}
		if (!real_metadata) {
			if (strcmp(metadata, "1") == 0 ||
			    strcmp(metadata, "1.0") == 0 ||
			    strcmp(metadata, "1.1") == 0 ||
			    strcmp(metadata, "1.2") == 0)
				real_metadata = super1.name;
		}
		if (!real_metadata) {

			if (prev_metadata != metadata) {
				bb_error_msg("metadata=%s unrecognised - ignoring rule",
					metadata);
				prev_metadata = metadata;
			}
			real_metadata = "unknown";
		}
	}

	n->metadata = real_metadata;
	n->next = *pol;
	*pol = n;
}

static int pol_lesseq(struct dev_policy *a, struct dev_policy *b)
{
	int cmp;

	if (a->name < b->name)
		return 1;
	if (a->name > b->name)
		return 0;

	cmp = strcmp(a->value, b->value);
	if (cmp < 0)
		return 1;
	if (cmp > 0)
		return 0;

	return (a->metadata <= b->metadata);
}

static void pol_sort(struct dev_policy **pol)
{
	/* sort policy list in *pol by name/metadata/value
	 * using merge sort
	 */

	struct dev_policy *pl[2];
	pl[0] = *pol;
	pl[1] = NULL;

	do {
		struct dev_policy **plp[2], *p[2];
		int curr = 0;
		struct dev_policy nul = { NULL, NULL, NULL, NULL };
		struct dev_policy *prev = &nul;
		int next = 0;

		/* p[] are the two lists that we are merging.
		 * plp[] are the ends of the two lists we create
		 * from the merge.
		 * 'curr' is which of plp[] that we are currently
		 *   adding items to.
		 * 'next' is which if p[] we will take the next
		 *   item from.
		 * 'prev' is that last value, which was placed in
		 * plp[curr].
		 */
		plp[0] = &pl[0];
		plp[1] = &pl[1];
		p[0] = pl[0];
		p[1] = pl[1];

		/* take least of p[0] and p[1]
		 * if it is larger than prev, add to
		 * plp[curr], else swap curr then add
		 */
		while (p[0] || p[1]) {
			if (p[next] == NULL ||
			    (p[1-next] != NULL &&
			     !(pol_lesseq(prev, p[1-next])
			       ^pol_lesseq(prev, p[next])
			       ^pol_lesseq(p[next], p[1-next])))
				)
				next = 1 - next;

			if (!pol_lesseq(prev, p[next]))
				curr = 1 - curr;

			*plp[curr] = prev = p[next];
			plp[curr] = &p[next]->next;
			p[next] = p[next]->next;
		}
		*plp[0] = NULL;
		*plp[1] = NULL;
	} while (pl[0] && pl[1]);
	if (pl[0])
		*pol = pl[0];
	else
		*pol = pl[1];
}

static void pol_dedup(struct dev_policy *pol)
{
	/* This is a sorted list - remove duplicates. */
	while (pol && pol->next) {
		if (pol_lesseq(pol->next, pol)) {
			struct dev_policy *tmp = pol->next;
			pol->next = tmp->next;
			free(tmp);
		} else
			pol = pol->next;
	}
}

/*
 * pol_find finds the first entry in the policy
 * list to match name.
 * If it returns non-NULL there is at least one
 * value, but how many can only be found by
 * iterating through the list.
 */
struct dev_policy *pol_find(struct dev_policy *pol, const char *name)
{
	while (pol && pol->name < name)
		pol = pol->next;

	if (!pol || pol->name != name)
		return NULL;
	return pol;
}

static char *disk_path(struct mdinfo *disk)
{
	struct stat stb;
	int prefix_len;
	DIR *by_path;
	char d_symlink[PATH_MAX] = "/dev/disk/by-path/";
	char nm[PATH_MAX];
	struct dirent *ent;
	int rv;

	by_path = opendir(d_symlink);
	if (!by_path)
		return NULL;
	prefix_len = strlen(d_symlink);

	while ((ent = readdir(by_path)) != NULL) {
		if (ent->d_type != DT_LNK)
			continue;
		strncpy(d_symlink + prefix_len,
			ent->d_name,
			sizeof(d_symlink) - prefix_len);
		if (stat(d_symlink, &stb) < 0)
			continue;
		if ((stb.st_mode & S_IFMT) != S_IFBLK)
			continue;
		if (stb.st_rdev != makedev(disk->disk.major, disk->disk.minor))
			continue;
		closedir(by_path);
		return strdup(ent->d_name);
	}
	closedir(by_path);
	/* A NULL path isn't really acceptable - use the devname.. */
	sprintf(d_symlink, "/sys/dev/block/%d:%d", disk->disk.major, disk->disk.minor);
	rv = readlink(d_symlink, nm, sizeof(nm)-1);
	if (rv > 0) {
		char *dname;
		nm[rv] = 0;
		dname = strrchr(nm, '/');
		if (dname)
			return strdup(dname + 1);
	}
	return strdup("unknown");
}

static const char type_part[] = "part";
static const char type_disk[] = "disk";

static const char *disk_type(struct mdinfo *disk)
{
	char buf[30+20+20];
	struct stat stb;
	sprintf(buf, "/sys/dev/block/%d:%d/partition",
		disk->disk.major, disk->disk.minor);
	if (stat(buf, &stb) == 0)
		return type_part;
	else
		return type_disk;
}

static int pol_match(struct rule *rule, char *path, const char *type)
{
	/* check if this rule matches on path and type */
	int pathok = 0; /* 0 == no path, 1 == match, -1 == no match yet */
	int typeok = 0;

	while (rule) {
		if (rule->name == rule_path) {
			if (pathok == 0)
				pathok = -1;
			if (path && fnmatch(rule->value, path, 0) == 0)
				pathok = 1;
		}
		if (rule->name == rule_type) {
			if (typeok == 0)
				typeok = -1;
			if (type && strcmp(rule->value, type) == 0)
				typeok = 1;
		}
		rule = rule->next;
	}
	return pathok >= 0 && typeok >= 0;
}

static void pol_merge(struct dev_policy **pol, struct rule *rule)
{
	/* copy any name assignments from rule into pol */
	struct rule *r;
	char *metadata = NULL;
	for (r = rule; r ; r = r->next)
		if (r->name == pol_metadata)
			metadata = r->value;

	for (r = rule; r ; r = r->next)
		if (r->name == pol_act ||
		    r->name == pol_domain ||
		    r->name == pol_auto)
			pol_new(pol, r->name, r->value, metadata);
}

static int path_has_part(char *path, char **part)
{
	/* check if path ends with "-partNN" and
	 * if it does, place a pointer to "-pathNN"
	 * in 'part'.
	 */
	int l;
	if (!path)
		return 0;
	l = strlen(path);
	while (l > 1 && isdigit(path[l-1]))
		l--;
	if (l < 5 || strncmp(path+l-5, "-part", 5) != 0)
		return 0;
	*part = path+l-4;
	return 1;
}

static void pol_merge_part(struct dev_policy **pol, struct rule *rule, char *part)
{
	/* copy any name assignments from rule into pol, appending
	 * -part to any domain.  The string with -part appended is
	 * stored with the rule so it has a lifetime to match
	 * the rule.
	 */
	struct rule *r;
	char *metadata = NULL;
	for (r = rule; r ; r = r->next)
		if (r->name == pol_metadata)
			metadata = r->value;

	for (r = rule; r ; r = r->next) {
		if (r->name == pol_act)
			pol_new(pol, r->name, r->value, metadata);
		else if (r->name == pol_domain) {
			char *dom;
			int len;
			if (r->dups == NULL)
				r->dups = dl_head();
			len = strlen(r->value);
			for (dom = dl_next(r->dups); dom != r->dups;
			     dom = dl_next(dom))
				if (strcmp(dom+len+1, part)== 0)
					break;
			if (dom == r->dups) {
				char *newdom = dl_strndup(
					r->value, len + 1 + strlen(part));
				strcat(strcat(newdom, "-"), part);
				dl_add(r->dups, newdom);
				dom = newdom;
			}
			pol_new(pol, r->name, dom, metadata);
		}
	}
}


/*
 * most policy comes from a set policy rules that are
 * read from the config file.
 * path_policy() gathers policy information for the
 * disk described in the given a 'path' and a 'type'.
 */
static struct dev_policy *path_policy(char *path, const char *type)
{
	struct pol_rule *rules;
	struct dev_policy *pol = NULL;
	int i;

	rules = config_rules;

	while (rules) {
		char *part;
		if (rules->type == rule_policy)
			if (pol_match(rules->rule, path, type))
				pol_merge(&pol, rules->rule);
		if (rules->type == rule_part && strcmp(type, type_part) == 0)
			if (path_has_part(path, &part)) {
				*part = 0;
				if (pol_match(rules->rule, path, type_disk))
					pol_merge_part(&pol, rules->rule, part+1);
				*part = '-';
			}
		rules = rules->next;
	}

	/* Now add any metadata-specific internal knowledge
	 * about this path
	 */
	for (i=0; path && superlist[i]; i++)
		if (superlist[i]->get_disk_controller_domain) {
			const char *d =
				superlist[i]->get_disk_controller_domain(path);
			if (d)
				pol_new(&pol, pol_domain, d, superlist[i]->name);
		}

	pol_sort(&pol);
	pol_dedup(pol);
	return pol;
}

/*
 * disk_policy() gathers policy information for the
 * disk described in the given mdinfo (disk.{major,minor}).
 */
struct dev_policy *disk_policy(struct mdinfo *disk)
{
	char *path = NULL;
	const char *type = disk_type(disk);
	struct dev_policy *pol = NULL;

	if (config_rules_has_path)
		path = disk_path(disk);

	pol = path_policy(path, type);

	free(path);
	return pol;
}

struct dev_policy *devnum_policy(int dev)
{
	struct mdinfo disk;
	disk.disk.major = major(dev);
	disk.disk.minor = minor(dev);
	return disk_policy(&disk);
}

/*
 * process policy rules read from config file.
 */


static int try_rule(char *w, const char *name, struct rule **rp)
{
	struct rule *r;
	int len = strlen(name);
	if (strncmp(w, name, len) != 0 ||
	    w[len] != '=')
		return 0;
	r = xmalloc(sizeof(*r));
	r->next = *rp;
	r->name = name;
	r->value = strdup(w+len+1);
	r->dups = NULL;
	*rp = r;
	return 1;
}

void policyline(char *line, const char *type)
{
	struct pol_rule *pr;
	char *w;

	pr = xmalloc(sizeof(*pr));
	pr->type = type;
	pr->rule = NULL;
	for (w = dl_next(line); w != line ; w = dl_next(w)) {
		if (try_rule(w, rule_path, &pr->rule))
			config_rules_has_path = 1;
		else if (! try_rule(w, rule_type, &pr->rule) &&
			 ! try_rule(w, pol_metadata, &pr->rule) &&
			 ! try_rule(w, pol_act, &pr->rule) &&
			 ! try_rule(w, pol_domain, &pr->rule) &&
			 ! try_rule(w, pol_auto, &pr->rule))
			bb_error_msg("policy rule %s unrecognised and ignored",
				w);
	}
	pr->next = config_rules;
	config_rules = pr;
}

void policy_add(const char *type, ...)
{
	va_list ap;
	struct pol_rule *pr;
	char *name, *val;

	pr = xmalloc(sizeof(*pr));
	pr->type = type;
	pr->rule = NULL;

	va_start(ap, type);
	while ((name = va_arg(ap, char*)) != NULL) {
		struct rule *r;

		val = va_arg(ap, char*);
		r = xmalloc(sizeof(*r));
		r->next = pr->rule;
		r->name = name;
		r->value = strdup(val);
		r->dups = NULL;
		pr->rule = r;
	}
	pr->next = config_rules;
	config_rules = pr;
	va_end(ap);
}

void dev_policy_free(struct dev_policy *p)
{
	struct dev_policy *t;
	while (p) {
		t = p;
		p = p->next;
		free(t);
	}
}

static enum policy_action map_act(const char *act)
{
	if (strcmp(act, "include") == 0)
		return act_include;
	if (strcmp(act, "re-add") == 0)
		return act_re_add;
	if (strcmp(act, "spare") == 0)
		return act_spare;
	if (strcmp(act, "spare-same-slot") == 0)
		return act_spare_same_slot;
	if (strcmp(act, "force-spare") == 0)
		return act_force_spare;
	return act_err;
}

static enum policy_action policy_action(struct dev_policy *plist, const char *metadata)
{
	enum policy_action rv = act_default;
	struct dev_policy *p;

	plist = pol_find(plist, pol_act);
	pol_for_each(p, plist, metadata) {
		enum policy_action a = map_act(p->value);
		if (a > rv)
			rv = a;
	}
	return rv;
}

int policy_action_allows(struct dev_policy *plist, const char *metadata, enum policy_action want)
{
	enum policy_action act = policy_action(plist, metadata);

	if (act == act_err)
		return 0;
	return (act >= want);
}

int disk_action_allows(struct mdinfo *disk, const char *metadata, enum policy_action want)
{
	struct dev_policy *pol = disk_policy(disk);
	int rv = policy_action_allows(pol, metadata, want);

	dev_policy_free(pol);
	return rv;
}


/* Domain policy:
 * Any device can have a list of domains asserted by different policy
 * statements.
 * An array also has a list of domains comprising all the domains of
 * all the devices in an array.
 * Where an array has a spare-group, that becomes an addition domain for
 * every device in the array and thus for the array.
 *
 * We keep the list of domains in a sorted linked list
 * As dev policies are already sorted, this is fairly easy to manage.
 */

static struct domainlist **domain_merge_one(struct domainlist **domp,
					    const char *domain)
{
	/* merge a domain name into a sorted list and return the
	 * location of the insertion or match
	 */
	struct domainlist *dom = *domp;

	while (dom && strcmp(dom->dom, domain) < 0) {
		domp = &dom->next;
		dom = *domp;
	}
	if (dom == NULL || strcmp(dom->dom, domain) != 0) {
		dom = xmalloc(sizeof(*dom));
		dom->next = *domp;
		dom->dom = domain;
		*domp = dom;
	}
	return domp;
}

void domain_merge(struct domainlist **domp, struct dev_policy *pollist,
			 const char *metadata)
{
	/* Add to 'domp' all the domains in pol that apply to 'metadata'
	 * which are not already in domp
	 */
	struct dev_policy *pol;
	pollist = pol_find(pollist, pol_domain);
	pol_for_each(pol, pollist, metadata)
		domain_merge_one(domp, pol->value);
}

int domain_test(struct domainlist *dom, struct dev_policy *pol,
		const char *metadata)
{
	/* Check that all domains in pol (for metadata) are also in
	 * dom.  Both lists are sorted.
	 * If pol has no domains, we don't really know about this device
	 * so we allow caller to choose:
	 * -1:  has no domains
	 *  0:  has domains, not all match
	 *  1:  has domains, all match
	 */
	int found_any = -1;
	struct dev_policy *p;

	pol = pol_find(pol, pol_domain);
	pol_for_each(p, pol, metadata) {
		found_any = 1;
		while (dom && strcmp(dom->dom, p->value) < 0)
			dom = dom->next;
		if (!dom || strcmp(dom->dom, p->value) != 0)
			return 0;
	}
	return found_any;
}

void domain_free(struct domainlist *dl)
{
	while (dl) {
		struct domainlist *head = dl;
		dl = dl->next;
		free(head);
	}
}

/* doubly linked lists */
/* This is free software. No strings attached. No copyright claimed */

#ifdef __dietlibc__
char *strncpy(char *dest, const char *src, size_t n) __THROW;
#endif


void *dl_head()
{
    void *h;
    h = dl_alloc(0);
    dl_next(h) = h;
    dl_prev(h) = h;
    return h;
}

void dl_free(void *v)
{
    struct __dl_head *vv  = v;
    free(vv-1);
}

void dl_init(void *v)
{
    dl_next(v) = v;
    dl_prev(v) = v;
}

void dl_add(void *head, void *val)
{
    dl_prev(val) = dl_prev(head);
    dl_next(val) = head;
    dl_next(dl_prev(val)) = val;
    dl_prev(dl_next(val)) = val;
}

void dl_del(void *val)
{
    if (dl_prev(val) == 0 || dl_next(val) == 0)
	return;
    dl_prev(dl_next(val)) = dl_prev(val);
    dl_next(dl_prev(val)) = dl_next(val);
    dl_prev(val) = dl_next(val) = 0;
}

char *dl_strndup(char *s, int l)
{
    char *n;
    if (s == NULL)
	return NULL;
    n = dl_newv(char, l+1);
    if (n == NULL)
	return NULL;
    else
    {
	strncpy(n, s, l);
	n[l] = 0;
	return n;
    }
}

char *dl_strdup(char *s)
{
    return dl_strndup(s, (int)strlen(s));
}


/*
 * following taken from linux/blkpg.h because they aren't
 * anywhere else and it isn't safe to #include linux/ * stuff.
 */

#define BLKPG      _IO(0x12,105)

/* The argument structure */
struct blkpg_ioctl_arg {
	int op;
	int flags;
	int datalen;
	void *data;
};

/* The subfunctions (for the op field) */
#define BLKPG_ADD_PARTITION     1
#define BLKPG_DEL_PARTITION     2

/* Sizes of name fields. Unused at present. */
#define BLKPG_DEVNAMELTH        64
#define BLKPG_VOLNAMELTH        64

/* The data structure for ADD_PARTITION and DEL_PARTITION */
struct blkpg_partition {
	long long start;                /* starting offset in bytes */
	long long length;               /* length in bytes */
	int pno;                        /* partition number */
	char devname[BLKPG_DEVNAMELTH]; /* partition name, like sda5 or c0d1p2,
					   to be used in kernel messages */
	char volname[BLKPG_VOLNAMELTH]; /* volume label */
};


/* Force a compilation error if condition is true */
#define BUILD_BUG_ON(condition) ((void)BUILD_BUG_ON_ZERO(condition))

/* Force a compilation error if condition is true, but also produce a
   result (of value 0 and type size_t), so the expression can be used
   e.g. in a structure initializer (or where-ever else comma expressions
   aren't permitted). */
#define BUILD_BUG_ON_ZERO(e) (sizeof(struct { int:-!!(e); }))

/*
 * Parse a 128 bit uuid in 4 integers
 * format is 32 hexx nibbles with options :.<space> separator
 * If not exactly 32 hex digits are found, return 0
 * else return 1
 */
int parse_uuid(char *str, int uuid[4])
{
	int hit = 0; /* number of Hex digIT */
	int i;
	char c;
	for (i=0; i<4; i++) uuid[i]=0;

	while ((c= *str++)) {
		int n;
		if (c>='0' && c<='9')
			n = c-'0';
		else if (c>='a' && c <= 'f')
			n = 10 + c - 'a';
		else if (c>='A' && c <= 'F')
			n = 10 + c - 'A';
		else if (strchr(":. -", c))
			continue;
		else return 0;

		if (hit<32) {
			uuid[hit/8] <<= 4;
			uuid[hit/8] += n;
		}
		hit++;
	}
	if (hit == 32)
		return 1;
	return 0;
}


/*
 * Get the md version number.
 * We use the RAID_VERSION ioctl if it is supported
 * If not, but we have a block device with major '9', we assume
 * 0.36.0
 *
 * Return version number as 24 but number - assume version parts
 * always < 255
 */

int md_get_version(int fd)
{
    struct stat stb;
    mdu_version_t vers;

    if (fstat(fd, &stb)<0)
	return -1;
    if ((S_IFMT&stb.st_mode) != S_IFBLK)
	return -1;

    if (ioctl(fd, RAID_VERSION, &vers) == 0)
	return  (vers.major*10000) + (vers.minor*100) + vers.patchlevel;
    if (errno == EACCES)
	    return -1;
    if (major(stb.st_rdev) == MD_MAJOR)
	return (3600);
    return -1;
}

void remove_partitions(int fd)
{
	/* remove partitions from this block devices.
	 * This is used for components added to an array
	 */
#ifdef BLKPG_DEL_PARTITION
	struct blkpg_ioctl_arg a;
	struct blkpg_partition p;

	a.op = BLKPG_DEL_PARTITION;
	a.data = (void*)&p;
	a.datalen = sizeof(p);
	a.flags = 0;
	memset(a.data, 0, a.datalen);
	for (p.pno=0; p.pno < 16; p.pno++)
		ioctl(fd, BLKPG, &a);
#endif
}

int test_partition(int fd)
{
	/* Check if fd is a whole-disk or a partition.
	 * BLKPG will return EINVAL on a partition, and BLKPG_DEL_PARTITION
	 * will return ENXIO on an invalid partition number.
	 */
	struct blkpg_ioctl_arg a;
	struct blkpg_partition p;
	a.op = BLKPG_DEL_PARTITION;
	a.data = (void*)&p;
	a.datalen = sizeof(p);
	a.flags = 0;
	memset(a.data, 0, a.datalen);
	p.pno = 1<<30;
	if (ioctl(fd, BLKPG, &a) == 0)
		/* Very unlikely, but not a partition */
		return 0;
	if (errno == ENXIO)
		/* not a partition */
		return 0;

	return 1;
}

int enough(int level, int raid_disks, int layout, int clean, char *avail)
{
	int copies, first;
	int i;
	int avail_disks = 0;

	for (i = 0; i < raid_disks; i++)
		avail_disks += !!avail[i];

	switch (level) {
	case 10:
		/* This is the tricky one - we need to check
		 * which actual disks are present.
		 */
		copies = (layout&255)* ((layout>>8) & 255);
		first=0;
		do {
			/* there must be one of the 'copies' form 'first' */
			int n = copies;
			int cnt=0;
			int this = first;
			while (n--) {
				if (avail[this])
					cnt++;
				this = (this+1) % raid_disks;
			}
			if (cnt == 0)
				return 0;
			first = (first+(layout&255)) % raid_disks;
		} while (first != 0);
		return 1;

	case LEVEL_MULTIPATH:
		return avail_disks>= 1;
	case LEVEL_LINEAR:
	case 0:
		return avail_disks == raid_disks;
	case 1:
		return avail_disks >= 1;
	case 4:
	case 5:
		if (clean)
			return avail_disks >= raid_disks-1;
		else
			return avail_disks >= raid_disks;
	case 6:
		if (clean)
			return avail_disks >= raid_disks-2;
		else
			return avail_disks >= raid_disks;
	default:
		return 0;
	}
}


int same_uuid(int a[4], int b[4], int swapuuid)
{
	if (swapuuid) {
		/* parse uuids are hostendian.
		 * uuid's from some superblocks are big-ending
		 * if there is a difference, we need to swap..
		 */
		unsigned char *ac = (unsigned char *)a;
		unsigned char *bc = (unsigned char *)b;
		int i;
		for (i=0; i<16; i+= 4) {
			if (ac[i+0] != bc[i+3] ||
			    ac[i+1] != bc[i+2] ||
			    ac[i+2] != bc[i+1] ||
			    ac[i+3] != bc[i+0])
				return 0;
		}
		return 1;
	} else {
		if (a[0]==b[0] &&
		    a[1]==b[1] &&
		    a[2]==b[2] &&
		    a[3]==b[3])
			return 1;
		return 0;
	}
}
void copy_uuid(void *a, int b[4], int swapuuid)
{
	if (swapuuid) {
		/* parse uuids are hostendian.
		 * uuid's from some superblocks are big-ending
		 * if there is a difference, we need to swap..
		 */
		unsigned char *ac = (unsigned char *)a;
		unsigned char *bc = (unsigned char *)b;
		int i;
		for (i=0; i<16; i+= 4) {
			ac[i+0] = bc[i+3];
			ac[i+1] = bc[i+2];
			ac[i+2] = bc[i+1];
			ac[i+3] = bc[i+0];
		}
	} else
		memcpy(a, b, 16);
}

unsigned long calc_csum(void *super, int bytes)
{
	unsigned long long newcsum = 0;
	int i;
	unsigned int csum;
	unsigned int *superc = (unsigned int*) super;

	for(i=0; i<bytes/4; i++)
		newcsum+= superc[i];
	csum = (newcsum& 0xffffffff) + (newcsum>>32);
#ifdef __alpha__
/* The in-kernel checksum calculation is always 16bit on
 * the alpha, though it is 32 bit on i386...
 * I wonder what it is elsewhere... (it uses and API in
 * a way that it shouldn't).
 */
	csum = (csum & 0xffff) + (csum >> 16);
	csum = (csum & 0xffff) + (csum >> 16);
#endif
	return csum;
}

int dev_open(char *dev, int flags)
{
	/* like 'open', but if 'dev' matches %d:%d, create a temp
	 * block device and open that
	 */
	char *e;
	int fd = -1;
	char devname[32];
	int major;
	int minor;

	if (!dev) return -1;
	flags |= O_DIRECT;

	major = strtoul(dev, &e, 0);
	if (e > dev && *e == ':' && e[1] &&
	    (minor = strtoul(e+1, &e, 0)) >= 0 &&
	    *e == 0) {
		char *path = map_dev(major, minor, 0);
		if (path)
			fd = open(path, flags);
		if (fd < 0) {
			snprintf(devname, sizeof(devname), "/dev/.tmp.md.%d:%d:%d",
				 (int)getpid(), major, minor);
			if (mknod(devname, S_IFBLK|0600, makedev(major, minor))==0) {
				fd = open(devname, flags);
				unlink(devname);
			}
		}
		if (fd < 0) {
			snprintf(devname, sizeof(devname), "/tmp/.tmp.md.%d:%d:%d",
				 (int)getpid(), major, minor);
			if (mknod(devname, S_IFBLK|0600, makedev(major, minor))==0) {
				fd = open(devname, flags);
				unlink(devname);
			}
		}
	} else
		fd = open(dev, flags);
	return fd;
}

int same_dev(char *one, char *two)
{
	struct stat st1, st2;
	if (stat(one, &st1) != 0)
		return 0;
	if (stat(two, &st2) != 0)
		return 0;
	if ((st1.st_mode & S_IFMT) != S_IFBLK)
		return 0;
	if ((st2.st_mode & S_IFMT) != S_IFBLK)
		return 0;
	return st1.st_rdev == st2.st_rdev;
}

void wait_for(char *dev, int fd)
{
	int i;
	struct stat stb_want;

	if (fstat(fd, &stb_want) != 0 ||
	    (stb_want.st_mode & S_IFMT) != S_IFBLK)
		return;

	for (i=0 ; i<25 ; i++) {
		struct stat stb;
		if (stat(dev, &stb) == 0 &&
		    (stb.st_mode & S_IFMT) == S_IFBLK &&
		    (stb.st_rdev == stb_want.st_rdev))
			return;
		xusleep(200000);
	}
	if (i == 25)
		dprintf("%s: timeout waiting for %s\n", __func__, dev);
}


struct supertype *dup_super(struct supertype *orig)
{
	struct supertype *st;

	if (!orig)
		return orig;
	st = zmalloc(sizeof(*st));
	st->ss = orig->ss;
	st->max_devs = orig->max_devs;
	st->minor_version = orig->minor_version;
	st->sb = NULL;
	st->info = NULL;
	return st;
}

struct supertype *guess_super_type(int fd, enum guess_types guess_type)
{
	/* try each load_super to find the best match,
	 * and return the best superswitch
	 */
	const struct superswitch  *ss;
	struct supertype *st;
	time_t besttime = 0;
	int bestsuper = -1;
	int i;

	st = zmalloc(sizeof(*st));
	st->container_dev = NoMdDev;

	for (i=0 ; superlist[i]; i++) {
		int rv;
		ss = superlist[i];
		if (guess_type == guess_array && ss->add_to_super == NULL)
			continue;
		if (guess_type == guess_partitions && ss->add_to_super != NULL)
			continue;
		memset(st, 0, sizeof(*st));
		st->ignore_hw_compat = 1;
		rv = ss->load_super(st, fd, NULL);
		if (rv == 0) {
			struct mdinfo info;
			st->ss->getinfo_super(st, &info, NULL);
			if (bestsuper == -1 ||
			    besttime < info.array.ctime) {
				bestsuper = i;
				besttime = info.array.ctime;
			}
			ss->free_super(st);
		}
	}
	if (bestsuper != -1) {
		int rv;
		memset(st, 0, sizeof(*st));
		st->ignore_hw_compat = 1;
		rv = superlist[bestsuper]->load_super(st, fd, NULL);
		if (rv == 0) {
			superlist[bestsuper]->free_super(st);
			st->ignore_hw_compat = 0;
			return st;
		}
	}
	free(st);
	return NULL;
}

/* Return size of device in bytes */
int get_dev_size(int fd, char *dname, unsigned long long *sizep)
{
	unsigned long long ldsize;
	struct stat st;

	if (fstat(fd, &st) != -1 && S_ISREG(st.st_mode))
		ldsize = (unsigned long long)st.st_size;
	else
#ifdef BLKGETSIZE64
	if (ioctl(fd, BLKGETSIZE64, &ldsize) != 0)
#endif
	{
		unsigned long dsize;
		if (ioctl(fd, BLKGETSIZE, &dsize) == 0) {
			ldsize = dsize;
			ldsize <<= 9;
		} else {
			if (dname)
				bb_perror_msg("Cannot get size of %s", dname);
			return 0;
		}
	}
	*sizep = ldsize;
	return 1;
}

/* Return true if this can only be a container, not a member device.
 * i.e. is and md device and size is zero
 */
int must_be_container(int fd)
{
	unsigned long long size;
	if (md_get_version(fd) < 0)
		return 0;
	if (get_dev_size(fd, NULL, &size) == 0)
		return 1;
	if (size == 0)
		return 1;
	return 0;
}

int add_disk(int mdfd, struct mdinfo *info)
{
	/* Add a device to an array, in one of 2 ways. */
	return ioctl(mdfd, ADD_NEW_DISK, &info->disk);
}

int set_array_info(int mdfd, struct mdinfo *info)
{
	/* Initialise kernel's knowledge of array.
	 * This varies between externally managed arrays
	 * and older kernels
	 */
	int vers = md_get_version(mdfd);
	int rv;

	if ((vers % 100) >= 1) { /* can use different versions */
		mdu_array_info_t inf;
		memset(&inf, 0, sizeof(inf));
		inf.major_version = info->array.major_version;
		inf.minor_version = info->array.minor_version;
		rv = ioctl(mdfd, SET_ARRAY_INFO, &inf);
	} else
		rv = ioctl(mdfd, SET_ARRAY_INFO, NULL);
	return rv;
}

int check_env(char *name)
{
	char *val = getenv(name);

	if (val && atoi(val) == 1)
		return 1;

	return 0;
}

#ifdef __TINYC__
/* tinyc doesn't optimize this check in ioctl.h out ... */
unsigned int __invalid_size_argument_for_IOC = 0;
#endif

/* name/number mappings */


/* This fill contains various 'library' style function.  They
 * have no dependency on anything outside this file.
 */

int get_mdp_major(void)
{
	FILE *fl;
	char *w;
	int have_block = 0;
	int have_devices = 0;
	int last_num = -1;

	if (mdp_major != -1)
		return mdp_major;
	fl = fopen("/proc/devices", "r");
	if (!fl)
		return -1;
	while ((w = conf_word(fl, 1))) {
		if (have_block && strcmp(w, "devices:")==0)
			have_devices = 1;
		have_block =  (strcmp(w, "Block")==0);
		if (isdigit(w[0]))
			last_num = atoi(w);
		if (have_devices && strcmp(w, "mdp")==0)
			mdp_major = last_num;
		free(w);
	}
	fclose(fl);
	return mdp_major;
}


void fmt_devname(char *name, int num)
{
	if (num >= 0)
		sprintf(name, "md%d", num);
	else
		sprintf(name, "md_d%d", -1-num);
}

char *devnum2devname(int num)
{
	char name[100];
	fmt_devname(name,num);
	return strdup(name);
}

int devname2devnum(char *name)
{
	char *ep;
	int num;
	if (strncmp(name, "md_d", 4)==0)
		num = -1-strtoul(name+4, &ep, 10);
	else
		num = strtoul(name+2, &ep, 10);
	return num;
}

int stat2devnum(struct stat *st)
{
	char path[30];
	char sys_link[200];
	char *cp;
	int n;

	if ((S_IFMT & st->st_mode) == S_IFBLK) {
		if (major(st->st_rdev) == MD_MAJOR)
			return minor(st->st_rdev);
		else if (major(st->st_rdev) == (unsigned)get_mdp_major())
			return -1- (minor(st->st_rdev)>>MdpMinorShift);

		/* must be an extended-minor partition. Look at the
		 * /sys/dev/block/%d:%d link which must look like
		 * ../../block/mdXXX/mdXXXpYY
		 */
		sprintf(path, "/sys/dev/block/%d:%d", major(st->st_rdev),
			minor(st->st_rdev));
		n = readlink(path, sys_link, sizeof(sys_link)-1);
		if (n <= 0)
			return NoMdDev;
		sys_link[n] = 0;
		cp = strrchr(sys_link, '/');
		if (cp) *cp = 0;
		cp = strrchr(sys_link, '/');
		if (cp && strncmp(cp, "/md", 3) == 0)
			return devname2devnum(cp+1);
	}
	return NoMdDev;

}

int fd2devnum(int fd)
{
	struct stat stb;
	if (fstat(fd, &stb) == 0)
		return stat2devnum(&stb);
	return NoMdDev;
}



/*
 * convert a major/minor pair for a block device into a name in /dev, if possible.
 * On the first call, walk /dev collecting name.
 * Put them in a simple linked listfor now.
 */

static int add_dev(const char *name, const struct stat *stb, int flag, struct FTW *s)
{
	struct stat st;

	(void)&flag;
	(void)&s;

	if (S_ISLNK(stb->st_mode)) {
		if (stat(name, &st) != 0)
			return 0;
		stb = &st;
	}

	if ((stb->st_mode&S_IFMT)== S_IFBLK) {
		char *n = strdup(name);
		struct devmap *dm = xmalloc(sizeof(*dm));
		if (strncmp(n, "/dev/./", 7)==0)
			strcpy(n+4, name+6);
		dm->major = major(stb->st_rdev);
		dm->minor = minor(stb->st_rdev);
		dm->name = n;
		dm->next = devmap_devlist;
		devmap_devlist = dm;
	}
	return 0;
}

#ifndef HAVE_NFTW
#ifdef HAVE_FTW
int add_dev_1(const char *name, const struct stat *stb, int flag)
{
	return add_dev(name, stb, flag, NULL);
}
int nftw(const char *path, int (*han)(const char *name, const struct stat *stb, int flag, struct FTW *s), int nopenfd, int flags)
{
	return ftw(path, add_dev_1, nopenfd);
}
#else
int nftw(const char *path, int (*han)(const char *name, const struct stat *stb, int flag, struct FTW *s), int nopenfd, int flags)
{
	return 0;
}
#endif /* HAVE_FTW */
#endif /* HAVE_NFTW */

/*
 * Find a block device with the right major/minor number.
 * If we find multiple names, choose the shortest.
 * If we find a name in /dev/md/, we prefer that.
 * This applies only to names for MD devices.
 * If 'prefer' is set (normally to e.g. /by-path/)
 * then we prefer a name which contains that string.
 */
char *map_dev(int major, int minor, int create)
{
	struct devmap *p;
	char *regular = NULL, *preferred=NULL;
	int did_check = 0;

	if (major == 0 && minor == 0)
			return NULL;

 retry:
	if (!devlist_ready) {
		char *dev = "/dev";
		struct stat stb;
		while(devmap_devlist) {
			struct devmap *d = devmap_devlist;
			devmap_devlist = d->next;
			free(d->name);
			free(d);
		}
		if (lstat(dev, &stb)==0 &&
		    S_ISLNK(stb.st_mode))
			dev = "/dev/.";
		nftw(dev, add_dev, 10, FTW_PHYS);
		devlist_ready=1;
		did_check = 1;
	}

	for (p=devmap_devlist; p; p=p->next)
		if (p->major == major &&
		    p->minor == minor) {
			if (strncmp(p->name, "/dev/md/",8) == 0) {
				if (preferred == NULL ||
				    strlen(p->name) < strlen(preferred))
					preferred = p->name;
			} else {
				if (regular == NULL ||
				    strlen(p->name) < strlen(regular))
					regular = p->name;
			}
		}
	if (!regular && !preferred && !did_check) {
		devlist_ready = 0;
		goto retry;
	}
	if (create && !regular && !preferred) {
		snprintf(major_minor_buf, sizeof(major_minor_buf), "%d:%d", major, minor);
		regular = major_minor_buf;
	}

	return preferred ? preferred : regular;
}



/* conf_word gets one word from the conf file.
 * if "allow_key", then accept words at the start of a line,
 * otherwise stop when such a word is found.
 * We assume that the file pointer is at the end of a word, so the
 * next character is a space, or a newline.  If not, it is the start of a line.
 */

char *conf_word(FILE *file, int allow_key)
{
	int wsize = 100;
	int len = 0;
	int c;
	int quote;
	int wordfound = 0;
	char *word = xmalloc(wsize);

	while (wordfound==0) {
		/* at the end of a word.. */
		c = getc(file);
		if (c == '#')
			while (c != EOF && c != '\n')
				c = getc(file);
		if (c == EOF) break;
		if (c == '\n') continue;

		if (c != ' ' && c != '\t' && ! allow_key) {
			ungetc(c, file);
			break;
		}
		/* looks like it is safe to get a word here, if there is one */
		quote = 0;
		/* first, skip any spaces */
		while (c == ' ' || c == '\t')
			c = getc(file);
		if (c != EOF && c != '\n' && c != '#') {
			/* we really have a character of a word, so start saving it */
			while (c != EOF && c != '\n' && (quote || (c!=' ' && c != '\t'))) {
				wordfound = 1;
				if (quote && c == quote) quote = 0;
				else if (quote == 0 && (c == '\'' || c == '"'))
					quote = c;
				else {
					if (len == wsize-1) {
						wsize += 100;
						word = realloc(word, wsize);
						if (!word) abort();
					}
					word[len++] = c;
				}
				c = getc(file);
				/* Hack for broken kernels (2.6.14-.24) that put
				 *        "active(auto-read-only)"
				 * in /proc/mdstat instead of
				 *        "active (auto-read-only)"
				 */
				if (c == '(' && len >= 6
				    && strncmp(word+len-6, "active", 6) == 0)
					c = ' ';
			}
		}
		if (c != EOF) ungetc(c, file);
	}
	word[len] = 0;

	/* Further HACK for broken kernels.. 2.6.14-2.6.24 */
	if (strcmp(word, "auto-read-only)") == 0)
		strcpy(word, "(auto-read-only)");

/*    printf("word is <%s>\n", word); */
	if (!wordfound) {
		free(word);
		word = NULL;
	}
	return word;
}

/*
 * All handling for the 0.90.0 version superblock is in
 * this file.
 * This includes:
 *   - finding, loading, and writing the superblock.
 *   - initialising a new superblock
 *   - printing the superblock for --examine
 *   - printing part of the superblock for --detail
 * .. other stuff
 */


static unsigned long calc_sb0_csum(mdp_super_t *super)
{
	unsigned long csum = super->sb_csum;
	unsigned long newcsum;
	super->sb_csum= 0 ;
	newcsum = calc_csum(super, MD_SB_BYTES);
	super->sb_csum = csum;
	return newcsum;
}


static void super0_swap_endian(struct mdp_superblock_s *sb)
{
	/* as super0 superblocks are host-endian, it is sometimes
	 * useful to be able to swap the endianness
	 * as (almost) everything is u32's we byte-swap every 4byte
	 * number.
	 * We then also have to swap the events_hi and events_lo
	 */
	char *sbc = (char *)sb;
	__u32 t32;
	int i;

	for (i=0; i < MD_SB_BYTES ; i+=4) {
		char t = sbc[i];
		sbc[i] = sbc[i+3];
		sbc[i+3] = t;
		t=sbc[i+1];
		sbc[i+1]=sbc[i+2];
		sbc[i+2]=t;
	}
	t32 = sb->events_hi;
	sb->events_hi = sb->events_lo;
	sb->events_lo = t32;

	t32 = sb->cp_events_hi;
	sb->cp_events_hi = sb->cp_events_lo;
	sb->cp_events_lo = t32;

}

static int match_home0(struct supertype *st, char *homehost)
{
	mdp_super_t *sb = st->sb;
	char buf[20];
	char *hash;

	if (!homehost)
		return 0;
	hash = sha1_buffer(homehost, strlen(homehost), buf);

	return (memcmp(&sb->set_uuid2, hash, 8)==0);
}

static void uuid_from_super0(struct supertype *st, int uuid[4])
{
	mdp_super_t *super = st->sb;
	uuid[0] = super->set_uuid0;
	if (super->minor_version >= 90) {
		uuid[1] = super->set_uuid1;
		uuid[2] = super->set_uuid2;
		uuid[3] = super->set_uuid3;
	} else {
		uuid[1] = 0;
		uuid[2] = 0;
		uuid[3] = 0;
	}
}

static void getinfo_super0(struct supertype *st, struct mdinfo *info, char *map)
{
	mdp_super_t *sb = st->sb;
	int working = 0;
	int i;
	int map_disks = info->array.raid_disks;

	memset(info, 0, sizeof(*info));
	info->array.major_version = sb->major_version;
	info->array.minor_version = sb->minor_version;
	info->array.patch_version = sb->patch_version;
	info->array.raid_disks = sb->raid_disks;
	info->array.level = sb->level;
	info->array.layout = sb->layout;
	info->array.md_minor = sb->md_minor;
	info->array.ctime = sb->ctime;
	info->array.utime = sb->utime;
	info->array.chunk_size = sb->chunk_size;
	info->array.state = sb->state;
	info->component_size = sb->size*2;

	if (sb->state & (1<<MD_SB_BITMAP_PRESENT))
		info->bitmap_offset = 8;

	info->disk.state = sb->this_disk.state;
	info->disk.major = sb->this_disk.major;
	info->disk.minor = sb->this_disk.minor;
	info->disk.raid_disk = sb->this_disk.raid_disk;
	info->disk.number = sb->this_disk.number;

	info->events = md_event(sb);
	info->data_offset = 0;

	sprintf(info->text_version, "0.%d", sb->minor_version);
	info->safe_mode_delay = 200;

	uuid_from_super0(st, info->uuid);

	info->recovery_start = MaxSector;
	if (sb->minor_version > 90 && (sb->reshape_position+1) != 0) {
		info->reshape_active = 1;
		info->reshape_progress = sb->reshape_position;
		info->new_level = sb->new_level;
		info->delta_disks = sb->delta_disks;
		info->new_layout = sb->new_layout;
		info->new_chunk = sb->new_chunk;
		if (info->delta_disks < 0)
			info->array.raid_disks -= info->delta_disks;
	} else
		info->reshape_active = 0;

	info->recovery_blocked = info->reshape_active;

	sprintf(info->name, "%d", sb->md_minor);
	/* work_disks is calculated rather than read directly */
	for (i=0; i < MD_SB_DISKS; i++)
		if ((sb->disks[i].state & (1<<MD_DISK_SYNC)) &&
		    (sb->disks[i].raid_disk < (unsigned)info->array.raid_disks) &&
		    (sb->disks[i].state & (1<<MD_DISK_ACTIVE)) &&
		    !(sb->disks[i].state & (1<<MD_DISK_FAULTY))) {
			working ++;
			if (map && i < map_disks)
				map[i] = 1;
		} else if (map && i < map_disks)
			map[i] = 0;
	info->array.working_disks = working;
}

static struct mdinfo *container_content0(struct supertype *st, char *subarray)
{
	struct mdinfo *info;

	if (subarray)
		return NULL;

	info = xmalloc(sizeof(*info));
	getinfo_super0(st, info, NULL);
	return info;
}

static int update_super0(struct supertype *st, struct mdinfo *info,
			 char *update,
			 char *devname, int verbose,
			 int uuid_set, char *homehost)
{
	/* NOTE: for 'assemble' and 'force' we need to return non-zero
	 * if any change was made.  For others, the return value is
	 * ignored.
	 */
	int rv = 0;
	int uuid[4];
	mdp_super_t *sb = st->sb;
	if (strcmp(update, "sparc2.2")==0 ) {
		/* 2.2 sparc put the events in the wrong place
		 * So we copy the tail of the superblock
		 * up 4 bytes before continuing
		 */
		__u32 *sb32 = (__u32*)sb;
		memcpy(sb32+MD_SB_GENERIC_CONSTANT_WORDS+7,
		       sb32+MD_SB_GENERIC_CONSTANT_WORDS+7+1,
		       (MD_SB_WORDS - (MD_SB_GENERIC_CONSTANT_WORDS+7+1))*4);
		if (verbose >= 0)
			bb_error_msg("adjusting superblock of %s for 2.2/sparc compatibility.",
				 devname);
	} else if (strcmp(update, "super-minor") ==0) {
		sb->md_minor = info->array.md_minor;
		if (verbose > 0)
			bb_error_msg("updating superblock of %s with minor number %d",
				devname, info->array.md_minor);
	} else if (strcmp(update, "summaries") == 0) {
		unsigned int i;
		/* set nr_disks, active_disks, working_disks,
		 * failed_disks, spare_disks based on disks[]
		 * array in superblock.
		 * Also make sure extra slots aren't 'failed'
		 */
		sb->nr_disks = sb->active_disks =
			sb->working_disks = sb->failed_disks =
			sb->spare_disks = 0;
		for (i=0; i < MD_SB_DISKS ; i++)
			if (sb->disks[i].major ||
			    sb->disks[i].minor) {
				int state = sb->disks[i].state;
				if (state & (1<<MD_DISK_REMOVED))
					continue;
				sb->nr_disks++;
				if (state & (1<<MD_DISK_ACTIVE))
					sb->active_disks++;
				if (state & (1<<MD_DISK_FAULTY))
					sb->failed_disks++;
				else
					sb->working_disks++;
				if (state == 0)
					sb->spare_disks++;
			} else if (i >= sb->raid_disks && sb->disks[i].number == 0)
				sb->disks[i].state = 0;
	} else if (strcmp(update, "force-one")==0) {
		/* Not enough devices for a working array, so
		 * bring this one up-to-date.
		 */
		__u32 ehi = sb->events_hi, elo = sb->events_lo;
		sb->events_hi = (info->events>>32) & 0xFFFFFFFF;
		sb->events_lo = (info->events) & 0xFFFFFFFF;
		if (sb->events_hi != ehi ||
		    sb->events_lo != elo)
			rv = 1;
	} else if (strcmp(update, "force-array")==0) {
		/* degraded array and 'force' requested, so
		 * maybe need to mark it 'clean'
		 */
		if ((sb->level == 5 || sb->level == 4 || sb->level == 6) &&
		    (sb->state & (1 << MD_SB_CLEAN)) == 0) {
			/* need to force clean */
			sb->state |= (1 << MD_SB_CLEAN);
			rv = 1;
		}
	} else if (strcmp(update, "assemble")==0) {
		int d = info->disk.number;
		int wonly = sb->disks[d].state & (1<<MD_DISK_WRITEMOSTLY);
		int mask = (1<<MD_DISK_WRITEMOSTLY);
		int add = 0;
		if (sb->minor_version >= 91)
			/* During reshape we don't insist on everything
			 * being marked 'sync'
			 */
			add = (1<<MD_DISK_SYNC);
		if (((sb->disks[d].state & ~mask) | add)
		    != (unsigned)info->disk.state) {
			sb->disks[d].state = info->disk.state | wonly;
			rv = 1;
		}
		if (info->reshape_active &&
		    sb->minor_version > 90 && (sb->reshape_position+1) != 0 &&
		    info->delta_disks >= 0 &&
		    info->reshape_progress < sb->reshape_position) {
			sb->reshape_position = info->reshape_progress;
			rv = 1;
		}
		if (info->reshape_active &&
		    sb->minor_version > 90 && (sb->reshape_position+1) != 0 &&
		    info->delta_disks < 0 &&
		    info->reshape_progress > sb->reshape_position) {
			sb->reshape_position = info->reshape_progress;
			rv = 1;
		}
	} else if (strcmp(update, "linear-grow-new") == 0) {
		memset(&sb->disks[info->disk.number], 0, sizeof(sb->disks[0]));
		sb->disks[info->disk.number].number = info->disk.number;
		sb->disks[info->disk.number].major = info->disk.major;
		sb->disks[info->disk.number].minor = info->disk.minor;
		sb->disks[info->disk.number].raid_disk = info->disk.raid_disk;
		sb->disks[info->disk.number].state = info->disk.state;
		sb->this_disk = sb->disks[info->disk.number];
	} else if (strcmp(update, "linear-grow-update") == 0) {
		sb->raid_disks = info->array.raid_disks;
		sb->nr_disks = info->array.nr_disks;
		sb->active_disks = info->array.active_disks;
		sb->working_disks = info->array.working_disks;
		memset(&sb->disks[info->disk.number], 0, sizeof(sb->disks[0]));
		sb->disks[info->disk.number].number = info->disk.number;
		sb->disks[info->disk.number].major = info->disk.major;
		sb->disks[info->disk.number].minor = info->disk.minor;
		sb->disks[info->disk.number].raid_disk = info->disk.raid_disk;
		sb->disks[info->disk.number].state = info->disk.state;
	} else if (strcmp(update, "resync") == 0) {
		/* make sure resync happens */
		sb->state &= ~(1<<MD_SB_CLEAN);
		sb->recovery_cp = 0;
	} else if (strcmp(update, "homehost") == 0 &&
		   homehost) {
		uuid_set = 0;
		update = "uuid";
		info->uuid[0] = sb->set_uuid0;
		info->uuid[1] = sb->set_uuid1;
	} else if (strcmp(update, "uuid") == 0) {
		if (!uuid_set && homehost) {
			char buf[20];
			char *hash = sha1_buffer(homehost, strlen(homehost), buf);
			memcpy(info->uuid+2, hash, 8);
		}
		sb->set_uuid0 = info->uuid[0];
		sb->set_uuid1 = info->uuid[1];
		sb->set_uuid2 = info->uuid[2];
		sb->set_uuid3 = info->uuid[3];
		if (sb->state & (1<<MD_SB_BITMAP_PRESENT)) {
			struct bitmap_super_s *bm;
			bm = (struct bitmap_super_s*)(sb+1);
			uuid_from_super0(st, uuid);
			memcpy(bm->uuid, uuid, 16);
		}
	} else if (strcmp(update, "no-bitmap") == 0) {
		sb->state &= ~(1<<MD_SB_BITMAP_PRESENT);
	} else if (strcmp(update, "_reshape_progress")==0)
		sb->reshape_position = info->reshape_progress;
	else if (strcmp(update, "writemostly")==0)
		sb->state |= (1<<MD_DISK_WRITEMOSTLY);
	else if (strcmp(update, "readwrite")==0)
		sb->state &= ~(1<<MD_DISK_WRITEMOSTLY);
	else
		rv = -1;

	sb->sb_csum = calc_sb0_csum(sb);
	return rv;
}

struct devinfo {
	int fd;
	char *devname;
	mdu_disk_info_t disk;
	struct devinfo *next;
};

static int store_super0(struct supertype *st, int fd)
{
	unsigned long long dsize;
	unsigned long long offset;
	mdp_super_t *super = st->sb;

	if (!get_dev_size(fd, NULL, &dsize))
		return 1;

	if (dsize < MD_RESERVED_SECTORS*512)
		return 2;

	offset = MD_NEW_SIZE_SECTORS(dsize>>9);

	offset *= 512;

	if (lseek(fd, offset, 0)< 0LL)
		return 3;

	if (write(fd, super, sizeof(*super)) != sizeof(*super))
		return 4;

	if (super->state & (1<<MD_SB_BITMAP_PRESENT)) {
		struct bitmap_super_s * bm = (struct bitmap_super_s*)(super+1);
		if (__le32_to_cpu(bm->magic) == BITMAP_MAGIC)
			if (write(fd, bm, ROUND_UP(sizeof(*bm),4096)) !=
			    ROUND_UP(sizeof(*bm),4096))
			    return 5;
	}

	fsync(fd);
	return 0;
}

static int compare_super0(struct supertype *st, struct supertype *tst)
{
	/*
	 * return:
	 *  0 same, or first was empty, and second was copied
	 *  1 second had wrong number
	 *  2 wrong uuid
	 *  3 wrong other info
	 */
	mdp_super_t *first = st->sb;
	mdp_super_t *second = tst->sb;
	int uuid1[4], uuid2[4];

	if (second->md_magic != MD_SB_MAGIC)
		return 1;
	if (!first) {
		if (posix_memalign((void**)&first, 4096,
			     MD_SB_BYTES +
			     ROUND_UP(sizeof(struct bitmap_super_s), 4096)) != 0) {
			bb_error_msg("%s could not allocate superblock", __func__);
			return 1;
		}
		memcpy(first, second, MD_SB_BYTES + sizeof(struct bitmap_super_s));
		st->sb = first;
		return 0;
	}

	uuid_from_super0(st, uuid1);
	uuid_from_super0(tst, uuid2);
	if (!same_uuid(uuid1, uuid2, 0))
		return 2;
	if (first->major_version != second->major_version ||
	    first->minor_version != second->minor_version ||
	    first->patch_version != second->patch_version ||
	    first->gvalid_words  != second->gvalid_words  ||
	    first->ctime         != second->ctime         ||
	    first->level         != second->level         ||
	    first->size          != second->size          ||
	    first->raid_disks    != second->raid_disks    )
		return 3;

	return 0;
}


static void free_super0(struct supertype *st);

static int load_super0(struct supertype *st, int fd, char *devname)
{
	/* try to read in the superblock
	 * Return:
	 *  0 on success
	 *  1 on cannot get superblock
	 *  2 on superblock meaningless
	 */
	unsigned long long dsize;
	unsigned long long offset;
	mdp_super_t *super;
	int uuid[4];
	struct bitmap_super_s *bsb;

	free_super0(st);

	if (!get_dev_size(fd, devname, &dsize))
		return 1;

	if (dsize < MD_RESERVED_SECTORS*512) {
		if (devname)
			bb_error_msg("%s is too small for md: size is %llu sectors.",
				devname, dsize);
		return 1;
	}

	offset = MD_NEW_SIZE_SECTORS(dsize>>9);

	offset *= 512;

	ioctl(fd, BLKFLSBUF, 0); /* make sure we read current data */

	if (lseek(fd, offset, 0)< 0LL) {
		if (devname)
			bb_perror_msg("Cannot seek to superblock on %s", devname);
		return 1;
	}

	if (posix_memalign((void**)&super, 4096,
			   MD_SB_BYTES +
			   ROUND_UP(sizeof(bitmap_super_t), 4096)) != 0) {
		bb_error_msg("%s could not allocate superblock", __func__);
		return 1;
	}

	if (read(fd, super, sizeof(*super)) != MD_SB_BYTES) {
		if (devname)
			bb_perror_msg("Cannot read superblock on %s",
				devname);
		free(super);
		return 1;
	}

	if (st->ss && st->minor_version == 9)
		super0_swap_endian(super);

	if (super->md_magic != MD_SB_MAGIC) {
		if (devname)
			bb_error_msg("No super block found on %s (Expected magic %08x, got %08x)",
				devname, MD_SB_MAGIC, super->md_magic);
		free(super);
		return 2;
	}

	if (super->major_version != 0) {
		if (devname)
			bb_error_msg("Cannot interpret superblock on %s - version is %d",
				devname, super->major_version);
		free(super);
		return 2;
	}
	st->sb = super;

	if (st->ss == NULL) {
		st->ss = &super0;
		st->minor_version = super->minor_version;
		st->max_devs = MD_SB_DISKS;
		st->info = NULL;
	}

	/* Now check on the bitmap superblock */
	if ((super->state & (1<<MD_SB_BITMAP_PRESENT)) == 0)
		return 0;
	/* Read the bitmap superblock and make sure it looks
	 * valid.  If it doesn't clear the bit.  An --assemble --force
	 * should get that written out.
	 */
	if (read(fd, super+1, ROUND_UP(sizeof(struct bitmap_super_s),4096))
	    != ROUND_UP(sizeof(struct bitmap_super_s),4096))
		goto no_bitmap;

	uuid_from_super0(st, uuid);
	bsb = (struct bitmap_super_s *)(super+1);
	if (__le32_to_cpu(bsb->magic) != BITMAP_MAGIC ||
	    memcmp(bsb->uuid, uuid, 16) != 0)
		goto no_bitmap;
	return 0;

 no_bitmap:
	super->state &= ~(1<<MD_SB_BITMAP_PRESENT);

	return 0;
}

static struct supertype *match_metadata_desc0(char *arg)
{
	struct supertype *st = calloc(1, sizeof(*st));
	if (!st)
		return st;

	st->container_dev = NoMdDev;
	st->ss = &super0;
	st->info = NULL;
	st->minor_version = 90;
	st->max_devs = MD_SB_DISKS;
	st->sb = NULL;
	/* we sometimes get 00.90 */
	while (arg[0] == '0' && arg[1] == '0')
		arg++;
	if (strcmp(arg, "0") == 0 ||
#ifdef DEFAULT_OLD_METADATA /* ifndef in super1.c */
	    strcmp(arg, bb_default) == 0 ||
#endif /* DEFAULT_OLD_METADATA */
	    strcmp(arg, "0.90") == 0 ||
	    strcmp(arg, "") == 0 /* no metadata  - i.e. non_persistent */
		)
		return st;

	st->minor_version = 91; /* reshape in progress */
	if (strcmp(arg, "0.91") == 0) /* For dup_super support */
		return st;

	st->minor_version = 9; /* flag for 'byte-swapped' */
	if (strcmp(arg, "0.swap")==0 ||
	    strcmp(arg, "0.9") == 0) /* For dup_super support */
		return st;

	free(st);
	return NULL;
}

static void free_super0(struct supertype *st)
{
	if (st->sb)
		free(st->sb);
	while (st->info) {
		struct devinfo *di = st->info;
		st->info = di->next;
		if (di->fd >= 0)
			close(di->fd);
		free(di);
	}
	st->sb = NULL;
}

static const struct superswitch super0 = {
	.match_home = match_home0,
	.uuid_from_super = uuid_from_super0,
	.getinfo_super = getinfo_super0,
	.container_content = container_content0,
	.update_super = update_super0,
	.store_super = store_super0,
	.compare_super = compare_super0,
	.load_super = load_super0,
	.match_metadata_desc = match_metadata_desc0,
	.free_super = free_super0,
	.name = "0.90",
};

/*
 * The version-1 superblock :
 * All numeric fields are little-endian.
 *
 * total size: 256 bytes plus 2 per device.
 *  1K allows 384 devices.
 */
struct mdp_superblock_1 {
	/* constant array information - 128 bytes */
	__u32   magic;          /* MD_SB_MAGIC: 0xa92b4efc - little endian */
	__u32   major_version;  /* 1 */
	__u32   feature_map;    /* 0 for now */
	__u32   pad0;           /* always set to 0 when writing */

	__u8    set_uuid[16];   /* user-space generated. */
	char    set_name[32];   /* set and interpreted by user-space */

	__u64   ctime;          /* lo 40 bits are seconds, top 24 are microseconds or 0*/
	__u32   level;          /* -4 (multipath), -1 (linear), 0,1,4,5 */
	__u32   layout;         /* only for raid5 currently */
	__u64   size;           /* used size of component devices, in 512byte sectors */

	__u32   chunksize;      /* in 512byte sectors */
	__u32   raid_disks;
	__u32   bitmap_offset;  /* sectors after start of superblock that bitmap starts
				 * NOTE: signed, so bitmap can be before superblock
				 * only meaningful of feature_map[0] is set.
				 */

	/* These are only valid with feature bit '4' */
	__u32   new_level;      /* new level we are reshaping to                */
	__u64   reshape_position;       /* next address in array-space for reshape */
	__u32   delta_disks;    /* change in number of raid_disks               */
	__u32   new_layout;     /* new layout                                   */
	__u32   new_chunk;      /* new chunk size (bytes)                       */
	__u8    pad1[128-124];  /* set to 0 when written */

	/* constant this-device information - 64 bytes */
	__u64   data_offset;    /* sector start of data, often 0 */
	__u64   data_size;      /* sectors in this device that can be used for data */
	__u64   super_offset;   /* sector start of this superblock */
	__u64   recovery_offset;/* sectors before this offset (from data_offset) have been recovered */
	__u32   dev_number;     /* permanent identifier of this  device - not role in raid */
	__u32   cnt_corrected_read; /* number of read errors that were corrected by re-writing */
	__u8    device_uuid[16]; /* user-space setable, ignored by kernel */
	__u8    devflags;        /* per-device flags.  Only one defined...*/
#define WriteMostly1    1        /* mask for writemostly flag in above */
	__u8    pad2[64-57];    /* set to 0 when writing */

	/* array state information - 64 bytes */
	__u64   utime;          /* 40 bits second, 24 btes microseconds */
	__u64   events;         /* incremented when superblock updated */
	__u64   resync_offset;  /* data before this offset (from data_offset) known to be in sync */
	__u32   sb_csum;        /* checksum upto dev_roles[max_dev] */
	__u32   max_dev;        /* size of dev_roles[] array to consider */
	__u8    pad3[64-32];    /* set to 0 when writing */

	/* device state information. Indexed by dev_number.
	 * 2 bytes per device
	 * Note there are no per-device state flags. State information is rolled
	 * into the 'roles' value.  If a device is spare or faulty, then it doesn't
	 * have a meaningful role.
	 */
	__u16   dev_roles[0];   /* role in array, or 0xffff for a spare, or 0xfffe for faulty */
};

#define MAX_SB_SIZE 4096
/* bitmap super size is 256, but we round up to a sector for alignment */
#define BM_SUPER_SIZE 512
#define MAX_DEVS ((int)(MAX_SB_SIZE - sizeof(struct mdp_superblock_1)) / 2)
#define SUPER1_SIZE     (MAX_SB_SIZE + BM_SUPER_SIZE \
			 + sizeof(struct misc_dev_info))

struct misc_dev_info {
	__u64 device_size;
};

/* feature_map bits */
#define MD_FEATURE_BITMAP_OFFSET        1
#define MD_FEATURE_RECOVERY_OFFSET      2 /* recovery_offset is present and
					   * must be honoured
					   */
#define MD_FEATURE_RESHAPE_ACTIVE       4

#define MD_FEATURE_ALL                  (1|2|4)

#ifndef offsetof
#define offsetof(t,f) ((size_t)&(((t*)0)->f))
#endif
static unsigned int calc_sb_1_csum(struct mdp_superblock_1 * sb)
{
	unsigned int disk_csum, csum;
	unsigned long long newcsum;
	int size = sizeof(*sb) + __le32_to_cpu(sb->max_dev)*2;
	unsigned int *isuper = (unsigned int*)sb;

/* make sure I can count... */
	if (offsetof(struct mdp_superblock_1,data_offset) != 128 ||
	    offsetof(struct mdp_superblock_1, utime) != 192 ||
	    sizeof(struct mdp_superblock_1) != 256) {
		bb_error_msg("WARNING - superblock isn't sized correctly");
	}

	disk_csum = sb->sb_csum;
	sb->sb_csum = 0;
	newcsum = 0;
	for (; size>=4; size -= 4 ) {
		newcsum += __le32_to_cpu(*isuper);
		isuper++;
	}

	if (size == 2)
		newcsum += __le16_to_cpu(*(unsigned short*) isuper);

	csum = (newcsum & 0xffffffff) + (newcsum >> 32);
	sb->sb_csum = disk_csum;
	return __cpu_to_le32(csum);
}

/*
 * Information related to file descriptor used for aligned reads/writes.
 * Cache the block size.
 */
struct align_fd {
	int fd;
	int blk_sz;
};

static void init_afd(struct align_fd *afd, int fd)
{
	afd->fd = fd;

	if (ioctl(afd->fd, BLKSSZGET, &afd->blk_sz) != 0)
		afd->blk_sz = 512;
}

static int aread(struct align_fd *afd, void *buf, int len)
{
	/* aligned read.
	 * On devices with a 4K sector size, we need to read
	 * the full sector and copy relevant bits into
	 * the buffer
	 */
	int bsize, iosize;
	char *b;
	int n;

	bsize = afd->blk_sz;

	if (!bsize || bsize > 4096 || len > 4096) {
		if (!bsize)
			bb_error_msg("WARNING - aread() called with "
				"invalid block size");
		return -1;
	}
	b = ROUND_UP_PTR((char *)abuf, 4096);

	for (iosize = 0; iosize < len; iosize += bsize)
		;
	n = read(afd->fd, b, iosize);
	if (n <= 0)
		return n;
	lseek(afd->fd, len - n, 1);
	if (n > len)
		n = len;
	memcpy(buf, b, n);
	return n;
}

static int awrite(struct align_fd *afd, void *buf, int len)
{
	/* aligned write.
	 * On devices with a 4K sector size, we need to write
	 * the full sector.  We pre-read if the sector is larger
	 * than the write.
	 * The address must be sector-aligned.
	 */
	int bsize, iosize;
	char *b;
	int n;

	bsize = afd->blk_sz;
	if (!bsize || bsize > 4096 || len > 4096) {
		if (!bsize)
			bb_error_msg("WARNING - awrite() called with "
				"invalid block size");
		return -1;
	}
	b = ROUND_UP_PTR((char *)abuf, 4096);

	for (iosize = 0; iosize < len ; iosize += bsize)
		;

	if (len != iosize) {
		n = read(afd->fd, b, iosize);
		if (n <= 0)
			return n;
		lseek(afd->fd, -n, 1);
	}

	memcpy(b, buf, len);
	n = write(afd->fd, b, iosize);
	if (n <= 0)
		return n;
	lseek(afd->fd, len - n, 1);
	return len;
}

static int match_home1(struct supertype *st, char *homehost)
{
	struct mdp_superblock_1 *sb = st->sb;
	int l = homehost ? strlen(homehost) : 0;

	return (l > 0 && l < 32 &&
		sb->set_name[l] == ':' &&
		strncmp(sb->set_name, homehost, l) == 0);
}

static void uuid_from_super1(struct supertype *st, int uuid[4])
{
	struct mdp_superblock_1 *super = st->sb;
	char *cuuid = (char*)uuid;
	int i;
	for (i=0; i<16; i++)
		cuuid[i] = super->set_uuid[i];
}

static void getinfo_super1(struct supertype *st, struct mdinfo *info, char *map)
{
	struct mdp_superblock_1 *sb = st->sb;
	int working = 0;
	unsigned int i;
	unsigned int role;
	unsigned int map_disks = info->array.raid_disks;

	memset(info, 0, sizeof(*info));
	info->array.major_version = 1;
	info->array.minor_version = st->minor_version;
	info->array.patch_version = 0;
	info->array.raid_disks = __le32_to_cpu(sb->raid_disks);
	info->array.level = __le32_to_cpu(sb->level);
	info->array.layout = __le32_to_cpu(sb->layout);
	info->array.md_minor = -1;
	info->array.ctime = __le64_to_cpu(sb->ctime);
	info->array.utime = __le64_to_cpu(sb->utime);
	info->array.chunk_size = __le32_to_cpu(sb->chunksize)*512;
	info->array.state =
		(__le64_to_cpu(sb->resync_offset) == MaxSector)
		? 1 : 0;

	info->data_offset = __le64_to_cpu(sb->data_offset);
	info->component_size = __le64_to_cpu(sb->size);
	if (sb->feature_map & __le32_to_cpu(MD_FEATURE_BITMAP_OFFSET))
		info->bitmap_offset = (int32_t)__le32_to_cpu(sb->bitmap_offset);

	info->disk.major = 0;
	info->disk.minor = 0;
	info->disk.number = __le32_to_cpu(sb->dev_number);
	if (__le32_to_cpu(sb->dev_number) >= __le32_to_cpu(sb->max_dev) ||
	    __le32_to_cpu(sb->dev_number) >= MAX_DEVS)
		role = 0xfffe;
	else
		role = __le16_to_cpu(sb->dev_roles[__le32_to_cpu(sb->dev_number)]);

	info->disk.raid_disk = -1;
	switch(role) {
	case 0xFFFF:
		info->disk.state = 0; /* spare: not active, not sync, not faulty */
		break;
	case 0xFFFE:
		info->disk.state = 1; /* faulty */
		break;
	default:
		info->disk.state = 6; /* active and in sync */
		info->disk.raid_disk = role;
	}
	if (sb->devflags & WriteMostly1)
		info->disk.state |= (1 << MD_DISK_WRITEMOSTLY);
	info->events = __le64_to_cpu(sb->events);
	sprintf(info->text_version, "1.%d", st->minor_version);
	info->safe_mode_delay = 200;

	memcpy(info->uuid, sb->set_uuid, 16);

	strncpy(info->name, sb->set_name, 32);
	info->name[32] = 0;

	if (sb->feature_map & __le32_to_cpu(MD_FEATURE_RECOVERY_OFFSET))
		info->recovery_start = __le32_to_cpu(sb->recovery_offset);
	else
		info->recovery_start = MaxSector;

	if (sb->feature_map & __le32_to_cpu(MD_FEATURE_RESHAPE_ACTIVE)) {
		info->reshape_active = 1;
		info->reshape_progress = __le64_to_cpu(sb->reshape_position);
		info->new_level = __le32_to_cpu(sb->new_level);
		info->delta_disks = __le32_to_cpu(sb->delta_disks);
		info->new_layout = __le32_to_cpu(sb->new_layout);
		info->new_chunk = __le32_to_cpu(sb->new_chunk)<<9;
		if (info->delta_disks < 0)
			info->array.raid_disks -= info->delta_disks;
	} else
		info->reshape_active = 0;

	info->recovery_blocked = info->reshape_active;

	if (map)
		for (i=0; i<map_disks; i++)
			map[i] = 0;
	for (i = 0; i < __le32_to_cpu(sb->max_dev); i++) {
		role = __le16_to_cpu(sb->dev_roles[i]);
		if (/*role == 0xFFFF || */role < (unsigned) info->array.raid_disks) {
			working++;
			if (map && role < map_disks)
				map[role] = 1;
		}
	}

	info->array.working_disks = working;
}

static struct mdinfo *container_content1(struct supertype *st, char *subarray)
{
	struct mdinfo *info;

	if (subarray)
		return NULL;

	info = xmalloc(sizeof(*info));
	getinfo_super1(st, info, NULL);
	return info;
}

static int update_super1(struct supertype *st, struct mdinfo *info,
			 char *update,
			 char *devname, int verbose,
			 int uuid_set, char *homehost)
{
	/* NOTE: for 'assemble' and 'force' we need to return non-zero
	 * if any change was made.  For others, the return value is
	 * ignored.
	 */
	int rv = 0;
	struct mdp_superblock_1 *sb = st->sb;

	(void)&verbose;
	(void)&uuid_set;

	if (strcmp(update, "force-one")==0) {
		/* Not enough devices for a working array,
		 * so bring this one up-to-date
		 */
		if (sb->events != __cpu_to_le64(info->events))
			rv = 1;
		sb->events = __cpu_to_le64(info->events);
	} else if (strcmp(update, "force-array")==0) {
		/* Degraded array and 'force' requests to
		 * maybe need to mark it 'clean'.
		 */
		switch(__le32_to_cpu(sb->level)) {
		case 5: case 4: case 6:
			/* need to force clean */
			if (sb->resync_offset != MaxSector)
				rv = 1;
			sb->resync_offset = MaxSector;
		}
	} else if (strcmp(update, "assemble")==0) {
		int d = info->disk.number;
		int want;
		if (info->disk.state == 6)
			want = info->disk.raid_disk;
		else
			want = 0xFFFF;
		if (sb->dev_roles[d] != __cpu_to_le16(want)) {
			sb->dev_roles[d] = __cpu_to_le16(want);
			rv = 1;
		}
		if (info->reshape_active &&
		    sb->feature_map & __le32_to_cpu(MD_FEATURE_RESHAPE_ACTIVE) &&
		    info->delta_disks >= 0 &&
		    info->reshape_progress < __le64_to_cpu(sb->reshape_position)) {
			sb->reshape_position = __cpu_to_le64(info->reshape_progress);
			rv = 1;
		}
		if (info->reshape_active &&
		    sb->feature_map & __le32_to_cpu(MD_FEATURE_RESHAPE_ACTIVE) &&
		    info->delta_disks < 0 &&
		    info->reshape_progress > __le64_to_cpu(sb->reshape_position)) {
			sb->reshape_position = __cpu_to_le64(info->reshape_progress);
			rv = 1;
		}
	} else if (strcmp(update, "linear-grow-new") == 0) {
		unsigned int i;
		int fd;
		unsigned int max = __le32_to_cpu(sb->max_dev);

		for (i=0 ; i < max ; i++)
			if (__le16_to_cpu(sb->dev_roles[i]) >= 0xfffe)
				break;
		sb->dev_number = __cpu_to_le32(i);
		info->disk.number = i;
		if (max >= __le32_to_cpu(sb->max_dev))
			sb->max_dev = __cpu_to_le32(max+1);
		random_get_bytes(sb->device_uuid, sizeof(sb->device_uuid));

		sb->dev_roles[i] =
			__cpu_to_le16(info->disk.raid_disk);

		fd = open(devname, O_RDONLY);
		if (fd >= 0) {
			unsigned long long ds;
			get_dev_size(fd, devname, &ds);
			close(fd);
			ds >>= 9;
			if (__le64_to_cpu(sb->super_offset) <
			    __le64_to_cpu(sb->data_offset)) {
				sb->data_size = __cpu_to_le64(
					ds - __le64_to_cpu(sb->data_offset));
			} else {
				ds -= 8*2;
				ds &= ~(unsigned long long)(4*2-1);
				sb->super_offset = __cpu_to_le64(ds);
				sb->data_size = __cpu_to_le64(
					ds - __le64_to_cpu(sb->data_offset));
			}
		}
	} else if (strcmp(update, "linear-grow-update") == 0) {
		sb->raid_disks = __cpu_to_le32(info->array.raid_disks);
		sb->dev_roles[info->disk.number] =
			__cpu_to_le16(info->disk.raid_disk);
	} else if (strcmp(update, "resync") == 0) {
		/* make sure resync happens */
		sb->resync_offset = 0ULL;
	} else if (strcmp(update, "uuid") == 0) {
		copy_uuid(sb->set_uuid, info->uuid, super1.swapuuid);

		if (__le32_to_cpu(sb->feature_map)&MD_FEATURE_BITMAP_OFFSET) {
			struct bitmap_super_s *bm;
			bm = (struct bitmap_super_s*)(st->sb+MAX_SB_SIZE);
			memcpy(bm->uuid, sb->set_uuid, 16);
		}
	} else if (strcmp(update, "no-bitmap") == 0) {
		sb->feature_map &= ~__cpu_to_le32(MD_FEATURE_BITMAP_OFFSET);
	} else if (strcmp(update, "homehost") == 0 &&
		   homehost) {
		char *c;
		update = "name";
		c = strchr(sb->set_name, ':');
		if (c)
			strncpy(info->name, c+1, 31 - (c-sb->set_name));
		else
			strncpy(info->name, sb->set_name, 32);
		info->name[32] = 0;
	} else if (strcmp(update, "name") == 0) {
		if (info->name[0] == 0)
			sprintf(info->name, "%d", info->array.md_minor);
		memset(sb->set_name, 0, sizeof(sb->set_name));
		if (homehost &&
		    strchr(info->name, ':') == NULL &&
		    strlen(homehost)+1+strlen(info->name) < 32) {
			strcpy(sb->set_name, homehost);
			strcat(sb->set_name, ":");
			strcat(sb->set_name, info->name);
		} else
			strcpy(sb->set_name, info->name);
	} else if (strcmp(update, "devicesize") == 0 &&
	    __le64_to_cpu(sb->super_offset) <
	    __le64_to_cpu(sb->data_offset)) {
		/* set data_size to device size less data_offset */
		struct misc_dev_info *misc = (struct misc_dev_info*)
			(st->sb + MAX_SB_SIZE + BM_SUPER_SIZE);
		printf("Size was %llu\n", (unsigned long long)
		       __le64_to_cpu(sb->data_size));
		sb->data_size = __cpu_to_le64(
			misc->device_size - __le64_to_cpu(sb->data_offset));
		printf("Size is %llu\n", (unsigned long long)
		       __le64_to_cpu(sb->data_size));
	} else if (strcmp(update, "_reshape_progress")==0)
		sb->reshape_position = __cpu_to_le64(info->reshape_progress);
	else if (strcmp(update, "writemostly")==0)
		sb->devflags |= WriteMostly1;
	else if (strcmp(update, "readwrite")==0)
		sb->devflags &= ~WriteMostly1;
	else
		rv = -1;

	sb->sb_csum = calc_sb_1_csum(sb);
	return rv;
}

static void locate_bitmap1(struct supertype *st, int fd);

static int store_super1(struct supertype *st, int fd)
{
	struct mdp_superblock_1 *sb = st->sb;
	unsigned long long sb_offset;
	struct align_fd afd;
	int sbsize;
	unsigned long long dsize;

	if (!get_dev_size(fd, NULL, &dsize))
		return 1;

	dsize >>= 9;

	if (dsize < 24)
		return 2;

	init_afd(&afd, fd);

	/*
	 * Calculate the position of the superblock.
	 * It is always aligned to a 4K boundary and
	 * depending on minor_version, it can be:
	 * 0: At least 8K, but less than 12K, from end of device
	 * 1: At start of device
	 * 2: 4K from start of device.
	 */
	switch(st->minor_version) {
	case 0:
		sb_offset = dsize;
		sb_offset -= 8*2;
		sb_offset &= ~(4*2-1);
		break;
	case 1:
		sb_offset = 0;
		break;
	case 2:
		sb_offset = 4*2;
		break;
	default:
		return -EINVAL;
	}



	if (sb_offset != __le64_to_cpu(sb->super_offset) &&
	    0 != __le64_to_cpu(sb->super_offset)
		) {
		bb_error_msg("internal error - sb_offset is wrong");
		abort();
	}

	if (lseek(fd, sb_offset << 9, 0)< 0LL)
		return 3;

	sbsize = ROUND_UP(sizeof(*sb) + 2 * __le32_to_cpu(sb->max_dev), 512);

	if (awrite(&afd, sb, sbsize) != sbsize)
		return 4;

	if (sb->feature_map & __cpu_to_le32(MD_FEATURE_BITMAP_OFFSET)) {
		struct bitmap_super_s *bm = (struct bitmap_super_s*)
			(((char*)sb)+MAX_SB_SIZE);
		if (__le32_to_cpu(bm->magic) == BITMAP_MAGIC) {
			locate_bitmap1(st, fd);
			if (awrite(&afd, bm, sizeof(*bm)) != sizeof(*bm))
				return 5;
		}
	}
	fsync(fd);
	return 0;
}

static int compare_super1(struct supertype *st, struct supertype *tst)
{
	/*
	 * return:
	 *  0 same, or first was empty, and second was copied
	 *  1 second had wrong number
	 *  2 wrong uuid
	 *  3 wrong other info
	 */
	struct mdp_superblock_1 *first = st->sb;
	struct mdp_superblock_1 *second = tst->sb;

	if (second->magic != __cpu_to_le32(MD_SB_MAGIC))
		return 1;
	if (second->major_version != __cpu_to_le32(1))
		return 1;

	if (!first) {
		if (posix_memalign((void**)&first, 4096, SUPER1_SIZE) != 0) {
			bb_error_msg("%s could not allocate superblock", __func__);
			return 1;
		}
		memcpy(first, second, SUPER1_SIZE);
		st->sb = first;
		return 0;
	}
	if (memcmp(first->set_uuid, second->set_uuid, 16)!= 0)
		return 2;

	if (first->ctime      != second->ctime     ||
	    first->level      != second->level     ||
	    first->layout     != second->layout    ||
	    first->size       != second->size      ||
	    first->chunksize  != second->chunksize ||
	    first->raid_disks != second->raid_disks)
		return 3;
	return 0;
}

static void free_super1(struct supertype *st)
{
	if (st->sb)
		free(st->sb);
	while (st->info) {
		struct devinfo *di = st->info;
		st->info = di->next;
		if (di->fd >= 0)
			close(di->fd);
		free(di);
	}
	st->sb = NULL;
}

static int load_super1(struct supertype *st, int fd, char *devname)
{
	unsigned long long dsize;
	unsigned long long sb_offset;
	struct mdp_superblock_1 *super;
	int uuid[4];
	struct bitmap_super_s *bsb;
	struct misc_dev_info *misc;
	struct align_fd afd;

	free_super1(st);

	init_afd(&afd, fd);

	if (st->ss == NULL || st->minor_version == -1) {
		int bestvers = -1;
		struct supertype tst;
		__u64 bestctime = 0;
		/* guess... choose latest ctime */
		memset(&tst, 0, sizeof(tst));
		tst.ss = &super1;
		for (tst.minor_version = 0; tst.minor_version <= 2 ; tst.minor_version++) {
			switch(load_super1(&tst, fd, devname)) {
			case 0: super = tst.sb;
				if (bestvers == -1 ||
				    bestctime < __le64_to_cpu(super->ctime)) {
					bestvers = tst.minor_version;
					bestctime = __le64_to_cpu(super->ctime);
				}
				free(super);
				tst.sb = NULL;
				break;
			case 1: return 1; /*bad device */
			case 2: break; /* bad, try next */
			}
		}
		if (bestvers != -1) {
			int rv;
			tst.minor_version = bestvers;
			tst.ss = &super1;
			tst.max_devs = MAX_DEVS;
			rv = load_super1(&tst, fd, devname);
			if (rv == 0)
				*st = tst;
			return rv;
		}
		return 2;
	}
	if (!get_dev_size(fd, devname, &dsize))
		return 1;
	dsize >>= 9;

	if (dsize < 24) {
		if (devname)
			bb_error_msg("%s is too small for md: size is %llu sectors.",
				devname, dsize);
		return 1;
	}

	/*
	 * Calculate the position of the superblock.
	 * It is always aligned to a 4K boundary and
	 * depending on minor_version, it can be:
	 * 0: At least 8K, but less than 12K, from end of device
	 * 1: At start of device
	 * 2: 4K from start of device.
	 */
	switch(st->minor_version) {
	case 0:
		sb_offset = dsize;
		sb_offset -= 8*2;
		sb_offset &= ~(4*2-1);
		break;
	case 1:
		sb_offset = 0;
		break;
	case 2:
		sb_offset = 4*2;
		break;
	default:
		return -EINVAL;
	}

	ioctl(fd, BLKFLSBUF, 0); /* make sure we read current data */


	if (lseek(fd, sb_offset << 9, 0)< 0LL) {
		if (devname)
			bb_perror_msg("Cannot seek to superblock on %s", devname);
		return 1;
	}

	if (posix_memalign((void**)&super, 4096, SUPER1_SIZE) != 0) {
		bb_error_msg("%s could not allocate superblock", __func__);
		return 1;
	}

	if (aread(&afd, super, MAX_SB_SIZE) != MAX_SB_SIZE) {
		if (devname)
			bb_error_msg("Cannot read superblock on %s", devname);
		free(super);
		return 1;
	}

	if (__le32_to_cpu(super->magic) != MD_SB_MAGIC) {
		if (devname)
			bb_error_msg("No super block found on %s (Expected magic %08x, got %08x)",
				devname, MD_SB_MAGIC, __le32_to_cpu(super->magic));
		free(super);
		return 2;
	}

	if (__le32_to_cpu(super->major_version) != 1) {
		if (devname)
			bb_error_msg("Cannot interpret superblock on %s - version is %d",
				devname, __le32_to_cpu(super->major_version));
		free(super);
		return 2;
	}
	if (__le64_to_cpu(super->super_offset) != sb_offset) {
		if (devname)
			bb_error_msg("No superblock found on %s (super_offset is wrong)",
				devname);
		free(super);
		return 2;
	}
	st->sb = super;

	bsb = (struct bitmap_super_s *)(((char*)super)+MAX_SB_SIZE);

	misc = (struct misc_dev_info*) (((char*)super)+MAX_SB_SIZE+BM_SUPER_SIZE);
	misc->device_size = dsize;

	/* Now check on the bitmap superblock */
	if ((__le32_to_cpu(super->feature_map)&MD_FEATURE_BITMAP_OFFSET) == 0)
		return 0;
	/* Read the bitmap superblock and make sure it looks
	 * valid.  If it doesn't clear the bit.  An --assemble --force
	 * should get that written out.
	 */
	locate_bitmap1(st, fd);
	if (aread(&afd, bsb, 512) != 512)
		goto no_bitmap;

	uuid_from_super1(st, uuid);
	if (__le32_to_cpu(bsb->magic) != BITMAP_MAGIC ||
	    memcmp(bsb->uuid, uuid, 16) != 0)
		goto no_bitmap;
	return 0;

 no_bitmap:
	super->feature_map = __cpu_to_le32(__le32_to_cpu(super->feature_map)
					   & ~MD_FEATURE_BITMAP_OFFSET);
	return 0;
}


static struct supertype *match_metadata_desc1(char *arg)
{
	struct supertype *st = calloc(1, sizeof(*st));
	if (!st)
		return st;

	st->container_dev = NoMdDev;
	st->ss = &super1;
	st->max_devs = MAX_DEVS;
	st->sb = NULL;
	/* leading zeros can be safely ignored.  --detail generates them. */
	while (*arg == '0')
		arg++;
	if (strcmp(arg, "1.0") == 0 ||
	    strcmp(arg, "1.00") == 0) {
		st->minor_version = 0;
		return st;
	}
	if (strcmp(arg, "1.1") == 0 ||
	    strcmp(arg, "1.01") == 0
		) {
		st->minor_version = 1;
		return st;
	}
	if (strcmp(arg, "1.2") == 0 ||
#ifndef DEFAULT_OLD_METADATA /* ifdef in super0.c */
	    strcmp(arg, bb_default) == 0 ||
#endif /* DEFAULT_OLD_METADATA */
	    strcmp(arg, "1.02") == 0) {
		st->minor_version = 2;
		return st;
	}
	if (strcmp(arg, "1") == 0 ||
	    strcmp(arg, bb_default) == 0) {
		st->minor_version = -1;
		return st;
	}

	free(st);
	return NULL;
}

static void locate_bitmap1(struct supertype *st, int fd)
{
	unsigned long long offset;
	struct mdp_superblock_1 *sb;
	int mustfree = 0;

	if (!st->sb) {
		if (st->ss->load_super(st, fd, NULL))
			return; /* no error I hope... */
		mustfree = 1;
	}
	sb = st->sb;

	offset = __le64_to_cpu(sb->super_offset);
	offset += (int32_t) __le32_to_cpu(sb->bitmap_offset);
	if (mustfree)
		free(sb);
	lseek(fd, offset<<9, 0);
}

static const struct superswitch super1 = {
	.match_home = match_home1,
	.uuid_from_super = uuid_from_super1,
	.getinfo_super = getinfo_super1,
	.container_content = container_content1,
	.update_super = update_super1,
	.store_super = store_super1,
	.compare_super = compare_super1,
	.load_super = load_super1,
	.match_metadata_desc = match_metadata_desc1,
	.free_super = free_super1,
#if __BYTE_ORDER == BIG_ENDIAN
	.swapuuid = 0,
#else
	.swapuuid = 1,
#endif
	.name = "1.x",
};


/* a non-official T10 name for creation GUIDs */
static const char T10[] = "Linux-MD";

/* DDF timestamps are 1980 based, so we need to add
 * second-in-decade-of-seventies to convert to linux timestamps.
 * 10 years with 2 leap years.
 */
#define DECADE (3600*24*(365*10+2))
static unsigned long crc32(unsigned long crc, const unsigned char *buf,
				unsigned len);

/* The DDF metadata handling.
 * DDF metadata lives at the end of the device.
 * The last 512 byte block provides an 'anchor' which is used to locate
 * the rest of the metadata which usually lives immediately behind the anchor.
 *
 * Note:
 *  - all multibyte numeric fields are bigendian.
 *  - all strings are space padded.
 *
 */

/* Primary Raid Level (PRL) */
#define DDF_RAID0       0x00
#define DDF_RAID1       0x01
#define DDF_RAID3       0x03
#define DDF_RAID4       0x04
#define DDF_RAID5       0x05
#define DDF_RAID1E      0x11
#define DDF_JBOD        0x0f
#define DDF_CONCAT      0x1f
#define DDF_RAID5E      0x15
#define DDF_RAID5EE     0x25
#define DDF_RAID6       0x06

/* Raid Level Qualifier (RLQ) */
#define DDF_RAID0_SIMPLE        0x00
#define DDF_RAID1_SIMPLE        0x00 /* just 2 devices in this plex */
#define DDF_RAID1_MULTI         0x01 /* exactly 3 devices in this plex */
#define DDF_RAID3_0             0x00 /* parity in first extent */
#define DDF_RAID3_N             0x01 /* parity in last extent */
#define DDF_RAID4_0             0x00 /* parity in first extent */
#define DDF_RAID4_N             0x01 /* parity in last extent */
/* these apply to raid5e and raid5ee as well */
#define DDF_RAID5_0_RESTART     0x00 /* same as 'right asymmetric' - layout 1 */
#define DDF_RAID6_0_RESTART     0x01 /* raid6 different from raid5 here!!! */
#define DDF_RAID5_N_RESTART     0x02 /* same as 'left asymmetric' - layout 0 */
#define DDF_RAID5_N_CONTINUE    0x03 /* same as 'left symmetric' - layout 2 */

#define DDF_RAID1E_ADJACENT     0x00 /* raid10 nearcopies==2 */
#define DDF_RAID1E_OFFSET       0x01 /* raid10 offsetcopies==2 */

/* Secondary RAID Level (SRL) */
#define DDF_2STRIPED    0x00    /* This is weirder than RAID0 !! */
#define DDF_2MIRRORED   0x01
#define DDF_2CONCAT     0x02
#define DDF_2SPANNED    0x03    /* This is also weird - be careful */

/* Magic numbers */
#define DDF_HEADER_MAGIC        __cpu_to_be32(0xDE11DE11)
#define DDF_CONTROLLER_MAGIC    __cpu_to_be32(0xAD111111)
#define DDF_PHYS_RECORDS_MAGIC  __cpu_to_be32(0x22222222)
#define DDF_PHYS_DATA_MAGIC     __cpu_to_be32(0x33333333)
#define DDF_VIRT_RECORDS_MAGIC  __cpu_to_be32(0xDDDDDDDD)
#define DDF_VD_CONF_MAGIC       __cpu_to_be32(0xEEEEEEEE)
#define DDF_SPARE_ASSIGN_MAGIC  __cpu_to_be32(0x55555555)
#define DDF_VU_CONF_MAGIC       __cpu_to_be32(0x88888888)
#define DDF_VENDOR_LOG_MAGIC    __cpu_to_be32(0x01dBEEF0)
#define DDF_BBM_LOG_MAGIC       __cpu_to_be32(0xABADB10C)

#define DDF_GUID_LEN    24
#define DDF_REVISION_0  "01.00.00"
#define DDF_REVISION_2  "01.02.00"

struct ddf_header {
	__u32   magic;          /* DDF_HEADER_MAGIC */
	__u32   crc;
	char    guid[DDF_GUID_LEN];
	char    revision[8];    /* 01.02.00 */
	__u32   seq;            /* starts at '1' */
	__u32   timestamp;
	__u8    openflag;
	__u8    foreignflag;
	__u8    enforcegroups;
	__u8    pad0;           /* 0xff */
	__u8    pad1[12];       /* 12 * 0xff */
	/* 64 bytes so far */
	__u8    header_ext[32]; /* reserved: fill with 0xff */
	__u64   primary_lba;
	__u64   secondary_lba;
	__u8    type;
	__u8    pad2[3];        /* 0xff */
	__u32   workspace_len;  /* sectors for vendor space -
				 * at least 32768(sectors) */
	__u64   workspace_lba;
	__u16   max_pd_entries; /* one of 15, 63, 255, 1023, 4095 */
	__u16   max_vd_entries; /* 2^(4,6,8,10,12)-1 : i.e. as above */
	__u16   max_partitions; /* i.e. max num of configuration
				   record entries per disk */
	__u16   config_record_len; /* 1 +ROUNDUP(max_primary_element_entries
						 *12/512) */
	__u16   max_primary_element_entries; /* 16, 64, 256, 1024, or 4096 */
	__u8    pad3[54];       /* 0xff */
	/* 192 bytes so far */
	__u32   controller_section_offset;
	__u32   controller_section_length;
	__u32   phys_section_offset;
	__u32   phys_section_length;
	__u32   virt_section_offset;
	__u32   virt_section_length;
	__u32   config_section_offset;
	__u32   config_section_length;
	__u32   data_section_offset;
	__u32   data_section_length;
	__u32   bbm_section_offset;
	__u32   bbm_section_length;
	__u32   diag_space_offset;
	__u32   diag_space_length;
	__u32   vendor_offset;
	__u32   vendor_length;
	/* 256 bytes so far */
	__u8    pad4[256];      /* 0xff */
};

/* type field */
#define DDF_HEADER_ANCHOR       0x00
#define DDF_HEADER_PRIMARY      0x01
#define DDF_HEADER_SECONDARY    0x02

/* The content of the 'controller section' - global scope */
struct ddf_controller_data {
	__u32   magic;                  /* DDF_CONTROLLER_MAGIC */
	__u32   crc;
	char    guid[DDF_GUID_LEN];
	struct controller_type {
		__u16 vendor_id;
		__u16 device_id;
		__u16 sub_vendor_id;
		__u16 sub_device_id;
	} type;
	char    product_id[16];
	__u8    pad[8]; /* 0xff */
	__u8    vendor_data[448];
};

/* The content of phys_section - global scope */
struct phys_disk {
	__u32   magic;          /* DDF_PHYS_RECORDS_MAGIC */
	__u32   crc;
	__u16   used_pdes;
	__u16   max_pdes;
	__u8    pad[52];
	struct phys_disk_entry {
		char    guid[DDF_GUID_LEN];
		__u32   refnum;
		__u16   type;
		__u16   state;
		__u64   config_size; /* DDF structures must be after here */
		char    path[18];       /* another horrible structure really */
		__u8    pad[6];
	} entries[0];
};

/* phys_disk_entry.type is a bitmap - bigendian remember */
#define DDF_Forced_PD_GUID              1
#define DDF_Active_in_VD                2
#define DDF_Global_Spare                4 /* VD_CONF records are ignored */
#define DDF_Spare                       8 /* overrides Global_spare */
#define DDF_Foreign                     16
#define DDF_Legacy                      32 /* no DDF on this device */

#define DDF_Interface_mask              0xf00
#define DDF_Interface_SCSI              0x100
#define DDF_Interface_SAS               0x200
#define DDF_Interface_SATA              0x300
#define DDF_Interface_FC                0x400

/* phys_disk_entry.state is a bigendian bitmap */
#define DDF_Online                      1
#define DDF_Failed                      2 /* overrides  1,4,8 */
#define DDF_Rebuilding                  4
#define DDF_Transition                  8
#define DDF_SMART                       16
#define DDF_ReadErrors                  32
#define DDF_Missing                     64

/* The content of the virt_section global scope */
struct virtual_disk {
	__u32   magic;          /* DDF_VIRT_RECORDS_MAGIC */
	__u32   crc;
	__u16   populated_vdes;
	__u16   max_vdes;
	__u8    pad[52];
	struct virtual_entry {
		char    guid[DDF_GUID_LEN];
		__u16   unit;
		__u16   pad0;   /* 0xffff */
		__u16   guid_crc;
		__u16   type;
		__u8    state;
		__u8    init_state;
		__u8    pad1[14];
		char    name[16];
	} entries[0];
};

/* virtual_entry.type is a bitmap - bigendian */
#define DDF_Shared              1
#define DDF_Enforce_Groups      2
#define DDF_Unicode             4
#define DDF_Owner_Valid         8

/* virtual_entry.state is a bigendian bitmap */
#define DDF_state_mask          0x7
#define DDF_state_optimal       0x0
#define DDF_state_degraded      0x1
#define DDF_state_deleted       0x2
#define DDF_state_missing       0x3
#define DDF_state_failed        0x4
#define DDF_state_part_optimal  0x5

#define DDF_state_morphing      0x8
#define DDF_state_inconsistent  0x10

/* virtual_entry.init_state is a bigendian bitmap */
#define DDF_initstate_mask      0x03
#define DDF_init_not            0x00
#define DDF_init_quick          0x01 /* initialisation is progress.
				      * i.e. 'state_inconsistent' */
#define DDF_init_full           0x02

#define DDF_access_mask         0xc0
#define DDF_access_rw           0x00
#define DDF_access_ro           0x80
#define DDF_access_blocked      0xc0

/* The content of the config_section - local scope
 * It has multiple records each config_record_len sectors
 * They can be vd_config or spare_assign
 */

struct vd_config {
	__u32   magic;          /* DDF_VD_CONF_MAGIC */
	__u32   crc;
	char    guid[DDF_GUID_LEN];
	__u32   timestamp;
	__u32   seqnum;
	__u8    pad0[24];
	__u16   prim_elmnt_count;
	__u8    chunk_shift;    /* 0 == 512, 1==1024 etc */
	__u8    prl;
	__u8    rlq;
	__u8    sec_elmnt_count;
	__u8    sec_elmnt_seq;
	__u8    srl;
	__u64   blocks;         /* blocks per component could be different
				 * on different component devices...(only
				 * for concat I hope) */
	__u64   array_blocks;   /* blocks in array */
	__u8    pad1[8];
	__u32   spare_refs[8];
	__u8    cache_pol[8];
	__u8    bg_rate;
	__u8    pad2[3];
	__u8    pad3[52];
	__u8    pad4[192];
	__u8    v0[32]; /* reserved- 0xff */
	__u8    v1[32]; /* reserved- 0xff */
	__u8    v2[16]; /* reserved- 0xff */
	__u8    v3[16]; /* reserved- 0xff */
	__u8    vendor[32];
	__u32   phys_refnum[0]; /* refnum of each disk in sequence */
      /*__u64   lba_offset[0];  LBA offset in each phys.  Note extents in a
				bvd are always the same size */
};

/* vd_config.cache_pol[7] is a bitmap */
#define DDF_cache_writeback     1       /* else writethrough */
#define DDF_cache_wadaptive     2       /* only applies if writeback */
#define DDF_cache_readahead     4
#define DDF_cache_radaptive     8       /* only if doing read-ahead */
#define DDF_cache_ifnobatt      16      /* even to write cache if battery is poor */
#define DDF_cache_wallowed      32      /* enable write caching */
#define DDF_cache_rallowed      64      /* enable read caching */

struct spare_assign {
	__u32   magic;          /* DDF_SPARE_ASSIGN_MAGIC */
	__u32   crc;
	__u32   timestamp;
	__u8    reserved[7];
	__u8    type;
	__u16   populated;      /* SAEs used */
	__u16   max;            /* max SAEs */
	__u8    pad[8];
	struct spare_assign_entry {
		char    guid[DDF_GUID_LEN];
		__u16   secondary_element;
		__u8    pad[6];
	} spare_ents[0];
};
/* spare_assign.type is a bitmap */
#define DDF_spare_dedicated     0x1     /* else global */
#define DDF_spare_revertible    0x2     /* else committable */
#define DDF_spare_active        0x4     /* else not active */
#define DDF_spare_affinity      0x8     /* enclosure affinity */

/* The data_section contents - local scope */
struct disk_data {
	__u32   magic;          /* DDF_PHYS_DATA_MAGIC */
	__u32   crc;
	char    guid[DDF_GUID_LEN];
	__u32   refnum;         /* crc of some magic drive data ... */
	__u8    forced_ref;     /* set when above was not result of magic */
	__u8    forced_guid;    /* set if guid was forced rather than magic */
	__u8    vendor[32];
	__u8    pad[442];
};

/* bbm_section content */
struct bad_block_log {
	__u32   magic;
	__u32   crc;
	__u16   entry_count;
	__u32   spare_count;
	__u8    pad[10];
	__u64   first_spare;
	struct mapped_block {
		__u64   defective_start;
		__u32   replacement_start;
		__u16   remap_count;
		__u8    pad[2];
	} entries[0];
};

/* Struct for internally holding ddf structures */
/* The DDF structure stored on each device is potentially
 * quite different, as some data is global and some is local.
 * The global data is:
 *   - ddf header
 *   - controller_data
 *   - Physical disk records
 *   - Virtual disk records
 * The local data is:
 *   - Configuration records
 *   - Physical Disk data section
 *  (  and Bad block and vendor which I don't care about yet).
 *
 * The local data is parsed into separate lists as it is read
 * and reconstructed for writing.  This means that we only need
 * to make config changes once and they are automatically
 * propagated to all devices.
 * Note that the ddf_super has space of the conf and disk data
 * for this disk and also for a list of all such data.
 * The list is only used for the superblock that is being
 * built in Create or Assemble to describe the whole array.
 */
struct ddf_super {
	struct ddf_header anchor, primary, secondary;
	struct ddf_controller_data controller;
	struct ddf_header *active;
	struct phys_disk        *phys;
	struct virtual_disk     *virt;
	int pdsize, vdsize;
	unsigned int max_part, mppe, conf_rec_len;
	int currentdev;
	int updates_pending;
	struct vcl {
		union {
			char space[512];
			struct {
				struct vcl      *next;
				__u64           *lba_offset; /* location in 'conf' of
							      * the lba table */
				unsigned int    vcnum; /* index into ->virt */
				__u64           *block_sizes; /* NULL if all the same */
			};
		};
		struct vd_config conf;
	} *conflist, *currentconf;
	struct ddf_dl {
		union {
			char space[512];
			struct {
				struct ddf_dl       *next;
				int major, minor;
				char *devname;
				int fd;
				unsigned long long size; /* sectors */
				int pdnum;      /* index in ->phys */
				struct spare_assign *spare;
				void *mdupdate; /* hold metadata update */

				/* These fields used by auto-layout */
				int raiddisk; /* slot to fill in autolayout */
				__u64 esize;
			};
		};
		struct disk_data disk;
		struct vcl *vlist[0]; /* max_part in size */
	} *dlist, *add_list;
};

#ifndef offsetof
#define offsetof(t,f) ((size_t)&(((t*)0)->f))
#endif


static unsigned int calc_crc(void *buf, int len)
{
	/* crcs are always at the same place as in the ddf_header */
	struct ddf_header *ddf = buf;
	__u32 oldcrc = ddf->crc;
	__u32 newcrc;
	ddf->crc = 0xffffffff;

	newcrc = crc32(0, buf, len);
	ddf->crc = oldcrc;
	/* The crc is store (like everything) bigendian, so convert
	 * here for simplicity
	 */
	return __cpu_to_be32(newcrc);
}

static int load_ddf_header(int fd, unsigned long long lba,
			   unsigned long long size,
			   int type,
			   struct ddf_header *hdr, struct ddf_header *anchor)
{
	/* read a ddf header (primary or secondary) from fd/lba
	 * and check that it is consistent with anchor
	 * Need to check:
	 *   magic, crc, guid, rev, and LBA's header_type, and
	 *  everything after header_type must be the same
	 */
	if (lba >= size-1)
		return 0;

	if (lseek(fd, lba<<9, 0) < 0)
		return 0;

	if (read(fd, hdr, 512) != 512)
		return 0;

	if (hdr->magic != DDF_HEADER_MAGIC)
		return 0;
	if (calc_crc(hdr, 512) != hdr->crc)
		return 0;
	if (memcmp(anchor->guid, hdr->guid, DDF_GUID_LEN) != 0 ||
	    memcmp(anchor->revision, hdr->revision, 8) != 0 ||
	    anchor->primary_lba != hdr->primary_lba ||
	    anchor->secondary_lba != hdr->secondary_lba ||
	    hdr->type != type ||
	    memcmp(anchor->pad2, hdr->pad2, 512 -
		   offsetof(struct ddf_header, pad2)) != 0)
		return 0;

	/* Looks good enough to me... */
	return 1;
}

static void *load_section(int fd, struct ddf_super *super, void *buf,
			  __u32 offset_be, __u32 len_be, int check)
{
	unsigned long long offset = __be32_to_cpu(offset_be);
	unsigned long long len = __be32_to_cpu(len_be);
	int dofree = (buf == NULL);

	if (check)
		if (len != 2 && len != 8 && len != 32
		    && len != 128 && len != 512)
			return NULL;

	if (len > 1024)
		return NULL;
	if (buf) {
		/* All pre-allocated sections are a single block */
		if (len != 1)
			return NULL;
	} else if (posix_memalign(&buf, 512, len<<9) != 0)
		buf = NULL;

	if (!buf)
		return NULL;

	if (super->active->type == 1)
		offset += __be64_to_cpu(super->active->primary_lba);
	else
		offset += __be64_to_cpu(super->active->secondary_lba);

	if ((unsigned long long)lseek(fd, offset<<9, 0) != (offset<<9)) {
		if (dofree)
			free(buf);
		return NULL;
	}
	if ((unsigned long long)read(fd, buf, len<<9) != (len<<9)) {
		if (dofree)
			free(buf);
		return NULL;
	}
	return buf;
}

static int load_ddf_headers(int fd, struct ddf_super *super, char *devname)
{
	unsigned long long dsize;

	get_dev_size(fd, NULL, &dsize);

	if (lseek(fd, dsize-512, 0) < 0) {
		if (devname)
			bb_perror_msg("Cannot seek to anchor block on %s", devname);
		return 1;
	}
	if (read(fd, &super->anchor, 512) != 512) {
		if (devname)
			bb_perror_msg("Cannot read anchor block on %s", devname);
		return 1;
	}
	if (super->anchor.magic != DDF_HEADER_MAGIC) {
		if (devname)
			bb_error_msg("no DDF anchor found on %s", devname);
		return 2;
	}
	if (calc_crc(&super->anchor, 512) != super->anchor.crc) {
		if (devname)
			bb_error_msg("bad CRC on anchor on %s", devname);
		return 2;
	}
	if (memcmp(super->anchor.revision, DDF_REVISION_0, 8) != 0 &&
	    memcmp(super->anchor.revision, DDF_REVISION_2, 8) != 0) {
		if (devname)
			bb_error_msg("can only support super revision"
				" %.8s and earlier, not %.8s on %s",
				DDF_REVISION_2, super->anchor.revision,devname);
		return 2;
	}
	if (load_ddf_header(fd, __be64_to_cpu(super->anchor.primary_lba),
			    dsize >> 9,  1,
			    &super->primary, &super->anchor) == 0) {
		if (devname)
			bb_error_msg("Failed to load primary DDF header on %s", devname);
		return 2;
	}
	super->active = &super->primary;
	if (load_ddf_header(fd, __be64_to_cpu(super->anchor.secondary_lba),
			    dsize >> 9,  2,
			    &super->secondary, &super->anchor)) {
		if ((__be32_to_cpu(super->primary.seq)
		     < __be32_to_cpu(super->secondary.seq) &&
		     !super->secondary.openflag)
		    || (__be32_to_cpu(super->primary.seq)
			== __be32_to_cpu(super->secondary.seq) &&
			super->primary.openflag && !super->secondary.openflag)
			)
			super->active = &super->secondary;
	}
	return 0;
}

static int load_ddf_global(int fd, struct ddf_super *super)
{
	void *ok;
	ok = load_section(fd, super, &super->controller,
			  super->active->controller_section_offset,
			  super->active->controller_section_length,
			  0);
	super->phys = load_section(fd, super, NULL,
				   super->active->phys_section_offset,
				   super->active->phys_section_length,
				   1);
	super->pdsize = __be32_to_cpu(super->active->phys_section_length) * 512;

	super->virt = load_section(fd, super, NULL,
				   super->active->virt_section_offset,
				   super->active->virt_section_length,
				   1);
	super->vdsize = __be32_to_cpu(super->active->virt_section_length) * 512;
	if (!ok ||
	    !super->phys ||
	    !super->virt) {
		free(super->phys);
		free(super->virt);
		super->phys = NULL;
		super->virt = NULL;
		return 2;
	}
	super->conflist = NULL;
	super->dlist = NULL;

	super->max_part = __be16_to_cpu(super->active->max_partitions);
	super->mppe = __be16_to_cpu(super->active->max_primary_element_entries);
	super->conf_rec_len = __be16_to_cpu(super->active->config_record_len);
	return 0;
}

static int load_ddf_local(int fd, struct ddf_super *super,
			  char *devname, int keep)
{
	struct ddf_dl *dl;
	struct stat stb;
	char *conf;
	unsigned int i;
	unsigned int confsec;
	int vnum;
	unsigned int max_virt_disks = __be16_to_cpu(super->active->max_vd_entries);
	unsigned long long dsize;

	/* First the local disk info */
	if (posix_memalign((void**)&dl, 512,
		       sizeof(*dl) +
		       (super->max_part) * sizeof(dl->vlist[0])) != 0) {
		bb_error_msg("%s could not allocate disk info buffer", __func__);
		return 1;
	}

	load_section(fd, super, &dl->disk,
		     super->active->data_section_offset,
		     super->active->data_section_length,
		     0);
	dl->devname = devname ? strdup(devname) : NULL;

	fstat(fd, &stb);
	dl->major = major(stb.st_rdev);
	dl->minor = minor(stb.st_rdev);
	dl->next = super->dlist;
	dl->fd = keep ? fd : -1;

	dl->size = 0;
	if (get_dev_size(fd, devname, &dsize))
		dl->size = dsize >> 9;
	dl->spare = NULL;
	for (i = 0 ; i < super->max_part ; i++)
		dl->vlist[i] = NULL;
	super->dlist = dl;
	dl->pdnum = -1;
	for (i = 0; i < __be16_to_cpu(super->active->max_pd_entries); i++)
		if (memcmp(super->phys->entries[i].guid,
			   dl->disk.guid, DDF_GUID_LEN) == 0)
			dl->pdnum = i;

	/* Now the config list. */
	/* 'conf' is an array of config entries, some of which are
	 * probably invalid.  Those which are good need to be copied into
	 * the conflist
	 */

	conf = load_section(fd, super, NULL,
			    super->active->config_section_offset,
			    super->active->config_section_length,
			    0);

	vnum = 0;
	for (confsec = 0;
	     confsec < __be32_to_cpu(super->active->config_section_length);
	     confsec += super->conf_rec_len) {
		struct vd_config *vd =
			(struct vd_config *)((char*)conf + confsec*512);
		struct vcl *vcl;

		if (vd->magic == DDF_SPARE_ASSIGN_MAGIC) {
			if (dl->spare)
				continue;
			if (posix_memalign((void**)&dl->spare, 512,
				       super->conf_rec_len*512) != 0) {
				bb_error_msg("%s could not allocate spare info buf", __func__);
				return 1;
			}

			memcpy(dl->spare, vd, super->conf_rec_len*512);
			continue;
		}
		if (vd->magic != DDF_VD_CONF_MAGIC)
			continue;
		for (vcl = super->conflist; vcl; vcl = vcl->next) {
			if (memcmp(vcl->conf.guid,
				   vd->guid, DDF_GUID_LEN) == 0)
				break;
		}

		if (vcl) {
			dl->vlist[vnum++] = vcl;
			if (__be32_to_cpu(vd->seqnum) <=
			    __be32_to_cpu(vcl->conf.seqnum))
				continue;
		} else {
			if (posix_memalign((void**)&vcl, 512,
				       (super->conf_rec_len*512 +
					offsetof(struct vcl, conf))) != 0) {
				bb_error_msg("%s could not allocate vcl buf", __func__);
				return 1;
			}
			vcl->next = super->conflist;
			vcl->block_sizes = NULL; /* FIXME not for CONCAT */
			super->conflist = vcl;
			dl->vlist[vnum++] = vcl;
		}
		memcpy(&vcl->conf, vd, super->conf_rec_len*512);
		vcl->lba_offset = (__u64*)
			&vcl->conf.phys_refnum[super->mppe];

		for (i=0; i < max_virt_disks ; i++)
			if (memcmp(super->virt->entries[i].guid,
				   vcl->conf.guid, DDF_GUID_LEN)==0)
				break;
		if (i < max_virt_disks)
			vcl->vcnum = i;
	}
	free(conf);

	return 0;
}

static void free_super_ddf(struct supertype *st);

static int load_super_ddf(struct supertype *st, int fd,
			  char *devname)
{
	unsigned long long dsize;
	struct ddf_super *super;
	int rv;

	if (get_dev_size(fd, devname, &dsize) == 0)
		return 1;

	if (test_partition(fd))
		/* DDF is not allowed on partitions */
		return 1;

	/* 32M is a lower bound */
	if (dsize <= 32*1024*1024) {
		if (devname)
			bb_error_msg("%s is too small for ddf: "
				"size is %llu sectors.", devname, dsize>>9);
		return 1;
	}
	if (dsize & 511) {
		if (devname)
			bb_error_msg("%s is an odd size for ddf: "
				"size is %llu bytes.", devname, dsize);
		return 1;
	}

	free_super_ddf(st);

	if (posix_memalign((void**)&super, 512, sizeof(*super))!= 0) {
		bb_error_msg("malloc of %zu failed.", sizeof(*super));
		return 1;
	}
	memset(super, 0, sizeof(*super));

	rv = load_ddf_headers(fd, super, devname);
	if (rv) {
		free(super);
		return rv;
	}

	/* Have valid headers and have chosen the best. Let's read in the rest*/

	rv = load_ddf_global(fd, super);

	if (rv) {
		if (devname)
			bb_error_msg("Failed to load all information "
				"sections on %s", devname);
		free(super);
		return rv;
	}

	rv = load_ddf_local(fd, super, devname, 0);

	if (rv) {
		if (devname)
			bb_error_msg("Failed to load all information "
				"sections on %s", devname);
		free(super);
		return rv;
	}

	/* Should possibly check the sections .... */

	st->sb = super;
	if (st->ss == NULL) {
		st->ss = &super_ddf;
		st->minor_version = 0;
		st->max_devs = 512;
	}
	return 0;

}

static void free_super_ddf(struct supertype *st)
{
	struct ddf_super *ddf = st->sb;
	if (ddf == NULL)
		return;
	free(ddf->phys);
	free(ddf->virt);
	while (ddf->conflist) {
		struct vcl *v = ddf->conflist;
		ddf->conflist = v->next;
		if (v->block_sizes)
			free(v->block_sizes);
		free(v);
	}
	while (ddf->dlist) {
		struct ddf_dl *d = ddf->dlist;
		ddf->dlist = d->next;
		if (d->fd >= 0)
			close(d->fd);
		if (d->spare)
			free(d->spare);
		free(d);
	}
	while (ddf->add_list) {
		struct ddf_dl *d = ddf->add_list;
		ddf->add_list = d->next;
		if (d->fd >= 0)
			close(d->fd);
		if (d->spare)
			free(d->spare);
		free(d);
	}
	free(ddf);
	st->sb = NULL;
}

static struct supertype *match_metadata_desc_ddf(char *arg)
{
	/* 'ddf' only support containers */
	struct supertype *st;
	if (strcmp(arg, "ddf") != 0 &&
	    strcmp(arg, bb_default) != 0
		)
		return NULL;

	st = zmalloc(sizeof(*st));
	st->container_dev = NoMdDev;
	st->ss = &super_ddf;
	st->max_devs = 512;
	st->minor_version = 0;
	st->sb = NULL;
	return st;
}


struct num_mapping {
	int num1, num2;
};

static const struct num_mapping ddf_level_num[] = {
	{ DDF_RAID0, 0 },
	{ DDF_RAID1, 1 },
	{ DDF_RAID3, LEVEL_UNSUPPORTED },
	{ DDF_RAID4, 4 },
	{ DDF_RAID5, 5 },
	{ DDF_RAID1E, LEVEL_UNSUPPORTED },
	{ DDF_JBOD, LEVEL_UNSUPPORTED },
	{ DDF_CONCAT, LEVEL_LINEAR },
	{ DDF_RAID5E, LEVEL_UNSUPPORTED },
	{ DDF_RAID5EE, LEVEL_UNSUPPORTED },
	{ DDF_RAID6, 6},
	{ INT_MAX, INT_MAX }
};

static int map_num1(const struct num_mapping *map, int num)
{
	int i;
	for (i=0 ; map[i].num1 != INT_MAX; i++)
		if (map[i].num1 == num)
			break;
	return map[i].num2;
}

static int match_home_ddf(struct supertype *st, char *homehost)
{
	/* It matches 'this' host if the controller is a
	 * Linux-MD controller with vendor_data matching
	 * the hostname
	 */
	struct ddf_super *ddf = st->sb;
	unsigned int len;

	if (!homehost)
		return 0;
	len = strlen(homehost);

	return (memcmp(ddf->controller.guid, T10, 8) == 0 &&
		len < sizeof(ddf->controller.vendor_data) &&
		memcmp(ddf->controller.vendor_data, homehost,len) == 0 &&
		ddf->controller.vendor_data[len] == 0);
}

static int find_phys(struct ddf_super *ddf, __u32 phys_refnum)
{
	/* Find the entry in phys_disk which has the given refnum
	 * and return it's index
	 */
	unsigned int i;
	for (i = 0; i < __be16_to_cpu(ddf->phys->max_pdes); i++)
		if (ddf->phys->entries[i].refnum == phys_refnum)
			return i;
	return -1;
}

static void uuid_from_super_ddf(struct supertype *st, int uuid[4])
{
	/* The uuid returned here is used for:
	 *  uuid to put into bitmap file (Create, Grow)
	 *  uuid for backup header when saving critical section (Grow)
	 *  comparing uuids when re-adding a device into an array
	 *    In these cases the uuid required is that of the data-array,
	 *    not the device-set.
	 *  uuid to recognise same set when adding a missing device back
	 *    to an array.   This is a uuid for the device-set.
	 *
	 * For each of these we can make do with a truncated
	 * or hashed uuid rather than the original, as long as
	 * everyone agrees.
	 * In the case of SVD we assume the BVD is of interest,
	 * though that might be the case if a bitmap were made for
	 * a mirrored SVD - worry about that later.
	 * So we need to find the VD configuration record for the
	 * relevant BVD and extract the GUID and Secondary_Element_Seq.
	 * The first 16 bytes of the sha1 of these is used.
	 */
	struct ddf_super *ddf = st->sb;
	struct vcl *vcl = ddf->currentconf;
	char *guid;
	char buf[20];
	sha1_ctx_t ctx;

	if (vcl)
		guid = vcl->conf.guid;
	else
		guid = ddf->anchor.guid;

	sha1_begin(&ctx);
	sha1_hash(guid, DDF_GUID_LEN, &ctx);
	sha1_end(buf, &ctx);
	memcpy(uuid, buf, 4*4);
}

static void getinfo_super_ddf_bvd(struct supertype *st, struct mdinfo *info, char *map);

static void getinfo_super_ddf(struct supertype *st, struct mdinfo *info, char *map)
{
	struct ddf_super *ddf = st->sb;
	int map_disks = info->array.raid_disks;
	__u32 *cptr;

	if (ddf->currentconf) {
		getinfo_super_ddf_bvd(st, info, map);
		return;
	}
	memset(info, 0, sizeof(*info));

	info->array.raid_disks    = __be16_to_cpu(ddf->phys->used_pdes);
	info->array.level         = LEVEL_CONTAINER;
	info->array.layout        = 0;
	info->array.md_minor      = -1;
	cptr = (__u32 *)(ddf->anchor.guid + 16);
	info->array.ctime         = DECADE + __be32_to_cpu(*cptr);

	info->array.utime         = 0;
	info->array.chunk_size    = 0;
	info->container_enough    = 1;


	info->disk.major = 0;
	info->disk.minor = 0;
	if (ddf->dlist) {
		info->disk.number = __be32_to_cpu(ddf->dlist->disk.refnum);
		info->disk.raid_disk = find_phys(ddf, ddf->dlist->disk.refnum);

		info->data_offset = __be64_to_cpu(ddf->phys->
					  entries[info->disk.raid_disk].
					  config_size);
		info->component_size = ddf->dlist->size - info->data_offset;
	} else {
		info->disk.number = -1;
		info->disk.raid_disk = -1;
//              info->disk.raid_disk = find refnum in the table and use index;
	}
	info->disk.state = (1 << MD_DISK_SYNC) | (1 << MD_DISK_ACTIVE);


	info->recovery_start = MaxSector;
	info->reshape_active = 0;
	info->recovery_blocked = 0;
	info->name[0] = 0;

	info->array.major_version = -1;
	info->array.minor_version = -2;
	strcpy(info->text_version, "ddf");
	info->safe_mode_delay = 0;

	uuid_from_super_ddf(st, info->uuid);

	if (map) {
		int i;
		for (i = 0 ; i < map_disks; i++) {
			if (i < info->array.raid_disks &&
			    (__be16_to_cpu(ddf->phys->entries[i].state) & DDF_Online) &&
			    !(__be16_to_cpu(ddf->phys->entries[i].state) & DDF_Failed))
				map[i] = 1;
			else
				map[i] = 0;
		}
	}
}

static int rlq_to_layout(int rlq, int prl);

static void getinfo_super_ddf_bvd(struct supertype *st, struct mdinfo *info, char *map)
{
	struct ddf_super *ddf = st->sb;
	struct vcl *vc = ddf->currentconf;
	int cd = ddf->currentdev;
	int j;
	struct ddf_dl *dl;
	int map_disks = info->array.raid_disks;
	__u32 *cptr;

	memset(info, 0, sizeof(*info));
	/* FIXME this returns BVD info - what if we want SVD ?? */

	info->array.raid_disks    = __be16_to_cpu(vc->conf.prim_elmnt_count);
	info->array.level         = map_num1(ddf_level_num, vc->conf.prl);
	info->array.layout        = rlq_to_layout(vc->conf.rlq, vc->conf.prl);
	info->array.md_minor      = -1;
	cptr = (__u32 *)(vc->conf.guid + 16);
	info->array.ctime         = DECADE + __be32_to_cpu(*cptr);
	info->array.utime         = DECADE + __be32_to_cpu(vc->conf.timestamp);
	info->array.chunk_size    = 512 << vc->conf.chunk_shift;
	info->custom_array_size   = 0;

	if (cd >= 0 && (unsigned)cd < ddf->mppe) {
		info->data_offset         = __be64_to_cpu(vc->lba_offset[cd]);
		if (vc->block_sizes)
			info->component_size = vc->block_sizes[cd];
		else
			info->component_size = __be64_to_cpu(vc->conf.blocks);
	}

	for (dl = ddf->dlist; dl ; dl = dl->next)
		if (dl->raiddisk == ddf->currentdev)
			break;

	info->disk.major = 0;
	info->disk.minor = 0;
	info->disk.state = 0;
	if (dl) {
		info->disk.major = dl->major;
		info->disk.minor = dl->minor;
		info->disk.raid_disk = dl->raiddisk;
		info->disk.number = dl->pdnum;
		info->disk.state = (1<<MD_DISK_SYNC)|(1<<MD_DISK_ACTIVE);
	}

	info->container_member = ddf->currentconf->vcnum;

	info->recovery_start = MaxSector;
	info->resync_start = 0;
	info->reshape_active = 0;
	info->recovery_blocked = 0;
	if (!(ddf->virt->entries[info->container_member].state
	      & DDF_state_inconsistent)  &&
	    (ddf->virt->entries[info->container_member].init_state
	     & DDF_initstate_mask)
	    == DDF_init_full)
		info->resync_start = MaxSector;

	uuid_from_super_ddf(st, info->uuid);

	info->array.major_version = -1;
	info->array.minor_version = -2;
	sprintf(info->text_version, "/%s/%d",
		devnum2devname(st->container_dev),
		info->container_member);
	info->safe_mode_delay = 200;

	memcpy(info->name, ddf->virt->entries[info->container_member].name, 16);
	info->name[16]=0;
	for(j=0; j<16; j++)
		if (info->name[j] == ' ')
			info->name[j] = 0;

	if (map)
		for (j = 0; j < map_disks; j++) {
			map[j] = 0;
			if (j <  info->array.raid_disks) {
				int i = find_phys(ddf, vc->conf.phys_refnum[j]);
				if (i >= 0 &&
				    (__be16_to_cpu(ddf->phys->entries[i].state) & DDF_Online) &&
				    !(__be16_to_cpu(ddf->phys->entries[i].state) & DDF_Failed))
					map[i] = 1;
			}
		}
}


static int update_super_ddf(struct supertype *st, struct mdinfo *info,
			    char *update,
			    char *devname, int verbose,
			    int uuid_set, char *homehost)
{
	/* For 'assemble' and 'force' we need to return non-zero if any
	 * change was made.  For others, the return value is ignored.
	 * Update options are:
	 *  force-one : This device looks a bit old but needs to be included,
	 *        update age info appropriately.
	 *  assemble: clear any 'faulty' flag to allow this device to
	 *              be assembled.
	 *  force-array: Array is degraded but being forced, mark it clean
	 *         if that will be needed to assemble it.
	 *
	 *  newdev:  not used ????
	 *  grow:  Array has gained a new device - this is currently for
	 *              linear only
	 *  resync: mark as dirty so a resync will happen.
	 *  uuid:  Change the uuid of the array to match what is given
	 *  homehost:  update the recorded homehost
	 *  name:  update the name - preserving the homehost
	 *  _reshape_progress: record new reshape_progress position.
	 *
	 * Following are not relevant for this version:
	 *  sparc2.2 : update from old dodgey metadata
	 *  super-minor: change the preferred_minor number
	 *  summaries:  update redundant counters.
	 */
	int rv = 0;

	(void)&st;
	(void)&info;
	(void)&devname;
	(void)&verbose;
	(void)&uuid_set;
	(void)&homehost;

	/* we don't need to handle "force-*" or "assemble" as
	 * there is no need to 'trick' the kernel.  We the metadata is
	 * first updated to activate the array, all the implied modifications
	 * will just happen.
	 */

	if (strcmp(update, "grow") == 0) {
		/* FIXME */
	} else if (strcmp(update, "resync") == 0) {
//              info->resync_checkpoint = 0;
	} else if (strcmp(update, "homehost") == 0) {
		/* homehost is stored in controller->vendor_data,
		 * or it is when we are the vendor
		 */
//              if (info->vendor_is_local)
//                      strcpy(ddf->controller.vendor_data, homehost);
		rv = -1;
	} else if (strcmp(update, "name") == 0) {
		/* name is stored in virtual_entry->name */
//              memset(ve->name, ' ', 16);
//              strncpy(ve->name, info->name, 16);
		rv = -1;
	} else if (strcmp(update, "_reshape_progress") == 0) {
		/* We don't support reshape yet */
	} else if (strcmp(update, "assemble") == 0 ) {
		/* Do nothing, just succeed */
		rv = 0;
	} else
		rv = -1;

//      update_all_csum(ddf);

	return rv;
}

static int rlq_to_layout(int rlq, int prl)
{
	switch(prl) {
	case DDF_RAID0:
		return 0; /* hopefully rlq == DDF_RAID0_SIMPLE */
	case DDF_RAID1:
		return 0; /* hopefully rlq == SIMPLE or MULTI depending
			     on raiddisks*/
	case DDF_RAID4:
		switch(rlq) {
		case DDF_RAID4_N:
			return 0;
		default:
			/* not supported */
			return -1; /* FIXME this isn't checked */
		}
	case DDF_RAID5:
		switch(rlq) {
		case DDF_RAID5_N_RESTART:
			return ALGORITHM_LEFT_ASYMMETRIC;
		case DDF_RAID5_0_RESTART:
			return ALGORITHM_RIGHT_ASYMMETRIC;
		case DDF_RAID5_N_CONTINUE:
			return ALGORITHM_LEFT_SYMMETRIC;
		default:
			return -1;
		}
	case DDF_RAID6:
		switch(rlq) {
		case DDF_RAID5_N_RESTART:
			return ALGORITHM_ROTATING_N_RESTART;
		case DDF_RAID6_0_RESTART:
			return ALGORITHM_ROTATING_ZERO_RESTART;
		case DDF_RAID5_N_CONTINUE:
			return ALGORITHM_ROTATING_N_CONTINUE;
		default:
			return -1;
		}
	}
	return -1;
}

static struct mdinfo *container_content_ddf(struct supertype *st, char *subarray)
{
	/* Given a container loaded by load_super_ddf_all,
	 * extract information about all the arrays into
	 * an mdinfo tree.
	 *
	 * For each vcl in conflist: create an mdinfo, fill it in,
	 *  then look for matching devices (phys_refnum) in dlist
	 *  and create appropriate device mdinfo.
	 */
	struct ddf_super *ddf = st->sb;
	struct mdinfo *rest = NULL;
	struct vcl *vc;

	for (vc = ddf->conflist ; vc ; vc=vc->next)
	{
		unsigned int i;
		unsigned int j;
		struct mdinfo *this;
		char *ep;
		__u32 *cptr;

		if (subarray &&
		    (strtoul(subarray, &ep, 10) != vc->vcnum ||
		     *ep != '\0'))
			continue;

		this = zmalloc(sizeof(*this));
		this->next = rest;
		rest = this;

		this->array.level = map_num1(ddf_level_num, vc->conf.prl);
		this->array.raid_disks =
			__be16_to_cpu(vc->conf.prim_elmnt_count);
		this->array.layout = rlq_to_layout(vc->conf.rlq, vc->conf.prl);
		this->array.md_minor      = -1;
		this->array.major_version = -1;
		this->array.minor_version = -2;
		cptr = (__u32 *)(vc->conf.guid + 16);
		this->array.ctime         = DECADE + __be32_to_cpu(*cptr);
		this->array.utime         = DECADE +
			__be32_to_cpu(vc->conf.timestamp);
		this->array.chunk_size    = 512 << vc->conf.chunk_shift;

		i = vc->vcnum;
		if ((ddf->virt->entries[i].state & DDF_state_inconsistent) ||
		    (ddf->virt->entries[i].init_state & DDF_initstate_mask) !=
		    DDF_init_full) {
			this->array.state = 0;
			this->resync_start = 0;
		} else {
			this->array.state = 1;
			this->resync_start = MaxSector;
		}
		memcpy(this->name, ddf->virt->entries[i].name, 16);
		this->name[16]=0;
		for(j=0; j<16; j++)
			if (this->name[j] == ' ')
				this->name[j] = 0;

		memset(this->uuid, 0, sizeof(this->uuid));
		this->component_size = __be64_to_cpu(vc->conf.blocks);
		this->array.size = this->component_size / 2;
		this->container_member = i;

		ddf->currentconf = vc;
		uuid_from_super_ddf(st, this->uuid);
		ddf->currentconf = NULL;

		sprintf(this->text_version, "/%s/%d",
			devnum2devname(st->container_dev),
			this->container_member);

		for (i = 0 ; i < ddf->mppe ; i++) {
			struct mdinfo *dev;
			struct ddf_dl *d;
			int stt;
			int pd;

			if (vc->conf.phys_refnum[i] == 0xFFFFFFFF)
				continue;

			for (pd = __be16_to_cpu(ddf->phys->used_pdes);
			     pd--;)
				if (ddf->phys->entries[pd].refnum
				    == vc->conf.phys_refnum[i])
					break;
			if (pd < 0)
				continue;

			stt = __be16_to_cpu(ddf->phys->entries[pd].state);
			if ((stt & (DDF_Online|DDF_Failed|DDF_Rebuilding))
			    != DDF_Online)
				continue;

			this->array.working_disks++;

			for (d = ddf->dlist; d ; d=d->next)
				if (d->disk.refnum == vc->conf.phys_refnum[i])
					break;
			if (d == NULL)
				/* Haven't found that one yet, maybe there are others */
				continue;

			dev = zmalloc(sizeof(*dev));
			dev->next = this->devs;
			this->devs = dev;

			dev->disk.number = __be32_to_cpu(d->disk.refnum);
			dev->disk.major = d->major;
			dev->disk.minor = d->minor;
			dev->disk.raid_disk = i;
			dev->disk.state = (1<<MD_DISK_SYNC)|(1<<MD_DISK_ACTIVE);
			dev->recovery_start = MaxSector;

			dev->events = __be32_to_cpu(ddf->primary.seq);
			dev->data_offset = __be64_to_cpu(vc->lba_offset[i]);
			dev->component_size = __be64_to_cpu(vc->conf.blocks);
			if (d->devname)
				strcpy(dev->name, d->devname);
		}
	}
	return rest;
}

static int store_super_ddf(struct supertype *st, int fd)
{
	struct ddf_super *ddf = st->sb;
	unsigned long long dsize;
	void *buf;
	int rc;

	if (!ddf)
		return 1;

	/* ->dlist and ->conflist will be set for updates, currently not
	 * supported
	 */
	if (ddf->dlist || ddf->conflist)
		return 1;

	if (!get_dev_size(fd, NULL, &dsize))
		return 1;

	if (posix_memalign(&buf, 512, 512) != 0)
		return 1;
	memset(buf, 0, 512);

	lseek(fd, dsize-512, 0);
	rc = write(fd, buf, 512);
	free(buf);
	if (rc < 0)
		return 1;
	return 0;
}

static int compare_super_ddf(struct supertype *st, struct supertype *tst)
{
	/*
	 * return:
	 *  0 same, or first was empty, and second was copied
	 *  1 second had wrong number
	 *  2 wrong uuid
	 *  3 wrong other info
	 */
	struct ddf_super *first = st->sb;
	struct ddf_super *second = tst->sb;

	if (!first) {
		st->sb = tst->sb;
		tst->sb = NULL;
		return 0;
	}

	if (memcmp(first->anchor.guid, second->anchor.guid, DDF_GUID_LEN) != 0)
		return 2;

	/* FIXME should I look at anything else? */
	return 0;
}

static int ddf_level_to_layout(int level)
{
	switch(level) {
	case 0:
	case 1:
		return 0;
	case 5:
		return ALGORITHM_LEFT_SYMMETRIC;
	case 6:
		return ALGORITHM_ROTATING_N_CONTINUE;
	case 10:
		return 0x102;
	default:
		return UnSet;
	}
}

static void default_geometry_ddf(struct supertype *st, int *level, int *layout, int *chunk)
{
	(void)&st;
	(void)&chunk;

	if (level && *level == UnSet)
		*level = LEVEL_CONTAINER;

	if (level && layout && *layout == UnSet)
		*layout = ddf_level_to_layout(*level);
}

static const struct superswitch super_ddf = {
	.match_home     = match_home_ddf,
	.uuid_from_super= uuid_from_super_ddf,
	.getinfo_super  = getinfo_super_ddf,
	.update_super   = update_super_ddf,

	.compare_super  = compare_super_ddf,

	.load_super     = load_super_ddf,
	.store_super    = store_super_ddf,
	.free_super     = free_super_ddf,
	.match_metadata_desc = match_metadata_desc_ddf,
	.container_content = container_content_ddf,
	.default_geometry = default_geometry_ddf,

	.external       = 1,

	.name = "ddf",
};

/* MPB == Metadata Parameter Block */
#define MPB_SIGNATURE "Intel Raid ISM Cfg Sig. "
#define MPB_SIG_LEN (strlen(MPB_SIGNATURE))
#define MPB_VERSION_RAID0 "1.0.00"
#define MPB_VERSION_RAID1 "1.1.00"
#define MPB_VERSION_MANY_VOLUMES_PER_ARRAY "1.2.00"
#define MPB_VERSION_3OR4_DISK_ARRAY "1.2.01"
#define MPB_VERSION_RAID5 "1.2.02"
#define MPB_VERSION_5OR6_DISK_ARRAY "1.2.04"
#define MPB_VERSION_CNG "1.2.06"
#define MPB_VERSION_ATTRIBS "1.3.00"
#define MAX_SIGNATURE_LENGTH  32
#define MAX_RAID_SERIAL_LEN   16

/* supports RAID0 */
#define MPB_ATTRIB_RAID0                __cpu_to_le32(0x00000001)
/* supports RAID1 */
#define MPB_ATTRIB_RAID1                __cpu_to_le32(0x00000002)
/* supports RAID10 */
#define MPB_ATTRIB_RAID10               __cpu_to_le32(0x00000004)
/* supports RAID1E */
#define MPB_ATTRIB_RAID1E               __cpu_to_le32(0x00000008)
/* supports RAID5 */
#define MPB_ATTRIB_RAID5                __cpu_to_le32(0x00000010)
/* supports RAID CNG */
#define MPB_ATTRIB_RAIDCNG              __cpu_to_le32(0x00000020)
/* supports expanded stripe sizes of  256K, 512K and 1MB */
#define MPB_ATTRIB_EXP_STRIPE_SIZE      __cpu_to_le32(0x00000040)

/* The OROM Support RST Caching of Volumes */
#define MPB_ATTRIB_NVM                  __cpu_to_le32(0x02000000)
/* The OROM supports creating disks greater than 2TB */
#define MPB_ATTRIB_2TB_DISK             __cpu_to_le32(0x04000000)
/* The OROM supports Bad Block Management */
#define MPB_ATTRIB_BBM                  __cpu_to_le32(0x08000000)

/* THe OROM Supports NVM Caching of Volumes */
#define MPB_ATTRIB_NEVER_USE2           __cpu_to_le32(0x10000000)
/* The OROM supports creating volumes greater than 2TB */
#define MPB_ATTRIB_2TB                  __cpu_to_le32(0x20000000)
/* originally for PMP, now it's wasted b/c. Never use this bit! */
#define MPB_ATTRIB_NEVER_USE            __cpu_to_le32(0x40000000)
/* Verify MPB contents against checksum after reading MPB */
#define MPB_ATTRIB_CHECKSUM_VERIFY      __cpu_to_le32(0x80000000)

/* Define all supported attributes that have to be accepted by mdadm
 */
#define MPB_ATTRIB_SUPPORTED           (MPB_ATTRIB_CHECKSUM_VERIFY | \
					MPB_ATTRIB_2TB             | \
					MPB_ATTRIB_2TB_DISK        | \
					MPB_ATTRIB_RAID0           | \
					MPB_ATTRIB_RAID1           | \
					MPB_ATTRIB_RAID10          | \
					MPB_ATTRIB_RAID5           | \
					MPB_ATTRIB_EXP_STRIPE_SIZE)

/* Define attributes that are unused but not harmful */
#define MPB_ATTRIB_IGNORED              (MPB_ATTRIB_NEVER_USE)

#define MPB_SECTOR_CNT 2210
#define IMSM_RESERVED_SECTORS 4096
#define NUM_BLOCKS_DIRTY_STRIPE_REGION 2056
#define SECT_PER_MB_SHIFT 11

/* Disk configuration info. */
#define IMSM_MAX_DEVICES 255
struct imsm_disk {
	__u8 serial[MAX_RAID_SERIAL_LEN];/* 0xD8 - 0xE7 ascii serial number */
	__u32 total_blocks_lo;           /* 0xE8 - 0xEB total blocks lo */
	__u32 scsi_id;                   /* 0xEC - 0xEF scsi ID */
#define SPARE_DISK      __cpu_to_le32(0x01)  /* Spare */
#define CONFIGURED_DISK __cpu_to_le32(0x02)  /* Member of some RaidDev */
#define FAILED_DISK     __cpu_to_le32(0x04)  /* Permanent failure */
	__u32 status;                    /* 0xF0 - 0xF3 */
	__u32 owner_cfg_num; /* which config 0,1,2... owns this disk */
	__u32 total_blocks_hi;           /* 0xF4 - 0xF5 total blocks hi */
#define IMSM_DISK_FILLERS       3
	__u32 filler[IMSM_DISK_FILLERS]; /* 0xF5 - 0x107 MPB_DISK_FILLERS for future expansion */
};

/* map selector for map managment
 */
#define MAP_0           0
#define MAP_1           1
#define MAP_X           -1

/* RAID map configuration infos. */
struct imsm_map {
	__u32 pba_of_lba0_lo;   /* start address of partition */
	__u32 blocks_per_member_lo;/* blocks per member */
	__u32 num_data_stripes_lo;      /* number of data stripes */
	__u16 blocks_per_strip;
	__u8  map_state;        /* Normal, Uninitialized, Degraded, Failed */
#define IMSM_T_STATE_NORMAL 0
#define IMSM_T_STATE_UNINITIALIZED 1
#define IMSM_T_STATE_DEGRADED 2
#define IMSM_T_STATE_FAILED 3
	__u8  raid_level;
#define IMSM_T_RAID0 0
#define IMSM_T_RAID1 1
#define IMSM_T_RAID5 5          /* since metadata version 1.2.02 ? */
	__u8  num_members;      /* number of member disks */
	__u8  num_domains;      /* number of parity domains */
	__u8  failed_disk_num;  /* valid only when state is degraded */
	__u8  ddf;
	__u32 pba_of_lba0_hi;
	__u32 blocks_per_member_hi;
	__u32 num_data_stripes_hi;
	__u32 filler[4];        /* expansion area */
#define IMSM_ORD_REBUILD (1 << 24)
	__u32 disk_ord_tbl[1];  /* disk_ord_tbl[num_members],
				 * top byte contains some flags
				 */
} __attribute__ ((packed));

struct imsm_vol {
	__u32 curr_migr_unit;
	__u32 checkpoint_id;    /* id to access curr_migr_unit */
	__u8  migr_state;       /* Normal or Migrating */
#define MIGR_INIT 0
#define MIGR_REBUILD 1
#define MIGR_VERIFY 2 /* analagous to echo check > sync_action */
#define MIGR_GEN_MIGR 3
#define MIGR_STATE_CHANGE 4
#define MIGR_REPAIR 5
	__u8  migr_type;        /* Initializing, Rebuilding, ... */
	__u8  dirty;
	__u8  fs_state;         /* fast-sync state for CnG (0xff == disabled) */
	__u16 verify_errors;    /* number of mismatches */
	__u16 bad_blocks;       /* number of bad blocks during verify */
	__u32 filler[4];
	struct imsm_map map[1];
	/* here comes another one if migr_state */
} __attribute__ ((packed));

struct imsm_dev {
	__u8  volume[MAX_RAID_SERIAL_LEN];
	__u32 size_low;
	__u32 size_high;
#define DEV_BOOTABLE            __cpu_to_le32(0x01)
#define DEV_BOOT_DEVICE         __cpu_to_le32(0x02)
#define DEV_READ_COALESCING     __cpu_to_le32(0x04)
#define DEV_WRITE_COALESCING    __cpu_to_le32(0x08)
#define DEV_LAST_SHUTDOWN_DIRTY __cpu_to_le32(0x10)
#define DEV_HIDDEN_AT_BOOT      __cpu_to_le32(0x20)
#define DEV_CURRENTLY_HIDDEN    __cpu_to_le32(0x40)
#define DEV_VERIFY_AND_FIX      __cpu_to_le32(0x80)
#define DEV_MAP_STATE_UNINIT    __cpu_to_le32(0x100)
#define DEV_NO_AUTO_RECOVERY    __cpu_to_le32(0x200)
#define DEV_CLONE_N_GO          __cpu_to_le32(0x400)
#define DEV_CLONE_MAN_SYNC      __cpu_to_le32(0x800)
#define DEV_CNG_MASTER_DISK_NUM __cpu_to_le32(0x1000)
	__u32 status;   /* Persistent RaidDev status */
	__u32 reserved_blocks; /* Reserved blocks at beginning of volume */
	__u8  migr_priority;
	__u8  num_sub_vols;
	__u8  tid;
	__u8  cng_master_disk;
	__u16 cache_policy;
	__u8  cng_state;
	__u8  cng_sub_state;
#define IMSM_DEV_FILLERS 10
	__u32 filler[IMSM_DEV_FILLERS];
	struct imsm_vol vol;
} __attribute__ ((packed));

struct imsm_super {
	__u8 sig[MAX_SIGNATURE_LENGTH]; /* 0x00 - 0x1F */
	__u32 check_sum;                /* 0x20 - 0x23 MPB Checksum */
	__u32 mpb_size;                 /* 0x24 - 0x27 Size of MPB */
	__u32 family_num;               /* 0x28 - 0x2B Checksum from first time this config was written */
	__u32 generation_num;           /* 0x2C - 0x2F Incremented each time this array's MPB is written */
	__u32 error_log_size;           /* 0x30 - 0x33 in bytes */
	__u32 attributes;               /* 0x34 - 0x37 */
	__u8 num_disks;                 /* 0x38 Number of configured disks */
	__u8 num_raid_devs;             /* 0x39 Number of configured volumes */
	__u8 error_log_pos;             /* 0x3A  */
	__u8 fill[1];                   /* 0x3B */
	__u32 cache_size;               /* 0x3c - 0x40 in mb */
	__u32 orig_family_num;          /* 0x40 - 0x43 original family num */
	__u32 pwr_cycle_count;          /* 0x44 - 0x47 simulated power cycle count for array */
	__u32 bbm_log_size;             /* 0x48 - 0x4B - size of bad Block Mgmt Log in bytes */
#define IMSM_FILLERS 35
	__u32 filler[IMSM_FILLERS];     /* 0x4C - 0xD7 RAID_MPB_FILLERS */
	struct imsm_disk disk[1];       /* 0xD8 diskTbl[numDisks] */
	/* here comes imsm_dev[num_raid_devs] */
	/* here comes BBM logs */
} __attribute__ ((packed));

#define BBM_LOG_MAX_ENTRIES 254

struct bbm_log_entry {
	__u64 defective_block_start;
#define UNREADABLE 0xFFFFFFFF
	__u32 spare_block_offset;
	__u16 remapped_marked_count;
	__u16 disk_ordinal;
} __attribute__ ((__packed__));

struct bbm_log {
	__u32 signature; /* 0xABADB10C */
	__u32 entry_count;
	__u32 reserved_spare_block_count; /* 0 */
	__u32 reserved; /* 0xFFFF */
	__u64 first_spare_lba;
	struct bbm_log_entry mapped_block_entries[BBM_LOG_MAX_ENTRIES];
} __attribute__ ((__packed__));


#define RAID_DISK_RESERVED_BLOCKS_IMSM_HI 2209

#define GEN_MIGR_AREA_SIZE 2048 /* General Migration Copy Area size in blocks */

#define MIGR_REC_BUF_SIZE 512 /* size of migr_record i/o buffer */
#define MIGR_REC_POSITION 512 /* migr_record position offset on disk,
			       * MIGR_REC_BUF_SIZE <= MIGR_REC_POSITION
			       */


#define UNIT_SRC_NORMAL     0   /* Source data for curr_migr_unit must
				 *  be recovered using srcMap */
#define UNIT_SRC_IN_CP_AREA 1   /* Source data for curr_migr_unit has
				 *  already been migrated and must
				 *  be recovered from checkpoint area */
struct migr_record {
	__u32 rec_status;           /* Status used to determine how to restart
				     * migration in case it aborts
				     * in some fashion */
	__u32 curr_migr_unit;       /* 0..numMigrUnits-1 */
	__u32 family_num;           /* Family number of MPB
				     * containing the RaidDev
				     * that is migrating */
	__u32 ascending_migr;       /* True if migrating in increasing
				     * order of lbas */
	__u32 blocks_per_unit;      /* Num disk blocks per unit of operation */
	__u32 dest_depth_per_unit;  /* Num member blocks each destMap
				     * member disk
				     * advances per unit-of-operation */
	__u32 ckpt_area_pba;        /* Pba of first block of ckpt copy area */
	__u32 dest_1st_member_lba;  /* First member lba on first
				     * stripe of destination */
	__u32 num_migr_units;       /* Total num migration units-of-op */
	__u32 post_migr_vol_cap;    /* Size of volume after
				     * migration completes */
	__u32 post_migr_vol_cap_hi; /* Expansion space for LBA64 */
	__u32 ckpt_read_disk_num;   /* Which member disk in destSubMap[0] the
				     * migration ckpt record was read from
				     * (for recovered migrations) */
} __attribute__ ((__packed__));

struct md_list {
	/* usage marker:
	 *  1: load metadata
	 *  2: metadata does not match
	 *  4: already checked
	 */
	int   used;
	char  *devname;
	int   found;
	int   container;
	dev_t st_rdev;
	struct md_list *next;
};

#define pr_vrb(fmt, arg...) (void) (verbose && bb_error_msg(fmt, ##arg))

static __u8 migr_type(struct imsm_dev *dev)
{
	if (dev->vol.migr_type == MIGR_VERIFY &&
	    dev->status & DEV_VERIFY_AND_FIX)
		return MIGR_REPAIR;
	else
		return dev->vol.migr_type;
}

static unsigned int sector_count(__u32 bytes)
{
	return ROUND_UP(bytes, 512) / 512;
}

static unsigned int mpb_sectors(struct imsm_super *mpb)
{
	return sector_count(__le32_to_cpu(mpb->mpb_size));
}

struct intel_dev {
	struct imsm_dev *dev;
	struct intel_dev *next;
	unsigned index;
};

struct intel_hba {
	enum sys_dev_type type;
	char *path;
	char *pci_id;
	struct intel_hba *next;
};

enum action {
	DISK_REMOVE = 1,
	DISK_ADD
};

/* internal representation of IMSM metadata */
struct intel_super {
	union {
		void *buf; /* O_DIRECT buffer for reading/writing metadata */
		struct imsm_super *anchor; /* immovable parameters */
	};
	union {
		void *migr_rec_buf; /* buffer for I/O operations */
		struct migr_record *migr_rec; /* migration record */
	};
	int clean_migration_record_by_mdmon; /* when reshape is switched to next
		array, it indicates that mdmon is allowed to clean migration
		record */
	size_t len; /* size of the 'buf' allocation */
	void *next_buf; /* for realloc'ing buf from the manager */
	size_t next_len;
	int updates_pending; /* count of pending updates for mdmon */
	int current_vol; /* index of raid device undergoing creation */
	unsigned long long create_offset; /* common start for 'current_vol' */
	__u32 random; /* random data for seeding new family numbers */
	struct intel_dev *devlist;
	struct dl {
		struct dl *next;
		int index;
		__u8 serial[MAX_RAID_SERIAL_LEN];
		int major, minor;
		char *devname;
		struct imsm_disk disk;
		int fd;
		int extent_cnt;
		struct extent *e; /* for determining freespace @ create */
		int raiddisk; /* slot to fill in autolayout */
		enum action action;
	} *disks, *current_disk;
	struct dl *disk_mgmt_list; /* list of disks to add/remove while mdmon
				      active */
	struct dl *missing; /* disks removed while we weren't looking */
	struct bbm_log *bbm_log;
	struct intel_hba *hba; /* device path of the raid controller for this metadata */
	const struct imsm_orom *orom; /* platform firmware support */
	struct intel_super *next; /* (temp) list for disambiguating family_num */
};

struct extent {
	unsigned long long start, size;
};

static const char * const _sys_dev_type[] = {
	[SYS_DEV_UNKNOWN] = "Unknown",
	[SYS_DEV_SAS] = "SAS",
	[SYS_DEV_SATA] = "SATA"
};

static const char *get_sys_dev_type(enum sys_dev_type type)
{
	if (type >= SYS_DEV_MAX)
		type = SYS_DEV_UNKNOWN;

	return _sys_dev_type[type];
}

static struct intel_hba * alloc_intel_hba(struct sys_dev *device)
{
	struct intel_hba *result = xmalloc(sizeof(*result));
	result->type = device->type;
	result->path = strdup(device->path);
	result->next = NULL;
	if (result->path && (result->pci_id = strrchr(result->path, '/')) != NULL)
		result->pci_id++;
	return result;
}

static struct intel_hba * find_intel_hba(struct intel_hba *hba, struct sys_dev *device)
{
	struct intel_hba *result=NULL;
	for (result = hba; result; result = result->next) {
		if (result->type == device->type && strcmp(result->path, device->path) == 0)
			break;
	}
	return result;
}

static void free_sys_dev(struct sys_dev **list)
{
	while (*list) {
		struct sys_dev *next = (*list)->next;

		if ((*list)->path)
			free((*list)->path);
		free(*list);
		*list = next;
	}
}

static int attach_hba_to_super(struct intel_super *super, struct sys_dev *device)
{
	struct intel_hba *hba;

	/* check if disk attached to Intel HBA */
	hba = find_intel_hba(super->hba, device);
	if (hba != NULL)
		return 1;
	/* Check if HBA is already attached to super */
	if (super->hba == NULL) {
		super->hba = alloc_intel_hba(device);
		return 1;
	}

	hba = super->hba;
	/* Intel metadata allows for all disks attached to the same type HBA.
	 * Do not sypport odf HBA types mixing
	 */
	if (device->type != hba->type)
		return 2;

	while (hba->next)
		hba = hba->next;

	hba->next = alloc_intel_hba(device);
	return 1;
}

static struct sys_dev* find_disk_attached_hba(int fd, const char *devname)
{
	struct sys_dev *list, *elem, *prev;
	char *disk_dev_path;

	if ((list = find_intel_devices()) == NULL)
		return 0;

	if (fd < 0)
		disk_dev_path  = (char *) devname;
	else
		disk_dev_path = diskfd_to_devpath(fd);

	if (!disk_dev_path) {
		free_sys_dev(&list);
		return 0;
	}

	for (prev = NULL, elem = list; elem; prev = elem, elem = elem->next) {
		if (path_attached_to_hba(disk_dev_path, elem->path)) {
			if (prev == NULL)
				list = list->next;
			else
				prev->next = elem->next;
			elem->next = NULL;
			if (disk_dev_path != devname)
				free(disk_dev_path);
			free_sys_dev(&list);
			return elem;
		}
	}
	if (disk_dev_path != devname)
		free(disk_dev_path);
	free_sys_dev(&list);

	return NULL;
}


static int find_intel_hba_capability(int fd, struct intel_super *super,
				     char *devname);

static struct supertype *match_metadata_desc_imsm(char *arg)
{
	struct supertype *st;

	if (strcmp(arg, "imsm") != 0 && strcmp(arg, bb_default) != 0)
		return NULL;

	st = zmalloc(sizeof(*st));
	st->container_dev = NoMdDev;
	st->ss = &super_imsm;
	st->max_devs = IMSM_MAX_DEVICES;
	st->minor_version = 0;
	st->sb = NULL;
	return st;
}

/* retrieve a disk directly from the anchor when the anchor is known to be
 * up-to-date, currently only at load time
 */
static struct imsm_disk *__get_imsm_disk(struct imsm_super *mpb, __u8 idx)
{
	if (idx >= mpb->num_disks)
		return NULL;
	return &mpb->disk[idx];
}

/* retrieve the disk description based on a index of the disk
 * in the sub-array
 */
static struct dl *get_imsm_dl_disk(struct intel_super *super, __u8 idx)
{
	struct dl *d;

	for (d = super->disks; d; d = d->next)
		if (d->index == idx)
			return d;

	return NULL;
}
/* retrieve a disk from the parsed metadata */
static struct imsm_disk *get_imsm_disk(struct intel_super *super, __u8 idx)
{
	struct dl *dl;

	dl = get_imsm_dl_disk(super, idx);
	if (dl)
		return &dl->disk;

	return NULL;
}

/* generate a checksum directly from the anchor when the anchor is known to be
 * up-to-date, currently only at load or write_super after coalescing
 */
static __u32 __gen_imsm_checksum(struct imsm_super *mpb)
{
	__u32 end = mpb->mpb_size / sizeof(end);
	__u32 *p = (__u32 *) mpb;
	__u32 sum = 0;

	while (end--) {
		sum += __le32_to_cpu(*p);
		p++;
	}

	return sum - __le32_to_cpu(mpb->check_sum);
}

static size_t sizeof_imsm_map(struct imsm_map *map)
{
	return sizeof(struct imsm_map) + sizeof(__u32) * (map->num_members - 1);
}

static struct imsm_map *get_imsm_map(struct imsm_dev *dev, int second_map)
{
	/* A device can have 2 maps if it is in the middle of a migration.
	 * If second_map is:
	 *    MAP_0 - we return the first map
	 *    MAP_1 - we return the second map if it exists, else NULL
	 *    MAP_X - we return the second map if it exists, else the first
	 */
	struct imsm_map *map = &dev->vol.map[0];
	struct imsm_map *map2 = NULL;

	if (dev->vol.migr_state)
		map2 = (void *)map + sizeof_imsm_map(map);

	switch (second_map) {
	case MAP_0:
		break;
	case MAP_1:
		map = map2;
		break;
	case MAP_X:
		if (map2)
			map = map2;
		break;
	default:
		map = NULL;
	}
	return map;

}

/* return the size of the device.
 * migr_state increases the returned size if map[0] were to be duplicated
 */
static size_t sizeof_imsm_dev(struct imsm_dev *dev, int migr_state)
{
	size_t size = sizeof(*dev) - sizeof(struct imsm_map) +
		      sizeof_imsm_map(get_imsm_map(dev, MAP_0));

	/* migrating means an additional map */
	if (dev->vol.migr_state)
		size += sizeof_imsm_map(get_imsm_map(dev, MAP_1));
	else if (migr_state)
		size += sizeof_imsm_map(get_imsm_map(dev, MAP_0));

	return size;
}

static struct imsm_dev *__get_imsm_dev(struct imsm_super *mpb, __u8 idx)
{
	int offset;
	int i;
	void *_mpb = mpb;

	if (idx >= mpb->num_raid_devs)
		return NULL;

	/* devices start after all disks */
	offset = ((void *) &mpb->disk[mpb->num_disks]) - _mpb;

	for (i = 0; i <= idx; i++)
		if (i == idx)
			return _mpb + offset;
		else
			offset += sizeof_imsm_dev(_mpb + offset, 0);

	return NULL;
}

static struct imsm_dev *get_imsm_dev(struct intel_super *super, __u8 idx)
{
	struct intel_dev *dv;

	if (idx >= super->anchor->num_raid_devs)
		return NULL;
	for (dv = super->devlist; dv; dv = dv->next)
		if (dv->index == idx)
			return dv->dev;
	return NULL;
}

/*
 * for second_map:
 *  == MAP_0 get first map
 *  == MAP_1 get second map
 *  == MAP_X than get map according to the current migr_state
 */
static __u32 get_imsm_ord_tbl_ent(struct imsm_dev *dev,
				  int slot,
				  int second_map)
{
	struct imsm_map *map;

	map = get_imsm_map(dev, second_map);

	/* top byte identifies disk under rebuild */
	return __le32_to_cpu(map->disk_ord_tbl[slot]);
}

#define ord_to_idx(ord) (((ord) << 8) >> 8)
static __u32 get_imsm_disk_idx(struct imsm_dev *dev, int slot, int second_map)
{
	__u32 ord = get_imsm_ord_tbl_ent(dev, slot, second_map);

	return ord_to_idx(ord);
}

static int get_imsm_disk_slot(struct imsm_map *map, unsigned idx)
{
	int slot;
	__u32 ord;

	for (slot = 0; slot < map->num_members; slot++) {
		ord = __le32_to_cpu(map->disk_ord_tbl[slot]);
		if (ord_to_idx(ord) == idx)
			return slot;
	}

	return -1;
}

static int get_imsm_raid_level(struct imsm_map *map)
{
	if (map->raid_level == 1) {
		if (map->num_members == 2)
			return 1;
		else
			return 10;
	}

	return map->raid_level;
}

static int cmp_extent(const void *av, const void *bv)
{
	const struct extent *a = av;
	const struct extent *b = bv;
	if (a->start < b->start)
		return -1;
	if (a->start > b->start)
		return 1;
	return 0;
}

static int count_memberships(struct dl *dl, struct intel_super *super)
{
	int memberships = 0;
	int i;

	for (i = 0; i < super->anchor->num_raid_devs; i++) {
		struct imsm_dev *dev = get_imsm_dev(super, i);
		struct imsm_map *map = get_imsm_map(dev, MAP_0);

		if (get_imsm_disk_slot(map, dl->index) >= 0)
			memberships++;
	}

	return memberships;
}

static __u32 imsm_min_reserved_sectors(struct intel_super *super);

static unsigned long long join_u32(__u32 lo, __u32 hi)
{
	return (unsigned long long)__le32_to_cpu(lo) |
	       (((unsigned long long)__le32_to_cpu(hi)) << 32);
}

static unsigned long long total_blocks(struct imsm_disk *disk)
{
	if (disk == NULL)
		return 0;
	return join_u32(disk->total_blocks_lo, disk->total_blocks_hi);
}

static unsigned long long pba_of_lba0(struct imsm_map *map)
{
	if (map == NULL)
		return 0;
	return join_u32(map->pba_of_lba0_lo, map->pba_of_lba0_hi);
}

static unsigned long long blocks_per_member(struct imsm_map *map)
{
	if (map == NULL)
		return 0;
	return join_u32(map->blocks_per_member_lo, map->blocks_per_member_hi);
}

static struct extent *get_extents(struct intel_super *super, struct dl *dl)
{
	/* find a list of used extents on the given physical device */
	struct extent *rv, *e;
	int i;
	int memberships = count_memberships(dl, super);
	__u32 reservation;

	/* trim the reserved area for spares, so they can join any array
	 * regardless of whether the OROM has assigned sectors from the
	 * IMSM_RESERVED_SECTORS region
	 */
	if (dl->index == -1)
		reservation = imsm_min_reserved_sectors(super);
	else
		reservation = MPB_SECTOR_CNT + IMSM_RESERVED_SECTORS;

	rv = xmalloc(sizeof(struct extent) * (memberships + 1));
	e = rv;

	for (i = 0; i < super->anchor->num_raid_devs; i++) {
		struct imsm_dev *dev = get_imsm_dev(super, i);
		struct imsm_map *map = get_imsm_map(dev, MAP_0);

		if (get_imsm_disk_slot(map, dl->index) >= 0) {
			e->start = pba_of_lba0(map);
			e->size = blocks_per_member(map);
			e++;
		}
	}
	qsort(rv, memberships, sizeof(*rv), cmp_extent);

	/* determine the start of the metadata
	 * when no raid devices are defined use the default
	 * ...otherwise allow the metadata to truncate the value
	 * as is the case with older versions of imsm
	 */
	if (memberships) {
		struct extent *last = &rv[memberships - 1];
		unsigned long long remainder;

		remainder = total_blocks(&dl->disk) - (last->start + last->size);
		/* round down to 1k block to satisfy precision of the kernel
		 * 'size' interface
		 */
		remainder &= ~1UL;
		/* make sure remainder is still sane */
		if (remainder < (unsigned)ROUND_UP(super->len, 512) >> 9)
			remainder = ROUND_UP(super->len, 512) >> 9;
		if (reservation > remainder)
			reservation = remainder;
	}
	e->start = total_blocks(&dl->disk) - reservation;
	e->size = 0;
	return rv;
}

/* try to determine how much space is reserved for metadata from
 * the last get_extents() entry, otherwise fallback to the
 * default
 */
static __u32 imsm_reserved_sectors(struct intel_super *super, struct dl *dl)
{
	struct extent *e;
	int i;
	__u32 rv;

	/* for spares just return a minimal reservation which will grow
	 * once the spare is picked up by an array
	 */
	if (dl->index == -1)
		return MPB_SECTOR_CNT;

	e = get_extents(super, dl);
	if (!e)
		return MPB_SECTOR_CNT + IMSM_RESERVED_SECTORS;

	/* scroll to last entry */
	for (i = 0; e[i].size; i++)
		continue;

	rv = total_blocks(&dl->disk) - e[i].start;

	free(e);

	return rv;
}

static int is_spare(struct imsm_disk *disk)
{
	return (disk->status & SPARE_DISK) == SPARE_DISK;
}

static int is_configured(struct imsm_disk *disk)
{
	return (disk->status & CONFIGURED_DISK) == CONFIGURED_DISK;
}

static int is_failed(struct imsm_disk *disk)
{
	return (disk->status & FAILED_DISK) == FAILED_DISK;
}

/* try to determine how much space is reserved for metadata from
 * the last get_extents() entry on the smallest active disk,
 * otherwise fallback to the default
 */
static __u32 imsm_min_reserved_sectors(struct intel_super *super)
{
	struct extent *e;
	int i;
	unsigned long long min_active;
	__u32 remainder;
	__u32 rv = MPB_SECTOR_CNT + IMSM_RESERVED_SECTORS;
	struct dl *dl, *dl_min = NULL;

	if (!super)
		return rv;

	min_active = 0;
	for (dl = super->disks; dl; dl = dl->next) {
		if (dl->index < 0)
			continue;
		unsigned long long blocks = total_blocks(&dl->disk);
		if (blocks < min_active || min_active == 0) {
			dl_min = dl;
			min_active = blocks;
		}
	}
	if (!dl_min)
		return rv;

	/* find last lba used by subarrays on the smallest active disk */
	e = get_extents(super, dl_min);
	if (!e)
		return rv;
	for (i = 0; e[i].size; i++)
		continue;

	remainder = min_active - e[i].start;
	free(e);

	/* to give priority to recovery we should not require full
	   IMSM_RESERVED_SECTORS from the spare */
	rv = MPB_SECTOR_CNT + NUM_BLOCKS_DIRTY_STRIPE_REGION;

	/* if real reservation is smaller use that value */
	return  (remainder < rv) ? remainder : rv;
}

/* Return minimum size of a spare that can be used in this array*/
static unsigned long long min_acceptable_spare_size_imsm(struct supertype *st)
{
	struct intel_super *super = st->sb;
	struct dl *dl;
	struct extent *e;
	int i;
	unsigned long long rv = 0;

	if (!super)
		return rv;
	/* find first active disk in array */
	dl = super->disks;
	while (dl && (is_failed(&dl->disk) || dl->index == -1))
		dl = dl->next;
	if (!dl)
		return rv;
	/* find last lba used by subarrays */
	e = get_extents(super, dl);
	if (!e)
		return rv;
	for (i = 0; e[i].size; i++)
		continue;
	if (i > 0)
		rv = e[i-1].start + e[i-1].size;
	free(e);

	/* add the amount of space needed for metadata */
	rv = rv + imsm_min_reserved_sectors(super);

	return rv * 512;
}

static int is_gen_migration(struct imsm_dev *dev);

/*******************************************************************************
 * function: imsm_check_attributes
 * Description: Function checks if features represented by attributes flags
 *              are supported by mdadm.
 * Parameters:
 *              attributes - Attributes read from metadata
 * Returns:
 *              0 - passed attributes contains unsupported features flags
 *              1 - all features are supported
 ******************************************************************************/
static int imsm_check_attributes(__u32 attributes)
{
	int ret_val = 1;
	__u32 not_supported = MPB_ATTRIB_SUPPORTED^0xffffffff;

	not_supported &= ~MPB_ATTRIB_IGNORED;

	not_supported &= attributes;
	if (not_supported) {
		bb_error_msg("(IMSM): Unsupported attributes : %x",
			(unsigned)__le32_to_cpu(not_supported));
		if (not_supported & MPB_ATTRIB_CHECKSUM_VERIFY) {
			dprintf("\t\tMPB_ATTRIB_CHECKSUM_VERIFY \n");
			not_supported ^= MPB_ATTRIB_CHECKSUM_VERIFY;
		}
		if (not_supported & MPB_ATTRIB_2TB) {
			dprintf("\t\tMPB_ATTRIB_2TB\n");
			not_supported ^= MPB_ATTRIB_2TB;
		}
		if (not_supported & MPB_ATTRIB_RAID0) {
			dprintf("\t\tMPB_ATTRIB_RAID0\n");
			not_supported ^= MPB_ATTRIB_RAID0;
		}
		if (not_supported & MPB_ATTRIB_RAID1) {
			dprintf("\t\tMPB_ATTRIB_RAID1\n");
			not_supported ^= MPB_ATTRIB_RAID1;
		}
		if (not_supported & MPB_ATTRIB_RAID10) {
			dprintf("\t\tMPB_ATTRIB_RAID10\n");
			not_supported ^= MPB_ATTRIB_RAID10;
		}
		if (not_supported & MPB_ATTRIB_RAID1E) {
			dprintf("\t\tMPB_ATTRIB_RAID1E\n");
			not_supported ^= MPB_ATTRIB_RAID1E;
		}
		if (not_supported & MPB_ATTRIB_RAID5) {
		dprintf("\t\tMPB_ATTRIB_RAID5\n");
			not_supported ^= MPB_ATTRIB_RAID5;
		}
		if (not_supported & MPB_ATTRIB_RAIDCNG) {
			dprintf("\t\tMPB_ATTRIB_RAIDCNG\n");
			not_supported ^= MPB_ATTRIB_RAIDCNG;
		}
		if (not_supported & MPB_ATTRIB_BBM) {
			dprintf("\t\tMPB_ATTRIB_BBM\n");
		not_supported ^= MPB_ATTRIB_BBM;
		}
		if (not_supported & MPB_ATTRIB_CHECKSUM_VERIFY) {
			dprintf("\t\tMPB_ATTRIB_CHECKSUM_VERIFY (== MPB_ATTRIB_LEGACY)\n");
			not_supported ^= MPB_ATTRIB_CHECKSUM_VERIFY;
		}
		if (not_supported & MPB_ATTRIB_EXP_STRIPE_SIZE) {
			dprintf("\t\tMPB_ATTRIB_EXP_STRIP_SIZE\n");
			not_supported ^= MPB_ATTRIB_EXP_STRIPE_SIZE;
		}
		if (not_supported & MPB_ATTRIB_2TB_DISK) {
			dprintf("\t\tMPB_ATTRIB_2TB_DISK\n");
			not_supported ^= MPB_ATTRIB_2TB_DISK;
		}
		if (not_supported & MPB_ATTRIB_NEVER_USE2) {
			dprintf("\t\tMPB_ATTRIB_NEVER_USE2\n");
			not_supported ^= MPB_ATTRIB_NEVER_USE2;
		}
		if (not_supported & MPB_ATTRIB_NEVER_USE) {
			dprintf("\t\tMPB_ATTRIB_NEVER_USE\n");
			not_supported ^= MPB_ATTRIB_NEVER_USE;
		}

		if (not_supported)
			dprintf(Name "(IMSM): Unknown attributes : %x\n", not_supported);

		ret_val = 0;
	}

	return ret_val;
}

static int match_home_imsm(struct supertype *st, char *homehost)
{
	/* the imsm metadata format does not specify any host
	 * identification information.  We return -1 since we can never
	 * confirm nor deny whether a given array is "meant" for this
	 * host.  We rely on compare_super and the 'family_num' fields to
	 * exclude member disks that do not belong, and we rely on
	 * mdadm.conf to specify the arrays that should be assembled.
	 * Auto-assembly may still pick up "foreign" arrays.
	 */

	(void)&st;
	(void)&homehost;
	return -1;
}

static void uuid_from_super_imsm(struct supertype *st, int uuid[4])
{
	/* The uuid returned here is used for:
	 *  uuid to put into bitmap file (Create, Grow)
	 *  uuid for backup header when saving critical section (Grow)
	 *  comparing uuids when re-adding a device into an array
	 *    In these cases the uuid required is that of the data-array,
	 *    not the device-set.
	 *  uuid to recognise same set when adding a missing device back
	 *    to an array.   This is a uuid for the device-set.
	 *
	 * For each of these we can make do with a truncated
	 * or hashed uuid rather than the original, as long as
	 * everyone agrees.
	 * In each case the uuid required is that of the data-array,
	 * not the device-set.
	 */
	/* imsm does not track uuid's so we synthesis one using sha1 on
	 * - The signature (Which is constant for all imsm array, but no matter)
	 * - the orig_family_num of the container
	 * - the index number of the volume
	 * - the 'serial' number of the volume.
	 * Hopefully these are all constant.
	 */
	struct intel_super *super = st->sb;

	char buf[20];
	sha1_ctx_t ctx;
	struct imsm_dev *dev = NULL;
	__u32 family_num;

	/* some mdadm versions failed to set ->orig_family_num, in which
	 * case fall back to ->family_num.  orig_family_num will be
	 * fixed up with the first metadata update.
	 */
	family_num = super->anchor->orig_family_num;
	if (family_num == 0)
		family_num = super->anchor->family_num;
	sha1_begin(&ctx);
	sha1_hash(super->anchor->sig, MPB_SIG_LEN, &ctx);
	sha1_hash(&family_num, sizeof(__u32), &ctx);
	if (super->current_vol >= 0)
		dev = get_imsm_dev(super, super->current_vol);
	if (dev) {
		__u32 vol = super->current_vol;
		sha1_hash(&vol, sizeof(vol), &ctx);
		sha1_hash(dev->volume, MAX_RAID_SERIAL_LEN, &ctx);
	}
	sha1_end(buf, &ctx);
	memcpy(uuid, buf, 4*4);
}

static __u32 migr_strip_blocks_resync(struct imsm_dev *dev)
{
	/* migr_strip_size when repairing or initializing parity */
	struct imsm_map *map = get_imsm_map(dev, MAP_0);
	__u32 chunk = __le32_to_cpu(map->blocks_per_strip);

	switch (get_imsm_raid_level(map)) {
	case 5:
	case 10:
		return chunk;
	default:
		return 128*1024 >> 9;
	}
}

static __u32 migr_strip_blocks_rebuild(struct imsm_dev *dev)
{
	/* migr_strip_size when rebuilding a degraded disk, no idea why
	 * this is different than migr_strip_size_resync(), but it's good
	 * to be compatible
	 */
	struct imsm_map *map = get_imsm_map(dev, MAP_1);
	__u32 chunk = __le32_to_cpu(map->blocks_per_strip);

	switch (get_imsm_raid_level(map)) {
	case 1:
	case 10:
		if (map->num_members % map->num_domains == 0)
			return 128*1024 >> 9;
		else
			return chunk;
	case 5:
		return max((__u32) 64*1024 >> 9, chunk);
	default:
		return 128*1024 >> 9;
	}
}

static __u32 num_stripes_per_unit_resync(struct imsm_dev *dev)
{
	struct imsm_map *lo = get_imsm_map(dev, MAP_0);
	struct imsm_map *hi = get_imsm_map(dev, MAP_1);
	__u32 lo_chunk = __le32_to_cpu(lo->blocks_per_strip);
	__u32 hi_chunk = __le32_to_cpu(hi->blocks_per_strip);

	return max((__u32) 1, hi_chunk / lo_chunk);
}

static __u32 num_stripes_per_unit_rebuild(struct imsm_dev *dev)
{
	struct imsm_map *lo = get_imsm_map(dev, MAP_0);
	int level = get_imsm_raid_level(lo);

	if (level == 1 || level == 10) {
		struct imsm_map *hi = get_imsm_map(dev, MAP_1);

		return hi->num_domains;
	} else
		return num_stripes_per_unit_resync(dev);
}

static __u8 imsm_num_data_members(struct imsm_dev *dev, int second_map)
{
	/* named 'imsm_' because raid0, raid1 and raid10
	 * counter-intuitively have the same number of data disks
	 */
	struct imsm_map *map = get_imsm_map(dev, second_map);

	switch (get_imsm_raid_level(map)) {
	case 0:
		return map->num_members;
		break;
	case 1:
	case 10:
		return map->num_members/2;
	case 5:
		return map->num_members - 1;
	default:
		dprintf("%s: unsupported raid level\n", __func__);
		return 0;
	}
}

static __u32 parity_segment_depth(struct imsm_dev *dev)
{
	struct imsm_map *map = get_imsm_map(dev, MAP_0);
	__u32 chunk =  __le32_to_cpu(map->blocks_per_strip);

	switch(get_imsm_raid_level(map)) {
	case 1:
	case 10:
		return chunk * map->num_domains;
	case 5:
		return chunk * map->num_members;
	default:
		return chunk;
	}
}

static __u32 map_migr_block(struct imsm_dev *dev, __u32 block)
{
	struct imsm_map *map = get_imsm_map(dev, MAP_1);
	__u32 chunk = __le32_to_cpu(map->blocks_per_strip);
	__u32 strip = block / chunk;

	switch (get_imsm_raid_level(map)) {
	case 1:
	case 10: {
		__u32 vol_strip = (strip * map->num_domains) + 1;
		__u32 vol_stripe = vol_strip / map->num_members;

		return vol_stripe * chunk + block % chunk;
	} case 5: {
		__u32 stripe = strip / (map->num_members - 1);

		return stripe * chunk + block % chunk;
	}
	default:
		return 0;
	}
}

static __u64 blocks_per_migr_unit(struct intel_super *super,
				  struct imsm_dev *dev)
{
	/* calculate the conversion factor between per member 'blocks'
	 * (md/{resync,rebuild}_start) and imsm migration units, return
	 * 0 for the 'not migrating' and 'unsupported migration' cases
	 */
	if (!dev->vol.migr_state)
		return 0;

	switch (migr_type(dev)) {
	case MIGR_GEN_MIGR: {
		struct migr_record *migr_rec = super->migr_rec;
		return __le32_to_cpu(migr_rec->blocks_per_unit);
	}
	case MIGR_VERIFY:
	case MIGR_REPAIR:
	case MIGR_INIT: {
		struct imsm_map *map = get_imsm_map(dev, MAP_0);
		__u32 stripes_per_unit;
		__u32 blocks_per_unit;
		__u32 parity_depth;
		__u32 migr_chunk;
		__u32 block_map;
		__u32 block_rel;
		__u32 segment;
		__u32 stripe;
		__u8  disks;

		/* yes, this is really the translation of migr_units to
		 * per-member blocks in the 'resync' case
		 */
		stripes_per_unit = num_stripes_per_unit_resync(dev);
		migr_chunk = migr_strip_blocks_resync(dev);
		disks = imsm_num_data_members(dev, MAP_0);
		blocks_per_unit = stripes_per_unit * migr_chunk * disks;
		stripe = __le16_to_cpu(map->blocks_per_strip) * disks;
		segment = blocks_per_unit / stripe;
		block_rel = blocks_per_unit - segment * stripe;
		parity_depth = parity_segment_depth(dev);
		block_map = map_migr_block(dev, block_rel);
		return block_map + parity_depth * segment;
	}
	case MIGR_REBUILD: {
		__u32 stripes_per_unit;
		__u32 migr_chunk;

		stripes_per_unit = num_stripes_per_unit_rebuild(dev);
		migr_chunk = migr_strip_blocks_rebuild(dev);
		return migr_chunk * stripes_per_unit;
	}
	case MIGR_STATE_CHANGE:
	default:
		return 0;
	}
}

static int imsm_level_to_layout(int level)
{
	switch (level) {
	case 0:
	case 1:
		return 0;
	case 5:
	case 6:
		return ALGORITHM_LEFT_ASYMMETRIC;
	case 10:
		return 0x102;
	}
	return UnSet;
}

/*******************************************************************************
 * Function:    read_imsm_migr_rec
 * Description: Function reads imsm migration record from last sector of disk
 * Parameters:
 *      fd      : disk descriptor
 *      super   : metadata info
 * Returns:
 *       0 : success,
 *      -1 : fail
 ******************************************************************************/
static int read_imsm_migr_rec(int fd, struct intel_super *super)
{
	int ret_val = -1;
	unsigned long long dsize;

	get_dev_size(fd, NULL, &dsize);
	if (lseek(fd, dsize - MIGR_REC_POSITION, SEEK_SET) < 0) {
		bb_perror_msg("Cannot seek to anchor block");
		goto out;
	}
	if (read(fd, super->migr_rec_buf, MIGR_REC_BUF_SIZE) !=
							    MIGR_REC_BUF_SIZE) {
		bb_perror_msg("Cannot read migr record block");
		goto out;
	}
	ret_val = 0;

out:
	return ret_val;
}

static struct imsm_dev *imsm_get_device_during_migration(
	struct intel_super *super)
{

	struct intel_dev *dv;

	for (dv = super->devlist; dv; dv = dv->next) {
		if (is_gen_migration(dv->dev))
			return dv->dev;
	}
	return NULL;
}

/*******************************************************************************
 * Function:    load_imsm_migr_rec
 * Description: Function reads imsm migration record (it is stored at the last
 *              sector of disk)
 * Parameters:
 *      super   : imsm internal array info
 *      info    : general array info
 * Returns:
 *       0 : success
 *      -1 : fail
 *      -2 : no migration in progress
 ******************************************************************************/
static int load_imsm_migr_rec(struct intel_super *super, struct mdinfo *info)
{
	struct mdinfo *sd;
	struct dl *dl = NULL;
	char nm[30];
	int retval = -1;
	int fd = -1;
	struct imsm_dev *dev;
	struct imsm_map *map = NULL;
	int slot = -1;

	/* find map under migration */
	dev = imsm_get_device_during_migration(super);
	/* nothing to load,no migration in progress?
	*/
	if (dev == NULL)
		return -2;
	map = get_imsm_map(dev, MAP_0);

	if (info) {
		for (sd = info->devs ; sd ; sd = sd->next) {
			/* skip spare and failed disks
			 */
			if (sd->disk.raid_disk < 0)
				continue;
			/* read only from one of the first two slots */
			if (map)
				slot = get_imsm_disk_slot(map,
							  sd->disk.raid_disk);
			if ((map == NULL) || (slot > 1) || (slot < 0))
				continue;

			sprintf(nm, "%d:%d", sd->disk.major, sd->disk.minor);
			fd = dev_open(nm, O_RDONLY);
			if (fd >= 0)
				break;
		}
	}
	if (fd < 0) {
		for (dl = super->disks; dl; dl = dl->next) {
			/* skip spare and failed disks
			*/
			if (dl->index < 0)
				continue;
			/* read only from one of the first two slots */
			if (map)
				slot = get_imsm_disk_slot(map, dl->index);
			if ((map == NULL) || (slot > 1) || (slot < 0))
				continue;
			sprintf(nm, "%d:%d", dl->major, dl->minor);
			fd = dev_open(nm, O_RDONLY);
			if (fd >= 0)
				break;
		}
	}
	if (fd < 0)
		goto out;
	retval = read_imsm_migr_rec(fd, super);

out:
	if (fd >= 0)
		close(fd);
	return retval;
}

/* spare/missing disks activations are not allowe when
 * array/container performs reshape operation, because
 * all arrays in container works on the same disks set
 */
static int imsm_reshape_blocks_arrays_changes(struct intel_super *super)
{
	int rv = 0;
	struct intel_dev *i_dev;
	struct imsm_dev *dev;

	/* check whole container
	 */
	for (i_dev = super->devlist; i_dev; i_dev = i_dev->next) {
		dev = i_dev->dev;
		if (is_gen_migration(dev)) {
			/* No repair during any migration in container
			 */
			rv = 1;
			break;
		}
	}
	return rv;
}
static unsigned long long imsm_component_size_aligment_check(int level,
					      int chunk_size,
					      unsigned long long component_size)
{
	unsigned int component_size_alligment;

	/* check component size aligment
	*/
	component_size_alligment = component_size % (chunk_size/512);

	dprintf("imsm_component_size_aligment_check(Level: %i, "
		"chunk_size = %i, component_size = %llu), "
		"component_size_alligment = %u\n",
		level, chunk_size, component_size,
		component_size_alligment);

	if (component_size_alligment && (level != 1) && (level != UnSet)) {
		dprintf("imsm: reported component size alligned from %llu ",
			component_size);
		component_size -= component_size_alligment;
		dprintf("to %llu (%i).\n",
			component_size, component_size_alligment);
	}

	return component_size;
}

static void getinfo_super_imsm_volume(struct supertype *st, struct mdinfo *info, char *dmap)
{
	struct intel_super *super = st->sb;
	struct migr_record *migr_rec = super->migr_rec;
	struct imsm_dev *dev = get_imsm_dev(super, super->current_vol);
	struct imsm_map *map = get_imsm_map(dev, MAP_0);
	struct imsm_map *prev_map = get_imsm_map(dev, MAP_1);
	struct imsm_map *map_to_analyse = map;
	struct dl *dl;
	char *devname;
	int map_disks = info->array.raid_disks;

	memset(info, 0, sizeof(*info));
	if (prev_map)
		map_to_analyse = prev_map;

	dl = super->current_disk;

	info->container_member    = super->current_vol;
	info->array.raid_disks    = map->num_members;
	info->array.level         = get_imsm_raid_level(map_to_analyse);
	info->array.layout        = imsm_level_to_layout(info->array.level);
	info->array.md_minor      = -1;
	info->array.ctime         = 0;
	info->array.utime         = 0;
	info->array.chunk_size    =
		__le16_to_cpu(map_to_analyse->blocks_per_strip) << 9;
	info->array.state         = !dev->vol.dirty;
	info->custom_array_size   = __le32_to_cpu(dev->size_high);
	info->custom_array_size   <<= 32;
	info->custom_array_size   |= __le32_to_cpu(dev->size_low);
	info->recovery_blocked = imsm_reshape_blocks_arrays_changes(st->sb);

	if (is_gen_migration(dev)) {
		info->reshape_active = 1;
		info->new_level = get_imsm_raid_level(map);
		info->new_layout = imsm_level_to_layout(info->new_level);
		info->new_chunk = __le16_to_cpu(map->blocks_per_strip) << 9;
		info->delta_disks = map->num_members - prev_map->num_members;
		if (info->delta_disks) {
			/* this needs to be applied to every array
			 * in the container.
			 */
			info->reshape_active = CONTAINER_RESHAPE;
		}
		/* We shape information that we give to md might have to be
		 * modify to cope with md's requirement for reshaping arrays.
		 * For example, when reshaping a RAID0, md requires it to be
		 * presented as a degraded RAID4.
		 * Also if a RAID0 is migrating to a RAID5 we need to specify
		 * the array as already being RAID5, but the 'before' layout
		 * is a RAID4-like layout.
		 */
		switch (info->array.level) {
		case 0:
			switch(info->new_level) {
			case 0:
				/* conversion is happening as RAID4 */
				info->array.level = 4;
				info->array.raid_disks += 1;
				break;
			case 5:
				/* conversion is happening as RAID5 */
				info->array.level = 5;
				info->array.layout = ALGORITHM_PARITY_N;
				info->delta_disks -= 1;
				break;
			default:
				/* FIXME error message */
				info->array.level = UnSet;
				break;
			}
			break;
		}
	} else {
		info->new_level = UnSet;
		info->new_layout = UnSet;
		info->new_chunk = info->array.chunk_size;
		info->delta_disks = 0;
	}

	if (dl) {
		info->disk.major = dl->major;
		info->disk.minor = dl->minor;
		info->disk.number = dl->index;
		info->disk.raid_disk = get_imsm_disk_slot(map_to_analyse,
							  dl->index);
	}

	info->data_offset         = pba_of_lba0(map_to_analyse);
	info->component_size      = blocks_per_member(map_to_analyse);

	info->component_size = imsm_component_size_aligment_check(
							info->array.level,
							info->array.chunk_size,
							info->component_size);

	memset(info->uuid, 0, sizeof(info->uuid));
	info->recovery_start = MaxSector;

	info->reshape_progress = 0;
	info->resync_start = MaxSector;
	if ((map_to_analyse->map_state == IMSM_T_STATE_UNINITIALIZED ||
	    dev->vol.dirty) &&
	    imsm_reshape_blocks_arrays_changes(super) == 0) {
		info->resync_start = 0;
	}
	if (dev->vol.migr_state) {
		switch (migr_type(dev)) {
		case MIGR_REPAIR:
		case MIGR_INIT: {
			__u64 blocks_per_unit = blocks_per_migr_unit(super,
								     dev);
			__u64 units = __le32_to_cpu(dev->vol.curr_migr_unit);

			info->resync_start = blocks_per_unit * units;
			break;
		}
		case MIGR_GEN_MIGR: {
			__u64 blocks_per_unit = blocks_per_migr_unit(super,
								     dev);
			__u64 units = __le32_to_cpu(migr_rec->curr_migr_unit);
			unsigned long long array_blocks;
			int used_disks;

			if (__le32_to_cpu(migr_rec->ascending_migr) &&
			    (units <
				(__le32_to_cpu(migr_rec->num_migr_units)-1)) &&
			    (super->migr_rec->rec_status ==
					__cpu_to_le32(UNIT_SRC_IN_CP_AREA)))
				units++;

			info->reshape_progress = blocks_per_unit * units;

			dprintf("IMSM: General Migration checkpoint : %llu "
			       "(%llu) -> read reshape progress : %llu\n",
				(unsigned long long)units,
				(unsigned long long)blocks_per_unit,
				info->reshape_progress);

			used_disks = imsm_num_data_members(dev, MAP_1);
			if (used_disks > 0) {
				array_blocks = blocks_per_member(map) *
					used_disks;
				/* round array size down to closest MB
				 */
				info->custom_array_size = (array_blocks
						>> SECT_PER_MB_SHIFT)
						<< SECT_PER_MB_SHIFT;
			}
		}
		case MIGR_VERIFY:
			/* we could emulate the checkpointing of
			 * 'sync_action=check' migrations, but for now
			 * we just immediately complete them
			 */
		case MIGR_REBUILD:
			/* this is handled by container_content_imsm() */
		case MIGR_STATE_CHANGE:
			/* FIXME handle other migrations */
		default:
			/* we are not dirty, so... */
			info->resync_start = MaxSector;
		}
	}

	strncpy(info->name, (char *) dev->volume, MAX_RAID_SERIAL_LEN);
	info->name[MAX_RAID_SERIAL_LEN] = 0;

	info->array.major_version = -1;
	info->array.minor_version = -2;
	devname = devnum2devname(st->container_dev);
	*info->text_version = '\0';
	if (devname)
		sprintf(info->text_version, "/%s/%d", devname, info->container_member);
	free(devname);
	info->safe_mode_delay = 4000;  /* 4 secs like the Matrix driver */
	uuid_from_super_imsm(st, info->uuid);

	if (dmap) {
		int i, j;
		for (i=0; i<map_disks; i++) {
			dmap[i] = 0;
			if (i < info->array.raid_disks) {
				struct imsm_disk *dsk;
				j = get_imsm_disk_idx(dev, i, MAP_X);
				dsk = get_imsm_disk(super, j);
				if (dsk && (dsk->status & CONFIGURED_DISK))
					dmap[i] = 1;
			}
		}
	}
}

static __u8 imsm_check_degraded(struct intel_super *super, struct imsm_dev *dev,
				int failed, int look_in_map);

static int imsm_count_failed(struct intel_super *super, struct imsm_dev *dev,
			     int look_in_map);


static struct imsm_disk *get_imsm_missing(struct intel_super *super, __u8 idx)
{
	struct dl *d;

	for (d = super->missing; d; d = d->next)
		if (d->index == idx)
			return &d->disk;
	return NULL;
}

static void getinfo_super_imsm(struct supertype *st, struct mdinfo *info, char *map)
{
	struct intel_super *super = st->sb;
	struct imsm_disk *disk;
	int map_disks = info->array.raid_disks;
	int max_enough = -1;
	int i;
	struct imsm_super *mpb;

	if (super->current_vol >= 0) {
		getinfo_super_imsm_volume(st, info, map);
		return;
	}
	memset(info, 0, sizeof(*info));

	/* Set raid_disks to zero so that Assemble will always pull in valid
	 * spares
	 */
	info->array.raid_disks    = 0;
	info->array.level         = LEVEL_CONTAINER;
	info->array.layout        = 0;
	info->array.md_minor      = -1;
	info->array.ctime         = 0; /* N/A for imsm */
	info->array.utime         = 0;
	info->array.chunk_size    = 0;

	info->disk.major = 0;
	info->disk.minor = 0;
	info->disk.raid_disk = -1;
	info->reshape_active = 0;
	info->array.major_version = -1;
	info->array.minor_version = -2;
	strcpy(info->text_version, "imsm");
	info->safe_mode_delay = 0;
	info->disk.number = -1;
	info->disk.state = 0;
	info->name[0] = 0;
	info->recovery_start = MaxSector;
	info->recovery_blocked = imsm_reshape_blocks_arrays_changes(st->sb);

	/* do we have the all the insync disks that we expect? */
	mpb = super->anchor;

	for (i = 0; i < mpb->num_raid_devs; i++) {
		struct imsm_dev *dev = get_imsm_dev(super, i);
		int failed, cur_enough, j, missing = 0;
		struct imsm_map *map_from_imsm;
		__u8 state;

		failed = imsm_count_failed(super, dev, MAP_0);
		state = imsm_check_degraded(super, dev, failed, MAP_0);
		map_from_imsm = get_imsm_map(dev, MAP_0);

		/* any newly missing disks?
		 * (catches single-degraded vs double-degraded)
		 */
		for (j = 0; j < map_from_imsm->num_members; j++) {
			__u32 ord = get_imsm_ord_tbl_ent(dev, j, MAP_0);
			__u32 idx = ord_to_idx(ord);

			if (!(ord & IMSM_ORD_REBUILD) &&
			    get_imsm_missing(super, idx)) {
				missing = 1;
				break;
			}
		}

		if (state == IMSM_T_STATE_FAILED)
			cur_enough = -1;
		else if (state == IMSM_T_STATE_DEGRADED &&
			 (state != map_from_imsm->map_state || missing))
			cur_enough = 0;
		else /* we're normal, or already degraded */
			cur_enough = 1;
		if (is_gen_migration(dev) && missing) {
			/* during general migration we need all disks
			 * that process is running on.
			 * No new missing disk is allowed.
			 */
			max_enough = -1;
			cur_enough = -1;
			/* no more checks necessary
			 */
			break;
		}
		/* in the missing/failed disk case check to see
		 * if at least one array is runnable
		 */
		max_enough = max(max_enough, cur_enough);
	}
	dprintf("%s: enough: %d\n", __func__, max_enough);
	info->container_enough = max_enough;

	if (super->disks) {
		__u32 reserved = imsm_reserved_sectors(super, super->disks);

		disk = &super->disks->disk;
		info->data_offset = total_blocks(&super->disks->disk) - reserved;
		info->component_size = reserved;
		info->disk.state  = is_configured(disk) ? (1 << MD_DISK_ACTIVE) : 0;
		/* we don't change info->disk.raid_disk here because
		 * this state will be finalized in mdmon after we have
		 * found the 'most fresh' version of the metadata
		 */
		info->disk.state |= is_failed(disk) ? (1 << MD_DISK_FAULTY) : 0;
		info->disk.state |= is_spare(disk) ? 0 : (1 << MD_DISK_SYNC);
	}

	/* only call uuid_from_super_imsm when this disk is part of a populated container,
	 * ->compare_super may have updated the 'num_raid_devs' field for spares
	 */
	if (info->disk.state & (1 << MD_DISK_SYNC) || super->anchor->num_raid_devs)
		uuid_from_super_imsm(st, info->uuid);
	else
		memcpy(info->uuid, uuid_zero, sizeof(uuid_zero));

	/* I don't know how to compute 'map' on imsm, so use safe default */
	if (map) {
		for (i = 0; i < map_disks; i++)
			map[i] = 1;
	}

}

/* allocates memory and fills disk in mdinfo structure
 * for each disk in array */
static struct mdinfo *getinfo_super_disks_imsm(struct supertype *st)
{
	struct mdinfo *mddev = NULL;
	struct intel_super *super = st->sb;
	struct imsm_disk *disk;
	int count = 0;
	struct dl *dl;
	if (!super || !super->disks)
		return NULL;
	dl = super->disks;
	mddev = zmalloc(sizeof(*mddev));
	while (dl) {
		struct mdinfo *tmp;
		disk = &dl->disk;
		tmp = zmalloc(sizeof(*tmp));
		if (mddev->devs)
			tmp->next = mddev->devs;
		mddev->devs = tmp;
		tmp->disk.number = count++;
		tmp->disk.major = dl->major;
		tmp->disk.minor = dl->minor;
		tmp->disk.state = is_configured(disk) ?
				  (1 << MD_DISK_ACTIVE) : 0;
		tmp->disk.state |= is_failed(disk) ? (1 << MD_DISK_FAULTY) : 0;
		tmp->disk.state |= is_spare(disk) ? 0 : (1 << MD_DISK_SYNC);
		tmp->disk.raid_disk = -1;
		dl = dl->next;
	}
	return mddev;
}

static int update_super_imsm(struct supertype *st, struct mdinfo *info,
			     char *update, char *devname, int verbose,
			     int uuid_set, char *homehost)
{
	/* For 'assemble' and 'force' we need to return non-zero if any
	 * change was made.  For others, the return value is ignored.
	 * Update options are:
	 *  force-one : This device looks a bit old but needs to be included,
	 *        update age info appropriately.
	 *  assemble: clear any 'faulty' flag to allow this device to
	 *              be assembled.
	 *  force-array: Array is degraded but being forced, mark it clean
	 *         if that will be needed to assemble it.
	 *
	 *  newdev:  not used ????
	 *  grow:  Array has gained a new device - this is currently for
	 *              linear only
	 *  resync: mark as dirty so a resync will happen.
	 *  name:  update the name - preserving the homehost
	 *  uuid:  Change the uuid of the array to match watch is given
	 *
	 * Following are not relevant for this imsm:
	 *  sparc2.2 : update from old dodgey metadata
	 *  super-minor: change the preferred_minor number
	 *  summaries:  update redundant counters.
	 *  homehost:  update the recorded homehost
	 *  _reshape_progress: record new reshape_progress position.
	 */
	int rv = 1;
	struct intel_super *super = st->sb;
	struct imsm_super *mpb;

	(void)&devname;
	(void)&verbose;
	(void)&homehost;

	/* we can only update container info */
	if (!super || super->current_vol >= 0 || !super->anchor)
		return 1;

	mpb = super->anchor;

	if (strcmp(update, "uuid") == 0) {
		/* We take this to mean that the family_num should be updated.
		 * However that is much smaller than the uuid so we cannot really
		 * allow an explicit uuid to be given.  And it is hard to reliably
		 * know if one was.
		 * So if !uuid_set we know the current uuid is random and just used
		 * the first 'int' and copy it to the other 3 positions.
		 * Otherwise we require the 4 'int's to be the same as would be the
		 * case if we are using a random uuid.  So an explicit uuid will be
		 * accepted as long as all for ints are the same... which shouldn't hurt
		 */
		if (!uuid_set) {
			info->uuid[1] = info->uuid[2] = info->uuid[3] = info->uuid[0];
			rv = 0;
		} else {
			if (info->uuid[0] != info->uuid[1] ||
			    info->uuid[1] != info->uuid[2] ||
			    info->uuid[2] != info->uuid[3])
				rv = -1;
			else
				rv = 0;
		}
		if (rv == 0)
			mpb->orig_family_num = info->uuid[0];
	} else if (strcmp(update, "assemble") == 0)
		rv = 0;
	else
		rv = -1;

	/* successful update? recompute checksum */
	if (rv == 0)
		mpb->check_sum = __le32_to_cpu(__gen_imsm_checksum(mpb));

	return rv;
}

static void free_devlist(struct intel_super *super)
{
	struct intel_dev *dv;

	while (super->devlist) {
		dv = super->devlist->next;
		free(super->devlist->dev);
		free(super->devlist);
		super->devlist = dv;
	}
}

static void imsm_copy_dev(struct imsm_dev *dest, struct imsm_dev *src)
{
	memcpy(dest, src, sizeof_imsm_dev(src, 0));
}

static int compare_super_imsm(struct supertype *st, struct supertype *tst)
{
	/*
	 * return:
	 *  0 same, or first was empty, and second was copied
	 *  1 second had wrong number
	 *  2 wrong uuid
	 *  3 wrong other info
	 */
	struct intel_super *first = st->sb;
	struct intel_super *sec = tst->sb;

	if (!first) {
		st->sb = tst->sb;
		tst->sb = NULL;
		return 0;
	}
	/* in platform dependent environment test if the disks
	 * use the same Intel hba
	 */
	if (!check_env("IMSM_NO_PLATFORM")) {
		if (!first->hba || !sec->hba ||
		    (first->hba->type != sec->hba->type))  {
			bb_error_msg("HBAs of devices does not match %s != %s",
				first->hba ? get_sys_dev_type(first->hba->type) : NULL,
				sec->hba ? get_sys_dev_type(sec->hba->type) : NULL);
			return 3;
		}
	}

	/* if an anchor does not have num_raid_devs set then it is a free
	 * floating spare
	 */
	if (first->anchor->num_raid_devs > 0 &&
	    sec->anchor->num_raid_devs > 0) {
		/* Determine if these disks might ever have been
		 * related.  Further disambiguation can only take place
		 * in load_super_imsm_all
		 */
		__u32 first_family = first->anchor->orig_family_num;
		__u32 sec_family = sec->anchor->orig_family_num;

		if (memcmp(first->anchor->sig, sec->anchor->sig,
			   MAX_SIGNATURE_LENGTH) != 0)
			return 3;

		if (first_family == 0)
			first_family = first->anchor->family_num;
		if (sec_family == 0)
			sec_family = sec->anchor->family_num;

		if (first_family != sec_family)
			return 3;

	}


	/* if 'first' is a spare promote it to a populated mpb with sec's
	 * family number
	 */
	if (first->anchor->num_raid_devs == 0 &&
	    sec->anchor->num_raid_devs > 0) {
		int i;
		struct intel_dev *dv;
		struct imsm_dev *dev;

		/* we need to copy raid device info from sec if an allocation
		 * fails here we don't associate the spare
		 */
		for (i = 0; i < sec->anchor->num_raid_devs; i++) {
			dv = xmalloc(sizeof(*dv));
			dev = xmalloc(sizeof_imsm_dev(get_imsm_dev(sec, i), 1));
			dv->dev = dev;
			dv->index = i;
			dv->next = first->devlist;
			first->devlist = dv;
		}
		if (i < sec->anchor->num_raid_devs) {
			/* allocation failure */
			free_devlist(first);
			bb_error_msg("imsm: failed to associate spare");
			return 3;
		}
		first->anchor->num_raid_devs = sec->anchor->num_raid_devs;
		first->anchor->orig_family_num = sec->anchor->orig_family_num;
		first->anchor->family_num = sec->anchor->family_num;
		memcpy(first->anchor->sig, sec->anchor->sig, MAX_SIGNATURE_LENGTH);
		for (i = 0; i < sec->anchor->num_raid_devs; i++)
			imsm_copy_dev(get_imsm_dev(first, i), get_imsm_dev(sec, i));
	}

	return 0;
}

static void fd2devname(int fd, char *name)
{
	struct stat st;
	char path[256];
	char dname[PATH_MAX];
	const char *nm;
	int rv;

	name[0] = '\0';
	if (fstat(fd, &st) != 0)
		return;
	sprintf(path, "/sys/dev/block/%d:%d",
		major(st.st_rdev), minor(st.st_rdev));

	rv = readlink(path, dname, sizeof(dname)-1);
	if (rv <= 0)
		return;

	dname[rv] = '\0';
	if ((nm = bb_basename(dname)) != dname)
		snprintf(name, MAX_RAID_SERIAL_LEN, "/dev/%s", nm);
}

static int scsi_get_serial(int fd, void *buf, size_t buf_len);

static int imsm_read_serial(int fd, char *devname,
			    __u8 serial[MAX_RAID_SERIAL_LEN])
{
	unsigned char scsi_serial[255];
	int rv;
	int rsp_len;
	int len;
	char *dest;
	char *src;
	char *rsp_buf;
	int i;

	memset(scsi_serial, 0, sizeof(scsi_serial));

	rv = scsi_get_serial(fd, scsi_serial, sizeof(scsi_serial));

	if (rv && check_env("IMSM_DEVNAME_AS_SERIAL")) {
		memset(serial, 0, MAX_RAID_SERIAL_LEN);
		fd2devname(fd, (char *) serial);
		return 0;
	}

	if (rv != 0) {
		if (devname)
			bb_error_msg("Failed to retrieve serial for %s", devname);
		return rv;
	}

	rsp_len = scsi_serial[3];
	if (!rsp_len) {
		if (devname)
			bb_error_msg("Failed to retrieve serial for %s", devname);
		return 2;
	}
	rsp_buf = (char *) &scsi_serial[4];

	/* trim all whitespace and non-printable characters and convert
	 * ':' to ';'
	 */
	for (i = 0, dest = rsp_buf; i < rsp_len; i++) {
		src = &rsp_buf[i];
		if (*src > 0x20) {
			/* ':' is reserved for use in placeholder serial
			 * numbers for missing disks
			 */
			if (*src == ':')
				*dest++ = ';';
			else
				*dest++ = *src;
		}
	}
	len = dest - rsp_buf;
	dest = rsp_buf;

	/* truncate leading characters */
	if (len > MAX_RAID_SERIAL_LEN) {
		dest += len - MAX_RAID_SERIAL_LEN;
		len = MAX_RAID_SERIAL_LEN;
	}

	memset(serial, 0, MAX_RAID_SERIAL_LEN);
	memcpy(serial, dest, len);

	return 0;
}

static int serialcmp(__u8 *s1, __u8 *s2)
{
	return strncmp((char *) s1, (char *) s2, MAX_RAID_SERIAL_LEN);
}

static void serialcpy(__u8 *dest, __u8 *src)
{
	strncpy((char *) dest, (char *) src, MAX_RAID_SERIAL_LEN);
}

static struct imsm_disk *
__serial_to_disk(__u8 *serial, struct imsm_super *mpb, int *idx)
{
	int i;

	for (i = 0; i < mpb->num_disks; i++) {
		struct imsm_disk *disk = __get_imsm_disk(mpb, i);

		if (serialcmp(disk->serial, serial) == 0) {
			if (idx)
				*idx = i;
			return disk;
		}
	}

	return NULL;
}

static int
load_imsm_disk(int fd, struct intel_super *super, char *devname, int keep_fd)
{
	struct imsm_disk *disk;
	struct dl *dl;
	struct stat stb;
	int rv;
	char name[40];
	__u8 serial[MAX_RAID_SERIAL_LEN];

	rv = imsm_read_serial(fd, devname, serial);

	if (rv != 0)
		return 2;

	dl = calloc(1, sizeof(*dl));
	if (!dl) {
		if (devname)
			bb_error_msg("failed to allocate disk buffer for %s", devname);
		return 2;
	}

	fstat(fd, &stb);
	dl->major = major(stb.st_rdev);
	dl->minor = minor(stb.st_rdev);
	dl->next = super->disks;
	dl->fd = keep_fd ? fd : -1;
	assert(super->disks == NULL);
	super->disks = dl;
	serialcpy(dl->serial, serial);
	dl->index = -2;
	dl->e = NULL;
	fd2devname(fd, name);
	if (devname)
		dl->devname = strdup(devname);
	else
		dl->devname = strdup(name);

	/* look up this disk's index in the current anchor */
	disk = __serial_to_disk(dl->serial, super->anchor, &dl->index);
	if (disk) {
		dl->disk = *disk;
		/* only set index on disks that are a member of a
		 * populated contianer, i.e. one with raid_devs
		 */
		if (is_failed(&dl->disk))
			dl->index = -2;
		else if (is_spare(&dl->disk))
			dl->index = -1;
	}

	return 0;
}

static int parse_raid_devices(struct intel_super *super)
{
	int i;
	struct imsm_dev *dev_new;
	size_t len, len_migr;
	size_t max_len = 0;
	size_t space_needed = 0;
	struct imsm_super *mpb = super->anchor;

	for (i = 0; i < super->anchor->num_raid_devs; i++) {
		struct imsm_dev *dev_iter = __get_imsm_dev(super->anchor, i);
		struct intel_dev *dv;

		len = sizeof_imsm_dev(dev_iter, 0);
		len_migr = sizeof_imsm_dev(dev_iter, 1);
		if (len_migr > len)
			space_needed += len_migr - len;

		dv = xmalloc(sizeof(*dv));
		if (max_len < len_migr)
			max_len = len_migr;
		if (max_len > len_migr)
			space_needed += max_len - len_migr;
		dev_new = xmalloc(max_len);
		imsm_copy_dev(dev_new, dev_iter);
		dv->dev = dev_new;
		dv->index = i;
		dv->next = super->devlist;
		super->devlist = dv;
	}

	/* ensure that super->buf is large enough when all raid devices
	 * are migrating
	 */
	if (__le32_to_cpu(mpb->mpb_size) + space_needed > super->len) {
		void *buf;

		len = ROUND_UP(__le32_to_cpu(mpb->mpb_size) + space_needed, 512);
		if (posix_memalign(&buf, 512, len) != 0)
			return 1;

		memcpy(buf, super->buf, super->len);
		memset(buf + super->len, 0, len - super->len);
		free(super->buf);
		super->buf = buf;
		super->len = len;
	}

	return 0;
}

/* retrieve a pointer to the bbm log which starts after all raid devices */
static struct bbm_log *__get_imsm_bbm_log(struct imsm_super *mpb)
{
	void *ptr = NULL;

	if (__le32_to_cpu(mpb->bbm_log_size)) {
		ptr = mpb;
		ptr += mpb->mpb_size - __le32_to_cpu(mpb->bbm_log_size);
	}

	return ptr;
}

/*******************************************************************************
 * Function:    check_mpb_migr_compatibility
 * Description: Function checks for unsupported migration features:
 *              - migration optimization area (pba_of_lba0)
 *              - descending reshape (ascending_migr)
 * Parameters:
 *      super   : imsm metadata information
 * Returns:
 *       0 : migration is compatible
 *      -1 : migration is not compatible
 ******************************************************************************/
static int check_mpb_migr_compatibility(struct intel_super *super)
{
	struct imsm_map *map0, *map1;
	struct migr_record *migr_rec = super->migr_rec;
	int i;

	for (i = 0; i < super->anchor->num_raid_devs; i++) {
		struct imsm_dev *dev_iter = __get_imsm_dev(super->anchor, i);

		if (dev_iter &&
		    dev_iter->vol.migr_state == 1 &&
		    dev_iter->vol.migr_type == MIGR_GEN_MIGR) {
			/* This device is migrating */
			map0 = get_imsm_map(dev_iter, MAP_0);
			map1 = get_imsm_map(dev_iter, MAP_1);
			if (pba_of_lba0(map0) != pba_of_lba0(map1))
				/* migration optimization area was used */
				return -1;
			if (migr_rec->ascending_migr == 0
				&& migr_rec->dest_depth_per_unit > 0)
				/* descending reshape not supported yet */
				return -1;
		}
	}
	return 0;
}

static void __free_imsm(struct intel_super *super, int free_disks);

/* load_imsm_mpb - read matrix metadata
 * allocates super->mpb to be freed by free_imsm
 */
static int load_imsm_mpb(int fd, struct intel_super *super, char *devname)
{
	unsigned long long dsize;
	unsigned long long sectors;
	struct stat;
	struct imsm_super *anchor;
	__u32 check_sum;

	get_dev_size(fd, NULL, &dsize);
	if (dsize < 1024) {
		if (devname)
			bb_error_msg("%s: device to small for imsm", devname);
		return 1;
	}

	if (lseek(fd, dsize - (512 * 2), SEEK_SET) < 0) {
		if (devname)
			bb_perror_msg("Cannot seek to anchor block on %s", devname);
		return 1;
	}

	if (posix_memalign((void**)&anchor, 512, 512) != 0) {
		if (devname)
			bb_error_msg("Failed to allocate imsm anchor buffer"
				" on %s", devname);
		return 1;
	}
	if (read(fd, anchor, 512) != 512) {
		if (devname)
			bb_perror_msg("Cannot read anchor block on %s", devname);
		free(anchor);
		return 1;
	}

	if (strncmp((char *) anchor->sig, MPB_SIGNATURE, MPB_SIG_LEN) != 0) {
		if (devname)
			bb_error_msg("no IMSM anchor on %s", devname);
		free(anchor);
		return 2;
	}

	__free_imsm(super, 0);
	/*  reload capability and hba */

	/* capability and hba must be updated with new super allocation */
	find_intel_hba_capability(fd, super, devname);
	super->len = ROUND_UP(anchor->mpb_size, 512);
	if (posix_memalign(&super->buf, 512, super->len) != 0) {
		if (devname)
			bb_error_msg("unable to allocate %zu byte mpb buffer",
				super->len);
		free(anchor);
		return 2;
	}
	memcpy(super->buf, anchor, 512);

	sectors = mpb_sectors(anchor) - 1;
	free(anchor);

	if (posix_memalign(&super->migr_rec_buf, 512, MIGR_REC_BUF_SIZE) != 0) {
		bb_error_msg("%s could not allocate migr_rec buffer", __func__);
		free(super->buf);
		return 2;
	}
	super->clean_migration_record_by_mdmon = 0;

	if (!sectors) {
		check_sum = __gen_imsm_checksum(super->anchor);
		if (check_sum != __le32_to_cpu(super->anchor->check_sum)) {
			if (devname)
				bb_error_msg("IMSM checksum %x != %x on %s",
					check_sum,
					__le32_to_cpu(super->anchor->check_sum),
					devname);
			return 2;
		}

		return 0;
	}

	/* read the extended mpb */
	if (lseek(fd, dsize - (512 * (2 + sectors)), SEEK_SET) < 0) {
		if (devname)
			bb_perror_msg("Cannot seek to extended mpb on %s", devname);
		return 1;
	}

	if ((unsigned)read(fd, super->buf + 512, super->len - 512) != super->len - 512) {
		if (devname)
			bb_perror_msg("Cannot read extended mpb on %s", devname);
		return 2;
	}

	check_sum = __gen_imsm_checksum(super->anchor);
	if (check_sum != __le32_to_cpu(super->anchor->check_sum)) {
		if (devname)
			bb_error_msg("IMSM checksum %x != %x on %s",
				check_sum, __le32_to_cpu(super->anchor->check_sum),
				devname);
		return 3;
	}

	/* FIXME the BBM log is disk specific so we cannot use this global
	 * buffer for all disks.  Ok for now since we only look at the global
	 * bbm_log_size parameter to gate assembly
	 */
	super->bbm_log = __get_imsm_bbm_log(super->anchor);

	return 0;
}

static int read_imsm_migr_rec(int fd, struct intel_super *super);

/* clears hi bits in metadata if MPB_ATTRIB_2TB_DISK not set */
static void clear_hi(struct intel_super *super)
{
	struct imsm_super *mpb = super->anchor;
	int i, n;
	if (mpb->attributes & MPB_ATTRIB_2TB_DISK)
		return;
	for (i = 0; i < mpb->num_disks; ++i) {
		struct imsm_disk *disk = &mpb->disk[i];
		disk->total_blocks_hi = 0;
	}
	for (i = 0; i < mpb->num_raid_devs; ++i) {
		struct imsm_dev *dev = get_imsm_dev(super, i);
		if (!dev)
			return;
		for (n = 0; n < 2; ++n) {
			struct imsm_map *map = get_imsm_map(dev, n);
			if (!map)
				continue;
			map->pba_of_lba0_hi = 0;
			map->blocks_per_member_hi = 0;
			map->num_data_stripes_hi = 0;
		}
	}
}

static int
load_and_parse_mpb(int fd, struct intel_super *super, char *devname, int keep_fd)
{
	int err;

	err = load_imsm_mpb(fd, super, devname);
	if (err)
		return err;
	err = load_imsm_disk(fd, super, devname, keep_fd);
	if (err)
		return err;
	err = parse_raid_devices(super);
	clear_hi(super);
	return err;
}

static void __free_imsm_disk(struct dl *d)
{
	if (d->fd >= 0)
		close(d->fd);
	if (d->devname)
		free(d->devname);
	if (d->e)
		free(d->e);
	free(d);

}

static void free_imsm_disks(struct intel_super *super)
{
	struct dl *d;

	while (super->disks) {
		d = super->disks;
		super->disks = d->next;
		__free_imsm_disk(d);
	}
	while (super->disk_mgmt_list) {
		d = super->disk_mgmt_list;
		super->disk_mgmt_list = d->next;
		__free_imsm_disk(d);
	}
	while (super->missing) {
		d = super->missing;
		super->missing = d->next;
		__free_imsm_disk(d);
	}

}

/* free all the pieces hanging off of a super pointer */
static void __free_imsm(struct intel_super *super, int free_disks)
{
	struct intel_hba *elem, *next;

	if (super->buf) {
		free(super->buf);
		super->buf = NULL;
	}
	/* unlink capability description */
	super->orom = NULL;
	if (super->migr_rec_buf) {
		free(super->migr_rec_buf);
		super->migr_rec_buf = NULL;
	}
	if (free_disks)
		free_imsm_disks(super);
	free_devlist(super);
	elem = super->hba;
	while (elem) {
		if (elem->path)
			free((void *)elem->path);
		next = elem->next;
		free(elem);
		elem = next;
	}
	super->hba = NULL;
}

static void free_imsm(struct intel_super *super)
{
	__free_imsm(super, 1);
	free(super);
}

static void free_super_imsm(struct supertype *st)
{
	struct intel_super *super = st->sb;

	if (!super)
		return;

	free_imsm(super);
	st->sb = NULL;
}

static struct intel_super *alloc_super(void)
{
	struct intel_super *super = zmalloc(sizeof(*super));

	super->current_vol = -1;
	super->create_offset = ~((unsigned long long) 0);
	return super;
}

/*
 * find and allocate hba and OROM/EFI based on valid fd of RAID component device
 */
static int find_intel_hba_capability(int fd, struct intel_super *super, char *devname)
{
	struct sys_dev *hba_name;
	int rv = 0;

	if ((fd < 0) || check_env("IMSM_NO_PLATFORM")) {
		super->orom = NULL;
		super->hba = NULL;
		return 0;
	}
	hba_name = find_disk_attached_hba(fd, NULL);
	if (!hba_name) {
		if (devname)
			bb_error_msg("%s is not attached to Intel(R) RAID controller.",
				devname);
		return 1;
	}
	rv = attach_hba_to_super(super, hba_name);
	if (rv == 2) {
		if (devname) {
			struct intel_hba *hba = super->hba;

			fprintf(stderr, "%s: %s is attached to Intel(R) %s RAID "
				"controller (%s),\n"
				"    but the container is assigned to Intel(R) "
				"%s RAID controller (",
				bb_applet_name, devname,
				hba_name->path,
				hba_name->pci_id ? : "Err!",
				get_sys_dev_type(hba_name->type));

			while (hba) {
				fprintf(stderr, "%s", hba->pci_id ? : "Err!");
				if (hba->next)
					fprintf(stderr, ", ");
				hba = hba->next;
			}

			fprintf(stderr, ").\n"
				"    Mixing devices attached to different controllers "
				"is not allowed.\n");
		}
		free_sys_dev(&hba_name);
		return 2;
	}
	super->orom = find_imsm_capability(hba_name->type);
	free_sys_dev(&hba_name);
	if (!super->orom)
		return 3;
	return 0;
}

static int load_super_imsm(struct supertype *st, int fd, char *devname)
{
	struct intel_super *super;
	int rv;

	if (test_partition(fd))
		/* IMSM not allowed on partitions */
		return 1;

	free_super_imsm(st);

	super = alloc_super();
	/* Load hba and capabilities if they exist.
	 * But do not preclude loading metadata in case capabilities or hba are
	 * non-compliant and ignore_hw_compat is set.
	 */
	rv = find_intel_hba_capability(fd, super, devname);
	/* no orom/efi or non-intel hba of the disk */
	if ((rv != 0) && (st->ignore_hw_compat == 0)) {
		if (devname)
			bb_error_msg("No OROM/EFI properties for %s", devname);
		free_imsm(super);
		return 2;
	}
	rv = load_and_parse_mpb(fd, super, devname, 0);

	if (rv) {
		if (devname)
			bb_error_msg("Failed to load all information "
				"sections on %s", devname);
		free_imsm(super);
		return rv;
	}

	st->sb = super;
	if (st->ss == NULL) {
		st->ss = &super_imsm;
		st->minor_version = 0;
		st->max_devs = IMSM_MAX_DEVICES;
	}

	/* load migration record */
	if (load_imsm_migr_rec(super, NULL) == 0) {
		/* Check for unsupported migration features */
		if (check_mpb_migr_compatibility(super) != 0) {
			bb_error_msg("Unsupported migration detected");
			if (devname)
				fprintf(stderr, " on %s\n", devname);
			else
				fprintf(stderr, " (IMSM).\n");
			return 3;
		}
	}

	return 0;
}

static int store_super_imsm(struct supertype *st, int fd)
{
	struct intel_super *super = st->sb;
	struct imsm_super *mpb = super ? super->anchor : NULL;

	(void)&fd;

	if (!mpb)
		return 1;

	return 1;
}

static int imsm_bbm_log_size(struct imsm_super *mpb)
{
	return __le32_to_cpu(mpb->bbm_log_size);
}

static int is_gen_migration(struct imsm_dev *dev)
{
	if (dev == NULL)
		return 0;

	if (!dev->vol.migr_state)
		return 0;

	if (migr_type(dev) == MIGR_GEN_MIGR)
		return 1;

	return 0;
}

static int is_rebuilding(struct imsm_dev *dev)
{
	struct imsm_map *migr_map;

	if (!dev->vol.migr_state)
		return 0;

	if (migr_type(dev) != MIGR_REBUILD)
		return 0;

	migr_map = get_imsm_map(dev, MAP_1);

	if (migr_map->map_state == IMSM_T_STATE_DEGRADED)
		return 1;
	else
		return 0;
}

static void update_recovery_start(struct intel_super *super,
					struct imsm_dev *dev,
					struct mdinfo *array)
{
	struct mdinfo *rebuild = NULL;
	struct mdinfo *d;
	__u32 units;

	if (!is_rebuilding(dev))
		return;

	/* Find the rebuild target, but punt on the dual rebuild case */
	for (d = array->devs; d; d = d->next)
		if (d->recovery_start == 0) {
			if (rebuild)
				return;
			rebuild = d;
		}

	if (!rebuild) {
		/* (?) none of the disks are marked with
		 * IMSM_ORD_REBUILD, so assume they are missing and the
		 * disk_ord_tbl was not correctly updated
		 */
		dprintf("%s: failed to locate out-of-sync disk\n", __func__);
		return;
	}

	units = __le32_to_cpu(dev->vol.curr_migr_unit);
	rebuild->recovery_start = units * blocks_per_migr_unit(super, dev);
}

static struct mdinfo *container_content_imsm(struct supertype *st, char *subarray)
{
	/* Given a container loaded by load_super_imsm_all,
	 * extract information about all the arrays into
	 * an mdinfo tree.
	 * If 'subarray' is given, just extract info about that array.
	 *
	 * For each imsm_dev create an mdinfo, fill it in,
	 *  then look for matching devices in super->disks
	 *  and create appropriate device mdinfo.
	 */
	struct intel_super *super = st->sb;
	struct imsm_super *mpb = super->anchor;
	struct mdinfo *rest = NULL;
	unsigned int i;
	int sb_errors = 0;
	struct dl *d;
	int spare_disks = 0;

	/* do not assemble arrays when not all attributes are supported */
	if (imsm_check_attributes(mpb->attributes) == 0) {
		sb_errors = 1;
		bb_error_msg("Unsupported attributes in IMSM metadata."
			"Arrays activation is blocked.");
	}

	/* check for bad blocks */
	if (imsm_bbm_log_size(super->anchor)) {
		bb_error_msg("BBM log found in IMSM metadata."
			"Arrays activation is blocked.");
		sb_errors = 1;
	}


	/* count spare devices, not used in maps
	 */
	for (d = super->disks; d; d = d->next)
		if (d->index == -1)
			spare_disks++;

	for (i = 0; i < mpb->num_raid_devs; i++) {
		struct imsm_dev *dev;
		struct imsm_map *map;
		struct imsm_map *map2;
		struct mdinfo *this;
		int slot;
		char *ep;

		if (subarray &&
		    (i != strtoul(subarray, &ep, 10) || *ep != '\0'))
			continue;

		dev = get_imsm_dev(super, i);
		map = get_imsm_map(dev, MAP_0);
		map2 = get_imsm_map(dev, MAP_1);

		/* do not publish arrays that are in the middle of an
		 * unsupported migration
		 */
		if (dev->vol.migr_state &&
		    (migr_type(dev) == MIGR_STATE_CHANGE)) {
			bb_error_msg("cannot assemble volume '%.16s':"
				" unsupported migration in progress",
				dev->volume);
			continue;
		}
		/* do not publish arrays that are not support by controller's
		 * OROM/EFI
		 */

		this = xmalloc(sizeof(*this));

		super->current_vol = i;
		getinfo_super_imsm_volume(st, this, NULL);
		this->next = rest;

		/* if array has bad blocks, set suitable bit in all arrays state */
		if (sb_errors)
			this->array.state |=
			  (1<<MD_SB_BLOCK_CONTAINER_RESHAPE) |
			  (1<<MD_SB_BLOCK_VOLUME);

		for (slot = 0 ; slot <  map->num_members; slot++) {
			unsigned long long recovery_start;
			struct mdinfo *info_d;
			int idx;
			int skip;
			__u32 ord;

			skip = 0;
			idx = get_imsm_disk_idx(dev, slot, MAP_0);
			ord = get_imsm_ord_tbl_ent(dev, slot, MAP_X);
			for (d = super->disks; d ; d = d->next)
				if (d->index == idx)
					break;

			recovery_start = MaxSector;
			if (d == NULL)
				skip = 1;
			if (d && is_failed(&d->disk))
				skip = 1;
			if (ord & IMSM_ORD_REBUILD)
				recovery_start = 0;

			/*
			 * if we skip some disks the array will be assmebled degraded;
			 * reset resync start to avoid a dirty-degraded
			 * situation when performing the intial sync
			 *
			 * FIXME handle dirty degraded
			 */
			if ((skip || recovery_start == 0) && !dev->vol.dirty)
				this->resync_start = MaxSector;
			if (skip)
				continue;

			info_d = calloc(1, sizeof(*info_d));
			if (!info_d) {
				bb_error_msg("failed to allocate disk"
					" for volume %.16s", dev->volume);
				info_d = this->devs;
				while (info_d) {
					struct mdinfo *d_next = info_d->next;

					free(info_d);
					info_d = d_next;
				}
				free(this);
				this = rest;
				break;
			}
			info_d->next = this->devs;
			this->devs = info_d;

			info_d->disk.number = d->index;
			info_d->disk.major = d->major;
			info_d->disk.minor = d->minor;
			info_d->disk.raid_disk = slot;
			info_d->recovery_start = recovery_start;
			if (map2) {
				if (slot < map2->num_members)
					info_d->disk.state = (1 << MD_DISK_ACTIVE);
				else
					this->array.spare_disks++;
			} else {
				if (slot < map->num_members)
					info_d->disk.state = (1 << MD_DISK_ACTIVE);
				else
					this->array.spare_disks++;
			}
			if (info_d->recovery_start == MaxSector)
				this->array.working_disks++;

			info_d->events = __le32_to_cpu(mpb->generation_num);
			info_d->data_offset = pba_of_lba0(map);
			info_d->component_size = blocks_per_member(map);
		}
		/* now that the disk list is up-to-date fixup recovery_start */
		update_recovery_start(super, dev, this);
		this->array.spare_disks += spare_disks;

		rest = this;
	}

	return rest;
}


static __u8 imsm_check_degraded(struct intel_super *super, struct imsm_dev *dev,
				int failed, int look_in_map)
{
	struct imsm_map *map;

	map = get_imsm_map(dev, look_in_map);

	if (!failed)
		return map->map_state == IMSM_T_STATE_UNINITIALIZED ?
			IMSM_T_STATE_UNINITIALIZED : IMSM_T_STATE_NORMAL;

	switch (get_imsm_raid_level(map)) {
	case 0:
		return IMSM_T_STATE_FAILED;
		break;
	case 1:
		if (failed < map->num_members)
			return IMSM_T_STATE_DEGRADED;
		else
			return IMSM_T_STATE_FAILED;
		break;
	case 10:
	{
		/**
		 * check to see if any mirrors have failed, otherwise we
		 * are degraded.  Even numbered slots are mirrored on
		 * slot+1
		 */
		int i;
		/* gcc -Os complains that this is unused */
		int insync = insync;

		for (i = 0; i < map->num_members; i++) {
			__u32 ord = get_imsm_ord_tbl_ent(dev, i, MAP_X);
			int idx = ord_to_idx(ord);
			struct imsm_disk *disk;

			/* reset the potential in-sync count on even-numbered
			 * slots.  num_copies is always 2 for imsm raid10
			 */
			if ((i & 1) == 0)
				insync = 2;

			disk = get_imsm_disk(super, idx);
			if (!disk || is_failed(disk) || ord & IMSM_ORD_REBUILD)
				insync--;

			/* no in-sync disks left in this mirror the
			 * array has failed
			 */
			if (insync == 0)
				return IMSM_T_STATE_FAILED;
		}

		return IMSM_T_STATE_DEGRADED;
	}
	case 5:
		if (failed < 2)
			return IMSM_T_STATE_DEGRADED;
		else
			return IMSM_T_STATE_FAILED;
		break;
	default:
		break;
	}

	return map->map_state;
}

static int imsm_count_failed(struct intel_super *super, struct imsm_dev *dev,
			     int look_in_map)
{
	int i;
	int failed = 0;
	struct imsm_disk *disk;
	struct imsm_map *map = get_imsm_map(dev, MAP_0);
	struct imsm_map *prev = get_imsm_map(dev, MAP_1);
	struct imsm_map *map_for_loop;
	__u32 ord;
	int idx;
	int idx_1;

	/* at the beginning of migration we set IMSM_ORD_REBUILD on
	 * disks that are being rebuilt.  New failures are recorded to
	 * map[0].  So we look through all the disks we started with and
	 * see if any failures are still present, or if any new ones
	 * have arrived
	 */
	map_for_loop = map;
	if (prev && (map->num_members < prev->num_members))
		map_for_loop = prev;

	for (i = 0; i < map_for_loop->num_members; i++) {
		idx_1 = -255;
		/* when MAP_X is passed both maps failures are counted
		 */
		if (prev &&
		    ((look_in_map == MAP_1) || (look_in_map == MAP_X)) &&
		    (i < prev->num_members)) {
			ord = __le32_to_cpu(prev->disk_ord_tbl[i]);
			idx_1 = ord_to_idx(ord);

			disk = get_imsm_disk(super, idx_1);
			if (!disk || is_failed(disk) || ord & IMSM_ORD_REBUILD)
				failed++;
		}
		if (((look_in_map == MAP_0) || (look_in_map == MAP_X)) &&
		    (i < map->num_members)) {
			ord = __le32_to_cpu(map->disk_ord_tbl[i]);
			idx = ord_to_idx(ord);

			if (idx != idx_1) {
				disk = get_imsm_disk(super, idx);
				if (!disk || is_failed(disk) ||
				    ord & IMSM_ORD_REBUILD)
					failed++;
			}
		}
	}

	return failed;
}

static const struct superswitch super_imsm = {
	.match_home     = match_home_imsm,
	.uuid_from_super= uuid_from_super_imsm,
	.getinfo_super  = getinfo_super_imsm,
	.getinfo_super_disks = getinfo_super_disks_imsm,
	.update_super   = update_super_imsm,

	.min_acceptable_spare_size = min_acceptable_spare_size_imsm,

	.compare_super  = compare_super_imsm,

	.load_super     = load_super_imsm,
	.store_super    = store_super_imsm,
	.free_super     = free_super_imsm,
	.match_metadata_desc = match_metadata_desc_imsm,
	.container_content = container_content_imsm,


	.external       = 1,
	.name = "imsm",

};

/* crc32.c -- compute the CRC-32 of a data stream
 * Thanks to Rodney Brown <rbrown64@csc.com.au> for his contribution of faster
 * CRC methods: exclusive-oring 32 bits of data at a time, and pre-computing
 * tables for updating the shift register in one step with three exclusive-ors
 * instead of four steps with four exclusive-ors.  This results about a factor
 * of two increase in speed on a Power PC G4 (PPC7455) using gcc -O3.
 */

/*
  Note on the use of DYNAMIC_CRC_TABLE: there is no mutex or semaphore
  protection on the static variables used to control the first-use generation
  of the crc tables.  Therefore, if you #define DYNAMIC_CRC_TABLE, you should
  first call get_crc_table() to initialize the tables before allowing more than
  one thread to use crc32().
 */

#ifdef MAKECRCH
#  ifndef DYNAMIC_CRC_TABLE
#    define DYNAMIC_CRC_TABLE
#  endif /* !DYNAMIC_CRC_TABLE */
#endif /* MAKECRCH */

#define Z_NULL ((void*)0)
#define OF(X) X
#define ZEXPORT
#define NOBYFOUR

#define local static

/* Find a four-byte integer type for crc32_little() and crc32_big(). */
#ifndef NOBYFOUR
#    define BYFOUR
#    if (UINT_MAX == 0xffffffffUL)
       typedef unsigned int u4;
#    else
#      if (ULONG_MAX == 0xffffffffUL)
	 typedef unsigned long u4;
#      else
#        if (USHRT_MAX == 0xffffffffUL)
	   typedef unsigned short u4;
#        else
#          undef BYFOUR     /* can't find a four-byte integer type! */
#        endif
#      endif
#    endif
#endif /* !NOBYFOUR */

/* Definitions for doing the crc four data bytes at a time. */
#ifdef BYFOUR
#  define REV(w) (((w)>>24)+(((w)>>8)&0xff00)+ \
		(((w)&0xff00)<<8)+(((w)&0xff)<<24))
   local unsigned long crc32_little OF((unsigned long,
			const unsigned char *, unsigned));
   local unsigned long crc32_big OF((unsigned long,
			const unsigned char *, unsigned));
#  define TBLS 8
#else
#  define TBLS 1
#endif /* BYFOUR */

#ifdef DYNAMIC_CRC_TABLE

local volatile int crc_table_empty = 1;
local unsigned long crc_table[TBLS][256];
local void make_crc_table OF((void));
#ifdef MAKECRCH
   local void write_table OF((FILE *, const unsigned long *));
#endif /* MAKECRCH */

/*
  Generate tables for a byte-wise 32-bit CRC calculation on the polynomial:
  x^32+x^26+x^23+x^22+x^16+x^12+x^11+x^10+x^8+x^7+x^5+x^4+x^2+x+1.

  Polynomials over GF(2) are represented in binary, one bit per coefficient,
  with the lowest powers in the most significant bit.  Then adding polynomials
  is just exclusive-or, and multiplying a polynomial by x is a right shift by
  one.  If we call the above polynomial p, and represent a byte as the
  polynomial q, also with the lowest power in the most significant bit (so the
  byte 0xb1 is the polynomial x^7+x^3+x+1), then the CRC is (q*x^32) mod p,
  where a mod b means the remainder after dividing a by b.

  This calculation is done using the shift-register method of multiplying and
  taking the remainder.  The register is initialized to zero, and for each
  incoming bit, x^32 is added mod p to the register if the bit is a one (where
  x^32 mod p is p+x^32 = x^26+...+1), and the register is multiplied mod p by
  x (which is shifting right by one and adding x^32 mod p if the bit shifted
  out is a one).  We start with the highest power (least significant bit) of
  q and repeat for all eight bits of q.

  The first table is simply the CRC of all possible eight bit values.  This is
  all the information needed to generate CRCs on data a byte at a time for all
  combinations of CRC register values and incoming bytes.  The remaining tables
  allow for word-at-a-time CRC calculation for both big-endian and little-
  endian machines, where a word is four bytes.
*/
local void make_crc_table()
{
    unsigned long c;
    int n, k;
    unsigned long poly;                 /* polynomial exclusive-or pattern */
    /* terms of polynomial defining this crc (except x^32): */
    static volatile int first = 1;      /* flag to limit concurrent making */
    static const unsigned char p[] = {0,1,2,4,5,7,8,10,11,12,16,22,23,26};

    /* See if another task is already doing this (not thread-safe, but better
       than nothing -- significantly reduces duration of vulnerability in
       case the advice about DYNAMIC_CRC_TABLE is ignored) */
    if (first) {
	first = 0;

	/* make exclusive-or pattern from polynomial (0xedb88320UL) */
	poly = 0UL;
	for (n = 0; n < sizeof(p)/sizeof(unsigned char); n++)
	    poly |= 1UL << (31 - p[n]);

	/* generate a crc for every 8-bit value */
	for (n = 0; n < 256; n++) {
	    c = (unsigned long)n;
	    for (k = 0; k < 8; k++)
		c = c & 1 ? poly ^ (c >> 1) : c >> 1;
	    crc_table[0][n] = c;
	}

#ifdef BYFOUR
	/* generate crc for each value followed by one, two, and three zeros,
	   and then the byte reversal of those as well as the first table */
	for (n = 0; n < 256; n++) {
	    c = crc_table[0][n];
	    crc_table[4][n] = REV(c);
	    for (k = 1; k < 4; k++) {
		c = crc_table[0][c & 0xff] ^ (c >> 8);
		crc_table[k][n] = c;
		crc_table[k + 4][n] = REV(c);
	    }
	}
#endif /* BYFOUR */

	crc_table_empty = 0;
    }
    else {      /* not first */
	/* wait for the other guy to finish (not efficient, but rare) */
	while (crc_table_empty)
	    ;
    }

#ifdef MAKECRCH
    /* write out CRC tables to crc32.h */
    {
	FILE *out;

	out = fopen("crc32.h", "w");
	if (out == NULL) return;
	fprintf(out, "/* crc32.h -- tables for rapid CRC calculation\n");
	fprintf(out, " * Generated automatically by crc32.c\n */\n\n");
	fprintf(out, "local const unsigned long ");
	fprintf(out, "crc_table[TBLS][256] =\n{\n  {\n");
	write_table(out, crc_table[0]);
#  ifdef BYFOUR
	fprintf(out, "#ifdef BYFOUR\n");
	for (k = 1; k < 8; k++) {
	    fprintf(out, "  },\n  {\n");
	    write_table(out, crc_table[k]);
	}
	fprintf(out, "#endif\n");
#  endif /* BYFOUR */
	fprintf(out, "  }\n};\n");
	fclose(out);
    }
#endif /* MAKECRCH */
}

#ifdef MAKECRCH
local void write_table(out, table)
    FILE *out;
    const unsigned long *table;
{
    int n;

    for (n = 0; n < 256; n++)
	fprintf(out, "%s0x%08lxUL%s", n % 5 ? "" : "    ", table[n],
		n == 255 ? "\n" : (n % 5 == 4 ? ",\n" : ", "));
}
#endif /* MAKECRCH */

#else /* !DYNAMIC_CRC_TABLE */
/* ========================================================================
 * Tables of CRC-32s of all single-byte values, made by make_crc_table().
 */
local const unsigned long crc_table[TBLS][256] =
{
  {
    0x00000000UL, 0x77073096UL, 0xee0e612cUL, 0x990951baUL, 0x076dc419UL,
    0x706af48fUL, 0xe963a535UL, 0x9e6495a3UL, 0x0edb8832UL, 0x79dcb8a4UL,
    0xe0d5e91eUL, 0x97d2d988UL, 0x09b64c2bUL, 0x7eb17cbdUL, 0xe7b82d07UL,
    0x90bf1d91UL, 0x1db71064UL, 0x6ab020f2UL, 0xf3b97148UL, 0x84be41deUL,
    0x1adad47dUL, 0x6ddde4ebUL, 0xf4d4b551UL, 0x83d385c7UL, 0x136c9856UL,
    0x646ba8c0UL, 0xfd62f97aUL, 0x8a65c9ecUL, 0x14015c4fUL, 0x63066cd9UL,
    0xfa0f3d63UL, 0x8d080df5UL, 0x3b6e20c8UL, 0x4c69105eUL, 0xd56041e4UL,
    0xa2677172UL, 0x3c03e4d1UL, 0x4b04d447UL, 0xd20d85fdUL, 0xa50ab56bUL,
    0x35b5a8faUL, 0x42b2986cUL, 0xdbbbc9d6UL, 0xacbcf940UL, 0x32d86ce3UL,
    0x45df5c75UL, 0xdcd60dcfUL, 0xabd13d59UL, 0x26d930acUL, 0x51de003aUL,
    0xc8d75180UL, 0xbfd06116UL, 0x21b4f4b5UL, 0x56b3c423UL, 0xcfba9599UL,
    0xb8bda50fUL, 0x2802b89eUL, 0x5f058808UL, 0xc60cd9b2UL, 0xb10be924UL,
    0x2f6f7c87UL, 0x58684c11UL, 0xc1611dabUL, 0xb6662d3dUL, 0x76dc4190UL,
    0x01db7106UL, 0x98d220bcUL, 0xefd5102aUL, 0x71b18589UL, 0x06b6b51fUL,
    0x9fbfe4a5UL, 0xe8b8d433UL, 0x7807c9a2UL, 0x0f00f934UL, 0x9609a88eUL,
    0xe10e9818UL, 0x7f6a0dbbUL, 0x086d3d2dUL, 0x91646c97UL, 0xe6635c01UL,
    0x6b6b51f4UL, 0x1c6c6162UL, 0x856530d8UL, 0xf262004eUL, 0x6c0695edUL,
    0x1b01a57bUL, 0x8208f4c1UL, 0xf50fc457UL, 0x65b0d9c6UL, 0x12b7e950UL,
    0x8bbeb8eaUL, 0xfcb9887cUL, 0x62dd1ddfUL, 0x15da2d49UL, 0x8cd37cf3UL,
    0xfbd44c65UL, 0x4db26158UL, 0x3ab551ceUL, 0xa3bc0074UL, 0xd4bb30e2UL,
    0x4adfa541UL, 0x3dd895d7UL, 0xa4d1c46dUL, 0xd3d6f4fbUL, 0x4369e96aUL,
    0x346ed9fcUL, 0xad678846UL, 0xda60b8d0UL, 0x44042d73UL, 0x33031de5UL,
    0xaa0a4c5fUL, 0xdd0d7cc9UL, 0x5005713cUL, 0x270241aaUL, 0xbe0b1010UL,
    0xc90c2086UL, 0x5768b525UL, 0x206f85b3UL, 0xb966d409UL, 0xce61e49fUL,
    0x5edef90eUL, 0x29d9c998UL, 0xb0d09822UL, 0xc7d7a8b4UL, 0x59b33d17UL,
    0x2eb40d81UL, 0xb7bd5c3bUL, 0xc0ba6cadUL, 0xedb88320UL, 0x9abfb3b6UL,
    0x03b6e20cUL, 0x74b1d29aUL, 0xead54739UL, 0x9dd277afUL, 0x04db2615UL,
    0x73dc1683UL, 0xe3630b12UL, 0x94643b84UL, 0x0d6d6a3eUL, 0x7a6a5aa8UL,
    0xe40ecf0bUL, 0x9309ff9dUL, 0x0a00ae27UL, 0x7d079eb1UL, 0xf00f9344UL,
    0x8708a3d2UL, 0x1e01f268UL, 0x6906c2feUL, 0xf762575dUL, 0x806567cbUL,
    0x196c3671UL, 0x6e6b06e7UL, 0xfed41b76UL, 0x89d32be0UL, 0x10da7a5aUL,
    0x67dd4accUL, 0xf9b9df6fUL, 0x8ebeeff9UL, 0x17b7be43UL, 0x60b08ed5UL,
    0xd6d6a3e8UL, 0xa1d1937eUL, 0x38d8c2c4UL, 0x4fdff252UL, 0xd1bb67f1UL,
    0xa6bc5767UL, 0x3fb506ddUL, 0x48b2364bUL, 0xd80d2bdaUL, 0xaf0a1b4cUL,
    0x36034af6UL, 0x41047a60UL, 0xdf60efc3UL, 0xa867df55UL, 0x316e8eefUL,
    0x4669be79UL, 0xcb61b38cUL, 0xbc66831aUL, 0x256fd2a0UL, 0x5268e236UL,
    0xcc0c7795UL, 0xbb0b4703UL, 0x220216b9UL, 0x5505262fUL, 0xc5ba3bbeUL,
    0xb2bd0b28UL, 0x2bb45a92UL, 0x5cb36a04UL, 0xc2d7ffa7UL, 0xb5d0cf31UL,
    0x2cd99e8bUL, 0x5bdeae1dUL, 0x9b64c2b0UL, 0xec63f226UL, 0x756aa39cUL,
    0x026d930aUL, 0x9c0906a9UL, 0xeb0e363fUL, 0x72076785UL, 0x05005713UL,
    0x95bf4a82UL, 0xe2b87a14UL, 0x7bb12baeUL, 0x0cb61b38UL, 0x92d28e9bUL,
    0xe5d5be0dUL, 0x7cdcefb7UL, 0x0bdbdf21UL, 0x86d3d2d4UL, 0xf1d4e242UL,
    0x68ddb3f8UL, 0x1fda836eUL, 0x81be16cdUL, 0xf6b9265bUL, 0x6fb077e1UL,
    0x18b74777UL, 0x88085ae6UL, 0xff0f6a70UL, 0x66063bcaUL, 0x11010b5cUL,
    0x8f659effUL, 0xf862ae69UL, 0x616bffd3UL, 0x166ccf45UL, 0xa00ae278UL,
    0xd70dd2eeUL, 0x4e048354UL, 0x3903b3c2UL, 0xa7672661UL, 0xd06016f7UL,
    0x4969474dUL, 0x3e6e77dbUL, 0xaed16a4aUL, 0xd9d65adcUL, 0x40df0b66UL,
    0x37d83bf0UL, 0xa9bcae53UL, 0xdebb9ec5UL, 0x47b2cf7fUL, 0x30b5ffe9UL,
    0xbdbdf21cUL, 0xcabac28aUL, 0x53b39330UL, 0x24b4a3a6UL, 0xbad03605UL,
    0xcdd70693UL, 0x54de5729UL, 0x23d967bfUL, 0xb3667a2eUL, 0xc4614ab8UL,
    0x5d681b02UL, 0x2a6f2b94UL, 0xb40bbe37UL, 0xc30c8ea1UL, 0x5a05df1bUL,
    0x2d02ef8dUL
#ifdef BYFOUR
  },
  {
    0x00000000UL, 0x191b3141UL, 0x32366282UL, 0x2b2d53c3UL, 0x646cc504UL,
    0x7d77f445UL, 0x565aa786UL, 0x4f4196c7UL, 0xc8d98a08UL, 0xd1c2bb49UL,
    0xfaefe88aUL, 0xe3f4d9cbUL, 0xacb54f0cUL, 0xb5ae7e4dUL, 0x9e832d8eUL,
    0x87981ccfUL, 0x4ac21251UL, 0x53d92310UL, 0x78f470d3UL, 0x61ef4192UL,
    0x2eaed755UL, 0x37b5e614UL, 0x1c98b5d7UL, 0x05838496UL, 0x821b9859UL,
    0x9b00a918UL, 0xb02dfadbUL, 0xa936cb9aUL, 0xe6775d5dUL, 0xff6c6c1cUL,
    0xd4413fdfUL, 0xcd5a0e9eUL, 0x958424a2UL, 0x8c9f15e3UL, 0xa7b24620UL,
    0xbea97761UL, 0xf1e8e1a6UL, 0xe8f3d0e7UL, 0xc3de8324UL, 0xdac5b265UL,
    0x5d5daeaaUL, 0x44469febUL, 0x6f6bcc28UL, 0x7670fd69UL, 0x39316baeUL,
    0x202a5aefUL, 0x0b07092cUL, 0x121c386dUL, 0xdf4636f3UL, 0xc65d07b2UL,
    0xed705471UL, 0xf46b6530UL, 0xbb2af3f7UL, 0xa231c2b6UL, 0x891c9175UL,
    0x9007a034UL, 0x179fbcfbUL, 0x0e848dbaUL, 0x25a9de79UL, 0x3cb2ef38UL,
    0x73f379ffUL, 0x6ae848beUL, 0x41c51b7dUL, 0x58de2a3cUL, 0xf0794f05UL,
    0xe9627e44UL, 0xc24f2d87UL, 0xdb541cc6UL, 0x94158a01UL, 0x8d0ebb40UL,
    0xa623e883UL, 0xbf38d9c2UL, 0x38a0c50dUL, 0x21bbf44cUL, 0x0a96a78fUL,
    0x138d96ceUL, 0x5ccc0009UL, 0x45d73148UL, 0x6efa628bUL, 0x77e153caUL,
    0xbabb5d54UL, 0xa3a06c15UL, 0x888d3fd6UL, 0x91960e97UL, 0xded79850UL,
    0xc7cca911UL, 0xece1fad2UL, 0xf5facb93UL, 0x7262d75cUL, 0x6b79e61dUL,
    0x4054b5deUL, 0x594f849fUL, 0x160e1258UL, 0x0f152319UL, 0x243870daUL,
    0x3d23419bUL, 0x65fd6ba7UL, 0x7ce65ae6UL, 0x57cb0925UL, 0x4ed03864UL,
    0x0191aea3UL, 0x188a9fe2UL, 0x33a7cc21UL, 0x2abcfd60UL, 0xad24e1afUL,
    0xb43fd0eeUL, 0x9f12832dUL, 0x8609b26cUL, 0xc94824abUL, 0xd05315eaUL,
    0xfb7e4629UL, 0xe2657768UL, 0x2f3f79f6UL, 0x362448b7UL, 0x1d091b74UL,
    0x04122a35UL, 0x4b53bcf2UL, 0x52488db3UL, 0x7965de70UL, 0x607eef31UL,
    0xe7e6f3feUL, 0xfefdc2bfUL, 0xd5d0917cUL, 0xcccba03dUL, 0x838a36faUL,
    0x9a9107bbUL, 0xb1bc5478UL, 0xa8a76539UL, 0x3b83984bUL, 0x2298a90aUL,
    0x09b5fac9UL, 0x10aecb88UL, 0x5fef5d4fUL, 0x46f46c0eUL, 0x6dd93fcdUL,
    0x74c20e8cUL, 0xf35a1243UL, 0xea412302UL, 0xc16c70c1UL, 0xd8774180UL,
    0x9736d747UL, 0x8e2de606UL, 0xa500b5c5UL, 0xbc1b8484UL, 0x71418a1aUL,
    0x685abb5bUL, 0x4377e898UL, 0x5a6cd9d9UL, 0x152d4f1eUL, 0x0c367e5fUL,
    0x271b2d9cUL, 0x3e001cddUL, 0xb9980012UL, 0xa0833153UL, 0x8bae6290UL,
    0x92b553d1UL, 0xddf4c516UL, 0xc4eff457UL, 0xefc2a794UL, 0xf6d996d5UL,
    0xae07bce9UL, 0xb71c8da8UL, 0x9c31de6bUL, 0x852aef2aUL, 0xca6b79edUL,
    0xd37048acUL, 0xf85d1b6fUL, 0xe1462a2eUL, 0x66de36e1UL, 0x7fc507a0UL,
    0x54e85463UL, 0x4df36522UL, 0x02b2f3e5UL, 0x1ba9c2a4UL, 0x30849167UL,
    0x299fa026UL, 0xe4c5aeb8UL, 0xfdde9ff9UL, 0xd6f3cc3aUL, 0xcfe8fd7bUL,
    0x80a96bbcUL, 0x99b25afdUL, 0xb29f093eUL, 0xab84387fUL, 0x2c1c24b0UL,
    0x350715f1UL, 0x1e2a4632UL, 0x07317773UL, 0x4870e1b4UL, 0x516bd0f5UL,
    0x7a468336UL, 0x635db277UL, 0xcbfad74eUL, 0xd2e1e60fUL, 0xf9ccb5ccUL,
    0xe0d7848dUL, 0xaf96124aUL, 0xb68d230bUL, 0x9da070c8UL, 0x84bb4189UL,
    0x03235d46UL, 0x1a386c07UL, 0x31153fc4UL, 0x280e0e85UL, 0x674f9842UL,
    0x7e54a903UL, 0x5579fac0UL, 0x4c62cb81UL, 0x8138c51fUL, 0x9823f45eUL,
    0xb30ea79dUL, 0xaa1596dcUL, 0xe554001bUL, 0xfc4f315aUL, 0xd7626299UL,
    0xce7953d8UL, 0x49e14f17UL, 0x50fa7e56UL, 0x7bd72d95UL, 0x62cc1cd4UL,
    0x2d8d8a13UL, 0x3496bb52UL, 0x1fbbe891UL, 0x06a0d9d0UL, 0x5e7ef3ecUL,
    0x4765c2adUL, 0x6c48916eUL, 0x7553a02fUL, 0x3a1236e8UL, 0x230907a9UL,
    0x0824546aUL, 0x113f652bUL, 0x96a779e4UL, 0x8fbc48a5UL, 0xa4911b66UL,
    0xbd8a2a27UL, 0xf2cbbce0UL, 0xebd08da1UL, 0xc0fdde62UL, 0xd9e6ef23UL,
    0x14bce1bdUL, 0x0da7d0fcUL, 0x268a833fUL, 0x3f91b27eUL, 0x70d024b9UL,
    0x69cb15f8UL, 0x42e6463bUL, 0x5bfd777aUL, 0xdc656bb5UL, 0xc57e5af4UL,
    0xee530937UL, 0xf7483876UL, 0xb809aeb1UL, 0xa1129ff0UL, 0x8a3fcc33UL,
    0x9324fd72UL
  },
  {
    0x00000000UL, 0x01c26a37UL, 0x0384d46eUL, 0x0246be59UL, 0x0709a8dcUL,
    0x06cbc2ebUL, 0x048d7cb2UL, 0x054f1685UL, 0x0e1351b8UL, 0x0fd13b8fUL,
    0x0d9785d6UL, 0x0c55efe1UL, 0x091af964UL, 0x08d89353UL, 0x0a9e2d0aUL,
    0x0b5c473dUL, 0x1c26a370UL, 0x1de4c947UL, 0x1fa2771eUL, 0x1e601d29UL,
    0x1b2f0bacUL, 0x1aed619bUL, 0x18abdfc2UL, 0x1969b5f5UL, 0x1235f2c8UL,
    0x13f798ffUL, 0x11b126a6UL, 0x10734c91UL, 0x153c5a14UL, 0x14fe3023UL,
    0x16b88e7aUL, 0x177ae44dUL, 0x384d46e0UL, 0x398f2cd7UL, 0x3bc9928eUL,
    0x3a0bf8b9UL, 0x3f44ee3cUL, 0x3e86840bUL, 0x3cc03a52UL, 0x3d025065UL,
    0x365e1758UL, 0x379c7d6fUL, 0x35dac336UL, 0x3418a901UL, 0x3157bf84UL,
    0x3095d5b3UL, 0x32d36beaUL, 0x331101ddUL, 0x246be590UL, 0x25a98fa7UL,
    0x27ef31feUL, 0x262d5bc9UL, 0x23624d4cUL, 0x22a0277bUL, 0x20e69922UL,
    0x2124f315UL, 0x2a78b428UL, 0x2bbade1fUL, 0x29fc6046UL, 0x283e0a71UL,
    0x2d711cf4UL, 0x2cb376c3UL, 0x2ef5c89aUL, 0x2f37a2adUL, 0x709a8dc0UL,
    0x7158e7f7UL, 0x731e59aeUL, 0x72dc3399UL, 0x7793251cUL, 0x76514f2bUL,
    0x7417f172UL, 0x75d59b45UL, 0x7e89dc78UL, 0x7f4bb64fUL, 0x7d0d0816UL,
    0x7ccf6221UL, 0x798074a4UL, 0x78421e93UL, 0x7a04a0caUL, 0x7bc6cafdUL,
    0x6cbc2eb0UL, 0x6d7e4487UL, 0x6f38fadeUL, 0x6efa90e9UL, 0x6bb5866cUL,
    0x6a77ec5bUL, 0x68315202UL, 0x69f33835UL, 0x62af7f08UL, 0x636d153fUL,
    0x612bab66UL, 0x60e9c151UL, 0x65a6d7d4UL, 0x6464bde3UL, 0x662203baUL,
    0x67e0698dUL, 0x48d7cb20UL, 0x4915a117UL, 0x4b531f4eUL, 0x4a917579UL,
    0x4fde63fcUL, 0x4e1c09cbUL, 0x4c5ab792UL, 0x4d98dda5UL, 0x46c49a98UL,
    0x4706f0afUL, 0x45404ef6UL, 0x448224c1UL, 0x41cd3244UL, 0x400f5873UL,
    0x4249e62aUL, 0x438b8c1dUL, 0x54f16850UL, 0x55330267UL, 0x5775bc3eUL,
    0x56b7d609UL, 0x53f8c08cUL, 0x523aaabbUL, 0x507c14e2UL, 0x51be7ed5UL,
    0x5ae239e8UL, 0x5b2053dfUL, 0x5966ed86UL, 0x58a487b1UL, 0x5deb9134UL,
    0x5c29fb03UL, 0x5e6f455aUL, 0x5fad2f6dUL, 0xe1351b80UL, 0xe0f771b7UL,
    0xe2b1cfeeUL, 0xe373a5d9UL, 0xe63cb35cUL, 0xe7fed96bUL, 0xe5b86732UL,
    0xe47a0d05UL, 0xef264a38UL, 0xeee4200fUL, 0xeca29e56UL, 0xed60f461UL,
    0xe82fe2e4UL, 0xe9ed88d3UL, 0xebab368aUL, 0xea695cbdUL, 0xfd13b8f0UL,
    0xfcd1d2c7UL, 0xfe976c9eUL, 0xff5506a9UL, 0xfa1a102cUL, 0xfbd87a1bUL,
    0xf99ec442UL, 0xf85cae75UL, 0xf300e948UL, 0xf2c2837fUL, 0xf0843d26UL,
    0xf1465711UL, 0xf4094194UL, 0xf5cb2ba3UL, 0xf78d95faUL, 0xf64fffcdUL,
    0xd9785d60UL, 0xd8ba3757UL, 0xdafc890eUL, 0xdb3ee339UL, 0xde71f5bcUL,
    0xdfb39f8bUL, 0xddf521d2UL, 0xdc374be5UL, 0xd76b0cd8UL, 0xd6a966efUL,
    0xd4efd8b6UL, 0xd52db281UL, 0xd062a404UL, 0xd1a0ce33UL, 0xd3e6706aUL,
    0xd2241a5dUL, 0xc55efe10UL, 0xc49c9427UL, 0xc6da2a7eUL, 0xc7184049UL,
    0xc25756ccUL, 0xc3953cfbUL, 0xc1d382a2UL, 0xc011e895UL, 0xcb4dafa8UL,
    0xca8fc59fUL, 0xc8c97bc6UL, 0xc90b11f1UL, 0xcc440774UL, 0xcd866d43UL,
    0xcfc0d31aUL, 0xce02b92dUL, 0x91af9640UL, 0x906dfc77UL, 0x922b422eUL,
    0x93e92819UL, 0x96a63e9cUL, 0x976454abUL, 0x9522eaf2UL, 0x94e080c5UL,
    0x9fbcc7f8UL, 0x9e7eadcfUL, 0x9c381396UL, 0x9dfa79a1UL, 0x98b56f24UL,
    0x99770513UL, 0x9b31bb4aUL, 0x9af3d17dUL, 0x8d893530UL, 0x8c4b5f07UL,
    0x8e0de15eUL, 0x8fcf8b69UL, 0x8a809decUL, 0x8b42f7dbUL, 0x89044982UL,
    0x88c623b5UL, 0x839a6488UL, 0x82580ebfUL, 0x801eb0e6UL, 0x81dcdad1UL,
    0x8493cc54UL, 0x8551a663UL, 0x8717183aUL, 0x86d5720dUL, 0xa9e2d0a0UL,
    0xa820ba97UL, 0xaa6604ceUL, 0xaba46ef9UL, 0xaeeb787cUL, 0xaf29124bUL,
    0xad6fac12UL, 0xacadc625UL, 0xa7f18118UL, 0xa633eb2fUL, 0xa4755576UL,
    0xa5b73f41UL, 0xa0f829c4UL, 0xa13a43f3UL, 0xa37cfdaaUL, 0xa2be979dUL,
    0xb5c473d0UL, 0xb40619e7UL, 0xb640a7beUL, 0xb782cd89UL, 0xb2cddb0cUL,
    0xb30fb13bUL, 0xb1490f62UL, 0xb08b6555UL, 0xbbd72268UL, 0xba15485fUL,
    0xb853f606UL, 0xb9919c31UL, 0xbcde8ab4UL, 0xbd1ce083UL, 0xbf5a5edaUL,
    0xbe9834edUL
  },
  {
    0x00000000UL, 0xb8bc6765UL, 0xaa09c88bUL, 0x12b5afeeUL, 0x8f629757UL,
    0x37def032UL, 0x256b5fdcUL, 0x9dd738b9UL, 0xc5b428efUL, 0x7d084f8aUL,
    0x6fbde064UL, 0xd7018701UL, 0x4ad6bfb8UL, 0xf26ad8ddUL, 0xe0df7733UL,
    0x58631056UL, 0x5019579fUL, 0xe8a530faUL, 0xfa109f14UL, 0x42acf871UL,
    0xdf7bc0c8UL, 0x67c7a7adUL, 0x75720843UL, 0xcdce6f26UL, 0x95ad7f70UL,
    0x2d111815UL, 0x3fa4b7fbUL, 0x8718d09eUL, 0x1acfe827UL, 0xa2738f42UL,
    0xb0c620acUL, 0x087a47c9UL, 0xa032af3eUL, 0x188ec85bUL, 0x0a3b67b5UL,
    0xb28700d0UL, 0x2f503869UL, 0x97ec5f0cUL, 0x8559f0e2UL, 0x3de59787UL,
    0x658687d1UL, 0xdd3ae0b4UL, 0xcf8f4f5aUL, 0x7733283fUL, 0xeae41086UL,
    0x525877e3UL, 0x40edd80dUL, 0xf851bf68UL, 0xf02bf8a1UL, 0x48979fc4UL,
    0x5a22302aUL, 0xe29e574fUL, 0x7f496ff6UL, 0xc7f50893UL, 0xd540a77dUL,
    0x6dfcc018UL, 0x359fd04eUL, 0x8d23b72bUL, 0x9f9618c5UL, 0x272a7fa0UL,
    0xbafd4719UL, 0x0241207cUL, 0x10f48f92UL, 0xa848e8f7UL, 0x9b14583dUL,
    0x23a83f58UL, 0x311d90b6UL, 0x89a1f7d3UL, 0x1476cf6aUL, 0xaccaa80fUL,
    0xbe7f07e1UL, 0x06c36084UL, 0x5ea070d2UL, 0xe61c17b7UL, 0xf4a9b859UL,
    0x4c15df3cUL, 0xd1c2e785UL, 0x697e80e0UL, 0x7bcb2f0eUL, 0xc377486bUL,
    0xcb0d0fa2UL, 0x73b168c7UL, 0x6104c729UL, 0xd9b8a04cUL, 0x446f98f5UL,
    0xfcd3ff90UL, 0xee66507eUL, 0x56da371bUL, 0x0eb9274dUL, 0xb6054028UL,
    0xa4b0efc6UL, 0x1c0c88a3UL, 0x81dbb01aUL, 0x3967d77fUL, 0x2bd27891UL,
    0x936e1ff4UL, 0x3b26f703UL, 0x839a9066UL, 0x912f3f88UL, 0x299358edUL,
    0xb4446054UL, 0x0cf80731UL, 0x1e4da8dfUL, 0xa6f1cfbaUL, 0xfe92dfecUL,
    0x462eb889UL, 0x549b1767UL, 0xec277002UL, 0x71f048bbUL, 0xc94c2fdeUL,
    0xdbf98030UL, 0x6345e755UL, 0x6b3fa09cUL, 0xd383c7f9UL, 0xc1366817UL,
    0x798a0f72UL, 0xe45d37cbUL, 0x5ce150aeUL, 0x4e54ff40UL, 0xf6e89825UL,
    0xae8b8873UL, 0x1637ef16UL, 0x048240f8UL, 0xbc3e279dUL, 0x21e91f24UL,
    0x99557841UL, 0x8be0d7afUL, 0x335cb0caUL, 0xed59b63bUL, 0x55e5d15eUL,
    0x47507eb0UL, 0xffec19d5UL, 0x623b216cUL, 0xda874609UL, 0xc832e9e7UL,
    0x708e8e82UL, 0x28ed9ed4UL, 0x9051f9b1UL, 0x82e4565fUL, 0x3a58313aUL,
    0xa78f0983UL, 0x1f336ee6UL, 0x0d86c108UL, 0xb53aa66dUL, 0xbd40e1a4UL,
    0x05fc86c1UL, 0x1749292fUL, 0xaff54e4aUL, 0x322276f3UL, 0x8a9e1196UL,
    0x982bbe78UL, 0x2097d91dUL, 0x78f4c94bUL, 0xc048ae2eUL, 0xd2fd01c0UL,
    0x6a4166a5UL, 0xf7965e1cUL, 0x4f2a3979UL, 0x5d9f9697UL, 0xe523f1f2UL,
    0x4d6b1905UL, 0xf5d77e60UL, 0xe762d18eUL, 0x5fdeb6ebUL, 0xc2098e52UL,
    0x7ab5e937UL, 0x680046d9UL, 0xd0bc21bcUL, 0x88df31eaUL, 0x3063568fUL,
    0x22d6f961UL, 0x9a6a9e04UL, 0x07bda6bdUL, 0xbf01c1d8UL, 0xadb46e36UL,
    0x15080953UL, 0x1d724e9aUL, 0xa5ce29ffUL, 0xb77b8611UL, 0x0fc7e174UL,
    0x9210d9cdUL, 0x2aacbea8UL, 0x38191146UL, 0x80a57623UL, 0xd8c66675UL,
    0x607a0110UL, 0x72cfaefeUL, 0xca73c99bUL, 0x57a4f122UL, 0xef189647UL,
    0xfdad39a9UL, 0x45115eccUL, 0x764dee06UL, 0xcef18963UL, 0xdc44268dUL,
    0x64f841e8UL, 0xf92f7951UL, 0x41931e34UL, 0x5326b1daUL, 0xeb9ad6bfUL,
    0xb3f9c6e9UL, 0x0b45a18cUL, 0x19f00e62UL, 0xa14c6907UL, 0x3c9b51beUL,
    0x842736dbUL, 0x96929935UL, 0x2e2efe50UL, 0x2654b999UL, 0x9ee8defcUL,
    0x8c5d7112UL, 0x34e11677UL, 0xa9362eceUL, 0x118a49abUL, 0x033fe645UL,
    0xbb838120UL, 0xe3e09176UL, 0x5b5cf613UL, 0x49e959fdUL, 0xf1553e98UL,
    0x6c820621UL, 0xd43e6144UL, 0xc68bceaaUL, 0x7e37a9cfUL, 0xd67f4138UL,
    0x6ec3265dUL, 0x7c7689b3UL, 0xc4caeed6UL, 0x591dd66fUL, 0xe1a1b10aUL,
    0xf3141ee4UL, 0x4ba87981UL, 0x13cb69d7UL, 0xab770eb2UL, 0xb9c2a15cUL,
    0x017ec639UL, 0x9ca9fe80UL, 0x241599e5UL, 0x36a0360bUL, 0x8e1c516eUL,
    0x866616a7UL, 0x3eda71c2UL, 0x2c6fde2cUL, 0x94d3b949UL, 0x090481f0UL,
    0xb1b8e695UL, 0xa30d497bUL, 0x1bb12e1eUL, 0x43d23e48UL, 0xfb6e592dUL,
    0xe9dbf6c3UL, 0x516791a6UL, 0xccb0a91fUL, 0x740cce7aUL, 0x66b96194UL,
    0xde0506f1UL
  },
  {
    0x00000000UL, 0x96300777UL, 0x2c610eeeUL, 0xba510999UL, 0x19c46d07UL,
    0x8ff46a70UL, 0x35a563e9UL, 0xa395649eUL, 0x3288db0eUL, 0xa4b8dc79UL,
    0x1ee9d5e0UL, 0x88d9d297UL, 0x2b4cb609UL, 0xbd7cb17eUL, 0x072db8e7UL,
    0x911dbf90UL, 0x6410b71dUL, 0xf220b06aUL, 0x4871b9f3UL, 0xde41be84UL,
    0x7dd4da1aUL, 0xebe4dd6dUL, 0x51b5d4f4UL, 0xc785d383UL, 0x56986c13UL,
    0xc0a86b64UL, 0x7af962fdUL, 0xecc9658aUL, 0x4f5c0114UL, 0xd96c0663UL,
    0x633d0ffaUL, 0xf50d088dUL, 0xc8206e3bUL, 0x5e10694cUL, 0xe44160d5UL,
    0x727167a2UL, 0xd1e4033cUL, 0x47d4044bUL, 0xfd850dd2UL, 0x6bb50aa5UL,
    0xfaa8b535UL, 0x6c98b242UL, 0xd6c9bbdbUL, 0x40f9bcacUL, 0xe36cd832UL,
    0x755cdf45UL, 0xcf0dd6dcUL, 0x593dd1abUL, 0xac30d926UL, 0x3a00de51UL,
    0x8051d7c8UL, 0x1661d0bfUL, 0xb5f4b421UL, 0x23c4b356UL, 0x9995bacfUL,
    0x0fa5bdb8UL, 0x9eb80228UL, 0x0888055fUL, 0xb2d90cc6UL, 0x24e90bb1UL,
    0x877c6f2fUL, 0x114c6858UL, 0xab1d61c1UL, 0x3d2d66b6UL, 0x9041dc76UL,
    0x0671db01UL, 0xbc20d298UL, 0x2a10d5efUL, 0x8985b171UL, 0x1fb5b606UL,
    0xa5e4bf9fUL, 0x33d4b8e8UL, 0xa2c90778UL, 0x34f9000fUL, 0x8ea80996UL,
    0x18980ee1UL, 0xbb0d6a7fUL, 0x2d3d6d08UL, 0x976c6491UL, 0x015c63e6UL,
    0xf4516b6bUL, 0x62616c1cUL, 0xd8306585UL, 0x4e0062f2UL, 0xed95066cUL,
    0x7ba5011bUL, 0xc1f40882UL, 0x57c40ff5UL, 0xc6d9b065UL, 0x50e9b712UL,
    0xeab8be8bUL, 0x7c88b9fcUL, 0xdf1ddd62UL, 0x492dda15UL, 0xf37cd38cUL,
    0x654cd4fbUL, 0x5861b24dUL, 0xce51b53aUL, 0x7400bca3UL, 0xe230bbd4UL,
    0x41a5df4aUL, 0xd795d83dUL, 0x6dc4d1a4UL, 0xfbf4d6d3UL, 0x6ae96943UL,
    0xfcd96e34UL, 0x468867adUL, 0xd0b860daUL, 0x732d0444UL, 0xe51d0333UL,
    0x5f4c0aaaUL, 0xc97c0dddUL, 0x3c710550UL, 0xaa410227UL, 0x10100bbeUL,
    0x86200cc9UL, 0x25b56857UL, 0xb3856f20UL, 0x09d466b9UL, 0x9fe461ceUL,
    0x0ef9de5eUL, 0x98c9d929UL, 0x2298d0b0UL, 0xb4a8d7c7UL, 0x173db359UL,
    0x810db42eUL, 0x3b5cbdb7UL, 0xad6cbac0UL, 0x2083b8edUL, 0xb6b3bf9aUL,
    0x0ce2b603UL, 0x9ad2b174UL, 0x3947d5eaUL, 0xaf77d29dUL, 0x1526db04UL,
    0x8316dc73UL, 0x120b63e3UL, 0x843b6494UL, 0x3e6a6d0dUL, 0xa85a6a7aUL,
    0x0bcf0ee4UL, 0x9dff0993UL, 0x27ae000aUL, 0xb19e077dUL, 0x44930ff0UL,
    0xd2a30887UL, 0x68f2011eUL, 0xfec20669UL, 0x5d5762f7UL, 0xcb676580UL,
    0x71366c19UL, 0xe7066b6eUL, 0x761bd4feUL, 0xe02bd389UL, 0x5a7ada10UL,
    0xcc4add67UL, 0x6fdfb9f9UL, 0xf9efbe8eUL, 0x43beb717UL, 0xd58eb060UL,
    0xe8a3d6d6UL, 0x7e93d1a1UL, 0xc4c2d838UL, 0x52f2df4fUL, 0xf167bbd1UL,
    0x6757bca6UL, 0xdd06b53fUL, 0x4b36b248UL, 0xda2b0dd8UL, 0x4c1b0aafUL,
    0xf64a0336UL, 0x607a0441UL, 0xc3ef60dfUL, 0x55df67a8UL, 0xef8e6e31UL,
    0x79be6946UL, 0x8cb361cbUL, 0x1a8366bcUL, 0xa0d26f25UL, 0x36e26852UL,
    0x95770cccUL, 0x03470bbbUL, 0xb9160222UL, 0x2f260555UL, 0xbe3bbac5UL,
    0x280bbdb2UL, 0x925ab42bUL, 0x046ab35cUL, 0xa7ffd7c2UL, 0x31cfd0b5UL,
    0x8b9ed92cUL, 0x1daede5bUL, 0xb0c2649bUL, 0x26f263ecUL, 0x9ca36a75UL,
    0x0a936d02UL, 0xa906099cUL, 0x3f360eebUL, 0x85670772UL, 0x13570005UL,
    0x824abf95UL, 0x147ab8e2UL, 0xae2bb17bUL, 0x381bb60cUL, 0x9b8ed292UL,
    0x0dbed5e5UL, 0xb7efdc7cUL, 0x21dfdb0bUL, 0xd4d2d386UL, 0x42e2d4f1UL,
    0xf8b3dd68UL, 0x6e83da1fUL, 0xcd16be81UL, 0x5b26b9f6UL, 0xe177b06fUL,
    0x7747b718UL, 0xe65a0888UL, 0x706a0fffUL, 0xca3b0666UL, 0x5c0b0111UL,
    0xff9e658fUL, 0x69ae62f8UL, 0xd3ff6b61UL, 0x45cf6c16UL, 0x78e20aa0UL,
    0xeed20dd7UL, 0x5483044eUL, 0xc2b30339UL, 0x612667a7UL, 0xf71660d0UL,
    0x4d476949UL, 0xdb776e3eUL, 0x4a6ad1aeUL, 0xdc5ad6d9UL, 0x660bdf40UL,
    0xf03bd837UL, 0x53aebca9UL, 0xc59ebbdeUL, 0x7fcfb247UL, 0xe9ffb530UL,
    0x1cf2bdbdUL, 0x8ac2bacaUL, 0x3093b353UL, 0xa6a3b424UL, 0x0536d0baUL,
    0x9306d7cdUL, 0x2957de54UL, 0xbf67d923UL, 0x2e7a66b3UL, 0xb84a61c4UL,
    0x021b685dUL, 0x942b6f2aUL, 0x37be0bb4UL, 0xa18e0cc3UL, 0x1bdf055aUL,
    0x8def022dUL
  },
  {
    0x00000000UL, 0x41311b19UL, 0x82623632UL, 0xc3532d2bUL, 0x04c56c64UL,
    0x45f4777dUL, 0x86a75a56UL, 0xc796414fUL, 0x088ad9c8UL, 0x49bbc2d1UL,
    0x8ae8effaUL, 0xcbd9f4e3UL, 0x0c4fb5acUL, 0x4d7eaeb5UL, 0x8e2d839eUL,
    0xcf1c9887UL, 0x5112c24aUL, 0x1023d953UL, 0xd370f478UL, 0x9241ef61UL,
    0x55d7ae2eUL, 0x14e6b537UL, 0xd7b5981cUL, 0x96848305UL, 0x59981b82UL,
    0x18a9009bUL, 0xdbfa2db0UL, 0x9acb36a9UL, 0x5d5d77e6UL, 0x1c6c6cffUL,
    0xdf3f41d4UL, 0x9e0e5acdUL, 0xa2248495UL, 0xe3159f8cUL, 0x2046b2a7UL,
    0x6177a9beUL, 0xa6e1e8f1UL, 0xe7d0f3e8UL, 0x2483dec3UL, 0x65b2c5daUL,
    0xaaae5d5dUL, 0xeb9f4644UL, 0x28cc6b6fUL, 0x69fd7076UL, 0xae6b3139UL,
    0xef5a2a20UL, 0x2c09070bUL, 0x6d381c12UL, 0xf33646dfUL, 0xb2075dc6UL,
    0x715470edUL, 0x30656bf4UL, 0xf7f32abbUL, 0xb6c231a2UL, 0x75911c89UL,
    0x34a00790UL, 0xfbbc9f17UL, 0xba8d840eUL, 0x79dea925UL, 0x38efb23cUL,
    0xff79f373UL, 0xbe48e86aUL, 0x7d1bc541UL, 0x3c2ade58UL, 0x054f79f0UL,
    0x447e62e9UL, 0x872d4fc2UL, 0xc61c54dbUL, 0x018a1594UL, 0x40bb0e8dUL,
    0x83e823a6UL, 0xc2d938bfUL, 0x0dc5a038UL, 0x4cf4bb21UL, 0x8fa7960aUL,
    0xce968d13UL, 0x0900cc5cUL, 0x4831d745UL, 0x8b62fa6eUL, 0xca53e177UL,
    0x545dbbbaUL, 0x156ca0a3UL, 0xd63f8d88UL, 0x970e9691UL, 0x5098d7deUL,
    0x11a9ccc7UL, 0xd2fae1ecUL, 0x93cbfaf5UL, 0x5cd76272UL, 0x1de6796bUL,
    0xdeb55440UL, 0x9f844f59UL, 0x58120e16UL, 0x1923150fUL, 0xda703824UL,
    0x9b41233dUL, 0xa76bfd65UL, 0xe65ae67cUL, 0x2509cb57UL, 0x6438d04eUL,
    0xa3ae9101UL, 0xe29f8a18UL, 0x21cca733UL, 0x60fdbc2aUL, 0xafe124adUL,
    0xeed03fb4UL, 0x2d83129fUL, 0x6cb20986UL, 0xab2448c9UL, 0xea1553d0UL,
    0x29467efbUL, 0x687765e2UL, 0xf6793f2fUL, 0xb7482436UL, 0x741b091dUL,
    0x352a1204UL, 0xf2bc534bUL, 0xb38d4852UL, 0x70de6579UL, 0x31ef7e60UL,
    0xfef3e6e7UL, 0xbfc2fdfeUL, 0x7c91d0d5UL, 0x3da0cbccUL, 0xfa368a83UL,
    0xbb07919aUL, 0x7854bcb1UL, 0x3965a7a8UL, 0x4b98833bUL, 0x0aa99822UL,
    0xc9fab509UL, 0x88cbae10UL, 0x4f5def5fUL, 0x0e6cf446UL, 0xcd3fd96dUL,
    0x8c0ec274UL, 0x43125af3UL, 0x022341eaUL, 0xc1706cc1UL, 0x804177d8UL,
    0x47d73697UL, 0x06e62d8eUL, 0xc5b500a5UL, 0x84841bbcUL, 0x1a8a4171UL,
    0x5bbb5a68UL, 0x98e87743UL, 0xd9d96c5aUL, 0x1e4f2d15UL, 0x5f7e360cUL,
    0x9c2d1b27UL, 0xdd1c003eUL, 0x120098b9UL, 0x533183a0UL, 0x9062ae8bUL,
    0xd153b592UL, 0x16c5f4ddUL, 0x57f4efc4UL, 0x94a7c2efUL, 0xd596d9f6UL,
    0xe9bc07aeUL, 0xa88d1cb7UL, 0x6bde319cUL, 0x2aef2a85UL, 0xed796bcaUL,
    0xac4870d3UL, 0x6f1b5df8UL, 0x2e2a46e1UL, 0xe136de66UL, 0xa007c57fUL,
    0x6354e854UL, 0x2265f34dUL, 0xe5f3b202UL, 0xa4c2a91bUL, 0x67918430UL,
    0x26a09f29UL, 0xb8aec5e4UL, 0xf99fdefdUL, 0x3accf3d6UL, 0x7bfde8cfUL,
    0xbc6ba980UL, 0xfd5ab299UL, 0x3e099fb2UL, 0x7f3884abUL, 0xb0241c2cUL,
    0xf1150735UL, 0x32462a1eUL, 0x73773107UL, 0xb4e17048UL, 0xf5d06b51UL,
    0x3683467aUL, 0x77b25d63UL, 0x4ed7facbUL, 0x0fe6e1d2UL, 0xccb5ccf9UL,
    0x8d84d7e0UL, 0x4a1296afUL, 0x0b238db6UL, 0xc870a09dUL, 0x8941bb84UL,
    0x465d2303UL, 0x076c381aUL, 0xc43f1531UL, 0x850e0e28UL, 0x42984f67UL,
    0x03a9547eUL, 0xc0fa7955UL, 0x81cb624cUL, 0x1fc53881UL, 0x5ef42398UL,
    0x9da70eb3UL, 0xdc9615aaUL, 0x1b0054e5UL, 0x5a314ffcUL, 0x996262d7UL,
    0xd85379ceUL, 0x174fe149UL, 0x567efa50UL, 0x952dd77bUL, 0xd41ccc62UL,
    0x138a8d2dUL, 0x52bb9634UL, 0x91e8bb1fUL, 0xd0d9a006UL, 0xecf37e5eUL,
    0xadc26547UL, 0x6e91486cUL, 0x2fa05375UL, 0xe836123aUL, 0xa9070923UL,
    0x6a542408UL, 0x2b653f11UL, 0xe479a796UL, 0xa548bc8fUL, 0x661b91a4UL,
    0x272a8abdUL, 0xe0bccbf2UL, 0xa18dd0ebUL, 0x62defdc0UL, 0x23efe6d9UL,
    0xbde1bc14UL, 0xfcd0a70dUL, 0x3f838a26UL, 0x7eb2913fUL, 0xb924d070UL,
    0xf815cb69UL, 0x3b46e642UL, 0x7a77fd5bUL, 0xb56b65dcUL, 0xf45a7ec5UL,
    0x370953eeUL, 0x763848f7UL, 0xb1ae09b8UL, 0xf09f12a1UL, 0x33cc3f8aUL,
    0x72fd2493UL
  },
  {
    0x00000000UL, 0x376ac201UL, 0x6ed48403UL, 0x59be4602UL, 0xdca80907UL,
    0xebc2cb06UL, 0xb27c8d04UL, 0x85164f05UL, 0xb851130eUL, 0x8f3bd10fUL,
    0xd685970dUL, 0xe1ef550cUL, 0x64f91a09UL, 0x5393d808UL, 0x0a2d9e0aUL,
    0x3d475c0bUL, 0x70a3261cUL, 0x47c9e41dUL, 0x1e77a21fUL, 0x291d601eUL,
    0xac0b2f1bUL, 0x9b61ed1aUL, 0xc2dfab18UL, 0xf5b56919UL, 0xc8f23512UL,
    0xff98f713UL, 0xa626b111UL, 0x914c7310UL, 0x145a3c15UL, 0x2330fe14UL,
    0x7a8eb816UL, 0x4de47a17UL, 0xe0464d38UL, 0xd72c8f39UL, 0x8e92c93bUL,
    0xb9f80b3aUL, 0x3cee443fUL, 0x0b84863eUL, 0x523ac03cUL, 0x6550023dUL,
    0x58175e36UL, 0x6f7d9c37UL, 0x36c3da35UL, 0x01a91834UL, 0x84bf5731UL,
    0xb3d59530UL, 0xea6bd332UL, 0xdd011133UL, 0x90e56b24UL, 0xa78fa925UL,
    0xfe31ef27UL, 0xc95b2d26UL, 0x4c4d6223UL, 0x7b27a022UL, 0x2299e620UL,
    0x15f32421UL, 0x28b4782aUL, 0x1fdeba2bUL, 0x4660fc29UL, 0x710a3e28UL,
    0xf41c712dUL, 0xc376b32cUL, 0x9ac8f52eUL, 0xada2372fUL, 0xc08d9a70UL,
    0xf7e75871UL, 0xae591e73UL, 0x9933dc72UL, 0x1c259377UL, 0x2b4f5176UL,
    0x72f11774UL, 0x459bd575UL, 0x78dc897eUL, 0x4fb64b7fUL, 0x16080d7dUL,
    0x2162cf7cUL, 0xa4748079UL, 0x931e4278UL, 0xcaa0047aUL, 0xfdcac67bUL,
    0xb02ebc6cUL, 0x87447e6dUL, 0xdefa386fUL, 0xe990fa6eUL, 0x6c86b56bUL,
    0x5bec776aUL, 0x02523168UL, 0x3538f369UL, 0x087faf62UL, 0x3f156d63UL,
    0x66ab2b61UL, 0x51c1e960UL, 0xd4d7a665UL, 0xe3bd6464UL, 0xba032266UL,
    0x8d69e067UL, 0x20cbd748UL, 0x17a11549UL, 0x4e1f534bUL, 0x7975914aUL,
    0xfc63de4fUL, 0xcb091c4eUL, 0x92b75a4cUL, 0xa5dd984dUL, 0x989ac446UL,
    0xaff00647UL, 0xf64e4045UL, 0xc1248244UL, 0x4432cd41UL, 0x73580f40UL,
    0x2ae64942UL, 0x1d8c8b43UL, 0x5068f154UL, 0x67023355UL, 0x3ebc7557UL,
    0x09d6b756UL, 0x8cc0f853UL, 0xbbaa3a52UL, 0xe2147c50UL, 0xd57ebe51UL,
    0xe839e25aUL, 0xdf53205bUL, 0x86ed6659UL, 0xb187a458UL, 0x3491eb5dUL,
    0x03fb295cUL, 0x5a456f5eUL, 0x6d2fad5fUL, 0x801b35e1UL, 0xb771f7e0UL,
    0xeecfb1e2UL, 0xd9a573e3UL, 0x5cb33ce6UL, 0x6bd9fee7UL, 0x3267b8e5UL,
    0x050d7ae4UL, 0x384a26efUL, 0x0f20e4eeUL, 0x569ea2ecUL, 0x61f460edUL,
    0xe4e22fe8UL, 0xd388ede9UL, 0x8a36abebUL, 0xbd5c69eaUL, 0xf0b813fdUL,
    0xc7d2d1fcUL, 0x9e6c97feUL, 0xa90655ffUL, 0x2c101afaUL, 0x1b7ad8fbUL,
    0x42c49ef9UL, 0x75ae5cf8UL, 0x48e900f3UL, 0x7f83c2f2UL, 0x263d84f0UL,
    0x115746f1UL, 0x944109f4UL, 0xa32bcbf5UL, 0xfa958df7UL, 0xcdff4ff6UL,
    0x605d78d9UL, 0x5737bad8UL, 0x0e89fcdaUL, 0x39e33edbUL, 0xbcf571deUL,
    0x8b9fb3dfUL, 0xd221f5ddUL, 0xe54b37dcUL, 0xd80c6bd7UL, 0xef66a9d6UL,
    0xb6d8efd4UL, 0x81b22dd5UL, 0x04a462d0UL, 0x33cea0d1UL, 0x6a70e6d3UL,
    0x5d1a24d2UL, 0x10fe5ec5UL, 0x27949cc4UL, 0x7e2adac6UL, 0x494018c7UL,
    0xcc5657c2UL, 0xfb3c95c3UL, 0xa282d3c1UL, 0x95e811c0UL, 0xa8af4dcbUL,
    0x9fc58fcaUL, 0xc67bc9c8UL, 0xf1110bc9UL, 0x740744ccUL, 0x436d86cdUL,
    0x1ad3c0cfUL, 0x2db902ceUL, 0x4096af91UL, 0x77fc6d90UL, 0x2e422b92UL,
    0x1928e993UL, 0x9c3ea696UL, 0xab546497UL, 0xf2ea2295UL, 0xc580e094UL,
    0xf8c7bc9fUL, 0xcfad7e9eUL, 0x9613389cUL, 0xa179fa9dUL, 0x246fb598UL,
    0x13057799UL, 0x4abb319bUL, 0x7dd1f39aUL, 0x3035898dUL, 0x075f4b8cUL,
    0x5ee10d8eUL, 0x698bcf8fUL, 0xec9d808aUL, 0xdbf7428bUL, 0x82490489UL,
    0xb523c688UL, 0x88649a83UL, 0xbf0e5882UL, 0xe6b01e80UL, 0xd1dadc81UL,
    0x54cc9384UL, 0x63a65185UL, 0x3a181787UL, 0x0d72d586UL, 0xa0d0e2a9UL,
    0x97ba20a8UL, 0xce0466aaUL, 0xf96ea4abUL, 0x7c78ebaeUL, 0x4b1229afUL,
    0x12ac6fadUL, 0x25c6adacUL, 0x1881f1a7UL, 0x2feb33a6UL, 0x765575a4UL,
    0x413fb7a5UL, 0xc429f8a0UL, 0xf3433aa1UL, 0xaafd7ca3UL, 0x9d97bea2UL,
    0xd073c4b5UL, 0xe71906b4UL, 0xbea740b6UL, 0x89cd82b7UL, 0x0cdbcdb2UL,
    0x3bb10fb3UL, 0x620f49b1UL, 0x55658bb0UL, 0x6822d7bbUL, 0x5f4815baUL,
    0x06f653b8UL, 0x319c91b9UL, 0xb48adebcUL, 0x83e01cbdUL, 0xda5e5abfUL,
    0xed3498beUL
  },
  {
    0x00000000UL, 0x6567bcb8UL, 0x8bc809aaUL, 0xeeafb512UL, 0x5797628fUL,
    0x32f0de37UL, 0xdc5f6b25UL, 0xb938d79dUL, 0xef28b4c5UL, 0x8a4f087dUL,
    0x64e0bd6fUL, 0x018701d7UL, 0xb8bfd64aUL, 0xddd86af2UL, 0x3377dfe0UL,
    0x56106358UL, 0x9f571950UL, 0xfa30a5e8UL, 0x149f10faUL, 0x71f8ac42UL,
    0xc8c07bdfUL, 0xada7c767UL, 0x43087275UL, 0x266fcecdUL, 0x707fad95UL,
    0x1518112dUL, 0xfbb7a43fUL, 0x9ed01887UL, 0x27e8cf1aUL, 0x428f73a2UL,
    0xac20c6b0UL, 0xc9477a08UL, 0x3eaf32a0UL, 0x5bc88e18UL, 0xb5673b0aUL,
    0xd00087b2UL, 0x6938502fUL, 0x0c5fec97UL, 0xe2f05985UL, 0x8797e53dUL,
    0xd1878665UL, 0xb4e03addUL, 0x5a4f8fcfUL, 0x3f283377UL, 0x8610e4eaUL,
    0xe3775852UL, 0x0dd8ed40UL, 0x68bf51f8UL, 0xa1f82bf0UL, 0xc49f9748UL,
    0x2a30225aUL, 0x4f579ee2UL, 0xf66f497fUL, 0x9308f5c7UL, 0x7da740d5UL,
    0x18c0fc6dUL, 0x4ed09f35UL, 0x2bb7238dUL, 0xc518969fUL, 0xa07f2a27UL,
    0x1947fdbaUL, 0x7c204102UL, 0x928ff410UL, 0xf7e848a8UL, 0x3d58149bUL,
    0x583fa823UL, 0xb6901d31UL, 0xd3f7a189UL, 0x6acf7614UL, 0x0fa8caacUL,
    0xe1077fbeUL, 0x8460c306UL, 0xd270a05eUL, 0xb7171ce6UL, 0x59b8a9f4UL,
    0x3cdf154cUL, 0x85e7c2d1UL, 0xe0807e69UL, 0x0e2fcb7bUL, 0x6b4877c3UL,
    0xa20f0dcbUL, 0xc768b173UL, 0x29c70461UL, 0x4ca0b8d9UL, 0xf5986f44UL,
    0x90ffd3fcUL, 0x7e5066eeUL, 0x1b37da56UL, 0x4d27b90eUL, 0x284005b6UL,
    0xc6efb0a4UL, 0xa3880c1cUL, 0x1ab0db81UL, 0x7fd76739UL, 0x9178d22bUL,
    0xf41f6e93UL, 0x03f7263bUL, 0x66909a83UL, 0x883f2f91UL, 0xed589329UL,
    0x546044b4UL, 0x3107f80cUL, 0xdfa84d1eUL, 0xbacff1a6UL, 0xecdf92feUL,
    0x89b82e46UL, 0x67179b54UL, 0x027027ecUL, 0xbb48f071UL, 0xde2f4cc9UL,
    0x3080f9dbUL, 0x55e74563UL, 0x9ca03f6bUL, 0xf9c783d3UL, 0x176836c1UL,
    0x720f8a79UL, 0xcb375de4UL, 0xae50e15cUL, 0x40ff544eUL, 0x2598e8f6UL,
    0x73888baeUL, 0x16ef3716UL, 0xf8408204UL, 0x9d273ebcUL, 0x241fe921UL,
    0x41785599UL, 0xafd7e08bUL, 0xcab05c33UL, 0x3bb659edUL, 0x5ed1e555UL,
    0xb07e5047UL, 0xd519ecffUL, 0x6c213b62UL, 0x094687daUL, 0xe7e932c8UL,
    0x828e8e70UL, 0xd49eed28UL, 0xb1f95190UL, 0x5f56e482UL, 0x3a31583aUL,
    0x83098fa7UL, 0xe66e331fUL, 0x08c1860dUL, 0x6da63ab5UL, 0xa4e140bdUL,
    0xc186fc05UL, 0x2f294917UL, 0x4a4ef5afUL, 0xf3762232UL, 0x96119e8aUL,
    0x78be2b98UL, 0x1dd99720UL, 0x4bc9f478UL, 0x2eae48c0UL, 0xc001fdd2UL,
    0xa566416aUL, 0x1c5e96f7UL, 0x79392a4fUL, 0x97969f5dUL, 0xf2f123e5UL,
    0x05196b4dUL, 0x607ed7f5UL, 0x8ed162e7UL, 0xebb6de5fUL, 0x528e09c2UL,
    0x37e9b57aUL, 0xd9460068UL, 0xbc21bcd0UL, 0xea31df88UL, 0x8f566330UL,
    0x61f9d622UL, 0x049e6a9aUL, 0xbda6bd07UL, 0xd8c101bfUL, 0x366eb4adUL,
    0x53090815UL, 0x9a4e721dUL, 0xff29cea5UL, 0x11867bb7UL, 0x74e1c70fUL,
    0xcdd91092UL, 0xa8beac2aUL, 0x46111938UL, 0x2376a580UL, 0x7566c6d8UL,
    0x10017a60UL, 0xfeaecf72UL, 0x9bc973caUL, 0x22f1a457UL, 0x479618efUL,
    0xa939adfdUL, 0xcc5e1145UL, 0x06ee4d76UL, 0x6389f1ceUL, 0x8d2644dcUL,
    0xe841f864UL, 0x51792ff9UL, 0x341e9341UL, 0xdab12653UL, 0xbfd69aebUL,
    0xe9c6f9b3UL, 0x8ca1450bUL, 0x620ef019UL, 0x07694ca1UL, 0xbe519b3cUL,
    0xdb362784UL, 0x35999296UL, 0x50fe2e2eUL, 0x99b95426UL, 0xfcdee89eUL,
    0x12715d8cUL, 0x7716e134UL, 0xce2e36a9UL, 0xab498a11UL, 0x45e63f03UL,
    0x208183bbUL, 0x7691e0e3UL, 0x13f65c5bUL, 0xfd59e949UL, 0x983e55f1UL,
    0x2106826cUL, 0x44613ed4UL, 0xaace8bc6UL, 0xcfa9377eUL, 0x38417fd6UL,
    0x5d26c36eUL, 0xb389767cUL, 0xd6eecac4UL, 0x6fd61d59UL, 0x0ab1a1e1UL,
    0xe41e14f3UL, 0x8179a84bUL, 0xd769cb13UL, 0xb20e77abUL, 0x5ca1c2b9UL,
    0x39c67e01UL, 0x80fea99cUL, 0xe5991524UL, 0x0b36a036UL, 0x6e511c8eUL,
    0xa7166686UL, 0xc271da3eUL, 0x2cde6f2cUL, 0x49b9d394UL, 0xf0810409UL,
    0x95e6b8b1UL, 0x7b490da3UL, 0x1e2eb11bUL, 0x483ed243UL, 0x2d596efbUL,
    0xc3f6dbe9UL, 0xa6916751UL, 0x1fa9b0ccUL, 0x7ace0c74UL, 0x9461b966UL,
    0xf10605deUL
#endif
  }
};
#endif /* DYNAMIC_CRC_TABLE */

/* ========================================================================= */
#define DO1 crc = crc_table[0][((int)crc ^ (*buf++)) & 0xff] ^ (crc >> 8)
#define DO8 DO1; DO1; DO1; DO1; DO1; DO1; DO1; DO1

/* ========================================================================= */
unsigned long ZEXPORT crc32(
	unsigned long crc,
	const unsigned char *buf,
	unsigned len)
{
    if (buf == Z_NULL) return 0UL;

#ifdef DYNAMIC_CRC_TABLE
    if (crc_table_empty)
	make_crc_table();
#endif /* DYNAMIC_CRC_TABLE */

#ifdef BYFOUR
    if (sizeof(void *) == sizeof(ptrdiff_t)) {
	u4 endian;

	endian = 1;
	if (*((unsigned char *)(&endian)))
	    return crc32_little(crc, buf, len);
	else
	    return crc32_big(crc, buf, len);
    }
#endif /* BYFOUR */
/*    crc = crc ^ 0xffffffffUL;*/
    while (len >= 8) {
	DO8;
	len -= 8;
    }
    if (len) do {
	DO1;
    } while (--len);
    return crc /* ^ 0xffffffffUL*/;
}

#ifdef BYFOUR

/* ========================================================================= */
#define DOLIT4 c ^= *buf4++; \
	c = crc_table[3][c & 0xff] ^ crc_table[2][(c >> 8) & 0xff] ^ \
	    crc_table[1][(c >> 16) & 0xff] ^ crc_table[0][c >> 24]
#define DOLIT32 DOLIT4; DOLIT4; DOLIT4; DOLIT4; DOLIT4; DOLIT4; DOLIT4; DOLIT4

/* ========================================================================= */
local unsigned long crc32_little(crc, buf, len)
    unsigned long crc;
    const unsigned char *buf;
    unsigned len;
{
    u4 c;
    const u4 *buf4;

    c = (u4)crc;
    c = ~c;
    while (len && ((ptrdiff_t)buf & 3)) {
	c = crc_table[0][(c ^ *buf++) & 0xff] ^ (c >> 8);
	len--;
    }

    buf4 = (const u4 *)buf;
    while (len >= 32) {
	DOLIT32;
	len -= 32;
    }
    while (len >= 4) {
	DOLIT4;
	len -= 4;
    }
    buf = (const unsigned char *)buf4;

    if (len) do {
	c = crc_table[0][(c ^ *buf++) & 0xff] ^ (c >> 8);
    } while (--len);
    c = ~c;
    return (unsigned long)c;
}

/* ========================================================================= */
#define DOBIG4 c ^= *++buf4; \
	c = crc_table[4][c & 0xff] ^ crc_table[5][(c >> 8) & 0xff] ^ \
	    crc_table[6][(c >> 16) & 0xff] ^ crc_table[7][c >> 24]
#define DOBIG32 DOBIG4; DOBIG4; DOBIG4; DOBIG4; DOBIG4; DOBIG4; DOBIG4; DOBIG4

/* ========================================================================= */
local unsigned long crc32_big(crc, buf, len)
    unsigned long crc;
    const unsigned char *buf;
    unsigned len;
{
    u4 c;
    const u4 *buf4;

    c = REV((u4)crc);
    c = ~c;
    while (len && ((ptrdiff_t)buf & 3)) {
	c = crc_table[4][(c >> 24) ^ *buf++] ^ (c << 8);
	len--;
    }

    buf4 = (const u4 *)buf;
    buf4--;
    while (len >= 32) {
	DOBIG32;
	len -= 32;
    }
    while (len >= 4) {
	DOBIG4;
	len -= 4;
    }
    buf4++;
    buf = (const unsigned char *)buf4;

    if (len) do {
	c = crc_table[4][(c >> 24) ^ *buf++] ^ (c << 8);
    } while (--len);
    c = ~c;
    return (unsigned long)(REV(c));
}

#endif /* BYFOUR */

static int scsi_get_serial(int fd, void *buf, size_t buf_len)
{
	unsigned char inq_cmd[] = {INQUIRY, 1, 0x80, 0, buf_len, 0};
	unsigned char sense[32];
	struct sg_io_hdr io_hdr;

	memset(&io_hdr, 0, sizeof(io_hdr));
	io_hdr.interface_id = 'S';
	io_hdr.cmdp = inq_cmd;
	io_hdr.cmd_len = sizeof(inq_cmd);
	io_hdr.dxferp = buf;
	io_hdr.dxfer_len = buf_len;
	io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
	io_hdr.sbp = sense;
	io_hdr.mx_sb_len = sizeof(sense);
	io_hdr.timeout = 5000;

	return ioctl(fd, SG_IO, &io_hdr);
}

/*
 * The /proc/mdstat file comes in at least 3 flavours:
 * In an unpatched 2.2 kernel (md 0.36.6):
 *  Personalities : [n raidx] ...
 *  read_ahead {not set|%d sectors}
 *  md0 : {in}active{ raidX /dev/hda...  %d blocks{ maxfault=%d}}
 *  md1 : .....
 *
 * Normally only 4 md lines, but all are listed.
 *
 * In a patched 2.2 kernel (md 0.90.0)
 *  Personalities : [raidx] ...
 *  read_ahead {not set|%d sectors}
 *  mdN : {in}active {(readonly)} raidX dev[%d]{(F)} ... %d blocks STATUS RESYNC
 *  ... Only initialised arrays listed
 *  unused devices: {dev dev ... | <none>}
 *
 * STATUS is personality dependant:
 *    linear:  %dk rounding
 *    raid0:   %dk chunks
 *    raid1:   [%d/%d] [U_U]   ( raid/working.  operational or not)
 *    raid5:   level 4/5, %dk chunk, algorithm %d [%d/%d] [U_U]
 *
 * RESYNC is  empty or:
 *    {resync|recovery}=%u%% finish=%u.%umin
 *  or
 *    resync=DELAYED
 *
 * In a 2.4 kernel (md 0.90.0/2.4)
 *  Personalities : [raidX] ...
 *  read_ahead {not set|%d sectors}
 *  mdN : {in}active {(read-only)} raidX dev[%d]{(F)} ...
 *       %d blocks STATUS
 *       RESYNC
 *  unused devices: {dev dev .. | <none>}
 *
 *  STATUS matches 0.90.0/2.2
 *  RESYNC includes [===>....],
 *          adds a space after {resync|recovery} and before and after '='
 *          adds a decimal to the recovery percent.
 *          adds (%d/%d) resync amount  and max_blocks, before finish.
 *          adds speed=%dK/sec after finish
 *
 *
 *
 * Out of this we want to extract:
 *   list of devices, active or not
 *   pattern of failed drives (so need number of drives)
 *   percent resync complete
 *
 * As continuation is indicated by leading space, we use
 *  conf_line from config.c to read logical lines
 *
 */


static void free_member_devnames(struct dev_member *m)
{
	while(m) {
		struct dev_member *t = m;

		m = m->next;
		free(t->name);
		free(t);
	}
}

static int add_member_devname(struct dev_member **m, char *name)
{
	struct dev_member *new;
	char *t;

	if ((t = strchr(name, '[')) == NULL)
		/* not a device */
		return 0;

	new = xmalloc(sizeof(*new));
	new->name = strndup(name, t - name);
	new->next = *m;
	*m = new;
	return 1;
}

void free_mdstat(struct mdstat_ent *ms)
{
	while (ms) {
		struct mdstat_ent *t;
		free(ms->dev);
		free(ms->level);
		free(ms->pattern);
		free(ms->metadata_version);
		free_member_devnames(ms->members);
		t = ms;
		ms = ms->next;
		free(t);
	}
}

struct mdstat_ent *mdstat_read(int hold, int start)
{
	FILE *f;
	struct mdstat_ent *all, *rv, **end, **insert_here;
	char *line;
	int fd;

	if (hold && mdstat_fd != -1) {
		lseek(mdstat_fd, 0L, 0);
		fd = dup(mdstat_fd);
		if (fd >= 0)
			f = fdopen(fd, "r");
		else
			return NULL;
	} else
		f = fopen("/proc/mdstat", "r");
	if (f == NULL)
		return NULL;
	else
		bb_cloexec(fileno(f));

	all = NULL;
	end = &all;
	for (; (line = conf_line(f)) ; free_line(line)) {
		struct mdstat_ent *ent;
		char *w;
		int devnum;
		int in_devs = 0;
		char *ep;

		if (strcmp(line, "Personalities")==0)
			continue;
		if (strcmp(line, "read_ahead")==0)
			continue;
		if (strcmp(line, "unused")==0)
			continue;
		insert_here = NULL;
		/* Better be an md line.. */
		if (strncmp(line, "md", 2)!= 0)
			continue;
		if (strncmp(line, "md_d", 4) == 0)
			devnum = -1-strtoul(line+4, &ep, 10);
		else if (strncmp(line, "md", 2) == 0)
			devnum = strtoul(line+2, &ep, 10);
		else
			continue;
		if (ep == NULL || *ep ) {
			/* bb_error_msg("bad /proc/mdstat line starts: %s", line); */
			continue;
		}

		ent = xmalloc(sizeof(*ent));
		ent->dev = ent->level = ent->pattern= NULL;
		ent->next = NULL;
		ent->percent = RESYNC_NONE;
		ent->active = -1;
		ent->resync = 0;
		ent->metadata_version = NULL;
		ent->raid_disks = 0;
		ent->devcnt = 0;
		ent->members = NULL;

		ent->dev = strdup(line);
		ent->devnum = devnum;

		for (w=dl_next(line); w!= line ; w=dl_next(w)) {
			int l = strlen(w);
			char *eq;
			if (strcmp(w, "active")==0)
				ent->active = 1;
			else if (strcmp(w, "inactive")==0) {
				ent->active = 0;
				in_devs = 1;
			} else if (ent->active > 0 &&
				 ent->level == NULL &&
				 w[0] != '(' /*readonly*/) {
				ent->level = strdup(w);
				in_devs = 1;
			} else if (in_devs && strcmp(w, "blocks")==0)
				in_devs = 0;
			else if (in_devs) {
				ent->devcnt +=
					add_member_devname(&ent->members, w);
				if (strncmp(w, "md", 2)==0) {
					/* This has an md device as a component.
					 * If that device is already in the
					 * list, make sure we insert before
					 * there.
					 */
					struct mdstat_ent **ih;
					int dn2 = devname2devnum(w);
					ih = &all;
					while (ih != insert_here && *ih &&
					       (*ih)->devnum != dn2)
						ih = & (*ih)->next;
					insert_here = ih;
				}
			} else if (strcmp(w, "super") == 0 &&
				   dl_next(w) != line) {
				w = dl_next(w);
				ent->metadata_version = strdup(w);
			} else if (w[0] == '[' && isdigit(w[1])) {
				ent->raid_disks = atoi(w+1);
			} else if (!ent->pattern &&
				 w[0] == '[' &&
				 (w[1] == 'U' || w[1] == '_')) {
				ent->pattern = strdup(w+1);
				if (ent->pattern[l-2]==']')
					ent->pattern[l-2] = '\0';
			} else if (ent->percent == RESYNC_NONE &&
				   strncmp(w, "re", 2)== 0 &&
				   w[l-1] == '%' &&
				   (eq=strchr(w, '=')) != NULL ) {
				ent->percent = atoi(eq+1);
				if (strncmp(w,"resync", 6)==0)
					ent->resync = 1;
				else if (strncmp(w, "reshape", 7)==0)
					ent->resync = 2;
				else
					ent->resync = 0;
			} else if (ent->percent == RESYNC_NONE &&
				   (w[0] == 'r' || w[0] == 'c')) {
				if (strncmp(w, "resync", 4)==0)
					ent->resync = 1;
				if (strncmp(w, "reshape", 7)==0)
					ent->resync = 2;
				if (strncmp(w, "recovery", 8)==0)
					ent->resync = 0;
				if (strncmp(w, "check", 5)==0)
					ent->resync = 3;

				if (l > 8 && strcmp(w+l-8, "=DELAYED") == 0)
					ent->percent = RESYNC_DELAYED;
				if (l > 8 && strcmp(w+l-8, "=PENDING") == 0)
					ent->percent = RESYNC_PENDING;
			} else if (ent->percent == RESYNC_NONE &&
				   w[0] >= '0' &&
				   w[0] <= '9' &&
				   w[l-1] == '%') {
				ent->percent = atoi(w);
			}
		}
		if (insert_here && (*insert_here)) {
			ent->next = *insert_here;
			*insert_here = ent;
		} else {
			*end = ent;
			end = &ent->next;
		}
	}
	if (hold && mdstat_fd == -1) {
		mdstat_fd = dup(fileno(f));
		bb_cloexec(mdstat_fd);
	}
	fclose(f);

	/* If we might want to start array,
	 * reverse the order, so that components comes before composites
	 */
	if (start) {
		rv = NULL;
		while (all) {
			struct mdstat_ent *e = all;
			all = all->next;
			e->next = rv;
			rv = e;
		}
	} else rv = all;
	return rv;
}

int mddev_busy(int devnum)
{
	struct mdstat_ent *mdstat = mdstat_read(0, 0);
	struct mdstat_ent *me;

	for (me = mdstat ; me ; me = me->next)
		if (me->devnum == devnum)
			break;
	free_mdstat(mdstat);
	return me != NULL;
}

/*
 * Intel(R) Matrix Storage Manager hardware and firmware support routines
 */


static int devpath_to_ll(const char *dev_path, const char *entry,
			 unsigned long long *val);

static __u16 devpath_to_vendor(const char *dev_path);


static struct sys_dev *find_driver_devices(const char *bus, const char *driver)
{
	/* search sysfs for devices driven by 'driver' */
	char path[292];
	char sys_link[256];
	char *c;
	DIR *driver_dir;
	struct dirent *de;
	struct sys_dev *head = NULL;
	struct sys_dev *list = NULL;
	enum sys_dev_type type;
	unsigned long long dev_id;

	if (strcmp(driver, "isci") == 0)
		type = SYS_DEV_SAS;
	else if (strcmp(driver, "ahci") == 0)
		type = SYS_DEV_SATA;
	else
		type = SYS_DEV_UNKNOWN;

	sprintf(path, "/sys/bus/%s/drivers/%s", bus, driver);
	driver_dir = opendir(path);
	if (!driver_dir)
		return NULL;
	for (de = readdir(driver_dir); de; de = readdir(driver_dir)) {
		int n;

		/* is 'de' a device? check that the 'subsystem' link exists and
		 * that its target matches 'bus'
		 */
		sprintf(path, "/sys/bus/%s/drivers/%s/%s/subsystem",
			bus, driver, de->d_name);
		n = readlink(path, sys_link, sizeof(sys_link));
		if (n < 0 || n >= (int)sizeof(sys_link))
			continue;
		sys_link[n] = '\0';
		c = strrchr(sys_link, '/');
		if (!c)
			continue;
		if (strncmp(bus, c+1, strlen(bus)) != 0)
			continue;

		sprintf(path, "/sys/bus/%s/drivers/%s/%s",
			bus, driver, de->d_name);

		/* if it's not Intel device skip it. */
		if (devpath_to_vendor(path) != 0x8086)
			continue;

		if (devpath_to_ll(path, "device", &dev_id) != 0)
			continue;

		/* start / add list entry */
		if (!head) {
			head = xmalloc(sizeof(*head));
			list = head;
		} else {
			list->next = xmalloc(sizeof(*head));
			list = list->next;
		}

		if (!list) {
			free_sys_dev(&head);
			break;
		}

		list->dev_id = (__u16) dev_id;
		list->type = type;
		list->path = bb_simplify_path(path);
		list->next = NULL;
		if ((list->pci_id = strrchr(list->path, '/')) != NULL)
			list->pci_id++;
	}
	closedir(driver_dir);
	return head;
}


static enum sys_dev_type device_type_by_id(__u16 device_id)
{
	struct sys_dev *iter;

	for(iter = intel_devices; iter != NULL; iter = iter->next)
		if (iter->dev_id == device_id)
			return iter->type;
	return SYS_DEV_UNKNOWN;
}

static int devpath_to_ll(const char *dev_path, const char *entry, unsigned long long *val)
{
	char path[strlen(dev_path) + strlen(entry) + 2];
	int fd;
	int n;

	sprintf(path, "%s/%s", dev_path, entry);

	fd = open(path, O_RDONLY);
	if (fd < 0)
		return -1;
	n = sysfs_fd_get_ll(fd, val);
	close(fd);
	return n;
}


static __u16 devpath_to_vendor(const char *dev_path)
{
	char path[strlen(dev_path) + strlen("/vendor") + 1];
	char vendor[7];
	int fd;
	__u16 id = 0xffff;
	int n;

	sprintf(path, "%s/vendor", dev_path);

	fd = open(path, O_RDONLY);
	if (fd < 0)
		return 0xffff;

	n = read(fd, vendor, sizeof(vendor));
	if (n == sizeof(vendor)) {
		vendor[n - 1] = '\0';
		id = strtoul(vendor, NULL, 16);
	}
	close(fd);

	return id;
}

struct sys_dev *find_intel_devices(void)
{
	struct sys_dev *ahci, *isci;

	isci = find_driver_devices("pci", "isci");
	ahci = find_driver_devices("pci", "ahci");

	if (!ahci) {
		ahci = isci;
	} else {
		struct sys_dev *elem = ahci;
		while (elem->next)
			elem = elem->next;
		elem->next = isci;
	}
	return ahci;
}

/*
 * PCI Expansion ROM Data Structure Format */
struct pciExpDataStructFormat {
	__u8  ver[4];
	__u16 vendorID;
	__u16 deviceID;
} __attribute__ ((packed));


static int scan(const void *start, const void *end, const void *data)
{
	int offset;
	const struct imsm_orom *imsm_mem;
	int dev;
	int len = (end - start);
	struct pciExpDataStructFormat *ptr= (struct pciExpDataStructFormat *)data;

	if (data + 0x18 > end) {
		dprintf("cannot find pciExpDataStruct \n");
		return 0;
	}

	dprintf("ptr->vendorID: %lx __le16_to_cpu(ptr->deviceID): %lx \n",
		(unsigned long) __le16_to_cpu(ptr->vendorID),
		(unsigned long) __le16_to_cpu(ptr->deviceID));

	if (__le16_to_cpu(ptr->vendorID) == 0x8086) {
		/* serach  attached intel devices by device id from OROM */
		dev = device_type_by_id(__le16_to_cpu(ptr->deviceID));
		if (dev == SYS_DEV_UNKNOWN)
			return 0;
	}
	else
		return 0;

	for (offset = 0; offset < len; offset += 4) {
		imsm_mem = start + offset;
		if ((memcmp(imsm_mem->signature, "$VER", 4) == 0)) {
			g_imsm_orom[dev] = *imsm_mem;
			populated_orom[dev] = 1;
			return populated_orom[SYS_DEV_SATA] && populated_orom[SYS_DEV_SAS];
		}
	}
	return 0;
}


static const struct imsm_orom *
imsm_platform_test(enum sys_dev_type hba_id, int *populated,
					   struct imsm_orom *imsm_orom)
{
	memset(imsm_orom, 0, sizeof(*imsm_orom));
	imsm_orom->rlc = IMSM_OROM_RLC_RAID0 | IMSM_OROM_RLC_RAID1 |
				IMSM_OROM_RLC_RAID10 | IMSM_OROM_RLC_RAID5;
	imsm_orom->sss = IMSM_OROM_SSS_4kB | IMSM_OROM_SSS_8kB |
				IMSM_OROM_SSS_16kB | IMSM_OROM_SSS_32kB |
				IMSM_OROM_SSS_64kB | IMSM_OROM_SSS_128kB |
				IMSM_OROM_SSS_256kB | IMSM_OROM_SSS_512kB |
				IMSM_OROM_SSS_1MB | IMSM_OROM_SSS_2MB;
	imsm_orom->dpa = IMSM_OROM_DISKS_PER_ARRAY;
	imsm_orom->tds = IMSM_OROM_TOTAL_DISKS;
	imsm_orom->vpa = IMSM_OROM_VOLUMES_PER_ARRAY;
	imsm_orom->vphba = IMSM_OROM_VOLUMES_PER_HBA;
	imsm_orom->attr = imsm_orom->rlc | IMSM_OROM_ATTR_ChecksumVerify;
	*populated = 1;

	if (check_env("IMSM_TEST_OROM_NORAID5")) {
		imsm_orom->rlc = IMSM_OROM_RLC_RAID0 | IMSM_OROM_RLC_RAID1 |
				IMSM_OROM_RLC_RAID10;
	}
	if (check_env("IMSM_TEST_AHCI_EFI_NORAID5") && (hba_id == SYS_DEV_SAS)) {
		imsm_orom->rlc = IMSM_OROM_RLC_RAID0 | IMSM_OROM_RLC_RAID1 |
				IMSM_OROM_RLC_RAID10;
	}
	if (check_env("IMSM_TEST_SCU_EFI_NORAID5") && (hba_id == SYS_DEV_SATA)) {
		imsm_orom->rlc = IMSM_OROM_RLC_RAID0 | IMSM_OROM_RLC_RAID1 |
				IMSM_OROM_RLC_RAID10;
	}

	return imsm_orom;
}


static const struct imsm_orom *find_imsm_hba_orom(enum sys_dev_type hba_id)
{
	unsigned long align;

	if (hba_id >= SYS_DEV_MAX)
		return NULL;

	/* it's static data so we only need to read it once */
	if (populated_orom[hba_id]) {
		dprintf("OROM CAP: %p, pid: %d pop: %d\n",
			&g_imsm_orom[hba_id], (int) getpid(), populated_orom[hba_id]);
		return &g_imsm_orom[hba_id];
	}
	if (check_env("IMSM_TEST_OROM")) {
		dprintf("OROM CAP: %p,  pid: %d pop: %d\n",
		     &g_imsm_orom[hba_id], (int) getpid(), populated_orom[hba_id]);
		return imsm_platform_test(hba_id, &populated_orom[hba_id], &g_imsm_orom[hba_id]);
	}
	/* return empty OROM capabilities in EFI test mode */
	if (check_env("IMSM_TEST_AHCI_EFI") ||
	    check_env("IMSM_TEST_SCU_EFI"))
		return NULL;


	if (intel_devices != NULL)
		free_sys_dev(&intel_devices);

	intel_devices = find_intel_devices();

	if (intel_devices == NULL)
		return NULL;

	/* scan option-rom memory looking for an imsm signature */
	if (check_env("IMSM_SAFE_OROM_SCAN"))
		align = 2048;
	else
		align = 512;
	if (probe_roms_init(align) != 0)
		return NULL;
	probe_roms();
	/* ignore return value - True is returned if both adapater roms are found */
	scan_adapter_roms(scan);
	probe_roms_exit();

	if (intel_devices != NULL)
		free_sys_dev(&intel_devices);
	intel_devices = NULL;

	if (populated_orom[hba_id])
		return &g_imsm_orom[hba_id];
	return NULL;
}

#define GUID_STR_MAX    37  /* according to GUID format:
			     * xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx" */

#define EFI_GUID(a, b, c, d0, d1, d2, d3, d4, d5, d6, d7) \
((struct efi_guid) \
{{ (a) & 0xff, ((a) >> 8) & 0xff, ((a) >> 16) & 0xff, ((a) >> 24) & 0xff, \
  (b) & 0xff, ((b) >> 8) & 0xff, \
  (c) & 0xff, ((c) >> 8) & 0xff, \
  (d0), (d1), (d2), (d3), (d4), (d5), (d6), (d7) }})


#define SYS_EFI_VAR_PATH "/sys/firmware/efi/vars"
#define SCU_PROP "RstScuV"
#define AHCI_PROP "RstSataV"

#define VENDOR_GUID \
	EFI_GUID(0x193dfefa, 0xa445, 0x4302, 0x99, 0xd8, 0xef, 0x3a, 0xad, 0x1a, 0x04, 0xc6)


static int read_efi_variable(void *buffer, ssize_t buf_size, char *variable_name, struct efi_guid guid)
{
	char path[PATH_MAX];
	char buf[GUID_STR_MAX];
	int dfd;
	ssize_t n, var_data_len;

	snprintf(path, PATH_MAX, "%s/%s-%s/size", SYS_EFI_VAR_PATH, variable_name, guid_str(buf, guid));

	dprintf("EFI VAR: path=%s\n", path);
	/* get size of variable data */
	dfd = open(path, O_RDONLY);
	if (dfd < 0)
		return 1;

	n = read(dfd, &buf, sizeof(buf));
	close(dfd);
	if (n < 0)
		return 1;
	buf[n] = '\0';

	errno = 0;
	var_data_len = strtoul(buf, NULL, 16);
	if ((errno == ERANGE && (var_data_len == LONG_MAX))
	    || (errno != 0 && var_data_len == 0))
		return 1;

	/* get data */
	snprintf(path, PATH_MAX, "%s/%s-%s/data", SYS_EFI_VAR_PATH, variable_name, guid_str(buf, guid));

	dprintf("EFI VAR: path=%s\n", path);
	dfd = open(path, O_RDONLY);
	if (dfd < 0)
		return 1;

	n = read(dfd, buffer, buf_size);
	close(dfd);
	if (n != var_data_len || n < buf_size) {
		return 1;
	}

	return 0;
}

static const struct imsm_orom *find_imsm_efi(enum sys_dev_type hba_id)
{
	if (hba_id >= SYS_DEV_MAX)
		return NULL;

	dprintf("EFI CAP: %p,  pid: %d pop: %d\n",
		&imsm_efi[hba_id], (int) getpid(), populated_efi[hba_id]);

	/* it's static data so we only need to read it once */
	if (populated_efi[hba_id]) {
		dprintf("EFI CAP: %p, pid: %d pop: %d\n",
			&imsm_efi[hba_id], (int) getpid(), populated_efi[hba_id]);
		return &imsm_efi[hba_id];
	}
	if (check_env("IMSM_TEST_AHCI_EFI") ||
	    check_env("IMSM_TEST_SCU_EFI")) {
		dprintf("OROM CAP: %p,  pid: %d pop: %d\n",
			&imsm_efi[hba_id], (int) getpid(), populated_efi[hba_id]);
		return imsm_platform_test(hba_id, &populated_efi[hba_id], &imsm_efi[hba_id]);
	}
	/* OROM test is set, return that there is no EFI capabilities */
	if (check_env("IMSM_TEST_OROM"))
		return NULL;

	if (read_efi_variable(&imsm_efi[hba_id], sizeof(imsm_efi[0]), hba_id == SYS_DEV_SAS ? SCU_PROP : AHCI_PROP, VENDOR_GUID)) {
		populated_efi[hba_id] = 0;
		return NULL;
	}

	populated_efi[hba_id] = 1;
	return &imsm_efi[hba_id];
}

/*
 * backward interface compatibility
 */
const struct imsm_orom *find_imsm_capability(enum sys_dev_type hba_id)
{
	const struct imsm_orom *cap=NULL;


	if ((cap = find_imsm_efi(hba_id)) != NULL)
		return cap;
	if ((cap = find_imsm_hba_orom(hba_id)) != NULL)
		return cap;
	return NULL;
}

static char *devt_to_devpath(dev_t dev)
{
	char device[46];

	sprintf(device, "/sys/dev/block/%d:%d/device", major(dev), minor(dev));
	return bb_simplify_path(device);
}

char *diskfd_to_devpath(int fd)
{
	/* return the device path for a disk, return NULL on error or fd
	 * refers to a partition
	 */
	struct stat st;

	if (fstat(fd, &st) != 0)
		return NULL;
	if (!S_ISBLK(st.st_mode))
		return NULL;

	return devt_to_devpath(st.st_rdev);
}

int path_attached_to_hba(const char *disk_dev_path, const char *hba_path)
{
	int rc;

	if (check_env("IMSM_TEST_AHCI_DEV") ||
	    check_env("IMSM_TEST_SCU_DEV")) {
		return 1;
	}

	if (!disk_dev_path || !hba_path)
		return 0;
	dprintf("hba: %s - disk: %s\n", hba_path, disk_dev_path);
	if (strncmp(disk_dev_path, hba_path, strlen(hba_path)) == 0)
		rc = 1;
	else
		rc = 0;

	return rc;
}

/*
 * probe_roms - scan for Adapter ROMS
 *
 * (based on linux-2.6:arch/x86/kernel/probe_roms_32.c)
 *
 * Copyright (C) 2008 Intel Corporation
 */


static const int rom_len = 0xf0000 - 0xc0000; /* option-rom memory region */

#define ARRAY_SIZE(x) (sizeof(x)/sizeof(x[0]))

static void sigbus(int sig)
{
	(void)&sig;
	_sigbus = 1;
}

static int probe_address8(const __u8 *ptr, __u8 *val)
{
	int rc = 0;

	*val = *ptr;
	if (_sigbus)
		rc = -1;
	_sigbus = 0;

	return rc;
}

static int probe_address16(const __u16 *ptr, __u16 *val)
{
	int rc = 0;

	*val = *ptr;
	if (_sigbus)
		rc = -1;
	_sigbus = 0;

	return rc;
}

void probe_roms_exit(void)
{
	signal(SIGBUS, SIG_DFL);
	if (rom_fd >= 0) {
		close(rom_fd);
		rom_fd = -1;
	}
	if (rom_mem != MAP_FAILED) {
		munmap(rom_mem, rom_len);
		rom_mem = MAP_FAILED;
	}
}

int probe_roms_init(unsigned long align)
{
	int fd = -1;
	int rc = 0;

	/* valid values are 2048 and 512.  512 is for PCI-3.0 compliant
	 * systems, or systems that do not have dangerous/legacy ISA
	 * devices.  2048 should always be safe
	 */
	if (align == 512 || align == 2048)
		rom_align = align;
	else
		return -1;

	if (signal(SIGBUS, sigbus) == SIG_ERR)
		rc = -1;
	if (rc == 0) {
		fd = open("/dev/mem", O_RDONLY);
		if (fd < 0)
			rc = -1;
	}
	if (rc == 0) {
		rom_mem = mmap(NULL, rom_len, PROT_READ, MAP_PRIVATE, fd, 0xc0000);
		if (rom_mem == MAP_FAILED)
			rc = -1;
	}

	if (rc == 0)
		rom_fd = fd;
	else {
		if (fd >= 0)
			close(fd);
		probe_roms_exit();
	}
	return rc;
}

/**
 * isa_bus_to_virt - convert physical address to mmap'd region
 * @addr - address to convert
 *
 * Only valid between a successful call to probe_roms_init and the
 * corresponding probe_roms_exit
 */
static void *isa_bus_to_virt(unsigned long addr)
{
	return rom_mem + (addr - 0xc0000);
}


static const struct resource system_rom_resource = {
	.name   = "System ROM",
	.start  = 0xf0000,
	.data   = 0,
	.end    = 0xfffff,
};

static const struct resource extension_rom_resource = {
	.name   = "Extension ROM",
	.start  = 0xe0000,
	.data   = 0,
	.end    = 0xeffff,
};


static const struct resource video_rom_resource_st = {
	.name   = "Video ROM",
	.start  = 0xc0000,
	.data   = 0,
	.end    = 0xc7fff,
};

#define ROMSIGNATURE 0xaa55

static int romsignature(const unsigned char *rom)
{
	const unsigned short * const ptr = (const unsigned short *)rom;
	unsigned short sig = 0;

	return probe_address16(ptr, &sig) == 0 && sig == ROMSIGNATURE;
}

static int romchecksum(const unsigned char *rom, unsigned long length)
{
	unsigned char sum, c;

	for (sum = 0; length && probe_address8(rom++, &c) == 0; length--)
		sum += c;
	return !length && !sum;
}

int scan_adapter_roms(scan_fn fn)
{
	/* let scan_fn examing each of the adapter roms found by probe_roms */
	int found = 0;

	if (rom_fd >= 0) {
		const struct resource *res;

		for (res = adapter_rom_resources; res < (adapter_rom_resources + ARR_SIZE); res++) {
			if (res->start) {
				found = fn(isa_bus_to_virt(res->start),
				   isa_bus_to_virt(res->end),
				   isa_bus_to_virt(res->data));
				if (found)
					break;
			} else
				break;
		}
	}
	return found;
}

static unsigned long align(unsigned long addr, unsigned long alignment)
{
	return (addr + alignment - 1) & ~(alignment - 1);
}

void probe_roms(void)
{
	const void *rom;
	unsigned long start, length, upper;
	unsigned char c;
	unsigned int i;
	__u16 val=0;

	if (rom_fd < 0)
		return;

	/* video rom */
	upper = adapter_rom_resources[0].start;
	for (start = video_rom_resource.start; start < upper; start += rom_align) {
		rom = isa_bus_to_virt(start);
		if (!romsignature(rom))
			continue;

		video_rom_resource.start = start;

		if (probe_address8(rom + 2, &c) != 0)
			continue;

		/* 0 < length <= 0x7f * 512, historically */
		length = c * 512;

		/* if checksum okay, trust length byte */
		if (length && romchecksum(rom, length))
			video_rom_resource.end = start + length - 1;
		break;
	}

	start = align(video_rom_resource.end + 1, rom_align);
	if (start < upper)
		start = upper;

	/* system rom */
	upper = system_rom_resource.start;

	/* check for extension rom (ignore length byte!) */
	rom = isa_bus_to_virt(extension_rom_resource.start);
	if (romsignature(rom)) {
		length = extension_rom_resource.end - extension_rom_resource.start + 1;
		if (romchecksum(rom, length))
			upper = extension_rom_resource.start;
	}

	/* check for adapter roms on 2k boundaries */
	for (i = 0; i < ARR_SIZE && start < upper; start += rom_align) {
		rom = isa_bus_to_virt(start);
		if (!romsignature(rom))
			continue;

		if (probe_address8(rom + 2, &c) != 0)
			continue;

		/* 0 < length <= 0x7f * 512, historically */
		length = c * 512;

		/* Retrieve 16-bit pointer to PCI Data Structure (offset 18h-19h)
		 * The data can be within 64KB forward of the first location
		 * of this code image. The pointer is in little-endian order
		 */

		if (probe_address16(rom + 0x18, &val) != 0)
			continue;
		val = __le16_to_cpu(val);

		/* but accept any length that fits if checksum okay */
		if (!length || start + length > upper || !romchecksum(rom, length))
			continue;

		adapter_rom_resources[i].start = start;
		adapter_rom_resources[i].data = start + (unsigned long) val;
		adapter_rom_resources[i].end = start + length - 1;

		start = adapter_rom_resources[i++].end & ~(rom_align - 1);
	}
}

static int load_sys(char *path, char *buf)
{
	int fd = open(path, O_RDONLY);
	int n;
	if (fd < 0)
		return -1;
	n = read(fd, buf, 1024);
	close(fd);
	if (n <0 || n >= 1024)
		return -1;
	buf[n] = 0;
	if (n && buf[n-1] == '\n')
		buf[n-1] = 0;
	return 0;
}

void sysfs_free(struct mdinfo *sra)
{
	while (sra) {
		struct mdinfo *sra2 = sra->next;
		while (sra->devs) {
			struct mdinfo *d = sra->devs;
			sra->devs = d->next;
			free(d);
		}
		free(sra);
		sra = sra2;
	}
}

static void sysfs_init(struct mdinfo *mdi, int fd, int devnum)
{
	mdi->sys_name[0] = 0;
	if (fd >= 0) {
		mdu_version_t vers;
		if (ioctl(fd, RAID_VERSION, &vers) != 0)
			return;
		devnum = fd2devnum(fd);
	}
	if (devnum == NoMdDev)
		return;
	fmt_devname(mdi->sys_name, devnum);
}


struct mdinfo *sysfs_read(int fd, int devnum, unsigned long options)
{
	char fname[PATH_MAX];
	char buf[PATH_MAX];
	char *base;
	char *dbase;
	struct mdinfo *sra;
	struct mdinfo *dev;
	DIR *dir = NULL;
	struct dirent *de;

	sra = zmalloc(sizeof(*sra));
	sysfs_init(sra, fd, devnum);
	if (sra->sys_name[0] == 0) {
		free(sra);
		return NULL;
	}

	sprintf(fname, "/sys/block/%s/md/", sra->sys_name);
	base = fname + strlen(fname);

	sra->devs = NULL;
	if (options & GET_VERSION) {
		strcpy(base, "metadata_version");
		if (load_sys(fname, buf))
			goto abort;
		if (strncmp(buf, "none", 4) == 0) {
			sra->array.major_version =
				sra->array.minor_version = -1;
			strcpy(sra->text_version, "");
		} else if (strncmp(buf, "external:", 9) == 0) {
			sra->array.major_version = -1;
			sra->array.minor_version = -2;
			strcpy(sra->text_version, buf+9);
		} else {
			sscanf(buf, "%d.%d",
			       &sra->array.major_version,
			       &sra->array.minor_version);
			strcpy(sra->text_version, buf);
		}
	}
	if (options & GET_LEVEL) {
		strcpy(base, "level");
		if (load_sys(fname, buf))
			goto abort;
		sra->array.level = map_name(buf);
	}
	if (options & GET_LAYOUT) {
		strcpy(base, "layout");
		if (load_sys(fname, buf))
			goto abort;
		sra->array.layout = strtoul(buf, NULL, 0);
	}
	if (options & GET_DISKS) {
		strcpy(base, "raid_disks");
		if (load_sys(fname, buf))
			goto abort;
		sra->array.raid_disks = strtoul(buf, NULL, 0);
	}
	if (options & GET_DEGRADED) {
		strcpy(base, "degraded");
		if (load_sys(fname, buf))
			goto abort;
		sra->array.failed_disks = strtoul(buf, NULL, 0);
	}
	if (options & GET_COMPONENT) {
		strcpy(base, "component_size");
		if (load_sys(fname, buf))
			goto abort;
		sra->component_size = strtoull(buf, NULL, 0);
		/* sysfs reports "K", but we want sectors */
		sra->component_size *= 2;
	}
	if (options & GET_CHUNK) {
		strcpy(base, "chunk_size");
		if (load_sys(fname, buf))
			goto abort;
		sra->array.chunk_size = strtoul(buf, NULL, 0);
	}
	if (options & GET_CACHE) {
		strcpy(base, "stripe_cache_size");
		if (load_sys(fname, buf))
			goto abort;
		sra->cache_size = strtoul(buf, NULL, 0);
	}
	if (options & GET_MISMATCH) {
		strcpy(base, "mismatch_cnt");
		if (load_sys(fname, buf))
			goto abort;
		sra->mismatch_cnt = strtoul(buf, NULL, 0);
	}
	if (options & GET_SAFEMODE) {
		int scale = 1;
		int dot = 0;
		unsigned i;
		unsigned long msec;
		size_t len;

		strcpy(base, "safe_mode_delay");
		if (load_sys(fname, buf))
			goto abort;

		/* remove a period, and count digits after it */
		len = strlen(buf);
		for (i = 0; i < len; i++) {
			if (dot) {
				if (isdigit(buf[i])) {
					buf[i-1] = buf[i];
					scale *= 10;
				}
				buf[i] = 0;
			} else if (buf[i] == '.') {
				dot=1;
				buf[i] = 0;
			}
		}
		msec = strtoul(buf, NULL, 10);
		msec = (msec * 1000) / scale;
		sra->safe_mode_delay = msec;
	}
	if (options & GET_BITMAP_LOCATION) {
		strcpy(base, "bitmap/location");
		if (load_sys(fname, buf))
			goto abort;
		if (strncmp(buf, "file", 4) == 0)
			sra->bitmap_offset = 1;
		else if (strncmp(buf, "none", 4) == 0)
			sra->bitmap_offset = 0;
		else if (buf[0] == '+')
			sra->bitmap_offset = bb_xparse_l_number(10 |
						BB_PN_SILENT_RE, buf + 1); /* BB_PN_LOWER_MIN ? */
		else
			goto abort;
	}

	if (! (options & GET_DEVS))
		return sra;

	/* Get all the devices as well */
	*base = 0;
	dir = opendir(fname);
	if (!dir)
		goto abort;
	sra->array.spare_disks = 0;

	while ((de = readdir(dir)) != NULL) {
		char *ep;
		if (de->d_ino == 0 ||
		    strncmp(de->d_name, "dev-", 4) != 0)
			continue;
		strcpy(base, de->d_name);
		dbase = base + strlen(base);
		*dbase++ = '/';

		dev = xmalloc(sizeof(*dev));

		/* Always get slot, major, minor */
		strcpy(dbase, "slot");
		if (load_sys(fname, buf)) {
			/* hmm... unable to read 'slot' maybe the device
			 * is going away?
			 */
			strcpy(dbase, "block");
			if (readlink(fname, buf, sizeof(buf)) < 0 &&
			    errno != ENAMETOOLONG) {
				/* ...yup device is gone */
				free(dev);
				continue;
			} else {
				/* slot is unreadable but 'block' link
				 * still intact... something bad is happening
				 * so abort
				 */
				free(dev);
				goto abort;
			}

		}
		strcpy(dev->sys_name, de->d_name);
		dev->disk.raid_disk = strtoul(buf, &ep, 10);
		if (*ep) dev->disk.raid_disk = -1;

		strcpy(dbase, "block/dev");
		if (load_sys(fname, buf)) {
			/* assume this is a stale reference to a hot
			 * removed device
			 */
			free(dev);
			continue;
		}
		sscanf(buf, "%d:%d", &dev->disk.major, &dev->disk.minor);

		/* special case check for block devices that can go 'offline' */
		strcpy(dbase, "block/device/state");
		if (load_sys(fname, buf) == 0 &&
		    strncmp(buf, "offline", 7) == 0) {
			free(dev);
			continue;
		}

		/* finally add this disk to the array */
		dev->next = sra->devs;
		sra->devs = dev;

		if (options & GET_OFFSET) {
			strcpy(dbase, "offset");
			if (load_sys(fname, buf))
				goto abort;
			dev->data_offset = strtoull(buf, NULL, 0);
		}
		if (options & GET_SIZE) {
			strcpy(dbase, "size");
			if (load_sys(fname, buf))
				goto abort;
			dev->component_size = strtoull(buf, NULL, 0) * 2;
		}
		if (options & GET_STATE) {
			dev->disk.state = 0;
			strcpy(dbase, "state");
			if (load_sys(fname, buf))
				goto abort;
			if (strstr(buf, "in_sync"))
				dev->disk.state |= (1<<MD_DISK_SYNC);
			if (strstr(buf, "faulty"))
				dev->disk.state |= (1<<MD_DISK_FAULTY);
			if (dev->disk.state == 0)
				sra->array.spare_disks++;
		}
		if (options & GET_ERROR) {
			strcpy(buf, "errors");
			if (load_sys(fname, buf))
				goto abort;
			dev->errors = strtoul(buf, NULL, 0);
		}
	}
	closedir(dir);
	return sra;

 abort:
	if (dir)
		closedir(dir);
	sysfs_free(sra);
	return NULL;
}

static int sysfs_set_str(struct mdinfo *sra, struct mdinfo *dev,
		  char *name, char *val)
{
	char fname[50];
	unsigned int n;
	int fd;

	sprintf(fname, "/sys/block/%s/md/%s/%s",
		sra->sys_name, dev?dev->sys_name:"", name);
	fd = open(fname, O_WRONLY);
	if (fd < 0)
		return -1;
	n = write(fd, val, strlen(val));
	close(fd);
	if (n != strlen(val)) {
		dprintf(Name ": failed to write '%s' to '%s' (%m)\n", val, fname);
		return -1;
	}
	return 0;
}

int sysfs_set_num(struct mdinfo *sra, struct mdinfo *dev,
		  char *name, unsigned long long val)
{
	char valstr[50];
	sprintf(valstr, "%llu", val);
	return sysfs_set_str(sra, dev, name, valstr);
}

int sysfs_uevent(struct mdinfo *sra, char *event)
{
	char fname[50];
	int n;
	int fd;

	sprintf(fname, "/sys/block/%s/uevent",
		sra->sys_name);
	fd = open(fname, O_WRONLY);
	if (fd < 0)
		return -1;
	n = write(fd, event, strlen(event));
	close(fd);
	if (n != (int)strlen(event)) {
		dprintf(Name ": failed to write '%s' to '%s' (%m)\n", event, fname);
		return -1;
	}
	return 0;
}

int sysfs_fd_get_ll(int fd, unsigned long long *val)
{
	char buf[50];
	int n;
	char *ep;

	lseek(fd, 0, 0);
	n = read(fd, buf, sizeof(buf));
	if (n <= 0)
		return -2;
	buf[n] = 0;
	*val = strtoull(buf, &ep, 0);
	if (ep == buf || (*ep != 0 && *ep != '\n' && *ep != ' '))
		return -1;
	return 0;
}

/*
 * 'mbr' is a pseudo metadata type for devices which have a
 * partition table in the Master Boot Record (mbr) also known
 * as a dos partition table.
 *
 * Obviously arrays cannot be created or assembled for this type.
 * It is used to allow a new bare device to have an partition table
 * added so the member partitions can then be included in other
 * arrays as relevant.
 *
 * The meaning operations are:
 * load_super
 * store_super
 */

static void free_mbr(struct supertype *st)
{
	free(st->sb);
	st->sb = NULL;
}

static int load_super_mbr(struct supertype *st, int fd, char *devname)
{
	/* try to read an mbr
	 * Return
	 *  0 on success
	 *  1 cannot get record
	 *  2 record is meaningless
	 */
	struct MBR *super;

	free_mbr(st);

	if (posix_memalign((void**)&super, 512, 512) != 0) {
		bb_error_msg("%s could not allocate superblock", __func__);
		return 1;
	}

	ioctl(fd, BLKFLSBUF, 0); /* make sure we read current data */

	lseek(fd, 0, 0);
	if (read(fd, super, sizeof(*super)) != sizeof(*super)) {
		if (devname)
			bb_error_msg("Cannot read partition table on %s", devname);
		free(super);
		return 1;
	}

	if (super->magic != MBR_SIGNATURE_MAGIC) {
		if (devname)
			bb_error_msg("No partition table found on %s", devname);
		free(super);
		return 1;
	}

	st->sb = super;

	if (st->ss == NULL) {
		st->ss = &mbr;
		st->minor_version = 0;
		st->max_devs = 1;
		st->info = NULL;
	}
	return 0;
}

static int store_mbr(struct supertype *st, int fd)
{
	struct MBR *old, *super;

	if (posix_memalign((void**)&old, 512, 512) != 0) {
		bb_error_msg("%s could not allocate superblock", __func__);
		return 1;
	}

	ioctl(fd, BLKFLSBUF, 0); /* make sure we read current data */

	lseek(fd, 0, 0);
	if (read(fd, old, sizeof(*old)) != sizeof(*old)) {
		free(old);
		return 1;
	}

	super = st->sb;
	memcpy(super->pad, old->pad, sizeof(super->pad));
	free(old);
	lseek(fd, 0, 0);
	if (write(fd, super, sizeof(*super)) != sizeof(*super))
		return 4;
	fsync(fd);
	ioctl(fd, BLKRRPART, 0);
	return 0;
}

static void getinfo_mbr(struct supertype *st, struct mdinfo *info, char *map)
{
	struct MBR *sb = st->sb;
	int i;

	(void)&map;

	memset(&info->array, 0, sizeof(info->array));
	memset(&info->disk, 0, sizeof(info->disk));
	strcpy(info->text_version, "mbr");
	strcpy(info->name, "mbr");
	info->component_size = 0;

	for (i = 0; i < MBR_PARTITIONS ; i++)
		if (sb->parts[i].blocks_num) {
			unsigned long last =
				(unsigned long)__le32_to_cpu(sb->parts[i].blocks_num)
				+ (unsigned long)__le32_to_cpu(sb->parts[i].first_sect_lba);
			if (last > info->component_size)
				info->component_size = last;
		}

}

static struct supertype *match_metadata_desc_mbr(char *arg)
{
	struct supertype *st;

	if (strcmp(arg, "mbr") != 0)
		return NULL;

	st = xmalloc(sizeof(*st));
	st->ss = &mbr;
	st->info = NULL;
	st->minor_version = 0;
	st->max_devs = 1;
	st->sb = NULL;
	return st;
}

static const struct superswitch mbr = {
	.match_metadata_desc = match_metadata_desc_mbr,
	.load_super = load_super_mbr,
	.store_super = store_mbr,
	.getinfo_super = getinfo_mbr,
	.free_super = free_mbr,
	.name = "mbr",
};

/*
 * 'gpt' is a pseudo metadata type for devices which have a
 * GPT partition table.
 *
 * Obviously arrays cannot be created or assembled for this type.
 * It is used to allow a new bare device to have an partition table
 * added so the member partitions can then be included in other
 * arrays as relevant.
 *
 * The meaning operations are:
 * load_super
 * store_super
 */

static void free_gpt(struct supertype *st)
{
	free(st->sb);
	st->sb = NULL;
}

static int load_gpt(struct supertype *st, int fd, char *devname)
{
	struct MBR *super;
	struct GPT *gpt_head;
	int to_read;

	free_gpt(st);

	if (posix_memalign((void**)&super, 4096, 32*512) != 0) {
		bb_error_msg("%s could not allocate superblock", __func__);
		return 1;
	}

	ioctl(fd, BLKFLSBUF, 0); /* make sure we read current data */

	lseek(fd, 0, 0);
	if (read(fd, super, sizeof(*super)) != sizeof(*super)) {
	no_read:
		if (devname)
			bb_error_msg("Cannot read partition table on %s", devname);
		free(super);
		return 1;
	}

	if (super->magic != MBR_SIGNATURE_MAGIC ||
	    super->parts[0].part_type != MBR_GPT_PARTITION_TYPE) {
	not_found:
		if (devname)
			bb_error_msg("No partition table found on %s", devname);
		free(super);
		return 1;
	}
	/* Seem to have GPT, load the header */
	gpt_head = (struct GPT*)(super+1);
	if (read(fd, gpt_head, sizeof(*gpt_head)) != sizeof(*gpt_head))
		goto no_read;
	if (gpt_head->magic != GPT_SIGNATURE_MAGIC)
		goto not_found;
	if (__le32_to_cpu(gpt_head->part_cnt) >= 128)
		goto not_found;

	to_read = __le32_to_cpu(gpt_head->part_cnt) * sizeof(struct GPT_part_entry);
	to_read =  ((to_read+511)/512) * 512;
	if (read(fd, gpt_head+1, to_read) != to_read)
		goto no_read;

	st->sb = super;

	if (st->ss == NULL) {
		st->ss = &gpt;
		st->minor_version = 0;
		st->max_devs = 1;
		st->info = NULL;
	}
	return 0;
}

static int store_gpt(struct supertype *st, int fd)
{
	/* FIXME should I save the boot loader */
	/* need to write two copies! */
	/* FIXME allow for blocks != 512 bytes
	 *etc
	 */
	struct MBR *super = st->sb;
	struct GPT *p_gpt;
	int to_write;

	p_gpt = (struct GPT*)(super+1);

	to_write = __le32_to_cpu(p_gpt->part_cnt) * sizeof(struct GPT_part_entry);
	to_write =  ((to_write+511)/512) * 512;

	lseek(fd, 0, 0);
	if (write(fd, st->sb, to_write) != to_write)
		return 4;

	fsync(fd);
	ioctl(fd, BLKRRPART, 0);
	return 0;
}

static void getinfo_gpt(struct supertype *st, struct mdinfo *info, char *map)
{
	struct GPT *p_gpt = st->sb + 512;
	struct GPT_part_entry *gpe = st->sb + 1024;
	unsigned int i;

	(void)&map;

	memset(&info->array, 0, sizeof(info->array));
	memset(&info->disk, 0, sizeof(info->disk));
	strcpy(info->text_version, "gpt");
	strcpy(info->name, "gpt");
	info->component_size = 0;

	for (i = 0; i < __le32_to_cpu(p_gpt->part_cnt); i++) {
		unsigned long long last =
			(unsigned long long)__le64_to_cpu(gpe[i].ending_lba);
		if (last > info->component_size)
			info->component_size = last;
	}
}

static struct supertype *match_metadata_desc_gpt(char *arg)
{
	struct supertype *st = xmalloc(sizeof(*st));

	if (strcmp(arg, "gpt") != 0) {
		free(st);
		return NULL;
	}

	st->ss = &gpt;
	st->info = NULL;
	st->minor_version = 0;
	st->max_devs = 1;
	st->sb = NULL;
	return st;
}

static const struct superswitch gpt = {
	.match_metadata_desc = match_metadata_desc_gpt,
	.load_super = load_gpt,
	.store_super = store_gpt,
	.getinfo_super = getinfo_gpt,
	.free_super = free_gpt,
	.name = "gpt",
};

int mdassemble_main(void)
{
	int rv = 0;
	struct mddev_ident *array_list;

	bb_xgetopt_ulflags("\0??");

	conffile = DefaultConfFile;
	mddevlp = &mddevlist;
	mdp_major = -1;
	mdstat_fd = -1;
	rom_fd = -1;
	rom_mem = MAP_FAILED;
	memcpy(adapter_rom_resources, adapter_rom_resources_st, sizeof(adapter_rom_resources_st));
	memcpy(&video_rom_resource, &video_rom_resource_st, sizeof(video_rom_resource_st));

	array_list =  conf_get_ident(NULL);
	if (!array_list)
		bb_error_msg_and_die("No arrays found in config file");
	 else
		for (; array_list; array_list = array_list->next) {
			mdu_array_info_t array;
			int mdfd;

			if (strcasecmp(array_list->devname, "<ignore>") == 0)
				continue;
			mdfd = open_mddev(array_list->devname);
			if (mdfd >= 0 && ioctl(mdfd, GET_ARRAY_INFO, &array) == 0) {
				rv |= Manage_ro(array_list->devname, mdfd, -1); /* make it readwrite */
				continue;
			}
			if (mdfd >= 0)
				close(mdfd);
			rv |= Assemble(array_list->st, array_list->devname,
				       array_list, NULL,
				       0, NULL, NULL, 0,
				       0, 0); /* runstop = verbose = force = 0 */
		}
	return rv;
}
