/*
 * mount.nfs umount.nfs statd sm-notify
 *
 * Authors:     Donald J. Becker, <becker@super.org>
 *              Rick Sladkey, <jrs@world.std.com>
 *              Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
 *              Olaf Kirch, <okir@monad.swb.de>
 *              Copyright (C) 2008-2009 Oracle Corporation.
 * Modified by Lon Hohberger, Oct. 2000.
 * svc_run: Copyright (c) 1984-2009, Sun Microsystems, Inc.
 * Copyright (C) 1995, 1997-1999 Jeffrey A. Uphoff
 * Copyright (C) 1993 Rick Sladkey <jrs@world.std.com>
 *
 *  Busybox port by Vladimir Oleynik (c) 2013 <dzo@simtreas.ru>
 *
 *              This software maybe be used for any purpose provided
 *              the above copyright notice is retained.  It is supplied
 *              as is, with no warranty expressed or implied.
 */
#ifdef APPLET_USAGE
#define mount_nfs_trivial_usage \
	"[flags] remotetarget dir [-rvVwfnsih] [-o nfsoptions]"
#define mount_nfs_full_usage \
	"options:\n" \
	"\t-r\t\tMount file system readonly\n" \
	"\t-v\t\tVerbose\n" \
	"\t-V\t\tPrint version\n" \
	"\t-w\t\tMount file system read-write\n" \
	"\t-f\t\tFake mount, do not actually mount\n" \
	"\t-n\t\tDo not update /etc/mtab\n" \
	"\t-s\t\tTolerate sloppy mount options rather than fail\n" \
	"\t-h\t\tPrint this help\n" \
	"\tnfsoptions\tRefer to mount.nfs(8) or nfs(5)"

#define umount_nfs_trivial_usage \
	"dir [-fvnrlh]"
#define umount_nfs_full_usage \
	"options:\n\t-f\t\tforce unmount\n" \
	"\t-v\tverbose\n" \
	"\t-n\tDo not update /etc/mtab\n" \
	"\t-r\tremount\n" \
	"\t-l\tlazy unmount\n" \
	"\t-h\tprint this help"

#define statd_trivial_usage \
	"[-FLNd] [-H prog] [-p port] [-o out_port] [-n host] [-P state_path]"
#define statd_full_usage \
	"\t-F, --foreground     Foreground (no-daemon mode)\n" \
	"\t-d, --no-syslog      Verbose logging to stderr.  Foreground mode only.\n" \
	"\t-p, --port           Port to listen on\n" \
	"\t-o, --outgoing-port  Port for outgoing connections\n" \
	"\t-V, -v, --version    Display version information and exit.\n" \
	"\t-n, --name           Specify a local hostname.\n" \
	"\t-P                   State directory path.\n" \
	"\t-N                   Run in notify only mode.\n" \
	"\t-L, --no-notify      Do not perform any notification.\n" \
	"\t-H                   Specify a high-availability callout program."

#define sm_notify_trivial_usage \
	"[-dfn] [-m max-retry-minutes] [-p srcport] [-P /path/to/state/directory] [-v my_host_name]"
#define sm_notify_full_usage \
	"\t-f\tForce\n" \
	"\t-d\tDebug mode on\n" \
	"\t-n\tNo update state"
#else

/* #undef HAVE_TCP_WRAPPER */
/* #undef LIBWRAP */
/* #undef HAVE_SYS_CAPABILITY_H */ /* need -lcap */

/* Name of package */
#define PACKAGE "nfs-utils"
/* Version number of package */
#define VERSION "1.2.3"
/* Define to the address where bug reports for this package should be sent. */
#define PACKAGE_BUGREPORT "linux-nfs@vger.kernel.org"

#define HAVE_DECL_AI_ADDRCONFIG 1
#define HAVE_GETRPCBYNAME 1
/* This defines the location of the NFS state files. */
#define NFS_STATEDIR "/var/lib/nfs"
/* Define to the full name and version of this package. */
#define PACKAGE_STRING "linux nfs-utils 1.2.3"
/* Define this to a script which can start statd on mount */
#ifndef CONFIG_INSTALL_NO_USR
#define START_STATD "/usr/sbin/start-statd"
#else
#define START_STATD "/sbin/start-statd"
#endif

#ifndef HAVE_DECL_AI_ADDRCONFIG
#define AI_ADDRCONFIG   0
#endif

#include "busybox.h"
#if defined(__MUSL__)
#error "MUSL libc have not RPC support. We can not compile nfs-like applets. :("
#endif
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/prctl.h>
#include <sys/mount.h>
#include <sys/queue.h>
#include <linux/posix_types.h>
#include <linux/version.h>
#include <stdbool.h>
#include <paths.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <strings.h>
#include <syslog.h>
#include <ctype.h>
#include <dirent.h>
#include <libgen.h>
#include <memory.h>
#include <poll.h>
#include <mntent.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <net/if.h>
#include <rpc/types.h>
#include <rpc/clnt.h>
#include <rpc/rpc.h>
#include <rpc/pmap_prot.h>
#include <rpc/pmap_rmt.h>
#include <rpc/pmap_clnt.h>
#include <rpc/svc.h>
#include <rpc/auth.h>
#ifdef HAVE_SYS_CAPABILITY_H
#include <sys/capability.h>
#endif
#include "bb_sig.h"
#include "bb_xopen.h"
#include "bb_inet.h"

#define NFS_MAXPATHLEN 1024
#define NFS_FHSIZE 32
#define NFS_COOKIESIZE 4

enum nfsstat {
	NFS_OK = 0,
	NFSERR_PERM = 1,
	NFSERR_NOENT = 2,
	NFSERR_IO = 5,
	NFSERR_NXIO = 6,
	NFSERR_ACCES = 13,
	NFSERR_EXIST = 17,
	NFSERR_NODEV = 19,
	NFSERR_NOTDIR = 20,
	NFSERR_ISDIR = 21,
	NFSERR_FBIG = 27,
	NFSERR_NOSPC = 28,
	NFSERR_ROFS = 30,
	NFSERR_NAMETOOLONG = 63,
	NFSERR_NOTEMPTY = 66,
	NFSERR_DQUOT = 69,
	NFSERR_STALE = 70,
	NFSERR_WFLUSH = 99,
};
typedef enum nfsstat nfsstat;


#ifndef _PATH_EXPORTS
#define _PATH_EXPORTS           "/etc/exports"
#endif
#ifndef _PATH_XTAB
#define _PATH_XTAB              NFS_STATEDIR "/xtab"
#endif
#ifndef _PATH_XTABTMP
#define _PATH_XTABTMP           NFS_STATEDIR "/xtab.tmp"
#endif
#ifndef _PATH_XTABLCK
#define _PATH_XTABLCK           NFS_STATEDIR "/.xtab.lock"
#endif
#ifndef _PATH_ETAB
#define _PATH_ETAB              NFS_STATEDIR "/etab"
#endif
#ifndef _PATH_ETABTMP
#define _PATH_ETABTMP           NFS_STATEDIR "/etab.tmp"
#endif
#ifndef _PATH_ETABLCK
#define _PATH_ETABLCK           NFS_STATEDIR "/.etab.lock"
#endif
#ifndef _PATH_RMTAB
#define _PATH_RMTAB             NFS_STATEDIR "/rmtab"
#endif
#ifndef _PATH_RMTABTMP
#define _PATH_RMTABTMP          _PATH_RMTAB ".tmp"
#endif
#ifndef _PATH_RMTABLCK
#define _PATH_RMTABLCK          NFS_STATEDIR "/.rmtab.lock"
#endif
#ifndef _PATH_PROC_EXPORTS
#define _PATH_PROC_EXPORTS      "/proc/fs/nfs/exports"
#define _PATH_PROC_EXPORTS_ALT  "/proc/fs/nfsd/exports"
#endif

#define UNUSED(x) UNUSED_ ## x __attribute__((unused))

/* This type is for defining buffers that contain network socket addresses.
 *
 * Casting a "struct sockaddr *" to the address of a "struct
 * sockaddr_storage" breaks C aliasing rules.  The "union
 * nfs_sockaddr" type follows C aliasing rules yet specifically
 * allows converting pointers to it between "struct sockaddr *"
 * and a few other network sockaddr-related pointer types.
 *
 * Note that this union is much smaller than a sockaddr_storage.
 * It should be used only for AF_INET or AF_INET6 socket addresses.
 * An AF_LOCAL sockaddr_un, for example, will clearly not fit into
 * a buffer of this type. */
union nfs_sockaddr {
	struct sockaddr         sa;
	struct sockaddr_in      s4;
#ifdef CONFIG_FEATURE_IPV6
	struct sockaddr_in6     s6;
#endif
};

#define SIZEOF_SOCKADDR_UNKNOWN (socklen_t)0
#define SIZEOF_SOCKADDR_IN      (socklen_t)sizeof(struct sockaddr_in)

#ifdef CONFIG_FEATURE_IPV6
#define SIZEOF_SOCKADDR_IN6     (socklen_t)sizeof(struct sockaddr_in6)
#else   /* !CONFIG_FEATURE_IPV6 */
#define SIZEOF_SOCKADDR_IN6     SIZEOF_SOCKADDR_UNKNOWN
#ifndef INET6_ADDRSTRLEN
#define INET6_ADDRSTRLEN 46
#endif
#endif  /* !CONFIG_FEATURE_IPV6 */

#define streq(s, t)     (strcmp ((s), (t)) == 0)


#define isoctal(a) (((a) & ~7) == '0')

/* exit status - bits below are ORed */
#define EX_SUCCESS      0       /* no failure occurred */
#define EX_USAGE        1       /* incorrect invocation or permission */
#define EX_SYSERR       2       /* out of memory, cannot fork, ... */
#define EX_SOFTWARE     4       /* internal mount bug or wrong version */
#define EX_USER         8       /* user interrupt */
#define EX_FILEIO      16       /* problems writing, locking, ... mtab/fstab */
#define EX_FAIL        32       /* mount failure */
#define EX_SOMEOK      64       /* some mount succeeded */
#define EX_BG         256       /* retry in background (internal only) */

static inline uint16_t get_port4(const struct sockaddr *sap)
{
	const struct sockaddr_in *sin = (const struct sockaddr_in *)sap;
	return ntohs(sin->sin_port);
}

#ifdef CONFIG_FEATURE_IPV6
static inline uint16_t get_port6(const struct sockaddr *sap)
{
	const struct sockaddr_in6 *sin6 = (const struct sockaddr_in6 *)sap;
	return ntohs(sin6->sin6_port);
}
#else   /* !CONFIG_FEATURE_IPV6 */
static inline uint16_t get_port6(const struct sockaddr *UNUSED(sap))
{
	return 0;
}
#endif  /* !CONFIG_FEATURE_IPV6 */

/* nfs_get_port - extract port value from a socket address
 * @sap: pointer to socket address
 *
 * Returns port value in host byte order, or zero if the
 * socket address contains an unrecognized family. */
static inline uint16_t nfs_get_port(const struct sockaddr *sap)
{
	switch (sap->sa_family) {
	case AF_INET:
		return get_port4(sap);
	case AF_INET6:
		return get_port6(sap);
	}
	return 0;
}

static inline void set_port4(struct sockaddr *sap, const uint16_t port)
{
	struct sockaddr_in *sin = (struct sockaddr_in *)sap;
	sin->sin_port = htons(port);
}

#ifdef CONFIG_FEATURE_IPV6
static inline void set_port6(struct sockaddr *sap, const uint16_t port)
{
	struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)sap;
	sin6->sin6_port = htons(port);
}
#else   /* !CONFIG_FEATURE_IPV6 */
static inline void set_port6(struct sockaddr *UNUSED(sap), const uint16_t UNUSED(port))
{
}
#endif  /* !CONFIG_FEATURE_IPV6 */

/* nfs_set_port - set port value in a socket address
 * @sap: pointer to socket address
 * @port: port value to set */
static inline void nfs_set_port(struct sockaddr *sap, const uint16_t port)
{
	switch (sap->sa_family) {
	case AF_INET:
		set_port4(sap, port);
		break;
	case AF_INET6:
		set_port6(sap, port);
		break;
	}
}

/* nfs_is_v4_loopback - test to see if socket address is AF_INET loopback
 * @sap: pointer to socket address
 *
 * Returns true if the socket address is the standard IPv4 loopback
 * address; otherwise false is returned. */
static inline _Bool nfs_is_v4_loopback(const struct sockaddr *sap)
{
	const struct sockaddr_in *sin = (const struct sockaddr_in *)sap;

	if (sin->sin_family != AF_INET)
		return false;
	if (sin->sin_addr.s_addr != htonl(INADDR_LOOPBACK))
		return false;
	return true;
}

static inline _Bool
compare_sockaddr4(const struct sockaddr *sa1, const struct sockaddr *sa2)
{
	const struct sockaddr_in *sin1 = (const struct sockaddr_in *)sa1;
	const struct sockaddr_in *sin2 = (const struct sockaddr_in *)sa2;
	return sin1->sin_addr.s_addr == sin2->sin_addr.s_addr;
}

#ifdef CONFIG_FEATURE_IPV6
static inline _Bool
compare_sockaddr6(const struct sockaddr *sa1, const struct sockaddr *sa2)
{
	const struct sockaddr_in6 *sin1 = (const struct sockaddr_in6 *)sa1;
	const struct sockaddr_in6 *sin2 = (const struct sockaddr_in6 *)sa2;

	if ((IN6_IS_ADDR_LINKLOCAL((char *)&sin1->sin6_addr) &&
	     IN6_IS_ADDR_LINKLOCAL((char *)&sin2->sin6_addr)) ||
	    (IN6_IS_ADDR_SITELOCAL((char *)&sin1->sin6_addr) &&
	     IN6_IS_ADDR_SITELOCAL((char *)&sin2->sin6_addr)))
		if (sin1->sin6_scope_id != sin2->sin6_scope_id)
			return false;

	return IN6_ARE_ADDR_EQUAL((char *)&sin1->sin6_addr,
					(char *)&sin2->sin6_addr);
}
#else   /* !CONFIG_FEATURE_IPV6 */
static inline _Bool
compare_sockaddr6(const struct sockaddr *UNUSED(sa1), const struct sockaddr *UNUSED(sa2))
{
	return false;
}
#endif  /* !CONFIG_FEATURE_IPV6 */

/* nfs_compare_sockaddr - compare two socket addresses for equality
 * @sa1: pointer to a socket address
 * @sa2: pointer to a socket address
 *
 * Returns true if the two socket addresses contain equivalent
 * network addresses; otherwise false is returned.
 */
static inline _Bool
nfs_compare_sockaddr(const struct sockaddr *sa1, const struct sockaddr *sa2)
{
	if (sa1 == NULL || sa2 == NULL)
		return false;

	if (sa1->sa_family == sa2->sa_family)
		switch (sa1->sa_family) {
		case AF_INET:
			return compare_sockaddr4(sa1, sa2);
		case AF_INET6:
			return compare_sockaddr6(sa1, sa2);
		}

	return false;
}


/* IANA does not define an IP protocol number for RDMA transports.
 * Choose an arbitrary value we can use locally. */
#define NFSPROTO_RDMA           (3939)

/* Conventional RPC program numbers */
#ifndef RPCBPROG
#define RPCBPROG        ((rpcprog_t)100000)
#endif

/* nfs_clear_rpc_createerr - zap all error reporting fields */
static inline void nfs_clear_rpc_createerr(void)
{
	memset(&rpc_createerr, 0, sizeof(rpc_createerr));
}

struct mntentchn {
	struct mntentchn *nxt, *prev;
	struct mntent m;
};

#define SM_MAXSTRLEN 1024

struct sm_name {
	char *mon_name;
};
typedef struct sm_name sm_name;

struct my_id {
	char *my_name;
	int my_prog;
	int my_vers;
	int my_proc;
};
typedef struct my_id my_id;

struct mon_id {
	char *mon_name;
	struct my_id my_id;
};
typedef struct mon_id mon_id;

#define SM_PRIV_SIZE 16

struct mon {
	struct mon_id mon_id;
	char priv[SM_PRIV_SIZE];
};
typedef struct mon mon;

struct stat_chge {
	char *mon_name;
	int state;
};
typedef struct stat_chge stat_chge;

struct sm_stat {
	int state;
};
typedef struct sm_stat sm_stat;

enum res {
	STAT_SUCC = 0,
	STAT_FAIL = 1,
};
typedef enum res res;

struct sm_stat_res {
	res res_stat;
	int state;
};
typedef struct sm_stat_res sm_stat_res;

struct status {
	char *mon_name;
	int state;
	char priv[SM_PRIV_SIZE];
};
typedef struct status status;

/* Primary information structure. */
struct notify_list {
  mon                   mon;    /* Big honkin' NSM structure. */
  in_port_t             port;   /* port number for callback */
  short int             times;  /* Counter used for various things. */
  int                   state;  /* For storing notified state for callbacks. */
  char                  *dns_name; /* used for matching incoming
				    * NOTIFY requests */
  struct notify_list    *next;  /* Linked list forward pointer. */
  struct notify_list    *prev;  /* Linked list backward pointer. */
  uint32_t              xid;    /* XID of MS_NOTIFY RPC call */
  time_t                when;   /* notify: timeout for re-xmit */
};

typedef struct notify_list notify_list;

struct nsm_host {
	struct nsm_host *       next;
	char *                  name;
	char *                  mon_name;
	char *                  my_name;
	struct addrinfo         *ai;
	time_t                  last_used;
	time_t                  send_next;
	unsigned int            timeout;
	unsigned int            retries;
	uint32_t                xid;
};

#define MAX_USER_FLAVOUR        16

CALC_BSS struct BB_COMMON_BSS {
	/* Flag for already existing lock file. */
	int we_created_lockfile;
	int lockfile_fd;
	char ip_addr[16];
	int pidfd;
	int sockfd;     /* notify socket */
	char nsm_base_dirname[PATH_MAX];
	int  nsm_family;
	/* Flag to indicate that signals have been set up. */
	int signals_have_been_setup;
	int verbose;
	int have_mtab_info;
	int var_mtab_does_not_exist;
	int var_mtab_is_a_symlink;
	int nfs_mount_data_version;
	char errbuf[BUFSIZ];
	char bss_common_buf[BUFSIZ];
	char *SM_stat_chge_mon_name;
	int SM_stat_chge_state;
	int _rpcpmstart;
	int _rpcfdtype;
	struct mntentchn mounttable;
	struct mntentchn fstab;
	int got_mtab;
	int got_fstab;
	/* PRC: a high-availability callout program can be specified with -H
	 * When this is done, the program will receive callouts whenever clients
	 * are added or deleted to the notify list */
	char *ha_callout_prog;
	char *mount_opts;
	char nsm_hostname[SM_MAXSTRLEN + 1];
	unsigned int opt_max_retry;
	char *opt_srcaddr;
	int nsm_state;
	/* Global Variables */
	notify_list *rtnl;   /* Run-time notify list */
	/* This is the global statd_notify list onto which all SM_NOTIFY and CALLBACK
	 * requests are put. Pending RPC calls */
	notify_list *statd_notify;
	/* Modes of operation - Lon */
	int run_mode;        /* foreground logging mode */
	volatile int svc_stop;
	struct nsm_host *hosts;
	int nfsmount_string_flg;
	char hostdir[1024];
	int pseudoflavour[MAX_USER_FLAVOUR];
	char clnt_res;
	int iter_tries;
	SVCXPRT *last_transp_SOCK_DGRAM;
	SVCXPRT *last_transp_SOCK_STREAM;
	uint32_t nsm_xid;
	int onlyonce;
	struct mntent me;
	char *result1;
	char *result2;
	sm_stat_res result3;
	sm_stat_res result4;
	sm_stat result5;
	sm_stat result6;
	int st_getdtablesize;
	int kernel_version;
} bb_common_bss;
#define we_created_lockfile bb_common_bss.we_created_lockfile
#define lockfile_fd bb_common_bss.lockfile_fd
#define ip_addr bb_common_bss.ip_addr
#define pidfd bb_common_bss.pidfd
#define sockfd bb_common_bss.sockfd
#define nsm_base_dirname bb_common_bss.nsm_base_dirname
#define nsm_family bb_common_bss.nsm_family
#define signals_have_been_setup bb_common_bss.signals_have_been_setup
#define verbose bb_common_bss.verbose
#define have_mtab_info bb_common_bss.have_mtab_info
#define var_mtab_does_not_exist bb_common_bss.var_mtab_does_not_exist
#define var_mtab_is_a_symlink bb_common_bss.var_mtab_is_a_symlink
#define nfs_mount_data_version bb_common_bss.nfs_mount_data_version
#define errbuf bb_common_bss.errbuf
#define bss_common_buf bb_common_bss.bss_common_buf
#define MY_NAME         bb_common_bss.SM_stat_chge_mon_name
#define MY_STATE        bb_common_bss.SM_stat_chge_state
#define _rpcpmstart bb_common_bss._rpcpmstart
#define _rpcfdtype bb_common_bss._rpcfdtype
#define mounttable bb_common_bss.mounttable
#define fstab bb_common_bss.fstab
#define got_mtab bb_common_bss.got_mtab
#define got_fstab bb_common_bss.got_fstab
#define ha_callout_prog bb_common_bss.ha_callout_prog
#define mount_opts bb_common_bss.mount_opts
#define nsm_hostname bb_common_bss.nsm_hostname
#define opt_max_retry bb_common_bss.opt_max_retry
#define opt_srcaddr bb_common_bss.opt_srcaddr
#define nsm_state bb_common_bss.nsm_state
#define rtnl bb_common_bss.rtnl
#define statd_notify bb_common_bss.statd_notify
#define run_mode bb_common_bss.run_mode
#define svc_stop bb_common_bss.svc_stop
#define hosts bb_common_bss.hosts
#define nfsmount_string_flg bb_common_bss.nfsmount_string_flg
#define hostdir bb_common_bss.hostdir
#define pseudoflavour bb_common_bss.pseudoflavour
#define clnt_res bb_common_bss.clnt_res
#define iter_tries bb_common_bss.iter_tries
#define last_transp_SOCK_DGRAM bb_common_bss.last_transp_SOCK_DGRAM
#define last_transp_SOCK_STREAM bb_common_bss.last_transp_SOCK_STREAM
#define nsm_xid bb_common_bss.nsm_xid
#define onlyonce bb_common_bss.onlyonce
#define me bb_common_bss.me
#define result1 bb_common_bss.result1
#define result2 bb_common_bss.result2
#define result3 bb_common_bss.result3
#define result4 bb_common_bss.result4
#define result5 bb_common_bss.result5
#define result6 bb_common_bss.result6
#define st_getdtablesize bb_common_bss.st_getdtablesize
#define kernel_version bb_common_bss.kernel_version


static void xlog_open(void)
{
	bb_openlog(LOG_PID, LOG_DAEMON);
	/* WARNING: I do remove xlog_toggle! */
	bb_signal((1 << SIGUSR1) | (1 << SIGUSR2), SIG_IGN);
}

static const rpcvers_t default_rpcb_version = PMAPVERS;

/* Historical: Map TCP connect timeouts to timeout
 * error code used by UDP. */
static void nfs_gp_map_tcp_errorcodes(const unsigned short protocol)
{
	if (protocol != IPPROTO_TCP)
		return;

	switch (rpc_createerr.cf_error.re_errno) {
	case ETIMEDOUT:
		rpc_createerr.cf_stat = RPC_TIMEDOUT;
		break;
	case ECONNREFUSED:
		rpc_createerr.cf_stat = RPC_CANTRECV;
		break;
	}
}

/* Discover the port number that should be used to contact an
 * rpcbind service.  This will detect if the port has a local
 * value that may have been set in /etc/services.
 *
 * Returns network byte-order port number of rpcbind service
 * on this system. */
static in_port_t nfs_gp_get_rpcb_port(const unsigned short protocol)
{
	static const char * const rpcb_netnametbl[] = {
		"rpcbind", "portmapper", "sunrpc", NULL
	};
	unsigned int i;
	struct addrinfo gai_hint;
	struct addrinfo *gai_results;
	const struct sockaddr_in *sin;
	in_port_t port;

	for (i = 0; rpcb_netnametbl[i] != NULL; i++) {
		gai_hint.ai_family = AF_INET;
		gai_hint.ai_protocol = protocol;
		gai_hint.ai_flags = AI_PASSIVE;
		if (getaddrinfo(NULL, rpcb_netnametbl[i], &gai_hint,
							&gai_results) != 0)
			continue;

		sin = (const struct sockaddr_in *)gai_results->ai_addr;
		port = sin->sin_port;
		freeaddrinfo(gai_results);
		if (port != 0)
			return port;
	}
	return (in_port_t)htons((uint16_t)PMAPPORT);
}

/* Acquire an RPC CLIENT * with an ephemeral source port */
static CLIENT           *nfs_get_rpcclient(const struct sockaddr *,
				const socklen_t, const unsigned short,
				const rpcprog_t, const rpcvers_t,
				struct timeval *);

/*
 * nfs_getrpcbyname - convert an RPC program name to a rpcprog_t
 * @program: default program number to use if names not found in db
 * @table: pointer to table of 'char *' names to try to find
 *
 * Returns program number of first name to be successfully looked
 * up, or the default program number if all lookups fail.
 */
static rpcprog_t nfs_getrpcbyname(const rpcprog_t program, const char * const table[])
{
#ifdef HAVE_GETRPCBYNAME
	struct rpcent *entry;
	unsigned int i;

	if (table != NULL)
		for (i = 0; table[i] != NULL; i++) {
			entry = getrpcbyname(table[i]);
			if (entry)
				return (rpcprog_t)entry->r_number;
		}
#endif  /* HAVE_GETRPCBYNAME */

	return program;
}

/*
 * Set up an RPC client for communicating with an rpcbind daemon at
 * @sap over @transport with protocol version @version.
 *
 * Returns a pointer to a prepared RPC client if successful, and
 * @timeout is initialized; caller must destroy a non-NULL returned RPC
 * client.  Otherwise returns NULL, and rpc_createerr.cf_stat is set to
 * reflect the error.
 */
static CLIENT *nfs_gp_get_rpcbclient(struct sockaddr *sap,
				     const socklen_t salen,
				     const unsigned short transport,
				     const rpcvers_t version,
				     struct timeval *timeout)
{
	static const char * const rpcb_pgmtbl[] = {
		"rpcbind",
		"portmap",
		"portmapper",
		"sunrpc",
		NULL,
	};
	rpcprog_t rpcb_prog = nfs_getrpcbyname(RPCBPROG, rpcb_pgmtbl);
	CLIENT *clnt;

	nfs_set_port(sap, ntohs(nfs_gp_get_rpcb_port(transport)));
	clnt = nfs_get_rpcclient(sap, salen, transport, rpcb_prog,
							version, timeout);
	nfs_gp_map_tcp_errorcodes(transport);
	return clnt;
}

/**
 * nfs_get_proto - Convert a netid to an address family and protocol number
 * @netid: C string containing a netid
 * @family: OUT: address family
 * @protocol: OUT: protocol number
 *
 * Returns 1 and fills in @protocol if the netid was recognized;
 * otherwise zero is returned.
 */
static int
nfs_get_proto(const char *netid, sa_family_t *family, unsigned long *protocol)
{
	struct protoent *proto;

	/*
	 * IANA does not define a protocol number for rdma netids,
	 * since "rdma" is not an IP protocol.
	 */
	if (strcmp(netid, "rdma") == 0) {
		*family = AF_INET;
		*protocol = NFSPROTO_RDMA;
		return 1;
	}

	proto = getprotobyname(netid);
	if (proto == NULL)
		return 0;

	*family = AF_INET;
	*protocol = (unsigned long)proto->p_proto;
	return 1;
}

/**
 * nfs_get_netid - Convert a protocol family and protocol name to a netid
 * @family: protocol family
 * @protocol: protocol number
 *
 * One of the arguments passed when querying remote rpcbind services
 * via rpcbind v3 or v4 is a netid string.  This replaces the pm_prot
 * field used in legacy PMAP_GETPORT calls.
 *
 * RFC 1833 says netids are not standard but rather defined on the local
 * host.  There are, however, standard definitions for nc_protofmly and
 * nc_proto that can be used to derive a netid string on the local host,
 * based on the contents of /etc/netconfig.
 *
 * Walk through the local netconfig database and grab the netid of the
 * first entry that matches @family and @protocol and whose netid string
 * fits in the provided buffer.
 *
 * Returns a '\0'-terminated string if successful.  Caller must
 * free the returned string.  Otherwise NULL is returned, and
 * rpc_createerr.cf_stat is set to reflect the error.
 */
static char *
nfs_get_netid(const sa_family_t family, const unsigned long protocol)
{
	struct protoent *proto;
	char *netid;

	if (family != AF_INET)
		goto out;
	proto = getprotobynumber((int)protocol);
	if (proto == NULL)
		goto out;

	netid = strdup(proto->p_name);
	if (netid == NULL)
		rpc_createerr.cf_stat = RPC_SYSTEMERROR;
	return netid;

out:
	rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
	return NULL;
}

/*
 * Extract a port number from a universal address, and terminate the
 * string in @addrstr just after the address part.
 *
 * Returns -1 if unsuccesful; otherwise a decoded port number (possibly 0)
 * is returned.
 */
static int nfs_gp_universal_porthelper(char *addrstr)
{
	char *p, *endptr;
	unsigned long portlo, porthi;
	int port = -1;

	p = strrchr(addrstr, '.');
	if (p == NULL)
		goto out;
	portlo = strtoul(p + 1, &endptr, 10);
	if (*endptr != '\0' || portlo > 255)
		goto out;
	*p = '\0';

	p = strrchr(addrstr, '.');
	if (p == NULL)
		goto out;
	porthi = strtoul(p + 1, &endptr, 10);
	if (*endptr != '\0' || porthi > 255)
		goto out;
	*p = '\0';
	port = (porthi << 8) | portlo;

out:
	return port;
}

/* nfs_universal2port - extract port number from a "universal address"
 * @uaddr: '\0'-terminated C string containing a universal address
 *
 * Universal addresses (defined in RFC 1833) are used when calling an
 * rpcbind daemon via protocol versions 3 or 4..
 *
 * Returns -1 if unsuccesful; otherwise a decoded port number (possibly 0)
 * is returned. */
static int nfs_universal2port(const char *uaddr)
{
	char *addrstr;
	int port = -1;

	addrstr = strdup(uaddr);
	if (addrstr != NULL) {
		port = nfs_gp_universal_porthelper(addrstr);
		free(addrstr);
	}
	return port;
}

/*
 * Send a NULL request to the indicated RPC service.
 *
 * Returns 1 if the service responded; otherwise 0;
 */
static int nfs_gp_ping(CLIENT *client, struct timeval timeout)
{
	enum clnt_stat call_status;

	call_status = CLNT_CALL(client, NULLPROC,
			   (xdrproc_t)xdr_void, NULL,
			   (xdrproc_t)xdr_void, NULL,
			   timeout);

	if (call_status != RPC_SUCCESS) {
		rpc_createerr.cf_stat = call_status;
		CLNT_GETERR(client, &rpc_createerr.cf_error);
	}
	return (int)(call_status == RPC_SUCCESS);
}

/*
 * Try GETPORT request via rpcbind version 2.
 *
 * Returns non-zero port number on success; otherwise returns
 * zero.  rpccreateerr is set to reflect the nature of the error.
 */
static unsigned long nfs_gp_pmap_getport(CLIENT *client,
					 struct pmap *parms,
					 struct timeval timeout)
{
	enum clnt_stat call_status;
	unsigned long port;

	call_status = CLNT_CALL(client, (rpcproc_t)PMAPPROC_GETPORT,
			   (xdrproc_t)xdr_pmap, (void *)parms,
			   (xdrproc_t)xdr_u_long, (void *)&port,
			   timeout);

	if (call_status != RPC_SUCCESS) {
		rpc_createerr.cf_stat = call_status;
		CLNT_GETERR(client, &rpc_createerr.cf_error);
		port = 0;
	} else if (port == 0)
		rpc_createerr.cf_stat = RPC_PROGNOTREGISTERED;

	return port;
}

static unsigned long nfs_gp_getport_pmap(CLIENT *client,
					 const rpcprog_t program,
					 const rpcvers_t version,
					 const unsigned short protocol,
					 struct timeval timeout)
{
	struct pmap parms = {
		.pm_prog        = program,
		.pm_vers        = version,
		.pm_prot        = protocol,
	};
	rpcvers_t pmap_version = PMAPVERS;

	CLNT_CONTROL(client, CLSET_VERS, (void *)&pmap_version);
	return nfs_gp_pmap_getport(client, &parms, timeout);
}

/*
 * Try an AF_INET6 request via rpcbind v4/v3; try an AF_INET
 * request via rpcbind v2.
 *
 * Returns non-zero port number on success; otherwise returns
 * zero.  rpccreateerr is set to reflect the nature of the error.
 */
static unsigned short nfs_gp_getport(CLIENT *client,
				     const struct sockaddr *sap,
				     const rpcprog_t program,
				     const rpcvers_t version,
				     const unsigned short protocol,
				     struct timeval timeout)
{
	switch (sap->sa_family) {
	case AF_INET:
		return nfs_gp_getport_pmap(client, program, version,
							protocol, timeout);
	}

	rpc_createerr.cf_stat = RPC_UNKNOWNADDR;
	return 0;
}

/* nfs_rpc_ping - Determine if RPC service is responding to requests
 * @sap: pointer to address of server to query (port is already filled in)
 * @salen: length of server address
 * @program: requested RPC program number
 * @version: requested RPC version number
 * @protocol: requested IPPROTO_ value of transport protocol
 * @timeout: pointer to request timeout (NULL means use default timeout)
 *
 * Returns 1 if the remote service responded without an error; otherwise
 * zero. */
static int nfs_rpc_ping(const struct sockaddr *sap, const socklen_t salen,
		 const rpcprog_t program, const rpcvers_t version,
		 const unsigned short protocol, const struct timeval *timeout)
{
	union nfs_sockaddr address;
	struct sockaddr *saddr = &address.sa;
	CLIENT *client;
	struct timeval tout = { -1, 0 };
	int result = 0;

	if (timeout != NULL)
		tout = *timeout;

	nfs_clear_rpc_createerr();

	memcpy(saddr, sap, (size_t)salen);
	client = nfs_get_rpcclient(saddr, salen, protocol,
						program, version, &tout);
	if (client != NULL) {
		result = nfs_gp_ping(client, tout);
		nfs_gp_map_tcp_errorcodes(protocol);
		CLNT_DESTROY(client);
	}

	return result;
}

/**
 * nfs_getport - query server's rpcbind to get port number for an RPC service
 * @sap: pointer to address of server to query
 * @salen: length of server's address
 * @program: requested RPC program number
 * @version: requested RPC version number
 * @protocol: IPPROTO_ value of requested transport protocol
 *
 * Uses any acceptable rpcbind version to discover the port number for the
 * RPC service described by the given [program, version, transport] tuple.
 * Uses a quick timeout and an ephemeral source port.  Supports AF_INET and
 * AF_INET6 server addresses.
 *
 * Returns a positive integer representing the port number of the RPC
 * service advertised by the server (in host byte order), or zero if the
 * service is not advertised or there was some problem querying the server's
 * rpcbind daemon.  rpccreateerr is set to reflect the underlying cause of
 * the error.
 *
 * There are a variety of ways to choose which transport and rpcbind versions
 * to use.  We chose to conserve local resources and try to avoid incurring
 * timeouts.
 *
 * Transport
 * To provide rudimentary support for traversing firewalls, query the remote
 * using the same transport as the requested service.  This provides some
 * guarantee that the requested transport is available between this client
 * and the server, and if the caller specifically requests TCP, for example,
 * this may be becuase a firewall is in place that blocks UDP traffic.  We
 * could try both, but that could involve a lengthy timeout in several cases,
 * and would often consume an extra ephemeral port.
 *
 * Rpcbind version
 * To avoid using up too many ephemeral ports, AF_INET queries use tried-and-
 * true rpcbindv2, and don't try the newer versions; and AF_INET6 queries use
 * rpcbindv4, then rpcbindv3 on the same socket.  The newer rpcbind protocol
 * versions can adequately detect if a remote RPC service does not support
 * AF_INET6 at all.  The rpcbind socket is re-used in an attempt to keep the
 * overall number of consumed ephemeral ports low.
 */
static unsigned short nfs_getport(const struct sockaddr *sap,
			   const socklen_t salen,
			   const rpcprog_t program,
			   const rpcvers_t version,
			   const unsigned short protocol)
{
	union nfs_sockaddr address;
	struct sockaddr *saddr = &address.sa;
	struct timeval timeout = { -1, 0 };
	unsigned short port = 0;
	CLIENT *client;

	nfs_clear_rpc_createerr();

	memcpy(saddr, sap, (size_t)salen);
	client = nfs_gp_get_rpcbclient(saddr, salen, protocol,
						default_rpcb_version, &timeout);
	if (client != NULL) {
		port = nfs_gp_getport(client, saddr, program,
					version, protocol, timeout);
		CLNT_DESTROY(client);
	}

	return port;
}

/* nfs_getport_ping - query server's rpcbind and do RPC ping to verify result
 * @sap: IN: pointer to address of server to query;
 *       OUT: pointer to updated address
 * @salen: length of server's address
 * @program: requested RPC program number
 * @version: requested RPC version number
 * @protocol: IPPROTO_ value of requested transport protocol
 *
 * Uses any acceptable rpcbind version to discover the port number for the
 * RPC service described by the given [program, version, transport] tuple.
 * Uses a quick timeout and an ephemeral source port.  Supports AF_INET and
 * AF_INET6 server addresses.
 *
 * Returns a 1 and sets the port number in the passed-in server address
 * if both the query and the ping were successful; otherwise zero.
 * rpccreateerr is set to reflect the underlying cause of the error. */
static int nfs_getport_ping(struct sockaddr *sap, const socklen_t salen,
		     const rpcprog_t program, const rpcvers_t version,
		     const unsigned short protocol)
{
	struct timeval timeout = { -1, 0 };
	unsigned short port = 0;
	CLIENT *client;
	int result = 0;

	nfs_clear_rpc_createerr();

	client = nfs_gp_get_rpcbclient(sap, salen, protocol,
						default_rpcb_version, &timeout);
	if (client != NULL) {
		port = nfs_gp_getport(client, sap, program,
					version, protocol, timeout);
		CLNT_DESTROY(client);
		client = NULL;
	}

	if (port != 0) {
		union nfs_sockaddr address;
		struct sockaddr *saddr = &address.sa;

		memcpy(saddr, sap, (size_t)salen);
		nfs_set_port(saddr, port);

		nfs_clear_rpc_createerr();

		client = nfs_get_rpcclient(saddr, salen, protocol,
						program, version, &timeout);
		if (client != NULL) {
			result = nfs_gp_ping(client, timeout);
			nfs_gp_map_tcp_errorcodes(protocol);
			CLNT_DESTROY(client);
		}
	}

	if (result)
		nfs_set_port(sap, port);

	return result;
}

/*
 * Generic RPC client socket-level APIs for nfs-utils
 *
 * Copyright (C) 2008 Oracle Corporation.  All rights reserved.
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */


/*
 * If "-1" is specified in the tv_sec field, use these defaults instead.
 */
#define NFSRPC_TIMEOUT_UDP      (3)
#define NFSRPC_TIMEOUT_TCP      (10)


/*
 * Set up an RPC client for communicating via a AF_LOCAL socket.
 *
 * @timeout is initialized upon return
 *
 * Returns a pointer to a prepared RPC client if successful; caller
 * must destroy a non-NULL returned RPC client.  Otherwise NULL, and
 * rpc_createerr.cf_stat is set to reflect the error.
 */
static CLIENT *nfs_get_localclient(const struct sockaddr *sap,
				   const socklen_t UNUSED(salen),
				   const rpcprog_t program,
				   const rpcvers_t version,
				   struct timeval *timeout)
{
	CLIENT *client;
	int sock;

	sock = socket(AF_LOCAL, SOCK_STREAM, 0);
	if (sock == -1) {
		rpc_createerr.cf_stat = RPC_SYSTEMERROR;
		rpc_createerr.cf_error.re_errno = errno;
		return NULL;
	}

	if (timeout->tv_sec == -1)
		timeout->tv_sec = NFSRPC_TIMEOUT_TCP;

	client = clntunix_create((struct sockaddr_un *)sap,
					program, version, &sock, 0, 0);
	if (client != NULL)
		CLNT_CONTROL(client, CLSET_FD_CLOSE, NULL);
	else
		(void)close(sock);

	return client;
}

/*
 * Bind a socket using an unused ephemeral source port.
 *
 * Returns zero on success, or returns -1 on error.  errno is
 * set to reflect the nature of the error.
 */
static int nfs_bind(const int sock, const sa_family_t family)
{
	struct sockaddr_in sin = {
		.sin_family             = AF_INET,
		.sin_addr.s_addr        = htonl(INADDR_ANY),
	};
#ifdef CONFIG_FEATURE_IPV6
	struct sockaddr_in6 sin6 = {
		.sin6_family            = AF_INET6,
		.sin6_addr              = IN6ADDR_ANY_INIT,
	};
#endif

	switch (family) {
	case AF_INET:
		return bind(sock, (struct sockaddr *)(char *)&sin,
					(socklen_t)sizeof(sin));
#ifdef CONFIG_FEATURE_IPV6
	case AF_INET6:
		return bind(sock, (struct sockaddr *)(char *)&sin6,
					(socklen_t)sizeof(sin6));
#endif
	}

	errno = EAFNOSUPPORT;
	return -1;
}

/* Bind a socket using an unused privileged source port.
 *
 * Returns zero on success, or returns -1 on error.  errno is
 * set to reflect the nature of the error. */
static int nfs_bindresvport(const int sock, const sa_family_t family)
{
	if (family != AF_INET) {
		errno = EAFNOSUPPORT;
		return -1;
	}

	return bindresvport(sock, NULL);
}

/* Perform a non-blocking connect on the socket fd.
 *
 * @timeout is modified to contain the time remaining (i.e. time provided
 * minus time elasped).
 *
 * Returns zero on success, or returns -1 on error.  errno is
 * set to reflect the nature of the error. */
static int nfs_connect_nb(const int fd, const struct sockaddr *sap,
			  const socklen_t salen, struct timeval *timeout)
{
	int flags, ret;
	fd_set rset;

	flags = fcntl(fd, F_GETFL, 0);
	if (flags < 0)
		return -1;

	ret = fcntl(fd, F_SETFL, flags | O_NONBLOCK);
	if (ret < 0)
		return -1;

	/*
	 * From here on subsequent sys calls could change errno so
	 * we set ret = -errno to capture it in case we decide to
	 * use it later.
	 */
	ret = connect(fd, sap, salen);
	if (ret < 0 && errno != EINPROGRESS) {
		ret = -1;
		goto done;
	}

	if (ret == 0)
		goto done;

	/* now wait */
	FD_ZERO(&rset);
	FD_SET(fd, &rset);

	ret = select(fd + 1, NULL, &rset, NULL, timeout);
	if (ret <= 0) {
		if (ret == 0)
			errno = ETIMEDOUT;
		ret = -1;
		goto done;
	}

	if (FD_ISSET(fd, &rset)) {
		socklen_t len = (socklen_t)sizeof(ret);

		if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &ret, &len) < 0) {
			ret = -1;
			goto done;
		}

		/* Oops - something wrong with connect */
		if (ret != 0) {
			errno = ret;
			ret = -1;
		}
	}

done:
	(void)fcntl(fd, F_SETFL, flags);
	return ret;
}

/*
 * Set up an RPC client for communicating via a datagram socket.
 * A connected UDP socket is used to detect a missing remote
 * listener as quickly as possible.
 *
 * @timeout is initialized upon return
 *
 * Returns a pointer to a prepared RPC client if successful; caller
 * must destroy a non-NULL returned RPC client.  Otherwise NULL, and
 * rpc_createerr.cf_stat is set to reflect the error.
 */
static CLIENT *nfs_get_udpclient(const struct sockaddr *sap,
				 const socklen_t salen,
				 const rpcprog_t program,
				 const rpcvers_t version,
				 struct timeval *timeout,
				 const int resvport)
{
	CLIENT *client;
	int ret, sock;

	if (sap->sa_family != AF_INET) {
		rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
		return NULL;
	}

	sock = socket((int)sap->sa_family, SOCK_DGRAM, IPPROTO_UDP);
	if (sock == -1) {
		rpc_createerr.cf_stat = RPC_SYSTEMERROR;
		rpc_createerr.cf_error.re_errno = errno;
		return NULL;
	}

	if (resvport)
		ret = nfs_bindresvport(sock, sap->sa_family);
	else
		ret = nfs_bind(sock, sap->sa_family);
	if (ret < 0) {
		rpc_createerr.cf_stat = RPC_SYSTEMERROR;
		rpc_createerr.cf_error.re_errno = errno;
		(void)close(sock);
		return NULL;
	}

	if (timeout->tv_sec == -1)
		timeout->tv_sec = NFSRPC_TIMEOUT_UDP;

	ret = nfs_connect_nb(sock, sap, salen, timeout);
	if (ret != 0) {
		rpc_createerr.cf_stat = RPC_SYSTEMERROR;
		rpc_createerr.cf_error.re_errno = errno;
		(void)close(sock);
		return NULL;
	}

	client = clntudp_create((struct sockaddr_in *)sap, program,
					version, *timeout, &sock);
	if (client != NULL) {
		struct timeval retry_timeout = { 1, 0 };
		CLNT_CONTROL(client, CLSET_RETRY_TIMEOUT,
						(char *)&retry_timeout);
		CLNT_CONTROL(client, CLSET_FD_CLOSE, NULL);
	} else
		(void)close(sock);

	return client;
}

/*
 * Set up and connect an RPC client for communicating via a stream socket.
 *
 * @timeout is initialized upon return
 *
 * Returns a pointer to a prepared and connected RPC client if
 * successful; caller must destroy a non-NULL returned RPC client.
 * Otherwise NULL, and rpc_createerr.cf_stat is set to reflect the
 * error.
 */
static CLIENT *nfs_get_tcpclient(const struct sockaddr *sap,
				 const socklen_t salen,
				 const rpcprog_t program,
				 const rpcvers_t version,
				 struct timeval *timeout,
				 const int resvport)
{
	CLIENT *client;
	int ret, sock;

	if (sap->sa_family != AF_INET) {
		rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
		return NULL;
	}

	sock = socket((int)sap->sa_family, SOCK_STREAM, IPPROTO_TCP);
	if (sock == -1) {
		rpc_createerr.cf_stat = RPC_SYSTEMERROR;
		rpc_createerr.cf_error.re_errno = errno;
		return NULL;
	}

	if (resvport)
		ret = nfs_bindresvport(sock, sap->sa_family);
	else
		ret = nfs_bind(sock, sap->sa_family);
	if (ret < 0) {
		rpc_createerr.cf_stat = RPC_SYSTEMERROR;
		rpc_createerr.cf_error.re_errno = errno;
		(void)close(sock);
		return NULL;
	}

	if (timeout->tv_sec == -1)
		timeout->tv_sec = NFSRPC_TIMEOUT_TCP;

	ret = nfs_connect_nb(sock, sap, salen, timeout);
	if (ret != 0) {
		rpc_createerr.cf_stat = RPC_SYSTEMERROR;
		rpc_createerr.cf_error.re_errno = errno;
		(void)close(sock);
		return NULL;
	}

	client = clnttcp_create((struct sockaddr_in *)sap,
					program, version, &sock, 0, 0);
	if (client != NULL)
		CLNT_CONTROL(client, CLSET_FD_CLOSE, NULL);
	else
		(void)close(sock);

	return client;
}

/* nfs_get_rpcclient - acquire an RPC client
 * @sap: pointer to socket address of RPC server
 * @salen: length of socket address
 * @transport: IPPROTO_ value of transport protocol to use
 * @program: RPC program number
 * @version: RPC version number
 * @timeout: pointer to request timeout (must not be NULL)
 *
 * Set up an RPC client for communicating with an RPC program @program
 * and @version on the server @sap over @transport.  An unprivileged
 * source port is used.
 *
 * Returns a pointer to a prepared RPC client if successful, and
 * @timeout is initialized; caller must destroy a non-NULL returned RPC
 * client.  Otherwise returns NULL, and rpc_createerr.cf_stat is set to
 * reflect the error. */
static CLIENT *nfs_get_rpcclient(const struct sockaddr *sap,
			  const socklen_t salen,
			  const unsigned short transport,
			  const rpcprog_t program,
			  const rpcvers_t version,
			  struct timeval *timeout)
{
	nfs_clear_rpc_createerr();

	switch (sap->sa_family) {
	case AF_LOCAL:
		return nfs_get_localclient(sap, salen, program,
						version, timeout);
	case AF_INET:
	case AF_INET6:
		if (nfs_get_port(sap) == 0) {
			rpc_createerr.cf_stat = RPC_UNKNOWNADDR;
			return NULL;
		}
		break;
	default:
		rpc_createerr.cf_stat = RPC_UNKNOWNADDR;
		return NULL;
	}

	switch (transport) {
	case IPPROTO_TCP:
		return nfs_get_tcpclient(sap, salen, program, version,
						timeout, 0);
	case 0:
	case IPPROTO_UDP:
		return nfs_get_udpclient(sap, salen, program, version,
						timeout, 0);
	}

	rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
	return NULL;
}

/**
 * nfs_get_priv_rpcclient - acquire an RPC client
 * @sap: pointer to socket address of RPC server
 * @salen: length of socket address
 * @transport: IPPROTO_ value of transport protocol to use
 * @program: RPC program number
 * @version: RPC version number
 * @timeout: pointer to request timeout (must not be NULL)
 *
 * Set up an RPC client for communicating with an RPC program @program
 * and @version on the server @sap over @transport.  A privileged
 * source port is used.
 *
 * Returns a pointer to a prepared RPC client if successful, and
 * @timeout is initialized; caller must destroy a non-NULL returned RPC
 * client.  Otherwise returns NULL, and rpc_createerr.cf_stat is set to
 * reflect the error.
 */
static CLIENT *nfs_get_priv_rpcclient(const struct sockaddr *sap,
			       const socklen_t salen,
			       const unsigned short transport,
			       const rpcprog_t program,
			       const rpcvers_t version,
			       struct timeval *timeout)
{
	nfs_clear_rpc_createerr();

	switch (sap->sa_family) {
	case AF_LOCAL:
		return nfs_get_localclient(sap, salen, program,
						version, timeout);
	case AF_INET:
	case AF_INET6:
		if (nfs_get_port(sap) == 0) {
			rpc_createerr.cf_stat = RPC_UNKNOWNADDR;
			return NULL;
		}
		break;
	default:
		rpc_createerr.cf_stat = RPC_UNKNOWNADDR;
		return NULL;
	}

	switch (transport) {
	case IPPROTO_TCP:
		return nfs_get_tcpclient(sap, salen, program, version,
						timeout, 1);
	case 0:
	case IPPROTO_UDP:
		return nfs_get_udpclient(sap, salen, program, version,
						timeout, 1);
	}

	rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
	return NULL;
}


/* AUTH_SYS doesn't allow more than 16 gids in the supplemental group list.
 * If there are more than that, trying to determine which ones to include
 * in the list is problematic. This function creates an auth handle that
 * only has the primary gid in the supplemental gids list. It's intended to
 * be used for protocols where credentials really don't matter much (the MNT
 * protocol, for instance). */
static AUTH * nfs_authsys_create(void)
{
	char machname[MAXHOSTNAMELEN + 1];
	uid_t   uid = geteuid();
	gid_t   gid = getegid();

	if (gethostname(machname, sizeof(machname)) == -1)
		return NULL;

	return authunix_create(machname, uid, gid, 1, &gid);
}

#ifndef PMAPPROG
#define PMAPPROG        ((rpcprog_t)100000)
#endif

static inline struct sockaddr *nfs_getrpccaller(SVCXPRT *xprt)
{
	return (struct sockaddr *)(char *)svc_getcaller(xprt);
}

#ifndef _SM_INTER_H_RPCGEN
#define _SM_INTER_H_RPCGEN


#define SM_INTER_X

#define SM_PROG 100024
#define SM_VERS 1

#define SM_STAT 1
#define SM_MON 2
#define SM_UNMON 3
#define SM_UNMON_ALL 4
#define SM_SIMU_CRASH 5
#define SM_NOTIFY 6

#endif /* !_SM_INTER_H_RPCGEN */

/* System-dependent declarations */

#ifdef FD_SETSIZE
# define FD_SET_TYPE    fd_set
# define SVC_FDSET      svc_fdset
#else
# define FD_SET_TYPE    int
# define SVC_FDSET      svc_fds
#endif

/* Some timeout values.  (Timeout values are in whole seconds.) */
#define CALLBACK_TIMEOUT         3 /* For client call-backs. */
#define NOTIFY_TIMEOUT           5 /* For status-change notifications. */
#define SELECT_TIMEOUT          10 /* Max select() timeout when work to do. */
#define MAX_TRIES                5 /* Max number of tries for any host. */

#define MODE_NODAEMON 1         /* No-daemon/foreground mode. */
#define MODE_LOG_STDERR 2       /* in foreground mode, log to stderr */
#define MODE_NOTIFY_ONLY 4      /* Send SM_NOTIFY to everyone monitored on
				   a single interface/alias */
/* LH - notify_only mode would be for notifying hosts on an IP alias
 * that just came back up, for ex, when failing over a HA service to
 * another host.... */
#define STATIC_HOSTNAME 8       /* Always use the hostname set by -n */
#define MODE_NO_NOTIFY  16      /* Don't notify peers of a reboot */


/* List-handling macros.
 * THESE INHERIT INFORMATION FROM PREVIOUSLY-DEFINED MACROS.
 * (So don't change their order unless you study them first!) */
#define NL_NEXT(L)      ((L)->next)
#define NL_FIRST        NL_NEXT
#define NL_PREV(L)      ((L)->prev)
#define NL_DATA(L)      ((L)->mon)
#define NL_STATE(L)     ((L)->state)
#define NL_TIMES(L)     ((L)->times)
#define NL_MON_ID(L)    (NL_DATA((L)).mon_id)
#define NL_PRIV(L)      (NL_DATA((L)).priv)
#define NL_MON_NAME(L)  (NL_MON_ID((L)).mon_name)
#define NL_MY_ID(L)     (NL_MON_ID((L)).my_id)
#define NL_MY_NAME(L)   (NL_MY_ID((L)).my_name)
#define NL_MY_PROC(L)   (NL_MY_ID((L)).my_proc)
#define NL_MY_PROG(L)   (NL_MY_ID((L)).my_prog)
#define NL_MY_VERS(L)   (NL_MY_ID((L)).my_vers)
#define NL_WHEN(L)      ((L)->when)

typedef unsigned int (*nsm_populate_t)(const char *hostname,
				const struct sockaddr *sap,
				const struct mon *mon,
				const time_t timestamp);
#define NSM_MAXMSGSIZE  (2048u)


static inline void ha_callout(char *event, char *arg1, char *arg2, int arg3)
{
	char buf[16]; /* should be plenty */
	pid_t pid;
	int ret = -1;
	struct sigaction oldact, newact;

	if (!ha_callout_prog) /* HA callout is not enabled */
		return;

	sprintf(buf, "%d", arg3);

	/* many daemons ignore SIGCHLD as tcpwrappers will
	 * fork a child to do logging.  We need to wait
	 * for a child here, so we need to un-ignore
	 * SIGCHLD temporarily
	 */
	memset(&newact, 0, sizeof(newact));
	newact.sa_handler = SIG_DFL;
	sigaction(SIGCHLD, &newact, &oldact);
	pid = fork();
	switch (pid) {
		case 0: execl(ha_callout_prog, ha_callout_prog,
				event, arg1, arg2,
			      arg3 < 0 ? NULL : buf,
			      NULL);
			bb_default_error_retval = 2;
			bb_perror_msg_and_die("execl");
		case -1: bb_perror_msg("fork");
			break;
		default: pid = waitpid(pid, &ret, 0);
	}
	bb_sigaction2(SIGCHLD, &oldact);
}

static _Bool statd_present_address(const struct sockaddr *sap, char *buf, const size_t buflen)
{
	socklen_t salen;
	int error;

	switch (sap->sa_family) {
	case AF_INET:
		salen = SIZEOF_SOCKADDR_IN;
		break;
	case AF_INET6:
		salen = SIZEOF_SOCKADDR_IN6;
		break;
	default:
		/* unsupported address family */
		return false;
	}

	error = getnameinfo(sap, salen, buf, (socklen_t)buflen,
						NULL, 0, NI_NUMERICHOST);
	if (error != 0) {
		/* getnameinfo(3): gai_strerror(error) */
		return false;
	}
	return true;
}

/* Look up the hostname; report exceptional errors.  Caller must
 * call freeaddrinfo(3) if a valid addrinfo is returned. */
__attribute_malloc__ static struct addrinfo *
get_addrinfo(const char *hostname, const struct addrinfo *hint)
{
	struct addrinfo *ai = NULL;
	int error;

	error = getaddrinfo(hostname, NULL, hint, &ai);
	switch (error) {
	case 0:
		return ai;
	case EAI_NONAME:
		break;
	/* default: failed to resolve host hostname: gai_strerror(error) */
	}
	return NULL;
}

static _Bool get_nameinfo(const struct sockaddr *sap, const socklen_t salen,
		/*@out@*/ char *buf, const socklen_t buflen)
{
	int error;

	error = getnameinfo(sap, salen, buf, buflen, NULL, 0, NI_NAMEREQD);
	if (error != 0) {
		/* failed to resolve address: gai_strerror(error) */
		return false;
	}

	return true;
}

/* statd_canonical_name - choose file name for monitor record files
 * @hostname: C string containing hostname or presentation address
 *
 * Returns a '\0'-terminated ASCII string containing a fully qualified
 * canonical hostname, or NULL if @hostname does not have a reverse
 * mapping.  Caller must free the result with free(3).
 *
 * Incoming hostnames are looked up to determine the canonical hostname,
 * and incoming presentation addresses are converted to canonical
 * hostnames.
 *
 * We won't monitor peers that don't have a reverse map.  The canonical
 * name gives us a key for our monitor list. */
__attribute_malloc__ static char *statd_canonical_name(const char *hostname)
{
	struct addrinfo hint = {
#ifdef CONFIG_FEATURE_IPV6
		.ai_family      = AF_UNSPEC,
#else   /* !CONFIG_FEATURE_IPV6 */
		.ai_family      = AF_INET,
#endif  /* !CONFIG_FEATURE_IPV6 */
		.ai_flags       = AI_NUMERICHOST,
		.ai_protocol    = (int)IPPROTO_UDP,
	};
	char buf[NI_MAXHOST];
	struct addrinfo *ai;

	ai = get_addrinfo(hostname, &hint);
	if (ai != NULL) {
		/* @hostname was a presentation address */
		_Bool result;
		result = get_nameinfo(ai->ai_addr, ai->ai_addrlen,
					buf, (socklen_t)sizeof(buf));
		freeaddrinfo(ai);
		if (!result)
			/* OK to use presentation address,
			 * if no reverse map exists */
			return strdup(hostname);
		return strdup(buf);
	}

	/* @hostname was a hostname */
	hint.ai_flags = AI_CANONNAME;
	ai = get_addrinfo(hostname, &hint);
	if (ai == NULL)
		return NULL;
	strcpy(buf, ai->ai_canonname);
	freeaddrinfo(ai);

	return strdup(buf);
}

/* statd_matchhostname - check if two hostnames are equivalent
 * @hostname1: C string containing hostname
 * @hostname2: C string containing hostname
 *
 * Returns true if the hostnames are the same, the hostnames resolve
 * to the same canonical name, or the hostnames resolve to at least
 * one address that is the same.  False is returned if the hostnames
 * do not match in any of these ways, if either hostname contains
 * wildcard characters, if either hostname is a netgroup name, or
 * if an error occurs. */
static _Bool statd_matchhostname(const char *hostname1, const char *hostname2)
{
	struct addrinfo *ai1, *ai2, *results1 = NULL, *results2 = NULL;
	struct addrinfo hint = {
		.ai_family      = AF_UNSPEC,
		.ai_flags       = AI_CANONNAME,
		.ai_protocol    = (int)IPPROTO_UDP,
	};
	_Bool result = false;

	if (strcasecmp(hostname1, hostname2) == 0) {
		result = true;
		goto out;
	}

	results1 = get_addrinfo(hostname1, &hint);
	if (results1 == NULL)
		goto out;
	results2 = get_addrinfo(hostname2, &hint);
	if (results2 == NULL)
		goto out;

	if (strcasecmp(results1->ai_canonname, results2->ai_canonname) == 0) {
		result = true;
		goto out;
	}

	for (ai1 = results1; ai1 != NULL; ai1 = ai1->ai_next)
		for (ai2 = results2; ai2 != NULL; ai2 = ai2->ai_next)
			if (nfs_compare_sockaddr(ai1->ai_addr, ai2->ai_addr)) {
				result = true;
				break;
			}

out:
	freeaddrinfo(results2);
	freeaddrinfo(results1);
	/* hostnames (result ? matched : did not match) */
	return result;
}


static void closeall(int min)
{
	char *endp;
	long n;
	DIR *dir = opendir("/proc/self/fd");

	if (dir != NULL) {
		int dfd = dirfd(dir);
		struct dirent *d;

		while ((d = readdir(dir)) != NULL) {
			errno = 0;
			n = strtol(d->d_name, &endp, 10);
			if (!errno && *endp == '\0' && endp != d->d_name &&
			    n >= min && n != dfd)
				(void) close(n);
		}
		closedir(dir);
	} else {
		int fd = sysconf(_SC_OPEN_MAX);
		while (--fd >= min)
			(void) close(fd);
	}
}

/* Simple list management for notify list */

/* Allocate memory and set up a new notify list entry. */
static notify_list *nlist_new(char *my_name, char *mon_name, int state)
{
	notify_list     *new;

	new = (notify_list *) xmalloc(sizeof(notify_list));
	memset(new, 0, sizeof(*new));

	NL_TIMES(new) = MAX_TRIES;
	NL_STATE(new) = state;
	NL_MY_NAME(new) = xstrdup(my_name);
	NL_MON_NAME(new) = xstrdup(mon_name);

	return new;
}

/* Insert *entry into a notify list at the point specified by
 * **head.  This can be in the middle.  However, we do not handle
 * list _append_ in this function; rather, the only place we should
 * have to worry about this case is in nlist_insert_timer below.
 * - entry must not be NULL. */
static void nlist_insert(notify_list **head, notify_list *entry)
{
	if (*head) {
		/* Cases where we're prepending a non-empty list
		 * or inserting possibly in the middle somewhere (eg,
		 * nlist_insert_timer...) */
		entry->next = (*head);          /* Forward pointer */
		entry->prev = (*head)->prev;    /* Back pointer */
		(*head)->prev = entry;          /* head's new back pointer */
	}

	/* Common to all cases, including new list creation */
	*head = entry;                  /* New head */
}

/* (re)insert *entry into notify_list **head.  This requires that
 * NL_WHEN(entry) has been set (usually, this is time() + 5 seconds).
 * - entry must not be NULL
 *
 * LH - This used to cause (a) a memory leak and (b) dropped notify-list
 * entries.  The pointer ran off the end of the list, and changed the
 * head-end to point to the new, one-entry list.  All other entries became garbage.
 *
 * FIXME: Optimize this function. (I'll work on it - LH) */
static void nlist_insert_timer(notify_list **head, notify_list *entry)
{
	notify_list     *spot = *head,          /* Insertion location */
						/* ...Start at head */
			*back = NULL;           /* Back pointer */


	/* Find first entry with higher timeout value or end of list */
	while (spot && NL_WHEN(spot) <= NL_WHEN(entry)) {
		/*
		 * Keep the back pointer in case we
		 * run off the end...  (see below)
		 */
		back = spot;
		spot = spot->next;
	}

	if (spot == (*head)) {
		/*
		 * case where we're prepending an empty or non-empty
		 * list or inserting in the middle somewhere.  Pass
		 * the real head of the list, since we'll be changing
		 * during the insert...
		 */
		nlist_insert(head, entry);
	} else {
		/* all other cases - don't move the real head pointer */
		nlist_insert(&spot, entry);

		/*
		 * If spot == entry, then spot was NULL when we called
		 * nlist_insert.  This happened because we had run off
		 * the end of the list.  Append entry to original list.
		 */
		if (spot == entry) {
			back->next = entry;
			entry->prev = back;
		}
	}
}

/* Remove *entry from the list pointed to by **head.
 * Do not destroy *entry.  This is normally done before
 * a re-insertion with a timer, but can be done anywhere.
 * - entry must not be NULL. */
static void nlist_remove(notify_list **head, notify_list *entry)
{
	notify_list     *prev = entry->prev,
			*next = entry->next;

	if (next) {
		next->prev = prev;
	}

	if (prev) {
		/* Case(s) where entry isn't at the front */
		prev->next = next;
	} else {
		/* cases where entry is at the front */
		*head = next;
	}

	entry->next = entry->prev = NULL;
}

/* Clone an entry in the notify list - * - entry must not be NULL */
static notify_list * nlist_clone(notify_list *entry)
{
	notify_list     *new;

	new = nlist_new(NL_MY_NAME(entry), NL_MON_NAME(entry), NL_STATE(entry));
	NL_MY_PROG(new) = NL_MY_PROG(entry);
	NL_MY_VERS(new) = NL_MY_VERS(entry);
	NL_MY_PROC(new) = NL_MY_PROC(entry);
	memcpy(NL_PRIV(new), NL_PRIV(entry), SM_PRIV_SIZE);

	return new;
}

/* Destroy an entry in a notify list and free the memory.
 * If *head is NULL, just free the entry.  This would be
 * done only when we know entry isn't in any list.
 * - entry must not be NULL. */
static void nlist_free(notify_list **head, notify_list *entry)
{
	if (head && (*head))
		nlist_remove(head, entry);
	if (NL_MY_NAME(entry))
		free(NL_MY_NAME(entry));
	if (NL_MON_NAME(entry))
		free(NL_MON_NAME(entry));
	free(entry->dns_name);
	free(entry);
}

/* Destroy an entire notify list */
static void nlist_kill(notify_list **head)
{
	while (*head)
		nlist_free(head, *head);
}

/* Walk a list looking for a matching name in the NL_MON_NAME field. */
static notify_list *nlist_gethost(notify_list *list, char *host, int myname)
{
	notify_list     *lp;

	for (lp = list; lp; lp = lp->next) {
		if (statd_matchhostname(host,
				myname? NL_MY_NAME(lp) : NL_MON_NAME(lp)))
			return lp;
	}

	return (notify_list *) NULL;
}

/* Callback notify list. */
/* notify_list *cbnl = NULL; ... never used */


/*
 * Services SM_NOTIFY requests.
 *
 * When NLM uses an SM_MON request to tell statd to monitor a remote,
 * the request contains a "mon_name" argument.  This is usually the
 * "caller_name" argument of an NLMPROC_LOCK request.  On Linux, the
 * NLM can send statd the remote's IP address instead of its
 * caller_name.  The NSM protocol does not allow both the remote's
 * caller_name and it's IP address to be sent in the same SM_MON
 * request.
 *
 * The remote's caller_name is useful because it makes it simple
 * to identify rebooting remotes by matching the "mon_name" argument
 * they sent via an SM_NOTIFY request.
 *
 * The caller_name string may not be a fully qualified domain name,
 * or even registered in the DNS database, however.  Having the
 * remote's IP address is useful because then there is no ambiguity
 * about where to send an SM_NOTIFY after the local system reboots.
 *
 * Without the actual caller_name, however, statd must use an
 * heuristic to match an incoming SM_NOTIFY request to one of the
 * hosts it is currently monitoring.  The incoming mon_name in an
 * SM_NOTIFY address is converted to a list of IP addresses using
 * DNS.  Each mon_name on statd's monitor list is also converted to
 * an address list, and the two lists are checked to see if there is
 * a matching address.
 *
 * There are some risks to this strategy:
 *
 *   1.  The external DNS database is not reliable.  It can change
 *       over time, or the forward and reverse mappings could be
 *       inconsistent.
 *
 *   2.  If statd's monitor list becomes substantial, finding a match
 *       can generate a not inconsequential amount of DNS traffic.
 *
 *   3.  statd is a single-threaded service.  When DNS becomes slow or
 *       unresponsive, statd also becomes slow or unresponsive.
 *
 *   4.  If the remote does not have a DNS entry at all (or if the
 *       remote can resolve itself, but the local host can't resolve
 *       the remote's hostname), the remote cannot be monitored, and
 *       therefore NLM locking cannot be provided for that host.
 *
 *   5.  Local DNS resolution can produce different results for the
 *       mon_name than the results the remote might see for the same
 *       query, especially if the remote did not send a caller_name
 *       or mon_name that is a fully qualified domain name.
 *
 *       Note that a caller_name is passed from NFS client to server,
 *       but the client never knows what mon_name the server might use
 *       to notify it of a reboot.  On Linux, the client extracts the
 *       server's name from the devname it was passed by the mount
 *       command.  This is often not a fully-qualified domain name.
 */
static void *sm_notify_1_svc(struct stat_chge *argp, struct svc_req *rqstp)
{
	notify_list    *lp, *call;
	struct sockaddr *sap = nfs_getrpccaller(rqstp->rq_xprt);
	char            sm_notify_ip_addr[INET6_ADDRSTRLEN];

	/* "Received SM_NOTIFY from argp->mon_name, state: argp->state" */

	/* quick check - don't bother if we're not monitoring anyone */
	if (rtnl == NULL) {
		xlog_warn("SM_NOTIFY from %s while not monitoring any hosts",
				argp->mon_name);
		return ((void *) &result1);
	}

	if (!statd_present_address(sap, sm_notify_ip_addr, sizeof(sm_notify_ip_addr))) {
		xlog_warn("Unrecognized sender address");
		return ((void *) &result1);
	}

	/* okir change: statd doesn't remove the remote host from its
	 * internal monitor list when receiving an SM_NOTIFY call from
	 * it. Lockd will want to continue monitoring the remote host
	 * until it issues an SM_UNMON call.
	 */
	for (lp = rtnl ; lp ; lp = lp->next)
		if (NL_STATE(lp) != argp->state &&
		    (statd_matchhostname(argp->mon_name, lp->dns_name) ||
		     statd_matchhostname(sm_notify_ip_addr, lp->dns_name))) {
			NL_STATE(lp) = argp->state;
			call = nlist_clone(lp);
			nlist_insert(&statd_notify, call);
		}


	return ((void *) &result1);
}



/* Reject requests from non-loopback addresses in order
 * to prevent attack described in CERT CA-99.05.
 *
 * Although the kernel contacts the statd service via only IPv4
 * transports, the statd service can receive other requests, such
 * as SM_NOTIFY, from remote peers via IPv6. */
static _Bool caller_is_localhost(struct svc_req *rqstp)
{
	struct sockaddr *sap = nfs_getrpccaller(rqstp->rq_xprt);
	char buf[INET6_ADDRSTRLEN];

	if (!nfs_is_v4_loopback(sap))
		goto out_nonlocal;
	return true;

out_nonlocal:
	if (!statd_present_address(sap, buf, sizeof(buf)))
		buf[0] = '\0';
	xlog_warn("SM_MON/SM_UNMON call from non-local host %s", buf);
	return false;
}


static unsigned int load_one_host(const char *hostname,
		const struct sockaddr *UNUSED(sap),
		const struct mon *m,
		const time_t UNUSED(timestamp))
{
	notify_list *clnt;

	clnt = nlist_new(m->mon_id.my_id.my_name,
				m->mon_id.mon_name, 0);
	if (clnt == NULL)
		return 0;

	clnt->dns_name = strdup(hostname);
	if (clnt->dns_name == NULL) {
		nlist_free(NULL, clnt);
		return 0;
	}

	/* Adding record for hostname to the monitor list... */
	NL_MY_PROG(clnt) = m->mon_id.my_id.my_prog;
	NL_MY_VERS(clnt) = m->mon_id.my_id.my_vers;
	NL_MY_PROC(clnt) = m->mon_id.my_id.my_proc;
	memcpy(NL_PRIV(clnt), m->priv, SM_PRIV_SIZE);

	nlist_insert(&rtnl, clnt);
	return 1;
}


/* After reboot, notify all hosts on our notify list. In order not to
 * hang statd with delivery to dead hosts, we perform all RPC calls in
 * parallel.
 *
 * It would have been nice to use the portmapper's rmtcall feature,
 * but that's not possible for security reasons (the portmapper would
 * have to forward the call with root privs for most statd's, which
 * it won't if it's worth its money). */


/* Initialize socket used to notify lockd of peer reboots.
 *
 * Returns the file descriptor of the new socket if successful;
 * otherwise returns -1 and logs an error.
 *
 * Lockd rejects such requests if the source port is not privileged.
 * statd_get_socket() must be invoked while statd still holds root
 * privileges in order for the socket to acquire a privileged source
 * port. */
static int statd_get_socket(void)
{
	struct sockaddr_in      sin;
	struct servent *se;
	int loopcnt = 100;

	if (sockfd >= 0)
		return sockfd;

	while (loopcnt-- > 0) {

		if (sockfd >= 0) close(sockfd);

		if ((sockfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
			bb_perror_msg("%s: Can't create socket", __func__);
			return -1;
		}


		memset(&sin, 0, sizeof(sin));
		sin.sin_family = AF_INET;
		sin.sin_addr.s_addr = INADDR_ANY;

		if (bindresvport(sockfd, &sin) < 0) {
			/* can't bind to reserved port */
			break;
		}
		se = getservbyport(sin.sin_port, "udp");
		if (se == NULL)
			break;
		/* rather not use that port, try again */
	}
	FD_SET(sockfd, &SVC_FDSET);
	return sockfd;
}

/* nsm_parse_reply - parse and validate the header in an RPC reply
 * @xdrs: pointer to XDR
 *
 * Returns the XID of the reply, or zero if an error occurred. */
static uint32_t nsm_parse_reply(XDR *xdrs)
{
	struct rpc_msg mesg = {
		.rm_reply.rp_acpt.ar_results.proc = (xdrproc_t)xdr_void,
	};
	uint32_t xid;

	if (xdr_replymsg(xdrs, &mesg) == FALSE) {
		bb_error_msg("%s: can't decode RPC reply", __func__);
		return 0;
	}
	xid = (uint32_t)mesg.rm_xid;

	if (mesg.rm_reply.rp_stat != MSG_ACCEPTED) {
		bb_error_msg("%s: [0x%x] RPC status %d",
			__func__, xid, mesg.rm_reply.rp_stat);
		return 0;
	}

	if (mesg.rm_reply.rp_acpt.ar_stat != SUCCESS) {
		bb_error_msg("%s: [0x%x] RPC accept status %d",
			__func__, xid, mesg.rm_reply.rp_acpt.ar_stat);
		return 0;
	}

	return xid;
}

/* nsm_recv_getport - parse PMAP_GETPORT reply
 * @xdrs: pointer to XDR
 *
 * Returns the port number from the RPC reply, or zero
 * if an error occurred. */
static unsigned long nsm_recv_getport(XDR *xdrs)
{
	unsigned long port = 0;

	if (xdr_u_long(xdrs, &port) == FALSE)
		bb_error_msg("%s: can't decode pmap reply", __func__);
	if (port > UINT16_MAX) {
		bb_error_msg("%s: bad port number", __func__);
		port = 0;
	}

	/* Received PMAP_GETPORT result: port */
	return port;
}

static notify_list *recv_rply(u_long *portp)
{
	char                    msgbuf[NSM_MAXMSGSIZE];
	ssize_t                 msglen;
	notify_list             *lp = NULL;
	XDR                     xdr;
	struct sockaddr_in      sin;
	socklen_t               alen = (socklen_t)sizeof(sin);
	uint32_t                xid;

	memset(msgbuf, 0, sizeof(msgbuf));
	msglen = recvfrom(sockfd, msgbuf, sizeof(msgbuf), 0,
				(struct sockaddr *)(char *)&sin, &alen);
	if (msglen == (ssize_t)-1) {
		xlog_warn("%s: recvfrom failed: %m", __func__);
		return NULL;
	}

	memset(&xdr, 0, sizeof(xdr));
	xdrmem_create(&xdr, msgbuf, (unsigned int)msglen, XDR_DECODE);
	xid = nsm_parse_reply(&xdr);
	if (xid == 0)
		goto done;
	if (sin.sin_addr.s_addr != htonl(INADDR_LOOPBACK)) {
		struct in_addr addr = sin.sin_addr;
		char buf[INET_ADDRSTRLEN];

		xlog_warn("%s: Unrecognized reply from %s", __func__,
				inet_ntop(AF_INET, &addr, buf,
						(socklen_t)sizeof(buf)));
		goto done;
	}

	for (lp = statd_notify; lp != NULL; lp = lp->next) {
		/* LH - this was a bug... it should have been checking
		 * the xid from the response message from the client,
		 * not the static, internal xid */
		if (lp->xid != xid)
			continue;
		if (lp->port == 0)
			*portp = nsm_recv_getport(&xdr);
		break;
	}

done:
	xdr_destroy(&xdr);
	return lp;
}



/* Services SM_SIMU_CRASH requests.
 *
 * Although the kernel contacts the statd service via only IPv4
 * transports, the statd service can receive other requests, such
 * as SM_NOTIFY, from remote peers via IPv6. */
static void *
sm_simu_crash_1_svc (void *UNUSED(argp), struct svc_req *rqstp)
{
  struct sockaddr *sap = nfs_getrpccaller(rqstp->rq_xprt);
  char buf[INET6_ADDRSTRLEN];

  /* Received SM_SIMU_CRASH */
  if (!nfs_is_v4_loopback(sap))
    goto out_nonlocal;

  if ((int)nfs_get_port(sap) >= IPPORT_RESERVED) {
    xlog_warn("SM_SIMU_CRASH call from unprivileged port");
    goto failure;
  }

  svc_stop = 1;

  if (rtnl)
    nlist_kill (&rtnl);

 failure:
  return ((void *)&result2);

 out_nonlocal:
  if (!statd_present_address(sap, buf, sizeof(buf)))
    buf[0] = '\0';
  xlog_warn("SM_SIMU_CRASH call from non-local host %s", buf);
  goto failure;
}


/* Services SM_STAT requests.
 *
 * According the the X/Open spec's on this procedure: "Implementations
 * should not rely on this procedure being operative.  In many current
 * implementations of the NSM it will always return a 'STAT_FAIL'
 * status."  My implementation is operative; it returns 'STAT_SUCC'
 * whenever it can resolve the hostname that it's being asked to
 * monitor, and returns 'STAT_FAIL' otherwise.
 *
 * sm_inter.x says the 'state' returned should be
 *   "state number of site sm_name".  It is not clear how to get this.
 * X/Open says:
 *   STAT_SUCC
 *      The NSM will monitor the given host. "sm_stat_res.state" contains
 *      the state of the NSM.
 * Which implies that 'state' is the state number of the *local* NSM.
 * href=http://www.opengroup.org/onlinepubs/9629799/SM_STAT.htm
 *
 * We return the *local* state as
 *   1/ We have easy access to it.
 *   2/ It might be useful to a remote client who needs it and has no
 *      other way to get it.
 *   3/ That's what we always did in the past.
 */
static struct sm_stat_res *
sm_stat_1_svc(struct sm_name *argp, struct svc_req *UNUSED(rqstp))
{
  char *name;

  /* Received SM_STAT from argp->mon_name */
  name = statd_canonical_name(argp->mon_name);
  if (name == NULL) {
    result3.res_stat = STAT_FAIL;
  } else {
    result3.res_stat = STAT_SUCC;
    free(name);
  }
  result3.state = MY_STATE;
  return(&result3);
}

#ifdef CONFIG_FEATURE_LONG_OPTIONS
static const struct option statd_longopts[] =
{
	{ "foreground", 0, 0, 'F' },
	{ "no-syslog", 0, 0, 'd' },
	{ "help", 0, 0, 'h' },
	{ "version", 0, 0, 'v' },
	{ "outgoing-port", 1, 0, 'o' },
	{ "port", 1, 0, 'p' },
	{ "name", 1, 0, 'n' },
	{ "state-directory-path", 1, 0, 'P' },
	{ "notify-mode", 0, 0, 'N' },
	{ "ha-callout", 1, 0, 'H' },
	{ "no-notify", 0, 0, 'L' },
	{ NULL, 0, 0, 0 }
};
#endif

static void sigusr (int UNUSED(sig))
{
	/* Caught signal sig, re-notifying (state MY_STATE) */
	svc_stop = 1;
}

/* Startup information. */
static void log_modes(void)
{
	/* No flags = no message */
	if (!run_mode) return;
	xlog_warn("Flags: %s%s",
			(run_mode & MODE_NODAEMON)? "No-Daemon " : "",
			(run_mode & MODE_LOG_STDERR)? "Log-STDERR " : "");
}

static const char pidfile[] = "/var/run/rpc.statd.pid";

static void create_pidfile(void)
{
	FILE *fp;

	unlink(pidfile);
	fp = bb_xfopen(pidfile, "Aw");
	fprintf(fp, "%d\n", getpid());
	pidfd = dup(fileno(fp));
	if (fclose(fp) < 0)
		xlog_warn("Flushing pid file failed: errno %d (%m)", errno);
}

static void truncate_pidfile(void)
{
	if (pidfd >= 0) {
		if (ftruncate(pidfd, 0) < 0)
			xlog_warn("truncating pid file failed: errno %d (%m)", errno);
	}
}

static void run_sm_notify(int outport)
{
	char op[20];
	char *av[6];
	int ac = 0;

	av[ac++] = "sm-notify";
	if (run_mode & MODE_NODAEMON)
		av[ac++] = "-d";
	if (outport) {
		sprintf(op, "-p%d", outport);
		av[ac++] = op;
	}
	if (run_mode & STATIC_HOSTNAME) {
		av[ac++] = "-v";
		av[ac++] = MY_NAME;
	}
	av[ac] = NULL;
	execv("/bin/busybox", av);
	bb_default_error_retval = 2;
	bb_error_msg_and_die("failed to run %s", av[0]);
}


#define _RPCSVC_CLOSEDOWN       120

static void
closedown(int sig)
{
	int i, openfd;

	(void) signal(sig, closedown);

	if (_rpcfdtype == SOCK_DGRAM)
		exit(0);

	if (st_getdtablesize == 0)
		st_getdtablesize = getdtablesize();

	for (i = 0, openfd = 0; i < st_getdtablesize && openfd < 2; i++)
		if (FD_ISSET(i, &svc_fdset))
			openfd++;
	if (openfd <= 1)
		exit(0);
	(void) alarm(_RPCSVC_CLOSEDOWN);
}

/* Create listener socket for a given port
 *
 * Return an open network socket on success; otherwise return -1
 * if some error occurs. */
static int makesock(int port, int proto)
{
	struct sockaddr_in sin;
	int     sock = bb_xsocket(((proto == IPPROTO_UDP) ? BB_XSOCKET_FLG_SOCK_DGRAM : 0 )
				| BB_XSOCKET_FLG_ERR_DIE);
	memset((char *) &sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl(INADDR_ANY);
	sin.sin_port = htons(port);

	if (proto == IPPROTO_TCP)
		if (setsockopt_1i(sock, SOL_SOCKET, SO_REUSEADDR, 1) < 0)
			bb_perror_msg("setsockopt failed");

	bb_xbind(sock, (struct sockaddr *) &sin, sizeof(sin));

	return sock;
}

static int svc_socket (u_long number, int type, int protocol, int reuse)
{
  struct sockaddr_in addr;
  socklen_t len = sizeof (struct sockaddr_in);
  char rpcdata [1024], servdata [1024];
  struct rpcent rpcbuf, *rpcp;
  struct servent servbuf, *servp = NULL;
  int sock, ret;
  const char *proto = protocol == IPPROTO_TCP ? "tcp" : "udp";

  if ((sock = socket (AF_INET, type, protocol)) < 0) {
      bb_perror_msg ("svc_socket: socket creation problem");
      return sock;
  }

  if (reuse) {
      ret = setsockopt_1i (sock, SOL_SOCKET, SO_REUSEADDR, 1);
      if (ret < 0) {
	  bb_perror_msg ("svc_socket: socket reuse problem");
	  return ret;
      }
  }

  memset (&addr, 0, sizeof (addr));
  addr.sin_family = AF_INET;

  ret = getrpcbynumber_r (number, &rpcbuf, rpcdata, sizeof rpcdata,
			  &rpcp);
  if (ret == 0 && rpcp != NULL) {
      /* First try name.  */
      ret = getservbyname_r (rpcp->r_name, proto, &servbuf, servdata,
			     sizeof servdata, &servp);
      if ((ret != 0 || servp == NULL) && rpcp->r_aliases) {
	  const char **a;

	  /* Then we try aliases.  */
	  for (a = (const char **) rpcp->r_aliases; *a != NULL; a++) {
	      ret = getservbyname_r (*a, proto, &servbuf, servdata,
				     sizeof servdata, &servp);
	      if (ret == 0 && servp != NULL)
		break;
	  }
      }
  }

  if (ret == 0 && servp != NULL) {
      addr.sin_port = servp->s_port;
      if (bind (sock, (struct sockaddr *) &addr, len) < 0) {
	  bb_perror_msg ("svc_socket: bind problem");
	  (void) close (sock);
	  sock = -1;
      }
  } else {
	  addr.sin_port = 0;
	  if (bind (sock, (struct sockaddr *) &addr, len) < 0) {
	      bb_perror_msg ("svc_socket: bind problem");
	      (void) close (sock);
	      sock = -1;
	  }
  }

  if (sock >= 0) {
	    /* This socket might be shared among multiple processes
	     * if mountd is run multi-threaded.  So it is safest to
	     * make it non-blocking, else all threads might wake
	     * one will get the data, and the others will block
	     * indefinitely.
	     * In all cases, transaction on this socket are atomic
	     * (accept for TCP, packet-read and packet-write for UDP)
	     * so O_NONBLOCK will not confuse unprepared code causing
	     * it to corrupt messages.
	     * It generally safest to have O_NONBLOCK when doing an accept
	     * as if we get a RST after the SYN and before accept runs,
	     * we can block despite being told there was an acceptable
	     * connection.
	     */
	int flags;
	if ((flags = fcntl(sock, F_GETFL)) < 0) {
	      bb_perror_msg ("svc_socket: can't get socket flags");
	      (void) close (sock);
	      sock = -1;
	} else if (fcntl(sock, F_SETFL, flags|O_NONBLOCK) < 0) {
	      bb_perror_msg ("svcsocket: can't set socket flags");
	      (void) close (sock);
	      sock = -1;
	}
  }

  return sock;
}

/*
 * Create and bind a TCP socket based on program number
 */
static int svctcp_socket (u_long number, int reuse)
{
  return svc_socket (number, SOCK_STREAM, IPPROTO_TCP, reuse);
}

/*
 * Create and bind a UDP socket based on program number
 */
static int svcudp_socket (u_long number)
{
  return svc_socket (number, SOCK_DGRAM, IPPROTO_UDP, FALSE);
}

static void rpc_init(char *name, int prog, int vers,
	 void (*dispatch)(struct svc_req *, register SVCXPRT *),
	 int defport)
{
	struct sockaddr_in saddr;
	SVCXPRT *transp;
	int     sock;
	socklen_t asize;

	asize = sizeof(saddr);
	sock = 0;
	if (getsockname(0, (struct sockaddr *) &saddr, &asize) == 0
	    && saddr.sin_family == AF_INET) {
		socklen_t ssize = sizeof(int);
		int fdtype = 0;
		if (getsockopt(0, SOL_SOCKET, SO_TYPE,
				(char *)&fdtype, &ssize) == -1)
			bb_perror_msg_and_die("getsockopt failed");
		/* inetd passes a UDP socket or a listening TCP socket.
		 * listen will fail on a connected TCP socket(passed by rsh).
		 */
		if (!(fdtype == SOCK_STREAM && listen(0,5) == -1)) {
			_rpcfdtype = fdtype;
			_rpcpmstart = 1;
		}
	}
	if (!_rpcpmstart) {
		pmap_unset(prog, vers);
		sock = RPC_ANYSOCK;
	}

	if ((_rpcfdtype == 0) || (_rpcfdtype == SOCK_DGRAM)) {
		if (_rpcpmstart == 0) {
			if (last_transp_SOCK_DGRAM
			    && (!defport || defport == last_transp_SOCK_DGRAM->xp_port)) {
				transp = last_transp_SOCK_DGRAM;
				goto udp_transport;
			}
			if (defport == 0)
				sock = RPC_ANYSOCK;
			else
				sock = makesock(defport, IPPROTO_UDP);
		}
		if (sock == RPC_ANYSOCK)
			sock = svcudp_socket (prog);
		transp = svcudp_create(sock);
		if (transp == NULL) {
			bb_error_msg_and_die("cannot create udp service.");
		}
      udp_transport:
		if (!svc_register(transp, prog, vers, dispatch, IPPROTO_UDP)) {
			bb_error_msg_and_die("unable to register (%s, %d, udp).",
					name, vers);
		}
		last_transp_SOCK_DGRAM = transp;
	}

	if ((_rpcfdtype == 0) || (_rpcfdtype == SOCK_STREAM)) {
		if (_rpcpmstart == 0) {
			if (last_transp_SOCK_STREAM
			    && (!defport || defport == last_transp_SOCK_STREAM->xp_port)) {
				transp = last_transp_SOCK_STREAM;
				goto tcp_transport;
			}
			if (defport == 0)
				sock = RPC_ANYSOCK;
			else
				sock = makesock(defport, IPPROTO_TCP);
		}
		if (sock == RPC_ANYSOCK)
			sock = svctcp_socket (prog, 1);
		transp = svctcp_create(sock, 0, 0);
		if (transp == NULL) {
			bb_error_msg_and_die("cannot create tcp service.");
		}
      tcp_transport:
		if (!svc_register(transp, prog, vers, dispatch, IPPROTO_TCP)) {
			bb_error_msg_and_die("unable to register (%s, %d, tcp).",
					name, vers);
		}
		last_transp_SOCK_STREAM = transp;
	}

	if (_rpcpmstart) {
		signal(SIGALRM, closedown);
		alarm(_RPCSVC_CLOSEDOWN);
	}
}

/* nfs_svc_create - start up RPC svc listeners
 * @name: C string containing name of new service
 * @program: RPC program number to register
 * @version: RPC version number to register
 * @dispatch: address of function that handles incoming RPC requests
 * @port: if not zero, transport listens on this port
 *
 * Sets up network transports for receiving RPC requests, and starts
 * the RPC dispatcher.  Returns the number of started network transports. */
static unsigned int
nfs_svc_create(char *name, const rpcprog_t program, const rpcvers_t version,
		void (*dispatch)(struct svc_req *, SVCXPRT *),
		const uint16_t port)
{
	rpc_init(name, (int)program, (int)version, dispatch, (int)port);
	return 1;
}

/* nfs_svc_unregister - remove service registrations from local rpcbind database
 * @program: RPC program number to unregister
 * @version: RPC version number to unregister
 *
 * Removes all registrations for [ @program, @version ] . */
static void
nfs_svc_unregister(const rpcprog_t program, const rpcvers_t version)
{
	pmap_unset((unsigned long)program, (unsigned long)version);
	/* if false: Failed to unregister program version=version */
}

/* NSM for Linux.
 *
 * Callback information and NSM state is stored in files, usually
 * under /var/lib/nfs.  A database of information contained in local
 * files stores NLM callback data and what remote peers to notify of
 * reboots.
 *
 * For each monitored remote peer, a text file is created under the
 * directory specified by NSM_MONITOR_DIR.  The name of the file
 * is a valid DNS hostname.  The hostname string must be a valid
 * ASCII DNS name, and must not contain slash characters, white space,
 * or '\0' (ie. anything that might have some special meaning in a
 * path name).
 *
 * The contents of each file include seven blank-separated fields of
 * text, finished with '\n'.  The first field contains the network
 * address of the NLM service to call back.  The current implementation
 * supports using only IPv4 addresses, so the only contents of this
 * field are a network order IPv4 address expressed in 8 hexadecimal
 * characters.
 *
 * The next four fields are text strings of hexadecimal characters,
 * representing:
 *
 * 2. A 4 byte RPC program number of the NLM service to call back
 * 3. A 4 byte RPC version number of the NLM service to call back
 * 4. A 4 byte RPC procedure number of the NLM service to call back
 * 5. A 16 byte opaque cookie that the NLM service uses to identify
 *    the monitored host
 *
 * The sixth field is the monitored host's mon_name, passed to statd
 * via an SM_MON request.
 *
 * The seventh field is the my_name for this peer, which is the
 * hostname of the local NLM (currently on Linux, the result of
 * `uname -n`).  This can be used as the source address/hostname
 * when sending SM_NOTIFY requests.
 *
 * The NSM protocol does not limit the contents of these strings
 * in any way except that they must fit into 1024 bytes.  Our
 * implementation requires that these strings not contain
 * white space or '\0'. */


#define RPCARGSLEN      (4 * (8 + 1))
#define LINELEN         (RPCARGSLEN + SM_PRIV_SIZE * 2 + 1)

#define NSM_KERNEL_STATE_FILE   "/proc/sys/fs/nfs/nsm_local_state"

/*
 * Some distributions place statd's files in a subdirectory
 */
#define NSM_PATH_EXTENSION
/* #define NSM_PATH_EXTENSION   "/statd" */

#define NSM_DEFAULT_STATEDIR            NFS_STATEDIR NSM_PATH_EXTENSION


#define NSM_MONITOR_DIR "sm"
#define NSM_NOTIFY_DIR  "sm.bak"
#define NSM_STATE_FILE  "state"


static _Bool error_check(const int len, const size_t buflen)
{
	return (len < 0) || ((size_t)len >= buflen);
}

static _Bool exact_error_check(const ssize_t len, const size_t buflen)
{
	return (len < 0) || ((size_t)len != buflen);
}

/* Returns a dynamically allocated, '\0'-terminated buffer
 * containing an appropriate pathname, or NULL if an error
 * occurs.  Caller must free the returned result with free(3). */
static __attribute_malloc__ char *
nsm_make_record_pathname(const char *directory, const char *hostname)
{
	const char *c;
	size_t size;
	char *path;
	int len;

	/*
	 * Block hostnames that contain characters that have
	 * meaning to the file system (like '/'), or that can
	 * be confusing on visual inspection (like ' ').
	 */
	for (c = hostname; *c != '\0'; c++)
		if (*c == '/' || isspace((int)*c) != 0) {
			/* Hostname contains invalid characters */
			return NULL;
		}

	size = strlen(nsm_base_dirname) + strlen(directory) + strlen(hostname) + 3;
	if (size > PATH_MAX) {
		/* Hostname results in pathname that is too long */
		return NULL;
	}

	path = malloc(size);
	if (path == NULL) {
		/* Failed to allocate memory for pathname */
		return NULL;
	}

	len = snprintf(path, size, "%s/%s/%s",
			nsm_base_dirname, directory, hostname);
	if (error_check(len, size)) {
		/* Pathname did not fit in specified buffer */
		free(path);
		return NULL;
	}

	return path;
}

/* Returns a dynamically allocated, '\0'-terminated buffer
 * containing an appropriate pathname, or NULL if an error
 * occurs.  Caller must free the returned result with free(3). */
static __attribute_malloc__ char *nsm_make_pathname(const char *directory)
{
	size_t size;
	char *path;
	int len;

	size = strlen(nsm_base_dirname) + strlen(directory) + 2;
	if (size > PATH_MAX)
		return NULL;

	path = malloc(size);
	if (path == NULL)
		return NULL;

	len = snprintf(path, size, "%s/%s", nsm_base_dirname, directory);
	if (error_check(len, size)) {
		free(path);
		return NULL;
	}

	return path;
}

/* Returns a dynamically allocated, '\0'-terminated buffer
 * containing an appropriate pathname, or NULL if an error
 * occurs.  Caller must free the returned result with free(3). */
static __attribute_malloc__ char *nsm_make_temp_pathname(const char *pathname)
{
	size_t size;
	char *path;
	int len;

	size = strlen(pathname) + sizeof(".new") + 2;
	if (size > PATH_MAX)
		return NULL;

	path = malloc(size);
	if (path == NULL)
		return NULL;

	len = snprintf(path, size, "%s.new", pathname);
	if (error_check(len, size)) {
		free(path);
		return NULL;
	}

	return path;
}

/* Use "mktemp, write, rename" to update the contents of a file atomically.
 *
 * Returns true if completely successful, or false if some error occurred. */
static _Bool
nsm_atomic_write(const char *path, const void *buf, const size_t buflen)
{
	_Bool result = false;
	ssize_t len;
	char *temp;
	int fd;

	temp = nsm_make_temp_pathname(path);
	if (temp == NULL) {
		bb_error_msg("Failed to create new path for %s", path);
		goto out;
	}

	fd = bb_xopen(temp, O_CREAT | O_TRUNC | O_SYNC | O_WRONLY | BB_OPEN_WARN, 0644);
	if (fd == -1)
		goto out;

	len = write(fd, buf, buflen);
	if (exact_error_check(len, buflen)) {
		bb_perror_msg("Failed to write %s", temp);
		(void)close(fd);
		(void)unlink(temp);
		goto out;
	}

	if (close(fd) == -1) {
		bb_perror_msg("Failed to close %s", temp);
		(void)unlink(temp);
		goto out;
	}

	if (rename(temp, path) == -1) {
		bb_perror_msg("Failed to rename %s -> %s", temp, path);
		(void)unlink(temp);
		goto out;
	}

	/* Ostensibly, a sync(2) is not needed here because
	 * open(O_CREAT), write(O_SYNC), and rename(2) are
	 * already synchronous with persistent storage, for
	 * any file system we care about. */

	result = true;

out:
	free(temp);
	return result;
}

/* nsm_setup_pathnames - set up pathname
 * @parentdir: C string containing pathname to on-disk state, or NULL
 *
 * This runs before logging is set up, so error messages are directed
 * to stderr. */
static void nsm_setup_pathnames(const char *parentdir)
{
	char buf[PATH_MAX];
	struct stat st;
	char *path;

	/* First: test length of name and whether it exists */
	if (lstat(parentdir, &st) == -1)
		bb_perror_msg_and_die("Failed to stat %s", parentdir);

	/* Ensure we have a clean directory pathname */
	strncpy(buf, parentdir, sizeof(buf));
	path = dirname(buf);
	if (*path == '.')
		bb_perror_msg_and_die("Unusable directory %s", parentdir);

	/* Using parentdir as the state directory */
	strncpy(nsm_base_dirname, parentdir, sizeof(nsm_base_dirname));
}

/* Clear all capabilities but CAP_NET_BIND_SERVICE.  This permits
 * callers to acquire privileged source ports, but all other root
 * capabilities are disallowed.
 *
 * Returns true if successful, or false if some error occurred. */
static _Bool nsm_clear_capabilities(void)
{
#ifdef HAVE_SYS_CAPABILITY_H
	cap_t caps;

	caps = cap_from_text("cap_net_bind_service=ep");
	if (caps == NULL) {
		bb_perror_msg("Failed to allocate capability");
		return false;
	}

	if (cap_set_proc(caps) == -1) {
		bb_perror_msg("Failed to set capability flags");
		(void)cap_free(caps);
		return false;
	}

	(void)cap_free(caps);
#endif
	return true;
}

/* nsm_drop_privileges - drop root privileges
 * @pid_fd: file descriptor of a pid file
 *
 * Returns true if successful, or false if some error occurred.
 *
 * Set our effective UID and GID to that of our on-disk database. */
static _Bool nsm_drop_privileges(const int pid_fd)
{
	struct stat st;

	(void)umask(S_IRWXO);

	/* XXX: If we can't stat dirname, or if dirname is owned by
	 *      root, we should use "statduser" instead, which is set up
	 *      by configure.ac.  Nothing in nfs-utils seems to use
	 *      "statduser," though. */
	if (lstat(nsm_base_dirname, &st) == -1) {
		bb_perror_msg("Failed to stat %s", nsm_base_dirname);
		return false;
	}

	if (st.st_uid == 0) {
		xlog_warn("Running as root.  "
			"chown %s to choose different user", nsm_base_dirname);
		return true;
	}

	xchdir(nsm_base_dirname);

	/* If the pidfile happens to reside on NFS, dropping privileges
	 * will probably cause us to lose access, even though we are
	 * holding it open.  Chown it to prevent this. */
	if (pid_fd >= 0)
		if (fchown(pid_fd, st.st_uid, st.st_gid) == -1)
			xlog_warn("Failed to change owner of pidfile: %m");

	/* Don't clear capabilities when dropping root. */
	if (prctl(PR_SET_KEEPCAPS, 1, 0, 0, 0) == -1) {
		bb_perror_msg("prctl(PR_SET_KEEPCAPS) failed");
		return 0;
	}

	if (setgroups(0, NULL) == -1) {
		bb_perror_msg("Failed to drop supplementary groups");
		return false;
	}

	/* ORDER
	 *
	 * setgid(2) first, as setuid(2) may remove privileges needed
	 * to set the group id. */
	if (setgid(st.st_gid) == -1 || setuid(st.st_uid) == -1) {
		bb_perror_msg("Failed to drop privileges");
		return false;
	}

	/* Effective UID, GID: st.st_uid, st.st_gid */
	return nsm_clear_capabilities();
}

/* nsm_get_state - retrieve on-disk NSM state number
 *
 * Returns an odd NSM state number read from disk, or an initial
 * state number.  Zero is returned if some error occurs. */
static int nsm_get_state(_Bool update)
{
	int fd, state = 0;
	ssize_t result;
	char *path = NULL;

	path = nsm_make_pathname(NSM_STATE_FILE);
	if (path == NULL) {
		bb_error_msg("Failed to allocate path for " NSM_STATE_FILE);
		goto out;
	}

	fd = open(path, O_RDONLY);
	if (fd == -1) {
		if (errno != ENOENT) {
			bb_perror_msg("Failed to open %s", path);
			goto out;
		}

		state = 1;
		update = true;
		goto update;
	}

	result = read(fd, &state, sizeof(state));
	if (exact_error_check(result, sizeof(state))) {
		xlog_warn("Failed to read %s: %m", path);

		state = 1;
		update = true;
		goto update;
	}

	if ((state & 1) == 0)
		state++;

update:
	(void)close(fd);

	if (update) {
		state += 2;
		if (!nsm_atomic_write(path, &state, sizeof(state)))
			state = 0;
	}

out:
	free(path);
	return state;
}

/* nsm_update_kernel_state - attempt to post new NSM state to kernel
 * @state: NSM state number */
static void nsm_update_kernel_state(const int state)
{
	ssize_t result;
	char buf[20];
	int fd, len;

	fd = open(NSM_KERNEL_STATE_FILE, O_WRONLY);
	if (fd == -1) {
		/* Failed to open NSM_KERNEL_STATE_FILE: %m */
		return;
	}

	len = snprintf(buf, sizeof(buf), "%d", state);
	if (error_check(len, sizeof(buf))) {
		xlog_warn("Failed to form NSM state number string");
		return;
	}

	result = write(fd, buf, strlen(buf));
	if (exact_error_check(result, strlen(buf)))
		xlog_warn("Failed to write NSM state number: %m");

	if (close(fd) == -1)
		bb_perror_msg("Failed to close NSM state file "
				NSM_KERNEL_STATE_FILE);
}

/* nsm_priv_to_hex - convert a NSM private cookie to a hex string.
 *
 * @priv: buffer holding the binary NSM private cookie
 * @buf: output buffer for NULL terminated hex string
 * @buflen: size of output buffer
 *
 * Returns the length of the resulting string or 0 on error */
static size_t
nsm_priv_to_hex(const char *priv, char *buf, const size_t buflen)
{
	int i, len;
	size_t remaining = buflen;

	for (i = 0; i < SM_PRIV_SIZE; i++) {
		len = snprintf(buf, remaining, "%02x",
				(unsigned int)(0xff & priv[i]));
		if (error_check(len, remaining))
			return 0;
		buf += len;
		remaining -= (size_t)len;
	}

	return buflen - remaining;
}

/* Returns the length in bytes of the created record. */
static __attribute_noinline__ size_t
nsm_create_monitor_record(char *buf, const size_t buflen,
		const struct sockaddr *sap, const struct mon *m)
{
	const struct sockaddr_in *sin = (const struct sockaddr_in *)sap;
	size_t hexlen, remaining = buflen;
	int len;

	len = snprintf(buf, remaining, "%08x %08x %08x %08x ",
			(unsigned int)sin->sin_addr.s_addr,
			(unsigned int)m->mon_id.my_id.my_prog,
			(unsigned int)m->mon_id.my_id.my_vers,
			(unsigned int)m->mon_id.my_id.my_proc);
	if (error_check(len, remaining))
		return 0;
	buf += len;
	remaining -= (size_t)len;

	hexlen = nsm_priv_to_hex(m->priv, buf, remaining);
	if (hexlen == 0)
		return 0;
	buf += hexlen;
	remaining -= hexlen;

	len = snprintf(buf, remaining, " %s %s\n",
			m->mon_id.mon_name, m->mon_id.my_id.my_name);
	if (error_check(len, remaining))
		return 0;
	remaining -= (size_t)len;

	return buflen - remaining;
}

static _Bool nsm_append_monitored_host(const char *path, const char *line)
{
	_Bool result = false;
	char *buf = NULL;
	struct stat stb;
	size_t buflen;
	ssize_t len;
	int fd;

	if (stat(path, &stb) == -1) {
		bb_perror_msg("Failed to insert: "
			"could not stat original file %s", path);
		goto out;
	}
	buflen = (size_t)stb.st_size + strlen(line);

	buf = malloc(buflen + 1);
	if (buf == NULL) {
		bb_error_msg("Failed to insert: no memory");
		goto out;
	}
	memset(buf, 0, buflen + 1);

	fd = open(path, O_RDONLY);
	if (fd == -1) {
		bb_perror_msg("Failed to insert: "
			"could not open original file %s", path);
		goto out;
	}

	len = read(fd, buf, (size_t)stb.st_size);
	if (exact_error_check(len, (size_t)stb.st_size)) {
		bb_perror_msg("Failed to insert: "
			"could not read original file %s", path);
		(void)close(fd);
		goto out;
	}
	(void)close(fd);

	strcat(buf, line);

	if (nsm_atomic_write(path, buf, buflen))
		result = true;

out:
	free(buf);
	return result;
}

/* nsm_insert_monitored_host - write callback data for one host to disk
 * @hostname: C string containing a hostname
 * @sap: sockaddr containing NLM callback address
 * @mon: SM_MON arguments to save
 *
 * Returns true if successful, otherwise false if some error occurs. */
static _Bool
nsm_insert_monitored_host(const char *hostname, const struct sockaddr *sap,
		const struct mon *m)
{
	char buf[LINELEN + 1 + SM_MAXSTRLEN + 2];
	char *path;
	_Bool result = false;
	ssize_t len;
	size_t size;
	int fd;

	path = nsm_make_record_pathname(NSM_MONITOR_DIR, hostname);
	if (path == NULL) {
		bb_error_msg("Failed to insert: bad monitor hostname '%s'",
				hostname);
		return false;
	}

	size = nsm_create_monitor_record(buf, sizeof(buf), sap, m);
	if (size == 0) {
		bb_error_msg("Failed to insert: record too long");
		goto out;
	}

	/*
	 * If exclusive create fails, we're adding a new line to an
	 * existing file.
	 */
	fd = open(path, O_WRONLY | O_CREAT | O_EXCL | O_SYNC, S_IRUSR | S_IWUSR);
	if (fd == -1) {
		if (errno != EEXIST) {
			bb_perror_msg("Failed to insert: creating %s", path);
			goto out;
		}

		result = nsm_append_monitored_host(path, buf);
		goto out;
	}
	result = true;

	len = write(fd, buf, size);
	if (exact_error_check(len, size)) {
		xlog_warn("Failed to insert: writing %s: %m", path);
		(void)unlink(path);
		result = false;
	}

	if (close(fd) == -1) {
		bb_perror_msg("Failed to insert: closing %s", path);
		(void)unlink(path);
		result = false;
	}

out:
	free(path);
	return result;
}

static __attribute_noinline__ _Bool
nsm_parse_line(char *line, struct sockaddr_in *sin, struct mon *m)
{
	unsigned int i, tmp;
	int count;
	char *c;

	strchr_set0(line, '\n');

	count = sscanf(line, "%8x %8x %8x %8x ",
			(unsigned int *)&sin->sin_addr.s_addr,
			(unsigned int *)&m->mon_id.my_id.my_prog,
			(unsigned int *)&m->mon_id.my_id.my_vers,
			(unsigned int *)&m->mon_id.my_id.my_proc);
	if (count != 4)
		return false;

	c = line + RPCARGSLEN;
	for (i = 0; i < SM_PRIV_SIZE; i++) {
		if (sscanf(c, "%2x", &tmp) != 1)
			return false;
		m->priv[i] = (char)tmp;
		c += 2;
	}

	c++;
	m->mon_id.mon_name = c;
	while (*c != '\0' && *c != ' ')
		c++;
	if (*c != '\0')
		*c++ = '\0';
	while (*c == ' ')
		c++;
	m->mon_id.my_id.my_name = c;

	return true;
}

/* Stuff a 'struct mon' with callback data, and call @func.
 *
 * Returns the count of in-core records created. */
static unsigned int
nsm_read_line(const char *hostname, const time_t timestamp, char *line,
		nsm_populate_t func)
{
	struct sockaddr_in sin = {
		.sin_family             = AF_INET,
	};
	struct mon m;

	if (!nsm_parse_line(line, &sin, &m))
		return 0;

	return func(hostname, (struct sockaddr *)(char *)&sin, &m, timestamp);
}

/* Given a filename, reads data from a file under NSM_MONITOR_DIR
 * and invokes @func so caller can populate their in-core
 * database with this data. */
static unsigned int
nsm_load_host(const char *directory, const char *filename, nsm_populate_t func)
{
	char buf[LINELEN + 1 + SM_MAXSTRLEN + 2];
	unsigned int result = 0;
	struct stat stb;
	char *path;
	FILE *f;

	path = nsm_make_record_pathname(directory, filename);
	if (path == NULL)
		goto out_err;

	if (stat(path, &stb) == -1) {
		bb_perror_msg("Failed to stat %s", path);
		goto out_freepath;
	}

	f = bb_xfopen(path, "Wr");
	if (f == NULL)
		goto out_freepath;

	while (fgets(buf, (int)sizeof(buf), f) != NULL) {
		buf[sizeof(buf) - 1] = '\0';
		result += nsm_read_line(filename, stb.st_mtime, buf, func);
	}
	if (result == 0)
		bb_error_msg("Failed to read monitor data from %s", path);

	(void)fclose(f);

out_freepath:
	free(path);
out_err:
	return result;
}

static unsigned int
nsm_load_dir(const char *directory, nsm_populate_t func)
{
	unsigned int count = 0;
	struct dirent *de;
	char *path;
	DIR *dir;

	path = nsm_make_pathname(directory);
	if (path == NULL) {
		bb_error_msg("Failed to allocate path for directory %s",
				directory);
		return 0;
	}

	dir = opendir(path);
	free(path);
	if (dir == NULL) {
		bb_perror_msg("Failed to open directory %s", directory);
		return 0;
	}

	while ((de = readdir(dir)) != NULL) {
		if (de->d_type != (unsigned char)DT_REG)
			continue;
		if (de->d_name[0] == '.')
			continue;

		count += nsm_load_host(directory, de->d_name, func);
	}

	(void)closedir(dir);
	return count;
}

/* nsm_load_monitor_list - load list of hosts to monitor
 * @func: callback function to create entry for one host
 *
 * Returns the count of hosts that were found in the directory. */
static unsigned int
nsm_load_monitor_list(nsm_populate_t func)
{
	return nsm_load_dir(NSM_MONITOR_DIR, func);
}

static void
nsm_delete_host(const char *directory, const char *hostname,
		const char *mon_name, const char *my_name)
{
	char line[LINELEN + 1 + SM_MAXSTRLEN + 2];
	char *outbuf = NULL;
	struct stat stb;
	char *path, *next;
	size_t remaining;
	FILE *f;

	path = nsm_make_record_pathname(directory, hostname);
	if (path == NULL) {
		bb_error_msg("Bad filename, not deleting");
		return;
	}

	if (stat(path, &stb) == -1) {
		bb_perror_msg("Failed to delete: "
			"could not stat original file %s", path);
		goto out;
	}
	remaining = (size_t)stb.st_size + 1;

	outbuf = malloc(remaining);
	if (outbuf == NULL) {
		bb_error_msg("Failed to delete: no memory");
		goto out;
	}

	f = fopen(path, "r");
	if (f == NULL) {
		bb_perror_msg("Failed to delete: "
			"could not open original file %s", path);
		goto out;
	}

	/*
	 * Walk the records in the file, and copy the non-matching
	 * ones to our output buffer.
	 */
	next = outbuf;
	while (fgets(line, (int)sizeof(line), f) != NULL) {
		struct sockaddr_in sin;
		struct mon m;
		size_t len;

		if (!nsm_parse_line(line, &sin, &m)) {
			bb_error_msg("Failed to delete: "
				"could not parse original file %s", path);
			(void)fclose(f);
			goto out;
		}

		if (strcmp(mon_name, m.mon_id.mon_name) == 0 &&
			 strcmp(my_name, m.mon_id.my_id.my_name) == 0)
			continue;

		/* nsm_parse_line destroys the contents of line[], so
		 * reconstruct the copy in our output buffer. */
		len = nsm_create_monitor_record(next, remaining,
					(struct sockaddr *)(char *)&sin, &m);
		if (len == 0) {
			bb_error_msg("Failed to delete: "
				"could not construct output record");
			(void)fclose(f);
			goto out;
		}
		next += len;
		remaining -= len;
	}

	(void)fclose(f);

	/*
	 * If nothing was copied when we're done, then unlink the file.
	 * Otherwise, atomically update the contents of the file.
	 */
	if (next != outbuf) {
		if (!nsm_atomic_write(path, outbuf, strlen(outbuf)))
			bb_perror_msg("Failed to delete: "
				"could not write new file %s", path);
	} else {
		if (unlink(path) == -1)
			bb_perror_msg("Failed to delete: "
				"could not unlink file %s", path);
	}

out:
	free(outbuf);
	free(path);
}

/* nsm_delete_monitored_host - delete on-disk record for monitored host
 * @hostname: '\0'-terminated C string containing hostname of record to delete
 * @mon_name: '\0'-terminated C string containing monname of record to delete
 * @my_name: '\0'-terminated C string containing myname of record to delete */
static void nsm_delete_monitored_host(const char *hostname,
		const char *mon_name, const char *my_name)
{
	nsm_delete_host(NSM_MONITOR_DIR, hostname, mon_name, my_name);
}

/* Instead of using ONC or TI RPC library calls, statd constructs
 * RPC calls directly in socket buffers.  This allows a single
 * socket to be concurrently shared among several different RPC
 * programs and versions using a simple RPC request dispatcher.
 *
 * This file contains the details of RPC header and call
 * construction and reply parsing, and a method for creating a
 * socket for use with these functions. */

/* Returns a fresh XID appropriate for RPC over UDP -- never zero. */
static uint32_t nsm_next_xid(void)
{
	struct timeval now;

	if (nsm_xid == 0) {
		(void)gettimeofday(&now, NULL);
		nsm_xid = (uint32_t)getpid() ^
				(uint32_t)now.tv_sec ^ (uint32_t)now.tv_usec;
	}

	return nsm_xid++;
}

/* Select a fresh XID and construct an RPC header in @mesg.
 * Always use AUTH_NULL credentials and verifiers.
 *
 * Returns the new XID. */
static uint32_t
nsm_init_rpc_header(const rpcprog_t program, const rpcvers_t version,
			const rpcproc_t procedure, struct rpc_msg *mesg)
{
	struct call_body *cb = &mesg->rm_call;
	uint32_t xid = nsm_next_xid();

	memset(mesg, 0, sizeof(*mesg));

	mesg->rm_xid = (unsigned long)xid;
	mesg->rm_direction = CALL;

	cb->cb_rpcvers = RPC_MSG_VERSION;
	cb->cb_prog = program;
	cb->cb_vers = version;
	cb->cb_proc = procedure;

	cb->cb_cred.oa_flavor = AUTH_NULL;
	cb->cb_cred.oa_base = (caddr_t) NULL;
	cb->cb_cred.oa_length = 0;
	cb->cb_verf.oa_flavor = AUTH_NULL;
	cb->cb_verf.oa_base = (caddr_t) NULL;
	cb->cb_verf.oa_length = 0;

	return xid;
}

/* Initialize the network send buffer and XDR memory for encoding. */
static void nsm_init_xdrmem(char *msgbuf, const unsigned int msgbuflen,
		XDR *xdrp)
{
	memset(msgbuf, 0, (size_t)msgbuflen);
	memset(xdrp, 0, sizeof(*xdrp));
	xdrmem_create(xdrp, msgbuf, msgbuflen, XDR_ENCODE);
}

/* Send a completed RPC call on a socket.
 *
 * Returns true if all the bytes were sent successfully; otherwise
 * false if any error occurred. */
static _Bool
nsm_rpc_sendto(const int sock, const struct sockaddr *sap,
			const socklen_t salen, XDR *xdrs, void *buf)
{
	const size_t buflen = (size_t)xdr_getpos(xdrs);
	ssize_t err;

	err = sendto(sock, buf, buflen, 0, sap, salen);
	if ((err < 0) || ((size_t)err != buflen)) {
		bb_perror_msg("%s: sendto failed", __func__);
		return false;
	}
	return true;
}

/* nsm_xmit_getport - post a PMAP_GETPORT call on a socket descriptor
 * @sock: datagram socket descriptor
 * @sin: pointer to AF_INET socket address of server
 * @program: RPC program number to query
 * @version: RPC version number to query
 *
 * Send a PMAP_GETPORT call to the portmap daemon at @sin using
 * socket descriptor @sock.  This request queries the RPC program
 * [program, version, IPPROTO_UDP].
 *
 * NB: PMAP_GETPORT works only for IPv4 hosts.  This implementation
 *     works only over UDP, and queries only UDP registrations.
 *
 * Returns the XID of the call, or zero if an error occurred. */
static uint32_t
nsm_xmit_getport(const int sock, const struct sockaddr_in *sin,
			const unsigned long program,
			const unsigned long version)
{
	char msgbuf[NSM_MAXMSGSIZE];
	struct sockaddr_in addr;
	struct rpc_msg mesg;
	_Bool sent = false;
	struct pmap parms = {
		.pm_prog        = program,
		.pm_vers        = version,
		.pm_prot        = (unsigned long)IPPROTO_UDP,
	};
	uint32_t xid;
	XDR xdr;

	/* Sending PMAP_GETPORT for program, version, udp */

	nsm_init_xdrmem(msgbuf, NSM_MAXMSGSIZE, &xdr);
	xid = nsm_init_rpc_header(PMAPPROG, PMAPVERS,
					(rpcproc_t)PMAPPROC_GETPORT, &mesg);

	addr = *sin;
	addr.sin_port = htons(PMAPPORT);

	if (xdr_callmsg(&xdr, &mesg) == TRUE &&
	    xdr_pmap(&xdr, &parms) == TRUE)
		sent = nsm_rpc_sendto(sock, (struct sockaddr *)(char *)&addr,
					(socklen_t)sizeof(addr), &xdr, msgbuf);
	else
		bb_error_msg("%s: can't encode PMAP_GETPORT call", __func__);

	xdr_destroy(&xdr);

	if (sent == false)
		return 0;
	return xid;
}

static bool_t xdr_stat_chge (XDR *xdrs, stat_chge *objp)
{
	if (!xdr_string (xdrs, &objp->mon_name, SM_MAXSTRLEN))
		 return FALSE;
	if (!xdr_int (xdrs, &objp->state))
		 return FALSE;
	return TRUE;
}

static bool_t xdr_status (XDR *xdrs, status *objp)
{
	if (!xdr_string (xdrs, &objp->mon_name, SM_MAXSTRLEN))
		 return FALSE;
	if (!xdr_int (xdrs, &objp->state))
		 return FALSE;
	if (!xdr_opaque (xdrs, objp->priv, SM_PRIV_SIZE))
		 return FALSE;
	return TRUE;
}

/* nsm_xmit_nlmcall - post an unnamed call to local NLM on a socket descriptor
 * @sock: datagram socket descriptor
 * @sap: address/port of NLM service to contact
 * @salen: size of @sap
 * @m: callback data defining RPC call to make
 * @state: state of rebooting host
 *
 * Send an unnamed call (previously requested via NSMPROC_MON) to the
 * specified local UDP-based RPC service using socket descriptor @sock.
 *
 * NB: This implementation works only over UDP, but supports both AF_INET
 *     and AF_INET6.
 *
 * Returns the XID of the call, or zero if an error occurred. */
static uint32_t
nsm_xmit_nlmcall(const int sock, const struct sockaddr *sap,
			const socklen_t salen, const struct mon *m,
			const int state)
{
	const struct my_id *id = &m->mon_id.my_id;
	char msgbuf[NSM_MAXMSGSIZE];
	struct status new_status;
	struct rpc_msg mesg;
	_Bool sent = false;
	uint32_t xid;
	XDR xdr;

	/* Sending NLM downcall for m->mon_id.mon_name */
	nsm_init_xdrmem(msgbuf, NSM_MAXMSGSIZE, &xdr);
	xid = nsm_init_rpc_header((rpcprog_t)id->my_prog,
					(rpcvers_t)id->my_vers,
					(rpcproc_t)id->my_proc, &mesg);

	new_status.mon_name = m->mon_id.mon_name;
	new_status.state = state;
	memcpy(&new_status.priv, &m->priv, sizeof(new_status.priv));

	if (xdr_callmsg(&xdr, &mesg) == TRUE && xdr_status(&xdr, &new_status) == TRUE)
		sent = nsm_rpc_sendto(sock, sap, salen, &xdr, msgbuf);
	else
		bb_error_msg("%s: can't encode NLM downcall", __func__);

	xdr_destroy(&xdr);

	if (sent == false)
		return 0;
	return xid;
}

static bool_t xdr_sm_name (XDR *xdrs, sm_name *objp)
{
	if (!xdr_string (xdrs, &objp->mon_name, SM_MAXSTRLEN))
		return FALSE;
	return TRUE;
}

static bool_t xdr_my_id (XDR *xdrs, my_id *objp)
{
	if (!xdr_string (xdrs, &objp->my_name, SM_MAXSTRLEN))
		 return FALSE;
	if (!xdr_int (xdrs, &objp->my_prog))
		 return FALSE;
	if (!xdr_int (xdrs, &objp->my_vers))
		 return FALSE;
	if (!xdr_int (xdrs, &objp->my_proc))
		 return FALSE;
	return TRUE;
}

static bool_t xdr_mon_id (XDR *xdrs, mon_id *objp)
{
	if (!xdr_string (xdrs, &objp->mon_name, SM_MAXSTRLEN))
		 return FALSE;
	if (!xdr_my_id (xdrs, &objp->my_id))
		 return FALSE;
	return TRUE;
}

static bool_t xdr_mon (XDR *xdrs, mon *objp)
{
	if (!xdr_mon_id (xdrs, &objp->mon_id))
		 return FALSE;
	if (!xdr_opaque (xdrs, objp->priv, SM_PRIV_SIZE))
		 return FALSE;
	return TRUE;
}

static bool_t xdr_sm_stat (XDR *xdrs, sm_stat *objp)
{
	if (!xdr_int (xdrs, &objp->state))
		 return FALSE;
	return TRUE;
}

static bool_t xdr_res (XDR *xdrs, res *objp)
{
	if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

static bool_t xdr_sm_stat_res (XDR *xdrs, sm_stat_res *objp)
{
	if (!xdr_res (xdrs, &objp->res_stat))
		 return FALSE;
	if (!xdr_int (xdrs, &objp->state))
		 return FALSE;
	return TRUE;
}

static void statd_unregister(void) {
	nfs_svc_unregister(SM_PROG, SM_VERS);
}

/* Signal handler. */
static void killer (int sig)
{
	statd_unregister ();
	bb_error_msg_and_die("Caught signal %d, un-registering and exiting", sig);
}


/* Services SM_MON requests. */
static struct sm_stat_res *sm_mon_1_svc(struct mon *argp, struct svc_req *rqstp)
{
	char            *mon_name = argp->mon_id.mon_name,
			*my_name  = argp->mon_id.my_id.my_name;
	struct my_id    *id = &argp->mon_id.my_id;
	char            *cp;
	notify_list     *clnt;
	struct sockaddr_in my_addr = {
		.sin_family             = AF_INET,
		.sin_addr.s_addr        = htonl(INADDR_LOOPBACK),
	};
	char *dnsname = NULL;

	/* Received SM_MON for mon_name from my_name */
	/* Assume that we'll fail. */
	result4.res_stat = STAT_FAIL;
	result4.state = -1;      /* State is undefined for STAT_FAIL. */

	/* 1.   Reject any remote callers.
	 *      Ignore the my_name specified by the caller, and
	 *      use "127.0.0.1" instead.
	 */
	if (!caller_is_localhost(rqstp))
		goto failure;

	/* 2.   Reject any registrations for non-lockd services.
	 *
	 *      This is specific to the linux kernel lockd, which
	 *      makes the callback procedure part of the lockd interface.
	 *      It is also prone to break when lockd changes its callback
	 *      procedure number -- which, in fact, has now happened once.
	 *      There must be a better way....   XXX FIXME
	 */
	if (id->my_prog != 100021 || (id->my_proc != 16 && id->my_proc != 24)) {
		xlog_warn("Attempt to register callback to %d/%d", id->my_prog, id->my_proc);
		goto failure;
	}

	/*
	 * Check hostnames.  If I can't look them up, I won't monitor.  This
	 * might not be legal, but it adds a little bit of safety and sanity.
	 */

	/* must check for /'s in hostname!  See CERT's CA-96.09 for details. */
	if (strchr(mon_name, '/') || mon_name[0] == '.') {
		bb_error_msg("SM_MON request for hostname containing '/' "
		     "or starting '.': %s", mon_name);
		bb_error_msg("POSSIBLE SPOOF/ATTACK ATTEMPT!");
		goto failure;
	}

	/* my_name must not have white space */
	for (cp=my_name ; *cp ; cp++)
		if (*cp == ' ' || *cp == '\t' || *cp == '\r' || *cp == '\n')
			*cp = '_';

	/*
	 * Hostnames checked OK.
	 * Now choose a hostname to use for matching.  We cannot
	 * really trust much in the incoming NOTIFY, so to make
	 * sure that multi-homed hosts work nicely, we get an
	 * FQDN now, and use that for matching.
	 */
	dnsname = statd_canonical_name(mon_name);
	if (dnsname == NULL) {
		xlog_warn("No canonical hostname found for %s", mon_name);
		goto failure;
	}

	/* Now check to see if this is a duplicate, and warn if so.
	 * I will also return STAT_FAIL. (I *think* this is how I should
	 * handle it.)
	 *
	 * Olaf requests that I allow duplicate SM_MON requests for
	 * hosts due to the way he is coding lockd. No problem,
	 * I'll just do a quickie success return and things should
	 * be happy.
	 */
	clnt = rtnl;

	while ((clnt = nlist_gethost(clnt, mon_name, 0))) {
		if (statd_matchhostname(NL_MY_NAME(clnt), my_name) &&
		    NL_MY_PROC(clnt) == id->my_proc &&
		    NL_MY_PROG(clnt) == id->my_prog &&
		    NL_MY_VERS(clnt) == id->my_vers &&
		    memcmp(NL_PRIV(clnt), argp->priv, SM_PRIV_SIZE) == 0) {
			/* Hey!  We already know you guys! */
			/* Duplicate SM_MON request for mon_name
			   from procedure on my_name */
			/* But we'll let you pass anyway. */
			free(dnsname);
			goto success;
		}
		clnt = NL_NEXT(clnt);
	}

	/*
	 * We're committed...ignoring errors.  Let's hope that a malloc()
	 * doesn't fail.  (I should probably fix this assumption.)
	 */
	if (!(clnt = nlist_new(my_name, mon_name, 0))) {
		free(dnsname);
		xlog_warn("out of memory");
		goto failure;
	}

	NL_MY_PROG(clnt) = id->my_prog;
	NL_MY_VERS(clnt) = id->my_vers;
	NL_MY_PROC(clnt) = id->my_proc;
	memcpy(NL_PRIV(clnt), argp->priv, SM_PRIV_SIZE);
	clnt->dns_name = dnsname;

	/*
	 * Now, Create file on stable storage for host.
	 */
	if (!nsm_insert_monitored_host(dnsname,
				(struct sockaddr *)(char *)&my_addr, argp)) {
		nlist_free(NULL, clnt);
		goto failure;
	}

	/* PRC: do the HA callout: */
	ha_callout("add-client", mon_name, my_name, -1);
	nlist_insert(&rtnl, clnt);
	/* MONITORING mon_name for my_name */
 success:
	result4.res_stat = STAT_SUCC;
	/* SUN's sm_inter.x says this should be "state number of local site".
	 * X/Open says '"state" will be contain the state of the remote NSM.'
	 * href=http://www.opengroup.org/onlinepubs/9629799/SM_MON.htm
	 * Linux lockd currently (2.6.21 and prior) ignores whatever is
	 * returned, and given the above contraction, it probably always will..
	 * So we just return what we always returned.  If possible, we
	 * have already told lockd about our state number via a sysctl.
	 * If lockd wants the remote state, it will need to
	 * use SM_STAT (and prayer).
	 */
	result4.state = MY_STATE;
	return (&result4);

failure:
	xlog_warn("STAT_FAIL to %s for SM_MON of %s", my_name, mon_name);
	return (&result4);
}

/* Services SM_UNMON requests.
 *
 * There is no statement in the X/Open spec's about returning an error
 * for requests to unmonitor a host that we're *not* monitoring.  I just
 * return the state of the NSM when I get such foolish requests for lack
 * of any better ideas.  (I also log the "offense.") */
static struct sm_stat *sm_unmon_1_svc(struct mon_id *argp, struct svc_req *rqstp)
{
	notify_list     *clnt;
	char            *mon_name = argp->mon_name,
			*my_name  = argp->my_id.my_name;
	struct my_id    *id = &argp->my_id;
	char            *cp;

	/* Received SM_UNMON for mon_name from my_name */
	result5.state = MY_STATE;

	if (!caller_is_localhost(rqstp))
		goto failure;

	/* my_name must not have white space */
	for (cp=my_name ; *cp ; cp++)
		if (*cp == ' ' || *cp == '\t' || *cp == '\r' || *cp == '\n')
			*cp = '_';


	/* Check if we're monitoring anyone. */
	if (rtnl == NULL) {
		xlog_warn("Received SM_UNMON request from %s for %s while not "
			"monitoring any hosts", my_name, argp->mon_name);
		return (&result5);
	}
	clnt = rtnl;

	/*
	 * OK, we are.  Now look for appropriate entry in run-time list.
	 * There should only be *one* match on this, since I block "duplicate"
	 * SM_MON calls.  (Actually, duplicate calls are allowed, but only one
	 * entry winds up in the list the way I'm currently handling them.)
	 */
	while ((clnt = nlist_gethost(clnt, mon_name, 0))) {
		if (statd_matchhostname(NL_MY_NAME(clnt), my_name) &&
			NL_MY_PROC(clnt) == id->my_proc &&
			NL_MY_PROG(clnt) == id->my_prog &&
			NL_MY_VERS(clnt) == id->my_vers) {
			/* Match! */
			/* UNMONITORING mon_name for my_name */
			/* PRC: do the HA callout: */
			ha_callout("del-client", mon_name, my_name, -1);

			nsm_delete_monitored_host(clnt->dns_name,
							mon_name, my_name);
			nlist_free(&rtnl, clnt);

			return (&result5);
		} else
			clnt = NL_NEXT(clnt);
	}

 failure:
	xlog_warn("Received erroneous SM_UNMON request from %s for %s",
		my_name, mon_name);
	return (&result5);
}

static struct sm_stat *sm_unmon_all_1_svc(struct my_id *argp, struct svc_req *rqstp)
{
	short int       count = 0;
	notify_list     *clnt;
	char            *my_name = argp->my_name;

	/* Received SM_UNMON_ALL for my_name */
	if (!caller_is_localhost(rqstp))
		goto failure;

	result6.state = MY_STATE;

	if (rtnl == NULL) {
		xlog_warn("Received SM_UNMON_ALL request from %s "
			"while not monitoring any hosts", my_name);
		return (&result6);
	}
	clnt = rtnl;

	while ((clnt = nlist_gethost(clnt, my_name, 1))) {
		if (NL_MY_PROC(clnt) == argp->my_proc &&
			NL_MY_PROG(clnt) == argp->my_prog &&
			NL_MY_VERS(clnt) == argp->my_vers) {
			/* Watch stack! */
			char            mon_name[SM_MAXSTRLEN + 1];
			notify_list     *temp;

			/* UNMONITORING (SM_UNMON_ALL) NL_MON_NAME(clnt)
			   for NL_MY_NAME(clnt) */
			strncpy(mon_name, NL_MON_NAME(clnt),
				sizeof (mon_name) - 1);
			mon_name[sizeof (mon_name) - 1] = '\0';
			temp = NL_NEXT(clnt);
			/* PRC: do the HA callout: */
			ha_callout("del-client", mon_name, my_name, -1);
			nsm_delete_monitored_host(clnt->dns_name,
							mon_name, my_name);
			nlist_free(&rtnl, clnt);
			++count;
			clnt = temp;
		} else
			clnt = NL_NEXT(clnt);
	}

	if (!count) {
		/* SM_UNMON_ALL request from my_name with no
		   SM_MON requests from it */
	}

 failure:
	return (&result6);
}

static void sm_prog_1(struct svc_req *rqstp, register SVCXPRT *transp)
{
	union {
		struct sm_name sm_stat_1_arg;
		struct mon sm_mon_1_arg;
		struct mon_id sm_unmon_1_arg;
		struct my_id sm_unmon_all_1_arg;
		struct stat_chge sm_notify_1_arg;
	} argument;
	char *result;
	xdrproc_t _xdr_argument, _xdr_result;
	char *(*local)(char *, struct svc_req *);

	switch (rqstp->rq_proc) {
	case NULLPROC:
		(void) svc_sendreply (transp, (xdrproc_t) xdr_void, (char *)NULL);
		return;

	case SM_STAT:
		_xdr_argument = (xdrproc_t) xdr_sm_name;
		_xdr_result = (xdrproc_t) xdr_sm_stat_res;
		local = (char *(*)(char *, struct svc_req *)) sm_stat_1_svc;
		break;

	case SM_MON:
		_xdr_argument = (xdrproc_t) xdr_mon;
		_xdr_result = (xdrproc_t) xdr_sm_stat_res;
		local = (char *(*)(char *, struct svc_req *)) sm_mon_1_svc;
		break;

	case SM_UNMON:
		_xdr_argument = (xdrproc_t) xdr_mon_id;
		_xdr_result = (xdrproc_t) xdr_sm_stat;
		local = (char *(*)(char *, struct svc_req *)) sm_unmon_1_svc;
		break;

	case SM_UNMON_ALL:
		_xdr_argument = (xdrproc_t) xdr_my_id;
		_xdr_result = (xdrproc_t) xdr_sm_stat;
		local = (char *(*)(char *, struct svc_req *)) sm_unmon_all_1_svc;
		break;

	case SM_SIMU_CRASH:
		_xdr_argument = (xdrproc_t) xdr_void;
		_xdr_result = (xdrproc_t) xdr_void;
		local = (char *(*)(char *, struct svc_req *)) sm_simu_crash_1_svc;
		break;

	case SM_NOTIFY:
		_xdr_argument = (xdrproc_t) xdr_stat_chge;
		_xdr_result = (xdrproc_t) xdr_void;
		local = (char *(*)(char *, struct svc_req *)) sm_notify_1_svc;
		break;

	default:
		svcerr_noproc (transp);
		return;
	}
	memset ((char *)&argument, 0, sizeof (argument));
	if (!svc_getargs (transp, (xdrproc_t) _xdr_argument, (caddr_t) &argument)) {
		svcerr_decode (transp);
		return;
	}
	result = (*local)((char *)&argument, rqstp);
	if (result != NULL && !svc_sendreply(transp, (xdrproc_t) _xdr_result, result)) {
		svcerr_systemerr (transp);
	}
	if (!svc_freeargs (transp, (xdrproc_t) _xdr_argument, (caddr_t) &argument)) {
		fprintf (stderr, "%s", "unable to free arguments");
		exit (1);
	}
	return;
}

static void load_state(void)
{
	/* unsigned int count = */
		nsm_load_monitor_list(load_one_host);
	/* if (count) Loaded count previously monitored hosts */
}

#ifdef HAVE_TCP_WRAPPER
#include "tcpwrapper.h"

static void
sm_prog_1_wrapper (struct svc_req *rqstp, register SVCXPRT *transp)
{
	/* remote host authorization check */
	if (!check_default("statd", nfs_getrpccaller(transp), SM_PROG)) {
		svcerr_auth (transp, AUTH_FAILED);
		return;
	}

	sm_prog_1 (rqstp, transp);
}

#define sm_prog_1 sm_prog_1_wrapper
#endif

/* Notify operation for a single list entry */
static int process_entry(notify_list *lp)
{
	struct sockaddr_in      sin;

	if (NL_TIMES(lp) == 0) {
		/* Cannot notify localhost, giving up */
		return 0;
	}

	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_port   = lp->port;
	/* LH - moved address into switch */

	/* __FORCE__ loopback for callbacks to lockd ... */
	/* Just in case we somehow ignored it thus far */
	sin.sin_addr.s_addr = htonl(INADDR_LOOPBACK);

	if (sin.sin_port == 0)
		lp->xid = nsm_xmit_getport(sockfd, &sin,
					(rpcprog_t)NL_MY_PROG(lp),
					(rpcvers_t)NL_MY_VERS(lp));
	else {
		struct mon m;

		memcpy(m.priv, NL_PRIV(lp), SM_PRIV_SIZE);

		m.mon_id.mon_name = NL_MON_NAME(lp);
		m.mon_id.my_id.my_name = NULL;
		m.mon_id.my_id.my_prog = NL_MY_PROG(lp);
		m.mon_id.my_id.my_vers = NL_MY_VERS(lp);
		m.mon_id.my_id.my_proc = NL_MY_PROC(lp);

		lp->xid = nsm_xmit_nlmcall(sockfd,
				(struct sockaddr *)(char *)&sin,
				(socklen_t)sizeof(sin), &m, NL_STATE(lp));
	}
	if (lp->xid == 0)
		xlog_warn("%s: failed to notify port %d", __func__, ntohs(lp->port));
	NL_TIMES(lp) -= 1;

	return 1;
}

/* Process a datagram received on the notify socket */
static int process_reply(FD_SET_TYPE *rfds)
{
	notify_list             *lp;
	u_long                  port;

	if (sockfd == -1 || !FD_ISSET(sockfd, rfds))
		return 0;

	if (!(lp = recv_rply(&port)))
		return 1;

	if (lp->port == 0) {
		if (port != 0) {
			lp->port = htons((unsigned short) port);
			process_entry(lp);
			NL_WHEN(lp) = time(NULL) + NOTIFY_TIMEOUT;
			nlist_remove(&statd_notify, lp);
			nlist_insert_timer(&statd_notify, lp);
			return 1;
		}
		xlog_warn("%s: service %d not registered on localhost",
			__func__, NL_MY_PROG(lp));
	} else {
		/* Callback to NL_MY_NAME(lp) (for NL_MON_NAME(lp)) succeeded */
	}
	nlist_free(&statd_notify, lp);
	return 1;
}

/* Process a notify list, either for notifying remote hosts after reboot
 * or for calling back (local) statd clients when the remote has notified
 * us of a crash. */
static int process_notify_list(void)
{
	notify_list     *entry;
	time_t          now;

	while ((entry = statd_notify) != NULL && NL_WHEN(entry) < time(&now)) {
		if (process_entry(entry)) {
			NL_WHEN(entry) = time(NULL) + NOTIFY_TIMEOUT;
			nlist_remove(&statd_notify, entry);
			nlist_insert_timer(&statd_notify, entry);
		} else {
			bb_error_msg("%s: Can't callback %s (%d,%d), giving up",
					__func__,
					NL_MY_NAME(entry),
					NL_MY_PROG(entry),
					NL_MY_VERS(entry));
			nlist_free(&statd_notify, entry);
		}
	}

	return 1;
}

/* This has been modified for my own evil purposes to prevent deadlocks
 * when two hosts start NSM's simultaneously and try to notify each
 * other (which mainly occurs during testing), or to stop and smell the
 * roses when I have callbacks due.
 * --Jeff Uphoff. */

/* This is the RPC server side idle loop.
 * Wait for input, call server program. */

/* The heart of the server.  A crib from libc for the most part... */
static void my_svc_run(void)
{
	FD_SET_TYPE     readfds;
	int             selret;
	time_t          now;

	svc_stop = 0;

	for (;;) {
		if (svc_stop)
			return;

		/* Ah, there are some notifications to be processed */
		while (statd_notify && NL_WHEN(statd_notify) <= time(&now)) {
			process_notify_list();
		}

		readfds = SVC_FDSET;
		if (statd_notify) {
			struct timeval  tv;

			tv.tv_sec  = NL_WHEN(statd_notify) - now;
			tv.tv_usec = 0;
			/* Waiting for reply... (timeo tv.tv_sec) */
			selret = select(FD_SETSIZE, &readfds,
				(void *) 0, (void *) 0, &tv);
		} else {
			/* Waiting for client connections */
			selret = select(FD_SETSIZE, &readfds,
				(void *) 0, (void *) 0, (struct timeval *) 0);
		}

		switch (selret) {
		case -1:
			if (errno == EINTR || errno == ECONNREFUSED
			 || errno == ENETUNREACH || errno == EHOSTUNREACH)
				continue;
			bb_perror_msg("my_svc_run() - select");
			return;

		case 0:
			/* A notify/callback timed out. */
			continue;

		default:
			selret -= process_reply(&readfds);
			if (selret)
				svc_getreqset(&readfds);
		}
	}
}

/* Entry routine/main loop. */
int statd_main(void)
{
	int pid;
	int port = 0, out_port = 0;
	struct rlimit rlim;
	int pipefds[2] = { -1, -1};
	char c_status;
	char *s_out_port = NULL;
	char *s_port = NULL;

	/* Set hostname */
	pidfd = -1;
	sockfd = -1;

	/* Process command line switches */
#ifdef CONFIG_FEATURE_LONG_OPTIONS
	bb_applet_long_options = statd_longopts;
#endif
	run_mode = bb_xgetopt_ulflags("FdNn:LVvo:p:P:H:h\0h-:=P", &MY_NAME,
					&s_out_port, &s_port,
					nsm_setup_pathnames,
					&ha_callout_prog);
	if((run_mode & (32|64))) {
		bb_default_error_retval = 0;
		bb_error_msg_and_die("version " VERSION);
	}
	run_mode &= MODE_NODAEMON | MODE_LOG_STDERR | MODE_NOTIFY_ONLY |
			STATIC_HOSTNAME | MODE_NO_NOTIFY;
	if(s_out_port != NULL) {
		out_port = atoi(s_out_port);
		if (out_port < 1 || out_port > 65535)
			bb_error_msg_and_die("-o %s: bad port number", s_out_port);
	}
	if(s_port != NULL) {
		port = atoi(s_port);
		if (port < 1 || port > 65535)
			bb_error_msg_and_die("-p %s: bad port number", s_port);
	}

	if (port == out_port && port != 0) {
		bb_default_error_retval = -1;
		bb_error_msg_and_die("Listening and outgoing ports cannot be the same!");
	}

	if (run_mode & MODE_NOTIFY_ONLY) {
		bb_error_msg("-N deprecated, consider using /usr/sbin/sm-notify directly");
		run_sm_notify(out_port);
	}

	if (!(run_mode & MODE_NODAEMON)) {
		/* Never log to console in daemon mode. */
		run_mode &= ~MODE_LOG_STDERR;
	}

	if (getrlimit (RLIMIT_NOFILE, &rlim) != 0)
		bb_perror_msg("getrlimit (RLIMIT_NOFILE) failed");
	else {
		/* glibc sunrpc code dies if getdtablesize > FD_SETSIZE */
		if (rlim.rlim_cur > FD_SETSIZE) {
			rlim.rlim_cur = FD_SETSIZE;
			if (setrlimit (RLIMIT_NOFILE, &rlim) != 0)
				bb_perror_msg("setrlimit (RLIMIT_NOFILE) failed");
		}
	}

	if (!(run_mode & MODE_NODAEMON)) {
		int tempfd;

		if (pipe(pipefds)<0)
			bb_perror_msg_and_die("unable to create pipe");
		if ((pid = fork ()) < 0) {
			bb_perror_msg_and_die("Could not fork");
		} else if (pid != 0) {
			/* Parent. Wait for status from child. */
			close(pipefds[1]);
			if (read(pipefds[0], &c_status, 1) != 1)
				return 1;
			return 0;
		}
		/* Child.       */
		close(pipefds[0]);
		setsid ();

		while (pipefds[1] <= 2) {
			pipefds[1] = dup(pipefds[1]);
			if (pipefds[1]<0)
				bb_perror_msg_and_die("dup");
		}
		tempfd = open(bb_dev_null, O_RDWR);
		dup2(tempfd, 0);
		dup2(tempfd, 1);
		dup2(tempfd, 2);
		dup2(pipefds[1], 3);
		pipefds[1] = 3;
		closeall(4);
	}

	/* Child. */

	if (run_mode & MODE_LOG_STDERR) {
		log_syslog = 0;
		log_stderr = 1;
	} else {
		/* Default: daemon mode */
		log_stderr = 0;
		log_syslog = LOG_ERR;
		xlog_open();
	}

	log_modes();

	bb_signal ((1 << SIGHUP) | (1 << SIGINT) | (1 << SIGTERM), killer);
	/* PRC: trap SIGUSR1 to re-read notify list from disk */
	signal(SIGUSR1, sigusr);
	/* WARNING: the following works on Linux and SysV, but not BSD! */
	/* Ignore SIGPIPE to avoid statd dying when peers close their
	 * TCP connection while we're trying to reply to them. */
	bb_signal((1 << SIGCHLD) | (1 << SIGPIPE), SIG_IGN);

	create_pidfile();
	atexit(truncate_pidfile);

	if (! (run_mode & MODE_NO_NOTIFY))
		switch (pid = fork()) {
		case 0:
			run_sm_notify(out_port);
			break;
		case -1:
			break;
		default:
			waitpid(pid, NULL, 0);
		}

	/* Make sure we have a privilege port for calling into the kernel */
	if (statd_get_socket() < 0)
		return 1;

	/* If sm-notify didn't take all the state files, load
	 * state information into our notify-list so we can
	 * pass on any SM_NOTIFY that arrives */
	load_state();

	MY_STATE = nsm_get_state(0);
	if (MY_STATE == 0)
		return 1;
	/* Local NSM state number: MY_STATE */
	nsm_update_kernel_state(MY_STATE);

	/* ORDER
	 * Clear old listeners while still root, to override any
	 * permission checking done by rpcbind. */
	statd_unregister();

	/* ORDER */
	if (!nsm_drop_privileges(pidfd))
		return 1;

	/* ORDER
	 * Create RPC listeners after dropping privileges.  This permits
	 * statd to unregister its own listeners when it exits. */
	if (nfs_svc_create("statd", SM_PROG, SM_VERS, sm_prog_1, port) == 0) {
		bb_error_msg("failed to create RPC listeners, exiting");
		return 1;
	}
	atexit(statd_unregister);

	/* If we got this far, we have successfully started, so notify parent */
	if (pipefds[1] > 0) {
		c_status = 0;
		if (write(pipefds[1], &c_status, 1) != 1) {
			xlog_warn("writing to parent pipe failed: errno %d (%s)",
				errno, strerror(errno));
		}
		close(pipefds[1]);
		pipefds[1] = -1;
	}

	for (;;) {
		/* Handle incoming requests:  SM_NOTIFY socket requests, as
		 * well as callbacks from lockd. */
		my_svc_run();   /* I rolled my own, Olaf made it better... */

		/* Only get here when simulating a crash so we should probably
		 * start sm-notify running again.  As we have already dropped
		 * privileges, this might not work, but I don't think
		 * responding to SM_SIMU_CRASH is an important use cases to
		 * get perfect. */
		if (! (run_mode & MODE_NO_NOTIFY))
			switch (pid = fork()) {
			case 0:
				run_sm_notify(out_port);
				break;
			case -1:
				break;
			default:
				waitpid(pid, NULL, 0);
			}

	}
	return 0;
}


/* Send NSM notify calls to all hosts listed in /var/lib/sm */


#define NSM_TIMEOUT     2
#define NSM_MAX_TIMEOUT 120     /* don't make this too big */


__attribute_malloc__
static struct addrinfo *
smn_lookup(const char *name)
{
	struct addrinfo *ai = NULL;
	struct addrinfo hint = {
		.ai_flags       = AI_ADDRCONFIG,
		.ai_family      = (nsm_family == AF_INET ? AF_INET: AF_UNSPEC),
		.ai_protocol    = (int)IPPROTO_UDP,
	};
	int error;

	error = getaddrinfo(name, NULL, &hint, &ai);
	if (error != 0) {
		/* getaddrinfo(3): gai_strerror(error) */
		return NULL;
	}

	return ai;
}

__attribute_malloc__
static struct nsm_host *
smn_alloc_host(const char *hostname, const char *mon_name,
		const char *my_name, const time_t timestamp)
{
	struct nsm_host *host;

	host = calloc(1, sizeof(*host));
	if (host == NULL)
		goto out_nomem;

	host->name = strdup(hostname);
	host->mon_name = strdup(mon_name);
	host->my_name = strdup(my_name);
	if (host->name == NULL ||
	    host->mon_name == NULL ||
	    host->my_name == NULL) {
		free(host->my_name);
		free(host->mon_name);
		free(host->name);
		free(host);
		goto out_nomem;
	}

	host->last_used = timestamp;
	host->timeout = NSM_TIMEOUT;
	host->retries = 100;            /* force address retry */

	return host;

out_nomem:
	xlog_warn("Unable to allocate memory");
	return NULL;
}

/* nsm_delete_notified_host - delete on-disk host record after notification
 * @hostname: '\0'-terminated C string containing hostname of record to delete
 * @mon_name: '\0'-terminated C string containing monname of record to delete
 * @my_name: '\0'-terminated C string containing myname of record to delete */
static void nsm_delete_notified_host(const char *hostname, const char *mon_name,
		const char *my_name)
{
	nsm_delete_host(NSM_NOTIFY_DIR, hostname, mon_name, my_name);
}

static void smn_forget_host(struct nsm_host *host)
{
	/* Removing host->name (host->mon_name, host->my_name)
	   from notify list */
	nsm_delete_notified_host(host->name, host->mon_name, host->my_name);

	free(host->my_name);
	free(host->mon_name);
	free(host->name);
	if (host->ai)
		freeaddrinfo(host->ai);

	free(host);
}


#ifdef CONFIG_FEATURE_IPV6
static int smn_socket(void)
{
	int sock;

	/*
	 * Use an AF_INET socket if IPv6 is disabled on the
	 * local system.
	 */
	sock = bb_xsocket(BB_XSOCKET_FLG_AF_INET6 | BB_XSOCKET_FLG_SOCK_DGRAM | BB_XSOCKET_FLG_NO_WARN);
	if (sock == -1) {
		if (errno != EAFNOSUPPORT) {
			bb_perror_msg("Failed to create RPC socket");
			return -1;
		}
		sock = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM);
		if (sock < 0)
			return -1;
	} else
		nsm_family = AF_INET6;

	if (fcntl(sock, F_SETFL, O_NONBLOCK) == -1) {
		bb_perror_msg("fcntl(3) on RPC socket failed");
		goto out_close;
	}

	/*
	 * TI-RPC over IPv6 (udp6/tcp6) does not handle IPv4.  However,
	 * since sm-notify open-codes all of its RPC support, it can
	 * use a single socket and let the local network stack provide
	 * the correct mapping between address families automatically.
	 * This is the same thing that is done in the kernel.
	 */
	if (nsm_family == AF_INET6) {
		if (setsockopt_1i(sock, SOL_IPV6, IPV6_V6ONLY, 0) == -1) {
			bb_perror_msg("setsockopt(3) on RPC socket failed");
			goto out_close;
		}
	}

	return sock;

out_close:
	(void)close(sock);
	return -1;
}
#else   /* !CONFIG_FEATURE_IPV6 */
static int smn_socket(void)
{
	int sock;

	sock = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM);
	if (sock == -1)
		return -1;

	if (fcntl(sock, F_SETFL, O_NONBLOCK) == -1) {
		bb_perror_msg("fcntl(3) on RPC socket failed");
		(void)close(sock);
		return -1;
	}

	return sock;
}
#endif  /* !CONFIG_FEATURE_IPV6 */

/*
 * If admin specified a source address or srcport, then convert those
 * to a sockaddr and return it.   Otherwise, return an ANYADDR address.
 */
__attribute_malloc__
static struct addrinfo *
smn_bind_address(const char *srcaddr, const char *srcport)
{
	struct addrinfo *ai = NULL;
	struct addrinfo hint = {
		.ai_flags       = AI_NUMERICSERV,
		.ai_family      = nsm_family,
		.ai_protocol    = (int)IPPROTO_UDP,
	};
	int error;

	if (srcaddr == NULL)
		hint.ai_flags |= AI_PASSIVE;

	if (srcport == NULL)
		error = getaddrinfo(srcaddr, "", &hint, &ai);
	else
		error = getaddrinfo(srcaddr, srcport, &hint, &ai);
	if (error != 0) {
		bb_error_msg("Invalid bind address or port for RPC socket: %s",
				gai_strerror(error));
		return NULL;
	}

	return ai;
}

static int
smn_bindresvport(int sock, struct sockaddr *sap)
{
	if (sap->sa_family != AF_INET) {
		errno = EAFNOSUPPORT;
		return -1;
	}

	return bindresvport(sock, (struct sockaddr_in *)(char *)sap);
}

/* Prepare a socket for sending RPC requests
 *
 * Returns a bound datagram socket file descriptor, or -1 if
 * an error occurs. */
static int
smn_create_socket(const char *srcport)
{
	int sock, retry_cnt = 0;
	struct addrinfo *ai;

retry:
	sock = smn_socket();
	if (sock == -1)
		return -1;

	ai = smn_bind_address(opt_srcaddr, srcport);
	if (ai == NULL) {
		(void)close(sock);
		return -1;
	}

	/* Use source port if provided on the command line,
	 * otherwise use bindresvport */
	if (srcport) {
		if (bind(sock, ai->ai_addr, ai->ai_addrlen) == -1) {
			bb_perror_msg("Failed to bind RPC socket");
			freeaddrinfo(ai);
			(void)close(sock);
			return -1;
		}
	} else {
		struct servent *se;

		if (smn_bindresvport(sock, ai->ai_addr) == -1) {
			bb_perror_msg("bindresvport on RPC socket failed");
			freeaddrinfo(ai);
			(void)close(sock);
			return -1;
		}

		/* try to avoid known ports */
		se = getservbyport((int)nfs_get_port(ai->ai_addr), "udp");
		if (se != NULL && retry_cnt < 100) {
			retry_cnt++;
			freeaddrinfo(ai);
			(void)close(sock);
			goto retry;
		}
	}

	freeaddrinfo(ai);
	return sock;
}

static uint32_t nsm_xmit_getaddr(const int UNUSED(sock),
			const struct sockaddr_in6 *UNUSED(sin6),
			const rpcprog_t UNUSED(program),
			const rpcvers_t UNUSED(version))
{
	return 0;
}

/* nsm_xmit_rpcbind - post an rpcbind request
 * @sock: datagram socket descriptor
 * @sap: pointer to socket address of server
 * @program: RPC program number to query
 * @version: RPC version number to query
 *
 * Send an rpcbind query to the rpcbind daemon at @sap using
 * socket descriptor @sock.
 *
 * NB: This implementation works only over UDP, but can query IPv4 or IPv6
 *     hosts.  It queries only UDP registrations.
 *
 * Returns the XID of the call, or zero if an error occurred. */
static uint32_t
nsm_xmit_rpcbind(const int sock, const struct sockaddr *sap,
			const rpcprog_t program, const rpcvers_t version)
{
	switch (sap->sa_family) {
	case AF_INET:
		return nsm_xmit_getport(sock, (const struct sockaddr_in *)sap,
						program, version);
	case AF_INET6:
		return nsm_xmit_getaddr(sock, (const struct sockaddr_in6 *)sap,
						program, version);
	}
	return 0;
}

/* nsm_xmit_notify - post an NSMPROC_NOTIFY call on a socket descriptor
 * @sock: datagram socket descriptor
 * @sap: pointer to socket address of peer to notify (port already filled in)
 * @salen: length of socket address
 * @program: RPC program number to use
 * @mon_name: mon_name of local peer (ie the rebooting system)
 * @state: state of local peer
 *
 * Send an NSMPROC_NOTIFY call to the peer at @sap using socket descriptor @sock.
 * This request notifies the peer that we have rebooted.
 *
 * NB: This implementation works only over UDP, but supports both AF_INET
 *     and AF_INET6.
 *
 * Returns the XID of the call, or zero if an error occurred. */
static uint32_t
nsm_xmit_notify(const int sock, const struct sockaddr *sap,
			const socklen_t salen, const rpcprog_t program,
			const char *mon_name, const int state)
{
	char msgbuf[NSM_MAXMSGSIZE];
	struct stat_chge state_change;
	struct rpc_msg mesg;
	_Bool sent = false;
	uint32_t xid;
	XDR xdr;

	state_change.mon_name = strdup(mon_name);
	if (state_change.mon_name == NULL) {
		bb_error_msg("%s: no memory", __func__);
		return 0;
	}
	state_change.state = state;

	/* Sending SM_NOTIFY for mon_name */
	nsm_init_xdrmem(msgbuf, NSM_MAXMSGSIZE, &xdr);
	xid = nsm_init_rpc_header(program, SM_VERS, SM_NOTIFY, &mesg);

	if (xdr_callmsg(&xdr, &mesg) == TRUE &&
	    xdr_stat_chge(&xdr, &state_change) == TRUE)
		sent = nsm_rpc_sendto(sock, sap, salen, &xdr, msgbuf);
	else
		bb_error_msg("%s: can't encode NSMPROC_NOTIFY call",
				__func__);

	xdr_destroy(&xdr);
	free(state_change.mon_name);

	if (sent == false)
		return 0;
	return xid;
}

/* Send notification to a single host */
static int notify_host(int sock, struct nsm_host *host)
{
	const char *my_name = (opt_srcaddr != NULL ?
					nsm_hostname : host->my_name);
	struct sockaddr *sap;
	socklen_t salen;

	if (host->ai == NULL) {
		host->ai = smn_lookup(host->name);
		if (host->ai == NULL) {
			xlog_warn("DNS resolution of %s failed; "
				"retrying later", host->name);
			return 0;
		}
	}

	/* If we retransmitted 4 times, reset the port to force
	 * a new portmap lookup (in case statd was restarted).
	 * We also rotate through multiple IP addresses at this
	 * point.
	 */
	if (host->retries >= 4) {
		/* don't rotate if there is only one addrinfo */
		if (host->ai->ai_next != NULL) {
			struct addrinfo *first = host->ai;
			struct addrinfo **next = &host->ai;

			/* remove the first entry from the list */
			host->ai = first->ai_next;
			first->ai_next = NULL;
			/* find the end of the list */
			next = &first->ai_next;
			while ( *next )
				next = & (*next)->ai_next;
			/* put first entry at end */
			*next = first;
		}

		nfs_set_port(host->ai->ai_addr, 0);
		host->retries = 0;
	}

	sap = host->ai->ai_addr;
	salen = host->ai->ai_addrlen;

	if (nfs_get_port(sap) == 0)
		host->xid = nsm_xmit_rpcbind(sock, sap, SM_PROG, SM_VERS);
	else
		host->xid = nsm_xmit_notify(sock, sap, salen,
					SM_PROG, my_name, nsm_state);

	return 0;
}

/* Insert host into sorted list */
static void insert_host(struct nsm_host *host)
{
	struct nsm_host **where, *p;

	where = &hosts;
	while ((p = *where) != 0) {
		/* Sort in ascending order of timeout */
		if (host->send_next < p->send_next)
			break;
		/* If we have the same timeout, put the
		 * most recently used host first.
		 * This makes sure that "recent" hosts
		 * get notified first.
		 */
		if (host->send_next == p->send_next
		 && host->last_used > p->last_used)
			break;
		where = &p->next;
	}

	host->next = *where;
	*where = host;
}

/* Successful NOTIFY call. Server returns void.
 *
 * Try sending another SM_NOTIFY with an unqualified "my_name"
 * argument.  Reuse the port number.  If "my_name" is already
 * unqualified, we're done. */
static void recv_notify_reply(struct nsm_host *host)
{
	if (strchr_set0(host->my_name, '.') != NULL) {
		host->send_next = time(NULL);
		host->xid = 0;
		if (host->timeout >= NSM_MAX_TIMEOUT / 4)
			host->timeout = NSM_MAX_TIMEOUT / 4;
		insert_host(host);
	} else {
		/* Host host->name notified successfully */
		smn_forget_host(host);
	}
}

/* nsm_recv_getaddr - parse RPCB_GETADDR reply
 * @xdrs: pointer to XDR
 *
 * Returns the port number from the RPC reply, or zero
 * if an error occurred. */
static uint16_t nsm_recv_getaddr(XDR *xdrs)
{
	char *uaddr = NULL;
	int port;

	if (xdr_wrapstring(xdrs, &uaddr) == FALSE)
		bb_error_msg("%s: can't decode rpcb reply", __func__);

	if ((uaddr == NULL) || (uaddr[0] == '\0')) {
		/* Received RPCB_GETADDR result program not registered */
		return 0;
	}

	port = nfs_universal2port(uaddr);

	xdr_free((xdrproc_t)xdr_wrapstring, (char *)&uaddr);

	if (port < 0 || port > UINT16_MAX) {
		bb_error_msg("%s: bad port number", __func__);
		return 0;
	}

	/* Received RPCB_GETADDR result: port */
	return (uint16_t)port;
}

/* nsm_recv_rpcbind - parse rpcbind reply
 * @af: address family of reply
 * @xdrs: pointer to XDR
 *
 * Returns the port number from the RPC reply, or zero
 * if an error occurred. */
static uint16_t nsm_recv_rpcbind(const sa_family_t family, XDR *xdrs)
{
	switch (family) {
	case AF_INET:
		return (uint16_t)nsm_recv_getport(xdrs);
	case AF_INET6:
		return nsm_recv_getaddr(xdrs);
	}
	return 0;
}

/* Extract the returned port number and set up the SM_NOTIFY call. */
static void recv_rpcbind_reply(struct sockaddr *sap, struct nsm_host *host, XDR *xdr)
{
	uint16_t port = nsm_recv_rpcbind(sap->sa_family, xdr);

	host->send_next = time(NULL);
	host->xid = 0;

	if (port == 0) {
		/* No binding for statd... */
		/* No statd on host host->name */
		host->timeout = NSM_MAX_TIMEOUT;
		host->send_next += NSM_MAX_TIMEOUT;
	} else {
		nfs_set_port(sap, port);
		if (host->timeout >= NSM_MAX_TIMEOUT / 4)
			host->timeout = NSM_MAX_TIMEOUT / 4;
	}

	insert_host(host);
}

/* Find host given the XID */
static struct nsm_host *find_host(uint32_t xid)
{
	struct nsm_host **where, *p;

	where = &hosts;
	while ((p = *where) != 0) {
		if (p->xid == xid) {
			*where = p->next;
			return p;
		}
		where = &p->next;
	}
	return NULL;
}

/* Receive reply from remote host */
static void recv_reply(int sock)
{
	struct nsm_host *hp;
	struct sockaddr *sap;
	char msgbuf[NSM_MAXMSGSIZE];
	uint32_t        xid;
	ssize_t         msglen;
	XDR             xdr;

	memset(msgbuf, 0 , sizeof(msgbuf));
	msglen = recv(sock, msgbuf, sizeof(msgbuf), 0);
	if (msglen < 0)
		return;

	/* Received packet... */
	memset(&xdr, 0, sizeof(xdr));
	xdrmem_create(&xdr, msgbuf, (unsigned int)msglen, XDR_DECODE);
	xid = nsm_parse_reply(&xdr);
	if (xid == 0)
		goto out;

	/* Before we look at the data, find the host struct for
	   this reply */
	if ((hp = find_host(xid)) == NULL)
		goto out;

	sap = hp->ai->ai_addr;
	if (nfs_get_port(sap) == 0)
		recv_rpcbind_reply(sap, hp, &xdr);
	else
		recv_notify_reply(hp);

out:
	xdr_destroy(&xdr);
}

/* Record pid in /var/run/sm-notify.pid
 * This file should remain until a reboot, even if the
 * program exits.
 * If file already exists, fail. */
static int record_pid(void)
{
	char pid[20];
	ssize_t len;
	int fd;

	(void)snprintf(pid, sizeof(pid), "%d\n", (int)getpid());
	fd = open("/var/run/sm-notify.pid", O_CREAT|O_EXCL|O_WRONLY, 0600);
	if (fd < 0)
		return 0;

	len = write(fd, pid, strlen(pid));
	if ((len < 0) || ((size_t)len != strlen(pid)))
		xlog_warn("Writing to pid file failed: errno %d (%m)", errno);

	(void)close(fd);
	return 1;
}

/* Notify hosts */
static void notify(const int sock)
{
	time_t  failtime = 0;

	if (opt_max_retry)
		failtime = time(NULL) + opt_max_retry;

	while (hosts) {
		struct pollfd   pfd;
		time_t          now = time(NULL);
		unsigned int    sent = 0;
		struct nsm_host *hp;
		long            wait_sec;

		if (failtime && now >= failtime)
			break;

		while (hosts && ((wait_sec = hosts->send_next - now) <= 0)) {
			/* Never send more than 10 packets at once */
			if (sent++ >= 10)
				break;

			/* Remove queue head */
			hp = hosts;
			hosts = hp->next;

			if (notify_host(sock, hp))
				continue;

			/* Set the timeout for this call, using an
			   exponential timeout strategy */
			wait_sec = hp->timeout;
			if ((hp->timeout <<= 1) > NSM_MAX_TIMEOUT)
				hp->timeout = NSM_MAX_TIMEOUT;
			hp->send_next = now + wait_sec;
			hp->retries++;

			insert_host(hp);
		}
		if (hosts == NULL)
			return;

		/* Host hosts->name due in wait seconds */
		pfd.fd = sock;
		pfd.events = POLLIN;

		wait_sec *= 1000;
		if (wait_sec < 100)
			wait_sec = 100;
		if (poll(&pfd, 1, wait_sec) != 1)
			continue;

		recv_reply(sock);
	}
}

static unsigned int smn_get_host(const char *hostname,
		const struct sockaddr *UNUSED(sap),
		const struct mon *m, const time_t timestamp)
{
	struct nsm_host *host;

	host = smn_alloc_host(hostname,
		m->mon_id.mon_name, m->mon_id.my_id.my_name, timestamp);
	if (host == NULL)
		return 0;

	insert_host(host);
	/* Added host hostname to notify list */
	return 1;
}

/* nsm_retire_monitored_hosts - back up all hosts from "sm/" to "sm.bak/"
 *
 * Returns the count of host records that were moved.
 *
 * Note that if any error occurs during this process, some monitor
 * records may be left in the "sm" directory. */
static unsigned int nsm_retire_monitored_hosts(void)
{
	unsigned int count = 0;
	struct dirent *de;
	char *path;
	DIR *dir;

	path = nsm_make_pathname(NSM_MONITOR_DIR);
	if (path == NULL) {
		bb_error_msg("Failed to allocate path for " NSM_MONITOR_DIR);
		return count;
	}

	dir = opendir(path);
	free(path);
	if (dir == NULL) {
		xlog_warn("Failed to open " NSM_MONITOR_DIR ": %m");
		return count;
	}

	while ((de = readdir(dir)) != NULL) {
		char *src, *dst;

		if (de->d_type != (unsigned char)DT_REG)
			continue;
		if (de->d_name[0] == '.')
			continue;

		src = nsm_make_record_pathname(NSM_MONITOR_DIR, de->d_name);
		if (src == NULL) {
			xlog_warn("Bad monitor file name, skipping");
			continue;
		}

		dst = nsm_make_record_pathname(NSM_NOTIFY_DIR, de->d_name);
		if (dst == NULL) {
			free(src);
			xlog_warn("Bad notify file name, skipping");
			continue;
		}

		if (rename(src, dst) == -1)
			xlog_warn("Failed to rename %s -> %s: %m", src, dst);
		else {
			/* Retired record for mon_name de->d_name */
			count++;
		}

		free(dst);
		free(src);
	}

	(void)closedir(dir);
	return count;
}

/* nsm_load_notify_list - load list of hosts to notify
 * @func: callback function to create entry for one host
 *
 * Returns the count of hosts that were found in the directory. */
static unsigned int nsm_load_notify_list(nsm_populate_t func)
{
	return nsm_load_dir(NSM_NOTIFY_DIR, func);
}

int sm_notify_main(void)
{
	int     c, sock;
	char *s_opt_max_retry = NULL;
	char *opt_srcport = NULL;

	strcpy(nsm_base_dirname, NSM_DEFAULT_STATEDIR);
	nsm_family = AF_INET;
	c = bb_xgetopt_ulflags("fdnm:p:v:P:\0n--:=P:?0", &s_opt_max_retry,
				&opt_srcport, &opt_srcaddr,
				nsm_setup_pathnames);
#define FORCE (c & 1)
#define opt_debug (c & 2)
#define opt_update_state (c & 4)

	if(s_opt_max_retry != NULL)
		opt_max_retry = atoi(optarg) * 60;
	else
		opt_max_retry = 15 * 60;

	log_syslog = LOG_ERR;
	log_stderr = opt_debug;

	xlog_open();

	if (record_pid() == 0 && FORCE == 0 && opt_update_state) {
		/* already run, don't try again */
		return 0;
	}

	if (opt_srcaddr != NULL) {
		struct addrinfo *ai = NULL;
		struct addrinfo hint = {
			.ai_family      = AF_UNSPEC,
			.ai_flags       = AI_NUMERICHOST,
		};

		if (getaddrinfo(opt_srcaddr, NULL, &hint, &ai))
			/* not a presentation address - use it */
			strncpy(nsm_hostname, opt_srcaddr, sizeof(nsm_hostname));
		else {
			/* was a presentation address - look it up in
			 * /etc/hosts, so it can be used for my_name */
			int error;

			freeaddrinfo(ai);
			hint.ai_flags = AI_CANONNAME;
			error = getaddrinfo(opt_srcaddr, NULL, &hint, &ai);
			if (error != 0) {
				bb_error_msg("Bind address %s is unusable: %s",
						opt_srcaddr, gai_strerror(error));
				return 1;
			}
			strncpy(nsm_hostname, ai->ai_canonname,
							sizeof(nsm_hostname));
			freeaddrinfo(ai);
		}
	}

	(void)nsm_retire_monitored_hosts();
	if (nsm_load_notify_list(smn_get_host) == 0) {
		/* No hosts to notify; exiting */
		return 0;
	}

	nsm_state = nsm_get_state(opt_update_state);
	if (nsm_state == 0)
		return 1;
	nsm_update_kernel_state(nsm_state);

	if (!opt_debug) {
		if (daemon(0, 0) < 0) {
			bb_perror_msg("unable to background");
			return 1;
		}

		close(0);
		close(1);
		close(2);
	}

	sock = smn_create_socket(opt_srcport);
	if (sock == -1)
		return 1;

	if (!nsm_drop_privileges(-1))
		return 1;

	notify(sock);

	if (hosts) {
		struct nsm_host *hp;

		while ((hp = hosts) != 0) {
			hosts = hp->next;
		}
		return 1;
	}

	return 0;
}

/* nfsmount.c -- Linux NFS mount */


#define NFS_PORT 2049
#define NFS_PROGRAM ((u_long)100003)

#ifndef NFSPROG
#define NFSPROG         ((rpcprog_t)100003)
#endif
#define MOUNTPROG       ((rpcprog_t)100005)
#ifndef NSMPROG
#define NSMPROG         ((rpcprog_t)100024)
#endif

#ifndef NFS_MAXHOSTNAME
#define NFS_MAXHOSTNAME         (255)
#endif

#ifndef NFS_MAXPATHNAME
#define NFS_MAXPATHNAME         (1024)
#endif

/* Maximum number of security flavors on an export: */
#define SECFLAVOR_COUNT 8

struct flav_info {
	char    *flavour;
	int     fnum;
};

struct mount_option {
	struct mount_option *next, *prev;
	char *keyword;
	char *value;
};

struct mount_options {
	struct mount_option *head, *tail;
	unsigned int count;
};

#define MNT_SENDBUFSIZE (2048U)
#define MNT_RECVBUFSIZE (1024U)

typedef struct {
	char **hostname;
	struct sockaddr_in saddr;
	struct pmap pmap;
} clnt_addr_t;

/* RPC call timeout values */
static const struct timeval TIMEOUT = { 20, 0 };
static const struct timeval RETRY_TIMEOUT = { 3, 0 };

struct tokenizer_state {
	char *pos;
	char delimiter;
	int error;
};


#ifndef _MOUNT_H_RPCGEN
#define _MOUNT_H_RPCGEN

/* from @(#)mount.x     1.3 91/03/11 TIRPC 1.0 */
/*
 * Copyright (c) 1985, 1990 by Sun Microsystems, Inc.
 */
#ifndef _rpcsvc_mount_h
#define _rpcsvc_mount_h
#define MNTPATHLEN 1024
#define MNTNAMLEN 255
#define FHSIZE 32

typedef char fhandle[FHSIZE];

struct fhstatus {
	u_int fhs_status;
	union {
		fhandle fhs_fhandle;
	} fhstatus_u;
};
typedef struct fhstatus fhstatus;

typedef char *dirpath;

typedef struct mountbody *mountlist;

struct mountbody {
	char *ml_hostname;
	dirpath ml_directory;
	mountlist ml_next;
};
typedef struct mountbody mountbody;

typedef struct groupnode *groups;

struct groupnode {
	char *gr_name;
	groups gr_next;
};
typedef struct groupnode groupnode;

typedef struct exportnode *exports;

struct exportnode {
	dirpath ex_dir;
	groups ex_groups;
	exports ex_next;
};
typedef struct exportnode exportnode;

#define FHSIZE3 64

typedef struct {
	u_int fhandle3_len;
	char *fhandle3_val;
} fhandle3;

enum mountstat3 {
	MNT_OK = 0,
	MNT3ERR_PERM = 1,
	MNT3ERR_NOENT = 2,
	MNT3ERR_IO = 5,
	MNT3ERR_ACCES = 13,
	MNT3ERR_NOTDIR = 20,
	MNT3ERR_INVAL = 22,
	MNT3ERR_NAMETOOLONG = 63,
	MNT3ERR_NOTSUPP = 10004,
	MNT3ERR_SERVERFAULT = 10006,
};
typedef enum mountstat3 mountstat3;

struct mountres3_ok {
	fhandle3 fhandle;
	struct {
		u_int auth_flavors_len;
		int *auth_flavors_val;
	} auth_flavors;
};
typedef struct mountres3_ok mountres3_ok;

struct mountres3 {
	mountstat3 fhs_status;
	union {
		mountres3_ok mountinfo;
	} mountres3_u;
};
typedef struct mountres3 mountres3;
#endif /*!_rpcsvc_mount_h*/

#define MOUNTVERS 1

#define MOUNTPROC_NULL 0
#define MOUNTPROC_MNT 1
#define MOUNTPROC_DUMP 2
#define MOUNTPROC_UMNT 3
#define MOUNTPROC_UMNTALL 4
#define MOUNTPROC_EXPORT 5
#define MOUNTPROC_EXPORTALL 6

#define MOUNTVERS_NFSV3 3

#define MOUNTPROC3_NULL 0
#define MOUNTPROC3_MNT 1
#define MOUNTPROC3_DUMP 2
#define MOUNTPROC3_UMNT 3
#define MOUNTPROC3_UMNTALL 4
#define MOUNTPROC3_EXPORT 5

#endif /* !_MOUNT_H_RPCGEN */

#define RPC_AUTH_GSS_KRB5       390003
#define RPC_AUTH_GSS_KRB5I      390004
#define RPC_AUTH_GSS_KRB5P      390005
#define RPC_AUTH_GSS_LKEY       390006
#define RPC_AUTH_GSS_LKEYI      390007
#define RPC_AUTH_GSS_LKEYP      390008
#define RPC_AUTH_GSS_SPKM       390009
#define RPC_AUTH_GSS_SPKMI      390010
#define RPC_AUTH_GSS_SPKMP      390011


static const char * const erreob = &errbuf[BUFSIZ];

/* frees its first arg - typical use: s = xstrconcat3(s,t,u); */
static char *xstrconcat3 (char *s, const char *t, const char *u)
{
     char *ret;

     if (!s) s = "";
     if (!u) u = "";
     ret = xasprintf("%s%s%s", s, t, u);
     free(s);
     return ret;
}

/* Convert RPC errors into strings */
static int rpc_strerror(int spos)
{
	int cf_stat = rpc_createerr.cf_stat;
	int pos = 0, cf_errno = rpc_createerr.cf_error.re_errno;
	char *ptr, *estr = clnt_sperrno(cf_stat);
	char *tmp;

	if (estr) {
		if ((ptr = strchr(estr, ':')))
			estr = ++ptr;

		tmp = &errbuf[spos];
		if (cf_stat == RPC_SYSTEMERROR)
			pos = snprintf(tmp, (erreob - tmp),
					"System Error: %s",
						strerror(cf_errno));
		else {
			if (cf_errno)
				pos = snprintf(tmp, (erreob - tmp),
					"RPC Error:%s; errno = %s",
					estr, strerror(cf_errno));
			else
				pos = snprintf(tmp, (erreob - tmp),
						"RPC Error:%s", estr);
		}
	}
	return pos;
}

/* rpc_mount_errors - log an RPC error that occurred during a user-space mount
 * @server: C string containing name of server we are attempting to mount
 * @will_retry: one indicates mount will retry at some later point
 * @bg: one indicates this is a background mount
 *
 * Extracts the error code from the user-space RPC library, and reports it
 * on stderr (fg mount) or in the system log (bg mount). */
static void rpc_mount_errors(char *server, int will_retry, int bg)
{
	int pos = 0;
	char *tmp;

	tmp = &errbuf[pos];
	pos = snprintf(tmp, (erreob - tmp), "mount to NFS server '%s' failed: ", server);

	tmp = &errbuf[pos];
	if (rpc_createerr.cf_stat == RPC_TIMEDOUT) {
		if (will_retry)
			pos = snprintf(tmp, (erreob - tmp),
					"timed out, retrying");
		else
			pos = snprintf(tmp, (erreob - tmp),
					"timed out, giving up");
	} else {
		pos += rpc_strerror(pos);
		tmp = &errbuf[pos];
		if (bg) {
			if (will_retry)
				pos = snprintf(tmp, (erreob - tmp),
						", retrying");
			else
				pos = snprintf(tmp, (erreob - tmp),
						", giving up");
		}
	}

	if (bg) {
		if (onlyonce++ < 1)
			bb_openlog(LOG_CONS|LOG_PID, LOG_AUTH);
		syslog(LOG_ERR, "%s", errbuf);
	} else
		bb_error_msg("%s", errbuf);
}

/* sys_mount_errors - log an error that occurred during a mount system call
 * @server: C string containing name of server we are attempting to mount
 * @bg: one indicates this is a background mount
 *
 * Passed an errno value generated by a mount system call, and reports it
 * on stderr (fg mount) or in the system log (bg mount). */
static void sys_mount_errors(char *server, int bg)
{
	int pos;
	char *tmp;
	int error = errno;

	tmp = errbuf;
	pos = snprintf(tmp, (erreob - tmp), "mount to NFS server '%s' failed: ", server);
	tmp = &errbuf[pos];
	if (error == ETIMEDOUT) {
		snprintf(tmp, (erreob - tmp), "timed out, retrying");
	} else {
		if (bg)
			snprintf(tmp, (erreob - tmp), "%s, retrying", strerror(error));
	}

	bb_error_msg("%s", errbuf);
}

/* mount_error - report a foreground mount error
 * @spec: C string containing the device name being mounted
 * @mount_point: C string containing the pathname of the local mounted on dir
 * @error: errno value to report */
static void mount_error(const char *spec, const char *mount_point)
{
	switch(errno) {
	case EACCES:
		bb_error_msg("access denied by server while mounting %s", spec);
		break;
	case EINVAL:
		bb_error_msg("an incorrect mount option was specified");
		break;
	case EOPNOTSUPP:
		bb_error_msg("requested NFS version or transport"
				" protocol is not supported");
		break;
	case ENOTDIR:
		bb_error_msg("mount point %s is not a directory", mount_point);
		break;
	case EBUSY:
		bb_error_msg("%s is busy or already mounted", mount_point);
		break;
	case ENOENT:
		if (spec)
			bb_perror_msg("mounting %s failed, "
				"reason given by server", spec);
		else
			bb_error_msg("mount point %s does not exist", mount_point);
		break;
	case ESPIPE:
		rpc_mount_errors((char *)spec, 0, 0);
		break;
	case EIO:
		bb_error_msg("mount system call failed");
		break;
	case EFAULT:
		bb_error_msg("encountered unexpected error condition.");
		bb_error_msg("please report the error to " PACKAGE_BUGREPORT);
		break;
	default:
		bb_perror_nomsg();
	}
}

/* umount_error - report a failed umount request
 * @err: errno value to report
 * @dev: C string containing the pathname of the local mounted on dir */
static void umount_error(const char *dev)
{
	switch (errno) {
	case ENXIO:
		bb_error_msg("%s: invalid block device", dev);
		break;
	case EINVAL:
		bb_error_msg("%s: not mounted", dev);
		break;
	case EIO:
		bb_error_msg("%s: can't write superblock", dev);
		break;
	case EBUSY:
		bb_error_msg("%s: device is busy", dev);
		break;
	case ENOENT:
		bb_error_msg("%s: not found", dev);
		break;
	case EPERM:
		bb_error_msg("%s: must be superuser to umount", dev);
		break;
	case EACCES:
		bb_error_msg("%s: block devices not permitted on fs", dev);
		break;
	default:
		bb_perror_msg("%s", dev);
		break;
	}
}

/* We need to translate between nfs status return values and
 * the local errno values which may not be the same.
*/

#ifndef EDQUOT
#define EDQUOT  ENOSPC
#endif

#define ARRAY_SIZE(x)           (sizeof(x) / sizeof((x)[0]))

static const struct {
	enum nfsstat stat;
	int errnum;
} nfs_errtbl[] = {
	{ NFS_OK,               0               },
	{ NFSERR_PERM,          EPERM           },
	{ NFSERR_NOENT,         ENOENT          },
	{ NFSERR_IO,            EIO             },
	{ NFSERR_NXIO,          ENXIO           },
	{ NFSERR_ACCES,         EACCES          },
	{ NFSERR_EXIST,         EEXIST          },
	{ NFSERR_NODEV,         ENODEV          },
	{ NFSERR_NOTDIR,        ENOTDIR         },
	{ NFSERR_ISDIR,         EISDIR          },
#ifdef NFSERR_INVAL
	{ NFSERR_INVAL,         EINVAL          },      /* that Sun forgot */
#endif
	{ NFSERR_FBIG,          EFBIG           },
	{ NFSERR_NOSPC,         ENOSPC          },
	{ NFSERR_ROFS,          EROFS           },
	{ NFSERR_NAMETOOLONG,   ENAMETOOLONG    },
	{ NFSERR_NOTEMPTY,      ENOTEMPTY       },
	{ NFSERR_DQUOT,         EDQUOT          },
	{ NFSERR_STALE,         ESTALE          },
#ifdef EWFLUSH
	{ NFSERR_WFLUSH,        EWFLUSH         },
#endif
	/* Throw in some NFSv3 values for even more fun (HP returns these) */
	{ 71,                   EREMOTE         },
};

static char *nfs_strerror(unsigned int nfs_stat, char buf[256])
{
	unsigned int i;

	for (i = 0; i < ARRAY_SIZE(nfs_errtbl); i++) {
		if (nfs_errtbl[i].stat == nfs_stat)
			return strerror(nfs_errtbl[i].errnum);
	}
	sprintf(buf, "unknown nfs status return value: %u", nfs_stat);
	return buf;
}

/* Make a canonical pathname from PATH.  Returns a freshly malloced string.
   It is up the *caller* to ensure that the PATH is sensible.  i.e.
   canonicalize ("/dev/fd0/.") returns "/dev/fd0" even though ``/dev/fd0/.''
   is not a legal pathname for ``/dev/fd0''.  Anything we cannot parse
   we return unmodified.   */
static char *canonicalize (const char *path)
{
	char canonical[PATH_MAX+2];

	if (path == NULL)
		return NULL;

#if 1
	if (streq(path, "none") ||
	    streq(path, "proc") ||
	    streq(path, "devpts"))
		return xstrdup(path);
#endif
	if (realpath (path, canonical))
		return xstrdup(canonical);

	return xstrdup(path);
}


static const struct flav_info flav_map[] = {
	{ "krb5",       RPC_AUTH_GSS_KRB5       },
	{ "krb5i",      RPC_AUTH_GSS_KRB5I      },
	{ "krb5p",      RPC_AUTH_GSS_KRB5P      },
	{ "lipkey",     RPC_AUTH_GSS_LKEY       },
	{ "lipkey-i",   RPC_AUTH_GSS_LKEYI      },
	{ "lipkey-p",   RPC_AUTH_GSS_LKEYP      },
	{ "spkm3",      RPC_AUTH_GSS_SPKM       },
	{ "spkm3i",     RPC_AUTH_GSS_SPKMI      },
	{ "spkm3p",     RPC_AUTH_GSS_SPKMP      },
	{ "unix",       AUTH_UNIX               },
	{ "sys",        AUTH_SYS                },
	{ "null",       AUTH_NULL               },
	{ "none",       AUTH_NONE               },
};

static const int flav_map_size = sizeof(flav_map)/sizeof(flav_map[0]);


#ifndef _NFS_UTILS_MOUNT_FSTAB_H
#define _NFS_UTILS_MOUNT_FSTAB_H


#define ERR_MAX 5

typedef struct mntFILEstruct {
	FILE *mntent_fp;
	char *mntent_file;
	int mntent_lineno;
	int mntent_errs;
	int mntent_softerrs;
} mntFILE;

#ifndef _PATH_FSTAB
#define _PATH_FSTAB "/etc/fstab"
#endif


struct mntentchn *getmntoptfile (const char *file);

#endif  /* _NFS_UTILS_MOUNT_FSTAB_H */

#ifndef _PATH_MOUNTED
#define _PATH_MOUNTED "/etc/mtab"
#endif
#define MOUNTED_LOCK    _PATH_MOUNTED "~"
#define MOUNTED_TEMP    _PATH_MOUNTED ".tmp"

#define LOCK_TIMEOUT    10
#define PROC_MOUNTS             "/proc/mounts"


/* Information about mtab. ------------------------------------*/
static void
get_mtab_info(void) {
	struct stat mtab_stat;

	if (!have_mtab_info) {
		if (lstat(MOUNTED, &mtab_stat))
			var_mtab_does_not_exist = 1;
		else if (S_ISLNK(mtab_stat.st_mode))
			var_mtab_is_a_symlink = 1;
		have_mtab_info = 1;
	}
}

static inline void reset_mtab_info(void) { have_mtab_info = 0; }

static int mtab_does_not_exist(void)
{
	get_mtab_info();
	return var_mtab_does_not_exist;
}

static int
mtab_is_a_symlink(void) {
	get_mtab_info();
	return var_mtab_is_a_symlink;
}

static int mtab_is_writable(void)
{
	int fd;

	/* Should we write to /etc/mtab upon an update?
	   Probably not if it is a symlink to /proc/mounts, since that
	   would create a file /proc/mounts in case the proc filesystem
	   is not mounted. */
	if (mtab_is_a_symlink())
		return 0;

	fd = open(MOUNTED, O_RDWR | O_CREAT, 0644);
	if (fd >= 0) {
		close(fd);
		return 1;
	} else
		return 0;
}

/* Contents of mtab and fstab ---------------------------------*/

static void nfs_endmntent (mntFILE *mfp)
{
	if (mfp) {
		if (mfp->mntent_fp)
			fclose(mfp->mntent_fp);
		if (mfp->mntent_file)
			free(mfp->mntent_file);
		free(mfp);
	}
}

static inline int is_space_or_tab (char c) {
	return (c == ' ' || c == '\t');
}

static char * skip_spaces(char *s) {
	while (is_space_or_tab(*s))
		s++;
	return s;
}

static char * skip_nonspaces(char *s)
{
	while (*s && !is_space_or_tab(*s))
		s++;
	return s;
}

/* returns malloced pointer - no more strdup required */
static char *unmangle(char *s)
{
	char *ret, *ss, *sp;

	ss = skip_nonspaces(s);
	ret = sp = xmalloc(ss-s+1);
	while(s != ss) {
		if (*s == '\\' && isoctal(s[1]) && isoctal(s[2]) && isoctal(s[3])) {
			*sp++ = 64*(s[1] & 7) + 8*(s[2] & 7) + (s[3] & 7);
			s += 4;
		} else
			*sp++ = *s++;
	}
	*sp = 0;
	return ret;
}

/* Read the next entry from the file fp. Stop reading at an incorrect entry. */
static struct mntent * nfs_getmntent (mntFILE *mfp, struct mntent *p_me)
{
	char *s;

 again:
	if (mfp->mntent_errs || mfp->mntent_softerrs >= ERR_MAX)
		return NULL;

	/* read the next non-blank non-comment line */
	do {
		if (fgets (bss_common_buf, sizeof(bss_common_buf), mfp->mntent_fp) == NULL)
			return NULL;

		mfp->mntent_lineno++;
		s = strchr_set0 (bss_common_buf, '\n');
		if (s == NULL) {
			/* Missing final newline?  Otherwise extremely */
			/* long line - assume file was corrupted */
			if (feof(mfp->mntent_fp)) {
				fprintf(stderr, "[mntent]: warning: no final "
					"newline at the end of %s\n",
					mfp->mntent_file);
				s = strchr (bss_common_buf, 0);
			} else {
				mfp->mntent_errs = 1;
				goto err;
			}
		}
		if (--s >= bss_common_buf && *s == '\r')
			*s = 0;
		s = skip_spaces(bss_common_buf);
	} while (*s == '\0' || *s == '#');

	p_me->mnt_fsname = unmangle(s);
	s = skip_nonspaces(s);
	s = skip_spaces(s);
	p_me->mnt_dir = unmangle(s);
	s = skip_nonspaces(s);
	s = skip_spaces(s);
	p_me->mnt_type = unmangle(s);
	s = skip_nonspaces(s);
	s = skip_spaces(s);
	p_me->mnt_opts = unmangle(s);
	s = skip_nonspaces(s);
	s = skip_spaces(s);

	if (isdigit(*s)) {
		p_me->mnt_freq = atoi(s);
		while(isdigit(*s)) s++;
	} else
		p_me->mnt_freq = 0;
	if(*s && !is_space_or_tab(*s))
		goto err;

	s = skip_spaces(s);
	if(isdigit(*s)) {
		p_me->mnt_passno = atoi(s);
		while(isdigit(*s)) s++;
	} else
		p_me->mnt_passno = 0;
	if(*s && !is_space_or_tab(*s))
		goto err;

	/* allow more stuff, e.g. comments, on this line */

	return p_me;

 err:
	mfp->mntent_softerrs++;
	fprintf(stderr, "[mntent]: line %d in %s is bad%s\n",
		mfp->mntent_lineno, mfp->mntent_file,
		(mfp->mntent_errs || mfp->mntent_softerrs >= ERR_MAX) ?
		"; rest of file ignored" : "");
	goto again;
}

static void
read_mntentchn(mntFILE *mfp, const char *fnam, struct mntentchn *mc0)
{
	struct mntentchn *mc = mc0;
	struct mntent *mnt;

	while ((mnt = nfs_getmntent(mfp, &me)) != NULL) {
		if (!streq(mnt->mnt_type, MNTTYPE_IGNORE)) {
			mc->nxt = (struct mntentchn *) xmalloc(sizeof(*mc));
			mc->nxt->prev = mc;
			mc = mc->nxt;
			mc->m = *mnt;
			mc->nxt = mc0;
		}
	}
	mc0->prev = mc;
	if (ferror(mfp->mntent_fp)) {
		bb_perror_msg("warning: error reading %s", fnam);
		mc0->nxt = mc0->prev = NULL;
	}
	nfs_endmntent(mfp);
}


/* fstat'ing the file and allocating a buffer holding all of it
 * may be a bad idea: if the file is /proc/mounts, the stat
 * returns 0.
 * (On the other hand, mangling and unmangling is meaningless
 *  for /proc/mounts.) */
static mntFILE * nfs_setmntent (const char *file, char *mode)
{
	mntFILE *mfp = xmalloc(sizeof(*mfp));
	mode_t old_umask = umask(077);

	mfp->mntent_fp = fopen(file, mode);
	umask(old_umask);
	mfp->mntent_file = xstrdup(file);
	mfp->mntent_errs = (mfp->mntent_fp == NULL);
	mfp->mntent_softerrs = 0;
	mfp->mntent_lineno = 0;
	return mfp;
}

/* Read /etc/mtab.  If that fails, try /proc/mounts.
 * This produces a linked list. The list head mounttable is a dummy.
 * Return 0 on success. */
static void read_mounttable(void)
{
	mntFILE *mfp;
	const char *fnam;
	struct mntentchn *mc = &mounttable;

	got_mtab = 1;
	mc->nxt = mc->prev = NULL;

	fnam = MOUNTED;
	mfp = nfs_setmntent (fnam, "r");
	if (mfp == NULL || mfp->mntent_fp == NULL) {
		fnam = PROC_MOUNTS;
		mfp = nfs_setmntent (fnam, "r");
		if (mfp == NULL || mfp->mntent_fp == NULL) {
			bb_perror_msg("warning: can't open %s", MOUNTED);
			return;
		}
		if (verbose)
			printf("%s: could not open %s; using %s instead\n",
				bb_applet_name, MOUNTED, PROC_MOUNTS);
	}
	read_mntentchn(mfp, fnam, mc);
}

static struct mntentchn *
mtab_head(void)
{
	if (!got_mtab)
		read_mounttable();
	return &mounttable;
}

static void read_fstab(void)
{
	mntFILE *mfp = NULL;
	const char *fnam;
	struct mntentchn *mc = &fstab;

	got_fstab = 1;
	mc->nxt = mc->prev = NULL;

	fnam = _PATH_FSTAB;
	mfp = nfs_setmntent (fnam, "r");
	if (mfp == NULL || mfp->mntent_fp == NULL) {
		bb_perror_msg("warning: can't open %s", _PATH_FSTAB);
		return;
	}
	read_mntentchn(mfp, fnam, mc);
}

static struct mntentchn *fstab_head(void)
{
	if (!got_fstab)
		read_fstab();
	return &fstab;
}

/* Given the directory name NAME, and the place MCPREV we found it last time,
 * try to find more occurrences. */
static struct mntentchn *
getmntdirbackward (const char *name, struct mntentchn *mcprev) {
	struct mntentchn *mc, *mc0;

	mc0 = mtab_head();
	if (!mcprev)
		mcprev = mc0;
	for (mc = mcprev->prev; mc && mc != mc0; mc = mc->prev)
		if (streq(mc->m.mnt_dir, name))
			return mc;
	return NULL;
}

/*
 * Given the device name NAME, and the place MCPREV we found it last time,
 * try to find more occurrences.
 */
static struct mntentchn *
getmntdevbackward (const char *name, struct mntentchn *mcprev)
{
	struct mntentchn *mc, *mc0;

	mc0 = mtab_head();
	if (!mcprev)
		mcprev = mc0;
	for (mc = mcprev->prev; mc && mc != mc0; mc = mc->prev)
		if (streq(mc->m.mnt_fsname, name))
			return mc;
	return NULL;
}

/* Find the dir FILE in fstab.  */
static struct mntentchn *
get_fsfile (const char *file)
{
	struct mntentchn *mc, *mc0;

	mc0 = fstab_head();
	for (mc = mc0->nxt; mc && mc != mc0; mc = mc->nxt)
		if (streq(mc->m.mnt_dir, file))
			return mc;
	return NULL;
}

/* Updating mtab ----------------------------------------------*/


/* Ensure that the lock is released if we are interrupted.  */
static void handler (int sig)
{
     char buf[16];
     bb_default_error_retval = EX_USER;
     bb_error_msg_and_die("%s", u_signal_num2name(sig, 1, buf));
}

/* nothing, fcntl will fail anyway */
static void setlkw_timeout (int UNUSED(sig)) { }

/* Remove lock file.  */
static void unlock_mtab (void)
{
	if (we_created_lockfile) {
		close(lockfile_fd);
		unlink (MOUNTED_LOCK);
		we_created_lockfile = 0;
	}
}

/* Create the lock file.
   The lock file will be removed if we catch a signal or when we exit. */
/* The old code here used flock on a lock file /etc/mtab~ and deleted
   this lock file afterwards. However, as rgooch remarks, that has a
   race: a second mount may be waiting on the lock and proceed as
   soon as the lock file is deleted by the first mount, and immediately
   afterwards a third mount comes, creates a new /etc/mtab~, applies
   flock to that, and also proceeds, so that the second and third mount
   now both are scribbling in /etc/mtab.
   The new code uses a link() instead of a creat(), where we proceed
   only if it was us that created the lock, and hence we always have
   to delete the lock afterwards. Now the use of flock() is in principle
   superfluous, but avoids an arbitrary sleep(). */

/* Where does the link point to? Obvious choices are mtab and mtab~~.
   HJLu points out that the latter leads to races. Right now we use
   mtab~.<pid> instead. Use 20 as upper bound for the length of %d. */
#define MOUNTLOCK_LINKTARGET            MOUNTED_LOCK "%d"
#define MOUNTLOCK_LINKTARGET_LTH        (sizeof(MOUNTED_LOCK)+20)

static void lock_mtab (void)
{
	int tries = 100000, i;
	char linktargetfile[MOUNTLOCK_LINKTARGET_LTH];

	atexit(unlock_mtab);

	if (!signals_have_been_setup) {
		int sig = 0;
		struct sigaction sa;

		sa.sa_handler = handler;
		sa.sa_flags = 0;
		sigfillset (&sa.sa_mask);

		while (sigismember (&sa.sa_mask, ++sig) != -1
		       && sig != SIGCHLD) {
			if (sig == SIGALRM)
				sa.sa_handler = setlkw_timeout;
			else
				sa.sa_handler = handler;
			sigaction (sig, &sa, (struct sigaction *) 0);
		}
		signals_have_been_setup = 1;
	}

	sprintf(linktargetfile, MOUNTLOCK_LINKTARGET, getpid ());

	i = open (linktargetfile, O_WRONLY|O_CREAT, 0);
	if (i < 0) {
		/* linktargetfile does not exist (as a file)
		   and we cannot create it. Read-only filesystem?
		   Too many files open in the system?
		   Filesystem full? */
		bb_default_error_retval = EX_FILEIO;
		bb_error_msg_and_die ("can't create lock file %s: %s "
					"(use -n flag to override)",
			 linktargetfile, strerror(errno));
	}
	close(i);

	/* Repeat until it was us who made the link */
	while (!we_created_lockfile) {
		struct flock flock;
		int errsv, j;

		j = link(linktargetfile, MOUNTED_LOCK);
		errsv = errno;

		if (j == 0)
			we_created_lockfile = 1;

		if (j < 0 && errsv != EEXIST) {
			(void) unlink(linktargetfile);
			bb_default_error_retval = EX_FILEIO;
			errno = errsv;
			bb_error_msg_and_die ("can't link lock file %s: %s "
			     "(use -n flag to override)",
			     MOUNTED_LOCK, strerror (errsv));
		}

		lockfile_fd = open (MOUNTED_LOCK, O_WRONLY);

		if (lockfile_fd < 0) {
			errsv = errno;
			/* Strange... Maybe the file was just deleted? */
			if (errsv == ENOENT && tries-- > 0) {
				if (tries % 200 == 0)
					xusleep(30);
				continue;
			}
			(void) unlink(linktargetfile);
			bb_default_error_retval = EX_FILEIO;
			bb_error_msg_and_die ("can't open lock file %s: %s "
			     "(use -n flag to override)",
			     MOUNTED_LOCK, strerror (errsv));
		}

		flock.l_type = F_WRLCK;
		flock.l_whence = SEEK_SET;
		flock.l_start = 0;
		flock.l_len = 0;

		if (j == 0) {
			/* We made the link. Now claim the lock. */
			if (fcntl (lockfile_fd, F_SETLK, &flock) == -1) {
				if (verbose)
				    bb_perror_msg("Can't lock lock file %s", MOUNTED_LOCK);
				/* proceed anyway */
			}
			(void) unlink(linktargetfile);
		} else {
			/* Someone else made the link. Wait. */
			alarm(LOCK_TIMEOUT);
			if (fcntl (lockfile_fd, F_SETLKW, &flock) == -1) {
				errsv = errno;
				(void) unlink(linktargetfile);
				bb_default_error_retval = EX_FILEIO;
				bb_error_msg_and_die ("can't lock lock file %s: %s",
				     MOUNTED_LOCK, (errno == EINTR) ?
				     "timed out" : strerror (errsv));
			}
			alarm(0);
			/* Limit the number of iterations - maybe there
			   still is some old /etc/mtab~ */
			++iter_tries;
			if (iter_tries % 200 == 0)
			   xusleep(30);
			if (iter_tries > 100000) {
				(void) unlink(linktargetfile);
				close(lockfile_fd);
				bb_default_error_retval = EX_FILEIO;
				bb_error_msg_and_die ("Cannot create link %s\n"
						  "Perhaps there is a stale lock file?",
					 MOUNTED_LOCK);
			}
			close(lockfile_fd);
		}
	}
}

/* Unfortunately the classical Unix /etc/mtab and /etc/fstab
   do not handle directory names containing spaces.
   Here we mangle them, replacing a space by \040.
   What do other Unices do? */
static const unsigned char need_escaping[] = { ' ', '\t', '\n', '\\' };

static char *mangle(const char *arg) {
	const unsigned char *s = (const unsigned char *)arg;
	char *ss, *sp;
	unsigned int n;

	n = strlen(arg);
	ss = sp = xmalloc(4*n+1);
	while(1) {
		for (n = 0; n < sizeof(need_escaping); n++) {
			if (*s == need_escaping[n]) {
				*sp++ = '\\';
				*sp++ = '0' + ((*s & 0300) >> 6);
				*sp++ = '0' + ((*s & 070) >> 3);
				*sp++ = '0' + (*s & 07);
				goto next;
			}
		}
		*sp++ = *s;
		if (*s == 0)
			break;
	next:
		s++;
	}
	return ss;
}

static int nfs_addmntent (mntFILE *mfp, struct mntent *mnt) {
	char *m1, *m2, *m3, *m4;
	int ret;

	if (fseek (mfp->mntent_fp, 0, SEEK_END))
		return 1;                       /* failure */

	m1 = mangle(mnt->mnt_fsname);
	m2 = mangle(mnt->mnt_dir);
	m3 = mangle(mnt->mnt_type);
	m4 = mangle(mnt->mnt_opts);

	ret = fprintf (mfp->mntent_fp, "%s %s %s %s %d %d\n",
		       m1, m2, m3, m4, mnt->mnt_freq, mnt->mnt_passno);

	free(m1);
	free(m2);
	free(m3);
	free(m4);
	return (ret < 0) ? 1 : 0;
}

/*
 * Update the mtab.
 *  Used by umount with null INSTEAD: remove the last DIR entry.
 *  Used by mount upon a remount: update option part,
 *   and complain if a wrong device or type was given.
 *   [Note that often a remount will be a rw remount of /
 *    where there was no entry before, and we'll have to believe
 *    the values given in INSTEAD.]
 */

static void update_mtab (const char *dir, struct mntent *instead)
{
	mntFILE *mfp, *mftmp;
	const char *fnam = MOUNTED;
	struct mntentchn mtabhead;      /* dummy */
	struct mntentchn *mc, *mc0, *absent = NULL;

	if (mtab_does_not_exist() || !mtab_is_writable())
		return;

	lock_mtab();

	/* having locked mtab, read it again */
	mc0 = mc = &mtabhead;
	mc->nxt = mc->prev = NULL;

	mfp = nfs_setmntent(fnam, "r");
	if (mfp == NULL || mfp->mntent_fp == NULL) {
		bb_perror_msg ("mtab not updated: cannot open %s", fnam);
		goto leave;
	}

	read_mntentchn(mfp, fnam, mc);

	/* find last occurrence of dir */
	for (mc = mc0->prev; mc && mc != mc0; mc = mc->prev)
		if (streq(mc->m.mnt_dir, dir))
			break;
	if (mc && mc != mc0) {
		if (instead == NULL) {
			/* An umount - remove entry */
			if (mc && mc != mc0) {
				mc->prev->nxt = mc->nxt;
				mc->nxt->prev = mc->prev;
				free(mc);
			}
		} else {
			/* A remount */
			mc->m.mnt_opts = instead->mnt_opts;
		}
	} else if (instead) {
		/* not found, add a new entry */
		absent = xmalloc(sizeof(*absent));
		absent->m = *instead;
		absent->nxt = mc0;
		absent->prev = mc0->prev;
		mc0->prev = absent;
		if (mc0->nxt == NULL)
			mc0->nxt = absent;
	}

	/* write chain to mtemp */
	mftmp = nfs_setmntent (MOUNTED_TEMP, "w");
	if (mftmp == NULL || mftmp->mntent_fp == NULL) {
		bb_perror_msg ("mtab not updated: cannot open %s", MOUNTED_TEMP);
		goto leave;
	}

	for (mc = mc0->nxt; mc && mc != mc0; mc = mc->nxt) {
		if (nfs_addmntent(mftmp, &(mc->m)) == 1) {
			bb_default_error_retval = EX_FILEIO;
			bb_perror_msg_and_die ("error writing %s", MOUNTED_TEMP);
		}
	}

	if (fchmod (fileno (mftmp->mntent_fp), S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH) < 0)
		bb_perror_msg("error changing mode of %s", MOUNTED_TEMP);
	nfs_endmntent (mftmp);

	{ /*
	   * If mount is setuid and some non-root user mounts sth,
	   * then mtab.tmp might get the group of this user. Copy uid/gid
	   * from the present mtab before renaming.
	   */
	    struct stat sbuf;
	    if (stat (MOUNTED, &sbuf) == 0) {
			if (chown (MOUNTED_TEMP, sbuf.st_uid, sbuf.st_gid) < 0)
				bb_perror_msg("error changing owner of %s", MOUNTED_TEMP);
	    }
	}

	/* rename mtemp to mtab */
	if (rename (MOUNTED_TEMP, MOUNTED) < 0)
		bb_perror_msg("can't rename %s to %s", MOUNTED_TEMP, MOUNTED);

 leave:
	unlock_mtab();
}


#ifndef _NFS_UTILS_MOUNT_CONSTANTS_H
#define _NFS_UTILS_MOUNT_CONSTANTS_H

#ifndef MS_DIRSYNC
#define MS_DIRSYNC      128     /* Directory modifications are synchronous */
#endif

#ifndef MS_ACTION_MASK
#define MS_ACTION_MASK  0x380
/* Remount, but new filesystem may be different from old. Atomic
   (i.e. there is no interval when nothing is mounted at the mountpoint).
   If new fs differs from the old one and old is busy - -EBUSY. */
#define MS_REPLACE      0x080   /* 128 */
/* After, Before: as soon as we get unions these will add a new member
   in the end or beginning of the chain. Fail if there is a stack
   on the mountpoint. */
#define MS_AFTER        0x100   /* 256 */
#define MS_BEFORE       0x180
/* Over: if nothing mounted on a mountpoint - same as if none of these
flags had been set; if we have a union with more than one element - fail;
if we have a stack or plain mount - mount atop of it, forming a stack. */
#define MS_OVER         0x200   /* 512 */
#endif
#ifndef MS_NOATIME
#define MS_NOATIME      0x400   /* 1024: Do not update access times. */
#endif
#ifndef MS_NODIRATIME
#define MS_NODIRATIME   0x800   /* 2048: Don't update directory access times */
#endif
#ifndef MS_BIND
#define MS_BIND         0x1000  /* 4096: Mount existing tree also elsewhere */
#endif
#ifndef MS_MOVE
#define MS_MOVE         0x2000  /* 8192: Atomically move tree */
#endif
#ifndef MS_REC
#define MS_REC          0x4000  /* 16384: Recursive loopback */
#endif
#ifndef MS_VERBOSE
#define MS_VERBOSE      0x8000  /* 32768 */
#endif
#ifndef MS_RELATIME
#define MS_RELATIME 0x200000 /* 200000: Update access times relative
				  to mtime/ctime */
#endif
/*
 * NFS fs-specific mount option flags
 *
 * MS_DUMMY is assigned to mount options that are not to be
 * passed to the kernel via the "flags" argument.  These are
 * generally ignored or handled entirely in user space.
 */
#define MS_DUMMY        0x00000000
#define MS_USERS        0x40000000
#define MS_USER         0x80000000

/* Magic mount flag number. Had to be or-ed to the flag values. */
#ifndef MS_MGC_VAL
#define MS_MGC_VAL 0xC0ED0000   /* magic flag number to indicate "new" flags */
#endif
#ifndef MS_MGC_MSK
#define MS_MGC_MSK 0xffff0000   /* magic flag number mask */
#endif

#endif  /* _NFS_UTILS_MOUNT_CONSTANTS_H */

#define NFS_MOUNT_VERSION       6
#define NFS_MAX_CONTEXT_LEN     256

struct nfs2_fh {
	char                    data[32];
};
struct nfs3_fh {
	unsigned short          size;
	unsigned char           data[64];
};

struct nfs_mount_data {
	int             version;                /* 1 */
	int             fd;                     /* 1 */
	struct nfs2_fh  old_root;               /* 1 */
	int             flags;                  /* 1 */
	int             rsize;                  /* 1 */
	int             wsize;                  /* 1 */
	int             timeo;                  /* 1 */
	int             retrans;                /* 1 */
	int             acregmin;               /* 1 */
	int             acregmax;               /* 1 */
	int             acdirmin;               /* 1 */
	int             acdirmax;               /* 1 */
	struct sockaddr_in addr;                /* 1 */
	char            hostname[256];          /* 1 */
	int             namlen;                 /* 2 */
	unsigned int    bsize;                  /* 3 */
	struct nfs3_fh  root;                   /* 4 */
	int             pseudoflavor;           /* 5 */
	char    context[NFS_MAX_CONTEXT_LEN + 1]; /* 6 */

};

/* bits in the flags field */

#define NFS_MOUNT_SOFT          0x0001  /* 1 */
#define NFS_MOUNT_INTR          0x0002  /* 1 */
#define NFS_MOUNT_SECURE        0x0004  /* 1 */
#define NFS_MOUNT_POSIX         0x0008  /* 1 */
#define NFS_MOUNT_NOCTO         0x0010  /* 1 */
#define NFS_MOUNT_NOAC          0x0020  /* 1 */
#define NFS_MOUNT_TCP           0x0040  /* 2 */
#define NFS_MOUNT_VER3          0x0080  /* 3 */
#define NFS_MOUNT_KERBEROS      0x0100  /* 3 */
#define NFS_MOUNT_NONLM         0x0200  /* 3 */
#define NFS_MOUNT_BROKEN_SUID   0x0400  /* 4 */
#define NFS_MOUNT_NOACL     0x0800  /* 4 */
#define NFS_MOUNT_SECFLAVOUR    0x2000  /* 5 */
#define NFS_MOUNT_NORDIRPLUS    0x4000  /* 5 */
#define NFS_MOUNT_UNSHARED      0x8000  /* 5 */

/*
 * WARNING!  Do not delete or change the order of these fields.  If
 * a new field is required then add it to the end.  The version field
 * tracks which fields are present.  This will ensure some measure of
 * mount-to-kernel version compatibility.  Some of these aren't used yet
 * but here they are anyway.
 */
#define NFS4_MOUNT_VERSION      1

struct nfs_string {
	unsigned int len;
	const char* data;
};

struct nfs4_mount_data {
	int version;                            /* 1 */
	int flags;                              /* 1 */
	int rsize;                              /* 1 */
	int wsize;                              /* 1 */
	int timeo;                              /* 1 */
	int retrans;                            /* 1 */
	int acregmin;                           /* 1 */
	int acregmax;                           /* 1 */
	int acdirmin;                           /* 1 */
	int acdirmax;                           /* 1 */

	/* see the definition of 'struct clientaddr4' in RFC3010 */
	struct nfs_string client_addr;          /* 1 */

	/* Mount path */
	struct nfs_string mnt_path;             /* 1 */

	/* Server details */
	struct nfs_string hostname;             /* 1 */
	/* Server IP address */
	unsigned int host_addrlen;              /* 1 */
	struct sockaddr* host_addr;             /* 1 */

	/* Transport protocol to use */
	int proto;                              /* 1 */

	/* Pseudo-flavours to use for authentication. See RFC2623 */
	int auth_flavourlen;                    /* 1 */
	int *auth_flavours;                     /* 1 */
};

/* bits in the flags field */
/* Note: the fields that correspond to existing NFSv2/v3 mount options
 *       should mirror the values from include/linux/nfs_mount.h
 */

#define NFS4_MOUNT_SOFT         0x0001  /* 1 */
#define NFS4_MOUNT_INTR         0x0002  /* 1 */
#define NFS4_MOUNT_NOCTO        0x0010  /* 1 */
#define NFS4_MOUNT_NOAC         0x0020  /* 1 */
#define NFS4_MOUNT_STRICTLOCK   0x1000  /* 1 */
#define NFS4_MOUNT_UNSHARED     0x8000  /* 5 */
#define NFS4_MOUNT_FLAGMASK     0xFFFF


static inline unsigned int MAKE_VERSION(unsigned int p, unsigned int q,
					unsigned int r)
{
	return (65536 * p) + (256 * q) + r;
}

static int linux_version_code(void)
{
	if(kernel_version)
		return kernel_version;
	kernel_version = get_kernel_revision();
	return kernel_version;
}

#define NOMTAB    (bss_flags & 8)
#define SLOPPY    (bss_flags & 16)

#define FOREGROUND      (0)
#define BACKGROUND      (1)


/*
 * Map from -o and fstab option strings to the flag argument to mount(2).
 */
struct opt_map {
	const char *opt;        /* option name */
	int skip;               /* skip in mtab option string */
	int inv;                /* true if flag value should be inverted */
	int mask;               /* flag mask value */
};

static const struct opt_map opt_map[] = {
  { "defaults", 0, 0, 0         },      /* default options */
  { "ro",       1, 0, MS_RDONLY },      /* read-only */
  { "rw",       1, 1, MS_RDONLY },      /* read-write */
  { "exec",     0, 1, MS_NOEXEC },      /* permit execution of binaries */
  { "noexec",   0, 0, MS_NOEXEC },      /* don't execute binaries */
  { "suid",     0, 1, MS_NOSUID },      /* honor suid executables */
  { "nosuid",   0, 0, MS_NOSUID },      /* don't honor suid executables */
  { "dev",      0, 1, MS_NODEV  },      /* interpret device files  */
  { "nodev",    0, 0, MS_NODEV  },      /* don't interpret devices */
  { "sync",     0, 0, MS_SYNCHRONOUS},  /* synchronous I/O */
  { "async",    0, 1, MS_SYNCHRONOUS},  /* asynchronous I/O */
  { "dirsync",  0, 0, MS_DIRSYNC},      /* synchronous directory modifications */
  { "remount",  0, 0, MS_REMOUNT},      /* Alter flags of mounted FS */
  { "bind",     0, 0, MS_BIND   },      /* Remount part of tree elsewhere */
  { "rbind",    0, 0, MS_BIND|MS_REC }, /* Idem, plus mounted subtrees */
  { "auto",     0, 0, MS_DUMMY },       /* Can be mounted using -a */
  { "noauto",   0, 0, MS_DUMMY },       /* Can  only be mounted explicitly */
  { "users",    1, 0, MS_USERS },       /* Allow ordinary user to mount */
  { "nousers",  0, 1, MS_DUMMY  },      /* Forbid ordinary user to mount */
  { "user",     1, 0, MS_USER  },       /* Allow ordinary user to mount */
  { "nouser",   0, 1, MS_DUMMY   },     /* Forbid ordinary user to mount */
  { "owner",    0, 0, MS_DUMMY  },      /* Let the owner of the device mount */
  { "noowner",  0, 0, MS_DUMMY  },      /* Device owner has no special privs */
  { "group",    0, 0, MS_DUMMY  },      /* Let the group of the device mount */
  { "nogroup",  0, 0, MS_DUMMY  },      /* Device group has no special privs */
  { "_netdev",  0, 0, MS_DUMMY},        /* Device requires network */
  { "comment",  0, 0, MS_DUMMY},        /* fstab comment only (kudzu,_netdev)*/

  /* add new options here */
#ifdef MS_NOSUB
  { "sub",      0, 1, MS_NOSUB  },      /* allow submounts */
  { "nosub",    0, 0, MS_NOSUB  },      /* don't allow submounts */
#endif
#ifdef MS_SILENT
  { "quiet",    0, 0, MS_SILENT    },   /* be quiet  */
  { "loud",     0, 1, MS_SILENT    },   /* print out messages. */
#endif
#ifdef MS_MANDLOCK
  { "mand",     0, 0, MS_MANDLOCK },    /* Allow mandatory locks on this FS */
  { "nomand",   0, 1, MS_MANDLOCK },    /* Forbid mandatory locks on this FS */
#endif
  { "loop",     1, 0, MS_DUMMY   },      /* use a loop device */
#ifdef MS_NOATIME
  { "atime",    0, 1, MS_NOATIME },     /* Update access time */
  { "noatime",  0, 0, MS_NOATIME },     /* Do not update access time */
#endif
#ifdef MS_NODIRATIME
  { "diratime", 0, 1, MS_NODIRATIME },  /* Update dir access times */
  { "nodiratime", 0, 0, MS_NODIRATIME },/* Do not update dir access times */
#endif
#ifdef MS_RELATIME
  { "relatime", 0, 0, MS_RELATIME },   /* Update access times relative to
		      mtime/ctime */
  { "norelatime", 0, 1, MS_RELATIME }, /* Update access time without regard
		      to mtime/ctime */
#endif
  { "noquota", 0, 0, MS_DUMMY },        /* Don't enforce quota */
  { "quota", 0, 0, MS_DUMMY },          /* Enforce user quota */
  { "usrquota", 0, 0, MS_DUMMY },       /* Enforce user quota */
  { "grpquota", 0, 0, MS_DUMMY },       /* Enforce group quota */
  { NULL,       0, 0, 0         }
};

static void parse_opts(const char *options, int *flags, char **extra_opts);

/*
 * Choose the version of the nfs_mount_data structure that is appropriate
 * for the kernel that is doing the mount.
 *
 * NFS_MOUNT_VERSION:           maximum version supported by these sources
 * nfs_mount_data_version:      maximum version supported by the running kernel
 */
static void discover_nfs_mount_data_version(void)
{
	linux_version_code();

	if (kernel_version) {
		if (kernel_version < MAKE_VERSION(2, 1, 32))
			nfs_mount_data_version = 1;
		else if (kernel_version < MAKE_VERSION(2, 2, 18))
			nfs_mount_data_version = 3;
		else if (kernel_version < MAKE_VERSION(2, 3, 0))
			nfs_mount_data_version = 4;
		else if (kernel_version < MAKE_VERSION(2, 3, 99))
			nfs_mount_data_version = 3;
		else if (kernel_version < MAKE_VERSION(2, 6, 3))
			nfs_mount_data_version = 4;
		else
			nfs_mount_data_version = 6;
	}
	if (nfs_mount_data_version > NFS_MOUNT_VERSION)
		nfs_mount_data_version = NFS_MOUNT_VERSION;
	else
		if (kernel_version > MAKE_VERSION(2, 6, 22))
			nfsmount_string_flg++;
}

static void print_one(char *spec, char *node, char *type, char *opts)
{
	if (!verbose)
		return;

	if (opts)
		printf("%s on %s type %s (%s)\n", spec, node, type, opts);
	else
		printf("%s on %s type %s\n", spec, node, type);
}

/*
 * Build a canonical mount option string for /etc/mtab.
 */
static char *fix_opts_string(int flags, const char *extra_opts)
{
	const struct opt_map *om;
	char *new_opts;

	new_opts = xstrdup((flags & MS_RDONLY) ? "ro" : "rw");
	if (flags & MS_USER) {
		/* record who mounted this so they can unmount */
		struct passwd *pw = getpwuid(getuid());
		if(pw)
			new_opts = xstrconcat3(new_opts, ",user=", pw->pw_name);
	}
	if (flags & MS_USERS)
		new_opts = xstrconcat3(new_opts, ",users", "");

	for (om = opt_map; om->opt != NULL; om++) {
		if (om->skip)
			continue;
		if (om->inv || !om->mask || (flags & om->mask) != om->mask)
			continue;
		new_opts = xstrconcat3(new_opts, ",", om->opt);
		flags &= ~om->mask;
	}
	if (extra_opts && *extra_opts) {
		new_opts = xstrconcat3(new_opts, ",", extra_opts);
	}
	return new_opts;
}

/* Create mtab with a root entry.  */
static void create_mtab (void)
{
	struct mntentchn *fstab_ent;
	struct mntent mnt;
	int flags;
	mntFILE *mfp;

	lock_mtab();

	mfp = nfs_setmntent (MOUNTED, "a+");
	if (mfp == NULL || mfp->mntent_fp == NULL) {
		bb_default_error_retval = EX_FILEIO;
		bb_perror_msg_and_die ("can't open %s for writing", MOUNTED);
	}

	/* Find the root entry by looking it up in fstab */
	if ((fstab_ent = get_fsfile ("/")) || (fstab_ent = get_fsfile ("root"))) {
		char *extra_opts;
		parse_opts (fstab_ent->m.mnt_opts, &flags, &extra_opts);
		mnt.mnt_dir = "/";
		mnt.mnt_fsname = xstrdup(fstab_ent->m.mnt_fsname);
		mnt.mnt_type = fstab_ent->m.mnt_type;
		mnt.mnt_opts = fix_opts_string (flags, extra_opts);
		mnt.mnt_freq = mnt.mnt_passno = 0;
		free(extra_opts);

		if (nfs_addmntent (mfp, &mnt) == 1) {
			bb_default_error_retval = EX_FILEIO;
			bb_perror_msg_and_die ("error writing %s", _PATH_MOUNTED);
		}
	}
	if (fchmod (fileno (mfp->mntent_fp), 0644) < 0)
		if (errno != EROFS) {
			bb_default_error_retval = EX_FILEIO;
			bb_perror_msg_and_die ("error changing mode of %s", _PATH_MOUNTED);
		}
	nfs_endmntent (mfp);

	unlock_mtab();

	reset_mtab_info();
}

static int add_mtab(char *spec, char *mount_point, char *fstype,
			int flags, char *opts, int freq, int pass)
{
	struct mntent ment;
	int result = EX_SUCCESS;

	ment.mnt_fsname = spec;
	ment.mnt_dir = mount_point;
	ment.mnt_type = fstype;
	ment.mnt_opts = fix_opts_string(flags, opts);
	ment.mnt_freq = freq;
	ment.mnt_passno = pass;

	if (!NOMTAB && mtab_does_not_exist()) {
		if (verbose > 1)
			printf("mount: no %s found - creating it..\n",
			       MOUNTED);
		create_mtab ();
	}

	if (!NOMTAB && mtab_is_writable()) {
		if (flags & MS_REMOUNT)
			update_mtab(ment.mnt_dir, &ment);
		else {
			mntFILE *mtab;

			lock_mtab();
			mtab = nfs_setmntent(MOUNTED, "a+");
			if (mtab == NULL || mtab->mntent_fp == NULL) {
				bb_perror_msg("Can't open mtab");
				result = EX_FILEIO;
			} else {
				if (nfs_addmntent(mtab, &ment) == 1) {
					bb_perror_msg("Can't write mount entry to mtab");
					result = EX_FILEIO;
				}
			}
			nfs_endmntent(mtab);
			unlock_mtab();
		}
	}

	free(ment.mnt_opts);

	return result;
}

static void parse_opt(const char *opt, int *mask, char *extra_opts, int len)
{
	const struct opt_map *om = find_string_from_struct(opt_map, sizeof(struct opt_map), opt);
	if(om->opt != NULL) {
		if (om->inv)
			*mask &= ~om->mask;
		else
			*mask |= om->mask;
		return;
	}

	len -= strlen(extra_opts);

	if (*extra_opts && --len > 0)
		strcat(extra_opts, ",");

	if ((len -= strlen(opt)) > 0)
		strcat(extra_opts, opt);
}

/*
 * Convert the provided mount command-line options into the 4th &
 * 5th arguments to mount(2).  Output parameter "@flags" gets the
 * standard options (indicated by MS_ bits), and output parameter
 * "@extra_opts" gets all the filesystem-specific options.
 */
static void parse_opts(const char *options, int *flags, char **extra_opts)
{
	if (options != NULL) {
		char *opts = xstrdup(options);
		char *opt, *p;
		int len = strlen(opts) + 1;     /* include room for a null */
		int open_quote = 0;

		*extra_opts = xmalloc(len);
		**extra_opts = '\0';

		for (p = opts, opt = NULL; p && *p; p++) {
			if (!opt)
				opt = p;        /* begin of the option item */
			if (*p == '"')
				open_quote ^= 1; /* reverse the status */
			if (open_quote)
				continue;       /* still in a quoted block */
			if (*p == ',')
				*p = '\0';      /* terminate the option item */

			/* end of option item or last item */
			if (*p == '\0' || *(p + 1) == '\0') {
				parse_opt(opt, flags, *extra_opts, len);
				opt = NULL;
			}
		}
		free(opts);
	}
}

static int chk_mountpoint(const char *mount_point)
{
	if (is_directory(mount_point, 1, NULL) <= 0 || access(mount_point, X_OK) < 0) {
		mount_error(NULL, mount_point);
		return 1;
	}

	return 0;
}



typedef enum {
	PO_FAILED = 0,
	PO_SUCCEEDED = 1,
} po_return_t;

typedef enum {
	PO_NOT_FOUND = 0,
	PO_FOUND = 1,
	PO_BAD_VALUE = 2,
} po_found_t;

#define PMAP_TIMEOUT    (10)
#define CONNECT_TIMEOUT (20)
#define MOUNT_TIMEOUT   (30)

static const char * const nfs_ns_pgmtbl[] = {
	"status",
	NULL,
};

static const char * const nfs_mnt_pgmtbl[] = {
	"mount",
	"mountd",
	NULL,
};

static const char * const nfs_nfs_pgmtbl[] = {
	"nfs",
	"nfsprog",
	NULL,
};

static const char * const nfs_transport_opttbl[] = {
	"udp",
	"tcp",
	"rdma",
	"proto",
	NULL,
};

static const char * const nfs_version_opttbl[] = {
	"v2",
	"v3",
	"v4",
	"vers",
	"nfsvers",
	NULL,
};

static const unsigned long nfs_to_mnt[] = {
	0,
	0,
	1,
	3,
};

static const unsigned long mnt_to_nfs[] = {
	0,
	2,
	2,
	3,
};

/*
 * Map an NFS version into the corresponding Mountd version
 */
static unsigned long nfsvers_to_mnt(const unsigned long vers)
{
	if (vers <= 3)
		return nfs_to_mnt[vers];
	return 0;
}

/*
 * Map a Mountd version into the corresponding NFS version
 */
static unsigned long mntvers_to_nfs(const unsigned long vers)
{
	if (vers <= 3)
		return mnt_to_nfs[vers];
	return 0;
}

static const unsigned int probe_udp_only[] = {
	IPPROTO_UDP,
	0,
};

static const unsigned int probe_udp_first[] = {
	IPPROTO_UDP,
	IPPROTO_TCP,
	0,
};

static const unsigned int probe_tcp_first[] = {
	IPPROTO_TCP,
	IPPROTO_UDP,
	0,
};

static const unsigned long probe_nfs2_only[] = {
	2,
	0,
};

static const unsigned long probe_nfs3_first[] = {
	3,
	2,
	0,
};

static const unsigned long probe_mnt1_first[] = {
	1,
	2,
	0,
};

static const unsigned long probe_mnt3_first[] = {
	3,
	1,
	2,
	0,
};

static inline const unsigned int *nfs_default_proto(void)
{
	return probe_tcp_first;
}

/* nfs_lookup - resolve hostname to an IPv4 or IPv6 socket address
 * @hostname: pointer to C string containing DNS hostname to resolve
 * @family: address family hint
 * @sap: pointer to buffer to fill with socket address
 * @len: IN: size of buffer to fill; OUT: size of socket address
 *
 * Returns 1 and places a socket address at @sap if successful;
 * otherwise zero. */
static int nfs_lookup(const char *hostname, const sa_family_t family,
		struct sockaddr *sap, socklen_t *salen)
{
	struct addrinfo *gai_results;
	struct addrinfo gai_hint = {
#ifdef HAVE_DECL_AI_ADDRCONFIG
		.ai_flags       = AI_ADDRCONFIG,
#endif  /* HAVE_DECL_AI_ADDRCONFIG */
		.ai_family      = family,
	};
	socklen_t len = *salen;
	int error, ret = 0;

	*salen = 0;

	error = getaddrinfo(hostname, NULL, &gai_hint, &gai_results);
	switch (error) {
	case 0:
		break;
	case EAI_SYSTEM:
		bb_perror_msg("DNS resolution failed for %s", hostname);
		return ret;
	default:
		bb_error_msg("DNS resolution failed for %s: %s",
				hostname, gai_strerror(error));
		return ret;
	}

	switch (gai_results->ai_addr->sa_family) {
	case AF_INET:
	case AF_INET6:
		if (len >= gai_results->ai_addrlen) {
			*salen = gai_results->ai_addrlen;
			memcpy(sap, gai_results->ai_addr, *salen);
			ret = 1;
		}
		break;
	default:
		/* things are really broken if we get here, so warn */
		bb_error_msg("unrecognized DNS resolution results for %s", hostname);
		break;
	}

	freeaddrinfo(gai_results);
	return ret;
}

/* nfs_gethostbyname - resolve a hostname to an IPv4 address
 * @hostname: pointer to a C string containing a DNS hostname
 * @sin: returns an IPv4 address
 *
 * Returns 1 if successful, otherwise zero. */
static int nfs_gethostbyname(const char *hostname, struct sockaddr_in *sin)
{
	socklen_t len = sizeof(*sin);

	return nfs_lookup(hostname, AF_INET, (struct sockaddr *)sin, &len);
}

/* nfs_string_to_sockaddr - convert string address to sockaddr
 * @address:    pointer to presentation format address to convert
 * @sap:        pointer to socket address buffer to fill in
 * @salen:      IN: length of address buffer
 *              OUT: length of converted socket address
 *
 * Convert a presentation format address string to a socket address.
 * Similar to nfs_lookup(), but the DNS query is squelched, and it
 * won't make any noise if the getaddrinfo() call fails.
 *
 * Returns 1 and fills in @sap and @salen if successful; otherwise zero.
 *
 * See RFC 4038 section 5.1 or RFC 3513 section 2.2 for more details
 * on presenting IPv6 addresses as text strings. */
static int nfs_string_to_sockaddr(const char *address, struct sockaddr *sap,
			   socklen_t *salen)
{
	struct addrinfo *gai_results;
	struct addrinfo gai_hint = {
		.ai_flags       = AI_NUMERICHOST,
	};
	socklen_t len = *salen;
	int ret = 0;

	*salen = 0;

	if (getaddrinfo(address, NULL, &gai_hint, &gai_results) == 0) {
		switch (gai_results->ai_addr->sa_family) {
		case AF_INET:
		case AF_INET6:
			if (len >= gai_results->ai_addrlen) {
				*salen = gai_results->ai_addrlen;
				memcpy(sap, gai_results->ai_addr, *salen);
				ret = 1;
			}
			break;
		}
		freeaddrinfo(gai_results);
	}

	return ret;
}

/* nfs_present_sockaddr - convert sockaddr to string
 * @sap: pointer to socket address to convert
 * @salen: length of socket address
 * @buf: pointer to buffer to fill in
 * @buflen: length of buffer
 *
 * Convert the passed-in sockaddr-style address to presentation format.
 * The presentation format address is placed in @buf and is
 * '\0'-terminated.
 *
 * Returns 1 if successful; otherwise zero.
 *
 * See RFC 4038 section 5.1 or RFC 3513 section 2.2 for more details
 * on presenting IPv6 addresses as text strings. */
static int
nfs_present_sockaddr(const struct sockaddr *sap, const socklen_t salen,
			 char *buf, const size_t buflen)
{
	int result;

	result = getnameinfo(sap, salen, buf, buflen,
					NULL, 0, NI_NUMERICHOST);
	if (!result)
		return 1;

	bb_error_msg("invalid server address: %s", gai_strerror(result));
	return 0;
}

/* Attempt to connect a socket, but time out after "timeout" seconds.
 * On error return, caller closes the socket. */
static int connect_to(int fd, struct sockaddr *addr,
			socklen_t addrlen, int timeout)
{
	int ret, saved;
	fd_set rset, wset;
	struct timeval tv = {
		.tv_sec = timeout,
	};

	saved = fcntl(fd, F_GETFL, 0);
	fcntl(fd, F_SETFL, saved | O_NONBLOCK);

	ret = connect(fd, addr, addrlen);
	if (ret < 0 && errno != EINPROGRESS)
		return -1;
	if (ret == 0)
		goto out;

	FD_ZERO(&rset);
	FD_SET(fd, &rset);
	wset = rset;
	ret = select(fd + 1, &rset, &wset, NULL, &tv);
	if (ret == 0) {
		errno = ETIMEDOUT;
		return -1;
	}
	if (FD_ISSET(fd, &rset) || FD_ISSET(fd, &wset)) {
		int error;
		socklen_t len = sizeof(error);
		if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &len) < 0)
			return -1;
		if (error) {
			errno = error;
			return -1;
		}
	} else
		return -1;

out:
	fcntl(fd, F_SETFL, saved);
	return 0;
}

/* Create a socket that is locally bound to a reserved or non-reserved port.
 * The caller should check rpc_createerr to determine the cause of any error.
 */
static int get_socket(struct sockaddr_in *saddr, unsigned int p_prot,
			unsigned int timeout, int resvp, int conn)
{
	int so, cc, type;
	struct sockaddr_in laddr;
	socklen_t namelen = sizeof(laddr);

	type = (p_prot == IPPROTO_UDP ? SOCK_DGRAM : SOCK_STREAM);
	if ((so = socket (AF_INET, type, p_prot)) < 0)
		goto err_socket;

	laddr.sin_family = AF_INET;
	laddr.sin_port = 0;
	laddr.sin_addr.s_addr = htonl(INADDR_ANY);
	if (resvp) {
		if (bindresvport(so, &laddr) < 0)
			goto err_bindresvport;
	} else {
		cc = bind(so, (struct sockaddr *)&laddr, namelen);
		if (cc < 0)
			goto err_bind;
	}
	if (type == SOCK_STREAM || (conn && type == SOCK_DGRAM)) {
		cc = connect_to(so, (struct sockaddr *)saddr, namelen,
				timeout);
		if (cc < 0)
			goto err_connect;
	}
	return so;

err_socket:
	rpc_createerr.cf_stat = RPC_SYSTEMERROR;
	rpc_createerr.cf_error.re_errno = errno;
	if (verbose)
		bb_perror_msg("Unable to create %s socket",
			p_prot == IPPROTO_UDP ? "UDP" : "TCP");
	return RPC_ANYSOCK;

err_bindresvport:
	rpc_createerr.cf_stat = RPC_SYSTEMERROR;
	rpc_createerr.cf_error.re_errno = errno;
	if (verbose)
		bb_perror_msg("Unable to bindresvport %s socket",
				p_prot == IPPROTO_UDP ? "UDP" : "TCP");
	close(so);
	return RPC_ANYSOCK;

err_bind:
	rpc_createerr.cf_stat = RPC_SYSTEMERROR;
	rpc_createerr.cf_error.re_errno = errno;
	if (verbose)
		bb_perror_msg("Unable to bind to %s socket",
				p_prot == IPPROTO_UDP ? "UDP" : "TCP");
	close(so);
	return RPC_ANYSOCK;

err_connect:
	rpc_createerr.cf_stat = RPC_SYSTEMERROR;
	rpc_createerr.cf_error.re_errno = errno;
	if (verbose)
		bb_perror_msg("Unable to connect to %s:%d",
				inet_ntoa(saddr->sin_addr),
				ntohs(saddr->sin_port));
	close(so);
	return RPC_ANYSOCK;
}

static void nfs_pp_debug(const struct sockaddr *sap, const socklen_t salen,
			 const rpcprog_t program, const rpcvers_t version,
			 const unsigned short protocol,
			 const unsigned short port)
{
	char buf[NI_MAXHOST];

	if (!verbose)
		return;

	if (nfs_present_sockaddr(sap, salen, buf, sizeof(buf)) == 0)
		strcpy(buf, "unknown host");

	bb_error_msg("trying %s prog %lu vers %lu prot %s port %d\n",
			buf, (unsigned long)program,
			(unsigned long)version,
			(protocol == IPPROTO_UDP ? "UDP" : "TCP"),
			port);
}

static void nfs_pp_debug2(const char *str)
{
	if (!verbose)
		return;

	if (rpc_createerr.cf_error.re_status == RPC_CANTRECV ||
	    rpc_createerr.cf_error.re_status == RPC_CANTSEND)
		bb_error_msg("portmap query %s%s - %s",
				str, clnt_spcreateerror(""),
				strerror(rpc_createerr.cf_error.re_errno));
	else
		bb_error_msg("portmap query %s%s", str, clnt_spcreateerror(""));
}

/* Use the portmapper to discover whether or not the service we want is
 * available. The lists 'versions' and 'protos' define ordered sequences
 * of service versions and udp/tcp protocols to probe for.
 *
 * Returns 1 if the requested service port is unambiguous and pingable;
 * @pmap is filled in with the version, port, and transport protocol used
 * during the successful ping.  Note that if a port is already specified
 * in @pmap and it matches the rpcbind query result, nfs_probe_port() does
 * not perform an RPC ping.
 *
 * If an error occurs or the requested service isn't available, zero is
 * returned; rpccreateerr.cf_stat is set to reflect the nature of the error.
 */
static int nfs_probe_port(const struct sockaddr *sap, const socklen_t salen,
			  struct pmap *pmap, const unsigned long *versions,
			  const unsigned int *protos)
{
	union nfs_sockaddr address;
	struct sockaddr *saddr = &address.sa;
	const unsigned long prog = pmap->pm_prog, *p_vers;
	const unsigned int prot = (u_int)pmap->pm_prot, *p_prot;
	const u_short port = (u_short) pmap->pm_port;
	unsigned long vers = pmap->pm_vers;
	unsigned short p_port;

	memcpy(saddr, sap, salen);
	p_prot = prot ? &prot : protos;
	p_vers = vers ? &vers : versions;

	for (;;) {
		if (verbose)
			printf("%s: prog %lu, trying vers=%lu, prot=%u\n",
				bb_applet_name, prog, *p_vers, *p_prot);
		p_port = nfs_getport(saddr, salen, prog, *p_vers, *p_prot);
		if (p_port) {
			if (!port || port == p_port) {
				nfs_set_port(saddr, p_port);
				nfs_pp_debug(saddr, salen, prog, *p_vers,
						*p_prot, p_port);
				if (nfs_rpc_ping(saddr, salen, prog,
							*p_vers, *p_prot, NULL))
					goto out_ok;
			} else
				rpc_createerr.cf_stat = RPC_PROGNOTREGISTERED;
		}
		if (rpc_createerr.cf_stat != RPC_PROGNOTREGISTERED &&
		    rpc_createerr.cf_stat != RPC_TIMEDOUT &&
		    rpc_createerr.cf_stat != RPC_CANTRECV &&
		    rpc_createerr.cf_stat != RPC_PROGVERSMISMATCH)
			break;

		if (!prot) {
			if (*++p_prot) {
				nfs_pp_debug2("retrying");
				continue;
			}
			p_prot = protos;
		}
		if (rpc_createerr.cf_stat == RPC_TIMEDOUT ||
		    rpc_createerr.cf_stat == RPC_CANTRECV)
			break;

		if (vers || !*++p_vers)
			break;
	}

	nfs_pp_debug2("failed");
	return 0;

out_ok:
	if (!vers)
		pmap->pm_vers = *p_vers;
	if (!prot)
		pmap->pm_prot = *p_prot;
	if (!port)
		pmap->pm_port = p_port;
	nfs_clear_rpc_createerr();
	return 1;
}
/* Probe a server's NFS service to determine which versions and
 * transport protocols are supported.
 *
 * Returns 1 if the requested service port is unambiguous and pingable;
 * @pmap is filled in with the version, port, and transport protocol used
 * during the successful ping.  If all three are already specified, simply
 * return success without an rpcbind query or RPC ping (we may be trying
 * to mount an NFS service that is not advertised via rpcbind).
 *
 * If an error occurs or the requested service isn't available, zero is
 * returned; rpccreateerr.cf_stat is set to reflect the nature of the error.
 */
static int nfs_probe_nfsport(const struct sockaddr *sap, const socklen_t salen,
				struct pmap *pmap)
{
	if (pmap->pm_vers && pmap->pm_prot && pmap->pm_port)
		return 1;

	if (nfs_mount_data_version >= 4) {
		const unsigned int *probe_proto;

		probe_proto = nfs_default_proto();

		return nfs_probe_port(sap, salen, pmap,
					probe_nfs3_first, probe_proto);
	} else
		return nfs_probe_port(sap, salen, pmap,
					probe_nfs2_only, probe_udp_only);
}

/* Probe a server's mountd service to determine which versions and
 * transport protocols are supported.
 *
 * Returns 1 if the requested service port is unambiguous and pingable;
 * @pmap is filled in with the version, port, and transport protocol used
 * during the successful ping.  If all three are already specified, simply
 * return success without an rpcbind query or RPC ping (we may be trying
 * to mount an NFS service that is not advertised via rpcbind).
 *
 * If an error occurs or the requested service isn't available, zero is
 * returned; rpccreateerr.cf_stat is set to reflect the nature of the error.
 */
static int nfs_probe_mntport(const struct sockaddr *sap, const socklen_t salen,
				struct pmap *pmap)
{
	if (pmap->pm_vers && pmap->pm_prot && pmap->pm_port)
		return 1;

	if (nfs_mount_data_version >= 4)
		return nfs_probe_port(sap, salen, pmap,
					probe_mnt3_first, probe_udp_first);
	else
		return nfs_probe_port(sap, salen, pmap,
					probe_mnt1_first, probe_udp_only);
}

/* Probe a server's mountd service to determine which versions and
 * transport protocols are supported.  Invoked when the protocol
 * version is already known for both the NFS and mountd service.
 *
 * Returns 1 and fills in both @pmap structs if the requested service
 * ports are unambiguous and pingable.  Otherwise zero is returned;
 * rpccreateerr.cf_stat is set to reflect the nature of the error. */
static int nfs_probe_version_fixed(const struct sockaddr *mnt_saddr,
			const socklen_t mnt_salen,
			struct pmap *mnt_pmap,
			const struct sockaddr *nfs_saddr,
			const socklen_t nfs_salen,
			struct pmap *nfs_pmap)
{
	if (!nfs_probe_nfsport(nfs_saddr, nfs_salen, nfs_pmap))
		return 0;
	return nfs_probe_mntport(mnt_saddr, mnt_salen, mnt_pmap);
}

/* nfs_probe_bothports - discover the RPC endpoints of mountd and NFS server
 * @mnt_saddr:  pointer to socket address of mountd server
 * @mnt_salen:  length of mountd server's address
 * @mnt_pmap:   IN: partially filled-in mountd RPC service tuple;
 *              OUT: fully filled-in mountd RPC service tuple
 * @nfs_saddr:  pointer to socket address of NFS server
 * @nfs_salen:  length of NFS server's address
 * @nfs_pmap:   IN: partially filled-in NFS RPC service tuple;
 *              OUT: fully filled-in NFS RPC service tuple
 *
 * Returns 1 and fills in both @pmap structs if the requested service
 * ports are unambiguous and pingable.  Otherwise zero is returned;
 * rpccreateerr.cf_stat is set to reflect the nature of the error. */
static int
nfs_probe_bothports(const struct sockaddr *mnt_saddr,
			const socklen_t mnt_salen,
			struct pmap *mnt_pmap,
			const struct sockaddr *nfs_saddr,
			const socklen_t nfs_salen,
			struct pmap *nfs_pmap)
{
	struct pmap save_nfs, save_mnt;
	const unsigned long *probe_vers;

	if (mnt_pmap->pm_vers && !nfs_pmap->pm_vers)
		nfs_pmap->pm_vers = mntvers_to_nfs(mnt_pmap->pm_vers);
	else if (nfs_pmap->pm_vers && !mnt_pmap->pm_vers)
		mnt_pmap->pm_vers = nfsvers_to_mnt(nfs_pmap->pm_vers);

	if (nfs_pmap->pm_vers)
		return nfs_probe_version_fixed(mnt_saddr, mnt_salen, mnt_pmap,
					       nfs_saddr, nfs_salen, nfs_pmap);

	memcpy(&save_nfs, nfs_pmap, sizeof(save_nfs));
	memcpy(&save_mnt, mnt_pmap, sizeof(save_mnt));
	probe_vers = (nfs_mount_data_version >= 4) ?
			probe_mnt3_first : probe_mnt1_first;

	for (; *probe_vers; probe_vers++) {
		nfs_pmap->pm_vers = mntvers_to_nfs(*probe_vers);
		if (nfs_probe_nfsport(nfs_saddr, nfs_salen, nfs_pmap) != 0) {
			mnt_pmap->pm_vers = *probe_vers;
			if (nfs_probe_mntport(mnt_saddr, mnt_salen, mnt_pmap) != 0)
				return 1;
			memcpy(mnt_pmap, &save_mnt, sizeof(*mnt_pmap));
		}
		switch (rpc_createerr.cf_stat) {
		case RPC_PROGVERSMISMATCH:
		case RPC_PROGNOTREGISTERED:
			break;
		default:
			return 0;
		}
		memcpy(nfs_pmap, &save_nfs, sizeof(*nfs_pmap));
	}

	return 0;
}

/* probe_bothports - discover the RPC endpoints of mountd and NFS server
 * @mnt_server: pointer to address and pmap argument for mountd results
 * @nfs_server: pointer to address and pmap argument for NFS server
 *
 * This is the legacy API that takes "clnt_addr_t" for both servers,
 * but supports only AF_INET addresses.
 *
 * Returns 1 and fills in the pmap field in both clnt_addr_t structs
 * if the requested service ports are unambiguous and pingable.
 * Otherwise zero is returned; rpccreateerr.cf_stat is set to reflect
 * the nature of the error. */
static int probe_bothports(clnt_addr_t *mnt_server, clnt_addr_t *nfs_server)
{
	return nfs_probe_bothports((struct sockaddr *)&mnt_server->saddr,
					sizeof(mnt_server->saddr),
					&mnt_server->pmap,
					(struct sockaddr *)&nfs_server->saddr,
					sizeof(nfs_server->saddr),
					&nfs_server->pmap);
}

static int nfs_probe_statd(void)
{
	struct sockaddr_in addr = {
		.sin_family             = AF_INET,
		.sin_addr.s_addr        = htonl(INADDR_LOOPBACK),
	};
	rpcprog_t program = nfs_getrpcbyname(NSMPROG, nfs_ns_pgmtbl);

	return nfs_getport_ping((struct sockaddr *)&addr, sizeof(addr),
				program, (rpcvers_t)1, IPPROTO_UDP);
}

/* start_statd - attempt to start rpc.statd
 * Returns 1 if statd is running; otherwise zero. */
static int start_statd(void)
{
#ifdef START_STATD
	struct stat stb;
#endif

	if (nfs_probe_statd())
		return 1;

#ifdef START_STATD
	if (stat(START_STATD, &stb) == 0) {
		if (S_ISREG(stb.st_mode) && (stb.st_mode & S_IXUSR)) {
			pid_t pid = fork();
			switch (pid) {
			case 0: /* child */
				execl(START_STATD, START_STATD, NULL);
				exit(1);
			case -1: /* error */
				bb_perror_msg("fork failed");
				break;
			default: /* parent */
				waitpid(pid, NULL,0);
				break;
			}
			if (nfs_probe_statd())
				return 1;
		}
	}
#endif

	return 0;
}


/* mnt_closeclnt - terminate a handle for a remote mountd service
 * @clnt: pointer to an active handle for a remote mountd service
 * @msock: file descriptor of the underlying transport socket */
static void mnt_closeclnt(CLIENT *clnt, int msock)
{
	auth_destroy(clnt->cl_auth);
	clnt_destroy(clnt);
	close(msock);
}

/* mnt_openclnt - get a handle for a remote mountd service
 * @mnt_server: address and pmap arguments of mountd service
 * @msock: returns a file descriptor of the underlying transport socket
 *
 * Returns an active handle for the remote's mountd service */
static CLIENT *mnt_openclnt(clnt_addr_t *mnt_server, int *msock)
{
	struct sockaddr_in *mnt_saddr = &mnt_server->saddr;
	struct pmap *mnt_pmap = &mnt_server->pmap;
	CLIENT *clnt = NULL;

	mnt_saddr->sin_port = htons((u_short)mnt_pmap->pm_port);
	*msock = get_socket(mnt_saddr, mnt_pmap->pm_prot, MOUNT_TIMEOUT,
				TRUE, FALSE);
	if (*msock == RPC_ANYSOCK) {
		if (rpc_createerr.cf_error.re_errno == EADDRINUSE)
			/*
			 * Probably in-use by a TIME_WAIT connection,
			 * It is worth waiting a while and trying again.
			 */
			rpc_createerr.cf_stat = RPC_TIMEDOUT;
		return NULL;
	}

	switch (mnt_pmap->pm_prot) {
	case IPPROTO_UDP:
		clnt = clntudp_bufcreate(mnt_saddr,
					 mnt_pmap->pm_prog, mnt_pmap->pm_vers,
					 RETRY_TIMEOUT, msock,
					 MNT_SENDBUFSIZE, MNT_RECVBUFSIZE);
		break;
	case IPPROTO_TCP:
		clnt = clnttcp_create(mnt_saddr,
				      mnt_pmap->pm_prog, mnt_pmap->pm_vers,
				      msock,
				      MNT_SENDBUFSIZE, MNT_RECVBUFSIZE);
		break;
	}
	if (clnt) {
		/* try to mount hostname:dirname */
		clnt->cl_auth = nfs_authsys_create();
		if (clnt->cl_auth)
			return clnt;
		CLNT_DESTROY(clnt);
	}
	return NULL;
}



/* clnt_ping - send an RPC ping to the remote RPC service endpoint
 * @saddr: server's address
 * @prog: target RPC program number
 * @vers: target RPC version number
 * @prot: target RPC protocol
 * @caddr: filled in with our network address
 *
 * Sigh... GETPORT queries don't actually check the version number.
 * In order to make sure that the server actually supports the service
 * we're requesting, we open an RPC client, and fire off a NULL
 * RPC call.
 *
 * caddr is the network address that the server will use to call us back.
 * On multi-homed clients, this address depends on which NIC we use to
 * route requests to the server.
 *
 * Returns one if successful, otherwise zero. */
static int clnt_ping(struct sockaddr_in *saddr, const unsigned long prog,
		const unsigned long vers, const unsigned int prot,
		struct sockaddr_in *caddr)
{
	CLIENT *clnt = NULL;
	int sock, cf_stat;
	struct sockaddr dissolve;

	rpc_createerr.cf_stat = cf_stat = 0;
	sock = get_socket(saddr, prot, CONNECT_TIMEOUT, FALSE, TRUE);
	if (sock == RPC_ANYSOCK) {
		if (rpc_createerr.cf_error.re_errno == ETIMEDOUT) {
			/*
			 * TCP timeout. Bubble up the error to see
			 * how it should be handled.
			 */
			rpc_createerr.cf_stat = RPC_TIMEDOUT;
		}
		return 0;
	}

	if (caddr) {
		/* Get the address of our end of this connection */
		socklen_t len = sizeof(*caddr);
		if (getsockname(sock, (struct sockaddr *)caddr, &len) != 0)
			caddr->sin_family = 0;
	}

	switch(prot) {
	case IPPROTO_UDP:
		/* The socket is connected (so we could getsockname successfully),
		 * but some servers on multi-homed hosts reply from
		 * the wrong address, so if we stay connected, we lose the reply.
		 */
		dissolve.sa_family = AF_UNSPEC;
		connect(sock, &dissolve, sizeof(dissolve));

		clnt = clntudp_bufcreate(saddr, prog, vers,
					 RETRY_TIMEOUT, &sock,
					 RPCSMALLMSGSIZE, RPCSMALLMSGSIZE);
		break;
	case IPPROTO_TCP:
		clnt = clnttcp_create(saddr, prog, vers, &sock,
				      RPCSMALLMSGSIZE, RPCSMALLMSGSIZE);
		break;
	}
	if (!clnt) {
		close(sock);
		return 0;
	}
	clnt_res = '\0';
	cf_stat = clnt_call(clnt, NULLPROC,
			 (xdrproc_t)xdr_void, (caddr_t)NULL,
			 (xdrproc_t)xdr_void, (caddr_t)&clnt_res,
			 TIMEOUT);
	if (cf_stat) {
		clnt_geterr(clnt, &rpc_createerr.cf_error);
		rpc_createerr.cf_stat = cf_stat;
	}
	clnt_destroy(clnt);
	close(sock);

	if (cf_stat == RPC_SUCCESS)
		return 1;
	else
		return 0;
}

/* Try a getsockname() on a connected datagram socket.
 *
 * Returns 1 and fills in @buf if successful; otherwise, zero.
 *
 * A connected datagram socket prevents leaving a socket in TIME_WAIT.
 * This conserves the ephemeral port number space, helping reduce failed
 * socket binds during mount storms. */
static int nfs_ca_sockname(const struct sockaddr *sap, const socklen_t salen,
			   struct sockaddr *buf, socklen_t *buflen)
{
	struct sockaddr_in sin = {
		.sin_family             = AF_INET,
		.sin_addr.s_addr        = htonl(INADDR_ANY),
	};
#ifdef CONFIG_FEATURE_IPV6
	struct sockaddr_in6 sin6 = {
		.sin6_family            = AF_INET6,
		.sin6_addr              = IN6ADDR_ANY_INIT,
	};
#endif
	int sock;

	sock = socket(sap->sa_family, SOCK_DGRAM, IPPROTO_UDP);
	if (sock < 0)
		return 0;

	switch (sap->sa_family) {
	case AF_INET:
		if (bind(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
			close(sock);
			return 0;
		}
		break;
#ifdef CONFIG_FEATURE_IPV6
	case AF_INET6:
		if (bind(sock, (struct sockaddr *)&sin6, sizeof(sin6)) < 0) {
			close(sock);
			return 0;
		}
		break;
#endif
	default:
		errno = EAFNOSUPPORT;
		return 0;
	}

	if (connect(sock, sap, salen) < 0) {
		close(sock);
		return 0;
	}

	return !getsockname(sock, buf, buflen);
}

/* Try to generate an address that prevents the server from calling us.
 *
 * Returns 1 and fills in @buf if successful; otherwise, zero. */
static int nfs_ca_gai(const struct sockaddr *sap,
		      struct sockaddr *buf, socklen_t *buflen)
{
	struct addrinfo *gai_results;
	struct addrinfo gai_hint = {
		.ai_family      = sap->sa_family,
		.ai_flags       = AI_PASSIVE,   /* ANYADDR */
	};

	if (getaddrinfo(NULL, "", &gai_hint, &gai_results))
		return 0;

	*buflen = gai_results->ai_addrlen;
	memcpy(buf, gai_results->ai_addr, *buflen);

	freeaddrinfo(gai_results);

	return 1;
}

/* nfs_callback_address - acquire our local network address
 * @sap: pointer to address of remote
 * @sap_len: length of address
 * @buf: pointer to buffer to be filled in with local network address
 * @buflen: IN: length of buffer to fill in; OUT: length of filled-in address
 *
 * Discover a network address that an NFSv4 server can use to call us back.
 * On multi-homed clients, this address depends on which NIC we use to
 * route requests to the server.
 *
 * Returns 1 and fills in @buf if an unambiguous local address is
 * available; returns 1 and fills in an appropriate ANYADDR address
 * if a local address isn't available; otherwise, returns zero. */
static int
nfs_callback_address(const struct sockaddr *sap, const socklen_t salen,
			 struct sockaddr *buf, socklen_t *buflen)
{
#ifdef CONFIG_FEATURE_IPV6
	struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)buf;
#endif

	if (nfs_ca_sockname(sap, salen, buf, buflen) == 0)
		if (nfs_ca_gai(sap, buf, buflen) == 0)
			goto out_failed;

#ifdef CONFIG_FEATURE_IPV6
	/*
	 * The server can't use an interface ID that was generated
	 * here on the client, so always clear sin6_scope_id.
	 */
	if (sin6->sin6_family == AF_INET6)
		sin6->sin6_scope_id = 0;
#endif

	return 1;

out_failed:
	*buflen = 0;
	if (verbose)
		bb_error_msg("failed to construct callback address");
	return 0;
}

/* po_get - return the value of the rightmost instance of an option
 * @options: pointer to mount options
 * @keyword: pointer to a C string containing option keyword for which to search
 *
 * If multiple instances of the same option are present in a mount option
 * list, the rightmost instance is always the effective one.
 *
 * Returns pointer to C string containing the value of the option.
 * Returns NULL if the option isn't found, or if the option doesn't
 * have a value. */
static char *po_get(struct mount_options *options, char *keyword)
{
	struct mount_option *option;

	if (options && keyword) {
		for (option = options->tail; option; option = option->prev)
			if (strcmp(option->keyword, keyword) == 0)
				return option->value;
	}

	return NULL;
}

/* po_get_numeric - return numeric value of rightmost instance of keyword option
 * @options: pointer to mount options
 * @keyword: pointer to a C string containing option keyword for which to search
 * @value: OUT: set to the value of the keyword
 *
 * This is specifically for parsing keyword options that take only a numeric
 * value.  If multiple instances of the same option are present in a mount
 * option list, the rightmost instance is always the effective one.
 *
 * Returns:
 *      * PO_FOUND if the keyword was found and the value is numeric; @value is
 *        set to the keyword's value
 *      * PO_NOT_FOUND if the keyword was not found
 *      * PO_BAD_VALUE if the keyword was found, but the value is not numeric
 *
 * These last two are separate in case the caller wants to warn about bad mount
 * options instead of silently using a default. */
static po_found_t
po_get_numeric(struct mount_options *options, char *keyword, long *value)
{
	char *option, *endptr;
	long tmp;

	option = po_get(options, keyword);
	if (option == NULL)
		return PO_NOT_FOUND;

	errno = 0;
	tmp = strtol(option, &endptr, 10);
	if (errno == 0 && endptr != option) {
		*value = tmp;
		return PO_FOUND;
	}
	return PO_BAD_VALUE;
}

/* "nfsprog" is supported only by the legacy mount command.  The
 * kernel mount client does not support this option.
 *
 * Returns TRUE if @program contains a valid value for this option,
 * or FALSE if the option was specified with an invalid value. */
static int
nfs_nfs_program(struct mount_options *options, unsigned long *program)
{
	long tmp;

	switch (po_get_numeric(options, "nfsprog", &tmp)) {
	case PO_NOT_FOUND:
		break;
	case PO_FOUND:
		if (tmp > 0) {
			*program = tmp;
			return 1;
		}
	case PO_BAD_VALUE:
		bb_error_msg("invalid value for 'nfsprog=' option");
		return 0;
	}

	/*
	 * NFS RPC program wasn't specified.  The RPC program
	 * cannot be determined via an rpcbind query.
	 */
	*program = nfs_getrpcbyname(NFSPROG, nfs_nfs_pgmtbl);
	return 1;
}

/* po_rightmost - determine the relative position of several options
 * @options: pointer to mount options
 * @keys: pointer to an array of C strings containing option keywords
 *
 * This function can be used to determine which of several similar
 * options will be the one to take effect.
 *
 * The kernel parses the mount option string from left to right.
 * If an option is specified more than once (for example, "intr"
 * and "nointr", the rightmost option is the last to be parsed,
 * and it therefore takes precedence over previous similar options.
 *
 * This can also distinguish among multiple synonymous options, such
 * as "proto=," "udp" and "tcp."
 *
 * Returns the index into @keys of the option that is rightmost.
 * If none of the options listed in @keys is present in @options, or
 * if @options is NULL, returns -1. */
static int po_rightmost(struct mount_options *options, const char * const keys[])
{
	struct mount_option *option;

	if (options) {
		for (option = options->tail; option; option = option->prev) {
			int i = compare_string_array(keys, option->keyword);
			if(i >= 0)
				return i;
		}
	}

	return -1;
}

/* Returns TRUE if @version contains a valid value for this option,
 * or FALSE if the option was specified with an invalid value. */
static int
nfs_nfs_version(struct mount_options *options, unsigned long *version)
{
	long tmp;

	switch (po_rightmost(options, nfs_version_opttbl)) {
	case 0: /* v2 */
		*version = 2;
		return 1;
	case 1: /* v3 */
		*version = 3;
		return 1;
	case 2: /* v4 */
		*version = 4;
		return 1;
	case 3: /* vers */
		switch (po_get_numeric(options, "vers", &tmp)) {
		case PO_FOUND:
			if (tmp >= 2 && tmp <= 4) {
				*version = tmp;
				return 1;
			}
			return 0;
		case PO_NOT_FOUND:
			bb_error_msg("parsing error on 'vers=' option");
			return 0;
		case PO_BAD_VALUE:
			bb_error_msg("invalid value for 'vers=' option");
			return 0;
		}
	case 4: /* nfsvers */
		switch (po_get_numeric(options, "nfsvers", &tmp)) {
		case PO_FOUND:
			if (tmp >= 2 && tmp <= 4) {
				*version = tmp;
				return 1;
			}
			return 0;
		case PO_NOT_FOUND:
			bb_error_msg("parsing error on 'nfsvers=' option");
			return 0;
		case PO_BAD_VALUE:
			bb_error_msg("invalid value for 'nfsvers=' option");
			return 0;
		}
	}

	/*
	 * NFS version wasn't specified.  The pmap version value
	 * will be filled in later by an rpcbind query in this case.
	 */
	*version = 0;
	return 1;
}



/* Returns TRUE if @protocol contains a valid value for this option,
 * or FALSE if the option was specified with an invalid value. On
 * error, errno is set. */
static int
nfs_nfs_protocol(struct mount_options *options, unsigned long *protocol)
{
	sa_family_t family;
	char *option;

	switch (po_rightmost(options, nfs_transport_opttbl)) {
	case 0: /* udp */
		*protocol = IPPROTO_UDP;
		return 1;
	case 1: /* tcp */
		*protocol = IPPROTO_TCP;
		return 1;
	case 2: /* rdma */
		*protocol = NFSPROTO_RDMA;
		return 1;
	case 3: /* proto */
		option = po_get(options, "proto");
		if (option != NULL) {
			if (!nfs_get_proto(option, &family, protocol)) {
				errno = EPROTONOSUPPORT;
				bb_error_msg("Failed to find '%s' protocol", option);
				return 0;
			}
			return 1;
		}
	}

	/*
	 * NFS transport protocol wasn't specified.  The pmap
	 * protocol value will be filled in later by an rpcbind
	 * query in this case.
	 */
	*protocol = 0;
	return 1;
}

/* Returns TRUE if @port contains a valid value for this option,
 * or FALSE if the option was specified with an invalid value. */
static int
nfs_nfs_port(struct mount_options *options, unsigned long *port)
{
	long tmp;

	switch (po_get_numeric(options, "port", &tmp)) {
	case PO_NOT_FOUND:
		break;
	case PO_FOUND:
		if (tmp >= 1 && tmp <= 65535) {
			*port = tmp;
			return 1;
		}
	case PO_BAD_VALUE:
		bb_error_msg("invalid value for 'port=' option");
		return 0;
	}

	/*
	 * NFS service port wasn't specified.  The pmap port value
	 * will be filled in later by an rpcbind query in this case.
	 */
	*port = 0;
	return 1;
}

#ifdef CONFIG_FEATURE_IPV6
static const sa_family_t     config_default_family = AF_UNSPEC;

static int
nfs_verify_family(sa_family_t UNUSED(family))
{
	return 1;
}
#else /* CONFIG_FEATURE_IPV6 */
static const sa_family_t     config_default_family = AF_INET;

static int
nfs_verify_family(sa_family_t family)
{
	if (family != AF_INET)
		return 0;

	return 1;
}
#endif /* CONFIG_FEATURE_IPV6 */

/* Returns TRUE and fills in @family if a valid NFS protocol option
 * is found, or FALSE if the option was specified with an invalid value
 * or if the protocol family isn't supported. On error, errno is set. */
static int nfs_nfs_proto_family(struct mount_options *options,
				sa_family_t *family)
{
	unsigned long protocol;
	char *option;
	sa_family_t tmp_family = config_default_family;

	switch (po_rightmost(options, nfs_transport_opttbl)) {
	case 0: /* udp */
	case 1: /* tcp */
	case 2: /* rdma */
		/* for compatibility; these are always AF_INET */
		*family = AF_INET;
		return 1;
	case 3: /* proto */
		option = po_get(options, "proto");
		if (option != NULL &&
		    !nfs_get_proto(option, &tmp_family, &protocol)) {

			bb_error_msg("Failed to find '%s' protocol", option);
			errno = EPROTONOSUPPORT;
			return 0;
		}
	}

	if (!nfs_verify_family(tmp_family))
		goto out_err;
	*family = tmp_family;
	return 1;
out_err:
	errno = EAFNOSUPPORT;
	return 0;
}

/* "mountprog" is supported only by the legacy mount command.  The
 * kernel mount client does not support this option.
 *
 * Returns TRUE if @program contains a valid value for this option,
 * or FALSE if the option was specified with an invalid value. */
static int
nfs_mount_program(struct mount_options *options, unsigned long *program)
{
	long tmp;

	switch (po_get_numeric(options, "mountprog", &tmp)) {
	case PO_NOT_FOUND:
		break;
	case PO_FOUND:
		if (tmp > 0) {
			*program = tmp;
			return 1;
		}
	case PO_BAD_VALUE:
		bb_error_msg("invalid value for 'mountprog=' option");
		return 0;
	}

	/*
	 * MNT RPC program wasn't specified.  The RPC program
	 * cannot be determined via an rpcbind query.
	 */
	*program = nfs_getrpcbyname(MOUNTPROG, nfs_mnt_pgmtbl);
	return 1;
}

/* Returns TRUE if @version contains a valid value for this option,
 * or FALSE if the option was specified with an invalid value. */
static int
nfs_mount_version(struct mount_options *options, unsigned long *version)
{
	long tmp;

	switch (po_get_numeric(options, "mountvers", &tmp)) {
	case PO_NOT_FOUND:
		break;
	case PO_FOUND:
		if (tmp >= 1 && tmp <= 4) {
			*version = tmp;
			return 1;
		}
	case PO_BAD_VALUE:
		bb_error_msg("invalid value for 'mountvers=' option");
		return 0;
	}

	/* MNT version wasn't specified.  The pmap version value
	 * will be filled in later by an rpcbind query in this case. */
	*version = 0;
	return 1;
}

/* Returns TRUE if @protocol contains a valid value for this option,
 * or FALSE if the option was specified with an invalid value. On
 * error, errno is set. */
static int
nfs_mount_protocol(struct mount_options *options, unsigned long *protocol)
{
	sa_family_t family;
	char *option;

	option = po_get(options, "mountproto");
	if (option != NULL) {
		if (!nfs_get_proto(option, &family, protocol)) {
			errno = EPROTONOSUPPORT;
			bb_error_msg("Failed to find '%s' protocol", option);
			return 0;
		}
		return 1;
	}

	/* MNT transport protocol wasn't specified.  If the NFS
	 * transport protocol was specified, use that; otherwise
	 * set @protocol to zero.  The pmap protocol value will
	 * be filled in later by an rpcbind query in this case. */
	return nfs_nfs_protocol(options, protocol);
}

/* Returns TRUE if @port contains a valid value for this option,
 * or FALSE if the option was specified with an invalid value. */
static int
nfs_mount_port(struct mount_options *options, unsigned long *port)
{
	long tmp;

	switch (po_get_numeric(options, "mountport", &tmp)) {
	case PO_NOT_FOUND:
		break;
	case PO_FOUND:
		if (tmp >= 1 && tmp <= 65535) {
			*port = tmp;
			return 1;
		}
	case PO_BAD_VALUE:
		bb_error_msg("invalid value for 'mountport=' option");
		return 0;
	}

	/* MNT service port wasn't specified.  The pmap port value
	 * will be filled in later by an rpcbind query in this case. */
	*port = 0;
	return 1;
}

/* Returns TRUE and fills in @family if a valid MNT protocol option
 * is found, or FALSE if the option was specified with an invalid value
 * or if the protocol family isn't supported. On error, errno is set. */
static int nfs_mount_proto_family(struct mount_options *options,
				sa_family_t *family)
{
	unsigned long protocol;
	char *option;
	sa_family_t tmp_family = config_default_family;

	option = po_get(options, "mountproto");
	if (option != NULL) {
		if (!nfs_get_proto(option, &tmp_family, &protocol)) {
			bb_error_msg("Failed to find '%s' protocol", option);
			errno = EPROTONOSUPPORT;
			goto out_err;
		}
		if (!nfs_verify_family(tmp_family))
			goto out_err;
		*family = tmp_family;
		return 1;
	}

	/* MNT transport protocol wasn't specified.  If the NFS
	 * transport protocol was specified, derive the family
	 * from that; otherwise, return the default family for NFS. */
	return nfs_nfs_proto_family(options, family);
out_err:
	errno = EAFNOSUPPORT;
	return 0;
}

/* nfs_options2pmap - set up pmap structs based on mount options
 * @options: pointer to mount options
 * @nfs_pmap: OUT: pointer to pmap arguments for NFS server
 * @mnt_pmap: OUT: pointer to pmap arguments for mountd server
 *
 * Returns TRUE if the pmap options specified in @options have valid
 * values; otherwise FALSE is returned. */
static int nfs_options2pmap(struct mount_options *options,
		     struct pmap *nfs_pmap, struct pmap *mnt_pmap)
{
	if (!nfs_nfs_program(options, &nfs_pmap->pm_prog))
		return 0;
	if (!nfs_nfs_version(options, &nfs_pmap->pm_vers))
		return 0;
	if (!nfs_nfs_protocol(options, &nfs_pmap->pm_prot))
		return 0;
	if (!nfs_nfs_port(options, &nfs_pmap->pm_port))
		return 0;

	if (!nfs_mount_program(options, &mnt_pmap->pm_prog))
		return 0;
	if (!nfs_mount_version(options, &mnt_pmap->pm_vers))
		return 0;
	if (!nfs_mount_protocol(options, &mnt_pmap->pm_prot))
		return 0;
	if (!nfs_mount_port(options, &mnt_pmap->pm_port))
		return 0;

	return 1;
}



static int parse_sec(char *sec)
{
	int i, num_flavour = 0;

	for (sec = strtok(sec, ":"); sec; sec = strtok(NULL, ":")) {
		if (num_flavour >= MAX_USER_FLAVOUR) {
			bb_error_msg("maximum number of security flavors exceeded");
			return 0;
		}
		for (i = 0; i < flav_map_size; i++) {
			if (strcmp(sec, flav_map[i].flavour) == 0) {
				pseudoflavour[num_flavour++] = flav_map[i].fnum;
				break;
			}
		}
		if (i == flav_map_size) {
			bb_error_msg("unknown security type %s\n", sec);
			return 0;
		}
	}
	if (!num_flavour)
		bb_error_msg("no security flavors passed to sec= option");
	return num_flavour;
}

static int parse_devname(const char *spec, char **hostname, char **dir_name)
{
	char *s;

	if (strlen(spec) >= sizeof(hostdir)) {
		bb_error_msg("excessively long host:dir argument");
		return -1;
	}
	strcpy(hostdir, spec);

	if (!(s = strchr_set0(hostdir, ':'))) {
		bb_error_msg("directory to mount not in host:dir format");
		return -1;
	}
	*hostname = hostdir;
	*dir_name = s + 1;
	/* Ignore all but first hostname in replicated mounts
	   until they can be fully supported. (mack@sgi.com) */
	if (strchr_set0(hostdir, ','))
		bb_error_msg("warning: multiple hostnames not supported");
	return 0;
}

static int fill_ipv4_sockaddr(const char *hostname, struct sockaddr_in *addr)
{
	if(bb_gethostbyname(BB_GHBN_SOCKADDR_IN, hostname, addr) == NULL)
		return -1;
	return 0;
}

static int get_my_ipv4addr(void)
{
	char myname[1024];
	struct sockaddr_in myaddr;

	if (gethostname(myname, sizeof(myname))) {
		bb_error_msg("can't determine client address");
		return -1;
	}
	if (fill_ipv4_sockaddr(myname, &myaddr))
		return -1;
	safe_strncpy(ip_addr, inet_ntoa(myaddr.sin_addr), sizeof(ip_addr));
	return 0;
}

static int nfs4mount(const char *spec, const char *node, int flags,
	      char **extra_opts, int fake, int running_bg)
{
	struct nfs4_mount_data data;
	struct sockaddr_in server_addr, client_addr;
	int num_flavour = 0;
	int ip_addr_in_opts = 0;

	char *hostname, *dir_name, *old_opts;
	char new_opts[1024];
	char *opt, *opteq;
	char *s;
	int val;
	int bg, soft, intr;
	int nocto, noac, unshared;
	int retry;
	int retval = EX_FAIL;
	time_t timeout, t;

	memset(&server_addr, 0, sizeof(server_addr));
	memset(&client_addr, 0, sizeof(client_addr));
	if(ip_addr[0] == '\0')
		strcpy(ip_addr, "127.0.0.1");

	if (parse_devname(spec, &hostname, &dir_name))
		goto fail;

	if (fill_ipv4_sockaddr(hostname, &server_addr))
		goto fail;
	if (get_my_ipv4addr())
		goto fail;

	/* add IP address to mtab options for use when unmounting */
	s = inet_ntoa(server_addr.sin_addr);
	old_opts = *extra_opts;
	if (!old_opts)
		old_opts = "";
	if (strlen(old_opts) + strlen(s) + 10 >= sizeof(new_opts)) {
		bb_error_msg("excessively long option argument");
		goto fail;
	}
	if (running_bg)
		strncpy(new_opts, old_opts, sizeof(new_opts));
	else
		snprintf(new_opts, sizeof(new_opts), "%s%saddr=%s",
			 old_opts, *old_opts ? "," : "", s);
	*extra_opts = xstrdup(new_opts);

	/* Set default options.
	 * rsize/wsize and timeo are left 0 in order to
	 * let the kernel decide.
	 */
	memset(&data, 0, sizeof(data));
	data.retrans    = 3;
	data.acregmin   = 3;
	data.acregmax   = 60;
	data.acdirmin   = 30;
	data.acdirmax   = 60;
	data.proto      = IPPROTO_TCP;

	bg = 0;
	soft = 0;
	intr = NFS4_MOUNT_INTR;
	nocto = 0;
	noac = 0;
	unshared = 0;
	retry = -1;

	/*
	 * NFSv4 specifies that the default port should be 2049
	 */
	server_addr.sin_port = htons(NFS_PORT);

	/* parse options */

	for (opt = strtok(old_opts, ","); opt; opt = strtok(NULL, ",")) {
		if ((opteq = strchr_set0(opt, '='))) {
			val = atoi(opteq + 1);
			if (!strcmp(opt, "rsize"))
				data.rsize = val;
			else if (!strcmp(opt, "wsize"))
				data.wsize = val;
			else if (!strcmp(opt, "timeo"))
				data.timeo = val;
			else if (!strcmp(opt, "retrans"))
				data.retrans = val;
			else if (!strcmp(opt, "acregmin"))
				data.acregmin = val;
			else if (!strcmp(opt, "acregmax"))
				data.acregmax = val;
			else if (!strcmp(opt, "acdirmin"))
				data.acdirmin = val;
			else if (!strcmp(opt, "acdirmax"))
				data.acdirmax = val;
			else if (!strcmp(opt, "actimeo")) {
				data.acregmin = val;
				data.acregmax = val;
				data.acdirmin = val;
				data.acdirmax = val;
			}
			else if (!strcmp(opt, "retry"))
				retry = val;
			else if (!strcmp(opt, "port"))
				server_addr.sin_port = htons(val);
			else if (!strcmp(opt, "proto")) {
				if (!strncmp(opteq+1, "tcp", 3))
					data.proto = IPPROTO_TCP;
				else if (!strncmp(opteq+1, "udp", 3))
					data.proto = IPPROTO_UDP;
				else
					printf("Warning: Unrecognized proto= option.\n");
			} else if (!strcmp(opt, "clientaddr")) {
				if (strlen(opteq+1) >= sizeof(ip_addr))
					printf("Invalid client address %s",
								opteq+1);
				strncpy(ip_addr, opteq+1, sizeof(ip_addr));
				ip_addr[sizeof(ip_addr)-1] = '\0';
				ip_addr_in_opts = 1;
			} else if (!strcmp(opt, "sec")) {
				num_flavour = parse_sec(opteq+1);
				if (!num_flavour)
					goto fail;
			} else if (!strcmp(opt, "addr") || SLOPPY) {
				/* ignore */;
			} else {
				printf("unknown nfs mount parameter: "
					 "%s=%d\n", opt, val);
				goto fail;
			}
		} else {
			val = 1;
			if (!strncmp(opt, "no", 2)) {
				val = 0;
				opt += 2;
			}
			if (!strcmp(opt, "bg"))
				bg = val;
			else if (!strcmp(opt, "fg"))
				bg = !val;
			else if (!strcmp(opt, "soft"))
				soft = val;
			else if (!strcmp(opt, "hard"))
				soft = !val;
			else if (!strcmp(opt, "intr"))
				intr = val;
			else if (!strcmp(opt, "cto"))
				nocto = !val;
			else if (!strcmp(opt, "ac"))
				noac = !val;
			else if (!strcmp(opt, "sharecache"))
				unshared = !val;
			else if (!SLOPPY) {
				printf("unknown nfs mount option: %s%s\n",
						val ? "" : "no", opt);
				goto fail;
			}
		}
	}

	/* if retry is still -1, then it wasn't set via an option */
	if (retry == -1) {
		if (bg)
			retry = 10000;  /* 10000 mins == ~1 week */
		else
			retry = 2;      /* 2 min default on fg mounts */
	}

	data.flags = (soft ? NFS4_MOUNT_SOFT : 0)
		| (intr ? NFS4_MOUNT_INTR : 0)
		| (nocto ? NFS4_MOUNT_NOCTO : 0)
		| (noac ? NFS4_MOUNT_NOAC : 0)
		| (unshared ? NFS4_MOUNT_UNSHARED : 0);

	data.auth_flavourlen = num_flavour;
	data.auth_flavours = pseudoflavour;

	data.client_addr.data = ip_addr;
	data.client_addr.len = strlen(ip_addr);

	data.mnt_path.data = dir_name;
	data.mnt_path.len = strlen(dir_name);

	data.hostname.data = hostname;
	data.hostname.len = strlen(hostname);
	data.host_addr = (struct sockaddr *)&server_addr;
	data.host_addrlen = sizeof(server_addr);

	timeout = time(NULL) + 60 * retry;
	data.version = NFS4_MOUNT_VERSION;
	for (;;) {
		if (verbose) {
			printf("%s: pinging: prog %lu vers %d prot %s port %d\n",
				bb_applet_name, NFS_PROGRAM, 4,
				data.proto == IPPROTO_UDP ? "udp" : "tcp",
				ntohs(server_addr.sin_port));
		}
		client_addr.sin_family = 0;
		client_addr.sin_addr.s_addr = 0;
		clnt_ping(&server_addr, NFS_PROGRAM, 4, data.proto, &client_addr);
		if (rpc_createerr.cf_stat == RPC_SUCCESS) {
			if (!ip_addr_in_opts &&
			    client_addr.sin_family != 0 &&
			    client_addr.sin_addr.s_addr != 0) {
				snprintf(ip_addr, sizeof(ip_addr), "%s",
					 inet_ntoa(client_addr.sin_addr));
				data.client_addr.len = strlen(ip_addr);
			}
			break;
		}

		if (!bg) {
			switch(rpc_createerr.cf_stat) {
			case RPC_TIMEDOUT:
				break;
			case RPC_SYSTEMERROR:
				if (errno == ETIMEDOUT)
					break;
			default:
				rpc_mount_errors(hostname, 0, bg);
				goto fail;
			}
		}

		if (bg && !running_bg) {
			if (retry > 0)
				retval = EX_BG;
			goto fail;
		}

		t = time(NULL);
		if (t >= timeout) {
			rpc_mount_errors(hostname, 0, bg);
			goto fail;
		}
		rpc_mount_errors(hostname, 1, bg);
		continue;
	}

	if (!fake) {
		if (mount(spec, node, "nfs4",
				flags & ~(MS_USER|MS_USERS), &data)) {
			mount_error(spec, node);
			goto fail;
		}
	}

	return EX_SUCCESS;

fail:
	return retval;
}


typedef dirpath mnt2arg_t;
typedef dirpath mnt3arg_t;
typedef dirpath mntarg_t;

typedef struct fhstatus  mnt2res_t;
typedef struct mountres3 mnt3res_t;
typedef union {
	mnt2res_t nfsv2;
	mnt3res_t nfsv3;
} mntres_t;



static bool_t
xdr_fhandle (XDR *xdrs, fhandle objp)
{
	if (!xdr_opaque (xdrs, objp, FHSIZE))
		 return FALSE;
	return TRUE;
}

static bool_t
xdr_fhstatus (XDR *xdrs, fhstatus *objp)
{

	if (!xdr_u_int (xdrs, &objp->fhs_status))
		 return FALSE;
	switch (objp->fhs_status) {
	case 0:
		if (!xdr_fhandle (xdrs, objp->fhstatus_u.fhs_fhandle))
			 return FALSE;
		break;
	default:
		break;
	}
	return TRUE;
}

static bool_t
xdr_dirpath (XDR *xdrs, dirpath *objp)
{

	if (!xdr_string (xdrs, objp, MNTPATHLEN))
		 return FALSE;
	return TRUE;
}

static inline enum clnt_stat
nfs2_mount(CLIENT *clnt, mnt2arg_t *mnt2arg, mnt2res_t *mnt2res)
{
	return clnt_call(clnt, MOUNTPROC_MNT,
			 (xdrproc_t) xdr_dirpath, (caddr_t) mnt2arg,
			 (xdrproc_t) xdr_fhstatus, (caddr_t) mnt2res,
			 TIMEOUT);
}

static bool_t
xdr_mountstat3 (XDR *xdrs, mountstat3 *objp)
{
	if (!xdr_enum (xdrs, (enum_t *) objp))
		return FALSE;
	return TRUE;
}

static bool_t
xdr_fhandle3 (XDR *xdrs, fhandle3 *objp)
{
	if (!xdr_bytes (xdrs, (char **)&objp->fhandle3_val, (u_int *) &objp->fhandle3_len, FHSIZE3))
		return FALSE;
	return TRUE;
}

static bool_t
xdr_mountres3_ok (XDR *xdrs, mountres3_ok *objp)
{
	if (!xdr_fhandle3 (xdrs, &objp->fhandle))
		return FALSE;
	if (!xdr_array (xdrs, (char **)&objp->auth_flavors.auth_flavors_val,
			(u_int *) &objp->auth_flavors.auth_flavors_len, ~0,
			sizeof (int), (xdrproc_t) xdr_int))
		return FALSE;
	return TRUE;
}

static bool_t xdr_mountres3 (XDR *xdrs, mountres3 *objp)
{
	if (!xdr_mountstat3 (xdrs, &objp->fhs_status))
		return FALSE;
	switch (objp->fhs_status) {
	case MNT_OK:
		if (!xdr_mountres3_ok (xdrs, &objp->mountres3_u.mountinfo))
			return FALSE;
		break;
	default:
		break;
	}
	return TRUE;
}

static inline enum clnt_stat
nfs3_mount(CLIENT *clnt, mnt3arg_t *mnt3arg, mnt3res_t *mnt3res)
{
	return clnt_call(clnt, MOUNTPROC3_MNT,
			 (xdrproc_t) xdr_dirpath, (caddr_t) mnt3arg,
			 (xdrproc_t) xdr_mountres3, (caddr_t) mnt3res,
			 TIMEOUT);
}

static int
nfs_call_mount(clnt_addr_t *mnt_server, clnt_addr_t *nfs_server,
	       mntarg_t *mntarg, mntres_t *mntres)
{
	CLIENT *clnt;
	enum clnt_stat cf_stat;
	int msock;

	if (!probe_bothports(mnt_server, nfs_server))
		goto out_bad;

	clnt = mnt_openclnt(mnt_server, &msock);
	if (!clnt)
		goto out_bad;
	/* make pointers in xdr_mountres3 NULL so
	 * that xdr_array allocates memory for us
	 */
	memset(mntres, 0, sizeof(*mntres));
	switch (mnt_server->pmap.pm_vers) {
	case 3:
		cf_stat = nfs3_mount(clnt, mntarg, &mntres->nfsv3);
		break;
	case 2:
	case 1:
		cf_stat = nfs2_mount(clnt, mntarg, &mntres->nfsv2);
		break;
	default:
		goto out_bad;
	}
	if (cf_stat != RPC_SUCCESS) {
		clnt_geterr(clnt, &rpc_createerr.cf_error);
		rpc_createerr.cf_stat = cf_stat;
	}
	mnt_closeclnt(clnt, msock);
	if (cf_stat == RPC_SUCCESS)
		return 1;
 out_bad:
	return 0;
}

static int
parse_options(char *old_opts, struct nfs_mount_data *data,
	      int *bg, int *retry, clnt_addr_t *mnt_server,
	      clnt_addr_t *nfs_server, char *new_opts, const int opt_size)
{
	struct sockaddr_in *mnt_saddr = &mnt_server->saddr;
	struct pmap *mnt_pmap = &mnt_server->pmap;
	struct pmap *nfs_pmap = &nfs_server->pmap;
	int len;
	char *opt, *opteq, *p, *opt_b, *tmp_opts;
	char *mounthost = NULL;
	char cbuf[128];
	int open_quote = 0;

	data->flags = 0;
	*bg = 0;

	len = strlen(new_opts);
	tmp_opts = xstrdup(old_opts);
	for (p=tmp_opts, opt_b=NULL; p && *p; p++) {
		if (!opt_b)
			opt_b = p;              /* begin of the option item */
		if (*p == '"')
			open_quote ^= 1;        /* reverse the status */
		if (open_quote)
			continue;               /* still in a quoted block */
		if (*p == ',')
			*p = '\0';              /* terminate the option item */
		if (*p == '\0' || *(p+1) == '\0') {
			opt = opt_b;            /* opt is useful now */
			opt_b = NULL;
		}
		else
			continue;               /* still somewhere in the option item */

		if (strlen(opt) >= sizeof(cbuf))
			goto bad_parameter;
		if ((opteq = strchr_set0(opt, '=')) && isdigit(opteq[1])) {
			int val = atoi(opteq + 1);
			if (!strcmp(opt, "rsize"))
				data->rsize = val;
			else if (!strcmp(opt, "wsize"))
				data->wsize = val;
			else if (!strcmp(opt, "timeo"))
				data->timeo = val;
			else if (!strcmp(opt, "retrans"))
				data->retrans = val;
			else if (!strcmp(opt, "acregmin"))
				data->acregmin = val;
			else if (!strcmp(opt, "acregmax"))
				data->acregmax = val;
			else if (!strcmp(opt, "acdirmin"))
				data->acdirmin = val;
			else if (!strcmp(opt, "acdirmax"))
				data->acdirmax = val;
			else if (!strcmp(opt, "actimeo")) {
				data->acregmin = val;
				data->acregmax = val;
				data->acdirmin = val;
				data->acdirmax = val;
			}
			else if (!strcmp(opt, "retry"))
				*retry = val;
			else if (!strcmp(opt, "port"))
				nfs_pmap->pm_port = val;
			else if (!strcmp(opt, "mountport"))
				mnt_pmap->pm_port = val;
			else if (!strcmp(opt, "mountprog"))
				mnt_pmap->pm_prog = val;
			else if (!strcmp(opt, "mountvers"))
				mnt_pmap->pm_vers = val;
			else if (!strcmp(opt, "mounthost"))
				mounthost=xstrndup(opteq+1, strcspn(opteq+1," \t\n\r,"));
			else if (!strcmp(opt, "nfsprog"))
				nfs_pmap->pm_prog = val;
			else if (!strcmp(opt, "nfsvers") ||
				 !strcmp(opt, "vers")) {
				nfs_pmap->pm_vers = val;
				opt = "nfsvers";
#if NFS_MOUNT_VERSION >= 2
			} else if (!strcmp(opt, "namlen")) {
				if (nfs_mount_data_version >= 2)
					data->namlen = val;
				else if (SLOPPY)
					continue;
				else
					goto bad_parameter;
#endif
			} else if (!strcmp(opt, "addr")) {
				/* ignore */;
				continue;
			} else if (SLOPPY)
				continue;
			else
				goto bad_parameter;
			sprintf(cbuf, "%s=%s,", opt, opteq+1);
		} else if (opteq) {
			*opteq = '\0';
			if (!strcmp(opt, "proto")) {
				if (!strcmp(opteq+1, "udp")) {
					nfs_pmap->pm_prot = IPPROTO_UDP;
					mnt_pmap->pm_prot = IPPROTO_UDP;
#if NFS_MOUNT_VERSION >= 2
					data->flags &= ~NFS_MOUNT_TCP;
				} else if (!strcmp(opteq+1, "tcp") &&
					   nfs_mount_data_version > 2) {
					nfs_pmap->pm_prot = IPPROTO_TCP;
					mnt_pmap->pm_prot = IPPROTO_TCP;
					data->flags |= NFS_MOUNT_TCP;
#endif
				} else if (SLOPPY)
					continue;
				else
					goto bad_parameter;
#if NFS_MOUNT_VERSION >= 5
			} else if (!strcmp(opt, "sec")) {
				char *secflavor = opteq+1;
				/* see RFC 2623 */
				if (nfs_mount_data_version < 5) {
					printf("Warning: ignoring sec=%s option\n",
							secflavor);
					continue;
				} else if (!strcmp(secflavor, "none"))
					data->pseudoflavor = AUTH_NONE;
				else if (!strcmp(secflavor, "sys"))
					data->pseudoflavor = AUTH_SYS;
				else if (!strcmp(secflavor, "krb5"))
					data->pseudoflavor = RPC_AUTH_GSS_KRB5;
				else if (!strcmp(secflavor, "krb5i"))
					data->pseudoflavor = RPC_AUTH_GSS_KRB5I;
				else if (!strcmp(secflavor, "krb5p"))
					data->pseudoflavor = RPC_AUTH_GSS_KRB5P;
				else if (!strcmp(secflavor, "lipkey"))
					data->pseudoflavor = RPC_AUTH_GSS_LKEY;
				else if (!strcmp(secflavor, "lipkey-i"))
					data->pseudoflavor = RPC_AUTH_GSS_LKEYI;
				else if (!strcmp(secflavor, "lipkey-p"))
					data->pseudoflavor = RPC_AUTH_GSS_LKEYP;
				else if (!strcmp(secflavor, "spkm3"))
					data->pseudoflavor = RPC_AUTH_GSS_SPKM;
				else if (!strcmp(secflavor, "spkm3i"))
					data->pseudoflavor = RPC_AUTH_GSS_SPKMI;
				else if (!strcmp(secflavor, "spkm3p"))
					data->pseudoflavor = RPC_AUTH_GSS_SPKMP;
				else if (SLOPPY)
					continue;
				else {
					printf("Warning: Unrecognized security flavor %s.\n",
						secflavor);
					goto bad_parameter;
				}
				data->flags |= NFS_MOUNT_SECFLAVOUR;
#endif
			} else if (!strcmp(opt, "mounthost"))
				mounthost=xstrndup(opteq+1,
						   strcspn(opteq+1," \t\n\r,"));
			 else if (!strcmp(opt, "context")) {
				char *context = opteq + 1;
				int ctxlen = strlen(context);

				if (ctxlen > NFS_MAX_CONTEXT_LEN) {
					bb_error_msg("context parameter exceeds"
							" limit of %d",
							NFS_MAX_CONTEXT_LEN);
					goto bad_parameter;
				}
				/* The context string is in the format of
				 * "system_u:object_r:...".  We only want
				 * the context str between the quotes.
				 */
				if (*context == '"')
					strncpy(data->context, context+1,
							ctxlen-2);
				else
					strncpy(data->context, context,
							NFS_MAX_CONTEXT_LEN);
			} else if (SLOPPY)
				continue;
			else
				goto bad_parameter;
			sprintf(cbuf, "%s=%s,", opt, opteq+1);
		} else {
			int val = 1;
			if (!strncmp(opt, "no", 2)) {
				val = 0;
				opt += 2;
			}
			if (!strcmp(opt, "bg"))
				*bg = val;
			else if (!strcmp(opt, "fg"))
				*bg = !val;
			else if (!strcmp(opt, "soft")) {
				data->flags &= ~NFS_MOUNT_SOFT;
				if (val)
					data->flags |= NFS_MOUNT_SOFT;
			} else if (!strcmp(opt, "hard")) {
				data->flags &= ~NFS_MOUNT_SOFT;
				if (!val)
					data->flags |= NFS_MOUNT_SOFT;
			} else if (!strcmp(opt, "intr")) {
				data->flags &= ~NFS_MOUNT_INTR;
				if (val)
					data->flags |= NFS_MOUNT_INTR;
			} else if (!strcmp(opt, "posix")) {
				data->flags &= ~NFS_MOUNT_POSIX;
				if (val)
					data->flags |= NFS_MOUNT_POSIX;
			} else if (!strcmp(opt, "cto")) {
				data->flags &= ~NFS_MOUNT_NOCTO;
				if (!val)
					data->flags |= NFS_MOUNT_NOCTO;
			} else if (!strcmp(opt, "ac")) {
				data->flags &= ~NFS_MOUNT_NOAC;
				if (!val)
					data->flags |= NFS_MOUNT_NOAC;
#if NFS_MOUNT_VERSION >= 2
			} else if (!strcmp(opt, "tcp")) {
				data->flags &= ~NFS_MOUNT_TCP;
				if (val) {
					if (nfs_mount_data_version < 2)
						goto bad_option;
					nfs_pmap->pm_prot = IPPROTO_TCP;
					mnt_pmap->pm_prot = IPPROTO_TCP;
					data->flags |= NFS_MOUNT_TCP;
				} else {
					mnt_pmap->pm_prot = IPPROTO_UDP;
					nfs_pmap->pm_prot = IPPROTO_UDP;
				}
			} else if (!strcmp(opt, "udp")) {
				data->flags &= ~NFS_MOUNT_TCP;
				if (!val) {
					if (nfs_mount_data_version < 2)
						goto bad_option;
					nfs_pmap->pm_prot = IPPROTO_TCP;
					mnt_pmap->pm_prot = IPPROTO_TCP;
					data->flags |= NFS_MOUNT_TCP;
				} else {
					nfs_pmap->pm_prot = IPPROTO_UDP;
					mnt_pmap->pm_prot = IPPROTO_UDP;
				}
#endif
#if NFS_MOUNT_VERSION >= 3
			} else if (!strcmp(opt, "lock")) {
				data->flags &= ~NFS_MOUNT_NONLM;
				if (!val) {
					if (nfs_mount_data_version < 3)
						goto bad_option;
					data->flags |= NFS_MOUNT_NONLM;
				}
#endif
#if NFS_MOUNT_VERSION >= 4
			} else if (!strcmp(opt, "broken_suid")) {
				data->flags &= ~NFS_MOUNT_BROKEN_SUID;
				if (val) {
					if (nfs_mount_data_version < 4)
						goto bad_option;
					data->flags |= NFS_MOUNT_BROKEN_SUID;
				}
			} else if (!strcmp(opt, "acl")) {
				data->flags &= ~NFS_MOUNT_NOACL;
				if (!val)
					data->flags |= NFS_MOUNT_NOACL;
			} else if (!strcmp(opt, "rdirplus")) {
				data->flags &= ~NFS_MOUNT_NORDIRPLUS;
				if (!val)
					data->flags |= NFS_MOUNT_NORDIRPLUS;
			} else if (!strcmp(opt, "sharecache")) {
				data->flags &= ~NFS_MOUNT_UNSHARED;
				if (!val)
					data->flags |= NFS_MOUNT_UNSHARED;
#endif
			} else {
			bad_option:
				if (SLOPPY)
					continue;
				bb_error_msg("Unsupported nfs mount option:"
						" %s%s", val ? "" : "no", opt);
				goto out_bad;
			}
			sprintf(cbuf, val ? "%s," : "no%s,", opt);
		}
		len += strlen(cbuf);
		if (len >= opt_size) {
			bb_error_msg("excessively long option argument");
			goto out_bad;
		}
		strcat(new_opts, cbuf);
	}
	/* See if the nfs host = mount host. */
	if (mounthost) {
		if (!nfs_gethostbyname(mounthost, mnt_saddr))
			goto out_bad;
		*mnt_server->hostname = mounthost;
	}
	free(tmp_opts);
	return 1;
 bad_parameter:
	bb_error_msg("Bad nfs mount parameter: %s", opt);
 out_bad:
	free(tmp_opts);
	return 0;
}

static int nfsmnt_check_compat(const struct pmap *nfs_pmap,
				const struct pmap *mnt_pmap)
{
	unsigned int max_nfs_vers = (nfs_mount_data_version >= 4) ? 3 : 2;
	unsigned int max_mnt_vers = (nfs_mount_data_version >= 4) ? 3 : 2;

	if (nfs_pmap->pm_vers == 4) {
		bb_error_msg("Please use '-t nfs4' instead of '-o vers=4'");
		goto out_bad;
	}

	if (nfs_pmap->pm_vers) {
		if (nfs_pmap->pm_vers > max_nfs_vers || nfs_pmap->pm_vers < 2) {
			bb_error_msg("NFS version %ld is not supported", nfs_pmap->pm_vers);
			goto out_bad;
		}
	}

	if (mnt_pmap->pm_vers > max_mnt_vers) {
		bb_error_msg("NFS mount version %ld is not supported", mnt_pmap->pm_vers);
		goto out_bad;
	}

	return 1;

out_bad:
	return 0;
}

/* nfs_call_umount - ask the server to remove a share from it's rmtab
 * @mnt_server: address of RPC MNT program server
 * @argp: directory path of share to "unmount"
 *
 * Returns one if the unmount call succeeded; zero if the unmount
 * failed for any reason.
 *
 * Note that a side effect of calling this function is that rpccreateerr
 * is set. */
static int nfs_call_umount(clnt_addr_t *mnt_server, dirpath *argp)
{
	struct sockaddr *sap = (struct sockaddr *)&mnt_server->saddr;
	socklen_t salen = sizeof(mnt_server->saddr);
	struct pmap *pmap = &mnt_server->pmap;
	CLIENT *clnt;
	enum clnt_stat result;
	int msock;

	if (!nfs_probe_mntport(sap, salen, pmap))
		return 0;
	clnt = mnt_openclnt(mnt_server, &msock);
	if (!clnt)
		return 0;
	result = clnt_call(clnt, MOUNTPROC_UMNT,
			(xdrproc_t)xdr_dirpath, (caddr_t)argp,
			(xdrproc_t)xdr_void, NULL,
			TIMEOUT);
	mnt_closeclnt(clnt, msock);

	if (result == RPC_SUCCESS)
		return 1;
	return 0;
}

static int
nfsmount(const char *spec, const char *node, int flags,
	 char **extra_opts, int fake, int running_bg)
{
	char *hostname, *dir_name, *old_opts, *mounthost = NULL;
	char new_opts[1024], cbuf[1024];
	struct nfs_mount_data data;
	int val;
	int doonce = 0;

	clnt_addr_t mnt_server = {
		.hostname = &mounthost
	};
	clnt_addr_t nfs_server = {
		.hostname = &hostname
	};
	struct sockaddr_in *nfs_saddr = &nfs_server.saddr;
	struct pmap  *mnt_pmap = &mnt_server.pmap,
		     *nfs_pmap = &nfs_server.pmap;
	struct pmap  save_mnt, save_nfs;

	int fsock = -1;

	mntres_t mntres;

	struct stat statbuf;
	char *s;
	int bg, retry;
	int retval = EX_FAIL;
	time_t t;
	time_t prevt;
	time_t timeout;

	if (parse_devname(spec, &hostname, &dir_name))
		goto fail;

	if (!nfs_gethostbyname(hostname, nfs_saddr))
		goto fail;
	mounthost = hostname;
	memcpy (&mnt_server.saddr, nfs_saddr, sizeof (mnt_server.saddr));

	/* add IP address to mtab options for use when unmounting */

	s = inet_ntoa(nfs_saddr->sin_addr);
	old_opts = *extra_opts;
	if (!old_opts)
		old_opts = "";

	/* Set default options.
	 * rsize/wsize (and bsize, for ver >= 3) are left 0 in order to
	 * let the kernel decide.
	 * timeo is filled in after we know whether it'll be TCP or UDP. */
	memset(&data, 0, sizeof(data));
	data.acregmin   = 3;
	data.acregmax   = 60;
	data.acdirmin   = 30;
	data.acdirmax   = 60;
#if NFS_MOUNT_VERSION >= 2
	data.namlen     = NAME_MAX;
#endif

	bg = 0;
	retry = -1;

	memset(mnt_pmap, 0, sizeof(*mnt_pmap));
	mnt_pmap->pm_prog = MOUNTPROG;
	memset(nfs_pmap, 0, sizeof(*nfs_pmap));
	nfs_pmap->pm_prog = NFS_PROGRAM;

	/* parse options */
	new_opts[0] = 0;
	if (!parse_options(old_opts, &data, &bg, &retry, &mnt_server, &nfs_server,
			   new_opts, sizeof(new_opts)))
		goto fail;
	if (!nfsmnt_check_compat(nfs_pmap, mnt_pmap))
		goto fail;

	if (retry == -1) {
		if (bg)
			retry = 10000;  /* 10000 mins == ~1 week*/
		else
			retry = 2;      /* 2 min default on fg mounts */
	}

#ifdef NFS_MOUNT_DEBUG
	printf("rsize = %d, wsize = %d, timeo = %d, retrans = %d\n",
	       data.rsize, data.wsize, data.timeo, data.retrans);
	printf("acreg (min, max) = (%d, %d), acdir (min, max) = (%d, %d)\n",
	       data.acregmin, data.acregmax, data.acdirmin, data.acdirmax);
	printf("port = %lu, bg = %d, retry = %d, flags = %.8x\n",
	       nfs_pmap->pm_port, bg, retry, data.flags);
	printf("mountprog = %lu, mountvers = %lu, nfsprog = %lu, nfsvers = %lu\n",
	       mnt_pmap->pm_prog, mnt_pmap->pm_vers,
	       nfs_pmap->pm_prog, nfs_pmap->pm_vers);
	printf("soft = %d, intr = %d, posix = %d, nocto = %d, noac = %d",
	       (data.flags & NFS_MOUNT_SOFT) != 0,
	       (data.flags & NFS_MOUNT_INTR) != 0,
	       (data.flags & NFS_MOUNT_POSIX) != 0,
	       (data.flags & NFS_MOUNT_NOCTO) != 0,
	       (data.flags & NFS_MOUNT_NOAC) != 0);
#if NFS_MOUNT_VERSION >= 2
	printf(", tcp = %d",
	       (data.flags & NFS_MOUNT_TCP) != 0);
#endif
#if NFS_MOUNT_VERSION >= 4
	printf(", noacl = %d", (data.flags & NFS_MOUNT_NOACL) != 0);
#endif
#if NFS_MOUNT_VERSION >= 5
	printf(", sec = %u", data.pseudoflavor);
	printf(", readdirplus = %d", (data.flags & NFS_MOUNT_NORDIRPLUS) != 0);
#endif
	printf("\n");
#endif

	data.version = nfs_mount_data_version;

	if (flags & MS_REMOUNT)
		goto out_ok;

	/* create mount deamon client */

	/*
	 * The following loop implements the mount retries. On the first
	 * call, "running_bg" is 0. When the mount times out, and the
	 * "bg" option is set, the exit status EX_BG will be returned.
	 * For a backgrounded mount, there will be a second call by the
	 * child process with "running_bg" set to 1.
	 *
	 * The case where the mount point is not present and the "bg"
	 * option is set, is treated as a timeout. This is done to
	 * support nested mounts.
	 *
	 * The "retry" count specified by the user is the number of
	 * minutes to retry before giving up.
	 *
	 * Only the first error message will be displayed.
	 */
	timeout = time(NULL) + 60 * retry;
	prevt = 0;
	t = 30;
	val = 1;

	memcpy(&save_nfs, nfs_pmap, sizeof(save_nfs));
	memcpy(&save_mnt, mnt_pmap, sizeof(save_mnt));
	for (;;) {
		if (bg && stat(node, &statbuf) == -1) {
			/* no mount point yet - sleep */
			if (running_bg) {
				sleep(val);     /* 1, 2, 4, 8, 16, 30, ... */
				val *= 2;
				if (val > 30)
					val = 30;
			}
		} else {
			int result;
			/* be careful not to use too many CPU cycles */
			if (t - prevt < 30)
				sleep(30);

			result = nfs_call_mount(&mnt_server, &nfs_server,
					      &dir_name, &mntres);
			if (result)
				break;
			memcpy(nfs_pmap, &save_nfs, sizeof(*nfs_pmap));
			memcpy(mnt_pmap, &save_mnt, sizeof(*mnt_pmap));
			prevt = t;
		}
		if (!bg) {
			switch(rpc_createerr.cf_stat){
			case RPC_TIMEDOUT:
				break;
			case RPC_SYSTEMERROR:
				if (errno == ETIMEDOUT)
					break;
			default:
				rpc_mount_errors(*nfs_server.hostname, 0, bg);
			goto fail;
			}
			t = time(NULL);
			if (t >= timeout) {
				rpc_mount_errors(*nfs_server.hostname, 0, bg);
				goto fail;
			}
			rpc_mount_errors(*nfs_server.hostname, 1, bg);
			continue;
		}
		if (!running_bg) {
			if (retry > 0)
				retval = EX_BG;
			goto fail;
		}
		t = time(NULL);
		if (t >= timeout) {
			rpc_mount_errors(*nfs_server.hostname, 0, bg);
			goto fail;
		}
		if (doonce++ < 1)
			rpc_mount_errors(*nfs_server.hostname, 1, bg);
	}

	if (mnt_pmap->pm_vers <= 2) {
		if (mntres.nfsv2.fhs_status != 0) {
			char buf[256];
			bb_error_msg("%s:%s failed, reason given by server: %s",
					hostname, dir_name,
					nfs_strerror(mntres.nfsv2.fhs_status, buf));
			goto fail;
		}
		memcpy(data.root.data,
		       (char *) mntres.nfsv2.fhstatus_u.fhs_fhandle,
		       NFS_FHSIZE);
#if NFS_MOUNT_VERSION >= 4
		data.root.size = NFS_FHSIZE;
		memcpy(data.old_root.data,
		       (char *) mntres.nfsv2.fhstatus_u.fhs_fhandle,
		       NFS_FHSIZE);
#endif
	} else {
#if NFS_MOUNT_VERSION >= 4
		mountres3_ok *mountres;
		fhandle3 *f_handle;
		int i,  n_flavors, *flavor, yum = 0;
		char buf[256];

		if (mntres.nfsv3.fhs_status != 0) {
			bb_error_msg("%s:%s failed, reason given by server: %s",
					hostname, dir_name,
					nfs_strerror(mntres.nfsv3.fhs_status, buf));
			goto fail;
		}
#if NFS_MOUNT_VERSION >= 5
		mountres = &mntres.nfsv3.mountres3_u.mountinfo;
		n_flavors = mountres->auth_flavors.auth_flavors_len;
		if (n_flavors <= 0)
			goto noauth_flavors;

		flavor = mountres->auth_flavors.auth_flavors_val;
		for (i = 0; i < n_flavors; ++i) {
			/*
			 * Per RFC2623, section 2.7, we should prefer the
			 * flavour listed first.
			 * If no flavour requested, use the first simple
			 * flavour that is offered.
			 */
			if (! (data.flags & NFS_MOUNT_SECFLAVOUR) &&
			    (flavor[i] == AUTH_SYS ||
			     flavor[i] == AUTH_NONE)) {
				data.pseudoflavor = flavor[i];
				data.flags |= NFS_MOUNT_SECFLAVOUR;
			}
			if (flavor[i] == data.pseudoflavor)
				yum = 1;
#ifdef NFS_MOUNT_DEBUG
			printf("auth flavor %d: %d\n", i, flavor[i]);
#endif
		}
		if (!yum) {
			bb_error_msg("%s:%s failed, security flavor "
					"not supported", hostname, dir_name);
			/* server has registered us in rmtab, send umount */
			nfs_call_umount(&mnt_server, &dir_name);
			goto fail;
		}
noauth_flavors:
#endif
		f_handle = &mntres.nfsv3.mountres3_u.mountinfo.fhandle;
		memset(data.old_root.data, 0, NFS_FHSIZE);
		memset(&data.root, 0, sizeof(data.root));
		data.root.size = f_handle->fhandle3_len;
		memcpy(data.root.data,
		       (char *) f_handle->fhandle3_val,
		       f_handle->fhandle3_len);

		data.flags |= NFS_MOUNT_VER3;
#endif
	}

	if (nfs_mount_data_version == 1) {
		/* create nfs socket for kernel */
		if (nfs_pmap->pm_prot == IPPROTO_TCP)
			fsock = bb_xsocket(0);
		else
			fsock = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM);
		if (fsock < 0)
			goto fail;
		if (bindresvport(fsock, 0) < 0) {
			bb_perror_msg("nfs bindresvport");
			goto fail;
		}
	}

#ifdef NFS_MOUNT_DEBUG
	printf("using port %lu for nfs deamon\n", nfs_pmap->pm_port);
#endif
	nfs_saddr->sin_port = htons(nfs_pmap->pm_port);
	/*
	 * connect() the socket for kernels 1.3.10 and below only,
	 * to avoid problems with multihomed hosts.
	 * --Swen
	 */
	if (linux_version_code() <= MAKE_VERSION(1, 3, 10) && fsock != -1
	    && connect(fsock, (struct sockaddr *) nfs_saddr,
		       sizeof (*nfs_saddr)) < 0) {
		bb_perror_msg("nfs connect");
		goto fail;
	}

#if NFS_MOUNT_VERSION >= 2
	if (nfs_pmap->pm_prot == IPPROTO_TCP)
		data.flags |= NFS_MOUNT_TCP;
	else
		data.flags &= ~NFS_MOUNT_TCP;
#endif

	/* prepare data structure for kernel */

	data.fd = fsock;
	memcpy((char *) &data.addr, (char *) nfs_saddr, sizeof(data.addr));
	strncpy(data.hostname, hostname, sizeof(data.hostname));

 out_ok:
	/* Ensure we have enough padding for the following strcat()s */
	if (strlen(new_opts) + strlen(s) + 30 >= sizeof(new_opts)) {
		bb_error_msg("excessively long option argument");
		goto fail;
	}

	snprintf(cbuf, sizeof(cbuf)-1, "addr=%s", s);
	strcat(new_opts, cbuf);

	*extra_opts = xstrdup(new_opts);

	if (!fake && !(data.flags & NFS_MOUNT_NONLM)) {
		if (!start_statd()) {
			bb_error_msg("rpc.statd is not running but is "
				"required for remote locking.\n"
				"   Either use '-o nolock' to keep "
				"locks local, or start statd.");
			goto fail;
		}
	}

	if (!fake) {
		if (mount(spec, node, "nfs",
				flags & ~(MS_USER|MS_USERS), &data)) {
			mount_error(spec, node);
			goto fail;
		}
	}

	return EX_SUCCESS;

	/* abort */
 fail:
	if (fsock != -1)
		close(fsock);
	return retval;
}

#if !defined(MNT_FORCE)
/* dare not try to include <linux/mount.h> -- lots of errors */
#define MNT_FORCE 1
#endif

#if !defined(MNT_DETACH)
#define MNT_DETACH 2
#endif

#undef FORCE
#define FORCE   (bss_flags & 1)
#define LAZY    (bss_flags & 2)
#define REMOUNT (bss_flags & 4)


static int try_remount(const char *spec, const char *node)
{
	int result;

	result = mount(spec, node, NULL,
		    MS_MGC_VAL | MS_REMOUNT | MS_RDONLY, NULL);
	if (result == 0) {
		struct mntent remnt;
		bb_error_msg("%s busy - remounted read-only", spec);
		remnt.mnt_type = remnt.mnt_fsname = NULL;
		remnt.mnt_dir = xstrdup(node);
		remnt.mnt_opts = xstrdup("ro");
		if (!NOMTAB)
			update_mtab(node, &remnt);
	} else if (errno != EBUSY) {    /* hmm ... */
		bb_perror_msg("remount: could not remount %s read-only", spec);
	}
	return result;
}

static int del_mtab(const char *spec, const char *node)
{
	int umnt_err, result;

	umnt_err = 0;
	if (LAZY) {
		result = umount2 (node, MNT_DETACH);
		if (result < 0)
			umnt_err = errno;
		goto writemtab;
	}

	if (FORCE) {
		result = umount2 (node, MNT_FORCE);
		if (result == -1) {
			int errsv = errno;
			bb_perror_msg("umount2");
			errno = errsv;
			if (errno == ENOSYS) {
				if (verbose)
					printf("no umount2, trying umount...\n");
				result = umount (node);
			}
		}
	} else
		result = umount (node);

	if (result < 0) {
		if (REMOUNT && errno == EBUSY && spec) {
			result = try_remount(spec, node);
			if (result)
				goto writemtab;
			return 0;
		} else
			umnt_err = errno;
	}

	if (result >= 0) {
		/* Umount succeeded */
		if (verbose)
			printf("%s umounted\n", spec ? spec : node);
	}

 writemtab:
	if (!NOMTAB &&
	    (umnt_err == 0 || umnt_err == EINVAL || umnt_err == ENOENT)) {
		update_mtab(node, NULL);
	}

	if (result >= 0)
		return 0;

	if (umnt_err) {
		errno = umnt_err;
		umount_error(node);
	}
	return EX_FILEIO;
}

/*
 * Discover mount server's hostname/address by examining mount options
 *
 * Returns a pointer to a string that the caller must free, on
 * success; otherwise NULL is returned.
 */
static char *nfs_umount_hostname(struct mount_options *options,
				 char *hostname)
{
	char *option;

	option = po_get(options, "mountaddr");
	if (option)
		goto out;
	option = po_get(options, "mounthost");
	if (option)
		goto out;
	option = po_get(options, "addr");
	if (option)
		goto out;

	return hostname;

out:
	free(hostname);
	return xstrdup(option);
}

/* nfs_advise_umount - ask the server to remove a share from it's rmtab
 * @sap: pointer to IP address of server to call
 * @salen: length of server address
 * @pmap: partially filled-in mountd RPC service tuple
 * @argp: directory path of share to "unmount"
 *
 * Returns one if the unmount call succeeded; zero if the unmount
 * failed for any reason;  rpccreateerr.cf_stat is set to reflect
 * the nature of the error.
 *
 * We use a fast timeout since this call is advisory only. */
static int
nfs_advise_umount(const struct sockaddr *sap, const socklen_t salen,
		      const struct pmap *pmap, const dirpath *argp)
{
	union nfs_sockaddr address;
	struct sockaddr *saddr = &address.sa;
	struct pmap mnt_pmap = *pmap;
	struct timeval timeout = {
		.tv_sec         = MOUNT_TIMEOUT >> 3,
	};
	CLIENT *client;
	enum clnt_stat result;

	memcpy(saddr, sap, salen);
	if (nfs_probe_mntport(saddr, salen, &mnt_pmap) == 0) {
		if (verbose)
			bb_error_msg("Failed to discover mountd port%s",
				clnt_spcreateerror(""));
		return 0;
	}
	nfs_set_port(saddr, mnt_pmap.pm_port);

	client = nfs_get_priv_rpcclient(saddr, salen, mnt_pmap.pm_prot,
					mnt_pmap.pm_prog, mnt_pmap.pm_vers,
					&timeout);
	if (client == NULL) {
		if (verbose)
			bb_error_msg("Failed to create RPC client%s",
					clnt_spcreateerror(""));
		return 0;
	}

	client->cl_auth = nfs_authsys_create();
	if (client->cl_auth == NULL) {
		if (verbose)
			bb_error_msg("Failed to create RPC auth handle");
		CLNT_DESTROY(client);
		return 0;
	}

	result = CLNT_CALL(client, MOUNTPROC_UMNT,
			(xdrproc_t)xdr_dirpath, (caddr_t)argp,
			(xdrproc_t)xdr_void, NULL,
			timeout);
	if (result != RPC_SUCCESS) {
		rpc_createerr.cf_stat = result;
		CLNT_GETERR(client, &rpc_createerr.cf_error);
		if (verbose)
			bb_error_msg("UMNT call failed: %s", clnt_sperrno(result));

	}
	auth_destroy(client->cl_auth);
	CLNT_DESTROY(client);

	if (result != RPC_SUCCESS)
		return 0;
	return 1;
}

/* Returns EX_SUCCESS if mount options and device name have been
 * parsed successfully; otherwise EX_FAIL. */
static int nfs_umount_do_umnt(struct mount_options *options,
			      char **hostname, char **dir_name)
{
	union {
		struct sockaddr         sa;
		struct sockaddr_in      s4;
#ifdef CONFIG_FEATURE_IPV6
		struct sockaddr_in6     s6;
#endif
	} address;
	struct sockaddr *sap = &address.sa;
	socklen_t salen = sizeof(address);
	struct pmap nfs_pmap, mnt_pmap;
	sa_family_t family;

	if (!nfs_options2pmap(options, &nfs_pmap, &mnt_pmap))
		return EX_FAIL;

	/* Skip UMNT call for vers=4 mounts */
	if (nfs_pmap.pm_vers == 4)
		return EX_SUCCESS;

	*hostname = nfs_umount_hostname(options, *hostname);
	if (!*hostname) {
		bb_error_msg("out of memory");
		return EX_FAIL;
	}

	if (!nfs_mount_proto_family(options, &family))
		return 0;
	if (!nfs_lookup(*hostname, family, sap, &salen))
		/* nfs_lookup reports any errors */
		return EX_FAIL;

	if (nfs_advise_umount(sap, salen, &mnt_pmap, dir_name) == 0)
		/* nfs_advise_umount reports any errors */
		return EX_FAIL;

	return EX_SUCCESS;
}

static int nfs_pdn_hostname_too_long_err(void)
{
	bb_error_msg("server hostname is too long");
	return 0;
}

static int nfs_pdn_pathname_too_long_err(void)
{
	bb_error_msg("export pathname is too long");
	return 0;
}

static int nfs_pdn_bad_format_err(void)
{
	bb_error_msg("remote share not in 'host:dir' format");
	return 0;
}

static int nfs_pdn_missing_brace_err(void)
{
	bb_error_msg("closing bracket missing from server address");
	return 0;
}

/* To handle raw IPv6 addresses (which contain colons), the
 * server's address is enclosed in square brackets.  Return
 * what's between the brackets.
 *
 * There could be anything in between the brackets, but we'll
 * let DNS resolution sort it out later. */
static int
nfs_parse_square_bracket(char *dev, char **hostname, char **pathname)
{
	size_t host_len, path_len;
	char *cbrace;

	dev++;

	/* Must have a closing square bracket */
	cbrace = strchr_set0(dev, ']');
	if (cbrace == NULL)
		return nfs_pdn_missing_brace_err();
	host_len = cbrace - dev;

	/* Must have a colon just after the closing bracket */
	cbrace++;
	if (*cbrace != ':')
		return nfs_pdn_bad_format_err();

	if (host_len > NFS_MAXHOSTNAME)
		return nfs_pdn_hostname_too_long_err();

	cbrace++;
	path_len = strlen(cbrace);
	if (path_len > NFS_MAXPATHNAME)
		return nfs_pdn_pathname_too_long_err();

	if (hostname)
		*hostname = xstrndup(dev, host_len);
	if (pathname)
		*pathname = xstrndup(cbrace, path_len);
	return 1;
}

static int nfs_pdn_no_devname_err(void)
{
	bb_error_msg("no device name was provided");
	return 0;
}

/* Standard hostname:path format */
static int nfs_parse_simple_hostname(char *dev, char **hostname, char **pathname)
{
	size_t host_len, path_len;
	char *colon, *comma;

	/* Must have a colon */
	colon = strchr_set0(dev, ':');
	if (colon == NULL)
		return nfs_pdn_bad_format_err();
	host_len = colon - dev;

	if (host_len > NFS_MAXHOSTNAME)
		return nfs_pdn_hostname_too_long_err();

	/* If there's a comma before the colon, take only the
	 * first name in list */
	comma = strchr_set0(dev, ',');
	if (comma != NULL) {
		host_len = comma - dev;
		bb_error_msg("warning: multiple hostnames not supported");
	} else
		colon++;
	path_len = strlen(colon);
	if (path_len > NFS_MAXPATHNAME)
		return nfs_pdn_pathname_too_long_err();

	if (hostname)
		*hostname = xstrndup(dev, host_len);
	if (pathname)
		*pathname = xstrndup(colon, path_len);
	return 1;
}

/* RFC 2224 says an NFS client must grok "public file handles" to
 * support NFS URLs.  Linux doesn't do that yet.  Print a somewhat
 * helpful error message in this case instead of pressing forward
 * with the mount request and failing with a cryptic error message
 * later. */
static int nfs_parse_nfs_url(const char *UNUSED(dev),
			     char **UNUSED(hostname),
			     char **UNUSED(pathname))
{
	bb_error_msg("NFS URLs are not supported");
	return 0;
}

/* nfs_parse_devname - Determine the server's hostname by looking at "devname".
 * @devname: pointer to mounted device name (first argument of mount command)
 * @hostname: OUT: pointer to server's hostname
 * @pathname: OUT: pointer to export path on server
 *
 * Returns 1 if succesful, or zero if some error occurred.  On success,
 * @hostname and @pathname point to dynamically allocated buffers containing
 * the hostname of the server and the export pathname (both '\0'-terminated).
 *
 * @hostname or @pathname may be NULL if caller doesn't want a copy of those
 * parts of @devname.
 *
 * Note that this will not work if @devname is a wide-character string. */
static int nfs_parse_devname(const char *devname,
		      char **hostname, char **pathname)
{
	char *dev;
	int result;

	if (devname == NULL)
		return nfs_pdn_no_devname_err();

	/* Parser is destructive, so operate on a copy of the device name. */
	dev = xstrdup(devname);
	if (*dev == '[')
		result = nfs_parse_square_bracket(dev, hostname, pathname);
	else if (strncmp(dev, "nfs://", 6) == 0)
		result = nfs_parse_nfs_url(dev, hostname, pathname);
	else
		result = nfs_parse_simple_hostname(dev, hostname, pathname);

	free(dev);
	return result;
}

static inline int options_empty(struct mount_options *options)
{
	return options->count == 0;
}

static void option_destroy(struct mount_option *option)
{
	free(option->keyword);
	free(option->value);
	free(option);
}

static void options_delete(struct mount_options *options,
			   struct mount_option *option)
{
	struct mount_option *prev = option->prev;
	struct mount_option *next = option->next;

	if (!options_empty(options)) {
		if (prev)
			prev->next = option->next;
		if (next)
			next->prev = option->prev;

		if (options->head == option)
			options->head = option->next;
		if (options->tail == option)
			options->tail = prev;

		options->count--;

		option_destroy(option);
	}
}

/* po_destroy - deallocate a group of mount options
 * @options: pointer to mount options to free */
static void po_destroy(struct mount_options *options)
{
	if (options) {
		while (!options_empty(options))
			options_delete(options, options->head);
		free(options);
	}
}

/* tokenizer_error - digs error value out of tokenizer context
 * @tstate: pointer to tokenizer context object */
static int tokenizer_error(struct tokenizer_state *tstate)
{
	return tstate ? tstate->error : 0;
}

static struct mount_option *option_create(char *str)
{
	struct mount_option *option;
	char *opteq;

	if (!str)
		return NULL;

	option = malloc(sizeof(*option));
	if (!option)
		return NULL;

	option->next = NULL;
	option->prev = NULL;

	opteq = strchr(str, '=');
	if (opteq) {
		option->keyword = xstrndup(str, opteq - str);
		if (!option->keyword)
			goto fail;
		option->value = xstrdup(opteq + 1);
		if (!option->value) {
			free(option->keyword);
			goto fail;
		}
	} else {
		option->keyword = xstrdup(str);
		if (!option->keyword)
			goto fail;
		option->value = NULL;
	}

	return option;

fail:
	free(option);
	return NULL;
}

static void options_init(struct mount_options *options)
{
	options->head = options->tail = NULL;
	options->count = 0;
}

static struct mount_options *options_create(void)
{
	struct mount_options *options;

	options = malloc(sizeof(*options));
	if (options)
		options_init(options);

	return options;
}

static void options_tail_insert(struct mount_options *options,
				struct mount_option *option)
{
	struct mount_option *prev = options->tail;

	option->next = NULL;
	option->prev = prev;

	if (prev)
		prev->next = option;
	else
		options->head = option;
	options->tail = option;

	options->count++;
}

/* end_tokenizer - free a tokenizer context object
 * @tstate: pointer to tokenizer context object */
static inline void end_tokenizer(struct tokenizer_state *tstate)
{
	free(tstate);
}

/* init_tokenizer - return an initialized tokenizer context object
 * @p_string: pointer to C string
 * @delimiter: single character that delimits tokens in @string
 *
 * Returns an initialized tokenizer context object */
static struct tokenizer_state *init_tokenizer(char *p_string, char delimiter)
{
	struct tokenizer_state *tstate;

	tstate = malloc(sizeof(*tstate));
	if (tstate) {
		tstate->pos = p_string;
		tstate->delimiter = delimiter;
		tstate->error = 0;
	}
	return tstate;
}

/* We've constructed a simple string tokenizer that is better than
 * strtok(3) in several ways:
 *
 * 1.  It doesn't interfere with ongoing tokenizations using strtok(3).
 * 2.  It's re-entrant so we can nest tokenizations, if needed.
 * 3.  It can handle double-quoted delimiters (needed for 'context="sd,fslj"').
 * 4.  It doesn't alter the string we're tokenizing, so it can work
 *     on write-protected strings as well as writable strings. */
static void find_next_nondelimiter(struct tokenizer_state *tstate)
{
	while (*tstate->pos != '\0' && *tstate->pos == tstate->delimiter)
		tstate->pos++;
}

static size_t find_next_delimiter(struct tokenizer_state *tstate)
{
	size_t len = 0;
	int quote_seen = 0;

	while (*tstate->pos != '\0') {
		if (*tstate->pos == '"')
			quote_seen ^= 1;

		if (!quote_seen && *tstate->pos == tstate->delimiter)
			break;

		len++;
		tstate->pos++;
	}

	/* did the string terminate before the close quote? */
	if (quote_seen) {
		tstate->error = EINVAL;
		return 0;
	}

	return len;
}

/* next_token - find the next token in a string and return it
 * @tstate: pointer to tokenizer context object
 *
 * Returns the next token found in the current string.
 * Returns NULL if there are no more tokens in the string,
 * or if an error occurs.
 *
 * Side effect: tstate is updated */
static char *next_token(struct tokenizer_state *tstate)
{
	char *token;
	size_t len;

	if (!tstate || !tstate->pos || tstate->error)
		return NULL;

	find_next_nondelimiter(tstate);
	if (*tstate->pos == '\0')
		goto fail;
	token = tstate->pos;

	len = find_next_delimiter(tstate);
	if (len) {
		token = xstrndup(token, len);
		if (token)
			return token;
		tstate->error = ENOMEM;
	}

fail:
	tstate->pos = NULL;
	return NULL;                    /* no tokens found in this string */
}

/* po_split - split options string into group of options
 * @options: pointer to C string containing zero or more comma-delimited options
 *
 * Convert our mount options string to a list to make it easier
 * to adjust the options as we go.  This is just an exercise in
 * lexical parsing -- this function doesn't pay attention to the
 * meaning of the options themselves.
 *
 * Returns a new group of mount options if successful; otherwise NULL
 * is returned if some failure occurred. */
static struct mount_options *po_split(char *str)
{
	struct mount_options *options;
	struct tokenizer_state *tstate;
	char *opt;

	if (!str)
		return options_create();

	options = options_create();
	if (options) {
		tstate = init_tokenizer(str, ',');
		for (opt = next_token(tstate); opt; opt = next_token(tstate)) {
			struct mount_option *option = option_create(opt);
			free(opt);
			if (!option)
				goto fail;
			options_tail_insert(options, option);
		}
		if (tokenizer_error(tstate))
			goto fail;
		end_tokenizer(tstate);
	}
	return options;

fail:
	end_tokenizer(tstate);
	po_destroy(options);
	return NULL;
}

/* Pick up certain mount options used during the original mount
 * from /etc/mtab.  The basics include the server's IP address and
 * the server pathname of the share to unregister.
 *
 * These options might also describe the mount port, mount protocol
 * version, and transport protocol used to punch through a firewall.
 * We will need this information to get through the firewall again
 * to do the umount.
 *
 * Note that option parsing failures won't necessarily cause the
 * umount request to fail.  Those values will be left zero in the
 * pmap tuple.  If the GETPORT call later fails to disambiguate them,
 * then we fail. */
static int nfs_umount23(const char *devname, char *po_string)
{
	char *hostname, *dir_name;
	struct mount_options *options;
	int result = EX_FAIL;

	if (!nfs_parse_devname(devname, &hostname, &dir_name))
		return EX_USAGE;

	options = po_split(po_string);
	if (options) {
		result = nfs_umount_do_umnt(options, &hostname, &dir_name);
		po_destroy(options);
	} else
		bb_error_msg("option parsing error");

	free(hostname);
	free(dir_name);
	return result;
}

#ifdef CONFIG_FEATURE_LONG_OPTIONS
static const struct option umount_longopts[] =
{
  { "force", 0, 0, 'f' },
  { "help", 0, 0, 'h' },
  { "no-mtab", 0, 0, 'n' },
  { "verbose", 0, 0, 'v' },
  { "read-only", 0, 0, 'r' },
  { NULL, 0, 0, 0 }
};
#endif

int umount_nfs_main(void)
{
	int ret;
	char *spec;
	struct mntentchn *mc;

#ifdef CONFIG_FEATURE_LONG_OPTIONS
	bb_applet_long_options = umount_longopts;
#endif
	bb_xgetopt_ulflags("flrnhv\0=1:h-:vv", &verbose);
	/* FORCE = flags & 1; */
	/* LAZY = flags & 2; */
	/* REMOUNT = flags & 4; */
	/* NOMTAB = flags & 8; */

	spec = Gargv[0];

	if (spec == NULL || (*spec != '/' && strchr(spec,':') == NULL))
		bb_error_msg_and_die("%s: not found", spec);

	if (*spec == '/')
		mc = getmntdirbackward(spec, NULL);
	else
		mc = getmntdevbackward(spec, NULL);
	if (!mc && verbose)
		printf("Could not find %s in mtab\n", spec);

	if (mc && strcmp(mc->m.mnt_type, "nfs") != 0 &&
	    strcmp(mc->m.mnt_type, "nfs4") != 0) {
		bb_error_msg_and_die("%s on %s is not an NFS filesystem",
				mc->m.mnt_fsname, mc->m.mnt_dir);
	}

	if (getuid() != 0) {
		/* only permitted if "user=" or "users" is in mount options */
		if (!mc) {
			/* umount might call us twice.  The second time there will
			 * be no entry in mtab and we should just exit quietly
			 */
			return EX_SUCCESS;

		only_root:
			bb_error_msg_and_die("You are not permitted to unmount %s", spec);
		}
		if (hasmntopt(&mc->m, "users") == NULL) {
			char *opt = hasmntopt(&mc->m, "user");
			struct passwd *pw;
			char *comma;
			size_t len;
			if (!opt)
				goto only_root;
			if (opt[4] != '=')
				goto only_root;
			comma = strchr(opt, ',');
			if (comma)
				len = comma - (opt + 5);
			else
				len = strlen(opt+5);
			pw = getpwuid(getuid());
			if (pw == NULL || strlen(pw->pw_name) != len
			    || strncmp(pw->pw_name, opt+5, len) != 0)
				goto only_root;
		}
	}

	ret = 0;
	if (mc) {
		if (!LAZY && strcmp(mc->m.mnt_type, "nfs4") != 0)
			/* We ignore the error from nfs_umount23.
			 * If the actual umount succeeds (in del_mtab),
			 * we don't want to signal an error, as that
			 * could cause /sbin/mount to retry!
			 */
			nfs_umount23(mc->m.mnt_fsname, mc->m.mnt_opts);
		ret = del_mtab(mc->m.mnt_fsname, mc->m.mnt_dir) ?: ret;
	} else if (*spec != '/') {
		if (!LAZY)
			ret = nfs_umount23(spec, "tcp,v3");
	} else
		ret = del_mtab(NULL, spec);

	return ret;
}


static struct mount_option *option_dup(const struct mount_option *option)
{
	struct mount_option *new;

	new = malloc(sizeof(*new));
	if (!new)
		return NULL;

	new->next = NULL;
	new->prev = NULL;

	new->keyword = xstrdup(option->keyword);
	if (!new->keyword)
		goto fail;

	new->value = NULL;
	if (option->value) {
		new->value = xstrdup(option->value);
		if (!new->value) {
			free(new->keyword);
			goto fail;
		}
	}

	return new;

fail:
	free(new);
	return NULL;
}


/* po_dup - duplicate an existing list of options
 * @options: pointer to mount options */
static struct mount_options *po_dup(struct mount_options *source)
{
	struct mount_options *target;
	struct mount_option *current;

	if (!source)
		return NULL;

	target = options_create();
	if (options_empty(source) || target == NULL)
		return target;

	current = source->head;
	while (target->count < source->count) {
		struct mount_option *option;

		option = option_dup(current);
		if (!option) {
			po_destroy(target);
			return NULL;
		}

		options_tail_insert(target, option);
		current = current->next;
	}

	return target;
}

/* po_join - recombine group of mount options into a C string
 * @options: pointer to mount options to recombine
 * @str: handle on string to replace (input and output)
 *
 * Convert our mount options object back into a string that the
 * rest of the world can use.
 *
 * Upon return, @string contains the address of a replacement
 * C string containing a comma-delimited list of mount options
 * and values; or the passed-in string is freed and NULL is
 * returned if some failure occurred. */
static po_return_t po_join(struct mount_options *options, char **str)
{
	size_t len = 0;
	struct mount_option *option;

	if (!str || !options)
		return PO_FAILED;

	free(*str);
	*str = NULL;

	if (options_empty(options)) {
		*str = xstrdup("");
		return *str ? PO_SUCCEEDED : PO_FAILED;
	}

	for (option = options->head; option; option = option->next) {
		len += strlen(option->keyword);
		if (option->value)
			len +=strlen(option->value) + 1;  /* equals sign */
		if (option->next)
			len++;  /* comma */
	}

	len++;  /* NULL on the end */

	*str = malloc(len);
	if (!*str)
		return PO_FAILED;
	*str[0] = '\0';

	for (option = options->head; option; option = option->next) {
		strcat(*str, option->keyword);
		if (option->value) {
			strcat(*str, "=");
			strcat(*str, option->value);
		}
		if (option->next)
			strcat(*str, ",");
	}

	return PO_SUCCEEDED;
}

/* po_append - concatenate an option onto a group of options
 * @options: pointer to mount options
 * @option: pointer to a C string containing the option to add */
static po_return_t po_append(struct mount_options *options, char *str)
{
	struct mount_option *option = option_create(str);

	if (option) {
		options_tail_insert(options, option);
		return PO_SUCCEEDED;
	}
	return PO_FAILED;
}

/* po_contains - check for presense of an option in a group
 * @options: pointer to mount options
 * @keyword: pointer to a C string containing option keyword
 * for which to search */
static po_found_t po_contains(struct mount_options *options, char *keyword)
{
	struct mount_option *option;

	if (options && keyword) {
		for (option = options->head; option; option = option->next)
			if (strcmp(option->keyword, keyword) == 0)
				return PO_FOUND;
	}

	return PO_NOT_FOUND;
}

/* po_remove_all - remove instances of an option from a group
 * @options: pointer to mount options
 * @keyword: pointer to a C string containing an option keyword to remove
 *
 * Side-effect: the passed-in list is truncated on success. */
static po_found_t po_remove_all(struct mount_options *options, char *keyword)
{
	struct mount_option *option, *next;
	int found = PO_NOT_FOUND;

	if (options && keyword) {
		for (option = options->head; option; option = next) {
			next = option->next;
			if (strcmp(option->keyword, keyword) == 0) {
				options_delete(options, option);
				found = PO_FOUND;
			}
		}
	}

	return found;
}


#ifndef NFS_DEF_FG_TIMEOUT_MINUTES
#define NFS_DEF_FG_TIMEOUT_MINUTES      (2u)
#endif

#ifndef NFS_DEF_BG_TIMEOUT_MINUTES
#define NFS_DEF_BG_TIMEOUT_MINUTES      (10000u)
#endif

struct nfsmount_info {
	const char              *spec,          /* server:/path */
				*node,          /* mounted-on dir */
				*type;          /* "nfs" or "nfs4" */
	char                    *hostname;      /* server's hostname */
	struct addrinfo         *address;       /* server's addresses */

	struct mount_options    *options;       /* parsed mount options */
	char                    **extra_opts;   /* string for /etc/mtab */

	unsigned long           version;        /* NFS version */
	int                     flags,          /* MS_ flags */
				fake,           /* actually do the mount? */
				child;          /* forked bg child? */
};

/*
 * Obtain a retry timeout value based on the value of the "retry=" option.
 *
 * Returns a time_t timeout timestamp, in seconds.
 */
static time_t nfs_parse_retry_option(struct mount_options *options,
				     unsigned int timeout_minutes)
{
	long tmp;

	switch (po_get_numeric(options, "retry", &tmp)) {
	case PO_NOT_FOUND:
		break;
	case PO_FOUND:
		if (tmp >= 0) {
			timeout_minutes = tmp;
			break;
		}
	case PO_BAD_VALUE:
		if (verbose)
			bb_error_msg("invalid retry timeout was specified; "
					"using default timeout");
		break;
	}

	return time(NULL) + (time_t)(timeout_minutes * 60);
}

/* Convert the passed-in sockaddr-style address to presentation
 * format, then append an option of the form "keyword=address".
 *
 * Returns 1 if the option was appended successfully; otherwise zero. */
static int nfs_append_generic_address_option(const struct sockaddr *sap,
					     const socklen_t salen,
					     const char *keyword,
					     struct mount_options *options)
{
	char address[NI_MAXHOST];
	char new_option[512];
	int len;

	if (!nfs_present_sockaddr(sap, salen, address, sizeof(address)))
		goto out_err;

	len = snprintf(new_option, sizeof(new_option), "%s=%s",
						keyword, address);
	if (len < 0 || (size_t)len >= sizeof(new_option))
		goto out_err;

	if (po_append(options, new_option) != PO_SUCCEEDED)
		goto out_err;

	return 1;

out_err:
	bb_error_msg("failed to construct %s option", keyword);
	return 0;
}

/* Append the 'addr=' option to the options string to pass a resolved
 * server address to the kernel.  After a successful mount, this address
 * is also added to /etc/mtab for use when unmounting.
 *
 * If 'addr=' is already present, we strip it out.  This prevents users
 * from setting a bogus 'addr=' option themselves, and also allows bg
 * retries to recompute the server's address, in case it has changed.
 *
 * Returns 1 if 'addr=' option appended successfully;
 * otherwise zero. */
static int nfs_append_addr_option(const struct sockaddr *sap,
				  socklen_t salen,
				  struct mount_options *options)
{
	po_remove_all(options, "addr");
	return nfs_append_generic_address_option(sap, salen, "addr", options);
}

/* Called to discover our address and append an appropriate 'clientaddr='
 * option to the options string.
 *
 * Returns 1 if 'clientaddr=' option created successfully or if
 * 'clientaddr=' option is already present; otherwise zero. */
static int nfs_append_clientaddr_option(const struct sockaddr *sap,
					socklen_t salen,
					struct mount_options *options)
{
	union nfs_sockaddr address;
	struct sockaddr *my_addr = &address.sa;
	socklen_t my_len = sizeof(address);

	if (po_contains(options, "clientaddr") == PO_FOUND)
		return 1;

	nfs_callback_address(sap, salen, my_addr, &my_len);

	return nfs_append_generic_address_option(my_addr, my_len,
							"clientaddr", options);
}

/* Determine whether to append a 'mountaddr=' option.  The option is needed if:
 *
 *   1. "mounthost=" was specified, or
 *   2. The address families for proto= and mountproto= are different. */
static int nfs_fix_mounthost_option(struct mount_options *options,
		const char *nfs_hostname)
{
	union nfs_sockaddr address;
	struct sockaddr *sap = &address.sa;
	socklen_t salen = sizeof(address);
	sa_family_t nfs_family, mnt_family;
	char *mounthost;

	if (!nfs_nfs_proto_family(options, &nfs_family))
		return 0;
	if (!nfs_mount_proto_family(options, &mnt_family))
		return 0;

	mounthost = po_get(options, "mounthost");
	if (mounthost == NULL) {
		if (nfs_family == mnt_family)
			return 1;
		mounthost = (char *)nfs_hostname;
	}

	if (!nfs_lookup(mounthost, mnt_family, sap, &salen)) {
		bb_error_msg("unable to determine mount server's address");
		return 0;
	}

	return nfs_append_generic_address_option(sap, salen,
							"mountaddr", options);
}

/* Returns zero if the "lock" option is in effect, but statd
 * can't be started.  Otherwise, returns 1. */
static const char * const nfs_lock_opttbl[] = {
	"nolock",
	"lock",
	NULL,
};

static int nfs_verify_lock_option(struct mount_options *options)
{
	if (po_rightmost(options, nfs_lock_opttbl) == 0)
		return 1;

	if (!start_statd()) {
		bb_error_msg("rpc.statd is not running but is "
			    "required for remote locking.\n"
			    "Either use '-o nolock' to keep "
			    "locks local, or start statd.");
		return 0;
	}

	return 1;
}

static int nfs_append_sloppy_option(struct mount_options *options)
{
	if (!SLOPPY || linux_version_code() < MAKE_VERSION(2, 6, 27))
		return 1;

	if (po_append(options, "sloppy") == PO_FAILED)
		return 0;
	return 1;
}

static int nfs_set_version(struct nfsmount_info *mi)
{
	if (!nfs_nfs_version(mi->options, &mi->version))
		return 0;

	if (strncmp(mi->type, "nfs4", 4) == 0)
		mi->version = 4;

	/*
	 * Before 2.6.32, the kernel NFS client didn't
	 * support "-t nfs vers=4" mounts, so NFS version
	 * 4 cannot be included when autonegotiating
	 * while running on those kernels.
	 */
	if (mi->version == 0 &&
	    linux_version_code() <= MAKE_VERSION(2, 6, 31))
		mi->version = 3;

	/*
	 * If we still don't know, check for version-specific
	 * mount options.
	 */
	if (mi->version == 0) {
		if (po_contains(mi->options, "mounthost") ||
		    po_contains(mi->options, "mountaddr") ||
		    po_contains(mi->options, "mountvers") ||
		    po_contains(mi->options, "mountproto"))
			mi->version = 3;
	}

	return 1;
}

/* Set up mandatory non-version specific NFS mount options.
 * Returns 1 if successful; otherwise zero. */
static int nfs_validate_options(struct nfsmount_info *mi)
{
	struct addrinfo hint = {
		.ai_protocol    = (int)IPPROTO_UDP,
		.ai_flags       = AI_ADDRCONFIG,
	};
	sa_family_t family;
	int error;

	if (!nfs_parse_devname(mi->spec, &mi->hostname, NULL))
		return 0;

	if (!nfs_nfs_proto_family(mi->options, &family))
		return 0;

	hint.ai_family = (int)family;
	error = getaddrinfo(mi->hostname, NULL, &hint, &mi->address);
	if (error != 0) {
		bb_error_msg("Failed to resolve server %s: %s",
				mi->hostname, gai_strerror(error));
		mi->address = NULL;
		return 0;
	}

	if (!nfs_set_version(mi))
		return 0;

	if (!nfs_append_sloppy_option(mi->options))
		return 0;

	if (!nfs_append_addr_option(mi->address->ai_addr,
					mi->address->ai_addrlen, mi->options))
		return 0;

	return 1;
}

/* Get NFS/mnt server addresses from mount options
 *
 * Returns 1 and fills in @nfs_saddr, @nfs_salen, @mnt_saddr, and @mnt_salen
 * if all goes well; otherwise zero. */
static int nfs_extract_server_addresses(struct mount_options *options,
					struct sockaddr *nfs_saddr,
					socklen_t *nfs_salen,
					struct sockaddr *mnt_saddr,
					socklen_t *mnt_salen)
{
	char *option;

	option = po_get(options, "addr");
	if (option == NULL)
		return 0;
	if (!nfs_string_to_sockaddr(option, nfs_saddr, nfs_salen))
		return 0;

	option = po_get(options, "mountaddr");
	if (option == NULL) {
		memcpy(mnt_saddr, nfs_saddr, *nfs_salen);
		*mnt_salen = *nfs_salen;
	} else if (!nfs_string_to_sockaddr(option, mnt_saddr, mnt_salen))
		return 0;

	return 1;
}

static int nfs_construct_new_options(struct mount_options *options,
				     struct sockaddr *nfs_saddr,
				     struct pmap *nfs_pmap,
				     struct sockaddr *mnt_saddr,
				     struct pmap *mnt_pmap)
{
	char new_option[64];
	char *netid;

	po_remove_all(options, "nfsprog");
	po_remove_all(options, "mountprog");

	po_remove_all(options, "v2");
	po_remove_all(options, "v3");
	po_remove_all(options, "vers");
	po_remove_all(options, "nfsvers");
	snprintf(new_option, sizeof(new_option) - 1,
		 "vers=%lu", nfs_pmap->pm_vers);
	if (po_append(options, new_option) == PO_FAILED)
		return 0;

	po_remove_all(options, "proto");
	po_remove_all(options, "udp");
	po_remove_all(options, "tcp");
	netid = nfs_get_netid(nfs_saddr->sa_family, nfs_pmap->pm_prot);
	if (netid == NULL)
		return 0;
	snprintf(new_option, sizeof(new_option) - 1,
			 "proto=%s", netid);
	free(netid);
	if (po_append(options, new_option) == PO_FAILED)
		return 0;

	po_remove_all(options, "port");
	if (nfs_pmap->pm_port != NFS_PORT) {
		snprintf(new_option, sizeof(new_option) - 1,
			 "port=%lu", nfs_pmap->pm_port);
		if (po_append(options, new_option) == PO_FAILED)
			return 0;
	}

	po_remove_all(options, "mountvers");
	snprintf(new_option, sizeof(new_option) - 1,
		 "mountvers=%lu", mnt_pmap->pm_vers);
	if (po_append(options, new_option) == PO_FAILED)
		return 0;

	po_remove_all(options, "mountproto");
	netid = nfs_get_netid(mnt_saddr->sa_family, mnt_pmap->pm_prot);
	if (netid == NULL)
		return 0;
	snprintf(new_option, sizeof(new_option) - 1,
			 "mountproto=%s", netid);
	free(netid);
	if (po_append(options, new_option) == PO_FAILED)
		return 0;

	po_remove_all(options, "mountport");
	snprintf(new_option, sizeof(new_option) - 1,
		 "mountport=%lu", mnt_pmap->pm_port);
	if (po_append(options, new_option) == PO_FAILED)
		return 0;

	return 1;
}

/* Reconstruct the mount option string based on a portmapper probe
 * of the server.  Returns one if the server's portmapper returned
 * something we can use, otherwise zero.
 *
 * To handle version and transport protocol fallback properly, we
 * need to parse some of the mount options in order to set up a
 * portmap probe.  Mount options that nfs_rewrite_pmap_mount_options()
 * doesn't recognize are left alone.
 *
 * Returns TRUE if rewriting was successful; otherwise
 * FALSE is returned if some failure occurred. */
static int
nfs_rewrite_pmap_mount_options(struct mount_options *options)
{
	union nfs_sockaddr nfs_address;
	struct sockaddr *nfs_saddr = &nfs_address.sa;
	socklen_t nfs_salen = sizeof(nfs_address);
	struct pmap nfs_pmap;
	union nfs_sockaddr mnt_address;
	struct sockaddr *mnt_saddr = &mnt_address.sa;
	socklen_t mnt_salen = sizeof(mnt_address);
	unsigned long protocol;
	struct pmap mnt_pmap;

	/* Version and transport negotiation is not required
	 * and does not work for RDMA mounts. */
	if (!nfs_nfs_protocol(options, &protocol)) {
		errno = EINVAL;
		return 0;
	}
	if (protocol == NFSPROTO_RDMA)
		goto out;

	/* Extract just the options needed to contact server.
	 * Bail now if any of these have bad values. */
	if (!nfs_extract_server_addresses(options, nfs_saddr, &nfs_salen,
						mnt_saddr, &mnt_salen)) {
		errno = EINVAL;
		return 0;
	}
	if (!nfs_options2pmap(options, &nfs_pmap, &mnt_pmap)) {
		errno = EINVAL;
		return 0;
	}

	/* The kernel NFS client doesn't support changing the RPC
	 * program number for these services, so force the value of
	 * these fields before probing the server's ports. */
	nfs_pmap.pm_prog = NFS_PROGRAM;
	mnt_pmap.pm_prog = MOUNTPROG;

	/* If the server's rpcbind service isn't available, we can't
	 * negotiate.  Bail now if we can't contact it. */
	if (!nfs_probe_bothports(mnt_saddr, mnt_salen, &mnt_pmap,
				 nfs_saddr, nfs_salen, &nfs_pmap)) {
		errno = ESPIPE;
		if (rpc_createerr.cf_stat == RPC_PROGNOTREGISTERED)
			errno = EOPNOTSUPP;
		else if (rpc_createerr.cf_error.re_errno != 0)
			errno = rpc_createerr.cf_error.re_errno;
		return 0;
	}

	if (!nfs_construct_new_options(options, nfs_saddr, &nfs_pmap,
					mnt_saddr, &mnt_pmap)) {
		if (rpc_createerr.cf_stat == RPC_UNKNOWNPROTO)
			errno = EPROTONOSUPPORT;
		else
			errno = EINVAL;
		return 0;
	}

out:
	errno = 0;
	return 1;
}

/* Do the mount(2) system call.
 *
 * Returns TRUE if successful, otherwise FALSE.
 * "errno" is set to reflect the individual error. */
static int nfs_sys_mount(struct nfsmount_info *mi, struct mount_options *opts)
{
	char *options = NULL;
	int result;

	if (po_join(opts, &options) == PO_FAILED) {
		errno = EIO;
		return 0;
	}

	if (mi->fake)
		return 1;

	result = mount(mi->spec, mi->node, mi->type,
			mi->flags & ~(MS_USER|MS_USERS), options);
	if (verbose && result) {
		int e = errno;
		bb_perror_msg("mount(2)");
		errno = e;
	}
	return !result;
}

static int nfs_do_mount_v3v2(struct nfsmount_info *mi,
		struct sockaddr *sap, socklen_t salen)
{
	struct mount_options *options = po_dup(mi->options);
	int result = 0;

	if (!options) {
		errno = ENOMEM;
		return result;
	}
	errno = 0;
	if (!nfs_append_addr_option(sap, salen, options)) {
		if (errno == 0)
			errno = EINVAL;
		goto out_fail;
	}

	if (!nfs_fix_mounthost_option(options, mi->hostname)) {
		if (errno == 0)
			errno = EINVAL;
		goto out_fail;
	}
	if (!mi->fake && !nfs_verify_lock_option(options)) {
		if (errno == 0)
			errno = EINVAL;
		goto out_fail;
	}

	/*
	 * Options we negotiate below may be stale by the time this
	 * file system is unmounted.  In order to force umount.nfs
	 * to renegotiate with the server, only write the user-
	 * specified options, and not negotiated options, to /etc/mtab.
	 */
	if (po_join(options, mi->extra_opts) == PO_FAILED) {
		errno = ENOMEM;
		goto out_fail;
	}

	if (verbose)
		printf("%s: trying text-based options '%s'\n",
			bb_applet_name, *mi->extra_opts);

	if (!nfs_rewrite_pmap_mount_options(options))
		goto out_fail;

	result = nfs_sys_mount(mi, options);

out_fail:
	po_destroy(options);
	return result;
}

/* Attempt a "-t nfs vers=2" or "-t nfs vers=3" mount.
 *
 * Returns TRUE if successful, otherwise FALSE.
 * "errno" is set to reflect the individual error. */
static int nfs_try_mount_v3v2(struct nfsmount_info *mi)
{
	struct addrinfo *ai;
	int ret;

	for (ai = mi->address; ai != NULL; ai = ai->ai_next) {
		ret = nfs_do_mount_v3v2(mi, ai->ai_addr, ai->ai_addrlen);
		if (ret != 0)
			return ret;

		switch (errno) {
		case ECONNREFUSED:
		case EOPNOTSUPP:
		case EHOSTUNREACH:
			continue;
		default:
			break;
		}
	}
	return ret;
}

static int nfs_do_mount_v4(struct nfsmount_info *mi,
		struct sockaddr *sap, socklen_t salen)
{
	struct mount_options *options = po_dup(mi->options);
	int result = 0;

	if (!options) {
		errno = ENOMEM;
		return result;
	}

	if (mi->version == 0) {
		if (po_contains(options, "mounthost") ||
			po_contains(options, "mountaddr") ||
			po_contains(options, "mountvers") ||
			po_contains(options, "mountproto")) {
		/*
		 * Since these mountd options are set assume version 3
		 * is wanted so error out with EPROTONOSUPPORT so the
		 * protocol negation starts with v3.
		 */
			errno = EPROTONOSUPPORT;
			goto out_fail;
		}
		if (po_append(options, "vers=4") == PO_FAILED) {
			errno = EINVAL;
			goto out_fail;
		}
	}

	if (!nfs_append_addr_option(sap, salen, options)) {
		errno = EINVAL;
		goto out_fail;
	}

	if (!nfs_append_clientaddr_option(sap, salen, options)) {
		errno = EINVAL;
		goto out_fail;
	}

	/*
	 * Update option string to be recorded in /etc/mtab.
	 */
	if (po_join(options, mi->extra_opts) == PO_FAILED) {
		errno = ENOMEM;
		goto out_fail;
	}

	if (verbose)
		printf("%s: trying text-based options '%s'\n",
			bb_applet_name, *mi->extra_opts);

	result = nfs_sys_mount(mi, options);

out_fail:
	po_destroy(options);
	return result;
}

/* Attempt a "-t nfs -o vers=4" or "-t nfs4" mount.
 *
 * Returns TRUE if successful, otherwise FALSE.
 * "errno" is set to reflect the individual error. */
static int nfs_try_mount_v4(struct nfsmount_info *mi)
{
	struct addrinfo *ai;
	int ret;

	for (ai = mi->address; ai != NULL; ai = ai->ai_next) {
		ret = nfs_do_mount_v4(mi, ai->ai_addr, ai->ai_addrlen);
		if (ret != 0)
			return ret;

		switch (errno) {
		case ECONNREFUSED:
		case EHOSTUNREACH:
			continue;
		default:
			break;
		}
	}
	return ret;
}

/* Handle NFS version and transport protocol autonegotiation.
 *
 * When no version or protocol is specified on the
 * command line, mount.nfs negotiates with the server
 * to determine appropriate settings for the new
 * mount point.
 *
 * Returns TRUE if successful, otherwise FALSE.
 * "errno" is set to reflect the individual error. */
static int nfs_autonegotiate(struct nfsmount_info *mi)
{
	int result;

	result = nfs_try_mount_v4(mi);
	if (result)
		return result;

	switch (errno) {
	case EPROTONOSUPPORT:
		/* A clear indication that the server or our
		 * client does not support NFS version 4. */
		goto fall_back;
	case ENOENT:
		/* Legacy Linux servers don't export an NFS
		 * version 4 pseudoroot. */
		goto fall_back;
	case EPERM:
		/* Linux servers prior to 2.6.25 may return
		 * EPERM when NFS version 4 is not supported. */
		goto fall_back;
	default:
		return result;
	}

fall_back:
	return nfs_try_mount_v3v2(mi);
}

/* This is a single pass through the fg/bg loop.
 *
 * Returns TRUE if successful, otherwise FALSE.
 * "errno" is set to reflect the individual error. */
static int nfs_try_mount(struct nfsmount_info *mi)
{
	int result = 0;

	switch (mi->version) {
	case 0:
		result = nfs_autonegotiate(mi);
		break;
	case 2:
	case 3:
		result = nfs_try_mount_v3v2(mi);
		break;
	case 4:
		result = nfs_try_mount_v4(mi);
		break;
	default:
		errno = EIO;
	}

	return result;
}

/* Distinguish between permanent and temporary errors.
 *
 * Basically, we retry if communication with the server has
 * failed so far, but fail immediately if there is a local
 * error (like a bad mount option).
 *
 * ESTALE is also a temporary error because some servers
 * return ESTALE when a share is temporarily offline.
 *
 * Returns 1 if we should fail immediately, or 0 if we
 * should retry. */
static int nfs_is_permanent_error(int error)
{
	switch (error) {
	case ESTALE:
	case ETIMEDOUT:
	case ECONNREFUSED:
	case EHOSTUNREACH:
		return 0;       /* temporary */
	default:
		return 1;       /* permanent */
	}
}

/* Handle "foreground" NFS mounts.
 *
 * Retry the mount request for as long as the 'retry=' option says.
 *
 * Returns a valid mount command exit code. */
static int nfsmount_fg(struct nfsmount_info *mi)
{
	unsigned int secs = 1;
	time_t timeout;

	timeout = nfs_parse_retry_option(mi->options,
					 NFS_DEF_FG_TIMEOUT_MINUTES);
	if (verbose)
		printf("%s: timeout set for %s", bb_applet_name, ctime(&timeout));

	for (;;) {
		if (nfs_try_mount(mi))
			return EX_SUCCESS;

		if (nfs_is_permanent_error(errno))
			break;

		if (time(NULL) > timeout) {
			errno = ETIMEDOUT;
			break;
		}

		if (errno != ETIMEDOUT) {
			if (sleep(secs))
				break;
			secs <<= 1;
			if (secs > 10)
				secs = 10;
		}
	};

	mount_error(mi->spec, mi->node);
	return EX_FAIL;
}

/* Handle "background" NFS mount [first try]
 *
 * Returns a valid mount command exit code.
 *
 * EX_BG should cause the caller to fork and invoke nfsmount_child. */
static int nfsmount_parent(struct nfsmount_info *mi)
{
	if (nfs_try_mount(mi))
		return EX_SUCCESS;

	if (nfs_is_permanent_error(errno)) {
		mount_error(mi->spec, mi->node);
		return EX_FAIL;
	}

	sys_mount_errors(mi->hostname, 1);
	return EX_BG;
}

/* Handle "background" NFS mount [retry daemon]
 *
 * Returns a valid mount command exit code: EX_SUCCESS if successful,
 * EX_FAIL if a failure occurred.  There's nothing to catch the
 * error return, though, so we use sys_mount_errors to log the failure. */
static int nfsmount_child(struct nfsmount_info *mi)
{
	unsigned int secs = 1;
	time_t timeout;

	timeout = nfs_parse_retry_option(mi->options,
					 NFS_DEF_BG_TIMEOUT_MINUTES);

	for (;;) {
		if (sleep(secs))
			break;
		secs <<= 1;
		if (secs > 120)
			secs = 120;

		if (nfs_try_mount(mi))
			return EX_SUCCESS;

		if (nfs_is_permanent_error(errno))
			break;

		if (time(NULL) > timeout)
			break;

		sys_mount_errors(mi->hostname, 1);
	}

	sys_mount_errors(mi->hostname, 0);
	return EX_FAIL;
}

/* Handle "background" NFS mount
 * Returns a valid mount command exit code. */
static int nfsmount_bg(struct nfsmount_info *mi)
{
	if (!mi->child)
		return nfsmount_parent(mi);
	else
		return nfsmount_child(mi);
}

/* Usually all that is needed for an NFS remount is to change
 * generic mount options like "sync" or "ro".  These generic
 * options are controlled by mi->flags, not by text-based
 * options, and no contact with the server is needed.
 *
 * Take care with the /etc/mtab entry for this mount; just
 * calling update_mtab() will change an "-t nfs -o vers=4"
 * mount to an "-t nfs -o remount" mount, and that will
 * confuse umount.nfs.
 *
 * Returns a valid mount command exit code. */
static int nfs_remount(struct nfsmount_info *mi)
{
	if (nfs_sys_mount(mi, mi->options))
		return EX_SUCCESS;
	return EX_FAIL;
}

/* Process mount options and try a mount system call.
 * Returns a valid mount command exit code. */
static const char * const nfs_background_opttbl[] = {
	"bg",
	"fg",
	NULL,
};

static int nfsmount_start(struct nfsmount_info *mi)
{
	if (!nfs_validate_options(mi))
		return EX_FAIL;

	/*
	 * Avoid retry and negotiation logic when remounting
	 */
	if (mi->flags & MS_REMOUNT)
		return nfs_remount(mi);

	if (po_rightmost(mi->options, nfs_background_opttbl) == 0)
		return nfsmount_bg(mi);
	else
		return nfsmount_fg(mi);
}

/* nfsmount_string - Mount an NFS file system using C string options
 * @spec: C string specifying remote share to mount ("hostname:path")
 * @node: C string pathname of local mounted-on directory
 * @type: C string that represents file system type ("nfs" or "nfs4")
 * @flags: MS_ style mount flags
 * @extra_opts: pointer to C string containing fs-specific mount options
 *              (input and output argument)
 * @fake: flag indicating whether to carry out the whole operation
 * @child: one if this is a mount daemon (bg)
 *
 * Returns a valid mount command exit code. */
static int
nfsmount_string(const char *spec, const char *node, const char *type,
		    int flags, char **extra_opts, int fake, int child)
{
	struct nfsmount_info mi = {
		.spec           = spec,
		.node           = node,
		.address        = NULL,
		.type           = type,
		.extra_opts     = extra_opts,
		.flags          = flags,
		.fake           = fake,
		.child          = child,
	};
	int retval = EX_FAIL;

	mi.options = po_split(*extra_opts);
	if (mi.options) {
		retval = nfsmount_start(&mi);
		po_destroy(mi.options);
	} else
		bb_error_msg("internal option parsing error");

	freeaddrinfo(mi.address);
	free(mi.hostname);
	return retval;
}

static bool_t xdr_name (XDR *xdrs, char **objp)
{
       if (!xdr_string (xdrs, objp, MNTNAMLEN))
		 return FALSE;
	return TRUE;
}

static bool_t xdr_mountbody (XDR *xdrs, mountbody *objp);

static bool_t
xdr_mountlist (XDR *xdrs, mountlist *objp)
{
	if (!xdr_pointer (xdrs, (char **)objp, sizeof (struct mountbody), (xdrproc_t) xdr_mountbody))
		return FALSE;
	return TRUE;
}

static bool_t xdr_mountbody (XDR *xdrs, mountbody *objp)
{
	if (!xdr_name (xdrs, &objp->ml_hostname))
		return FALSE;
	if (!xdr_dirpath (xdrs, &objp->ml_directory))
		return FALSE;
	if (!xdr_mountlist (xdrs, &objp->ml_next))
		return FALSE;
	return TRUE;
}

static bool_t xdr_groups (XDR *xdrs, groups *objp);

static bool_t xdr_groupnode (XDR *xdrs, groupnode *objp)
{
	if (!xdr_name (xdrs, &objp->gr_name))
		return FALSE;
	if (!xdr_groups (xdrs, &objp->gr_next))
		return FALSE;
	return TRUE;
}

static bool_t xdr_groups (XDR *xdrs, groups *objp)
{
	if (!xdr_pointer (xdrs, (char **)objp, sizeof (struct groupnode), (xdrproc_t) xdr_groupnode))
		return FALSE;
	return TRUE;
}

static bool_t xdr_exports (XDR *xdrs, exports *objp);

static bool_t xdr_exportnode (XDR *xdrs, exportnode *objp)
{
	if (!xdr_dirpath (xdrs, &objp->ex_dir))
		return FALSE;
	if (!xdr_groups (xdrs, &objp->ex_groups))
		return FALSE;
	if (!xdr_exports (xdrs, &objp->ex_next))
		return FALSE;
	return TRUE;
}

static bool_t xdr_exports (XDR *xdrs, exports *objp)
{
	if (!xdr_pointer (xdrs, (char **)objp, sizeof (struct exportnode), (xdrproc_t) xdr_exportnode))
		return FALSE;
	return TRUE;
}


static int try_mount(char *spec, char *mount_point, int flags,
			char *fs_type, char **extra_opts, int fake, int bg)
{
	int ret;

	if (nfsmount_string_flg)
		ret = nfsmount_string(spec, mount_point, fs_type, flags,
					extra_opts, fake, bg);
	else {
		if (strcmp(fs_type, "nfs4") == 0)
			ret = nfs4mount(spec, mount_point, flags,
					extra_opts, fake, bg);
		else
			ret = nfsmount(spec, mount_point, flags,
					extra_opts, fake, bg);
	}

	if (ret)
		return ret;

	if (!fake)
		print_one(spec, mount_point, fs_type, mount_opts);

	ret = add_mtab(spec, mount_point, fs_type, flags, *extra_opts,
			0, 0 /* these are always zero for NFS */ );
	return ret;
}

#ifdef CONFIG_FEATURE_LONG_OPTIONS
static const struct option longopts[] = {
  { "fake", 0, 0, 'f' },
  { "help", 0, 0, 'h' },
  { "no-mtab", 0, 0, 'n' },
  { "read-only", 0, 0, 'r' },
  { "ro", 0, 0, 'r' },
  { "verbose", 0, 0, 'v' },
  { "version", 0, 0, 'V' },
  { "read-write", 0, 0, 'w' },
  { "rw", 0, 0, 'w' },
  { "options", 1, 0, 'o' },
  { NULL, 0, 0, 0 }
};
#endif

static void split_minus_o(const char *opt_arg)
{
	if(mount_opts == NULL) {
		mount_opts = xstrdup(opt_arg);
	} else {
		char *new_o = xasprintf("%s,%s", mount_opts, opt_arg);
		free(mount_opts);
		mount_opts = new_o;
	}
}

int mount_nfs_main(void)
{
	int c, flags = 0, mnt_err = 1, fake;
	char *spec, *mount_point, *fs_type = "nfs";
	char *extra_opts = NULL;
	uid_t uid = getuid();

	discover_nfs_mount_data_version();

	if(!strncmp(bb_applet_name, "umount", strlen("umount")))
		return umount_nfs_main();
	if(!strcmp(bb_applet_name, "statd"))
		return statd_main();
	if(!strcmp(bb_applet_name, "sm-notify"))
		return sm_notify_main();

#ifdef CONFIG_FEATURE_LONG_OPTIONS
	bb_applet_long_options = longopts;
#endif
	c = bb_xgetopt_ulflags("Vrwnsfo:hv\0=o:h-:=2:vv",
					split_minus_o, &verbose);
	if(c & 1) {
		printf("%s: ("PACKAGE_STRING")\n", bb_applet_name);
		return EX_SUCCESS;
	}
	if(c & 2)
		flags |= MS_RDONLY;
	if(c & 4)
		flags &= ~MS_RDONLY;
	/* NOMTAB = c & 8; */
	/* SLOPPY = c & 16; */
	fake = c & 32;
	spec = Gargv[0];
	mount_point = Gargv[1];
	if (strcmp(bb_applet_name, "mount.nfs4") == 0)
		fs_type = "nfs4";

	/* If a non-root user is attempting to mount, make sure the
	 * user's requested options match the options specified in
	 * /etc/fstab; otherwise, don't allow the mount. */
	if (uid != 0) {
		struct mntentchn *mc;

		if ((mc = get_fsfile(mount_point)) == NULL ||
		    strcmp(mc->m.mnt_fsname, spec) != 0 ||
		    strcmp(mc->m.mnt_type, fs_type) != 0) {
			bb_error_msg_and_die("permission denied: no match for %s "
				"found in /etc/fstab", mount_point);
		}

		/* 'mount' munges the options from fstab before passing them
		 * to us, so it is non-trivial to test that we have the correct
		 * set of options and we don't want to trust what the user
		 * gave us, so just take whatever is in /etc/fstab. */
		mount_opts = xstrdup(mc->m.mnt_opts);
	}

	mount_point = canonicalize(mount_point);
	if (!mount_point)
		bb_error_msg_and_die("no mount point provided");
	if (mount_point[0] != '/')
		bb_error_msg_and_die("unrecognized mount point %s", mount_point);

	parse_opts(mount_opts, &flags, &extra_opts);

	if (uid != 0) {
		if (!(flags & (MS_USERS|MS_USER)))
			bb_error_msg_and_die("permission denied");

		if (geteuid() != 0) {
			bb_error_msg("not installed setuid - "
				    "\"user\" NFS mounts not supported.");
			return EX_FAIL;
		}
	}

	if (chk_mountpoint(mount_point))
		return EX_USAGE;

	mnt_err = try_mount(spec, mount_point, flags, fs_type, &extra_opts,
				fake, FOREGROUND);
	if (mnt_err == EX_BG) {
		printf("%s: backgrounding \"%s\"\n", bb_applet_name, spec);
		printf("%s: mount options: \"%s\"\n", bb_applet_name, extra_opts);

		fflush(stdout);

		/* Parent exits immediately with success. */
		if (daemon(0, 0)) {
			bb_default_error_retval = EX_FAIL;
			bb_perror_msg_and_die("failed to start background process");
		}

		mnt_err = try_mount(spec, mount_point, flags, fs_type,
					&extra_opts, fake, BACKGROUND);
		if (verbose && mnt_err)
			printf("%s: giving up \"%s\"\n", bb_applet_name, spec);
	}
	return mnt_err;
}
#endif /* APPLET_USAGE */
